/*!
 * @file advcal_dlg.cpp
 * @brief This object allows the user to control how the black reference
 *        frame calibration proceeds.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "advcal_dlg.h"
#include "ui_advcal_dlg.h"

#include "camera_mdi.h"
#include "fi_sensor.h"
#include "histo_dock.h"
#include "logger.h"
#include "mainwindow.h"

namespace ADV_CAL_DIALOG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The valid Lupa sensor gain values.
const int LUPA_GAIN_SETTINGS[]      = { 100, 150, 200, 250, 300, 400 };

//! The number of Lupa gain settings in the LUPA_GAIN_SETTINGS array.
const int LUPA_GAIN_SETTINGS_COUNT  = (sizeof(LUPA_GAIN_SETTINGS) / sizeof(LUPA_GAIN_SETTINGS[0]));

//! The valid Lince sensor gain values.
const int LINCE_GAIN_SETTINGS[]     = { 100, 200, 400 };

//! The number of Lince gain settings in the LINCE_GAIN_SETTINGS array.
const int LINCE_GAIN_SETTINGS_COUNT = (sizeof(LINCE_GAIN_SETTINGS) / sizeof(LINCE_GAIN_SETTINGS[0]));

//! The valid Lux2100 sensor gain values.
const int LUX2100_GAIN_SETTINGS[]      = { 100, 150, 200, 250, 300, 350, 400, 500, 600, 700, 800, 900, 1000, 1100, 1200, 1300, 1400, 1500, 1600 };

//! The number of Lux2100 gain settings in the LUX2100_GAIN_SETTINGS array.
const int LUX2100_GAIN_SETTINGS_COUNT  = (sizeof(LUX2100_GAIN_SETTINGS) / sizeof(LUX2100_GAIN_SETTINGS[0]));

}   // End of namespace ADV_CAL_DIALOG.

using namespace ADV_CAL_DIALOG;


/*!
 * @brief The class constructor.
 */
AdvCalDlg::AdvCalDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::AdvCalDlg)
{
    FI_comm     *comm_ptr;

    // Setup our UI elements.
    ui->setupUi(this);

    // Initialize our controls.
    orig_sensor_dac = -1U;
    orig_sensor_gain = -1U;
    need_to_hide_histo = true;

    // Sanity check.
    if (MainWindow::GetInstance().isNull())
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null main window!";
        return;
    }

    // Get our camera object.
    camera_ptr = MainWindow::GetInstance()->active_mdi_child();
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No active camera!?";
        return;
    }

    // Sanity check.
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No sensor?";
        return;
    }

    // What sensor are we connecting to?
    switch (camera_ptr->get_sensor()->get_type())
    {
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO:
        // No offset controls for the Lince, so make sure they are not visible.
        offset_allowed = false;

        // Initialize our gain control(s).
        num_gain_settings = LINCE_GAIN_SETTINGS_COUNT;
        gain_settings_array = LINCE_GAIN_SETTINGS;
        break;
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO:
        // Make sure our offset controls are visible.
        offset_allowed = true;

        // Initialize our gain control(s).
        num_gain_settings = LUPA_GAIN_SETTINGS_COUNT;
        gain_settings_array = LUPA_GAIN_SETTINGS;
        break;
    case FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX2100_MONO:
        // No offset controls for the Lux2100, so make sure they are not visible.
        offset_allowed = false;

        // Initialize our gain control(s).
        num_gain_settings = LUX2100_GAIN_SETTINGS_COUNT;
        gain_settings_array = LUX2100_GAIN_SETTINGS;
        break;
    default:
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unknown sensor type?";
        return;
    }

    // Update our controls.
    ui->sld_gain->setRange(0, num_gain_settings - 1);
    ui->spin_gain->setRange(0, gain_settings_array[ num_gain_settings - 1 ]);
    ui->offset_frame->setVisible(offset_allowed);

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Connect our signals/slots.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &AdvCalDlg::read_complete, Qt::QueuedConnection);
    connect(ui->chk_auto_offset, &QCheckBox::clicked, this, &AdvCalDlg::auto_offset_change);
    connect(ui->sld_gain, &QSlider::valueChanged, this, &AdvCalDlg::handle_gain_change);
    connect(ui->sld_offset, &QSlider::valueChanged, this, &AdvCalDlg::handle_offset_change);
    connect(ui->spin_offset, QOverload<int>::of(&QSpinBox::valueChanged), this, &AdvCalDlg::handle_offset_change);

    connect(this, &AdvCalDlg::rejected, this, &AdvCalDlg::restore_settings);

    // Read the current sensor gain.
    camera_ptr->start_camera_read_reg(GIGE_REG_SENSOR_GAIN);
    if (offset_allowed)
        camera_ptr->start_camera_read_reg(GIGE_REG_SENSOR_DAC);

    // Update the layout.
    update();
}

/*!
 * @brief The class destructor.
 */
AdvCalDlg::~AdvCalDlg()
{
    delete ui;
}

/*!
 * @brief   Returns true if the user wants the camera to automatically adjust the offset/DAC
 *          setting. False if the user has adjusted the setting manually.
 */
bool AdvCalDlg::do_auto_offset(void)
{
    return ui->chk_auto_offset->isChecked();
}

/*!
 * @brief The user has modified the gain setting.
 * @param new_value     The new gain value.
 */
void AdvCalDlg::handle_gain_change(int new_value)
{
    double  display_value;

    // Sanity check.
    if (new_value >= num_gain_settings)
        new_value = num_gain_settings - 1;

    // Calculate our displayed value.
    display_value = ((double)gain_settings_array[ new_value ]) / 100.0;

    // Setup our controls.
    ui->spin_gain->setValue(display_value);

    // Send the new value to the camera.
    if (camera_ptr != nullptr)
        camera_ptr->write_camera_register(GIGE_REG_SENSOR_GAIN, (uint32_t)gain_settings_array[ new_value ]);
}

/*!
 * @brief The user has modified the offset/DAC setting.
 * @param new_value     The new offset/DAC value.
 */
void AdvCalDlg::handle_offset_change(int new_value)
{
    // Sanity check.
    if (!offset_allowed)
        return;

    // Update our control.
    if (ui->spin_offset->value() != new_value)
    {
        ui->spin_offset->blockSignals(true);
        ui->spin_offset->setValue(new_value);
        ui->spin_offset->blockSignals(false);
    }
    if (ui->sld_offset->value() != new_value)
    {
        ui->sld_offset->blockSignals(true);
        ui->sld_offset->setValue(new_value);
        ui->sld_offset->blockSignals(false);
    }

    // Send the new value to the camera.
    if (camera_ptr != nullptr)
        camera_ptr->write_camera_register(GIGE_REG_SENSOR_DAC, (uint32_t)new_value);
}

/*!
 * @brief Update our offset (sensor DAC) controls with the value we read
 *        from the camera.
 * @param dac_value     The offset/DAC value read from the camera.
 */
void AdvCalDlg::update_sensor_dac(uint32_t dac_value)
{
    int sensor_dac;

    // Sanity check.
    if (!offset_allowed)
        return;

    // Extract the value from the register.
    sensor_dac = (int)(dac_value & GIGE_REG_SENSOR_DAC_MASK);

    // Setup our controls.
    ui->sld_offset->blockSignals(true);
    ui->sld_offset->setValue(sensor_dac);
    ui->sld_offset->blockSignals(false);
    ui->spin_offset->blockSignals(true);
    ui->spin_offset->setValue(sensor_dac);
    ui->spin_offset->blockSignals(false);
}

/*!
 * @brief Update our gain controls with the value we read from the camera.
 * @param gain_value    The gain value read from the camera.
 */
void AdvCalDlg::update_sensor_gain(uint32_t gain_value)
{
    int     i;
    double  display_value;

    // Find the sensor gain index.
    for (i = 0; i < num_gain_settings; ++i)
    {
        // Is this a good choice?
        if (gain_value <= (uint32_t)gain_settings_array[i])
            break;
    }

    // Sanity check.
    if (i == num_gain_settings)
        i = num_gain_settings - 1;

    // Calculate our displayed value.
    display_value = ((double)gain_settings_array[i]) / 100.0;

    // Setup our controls.
    ui->sld_gain->setValue(i);
    ui->spin_gain->setValue(display_value);

    // Is the offset information supported?
    if (ui->offset_frame->isVisible())
    {
        // By default, assume automatic offset calibration.
        ui->chk_auto_offset->setChecked(true);
        auto_offset_change();
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has set or cleared the auto checkbox.
 */
void AdvCalDlg::auto_offset_change(void)
{
    ui->lbl_offset->setEnabled(!ui->chk_auto_offset->isChecked());
    ui->sld_offset->setEnabled(!ui->chk_auto_offset->isChecked());
    ui->spin_offset->setEnabled(!ui->chk_auto_offset->isChecked());
}

/*!
 * @brief The dialog is finished. Restore the histogram as it was before we got involved.
 * @param result        The dialog result code.
 */
void AdvCalDlg::done(int result)
{
    HistoDock   *histo_dock;

    // Sanity check.
    if (MainWindow::GetInstance().isNull())
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No main window!?";
        goto done_bail;
    }

    // Get a pointer to the histogram dock widget.
    histo_dock = MainWindow::GetInstance()->get_histo_dock();

    // Sanity check.
    if (histo_dock == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No histodock!?";
        goto done_bail;
    }

    // Do we need to hide the histogram?
    if (need_to_hide_histo)
        histo_dock->setVisible(false);

    // Tell the histogram to use processed samples.
    histo_dock->set_raw_mode(false);

done_bail:
    // Pass it on.
    QDialog::done(result);
}

/*!
 * @brief Bring up the histogram view before we show the dialog.
 * @return The result of QDialog::exec().
 */
int AdvCalDlg::exec()
{
    bool raw_histo_avail(false);

    // Sanity check.
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No sensor?";
        return QDialog::Rejected;
    }

    // What sensor are we connecting to?
    switch (camera_ptr->get_sensor()->get_type())
    {
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO:
        // Raw histogram is available.
        raw_histo_avail = true;
        break;
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO:
    case FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX2100_MONO:
    default:
        break;
    }

    // Only put up the histogram if we have a sensor that supports
    // raw mode.
    if (raw_histo_avail)
    {
        HistoDock   *histo_dock;

        // Sanity check.
        if (MainWindow::GetInstance().isNull())
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "No main window!?";
            return QDialog::Rejected;
        }

        // Get a pointer to the histogram dock widget.
        histo_dock = MainWindow::GetInstance()->get_histo_dock();

        // Sanity check.
        if (histo_dock == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "No histodock!?";
            return QDialog::Rejected;
        }

        // Tell the histogram to use raw samples.
        histo_dock->set_raw_mode(true);

        // Is the histogram visible?
        if (!histo_dock->isVisible())
        {
            // Make the histogram visible.
            need_to_hide_histo = true;
            histo_dock->setVisible(true);
        }
        else
        {
            // It's already visible, so we'll just keep it that way.
            need_to_hide_histo = false;
        }
    }
    else
    {
        // Leave the histogram alone.
        need_to_hide_histo = false;
    }

    // Carry on.
    return QDialog::exec();
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 * @param ip_addr       IP address of the camera associated with the read operation.
 * @param reg_addr      Register address which was read.
 */
void AdvCalDlg::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;

    // Make sure this is our camera.
    if (camera_ptr == nullptr)
        return;
    if (camera_ptr->get_ip_addr() != ip_addr)
        return;

    // Get the value.
    reg_value = camera_ptr->read_camera_cache(reg_addr);

    // Only interested in the following registers.
    switch (reg_addr)
    {
    case GIGE_REG_SENSOR_GAIN:
        // Is this the first read? If so, we want to save this value in case of a cancel.
        // Ignore any others.
        if (orig_sensor_gain == (uint32_t) -1)
        {
            // Save this value.
            orig_sensor_gain = reg_value;

            // Process the register setting.
            update_sensor_gain(reg_value);
        }
        break;
    case GIGE_REG_SENSOR_DAC:
        // Is this the first read? If so, we want to save this value in case of a cancel.
        // Ignore any others.
        if (offset_allowed && (orig_sensor_dac == (uint32_t) -1))
        {
            // Save this value.
            orig_sensor_dac = reg_value;

            // Process the register setting.
            update_sensor_dac(reg_value);
        }
        break;
    default:
        break;
    }
}

void AdvCalDlg::restore_settings(void)
{
    // Sanity check.
    if (camera_ptr == nullptr)
        return;

    // Restore our settings.
    if (offset_allowed && (orig_sensor_dac != (uint32_t) -1))
        camera_ptr->write_camera_register(GIGE_REG_SENSOR_DAC, orig_sensor_dac);
    if (orig_sensor_gain != (uint32_t) -1)
        camera_ptr->write_camera_register(GIGE_REG_SENSOR_GAIN, orig_sensor_gain);
}

