/*!
 * @file calibrate_dlg.cpp
 * @brief This object allows the user to control how the black reference
 *        frame calibration proceeds.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QString>
#include <QStringLiteral>

#include "calibrate_dlg.h"
#include "ui_calibrate_dlg.h"

#include "advcal_dlg.h"
#include "camera_mdi.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"

CalibrateDlg::CalibrateDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::CalibrateDlg)
{
    CameraMDI   *mdi_parent;
    const QString s3_label = tr("<html><head/><body><p>Please confirm that you have lens cap on.</p><p><span style=\" font-weight:600;\">NOTE: This will require the connection to be closed and reopened. Any images in memory will be lost.</span></p><p>Do you want to proceed?</p></body></html>");
    const QString sN_label = tr("<html><head/><body><p><span style=\" font-weight:600;\">NOTE: This will require the connection to be closed and reopened. Any images in memory will be lost.</span></p><p>Do you want to proceed?</p></body></html>");

    // Initialize the UI object.
    ui->setupUi(this);

    // By default, we want to automatically adjust the offset.
    auto_cal_offset = true;

    // Make sure we have a parent.
    mdi_parent = qobject_cast<CameraMDI *>(parent);
    if (mdi_parent == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Parent not a CameraMDI!";
        return;
    }

    // Are we series3?
    if (mdi_parent->is_series3())
        ui->lbl_info->setText(s3_label);
    else
        ui->lbl_info->setText(sN_label);

    // Does our sensor support any advanced calibration?
    if ((mdi_parent->get_sensor() != nullptr) && mdi_parent->get_sensor()->supports_adv_cal())
    {
        // Show the advanced calibration feature(s).
        ui->pb_advanced->setVisible(true);

        // Connect up our signals/slots.
        connect(ui->pb_advanced, SIGNAL(clicked()), this, SLOT(do_advanced_cal()));
    }
    else
    {
        // Hide the advanced calibration feature(s).
        ui->pb_advanced->setVisible(false);
    }
}

CalibrateDlg::~CalibrateDlg()
{
    delete ui;
}

void CalibrateDlg::set_info_text(const QString& info_str)
{
    ui->lbl_info->setText(info_str);
    update();
}

void CalibrateDlg::set_title(const QString& title_str)
{
    setWindowTitle(title_str);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void CalibrateDlg::do_advanced_cal()
{
    AdvCalDlg   adv_dlg(this);

    // Run the advance dialog. If the user clicks OK, then that means we're
    // good to run the calibration.
    if (adv_dlg.exec() == QDialog::Accepted)
    {
        // Remember if the user wanted to do automatic calibration of the offset.
        auto_cal_offset = adv_dlg.do_auto_offset();

        // Carry on.
        accept();
    }
}

