/*!
 * @file cam_update_dlg.cpp
 * @brief This is the implementation of the CamUpdateDlg dialog.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QByteArray>
#include <QDir>
#include <QFile>
#include <QFileDialog>
#include <QMessageBox>
#include <QVersionNumber>

#include "cam_update_dlg.h"
#include "ui_cam_update_dlg.h"

#include "camera_mdi.h"
#include "fi_comm.h"
#include "fi_updater.h"
#include "logger.h"
#include "mainwindow.h"

CamUpdateDlg::CamUpdateDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::CamUpdateDlg)
{
    CameraMDI *my_parent = qobject_cast<CameraMDI *>(parent);

    // Make sure we got our parent.
    if (my_parent == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need CameraMDI as parent!";
        return;
    }

    // Setup the UI elements.
    ui->setupUi(this);

    // Connect slots/signals.
    connect(ui->pb_cancel, &QPushButton::clicked, this, &CamUpdateDlg::reject);
    connect(ui->pb_update, &QPushButton::clicked, this, &CamUpdateDlg::startUpdate);

    // If the parent is closing, we need to bail.
    connect(my_parent, &CameraMDI::closing, this, &CamUpdateDlg::reject);

    // Set our icon.
    ui->lbl_icon->setPixmap(this->style()->standardPixmap(QStyle::SP_MessageBoxInformation));

    // Display the confirmation message.
    if (my_parent->is_series3())
    {
#ifdef Q_OS_WIN
        // Get our update object.
        const FiUpdater *updater;
        updater = MainWindow::GetInstance()->get_updater();

        // Series3 cameras require a built SDCard.
        ui->pb_build->setVisible(true);
        ui->pb_build->setEnabled(QFile::exists(updater->GetPackage(FI_UPDATER::AVAIL_CAM_S3_UPDATE)));
        connect(ui->pb_build, &QPushButton::clicked, this, &CamUpdateDlg::doBuildSDCard);
#endif

        // Series3 cameras don't get to specify where the package is.
        ui->update_location->setVisible(false);

        // The confirmation is specific to a series3 camera.
        ui->lbl_confirmation->setText(tr("Are you sure you want to update the camera?\n\nPlease make sure the update media is in the SDCARD slot. Communication will be lost and will need to be reestablished once the update is complete."));
    }
    else
    {
#ifdef ALLOW_DEVICE_UPDATES
        // Connect the communications interfaces that we're interested in.
        connect(my_parent, &CameraMDI::media_info_ready, this, &CamUpdateDlg::handleJsonMedia, Qt::QueuedConnection);

        // Ask the camera for the available media. This will be caught
        // by the camera object, which will start the ball rolling.
        my_parent->start_camera_read_reg(GIGE_REG_AVAIL_MEDIA);

        // SeriesNC cameras get to specify where the package is.
        ui->update_location->setVisible(true);

        // Default to using the network.
        connect(ui->rb_device, &QRadioButton::toggled, this, &CamUpdateDlg::locationChange);
        connect(ui->rb_network, &QRadioButton::toggled, this, &CamUpdateDlg::locationChange);
        ui->rb_network->setChecked(true);
#else
        // For now, we can't update from a device.
        ui->rb_device->setEnabled(false);
        ui->rb_network->setChecked(true);
        ui->update_location->setVisible(false);
#endif  /* ALLOW_DEVICE_UPDATES */

        // SeriesNC cameras do not use an SDCard.
        ui->pb_build->setVisible(false);

        // The confirmation is specific to a seriesNC camera.
        ui->lbl_confirmation->setText(tr("Are you sure you want to update the camera?\n\nCommunication will be lost and will need to be reestablished once the update is complete."));
    }

    // Update our dialog.
    adjustSize();
    update();
}

CamUpdateDlg::~CamUpdateDlg()
{
    delete ui;
}

/*!
 * @brief Return the path for the update package, if any.
 */
const QString CamUpdateDlg::GetPackagePath() const
{
    QJsonObject dev_object;

    // If the network radiobutton is selected, then there's no path.
    if (ui->rb_network->isChecked())
        return QString();

    // Get the selected object.
    dev_object = ui->cb_devices->currentData().toJsonObject();

    // Return the mount path (this is where we want to write to).
    return dev_object["mountdir"].toString();
}

/*!
 * @brief Return the device selected by the user for the update package.
 * @note This is only valid for newer cameras.
 */
const QString CamUpdateDlg::GetUpdateDevice() const
{
    return ui->cb_devices->currentText();
}

/*!
 * @brief Return the update version as an "X.Y.Z" string.
 * @note This is only valid for newer cameras.
 */
const QString CamUpdateDlg::GetUpdateVersion() const
{
    const FiUpdater *updater;

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    // Return the seriesNC version.
    return updater->GetVersionCamNC().toString();
}

/*!
 * @brief Return true if we're doing a network update, false for a device update.
 */
bool CamUpdateDlg::DoNetworkUpdate()
{
    return ui->rb_network->isChecked();
}

/*!
 * @brief Utility function for building the update SDCard.
 */
void CamUpdateDlg::BuildSDCard()
{
    const FiUpdater *updater;

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    // Make sure the file exists.
    if (QFile::exists(updater->GetPackage(FI_UPDATER::AVAIL_CAM_S3_UPDATE)))
    {
        // Do the update!
        QProcess::startDetached(updater->GetPackage(FI_UPDATER::AVAIL_CAM_S3_UPDATE), QStringList());
    }
}

/*!
 * @brief The user has pressed the button to build the update SDCard.
 */
void CamUpdateDlg::doBuildSDCard()
{
    // Build the SDCard.
    CamUpdateDlg::BuildSDCard();

    // Close the dialog with a reject so we don't carry on as though we're doing an update.
    reject();
}

/*!
 * @brief The user changed where to update the cameras from.
 */
void CamUpdateDlg::locationChange()
{
    // The combobox is only enabled if the devices is selected.
    ui->cb_devices->setEnabled(ui->rb_device->isChecked());
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_AVAIL_MEDIA register value.
 */
void CamUpdateDlg::handleJsonMedia(const QJsonDocument& jsonMedia)
{
    int item_index;
    double dev_size;
    QString base_label;
    QString device_label;
    QString device_sel;
    QJsonObject mediaObj;

    // Get the object from the document.
    mediaObj = jsonMedia.object();

    // Block updates from the combobox.
    ui->cb_devices->blockSignals(true);

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
        device_sel = ui->cb_devices->currentText();

    // Clear the combobox.
    ui->cb_devices->clear();

    // Add back in found devices starting with the SDCard.
    if (mediaObj.contains("sdcard") && mediaObj["sdcard"].isArray())
    {
        QJsonArray dev_array = mediaObj["sdcard"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an sdcard device.
            dev_object["type"] = GIGE_MEDIA_SDCARD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = tr("SDCard %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" %1GB").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // Check for NVME SSD devices.
    if (mediaObj.contains("nvme") && mediaObj["nvme"].isArray())
    {
        QJsonArray dev_array = mediaObj["nvme"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Don't select the internal drive.
            if (dev_object["internal"].toBool())
                continue;

            // Remember this is an NVME device.
            dev_object["type"] = GIGE_MEDIA_SSD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                if (dev_object["internal"].toBool())
                    device_label = tr("SSD %1GB [Internal]").arg(dev_size, 0, 'f', 0);
                else
                    device_label = tr("SSD %1GB [External]").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" %1GB [External]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // Check for USB drives.
    if (mediaObj.contains("usb") && mediaObj["usb"].isArray())
    {
        QJsonArray dev_array = mediaObj["usb"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an USB device.
            dev_object["type"] = GIGE_MEDIA_USB;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = tr("USB %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // If there is no external media then disable the devices controls.
    if (ui->cb_devices->count() == 0)
    {
        // Make sure we're disabled.
        ui->rb_device->setEnabled(false);
        ui->rb_network->setChecked(true);
    }
    else
    {
        int sel_index = 0;

        // Make sure we're enabled.
        ui->rb_device->setEnabled(true);

        // Attempt to re-select the old value.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);
            if (sel_index < 0)
                sel_index = 0;
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);
    }

    // Allow signals to flow again.
    ui->cb_devices->blockSignals(false);
}


void CamUpdateDlg::startUpdate()
{
    // Our job is done here.
    accept();
}

