/*!
 * @file camera_actions.cpp
 * @brief This file contains the implementation of the CameraActions
 *        class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QAction>

#include "camera_actions.h"
#include "ui_camera_actions.h"

#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The "name" for the ARM action.
const QString CameraActions::ARM_ACTION_NAME     = QStringLiteral("action_arm");

//! The "name" for the Autosave Enable action.
const QString CameraActions::AUTOSAVE_ENABLE_ACTION_NAME = QStringLiteral("action_autosave_enable");

//! The "name" for the Trigger action.
const QString CameraActions::TRIGGER_ACTION_NAME = QStringLiteral("action_trigger");

//! The "name" for the Live Mode action.
const QString CameraActions::LIVE_ACTION_NAME    = QStringLiteral("action_live_mode");

//! The "name" for the Review Mode action.
const QString CameraActions::REVIEW_ACTION_NAME  = QStringLiteral("action_review_mode");

/*!
 * @brief Constructor for the CameraActions object.
 * @param parent        Pointer to the parent widget.
 */
CameraActions::CameraActions(QWidget *parent) :
    QFrame(parent),
    ui(new Ui::CameraActions)
{
    // Setup our UI.
    ui->setupUi(this);

    // Initialize our variables.
    action_arm = nullptr;
    action_autosave_enable = nullptr;
    action_live = nullptr;
    action_review = nullptr;
    action_trigger = nullptr;
    multicam_mode = false;

    // We need to wait for the main window to be created before we can continue.
    QTimer::singleShot(10, this, &CameraActions::delayed_init);
}

/*!
 * @brief Destructor for CameraActions object.
 */
CameraActions::~CameraActions()
{
    delete ui;
}

void CameraActions::delayed_init()
{
    int     i;
    QList<QAction *>    main_actions;

    // Get the actions for the Camera menu. In some cases, the main window may not exist just
    // yet. We'll skip any action processing in that case.
    if (MainWindow::GetInstance() != nullptr)
        main_actions = MainWindow::GetInstance()->get_camera_actions();
    else
        main_actions.clear();

    // Parse through them looking for the actions associated with our controls.
    for (i = 0; i < main_actions.count(); ++i)
    {
        QString action_name = main_actions.at(i)->objectName();

        // See if we have a match.
        if (action_name.compare(ARM_ACTION_NAME) == 0)
        {
            // If we have a match, copy the resulting action over to our list.
            addAction(main_actions.at(i));

            // Setup the connection for this action.
            connect(ui->tb_arm, &QToolButton::clicked, main_actions.at(i), &QAction::triggered, Qt::UniqueConnection);
        }
        else if (action_name.compare(AUTOSAVE_ENABLE_ACTION_NAME) == 0)
        {
            // If we have a match, copy the resulting action over to our list.
            addAction(main_actions.at(i));

            // Setup the connection for this action.
            connect(ui->tb_autosave, &QToolButton::clicked, main_actions.at(i), &QAction::triggered, Qt::UniqueConnection);
        }
        else if (action_name.compare(TRIGGER_ACTION_NAME) == 0)
        {
            // If we have a match, copy the resulting action over to our list.
            addAction(main_actions.at(i));

            // Setup the connection for this action.
            connect(ui->tb_trigger, &QToolButton::clicked, main_actions.at(i), &QAction::triggered, Qt::UniqueConnection);
        }
        else if (action_name.compare(LIVE_ACTION_NAME) == 0)
        {
            // If we have a match, copy the resulting action over to our list.
            addAction(main_actions.at(i));

            // Setup the connection for this action.
            connect(ui->tb_live, &QToolButton::clicked, main_actions.at(i), &QAction::triggered, Qt::UniqueConnection);
        }
        else if (action_name.compare(REVIEW_ACTION_NAME) == 0)
        {
            // If we have a match, copy the resulting action over to our list.
            addAction(main_actions.at(i));

            // Setup the connection for this action.
            connect(ui->tb_review, &QToolButton::clicked, main_actions.at(i), &QAction::triggered, Qt::UniqueConnection);
        }
    }
}

/*!
 * @brief Disable our action buttons.
 */
void CameraActions::disable_actions()
{
    int i;
    QList<QAction *>    cam_actions;

    // Get our actions.
    cam_actions = actions();

    // Loop through the list and disable all the actions.
    for (i = 0; i < cam_actions.count(); ++i)
        cam_actions.at(i)->setEnabled(false);

    // Disable alternate actions too.
    if (action_arm != nullptr)
        action_arm->setEnabled(false);
    if (action_autosave_enable != nullptr)
        action_autosave_enable->setEnabled(false);
    if (action_live != nullptr)
        action_live->setEnabled(false);
    if (action_review != nullptr)
        action_review->setEnabled(false);
    if (action_trigger != nullptr)
        action_trigger->setEnabled(false);

    // Disable the panel too.
    setEnabled(false);
}

/*!
 * @brief Enable our action buttons.
 */
void CameraActions::enable_actions()
{
    int i;
    QList<QAction *>    cam_actions;

    // Enable the panel.
    setEnabled(true);

    // Get our actions.
    cam_actions = actions();

    // Loop through the list and restore our actions.
    for (i = 0; i < cam_actions.count(); ++i)
    {
        QString action_name = cam_actions.at(i)->objectName();

        // See if we have a match.
        if (action_name.compare(ARM_ACTION_NAME) == 0)
        {
            // Arm the action if the button is armed.
            cam_actions.at(i)->setEnabled(ui->tb_arm->isEnabled());
        }
        else if (action_name.compare(AUTOSAVE_ENABLE_ACTION_NAME) == 0)
        {
            // Arm the action if the button is armed.
            cam_actions.at(i)->setEnabled(ui->tb_autosave->isEnabled());
        }
        else if (action_name.compare(TRIGGER_ACTION_NAME) == 0)
        {
            // Arm the action if the button is armed.
            cam_actions.at(i)->setEnabled(ui->tb_trigger->isEnabled());
        }
        else if (action_name.compare(LIVE_ACTION_NAME) == 0)
        {
            // Arm the action if the button is armed.
            cam_actions.at(i)->setEnabled(ui->tb_live->isEnabled());
        }
        else if (action_name.compare(REVIEW_ACTION_NAME) == 0)
        {
            // Arm the action if the button is armed.
            cam_actions.at(i)->setEnabled(ui->tb_review->isEnabled());
        }
    }

    // Restore alternate actions.
    if (action_arm != nullptr)
        action_arm->setEnabled(ui->tb_arm->isEnabled());
    if (action_autosave_enable != nullptr)
        action_autosave_enable->setEnabled(ui->tb_autosave->isEnabled());
    if (action_live != nullptr)
        action_live->setEnabled(ui->tb_live->isEnabled());
    if (action_review != nullptr)
        action_review->setEnabled(ui->tb_review->isEnabled());
    if (action_trigger != nullptr)
        action_trigger->setEnabled(ui->tb_trigger->isEnabled());
}

/*!
 * @brief Enable special process used by multicamera mode.
 */
void CameraActions::enable_multicam_mode()
{
    // Entering into multicamera mode, we want to disable all the controls.
    ui->tb_arm->setEnabled(false);
    ui->tb_autosave->setEnabled(false);
    ui->tb_live->setEnabled(false);
    ui->tb_trigger->setEnabled(false);
    ui->tb_review->setEnabled(false);

    // Remember that we're in multicamera mode.
    multicam_mode = true;

    // Disconnect all our signal/slot connections.
    ui->tb_arm->disconnect();
    ui->tb_autosave->disconnect();
    ui->tb_live->disconnect();
    ui->tb_review->disconnect();
    ui->tb_trigger->disconnect();

    // Clear out all the actions.
    QList<QAction *> old_actions = actions();
    for (int i = 0; i < old_actions.count(); ++i)
    {
        removeAction(old_actions.at(i));
    }
}

/*!
 * @brief In multicamera mode, updating the controls is a multiple pass
 *        operation. Each pass looks at the state of a particular camera.
 *        This is the final step in the multicamera update process.
 */
void CameraActions::finish_update()
{
    // Catch the case when there are no cameras. All the actions can't be enabled
    // at the same time.
    if (enable_arm && enable_live && enable_review && enable_trigger)
    {
        // Reset all our flags.
        enable_arm = false;
        enable_autosave = false;
        enable_live = false;
        enable_review = false;
        enable_trigger = false;
        use_dearm_icon = false;
        use_detrig_icon = false;
    }

    // Setup our controls.
    ui->tb_arm->setEnabled(enable_arm);
    ui->tb_autosave->setEnabled(enable_autosave);
    ui->tb_live->setEnabled(enable_live);
    ui->tb_review->setEnabled(enable_review);
    ui->tb_trigger->setEnabled(enable_trigger);

    // Initialize our actions.
    if (action_arm != nullptr)
        action_arm->setEnabled(enable_arm);
    if (action_autosave_enable != nullptr)
        action_autosave_enable->setEnabled(enable_autosave);
    if (action_live != nullptr)
        action_live->setEnabled(enable_live);
    if (action_review != nullptr)
        action_review->setEnabled(enable_review);
    if (action_trigger != nullptr)
        action_trigger->setEnabled(enable_trigger);

    // Which Arm icon should we use?
    if (use_dearm_icon)
        ui->tb_arm->setIcon(QIcon(QString::fromUtf8(":/images/dearmbtn.png")));
    else if (use_detrig_icon)
        ui->tb_arm->setIcon(QIcon(QString::fromUtf8(":/images/detrigbtn.png")));
    else
        ui->tb_arm->setIcon(QIcon(QString::fromUtf8(":/images/armbtn.png")));
}

/*!
 * @brief Update the user interface to reflect the current camera state.
 *
 * This is done independently from the main window's similar operation because this
 * camera may not be the active camera. The main window actions have to reflect the
 * active camera state.
 */
void CameraActions::reflect_camera_state(uint32_t camera_state, uint32_t part_state, uint32_t ltr_mode)
{
    bool arm_enable;
    bool autosave_enable;
    bool live_enable;
    bool review_enable;
    bool trig_enable;
    bool do_dearm;
    bool do_detrig;

    // Update as required.
    switch (camera_state)
    {
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "What?! Wrong camera state:" << Qt::showbase << Qt::hex << camera_state;

        //
        // Fall through intentional...
        //
        [[fallthrough]];

    case GIGE_CAMERA_STATE_IDLE:
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_SNAPSHOT:
    case GIGE_CAMERA_STATE_SHUTDOWN:
        arm_enable = false;
        live_enable = false;
        review_enable = false;
        trig_enable = false;
        autosave_enable = false;
        do_dearm = false;
        do_detrig = false;
        break;
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        arm_enable = ((part_state == GIGE_PART_STATE_COMPLETE) || (part_state == GIGE_PART_STATE_EMPTY));
        live_enable = false;
        review_enable = (part_state == GIGE_PART_STATE_COMPLETE);
        trig_enable = true;
        autosave_enable = true;
        do_dearm = false;
        do_detrig = false;
        break;
    case GIGE_CAMERA_STATE_ARMED:
        arm_enable = true;
        live_enable = false;
        review_enable = false;
        trig_enable = true;
        autosave_enable = false;
        do_dearm = true;
        do_detrig = false;
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        arm_enable = true;
        live_enable = false;
        review_enable = false;
        if (((ltr_mode & GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE) ||
            ((ltr_mode & GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR))
            trig_enable = ((ltr_mode & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0);
        else
            trig_enable = false;
        autosave_enable = false;
        do_dearm = false;
        do_detrig = true;
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        arm_enable = true;
        live_enable = false;
        review_enable = false;
        trig_enable = false;
        autosave_enable = false;
        do_dearm = false;
        do_detrig = false;
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        arm_enable = true;
        live_enable = true;
        review_enable = false;
        trig_enable = false;
        autosave_enable = false;
        do_dearm = false;
        do_detrig = false;
        break;
    }

    // Are we in multicamera mode?
    if (multicam_mode)
    {
        if (enable_arm)
            enable_arm = arm_enable;
        if (enable_autosave)
            enable_autosave = autosave_enable;
        if (enable_live)
            enable_live = live_enable;
        if (enable_review)
            enable_review = review_enable;
        if (enable_trigger)
            enable_trigger = trig_enable;
        if (!use_dearm_icon)
            use_dearm_icon = do_dearm;
        if (!use_detrig_icon)
            use_detrig_icon = do_detrig;
    }
    else
    {
        int     i;
        QList<QAction *>    cam_actions;

        // For autosave, reflect what is in the main window menu.
        autosave_enable = MainWindow::GetInstance()->get_cam_autosave_enable()->isEnabled();

        // Get our actions.
        cam_actions = actions();

        // Loop through the list and restore our actions.
        for (i = 0; i < cam_actions.count(); ++i)
        {
            QString action_name = cam_actions.at(i)->objectName();

            // See if we have a match.
            if (action_name.compare(ARM_ACTION_NAME) == 0)
            {
                // Arm the action if the button is armed.
                cam_actions.at(i)->setEnabled(arm_enable);
            }
            else if (action_name.compare(AUTOSAVE_ENABLE_ACTION_NAME) == 0)
            {
                // Arm the action if the button is armed.
                cam_actions.at(i)->setEnabled(autosave_enable);
            }
            else if (action_name.compare(TRIGGER_ACTION_NAME) == 0)
            {
                // Arm the action if the button is armed.
                cam_actions.at(i)->setEnabled(trig_enable);
            }
            else if (action_name.compare(LIVE_ACTION_NAME) == 0)
            {
                // Arm the action if the button is armed.
                cam_actions.at(i)->setEnabled(live_enable);
            }
            else if (action_name.compare(REVIEW_ACTION_NAME) == 0)
            {
                // Arm the action if the button is armed.
                cam_actions.at(i)->setEnabled(review_enable);
            }
        }

        // Setup our controls.
        ui->tb_arm->setEnabled(arm_enable);
        ui->tb_autosave->setEnabled(autosave_enable);
        ui->tb_live->setEnabled(live_enable);
        ui->tb_review->setEnabled(review_enable);
        ui->tb_trigger->setEnabled(trig_enable);

        // Initialize our actions.
        if (action_arm != nullptr)
            action_arm->setEnabled(arm_enable);
        if (action_autosave_enable != nullptr)
            action_autosave_enable->setEnabled(autosave_enable);
        if (action_live != nullptr)
            action_live->setEnabled(live_enable);
        if (action_review != nullptr)
            action_review->setEnabled(review_enable);
        if (action_trigger != nullptr)
            action_trigger->setEnabled(trig_enable);

        // Which Arm icon should we use?
        if (do_dearm)
            ui->tb_arm->setIcon(QIcon(QString::fromUtf8(":/images/dearmbtn.png")));
        else if (do_detrig)
            ui->tb_arm->setIcon(QIcon(QString::fromUtf8(":/images/detrigbtn.png")));
        else
            ui->tb_arm->setIcon(QIcon(QString::fromUtf8(":/images/armbtn.png")));
    }
}

/*!
 * @brief Set the action to associate with the ARM button.
 * @param action_ptr    The ARM action.
 */
void CameraActions::set_arm_action(QAction *action_ptr)
{
    // Connect up the new ARM action.
    action_arm = action_ptr;
    addAction(action_ptr);
    ui->tb_arm->setStatusTip(action_ptr->statusTip());
    connect(ui->tb_arm, &QToolButton::clicked, action_ptr, &QAction::triggered, Qt::UniqueConnection);
}

void CameraActions::set_autosave_checked(bool checked)
{
    ui->tb_autosave->setChecked(checked);
}

/*!
 * @brief Update the checked status of the autosave enable toolbutton.
 * @param enabled   True to check, false to uncheck.
 */
void CameraActions::set_autosave_enabled(bool enabled)
{
    enable_autosave = enabled;
}

/*!
 * @brief Set the action to associate with the Autosave Enable button.
 * @param action_ptr    The Autosave Enable action.
 */
void CameraActions::set_autosave_action(QAction *action_ptr)
{
    action_autosave_enable = action_ptr;
    addAction(action_ptr);
    ui->tb_autosave->setStatusTip(action_ptr->statusTip());
    connect(ui->tb_autosave, &QToolButton::clicked, action_ptr, &QAction::triggered, Qt::UniqueConnection);
}

/*!
 * @brief Set the action to associate with the Live button.
 * @param action_ptr    The Live action.
 */
void CameraActions::set_live_action(QAction *action_ptr)
{
    action_live = action_ptr;
    addAction(action_ptr);
    ui->tb_live->setStatusTip(action_ptr->statusTip());
    connect(ui->tb_live, &QToolButton::clicked, action_ptr, &QAction::triggered, Qt::UniqueConnection);
}

/*!
 * @brief Set the action to associate with the Review button.
 * @param action_ptr    The Review action.
 */
void CameraActions::set_review_action(QAction *action_ptr)
{
    action_review = action_ptr;
    addAction(action_ptr);
    ui->tb_review->setStatusTip(action_ptr->statusTip());
    connect(ui->tb_review, &QToolButton::clicked, action_ptr, &QAction::triggered, Qt::UniqueConnection);
}

/*!
 * @brief Set the action to associate with the Trigger button.
 * @param action_ptr    The Trigger action.
 */
void CameraActions::set_trigger_action(QAction *action_ptr)
{
    action_trigger = action_ptr;
    addAction(action_ptr);
    ui->tb_trigger->setStatusTip(action_ptr->statusTip());
    connect(ui->tb_trigger, &QToolButton::clicked, action_ptr, &QAction::triggered, Qt::UniqueConnection);
}

/*!
 * @brief This is the entry into the multiple pass update process for
 *        multicamera updates.
 */
void CameraActions::start_update()
{
    // Initialize our variables to start an update in multicamera mode.
    enable_arm = true;
    enable_autosave = true;
    enable_live = true;
    enable_review = true;
    enable_trigger = true;

    use_dearm_icon = false;
    use_detrig_icon = false;
}

