/*!
 * @file camera_dock.cpp
 * @brief Displays all found cameras.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "camera_dock.h"
#include "ui_camera_dock.h"

#include <QtEndian>

#include <QContextMenuEvent>
#include <QDesktopServices>
#include <QHostInfo>
#include <QInputDialog>
#include <QListView>
#include <QMdiArea>
#include <QMdiSubWindow>
#include <QMenu>
#include <QMessageBox>
#include <QtAlgorithms>
#include <QtNetwork>
#include <QScrollBar>
#include <QStandardItem>
#include <QStandardItemModel>
#include <QString>
#include <QTableView>
#include <QTimer>

#include "camera_mdi.h"
#include "fcp_socket.h"
#include "force_ip.h"
#include "logger.h"
#include "mainwindow.h"
#include "multicam_dock.h"
#include "show_columns.h"
#include "user_preferences.h"

// Avoid polluting the global namespace.
namespace CAMERA_DOCK
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macro Definitions
//--------------------------------------------------------------------
#define LOG_DISCOVERY_POLLING           0

//! Amount of time to wait before opening a camera.
const int   OPEN_DELAY                  = 500;

//! Our interval for updating scan results, in milliseconds.
const int   SCAN_UPDATE_INTERVAL        = 250;

//! The number of timer events to wait for responses from cameras.
const int   WAIT_FOR_RESPONSE_TIMEOUT   = (2000 / SCAN_UPDATE_INTERVAL);

//! The maximum number of scan cycles a camera can ignore us before we declare it missing.
const int   MAX_MISSING_COUNT           = (10 * WAIT_FOR_RESPONSE_TIMEOUT);

}   // End of namespace CAMERA_DOCK.

//--------------------------------------------------------------------
// Local Variable Declarations
//--------------------------------------------------------------------

// Context menu actions
QMenu           *CameraDock::arrange_menu = nullptr;
QList<QAction *> CameraDock::arrange_act;
QActionGroup    *CameraDock::arrange_group = nullptr;

//! Common place for the QSettings strings associated with the camera dock window.
const QString CameraDock::CAMERA_DOCK_GROUP_NAME            = QStringLiteral("CameraDock");

//! Camera Dock network interface key string.
const QString CameraDock::CAMERA_DOCK_NET_IF                = QStringLiteral("net_if");

//! Camera Dock sorted column key string.
const QString CameraDock::CAMERA_DOCK_SORT_COL              = QStringLiteral("sorted_col");

//! Camera Dock sorted order key string.
const QString CameraDock::CAMERA_DOCK_SORT_ORDER            = QStringLiteral("sort_order");

//! Camera Dock stream key string.
const QString CameraDock::CAMERA_DOCK_TAB_DETAIL_STATE      = QStringLiteral("detail_state");

//! Camera Dock stream key string.
const QString CameraDock::CAMERA_DOCK_STREAM_ON_CONNECT     = QStringLiteral("stream");

// Make sure we use our namespace.
using namespace CAMERA_DOCK;

/*!
 * @brief CameraDock class constructor.
 */
CameraDock::CameraDock(QWidget *parent) :
    QDockWidget(parent),
    ui(new Ui::CameraDock)
{
    // Initialize the form UI.
    ui->setupUi(this);

    // Initialize our variables.
    delay_index = -1;
    ignore_update = false;
    last_col = 0;
    sort_order = Qt::AscendingOrder;
    last_hscroll_pos = 0;
    last_vscroll_pos = 0;
    pause_discovery = false;
    refresh_index = -1;
    scan_timer_id = 0;
    pending_index = -1;

    // Create our UI elements.
    init_elements();

    // Initialize our connections.
    init_connects();

    // Initialize our network adapter control.
    init_network();

    // Initialize our context menu.
    init_context_menu();

    // Restore our settings.
    read_settings();

    // We need to delay a little bit for the main window to finish it's setup.
    QTimer::singleShot(FASMOTION_SETTINGS::STARTUP_DELAY, this, SLOT(finish_setup()));
}

/*!
 * @brief CameraDock destructor.
 */
CameraDock::~CameraDock()
{
    int     i;

    // Make sure we save our settings.
    write_settings();

    // Do we have any timers running?
    if (scan_timer_id)
    {
        killTimer(scan_timer_id);
        scan_timer_id = 0;
    }

    // Remove our UI element.
    delete ui;

    // Clean up our controls.
    delete camera_data;

    // Clean up our context menu.
    for (i = 0; i < CameraDock::arrange_act.count(); ++i)
    {
        // Get the pointer to the action.
        QAction *act_ptr = CameraDock::arrange_act.takeFirst();

        // Delete it.
        delete act_ptr;
    }

    // Now we can delete the submenu.
    if (CameraDock::arrange_menu != nullptr)
    {
        delete CameraDock::arrange_menu;
        CameraDock::arrange_menu = nullptr;
    }
    if (CameraDock::arrange_group != nullptr)
    {
        delete CameraDock::arrange_group;
        CameraDock::arrange_group = nullptr;
    }

    // Clean out our list of cameras.
    qDeleteAll(camera_list);
}

/*!
 * @brief Fill in the camera data object with information about the selected or the current camera.
 * @note It is assumed that sorting has been disabled!
 *
 * @param cam_info  Information about the camera we are interested in connecting to or are connected to.
 */
void CameraDock::add_row(int row_index, const FI_COMM::FI_CAMERA_INFO *cam_info)
{
    QStandardItem   *tbl_item;
    quint8          mac_addr[6];
    QHostAddress    addr;
    QString         tmp_string;
    QLatin1Char     zero_fill('0');
    QFont           item_font;
    uint16_t         fpga_patch;
    uint8_t         fpga_minor;
    uint8_t         fpga_major;

    // Sanity check.
    if (camera_data == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "camera_data pointer is NULL!";
        return;
    }

    // Are we simply resetting the table?
    if (cam_info == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "camera info pointer is NULL!";
        return;
    }

    // Do we need to grow the table?
    if (row_index >= camera_data->rowCount())
        camera_data->setRowCount(row_index + 1);

    // Has this camera gone missing?
    if (cam_info->offline)
        item_font.setStrikeOut(true);

    // User defined name
    tmp_string = QString::fromUtf8(cam_info->user_name);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_USER_NAME, tbl_item);

    // MAC address
    memcpy(&mac_addr, cam_info->mac_addr, 6);
    tbl_item = new QStandardItem(QString::fromUtf8("%1:%2:%3:%4:%5:%6").
                                 arg(mac_addr[0], 2, 16, zero_fill).
                                 arg(mac_addr[1], 2, 16, zero_fill).
                                 arg(mac_addr[2], 2, 16, zero_fill).
                                 arg(mac_addr[3], 2, 16, zero_fill).
                                 arg(mac_addr[4], 2, 16, zero_fill).
                                 arg(mac_addr[5], 2, 16, zero_fill));
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_MAC_ADDR, tbl_item);

    // IP Address
    addr.setAddress(qFromBigEndian(cam_info->ip_addr));
    tbl_item = new QStandardItem(addr.toString());
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_IP_ADDR, tbl_item);

    // Subnet mask
    addr.setAddress(qFromBigEndian(cam_info->subnet_mask));
    tbl_item = new QStandardItem(addr.toString());
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_SUBNET_MASK, tbl_item);

    // Default gateway
    addr.setAddress(qFromBigEndian(cam_info->gateway));
    tbl_item = new QStandardItem(addr.toString());
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_DEFAULT_GW, tbl_item);

    // Vendor
    tmp_string = QString::fromUtf8(cam_info->man_name);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_VENDOR_ID, tbl_item);

    // Model
    tmp_string = QString::fromUtf8(cam_info->model_name);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_MODEL, tbl_item);

    // Serial Number
    tmp_string = QString::fromUtf8(cam_info->serial_num);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_SERIAL_NUM, tbl_item);

    // Manufactures info
    tmp_string = QString::fromUtf8(cam_info->man_info);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_MANUFACTURER_INFO, tbl_item);

    // Software Version
    tmp_string = QString::fromUtf8(cam_info->dev_version);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_SOFTWARE_VER, tbl_item);

    // Protocol version
    tbl_item = new QStandardItem(QString::fromUtf8("%1.%2").
                                 arg(qFromBigEndian(cam_info->gige_maj_ver), 2, 10).
                                 arg(qFromBigEndian(cam_info->gige_min_ver), 2, 10, zero_fill));
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_PROTOCOL_VER, tbl_item);

    // Bootloader Version
    tmp_string = QString::fromUtf8("%1").arg(QString::fromUtf8(cam_info->bootloader_version));
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_BOOTLOADER_VER, tbl_item);

    // FPGA Version
    fpga_patch = cam_info->fpga_version & 0xFFFF;
    fpga_minor = (cam_info->fpga_version >> 16) & 0xFF;
    fpga_major = (cam_info->fpga_version >> 24) & 0xFF;
    tbl_item = new QStandardItem(QString::fromUtf8("%1.%2.%3").arg(fpga_major).arg(fpga_minor).arg(fpga_patch));
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_FPGA_VER, tbl_item);

    // Sensor Version
    tmp_string = QString::fromUtf8(cam_info->sensor_version);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_SENSOR_VER, tbl_item);

    // Index back into the camera_list array. For now the row_index and
    // the list index are one and the same. After sorting all bets are off.
    tmp_string = QString::fromUtf8("%1").arg(row_index);
    tbl_item = new QStandardItem(tmp_string);
    tbl_item->setFont(item_font);
    camera_data->setItem(row_index, CD_LIST_INDEX, tbl_item);
}

/*!
 * @brief The user clicked on one of the "arrange" actions in the context
 *        menu. This has the effect of sorting the data.
 *
 * @param obj_ptr   The pointer to the action item the user clicked/selected.
 */
void CameraDock::arrange_clicked(QObject *obj_ptr)
{
    int     index;
    QAction *act_ptr = qobject_cast<QAction *>(obj_ptr);

    // Get the index of the action object that was selected.
    index = CameraDock::arrange_act.indexOf(act_ptr);

    // Pass the information on as though the header was clicked.
    hdr_change(index);
}

/*!
 * @brief Given an IP address, try to find a matching entry in our table.
 *
 * @param ip_addr   The IP address to search for.
 *
 * @return The row index of the found table, or 0 if one is not found.
 */
int CameraDock::find_in_table(uint32_t ip_addr)
{
    int     row_index;

    // Loop through all entries in the table
    for (row_index = 0; row_index < camera_data->rowCount(); ++row_index)
    {
        const FI_COMM::FI_CAMERA_INFO *cam_info;

        // Get the camera information for this row.
        cam_info = get_info_by_row(row_index);
        if (cam_info == nullptr)
        {
            QString strMsg = QString::fromUtf8("%1: get_info_by_row( %2 ) returned NULL!").arg(__PRETTY_FUNCTION__).arg(row_index);
            QLOG_ERROR() << strMsg;
            return 0;
        }

        // Do we have a match?
        if (cam_info->ip_addr == ip_addr)
            break;
    }

    // If we didn't find a match, then we need to adjust our row_index value.
    if (row_index == camera_data->rowCount())
        row_index = 0;

    // Return the row index of the item, or 0.
    return row_index;
}

/*!
 * @brief Let others access our arrange_group so that it can be reused in a context menu, etc.
 */
const QActionGroup *CameraDock::get_arrange_group()
{
    return CameraDock::arrange_group;
}

/*!
 * @brief From the sel_row value, determine the index into the camera_list.
 *
 * @param sel_row   Which row of the table we're interested in, -1 for the currently active row.
 *
 * @return Associated index into the camera_list. -1 on error.
 */
int CameraDock::get_list_index(int sel_row) const
{
    int     camera_index;
    QStandardItem   *item_ptr;
    QString         item_str;

    // If there are no cameras found then we can't select one.
    if (camera_list.size() <= 0)
        return -1;
    if (camera_data->rowCount() <= 0)
        return -1;

    // Did the caller give us a row to use, or should we just grab the
    // currently selected row?
    if (sel_row == -1)
    {
        // Get the index to the current row.
        sel_row = ui->table_view->currentIndex().row();

        // Check to see if there IS a current row.
        if (sel_row == -1)
            return -1;
    }

    // Find out which camera has been selected
    item_ptr = camera_data->item(sel_row, CD_LIST_INDEX);
    if (item_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No item at row" << sel_row;
        return -1;
    }
    item_str = item_ptr->text();
    if (item_str.isEmpty())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No text at row" << sel_row;
        return -1;
    }
    camera_index = item_str.toInt();

    // Return the index into the camera_list array.
    return camera_index;
}

/*!
 * @brief Utility function to retrieve the camera information object at
 *        a specified row in the table of found cameras.
 *
 * @param sel_row   [OPTIONAL] Specify the desired row. If not specified, the currently
 *                             selected row will be used.
 *
 * @return Pointer to the selected camera information object, or NULL.
 */
FI_COMM::FI_CAMERA_INFO *CameraDock::get_info_by_row(int sel_row) const
{
    int     camera_index;
    FI_COMM::FI_CAMERA_INFO *info_ptr;

    // Get the index for the entry at the specified row.
    camera_index = get_list_index(sel_row);
    if (camera_index < 0)
        info_ptr = nullptr;
    else
        info_ptr = camera_list.at(camera_index);

    // Return what we found.
    return info_ptr;
}

/*!
 * @brief Return the row index of the selected camera.
 */
int CameraDock::get_selected() const
{
    // Which is the row selected?
    return ui->table_view->currentIndex().row();
}

void CameraDock::set_camera_autoconnect(uint32_t ip_addr)
{
    int camera_index;

    // Walk through our list of cameras.
    for (camera_index = 0; camera_index < camera_list.size(); ++camera_index)
    {
        FI_COMM::FI_CAMERA_INFO *cam_info = camera_list[camera_index];

        // Is this the one?
        if (cam_info->ip_addr == ip_addr)
        {
            // Now clean up the database.
            cam_info->rebooting = true;
            break;
        }
    }
}

/*!
 * @brief Mark the specified (by ip_addr) camera as "offline".
 *
 * The "offline" status will automatically clear (assuming the camera is not offline)
 * after the system queries and gets responses from the camera.
 *
 * @param ip_addr   The IP address of the camera we're interested in.
 */
void CameraDock::set_camera_offline(uint32_t ip_addr)
{
    int camera_index;

    // Walk through our list of cameras.
    for (camera_index = 0; camera_index < camera_list.size(); ++camera_index)
    {
        FI_COMM::FI_CAMERA_INFO *cam_info = camera_list[camera_index];

        // Is this the one?
        if (cam_info->ip_addr == ip_addr)
        {
            // Now clean up the database.
            cam_info->offline = true;
            cam_info->offline_skip = SCAN_SKIP_COUNT;
            cam_info->camera_missing_count = (uint32_t)MAX_MISSING_COUNT;

            // Refresh the display.
            update_display();
            break;
        }
    }
}

/*!
 * @brief Allows user to pause or start camera discovery checks.
 */
void CameraDock::set_discovery_check(bool check_for_cameras)
{
#if LOG_DISCOVERY_POLLING
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "*** Setting Check:" << check_for_cameras << "***";
#endif
    // If pausing discovery checks and we have an outstanding timer, kill that timer and set our discovery checking state.
    if (!check_for_cameras)
    {
        pause_discovery = true;
        if (scan_timer_id)
        {
            killTimer(scan_timer_id);
            scan_timer_id = 0;
        }
    }
    // Otherwise restart discovery check and our discovery checking state.
    else
    {
        pause_discovery = false;
        if (!scan_timer_id)
        {
            search_for_cameras();
        }
    }
}

/*!
 * @brief Perform processing associated with the "scan" timer.
 *
 *        Essentially we're waiting for discovery ACK messages from cameras and
 *        adding those cameras to a list for later processing. After waiting
 *        for WAIT_FOR_RESPONSE_TIMEOUT "ticks" we re-issue a
 *        discovery message (in case any cameras have come on-line) and we emit
 *        a search_complete signal to begin processing the found cameras.
 */
void CameraDock::handle_scan_event()
{
    int         index;
    FI_comm     *comm_ptr;
    bool        any_search;

    // Sanity check.
    if (MainWindow::GetInstance().isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "MainWindow::GetInstance() is NULL!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Are we doing a search for any cameras?
    any_search = ((net_mask == FCP_COMM::GLOBAL_BROADCAST_IPADDR) && (adapter_ip == FCP_COMM::GLOBAL_BROADCAST_IPADDR));

    // Increment our missing count.
    for (index = 0; index < camera_list.size(); ++index)
    {
        FI_COMM::FI_CAMERA_INFO *p_list_entry;

        // Get the list entry here.
        p_list_entry = camera_list[ index ];

        // Make sure we haven't already determined the camera is off-line.
        if (p_list_entry->offline)
        {
            // Decrement our skip counter if need be.
            if (p_list_entry->offline_skip > 0)
            {
                --p_list_entry->offline_skip;
            }
        }
        else
        {
            // Do we need to increment our missing counter?
            if (p_list_entry->camera_missing_count < (uint32_t)MAX_MISSING_COUNT)
            {
                p_list_entry->camera_missing_count++;
            }
            else
            {
                // Make a pretty string out of the address.
                QHostAddress our_addr;
                our_addr.setAddress(qFromBigEndian(p_list_entry->ip_addr));
                QLOG_INFO() << __PRETTY_FUNCTION__ << QString("Camera [%1] went missing for too long, marking it offline.").arg(our_addr.toString());

                // Go off-line.
                p_list_entry->offline = true;
                p_list_entry->offline_skip = SCAN_SKIP_COUNT;
                comm_ptr->set_offline(p_list_entry->ip_addr, true);
            }
        }
    }

    // Loop through the "found" cameras list.
    while (comm_ptr->get_num_cameras() > 0)
    {
        bool                    discard_info;
        bool                    got_a_match;
        uint32_t                cam_ip_addr;
        FI_COMM::FI_CAMERA_INFO *p_cam_info;

        // Get the camera information.
        p_cam_info = comm_ptr->takeFirst();
        if (p_cam_info == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera info available?!";
            continue;
        }

        // Get the IP address in the correct byte order.
        cam_ip_addr = qFromBigEndian(p_cam_info->ip_addr);

        // See if we got a match.
        got_a_match = ((cam_ip_addr & net_mask) == (adapter_ip & net_mask));

        // Assume we'll want to discard it.
        discard_info = true;

        // Now see if this camera is in the right subnet for us to
        // be able to communicate with it.
        if (any_search || got_a_match)
        {
            // Check to see if this one is already in the list.
            for (index = 0; index < camera_list.size(); ++index)
            {
                FI_COMM::FI_CAMERA_INFO *p_list_entry;

                // Get the list entry here.
                p_list_entry = camera_list[ index ];

                // Did we find it?
                if (p_list_entry && (p_list_entry->ip_addr == p_cam_info->ip_addr))
                {
                    // Reset our missing counter.
                    p_list_entry->camera_missing_count = 0;

                    // Were we offline and have we waited long enough?
                    if (p_list_entry->offline && (p_list_entry->offline_skip <= 0))
                    {
                        // Save our index as the refresh index.
                        refresh_index = index;

                        // We've waited long enough.
                        p_list_entry->offline = false;

                        // Move the camera back on-line.
                        comm_ptr->set_offline(p_list_entry->ip_addr, false);

                        // Make sure we update our information fields in case they changed while the camera was offline.
                        strncpy(p_list_entry->man_name, p_cam_info->man_name, sizeof(p_list_entry->man_name) - 1);
                        p_list_entry->man_name[ sizeof(p_cam_info->man_name) - 1 ] = '\0';
                        strncpy(p_list_entry->model_name, p_cam_info->model_name, sizeof(p_list_entry->model_name) - 1);
                        p_list_entry->model_name[ sizeof(p_cam_info->model_name) - 1 ] = '\0';
                        strncpy(p_list_entry->dev_version, p_cam_info->dev_version, sizeof(p_list_entry->dev_version) - 1);
                        p_list_entry->dev_version[ sizeof(p_cam_info->dev_version) - 1 ] = '\0';
                        strncpy(p_list_entry->man_info, p_cam_info->man_info, sizeof(p_list_entry->man_info) - 1);
                        p_list_entry->man_info[ sizeof(p_cam_info->man_info) - 1 ] = '\0';
                        strncpy(p_list_entry->serial_num, p_cam_info->serial_num, sizeof(p_list_entry->serial_num) - 1);
                        p_list_entry->serial_num[ sizeof(p_cam_info->serial_num) - 1 ] = '\0';
                        strncpy(p_list_entry->user_name, p_cam_info->user_name, sizeof(p_list_entry->user_name) - 1);
                        p_list_entry->user_name[ sizeof(p_cam_info->user_name) - 1 ] = '\0';

                        // Re-read these fields.
                        comm_ptr->start_read_reg(p_list_entry->ip_addr, GIGE_REG_FPGA_VER);
                        comm_ptr->start_read_mem(p_list_entry->ip_addr, GIGE_REG_SENSOR_VER, GIGE_SENSOR_VER_LEN);
                        comm_ptr->start_read_mem(p_list_entry->ip_addr, GIGE_REG_BOOTLOADER_VER, FI_COMM::BOOTLOADER_VER_LEN);

                        // Were we rebooting the camera?
                        if (p_list_entry->rebooting)
                        {
                            // We can clear our rebooting flag now.
                            p_list_entry->rebooting = false;

                            // Re-open the camera.
                            delay_index = index;
                            QTimer::singleShot(OPEN_DELAY, this, SLOT(delay_camera_open()));
                        }
                    }

                    // Done.
                    break;
                }
            }

            // Did we find it in the table?
            if (index == camera_list.size())
            {
                // It's a new one - add it to our list.
                camera_list.append(p_cam_info);

                // Don't discard it!
                discard_info = false;
            }
        }

        // Do we want to discard the camera info element?
        if (discard_info)
            delete p_cam_info;
    }

    // Increment our count of times through here.
    ++ticks_waited;

    // Should this be the last one?
    if ((ticks_waited > WAIT_FOR_RESPONSE_TIMEOUT) && (scan_timer_id != 0) && !pause_discovery)
    {
#if LOG_DISCOVERY_POLLING
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "scan_timer_id:" << scan_timer_id << "pause_discovery:" << pause_discovery << "***";
#endif

        // Issue another discover request.
        search_for_cameras(false);

        // Indicate that we're done.
        emit search_complete();
    }
}

/*!
 * @brief Initialize the various signal/slot connections for CameraDock.
 */
void CameraDock::init_connects() const
{
    connect(ui->cb_ni, qOverload<int>(&QComboBox::currentIndexChanged), this, &CameraDock::do_interface_changed);
    connect(ui->table_view, &fiTableView::doubleClicked, this, &CameraDock::do_item_selected);
    connect(ui->table_view->horizontalHeader(), &QHeaderView::sectionClicked, this, &CameraDock::do_hdr_clicked);

    connect(ui->table_view->verticalScrollBar(), &QScrollBar::valueChanged, this, &CameraDock::scroll_changed);
    connect(ui->table_view->horizontalScrollBar(), &QScrollBar::valueChanged, this, &CameraDock::scroll_changed);

    connect(this, &CameraDock::search_complete, this, &CameraDock::refresh_cameras);
    connect(this, &CameraDock::update_cam_info, this, &CameraDock::do_update_data);
}

/*!
 * @brief Initialize the context menu.
 */
void CameraDock::init_context_menu()
{
    int     i;
    QAction *act_ptr;

    // Sanity check.
    if (camera_data == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "camera_data pointer is NULL!";
        return;
    }

    // Allocate our action group so we can deal with them as an entity.
    CameraDock::arrange_group = new QActionGroup(this);
    if (CameraDock::arrange_group == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate arrange_menu!";
        return;
    }

    // For the context menu, our actions operate exclusive of each other.
    CameraDock::arrange_group->setExclusive(true);

    // Loop through all the headers in the camera data table.
    for (i = 0; i < CD_COL_COUNT; ++i)
    {
        QString strLabel;
        QStandardItem   *hdr_item;

        // Get the header item pointer. Make sure it has an entry (not all do).
        hdr_item = camera_data->horizontalHeaderItem(i);
        if (hdr_item == nullptr)
            continue;

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "strLabel:" << hdr_item->text();

        // Get the header label for this column.
        strLabel = hdr_item->text();
        act_ptr = new QAction(strLabel, this);
        act_ptr->setCheckable(true);
        act_ptr->setIconVisibleInMenu(false);

        // Connect our signals/slots.
        connect(act_ptr, &QAction::triggered, this, [this, act_ptr]() { arrange_clicked(act_ptr); });

        // Append the action to our list.
        CameraDock::arrange_act.append(act_ptr);

        // Add the action to our group.
        CameraDock::arrange_group->addAction(act_ptr);
    }

    // Allocate our arrange submenu.
    CameraDock::arrange_menu = new QMenu(tr("&Arrange By"), this);
    if (CameraDock::arrange_menu == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate arrange_menu!";
    }

    // Add our "View Group" to the menu.
    CameraDock::arrange_menu->addActions(CameraDock::arrange_group->actions());
}

/*!
 * @brief Initialize the various elements of the CameraDock object.
 */
void CameraDock::init_elements()
{
    QStandardItem   *tbl_item;

    // Initialize our member variables.
    scan_timer_id = 0;

    // Set our stream checkbox - always want to assume we'll stream.
    ui->cb_stream->setChecked(true);

    // Create our data model. 2 rows (Header and one empty) and
    // 13 columns (all the details we'll every show).
    camera_data = new QStandardItemModel(1, CD_COL_COUNT, this);

    // Setup the the camera detail table
    tbl_item = new QStandardItem(tr("User Defined Name"));
    camera_data->setHorizontalHeaderItem(CD_USER_NAME, tbl_item);
    tbl_item = new QStandardItem(tr("MAC"));
    camera_data->setHorizontalHeaderItem(CD_MAC_ADDR, tbl_item);
    tbl_item = new QStandardItem(tr("IP"));
    camera_data->setHorizontalHeaderItem(CD_IP_ADDR, tbl_item);
    tbl_item = new QStandardItem(tr("Subnet Mask"));
    camera_data->setHorizontalHeaderItem(CD_SUBNET_MASK, tbl_item);
    tbl_item = new QStandardItem(tr("Default Gateway"));
    camera_data->setHorizontalHeaderItem(CD_DEFAULT_GW, tbl_item);
    tbl_item = new QStandardItem(tr("Vendor"));
    camera_data->setHorizontalHeaderItem(CD_VENDOR_ID, tbl_item);
    tbl_item = new QStandardItem(tr("Model"));
    camera_data->setHorizontalHeaderItem(CD_MODEL, tbl_item);
    tbl_item = new QStandardItem(tr("Serial Number"));
    camera_data->setHorizontalHeaderItem(CD_SERIAL_NUM, tbl_item);
    tbl_item = new QStandardItem(tr("Manufacture Info"));
    camera_data->setHorizontalHeaderItem(CD_MANUFACTURER_INFO, tbl_item);
    tbl_item = new QStandardItem(tr("Software Version"));
    camera_data->setHorizontalHeaderItem(CD_SOFTWARE_VER, tbl_item);
    tbl_item = new QStandardItem(tr("Protocol Version"));
    camera_data->setHorizontalHeaderItem(CD_PROTOCOL_VER, tbl_item);
    tbl_item = new QStandardItem(tr("Bootloader Version"));
    camera_data->setHorizontalHeaderItem(CD_BOOTLOADER_VER, tbl_item);
    tbl_item = new QStandardItem(tr("FPGA Version"));
    camera_data->setHorizontalHeaderItem(CD_FPGA_VER, tbl_item);
    tbl_item = new QStandardItem(tr("Sensor Version"));
    camera_data->setHorizontalHeaderItem(CD_SENSOR_VER, tbl_item);

    // Set the model for the views to be our camera data.
    ui->table_view->setModel(camera_data);

    // Hide the list index column of the table data. There's no need for the
    // user to see that.
    ui->table_view->setColumnHidden(CD_LIST_INDEX, true);

    // Add our views to the dock.
    ui->verticalLayout->addWidget(ui->table_view);
}

/*!
 * @brief Initialize the network selection combo box.
 */
void CameraDock::init_network()
{
    QList<QNetworkAddressEntry> all_address;

    // Do not propogate any signals from the combobox while we're doing this!
    ui->cb_ni->blockSignals(true);

    // Find out all the interfaces which are on the system.
    QList<QNetworkInterface> net_interfaces = QNetworkInterface::allInterfaces();

    // Since there may be multiple network interfaces on the client machine we need
    // to loop through all of them and let the user pick which one they want to
    // send the discovery message on.
    for (int i = 0; i < net_interfaces.size(); ++i)
    {
        const QNetworkInterface             net_interface = net_interfaces.at(i);
        QNetworkInterface::InterfaceFlags   flags;

        // Get the network interface's flags.
        flags = net_interface.flags();

        // Make sure the interface is valid.
        if (((flags & QNetworkInterface::IsUp) == 0) ||                     // Network isn't up.
            ((flags & QNetworkInterface::IsRunning) == 0))                  // Network isn't running.
        {
            // Not interested.
            continue;
        }

        // The network is a loopback interface and we're not interested in local only operation.
        if (((flags & QNetworkInterface::IsLoopBack) == QNetworkInterface::IsLoopBack) && !MainWindow::local_control())
            continue;

        // Get the address information for this interface.
        all_address = net_interface.addressEntries();

        // Get the relevant IP info. We will need this to setup which interface to use.
        // Note that we are interested in the IPv4 address and we want to skip the IPv6
        // addresses.
        for (int j = 0; j < all_address.size(); ++j)
        {
            QString     cb_str;                 // The string we show in the combo box.
            QString     inter_name_str;         // The name of the interface.
            QStringList interface_info;         // A string list containing all the information we'll need on this interface.
            QHostAddress    ip, netmask;        // IP address and netmask information.

            // Not interested if it doesn't have a broadcast address.
            if (all_address.at(j).broadcast().isNull())
                continue;

            // We found a good network. Get the name of the interface.
            inter_name_str = net_interface.humanReadableName();

            // Get the IP address information, as well as the netmask and broadcast address.
            ip = all_address.at(j).ip();
            netmask = all_address.at(j).netmask();

            // Build up our string list for the combo box.
            interface_info << inter_name_str << QString::fromUtf8("%1").arg(netmask.toIPv4Address(), 0, 16) <<
                           QString::fromUtf8("%1").arg(ip.toIPv4Address(), 0, 16);

            // Build up the combo box string.
            cb_str = inter_name_str + QString::fromUtf8(" (") + ip.toString() + QString::fromUtf8(")");
            ui->cb_ni->addItem(cb_str, interface_info);
            break;
        }
    }

    // It's ok to turn signals back on now.
    ui->cb_ni->blockSignals(false);
}

/*!
 * @brief Given a camera information structure, add the camera to our table.
 * @param cam_info  Pointer to the camera information structure we're interested in.
 */
void CameraDock::insert_camera_info(const FI_COMM::FI_CAMERA_INFO *cam_info)
{
    // Add the information to the table.
    add_row(update_index, cam_info);

    // Move on to the next.
    update_index++;

    // Update the display.
    update_display();

    // Are we done?
    if (update_index < camera_list.size())
    {
        // Go around another time.
        emit update_cam_info();
    }
}

/*!
 * @brief Open the camera which is currently being selected in our table.
 */
void CameraDock::open_selected_camera()
{
    int sel_row;
    FI_COMM::FI_CAMERA_INFO *cam_info;

    // Which is the row selected?
    sel_row = ui->table_view->currentIndex().row();

    // Get the selected camera information.
    cam_info = get_info_by_row(sel_row);
    if (cam_info == nullptr)
        return;

    // Check to see if we're off-line.
    if (cam_info->offline)
        return;

    // Save if the user wants to stream video from the camera.
    cam_info->stream = ui->cb_stream->isChecked();

    // Tell the main window we've got a connect request.
    emit MainWindow::GetInstance()->connect_to_camera(sel_row);
}

/*!
 * @brief If we are on-camera and the camera information is for our
 *        camera, then we want to "connect".
 * @param cam_info  Pointer to the camera information structure.
 * @returns True if this is the camera information for our camera.
 */
bool CameraDock::our_camera(const FI_COMM::FI_CAMERA_INFO *cam_info)
{
    QString cam_name;

    // Sanity check.
    if (cam_info == nullptr)
        return false;

    // Are we running locally?
    if (!MainWindow::local_control())
        return false;
    if (!MainWindow::do_autoconnect())
        return false;

    // Put the camera name into an object we can use.
    cam_name = QString::fromUtf8(cam_info->user_name);

    // Get our name.
    if (QHostInfo::localHostName().compare(cam_name, Qt::CaseInsensitive) == 0)
        return true;

    // No match.
    return false;
}

/*!
 * @brief Handle user requested network changes.
 *
 * @param do_dhcp       True if we're supposed to enable DHCP on the selected camera.
 * @param ip_addr       If static, the IP address to use.
 * @param subnet_mask   If static, the subnet mask to use.
 * @param gw_addr       If static, the gateway address to use.
 * @param mac_addr      If we're forcing the IP address, the MAC address of the camera to talk to.
 */
void CameraDock::process_net_changes(bool do_dhcp, QHostAddress& ip_addr, QHostAddress& subnet_mask, QHostAddress& gw_addr, const uint8_t *mac_addr)
{
    int         sel_row;
    uint32_t    cam_ip_addr;
    const FI_COMM::FI_CAMERA_INFO *cam_info;

    // Figure out which element we're talking about.
    sel_row = ui->table_view->currentIndex().row();

    // Did the user click on a particular camera?
    if (sel_row < 0)
    {
        // No existing camera info to use.
        cam_ip_addr = FCP_COMM::GLOBAL_BROADCAST_IPADDR;
    }
    else
    {
        // Get the selected camera information, if any.
        cam_info = get_info_by_row(sel_row);
        cam_ip_addr = cam_info->ip_addr;
    }

    // Request a change to the camera's network configuration.
    if (MainWindow::GetInstance()->change_camera_network(cam_ip_addr, do_dhcp, ip_addr, subnet_mask, gw_addr, mac_addr))
    {
        // Reset the list and start a fresh scan.
        do_reset();
    }
}

/*!
 * @brief Reset the camera information data, clearing out our table as well as
 *        the list of cameras. Make an exception for any open cameras.
 */
void CameraDock::reset_camera_info()
{
    int count;
    int index;
    int loop;

    // Clear any prior result information.
    camera_data->removeRows(0, camera_data->rowCount());

    // Clear out any previous cameras found.
    count = camera_list.count();
    for (index = 0, loop = 0; loop < count; ++loop)
    {
        FI_COMM::FI_CAMERA_INFO *cam_info;

        // Get the camera information at this index.
        cam_info = camera_list.at(index);

        // We have to be a little careful here. If we're actually connected to the camera
        // we can't go blow it's database away!
        if (MainWindow::GetInstance()->find_mdi_child(cam_info->ip_addr) == nullptr)
        {
            // No activity on this camera, go ahead and delete it. There's no need
            // to update index, 'cuz the next camera will fill this slot.
            delete camera_list.takeAt(index);
        }
        else
        {
            // Skip this camera - just move on to the next.
            index++;
        }
    }

    // Make sure we reset our update index.
    update_index = 0;
}

/*!
 * @brief Slot to handle scrollbar changes.
 * @param newValue  The new scrollbar setting.
 */
void CameraDock::scroll_changed(int newValue)
{
    QScrollBar *scroll_ptr = qobject_cast<QScrollBar *>(sender());

    // Make sure we have a scrollbar. Also, ignore the update if
    // we're told to.
    if (ignore_update || (scroll_ptr == nullptr))
        return;

    // Make sure we save the correct value.
    if (scroll_ptr->orientation() == Qt::Vertical)
        last_vscroll_pos = newValue;
    else
        last_hscroll_pos = newValue;
}

/*!
 * @brief This is the worker function which actually goes out to the selected
 *        network interface and determines if there are any cameras we might
 *        be able to connect to.
 *
 * @param reset_table   True if we want to reset the table, false to simply add to it.
 */
void CameraDock::search_for_cameras(bool reset_table)
{
    QStringList interface_info;
    FI_comm     *comm_ptr;

    // Are we supposed to reset the table?
    if (reset_table)
        reset_camera_info();

    // Get the selected interface.
    interface_info = ui->cb_ni->itemData(ui->cb_ni->currentIndex()).toStringList();
    if (!interface_info.isEmpty())
    {
        // Parse the information from the interface_info list.
        net_mask = interface_info.at(1).toUInt(nullptr, 16);
        adapter_ip = interface_info.at(2).toUInt(nullptr, 16);
    }
    else
    {
        QMessageBox::critical(this, tr("Find Cameras"), tr("A network adapter must be selected"));
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // See if we can find any cameras to talk to.
    if (comm_ptr->find_cameras(adapter_ip, 0) != FI_COMM::FI_SUCCESS)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Comms error, resetting...";
        MainWindow::GetInstance()->reset_comm();
        return;
    }

    // Reset our counter.
    ticks_waited = 0;

#if LOG_DISCOVERY_POLLING
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "***ticks_waited:" << ticks_waited << "pause_discovery:" << pause_discovery << "***";
#endif

    if (reset_table)
    {
        // Setup a timer to query the response.
        if ((scan_timer_id == 0) && !pause_discovery)
        {
#if LOG_DISCOVERY_POLLING
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "***Restarting Discovery Checks!***";
#endif
            scan_timer_id = startTimer(SCAN_UPDATE_INTERVAL);
        }
    }
}

/*!
 * @brief Open the camera indexed by "index" in our list of cameras.
 * @param index     The index into our camera_list object for the camera we're interested in.
 */
void CameraDock::open_camera(int index)
{
    int cam_row;

    // Sanity check the index.
    if (index >= camera_list.count())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index is too large:" << index;
        return;
    }

    // Get the index in the model for our item.
    cam_row = find_in_table(camera_list[ index ]->ip_addr);

    // Save if the user wants to stream video from the camera.
    camera_list[ index ]->stream = ui->cb_stream->isChecked();

    // Tell the main window we've got a connect request.
    emit MainWindow::GetInstance()->connect_to_camera(cam_row);
}

/*!
 * @brief Translate the UI elements for this object.
 */
void CameraDock::translate_ui()
{
    // Have the UI object perform it's translations.
    ui->retranslateUi(this);

    // Adjust our labels.
    camera_data->horizontalHeaderItem(CD_USER_NAME)->setText(tr("User Defined Name"));
    camera_data->horizontalHeaderItem(CD_MAC_ADDR)->setText(tr("MAC"));
    camera_data->horizontalHeaderItem(CD_IP_ADDR)->setText(tr("IP"));
    camera_data->horizontalHeaderItem(CD_SUBNET_MASK)->setText(tr("Subnet Mask"));
    camera_data->horizontalHeaderItem(CD_DEFAULT_GW)->setText(tr("Default Gateway"));
    camera_data->horizontalHeaderItem(CD_VENDOR_ID)->setText(tr("Vendor"));
    camera_data->horizontalHeaderItem(CD_MODEL)->setText(tr("Model"));
    camera_data->horizontalHeaderItem(CD_SERIAL_NUM)->setText(tr("Serial Number"));
    camera_data->horizontalHeaderItem(CD_MANUFACTURER_INFO)->setText(tr("Manufacture Info"));
    camera_data->horizontalHeaderItem(CD_SOFTWARE_VER)->setText(tr("Software Version"));
    camera_data->horizontalHeaderItem(CD_PROTOCOL_VER)->setText(tr("Protocol Version"));
    camera_data->horizontalHeaderItem(CD_BOOTLOADER_VER)->setText(tr("Bootloader Version"));
    camera_data->horizontalHeaderItem(CD_FPGA_VER)->setText(tr("FPGA Version"));
    camera_data->horizontalHeaderItem(CD_SENSOR_VER)->setText(tr("Sensor Version"));

    // Cleanup our context menu.
    if (CameraDock::arrange_menu != nullptr)
    {
        int i;

        // Clean up our context menu.
        for (i = 0; i < CameraDock::arrange_act.count(); ++i)
        {
            // Get the pointer to the action.
            QAction *act_ptr = CameraDock::arrange_act.takeFirst();

            // Delete it.
            delete act_ptr;
        }

        // Now we can delete the submenu.
        delete CameraDock::arrange_menu;
        delete CameraDock::arrange_group;
    }

    // Re-build our context menu.
    init_context_menu();
}

/*!
 * @brief Force the views to update, resizing the columns, reformatting, etc.
 */
void CameraDock::update_display()
{
    int sel_index;
    QModelIndex my_cell;

    // We've updated all the cameras. Resize the columns in our table view.
    ui->table_view->resizeRowsToContents();
    ui->table_view->resizeColumnsToContents();
    ui->table_view->horizontalHeader()->setStretchLastSection(true);

    // We can now re-enable sorting.
    ui->table_view->setSortingEnabled(true);

    // Sort the data according to the IP address.
    hdr_change(((last_col == -1) ? 0 : last_col), false);

    // Get the index in the model for our item.
    sel_index = find_in_table(selected_ip);

    // Make sure we ignore any changes to the scroll bars when we
    // restore the current index of the table.
    ignore_update = true;
    my_cell = camera_data->index(sel_index, last_col);
    ui->table_view->setCurrentIndex(my_cell);
    ignore_update = false;

    // Restore the scrollbar position.
    if (ui->table_view->horizontalScrollBar()->maximum() < last_hscroll_pos)
        ui->table_view->horizontalScrollBar()->setMaximum(last_hscroll_pos);
    ui->table_view->horizontalScrollBar()->setValue(last_hscroll_pos);
    if (ui->table_view->verticalScrollBar()->maximum() < last_vscroll_pos)
        ui->table_view->verticalScrollBar()->setMaximum(last_vscroll_pos);
    ui->table_view->verticalScrollBar()->setValue(last_vscroll_pos);

    // Make sure to restore the focus, if we're supposed to.
    if (restore_focus)
        ui->table_view->setFocus();
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void CameraDock::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    QDockWidget::changeEvent(event);
}

/*!
 * @brief Handle a right click from the user - this will pop-up a context menu.
 * @param event_ptr     Pointer to the right-click event.
 */
void CameraDock::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    int     sel_row;
    QString menu_str;
    QMenu   context_menu(this);

    // Add our "View Group" to the menu.
    context_menu.addMenu(CameraDock::arrange_menu);
    context_menu.addAction(tr("Add &Column"), this, SLOT(do_add_column()));

    // Add a seperator before the next section.
    context_menu.addSeparator();

    // Add some menus for the user to connect to the camera in different ways
    context_menu.addAction(MainWindow::GetInstance()->get_cam_connect());
    context_menu.addAction(MainWindow::GetInstance()->get_cam_multicam());

    // Add a section to let the user remove one or all of the entries in the display.
    context_menu.addSeparator();

    // Get the currently selected row.
    sel_row = ui->table_view->currentIndex().row();

    // Did the user click on a particular camera?
    if (sel_row >= 0)
        context_menu.addAction(tr("Clear"), this, SLOT(do_clear_entry()));
    context_menu.addAction(tr("Clear All"), this, SLOT(do_clear_all_entries()));

    // Add our seperator.
    context_menu.addSeparator();

    // Did the user click on a particular camera?
    if (sel_row >= 0)
    {
        const FI_COMM::FI_CAMERA_INFO *cam_info;

        // Get the camera associated with this row.
        cam_info = get_info_by_row(sel_row);
        if ((cam_info != nullptr) && !cam_info->offline && CameraMDI::is_series3(cam_info))
        {
            menu_str = tr("Packet Delay (%1us)...").arg(cam_info->packet_delay);
            context_menu.addAction(menu_str, this, SLOT(do_set_packet_delay()));
        }
    }

    // Let the user get to the force IP dialog.
    menu_str = tr("Change IP Address...");
    context_menu.addAction(menu_str, this, SLOT(do_force_ip()));

    // Add a menu to re-dock the panel.
    if (isFloating())
    {
        context_menu.addSeparator();
        context_menu.addAction(tr("Dock"), this, SLOT(do_dock()));
    }

    // Is the factory menu enabled?
    if (MainWindow::GetInstance()->is_factory_enabled())
    {
        context_menu.addSeparator();
        context_menu.addAction(tr("Open Console"), this, SLOT(do_factory_console()));
    }

    // Select the action which is currently used for sorting.
    if (last_col < CameraDock::arrange_act.count())
    {
        QAction *act_ptr = CameraDock::arrange_act.at(last_col);
        act_ptr->setChecked(true);
    }

    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

/*!
 * @brief Poll the communications object to see if there are any new cameras
 *        which have responded to our discovery message.
 *
 * @param event_ptr     Pointer to the object which generated this event.
 */
void CameraDock::timerEvent(QTimerEvent *event_ptr)
{
    // Sanity check.
    if (event_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null event pointer?!";
        return;
    }

    // Proceess based on the event ID.
    if (event_ptr->timerId() == scan_timer_id)
        handle_scan_event();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief It's time to open that camera!
 */
void CameraDock::delay_camera_open()
{
    // Sanity check.
    if ((delay_index >= 0) && (delay_index < camera_list.count()))
        open_camera(delay_index);

    // Reset our delay index.
    delay_index = -1;
}

/*!
 * @brief The user wants to modify the columns being shown in the detailed view.
 */
void CameraDock::do_add_column()
{
    int     i;
    int     data_index;
    ShowColumnsDialog   dlg(this);
    QStandardItem       *tbl_item;
    QHeaderView         *hdr_ptr;

    // Get a poitner to our header.
    hdr_ptr = ui->table_view->horizontalHeader();

    // Loop through all of our columns.
    for (i = 0; i < CD_COL_COUNT; ++i)
    {
        // Get our header item. We want them "as displayed", not in sequential
        // "stored" order.
        for (data_index = 0; data_index < hdr_ptr->count(); ++data_index)
        {
            // If we found the "next" one, break out of the loop.
            if (i == hdr_ptr->visualIndex(data_index))
                break;
        }
        tbl_item = camera_data->horizontalHeaderItem(data_index);
        if (tbl_item == nullptr)
            continue;

        // We need to build up the used and available lists.
        //
        // If a column is in use, then it's not available. We define a column
        // to be in use if it's visible.
        if (ui->table_view->isColumnHidden(data_index))
            dlg.avail_cols.append(tbl_item->text());
        else
            dlg.used_cols.append(tbl_item->text());

        // The factory list gets them in sequential order.
        if (i != data_index)
            tbl_item = camera_data->horizontalHeaderItem(i);
        dlg.factory_cols.append(tbl_item->text());
    }

    // Pop up the dialog.
    if (dlg.exec() == QDialog::Accepted)
    {
        // Hide all columns of the table.
        for (i = 0; i < hdr_ptr->count(); ++i)
            ui->table_view->hideColumn(i);

        // Loop through all columns of the camera data.
        for (i = 0; i < dlg.used_cols.count(); ++i)
        {
            QString label_str;

            // Get our entry.
            label_str = dlg.used_cols.at(i);

            // Find the entry in our table.
            for (data_index = 0; data_index < CD_COL_COUNT; data_index++)
            {
                // Get our header item.
                tbl_item = camera_data->horizontalHeaderItem(data_index);
                if (tbl_item == nullptr)
                    continue;

                // Is this a match?
                if (label_str == tbl_item->text())
                {
                    // Found it.
                    break;
                }
            }

            if (data_index == CD_COL_COUNT)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << label_str << "not found!";
            }

            else
            {
                // Show this column.
                ui->table_view->showColumn(data_index);

                // Do we need to remap this column?
                if (i != hdr_ptr->visualIndex(data_index))
                {
                    // Move from the old column to the new one.
                    ui->table_view->horizontalHeader()->moveSection(hdr_ptr->visualIndex(data_index), i);
                }
            }
        }

        // Force the display to update.
        update_display();
    }
}

/*!
 * @brief Let the user remove all the entries from the table.
 */
void CameraDock::do_clear_all_entries()
{
    do_reset();
}

/*!
 * @brief Let the user clear a single entry from the table. This allows the user
 *        to "age" and remove old/defunct entries.
 */
void CameraDock::do_clear_entry()
{
    int     index;
    int     camera_index;
    QMdiSubWindow           *mdi_ptr;
    FI_COMM::FI_CAMERA_INFO *cam_info;

    // Figure out which element we're talking about.
    index = ui->table_view->currentIndex().row();

    // Get the index into our camera list.
    camera_index = get_list_index(index);

    // Get the pointer to the camera information.
    cam_info = camera_list.at(camera_index);

    // We have to be a little careful here. If we're actually connected to the camera
    // we have to notify the user!
    mdi_ptr = MainWindow::GetInstance()->find_mdi_child(cam_info->ip_addr);
    if (mdi_ptr != nullptr)
    {
        QString     title_str;
        CameraMDI   *camera_ptr;
        QMdiArea    *mdi_area = qobject_cast<QMdiArea *>(MainWindow::GetInstance()->centralWidget());

        // Make this child the active window.
        mdi_area->setActiveSubWindow(mdi_ptr);

        // Get a pointer to the camera object.
        camera_ptr = qobject_cast<CameraMDI *>(mdi_ptr->widget());
        if (camera_ptr == nullptr)
            title_str = tr("FasMotion");
        else
            title_str = camera_ptr->get_camera_name();

        // Let the user know they have more work to do.
        QMessageBox::information(MainWindow::GetInstance().get(), title_str, tr("Please close the window before trying to clear the camera."));
        return;
    }

    // Now clean up the database.
    cam_info = camera_list.takeAt(camera_index);
    delete cam_info;

    // Rebuild our camera database.
    camera_data->removeRows(0, camera_data->rowCount());
    for (index = 0; index < camera_list.size(); ++index)
    {
        // Add this camera back into the table.
        add_row(index, camera_list.at(index));
    }

    // Clear the selection.
    selected_ip = 0;

    // Refresh the display.
    update_display();
}

/*!
 * @brief Dock ourselves.
 */
void CameraDock::do_dock()
{
    setFloating(false);
}

/*!
 * @brief The user is requesting that we open a console window to the selected camera.
 */
void CameraDock::do_factory_console()
{
    int             sel_row;
    QString         url_string;
    QHostAddress    our_addr;
    const FI_COMM::FI_CAMERA_INFO *cam_info;

    // Figure out which element we're talking about.
    sel_row = ui->table_view->currentIndex().row();

    // Did the user click on a particular camera?
    if (sel_row < 0)
        return;

    // Get the selected camera information.
    cam_info = get_info_by_row(sel_row);
    if (cam_info == nullptr)
        return;

    // Convert the packed binary format into a human readable ip address.
    our_addr.setAddress(qFromBigEndian(cam_info->ip_addr));

    // If we are a series3 camera, use "Shell In A Box"
    if (CameraMDI::is_series3(cam_info))
    {
        // We get there by specifying the port number.
        url_string = QString::fromUtf8("http://%1:4200").arg(our_addr.toString());
        QDesktopServices::openUrl(url_string);
    }
    else
    {
#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
        // For newer cameras, just open up an SSH terminal.
        QProcess *ssh = new QProcess(this);
        QStringList ssh_args;

        ssh_args << our_addr.toString();
        ssh->start("ssh", ssh_args);
#else
        QString putty_path;
        QSettings settings;
        QProcess *putty = new QProcess(this);
        QStringList putty_args;

        // Get the full path to PuTTY.
        settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
        putty_path = settings.value("putty_path").toString();
        settings.endGroup();

        // WinOS: For newer cameras, just open up an SSH terminal.
        // Assumption: putty is on the path.
        putty_args << "-ssh" << our_addr.toString();
        putty->start(putty_path, putty_args);
#endif
    }
}

/*!
 * @brief User pressed the Change Address button. We want to bring up the
 *        network dialog so that the user can change the IP address,
 *        mask, etc. of the selected camera.
 */
void CameraDock::do_force_ip()
{
    int     index;
    int     sel_row;
    const FI_COMM::FI_CAMERA_INFO *cam_info;

    // Figure out which element we're talking about.
    sel_row = ui->table_view->currentIndex().row();

    // Did the user click on a particular camera?
    if (sel_row < 0)
    {
        // No existing information structure for us to use.
        cam_info = nullptr;
    }
    else
    {
        // Get the selected camera information.
        cam_info = get_info_by_row(sel_row);
    }

    // Create our dialog.
    ForceIPDialog dlg(MainWindow::GetInstance().get());
    dlg.setMinimumSize(200, 100);

    // When entering the dialog via this path, we'll allow the user
    // to modify the MAC address.
    dlg.allow_mac_edit(true);

    // Can we talk to this camera "normally"?
    for (index = 0; (cam_info != nullptr) && (index < ui->cb_ni->count()); ++index)
    {
        bool        got_a_match;
        uint32_t    cam_ip_addr;
        uint32_t    net_mask;
        uint32_t    adapter_ip;
        QStringList interface_info;

        // Get the interface specific information.
        interface_info = ui->cb_ni->itemData(index).toStringList();
        if (interface_info.count() == 0)
            continue;

        // Extract our mask information from the string list.
        net_mask = interface_info.at(1).toUInt(nullptr, 16);
        adapter_ip = interface_info.at(2).toUInt(nullptr, 16);

        // Get the IP address in the correct byte order.
        cam_ip_addr = qFromBigEndian(cam_info->ip_addr);

        // See if we got a match.
        got_a_match = ((cam_ip_addr & net_mask) == (adapter_ip & net_mask));
        if (got_a_match)
            break;
    }

    // Did we find an interface we can use to talk directly to this camera?
    if ((index == ui->cb_ni->count()) || (cam_info == nullptr))
    {
        // Don't allow DHCP because the only thing we can do is a "Force IP" command
        // which only lets us set a static IP address.
        dlg.set_dhcp_allowed(false);
    }
    else
    {
        // If the camera is off-line, we cannot talk directly to this camera, so
        // DHCP is not allowed, otherwise DHCP is fine.
        if (cam_info->offline)
            dlg.set_dhcp_allowed(false);
        else
            dlg.set_dhcp_allowed(true);
    }

    // Pass the information on the selected camera to the dialog.
    if (cam_info != nullptr)
        dlg.set_info(cam_info);

    // Run the dialog for the user.
    if (dlg.exec() == QDialog::Accepted)
    {
        // Process the changes requested by the user.
        process_net_changes(dlg.dhcp_is_enabled(), dlg.get_ip(), dlg.get_net_mask(), dlg.get_gw(), dlg.get_mac());
    }
}

/*!
 * @brief Slot invoked when we want to sort the found table.
 *
 * @param index         Which column are we sorting on?
 */
void CameraDock::do_hdr_clicked(int index)
{
    hdr_change(index);
}

/*!
 * @brief Method to change the sorting of our table.
 *
 * @param index         Which column are we sorting on?
 * @param toggle_sort   [OPTIONAL] True if we want to toggle the sort direction, false otherwise.
 */
void CameraDock::hdr_change(int index, bool toggle_sort)
{
    // If we clicked on this column the last time, flip the sort order.
    if (toggle_sort && (last_col == index))
    {
        if (sort_order == Qt::AscendingOrder)
            sort_order = Qt::DescendingOrder;
        else
            sort_order = Qt::AscendingOrder;
    }

    // Sort the data.
    camera_data->sort(index, sort_order);

    // Indicate which column and the direction.
    ui->table_view->horizontalHeader()->setSortIndicator(index, sort_order);

    // Remember this as the last column selected.
    last_col = index;
}

/*!
 * @brief Respond to the user changing the network interface. This results in clearing
 *        the existing list of cameras so we can build up a new one.
 */
void CameraDock::do_interface_changed(int new_index)
{
    Q_UNUSED(new_index);

    // Force a reset of the list.
    do_clear_all_entries();
}

/*!
 * @brief Respond to the user double-clicking on an item, either in list or table view.
 * @param index     Which item was selected.
 */
void CameraDock::do_item_selected(QModelIndex index)
{
    Q_UNUSED(index);

    // Open the cameera.
    open_selected_camera();
}

/*!
 * @brief Respond to the user selecting the menu item to add the selected camera to the
 *        multicamera dock.
 */
void CameraDock::do_multicam()
{
    int             sel_row;
    MultiCamDock    *multicam_dock;
    const FI_COMM::FI_CAMERA_INFO *cam_info;

    // Which is the row selected?
    sel_row = ui->table_view->currentIndex().row();

    // Get the selected camera information.
    cam_info = get_info_by_row(sel_row);
    if (cam_info == nullptr)
        return;

    // Check to see if we're off-line.
    if (cam_info->offline)
        return;

    // Get the multi-camera dock.
    multicam_dock = MainWindow::GetInstance()->get_multicam_dock();
    if (multicam_dock == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No multi-camera dock!";
        return;
    }

    // Make sure the dock is visible.
    multicam_dock->setVisible(true);

    // Add the selected camera to the dock.
    multicam_dock->add_ip_addr(cam_info->ip_addr);
}

/*!
 * @brief Automatically try to find all cameras we could possibly connect with.
 *        In addition, this resets the data so "stale" entries are removed.
 */
void CameraDock::do_reset()
{
    // Do we have a scan update timer running?
    if (scan_timer_id)
    {
        killTimer(scan_timer_id);
        scan_timer_id = 0;
    }

    // Start the search.
    search_for_cameras();
}

/*!
 * @brief The user wants to adjust the inter-packet delay for the packets being
 *        sent from the camera.
 */
void CameraDock::do_set_packet_delay()
{
    int     sel_row;
    int     new_delay;
    bool    ok;
    FI_COMM::FI_CAMERA_INFO *cam_info;

    // Figure out which element we're talking about.
    sel_row = ui->table_view->currentIndex().row();

    // Did the user click on a particular camera?
    if (sel_row < 0)
        return;

    // Get the camera associated with this row.
    cam_info = get_info_by_row(sel_row);
    if (cam_info == nullptr)
        return;

    // Ask the user for a new delay value.
    new_delay = QInputDialog::getInt(MainWindow::GetInstance().get(), tr("Packet Delay"), tr("Enter delay (microseconds):"),
                                     (int)cam_info->packet_delay,
                                     FI_COMM::PKT_DELAY_MIN, FI_COMM::PKT_DELAY_MAX,
                                     1, &ok);

    // Did the user accept a new value?
    if (ok)
    {
        CameraMDI   *camera_ptr;

        // Save the new value into our structure.
        cam_info->packet_delay = (uint32_t)new_delay;

        // Is this camera connected?
        camera_ptr = MainWindow::GetInstance()->active_mdi_child();
        if ((camera_ptr != nullptr) && (camera_ptr->get_ip_addr() == cam_info->ip_addr))
        {
            FI_comm     *comm_ptr;

            // Get our communications object pointer.
            comm_ptr = MainWindow::GetInstance()->get_comm();
            if (comm_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
                return;
            }

            // Tell the camera about our new delay.
            comm_ptr->start_write_reg(cam_info->ip_addr, GIGE_REG_SCPD0, (uint32_t)new_delay);
        }
    }
}


/*!
 * @brief Update fields in the camera data, and once complete, add the camera
 *        to our list of cameras.
 *
 * @param info_ptr      Pointer to our camera information structure we're working on.
 */
void CameraDock::do_update_data()
{
    bool        add_the_camera;
    int         status;
    FI_comm     *comm_ptr;
    FI_COMM::FI_CAMERA_INFO *info_ptr;

    // Are we still waiting for a response from this camera?
    if (update_index == pending_index)
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        QApplication::restoreOverrideCursor();
        return;
    }

    // Get our camera data.
    info_ptr = camera_list.at(update_index);

    // Assume we're not ready to add it yet.
    add_the_camera = false;

    // See if we have the FPGA version in the camera information.
    if (info_ptr->fpga_version == 0)
    {
        // Remember we're waiting on this camera.
        pending_index = update_index;

        // Go read the FPGA version register.
        status = comm_ptr->start_read_reg(info_ptr->ip_addr, GIGE_REG_FPGA_VER);
        if (status != FI_COMM::FI_SUCCESS)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "start_read_reg( GIGE_REG_FPGA_VER ) failed:" << status;
        }
    }
    else if (strlen(info_ptr->sensor_version) == 0)
    {
        // Remember we're waiting on this camera.
        pending_index = update_index;

        // Go read the sensor version information.
        status = comm_ptr->start_read_mem(info_ptr->ip_addr, GIGE_REG_SENSOR_VER, GIGE_SENSOR_VER_LEN);
        if (status != FI_COMM::FI_SUCCESS)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "start_read_mem( GIGE_REG_SENSOR_VER ) failed:" << status;
        }
    }
    else if (info_ptr->bootloader_version[0] == 0)
    {
        // Remember we're waiting on this camera.
        pending_index = update_index;

        // Go read the bootloader version information.
        status = comm_ptr->start_read_mem(info_ptr->ip_addr, GIGE_REG_BOOTLOADER_VER, FI_COMM::BOOTLOADER_VER_LEN);
        if (status != FI_COMM::FI_SUCCESS)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "start_read_mem( GIGE_REG_BOOTLOADER_VER ) failed:" << status;
        }
    }
    else
    {
        // We're ready to add the camera to the list.
        add_the_camera = true;
    }

    // Are we ready to add the camera?
    if (add_the_camera)
    {
        // Save the index.
        int our_index = update_index;

        // Insert the camera information into the database. Note that this
        // will modify the "update_index".
        insert_camera_info(info_ptr);

        // If we added our local camera, then automatically connect to it.
        if ((our_index >= 0) && !info_ptr->startup_open && our_camera(info_ptr))
        {
            // Open the camera.
            delay_index = our_index;
            QTimer::singleShot(OPEN_DELAY, this, SLOT(delay_camera_open()));

            // Remember that we've done this once.
            info_ptr->startup_open = true;
        }
    }
}

/*!
 * @brief Handle the "delay" timer event.
 *
 *        This is needed to let the main window finish its startup sequence
 *        so that we'll have the resource we need to do network communications.
 */
void CameraDock::finish_setup()
{
    FI_comm     *comm_ptr;

    // Are we ready to play yet?
    if (MainWindow::GetInstance().isNull())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
        return;

    // Connect the communications interfaces that we're interested in.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &CameraDock::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &CameraDock::handle_msg_timeout, Qt::QueuedConnection);

    // Connect the signal to save and load settings.
    connect(MainWindow::GetInstance().get(), &MainWindow::save_settings, this, &CameraDock::write_settings);
    connect(MainWindow::GetInstance().get(), &MainWindow::load_settings, this, &CameraDock::read_settings);

    connect(MainWindow::GetInstance()->get_cam_connect(), &QAction::triggered, this, &CameraDock::open_selected_camera);
    connect(MainWindow::GetInstance()->get_cam_multicam(), &QAction::triggered, this, &CameraDock::do_multicam);

    // Tell the user that we've started.
    MainWindow::GetInstance()->set_statusbar_message(tr("Searching for cameras..."), 2000);

    // Start our scan.
    do_reset();
}

/*!
 * @brief We timed out waiting for a message to be sent to the camera. We use the broadcast address
 *        so make sure that's the IP address we're timing out on.
 *
 *        The assumption is that the camera just fell off the grid. We'll clear it out of our list, and
 *        if this was a mistake, it'll respond to the next discovery message and come back in.
 *
 * @param ip_addr   IP address of the camera that timed out. We're only interested in the broadcast address.
 */
void CameraDock::handle_msg_timeout(uint32_t ip_addr, uint32_t /* reg_addr */, bool /* read_op */)
{
    QHostAddress    our_addr;
    const FI_COMM::FI_CAMERA_INFO *p_list_entry;

    // Make sure this is for us.
    if (ip_addr != FCP_COMM::GLOBAL_BROADCAST_IPADDR)
        return;
    if (update_index >= camera_list.count())
        return;

    // Remove this from the list.
    p_list_entry = camera_list.takeAt(update_index);
    our_addr.setAddress(qFromBigEndian(p_list_entry->ip_addr));
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing camera:" << our_addr;
    delete p_list_entry;

    // Reset our pending index.
    pending_index = -1;

    // Are there any other cameras in the list?
    if (!camera_list.isEmpty())
    {
        // Try updating a different camera.
        update_index = 0;
        emit update_cam_info();
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void CameraDock::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    bool        do_update = false;
    FI_comm     *comm_ptr;
    QByteArray  *data_ptr;
    FI_COMM::FI_CAMERA_INFO *p_list_entry;

    // Make sure this is for us.
    if (ip_addr != FCP_COMM::GLOBAL_BROADCAST_IPADDR)
        return;

    // Initialize our list pointer.
    p_list_entry = nullptr;

    // Reset our pending index.
    pending_index = -1;

    // Sanity check.
    if (update_index < camera_list.size())
    {
        // Get the selected camera information.
        p_list_entry = camera_list.at(update_index);

        // We must be in the update process, so make sure we carry on with it.
        do_update = true;
    }
    else if ((refresh_index >= 0) && (refresh_index < camera_list.size()))
    {
        // If we're here, then we're not in the "normal" update process. Make sure
        // we don't emit any signals.
        do_update = false;

        // Get the selected camera information.
        p_list_entry = camera_list.at(refresh_index);
    }

    // Sanity check.
    if (p_list_entry == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No selected camera!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_FPGA_VER:
        // Read the FPGA version
        p_list_entry->fpga_version = comm_ptr->get_reg(ip_addr, reg_addr);
        break;
    case GIGE_REG_SENSOR_VER:
        // Read the sensor version information.
        data_ptr = comm_ptr->get_reg_buffer(ip_addr, reg_addr);
        if (data_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_reg_buffer( GIGE_REG_SENSOR_VER ) returned NULL!";
            return;
        }

        // Copy the data over to our field.
        memcpy(p_list_entry->sensor_version, data_ptr->constData(), (size_t)data_ptr->size());
        delete data_ptr;
        break;
    case GIGE_REG_BOOTLOADER_VER:
        // Read the bootloader version information.
        data_ptr = comm_ptr->get_reg_buffer(ip_addr, reg_addr);
        if (data_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_reg_buffer( GIGE_REG_BOOTLOADER_VER ) returned NULL!";
            return;
        }
        if (data_ptr->size() == 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "NULL GIGE_REG_BOOTLOADER_VER from" << p_list_entry->user_name;
            data_ptr->append('X');
        }

        // Copy the data over to our field.
        memcpy(p_list_entry->bootloader_version, data_ptr->constData(), (size_t)data_ptr->size());
        delete data_ptr;

        // Reset the refresh index.
        if (refresh_index >= 0)
            refresh_index = -1;
        break;
    default:
        // This isn't one of ours, so carry on.
        do_update = false;
        break;
    }

    // Are we in the update process?
    if (do_update)
    {
        // See if we're all done.
        emit update_cam_info();
    }
}

/*!
 * @brief Read the CameraDock specific settings from the application parameter store.
 */
void CameraDock::read_settings()
{
    bool        stream_on_connect;
    int         index;
    QString     net_if;

    // Load our previous settings.
    QSettings settings;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Settings is located:" << settings.fileName();

    settings.beginGroup(CameraDock::CAMERA_DOCK_GROUP_NAME);
    last_col = settings.value(CameraDock::CAMERA_DOCK_SORT_COL, CD_USER_NAME).toInt();
    sort_order = (Qt::SortOrder)settings.value(CameraDock::CAMERA_DOCK_SORT_ORDER, Qt::AscendingOrder).toUInt();
    stream_on_connect = settings.value(CameraDock::CAMERA_DOCK_STREAM_ON_CONNECT, true).toBool();
    net_if = settings.value(CameraDock::CAMERA_DOCK_NET_IF).toString();
    ui->table_view->horizontalHeader()->restoreState(settings.value(CameraDock::CAMERA_DOCK_TAB_DETAIL_STATE).toByteArray());
    settings.endGroup();

    // Are we streaming on connect?
    ui->cb_stream->setChecked(stream_on_connect);

    // Do not propogate any signals from the combobox while we're doing this!
    ui->cb_ni->blockSignals(true);

    // Select the network interface previously used.
    index = ui->cb_ni->findText(net_if);
    if (index >= 0)
        ui->cb_ni->setCurrentIndex(index);

    // It's ok to turn on signalling again.
    ui->cb_ni->blockSignals(false);

    // Sort the data.
    camera_data->sort(last_col, sort_order);

    // Indicate which column and the direction.
    ui->table_view->horizontalHeader()->setSortIndicator(last_col, sort_order);
}

/*!
 * @brief Reset and reload the camera information tables.
 */
void CameraDock::refresh_cameras()
{
    const FI_COMM::FI_CAMERA_INFO *cam_info;

    // Get the selected camera information.
    cam_info = get_info_by_row();
    if (cam_info == nullptr)
        selected_ip = 0;
    else
        selected_ip = cam_info->ip_addr;

    // Does the table have the focus?
    restore_focus = ui->table_view->hasFocus();

    // Clear out our table if necessary.
    if (camera_data->rowCount() > 0)
        camera_data->removeRows(0, camera_data->rowCount());

    // Do we have any cameras?
    if (camera_list.size() == 0)
        return;

    // Disable sorting while we add our rows.
    ui->table_view->setSortingEnabled(false);

    // Start the update process to make sure all the fields are populated.
    update_index = 0;
    emit update_cam_info();
}

/*!
 * @brief Save our settings to the application's parameter store.
 */
void CameraDock::write_settings()
{
    // Save the current settings.
    QSettings settings;
    settings.beginGroup(CameraDock::CAMERA_DOCK_GROUP_NAME);
    settings.setValue(CameraDock::CAMERA_DOCK_SORT_COL, last_col);
    settings.setValue(CameraDock::CAMERA_DOCK_SORT_ORDER, sort_order);
    settings.setValue(CameraDock::CAMERA_DOCK_TAB_DETAIL_STATE, ui->table_view->horizontalHeader()->saveState());
    settings.setValue(CameraDock::CAMERA_DOCK_STREAM_ON_CONNECT, ui->cb_stream->isChecked());
    settings.setValue(CameraDock::CAMERA_DOCK_NET_IF, ui->cb_ni->currentText());
    settings.endGroup();
}
