/*!
 * @file camera_dock.h
 * @brief This file contains the definition of the CameraDock class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_DOCK_H
#define CAMERA_DOCK_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QActionGroup>
#include <QAbstractItemModel>
#include <QAbstractItemView>
#include <QDockWidget>
#include <stdint.h>

#include "fi_comm.h"

namespace Ui
{
class CameraDock;
}

namespace CAMERA_DOCK
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macro Definitions
//--------------------------------------------------------------------

//! An enumeration of the various columns of data we'll support in our table.
enum CD_COLUMN_ID
{
    CD_USER_NAME = 0,
    CD_MAC_ADDR,
    CD_IP_ADDR,
    CD_SUBNET_MASK,
    CD_DEFAULT_GW,
    CD_VENDOR_ID,
    CD_MODEL,
    CD_SERIAL_NUM,
    CD_MANUFACTURER_INFO,
    CD_SOFTWARE_VER,
    CD_PROTOCOL_VER,
    CD_BOOTLOADER_VER,
    CD_FPGA_VER,
    CD_SENSOR_VER,
    CD_LIST_INDEX,                  //!< Non-visible index back into the camera_list array.
    CD_COL_COUNT
};

//! The number of times to ignore scans before we start looking for a camera again.
const int   SCAN_SKIP_COUNT             = 10;

}   // End of namespace CAMERA_DOCK.

//--------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------

class QActionGroup;
class QHostAddress;
class QListView;
class QTableView;
class QStandardItemModel;

//--------------------------------------------------------------------
// The class definition
//--------------------------------------------------------------------
class CameraDock : public QDockWidget
{
    Q_OBJECT

public:
    explicit CameraDock(QWidget *parent = nullptr);
    ~CameraDock();

    FI_COMM::FI_CAMERA_INFO *get_info_by_row(int sel_row = -1) const;
    int get_selected() const;

    void set_camera_autoconnect(uint32_t ip_addr);
    void set_camera_offline(uint32_t ip_addr);
    void set_discovery_check(bool check_for_cameras);

    static const QActionGroup *get_arrange_group();

signals:
    void    search_complete();
    void    update_cam_info();

public slots:
    void    do_multicam();
    void    open_selected_camera();

protected:
    virtual void changeEvent(QEvent *event);
    virtual void contextMenuEvent(QContextMenuEvent *event_ptr);
    virtual void timerEvent(QTimerEvent *event_ptr);

private slots:
    void    arrange_clicked(QObject *obj_ptr);
    void    delay_camera_open();
    void    do_add_column();
    void    do_clear_entry();
    void    do_clear_all_entries();
    void    do_dock();
    void    do_factory_console();
    void    do_force_ip();
    void    do_hdr_clicked(int index);
    void    do_interface_changed(int new_index);
    void    do_item_selected(QModelIndex index);
    void    do_reset();
    void    do_set_packet_delay();
    void    do_update_data();
    void    finish_setup();
    void    handle_msg_timeout(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    void    refresh_cameras();

    void    read_settings();
    void    write_settings();

    void    scroll_changed(int newValue);

private:
    void    init_connects() const;
    void    init_context_menu();
    void    init_elements();
    void    init_network();

    void    handle_scan_event();
    void    hdr_change(int index, bool toggle_sort = true);

    void    add_row(int index, const FI_COMM::FI_CAMERA_INFO *cam_info);
    int     find_in_table(uint32_t ip_addr);
    int     get_list_index(int sel_row) const;
    void    insert_camera_info(const FI_COMM::FI_CAMERA_INFO *cam_info);
    void    open_camera(int index);
    bool    our_camera(const FI_COMM::FI_CAMERA_INFO *cam_info);
    void    process_net_changes(bool do_dhcp, QHostAddress& ip_addr, QHostAddress& subnet_mask, QHostAddress& gw_addr, const uint8_t *mac_addr);
    void    reset_camera_info();
    void    search_for_cameras(bool reset_table = true);
    void    translate_ui();
    void    update_display();

private:
    Ui::CameraDock  *ui;

    QStandardItemModel  *camera_data;               //!< Camera data stored which the view uses.
    QList<FI_COMM::FI_CAMERA_INFO *> camera_list;   //!< List of found cameras.
    Qt::SortOrder   sort_order;                     //!< Sort order for the table of found cameras.

    uint32_t    adapter_ip;                     //!< The adapter's IP address for the interface we're searching on.
    int         delay_index;                    //!< Index of the camera we want to open in delay_camera_open().
    bool        ignore_update;
    int         last_col;                       //!< Last column in the header, which was clicked.
    int         last_hscroll_pos;               //!< The save horizontal scrollbar position.
    int         last_vscroll_pos;               //!< The save vertical scrollbar position.
    uint32_t    net_mask;                       //!< The net mask for the interface we're searching on.
    bool        pause_discovery;                //!< Pause discovery(scan_time_id) requests.
    int         refresh_index;
    bool        restore_focus;                  //!< True if we're supposed to restore focus to the table or list view.
    int         scan_timer_id;                  //!< Timer to update our scanning results.
    uint32_t    selected_ip;                    //!< The IP address of the currently selected camera.
    int         ticks_waited;                   //!< The number of ticks we've spent waiting for responses.
    int         update_index;                   //!< Index into our list of cameras we are updating.
    int         pending_index;                  //!< Waiting for a response from this camera.

    // Context menu actions
    static QMenu           *arrange_menu;       //!< Our "Arrange By" sub menu.
    static QList<QAction *> arrange_act;        //!< Our arrange actions.
    static QActionGroup    *arrange_group;      //!< Group our arrange actions so they behave together properly.

    // The strings used for the settings values.
    static const QString CAMERA_DOCK_GROUP_NAME;
    static const QString CAMERA_DOCK_NET_IF;
    static const QString CAMERA_DOCK_SORT_COL;
    static const QString CAMERA_DOCK_SORT_ORDER;
    static const QString CAMERA_DOCK_TAB_DETAIL_STATE;
    static const QString CAMERA_DOCK_STREAM_ON_CONNECT;
};

#endif // CAMERA_DOCK_H
