/*!
 * @file camera_explore.h
 * @brief This file contains the definitions for the CamExplore class
 *        which is used to explore the files stored on the camera's media.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_EXPLORE_H
#define CAMERA_EXPLORE_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QDialog>
#include <QFile>
#include <QIcon>
#include <QJsonArray>
#include <QJsonObject>
#include <QModelIndex>
#include <QTreeWidgetItem>

#include <memory>

#include <fi_saver.h>

#include <stdint.h>

namespace Ui
{
class CamExplore;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QTimerEvent;

class fiLoadCaptureDlg;
class fiProgressDialog;
class CameraMDI;
class SaveDialog;

//--------------------------------------------------------------------
// Namespace: CAM_EXPLORE
//--------------------------------------------------------------------
namespace CAM_EXPLORE
{
class FiTreeWidgetItem;

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The indecies into our tree widget.
enum
{
    COLUMN_TITLE    = 0,        //!< The name of the row: ROW_STILLS, ROW_STACKS, ROW_VIDEO, ROW_CAPTURE.
    COLUMN_NAME,                //!< The name of the file/folder.
    COLUMN_SIZE,                //!< The size of the file/folder.
    COLUMN_TIME                 //!< The time of the file/folder.
};

//! Used to enumerate the rows of the tree structure.
enum
{
    ROW_STILLS,                 //!< The row for still images.
    ROW_STACKS,                 //!< The row for image stacks.
    ROW_VIDEO,                  //!< The row for videos.
    ROW_CAPTURE                 //!< The row for captures.
};

// Structure to hold the data parsed from the file listing XML file downloaded from the camera.
typedef struct FILE_INFO
{
public:
    void clear()    { tree_widget = nullptr; }

    QString     name;           // The name of the file or directory
    QString     time_stamp;     // its timestamp
    QString     size;           // it's size in bytes
    QString     width;          // the width of the images
    QString     height;         // height of the images
    QString     frame_count;    // how many frames are there for a stack directory
    QString     meta_count;     // how many metafiles, with will either be 1, 2 or 3 depending on user configurations.
    QString     file_ext;       // The file extension component of the file.
    QString     fps;            // what frame rate was it recorded at
    QString     bit_mode;       // what bit mode was it recorded at.
    QString     raw;            // is the file or directory raw
    QString     start_frame;    // what is the starting frame number for stacks
    QString     end_frame;      // what is the ending frame number for stacks
    QString     comment;        // User provided comment, if any.
    QString     incomplete;     // Is the video in an incomplete state?  Will be 1 if so, 0 if not.
    QString     start_cap;      // The first CAP file, only applicable for CAP stacks.
    QString     end_cap;        // The last CAP file, only applicable for CAP stacks.
    FiTreeWidgetItem *tree_widget;

} FILE_INFO;

typedef struct ROOT_FOLDER_INFO
{
public:
    void clear()
    {
        target.clear();
        capacity = 0;
        free = 0;
        inodes = 0;
    }

    QString     target;         //!< What directory of the camera is this pointing to.
    qulonglong  capacity;       //!< How much space does the device have in bytes.
    qulonglong  free;           //!< How much space is left in bytes.
    qulonglong  inodes;         //!< How many inodes are available (0 if not applicable).
} ROOT_FOLDER_INFO;

typedef QVector<CAM_EXPLORE::FILE_INFO> LISTING_INFO;

/*!
 * @brief Derive a new class from the QTreeWidgetItem so we can sort numbers as numbers
 *        and not text.
 */
class FiTreeWidgetItem : public QTreeWidgetItem
{
public:
    explicit FiTreeWidgetItem(QTreeWidget *parent) : QTreeWidgetItem(parent) {}
    explicit FiTreeWidgetItem(QTreeWidgetItem *parent) : QTreeWidgetItem(parent) {}

protected:
    bool operator<(const QTreeWidgetItem& other) const
    {
        bool less_than;

        // What column are we looking at?
        int column = treeWidget()->sortColumn();
        switch (column)
        {
        case COLUMN_SIZE:
            // We need to compare these as numbers. String compare just won't do it.
            less_than = text(column).toULongLong() < other.text(column).toULongLong();
            break;
        case COLUMN_TIME:
        case COLUMN_NAME:
        case COLUMN_TITLE:
        default:
            // Normal string compare.
            less_than = text(column) < other.text(column);
            break;
        }
        return less_than;
    }
};

}   // End namespace (CAM_EXPLORE)

//--------------------------------------------------------------------
// Class CamExplore
//--------------------------------------------------------------------
class CamExplore : public QDialog
{
    Q_OBJECT

public:
    explicit CamExplore(QWidget *parent = nullptr);
    ~CamExplore();

    static const QString FILE_INFO_INCOMPLETE;
    static const QString FILE_INFO_INCOMPLETE_STR;
    static const QString FILE_INFO_COMPLETE_STR;
    static const QString FILE_DIRECTORY_SELECT_STR;

    static const QString STACKS_DIR_LISTING;
    static const QString VIDEOS_DIR_LISTING;
    static const QString CAPTURES_DIR_LISTING;

    static const QString DEST_SDCARD_STR;
    static const QString DEST_SSD_STR;
    static const QString DEST_USB_STR;

    // Define some common labels to be used whenever you need one.
    static const QString SDCARD_LABEL;
    static const QString SSD_LABEL;
    static const QString USB_LABEL;

    static const QString CAMERA_EXPLORE_GROUP_NAME;
    static const QString LAST_DEVICE;

    static QString generateMetadataString(const CAM_EXPLORE::FILE_INFO *file_info);

    bool loadCapFile(const QString& file_path);

    void setCapFilesLoad(bool load_cap)        { cap_file_load = load_cap; }
    void setCamera(CameraMDI *camera_ptr);
    void setJsonInfo(const QJsonObject& json_info);
    void setMediaType(int sel_type);

    bool capLoadComplete()    { return cap_load_success; }

public slots:
    void doRefresh();

signals:
    void copyProgressValueChange(int progress);
    void loadComplete();
    void progressUpdate(QString progress_info);

protected:
    void customEvent(QEvent *event) override;
    void showEvent(QShowEvent *event) override;
    void timerEvent(QTimerEvent *event) override;

private slots:
    void reject() override;

    void    cameraOffline(uint32_t ip_addr, bool offline);
    void    capStartLocalSave();
    void    capLoadedCamReviewPanelReady();
    void    capLoadedReviewCheck();
    void    capLoadedReviewReady();
    void    capXferStartReady();
    void    capXferFileSaved();
    void    capXferImageLoaded();
    void    copyWorkerFileProgress(int progress, const QString& name);
    void    copyStackProgress(uint32_t bufferSeqNbr);
    void    copyBlobProgress(uint32_t bufferSeqNbr);

    void    doCopy();
    void    doDelete();
    void    doDeviceSelected(int index);
    void    doEject();
    void    doFormat();
    void    doLoad();
    void    doMove();
    void    doOptions(bool clicked);
    void    doTreeExpansion(QModelIndex model_index);
    void    doTreeItemChanged(QTreeWidgetItem *item, QTreeWidgetItem *prev);
    void    doTreeItemDoubleClicked(QTreeWidgetItem *item, int column);
    void    doTreeSelectionChanged();

    void    handleCameraClosing();
    void    handleCopyTimeout();
    void    handleJsonMedia(const QJsonDocument& jsonMedia);
    void    handleProgressCancelledSignal();
    void    handleSaverProgressCancelledSignal();
    void    handleSaverErrorReported();

    void    formatPoll();
    void    loadingFinish();
    void    msgTimeoutEvent(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    void    registerPoll();
    void    readComplete(uint32_t ip_addr, uint32_t reg_addr);
    void    setCapturesAllowed(bool do_allow);
    void    streamPaused();
    void    userCanceledCopy();
    void    writeComplete(uint32_t ip_addr, uint32_t reg_addr);

private:
    bool    buildMediaPath(int root_row);
    QString buildSmbPath(const CAM_EXPLORE::FILE_INFO *file_info, int row);

    bool    capFileTransfer();
    void    capLoadStart(int sel_row);
    void    capXferFinish();
    void    capXferImageSave();
    void    capXferStartCheck();
    void    capXferStart();

    void    cleanupState();
    void    clearFolders();

    void    copyComplete();
    void    copyContinue();
    void    copyFileStart();
    bool    copyLocalSetup(int sel_row, const CAM_EXPLORE::FILE_INFO *file_info);
    bool    copyLocalStart(int sel_row, const CAM_EXPLORE::FILE_INFO *file_info);
    void    copyMetafile();
    void    copyStartAction();
    void    copyStoppedForError();

    void    copyDirect();
    static void copyWorker(QPromise<QString>& promise, QWidget *parent);

    bool    createXferCopyDialog(FI_SAVER::STREAM_TYPE streamType, QString& destValueStr);
    void    deleteFinish();
    void    deleteStart();
    void    displayMetadata(const CAM_EXPLORE::FILE_INFO *file_info);
    int     findDevice(QJsonObject& dev_object);

    void    getDestination();
    void    getDestination(int num_videos);
    const CAM_EXPLORE::FILE_INFO *getFileInfo(const QTreeWidgetItem *item, int& row_out);
    uint32_t     getMetafileCountFromFlag(uint32_t metafile_count_as_flag);
    int     getRootRow(const QTreeWidgetItem *item_ptr);
    void    getSelectionCounts(int& num_roots, int& num_stills, int& num_stacks, int& num_videos, int& num_caps, bool& convertable_found, bool& incomplete_found);
    QString getCurStateMsg();

    void    listingDownloadProcess();
    void    listingFinish();
    void    listingParse();
    void    listingStart(int listing_row);

    void    parseFileActionStatus(uint32_t fa_status);
    uint32_t parseID(const QString& file_path);

    void    stateCapXfer(QEvent *event);
    void    stateCopy(QEvent *event);
    void    stateDelete(QEvent *event);
    void    stateRefresh(QEvent *event);

    void    updateFormatProgress(uint32_t reg_value);
    void    updateMedia(uint32_t reg_value);
    void    userCanceled(const QString& whoCalledMe);
    void    writeFilepathDone(uint32_t write_status);
    void    writeLockDone(uint32_t write_status);

private:
    Ui::CamExplore  *ui;
    std::unique_ptr<fiProgressDialog> xferCopyDialog;   //!< A xfer progress dialog managed internally here.
    std::unique_ptr<fiProgressDialog> progressDialog;   //!< A general progress dialog managed internally here.
    std::unique_ptr<SaveDialog> xferCapDialog;          //!< A Save Dialog used for transcoding CAP data.
    std::unique_ptr<fiLoadCaptureDlg> loadCaptureDlg;   //!< A load capture file progress dialog only one can exist at a time.

    static const QString CAPTURES_BASE_PATH;            //!< Base path for the capture files.
    static const QString IGNORE_FILENAME;
    static const QString COPY_SRC_RECEIVE_FRAME;

    CameraMDI   *camera_ptr;                            //!< The camera object we're working with.
    uint32_t    cam_ip_addr;                            //!< Read this once and keep it around.
    volatile bool cancel_in_progress;                   //!< Flag to let us know we're in the middle of a cancel.
    bool        cap_copy_in_progress;                   //!< Flag to indicate if a CAP file copy is in progress.
    bool        cap_file_load;                          //!< True if we're only interested in loading CAP files.
    int         cap_load_sel_row;                       //!< Which selection row for the CAP being loaded.
    bool        cap_load_success;                       //!< True if a CAP file has successfully been loaded into the camera.
    bool        camera_review_panel_ready;              //!< True if camera that loaded CAP considers the review panel ready.
    bool        captures_allowed;                       //!< True if the camera supports CAP files.
    bool        cleanup_in_progress;                    //!< Flag to let us know we're in the middle of a cleanup.
    int         copy_destination;                       //!< Target media for our copy operation.
    QString     copy_dest_path;                         //!< Destination path requested by the user.
    uint32_t    copyMetafileCount;                      //!< How many metafiles are associated with current copy?
    uint32_t    copyMetafileId;                         //!< The current metafile (zero-based) of copyMetaFileCount being processed.
    QMap<uint32_t, uint32_t> copyMetafileMap;           //!< The currentMetafileId->FISAVER::METADATA_* map.
    int         cur_state;                              //!< The currently active state, if any.
    QString     download_base_path;                     //!< The local root path for the downloads.
    bool        force_update;                           //!< True if an update of the selected device is required.
    uint32_t    format_cmd_status;                      //!< The status back from the camera when we try to start a format.
    bool        format_in_progress;                     //!< Formatting of drive in progress.
    QString     last_device;                            //!< The last selected device.
    uint32_t    listing_len;                            //!< The number of bytes in the listing from the camera.
    QJsonObject media_info;                             //!< May be NULL, the JSON info for the media selected.
    QString     media_path;                             //!< Camera path for the folder we want a listing from.
    int         media_type;                             //!< The type of media we're working on: SDCARD, SSD, USB.
    bool        move_in_progress;                       //!< Flag to indicate if we're doing a move operation.
    int         refresh_row;                            //!< The row the refresh logic is currently working on.
    bool        video_was_running;                      //!< Flag to remember the state of the video stream.
    bool        we_own_lock;                            //!< True if we have the camera locked, false otherwise.
    QFile       listing_file;                           //!< File name for the XML directory listing from the camera.
    int         blob_count;                             //!< Number of "blobs" required to transfer a file.
    bool        rcv_retrying;                           //!< True when we are transitioned to a retry frame state.
    int         rcv_count;                              //!< Count of the number of buffers we've received.
    int         rcv_metadata_count;                     //!< Count of the number of metadata buffers we have received.
    int         rcv_expected;                           //!< What is the expected buffer count.
    int         retry_count;                            //!< Retry counter.
    uint32_t    next_metafile;                          //!< Index of the metafile to be transferred.
    QString     last_metafile_ext;                      //!< Since metafiles can be in any order we cannot rely on a sequence anymore.
    int         wait_for_state_count;                   //!< Number of times we've timed out waiting for a state transition.

    int         file_timer_id;                          //!< Timer used when doing a file operation to update the progress dialog.
    int         live_timer_id;                          //!< Timer used to make sure the camera goes into live mode.
    int         review_timer_id;                        //!< Timer used to make sure the camera goes into review mode.

    bool        enable_merge;                           //!< The user specified to enable overlays when transferring CAP files.

    QIcon       capture_icon;                           //!< Icon used for capture files.
    QIcon       folder_icon;                            //!< Icon used for folders.
    QIcon       still_icon;                             //!< Icon used for still images.
    QIcon       stack_icon;                             //!< Icon used for stacks.
    QIcon       video_icon;                             //!< Icon used for videos.

    std::unique_ptr<QFutureWatcher<QString>> copy_worker_thread;    //!< Our worker thread "handle" for the copy process.

    QList<CAM_EXPLORE::FiTreeWidgetItem *> root_items;  //!< The root widget items for each of the groups
    QList<QTreeWidgetItem *> sel_items;                 //!< List of selected items, used for operations like copy, delete & move.

    CAM_EXPLORE::ROOT_FOLDER_INFO   root_folder;        //!< The root folder type stores capacity information about the root folder
    QList<CAM_EXPLORE::LISTING_INFO *>  cam_folders;    //!< Collection to hold the folder data parsed from the XML file
    const CAM_EXPLORE::FILE_INFO    *file_info;         //!< The information on the entry currently being operated on.
};

#endif // CAMERA_EXPLORE_H
