/*!
 * @file camera_info.cpp
 * @brief This file contains the implementation of the CameraInfo
 *        class. This object is responsible for showing some of the
 *        camera specific information to the user.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "camera_info.h"
#include "ui_camera_info.h"

#include <QByteArray>
#include <QRegularExpression>
#include <QRegularExpressionValidator>
#include <QVector>
#include <QWidget>

#include "camera_mdi.h"
#include "camera_video.h"
#include "client_math.h"
#include "fi_rectitem.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"

namespace CAMERA_INFO_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The default framerate we'll support.
const int   DEFAULT_FRAMERATE               = 20;

//! The minimum framerate we'll support.
const int   MIN_FRAMERATE                   = 1;

//! The per-frame metadata IRIG decode lock bit.
const uint32_t  IRIG_DECODE_LOCK            = (1U << 31);

// The per-frame metadata SYNC MODE values.
const uint32_t  TS3_SYNC_MODE_SHIFT         = 24;
const uint32_t  TS3_SYNC_MODE_MASK          = (3 << TS3_SYNC_MODE_SHIFT);
enum
{
    TS3_SYNC_MODE_MASTER = 0,
    TS3_SYNC_MODE_1PPF,
    TS3_SYNC_MODE_1PPS,
};

// The per-frame metadata TIME STATE values.
const uint32_t  TS3_TIME_STATE_SHIFT        = 20;
const uint32_t  TS3_TIME_STATE_MASK         = (7 << TS3_TIME_STATE_SHIFT);
const uint32_t  TS3_TIME_STATE_IRIG_MIN     = 4;

const uint32_t  MARKERS_MASK                = 0xFFFF;
const int       MARKERS_EDGE_SHIFT          = 8;
const int       MARKERS_COUNT               = 8;

const QString green_style   = QString::fromUtf8("QProgressBar {text-align: center;}::chunk{background-color: green;width 10px;}");
const QString red_style     = QString::fromUtf8("QProgressBar {text-align: center;}::chunk{background-color: red;width 10px;}");

}   // End of namespace CAMERA_INFO_SETTINGS.

// #define DUMP_PFMD 1
using namespace CAMERA_INFO_SETTINGS;

/*!
 * @brief Our constructor
 */
CameraInfo::CameraInfo(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::CameraInfo)
{
    ui->setupUi(this);

    // Setup the battery display.
    ui->lbl_charging->setVisible(false);
    ui->lbl_charging->keepAspect(true);
    ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/batteryNotThere.png")));

    // By default, hide the update button.
    ui->pb_update->setVisible(false);
    connect(ui->pb_update, &QPushButton::clicked, this, &CameraInfo::update_clicked);

    // Initialize our members.
    player_mode = false;
    vt_flags = 0;
    vt_area_change = 0;
    vt_area_percent = 0.0;

    // Read our settings.
    QSettings   settings;
    settings.beginGroup("CameraInfo");
    irig_enable = settings.value("showIRIG", false).toBool();
    local_frametime = settings.value("showLocalFrameTime", false).toBool();
    pfmd_enable = settings.value("showPerFrameMetadata", false).toBool();

    //QLOG_INFO() << __PRETTY_FUNCTION__ << "[CONSTRUCT pfmd_enabled][" << pfmd_enable;

    summary_enable = settings.value("showSummary", true).toBool();
    settings.endGroup();

    // Hide these controls until they're needed.
    ui->chk_irig_locked->setVisible(irig_enable);
    ui->frm_pfmd->setVisible(false);
    ui->progress_autosave->setVisible(false);
    ui->progress_save->setVisible(false);
    ui->gain_frame->setVisible(false);

    // Assume we are NOT in a group right away.
    ui->cb_grouped->setChecked(false);
    ui->cb_grouped->setVisible(false);

    // Restore the summary controls based on the last settings.
    ui->frm_summary->setVisible(summary_enable);

    // Initialize other controls.
    ui->spin_framerate->setMinimum(MIN_FRAMERATE);
    ui->spin_framerate->setMaximum(DEFAULT_FRAMERATE);
    ui->spin_framerate->setValue(DEFAULT_FRAMERATE);

    // Make sure the new camera name conforms to standard host naming rules as per RFC 1123,
    // which are, basically, a name can contain a-z, A-Z, 0-9 and '-' but '-' can't be the
    // first character.
    QRegularExpression ValidHostnameRegex("^(([a-zA-Z0-9]|[a-zA-Z0-9][a-zA-Z0-9\\-]*[a-zA-Z0-9])\\.)*([A-Za-z0-9]|[A-Za-z0-9][A-Za-z0-9\\-]*[A-Za-z0-9])$");
    QRegularExpressionValidator *validator = new QRegularExpressionValidator(ValidHostnameRegex, ui->txt_cam_name);
    ui->txt_cam_name->setValidator(validator);

    // Resize our widget to meet the needs of our controls.
    resize(geometry().width(), ui->frm_info->geometry().height());
    setMinimumHeight(ui->frm_info->geometry().height());

    connect(ui->spin_framerate, SIGNAL(valueChanged(int)), this, SIGNAL(framerate_changed(int)));
    connect(ui->tb_streaming, SIGNAL(clicked()), this, SIGNAL(stream_toggle()));
    connect(ui->txt_cam_name, SIGNAL(editingFinished()), this, SIGNAL(camera_name_changed()));

    connect(ui->tb_zoom_in, SIGNAL(clicked()), MainWindow::GetInstance().get(), SIGNAL(zoom_in()));
    connect(ui->tb_zoom_out, SIGNAL(clicked()), MainWindow::GetInstance().get(), SIGNAL(zoom_out()));

    pfmd_file_exists            = false;
    metadata_txt_file_exists    = false;
}

/*!
 * @brief Our destructor.
 */
CameraInfo::~CameraInfo()
{
    // Clean up the UI object.
    delete ui;
}

/*!
 * @brief This method parses the "autosave_value" parameter and displays/updates the
 *        UI elements associated with the autosave functionality.
 *
 * @param autosave_value    The value of the GIGE_REG_AUTOSAVE register.
 * @param fflr_recovery     If camera is FasFireLR recovery the GIGE_AUTOSAVE_ENABLE may be false...
 */
void CameraInfo::display_autosave_settings(uint32_t autosave_value)
{
    bool    enabled;
    uint32_t    dirty_count;
    uint32_t    avail_count;
    uint32_t    total_count;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    bool        fflr_is_enabled = false;
    uint32_t    fflr_type;

    // Determine if we're in FFLR mode.
    if (my_parent)
    {
        fflr_is_enabled = my_parent->is_fflr_enabled(fflr_type);
    }

    // Parse out the fields we're interested in.
    enabled = (autosave_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE;

    avail_count = GIGE_AUTOSAVE_AVAIL_COUNT(autosave_value);
    total_count = GIGE_AUTOSAVE_TOTAL_COUNT(autosave_value);
    dirty_count = total_count - avail_count;

//    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << my_parent->get_ip_addr_pretty() << "enabled:" << enabled << "avail_count:" << avail_count << "total_count:" << total_count;

    // Update the autosave enable check.
    ui->cam_actions->set_autosave_checked(enabled);

    // For FasFireLR make sure we are showing the buffer information.
    if (fflr_is_enabled)
    {
        enabled = fflr_is_enabled;
    }

    // Update our autosave control.
    if (enabled || (total_count > 1))
    {
        ui->progress_autosave->setMinimum(0);
        ui->progress_autosave->setMaximum((int)total_count);
        ui->progress_autosave->setValue((int)avail_count);

        if (avail_count > 1)
            ui->progress_autosave->setStyleSheet(green_style);
        else
            ui->progress_autosave->setStyleSheet(red_style);

        // If we're not enabled we MIGHT want to enable the display if we have multiple dirty buffers.
        if (!enabled)
            enabled = (dirty_count > 1);
    }
    else
    {
        // Don't bother showing the controls.
        enabled = false;
    }

    // Only show it if autosave is enabled.
    ui->progress_autosave->setVisible(enabled);

    // Also, if autosave is not enabled, hide the save progress control.
    if (!enabled)
    {
        ui->progress_save->setVisible(false);
    }

    // Update our buffer settings.
    if (enabled || ((total_count > 1) && (dirty_count > 0)))
        set_buffer_info((int)dirty_count, (int)total_count);
    else
        set_buffer_info(0, 0);
}

/*!
 * @brief Display the current state of the battery.
 * @param battery_state     The value of GIGE_REG_BATTERY_CHARGE.
 */
void CameraInfo::display_battery_state(uint32_t battery_state)
{
    int         charge_percent;
    bool        show_ctrl(true);
    uint32_t    charge_state;
    QString     tool_tip_str;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
        return;

    // Parse out the fields from the register value.
    charge_state = battery_state & GIGE_BATTERY_STATE_MASK;
    charge_percent = (int)((battery_state >> GIGE_BATTERY_PERCENT_SHIFT) & GIGE_BATTERY_PERCENT_MASK);

    // Process based on the state.
    switch (charge_state)
    {
    case GIGE_BATTERY_NO_BATTERY:
    case GIGE_BATTERY_NO_INFO:
        // If we are a seriesNC camera (not series3) and there is no battery, then
        // make sure we do not show the control.
        if (!my_parent->is_series3())
            show_ctrl = false;

        // Series3 cameras understand batteries, even when they are disconnected.
        ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/batteryNotThere.png")));
        tool_tip_str = tr("The Battery is Disconnected");
        break;
    case GIGE_BATTERY_CHARGING:
        tool_tip_str = QString(tr("Battery Reads %1%")).arg(charge_percent);
        if (charge_percent <= 5)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery5Charging.png")));
        else if (charge_percent <= 15)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery6to15Charging.png")));
        else if (charge_percent <= 50)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery25to50Charging.png")));
        else if (charge_percent <= 60)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery50Charging.png")));
        else if (charge_percent <= 80)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery75Charging.png")));
        else
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery100Charging.png")));
        break;
    default:
        tool_tip_str = QString(tr("Battery Reads %1%")).arg(charge_percent);
        if (charge_percent <= 5)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery5.png")));
        else if (charge_percent <= 15)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery6to15.png")));
        else if (charge_percent <= 50)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery25to50.png")));
        else if (charge_percent <= 60)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery50.png")));
        else if (charge_percent <= 80)
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery75.png")));
        else
            ui->lbl_charging->setPixmap(QPixmap(QString::fromUtf8(":/images/battery100.png")));
        break;
    }

    // Set the tool tip string.
    ui->lbl_charging->setToolTip(tool_tip_str);

    // Display or hide the symbol.
    ui->lbl_charging->setVisible(show_ctrl);
}

/*!
 * @brief Display the camera name in the txt_cam_name control.
 */
void CameraInfo::display_camera_name()
{
    // Get the camera name. This information arrived as part of the discovery message
    // so just pull it from the camera information structure.
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    const char  *name_ptr;

    // Sanity checks.
    if ((my_parent == nullptr) || (my_parent->get_camera_info() == nullptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No parent or no camera information!";
        return;
    }

    // Try to retrieve the user name first. If that doesn't exist, try to use
    // the model name.
    name_ptr = (const char *)my_parent->get_camera_info()->user_name;
    if (name_ptr[0] == '\0')
        name_ptr = (const char *)my_parent->get_camera_info()->model_name;
    if (name_ptr[0] != '\0')
        ui->txt_cam_name->setText(name_ptr);
}

/*!
 * @brief Display the camera state.
 */
void CameraInfo::display_camera_state(uint32_t state)
{
    bool    streaming_enable(false);

    switch (state)
    {
    case GIGE_CAMERA_STATE_IDLE:
        ui->lbl_cam_mode->setText(tr("Idle"));
        break;
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        ui->lbl_cam_mode->setText(tr("Live Video"));
        streaming_enable = true;
        break;
    case GIGE_CAMERA_STATE_ARMED:
        ui->lbl_cam_mode->setText(tr("Armed"));
        streaming_enable = true;
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        ui->lbl_cam_mode->setText(tr("Triggered"));
        streaming_enable = true;
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        ui->lbl_cam_mode->setText(tr("Video Review"));
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
        ui->lbl_cam_mode->setText(tr("Snapshot"));
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        ui->lbl_cam_mode->setText(tr("Autosave"));
        break;
    case GIGE_CAMERA_STATE_DEARMING:
        ui->lbl_cam_mode->setText(tr("Dearming"));
        break;
    case GIGE_CAMERA_STATE_SHUTDOWN:
        ui->lbl_cam_mode->setText(tr("Shutdown"));
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected camera state:" << state;
        break;
    }

    // Enable/disable the streaming button.
    ui->tb_streaming->setEnabled(streaming_enable);
}

/*!
 * @brief Update the autosave progress indicator.
 * @param save_progress     A value of 0..100, indicating the save progress.
 */
void CameraInfo::display_save_progress(uint32_t save_progress)
{
    if (save_progress > 0)
    {
        ui->progress_save->setVisible(true);
        ui->progress_save->setValue((int)save_progress);
    }
    else
    {
        ui->progress_save->setVisible(false);
    }
}

/*!
 * @brief Disable the state actions (Live, Arm, etc.).
 */
void CameraInfo::disable_actions()
{
    // Disable the action buttons.
    ui->cam_actions->disable_actions();
}

/*!
 * @brief Enable the state actions (Live, Arm, etc.).
 */
void CameraInfo::enable_actions()
{
    // Enable the action buttons.
    ui->cam_actions->enable_actions();
}

/*!
 * @brief Enable/disable the IRIG specific UI elements. When enabled, the
 *        UI elements are displayed, when disabled they are hidden.
 *
 * @param do_enable     True to enable (show), false to disable (hide).
 */
void CameraInfo::enable_irig(bool do_enable)
{
    // Remember the new setting.
    irig_enable = do_enable;

    // Save our settings.
    QSettings   settings;
    settings.beginGroup("CameraInfo");
    settings.setValue("showIRIG", irig_enable);
    settings.endGroup();
}

/*!
 * @brief Enable/disable the display of time using a local format.
 *
 * @param do_enable     True to enable the displaying time in a local format.
 */
void CameraInfo::enable_local_frametime(bool do_enable)
{
    // Remember the new setting.
    local_frametime = do_enable;

    // Save our settings.
    QSettings   settings;
    settings.beginGroup("CameraInfo");
    settings.setValue("showLocalFrameTime", local_frametime);
    settings.endGroup();
}


/*!
 * @brief Enable the per-frame metadata information. Also, if enabled, enable the IRIG
 *        specific information.
 *
 * @param do_enable         True to enable the per-frame metadata.
 * @param check_settings    True if need to override with CameraInfo.showPerFrameMetadata.
 */
void CameraInfo::enable_pfmd(bool do_enable)
{
    QSettings   settings;

    // Remember the new setting.
    pfmd_enable = do_enable;

    //QLOG_INFO() << __PRETTY_FUNCTION__ << "[NEW SETTING pfmd_enable][" << pfmd_enable << "][do_enable][" << do_enable << "]";

    // Save our settings.
    settings.beginGroup("CameraInfo");
    settings.setValue("showPerFrameMetadata", pfmd_enable);
    settings.endGroup();
}

/*!
 * @brief Place the CameraReview object into "player mode". That means there
 *        is no camera attached because we're playing a saved video.
 */
void CameraInfo::enable_player_mode()
{
    // Remember we're in player mode.
    player_mode = true;

    // Hide the camera info controls.
    ui->frm_info->setVisible(false);
}

/*!
 * @brief Enable or disable the stream pause/resume control
 * @param do_enable     True to enable the control, false to disable it.
 */
void CameraInfo::enable_stream(bool do_enable)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Update the button look.
    update_stream_btn();

    // You can enable/disable the button if the state allows the button
    // to be enabled. A state that would disable the button, keeps the button
    // in a disabled state.
    if (do_enable && (my_parent != nullptr))
    {
        uint32_t    camera_state;

        camera_state = my_parent->read_camera_cache(GIGE_REG_CAMERA_STATE);
        switch (camera_state)
        {
        case GIGE_CAMERA_STATE_IDLE:
        case GIGE_CAMERA_STATE_REVIEW:
        case GIGE_CAMERA_STATE_SNAPSHOT:
        case GIGE_CAMERA_STATE_AUTOSAVE:
        case GIGE_CAMERA_STATE_DEARMING:
        case GIGE_CAMERA_STATE_SHUTDOWN:
            // The button is always disabled in these states.
            do_enable = false;
            break;
        case GIGE_CAMERA_STATE_LIVE_VIDEO:
        case GIGE_CAMERA_STATE_ARMED:
        case GIGE_CAMERA_STATE_TRIGGERED:
            // Allow the client to enable/disable the button.
            break;
        default:
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected camera state:" << camera_state;
            do_enable = false;
            break;
        }
    }

    // Update our button.
    ui->tb_streaming->setEnabled(do_enable);
}

/*!
 * @brief Enable the summary information.
 *
 * @param do_enable     True to enable the summary display.
 */
void CameraInfo::enable_summary(bool do_enable)
{
    // Remember the new setting.
    summary_enable = do_enable;
    ui->frm_summary->setVisible(summary_enable);
    adjustSize();

    // Save our settings.
    QSettings   settings;
    settings.beginGroup("CameraInfo");
    settings.setValue("showSummary", summary_enable);
    settings.endGroup();
}

void CameraInfo::enable_update(bool do_enable)
{
    ui->pb_update->setVisible(do_enable);
    ui->pb_update->setEnabled(do_enable);
}

/*!
 * @brief Return a list of actions (Arm, Trigger, etc.) for the camera.
 */
QList<QAction *> CameraInfo::get_camera_actions()
{
    return ui->cam_actions->actions();
}

/*!
 * @brief Return the camera name being displayed. This is possibly a new name which
 *        the user has entered.
 */
QString CameraInfo::get_camera_name()
{
    return ui->txt_cam_name->text();
}

QString CameraInfo::get_exposure()
{
    return ui->exp_value->text();
}

int CameraInfo::get_fps()
{
    return ui->fps_value->text().toInt();
}

/*!
 * @brief Return the framerate being displayed. This is the "display" framerate - the number
 *        of frames being sent from the camera per second for us to display.
 */
int CameraInfo::get_framerate()
{
    return ui->spin_framerate->value();
}

QString CameraInfo::get_gain()
{
    return ui->gain_value->text();
}

QString CameraInfo::get_markers()
{
    return ui->markers_value->text();
}

/*!
 * @brief Get the pfmd camera info settings if it exists.
 */
bool CameraInfo::get_pfmd_setting()
{
    bool pfmd_setting;

    QSettings   settings;
    settings.beginGroup("CameraInfo");
    pfmd_setting = settings.value("showPerFrameMetadata", pfmd_enable).toBool();
    settings.endGroup();

    //QLOG_INFO() << __PRETTY_FUNCTION__ << "[GET pfmd_setting][" << pfmd_setting << "]";

    return pfmd_setting;
}

QString CameraInfo::get_roi()
{
    return ui->roi_value->text();
}
QRect CameraInfo::get_roi_rect() const
{
    return roi_rect;
}

QString CameraInfo::get_sync()
{
    return ui->sync_value->text();
}

QString CameraInfo::get_time()
{
    return ui->time_value->text();
}

QString CameraInfo::get_time_source()
{
    return ui->time_source_value->text();
}

QString CameraInfo::get_trig_pos()
{
    return ui->trig_value->text();
}

/*!
 * @brief Return true if the camera is in a group, false otherwise.
 */
bool CameraInfo::is_grouped()
{
    return ui->cb_grouped->isChecked();
}

/*!
 * @brief Update the user interface to reflect the current camera state.
 */
void CameraInfo::reflect_camera_state(uint32_t camera_state, uint32_t part_state, uint32_t ltr_mode)
{
    // Pass it on to the control.
    ui->cam_actions->reflect_camera_state(camera_state, part_state, ltr_mode);
}

/*!
 * @brief - Convert a raw sbs time to a struct tm structure.
 *
 * @param sbs - the sbs time gotten from the camera.
 * @param my_tm - the tm structure
 *
 * @return none
 */
void CameraInfo::sbs_to_tm(uint32_t sbs, struct tm& my_tm)
{
    // Clear out the structure.
    memset(&my_tm, 0, sizeof(struct tm));

    my_tm.tm_year = sbs / (365 * 24 * 60 * 60);
    sbs -= (uint32_t)(my_tm.tm_year * (365 * 24 * 60 * 60));
    my_tm.tm_yday = sbs / (24 * 60 * 60);
    sbs -= (uint32_t)(my_tm.tm_yday * (24 * 60 * 60));
    my_tm.tm_hour = sbs / (60 * 60);
    sbs -= (uint32_t)(my_tm.tm_hour * (60 * 60));
    my_tm.tm_min = sbs / 60;
    my_tm.tm_sec = (int)sbs - (my_tm.tm_min * 60);
}

/*!
 * @brief Set the number of used and the total number of images buffers on the camera.
 *
 * @param used      The number of used buffers.
 * @param total     The total number of buffers.
 */
void CameraInfo::set_buffer_info(int used, int total)
{
    // Show or hide based on the total number of buffers.
    ui->buf_info_frame->setVisible(total > 1);

    // Build up our string.
    ui->buf_value->setText(QString("%1/%2").arg(used).arg(total));
}

/*!
 * @brief Set or clear the summary "grouped" checkbox.
 * @param grouped   True to check, false to clear, the "grouped" checkbox.
 */
void CameraInfo::set_grouped(bool grouped)
{
    // If we're grouped, check the box and show it. Otherwise, hide the control.
    ui->cb_grouped->setChecked(grouped);
    ui->cb_grouped->setVisible(grouped);
    update();
}

/*!
 * @brief Set the flags for external supporting metadata files associated
 * with this "camera".  It may be a video loaded in a camera window so it will
 * have this information.
 *
 * @param has_pfmd_file Flag is true if there is a .pm file for this "camera".
 * @param has_txt_file Flag is true if there is a .txt file for this "camera".
 */
void CameraInfo::set_metadata_file_info(bool has_pfmd_file, bool has_txt_file)
{
    pfmd_file_exists            = has_pfmd_file;
    metadata_txt_file_exists    = has_txt_file;
}

/*!
 * @brief Parse the raw metadata information into something that will make
 *        sense to the user.
 *
 * @param metadata_ptr      The raw per-frame metadata.
 */
void CameraInfo::set_pfmd(const QByteArray *metadata_ptr)
{
    int         index;
    QString value_str;
    uint32_t    markers;
    uint32_t    time_state;
    struct tm   sbs_tm;
    QLatin1Char zero_fill = QLatin1Char('0');
    uint32_t pfmd_info[8];

    // Copy our PFMD data.
    memcpy(pfmd_info, metadata_ptr->constData(), sizeof(pfmd_info));

#ifdef DUMP_PFMD
    {
        QString msg_str;
        msg_str = tr("%1 %2 %3 %4").arg(pfmd_info[0], 8, 16, QChar('0')).arg(pfmd_info[1], 8, 16, QChar('0')).arg(pfmd_info[2], 8, 16, QChar('0')).arg(pfmd_info[3], 8, 16, QChar('0'));
        QLOG_TRACE() << __PRETTY_FUNCTION__ << msg_str;
        msg_str = tr("%1 %2 %3 %4").arg(pfmd_info[4], 8, 16, QChar('0')).arg(pfmd_info[5], 8, 16, QChar('0')).arg(pfmd_info[6], 8, 16, QChar('0')).arg(pfmd_info[7], 8, 16, QChar('0'));
        QLOG_TRACE() << __PRETTY_FUNCTION__ << msg_str;
    }
#endif

    // RECHECK_JLC
    // Is this time value zero?  It shouldn't ever be zero...
    if (pfmd_info[0] == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ZERO TIME PROVIDED in PER FRAME METADATA!!!";
    }

    // Convert the time.
    sbs_to_tm(pfmd_info[0], sbs_tm);

    // Build up our time string.
    if (local_frametime)
    {
        // Get the date in a Date object so it'll do our leap year calculations.
        QDate   frameDate(sbs_tm.tm_year + 2000, 1, 1);
        frameDate = frameDate.addDays(sbs_tm.tm_yday);

        // Display according to the system's short date format and a normal hour:min:sec.microsecond.
        value_str = frameDate.toString(QLocale::system().dateFormat(QLocale::ShortFormat)) +
                    QString(" %1:%2:%3.%4").arg(sbs_tm.tm_hour, 2, 10, zero_fill)
                    .arg(sbs_tm.tm_min, 2, 10, zero_fill).arg(sbs_tm.tm_sec, 2, 10, zero_fill)
                    .arg(pfmd_info[1] & 0xFFFFF, 6, 10, zero_fill);
    }
    else
    {
        // Display in Fastec "standard".
        value_str = tr("%1:%2:%3:%4:%5.%6")
                    .arg(sbs_tm.tm_year, 2, 10, zero_fill).arg(sbs_tm.tm_yday + 1, 3, 10, zero_fill).arg(sbs_tm.tm_hour, 2, 10, zero_fill)
                    .arg(sbs_tm.tm_min, 2, 10, zero_fill).arg(sbs_tm.tm_sec, 2, 10, zero_fill)
                    .arg(pfmd_info[1] & 0xFFFFF, 6, 10, zero_fill);
    }
    ui->time_value->setText(value_str);

    // Are we locked to IRIG?
    ui->chk_irig_locked->setChecked((pfmd_info[1] & IRIG_DECODE_LOCK) == (uint32_t)IRIG_DECODE_LOCK);

    // Build up our string for the sync mode.
    switch ((pfmd_info[1] & TS3_SYNC_MODE_MASK) >> TS3_SYNC_MODE_SHIFT)
    {
    case TS3_SYNC_MODE_MASTER:
        value_str = tr("Internal");
        break;
    case TS3_SYNC_MODE_1PPF:
        value_str = tr("1PPF");
        break;
    case TS3_SYNC_MODE_1PPS:
        value_str = tr("1PPS");
        break;
    default:
        value_str = tr("Unknown");
        break;
    }
    ui->sync_value->setText(value_str);

    // Build up our string for the time state.
    time_state = (pfmd_info[1] & TS3_TIME_STATE_MASK) >> TS3_TIME_STATE_SHIFT;
    if (time_state == 0)
        value_str = tr("Internal");
    else if (time_state < TS3_TIME_STATE_IRIG_MIN)
        value_str = tr("1PPS");
    else
        value_str = tr("IRIG");
    ui->time_source_value->setText(value_str);

    // Save our marker information.
    markers = pfmd_info[2] & MARKERS_MASK;

    // Build up our marker display.
    value_str.clear();
    for (index = (MARKERS_COUNT - 1); index >= 0; --index)
    {
        // Is there an edge event?
        if ((markers & (0x01 << (MARKERS_EDGE_SHIFT + index))) != 0)
        {
            // Got an edge event. Was it ending high or low?
            if ((markers & (0x01 << index)) == 0)
                value_str += QString::fromUtf8("\342\206\223");
            else
                value_str += QString::fromUtf8("\342\206\221");
        }
        else
        {
            // No edge, just the signal value.
            if ((markers & (0x01 << index)) == 0)
                value_str += QString::fromUtf8("0");
            else
                value_str += QString::fromUtf8("1");
        }
    }
    ui->markers_value->setText(value_str);

    // Dump some hardware information to the debug log.
//    if (pfmd_info[7] != 0)
    {
        CameraMDI *my_parent = qobject_cast<CameraMDI *>(parent());

        // Save the number of pixels counter so we can use it a few times.
        vt_area_change = pfmd_info[7] & CAMERA_INFO::VTRIG_ROI_PIXEL_CHANGE_COUNT_MASK;
        vt_flags = (pfmd_info[7] & CAMERA_INFO::VTRIG_FLAGS_MASK) >> CAMERA_INFO::VTRIG_FLAGS_SHIFT;

        // Calculate the area percentage.
        if (my_parent)
        {
            fiRectItem *vt_reticle;

            // Get the reticle.
            vt_reticle = my_parent->get_player()->get_vt_reticle();
            if (vt_reticle != nullptr)
            {
                QRectF trig_reticle;
                QPointF trig_pos;
                QRect trig_rect;

                trig_reticle = vt_reticle->adjrect();
                trig_pos = vt_reticle->pos();

                // Move the rectangle to the right position.
                trig_reticle.moveTopLeft(trig_pos);
                trig_rect = trig_reticle.toAlignedRect();

//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "VideoTrigger:" << trig_rect;
                vt_area_percent = (((float)vt_area_change) / (trig_rect.width() * trig_rect.height())) * 100.0;
            }
            else
            {
                vt_area_percent = 0.0;
            }
        }
        else
        {
            vt_area_percent = 0.0;
        }

#if 0
        // Display our counts.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "HW Frame:" << QString::number(pfmd_info[3])
                     << "Counts:" << QString::number(vt_area_change) << "/" << QString::number(vt_area_change, 16)
                     << "Area Change:" << QString::number(vt_area_percent, 'f', 2) + QStringLiteral("%")
                     << "Flags:" << Qt::showbase << Qt::hex << vt_flags;
#endif
    }
}

/*!
 * @brief Display the camera settings summary.
 *
 * @param width             The width setting for the camera.
 * @param height            The height setting for the camera.
 * @param frame_rate        The current frame rate.
 * @param ss                The shutter speed.
 * @param trig_pos          The trigger position, either percent or actual frame.
 * @param gain              The analog sensor gain being used.
 */
void CameraInfo::set_summary(uint32_t width, uint32_t height, uint32_t frame_rate, uint32_t ss, uint32_t trig_pos, uint32_t gain)
{
    bool frame_based;
    double ss_f;
    QString trig_info;

    // Is the trigger position frame number or percent?
    frame_based = ((trig_pos & GIGE_TRIGGER_REQ_FRAME_MASK) == GIGE_TRIGGER_REQ_FRAME_MASK);
    if (frame_based)
    {
        trig_pos &= GIGE_TRIGGER_REQ_POS_MASK;
        trig_info = QString("%1").arg(trig_pos);
    }
    else
    {
        trig_info = QString("%1%").arg(trig_pos);
    }

    // Get our shutter speed as a floating point number.
    ss_f = (double)ss / 1000.0;

    // Set our values.
    roi_rect.setX(0);
    roi_rect.setY(0);
    roi_rect.setWidth((int)width);
    roi_rect.setHeight((int)height);
    ui->roi_value->setText(QString("%1x%2").arg(width).arg(height));
    ui->fps_value->setText(QString::number(ClientMath::reg_to_frame_rate(frame_rate)));
    ui->exp_value->setText(QString::number(ss_f, 'f', 3));
    ui->trig_value->setText(trig_info);

    // Figure out our sensor gain value.
    if (gain == 0)
    {
        ui->gain_frame->setVisible(false);
    }
    else
    {
        ui->gain_frame->setVisible(true);
        ui->gain_value->setText(QString::number((gain / 100.0), 'f', 1));
    }
}

/*!
 * @brief Show/hide the enabled IRIG or per-frame metadata controls. This
 *        also updates the height of the CameraInfo pane.
 */
void CameraInfo::show_info()
{
    int         new_height;
    bool        ok_to_show(true);
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // We only need to worry about this if we're connected to a camera.
    if (my_parent != nullptr)
    {
        uint32_t    camera_state;

        // Get the parent's state.
        camera_state = my_parent->read_camera_cache(GIGE_REG_CAMERA_STATE);

        // We can only show PFMD in review mode.
        ok_to_show = (camera_state == GIGE_CAMERA_STATE_REVIEW);
    }

    // Do we have the info frame?
    if (ui->frm_info->isVisible())
        new_height = ui->frm_info->geometry().height();
    else
        new_height = 0;

    // Are we enabling or disabling the per-frame metadata view.
    //QLOG_INFO() << __PRETTY_FUNCTION__ << "[pfmd_enable][" << pfmd_enable << "][ok_to_show][" << ok_to_show << "][summary_enable][" << summary_enable << "]";
    if (pfmd_enable)
    {
        // Show or hide as appropriate.
        if (ok_to_show != ui->frm_pfmd->isVisible())
            ui->frm_pfmd->setVisible(ok_to_show);

        // Only worry about the height if we're showing.
        if (ok_to_show)
        {
            // Show/hide the IRIG information based on our flag.
            if (irig_enable != ui->chk_irig_locked->isVisible())
                ui->chk_irig_locked->setVisible(irig_enable);

            if (new_height != 0)
                new_height += ui->verticalLayout->spacing();
            new_height += ui->frm_pfmd->geometry().height();
        }
    }
    else
    {
        // Hide the PFMD controls if they're not already hidden.
        if (ui->frm_pfmd->isVisible())
        {
            ui->frm_pfmd->setVisible(false);
        }
    }

    // Are we enabling our summary line?
    if (summary_enable)
    {
        if (!ui->frm_summary->isVisible())
            ui->frm_summary->setVisible(true);

        if (new_height != 0)
            new_height += ui->verticalLayout->spacing();
        new_height += ui->frm_summary->geometry().height();
    }
    else
    {
        // Hide the summary.
        if (ui->frm_summary->isVisible())
            ui->frm_summary->setVisible(false);
    }

    // Update our geometry.
    if (new_height == 0)
    {
        setVisible(false);
    }
    else
    {
        if (!isVisible())
        {
            adjustSize();
            setVisible(true);
        }
    }
}

/*!
 * @brief Translate the UI elements for this object.
 */
void CameraInfo::translate_ui()
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Have the UI object perform it's translations.
    ui->retranslateUi(this);

    // Update our displays.
    if (my_parent != nullptr)
        display_camera_state(my_parent->read_camera_cache(GIGE_REG_CAMERA_STATE));
}

/*!
 * @brief Update the look of the stream button to be either a "play" or a "pause" button.
 */
void CameraInfo::update_stream_btn()
{
    QString     tool_tip;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity checks.
    if ((my_parent == nullptr) || (my_parent->get_camera_info() == nullptr))
    {
        // We sometimes get here to early. In that case, we'll come back around.
        return;
    }

    // If we're currently streaming, we'll be turning that off.
    if (my_parent->get_camera_info()->stream_running)
    {
        ui->tb_streaming->setIcon(QIcon(QString::fromUtf8(":/images/pause.png")));
        tool_tip = tr("Stop streaming video from the camera");
    }
    else
    {
        ui->tb_streaming->setIcon(QIcon(QString::fromUtf8(":/images/play_forward.png")));
        tool_tip = tr("Start streaming video from the camera");
    }

    // Update the tooltip.
    ui->tb_streaming->setToolTip(tool_tip);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void CameraInfo::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    QWidget::changeEvent(event);
}

/*!
 * @brief Generate a "clicked" signal for the autosave and save progress bars
 *        on a mouse release event.
 *
 * @param event     Pointer to a QMouseEvent structure, this is used to determine
 *                  which, if any, of the controls contained the mouse event.
 */
void CameraInfo::mouseReleaseEvent(QMouseEvent *event)
{
    // Did the user click on the autosave progress?
    if (ui->progress_autosave->geometry().contains(event->position().toPoint()))
    {
        // Let folks know.
        emit autosave_clicked();
    }

    // Did the user click on the save progress?
    else if (ui->progress_save->geometry().contains(event->position().toPoint()))
    {
        emit save_clicked();
    }

    // Pass it on.
    QWidget::mouseReleaseEvent(event);
}

