/*!
 * @file camera_info.h
 * @brief This is the header file for the CameraInfo class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_INFO_H
#define CAMERA_INFO_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <time.h>

#include <QByteArray>
#include <QWidget>

namespace Ui
{
class CameraInfo;
}

namespace CAMERA_INFO
{
//--------------------------------------------------------------------
// Constants
//--------------------------------------------------------------------
const uint32_t VTRIG_FLAGS_SHIFT                    = 20;
const uint32_t VTRIG_FLAGS_MASK                     = (3 << VTRIG_FLAGS_SHIFT);
const uint32_t VTRIG_ROI_PIXEL_CHANGE_COUNT_MASK    = 0x0001FFFF;

// For per-frame metadata, these are the flags we'll see (after shifting).
const uint32_t VTRIG_FLAGS_REFRESH                  = (1 << 0);
const uint32_t VTRIG_FLAGS_TRIGGER                  = (1 << 1);
}

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class CameraInfo : public QWidget
{
    Q_OBJECT

public:
    explicit CameraInfo(QWidget *parent = nullptr);
    ~CameraInfo();

    void    display_autosave_settings(uint32_t autosave_value);
    void    display_battery_state(uint32_t battery_state);
    void    display_camera_name();
    void    display_camera_state(uint32_t state);
    void    display_save_progress(uint32_t save_progress);

    void    disable_actions();
    void    enable_actions();
    void    enable_irig(bool do_enable);
    void    enable_local_frametime(bool do_enable);
    void    enable_pfmd(bool do_enable);
    void    enable_player_mode();
    void    enable_stream(bool do_enable);
    void    enable_summary(bool do_enable);
    void    enable_update(bool do_enable);

    QList<QAction *> get_camera_actions();
    QString get_camera_name();
    QString get_exposure();
    int     get_fps();
    int     get_framerate();
    QString get_gain();
    QString get_markers();
    bool    get_pfmd_setting();

    QString get_roi();
    QRect   get_roi_rect() const;
    QString get_sync();
    QString get_time();
    QString get_time_source();
    QString get_trig_pos();

    int     get_vt_flags()              { return vt_flags; }
    uint32_t get_vt_area_change()       { return vt_area_change; }
    float   get_vt_area_percent()       { return vt_area_percent; }

    bool    irig_is_enabled()           { return irig_enable; }
    bool    is_grouped();
    bool    localtime_is_enabled()      { return local_frametime; }
    bool    pfmd_is_enabled()           { return pfmd_enable; }
    bool    summary_is_enabled()        { return summary_enable; }


    void    reflect_camera_state(uint32_t camera_state, uint32_t part_state, uint32_t ltr_mode);
    void    set_buffer_info(int used, int total);
    void    set_grouped(bool grouped);
    void    set_pfmd(const QByteArray *metadata_ptr);
    void    set_summary(uint32_t width, uint32_t height, uint32_t frame_rate, uint32_t ss, uint32_t trig_pos, uint32_t gain = 0);
    void    set_metadata_file_info(bool has_pfmd_file, bool has_txt_file);
    void    show_info();

signals:
    void    camera_name_changed();
    void    framerate_changed(int new_value);
    void    stream_toggle();

    void    autosave_clicked();
    void    save_clicked();
    void    update_clicked();

protected:
    virtual void changeEvent(QEvent *event);
    virtual void mouseReleaseEvent(QMouseEvent *event);

private:
    void    sbs_to_tm(uint32_t sbs, struct tm& my_tm);
    void    translate_ui();
    void    update_stream_btn();

private:
    Ui::CameraInfo *ui;                 //!< Pointer to our UI object.

    QRect   roi_rect;                   //!< Image ROI.
    bool    irig_enable;                //!< Flag to remember if the IRIG controls are to be shown (true) or hidden (false).
    bool    local_frametime;            //!< Flag to remember if the time is displayed using a local format.
    bool    pfmd_enable;                //!< Flag to remember if the per-frame metadata controls are to be shown (true) or hidden (false).
    bool    player_mode;                //!< Flag to indicate if we're in "player" mode or not.
    bool    summary_enable;             //!< Flag to remember if the camera settings summary is to be shown.

    bool    pfmd_file_exists;           //!< Does a pfmd file exist for this "camera"?
    bool    metadata_txt_file_exists;   //!< Does a partition metadata txt file exist for this "camera"?

    int     vt_flags;                   //!< Video trigger flags from the PFMD.
    uint32_t vt_area_change;            //!< Video trigger area change value from the PFMD.
    float vt_area_percent;             //!< Video trigger area change in percent of total area.
};

#endif // CAMERA_INFO_H
