/*!
 * @file camera_log_search_dlg.cpp
 * @brief This is the source file for the Fastec Camera Log Search Dialog.
 *
 * This class is modelled after the Notepad "Find" dialog to provide
 * primative search functionality for the Camera Log Docking window.
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */


//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include "camera_log_search_dlg.h"
#include "ui_camera_log_search_dlg.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

namespace CAMLOG_SEARCH
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The maximum number of characters to display in a path. Anything
//! longer will be elided.
//const int       MAX_PATH_LEN            = 48;

}   // End of namespace CAMLOG_SEARCH.

CameraLogSearchDialog::CameraLogSearchDialog(QWidget *parent, Qt::WindowFlags flags) :
    QDialog(parent, flags | Qt::CustomizeWindowHint | Qt::WindowTitleHint),
    ui(new Ui::CameraLogSearchDialog)
{
    ui->setupUi(this);

    // Connect up our signal and slots.
    connect(ui->pb_find_cancel, &QPushButton::clicked, this, &CameraLogSearchDialog::cancelDialog);
    connect(ui->pb_find_next, &QPushButton::clicked, this, &CameraLogSearchDialog::findNext);
    connect(this, &QDialog::rejected, this, &CameraLogSearchDialog::cancelDialog);

    got_cancel_sig          = false;

    // Typically, we want to be application modal.
    setWindowModality(Qt::ApplicationModal);
}

CameraLogSearchDialog::~CameraLogSearchDialog()
{
    delete ui;
}


/*!
* @brief Factory to create a simple progress dialog.
*
* @param titleString        What is the title for this simple progress dialog.
* @param statusInfoString   What is the status text to display for this simple progress dialog.
* @param useCancelButton    Does the user want to have a cancel button?
*
* @returns A std::unique_ptr<fiProgressDialog> containing a pointer to the progress dialog.
*/
std::unique_ptr<CameraLogSearchDialog> CameraLogSearchDialog::CreateDialog(const QString& dialogTitleStr,
                                                                           QWidget *parent, Qt::WindowFlags flags)
{
    std::unique_ptr<CameraLogSearchDialog> theSimpleDlg =
        std::unique_ptr<CameraLogSearchDialog>(new CameraLogSearchDialog(parent, flags));

    // Install the title for the search dialog.
    if (theSimpleDlg)
    {
        theSimpleDlg->setWindowTitle(dialogTitleStr);
    }

    return theSimpleDlg;
}

/*!
* @brief Retrieves the user-entered search string.
*
* @returns A constant reference to the user-entered search string.
*/
const QString& CameraLogSearchDialog::getSearchString()
{
    return searchString;
}

/*!
* @brief Retrive search direction.
*
* @returns The direction the user has selected for searching.
*/
CAMLOG_SEARCH::SEARCH_DIRECTION CameraLogSearchDialog::getSearchDirection()
{
    // Which direction has the user selected?
    if (ui->rb_direction_down->isChecked())
    {
        return CAMLOG_SEARCH::SEARCH_DOWN;
    }
    return CAMLOG_SEARCH::SEARCH_UP;
}

/*!
* @brief Retrieves the QT find flags based on the search settings in dialog.
*
* @param findFlagsToSet - the find flags set up for QPlainTextEdit finds.
*
*/
void CameraLogSearchDialog::getSearchFlags(QTextDocument::FindFlags& findFlagsToSet)
{
    // Going backwards in the log?
    if (ui->rb_direction_up->isChecked())
    {
        findFlagsToSet = QTextDocument::FindBackward;
    }
    // Otherwise going forwards in the log.
    else
    {
        findFlagsToSet = QTextDocument::FindFlags();    // No special search flags use default.
    }

    // Are we looking for case sensitivity?
    if (getSearchMatchCase())
    {
        findFlagsToSet = findFlagsToSet | QTextDocument::FindCaseSensitively;
    }

    // Are we searching for whole words?
    if (getSearchWholeWord())
    {
        findFlagsToSet = findFlagsToSet | QTextDocument::FindWholeWords;
    }
}

/*!
* @brief Retrive match case setting.
*
* @returns True if user wants to match case in the search, false if not.
*/
bool CameraLogSearchDialog::getSearchMatchCase()
{
    return (ui->cb_match_case->isChecked());
}

/*!
* @brief Retrive find whole words only state.
*
* @returns True if user wants to match whole words only during search, false if not.
*/
bool CameraLogSearchDialog::getSearchWholeWord()
{
    return (ui->cb_find_while_word->isChecked());
}

/*!
* @brief Retrive search wrap around setting.
*
* @returns True if user wants to wrap around with the search, false if not.
*/
bool CameraLogSearchDialog::getSearchWrapAround()
{
    return (ui->cb_wrap_around->isChecked());
}

/*!
* @brief Set the search direction.
*
* @param searchDirectionToSet - which direction to set.
*/
void CameraLogSearchDialog::setSearchDirection(CAMLOG_SEARCH::SEARCH_DIRECTION searchDirectionToSet)
{
    if (searchDirectionToSet == CAMLOG_SEARCH::SEARCH_DOWN)
        ui->rb_direction_down->setChecked(true);
    else
        ui->rb_direction_up->setChecked(true);
}

/*!
* @brief Set an information message.
*
* @param messageToSet - the message to write into the information area of the dialog.
*/
void CameraLogSearchDialog::setInfoMessage(const QString& messageToSet)
{
    ui->lbl_search_info->setText(messageToSet);
}

/*!
* @brief Set the search match case.
*
* @param useMatchCase - True if setting or false if not.
*/
void CameraLogSearchDialog::setSearchMatchCase(bool useMatchCase)
{
    ui->cb_match_case->setChecked(useMatchCase);
}

/*!
* @brief Set the search match case.
*
* @param wholeWordsOnly - True if looking for whole words only, false if not.
*/
void CameraLogSearchDialog::setSearchWholeWord(bool wholeWordsOnly)
{
    ui->cb_find_while_word->setChecked(wholeWordsOnly);
}

/*!
* @brief Set the search wrap around setting.
*
* @param useWrapAround - True if setting or false if not.
*/
void CameraLogSearchDialog::setSearchWrapAround(bool useWrapAround)
{
    ui->cb_wrap_around->setChecked(useWrapAround);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has pressed the cancel button.
 */
void CameraLogSearchDialog::cancelDialog()
{
    // Have we already emitted a cancel signal?
    if (got_cancel_sig)
        return;

    // Remember we've gotten canceled.
    got_cancel_sig = true;

    disconnect(this, &QDialog::rejected, this, &CameraLogSearchDialog::cancelDialog);

    reject();   // We are done with this dialog - interested in signal connect to "rejected".
}

/*!
 * @brief The user has pressed the find next button.
 */
void CameraLogSearchDialog::findNext()
{
    // Get the search string and persist its contents.
    searchString = ui->le_find_what->text();
    // Let folks know we're canceled.
    emit searchForNext();
}
