/*!
 * @file camera_log.cpp
 * @brief This factory save log dialog, requesting the user to provide all the information
 *        necessary to save the filtered log data.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QMenu>
#include <QFileDialog>
#include <QHostInfo>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonParseError>
#include <QMessageBox>
#include <QSettings>
#include <QTextStream>
#include <QMdiSubWindow>
#include <QContextMenuEvent>
#include <QActionGroup>

#include "camera_logs.h"
#include "ui_camera_log.h"
#include "camera_log_search_dlg.h"
#include "camera_mdi.h"
#include "fi_plain_textedit.h"
#include "fi_journalctl.h"
#include "logger.h"
#include "user_preferences.h"
#include "mainwindow.h"

namespace CAMERA_LOGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

const int LOG_RANGE_DAYS_DEFAULT    = 1;    //!< How many days to set as default.
const int LOG_RANGE_DAYS_MIN        = 1;    //!< Minimum number of days allowed.
const int LOG_RANGE_DAYS_MAX        = 365;  //!< Maximum number of days allowed.

const bool LOG_REAL_TIME_DEFAULT    = true; //!< Default setting for real-time logs request.

//! If user wants to filter on camera server log entries use this ID.
const QString   CSRV_FILTER_ID = QStringLiteral("camera_server");

//! If user wants to filter on FasMotion log entries use this ID.
const QString   FASMOTION_FILTER_ID = QStringLiteral("FasMotion");

//! Title for the camera log search dialog.
const QString LOG_SEARCH_DIALOG_TITLE = QStringLiteral("Camera Log Search");

//! Log Filter Index values for logFilterActionList.
const int LOG_FILTER_RT_MENU_INDEX    = 0;
const int LOG_FILTER_ALL_MENU_INDEX   = 1;
const int LOG_FILTER_FM_MENU_INDEX    = 2;
const int LOG_FILTER_CSRV_MENU_INDEX  = 3;

//! Log Index values for logMenuActionList collection.
const int LOG_MENU_ACTION_SEARCH        = 0;
const int LOG_MENU_ACTION_SAVE          = 1;
const int LOG_MENU_ACTION_WORDWRAP      = 2;

}   // End of namespace SAVE_DLG.

using namespace CAMERA_LOGS;

//--------------------------------------------------------------------
// Local Variable Declarations
//--------------------------------------------------------------------

/*!
 * @brief The Camera Logs Dock Object constructor.
 * @param ip_addr   IP address of the camera we're interested in.
 * @param parent    Our parent widget.
 */
CameraLogs::CameraLogs(QWidget *parent) :
    QDockWidget(parent),
    ui(new Ui::CameraLogDock)
{
    initialized = false;
    haveSetCameraChangeConnect = false;

    // Setup our UI.
    ui->setupUi(this);

    // Install the log levels in a drop down control.
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_EMERGENCY));
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_ALERT));
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_CRITICAL));
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_ERROR));
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_WARNING));
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_NOTICE));
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_INFORMATIONAL));
    ui->dd_log_levels->addItem(CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY::LEVEL_DEBUG));

    // Initialize the default values.
    useFasMotionFilter  = true;
    useCSRVFilter       = true;
    useNoFilter         = false;
    useWordWrap         = false;

    activePriority      = CAMERA_LOGS::PRIORITY::LEVEL_DEBUG;

    // Initialize our context menu.
    initializeContextMenus();

    // Initially there is no attached camera.
    attachedCameraPtr = nullptr;

    // Initialize our signals/slots.
    connect(ui->range_grp, &QGroupBox::clicked, this, &CameraLogs::doRangeGroupSelect);
    connect(ui->grp_filter_options, &QGroupBox::clicked, this, &CameraLogs::doOptionGroupSelect);
    connect(ui->rb_range_by_day, &QRadioButton::clicked, this, &CameraLogs::doByDaySelect);
    connect(ui->rb_range_by_date, &QRadioButton::clicked, this, &CameraLogs::doByDateSelect);
    connect(ui->cb_follow_log, &QCheckBox::clicked, this, &CameraLogs::doFollowLog);
    connect(ui->cb_filter_fasmotion, &QCheckBox::clicked, this, &CameraLogs::doFasMotionFilterSelect);
    connect(ui->cb_filter_csrv, &QCheckBox::clicked, this, &CameraLogs::doCameraServerFilterSelect);
    connect(ui->cb_filter_none, &QCheckBox::clicked, this, &CameraLogs::doNoFilterSelect);
    connect(ui->cb_wrap_lines, &QCheckBox::clicked, this, &CameraLogs::doWrapLines);
    connect(ui->sp_log_range_days, QOverload<int>::of(&QSpinBox::valueChanged), this, &CameraLogs::updateDayRange);
    connect(ui->de_range_start, &QDateEdit::dateChanged, this, &CameraLogs::updateDateTimeRange);
    connect(ui->de_range_end, &QDateTimeEdit::dateChanged, this, &CameraLogs::updateDateTimeRange);
    connect(ui->te_range_start, &QDateTimeEdit::timeChanged, this, &CameraLogs::updateDateTimeRange);
    connect(ui->te_range_end, &QDateTimeEdit::timeChanged, this, &CameraLogs::updateDateTimeRange);
    connect(ui->pb_sync_end_to_now, &QPushButton::clicked, this, &CameraLogs::doSyncEndDateTimeToNow);
    connect(ui->dd_log_levels, QOverload<int>::of(&QComboBox::activated), this, &CameraLogs::doLogLevelChanged);

    connect(ui->pb_search_log, &QPushButton::clicked, this, &CameraLogs::doSearch);
    connect(ui->pb_save_log, &QPushButton::clicked, this, &CameraLogs::doSave);

    setVisible(false);
    setFloating(false);
}

/*!
 * @brief FactorySaveLog destructor.
 */
CameraLogs::~CameraLogs()
{
    // Closeup any resources that may be open for journal control object.
    closeupJournal();

    // Pop a log message to make sure the deleteLater is invoked.
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Destructing CameraLogs Docking Widget.";

    delete ui;

    // Clean up our log level context menu.
    for (int i = 0; i < CameraLogs::logLevelActionList.count(); ++i)
    {
        // Get the pointer to the action.
        QAction *act_ptr = CameraLogs::logLevelActionList.takeFirst();

        if (act_ptr)
        {
            // Delete it.
            delete act_ptr;
        }
    }

    // Now we can delete the submenu.
    if (logLevelMenu)
    {
        delete logLevelMenu;
    }
    if (logLevelActionGroup)
    {
        delete logLevelActionGroup;
    }

    // Clean up our log filter context menu.
    for (int i = 0; i < logFilterActionList.count(); ++i)
    {
        // Get the pointer to the action.
        QAction *act_ptr = logFilterActionList.takeFirst();

        if (act_ptr)
        {
            // Delete it.
            delete act_ptr;
        }
    }

    // Now we can delete the submenu.
    if (logFilterMenu)
    {
        delete logFilterMenu;
    }
    if (logFilterActionGroup)
    {
        delete logFilterActionGroup;
    }

    // And clean up any other actions we created.
    for (int i = 0; i < logMenuActionList.count(); ++i)
    {
        // Get the pointer to the action.
        QAction *act_ptr = logMenuActionList.takeFirst();

        if (act_ptr)
        {
            // Delete it.
            delete act_ptr;
        }
    }
}

/*!
 * @brief Handle a right click from the user - this will pop-up a context menu.
 * @param event_ptr     Pointer to the right-click event.
 */
void CameraLogs::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    // Do we have a context menu?
    if (logContextMenu)
    {
        logContextMenu->exec(event_ptr->globalPos());
    }
}

/*!
 * @brief Resets the logContextMenu to match current state.
 *
 * @returns True if successful, false if there something wrong with menu items.
 */
bool CameraLogs::resetContextMenu()
{
    QAction *act_ptr = nullptr;

    // Make sure we have a log context menu.
    if (!logContextMenu)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "logContextMenu is not allocated.";
        return false;
    }

    // Select the log level which is currently used.
    if (activePriority < logLevelActionList.count())
    {
        QAction *act_ptr = logLevelActionList.at(activePriority);
        act_ptr->setChecked(true);
    }

    // Setup the filter(s) checked state.
    act_ptr = logFilterActionList.at(LOG_FILTER_RT_MENU_INDEX);
    if (act_ptr)
    {
        act_ptr->setChecked(ui->cb_follow_log->isChecked());
    }
    act_ptr = logFilterActionList.at(LOG_FILTER_ALL_MENU_INDEX);
    if (act_ptr)
    {
        if (useNoFilter)
            act_ptr->setChecked(true);
        else
            act_ptr->setChecked(ui->cb_filter_none->isChecked());
    }
    act_ptr = logFilterActionList.at(LOG_FILTER_CSRV_MENU_INDEX);
    if (act_ptr)
    {
        if (useNoFilter)
            act_ptr->setChecked(false);
        else
            act_ptr->setChecked(ui->cb_filter_csrv->isChecked());
    }
    act_ptr = logFilterActionList.at(LOG_FILTER_FM_MENU_INDEX);
    if (act_ptr)
    {
        if (useNoFilter)
            act_ptr->setChecked(false);
        else
            act_ptr->setChecked(ui->cb_filter_fasmotion->isChecked());
    }

    act_ptr = logMenuActionList.at(LOG_MENU_ACTION_WORDWRAP);
    if (act_ptr)
    {
        if (ui->eb_log_info->wordWrapMode() == QTextOption::NoWrap)
        {
            act_ptr->setChecked(false);
        }
        else
        {
            act_ptr->setChecked(true);
        }
    }

    return true;
}

/*!
 * @brief Return the FasMotion filter setting.
 */
bool CameraLogs::usingFasMotionFilter()
{
    bool useFasMotionFilter = false;

    if (ui->cb_filter_fasmotion->isChecked())
        useFasMotionFilter = true;
    return useFasMotionFilter;
}

/*!
 * @brief Return the camera server filter setting.
 */
bool CameraLogs::usingCameraServerFilter()
{
    bool useCSRVFilter = false;

    if (ui->cb_filter_csrv->isChecked())
        useCSRVFilter = true;
    return useCSRVFilter;
}

/*!
 * @brief Return the use no filter setting.
 */
bool CameraLogs::usingNoFilter()
{
    bool useNoFilter = false;

    if (ui->cb_filter_none->isChecked())
        useNoFilter = true;
    return useNoFilter;
}

/*!
 * @brief Return the priority leve the user wants.
 */
void CameraLogs::setPriorityLevel(CAMERA_LOGS::PRIORITY priorityLevelToSet)
{
    ui->dd_log_levels->setCurrentIndex(priorityLevelToSet);
    ui->dd_log_levels->setCurrentText(CameraLogs::getPriorityString(priorityLevelToSet));
}

/*!
 * @brief Sync up the range type in use.
 */
void CameraLogs::syncRangeType()
{
    // Using By Date.
    ui->lbl_range_start->setEnabled(!useRangeByDay);
    ui->de_range_start->setEnabled(!useRangeByDay);
    ui->te_range_start->setEnabled(!useRangeByDay);
    ui->lbl_range_end->setEnabled(!useRangeByDay);
    ui->de_range_end->setEnabled(!useRangeByDay);
    ui->te_range_end->setEnabled(!useRangeByDay);
    ui->pb_sync_end_to_now->setEnabled(!useRangeByDay);
    // Using By Number of Days.
    ui->sp_log_range_days->setEnabled(useRangeByDay);
    ui->cb_follow_log->setEnabled(useRangeByDay);
}

/*!
 * @brief Initialize our dialog boxes controls.
 *
 * @param cameraToAttach - Camera that supplies the log information.
 *
 * @returns True if successful initialization, false if not.
 */
bool CameraLogs::initializeCameraLog(CameraMDI *cameraToAttach)
{
    bool initializedOk = false;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cameraToAttach:"
                 << (cameraToAttach ? cameraToAttach->get_ip_addr_pretty() : "Null") << "attachedCameraPtr:" << (attachedCameraPtr ? attachedCameraPtr->get_ip_addr_pretty() : "Null");

    // Initialize the log default setup.
    ui->cb_filter_fasmotion->setChecked(false);
    ui->cb_filter_csrv->setChecked(false);
    ui->cb_filter_none->setChecked(true);
    ui->pb_save_log->setEnabled(false);

    // Set up our range in days.  Do this before setting up the date/time edit control.
    ui->sp_log_range_days->setMinimum(LOG_RANGE_DAYS_MIN);
    ui->sp_log_range_days->setMaximum(LOG_RANGE_DAYS_MAX);
    ui->sp_log_range_days->setValue(LOG_RANGE_DAYS_DEFAULT);

    // By default, let's grab everything.
    activePriority = CAMERA_LOGS::PRIORITY::LEVEL_DEBUG;
    setPriorityLevel(activePriority);

    // Set the default to real-time default setting.
    ui->cb_follow_log->setChecked(LOG_REAL_TIME_DEFAULT);

    // Call the handler for using range by day.
    useRangeByDay = true;
    doByDaySelect();

    // Specify the range type based on useRangeByDay flag.
    ui->rb_range_by_day->setChecked(useRangeByDay);
    ui->rb_range_by_date->setChecked(!useRangeByDay);

    // Validate the camera information provided.
    if (cameraToAttach == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Camera provided was a nullptr!";
        return false;
    }

    // If the camera is a series 3 camera this functionality is not available.
    if (cameraToAttach->is_series3())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Saving journalctl logs is not available for series 3 cameras.";
        return false;
    }

    // Persist the attached camera pointer for use in this object.
    // Since this isn't a smart pointer it is important to point out
    // this object is just borrowing this pointer from the caller it
    // is not managed or freed in this object.
    attachedCameraPtr = cameraToAttach;

    // Set the internal initialized state to true.
    initialized = true;

    // Set up the log for the first pull using the settings provided here.
    initializedOk = resetJournalCtl();

    // Set up a way to catch when active children change in the system.
    //if (!haveSetCameraChangeConnect && !MainWindow::GetInstance().isNull())
    if (!haveSetCameraChangeConnect && MainWindow::GetInstance() != nullptr)
    {
        FI_comm     *comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr != nullptr)
        {
            // Connect our signals and slots.
            connect(MainWindow::GetInstance().get(), &MainWindow::active_child_changed, this, &CameraLogs::doActiveCameraChanged);
            haveSetCameraChangeConnect = true;
        }
    }

    return initializedOk;
}

/*!
 * @brief Initialize our dialog boxes controls.
 *
 * @param cameraToAttach - Camera that supplies the log information.
 *
 * @returns True if camera belongs to this log.
 */
bool CameraLogs::isCameraAttached(CameraMDI *camera_ptr)
{
    bool ourCamera = false;
    // Is this ours?
    if (camera_ptr == attachedCameraPtr)
    {
        ourCamera = true;
    }
    return ourCamera;
}

/*!
 * @brief Removes an attached camera log from the docking window.
 */
void CameraLogs::removeCameraLog(CameraMDI *camera_ptr)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ <<
                 "camera_ptr:" << (camera_ptr ? camera_ptr->get_ip_addr_pretty() : "Null") <<
                 "attachedCameraPtr:" << (attachedCameraPtr ? attachedCameraPtr->get_ip_addr_pretty() : "Null");

    // Is this ours?
    if (isCameraAttached(camera_ptr))
    {
        // Reset the initialized flag.
        initialized = false;

        // Close down any journal control resources in use.
        closeupJournal();

        attachedCameraPtr = nullptr;
    }
}

/*!
 * @brief Install the camera pointer that contains the log file the user
 * wants to save.
 *
 * @param cameraToAttach the pointer to connected camera.
 *
 * @return If successful it returns true otherwise an error occurred.
 */
bool CameraLogs::setCamera(CameraMDI *cameraToAttach)
{
    bool setCameraOk = false;

    QString titleActionStr  = QStringLiteral("System Logs");
    QString titleFullStr    = titleActionStr;
    QString setCameraErrorString;
    QString allocateErrorStr = QStringLiteral("Unable to allocate the JournalCtl object.");

    QLOG_TRACE() << __PRETTY_FUNCTION__ << (cameraToAttach ? cameraToAttach->get_ip_addr_pretty() : "Local Camera");

    // Cleanup and reset the journal control object if it is allocated from
    // a previous log request.
    closeupJournal();

    // Validate the camera information provided.
    if (cameraToAttach == nullptr)
    {
        setCameraErrorString = QStringLiteral("%1 Received Null Camera Pointer.").arg(__PRETTY_FUNCTION__);
        goto setCameraExit;
    }

    // If the camera is a series 3 camera this functionality is not available.
    if (cameraToAttach->is_series3())
    {
        setCameraErrorString = QStringLiteral("%1 Saving journalctl log is not available for series 3 cameras.").arg(__PRETTY_FUNCTION__);
        QLOG_ERROR() << setCameraErrorString;
        goto setCameraExit;
    }

    // Persist the attached camera pointer for use in this object.
    // Since this isn't a smart pointer it is important to point out
    // this object is just borrowing this pointer from the caller it
    // is not managed or freed in this object.
    attachedCameraPtr = cameraToAttach;

    // Is the camera remote?
    if (!attachedCameraPtr->is_local())
    {
        QSettings settings;
        FI_JOURNALCTL::SSH_SETTINGS ssh_info;

        // Build up a title for our window.
        titleFullStr = QStringLiteral("%1 For %2 [%3]").arg(titleActionStr, attachedCameraPtr->get_camera_name(), attachedCameraPtr->get_ip_addr_pretty());

        // Initialize our information.
        ssh_info.hostname = attachedCameraPtr->get_ip_addr_pretty();
        settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
        ssh_info.username = settings.value(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
        ssh_info.crypt_info = settings.value(UserPreferences::SERIESNC_CAM_PASS_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
        settings.endGroup();

        // Create our remote journal control object;
        journal = std::unique_ptr<FiJournalCtl>(new FiJournalCtl(this, &ssh_info));
        if (journal.get() == nullptr)
        {
            setCameraErrorString = QStringLiteral("%1 %2").arg(__PRETTY_FUNCTION__, allocateErrorStr);
            QLOG_DEBUG() << setCameraErrorString;
            goto setCameraExit;
        }
    }
    else
    {
        // Build up a title for our window.
        titleFullStr = QStringLiteral("%1 For %2 [local]").arg(titleActionStr, attachedCameraPtr->get_camera_name());

        // Create our journal control object;
        journal = std::unique_ptr<FiJournalCtl>(new FiJournalCtl(this));
        if (journal.get() == nullptr)
        {
            setCameraErrorString = QStringLiteral("%1 %2").arg(__PRETTY_FUNCTION__, allocateErrorStr);
            QLOG_DEBUG() << setCameraErrorString;
            goto setCameraExit;
        }
    }

    // Is the journal control object OK?
    if (!journal->isCreated())
    {
        setCameraErrorString = QStringLiteral("%1 JournalCtl object was not created successfully.").arg(__PRETTY_FUNCTION__);
        QLOG_DEBUG() << setCameraErrorString;

        // Show an error message.
        ui->eb_log_info->setUpdatesEnabled(false);
        ui->eb_log_info->appendPlainText(journal->errorStr());
        ui->eb_log_info->setUpdatesEnabled(true);

        // Close the journal control object.
        closeupJournal();

        goto setCameraExit;
    }

    // Connect the signals up.
    connect(journal.get(), &FiJournalCtl::dataAvailable, this, &CameraLogs::updateLogData, Qt::QueuedConnection);

    setCameraOk = true;

setCameraExit:

    // Set a title.
    setWindowTitle(titleFullStr);

    // If this was not successful put the error condition we set in the
    // user visible log information.
    if (!setCameraOk)
    {
        // Show our internal error message.
        ui->eb_log_info->setUpdatesEnabled(false);
        ui->eb_log_info->appendPlainText(setCameraErrorString);
        ui->eb_log_info->setUpdatesEnabled(true);
    }

    // Return the status of setting up the camera.
    return setCameraOk;
}

/*!
 * @brief Installs the information for the journal control to request the
 * log data from the camera system.
 */
void CameraLogs::requestLogData()
{
    QStringList my_id;
    QStringList my_args;

    // Make sure there is a journal object to use.
    if (!journal)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Missing journal control object!";
        return;
    }

    // Does the user want to follow the log in real time?
    journal->setFollow(ui->cb_follow_log->isChecked());

    // We also need to run as root.
    journal->setUseSudo(true);

    // Set up any filters specified.
    if (usingFasMotionFilter())
    {
        my_id.append(FASMOTION_FILTER_ID);
    }
    if (usingCameraServerFilter())
    {
        my_id.append(CSRV_FILTER_ID);
    }
    if (!my_id.isEmpty())
    {
        journal->setIdentifiers(my_id);
    }

    my_args.append(QStringLiteral("-a"));
    my_args.append(QStringLiteral("-p"));
    my_args.append(QString::number(ui->dd_log_levels->currentIndex()));

    QDateTime sinceDateTime(ui->de_range_start->date(), ui->te_range_start->time());
    QString sinceStr;

    // Is the camera local?
    if (attachedCameraPtr->is_local())
        sinceStr = QStringLiteral("--since=") + sinceDateTime.toString("yyyy-MM-dd hh:mm:00");
    else
        sinceStr = QStringLiteral("--since=\"") + sinceDateTime.toString("yyyy-MM-dd hh:mm:00") + QStringLiteral("\"");
    my_args.append(sinceStr);

    if (!ui->cb_follow_log->isChecked())
    {
        QDateTime untilDateTime(ui->de_range_end->date(), ui->te_range_end->time());
        QString untilStr;

        // Is the camera local?
        if (attachedCameraPtr->is_local())
            untilStr = QStringLiteral("--until=") + untilDateTime.toString("yyyy-MM-dd hh:mm:00");
        else
            untilStr = QStringLiteral("--until=\"") + untilDateTime.toString("yyyy-MM-dd hh:mm:00") + QStringLiteral("\"");
        my_args.append(untilStr);
    }

    QLOG_TRACE() << "[Factory Save Log Command][journalctl " << my_args << "]";

    // Send the custom command list to the journal.
    journal->setAdditionalArgs(my_args);

    // Tell the journal control object to go!
    journal->run();
}

/*!
 * @brief Common close up logic for the journal control object.
 *
 */
void CameraLogs::closeupJournal()
{
    // Do we have a journal control object?
    if (journal.get() != nullptr)
    {
        // Disconnect any signals that were setup.
        disconnect(journal.get(), &FiJournalCtl::dataAvailable, this, &CameraLogs::updateLogData);

        // Close and reset the control object.
        journal->close();
        journal.reset();
    }
}


/*!
 * @brief Logic to generate a menu action object for context menus.
 *
 * @param menuItemString - the string to use for the menu item.
 *
 * @returns The QAction pointer or null if unable to create one.
 */
QAction *CameraLogs::createMenuAction(const QString& menuItemString,
                                      bool checkable,
                                      bool visibleIcon,
                                      QList<QAction *> *actionListPtr, QActionGroup *actionGroupPtr)
{
    // Allocate a new QAction for a menu item.
    QAction *act_ptr = new QAction(menuItemString, this);
    if (act_ptr)
    {
        act_ptr->setCheckable(checkable);
        act_ptr->setIconVisibleInMenu(visibleIcon);

        if (actionListPtr != nullptr)
        {
            // Append the action to our list (ownership moves to list).
            actionListPtr->append(act_ptr);
        }

        if (actionGroupPtr != nullptr)
        {
            // Add the action to our group.
            actionGroupPtr->addAction(act_ptr);
        }
    }
    return act_ptr;
}

/*!
 * @brief Initializes the context menus.
 */
bool CameraLogs::initializeContextMenus()
{
    bool created_ok = false;

    // LOG LEVEL SUBMENU
    // Allocate our log-level action group so we can deal with them as an entity.
    logLevelActionGroup = new QActionGroup(this);
    if (logLevelActionGroup == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate log level action group!";
        goto initContextMenuItemsExit;
    }
    else
    {
        // For the context menu, our actions operate exclusive of each other.
        logLevelActionGroup->setExclusive(true);

        // Install all of the log-level actions.
        for (int i = CAMERA_LOGS::PRIORITY::LEVEL_EMERGENCY; i < CAMERA_LOGS::PRIORITY::LEVEL_END; i++)
        {
            // Build up the menu based on the log
            QAction *act_ptr = createMenuAction(getPriorityString((CAMERA_LOGS::PRIORITY)i), true, false, &logLevelActionList, logLevelActionGroup);
            if (act_ptr)
            {
                // Connect our signals/slots.
                connect(act_ptr, &QAction::triggered, this, [this, act_ptr]() { doLogLevelMenuChanged(act_ptr); });
            }
        }

        // Allocate our log level submenu.
        logLevelMenu = new QMenu(tr("Log Level"), this);
        if (logLevelMenu == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate logLevelMenu!";
            goto initContextMenuItemsExit;
        }
        else
        {
            // Add our "Log Level Group" to the menu.
            logLevelMenu->addActions(logLevelActionGroup->actions());
        }
    }

    // LOG FILTER SUBMENU
    // Allocate our log filter action group so we can deal with them as an entity.
    logFilterActionGroup = new QActionGroup(this);
    if (logFilterActionGroup == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate log filter action group!";
        goto initContextMenuItemsExit;
    }
    else
    {
        // For filter options we are not exclusive of each other.
        logFilterActionGroup->setExclusive(false);
        QAction *act_ptr = createMenuAction(ui->cb_follow_log->text(), true, false, &logFilterActionList, logFilterActionGroup);
        if (act_ptr)
        {
            // Connect our signals/slots.
            connect(act_ptr, &QAction::triggered, this, [this, act_ptr]() { doLogFilterMenuChanged(act_ptr); });
        }
        act_ptr = createMenuAction(ui->cb_filter_none->text(), true, false, &logFilterActionList, logFilterActionGroup);
        if (act_ptr)
        {
            // Connect our signals/slots.
            connect(act_ptr, &QAction::triggered, this, [this, act_ptr]() { doLogFilterMenuChanged(act_ptr); });
        }
        act_ptr = createMenuAction(ui->cb_filter_fasmotion->text(), true, false, &logFilterActionList, logFilterActionGroup);
        if (act_ptr)
        {
            // Connect our signals/slots.
            connect(act_ptr, &QAction::triggered, this, [this, act_ptr]() { doLogFilterMenuChanged(act_ptr); });
        }
        act_ptr = createMenuAction(ui->cb_filter_csrv->text(), true, false, &logFilterActionList, logFilterActionGroup);
        if (act_ptr)
        {
            // Connect our signals/slots.
            connect(act_ptr, &QAction::triggered, this, [this, act_ptr]() { doLogFilterMenuChanged(act_ptr); });
        }

        // Allocate our Log Filter submenu.
        logFilterMenu = new QMenu(tr("Filter Options"), this);
        if (logFilterMenu == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate logFilterMenu!";
            goto initContextMenuItemsExit;
        }

        // Add our "Log Level Group" to the menu.
        logFilterMenu->addActions(logFilterActionGroup->actions());
    }

    // Create our filter options submenu.
    logContextMenu = std::shared_ptr<QMenu>(new QMenu(ui->grp_filter_options->title(), this));
    if (logContextMenu)
    {
        QAction *act_ptr;

        // Note order of insertion into logMenuActionList is important!!!
        // These constants leverage the index position:
        //const int LOG_MENU_ACTION_SEARCH        = 0;
        //const int LOG_MENU_ACTION_SAVE          = 1;
        //const int LOG_MENU_ACTION_WORDWRAP      = 2;

        act_ptr = createMenuAction(ui->pb_search_log->text(), false, false, &logMenuActionList, nullptr);
        if (act_ptr)
        {
            // Connect our signals/slots.
            connect(act_ptr, &QAction::triggered, this, [this]() { doSearch(); });
            logContextMenu->addAction(act_ptr);
        }

        act_ptr = createMenuAction(ui->pb_save_log->text(), false, false, &logMenuActionList, nullptr);
        if (act_ptr)
        {
            // Connect our signals/slots.
            connect(act_ptr, &QAction::triggered, this, [this]() { doSave(); });
            logContextMenu->addAction(act_ptr);
        }

        // Install the submenu sections.
        logContextMenu->addSeparator();

        act_ptr = createMenuAction(ui->cb_wrap_lines->text(), true, false, &logMenuActionList, nullptr);
        if (act_ptr)
        {
            // Connect our signals/slots.
            connect(act_ptr, &QAction::triggered, this, [this, act_ptr]() { doWordLinesMenuChange(act_ptr); });
            logContextMenu->addAction(act_ptr);
        }

        // Install the submenu sections.
        logContextMenu->addSeparator();

        // Install the "Log Filter Group Submenu" in the contextMenu.
        logContextMenu->addMenu(logFilterMenu);

        // Install the "Log Level Group Submenu" to the contextMenu.
        logContextMenu->addMenu(logLevelMenu);

        resetContextMenu();

        // Tell our plain text editor about our menu and instruct it to use it as a sub-context menu.
        ui->eb_log_info->setParentContextMenu(logContextMenu, true);
    }

initContextMenuItemsExit:
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "created_ok:" << created_ok;

    return created_ok;
}

/*!
 * @brief Get the priority string associated with a priority level..
 *
 * The index values should correspond to the enumeration in
 * FACTORY_SAVELOG::PRIORITY_LEVELS
 *
 * @param priorityLevelToGet - which priority level are we interested in?
 *
 */
QString CameraLogs::getPriorityString(CAMERA_LOGS::PRIORITY priorityLevelToGet)
{
    // Change the index to match the slider type.
    QString priorityString;

    switch (priorityLevelToGet)
    {
    case CAMERA_LOGS::PRIORITY::LEVEL_EMERGENCY:
        priorityString = tr("Emergency");
        break;
    case CAMERA_LOGS::PRIORITY::LEVEL_ALERT:
        priorityString = tr("Alert");
        break;
    case CAMERA_LOGS::PRIORITY::LEVEL_CRITICAL:
        priorityString = tr("Critical");
        break;
    case CAMERA_LOGS::PRIORITY::LEVEL_ERROR:
        priorityString = tr("Error");
        break;
    case CAMERA_LOGS::PRIORITY::LEVEL_WARNING:
        priorityString = tr("Warning");
        break;
    case CAMERA_LOGS::PRIORITY::LEVEL_NOTICE:
        priorityString = tr("Notice");
        break;
    case CAMERA_LOGS::PRIORITY::LEVEL_INFORMATIONAL:
        priorityString = tr("Informational");
        break;
    case CAMERA_LOGS::PRIORITY::LEVEL_DEBUG:
    default:
        priorityString = tr("Debug");
        break;
    }
    return priorityString;
}

/*!
 * @brief The user clicked on one of the log filter actions in the context
 *        menu. This has the effect of changing the log filter.
 *
 * @param obj_ptr   The pointer to the action item the user clicked/selected.
 */
void CameraLogs::doLogFilterMenuChanged(QObject *obj_ptr)
{
    int     index;
    QAction *act_ptr = qobject_cast<QAction *>(obj_ptr);

    // Process the action.
    if (act_ptr)
    {
        // Get the index of the action object that was selected.
        index = logFilterActionList.indexOf(act_ptr);

        switch (index)
        {
        case LOG_FILTER_CSRV_MENU_INDEX:
            // Toggle this.
            ui->cb_filter_csrv->setChecked(act_ptr->isChecked());
            doCameraServerFilterSelect();
            break;
        case LOG_FILTER_FM_MENU_INDEX:
            // Toggle this.
            ui->cb_filter_fasmotion->setChecked(act_ptr->isChecked());
            doFasMotionFilterSelect();
            break;
        case LOG_FILTER_RT_MENU_INDEX:
            // Toggle this.
            ui->cb_follow_log->setChecked(act_ptr->isChecked());
            doFollowLog();
            break;
        case LOG_FILTER_ALL_MENU_INDEX:
        default:
            // This is a setting to true.
            ui->cb_filter_none->setChecked(act_ptr->isChecked());
            doNoFilterSelect();
            break;
        }
    }
}

/*!
 * @brief The user clicked on one of the log level actions in the context
 *        menu. This has the effect of changing the log contents to match the level.
 *
 * @param obj_ptr   The pointer to the action item the user clicked/selected.
 */
void CameraLogs::doLogLevelMenuChanged(QObject *obj_ptr)
{
    int     index;
    QAction *act_ptr = qobject_cast<QAction *>(obj_ptr);

    // Process the action.
    if (act_ptr)
    {
        // Get the index of the action object that was selected.
        index = logLevelActionList.indexOf(act_ptr);

        // Pass the information on as though the header was clicked.
        doLogLevelChanged(index);
    }
}



/*!
 * @brief The user has made a selection in the word wrap context menu.
 */
void CameraLogs::doWordLinesMenuChange(QObject *obj_ptr)
{
    QAction *act_ptr = qobject_cast<QAction *>(obj_ptr);

    // Process the action.
    if (act_ptr)
    {
        if (act_ptr->isChecked())
        {
            ui->cb_wrap_lines->setChecked(true);
        }
        else
        {
            ui->cb_wrap_lines->setChecked(false);
        }

        // Pass the information on as though the checkbox was clicked.
        doWrapLines();
    }
}

/*!
 * @brief Initializes the camera and journal control and requests log data
 * based on the settings in this dialog.
 *
 * @return Result from setting up the camera/journal control.
 */
bool CameraLogs::resetJournalCtl()
{
    bool resetJournalCtlOk = false;

    // If this dialog isn't initialized do not do anything here yet.
    if (!initialized)
        return false;

    ui->eb_log_info->setUpdatesEnabled(false);
    ui->eb_log_info->clear();
    ui->eb_log_info->clearBookmarks();
    ui->eb_log_info->setUpdatesEnabled(true);
    ui->eb_log_info->setWordWrapMode(useWordWrap ? QTextOption::WordWrap : QTextOption::NoWrap);
    ui->pb_save_log->setEnabled(false);

    // Make sure our context menu matches our state.
    resetContextMenu();

    // Set the camera and setup the journal control to retrieve log data.
    resetJournalCtlOk = setCamera(attachedCameraPtr);
    if (resetJournalCtlOk)
    {
        // If the camera and journal control all set then tell the journal
        // control what to retrieve.
        requestLogData();
    }
    return resetJournalCtlOk;
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void CameraLogs::updateLogData(QString log_data)
{
    // Add the new entries to our log display.
    ui->eb_log_info->setUpdatesEnabled(false);
    ui->eb_log_info->appendPlainText(log_data);
    ui->eb_log_info->setUpdatesEnabled(true);

    // If we added log data and the save control is disabled then renable it.
    if (!ui->pb_save_log->isEnabled())
        ui->pb_save_log->setEnabled(true);
}

/*!
 * @brief A new camera (or possibly none) has become active.
 * @param window    The QMdiSubWindow pointer to the related camera window.
 */
void CameraLogs::doActiveCameraChanged(QMdiSubWindow *window)
{
    CameraMDI   *new_camera;

    // Is there no new active camera?
    if (window == nullptr)
        return;

    // Get our new camera object.
    new_camera = qobject_cast<CameraMDI *>(window->widget());
    if (new_camera == nullptr)
        return;

    // Don't support series3 cameras.
    if (new_camera->is_series3())
        return;

    // Do we have an attached camera pointer?
    if (attachedCameraPtr)
    {
        // Is it really new?
        if (new_camera != attachedCameraPtr)
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "new_camera:" << new_camera->get_ip_addr_pretty() << "attachedCameraPtr:" << attachedCameraPtr->get_ip_addr_pretty();

            removeCameraLog(attachedCameraPtr);
            initializeCameraLog(new_camera);
        }
    }
}

/*!
 * @brief The user clicked on the select range by day radio button.
 */
void CameraLogs::doByDaySelect()
{
    useRangeByDay = true;
    syncRangeType();
    doSyncEndDateTimeToNow();
    updateDateTimeForDayRange();
    resetJournalCtl();
}

/*!
 * @brief The user clicked on the select range by date radio button.
 */
void CameraLogs::doByDateSelect()
{
    // Moving to by date don't allow a follow setting.
    ui->cb_follow_log->setChecked(false);

    useRangeByDay   = false;
    syncRangeType();
    doSyncEndDateTimeToNow();

    resetJournalCtl();
}



/*!
 * @brief The user pressed the OK button. Save the settings entered by the user
 *        so we can re-use them the next time we present the save dialog.
 */
void CameraLogs::doSave()
{
    QString     filename;
    QString     load_path;
    QFile       log_file;

    // Let the downloads path be the default path to start the save location each time.
    load_path = QStandardPaths::standardLocations(QStandardPaths::DownloadLocation).at(0);

    // Ask the user for a configuration file.
    filename = QFileDialog::getSaveFileName(this, tr("Factory Save Log Data"), load_path, tr("Log File (*.log *.txt)"));
    if (filename.isEmpty())
        return;

    // Make sure the path will look "right" to the user.
    filename = QDir::toNativeSeparators(filename);

    // Get the path for the log file.
    log_file.setFileName(filename);

    // Open the file to be written to.
    if (!log_file.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        QString msg_str = tr("Unable to open %1, Err=%2").arg(filename, log_file.errorString());
        QMessageBox::warning(this, "Error Opening Log File", msg_str);
        return;
    }
    log_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // Walk through the data and save off the information.
    QTextStream out_stream(&log_file);

    out_stream << ui->eb_log_info->toPlainText() << Qt::endl;

    // All done.
    log_file.close();

    // Reset the initialized flag.
    initialized = false;

    // Pass it on up.
    //QDialog::accept();
}

/*!
 * @brief The user pressed the Search Log button. Need to present the search dialog and
 * allow them to search the log information.
 */
void CameraLogs::doSearch()
{
    ui->pb_search_log->setEnabled(false);

    // Is there already a search dialog showing?
    if (!searchDlg)
    {
        // Generate a new search dialog.
        searchDlg  = CameraLogSearchDialog::CreateDialog(LOG_SEARCH_DIALOG_TITLE, this);
    }

    // Set it up for a new search.
    if (searchDlg)
    {
        // We want the user to be able to do other things when using this dialog.
        searchDlg->setModal(false);

        // Anchor the cursor at the top of the log.
        QTextCursor tempCursor = ui->eb_log_info->textCursor();
        ui->eb_log_info->setCursorWidth(10);
        ui->eb_log_info->moveCursor(QTextCursor::Start);
        tempCursor = ui->eb_log_info->textCursor();
        //ui->eb_log_info->moveCursor(QTextCursor::End);

        searchDlg->setSearchDirection(CAMLOG_SEARCH::SEARCH_DOWN);
        searchDlg->setSearchMatchCase(false);
        searchDlg->setSearchWrapAround(true);
        searchDlg->setSearchWholeWord(false);

        // Connect up any signals.
        connect(searchDlg.get(), &CameraLogSearchDialog::rejected, this, &CameraLogs::stopLogSearch);
        connect(searchDlg.get(), &CameraLogSearchDialog::searchForNext, this, &CameraLogs::nextLogSearch);
        searchDlg->showNormal();
    }
}


/*!
 * @brief Responds to user clicking to close the log docking widget.
 *
 * @param event     Pointer to the event structure.
 */
void CameraLogs::closeEvent(QCloseEvent *event)
{
    removeCameraLog(attachedCameraPtr);

    // Accept the event.
    event->accept();

    // Pass it on.
    QDockWidget::closeEvent(event);
}

/*!
 * @brief The user clicked on Real-time Log Updates checkbox.
 */
void CameraLogs::doFollowLog()
{
    resetJournalCtl();
}

/*!
 * @brief The user clicked on the Fasmotion filter checkbox.
 */
void CameraLogs::doFasMotionFilterSelect()
{
    useFasMotionFilter = ui->cb_filter_fasmotion->isChecked();
    // If selecting the FasMotion filter then uncheck the no-filter option.
    if (useFasMotionFilter)
    {
        ui->cb_filter_none->setChecked(false);
        useNoFilter = false;
    }
    resetJournalCtl();
}

/*!
 * @brief The user clicked on the camera server filter checkbox.
 */
void CameraLogs::doCameraServerFilterSelect()
{
    useCSRVFilter = ui->cb_filter_csrv->isChecked();
    // If selecting the camera server filter then uncheck the no-filter option.
    if (useCSRVFilter)
    {
        ui->cb_filter_none->setChecked(false);
        useNoFilter = false;
    }
    resetJournalCtl();
}

/*!
 * @brief The user clicked on the no-filter checkbox.
 */
void CameraLogs::doNoFilterSelect()
{
    useNoFilter = ui->cb_filter_none->isChecked();
    // If selecting the camera server filter then uncheck the no-filter option.
    if (useNoFilter)
    {
        ui->cb_filter_fasmotion->setChecked(false);
        useFasMotionFilter = false;
        ui->cb_filter_csrv->setChecked(false);
        useCSRVFilter = false;
    }
    else
    {
        // If unchecking this then recheck both the other options.
        if (!ui->cb_filter_fasmotion->isChecked())
        {
            ui->cb_filter_fasmotion->setChecked(true);
        }
        if (!ui->cb_filter_csrv->isChecked())
        {
            ui->cb_filter_csrv->setChecked(true);
        }
    }
    resetJournalCtl();
}

/*!
 * @brief The user clicked on the wrap lines checkbox.
 */
void CameraLogs::doWrapLines()
{
    useWordWrap = ui->cb_wrap_lines->isChecked();
    if (useWordWrap)
    {
        ui->eb_log_info->setWordWrapMode(QTextOption::WordWrap);
    }
    else
    {
        ui->eb_log_info->setWordWrapMode(QTextOption::NoWrap);
    }
}

/*!
 * @brief User has modified the drop down combo box for the log level.
 */
void CameraLogs::doLogLevelChanged(int index)
{
    // Is there a change?
    if (index != (int) activePriority)
    {
        // Modify the active priority and we need to reset the journal control.
        activePriority = (CAMERA_LOGS::PRIORITY)index;

        // Make sure that our drop-down control matches the priority level.
        if (ui->dd_log_levels->currentIndex() != index)
        {
            setPriorityLevel((CAMERA_LOGS::PRIORITY)index);
        }

        // Setup the journal control.
        resetJournalCtl();
    }
}

/*!
 * @brief The user clicked on the range group check box.
 */
void CameraLogs::doRangeGroupSelect()
{
    bool showRangeGroup = ui->range_grp->isChecked();

    // Change visibility of the group members based on the show state.
    ui->rb_range_by_day->setVisible(showRangeGroup);
    ui->sp_log_range_days->setVisible(showRangeGroup);
    ui->rb_range_by_date->setVisible(showRangeGroup);
    ui->lbl_range_start->setVisible(showRangeGroup);
    ui->de_range_start->setVisible(showRangeGroup);
    ui->te_range_start->setVisible(showRangeGroup);
    ui->lbl_range_end->setVisible(showRangeGroup);
    ui->de_range_end->setVisible(showRangeGroup);
    ui->te_range_end->setVisible(showRangeGroup);
    ui->pb_sync_end_to_now->setVisible(showRangeGroup);

    ui->range_grp->setMaximumHeight(showRangeGroup ? 16777215 : 20);
}

/*!
 * @brief The user clicked on the option group check box.
 */
void CameraLogs::doOptionGroupSelect()
{
    bool showOptionGroup = ui->grp_filter_options->isChecked();

    // Change visibility of the group members based on the show state.
    ui->cb_follow_log->setVisible(showOptionGroup);
    ui->cb_filter_fasmotion->setVisible(showOptionGroup);
    ui->cb_filter_csrv->setVisible(showOptionGroup);
    ui->cb_filter_none->setVisible(showOptionGroup);
    ui->dd_log_levels->setVisible(showOptionGroup);
    ui->pb_save_log->setVisible(showOptionGroup);

    ui->grp_filter_options->setMaximumHeight(showOptionGroup ? 16777215 : 20);
}

/*!
 * @brief The user wants to set the end time to right now.
 *
 */
void CameraLogs::doSyncEndDateTimeToNow()
{
    QDateTime dateTimeNow = QDateTime::currentDateTime();
    QDateTime endDateTime(ui->de_range_end->date(), ui->te_range_end->time());
    bool timeChanged = (dateTimeNow != endDateTime);

    // Only do anything if time has changed.
    if (timeChanged)
    {
        ui->de_range_end->blockSignals(true);
        ui->de_range_end->setMaximumDateTime(dateTimeNow);
        ui->de_range_end->setDateTime(dateTimeNow);
        ui->de_range_end->blockSignals(false);

        ui->te_range_end->blockSignals(true);
        ui->te_range_end->setMaximumDateTime(dateTimeNow);
        ui->te_range_end->setDateTime(dateTimeNow);
        ui->te_range_end->blockSignals(false);

        resetJournalCtl();    // time changed.
    }
}

/*!
 * @brief The user has cancelled the log search time to remove the dialog.
 *
 */
void CameraLogs::stopLogSearch()
{
    if (searchDlg)
    {
        disconnect(searchDlg.get(), &CameraLogSearchDialog::rejected, this, &CameraLogs::stopLogSearch);
        disconnect(searchDlg.get(), &CameraLogSearchDialog::searchForNext, this, &CameraLogs::nextLogSearch);
        searchDlg.release();
    }
    ui->pb_search_log->setEnabled(true);
}

/*!
 * @brief The user requested to find the next string in the log.
 *
 */
void CameraLogs::nextLogSearch()
{
    if (searchDlg)
    {
        bool foundString = false;
        QTextDocument::FindFlags searchFlags;
        CAMLOG_SEARCH::SEARCH_DIRECTION whichDirection = searchDlg->getSearchDirection();

        // Get the string from the search dialog. (actually this can be signal if it changes)
        logSearchString = searchDlg->getSearchString();

        // Get the search flags to use for the QT find operation from search dialog.
        searchDlg->getSearchFlags(searchFlags);

        // Allow two passes through in case need to wrap around.
        for (int passCount = 0; !foundString && (passCount < 2); passCount++)
        {
            // Search for the string.
            foundString = ui->eb_log_info->find(logSearchString, searchFlags);

            // Check for wrap-around setting and if set go around one more pass.
            if (!foundString)
            {
                // If user has it set to wrap around - go around and check.
                if (searchDlg->getSearchWrapAround())
                {
                    if (whichDirection == CAMLOG_SEARCH::SEARCH_DOWN)
                    {
                        // If searching down attempt to go around again from beginning.
                        ui->eb_log_info->moveCursor(QTextCursor::Start);
                    }
                    else
                    {
                        // If searching up attempt to go around again from end.
                        ui->eb_log_info->moveCursor(QTextCursor::End);
                    }
                }
            }
        }
        // Log the message if found.
        if (foundString)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Found:" << logSearchString;
            QString strMsg; // Empty string clears out the information message line.
            searchDlg->setInfoMessage(strMsg);
        }
        else
        {
            QString strMsg;
            if (logSearchString.isEmpty())
            {
                strMsg = tr("Please enter a string to find.");
            }
            else
            {
                strMsg = QString::fromUtf8("Unable to find \"%1\" in the log.").arg(logSearchString);
            }
            searchDlg->setInfoMessage(strMsg);
        }
    }
}

/*!
 * @brief The user has modified number of days they want to log.
 *
 * @param index     The newly selected number of days to save the log inforamtion.
 *
 */
void CameraLogs::updateDayRange(int value)
{
    (void)value;

    // Just update the date and time for the new day range.
    updateDateTimeForDayRange();

    // Reset the log to reflect any changed values.
    resetJournalCtl();
}

/*!
 * @brief Update the date and time ranges using current date and time.
 *
 */
void CameraLogs::updateDateTimeForDayRange()
{
    QDateTime datetimeNow = QDateTime::currentDateTime();
    QDateTime datetimeOld = datetimeNow.addDays((qint64)(ui->sp_log_range_days->value() * -1));

    // Don't want to get signals when changing this data.
    ui->de_range_start->blockSignals(true);
    ui->de_range_end->blockSignals(true);
    ui->te_range_start->blockSignals(true);
    ui->te_range_end->blockSignals(true);

    ui->de_range_start->setMaximumDateTime(datetimeNow);
    ui->de_range_start->setDateTime(datetimeOld);
    ui->de_range_end->setMaximumDateTime(datetimeNow);
    ui->de_range_end->setDateTime(datetimeNow);

    // Make sure we have a minimum of one day.
    if (datetimeOld < datetimeNow.addDays(-1))
    {
        // We have more than enough time, so no need for a maximum.
        ui->te_range_start->clearMaximumDateTime();
    }
    else
    {
        // Force the maximum time so we get at least one day in the log.
        ui->te_range_start->setMaximumDateTime(datetimeNow);
    }
    ui->te_range_start->setDateTime(datetimeOld);
    ui->te_range_end->setMaximumDateTime(datetimeNow);
    ui->te_range_end->setDateTime(datetimeNow);

    ui->de_range_start->blockSignals(false);
    ui->te_range_start->blockSignals(false);
    ui->de_range_end->blockSignals(false);
    ui->te_range_end->blockSignals(false);
}

/*!
 * @brief Update the date and time ranges based on any changes made to them.
 *
 */
void CameraLogs::updateDateTimeRange()
{
    QDateTime dateTimeNow   = QDateTime::currentDateTime();
    QDateTime dateTimeStart(ui->de_range_start->date(), ui->te_range_start->time());
    QDateTime dateTimeEnd(ui->de_range_end->date(), ui->te_range_end->time());

    // Don't want to get signals when changing this data.
    ui->de_range_start->blockSignals(true);
    ui->de_range_end->blockSignals(true);
    ui->te_range_start->blockSignals(true);
    ui->te_range_end->blockSignals(true);

    // Make sure the ending date/time make sense.
    if (dateTimeEnd > dateTimeNow)
    {
        dateTimeEnd = dateTimeNow;
    }

    // Can continue to move forward the maximum date/time for the end.
    ui->de_range_end->setMaximumDateTime(dateTimeNow);

    // If the ending date is prior to the current date, then there
    // is no limit on the time setting. Otherwise, we cannot allow
    // the user to enter in later than "now".
    if (dateTimeEnd.date() < dateTimeNow.date())
        ui->te_range_end->clearMaximumTime();
    else
        ui->te_range_end->setMaximumDateTime(dateTimeNow);

    // If the new end is greater than the starting time push it back.
    if (dateTimeStart >= dateTimeEnd.addDays(-1))
    {
        // Push it back by at least 1 day.
        dateTimeStart = dateTimeEnd.addDays(-1);
        ui->de_range_start->setDateTime(dateTimeStart);
    }

    // Check to make sure the time edit is setup properly.
    if (dateTimeStart < dateTimeEnd.addDays(-1))
    {
        // We are more than a day from the ending time, so the user can pick any
        // time for starting the logs.
        ui->te_range_start->clearMaximumDateTime();
    }
    else
    {
        // We have to limit the maximum start time so that we have at
        // least one day in the range.
        ui->te_range_start->setMaximumDateTime(dateTimeEnd);
    }
    ui->te_range_start->setDateTime(dateTimeStart);

    ui->de_range_start->blockSignals(false);
    ui->de_range_end->blockSignals(false);
    ui->te_range_start->blockSignals(false);
    ui->te_range_end->blockSignals(false);

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "DATE/TIME CHANGED = " << dateTime.toString("yyyy-MM-dd hh:mm:00");
    resetJournalCtl();
}

