/*!
 * @file camera_log.h
 * @brief This file contains the definitions for the CameraLog class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_LOGS_H
#define CAMERA_LOGS_H

//--------------------------------------------------------------------
// Include Files
//--------------------------------------------------------------------
#include <QDockWidget>
#include <stdint.h>
#include <memory>
#include "mainwindow.h"

namespace Ui
{
class CameraLogDock;
}

namespace CAMERA_LOGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define the priority levels for log messages being saved.
//! The comments for each one match the "Description" column in systemd/journal
//! for Arch Linux.
typedef enum
{
    LEVEL_EMERGENCY = 0,    //!< System is unusable.
    LEVEL_ALERT,            //!< Should be corrected immediately.
    LEVEL_CRITICAL,         //!< Critical conditions.
    LEVEL_ERROR,            //!< Error conditions.
    LEVEL_WARNING,          //!< May indicate that an error will occur if action not taken.
    LEVEL_NOTICE,           //!< Events that are unusual, but not error conditions.
    LEVEL_INFORMATIONAL,    //!< Normal operational messages requiring no action.
    LEVEL_DEBUG,            //!< Information useful to developers for debugging the application.
    LEVEL_END               // Keep at the end.
} PRIORITY;

}   // End of namespace CAMERA_LOGS.

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class FiJournalCtl;
class CameraMDI;
class CameraLogSearchDialog;

//--------------------------------------------------------------------
// Class CameraLogs
//--------------------------------------------------------------------
class CameraLogs : public QDockWidget
{
    Q_OBJECT

public:
    explicit CameraLogs(QWidget *parent = nullptr);
    ~CameraLogs();

    static QString getPriorityString(CAMERA_LOGS::PRIORITY priorityLevelToSet);

    bool    initializeCameraLog(CameraMDI *camera_ptr);
    bool    isCameraAttached(CameraMDI *camera_ptr);
    void    removeCameraLog(CameraMDI *camera_ptr);

protected:
    //virtual void    changeEvent( QEvent *event );
    virtual void    closeEvent(QCloseEvent *event);
    virtual void    contextMenuEvent(QContextMenuEvent *event_ptr);

private slots:
    void    doActiveCameraChanged(QMdiSubWindow *window);

    void    doByDaySelect();
    void    doByDateSelect();

    void    doSave();
    void    doSearch();

    void    doFollowLog();

    void    doFasMotionFilterSelect();
    void    doCameraServerFilterSelect();
    void    doNoFilterSelect();
    void    doWrapLines();

    void    doLogFilterMenuChanged(QObject *obj_ptr);
    void    doLogLevelChanged(int index);
    void    doLogLevelMenuChanged(QObject *obj_ptr);
    void    doWordLinesMenuChange(QObject *obj_ptr);

    void    doSyncEndDateTimeToNow();

    void    doOptionGroupSelect();
    void    doRangeGroupSelect();

    void    stopLogSearch();
    void    nextLogSearch();

    void    updateDayRange(int index);
    void    updateDateTimeForDayRange();
    void    updateDateTimeRange();
    void    updateLogData(QString log_data);

signals:
    //void    dateTimeChangedSignal();  //!< If object wants to signal date/time change.

private:
    Ui::CameraLogDock *ui;

    // Context menu actions
    QMenu           *logLevelMenu;          //!< Log level reset by sub-menu.
    QList<QAction *> logLevelActionList;    //!< Log level submenu actions.
    QActionGroup    *logLevelActionGroup;   //!< Group our log level actions so they behave together properly.

    QMenu           *logFilterMenu;         //!< Log filter reset by sub-menu.
    QList<QAction *> logFilterActionList;   //!< Log filter submenu actions.
    QActionGroup    *logFilterActionGroup;  //!< Group our log filter actions so they behave together properly.

    QList<QAction *> logMenuActionList;     //!< List of our log menu actions.

    std::shared_ptr<QMenu> logContextMenu;    //!< The context menu for our log area.

    void    closeupJournal();
    QAction *createMenuAction(const QString& menuItemString,
                              bool checkable,
                              bool visibleIcon,
                              QList<QAction *> *actionListPtr, QActionGroup *actionGroupPtr);

    bool    initializeContextMenus();
    bool    resetContextMenu();
    void    requestLogData();
    bool    resetJournalCtl();
    bool    setCamera(CameraMDI *camera_ptr);
    void    setPriorityLevel(CAMERA_LOGS::PRIORITY priorityLevelToSet);
    void    syncRangeType();
    bool    usingFasMotionFilter();
    bool    usingCameraServerFilter();
    bool    usingNoFilter();

    CAMERA_LOGS::PRIORITY activePriority;

    bool initialized;                   //!< Is this dialog properly initialized?
    bool haveSetCameraChangeConnect;    //!< This will be set first initialization.

    CameraMDI   *attachedCameraPtr;     //!< The log data being saved belongs to this camera.

    QString logSearchString;            //!< The search string we are using to find in our log entry.

    bool useFasMotionFilter;            //!< True if filter on FasMotion entries.
    bool useCSRVFilter;                 //!< True if filter on camera server entries.
    bool useNoFilter;                   //!< True if no filters being used on log request.
    bool useWordWrap;                   //!< True if log information is word-wrapped.

    bool useRangeByDay;                 //!< True if using a day range rather than a date range.
    std::unique_ptr<CameraLogSearchDialog> searchDlg;   //!< A search dialog for finding strings in the log.
    std::unique_ptr<FiJournalCtl> journal;
};

#endif // CAMERA_LOGS_H
