/*!
 * @file camera_mdi.h
 * @brief This is the header file for the CameraMDI class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_MDI_H
#define CAMERA_MDI_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <QWidget>

#include "fi_comm.h"
#include "fi_camreg_group.h"
#include "fsp_socket.h"
#include "fi_saver.h"

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QEvent;
class QLabel;
class QMdiSubWindow;
class QTimer;
class QTimerEvent;
class QVBoxLayout;

class CameraInfo;
class CameraReview;
class CameraStream;
class CameraVideo;
class fiLoadCaptureDlg;
class fiProgBar;
class fiProgressDialog;
class fiSensor;
class FSPSocket;
class SaveVideo;

namespace CAMERA_MDI_SETTINGS
{
class PfmdCache;
}

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class CameraMDI : public QWidget
{
    Q_OBJECT

public:
    explicit CameraMDI(QWidget *parent = nullptr);
    ~CameraMDI();

    static const QString fflr_feature_name;
    static QString  get_config_file();
    static bool     is_series3(const FI_COMM::FI_CAMERA_INFO *info_ptr);

    void        set_camera_info(FI_COMM::FI_CAMERA_INFO *cam_info);
    const FI_COMM::FI_CAMERA_INFO *get_camera_info() const;

    void        allow_multicam_autosave()             { multicam_save_allowed = true; start_autosave(); }
    void        attach_to_playback(bool play_forward);
    bool        autoplay();
    uint32_t    blob_size();
    double      calc_frame_time(uint32_t frame_num);
    double      calc_frame_time_from_PFMD(const uint8_t *pfmd_ptr);
    uint64_t    calc_max_frames();
    void        calc_trigger_frame_start(uint32_t start_frame, uint32_t end_frame);
    bool        cap_is_allowed();
    void        change_heartbeat(uint32_t new_rate);
    bool        comms_are_ready()                     { return ok_to_talk; }
    void        copy_finished(QString whoCalledMe);
    void        disable_event_markers()               { use_event_markers = false; }
    void        disable_pfmd_autodownload()           { autodownload_metadata = false; }
    void        fpn_calibration(bool do_factory = false);
    void        free_partition(bool free_all);
    void        pause_cache_refresh(bool do_pause = true);
    bool        prepare_smb_filepath(const QString& media_path, const QString& name_part, const QString& file_id, const QString& file_ext, QString& smb_file_name);
    void        prnu_calibration();
    void        refresh_autosave_buffers();
    bool        verify_dirty_overwrite(bool free_all_on_overwrite, const QString& operation_label);

    QList<QAction *> get_camera_actions();
    CameraInfo  *get_info_widget()                    { return camera_info; }
    QString     get_camera_name();
    uint32_t    get_clip_end();
    uint32_t    get_clip_start();
    uint32_t    get_current_frame_num();

    constexpr QString& get_fflr_device_label()        { return fflr_device_label; }
    constexpr QString& get_fflr_device_name()         { return fflr_device_name; }
    constexpr QString& get_fflr_device_mountdir()     { return fflr_device_mountdir; }
    constexpr QString& get_fflr_device_tag()          { return fflr_device_tag; }
    constexpr uint32_t get_fflr_device_type()         { return fflr_device_type; }
    constexpr uint32_t get_fflr_session_len()         { return fflr_session_len; }

    uint32_t    get_ip_addr();
    QString     get_ip_addr_pretty();
    QJsonDocument get_json_str()                      { return jsonMedia; }
    QString     get_mac();
    int         get_missed_frames()                   { return missed_frame_count; }
    int         get_pfmd_len();
    int         get_packet_len();
    const QByteArray  *get_partition_metadata() const { return &partition_metadata; }
    CameraVideo *get_player()                         { return video_player; }
    QByteArray  *get_raw_metadata()                   { return &raw_metadata; }
    int         get_retry_count()                     { return retry_frame_list.count(); }
    CameraReview *get_review_controls()               { return review_controls; }
    SaveVideo   *get_saver()                          { return save_video; }
    fiSensor    *get_sensor()                         { return sensor_ptr; }
    CameraStream    *get_streamer()                   { return streamer; }
    int         get_tt_base()                         { return tt_time; }
    int         get_tt_count()                        { return tt_count; }
    int         get_tt_remaining()                    { return tt_timer.remainingTime() / 1000; }
    double      get_trigger_event();
    int         get_trigger_frame_for_save()          { return trigger_frame; }
    double      get_trig_frame_start()                { return trigger_frame_start; }
    const QString   get_version();
    int         get_vt_area()                         { return vt_area; }
    int         get_vt_change()                       { return vt_change; }
    bool        get_vt_mark_only()                    { return (!is_series3() && vt_mark_only); }
    int         get_vt_refresh()                      { return vt_refresh; }

    void        group_read_cancel(uint32_t reg_addr);

    void        ae_reticle_show(bool show_reticle);
    void        ae_reticle_reset();
    bool        ae_reticle_visible();
    void        vt_reticle_show(bool show_reticle);
    void        vt_reticle_reset();
    bool        vt_reticle_visible();
    bool        vt_software();
    bool        in_calibration();
    bool        is_autosave_configured();
    bool        is_autosave_enabled();
    bool        is_autosave_overlapping();
    bool        is_fflr_enabled(uint32_t& ltr_type);
    bool        is_fflr_underflow_recovering();
    bool        is_busy();
    bool        is_grouped();
    bool        is_local();
    bool        is_ltr_enabled(uint32_t& ltr_type);
    bool        is_offline();
    bool        is_pm_downloading()                   { return pm_download_in_progress; }
    bool        is_series3();
    bool        is_sync_in_enabled();
    bool        is_sync_in_receiving_pulse();
    bool        is_sync_in_warning_in_threshold(int msThreshold);
    void        pause_video_request(const QString whoCalledMe = "Caller wasn't set.");
    void        process_arm(bool user_queried, bool allow_discard);
    void        push_config_file(QString& file_name);
    bool        raw_metadata_ok();
    void        reboot();
    void        release_buffer(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr);
    void        reset();
    void        resume_video_request();
    void        report_sync_in_suspension();
    void        retry_missing_frames();
    bool        retrying_frames()                     { return retry_in_progress; }
    bool        review_has_received_a_frame();
    bool        safe_to_arm(bool& user_queried, bool& allow_discard);
    void        set_grouped(bool is_grouped);
    void        set_inactive();
    bool        set_fflr_info(const QString& fflr_device_label, const QString& fflr_device_name, const QString& fflr_device_mountdir, const QString& fflr_tag, uint32_t fflr_device_type, uint32_t fflr_session_len);
    void        set_fflr_user_discarded_parts(bool discarded)   { fflr_user_discards_parts = discarded; };
    void        set_offline(bool close_window);
    void        set_reboot(bool rebooting)                 { camera_ptr->rebooting = rebooting; }
    void        set_vt_area(int new_area, bool send_to_camera = true);
    void        set_vt_change(int new_change, bool send_to_camera = true);
    void        set_vt_mark_only(bool only_mark, bool send_to_camera = true);
    void        set_vt_refresh(int new_refresh, bool send_to_camera = true);
    void        set_vt_region(const QRect& new_rect);
    void        set_window(QMdiSubWindow *win_ptr);
    bool        shutdown_in_progress()                { return in_shutdown; }
    void        start_copy(QString& camera_path);
    void        start_local_save(bool reset_retry_count = false);
    void        start_session_save();
    void        stream_save_video(uint32_t clip_start, uint32_t clip_end,
                                  FI_SAVER::STREAM_TYPE streamType, FI_SAVER::IMAGE_TYPE imageType);
    void        track_exposure(bool enable_track);
    bool        tracking_exposure();
    void        update_frame();
    void        update_save_progress(uint32_t progress);
    void        update_wifi_status();
    QString     user_friendly_name();

    // Utility methods to perform common operations.
    QByteArray  *read_camera_buffer(uint32_t buf_addr, bool *ok = nullptr);
    uint32_t    read_camera_cache(uint32_t reg_addr, bool *ok = nullptr);
    void        read_camera_registers(QSharedPointer<FiCamRegGroup> camGroupRead);
    void        start_camera_read_buf(uint32_t buf_addr, uint32_t buf_len, bool do_expire = true);
    void        start_camera_read_reg(uint32_t reg_addr);
    bool        waiting_for_retry_frame()           { return waiting_for_frame; }
    void        write_camera_buffer(uint32_t buf_addr, const QByteArray& buffer);
    void        write_camera_register(uint32_t reg_addr, uint32_t reg_value);

signals:
    void        ext_arm_in_enabling();
    void        ext_arm_in_disabling();
    void        closing();
    void        copy_timeout();
    void        dearm_session_discard();
    void        dearm_session_keep();
    void        event_search_complete();
    void        group_read_complete(uint32_t ip_addr, uint32_t last_reg_read);
    void        fflr_status_ready(const QJsonDocument& jsonFFLRDoc);
    void        media_info_ready(const QJsonDocument& jsonMedia);
    void        pm_metadata_received(uint32_t ip_addr);
    void        raw_metadata_received();
    void        review_panel_ready_to_go();
    void        save_error_reported();
    void        save_timeout();
    void        save_finished();
    void        sync_in_resumed();
    void        sync_in_timeout();
    void        tt_expired();
    void        video_is_paused();
    void        video_is_streaming();

public slots:
    void    camera_offline(uint32_t ip_addr, bool offline);
    void    cancel_save();
    void    do_camera_arm();
    void    do_camera_autosave_enable(bool do_enable);
    void    do_camera_live();
    void    do_camera_review();
    void    do_camera_save();
    void    do_camera_save_all();
    void    do_camera_trigger();
    void    do_ext_arm_in_disabling();
    void    do_ext_arm_in_enabling();
    void    do_fflr_underflow_discard();
    void    time_trigger(int time_value);
    void    update_camera_menu();

protected:
    void closeEvent(QCloseEvent *event) override;
    void contextMenuEvent(QContextMenuEvent *event_ptr) override;
    void focusInEvent(QFocusEvent *event) override;
    void showEvent(QShowEvent *event) override;
    void timerEvent(QTimerEvent *event) override;

private slots:
    void    ae_track();
    void    calibration_continuation();
    void    calibration_finalize(uint32_t ip_addr, uint32_t reg_addr);
    void    check_for_busy();
    void    commitData();
    void    delayed_state_check();
    void    do_autosave_setup();
    void    do_black_level();
    void    do_camera_events();
    void    do_camera_load();
    void    do_camera_pfmd();
    void    do_camera_power_down();
    void    do_camera_preferences();
    void    do_camera_reboot();
    void    do_camera_samba();
    void    do_camera_snapshot();
    void    do_camera_update();
    void    do_config_factory_reset();
    void    do_config_load();
    void    do_config_save();
    void    do_config_time_date();
    void    do_config_wifi();
    void    do_fflr_review(const QJsonDocument& json_FFLR_doc);
    void    do_open_device();
    void    do_release_buffer();
    void    do_release_all_buffers();
    void    do_truncate_clip();
    void    do_truncate_release();
    void    do_view_aiming_cursor();
    void    do_view_info_cursor();
    void    do_view_localtime();
    void    do_view_review();
    void    do_view_summary();
    void    done_saving();
    void    error_saving();
    void    event_search_canceled();
    void    fflr_playback_closing();
    void    finish_arm();
    void    finish_shutdown();
    void    framerate_changed(int new_value);
    void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    bool    last_frame_missing_check();
    void    load_complete();
    void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    void    new_camera_name();
    void    progress_check();
    void    raw_metadata_cancel();
    void    raw_metadata_download();
    void    raw_metadata_ready();
    void    raw_metadata_start();
    void    raw_metadata_timeout();
    void    raw_metadata_wrapup();
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr);

    void    chk_refresh_cache_read(uint32_t ip_addr, uint32_t reg_addr);
    void    chk_camera_preference_read(uint32_t ip_addr, uint32_t reg_addr);
    void    chk_rec_progress_read(uint32_t ip_addr, uint32_t reg_addr);

    void    review_panel_ready();
    void    save_receive_count_complete();
    void    saver_is_idle();
    void    search_for_events();
    bool    set_vtrig_marker(int index, uint32_t& event_word, uint32_t *compare_ptr, uint32_t vtrig_word, uint32_t last_vtrig_word);
    void    stream_metadata();
    void    streamer_faulted();
    void    streamer_image_received();
    void    streamer_image_timeout();
    void    streamer_is_closed();
    void    streamer_is_paused();
    void    streamer_is_ready();
    void    streamer_is_streaming();
    void    sync_in_warning_closed();
    void    time_trigger_expired();
    void    unsol_socket_is_ready();
    void    update_pfmd();
    void    update_rec_progress();
    void    user_stream_request();
    void    vt_track();
    void    vt_changed();
    void    wait_to_transition();
    void    window_activating();
    void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

    void    video_current_frame_changed();
    void    video_jump_time();
    void    video_play_forward();
    void    video_play_reverse();
    void    video_rate_changed(int new_rate);
    void    video_review_stop();
    void    video_review_stop_wait();

private:
    bool        active_window();
    bool        add_to_retry_list(uint32_t frame_0_based);
    void        ae_baseline();
    void        ae_process();
    void        block_download_start(uint32_t reg_addr, int mem_len, int max_xfer);
    void        block_download_next();
    void        block_download_finish();
    void        build_data();
    void        check_for_metadata();
    void        cleanup_save();
    void        configure_wifi();
    void        copy_video_timeout();
    void        disconnect_video_player();
    bool        dearm_camera(bool user_queried, bool allow_discard);

    bool        fflr_check_before_arming();

    void        merge_overlay(uint32_t seq_number, FI_MEMPOOL::FiImageBuffer *imageBufferPtr);

    bool        parse_config_file(QString& file_name, uint32_t reg_addr, uint32_t& reg_value);
    bool        parse_config_file(QString& file_name, uint32_t reg_addr, QString& reg_value);

    void        play_video(bool play_forward);
    void        prepare_to_download();
    void        process_autosave(uint32_t reg_value);
    void        process_ccp_write(uint32_t write_status);
    void        process_fpn_change(uint32_t reg_value);
    void        process_json_fflr();
    void        process_json_media();
    void        process_partition_capture();
    void        process_partition_state(void);
    void        process_received_chunk(uint32_t seq_number);
    void        process_received_image(uint32_t seq_number);
    void        process_received_metadata(std::unique_ptr<FI_MEMPOOL::FiImageBuffer>meta_ptr);
    void        process_state_change(uint32_t new_state, bool state_change_error = false);
    void        process_timeofday(uint32_t timeofday);
    void        push_config_file_next();
    void        read_frame_metadata();
    void        read_settings();
    void        recover_stream();
    void        rec_progress_state_check(bool capturing_video);
    void        refresh_cache();
    void        release_buffer(bool quiet);
    void        release_all_buffers(bool quiet);
    void        request_active_metadata(const QString& who_called);
    void        request_pfmd(uint32_t frame_num, bool active = true);
    void        reset_event_info();
    QString     save_raw_metadata();
    void        save_settings();
    void        save_video_timeout();
    void        set_rec_progress(uint32_t camera_state);
    int         setup_color_prnu_calibration();
    void        start_autosave();
    void        stop_lock_timer();
    void        stream_setup(uint32_t new_state, const QString& whoCalledMe);
    void        test_complete(bool got_timeout);
    void        update_camera_save_progress();
    void        update_event_info();
    void        update_media();
    void        update_rec_info();
    void        update_status();
    bool        valid_config_file(QString& file_name);
    void        vt_baseline_calc();
    void        vt_process();

private:
    FI_COMM::FI_CAMERA_INFO *camera_ptr;        //!< The camera specific information for the camera we're associated with.

    static const QString sync_in_warning_msg;
    static const QString sync_in_warning_title_msg;
    static const QString prnu_setup_msg;
    static const QString prnu_title_str;
    static const QString prnu_red_path;
    static const QString prnu_green_path;
    static const QString prnu_blue_path;
    static const char XDG_RUNTIME_DIR[];

    bool        active_metadata_requested;      //!< The GIGE_REG_ACTIVE_METADATA has been requested we are waiting for confirmation to start download.
    int         ae_histo_count;                 //!< Number of histograms captured for the baseline.
    bool        ae_ready;                       //!< Flag to indicate when we are ready to start processing.
    int         ae_track_state;                 //!< Current status of our tracker.
    int         ae_target_Y;                    //!< Target luminance for autoexposure.
    bool        autodownload_metadata;          //!< True if we're to automatically download the per-frame metadata, false otherwise.
    bool        auto_offset_cal;                //!< True if the camera is to automatically adjust the offset during calibration.
    bool        autosave_pending;               //!< True if we've started the save process, false otherwise.
    bool        autosave_lockout;               //!< True if we're ignoring GIGE_REG_AUTOSAVE (locked out).
    int         block_download_id;              //!< The register address of the block we're currently downloading.
    int         block_download_len;             //!< The number of bytes we're expecting to read.
    int         block_download_max_xfer;        //!< The maximum number of bytes we can transfer in one read.
    bool        cache_refresh_holdoff;          //!< Flag to indicate if we should hold off refreshing the cache.
    int         calibration_event;              //!< Indicator of what, if any, calibration event we're processing.
    QFile       config_file;                    //!< Configuration file object to read/write the camera's configuration information.
    int         config_length;                  //!< The number of bytes the configuration information require.
    uint32_t    count_metafiles_copied;         //!< How many metafile sequence numbers have come through stream image received?
    bool        force_frame_update;             //!< Regardless of if this is the same frame, we want the camera to resend it.
    QString     fflr_device_label;              //!< FasFireLR device "label" from the GIGE_REG_MEDIA_JSON data.
    QString     fflr_device_name;               //!< FasFireLR device "name" from the GIGE_REG_MEDIA_JSON data.
    QString     fflr_device_mountdir;           //!< FasFireLR device "mountdir" from the GIGE_REG_MEDIA_JSON data.
    QString     fflr_device_tag;                //!< FasFireLR device "tag" built from the GIGE_REG_MEDIA_JSON data. This is what the user sees in the combobox.
    uint32_t    fflr_device_type;               //!< FasFireLR device "type" from the GIGE_REG_MEDIA_JSON data which is a GIGE_MEDIA_XXX type.
    uint64_t    fflr_session_len;               //!< FasFireLR session partition size in 256 MiB units.
    bool        fflr_underflow_discarded;       //!< FasFireLR was the underflow partial recording discarded by user.
    bool        fflr_user_discards_parts;       //!< Initialized false when arming FasFireLR, if dearmed or if underflow error and user discards, this is true.
    int         heartbeat_rate;                 //!< The number of milliseconds between access to the camera.
    bool        in_shutdown;                    //!< True if we (the MDI child window) are closing (shutting down).
    bool        in_startup;                     //!< Flag to indicate if we're in startup or not.
    bool        in_truncate;                    //!< Flag to indicate if we're in the process of truncating the recording.
    QString     json_fflr_str;                  //!< Accumulation buffer for JSON media data.
    QString     json_build_str;                 //!< Accumulation buffer for JSON media data.
    QJsonDocument jsonFFLRDoc;                  //!< JSON FasFireLR JSON status to be shared to all.
    QJsonDocument jsonMedia;                    //!< JSON media data to be shared to all.
    uint32_t    last_received_frame;            //!< Sequence number of the last frame we've received.
    uint32_t    last_requested_frame;           //!< Frame number of the last requested frame, so we don't issue duplicate requests.
    int         last_window_state;                //!< The state the window was in before it was minimized.
    int         lock_timer_id;                  //!< Timer ID we use to poll the camera. Set to a little less than the heartbeat_rate value.
    int         missed_frame_count;             //!< The final number of missed frames in a download attempt.
    bool        multicam_save_allowed;          //!< Flag to control autosave processing when in a group.

    QList<QSharedPointer<FiCamRegGroup>> group_reads;   //!< Manages all camera register group reads through the camera.

    bool        ok_to_talk;                     //!< True if our communications are ready, false otherwise.
    uint32_t    old_camera_state;               //!< Camera state we save away to tell if the camera actually changed state.
    uint32_t    requested_state;                //!< The requested camera state.
    QByteArray  partition_metadata;             //!< The metadata for the active partition on the camera.
    QList<uint32_t> active_frame_requests;      //!< Frame number we're requesting PFMD info using the "active" frame register interface.
    QList<uint32_t> pfmd_frame_requests;        //!< Frame number we're requesting PFMD info for if not active.
    CAMERA_MDI_SETTINGS::PfmdCache   *pfmd_info; //!< The PFMD information cache.
    bool        pm_download_only;               //!< True if we only want the pm_metadata without a save going on.
    bool        pm_download_in_progress;        //!< True if we are in the middle of a download.
    bool        play_video_forward;             //!< True if we're playing the video normally, false if in reverse.


    QByteArray  raw_metadata;                   //!< The raw per-frame metadata we download.
    int         raw_metadata_len;               //!< The expected size of the raw per-frame metadata being downloaded.
    QString     raw_path;                       //!< The full path name to the temporary raw metadata file, if it exists.
    bool        release_all;                    //!< True if we're in the process of releasing all dirty partitions.
    bool        restart_timer;                  //!< A flag to indicate if we should (or not) restart the frame timer.
    int         retry_frame_count;              //!< Number of times we've tried to retransmit a frame.
    QList<uint32_t> retry_frame_list;           //!< List of frame sequence numbers that we need to retry.
    QList<uint32_t> retry_frames_reqd;          //!< List of frame sequence numbers pulled from retry frame list and requested.
    bool        retry_in_progress;              //!< True if we're retrying missing frames, false otherwise.
    int         retry_count_raw_metadata;       //!< Retry counter for the raw metadata.
    bool        save_cleanup;                   //!< True if a save operation completed.
    SaveVideo   *save_video;                    //!< Pointer to our video saving object.
    fiSensor    *sensor_ptr;                    //!< Pointer to our sensor object.
    int         ss_delay_counter;               //!< Single step delay counter to hold off single stepping to give the camera a chance to settle.
    bool        stream_active;                  //!< True if the stream is currently running (active).

    std::unique_ptr<fiProgressDialog> sync_in_warning_dlg;    //!< The sync-in warning "progress" dialog if frame stream time-outs are active.
    int         sync_in_warning_state;          //!< Keeps track of what the user is seeing when sync-in suspended signal warning occurs.
    QDateTime   time_of_sync_in_suspension;     //!< Time of recorded sync-in suspension.  Initial construction sets an empty time.  Use isValid() to determine.

    CameraStream    *streamer;                  //!< Pointer to our streamer object.
    QTimer      tt_timer;                       //!< Our time trigger timer.
    int         tt_count;                       //!< Number of times our time trigger expired.
    int         tt_time;                        //!< Our time trigger base value.
    uint32_t    trigger_frame;                  //!< Hold the trigger frame for the partition, so we don't need to read it each time.
    double      trigger_frame_start;            //!< Start time for the trigger frame.
    int         vt_area;                        //!< Number of pixels to monitor for video trigger.
    int         vt_change;                      //!< Delta change in pixel values for video trigger.
    QVector<int> vt_baseline;                   //!< Our baseline data we use for the video trigger calculations.
    int         vt_histo_count;                 //!< Number of histograms captured for the baseline.
    bool        vt_mark_only;                   //!< True if we're only marking events, not actually triggering.
    int         vt_refresh;                     //!< Number of frames before we reset the baseline.
    bool        vt_ready;                       //!< Flag to indicate when we are ready to start processing.
    int         vt_track_state;                 //!< Current state of the video trigger algorithm.
    bool        waiting_for_frame;              //!< True if we're waiting on a missing frame.

    std::unique_ptr<fiLoadCaptureDlg> loadCaptureDlg;   //!< A load capture progress dialog.
    std::unique_ptr<fiProgressDialog> progressDialog;   //!< A simple progress dialog for operational status.
    std::unique_ptr<fiProgressDialog> xferDialog;       //!< A xfer progress dialog.

    QVBoxLayout *layout_ptr;                    //!< Our window layout.
    CameraInfo  *camera_info;                   //!< Top pane of the window.
    CameraVideo *video_player;                  //!< Middle pane of the window.
    fiProgBar   *rec_progress;                  //!< Bottom pane of the window.
    CameraReview    *review_controls;           //!< Also bottom pane, used when in review mode.

    bool        use_event_markers;              //!< True if we're using event markers.
    bool        event_logic_is_and;             //!< True if we're using AND for the event logic, false for OR.
    bool        downloading_pfmd;               //!< True if we're downloading the raw per-frame metadata.
    QList<int>  event_selections;               //!< Which events we're interested in on which lines.
    QList<int>  marker_list;                    //!< List of frames with events which pass the selection criteria.
    int         search_start_index;             //!< Which frame to start searching on.

    int         frame_rate_counter;             //!< Simple counter to indicate the number of frames/second we're getting.
    int         frame_rate_timer;               //!< Timer ID used for the counter.
};

#endif // CAMERA_MDI_H
