/*!
 * @file camera_preferences.cpp
 * @brief This is the implementation of the CameraPreferences dialog.
 *
 * This dialog allows the user to configure some of the per-camera
 * settings, such as gamma, etc.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "camera_preferences.h"
#include "ui_camera_preferences.h"

#include "camera_mdi.h"
#include "gige_msg.h"
#include "logger.h"

namespace CAMERA_PREFS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The value for a Gamma of 1.0.
const int GAMMA_1pt0            = 100;

//! The value for a Gamma of 2.2.
const int GAMMA_2pt2            = 220;
}

using namespace CAMERA_PREFS;

/*!
 * @brief The CameraPreferences dialog constructor.
 */
CameraPreferences::CameraPreferences(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::CameraPreferences)
{
    CameraMDI *my_camera = qobject_cast<CameraMDI *>(parent);

    // Setup the UI.
    ui->setupUi(this);


    // The External I/O voltage isn't for series3 cameras.
    if (my_camera && (my_camera->is_series3() == false))
    {
        // We are a seriesNC camera.
        ui->grp_ext_io->setVisible(true);
    }
    else
    {
        // We either do not have a camera pointer, or the camera is series3.
        ui->grp_ext_io->setVisible(false);
    }

    // Connect our signal/slots.
    connect(ui->spin_jpeg, QOverload<int>::of(&QSpinBox::valueChanged), ui->sld_jpeg, &QSlider::setValue);
    connect(ui->sld_jpeg, &QSlider::valueChanged, ui->spin_jpeg, &QSpinBox::setValue);

    // Update the layout.
    adjustSize();
    update();
}

/*!
 * @brief The CameraPreferences dialog destructor.
 */
CameraPreferences::~CameraPreferences()
{
    delete ui;
}

/*!
 * @brief Return true if the external I/O should be 5V, false for 3.3V.
 */
bool CameraPreferences::isExtIO_5V()
{
    return ui->rb_5V->isChecked();
}


/*!
 * @brief Set the appropriate indicate for the external I/O voltate.
 * @param enable    True for 5V, false for 3.3V.
 */
void CameraPreferences::extIO_5V(bool enable)
{
    if (enable)
        ui->rb_5V->setChecked(true);
    else
        ui->rb_3_3V->setChecked(true);
}

/*!
 * @brief Return true if the autodownload checkbox is checked, false otherwise.
 */
bool CameraPreferences::get_autodownload_enable(void)
{
    return ui->auto_download_metadata->isChecked();
}

/*!
 * @brief Return true if the autoplayback checkbox is checked, false otherwise.
 */
bool CameraPreferences::get_autoplayback_enable(void)
{
    return ui->auto_playback->isChecked();
}

/*!
 * @brief Return the gamma value the user selected (1.0 or 2.2).
 */
int CameraPreferences::get_gamma(void)
{
    int gamma_value;

    // Has the user selected 1.0 or 2.2 gamma?
    if (ui->rb_10->isChecked())
        gamma_value = GAMMA_1pt0;
    else
        gamma_value = GAMMA_2pt2;

    return gamma_value;
}

/*!
 * @brief Return the JPEG quality value the user selected.
 */
int CameraPreferences::get_jpeg_quality(void)
{
    return ui->sld_jpeg->value();
}

/*!
 * @brief Check or clear the autoplayback checkbox based on the flag passed in.
 * @param enabled   True to check, false to clear, the control.
 */
void CameraPreferences::set_autoplayback(bool enabled)
{
    ui->auto_playback->setChecked(enabled);
}

/*!
 * @brief Set the appropriate gamma radio button based on the passed in value.
 * @param gamma_value   The gamma value of 2.2 (200) or 1.0 (100).
 */
void CameraPreferences::set_gamma(int gamma_value)
{
    // Update our UI based on the setting.
    switch (gamma_value)
    {
    case GAMMA_1pt0:
        ui->rb_10->setChecked(true);
        break;
    case GAMMA_2pt2:
        ui->rb_22->setChecked(true);
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Invalid gamma_value:" << gamma_value;

        // By default, use 1.0.
        ui->rb_10->setChecked(true);
        break;
    }
}

/*!
 * @brief Set the UI controls to the JPEG value passed in.
 * @param quality_value     The current JPEG quality value.
 */
void CameraPreferences::set_jpeg_quality(int quality_value)
{
    // Put out a log if we get a strangely low value.
    if (quality_value < 20)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpectedly low quality_value:" << quality_value;
    }

    ui->sld_jpeg->setValue(quality_value);
    ui->spin_jpeg->setValue(quality_value);
}

/*!
 * @brief Check or clear the autodownload checkbox based on the flag passed in.
 * @param do_autodownload   True to check, false to clear, the control.
 */
void CameraPreferences::set_metadata_autodownload(bool do_autodownload)
{
    ui->auto_download_metadata->setChecked(do_autodownload);
}
