/*!
 * @file camera_review.cpp
 * @brief This file contains the implementation of the CameraReview class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "camera_review.h"
#include "ui_camera_review.h"

#include <QTimerEvent>

#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qwt_plot_canvas.h>
#include <qwt_plot_layout.h>

#include "camera_mdi.h"
#include "camera_stream.h"
#include "client_math.h"
#include "control_panel.h"
#include "eventfilter.h"
#include "fi_comm.h"
#include "fi_slider.h"
#include "fi_stringspinbox.h"
#include "logger.h"
#include "mainwindow.h"
#include "panel_reccontrols.h"
#include "playback_mdi.h"

namespace REVIEW_CONTROLS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Set to 0 if want QLOG_DEBUG logging for camera register group read complete status.
#define LOG_GROUP_READS     0

//! When the user "scrubs" the current frame slider, this generates a number of events which
//! take a fair amount of time to propogate through the system. In order to avoid a feedback
//! condition with the camera, we want to ignore unsolicited frame change events from the
//! camera for a little while after the user releases the slider.
const int           BLACKOUT_PERIOD         = 1500;

//! Period to delay waiting for communications to be ready.
const int           DELAY_UPDATE_TIME       = 500;

//! When we are in playback, update the frame at this rate.
const int           PLAY_UPDATE_TIME        = 200;

//! The maximum number of points along a plot we'll support if we're decimating.
const int           MAX_SAMPLE_LEN          = 10000;

//! The maximum number of points along a plot we'll support before we start to decimate the data.
const int           DECIMATE_THRESHOLD      = (4 * REVIEW_CONTROLS::MAX_SAMPLE_LEN);

//! Period we poll the camera for our register information.
const int           PANEL_UPDATE_PERIOD     = 1000;

//! Determines whether or not to output synchronization efforts by external source as
//! is in use for Ticket #302.
//static const int VERBOSE_SYNC_CLIP_LOG = 1;
static const int VERBOSE_SYNC_CLIP_LOG = 0;

}   // End of namespace REVIEW_CONTROLS.

using namespace REVIEW_CONTROLS;

//! Keep a constant of when a sanity check on CameraMDI parent pointer fails.
const QString CameraReview::PLAYER_MODE_RESTRICTION = QStringLiteral("Not available in player mode!");

/*!
 * @brief The CameraReview constructor.
 */
CameraReview::CameraReview(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::CameraReview)
{
    FI_comm     *comm_ptr;

    // We are being used in either a camera object or a playback object.
    CameraMDI   *camera_ptr     = qobject_cast<CameraMDI *>(parent);

    // Initialize our UI object.
    ui->setupUi(this);

    // Initialize the frame rate controls.
    ui->spin_frame_rate->setMinimum(1);
    ui->spin_frame_rate->setMaximum(100);
    ui->spin_frame_rate->setValue(DEF_PLAYBACK_RATE);

    // Initialize our variables.
    allow_frame_updates = true;
    blackout_id = 0;

    last_frame_num = 0;
    last_part_state = GIGE_PART_STATE_EMPTY;
    play_update_id = 0;
    player_mode = false;
    showing = true;
    ignore_updates = true;
    pb_stop = false;

    // Initialize our graph support.
    create_zoom_controls();
    create_chart_controls();

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // We are going to use a Queued connection type later, and Qt needs to know about uint32_t types.
    qRegisterMetaType<uint32_t>("uint32_t");

    // Initialize our signal/slots.
    if (camera_ptr != nullptr)
    {
        connect(comm_ptr, &FI_comm::read_reg_complete, this, &CameraReview::read_complete, Qt::QueuedConnection);
        connect(comm_ptr, &FI_comm::unsol_msg_received, this, &CameraReview::handle_unsol_msg, Qt::QueuedConnection);
        connect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraReview::write_complete, Qt::QueuedConnection);

        // Group register reads funnel through the CameraMDI object.
        connect(camera_ptr, &CameraMDI::group_read_complete, this, &CameraReview::chk_read_panel_regs, Qt::QueuedConnection);
        connect(camera_ptr, &CameraMDI::group_read_complete, this, &CameraReview::chk_read_playback_regs, Qt::QueuedConnection);
    }

    connect(ui->tb_play_reverse, &QToolButton::clicked, this, &CameraReview::play_reverse);
    connect(ui->tb_stop, &QToolButton::clicked, this, &CameraReview::playback_stop);
    connect(ui->tb_play_forward, &QToolButton::clicked, this, &CameraReview::play_forward);
    connect(ui->tb_time_jump, &QToolButton::clicked, this, &CameraReview::jump_time);
    connect(ui->tb_trash, &QToolButton::clicked, this, &CameraReview::truncate);

    connect(ui->tb_jump_first, &QToolButton::clicked, this, &CameraReview::jump_forward);
    connect(ui->tb_jump_last, &QToolButton::clicked, this, &CameraReview::jump_reverse);
    connect(ui->tb_trigger, &QToolButton::clicked, this, &CameraReview::jump_trigger);
    connect(ui->tb_step_forward, &QToolButton::clicked, this, &CameraReview::stepping_forward);
    connect(ui->tb_step_reverse, &QToolButton::clicked, this, &CameraReview::stepping_reverse);

    connect(ui->spin_frame_rate, qOverload<int>(&QSpinBox::valueChanged), this, &CameraReview::playback_rate_changed);
    connect(ui->spin_clip_end, qOverload<int>(&QSpinBox::valueChanged), this, &CameraReview::clip_end_changing, Qt::QueuedConnection);
    connect(ui->spin_clip_end, qOverload<int>(&QSpinBox::valueChanged), ui->clip_slider, &fiSpanSlider::setUpperPosition, Qt::DirectConnection);
    connect(ui->spin_clip_end, &QSpinBox::editingFinished, this, &CameraReview::clip_end_changed);
    connect(ui->pb_clip_end, &QPushButton::clicked, this, &CameraReview::clip_end_clicked);

    connect(ui->spin_clip_start, qOverload<int>(&QSpinBox::valueChanged), this, &CameraReview::clip_start_changing, Qt::QueuedConnection);
    connect(ui->spin_clip_start, qOverload<int>(&QSpinBox::valueChanged), ui->clip_slider, &fiSpanSlider::setLowerPosition, Qt::DirectConnection);
    connect(ui->spin_clip_start, &QSpinBox::editingFinished, this, &CameraReview::clip_start_changed);
    connect(ui->pb_clip_start, &QPushButton::clicked, this, &CameraReview::clip_start_clicked);

    connect(ui->clip_slider, &fiSpanSlider::lowerValueChanged, ui->spin_clip_start, &QSpinBox::setValue);
    connect(ui->clip_slider, &fiSpanSlider::upperValueChanged, ui->spin_clip_end, &QSpinBox::setValue);
    connect(ui->clip_slider, &fiSpanSlider::zoom_changing, this, &CameraReview::zoom_change);
    connect(ui->clip_slider, &fiSpanSlider::zoom_hiding, this, &CameraReview::zoom_hide);
    connect(ui->clip_slider, &fiSpanSlider::zoom_showing, this, &CameraReview::zoom_show);
    connect(ui->clip_slider, &fiSpanSlider::zoom_moving, this, &CameraReview::zoom_move);

    // We don't know which slider was released, so act like both of them did.
    connect(ui->clip_slider, &fiSpanSlider::sliderReleased, this, &CameraReview::clip_start_changed);
    connect(ui->clip_slider, &fiSpanSlider::sliderReleased, this, &CameraReview::clip_end_changed);
    connect(ui->clip_slider, &fiSpanSlider::sliderReleased, this, &CameraReview::clip_changed);

    // Make sure the span slider doesn't overlap the clip start/end.
    ui->clip_slider->setHandleMovementMode(FI_SPAN_SLIDER::NoOverlapping);

    connect(ui->spin_current_frame, &QSpinBox::editingFinished, this, &CameraReview::current_frame_changed);
    connect(ui->spin_current_frame, qOverload<int>(&QSpinBox::valueChanged), this, &CameraReview::cur_frame_changing);

    ui->sld_current_frame->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    connect(ui->sld_current_frame, &fiSlider::valueSlowChanged, this, &CameraReview::cur_frame_slow_update, Qt::QueuedConnection);
    connect(ui->sld_current_frame, &fiSlider::actionTriggered, this, &CameraReview::cur_frame_action);
    connect(ui->sld_current_frame, &fiSlider::sliderPressed, this, &CameraReview::cur_frame_pressed);
    connect(ui->sld_current_frame, &fiSlider::released, this, &CameraReview::cur_frame_released);
    connect(ui->sld_current_frame, &fiSlider::released, this, &CameraReview::current_frame_changed);
}

/*!
 * @brief The CameraReview destructor.
 */
CameraReview::~CameraReview()
{
    // Shutdown any timers which might still be running.
    if (blackout_id != 0)
    {
        killTimer(blackout_id);
        blackout_id = 0;
    }
    if (play_update_id != 0)
    {
        killTimer(play_update_id);
        play_update_id = 0;
    }

    // Clean up our zoom controls.
    delete pb_zoom_in;
    delete pb_zoom_out;
    delete zoom_scrollbar;

    // Release any raw data pointer we may have allocated.
    if (raw_data_ptr != nullptr)
    {
        delete []raw_data_ptr;
        raw_data_ptr = nullptr;
        raw_data_len = 0;
    }

    // Do we have existing arrays? If so, delete them.
    if (x_axis)
    {
        // We won't have one with out the others.
        delete []x_axis;
        delete []y_ao_axis;
        delete []y_ai_axis;
        delete []y_so_axis;
        delete []y_si_axis;
        delete []y_to_axis;
        delete []y_ti_axis;
    }

    // Release the UI object.
    delete ui;
}

/*!
 * @brief Attach to the camera's existing playback stream.
 */
void CameraReview::attach_playback()
{
    bool            play_forward;
    uint32_t        pb_control;
    CameraMDI       *my_parent      = qobject_cast<CameraMDI *>(parent());
    CameraStream    *my_stream      = nullptr;

    // Not available in player mode.
    if (player_mode)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Not available in player mode!";
        return;
    }

    // Make sure we're still playing.
    if (!is_playing())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Not playing.";
        return;
    }

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No parent!";
        return;
    }

    // See if we're already playing.
    if ((my_stream = my_parent->get_streamer()) && my_stream->is_playing())
    {
        // This is OK. It just means we probably initiated the playback
        // and we're getting here as a result of events on the camera causing
        // unsolicited messages to propogate back to us.
        // No need to carry on.
        return;
    }

    // What direction are we moving? The GIGE_PBC_REVERSE has the bit set
    // so we need to use that as the mask.
    pb_control = read_camera_cache(GIGE_REG_PB_CONTROL);
    play_forward = ((pb_control & GIGE_PBC_REVERSE) != GIGE_PBC_REVERSE);

    // Tell our parent to attach to the stream.
    my_parent->attach_to_playback(play_forward);
}

/*!
 * @brief Make the labels for the signals.
 */
void CameraReview::build_labels()
{
    QString label_str;
    REC_CONTROLS::EXT_IO_TYPES ext_io;
    PanelRecControls *controls_ptr;

    // Sanity check. Not for series3 cameras.
    if (is_series3())
        return;

    // Reset our list.
    signal_names.clear();

    // Get our record control panel object.
    controls_ptr = static_cast<PanelRecControls *>(MainWindow::GetInstance()->get_control_panel()->getPanel(CONTROL_PANEL::PANEL_REC_CONTROLS));

    // Change the labels of the remaining controls.
    label_str.clear();
    label_str.append("Ch0");
    ext_io = controls_ptr->getType(0);
    if (ext_io != REC_CONTROLS::EXT_IO_TYPES::EXTIO_NONE)
        label_str.append(QStringLiteral("/") + controls_ptr->getExtioLabel(ext_io));
    signal_names << label_str;

    label_str.clear();
    label_str.append("Ch1");
    ext_io = controls_ptr->getType(1);
    if (ext_io != REC_CONTROLS::EXT_IO_TYPES::EXTIO_NONE)
        label_str.append(QStringLiteral("/") + controls_ptr->getExtioLabel(ext_io));
    signal_names << label_str;

    label_str.clear();
    label_str.append("Ch2");
    ext_io = controls_ptr->getType(2);
    if (ext_io != REC_CONTROLS::EXT_IO_TYPES::EXTIO_NONE)
        label_str.append(QStringLiteral("/") + controls_ptr->getExtioLabel(ext_io));
    signal_names << label_str;
}

/*!
 * @brief Create our chart controls at runtime, rather than using Qt Designer.
 *
 * The reason for this is that, for some controls (the QwtGraph specifically)
 * Qt Designer doesn't allow me to show the control as I want it displayed (not WYSIWYG).
 * I found it very distracting to add these in Qt Designer. By doing it here,
 * I control exactly the options I want.
 */
void CameraReview::create_chart_controls()
{
    int             index;
    QSizePolicy     size_policy(QSizePolicy::Expanding, QSizePolicy::Preferred);

    // Initialize our variables.
    graph_enable = false;
    raw_data_ptr = nullptr;
    raw_data_len = 0;
    x_axis = nullptr;
    y_ao_axis = nullptr;
    y_ai_axis = nullptr;
    y_so_axis = nullptr;
    y_si_axis = nullptr;
    y_to_axis = nullptr;
    y_ti_axis = nullptr;

    // For now, just use default names.
    signal_names << tr("A-O") << tr("A-I") << tr("S-O") << tr("S-I") << tr("T-O") << tr("T-I");

    // Initialize our spacers.
    left_spacer = nullptr;
    right_spacer = nullptr;

    // Need to create our controls.
    pb_add_chart = new QPushButton(this);
    pb_add_chart->setObjectName(QString::fromUtf8("pb_add_chart"));
    pb_add_chart->setText(tr("Add Chart"));
    pb_add_chart->hide();
    ui->leftside_layout->addWidget(pb_add_chart);
    connect(pb_add_chart, &QPushButton::clicked, this, &CameraReview::add_chart);

    // Initialize our size policy.
    size_policy.setHorizontalStretch(0);
    size_policy.setVerticalStretch(0);

    for (index = 0; index < FM_EVENT_FILTERS::INDEX_COUNT - 1; ++index)
    {
        // Create our new spin control. Give it the strings, and set it's
        // alignment properly. Finally, append the spin control into our
        // list and add it to the proper layout.
        fiStringSpinBox *p_spin = new fiStringSpinBox(this);
        p_spin->setStrings(signal_names);
        p_spin->setAlignment(Qt::AlignCenter);
        p_spin->hide();
        spin_signal.append(p_spin);
        ui->midleft_layout->addWidget(p_spin);

        // Connect our spin control to the spin signal mapper, and identify
        // the signal withour index into the list control.
        connect(p_spin, &fiStringSpinBox::value_changed, [this, p_spin]() { spin_signal_change(p_spin); });

        // Create our new plot curve control.
        QwtPlotCurve *p_curve = new QwtPlotCurve();
        p_curve->setStyle(QwtPlotCurve::Lines);
        p_curve->setRenderHint(QwtPlotItem::RenderAntialiased);

        // Create our plot widget and attach our curve to it.
        QwtPlot *p_plot = new QwtPlot(this);
        p_plot->enableAxis(QwtPlot::yLeft, false);
        p_plot->enableAxis(QwtPlot::xBottom, false);
        p_plot->setAxisAutoScale(QwtPlot::xBottom, false);
        p_plot->setLineWidth(1);
        p_plot->setFocusPolicy(Qt::NoFocus);
        p_plot->setSizePolicy(size_policy);
        p_plot->plotLayout()->setCanvasMargin(0, -1);
        p_plot->setFrameStyle(QFrame::Plain | QFrame::NoFrame);
        p_plot->setAutoReplot(true);
        p_plot->hide();

        // Attach the curve control to the plot widget.
        p_curve->attach(p_plot);

        // Put the plot widget into our list and add it to the layout control.
        plot_curve.append(p_plot);
        ui->mid_layout->addWidget(p_plot);

        // Create our new button and add it to all the usual controls.
        QPushButton *p_button = new QPushButton(this);
        p_button->setText(tr("Remove"));
        p_button->hide();
        pb_remove_chart.append(p_button);
        ui->midright_layout->addWidget(p_button);

        // Connect the button to it's mapper, etc.
        connect(p_button, &QPushButton::clicked, [this, p_button]() { remove_chart(p_button);});
    }
}

/*!
 * @brief Construct the controls used for zooming around the frame bar.
 */
void CameraReview::create_zoom_controls()
{
    // Create our "zoom" controls.
    pb_zoom_in = new QPushButton(this);
    pb_zoom_in->setText(tr("Zoom In"));
    pb_zoom_in->hide();
    ui->midleft_layout->addWidget(pb_zoom_in);

    pb_zoom_out = new QPushButton(this);
    pb_zoom_out->setText(tr("Zoom Out"));
    pb_zoom_out->hide();
    ui->midright_layout->addWidget(pb_zoom_out);

    zoom_scrollbar = new fiSlider(this);
    zoom_scrollbar->setOrientation(Qt::Horizontal);
    zoom_scrollbar->setSingleStep(1);
    zoom_scrollbar->hide();
    ui->mid_layout->addWidget(zoom_scrollbar);

    // Setup our "zoom" connections.
    connect(pb_zoom_in, &QPushButton::clicked, this, &CameraReview::zoom_in);
    connect(pb_zoom_out, &QPushButton::clicked, this, &CameraReview::zoom_out);

    zoom_scrollbar->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    connect(zoom_scrollbar, &fiSlider::valueSlowChanged, ui->spin_current_frame, &QSpinBox::setValue);
    connect(zoom_scrollbar, &fiSlider::valueSlowChanged, ui->sld_current_frame, &fiSlider::setValue);
    connect(zoom_scrollbar, &fiSlider::actionTriggered, this, &CameraReview::zoom_action);
    connect(zoom_scrollbar, &fiSlider::sliderPressed, this, &CameraReview::cur_frame_pressed);
    connect(zoom_scrollbar, &fiSlider::released, this, &CameraReview::cur_frame_released);
    connect(zoom_scrollbar, &fiSlider::released, this, &CameraReview::current_frame_changed);

    connect(ui->sld_current_frame, &fiSlider::valueSlowChanged, zoom_scrollbar, &fiSlider::setValue);
}

/*!
 * @brief Enable or disable the signal I/O graphing display.
 *
 * @param do_enable     True to enable (show) the graphing display, false to hide it.
 */
void CameraReview::enable_graph(bool do_enable)
{
    // Remember the setting.
    graph_enable = do_enable;

    // Are we enabling or disabling the graph(s)?
    if (do_enable)
    {
        // Show the first level of controls.
        pb_add_chart->show();
        if (is_series3())
        {
            spin_signal.at(FM_EVENT_FILTERS::AO_INDEX)->show();
            plot_curve.at(FM_EVENT_FILTERS::AO_INDEX)->show();
            pb_remove_chart.at(FM_EVENT_FILTERS::AO_INDEX)->show();
        }
        else
        {
            spin_signal.at(FM_EVENT_FILTERS::SI_INDEX)->show();
            plot_curve.at(FM_EVENT_FILTERS::SI_INDEX)->show();
            pb_remove_chart.at(FM_EVENT_FILTERS::SI_INDEX)->show();
        }

        // Add our spacer and put it in the appropraite layout control.
        if (right_spacer == nullptr)
        {
            right_spacer = new QSpacerItem(20, 20, QSizePolicy::Minimum, QSizePolicy::Expanding);
            ui->rightside_layout->addItem(right_spacer);
        }

        // Make sure we need a spacer.
        if (left_spacer == nullptr)
        {
            left_spacer = new QSpacerItem(20, 20, QSizePolicy::Minimum, QSizePolicy::Expanding);
            ui->leftside_layout->insertSpacerItem(ui->leftside_layout->count() - 1, left_spacer);
        }
    }
    else
    {
        int index;

        // Make sure all the controls are hidden.
        pb_add_chart->hide();
        pb_add_chart->setEnabled(true);
        for (index = 0; index < FM_EVENT_FILTERS::INDEX_COUNT - 1; ++index)
        {
            spin_signal.at(index)->hide();
            plot_curve.at(index)->hide();
            pb_remove_chart.at(index)->hide();
        }

        // Only remove the spacers if we don't have the zoom window up.
        if (!zoom_scrollbar->isVisible())
        {
            // Remove our spacers.
            ui->rightside_layout->removeItem(right_spacer);
            delete right_spacer;
            right_spacer = nullptr;
            ui->leftside_layout->removeItem(left_spacer);
            delete left_spacer;
            left_spacer = nullptr;
        }
    }
}

/*!
 * @brief Place the CameraReview object into "player mode". That means there
 *        is no camera attached because we're playing a saved video.
 */
void CameraReview::enable_player_mode()
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Disconnect from the communications object.
    disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &CameraReview::read_complete);
    disconnect(comm_ptr, &FI_comm::unsol_msg_received, this, &CameraReview::handle_unsol_msg);
    disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraReview::write_complete);

    // Remember we're in player mode.
    player_mode = true;
}

/*!
 * @brief Return the absolute frame number for the clip end frame.
 */
uint32_t CameraReview::get_clip_end()
{
    int         rel_frame;
    uint32_t    trig_frame;
    uint32_t    clip_frame;

    // Get the relative frame number from the UI control.
    rel_frame = ui->clip_slider->upperValue();

    // Get the absolute trigger frame so we can calculate the absolute
    // frame value.
    if (player_mode)
        trig_frame = pb_trigger;
    else
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
    clip_frame = (uint32_t)rel_frame + trig_frame;

    // Return the absolute frame number.
    return clip_frame;
}

/*!
 * @brief Return the absolute frame number for the clip start frame.
 */
uint32_t CameraReview::get_clip_start()
{
    int         rel_frame;
    uint32_t    trig_frame;
    uint32_t    clip_frame;

    // Get the relative frame number from the UI control.
    rel_frame = ui->clip_slider->lowerValue();

    // Get the absolute trigger frame so we can calculate the absolute
    // frame value.
    if (player_mode)
        trig_frame = pb_trigger;
    else
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
    clip_frame = (uint32_t)rel_frame + trig_frame;

    // Return the absolute frame number.
    return clip_frame;
}

/*!
 * @brief Return the absolute frame number for the current frame.
 */
uint32_t CameraReview::get_current_frame_num()
{
    int         rel_frame;
    uint32_t    cur_frame;

    // Get the relative frame number from the UI control.
    rel_frame = ui->spin_current_frame->value();

    // Calculate our absolute current frame number.
    cur_frame = (uint32_t)rel_frame + get_trigger_frame_num();

    // Return the absolute frame number.
    return cur_frame;
}

/*!
 * @brief Return the relative frame information for active review frame.
 *
 * @param [OUT] rel_frame_num   - the current relative frame number in review.
 * @param [OUT] rel_clip_start  - the current clip start for frame in review.
 * @param [OUT] rel_clip_end    - the current clip end for frame in review.
 */
void CameraReview::get_current_rel_frame_info(int& rel_frame_num, int& rel_clip_start, int& rel_clip_end)
{
    rel_frame_num = ui->spin_current_frame->value();
    rel_clip_start = ui->spin_clip_start->value();
    rel_clip_end = ui->spin_clip_end->value();
}

/*!
 * @brief Return the user selected playback rate.
 */
int CameraReview::get_playback_rate()
{
    return ui->spin_frame_rate->value();
}

/*!
 * @brief Return the absolute frame number for the trigger frame.
 */
uint32_t CameraReview::get_trigger_frame_num()
{
    uint32_t    trig_frame;

    // Get the absolute trigger frame so we can calculate the absolute
    // current frame value.
    if (player_mode)
        trig_frame = pb_trigger;
    else
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

    // Return the absolute frame number.
    return trig_frame;
}

bool CameraReview::is_playing()
{
    uint32_t    pb_status;

    // Read up the playback control register.
    pb_status = read_camera_cache(GIGE_REG_PB_STATUS);

    //QLOG_TRACE() << __PRETTY_FUNCTION__ << "pb_status:" << pb_status;

    // Make sure we're still playing.
    return ((pb_status & (GIGE_PBS_BUSY | GIGE_PBS_SINGLEMODE)) != 0);
}


/*!
 * @brief Determines what type of video we are reviewing whether it is coming from a camera
 * or a disk-based video playback engine.
 *
 * @return True if player is connected to a camera, false if not and in that case it is assumed
 * that the player is using the playback engine.
 */
bool CameraReview::is_playback_from_camera()
{
    bool camera_connected = false;
    CameraMDI   *camera_ptr;

    // Do we have a camera pointer?
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr != nullptr)
        camera_connected = true;

    // Return the camera_connected setting.
    return camera_connected;
}

/*!
 * @brief Return a true if we're a series3 camera, false otherwise.
 * @return true if we don't know what the parent is, otherwise whatever the parent tells us.
 */
bool CameraReview::is_series3()
{
    PlaybackMDI *playback_ptr;
    CameraMDI   *camera_ptr;

    // Do we have a camera pointer?
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr != nullptr)
        return camera_ptr->is_series3();

    // Do we have a playback pointer?
    playback_ptr = qobject_cast<PlaybackMDI *>(parent());
    if (playback_ptr != nullptr)
        return playback_ptr->isSeries3();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unknown parent object!";
    return true;
}

/*!
 *  @brief  Return our list of markers.
 */
const QList<int>& CameraReview::markers() const
{
    return ui->clip_slider->markers();
}

/*!
 * @brief Utility method to read the camera's register cache. The assumption
 *        being made is that the register value has already been loaded from
 *        the camera.
 *
 * @param reg_addr      The address of the register we're interested in.
 * @param ok            [OPTIONAL] pointer to a bool to indicate if the register is in cache.
 *
 * @return The register value or 0 if an error (*ok == false).
 */
uint32_t CameraReview::read_camera_cache(uint32_t reg_addr, bool *ok)
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Not available in player mode.
    if (player_mode)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CameraReview::PLAYER_MODE_RESTRICTION;

        if (ok != nullptr)
            *ok = false;
        return 0;
    }
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null my_parent pointer!";

        if (ok != nullptr)
            *ok = false;
        return 0;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";

        if (ok != nullptr)
            *ok = false;
        return 0;
    }

    // Return the cached value.
    return comm_ptr->get_reg(my_parent->get_ip_addr(), reg_addr, ok);
}

/*!
 * @brief Setup a block read to pull the panel's register values down
 *        from the camera.
 */
void CameraReview::read_panel_regs()
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Not available in player mode.
    if (player_mode)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CameraReview::PLAYER_MODE_RESTRICTION;
        return;
    }
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null my_parent pointer!";
        return;
    }

    // Are we ready to talk yet?
    if (!my_parent->comms_are_ready())
    {
        // Try again in a little bit.
        QTimer::singleShot(DELAY_UPDATE_TIME, this, &CameraReview::read_panel_regs);
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Set up the camera group register read.
    cam_register_group->addReg(GIGE_REG_WIDTH);
    cam_register_group->addReg(GIGE_REG_HEIGHT);
    cam_register_group->addReg(GIGE_REG_FRAME_RATE);
    cam_register_group->addReg(GIGE_REG_JPEG_QUALITY);
    cam_register_group->addReg(GIGE_REG_PB_FRAME_PERIOD);
    cam_register_group->addReg(GIGE_REG_PART_NUM_FRAMES);
    cam_register_group->addReg(GIGE_REG_PART_TRIG_FRAME);
    cam_register_group->addReg(GIGE_REG_CUR_FRAME_POS);
    cam_register_group->lastReg(GIGE_REG_CSI_RVW_FRAME_NUM);

    // Start a camera group register read.
    my_parent->read_camera_registers(cam_register_group);
}

/*!
 * @brief Setup a block read to get the playback associated registers
 *        from the camera.
 */
void CameraReview::read_playback_regs()
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Not available in player mode.
    if (player_mode)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CameraReview::PLAYER_MODE_RESTRICTION;
        return;
    }
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Set up the camera group register read.
    cam_register_group->addReg(GIGE_REG_PB_CUT_IN_FRAME);
    cam_register_group->addReg(GIGE_REG_PB_CUT_OUT_FRAME);
    cam_register_group->addReg(GIGE_REG_PB_CONTROL);
    cam_register_group->addReg(GIGE_REG_PB_STATUS);
    cam_register_group->addReg(GIGE_REG_STREAM_CONTROL);
    cam_register_group->lastReg(GIGE_REG_PB_FRAME_PERIOD);

    // Start a register read.
    my_parent->read_camera_registers(cam_register_group);
}

/*!
 * @brief Reset the clip UI controls.
 */
void CameraReview::reset_clip()
{
    // Reset our flags and variables.
    last_frame_num = 0;

    // Ignore updates until we are connected to a camera.
    ignore_updates = true;
}

/*!
 * @brief Set the current frame number.
 * @param cur_frame     The new current frame number.
 */
void CameraReview::set_current_frame_num(uint32_t cur_frame)
{
    int         rel_frame;
    uint32_t    trig_frame;

    // Get the absolute trigger frame so we can calculate the relative
    // current frame value.
    if (player_mode)
        trig_frame = pb_trigger;
    else
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
    rel_frame = (int)cur_frame - (int)trig_frame;

    // Set the current frame to the new value.
    ui->spin_current_frame->setValue(rel_frame);
    ui->sld_current_frame->setValue(rel_frame);
    if (zoom_scrollbar->isVisible())
        zoom_scrollbar->setValue(rel_frame);

    // Remember the new setting.
    last_frame_num = rel_frame;

    // Start our blackout timer.
    start_blackout();

    // Update our time display to reflect where the new frame is at.
    update_time_display(ui->sld_current_frame->value());

    // Let folks know that we're changing.
    emit current_frame_changed();
}

/*!
 * @brief Set the graph data to be used when the signal I/O graphing is enabled.
 *
 * All the signals fit in a single byte, so this is a byte of signal information
 * per frame of video.
 *
 * @param data_ptr      Pointer to the signal I/O data array.
 * @param data_len      The number of bytes in the array.
 */
void CameraReview::set_graph_data(uint8_t *data_ptr, int data_len)
{
    int     i;

    // Sanity check.
    if (data_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "data_ptr is NULL!";
        return;
    }

    // Do we already have a data array?
    if (raw_data_ptr != nullptr)
        delete []raw_data_ptr;

    // Save the new pointer and length.
    raw_data_ptr = data_ptr;
    raw_data_len = (uint32_t)data_len;

    // Do we have existing arrays? If so, delete them.
    if (x_axis)
    {
        // We won't have one with out the others.
        delete []x_axis;
        delete []y_ao_axis;
        delete []y_ai_axis;
        delete []y_so_axis;
        delete []y_si_axis;
        delete []y_to_axis;
        delete []y_ti_axis;
    }

    // Allocate memory for our new data sets.
    x_axis = new double[ data_len + 1 ];
    y_ao_axis = new double[ data_len + 1 ];
    y_ai_axis = new double[ data_len + 1 ];
    y_so_axis = new double[ data_len + 1 ];
    y_si_axis = new double[ data_len + 1 ];
    y_to_axis = new double[ data_len + 1 ];
    y_ti_axis = new double[ data_len + 1 ];
    axis_len = data_len;

    // Build up the various axis of information.
    for (i = 0; i < data_len; ++i)
    {
        x_axis[i] = (double)i;

        if (is_series3())
        {
            y_ao_axis[i] = (data_ptr[i] & (1 << FM_EVENT_FILTERS::AO_INDEX)) ? 1.0 : 0.0;
            y_ai_axis[i] = (data_ptr[i] & (1 << FM_EVENT_FILTERS::AI_INDEX)) ? 1.0 : 0.0;
            y_so_axis[i] = (data_ptr[i] & (1 << FM_EVENT_FILTERS::SO_INDEX)) ? 1.0 : 0.0;
            y_si_axis[i] = (data_ptr[i] & (1 << FM_EVENT_FILTERS::SI_INDEX)) ? 1.0 : 0.0;
            y_to_axis[i] = (data_ptr[i] & (1 << FM_EVENT_FILTERS::TO_INDEX)) ? 1.0 : 0.0;
            y_ti_axis[i] = (data_ptr[i] & (1 << FM_EVENT_FILTERS::TI_INDEX)) ? 1.0 : 0.0;
        }
        else
        {
            y_ao_axis[i] = 0.0;
            y_ai_axis[i] = 0.0;
            y_so_axis[i] = 0.0;
            y_si_axis[i] = (data_ptr[i] & (1 << 0)) ? 1.0 : 0.0;    // Channel 0
            y_to_axis[i] = (data_ptr[i] & (1 << 1)) ? 1.0 : 0.0;    // Channel 1
            y_ti_axis[i] = (data_ptr[i] & (1 << 2)) ? 1.0 : 0.0;    // Channel 2
        }
    }

    // Set the last point to be the opposite of the first for the chart scaling.
    x_axis[ data_len ] = (double)data_len;
    y_ao_axis[ data_len ] = (y_ao_axis[0] == 0.0) ? 1.0 : 0.0;
    y_ai_axis[ data_len ] = (y_ai_axis[0] == 0.0) ? 1.0 : 0.0;
    y_so_axis[ data_len ] = (y_so_axis[0] == 0.0) ? 1.0 : 0.0;
    y_si_axis[ data_len ] = (y_si_axis[0] == 0.0) ? 1.0 : 0.0;
    y_to_axis[ data_len ] = (y_to_axis[0] == 0.0) ? 1.0 : 0.0;
    y_ti_axis[ data_len ] = (y_ti_axis[0] == 0.0) ? 1.0 : 0.0;

    // Loop through the charts and display the information.
    for (i = 0; i < plot_curve.size(); ++i)
    {
        fiStringSpinBox *p_spin = spin_signal.at(i);

        // Only update it if it's visible.
        if (p_spin->isVisible())
        {
            // Pretend like we've changed the spinner. We haven't, but
            // this will load the graph.
            spin_signal_change(p_spin);
        }
    }
}

/*!
 * @brief   Set the list of markers. If we are getting a non-empty
 *          list, then change our jump button icons to indicate we're
 *          in "event tracking" mode. Otherwise, adjust the icons to
 *          reflect a "normal" mode.
 *
 * @param   markers     A list of markers, may be empty if we're exiting "event tracking" mode.
 */
void CameraReview::set_markers(const QList<int>& markers)
{
    QIcon   prev_icon;
    QIcon   next_icon;

    // Setup the markers.
    ui->clip_slider->setEventMarkers(markers);

    // Do we have a zoom control?
    if (zoom_scrollbar != nullptr)
        zoom_scrollbar->setEventMarkers(markers);

    // If we have markers, then adjust our button icons.
    if (markers.size() > 0)
    {
        // Get the "event" icons for the buttons.
        prev_icon.addFile(QString::fromUtf8(":/images/jump_to_prev_event.png"));
        next_icon.addFile(QString::fromUtf8(":/images/jump_to_next_event.png"));
    }
    else
    {
        prev_icon.addFile(QString::fromUtf8(":/images/jump_to_start.png"));
        next_icon.addFile(QString::fromUtf8(":/images/jump_to_end.png"));
    }

    ui->tb_jump_first->setIcon(prev_icon);
    ui->tb_jump_first->setEnabled(true);
    ui->tb_jump_last->setIcon(next_icon);
    ui->tb_jump_last->setEnabled(true);
}

/*!
 * @brief Modifies the playback control clip setting and updates the review controls.
 */
void CameraReview::set_pb_sync_clip(uint32_t clip_start, uint32_t clip_end)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Make sure we have a parent camera.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    // Make sure they are set properly.
    if (clip_end < clip_start)
        clip_end = clip_start;

    // We want to update the camera registers for the synchronized clipping range.
    my_parent->write_camera_register(GIGE_REG_PB_CUT_IN_FRAME, clip_start);
    my_parent->write_camera_register(GIGE_REG_PB_CUT_OUT_FRAME, clip_end);

    // Refresh these in our cache if a series 3 camera as it doesn't get unsol notification for these.
    if (my_parent->is_series3())
    {
        QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
        if (cam_register_group.isNull())
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
            return;
        }

        // Set up the camera group register read.
        cam_register_group->addReg(GIGE_REG_PB_CUT_IN_FRAME);
        cam_register_group->lastReg(GIGE_REG_PB_CUT_OUT_FRAME);
        my_parent->read_camera_registers(cam_register_group);
    }

    // Put some order into this because of how the controls will work.
    // If out of range they self-clip to keep inside of it.
    if (clip_end > pb_clip_start)
    {
        set_pb_clip_end(clip_end);
        set_pb_clip_start(clip_start);
    }
    else
    {
        set_pb_clip_start(clip_start);
        set_pb_clip_end(clip_end);
    }
}

/*!
 * @brief Set the playback end frame number.
 * @param clip_end  The absolute ending frame number for the playback video.
 */
void CameraReview::set_pb_clip_end(uint32_t clip_end)
{
    pb_clip_end = clip_end;
    update_clip_end();
}

/*!
 * @brief Set the playback start frame number.
 * @param clip_start    The absolute starting frame number for the playback video.
 */
void CameraReview::set_pb_clip_start(uint32_t clip_start)
{
    pb_clip_start = clip_start;
    update_clip_start();
}

/*!
 * @brief Set the number of frames for playback.
 * @param frame_count       The number of frames.
 */
void CameraReview::set_pb_frame_count(uint32_t frame_count)
{
    // Save the frame count.
    pb_frame_count = frame_count;

    // Sanity check.
    if (((pb_clip_end - pb_clip_start) + 1) != pb_frame_count)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "frame count:" << pb_frame_count << "does not equal end:" << pb_clip_end << "minus start:" << pb_clip_start << "plus 1.";
    }

    // Update the controls.
    update_frame_info();
    update_clip_start();
    update_clip_end();
}

/*!
 * @brief Update the playback frame rate.
 * @param frame_rate        The new rate.
 */
void CameraReview::set_pb_frame_rate(uint32_t frame_rate)
{
    pb_frame_rate = frame_rate;
    update_roi_rate();
}

/*!
 * @brief Start a blackout period to ignore recursive updates from the camera.
 */
void CameraReview::start_blackout()
{
    // Frame updates are not allowed.
    allow_frame_updates = false;

    // If the time is already going, kill it first.
    if (blackout_id != 0)
        killTimer(blackout_id);

    // Startup our blackout timer for the asynchronous updates.
    blackout_id = startTimer(BLACKOUT_PERIOD);
}

/*!
 * @brief Translate the UI elements for this object.
 */
void CameraReview::translate_ui()
{
    // Have the UI object perform it's translations.
    ui->retranslateUi(this);

    // Translate the controls we've created.
    pb_add_chart->setText(tr("Add Chart"));
    pb_zoom_in->setText(tr("Zoom In"));
    pb_zoom_out->setText(tr("Zoom Out"));

    for (int index = 0; index < FM_EVENT_FILTERS::INDEX_COUNT - 1; ++index)
    {
        pb_remove_chart.at(index)->setText(tr("Remove"));
    }

    // Update our display.
    update_time_display(ui->sld_current_frame->value());
    update_frame_info();
}

/*!
 * @brief Enable/disable various playback controls.
 * @param do_enable     True to enable, false to disable.
 */
void CameraReview::update_buttons(bool do_enable)
{
    ui->tb_jump_first->setEnabled(do_enable);
    ui->tb_jump_last->setEnabled(do_enable);
    ui->tb_play_forward->setEnabled(do_enable);
    ui->tb_play_reverse->setEnabled(do_enable);
    ui->tb_step_forward->setEnabled(do_enable);
    ui->tb_step_reverse->setEnabled(do_enable);
    ui->tb_time_jump->setEnabled(do_enable);
    ui->tb_trigger->setEnabled(do_enable);
    ui->tb_trash->setEnabled(do_enable);
    ui->pb_clip_end->setEnabled(do_enable);
    ui->pb_clip_start->setEnabled(do_enable);
    ui->spin_clip_end->setEnabled(do_enable);
    ui->spin_clip_start->setEnabled(do_enable);
    ui->spin_current_frame->setEnabled(do_enable);
    ui->sld_current_frame->setEnabled(do_enable);
    ui->clip_slider->setEnabled(do_enable);
}

/*!
 * @brief Update the playback controls based on the camera register settings.
 */
void CameraReview::update_clip_end()
{
    uint32_t    clip_end;
    uint32_t    trig_frame;

    // Read up our registers.
    if (player_mode)
    {
        // Load our playback settings.
        trig_frame = pb_trigger;
        clip_end = pb_clip_end;

        // Adjust the upper end of the range.
        ui->clip_slider->setMaximum((int)(clip_end - trig_frame));
    }
    else
    {
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
        clip_end = read_camera_cache(GIGE_REG_PB_CUT_OUT_FRAME);
    }

    // Update the clip UI.
    ui->spin_clip_end->blockSignals(true);
    ui->spin_clip_end->setValue((int)(clip_end - trig_frame));
    ui->spin_clip_end->blockSignals(false);
    ui->clip_slider->setUpperValue((int)(clip_end - trig_frame));
}

/*!
 * @brief Update the playback controls based on the camera register settings.
 */
void CameraReview::update_clip_start()
{
    uint32_t    clip_start;
    uint32_t    trig_frame;

    // Read up our registers.
    if (player_mode)
    {
        // Load our playback settings.
        trig_frame = pb_trigger;
        clip_start = pb_clip_start;

        // Adjust the bottom end of the range.
        ui->clip_slider->setMinimum((int)(clip_start - trig_frame));
    }
    else
    {
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
        clip_start = read_camera_cache(GIGE_REG_PB_CUT_IN_FRAME);
    }

    // Update the clip UI.
    ui->spin_clip_start->blockSignals(true);
    ui->spin_clip_start->setValue((int)(clip_start - trig_frame));
    ui->spin_clip_start->blockSignals(false);
    ui->clip_slider->setLowerValue((int)(clip_start - trig_frame));
}

/*!
 * @brief Update the displayed frame information.
 */
void CameraReview::update_frame_info()
{
    int         rel_frame;
    int         rel_min;
    int         rel_max;
    uint32_t    num_frames;

    // If the user is scrubbing the current frame slider, then we need to ignore this.
    if (!allow_frame_updates)
    {
        return;
    }

    // Read up our registers.
    if (player_mode)
    {
        // Copy our number of frames.
        num_frames = pb_frame_count;

        // Sanity check.
        if (last_frame_num > (int)pb_frame_count)
            last_frame_num = 0;

        // Since trigger changed, all the other displayed fields need to change as well
        // since they are relative to the trigger.
        rel_frame = last_frame_num;
        rel_min = (int)pb_clip_start - (int)pb_trigger;
        rel_max = (int)pb_clip_end - (int)pb_trigger;
    }
    else
    {
        uint32_t    trig_frame;
        uint32_t    part_state;

        // Get our partition state.
        part_state = read_camera_cache(GIGE_REG_PART_STATE);

        // Only interested if the partition state is done.
        if (part_state != GIGE_PART_STATE_COMPLETE)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "part state not complete.";
            return;
        }

        // Get the camera values.
        num_frames = read_camera_cache(GIGE_REG_PART_NUM_FRAMES);
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

        // Sanity check.
        if (last_frame_num > (int)((num_frames - trig_frame) - 1))
            last_frame_num = (int)((num_frames - trig_frame) - 1);
        else if (last_frame_num < -((int)trig_frame))
            last_frame_num = -((int)trig_frame);

        // Since trigger changed, all the other displayed fields need to change as well
        // since they are relative to the trigger.
        rel_frame = last_frame_num;
        rel_min = 0 - (int)trig_frame;
        rel_max = (int)(num_frames - trig_frame) - 1;
    }

    ui->spin_current_frame->setMinimum(rel_min);
    ui->sld_current_frame->setMinimum(rel_min);
    ui->spin_current_frame->setMaximum(rel_max);
    ui->sld_current_frame->setMaximum(rel_max);

    ui->spin_clip_start->blockSignals(true);
    ui->spin_clip_end->blockSignals(true);
    ui->spin_clip_start->setMinimum(rel_min);
    ui->clip_slider->setMinimum(rel_min);
    ui->spin_clip_end->setMinimum(rel_min + 1);
    ui->spin_clip_start->setMaximum(rel_max - 1);
    ui->clip_slider->setMaximum(rel_max);
    ui->spin_clip_end->setMaximum(rel_max);
    ui->spin_clip_start->blockSignals(false);
    ui->spin_clip_end->blockSignals(false);

    // Initialize the span to cover the entire recording.
    ui->clip_slider->setSpan(rel_min, rel_max);

    // Is the new frame different than the old frame?
    if (rel_frame != ui->spin_current_frame->value())
    {
        ui->spin_current_frame->setValue(rel_frame);
        ui->sld_current_frame->setValue(rel_frame);
        if (zoom_scrollbar->isVisible())
            zoom_scrollbar->setValue(rel_frame);

        emit current_frame_changed();
    }
    ui->lbl_num_frames->setText(QString::number(num_frames));

    // Now calculate the frame time relative to the trigger frame.
    update_time_display(rel_frame);

    // Update the signal information.
    update_graph();
}

void CameraReview::update_graph()
{
    int index;

    // The default layout is for series3 cameras, so we're done.
    if (is_series3())
        return;

    // Reset the signal names list.
    build_labels();

    for (index = FM_EVENT_FILTERS::CH0_INDEX; index < FM_EVENT_FILTERS::INDEX_COUNT - 1; ++index)
    {
        fiStringSpinBox *p_spin;

        // Get a pointer to our spin control object. Give it the strings associated
        // with the external I/O signals.
        p_spin = spin_signal.at(index);
        p_spin->setStrings(signal_names);
    }
}

void CameraReview::update_part_info(uint32_t part_state)
{
    // Are we empty? If so, reset the review controls.
    if (part_state == GIGE_PART_STATE_EMPTY)
    {
        // Disable our controls until we get things initialized.
        setEnabled(false);

        // Reset our last frame value.
        last_frame_num = 0;

        // Ignore updates. This will get reset once we get a complete partition
        // and have finished reading all the registers.
        ignore_updates = true;
    }

    // Only interested if the partition state is done.
    else if (part_state == GIGE_PART_STATE_COMPLETE)
    {
        // Are we visible?
        if (isVisible() && (last_part_state != part_state))
        {
            // Read the registers we'll need.
            read_panel_regs();
        }
    }

    // Remember the state.
    last_part_state = part_state;
}

/*!
 * @brief We've received an unsolicited update to the playback control register.
 *        Determine if we are starting or stopping a stream, or if the user is
 *        manipulating the stream on a different UI.
 */
void CameraReview::update_playback_control(uint32_t pb_control)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    //QLOG_TRACE() << __PRETTY_FUNCTION__ << "pb_control:" << Qt::showbase << Qt::hex << pb_control;

    // Sanity check.
    if (my_parent == nullptr)
        return;

    // Check to see if we are still playing.
    if (!is_series3())
    {
        // Series3 cameras don't need this.
        if (is_playing() && ((pb_control & (GIGE_PBC_ENABLE | GIGE_PBC_SINGLE_START)) == 0))
        {
            // Make sure we stop the camera from playback.
            playback_stop();
        }
    }

    // Update our registers.
    read_playback_regs();
}

/*!
 * @brief Update the frame information during playback.
 */
void CameraReview::update_playback_frame()
{
    int         rel_frame;
    uint32_t    trig_frame;
    uint32_t    pb_frame;
    int         cur_frame;

    // Not available in player mode.
    if (player_mode)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CameraReview::PLAYER_MODE_RESTRICTION;
        return;
    }

    // Make sure we're in playback mode.
    if (play_update_id == 0)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "play_update_id is 0";
        return;
    }

    // Read up our registers.
    pb_frame = read_camera_cache(GIGE_REG_PB_LAST_FRAME);
    trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

    // Get the playback frame relative to the trigger frame.
    rel_frame = (int)(pb_frame - trig_frame);
    cur_frame = ui->sld_current_frame->sliderPosition();

    // Are we repeating the same frame?
    if (cur_frame == rel_frame)
    {
        // Maybe we should update our idea of what the registers contain?
        read_playback_regs();
    }

    // If the user is scrubbing the current frame slider, then we need to ignore this.
    if (ui->sld_current_frame->throttleIsActive() == false)
    {
        // Go ahead and update the UI.
        ui->spin_current_frame->setValue(rel_frame);
        ui->sld_current_frame->setValue(rel_frame);
        if (zoom_scrollbar->isVisible())
            zoom_scrollbar->setValue(rel_frame);
    }

    // Now calculate the frame time relative to the trigger frame.
    update_time_display(rel_frame);

    // Let folks know it's time to update.
    emit play_update();
}

/*!
 * @brief Read the playback status register and figure out if we are
 *        playing a stream or not.
 *
 * @return True if the stream is playing, false if we're stopped.
 */
bool CameraReview::update_playback_status()
{
    bool playing = is_playing();
    bool stepping;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check
    if (my_parent == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No parent?!";
        return false;
    }
    if (my_parent->get_streamer() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No streamer?!";
        return false;
    }

    // We might be playing and we might be stepping...big difference in how we
    // handle these.
    // For series3 cameras, since the FPGA is in charge of playback, we only need to check
    // to see if we're stepping.
    // For seriesNC cameras, we have to also see if the streamer is paused before deciding if
    // we're stepping.
    if (is_series3())
        stepping = my_parent->get_streamer()->is_stepping();
    else
        stepping = (my_parent->get_streamer()->is_paused() || my_parent->get_streamer()->is_stepping());
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "playing:" << playing << "stepping:" << stepping << "is_paused:" << my_parent->get_streamer()->is_paused() << "is_stepping:" << my_parent->get_streamer()->is_stepping();

    // Are we playing?
    if (playing && !stepping)
    {
        // Update our buttons.
        update_buttons(false);

        // Ticket #328 UPDATED APPROACH:  I am revisiting this logic for Ticket #328.
        // If you get here in this scope then the review video is playing.
        //
        // Therefore should we arrive here and there is a player timer already operating
        // perhaps we just let it run.  But in doing so it wasn't consistent and behaved
        // poorly in fact the timer doesn't seem to continue firing!  Did someone kill it
        // underneath us?
        //
        // Investigation on the play_update_id usage seems to yield a proper handling of the
        // timer so it's a little puzzling how we end up with one already running but the
        // assumption in the ticket was thinking a double-click on a record playback control
        // got us into this scenario.
        //
        // The right approach is keep the original fix in play.  It seems to keep the playback
        // operational and the debug statements can be turned back on if someone wants to continue
        // chasing down what's happening.  Perhaps it's a threaded timer glitch with the register
        // read I don't know.  But as long as someone isn't double clicking on the controls we
        // really shouldn't get into this scenario.
        if (play_update_id != 0)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "UNEXPECTED: play_update_id:" << play_update_id << "timer was already set!" << "is_playing()=" << is_playing() << "is_paused:" << my_parent->get_streamer()->is_paused() << "is_stepping:" << my_parent->get_streamer()->is_stepping();
            killTimer(play_update_id);
            play_update_id = 0;
        }
        if (play_update_id == 0)
        {
            // Get our timer going so we can track the playback progress.
            play_update_id = startTimer(PLAY_UPDATE_TIME);
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Started New play_update_id:" << play_update_id;
        }
    }
    else
    {
        // We are not "playing" we are, at best, single stepping.
        playing = false;

        // Shutdown our update timer.
        if (play_update_id != 0)
        {
            killTimer(play_update_id);
            play_update_id = 0;
        }

        // Update our buttons.
        update_buttons(true);
    }

    // Return our playing status.
    return playing;
}


/*!
 * @brief Update the IO graph being displayed based on the selection of the
 *        spin control passed in.
 * @param p_active_spin     The pointer to the spin control being modified.
 */
void CameraReview::update_plot(fiStringSpinBox *p_active_spin)
{
    int             min_frame;
    int             trig_frame;
    int             frame_count;
    int             index;
    QwtPlot         *p_plot;
    QwtPlotCurve    *p_curve;

    int             decimate_len;
    bool            allocated_decs = false;
    double          *dec_x = nullptr;
    double          *dec_y_ao = nullptr;
    double          *dec_y_ai = nullptr;
    double          *dec_y_so = nullptr;
    double          *dec_y_si = nullptr;
    double          *dec_y_to = nullptr;
    double          *dec_y_ti = nullptr;

    // Update the I/O graphs, if they're enabled.
    if (!graph_is_enabled())
        return;

    // Assume we won't have to decimate.
    decimate_len = 0;

    // Get our trigger frame.
    if (player_mode)
        trig_frame = (int)pb_trigger;
    else
        trig_frame = (int)read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

    // Do we have the zoom window enabled.
    if (zoom_scrollbar->isVisible())
    {
        // Calculate our absolute minimum frame value.
        min_frame = trig_frame + zoom_scrollbar->minimum();
        frame_count = zoom_scrollbar->maximum() - zoom_scrollbar->minimum();

        // If we're in "player" mode, we need to adjust the minimum frame number
        // by the clip start value.
        if (player_mode)
            min_frame -= pb_clip_start;
    }
    else
    {
        min_frame = 0;
        frame_count = axis_len - 1;
    }

    // Is the data space too large?
    if (axis_len > DECIMATE_THRESHOLD)
    {
        int skip_count;
        int axis_index;

        // We have to use a smaller data space. Use the frame
        // count so we can populate the data with as much resolution
        // as possible.
        if (frame_count < MAX_SAMPLE_LEN)
            skip_count = 1;     // No skip.
        else
            skip_count = frame_count / MAX_SAMPLE_LEN;

        // Initially we'll start our array at this size.
        decimate_len = MAX_SAMPLE_LEN;

        // Most likely, our skip count does not evenly match our frame count.
        // Overallocate our buffer so we can dribble in the last few elements.
        if (frame_count > (skip_count * MAX_SAMPLE_LEN))
            decimate_len += (frame_count - (skip_count * MAX_SAMPLE_LEN));

        // Allocate our buffers.
        dec_x = new double[ decimate_len ];
        dec_y_ao = new double[ decimate_len ];
        dec_y_ai = new double[ decimate_len ];
        dec_y_so = new double[ decimate_len ];
        dec_y_si = new double[ decimate_len ];
        dec_y_to = new double[ decimate_len ];
        dec_y_ti = new double[ decimate_len ];
        allocated_decs = true;

        // Decimate the data.
        for (index = 0, axis_index = min_frame; index < MAX_SAMPLE_LEN; ++index, axis_index += skip_count)
        {
            // Make sure our new_index isn't too big.
            if (axis_index >= axis_len)
            {
                // Adjust our length.
                decimate_len = index;
                break;
            }

            // Copy the data over.
            dec_x[ index ] = axis_index;
            dec_y_ao[ index ] = y_ao_axis[ axis_index ];
            dec_y_ai[ index ] = y_ai_axis[ axis_index ];
            dec_y_so[ index ] = y_so_axis[ axis_index ];
            dec_y_si[ index ] = y_si_axis[ axis_index ];
            dec_y_to[ index ] = y_to_axis[ axis_index ];
            dec_y_ti[ index ] = y_ti_axis[ axis_index ];
        }

        // Dribble in the last few.
        while ((index < decimate_len) && (axis_index < axis_len))
        {
            // Copy the data over.
            dec_x[ index ] = axis_index;
            dec_y_ao[ index ] = y_ao_axis[ axis_index ];
            dec_y_ai[ index ] = y_ai_axis[ axis_index ];
            dec_y_so[ index ] = y_so_axis[ axis_index ];
            dec_y_si[ index ] = y_si_axis[ axis_index ];
            dec_y_to[ index ] = y_to_axis[ axis_index ];
            dec_y_ti[ index ] = y_ti_axis[ axis_index ];

            // Increment our indicies.
            ++axis_index;
            ++index;
        }

        // Make sure to check to see if we need to adjust our length.
        if (index != decimate_len)
            decimate_len = index;
    }

    // Update the plots with the data.
    for (index = 0; index < plot_curve.size(); ++index)
    {
        // Get our spin control.
        fiStringSpinBox *p_spin = spin_signal.at(index);

        // Was there a spin control we're particularly interested in?
        if ((p_active_spin != nullptr) && (p_spin != p_active_spin))
        {
            // This isn't the one, so keep looking.
            continue;
        }

        // Also, we don't need to do any updates if the control isn't visible.
        if (!p_spin->isVisible())
            continue;

        // Get our plot and curve.
        p_plot = plot_curve.at(index);
        p_curve = static_cast<QwtPlotCurve *>(p_plot->itemList().at(0));
        if (p_curve == nullptr)
        {
            qDebug() << __FUNCTION__ << "p_curve is NULL!";
            continue;
        }

        // Set the data appropriate for the selected signal. If we are decimating,
        // then we need to select the decimated array. For the decimated array we
        // use the setSamples() method, which uses a "deep copy" to move the samples.
        // Otherwise we use setRawSamples() which simply references our standard arrays.
        if (is_series3())
        {
            if (p_spin->text() == signal_names.at(FM_EVENT_FILTERS::AO_INDEX))
            {
                if (dec_x && dec_y_ao && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_ao, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_ao_axis, axis_len + 1);
            }
            else if (p_spin->text() == signal_names.at(FM_EVENT_FILTERS::AI_INDEX))
            {
                if (dec_x && dec_y_ai && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_ai, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_ai_axis, axis_len + 1);
            }
            else if (p_spin->text() == signal_names.at(FM_EVENT_FILTERS::SO_INDEX))
            {
                if (dec_x && dec_y_so && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_so, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_so_axis, axis_len + 1);
            }
            else if (p_spin->text() == signal_names.at(FM_EVENT_FILTERS::SI_INDEX))
            {
                if (dec_x && dec_y_si && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_si, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_si_axis, axis_len + 1);
            }
            else if (p_spin->text() == signal_names.at(FM_EVENT_FILTERS::TO_INDEX))
            {
                if (dec_x && dec_y_to && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_to, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_to_axis, axis_len + 1);
            }
            else if (p_spin->text() == signal_names.at(FM_EVENT_FILTERS::TI_INDEX))
            {
                if (dec_x && dec_y_ti && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_ti, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_ti_axis, axis_len + 1);
            }
        }
        else
        {
            // Channel 0?
            if (p_spin->text() == signal_names.at(0))
            {
                if (dec_x && dec_y_si && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_si, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_si_axis, axis_len + 1);
            }

            // Chanel 1?
            else if (p_spin->text() == signal_names.at(1))
            {
                if (dec_x && dec_y_to && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_to, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_to_axis, axis_len + 1);
            }

            // Channel 2?
            else if (p_spin->text() == signal_names.at(2))
            {
                if (dec_x && dec_y_ti && (decimate_len > 0))
                    p_curve->setSamples(dec_x, dec_y_ti, decimate_len);
                else
                    p_curve->setRawSamples(x_axis, y_ti_axis, axis_len + 1);
            }
        }

        // Update the scale.
        p_plot->setAxisScale(QwtPlot::xBottom, (double)(min_frame), (double)(min_frame + frame_count));
        p_plot->updateAxes();
        p_plot->update();

        // If the spin control was specified, then we can leave. No need to do all of them
        if (p_active_spin != nullptr)
            break;
    }

    // Do we need to release memory?
    if (allocated_decs)
    {
        delete []dec_x;
        delete []dec_y_ao;
        delete []dec_y_ai;
        delete []dec_y_so;
        delete []dec_y_si;
        delete []dec_y_to;
        delete []dec_y_ti;
    }
}

/*!
 * @brief Update the ROI and frame rate displays based on the camera settings.
 */
void CameraReview::update_roi_rate()
{
    uint32_t    frame_rate;

    if (player_mode)
        frame_rate = pb_frame_rate;
    else
        frame_rate = read_camera_cache(GIGE_REG_FRAME_RATE);

    // Update the frame rate controls.
    ui->spin_frame_rate->setMinimum(1);
    ui->spin_frame_rate->setMaximum((int)ClientMath::reg_to_frame_rate(frame_rate));
}

/*!
 * @brief Update the time display based on the frame number passed in.
 * @param frame_no      The frame number we're interested in.
 */
void CameraReview::update_time_display(int frame_no)
{
    uint32_t    frame_rate;
    double      frame_time;
    QString     field_str;

    // We need the frame rate to figure out the relative time of the current frame.
    if (player_mode)
        frame_rate = pb_frame_rate;
    else
        frame_rate = read_camera_cache(GIGE_REG_FRAME_RATE);

    frame_time = (double)frame_no;
    frame_time = frame_time / ClientMath::reg_to_frame_rate(frame_rate);
    field_str = QString::fromUtf8("Time %1 sec").arg(frame_time, 0, 'f', 4);
    ui->lbl_date_time->setText(field_str);
}

/*!
 * @brief Write a register on the camera.
 * @param reg_addr      The register address we want to write.
 * @param reg_value     The register value we want to write.
 */
void CameraReview::write_register(uint32_t reg_addr, uint32_t reg_value)
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Not available in player mode.
    if (player_mode)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "reg_addr:" << Qt::showbase << Qt::hex << reg_addr << "reg_value:" << Qt::showbase << Qt::hex << reg_value << CameraReview::PLAYER_MODE_RESTRICTION;
        return;
    }
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null my_parent pointer!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

#if VERBOSE_SYNC_CLIP_LOG
    if (reg_addr == GIGE_REG_PB_CUT_IN_FRAME)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Writing GIGE_REG_PB_CUT_IN_FRAME:" << reg_value;
    }
    if (reg_addr == GIGE_REG_PB_CUT_OUT_FRAME)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Writing GIGE_REG_PB_CUT_OUT_FRAME:" << reg_value;
    }
#endif

    // Issue the write request.
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), reg_addr, reg_value);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void CameraReview::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    QWidget::changeEvent(event);
}

/*!
 * @brief Handle timer expiration events.
 * @param event     Pointer to the timer which generated the event.
 */
void CameraReview::timerEvent(QTimerEvent *event)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Is this our playback update timer?
    if ((my_parent != nullptr) && (event->timerId() == play_update_id))
    {
        // Read the current playback frame.
        my_parent->start_camera_read_reg(GIGE_REG_PB_LAST_FRAME);
    }
    else if (event->timerId() == blackout_id)
    {
        // Don't need the timer anymore.
        killTimer(blackout_id);
        blackout_id = 0;

        // Check to make sure we're not scrubbing again.
        if (ui->sld_current_frame->throttleIsActive() == false)
        {
            // Allow unsolicited frame updates through now.
            allow_frame_updates = true;
        }
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Handle the event when the user presses the "Add Chart" button.
 *
 * Add another row of signal I/O graph display controls, adjusting the
 * layout as required. This doesn't actually add controls but rather
 * "shows" the appropriate controls. Also, if we've reached our maximum number of
 *  rows, make sure we disable the "Add Chart" button.
 */
void CameraReview::add_chart()
{
    int i;
    int index;
    int hidden_count;
    int start_index;

    if (is_series3())
        start_index = FM_EVENT_FILTERS::AO_INDEX;
    else
        start_index = FM_EVENT_FILTERS::CH0_INDEX;

    // Find the first "hidden" control.
    for (i = start_index, index = -1, hidden_count = 0; i < plot_curve.size(); ++i)
    {
        // Is this one hidden?
        if (!plot_curve.at(i)->isVisible())
        {
            // Count the hidden rows.
            ++hidden_count;

            // Do we need to remember this one?
            if (index < 0)
                index = i;
        }
    }
    if (hidden_count == 0)
    {
        QLOG_DEBUG() << __FUNCTION__ << "No hidden controls found!";
        pb_add_chart->setEnabled(false);
        return;
    }

    spin_signal.at(index)->show();
    plot_curve.at(index)->show();
    pb_remove_chart.at(index)->show();

    // Is this the last one we should add?
    if (hidden_count == 1)
        pb_add_chart->setEnabled(false);

    // Update our data.
    spin_signal_change(spin_signal.at(index));
}

/*!
 * @brief The clip range has changed.
 */
void CameraReview::clip_changed()
{
    int clip_start;
    int clip_end;

    clip_start = ui->clip_slider->lowerPosition();
    clip_end = ui->clip_slider->upperPosition();

    if (clip_start != ui->spin_clip_start->value())
        ui->spin_clip_start->setValue(clip_start);
    if (clip_end != ui->spin_clip_end->value())
        ui->spin_clip_end->setValue(clip_end);
}

/*!
 * @brief The clip ending position has changed.
 * @param new_value     The frame number of the new clip ending.
 */
void CameraReview::clip_end_changing(int new_value)
{
    uint32_t abs_frame;

    // Update the start-clip control's maximum value based on the end-clip's value.
    if (new_value > ui->spin_clip_start->value())
        ui->spin_clip_start->setMaximum(new_value - 1);

    // Determine the player type.  If it is not connected to a camera then there is
    // no register to write.
    if (is_playback_from_camera())
    {
        // Update the camera. By using the get_clip_end() function we get the absolute
        // frame number.
        abs_frame = get_clip_end();
        write_register(GIGE_REG_PB_CUT_OUT_FRAME, abs_frame);
    }
}

/*!
 * @brief The user clicked the clip end button to make the current frame
 *        the new end of the clip.
 */
void CameraReview::clip_end_clicked()
{
    int rel_clip_start;

    // Make the clip start value match the current frame.
    // ORDER IS IMPORTANT! Update the slider first, then the spinner!
    rel_clip_start = ui->spin_current_frame->value();
    ui->clip_slider->setUpperValue(rel_clip_start);
    ui->spin_clip_end->setValue(rel_clip_start);

    // Now emit the signal to let interested parties know that the clip end has changed
    emit clip_end_changed();
}

/*!
 * @brief The clip start position has changed.
 * @param new_value     The frame number of the new clip start.
 */
void CameraReview::clip_start_changing(int new_value)
{
    uint32_t abs_frame;

    // Update the end-clip control's minimum value based on the start-clip's value.
    if (new_value < ui->spin_clip_end->value())
        ui->spin_clip_end->setMinimum(new_value + 1);

    // Determine the player type.  If it is not connected to a camera then there is
    // no register to write.
    if (is_playback_from_camera())
    {
        // Update the camera. By using the get_clip_start() function we get the absolute
        // frame number.
        abs_frame = get_clip_start();
        write_register(GIGE_REG_PB_CUT_IN_FRAME, abs_frame);
    }
}

/*!
 * @brief The user clicked the clip start button to make the current frame
 *        the new start of the clip.
 */
void CameraReview::clip_start_clicked()
{
    int rel_clip_start;

    // Make the clip start value match the current frame.
    // ORDER IS IMPORTANT! Update the slider first, then the spinner!
    rel_clip_start = ui->spin_current_frame->value();
    ui->clip_slider->setLowerValue(rel_clip_start);
    ui->spin_clip_start->setValue(rel_clip_start);

    // Now emit the signal to let interested parties know that the clip start has changed
    emit clip_start_changed();
}

/*!
 * @brief The user has performed some action on the current frame slider.
 * @param action        The action performed on the current frame slider.
 */
void CameraReview::cur_frame_action(int action)
{
    // We're only interested in a few of the actions. We can
    // safely ignore the others.
    switch (action)
    {
    case QAbstractSlider::SliderSingleStepAdd:
    case QAbstractSlider::SliderSingleStepSub:
    case QAbstractSlider::SliderPageStepAdd:
    case QAbstractSlider::SliderPageStepSub:
        // Sanity check.
        if ((ui->sld_current_frame->sliderPosition() < ui->sld_current_frame->minimum()) || (ui->sld_current_frame->sliderPosition() > ui->sld_current_frame->maximum()))
            break;

        // The change event doesn't propagate, so let's make sure things happen.
        ui->spin_current_frame->setValue(ui->sld_current_frame->sliderPosition());

        // Remember the new setting.
        last_frame_num = ui->sld_current_frame->sliderPosition();

        // Start our blackout timer.
        start_blackout();

        // Update our display.
        update_time_display(ui->sld_current_frame->sliderPosition());

        // Let folks know the current frame is different.
        emit current_frame_changed();
        break;
    case QAbstractSlider::SliderMove:
    default:
        break;
    }
}

/*!
 * @brief The user is dragging the current frame slider around.
 * @param new_value     The new current frame value.
 */
void CameraReview::cur_frame_changing(int new_value)
{
    // Make sure our spinbox and slider agree.
    if (new_value != ui->sld_current_frame->value())
    {
        // Set the new value into the slider.
        ui->sld_current_frame->setValue(new_value);
        if (zoom_scrollbar->isVisible())
        {
            // Is the new value outside or highlighted region?
            if ((new_value < zoom_scrollbar->minimum()) || (new_value > zoom_scrollbar->maximum()))
                zoom_change();

            // Now update the scrollbar position.
            zoom_scrollbar->setValue(new_value);
        }
    }

    // Remember the new setting.
    last_frame_num = new_value;

    // Start our blackout timer.
    start_blackout();

    // Update our display.
    update_time_display(new_value);

    // Let folks know that we're changing.
    emit current_frame_changed();
}

/*!
 * @brief The user has pressed the current frame slider to begin
 *        dragging (scrubbing) the current frame.
 */
void CameraReview::cur_frame_pressed()
{
    // Kill the blackout timer if it's running.
    if (blackout_id != 0)
    {
        killTimer(blackout_id);
        blackout_id = 0;
    }

    // Do not allow asynchronous frame updates.
    allow_frame_updates = false;
}

/*!
 * @brief The user has released the current frame slider.
 */
void CameraReview::cur_frame_released()
{
    // Make the release point the "last" point.
    ui->spin_current_frame->setValue(ui->sld_current_frame->value());
}

/*!
 * @brief Make sure we only update the spinner (on a "slow" update) when the user
 *        has the slider down.
 * @param new_value     The new slider position.
 */
void CameraReview::cur_frame_slow_update(int new_value)
{
    // Only pass these along if the slider is down. Ignore others.
    if (ui->sld_current_frame->throttleIsActive())
    {
        // Update the spinner. This will have a ripple effect, eventually
        // causing a new frame to be fetched.
        ui->spin_current_frame->setValue(new_value);
    }
}

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void CameraReview::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    Q_UNUSED(reg_value);

    // Sanity check.
    if (my_parent == nullptr)
        return;

    // Is this event for us?
    if (ignore_updates || (ip_addr != my_parent->get_ip_addr()))
        return;

    // Update our display.
    switch (reg_addr)
    {
    case GIGE_REG_FRAME_RATE:
        update_frame_info();
        update_roi_rate();
        break;
    case GIGE_REG_PB_CONTROL:
        update_playback_control(reg_value);
        break;
    case GIGE_REG_PB_CUT_IN_FRAME:
#if VERBOSE_SYNC_CLIP_LOG
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_PB_CUT_IN_FRAME:" << reg_value;
#endif
        set_pb_clip_start(reg_value);
        break;
    case GIGE_REG_PB_CUT_OUT_FRAME:
#if VERBOSE_SYNC_CLIP_LOG
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_PB_CUT_OUT_FRAME:" << reg_value;
#endif
        set_pb_clip_end(reg_value);
        break;
    case GIGE_REG_PART_NUM_FRAMES:
    case GIGE_REG_PART_TRIG_FRAME:
        break;
    case GIGE_REG_PART_STATE:
        update_part_info(reg_value);
        break;
    case GIGE_REG_WIDTH:
    case GIGE_REG_HEIGHT:
        update_roi_rate();
        break;
    default:
        break;
    }
}

/*!
 * @brief The user pressed the jump forward button.
 */
void CameraReview::jump_forward()
{
    int i;
    int cur_frame;
    int new_frame;

    // Move towards the start of the video.
    cur_frame = ui->sld_current_frame->value();
    if (cur_frame <= ui->spin_clip_start->value())
        new_frame = ui->sld_current_frame->minimum();
    else
        new_frame = ui->spin_clip_start->value();

    // If we have markers, find the one "before" our current position.
    for (i = 0; i < markers().size(); ++i)
    {
        // Get the event frame as an relative frame.
        int event_frame = markers().at(i);

        // Skip the RUN_FLAG marker.
        if (event_frame == INT_MAX)
            continue;

        // See if we've gone too far. If we have, then we're done.
        if (cur_frame <= event_frame)
            break;

        // Save the current marker as a potential jump point.
        // Our "new_frame" is initialized to either the start
        // frame or the "cut_in" frame. Need to make sure that
        // we only take events between that point and the "cur_frame".
        if (new_frame < event_frame)
            new_frame = event_frame;
    }

    // Are things changing?
    if (new_frame != cur_frame)
    {
        // Update the UI.
        ui->spin_current_frame->setValue(new_frame);
        ui->sld_current_frame->setValue(new_frame);
        if (zoom_scrollbar->isVisible())
            zoom_scrollbar->setValue(new_frame);

        // Remember the new setting.
        last_frame_num = new_frame;

        // Start our blackout timer.
        start_blackout();

        // Update our display.
        update_time_display(new_frame);

        // Let folks know that we're changing.
        emit current_frame_changed();
    }
}

/*!
 * @brief The user pressed the jump reverse button.
 */
void CameraReview::jump_reverse()
{
    int i;
    int cur_frame;
    int new_frame;

    // Move towards the end of the video.
    cur_frame = ui->sld_current_frame->value();
    if (cur_frame >= ui->spin_clip_end->value())
        new_frame = ui->sld_current_frame->maximum();
    else
        new_frame = ui->spin_clip_end->value();

    // If we have markers, find the one "after" our current position.
    for (i = 0; i < markers().size(); ++i)
    {
        // Get the event frame as an relative frame.
        int event_frame = markers().at(i);

        // Skip the RUN_FLAG marker.
        if (event_frame == INT_MAX)
            continue;

        // See if we've found "the one".
        if (cur_frame < event_frame)
        {
            // Is this event frame before what we thought our "new_frmae"
            // was going to be? If so, adjust the "new_frame".
            if (event_frame < new_frame)
                new_frame = event_frame;

            // Done.
            break;
        }
    }

    // Are things changing?
    if (new_frame != cur_frame)
    {
        // Update the UI.
        ui->spin_current_frame->setValue(new_frame);
        ui->sld_current_frame->setValue(new_frame);
        if (zoom_scrollbar->isVisible())
            zoom_scrollbar->setValue(new_frame);

        // Remember the new setting.
        last_frame_num = new_frame;

        // Start our blackout timer.
        start_blackout();

        // Update our display.
        update_time_display(new_frame);

        // Let folks know that we're changing.
        emit current_frame_changed();
    }
}

/*!
 * @brief The user pressed the jump to trigger button.
 */
void CameraReview::jump_trigger()
{
    // Are we already at the trigger frame?
    if (ui->sld_current_frame->value() == 0)
        return;

    // Set the current frame to the trigger.
    ui->spin_current_frame->setValue(0);
    ui->sld_current_frame->setValue(0);
    if (zoom_scrollbar->isVisible())
        zoom_scrollbar->setValue(0);

    // Remember the new setting.
    last_frame_num = 0;

    // Start our blackout timer.
    start_blackout();

    // Update our display.
    update_time_display(ui->sld_current_frame->value());

    // Let folks know that we're changing.
    emit current_frame_changed();
}

/*!
 * @brief The user pressed the stop playback button.
 */
void CameraReview::playback_stop()
{

    // Make sure we kill our update timer.
    if (play_update_id != 0)
    {
        killTimer(play_update_id);
        play_update_id = 0;
    }

    // Don't need the timer anymore.
    if (blackout_id != 0)
    {
        killTimer(blackout_id);
        blackout_id = 0;
    }
    allow_frame_updates = true;

    // Not available in player mode.
    if (player_mode)
    {
        // There's no camera, so we won't get the write event coming back.
        emit review_stop();
    }
    else
    {
        uint32_t    pb_control;

        // Remember user requested a stop.
        pb_stop = true;

        // Determine the player type.  If it is not connected to a camera then there is
        // no register to write.
        if (is_playback_from_camera())
        {
            // Modify the playback control register to stop the playback.
            pb_control = read_camera_cache(GIGE_REG_PB_CONTROL);
            pb_control &= ~GIGE_PBC_ENABLE;
            write_register(GIGE_REG_PB_CONTROL, pb_control);
        }
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void CameraReview::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
        return;

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Perform any special action we might want on a particular register.
    switch (reg_addr)
    {
    case GIGE_REG_PB_CUT_IN_FRAME:
#if VERBOSE_SYNC_CLIP_LOG
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_PB_CUT_IN_FRAME:" << read_camera_cache(GIGE_REG_PB_CUT_IN_FRAME);
#endif
        update_clip_start();
        break;
    case GIGE_REG_PB_CUT_OUT_FRAME:
#if VERBOSE_SYNC_CLIP_LOG
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_PB_CUT_OUT_FRAME:" << read_camera_cache(GIGE_REG_PB_CUT_OUT_FRAME);
#endif
        update_clip_end();
        break;
    case GIGE_REG_PB_LAST_FRAME:
        update_playback_frame();
        break;
    case GIGE_REG_PB_STATUS:
        // QLOG_TRACE() << __PRETTY_FUNCTION__ << "GIGE_REG_PB_STATUS:" << read_camera_cache(GIGE_REG_PB_STATUS);
        break;
    default:
        break;
    }
}

/*!
 * @brief   Handle the event when the user presses the "Remove" button.
 *
 *          Determine which row the user wants to be "removed". This doesn't
 *          actually remove the controls, but rather the row of controls is
 *          hidden. Layout is adjust as required. Also, if the "Add Chart"
 *          button was disabled, we can re-enable it.
 */
void CameraReview::remove_chart(QWidget *p_caller)
{
    int i;
    int index;
    int visible_count;
    QPushButton *p_button = static_cast<QPushButton *>(p_caller);

    // Find the control in the list, so we can know which plot to use.
    // Also, count the number of "visible" controls.
    for (index = -1, visible_count = 0, i = 0; i < pb_remove_chart.size(); ++i)
    {
        // Is this a match?
        if (pb_remove_chart.at(i) == p_button)
        {
            // Remember this place so we can get back here.
            index = i;
        }

        // If the plot is visible, then increment our counter.
        if (plot_curve.at(i)->isVisible())
            visible_count++;
    }

    // Sanity check.
    if (index == -1)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Caller not found!";
        return;
    }

    // If there's only the one control visible, then disable the graph display.
    if (visible_count == 1)
    {
        enable_graph(false);
        return;
    }

    // Hide the controls.
    spin_signal.at(index)->hide();
    plot_curve.at(index)->hide();
    pb_remove_chart.at(index)->hide();

    // Is the add button disabled? Since we just deleted an entry
    // we can re-enable it.
    if (!pb_add_chart->isEnabled())
        pb_add_chart->setEnabled(true);
}

void CameraReview::setVisible(bool visible)
{
    // Are we going visible?
    if (visible)
    {
        // Are we in the "player" mode, with a saved capture? Or are we connected
        // to a camera?
        if (!player_mode)
        {
            // Must be connected to a camera, so disable our controls until we get things initialized.
            setEnabled(false);

            // Read the registers we'll need.
            read_panel_regs();

            // Ticket #343:  Allow the user to "truncate" since reviewing a live capture in a camera.
            ui->tb_trash->setHidden(false);
        }
        else
        {
            // No camera, so just make sure our controls are enabled.
            setEnabled(true);
            // Ticket #343:  We don't want to allow user to truncate since this is a captured save on disk.
            ui->tb_trash->setHidden(true);
        }
    }
    else
    {
        CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

        // We may not be a review for a camera.
        if (my_parent != nullptr)
        {
            // Not interested in our register update. Clear them out.
            my_parent->group_read_cancel(GIGE_REG_PB_FRAME_PERIOD);
            my_parent->group_read_cancel(GIGE_REG_CSI_RVW_FRAME_NUM);
        }

        // Don't need the timer anymore.
        if (blackout_id != 0)
        {
            killTimer(blackout_id);
            blackout_id = 0;
        }

        allow_frame_updates = true;
    }

    QWidget::setVisible(visible);
}

/*!
 * @brief   Handle the event of the user changing which signal to be
 *          displayed on a particular row.
 *
 *          Determine which row the user is modifying and update the
 *          graph to reflect the selected signal.
 */
void CameraReview::spin_signal_change(QWidget *p_caller)
{
    fiStringSpinBox *p_spin = static_cast<fiStringSpinBox *>(p_caller);

    // Update the plot for this spin control.
    update_plot(p_spin);
}

/*!
 * @brief The user has pressed the step forward button.
 */
void CameraReview::stepping_forward()
{
    int cur_frame;

    // Bump the current frame by 1.
    cur_frame = ui->sld_current_frame->value() + 1;
    if (cur_frame <= ui->sld_current_frame->maximum())
    {
        // Update the UI.
        ui->spin_current_frame->setValue(cur_frame);
        ui->sld_current_frame->setValue(cur_frame);
        if (zoom_scrollbar->isVisible())
            zoom_scrollbar->setValue(cur_frame);

        // Remember the new setting.
        last_frame_num = cur_frame;

        // Start our blackout timer.
        start_blackout();

        // Update our display.
        update_time_display(cur_frame);

        // Let folks know that we're stepping.
        emit current_frame_changed();
    }
}

/*!
 * @brief The user has pressed the step reverse button.
 */
void CameraReview::stepping_reverse()
{
    int cur_frame;

    // Decrement the current frame by 1.
    cur_frame = ui->sld_current_frame->value() - 1;
    if (cur_frame >= ui->sld_current_frame->minimum())
    {
        // Update the UI.
        ui->spin_current_frame->setValue(cur_frame);
        ui->sld_current_frame->setValue(cur_frame);
        if (zoom_scrollbar->isVisible())
            zoom_scrollbar->setValue(cur_frame);

        // Remember the new setting.
        last_frame_num = cur_frame;

        // Start our blackout timer.
        start_blackout();

        // Update our display.
        update_time_display(cur_frame);

        // Let folks know that we're stepping.
        emit current_frame_changed();
    }
}

/*!
 * @brief A register write operation has completed.
 *
 * @param ip_addr       IP address of the camera written to.
 * @param reg_addr      Register address that was written.
 */
void CameraReview::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
        return;

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Process based on the register that was just written.
    switch (reg_addr)
    {
    case GIGE_REG_PB_CUT_IN_FRAME:
        // Refresh these in our cache if a series 3 camera as it doesn't get unsol notification for these.
        if (my_parent->is_series3())
        {
            my_parent->start_camera_read_reg(GIGE_REG_PB_CUT_IN_FRAME);
        }
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status != GEV_STATUS_SUCCESS)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to write CUT_IN register.";
        }
#if VERBOSE_SYNC_CLIP_LOG
        QLOG_DEBUG() << __PRETTY_FUNCTION__  << "camera:" << my_parent->get_camera_name()
                     << "GIGE_REG_PB_CUT_IN_FRAME:" << read_camera_cache(GIGE_REG_PB_CUT_IN_FRAME) << "series3:" << my_parent->is_series3();
#endif
        break;
    case GIGE_REG_PB_CUT_OUT_FRAME:
        // Refresh these in our cache if a series 3 camera as it doesn't get unsol notification for these.
        if (my_parent->is_series3())
        {
            my_parent->start_camera_read_reg(GIGE_REG_PB_CUT_OUT_FRAME);
        }
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status != GEV_STATUS_SUCCESS)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to write CUT_OUT register.";
        }
#if VERBOSE_SYNC_CLIP_LOG
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "camera:" << my_parent->get_camera_name()
                     << "GIGE_REG_PB_CUT_OUT_FRAME:" << read_camera_cache(GIGE_REG_PB_CUT_OUT_FRAME) << "series3:" << my_parent->is_series3();
#endif
        break;
    case GIGE_REG_PB_CONTROL:
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (pb_stop && (write_status == GEV_STATUS_SUCCESS))
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "User stop complete.";

            // Invalidate and update the PB_STATUS.
            comm_ptr->expire_entry(ip_addr, GIGE_REG_PB_STATUS);
            comm_ptr->start_read_reg(ip_addr, GIGE_REG_PB_STATUS);

            // Let others know that we're stopping.
            emit review_stop();
            pb_stop = false;
        }
        break;
    default:
        break;
    }
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the update record progress' group read's last address == GIGE_REG_CSI_RVW_FRAME_NUM.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void CameraReview::chk_read_panel_regs(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr     = qobject_cast<CameraMDI *>(parent());

    // If we are not connected to any camera we don't care about this signal.
    if (camera_ptr != nullptr)
    {
        // Is this event for us?
        if (ip_addr != camera_ptr->get_ip_addr())
            return;

        // Verify it's our register.
        if (reg_addr == GIGE_REG_CSI_RVW_FRAME_NUM)
        {
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_CSI_RVW_FRAME_NUM]";
#endif
            // Ok. We can start accepting updates now.
            ignore_updates = false;

            // Update some of the display elements that we now know about.
            update_roi_rate();
            update_frame_info();

            // Let the parent know we're ready.
            emit panel_ready();
        }
    }
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the update record progress' group read's last address == GIGE_REG_PB_FRAME_PERIOD.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void CameraReview::chk_read_playback_regs(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr     = qobject_cast<CameraMDI *>(parent());

    // If we are not connected to any camera we don't care about this signal.
    if (camera_ptr != nullptr)
    {
        // Is this event for us?
        if (ip_addr != camera_ptr->get_ip_addr())
            return;

        // Verify it's our register.
        if (reg_addr == GIGE_REG_PB_FRAME_PERIOD)
        {
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_PB_FRAME_PERIOD]";
#endif
            // Use the register information we just finished reading to
            // determine if the camera is in playback mode.
            if (update_playback_status())
            {
                // See if we want to attach to the camera's video stream.
                QTimer::singleShot(10, this, &CameraReview::attach_playback);
            }
        }
    }
}

/*!
 * @brief The user has performed an action on the zoom slider.
 * @param action        The action the user performed.
 */
void CameraReview::zoom_action(int action)
{
    // We're only interested in a few of the actions. We can
    // safely ignore the others.
    switch (action)
    {
    case QAbstractSlider::SliderSingleStepAdd:
    case QAbstractSlider::SliderSingleStepSub:
    case QAbstractSlider::SliderPageStepAdd:
    case QAbstractSlider::SliderPageStepSub:
        // Sanity check.
        if ((zoom_scrollbar->sliderPosition() < zoom_scrollbar->minimum()) || (zoom_scrollbar->sliderPosition() > zoom_scrollbar->maximum()))
            break;

        // The change event doesn't propagate, so let's make sure things happen.
        ui->spin_current_frame->setValue(zoom_scrollbar->sliderPosition());

        // Remember the new setting.
        last_frame_num = zoom_scrollbar->sliderPosition();

        // Start our blackout timer.
        start_blackout();

        // Update our display.
        update_time_display(zoom_scrollbar->sliderPosition());

        emit current_frame_changed();
        break;
    case QAbstractSlider::SliderMove:
    default:
        break;
    }
}

/*!
 * @brief   Update the displayed controls when the user changes the zoom, if the
 *          user is in a "zoomed" mode. Otherwise we'll ignore this event.
 */
void CameraReview::zoom_change()
{
    int cur_frame_rel;

    // Get our "relative" current frame - the frame being shown to the user.
    cur_frame_rel = ui->spin_current_frame->value();

    // If the zoom region of the slider does not contain our current position,
    // then we need to adjust the location of the zoom region.
    if ((ui->clip_slider->get_zoom_start_frame() > cur_frame_rel) ||
        (ui->clip_slider->get_zoom_stop_frame() < cur_frame_rel))
    {
        ui->clip_slider->set_zoom_center(cur_frame_rel);
    }

    // Adjust our slider's min/max values to the zoomed value.
    zoom_scrollbar->setMinimum(ui->clip_slider->get_zoom_start_frame());
    zoom_scrollbar->setMaximum(ui->clip_slider->get_zoom_stop_frame());
    zoom_scrollbar->setPageStep((ui->clip_slider->get_zoom_stop_frame() - ui->clip_slider->get_zoom_start_frame()) / 10);
    zoom_scrollbar->setValue(cur_frame_rel);

    // Update our plot information.
    update_plot();
}

/*!
 * @brief   Handle the event when the user wants to resume "normal" (not zoomed)
 *          timeline operation.
 */
void CameraReview::zoom_hide()
{
    // Hide our zoom controls.
    pb_zoom_in->hide();
    pb_zoom_out->hide();
    zoom_scrollbar->hide();

    // Remove the spacers if an I/O graph isn't being displayed.
    if (!graph_is_enabled())
    {
        // Remove our spacers.
        ui->rightside_layout->removeItem(right_spacer);
        delete right_spacer;
        right_spacer = nullptr;
        ui->leftside_layout->removeItem(left_spacer);
        delete left_spacer;
        left_spacer = nullptr;
    }

    // Update our plot information.
    update_plot();
}

/*!
 * @brief The user pressed the zoom in button.
 */
void CameraReview::zoom_in()
{
    // Tell the control to zoom in.
    ui->clip_slider->update_zoom(true);
    ui->clip_slider->update();
}

/*!
 * @brief   Handle the event when the user moves the zoom region.
 */
void CameraReview::zoom_move()
{
    // Adjust our slider's min/max values to the zoomed value.
    zoom_scrollbar->setMinimum(ui->clip_slider->get_zoom_start_frame());
    zoom_scrollbar->setMaximum(ui->clip_slider->get_zoom_stop_frame());

    // Update our plot information.
    update_plot();
}

/*!
 * @brief The user pressed the zoom out button.
 */
void CameraReview::zoom_out()
{
    // Tell the control to zoom out.
    ui->clip_slider->update_zoom(false);
    ui->clip_slider->update();
}

/*!
 * @brief   Handle the event when the user wants to zoom in on the timeline.
 */
void CameraReview::zoom_show()
{
    // Show our zoom controls.
    pb_zoom_in->show();
    pb_zoom_out->show();
    zoom_scrollbar->show();

    // Add our spacers and put them in the appropraite layout control, if necessary.
    if (right_spacer == nullptr)
    {
        right_spacer = new QSpacerItem(20, 20, QSizePolicy::Minimum, QSizePolicy::Expanding);
        ui->rightside_layout->addItem(right_spacer);
    }
    if (left_spacer == nullptr)
    {
        left_spacer = new QSpacerItem(20, 20, QSizePolicy::Minimum, QSizePolicy::Expanding);
        ui->leftside_layout->insertSpacerItem(ui->leftside_layout->count() - 1, left_spacer);
    }

    // Update our plot information.
    update_plot();
}

