/*!
 * @file camera_review.h
 * @brief This is the header file for the CameraReview class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_REVIEW_H
#define CAMERA_REVIEW_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QWidget>
#include <stdint.h>

namespace Ui
{
class CameraReview;
}

namespace REVIEW_CONTROLS
{

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

const int   MAX_GRAPH_COUNTROLS     = 3;        //!< The maximum number of graphing controls displayed.
const int   DEF_PLAYBACK_RATE       = 20;       //!< The default playback rate.

}   // End of namespace REVIEW_CONTROLS

//-----------------------------------------------------------------------------
// Class Forward Declarations
//-----------------------------------------------------------------------------

class QPushButton;
class QSlider;
class QSpacerItem;
class QTimerEvent;
class QToolButton;
class QwtPlot;

class fiSlider;
class fiStringSpinBox;

class CameraReview : public QWidget
{
    Q_OBJECT

public:
    explicit CameraReview(QWidget *parent = nullptr);
    ~CameraReview();

    // Put common sanity failure constant strings here so all objects can reference them.
    static const QString PLAYER_MODE_RESTRICTION;

    bool        controls_enabled()                  { return showing; }
    void        enable_graph(bool do_enable);
    void        enable_panel(bool do_enable)      { setEnabled(do_enable); }
    void        enable_player_mode();
    uint32_t    get_clip_end();
    uint32_t    get_clip_start();
    uint32_t    get_current_frame_num();
    void        get_current_rel_frame_info(int& rel_frame_num, int& rel_clip_start, int& rel_clip_end);
    int         get_playback_rate();
    uint32_t    get_trigger_frame_num();
    bool        graph_is_enabled()                  { return graph_enable; }
    bool        is_playing();
    bool        is_playback_from_camera();
    void        reset_clip();
    void        set_current_frame_num(uint32_t cur_frame);
    void        set_graph_data(uint8_t *data_ptr, int data_len);
    void        set_markers(const QList<int>& markers);
    void        set_pb_sync_clip(uint32_t clip_start, uint32_t clip_end);
    void        set_pb_clip_end(uint32_t clip_end);
    void        set_pb_clip_start(uint32_t clip_start);
    void        set_pb_frame_count(uint32_t frame_count);
    void        set_pb_frame_rate(uint32_t frame_rate);
    void        set_pb_trigger(uint32_t trig_frame)   { pb_trigger = trig_frame; }
    void        update_buttons(bool do_enable);
    void        update_graph();
    void        update_playback_control(uint32_t pb_control);
    void        view_controls(bool do_view)           { showing = do_view; setVisible(do_view); }

    const QList<int>& markers() const;

signals:
    void    clip_end_changed();
    void    clip_start_changed();
    void    current_frame_changed();

    void    jump_time();
    void    play_forward();
    void    play_reverse();
    void    play_update();
    void    playback_rate_changed(int new_rate);
    void    review_stop();
    void    truncate();

    void    panel_ready();

public slots:
    void    playback_stop();
    void    read_panel_regs();

    void    jump_forward();
    void    jump_reverse();
    void    jump_trigger();
    void    stepping_forward();
    void    stepping_reverse();

    virtual void    setVisible(bool visible);

protected:
    virtual void    changeEvent(QEvent *event);
    virtual void    timerEvent(QTimerEvent *event);

private slots:
    void    attach_playback();
    void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

    void    chk_read_panel_regs(uint32_t ip_addr, uint32_t reg_addr);
    void    chk_read_playback_regs(uint32_t ip_addr, uint32_t reg_addr);

    void    clip_changed();
    void    clip_end_changing(int new_value);
    void    clip_end_clicked();
    void    clip_start_changing(int new_value);
    void    clip_start_clicked();

    void    cur_frame_action(int action);
    void    cur_frame_changing(int new_value);
    void    cur_frame_pressed();
    void    cur_frame_released();
    void    cur_frame_slow_update(int new_value);

    void    add_chart();
    void    remove_chart(QWidget *p_caller);
    void    spin_signal_change(QWidget *p_caller);

    void    zoom_action(int action);
    void    zoom_change();
    void    zoom_hide();
    void    zoom_in();
    void    zoom_move();
    void    zoom_out();
    void    zoom_show();

private:
    void        build_labels();
    void        create_chart_controls();
    void        create_zoom_controls();
    bool        is_series3();
    uint32_t    read_camera_cache(uint32_t reg_addr, bool *ok = nullptr);
    void        read_playback_regs();
    void        start_blackout();
    void        translate_ui();
    void        update_clip_end();
    void        update_clip_start();
    void        update_frame_info();
    void        update_part_info(uint32_t part_state);
    void        update_playback_frame();
    bool        update_playback_status();
    void        update_plot(fiStringSpinBox *p_active_spin = nullptr);
    void        update_roi_rate();
    void        update_time_display(int frame_no);
    void        write_register(uint32_t reg_addr, uint32_t reg_value);

private:
    Ui::CameraReview *ui;

    bool        allow_frame_updates;                //!< Flag to allow/disallow frame updates.

    int         blackout_id;                        //!< Timer ID to blackout updates when we're dynamically moving around a video.
    int         play_update_id;                     //!< Timer ID for pacing the playback of the video.
    bool        player_mode;                        //!< Flag to indicate if we're in "player" mode or not.
    int         scrub_update_id;                    //!< Timer ID to pace the updates when the user is "scrubbing" the slider.

    uint32_t    pb_clip_end;                        //!< The absolute frame number for the end of the clip in playback mode (not attached to a camera).
    uint32_t    pb_clip_start;                      //!< The absolute frame number for the start of the clip in playback mode (not attached to a camera).
    uint32_t    pb_frame_count;                     //!< The number of frames in our playback video.
    uint32_t    pb_frame_rate;                      //!< The frame rate the video was captured at.
    uint32_t    pb_trigger;                         //!< The absolute frame number for the trigger frame when in playback mode (not attached to a camera).
    bool        pb_stop;                            //!< True if the user has requested a stop.

    bool        showing;                            //!< Flag to indiciate if the user wants to see the controls or not. Slightly different from isVisible().
    bool        ignore_updates;                     //!< Flag to ignore updates until we get connected.
    int         last_frame_num;                     //!< Used to remember the last (relative) frame the user selected.
    uint32_t    last_part_state;                    //!< The last partition state we saw.

    // Signal graph support.
    QList<fiStringSpinBox *>      spin_signal;      //!< List of spinner controls to allow the user to specify a signal (Arm-In, Sync-Out, etc.)
    QList<QwtPlot *>        plot_curve;             //!< List of plots showing the signals values.
    QList<QPushButton *>    pb_remove_chart;        //!< List of push button controls to remove an associated IO graph.
    QPushButton     *pb_add_chart;                  //!< Push button to allow the user to add an IO graph.
    QSpacerItem     *left_spacer;                   //!< Spacer control required when more than one IO graph is displayed.
    QSpacerItem     *right_spacer;                  //!< Spacer control required when more than one IO graph is displayed.
    QStringList     signal_names;                   //!< List of the various signals (Arm-In, Sync-Out, etc.)
    bool            graph_enable;                   //!< Flag to indicate if IO graphs are enabled (true) or disabled (false).
    uint8_t         *raw_data_ptr;                  //!< Pointer to the raw per-frame metadata buffer.
    uint32_t        raw_data_len;                   //!< The number of bytes in the raw per-frame metadata buffer.
    int             axis_len;                       //!< The number of entries (aka frames) along the x-axis.
    double          *x_axis;                        //!< The array of x-axis values.
    double          *y_ao_axis;                     //!< The array of Arm-Out values.
    double          *y_ai_axis;                     //!< The array of Arm-In values.
    double          *y_so_axis;                     //!< The array of Sync-Out values.
    double          *y_si_axis;                     //!< The array of Sync-In values.
    double          *y_to_axis;                     //!< The array of Trigger-Out values.
    double          *y_ti_axis;                     //!< The array of Trigger-In values.

    // Zoom controls we need.
    QPushButton     *pb_zoom_in;                    //!< Push button control to allow the user to zoom in on the frame bar.
    QPushButton     *pb_zoom_out;                   //!< Push button control to allow the user to zoom out of the frame bar.
    fiSlider        *zoom_scrollbar;                //!< When zoomed in, the "magnified" frame bar.
};

#endif // CAMERA_REVIEW_H
