/*!
 * @file camera_mdi.cpp
 * @brief This file contains the implementation of the CameraMDI
 *        class. This is the child window class which connects and
 *        displays video, etc. from a specified camera.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "camera_stream.h"

#include <QApplication>

#include "camera_mdi.h"
#include "fi_comm.h"
#include "fsp_socket.h"
#include "logger.h"
#include "mainwindow.h"
#include "panel_reccontrols.h"

namespace CAMERA_STREAM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// When testing additional log messages can appear if these definitions are
// compiled with a 1 instead of a zero.
#define LOG_TIMEOUT_FRAMES                  0
#define LOG_TIMEOUT_SAVER_FRAMES            0

//! The interframe delay for background mode. This is in milliseconds,
//! and we'd like to get around 2 frames per second. We want to run just
//! a little faster than 500 milliseconds.
const int       BACKGROUND_STREAM_RATE      = 400;

//! The maximum number of retries we attempt when receiving a busy response.
const int       MAX_BUSY_RETRIES            = 5;

//! When copying files, we need to give the camera time to move the file from
//! the external media to the internal medai to offload it to us.
const int       COPY_ARRIVAL_TIMEOUT        = 10000;

//! We better get a streaming packet in this time, or else!
const int       FRAME_ARRIVAL_TIMEOUT       = 2000;

//! Pause for a little bit before retrying the stream control register.
const int       RETRY_PAUSE_PERIOD          = 200;

//! The period we'll leave the stream open for quick stepping.
const int       STEP_STREAM_PERIOD          = 1000;

//! The default delay between frames.
const uint32_t  STREAM_DEF_DELAY        = 50;

//! The default size of a packet for streaming video.
const uint32_t  STREAM_DEF_PACKET_SIZE  = 1500;

//! The maximum size of a packet that the camera supports.
const uint32_t  STREAM_MAX_PACKET_SIZE  = (uint32_t)FSP_COMM::JUMBO_PACKET_SIZE;

//! The size we increment our packet length during each phase of the test.
const uint32_t  STREAM_PACKET_INCR          = 500;

//! We better get a packet in this time, or we declare failure.
const int       TEST_PKT_TIMEOUT            = 2000;

//! The maximum number of timeout events before we declare a failure.
const int       TIMEOUT_FAILURE_COUNT       = 3;

//! The maximum number of timeout events before we declare a save failure.
const int       TIMEOUT_SAVE_FAILURE_COUNT  = 5;


//! Define the states we transition through to receive video from the camera.
enum
{
    STATE_INIT,                         //!< Initialize the streaming interface.
    STATE_LINK_NEGOTIATE,               //!< Determine the optimal packet size for the streaming interface. See [stream_fire_test_packet](@ref CameraMDI::stream_fire_test_packet).
    STATE_LIVE,                         //!< A "virtual" state, when we're actually streamining.
    STATE_PAUSE,                        //!< The camera is requested to pause it's stream. See [stream_pause](@ref CameraMDI::stream_pause).
    STATE_SINGLE_FRAME,                 //!< Used for single-stepping a captured video. See [stream_single_frame](@ref CameraMDI::stream_single_frame).
    STATE_PLAYBACK,                     //!< Used to playback a captured video. See [stream_playback](@ref CameraMDI::stream_playback).
    STATE_SAVE_METADATA,                //!< Our streamer is downloading the raw per-frame metadata from the camera. See [stream_save_metadata](@ref CameraMDI::stream_save_metadata).
    STATE_SAVE_SESSION,                 //!< Downloading the camera session to save to a local PC file.
    STATE_SAVE_VIDEO,                   //!< Configures the camera to stream a captured video using a particular image format. See [stream_save_start](@ref CameraMDI::stream_save_video).
    STATE_COPY_FILES,                   //!< Start up a copy stream.
    STATE_IDLE                          //!< A "virutal" state, when we're not initialized or shutting down.
};

//! Define a union of possible parameters for the StreamerStateEvent operations.
typedef union STREAM_EVENT_PARAMS
{
    uint32_t    frame_request;
    int         test_pkt_len;
    STREAM_PARAMS   params;

} STREAM_EVENT_PARAMS;

//! Define a custom event to interface with the streaming state machine.
const QEvent::Type STREAMER_STATE_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_STREAMER_EVENT);

//! Subclass QEvent for our custom streaming state machine event.
class StreamerStateEvent : public QEvent
{
public:
    explicit StreamerStateEvent(int state, CAMERA_STREAM::STREAM_EVENT_PARAMS *param_ptr = nullptr):
        QEvent(STREAMER_STATE_EVENT)
    {
        new_state = state;
        stream_info = param_ptr;
    }

    ~StreamerStateEvent()
    {
        // Release any allocated memory.
        if (stream_info != nullptr)
        {
            delete stream_info;
            stream_info = nullptr;
        }
        new_state = STATE_IDLE;
    }

    int     get_state()                               { return new_state; }
    CAMERA_STREAM::STREAM_EVENT_PARAMS *get_info()    { return stream_info; }

    uint32_t    get_frame()           { return stream_info->frame_request; }
    int         get_test_pkt_len()    { return stream_info->test_pkt_len; }
    CAMERA_STREAM::STREAM_PARAMS    get_stream_params()           { return stream_info->params; }

    void    set_info(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr)    { stream_info = info_ptr; }

private:
    int     new_state;
    CAMERA_STREAM::STREAM_EVENT_PARAMS *stream_info;
};
}   // End of namespace CAMERA_STREAM.

using namespace CAMERA_STREAM;

/*!
 * @brief CameraStream class constructor.
 */
CameraStream::CameraStream(QObject *parent) :
    QObject(parent)
{
    FI_comm     *comm_ptr;

    // Get our parent object cast in a usable form.
    mdi_parent = qobject_cast<CameraMDI *>(parent);

    // Initialize our variables.
    busy_retry_counter = 0;
    current_state = STATE_IDLE;
    detach_pending = false;
    error_recovery = false;
    pending_control_write = 0;
    pending_transition = false;
    next_state = STATE_IDLE;
    next_info = nullptr;
    frames_received = 0;
    step_timer_id = 0;
    fsp_streamer = nullptr;
    test_pkt_timer_id = 0;
    timeout_counter = 0;
    timeout_frames_received = 0;
    stream_packet_len = STREAM_DEF_PACKET_SIZE;
    stream_frame_delay = STREAM_DEF_DELAY;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // We are going to use a Queued connection type later, and Qt needs to know about uint32_t types.
    qRegisterMetaType<uint32_t>("uint32_t");

    // Connect signals/slots.
    connect(comm_ptr, &FI_comm::msg_timeout, this, &CameraStream::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &CameraStream::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraStream::write_complete, Qt::QueuedConnection);

    // Setup our timer which catches dropped frames.
    frame_timer = new QTimer(this);
    if (frame_timer == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate frame_timer!";
        return;
    }
    connect(frame_timer, &QTimer::timeout, this, &CameraStream::frame_timeout);
}

/*!
 * @brief CameraStream class destructor.
 */
CameraStream::~CameraStream()
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr != nullptr)
    {
        // Tell the communciations object we're done.
        comm_ptr->stream_finish(this);
    }

    // Make sure our timer is cleaned up.
    if (frame_timer != nullptr)
    {
        frame_timer->stop();
        frame_timer->deleteLater();
        frame_timer = nullptr;
    }
    if (step_timer_id != 0)
    {
        killTimer(step_timer_id);
        step_timer_id = 0;
    }
    if (test_pkt_timer_id != 0)
    {
        killTimer(test_pkt_timer_id);
        test_pkt_timer_id = 0;
    }
}

/*!
 * @brief Connect to the camera's FSP socket.
 */
void CameraStream::attach()
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get our streamer object.
    fsp_streamer = comm_ptr->get_streamer(mdi_parent->get_ip_addr());
    if (fsp_streamer == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No streamer object!";
        return;
    }

    // Connect our streamer's signals to our slots.
    connect(fsp_streamer, &FSPSocket::image_received, this, &CameraStream::fsp_image_received, Qt::QueuedConnection);
    connect(fsp_streamer, &FSPSocket::fsp_is_ready, this, &CameraStream::fsp_socket_is_ready);
    connect(fsp_streamer, &FSPSocket::test_packet_received, this, &CameraStream::fsp_test_packet_received);

    // Have the streamer initialize its network interface.
    fsp_streamer->init_socket(mdi_parent->is_series3());
}

/*!
 * @brief We had a frame timeout while copying files from the camera.
 */
void CameraStream::copy_files_timeout()
{
    // Don't bother if the camera is off-line.
    if ((mdi_parent == nullptr) || (mdi_parent && mdi_parent->is_offline()))
        return;

    // Are we in a paused state, waiting to catch up?
    if (fsp_streamer->is_paused())
    {
        // Reset our frame timer
        reset_timeout_counter();
        if (frame_timer != nullptr)
        {
            frame_timer->start(COPY_ARRIVAL_TIMEOUT);
        }
    }
    else
    {
        // Make sure we give the system a chance.
        if (timeout_counter >= TIMEOUT_FAILURE_COUNT)
        {
            // Time to let folks know.
            emit timeout();
        }

        else if (frame_timer != nullptr)
        {
            // Restart the timer.
            frame_timer->start(COPY_ARRIVAL_TIMEOUT);
        }
    }
}

/*!
 * @brief Method to start copying from a path on the camera, pulling the data down via the streaming
 *        interface. As either files or "chunks" of files come in, the client will be notified via
 *        the "image_received" signal.
 *
 * @param camera_path   Fully qualified path on the camera for the source material to be downloaded.
 */
void CameraStream::copy_from_camera(QString& camera_path)
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Tell the camera the file/folder we want to copy.
    comm_ptr->start_write_mem(mdi_parent->get_ip_addr(), GIGE_REG_STREAM_FILEPATH, camera_path.toUtf8());
}

/*!
 * @brief Detach the streamer from the camera.
 */
void CameraStream::detach()
{
    // Shutdown the timers, if any are running.
    if (frame_timer != nullptr)
        frame_timer->stop();
    if (step_timer_id != 0)
    {
        killTimer(step_timer_id);
        step_timer_id = 0;
    }
    if (test_pkt_timer_id != 0)
    {
        killTimer(test_pkt_timer_id);
        test_pkt_timer_id = 0;
    }

    // Are we connected to a streamer object?
    if (fsp_streamer != nullptr)
    {
        // Disconnect our streamer's signals from our slots.
        disconnect(fsp_streamer, &FSPSocket::image_received, this, &CameraStream::fsp_image_received);
        disconnect(fsp_streamer, &FSPSocket::fsp_is_ready, this, &CameraStream::fsp_socket_is_ready);
        disconnect(fsp_streamer, &FSPSocket::test_packet_received, this, &CameraStream::fsp_test_packet_received);

        // We're done with the streamer object.
        fsp_streamer = nullptr;
    }

    // Is there anything more to do?
    if (current_state != STATE_IDLE)
    {
        // Remember that we're starting a detach process.
        detach_pending = true;

        // Tell the machine to transition to live streaming.
        QApplication::postEvent(this, new StreamerStateEvent(STATE_IDLE));
        // DELETE ME?? machine(new StreamerStateEvent(STATE_IDLE));
    }
}

/*!
 * @brief The streamer's error handler, we reset & re-initialize the streaming interface
 *        an attempt to restore it to the current state.
 */
void CameraStream::error_handler()
{
    int     stream_state;

    // Shutdown the timers, if any are running.
    if (frame_timer != nullptr)
        frame_timer->stop();
    if (step_timer_id != 0)
    {
        killTimer(step_timer_id);
        step_timer_id = 0;
    }
    if (test_pkt_timer_id != 0)
    {
        killTimer(test_pkt_timer_id);
        test_pkt_timer_id = 0;
    }

    // Sanity checks.
    if (mdi_parent == nullptr)
        return;

    // Are we in shutdown?
    if (mdi_parent->shutdown_in_progress())
        return;

    // Don't bother if the camera is off-line.
    if (mdi_parent->is_offline())
        return;

    // Remember what state we're currently in, so we can get back.
    stream_state = current_state;

    // Do we want to try to recover and resume the stream state?
    switch (stream_state)
    {
    case STATE_LIVE:
        // Tell the machine to transition to live streaming.
        QApplication::postEvent(this, new StreamerStateEvent(STATE_IDLE));

        // Remember that we're in error recovery mode and that we want to transition
        // automatically from the idle state to the live state.
        error_recovery = true;
        recovery_state = STATE_LIVE;
        break;

    case STATE_SINGLE_FRAME:
    case STATE_PLAYBACK:
    case STATE_PAUSE:
        QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "setup to transition from IDLE stream_state:" << stream_state << "to PAUSE(3).";

        // Tell the machine to transition to live streaming.
        QApplication::postEvent(this, new StreamerStateEvent(STATE_PAUSE));
        break;

    case STATE_SAVE_METADATA:
    case STATE_SAVE_SESSION:
    case STATE_SAVE_VIDEO:
    case STATE_COPY_FILES:
        break;

    default:
        // We are in error recover right now, but we don't need any special
        // processing to automatically transition from the paused state.
        error_recovery = false;
        break;
    }
}

/*!
 * @brief Public interface to allow the client to start a stream to download the raw
 *        per-frame metadata.
 * @param stream_info   Parameters to describe the download.
 */
void CameraStream::get_raw_metadata(CAMERA_STREAM::STREAM_PARAMS& stream_info, bool retry)
{
    FI_comm     *comm_ptr;
    uint32_t    camera_state;
    STREAM_EVENT_PARAMS *param_ptr;

    // Retrieve the register value.
    camera_state = mdi_parent->read_camera_cache(GIGE_REG_CAMERA_STATE);
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "retry:" << retry << "current state:" << current_state << "camera_state:" << camera_state;

    // Only read the metadata if the camera is in review or autosave.
    if ((camera_state != GIGE_CAMERA_STATE_REVIEW) && (camera_state != GIGE_CAMERA_STATE_AUTOSAVE))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Wrong camera state:" << camera_state;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Allocate our event structure.
    param_ptr = new STREAM_EVENT_PARAMS;
    if (param_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate STREAM_EVENT_PARAMS structure!";
        return;
    }

    // Copy over our parameters.
    param_ptr->params.clip_end = stream_info.clip_end;
    param_ptr->params.clip_start = stream_info.clip_start;
    param_ptr->params.cur_frame = stream_info.cur_frame;
    param_ptr->params.desired_fps = stream_info.desired_fps;
    param_ptr->params.frame_delay = stream_info.frame_delay;
    param_ptr->params.frame_period = stream_info.frame_period;
    param_ptr->params.save_format = stream_info.save_format;
    param_ptr->params.skip_interval = stream_info.skip_interval;
    param_ptr->params.stream_control = stream_info.stream_control;

    // If this is a retry, pop the last attempt off the list.
    if (retry)
        state_save_metadata_exit(STATE_PAUSE);

    // Now that we have all the parameters, let's make sure we don't collide with someone else
    // trying to download their PFMD.
    comm_ptr->stream_queue(this, new StreamerStateEvent(STATE_SAVE_METADATA, param_ptr));
}

/*!
 * @brief Public method to switch the streaming to live.
 */
void CameraStream::go_live()
{
    FI_comm     *comm_ptr;

    // Sanity check.
    if (fsp_streamer == nullptr)
        return;
    if (!fsp_streamer->is_initialized())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process based on the new state.
    switch (current_state)
    {
    default:
    case STATE_LIVE:
        // Already there.
        stream_live_finished();
        return;

    case STATE_SAVE_METADATA:
        // Ignore it.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Ignoring go_live request.";
        return;

    case STATE_COPY_FILES:
    case STATE_SINGLE_FRAME:
    case STATE_SAVE_SESSION:
    case STATE_SAVE_VIDEO:
    case STATE_IDLE:
    case STATE_INIT:
    case STATE_LINK_NEGOTIATE:
        if (!pending_transition)
        {
            // Remember we've got a pending transition.
            pending_transition = true;

            // Set the next state information to be LIVE.
            next_state = STATE_LIVE;
            next_info = nullptr;
        }
        return;

    case STATE_PLAYBACK:
        // Tell the machine to transition to live streaming.
        QApplication::postEvent(this, new StreamerStateEvent(STATE_LIVE));
        return;
    case STATE_PAUSE:
        // We may be paused, but is the queue empty?
        if (!comm_ptr->stream_queue_is_empty())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Stream is paused, but Queue is not empty. =================================";

            // Reset our stream.
            comm_ptr->stream_finish(this);
            return;
        }
        break;
    }


    // We need to be paused before going to live streaming mode.
    if (!is_paused())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Streamer is not paused!";
    }

    // Tell the machine to transition to live streaming.
    QApplication::postEvent(this, new StreamerStateEvent(STATE_LIVE));
}

/*!
 * @brief Pass it on to our streamer.
 */
int CameraStream::image_count()
{
    int rc = 0;

    if (fsp_streamer != nullptr)
        rc = fsp_streamer->image_count();

    return rc;
}

/*!
 * @brief Pass it on to our streamer.
 * @param imageBufferPtr     Image buffer to be released.
 */
void CameraStream::image_release(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr)
{
    if (fsp_streamer != nullptr)
    {
        fsp_streamer->image_release(std::move(imageBufferPtr));
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "!![RESOURCE LEAK]!! fsp_streamer is NULL!  This will recycle the memory data back to system pool anyway.";
        if (imageBufferPtr)
        {
            QString myInfo = QStringLiteral("[CameraStream::image_release()]");
            imageBufferPtr->LogMyDataId(&myInfo);
        }
    }
}

/*!
 * @brief Pass it on to our streamer.
 */
std::unique_ptr<FI_MEMPOOL::FiImageBuffer> CameraStream::image_take_first()
{
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr;

    if (fsp_streamer != nullptr)
    {
        imageBufferPtr = fsp_streamer->image_take_first();

        // Log a diagnostic out there for this resource.
        if (imageBufferPtr)
        {
            QString infoMsg = QStringLiteral("[CameraStream::image_take_first()]");
            imageBufferPtr->LogMyDataId(&infoMsg);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESOURCE EMPTY:  imageBufferPtr was nullptr!";
        }
    }

    return imageBufferPtr;
}

/*!
 * @brief Map the current streaming state to the type of image which is
 *        being passed on.
 * @return One of the streaming types.
 */
int CameraStream::image_type()
{
    int type_value;

    // Process based on the new state.
    switch (current_state)
    {
    case STATE_COPY_FILES:
        type_value = STREAM_COPY_IMAGE;
        break;
    case STATE_LIVE:
    case STATE_PLAYBACK:
    case STATE_SINGLE_FRAME:
        type_value = STREAM_LIVE_IMAGE;
        break;
    case STATE_SAVE_METADATA:
        type_value = STREAM_RAW_METADATA;
        break;
    case STATE_SAVE_SESSION:
        type_value = STREAM_SAVE_SESSION;
        break;
    case STATE_SAVE_VIDEO:
        type_value = STREAM_SAVE_IMAGE;
        break;
    default:
    case STATE_IDLE:
    case STATE_INIT:
    case STATE_LINK_NEGOTIATE:
    case STATE_PAUSE:
        type_value = STREAM_UNKNOWN;
        break;
    }

    // Return the image type.
    return type_value;
}


/*!
 * @brief We can be considered "active" if we have a streamer object.
 * @return True if we're active, false otherwise.
 */
bool CameraStream::is_active()
{
    // Do we have a streamer object?
    return (fsp_streamer != nullptr);
}

/*!
 * @brief Return true if the streamer is copying a video.
 */
bool CameraStream::is_copying()
{
    return (current_state == STATE_COPY_FILES);
}

/*!
 * @brief Return true if the streamer is paused.
 */
bool CameraStream::is_paused()
{
    // Sanity check.
    if (fsp_streamer == nullptr)
        return true;

    // We're "paused" if the streamer is in the idle state.
    return fsp_streamer->is_idle();
}

/*!
 * @brief Return true if the streamer is playing a video.
 */
bool CameraStream::is_playing()
{
    return (current_state == STATE_PLAYBACK);
}

/*!
 * @brief Return true if the streamer is in the single-step state.
 */
bool CameraStream::is_stepping()
{
    // We're in the single step state if our stepping timer is non-zero.
    return (step_timer_id != 0);
}

/*!
 * @brief We had a frame timeout while in live streaming mode.
 */
void CameraStream::live_frame_timeout()
{
    bool tolerate_timeout = false;

    // Check for sync-in condition where we stopped receiving the signal pulse.
    // If we are in sync-in allow the camera object to notify the user about a
    // loss of sync-in frame stream.
    if (mdi_parent != nullptr)
    {
        // We could be in an autosave overlap condition where there are no buffers.
        // This needs to be tolerated.  Ticket #401.
        if (mdi_parent->is_busy())
        {
#if LOG_TIMEOUT_FRAMES
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "Tolerating live frame timeout for autosave condition.";
#endif
            tolerate_timeout = true;
        }

        // If the camera has enabled sync-in this time out is likely due to a loss
        // of that sync-in signal.  We give the camera a chance to report it here.
        if (mdi_parent->is_sync_in_enabled())
        {
            tolerate_timeout = true;

            // The reporting of the sync-in suspension is now handled by the monitoring
            // of the GIGE_REG_SYNC_IN_TIMING value. We will not initiate the process of
            // sending the sync-in suspended warning to the user from here anymore.
        }
        // Otherwise the timeout still may have been a leftover from a sync-in signal
        // that has been restored.  Using a threshold ask the camera if it had been
        // handling sync-in signal loss warnings and was the last one within this
        // threshold (I am useing 20000 here) - this should catch any leftover time out
        // cases that may have been due to sync-in signal loss.
        else if (mdi_parent->is_sync_in_warning_in_threshold(20000))
        {
            tolerate_timeout = true;
        }

        // Is ok to tolerate this frame time-out due to a sync-in suspended situation?
        // If so facilitate another round of live-frame time-outs.
        if (tolerate_timeout)
        {
            // Reset our counter.
            reset_timeout_counter();
        }
    }

#if 0
    PanelRecControls *controls_ptr;

// Get our record control panel object.
    controls_ptr = static_cast<PanelRecControls *>
                   (MainWindow::GetInstance()->get_control_panel()->getPanel(CONTROL_PANEL::PANEL_REC_CONTROLS));
    if (controls_ptr->syncInIsEnabled())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "SyncIn is enabled. Nothing to see here.";

        // Reset our counter.
        timeout_counter = 0;

        // Restart our timer and wait a little while longer.
        if (frame_timer != nullptr)
            frame_timer->start(FRAME_ARRIVAL_TIMEOUT);
    }
#endif

    // When we get a frame timeout, we always have to worry about the
    // camera going away. If the camera is gone, or if we've exceeded our
    // threshold, then reset the stream.
    if (((mdi_parent == nullptr) || (mdi_parent && mdi_parent->is_offline())) || (timeout_counter >= TIMEOUT_FAILURE_COUNT))
    {
        // This is more than we can deal with.
        error_handler();
    }
    else if (frame_timer != nullptr)
    {
        // Restart our timer and wait a little while longer.
        frame_timer->start(FRAME_ARRIVAL_TIMEOUT);
    }
}

/*!
 * @brief Turn the state value for the camera stream into a readable QString
 * @param stateValue     The state to translate.
 * @return A QString in human readable form.
 */
QString CameraStream::stateToString(int stateValue)
{
    QString resultString;

    switch (stateValue)
    {
    case STATE_INIT:
        resultString = "STATE_INIT";
        break;
    case STATE_LINK_NEGOTIATE:
        resultString = "STATE_LINK_NEGOTIATE";
        break;
    case STATE_LIVE:
        resultString = "STATE_LIVE";
        break;
    case STATE_PAUSE:
        resultString = "STATE_PAUSE";
        break;
    case STATE_SINGLE_FRAME:
        resultString = "STATE_SINGLE_FRAME";
        break;
    case STATE_PLAYBACK:
        resultString = "STATE_PLAYBACK";
        break;
    case STATE_SAVE_METADATA:
        resultString = "STATE_SAVE_METADATA";
        break;
    case STATE_SAVE_SESSION:
        resultString = "STATE_SAVE_SESSION";
        break;
    case STATE_SAVE_VIDEO:
        resultString = "STATE_SAVE_VIDEO";
        break;
    case STATE_COPY_FILES:
        resultString = "STATE_COPY_FILES";
        break;
    case STATE_IDLE:
        resultString = "STATE_IDLE";
        break;
    default:
        resultString = "UNKNOWN";
        break;
    }
    return resultString;
}

/*!
 * @brief Our streaming state machine. The state machine is driven by the StreamerStateEvent events.
 * @param event     The new event "feeding" the state machine.
 */
void CameraStream::machine(QEvent *event)
{
    StreamerStateEvent *event_ptr = static_cast<StreamerStateEvent *>(event);

    // Sanity check.
    if (event_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "No event pointer!";
        return;
    }
    if ((fsp_streamer == nullptr) && !detach_pending)
    {
        // If there are state changes happening before we actually start streaming,
        // we could end up here. Perfectly logical...
        return;
    }

    // Dump some trace when we have a transition.
    if (current_state != event_ptr->get_state())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "current state:" << stateToString(current_state) << "next_state:" << stateToString(event_ptr->get_state());
    }

    // If there is a pending transition (we're back after completing a task, such as
    // exiting the previous state), or if we're really not changing state, then there
    // is no need to exit (or re-exit) the current state.
    if (!pending_transition && (current_state != event_ptr->get_state()))
    {
        // Exit the current state and see if we want to continue, or do we need to wait?
        if (!machine_state_exit(event_ptr->get_state()))
        {
            // Remember we've got a pending transition.
            pending_transition = true;

            // Save the new state so we can (hopefully) transition there later.
            next_state = event_ptr->get_state();

            // Save the parameters (if any).
            next_info = event_ptr->get_info();

            // Reset the info pointer in the event structure so it doesn't get released.
            event_ptr->set_info(nullptr);
            return;
        }
    }

    // Are we OK to transition?
    if (!transition_ok(event_ptr->get_state()))
    {
        // Log the problem.
        QLOG_WARN() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << QString("Transition from current_state [%1] to next_state [%2] not allowed.").arg(current_state).arg(event_ptr->get_state());

        // Reset the info pointer in the event structure so it doesn't get released.
        event_ptr->set_info(nullptr);
        return;
    }

    // Enter the new state.
    machine_state_enter(event_ptr->get_state(), event_ptr->get_info());

    // Did we finish our pending transition?
    if (pending_transition && (current_state == next_state))
    {
        // Reset our pending information.
        pending_transition = false;
        next_state = STATE_IDLE;
        if (next_info != nullptr)
        {
            delete next_info;
            next_info = nullptr;
        }
    }
}

/*!
 * @brief Perform the processing needed to enter into the new state.
 * @param new_state     The new state we wish to enter.
 * @param info_ptr      Parameters associated with the new state.
 */
void CameraStream::machine_state_enter(int new_state, CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr)
{
    // Process based on the new state.
    switch (new_state)
    {
    case STATE_COPY_FILES:
        state_copy_files_enter();
        break;
    case STATE_LINK_NEGOTIATE:
        state_link_negotiate_enter(info_ptr->test_pkt_len);
        break;
    case STATE_LIVE:
        state_live_enter();
        break;
    case STATE_PAUSE:
        state_pause_enter();
        break;
    case STATE_PLAYBACK:
        state_playback_enter(info_ptr);
        break;
    case STATE_SAVE_METADATA:
        state_save_metadata_enter(info_ptr);
        break;
    case STATE_SAVE_SESSION:
        state_save_session_enter(info_ptr);
        break;
    case STATE_SAVE_VIDEO:
        state_save_video_enter(info_ptr);
        break;
    case STATE_SINGLE_FRAME:
        state_single_frame_enter(info_ptr->frame_request);
        break;
    case STATE_IDLE:
        state_stop();
        break;

    default:
    case STATE_INIT:
        state_init_enter();
        break;
    }

    // Were we pending and have we successfully entered the new state?
    if (pending_transition && (new_state == next_state))
    {
        // State transition is no longer pending.
        pending_transition = false;
    }
    if (error_recovery && (new_state == recovery_state))
    {
        // Error recovery is no longer pending.
        error_recovery = false;
    }

    // Update the state from the event information.
    current_state = new_state;

    // Reset our frames_received accumulator for the new state.
    frames_received = 0;
}

/*!
 * @brief Process the exiting from the current state. There are some transitions where
 *        we will have to wait for a response from the camera before we can safely transition
 *        to the new state.
 *
 * @return True if we can move to the new state, false if we have to wait.
 */
bool CameraStream::machine_state_exit(int new_state)
{
    bool    move_to_new_state = false;

    // Handle any exit conditions from the current state.
    switch (current_state)
    {
    case STATE_COPY_FILES:
    case STATE_LIVE:
    case STATE_PLAYBACK:
    case STATE_SAVE_SESSION:
    case STATE_SAVE_VIDEO:
        // Exit the active state. If we have to wait (it returns true), then we
        // do NOT want to move to the new state. Otherwise it's OK to move to the
        // new state.
        move_to_new_state = !state_exit(new_state);
        break;
    case STATE_SAVE_METADATA:
        move_to_new_state = !state_save_metadata_exit(new_state);
        break;
    case STATE_SINGLE_FRAME:
        move_to_new_state = !state_single_frame_exit(new_state);
        break;

    case STATE_INIT:
    case STATE_IDLE:
    case STATE_LINK_NEGOTIATE:
    case STATE_PAUSE:
    default:
        // No exit handling required.
        move_to_new_state = true;
        break;
    }

    return move_to_new_state;
}

/*!
 * @brief Public method to pause the streaming.
 */
void CameraStream::pause_streaming()
{
    // Make sure we're not already in the pause state.
    if (current_state != STATE_PAUSE)
    {
        // Have the streamer pause.
        QApplication::postEvent(this, new StreamerStateEvent(STATE_PAUSE));
    }
}

/*!
 * @brief Public interface to allow the client to initiate a playback stream.
 * @param stream_info   Parameters to describe the playback.
 */
void CameraStream::playback(const STREAM_PARAMS& stream_info)
{
    STREAM_EVENT_PARAMS *param_ptr;

    // Allocate our event structure.
    param_ptr = new STREAM_EVENT_PARAMS;
    if (param_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Unable to allocate STREAM_EVENT_PARAMS structure!";
        return;
    }

    // Copy over our parameters.
    param_ptr->params.clip_end = stream_info.clip_end;
    param_ptr->params.clip_start = stream_info.clip_start;
    param_ptr->params.cur_frame = stream_info.cur_frame;
    param_ptr->params.desired_fps = stream_info.desired_fps;
    param_ptr->params.frame_delay = stream_info.frame_delay;
    param_ptr->params.frame_period = stream_info.frame_period;
    param_ptr->params.save_format = stream_info.save_format;
    param_ptr->params.skip_interval = stream_info.skip_interval;
    param_ptr->params.stream_control = stream_info.stream_control;

    // Get the ball rolling.
    QApplication::postEvent(this, new StreamerStateEvent(STATE_PLAYBACK, param_ptr));
}

/*!
 * @brief On our successful write of the GIGE_REG_SCPD0 register, we want to start the
 *        link negotiation with the camera.
 */
void CameraStream::process_scpd0()
{
    STREAM_EVENT_PARAMS *param_ptr;

    // Are we in the INIT state? We only use the successful write to the SCPD0 register
    // as a transition for the INIT -> LINK_NEGOTIATE transition.
    if (current_state != STATE_INIT)
        return;

    // Tell the streamer to expect test packets.
    fsp_streamer->start_stream(true, mdi_parent->get_camera_info()->packet_delay);

    // Move on to the testing phase.
    param_ptr = new STREAM_EVENT_PARAMS;
    if (param_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Unable to allocate STREAM_EVENT_PARAMS structure!";
        return;
    }
    param_ptr->test_pkt_len = STREAM_DEF_PACKET_SIZE;
    QApplication::postEvent(this, new StreamerStateEvent(STATE_LINK_NEGOTIATE, param_ptr));
}

/*!
 * @brief Process the GIGE_REG_STREAM_CONTROL register.
 */
void CameraStream::process_stream_control()
{
    uint32_t    reg_value;

    // Retrieve the register value.
    reg_value = mdi_parent->read_camera_cache(GIGE_REG_STREAM_CONTROL) & ((uint32_t)~GIGE_STREAM_CONTROL_MASK);
    // QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "GIGE_REG_STREAM_CONTROL:" << Qt::showbase << Qt::hex << reg_value;

    // See if we care about the value.
    switch (reg_value)
    {
    case GIGE_STREAM_LIVE:
        stream_live_finished();
        break;
    case GIGE_STREAM_RECORDED:
    case GIGE_STREAM_SLAVE_ENABLE:
        // Let folks know we're streaming.
        emit streaming();
        break;
    case GIGE_STREAM_DOWNLOAD:
    case GIGE_STREAM_GET_FILE:
    case GIGE_STREAM_SESSION:
    case GIGE_STREAM_RAW_METADATA:
    case GIGE_STREAM_SINGLESTEP:
        // Nothing to do here - the parent will start receiving "image_received" signals.
        break;
    case GIGE_STREAM_STOP:
        stream_pause_finished();
        break;
    default:
        break;
    }
}

/*!
 * @brief Request a single frame from the camera.
 * @param frame_index   The absolute frame number.
 */
void CameraStream::request_frame(uint32_t frame_index)
{
    uint32_t            stream_control;
    STREAM_EVENT_PARAMS *param_ptr;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "current_state:" << current_state;

    // Retrieve the stream control register value.
    stream_control = mdi_parent->read_camera_cache(GIGE_REG_STREAM_CONTROL);
    if ((stream_control != GIGE_STREAM_STOP) && (stream_control != GIGE_STREAM_SINGLESTEP))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "GIGE_REG_STREAM_CONTROL:" << Qt::showbase << Qt::hex << stream_control;
        return;
    }

    // We can only request a frame in certain states.
    if ((current_state != STATE_PAUSE) && (current_state != STATE_SINGLE_FRAME))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "current_state:" << current_state;
        return;
    }

    // Are we in a pending transition.
    if (pending_transition && ((next_state != STATE_PAUSE) && (next_state != STATE_SINGLE_FRAME)))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "pending_transition && next_state:" << next_state << "current_state:" << current_state;
        return;
    }

    // Transition to the single frame state.
    param_ptr = new STREAM_EVENT_PARAMS;
    if (param_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Unable to allocate STREAM_EVENT_PARAMS structure!";
        return;
    }
    param_ptr->frame_request = frame_index;
    QApplication::postEvent(this, new StreamerStateEvent(STATE_SINGLE_FRAME, param_ptr));
}

/*!
 * @brief When downloading data, retry a frame or "chunk" from the camera.
 * @param retry_index   The frame or chunk number to retry.
 */
void CameraStream::retry(uint32_t retry_index)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << retry_index << "current_state:" << current_state;

    // Make sure we're in the correct state.
    if (current_state == STATE_SAVE_VIDEO)
    {
        // Write out our registers.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_STOP);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_START_NUM, retry_index);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_END_NUM, retry_index);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_SEQ_START, retry_index);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_DOWNLOAD);
        pending_control_write = GIGE_STREAM_DOWNLOAD;
        busy_retry_counter = 0;
    }
    else if (current_state == STATE_COPY_FILES)
    {
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_STOP);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_RETRY, retry_index);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_GET_FILE);
        pending_control_write = GIGE_STREAM_GET_FILE;
        busy_retry_counter = 0;
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Wrong stream state:" << current_state;
        return;
    }

    // Restart our frame timeout timer.
    if (frame_timer != nullptr)
    {
        reset_timeout_counter();
        frame_timer->start();
    }
}

/*!
 * @brief Retry setting the stream control register.
 */
void CameraStream::retry_stream_control()
{
    // Process based on the state.
    switch (current_state)
    {
    case STATE_SAVE_METADATA:
    case STATE_PLAYBACK:
    case STATE_COPY_FILES:
    case STATE_LIVE:
    case STATE_SAVE_VIDEO:
    case STATE_SINGLE_FRAME:
    case STATE_PAUSE:
        ++busy_retry_counter;
        if (busy_retry_counter > MAX_BUSY_RETRIES)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "busy_retry_counter exceeded! pending_control_write =" << Qt::showbase << Qt::hex << pending_control_write;

            // Stop the frame timeout timer.
            if (frame_timer != nullptr)
                frame_timer->stop();

            // Pretend that it went through.
            stream_pause_finished();

            // Force into the IDLE state.
            pending_transition = false;
            error_recovery = false;
            next_state = STATE_IDLE;
            current_state = STATE_IDLE;
        }
        else
        {
            mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, pending_control_write);
        }
        break;

    case STATE_LINK_NEGOTIATE:
    case STATE_IDLE:
        break;

    default:
    case STATE_INIT:
        break;
    }
}

/*!
 * @brief Public interface to allow the client to start a stream to download the video of
 *        a capture in a particular format.
 * @param stream_info   Parameters to describe the download.
 */
void CameraStream::save_video(CAMERA_STREAM::STREAM_PARAMS& stream_info)
{
    STREAM_EVENT_PARAMS *param_ptr;

    // Allocate our event structure.
    param_ptr = new STREAM_EVENT_PARAMS;
    if (param_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Unable to allocate STREAM_EVENT_PARAMS structure!";
        return;
    }

    // Copy over our parameters.
    param_ptr->params.clip_end = stream_info.clip_end;
    param_ptr->params.clip_start = stream_info.clip_start;
    param_ptr->params.cur_frame = stream_info.cur_frame;
    param_ptr->params.desired_fps = stream_info.desired_fps;
    param_ptr->params.frame_delay = stream_info.frame_delay;
    param_ptr->params.frame_period = stream_info.frame_period;
    param_ptr->params.save_format = stream_info.save_format;
    param_ptr->params.skip_interval = stream_info.skip_interval;
    param_ptr->params.stream_control = stream_info.stream_control;

    // Get the ball rolling.
    QApplication::postEvent(this, new StreamerStateEvent(STATE_SAVE_VIDEO, param_ptr));
}

/*!
 * @brief Common logic to reset_timeout_counter and images received accumulator.
 */
void CameraStream::reset_timeout_counter()
{
    timeout_counter = 0;
    timeout_frames_received = 0;
}

/*!
 * @brief We had a frame timeout while downloading metadata from the camera.
 */
void CameraStream::save_metadata_timeout()
{
    // Don't bother if the camera is off-line.
    if ((mdi_parent == nullptr) || (mdi_parent && mdi_parent->is_offline()))
        return;

    // Are we in a paused state, waiting to catch up?
    if (fsp_streamer->is_paused())
    {
        // Restart the timer.
        if (frame_timer != nullptr)
        {
            reset_timeout_counter();
            frame_timer->start(2 * FRAME_ARRIVAL_TIMEOUT);
        }
    }
    else
    {
        // Make sure we give the system a chance.
        if (timeout_counter >= TIMEOUT_FAILURE_COUNT)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "timeout_counter expired.";

            // Time to let folks know.
            emit timeout();
        }

        else if (frame_timer != nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Retry:" << timeout_counter;

            // Restart the timer.
            frame_timer->start(2 * FRAME_ARRIVAL_TIMEOUT);
        }
    }
}

/*!
 * @brief Method to start saving the session. As "chunks" of the session come in, the client will be notified via
 *        the "image_received" signal.
 */
void CameraStream::save_session(const STREAM_PARAMS& stream_info)
{
    STREAM_EVENT_PARAMS *param_ptr;

    // Allocate our event structure.
    param_ptr = new STREAM_EVENT_PARAMS;
    if (param_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Unable to allocate STREAM_EVENT_PARAMS structure!";
        return;
    }

    // Copy over our parameters.
    param_ptr->params.clip_end = stream_info.clip_end;
    param_ptr->params.clip_start = stream_info.clip_start;

    // Tell the stream state machine to start the session save.
    QApplication::postEvent(this, new StreamerStateEvent(STATE_SAVE_SESSION, param_ptr));
}

/*!
 * @brief We had a frame timeout while saving a video.
 */
void CameraStream::save_video_timeout()
{
    bool handled_timeout = false;
    bool image_count_in_streamer = 0;

    // Don't bother if the camera is off-line.
    if ((mdi_parent == nullptr) || (mdi_parent && mdi_parent->is_offline()))
        return;

    // What kind of state is our input socket streamer in at this point in time?
    if (fsp_streamer)
    {
        // Grab the buffer count.
        image_count_in_streamer = fsp_streamer->image_count();

        // Is the socket streamer paused?
        if (fsp_streamer->is_paused())
        {
            // Have we waited long enough?
            if (timeout_counter >= TIMEOUT_SAVE_FAILURE_COUNT)
            {
                // Do we have pending images?
                if (image_count_in_streamer > 0)
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "image_count:" << image_count_in_streamer;

                    // Let other's know that we've received an image.
                    emit image_received();
                }
                else
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "No images in the pipeline. Force FSP resume.";

                    // Try to get the stream moving again.
                    fsp_streamer->resume_stream();
                }

                // Reset our counter.
                reset_timeout_counter();
            }
#if LOG_TIMEOUT_SAVER_FRAMES
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() <<
                             "timeout_counter:" << timeout_counter
                             << "frames_received:" << frames_received
                             << "timeout_frames_received:" << timeout_frames_received
                             << "image_count:" << image_count_in_streamer;
            }
#endif

            // Restart the timer.
            if (frame_timer != nullptr)
            {
                frame_timer->start(FRAME_ARRIVAL_TIMEOUT);
            }

            handled_timeout = true; // We have processed this timeout here.
        }
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "fsp_streamer does not exist!";
    }

    // If this timeout was not handled we need to process it.
    if (!handled_timeout)
    {
#if LOG_TIMEOUT_SAVER_FRAMES
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() <<
                     "streamer is not paused timeout_counter:" << timeout_counter
                     << "frames_received:" << frames_received
                     << "timeout_frames_received:" << timeout_frames_received
                     << "image_count:" << image_count_in_streamer;
#endif
        // Make sure we give the system a chance.
        if (timeout_counter >= TIMEOUT_SAVE_FAILURE_COUNT)
        {
            // Time to let folks know.
            emit timeout();
        }
        else if (frame_timer != nullptr)
        {
            // Restart the timer.
            frame_timer->start(FRAME_ARRIVAL_TIMEOUT);
        }
    }
}

/*!
 * @brief The user would like to change the delay used between frames.
 * @param new_delay      The new delay, in milliseconds.
 */
void CameraStream::set_interframe_delay(uint32_t new_delay)
{
    // Sanity check.
    if (mdi_parent == nullptr)
        return;

    // Save the value.
    stream_frame_delay = new_delay;

    // Tell the camera what framerate we want. If we're not the active child, we want to run in the background.
    if (MainWindow::GetInstance()->active_mdi_child() != mdi_parent)
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, BACKGROUND_STREAM_RATE);
    else
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, stream_frame_delay);
}

/*!
 * @brief Setup the camera to start streaming the file(s) that were
 *        specified in an earlier write to GIGE_REG_STREAM_FILEPATH.
 */
void CameraStream::state_copy_files_enter()
{
    // Tell the streamer to expect "real" (aka not test) packets.
    fsp_streamer->start_stream(false, mdi_parent->get_camera_info()->packet_delay);

    // Clear the inter-frame delay.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, 0);

    // Reset the retry register.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_RETRY, 0);

    // Have the camera start streaming the data.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_GET_FILE);
    pending_control_write = GIGE_STREAM_GET_FILE;
    busy_retry_counter = 0;

    // Restart our frame timeout timer.
    if (frame_timer != nullptr)
    {
        reset_timeout_counter();
        frame_timer->start(COPY_ARRIVAL_TIMEOUT);
    }
}


/*!
 * @brief Default method to be called when we're exiting a state.
 * @return True if we need to wait for a camera response, false if we don't.
 */
bool CameraStream::state_exit(int new_state)
{
    bool        wait_for_action;

    // Assume there's no need to wait.
    wait_for_action = false;

    // Make sure we're paused before exiting the live state. Transitioning
    // to the paused state is as good as being paused.
    if (!is_paused() && (new_state != STATE_PAUSE))
    {
        // Make sure we're not in shutdown.
        if (mdi_parent && !mdi_parent->shutdown_in_progress())
        {
            // Come back around to pause the stream.
            QApplication::postEvent(this, new StreamerStateEvent(STATE_PAUSE));

            // We need to wait for the stream to pause.
            wait_for_action = true;
        }
        else
        {
            // We need to stop the stream now. Call it directly.
            stream_pause();
        }
    }

    // Return our wait flag.
    return wait_for_action;
}

/*!
 * @brief Initialize the streaming interface. This writes the streaming registers
 *        on the camera. Once we're convinced we have [written](@ref CameraStream::write_complete)
 *        the values, we'll transition to the [link negotiation state](@ref CAMERA_STREAM::STATE_LINK_NEGOTIATE).
 */
void CameraStream::state_init_enter()
{
    const FI_COMM::FI_CAMERA_INFO *camera_ptr;

    // Sanity checks.
    if ((fsp_streamer == nullptr) || (mdi_parent == nullptr))
        return;

    // Get our camera information object.
    camera_ptr = mdi_parent->get_camera_info();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "No FI_CAMERA_INFO pointer!";
        return;
    }

    // Handle series3 specific tasks.
    if (mdi_parent->is_series3())
    {
        // We need to tell the camera our IP address so it know where to stream to.
        mdi_parent->write_camera_register(GIGE_REG_SCDA0, fsp_streamer->get_local_address().toIPv4Address());

        // The camera also needs to know the port.
        mdi_parent->write_camera_register(GIGE_REG_SCP0, (uint32_t)fsp_streamer->get_port());

        // Although it's not really pertinent for this test phase, tell the camera what
        // inter-packet delay we think we want.
        mdi_parent->write_camera_register(GIGE_REG_SCPD0, camera_ptr->packet_delay);
    }
    else
    {
        // Go straight to expecting a video stream. The stream packet length is meaningless
        // for a TCP connection, but let's make folks happy.
        stream_packet_len = STREAM_MAX_PACKET_SIZE;

        // Bump our state to indicate we're done.
        QApplication::postEvent(this, new StreamerStateEvent(STATE_PAUSE));

        QLOG_INFO() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "(FAKED) Successfully negotiated with" << mdi_parent->get_ip_addr_pretty() << "PktLen:" << stream_packet_len << "bytes";

        // Let folks know we're ready to stream.
        emit ready();
    }
}

/*!
 * @brief Negotiate with the camera to determine the maximum sized packet we can use
 *        with the FSP interface.
 *
 * @param pkt_len   The size of test packet we would like the camera to send.
 */
void CameraStream::state_link_negotiate_enter(int pkt_len)
{
    uint32_t    reg_value;

    // Sanity checks.
    if ((fsp_streamer == nullptr) || (mdi_parent == nullptr))
        return;

    // Are we done?
    if (pkt_len > (int)STREAM_MAX_PACKET_SIZE)
    {
        // All done.
        test_complete(false);
        return;
    }

    // Update our test packet len.
    stream_packet_len = pkt_len;

    // Tell the camera what size test packet to send.
    reg_value = (uint32_t)stream_packet_len;
    reg_value |= GIGE_SCPSX_FTP_BE;  // we want a test packet on the stream channel
    reg_value |= GIGE_SCPSX_DNF_BE;

    // Have the camera fire a test packet.
    mdi_parent->write_camera_register(GIGE_REG_SCPS0, reg_value);

    // Setup a single shot timer to catch when we've gone too far.
    test_pkt_timer_id = startTimer(TEST_PKT_TIMEOUT);
}

/*!
 * @brief Setup the camera and the streamer object to start showing live video. This
 *        doesn't have much to do with the camera's state, except that the camera can't
 *        send live video in certain states.
 */
void CameraStream::state_live_enter()
{
    const FI_COMM::FI_CAMERA_INFO *camera_ptr;

    // Sanity checks.
    if ((fsp_streamer == nullptr) || (mdi_parent == nullptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "fsp_streamer or mdi_parent are null";
        return;
    }

    // Get our camera information object.
    camera_ptr = mdi_parent->get_camera_info();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "No FI_CAMERA_INFO pointer!";
        return;
    }

    // Does the user want to stream?
    if (!camera_ptr->stream)
    {
        // Go back to the "paused" state. We can't just call pause_streaming() because
        // that checks to see if we're paused. Well, we ARE, but we won't be as soon as
        // we take this return. We need to post the event so we'll come back and switch
        // back to "paused".
        QApplication::postEvent(this, new StreamerStateEvent(STATE_PAUSE));
        return;
    }

    // If we're not already running (we might be here 'cuz of state change)
    // make sure the camera is setup to stream.
    if (fsp_streamer->is_idle())
    {
        uint32_t    reg_value;

        // Tell the camera what size we've settled on.
        reg_value = (uint32_t)stream_packet_len;
        reg_value |= GIGE_SCPSX_DNF_BE;

        // Tell the camera the packet size we want to use.
        mdi_parent->write_camera_register(GIGE_REG_SCPS0, reg_value);

        // Tell the camera about our inter-packet delay.
        mdi_parent->write_camera_register(GIGE_REG_SCPD0, camera_ptr->packet_delay);

        // Tell the camera what framerate we want. If we're not the active child, we want to run in the background.
        if (MainWindow::GetInstance()->active_mdi_child() != mdi_parent)
            mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, BACKGROUND_STREAM_RATE);
        else
            mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, stream_frame_delay);

        // Make sure the format is JPEG.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FORMAT, GIGE_STREAM_FORMAT_JPEG);

        // Tell the streamer to expect "real" (aka not test) packets.
        fsp_streamer->start_stream(false, camera_ptr->packet_delay);

        // Have the camera turn on streaming.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_LIVE);
        pending_control_write = GIGE_STREAM_LIVE;
        busy_retry_counter = 0;
    }

    // Restart our timer to verify we're getting packets.
    if (frame_timer != nullptr)
    {
        reset_timeout_counter();
        frame_timer->start(FRAME_ARRIVAL_TIMEOUT);
    }
}

/*!
 * @brief Playback a recorded stream from the connected camera, using
 *        the parameters specified in the info_ptr argument.
 * @param info_ptr    Parameters describing how the playback should be configured.
 */
void CameraStream::state_playback_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr)
{
    uint32_t        playback_rate;
    STREAM_PARAMS   *playback_info = &info_ptr->params;

    // Sanity checks.
    if ((mdi_parent == nullptr) || (MainWindow::GetInstance().isNull()))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No MDI parent or main window.";
        return;
    }

    // Stop our frame timer.
    if (frame_timer != nullptr)
        frame_timer->stop();

    // Tell the streamer to expect "real" (aka not test) packets.
    fsp_streamer->start_stream(false, mdi_parent->get_camera_info()->packet_delay);

    // Are we the active camera? If we are, then we want to use the playback rate
    // specified by the user. For background cameras, we use the background refresh rate.
    if (MainWindow::GetInstance()->active_mdi_child() == mdi_parent)
        playback_rate = playback_info->frame_delay;
    else
        playback_rate = BACKGROUND_STREAM_RATE;

    // Update the common registers.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, playback_rate);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FORMAT, GIGE_STREAM_FORMAT_JPEG);

    // Write out our registers. Are we attaching to the stream?
    if (playback_info->stream_control == GIGE_STREAM_SLAVE_ENABLE)
    {
        // No need to set up the other parameters. Hook into the existing stream.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_SLAVE_ENABLE);
        pending_control_write = GIGE_STREAM_SLAVE_ENABLE;
        busy_retry_counter = 0;
    }
    else
    {
        // Update all the registers required to start a playback.
        mdi_parent->write_camera_register(GIGE_REG_PB_FRAME_PERIOD, playback_info->frame_period);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_START_NUM, playback_info->clip_start);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_END_NUM, playback_info->clip_end);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_SEQ_START, playback_info->cur_frame);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_INTERVAL, playback_info->skip_interval);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_RECORDED);
        pending_control_write = GIGE_STREAM_RECORDED;
        busy_retry_counter = 0;
    }
}

/*!
 * @brief Enter the "paused" streaming stated.
 */
void CameraStream::state_pause_enter()
{
    // Are we not already paused?
    if (current_state != STATE_PAUSE)
    {
        // Go pause the stream.
        stream_pause();
    }
}

/*!
 * @brief Start a streaming download from the camera to save the per-frame metadata
 *        to the local workstation. Use the settings in the info_ptr argument.
 *
 * @param info_ptr    Parameters describing what portion of the recording's per-frame
 *                    metadata should be saved.
 */
void CameraStream::state_save_metadata_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr)
{
    uint32_t        reg_value;
    STREAM_PARAMS   *stream_info = &info_ptr->params;

    // Stop our frame timer.
    if (frame_timer != nullptr)
        frame_timer->stop();

    // Tell the streamer to expect "real" (aka not test) packets.
    fsp_streamer->start_stream(false, mdi_parent->get_camera_info()->packet_delay);

    // Tell the camera what packet size we want to use. Also, we don't want
    // to fragment the packets.
    reg_value = (uint32_t)stream_packet_len;
    reg_value |= GIGE_SCPSX_DNF_BE;
    mdi_parent->write_camera_register(GIGE_REG_SCPS0, reg_value);

    // Write out our registers.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_START_NUM, stream_info->clip_start);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_END_NUM, stream_info->clip_end);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_SEQ_START, stream_info->cur_frame);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_INTERVAL, stream_info->skip_interval);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, stream_info->frame_delay);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_RAW_METADATA);
    pending_control_write = GIGE_STREAM_RAW_METADATA;
    busy_retry_counter = 0;

    // Restart our frame timeout timer.
    if (frame_timer != nullptr)
    {
        reset_timeout_counter();
        frame_timer->start(2 * FRAME_ARRIVAL_TIMEOUT);
    }
}

/*!
 * @brief Exit the save metadata state. We'll cleanup our entry in the stream queue,
 *        and then exit the streamer state.
 *
 * @param new_state     State we're transitioning to.
 * @return Result of the state_exit() call.
 */
bool CameraStream::state_save_metadata_exit(int new_state)
{
    FI_comm     *comm_ptr;

    // Stop our frame timer.
    if (frame_timer != nullptr)
        frame_timer->stop();

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return false;
    }

    // Tell the communciations object we're done.
    comm_ptr->stream_finish(this);

    // Exit the state.
    return state_exit(new_state);
}

/*!
 * @brief Setup the camera to start streaming the session data.
 */
void CameraStream::state_save_session_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr)
{
    STREAM_PARAMS   *stream_info = &info_ptr->params;

    // Tell the streamer to expect "real" (aka not test) packets.
    fsp_streamer->start_stream(false, mdi_parent->get_camera_info()->packet_delay);

    // Write out our registers.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_START_NUM, stream_info->clip_start);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_END_NUM, stream_info->clip_end);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, 0);

    // Reset the retry register.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_RETRY, 0);

    // Have the camera start streaming the data.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_SESSION);
    pending_control_write = GIGE_STREAM_SESSION;
    busy_retry_counter = 0;

    // Restart our frame timeout timer.
    if (frame_timer != nullptr)
    {
        reset_timeout_counter();
        frame_timer->start(COPY_ARRIVAL_TIMEOUT);
    }
}


/*!
 * @brief Start a streaming download from the camera to save the per-frame metadata
 *        to the local workstation. Use the settings in the info_ptr argument.
 *
 * @param info_ptr    Parameters describing what portion of the recording's per-frame
 *                    metadata should be saved.
 */
void CameraStream::state_save_video_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr)
{
    STREAM_PARAMS   *stream_info = &info_ptr->params;

    // Tell the streamer to expect "real" (aka not test) packets.
    fsp_streamer->start_stream(false, mdi_parent->get_camera_info()->packet_delay);

    // Write out our registers.
    mdi_parent->write_camera_register(GIGE_REG_STREAM_START_NUM, stream_info->clip_start);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_END_NUM, stream_info->clip_end);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_SEQ_START, stream_info->cur_frame);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_INTERVAL, stream_info->skip_interval);
    mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, stream_info->frame_delay);

    // In this case are we streaming a container format?  If we are then the stream we want
    // to pull will be JPEG images (at least right now).  NOTE_JLC - Come back here for Planar
    // formats.  Do we have to put information in the info_ptr->params to allow for specification
    // of the image type to stream for the container?
    if ((stream_info->save_format == GIGE_STREAM_FORMAT_AVI) ||
        (stream_info->save_format == GIGE_STREAM_FORMAT_MP4))
    {
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FORMAT, GIGE_STREAM_FORMAT_JPEG);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Writing to GIGE_REG_STREAM_FORMAT FOR AVI OR MP4: "
                     << FiSaver::getGIGEStreamStr(GIGE_STREAM_FORMAT_JPEG);
    }
    else
    {
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FORMAT, stream_info->save_format);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Writing to GIGE_REG_STREAM_FORMAT: "
                     << FiSaver::getGIGEStreamStr(stream_info->save_format);
    }

    mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_DOWNLOAD);
    pending_control_write = GIGE_STREAM_DOWNLOAD;
    busy_retry_counter = 0;

    // Restart our frame timeout timer.
    if (frame_timer != nullptr)
    {
        reset_timeout_counter();
        frame_timer->start(FRAME_ARRIVAL_TIMEOUT);
    }
}

/*!
 * @brief Request a single frame from the camera.
 *
 * @param frame_no  The absolute frame number of the frame we want to pull from the camera.
 */
void CameraStream::state_single_frame_enter(uint32_t frame_no)
{
    // Sanity checks.
    if ((mdi_parent == nullptr) || (frame_timer == nullptr))
        return;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "current_state:" << current_state << "frame_no:" << frame_no;

    // Is our streamer running?
    if (current_state == STATE_SINGLE_FRAME)
    {
        // Just need to tell the camera the new frame. In single step mode, it doesn't
        // check against start/end values. Also, while we're in single step mode, the
        // camera is just looking for changes to the sequence start register.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_SEQ_START, frame_no);

        // Kill the stepping timer. We're going to restart it later.
        killTimer(step_timer_id);
    }

    // We had better be in the paused or single frame state.
    else if (current_state != STATE_PAUSE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Cannot single step from the current state:" << current_state;
    }
    else
    {
        uint32_t    reg_value;
        const FI_COMM::FI_CAMERA_INFO *camera_ptr;

        // Don't bother if the camera is off-line.
        if ((mdi_parent == nullptr) || (mdi_parent && mdi_parent->is_offline()))
            return;

        // Get our camera information object.
        camera_ptr = mdi_parent->get_camera_info();
        if (camera_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "No FI_CAMERA_INFO pointer!";
            return;
        }

        // Tell the streamer to expect "real" (aka not test) packets.
        fsp_streamer->start_stream(false, mdi_parent->get_camera_info()->packet_delay);

        // Tell the camera what size we've settled on.
        reg_value = (uint32_t)stream_packet_len;
        reg_value |= GIGE_SCPSX_DNF_BE;

        // Set the packet length value.
        mdi_parent->write_camera_register(GIGE_REG_SCPS0, reg_value);

        // Tell the camera about our inter-packet delay.
        mdi_parent->write_camera_register(GIGE_REG_SCPD0, camera_ptr->packet_delay);

        // We don't need a frame delay when single stepping.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, 0);

        // Tell the camera the frame we're interested in.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_START_NUM, frame_no);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_END_NUM, frame_no);
        mdi_parent->write_camera_register(GIGE_REG_STREAM_SEQ_START, frame_no);

        // There's no skipping when single stepping.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_INTERVAL, 1);

        // Make sure the format is JPEG.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_FORMAT, GIGE_STREAM_FORMAT_JPEG);

        // Have the camera step.
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_SINGLESTEP);
        pending_control_write = GIGE_STREAM_SINGLESTEP;
        busy_retry_counter = 0;
    }

    // Start up our stepping timer.
    step_timer_id = startTimer(STEP_STREAM_PERIOD);

    // Restart our timer to verify we're getting packets.
    if (frame_timer != nullptr)
    {
        reset_timeout_counter();
        frame_timer->start(FRAME_ARRIVAL_TIMEOUT);
    }
}

/*!
 * @brief Exit from the single-frame state. We have a special exit routine,
 *        because we have to make sure to stop our timer.
 *
 * @param new_state     The new state we want to go to.
 * @return True to wait on state transition, false to carry on. Actually, this will
 *         be the return value from the state_exit() method.
 */
bool CameraStream::state_single_frame_exit(int new_state)
{
    // Make sure our stepping timer is turned off.
    if (step_timer_id != 0)
    {
        killTimer(step_timer_id);
        step_timer_id = 0;
    }

    // Perform normal exit routine.
    return state_exit(new_state);
}

/*!
 * @brief Stop the streamer.
 */
void CameraStream::state_stop()
{
    // Are we running?
    if (!is_paused())
    {
        // Pause the stream before stopping.
        stream_pause();
    }

    // Close the streaming port.
    if (mdi_parent != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Tell camera to stop streaming";
        mdi_parent->write_camera_register(GIGE_REG_SCP0, 0);
    }

    // Are we in error recovery mode?
    if (error_recovery)
    {
        // Re-initialize the stream.
        if (mdi_parent != nullptr)
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Re-initialize the stream";
        }
        QApplication::postEvent(this, new StreamerStateEvent(STATE_INIT));

        // Let folks know we've faulted.
        emit fault();
    }
}

/*!
 * @brief The streaming interface is now streaming live video, so wrap
 *        up any pending transitions.
 */
void CameraStream::stream_live_finished()
{
    // Let folks know we're streaming.
    emit streaming();

    // Do we have a pending state transition?
    if (pending_transition)
    {
        // Queue up the next state transition.
        QApplication::postEvent(this, new StreamerStateEvent(next_state, next_info));
    }
}

/*!
 * @brief Pause the streaming interface.
 */
void CameraStream::stream_pause()
{
    // Always Kill our timer 'cuz we aren't getting anymore frames for now.
    if (frame_timer != nullptr)
        frame_timer->stop();

    // Are we running?
    if (is_paused())
    {
        // We're done!
        stream_pause_finished();
        return;
    }

    // "Pausing" is really telling the camera to stop streaming.
    if (mdi_parent != nullptr)
    {
        mdi_parent->write_camera_register(GIGE_REG_STREAM_CONTROL, GIGE_STREAM_STOP);
        pending_control_write = GIGE_STREAM_STOP;
        busy_retry_counter = 0;
    }

    // Tell the streamer to stop.
    if (fsp_streamer != nullptr)
    {
        fsp_streamer->stop_stream();
    }
}

/*!
 * @brief The streaming interface has been paused, so wrap up any pending transition.
 */
void CameraStream::stream_pause_finished()
{
    // Let folks know we're not streaming anymore. Except in the case when we're
    // stopping the stream (STATE_IDLE). Then we don't want to tell anyone.
    if (current_state != STATE_IDLE)
        emit paused();

    // Do we have a pending state transition?
    if (pending_transition)
    {
        // Queue up the next state transition.
        QApplication::postEvent(this, new StreamerStateEvent(next_state, next_info));
    }

    // Are we in error recovery mode?
    else if (error_recovery)
    {
        // We want to automatically transition to a new state. Set things up.
        pending_transition = true;
        next_state = recovery_state;
        next_info = nullptr;

        // Clear our flag.
        error_recovery = false;
    }
}

/*!
 * @brief Our testing has finished. We've successfully negotiated as big a pipe as we're going to get.
 * @param got_timeout   True if we timed out waiting for the last packet.
 */
void CameraStream::test_complete(bool got_timeout)
{
    // Sanity checks.
    if ((fsp_streamer == nullptr) || (mdi_parent == nullptr))
        return;

    // Did we get a timeout event? If we did, we need to back off by our increment amount.
    if (got_timeout)
        stream_packet_len -= STREAM_PACKET_INCR;

    // Tell the streamer that we're not expecting anything right now.
    fsp_streamer->stop_stream();

    // Bump our state to indicate we're done.
    QApplication::postEvent(this, new StreamerStateEvent(STATE_PAUSE));

    QLOG_INFO() << __PRETTY_FUNCTION__ << "Successfully negotiated with" << mdi_parent->get_ip_addr_pretty() << "PktLen:" << stream_packet_len << "bytes";

    // Let folks know we're ready to stream.
    emit ready();
}

/*!
 * @brief This method tests preconditions for the various state transitions.
 * @param test_state    The state we want to test the preconditions for.
 * @return True if it would be OK to transition into the test_state, false otherwise.
 */
bool CameraStream::transition_ok(int test_state)
{
    bool    ok_to_go(true);

    // Process based on the new state.
    switch (test_state)
    {
    case STATE_COPY_FILES:
    case STATE_PLAYBACK:
    case STATE_SAVE_METADATA:
    case STATE_SAVE_SESSION:
    case STATE_SAVE_VIDEO:
    case STATE_SINGLE_FRAME:
        // We must be in either a PAUSED state or the test_state state.
        if ((current_state != STATE_PAUSE) && (current_state != test_state))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "current_state:" << current_state;
            ok_to_go = false;
        }

        // If we have a pending transition, it must be to either the PAUSED or test_state state.
        else if (pending_transition && ((next_state != STATE_PAUSE) && (next_state != test_state)))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "pending_transition && next_state:" << next_state << "current_state:" << current_state;
            ok_to_go = false;
        }
        break;

    default:
    case STATE_LINK_NEGOTIATE:
    case STATE_LIVE:
    case STATE_PAUSE:
    case STATE_IDLE:
    case STATE_INIT:
        // We can always transition to these states.
        break;
    }

    // Return our status.
    return ok_to_go;
}

/*!
 * @brief Used to let the streamer know when things change with the MDI parent.
 */
void CameraStream::update(bool set_active)
{
    uint32_t    reg_value;

    // Don't bother if the camera is off-line.
    if ((mdi_parent == nullptr) || (mdi_parent && mdi_parent->is_offline()))
        return;

    // We only care about changes in the parent for certain states.
    switch (current_state)
    {
    case STATE_LIVE:
    case STATE_PLAYBACK:
        // What is the current frame delay?
        reg_value = mdi_parent->read_camera_cache(GIGE_REG_STREAM_FRAME_DELAY);

        // Tell the camera what framerate we want. If we're not the active child, we want to run in the background.
        if (set_active)
        {
            // Is there a change?
            if (reg_value != stream_frame_delay)
            {
                // Pause the stream first.
                mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, stream_frame_delay);
            }
        }

        // Is there a change?
        else if (reg_value != BACKGROUND_STREAM_RATE)
        {
            // Slowing down, we can simply write the register value.
            mdi_parent->write_camera_register(GIGE_REG_STREAM_FRAME_DELAY, BACKGROUND_STREAM_RATE);
        }
        break;
    default:
        break;
    }
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief The receive method for the streamer state events.
 * @param event     Pointer to the event structure.
 */
void CameraStream::customEvent(QEvent *event)
{
    // Only respond to our custom event
    if (event->type() == STREAMER_STATE_EVENT)
    {
        // Run the machine.
        machine(event);
    }
}

/*!
 * @brief Event routine which is invoked when one of our timers expires.
 * @param event     Pointer to the event class.
 */
void CameraStream::timerEvent(QTimerEvent *event)
{
    // Did the test packet timeout?
    if (event->timerId() == test_pkt_timer_id)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Test packet timeout." << stream_packet_len << "bytes must be too long.";

        // Kill our timer.
        killTimer(test_pkt_timer_id);
        test_pkt_timer_id = 0;

        // Our testing is done.
        test_complete(true);
    }
    else if (event->timerId() == step_timer_id)
    {
        // Kill our timer.
        killTimer(step_timer_id);
        step_timer_id = 0;

        // Pause the stream.
        pause_streaming();
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Timed out waiting for an image frame from the camera.
 */
void CameraStream::frame_timeout()
{
#if LOG_TIMEOUT_FRAMES
    QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "stream_state:" << current_state;
#endif

    // Kill the timer.
    if (frame_timer != nullptr)
        frame_timer->stop();

    // If this is our first timeout_counter set the images_received count.
    if (timeout_counter == 0)
    {
        timeout_frames_received = frames_received;
    }

    // Increment our counter.
    ++timeout_counter;

    // If our timeout_counter is coming around again have we received any frames?
    if (timeout_counter > 1)
    {
        // Each image received signal increments the frames_received.
        if (frames_received > timeout_frames_received)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "TIMEOUT COUNT RESET DUE TO FSP SOCKET RECEIPT."
                         << "frames_received: " << frames_received
                         << "timeout_frames_received: " << timeout_frames_received;

            // Reset the timeout counter as FSP Socket has received frames since we started.
            reset_timeout_counter();
        }
    }

    // Error recovery depends on the state.
    switch (current_state)
    {
    case STATE_PAUSE:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "We are in STATE_PAUSE but frame_timer is running!";
        reset_timeout_counter();
        break;
    case STATE_COPY_FILES:
        copy_files_timeout();
        break;
    case STATE_LIVE:
        live_frame_timeout();
        break;
    case STATE_SAVE_VIDEO:
        save_video_timeout();
        break;
    case STATE_SAVE_METADATA:
        save_metadata_timeout();
        break;
    case STATE_SAVE_SESSION:
    default:
        // Default behavior is to pass the buck.
        emit timeout();
        break;
    }
}

/*!
 * @brief The streamer has indicated that it's received a frame.
 */
void CameraStream::fsp_image_received()
{
    bool    restart_timer = false;

    // Do we have a timer?
    if (frame_timer != nullptr)
    {
        // Stop the timer.
        restart_timer = frame_timer->isActive();
        frame_timer->stop();
    }

    // We have been notified of our socket receiving a frame so
    // increment our internal frame received accumulator.
    frames_received++;

    // Let other's know that we've received an image.
    emit image_received();

    // Check to see if we're interested in restarting the timer.
    switch (current_state)
    {
    case STATE_LIVE:
    case STATE_PLAYBACK:
    case STATE_SAVE_METADATA:
    case STATE_SAVE_SESSION:
    case STATE_SAVE_VIDEO:
    case STATE_COPY_FILES:
        // Do we want to restart the timer?
        if ((frame_timer != nullptr) && restart_timer)
        {
            // Restart the timer.
            reset_timeout_counter();
            frame_timer->start();
        }
        break;
    default:
        break;
    }
}

/*!
 * @brief Received a signal that the streamer object for the camera at "ip_addr" is ready.
 * @param ip_addr   IP Address of the camera that is ready to begin streaming operations.
 */
void CameraStream::fsp_socket_is_ready(uint32_t ip_addr)
{
    // Is this event for us?
    if (ip_addr != mdi_parent->get_ip_addr())
        return;

    // Make sure the streamer is initialized.
    if (!fsp_streamer->is_initialized())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "fsp_streamer is not initialized";
        return;
    }

    // Startup our streaming machine.
    QApplication::postEvent(this, new StreamerStateEvent(STATE_INIT));
}

/*!
 * @brief Received a test packet from the camera.
 */
void CameraStream::fsp_test_packet_received()
{
    STREAM_EVENT_PARAMS *param_ptr;

    // We got the test packet. Kill the timer.
    killTimer(test_pkt_timer_id);
    test_pkt_timer_id = 0;

    // Bump the size and go again.
    param_ptr = new STREAM_EVENT_PARAMS;
    if (param_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate STREAM_EVENT_PARAMS structure!";
        return;
    }
    param_ptr->test_pkt_len = stream_packet_len + (int)STREAM_PACKET_INCR;
    QApplication::postEvent(this, new StreamerStateEvent(STATE_LINK_NEGOTIATE, param_ptr));
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param reg_addr      The address of the register which was read.
 */
void CameraStream::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    // Is this event for us?
    if (ip_addr != mdi_parent->get_ip_addr())
        return;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_STREAM_CONTROL:
        // We really only care about our write failures.
        if (!read_op)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Timeout for GIGE_REG_STREAM_CONTROL";

            // Invoke our error handler.
            error_handler();
        }
        break;
    default:
        break;
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void CameraStream::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Is this event for us?
    if (ip_addr != mdi_parent->get_ip_addr())
        return;

    // Ignore if we're in shutdown.
    if (mdi_parent->shutdown_in_progress())
        return;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_STREAM_CONTROL:
        process_stream_control();
        break;
    default:
        break;
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void CameraStream::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;
    QString     err_msg;

    // Is this event for us?
    if (ip_addr != mdi_parent->get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Only deal with registers we're interested in.
    switch (reg_addr)
    {
    case GIGE_REG_STREAM_FILEPATH:
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Tell the stream state machine to start the copy.
            QApplication::postEvent(this, new StreamerStateEvent(STATE_COPY_FILES));
        }
        else
        {
            err_msg = QString::fromUtf8("Unable to write GIGE_REG_STREAM_FILEPATH to camera=%1, status=0x%2").arg(mdi_parent->get_ip_addr_pretty()).arg(write_status, 0, 16);
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }
        break;
    case GIGE_REG_SCP0:
        // At this point, it doesn't matter if the write was successful or not,
        // if we're shutting down, then we're detached.
        if ((mdi_parent && mdi_parent->shutdown_in_progress()) && detach_pending)
        {
            // No longer pending - we're detached.
            detach_pending = false;

            // Let folks know that we're detached.
            emit detached();
        }
        break;
    case GIGE_REG_SCPD0:
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Continue with our processing.
            process_scpd0();
        }
        else
        {
            err_msg = QString::fromUtf8("Unable to write GIGE_REG_SCPD0 to camera=%1, status=0x%2").arg(mdi_parent->get_ip_addr_pretty()).arg(write_status, 0, 16);
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }
        break;
    case GIGE_REG_STREAM_CONTROL:
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        QLOG_TRACE() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "pending_control_write:" << Qt::showbase << Qt::hex << pending_control_write << "write_status:" << write_status;

        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Make sure we're not in shutdown.
            if (mdi_parent && !mdi_parent->shutdown_in_progress())
            {
                // Read the register back from the camera so we know exactly what the camera thinks is going on.
                comm_ptr->expire_entry(mdi_parent->get_ip_addr(), GIGE_REG_STREAM_CONTROL);
                comm_ptr->start_read_reg(mdi_parent->get_ip_addr(), GIGE_REG_STREAM_CONTROL);
            }
        }

        // If the status is BUSY, then try to recover.
        else if (write_status == GEV_STATUS_BUSY)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Camera is busy, current_state:" << current_state;

            // We got a busy response from the camera. Retry setting the register.
            QTimer::singleShot(RETRY_PAUSE_PERIOD, this, &CameraStream::retry_stream_control);
        }

        // Is the camera reporting an incorrect configuration?
        else if (write_status == GEV_STATUS_WRONG_CONFIG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << mdi_parent->get_ip_addr_pretty() << "Camera reports 'Wrong Config', current_state:" << current_state;
        }
        else
        {
            // Log the issue.
            err_msg = QString::fromUtf8("Unable to write GIGE_REG_STREAM_CONTROL to camera=%1, status=0x%2").arg(mdi_parent->get_ip_addr_pretty()).arg(write_status, 0, 16);
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;

            // Make sure we're not in shutdown.
            if (mdi_parent && !mdi_parent->shutdown_in_progress())
            {
                // Handle the error.
                error_handler();
            }
        }
        break;
    case GIGE_REG_STREAM_FRAME_DELAY:
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status != GEV_STATUS_SUCCESS)
        {
            // Log the issue.
            err_msg = QString::fromUtf8("Unable to write GIGE_REG_STREAM_FRAME_DELAY to camera=%1, status=0x%2").arg(mdi_parent->get_ip_addr_pretty()).arg(write_status, 0, 16);
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }
        else
        {
            // Make sure we're not in shutdown.
            if (mdi_parent && !mdi_parent->shutdown_in_progress())
            {
                // Read the register back from the camera so we know exactly what the camera thinks is going on.
                comm_ptr->start_read_reg(mdi_parent->get_ip_addr(), GIGE_REG_STREAM_FRAME_DELAY);
            }
        }
        break;
    default:
        break;
    }
}
