/*!
 * @file camera_stream.h
 * @brief This is the header file for the CameraStream class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_STREAM_H
#define CAMERA_STREAM_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <QObject>

#include "fsp_socket.h"

namespace CAMERA_STREAM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define a structure to describe how we want a stream operation to be performed.
typedef struct STREAM_PARAMS
{
    uint32_t    clip_start;
    uint32_t    clip_end;
    uint32_t    cur_frame;
    uint32_t    skip_interval;
    uint32_t    desired_fps;
    uint32_t    frame_delay;
    uint32_t    frame_period;
    uint32_t    save_format;
    uint32_t    stream_control;
} STREAM_PARAMS;

//! Define a set of type values describing the various "image" buffer types.
enum
{
    STREAM_COPY_IMAGE,
    STREAM_LIVE_IMAGE,
    STREAM_RAW_METADATA,
    STREAM_SAVE_IMAGE,
    STREAM_SAVE_SESSION,
    STREAM_UNKNOWN
};

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
union STREAM_EVENT_PARAMS;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class QTimer;
class CameraMDI;

class CameraStream : public QObject
{
    Q_OBJECT

public:
    explicit CameraStream(QObject *parent = nullptr);
    ~CameraStream();

    void    attach();
    void    copy_from_camera(QString& camera_path);
    void    detach();
    void    get_raw_metadata(CAMERA_STREAM::STREAM_PARAMS& stream_info, bool retry);
    void    go_live();
    void    pause_streaming();
    void    playback(const CAMERA_STREAM::STREAM_PARAMS& stream_info);
    void    request_frame(uint32_t frame_index);
    void    retry(uint32_t retry_index);
    void    save_session(const CAMERA_STREAM::STREAM_PARAMS& stream_info);
    void    save_video(CAMERA_STREAM::STREAM_PARAMS& stream_info);
    void    set_interframe_delay(uint32_t new_delay);
    void    update(bool set_active);

    int     image_count();
    void    image_release(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr);
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> image_take_first();
    int     image_type();

    bool    is_active();
    bool    is_busy()                           { return is_copying() || is_playing() || is_stepping(); }
    bool    is_copying();
    bool    is_paused();
    bool    is_playing();
    bool    is_stepping();

    int     packet_len()                          { return stream_packet_len; }

signals:
    void    detached();
    void    fault();
    void    image_received();
    void    paused();
    void    ready();
    void    streaming();
    void    timeout();

public slots:

protected:
    virtual void    customEvent(QEvent *event);
    virtual void    timerEvent(QTimerEvent *event);

private slots:
    void    frame_timeout();
    void    fsp_image_received();
    void    fsp_socket_is_ready(uint32_t ip_addr);
    void    fsp_test_packet_received();
    void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    void    retry_stream_control();
    void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private:
    void    copy_files_timeout();
    void    error_handler();
    void    live_frame_timeout();
    void    machine(QEvent *event);
    void    machine_state_enter(int new_state, CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr);
    bool    machine_state_exit(int new_state);
    void    process_scpd0();
    void    process_stream_control();
    void    reset_timeout_counter();
    void    save_metadata_timeout();
    void    save_video_timeout();

    void    state_copy_files_enter();
    bool    state_exit(int new_state);
    void    state_init_enter();
    void    state_link_negotiate_enter(int pkt_len);
    void    state_live_enter();
    void    state_pause_enter();
    void    state_playback_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr);
    void    state_save_metadata_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr);
    bool    state_save_metadata_exit(int new_state);
    void    state_save_session_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr);
    void    state_save_video_enter(CAMERA_STREAM::STREAM_EVENT_PARAMS *info_ptr);
    void    state_single_frame_enter(uint32_t frame_no);
    bool    state_single_frame_exit(int new_state);
    void    state_stop();

    void    stream_live_finished();
    void    stream_pause();
    void    stream_pause_finished();

    void    stream_copy_files();
    void    stream_playback(void *params_ptr);
    void    stream_save_exit();
    void    stream_save_metadata(void *params_ptr);
    void    stream_save_start(void *params_ptr);
    void    stream_setup();
    void    stream_single_frame(uint32_t frame_no);
    void    stream_start_live();

    void    test_complete(bool got_timeout);
    bool    transition_ok(int test_state);

    QString stateToString(int stateValue);


private:
    CameraMDI       *mdi_parent;                    //!< Handy pointer to our parent object.
    FSPSocket      *fsp_streamer;                 //!< Pointer to our FCP streamer object.

    int             current_state;                  //!< The current state of the streamer.
    bool            detach_pending;                 //!< True if we're in process of detaching from the camera.
    uint32_t        pending_control_write;          //!< The stream control register value we're trying to write.
    bool            pending_transition;             //!< True if there is a pending state transition.
    uint32_t        stream_frame_delay;             //!< The delay between frames, in milliseconds.
    int             stream_packet_len;              //!< The number of bytes per stream packet that we've negotiated with the camera.

    uint32_t        frames_received;                //!< Accumulates a count of frames_received for a state.  Reset each state change.

    QTimer          *frame_timer;                   //!< A timer to catch missing frames.
    int             step_timer_id;                  //!< Timer used when the user is single stepping through a captured video.
    int             test_pkt_timer_id;              //!< Timer used to negotiate stream packet sizes.
    int             timeout_counter;                //!< We don't want to declare a failure immediately.
    uint32_t        timeout_frames_received;        //!< At first frame timeout - what was the frames_received accumulation.
    int             busy_retry_counter;             //!< Number of times we've retried on a BUSY return;

    bool            error_recovery;                 //!< True if we're in an error recovery mode, false otherwise.
    int             recovery_state;                 //!< If error_recovery is true, this is the state we want to recover to.

    int             next_state;                     //!< The pending state, when we're waiting on a transition event to complete.
    CAMERA_STREAM::STREAM_EVENT_PARAMS *next_info;  //!< The pending information associated with the "next_state";
};

#endif // CAMERA_STREAM_H
