#include "camera_telemetry.h"
#include "ui_camera_telemetry.h"

#include <QCloseEvent>
#include <QFile>
#include <QFileInfo>
#include <QFileDialog>
#include <QJsonDocument>
#include <QList>
#include <qmath.h>
#include <QMdiSubWindow>
#include <QMessageBox>
#include <QProcess>
#include <QSignalBlocker>
#include <QStandardItem>
#include <QtConcurrent>
#include <QTextStream>
#include <QThread>
#include <QTime>
#include <QTimerEvent>
#include <qwt.h>
#include <qwt_color_map.h>

#include "camera_mdi.h"
#include "logger.h"
#include "mainwindow.h"
#include "polldlg.h"

namespace CAM_TELEMETRY
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Our update frequency to gather sensor information.
const int UPDATE_INTERVAL           = 1;
const double SENSOR_TEMP_FAULT      = 80.0;

//! Our basic title.
const QString   WINDOW_TITLE_STR(CameraTelemetry::tr("Telemetry"));

enum COL_ID
{
    COL_TIME = 0,
    COL_TEMP_A10_INT,
    COL_TEMP_A10_EXT,
    COL_TEMP_SENSOR,
    COL_TEMP_NVME,
    COL_TEMP_FIREFLY,
    COL_TEMP_EM2120,
    COL_TEMP_FAN,
    COL_TEMP_PCI,
    COL_SENSOR_ALERT,
    COL_PMBUS_ALERT,
    COL_EM2120_CURR,
    COL_EM2120_VOLT,
    COL_A10_VEXT0,
    COL_A10_VEXT1,
    COL_A10_VCC,
    COL_A10_VCCP,
    COL_A10_VCCPT,
    COL_A10_VCCERAM,
    COL_A10_ADCGND,
    COL_FIREFLY_VOLT,
    COL_FAN_TACH,
    COL_FAN_PWM,
    COL_COUNT
};
}   // End of namespace CAM_TELEMETRY.

using namespace CAM_TELEMETRY;

CameraTelemetry::CameraTelemetry(QWidget *parent) :
    QDockWidget(parent),
    ui(new Ui::CameraTelemetry),
    active_camera(nullptr),
    log_data(nullptr)
{
    QwtLinearColorMap *thermo_map;
    QStandardItem *tbl_item;
    FI_comm     *comm_ptr;

    // Initialize our UI object.
    ui->setupUi(this);

    // We are going to use a Queued connection type later, and Qt needs to know about QList<QStandardItem *> types.
    qRegisterMetaType<QList<QStandardItem *>>("QList<QStandardItem *>");

    // Intialize our variables.
    degCelsius = true;
    logger = 0;
    json_data = nullptr;
    poll_period = UPDATE_INTERVAL;

    // Are we ready to play yet?
    if (MainWindow::GetInstance().isNull())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
        return;

    // Disable the "export" button until we get data to export.
    ui->pb_export->setEnabled(false);

    // Create our data model. Just the header with the
    // columns for the values.
    log_data = new QStandardItemModel(0, COL_COUNT, this);
    tbl_item = new QStandardItem(tr("Time"));
    log_data->setHorizontalHeaderItem(COL_TIME, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_a10_int->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_A10_INT, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_a10_ext->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_A10_EXT, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_sensor->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_SENSOR, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_nvme->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_NVME, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_firefly->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_FIREFLY, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_em2120->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_EM2120, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_fan->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_FAN, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_temp_pcie->text());
    log_data->setHorizontalHeaderItem(COL_TEMP_PCI, tbl_item);

    tbl_item = new QStandardItem(ui->rb_sensor_pwr_ok->text());
    log_data->setHorizontalHeaderItem(COL_SENSOR_ALERT, tbl_item);
    tbl_item = new QStandardItem(ui->rb_pmbus_alert->text());
    log_data->setHorizontalHeaderItem(COL_PMBUS_ALERT, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_em2120_I->text());
    log_data->setHorizontalHeaderItem(COL_EM2120_CURR, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_em2120_V->text());
    log_data->setHorizontalHeaderItem(COL_EM2120_VOLT, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_adc_vext0->text());
    log_data->setHorizontalHeaderItem(COL_A10_VEXT0, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_adc_vext1->text());
    log_data->setHorizontalHeaderItem(COL_A10_VEXT1, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_adc_vcc->text());
    log_data->setHorizontalHeaderItem(COL_A10_VCC, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_adc_vccp->text());
    log_data->setHorizontalHeaderItem(COL_A10_VCCP, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_adc_vccpt->text());
    log_data->setHorizontalHeaderItem(COL_A10_VCCPT, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_adc_vcceram->text());
    log_data->setHorizontalHeaderItem(COL_A10_VCCERAM, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_adc_gnd->text());
    log_data->setHorizontalHeaderItem(COL_A10_ADCGND, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_firefly->text());
    log_data->setHorizontalHeaderItem(COL_FIREFLY_VOLT, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_fan_tach->text());
    log_data->setHorizontalHeaderItem(COL_FAN_TACH, tbl_item);
    tbl_item = new QStandardItem(ui->lbl_fan_pwm->text());
    log_data->setHorizontalHeaderItem(COL_FAN_PWM, tbl_item);

    // Set the model for the views to be our camera data.
    ui->log_table->setModel(log_data);

    // Resize the columns in our table view.
    ui->log_table->resizeRowsToContents();
    ui->log_table->resizeColumnsToContents();
    ui->log_table->horizontalHeader()->setStretchLastSection(false);

    // Set up our thermometers.
    ui->temp_a10_int->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_a10_int->setColorMap(thermo_map);
    ui->temp_a10_ext->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_a10_ext->setColorMap(thermo_map);
    ui->temp_sensor->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_sensor->setColorMap(thermo_map);
    ui->temp_nvme->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_nvme->setColorMap(thermo_map);
    ui->temp_firefly->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_firefly->setColorMap(thermo_map);
    ui->temp_em2120->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_em2120->setColorMap(thermo_map);
    ui->temp_fan_external->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_fan_external->setColorMap(thermo_map);
    ui->temp_pcie->setScale(20.0, 80.0);
    thermo_map = new QwtLinearColorMap(Qt::green, Qt::red);
    thermo_map->addColorStop(0.55, Qt::yellow);
    thermo_map->addColorStop(0.6, QColor(255, 165, 0));     // orange
    ui->temp_pcie->setColorMap(thermo_map);

    // Connect our signals/slots.
    connect(ui->pb_temp_scale, &QPushButton::clicked, this, &CameraTelemetry::changeTempScale);
    connect(ui->pb_start_stop, &QPushButton::clicked, this, &CameraTelemetry::logStartStop);
    connect(ui->pb_export, &QPushButton::clicked, this, &CameraTelemetry::exportLog);
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &CameraTelemetry::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraTelemetry::write_complete, Qt::QueuedConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::active_child_changed, this, &CameraTelemetry::new_active_camera);
    connect(this, &CameraTelemetry::parse_done, this, &CameraTelemetry::parse_complete);
}

CameraTelemetry::~CameraTelemetry()
{
    // Stop our timer if it's still active.
    if (logger)
    {
        killTimer(logger);
        logger = 0;
    }

    delete ui;
}

void CameraTelemetry::changeEvent(QEvent *e)
{
    QDockWidget::changeEvent(e);
    switch (e->type())
    {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}

void CameraTelemetry::changeTempScale()
{
    double old_temp;

    if (degCelsius)
    {
        // Change to Fahrenheit.
        ui->pb_temp_scale->setText("°F");
        degCelsius = false;

        // Switch the temperature displays to be in Fahrenheit.
        old_temp = ui->temp_a10_int->value();
        ui->temp_a10_int->setLowerBound(c_to_f(ui->temp_a10_int->minimum()));
        ui->temp_a10_int->setUpperBound(c_to_f(ui->temp_a10_int->maximum()));
        ui->temp_a10_int->setValue(c_to_f(old_temp));
        old_temp = ui->temp_a10_ext->value();
        ui->temp_a10_ext->setLowerBound(c_to_f(ui->temp_a10_ext->minimum()));
        ui->temp_a10_ext->setUpperBound(c_to_f(ui->temp_a10_ext->maximum()));
        ui->temp_a10_ext->setValue(c_to_f(old_temp));
        old_temp = ui->temp_sensor->value();
        ui->temp_sensor->setLowerBound(c_to_f(ui->temp_sensor->minimum()));
        ui->temp_sensor->setUpperBound(c_to_f(ui->temp_sensor->maximum()));
        ui->temp_sensor->setValue(c_to_f(old_temp));
        old_temp = ui->temp_nvme->value();
        ui->temp_nvme->setLowerBound(c_to_f(ui->temp_nvme->minimum()));
        ui->temp_nvme->setUpperBound(c_to_f(ui->temp_nvme->maximum()));
        ui->temp_nvme->setValue(c_to_f(old_temp));
        old_temp = ui->temp_firefly->value();
        ui->temp_firefly->setLowerBound(c_to_f(ui->temp_firefly->minimum()));
        ui->temp_firefly->setUpperBound(c_to_f(ui->temp_firefly->maximum()));
        ui->temp_firefly->setValue(c_to_f(old_temp));
        old_temp = ui->temp_em2120->value();
        ui->temp_em2120->setLowerBound(c_to_f(ui->temp_em2120->minimum()));
        ui->temp_em2120->setUpperBound(c_to_f(ui->temp_em2120->maximum()));
        ui->temp_em2120->setValue(c_to_f(old_temp));
        old_temp = ui->temp_fan_external->value();
        ui->temp_fan_external->setLowerBound(c_to_f(ui->temp_fan_external->minimum()));
        ui->temp_fan_external->setUpperBound(c_to_f(ui->temp_fan_external->maximum()));
        ui->temp_fan_external->setValue(c_to_f(old_temp));
        old_temp = ui->temp_pcie->value();
        ui->temp_pcie->setLowerBound(c_to_f(ui->temp_pcie->minimum()));
        ui->temp_pcie->setUpperBound(c_to_f(ui->temp_pcie->maximum()));
        ui->temp_pcie->setValue(c_to_f(old_temp));
    }
    else
    {
        // Change to Celsius.
        ui->pb_temp_scale->setText("°C");
        degCelsius = true;

        // Switch the temperature displays to be in Fahrenheit.
        old_temp = ui->temp_a10_int->value();
        ui->temp_a10_int->setLowerBound(f_to_c(ui->temp_a10_int->minimum()));
        ui->temp_a10_int->setUpperBound(f_to_c(ui->temp_a10_int->maximum()));
        ui->temp_a10_int->setValue(f_to_c(old_temp));
        old_temp = ui->temp_a10_ext->value();
        ui->temp_a10_ext->setLowerBound(f_to_c(ui->temp_a10_ext->minimum()));
        ui->temp_a10_ext->setUpperBound(f_to_c(ui->temp_a10_ext->maximum()));
        ui->temp_a10_ext->setValue(f_to_c(old_temp));
        old_temp = ui->temp_sensor->value();
        ui->temp_sensor->setLowerBound(f_to_c(ui->temp_sensor->minimum()));
        ui->temp_sensor->setUpperBound(f_to_c(ui->temp_sensor->maximum()));
        ui->temp_sensor->setValue(f_to_c(old_temp));
        old_temp = ui->temp_nvme->value();
        ui->temp_nvme->setLowerBound(f_to_c(ui->temp_nvme->minimum()));
        ui->temp_nvme->setUpperBound(f_to_c(ui->temp_nvme->maximum()));
        ui->temp_nvme->setValue(f_to_c(old_temp));
        old_temp = ui->temp_firefly->value();
        ui->temp_firefly->setLowerBound(f_to_c(ui->temp_firefly->minimum()));
        ui->temp_firefly->setUpperBound(f_to_c(ui->temp_firefly->maximum()));
        ui->temp_firefly->setValue(f_to_c(old_temp));
        old_temp = ui->temp_em2120->value();
        ui->temp_em2120->setLowerBound(f_to_c(ui->temp_em2120->minimum()));
        ui->temp_em2120->setUpperBound(f_to_c(ui->temp_em2120->maximum()));
        ui->temp_em2120->setValue(f_to_c(old_temp));
        old_temp = ui->temp_fan_external->value();
        ui->temp_fan_external->setLowerBound(f_to_c(ui->temp_fan_external->minimum()));
        ui->temp_fan_external->setUpperBound(f_to_c(ui->temp_fan_external->maximum()));
        ui->temp_fan_external->setValue(f_to_c(old_temp));
        old_temp = ui->temp_pcie->value();
        ui->temp_pcie->setLowerBound(f_to_c(ui->temp_pcie->minimum()));
        ui->temp_pcie->setUpperBound(f_to_c(ui->temp_pcie->maximum()));
        ui->temp_pcie->setValue(f_to_c(old_temp));
    }
}

/*!
 * @brief We are closing the window, so make sure we stop our timer.
 *
 * @param event     Pointer to the event structure.
 */
void CameraTelemetry::closeEvent(QCloseEvent *event)
{
    // Uncheck our menu item.
    MainWindow::GetInstance()->close_telemetry();

    // Accept the event.
    event->accept();

    // Pass it on.
    QDockWidget::closeEvent(event);
}

void CameraTelemetry::logStartStop()
{
    // Are we logging (i.e. logger is non-zero).
    if (logger)
    {
        // Stop the timer.
        killTimer(logger);
        logger = 0;

        // Update our button text.
        ui->pb_start_stop->setText("Start Logging");

        // Add an null row to mark the transition.
        QTimer::singleShot(10, this, &CameraTelemetry::add_null_row);
        return;
    }

    // Do we have an active camera?
    if (active_camera == nullptr)
    {
        // Get the currently active camera.
        active_camera = MainWindow::GetInstance()->active_mdi_child();
        if (active_camera == nullptr)
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "No active camera!";
            return;
        }
        else if (active_camera->is_series3())
        {
            // Series3 cameras don't have telemetry.
            active_camera = nullptr;
            return;
        }
    }

    // Ask the user for the period between polls.
    PollDlg dlg(this);
    dlg.set_period(poll_period);
    if (dlg.exec() == QDialog::Rejected)
        return;

    // Start our timer.
    poll_period = dlg.get_period();
    logger = startTimer(poll_period * 1000);
    if (logger == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to start the timer?!";
        return;
    }

    // Update the button text.
    ui->pb_start_stop->setText("Stop Logging");

    // Don't wait, get a poll in right away.
    timerEvent(nullptr);
}

/*!
 * @brief A new camera (or possibly none) has become active.
 * @param window    The QMdiSubWindow pointer to the related camera window.
 */
void CameraTelemetry::new_active_camera(QMdiSubWindow *window)
{
    QString     title_str;
    CameraMDI   *new_camera;

    // Is there no new active camera?
    if (window == nullptr)
    {
        // Stop our update timer.
        if (logger > 0)
            logStartStop();

        // Reset our active camera.
        active_camera = nullptr;

        // Set our title.
        setWindowTitle(WINDOW_TITLE_STR);
        return;
    }

    // Get our new camera object.
    new_camera = qobject_cast<CameraMDI *>(window->widget());
    if (new_camera == nullptr)
        return;

    // Don't support series3 cameras.
    if (new_camera->is_series3())
        return;

    // Is it really new?
    if (new_camera != active_camera)
    {
        // Did we ever have an active camera?
        if (active_camera != nullptr)
        {
            // Stop our update timer.
            if (logger > 0)
                logStartStop();

            // Sanity check.
            QList<CameraMDI *>  child_mdi_list = MainWindow::GetInstance()->get_children();
            if (!child_mdi_list.contains(active_camera))
            {
                // The "active" camera window went away - crashed or something.
                active_camera = nullptr;
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Active camera has gone away. Resetting.";
            }
        }

        // Remember our active camera.
        active_camera = new_camera;

        // Update our title string.
        title_str = WINDOW_TITLE_STR + QString::fromUtf8(": ") + active_camera->get_camera_name();
        setWindowTitle(title_str);
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void CameraTelemetry::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Ignore this if we don't have an active camera.
    if (active_camera == nullptr)
        return;

    // Is this event for us?
    if (ip_addr != active_camera->get_ip_addr())
        return;

    // Perform any special action we might want on a particular register.
    switch (reg_addr)
    {
    case GIGE_REG_TELEMETRY_JSON:
        // If there's already data outstanding, yell.
        if (json_data != nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "json_data is not NULL!";
            break;
        }

        // Read the buffer.
        json_data = active_camera->read_camera_buffer(GIGE_REG_TELEMETRY_JSON);

        // We're good to go. Start up our parsing thread.
        QtConcurrent::run(parse_sensor_data, this, json_data);
        break;
    default:
        break;
    }
}

void CameraTelemetry::timerEvent(QTimerEvent *event)
{
    // Log if this is unexpected.
    if (event && (event->timerId() != logger))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "TimerID is not logger!";
    }

    // Sanity check.
    if (active_camera == nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No active camera!";
        return;
    }

    // Ask the camera for our information. The register write initiates the ADC conversion.
    active_camera->write_camera_register(GIGE_REG_TELEMETRY_JSON, 0);
}

void CameraTelemetry::updateValue(double dValue)
{
    if (dValue < ui->temp_a10_int->minimum())
        dValue = ui->temp_a10_int->minimum();
    if (dValue > ui->temp_a10_int->maximum())
        dValue = ui->temp_a10_int->maximum();

    ui->temp_a10_int->setValue(dValue);
    ui->em2120_V->setText(QString("%1").arg(dValue / 100.0, 5, 'f', 2));
}

void CameraTelemetry::add_null_row()
{
    int col;
    QList<QStandardItem *> new_row;
    QTime time_now = QTime::currentTime();

    // Add the time.
    new_row.append(new QStandardItem(time_now.toString()));

    // Create a row of zeros.
    for (col = 1; col < log_data->columnCount(); ++col)
        new_row.append(new QStandardItem("0"));

    // Append our new row to the data table.
    log_data->appendRow(new_row);
}

/*!
 * @brief Utility function to convert from degrees Celsius to Fahrenheit.
 * @param c_value   Degrees Celsius.
 * @return Degrees Fahrenheit.
 */
double CameraTelemetry::c_to_f(double c_value)
{
    return (((9.0 / 5.0) * c_value) + 32.0);
}

/*!
 * @brief Utility function to convert from degrees Fahrenheit to Celsius.
 * @param f_value   Degrees Fahrenheit.
 * @return Degrees Celsius.
 */
double CameraTelemetry::f_to_c(double f_value)
{
    return ((f_value - 32.0) * 5.0 / 9.0);
}

/*!
 * @brief Write out the logged data (found in log_data) to a file.
 */
void CameraTelemetry::exportLog()
{
    int row, col;
    QString filename;
    QFile log_file;

    // Ask the user for a configuration file.
    filename = QFileDialog::getSaveFileName(this, tr("Save Log Data"), nullptr, tr("Log File (*.csv *.txt)"));
    if (filename.isEmpty())
        return;

    // Get the path for the log file.
    log_file.setFileName(filename);

    // Open the file to be written to.
    if (!log_file.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        QString msg_str = tr("Unable to open %1, Err=%2").arg(filename, log_file.errorString());
        QMessageBox::warning(this, "Error Opening File", msg_str);
        return;
    }
    log_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // Walk through the data and save off the information.
    QTextStream out_stream(&log_file);
    for (col = 0; col < log_data->columnCount(); ++col)
    {
        out_stream << log_data->horizontalHeaderItem(col)->text();
        if ((col + 1) < log_data->columnCount())
            out_stream << ",";
    }
    out_stream << Qt::endl;
    for (row = 0; row < log_data->rowCount(); ++row)
    {
        for (col = 0; col < log_data->columnCount(); ++col)
        {
            out_stream << log_data->item(row, col)->text();
            if ((col + 1) < log_data->columnCount())
                out_stream << ",";
        }

        out_stream << Qt::endl;
    }

    // All done.
    log_file.close();
}

void CameraTelemetry::parse_complete(const QList<QStandardItem *>& new_row)
{
    int last_row;

    // Get rid of the old data pointer.
    if (json_data != nullptr)
    {
        delete json_data;
        json_data = nullptr;
    }

    // Append our new row to the data table.
    log_data->appendRow(new_row);

    // Do we have any data in our log?
    last_row = log_data->rowCount() - 1;
    if (last_row >= 0)
    {
        QStandardItem *tbl_item;

        // We can export data now.
        if (!ui->pb_export->isEnabled())
            ui->pb_export->setEnabled(true);

        // Resize the columns in our table view.
        ui->log_table->resizeRowsToContents();
        ui->log_table->resizeColumnsToContents();
        ui->log_table->horizontalHeader()->setStretchLastSection(false);

        // Make sure we make this new item visible.
        ui->log_table->scrollToBottom();

        // Populate our UI now.
        tbl_item = log_data->item(last_row, COL_TEMP_A10_INT);
        if (tbl_item != nullptr)
            ui->temp_a10_int->setValue(tbl_item->text().toDouble());
        tbl_item = log_data->item(last_row, COL_TEMP_A10_EXT);
        if (tbl_item != nullptr)
            ui->temp_a10_ext->setValue(tbl_item->text().toDouble());
        tbl_item = log_data->item(last_row, COL_TEMP_SENSOR);
        if (tbl_item != nullptr)
            ui->temp_sensor->setValue(tbl_item->text().toDouble());
        tbl_item = log_data->item(last_row, COL_TEMP_NVME);
        if (tbl_item != nullptr)
            ui->temp_nvme->setValue(tbl_item->text().toDouble());
        tbl_item = log_data->item(last_row, COL_TEMP_FIREFLY);
        if (tbl_item != nullptr)
            ui->temp_firefly->setValue(tbl_item->text().toDouble());
        tbl_item = log_data->item(last_row, COL_TEMP_EM2120);
        if (tbl_item != nullptr)
            ui->temp_em2120->setValue(tbl_item->text().toDouble());
        tbl_item = log_data->item(last_row, COL_TEMP_FAN);
        if (tbl_item != nullptr)
            ui->temp_fan_external->setValue(tbl_item->text().toDouble());
        tbl_item = log_data->item(last_row, COL_TEMP_PCI);
        if (tbl_item != nullptr)
            ui->temp_pcie->setValue(tbl_item->text().toDouble());

        tbl_item = log_data->item(last_row, COL_SENSOR_ALERT);
        if (tbl_item != nullptr)
            ui->rb_sensor_pwr_ok->setChecked(tbl_item->text().toInt());
        tbl_item = log_data->item(last_row, COL_PMBUS_ALERT);
        if (tbl_item != nullptr)
            ui->rb_pmbus_alert->setChecked(tbl_item->text().toInt());

        tbl_item = log_data->item(last_row, COL_EM2120_CURR);
        if (tbl_item != nullptr)
            ui->em2120_I->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_EM2120_VOLT);
        if (tbl_item != nullptr)
            ui->em2120_V->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_A10_VEXT0);
        if (tbl_item != nullptr)
            ui->adc_vext0->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_A10_VEXT1);
        if (tbl_item != nullptr)
            ui->adc_vext1->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_A10_VCC);
        if (tbl_item != nullptr)
            ui->adc_vcc->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_A10_VCCP);
        if (tbl_item != nullptr)
            ui->adc_vccp->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_A10_VCCPT);
        if (tbl_item != nullptr)
            ui->adc_vccpt->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_A10_VCCERAM);
        if (tbl_item != nullptr)
            ui->adc_vcceram->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_A10_ADCGND);
        if (tbl_item != nullptr)
            ui->adc_gnd->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_FIREFLY_VOLT);
        if (tbl_item != nullptr)
            ui->firefly_V->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_FAN_TACH);
        if (tbl_item != nullptr)
            ui->fan_tach->setText(tbl_item->text());
        tbl_item = log_data->item(last_row, COL_FAN_PWM);
        if (tbl_item != nullptr)
            ui->fan_pwm->setText(tbl_item->text());
    }
    else
    {
        // Can't export if we don't have any data.
        if (ui->pb_export->isEnabled())
            ui->pb_export->setEnabled(false);
    }
}

/*!
 * @brief Parse the JSON encoded sensor telemetry and update the UI
 *        and the log.
 * @param json_data Pointer to the raw JSON data.
 */
void CameraTelemetry::parse_sensor_data(CameraTelemetry *tel_ptr, QByteArray *json_data)
{
    double json_val;
    QJsonDocument jsonMedia;
    QJsonObject telemetryObj;
    QJsonParseError jp_err;
    QString json_str = QString::fromUtf8(*json_data);
    QList<QStandardItem *> new_row;
    QStandardItem *tbl_item;

// #define DUMP_TEL 1
#ifdef DUMP_TEL
    if (!json_str.isEmpty())
    {
        QFile dump_file("tel_dump.txt");

        // Open the file to be written to.
        if (!dump_file.open(QIODevice::WriteOnly | QIODevice::Text))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to open tel_dump.txt:" << dump_file.errorString();
            return;
        }
        dump_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

        // Walk through the data and save off the information.
        QTextStream out_stream(&dump_file);
        out_stream << json_str;

        // Close the file.
        dump_file.close();
    }
#endif /* DUMP_TEL */

    // Strip any trailing NULL characters.
    if (json_str.at(json_str.size() - 1) == '\0')
        json_str.chop(1);

    // Convert to a JSON document.
    jsonMedia = QJsonDocument::fromJson(json_str.toUtf8(), &jp_err);
    if (jsonMedia.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Got a parse err:" << jp_err.errorString();
        return;
    }

    // Get the object from the document.
    telemetryObj = jsonMedia.object();

    // Build up each of the column entries for the new row.
    for (int i = 0; i < COL_COUNT; ++i)
    {
        bool bOK;
        QTime time_now;

        switch (i)
        {
        case COL_TIME:
            // Record the values in our logger.
            time_now = QTime::currentTime();
            tbl_item = new QStandardItem(time_now.toString());
            break;
        case COL_TEMP_A10_EXT:
            // Extract the values and update the UI.
            json_val = telemetryObj["temp_a10_ext"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_TEMP_A10_INT:
            json_val = telemetryObj["temp_a10_int"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_TEMP_SENSOR:
            json_val = telemetryObj["temp_sensor"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_TEMP_NVME:
            json_val = telemetryObj["temp_nvme"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_TEMP_FIREFLY:
            json_val = telemetryObj["temp_firefly"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_TEMP_EM2120:
            json_val = telemetryObj["temp_em2120"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_TEMP_FAN:
            json_val = telemetryObj["temp_fan_external"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_TEMP_PCI:
            json_val = telemetryObj["temp_pcie"].toDouble();
            if (!tel_ptr->degCelsius)
                json_val = tel_ptr->c_to_f(json_val);
            tbl_item = new QStandardItem(QString::number(json_val));
            break;
        case COL_SENSOR_ALERT:
            bOK = telemetryObj["sensor_pwr_ok"].toBool();
            tbl_item = new QStandardItem((bOK ? "1" : "0"));
            break;
        case COL_PMBUS_ALERT:
            bOK = telemetryObj["pmbus_alert"].toBool();
            tbl_item = new QStandardItem((bOK ? "1" : "0"));
            break;
        case COL_EM2120_CURR:
            tbl_item = new QStandardItem(QString("%1 A").arg(telemetryObj["em2120_I"].toDouble(), 5, 'f', 2));
            break;
        case COL_EM2120_VOLT:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["em2120_V"].toDouble(), 5, 'f', 2));
            break;
        case COL_FIREFLY_VOLT:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["firefly_V"].toDouble(), 5, 'f', 2));
            break;
        case COL_A10_VEXT0:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["vext0"].toDouble(), 5, 'f', 2));
            break;
        case COL_A10_VEXT1:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["vext1"].toDouble(), 5, 'f', 2));
            break;
        case COL_A10_VCC:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["vcc"].toDouble(), 5, 'f', 2));
            break;
        case COL_A10_VCCP:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["vccp"].toDouble(), 5, 'f', 2));
            break;
        case COL_A10_VCCPT:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["vccpt"].toDouble(), 5, 'f', 2));
            break;
        case COL_A10_VCCERAM:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["vcceram"].toDouble(), 5, 'f', 2));
            break;
        case COL_A10_ADCGND:
            tbl_item = new QStandardItem(QString("%1 V").arg(telemetryObj["adcgnd"].toDouble(), 5, 'f', 2));
            break;
        case COL_FAN_TACH:
            tbl_item = new QStandardItem(QString("%1").arg(telemetryObj["fan_tach_reading"].toInt()));
            break;
        case COL_FAN_PWM:
            tbl_item = new QStandardItem(QString("%1").arg(telemetryObj["fan_pwm_setting"].toInt()));
            break;
        default:
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected index: " << i;
            break;
        }

        // Add our new table element to the row.
        new_row.append(tbl_item);
    }

    // Let folks know we're done. Send the new row onwards.
    emit tel_ptr->parse_done(new_row);
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void CameraTelemetry::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    FI_comm     *comm_ptr;

    // Ignore this if we don't have an active camera.
    if (active_camera == nullptr)
        return;

    // Is this event for us?
    if (ip_addr != active_camera->get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_TELEMETRY_JSON:
        // Ok, now go read the data.
        active_camera->start_camera_read_buf(GIGE_REG_TELEMETRY_JSON, GIGE_TELEMETRY_JSON_LEN);
        break;
    default:
        break;
    }
}
