/*!
 * @file camera_video.cpp
 * @brief The code to display images to the user.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QAction>
#include <QApplication>
#include <QGraphicsLinearLayout>
#include <QGraphicsPixmapItem>
#include <QGraphicsLineItem>
#include <QGraphicsRectItem>
#include <QGraphicsScene>
#include <QGraphicsView>
#include <QGraphicsWidget>
#include <QImage>
#include <QMutexLocker>
#include <QPainter>
#include <QPolygon>
#include <QRgb>
#include <QResizeEvent>
#include <QSettings>
#include <QtConcurrent>
#include <QToolBar>
#include <QVBoxLayout>
#include <qmath.h>

// Made this ifdef because clang-tidy was puking on the OpenCV logic.
#define DO_OPENCV   1
#ifdef DO_OPENCV
#include <opencv2/core/core.hpp>
#include <opencv2/imgproc/imgproc.hpp>
#include <opencv2/calib3d/calib3d.hpp>
#include <opencv2/highgui/highgui.hpp>
#include <opencv2/imgcodecs.hpp>
#include <opencv2/videoio.hpp>
#endif /* DO_OPENCV */

#include <vector>

#include "camera_video.h"
#include "camera_mdi.h"
#include "camera_review.h"
#include "fi_graphicsview.h"
#include "fi_overlay.h"
#include "fi_rectitem.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_mdi.h"
#include "playback_meta_info.h"

namespace CAMERA_VIDEO
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The maximum number of QImage buffers we'll maintain.
const int       MAX_IMAGE_BUFFERS       = 5;

//! This is what I saw the mouseWheelEvent report as an angle.
const double    ZOOM_ANGLE              = 120.0;

//! Scale factor 1:1.
const int       SCALE_NORMAL            = 100;

//! Scale calculation denominator.
const int       SCALE_DEM               = 100;

//! Special case scale which means best fit.
const int       SCALE_FIT               = 0;

//! Amount of time to delay before rescaling the image.
const int       ZOOM_DELAY              = 100;

//! Maximum area for the video trigger.
const int       MAX_VT_AREA             = (128 * 1024);

//! Define a custom event to indicate when a conversion is complete.
const QEvent::Type CONVERSION_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_CONVERT_EVENT);

//! Subclass QEvent for our custom conversion event.
class ConversionEvent : public QEvent
{
public:
    explicit ConversionEvent(QImage *qImagePtrToSet): QEvent(CONVERSION_EVENT)
    {
        this->qImagePtr = qImagePtrToSet;
    }
    QImage *get_image_ptr() { return qImagePtr; }

private:
    QImage *qImagePtr;
};

}   // End of namespace CAMERA_VIDEO.

using namespace CAMERA_VIDEO;

// Initalize the static "colorize" variable.
bool CameraVideo::do_colorize = false;

// Initialize the static "focus assist" variable.
bool CameraVideo::do_focus_assist = false;

/*!
 * @brief The class constructor.
 */
CameraVideo::CameraVideo(QWidget *parent) :
    QWidget(parent)
{
    int         i;

    // Initialize our variables.
    thread_stop = false;
    for (i = 0; i < MAX_IMAGE_BUFFERS; ++i)
        avail_list.append(new QImage());
    overlay_list.clear();
    overlay_enabled = false;
    pending_rezoom = false;
    ae_reticle = nullptr;
    vt_reticle = nullptr;
    histogram_enabled = false;
    received_at_least_1_image = false;

    CameraMDI *myParentCamera = static_cast<CameraMDI *>(parent);

    // Connect the fault signal from the histogram processor.
    connect(&histo_proc, &fiHistoProc::fault, this, &CameraVideo::histogram_fault);
    //connect(parent, SIGNAL(closing()), this, SLOT(parent_closing()));
    connect(myParentCamera, &CameraMDI::closing, this, &CameraVideo::parent_closing);

    // Create our default scene.
    scene_ptr = new QGraphicsScene(this);
    horz_line = nullptr;
    vert_line = nullptr;
    pixmap_item_ptr = nullptr;
    fit_to_window = false;
    scale_factor = SCALE_FIT;
    rot_angle = 0.0;

    // Initialize our render members.
    render_scene = nullptr;
    render_pixmap = nullptr;
    render_list.clear();

    // Get our layout object.
    layout_ptr = new QGraphicsLinearLayout(Qt::Vertical);
    layout_ptr->setSpacing(0);
    layout_ptr->setContentsMargins(0, 0, 0, 0);

    // Get the graphics widget allocated to handle the layout.
    QGraphicsWidget *container = new QGraphicsWidget;
    container->setLayout(layout_ptr);

    // Add the widget to our scene.
    scene_ptr->addItem(container);

    // Initialize our default view.
    view_ptr = new fiGraphicsView(scene_ptr, this);

    // Set the panning mode and other options on the view.
    view_ptr->setDragMode(QGraphicsView::ScrollHandDrag);
    view_ptr->setRenderHints(QPainter::Antialiasing);
    view_ptr->setBackgroundRole(QPalette::Dark);
    view_ptr->show();
    connect(view_ptr, &fiGraphicsView::zoomed, this, &CameraVideo::image_zoomed);

    // Get our overlay action initialized.
    action_overlay = new QAction(parent);
    action_overlay->setObjectName(QStringLiteral("action_overlay"));
    action_overlay->setCheckable(false);
    action_overlay->setText(tr("Add Overlay"));
    action_overlay->setStatusTip(tr("Add an overlay object to the image."));
    enable_overlay(false);

    // Setup the layout.
    set_initial_layout(view_ptr);
}

/*!
 * @brief The class destructor.
 */
CameraVideo::~CameraVideo()
{
    // Save our settings.
    save_settings();

    // Is the thread running?
    if (!thread_stop)
    {
        // Tell the threads to stop.
        thread_stop = true;

        // Wait for all threads to shutdown.
        if (QThreadPool::globalInstance()->activeThreadCount())
            QThreadPool::globalInstance()->waitForDone();

        if (image_thread.isRunning())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The image_thread is running?";
        }
    }

    // Check to see if there is an autoexposure reticle.
    if (ae_reticle != nullptr)
    {
        // Make sure there is a scene pointer before we remove the
        // autoexposure reticle.
        if (scene_ptr != nullptr)
            scene_ptr->removeItem(ae_reticle);
        delete ae_reticle;
        ae_reticle = nullptr;
    }

    // Check to see if there is a video trigger reticle.
    if (vt_reticle != nullptr)
    {
        // Make sure there is a scene pointer before we remove the
        // autoexposure reticle.
        if (scene_ptr != nullptr)
            scene_ptr->removeItem(vt_reticle);
        delete vt_reticle;
        vt_reticle = nullptr;
    }

    {
        // Grab a new image from the pool.
        QMutexLocker my_locker(&avail_mutex);

        // Clean out our lists.
        while (!avail_list.isEmpty())
        {
            QImage *qImagePtr = avail_list.takeFirst();
            if (qImagePtr != nullptr)
            {
                delete qImagePtr;
            }
            else
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "avail_list had a null QImage*";
            }
        }
    }

    if (action_overlay != nullptr)
    {
        delete action_overlay;
        action_overlay = nullptr;
    }
}


void CameraVideo::cleanup_overlays(void)
{
    int     i;

    // Walk through the list of overlays and remove any "non-pinned" overlays from the layout object.
    for (i = 0; i < overlay_list.count(); ++i)
    {
        // Is this item "pinned"?
        if (!overlay_list.at(i)->isPinned())
        {
            // Not pinned, so shouldn't be in the layout.
            layout_ptr->removeItem(overlay_list.at(i));
        }
    }
}

/*!
 * @brief Apply our colorization algorithm to the image.
 *
 * If any pixel value is above the CAMERA_VIDEO::COLORIZE_THRESHOLD then
 * force that pixel to be red.
 *
 * @param image_ptr The pointer to the QImage object we're interested in.
 */
void CameraVideo::colorize_image(QImage *image_ptr)
{
    int col, row, my_width, my_height;
    int format;
    uint32_t user_color;
    uint8_t *image_data;
    QSettings settings;

    settings.beginGroup("Colorize");
    user_color = settings.value("color", qRgb(255, 0, 0)).toUInt();
    settings.endGroup();

    // Gather information about our image for the loops below.
    format = image_ptr->format();
    my_width = image_ptr->width();
    my_height = image_ptr->height();
    image_data = image_ptr->bits();

    // Find if any pixel is saturated.
// *INDENT-OFF*
#pragma omp parallel for collapse(2)
// *INDENT-ON*
    for (row = 0; row < my_height; ++row)
    {
        for (col = 0; col < my_width; ++col)
        {
            int index;

            // Our images are either RGB888 or RGB32.
            index = (row * my_width) + col;
            if (format == QImage::Format_RGB888)
                index *= 3;
            else
                index *= 4;

            // If this pixel is in any way saturated, colorize it. Don't worry about the
            // alpha channel.
            if ((image_data[index] > CAMERA_VIDEO::COLORIZE_THRESHOLD) || (image_data[index + 1] > CAMERA_VIDEO::COLORIZE_THRESHOLD) || (image_data[index + 2] > CAMERA_VIDEO::COLORIZE_THRESHOLD))
            {
                if (format == QImage::Format_RGB888)
                {
                    image_data[index]       = (uint8_t)qRed(user_color);    // Red
                    image_data[index + 1]   = (uint8_t)qGreen(user_color);  // Green
                    image_data[index + 2]   = (uint8_t)qBlue(user_color);   // Blue
                }
                else
                {
                    // RGB32 is 0xFFrrggbb
                    image_data[index]       = (uint8_t)qBlue(user_color);   // Blue
                    image_data[index + 1]   = (uint8_t)qGreen(user_color);  // Green
                    image_data[index + 2]   = (uint8_t)qRed(user_color);    // Red
                }
            }
        }
    }
}

void CameraVideo::focus_assist(QImage *image_ptr)
{
#ifdef DO_OPENCV
    int my_width, my_height;
    int my_step;
    uint8_t *image_data;
    double lowThresh, highThresh;
    QImage imgEdges;
    QSettings settings;
    int apertureSize;
    int ksize;
    double sigmaX;
    double ltFactor;
    QString settingsStr;
    int penWidth;
    uint32_t user_color;

    // Read the settings from the registry. Normally there won't be anything there and
    // we'll just use the defaults, but this gives us the option to override.
    settings.beginGroup("FocusAssist");
    apertureSize = settings.value(QStringLiteral("apertureSize"), 3).toInt();
    ksize = settings.value(QStringLiteral("ksize"), 3).toInt();
    settingsStr = settings.value(QStringLiteral("sigmaX"), "0.0").toString();
    sigmaX = settingsStr.toDouble();
    settingsStr = settings.value(QStringLiteral("ltFactor"), "0.5").toString();
    ltFactor = settingsStr.toDouble();
    penWidth = settings.value(QStringLiteral("penWidth"), 3).toInt();
    user_color = settings.value("color", qRgb(255, 0, 0)).toUInt();
    settings.endGroup();

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "apertureSize:" << apertureSize << "ksize:" << ksize << "sigmaX:" << sigmaX << "ltFactor:" << ltFactor;

    // Make sure we're grayscale.
    if (image_ptr->format() != QImage::Format_Grayscale8)
    {
        // Convert the image.
        imgEdges = image_ptr->convertToFormat(QImage::Format_Grayscale8);
    }
    else
    {
        // Copy the image date over.
        imgEdges = image_ptr->copy();
    }

    // Gather information about our image for the loops below.
    my_width = imgEdges.width();
    my_height = imgEdges.height();
    image_data = imgEdges.bits();
    my_step = ((my_width + 3) / 4) * 4;

    // Put the QImage into a format OpenCV can deal with.
    cv::Mat img_mat(my_height, my_width, CV_8U, image_data, my_step);

    // Blur the image a bit. This removes the high frequency noise
    // associated with the image, and lets the edge detection algorithm
    // work better.
    cv::GaussianBlur(img_mat, img_mat, cv::Size(ksize, ksize), sigmaX);

    // To calculate our thresholds, we're going to make use of Otsu's threshold. Using
    // the Otsu's threshold value as the higher and 1/2 for the lower.
    // See: https://stackoverflow.com/questions/4292249/automatic-calculation-of-low-and-high-thresholds-for-the-canny-operation-in-open
    // See: https://riptutorial.com/opencv/topic/6099/edge-detection
    cv::Mat dummy;
    highThresh = cv::threshold(img_mat, dummy, 0, 255, cv::THRESH_BINARY | cv::THRESH_OTSU);
    lowThresh = highThresh * ltFactor;

    // Perform edge detection.
    cv::Canny(img_mat, img_mat, lowThresh, highThresh, apertureSize);

    // Dilate the Canny output to remove potential holes.
    cv::dilate(img_mat, img_mat, cv::Mat());

    {
        // Create our contour vector variables.
        std::vector<std::vector<cv::Point>> contours;
        std::vector<cv::Vec4i> hierarchy;

        // Find the contours.
        cv::findContours(img_mat, contours, hierarchy, cv::RETR_LIST, cv::CHAIN_APPROX_SIMPLE);

        // Make our output an color pixmap.
        QImage resultImage(my_width, my_height, QImage::Format_ARGB32);
        QPainter painter(&resultImage);

        // Create our pen.
        QPen contourPen(user_color);
        contourPen.setWidth(penWidth);

        painter.setPen(contourPen);
        painter.setCompositionMode(QPainter::CompositionMode_Source);
        painter.fillRect(resultImage.rect(), Qt::transparent);

        // Loop through the contours vector, drawing the polylines.
        for (int i = 0; i < (int)contours.size(); ++i)
        {
            // Convert the OpenCV vector of points to a Qt vector of points.
            QPolygon points;
            std::vector<cv::Point> the_vec = contours.at(i);
            for (int pt = 0; pt < (int)the_vec.size(); ++pt)
                points.append(QPoint(the_vec.at(pt).x, the_vec.at(pt).y));

            // Create a polygon and draw it.
            painter.drawPolyline(points);
        }

        // All done.
        painter.end();

        // Copy over the result.
        imgEdges = resultImage.copy();
    }

#if 1
    // Combine the two images.
    // See: https://doc.qt.io/qt-5/qtwidgets-painting-imagecomposition-example.html
    {
        QImage resultImage(my_width, my_height, QImage::Format_ARGB32);
        QPainter painter(&resultImage);
        QPainter::CompositionMode mode = QPainter::CompositionMode_Overlay;

        painter.setCompositionMode(QPainter::CompositionMode_Source);
        painter.fillRect(resultImage.rect(), Qt::transparent);
        painter.setCompositionMode(QPainter::CompositionMode_SourceOver);
        painter.drawImage(0, 0, imgEdges);      // Destination Image
        painter.setCompositionMode(mode);
        painter.drawImage(0, 0, *image_ptr);    // Source Image
        painter.setCompositionMode(QPainter::CompositionMode_DestinationOver);
        painter.end();

        // Copy over the result.
        *image_ptr = resultImage.copy();
    }
#else
    *image_ptr = imgEdges.copy();
#endif
#else
    (void)image_ptr;
#endif /* DO_OPENCV */
}

/*!
 * @brief Copy the currently active image to the dest_pixmap object.
 * @param dest_pixmap   Destination QPixmap object for the current image.
 */
void CameraVideo::copy_image(QPixmap& dest_pixmap)
{
    // Do we have an image?
    if (pixmap_item_ptr == nullptr)
        dest_pixmap = QPixmap();
    else
        dest_pixmap = pixmap_item_ptr->pixmap();
}

/*!
 * @brief This is a static method shared by all CameraVideo instantiations. It is used
 *        by the QtConcurrent::run() method to operate in a seperate thread. The purpose
 *        of this method is to convert the incoming image buffer into a QImage object.
 *
 * @param parent        Pointer to the instantiated CameraVideo class we're a member of.
 * @param data_ptr      Pointer to the image buffer we want to convert.
 */
void CameraVideo::convert_image(QWidget *parent)
{
    QImage  *qImagePtr = nullptr;
    bool    status;
    CameraVideo *my_class = qobject_cast<CameraVideo *>(parent);

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Make sure there is an active image buffer to process!
    if (my_class->activeImageBufferPtr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "EMPTY activeImageBufferPtr!";
        return;
    }

    // Make sure we're not supposed to stop.
    if (!my_class->thread_stop)
    {
        // Are there available image buffers and are we not stopping?
        if (!my_class->avail_list.isEmpty())
        {
            {
                // Grab a new image from the pool.
                QMutexLocker my_locker(&my_class->avail_mutex);
                qImagePtr = my_class->avail_list.takeFirst();
            }

// #define DO_SAVE_FILE 1
#ifdef DO_SAVE_FILE
            {
                QString file_name = "./test_image_" + QString::number(data_ptr->seq_number) + ".tif";
                QFile new_doc(file_name);
                if (new_doc.open(QIODevice::WriteOnly))
                    new_doc.write((const char *)data_ptr->image_ptr, data_ptr->image_len);
                new_doc.close();
            }
#endif

            // Convert the raw data to the image buffer.
            status = qImagePtr->loadFromData(my_class->activeImageBufferPtr->GetMemDataPtr(), (int)my_class->activeImageBufferPtr->GetImageDataLen());
            if (!status)
            {
                // Log the error, but we need to keep moving.
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "LoadFromData() failed.";
            }
            else
            {
                // Are we colorizing the image?
                if (CameraVideo::do_colorize)
                {
                    // If the image is greyscale, we need to convert it.
                    if (qImagePtr->isGrayscale())
                    {
                        QImage color_image = qImagePtr->convertToFormat(QImage::Format_RGB888);
                        qImagePtr->swap(color_image);
                    }

                    // Apply colorization to the image.
                    my_class->colorize_image(qImagePtr);
                }
                if (CameraVideo::do_focus_assist)
                {
                    my_class->focus_assist(qImagePtr);
                }
            }
        }
    }

    // NOTE:  Provide the converted QImage to our custom event handler
    // and allow the GUI thread to continue.  The allocated "new"
    // ConversionEvent is freed by the QT system.
    //
    // Post an event so the GUI thread can carry on.
    QApplication::postEvent(my_class, new ConversionEvent(qImagePtr));
}

/*!
 * @brief Display the pixmap in our "scene".
 * @param my_pixelmap   The pixmap we're interested in.
 */
void CameraVideo::display_image(const QPixmap& my_pixelmap)
{
    bool    first_image;

    // Add our pixmap item to the scene.
    if (pixmap_item_ptr == nullptr)
    {
        // Create the new pixmap item.
        pixmap_item_ptr = new fiLayoutPixmap(my_pixelmap);

        // Add the image to our scene.
        scene_ptr->addItem(pixmap_item_ptr);
        layout_ptr->addItem(pixmap_item_ptr);
        layout_ptr->setStretchFactor(pixmap_item_ptr, QSizePolicy::Expanding);

        // Load our settings.
        load_settings();

        // This is the first image.
        first_image = true;
    }
    else
    {
        // Break the parental link to reticles if one exists.
        if (vt_reticle && (vt_reticle->parentItem() == pixmap_item_ptr))
            vt_reticle->setParentItem(nullptr);
        if (ae_reticle && (ae_reticle->parentItem() == pixmap_item_ptr))
            ae_reticle->setParentItem(nullptr);

        // Remove the old pixmap item.
        layout_ptr->removeItem(pixmap_item_ptr);
        scene_ptr->removeItem(pixmap_item_ptr);
        delete pixmap_item_ptr;

        // Create the new pixmap item.
        pixmap_item_ptr = new fiLayoutPixmap(my_pixelmap);

        // If we are overlay_enabled check for the active reticle and establish image parentage if needed.
        if (pixmap_item_ptr)
        {
            fiRectItem *fi_vt_rect = nullptr;
            fiRectItem *fi_ae_rect = nullptr;

            // Pull the fiRectItem pointers for our reticles.
            if (vt_reticle)
                fi_vt_rect = static_cast<fiRectItem *>(vt_reticle);
            if (ae_reticle)
                fi_ae_rect = static_cast<fiRectItem *>(ae_reticle);


            // Modify our reticles.
            if (fi_vt_rect)
            {
                // If we are enabled for overlay support make reticle a child of the image.
                if (overlay_enabled)
                    fi_vt_rect->setParentItem(pixmap_item_ptr);

                // Mark the reticle as moveable or not.
                fi_vt_rect->setFlag(QGraphicsItem::ItemIsMovable, !overlay_enabled);
                fi_vt_rect->setMoveable(!overlay_enabled);
            }
            if (fi_ae_rect)
            {
                // If we are enabled for overlay support make reticle a child of the image.
                if (overlay_enabled)
                    fi_ae_rect->setParentItem(pixmap_item_ptr);

                // Mark the reticle as moveable or not.
                fi_ae_rect->setFlag(QGraphicsItem::ItemIsMovable, !overlay_enabled);
                fi_ae_rect->setMoveable(!overlay_enabled);
            }
        }

        scene_ptr->addItem(pixmap_item_ptr);

        // Walk through our overlays to determine where this pixmap will reside.
        if (overlay_enabled)
        {
            QList<int> overlays_pinned_to_bottom;

            // Process all of the overlays pinned to the top.
            for (int i = 0; i < overlay_list.count(); ++i)
            {
                // Is the overlay pinned into a position?
                if (overlay_list.at(i)->isPinned())
                {
                    // Is this pinned overlay a top border pin scenario?
                    if (overlay_list.at(i)->isPinnedAtTop())
                    {
                        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "TOP i=" << i;
                        // Pin the overlay to the top which inserts the overlay into position 0 of layout.
                        pin_overlay(overlay_list.at(i), true);
                    }
                    else
                    {
                        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "BOTTOM i=" << i;
                        // Save this bottom pinned overlay index for later additions to end of layout.
                        overlays_pinned_to_bottom.append(i);

                        // Take this overlay out of the layout.
                        layout_ptr->removeItem(overlay_list.at(i));
                    }
                }
            }

            // add the display image to the layout now after any top pinned overlays have been positioned.
            layout_ptr->addItem(pixmap_item_ptr);

            // Process any bottom pinned overlays to add to the layout at the end.
            while (overlays_pinned_to_bottom.count() > 0)
            {
                int overlay_index = overlays_pinned_to_bottom.takeFirst();
                pin_overlay(overlay_list.at(overlay_index), false);
            }
        }
        else
        {
            // No overlays to worry about just insert the pixmap at position 0.
            layout_ptr->insertItem(0, pixmap_item_ptr);
        }
        layout_ptr->setStretchFactor(pixmap_item_ptr, QSizePolicy::Expanding);

        // Not the first image.
        first_image = false;
    }

    // Is this the first image?
    if (first_image)
    {
        // We need a rezoom.
        pending_rezoom = true;
    }
    else
    {
        QRectF  scene_rect;

        // Adjust the scaling if necessary.
        scene_rect = size_image();

        // Adjust the scene ROI.
        scene_ptr->setSceneRect(scene_rect);
    }

    // Do we have a rezoom request pending?
    if (pending_rezoom)
    {
        QTimer::singleShot(ZOOM_DELAY, this, &CameraVideo::rezoom);
        pending_rezoom = false;
    }
}

/*!
 * @brief Enable or disable the aiming cursor function.
 * @param do_enable     True to enable, false to disable.
 */
void CameraVideo::enable_aiming_cursor(bool do_enable)
{
    // Make sure we clear out the old lines.
    if (horz_line != nullptr)
    {
        scene_ptr->removeItem(horz_line);
        horz_line = nullptr;
    }
    if (vert_line != nullptr)
    {
        scene_ptr->removeItem(vert_line);
        vert_line = nullptr;
    }

    // Are we displaying the lines?
    if (do_enable)
    {
        // Get the rectangle enclosing the image.
        QRectF sc_rect = get_image_rect();

        // Add our lines to the scene.
        horz_line = scene_ptr->addLine(sc_rect.left(), sc_rect.top() + (sc_rect.height() / 2.0), sc_rect.right(), sc_rect.top() + (sc_rect.height() / 2.0), QPen(Qt::red, 2));
        horz_line->setZValue(1);
        vert_line = scene_ptr->addLine(sc_rect.left() + (sc_rect.width() / 2.0), sc_rect.bottom(), sc_rect.left() + (sc_rect.width() / 2.0), sc_rect.top(), QPen(Qt::red, 2));
        vert_line->setZValue(1);
    }
}

/*!
 * @brief Enable or disable the information cursor function.
 * @param do_enable     True to enable, false to disable.
 */
void CameraVideo::enable_info_cursor(bool do_enable)
{
    // Pass it on to our graphics view object.
    view_ptr->enable_info_cursor(do_enable);
}

/*!
 * @brief Enable or disable the overlays.
 * @param do_enable     True to enable, false to disable.
 */
void CameraVideo::enable_overlay(bool do_enable)
{
    int i;

    // Remember the setting.
    overlay_enabled = do_enable;

    // Make sure our action reflects the setting.
    get_action_overlay()->setEnabled(do_enable);

    // If we are enabling the overlay, then we have somethings to take care of first.
    if (do_enable)
    {
        PlaybackMDI *playback_ptr;
        CameraMDI   *camera_ptr;

        // Establish a parental relationship for our reticles if reviewing.
        // Note: Establishing parentage here repositions any reticles as overlays
        // are being rendered in the scene. Without it old positioning remains on
        // the screen noticibly.
        set_reticle_parent(vt_reticle, pixmap_item_ptr);
        set_reticle_parent(ae_reticle, pixmap_item_ptr);

        // Need to figure out if we have a camera pointer or playback pointer.
        playback_ptr = qobject_cast<PlaybackMDI *>(parentWidget());
        camera_ptr = qobject_cast<CameraMDI *>(parentWidget());

        // Make sure our overlay widgets are configured.
        if (camera_ptr)
            update_overlay(camera_ptr->get_current_frame_num(), camera_ptr->get_info_widget(), nullptr);
        else
            update_overlay(playback_ptr->getLastFrame(), playback_ptr->getCameraInfo(), playback_ptr->getMetadata());

        // If the overlay is supposed to be pinned...
        for (i = 0; i < overlay_list.count(); ++i)
        {
            if (overlay_list.at(i)->isPinned())
            {
                // Pin it.
                pin_overlay(overlay_list.at(i), overlay_list.at(i)->isPinnedAtTop());
            }
        }
    }

    // Update the overlays according to the setting.
    for (i = 0; i < overlay_list.count(); ++i)
    {
        if (do_enable)
        {
            // Notes:
            // Is the overlay already part of the scene?  When first loading or
            // creating an overlay it is added to the scene and when the this enabler
            // is invoked a warning comes out about it already being part of the scene.
            // To be more efficient with scene management look to see if this overlay
            // is already part of QGraphicsScene list and if so we won't add it again.
            // In order to eliminate the overlay offset problem documented in Ticket #348
            // (see the else condition to the if(do_enable) above) the overlays are deleted
            // from the scene.  Therefore upon re-enabling an overlay it must be added back
            // to the scene if it's not there.  So this next logic accomplishes this task.
            // The case that occurs causing a re-insertion back to the scene is by moving
            // to the live view then back to the review view.  Coming back to review causes
            // this function to fire and in that case there is no overlay in the scene.
            QList<QGraphicsItem *> scene_items = scene_ptr->items();
            bool overlay_in_scene = false;
            for (int scene_index = 0; !overlay_in_scene && scene_index < scene_items.count(); scene_index++)
            {
                if (scene_items.at(scene_index) == overlay_list.at(i))
                {
                    overlay_in_scene = true;
                }
            }

            // Only add it if it's not there.
            if (!overlay_in_scene)
            {
                scene_ptr->addItem(overlay_list.at(i));
            }

            // Now pin the overlay if it is pinned.
            if (overlay_list.at(i)->isPinned())
            {
                // Pin it.
                pin_overlay(overlay_list.at(i), overlay_list.at(i)->isPinnedAtTop());
            }
        }
        else
        {
            // Ticket #348 we want to remove the overlay from our scene and layout for video review.
            // When the overlay is left in the scene the layout engine seems to still leave a spot for
            // it in the display frame.
            scene_ptr->removeItem(overlay_list.at(i));
            layout_ptr->removeItem(overlay_list.at(i));
        }
        overlay_list.at(i)->setVisible(do_enable);
    }

    // Delay for a bit to let the layout do it's thing.
    QTimer::singleShot(ZOOM_DELAY, this, &CameraVideo::rezoom);
}

void CameraVideo::enable_ae_reticle(bool do_enable)
{
    // Sanity check.
    if (ae_reticle == nullptr)
        return;

    // Show/hide the reticle.
    ae_reticle->setEnabled(do_enable);
}

void CameraVideo::enable_vt_reticle(bool do_enable)
{
    fiRectItem *fi_rect = static_cast<fiRectItem *>(vt_reticle);

    // Sanity check.
    if (fi_rect == nullptr)
        return;

    // Register/deregister the callback.
    fi_rect->registerCallback(do_enable ? this : nullptr);

    // Show/hide the reticle.
    fi_rect->setEnabled(do_enable);
}

QRect CameraVideo::get_ae_rect()
{
    QRect our_roi = QRect(0, 0, 0, 0);

    // Sanity check.
    if (ae_reticle != nullptr)
    {
        QRectF user_rect;

        // Get the ROI in scene coordinates.
        user_rect = ae_reticle->mapRectToScene(ae_reticle->rect());

        // Convert to integer values.
        our_roi = user_rect.toAlignedRect();
    }

    // Return the ROI.
    return our_roi;

}

fiRectItem *CameraVideo::get_ae_reticle()
{
    return static_cast<fiRectItem *>(ae_reticle);
}

QRect CameraVideo::get_image_rect()
{
    QRect our_roi = QRect(0, 0, 0, 0);

    // Sanity check.
    if (pixmap_item_ptr != nullptr)
    {
        QRectF user_rect;

        // Get the ROI in scene coordinates.
        user_rect = pixmap_item_ptr->mapRectToScene(pixmap_item_ptr->boundingRect());

        // Convert to integer values.
        our_roi = user_rect.toAlignedRect();
    }

    // Return the ROI.
    return our_roi;
}

QRect CameraVideo::get_vt_rect()
{
    QRect our_roi = QRect(0, 0, 0, 0);

    // Sanity check.
    if (vt_reticle != nullptr)
    {
        QRectF user_rect;
        QPointF user_pos;

        // Get the ROI in scene coordinates.
        user_rect = get_vt_reticle()->adjrect();
        user_pos = get_vt_reticle()->pos();

        // Move the rectangle to the right position.
        user_rect.moveTopLeft(user_pos);

        // Convert to integer values.
        our_roi = user_rect.toAlignedRect();
    }

    // Return the ROI.
    return our_roi;
}

fiRectItem *CameraVideo::get_vt_reticle()
{
    return static_cast<fiRectItem *>(vt_reticle);
}

/*!
 * @brief Return a reasonable idea for what the zoom is set to. This doesn't
 *        return the actual scale value, but just one of the constants so the
 *        UI code can know which button to "check".
 */
int CameraVideo::get_zoom_setting(void)
{
    int zoom_setting;

    if (fit_to_window)
        zoom_setting = FASMOTION_SETTINGS::ZOOM_TO_WINDOW;
    else
    {
        if (qFuzzyCompare(scale_factor, SCALE_NORMAL))
            zoom_setting = FASMOTION_SETTINGS::ZOOM_1_TO_1;
        else if (qFuzzyCompare(scale_factor, SCALE_FIT))
            zoom_setting = FASMOTION_SETTINGS::ZOOM_TO_FIT;
        else
            zoom_setting = FASMOTION_SETTINGS::ZOOM_UNKNOWN;
    }

    return zoom_setting;
}

bool CameraVideo::is_ae_reticle_enabled()
{
    if (ae_reticle == nullptr)
        return true;

    return ae_reticle->isEnabled();
}

bool CameraVideo::is_ae_reticle_visible()
{
    if (ae_reticle == nullptr)
        return false;
    return ae_reticle->isVisible();
}

bool CameraVideo::is_vt_reticle_enabled()
{
    if (vt_reticle == nullptr)
        return false;
    return vt_reticle->isEnabled();
}

bool CameraVideo::is_vt_reticle_visible()
{
    if (vt_reticle == nullptr)
        return false;
    return vt_reticle->isVisible();
}

void CameraVideo::load_reticle_settings(QGraphicsRectItem **reticle_ptr, QString settings_str)
{
    QPointF reticle_pos;
    QRectF  scene_rect;
    QRectF  reticle_rect;
    bool    reticle_enable;
    bool    reticle_visible;
    QSettings   settings;
    QGraphicsRectItem *reticle;

    // Load up the information for the reticle.
    settings.beginGroup(settings_str);

    // See if we were enabled previously.
    reticle_enable = settings.value(QStringLiteral("enabled"), false).toBool();

    // See if we were visible.
    reticle_visible = settings.value(QStringLiteral("visible"), false).toBool();

    // Read up the reticle ROI information.
    reticle_rect = settings.value(QStringLiteral("ROI"), QRectF(0, 0, 0, 0)).toRectF();
    reticle_pos = settings.value(QStringLiteral("pos"), QPointF(0, 0)).toPointF();
    settings.endGroup();

    // Get the bounding rectangle for the scene.
    scene_rect = scene_ptr->sceneRect();

    // Was there a reticle?
    if (reticle_rect.isEmpty())
    {
        // We want a reticle the user can drag around to specify the region
        // to use for calculating the custom white balance gains. Center the
        // reticle and make it 5% the size of the scene.
        reticle_rect.setHeight(scene_rect.height() * 0.05);
        reticle_rect.setWidth(scene_rect.width() * 0.05);

        // Have our reticle start in the middle of the image.
        reticle_pos.setX((scene_rect.width() - reticle_rect.width()) / 2);
        reticle_pos.setY((scene_rect.height() - reticle_rect.height()) / 2);
    }

    // Allocate our reticle object.
    reticle = new fiRectItem(reticle_rect);

    // Make sure the user can grab the rectangle and move it around.
    reticle->setFlag(QGraphicsItem::ItemIsMovable, true);

    // Add it to the scene.
    scene_ptr->addItem(reticle);

    // Now move the reticle into the scene.
    reticle->setPos(reticle_pos);

    // Make sure we are on top of the (default) scene. The default Z-value is 0.
    reticle->setZValue(1);

    // Show/hide it based on our previous setting.
    reticle->setVisible(reticle_visible);

    // Were we enabled?
    reticle->setEnabled(reticle_enable);

    // Do we have someplace to put this?
    if (reticle_ptr != nullptr)
        *reticle_ptr = reticle;
}

/*!
 * @brief load the user specified settings fromthe non-volatile parameter storage.
 */
void CameraVideo::load_settings(void)
{
    int i;
    int count;
    QRectF roi;
    QRectF scene_rect;
    QSettings   settings;
    CameraMDI   *camera_ptr;

    settings.beginGroup("Video");

    // Is the aiming cursor enabled?
    bool aiming = settings.value("aiming", false).toBool();
    MainWindow::GetInstance()->get_view_aiming_cursor()->setChecked(aiming);
    enable_aiming_cursor(aiming);

    settings.endGroup();

    // Check for colorization. It's really part of the display panel,
    // but we need that information here too.
    settings.beginGroup("Colorize");
    CameraVideo::do_colorize = settings.value("enable", false).toBool();
    settings.endGroup();

    settings.beginGroup("FocusAssist");
    CameraVideo::do_focus_assist = settings.value("enable", false).toBool();
    settings.endGroup();

    // Need to figure out if we have a camera pointer.
    camera_ptr = qobject_cast<CameraMDI *>(parentWidget());

    // Only worry about the reticles if we are connected to a camera.
    if (camera_ptr)
    {
        // Load in the autoexposure reticle information.
        load_reticle_settings(&ae_reticle, QStringLiteral("ae_reticle"));
        if (ae_reticle == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "ae_reticle is nullptr!?";
        }
        else
        {
            // Sanity check
            roi = get_ae_rect();
            scene_rect = get_image_rect();
            if (!scene_rect.contains(roi))
            {
                reset_ae_reticle();
            }

            // By default, we do not want the tracker active. The "enable"
            // means to enable the control, not enable tracking.
            enable_ae_reticle(true);
            show_ae_reticle(false);
        }

        // Load in the video trigger reticle information.
        load_reticle_settings(&vt_reticle, QStringLiteral("vt_reticle"));
        if (vt_reticle == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "vt_reticle is nullptr!?";
        }
        else
        {
            // Make sure our max area is set.
            get_vt_reticle()->setMaxArea(MAX_VT_AREA);

            // Sanity check
            roi = get_vt_rect();
            scene_rect = get_image_rect();
            if (!scene_rect.contains(roi))
            {
                reset_vt_reticle();
            }

            // Make sure our reticle is setup properly.
            get_vt_reticle()->setEnabledPen(QPen(Qt::white, 3));
            get_vt_reticle()->setEnabledBrush(QBrush(Qt::NoBrush));

            // By default, we do not want the tracker active.
            enable_vt_reticle(false);
            show_vt_reticle(false);
        }
    }

    // Read in the overlay information.
    settings.beginGroup(QStringLiteral("Overlays"));
    count = settings.value(QStringLiteral("count"), 0).toInt();
    if (count == 0)
        return;

    for (i = 0; i < count; ++i)
    {
        // Create a new overlay object.
        fiOverlay *overlay_ptr = new fiOverlay;
        if (overlay_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Error allocating overlay object!";
            return;
        }

        // Load the overlay settings.
        overlay_ptr->load_settings(settings, i);

        // Append the overlay to our list.
        overlay_list.append(overlay_ptr);
        connect(overlay_ptr, &fiOverlay::remove_overlay, this, &CameraVideo::do_delete_overlay);
        connect(overlay_ptr, &fiOverlay::pin_to_bottom, this, &CameraVideo::do_pin2bottom_overlay);
        connect(overlay_ptr, &fiOverlay::pin_to_top, this, &CameraVideo::do_pin2top_overlay);
        connect(overlay_ptr, &fiOverlay::need_update, this, &CameraVideo::do_update_overlay);
        connect(overlay_ptr, &fiOverlay::overlay_moved, this, &CameraVideo::do_resize_overlay);

        // Add it to our scene.
        scene_ptr->addItem(overlay_ptr);

        // Make sure we are on top of the (default) scene. The default Z-value is 0.
        overlay_ptr->setZValue(1);
    }
    settings.endGroup();

    // If we are adding overlays, make sure we add padding.
    set_overlay_padding(overlay_list.count() > 0);

    // Show/hide the overlay information as appropriate.
    enable_overlay(overlay_enabled);
}

void CameraVideo::pin_overlay(fiOverlay *overlay_ptr, bool to_top)
{
    // Tell the overlay that it's pinned.
    overlay_ptr->setPinned(true, to_top, pixmap_item_ptr->boundingRect().width());

    // Remove the overlay from the layout.
    layout_ptr->removeItem(overlay_ptr);

    // If we're going on top, insert it in the layout at the front, otherwise
    // append it to the end.
    if (to_top)
        layout_ptr->insertItem(0, overlay_ptr);
    else
        layout_ptr->addItem(overlay_ptr);

    // Make sure our layout is still setup properly.
    layout_ptr->setSpacing(0);
    layout_ptr->setContentsMargins(0, 0, 0, 0);
}


/*!
 * @brief Receive an image buffer to be processed and displayed.
 * @param image_ptr     The image buffer we're interested in.
 */
void CameraVideo::receive_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer>imageBufferPtr)
{
    // Is the imageBufferPtr valid?
    // Sanity check.  Log a diagnostic out there for this resource.
    if (imageBufferPtr)
    {
        QString myInfo = QStringLiteral("[CameraVideo::receive_image()]");
        imageBufferPtr->LogMyDataId(&myInfo);
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESOURCE EMPTY:  imageBufferPtr was nullptr!";
        return;
    }

    // Is there a thread already running?
    if (image_thread.isRunning())
    {
        // Drop this image and move on.
        release_image(std::move(imageBufferPtr));
        return;
    }

    // Make sure we're not stopping.
    if (thread_stop)
    {
        // We're stopping, so just release the image.
        release_image(std::move(imageBufferPtr));
        return;
    }

    // Make sure that there isn't an activeImageBuffer - active!
    if (activeImageBufferPtr != nullptr)
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Releasing activeImageBufferPtr as have a new image to process.";
        release_image(std::move(imageBufferPtr));
        return;
    }

    // Install the activeImageBufferPtr with the current imageBufferPtr.
    //
    // NOTE:  This is consumed by the convert_image thread.  It is the only
    // way that the thread knows about this image resource because we cannot
    // pass unique_ptr resources using QtConcurrent::run command.
    activeImageBufferPtr = std::move(imageBufferPtr);

    // We're good to go. In a seperate thread, convert from our image buffer
    // to a QImage object.
    image_thread = QtConcurrent::run(convert_image, this);

    // In case anyone wants to know if this video object has received at least one image.
    received_at_least_1_image = true;
}

/*!
 * @brief Return the relative frame information for active review frame.
 *
 * @param [OUT] rel_frame_num   - the current relative frame number in review.
 * @param [OUT] rel_frame_min   - the current clip start for frame in review.
 * @param [OUT] rel_frame_max   - the current clip end for frame in review.
 */
void CameraVideo::refresh_rel_frame_info(int& rel_frame_num, int& rel_frame_min, int& rel_frame_max)
{
    // Set it all up to zeros.
    rel_frame_num = rel_frame_min = rel_frame_max = 0;

    // Are we connected to a camera?
    CameraMDI   *camera_ptr = qobject_cast<CameraMDI *>(parentWidget());
    CameraReview *camera_review_ptr = nullptr;
    if (camera_ptr != nullptr)
    {
        camera_review_ptr = camera_ptr->get_review_controls();
    }
    else
    {
        // Are we doing a playback?
        PlaybackMDI *playback_ptr = qobject_cast<PlaybackMDI *>(parentWidget());
        if (playback_ptr != nullptr)
        {
            camera_review_ptr = playback_ptr->getReviewControls();
        }
    }

    // If we have the review object get the relative frame information.
    if (camera_review_ptr)
    {
        camera_review_ptr->get_current_rel_frame_info(rel_frame_num, rel_frame_min, rel_frame_max);
    }
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rel_frame_num:" << rel_frame_num  << "rel_frame_min:" << rel_frame_min  << "rel_frame_max:" << rel_frame_max;
}

/*!
 * @brief Release a consumed image buffer back to the free pool.
 * @param image_ptr     The image buffer we're done with.
 */
void CameraVideo::release_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer>imageBufferPtr)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parentWidget());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No parent!";
        return;
    }

    // Pass the release request on.
    my_parent->release_buffer(std::move(imageBufferPtr));
}

void CameraVideo::render_start(void)
{
    QPixmap temp_pixmap;

    // Sanity check.
    if (render_scene != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "render_scene is not nullptr!";
    }

    // Create our objects.
    render_scene = new QGraphicsScene();
    render_scene->setSceneRect(scene_ptr->sceneRect());
    render_scene->setBackgroundBrush(Qt::black);

    // Create our pixmap item.
    copy_image(temp_pixmap);
    render_pixmap = new fiLayoutPixmap(temp_pixmap);
    render_pixmap->setPos(pixmap_item_ptr->pos());
    render_scene->addItem(render_pixmap);

    // Add in the overlays.
    for (int i = 0; i < overlay_list.count(); ++i)
    {
        fiOverlay *temp_overlay = new fiOverlay(nullptr, overlay_list.at(i));
        render_scene->addItem(temp_overlay);
        render_list.append(temp_overlay);
    }
}

void CameraVideo::render_add_image(QPixmap& my_pixmap)
{
    // Sanity check.
    if (render_pixmap == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "render_pixmap is nullptr!";
        return;
    }

    // Update the pixmap item.
    render_pixmap->setPixmap(my_pixmap);
}

void CameraVideo::render_update(uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata)
{
    int     i;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parentWidget());

    for (i = 0; i < render_list.count(); ++i)
    {
        // Have the overlay update it's items.
        render_list.at(i)->update_items(frame_num, info_ptr, metadata, my_parent);
    }
}

void CameraVideo::render_end(void)
{
    // Sanity check.
    if (render_scene == nullptr)
        return;

    while (render_list.count() > 0)
    {
        fiOverlay *overlay_ptr = render_list.takeAt(0);
        if (overlay_ptr != nullptr)
        {
            overlay_ptr->setParentItem(nullptr);
            render_scene->removeItem(overlay_ptr);
            overlay_ptr->deleteLater();
        }
    }

    delete render_pixmap;
    render_pixmap = nullptr;
    delete render_scene;
    render_scene = nullptr;
}

/*!
 * @brief Initialize the layout for the QGraphicsView widget. We want it as a child
 *        widget, and we want it to take all of our realestate.
 * @param video_widget_ptr  Pointer to the QGraphicsView widget.
 */
void CameraVideo::set_initial_layout(QWidget *video_widget_ptr)
{
    // Make sure our derived class did what it was supposed to
    if (!video_widget_ptr)
        return;

    // We want our play to stack vertically.
    QVBoxLayout *layout = new QVBoxLayout;

    // Add our video widget.
    layout->addWidget(video_widget_ptr);
    layout->setStretchFactor(video_widget_ptr, QSizePolicy::Expanding);
    layout->setContentsMargins(0, 0, 0, 0);

    // Set the layout into our window.
    setLayout(layout);
}

/*!
 * @brief Set the overlay padding for connected camera.
 *
 * @param addPadding    True if we want to add padding, false otherwise.
 */
void CameraVideo::set_overlay_padding(bool addPadding)
{
    // Need to set an overlay padding in the communications object.
    // This increase in buffer size (padding) will support a possible
    // increase in the size of an image when the overlay is merged.
    //
    // There still may be a need to change resources when overlays are
    // applied but (empirically) this padding seems to catch most of
    // the conditions.
    CameraMDI   *my_parent  = qobject_cast<CameraMDI *>(parentWidget());
    FI_comm     *comm_ptr   = MainWindow::GetInstance()->get_comm();
    if ((my_parent != nullptr) && (comm_ptr != nullptr))
    {
        uint32_t ip_addr = my_parent->get_ip_addr();
        if (addPadding)
        {
            // This could be set to different values depending on the ROI of
            // the camera setting, but the value 32K seemed to be pretty good
            // at catching most overlays (even image overlays).
            comm_ptr->set_overlay_padding(ip_addr, 0x8000);
        }
        else
        {
            comm_ptr->set_overlay_padding(ip_addr, 0);
        }
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to get camera/comm pointers.  addPadding:" << addPadding;
    }
}

void CameraVideo::reset_ae_reticle()
{
    QPointF reticle_pos;
    QRectF  reticle_rect;
    QRectF  scene_rect;
    QGraphicsItem *ae_reticle_parent = nullptr;

    // Sanity check.
    if (ae_reticle == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ae_reticle is nullptr!?";
        return;
    }

    // What is the parental relationship for the ae_reticle.
    ae_reticle_parent = ae_reticle->parentItem();
    if (ae_reticle_parent != nullptr)
    {
        // Remove any parentage.
        ae_reticle->setParentItem(nullptr);
    }

    // Make sure there is a scene pointer before we remove the
    // reticle.
    if (scene_ptr != nullptr)
        scene_ptr->removeItem(ae_reticle);
    delete ae_reticle;
    ae_reticle = nullptr;

    // Get the bounding rectangle for the scene.
    scene_rect = scene_ptr->sceneRect();

    // We want a reticle the user can drag around to specify the region
    // to use for calculating the custom white balance gains. Center the
    // reticle and make it 5% the size of the scene.
    reticle_rect.setHeight(scene_rect.height() * 0.05);
    reticle_rect.setWidth(scene_rect.width() * 0.05);

    // Have our reticle start in the middle of the image.
    reticle_pos.setX((scene_rect.width() - reticle_rect.width()) / 2);
    reticle_pos.setY((scene_rect.height() - reticle_rect.height()) / 2);

    // Allocate our reticle object.
    ae_reticle = new fiRectItem(reticle_rect);
    ae_reticle->setPen(QPen(Qt::white, 2));

    // Re-establish a parental relationship if one was set before.
    if (ae_reticle != nullptr)
    {
        ae_reticle->setParentItem(ae_reticle_parent);
    }

    // Make sure the user can grab the rectangle and move it around.
    ae_reticle->setFlag(QGraphicsItem::ItemIsMovable, !overlay_enabled);

    // Add it to the scene.
    scene_ptr->addItem(ae_reticle);

    // Now move the reticle into the scene.
    ae_reticle->setPos(reticle_pos);

    // Make sure we are on top of the (default) scene. The default Z-value is 0.
    ae_reticle->setZValue(1);

    // Show the reticle.
    ae_reticle->setVisible(true);

    // We want to enable the reticle, which means "do not track".
    ae_reticle->setEnabled(true);

}

void CameraVideo::reset_vt_reticle(const QRect& new_rect)
{
    QPointF reticle_pos;
    QRectF  reticle_rect;
    QRectF  scene_rect;
    QGraphicsItem *vt_reticle_parent = nullptr;

    // Sanity check.
    if (vt_reticle == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "vt_reticle is nullptr!?";
        return;
    }

    // What is the parental relationship for the vt_reticle.
    vt_reticle_parent = vt_reticle->parentItem();
    if (vt_reticle_parent != nullptr)
    {
        // Remove any parentage.
        vt_reticle->setParentItem(nullptr);
    }

    // Make sure there is a scene pointer before we remove the
    // reticle.
    if (scene_ptr != nullptr)
        scene_ptr->removeItem(vt_reticle);
    delete vt_reticle;
    vt_reticle = nullptr;

    // Get the bounding rectangle for the scene.
    scene_rect = scene_ptr->sceneRect();

    if (new_rect.isEmpty())
    {
        qreal r_width;
        qreal r_height;

        // We want a reticle the user can drag around to specify the region
        // to use for calculating the custom white balance gains. Center the
        // reticle and make it 5% the size of the scene.
        r_height = scene_rect.height() * 0.05;
        r_width = scene_rect.width() * 0.05;
        if ((int)(r_height * r_width) > MAX_VT_AREA)
        {
            // Make this square if we can, but stay within the image.
            if (r_height > (MAX_VT_AREA / 2))
                r_height = MAX_VT_AREA / 2;
            if (r_height >= scene_rect.height())
                r_height = scene_rect.height() - 2;
            if (r_width > (MAX_VT_AREA / 2))
                r_width = MAX_VT_AREA / 2;
            if (r_width >= scene_rect.width())
                r_width = scene_rect.width() - 2;
        }

        reticle_rect.setHeight(r_height);
        reticle_rect.setWidth(r_width);

        // Have our reticle start in the middle of the image.
        reticle_pos.setX((scene_rect.width() - reticle_rect.width()) / 2);
        reticle_pos.setY((scene_rect.height() - reticle_rect.height()) / 2);
    }
    else
    {
        // Use the new rectangles coordinates/dimensions.
        reticle_rect.setHeight(new_rect.height());
        reticle_rect.setWidth(new_rect.width());
        reticle_pos.setX(new_rect.left());
        reticle_pos.setY(new_rect.top());
    }

    // Allocate our reticle object.
    vt_reticle = new fiRectItem(reticle_rect);

    static_cast<fiRectItem *>(vt_reticle)->setMaxArea(MAX_VT_AREA);
    vt_reticle->setPen(QPen(Qt::white, 2));

    // Re-establish a parental relationship if one was set before.
    if (vt_reticle_parent != nullptr)
    {
        vt_reticle->setParentItem(vt_reticle_parent);
    }

    // Make sure the user can grab the rectangle and move it around.
    vt_reticle->setFlag(QGraphicsItem::ItemIsMovable, !overlay_enabled);

    // Add it to the scene.
    scene_ptr->addItem(vt_reticle);

    // Now move the reticle into the scene.
    vt_reticle->setPos(reticle_pos);

    // Make sure we are on top of the (default) scene. The default Z-value is 0.
    vt_reticle->setZValue(1);

    // Make it visible.
    vt_reticle->setVisible(true);

    // Since it's visible, it must be enabled.
    vt_reticle->setEnabled(true);

}

/*!
 * @brief Establishes a parental relationship with a reticle graphics item.
 * @param parent The layout pixmap for active image.
 */
void CameraVideo::set_reticle_parent(QGraphicsRectItem *reticle_ptr, fiLayoutPixmap *parent)
{
    fiRectItem *fi_rect = nullptr;

    if (reticle_ptr)
        fi_rect = static_cast<fiRectItem *>(reticle_ptr);

    // May need to adjust the parental relationship with the vt_reticle.
    if (fi_rect)
    {
        // Don't install if one already exists.
        if (fi_rect->parentItem() == nullptr)
        {
            // Establish the parental relationship if the parent is allocated.
            if (parent)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Installing pixmap_item_ptr as parent.";
                fi_rect->setParentItem(parent);
            }
        }
    }
}

void CameraVideo::show_ae_reticle(bool visible)
{
    // Sanity check.
    if (ae_reticle == nullptr)
        return;

    // Show/hide the reticle.
    ae_reticle->setVisible(visible);
}

void CameraVideo::show_vt_reticle(bool visible)
{
    // Sanity check.
    if (vt_reticle == nullptr)
        return;

    // Show/hide the reticle.
    vt_reticle->setVisible(visible);
}

/*!
 * @brief Update the size of the displayed image based on the scaling information.
 */
const QRectF CameraVideo::size_image(void)
{
    QRectF scene_rect = QRectF();
    QPointF adj_point;
    QPointF pixmap_scene_pos;

    int bottom_pinned = 0;      // Keeps track of a count of bottom pinned overlays.

    // Sanity check.
    if (pixmap_item_ptr == nullptr)
        return scene_rect;

    // Initialize our return value.
    scene_rect = pixmap_item_ptr->boundingRect();

    // Grab the scene position for the image frame.
    pixmap_scene_pos = pixmap_item_ptr->scenePos();

    // Are we concerned about overlays?
    if (overlay_enabled)
    {
        int     i;

        // Build up the scene area we want visible.
        for (i = 0; i < overlay_list.count(); ++i)
        {
            // Account for any overlays.
            QRectF  ovr_rect = overlay_list.at(i)->boundingRect();

            // Make the overlay adjustment and accumulate the overlay space into scene rect.
            ovr_rect.moveTo(overlay_list.at(i)->pos());
            scene_rect = scene_rect.united(ovr_rect);

            // Do we have a bottom pinned overlay?
            if (overlay_list.at(i)->isPinned())
            {
                // Only care about counting the number of bottom overlays.
                if (!overlay_list.at(i)->isPinnedAtTop())
                {
                    bottom_pinned++;
                }
            }
        }
    }
    else
    {
        // Ticket #350 conditionalize this moveto for the scene to only happen if not overlay_enabled.
        // when it always happens the shifting of the display can occur during playback scrubbing.
        // If no overlay then setup the position for the image in the scene.
        scene_rect.moveTo(pixmap_item_ptr->pos());
    }

    // Make sure we have an OK rectangle (normalized creates rectangle with non-negative width and height).
    scene_rect = scene_rect.normalized();

    // How much to we need everything to adjust?
    // Note: These adjustment happens when overlays move outside of the image frame
    // and the scene rectangle changes origin.
    adj_point = scene_rect.topLeft();
    if (adj_point != QPointF(0, 0))
    {
        // Move the overall scene rectangle to sit at [0, 0].
        scene_rect.moveTo(0.0, 0.0);

        // Adjust the pixmap.
        pixmap_item_ptr->moveBy(-adj_point.x(), -adj_point.y());

        // Are we concerned about overlays?
        if (overlay_enabled)
        {
            // Adjust all the overlays.
            for (int i = 0; i < overlay_list.count(); ++i)
            {
                overlay_list.at(i)->moveBy(-adj_point.x(), -adj_point.y());
            }
        }
    }

    // Ticket #484:
    // Handle overlay rendering conditions that cause problems with the image display.
    if (overlay_enabled)
    {
        // The following condition is where there are overlays above the pixmap and nothing
        // pinned to the bottom. The adjustment made ensures the overlay height(s) is/are
        // accomodated in the scene. Without this the pixmap renders truncated at bottom in the scene.
        // Note: The reason why is because the top pinned overlays end up with 0,0 as their
        // left,top position and even though the layout engine places them properly depending on
        // when they were added to the scene, the scene_rect.united above doesn't have the true
        // rectangle scene position for all overlays to properly calculate the unionization of
        // the scene object rectangles. The code below ensures that the bottom right of the scene
        // rect accomodates all overlays positioned above it.
        if (pixmap_scene_pos.y() > 0 && (bottom_pinned == 0))
        {
            // We only make an adjustment on the bottom if there were overlays above it (and we know nothing pinned on bottom).
            if (adj_point.y() < 0)
            {
                QPointF adjust_bottom_right(scene_rect.width(), (scene_rect.bottom() + pixmap_scene_pos.y()));
                scene_rect.setBottomRight(adjust_bottom_right);
            }
        }
    }

    // Return our scene's rectangular area.
    return scene_rect;
}

/*!
 * @brief Propogate the fact that we want to stop our video processing.
 */
void CameraVideo::stop_video(void)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parentWidget());

    // Make sure everyone knows we want to stop.
    thread_stop = true;

    // Make sure we have a parent object.
    if (my_parent == nullptr)
        return;

    // Are we in shutdown?
    if (my_parent->shutdown_in_progress())
        return;

    // If there is a thread running, let's wait for it to finish.
    if (image_thread.isRunning())
    {
        image_thread.waitForFinished();
    }

    // The video is not going to be displaying anymore - is there an active
    // image buffer we have in play?  If so it must be released.
    if (activeImageBufferPtr)
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Releasing activeImageBufferPtr.";
        // Release the ownership of this back to the pool.
        release_image(std::move(activeImageBufferPtr));
    }
}

void CameraVideo::update_overlay(uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata)
{
    int     i;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parentWidget());

    // Ask for the relative frame information.
    int relative_frame = 0;
    int relative_min_frame = 0;
    int relative_max_frame = 0;

    // Ask for the review parameters so we can provide some guidance to overlay object
    // about relative frame information.
    refresh_rel_frame_info(relative_frame, relative_min_frame, relative_max_frame);

    for (i = 0; i < overlay_list.count(); ++i)
    {
        // Update the overlay relative frame range.
        overlay_list.at(i)->update_rel_frame_range(relative_min_frame, relative_max_frame);

        // Have the overlay update it's items.
        overlay_list.at(i)->update_items(frame_num, info_ptr, metadata, my_parent);
    }
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief We are closing the window, so make sure we set our flag to
 *        stop any further threaded processing.
 *
 * @param event     Pointer to the event structure.
 */
void CameraVideo::closeEvent(QCloseEvent *event)
{
    // Tell the threads to stop.
    thread_stop = true;

    // Accept the event.
    event->accept();

    // Pass it on.
    QWidget::closeEvent(event);
}

/*!
 * @brief Our conversion event, posted when the conversion thread
 *        completes it's processing.
 *
 * @param event     Pointer to the event structure.
 */
void CameraVideo::customEvent(QEvent *event)
{
    QPixmap working_pixmap;     // A "working" pixmap object to convert QImages to something we can display.

    // Is this an fiRectEvent?
    if (event->type() == fiRectItem::CHANGE_EVENT)
    {
        emit videoRectChanged();
        return;
    }

    // Only respond to our custom event. We also want to leave if we're in shutdown.
    if (event->type() != CONVERSION_EVENT)
    {
        QWidget::customEvent(event);
        return;
    }

    // !! IMPORTANT !!
    // At this point the live view image was converted, we have our QImage
    // to display on the screen and we can release the activeImageBuffer
    // back to its resource pool.
    if (!activeImageBufferPtr)
    {
        QLOG_DEBUG() << "[RESOURCE EMPTY][CameraVideo::customEvent()][activeImageBufferPtr]";
    }
    else
    {
        QString myInfo = QStringLiteral("[CameraVideo::CONVERSION_EVENT]");
        //QLOG_DEBUG() << "[RESOURCE RELEASE][CameraVideo::customEvent()][activeImageBufferPtr]";
        activeImageBufferPtr->LogMyDataId(&myInfo);

        // Release the ownership of this back to the pool.
        release_image(std::move(activeImageBufferPtr));
    }

    // Cast the event to the ConversionEvent type.
    ConversionEvent *our_event = dynamic_cast<ConversionEvent *>(event);
    if (our_event == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Casted to nullptr?";
        QWidget::customEvent(event);
        return;
    }

    // Are we done?
    if (thread_stop)
        goto event_bail;

    // Get the image as a pixmap.
    if (our_event->get_image_ptr() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "our_event has a null image_ptr?";
        goto event_bail;
    }
    working_pixmap.convertFromImage(*our_event->get_image_ptr());
    if (working_pixmap.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "QImage to QPixmap failed!";
        goto event_bail;
    }

    // Do we want to generate a histogram for the image?
    if (histogram_enabled)
    {
        // Caculate a histogram.
        histo_proc.calc(working_pixmap);
    }

    // Display the image for the user.
    display_image(working_pixmap);

event_bail:

    // Is there an image pointer?
    if (our_event->get_image_ptr() != nullptr)
    {
        // All done with the QImage, so put it on the available list.
        QMutexLocker my_locker(&avail_mutex);
        avail_list.append(our_event->get_image_ptr());
    }
}

/*!
 * @brief The window has resized. We need to update the image dimensions.
 * @param event     Pointer to the event structure.
 */
void CameraVideo::resizeEvent(QResizeEvent *event)
{
    // Pass it up.
    QWidget::resizeEvent(event);

    // Update our image scaling to fit the window.
    if (fit_to_window || qFuzzyCompare(scale_factor, SCALE_FIT))
    {
        QRectF  scene_rect;

        // Adjust the scaling if necessary.
        scene_rect = size_image();

        // Adjust the scene ROI.
        scene_ptr->setSceneRect(scene_rect);
        view_ptr->ensureVisible(scene_rect);
    }

    // Rescale the image.
    rezoom();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void CameraVideo::do_add_overlay(void)
{
    fiOverlay   *overlay_ptr;
    int relative_frame = 0;
    int relative_min_frame = 0;
    int relative_max_frame = 0;

    // Before creating a new overlay, let's cleanup the existing ones.
    cleanup_overlays();

    // Create a new overlay object.
    overlay_ptr = new fiOverlay;
    if (overlay_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Error allocating overlay object!";
        return;
    }

    // Ask for the review parameters so we can provide some guidance to overlay object
    // about relative frame information.
    refresh_rel_frame_info(relative_frame, relative_min_frame, relative_max_frame);

    // Give the overlay object a default based on whatever the current frame might
    // be at the time of this overlay creation.  This is only used if no relative frame
    // information exists in the settings for the overlay.
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Calculated relative_frame:" << relative_frame << "relative_min_frame:" << relative_min_frame << "relative_max_frame:" << relative_max_frame;
    overlay_ptr->set_default_rel_frame_info(relative_frame, relative_min_frame, relative_max_frame);

    // Append the overlay to our list.
    overlay_list.append(overlay_ptr);
    connect(overlay_ptr, &fiOverlay::remove_overlay, this, &CameraVideo::do_delete_overlay);
    connect(overlay_ptr, &fiOverlay::pin_to_bottom, this, &CameraVideo::do_pin2bottom_overlay);
    connect(overlay_ptr, &fiOverlay::pin_to_top, this, &CameraVideo::do_pin2top_overlay);
    connect(overlay_ptr, &fiOverlay::need_update, this, &CameraVideo::do_update_overlay);
    connect(overlay_ptr, &fiOverlay::overlay_moved, this, &CameraVideo::do_resize_overlay);

    // Add it to our scene pinning it to the top by default.
    overlay_ptr->setPinned(true, true, pixmap_item_ptr->boundingRect().width());
    scene_ptr->addItem(overlay_ptr);
    layout_ptr->insertItem(0, overlay_ptr);
    layout_ptr->setSpacing(0);
    layout_ptr->setContentsMargins(0, 0, 0, 0);

    // Make sure we are on top of the (default) scene. The default Z-value is 0.
    overlay_ptr->setZValue(1);

    // Have the user select stuff.
    overlay_ptr->do_select_item();

    // Update the overlay padding. We only need to do this the first
    // time we add an overlay.
    if (overlay_list.count() == 1)
        set_overlay_padding(true);

    // Zoom out so the user can see the whole picture.
    QTimer::singleShot(ZOOM_DELAY, this, &CameraVideo::rezoom);
}

void CameraVideo::do_delete_overlay(void)
{
    int i;
    fiOverlay *delete_sender = qobject_cast<fiOverlay *>(sender());

    // Walk through the list and find the one we're looking for.
    for (i = 0; i < overlay_list.count(); ++i)
    {
        // Is this a match?
        if (overlay_list.at(i) == delete_sender)
        {
            // Remove the overlay from our list.
            fiOverlay *overlay_ptr = overlay_list.takeAt(i);

            // Remove the overlay from the scene.
            scene_ptr->removeItem(overlay_ptr);

            // Delete the overlay object.
            delete overlay_ptr;

            // And done.
            break;
        }
    }

    // Clean up the remaining overlays.
    cleanup_overlays();

    // Do we have any overlays?
    if (overlay_list.isEmpty())
    {
        // No need for any padding.
        set_overlay_padding(false);
    }
}

void CameraVideo::do_pin2bottom_overlay(void)
{
    int i;
    fiOverlay *pin_sender = qobject_cast<fiOverlay *>(sender());

    // Clean up the overlays.
    cleanup_overlays();

    // Walk through the list and find the one we're looking for.
    for (i = 0; i < overlay_list.count(); ++i)
    {
        // Is this a match?
        if (overlay_list.at(i) == pin_sender)
        {
            // Get the overlay from our list.
            fiOverlay *overlay_ptr = overlay_list.at(i);

            // Pin the overlay.
            pin_overlay(overlay_ptr, false);

            // Zoom out so the user can see the whole picture.
            QTimer::singleShot(ZOOM_DELAY, this, &CameraVideo::rezoom);

            // And done.
            break;
        }
    }
}

void CameraVideo::do_pin2top_overlay(void)
{
    int i;
    fiOverlay *pin_sender = qobject_cast<fiOverlay *>(sender());

    // Clean up the overlays.
    cleanup_overlays();

    // Walk through the list and find the one we're looking for.
    for (i = 0; i < overlay_list.count(); ++i)
    {
        // Is this a match?
        if (overlay_list.at(i) == pin_sender)
        {
            // Remove the overlay from our list.
            fiOverlay *overlay_ptr = overlay_list.at(i);

            // Pin the overlay.
            pin_overlay(overlay_ptr, true);

            // Zoom out so the user can see the whole picture.
            QTimer::singleShot(ZOOM_DELAY, this, &CameraVideo::rezoom);

            // And done.
            break;
        }
    }
}

void CameraVideo::do_resize_overlay(void)
{
    QRectF  scene_rect;
    fiOverlay *update_sender = qobject_cast<fiOverlay *>(sender());

    // Remove the overlay from the layout.
    layout_ptr->removeItem(update_sender);

    // Adjust the scaling if necessary.
    scene_rect = size_image();

    // Adjust the scene ROI.
    scene_ptr->setSceneRect(scene_rect);
    view_ptr->ensureVisible(update_sender);
}

void CameraVideo::do_update_overlay(void)
{
    int i;
    fiOverlay *update_sender = qobject_cast<fiOverlay *>(sender());

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "update_sender:" << update_sender;

    // Walk through the list and find the one we're looking for.
    for (i = 0; i < overlay_list.count(); ++i)
    {
        // Is this a match?
        if (overlay_list.at(i) == update_sender)
        {
            // Are we connected to a camera?
            CameraMDI   *camera_ptr = qobject_cast<CameraMDI *>(parentWidget());
            if (camera_ptr != nullptr)
            {
                // Use the camera data to update the overlay items.
                update_sender->update_items(camera_ptr->get_current_frame_num(), camera_ptr->get_info_widget(), nullptr, camera_ptr);
            }
            else
            {
                // Are we doing a playback?
                PlaybackMDI *playback_ptr = qobject_cast<PlaybackMDI *>(parentWidget());
                if (playback_ptr == nullptr)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unknown parent!";
                    return;
                }

                // Update the overlay items with the playback data.
                update_sender->update_items(playback_ptr->getLastFrame(), playback_ptr->getCameraInfo(), playback_ptr->getMetadata(), nullptr);
            }

            // And done.
            break;
        }
    }

}

void CameraVideo::do_rotate_counter()
{
    // Rotate the image 90-degrees to the left (counter-clockwise).
    rot_angle -= 90.0;
    if (rot_angle < -359.0)
        rot_angle = 0.0;

    // Re-zoom updates all the transforms.
    rezoom();

    // Did we do the rotation?
    if (!qFuzzyCompare(view_ptr->rot_angle(), rot_angle))
    {
        do_zoom_to_fit();
    }
}

void CameraVideo::do_rotate_clockwise()
{
    // Rotate the image 90-degrees to the right (clockwise).
    rot_angle += 90.0;
    if (rot_angle > 359.0)
        rot_angle = 0.0;

    // Re-zoom updates all the transforms.
    rezoom();

    // Did we do the rotation?
    if (!qFuzzyCompare(view_ptr->rot_angle(), rot_angle))
    {
        do_zoom_to_fit();
    }
}

/*!
 * @brief This makes a given pixel in the image, equal to a pixel on the screen.
 */
void CameraVideo::do_zoom_1_to_1(void)
{
    if (!qFuzzyCompare(scale_factor, SCALE_NORMAL))
    {
        // Setup our flags.
        fit_to_window = false;
        scale_factor = SCALE_NORMAL;

        // We'll start out by resetting the transform values.
        // This has the side effect of getting us to 1:1.
        view_ptr->resetTransform();

        // Make sure we restore our rotation.
        view_ptr->rotate(rot_angle);

        // May need to adjust our aiming cursor.
        enable_aiming_cursor(MainWindow::GetInstance()->get_view_aiming_cursor()->isChecked());
    }
}

/*!
 * @brief Zoom in by a fixed ratio.
 */
void CameraVideo::do_zoom_in(void)
{
    double factor;

    factor = qPow(view_ptr->get_base_factor(), ZOOM_ANGLE);
    view_ptr->do_zooming(factor);
}

/*!
 * @brief Zoom out by a fixed ratio but only as far as 1 to one.
 */
void CameraVideo::do_zoom_out(void)
{
    double factor;

    factor = qPow(view_ptr->get_base_factor(), -ZOOM_ANGLE);
    view_ptr->do_zooming(factor);
}

/*!
 * @brief Zoom to make the image fit within the window, while
 *        maintaining the image's aspect ratio.
 */
void CameraVideo::do_zoom_to_fit(void)
{
    QRectF  scene_rect;

    if ((fit_to_window == true) || (!qFuzzyCompare(scale_factor, SCALE_FIT)))
    {
        fit_to_window = false;
        scale_factor = SCALE_FIT;
    }

    // Adjust the scaling if necessary.
    scene_rect = size_image();

    // We'll start out by resetting the transform values.
    // This has the side effect of getting us to 1:1.
    view_ptr->resetTransform();

    // Make sure we restore our rotation.
    view_ptr->rotate(rot_angle);

    // Make sure we fit the scene in the view port.
    view_ptr->fitInView(scene_rect, Qt::KeepAspectRatio);

    // Adjust the scene ROI.
    scene_ptr->setSceneRect(scene_rect);
    view_ptr->ensureVisible(scene_rect);

    // May need to adjust our aiming cursor.
    enable_aiming_cursor(MainWindow::GetInstance()->get_view_aiming_cursor()->isChecked());
}

/*!
 * @brief Zoom to make the image fit within the window, ignoring
 *        the image's aspect ratio.
 */
void CameraVideo::do_zoom_to_window(void)
{
    QRectF  scene_rect;

    if ((fit_to_window == false) || (!qFuzzyCompare(scale_factor, SCALE_FIT)))
    {
        fit_to_window = true;
        scale_factor = SCALE_FIT;
    }

    // Adjust the scaling if necessary.
    scene_rect = size_image();

    // We'll start out by resetting the transform values.
    // This has the side effect of getting us to 1:1.
    view_ptr->resetTransform();

    // Make sure we restore our rotation.
    view_ptr->rotate(rot_angle);

    // Make sure we fit the scene in the view port.
    view_ptr->fitInView(scene_rect, Qt::IgnoreAspectRatio);

    // Adjust the scene ROI.
    scene_ptr->setSceneRect(scene_rect);
    view_ptr->ensureVisible(scene_rect);

    // May need to adjust our aiming cursor.
    enable_aiming_cursor(MainWindow::GetInstance()->get_view_aiming_cursor()->isChecked());
}

/*!
 * @brief Slot to catch the histogram fault.
 */
void CameraVideo::histogram_fault()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
    enable_histogram(false);
}

/*!
 * @brief Our [graphics view object](@ref fiGraphicsView) has indicated
 *       that the user has changed the "zoom" level.
 */
void CameraVideo::image_zoomed(void)
{
    scale_factor = view_ptr->get_zoom_factor();

    // Update the selected zoom.
    MainWindow::GetInstance()->check_zoom_actions(get_zoom_setting());
}

void CameraVideo::parent_closing()
{
    //QLOG_DEBUG() << __PRETTY_FUNCTION__;
    stop_video();
}

void CameraVideo::rezoom(void)
{
    // Rescale the image according to the current zoom setting.
    switch (get_zoom_setting())
    {
    case FASMOTION_SETTINGS::ZOOM_1_TO_1:
        do_zoom_1_to_1();
        break;
    case FASMOTION_SETTINGS::ZOOM_TO_FIT:
        do_zoom_to_fit();
        break;
    case FASMOTION_SETTINGS::ZOOM_TO_WINDOW:
        do_zoom_to_window();
        break;
    default:
    case FASMOTION_SETTINGS::ZOOM_UNKNOWN:
    case FASMOTION_SETTINGS::ZOOM_IN:
    case FASMOTION_SETTINGS::ZOOM_OUT:
        break;
    }
}

/*!
 * @brief Save the user specified settings to the non-volatile parameter storage.
 */
void CameraVideo::save_settings(void)
{
    QSettings settings;

    // Save the aiming cursor setting.
    settings.beginGroup("Video");
    settings.setValue("aiming", MainWindow::GetInstance()->get_view_aiming_cursor()->isChecked());
    settings.endGroup();

    // Remove the overlay settings to clear out any old values.
    settings.remove(QStringLiteral("Overlays"));

    // Save the new values.
    if (overlay_list.count() > 0)
    {
        int     i;

        settings.beginGroup(QStringLiteral("Overlays"));
        settings.setValue(QStringLiteral("count"), overlay_list.count());
        for (i = 0; i < overlay_list.count(); ++i)
            overlay_list.at(i)->save_settings(settings, i);
        settings.endGroup();
    }

    // Add the reticle information so we can use it again later.
    if (ae_reticle != nullptr)
    {
        QRectF save_rect = static_cast<fiRectItem *>(ae_reticle)->adjrect();
        QPointF save_pos = ae_reticle->pos();

        settings.beginGroup(QStringLiteral("ae_reticle"));
        settings.setValue(QStringLiteral("enabled"), ae_reticle->isEnabled());
        settings.setValue(QStringLiteral("visible"), ae_reticle->isVisible());
        settings.setValue(QStringLiteral("ROI"), save_rect);
        settings.setValue(QStringLiteral("pos"), save_pos);
        settings.endGroup();
    }

    // Add the video trigger reticle information so we can use it again later.
    if (vt_reticle != nullptr)
    {
        QRectF save_rect = static_cast<fiRectItem *>(vt_reticle)->adjrect();
        QPointF save_pos = vt_reticle->pos();

        settings.beginGroup(QStringLiteral("vt_reticle"));
        settings.setValue(QStringLiteral("enabled"), vt_reticle->isEnabled());
        settings.setValue(QStringLiteral("visible"), vt_reticle->isVisible());
        settings.setValue(QStringLiteral("ROI"), save_rect);
        settings.setValue(QStringLiteral("pos"), save_pos);
        settings.endGroup();
    }
}

