/*!
 * @file camera_video.h
 * @brief This is the header file for the CameraVideo class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CAMERA_VIDEO_H
#define CAMERA_VIDEO_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QPixmap>
#include <QWidget>
#include <stdint.h>

#include "fi_histo_proc.h"
#include "fsp_socket.h"

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QAction;
class QCloseEvent;
class QContextMenuEvent;
class QGraphicsScene;
class QGraphicsLinearLayout;
class QGraphicsLineItem;
class QGraphicsRectItem;
class QImage;
class QResizeEvent;
class QWheelEvent;

class CameraInfo;
class fiGraphicsView;
class fiLayoutPixmap;
class fiOverlay;
class fiRectItem;
class PlaybackMetaInfo;

namespace CAMERA_VIDEO
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
typedef QList< fiOverlay * >    OVERLAY_LIST;

//! Threshold for colorizing pixels.
const int   COLORIZE_THRESHOLD          = 250;
}

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class CameraVideo : public QWidget
{
    Q_OBJECT

public:
    explicit CameraVideo(QWidget *parent = nullptr);
    ~CameraVideo();

    void    colorize(bool enable)               { do_colorize = enable; }
    void    copy_image(QPixmap& dest_pixmap);
    void    display_image(const QPixmap& my_pixelmap);
    void    enable_aiming_cursor(bool do_enable);
    void    enable_info_cursor(bool do_enable);
    void    enable_overlay(bool do_enable);
    void    enable_ae_reticle(bool do_enable);
    void    enable_focus_assist(bool do_enable) { do_focus_assist = do_enable; }
    void    enable_histogram(bool do_enable)    { histogram_enabled = do_enable; }
    void    enable_vt_reticle(bool do_enable);
    QRect   get_ae_rect();
    fiRectItem *get_ae_reticle();
    fiHistoProc *get_histogram()            { return &histo_proc; }
    QRect   get_image_rect();
    QRect   get_vt_rect();
    fiRectItem *get_vt_reticle();
    int     get_zoom_setting();
    bool    has_received_1_image()          { return received_at_least_1_image; }
    bool    is_ae_reticle_enabled();
    bool    is_ae_reticle_visible();
    bool    is_vt_reticle_enabled();
    bool    is_vt_reticle_visible();
    void    need_rezoom()                 { pending_rezoom = true; }
    void    receive_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer>imageBufferPtr);
    void    reset_ae_reticle();
    void    reset_vt_reticle(const QRect& new_rect = QRect());
    void    set_reticle_parent(QGraphicsRectItem  *reticle_ptr, fiLayoutPixmap *parent);
    void    show_ae_reticle(bool visible);
    void    show_vt_reticle(bool visible);
    void    stop_video();

    QAction         *get_action_overlay() { return action_overlay; }
    QGraphicsScene  *get_scene()          { return scene_ptr; }
    bool    has_overlay()                 { return !overlay_list.isEmpty(); }
    void    update_overlay(uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata);

    void    render_start();
    void    render_add_image(QPixmap& my_pixmap);
    void    render_update(uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata);
    QGraphicsScene  *render_get_scene()   { return render_scene; }
    void    render_end();

public slots:
    void    do_zoom_1_to_1();
    void    do_zoom_in();
    void    do_zoom_out();
    void    do_zoom_to_fit();
    void    do_zoom_to_window();

    void    do_rotate_counter();
    void    do_rotate_clockwise();

    void    do_add_overlay();
    void    rezoom();
    void    save_settings();

signals:
    void    videoRectChanged();

protected:
    virtual void    closeEvent(QCloseEvent *event);
    virtual void    customEvent(QEvent *event);
    virtual void    resizeEvent(QResizeEvent *event);

private slots:
    void    do_delete_overlay();
    void    do_pin2bottom_overlay();
    void    do_pin2top_overlay();
    void    do_resize_overlay();
    void    do_update_overlay();

    void    histogram_fault();
    void    image_zoomed();
    void    parent_closing();

private:
    static void convert_image(QWidget *parent);

    void    cleanup_overlays();
    void    colorize_image(QImage *image_ptr);
    void    focus_assist(QImage *image_ptr);
    void    load_settings();
    void    load_reticle_settings(QGraphicsRectItem **reticle_ptr, QString settings_str);
    void    pin_overlay(fiOverlay *overlay_ptr, bool to_top);
    void    refresh_rel_frame_info(int& rel_frame_num, int& rel_frame_min, int& rel_frame_max);
    void    release_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer>imageBufferPtr);
    void    set_initial_layout(QWidget *video_widget_ptr);
    void    set_overlay_padding(bool addPadding);
    const QRectF    size_image();

private:
    QList<QImage *>     avail_list;                     //!< Our list of available QImage objects.
    QMutex              avail_mutex;                    //!< Protect the available list.
    volatile bool       thread_stop;                    //!< Flag to tell our processing thread to stop.

    //! When processing live images this stores the active one in the stream
    //! that will be used for converting to QImage for storage on the image_list.
    //! This had to be created this way because std::unique_ptr cannot be conveyed
    //! through the QtConcurrent object's run method because QtConcurrent moves
    //! objects using a copy paradigm and std::unique_ptr objects does not support
    //! a copy or assignment interface therefore the usage fails.  We created this
    //! object in this class to be the unique pointer being consumed in convert_image
    //! interface.
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> activeImageBufferPtr;

    bool                received_at_least_1_image;

    bool                fit_to_window;                  //!< Zoom flag telling us to fit (e.g. "scale") the image to the window dimensions.
    bool                histogram_enabled;
    bool                pending_rezoom;                 //!< True to schedule a rezoom operation on the next received frame.
    double              scale_factor;                   //!< Zoom scale factor.
    double              rot_angle;                      //!< Rotation angle.

    QAction             *action_overlay;                //!< Action to add overlay(s).
    bool                overlay_enabled;                //!< Flag to show/hide overlay(s).
    CAMERA_VIDEO::OVERLAY_LIST  overlay_list;           //!< Our list of overlay object(s).

    // The display controls/variables.
    QGraphicsLinearLayout   *layout_ptr;                //!< Our layout object.
    QGraphicsScene  *scene_ptr;                         //!< Pointer to our Qt "scene" object.
    fiGraphicsView  *view_ptr;                          //!< Our graphics view object.
    fiLayoutPixmap  *pixmap_item_ptr;                   //!< The pixmap object pointer, used to manipulate the image we're displaying.
    QGraphicsRectItem   *ae_reticle;                    //!< The region we'll be sampling for use with auto exposure.
    QGraphicsRectItem   *vt_reticle;                    //!< The region we'll be sampling for use with video trigger.

    QGraphicsLineItem   *horz_line;
    QGraphicsLineItem   *vert_line;

    CAMERA_VIDEO::OVERLAY_LIST  render_list;            //!< Our list of overlay object(s).
    fiHistoProc             histo_proc;
    fiLayoutPixmap          *render_pixmap;
    QGraphicsScene          *render_scene;              //!< Pointer to our Qt "scene" object.

    QFuture<void> image_thread;                         //!< Our image processing thread "handle".

    static bool do_colorize;                            //!< True to colorize saturated pixels.
    static bool do_focus_assist;                        //!< Run through the focus assist algorithm.
};

#endif // CAMERA_VIDEO_H
