/*!
 * @file client_math.cpp
 * @brief This file contains the implementation of the ClientMath
 *        class. This class is a collection of routines used to calculate
 *        common values used by the application.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QDate>
#include <QString>
#include <QStringList>
#include <QTime>
#include <qmath.h>

#include "client_math.h"
#include "fi_comm.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"

namespace CLIENT_MATH
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The number of bytes in an SSD sector.
const uint32_t  SSD_SECTOR                  = 512;

//! The number of bytes in an SSD page.
const uint32_t  SSD_PAGE                    = 8 * SSD_SECTOR;

//! The number of bytes in a QuadWord.
const uint32_t  QUADWORD_BYTES              = 16;

//! The number of bytes in a HexWord.
const uint32_t HEXWORD_BYTES                = 64;

//! The per-frame overhead required for frame read-out, in microseconds.
const uint32_t  DEFAULT_FRAME_OVERHEAD      = 6;

//! The default overhead for FasFireLR metamap file.
const uint32_t DEFAULT_METAMAP_OVERHEAD     = 4096;

//! How many QWORDS are in a gigabyte.
const uint32_t  GBYTES_TO_QWORDS            = (64 * 1024 * 1024);

//! The conversion factor from sectors to QuadWords.
const uint32_t  SSD_SECTOR_LEN_QWORDS       = SSD_SECTOR / QUADWORD_BYTES;

//! The number of bytes in the SeriesNC camera frame's metadata.
const int SNC_FRAME_METADATA_LEN_BYTES      = 64;

//! The number of bytes in the Series3 camera frame's metadata.
const int S3_FRAME_METADATA_LEN_BYTES       = 32;

//! In the Series3 cameras, the FPGA packs 128 frames of
//! metadata in a page of RAM (4096 / 32).
const int S3_FRAME_METADATA_PER_PAGE        = 128;

//! Conversion to use when going from seconds to microseconds.
const uint32_t  USEC_CONVERSION             = 1000000;

}   // End of namespace CLIENT_MATH.

using namespace CLIENT_MATH;

/*!
 * @brief Constructor for the ClientMath class.
 */
ClientMath::ClientMath()
{
}

/*!
 * @brief Calculate the maximum shutter speed based upon a frame rate.
 *
 * @param framerate         Frame rate to use to calculate the maximum shutter speed.
 * @param frame_overhead    Per-frame overhead.
 *
 * @return 0 if the roi not valid, or the max frame rate if it is.
 */
uint32_t ClientMath::calc_max_exposure(uint32_t framerate, uint32_t frame_overhead)
{
    // A frame overhead of 0 is not allowed.
    if (!frame_overhead)
        frame_overhead = DEFAULT_FRAME_OVERHEAD;

    // A frame rate of 0 is not allowed.
    if (framerate == 0)
        return 0;

    float rate_value = ClientMath::reg_to_frame_rate(framerate);

    // Return the maximum exposure for the given framerate.
    return ((USEC_CONVERSION / rate_value) - frame_overhead);
}

/*!
 * @brief Calculate how many frames will fit in the partition given a given width and height.
 *
 * @param part_size     The partition size in bytes.
 * @param width         The width we want to use for the math.
 * @param height        The height we want to use for the math.
 * @param bit_mode      Which bit mode are we using.
 * @param ltr           Are we in ltr mode or not.
 *
 * @return The number of frames which will fit in the partition.
 */
uint32_t ClientMath::calc_frame_count(uint64_t part_size, uint32_t width, uint32_t height, uint32_t bit_mode, bool ltr, bool series3)
{
    uint32_t frame_count;
    uint32_t frame_size;

    if (ltr)
    {
        // If we are in LTR mode then the calculation is different.

        // Get size in bytes.
        frame_size = calc_ltr_frame_size(width, height, bit_mode, series3);
        if (frame_size == 0)
        {
            // This can happen during startup, before width & height are really initialized.
            return 1;
        }

        // How many frames will fit in our partition?
        frame_count = (uint32_t)(part_size / frame_size);

        // This "packing" of metadata only happens with the Series3 cameras.
        if (series3)
        {
            // Make it a multiple of 128, the number of frame metadata records per page.
            frame_count = (frame_count / S3_FRAME_METADATA_PER_PAGE) * S3_FRAME_METADATA_PER_PAGE;
        }
    }
    else
    {
        uint32_t value;

        // Calculate the number of bytes per frame.
        value = ClientMath::calc_frame_size(width, height, bit_mode);

        // Are we series3?
        if (series3)
        {
            // Round the frame size up to qword size.
            frame_size = ((value + (QUADWORD_BYTES - 1)) / QUADWORD_BYTES) * QUADWORD_BYTES;

            // Account for the metadata.
            frame_size += S3_FRAME_METADATA_LEN_BYTES;
        }
        else
        {
            // Round the frame size up to hexword size.
            frame_size = ((value + (HEXWORD_BYTES - 1)) / HEXWORD_BYTES) * HEXWORD_BYTES;

            // Account for the metadata.
            frame_size += SNC_FRAME_METADATA_LEN_BYTES;
        }

        // How many frames can we get.
        frame_count = (uint32_t)(part_size / frame_size);

        // SeriesNC cameras like things on page boundaries.
        if (!series3)
        {
            uint64_t temp_addr;

            // We want to make sure that, when loading in from SSD, we don't
            // overflow the end of the partition. Transfers to/from the SSD
            // happen on an SSD_PAGE_LEN boundary...so adjust accordingly.
            //
            // Calculate the start address of the last buffer of the partition.
            temp_addr = (uint64_t)(frame_count - 1) * (uint64_t)frame_size;

            // Calculate the ending address of the buffer when we're
            // transferring from the SSD. We want the ENDING address,
            // not the START address of the next buffer! Round the buffer
            // size up to an SSD_PAGE_LEN boundary. Subtract one to get the
            // ending address and add that length to our starting address.
            temp_addr += ((((frame_size + (SSD_PAGE - 1)) / SSD_PAGE) * SSD_PAGE) - 1);

            // Make sure this will fit.
            if (temp_addr >= part_size)
            {
                // We're going to slop over, so drop the total number of available buffers.
                frame_count--;
            }
        }
    }

    return frame_count;
}

/*!
 * @brief This routine calculates the frame size in image memory given an ROI and the bit depth.
 * @note If the camera is in LR Mode, you want to use [calc_ltr_frame_size](@ref ClientMath::calc_ltr_frame_size)
 *
 * @param width         The width of the image.
 * @param height        The height of the image.
 * @param bit_depth     The bit-depth used (GIGE_BITMODE_10_BITS, GIGE_BITMODE_LOWER_8_BITS, etc.)
 *
 * @return The frame size, in bytes. 0 if an error occurs.
 */
uint32_t ClientMath::calc_frame_size(uint32_t width, uint32_t height, uint32_t bit_depth)
{
    uint32_t    frame_size = 0;

    // Sanity check.
    if ((width == 0) || (height == 0))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "width:" << width << "height:" << height;
        return 0;
    }

    // Get the bit-depth we're capture at.
    frame_size = width;
    if (IS_12_BIT_MODE(bit_depth))
    {
        // For 12-bit pixels, we have 5 pixels per 64-bit double word with 4 bits left over.
        if (frame_size % 5)
            frame_size = (frame_size / 5) + 1;
        else
            frame_size = frame_size / 5;
        frame_size = (frame_size * 8);   // now make it bytes
    }
    else if (IS_10_BIT_MODE(bit_depth))
    {
        // For 10-bit pixels, we have 3 pixels per 32-bit word.
        // We use an even number of 32-bit words per row.
        if (frame_size % 3)
            frame_size = (frame_size / 3) + 1;
        else
            frame_size = frame_size / 3;
        frame_size = (frame_size * 4);   // now make it bytes
    }
    else
    {
        // For 8-bit pixels, we have 4 pixels per 32-bit word.
        // Make sure there is an even number of 32-bit words per row.
        if ((frame_size % 4) == 0)
            frame_size = frame_size / 4;
        else
            frame_size = (frame_size / 4) + 1;
        frame_size = frame_size * 4;  // put it back in bytes
    }

    // Multiple the row size by the number of rows.
    frame_size = frame_size * height;

    // Return our frame size.
    return frame_size;
}

/*!
 * @brief This routine calculates the frame size on [page boundaries](@ref CLIENT_MATH::SSD_PAGE)
 *        given an ROI and the bit depth.
 *
 * @param width         The width of the image.
 * @param height        The height of the image.
 * @param bit_depth     The bit-depth used (GIGE_BITMODE_10_BITS, GIGE_BITMODE_LOWER_8_BITS, etc.)
 *
 * @return The frame size, in [pages](@ref CLIENT_MATH::SSD_PAGE).
 */
uint32_t ClientMath::calc_ltr_frame_size(uint32_t width, uint32_t height, uint32_t bit_depth, bool series3)
{
    uint32_t    frame_size;

    // Calculate the number of bytes for the frame.
    frame_size = ClientMath::calc_frame_size(width, height, bit_depth);

    // Now adjust for LR Mode peculiarities.
    if (series3)
    {
        // Make sure the frame is in multiples of SSD pages and round up.
        frame_size = ((frame_size + (SSD_PAGE - 1)) / SSD_PAGE) * SSD_PAGE;
    }
    else
    {
        // Round the frame size up to hexword size.
        frame_size = ((frame_size + (HEXWORD_BYTES - 1)) / HEXWORD_BYTES) * HEXWORD_BYTES;

        // Add space for the per-frame metadata.
        frame_size += SNC_FRAME_METADATA_LEN_BYTES;
    }

    // Return our frame size.
    return frame_size;
}

/*!
 * @brief Given a number of [sectors](@ref CLIENT_MATH::SSD_SECTOR], return
 *        the number of [QWORD](@ref CLIENT_MATH::SSD_SECTOR_LEN_QWORDS).
 *
 * @param sector_count  The number of sectors.
 * @return The number of QWords.
 */
uint64_t ClientMath::calc_ltr_part_size(uint64_t sector_count)
{
    return (sector_count * (uint64_t)SSD_SECTOR_LEN_QWORDS);
}

/*!
 * @brief Calculate the maximum frame rate for a given width and height.
 *        This is Lupa1300-2 specific.
 *
 * @param startx                The start of the ROI in pixels.
 * @param width                 The width of the ROI in pixels.
 * @param height                The height of the ROI.
 * @param cam_max_framerate     The maximum licensed frame rate of the camera.
 * @param bit_depth             The number of bits per pixel.
 * @param ltr                   True if ltr mode is enabled.
 * @param ssd_bandwidth         The maximum number of bytes per second allowed to the SSD.
 *
 * @return The maximum allowable frame rate for the given ROI.
 */
uint32_t ClientMath::calc_max_framerate(fiSensor *p_sensor, uint32_t startx, uint32_t width, uint32_t height, int bit_depth, bool ltr, uint32_t ssd_bandwidth, bool series3)
{
    uint32_t    max_framerate;
    QRect       roi_rect;

    // Sanity check.
    if (p_sensor == nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Null sensor!?";
        return 0;
    }

    // Initialize our rect.
    roi_rect.setX((int)startx);
    roi_rect.setY(0);
    roi_rect.setWidth((int)width);
    roi_rect.setHeight((int)height);

    // Get the frame rate.
    max_framerate = p_sensor->max_frame_rate(&roi_rect, (uint32_t)bit_depth);

    // Are we in LTR mode?
    if (ltr)
    {
        uint32_t    frame_size;
        double      max_ltr_framerate;

        // if we are in ltr mode then the calculation is different
        frame_size = calc_ltr_frame_size(width, height, (uint32_t)bit_depth, series3);
        if (frame_size == 0)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "frame_size is 0?!";
            return 1;
        }

        max_ltr_framerate = ssd_bandwidth / frame_size;

        // Make sure our LTR maximum is the lower of the two.
        if (max_ltr_framerate < (double)max_framerate)
            max_framerate = (uint32_t)qFloor(max_ltr_framerate);
    }

    // Return our framerate.
    return max_framerate;
}

/*!
 * @brief Calculate the maximum number of 0.256GB sessions that can fit in the
 *        specified DDR2 space.
 *
 * @note The 'ddr2_size' is in QWords.
 *
 * @param ddr2_size     The number of sessions.
 * @return The number of sessions that will fit.
 */
uint32_t ClientMath::calc_max_session_count(uint32_t ddr2_size)
{
    double session_calc = (((double)ddr2_size / (double)GBYTES_TO_QWORDS) * 4.0);

    // Note: We round this calculation up so we can make full use of the available memory.
    //       That last "session" is going to be a runt, because part of it will be reserved
    //       for overhead (live view buffers, etc.)
    return (uint32_t)ceil(session_calc);
}

/*!
 * @brief Calculate the maximum number of 0.256GB sessions that can fit in the
 *        specified device space. Allows for large device space values.
 *
 * @note The 'device_space' is in QWords.
 *
 * @param device_space     The available space in quadwords for a storage device.
 * @return The number of "0.256GB sessions" that will fit.
 */
uint64_t ClientMath::calc_max_session_count64(uint64_t device_space)
{
    uint64_t session_calc = ((device_space / GBYTES_TO_QWORDS) * 4);
    return session_calc;
}

/*!
 * @brief Utility function to calculate the maximum number of frames represented by
 *        the memory and ROI settings.
 *
 * @return The maximum number of frames supported by the memory and ROI settings.
 */
uint32_t ClientMath::calc_max_total_frames(uint32_t ip_addr, uint32_t width, uint32_t height, bool ltr_enabled, bool series3)
{
    uint64_t    max_space;          // The maximum amount of space in a partition.
    uint32_t    frame_count;
    uint32_t    bit_mode;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return 1;
    }

    // Determining the maximum space depends on the mode the camera is in.
    // We can safely read the cache, because these things don't change that
    // frequently (LTR stuff only on boot) and when they do change, we'll
    // get an unsolicited message.
    if (ltr_enabled)
    {
        // Get the number of sectors we support in LTR mode.
        max_space = ((uint64_t)comm_ptr->get_reg(ip_addr, GIGE_REG_LTR_SECTOR_COUNT)) * SSD_SECTOR;
    }
    else
    {
        // Get the size of the partion.
        max_space = ClientMath::qword_to_bytes(comm_ptr->get_reg(ip_addr, GIGE_REG_PART_SIZE));

        // Adjust by the reserved space.
        max_space -= ClientMath::qword_to_bytes(comm_ptr->get_reg(ip_addr, GIGE_REG_PART_SIZE_RESERVED));
    }

    // We need to know what the current BIT mode being used by the camera is.
    bit_mode = comm_ptr->get_reg(ip_addr, GIGE_REG_BIT_MODE);

    // Figure out how many frames will fit in the space.
    frame_count = ClientMath::calc_frame_count(max_space, width, height, bit_mode, ltr_enabled, series3);

    // Return our result.
    return frame_count;
}

/*!
 * @brief Calcualte the amount of time in a recording given the parameters of (frame count, capture rate).
 *        The duration is returned as both a QTime structure and returned as a number of milliseconds.
 *
 * @param frame_count       The number of frames in the recording.
 * @param frame_rate        The frames per second for the recording.
 * @param duration          Pointer to the QTime structure to receive the duration.
 *
 * @return The number of milliseconds in the recording.
 */
uint64_t ClientMath::calc_record_ms(uint32_t frame_count, uint32_t frame_rate, QTime *duration)
{
    double  float_ms;
    uint64_t    total_ms;

    // Calculate how many milliseconds we can record given the number of frames
    // and the frame rate.
    float_ms = ((double)frame_count * 1000.0) / ClientMath::reg_to_frame_rate(frame_rate);
    total_ms = (uint64_t)(float_ms + 0.5);

    // Are we supposed to convert this to a QTime value?
    if (duration != nullptr)
    {
        int hours;
        int min;
        int sec;
        int ms;
        uint64_t    value;

        // Break down our time into the component parts.
        value = total_ms / 1000;
        ms = (int)(total_ms - (value * 1000));
        min = (int)(value / 60);
        sec = (int)(value - ((uint64_t)min * 60));
        hours = min / 60;
        min = min - (hours * 60);

        // Pass the new values back as a QTime structure.
        *duration = QTime(hours, min, sec, ms);
    }

    // Return the total number of milliseconds.
    return total_ms;
}

/*!
 * @brief Calculate the number of QWords available to the session.
 *
 * @param session           The session we're interested in.
 * @param max_session       The maximum number of sessions.
 * @param ddr2_size         The number of QWords in the camera's image memory DDR2.
 * @param ddr2_reserved     The number of reserved QWords (not available for image recording).
 *
 * @return The available number of QWords for image recording.
 */
uint32_t ClientMath::calc_session_space(uint32_t session, uint32_t max_session, uint32_t ddr2_size, uint32_t ddr2_reserved)
{
    uint32_t    max_space;

    // We need to worry about "actual" space when we are being told
    // to use "all" the memory.
    if (session < max_session)
        max_space = (session * SESSION_INCR_QWORDS);
    else
        max_space = (ddr2_size - ddr2_reserved);

    // Return our result.
    return max_space;
}

/*!
 * @brief Provide an interface to return the per-frame-metadata size used for the
 * math calculations based on a type of camera.
 *
 * @param series3      True if asking for a series3 camera, false if asking for a seriesNC camera.
 *
 * @return The number of bytes used for the math calculator routines.
 */
uint32_t ClientMath::get_pfmd_size(bool series3)
{
    uint32_t pfmd_calc_size = SNC_FRAME_METADATA_LEN_BYTES;

    // If series3 camera is asking it is a different valud.
    if (series3)
    {
        pfmd_calc_size = S3_FRAME_METADATA_LEN_BYTES;
    }
    return pfmd_calc_size;
}

/*!
 * @brief Convert the quad-word value qw_value into bytes.
 * @param qw_value      The quad word value to convert.
 * @return The number of byte represented by qw_value;
 */
uint64_t ClientMath::qword_to_bytes(uint64_t qw_value)
{
    return (qw_value * QUADWORD_BYTES);
}

/*!
 * @brief Convert the byte value into quadwords with rounding.
 * @param byte_value    The byte value to convert.
 * @param round_up      True if rounding quadword boundary up but only if it wasn't already aligned.
 * @return The number of quadwords represented in byte_value;
 */
uint64_t ClientMath::bytes_to_qwords_rounded(uint64_t byte_value, bool round_up)
{
    uint64_t qw_value;

    // Get the byte value rounded-down to a QWORD value.
    qw_value = byte_value / QUADWORD_BYTES;

    // Does user want it rounded up?
    if (round_up)
    {
        if ((byte_value % QUADWORD_BYTES) != 0)
        {
            qw_value++;
        }
    }
    // Return the value.
    return (qw_value);
}

/*!
 * @brief Convert the sector value sector_value into bytes.
 * @param sector_value      The number of sectors.
 * @return The number of byte represented by sector_value;
 */
uint64_t ClientMath::sector_to_bytes(uint64_t sector_value)
{
    return (sector_value * SSD_SECTOR);
}

QString ClientMath::fi_time_to_local(const QString& fi_time)
{
    QString localTime;
    QStringList tokens = fi_time.split(':');

    // Sanity check.
    if (fi_time.isEmpty() || (tokens.count() < 2))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Specified time is invalid:" << fi_time;
        return localTime;
    }

    // This will set local_date to 1/1/20<yr>.
    QDate local_date(tokens[0].toInt() + 2000, 1, 1);

    // A '0' would be 1/1/20<yr>, reflected as a '1' in the days. So we need
    // to subtract one.
    local_date = local_date.addDays(tokens[1].toInt() - 1);

    // Display according to the system's short date format and a normal hour:min:sec.microsecond.
    localTime = local_date.toString(Qt::ISODate) + QStringLiteral(" ") + fi_time.right(15);

    return localTime;
}

/*!
 * @brief Convert straight-binary-seconds to a "real" struct tm.
 *
 * This attempts to recreate a real struct tm so that calls made to other
 * time functions (e.g., mktime) will work correctly.
 */
void ClientMath::sbs_to_tm(uint32_t sbs, struct tm *tm_ptr)
{
    int year;
    int isLeap;
    time_t timeNow;
    struct tm temp_tm;

    static const int mon_yday[2][13] =
    {
        // Normal years
        { 0, 31, 59, 90, 120, 151, 181, 212, 243, 273, 304, 334, 365 },
        // Leap years
        { 0, 31, 60, 91, 121, 152, 182, 213, 244, 274, 305, 335, 366 }
    };

    // Clear out the structure.
    memset(&temp_tm, 0, sizeof(struct tm));

    // Initialize unknown elements of our structure by using the current time.
    timeNow = time(NULL);
    memcpy(&temp_tm, localtime(&timeNow), sizeof(struct tm));

    // Get years.
    temp_tm.tm_year = sbs / (365 * 24 * 60 * 60);
    sbs -= ((uint32_t)temp_tm.tm_year * (365 * 24 * 60 * 60));
    temp_tm.tm_yday = sbs / (24 * 60 * 60);
    sbs -= ((uint32_t)temp_tm.tm_yday * (24 * 60 * 60));
    temp_tm.tm_hour = sbs / (60 * 60);
    sbs -= ((uint32_t)temp_tm.tm_hour * (60 * 60));
    temp_tm.tm_min = sbs / 60;
    temp_tm.tm_sec = (int)sbs - (temp_tm.tm_min * 60);

    // We may need to add 100 to the years, because we're older than 2000.
    if (temp_tm.tm_year < 100)
        temp_tm.tm_year += 100;

    // Are we in a leap year?
    year = temp_tm.tm_year + 1900;

    // Divisible by 400?
    if ((year % 400) == 0)
        isLeap = 1;

    // Divisible by 100?
    else if ((year % 100) == 0)
        isLeap = 0;

    // Divisible by 4?
    else if ((year % 4) == 0)
        isLeap = 1;

    // Others are not.
    else
        isLeap = 0;

    int dayOfYear = temp_tm.tm_yday;
    for (temp_tm.tm_mon = 0; temp_tm.tm_mon < 12; ++temp_tm.tm_mon)
    {
        if (dayOfYear <= mon_yday[isLeap][temp_tm.tm_mon])
        {
            // Back up our month.
            temp_tm.tm_mon -= 1;
            if (temp_tm.tm_mon < 0)
                temp_tm.tm_mon = 0;

            // Figure out day of the month.
            temp_tm.tm_mday = (dayOfYear - mon_yday[isLeap][temp_tm.tm_mon]) + 1;
            break;
        }
    }

    // Sanity check.
    if (tm_ptr != nullptr)
    {
        // Copy the structure over before mktime goes messing around with it.
        memcpy(tm_ptr, &temp_tm, sizeof(struct tm));

        // Now see if DST is on.
        (void)mktime(&temp_tm);

        // Update the return values.
        tm_ptr->tm_isdst = temp_tm.tm_isdst;

#if defined( Q_OS_LINUX )
        tm_ptr->tm_gmtoff = temp_tm.tm_gmtoff;
#endif
    }
}

/*!
 * @brief Convert the passed in 'frame_rate' value to the fixed point value
 *        that will go in the GIGE_REG_FRAME_RATE register.
 *
 * @param frame_rate    The floating point frame rate value.
 * @return The uint32_t register value for GIGE_REG_FRAME_RATE.
 */
uint32_t ClientMath::frame_rate_to_reg(float frame_rate)
{
    uint32_t reg_value;

    // Grab the integer value.
    reg_value = (uint32_t)frame_rate;

    // Is there anything else to worry about?
    if (!qFuzzyIsNull((frame_rate - (float)reg_value)))
    {
        // We have a true floating point frame rate. Do the conversion.
        reg_value = (uint32_t)(frame_rate * 65536.0);
        reg_value |= GIGE_FRAME_RATE_FRACTIONAL;
    }

    // Return the register value for frame rate.
    return reg_value;
}

/*!
 * @brief Convert the passed in register value from GIGE_REG_FRAME_RATE
 *        to the floating point value it represents.
 *
 * @param reg_value The value from GIGE_REG_FRAME_RATE.
 * @return The floating point frame rate.
 */
float ClientMath::reg_to_frame_rate(uint32_t reg_value)
{
    float ret_value;

    if ((reg_value & GIGE_FRAME_RATE_FRACTIONAL) == GIGE_FRAME_RATE_FRACTIONAL)
        ret_value = (float)(reg_value & ~GIGE_FRAME_RATE_FRACTIONAL) / 65536.0;
    else
        ret_value = (float)reg_value;

    return ret_value;
}
