/*!
 * @file fi_img_dng.cpp
 * @brief This file contains the implementation of the fiDng class.
 *
 * @note All calculations are assuming sRGB and D65 for a reference
 *       white, unless specifically noted.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <cmath>
#include <float.h>

#include "color_tool.h"
#include "logger.h"

namespace FI_COLORTOOL
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

/*!
 *  3x3 conversion matrix from linear sRGB to CIE XYZ. From the table
 *  found here: http://www.brucelindbloom.com/index.html?Eqn_RGB_XYZ_Matrix.html.
 */
const float RGBtoXYZ_D65[9] =
{
    0.4124564f, 0.3575761f, 0.1804375f,
    0.2126729f, 0.7151522f, 0.0721750f,
    0.0193339f, 0.1191920f, 0.9503041f
};
const float RGBtoXYZ_D50[9] =
{
    0.4360747f, 0.3850649f, 0.1430804f,
    0.2225045f, 0.7168786f, 0.0606169f,
    0.0139322f, 0.0971045f, 0.7141733f
};

/* LERP(a,b,c) = linear interpolation macro, is 'a' when c == 0.0 and 'b' when c == 1.0 */
#define LERP(a,b,c)     (((b) - (a)) * (c) + (a))

typedef struct UVT
{
    double  u;
    double  v;
    double  t;
} UVT;

const float rt[31] =
{
    /* reciprocal temperature (K) */
    FLT_MIN,  10.0e-6,  20.0e-6,  30.0e-6,  40.0e-6,  50.0e-6,
    60.0e-6,  70.0e-6,  80.0e-6,  90.0e-6, 100.0e-6, 125.0e-6,
    150.0e-6, 175.0e-6, 200.0e-6, 225.0e-6, 250.0e-6, 275.0e-6,
    300.0e-6, 325.0e-6, 350.0e-6, 375.0e-6, 400.0e-6, 425.0e-6,
    450.0e-6, 475.0e-6, 500.0e-6, 525.0e-6, 550.0e-6, 575.0e-6,
    600.0e-6
};

const UVT uvt[31] =
{
    {0.18006, 0.26352, -0.24341},
    {0.18066, 0.26589, -0.25479},
    {0.18133, 0.26846, -0.26876},
    {0.18208, 0.27119, -0.28539},
    {0.18293, 0.27407, -0.30470},
    {0.18388, 0.27709, -0.32675},
    {0.18494, 0.28021, -0.35156},
    {0.18611, 0.28342, -0.37915},
    {0.18740, 0.28668, -0.40955},
    {0.18880, 0.28997, -0.44278},
    {0.19032, 0.29326, -0.47888},
    {0.19462, 0.30141, -0.58204},
    {0.19962, 0.30921, -0.70471},
    {0.20525, 0.31647, -0.84901},
    {0.21142, 0.32312, -1.0182},
    {0.21807, 0.32909, -1.2168},
    {0.22511, 0.33439, -1.4512},
    {0.23247, 0.33904, -1.7298},
    {0.24010, 0.34308, -2.0637},
    {0.24792, 0.34655, -2.4681},    /* Note: 0.24792 is a corrected value for the error found in W&S as 0.24702 */
    {0.25591, 0.34951, -2.9641},
    {0.26400, 0.35200, -3.5814},
    {0.27218, 0.35407, -4.3633},
    {0.28039, 0.35577, -5.3762},
    {0.28863, 0.35714, -6.7262},
    {0.29685, 0.35823, -8.5955},
    {0.30505, 0.35907, -11.324},
    {0.31320, 0.35968, -15.628},
    {0.32129, 0.36011, -23.325},
    {0.32931, 0.36038, -40.770},
    {0.33724, 0.36051, -116.45}
};

}   // End of namespace FI_COLORTOOL.

using namespace FI_COLORTOOL;

/*!
 * @brief Convert from an RGB to an XYZ color. Using the math described
 *        here: http://www.brucelindbloom.com/index.html?Eqn_RGB_to_XYZ.html.
 *        Assuming sRGB.
 *
 * @param c Our RGB color.
 * @return The calculated XYZ value.
 */
FI_COLORTOOL::XYZColor ColorTool::rgbToXYZ(FI_COLORTOOL::RGBColor c, bool useD50)
{
    float x, y, z, r, g, b;

    // Normalize our color values.
    r = c.r / 255.0;
    g = c.g / 255.0;
    b = c.b / 255.0;

    // Inverse sRGB companding.
    if (r > 0.04045)
        r = powf(((r + 0.055) / 1.055), 2.4);
    else
        r /= 12.92;
    if (g > 0.04045)
        g = powf(((g + 0.055) / 1.055), 2.4);
    else
        g /= 12.92;
    if (b > 0.04045)
        b = powf(((b + 0.055) / 1.055), 2.4);
    else
        b /= 12.92;

    // Are we using D50 or D65?
    if (useD50)
    {
        // Calibration for observer @2° with illumination = D65
        x = (r * RGBtoXYZ_D50[0]) + (g * RGBtoXYZ_D50[1]) + (b * RGBtoXYZ_D50[2]);
        y = (r * RGBtoXYZ_D50[3]) + (g * RGBtoXYZ_D50[4]) + (b * RGBtoXYZ_D50[5]);
        z = (r * RGBtoXYZ_D50[6]) + (g * RGBtoXYZ_D50[7]) + (b * RGBtoXYZ_D50[8]);
    }
    else
    {
        // Calibration for observer @2° with illumination = D65
        x = (r * RGBtoXYZ_D65[0]) + (g * RGBtoXYZ_D65[1]) + (b * RGBtoXYZ_D65[2]);
        y = (r * RGBtoXYZ_D65[3]) + (g * RGBtoXYZ_D65[4]) + (b * RGBtoXYZ_D65[5]);
        z = (r * RGBtoXYZ_D65[6]) + (g * RGBtoXYZ_D65[7]) + (b * RGBtoXYZ_D65[8]);
    }

    // Return our XYZ color.
    return XYZColor(x, y, z);
}

/*!
 * @brief Convert from an XYZ color to XY coordinates. Using the math described
 *        here: http://www.brucelindbloom.com/index.html?Eqn_XYZ_to_xyY.html.
 *
 * @param c Our XYZ color.
 * @param x The X-coordinate.
 * @param y the Y-coordinate.
 *
 * @return 0 if successful, -1 on an invalid XYZ value.
 */
int ColorTool::xyzToXY(FI_COLORTOOL::XYZColor c, float& x, float& y)
{
    float xyz_sum = c.x + c.y + c.z;

    // We are dealing with floats, so this is aproximately equal to zero.
    if (xyz_sum < 1.0e-20)
        return -1;

    // Perform our conversion.
    x = c.x / xyz_sum;
    y = c.y / xyz_sum;

    return 0;
}

/*!
 * @brief CIE 1931 x,y to correlated color temperature approximation function.
 *
 * Uses approximate formula of Hernandez-Andres, et al. 1999: "Calculating
 * correlated color temperatures across the entire gamut of daylight and
 * skylight chromaticities". Accurate within 2% inside 3K-50K CCT.
 *
 * @return The integer Correlated Color Temperature, or -1 on failure.
 */
int ColorTool::xyToCCT(float x, float y)
{
    const float x_e = 0.3366f;
    const float y_e = 0.1735f;
    const float A0 = -949.86315f;
    const float A1 = 6253.80338f;
    const float t1 = 0.92159f;
    const float A2 = 28.70599f;
    const float t2 = 0.20039f;
    const float A3 = 0.00004f;
    const float t3 = 0.07125f;
    float n;
    float cct;

    // Safety check.
    if (fabs(y - y_e) < 1.0e-20)
        return -1;

    n = (x - x_e) / (y - y_e);
    cct = A0 + (A1 * std::exp(-n / t1)) + (A2 * std::exp(-n / t2)) + (A3 * std::exp(-n / t3));
    if ((cct < 3000) || (cct > 50000))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Conversion only accurate within 3000 to 50000K, result was:" << cct;
    }
    return static_cast<int>(std::floor(cct + 0.5f));
}

/*!
 * @brief XYZ to Correlated Color Temperature.
 *
 * Derived heavily from: http://www.brucelindbloom.com/Eqn_XYZ_to_T.html.
 * From the web page:
 *      Copyright (c) 2003 Bruce Justin Lindbloom. All rights reserved.
 *
 *      This is an implementation of Robertson's method of computing the correlated color
 *      temperature of an XYZ color. It can compute correlated color temperatures in the
 *      range [1666.7K, infinity].
 *
 *      Reference:
 *          "Color Science: Concepts and Methods, Quantitative Data and Formulae", Second Edition,
 *          Gunter Wyszecki and W. S. Stiles, John Wiley & Sons, 1982, pp. 227, 228.
 *
 * @param c     Our XYZ color.
 * @return The computed correlated color temperature if successful, -1 on failure.
 */
int ColorTool::xyzToCCT(FI_COLORTOOL::XYZColor c)
{
    double us, vs, p, di, dm;
    int i;

    // Protect against divide by zero.
    if ((c.x < 1.0e-20) && (c.y < 1.0e-20) && (c.z < 1.0e-20))
        return -1;

    us = (4.0 * c.x) / (c.x + (15.0 * c.y) + (3.0 * c.z));
    vs = (6.0 * c.y) / (c.x + (15.0 * c.y) + (3.0 * c.z));
    dm = 0.0;
    for (i = 0; i < 31; i++)
    {
        di = (vs - uvt[i].v) - uvt[i].t * (us - uvt[i].u);
        if ((i > 0) && (((di < 0.0) && (dm >= 0.0)) || ((di >= 0.0) && (dm < 0.0))))
            break;  /* found lines bounding (us, vs) : i-1 and i */
        dm = di;
    }

    // Bad XYZ input, color temp would be less than minimum of 1666.7 degrees, or too far towards blue.
    if (i == 31)
        return -1;

    di = di / sqrt(1.0 + uvt[i    ].t * uvt[i    ].t);
    dm = dm / sqrt(1.0 + uvt[i - 1].t * uvt[i - 1].t);
    p = dm / (dm - di);     /* p = interpolation parameter, 0.0 : i-1, 1.0 : i */
    p = 1.0 / (LERP(rt[i - 1], rt[i], p));

    // All is well.
    return static_cast<int>(std::floor(p + 0.5f));
}


/*!
 * @brief Blackbody radiator color temperature to CIE 1931 x,y chromaticity
 *        approximation function.
 *
 * Uses approximate formula of Kim et al. 2002: "Design of Advanced Color - Temperature
 * Control System for HDTV Applications" and associated patent.
 */
void ColorTool::kelvinToXY(int T, float *x, float *y)
{
    if (!x || !y)
    {
        return;
    }

    if ((T < 1667) || (T > 25000))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Conversion only accurate within 1667K to 25000K, requested:" << T;
    }

    // Chromaticity x coefficients for T <= 4000K.
    const float A_x00 = -0.2661239f;
    const float A_x01 = -0.2343580f;
    const float A_x02 = 0.8776956f;
    const float A_x03 = 0.179910f;

    // Chromaticity x coefficients for T > 4000K.
    const float A_x10 = -3.0258469f;
    const float A_x11 = 2.1070379f;
    const float A_x12 = 0.2226347f;
    const float A_x13 = 0.24039f;

    // Chromaticity y coefficients for T <= 2222K.
    const float A_y00 = -1.1063814f;
    const float A_y01 = -1.34811020f;
    const float A_y02 = 2.18555832f;
    const float A_y03 = -0.20219683f;

    // Chromaticity y coefficients for 2222K < T <= 4000K.
    const float A_y10 = -0.9549476f;
    const float A_y11 = -1.37418593f;
    const float A_y12 = 2.09137015f;
    const float A_y13 = -0.16748867f;

    // Chromaticity y coefficients for T > 4000K.
    const float A_y20 = 3.0817580f;
    const float A_y21 = -5.87338670f;
    const float A_y22 = 3.75112997f;
    const float A_y23 = -0.37001483f;

    float invKiloK = 1000.0f / T;
    float xc, yc;
    if (T <= 4000)
    {
        xc = A_x00 * invKiloK * invKiloK * invKiloK +
             A_x01 * invKiloK * invKiloK +
             A_x02 * invKiloK +
             A_x03;
    }
    else
    {
        xc = A_x10 * invKiloK * invKiloK * invKiloK +
             A_x11 * invKiloK * invKiloK +
             A_x12 * invKiloK +
             A_x13;
    }

    if (T <= 2222)
    {
        yc = A_y00 * xc * xc * xc +
             A_y01 * xc * xc +
             A_y02 * xc +
             A_y03;
    }
    else if (T <= 4000)
    {
        yc = A_y10 * xc * xc * xc +
             A_y11 * xc * xc +
             A_y12 * xc +
             A_y13;
    }
    else
    {
        yc = A_y20 * xc * xc * xc +
             A_y21 * xc * xc +
             A_y22 * xc +
             A_y23;
    }

    *x = xc;
    *y = yc;
}

/*!
 * @brief  Compute a 3x3 matrix inverse.
 */
void ColorTool::invert3x3(float *in, float *out)
{
    out[0] = (in[4] * in[8]) - (in[5] * in[7]);
    out[3] = (in[5] * in[6]) - (in[3] * in[8]);
    out[6] = (in[3] * in[7]) - (in[4] * in[6]);

    float invdet = 1.0f / ((in[0] * out[0]) + (in[1] * out[3]) + (in[2] * out[6]));

    out[0] *= invdet;
    out[3] *= invdet;
    out[6] *= invdet;

    out[1] = invdet * ((in[7] * in[2]) - (in[8] * in[1]));
    out[4] = invdet * ((in[8] * in[0]) - (in[6] * in[2]));
    out[7] = invdet * ((in[6] * in[1]) - (in[7] * in[0]));

    out[2] = invdet * ((in[1] * in[5]) - (in[2] * in[4]));
    out[5] = invdet * ((in[2] * in[3]) - (in[0] * in[5]));
    out[8] = invdet * ((in[0] * in[4]) - (in[1] * in[3]));
}

void ColorTool::invert3x3(double *in, double *out)
{
    out[0] = in[4] * in[8] - in[5] * in[7];
    out[3] = in[5] * in[6] - in[3] * in[8];
    out[6] = in[3] * in[7] - in[4] * in[6];

    double invdet = 1.0 / (in[0] * out[0] + in[1] * out[3] + in[2] * out[6]);

    out[0] *= invdet;
    out[3] *= invdet;
    out[6] *= invdet;

    out[1] = invdet * (in[7] * in[2] - in[8] * in[1]);
    out[4] = invdet * (in[8] * in[0] - in[6] * in[2]);
    out[7] = invdet * (in[6] * in[1] - in[7] * in[0]);

    out[2] = invdet * (in[1] * in[5] - in[2] * in[4]);
    out[5] = invdet * (in[2] * in[3] - in[0] * in[5]);
    out[8] = invdet * (in[0] * in[4] - in[1] * in[3]);

}


/*!
 * @brief Linearly interpolates two 3x3 color matrices a and b using
 *        interpolant alpha.
 *
 * (result = a if alpha = 0, result = b if alpha = 1).
 * Rescales result matrix to have a minimum row sum of 1.
 */
void ColorTool::colorMatrixInterpolate(const float *a, const float *b, float alpha, float *result)
{
    // Check for identity interpolations
    if (alpha == 0.0f)
    {
        for (int i = 0; i < 9; i++)
        {
            result[i] = a[i];
        }
        return;
    }
    if (alpha == 1.0f)
    {
        for (int i = 0; i < 9; i++)
        {
            result[i] = b[i];
        }
        return;
    }

    // Do a linear interpolation between two known settings.
    for (int i = 0; i < 9; i++)
    {
        result[i] = ((1 - alpha) * a[i]) + (alpha * b[i]);
    }

    // Rescale the result so that the minimum row sum is one.
    float rowSum[3] = { result[0] + result[1] + result[2],
                        result[3] + result[4] + result[5],
                        result[6] + result[7] + result[8]
                      };
    float scale = 1.0f;
    if ((rowSum[0] < rowSum[1]) && (rowSum[0] < rowSum[2]))
    {
        scale = 1.0f / rowSum[0];
    }
    else if (rowSum[1] < rowSum[2])
    {
        scale = 1.0f / rowSum[1];
    }
    else
    {
        scale = 1.0f / rowSum[2];
    }

    for (int i = 0; i < 9; i++)
    {
        result[i] *= scale;
    }
}
