/*!
 * @file color_tool.h
 * @brief This file contains assorted color manipulation utilities.
 *
 * Some of the ideas and constructs are from: https://github.com/ThunderStruct/Color-Utilities.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef COLOR_TOOL_H
#define COLOR_TOOL_H

#include <QString>

namespace FI_COLORTOOL
{

//--------------------------------------------------------------------
// Class and Structure Definitions
//--------------------------------------------------------------------

/*! Represent an XYZ color value. */
struct XYZColor
{
    float   x;
    float   y;
    float   z;

    XYZColor()      {}

    XYZColor(float x, float y, float z)
    {
        this->x = x; this->y = y; this->z = z;
    }
    XYZColor(const XYZColor& c)
    {
        x = c.x; y = c.y; z = c.z;
    }
    XYZColor& operator=(const XYZColor& c)
    {
        x = c.x; y = c.y; z = c.z;
        return *this;
    }

    QString toString()
    {
        return QString("(%1, %2, %3)").arg(x).arg(y).arg(z);
    }
};

/*! Represent an RGB color value. */
struct RGBColor
{
    unsigned int r, g, b;
    float rF, gF, bF;

    RGBColor() {}
    RGBColor(unsigned int r, unsigned int g, unsigned int b)
    {
        this->r = r % 256; this->g = g % 256; this->b = b % 256;
        initFloats();
    }
    RGBColor(float r, float g, float b)
    {
        this->rF = r; this->gF = g; this->bF = b;
        initInts();
    }
    RGBColor(const RGBColor& c)
    {
        r = c.r; g = c.g; b = c.b;
        initFloats();
    }
    RGBColor& operator=(const RGBColor& c)
    {
        r = c.r; g = c.g; b = c.b;
        initFloats();
        return *this;
    }

    QString toString()
    {
        return QString("(%1, %2, %3)").arg(r).arg(g).arg(b);
    }

private:
    void initFloats()
    {
        rF = r / 255.0; gF = g / 255.0; bF = b / 255.0;
    }
    void initInts()
    {
        r = rF * 255.0; g = gF * 255.0; b = bF * 255.0;
    }
};

}   // End of namespace FI_COLORTOOL.

/*!
 * @brief The ColorTool class collects our color processing functions.
 */
class ColorTool
{
public:
    ColorTool()     {}
    ~ColorTool()    {}

    static FI_COLORTOOL::XYZColor rgbToXYZ(FI_COLORTOOL::RGBColor c, bool useD50);
    static int xyzToXY(FI_COLORTOOL::XYZColor c, float& x, float& y);
    static int xyToCCT(float x, float y);
    static int xyzToCCT(FI_COLORTOOL::XYZColor c);
    static void kelvinToXY(int T, float *x, float *y);
    static void invert3x3(float *in, float *out);
    static void invert3x3(double *in, double *out);
    static void colorMatrixInterpolate(const float *a, const float *b, float alpha, float *result);
};

#endif // COLOR_TOOL_H
