/*!
 * @file control_panel.cpp
 * @brief Manages the display of the various control panels.
 *
 * @note The layout of controls is EXTREMELY important to the operation of
 * the panel. The assumption is that the splitter has "children".
 * The splitter children are arranged in this EXACT order:
 *   + QFrame control #0 to show/hide the panels, etc.
 *   + fiPanel widget #0 with all controls.
 *   +  :
 *   + QFrame control #n to show/hide the panels, etc.
 *   + fiPanel widget #n with all controls.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "control_panel.h"
#include "ui_control_panel.h"

#include <QDebug>
#include <QMdiSubWindow>
#include <QMenu>
#include <QResizeEvent>
#include <QScrollArea>
#include <QTimer>

#include "logger.h"
#include "mainwindow.h"
#include "panel_display.h"
#include "panel_lens.h"
#include "panel_reccontrols.h"
#include "panel_recsettings.h"
#include "panel_capture.h"

namespace CONTROL_PANEL
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The panels MUST be at the offset to the frame they are associated with.
const int PANEL_OFFSET = 1;

//! The frame controls contain, in this EXACT order, the following children:
//! * Layout widget
//! * QToolButton to show/hide the panel
//! * QToolButton to lock the panel
//!
//! The above order is reflected in the following constants.
//! @{

const int LAYOUT_WIDGET_OFFSET = 0;     //!< Offset to the QLayout widget.
const int SHOW_HIDE_BUTTON_OFFSET = 1;  //!< Offset to the QToolButton with the up/down arrows.
const int LOCK_BUTTON_OFFSET = 2;       //!< Offset to the lock/unlock QToolButton.
const int FRAME_CHILDREN = 3;           //!< The number of expected children per frame.

//! @}

//! How often to check to see if it's time to update the panels.
const int   UPDATE_POLL_TIMER           = 250;

}   // End of namespace CONTROL_PANEL.

using namespace CONTROL_PANEL;

/*!
 * @brief ControlPanel class constructor.
 */
ControlPanel::ControlPanel(QWidget *parent) :
    QDockWidget(parent),
    ui(new Ui::control_panel)
{
    int     index;
    int     req_width(0);               // Widest width required for the panels.
    QSplitter   *p_splitter;            // Convenience variable to access our splitter.

    // Setup the UI elements.
    ui->setupUi(this);

    // Initialize our variables.
    camera_ptr = nullptr;

    // Get a pointer to our splitter.
    p_splitter = ui->splitter;
#if 0
    QRect orig_area = p_splitter->geometry();   // Pull the original starting area of our splitter.
#endif

    // Setup our control panel splitter object and it's children.
    for (index = 0; index < p_splitter->count(); ++index)
    {
        QToolButton *p_button;

        // Make our buttons NOT collapsable.
        p_splitter->setCollapsible(index, false);

        // Get the lock button.
        p_button = getLockButton(index);
        if (p_button != nullptr)
        {
            // Hide the button.
            p_button->setVisible(false);
        }
    }

    // Create our panels and insert them into the splitter immediately following
    // the panels associated frame with the show/hide button.
    addPanel(new PanelCapture(this), p_splitter->indexOf(ui->ss_frame) + 1, req_width);
    addPanel(new PanelRecSettings(this), p_splitter->indexOf(ui->rs_frame) + 1, req_width);
    addPanel(new PanelRecControls(this), p_splitter->indexOf(ui->rc_frame) + 1, req_width);
    addPanel(new PanelDisplay(this), p_splitter->indexOf(ui->ds_frame) + 1, req_width);
    addPanel(new PanelLens(this), p_splitter->indexOf(ui->lc_frame) + 1, req_width);

#if 0
    // Get our current width.
    cp_width = geometry().width();

    // Subtract the original width of the splitter and add back the final width.
    cp_width -= orig_area.width();
    cp_width += req_width;
#else
    cp_width = req_width;
#endif

    // Resize our widget to meet the needs of our panels.
    resize(cp_width, geometry().height());
    //    setMinimumWidth( cp_width );

    // Give all available left over space to our spacer.
    index = p_splitter->indexOf(ui->spacer);
    p_splitter->setStretchFactor(index, 1);

    // Create a scroll area to put the vertical splitter inside
    p_scroll_area = new QScrollArea;
    p_scroll_area->setWidget(p_splitter);
    p_scroll_area->setWidgetResizable(true);
    p_scroll_area->setHorizontalScrollBarPolicy(Qt::ScrollBarAlwaysOff);
    p_scroll_area->setVerticalScrollBarPolicy(Qt::ScrollBarAsNeeded);

    // Set the scroll area as our main widget of the control panel.
    setWidget(p_scroll_area);

    // Disable the handle used to resize the various panels.
    for (index = 0; index < p_splitter->count(); ++index)
        p_splitter->handle(index)->setEnabled(false);

    // Setup our connections.
    if (!initConnects())
        return;
}

/*!
 * @brief ControlPanel class destructor.
 */
ControlPanel::~ControlPanel()
{
    delete ui;
}

/*!
 * @brief Add a "panel", or dialog, into our control panel.
 *
 * @param p_panel   Pointer to the widget we're adding.
 * @param index     Where in the splitter we want the panel to be added.
 * @param req_width The required width.
 *
 * @return True on success, false otherwise.
 */
bool ControlPanel::addPanel(QWidget *p_panel, int index, int& req_width)
{
    QRect   panel_area;

    if (p_panel == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "p_pane is nullptr!";
        return false;
    }

    // Add the new panel to our splitter and our list of widgets.
    widget_list.append(qMakePair(index, p_panel));
    ui->splitter->insertWidget(index, p_panel);

    // The panel is not collapsible.
    ui->splitter->setCollapsible(index, false);

    // Figure out the width of the panel. If it's larger than
    // any other panel, update our minimum required width.
    panel_area = p_panel->geometry();
    if (panel_area.width() > req_width)
        req_width = panel_area.width();

    // All good.
    return true;
}

QToolButton *ControlPanel::getActivateButton(int frame_index)
{
    QFrame *p_frame;
    QToolButton *p_activate_button = nullptr;

    // Get the frame object.
    p_frame = qobject_cast<QFrame *>(ui->splitter->widget(frame_index));
    if (p_frame != nullptr)
    {
        // Make sure we have the correct number of elements.
        if (p_frame->children().count() == FRAME_CHILDREN)
        {
            // Access the lock button.
            p_activate_button = qobject_cast<QToolButton *>(p_frame->children().at(SHOW_HIDE_BUTTON_OFFSET));
        }
    }

    // Return our pointer.
    return p_activate_button;

}

QToolButton *ControlPanel::getLockButton(int frame_index)
{
    QFrame *p_frame;
    QToolButton *p_lock_button = nullptr;

    // Get the frame object.
    p_frame = qobject_cast<QFrame *>(ui->splitter->widget(frame_index));
    if (p_frame != nullptr)
    {
        // Make sure we have the correct number of elements.
        if (p_frame->children().count() == FRAME_CHILDREN)
        {
            // Access the lock button.
            p_lock_button = qobject_cast<QToolButton *>(p_frame->children().at(LOCK_BUTTON_OFFSET));
        }
    }

    // Return our pointer.
    return p_lock_button;
}

/*!
 * @brief Disconnect the camera from the control panels.
 */
void ControlPanel::cameraDisconnect()
{
    int     index;

    // Tell all the panels to disconnect.
    for (index = 0; index < widget_list.size(); ++index)
    {
        FMPanelWidget *p_panel = qobject_cast<FMPanelWidget *>(widget_list.at(index).second);
        if (p_panel)
            p_panel->camera_disconnect();
    }

    // We are no longer connected to a camera.
    camera_ptr = nullptr;
}

/*!
 * @brief Called when the p_widget child resizes, this handles updating
 *        the control panel display/layout.
 *
 * @param p_widget  Pointer to the child widget that is resizing.
 */
void ControlPanel::childResized(QWidget *p_widget)
{
    int     index;

    // Loop through all our panel widgets.
    for (index = 0; index < widget_list.size(); ++index)
    {
        // Is this the panel we're interested in?
        if (p_widget != widget_list.at(index).second)
        {
            // Not our calling widget, so minimize this one.
            FMPanelWidget *p_panel = qobject_cast<FMPanelWidget *>(widget_list.at(index).second);
            if (p_panel && p_panel->isVisible())
            {
                QToolButton *p_lock_button;
                int frame_index;

                // Get the index of the frame. The frame is the one before the panel.
                frame_index = widget_list.at(index).first - PANEL_OFFSET;
                if (frame_index < 0)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Invalid frame index!";
                    return;
                }

                // Access the lock button.
                p_lock_button = getLockButton(frame_index);
                if (p_lock_button != nullptr)
                {
                    // Is the lock active?
                    if (!p_lock_button->isChecked())
                    {
                        // Tell the panel to change size.
                        p_panel->toggle_visibility(false);

                        // Make sure the lock button is not visible.
                        p_lock_button->setVisible(false);
                    }
                }
            }
        }
    }
}

/*!
 * @brief Tell all of the control panel widgets to close themselves.
 */
void ControlPanel::closeAllPanels()
{
    int     index;
    FMPanelWidget   *p_panel;

    // Loop through all our panel widgets.
    for (index = 0; index < widget_list.size(); ++index)
    {
        int frame_index;
        QToolButton *p_lock_button;

        // Get the index of the frame. The frame is the one before the panel.
        frame_index = widget_list.at(index).first - PANEL_OFFSET;
        if (frame_index < 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Invalid frame index!";
            return;
        }

        // Get our lock button.
        p_lock_button = getLockButton(frame_index);
        if (p_lock_button != nullptr)
        {
            // Is the lock active?
            if (!p_lock_button->isChecked())
            {
                // Request the panel at this index to close.
                p_panel = qobject_cast<FMPanelWidget *>(widget_list.at(index).second);
                if (p_panel)
                    p_panel->close_panel();

                // Make sure the lock button is not visible.
                p_lock_button->setVisible(false);
            }
        }
    }
}

/*!
 * @brief Get the pointer to the panel object specified by the "index".
 *
 * @param index     Which [panel](@ref CONTROL_PANEL::PANEL_ID) we're interested in.
 * @return Pointer to the panel, or nullptr if index is invalid.
 */
QWidget *ControlPanel::getPanel(int index)
{
    // Sanity check.
    if ((index < 0) || (index > widget_list.size()))
        return nullptr;

    // Return the desired panel.
    return widget_list.at(index).second;
}

/*!
 * @brief Initialize our signal/slot connections.
 * @return True on success, false otherwise.
 */
bool ControlPanel::initConnects()
{
    MainWindow  *p_mainwnd;

    // Connect the signals to the function via a lambda.
    connect(ui->storage_activate, &QToolButton::clicked, [this]() { panelButtonClicked(this->ui->storage_activate); });
    connect(ui->recsettings_activate, &QToolButton::clicked, [this]() { panelButtonClicked(this->ui->recsettings_activate); });
    connect(ui->reccontrols_activate, &QToolButton::clicked, [this]() { panelButtonClicked(this->ui->reccontrols_activate); });
    connect(ui->display_activate, &QToolButton::clicked, [this]() { panelButtonClicked(this->ui->display_activate); });
    connect(ui->lens_activate, &QToolButton::clicked, [this]() { panelButtonClicked(this->ui->lens_activate); });

    // Connect the menu item(s).
    p_mainwnd = qobject_cast<MainWindow *>(parent());

    // Panel menu signals & slots.
    connect(p_mainwnd->get_cp_close_all(), &QAction::triggered, this, &ControlPanel::closeAllPanelsTriggered);
    connect(p_mainwnd, &MainWindow::active_child_changed, this, &ControlPanel::newActiveCamera);

    // All good.
    return true;
}

/*!
 * @brief Tell a panel to open.
 *
 * @param index     The index of the panel to open.
 */
void ControlPanel::openPanel(int index)
{
    FMPanelWidget   *p_panel;

    // Sanity check.
    if (index >= widget_list.size())
        return;

    // Make sure all the panels are closed first.
    closeAllPanels();

    // Open the requested panel.
    p_panel = qobject_cast<FMPanelWidget *>(widget_list.at(index).second);
    if (p_panel)
        p_panel->open_panel();
}

/*!
 * @brief This function is called to enable/disable a panel and it's associated
 *        activation button.
 *
 * @param p_panel       Pointer to the panel we're interested in.
 * @param do_enable     True to enable, false to disable.
 */
void ControlPanel::panelButtonEnable(QWidget *p_panel, bool do_enable)
{
    int splitter_index;

    // Get the splitter's index for the panel.
    splitter_index =  ui->splitter->indexOf(p_panel);
    if (splitter_index > 0)
    {
        QFrame *p_frame;

        // The associated frame is the previous index. Adjust and
        // retrieve the pointer for the frame.
        splitter_index -= PANEL_OFFSET;
        p_frame  = qobject_cast<QFrame *>(ui->splitter->widget(splitter_index));
        if (p_frame == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Frame isn't the selected child!";
        }
        else
        {
            // Enable/disable the button.
            p_frame->setEnabled(do_enable);
        }
    }
}

/*!
 * @brief Used to modify the activation button's icon based on the state
 *        (i.e. being opened or closed) of the panel.
 *
 * @param p_panel       Pointer to the panel we're interested in.
 * @param icon_type     The arrow type we want to have the button use.
 */
void ControlPanel::panelButtonIcon(QWidget *p_panel, Qt::ArrowType icon_type)
{
    int splitter_index;

    // Get the splitter's index for the panel.
    splitter_index =  ui->splitter->indexOf(p_panel);
    if (splitter_index > 0)
    {
        int frame_index;
        QToolButton *p_button;

        // The associated frame is the previous index. Adjust and
        // retrieve the pointer for the frame.
        frame_index = splitter_index - PANEL_OFFSET;

        // Access the show/hide button.
        p_button = getActivateButton(frame_index);
        if (p_button != nullptr)
        {
            // Enable/disable the button.
            p_button->setArrowType(icon_type);
        }
    }
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void ControlPanel::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        ui->retranslateUi(this);
    }

    // Pass it on.
    QDockWidget::changeEvent(event);
}

void ControlPanel::doDock()
{
    setFloating(false);
}

/*!
 * @brief Handle a right click from the user - this will pop-up a context menu.
 * @param event_ptr     Pointer to the right-click event.
 */
void ControlPanel::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QMenu   context_menu(this);

    // Only do this if we're floating.
    if (!isFloating())
        return;

    // Add a menu to re-dock the panel.
    context_menu.addAction(tr("Dock"), this, SLOT(doDock()));

    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Respond to the Close All Panels action.
 */
void ControlPanel::closeAllPanelsTriggered()
{
    closeAllPanels();
}

/*!
 * @brief There is a new camera we should connect to.
 * @param window    The window associated with the new camera.
 */
void ControlPanel::newActiveCamera(QMdiSubWindow *window)
{
    CameraMDI   *old_camera;
    CameraMDI   *new_camera;

    // Save the "old" camera.
    old_camera = camera_ptr;

    // Get the "new" active camera.
    if (window != nullptr)
        new_camera = qobject_cast<CameraMDI *>(window->widget());
    else
        new_camera = nullptr;

    // Is the old camera different from the new camera?
    if ((old_camera != nullptr) && (new_camera != old_camera))
    {
        // Make sure all the panels disconnect from the old camera.
        cameraDisconnect();
    }

    // Now, update our camera pointer to point to the new camera.
    camera_ptr = new_camera;
    if (camera_ptr == nullptr)
    {
        // Update the title to reflect there are no cameras connected.
        setWindowTitle(tr("Control Panel - No Camera"));
        return;
    }

    // Update the title to reflect our new camera.
    setWindowTitle(tr("Control Panel - ") + camera_ptr->user_friendly_name());

    // Have our panels update their settings.
    updatePanels();
}

/*!
 * @brief This slot is used to handle when the user clicks on a
 *        panel activate button.
 *
 * @param source_button     Pointer to the activate button that was clicked.
 */
void ControlPanel::panelButtonClicked(QWidget *source_button)
{
    int  frame_index;
    QToolButton *p_button;
    QToolButton *p_lock_button;
    FMPanelWidget   *p_panel;

    // Sanity check.
    p_button = qobject_cast<QToolButton *>(source_button);
    if (p_button == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "p_button is nullptr!";
        return;
    }

    // Find the index of the button in the splitter so we can
    // know which panel to work on.
    frame_index = ui->splitter->indexOf(p_button->parentWidget());

    // Get a pointer to our panel.
    p_panel = qobject_cast<FMPanelWidget *>(ui->splitter->widget(frame_index + PANEL_OFFSET));
    if (p_panel == nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "p_panel is null";
        return;
    }

    // Get our lock button.
    p_lock_button = getLockButton(frame_index);
    if (p_lock_button == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Lock button isn't the selected child!";
        return;
    }

    // Are we showing or hiding the panel?
    if (p_button->arrowType() == Qt::DownArrow)
    {
        // Ignore the click if we're locked.
        if (p_lock_button->isChecked())
            return;

        // Close the panel.
        if (p_panel)
            p_panel->close_panel();

        // Hide the lock button.
        p_lock_button->setVisible(false);
    }
    else
    {
        // Open the panel.
        p_panel->open_panel();

        // Show the lock button.
        p_lock_button->setVisible(true);
    }
}

/*!
 * @brief Iterate through each of the panels and have them update their settings.
 */
void ControlPanel::updatePanels()
{
    int     index;

    // Make sure we have a camera pointer.
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer.";
        return;
    }

    // Let's make sure we're connected.
    if (camera_ptr->get_sensor() == nullptr)
    {
        QTimer::singleShot(CONTROL_PANEL::UPDATE_POLL_TIMER, this, SLOT(updatePanels()));
        return;
    }

    // Loop through all our panel widgets.
    for (index = 0; index < widget_list.size(); ++index)
    {
        FMPanelWidget *p_panel = qobject_cast<FMPanelWidget *>(widget_list.at(index).second);

        // Is this panel open?
        if (p_panel)
        {
            // Is the camera off-line?
            if (!camera_ptr->get_camera_info()->offline)
            {
                // Have the panel setup it's connections to the camera.
                p_panel->camera_connect(camera_ptr->get_ip_addr());

                // Tell the panel to update.
                p_panel->update_panel();
            }
        }
    }
}

