/*!
 * @file control_panel.h
 * @brief This file contains the definition of the ControlPanel class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CONTROL_PANEL_H
#define CONTROL_PANEL_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QDockWidget>
#include <QList>
#include <QPair>

#include "camera_mdi.h"

namespace Ui
{
class control_panel;
}

namespace CONTROL_PANEL
{
//--------------------------------------------------------------------
// Constant, Macro and Type definitions
//--------------------------------------------------------------------

//! The mode which the control panel and it's children are operating under.
enum cp_mode
{
    CP_MODE_NONE,           //!< No control panel mode defined.
    CP_MODE_LIVE,           //!< Control panel is in live playback mode.
    CP_MODE_ARMED,          //!< Control panel is in armed live mode.
    CP_MODE_REVIEW,         //!< Both the camera and the PC App are in review mode.
    CP_MODE_AUTOSAVE        //!< Autosave mode.
};

//! The list of panels. Note that it is important that the panels are added
//! to the "widget_list" member in this order.
enum PANEL_ID
{
    PANEL_STORAGE,          //!< Identifier for the storage panel.
    PANEL_REC_SETTINGS,     //!< Identifier for the record settings panel.
    PANEL_REC_CONTROLS,     //!< Identifier for the record controls panel.
    PANEL_DISPLAY,          //!< Identifier for the display settings panel.
    PANEL_LENS              //!< Identifier for the lens control panel.
};

}   // End of namespace CONTROL_PANEL.

//--------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------
class QScrollArea;
class QToolButton;

//--------------------------------------------------------------------
// The ControlPanel class definition
//--------------------------------------------------------------------
class ControlPanel : public QDockWidget
{
    Q_OBJECT

public:
    explicit ControlPanel(QWidget *parent = nullptr);
    ~ControlPanel();

    void    childResized(QWidget *p_widget);
    void    closeAllPanels();
    void    openPanel(int index);

    QWidget *getPanel(int index);
    CameraMDI *getCamera()          { return camera_ptr; }

    void    panelButtonEnable(QWidget *p_panel, bool do_enable);
    void    panelButtonIcon(QWidget *p_panel, Qt::ArrowType icon_type);

    void    cameraDisconnect();

public slots:
    void    updatePanels();

private slots:
    void    closeAllPanelsTriggered();
    void    doDock();
    void    newActiveCamera(QMdiSubWindow *window);
    void    panelButtonClicked(QWidget *source_button);

protected:
    virtual void changeEvent(QEvent *event);
    virtual void contextMenuEvent(QContextMenuEvent *event_ptr);

private:
    bool    addPanel(QWidget *p_panel, int index, int& req_width);
    QToolButton *getActivateButton(int frame_index);
    QToolButton *getLockButton(int frame_index);
    bool    initConnects();

private:
    Ui::control_panel   *ui;                    //!< Our UI object.

    QList<QPair<int, QWidget *>> widget_list;   //!< The list of our widgets and their index into the splitter.
    QScrollArea         *p_scroll_area;         //!< The scroll control widget.
    CameraMDI           *camera_ptr;            //!< What we think of as the "active" child. May actually be the "last" active child.

    int         cp_width;                       //!< Used to calculate the control panel width.

};

#endif // CONTROL_PANEL_H
