/*!
 * @file cornergrabber.h
 * @brief This is the header file for the CornerGrabber class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CORNERGRABBER_H
#define CORNERGRABBER_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QObject>
#include <QGraphicsItem>
#include <QGraphicsRectItem>
#include <QGraphicsTextItem>
#include <QGraphicsSceneHoverEvent>
#include <QGraphicsSceneMouseEvent>
#include <QColor>
#include <QPainter>
#include <QPen>
#include <QPointF>

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------

class CornerGrabber : public QGraphicsItem
{

public:
    explicit CornerGrabber(QGraphicsItem *parent = nullptr, int corner=0);
    ~CornerGrabber() {}

    int  getCorner() const;             //!< Allows the owner to find out which corner this is.
    void setMouseState(int new_state);  //!< Allows the owner to record the current mouse state.
    int  getMouseState() const;         //!< Allows the owner to get the current mouse state.

    void setWidth(qreal width)      { _width = width; }
    void setHeight(qreal height)    { _height = height; }

    virtual QRectF boundingRect() const;    //!< Must be re-implemented in this class to provide the diminsions of the box to the QGraphicsView.

    qreal mouseDownX;                   //!< Convenient location for the "owner" to store where in the corner the mouse-X down position is.
    qreal mouseDownY;                   //!< Convenient location for the "owner" to store where in the corner the mouse-Y down position is.

    //! Define the mouse states.
    enum {kMouseReleased=0, kMouseDown, kMouseMoving};

private:
    virtual void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget); //!< Must be re-implemented here to paint the box on the paint-event.
    virtual void hoverEnterEvent(QGraphicsSceneHoverEvent *event); //!< Must be re-implemented to handle mouse hover enter events.
    virtual void hoverLeaveEvent(QGraphicsSceneHoverEvent *event); //!< Must be re-implemented to handle mouse hover leave events.

    // Once the hover event handlers are implemented in this class,
    // the mouse events must be implemented because of
    // some linkage issue - apparrently there is some connection
    // between the hover events and mouseMove/Press/Release
    // events which triggers a vtable issue.
    virtual void mouseMoveEvent(QGraphicsSceneMouseEvent * event);
    virtual void mouseMoveEvent(QGraphicsSceneDragDropEvent *event);
    virtual void mousePressEvent(QGraphicsSceneMouseEvent *event);
    virtual void mousePressEvent(QGraphicsSceneDragDropEvent *event);
    virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent * event);

    QColor  _outterborderColor;     //!< The hover event handlers will toggle this between red and black.
    QPen    _outterborderPen;       //!< The pen is used to paint the red/black border.

    qreal   _width;                 //!< The width of our corner box.
    qreal   _height;                //!< The height of our corner box.

    int     _corner;                //!< 0,1,2,3  - starting at x=0,y=0 and moving clockwise around the box.
    int     _mouseButtonState;      //!< One of our mouse states (kMouseReleased, kMouseDown, kMouseMoving).
};

#endif // CORNERGRABBER_H
