/*!
 * @file curve_settings.cpp
 * @brief The code to display a snapshot image and allow the user
 *        to modify the gamma via a spline curve control.
 *
 * Copyright 2016 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "curve_settings.h"
#include "ui_curve_settings.h"

#include <errno.h>
#include <QApplication>
#include <QDomDocument>
#include <QDomElement>
#include <QFileDialog>
#include <QFontMetrics>
#include <QGraphicsPixmapItem>
#include <QGraphicsScene>
#include <QMessageBox>
#include <QPixmap>
#include <QResizeEvent>
#include <QTimer>
#include <QtConcurrent>

#include "camera_mdi.h"
#include "fi_comm.h"
#include "fi_histo_proc.h"
#include "fi_lutfile.h"
#include "fi_msgbox.h"
#include "fi_splinewidget.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_mdi.h"

namespace CURVE_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The number of entries in our RGB lookup tables.
const int   LUT_LENGTH                  = 256;

//! Our polling period when waiting for a file action to complete.
const int   FA_POLLING_PERIOD           = 250;

//! Spline curve tab indices and curve identifier.
//! Note:  These are used for two purposes.  The tab
//! curve index and a curve identifier.
const int   SPLINE_TAB_NONE                 = -1;
const int   SPLINE_TAB_INDEX_RED_OR_MONO    = 0;
const int   SPLINE_TAB_INDEX_GREEN          = 1;
const int   SPLINE_TAB_INDEX_BLUE           = 2;

//! Spline XML tag strings for saving and loading data from xml file.
const QString LUT_SPLINE_LIST_DOM_DOCUMENT          = QStringLiteral("lut_list");
const QString LUT_SPLINE_LIST_DOM_ELEMENT           = QStringLiteral("lut_listing");
const QString LUT_SPLINE_LIST_COUNT_DOM_ELEMENT     = QStringLiteral("lut_count");
const QString LUT_SPLINE_REC_ACTIVE_DOM_ELEMENT     = QStringLiteral("active_lut");
const QString LUT_SPLINE_REC_DOM_ELEMENT            = QStringLiteral("lut_info");
const QString LUT_SPLINE_REC_DOM_ID                 = QStringLiteral("id");
const QString LUT_SPLINE_REC_DOM_DESC               = QStringLiteral("description");
const QString LUT_SPLINE_REC_DOM_LUT_DATA_TYPE      = QStringLiteral("lut_data_type");
const QString LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOTS = QStringLiteral("knots");
const QString LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOT  = QStringLiteral("knot");
const QString LUT_SPLINE_REC_DOM_GREEN_KNOTS        = QStringLiteral("green_knots");
const QString LUT_SPLINE_REC_DOM_GREEN_KNOT         = QStringLiteral("green_knot");
const QString LUT_SPLINE_REC_DOM_BLUE_KNOTS         = QStringLiteral("blue_knots");
const QString LUT_SPLINE_REC_DOM_BLUE_KNOT          = QStringLiteral("blue_knot");

//! Spline QSettings strings used to save and load data for LUT_SPLINE_RECORD.
const QString LUT_SPLINE_SETTINGS_GROUP             = QStringLiteral("gammaLut");
const QString LUT_SPLINE_SETTINGS_LASTFILE          = QStringLiteral("last_filename");
const QString LUT_SPLINE_SETTINGS_ACTIVE_ID         = QStringLiteral("activeID");
const QString LUT_SPLINE_SETTINGS_COUNT             = QStringLiteral("count");
const QString LUT_SPLINE_SETTINGS_GROUP_LUT_ID      = QStringLiteral("lut_%1");
const QString LUT_SPLINE_SETTINGS_LUT_ID            = QStringLiteral("id");
const QString LUT_SPLINE_SETTINGS_DESC              = QStringLiteral("desc");
const QString LUT_SPLINE_SETTINGS_LUT_DATA_TYPE     = QStringLiteral("lut_data_type");
const QString LUT_SPLINE_SETTINGS_TYPE_UNKNOWN      = QStringLiteral("%1").arg(LUT_SERIES_UNKNOWN);
const QString LUT_SPLINE_SETTINGS_KNOTS             = QStringLiteral("knots");
const QString LUT_SPLINE_SETTINGS_GREEN_KNOTS       = QStringLiteral("greenKnots");
const QString LUT_SPLINE_SETTINGS_BLUE_KNOTS        = QStringLiteral("blueKnots");

//! Initial state of spline lock control.
const bool  SPLINE_LOCK_INITIAL_SETTING             = true;

//! Set True for VERBOSE Logging.
const bool VERBOSE_LOGGING = false;

}   // End of namespace CURVE_SETTINGS.

using namespace CURVE_SETTINGS;

//! Spline curve tab names
const QString CurveSettings::SPLINE_TAB_GAMMA_NAME         = QT_TR_NOOP("Gamma");
const QString CurveSettings::SPLINE_TAB_RED_NAME           = QT_TR_NOOP("Red");
const QString CurveSettings::SPLINE_TAB_GREEN_NAME         = QT_TR_NOOP("Green");
const QString CurveSettings::SPLINE_TAB_BLUE_NAME          = QT_TR_NOOP("Blue");

/*!
 * @brief The class constructor.
 */
CurveSettings::CurveSettings(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::CurveSettings),
    myParent(parent),
    allowLutProcessing(false),
    activeLutTabIndex(SPLINE_TAB_INDEX_RED_OR_MONO),
    err_received(false),
    builtRedOrGammaCameraLut(false),
    builtGreenCameraLut(false),
    builtBlueCameraLut(false),
    busyCursorShowing(false),
    cameraLutLastModified(SPLINE_TAB_INDEX_RED_OR_MONO),
    initialLockSetting(SPLINE_LOCK_INITIAL_SETTING),
    isMonoSensor(false),
    isSeries3Camera(false),
    listing_len(0),
    okToUpdateDisplayImage(false),
    old_slot(0),
    poll_timer_id(0),
    refreshRedOrGammaCurve(false),
    refreshGreenCurve(false),
    refreshBlueCurve(false),
    scene_ptr(nullptr),
    pixmap_item_ptr(nullptr),
    snap_pixmap(nullptr),
    mod_pixmap(nullptr),
    splineChangedTimerStarted(false),
    splineChangedCount(0),
    splinesInitialized(false),
    splineIsLocked(false)
{
}

/*!
 * @brief The class destructor.
 */
CurveSettings::~CurveSettings()
{
    if (poll_timer_id != 0)
    {
        killTimer(poll_timer_id);
        poll_timer_id = 0;
    }
    if (scene_ptr != nullptr)
    {
        delete scene_ptr;
        scene_ptr = nullptr;
    }
    if (snap_pixmap != nullptr)
    {
        delete snap_pixmap;
        snap_pixmap = nullptr;
    }
    if (mod_pixmap != nullptr)
    {
        delete mod_pixmap;
        mod_pixmap = nullptr;
    }

    delete ui;
}

/*!
 * @brief Factory to create a curve settings dialog.
 *
 * @param setMonoSensor - true if the image data is from monochrome sensor, false if not.
 * @param setSeries3Camera - true if the image data is from series 3 camera, false if not.
 * @param displayImageSnap - the display image to supply to user as they manipulate curve spline(s).
 * @param initialLutId - the one-based radio LUT id to set for dialog.
 * @param parent - who is this dialog's parent widget default nullptr if not supplied.
 *
 * @returns the std::unique_ptr for the curve settings dialog.
 */
std::unique_ptr<CurveSettings> CurveSettings::CreateCurveDialog(bool setMonoSensor,
                                                                bool setSeries3Camera,
                                                                fiPixmap& displayImageSnap,
                                                                int InitialLutId,
                                                                QWidget *parent)
{
    std::unique_ptr<CurveSettings> theCurveSettingsDlg = std::unique_ptr<CurveSettings>(new CurveSettings(parent));

    // Install the configuration setup for the xfer dialog.
    if (theCurveSettingsDlg)
    {
        theCurveSettingsDlg->myParent           = parent;
        theCurveSettingsDlg->isMonoSensor       = setMonoSensor;
        theCurveSettingsDlg->isSeries3Camera    = setSeries3Camera;
        theCurveSettingsDlg->snap_pixmap        = new fiPixmap(displayImageSnap.copy());
        theCurveSettingsDlg->mod_pixmap         = new fiPixmap(displayImageSnap.copy());
        theCurveSettingsDlg->splinesInitialized = false;
        theCurveSettingsDlg->okToUpdateDisplayImage = false;
        theCurveSettingsDlg->initialLockSetting = SPLINE_LOCK_INITIAL_SETTING;

        // Is our parent a playback object?  And if so does it have a custom
        // lut spline record to use?
        PlaybackMDI *myPlaybackParent = qobject_cast<PlaybackMDI *>(parent);
        if (myPlaybackParent != nullptr)
        {
            // See if there is custom metadata lut data.  It will have a lot of information
            // that can be leveraged initializing the curve dialog.
            theCurveSettingsDlg->playbackLut = myPlaybackParent->GetMetadataLut();

            // Do we have a metadata LUT?
            if (theCurveSettingsDlg->playbackLut)
            {
                // If the id is > 0 this is a custom lut.  Make sure our spline record
                // and our internal Id matches this one.
                if (theCurveSettingsDlg->playbackLut->GetLutId() <= 0)
                {
                    // This is not usable.
                    theCurveSettingsDlg->playbackLut = nullptr;
                }
            }

            // If we still have a playback LUT then use it.
            if (theCurveSettingsDlg->playbackLut)
            {
                if (VERBOSE_LOGGING)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Processing playback LUT id:" << theCurveSettingsDlg->playbackLut->GetLutId();
                }

                // Override the initial lut id provided we will install the metadata LUT.
                InitialLutId = theCurveSettingsDlg->playbackLut->GetLutId();

                DISPLAY_LUT::LUT_SPLINE_RECORD playbackSplineRec;
                theCurveSettingsDlg->playbackLut->CopySplineRecord(playbackSplineRec);

                // Set the initial lock based on the knots being equal.
                theCurveSettingsDlg->initialLockSetting = FiDisplayLut::KnotsAreEqual(playbackSplineRec);

                if (VERBOSE_LOGGING)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Initialized Lock:" << theCurveSettingsDlg->initialLockSetting
                                 << "appending id:" << playbackSplineRec.id
                                 << "lutSplineRecs.count():" << theCurveSettingsDlg->lutSplineRecs.count()
                                 << "lutDataType:" << playbackSplineRec.lutDataType;
                }

                // Add the spline record to our list of spline records.
                theCurveSettingsDlg->lutSplineRecs.append(playbackSplineRec);
            }
        }

        // Setup the initial LUT id for the destination channel.
        if (InitialLutId > 0 && InitialLutId < 6)
            theCurveSettingsDlg->old_slot = InitialLutId;
        else
            theCurveSettingsDlg->old_slot = 1;

        // Initialize our UI.
        theCurveSettingsDlg->ui->setupUi(theCurveSettingsDlg.get());

        // Create a local histogram process object.
        theCurveSettingsDlg->histoProc = std::unique_ptr<fiHistoProc>(new fiHistoProc(theCurveSettingsDlg.get()));

        // Allocate memory for the display image lookup tables.
        theCurveSettingsDlg->displayLutRed.fill(0, LUT_LENGTH);
        theCurveSettingsDlg->displayLutGreen.fill(0, LUT_LENGTH);
        theCurveSettingsDlg->displayLutBlue.fill(0, LUT_LENGTH);

        // Create our default scene.
        theCurveSettingsDlg->scene_ptr = new QGraphicsScene(theCurveSettingsDlg.get());
        theCurveSettingsDlg->pixmap_item_ptr = nullptr;

        // Put this text into the scene to let the user know to wait.
        if (theCurveSettingsDlg->UsingRGBSplines())
            theCurveSettingsDlg->scene_ptr->addText("Preparing RGB Curves and Histograms.  Please Wait...");
        else
            theCurveSettingsDlg->scene_ptr->addText("Preparing Gamma Curve and Histogram.  Please Wait...");

        // Add our new scene to the graphics view object.
        theCurveSettingsDlg->ui->image_view->setScene(theCurveSettingsDlg->scene_ptr);

        // Initialize our controls.
        theCurveSettingsDlg->InitializeControls();

        // No matter what we always have a red/gamma spline curve.
        theCurveSettingsDlg->aaSplineRedOrGamma = std::shared_ptr<aaAaa::aaSpline>(new aaAaa::aaSpline());
        theCurveSettingsDlg->aaSplineRedOrGamma->name = theCurveSettingsDlg->ui->fi_spline->objectName().toStdString();
        theCurveSettingsDlg->aaSplineRedOrGamma->type = aaAaa::aaSpline::SPLINE_CUBIC;

        // And if we have RGB sensor set up the green and blue channels.
        if (theCurveSettingsDlg->UsingRGBSplines())
        {
            // Create the green and blue spline curve controls.
            theCurveSettingsDlg->aaSplineGreen = std::shared_ptr<aaAaa::aaSpline>(new aaAaa::aaSpline());
            theCurveSettingsDlg->aaSplineGreen->name = theCurveSettingsDlg->ui->fiSplineGreen->objectName().toStdString();
            theCurveSettingsDlg->aaSplineGreen->type = aaAaa::aaSpline::SPLINE_CUBIC;

            theCurveSettingsDlg->aaSplineBlue = std::shared_ptr<aaAaa::aaSpline>(new aaAaa::aaSpline());
            theCurveSettingsDlg->aaSplineBlue->name = theCurveSettingsDlg->ui->fiSplineBlue->objectName().toStdString();
            theCurveSettingsDlg->aaSplineBlue->type = aaAaa::aaSpline::SPLINE_CUBIC;

            // Initialize the size of the red, green and blue camera lut vectors.
            theCurveSettingsDlg->cameraLutRedOrGamma.resize(LUT_MAX_TABLE_COUNT);
            theCurveSettingsDlg->cameraLutGreen.resize(LUT_MAX_TABLE_COUNT);
            theCurveSettingsDlg->cameraLutBlue.resize(LUT_MAX_TABLE_COUNT);
        }
        else
        {
            // Initialize the size of the gamma camera lut vector.
            theCurveSettingsDlg->cameraLutRedOrGamma.resize(LUT_MAX_TABLE_COUNT);
        }

        // Initialize our signals/slots for the push buttons.
        connect(theCurveSettingsDlg->ui->pb_ok, &QPushButton::clicked, theCurveSettingsDlg.get(), &CurveSettings::do_accept);
        connect(theCurveSettingsDlg->ui->pb_cancel, &QPushButton::clicked, theCurveSettingsDlg.get(), &CurveSettings::do_cancel);
        connect(theCurveSettingsDlg->ui->pb_load, &QPushButton::clicked, theCurveSettingsDlg.get(), &CurveSettings::do_load);
        connect(theCurveSettingsDlg->ui->pb_reset, &QPushButton::clicked, theCurveSettingsDlg.get(), &CurveSettings::do_reset);
        connect(theCurveSettingsDlg->ui->pb_save, &QPushButton::clicked, theCurveSettingsDlg.get(), &CurveSettings::do_save);

        // If the parent is closing, we need to bail. Note, this only applies for a CameraMDI. PlaybackMDI objects
        // don't tend to close underneath us.
        CameraMDI *my_parent = qobject_cast<CameraMDI *>(parent);
        if (my_parent != nullptr)
            connect(my_parent, &CameraMDI::closing, theCurveSettingsDlg.get(), &CurveSettings::do_cancel);

        // Disable the control frame until initialization complete.
        theCurveSettingsDlg->ui->tb_curve_lock->setEnabled(false);
        theCurveSettingsDlg->ui->tabLutChannels->setEnabled(false);
        theCurveSettingsDlg->ui->lut_control_frame->setEnabled(false);

        // Spawn a one-shot timer that will finish the initialization of our dialog.
        connect(&theCurveSettingsDlg->splineInitTimer, &QTimer::timeout, theCurveSettingsDlg.get(), &CurveSettings::HandleSplineInitTimer, Qt::UniqueConnection);

        // Fire off the timer to start the spline curve initialization.
        theCurveSettingsDlg->splineInitTimer.setSingleShot(true);
        theCurveSettingsDlg->splineInitTimer.start(100);
    }

    // Return the curve settings dialog.
    return theCurveSettingsDlg;
}

/*!
 * @brief Returns the radio button id for the checked "Destination LUT" the current spline
 * curve(s) are modifying.
 */
int CurveSettings::GetCheckedRadioLutId()
{
    int slot;

    // Figure out which slot the user wants.
    if (ui->rb_1->isChecked())
        slot = 1;
    else if (ui->rb_2->isChecked())
        slot = 2;
    else if (ui->rb_3->isChecked())
        slot = 3;
    else if (ui->rb_4->isChecked())
        slot = 4;
    else
        slot = 5;
    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Checked Radio Lut Id:" << slot;
    }
    return slot;
}

/*!
 * @brief Return the camera lookup table LUT_FILE, as a byte array.
 */
QByteArray  *CurveSettings::GetCameraGammaLut()
{
    int     i;                      // Generic loop variable.
    QByteArray      *the_array;     // This is the QByteArray we'll return.

    LUT_FILE        *lutData;       // This pointer will overlay the QByteArray (the_array).
    int             lutDataSize = sizeof(LUT_FILE);

    // Allocate our memory.
    the_array = new QByteArray(lutDataSize, 0);
    if (the_array == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate our QByteArray!";
        return nullptr;
    }

    // Overlay our structure onto the QByteArray.
    lutData = (LUT_FILE *)the_array->data();

    // Copy the description.
    strncpy(lutData->hdr.description, ui->eb_description->text().toUtf8().constData(), LUT_DESCRIPTION_LEN);

    // This should already be zero from the allocation above but calling attention to the
    // fact that it is not an RGB LUT but a single channel.
    if (isMonoSensor && !isSeries3Camera)
    {
        lutData->hdr.lutDataType = LUT_SERIESNC_MONO;
    }
    else
    {
        lutData->hdr.lutDataType = LUT_SERIES3;
    }

    // Loop through the "knots" and save the coordinates.
    for (i = 0; (i < (int)aaSplineRedOrGamma->knots.size()) && (i < LUT_MAX_KNOT_COUNT); i++)
    {
        aaAaa::aaPoint  knot = aaSplineRedOrGamma->knots.at((uint64_t)i);
        lutData->knots[i].x = (uint16_t)knot.t;
        lutData->knots[i].y = (uint16_t)knot.y;
    }

    // Save the camera lookup table.
    for (i = 0; (i < cameraLutRedOrGamma.size()) && (i < LUT_MAX_TABLE_COUNT); i++)
    {
        lutData->table[i] = (uint8_t)cameraLutRedOrGamma.at(i);
    }
    // Return our byte array.
    return the_array;
}

/*!
 * @brief Return the camera lookup table LUT_RGB_FILE, as a byte array.
 */
QByteArray  *CurveSettings::GetCameraRGBLut()
{
    int     i;                      // Generic loop variable.
    QByteArray      *the_array;     // This is the QByteArray we'll return.

    LUT_RGB_FILE    *lutData;       // This pointer will overlay the QByteArray (the_array).
    int             lutDataSize = sizeof(LUT_RGB_FILE);

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Pulling Knots for RGB Camera Save.";
    }

    // Allocate our memory.
    the_array = new QByteArray(lutDataSize, 0);
    if (the_array == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate our QByteArray!";
        return nullptr;
    }

    // Overlay our structure onto the QByteArray.
    lutData = (LUT_RGB_FILE *)the_array->data();

    // Copy the description.
    strncpy(lutData->hdr.description, ui->eb_description->text().toUtf8().constData(), LUT_DESCRIPTION_LEN);

    // This is an RGB LUT with 3 channels.
    lutData->hdr.lutDataType = LUT_SERIESNC_COLOR;

    // Loop through the red "knots" and save the coordinates.
    for (i = 0; (i < (int)aaSplineRedOrGamma->knots.size()) && (i < LUT_MAX_KNOT_COUNT); i++)
    {
        aaAaa::aaPoint  knot = aaSplineRedOrGamma->knots.at((uint64_t)i);
        lutData->redKnots[i].x = (uint16_t)knot.t;
        lutData->redKnots[i].y = (uint16_t)knot.y;

        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << "[R" << lutData->redKnots[i].x << "," << lutData->redKnots[i].y << "]";
        }
    }

    // Save the red channel camera lookup table.
    for (i = 0; (i < cameraLutRedOrGamma.size()) && (i < LUT_MAX_TABLE_COUNT); i++)
    {
        lutData->redTable[i] = (uint8_t)cameraLutRedOrGamma.at(i);
    }

    // Loop through the green "knots" and save the coordinates.
    for (i = 0; (i < (int)aaSplineGreen->knots.size()) && (i < LUT_MAX_KNOT_COUNT); i++)
    {
        aaAaa::aaPoint  knot = aaSplineGreen->knots.at((uint64_t)i);
        lutData->greenKnots[i].x = (uint16_t)knot.t;
        lutData->greenKnots[i].y = (uint16_t)knot.y;

        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << "[G" << lutData->greenKnots[i].x << "," << lutData->greenKnots[i].y << "]";
        }
    }

    // Save the green channel camera lookup table.
    for (i = 0; (i < cameraLutGreen.size()) && (i < LUT_MAX_TABLE_COUNT); i++)
    {
        lutData->greenTable[i] = (uint8_t)cameraLutGreen.at(i);
    }

    // Loop through the blue "knots" and save the coordinates.
    for (i = 0; (i < (int)aaSplineBlue->knots.size()) && (i < LUT_MAX_KNOT_COUNT); i++)
    {
        aaAaa::aaPoint  knot = aaSplineBlue->knots.at((uint64_t)i);
        lutData->blueKnots[i].x = (uint16_t)knot.t;
        lutData->blueKnots[i].y = (uint16_t)knot.y;
        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << "[B" << lutData->blueKnots[i].x << "," << lutData->blueKnots[i].y << "]";
        }
    }

    // Save the blue channel camera lookup table.
    for (i = 0; (i < cameraLutBlue.size()) && (i < LUT_MAX_TABLE_COUNT); i++)
    {
        lutData->blueTable[i] = (uint8_t)cameraLutBlue.at(i);
    }

    // Return our byte array.
    return the_array;
}

/*!
 * @brief Return the camera lookup table, as a byte array.
 *
 * Note:
 * Depending on the type of camera lookup table (series3/mono or rgb) the
 * number of bytes in the lookup table will be different.  The data field
 * in LUT_FILE called lutDataType identifies the type of data returned.
 *
 * Imporant:  It is caller's responsibility to delete and free up memory allocated
 * in "the_array".
 *
 * @returns an allocated QByteArray containing the camera LUT(s) and it is
 * expected that the caller will take ownership of the allocated array and
 * cleanup memory when it is done using the data.  If a nullptr is returned
 * then something went wrong.
 */
QByteArray *CurveSettings::GetCameraLut()
{
    QByteArray      *the_array;     // This is the QByteArray we'll return.

    // What type of LUT are we supporting?
    if (UsingRGBSplines())
    {
        the_array = GetCameraRGBLut();
    }
    else
    {
        the_array = GetCameraGammaLut();
    }

    // Return our byte array.
    return the_array;
}

/*!
 * @brief Traverses the lutSplineRecs array looking for a match with a selectedRadioLutId to
 * locate the index into the array of records.
 *
 * @returns The zero-based index into lutSplineRecs matching the splineLutId provided
 * or a -1 if the match wasn't found.
 */

int CurveSettings::GetLutSplineRecId()
{
    bool foundMatch = false;
    int i;
    int selectedRadioLutId = GetCheckedRadioLutId();  // Which destination LUT radio button is checked?

    // Loop through our list to find the selected LUT.
    for (i = 0; i < lutSplineRecs.count(); i++)
    {
        // Break out of the loop if this is the one.
        if (lutSplineRecs.at(i).id == selectedRadioLutId)
        {
            foundMatch = true;
            break;
        }
    }

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "foundMatch:" << foundMatch
                     << "index:" << i
                     << "desc:" << (foundMatch ? lutSplineRecs.at(i).desc : "No Match");
    }

    // Did we get a match?
    if (!foundMatch)
        return -1;

    // Return the selection.
    return i;
}

/*!
 * @brief Return a reference to the selected LUT_SPLINE_RECORD (@sa CURVE_SETTINGS::LUT_SPLINE_RECORD).
 *        Remember that LUT_SPLINE_RECORD id's are not (necessarily) the same as an index!
 */
const DISPLAY_LUT::LUT_SPLINE_RECORD& CurveSettings::GetRadioLutSplineRec()
{
    int index;

    // Get the index of the selected LUT in our array of LUT_SPLINE_RECORDs.
    index = GetLutSplineRecId();

    // If we couldn't find one, return the first entry.
    if (index < 0)
    {
        DISPLAY_LUT::LUT_SPLINE_RECORD    empty_lut;

        empty_lut.lutDataType = LUT_SERIES3;

        // Add an empty LUT_SPLINE_RECORD and point at that.
        empty_lut.id = GetCheckedRadioLutId();
        lutSplineRecs.append(empty_lut);

        // Our empty lut will be the last entry.
        index = lutSplineRecs.count() - 1;
    }

    // Return the found entry.
    return lutSplineRecs.at(index);
}

/*!
 * @brief Build the red, green and blue display lookup tables (LUTs) to apply the
 *        changes to the image as specified by the user.
 */
void CurveSettings::BuildDisplayLuts()

{
    uint32_t    r_value;
    uint32_t    g_value;
    uint32_t    b_value;
    int         i;

    // Sanity check.
    if ((displayLutRed.empty()) || (displayLutGreen.empty()) || (displayLutBlue.empty()))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "One of the display LUTs is empty()!";
        return;
    }

    // What type of curve settings is this dialog managing?
    if (UsingRGBSplines())
    {
        // Now generate the LUT values for every pixel 0-255. Note that there
        // are 4 times more entries in the gamma LUT.
        for (i = 0; i < LUT_LENGTH; i++)
        {
            // Compute the individual RGB values, adjusting for any gain changes as well.
            r_value = (uint32_t)(cameraLutRedOrGamma[i * 4] + 0.5);
            g_value = (uint32_t)(cameraLutGreen[i * 4] + 0.5);
            b_value = (uint32_t)(cameraLutBlue[i * 4] + 0.5);

            // Make sure we're clipped at 255 - the maximum
            // output value.
            if (r_value > 0xFF)
                r_value = 0xFF;
            if (g_value > 0xFF)
                g_value = 0xFF;
            if (b_value > 0xFF)
                b_value = 0xFF;

            // Update the display image 8bpp LUTs.
            displayLutRed[ i ] = (uint8_t)r_value;
            displayLutGreen[ i ] = (uint8_t)g_value;
            displayLutBlue[ i ] = (uint8_t)b_value;
        }
    }
    else
    {
        // Now generate the LUT values for every pixel 0-255. Note that there
        // are 4 times more entries in the gamma LUT.
        for (i = 0; i < LUT_LENGTH; i++)
        {
            // Compute the individual RGB values, adjusting for any gain changes as well.
            r_value = (uint32_t)(cameraLutRedOrGamma[i * 4] + 0.5);
            g_value = (uint32_t)(cameraLutRedOrGamma[i * 4] + 0.5);
            b_value = (uint32_t)(cameraLutRedOrGamma[i * 4] + 0.5);

            // Make sure we're clipped at 255 - the maximum
            // output value.
            if (r_value > 0xFF)
                r_value = 0xFF;
            if (g_value > 0xFF)
                g_value = 0xFF;
            if (b_value > 0xFF)
                b_value = 0xFF;

            // Update the display image 8bpp LUTs.
            displayLutRed[ i ] = (uint8_t)r_value;
            displayLutGreen[ i ] = (uint8_t)g_value;
            displayLutBlue[ i ] = (uint8_t)b_value;
        }
    }

    // Remove any existing pixmap.
    if (mod_pixmap != nullptr)
    {
        delete mod_pixmap;
        mod_pixmap = nullptr;
    }

    // Make sure we have a pixmap. We won't during the dialog creation.
    if (snap_pixmap != nullptr)
    {
        // Copy the original lookup table.
        mod_pixmap = new fiPixmap(snap_pixmap->copy());

        // Pass the image through the lookup table.
        mod_pixmap->apply_lookup(displayLutRed.constData(), displayLutGreen.constData(), displayLutBlue.constData());
    }
}

/*!
 * @brief Once all of the spline curve(s) have been set up this will be called to
 * complete the initialization and allow the user to make edits to the color curves.
 */
void CurveSettings::ColorCurvesReadyForEditing()
{
    // And if we have RGB sensor set up the green and blue channels.
    if (UsingRGBSplines())
    {
        // Only connect up signals for the rgb tabs and support the three channel splines.
        connect(ui->fi_spline, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::RedSplineChanged);
        connect(ui->fi_spline, &fiSplineWidget::switchType, this, &CurveSettings::RedSplineTypeChanged);
        connect(ui->fiSplineGreen, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::GreenSplineChanged);
        connect(ui->fiSplineGreen, &fiSplineWidget::switchType, this, &CurveSettings::GreenSplineTypeChanged);
        connect(ui->fiSplineBlue, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::BlueSplineChanged);
        connect(ui->fiSplineBlue, &fiSplineWidget::switchType, this, &CurveSettings::BlueSplineTypeChanged);
    }
    else
    {
        // Only connect up signals for the single tabbed gamma/mono spline.
        connect(ui->fi_spline, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::GammaSplineChanged);
        connect(ui->fi_spline, &fiSplineWidget::switchType, this, &CurveSettings::GammaSplineTypeChanged);
    }

    splinesInitialized = true;

    // At this point let the image display in our QGraphicsView.
    okToUpdateDisplayImage = true;

    // Enable the lut controls now that initialization is complete.
    ui->lut_control_frame->setEnabled(true);
    ui->tabLutChannels->setEnabled(true);
    ui->tb_curve_lock->setEnabled(true);

    // Restore the application cursor.
    if (busyCursorShowing)
    {
        QApplication::restoreOverrideCursor();
        busyCursorShowing = false;
    }

    // Let anyone that is waiting on our curve data know it can be used now that curves are ready.
    emit CurvesAreReadyForEditing();
}

/*!
 * @brief Disconnects all timers and updates and curve controls in the dialog.
 */
void CurveSettings::DisableControls()
{
    splinesInitialized = false;

    // And if we have RGB sensor set up the green and blue channels.
    if (UsingRGBSplines())
    {
        // Only connect up signals for the rgb tabs and support the three channel splines.
        disconnect(ui->fi_spline, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::RedSplineChanged);
        disconnect(ui->fi_spline, &fiSplineWidget::switchType, this, &CurveSettings::RedSplineTypeChanged);
        disconnect(ui->fiSplineGreen, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::GreenSplineChanged);
        disconnect(ui->fiSplineGreen, &fiSplineWidget::switchType, this, &CurveSettings::GreenSplineTypeChanged);
        disconnect(ui->fiSplineBlue, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::BlueSplineChanged);
        disconnect(ui->fiSplineBlue, &fiSplineWidget::switchType, this, &CurveSettings::BlueSplineTypeChanged);
    }
    else
    {
        // Only connect up signals for the single tabbed gamma/mono spline.
        disconnect(ui->fi_spline, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::GammaSplineChanged);
        disconnect(ui->fi_spline, &fiSplineWidget::switchType, this, &CurveSettings::GammaSplineTypeChanged);
    }

    disconnect(ui->fi_spline, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::RedSplineChanged);
    disconnect(ui->fiSplineGreen, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::GreenSplineChanged);
    disconnect(ui->fiSplineBlue, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::BlueSplineChanged);

    disconnect(ui->rb_1, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    disconnect(ui->rb_2, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    disconnect(ui->rb_3, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    disconnect(ui->rb_4, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    disconnect(ui->rb_5, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);

    // Disconnect to capture any tab clicks changing active spline curve.
    disconnect(ui->tabLutChannels, &QTabWidget::currentChanged, this, &CurveSettings::HandleLutTabClick);

    // Disconnect our Lut processing complete signal to its handler.
    disconnect(this, &CurveSettings::LutProcessingComplete, this, &CurveSettings::HandleLutProcessingComplete);

    // Disconnect our Build Camera Lut completed signal to its handler.
    disconnect(this, &CurveSettings::BuildACameraLutCompleted, this, &CurveSettings::HandleBuildACameraLutCompleted);
}


/*!
 * @brief Utility function to process the write completion status of the FileAction register
 *        write performed by one of the operations.
 */
void CurveSettings::handle_fa_status(uint32_t fa_status)
{
    // Check to see where the camera is at.
    if ((fa_status & GIGE_FA_STATUS_BUSY) == (uint32_t)GIGE_FA_STATUS_BUSY)
    {
        // The camera is still busy generating the LUT listing.
    }
    else if ((fa_status & GIGE_FA_STATUS_ERROR) == (uint32_t)GIGE_FA_STATUS_ERROR)
    {
        QString err_info_str;

        // Turn off the timer.
        if (poll_timer_id != 0)
        {
            killTimer(poll_timer_id);
            poll_timer_id = 0;
        }

        // Build up our string with more information.
        err_info_str = tr("The camera has reported an error:\n%1 (%2)").arg(fa_status & GIGE_FA_STATUS_MASK).arg(strerror(fa_status & GIGE_FA_STATUS_MASK));
        QLOG_WARN() << __PRETTY_FUNCTION__ << err_info_str;
    }
    else
    {
        uint32_t    read_len;
        CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

        // Sanity check.
        if (my_parent == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
            return;
        }

        // Turn off the timer.
        if (poll_timer_id != 0)
        {
            killTimer(poll_timer_id);
            poll_timer_id = 0;
        }

        // Extract the number of bytes in the listing.
        listing_len = fa_status & GIGE_FA_STATUS_MASK;

        // How much can we grab?
        if (listing_len > GIGE_FILEPATH_LEN)
            read_len = GIGE_FILEPATH_LEN;
        else
            read_len = listing_len;

        // Do we have a listing to download?
        if (read_len > 0)
        {
            // Start pulling down the listing.
            my_parent->start_camera_read_buf(GIGE_REG_FILEPATH, read_len);
        }
    }
}

/*!
 * @brief This method performs the actual ReadMem calls to download the directory listing produced
 *        on the camera by a previous FileAction register write.
 *
 * The assumption is that everything is setup, the destination file has been created and left in
 * an open, usable state so that this method simply has to pull down the data and write it out.
 */
void CurveSettings::handle_listing_download()
{
    uint32_t    buffer_len;
    QByteArray  *read_chunk;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Get the chunk of data we just read from the camera.
    read_chunk = my_parent->read_camera_buffer(GIGE_REG_FILEPATH);
    if (read_chunk == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No buffer available!";
        return;
    }
    if (read_chunk->size() <= 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Bad buffer length: " << read_chunk->size();
        delete read_chunk;
        return;
    }

    // Get the number of bytes we've read...but we only care up to listing_len bytes.
    buffer_len = (uint32_t)read_chunk->size();
    if (buffer_len > listing_len)
        buffer_len = listing_len;

    // Write the chunk out to our file.
    listing_file.write(read_chunk->constData(), read_chunk->size());

    // Done with the chunk.
    delete read_chunk;

    // How much do we have left?
    listing_len -= buffer_len;
    if (listing_len != 0)
    {
        uint32_t    read_len;

        // How much can we grab?
        if (listing_len > GIGE_FILEPATH_LEN)
            read_len = GIGE_FILEPATH_LEN;
        else
            read_len = listing_len;

        // Request the next chunk.
        my_parent->start_camera_read_buf(GIGE_REG_FILEPATH, read_len);
    }
    else
    {
        // We're done with this listing. Close the file.
        listing_file.close();

        // Parse the information we've downloaded.
        parse_listing();
    }
}

/*!
 * @brief The register read of GIGE_REG_LUT has returned a value.
 * @param reg_value     Which LUT is active.
 */
void CurveSettings::handle_lut_change(uint32_t reg_value)
{
    // Setup our radio selection.
    switch (reg_value & GIGE_LUT_MASK)
    {
    default:
    case 1:
        ui->rb_1->setChecked(true);
        break;
    case 2:
        ui->rb_2->setChecked(true);
        break;
    case 3:
        ui->rb_3->setChecked(true);
        break;
    case 4:
        ui->rb_4->setChecked(true);
        break;
    case 5:
        ui->rb_5->setChecked(true);
        break;
    }

    // Update the curve.
    HandleLutRadioClick();
}

/*!
 * @brief Initialize our dialog boxes controls.
 */
void CurveSettings::InitializeControls()
{
    // Set the active lut channel to be the red/mono spline control.
    activeLutTabIndex = SPLINE_TAB_INDEX_RED_OR_MONO;
    ui->tabLutChannels->setCurrentIndex(activeLutTabIndex);

    // Make sure the tabs are set up depending on the type of sensor/camera series we are supporting.
    if (UsingRGBSplines())
    {
        // Set up the tab names for RGB curve manipulation.
        ui->tabLutChannels->setTabText(SPLINE_TAB_INDEX_RED_OR_MONO, SPLINE_TAB_RED_NAME);
        ui->tabLutChannels->setTabText(SPLINE_TAB_INDEX_GREEN, SPLINE_TAB_GREEN_NAME);
        ui->tabLutChannels->setTabText(SPLINE_TAB_INDEX_BLUE, SPLINE_TAB_BLUE_NAME);

        // All three tabs are visible.
        ui->tabLutChannels->setTabVisible(SPLINE_TAB_INDEX_RED_OR_MONO, true);
        ui->tabLutChannels->setTabVisible(SPLINE_TAB_INDEX_GREEN, true);
        ui->tabLutChannels->setTabVisible(SPLINE_TAB_INDEX_BLUE, true);

        // We support locking the three curves to one.
        ui->tb_curve_lock->setVisible(true);
        ui->tb_curve_lock->setEnabled(true);
        splineIsLocked = SPLINE_LOCK_INITIAL_SETTING;
        ui->tb_curve_lock->setChecked(splineIsLocked);
    }
    else
    {
        // Set up the tab names for gamma curve manipulation.
        ui->tabLutChannels->setTabText(SPLINE_TAB_INDEX_RED_OR_MONO, SPLINE_TAB_GAMMA_NAME);

        // All three tabs are visible.
        ui->tabLutChannels->setTabVisible(SPLINE_TAB_INDEX_RED_OR_MONO, true);
        ui->tabLutChannels->setTabVisible(SPLINE_TAB_INDEX_GREEN, false);
        ui->tabLutChannels->setTabVisible(SPLINE_TAB_INDEX_BLUE, false);

        // There is no support for locking three curves to one because there is only one curve.
        ui->tb_curve_lock->setVisible(false);
        ui->tb_curve_lock->setEnabled(false);
    }

    // Connect to capture any tab clicks changing active spline curve.
    connect(ui->tabLutChannels, &QTabWidget::currentChanged, this, &CurveSettings::HandleLutTabClick);

    // Connect our Lut processing complete signal to its handler.
    connect(this, &CurveSettings::LutProcessingComplete, this, &CurveSettings::HandleLutProcessingComplete, Qt::QueuedConnection);

    // Connect our Build Camera Lut completed signal to its handler.
    connect(this, &CurveSettings::BuildACameraLutCompleted, this, &CurveSettings::HandleBuildACameraLutCompleted, Qt::QueuedConnection);

    // Which channel should we set to be the default?
    if ((old_slot < 1) || (old_slot > 5))
        old_slot = 1;

    // Setup the initial setting for the destination LUT but don't expect it to
    // load up any spline data at this point in the dialog setup (second argument
    // specifies whether or not to load the LUT.
    SetupDestinationLut(old_slot, false);

    // Connect up the radiobuttons to a handler.
    connect(ui->rb_1, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    connect(ui->rb_2, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    connect(ui->rb_3, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    connect(ui->rb_4, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
    connect(ui->rb_5, &QRadioButton::clicked, this, &CurveSettings::HandleLutRadioClick);
}

/*!
 * @brief Common logic to setup the dialog when there isn't a spline lut record
 * to use for initialization.
 */
void CurveSettings::InitializeDefaultSetup()
{
    // Supply a default setup.
    if (UsingRGBSplines())
    {
        InitializeSpline(SPLINE_TAB_INDEX_RED_OR_MONO);
        InitializeSpline(SPLINE_TAB_INDEX_GREEN);
        InitializeSpline(SPLINE_TAB_INDEX_BLUE);
    }
    else
    {
        InitializeSpline(SPLINE_TAB_INDEX_RED_OR_MONO);
    }

    // Clear out the description.
    ui->eb_description->setText("");
    ui->eb_path->setText("");
    ui->eb_path->setVisible(false);
}

/*!
 * @brief Initialize the spline control which displays the curve for the user.
 *        The default is a linear "curve" (really, a line) from low to high.
 *
 * @param curveTabIndex - which curve is being initialized.
 */
void CurveSettings::InitializeSpline(int curveTabIndex)
{
    std::shared_ptr<aaAaa::aaSpline> aaSplineToInit;
    QMutex *aaSplineMutex               = nullptr;
    fiSplineWidget *fiSplineToInit      = nullptr;
    const QVector<QPoint> *customKnots  = nullptr;
    bool customCameraLutBuilt           = false;
    DISPLAY_LUT::LUT_SPLINE_RECORD customSplineRec;

    // Set the id to zero for empty and no custom information.
    customSplineRec.id = 0;

    // Do we have custom lut information?
    if (playbackLut)
    {
        playbackLut->CopySplineRecord(customSplineRec);
    }

    // Build the requested camera lut.
    switch (curveTabIndex)
    {
    case SPLINE_TAB_INDEX_RED_OR_MONO:
        fiSplineToInit = ui->fi_spline;
        aaSplineToInit = aaSplineRedOrGamma;
        aaSplineMutex = &aaSplineRedOrGammaMutex;
        if (customSplineRec.id > 0)
        {
            customKnots = &customSplineRec.redOrGammaKnots;
        }
        break;
    case SPLINE_TAB_INDEX_GREEN:
        fiSplineToInit = ui->fiSplineGreen;
        aaSplineToInit = aaSplineGreen;
        aaSplineMutex = &aaSplineGreenMutex;
        if (customSplineRec.id > 0)
        {
            customKnots = &customSplineRec.greenKnots;
        }
        break;
    case SPLINE_TAB_INDEX_BLUE:
        fiSplineToInit = ui->fiSplineBlue;
        aaSplineToInit = aaSplineBlue;
        aaSplineMutex = &aaSplineBlueMutex;
        if (customSplineRec.id > 0)
        {
            customKnots = &customSplineRec.blueKnots;
        }
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No curveTabIndex provided.";
        goto exitInitializeSpline;
    }

    // Make sure the spline widget provided is a valid pointer.
    if (fiSplineToInit == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fiSplineToInit was a nullptr!";
        goto exitInitializeSpline;
    }

    // Make sure the spline curve provided is a valid pointer.
    if (aaSplineToInit == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "aaSplineToInit was a nullptr!";
        goto exitInitializeSpline;
    }

    // Make sure the camera lut vector is a valid pointer.
    if (aaSplineMutex == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "aaSplineMutex was a nullptr!";
        goto exitInitializeSpline;
    }

    //Setup the initial layout of the curve.
    {
        aaSplineMutex->lock();

        // Clear the spline.
        aaSplineToInit->knots.clear();

        // Force the spline to limit itself to our bounds.
        aaSplineToInit->setLimit(0, LUT_LENGTH - 1, LUT_MAX_TABLE_COUNT - 1, 0);

        // Is there a metadata knot configuration?
        if (customKnots)
        {
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Custom Knots:";
            }

            // Transfer each knot into a spline point.
            // Note: It is assumed the knots are validated knots based on the internal
            // collection established when created in the curve settings factory.
            for (int knotId = 0; knotId < customKnots->count(); knotId++)
            {
                aaSplineToInit->addKnots(aaAaa::aaPoint((float)customKnots->at(knotId).x(), (float)customKnots->at(knotId).y()));
                if (VERBOSE_LOGGING)
                {
                    QLOG_DEBUG() << "[" << customKnots->at(knotId).x() << "," << customKnots->at(knotId).y() << "]";
                }
            }
        }
        else
        {
            // Add knots at the beginning, middle, and end creating a linear line.
            aaSplineToInit->addKnots(aaAaa::aaPoint(0, 0));
            aaSplineToInit->addKnots(aaAaa::aaPoint(LUT_MAX_TABLE_COUNT / 2, LUT_LENGTH / 2));
            aaSplineToInit->addKnots(aaAaa::aaPoint(LUT_MAX_TABLE_COUNT - 1, LUT_LENGTH - 1));
        }
        // Set the resolution of our spline.
        fiSplineToInit->set_increment(0.1);

        // Pass the data on to our control.
        fiSplineToInit->set_spline(aaSplineToInit);

        // Make sure we limit the maximum number of knots on the spline.
        fiSplineToInit->set_max_knot_count(LUT_MAX_KNOT_COUNT);

        aaSplineMutex->unlock();
    }

#if 0
    // If using custom knots from a metadata lut we should be able to get the
    // actual camera LUT data without having to build it.
    if (customKnots && playbackLut)
    {
        bool builtBlueCameraLut         = false;
        bool builtGreenCameraLut        = false;
        bool builtRedOrGammaCameraLut   = false;

        // Build the requested camera lut.
        switch (curveTabIndex)
        {
        case SPLINE_TAB_INDEX_RED_OR_MONO:
        {
            const QVector<double>& redOrGammaCameraLut = playbackLut->GetRedOrGammaCameraLut(builtRedOrGammaCameraLut);
            if (builtRedOrGammaCameraLut)
            {
                // These can be copied to our own camera lut.
                cameraLutRedOrGamma = redOrGammaCameraLut;
                HandleBuildACameraLutCompleted(curveTabIndex);
                customCameraLutBuilt = true;
            }
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Custom RED camera LUT:" << (builtRedOrGammaCameraLut ? "Exists" : "Does Not Exist");
            }
        }
        break;
        case SPLINE_TAB_INDEX_GREEN:
        {
            const QVector<double>& greenCameraLut = playbackLut->GetGreenCameraLut(builtGreenCameraLut);
            if (builtGreenCameraLut)
            {
                // These can be copied to our own camera lut.
                cameraLutGreen = greenCameraLut;
                HandleBuildACameraLutCompleted(curveTabIndex);
                customCameraLutBuilt = true;
            }
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Custom Green camera LUT:" << (builtGreenCameraLut ? "Exists" : "Does Not Exist");
            }
        }
        break;
        case SPLINE_TAB_INDEX_BLUE:
        {
            const QVector<double>& blueCameraLut = playbackLut->GetBlueCameraLut(builtBlueCameraLut);
            if (builtBlueCameraLut)
            {
                // These can be copied to our own camera lut.
                cameraLutBlue = blueCameraLut;
                HandleBuildACameraLutCompleted(curveTabIndex);
                customCameraLutBuilt = true;
            }
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Custom Blue camera LUT:" << (builtBlueCameraLut ? "Exists" : "Does Not Exist");
            }
        }
        break;
        default:
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No curveTabIndex provided.";
        }
    }
#endif

    // If we haven't already pulled custom camera lut data we will need to build it.
    if (!customCameraLutBuilt)
    {
        // Fire off the worker thread to build up the camera lut data.
        QFuture<void> future = QtConcurrent::run(BuildACameraLut, this, curveTabIndex);
    }
    else
    {
        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Using custom camera LUT skipped BuildACameraLut.";
        }
    }

exitInitializeSpline:

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Leaving.";
    }
}

/*!
 * @brief Load the curve information into the display based on the specified radio "id".
 * @param lut_id    Which LUT_SPLINE_RECORD to load (corresponds to a radio "slot" number).
 */
void CurveSettings::LoadLutSplineRec(int lut_id)
{
    bool foundSplineRec = false;
    int i;

    lutSplinexMutex.lock();

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut_id received:" << lut_id << "count:" << lutSplineRecs.count();
    }
    // Find out which LUT_SPLINE_RECORD we're looking at.
    for (i = 0; i < lutSplineRecs.count(); i++)
    {
        DISPLAY_LUT::LUT_SPLINE_RECORD& lutSplineRef = lutSplineRecs[i];


        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Comparing index: " << i << "lutSplineRef.id :" << lutSplineRef.id  << "with lut_id:" << lut_id;
        }

        // Does the stored spline record's id match the radio button slot id?
        if (lutSplineRef.id == lut_id)
        {
            const QVector<QPoint> *knot_ptr;

            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "MATCH id:" << lutSplineRef.id
                             << "lutDataType:" << lutSplineRef.lutDataType
                             << "desc:" << lutSplineRef.desc;
            }

            foundSplineRec = true;

            // Update the description.
            ui->eb_description->setText(lutSplineRecs.at(i).desc);

            // Update the local path.
            if (lutSplineRecs.at(i).localPath.isEmpty())
            {
                ui->eb_path->setText("");
                ui->eb_path->setVisible(false);
            }
            else
            {
                ui->eb_path->setText(lutSplineRecs.at(i).localPath);
                ui->eb_path->setVisible(true);
            }

            // Make sure our spline record matches our dialog setup.
            if (UsingRGBSplines())
            {
                // If our spline record doesn't have 3 curves copy the red to the green and blue.
                if (lutSplineRef.lutDataType != LUT_SERIESNC_COLOR)
                {

                    if (VERBOSE_LOGGING)
                    {
                        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Datatype: " << lutSplineRef.lutDataType
                                     << "not registered as LUT_SERIESNC_COLOR:" << LUT_SERIESNC_COLOR;
                    }
                    // Modify the spline record to be in sync with dialog setup by reproducing red channel for green and blue.
                    lutSplineRef.lutDataType    = LUT_SERIESNC_COLOR;
                    lutSplineRef.greenKnots     =  lutSplineRef.redOrGammaKnots;
                    lutSplineRef.blueKnots      =  lutSplineRef.redOrGammaKnots;
                }
            }
            else
            {
                // Set the record type to match the dialog setup for single channel LUT.
                if (isSeries3Camera)
                    lutSplineRef.lutDataType    = LUT_SERIES3;
                else
                    lutSplineRef.lutDataType    = LUT_SERIESNC_MONO;
            }

            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "index:" << i
                             << "lutDataType:" << lutSplineRef.lutDataType
                             << "lutId:" << lutSplineRef.id;
            }

            // What kind of curve is this a 3-channel rgb or gamma?
            if (lutSplineRef.lutDataType == LUT_SERIESNC_COLOR)
            {
                // There are three curves to update.
                aaSplineRedOrGamma->knots.clear();
                knot_ptr = &lutSplineRef.redOrGammaKnots;
                for (int k = 0; k < knot_ptr->count(); k++)
                {
                    // Add this knot to the spline.
                    aaSplineRedOrGamma->addKnots(aaAaa::aaPoint((float)knot_ptr->at(k).x(), (float)knot_ptr->at(k).y()));
                    if (VERBOSE_LOGGING)
                    {
                        QLOG_DEBUG() << "[" << knot_ptr->at(k).x() << "," << knot_ptr->at(k).y() << "]";
                    }
                }
                aaSplineGreen->knots.clear();
                knot_ptr = &lutSplineRef.greenKnots;
                for (int k = 0; k < knot_ptr->count(); k++)
                {
                    // Add this knot to the spline.
                    aaSplineGreen->addKnots(aaAaa::aaPoint((float)knot_ptr->at(k).x(), (float)knot_ptr->at(k).y()));
                    if (VERBOSE_LOGGING)
                    {
                        QLOG_DEBUG() << "[" << knot_ptr->at(k).x() << "," << knot_ptr->at(k).y() << "]";
                    }
                }
                aaSplineBlue->knots.clear();
                knot_ptr = &lutSplineRef.blueKnots;
                for (int k = 0; k < knot_ptr->count(); k++)
                {
                    // Add this knot to the spline.
                    aaSplineBlue->addKnots(aaAaa::aaPoint((float)knot_ptr->at(k).x(), (float)knot_ptr->at(k).y()));
                    if (VERBOSE_LOGGING)
                    {
                        QLOG_DEBUG() << "[" << knot_ptr->at(k).x() << "," << knot_ptr->at(k).y() << "]";
                    }
                }
                ui->fi_spline->set_spline(aaSplineRedOrGamma);
                ui->fiSplineGreen->set_spline(aaSplineGreen);
                ui->fiSplineBlue->set_spline(aaSplineBlue);

                // If knots are different between channels don't keep the lock button active as it will cause
                // a loss of state go ahead in this case set the spline lock false.
                splineIsLocked = FiDisplayLut::KnotsAreEqual(lutSplineRef);

                if (VERBOSE_LOGGING)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "LOADING LUT splineIsLocked:" << splineIsLocked;
                }
                ui->tb_curve_lock->setChecked(splineIsLocked);
            }
            // There is only one spline curve to setup.
            else
            {
                // Update the curve.
                aaSplineRedOrGamma->knots.clear();
                knot_ptr = &lutSplineRef.redOrGammaKnots;
                for (int k = 0; k < knot_ptr->count(); k++)
                {
                    // Add this knot to the spline.
                    aaSplineRedOrGamma->addKnots(aaAaa::aaPoint((float)knot_ptr->at(k).x(), (float)knot_ptr->at(k).y()));
                    if (VERBOSE_LOGGING)
                    {
                        QLOG_DEBUG() << "[" << knot_ptr->at(k).x() << "," << knot_ptr->at(k).y() << "]";
                    }
                }
                // Pass the data on to our control.
                ui->fi_spline->set_spline(aaSplineRedOrGamma);
            }

            // Set up the lock based on the curve data.
            // If knots are different between channels don't keep the lock button active as it will cause
            // a loss of state go ahead in this case set the spline lock false.
            splineIsLocked = FiDisplayLut::KnotsAreEqual(lutSplineRef);
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Loaded spline " << lutSplineRef.id
                             << " setting splineIsLocked:" << splineIsLocked;
            }
            ui->tb_curve_lock->setChecked(splineIsLocked);
            // All done.
            break;
        }
    }
    lutSplinexMutex.unlock();

    // If no spline record was found just initialize as a default setup.
    if (!foundSplineRec)
    {
        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "NO MATCH found for id:" << lut_id << "calling InitializeDefaultSetup().";
        }
        InitializeDefaultSetup();
    }
}

/*!
 * @brief This is the worker thread for processing the LUT changes. Because this process
 *        is very time consuming, we use this worker thread to work in parallel with the main GUI
 *        thread.
 *
 * @param parent    Pointer to the [parent](@ref CurveSettings) object.
 */
void CurveSettings::LutProcessing(QObject *parent)
{
    CurveSettings   *my_class = qobject_cast<CurveSettings *>(parent);

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    {
        bool lutProcessingIsAllowed = true;
        my_class->allowLutProcessingMutex.lock();
        lutProcessingIsAllowed = my_class->allowLutProcessing;
        my_class->allowLutProcessingMutex.unlock();
        if (!lutProcessingIsAllowed)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "I was called to process but lutProcessingIsAllowed:" << lutProcessingIsAllowed;
            return;
        }
    }

    // Make sure we are build display luts for the most current change.
    // Note that even though the fiSplineWidget UI graphs may need to be refreshed the
    // camera LUTs for any locked curve matches the data being set here so it is valid
    // to build up the display look up tables here.
    my_class->BuildDisplayLuts();

    // Update the histogram and the display image.
    my_class->UpdateHistogram();

    // Let folks know we are done.
    emit my_class->LutProcessingComplete();
}

/*!
 * @brief This is the worker thread for translating a fiSplineWidget Curve into a camera LUT.
 * Because this process is time consuming, a worker thread is being used to work in parallel with
 * the main GUI thread.
 *
 * @param parent    Pointer to the [parent](@ref CurveSettings) object.
 * @param curveTabIndex - which curve is being created?
*/
void CurveSettings::BuildACameraLut(QObject *parent, int curveTabIndex)
{
    CurveSettings   *my_class = qobject_cast<CurveSettings *>(parent);
    int cameraLutId = SPLINE_TAB_NONE;      // Set to none until one is built.

    fiSplineWidget *fiSplineToChange = nullptr;
    std::shared_ptr<aaAaa::aaSpline> aaSplineCurve;
    QVector<double> *cameraLutForSpline = nullptr;
    QMutex *cameraLutMutex = nullptr;

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        goto exitBuildCameraLuts;
    }

    // Build the requested camera lut.
    switch (curveTabIndex)
    {
    case SPLINE_TAB_INDEX_RED_OR_MONO:
        fiSplineToChange = my_class->ui->fi_spline;
        aaSplineCurve = my_class->aaSplineRedOrGamma;
        cameraLutForSpline = &my_class->cameraLutRedOrGamma;
        cameraLutMutex = &my_class->cameraLutRedOrGammaMutex;
        break;
    case SPLINE_TAB_INDEX_GREEN:
        fiSplineToChange = my_class->ui->fiSplineGreen;
        aaSplineCurve = my_class->aaSplineGreen;
        cameraLutForSpline = &my_class->cameraLutGreen;
        cameraLutMutex = &my_class->cameraLutGreenMutex;
        break;
    case SPLINE_TAB_INDEX_BLUE:
        fiSplineToChange = my_class->ui->fiSplineBlue;
        aaSplineCurve = my_class->aaSplineBlue;
        cameraLutForSpline = &my_class->cameraLutBlue;
        cameraLutMutex = &my_class->cameraLutBlueMutex;
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No curveTabIndex provided.";
        goto exitBuildCameraLuts;
    }

    // Make sure the spline widget provided is a valid pointer.
    if (fiSplineToChange == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fiSplineToChange was a nullptr!";
        goto exitBuildCameraLuts;
    }

    // Make sure the spline curve provided is a valid pointer.
    if (aaSplineCurve == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "aaSplineCurve was a nullptr!";
        goto exitBuildCameraLuts;
    }

    // Make sure the camera lut vector is a valid pointer.
    if (cameraLutForSpline == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cameraLutForSpline was a nullptr!";
        goto exitBuildCameraLuts;
    }

    // Make sure we have a camera lut mutex.
    if (cameraLutMutex == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cameraLutMutex was a nullptr!";
        goto exitBuildCameraLuts;
    }

    // Handle the building of the camera lut:
    {
        cameraLutMutex->lock();
        // Read the LUT information from the spline widget using the spline curve control attribute.
        if (fiSplineToChange->get_data((int)(aaSplineCurve->limit_right + 1) / LUT_MAX_TABLE_COUNT, *cameraLutForSpline) < 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to read the LUT from " << fiSplineToChange->objectName() << " control!";
            cameraLutId = SPLINE_TAB_NONE;
        }
        else
        {
            cameraLutId = curveTabIndex;
        }
        cameraLutMutex->unlock();
    }

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cameraLUT:" << *cameraLutForSpline;

exitBuildCameraLuts:

    // Let folks know we are done.
    emit my_class->BuildACameraLutCompleted(cameraLutId);
}


/*!
 * @brief Start the process of downloading the file listing for the lookup tables, if any.
 */
void CurveSettings::listing_start()
{
    QString         xml_filename;

    // Get a pointer to the communications object.
    FI_comm     *comm_ptr;
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Hook up the read/write completion signals.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &CurveSettings::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &CurveSettings::write_complete, Qt::QueuedConnection);

    // Cast our parent to a form we can use.
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Initiate the reading of the LUT(s) from the camera.
    my_parent->write_camera_register(GIGE_REG_FILE_ACTION, GIGE_FA_LUT_LISTING);

    // Create the path name for the XML file we are going to create.
    xml_filename = QDir::tempPath() + QStringLiteral("/lut_listing.xml");

    // Remove any old file that might still be laying around.
    QFile::remove(xml_filename);

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "xml_filename:" << xml_filename;
    }

    // Get ready for the new file.
    listing_file.setFileName(xml_filename);
    if (!listing_file.open(QIODevice::WriteOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open" << xml_filename << ". Err =" << listing_file.errorString();
        return;
    }
    listing_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
}

/*!
 * @brief Parse DOM knots into the vector that will hold them.
 */
void CurveSettings::ParseDOMKnots(QDomElement& DOMKnots, QVector<QPoint>& knotArray, const QString& knotDOMString)
{
    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "knotDOMString:" << knotDOMString;
    }

    while (!DOMKnots.isNull())
    {
        QPoint  coord;
        QString knot_str = DOMKnots.text();

        // Split the coordinate pair string into two parts (X and Y).
        QStringList point_list = knot_str.split(QChar(':'));

        // Convert the string list to an actual point.
        coord.setX(point_list.at(0).toInt());
        coord.setY(point_list.at(1).toInt());

        // Append that to our information.
        knotArray.append(coord);

        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << "[" << coord.x() << "," << coord.y() << "]";
        }

        // Get the next "knot" in the list.
        DOMKnots = DOMKnots.nextSiblingElement(knotDOMString);
    }
}

/*!
 * @brief Parse the XML file, listing_file, which contains the lookup table information.
 */
void CurveSettings::parse_listing()
{
    QString         msg_str;
    QDomDocument    doc(LUT_SPLINE_LIST_DOM_DOCUMENT);
    QDomElement     dom_root;           // xml token LUT_SPLINE_LIST_DOM_ELEMENT
    QDomElement     dom_child;          // xml token LUT_SPLINE_REC_ACTIVE_DOM_ELEMENT or LUT_SPLINE_REC_DOM_ELEMENT

    // Open our file so the XML parser can read it.
    if (!listing_file.open(QIODevice::ReadOnly))
    {
        msg_str = QStringLiteral("Error opening returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(listing_file.error()).arg(listing_file.errorString());
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        return;
    }

    // Set the XML file to the root document of the XML parser.
    if (!doc.setContent(&listing_file))
    {
        msg_str = QStringLiteral("Error parsing returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(errno).arg(QString::fromUtf8(strerror(errno)));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;

        listing_file.close();
        return;
    }

    // Don't need this anymore it's been read in.
    listing_file.close();

    // Get the root object.
    dom_root = doc.firstChildElement(LUT_SPLINE_LIST_DOM_ELEMENT);

    // Now get the active ID.
    dom_child = dom_root.firstChildElement(LUT_SPLINE_REC_ACTIVE_DOM_ELEMENT);
    int active_id = dom_child.attribute(LUT_SPLINE_REC_DOM_ID, QStringLiteral("0")).toInt();
    if (active_id == 0)
    {
        // There wasn't an active ID, so try something intelligent.
        if (old_slot != 0)
            active_id = old_slot;
        else
            active_id = 1;
    }

    // Clear out the existing LUT array.
    lutSplineRecs.clear();
    // Walk through the lut information.
    dom_child = dom_root.firstChildElement(LUT_SPLINE_REC_DOM_ELEMENT);
    while (!dom_child.isNull())
    {
        DISPLAY_LUT::LUT_SPLINE_RECORD    lutSplineRec;      // Holding spot for the information.

        QDomElement dom_knots;                  // xml token "*knots" reused for each channel if !useMono.
        QDomElement knot_info;                  // xml token "*knot" reused for each channel if !useMono.

        // Read the ID for this LUT.
        lutSplineRec.id = dom_child.attribute(LUT_SPLINE_REC_DOM_ID, QStringLiteral("0")).toInt();
        if (lutSplineRec.id == 0)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Malformed XML: expecting an ID for this lut_info.";
        }

        // Get the description information.
        lutSplineRec.desc = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_DESC).text();
        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lutSplineRec.desc:" << lutSplineRec.desc;
        }

        // Is this a rgb setting?  Default to zero if not found to indicate only one channel of data.
        // Look for the lut data type to determine number of channels.
        // NOTE:  This had been using dom_child.attribute but it doesn't work right.
        QString lutDataTypeString = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_LUT_DATA_TYPE).text();
        lutSplineRec.lutDataType = (uint8_t)lutDataTypeString.toInt();

        // If the LUT data type is unknown default it to series 3.
        if (lutSplineRec.lutDataType >= LUT_SERIES_UNKNOWN)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Parsed lutDataType:" << lutSplineRec.lutDataType
                        << " Setting it to safe default LUT_SERIES3";
            lutSplineRec.lutDataType = LUT_SERIES3;
        }

        // We always expect to find the single spline data in the xml data.
        dom_knots = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOTS);
        knot_info = dom_knots.firstChildElement(LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOT);
        ParseDOMKnots(knot_info, lutSplineRec.redOrGammaKnots, LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOT);

        // Make sure that the dialog setup matches the data being parsed.
        if (UsingRGBSplines())
        {
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RGB: lutSplineRec.lutDataType:" << lutSplineRec.lutDataType;
            }

            // If the DOM isn't set up for RGB duplicate the red channel for green and blue.
            if (lutSplineRec.lutDataType != LUT_SERIESNC_COLOR)
            {
                lutSplineRec.lutDataType = LUT_SERIESNC_COLOR;
                //lutSplineRec.greenKnots = lutSplineRec.redOrGammaKnots;
                //lutSplineRec.blueKnots = lutSplineRec.redOrGammaKnots;
            }
            //else
            {
                // The green spline data should be in this record.
                dom_knots = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_GREEN_KNOTS);
                knot_info = dom_knots.firstChildElement(LUT_SPLINE_REC_DOM_GREEN_KNOT);
                ParseDOMKnots(knot_info, lutSplineRec.greenKnots, LUT_SPLINE_REC_DOM_GREEN_KNOT);

                // And the blue spline data should also be in this record.
                dom_knots = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_BLUE_KNOTS);
                knot_info = dom_knots.firstChildElement(LUT_SPLINE_REC_DOM_BLUE_KNOT);
                ParseDOMKnots(knot_info, lutSplineRec.blueKnots, LUT_SPLINE_REC_DOM_BLUE_KNOT);
            }
        }

        bool knotsEqual = FiDisplayLut::KnotsAreEqual(lutSplineRec);
        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Appending lut id:" << lutSplineRec.id
                         << "desc:" << lutSplineRec.desc
                         << "localPath:" << lutSplineRec.localPath
                         << "lutDataType:" << lutSplineRec.lutDataType
                         << "knotsEqual:" << knotsEqual;
        }

        // Append the new LUT to our list.
        lutSplineRecs.append(lutSplineRec);

        // Get the next LUT in the list.
        dom_child = dom_child.nextSiblingElement(LUT_SPLINE_REC_DOM_ELEMENT);
    }

    // Now that everything is setup, setup the active LUT.
    SetupDestinationLut(active_id);
}

void CurveSettings::RestoreFromQSettings()
{
    int         active_id;
    int         lut_count;
    int         i;
    QSettings   settings;
    DISPLAY_LUT::LUT_SPLINE_RECORD    lutSplineRec;     // The lut spline record reusable buffer.

    lutSplinexMutex.lock();

    // Clear out the existing LUT array.
    lutSplineRecs.clear();

    // Initialize our data space.
    settings.beginGroup(LUT_SPLINE_SETTINGS_GROUP);
    active_id = settings.value(LUT_SPLINE_SETTINGS_ACTIVE_ID, 0).toInt();
    lut_count = settings.value(LUT_SPLINE_SETTINGS_COUNT, 0).toInt();

    // If we are overriding with a playback lut set the active id to that one.
    if (playbackLut)
    {
        active_id = playbackLut->GetLutId();
    }

    // Loop through the entries and get the information.
    for (i = 0; i < lut_count; i++)
    {
        bool    usePlaybackRecord = false;
        int     k;
        int     knot_count;

        // Setup for this entry.
        settings.beginGroup(QString(LUT_SPLINE_SETTINGS_GROUP_LUT_ID).arg(i));

        // Read the ID for this LUT.
        lutSplineRec.id  = settings.value(LUT_SPLINE_SETTINGS_LUT_ID, 0).toInt();

        // Do we have a playback LUT?  If so it will supercede any settings that
        // exist in our application registry.  It is part of the original playback
        // video and is maintained here.
        if (playbackLut)
        {
            if (lutSplineRec.id == playbackLut->GetLutId())
            {
                // We will skip this one in from the QSettings.
                usePlaybackRecord = true;
            }
        }

        if (!usePlaybackRecord)
        {
            // Get the description.
            lutSplineRec.desc = settings.value(LUT_SPLINE_SETTINGS_DESC).toString();

            // Get rgb state.  Defaults to single channel if not found otherwise maintains setting if found = 1.
            lutSplineRec.lutDataType = (uint8_t)settings.value(LUT_SPLINE_SETTINGS_LUT_DATA_TYPE, LUT_SPLINE_SETTINGS_TYPE_UNKNOWN).toInt();

            // If the LUT data type is unknown default it to series 3.
            if (lutSplineRec.lutDataType == (uint8_t)LUT_SPLINE_SETTINGS_TYPE_UNKNOWN.toUInt())
            {
                lutSplineRec.lutDataType = LUT_SERIES3;
            }

            // Load the "knots" for the red (or gamma) channel.
            knot_count = settings.beginReadArray(LUT_SPLINE_SETTINGS_KNOTS);
            for (k = 0; k < knot_count; k++)
            {
                settings.setArrayIndex(k);
                lutSplineRec.redOrGammaKnots.append(settings.value(LUT_SPLINE_SETTINGS_KNOTS).toPoint());
            }
            settings.endArray();

            // What type of data does our dialog want to support?
            if (UsingRGBSplines())
            {
                // We are expecting three channels of data.  If the QSettings data
                // does not have all three channels just copy the red channel for the
                // green and blue ones.
                if (lutSplineRec.lutDataType != LUT_SERIESNC_COLOR)
                {
                    // Set the data type to what is expected for this dialog.
                    lutSplineRec.lutDataType = LUT_SERIESNC_COLOR;

                    // Copy over the red channel for green and blue.
                    lutSplineRec.greenKnots = lutSplineRec.redOrGammaKnots;
                    lutSplineRec.blueKnots = lutSplineRec.redOrGammaKnots;
                }
                else
                {
                    // Assume the green and blue channel data is in the QSettings.

                    // Load the "knots" for the green spline.
                    knot_count = settings.beginReadArray(LUT_SPLINE_SETTINGS_GREEN_KNOTS);
                    for (k = 0; k < knot_count; k++)
                    {
                        settings.setArrayIndex(k);
                        lutSplineRec.greenKnots.append(settings.value(LUT_SPLINE_SETTINGS_GREEN_KNOTS).toPoint());
                    }
                    settings.endArray();

                    // Load the "knots" for the blue spline.
                    knot_count = settings.beginReadArray(LUT_SPLINE_SETTINGS_BLUE_KNOTS);
                    for (k = 0; k < knot_count; k++)
                    {
                        settings.setArrayIndex(k);
                        lutSplineRec.blueKnots.append(settings.value(LUT_SPLINE_SETTINGS_BLUE_KNOTS).toPoint());
                    }
                    settings.endArray();
                }
            }
            // This dialog is not expecting 3 channels.
            else
            {
                // Make sure we set the data type to match dialog setup.  We have already
                // pulled the red/mono channel above so we are done.
                if (isSeries3Camera)
                    lutSplineRec.lutDataType = LUT_SERIES3;
                else
                    lutSplineRec.lutDataType = LUT_SERIESNC_MONO;
            }
        }
        // The LUT group is completed.
        settings.endGroup();

        // If we are using a playback LUT we do nothing here because it was added
        // when the curve dialog was created.
        if (!usePlaybackRecord)
        {
            bool knotsEqual = FiDisplayLut::KnotsAreEqual(lutSplineRec);
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Appending QSettings lut id:" << lutSplineRec.id << "knotsEqual:" << knotsEqual;
            }
            // Append the new LUT_SPLINE_RECORD to our list.
            lutSplineRecs.append(lutSplineRec);
        }
    }

    // All done.
    settings.endGroup();

    // If we have a playback custom LUT don't forget to add it because we skipped any
    // record that may have been in QSettings to override with this one.
    if (playbackLut)
    {
        // We want this record not the one in the settings.
        playbackLut->CopySplineRecord(lutSplineRec);
        bool knotsEqual = FiDisplayLut::KnotsAreEqual(lutSplineRec);
        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Appending Playback lut id:" << lutSplineRec.id
                         << "knotsEqual:" << knotsEqual
                         << "lutDataType:" << lutSplineRec.lutDataType;
        }
        lutSplineRecs.append(lutSplineRec);
    }

    lutSplinexMutex.unlock();

    // Now that everything is setup, pick the active LUT.
    if (active_id > 0)
        SetupDestinationLut(active_id);
}

void CurveSettings::SaveToQSettings()
{
    int         i;
    QSettings   settings;

    // Initialize our data space.
    settings.beginGroup(LUT_SPLINE_SETTINGS_GROUP);
    settings.setValue(LUT_SPLINE_SETTINGS_ACTIVE_ID, GetCheckedRadioLutId());
    settings.setValue(LUT_SPLINE_SETTINGS_COUNT, lutSplineRecs.size());

    // Loop through the entries and get the information.
    for (i = 0; i < lutSplineRecs.size(); i++)
    {
        int     k;

        // Setup for this entry.
        settings.beginGroup(QString(LUT_SPLINE_SETTINGS_GROUP_LUT_ID).arg(i));

        // Save the ID for this LUT.
        settings.setValue(LUT_SPLINE_SETTINGS_LUT_ID, lutSplineRecs[i].id);

        // Save the description.
        settings.setValue(LUT_SPLINE_SETTINGS_DESC, lutSplineRecs[i].desc);

        // Save the number of knots depending on the type of LUT data.
        settings.setValue(LUT_SPLINE_SETTINGS_LUT_DATA_TYPE, lutSplineRecs[i].lutDataType);

        // Save the red or gamma "knots".
        settings.beginWriteArray(LUT_SPLINE_SETTINGS_KNOTS);
        for (k = 0; k < lutSplineRecs[i].redOrGammaKnots.size(); ++k)
        {
            settings.setArrayIndex(k);
            settings.setValue(LUT_SPLINE_SETTINGS_KNOTS, lutSplineRecs[i].redOrGammaKnots[k]);
        }
        settings.endArray();

        // If this is an rgb condition save the green and blue knots.
        if (lutSplineRecs[i].lutDataType == LUT_SERIESNC_COLOR)
        {
            // Save the green "knots".
            settings.beginWriteArray(LUT_SPLINE_SETTINGS_GREEN_KNOTS);
            for (k = 0; k < lutSplineRecs[i].greenKnots.size(); ++k)
            {
                settings.setArrayIndex(k);
                settings.setValue(LUT_SPLINE_SETTINGS_GREEN_KNOTS, lutSplineRecs[i].greenKnots[k]);
            }
            settings.endArray();

            // Save the blue "knots".
            settings.beginWriteArray(LUT_SPLINE_SETTINGS_BLUE_KNOTS);
            for (k = 0; k < lutSplineRecs[i].blueKnots.size(); ++k)
            {
                settings.setArrayIndex(k);
                settings.setValue(LUT_SPLINE_SETTINGS_BLUE_KNOTS, lutSplineRecs[i].blueKnots[k]);
            }
            settings.endArray();
        }
        // End of the lut spline record in the array of lut spline records.
        settings.endGroup();
    }

    // All done.
    settings.endGroup();
}

/*!
 * @brief Save the spline curve information into the specified LUT_SPLINE_RECORD.
 * @param lut_id    Which LUT_SPLINE_RECORD to overwrite in lutSplineRecs vector.
 */
void CurveSettings::SaveSplineData(int lut_id)
{
    int index;

    // Find the associated LUT.
    for (index = 0; index < lutSplineRecs.count(); index++)
    {
        if (lut_id == lutSplineRecs.at(index).id)
            break;
    }

    // If we didn't find the existing lut in the lutSplineRecs
    // we need to append an "empty" one so that the dialog settings
    // can be preserved.
    if (index == lutSplineRecs.count())
    {
        DISPLAY_LUT::LUT_SPLINE_RECORD    empty_lut;

        // Add an empty LUT and point at that.
        empty_lut.id = lut_id;
        empty_lut.lutDataType = LUT_SERIES3;

        lutSplineRecs.append(empty_lut);
    }

    // Our index is now referencing a LUT. Save the current information.
    lutSplineRecs[index].desc = ui->eb_description->text();

    // Are we an RGB or Gamma curve?
    if (UsingRGBSplines())
    {
        lutSplineRecs[index].lutDataType = LUT_SERIESNC_COLOR;
    }
    else
    {
        // Distinguish between a series3 camera and seriesNC monochrome.
        if (isSeries3Camera)
            lutSplineRecs[index].lutDataType = LUT_SERIES3;
        else
            lutSplineRecs[index].lutDataType = LUT_SERIESNC_MONO;
    }

    if (UsingRGBSplines())
    {
        lutSplineRecs[index].redOrGammaKnots.clear();
        for (int i = 0; i < (int)aaSplineRedOrGamma->knots.size(); i++)
        {
            aaAaa::aaPoint  knot = aaSplineRedOrGamma->knots.at((uint64_t)i);
            QPoint  coord((int)knot.t, (int)knot.y);

            // Append this coordinate to our LUT.
            lutSplineRecs[index].redOrGammaKnots.append(coord);
        }

        lutSplineRecs[index].greenKnots.clear();
        for (int i = 0; i < (int)aaSplineGreen->knots.size(); i++)
        {
            aaAaa::aaPoint  knot = aaSplineGreen->knots.at((uint64_t)i);
            QPoint  coord((int)knot.t, (int)knot.y);

            // Append this coordinate to our LUT.
            lutSplineRecs[index].greenKnots.append(coord);
        }

        lutSplineRecs[index].blueKnots.clear();
        for (int i = 0; i < (int)aaSplineBlue->knots.size(); i++)
        {
            aaAaa::aaPoint  knot = aaSplineBlue->knots.at((uint64_t)i);
            QPoint  coord((int)knot.t, (int)knot.y);

            // Append this coordinate to our LUT.
            lutSplineRecs[index].blueKnots.append(coord);
        }
    }
    // Gamma usage.
    else
    {
        lutSplineRecs[index].redOrGammaKnots.clear();
        for (int i = 0; i < (int)aaSplineRedOrGamma->knots.size(); i++)
        {
            aaAaa::aaPoint  knot = aaSplineRedOrGamma->knots.at((uint64_t)i);
            QPoint  coord((int)knot.t, (int)knot.y);

            // Append this coordinate to our LUT.
            lutSplineRecs[index].redOrGammaKnots.append(coord);
        }
    }
}

/*!
 * @brief Select a specific lookup table. This makes sure the appropriate radio
 *        button is selected, and if requested the specified lookup table is
 *        located and loaded for use in the dialog.
 * @param lut_id        The id of the lookup table to load.
 * @param loadLutRec    True if caller wants to load matching id from lut spline record array.
 */
void CurveSettings::SetupDestinationLut(int lut_id, bool loadLutRec)
{
    // Set the appropriate radio button.
    switch (lut_id)
    {
    case 1:
        ui->rb_1->setChecked(true);
        break;
    case 2:
        ui->rb_2->setChecked(true);
        break;
    case 3:
        ui->rb_3->setChecked(true);
        break;
    case 4:
        ui->rb_4->setChecked(true);
        break;
    case 5:
        ui->rb_5->setChecked(true);
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unknown lut:" << lut_id;
        lut_id = 1;
        ui->rb_1->setChecked(true);
        break;
    }

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut_id:" << lut_id << "old_slot:" << old_slot;
    }

    // Update the curve to match either a saved record or an empty one if none found.
    if (loadLutRec)
        LoadLutSplineRec(lut_id);

    // Remember the new "slot".
    old_slot = lut_id;
}

/*!
 * @brief Get a copy of the image we're going to use.
 * @param snap_image    The pixmap of our snapshot.
 */
void CurveSettings::set_image(QPixmap& snap_image)
{
    // Squirrel away the original pixmap.
    if (snap_pixmap != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Non-empty snap_pixmap!?";
        delete snap_pixmap;
    }
    snap_pixmap = new fiPixmap(snap_image.copy());
}

/*!
 * @brief Check if the RGB curve group is locked.
 *
 * If the group is locked, then one curve modification will effect the
 * other two so they move in unison.
 *
 * @return True if the control is enabled and checked.
 */
bool CurveSettings::IsCurveLocked()
{
    // The group is locked if the control is enabled and checked.
    return (ui->tb_curve_lock->isEnabled() && ui->tb_curve_lock->isChecked());
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void CurveSettings::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        ui->retranslateUi(this);
    }

    // Pass it on.
    QDialog::changeEvent(event);
}


/*!
 * @brief Update our image during a resize of the dialog.
 * @param event     [UNUSED]
 */
void CurveSettings::resizeEvent(QResizeEvent *event)
{
    // Ignore this until we are ready to display the image.
    if (!okToUpdateDisplayImage)
        return;

    // Sanity check.
    if ((scene_ptr != nullptr) && (mod_pixmap != nullptr))
    {
        // Create the new pixmap item.
        if (pixmap_item_ptr == nullptr)
            pixmap_item_ptr = scene_ptr->addPixmap(*mod_pixmap);
        else
            pixmap_item_ptr->setPixmap(*mod_pixmap);
    }

    // Reset the zoom factor.
    rezoom();

    // Pass it on.
    QDialog::resizeEvent(event);
}

/*!
 * @brief A timer has fired. Used for monitoring the progress of long operations.
 * @param event The actual timer event
 */
void CurveSettings::timerEvent(QTimerEvent *event)
{
    // Is this our file action timer?
    if (event->timerId() == poll_timer_id)
    {
        CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

        // Sanity check.
        if (my_parent == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
            return;
        }

        // Poll the file action status.
        my_parent->start_camera_read_reg(GIGE_REG_FILE_ACTION_STATUS);
    }
}

/*!
 * @brief The image has been modified or initialized this is the common logic to
 * regenerate and update the histogram graph.
 */
void CurveSettings::UpdateHistogram()
{
    // Set the ROI and build up the histogram proc data.
    if (histoProc && mod_pixmap && splinesInitialized)
    {
        // Set the mono state of this image.
        bool monoImage = mod_pixmap->depth() == 1 ? true : false;

        // Don't need to set the ROI as a 0 QRect uses entire image.
        //histoProc->set_roi(mod_pixmap->rect());
        // This doesn't seem to have any effect.
        //histoProc->set_mono(monoImage);
        histoProc->calc(*mod_pixmap);

        // Build up the histoData array(s).
        histoData.clear();
        histoData.append(*histoProc->get_red());
        if (!monoImage)
        {
            histoData.append(*histoProc->get_green());
            histoData.append(*histoProc->get_blue());
        }

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Updated histogram.  bpp=" << mod_pixmap->depth();
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to update.  histoProc[" << (histoProc ? "Allocated]" : "Nullptr]")
                     << "mod_pixmap[" << (mod_pixmap ? "Allocated]" : "Nullptr]");
    }
}

/*!
 * @brief The mod_pixmap has been modified and this is the logic to update the display image.
 */
void CurveSettings::UpdateDisplayImage()
{
    // Refresh the display.
    resizeEvent(nullptr);
}

/*!
 * @brief Determines if the curve settings support RGB spline curves.
 *
 * @returns True if this dialog is supporting all 3 channels or false if it is only
 * supporting the gamma (for series3) or monochrome (for seriesNC).
 */
bool CurveSettings::UsingRGBSplines()
{
    bool rgbSplines = true;

    // If the camera sensor that made the image we are modifying is a series3 camera
    // or if the camera sensor is a cameraNC but monochrome sensor then we are not
    // using the rgb splines we are only using one.
    if (isMonoSensor || isSeries3Camera)
    {
        rgbSplines = false;
    }
    return rgbSplines;
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user wants to upload the LUT.
 */
void CurveSettings::do_accept()
{
    QString str_text;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[[[[[[[Calling SaveSplineData(" << old_slot << ")]]]]]]]";
    }

    {
        allowLutProcessingMutex.lock();
        allowLutProcessing = false;
        allowLutProcessingMutex.unlock();
    }

    // Pulls current dialog settings and installs in lutSplineRecs for the record
    // that matches the working lut id settings (in the old_slot).
    SaveSplineData(old_slot);

    // Are we overriding a camera?
    if (my_parent != nullptr)
    {
        // Make sure the user is OK with this.
        str_text = tr("You are about to overwrite Lookup Table #%1 on the camera. Do you want to continue?").arg(GetCheckedRadioLutId());
        int user_response = QMessageBox::question(this, tr("LUT Curve Settings"), str_text, QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
        if (user_response == QMessageBox::No)
            return;     // Bail out.
    }
    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Calling save_lut() " << old_slot;
    }

    // Save the LUT information to our application registry.
    SaveToQSettings();

    // Pull the spline record index for the selected LUT.
    int splineRecIndex = GetLutSplineRecId();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Just called GetLutSplineRecId() splineRecIndex: " << splineRecIndex;

    // We want to save off an accepted Display LUT in case caller wants it.
    if (splineRecIndex != -1)
    {
        const DISPLAY_LUT::LUT_SPLINE_RECORD& activeSplineRec = lutSplineRecs[splineRecIndex];

        if (VERBOSE_LOGGING)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Accept id:" << activeSplineRec.id
                         << "index" << splineRecIndex
                         << "old_slot" << old_slot
                         << "lutDataType:" << activeSplineRec.lutDataType
                         << "desc:" << activeSplineRec.desc;
        }

        // Save off our accepted LUT in case the caller wants this data.
        acceptedLut = FiDisplayLut::CreateDisplayLut(activeSplineRec,
                                                     cameraLutRedOrGamma,
                                                     cameraLutGreen,
                                                     cameraLutBlue);
    }

    DisableControls();

    // All done.
    QDialog::accept();
}


/*!
 * @brief The user wants to exit dialog without any changes.
 */
void CurveSettings::do_cancel()
{
    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__;
    }
    {
        allowLutProcessingMutex.lock();
        allowLutProcessing = false;
        allowLutProcessingMutex.unlock();
    }

    DisableControls();

    // All done.
    QDialog::reject();
}

/*!
 * @brief The user has clicked on the "Load" button to load a LUT Spline Record Array from an XML file.
 */
void CurveSettings::do_load()
{
    uint8_t lutDataTypeLoaded = LUT_SERIES3;    // Set default.
    int lutSplineRecId;
    QString     filename;
    QString     last_filename;
    QSettings   settings;

    // Initialize the default path based on the saved settings.
    settings.beginGroup(LUT_SPLINE_SETTINGS_GROUP);
    last_filename = settings.value(LUT_SPLINE_SETTINGS_LASTFILE).toString();
    settings.endGroup();

    // Make sure we got something.
    if (last_filename.isEmpty())
        last_filename = QStandardPaths::writableLocation(QStandardPaths::DesktopLocation);
    else
        last_filename = QDir::toNativeSeparators(last_filename);

    // Have the user select a file.
    filename = QFileDialog::getOpenFileName(MainWindow::GetInstance().get(), tr("Select A LUT"),
                                            last_filename, tr("LUT File (*.xml)"));
    if (filename.isEmpty())
        return;

    // Save the folder the user selected.
    settings.beginGroup(LUT_SPLINE_SETTINGS_GROUP);
    settings.setValue(LUT_SPLINE_SETTINGS_LASTFILE, filename);
    settings.endGroup();

    // Make sure our listing file is closed.
    listing_file.close();

    // Use the new file.
    listing_file.setFileName(filename);

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Loading xml file:" << filename;
    }

    // Get the index of the selected LUT.
    lutSplineRecId = GetLutSplineRecId();
    if (lutSplineRecId < 0)
    {
        // We don't have a LUT_SPLINE_RECORD matching the selected destination LUT radio, so add an empty one.
        DISPLAY_LUT::LUT_SPLINE_RECORD    empty_lut;
        empty_lut.id = GetCheckedRadioLutId();  // Save the selected destination LUT radio id.
        empty_lut.lutDataType = LUT_SERIES3;    // Default it to a series3.
        lutSplineRecs.append(empty_lut);

        // Our "new" lut is the last one in the list save its list zero-based index.
        lutSplineRecId = lutSplineRecs.count() - 1;
    }

    // Part II - go through the XML data file and pull out the settings for the
    // LUT_SPLINE_RECORD.

    QString         msg_str;
    QDomDocument    doc(LUT_SPLINE_LIST_DOM_DOCUMENT);
    QDomElement     dom_root;           // xml token LUT_SPLINE_LIST_DOM_ELEMENT
    QDomElement     dom_child;          // xml token LUT_SPLINE_REC_ACTIVE_DOM_ELEMENT or LUT_SPLINE_REC_DOM_ELEMENT

    // Open our file so the XML parser can read it.
    if (!listing_file.open(QIODevice::ReadOnly))
    {
        msg_str = QStringLiteral("Error opening returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(listing_file.error()).arg(listing_file.errorString());
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        return;
    }

    // Set the XML file to the root document of the XML parser.
    if (!doc.setContent(&listing_file))
    {
        msg_str = QStringLiteral("Error parsing returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(errno).arg(QString::fromUtf8(strerror(errno)));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;

        listing_file.close();
        return;
    }

    // Don't need this anymore it's been read in.
    listing_file.close();

    // Get the root object.
    dom_root = doc.firstChildElement(LUT_SPLINE_LIST_DOM_ELEMENT);

    // Walk through the lut information.
    dom_child = dom_root.firstChildElement(LUT_SPLINE_REC_DOM_ELEMENT);
    if (dom_child.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No lut_info entry!?";
        return;
    }

    QDomElement dom_knots;                  // xml token "*knots" reused for green and blue if !useMono.
    QDomElement knot_info;                  // xml token "*knot" reused for green and blue if !useMono.
    bool hasGreenAndBlue = false;           // Whether or not we expect green and blue knots.

    // Get the description information.
    // NOTE:  This had been using dom_child.attribute but it doesn't work right.
    lutSplineRecs[lutSplineRecId].desc = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_DESC).text();
    lutSplineRecs[lutSplineRecId].localPath = filename;

    // Look for the lut data type to determine number of channels.
    QString lutDataTypeString = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_LUT_DATA_TYPE).text();
    lutDataTypeLoaded = (uint8_t)lutDataTypeString.toInt();

    // Is the loaded file expected to contain blue and green knots?
    if (lutDataTypeLoaded == LUT_SERIESNC_COLOR)
    {
        hasGreenAndBlue = true;
    }

    // Set up the lut data type based on the setting for the dialog.
    // If using RGB splines we can assume series NC color.
    if (UsingRGBSplines())
        lutSplineRecs[lutSplineRecId].lutDataType = LUT_SERIESNC_COLOR;
    else
        lutSplineRecs[lutSplineRecId].lutDataType = LUT_SERIES3;

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lutSplineRec.desc:" << lutSplineRecs[lutSplineRecId].desc
                     << "lutDataTypeLoaded:" << lutDataTypeLoaded
                     << "id:" << lutSplineRecs[lutSplineRecId].id
                     << "lutDataType:" << lutSplineRecs[lutSplineRecId].lutDataType
                     << "expects Green and Blue:" << hasGreenAndBlue;
    }

    // Reset all LUTS' knot information.
    lutSplineRecs[lutSplineRecId].redOrGammaKnots.clear();
    lutSplineRecs[lutSplineRecId].greenKnots.clear();
    lutSplineRecs[lutSplineRecId].blueKnots.clear();

    // Now load the knot info.
    // We always expect to find the single spline data in the xml data.
    dom_knots = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOTS);
    knot_info = dom_knots.firstChildElement(LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOT);
    ParseDOMKnots(knot_info, lutSplineRecs[lutSplineRecId].redOrGammaKnots, LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOT);

    // If the curve dialog has been configured to expect 3 channels of color some housekeeping is needed.
    if (lutSplineRecs[lutSplineRecId].lutDataType == LUT_SERIESNC_COLOR)
    {
        // We may or may not have green and blue data saved.
        if (!hasGreenAndBlue)
        {
            // Mark the red channel twice.
            lutSplineRecs[lutSplineRecId].greenKnots = lutSplineRecs[lutSplineRecId].redOrGammaKnots;
            lutSplineRecs[lutSplineRecId].blueKnots = lutSplineRecs[lutSplineRecId].redOrGammaKnots;
        }
        // We expect there to be green and blue knots.
        else
        {
            // The green spline data should be in this record.
            dom_knots = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_GREEN_KNOTS);
            knot_info = dom_knots.firstChildElement(LUT_SPLINE_REC_DOM_GREEN_KNOT);
            ParseDOMKnots(knot_info, lutSplineRecs[lutSplineRecId].greenKnots, LUT_SPLINE_REC_DOM_GREEN_KNOT);

            // And the blue spline data should also be in this record.
            dom_knots = dom_child.firstChildElement(LUT_SPLINE_REC_DOM_BLUE_KNOTS);
            knot_info = dom_knots.firstChildElement(LUT_SPLINE_REC_DOM_BLUE_KNOT);
            ParseDOMKnots(knot_info, lutSplineRecs[lutSplineRecId].blueKnots, LUT_SPLINE_REC_DOM_BLUE_KNOT);
        }
    }

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Leaving and knotsEqual:" << FiDisplayLut::KnotsAreEqual(lutSplineRecs[lutSplineRecId]);
    }

    // Now that everything is setup, make sure we're the active LUT.
    SetupDestinationLut(GetCheckedRadioLutId());
}

/*!
 * @brief Reset the currently selected curve to a linear setting and erases all
 * entries in lutSplineRecs.
 */
void CurveSettings::do_reset()
{
    int i;
    int lut_id;

    // Get the ID.
    lut_id = GetCheckedRadioLutId();

    // Find out which LUT we're looking at.
    for (i = 0; i < lutSplineRecs.count(); i++)
    {
        // Is this the one we're looking for?
        if (lutSplineRecs.at(i).id == lut_id)
        {
            // Delete this entry.
            lutSplineRecs.removeAt(i);

            // Does this match our playback_ptr?
            if (playbackLut)
            {
                if (lut_id == playbackLut->GetLutId())
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing playbackLut:" << lut_id;
                    // Set our playbackLut null now it's been reset.
                    playbackLut = nullptr;
                }
            }

            // All done.
            break;
        }
    }

    // Clear the QSettings for our LUT tables.
    QSettings settings;
    settings.remove(LUT_SPLINE_SETTINGS_GROUP);

    // Initialize the dialog for a default setup.
    InitializeDefaultSetup();
}

/*!
 * @brief The user clicked on the "Save" button to save the spline curves currently
 *        in the dialog.
 */
void CurveSettings::do_save()
{
    int         lut_index;
    int         knot_index;
    QString     filename;
    QString     last_filename;
    QSettings   settings;
    QXmlStreamWriter    writer;

    // Initialize the default path based on the saved settings.
    settings.beginGroup(LUT_SPLINE_SETTINGS_GROUP);
    last_filename = settings.value(LUT_SPLINE_SETTINGS_LASTFILE).toString();
    settings.endGroup();

    // Make sure we got something.
    if (last_filename.isEmpty())
        last_filename = QStandardPaths::writableLocation(QStandardPaths::DesktopLocation);
    else
        last_filename = QDir::toNativeSeparators(last_filename);

    // Ask the user for a file name.
    filename = QFileDialog::getSaveFileName(this, tr("Select a Directory and Filename to save the LUT File"),
                                            last_filename, tr("LUT File (*.xml)"));
    if (filename.isEmpty())
        return;

    // Save the folder the user selected.
    settings.beginGroup(LUT_SPLINE_SETTINGS_GROUP);
    settings.setValue(LUT_SPLINE_SETTINGS_LASTFILE, filename);
    settings.endGroup();

    // Make sure our listing file is closed.
    listing_file.close();

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Saving curve settings to filename:" << filename;
    }

    // Open our XML file.
    listing_file.setFileName(filename);
    if (!listing_file.open(QFile::WriteOnly | QFile::Text))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to open the XML file %1, err=%2").arg(filename, listing_file.errorString());
        return;
    }
    listing_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // Save the current curve.
    SaveSplineData(old_slot);

    // Setup our XML writer to work with our file.
    writer.setDevice(&listing_file);
    writer.setAutoFormatting(true);
    writer.writeStartDocument();

    // We are only saving the active LUT.
    writer.writeStartElement(LUT_SPLINE_LIST_DOM_ELEMENT);
    writer.writeAttribute(LUT_SPLINE_LIST_COUNT_DOM_ELEMENT, QString::number(1));

    // Save the active (selected) ID.
    writer.writeStartElement(LUT_SPLINE_REC_ACTIVE_DOM_ELEMENT);
    writer.writeAttribute(LUT_SPLINE_REC_DOM_ID, QString::number(GetCheckedRadioLutId()));
    writer.writeEndElement();

    // Start saving the LUTs.
    for (lut_index = 0; lut_index < lutSplineRecs.count(); lut_index++)
    {
        // Is this the one we're interested in?
        if (lutSplineRecs[lut_index].id != GetCheckedRadioLutId())
            continue;

        // Found it, so write it out.
        writer.writeStartElement(LUT_SPLINE_REC_DOM_ELEMENT);
        writer.writeAttribute(LUT_SPLINE_REC_DOM_ID, QString::number(lutSplineRecs[lut_index].id));
        writer.writeTextElement(LUT_SPLINE_REC_DOM_DESC, lutSplineRecs[lut_index].desc);

        // Save the LUT data type and make sure everything matches the state of this dialog.
        if (isSeries3Camera)
        {
            lutSplineRecs[lut_index].lutDataType = LUT_SERIES3;
            writer.writeTextElement(LUT_SPLINE_REC_DOM_LUT_DATA_TYPE, (QString("%1").arg(LUT_SERIES3)));
        }
        else if (UsingRGBSplines())
        {
            lutSplineRecs[lut_index].lutDataType = LUT_SERIESNC_COLOR;
            writer.writeTextElement(LUT_SPLINE_REC_DOM_LUT_DATA_TYPE, (QString("%1").arg(LUT_SERIESNC_COLOR)));
        }
        else
        {
            // Last but least it must be a seriesNC monochrome sensor.
            lutSplineRecs[lut_index].lutDataType = LUT_SERIESNC_MONO;
            writer.writeTextElement(LUT_SPLINE_REC_DOM_LUT_DATA_TYPE, (QString("%1").arg(LUT_SERIESNC_MONO)));
        }

        // Save the red/gamma knots.
        writer.writeStartElement(LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOTS);
        for (knot_index = 0; knot_index < lutSplineRecs[lut_index].redOrGammaKnots.count(); knot_index++)
        {
            writer.writeTextElement(LUT_SPLINE_REC_DOM_RED_OR_GAMMA_KNOT, QString("%1:%2").arg(lutSplineRecs[lut_index].redOrGammaKnots[knot_index].x()).arg(lutSplineRecs[lut_index].redOrGammaKnots[knot_index].y()));
        }
        writer.writeEndElement();   // Terminate red/gamma knots

        // If there are green and blue knots write them out.
        if (lutSplineRecs[lut_index].lutDataType == LUT_SERIESNC_COLOR)
        {
            // Save the green knots.
            writer.writeStartElement(LUT_SPLINE_REC_DOM_GREEN_KNOTS);
            for (knot_index = 0; knot_index < lutSplineRecs[lut_index].greenKnots.count(); knot_index++)
            {
                writer.writeTextElement(LUT_SPLINE_REC_DOM_GREEN_KNOT, QString("%1:%2").arg(lutSplineRecs[lut_index].greenKnots[knot_index].x()).arg(lutSplineRecs[lut_index].greenKnots[knot_index].y()));
            }
            writer.writeEndElement();   // Terminate green knots

            // Save the blue knots.
            writer.writeStartElement(LUT_SPLINE_REC_DOM_BLUE_KNOTS);
            for (knot_index = 0; knot_index < lutSplineRecs[lut_index].blueKnots.count(); knot_index++)
            {
                writer.writeTextElement(LUT_SPLINE_REC_DOM_BLUE_KNOT, QString("%1:%2").arg(lutSplineRecs[lut_index].blueKnots[knot_index].x()).arg(lutSplineRecs[lut_index].blueKnots[knot_index].y()));
            }
            writer.writeEndElement();   // Terminate green knots
        }

        // Terminate the LUT_SPLINE_REC_DOM_ELEMENT "lut_info" element.
        writer.writeEndElement();

        // We can break out of the loop now.
        break;
    }

    // All done.
    writer.writeEndDocument();
    listing_file.close();

    // Make sure everything is OK.
    if (listing_file.error() != QFile::NoError)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Error writing XML file %1, err=%2").arg(filename, listing_file.errorString());
    }
}

/*!
 * @brief The worker thread that translates a spline curve data into a
 * camera lut has completed for one of the channels.  Mark the channel
 * as being updated and perform any housekeeping.
 *
 * @param tabIndexOfCurve - which camera lut was just built?
 */
void CurveSettings::HandleBuildACameraLutCompleted(int tabIndexOfCurve)
{
    // Now set the data that was generated into the histogram plotter.
    switch (tabIndexOfCurve)
    {
    case SPLINE_TAB_INDEX_RED_OR_MONO:
        cameraLutLastModified = SPLINE_TAB_INDEX_RED_OR_MONO;
        builtRedOrGammaCameraLut = true;
        refreshRedOrGammaCurve = false; // Up to date.
        break;
    case SPLINE_TAB_INDEX_GREEN:
        cameraLutLastModified = SPLINE_TAB_INDEX_GREEN;
        builtGreenCameraLut = true;
        refreshGreenCurve = false; // Up to date.
        break;
    case SPLINE_TAB_INDEX_BLUE:
        cameraLutLastModified = SPLINE_TAB_INDEX_BLUE;
        builtBlueCameraLut = true;
        refreshBlueCurve = false; // Up to date.
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to build camera lut.";
        break;
    }

#if 0
    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__
                     << "tab:" << tabIndexOfCurve
                     << "builtRed:" << builtRedOrGammaCameraLut
                     << "builtGreen:" << builtGreenCameraLut
                     << "builtBlue:" << builtBlueCameraLut
                     << "splinesInitialized:" << splinesInitialized;
    }
#endif

    // Are we waiting on the camera luts to build in order to finish the
    // dialog initialization?
    if (!splinesInitialized)
    {
        if (UsingRGBSplines())
        {
            if (builtRedOrGammaCameraLut && builtGreenCameraLut && builtBlueCameraLut)
            {
                // Can complete the curve dialog initialization.
                ColorCurvesReadyForEditing();
            }
        }
        else
        {
            if (builtRedOrGammaCameraLut)
            {
                // Can complete the curve dialog initialization.
                ColorCurvesReadyForEditing();
            }
        }
    }

    // Test again and if the splines are initialized at this point we are ready
    // to get the display luts, the histogram, and the image updated.
    if (splinesInitialized)
    {
        // Update the display...
        QTimer::singleShot(0, this, &CurveSettings::HandleSplineChangedTimer);
    }
}

/*!
 * @brief Finish up the LUT processing by updating the display with
 *        the new values.
 */
void CurveSettings::HandleLutProcessingComplete()
{
    // Now set the data that was generated into the histogram plotter.
    // All done reading the histogram data.
    //ui->histogramPlotCtl->set_mono(monoImage);
    ui->histogramPlotCtl->populate_histogram(histoData);
    UpdateDisplayImage();

    // The spline change UI timer was serviced.
    splineChangedTimerStarted = false;

    // Are there pending changes? If there are, go back around.
    if (splineChangedCount)
        QTimer::singleShot(0, this, &CurveSettings::HandleSplineChangedTimer);
}

/*!
 * @brief The user has selected a lut spline curve for modification/view.  Handle
 * the logistics of changing spline controls.
 */
void CurveSettings::HandleLutTabClick(int tabIndex)
{
    switch (tabIndex)
    {
    case SPLINE_TAB_INDEX_RED_OR_MONO:
        activeLutTabIndex = tabIndex;
        // Does the red or mono spline widget need to be updated?
        if (refreshRedOrGammaCurve)
        {
            // Knots are out of sync with the graph.  Reset the spline data.
            ui->fi_spline->set_spline(aaSplineRedOrGamma);
            refreshRedOrGammaCurve = false;
        }
        break;
    case SPLINE_TAB_INDEX_GREEN:
        activeLutTabIndex = tabIndex;
        // Does the green spline widget need to be refreshed?
        if (refreshGreenCurve)
        {
            // Knots are out of sync with the graph.  Reset the spline data.
            ui->fiSplineGreen->set_spline(aaSplineGreen);
            refreshGreenCurve = false;
        }
        break;
    case SPLINE_TAB_INDEX_BLUE:
        activeLutTabIndex = tabIndex;
        // Does the blue spline widget need to be refreshed?
        if (refreshBlueCurve)
        {
            // Knots are out of sync with the graph.  Reset the spline data.
            ui->fiSplineBlue->set_spline(aaSplineBlue);
            refreshBlueCurve = false;
        }
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Tab Index provided was out of range:" << tabIndex;
        break;
    }

    // Make sure any lock change is caught.
    if (splineIsLocked != ui->tb_curve_lock->isChecked())
    {
        // Are we locked?
        splineIsLocked = ui->tb_curve_lock->isChecked();
        if (splineIsLocked)
        {
            // Fire off a timer to start processing spline changes.
            QTimer::singleShot(0, this, &CurveSettings::HandleSplineChangedTimer);
        }
    }
}

/*!
 * @brief A spline curve has been modified and this is a slot handler to respond to the
 * splineChangedTimer QTimer::timeout signal.  It was created to reduce the number of times
 * the histogram and the display image are updated in response to user mouse moves in the
 * spline curve control widget.
 */
void CurveSettings::HandleSplineChangedTimer()
{
    // Check for a double shot of the timer as we want to finish processing
    // the current one if so.
    if (splineChangedTimerStarted)
    {
        // Not setting another timer so just increment the changed count.
        splineChangedCount++;
        return;
    }

    // Save the fact that we are processing.
    splineChangedCount = 0;
    splineChangedTimerStarted = true;

    // Are we an RGB setup and are we locked?
    if (UsingRGBSplines())
    {
        if (IsCurveLocked())
        {
            // By syncing up the curves for a lock scenario we want to disconnect any modification
            // signals that will come through when syncing the knot data.
            disconnect(ui->fi_spline, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::RedSplineChanged);
            disconnect(ui->fiSplineGreen, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::GreenSplineChanged);
            disconnect(ui->fiSplineBlue, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::BlueSplineChanged);

            // Sync the non-active spline curves with the active spline curve data and camera lut.
            if (cameraLutLastModified == SPLINE_TAB_INDEX_RED_OR_MONO)
            {
                // Sync the green curve data and mark it in need of refresh to sync the fiSplineWidget.
                aaSplineGreen->knots = aaSplineRedOrGamma->knots;
                refreshGreenCurve = true;
                cameraLutGreen = cameraLutRedOrGamma;

                // Sync the blue curve data and mark it in need of refresh to sync the fiSplineWidget.
                aaSplineBlue->knots = aaSplineRedOrGamma->knots;
                refreshBlueCurve = true;
                cameraLutBlue = cameraLutRedOrGamma;
            }
            else if (cameraLutLastModified == SPLINE_TAB_INDEX_GREEN)
            {
                // Sync the red curve data and mark it in need of refresh to sync the fiSplineWidget.
                aaSplineRedOrGamma->knots = aaSplineGreen->knots;
                refreshRedOrGammaCurve = true;
                cameraLutRedOrGamma = cameraLutGreen;

                // Sync the blue curve data and mark it in need of refresh to sync the fiSplineWidget.
                aaSplineBlue->knots = aaSplineGreen->knots;
                refreshBlueCurve = true;
                cameraLutBlue = cameraLutGreen;
            }
            else if (cameraLutLastModified == SPLINE_TAB_INDEX_BLUE)
            {
                // Sync the red curve data and mark it in need of refresh to sync the fiSplineWidget.
                aaSplineRedOrGamma->knots = aaSplineBlue->knots;
                refreshRedOrGammaCurve = true;
                cameraLutRedOrGamma = cameraLutBlue;

                // Sync the green curve data and mark it in need of refresh to sync the fiSplineWidget.
                aaSplineGreen->knots = aaSplineBlue->knots;
                refreshGreenCurve = true;
                cameraLutGreen = cameraLutBlue;
            }

            // Make sure to reconnect up the modification signals.
            connect(ui->fi_spline, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::RedSplineChanged, Qt::UniqueConnection);
            connect(ui->fiSplineGreen, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::GreenSplineChanged, Qt::UniqueConnection);
            connect(ui->fiSplineBlue, &fiSplineWidget::selectValuesChanged, this, &CurveSettings::BlueSplineChanged, Qt::UniqueConnection);
        }
    }

    // Fire off the worker thread.
    QFuture future = QtConcurrent::run(LutProcessing, this);
}


/*!
 * @brief Show a busy cursor if the dialog is being executed.
 */
int CurveSettings::exec()
{
    // Show a busy cursor while the spline controls are being created.
    if (!busyCursorShowing)
    {
        QApplication::setOverrideCursor(Qt::BusyCursor);
        busyCursorShowing = true;
    }
    {
        allowLutProcessingMutex.lock();
        allowLutProcessing = true;
        allowLutProcessingMutex.unlock();
    }
    return (QDialog::exec());
}

/*!
 * @brief The curve dialog has been initialized but we still need to calculate all of the
 * initial LUT curves.  This can take some time so a one-shot timer went off to handle this
 * operation.
 */
void CurveSettings::HandleSplineInitTimer()
{
    // Make sure this isn't a bogus call.
    if (splinesInitialized)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Only one HandleSplineInitTimer event per dialog session - splinesInitialized was already true!";
        return;
    }

    // Initialize for a default dialog setup.
    InitializeDefaultSetup();

    // Are we connected to a camera?
    if (qobject_cast<CameraMDI *>(myParent) != nullptr)
    {
        // Start the process of reading the listing of lookup table(s) from the camera.
        listing_start();
    }
    else
    {
        // Restore our Spline LUT records from QSettings and perform any housekeeping.
        RestoreFromQSettings();
    }

    // If we already have a playback setup nothing to do here.
    if (!playbackLut)
    {
        // If we have a LUT selected, let's run the image through it.
        if (old_slot > 0)
            LoadLutSplineRec(old_slot);
    }
}

/*!
 * @brief Handles a destination LUT radio selection managing data persistence
 * for the lutSplineRecs vector.
 */
void CurveSettings::HandleLutRadioClick()
{
    int slot = GetCheckedRadioLutId();

    // Save the current information.
    SaveSplineData(old_slot);

    // Update our curve.
    LoadLutSplineRec(slot);

    // Update our slot ID.
    old_slot = slot;
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void CurveSettings::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_FILE_ACTION_STATUS:
        // Read the value we got back.
        reg_value = my_parent->read_camera_cache(GIGE_REG_FILE_ACTION_STATUS);

        handle_fa_status(reg_value);
        break;
    case GIGE_REG_FILEPATH:
        handle_listing_download();
        break;
    case GIGE_REG_LUT:
        // Read the value we got back.
        reg_value = my_parent->read_camera_cache(GIGE_REG_LUT);

        // Deal with it.
        handle_lut_change(reg_value);
        break;
    default:
        break;
    }
}

void CurveSettings::rezoom()
{
    // We'll start out by resetting the transform values.
    // This has the side effect of getting us to 1:1.
    ui->image_view->resetTransform();

    // Sanity check.
    if (pixmap_item_ptr != nullptr)
    {
        // Scale the image into our viewport.
        ui->image_view->fitInView(pixmap_item_ptr, Qt::KeepAspectRatio);
    }
}

/*!
 * @brief The user is modifying the blue curve. Process the change.
 *
 * @param x     [UNUSED]
 * @param y     [UNUSED]
 */
void CurveSettings::BlueSplineChanged(float x, float y)
{
    // We actually don't care what the new values are, we're just
    // interested that there was a change.
    Q_UNUSED(x)
    Q_UNUSED(y)

    QFuture future = QtConcurrent::run(BuildACameraLut, this, SPLINE_TAB_INDEX_BLUE);

    // Ferry the call down to process the change in the blue spline control.
    //ProcessSplineChange(ui->fiSplineBlue, aaSplineBlue, cameraLutBlue);
}

/*!
 * @brief The user is modifying the monochrome/gamma curve. Process the change.
 *
 * @param x     [UNUSED]
 * @param y     [UNUSED]
 */
void CurveSettings::GammaSplineChanged(float x, float y)
{
    // We actually don't care what the new values are, we're just
    // interested that there was a change.
    Q_UNUSED(x)
    Q_UNUSED(y)

    QFuture future = QtConcurrent::run(BuildACameraLut, this, SPLINE_TAB_INDEX_RED_OR_MONO);
    // Ferry the call down to process the change in the gamma spline control.
    //ProcessSplineChange(ui->fi_spline, aaSplineRedOrGamma, cameraLutRedOrGamma);
}

/*!
 * @brief The user is modifying the green curve. Process the change.
 *
 * @param x     [UNUSED]
 * @param y     [UNUSED]
 */
void CurveSettings::GreenSplineChanged(float x, float y)
{
    // We actually don't care what the new values are, we're just
    // interested that there was a change.
    Q_UNUSED(x)
    Q_UNUSED(y)

    // Ferry the call down to process the change in the green spline control.
    QFuture future = QtConcurrent::run(BuildACameraLut, this, SPLINE_TAB_INDEX_GREEN);
    //ProcessSplineChange(ui->fiSplineGreen, aaSplineGreen, cameraLutGreen);
}

/*!
 * @brief The user is modifying the red curve. Process the change.
 *
 * @param x     [UNUSED]
 * @param y     [UNUSED]
 */
void CurveSettings::RedSplineChanged(float x, float y)
{
    // We actually don't care what the new values are, we're just
    // interested that there was a change.
    Q_UNUSED(x)
    Q_UNUSED(y)

    // Ferry the call down to process the change in the red spline control.
    QFuture future = QtConcurrent::run(BuildACameraLut, this, SPLINE_TAB_INDEX_RED_OR_MONO);
    //ProcessSplineChange(ui->fi_spline, aaSplineRedOrGamma, cameraLutRedOrGamma);
}

/*!
 * @brief The spline curve type has changed from linear to cubic or cubic to linear.
 *
 * @param newCurveType - it should be aaAaa::aaSpline::SPLINE_CUBIC or aaAaa::aaSpline::SPLINE_LINEAR
 */
void CurveSettings::GammaSplineTypeChanged(int newCurveType)
{
    aaSplineRedOrGamma->type = newCurveType;
    LoadLutSplineRec(GetCheckedRadioLutId());
}

/*!
 * @brief The spline curve type has changed from linear to cubic or cubic to linear.
 *
 * @param newCurveType - it should be aaAaa::aaSpline::SPLINE_CUBIC or aaAaa::aaSpline::SPLINE_LINEAR
 */
void CurveSettings::RedSplineTypeChanged(int newCurveType)
{
    aaSplineRedOrGamma->type = newCurveType;
    LoadLutSplineRec(GetCheckedRadioLutId());
}

/*!
 * @brief The spline curve type has changed from linear to cubic or cubic to linear.
 *
 * @param newCurveType - it should be aaAaa::aaSpline::SPLINE_CUBIC or aaAaa::aaSpline::SPLINE_LINEAR
 */
void CurveSettings::GreenSplineTypeChanged(int newCurveType)
{
    aaSplineGreen->type = newCurveType;
    LoadLutSplineRec(GetCheckedRadioLutId());
}

/*!
 * @brief The spline curve type has changed from linear to cubic or cubic to linear.
 *
 * @param newCurveType - it should be aaAaa::aaSpline::SPLINE_CUBIC or aaAaa::aaSpline::SPLINE_LINEAR
 */
void CurveSettings::BlueSplineTypeChanged(int newCurveType)
{
    aaSplineBlue->type = newCurveType;
    LoadLutSplineRec(GetCheckedRadioLutId());
}

#if 0

/*!
 * @brief The user is playing with the curve. Update the look up table matching the spline
 * curve widget provided.  This builds up the cameraLutForSpline vector.  If the user wants
 * the display updated then the setTimerForUpdates will be true to spawn off a single shot
 * update timer.
 *
 * @param fiSplineToChange - identifies which spline widget for the curve being modified.
 * @param aaSplineCurve - identifies which curve control being modified.
 * @param cameraLutForSpline - will be set to the camera lut data matching the spline curve.
 * @param setTimerForUpdates - does caller want display updates to happen.
 */
void CurveSettings::ProcessSplineChange(fiSplineWidget *fiSplineToChange,
                                        std::shared_ptr<aaAaa::aaSpline> aaSplineCurve,
                                        QVector<double>& cameraLutForSpline,
                                        bool setTimerForUpdates)
{
    // Make sure to camera lut provided is the right size.
    if (cameraLutForSpline.size() != LUT_MAX_TABLE_COUNT)
        cameraLutForSpline.resize(LUT_MAX_TABLE_COUNT);

    // Make sure the spline widget provided is a valid pointer.
    if (fiSplineToChange == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fiSplineToChange was a nullptr!";
        return;
    }

    // Make sure the spline curve provided is a valid pointer.
    if (aaSplineCurve == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "aaSplineCurve was a nullptr!";
        return;
    }

    // Read the LUT information from the spline widget using the spline curve control attribute.
    if (fiSplineToChange->get_data((int)(aaSplineCurve->limit_right + 1) / LUT_MAX_TABLE_COUNT, cameraLutForSpline) < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to read the LUT from " << fiSplineToChange->objectName() << " control!";
        return;
    }

    // Which camera lut was built?
    // Look at the spline widget pointer and set an integer matching tab index.
    // Reset any refresh ui indicator for the spline widget as it should be fully up to date.
    if (fiSplineToChange == ui->fi_spline)
    {
        cameraLutLastModified = SPLINE_TAB_INDEX_RED_OR_MONO;
        refreshRedOrGammaCurve = false; // curve widget is up to date.
    }
    else if (fiSplineToChange == ui->fiSplineGreen)
    {
        cameraLutLastModified = SPLINE_TAB_INDEX_GREEN;
        refreshGreenCurve = false;      // curve widget is up to date.
    }
    else if (fiSplineToChange == ui->fiSplineBlue)
    {
        cameraLutLastModified = SPLINE_TAB_INDEX_BLUE;
        refreshBlueCurve = false;      // curve widget is up to date.
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't match fiSplineToChange pointer with ui spline widgets!";
    }

    // Does the caller want us to fire off a single shot timer to display any updates?
    if (setTimerForUpdates)
    {
        // Make sure we are a single shot timer and send a single shot timer event.
        // We are using the timer event as a means to separate the processing of the
        // data from the GUI. Start the processing as soon as we can get around to it.
        QTimer::singleShot(0, this, &CurveSettings::HandleSplineChangedTimer);
    }
}

#endif


/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void CurveSettings::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;

    // Cast our parent to a form we can use.
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    if (my_parent == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No parent?";
        return;
    }

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Get our communications object pointer.
    FI_comm     *comm_ptr;
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_FILE_ACTION:
        // Read the status of the write operation.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Was this a success?
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Start our timer to poll for the completion of the file action.
            if (poll_timer_id == 0)
                poll_timer_id = startTimer(FA_POLLING_PERIOD);
        }
        break;
    default:
        break;
    }
}
