/*!
 * @file curve_settings.h
 * @brief This is the header file for the CurveSettings class.
 *
 * Copyright 2016 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef CURVE_SETTINGS_H
#define CURVE_SETTINGS_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QDialog>
#include <QFile>
#include <stdint.h>

#include "fi_pixmap.h"
#include "fi_splinewidget.h"
#include "fi_histogram.h"
#include "fi_display_lut.h"

namespace Ui
{
class CurveSettings;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class QDomElement;
class QGraphicsScene;
class QGraphicsPixmapItem;
class fiHistoProc;

//--------------------------------------------------------------------
// Namespace: CURVE_SETTINGS
//--------------------------------------------------------------------
namespace CURVE_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

}   // End namespace (CURVE_SETTINGS)

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class CurveSettings : public QDialog
{
    Q_OBJECT

protected:
    explicit CurveSettings(QWidget *parent = nullptr);

public:
    ~CurveSettings();

    static std::unique_ptr<CurveSettings>
    CreateCurveDialog(bool isMonoSensor,
                      bool isSeries3Camera,
                      fiPixmap& displayImageSnap,
                      int InitialLutId,
                      QWidget *parent = nullptr);

    //! Returns an FiDisplayLut matching data associated with accepted curve settings.
    std::shared_ptr<FiDisplayLut> GetAcceptedFiDisplayLut() {return acceptedLut;}

    QByteArray  *GetCameraLut();
    int         GetCheckedRadioLutId();
    const DISPLAY_LUT::LUT_SPLINE_RECORD& GetRadioLutSplineRec();

signals:
    void    BuildACameraLutCompleted(int tabIndex);
    void    LutProcessingComplete();
    void    CurvesAreReadyForEditing();

protected:
    virtual void    resizeEvent(QResizeEvent *event);
    virtual void    set_image(QPixmap& snap_image);
    virtual void    timerEvent(QTimerEvent *event);

public slots:
    virtual int exec();

protected slots:
    virtual void    changeEvent(QEvent *event);

private slots:
    void    do_accept();
    void    do_cancel();
    void    do_load();
    void    do_reset();
    void    do_save();

    void    HandleBuildACameraLutCompleted(int tabIndex);
    void    HandleLutProcessingComplete();
    void    HandleLutRadioClick();
    void    HandleLutTabClick(int tabIndex);
    void    HandleSplineInitTimer();
    void    HandleSplineChangedTimer();

    void    GammaSplineChanged(float x, float y);
    void    RedSplineChanged(float x, float y);
    void    GreenSplineChanged(float x, float y);
    void    BlueSplineChanged(float x, float y);

    void    GammaSplineTypeChanged(int new_type);
    void    RedSplineTypeChanged(int new_type);
    void    GreenSplineTypeChanged(int new_type);
    void    BlueSplineTypeChanged(int new_type);

    void    rezoom();
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private:
    static void LutProcessing(QObject *parent);
    static void BuildACameraLut(QObject *parent, int curveTabIndex);

    void    BuildDisplayLuts();

    void    ColorCurvesReadyForEditing();

    void    DisableControls();

    QByteArray  *GetCameraGammaLut();
    QByteArray  *GetCameraRGBLut();

    int     GetLutSplineRecId();
    void    handle_fa_status(uint32_t fa_status);
    void    handle_listing_download();
    void    handle_lut_change(uint32_t reg_value);
    void    InitializeControls();
    void    InitializeDefaultSetup();
    void    InitializeSpline(int tabIndex);

    bool    IsCurveLocked();

    void    listing_start();
    void    LoadLutSplineRec(int lut_id);
    void    ParseDOMKnots(QDomElement& DOMKnots, QVector<QPoint>& knotArray, const QString& knotDOMString);
    void    parse_listing();
    void    RestoreFromQSettings();
    void    SaveToQSettings();
    void    SaveSplineData(int lut_id);
    void    SetupDestinationLut(int lut_id, bool loadLutRecord = true);
    void    set_gamma_lut(QVector<double> lut);
    void    UpdateHistogram();
    void    UpdateDisplayImage();
    bool    UsingRGBSplines();

private:
    Ui::CurveSettings   *ui;

    QWidget *myParent;                                  //!< The parent QWidget* provided when the dialog was created.

    bool                allowLutProcessing;             //!< When is lut processing allowed?
    QMutex              allowLutProcessingMutex;        //! < Mitigates access to allowLutProcessing setting.

    QVector<uint8_t>    displayLutRed;                  //!< The 8bpp red channel LUT used to create mod_pixmap.
    QVector<uint8_t>    displayLutGreen;                //!< The 8bpp green channel LUT used to create mod_pixmap.
    QVector<uint8_t>    displayLutBlue;                 //!< The 8bpp blue channel LUT used to create mod_pixmap.

    int                 activeLutTabIndex;              //!< Which LUT-Tab index is active.
    bool                err_received;                   //!< Flag to remember if we get an error.

    bool                builtRedOrGammaCameraLut;       //!< Set true when cameraLutRedOrGamma is built.
    bool                builtGreenCameraLut;            //!< Set true when cameraLutGreen is built.
    bool                builtBlueCameraLut;             //!< Set true when cameraLutBlue is built.

    bool                busyCursorShowing;              //!< Set true if the dialog was executed.

    QVector<double>     cameraLutRedOrGamma;            //!< Our camera gamma lookup table for series3 images and seriesNC monochrome images.
    QVector<double>     cameraLutGreen;                 //!< Our camera green lookup table.
    QVector<double>     cameraLutBlue;                  //!< Our camera blue lookup table.
    QMutex cameraLutRedOrGammaMutex;                    //!< Mutex to guard access to cameraLutRedOrGamma.
    QMutex cameraLutGreenMutex;                         //!< Mutex to guard access to cameraLutGreen.
    QMutex cameraLutBlueMutex;                          //!< Mutex to guard access to cameraLutBlue.

    int                 cameraLutLastModified;          //!< Which camera lut was last modified.

    std::unique_ptr<fiHistoProc> histoProc;             //!< Histogram processing object.
    QVector<uint32_t>   histoData;                      //!< The histogram data created by histoProc.

    bool                initialLockSetting;             //!< Determines how to set up initial lock.

    bool                isMonoSensor;                   //!< Are we working with mono-sensor image data?
    bool                isSeries3Camera;                //!< Are we working with series3 camera?

    QFile               listing_file;                   //!< File name for the XML directory listing from the camera.
    uint32_t            listing_len;                    //!< The number of bytes in the listing from the camera.

    bool                okToUpdateDisplayImage;         //!< When is it ok to actually update the display image?

    int                 old_slot;                       //!< The last destination LUT ID used.
    int                 poll_timer_id;                  //!< Timer used when to poll the file operation.

    std::shared_ptr<FiDisplayLut> playbackLut;          //! Playback pointer custom lut if one exists.
    std::shared_ptr<FiDisplayLut> acceptedLut;          //! A copy of the curve settings lut if accepted.

    QMutex lutSplinexMutex;                             //!< Mutex to guard access to lutSplineRecs.
    QVector<DISPLAY_LUT::LUT_SPLINE_RECORD> lutSplineRecs;  //!< An array of LUT_SPLINE_RECORDs collected.

    // Note:  These refresh indicators get set true when have RGB curves and tabs are
    // locked to be in sync.  Refreshing all 3 fiSplineWidgets for each move is expensive
    // in terms of user experience because performance hits take place as user moves
    // a locked curve.  Instead we set a condition that the widget needs to be refreshed
    // and only when a new tab is clicked and only then will the spline widget replot and
    // regenerate the spline.  The cameraLut* vectors and the aaSpline* knots are kept in
    // sync each time a curve modification is handled in the HandleSplineChangedTimer slot
    // so that the display image and spline knots are sync'd but the plotting for the fiSplineWidget
    // control is postponed until that graph is actually activated by a tab change.
    bool                refreshRedOrGammaCurve;         //!< Set true if spline widget is locked and needs refreshing.
    bool                refreshGreenCurve;              //!< Set true if spline widget is locked and needs refreshing.
    bool                refreshBlueCurve;               //!< Set true if spline widget is locked and needs refreshing.

    QGraphicsScene  *scene_ptr;                         //!< Pointer to our Qt "scene" object.
    QGraphicsPixmapItem *pixmap_item_ptr;               //!< Pointer to our Qt "scene pixmap" object.

    fiPixmap            *snap_pixmap;                   //!< The original pixmap for the snapshot.
    fiPixmap            *mod_pixmap;                    //!< The modified pixmap, after applying the LUT.

    std::shared_ptr<aaAaa::aaSpline> aaSplineRedOrGamma;    //!< Curve data for Red or Gamma tab.
    std::shared_ptr<aaAaa::aaSpline> aaSplineGreen;         //!< Curve data for green color channel.
    std::shared_ptr<aaAaa::aaSpline> aaSplineBlue;          //!< Curve data for blue color channel.
    QMutex aaSplineRedOrGammaMutex;                    //!< Mutex to guard access to aaSplineRedOrGamma.
    QMutex aaSplineGreenMutex;                         //!< Mutex to guard access to aaSplineGreen.
    QMutex aaSplineBlueMutex;                          //!< Mutex to guard access to aaSplineBlue.

    bool                splineChangedTimerStarted;      //!< Indicates when single shot timer fired to reduce double shots.
    uint32_t            splineChangedCount;             //!< How many spline changed events inbetween timer events.

    // Note:  When the dialog is created and initialized and it happens to be an RGB condition
    // where three tabs and three spline widgets are set up there is enough of a delay to warrant
    // pulling out the initialization of those splines and let the dialog pop up and cursor spin
    // for a bit until all can come together.  Otherwise the user may hit the button many times
    // due to the fact that the dialog isn't showing up...  This timer was created specifically to
    // separate out that final initialization segment.
    QTimer              splineInitTimer;                //!< Single one-shot timer to facilitate spline curve creation.
    bool                splinesInitialized;             //!< Indicates all of the spline curve(s) are ready to go.

    bool                splineIsLocked;                 //!< Spline lock condition.  All splines update in unison if locked.

    //! Spline curve tab names
    static const QString SPLINE_TAB_GAMMA_NAME;
    static const QString SPLINE_TAB_RED_NAME;
    static const QString SPLINE_TAB_GREEN_NAME;
    static const QString SPLINE_TAB_BLUE_NAME;
};

#endif // CURVE_SETTINGS_H
