/*!
 * @file dcraw_util.cpp
 *
 * @brief Dcraw utility class implementation.
 *
 * Copyright 2008-2017 LibRaw LLC (info@libraw.org)
 * Created: Sat Mar  8, 2008
 *
 * dcraw.c -- Dave Coffin's raw photo decoder
 * Copyright 1997-2014 by Dave Coffin, dcoffin a cybercom o net
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QElapsedTimer>
#include <QTime>
#include <libraw/libraw.h>

#include "logger.h"
#include "dcrawutil.h"

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

class fiLibRaw : public LibRaw
{
public:
    fiLibRaw(unsigned int flags = LIBRAW_OPTIONS_NONE) : LibRaw(flags) {}
    int firaw_process();
    libraw_processed_image_t *firaw_make_raw_image(int *errcode = NULL);
};

/*!
 * @brief Make an "image" from the debayered data. The libraw package doesn't
 *        do this natively, so make a routine for our use.
 *
 * @param errcode   [OUT, optional] Error code, if any.
 * @return The image structure containing the RGB debayered samples.
 */
libraw_processed_image_t *fiLibRaw::firaw_make_raw_image(int *errcode)
{
    int width, height, colors, bps;
    int stride;
    int num_pixels;
    uint32_t data_size;
    uint8_t *pixels;

    // Get the image layout information.
    get_mem_image_format(&width, &height, &colors, &bps);
    stride = width * (bps / 8) * colors;
    data_size = height * stride;
    num_pixels = width * height * (bps / 8);
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "width:" << width << "height:" << height << "colors:" << colors << "bps:" << bps;
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "num_pixels:" << num_pixels;

    // Allocate our return structure.
    libraw_processed_image_t *ret = (libraw_processed_image_t *)::malloc(sizeof(libraw_processed_image_t) + data_size);
    if (!ret)
    {
        if (errcode)
            *errcode = ENOMEM;
        return NULL;
    }

    // Clear the return structure. No need to clear the data part.
    memset(ret, 0, sizeof(libraw_processed_image_t));

    // Set up the return data structure information.
    ret->type = LIBRAW_IMAGE_BITMAP;
    ret->height = height;
    ret->width = width;
    ret->colors = colors;
    ret->bits = bps;
    ret->data_size = data_size;

    // Point to the return data area.
    pixels = ret->data;

    // Copy over the data. The raw image structure is RGBA, and we only want RGB so skip the
    // 4th channel.
// *INDENT-OFF*
#pragma omp parallel for
// *INDENT-ON*
    for (int i = 0; i < num_pixels; i++)
    {
        pixels[(i * 3)] = (uint8_t)(imgdata.image[i][0]);
        pixels[(i * 3) + 1] = (uint8_t)(imgdata.image[i][1]);
        pixels[(i * 3) + 2] = (uint8_t)(imgdata.image[i][2]);
    }

    // Return our information.
    return ret;
}


/*!
 * @brief Replace the standard LibRaw::dcraw_process because we do not want to
 *        do any color correction when loading the image.
 *
 * @return LIBRAW_SUCCESS, or an appropriate error.
 */
int fiLibRaw::firaw_process()
{
    int rc = LIBRAW_SUCCESS;
    int iterations = -1, dcb_enhance = 1;

    try
    {
        libraw_decoder_info_t di;
        get_decoder_info(&di);

        raw2image_ex(0); // allocate imgdata.image and copy data!

        if (imgdata.params.dark_frame)
        {
            subtract(imgdata.params.dark_frame);
            SET_PROC_FLAG(LIBRAW_PROGRESS_DARK_FRAME);
        }

        pre_interpolate();
        SET_PROC_FLAG(LIBRAW_PROGRESS_PRE_INTERPOLATE);

        if (imgdata.params.user_qual == 0)
            lin_interpolate();
        else if (imgdata.params.user_qual == 1 || imgdata.idata.colors > 3)
            vng_interpolate();
        else if (imgdata.params.user_qual == 2 && imgdata.idata.filters > 1000)
            ppg_interpolate();
        else if (imgdata.idata.filters == LIBRAW_XTRANS)
        {
            // Fuji X-Trans
            xtrans_interpolate(imgdata.params.user_qual > 2 ? 3 : 1);
        }
        else if (imgdata.params.user_qual == 3)
            ahd_interpolate(); // really don't need it here due to fallback op
        else if (imgdata.params.user_qual == 4)
            dcb(iterations, dcb_enhance);

        else if (imgdata.params.user_qual == 11)
            dht_interpolate();
        else if (imgdata.params.user_qual == 12)
            aahd_interpolate();
        // fallback to AHD
        else
        {
            ahd_interpolate();
            imgdata.process_warnings |= LIBRAW_WARN_FALLBACK_TO_AHD;
        }

        SET_PROC_FLAG(LIBRAW_PROGRESS_INTERPOLATE);

        if (!libraw_internal_data.output_data.histogram)
        {
            libraw_internal_data.output_data.histogram =
                (int(*)[LIBRAW_HISTOGRAM_SIZE])malloc(
                    sizeof(*libraw_internal_data.output_data.histogram) * 4);
        }
    }
    catch (LibRaw_exceptions err)
    {
        QLOG_DEBUG() << __FUNCTION__ << "Caught an exception:" << err;
        rc = (int)err;
    }

    return rc;
}

DcrawUtil::DcrawUtil()
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Even though the LIBRAW_OPTIONS_NONE parameter is the default value, by
    // specifying it here we can get rid of a compiler warning.
    rawProcessor = std::make_unique<fiLibRaw>(LIBRAW_OPTIONS_NONE);

    // Default gamma.
    gamma = 2.2;
}

DcrawUtil::~DcrawUtil()
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__;
}

/*!
 * @brief When creating a QImage, the pixel buffer has to exist for the lifetime of the
 *        QImage object. QImage allows us to specify a cleanup routine for when the last
 *        instance of the object is destroyed so that we can cleanup the memory.
 *
 * @param info  This will be the QByteArray object pointer.
 */
void DcrawUtil::cleanupByteArray(void *info)
{
    if (info != nullptr)
        delete (QByteArray *)info;
}

/*!
 * @brief When creating a QImage, the pixel buffer has to exist for the lifetime of the
 *        QImage object. QImage allows us to specify a cleanup routine for when the last
 *        instance of the object is destroyed so that we can cleanup the memory.
 *
 * @param info  This will be the LibRaw memory object pointer.
 */
void DcrawUtil::cleanupImageMemory(void *info)
{
    if (info != nullptr)
        LibRaw::dcraw_clear_mem((libraw_processed_image_t *)info);
}

/*!
 * @brief Generate the de-Bayered RGB values from the DNG image.
 *
 * @param dng_ptr       [IN] The location in memory containing the DNG data.
 * @param dng_length    [IN] The number of bytes for the DNG data.
 * @return A pointer to a QImage on success, or nullptr.
 */
QImage *DcrawUtil::debayer(uint8_t *dng_ptr, int dng_length)
{
    int ret;
    QImage *result_ptr = nullptr;                       // Our returned QImage pointer.
    libraw_processed_image_t *image_ptr = nullptr;      // Pointer to the processed image data.

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // We really do NOT want to do any processing on this image. We do, however,
    // want the system to DC-offset and interpolate the Bayer samples to produce
    // RGB colors for us.
    rawProcessor->imgdata.params.gamm[0] = rawProcessor->imgdata.params.gamm[1] = 1;
    rawProcessor->imgdata.params.user_qual = 0;
    rawProcessor->imgdata.idata.filters = 1;
    rawProcessor->imgdata.params.no_auto_bright = 1;
    rawProcessor->imgdata.params.no_auto_scale = 1;
    rawProcessor->imgdata.params.use_auto_wb = 0;
    rawProcessor->imgdata.params.exp_correc = 0;
    rawProcessor->imgdata.params.med_passes = 0;
    rawProcessor->imgdata.params.highlight = 0;
    rawProcessor->imgdata.params.use_fuji_rotate = 0;
    rawProcessor->imgdata.params.dark_frame = nullptr;

    // Load it into our image processor.
    ret = rawProcessor->open_buffer(dng_ptr, dng_length);
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot open_buffer():" << libraw_strerror(ret);
        goto rf_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "open_buffer:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Unpack the image.
    if ((ret = rawProcessor->unpack()) != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot unpack():" << libraw_strerror(ret);
        goto rf_done;
    }
#if 0
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "raw_bps:" << rawProcessor->imgdata.color.raw_bps;
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cam_mul:" << rawProcessor->imgdata.color.cam_mul[0] << rawProcessor->imgdata.color.cam_mul[1] << rawProcessor->imgdata.color.cam_mul[2];
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "asShotNeutral:" << rawProcessor->imgdata.color.dng_levels.asshotneutral[0] << rawProcessor->imgdata.color.dng_levels.asshotneutral[1] << rawProcessor->imgdata.color.dng_levels.asshotneutral[2];
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rgbCam:" << rawProcessor->imgdata.color.rgb_cam[0][0] << rawProcessor->imgdata.color.rgb_cam[0][1] << rawProcessor->imgdata.color.rgb_cam[0][2] << rawProcessor->imgdata.color.rgb_cam[0][3];
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rgbCam:" << rawProcessor->imgdata.color.rgb_cam[1][0] << rawProcessor->imgdata.color.rgb_cam[1][1] << rawProcessor->imgdata.color.rgb_cam[1][2] << rawProcessor->imgdata.color.rgb_cam[1][3];
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rgbCam:" << rawProcessor->imgdata.color.rgb_cam[2][0] << rawProcessor->imgdata.color.rgb_cam[2][1] << rawProcessor->imgdata.color.rgb_cam[2][2] << rawProcessor->imgdata.color.rgb_cam[2][3];
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cdesc:" << rawProcessor->imgdata.idata.cdesc << "color:" << rawProcessor->COLOR(0, 0) << rawProcessor->COLOR(0, 1) << rawProcessor->COLOR(1, 0) << rawProcessor->COLOR(1, 1);
    if (rawProcessor->imgdata.rawdata.raw_image != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "First raw word:" << Qt::showbase << Qt::hex << rawProcessor->imgdata.rawdata.raw_image[0] << rawProcessor->imgdata.rawdata.raw_image[1] << rawProcessor->imgdata.rawdata.raw_image[2] << rawProcessor->imgdata.rawdata.raw_image[3];
    }
#endif

    // Process the data, converting it to an image.
    ret = rawProcessor->firaw_process();
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot do postpocessing:" << libraw_strerror(ret);
        if (LIBRAW_FATAL_ERROR(ret))
            goto rf_done;
    }

//    QLOG_DEBUG() << __FUNCTION__ << "progress_flags:" << Qt::showbase << Qt::hex << rawProcessor->imgdata.progress_flags << "process_warnings:" << rawProcessor->imgdata.process_warnings;
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "First DeB word:" << Qt::showbase << Qt::hex << rawProcessor->imgdata.image[0][0] << rawProcessor->imgdata.image[0][1] << rawProcessor->imgdata.image[0][2] << rawProcessor->imgdata.image[0][3];


#ifdef DC_STATS
//-------------------------------------------------------------------
// Statistics
// Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "RGB32:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
// Statistics
//-------------------------------------------------------------------
#endif

    // Get the image data. Handle things differently depending on the raw BPS value.
    if (rawProcessor->imgdata.color.raw_bps == 8)
        image_ptr = rawProcessor->firaw_make_raw_image(&ret);
    else
        image_ptr = rawProcessor->dcraw_make_mem_image(&ret);

    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "dcraw_make_mem_image() failed:" << libraw_strerror(ret);
        goto rf_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "dcraw_make_mem_image:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Create a new scope so we can define some variables local to this processing.
    if (image_ptr->type == LIBRAW_IMAGE_BITMAP)
    {
        // Allocate a QImage for general consumption.
        result_ptr = new QImage((uint8_t *)image_ptr->data, image_ptr->width, image_ptr->height, QImage::Format_RGB888, DcrawUtil::cleanupImageMemory, image_ptr);
        if (result_ptr->isNull())
        {
            QLOG_DEBUG() << __FUNCTION__ << "Unable to allocate a QImage!";

            // That didn't work...clean up.
            delete result_ptr;
            result_ptr = nullptr;
        }
        else
        {
            // All done with the image_ptr. It will be cleaned up in the destruction of the QImage object.
            image_ptr = nullptr;
        }
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "RGB32:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

rf_done:
    // Recycle the processing object, freeing all the internal memory structures.
    rawProcessor->recycle();

    // Return our result.
    return result_ptr;
}

/*!
 * @brief Open and process a raw image file.
 *
 * @param file_name     [IN] The fully qualified filename to open.
 * @return A pointer to a QImage on success, or nullptr.
 */
QImage *DcrawUtil::raw_file(const QString& file_name)
{
    int ret;
    QImage *result_ptr = nullptr;                       // Our returned QImage pointer.
    libraw_processed_image_t *image_ptr = nullptr;      // Pointer to the processed image data.

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // We really do NOT want to do any processing on this image. We do, however,
    // want the system to DC-offset and interpolate the Bayer samples to produce
    // RGB colors for us.
    rawProcessor->imgdata.params.gamm[0] = rawProcessor->imgdata.params.gamm[1] = 1;
    rawProcessor->imgdata.params.user_qual = 0;
    rawProcessor->imgdata.idata.filters = 1;
    rawProcessor->imgdata.params.no_auto_bright = 1;
    rawProcessor->imgdata.params.no_auto_scale = 1;
    rawProcessor->imgdata.params.use_auto_wb = 0;
    rawProcessor->imgdata.params.exp_correc = 0;
    rawProcessor->imgdata.params.med_passes = 0;
    rawProcessor->imgdata.params.highlight = 0;
    rawProcessor->imgdata.params.use_fuji_rotate = 0;

    // Load it into our image processor.
    ret = rawProcessor->open_file(file_name.toLocal8Bit().constData());
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot open_buffer():" << libraw_strerror(ret);
        goto rf_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "open_buffer:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Unpack the image.
    if ((ret = rawProcessor->unpack()) != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot unpack():" << libraw_strerror(ret);
        goto rf_done;
    }

    // No need to provide a dark frame, as FPN is already done prior to reaching here.
    rawProcessor->imgdata.params.dark_frame = nullptr;

    // Process the data, converting it to an image.
    ret = rawProcessor->firaw_process();
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot do postpocessing:" << libraw_strerror(ret);
        if (LIBRAW_FATAL_ERROR(ret))
            goto rf_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "firaw_process:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

//    QLOG_DEBUG() << __FUNCTION__ << "progress_flags:" << Qt::showbase << Qt::hex << rawProcessor->imgdata.progress_flags << "process_warnings:" << rawProcessor->imgdata.process_warnings;

    // Get the image data.
    image_ptr = rawProcessor->dcraw_make_mem_image(&ret);
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "dcraw_make_mem_image() failed:" << libraw_strerror(ret);
        goto rf_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "dcraw_make_mem_image:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Create a new scope so we can define some variables local to this processing.
    if (image_ptr->type == LIBRAW_IMAGE_BITMAP)
    {
        // Allocate a QImage for general consumption.
        result_ptr = new QImage((uint8_t *)image_ptr->data, image_ptr->width, image_ptr->height, QImage::Format_RGB888, DcrawUtil::cleanupImageMemory, image_ptr);
        if (result_ptr->isNull())
        {
            QLOG_DEBUG() << __FUNCTION__ << "Unable to allocate a QImage!";

            // That didn't work...clean up.
            delete result_ptr;
            result_ptr = nullptr;
        }
        else
        {
            // All done with the image_ptr. It will be cleaned up in the destruction of the QImage object.
            image_ptr = nullptr;
        }
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "RGB32:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

rf_done:
    // If we have an image pointer, go ahead and release the memory.
    if (image_ptr)
        LibRaw::dcraw_clear_mem(image_ptr);

    // Recycle the processing object, freeing all the internal memory structures.
    rawProcessor->recycle();

    // Return our result.
    return result_ptr;
}

/*!
 * @brief Process a DNG located in memory.
 *
 * @param dng_ptr       [IN] The location in memory containing the DNG data.
 * @param dng_length    [IN] The number of bytes for the DNG data.
 * @return A pointer to a QImage on success, or nullptr.
 */
QImage *DcrawUtil::raw_process(uint8_t *dng_ptr, int dng_length)
{
    int ret;
    QImage *result_ptr = nullptr;                       // Our returned QImage pointer.
    libraw_processed_image_t *image_ptr = nullptr;      // Pointer to the processed image data.

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // The following are taken from https://www.libraw.org/docs/API-datastruct.html.
    //
    // Use automatic white balance obtained after averaging over the entire image.
//    rawProcessor->imgdata.params.use_auto_wb = 1;
    //
    // The gamma parameter sets the user gamma-curve.
    // gamm[0] - inverted gamma value.
    // gamm[1] - slope for linear part (so called toe slope). Set to zero for simple power curve.
    //
    // Are we supposed to apply gamma? We know the "knee" for sRGB, others we'll just
    // use a straight calculation.
    if (qFuzzyCompare(gamma, 2.2))
    {
        // Note that the CCM for the Fastec cameras are all calibrated to sRGB values. Of course,
        // not everyone makes use of this fact...
        // sRGB gamma:
        rawProcessor->imgdata.params.gamm[0] = 1.0 / 2.4; rawProcessor->imgdata.params.gamm[1] = 12.92;
    }
    else if (qFuzzyCompare(gamma, 1.0))
    {
        // Linear gamma
        rawProcessor->imgdata.params.gamm[0] = rawProcessor->imgdata.params.gamm[1] = 1.0;
    }
    else
    {
        // Other gamma values.
        rawProcessor->imgdata.params.gamm[0] = 1.0 / gamma; rawProcessor->imgdata.params.gamm[1] = 0.0;
    }

    //
    // Don't use automatic increase of brightness by histogram.
    rawProcessor->imgdata.params.no_auto_bright = 1;
    //
    // The following controls the use of the CCM:
    // 0: do not use embedded color profile
    // 1 (default): use embedded color profile (if present) for DNG files (always); for other files only if use_camera_wb is set;
    // 3: use embedded color data (if present) regardless of white balance setting.
    rawProcessor->imgdata.params.use_camera_matrix = 1;
    //
    // If possible, use the white balance from the camera.
    rawProcessor->imgdata.params.use_camera_wb = 1;
    //
    // Set the interpolation quality:
    // 0 - linear, 1 - VNG, 2 - PPG, 3 - AHD, 4 - DCB, 11 - DHT, 12 - AAHD
    rawProcessor->imgdata.params.user_qual = 12;

    // Load it into our image processor.
    ret = rawProcessor->open_buffer(dng_ptr, dng_length);
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot open_buffer():" << libraw_strerror(ret);
        goto rp_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "open_buffer:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Unpack the image.
    if ((ret = rawProcessor->unpack()) != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot unpack():" << libraw_strerror(ret);
        goto rp_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "unpack:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Process the data, converting it to an image.
    ret = rawProcessor->dcraw_process();
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cannot do postpocessing:" << libraw_strerror(ret);
        if (LIBRAW_FATAL_ERROR(ret))
            goto rp_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "dcraw_process:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Get the image data.
    image_ptr = rawProcessor->dcraw_make_mem_image(&ret);
    if (ret != LIBRAW_SUCCESS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "dcraw_make_mem_image() failed:" << libraw_strerror(ret);
        goto rp_done;
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "dcraw_make_mem_image:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // We don't expect anything else.
    if (image_ptr->type == LIBRAW_IMAGE_BITMAP)
    {
        // Convert to a QImage for general consumption.
        result_ptr = new QImage((uint8_t *)image_ptr->data, image_ptr->width, image_ptr->height, QImage::Format_RGB888, DcrawUtil::cleanupImageMemory, image_ptr);
        if (result_ptr->isNull())
        {
            QLOG_DEBUG() << __FUNCTION__ << "Unable to allocate a QImage!";

            // That didn't work...clean up.
            delete result_ptr;
            result_ptr = nullptr;
        }
        else
        {
            // All done with the image_ptr. It will be cleaned up in the destruction of the QImage object.
            image_ptr = nullptr;
        }
    }

#ifdef DC_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "RGB32:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

rp_done:
    // If we have an image pointer, go ahead and release the memory.
    if (image_ptr)
        LibRaw::dcraw_clear_mem(image_ptr);

    // Recycle the processing object, freeing all the internal memory structures.
    rawProcessor->recycle();

    // Return our result.
    return result_ptr;
}
