/*!
 * @file eventfilter.cpp
 * @brief This is the source file for the event filter dialog class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include "eventfilter.h"
#include "ui_eventfilter.h"

#include <QDebug>

#include "camera_mdi.h"
#include "control_panel.h"
#include "logger.h"
#include "mainwindow.h"
#include "panel_reccontrols.h"
#include "playback_mdi.h"

/*!
 * @brief The eventFilterDlg constructor.
 */
eventFilterDlg::eventFilterDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::eventFilterDlg)
{
    ui->setupUi(this);

    // Adjust our interface based on if we're a series3 camera or not. If we're
    // a series3 camera, then the default interface is fine.
    if (is_series3())
        return;

    // For the seriesNC cameras, we can only have channel 0-2 events plus the Run/Start.
    // We need to repurpose the existing controls.
    ui->so_frame->setVisible(false);
    ui->so_select->setCurrentIndex(FM_EVENT_FILTERS::SELECT_IGNORE);
    ui->ai_frame->setVisible(false);
    ui->ai_select->setCurrentIndex(FM_EVENT_FILTERS::SELECT_IGNORE);

    // Repurpose the A0 controls to be our Video Trigger event controls.
    ui->ao_label->setText("VT");

    // Build our labels.
    build_labels();
}

/*!
 * @brief The eventFilterDlg destructor.
 */
eventFilterDlg::~eventFilterDlg()
{
    delete ui;
}

/*!
 * @brief Return a true if we're a series3 camera, false otherwise.
 * @return true if we don't know what the parent is, otherwise whatever the parent tells us.
 */
bool eventFilterDlg::is_series3()
{
    PlaybackMDI *playback_ptr;
    CameraMDI   *camera_ptr;

    // Do we have a camera pointer?
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr != nullptr)
        return camera_ptr->is_series3();

    // Do we have a playback pointer?
    playback_ptr = qobject_cast<PlaybackMDI *>(parent());
    if (playback_ptr != nullptr)
        return playback_ptr->isSeries3();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unknown parent object!";
    return true;
}


/*!
 * @brief Enable (check) or disable (clear) the event marker enable checkbox.
 * @param do_enable     True to check, false to clear the event marker checkbox.
 */
void eventFilterDlg::set_event_markers_enable(bool do_enable)
{
    // Save the setting.
    event_markers_enabled = do_enable;

    // Adjust the UI to reflect the setting.
    ui->event_enable->setChecked(do_enable);
}

/*!
 * @brief Enable (check) or disable (clear) the display graph enable checkbox.
 * @param do_enable     True to check, false to clear the display graph checkbox.
 */
void eventFilterDlg::set_display_graph_enable(bool do_enable)
{
    // Save the setting.
    display_io_graph = do_enable;

    // Adjust the UI to reflect the setting.
    ui->graph_display->setChecked(do_enable);
}

/*!
 * @brief Set either the "and" or the "or" radio button to indicate the logic
 *        to use when combining the selections.
 *
 * @param use_and   True to use "and", false for "or".
 */
void eventFilterDlg::set_logic_to_and(bool use_and)
{
    // Save the setting.
    logic_is_and = use_and;

    // Adjust the UI to reflect the setting.
    ui->and_rb->setChecked(use_and);
    ui->or_rb->setChecked(!use_and);
}

/*!
 * @brief Return a list of [selections](@ref FM_EVENT_FILTERS::EVENT_TYPES).
 * @return A list of the selections, indexed by FM_EVENT_FILTERS::SIGNAL_INDICIES.
 */
const QList<int>& eventFilterDlg::event_selects(void)
{
    return event_selections;
}

/*!
 * @brief Set the selections to the "new_selects" parameter.
 * @param new_selects   A list of [selections](@ref FM_EVENT_FILTERS::EVENT_TYPES).
 */
void eventFilterDlg::set_selects(const QList<int>& new_selects)
{
    int i;

    // Make sure the list getting passed in makes sense.
    if (new_selects.size() < FM_EVENT_FILTERS::INDEX_COUNT)
        return;

    // Clear out the old list.
    event_selections.clear();

    // Copy over the new.
    for (i = 0; i < new_selects.size(); ++i)
        event_selections.append(new_selects.at(i));

    // Adjust the UI to reflect the settings.
    ui->rs_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::RS_INDEX));
    ui->ti_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::TI_INDEX));
    ui->to_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::TO_INDEX));
    ui->si_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::SI_INDEX));

    // If we're series3, we want to set the following. For seriesNC cameras, these
    // are don't cares.
    if (is_series3())
    {
        ui->so_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::SO_INDEX));
        ui->ai_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::AI_INDEX));
        ui->ao_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::AO_INDEX));
    }
    else
    {
        // Take AO to use for VT.
        ui->ao_select->setCurrentIndex(event_selections.at(FM_EVENT_FILTERS::VT_INDEX));

        ui->so_select->setCurrentIndex(FM_EVENT_FILTERS::SELECT_IGNORE);
        ui->ai_select->setCurrentIndex(FM_EVENT_FILTERS::SELECT_IGNORE);
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Standard dialog method invoked when the user presses the OK button.
 */
void eventFilterDlg::accept(void)
{
    int i;

    // Read in all the controls and set our local values to match.
    event_markers_enabled = ui->event_enable->isChecked();
    display_io_graph = ui->graph_display->isChecked();
    logic_is_and = ui->and_rb->isChecked();

    // Clear out the old selection setting.
    event_selections.clear();

    // This is a little over the top, but we need to make sure
    // that the order of items in the list matches the indecies.
    for (i = 0; i < FM_EVENT_FILTERS::INDEX_COUNT; ++i)
    {
        switch (i)
        {
        case FM_EVENT_FILTERS::VT_INDEX:
            if (is_series3())
                event_selections.append(FM_EVENT_FILTERS::SELECT_IGNORE);
            else
                event_selections.append(ui->ao_select->currentIndex());
            break;
        case FM_EVENT_FILTERS::RS_INDEX:
            event_selections.append(ui->rs_select->currentIndex());
            break;
        case FM_EVENT_FILTERS::TI_INDEX:
            event_selections.append(ui->ti_select->currentIndex());
            break;
        case FM_EVENT_FILTERS::TO_INDEX:
            event_selections.append(ui->to_select->currentIndex());
            break;
        case FM_EVENT_FILTERS::SI_INDEX:
            event_selections.append(ui->si_select->currentIndex());
            break;
        case FM_EVENT_FILTERS::SO_INDEX:
            event_selections.append(ui->so_select->currentIndex());
            break;
        case FM_EVENT_FILTERS::AI_INDEX:
            event_selections.append(ui->ai_select->currentIndex());
            break;
        case FM_EVENT_FILTERS::AO_INDEX:
            event_selections.append(ui->ao_select->currentIndex());
            break;
        default:
            qDebug() << __PRETTY_FUNCTION__ << "What?! Unexpected index!" << i;
            break;
        }
    }

    // Pass it on.
    QDialog::accept();
}

void eventFilterDlg::build_labels()
{
    QString label_str;
    REC_CONTROLS::EXT_IO_TYPES ext_io;
    PanelRecControls *controls_ptr;

    // Sanity check. Not for series3 cameras.
    if (is_series3())
        return;

    // Get our record control panel object.
    controls_ptr = static_cast<PanelRecControls *>(MainWindow::GetInstance()->get_control_panel()->getPanel(CONTROL_PANEL::PANEL_REC_CONTROLS));

    // Change the labels of the remaining controls.
    label_str.clear();
    ext_io = controls_ptr->getType(0);
    if (ext_io != REC_CONTROLS::EXT_IO_TYPES::EXTIO_NONE)
    {
        label_str = controls_ptr->getExtioLabel(ext_io) + QString("\n");
    }
    label_str.append("Ch0");
    ui->si_label->setText(label_str);

    label_str.clear();
    ext_io = controls_ptr->getType(1);
    if (ext_io != REC_CONTROLS::EXT_IO_TYPES::EXTIO_NONE)
    {
        label_str = controls_ptr->getExtioLabel(ext_io) + QString("\n");
    }
    label_str.append("Ch1");
    ui->to_label->setText(label_str);

    label_str.clear();
    ext_io = controls_ptr->getType(2);
    if (ext_io != REC_CONTROLS::EXT_IO_TYPES::EXTIO_NONE)
    {
        label_str = controls_ptr->getExtioLabel(ext_io) + QString("\n");
    }
    label_str.append("Ch2");
    ui->ti_label->setText(label_str);
}
