/*!
 * @file eventfilter.h
 *
 * @brief This is the header file for the event filter dialog.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef EVENTFILTER_H
#define EVENTFILTER_H

#include <QDialog>
#include <QList>

namespace Ui
{
class eventFilterDlg;
}

namespace FM_EVENT_FILTERS
{
//--------------------------------------------------------------------
// Constant, Macro and Type definitions
//--------------------------------------------------------------------

//! Set our maximum event marker list size.
const int       MAX_MARKER_LIST_SIZE    = 100000;

//! Where to find the event flags. Word is byte swapped, so this is the MSByte.
const int       EVENT_FLAG_OFFSET       = 24;

//! Where to find the edge flags. Word is byte swapped, so this is the upper middle byte.
const int       EDGE_FLAG_OFFSET        = 16;

//! An enumeration of the various events the user can select.
enum EVENT_TYPES
{
    SELECT_IGNORE = 0,          //!< Not interested in this signal.
    SELECT_LOW,                 //!< The signal is low.
    SELECT_EDGE_LOW,            //!< The signal transitioned to a low level.
    SELECT_HIGH,                //!< The signal is high.
    SELECT_EDGE_HIGH,           //!< The signal transitioned to a high level.
    SELECT_CHANGE,              //!< The signal changed from low to high or high to low.
    SELECT_GLITCH               //!< The signal changed, but so quick we couldn't record the new value.
};

//! Enumerate the various signals we use for events.
enum SIGNAL_INDICIES
{
    AO_INDEX = 0,               //!< The ARM-OUT signal.
    AI_INDEX,                   //!< The ARM-IN signal.
    SO_INDEX,                   //!< The SYNC-OUT signal.
    SI_INDEX,                   //!< The SYNC-IN signal.
    TO_INDEX,                   //!< The TRIGGER-OUT signal.
    TI_INDEX,                   //!< The TRIGGER-IN signal.
    RS_INDEX,                   //!< The RECORD-START event.
    VT_INDEX,                   //!< The VIDEO-TRIGGER event.
    INDEX_COUNT                 //!< Total number of event "sources".
};

//! For the seriesNC cameras, SI_INDEX is Channel0.
const int CH0_INDEX = FM_EVENT_FILTERS::SI_INDEX;

//! For the seriesNC cameras, TO_INDEX is Channel1.
const int CH1_INDEX = FM_EVENT_FILTERS::TO_INDEX;

//! For the seriesNC cameras, TI_INDEX is Channel2.
const int CH2_INDEX = FM_EVENT_FILTERS::TI_INDEX;

}   // End of the FM_EVENT_FILTERS namespace.

//--------------------------------------------------------------------
// The eventFilterDlg class definition
//--------------------------------------------------------------------
class eventFilterDlg : public QDialog
{
    Q_OBJECT

public:
    explicit eventFilterDlg(QWidget *parent = 0);
    ~eventFilterDlg();

    bool    event_markers_are_enabled(void)       { return event_markers_enabled; }
    void    set_event_markers_enable(bool do_enable);
    bool    display_graph_enabled(void)           { return display_io_graph; }
    void    set_display_graph_enable(bool do_enable);
    bool    use_logic_and(void)                   { return logic_is_and; }
    void    set_logic_to_and(bool use_and);

    const QList<int>& event_selects(void);
    void    set_selects(const QList<int>& new_selects);

protected slots:
    virtual void    accept(void);               //!< Invoked when the user presses the OK button.

private:
    Ui::eventFilterDlg *ui;

    bool    event_markers_enabled;              //!< True if event marker functionality is enabled, false otherwise.
    bool    display_io_graph;                   //!< True if we want to display the I/O graph, false otherwise.
    bool    logic_is_and;                       //!< True if we are using the logical AND function, false if we're doing OR.

    QList<int>  event_selections;               //!< The event marker selections.

    void    build_labels();
    bool    is_series3();
};

//--------------------------------------------------------------------
// The eventFilter class definition
//--------------------------------------------------------------------
class eventFilter
{
public:
    eventFilter() {}
    ~eventFilter() {}

    /*!
     * @brief Compare the marker from the per-frame metadata with our search criteria to
     *        see if there's a match.
     *
     * @note    The marker information is the raw data straight from the camera. It is
     *          byte swapped, so we need to offset our bit masks accordingly.
     *
     * @param last_markers  The last frame's markers, in case we're interested in changes.
     * @param cur_markers   The current frame's markers to compare with.
     * @param event_selections  The list of selected events.
     * @param logic_is_and      True if the bits are 'anded' or 'ored' together.
     *
     * @returns True if there is a match, False otherwise.
     */
    static bool match_criteria(uint32_t last_markers, uint32_t cur_markers, QList<int>& event_selections, bool logic_is_and, bool series3 = true)
    {
        int     i;
        int     start_index;
        int     mask(0);
        int     index_match(0);
        bool    got_a_match(false);

        // We have a different start index depending on if we're a series3 camera or
        // a seriesNC camera.
        if (series3)
            start_index = FM_EVENT_FILTERS::AO_INDEX;
        else
            start_index = FM_EVENT_FILTERS::CH0_INDEX;

        // Look through each of the indices and see how they compare.
        for (i = start_index; i < FM_EVENT_FILTERS::INDEX_COUNT; ++i)
        {
            uint32_t    event_mask;
            uint32_t    edge_mask;

            // Build our bitmask. The mask is different if we're looking at
            // a channel or not. Series3 cameras don't have to worry about this.
            if (series3)
            {
                event_mask = (1 << (i + FM_EVENT_FILTERS::EVENT_FLAG_OFFSET));
                edge_mask = (1 << (i + FM_EVENT_FILTERS::EDGE_FLAG_OFFSET));
            }
            else
            {
                // Not Series3. Are we looking at a channel index?
                if (i > FM_EVENT_FILTERS::CH2_INDEX)
                {
                    // Not a channel, so treat this as normal.
                    event_mask = (1 << (i + FM_EVENT_FILTERS::EVENT_FLAG_OFFSET));
                    edge_mask = (1 << (i + FM_EVENT_FILTERS::EDGE_FLAG_OFFSET));
                }
                else
                {
                    // The channel events start at bit 0, so we need to shift things.
                    event_mask = (1 << ((i - start_index) + FM_EVENT_FILTERS::EVENT_FLAG_OFFSET));
                    edge_mask = (1 << ((i - start_index) + FM_EVENT_FILTERS::EDGE_FLAG_OFFSET));
                }
            }

            // Look at this event.
            switch (event_selections.at(i))
            {
            default:
            case FM_EVENT_FILTERS::SELECT_IGNORE:
                break;
            case FM_EVENT_FILTERS::SELECT_LOW:
                // The event value is a '0'.
                index_match |= (((cur_markers & event_mask) == 0) << i);
                mask |= (1 << i);
                break;
            case FM_EVENT_FILTERS::SELECT_EDGE_LOW:
                // We have an edge indicator and the event value is a '0'.
                index_match |= ((((cur_markers & edge_mask) == edge_mask) && ((cur_markers & event_mask) == 0)) << i);
                mask |= (1 << i);
                break;
            case FM_EVENT_FILTERS::SELECT_HIGH:
                // The event value is a '1'.
                index_match |= (((cur_markers & event_mask) == event_mask) << i);
                mask |= (1 << i);
                break;
            case FM_EVENT_FILTERS::SELECT_EDGE_HIGH:
                // We have an edge indicator and the event value is a '1'.
                index_match |= ((((cur_markers & edge_mask) == edge_mask) && ((cur_markers & event_mask) == event_mask)) << i);
                mask |= (1 << i);
                break;
            case FM_EVENT_FILTERS::SELECT_CHANGE:
                // The event value is different than the previous event value.
                index_match |= (((cur_markers & event_mask) != (last_markers & event_mask)) << i);
                mask |= (1 << i);
                break;
            case FM_EVENT_FILTERS::SELECT_GLITCH:
                // A "glitch" is defined as: there was an edge indicator, but the event value didn't change.
                index_match |= ((((cur_markers & edge_mask) == edge_mask) && ((cur_markers & event_mask) == (last_markers & event_mask))) << i);
                mask |= (1 << i);
                break;
            }
        }

        // Do we care about any bit?
        if (mask)
        {
            // Determine how we're supposed to combine.
            if (logic_is_and)
            {
                // We are AND-ing.
                // First, grab the bits we're interested in (the AND operation), then
                // check to see if ALL the interesting bits are set (1 XOR 1 = 0, 1 XOR 0 = 1).
                got_a_match = (((index_match & mask) ^ mask) == 0);
            }
            else
            {
                // We are OR-ing.
                got_a_match = ((index_match & mask) != 0);
            }
        }

        // Return our decision.
        return got_a_match;
    }
};

#endif // EVENTFILTER_H
