/*!
 * @file export_dlg.cpp
 * @brief This object allows the user to export a capture to a particular format.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "export_dlg.h"
#include "ui_export_dlg.h"

#include <QDesktopServices>
#include <QString>
#include <QProcess>
#include <QScrollBar>
#include <QMessageBox>
#include <QFileInfo>
#include <QFileDialog>
#include <QSettings>

#include "fi_saver.h"
#include "logger.h"
#include "user_preferences.h"

namespace EXPORT_DIALOG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The maximum number of settings that we'll remember.
const int           MAX_SETTINGS        =   10;

}   // End of namespace EXPORT_DIALOG.

using namespace EXPORT_DIALOG;


ExportDlg::ExportDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::ExportDlg),
    DEFAULT_ARG_LIST(QStringList() << QStringLiteral("-vcodec prores -profile:v 2") << QStringLiteral("-vcodec mjpeg -qscale 8"))
{
    QSettings   settings;
    QStringList arg_list;
    bool        no_args;

    // Initialize our members.
    status_string.clear();
    transcoding_process = nullptr;
    clip_start = 0;
    clip_end = 1;
    export_path.clear();
    num_frames = 0;
    record_fps = 0;
    source_path.clear();
    trig_frame = 0;

    // Initialize the ui object.
    ui->setupUi(this);

    // Sanity check.
    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    ffmpeg_path = settings.value("ffmpeg_path").toString();
    settings.endGroup();
    if (ffmpeg_path.isEmpty() || !QFile::exists(ffmpeg_path))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No FFMPEG specified!";
        return;
    }

    // Initialize the default path based on the saved settings.
    settings.beginGroup("ExportInfo");
    export_path = settings.value("last_export").toString();
    arg_list = settings.value("user_args").toStringList();
    no_args = settings.value("no_args", true).toBool();
    settings.endGroup();

    // Update our combo box.
    if (arg_list.isEmpty())
    {
        arg_list = DEFAULT_ARG_LIST;
        no_args = true;
    }
    update_args(arg_list, no_args);

    // Initialize our controls.
    ui->toLineEdit->setText(QDir::toNativeSeparators(export_path));

    // Create our transcoding process object.
    transcoding_process = new QProcess(this);
    if (transcoding_process == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate our transcoding process object!";
        return;
    }

    // Connect signals/slots.
    connect(transcoding_process, SIGNAL(started()), this, SLOT(process_started()));
    connect(transcoding_process, SIGNAL(readyReadStandardOutput()), this, SLOT(stdout_ready()));
    connect(transcoding_process, SIGNAL(finished(int)), this, SLOT(encoding_finished()));
    connect(ui->closeButton, SIGNAL(clicked()), this, SLOT(close_clicked()));
    connect(ui->helpButton, SIGNAL(clicked()), this, SLOT(help_clicked()));
    connect(ui->toFileBrowser, SIGNAL(clicked()), this, SLOT(to_browser()));
    connect(ui->startButton, SIGNAL(clicked()), this, SLOT(start_clicked()));
}

ExportDlg::~ExportDlg()
{
    QSettings   settings;

    // Save our last used settings.
    settings.beginGroup("ExportInfo");
    settings.setValue("last_export", ui->toLineEdit->text());
    settings.endGroup();

    // Get rid of the UI object.
    delete ui;
}

/*!
 * @brief Initialize the start/end clip values using absolute frame numbers.
 *
 * @param start_clip    The start clip frame to use.
 * @param end_clip      The ending clip frame to use.
 *
 * @sa ExportDlg::set_rel_range
 */
void ExportDlg::set_clip_range(uint32_t start_clip, uint32_t end_clip)
{
    int rel_max;
    int rel_min;
    int rel_start;
    int rel_end;

    // Save the settings.
    clip_start = start_clip;
    clip_end = end_clip;

    // If the user specified a -1 as our end clip, that means to use the number of frames.
    if (end_clip == (uint32_t) -1)
        end_clip = num_frames;

    // Recalculate what our min/max values are.
    rel_min = (int)(start_clip - trig_frame);
    rel_start = rel_min;
    rel_max = (int)(end_clip - trig_frame);
    rel_end = rel_max;

    // Update our controls.
    ui->spin_clip_start->blockSignals(true);
    ui->spin_clip_end->blockSignals(true);
    ui->spin_clip_start->setMinimum(rel_min);
    ui->spin_clip_end->setMinimum(rel_min + 1);
    ui->spin_clip_start->setMaximum(rel_max - 1);
    ui->spin_clip_end->setMaximum(rel_max);
    ui->spin_clip_start->blockSignals(false);
    ui->spin_clip_end->blockSignals(false);

    // Update the starting/ending positions.
    ui->spin_clip_start->setValue(rel_start);
    ui->spin_clip_end->setValue(rel_end);
}

/*!
 * @brief Initialize the start/end clip values using trigger-relative
 *        values rather than absolute frame numbers.
 *
 * @param start_clip    The start clip frame to use.
 * @param end_clip      The ending clip frame to use.
 *
 * @sa ExportDlg::set_clip_range
 */
void ExportDlg::set_rel_range(int start_clip, int end_clip)
{
    // Sanity check.
    if (start_clip < ui->spin_clip_start->minimum())
        start_clip = ui->spin_clip_start->minimum();
    if (end_clip > ui->spin_clip_end->maximum())
        end_clip = ui->spin_clip_end->maximum();

    // Are they the same?
    if (start_clip == end_clip)
    {
        // Use the full range.
        start_clip = ui->spin_clip_start->minimum();
        end_clip = ui->spin_clip_end->maximum();
    }

    // Set the controls.
    ui->spin_clip_start->setValue(start_clip);
    ui->spin_clip_end->setValue(end_clip);
}

void ExportDlg::set_source_path(const QString& src_path)
{
    source_path = src_path;
}

void ExportDlg::update_args(const QStringList& arg_list, bool no_args)
{
    int i;

    // Sanity check.
    if (arg_list.isEmpty())
        return;

    // Clear the list.
    ui->argsCombo->clear();

    // Populate the list with our settings.
    for (i = 0; i < arg_list.count(); ++i)
        ui->argsCombo->insertItem(i, arg_list.at(i));

    // Were there any arguments to select?
    if (no_args)
    {
        // Don't select any.
        ui->argsCombo->setCurrentIndex(-1);
    }
    else
    {
        // Make the first one the active one.
        ui->argsCombo->setCurrentIndex(0);
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void ExportDlg::close_clicked(void)
{
    // Do we have a process and is it running?
    if ((transcoding_process != nullptr) && (transcoding_process->state() == QProcess::Running))
    {
        if (QMessageBox::question(this, tr("Export Video"),
                                  tr("Are you sure you want to cancel the export process?"),
                                  QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No)
        {
            return;
        }

        // Shutdown the transcode.
        transcoding_process->kill();
    }

    // Go ahead and close.
    close();
}

void ExportDlg::encoding_finished(void)
{
    // Enable the buttons.
    ui->startButton->setEnabled(true);
    ui->toFileBrowser->setEnabled(true);

    // Set the encoding status by checking output file's existence
    QString fileName = ui->toLineEdit->text();

    if (QFile::exists(fileName))
    {
        ui->transcodingStatusLabel->setText(tr("Transcoding Status: Successful!"));
    }
    else
    {
        ui->transcodingStatusLabel->setText(tr("Transcoding Status: Failed!"));
    }
}

void ExportDlg::help_clicked(void)
{
    QDesktopServices::openUrl(QUrl("http://ffmpeg.org/ffmpeg.html"));
}

void ExportDlg::process_started(void)
{
    ui->transcodingStatusLabel->setText(tr("Transcoding Status: Working..."));
}

void ExportDlg::start_clicked(void)
{
    QSettings   settings;
    QStringList arguments;
    QStringList input_args;
    QString     opt_args;
    QString     file_ext;
    QFileInfo   file_info(source_path);
    int         start_frame;
    int         frame_count;

    if (source_path.isEmpty())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No source";
        QMessageBox::information(this, tr("Export Video"), tr("Input file not specified"));
        return;
    }

    // Figure our starting frame and the number of frames.
    start_frame = ui->spin_clip_start->value() + (int)trig_frame;
    frame_count = ((ui->spin_clip_end->value() + (int)trig_frame) - start_frame) + 1;

    // If we're specifying a playback rate, set the input rate the same, otherwise
    // we'll be dropping frames.
    if (ui->spin_play_fps->value() > 0)
        input_args << "-r" << QString("%1").arg(ui->spin_play_fps->value());

    // Determine if we are playing an AVI or a stack of images.
    file_ext = file_info.suffix().toLower();
    if (file_ext.compare(FiSaver::IMAGE_EXT_AVI, Qt::CaseInsensitive) == 0)
    {
        // Sometimes the start frame number isn't 0 based, so adjust now.
        start_frame = start_frame - (int)clip_start;

        // Are we specifying a start frame?
        if (start_frame > 0)
        {
            // Sanity check.
            if (record_fps == 0)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "What! You forgot to set the recording FPS!";
            }
            else
            {
                // Need to convert this to a time.
                input_args << "-ss" << QString("%1").arg((double)start_frame / record_fps);
            }
        }

        // We're dealing with an AVI, so just move the path on over.
        input_args << "-i" << source_path;
    }
    else
    {
        int sub_index;
        QString     input_file;

        // We are working on an image stack, we need to know if sub folders are being used.
        sub_index = source_path.indexOf(QString("/") + FiSaver::SUB_FOLDER_PREFIX, Qt::CaseInsensitive);
        if (sub_index >= 0)
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "Can't do sub-folders!";
            QMessageBox::information(this, tr("Export Video"), tr("Images are split across sub-folders. FFMPEG does not support this."));
            return;
        }

        // For a list of individual files, we want to specify the "image2" demuxer.
        input_args << "-f" << "image2";
        if (start_frame > 0)
            input_args << "-start_number" << QString("%1").arg(start_frame);

        // We need to specify the file information as a regex expression.
        input_file = file_info.path() + QString("/%07d.") + file_ext;
        input_args << "-i" << input_file;
    }

    QString output_file = QDir::fromNativeSeparators(ui->toLineEdit->text());
    if (output_file.isEmpty())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No output";
        QMessageBox::information(this, tr("Export Video"), tr("Output file not specified"));
        return;
    }

    if (QFile::exists(output_file))
    {
        if (QMessageBox::question(this, tr("Export Video"),
                                  tr("There already exists a file called %1 in the current directory. Overwrite?").arg(QDir::toNativeSeparators(output_file)),
                                  QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::No)
        {
            return;
        }

        QFile::remove(output_file);
        while (QFile::exists(output_file))
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "Waiting for file to be gone.";
            qApp->processEvents();
        }
    }

    // Grab the optional arguments.
    opt_args = ui->argsCombo->currentText();

    // Build the argument list.
    arguments << input_args;

    // Any optional arguments are considered "output" arguments. Add them now.
    if (!opt_args.isEmpty())
        arguments << opt_args.split(" ");
    if (ui->spin_play_fps->value() > 0)
        arguments << "-r" << QString("%1").arg(ui->spin_play_fps->value());
    if (frame_count < (int)num_frames)
        arguments << "-vframes" << QString("%1").arg(frame_count);
    arguments << output_file;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << arguments;

    // Start the export.
    transcoding_process->setProcessChannelMode(QProcess::MergedChannels);
    transcoding_process->start(ffmpeg_path, arguments);

    // Disable some of the buttons.
    ui->startButton->setEnabled(false);
    ui->toFileBrowser->setEnabled(false);

    // Save our last used settings.
    settings.beginGroup("ExportInfo");

    // Now that we have the transcoding started, update our argument list.
    if (opt_args.isEmpty())
    {
        // Remember that there were no arguments selected.
        settings.setValue("no_args", true);
    }
    else
    {
        QStringList arg_list;

        // Get the list of saved arguments.
        arg_list = settings.value("user_args").toStringList();
        if (arg_list.isEmpty())
        {
            arg_list << DEFAULT_ARG_LIST;
        }

        // Make sure there aren't any duplicates.
        arg_list.removeAll(opt_args);

        // Put the new argument as the first entry.
        arg_list.prepend(opt_args);

        // Only allow a limited number to hang around.
        while (arg_list.size() > MAX_SETTINGS)
            arg_list.removeLast();

        // Save the list.
        settings.setValue("user_args", arg_list);

        // Remember that there was an argument selected.
        settings.setValue("no_args", false);

        // Update the combo box.
        update_args(arg_list, false);
    }


    // All done with the settings.
    settings.endGroup();
}

void ExportDlg::stdout_ready(void)
{
    // Append the stdout data to our string.
    status_string.append(transcoding_process->readAllStandardOutput());
    ui->statusEdit->setText(status_string);

    // Put the slider at the bottom.
    ui->statusEdit->verticalScrollBar()->setSliderPosition(ui->statusEdit->verticalScrollBar()->maximum());
}

void ExportDlg::to_browser(void)
{
    QString fileName;

    fileName = QFileDialog::getSaveFileName(this, tr("Save Video"), ui->toLineEdit->text(), tr("Videos (*.avi *.mov *.mp4)"));
    if (!fileName.isEmpty())
    {
        ui->toLineEdit->setText(QDir::toNativeSeparators(fileName));
    }
}

