/*!
 * @file fcp_socket.cpp
 * @brief This file contains the implementation of the FCPSocket
 *        class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "fcp_socket.h"

#include "fi_comm.h"
#include "logger.h"
#include "gige_socket.h"
#include "fsp_socket.h"
#include "mainwindow.h"

namespace FCP_SOCKET
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! In most situations, this timeout is sufficient.
const int   DEFAULT_TIMEOUT             = 6000;

//! In some situations, a longer timeout is required.
const int   LONG_MSG_TIMEOUT            = 10000;

//! Define a constant for the global broadcast IP address.
static const char *GLOBAL_BROADCAST_ADDRESS = "255.255.255.255";

//! Length for our dummy packet to open the FMP port(s).
const int   DUMMY_PACKET_LEN            = 10;

}   // End of namespace FCP_SOCKET.

using namespace FCP_COMM;
using namespace FCP_SOCKET;

#ifdef DEBUG_FORMAT_ROUTINE
/*!
 * @brief Utility function to create a Qt::hex dump from a QByteArray.
 *
 * @param ba    The QByteArray we're interested in.
 * @return The character Qt::hex dump.
 */
QString byte_array_to_string(QByteArray& ba)
{
    int     i;
    QString resQt::hex;

    for (i = 0; i < ba.size(); ++i)
    {
        uint8_t value = (uint8_t)ba.at(i);
        resQt::hex.append(QString::fromUtf8("%1 ").arg(value, 2, 16, QLatin1Char('0')));
    }

    return resQt::hex;
}
#endif

//--------------------------------------------------------------------
// FCPSocket: class implementation
//--------------------------------------------------------------------

/*!
 * @brief Constructor for the FCPSocket class.
 *
 * @param my_ip_addr    The camera's IP address.
 * @param default_port  The port to use for communicating with cameras.
 */
FCPSocket::FCPSocket(uint32_t my_ip_addr, uint32_t default_port) :
    QObject(nullptr)
{
    // Initialize our members.
    initialized = false;
    offline = true;
    gige_socket = nullptr;
    fmp_socket = nullptr;
    fsp_socket = nullptr;
    last_errcode = FI_COMM::FI_SUCCESS;
    cam_ip_addr = my_ip_addr;
    pending_read_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    pending_write_value = 0;
    pending_write_lock = false;
    pending_msg_id = 0;
    msg_queue.clear();

    msg_port = (int)default_port;
    msg_id = 0;
    msg_timer_id = 0;
}

/*!
 * @brief The FCPSocket class destructor.
 */
FCPSocket::~FCPSocket()
{
    // Empty out our receive message list.
    while (!rcv_list.isEmpty())
    {
        fcp_msg *p_msg = rcv_list.takeFirst();
        delete p_msg;
    }
}

/*!
 * @brief Provide a method to reset the pending read list.
 */
void FCPSocket::clear_pending_read_list()
{
    pending_read_list.clear();
}

/*!
 * @brief This method is called to find the network address information for a specific adapter.
 *
 * @param   adapter_name    The network adapter selected by the user.
 * @param   addr_entry      The return information for the specific adapter, if it was found.
 *
 * @return True if the adapter_name was found (which makes addr_entry valid), false otherwise
 *         (in which case, addr_entry is invalid).
 */
bool FCPSocket::find_adapter(const QString& adapter_name, QNetworkAddressEntry& addr_entry)
{
    bool    adapter_found(false);               // Assume we will not find the adapter.

    // Find out all the interfaces which are on the system
    QList<QNetworkInterface> net_interfaces = QNetworkInterface::allInterfaces();

    // Since there may be multiple network interfaces on the client machine we need
    // to find the one we're interested in.
    for (int i = 0; !adapter_found && (i < net_interfaces.size()); ++i)
    {
        const QNetworkInterface net_interface = net_interfaces.at(i);
        QNetworkInterface::InterfaceFlags   flags = net_interface.flags();
        QList<QNetworkAddressEntry>         all_address;

        // Make sure the interface is valid.
        if (((flags & QNetworkInterface::IsUp) == 0) ||                     // Network isn't up.
            ((flags & QNetworkInterface::IsRunning) == 0))                  // Network isn't running.
        {
            continue;  // not the right kind of interface
        }

        // The network is a loopback interface and we're not interested in local only operation.
        if (((flags & QNetworkInterface::IsLoopBack) == QNetworkInterface::IsLoopBack) && !MainWindow::local_control())
            continue;

        // Check to see if this is the adapter we're interested in.
        QString interface_name = net_interface.humanReadableName();
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "interface_name:" << interface_name << "adapter_name:" << adapter_name;
        if (interface_name.compare(adapter_name, Qt::CaseInsensitive) != 0)
            continue;

        // Get the list of addresses which are associated with this network interface.
        all_address = net_interface.addressEntries();

        // Get the relevant IP info.
        for (int j = 0; j < all_address.size(); j++)
        {
            // Note that we are interested in the IPv4 address and we want to skip the IPv6
            // address, which we can find because IPv6 address's don't have broadcast address's.
            if (all_address.at(j).broadcast().isNull())
                continue;

            // We found it. Save the address information and leave.
            adapter_found = true;
            addr_entry = all_address.at(j);
            break;
        }
    }

    // Return our status.
    return adapter_found;
}

bool FCPSocket::find_adapter(uint32_t adapter_ip, QNetworkAddressEntry& addr_entry)
{
    bool    adapter_found(false);               // Assume we will not find the adapter.

    // Find out all the interfaces which are on the system
    QList<QNetworkInterface> net_interfaces = QNetworkInterface::allInterfaces();

    // Since there may be multiple network interfaces on the client machine we need
    // to find the one we're interested in.
    for (int i = 0; !adapter_found && (i < net_interfaces.size()); ++i)
    {
        const QNetworkInterface net_interface = net_interfaces.at(i);
        QNetworkInterface::InterfaceFlags   flags = net_interface.flags();
        QList<QNetworkAddressEntry>         all_address;

        // Make sure the interface is valid.
        if (((flags & QNetworkInterface::IsUp) == 0) ||                     // Network isn't up.
            ((flags & QNetworkInterface::IsRunning) == 0))                  // Network isn't running.
            continue;

        // The network is a loopback interface and we're not interested in local only operation.
        if (((flags & QNetworkInterface::IsLoopBack) == QNetworkInterface::IsLoopBack) && !MainWindow::local_control())
            continue;

        // Get the list of addresses which are associated with this network interface.
        all_address = net_interface.addressEntries();

        // Get the relevant IP info.
        for (int j = 0; j < all_address.size(); j++)
        {
            // Note that we are interested in the IPv4 address and we want to skip the IPv6
            // address, which we can find because IPv6 address's don't have broadcast address's.
            if (all_address.at(j).broadcast().isNull())
                continue;

            // Make sure our adapter IP address matches.
            if (all_address.at(j).ip().toIPv4Address() != adapter_ip)
                continue;

            // We found it. Save the address information and leave.
            adapter_found = true;
            addr_entry = all_address.at(j);
            break;
        }
    }

    // Return our status.
    return adapter_found;
}

/*!
 * @brief Force the camera to change it's network settings. We send this
 *        as an IP broadcast (not a MAC broadcast), so we use the MAC
 *        address of the camera to target the camera.
 *
 * @param mac_addr      The MAC address for the camera.
 * @param new_ip        The new IP address for the camera.
 * @param new_mask      The new network mask for the camera.
 * @param new_gw        The new gateway IP address for the camera.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::force_ip(const uint8_t *mac_addr, uint32_t new_ip, uint32_t new_mask, uint32_t new_gw)
{
    fcp_msg    msg;                        // Our message structure.
    uint16_t    host_sval;
    QHostAddress    dest_addr;
    MSG_ENTRY       msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Specify the IP address we're going to. We need to send this as a broadcast
    // in case the target is in a different sub-net.
    dest_addr.setAddress(QString::fromUtf8(GLOBAL_BROADCAST_ADDRESS));

    // Build up the read register message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = GIGE_FLAG_ALLOW_BROADCAST_BE | GIGE_FLAG_MUST_ACK_BE;
    host_sval = GIGE_MSG_FORCEIP_CMD;
    msg.header.req.command = qToBigEndian(host_sval);
    host_sval = GIGE_FORCEIP_REQ_B_LEN;
    msg.header.req.length = qToBigEndian(host_sval);

    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Fill in the data fields
    memcpy(&msg.data_b[ GIGE_FORCEIP_REQ_MAC_B ], mac_addr, 6);
    msg.data_l[ GIGE_FORCEIP_REQ_IP_L ] = new_ip;
    msg.data_l[ GIGE_FORCEIP_REQ_SUBNET_L ] = new_mask;
    msg.data_l[ GIGE_FORCEIP_REQ_GATEWAY_L ] = new_gw;

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, sizeof(msg.header) + GIGE_FORCEIP_REQ_B_LEN);

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    msg_entry.do_timeout = 0;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_read_list.clear();
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_write_value = 0;
    msg_entry.pending_write_lock = false;

    // Send our message and return the status.
    return send_msg(msg_entry);
}

/*!
 * @brief Return the last error string and code.
 *
 * @param err_msg   A textual description of the error code.
 * @return The error code.
 */
int FCPSocket::get_last_err(QString& err_msg)
{
    err_msg = last_errstr;
    return last_errcode;
}

/*!
 * @brief Return the address of the local address, or QHostAddress::Null if not available.
 */
QHostAddress FCPSocket::get_local_address() const
{
    return gige_socket->localAddress();
}

/*!
 * @brief Return the list of pending register address from a multiregister read.
 */
const FI_COMM::REG_ADDR_LIST& FCPSocket::get_pending_read_list()
{
    // Return our read list.
    return pending_read_list;
}

/*!
 * @brief Return the UDP port we're using for communications.
 */
int FCPSocket::get_port()
{
    // If we have a socket object, update our port value with
    // the latest version.
    if (gige_socket != nullptr)
        msg_port = (int)gige_socket->get_port();

    return msg_port;
}

/*!
 * @brief Return the [streamer object](@ref FSPSocket) so that the caller can have
 *        direct access to the streaming protocol interface.
 *
 * @return A pointer to the streaming object.
 */
FSPSocket *FCPSocket::get_streamer()
{
    // Do we already have a streamer?
    if (fsp_socket == nullptr)
    {
        // Get our FSP object running so that it'll be ready to go when needed.
        fsp_socket = new FSPSocket(this);
        if (fsp_socket == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate FSPSocket!";
            return nullptr;
        }

        // Initialize our streamer.
        fsp_socket->set_port((int)gige_socket->get_port() + FSP_PORT_OFFSET);
        fsp_socket->init_buffers();

        // Report any socket errors to the FCPSocket class.
        connect(fsp_socket, SIGNAL(error()), this, SLOT(fsp_error_event()));
    }

    // Return our streamer object.
    return fsp_socket;
}

/*!
 * @brief Pop a message off the front of our message list.
 *
 * @return Pointer to the next message, or nullptr if the list is empty.
 */
fcp_msg *FCPSocket::pop_msg()
{
    fcp_msg    *p_msg(nullptr);

    // Sanity check.
    if (!rcv_list.isEmpty())
        p_msg = rcv_list.takeFirst();

    // Return our message pointer.
    return p_msg;
}

/*!
 * @brief Process the ACK message. Really, we just save the received message in
 *        our list to be handled by someone who cares.
 *
 * @param msg       The received message.
 */
void FCPSocket::process_ack(fcp_msg& msg)
{
    fcp_msg    *p_msg;
    quint16     rcv_msg_id;
    quint16     rcv_cmd;
    QString     debug_msg;

    // Now do the validation on what we received.
    // Make sure the message and id are what we are looking for.
    rcv_msg_id = qFromBigEndian(msg.header.resp.ack_id);

    // Get the command from the response.
    rcv_cmd = qFromBigEndian(msg.header.resp.command);

    // Check to see if we're already past this message.
    if (rcv_msg_id < pending_msg_id)
    {
        QHostAddress    our_addr;
        our_addr.setAddress(qFromBigEndian(cam_ip_addr));

        // This can happen if the last message timed out but we do eventually get it. In this case just toss this
        // message since the client has already timed out.
        debug_msg = QString::fromUtf8("%1: Received msg_id (%2) is less than sent id (%3). cmd was 0x%4")
                    .arg(QString::fromUtf8(__PRETTY_FUNCTION__)).arg(rcv_msg_id).arg(pending_msg_id).arg(rcv_cmd, 0, 16);
        QLOG_INFO() << debug_msg << "IP:" << our_addr.toString();

        last_errcode = FI_COMM::FI_MSG_ID_SYNC_ERR;
        last_errstr.clear();
        return;
    }
    else if (rcv_msg_id != pending_msg_id)
    {
        // This should never happen, the previous case should catch all of them.
        debug_msg = QString::fromUtf8("%1: Received msg_id (%2) doesn't match last sent id (%3) cmd:%4")
                    .arg(QString::fromUtf8(__PRETTY_FUNCTION__)).arg(rcv_msg_id).arg(pending_msg_id).arg(rcv_cmd);
        QLOG_ERROR() << debug_msg;

        last_errcode = FI_COMM::FI_MSG_ID_SYNC_ERR;
        last_errstr = debug_msg;

        // We want to keep going. We got a bad message ID and we want to make
        // sure there are no more message behind it.
        return;
    }

    // Kill our timer, if we've got one going.
    if (msg_timer_id != 0)
    {
        killTimer(msg_timer_id);
        msg_timer_id = 0;
    }

    // Allocate a message structure.
    p_msg = new fcp_msg;
    if (p_msg == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate fcp_msg structure!";
        last_errcode = FI_COMM::FI_ENOMEM;
        return;
    }

    // Copy the message over to the new structure.
    memcpy(p_msg, &msg, sizeof(fcp_msg));

    // Add the message to the list.
    rcv_list.append(p_msg);

    // Happy with this message.
    last_errcode = FI_COMM::FI_SUCCESS;
}

/*!
 * @brief Process the discovery ACK message, parsing the data into a new FI_COMM::FI_CAMERA_INFO structure.
 *
 * @param msg       The received message.
 */
void FCPSocket::process_discovery_ack(fcp_msg& msg)
{
    fcp_msg    *p_msg;

    // We recieved a response to the broadcast from the camera.
    // Make sure that the camera has a valid IP address. We can't connect to it
    // if it doesn't.
    if (msg.data_l[ GIGE_DSC_ACK_IP_L ] == 0)
    {
        return;
    }

    // Allocate a message structure.
    p_msg = new fcp_msg;
    if (p_msg == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate fcp_msg structure!";
        last_errcode = FI_COMM::FI_ENOMEM;
        return;
    }

    // Copy the message over to the new structure.
    memcpy(p_msg, &msg, sizeof(fcp_msg));

    // Add the message to the list.
    rcv_list.append(p_msg);
}

/*!
 * @brief This method processes a received FMP message.
 *
 * @param msg       The FMP message we've received.
 */
void FCPSocket::process_fmp_msg(fcp_msg& msg)
{
    quint16     rcv_len;
    quint16     rcv_cmd;
    uint32_t    reg_addr;
    uint32_t    reg_value;
    QString     debug_msg;
    bool        pass_it_on;

    // Get the command from the response.
    rcv_cmd = qFromBigEndian(msg.header.req.command);

    // Make sure the command is appropriate for an asynchronous messsage.
    if (rcv_cmd != GIGE_MSG_EVENTDATA_CMD)
    {
        // If the receive command is '0', then this is a normal part of connecting to the camera.
        // We should only log the event if it is noteworthy.
        if (rcv_cmd)
        {
            QLOG_DEBUG() << QString::fromUtf8("%1: Unexpected receive command: 0x%2").arg(__PRETTY_FUNCTION__).arg(rcv_cmd, 0, 16);
        }
        return;
    }

    // Get the receive length. It needs to be the right size too.
    rcv_len = qFromBigEndian(msg.header.req.length);
    if (rcv_len != GIGE_EVENTDATA_B_LEN)
    {
        debug_msg = QString::fromUtf8("%1: Unexpected receive length: 0x%2").arg(__PRETTY_FUNCTION__).arg(rcv_len, 0, 16);
        QLOG_ERROR() << debug_msg;
        return;
    }

    // Get the address and value which need updating.
    reg_addr = qFromBigEndian(msg.fed.event_id);
    reg_value = qFromBigEndian(msg.fed.reg_val_l);

    // See if we want to intercept the message.
    switch (reg_addr)
    {
    case GIGE_REG_CCP:
    case GIGE_REG_FASTEC_CCP:
        // This is control information. It uses a special signal.
        pass_it_on = false;
        emit ctrl_msg_received(cam_ip_addr, reg_addr, reg_value);
        break;
    default:
        // In general, we want to pass the information on.
        pass_it_on = true;
        break;
    }

    // Tell anyone who wants to know.
    if (pass_it_on)
        emit unsol_msg_received(cam_ip_addr, reg_addr, reg_value);
}

/*!
 * @brief Once the communications class has handled a message, it calls this method to process the
 *        message queue. The old message is popped off the top of the queue and, if there is a pending
 *        message, it is sent to the network.
 */
void FCPSocket::process_msg_queue()
{
    MSG_ENTRY   msg_entry;

    // Pop the top of our queue.
    if (!msg_queue.isEmpty())
        msg_queue.takeFirst();

    // Do we have a queue of messages?
    while (!msg_queue.isEmpty())
    {
        // Kill our timer, if we've got one going.
        if (msg_timer_id != 0)
        {
            killTimer(msg_timer_id);
            msg_timer_id = 0;
        }

        // Grab the entry at the top of the queue. This leaves the entry
        // on the queue.
        msg_entry = msg_queue.first();

        // Update the "pending" information.
        update_pending_info(msg_entry);

        // There's no other messages in the queue, so go ahead and send
        // this one out.
        if (gige_socket->send_msg(msg_entry.msg_blob, msg_entry.dest_ip_addr) != FI_COMM::FI_SUCCESS)
        {
            // The gige_socket object has already emitted an error() signal, so just log it here.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "gige_socket->send_msg() to " << Qt::showbase << Qt::hex << msg_entry.dest_ip_addr << "failed!";

            // Pop the message off the queue.
            msg_queue.takeFirst();
            return;
        }

        // Start our time-out timer if requested.
        if (msg_entry.do_timeout)
        {
            msg_timer_id = startTimer(msg_entry.do_timeout);
            break;
        }
        else
        {
            // There's no timeout - so there's no expected message. Pop the queue.
            msg_queue.takeFirst();
        }
    }
}

/*!
 * @brief This method processes a received message. It performs simple
 *        sanity checks to make sure this is an "expected" message. Then
 *        it passes the message on to command specific parsing methods.
 *
 * @param msg       The FCP message we've received.
 */
void FCPSocket::process_received_msg(fcp_msg& msg)
{
    quint16     rcv_cmd;
    QString     debug_msg;

    // Get the command from the response.
    rcv_cmd = qFromBigEndian(msg.header.resp.command);

    // Process based on received command.
    switch (rcv_cmd)
    {
    case GIGE_MSG_DISCOVERY_ACK:
        process_discovery_ack(msg);
        break;
    case GIGE_MSG_FORCEIP_ACK:
    case GIGE_MSG_READREG_ACK:
    case GIGE_MSG_WRITEREG_ACK:
    case GIGE_MSG_READMEM_ACK:
    case GIGE_MSG_WRITEMEM_ACK:
        process_ack(msg);
        break;
    default:
        debug_msg = QString::fromUtf8("%1: Unexpected receive command: 0x%2").arg(__PRETTY_FUNCTION__).arg(rcv_cmd, 0, 16);
        QLOG_ERROR() << debug_msg;
        break;
    }
}

/*!
 * @brief Build up a FCP message to read a GIGE memory register.
 *
 * @param ip_addr       The IP address of the camera we want to talk with.
 * @param reg_addr      The FCP register address we're interested in.
 * @param reg_len       How many bytes of memory we want to read.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::read_mem(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_len)
{
    fcp_msg    msg;                        // Our message structure.
    uint16_t    host_sval;
    uint32_t    host_lval;
    QHostAddress    dest_addr;
    MSG_ENTRY       msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Specify the IP address we're going to.
    dest_addr.setAddress(qFromBigEndian(ip_addr));

    // Build up the read register message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    host_sval = GIGE_MSG_READMEM_CMD;
    msg.header.req.command = qToBigEndian(host_sval);
    host_sval = GIGE_READMEM_REQ_B_LEN;
    msg.header.req.length = qToBigEndian(host_sval);
    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Fill in the data fields.
    host_lval = reg_addr;
    msg.data_l[ GIGE_READMEM_REQ_ADDR_L ] = qToBigEndian(host_lval);
    if (reg_len > GIGE_READMEM_B_LEN)
        reg_len = GIGE_READMEM_B_LEN;
    host_sval = (uint16_t)reg_len;
    msg.data_w[ GIGE_READMEM_REQ_COUNT_W ] = qToBigEndian(host_sval);

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, sizeof(msg.header) + GIGE_READMEM_REQ_B_LEN);

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    msg_entry.do_timeout = DEFAULT_TIMEOUT;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = reg_addr;
    msg_entry.pending_read_list.clear();
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_write_lock = false;
    msg_entry.pending_write_value = 0;

    // Send our message and return the status.
    return send_msg(msg_entry);
}

/*!
 * @brief Build up a FCP message to read a list of GIGE registers.
 *
 * @param ip_addr       The IP address of the camera we want to talk with.
 * @param reg_list      The list of FCP register addresses we're interested in.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::read_multireg(uint32_t ip_addr, const FI_COMM::REG_ADDR_LIST& reg_list)
{
    int         i;
    fcp_msg    msg;                        // Our message structure.
    uint16_t    host_sval;
    QHostAddress    dest_addr;
    MSG_ENTRY       msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Specify the IP address we're going to.
    dest_addr.setAddress(qFromBigEndian(ip_addr));

    // Build up the read register message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    host_sval = GIGE_MSG_READREG_CMD;
    msg.header.req.command = qToBigEndian(host_sval);
    host_sval = (uint16_t)(GIGE_READREG_REQ_B_LEN * reg_list.count());
    msg.header.req.length = qToBigEndian(host_sval);

    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Fill in the data fields.
    for (i = 0; i < reg_list.count(); ++i)
    {
        uint32_t host_lval = reg_list.at(i);
        msg.data_l[ i ] = qToBigEndian(host_lval);
    }

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, (int)sizeof(msg.header) + (GIGE_READREG_REQ_B_LEN * reg_list.count()));

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    msg_entry.do_timeout = DEFAULT_TIMEOUT;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = FCP_COMM::MULTI_PENDING_REG_ADDR;
    msg_entry.pending_read_list.clear();
    msg_entry.pending_read_list = reg_list;
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_write_lock = false;
    msg_entry.pending_write_value = 0;

    // Send our message and return the status.
    return send_msg(msg_entry);
}

/*!
 * @brief Build up a FCP message to read a GIGE register.
 *
 * @param ip_addr       The IP address of the camera we want to talk with.
 * @param reg_addr      The FCP register address we're interested in.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::read_reg(uint32_t ip_addr, uint32_t reg_addr)
{
    fcp_msg    msg;                        // Our message structure.
    uint16_t    host_sval;
    uint32_t    host_lval;
    QHostAddress    dest_addr;
    MSG_ENTRY       msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Specify the IP address we're going to.
    dest_addr.setAddress(qFromBigEndian(ip_addr));

    // Build up the read register message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = GIGE_FLAG_MUST_ACK_BE;
    host_sval = GIGE_MSG_READREG_CMD;
    msg.header.req.command = qToBigEndian(host_sval);
    host_sval = GIGE_READREG_REQ_B_LEN;
    msg.header.req.length = qToBigEndian(host_sval);

    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Fill in the data fields.
    host_lval = reg_addr;
    msg.data_l[0] = qToBigEndian(host_lval);

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, sizeof(msg.header) + GIGE_READREG_REQ_B_LEN);

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    msg_entry.do_timeout = DEFAULT_TIMEOUT;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = reg_addr;
    msg_entry.pending_read_list.clear();
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_write_lock = false;
    msg_entry.pending_write_value = 0;

    // Send our message and return the status.
    return send_msg(msg_entry);
}

/*!
 * @brief   Retry the last write attempt. The data and lock requirements are
 *          picked up from the last attempt and so are not passed in.
 *
 * @param ip_addr   Address of the camera we're communicating with.
 * @param reg_addr  Address of the register we're writing.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::retry_write(uint32_t ip_addr, uint32_t reg_addr, uint32_t write_value, bool write_lock)
{
    return write_reg(ip_addr, reg_addr, write_value, true, write_lock);
}

/*!
 * @brief Transmit a FCP Discover message via the adapter specified by "adapter_name".
 *        If the optional ip_addr is specified, then a unicast message is sent to that
 *        target. Otherwise a broadcast message is sent.
 *
 * @param adapter_name  Which network adapter to use.
 * @param ip_addr       [Optional] A specific IP address to send the message to.
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::send_discovery_msg(QString& adapter_name, uint32_t ip_addr)
{
    fcp_msg    msg;                        // Our message structure.
    quint16     host_sval;
    QHostAddress            dest_addr;
    QNetworkAddressEntry    addr_entry;
    MSG_ENTRY               msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // See if we can find our adapter information.
    if (!find_adapter(adapter_name, addr_entry))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << adapter_name << "was not found!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Did the user specify a target address/port?
    if ((ip_addr == 0) || (ip_addr == FCP_COMM::GLOBAL_BROADCAST_IPADDR))
    {
        // No specific target, so send a subnet broadcast message.
        dest_addr.setAddress(addr_entry.broadcast().toIPv4Address());
    }
    else
    {
        // The user has a specific target in mind.
        dest_addr.setAddress(ip_addr);
    }

    // Set our address for the socket.
    set_local_address(addr_entry.ip());

    // Build up the discovery message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = GIGE_FLAG_ALLOW_BROADCAST_BE | GIGE_FLAG_MUST_ACK_BE;
    host_sval = GIGE_MSG_DISCOVERY_CMD;
    msg.header.req.command = qToBigEndian(host_sval);
    msg.header.req.length = 0;          // No data portion on this command.

    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, sizeof(msg.header));

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    msg_entry.do_timeout = 0;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_read_list.clear();
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_write_value = 0;
    msg_entry.pending_write_lock = false;

    // Send our message and return the status. For the discovery message,
    // there's no data involved - just send the header.
    return send_msg(msg_entry);
}

int FCPSocket::send_discovery_msg(uint32_t adapter_ip, uint32_t ip_addr)
{
    fcp_msg    msg;                        // Our message structure.
    quint16     host_sval;
    QHostAddress            dest_addr;
    QNetworkAddressEntry    addr_entry;
    MSG_ENTRY               msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // See if we can find our adapter information.
    if (!find_adapter(adapter_ip, addr_entry))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Interface adapter was not found!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Did the user specify a target address/port?
    if ((ip_addr == 0) || (ip_addr == FCP_COMM::GLOBAL_BROADCAST_IPADDR))
    {
        // No specific target, so send a subnet broadcast message.
        dest_addr.setAddress(addr_entry.broadcast().toIPv4Address());
    }
    else
    {
        // The user has a specific target in mind.
        dest_addr.setAddress(ip_addr);
    }

    // Set our address for the socket.
    set_local_address(addr_entry.ip());

    // Build up the discovery message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = GIGE_FLAG_ALLOW_BROADCAST_BE | GIGE_FLAG_MUST_ACK_BE;
    host_sval = GIGE_MSG_DISCOVERY_CMD;
    msg.header.req.command = qToBigEndian(host_sval);
    msg.header.req.length = 0;          // No data portion on this command.

    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, sizeof(msg.header));

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    msg_entry.do_timeout = 0;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_read_list.clear();
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_write_value = 0;
    msg_entry.pending_write_lock = false;

    // Send our message and return the status. For the discovery message,
    // there's no data involved - just send the header.
    return send_msg(msg_entry);
}

/*!
 * @brief Helper function to send a message via the gige_socket object.
 *
 * @param msg_entry     The structure containing all the information for the GIGE message to be sent.
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::send_msg(const MSG_ENTRY& msg_entry)
{
    int status;

    // Reset our error information.
    last_errcode = FI_COMM::FI_SUCCESS;
    last_errstr.clear();

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Is this a priority message?
    if (msg_entry.do_timeout == 0)
    {
        // Send the message straight away.
        if ((status = gige_socket->send_msg(msg_entry.msg_blob, msg_entry.dest_ip_addr)) != FI_COMM::FI_SUCCESS)
        {
            // The gige_socket object has already emitted an error() signal, so just log it here.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "gige_socket->send_msg() to " << Qt::showbase << Qt::hex << msg_entry.dest_ip_addr << "failed!";
            return status;
        }

        // All is well.
        return FI_COMM::FI_SUCCESS;
    }

    // Append this message to the queue.
    msg_queue.append(msg_entry);

    // Do we have a queue of messages?
    if (msg_queue.count() == 1)
    {
        // Kill our timer, if we've got one going.
        if (msg_timer_id != 0)
        {
            killTimer(msg_timer_id);
            msg_timer_id = 0;
        }

        // Update the pending information.
        update_pending_info(msg_entry);

        // There's no other messages in the queue, so go ahead and send
        // this one out.
        if ((status = gige_socket->send_msg(msg_entry.msg_blob, msg_entry.dest_ip_addr)) != FI_COMM::FI_SUCCESS)
        {
            // The gige_socket object has already emitted an error() signal, so just log it here.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "gige_socket->send_msg() to " << Qt::showbase << Qt::hex << msg_entry.dest_ip_addr << "failed!";

            // Pop it back off the queue.
            msg_queue.takeFirst();
            return status;
        }

        // Start our time-out timer if requested.
        if (msg_entry.do_timeout)
        {
            msg_timer_id = startTimer(msg_entry.do_timeout);
        }
        else
        {
            // Kill our timer, if we've got one going.
            if (msg_timer_id != 0)
            {
                killTimer(msg_timer_id);
                msg_timer_id = 0;
            }

            // No timeout so no expected message. Pop the queue.
            msg_queue.takeFirst();
        }
    }

    // All is well.
    return FI_COMM::FI_SUCCESS;
}

/*!
 * @brief Set the local address for our socket.
 * @param address   The IP address of our end of the socket.
 */
void FCPSocket::set_local_address(const QHostAddress& address)
{
    gige_socket->setLocalAddress(address);
}

/*!
 * @brief Called to notify the socket when the camera is going on- or off-line.
 * @param is_offline    True for off-line, false for on-line.
 */
void FCPSocket::set_offline(bool is_offline)
{
    offline = is_offline;
}

/*!
 * @brief Specify the UDP port we want to use for communications.
 *
 * @param port      The port we're going to use.
 */
void FCPSocket::set_port(int port)
{
    // Save the port for later.
    msg_port = port;
}

/*!
 * @brief Start the communications object working.
 */
void FCPSocket::start()
{
    // Create our GiGE socket object which does all the UDP packet communications.
    gige_socket = new GiGESocket(nullptr, (uint32_t)msg_port);
    if (gige_socket == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Can't allocate a new GIGESocket!";
        return;
    }

    /*
     * Connect up the slots/signals so everyone can communicate.
     */

    // Report any socket errors to the FCPSocket class.
    connect(gige_socket, SIGNAL(error()), this, SLOT(fcp_error_event()));

    // Have the socket object notify us when it's up and running.
    connect(gige_socket, SIGNAL(started()), this, SLOT(fcp_socket_ready()));

    // Have the socket tell us when a data packet (datagram) is available.
    connect(gige_socket, SIGNAL(readyRead()), this, SLOT(fcp_read_pending_msg()), Qt::DirectConnection);

    // For shutdown, once the GiGE socket object is done have it invoke the delayed
    // cleanup method on exit.
    connect(gige_socket, SIGNAL(disconnected()), gige_socket, SLOT(deleteLater()));

    // Get the socket started. It'll either emit an error signal or, if everything
    // goes OK, it'll emit a "started" signal and we'll know we're initialized.
    gige_socket->start();
}

/*!
 * @brief Create the interface to receive unsolicited (i.e. FMP) messages from the camera.
 * @return A status code indicating success or failure.
 */
int FCPSocket::start_unsolicited()
{
    // If we've already got an unsolicited socket class, then we're done.
    if (fmp_socket != nullptr)
        return GEV_STATUS_SUCCESS;

    // Make sure we've got our GiGE socket.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return GEV_STATUS_ERROR;
    }

    // Create our GiGE socket object for asynchronous, unsolicited UDP packet communications.
    fmp_socket = new GiGESocket(nullptr, (uint32_t)get_port() + FMP_PORT_OFFSET);
    if (fmp_socket == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Can't allocate a new GIGESocket!";
        return GEV_STATUS_ERROR;
    }

    /*
     * Connect up the slots/signals so everyone can communicate.
     */

    // Report any socket errors to the FCPSocket class.
    connect(fmp_socket, SIGNAL(error()), this, SLOT(fmp_error_event()));

    // Have the socket object notify us when it's up and running.
    connect(fmp_socket, SIGNAL(started()), this, SLOT(fmp_socket_ready()));

    // Have the socket tell us when a data packet (datagram) is available.
    connect(fmp_socket, SIGNAL(readyRead()), this, SLOT(fmp_read_pending_msg()), Qt::DirectConnection);

    // For shutdown, once the GiGE socket object is done have it invoke the delayed
    // cleanup method on exit.
    connect(fmp_socket, SIGNAL(disconnected()), fmp_socket, SLOT(deleteLater()));

    // Get the socket started. It'll either emit an error signal or, if everything
    // goes OK, it'll emit a "started" signal and we'll know we're initialized.
    fmp_socket->start();

    // Assume success at this point.
    return GEV_STATUS_SUCCESS;
}

/*!
 * @brief Have our communications object shutdown.
 */
void FCPSocket::stop()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // We are definitely offline!
    offline = true;

    // Check for timer events.
    if (msg_timer_id != 0)
    {
        // No need for this timer.
        killTimer(msg_timer_id);
        msg_timer_id = 0;
    }

    // Shutdown our sockets.
    if (gige_socket != nullptr)
        gige_socket->quit();
    if (fmp_socket != nullptr)
        fmp_socket->quit();

    // Do we have a streamer?
    if (fsp_socket != nullptr)
        fsp_socket->quit();

    // Clear our message queue.
    msg_queue.clear();
}

/*!
 * @brief Close down the unsolicited message handling interface.
 */
void FCPSocket::stop_unsolicited()
{
    // If we don't have a message object, then we're done.
    if (fmp_socket == nullptr)
        return;

    // Tell the camera that we're not really interested in these messages anymore.
    write_reg(cam_ip_addr, GIGE_REG_MCP, 0, false);
    write_reg(cam_ip_addr, GIGE_REG_MCDA, 0, false);

    // Disconnect our signal/slots.
    disconnect(fmp_socket, SIGNAL(error()), this, SLOT(fmp_error_event()));
    disconnect(fmp_socket, SIGNAL(started()), this, SLOT(fmp_socket_ready()));
    disconnect(fmp_socket, SIGNAL(readyRead()), this, SLOT(fmp_read_pending_msg()));

    // Tell the message object that we're done.
    fmp_socket->quit();
    fmp_socket = nullptr;
}

/*!
 * @brief Update our pending_xxx values so we know what we're trying to accomplish in the
 *        case of a timeout or other error.
 * @param msg_entry     Structure describing the GIGE message operation.
 */
void FCPSocket::update_pending_info(const MSG_ENTRY& msg_entry)
{
    pending_read_addr = msg_entry.pending_read_addr;
    pending_read_list.clear();
    pending_read_list = msg_entry.pending_read_list;
    pending_write_addr = msg_entry.pending_write_addr;
    pending_write_lock = msg_entry.pending_write_lock;
    pending_write_value = msg_entry.pending_write_value;
    pending_msg_id = (uint16_t)msg_entry.msg_id;
}

/*!
 * @brief Build up a FCP message to write a GIGE memory register.
 *
 * @param ip_addr       The IP address of the camera we want to talk with.
 * @param reg_addr      The FCP register address we're interested in.
 * @param buffer        A QByteArray object containing the data we are writing.
 * @param need_ack      True if the caller wants a GIGE acknowledge message from the camera, false otherwise.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::write_mem(uint32_t ip_addr, uint32_t reg_addr, const QByteArray& buffer, bool need_ack)
{
    int         buffer_len;
    fcp_msg    msg;                        // Our message structure.
    uint16_t    msg_length;
    uint16_t    host_sval;
    QHostAddress    dest_addr;
    MSG_ENTRY       msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Specify the IP address we're going to.
    dest_addr.setAddress(qFromBigEndian(ip_addr));

    // Build up the read register message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = need_ack ? GIGE_FLAG_MUST_ACK_BE : 0;
    host_sval = GIGE_MSG_WRITEMEM_CMD;
    msg.header.req.command = qToBigEndian(host_sval);

    // Figure out our buffer length.
    buffer_len = buffer.size();
    if (buffer_len > GIGE_WRITEMEM_B_LEN)
        buffer_len = GIGE_WRITEMEM_B_LEN;

    msg_length = GIGE_WRITEMEM_ADDR_LEN + (uint16_t)buffer_len;
    msg.header.req.length = qToBigEndian(msg_length);

    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Fill in our data.
    msg.data_l[0] = qToBigEndian(reg_addr);
    memcpy(&msg.data_l[1], buffer.constData(), (size_t)buffer_len);

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, sizeof(msg.header) + msg_length);

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    if (need_ack)
        msg_entry.do_timeout = LONG_MSG_TIMEOUT;
    else
        msg_entry.do_timeout = 0;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_read_list.clear();
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = reg_addr;
    msg_entry.pending_write_value = 0;
    msg_entry.pending_write_lock = false;

    // Send our message and return the status.
    return send_msg(msg_entry);
}

/*!
 * @brief Build up a FCP message to write a register value.
 *
 * @param ip_addr       The IP address of the camera we want to talk with.
 * @param reg_addr      The FCP register address we're interested in.
 * @param reg_value     The value we want to write.
 * @param need_ack      True if the caller wants a GIGE acknowledge message from the camera, false otherwise.
 * @param need_lock     TRUE if we need to lock access, FALSE otherwise.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FCPSocket::write_reg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value, bool need_ack, bool need_lock)
{
    fcp_msg    msg;                        // Our message structure.
    uint16_t    msg_length;
    uint16_t    host_sval;
    QHostAddress    dest_addr;
    MSG_ENTRY       msg_entry;

    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Specify the IP address we're going to.
    dest_addr.setAddress(qFromBigEndian(ip_addr));

    // Build up the read register message.
    memset(&msg, 0, sizeof(msg));
    msg.header.req.code = GIGE_REQ_CODE;
    msg.header.req.flag = need_ack ? GIGE_FLAG_MUST_ACK_BE : 0;
    host_sval = GIGE_MSG_WRITEREG_CMD;
    msg.header.req.command = qToBigEndian(host_sval);

    // The length depends on if we're writing one register, or if we
    // need to get the lock as well as write the register.
    if (need_lock)
        msg_length = 3 * GIGE_WRITEREG_REQ_B_LEN;
    else
        msg_length = GIGE_WRITEREG_REQ_B_LEN;
    msg.header.req.length = qToBigEndian(msg_length);

    msg_id++;                           // Increment before using.
    host_sval = msg_id;
    msg.header.req.req_id = qToBigEndian(host_sval);

    // Fill in the data fields based on if we need the lock or not.
    if (need_lock)
    {
        // First the lock, then the acutal register write, then the unlock.
        msg.data_l[0] = qToBigEndian(quint32(GIGE_REG_FASTEC_CCP));
        msg.data_l[1] = qToBigEndian(quint32(GIGE_FASTEC_CCP_CA_BE));
        msg.data_l[2] = qToBigEndian(reg_addr);
        msg.data_l[3] = qToBigEndian(reg_value);
        msg.data_l[4] = qToBigEndian(quint32(GIGE_REG_FASTEC_CCP));
        msg.data_l[5] = 0;
    }
    else
    {
        msg.data_l[0] = qToBigEndian(reg_addr);
        msg.data_l[1] = qToBigEndian(reg_value);
    }

    // Make a "blob" of data out of our message.
    msg_entry.msg_blob.append((const char *)&msg, sizeof(msg.header) + msg_length);

    // Initialize the rest of our entry fields.
    msg_entry.dest_ip_addr = dest_addr;
    if (need_ack)
        msg_entry.do_timeout = DEFAULT_TIMEOUT;
    else
        msg_entry.do_timeout = 0;

    // Mirror all the "pending" information.
    msg_entry.pending_read_addr = FCP_COMM::NO_PENDING_REG_ADDR;
    msg_entry.pending_read_list.clear();
    msg_entry.msg_id = (int16_t)msg_id;
    msg_entry.pending_write_addr = reg_addr;
    msg_entry.pending_write_value = reg_value;
    msg_entry.pending_write_lock = need_lock;

    // Send our message and return the status. If we need an ACK, then we probably
    // want to timeout the message - no ACK, no timeout.
    return send_msg(msg_entry);
}

//--------------------------------------------------------------------
// Event Handler(s)
//--------------------------------------------------------------------

/*!
 * @brief Event routine which is invoked when one of our timers expires.
 * @param event     Pointer to the event class.
 */
void FCPSocket::timerEvent(QTimerEvent *event)
{
    // Is this our timer event?
    if (event->timerId() != msg_timer_id)
        return;

    // Sanity check.
    if (msg_timer_id == 0)
        return;

    // OK...we really got a timeout. Don't need the timer anymore.
    killTimer(msg_timer_id);
    msg_timer_id = 0;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << qToBigEndian(cam_ip_addr) << pending_read_addr << pending_write_addr;
    if (pending_read_addr == FCP_COMM::MULTI_PENDING_REG_ADDR)
    {
        int i;
        QString strmsg;

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "msg_id:" << Qt::showbase << Qt::hex << pending_msg_id;
        for (i = 0; i < pending_read_list.count(); i++)
        {
            strmsg.append(QString::number(pending_read_list.at(i), 16));
            strmsg.append(", ");
        }

        // Remove the last ", " before displaying.
        strmsg.truncate(strmsg.length() - 2);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "List [" << pending_read_list.count() << "]:" << strmsg;
    }

    // Did this camera get dropped?
    if (offline)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera is offline - purging message queue.";

        // No camera, so purge any pending messages.
        msg_queue.clear();
    }
    else
    {
        // Got a timeout!
        emit msg_timeout();
    }
}

//--------------------------------------------------------------------
// Slots
//--------------------------------------------------------------------

/*!
 * @brief Invoked when an error has happened. Grab the error information
 *        into our local variables, last_errcode and last_errstr.
 */
void FCPSocket::fcp_error_event()
{
    CameraMDI *camera_ptr;

    // If we have a socket object, grab the error information.
    if (gige_socket != nullptr)
        last_errcode = gige_socket->get_last_err(last_errstr);

    QString err_str = QString::fromUtf8("Error [%1] = %2").arg(last_errcode).arg(last_errstr);
    QLOG_INFO() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << qToBigEndian(cam_ip_addr) << err_str;

    // Is there an active camera?
    camera_ptr = MainWindow::GetInstance()->active_mdi_child();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No active child";
        return;
    }

    // Pass it on.
    emit fcp_error();
}

/*!
 * @brief Read and do preliminary processing on any pending datagrams from the gige_socket object.
 */
void FCPSocket::fcp_read_pending_msg()
{
    // Sanity check.
    if (gige_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No gige_socket!";
        return;
    }

    // Read in any and all pending datagrams.
    while (gige_socket->hasPendingDatagrams())
    {
        fcp_msg    cur_msg_received;

        // Clear our message structure.
        memset(&cur_msg_received, 0, sizeof(cur_msg_received));

        // Read our waiting message.
        qint64 read_len = gige_socket->readDatagram((char *)&cur_msg_received, sizeof(cur_msg_received));
        if (read_len <= 0)
        {
            QString err_msg;

            last_errstr = gige_socket->errorString();
            last_errcode = FI_COMM::FI_SOCKET_ERR;

            err_msg = QString::fromUtf8("%1: readDatagram returned err(%2), errno=%3(%4)").arg(__PRETTY_FUNCTION__, last_errstr).arg(errno).arg(strerror(errno));
            QLOG_ERROR() << err_msg;
            return;
        }

        // Process the message.
        process_received_msg(cur_msg_received);
    }

    // Let any listeners know that we got a message.
    if (last_errcode == FI_COMM::FI_SUCCESS)
        emit msg_received();
}

/*!
 * @brief Invoked when the socket has finished initializing.
 */
void FCPSocket::fcp_socket_ready()
{
    // We are initialized and ready for message processing.
    initialized = true;
    offline = false;
}

/*!
 * @brief Invoked when an error has happened. Grab the error information
 *        into our local variables, last_errcode and last_errstr.
 */
void FCPSocket::fmp_error_event()
{
    CameraMDI *camera_ptr;

    // If we have a socket object, grab the error information.
    if (fmp_socket != nullptr)
        last_errcode = fmp_socket->get_last_err(last_errstr);

    QString err_str = QString::fromUtf8("Error [%1] = %2").arg(last_errcode).arg(last_errstr);
    QLOG_INFO() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << qToBigEndian(cam_ip_addr) << err_str;

    // Is there an active camera?
    camera_ptr = MainWindow::GetInstance()->active_mdi_child();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No active child";
        return;
    }

    // Pass it on.
    emit fmp_error();
}

/*!
 * @brief Read and do preliminary processing on any pending datagrams from the fmp_socket object.
 */
void FCPSocket::fmp_read_pending_msg()
{
    // Sanity check.
    if (fmp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No fmp_socket!";
        return;
    }

    // Read in any and all pending datagrams.
    while (fmp_socket->hasPendingDatagrams())
    {
        fcp_msg    cur_msg_received;

        // Clear our message structure.
        memset(&cur_msg_received, 0, sizeof(cur_msg_received));

        // Read our waiting message.
        qint64 read_len = fmp_socket->readDatagram((char *)&cur_msg_received, sizeof(cur_msg_received));
        if (read_len > 0)
        {
            // Process the message.
            process_fmp_msg(cur_msg_received);
        }
        else if (errno != 0)
        {
            QString err_msg;

            last_errstr = fmp_socket->errorString();
            last_errcode = FI_COMM::FI_SOCKET_ERR;

            err_msg = QString::fromUtf8("%1: readDatagram returned err(%2), errno=%3(%4)").arg(__PRETTY_FUNCTION__, last_errstr).arg(errno).arg(strerror(errno));
            QLOG_ERROR() << err_msg;
        }
    }
}

/*!
 * @brief Invoked when the socket has finished initializing.
 */
void FCPSocket::fmp_socket_ready()
{
    char    buffer[ DUMMY_PACKET_LEN ];
    qint64      status;
    QHostAddress    cam_addr;

    // Get our local address.
    cam_addr = get_local_address();
    if (cam_addr == QHostAddress::Null)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "What?! We don't know our address?!";
        return;
    }

    // Our socket is ready, now we need to tell the camera we're interested in
    // unsolicited messages.
    write_reg(cam_ip_addr, GIGE_REG_MCDA, cam_addr.toIPv4Address(), false);

    // Send a fake message over the FMP port so that the connection looks like it originated at the
    // client. We've discovered this is a way to make it through firewalls.

    // Address the command to the camera. Note that the setAddress method requires
    // the IP address to be in host order, which might be different than network order(BigEndian).
    cam_addr.setAddress(qFromBigEndian(cam_ip_addr));

    // Send our fake packet of zeros.
    memset(buffer, 0, DUMMY_PACKET_LEN);
    status = fmp_socket->writeDatagram(buffer, DUMMY_PACKET_LEN, cam_addr, (uint16_t)fmp_socket->get_port());
    if (status < 0)
    {
        // Log this as a warning - things may still work fine.
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to send fake message. Err=" << status << fmp_socket->errorString();
    }

    // Tell the camera the port to use.
    write_reg(cam_ip_addr, GIGE_REG_MCP, fmp_socket->get_port(), false);

    // Let the world know.
    emit unsol_socket_ready();
}

/*!
 * @brief Invoked when an error has happened. Grab the error information
 *        into our local variables, last_errcode and last_errstr.
 */
void FCPSocket::fsp_error_event()
{
    // If we have a socket object, grab the error information.
    if (fsp_socket != nullptr)
        last_errcode = fsp_socket->get_last_err(last_errstr);

    QString err_str = QString::fromUtf8("Error [%1] = %2").arg(last_errcode).arg(last_errstr);
    QLOG_INFO() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << qToBigEndian(cam_ip_addr) << err_str;

    // Pass it on.
    emit fsp_error();
}

