/*!
 * @file fcp_socket.h
 * @brief This header file declares the FCPSocket class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FCP_SOCKET_H
#define FCP_SOCKET_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QObject>
#include <QList>
#include <QtNetwork>
#include <QNetworkInterface>

#include "fi_comm.h"
#include "gige_msg.h"

namespace FCP_COMM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The default port for control portocol communications.
const uint32_t  DEFAULT_FCP_PORT           = 5824;

const int   FCP_PORT_OFFSET            = 0;        //!< Offset from the base port for FCP message traffic.
const int   FSP_PORT_OFFSET            = 1;        //!< Offset from the base port for FSP data.
const int   FMP_PORT_OFFSET            = 2;        //!< Offset from the base port for FMP message traffic.

//! Define a constant for the global broadcast IP address.
const uint32_t  GLOBAL_BROADCAST_IPADDR     = 0xFFFFFFFF;

//! Define a constant for no pending register address.
const uint32_t  NO_PENDING_REG_ADDR         = ((uint32_t) -1);

//! Define a constant for multiple register pending.
const uint32_t  MULTI_PENDING_REG_ADDR      = ((uint32_t) -2);

//! Define a message type.
typedef struct MSG_ENTRY
{
    QByteArray      msg_blob;
    QHostAddress    dest_ip_addr;
    int             do_timeout;

    // Mirror the pending information so we can restore when the queue is popped.
    uint32_t        pending_read_addr;
    FI_COMM::REG_ADDR_LIST  pending_read_list;

    uint32_t        pending_write_addr;
    uint32_t        pending_write_value;
    bool            pending_write_lock;
    qint16          msg_id;

} MSG_ENTRY;

//! A type definition for queuing up network messages.
typedef QList< MSG_ENTRY >       MSG_QUEUE;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class GiGESocket;
class FSPSocket;

//--------------------------------------------------------------------
// Class definition for FCPSocket
//--------------------------------------------------------------------

class FCPSocket : public QObject
{
    Q_OBJECT

public:
    FCPSocket(uint32_t my_ip_addr = FCP_COMM::GLOBAL_BROADCAST_IPADDR, uint32_t default_port = FCP_COMM::DEFAULT_FCP_PORT);
    ~FCPSocket();

    bool    is_initialized()          { return initialized; }
    bool    is_offline()              { return offline; }
    void    set_offline(bool is_offline);

    void    start();
    void    stop();

    uint32_t get_cam_address()        { return cam_ip_addr; }

    int     get_port();
    void    set_port(int port);

    int     get_last_err(QString& err_msg);
    void    set_local_address(const QHostAddress& address);
    QHostAddress get_local_address() const;

    int     get_msg_count()       { return rcv_list.size(); }
    fcp_msg    *pop_msg();

    int     send_discovery_msg(QString& adapter_name, uint32_t ip_addr);
    int     send_discovery_msg(uint32_t adapter_ip, uint32_t ip_addr);

    int     read_reg(uint32_t ip_addr, uint32_t reg_addr);
    int     read_multireg(uint32_t ip_addr, const FI_COMM::REG_ADDR_LIST& reg_list);
    int     read_mem(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_len);
    int     retry_write(uint32_t ip_addr, uint32_t reg_addr, uint32_t write_value, bool write_lock);

    int     write_mem(uint32_t ip_addr, uint32_t reg_addr, const QByteArray& buffer, bool need_ack = true);
    int     write_reg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value, bool need_ack = true, bool need_lock = true);

    int     force_ip(const uint8_t *mac_addr, uint32_t new_ip, uint32_t new_mask, uint32_t new_gw);

    uint32_t    get_pending_read_register()   { uint32_t last_reg = pending_read_addr; pending_read_addr = FCP_COMM::NO_PENDING_REG_ADDR; return last_reg; }
    const FI_COMM::REG_ADDR_LIST& get_pending_read_list();
    void        clear_pending_read_list();
    uint32_t    get_pending_write_register()  { uint32_t last_reg = pending_write_addr; pending_write_addr = FCP_COMM::NO_PENDING_REG_ADDR; return last_reg; }
    uint32_t    get_pending_write_value()     { return pending_write_value; }
    bool        get_pending_write_lock()      { return pending_write_lock; }

    void    process_msg_queue();
    int     start_unsolicited();
    void    stop_unsolicited();

    FSPSocket  *get_streamer();

signals:
    // FCP related signals.
    void    msg_received();
    void    msg_timeout();

    // FMP related signals.
    void    ctrl_msg_received(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    void    unsol_socket_ready();
    void    unsol_msg_received(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);

    void    fsp_error();
    void    fcp_error();
    void    fmp_error();

protected:
    virtual void    timerEvent(QTimerEvent  *event);

private slots:
    void    fcp_error_event();
    void    fcp_read_pending_msg();
    void    fcp_socket_ready();

    void    fmp_error_event();
    void    fmp_read_pending_msg();
    void    fmp_socket_ready();

    void    fsp_error_event();

private:
    bool    find_adapter(const QString& adapter_name, QNetworkAddressEntry& addr_entry);
    bool    find_adapter(uint32_t adapter_ip, QNetworkAddressEntry& addr_entry);

    void    process_ack(fcp_msg& msg);
    void    process_discovery_ack(fcp_msg& msg);
    void    process_fmp_msg(fcp_msg& msg);
    void    process_received_msg(fcp_msg& msg);
    int     send_msg(const FCP_COMM::MSG_ENTRY& msg_entry);
    void    update_pending_info(const FCP_COMM::MSG_ENTRY& msg_entry);

private:
    GiGESocket      *gige_socket;       //!< The actual socket class for the GiGE FCP messages.
    GiGESocket      *fmp_socket;        //!< The actual socket class for the GiGE FMP messages.
    FSPSocket       *fsp_socket;        //!< The FSP streamer class.

    uint32_t        cam_ip_addr;        //!< The IP address we're using.
    bool            initialized;        //!< Has the socket finished it's initialization sequence (successful or not)?
    bool            offline;            //!< True if the camera isn't talking (offline), false otherwise.
    QList<fcp_msg *>   rcv_list;        //!< List of received messages.

    uint32_t        pending_read_addr;  //!< Last read register address.
    FI_COMM::REG_ADDR_LIST  pending_read_list;  //!< For multiregsiter reads, this is the last list of addresses read.

    uint32_t        pending_write_addr; //!< Last write register address.
    uint32_t        pending_write_value;//!< Last write value...for retry attempts.
    bool            pending_write_lock; //!< Last write lock requirement...for retry attempts.

    FCP_COMM::MSG_QUEUE    msg_queue;  //!< List of messages waiting to be sent.

    int             msg_port;           //!< Port we want to use for communicating with the camera.
    quint16         msg_id;             //!< FCP message ID. Keep a class variable so we know what the next message ID should be.
    int             msg_timer_id;       //!< Timer to catch a timeout on a message.
    quint16         pending_msg_id;     //!< The message ID we're actually waiting for a response on.

    QString     last_errstr;    //!< last message which occurred
    int         last_errcode;   //!< last error code which occurred
};

#endif // FCP_SOCKET_H
