/*!
 * @file fi_camreg_group.cpp
 * @brief This provides the implementation for camera register group I/O support.
 *
 * Copyright 2023 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QList>

#include "fi_comm.h"
#include "fi_camreg_group.h"
#include "logger.h"

namespace CAMERA_REG_GROUP
{

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Log messages for testing purposes.  Set to 1 to have them compiled for output.
#define LOG_GRPINFO_TRACING            0
#define LOG_GRPDEL_TRACING             0

}   // End of namespace CAMERA_REG_GROUP.

using namespace CAMERA_REG_GROUP;

//! Keep a constant of when a sanity check on FiCamRegGroup fails to create one in the factory.
const QString FiCamRegGroup::CAMERA_GRP_READ_FAILURE = QStringLiteral("[CHECK FAILED] Unable to create a camera group read object!");

//! Keep a constant used to identify camera group log messages.
const QString FiCamRegGroup::CAMERA_GRP_LOG_NAME = QStringLiteral("[GIGE_GROUP]");

/*!
 * @brief The FiCamRegGroup manages a collection of Camera GIGE registers for I/O
 * for a connected camera.
 */
FiCamRegGroup::FiCamRegGroup()
{
#if LOG_GRPINFO_TRACING
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
#endif
};

/*!
 * @brief The FiCamRegGroup destructor.
 */
FiCamRegGroup::~FiCamRegGroup()
{
#if LOG_GRPINFO_TRACING
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "camRegGroupSigStr:" << camRegGroupSigStr;
#endif
};

/*!
 * @brief Factory method to create a group of registers for camera I/O.
 */
QSharedPointer<FiCamRegGroup> FiCamRegGroup::createCamRegGroup()
{
    // This is a very simple factory.
    QSharedPointer<FiCamRegGroup> camRegGroup = QSharedPointer<FiCamRegGroup>(new FiCamRegGroup());

    camRegGroup->lastRegAddr = 0;   // There is no last register as this group is empty.

#if LOG_GRPINFO_TRACING
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[FACTORY]";
#endif

    return camRegGroup;
}

/*!
 * @brief Adds a GIGE register to the group collection.
 * Assumes that lastReg has not been set.
 *
 * @param[in] regAddr   The GIGE register to add to the collection.
 *
 * @returns True if the register was added, false if this group collection is frozen (the lastReg was called),
 * already exists in the group or is zero.
 */
bool FiCamRegGroup::addReg(uint32_t regAddr)
{
    int count = camRegGroup.count();

    // If the last register address was added already cannot add anymore.
    if( lastRegAddr != 0 )
        return false;

    if( regAddr == 0 )
        return false;

    // Make sure this register isn't in the group already.
    for(int x=0; x < count; x++ )
    {
        if( camRegGroup.at(x) == regAddr )
        {
            return false;
        }
    }
    // Increase our count to what we want it to be.
    count++;

    // Add the register to the group.
    camRegGroup.append(regAddr);

    // Added a register to the group add to our register "signature" string.
    QString regAddrSigStr = QString("%1").arg(regAddr, 2, 16, QChar('0'));
    camRegGroupSigStr += regAddrSigStr;

    // Return based on the count equaling what it should be after the append.
    return (camRegGroup.count() == count);
}

/*!
 * @brief Adds a GIGE register to the group collection.
 *
 * @param[in] regAddr   The GIGE register to delete from the collection.
 *
 * @returns True if the register was removed, false if this register wasn't in the group.
 */
bool FiCamRegGroup::delReg(uint32_t regAddr)
{
    bool deletedReg = false;

    // Find the register in the group collection and remove it.
    for(int x=0; x < camRegGroup.count(); x++ )
    {
        if( camRegGroup.at(x) == regAddr )
        {
            camRegGroup.removeAt(x);
            deletedReg = true;
#if LOG_GRPDEL_TRACING
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "regAddr:" << Qt::showbase << Qt::hex << regAddr << "count:"<< camRegGroup.count() << "camRegGroupSigStr:" << camRegGroupSigStr;
#endif
        }
    }

    // Return the result of the deletion.
    return (deletedReg);
}

/*!
 * @brief Marks the last register in the group and marks the group collection as frozen. No additional
 * registers can be added after this call.
 *
 * @param[in] regAddr   The GIGE register to add to end of collection and mark as the last one.
 *
 * @returns True if the last register successfully added, false if this group collection is already frozen (the lastReg was called),
 * or is the regAddr was zero.
 */
bool FiCamRegGroup::lastReg(uint32_t regAddr)
{
    int count = camRegGroup.count();

    // If the last register address was added already this group is frozen.
    if( lastRegAddr != 0 )
        return false;

    // Need to have a non-zero register.
    if( regAddr == 0 )
        return false;

    // Make sure this last register isn't in the group already.
    for(int x=0; x < count; x++ )
    {
        if( camRegGroup.at(x) == regAddr )
        {
            return false;
        }
    }
    // Increase our count to what we want it to be.
    count++;

    // Add the register to the group.
    camRegGroup.append(regAddr);

    // Mark the last register address.
    lastRegAddr = regAddr;

    // Added a register to the group add to our register "signature" string.
    QString regAddrSigStr = QString("%1").arg(regAddr, 2, 16, QChar('0'));
    camRegGroupSigStr += regAddrSigStr;

#if LOG_GRPINFO_TRACING
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "camRegGroupSigStr:" << camRegGroupSigStr;
#endif

    // Return based on the count equaling what it should be after the append.
    return (camRegGroup.count() == count);
}

/*!
 * @brief Provides a read-only reference to the internal camera register group to allow for usage with
 * existing methods in the FasMotion environment.
 *
 * @returns Unmodifiable FI_COMM GIGE Register Address List for our group of registers.
 */
const FI_COMM::REG_ADDR_LIST& FiCamRegGroup::getCamRegGroupList()
{
    return camRegGroup;
}

/*!
 * @brief Takes in an outside camera group and checks to see if their camRegGroupSigStr
 * matches ours.
 *
 * @returns True if the groups were created with the same group of registers.
 */
bool FiCamRegGroup::camRegGroupCompare(QSharedPointer<FiCamRegGroup> camRegGroupToCompare)
{
    return (camRegGroupSigStr.compare(camRegGroupToCompare->camRegGroupSigStr) == 0);
}

