/*!
 * @file fi_camreg_group.cpp
 * @brief This provides the implementation for camera register group I/O support.
 *
 * Copyright 2023 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_CAMREG_GROUP_H
#define FI_CAMREG_GROUP_H

#include "fi_comm.h"

//--------------------------------------------------------------------
// Class Definition FiCamRegGroup
//--------------------------------------------------------------------

//! Process I/O for a group of GIGE registers on a connected camera.
class FiCamRegGroup
{
public:
    FiCamRegGroup();
    virtual ~FiCamRegGroup();

    // This object is non-copyable and non-copy assignable.
    FiCamRegGroup(const FiCamRegGroup& i) = delete;
    FiCamRegGroup& operator=(const FiCamRegGroup& i) = delete;

    static QSharedPointer<FiCamRegGroup> createCamRegGroup();
    static const QString CAMERA_GRP_READ_FAILURE;
    static const QString CAMERA_GRP_LOG_NAME;

    bool    addReg(uint32_t regAddr);
    bool    delReg(uint32_t regAddr);
    bool    lastReg(uint32_t regAddr);

    const FI_COMM::REG_ADDR_LIST&   getCamRegGroupList();
    const QString&                  getCamRegGroupSigStr()  { return camRegGroupSigStr; }       //!< Returns the group signature string.
    uint32_t                        getCount()              { return camRegGroup.count(); }     //!< How many registers are in the group.
    uint32_t                        getLastRegAddr()        { return lastRegAddr; }             //!< Returns the terminating GIGE register for the group.

    bool camRegGroupCompare(QSharedPointer<FiCamRegGroup> camRegGroupToCompare);

private:
    FI_COMM::REG_ADDR_LIST camRegGroup;                     //!< The FI_COMM GIGE Register Address List maintained by this object.

    uint32_t lastRegAddr;                                   //!< Group terminating GIGE register.
    QString camRegGroupSigStr;                              //!< A "signature" string holding all registers in group up to last register when frozen.
};


#endif // FI_CAMREG_GROUP_H
