/*!
 * @file fi_comm.cpp
 * @brief This provides the interface for communicating with the camera.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_comm.h"

#include <QApplication>
#include <QMdiSubWindow>
#include <QMutexLocker>
#include <QScopedPointer>
#include <QTimer>
#include <QtAlgorithms>

#include "camera_mdi.h"
#include "logger.h"
#include "mainwindow.h"
#include "fcp_socket.h"
#include "regcache_init.h"

namespace FI_COMM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The number of times we'll attempt to issue a write request to the camera.
const int   MAX_WRITE_RETRIES           = 5;

}   // End of namespace FI_COMM.

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
REG_CACHE::CACHE_LIST FI_comm::cache_init_list;

using namespace FI_COMM;

/*!
 * @brief Utility method to generate a loggable message based on a GEV_STATUS_*
 * value.
 *
 * @param gev_status - the status to translate.
 *
 * @returns A read-only string with the gev_status information.
 */
const QString FI_comm::gev_status_to_msg(uint32_t gev_status)
{
    QString gev_status_msg;

    switch (gev_status)
    {
    case GEV_STATUS_SUCCESS:
        gev_status_msg = "GEV_STATUS_SUCCESS";
        break;
    case GEV_STATUS_PACKET_RESEND:
        gev_status_msg = "GEV_STATUS_PACKET_RESEND";
        break;
    case GEV_STATUS_NOT_IMPLEMENTED:
        gev_status_msg = "GEV_STATUS_NOT_IMPLEMENTED";
        break;
    case GEV_STATUS_INVALID_PARAMETER:
        gev_status_msg = "GEV_STATUS_INVALID_PARAMETER";
        break;
    case GEV_STATUS_INVALID_ADDRESS:
        gev_status_msg = "GEV_STATUS_INVALID_ADDRESS";
        break;
    case GEV_STATUS_ACCESS_DENIED:
        gev_status_msg = "GEV_STATUS_ACCESS_DENIED";
        break;
    case GEV_STATUS_BUSY:
        gev_status_msg = "GEV_STATUS_BUSY";
        break;
    case GEV_STATUS_PACKET_NOT_AVAIL:
        gev_status_msg = "GEV_STATUS_PACKET_NOT_AVAIL";
        break;
    case GEV_STATUS_DATA_OVERRUN:
        gev_status_msg = "GEV_STATUS_DATA_OVERRUN";
        break;
    case GEV_STATUS_INVALID_HEADER:
        gev_status_msg = "GEV_STATUS_INVALID_HEADER";
        break;
    case GEV_STATUS_WRONG_CONFIG:
        gev_status_msg = "GEV_STATUS_WRONG_CONFIG";
        break;
    case GEV_STATUS_ERROR:
        gev_status_msg = "GEV_STATUS_ERROR";
        break;
    default:
        gev_status_msg = "GEV_STATUS_UNKNOWN";
        break;
    }
    return gev_status_msg;
}

/*!
 * @brief Constructor for the FI_comm class.
 *
 * @param net_buffers   Number of network buffers to use.
 * @param backoff_buffers   How far into the net_buffers before you request a backoff.
 */
FI_comm::FI_comm(int net_buffers, int backoff_buffers) :
    QObject(nullptr)
{
    FCPSocket  *fcp_socket;

    // Initialize our variables.
    default_packet_delay = PKT_DELAY_DEFAULT;
    num_buffers = net_buffers;
    backoff_level = backoff_buffers;

    // Load our persistent settings.
    read_settings();

    // Initialize our FCP map.
    fcp_map.clear();

    // Initialize our default cache settings.
    init_cache();

    // We want to add an object associated with the broadcast
    // IP address (0xFFFFFFFF) so we can look for cameras.
    fcp_socket = add_socket(FCP_COMM::GLOBAL_BROADCAST_IPADDR);
    if (fcp_socket != nullptr)
    {
        // Have the socket thread start.
        fcp_socket->start();
    }

    connect(MainWindow::GetInstance().get(), &MainWindow::save_settings, this, &FI_comm::write_settings);
    connect(MainWindow::GetInstance().get(), &MainWindow::load_settings, this, &FI_comm::read_settings);
}

/*!
 * @brief Destructor for the FI_comm class.
 */
FI_comm::~FI_comm()
{
    // Save our persistent settings.
    write_settings();

    // Shutdown all the sockets.
    shutdown();
}

/*!
 * @brief   Check to see if "ip_addr" exists in our socket map. If it does, return
 *          the pointer to the socket class. If it doesn't exist, create a new
 *          socket class pointer for "ip_addr" and add it to the map.
 *
 * ASSUMPTION: fcp_map_mutex is already held.
 *
 * @param ip_addr   Address of the camera we're interested in talking with.
 *
 * @return Pointer to the socket class, or nullptr if we were unable to create one.
 */
FCPSocket *FI_comm::add_socket(uint32_t ip_addr)
{
    FCPSocket  *fcp_socket;

    // Check to see if it's already here.
    if (fcp_map.contains(ip_addr))
    {
        // Get the existing class pointer.
        fcp_socket = fcp_map.value(ip_addr);
    }
    else
    {
        // Create our control protocol object so we can talk with the cameras.
        fcp_socket = new FCPSocket(ip_addr);
        if (fcp_socket == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a new FCPSocket!";
        }
        else
        {
            // Add it to our map.
            fcp_map.insert(ip_addr, fcp_socket);

            // Connect our signals/slots.
            connect(fcp_socket, &FCPSocket::msg_received, this, [this, ip_addr]() { message_received(static_cast<int>(ip_addr));});
            connect(fcp_socket, &FCPSocket::msg_timeout, this, [this, ip_addr]() { message_timeout(static_cast<int>(ip_addr));});
            connect(fcp_socket, &FCPSocket::unsol_socket_ready, this, &FI_comm::unsol_socket_ready);

            // We want the error event to happen via the event queue, so force this by using
            // a single-shot timer lambda.
            connect(fcp_socket, &FCPSocket::fcp_error, this, [this, ip_addr]() { QTimer::singleShot(0, this, [this, ip_addr]() { error_received(true, static_cast<int>(ip_addr));});});
            connect(fcp_socket, &FCPSocket::fsp_error, this, [this, ip_addr]() { QTimer::singleShot(0, this, [this, ip_addr]() { error_received(false, static_cast<int>(ip_addr));});});

            connect(fcp_socket, &FCPSocket::unsol_msg_received, this, &FI_comm::handle_unsol_msg);
            connect(fcp_socket, &FCPSocket::ctrl_msg_received, this, &FI_comm::handle_ctrl_msg);
        }
    }

    // Return our class pointer.
    return fcp_socket;
}

/*!
 * @brief Retrieve the camera information structure at the given index. This
 *        is a non-destructive read of the list.
 *
 * @param index     Which camera information structure to retrieve.
 *
 * @return A pointer to the camera information structure.
 */
FI_COMM::FI_CAMERA_INFO *FI_comm::at(int index)
{
    // Sanity check.
    if (index >= discover_camera_list.size())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Invalid index:" << index;
        return nullptr;
    }

    // Return the structure pointer.
    return discover_camera_list.at(index);
}

/*!
 * @brief Public interface to create a communications object to a particular
 *        IP address.
 *
 * @param ip_addr   The IP address of the camera we want to talk with.
 * @return True on success, FALSE otherwise.
 */
bool FI_comm::connect_cam(uint32_t ip_addr)
{
    bool rc = false;
    FCPSocket  *fcp_socket;
    FCPSocket  *bcst_socket;
    QHostAddress our_addr;
    QMutexLocker map_locker(&fcp_map_mutex);

    // We shouldn't be connecting to a broadcast address.
    if (ip_addr == FCP_COMM::GLOBAL_BROADCAST_IPADDR)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Connecting to a broadcast address?";
        goto exitConnectCam;
    }

    // At this point, we should have our broadcast address socket setup because
    // of the discovery process.
    bcst_socket = fcp_map.value(FCP_COMM::GLOBAL_BROADCAST_IPADDR);
    if (bcst_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No broadcast socket?";
        goto exitConnectCam;
    }

    // Clean out the write queue. Make sure we don't have anything left over
    // from a previous connection.
    write_map.remove(ip_addr);

    // Create a socket for this IP address.
    fcp_socket = add_socket(ip_addr);
    if (fcp_socket)
    {
        fcp_socket->start();

        // Set our IP address for the socket to use later.
        our_addr = bcst_socket->get_local_address();
        fcp_socket->set_local_address(our_addr);

        // We have a socket setup set our return code true.
        rc = true;
    }

exitConnectCam:
    // Return TRUE on success, FALSE otherwise.
    return (rc);
}

/*!
 * @brief Disconnect from a camera by removing the communications object and
 *        deleting it.
 *
 * @param ip_addr   The IP address of the camera we're disconnecting from.
 */
void FI_comm::disconnect_cam(uint32_t ip_addr)
{
    QMutexLocker map_locker(&fcp_map_mutex);

    {
        QHostAddress    cam_addr;
        cam_addr.setAddress(qFromBigEndian(ip_addr));
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "IP:" << cam_addr.toString();
    }

    // Does our command protocol map know about this camera?
    if (fcp_map.contains(ip_addr))
    {
        // Get the socket object.
        FCPSocket  *fcp_socket = fcp_map.take(ip_addr);

        // Sanity check.
        if (fcp_socket == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Invalid socket object!";
        }
        else
        {
            // Tell the socket to stop.
            fcp_socket->stop();
        }
    }

    // Clear the cache entry for this IP address.
    if (camera_reg_cache.contains(ip_addr))
    {
        // Get the cache entry for this camera.
        RegCache *my_cache = camera_reg_cache.take(ip_addr);

        // Sanity check.
        if (my_cache == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Invalid cache object!";
        }
        else
        {
            // Clear out the cache object.
            my_cache->clear();

            // Delete the cache entry when convenient.
            my_cache->deleteLater();
        }
    }
}

/*!
 * @brief Allow the client to expire a cache entry.
 * @param ip_addr   The IP address of the entry we're interested in.
 * @param reg_addr  The actual entry we're interested in.
 */
void FI_comm::expire_entry(uint32_t ip_addr, uint32_t reg_addr)
{
    // Make sure our map holds an entry.
    if (camera_reg_cache.contains(ip_addr))
    {
        // Make sure the map contains our register address.
        camera_reg_cache.value(ip_addr)->expire_entry(reg_addr);
    }
}

/*!
 * @brief Send a discovery message as either a broadcast message or to a specific IP address.
 *
 * @param adapter_name      The name of the adapter we want to communicate with.
 * @param ip_addr           [Optional] The IP address of a specific camera to communicate with.
 *
 * @return Status of the command.
 */
int FI_comm::find_cameras(QString& adapter_name, uint32_t ip_addr)
{
    FCPSocket  *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // A value of 0 means that we want to go to ANY address (aka broadcast).
    if (ip_addr == 0)
        ip_addr = FCP_COMM::GLOBAL_BROADCAST_IPADDR;

    // Check to see if it's already here.
    if (fcp_map.contains(ip_addr))
    {
        // Get the existing class pointer.
        fcp_socket = fcp_map.value(ip_addr);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized for ip_addr:" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Ask our FCP socket to send a discovery message.
    return fcp_socket->send_discovery_msg(adapter_name, ip_addr);
}

int FI_comm::find_cameras(uint32_t adapter_ip, uint32_t ip_addr)
{
    FCPSocket  *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // A value of 0 means that we want to go to ANY address (aka broadcast).
    if (ip_addr == 0)
        ip_addr = FCP_COMM::GLOBAL_BROADCAST_IPADDR;

    // Check to see if it's already here.
    if (fcp_map.contains(ip_addr))
    {
        // Get the existing class pointer.
        fcp_socket = fcp_map.value(ip_addr);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized for ip_addr:" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return FI_COMM::FI_INIT_FAILURE;
    }

    // Ask our FCP socket to send a discovery message.
    return fcp_socket->send_discovery_msg(adapter_ip, ip_addr);
}

/*!
 * @brief Helper method to return a pointer to the cache for the camera at
 *        ip_addr, or create a new cache.
 *
 * @param ip_addr   IP address of the camera we're interested in.
 * @return Pointer to the cache for the camera at ip_addr, or nullptr on error.
 */
RegCache *FI_comm::get_cache(uint32_t ip_addr)
{
    RegCache    *reg_cache;

    // Do we have a register value map for this IP addr?
    if (camera_reg_cache.contains(ip_addr))
    {
        reg_cache = camera_reg_cache.value(ip_addr);
    }
    else
    {
        // Create one for this IP address.
        reg_cache = new RegCache(FI_comm::cache_init_list);
        if (reg_cache != nullptr)
            camera_reg_cache[ ip_addr ] = reg_cache;
    }

    // Return our cache pointer.
    return reg_cache;
}

/*!
 * @brief Return the value of the register requested by the previous start_read_reg().
 *
 * @param ip_addr       The IP address of the camera we're talking with.
 * @param reg_addr      The address of the register we're interested in.
 *
 * @return The value read.
 */
uint32_t FI_comm::get_reg(uint32_t ip_addr, uint32_t reg_addr, bool *ok)
{
    uint32_t    ret_value(0);

    // Make sure our map holds an entry.
    if (camera_reg_cache.contains(ip_addr))
    {
        bool reg_ok;
        bool expired;

        // Make sure the map contains our register address.
        ret_value = camera_reg_cache.value(ip_addr)->get_reg(reg_addr, &reg_ok, &expired);
        if (ok != nullptr)
        {
            // Everything is OK if we're able to read the register OK and it's not expired.
            *ok = (reg_ok && !expired);
        }
    }
    else if (ok != nullptr)
    {
        // Not OK.
        *ok = false;
    }

    // Return the value.
    return ret_value;
}

/*!
 * @brief Return the array of the memory address requested by the previous start_read_mem().
 *
 * @param ip_addr       The IP address of the camera we're talking with.
 * @param reg_addr      The address of the register we're interested in.
 *
 * @return The array which was read.
 */
QByteArray *FI_comm::get_reg_buffer(uint32_t ip_addr, uint32_t reg_addr, bool *ok)
{
    QByteArray  *ret_ptr = nullptr;

    // Make sure our map holds an entry.
    if (camera_reg_cache.contains(ip_addr))
    {
        bool reg_ok;
        bool expired;

        // Make sure the map holds our register address.
        ret_ptr = camera_reg_cache.value(ip_addr)->get_mem(reg_addr, &reg_ok, &expired);
        if (ok != nullptr)
        {
            // Everything is OK if we're able to read the register OK and it's not expired.
            *ok = (reg_ok && !expired);
        }
    }
    else if (ok != nullptr)
    {
        // Not OK.
        *ok = false;
    }

    // Return the pointer.
    return ret_ptr;
}

/*!
 * @brief Get the streaming class associated with the camera at the specified IP address.
 * @param ip_addr   The camera we're interested in.
 * @return Pointer to the FSPSocket class, or nullptr if there is an error.
 */
FSPSocket *FI_comm::get_streamer(uint32_t ip_addr)
{
    FCPSocket  *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return nullptr;
    }

    // Return the streaming socket.
    return fcp_socket->get_streamer();
}

/*!
 * @brief Return the status of the previous start_write_reg().
 *
 * Note:  The "status" returned from this method is one of the GEV_STATUS_* codes (see gigemsg.h).
 * This is different from most of the methods returning "status" from this object which in most
 * cases would be an FI_COMM:: FI_STATUS_CODES type.  For example when FI_comm::start_write_reg()
 * is called it will return an FI_COMM::FI_STATUS_CODE and in this case a FI_COMM::FI_SUCCESS return
 * status means that the method was able to successfully write to the camera server the register
 * data BUT it does not mean that the camera_server successfully did write it.  To determine if
 * the camera server was successful, FasMotion must query for the FCP status information this
 * method and the FCP status information is represented in GEV_STATUS_* values.
 *
 * @param ip_addr       The IP address of the camera we're talking with.
 * @param reg_addr      The address of the register we're interested in.
 *
 * @return One of the GEV_STATUS_* codes (see gigemsg.h) indicating the camera_server's write to the
 * register.  This status is the true result of the request to write a camera register. If reg_addr
 * is not in the write hash map GEV_STATUS_UNKNOWN is returned meaning no pending write status exists.
 */
uint32_t FI_comm::get_write_status(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    ret_value = GEV_STATUS_UNKNOWN;

    // Make sure our hash holds the entry.
    if (write_status.contains(ip_addr))
    {
        // Make sure the map holds our register address.
        if (write_status.value(ip_addr)->contains(reg_addr))
        {
            // Remove the value from our hash.
            ret_value = write_status.value(ip_addr)->take(reg_addr);
        }
    }

    // Return the status.
    return ret_value;
}

/*!
 * @brief   Handle the processing of a read register response message.
 *
 * ASSUMPTION: fcp_map_mutex is already held.
 *
 * @param ip_addr   Address of the camera we're talking with.
 * @param p_msg     Pointer to the response message we need to parse.
 */
void FI_comm::handle_read_reg(uint32_t ip_addr, fcp_msg *p_msg)
{
    FCPSocket  *fcp_socket;
    RegCache    *reg_cache;

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return;
    }

    // Get our cache pointer.
    reg_cache = get_cache(ip_addr);
    if (reg_cache != nullptr)
    {
        uint32_t    reg_addr;

        // Get the register address we were trying to read.
        reg_addr = fcp_socket->get_pending_read_register();
        if (reg_addr == FCP_COMM::NO_PENDING_REG_ADDR)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No pending read register address!";
        }
        else if (reg_addr == FCP_COMM::MULTI_PENDING_REG_ADDR)
        {
            int i;
            int length;
            int num_values;
            FI_COMM::REG_ADDR_LIST  reg_list;

            // Get the list of addresses we're waiting on.
            reg_list = fcp_socket->get_pending_read_list();

            // Reset the socket's list.
            fcp_socket->clear_pending_read_list();

            // Get the length of the data section.
            length = qFromBigEndian(p_msg->header.resp.length);
            num_values = length / GIGE_READREG_REQ_B_LEN;

            // We should have a match between the number of registers in our
            // pending list and the number of values we read.
            if (num_values != reg_list.count())
            {
                QHostAddress    cam_addr;
                QList< uint32_t >   new_value_list;

                cam_addr.setAddress(qFromBigEndian(ip_addr));
                for (i = 0; i < num_values; ++i)
                    new_value_list.append(p_msg->data_l[ i ]);

                QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_addr.toString() << "Mismatch between register list and values list";
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "reg_list.count =" << reg_list.count() << "values:" << Qt::showbase << Qt::hex << reg_list;
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "num_values =" << num_values << "values:" << Qt::showbase << Qt::hex << new_value_list;
            }
            else
            {
                // Loop through the registers, and let folks know they're ready.
                for (i = 0; i < num_values; ++i)
                {
                    // Get the next register address.
                    reg_addr = reg_list.at(i);

                    // Save the register value.
                    reg_cache->set_reg(reg_addr, qFromBigEndian(p_msg->data_l[ i ]));

                    // Let interested parties know that the read is finished.
                    emit read_reg_complete(ip_addr, reg_addr);
                }
            }
        }
        else
        {
            // Save the register value.
            reg_cache->set_reg(reg_addr, qFromBigEndian(p_msg->data_l[0]));

            // Let interested parties know that the read is finished.
            emit read_reg_complete(ip_addr, reg_addr);
        }
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find/create a reg_value_map for IP address =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
    }
}

/*!
 * @brief   Handle the processing of a read memory response message.
 *
 * ASSUMPTION: fcp_map_mutex is already held.
 *
 * @param ip_addr   Address of the camera we're talking with.
 * @param p_msg     Pointer to the response message we need to parse.
 */
void FI_comm::handle_read_mem_ack(uint32_t ip_addr, fcp_msg *p_msg)
{
    FCPSocket  *fcp_socket;
    RegCache    *reg_cache;

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return;
    }

    // Get our cache pointer.
    reg_cache = get_cache(ip_addr);
    if (reg_cache != nullptr)
    {
        uint32_t    reg_addr;

        // Get the register address we were trying to read.
        reg_addr = fcp_socket->get_pending_read_register();
        if (reg_addr == (uint32_t) -1)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No pending read register address!";
        }
        else
        {
            // Allocate our data array and move the data over.
            QScopedPointer<QByteArray> data_ptr(new QByteArray);
            if (data_ptr.isNull())
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a QByteArray object!";
                return;
            }

            // Is the camera still busy?
            if (qFromBigEndian(p_msg->header.resp.status) == GEV_STATUS_BUSY)
            {
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera busy reading reg_addr:" << Qt::showbase << Qt::hex << reg_addr;
                data_ptr->append(' ');
            }
            else
            {
                // How many bytes did we get?
                uint32_t data_len = qFromBigEndian(p_msg->header.resp.length);
                if ((data_len <= GIGE_READMEM_ACK_DATA_B) || (data_len > FCP_MSG_DATA_SIZE))
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "Error reading reg_addr:" << Qt::showbase << Qt::hex << reg_addr << "data_len:" << data_len;
                    data_ptr->append(' ');
                }
                else
                {
                    // Ignore the address field.
                    data_len -= GIGE_READMEM_ACK_DATA_B;
                    data_ptr->append((const char *)&p_msg->data_b[ GIGE_READMEM_ACK_DATA_B ], (int)data_len);
                }
            }

            // Save the register value.
            reg_cache->set_reg(reg_addr, data_ptr.get());

            // Let interested parties know that the read is finished.
            emit read_reg_complete(ip_addr, reg_addr);
        }
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find/create a mem_value_map for IP address =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
    }
}

/*!
 * @brief   Handle the processing of a write response message.
 *
 * ASSUMPTION: fcp_map_mutex is already held.
 *
 * @param ip_addr   Address of the camera we're talking with.
 * @param p_msg     Pointer to the response message we need to parse.
 */
void FI_comm::handle_write_mem_ack(uint32_t ip_addr, fcp_msg *p_msg)
{
    FCPSocket          *fcp_socket;
    FI_COMM::REG_VALUE  *reg_value_map;

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return;
    }

    // Do we have a register value map for this IP addr?
    if (write_status.contains(ip_addr))
    {
        reg_value_map = write_status.value(ip_addr);
    }
    else
    {
        // Create one for this IP address.
        reg_value_map = new FI_COMM::REG_VALUE;
        if (reg_value_map != nullptr)
            write_status[ ip_addr ] = reg_value_map;
    }
    if (reg_value_map != nullptr)
    {
        uint32_t    reg_addr;

        // Get the register address we were trying to write.
        reg_addr = fcp_socket->get_pending_write_register();
        if (reg_addr == (uint32_t) -1)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No pending write register address!";
        }
        else
        {
            // Parse the response status from the header.
            uint16_t cmd_status = qFromBigEndian(p_msg->header.resp.status);

            // Only keep the latest status for this register.
            if (reg_value_map->count(reg_addr))
                reg_value_map->remove(reg_addr);

            // Now save the status.
            reg_value_map->insert(reg_addr, ((uint32_t)cmd_status));

            // One way or another, we're done writing.
            emit write_reg_complete(ip_addr, reg_addr);
        }
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find/create a reg_value_map for IP address =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
    }
}

/*!
 * @brief   Handle the processing of a write response message.
 *
 * ASSUMPTION: fcp_map_mutex is already held.
 *
 * @param ip_addr   Address of the camera we're talking with.
 * @param p_msg     Pointer to the response message we need to parse.
 */
void FI_comm::handle_write_reg_ack(uint32_t ip_addr, fcp_msg *p_msg)
{
    FCPSocket          *fcp_socket;
    FI_COMM::REG_VALUE  *reg_value_map;

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return;
    }

    // Do we have a register value map for this IP addr?
    if (write_status.contains(ip_addr))
    {
        reg_value_map = write_status.value(ip_addr);
    }
    else
    {
        // Create one for this IP address.
        reg_value_map = new FI_COMM::REG_VALUE;
        if (reg_value_map != nullptr)
            write_status[ ip_addr ] = reg_value_map;
    }
    if (reg_value_map != nullptr)
    {
        uint32_t    reg_addr;

        // Get the register address we were trying to write.
        reg_addr = fcp_socket->get_pending_write_register();
        if (reg_addr == (uint32_t) -1)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No pending write register address!";
        }
        else
        {
            // Parse the response status from the header.
            uint16_t cmd_status = qFromBigEndian(p_msg->header.resp.status);

            // Only keep the latest status for this register.
            if (reg_value_map->count(reg_addr))
                reg_value_map->remove(reg_addr);

            // Now save the status.
            reg_value_map->insert(reg_addr, ((uint32_t)cmd_status));

            // Were we denied access?
            if (cmd_status == GEV_STATUS_ACCESS_DENIED)
            {
                // Increment our retry counter.
                write_retries++;

                // Do we have anymore tries left?
                if (write_retries < MAX_WRITE_RETRIES)
                {
                    QString debug_msg;

                    debug_msg = QString::fromUtf8("%1: write of reg 0x%2 was denied. Retry %3").arg(__PRETTY_FUNCTION__).arg(reg_addr, 0, 16).arg(write_retries);
                    QLOG_WARN() << debug_msg;

                    // Retry the write.
                    // fcp_socket->retry_write( ip_addr, reg_addr );

                    // Retry the write in a little bit.
                    //
                    // TODO: The retry_xxx variables should probable be a queue...
                    retry_ip_addr = ip_addr;
                    retry_reg_addr = reg_addr;
                    retry_write_value = fcp_socket->get_pending_write_value();
                    retry_write_lock = fcp_socket->get_pending_write_lock();
                    QTimer::singleShot(50, this, &FI_comm::delay_retry);
                    return;
                }
            }

            // One way or another, we're done writing.
            emit write_reg_complete(ip_addr, reg_addr);
        }
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find/create a reg_value_map for IP address =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
    }

    // Process any remaining pending writes.
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << ip_addr << "write_map:" << write_map.count(ip_addr);
    if (write_map.count(ip_addr) > 0)
        QTimer::singleShot(0, this, [ = ] { process_write_group(ip_addr); });
}


/*!
 * @brief Use the INIT_VALUES array to initialize a new cache.
 */
void FI_comm::init_cache(void)
{
    int i;

    // Start at a known location.
    FI_comm::cache_init_list.clear();

    // Build up our list.
    for (i = 0; i < INIT_LIST_LEN; ++i)
        FI_comm::cache_init_list.append(INIT_VALUES[ i ]);
}

/*!
 * @brief Process the response (ACK) message we received.
 *
 * ASSUMPTION: fcp_map_mutex is already held.
 *
 * @param ip_addr   Address of the camera we're talking with.
 * @param rcv_cmd   Which response message are we handling?
 * @param p_msg     Pointer to the actual message structure.
 */
void FI_comm::process_ack(uint32_t ip_addr, quint16 rcv_cmd, fcp_msg *p_msg)
{
    QString     debug_msg;

    // Process based on received command.
    switch (rcv_cmd)
    {
    case GIGE_MSG_DISCOVERY_ACK:
        break;
    case GIGE_MSG_FORCEIP_ACK:
        // Let interested parties know that the force IP command is done.
        emit force_ip_complete(ip_addr);
        break;
    case GIGE_MSG_READREG_ACK:
        handle_read_reg(ip_addr, p_msg);
        break;
    case GIGE_MSG_READMEM_ACK:
        handle_read_mem_ack(ip_addr, p_msg);
        break;
    case GIGE_MSG_WRITEREG_ACK:
        handle_write_reg_ack(ip_addr, p_msg);
        break;
    case GIGE_MSG_WRITEMEM_ACK:
        handle_write_mem_ack(ip_addr, p_msg);
        break;
    default:
        debug_msg = QString::fromUtf8("%1: IPaddr: 0x%2 Unexpected receive command: 0x%3").arg(__PRETTY_FUNCTION__).arg(qToBigEndian(ip_addr), 0, 16).arg(rcv_cmd, 0, 16);
        QLOG_ERROR() << debug_msg;
        break;
    }
}

/*!
 * @brief Process the discovery ACK message, parsing the data into a new FI_COMM::FI_CAMERA_INFO structure.
 *
 * @note It is ASSUMED (always dangerous) that the receiving thread (in the fcp_socket object)
 *       has already validated the response message, checking to make sure the model of camera
 *       is correct etc. We will not duplicate those checks here.
 *
 * @param p_msg     Pointer to the received message.
 */
void FI_comm::process_discovery_ack(fcp_msg *p_msg)
{
    // Allocate a new structure for this camera.
    FI_COMM::FI_CAMERA_INFO *p_camera_info = new FI_COMM::FI_CAMERA_INFO;
    if (p_camera_info == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a new camera_info structure!";
        return;
    }

    // Make sure the memory is cleared.
    memset(p_camera_info, 0, sizeof(FI_COMM::FI_CAMERA_INFO));

    // Initialize default values.
    p_camera_info->packet_delay = (uint32_t)default_packet_delay;

    // Move the data over to the new structure.
    p_camera_info->ip_addr = p_msg->data_l[ GIGE_DSC_ACK_IP_L ];
    p_camera_info->subnet_mask = p_msg->data_l[ GIGE_DSC_ACK_SUBNET_L ];
    p_camera_info->gateway = p_msg->data_l[ GIGE_DSC_ACK_GATEWAY_L ];
    p_camera_info->gige_maj_ver = p_msg->data_w[ GIGE_DSC_ACK_PROTO_VER_W ];
    p_camera_info->gige_min_ver = p_msg->data_w[ GIGE_DSC_ACK_PROTO_VER_W + 1 ];
    memcpy(p_camera_info->mac_addr, &p_msg->data_b[ GIGE_DSC_ACK_MAC_B ], MAX_ADDRESS_LEN);
    p_camera_info->ip_current_cfg = qFromBigEndian(p_msg->data_l[ GIGE_DSC_ACK_IP_CUR_L ]);
    strncpy(p_camera_info->man_name, &p_msg->data_b[ GIGE_DSC_ACK_VENDOR_B ], sizeof(p_camera_info->man_name));
    p_camera_info->man_name[ sizeof(p_camera_info->man_name) - 1 ] = '\0';
    strncpy(p_camera_info->model_name, &p_msg->data_b[ GIGE_DSC_ACK_MODEL_B ], sizeof(p_camera_info->model_name));
    p_camera_info->model_name[ sizeof(p_camera_info->model_name) - 1 ] = '\0';
    strncpy(p_camera_info->dev_version, &p_msg->data_b[ GIGE_DSC_ACK_VERSION_B ], sizeof(p_camera_info->dev_version));
    p_camera_info->dev_version[ sizeof(p_camera_info->dev_version) - 1 ] = '\0';
    strncpy(p_camera_info->man_info, &p_msg->data_b[ GIGE_DSC_ACK_MAN_INFO_B ], sizeof(p_camera_info->man_info));
    p_camera_info->man_info[ sizeof(p_camera_info->man_info) - 1 ] = '\0';
    strncpy(p_camera_info->serial_num, &p_msg->data_b[GIGE_DSC_ACK_SERIAL_B], sizeof(p_camera_info->serial_num));
    p_camera_info->serial_num[ sizeof(p_camera_info->serial_num) - 1 ] = '\0';
    strncpy(p_camera_info->user_name, &p_msg->data_b[ GIGE_DSC_ACK_USER_NAME_B ], sizeof(p_camera_info->user_name));
    p_camera_info->user_name[ sizeof(p_camera_info->user_name) - 1 ] = '\0';

    // Split out the serial number and find the board revision. The format of the serial number
    // is the lower 3 bytes of the MAC and then the board revision after the last '.'
    char *board_ptr = strrchr(p_camera_info->serial_num, '.');
    if (board_ptr != nullptr)
    {
        board_ptr++; // go past
        memcpy(p_camera_info->board_rev, board_ptr, 2);
    }
    else
    {
//        QHostAddress    cam_addr;
//        cam_addr.setAddress(qFromBigEndian((uint32_t)p_camera_info->ip_addr));
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Discarding response from" << cam_addr.toString() << "- serial_num:" << p_camera_info->serial_num;

        // Not interested in this one.
        delete p_camera_info;
        return;
    }

    // Add it to the list.
    discover_camera_list.append(p_camera_info);
}

void FI_comm::process_write_group(uint32_t ip_addr)
{
    QMutexLocker map_locker(&fcp_map_mutex);

    // Do we have any entries in the queue?
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << ip_addr << "write_map:" << write_map.count(ip_addr);
    if (write_map.count(ip_addr) != 0)
    {
        FCPSocket  *fcp_socket;
        bool        need_lock;
        bool        need_ack;
        WRITE_QUEUE *queue_ptr = write_map.value(ip_addr);

        // Sanity check.
        if (queue_ptr->count() == 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "There are no entries in the queue for IP address:" << Qt::showbase << Qt::hex << ip_addr;
            write_map.remove(ip_addr);
            return;
        }

        // Get a scoped pointer to the write group. It willl automatically free when
        // we leave scope.
        QScopedPointer<REG_WRITE_GROUP> write_group(queue_ptr->takeFirst());

        // Check to see if this was the last entry.
        if (queue_ptr->count() == 0)
        {
            // Clear out this entry in the map.
            write_map.remove(ip_addr);
        }

        // Get our socket pointer for the camera at "ip_addr".
        fcp_socket = fcp_map.value(ip_addr, nullptr);
        if (fcp_socket == nullptr)
        {
            // See if we can use the default socket.
            if (fcp_map.contains(FCP_COMM::GLOBAL_BROADCAST_IPADDR))
            {
                // Get the existing class pointer.
                fcp_socket = fcp_map.value(FCP_COMM::GLOBAL_BROADCAST_IPADDR);
                if (fcp_socket == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << "No broadcast socket!";
                    return;
                }
            }
            else
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized!";
                return;
            }
        }

        // Check to see if the camera is offline.
        if (fcp_socket->is_offline())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera is offline. IP:" << Qt::showbase << Qt::hex << ip_addr << "reg_addr:" << write_group->reg_addr << "reg_value:" << write_group->reg_value;
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Trying to write in any event.";

            // If the camera is offline, then we do not need an ack.
            need_ack = false;
        }
        else
        {
            // We expect to get an ack back.
            need_ack = true;
        }

        // Figure out if we need to grab the Fastec lock for the
        // particular register we're trying to write to.
        switch (write_group->reg_addr)
        {
        case GIGE_REG_FASTEC_CCP:
        case GIGE_REG_CCP:
            // Update our lock status.
            update_lock(ip_addr, write_group->reg_addr, write_group->reg_value);

            /*
             * ... Fall through intentional...
             */
            [[fallthrough]];

        case GIGE_REG_SCP0:
        case GIGE_REG_SCPS0:
        case GIGE_REG_SCPD0:
        case GIGE_REG_SCDA0:
        case GIGE_REG_STREAM_CONTROL:
        case GIGE_REG_STREAM_RETRY:
        case GIGE_REG_STREAM_SEQ_START:
        case GIGE_REG_STREAM_START_NUM:
        case GIGE_REG_STREAM_END_NUM:
        case GIGE_REG_STREAM_FRAME_INTERVAL:
        case GIGE_REG_STREAM_FRAME_DELAY:
        case GIGE_REG_STREAM_FORMAT:
        case GIGE_REG_STREAM_FILEPATH:
            need_lock = false;
            break;
        default:
            // If we own the lock, then we don't need to get the lock.
            if (we_own_lock(ip_addr))
                need_lock = false;
            else
                need_lock = true;
            break;
        }

        // Clear our retry counter.
        write_retries = 0;

        // Pass on the write.
        if (fcp_socket->write_reg(ip_addr, write_group->reg_addr, write_group->reg_value, need_ack, need_lock) != FI_COMM::FI_SUCCESS)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FCPSocket::write_reg failed. IP:" << Qt::showbase << Qt::hex << ip_addr << "reg_addr:" << write_group->reg_addr << "reg_value:" << write_group->reg_value;
        }
    }
}

/*!
 * @brief Shutdown the socket.
 */
void FI_comm::shutdown(FCPSocket *fcp_socket)
{
    QMutexLocker map_locker(&fcp_map_mutex);

    // If we have a socket, tell it to stop.
    if (fcp_socket == nullptr)
    {
        QMapIterator< uint32_t, FCPSocket *>    map_iter(fcp_map);

        // Tell all the sockets to shutdown.
        while (map_iter.hasNext())
        {
            // Move to the next item.
            map_iter.next();

            // Get our pointer, and have the socket shutdown.
            fcp_socket = map_iter.value();
            fcp_socket->stop();
        }
        fcp_map.clear();
    }
    else
    {
        fcp_socket->stop();
    }

    // Walk through our data objects and clean them up.
    qDeleteAll(camera_reg_cache);
    camera_reg_cache.clear();

    // Clear out the write status hash.
    qDeleteAll(write_status);
    write_status.clear();

    // Clear out the write map.
    write_map.clear();

    // Clean out our list of cameras.
    qDeleteAll(discover_camera_list);
    discover_camera_list.clear();
}

/*!
 * @brief Mark the camera referenced by the "ip_addr" parameter as off- or on-line.
 *
 * @param ip_addr       The IP address of the camera we're interested in.
 * @param go_offline    True to mark it off-line, false for on-line.
 */
void FI_comm::set_offline(uint32_t ip_addr, bool go_offline)
{
    QHostAddress our_addr;
    our_addr.setAddress(qFromBigEndian(ip_addr));
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ip_addr:" << our_addr.toString() << "go_offline:" << go_offline;

    {
        FCPSocket  *fcp_socket;
        QMutexLocker map_locker(&fcp_map_mutex);

        // Broadcast address can't go offline.
        if (ip_addr == FCP_COMM::GLOBAL_BROADCAST_IPADDR)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Broadcast IP can't go offline!";
            return;
        }

        // Get our socket pointer for the camera at "ip_addr".
        fcp_socket = fcp_map.value(ip_addr, nullptr);
        if (fcp_socket == nullptr)
        {
            // If we're not talking to the camera, then it's perfectly reasonable that there isn't
            // a socket yet.
            return;
        }

        // Make sure we push out any queued writes prior to going offline.
        if (write_map.count(ip_addr) > 0)
        {
            // We need to unlock because process_write_group is going to lock.
            map_locker.unlock();

            // Push out all messages that are still pending.
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Pushing out messages for:" << Qt::showbase << Qt::hex << ip_addr << "write_map:" << write_map.count(ip_addr);
            while (write_map.count(ip_addr) > 0)
            {
                process_write_group(ip_addr);
            }

            // Go ahead and re-lock.
            map_locker.relock();
        }

        // Pass the info on to the socket.
        fcp_socket->set_offline(go_offline);
    }

    // Let folks know what happened.
    emit camera_offline_status_changing(ip_addr, go_offline);
}

/*!
 * @brief Tell the fsp socket about an overlay padding amount.
 *
 * @param ip_addr       The IP address of the camera we're interested in.
 * @param pad_amount    The amount of padding for overlay support (could be zero
 * overlays are turned off).
 */
void FI_comm::set_overlay_padding(uint32_t ip_addr, uint32_t pad_amount)
{
    FCPSocket  *fcp_socket = nullptr;
    {
        QMutexLocker map_locker(&fcp_map_mutex);

        // Get our socket pointer for the camera at "ip_addr".
        fcp_socket = fcp_map.value(ip_addr, nullptr);
    }

    // Install an overlay padding amount in the streaming socket.
    if (fcp_socket != nullptr)
    {
        FSPSocket *fsp_socket = fcp_socket->get_streamer();
        if (fsp_socket)
        {
            fsp_socket->set_overlay_padding(pad_amount);
        }
    }
}

/*!
 * @brief Use the FCP Force IP command to force the camera to change
 *        it's network settings to a static IP address.
 *
 * @param mac_addr      The MAC address for the camera.
 * @param new_ip        The new IP address for the camera.
 * @param new_mask      The new network mask for the camera.
 * @param new_gw        The new gateway IP address for the camera.
 *
 * @return FI_COMM::FI_SUCCESS if everything is OK, error code otherwise.
 */
int FI_comm::start_force_ip(const uint8_t *mac_addr, uint32_t new_ip, uint32_t new_mask, uint32_t new_gw)
{
    FCPSocket *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(FCP_COMM::GLOBAL_BROADCAST_IPADDR, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized!";
        return FI_COMM::FI_INIT_FAILURE;
    }

    // The Force IP message is a special message, handled uniquely.
    return fcp_socket->force_ip(mac_addr, new_ip, new_mask, new_gw);
}

/*!
 * @brief Initiate a register read request to the camera server at ip_addr.
 *
 * @param ip_addr   The IP address of the camera we want to talk with.
 * @param reg_addr  The register address we want to read.
 *
 * @return The status of the request.
 */
int FI_comm::start_read_reg(uint32_t ip_addr, uint32_t reg_addr)
{
    FCPSocket  *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Don't check the cache for a broadcast IP address.
    if (ip_addr != FCP_COMM::GLOBAL_BROADCAST_IPADDR)
    {
        bool    reg_ok;

        // See if the register is in our cache. We'll ignore the value, we're
        // really only interested in if the value is in the cache.
        get_reg(ip_addr, reg_addr, &reg_ok);

        // Now check to see if our register is OK.
        if (reg_ok)
        {
            // The register is in cache, and it's good to go. Pretend that we
            // actually sent the message - just emit an "it's ready" message.
            emit read_reg_complete(ip_addr, reg_addr);
            return FI_COMM::FI_SUCCESS;
        }
    }

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        // See if we can use the default socket.
        if (fcp_map.contains(FCP_COMM::GLOBAL_BROADCAST_IPADDR))
        {
            // Get the existing class pointer.
            fcp_socket = fcp_map.value(FCP_COMM::GLOBAL_BROADCAST_IPADDR);
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized!";
            return FI_COMM::FI_INIT_FAILURE;
        }
    }

    // Check to see if the camera is offline.
    if (fcp_socket->is_offline())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera is offline. IP:" << Qt::showbase << Qt::hex << ip_addr;
        return FI_COMM::FI_CAMERA_OFFLINE;
    }

    // Ask our FCP socket to send a discovery message.
    return fcp_socket->read_reg(ip_addr, reg_addr);
}

/*!
 * @brief Initiate a memory read request to the camera server at ip_addr.
 *
 * @param ip_addr   The IP address of the camera we want to talk with.
 * @param reg_addr  The register address we want to read.
 * @param reg_len   The number of bytes we expect to read.
 *
 * @return The status of the request.
 */
int FI_comm::start_read_mem(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_len)
{
    FCPSocket  *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Don't check the cache for a broadcast IP address.
    if (ip_addr != FCP_COMM::GLOBAL_BROADCAST_IPADDR)
    {
        bool    reg_ok;
        QByteArray  *data_ptr;

        // See if the register is in our cache. We'll ignore the value, we're
        // really only interested in if the value is in the cache.
        data_ptr = get_reg_buffer(ip_addr, reg_addr, &reg_ok);

        // Don't care about the data, delete it if we got something.
        if (data_ptr != nullptr)
            delete data_ptr;

        // Was the register OK?
        if (reg_ok)
        {
            // The register is in cache, and it's good to go. Pretend that we
            // actually sent the message - just emit an "it's ready" message.
            emit read_reg_complete(ip_addr, reg_addr);
            return FI_COMM::FI_SUCCESS;
        }
    }

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        // See if we can use the default socket.
        if (fcp_map.contains(FCP_COMM::GLOBAL_BROADCAST_IPADDR))
        {
            // Get the existing class pointer.
            fcp_socket = fcp_map.value(FCP_COMM::GLOBAL_BROADCAST_IPADDR);
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized!";
            return FI_COMM::FI_INIT_FAILURE;
        }
    }

    // Check to see if the camera is offline.
    if (fcp_socket->is_offline())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera is offline. IP:" << Qt::showbase << Qt::hex << ip_addr;
        return FI_COMM::FI_CAMERA_OFFLINE;
    }

    // Ask our FCP socket to send a discovery message.
    return fcp_socket->read_mem(ip_addr, reg_addr, reg_len);
}

/*!
 * @brief Initiate a register read request to the camera server at ip_addr
 *        for multiple registers in the camera.
 *
 * @param ip_addr   The IP address of the camera we want to talk with.
 * @param reg_list  The list of register addresses we want to read. Assumes caller has verified list
 * contents exist. If it is empty this will return FI_COMM::SUCCESS because it didn't have to read
 * anything and is done early.
 *
 * @return The status of the request, one of the FI_COMM::FI_STATUS_CODES.
 */
int FI_comm::start_read_multireg(uint32_t ip_addr, const FI_COMM::REG_ADDR_LIST& reg_list)
{
    REG_ADDR_LIST   modified_list;
    FCPSocket      *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // If the user provided an empty reg_list just return a success condition.
    if (reg_list.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera group read request was empty. IP:" << Qt::showbase << Qt::hex << ip_addr;
        return FI_COMM::FI_SUCCESS;
    }

    // Initialize our modified list to a known (empty) state.
    modified_list.clear();

    // Don't check the cache for a broadcast IP address.
    if (ip_addr != FCP_COMM::GLOBAL_BROADCAST_IPADDR)
    {
        int     last_index          = reg_list.count() - 1;
        bool    last_reg_in_cache   = false;    // Assume we will need to fetch it.

        // See if any of the requested registers are already in our cache.
        for (int i = 0; i < last_index; ++i)
        {
            bool reg_ok;    // True if the register is in cache, false if not.

            // See if the register is in our cache. We'll ignore the value, we're
            // really only interested in if the value is in the cache.
            get_reg(ip_addr, reg_list.at(i), &reg_ok);

            // Now check to see if our register is available.
            if (reg_ok)
            {
                // The register is in cache, and it's good to go. Pretend that we
                // actually sent the message - just emit an "it's ready" message.
                emit read_reg_complete(ip_addr, reg_list.at(i));
            }
            else
            {
                // Need to go fetch this one.
                modified_list.append(reg_list.at(i));
            }
        }

        // The very last register address in a group read request serves as a terminating
        // register read for the group. Treat it differently and ensure it is processed last.
        // NOTE: The empty list condition is handled at the start of this function so we don't
        // get here unless there is at least one entry in the group read request.

        // If we have not put any registers on the modified_list it means they are all in
        // the cache. We want to leverage our cache should our last address also be cached.
        if (modified_list.isEmpty())
        {
            get_reg(ip_addr, reg_list.at(last_index), &last_reg_in_cache);
        }

        // If last register is in the cache, emit the read_reg_complete signal.
        if (last_reg_in_cache)
        {
            emit read_reg_complete(ip_addr, reg_list.at(last_index));
        }
        else
        {
            // The last register is not in the cache and need to fetch it.
            modified_list.append(reg_list.at(last_index));
        }
    }
    else
    {
        // Simply copy the entire list over.
        modified_list.append(reg_list);
    }

    // Do we need to go get any of the registers?
    if (modified_list.isEmpty())
        return FI_SUCCESS;

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        // See if we can use the default socket.
        if (fcp_map.contains(FCP_COMM::GLOBAL_BROADCAST_IPADDR))
        {
            // Get the existing class pointer.
            fcp_socket = fcp_map.value(FCP_COMM::GLOBAL_BROADCAST_IPADDR);
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized!";
            return FI_COMM::FI_INIT_FAILURE;
        }
    }

    // Check to see if the camera is offline.
    if (fcp_socket->is_offline())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera is offline. IP:" << Qt::showbase << Qt::hex << ip_addr;
        return FI_COMM::FI_CAMERA_OFFLINE;
    }

    // Ask our FCP socket to send a discovery message.
    return fcp_socket->read_multireg(ip_addr, modified_list);
}


/*!
 * @brief Register with the camera at ip_addr for FMP (aka "unsolicited") messages.
 *
 * @param ip_addr   IP address of the camera we're interested in.
 * @return GEV_STATUS_SUCCESS or one of the error codes.
 */
int FI_comm::start_unsolicited(uint32_t ip_addr)
{
    FCPSocket *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return GEV_STATUS_INVALID_ADDRESS;
    }

    // Tell our FCP class to start up unsolicited messages.
    return fcp_socket->start_unsolicited();
}

/*!
 * @brief Initiate a FCP memory write request to the camera server at ip_addr.
 *
 * @param ip_addr   The IP address of the camera we want to talk with.
 * @param reg_addr  The register address we want to write to.
 * @param buffer    The QByteArray we want to write to the camera.
 *
 * @return The status of the request.
 */
int FI_comm::start_write_mem(uint32_t ip_addr, uint32_t reg_addr, const QByteArray& buffer)
{
    FCPSocket  *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        // See if we can use the default socket.
        if (fcp_map.contains(FCP_COMM::GLOBAL_BROADCAST_IPADDR))
        {
            // Get the existing class pointer.
            fcp_socket = fcp_map.value(FCP_COMM::GLOBAL_BROADCAST_IPADDR);
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Socket is not initialized!";
            return FI_COMM::FI_INIT_FAILURE;
        }
    }

    // Check to see if the camera is offline.
    if (fcp_socket->is_offline())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera is offline. IP:" << Qt::showbase << Qt::hex << ip_addr;
        return FI_COMM::FI_CAMERA_OFFLINE;
    }

    // Pass on the write.
    return fcp_socket->write_mem(ip_addr, reg_addr, buffer, true);
}

/*!
 * @brief Initiate a FCP register write request to the camera server at ip_addr.
 *
 * @param ip_addr   The IP address of the camera we want to talk with.
 * @param reg_addr  The register address we want to write to.
 * @param reg_value The value we want to write to reg_addr.
 *
 * @return The status of the request.
 */
int FI_comm::start_write_reg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    REG_WRITE_GROUP *write_group;
    WRITE_QUEUE *queue_ptr;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Build up our group structure.
    write_group = new REG_WRITE_GROUP;
    if (write_group == nullptr)
        return FI_COMM::FI_ENOMEM;
    write_group->reg_addr = reg_addr;
    write_group->reg_value = reg_value;

    // Do we have an existing list going?
    queue_ptr = write_map.value(ip_addr);
    if (queue_ptr == nullptr)
    {
        // We need to allocate our queue.
        queue_ptr = new WRITE_QUEUE;
        if (queue_ptr == nullptr)
            return FI_COMM::FI_ENOMEM;

        // Add the queue to the map.
        write_map.insert(ip_addr, queue_ptr);
    }

    // Push it onto our list.
    queue_ptr->append(write_group);

    // Do we need to kick start the system?
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << "ip_addr:" << ip_addr << "reg_addr:" << reg_addr << "reg_value:" << reg_value << "write_queue:" << queue_ptr->count();
    if (queue_ptr->count() == 1)
        QTimer::singleShot(0, this, [ = ] { process_write_group(ip_addr); });

    return FI_COMM::FI_SUCCESS;
}

/*!
 * @brief Unregister with the camera at ip_addr for FMP messages.
 * @param ip_addr   IP address of the camera we're interested in.
 */
void FI_comm::stop_unsolicited(uint32_t ip_addr)
{
    FCPSocket *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(ip_addr);
        return;
    }

    // Tell our FCP class to stop unsolicited messages.
    fcp_socket->stop_unsolicited();
}

/*!
 * @brief Used to finish a stream event - or simply remove a queued
 *        object from the list.
 *
 * @param receiver  The receiver of the stream event.
 */
void FI_comm::stream_finish(QObject *receiver)
{
    int     index;
    bool    popped_top(false);

    // Loop through the list looking for the receiver.
    index = 0;
    while (index < stream_list.count())
    {
        // Do we have a match?
        if (receiver == stream_list.at(index)->first)
        {
            // Remove this item. Don't update the index because
            // there's a new entry at this index (once we delete
            // the old one).
            delete stream_list.takeAt(index);

            // Did we just pop the first entry?
            if (index == 0)
                popped_top = true;
        }
        else
        {
            // Move to the next entry.
            ++index;
        }
    }

    // If this was the first (aka "top") entry in the list,
    // and there are more - let's start the next one.
    if (popped_top && !stream_list.isEmpty())
    {
        // Pass the event directly on to the receiver.
        QApplication::postEvent(stream_list.at(0)->first, stream_list.at(0)->second);
    }
}

/*!
 * @brief This method is used to manage multicamera access to full-bandwidth
 *        streaming functions, such as pulling down the raw per-frame metadata.
 *
 * @param receiver      Pointer to the object to receive the event.
 * @param event         The event we want to send.
 */
void FI_comm::stream_queue(QObject *receiver, QEvent *event)
{
    // Is our stream list empty?
    if (stream_list.isEmpty())
    {
        // Pass the event directly on to the receiver.
        QApplication::postEvent(receiver, event);
    }
    else
    {
        // Make sure there isn't already a request for this receiver.
        for (int index = 0; index < stream_list.count(); ++index)
        {
            // Do we have a match?
            if (receiver == stream_list.at(index)->first)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Found duplicate at index:" << index;
                return;
            }
        }
    }

    // Queue up the request. We always queue up the event. That way,
    // if the list is not empty, then we know the streamer is active.
    stream_list.append(new STREAM_EVENT(receiver, event));
}

/*!
 * @brief Pop the first camera information structure off the list and return the pointer.
 *
 * @return The pointer to the first camera information structure in the list, or nullptr.
 */
FI_COMM::FI_CAMERA_INFO *FI_comm::takeFirst(void)
{
    // Sanity check.
    if (discover_camera_list.isEmpty())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "camera_list is empty!";
        return nullptr;
    }

    // Return the structure pointer.
    return discover_camera_list.takeFirst();
}

/*!
 * @brief Update our idea of lock ownership.
 *
 * @param ip_addr       Camera we're talking about.
 * @param reg_addr      Which lock register, GIGE_REG_CCP or GIGE_REG_FASTEC_CCP, we're working with.
 * @param reg_value     The new value.
 */
void FI_comm::update_lock(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    RegCache *reg_cache;

    // Get our cache pointer.
    reg_cache = get_cache(ip_addr);
    if (reg_cache == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No cache for IP=" << qToBigEndian(ip_addr);
        return;
    }

    // Update the cache for our lock status.
    reg_cache->set_reg(reg_addr, reg_value);
}

/*!
 * @brief Return true if we own the CCP and FASTEC_CCP locks, false otherwise.
 *
 * @param ip_addr   Camera we're concerned with.
 */
bool FI_comm::we_own_lock(uint32_t ip_addr)
{
    RegCache    *reg_cache;
    uint32_t    ccp_reg;
    uint32_t    fccp_reg;
    bool        own_it;

    // Get our cache pointer.
    reg_cache = get_cache(ip_addr);
    if (reg_cache == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No cache for IP=" << qToBigEndian(ip_addr);
        return false;
    }

    // Get our registers.
    ccp_reg = reg_cache->get_reg(GIGE_REG_CCP);
    fccp_reg = reg_cache->get_reg(GIGE_REG_FASTEC_CCP);

    // Assume we do NOT own the lock.
    own_it = false;

    // Do we have the CCP register locked?
    if ((ccp_reg & GIGE_CCP_CA_BE) == GIGE_CCP_CA_BE)
    {
        // Now check the Fastec lock.
        if ((fccp_reg & GIGE_FASTEC_CCP_CA_BE) == GIGE_FASTEC_CCP_CA_BE)
        {
            // We own it!
            own_it = true;
        }
    }

    // Return our flag.
    return own_it;
}


//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void FI_comm::delay_retry(void)
{
    FCPSocket *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "retry_ip_addr:" << Qt::showbase << Qt::hex << retry_ip_addr << "retry_reg_addr:" << retry_reg_addr
                 << "retry_write_value" << retry_write_value;

    // Get our socket pointer for the camera at "ip_addr".
    fcp_socket = fcp_map.value(retry_ip_addr, nullptr);
    if (fcp_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no socket for ip_addr =" << Qt::showbase << Qt::hex << qToBigEndian(retry_ip_addr);
        return;
    }

    fcp_socket->retry_write(retry_ip_addr, retry_reg_addr, retry_write_value, retry_write_lock);
}

/*!
 * @brief Process an unsolicited control message.
 *
 * @param ip_addr       The IP address of the camera which sent the unsolicited message.
 * @param reg_addr      The address of the register that was changed.
 * @param reg_value     The new value of the register.
 */
void FI_comm::handle_ctrl_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    RegCache *reg_cache;

    // Get our cache pointer.
    reg_cache = get_cache(ip_addr);
    if (reg_cache == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No cache for IP=" << qToBigEndian(ip_addr);
        return;
    }

    // If the lock was released, then no-one ones it. If the lock is being set,
    // it's either ours or someone elses. Either way, we'll figure it out later.
    if (reg_value == 0)
    {
        // Set the value and we're done.
        reg_cache->set_reg(reg_addr, reg_value);
    }
}

/*!
 * @brief When we receive an unsolicited message, make sure we update our cache entry. Then
 *        pass the message on to any other listener.
 *
 * @param ip_addr       The IP address of the camera which sent the unsolicited message.
 * @param reg_addr      The address of the register that was changed.
 * @param reg_value     The new value of the register.
 */
void FI_comm::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    RegCache *reg_cache;

    // Get our cache pointer.
    reg_cache = get_cache(ip_addr);
    if (reg_cache == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No cache for IP=" << qToBigEndian(ip_addr);
        return;
    }

    // Save the register value.
    reg_cache->set_reg(reg_addr, reg_value);

    // Pass on the information.
    emit unsol_msg_received(ip_addr, reg_addr, reg_value);
}


/*!
 * @brief   Slot which gets invoked when the socket times out waiting for a message.
 * @param ip_addr   The address of the camera we're talking with.
 */
void FI_comm::error_received(bool fcp_error, int ip_addr)
{
    int err_value = 0;
    QString err_str;
    CameraMDI *camera_ptr;
    QMdiSubWindow *win_ptr;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << qToBigEndian((uint32_t)ip_addr) << "fcp_error:" << fcp_error;

    // Get the point to the associated child window.
    win_ptr = MainWindow::GetInstance()->find_mdi_child((uint32_t)ip_addr);
    if (win_ptr == nullptr)
        return;

    // Get a pointer to the camera object associated with the child window.
    camera_ptr = qobject_cast<CameraMDI *>(win_ptr->widget());
    if (camera_ptr == nullptr)
    {
        // Camera is already gone.
        return;
    }

    // Are we offline?
    if (camera_ptr->is_offline())
    {
        // Camera is offline. No need to worry about errors.
        // Process any remaining pending writes.
        if (write_map.count(ip_addr) > 0)
            QTimer::singleShot(0, this, [ = ] { process_write_group(ip_addr); });
        return;
    }

    // Create a block for the mutex.
    {
        FCPSocket  *fcp_socket;
        QMutexLocker map_locker(&fcp_map_mutex);

        // Check to see if it's already here.
        if (fcp_map.contains((uint32_t)ip_addr))
        {
            // Get the existing class pointer.
            fcp_socket = fcp_map.value((uint32_t)ip_addr);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "IP address not found in the fcp_map!";
            fcp_socket = nullptr;
        }

        // Make sure we found an FCP object.
        if (fcp_socket != nullptr)
        {
            // Let folks know an error happened.
            err_value = fcp_socket->get_last_err(err_str);
            QLOG_INFO() << __PRETTY_FUNCTION__ << "FCP err_value:" << err_value << ":" << err_str;
        }
    }

    // Control or streaming error?
    if (fcp_error)
    {
        // This was a control protocol error, we'll just emit a signal to let others know what happened.
        emit error(ip_addr, err_value);
    }
    else
    {
        // This was a streaming error. Shutdown the camera window.
        camera_ptr->set_offline(true);
    }
}

/*!
 * @brief   Slot which gets invoked when the socket receives a message.
 * @param ip_addr   The address of the camera we're talking with.
 */
void FI_comm::message_received(int ip_addr)
{
    FCPSocket  *fcp_socket;
    QMutexLocker map_locker(&fcp_map_mutex);

    // Check to see if it's already here.
    if (fcp_map.contains((uint32_t)ip_addr))
    {
        // Get the existing class pointer.
        fcp_socket = fcp_map.value((uint32_t)ip_addr);
    }
    else
    {
        QHostAddress    cam_addr;
        cam_addr.setAddress(qFromBigEndian((uint32_t)ip_addr));

        // No match - don't know who this is. We're probably in the process of disconnecting.
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No fcp_socket for" << cam_addr.toString();
        return;
    }

    // Loop through all the received messages.
    while (fcp_socket->get_msg_count() > 0)
    {
        quint16     rcv_cmd;
        fcp_msg    *p_msg;
        QString     debug_msg;

        // Get the first message from the list.
        p_msg = fcp_socket->pop_msg();
        if (p_msg == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "What? get_msg_count() returned non-zero, but pop_msg() returned nullptr!";
            return;
        }

        // Get the command from the response.
        rcv_cmd = qFromBigEndian(p_msg->header.resp.command);

        // Process based on received command.
        switch (rcv_cmd)
        {
        case GIGE_MSG_DISCOVERY_ACK:
            process_discovery_ack(p_msg);
            break;
        case GIGE_MSG_FORCEIP_ACK:
        case GIGE_MSG_READREG_ACK:
        case GIGE_MSG_WRITEREG_ACK:
        case GIGE_MSG_READMEM_ACK:
        case GIGE_MSG_WRITEMEM_ACK:
            process_ack((uint32_t)ip_addr, rcv_cmd, p_msg);
            break;
        default:
            debug_msg = QString::fromUtf8("%1: Unexpected receive command: 0x%2:0x%3").arg(__PRETTY_FUNCTION__).arg(ip_addr, 0, 16).arg(rcv_cmd, 0, 16);
            QLOG_WARN() << debug_msg;
            break;
        }

        // Release the message structure.
        delete p_msg;
    }

    // Process any messages which might be queued up.
    fcp_socket->process_msg_queue();
}

/*!
 * @brief   Slot which gets invoked when the socket times out waiting for a message.
 * @param ip_addr   The address of the camera we're talking with.
 */
void FI_comm::message_timeout(int ip_addr)
{
    bool        read_op;
    uint32_t    reg_addr;
    FCPSocket  *fcp_socket;

    {
        QMutexLocker map_locker(&fcp_map_mutex);

        // Check to see if it's already here.
        if (fcp_map.contains((uint32_t)ip_addr))
        {
            // Get the existing class pointer.
            fcp_socket = fcp_map.value((uint32_t)ip_addr);
        }
        else
        {
            // No match - don't know who this is. We're probably in the process of disconnecting.
            // Process any remaining pending writes.
            if (write_map.count(ip_addr) > 0)
                QTimer::singleShot(0, this, [ = ] { process_write_group(ip_addr); });
            return;
        }

        // Figure out what the register address was...look at reads.
        read_op = true;
        reg_addr = fcp_socket->get_pending_read_register();
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "(read) ip_addr:" << Qt::showbase << Qt::hex << qToBigEndian((uint32_t)ip_addr) << "reg_addr:" << reg_addr;
        if (reg_addr == FCP_COMM::NO_PENDING_REG_ADDR)
        {
            // See if it was a write that timed out.
            reg_addr = fcp_socket->get_pending_write_register();
            read_op = false;
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "(write) ip_addr:" << Qt::showbase << Qt::hex << qToBigEndian((uint32_t)ip_addr) << "reg_addr:" << reg_addr;
        }

        // Process any messages which might be queued up.
        fcp_socket->process_msg_queue();

        // Process any remaining pending writes.
        if (write_map.count(ip_addr) > 0)
            QTimer::singleShot(0, this, [ = ] { process_write_group(ip_addr); });
    }

    // Got a timeout signal!
    emit msg_timeout((uint32_t)ip_addr, reg_addr, read_op);
}

/*!
 * @brief Read the persistent settings for our class.
 */
void FI_comm::read_settings(void)
{
    // Load our previous settings.
    QSettings settings;
    settings.beginGroup("Communications");

    default_packet_delay = settings.value("pkt_delay", PKT_DELAY_DEFAULT).toInt();
    settings.endGroup();
}

/*!
 * @brief Save the persistent settings for our class.
 */
void FI_comm::write_settings(void)
{
    // Save the current settings.
    QSettings settings;
    settings.beginGroup("Communications");

    settings.setValue("pkt_delay", default_packet_delay);
    settings.endGroup();
}

