/*!
 * @file fi_comm.h
 * @brief This file contains the definitions for the Fastec Imaging
 *        Communications class, FI_comm.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_COMM_H
#define FI_COMM_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <QObject>
#include <QMap>
#include <QMutex>
#include <QPair>

#include "gige_msg.h"
#include "regcache.h"

//--------------------------------------------------------------------
// Namespace: FI_COMM
//--------------------------------------------------------------------
namespace FI_COMM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! A constant which is not associated with any register address.
const uint32_t  NO_REGISTER             = ((uint32_t) -1);

//! A constant which indicates GEV status is unknown.
const uint32_t  GEV_STATUS_UNKNOWN      = ((uint32_t) -1);

//! Status values returned by FI_COMM related functions.
enum FI_STATUS_CODES
{
    FI_SUCCESS  = 0,                    //!< All is well.
    FI_ENOMEM,                          //!< We are out of memory!
    FI_INIT_FAILURE,                    //!< Initialization error.
    FI_SOCKET_ERR,                      //!< Got a socket error.
    FI_BAD_MSG_TYPE,                    //!< Got a bad (unexpected) message type.
    FI_MSG_ID_SYNC_ERR,                 //!< Message IDs are out of sync.
    FI_CAMERA_OFFLINE,                  //!< The camera is off-line.
    FI_UNKNOWN_ERR
};

// Lengths of various fields.
const int   MANUFACTURERS_NAME_LEN      = GIGE_DSC_ACK_VENDOR_B_LEN;
const int   MODEL_NAME_LEN              = GIGE_DSC_ACK_MODEL_B_LEN;
const int   DEV_VERSION_LEN             = GIGE_DSC_ACK_VERSION_B_LEN;
const int   MANUFACTURERS_INFO_LEN      = GIGE_DSC_ACK_MAN_INFO_B_LEN;
const int   SERIAL_NUM_LEN              = GIGE_DSC_ACK_SERIAL_B_LEN;
const int   USER_NAME_LEN               = GIGE_DSC_ACK_USER_NAME_B_LEN;
const int   SENSOR_VERSION_LEN          = GIGE_SENSOR_VER_LEN;
const int   MAX_ADDRESS_LEN             = 6;
const int   BOOTLOADER_VER_LEN          = 52;

const int   PKT_DELAY_MIN               = 10;           //!< Minimum packet delay (microseconds).
const int   PKT_DELAY_MAX               = 1000000;      //!< Maximum packet delay (microseconds).
const int   PKT_DELAY_DEFAULT           = 200;          //!< A default (aka "safe") value for the interpacket delay.

const char  CAM_BOARD_REV_A             = 'A';
const char  CAM_BOARD_REV_A3            = '3';

constexpr uint8_t FASTEC_OUI[3]         = { 0xA4, 0x1B, 0xC0 };

//! FI_CAMERA_INFO is a structure used to pass camera information between
//! the low level GigE protocol handlers and the application level code.
typedef struct FI_CAMERA_INFO
{
    // The following are all binary fields set by the results from cam_find_cameras()
    uint32_t    ip_addr;                            //!< IP address in network byte order.
    uint32_t    subnet_mask;                        //!< Subnet in network byte order.
    uint32_t    gateway;                            //!< Default gateway in network byte order.
    uint16_t    gige_maj_ver;                       //!< Version of the FCP spec implemented major number.
    uint16_t    gige_min_ver;                       //!< Version of the FCP spec implemented minor number.
    uint8_t mac_addr[ MAX_ADDRESS_LEN ];                //!< MAC address of the device.
    uint32_t    ip_current_cfg;                     //!< Current IP configuration, bit 31 means static ip, bit 30 means DHCP, bit 29 means LLA.

    // The following are all strings set by the results from cam_find_cameras()
    char    man_name[ MANUFACTURERS_NAME_LEN ];         //!< Manufacturers name.
    char    model_name[ MODEL_NAME_LEN ];               //!< Model name.
    char    dev_version[ DEV_VERSION_LEN ];             //!< Version of the firmware.
    char    man_info[ MANUFACTURERS_INFO_LEN ];         //!< Manufacturer info.
    char    serial_num[ SERIAL_NUM_LEN ];               //!< Serial number.
    char    user_name[ USER_NAME_LEN ];                 //!< User defined name.

    //! The board revision is a subset of the serial num for A3 and above. Format in hex is 'A3', 'A4', etc.
    char    board_rev[ 2 ];
    char    sensor_version[ SENSOR_VERSION_LEN ];       //!< Version of the sensor on the board
    char    bootloader_version[ BOOTLOADER_VER_LEN ];   //!< Bootloader version
    uint32_t    fpga_version;                           //!< FPGA version

    bool        stream;                                 //!< True if the user wants to stream video from the camera.
    bool        stream_running;                         //!< True if the stream is running, i.e. we are getting frames when asked for them.
    bool        offline;                                //!< True if the camera is off-line for some reason (such as FPN calibration).
    int         offline_skip;                           //!< The number of times we still need to skip before looking for the camera again.
    bool        rebooting;                              //!< True if we requested a reboot of the camera.
    bool        startup_open;                           //!< For local camera, did we already auto-open it once on startup?

    uint32_t    camera_missing_count;                   //!< How many scan cycles has it been since we've heard from this camera?
    uint32_t    packet_delay;                           //!< How much of a packet delay should there be.

} FI_CAMERA_INFO;

//! Used to queue up write requests.
typedef struct REG_WRITE_GROUP
{
    uint32_t reg_addr;
    uint32_t reg_value;
} REG_WRITE_GROUP;

typedef QList<FI_COMM::REG_WRITE_GROUP *> WRITE_QUEUE;  //!< A list of register/value pairs to be written.
typedef QList<uint32_t> REG_ADDR_LIST;                  //!< A list of register addresses to be read.
typedef QMap<uint32_t, uint32_t> REG_VALUE;             //!< A map of register [address, value] pairs we've read/written.
typedef QPair<QObject *, QEvent *> STREAM_EVENT;        //!< The [receiver, event] pair associated with a stream event.

}   // End namespace (FI_COMM)

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class   FCPSocket;
class   FSPSocket;

//--------------------------------------------------------------------
// Class FI_comm
//--------------------------------------------------------------------
class FI_comm : public QObject
{
    Q_OBJECT

public:
    explicit FI_comm(int net_buffers, int backoff_buffers);
    ~FI_comm();

    static const QString gev_status_to_msg(uint32_t gev_status);

    bool    connect_cam(uint32_t ip_addr);
    void    disconnect_cam(uint32_t ip_addr);
    void    expire_entry(uint32_t ip_addr, uint32_t reg_addr);
    int     find_cameras(QString& adapter_name, uint32_t ip_addr);
    int     find_cameras(uint32_t adapter_ip, uint32_t ip_addr);
    int     get_backoff_level()                   { return backoff_level; }
    int     get_num_cameras()                     { return discover_camera_list.size(); }
    int     get_num_net_buffers()                 { return num_buffers; }
    int     get_default_pkt_delay()               { return default_packet_delay; }
    uint32_t    get_reg(uint32_t ip_addr, uint32_t reg_addr, bool *ok = nullptr);
    QByteArray  *get_reg_buffer(uint32_t ip_addr, uint32_t reg_addr, bool *ok = nullptr);
    FSPSocket  *get_streamer(uint32_t ip_addr);
    uint32_t    get_write_status(uint32_t ip_addr, uint32_t reg_addr);
    void    set_backoff_level(int new_level)          { backoff_level = new_level; }
    void    set_default_pkt_delay(int pkt_delay)      { default_packet_delay = pkt_delay; write_settings(); }
    void    set_net_buffers(int new_count)            { num_buffers = new_count; }
    void    set_offline(uint32_t ip_addr, bool go_offline);
    void    set_overlay_padding(uint32_t ip_addr, uint32_t pad_amount);
    void    shutdown(FCPSocket *fcp_socket = nullptr);
    int     start_force_ip(const uint8_t *mac_addr, uint32_t new_ip, uint32_t new_mask, uint32_t new_gw);
    int     start_read_reg(uint32_t ip_addr, uint32_t reg_addr);
    int     start_read_mem(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_len);
    int     start_read_multireg(uint32_t ip_addr, const FI_COMM::REG_ADDR_LIST& reg_list);
    int     start_unsolicited(uint32_t ip_addr);
    int     start_write_mem(uint32_t ip_addr, uint32_t reg_addr, const QByteArray& buffer);
    int     start_write_reg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    void    stop_unsolicited(uint32_t ip_addr);

    void    stream_finish(QObject *receiver);
    void    stream_queue(QObject *receiver, QEvent *event);
    bool    stream_queue_is_empty()           { return stream_list.isEmpty(); }

    FI_COMM::FI_CAMERA_INFO *at(int index);
    FI_COMM::FI_CAMERA_INFO *takeFirst();

signals:
    void    camera_offline_status_changing(uint32_t ip_addr, bool offline);     //!< Emitted when the camera goes offline.
    void    force_ip_complete(uint32_t ip_addr);                                //!< Emitted when the FCP Force IP command completes.
    void    msg_timeout(uint32_t ip_addr, uint32_t reg_addr, bool read_op);     //!< Emitted when a FCP register access times out.
    void    read_reg_complete(uint32_t ip_addr, uint32_t reg_addr);             //!< Emitted when a FCP Read Register command completes.
    void    unsol_socket_ready();                                               //!< Emitted when a FMP socket is ready to receive messages.
    void    unsol_msg_received(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);    //!< Emitted when we got a FMP message.
    void    write_reg_complete(uint32_t ip_addr, uint32_t reg_addr);            //!< Emitted when a FCP Write Register command completes.
    void    error(uint32_t ip_addr, int err_code);                              //!< Passing the control protocol error.

private slots:
    void    delay_retry();
    void    handle_ctrl_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    void    error_received(bool fcp_error, int ip_addr);
    void    message_received(int ip_addr);
    void    message_timeout(int ip_addr);
    void    process_write_group(uint32_t ip_addr);
    void    read_settings();
    void    write_settings();

private:
    FCPSocket *add_socket(uint32_t ip_addr);
    RegCache *get_cache(uint32_t ip_addr);
    void    handle_read_reg(uint32_t ip_addr, fcp_msg *p_msg);
    void    handle_read_mem_ack(uint32_t ip_addr, fcp_msg *p_msg);
    void    handle_write_reg_ack(uint32_t ip_addr, fcp_msg *p_msg);
    void    handle_write_mem_ack(uint32_t ip_addr, fcp_msg *p_msg);
    void    init_cache();
    void    process_discovery_ack(fcp_msg *p_msg);
    void    process_ack(uint32_t ip_addr, quint16 rcv_cmd, fcp_msg *p_msg);
    void    update_lock(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    bool    we_own_lock(uint32_t ip_addr);

private:
    QMutex          fcp_map_mutex;                  //!< Mutex to guard access to the fcp_map.
    QMap<uint32_t, FCPSocket *>    fcp_map;         //!< A map of IP [address, socket ptr] pairs for cameras we're talking with.

    uint32_t    pending_reg_addr;                   //!< Register address of the pending operation.

    int         write_retries;                      //!< Number of times we've tried to perform a write.
    int         num_buffers;                        //!< Number of buffers to use for streaming.
    int         backoff_level;                      //!< Point at which we need to pause the stream.
    int         default_packet_delay;               //!< Default inter-packet delay (microseconds).

    uint32_t    retry_ip_addr;
    uint32_t    retry_reg_addr;
    uint32_t    retry_write_value;
    bool        retry_write_lock;

    QMap<uint32_t, RegCache *>              camera_reg_cache;       //!< A map of [ip_addr, reg_cache] pairs.
    QList<FI_COMM::FI_CAMERA_INFO *>        discover_camera_list;   //!< List of cameras we heard from in the last "discover" phase.
    QList<FI_COMM::STREAM_EVENT *>          stream_list;            //!< A list of camera objects and event information waiting for their turn to start streaming.
    QMap<uint32_t, FI_COMM::WRITE_QUEUE *>  write_map;              //!< A map of [ip_addr, WRITE_QUEUE] values;
    QMap<uint32_t, FI_COMM::REG_VALUE *>    write_status;           //!< A map of [ip_addr, status_value] pairs.

    static REG_CACHE::CACHE_LIST cache_init_list;    //!< A list of REG_CACHE::CACHE_VALUE entries to initialize our cache.
};

#endif // FI_COMM_H
