/*!
 * @file fi_countdown.cpp
 * @brief This file contains the implementation of the fiCountDown class.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QTime>

#include "camera_mdi.h"
#include "fi_countdown.h"
#include "logger.h"

namespace FI_COUNTDOWN
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! This is how frequently we update.
constexpr int COUNTDOWN_INTERVAL = 1000;

//! Our display is hours, minutes, seconds.
const QString   COUNTDOWN_FORMAT = QStringLiteral("hh:mm:ss");
}

using namespace FI_COUNTDOWN;

FiCountDown::FiCountDown(QWidget *parent) :
    QLCDNumber(parent),
    do_repeat(true),
    time_start(0),
    time_left(0),
    timer_id(0),
    camera_ptr(nullptr)
{
    initialize();
}

FiCountDown::FiCountDown(int secs, QWidget *parent) :
    QLCDNumber(parent),
    do_repeat(true),
    time_start(secs),
    time_left(secs),
    timer_id(0),
    camera_ptr(nullptr)
{
    initialize();
}

FiCountDown::~FiCountDown()
{
    // Nothing to do here.
}

/*!
 * @brief Associate the control with a camera.
 * @param camera_ptr    Camera we're interested in. nullptr if we're disconnecting.
 */
void FiCountDown::setCamera(CameraMDI *camera_ptr)
{
    // Remember our camera.
    this->camera_ptr = camera_ptr;
}

/*!
 * @brief A utility function to handle constructor initialization.
 *
 * This way I could combine all the common initialization code for use
 * amoungst the multiple constructors.
 */
void FiCountDown::initialize()
{
    // Make sure we have enough digits to display.
    setDigitCount(COUNTDOWN_FORMAT.length());

    // Initialize the display.
    QTime display_time(0, 0, 0);
    display_time = display_time.addSecs(time_left);
    display(display_time.toString(COUNTDOWN_FORMAT));
}

/*!
 * @brief Set the starting amount of time. Optionally, start
 *        the timer counting.
 *
 * @param secs      The number of seconds.
 * @param start     [OPTIONAL] Automatically start the timer.
 */
void FiCountDown::setTime(int secs, bool autostart)
{
    // Kill the timer if it's running.
    if (timer_id)
    {
        killTimer(timer_id);
        timer_id = 0;
    }

    // Save the amount of time.
    time_start = secs;
    time_left = secs;

    // Update our display.
    QTime display_time(0, 0, 0);
    display_time = display_time.addSecs(time_left);
    display(display_time.toString(COUNTDOWN_FORMAT));

    // A value of 0 means to stop. So if it's not zero, did the
    // user want to start?
    if ((secs != 0) && autostart)
    {
        // Start the timer.
        start();
    }
}

/*!
 * @brief Resume the countdown at a specific value.
 * @param secs  Time value to resume the countdown at.
 */
void FiCountDown::resume(int secs)
{
    // Reload the time left.
    time_left = secs;

    // Update our display.
    QTime display_time(0, 0, 0);
    display_time = display_time.addSecs(time_left);
    display(display_time.toString(COUNTDOWN_FORMAT));

    // Start the timer.
    start();
}

/*!
 * @brief Restart the timer. This will stop the current timer, reload
 *        the countdown with the original setting, and restart the timer.
 */
void FiCountDown::restart()
{
    // Stop the timer.
    stop();

    // Reload the time left.
    // Reload the time left.
    if (camera_ptr != nullptr)
        time_left = camera_ptr->get_tt_base();
    else
        time_left = time_start;

    // Start the timer.
    start();
}

/*!
 * @brief Kill any timer which might already be running, and start
 *        the countdown timer.
 */
void FiCountDown::start()
{
    // We only want one timer event, so if a timer is already
    // running, stop it.
    if (timer_id)
    {
        killTimer(timer_id);
        timer_id = 0;
    }

    // Sanity check.
    if (time_start != 0)
        timer_id = startTimer(COUNTDOWN_INTERVAL);
}

/*!
 * @brief Stop the timer and reset the display to 00:00:00.
 */
void FiCountDown::stop()
{
    // Stop the timer.
    killTimer(timer_id);
    timer_id = 0;

    // Clear our time value.
    time_left = 0;

    // Update the display.
    QTime display_time(0, 0, 0);
    display(display_time.toString(COUNTDOWN_FORMAT));
}

/*!
 * @brief Handle the timeout event.
 * @param event     [UNUSED] The event object.
 */
void FiCountDown::timerEvent(QTimerEvent *event)
{
    Q_UNUSED(event)

    // Do we have a camera?
    if (camera_ptr != nullptr)
    {
        // Use the camera's information.
        time_left = camera_ptr->get_tt_remaining();
    }
    else
    {
        // Sanity check.
        if (time_left <= 0)
            return;

        // Decrement our remaining time.
        --time_left;
    }

    // Update our display.
    QTime display_time(0, 0, 0);
    display_time = display_time.addSecs(time_left);
    display(display_time.toString(COUNTDOWN_FORMAT));

    // Turn off the timer if we hit 0.
    if ((camera_ptr == nullptr) && (time_left == 0))
    {
        // Let folks know we hit 0.
        emit expired();

        // Are we supposed to restart?
        if (do_repeat)
        {
            // Reload our counter.
            time_left = time_start;
        }
        else if (camera_ptr != nullptr)
        {
            // Stop the timer event.
            this->killTimer(timer_id);
            timer_id = 0;
        }
    }
}
