/*!
 * @file fi_display_lut.h
 * @brief This is the header file for the FiDisplayLut class.
 * It is responsible for all supporting functions for display
 * LUTs.  Their creation and their manipulation.
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */


//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "gige_msg.h"
#include "logger.h"

#include "fi_display_lut.h"
#include "fi_lutfile.h"

namespace DISPLAY_LUT
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The number of entries in our RGB lookup tables.
const int   LUT_LENGTH                  = 256;

//! The maximum lut id value.
const int SPLINE_RECS_MAX_LUT_ID        = GIGE_LUT_MAX;

//! A name to use as the spline record description for a metadata LUT.
const QString SPLINE_RECS_METADATA_DESC = QStringLiteral("Metadata LUT");

//! A description for an empty non-existant lut.
const QString EMPTY_CUSTOM_LUT_DESC = QStringLiteral("Empty LUT");

//! Name for the red or gamma spline widget.
const QString NAME_SPLINE_WIDGET_RED_OR_GAMMA = QStringLiteral("Red_or_Gamma fiSplineWidget");

//! Name for the red or gamma spline widget.
const QString NAME_SPLINE_WIDGET_GREEN = QStringLiteral("Green_fiSplineWidget");

//! Name for the red or gamma spline widget.
const QString NAME_SPLINE_WIDGET_BLUE = QStringLiteral("Blue_fiSplineWidget");

//! Verbose logging control.  Set true for verbose debug logging.
const bool VERBOSE_LOGGING = false;

}   // End of namespace CURVE_SETTINGS.

// Make sure we use our namespace.
using namespace DISPLAY_LUT;

/*!
 * @brief The class constructor.
 */
FiDisplayLut::FiDisplayLut()
{
    builtRedOrGammaLut  = false;
    builtGreenLut       = false;
    builtBlueLut        = false;
    usingColor          = false;
}

/*!
 * @brief The class destructor.
 */
FiDisplayLut::~FiDisplayLut()
{
}

/*!
 * @brief Factory to create a FiDisplayLut object.
 *
 * If there are no knots this still generates an default linear spline record with 3 knots.
 *
 * @param isMonoSensor          True if the image data is from monochrome sensor, false if not.
 * @param isSeries3Camera       True if the image data is from series 3 camera, false if not.
 * @param metadataLutId         The one-based radio LUT id to set for dialog or zero if none.
 * @param metadataLutKnots      The metadata lut knots from the pm_metadata structure.
 *
 * @returns the std::shared_ptr to the created FiDisplayLut object.
 */
std::shared_ptr<FiDisplayLut>
FiDisplayLut::CreateDisplayLut(bool isMonoSensor,
                               bool isSeries3Camera,
                               uint32_t metadataLutId,
                               const QVector<uint16_t>& metadataLutKnots)
{
    bool initializedOk  = false;

    std::shared_ptr<FiDisplayLut> fiDisplayLut =
        std::shared_ptr<FiDisplayLut>(new FiDisplayLut());

    if (fiDisplayLut)
    {
        // Determine if this is set up to use color or not.
        fiDisplayLut->usingColor = (isMonoSensor || isSeries3Camera) ? false : true;
        fiDisplayLut->lutSplineRec = FiDisplayLut::XlateMetadataKnots(isMonoSensor,
                                                                      isSeries3Camera,
                                                                      metadataLutId,
                                                                      metadataLutKnots);

        // If unable to get the lut spline record pop out a warning and exit.
        if (!fiDisplayLut->lutSplineRec)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate lutSplineRec.";
            goto exitCreateDisplayLut;
        }

        // Create the red spline widget.
        if (!fiDisplayLut->CreateSplineWidget(DISPLAY_LUT::CURVE_ID::RED_GAMMA,
                                              &fiDisplayLut->lutSplineRec->redOrGammaKnots))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate fiSplineRedOrGamma.";
            goto exitCreateDisplayLut;
        }

        // Build the red/gamma camera LUT.
        if (!fiDisplayLut->BuildRedOrGammaCameraLut())
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to build red/gamma camera lut.";
            goto exitCreateDisplayLut;
        }

        // Only if using color do we need spline widgets for green and blue.
        if (fiDisplayLut->usingColor)
        {
            // Create the green spline widget.
            if (!fiDisplayLut->CreateSplineWidget(DISPLAY_LUT::CURVE_ID::GREEN,
                                                  &fiDisplayLut->lutSplineRec->greenKnots))
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate fiSplineGreen.";
                goto exitCreateDisplayLut;
            }

            // Create the blue spline widget.
            if (!fiDisplayLut->CreateSplineWidget(DISPLAY_LUT::CURVE_ID::BLUE,
                                                  &fiDisplayLut->lutSplineRec->blueKnots))
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate fiSplineBlue.";
                goto exitCreateDisplayLut;
            }

            // Build the green camera LUT.
            if (!fiDisplayLut->BuildGreenCameraLut())
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to build green camera lut.";
                goto exitCreateDisplayLut;
            }

            // Build the blue camera LUT.
            if (!fiDisplayLut->BuildBlueCameraLut())
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to build blue camera lut.";
                goto exitCreateDisplayLut;
            }
        }
    }
    else
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate fiDisplayLut.";
        goto exitCreateDisplayLut;
    }

    // If make it here all should have been properly initialized.
    initializedOk = true;

exitCreateDisplayLut:

    // If not initialized properly we don't want to return partial creation.
    if (!initializedOk)
    {
        if (fiDisplayLut)
        {
            fiDisplayLut = nullptr;
        }
    }

    return fiDisplayLut;
}

/*!
 * @brief Common spline creation logic.
 *
 * @param whichCurve        [Input]Which curve is being created.
 * @param splineKnots       [Input]The vector of QPoints containing the knots.
 *
 * @returns True if the curve objects were created, false if error occurred.
 */
bool FiDisplayLut::CreateSplineWidget(DISPLAY_LUT::CURVE_ID whichCurve,
                                      QVector<QPoint> *splineKnots)
{
    bool createdOk = false;

    std::unique_ptr<fiSplineWidget> fiSplineWidgetCreated;
    std::shared_ptr<aaAaa::aaSpline> splineCurve = nullptr;
    QString splineName;

    switch (whichCurve)
    {
    case DISPLAY_LUT::CURVE_ID::RED_GAMMA:
    {
        splineName = NAME_SPLINE_WIDGET_RED_OR_GAMMA;
        aaSplineRedOrGamma = std::shared_ptr<aaAaa::aaSpline>(new aaAaa::aaSpline());
        splineCurve = aaSplineRedOrGamma;
        break;
    }
    case DISPLAY_LUT::CURVE_ID::GREEN:
    {
        splineName = NAME_SPLINE_WIDGET_GREEN;
        aaSplineGreen = std::shared_ptr<aaAaa::aaSpline>(new aaAaa::aaSpline());
        splineCurve = aaSplineGreen;
        break;
    }
    case DISPLAY_LUT::CURVE_ID::BLUE:
    {
        splineName = NAME_SPLINE_WIDGET_BLUE;
        aaSplineBlue = std::shared_ptr<aaAaa::aaSpline>(new aaAaa::aaSpline());
        splineCurve = aaSplineBlue;
        break;
    }
    default:
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Invalid curve id provided:" << whichCurve;
        goto exitCreateSplineWidget;
    }
    }

    // Was the spline curve allocated?.
    if (!splineCurve)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "unable to allocate aaAaa::aaSpline for:" << splineName;
        goto exitCreateSplineWidget;
    }

    // Set up the spline curve's attributes.
    splineCurve->name = splineName.toStdString();
    splineCurve->type = aaAaa::aaSpline::SPLINE_CUBIC;

    // Create the spline widget.
    fiSplineWidgetCreated = std::unique_ptr<fiSplineWidget>(new fiSplineWidget());
    if (fiSplineWidgetCreated)
    {
        fiSplineWidgetCreated->setObjectName(splineName);

        QSizePolicy sizePolicy(QSizePolicy::Preferred, QSizePolicy::Preferred);
        sizePolicy.setHorizontalStretch(0);
        sizePolicy.setVerticalStretch(0);
        sizePolicy.setHeightForWidth(fiSplineWidgetCreated->sizePolicy().hasHeightForWidth());
        fiSplineWidgetCreated->setSizePolicy(sizePolicy);
        fiSplineWidgetCreated->setMinimumSize(QSize(0, 0));
        fiSplineWidgetCreated->setMaximumSize(QSize(16777215, 16777215));

        // Clear any old knots in the spline curve.
        splineCurve->knots.clear();

        // Force the spline to limit itself to our LUT camera bounds.
        splineCurve->setLimit(0, LUT_LENGTH - 1, LUT_MAX_TABLE_COUNT - 1, 0);

        // Is there a custom knot configuration?
        if (splineKnots)
        {
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Using splineKnots to create splineCurve for" << splineName;
            }
            // Transfer each knot into a spline point.
            // Note: It is assumed the knots are validated knots based on the internal
            // collection established when created in the curve settings factory.
            for (int knotId = 0; knotId < splineKnots->count(); knotId++)
            {
                splineCurve->addKnots(aaAaa::aaPoint((float)splineKnots->at(knotId).x(), (float)splineKnots->at(knotId).y()));
                if (VERBOSE_LOGGING)
                {
                    QLOG_DEBUG() << "[" << splineKnots->at(knotId).x() << "," << splineKnots->at(knotId).y() << "]";
                }
            }
        }
        else
        {
            if (VERBOSE_LOGGING)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Using default linear line to create splineCurve for" << splineName;
            }

            // Add knots at the beginning, middle, and end creating a linear line.
            splineCurve->addKnots(aaAaa::aaPoint(0, 0));
            splineCurve->addKnots(aaAaa::aaPoint(LUT_MAX_TABLE_COUNT / 2, LUT_LENGTH / 2));
            splineCurve->addKnots(aaAaa::aaPoint(LUT_MAX_TABLE_COUNT - 1, LUT_LENGTH - 1));
        }

        // Set the resolution of our spline.
        fiSplineWidgetCreated->set_increment(0.1);

        // Pass the data on to our control.
        fiSplineWidgetCreated->set_spline(splineCurve);

        // Make sure we limit the maximum number of knots on the spline.
        fiSplineWidgetCreated->set_max_knot_count(LUT_MAX_KNOT_COUNT);

        // Now install our spline widget into our object.
        switch (whichCurve)
        {
        case DISPLAY_LUT::CURVE_ID::RED_GAMMA:
        {
            fiSplineRedOrGamma = std::move(fiSplineWidgetCreated);
            break;
        }
        case DISPLAY_LUT::CURVE_ID::GREEN:
        {
            fiSplineGreen = std::move(fiSplineWidgetCreated);
            break;
        }
        case DISPLAY_LUT::CURVE_ID::BLUE:
        {
            fiSplineBlue = std::move(fiSplineWidgetCreated);
            break;
        }
        default:
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Invalid curve id provided:" << whichCurve;
            goto exitCreateSplineWidget;
        }
        }

        // Make it here and all is good.
        createdOk = true;
    }

exitCreateSplineWidget:
    return createdOk;
}

/*!
 * @brief Creates a Display LUT using information provided by caller.
 *
 * Note:  The purpose for this interface is to provide the curve settings
 * dialog with a way to take its internal data and using an existing
 * FiDisplayLut object duplicate the information providing new settings
 * most importantly with a camera lut already built.
 *
 * @param lutSplineRecToUse     The LUT_SPLINE_RECORD data.
 * @param cameraLutRedOrGamma   The generated red/gamma camera lut matching thelut spline record knots.
 * @param cameraLutGreen        The generated green camera lut matching thelut spline record knots.
 * @param cameraLutBlue         The generated blue camera lut matching thelut spline record knots.
 *
 * @returns Valid shared pointer to the duplicated display lut or nullptr if error occurred.
 */
std::shared_ptr<FiDisplayLut> FiDisplayLut::CreateDisplayLut(const DISPLAY_LUT::LUT_SPLINE_RECORD& lutSplineRecToUse,
                                                             const QVector<double>& cameraLutRedOrGammaToUse,
                                                             const QVector<double>& cameraLutGreenToUse,
                                                             const QVector<double>& cameraLutBlueToUse)
{
    bool duplicatedOk = false;
    std::shared_ptr<FiDisplayLut> dupDisplayLut;

    // Only able to do minimal evaluation on data provided.  This was created in order to eliminate
    // the need for a build of the camera luts for the new spline record data when trying to
    // persist settings beyond the use of the curve settings dialog.  In the future it would make
    // more sense if we could leverage the object inside of curve settings and just return that
    // object.  In other words we would pass one of these to the curve settings dialog and it
    // would operate on that object directly.

    dupDisplayLut = std::shared_ptr<FiDisplayLut>(new FiDisplayLut);

    if (!dupDisplayLut)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate duplicate FiDisplayLut.";
        goto exitDuplicateDisplayLut;
    }

    // Generate the lut spline record holding the translation.
    dupDisplayLut->lutSplineRec =
        std::unique_ptr<struct DISPLAY_LUT::LUT_SPLINE_RECORD>(new struct DISPLAY_LUT::LUT_SPLINE_RECORD);

    // Make sure the lut spline record was allocated.
    if (!dupDisplayLut->lutSplineRec)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate duplicate dupDisplayLut->lutSplineRec.";
        goto exitDuplicateDisplayLut;
    }

    // Make sure the spline record has a valid id.
    if ((lutSplineRecToUse.id < 0) || (lutSplineRecToUse.id > SPLINE_RECS_MAX_LUT_ID))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The lutSplineRecToUse.id was out of range.";
        goto exitDuplicateDisplayLut;
    }

    // Transfer the spline record information.
    dupDisplayLut->lutSplineRec->id                = lutSplineRecToUse.id;
    dupDisplayLut->lutSplineRec->desc              = lutSplineRecToUse.desc;
    dupDisplayLut->lutSplineRec->localPath         = lutSplineRecToUse.localPath;
    dupDisplayLut->lutSplineRec->lutDataType       = lutSplineRecToUse.lutDataType;
    dupDisplayLut->lutSplineRec->redOrGammaKnots   = lutSplineRecToUse.redOrGammaKnots;
    dupDisplayLut->lutSplineRec->greenKnots        = lutSplineRecToUse.greenKnots;
    dupDisplayLut->lutSplineRec->blueKnots         = lutSplineRecToUse.blueKnots;

    // Create the red spline data.
    if (!dupDisplayLut->CreateSplineWidget(DISPLAY_LUT::CURVE_ID::RED_GAMMA,
                                           &dupDisplayLut->lutSplineRec->redOrGammaKnots))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The dupDisplayLut->fiSplineRedOrGamma wasn't created.";
        goto exitDuplicateDisplayLut;
    }

    // We can only really check the size of the supplied camera lut for red/gamma.
    if (cameraLutRedOrGammaToUse.count() != LUT_MAX_TABLE_COUNT)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The cameraLutRedOrGammaToUse size != LUT_MAX_TABLE_COUNT.";
        goto exitDuplicateDisplayLut;
    }

    dupDisplayLut->cameraLutRedOrGamma = cameraLutRedOrGammaToUse;
    dupDisplayLut->builtRedOrGammaLut = true;   // Assume this is valid.

    // Only if using color do we need spline widgets for green and blue.
    if (lutSplineRecToUse.lutDataType == LUT_SERIESNC_COLOR)
    {
        // Create the green spline data.
        if (!dupDisplayLut->CreateSplineWidget(DISPLAY_LUT::CURVE_ID::GREEN,
                                               &dupDisplayLut->lutSplineRec->greenKnots))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The dupDisplayLut->fiSplineGreen wasn't created.";
            goto exitDuplicateDisplayLut;
        }

        // Create the blue spline data.
        if (!dupDisplayLut->CreateSplineWidget(DISPLAY_LUT::CURVE_ID::BLUE,
                                               &dupDisplayLut->lutSplineRec->blueKnots))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The dupDisplayLut->fiSplineBlue wasn't created.";
            goto exitDuplicateDisplayLut;
        }

        // We can only really check the size of the supplied camera lut for validity.
        if (cameraLutGreenToUse.count() != LUT_MAX_TABLE_COUNT)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The cameraLutGreenToUse size != LUT_MAX_TABLE_COUNT.";
            goto exitDuplicateDisplayLut;
        }

        dupDisplayLut->cameraLutGreen = cameraLutGreenToUse;
        dupDisplayLut->builtGreenLut = true;   // Assume this is valid.

        // We can only really check the size of the supplied camera lut for validity.
        if (cameraLutBlueToUse.count() != LUT_MAX_TABLE_COUNT)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The cameraLutBlueToUse size != LUT_MAX_TABLE_COUNT.";
            goto exitDuplicateDisplayLut;
        }

        dupDisplayLut->cameraLutBlue = cameraLutBlueToUse;
        dupDisplayLut->builtBlueLut = true;   // Assume this is valid.
    }

    // Make it here and all duplication efforts were a success.
    duplicatedOk = true;

exitDuplicateDisplayLut:

    // If the duplication was not ok delete the duplicated display lut if started.
    if (!duplicatedOk)
    {
        if (dupDisplayLut)
        {
            dupDisplayLut = nullptr;
        }
    }
    return dupDisplayLut;
}

/*!
 * @brief Determines if all knots are equal in a spline record.
 * Note:  If all knots are empty they are still considered equal
 * the caller must test for that case outside of this function.
 *
 * @returns True if all are equal, false if there are differences.
*/
bool FiDisplayLut::KnotsAreEqual(const DISPLAY_LUT::LUT_SPLINE_RECORD& lutSplineRecToTest)
{
    // First just check the counts.
    if (lutSplineRecToTest.redOrGammaKnots.count() == lutSplineRecToTest.greenKnots.count())
    {
        // The red and green knots are equal are the red and blue knots equal?
        if (lutSplineRecToTest.redOrGammaKnots.count() != lutSplineRecToTest.blueKnots.count())
        {
            return false;    // They are not equal if counts aren't the same.
        }
    }
    else
    {
        // The red and green knots not equal we are done here.
        return false;
    }

    // The counts are equal.  Are the contents?
    for (int x = 0; x < lutSplineRecToTest.redOrGammaKnots.count(); x++)
    {
        // Are the red and green points equal?
        if ((lutSplineRecToTest.redOrGammaKnots.at(x).x() != lutSplineRecToTest.greenKnots.at(x).x()) ||
            (lutSplineRecToTest.redOrGammaKnots.at(x).y() != lutSplineRecToTest.greenKnots.at(x).y()))
        {
            return false;
        }
        else if ((lutSplineRecToTest.redOrGammaKnots.at(x).x() != lutSplineRecToTest.blueKnots.at(x).x()) ||
                 (lutSplineRecToTest.redOrGammaKnots.at(x).y() != lutSplineRecToTest.blueKnots.at(x).y()))
        {
            return false;
        }
    }

    // If make it here all is knots are the same.
    return true;
}

/*!
 * @brief Return the id for the custom lut data.
 * @returns A non-zero id if there is custom lut data.
*/
int FiDisplayLut::GetLutId()
{
    int lutId = 0;

    if (lutSplineRec)
    {
        lutId = lutSplineRec->id;
    }
    return lutId;
}

/*!
 * @brief Returns the description for the custom lut data.
*/
const QString& FiDisplayLut::GetLutDesc()
{
    if (lutSplineRec)
    {
        return (lutSplineRec->desc);
    }
    return EMPTY_CUSTOM_LUT_DESC;
}

/*!
 * @brief Build a camera lookup table for blue channel.
 *
*/
bool FiDisplayLut::BuildBlueCameraLut()
{
    builtBlueLut = false;

    // Make sure the spline widget provided is a valid pointer.
    if (fiSplineBlue == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fiSplineBlue was a nullptr!";
        return builtBlueLut;
    }

    // Make sure the spline curve provided is a valid pointer.
    if (aaSplineBlue == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "aaSplineBlue was a nullptr!";
        return builtBlueLut;
    }

    // Make sure the camera LUT is large enough
    cameraLutBlue.resize(LUT_MAX_TABLE_COUNT);

    // Read the LUT information from the spline widget using the spline curves.
    if (fiSplineBlue->get_data((int)(aaSplineBlue->limit_right + 1) / LUT_MAX_TABLE_COUNT, cameraLutBlue) < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to read the LUT from " << fiSplineBlue->objectName() << " control!";
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Created blue camera lut.";
        builtBlueLut = true;
    }

    return builtBlueLut;
}

/*!
 * @brief Build a camera lookup table for blue channel.
 *
*/
bool FiDisplayLut::BuildGreenCameraLut()
{
    builtGreenLut = false;

    // Make sure the spline widget provided is a valid pointer.
    if (fiSplineGreen == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fiSplineGreen was a nullptr!";
        return builtGreenLut;
    }

    // Make sure the spline curve provided is a valid pointer.
    if (aaSplineGreen == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "aaSplineGreen was a nullptr!";
        return builtGreenLut;
    }

    // Make sure the camera LUT is large enough
    cameraLutGreen.resize(LUT_MAX_TABLE_COUNT);

    // Read the LUT information from the spline widget using the spline curves.
    if (fiSplineGreen->get_data((int)(aaSplineGreen->limit_right + 1) / LUT_MAX_TABLE_COUNT, cameraLutGreen) < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to read the LUT from " << fiSplineGreen->objectName() << " control!";
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Created green camera lut.";
        builtGreenLut = true;
    }

    return builtGreenLut;
}

/*!
 * @brief Build a camera lookup table for red/gamma channel.
 *
*/
bool FiDisplayLut::BuildRedOrGammaCameraLut()
{
    builtRedOrGammaLut = false;

    // Make sure the spline widget provided is a valid pointer.
    if (fiSplineRedOrGamma == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fiSplineRedOrGamma was a nullptr!";
        return builtRedOrGammaLut;
    }

    // Make sure the spline curve provided is a valid pointer.
    if (aaSplineRedOrGamma == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "aaSplineRedOrGamma was a nullptr!";
        return builtRedOrGammaLut;
    }

    // Make sure the camera LUT is large enough
    cameraLutRedOrGamma.resize(LUT_MAX_TABLE_COUNT);

    // Read the LUT information from the spline widget using the spline curves.
    if (fiSplineRedOrGamma->get_data((int)(aaSplineRedOrGamma->limit_right + 1) / LUT_MAX_TABLE_COUNT, cameraLutRedOrGamma) < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to read the LUT from " << fiSplineRedOrGamma->objectName() << " control!";
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Created red/gamma camera lut.";
        builtRedOrGammaLut = true;
    }

    return builtRedOrGammaLut;
}

/*!
 * @brief Retrieve a reference to the blue camera lut.
 * Note:  Test the setting in builtBlueCameraLut to determine if the
 * camera lut reference is valid.
 *
 * @param builtBlueCameraLut    Sets reference true if built the lut or false if not.
 *
 * @returns A reference to the blue camera lut.
 */
const QVector<double>& FiDisplayLut::GetBlueCameraLut(bool& builtBlueCameraLut)
{
    const QVector<double>& theCameraLut = cameraLutBlue;
    builtBlueCameraLut = builtBlueLut;
    return theCameraLut;
}

/*!
 * @brief Retrieve a reference to the green camera lut.
 * Note:  Test the setting in builtGreenCameraLut to determine if the
 * camera lut reference is valid.
 *
 * @param builtGreenCameraLut    Sets reference true if built the lut or false if not.
 *
 * @returns A reference to the green camera lut.
 */
const QVector<double>& FiDisplayLut::GetGreenCameraLut(bool& builtGreenCameraLut)
{
    const QVector<double>& theCameraLut = cameraLutGreen;
    builtGreenCameraLut = builtGreenLut;
    return theCameraLut;
}

/*!
 * @brief Retrieve a reference to the red or gamma camera lut.
 * Note:  Test the setting in builtRedOrGammaCameraLut to determine if the
 * camera lut reference is valid.
 *
 * @param builtRedOrGammaCameraLut    Sets reference true if built the lut or false if not.
 *
 * @returns A reference to the red or gamma 10-bit camera lut.
 */
const QVector<double>& FiDisplayLut::GetRedOrGammaCameraLut(bool& builtRedOrGammaCameraLut)
{
    const QVector<double>& theCameraLut = cameraLutRedOrGamma;
    builtRedOrGammaCameraLut = builtRedOrGammaLut;
    return theCameraLut;
}

/*!
 * @brief Copies the contents of the internal LUT_SPLINE_RECORD into the reference
 * passed in by the caller.
 *
 * @param lutSplineRecCopy    <Output> Copies the contents of this object's lutSplineRec
 * into this reference.
 *
 * @returns True if the lutSplineRecCopy is valid and ok to use.
 */
bool FiDisplayLut::CopySplineRecord(DISPLAY_LUT::LUT_SPLINE_RECORD& lutSplineRecCopy)
{
    bool goodCopy = false;
    if (lutSplineRec)
    {
        lutSplineRecCopy.id                 = lutSplineRec->id;
        lutSplineRecCopy.desc               = lutSplineRec->desc;
        lutSplineRecCopy.localPath          = lutSplineRec->localPath;
        lutSplineRecCopy.lutDataType        = lutSplineRec->lutDataType;
        lutSplineRecCopy.redOrGammaKnots    = lutSplineRec->redOrGammaKnots;
        lutSplineRecCopy.greenKnots         = lutSplineRec->greenKnots;
        lutSplineRecCopy.blueKnots          = lutSplineRec->blueKnots;
        goodCopy = true;
    }
    else
    {
        lutSplineRecCopy.id = 0;    // No custom lut information exists.
    }
    return goodCopy;
}

/*!
 * @brief Factory to create a DISPLAY_LUT::LUT_SPLINE_RECORD using the knots
 * supplied by playback (pm_metadata) lut information.
 *
 * If there are no knots this still generates an default linear spline record with 3 knots.
 *
 * @param isMonoSensor          True if the image data is from monochrome sensor, false if not.
 * @param isSeries3Camera       True if the image data is from series 3 camera, false if not.
 * @param metadataLutId         The one-based radio LUT id to set for dialog or zero if none.
 * @param metadataLutKnots      The metadata lut knots from the pm_metadata structure.
 *
 * @returns the std::unique_ptr for the LUT_SPLINE_RECORD data.
 */
std::unique_ptr<struct DISPLAY_LUT::LUT_SPLINE_RECORD>
FiDisplayLut::XlateMetadataKnots(bool isMonoSensor,
                                 bool isSeries3Camera,
                                 uint32_t metadataLutId,
                                 const QVector<uint16_t>& metadataLutKnots)
{
    // What is our maximum point position.
    int maxX = LUT_MAX_TABLE_COUNT - 1;
    int maxY = LUT_LENGTH - 1;

    // Create the endpoints.
    QPoint  firstKnot(0, 0);
    QPoint  lastKnot(maxX, maxY);
    QPoint  midKnot(LUT_MAX_TABLE_COUNT / 2, LUT_LENGTH / 2);

    // Set the indices into the vector for red, green and blue knot channels.
    int red = 0;
    int green = red + (LUT_MAX_KNOT_COUNT * 2);
    int blue = green + (LUT_MAX_KNOT_COUNT * 2);

    // Generate the lut spline record holding the translation.
    std::unique_ptr<struct DISPLAY_LUT::LUT_SPLINE_RECORD> lutSplineRec =
            std::unique_ptr<struct DISPLAY_LUT::LUT_SPLINE_RECORD>(new struct DISPLAY_LUT::LUT_SPLINE_RECORD);

    // Make sure it is within our range.
    if (metadataLutId > DISPLAY_LUT::SPLINE_RECS_MAX_LUT_ID)
    {
        // Confine the metadataLutId to within our range of possible lut ids.
        metadataLutId = DISPLAY_LUT::SPLINE_RECS_MAX_LUT_ID;
    }

    // Install the default description for our spline record.
    lutSplineRec->desc = SPLINE_RECS_METADATA_DESC;

    // Set up the type of spline record based on the information provided about the sensor.
    if (isMonoSensor && !isSeries3Camera)
    {
        lutSplineRec->lutDataType = LUT_SERIESNC_MONO;
    }
    else if (!isMonoSensor && !isSeries3Camera)
    {
        lutSplineRec->lutDataType = LUT_SERIESNC_COLOR;
    }
    else
    {
        lutSplineRec->lutDataType = LUT_SERIES3;
    }

    // Install the metadata Lut Id.
    lutSplineRec->id = (int)metadataLutId;

    // Note:  If there were no knots provided this logic creates a linear line
    // for the spline knots.

    // Install the first knot which is *the* first endpoint at (0,0).
    lutSplineRec->redOrGammaKnots.append(firstKnot);
    lutSplineRec->greenKnots.append(firstKnot);
    lutSplineRec->blueKnots.append(firstKnot);

    // Can only copy over the knots if the metadata lut id is non-zero
    // otherwise we can safely assume any data in the metadata knot arrays
    // is suspect and we don't use it.
    if (metadataLutId > 0)
    {
        // Copy over the knots.
        // The camera server sets the pm_metadata lut array in channel order which means
        // the red knots for (LUT_MAX_KNOT_COUNT) in x,y uint16_t pairs up followed by
        // the green knots then the blue knots.  We will pull them all at once.
        for (int knotId = 0; knotId < LUT_MAX_KNOT_COUNT; knotId++)
        {
            // Pull in knot pairs stored as "spline_knot" structure in fi_lutfile.h
            // for count of LUT_MAX_KNOT_COUNT in channel order (red, green, blue).
            // For reference see part_mgr.cpp in the camera_server (PartMgr::setknots)
            // Note:  camera_server initially fills this data with zeroes and only sets
            // valid knot data for the red/gamma channel if not color HS camera.
            QPoint redKnot;
            QPoint greenKnot;
            QPoint blueKnot;

            // Pull the Xs.
            redKnot.setX(metadataLutKnots[red++]);
            greenKnot.setX(metadataLutKnots[green++]);
            blueKnot.setX(metadataLutKnots[blue++]);

            // Pull the Ys.
            redKnot.setY(metadataLutKnots[red++]);
            greenKnot.setY(metadataLutKnots[green++]);
            blueKnot.setY(metadataLutKnots[blue++]);

            // Append the points if they are not equal to the first knot.
            if (redKnot != firstKnot)
            {
                // Keep these knots in range.
                if (redKnot.x() > maxX)
                    redKnot.setX(maxX);
                if (redKnot.y() > maxY)
                    redKnot.setY(maxY);

                // Add if not the last knot endpoint.
                if (redKnot != lastKnot)
                {
                    lutSplineRec->redOrGammaKnots.append(redKnot);
                }
            }
            if (greenKnot != firstKnot)
            {
                // Keep these knots in range.
                if (greenKnot.x() > maxX)
                    greenKnot.setX(maxX);
                if (greenKnot.y() > maxY)
                    greenKnot.setY(maxY);

                // Add if not the last knot endpoint.
                if (greenKnot !=  lastKnot)
                {
                    lutSplineRec->greenKnots.append(greenKnot);
                }
            }
            if (blueKnot != firstKnot)
            {
                // Keep these knots in range.
                if (blueKnot.x() > maxX)
                    blueKnot.setX(maxX);
                if (blueKnot.y() > maxY)
                    blueKnot.setY(maxY);

                // Add if not the last knot endpoint.
                if (blueKnot != lastKnot)
                {
                    lutSplineRec->blueKnots.append(blueKnot);
                }
            }
        }  // End of the For Loop processing the metadata knots.
    }

    // If only the first endpoint was installed add a midpoint knot.
    if (lutSplineRec->redOrGammaKnots.count() == 1)
    {
        lutSplineRec->redOrGammaKnots.append(midKnot);
    }
    // Add the endpoint for the last knot.
    lutSplineRec->redOrGammaKnots.append(lastKnot);

    // If only the first endpoint was installed add a midpoint knot.
    if (lutSplineRec->greenKnots.count() == 1)
    {
        lutSplineRec->greenKnots.append(midKnot);
    }
    // Add the endpoint for the last knot.
    lutSplineRec->greenKnots.append(lastKnot);

    // If only the first endpoint was installed add a midpoint knot.
    if (lutSplineRec->blueKnots.count() == 1)
    {
        lutSplineRec->blueKnots.append(midKnot);
    }
    // Add the endpoint for the last knot.
    lutSplineRec->blueKnots.append(lastKnot);

    // Return the created spline record.
    return lutSplineRec;
}

/*!
 * @brief An object copy taking the contents of the display lut
 * provided and installing into this object.
 *
 * @param displayLutSource     <input> Source of copy.  This object
 * is used as a source for copying into our object.
 *
 * @returns True if the data was accepted and the object was modified.
 */
bool FiDisplayLut::CopyDisplayLut(std::shared_ptr<FiDisplayLut> displayLutSource)
{
    bool dataCopiedOk = false;

    // Do we have a lut spline record?
    if (!lutSplineRec)
    {
        // Allocate one.
        lutSplineRec =
            std::unique_ptr<struct DISPLAY_LUT::LUT_SPLINE_RECORD>(new struct DISPLAY_LUT::LUT_SPLINE_RECORD);
    }

    // Make sure the lut spline record was allocated.
    if (!lutSplineRec)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Missing lutSplineRec.";
        goto exitCopyDisplayLut;
    }
    // Make sure the source has a lut spline record.
    if (!displayLutSource->lutSplineRec)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Missing displayLutSource->lutSplineRec.";
        goto exitCopyDisplayLut;
    }

    // Make sure the source spline record has a valid id.
    if ((displayLutSource->lutSplineRec->id < 0) || (displayLutSource->lutSplineRec->id > SPLINE_RECS_MAX_LUT_ID))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The displayLutSource->lutSplineRec->id was out of range.";
        goto exitCopyDisplayLut;
    }

    // Transfer the spline record information.
    lutSplineRec->id                = displayLutSource->lutSplineRec->id;
    lutSplineRec->desc              = displayLutSource->lutSplineRec->desc;
    lutSplineRec->localPath         = displayLutSource->lutSplineRec->localPath;
    lutSplineRec->lutDataType       = displayLutSource->lutSplineRec->lutDataType;
    lutSplineRec->redOrGammaKnots   = displayLutSource->lutSplineRec->redOrGammaKnots;
    lutSplineRec->greenKnots        = displayLutSource->lutSplineRec->greenKnots;
    lutSplineRec->blueKnots         = displayLutSource->lutSplineRec->blueKnots;

    // Create the red/gamma spline data.
    if (!CreateSplineWidget(DISPLAY_LUT::CURVE_ID::RED_GAMMA,
                            &displayLutSource->lutSplineRec->redOrGammaKnots))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The fiSplineRedOrGamma wasn't created.";
        goto exitCopyDisplayLut;
    }

    // Check the size of the supplied camera lut for red/gamma.
    if (displayLutSource->cameraLutRedOrGamma.count() != LUT_MAX_TABLE_COUNT)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The cameraLutRedOrGamma size != LUT_MAX_TABLE_COUNT.";
        goto exitCopyDisplayLut;
    }

    // Standard QVector copy.
    cameraLutRedOrGamma = displayLutSource->cameraLutRedOrGamma;
    builtRedOrGammaLut = true;   // Assume this is valid.

    // Copy the using Color setting.
    usingColor = displayLutSource->usingColor;

    // Only if using color we need spline widgets for green and blue.
    if (!displayLutSource->usingColor)
    {
        fiSplineGreen = nullptr;
        aaSplineGreen = nullptr;
        builtGreenLut = false;

        fiSplineBlue = nullptr;
        aaSplineBlue = nullptr;
        builtBlueLut = false;
    }
    else
    {
        // Create the green spline data.
        if (!CreateSplineWidget(DISPLAY_LUT::CURVE_ID::GREEN,
                                &displayLutSource->lutSplineRec->greenKnots))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The fiSplineGreen wasn't created.";
            goto exitCopyDisplayLut;
        }

        // Create the green spline data.
        if (!CreateSplineWidget(DISPLAY_LUT::CURVE_ID::BLUE,
                                &displayLutSource->lutSplineRec->blueKnots))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The fiSplineBlue wasn't created.";
            goto exitCopyDisplayLut;
        }

        // We can only really check the size of the supplied camera lut for validity.
        if (displayLutSource->cameraLutGreen.count() != LUT_MAX_TABLE_COUNT)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The displayLutSource->cameraLutGreen size != LUT_MAX_TABLE_COUNT.";
            goto exitCopyDisplayLut;
        }

        // Standard QVector copy.
        cameraLutGreen = displayLutSource->cameraLutGreen;
        builtGreenLut = true;   // Assume this is valid.

        // We can only really check the size of the supplied camera lut for validity.
        if (displayLutSource->cameraLutBlue.count() != LUT_MAX_TABLE_COUNT)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The displayLutSource->cameraLutBlue size != LUT_MAX_TABLE_COUNT.";
            goto exitCopyDisplayLut;
        }

        // Standard QVector copy.
        cameraLutBlue = displayLutSource->cameraLutBlue;
        builtBlueLut = true;   // Assume this is valid.
    }

    // Make it here and all duplication efforts were a success.
    dataCopiedOk = true;

exitCopyDisplayLut:
    return dataCopiedOk;
}


