/*!
 * @file fi_display_lut.h
 * @brief This is the header file for the FiDisplayLut class.
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_DISPLAY_LUT_H
#define FI_DISPLAY_LUT_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <memory>
#include <stdint.h>

#include <QString>
#include <QPoint>
#include <QVector>

#include "fi_splinewidget.h"

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Namespace: DISPLAY_LUT
//--------------------------------------------------------------------
namespace DISPLAY_LUT
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

typedef enum CURVE_ID
{
    RED_GAMMA = 0,  //!< Identifies a red or gamma curve.
    GREEN,          //!< Identifies a green curve.
    BLUE,           //!< Identifies a blue curve.
    UNKNOWN,        //!< Identifies No Curve.
} CURVE_ID;

// Structure to hold the data parsed from the LUT listing XML file downloaded from the camera.
typedef struct LUT_SPLINE_RECORD
{
    int             id;                 //!< Position for this LUT.
    QString         desc;               //!< User supplied description.
    QString         localPath;          //!< Local path (if any) to the file with the LUT information.
    uint8_t         lutDataType;        //!< What type of data is this?  See LUT_DATA_TYPES in fi_lutfile.h.
    QVector<QPoint> redOrGammaKnots;    //!< The coordinate points defining the spline.
    QVector<QPoint> greenKnots;         //!< The coordinate points defining the spline.
    QVector<QPoint> blueKnots;          //!< The coordinate points defining the spline.
} LUT_SPLINE_RECORD;

}   // End namespace (DISPLAY_LUT)


class FiDisplayLut
{
protected:
    explicit FiDisplayLut();

public:
    ~FiDisplayLut();

    // This object, for now, is non-copyable and non-copy assignable.
    FiDisplayLut(const FiDisplayLut& i) = delete;
    FiDisplayLut& operator=(const FiDisplayLut& i) = delete;

    static std::shared_ptr<FiDisplayLut>
    CreateDisplayLut(bool isMonoSensor,
                     bool isSeries3Camera,
                     uint32_t metadataLutId,
                     const QVector<uint16_t>& metadataLutKnots);

    static std::shared_ptr<FiDisplayLut>
    CreateDisplayLut(const DISPLAY_LUT::LUT_SPLINE_RECORD& lutSplineRecToUse,
                     const QVector<double>& cameraLutRedOrGamma,
                     const QVector<double>& cameraLutGreen,
                     const QVector<double>& cameraLutBlue);

    static bool KnotsAreEqual(const DISPLAY_LUT::LUT_SPLINE_RECORD& lutSplineRecToTest);

    bool CopyDisplayLut(std::shared_ptr<FiDisplayLut> displayLutSource);
    bool CopySplineRecord(DISPLAY_LUT::LUT_SPLINE_RECORD& lutSplineRecCopy);

    int GetLutId();
    const QString& GetLutDesc();

    const QVector<double>& GetBlueCameraLut(bool& builtBlueCameraLut);
    const QVector<double>& GetGreenCameraLut(bool& builtGreenCameraLut);
    const QVector<double>& GetRedOrGammaCameraLut(bool& builtRedOrGammaCameraLut);


    //! Is this display lut supporting color channels or just gamma channel.
    bool IsUsingColor()     { return usingColor; }

protected:

    bool CreateSplineWidget(DISPLAY_LUT::CURVE_ID whichCurve,
                            QVector<QPoint> *splineKnots);
    bool BuildRedOrGammaCameraLut();
    bool BuildGreenCameraLut();
    bool BuildBlueCameraLut();

    static std::unique_ptr<struct DISPLAY_LUT::LUT_SPLINE_RECORD>
    XlateMetadataKnots(bool isMonoSensor,
                       bool isSeries3Camera,
                       uint32_t metadataLutId,
                       const QVector<uint16_t>& metadataLutKnots);

    std::unique_ptr<fiSplineWidget> fiSplineBlue;           //!< Blue FI spline object used to generate cameraLutBlue data.
    std::unique_ptr<fiSplineWidget> fiSplineGreen;          //!< Green FI spline object used to generate cameraLutGreen data.
    std::unique_ptr<fiSplineWidget> fiSplineRedOrGamma;     //!< Red or gamma FI spline object used to generate cameraLutRedOrGamma data.

    std::shared_ptr<aaAaa::aaSpline> aaSplineBlue;          //!< Curve data for blue color channel.
    std::shared_ptr<aaAaa::aaSpline> aaSplineGreen;         //!< Curve data for green color channel.
    std::shared_ptr<aaAaa::aaSpline> aaSplineRedOrGamma;    //!< Curve data for Red or Gamma tab.

    bool                builtBlueLut;                       //!< True if cameraLutBlue is built.
    bool                builtGreenLut;                      //!< True if cameraLutGreen is built.
    bool                builtRedOrGammaLut;                 //!< True if cameraLutRedOrGamma is built.

    QVector<double>     cameraLutBlue;                      //!< Camera blue lookup table.
    QVector<double>     cameraLutGreen;                     //!< Camera green lookup table.
    QVector<double>     cameraLutRedOrGamma;                //!< Camera gamma lookup table for series3 images and seriesNC monochrome images.

    std::unique_ptr<struct DISPLAY_LUT::LUT_SPLINE_RECORD> lutSplineRec;

    bool usingColor;                                        //!< True if this display lut object is supporting color.
};

#endif // FI_DISPLAY_LUT_H
