/*!
 * @file fi_fflr_underflow_dialog.cpp
 * @brief This is the header file for the fiFFLRUnderflowDlg class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QFileDialog>
#include <QMessageBox>
#include <QSettings>
#include <QString>
#include <QStyle>

#include "fi_fflr_underflow_dialog.h"
#include "ui_fi_fflr_underflow_dialog.h"

#include "logger.h"

namespace FI_FFLR_UNDERFLOW_DLG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
}

using namespace FI_FFLR_UNDERFLOW_DLG;

//--------------------------------------------------------------------
// Private static data members.
//--------------------------------------------------------------------

//--------------------------------------------------------------------
// Class Implementation for fiFFLRUnderflowDlg
//--------------------------------------------------------------------

/*!
 * @brief The FI User Documentation Dialog constructor.
 * Note: The use of the windows flags ends up getting rid of the CLOSE X at the top right!
 */
FiFFLRUnderflowDlg::FiFFLRUnderflowDlg(uint32_t dirtyBufferCount, const QString& deviceName, CameraMDI *camera_ptr,
                                       QWidget *parent, Qt::WindowFlags flags) :
    QDialog(parent, flags | Qt::CustomizeWindowHint | Qt::WindowTitleHint),
    ui(new Ui::FiFFLRUnderflowDlg)
{
    ui->setupUi(this);

    discardBuffers = false;

    ourCameraPtr = camera_ptr;

    QString fflrInfoUnderflowTxt     = tr("The %1 Drive for the FasFire LR recording has failed to maintain the data rate required for this capture!").arg(deviceName);
    QString fflrInfoBufferTxt        = tr("Currently %1 recorded 4GB capture buffers are unsaved.").arg(dirtyBufferCount);
    QString fflrInfoCBTxt            = tr("Release and discard %1 capture buffers").arg(dirtyBufferCount);

    ui->lbl_info->setText(fflrInfoUnderflowTxt);
    ui->lbl_buffer_info->setText(fflrInfoBufferTxt);
    ui->cb_discard_buffers->setText(fflrInfoCBTxt);
    ui->cb_discard_buffers->setChecked(discardBuffers);

    ui->tb_info_icon->setPixmap(style()->standardPixmap(QStyle::SP_MessageBoxWarning));
    ui->tb_info_icon->setVisible(true);

    connect(ui->pb_discard, &QPushButton::clicked, this, &FiFFLRUnderflowDlg::doDiscardPartialRecording);
    connect(ui->pb_complete, &QPushButton::clicked, this, &FiFFLRUnderflowDlg::doCompletePartialRecording);
    connect(ui->cb_discard_buffers, &QCheckBox::clicked, this, &FiFFLRUnderflowDlg::doDiscardBufferChecked);
}

/*!
 * @brief The FI User Documentation Dialog destructor.
 */
FiFFLRUnderflowDlg::~FiFFLRUnderflowDlg()
{
    delete ui;
}

//--------------------------------------------------------------------
// Event Methods
//--------------------------------------------------------------------

/*!
 * @brief If the user hits the close button, we want to ignore it.
 *    They have to hit one of our action buttons in this dialog.
 *
 * @param event Our Close event.
 */
void FiFFLRUnderflowDlg::closeEvent(QCloseEvent *event)
{
    Q_UNUSED(event)

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Got a closeEvent";
    //event->ignore();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Responds to the Discard Partial Recording push button.
 */
void FiFFLRUnderflowDlg::doDiscardPartialRecording()
{
    // Simply mark the dialog as being done with a QDialog::Rejected - this will mark our completion
    // for discarding.
    done(QDialog::Rejected);
}

/*!
 * @brief Responds to the Complete Partial Recording push button.
 */
void FiFFLRUnderflowDlg::doCompletePartialRecording()
{
    // Simply mark the dialog as being done with a QDialog::Accepted - this will mark our completion
    // for completing partial recording.
    done(QDialog::Accepted);
}

/*!
 * @brief Responds to the Complete Partial Recording push button.
 */
void FiFFLRUnderflowDlg::doDiscardBufferChecked()
{
    // Simply mark the persisted state so we can ask for it later.
    discardBuffers = ui->cb_discard_buffers->isChecked();
}


