/*!
 * @file fi_file.cpp
 * @brief A file utility class derived from QFile allowing enhancements
 * beyond the standard QFile functionality.
 *
 * Copyright 2023 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QObject>
#include <QFileInfo>
#include <QFile>

#include "fi_file.h"
#include "logger.h"

using namespace FI_FILE;

//! Collects active copy file operations and their cancel state.
QMap<QString, bool> FiFile::copyFileList;

/*!
 * @brief Class constructor.
 */
FiFile::FiFile(const QString& name, QObject *parent)
    : QFile{name, parent}
{
    FiFile::copyFileList.insert(name, false);
}

/*!
 * @brief Class destructor.
 */
FiFile::~FiFile()
{
    FiFile::copyFileList.remove(fileName());
}

/*!
 * @brief Custom copy function that can signal progress.
 *
 * @param newName           The name of the copy which includes its full path.
 * @param pctUpdatePeriod   A periodic percentage for signaling copyPercentComplete updates
 *                          to consumer ranging from 1 up to 100.
 *
 * @returns True if the file was copied without any errors, false if not.
 */
bool FiFile::copyFile(const QString& newName, uint32_t pctUpdatePeriod)
{
    qint64      bytesExpected       = 0;
    qint64      totalRead           = 0;
    bool        copyError           = false;
    bool        userCanceled        = false;
    const auto  fileTemplate        = QLatin1String("%1/fi_temp.XXXXXX");

    QFile outFile(fileTemplate.arg(QFileInfo(newName).path()));

    // Make sure the percent updating period is a number between 1 and 100.
    if (pctUpdatePeriod > 100)
        pctUpdatePeriod = 100;
    else if (pctUpdatePeriod == 0)
        pctUpdatePeriod = 1;

    // Can't copy a non-existing file.
    if (!exists())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("Cannot copy file %1, file doesn't exist.").arg(fileName());
        copyError = true;
        goto copyExit;
    }

    // Can't overwrite an existing file.
    if (QFile::exists(newName))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("Cannot create %1, file already exists.").arg(newName);
        copyError = true;
        goto copyExit;
    }

    // Make sure that our file isn't already opened.
    if (isOpen())
    {
        close();
    }

    // Attempt to open our existing file.
    if (!open(QFile::ReadOnly))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("Cannot open %1 for input").arg(fileName());
        copyError = true;
        goto copyExit;
    }

    // How many bytes have to be copied.
    bytesExpected = size();

    // If there is no size to the source file that's an error.
    if (bytesExpected <= 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("File %1 is empty.").arg(fileName());
        copyError = true;
        goto copyExit;
    }

    if (!outFile.open(QIODevice::ReadWrite))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("Cannot open for output: %1").arg(outFile.errorString());
        copyError = true;
        goto copyExit;
    }

    // Reset the user canceled state to false and insert if it not in list.
    if (!FiFile::copyFileList.contains(fileName()))
    {
        FiFile::copyFileList.insert(fileName(), false);
    }
    else
    {
        FiFile::copyFileList[fileName()] = false;
    }

    // Scope this because of variables and goto statements above.
    {
        uint32_t pctComplete    = 0;
        uint32_t pctSignal      = 0;

        char block[4096];
        while (!atEnd() && !userCanceled)
        {
            qint64 bytesRead = read(block, sizeof(block));
            if (bytesRead <= 0)
                break;
            totalRead += bytesRead;
            if (bytesRead != outFile.write(block, bytesRead))
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << QString("Failure to write block");
                break;
            }
            pctComplete = (totalRead * 100) / bytesExpected;

            // Emit progress update if an update period was provided.
            if (pctUpdatePeriod > 0)
            {
                // Only emit if threshold hit.
                if ((pctComplete - pctSignal) >= pctUpdatePeriod)
                {
                    pctSignal = pctComplete;
                    emit copyPercentComplete((int)pctSignal, fileName());
                    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "pctSignal:" << pctSignal;
                }
            }

            // Check to see if external processing marked our copy to be canceled.
            if (FiFile::copyFileList.contains(fileName()))
            {
                userCanceled = FiFile::copyFileList.value(fileName());
                //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "userCanceled:" << userCanceled;
            }
        }
    }

    // Did we copy the entire file?
    if (totalRead == bytesExpected)
    {
        if (!outFile.rename(newName))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << QString("Unable to rename temporary copy to").arg(newName);
            copyError = true;
            goto copyExit;
        }
    }
    else
    {
        QString copy_incomplete_str = "Copy incomplete";
        if (userCanceled)
            copy_incomplete_str += "(canceled by user)";
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("%1, had transferred %2% before stopping.").arg(copy_incomplete_str).arg((totalRead * 100) / size());
        copyError = true;
        goto copyExit;
    }

copyExit:

    // Make sure to close any output.
    if (outFile.isOpen())
    {
        outFile.close();
    }

    if (copyError)
    {
        // Remove any partial attempt.
        outFile.remove();
    }

    QFile::setPermissions(newName, permissions());
    close();

    return !copyError;
}


