/*!
 * @file fi_graphicsview.cpp
 * @brief This provides the implementation for the fiGraphicsView class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QApplication>
#include <QGraphicsItem>
#include <QGraphicsPixmapItem>
#include <QGraphicsView>
#include <QImage>
#include <QMouseEvent>
#include <QPixmap>
#include <QRgb>
#include <QWheelEvent>
#include <qmath.h>

#include "fi_graphicsview.h"
#include "logger.h"
#include "mainwindow.h"

namespace FI_GVIEW
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

const double                BASE_ZOOM_FACTOR        = 1.0015;
const Qt::KeyboardModifiers DEFAULT_MODIFIERS       = Qt::ControlModifier;
const int                   INFO_DISPLAY_TIME       = 60000;

}   // End of namespace FI_GVIEW.

using namespace FI_GVIEW;

/*!
 * @brief The fiGraphicsView constructor with only the parent pointer as a parameter.
 */
fiGraphicsView::fiGraphicsView(QWidget *parent) :
    QGraphicsView(parent),
    use_info_cursor(false),
    zoom_factor(1.0),
    zoom_factor_base(BASE_ZOOM_FACTOR),
    modifiers(DEFAULT_MODIFIERS),
    saved_rot_angle(0.0)
{
    // Make sure we track the mouse.
    setMouseTracking(true);
}

/*!
 * @brief The fiGraphicsView constructor used when the scene is specified.
 */
fiGraphicsView::fiGraphicsView(QGraphicsScene *scene, QWidget *parent) :
    QGraphicsView(scene, parent),
    use_info_cursor(false),
    zoom_factor(1.0),
    zoom_factor_base(BASE_ZOOM_FACTOR),
    modifiers(DEFAULT_MODIFIERS),
    saved_rot_angle(0.0)
{
    // Make sure we track the mouse.
    setMouseTracking(true);
}

/*!
 * @brief Used to scale the scene to a specific zoom factor.
 * @param factor    The zoom factor the user wants to use for the scene.
 */
void fiGraphicsView::do_zooming(double factor)
{
    QPointF delta_viewport_pos;
    QPointF viewport_center;

    // Save our zoom factor.
    zoom_factor = factor;

    // Update our image.
    scale(factor, factor);
    delta_viewport_pos = target_viewport_pos - QPointF(viewport()->width() / 2.0, viewport()->height() / 2.0);
    viewport_center = mapFromScene(target_scene_pos) - delta_viewport_pos;
    centerOn(mapToScene(viewport_center.toPoint()));

    emit zoomed();
}

/*!
 * @brief Enable or disable the information cursor function.
 * @param do_enable     True to enable, false to disable.
 */
void fiGraphicsView::enable_info_cursor(bool do_enable)
{
    // Save our enable flag.
    use_info_cursor = do_enable;

    // Are we using the information cursor?
    if (use_info_cursor)
    {
        // Change the cursor to be the crosshair shape.
        viewport()->setCursor(Qt::CrossCursor);

        // Show information, if it's available.
        update_info(pos());
    }
    else
    {
        // Set our cursor to the default.
        viewport()->setCursor(Qt::OpenHandCursor);
        setToolTip(QString());
    }
}

void fiGraphicsView::resetTransform()
{
    // Reset the rotation angle.
    saved_rot_angle = 0.0;

    // Pass it on.
    QGraphicsView::resetTransform();
}

void fiGraphicsView::rotate(qreal angle)
{
    // Save the angle.
    saved_rot_angle = angle;

    // Pass it on.
    QGraphicsView::rotate(angle);
}

/*!
 * @brief Update the cursors tool-tip with the position information.
 * @param pos       Position we're interested in.
 */
void fiGraphicsView::update_info(QPoint pos)
{
    int     i;
    QPointF mapped_pos;
    QRgb    pixel_value;
    QString tt_str;
    QList<QGraphicsItem *>  scene_items;
    QGraphicsPixmapItem     *pixmap_item;

    // Map from the screen coordinates to the image coordinates.
    mapped_pos = mapToScene(pos);
    if (mapped_pos.x() < 0)
        mapped_pos.setX(0.0);
    if (mapped_pos.y() < 0)
        mapped_pos.setY(0.0);

    // Find the pixmap item in the scene.
    scene_items = scene()->items();
    pixmap_item = nullptr;
    for (i = 0; i < scene_items.count(); ++i)
    {
        pixmap_item = qgraphicsitem_cast<QGraphicsPixmapItem *>(scene_items.at(i));
        if (pixmap_item != nullptr)
            break;
    }

    // Did we find the pixmap item?
    if (pixmap_item == nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No pixmap item found.";
        return;
    }

    // Adjust our mapped position to be within the image.
    if (mapped_pos.x() < pixmap_item->sceneBoundingRect().left())
        mapped_pos.setX(pixmap_item->sceneBoundingRect().left());
    if (mapped_pos.y() < pixmap_item->sceneBoundingRect().top())
        mapped_pos.setY(pixmap_item->sceneBoundingRect().top());
    if (mapped_pos.x() >= pixmap_item->sceneBoundingRect().right())
        mapped_pos.setX(pixmap_item->sceneBoundingRect().right() - 1.0);
    if (mapped_pos.y() >= pixmap_item->sceneBoundingRect().bottom())
        mapped_pos.setY(pixmap_item->sceneBoundingRect().bottom() - 1.0);

    // Make sure our mapped position is normalized to 0,0.
    mapped_pos.setX(mapped_pos.x() - pixmap_item->sceneBoundingRect().left());
    mapped_pos.setY(mapped_pos.y() - pixmap_item->sceneBoundingRect().top());

    // Get the pixel value at the position specified.
    pixel_value = pixmap_item->pixmap().toImage().pixel((int)mapped_pos.x(), (int)mapped_pos.y());

    // Build up our tool-tip.
    tt_str = QString("[%1, %2, %3]@(%4, %5)").arg(qRed(pixel_value)).arg(qGreen(pixel_value)).arg(qBlue(pixel_value)).arg((int)mapped_pos.x()).arg((int)mapped_pos.y());

    // Display the tool-tip.
    setToolTip(tt_str);
    setToolTipDuration(INFO_DISPLAY_TIME);

    // Also put the text in the main window's status bar. Allow
    // it to timeout after 2 seconds.
    MainWindow::GetInstance()->set_statusbar_message(tt_str, 2000);

    // Force the tool tip to be updated.
    update();
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Allow the user to "pan" the scene in the view.
 * @param event     Pointer to the event.
 */
void fiGraphicsView::mouseMoveEvent(QMouseEvent *event)
{
    QPointF delta = target_viewport_pos - event->pos();

    // Make sure the mouse has moved a minimum amount.
    if ((qAbs(delta.x()) > 5) || (qAbs(delta.y()) > 5))
    {
        // Get the new mouse position.
        target_viewport_pos = event->pos();
        target_scene_pos = mapToScene(event->pos());
    }

    // Pass it on up.
    QGraphicsView::mouseMoveEvent(event);

    // Update our cursor, if appropriate.
    if (use_info_cursor && (event->buttons() == Qt::NoButton))
    {
        // We're in info cursor mode!
        viewport()->setCursor(Qt::CrossCursor);

        // Update the cursor's tooltip.
        update_info(event->pos());
    }
}

/*!
 * @brief The user has pressed the mouse button (make sure it's the
 *        left one). Update the cursor if appropriate.
 * @param event     Pointer to the event.
 */
void fiGraphicsView::mousePressEvent(QMouseEvent *event)
{
    // Pass it on up.
    QGraphicsView::mousePressEvent(event);
}

/*!
 * @brief The user has released the mouse button. Update the cursor if appropriate.
 * @param event     Pointer to the event.
 */
void fiGraphicsView::mouseReleaseEvent(QMouseEvent *event)
{
    // Pass it on up.
    QGraphicsView::mouseReleaseEvent(event);

    // Update our cursor, if appropriate.
    if (use_info_cursor)
    {
        // Restore our cursor.
        viewport()->setCursor(Qt::CrossCursor);

        // Update the cursor's tooltip.
        update_info(event->pos());
    }
}

/*!
 * @brief Allow the user to "zoom" the scene.
 * @param event     Pointer to the event.
 */
void fiGraphicsView::wheelEvent(QWheelEvent *event)
{
    // Is the user holding down the correct key sequence while
    // using the mouse wheel?
    if (QApplication::keyboardModifiers() == modifiers)
    {
        // Update the zoom.
        double angle = event->angleDelta().y();
        double factor = qPow(zoom_factor_base, angle);
        do_zooming(factor);
    }
    else
    {
        // Pass it on up.
        QGraphicsView::wheelEvent(event);
    }
}
