/*!
 * @file fi_graphicsview.h
 * @brief This is the header file for the fiGraphicsView class.
 *
 * This class adds the ability to zoom QGraphicsView images using
 * the mouse wheel. The point under cursor remains motionless while it's possible.
 *
 * Note that it becomes not possible when the scene's size is not large enough compared
 * to the viewport size. QGraphicsView centers the picture when it's smaller than the view.
 * And QGraphicsView's scrolls boundaries don't allow to put any picture point at any viewport position.
 *
 * When the user starts scrolling, this class remembers original scene position and
 * keeps it until scrolling is completed. It's better than getting original scene position at
 * each scrolling step because that approach leads to position errors due to before-mentioned
 * positioning restrictions.
 *
 * When zommed using scroll, this class emits zoomed() signal.
 *
 * You can set keyboard modifiers used for zooming using set_modified(). Zooming will be
 * performed only on exact match of modifiers combination. The default modifier is Ctrl.
 *
 * You can change zoom velocity by calling set_zoom_factor_base().
 * Zoom coefficient is calculated as zoom_factor_base^angle_delta
 * (see QWheelEvent::angleDelta).
 * The default zoom factor base is 1.0015.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_GRAPHICSVIEW_H
#define FI_GRAPHICSVIEW_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QWidget>
#include <QGraphicsLayoutItem>
#include <QGraphicsPixmapItem>
#include <QGraphicsTextItem>
#include <QGraphicsView>

#include "logger.h"

//! Need to create a QGraphicsView subclass so we can use the mouse wheel
//! to zoom our images.
class fiGraphicsView : public QGraphicsView
{
    Q_OBJECT

public:
    explicit fiGraphicsView(QWidget *parent = nullptr);
    fiGraphicsView(QGraphicsScene *scene, QWidget *parent = nullptr);
    ~fiGraphicsView() {}

    void    do_zooming(double factor);
    void    enable_info_cursor(bool do_enable);

    double  get_base_factor(void)         { return zoom_factor_base; }
    double  get_zoom_factor(void)         { return zoom_factor; }

    void    set_base_factor(double new_factor)                    { zoom_factor_base = new_factor; }
    void    set_modifiers(Qt::KeyboardModifiers new_modifiers)    { modifiers = new_modifiers; }

    void    resetTransform();

    void    rotate(qreal angle);
    qreal   rot_angle(void) { return saved_rot_angle; }

signals:
    void    zoomed(void);

protected:
    virtual void    mouseMoveEvent(QMouseEvent *event);
    virtual void    mousePressEvent(QMouseEvent *event);
    virtual void    mouseReleaseEvent(QMouseEvent *event);
    virtual void    wheelEvent(QWheelEvent *event);

private:
    void    update_info(QPoint pos);

private:
    bool                    use_info_cursor;
    double                  zoom_factor;
    double                  zoom_factor_base;
    Qt::KeyboardModifiers   modifiers;

    QPointF                 target_scene_pos;
    QPointF                 target_viewport_pos;

    qreal                   saved_rot_angle;
};

class fiLayoutPixmap : public QGraphicsPixmapItem, public QGraphicsLayoutItem
{
public:
    explicit fiLayoutPixmap(const QPixmap& pixmap, QGraphicsItem *parent = nullptr) :
        QGraphicsPixmapItem(pixmap, parent),
        QGraphicsLayoutItem()
    {
    }

    ~fiLayoutPixmap()       {}

    virtual QSizeF sizeHint(Qt::SizeHint which, const QSizeF& constraint = QSizeF()) const
    {
        Q_UNUSED(which)
        Q_UNUSED(constraint)
        return boundingRect().size();
    }

    virtual void setGeometry(const QRectF& rect)
    {
        setPos(rect.topLeft());
    }
};

class fiLayoutText : public QGraphicsSimpleTextItem, public QGraphicsLayoutItem
{
//    Q_OBJECT
//    Q_INTERFACES(QGraphicsLayoutItem)

public:
    explicit fiLayoutText(QGraphicsItem *parent = nullptr) :
        QGraphicsSimpleTextItem(parent),
        QGraphicsLayoutItem()
    {
        id = -1;
    }

    explicit fiLayoutText(const QString& my_text, QGraphicsItem *parent = nullptr) :
        QGraphicsSimpleTextItem(my_text, parent),
        QGraphicsLayoutItem()
    {
        id = -1;
    }

    ~fiLayoutText()         {}

    virtual QSizeF sizeHint(Qt::SizeHint which, const QSizeF& constraint = QSizeF()) const
    {
        Q_UNUSED(which)
        Q_UNUSED(constraint)

        return boundingRect().size();
    }

    virtual void setGeometry(const QRectF& rect)
    {
        setPos(rect.topLeft());
    }

    void setPos(const QPointF& pos)
    {
        QGraphicsSimpleTextItem::setPos(pos);
    }
    inline void setPos(qreal x, qreal y);

public:
    int     id;
};

inline void fiLayoutText::setPos(qreal ax, qreal ay)    { setPos(QPointF(ax, ay)); }

#endif // FI_GRAPHICSVIEW_H
