/*!
 * @file fi_histo_proc.cpp
 * @brief The code to calculate and manipulate various histograms.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QImage>
#include <QPixmap>
#include <QtConcurrent>
#include <qmath.h>

#include "fi_histo_proc.h"
#include "logger.h"

namespace FI_HISTO_PROC
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

}   // End of namespace FI_HISTO_PROC.

using namespace FI_HISTO_PROC;

/*!
 * @brief The class constructor.
 */
fiHistoProc::fiHistoProc(QObject *parent) : QObject(parent),
    mono_mode(false),
    histogram_roi(QRect(0, 0, 0, 0)),
    sum_red(0),
    sum_sq_red(0),
    sum_green(0),
    sum_sq_green(0),
    sum_blue(0),
    sum_sq_blue(0)
{
}

void fiHistoProc::calc(const QPixmap& my_pixelmap)
{
    int x;
    int y;
    QRect img_rect;
    QImage my_image;

    // Convert the pixmap to an image so I can work with it.
    my_image = my_pixelmap.toImage();

    // Setup the histogram vector. Note that, with a QPixmap, we assume an
    // 8-bit histogram.
    histogram_red.fill(0, 256);
    histogram_green.fill(0, 256);
    histogram_blue.fill(0, 256);
    histogram_luminance.fill(0, 256);

    sum_red = 0;
    sum_sq_red = 0;
    sum_green = 0;
    sum_sq_green = 0;
    sum_blue = 0;
    sum_sq_blue = 0;

    // An "empty" roi means to use the full ROI.
    if ((histogram_roi.width() == 0) || (histogram_roi.height() == 0))
    {
        // Get the ROI for the image.
        img_rect = my_image.rect();
    }
    else
    {
        // Sanity check our ROI.
        img_rect = my_image.rect();
        if (!img_rect.contains(histogram_roi))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Histogram ROI:" << histogram_roi << "Not in range:" << img_rect;

            // Turn off the histogram so we don't come back in here.
            emit fault();
            return;
        }

        // Now re-use the image ROI to be our histogram ROI.
        img_rect = histogram_roi;
    }

    // Setup our map.
    histo_lum_map.fill(0, (img_rect.width() * img_rect.height()));

    // Start accumulating.
// #pragma omp parallel for collapse(2)
    for (y = img_rect.top(); y < img_rect.bottom(); y++)
    {
        for (x = img_rect.left(); x < img_rect.right(); x++)
        {
            QRgb pixel;
            uint32_t r_pix;
            uint32_t g_pix;
            uint32_t b_pix;
            int lum;
            int map_y = (y - img_rect.top()) * img_rect.width();
            int map_x = (x - img_rect.left());

            // Get the pixel at this x (column) and y (row)
            pixel = my_image.pixel(x, y);

            // Pull the RGB out of the pixel
            r_pix = (uint32_t)qRed(pixel);
            g_pix = (uint32_t)qGreen(pixel);
            b_pix = (uint32_t)qBlue(pixel);

            // Accumulate the histogram values for the red channel. Note that we are
            // using the red channel for "mono" as well.
            histogram_red[(int)r_pix] += 1;
            sum_red += r_pix;
            sum_sq_red += (((uint64_t)r_pix) * ((uint64_t)r_pix));

            // If we're not in mono mode, then accumulate the other channels too.
            if (!mono_mode)
            {
                histogram_green[(int)g_pix] += 1;
                sum_green += g_pix;
                sum_sq_green += (((uint64_t)g_pix) * ((uint64_t)g_pix));

                histogram_blue[(int)b_pix] += 1;
                sum_blue += b_pix;
                sum_sq_blue += (((uint64_t)b_pix) * ((uint64_t)b_pix));

                // Derive luminance for this pixel and update the histogram. This
                // formula is from ITU BT.709.
                lum = (int)ceil(0.2126 * qRed(pixel) + 0.7152 * qGreen(pixel) + 0.0722 * qBlue(pixel));
            }
            else
            {
                // Luminance doesn't really make sense for mono images...it's simply the pixel value.
                lum = (int)r_pix;
            }

            // Finish with the luminance processing.
            if (lum > 255)
                lum = 255;

            // Save our values.
            histogram_luminance[ lum ] += 1;
            histo_lum_map[map_y + map_x] = lum;
        }
    }

    // Tell folks the histogram is ready.
    emit ready();
}

void fiHistoProc::calc(const QByteArray& my_pixels, int bpp, const QRect& img_rect)
{
    int x;
    int y;
    int bin_count = (int)pow(2, bpp);
    QRect sub_rect;
    const uint8_t *buffer_ptr;

    // Setup the histogram vector. For "raw" mode, we're only using the red channel.
    histogram_red.resize(bin_count);
    histogram_red.fill(0);
    sum_red = 0;
    sum_sq_red = 0;

    // These are unused.
    histogram_green.clear();
    histogram_blue.clear();
    histogram_luminance.clear();

    // An "empty" roi means to use the full ROI.
    if ((histogram_roi.width() == 0) || (histogram_roi.height() == 0))
    {
        // Use the full ROI.
        sub_rect = img_rect;

        // Setup our map.
        histo_lum_map.resize(img_rect.width() * img_rect.height());
        histo_lum_map.fill(0);
    }
    else
    {
        // Sanity check our ROI.
        if (!img_rect.contains(histogram_roi))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Histogram ROI:" << histogram_roi << "Not in range:" << img_rect;

            // Turn off the histogram so we don't come back in here.
            emit fault();
            return;
        }

        // Remember the ROI we're using.
        sub_rect = histogram_roi;

        // Setup our map.
        histo_lum_map.resize(sub_rect.width() * sub_rect.height());
        histo_lum_map.fill(0);
    }

    // Setup our pointer.
    buffer_ptr = (const uint8_t *)my_pixels.constData();

    // Start accumulating.
// #pragma omp parallel for collapse(2)
    for (y = sub_rect.top(); y < sub_rect.bottom(); y++)
    {
        for (x = sub_rect.left(); x < sub_rect.right(); x++)
        {
            uint16_t raw_pix;
            int index;
            int map_y = (y - sub_rect.top()) * sub_rect.width();
            int map_x = (x - sub_rect.left());

            // What is the index into our byte array?
            index = (y * img_rect.width()) + x;

            // The pixel will be left justified in the word, if bpp is greater than 8 bits.
            if (bpp > 8)
            {
                // Get the pixel at this x (column) and y (row)
                memcpy(&raw_pix, &buffer_ptr[index * 2], 2);
                raw_pix = raw_pix >> (16 - bpp);
            }
            else
            {
                raw_pix = (uint16_t)buffer_ptr[index];
            }

            // Accumulate the histogram values for the red channel. Note that we are
            // using the red channel for "mono" as well.
            histogram_red[(int)raw_pix] += 1;
            sum_red += raw_pix;
            sum_sq_red += (((uint64_t)raw_pix) * ((uint64_t)raw_pix));

            // Finish with the luminance processing.
            histo_lum_map[map_y + map_x] = raw_pix;
        }
    }

    // Tell folks the histogram is ready.
    emit ready();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

