/*!
 * @file fi_histogram.cpp
 * @brief This file contains the implementation of the fiHistogram class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_histogram.h"
#include "ui_fi_histogram.h"

#include <QMouseEvent>
#include <qmath.h>

#include <qwt_plot.h>
#include <qwt_plot_canvas.h>
#include <qwt_plot_curve.h>
#include <qwt_plot_magnifier.h>
#include <qwt_plot_marker.h>
#include <qwt_plot_panner.h>
#include <qwt_picker.h>
#include <qwt_picker_machine.h>
#include <qwt_plot_picker.h>
#include <qwt_text.h>

#include <algorithm>

#include "logger.h"

namespace FI_HISTOGRAM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Set these to 1 if you want to see log output.
#define LOG_OPERATIONAL_INFO    0

const int   GP_COL_WIDTH    = 5;            //!< The number of columns we use for our "goal posts".

//! Create a common class to manage the histogram axis data elements.
class fiAxisData
{
public:
    /*!
    * @brief Static method to reuse or allocate an axis data shared pointer.
    *
    * @param numberOfEntries    How many entries in the axis data will be used.
    * @param theAxisData        Axis data that has already been allocated or an empty one if none exist.
    *
    * @returns The shared pointer to the fiAxisData if one was reallocated or allocated.
    */
    static std::shared_ptr<FI_HISTOGRAM::fiAxisData> allocateData(int numberOfEntries, std::shared_ptr<FI_HISTOGRAM::fiAxisData> theAxisData)
    {
        std::shared_ptr<FI_HISTOGRAM::fiAxisData> emptyData;  // Initialize to empty unique pointer.

        // Is there enough space in the axis data already allocated?
        if (theAxisData)
        {
            // Are there enough entries for this allocation?
            if (theAxisData->theDataAllocated >= numberOfEntries)
            {
                theAxisData->theDataUsed = numberOfEntries;
                return theAxisData;
            }
        }

        // No, cannot reuse, must allocate one.
        std::shared_ptr<FI_HISTOGRAM::fiAxisData> axisDataAllocated = std::shared_ptr<FI_HISTOGRAM::fiAxisData>(new FI_HISTOGRAM::fiAxisData(numberOfEntries));

        // Make sure we have the right number of elements created.
        if (axisDataAllocated)
        {
            if (axisDataAllocated->getElementsInUse() == numberOfEntries)
            {
                return axisDataAllocated;
            }
            else
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "theAxisData->getElementsInUse:" << axisDataAllocated->getElementsInUse()
                            << "not EQUAL to numberOfEntries:" << numberOfEntries << "!";
            }
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate theData!";
        }

        // If arrive here, the axis data could not be allocated or reused.
        return emptyData;
    }

    /*!
    * @brief Static method to generate a QVector of axis data from existing fiAxisData shared pointer.
    *
    * Note: If theAxisData provided as parameter doesn't have a full count the vector is filled with
    * 0.0 for any missing entries.
    *
    * @param countExpected      How many entries in the axis data will be copied.
    * @param theAxisData        Axis data that has already been allocated used as the source of the vector data.
    *
    * @returns The shared pointer to the fiAxisData if one was reallocated or allocated.
    */
    static const QVector<double> copyAxisData(int countExpected, std::shared_ptr<FI_HISTOGRAM::fiAxisData> theAxisData)
    {
        int i;
        double *axisDataPtr = nullptr;  // The pointer to the axis data.
        double emptyElement = 0.0;

        QVector<double> ret_data;

        // Make sure the vector is sized to hold the countExpected.
        ret_data.reserve(countExpected);

        // Get a pointer to the axis data.
        if (theAxisData)
            axisDataPtr = theAxisData->getAxisDataPtr();

        // If we don't have valid axis data pointer, fill QVector with 0.0.
        if (!axisDataPtr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "theAxisData pointer was null, filling QVector with 0.0!";

            for (i = 0; i < countExpected; ++i)
                ret_data.append(emptyElement);
        }
        else
        {
            bool had_overrun = false;
            for (i = 0; i < countExpected; ++i)
            {
                if (i < theAxisData->theDataUsed)
                    ret_data.append(axisDataPtr[i]);
                else
                {
                    ret_data.append(emptyElement);
                    had_overrun = true;
                }
            }
            if (had_overrun)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "theAxisData->theDataUsed:" << theAxisData->theDataUsed
                            << "LESS than countExpected:" << countExpected << "used 0.0 for overrun entries!";
            }
        }

        // Returning the QVector created here.
        return ret_data;
    }

    /*!
    * @brief Static method return the data associated with one axis element.
    *
    * @param theAxisData        Axis data that has already been allocated and filled.
    * @param index              Zero-based entry index for element requested.
    *
    * @returns The element data or 0.0 if no element exists for index.
    */
    static double getAxisElement(std::shared_ptr<FI_HISTOGRAM::fiAxisData> theAxisData, int index)
    {
        double result = 0.0;

        // If have the data set the result.
        if (theAxisData)
        {
            if (index < theAxisData->theDataUsed)
                result = theAxisData->theData[index];
            else
                QLOG_WARN() << __PRETTY_FUNCTION__ << "index:" << index << "was out of bounds for theAxisData::theDataUsed:" << theAxisData->theDataUsed;
        }
        return result;
    }

    /*!
    * @brief Provides access to the data pointer.
    * @returns A pointer to the data if it exists, nullptr if it does not exist.
    */
    double *getAxisDataPtr()
    {
        double *theDataPtr = nullptr;
        if (theData)
        {
            theDataPtr = theData.get();
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to get() theData as it was not allocated!";
        }
        return theDataPtr;
    }

    /*!
    * @brief Returns the number of entries available for use.
    */
    int getElementsInUse()
    {
        int elementsInUse = 0;

        if (theData)
        {
            elementsInUse = theDataUsed;
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read theDataUsed as theData was not allocated!";
        }
        return elementsInUse;
    }

private:
    /*!
    * @brief Constructor for the fiHistogramData class.
    * It must be constructed via the static allocateData method.
    *
    * @param numberOfElements   How many elements in this histogram axis data.
    */
    fiAxisData(int numberOfElements) : theDataAllocated(0), theDataUsed(0)
    {
        if (numberOfElements > 0)
        {
            theData = std::unique_ptr<double[]>(new double[numberOfElements]);
            if (theData)
            {
                theDataAllocated = theDataUsed  = numberOfElements;
            }
            else
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate theData!";
            }
        }
    }

    int theDataAllocated;               //!< Size of theData when allocated.
    int theDataUsed;                    //!< Size of theData being used.
    std::unique_ptr<double[]> theData;  //!< Histogram data for an axis or axis channel.
};

//! Subclass the plot picker object so we can display our own information,
//! namely the counts in the histogram at the cursors position.
class fiPlotPicker: public QwtPlotPicker
{
public:
    fiPlotPicker(QWidget *canvas, fiHistogram *parent):
        QwtPlotPicker(canvas)
    {
        my_parent = parent;
    }

protected:
    virtual QwtText trackerTextF(const QPointF& pos) const
    {
        int     x_pos;
        double  r_val(0);
        double  g_val(0);
        double  b_val(0);
        QString ret_text;

        // Sanity check.
        if (my_parent == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "What?! No Parent!";
            return QwtText("");
        }

        // Get the counts at this X position.
        x_pos = (int)pos.x();
        my_parent->get_count_at(x_pos, r_val, g_val, b_val);
        if (my_parent->is_mono())
            ret_text = QString("%1: %2").arg(x_pos).arg((int)r_val);
        else
            ret_text = QString("%1: [%2, %3, %4]").arg(x_pos).arg((int)r_val).arg((int)g_val).arg((int)b_val);

        // Return the text.
        return QwtText(ret_text);
    }

private:
    fiHistogram *my_parent;                 //!< Pointer to our histogram parent.
};

}   // End of namespace FI_HISTOGRAM.

using namespace FI_HISTOGRAM;


/*!
* @brief Constructor for the fiHistogram class. Initialize variables.
*/
fiHistogram::fiHistogram(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::fiHistogram)
{
    QPalette canvasPalette(Qt::white);
    QwtPlotCanvas       *canvas_ptr;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++CONSTRUCTING+++";
#endif

    // Initialize the UI.
    ui->setupUi(this);

    // Assume we're a color camera.
    b_mono = false;
    b_raw = false;
    x_count = 0;
    y_scale = 1.0;

    // Catch the check/uncheck of the group boxes.
    connect(ui->red_group, SIGNAL(clicked(bool)), this, SLOT(red_box_clicked(bool)));
    connect(ui->green_group, SIGNAL(clicked(bool)), this, SLOT(green_box_clicked(bool)));
    connect(ui->blue_group, SIGNAL(clicked(bool)), this, SLOT(blue_box_clicked(bool)));

    // All smart pointers for the x and y axis data are initially empty.

    // Initialize our pointers to null.
    p_gray_plot = nullptr;
    p_red_plot = nullptr;
    p_grn_plot = nullptr;
    p_blue_plot = nullptr;
    p_gp_min = nullptr;
    p_gp_max = nullptr;
    histo_tracker = nullptr;

    // Allocate space for our "goal post" plots.
    gp_min_x_points = std::unique_ptr<double[]>(new double[ GP_COL_WIDTH ]);
    gp_min_y_points = std::unique_ptr<double[]>(new double[ GP_COL_WIDTH ]);
    gp_max_x_points = std::unique_ptr<double[]>(new double[ GP_COL_WIDTH ]);
    gp_max_y_points = std::unique_ptr<double[]>(new double[ GP_COL_WIDTH ]);

    // Sanity check.
    if (!gp_min_x_points || !gp_min_y_points || !gp_max_x_points || !gp_max_y_points)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate goal-post memory!";
        return;
    }

    // Initialize our background drawing.
    ui->histo_plot->setAutoFillBackground(true);
    update_gradient();

    // Get a pointer to our canvas.
    canvas_ptr = qobject_cast<QwtPlotCanvas *>(ui->histo_plot->canvas());
    if (canvas_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No canvas?!";
        return;
    }

    // Initialize our canvas.
    canvas_ptr->setLineWidth(1);
    canvas_ptr->setFrameStyle(QFrame::Panel | QFrame::Sunken);
    canvas_ptr->setPalette(canvasPalette);

    // Initialize our default display.
    ui->histo_plot->enableAxis(QwtPlot::yLeft, false);
    ui->histo_plot->enableAxis(QwtPlot::xBottom, false);
    ui->histo_plot->setAxisAutoScale(QwtPlot::xBottom, false);
    ui->histo_plot->setLineWidth(1);
    ui->histo_plot->setFocusPolicy(Qt::NoFocus);
    ui->histo_plot->setAutoReplot(false);

    histo_tracker = qobject_cast< QwtPlotPicker *>(new fiPlotPicker(canvas_ptr, this));

    histo_tracker->setStateMachine(new QwtPickerTrackerMachine);
    histo_tracker->setTrackerMode(QwtPicker::ActiveOnly);

    // Start out with the statistics information hidden.
    show_statistics(false);
}

/*!
 * @brief The fiHistogram destructor.
 */
fiHistogram::~fiHistogram()
{
#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++DESTRUCTING+++";
#endif
    // Release the ui.
    delete ui;

    // Release the plot data.
    free_plots();

    // Release the histo_tracker.
    if (histo_tracker)
    {
        delete histo_tracker;
        histo_tracker = nullptr;
    }
}

/*!
 * @brief Given the "curve_data", build up the "blue" curve of the histogram.
 *
 * @param curve_data        Pointer to the blue histogram data.
 * @param num_entries       Number of points in the data buffer.
 *
 * @return The maximum value found.
 */
uint32_t fiHistogram::build_blue_curve(const uint32_t *curve_data, int num_entries)
{
    int         i;
    int         min_index;
    int         max_index;
    int         peak_index;
    uint32_t    peak_value;
    QString     info_str;

    double     *x_axis_data_ptr = nullptr;
    double     *blue_data_ptr   = nullptr;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++BLUE CURVE BUILD+++ num_entries:" << num_entries;
#endif

    // Initialize our counters.
    min_index = num_entries;
    max_index = 0;
    peak_index = 0;
    peak_value = 0;

    // Grab the double data pointers.
    if (x_axis_data)
        x_axis_data_ptr = x_axis_data->getAxisDataPtr();

    blue_data = fiAxisData::allocateData(num_entries, blue_data);

    // Grab the green data pointer for loading and plotting.
    blue_data_ptr = blue_data->getAxisDataPtr();

    // Make sure we have data to plot with, if not log fatal error.
    if (!blue_data_ptr || !x_axis_data_ptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Couldn't allocate blue_data array!";
        return 0;
    }
    for (i = 0; i < num_entries; i++)
    {
        // Convert the data to be floating point for the plotting library.
        blue_data_ptr[ i ] = double(curve_data[ i ]);

        // Is the data non-zero?
        if (curve_data[ i ] != 0)
        {
            // Did we find a new peak?
            if (peak_value < curve_data[ i ])
            {
                peak_value = curve_data[ i ];
                peak_index = i;
            }

            // Do we have a new min?
            if (i < min_index)
                min_index = i;

            // Do we have a new max?
            if (i > max_index)
                max_index = i;
        }
    }

    p_blue_plot = new QwtPlotCurve();
    p_blue_plot->setStyle(QwtPlotCurve::Lines);
    p_blue_plot->setRenderHint(QwtPlotItem::RenderAntialiased);
    p_blue_plot->setPen(QPen(Qt::blue));
    p_blue_plot->setRawSamples(x_axis_data_ptr, blue_data_ptr, num_entries);
    p_blue_plot->attach(ui->histo_plot);

    // Update our text.
    ui->min_value_blue->setValue(min_index);
    ui->max_value_blue->setValue(max_index);
    info_str = QString::fromUtf8("%1@%2").arg(peak_value).arg(peak_index);
    ui->peak_value_blue->setText(info_str);

    // Return our peak value.
    return peak_value;
}

/*!
 * @brief Given the "curve_data", build up the "green" curve of the histogram.
 *
 * @param curve_data        Pointer to the green histogram data.
 * @param num_entries       Number of points in the data buffer.
 *
 * @return The maximum value found.
 */
uint32_t fiHistogram::build_green_curve(const uint32_t *curve_data, int num_entries)
{
    int         i;
    int         min_index;
    int         max_index;
    int         peak_index;
    uint32_t    peak_value;
    QString     info_str;

    double     *x_axis_data_ptr  = nullptr;
    double     *grn_data_ptr     = nullptr;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++GREEN CURVE BUILD+++ num_entries:" << num_entries;
#endif

    // Initialize our counters.
    min_index = num_entries;
    max_index = 0;
    peak_index = 0;
    peak_value = 0;

    // Grab the x-axis data pointer for plotting.
    if (x_axis_data)
        x_axis_data_ptr = x_axis_data->getAxisDataPtr();

    grn_data = fiAxisData::allocateData(num_entries, grn_data);

    // Grab the green data pointer for loading and plotting.
    grn_data_ptr = grn_data->getAxisDataPtr();

    // Make sure we have data to plot with, if not log fatal error.
    if (!grn_data_ptr || !x_axis_data_ptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Couldn't allocate green_data array!";
        return 0;
    }
    for (i = 0; i < num_entries; i++)
    {
        // Convert the data to be floating point for the plotting library.
        grn_data_ptr[ i ] = double(curve_data[ i ]);

        // Is the data non-zero?
        if (curve_data[ i ] != 0)
        {
            // Did we find a new peak?
            if (peak_value < curve_data[ i ])
            {
                peak_value = curve_data[ i ];
                peak_index = i;
            }

            // Do we have a new min?
            if (i < min_index)
                min_index = i;

            // Do we have a new max?
            if (i > max_index)
                max_index = i;
        }
    }

    p_grn_plot = new QwtPlotCurve();
    p_grn_plot->setStyle(QwtPlotCurve::Lines);
    p_grn_plot->setRenderHint(QwtPlotItem::RenderAntialiased);
    p_grn_plot->setPen(QPen(Qt::green));
    p_grn_plot->setRawSamples(x_axis_data_ptr, grn_data_ptr, num_entries);
    p_grn_plot->attach(ui->histo_plot);

    // Update our text.
    ui->min_value_green->setValue(min_index);
    ui->max_value_green->setValue(max_index);
    info_str = QString::fromUtf8("%1@%2").arg(peak_value).arg(peak_index);
    ui->peak_value_green->setText(info_str);


    // Return our peak value.
    return peak_value;
}

/*!
 * @brief Given the "curve_data", build up the "red" curve of the histogram.
 *
 * @param curve_data        Pointer to the red histogram data.
 * @param num_entries       Number of points in the data buffer.
 *
 * @return The maximum value found.
 */
uint32_t fiHistogram::build_red_curve(const uint32_t *curve_data, int num_entries)
{
    int         i;
    int         min_index;
    int         max_index;
    int         peak_index;
    uint32_t    peak_value;
    QString     info_str;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++RED CURVE BUILD+++ num_entries:" << num_entries;
#endif

    double     *x_axis_data_ptr  = nullptr;
    double     *red_data_ptr     = nullptr;

    // Initialize our counters.
    min_index = num_entries;
    max_index = 0;
    peak_index = 0;
    peak_value = 0;

    // Grab the x-axis data pointer for plotting.
    if (x_axis_data)
        x_axis_data_ptr = x_axis_data->getAxisDataPtr();

    red_data = fiAxisData::allocateData(num_entries, red_data);

    // Grab the red data pointer for loading and plotting.
    red_data_ptr = red_data->getAxisDataPtr();

    // Make sure the pointers exist.
    if (!red_data_ptr || !x_axis_data_ptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Couldn't allocate red_data array!";
        return 0;
    }

    for (i = 0; i < num_entries; i++)
    {
        // Convert the data to be floating point for the plotting library.
        red_data_ptr[ i ] = double(curve_data[ i ]);

        // Is the data non-zero?
        if (curve_data[ i ] != 0)
        {
            // Did we find a new peak?
            if (peak_value < curve_data[ i ])
            {
                peak_value = curve_data[ i ];
                peak_index = i;
            }

            // Do we have a new min?
            if (i < min_index)
                min_index = i;

            // Do we have a new max?
            if (i > max_index)
                max_index = i;
        }
    }

    p_red_plot = new QwtPlotCurve();
    p_red_plot->setStyle(QwtPlotCurve::Lines);
    p_red_plot->setRenderHint(QwtPlotItem::RenderAntialiased);
    p_red_plot->setPen(QPen(Qt::red));
    p_red_plot->setRawSamples(x_axis_data_ptr, red_data_ptr, num_entries);
    p_red_plot->attach(ui->histo_plot);

    // Update our text.
    ui->min_value_red->setValue(min_index);
    ui->max_value_red->setValue(max_index);
    info_str = QString::fromUtf8("%1@%2").arg(peak_value).arg(peak_index);
    ui->peak_value_red->setText(info_str);

    // Return our peak value.
    return peak_value;
}

/*!
 * @brief Release the plot data buffers.
 */
void fiHistogram::free_plots(void)
{
#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++FREE PLOTS+++";
#endif

    if (p_gray_plot != nullptr)
    {
        p_gray_plot->detach();
        delete p_gray_plot;
        p_gray_plot = nullptr;
    }
    if (p_red_plot != nullptr)
    {
        p_red_plot->detach();
        delete p_red_plot;
        p_red_plot = nullptr;
    }
    if (p_grn_plot != nullptr)
    {
        p_grn_plot->detach();
        delete p_grn_plot;
        p_grn_plot = nullptr;
    }
    if (p_blue_plot != nullptr)
    {
        p_blue_plot->detach();
        delete p_blue_plot;
        p_blue_plot = nullptr;
    }
    if (p_gp_min != nullptr)
    {
        p_gp_min->detach();
        delete p_gp_min;
        p_gp_min = nullptr;
    }
    if (p_gp_max != nullptr)
    {
        p_gp_max->detach();
        delete p_gp_max;
        p_gp_max = nullptr;
    }
}

void fiHistogram::get_count_at(int& x_pos, double& r_val, double& g_val, double& b_val)
{
    // Make sure our index is clipped appropriately.
    if (x_pos < 0)
        x_pos = 0;
    else if (x_pos >= x_count)
        x_pos = x_count - 1;

    // Sanity check.
    if (x_pos < 0)
        return;

    // Are we mono?
    if (is_mono())
    {
        // Return the count in the "mono" value position.
        r_val = FI_HISTOGRAM::fiAxisData::getAxisElement(gray_data, x_pos);
        g_val = b_val = 0;
    }
    else
    {
        // Get the counts at this position.
        r_val = fiAxisData::getAxisElement(red_data, x_pos);
        g_val = fiAxisData::getAxisElement(grn_data, x_pos);
        b_val = fiAxisData::getAxisElement(blue_data, x_pos);
    }
}

const QVector<double> fiHistogram::get_mono_data()
{
    QVector<double> ret_data = FI_HISTOGRAM::fiAxisData::copyAxisData(x_count, gray_data);
    return ret_data;
}

const QVector<double> fiHistogram::get_red_data()
{
    QVector<double> ret_data = FI_HISTOGRAM::fiAxisData::copyAxisData(x_count, red_data);
    return ret_data;
}

const QVector<double> fiHistogram::get_green_data()
{
    QVector<double> ret_data = FI_HISTOGRAM::fiAxisData::copyAxisData(x_count, grn_data);
    return ret_data;
}

const QVector<double> fiHistogram::get_blue_data()
{
    QVector<double> ret_data = FI_HISTOGRAM::fiAxisData::copyAxisData(x_count, blue_data);
    return ret_data;
}

/*!
 * @brief Initialize the x-axis of the histogram.
 * @param num_entries   The number of points in the histogram.
 */
void fiHistogram::init_x_axis(int num_entries)
{
    int i;
    double *x_axis_data_ptr = nullptr;

    // Set our x_count to zero until everything is ready.
    x_count = 0;

    // Reserve the x-axis data.
    x_axis_data = fiAxisData::allocateData(num_entries, x_axis_data);

    if (x_axis_data)
        x_axis_data_ptr = x_axis_data->getAxisDataPtr();

    // Make sure x-axis data exists.
    if (!x_axis_data_ptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Couldn't allocate x_axis_data array!";
        return;
    }

    // Double check to make sure that the num_entries exist in x_axis_data.
    if (x_axis_data->getElementsInUse() != num_entries)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "The x_axis_data.theDataUsed" << x_axis_data->getElementsInUse() << "is not equal to num_entries:" << num_entries << "!";
        return;
    }

    // Remember the number of entries in use for x-axis of the histogram.
    x_count = num_entries;

    // Fill the x-axis data entries.
    for (i = 0; i < num_entries; i++)
    {
        x_axis_data_ptr[ i ] = double(i);
    }
}

/*!
 * @brief Use the "histo_data" buffer to build up the histogram plot.
 * @param histo_data    The array of mono/raw/RGB histogram data.
 */
void fiHistogram::populate_histogram(QVector<uint32_t>& histo_data)
{
    int     num_entries;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
#endif

    // Are we processing a raw image?
    if (b_raw)
    {
        // The number of entries is the number of 32-bit words.
        num_entries = histo_data.size();

        // Populate the data.
        populate_raw(histo_data.constData(), num_entries);
    }

    // Are we a mono image?
    else if (b_mono)
    {
        // The number of entries is the number of 32-bit words.
        num_entries = histo_data.size();

        // Populate the data.
        populate_mono(histo_data.constData(), num_entries);
    }

    // Must be color.
    else
    {
        int r_index;
        int g_index;
        int b_index;
        const uint32_t *rgb_buffer;

        // Calculate the number of entries per channel.
        num_entries = histo_data.size() / 3;

        // Initialize our indicies.
        r_index = 0;
        g_index = r_index + num_entries;
        b_index = g_index + num_entries;

        // Get an appropriately sized pointer to manipulate the data.
        rgb_buffer = histo_data.constData();

        // Populate the data.
        populate_rgb(&rgb_buffer[ r_index ], &rgb_buffer[ g_index ], &rgb_buffer[ b_index ], num_entries);
    }
}

/*!
 * @brief Populate a monochrome histogram data plot.
 *
 * @param mono_data     The array of monochrome histogram data.
 * @param num_entries   The number of entries in the array.
 */
void fiHistogram::populate_mono(const uint32_t *mono_data, int num_entries)
{
    int         i;
    int         min_index;
    int         max_index;
    int         peak_index;
    uint32_t    peak_value;
    QString     info_str;
    double     *gray_axis_data_ptr   = nullptr;
    double     *x_axis_data_ptr      = nullptr;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++MONO+++ num_entries:" << num_entries;
#endif

    // Sanity check.
    if (!b_mono)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Not mono camera!";
        return;
    }

    // if the curves already exist then detach and clean them up before doing the new ones.
    free_plots();

    // Initialize the x-axis information.
    init_x_axis(num_entries);

    // Initialize our counters.
    min_index = num_entries;
    max_index = 0;
    peak_index = 0;
    peak_value = 0;

    // Grab the x-axis data pointer for plotting.
    if (x_axis_data)
        x_axis_data_ptr = x_axis_data->getAxisDataPtr();

    // Insert new curves.
    gray_data = fiAxisData::allocateData(num_entries, gray_data);

    // Pull a pointer to the gray axis data for loading and plotting.
    if (gray_data)
        gray_axis_data_ptr = gray_data->getAxisDataPtr();

    // If we don't have a place to put this data we can't continue.
    if (!gray_axis_data_ptr || !x_axis_data_ptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Couldn't allocate axis data!";
        return;
    }

    // Load up the data.
    for (i = 0; i < num_entries; i++)
    {
        gray_axis_data_ptr[ i ] = double(mono_data[ i ]);

        // Is the data non-zero?
        if (mono_data[ i ] != 0)
        {
            // Did we find a new peak?
            if (peak_value < mono_data[ i ])
            {
                peak_value = mono_data[ i ];
                peak_index = i;
            }

            // Do we have a new min?
            if (i < min_index)
                min_index = i;

            // Do we have a new max?
            if (i > max_index)
                max_index = i;
        }
    }

    // Create our plot curve.
    p_gray_plot = new QwtPlotCurve();
    p_gray_plot->setStyle(QwtPlotCurve::Lines);
    p_gray_plot->setRenderHint(QwtPlotItem::RenderAntialiased);
    p_gray_plot->setPen(QPen(Qt::black));
    p_gray_plot->setRawSamples(x_axis_data_ptr, gray_axis_data_ptr, num_entries);
    p_gray_plot->attach(ui->histo_plot);

    // Update our goal post markers.
    update_goal_posts(mono_data[ 0 ], mono_data[ num_entries - 1 ], num_entries);

    // Update the X- and Y-axis scaling.
    set_x_axis(0, (uint32_t)num_entries);
    set_y_axis(0, peak_value);

    // Update the axis and replot.
    ui->histo_plot->updateAxes();
    ui->histo_plot->replot();

    // Update our text.
    ui->min_value_blue->setValue(min_index);
    ui->max_value_blue->setValue(max_index);
    info_str = QString::fromUtf8("%1@%2").arg(peak_value).arg(peak_index);
    ui->peak_value_blue->setText(info_str);
}

/*!
 * @brief Populate a raw histogram data plot. A raw histogram is like a monochrome
 *        plot, but reflects raw sensor values rather than process mono or RGB pixels.
 *
 * @param mono_data     The array of raw histogram data.
 * @param num_entries   The number of entries in the array.
 */
void fiHistogram::populate_raw(const uint32_t *raw_data, int num_entries)
{
    int         i;
    int         min_index;
    int         max_index;
    int         peak_index;
    uint32_t    peak_value;
    QString     info_str;

    double     *x_axis_data_ptr      = nullptr;
    double     *gray_axis_data_ptr   = nullptr;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++RAW+++ num_entries:" << num_entries;
#endif

    // Sanity check.
    if (!b_raw)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Not raw?";
        return;
    }

    // If the curves already exist then detach and clean them up before doing the new ones.
    free_plots();

    // Initialize the x-axis information.
    init_x_axis(num_entries);

    // Initialize our counters.
    min_index = num_entries;
    max_index = 0;
    peak_index = 0;
    peak_value = 0;

    // Grab the x-axis data pointer for plotting.
    if (x_axis_data)
        x_axis_data_ptr = x_axis_data->getAxisDataPtr();

    // Insert new curves.
    gray_data = fiAxisData::allocateData(num_entries, gray_data);

    // Get access to the axis data pointer.
    if (gray_data)
        gray_axis_data_ptr = gray_data->getAxisDataPtr();

    // If can't get access to the axis data pointer we can't fill it.
    if (!gray_axis_data_ptr || !x_axis_data_ptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Couldn't allocate axis data!";
        return;
    }

    // Fill the axis data.
    for (i = 0; i < num_entries; i++)
    {
        gray_axis_data_ptr[ i ] = double(raw_data[ i ]);

        // Is the data non-zero?
        if (raw_data[ i ] != 0)
        {
            // Did we find a new peak?
            if (peak_value < raw_data[ i ])
            {
                peak_value = raw_data[ i ];
                peak_index = i;
            }

            // Do we have a new min?
            if (i < min_index)
                min_index = i;

            // Do we have a new max?
            if (i > max_index)
                max_index = i;
        }
    }

    // Create our plot curve.
    p_gray_plot = new QwtPlotCurve();
    p_gray_plot->setStyle(QwtPlotCurve::Lines);
    p_gray_plot->setRenderHint(QwtPlotItem::RenderAntialiased);
    p_gray_plot->setPen(QPen(Qt::gray));
    p_gray_plot->setRawSamples(x_axis_data_ptr, gray_axis_data_ptr, num_entries);
    p_gray_plot->attach(ui->histo_plot);

    // Update our goal post markers.
    update_goal_posts(raw_data[ 0 ], raw_data[ num_entries - 1 ], num_entries);

    // Update the X- and Y-axis scaling.
    set_x_axis(0, (uint32_t)num_entries);
    set_y_axis(0, peak_value);

    // Update the axis and replot.
    ui->histo_plot->updateAxes();
    ui->histo_plot->replot();

    // Update our text.
    ui->min_value_blue->setValue(min_index);
    ui->max_value_blue->setValue(max_index);
    info_str = QString::fromUtf8("%1@%2").arg(peak_value).arg(peak_index);
    ui->peak_value_blue->setText(info_str);
}

/*!
 * @brief Populate a RGB histogram data plot.
 *
 * The array is a concatentated array, where the first 1/3rd is the red, the next
 * contains the green data and the last 1/3rd contain the blue data.
 *
 * @param mono_data     The array of RGB histogram data.
 * @param num_entries   The number of entries in the array.
 */
void fiHistogram::populate_rgb(const uint32_t *r_data, const uint32_t *g_data, const uint32_t *b_data, int num_entries)
{
    uint32_t    peak_value;
    uint32_t    peak_overall;
    uint32_t    peak_min_value;
    uint32_t    peak_max_value;

#if LOG_OPERATIONAL_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "+++RGB+++ num_entries:" << num_entries;
#endif

    // Check to see if we should be calling the mono plot.
    if (b_mono)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Should have called populate_mono()!";
        populate_mono(r_data, num_entries);
        return;
    }

    // If the curves already exist then detach and clean them up before doing the new ones.
    free_plots();

    // Initialize the x-axis information.
    init_x_axis(num_entries);

    // Save the peak values.
    peak_min_value = r_data[ 0 ];
    peak_max_value = r_data[ num_entries - 1 ];
    peak_overall = 0;

    // Build up the red curve.
    peak_value = build_red_curve(r_data, num_entries);

    // Is this the overall peak value?
    if (peak_overall < peak_value)
        peak_overall = peak_value;

    // Are the greens larger than the red?
    if (peak_min_value < g_data[ 0 ])
        peak_min_value = g_data[ 0 ];
    if (peak_max_value < g_data[ num_entries - 1 ])
        peak_max_value = g_data[ num_entries - 1 ];

    // Build up the green curve.
    peak_value = build_green_curve(g_data, num_entries);

    // Is this the overall peak value?
    if (peak_overall < peak_value)
        peak_overall = peak_value;

    // Are the blues the largest?
    if (peak_min_value < b_data[ 0 ])
        peak_min_value = b_data[ 0 ];
    if (peak_max_value < b_data[ num_entries - 1 ])
        peak_max_value = b_data[ num_entries - 1 ];

    // Build up the blue curve.
    peak_value = build_blue_curve(b_data, num_entries);

    // Is this the overall peak value?
    if (peak_overall < peak_value)
        peak_overall = peak_value;

    // Update our goal post markers.
    update_goal_posts(peak_min_value, peak_max_value, num_entries);

    // Update the X- and Y-axis scaling.
    set_x_axis(0, (uint32_t)num_entries);
    set_y_axis(0, peak_overall);

    // Update the axis and replot.
    ui->histo_plot->updateAxes();
    ui->histo_plot->replot();
}

/*!
 * @brief Set the histogram to a mono or color histogram.
 * @param mono  True for monochrom, false for RGB.
 */
void fiHistogram::set_mono(bool mono)
{
    // Any change?
    if (b_mono == mono)
        return;

    // Save the flag.
    b_mono = mono;

    // If we're mono, we want to change the blue group's text.
    if (b_mono)
    {
        // Since we're mono, we don't want the group name to be a color.
        ui->blue_group->setTitle(tr("Statistics"));
    }
    else
    {
        // Restore the group name.
        ui->blue_group->setTitle(tr("Blue"));
    }

    // Update our statistics display, if they are visible.
    if (stats_are_visible())
        show_statistics(true);

    // Update ourselves.
    update();
}

/*!
 * @brief Enable/disable raw mode display.
 * @param use_raw   True to use raw mode, false otherwise.
 */
void fiHistogram::set_raw_mode(bool use_raw)
{
    // Remember if we're using raw mode or not.
    b_raw = use_raw;

    // If we're using raw mode, then our histogram is in mono-mode too.
    if (use_raw)
        set_mono(true);
}

/*!
 * @brief Set the statistics for a given [curve](@ref FI_HISTOGRAM::CURVE_TYPE).
 *
 * @param curve         One of the FI_HISTOGRAM::CURVE_TYPE enumerations.
 * @param sum_x         The "sum" value.
 * @param sum_sq_x      The "sum of squares" value.
 */
void fiHistogram::set_statistics(int curve, uint32_t sum_x, uint64_t sum_sq_x)
{
    QString sum_str;
    QString sum_sq_str;

    // Build up our strings.
    sum_str = QString::number(sum_x);
    sum_sq_str = QString::number(sum_sq_x);

    // Add the statistics to the appropriate "curve".
    switch (curve)
    {
    case BLUE_CURVE:
    case MONO_CURVE:
        ui->sum_x_value_blue->setText(sum_str);
        ui->sum_sq_value_blue->setText(sum_sq_str);
        break;
    case GREEN_CURVE:
        ui->sum_x_value_green->setText(sum_str);
        ui->sum_sq_value_green->setText(sum_sq_str);
        break;
    case RED_CURVE:
        ui->sum_x_value_red->setText(sum_str);
        ui->sum_sq_value_red->setText(sum_sq_str);
        break;
    default:
        break;
    }
}

/*!
 * @brief Set the minimum and maximum value in the X-axis.
 *
 * @param min       The minimum value for the X-axis.
 * @param max       The maximum value for the X-axis.
 */
void fiHistogram::set_x_axis(uint32_t min, uint32_t max)
{
    ui->histo_plot->setAxisScale(QwtPlot::xBottom, double(min), double(max));
}

/*!
 * @brief Set the minimum and maximum values for the Y-axis.
 *
 * @param min       The minimum value for the Y-axis.
 * @param max       The maximum value for the Y-axis.
 */
void fiHistogram::set_y_axis(uint32_t min, uint32_t max)
{
    ui->histo_plot->setAxisScale(QwtPlot::yLeft, double(min), double(max) * y_scale);
}

/*!
 * @brief Show or hide the statistics information.
 * @param show  True to show the statistics, false to hide them.
 */
void fiHistogram::show_statistics(bool show)
{
    // Blue is used for mono, color and raw.
    ui->blue_group->setVisible(show);

    // Only do these if we're not mono.
    if (!b_mono)
    {
        ui->green_group->setVisible(show);
        ui->red_group->setVisible(show);
    }
    else
    {
        // Make sure these are hidden for mono.
        ui->green_group->setVisible(false);
        ui->red_group->setVisible(false);
    }
}

/*!
 * @brief Return true if the statistics are visible, false otherwise.
 */
bool fiHistogram::stats_are_visible(void)
{
    return ui->blue_group->isVisible();
}

/*!
 * @brief Update the "goal post" curves at each side of the histogram plot.
 *
 * These "goal posts" are used to more easily see saturation at each end of
 * the histogram.
 *
 * @param min_peak      The maximum number of entries at the minimum point of the X-axis.
 * @param max_peak      The maximum number of entries at the maximum point of the X-axis.
 * @param num_entries   The number of entries along the X-axis.
 */
void fiHistogram::update_goal_posts(uint32_t min_peak, uint32_t max_peak, int num_entries)
{
    int i;
    
    // Build up the low-side goal post.
    for (i = 0; i < GP_COL_WIDTH; ++i)
    {
        gp_min_x_points[ i ] = (double)i;
        gp_min_y_points[ i ] = (double)min_peak;
    }
    gp_min_y_points[ GP_COL_WIDTH - 1 ] = 0.0;

    // Create our goal posts.
    p_gp_min = new QwtPlotCurve();
    p_gp_min->setStyle(QwtPlotCurve::Lines);
    p_gp_min->setRenderHint(QwtPlotItem::RenderAntialiased);
    // p_gp_min->setLegendAttribute( QwtPlotCurve::LegendShowBrush, true );
    p_gp_min->setPen(QPen(Qt::magenta));
    p_gp_min->setBrush(QBrush(Qt::magenta));
    p_gp_min->setRawSamples(gp_min_x_points.get(), gp_min_y_points.get(), GP_COL_WIDTH);
    p_gp_min->attach(ui->histo_plot);

    // Build up the high-side goal post.
    for (i = 0; i < GP_COL_WIDTH; ++i)
    {
        gp_max_x_points[ i ] = (double)((num_entries - GP_COL_WIDTH) + i);
        gp_max_y_points[ i ] = (double)max_peak;
    }
    gp_max_y_points[ 0 ] = 0.0;

    // Create our goal posts.
    p_gp_max = new QwtPlotCurve();
    p_gp_max->setStyle(QwtPlotCurve::Lines);
    p_gp_max->setRenderHint(QwtPlotItem::RenderAntialiased);
    // p_gp_max->setLegendAttribute( QwtPlotCurve::LegendShowBrush, true );
    p_gp_max->setPen(QPen(Qt::magenta));
    p_gp_max->setBrush(QBrush(Qt::magenta));
    p_gp_max->setRawSamples(gp_max_x_points.get(), gp_max_y_points.get(), GP_COL_WIDTH);
    p_gp_max->attach(ui->histo_plot);
}

/*!
 * @brief Perform platform-specific updates to the gradient of the plot.
 */
void fiHistogram::update_gradient(void)
{
    QPalette pal = palette();

    const QColor buttonColor = pal.color(QPalette::Button);

#ifdef Q_OS_LINUX
    // Qt 4.7.1: QGradient::StretchToDeviceMode is buggy on X11
    QLinearGradient gradient(rect().topLeft(), rect().bottomLeft());
    gradient.setColorAt(0.0, Qt::white);
    gradient.setColorAt(0.7, buttonColor);
    gradient.setColorAt(1.0, buttonColor);
#else
    QLinearGradient gradient(0, 0, 0, 1);
    gradient.setCoordinateMode(QGradient::StretchToDeviceMode);
    gradient.setColorAt(0.0, Qt::white);
    gradient.setColorAt(0.7, buttonColor);
    gradient.setColorAt(1.0, buttonColor);
#endif

    pal.setBrush(QPalette::Window, gradient);
    ui->histo_plot->setPalette(pal);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void fiHistogram::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        ui->retranslateUi(this);
    }

    // Pass it on.
    QWidget::changeEvent(event);
}

/*!
 * @brief Catch the resize event.
 * @param event     The event pointer.
 */
void fiHistogram::resizeEvent(QResizeEvent *event)
{
    Q_UNUSED(event);

#ifdef Q_OS_LINUX
    update_gradient();
#endif
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Hide or show the statistics information for the blue curve.
 * @param clicked   True to show, false to hide.
 */
void fiHistogram::blue_box_clicked(bool clicked)
{
    ui->blue_frame->setVisible(clicked);
}

/*!
 * @brief Hide or show the statistics information for the green curve.
 * @param clicked   True to show, false to hide.
 */
void fiHistogram::green_box_clicked(bool clicked)
{
    ui->green_frame->setVisible(clicked);
}

/*!
 * @brief Hide or show the statistics information for the red curve.
 * @param clicked   True to show, false to hide.
 */
void fiHistogram::red_box_clicked(bool clicked)
{
    ui->red_frame->setVisible(clicked);
}
