/*!
 * @file fi_histogram.h
 * @brief This is the header file for the Fastec Imaging Histogram object.
 *
 * The purpose of this object is to place all the histogram display
 * functionality in one reusable "widget". This widget can then be
 * "placed" in dialogs and other UI elements as needed.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_HISTOGRAM_H
#define FI_HISTOGRAM_H

//--------------------------------------------------------------------
// Include Files
//--------------------------------------------------------------------
#include <QWidget>
#include <QVector>
#include <stdint.h>

namespace Ui
{
class fiHistogram;
}

namespace FI_HISTOGRAM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! An enumeration to define the curves of the histogram.
enum CURVE_TYPE
{
    RED_CURVE,      //!< Operations are to affect only the RED curve of the histogram.
    GREEN_CURVE,    //!< Operations are to affect only the GREEN curve of the histogram.
    BLUE_CURVE,     //!< Operations are to affect only the BLUE curve of the histogram.
    MONO_CURVE      //!< Operations are for the MONO/Raw curve ofthe histogram.
};

// Forward declaration of private axis data class.
class fiAxisData;

}   // End of namespace FI_HISTOGRAM.

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class QByteArray;
class QwtPlotCurve;
class QwtPlotPicker;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class fiHistogram : public QWidget
{
    Q_OBJECT

public:
    explicit fiHistogram(QWidget *parent = nullptr);
    ~fiHistogram();

    void    get_count_at(int& x_pos, double& r_val, double& g_val, double& b_val);
    double  get_y_scale() const                     { return y_scale; }

    const QVector<double> get_mono_data();
    const QVector<double> get_red_data();
    const QVector<double> get_green_data();
    const QVector<double> get_blue_data();

    bool    is_mono() const                         { return b_mono; }
    bool    is_raw() const                          { return b_raw; }

    void    populate_histogram(QVector<uint32_t>& histo_data);

    void    set_mono(bool mono);
    void    set_raw_mode(bool use_raw);
    void    set_statistics(int curve, uint32_t sum_x, uint64_t sum_sq_x);
    void    set_x_axis(uint32_t min, uint32_t max);
    void    set_y_axis(uint32_t min, uint32_t max);
    void    set_y_scale(double scale)               { y_scale = scale; }

    void    show_statistics(bool show);
    bool    stats_are_visible();

protected:
    virtual void    changeEvent(QEvent *event);
    virtual void    resizeEvent(QResizeEvent *event);

private slots:
    void    blue_box_clicked(bool clicked);
    void    green_box_clicked(bool clicked);
    void    red_box_clicked(bool clicked);

private:
    uint32_t    build_red_curve(const uint32_t *curve_data, int num_entries);
    uint32_t    build_green_curve(const uint32_t *curve_data, int num_entries);
    uint32_t    build_blue_curve(const uint32_t *curve_data, int num_entries);

    void    init_x_axis(int num_entries);
    void    free_plots();

    void    populate_mono(const uint32_t *mono_data, int num_entries);
    void    populate_raw(const uint32_t *raw_data, int num_entries);
    void    populate_rgb(const uint32_t *r_data, const uint32_t *g_data, const uint32_t *b_data, int num_entries);

    void    update_goal_posts(uint32_t min_peak, uint32_t max_peak, int num_entries);
    void    update_gradient();

private:
    Ui::fiHistogram *ui;                    //!< Our UI elements object.

    QwtPlotCurve    *p_gray_plot;       //!< The plot being done for a mono camera.
    QwtPlotCurve    *p_red_plot;        //!< The plot being done for red of a color camera.
    QwtPlotCurve    *p_grn_plot;        //!< The plot being done for green of a color camera.
    QwtPlotCurve    *p_blue_plot;       //!< The plot being done for blue of a color camera.
    QwtPlotPicker   *histo_tracker;     //!< Object to track mouse movement across the plot.

    bool    b_mono;                         //!< Flag indicating if we are a mono (true) or color (false) camera.
    bool    b_raw;                          //!< Flag indicating if we are display raw or RGB/mono data.
    int     x_count;                        //!< Number of entries on the X axis (and y-channel(s) axis).

    std::shared_ptr<FI_HISTOGRAM::fiAxisData> x_axis_data;  //!< X axis histogram data.
    std::shared_ptr<FI_HISTOGRAM::fiAxisData> gray_data;    //!< Y axis gray data for mono camera.
    std::shared_ptr<FI_HISTOGRAM::fiAxisData> red_data;     //!< Y axis red data for color camera.
    std::shared_ptr<FI_HISTOGRAM::fiAxisData> grn_data;     //!< Y axis green data for color camera.
    std::shared_ptr<FI_HISTOGRAM::fiAxisData> blue_data;    //!< Y axis blue data for color camera.

    QwtPlotCurve    *p_gp_min;              //!< "Goal Post" at 0.
    QwtPlotCurve    *p_gp_max;              //!< "Goal Post" at the maximum value.

    std::unique_ptr<double[]> gp_min_x_points;      //!< Our goal post x-axis data for the minimum side.
    std::unique_ptr<double[]> gp_min_y_points;      //!< Our goal post y-axis data for the minimum side.
    std::unique_ptr<double[]> gp_max_x_points;      //!< Our goal post x-axis data for the maximum side.
    std::unique_ptr<double[]> gp_max_y_points;      //!< Our goal post y-axis data for the maximum side.

    double y_scale;                         //!< Scaling to be applied to the Y-axis.
};

#endif // FI_HISTOGRAM_H
