/*!
 * @file fi_image_buffer.h
 *
 * @brief A class that manages information about an image data
 * resource in a stream.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <memory>
#include <QHostAddress>
#include <QMutexLocker>

#include "logger.h"
#include "fi_mem_pool.h"
#include "fi_image_buffer.h"

namespace FI_MEMPOOL
{

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------
const bool SHOW_ID_LOGS = true;

} // End of the FI_MEMPOOL namespace declaration.

// Make things easier for usage.
using namespace std;
using namespace FI_MEMPOOL;

//--------------------------------------------------------------------
// Implementation of the Image Buffer Class
//--------------------------------------------------------------------

/*!
 * @brief Constructor for the image buffer object.
 */
FiImageBuffer::FiImageBuffer() :
    seqNumber(0),
    warnIfDestructed(true)
{
}

/*!
 * @brief Destructor for the image buffer object.
 */
FiImageBuffer::~FiImageBuffer()
{
    if (warnIfDestructed)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Destructing an FiImageBuffer object before its time!";
    }

    // Go ahead and let the image resource destruct which will recycle the resource.
    if (imageResourcePtr)
    {
        imageResourcePtr.reset();
    }
}

/*!
 * @brief Factory to create a image buffer.
 * @returns The newly created image buffer transfers ownership to caller.
 */
std::unique_ptr<FI_MEMPOOL::FiImageBuffer> FiImageBuffer::CreateImageBuffer()
{
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> createdImageBuffPtr(new FiImageBuffer());

    //std::unique_ptr<FI_MEMPOOL::FiImageBuffer> createdImageBuffPtr =
    //    std::unique_ptr<FI_MEMPOOL::FiImageBuffer>(new FiImageBuffer());

    // If the creation failed write out a warning to the log.
    if (!createdImageBuffPtr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate an Image Buffer object!";
    }

    // Return the newly created image buffer object.
    return (createdImageBuffPtr);
}

/*!
 * @brief Attaches image data into this image buffer.
 *
 * @param imageDataPtrToAttach - The image data pointer that turns this image buffer
 * from a free image buffer into a receive image buffer.
 * @param recvSizeToExpect - how many received bytes are expected for this image data?
 * @param seqNumberToUse - a zero-based sequence number to associate with this image data.
 *
 * @returns The status of the installation.
 */
bool FiImageBuffer::AttachImageData(FI_MEMPOOL::FiMemoryResourcePtr imageResourcePtrToAttach)
{
    bool attachedImageData = false;

    if (imageResourcePtrToAttach == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Received an empty image data resource.";
        goto exitAttachImageData;
    }

    // Otherwise take ownership of the image data buffer.
    imageResourcePtr = std::move(imageResourcePtrToAttach);

    // Make sure to set the result to a valid attach state.
    attachedImageData = true;

exitAttachImageData:
    return attachedImageData;
}

/*!
 * @brief Returns the number of bytes comprising the image stored here.
 * @returns The number of bytes in use or zero if error or nothing attached.
 */
uint32_t FiImageBuffer::GetImageDataLen()
{
    uint32_t bytesInUse = 0;

    if (imageResourcePtr)
        bytesInUse = imageResourcePtr->GetLength();
    return bytesInUse;
}

/*!
 * @brief Returns a read-only pointer to the image data stored here.
 * @returns The valid read-only pointer if image data is attached or nullptr if not.
 */
const uint8_t *FiImageBuffer::GetMemDataPtr()
{
    const uint8_t *imageDataReadOnlyPtr = nullptr;

    if (imageResourcePtr)
    {
        uint32_t dataLength = 0;
        imageDataReadOnlyPtr = imageResourcePtr->GetReservedPtr(dataLength);
    }
    return imageDataReadOnlyPtr;
}

/*!
 * @brief Returns the sequence number set here.
 */
uint32_t FiImageBuffer::GetSequenceNumber()
{
    return seqNumber;
}

/*!
 * @brief Is this image buffer in a free state?
 * @returns True if there is no attached image data object, false if not.
 */
bool FiImageBuffer::isFree()
{
    // Just checks for the existence of the image resource object.
    if (imageResourcePtr)
        return true;
    return false;
}

/*!
 * @brief Gives this object an opportunity to prepare for destructing.
 *
 * At the moment this is for diagnostic purposes.  Each memory data will
 * get a unique id that recycles at 10,000.  This is being logged right now
 * to track resource movement.
 */
void FiImageBuffer::LogMyDataId(const QString *moreInfoStr)
{
    if (SHOW_ID_LOGS && (moreInfoStr != nullptr))
    {
        if (imageResourcePtr)
        {
            imageResourcePtr->LogMyId(moreInfoStr);
        }
        else
        {
            QLOG_DEBUG() << "[RESOURCE EMPTY]" << moreInfoStr << "[memDataPtr=nullptr]";
        }
    }
}

/*!
 * @brief Gives this object an opportunity to prepare for destructing.
 *
 * At the moment this is for diagnostic purposes.  Because the data
 * belonging to this object is managed by a unique_ptr it
 * is susceptable to a resource-leak in the system if it
 * is not properly transported.  This warning mechanism is strictly
 * that, if the object is destructed and the warning flag is true
 * (which it is after a factory creation) then the warning will
 * appear in the log file.
 */
void FiImageBuffer::PrepareToDestruct()
{
    warnIfDestructed = false;
}

/*!
 * @brief Releases the image buffer's image data back to the system image pool
 * resource and then resets this image buffer to a free and empty state.
 *
 * @param theImageBufferToFree - the image buffer that has been processed and will be
 * returned to the free buffer queue.
 *
 * @returns The result of the operation.
 */
FI_MEMPOOL::FiMemoryResourcePtr FiImageBuffer::ReleaseImageData()
{
    FI_MEMPOOL::FiMemoryResourcePtr imageDataPtrReleased;

    if (imageResourcePtr)
    {
        // Decouple the image data buffer from the image buffer record.
        imageDataPtrReleased = std::move(imageResourcePtr);
    }

    // Reset the local settings for the network receive buffer.
    seqNumber = 0;

    return imageDataPtrReleased;
}

/*!
 * @brief Replaces the image data buffer with new image bytes.
 *
 * @param memBytesToSet - A read only pointer to the new image bytes.
 * @param memBytesToSetLen - The number of image bytes to set.
 *
 * @returns The number of bytes that are now used in the image data.  If zero
 * the replacment failed.
 */
uint32_t FiImageBuffer::ReplaceImageBytes(uint32_t memBytesToSetLen, const uint8_t *memBytesToSetPtr)
{
    QString myLabelStr = QString("%1[FiImageBuffer::ReplaceImageBytes]").arg(memPoolLabelStr);
    uint32_t bytesInUse = 0;

    if (imageResourcePtr)
    {
        //! Will create a new resource if the current resource is not large enough to accomodate the data.
        FI_MEMPOOL::FiMemoryResourcePtr biggerImageResourcePtr;

        //! What is the existing image resource's capacity (how many bytes can it hold)?
        uint32_t dataCapacity = imageResourcePtr->GetCapacity();

        // Force a re-allocation of the resource.
        //dataCapacity = 0;

        // Need to make sure that there is enough room in the existing resource to
        // accomodate the size of the new image.
        if (dataCapacity < memBytesToSetLen)
        {
            // Pop out a diagnostic for logging purposes.
            // Note:  This was using VLOG_SHOW_ALWAYS but that is not really necessary
            // anymore since the memory pool is fully integrated.  This condition tends
            // to happen when large frames like RAW TIF images are being saved to PC from
            // a series3 camera and these messages becomes noisy so I removed VLOG_SHOW_ALWAYS.
            FiMemoryPool::LogVerbose((QString("%1 [dataCapacity=%2][#memBytesToSetLen=%3] Needs new resource to accomodate!").
                                      arg(myLabelStr).
                                      arg(dataCapacity).arg(memBytesToSetLen)).toUtf8().constData());

            // Attempt to get a replacement for the resource.  Allow emergency allocation here.
            biggerImageResourcePtr = FiMemoryPool::GetMemoryResource(memBytesToSetLen, true);
            if (!biggerImageResourcePtr)
            {
                FiMemoryPool::LogVerbose((QString("%1 [dataCapacity=%2][#memBytesToSetLen=%3] FAILED to aquire larger resource!").
                                          arg(myLabelStr).
                                          arg(dataCapacity).arg(memBytesToSetLen)).toUtf8().constData());
                goto exitReplaceImageBytes;
            }

            // Replace the imageResourcePtr to equal the replaceImageDataPtr.  The logic will
            // hold on to the old pointer just in case an error occurs during the attachment
            // process.  If the attachment process goes well then the old imageResourcePtr will
            // be reset and recycled internally through the system.
            FI_MEMPOOL::FiMemoryResourcePtr saveCurrentImageResourcePtr = std::move(imageResourcePtr);

            if (!AttachImageData(std::move(biggerImageResourcePtr)))
            {
                FiMemoryPool::LogVerbose((QString("%1 [dataCapacity=%2][#memBytesToSetLen=%3] FAILED to attach larger resource!").
                                          arg(myLabelStr).
                                          arg(dataCapacity).arg(memBytesToSetLen)).toUtf8().constData());

                imageResourcePtr = std::move(saveCurrentImageResourcePtr);  // Restore the old one...
                goto exitReplaceImageBytes;
            }

            // Go ahead and delete the old image resource ptr even though it could be deleted
            // by letting the attach replace it, this will ensure a valid time to delete it.
            saveCurrentImageResourcePtr.reset();

            // At this point simply append the larger data to the larger resource.
            if (imageResourcePtr)
            {
                bytesInUse = imageResourcePtr->AppendBytes(memBytesToSetLen, memBytesToSetPtr);
            }
            FiMemoryPool::LogVerbose((QString("%1 [dataCapacity=%2][#memBytesToSetLen=%3][bytesInUse=%4] REPLACED ENTIRE RESOURCE!").
                                      arg(myLabelStr).
                                      arg(dataCapacity).arg(memBytesToSetLen).arg(bytesInUse)).toUtf8().constData());
        }
        else
        {
            // There is enough capacity with existing resource so just replace the bytes.
            bytesInUse = imageResourcePtr->ReplaceBytes(memBytesToSetLen, memBytesToSetPtr);
        }
    }
exitReplaceImageBytes:
    return bytesInUse;
}

/*!
 * @brief Assigns a sequence number to this image buffer.
 *
 * @param seqNumberToSet - zero-based sequence number to tag to this image buffer.
 */
void FiImageBuffer::SetSequenceNumber(uint32_t seqNumberToSet)
{
    seqNumber = seqNumberToSet;
}

