/*!
 * @file fi_internal.cpp
 * @brief This provides the functionality for the FiInternal class.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QLibrary>
#include <QSettings>

#include "fi_internal.h"
#include "logger.h"

// ==============================================================
// Static methods should be defined outside the class.
// ==============================================================

std::unique_ptr<FiInternal> FiInternal::instance;
std::mutex FiInternal::instanceMutex;
QLibrary FiInternal::fasLib;

QString FiInternal::GET_PACKED_SIZE = QStringLiteral("fiGetPackedSize");
QString FiInternal::GET_HANDLE = QStringLiteral("fiCreatePackStreamer");
QString FiInternal::CLOSE_PACK_STREAMER = QStringLiteral("fiClosePackStreamer");
QString FiInternal::IS_PACK_STREAMER_FINISHED = QStringLiteral("fiIsPackStreamerFinished");
QString FiInternal::ADD_PACKED_DATA = QStringLiteral("fiAddPackedData");
QString FiInternal::ADD_TIFF_IMAGE_DATA = QStringLiteral("fiAddTiffImageData");
QString FiInternal::ADD_PLACEHOLDER = QStringLiteral("fiAddPlaceholder");
QString FiInternal::REPLACE_FRAME = QStringLiteral("fiReplaceFrame");
QString FiInternal::GET_PACK_STATISTICS = QStringLiteral("fiGetPackStatistics");

#if defined(Q_OS_WIN)
QString FiInternal::LIBRARY_NAME = QStringLiteral("libfas_internal.dll");
#elif defined(Q_OS_UNIX)
QString FiInternal::LIBRARY_NAME = QStringLiteral("fas_internal.so");
#else
QString FiInternal::LIBRARY_NAME = QStringLiteral("fas_internal.so");
#endif

FiInternal::FiInternal()
{
}

/*!
 * @brief This is the static method that controls the access to the singleton
 *        instance. On the first run, it creates a singleton object and places it
 *        into the static field. On subsequent runs, it returns the client existing
 *        object stored in the static field.
 */
FiInternal *FiInternal::factory()
{
    std::lock_guard<std::mutex> lock(instanceMutex);

    // Check to see if we have our singleton.
    if (instance.get() == nullptr)
    {
        // Attempt to load the library.
        fasLib.setFileName(LIBRARY_NAME);
        if (fasLib.load())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "has loaded.";

            // Create the one and only instance.
            instance = std::unique_ptr<FiInternal>(new FiInternal());
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "did not load.";
        }
    }

    return instance.get();
}

FiInternal::GET_PACKED_SIZE_FN FiInternal::getPackedSizeFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (GET_PACKED_SIZE_FN)fasLib.resolve(GET_PACKED_SIZE.toStdString().c_str());
}

FiInternal::GET_HANDLE_FN FiInternal::getHandleFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (GET_HANDLE_FN)fasLib.resolve(GET_HANDLE.toStdString().c_str());
}

FiInternal::CLOSE_PACK_STREAMER_FN FiInternal::getClosePackFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (CLOSE_PACK_STREAMER_FN)fasLib.resolve(CLOSE_PACK_STREAMER.toStdString().c_str());
}

FiInternal::IS_PACK_STREAMER_FINISHED_FN FiInternal::geIsFinishedFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (IS_PACK_STREAMER_FINISHED_FN)fasLib.resolve(IS_PACK_STREAMER_FINISHED.toStdString().c_str());
}

FiInternal::ADD_TIFF_IMAGE_DATA_FN FiInternal::getAddPackedFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (ADD_PACKED_DATA_FN)fasLib.resolve(ADD_PACKED_DATA.toStdString().c_str());
}

FiInternal::ADD_TIFF_IMAGE_DATA_FN FiInternal::getAddTiffFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (ADD_TIFF_IMAGE_DATA_FN)fasLib.resolve(ADD_TIFF_IMAGE_DATA.toStdString().c_str());
}

FiInternal::ADD_PLACEHOLDER_FN FiInternal::getAddPlaceholderFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (ADD_PLACEHOLDER_FN)fasLib.resolve(ADD_PLACEHOLDER.toStdString().c_str());
}

FiInternal::REPLACE_FRAME_FN FiInternal::getReplaceFrameFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (REPLACE_FRAME_FN)fasLib.resolve(REPLACE_FRAME.toStdString().c_str());
}

FiInternal::GET_PACK_STATISTICS_FN FiInternal::getGetStatisticsFuncPtr()
{
    // Sanity check.
    if (!fasLib.isLoaded())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << LIBRARY_NAME << "is not loaded.";
        return nullptr;
    }

    // Return the function pointer.
    return (GET_PACK_STATISTICS_FN)fasLib.resolve(GET_PACK_STATISTICS.toStdString().c_str());
}
