/*!
 * @file fi_internal.h
 *
 * @brief A class that manages access to the Fastec Internal
 *        Proprietary class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_INTERNAL_H
#define FI_INTERNAL_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QLibrary>
#include <memory>
#include <mutex>

#include "fas_internal.h"

//! The FiInternal class is a singleton object to control access to the
//! Fastec Internal Proprietary library.
class FiInternal
{
public:
    // Define the function "signature".
    typedef uint32_t (* GET_PACKED_SIZE_FN)(uint32_t width, uint32_t height, int bitsPerPixel);
    typedef FAS_INTERNAL::FIP_HANDLE(* GET_HANDLE_FN)(const std::string& streamName, const FAS_INTERNAL::FIP_STREAM_CONFIGURATION& configInfo);
    typedef FAS_INTERNAL::FIP_STATUS(* CLOSE_PACK_STREAMER_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle);
    typedef FAS_INTERNAL::FIP_STATUS(* IS_PACK_STREAMER_FINISHED_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle);
    typedef FAS_INTERNAL::FIP_STATUS(* ADD_PACKED_DATA_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle, const uint8_t *imagePtr);
    typedef FAS_INTERNAL::FIP_STATUS(* ADD_TIFF_IMAGE_DATA_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle, const uint8_t *imagePtr);
    typedef FAS_INTERNAL::FIP_STATUS(* ADD_PLACEHOLDER_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle);
    typedef FAS_INTERNAL::FIP_STATUS(* REPLACE_FRAME_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle, uint32_t seqNumber, const uint8_t *imagePtr);
    typedef FAS_INTERNAL::FIP_STATUS(* GET_PACK_STATISTICS_FN)(FAS_INTERNAL::FIP_HANDLE streamHandle, uint32_t& frameCount, uint32_t& lastFrame, uint32_t& encodedFrames);

public:
    ~FiInternal() {}

    // FiInternal should not be cloneable.
    FiInternal(FiInternal& other) = delete;

    // FiInternal should not be assignable.
    void operator=(const FiInternal&) = delete;

    /**
     * This is the static method that controls the access to the singleton
     * instance. On the first run, it creates a singleton object and places it
     * into the static field. On subsequent runs, it returns the client existing
     * object stored in the static field.
     */
    static FiInternal *factory();

    GET_PACKED_SIZE_FN getPackedSizeFuncPtr();
    GET_HANDLE_FN getHandleFuncPtr();
    CLOSE_PACK_STREAMER_FN getClosePackFuncPtr();
    IS_PACK_STREAMER_FINISHED_FN geIsFinishedFuncPtr();
    ADD_PACKED_DATA_FN getAddPackedFuncPtr();
    ADD_TIFF_IMAGE_DATA_FN getAddTiffFuncPtr();
    ADD_PLACEHOLDER_FN getAddPlaceholderFuncPtr();
    REPLACE_FRAME_FN getReplaceFrameFuncPtr();
    GET_PACK_STATISTICS_FN getGetStatisticsFuncPtr();

private:
    FiInternal();

private:
    static std::unique_ptr<FiInternal> instance;
    static std::mutex instanceMutex;

    static QString LIBRARY_NAME;
    static QString GET_PACKED_SIZE;
    static QString GET_HANDLE;
    static QString CLOSE_PACK_STREAMER;
    static QString IS_PACK_STREAMER_FINISHED;
    static QString ADD_PACKED_DATA;
    static QString ADD_TIFF_IMAGE_DATA;
    static QString ADD_PLACEHOLDER;
    static QString REPLACE_FRAME;
    static QString GET_PACK_STATISTICS;

    static QLibrary fasLib;
};

#endif // FI_INTERNAL_H
