/*!
 * @file fi_journalctl.cpp
 * @brief The implementation of the Fastec JournalCtl object.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QProcess>

#include "fi_journalctl.h"
#include "logger.h"
#include "jc_local.h"
#include "jc_remote.h"

namespace FI_JOURNALCTL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------
} // End of FI_JOURNALCTL

using namespace FI_JOURNALCTL;

FiJournalCtl::FiJournalCtl(QObject *parent) : QObject(parent)
{
    // Create our local object.
    jc_conn = std::unique_ptr<jcLocal>(new jcLocal(this));
    is_remote = false;

    // Connect the local object to our signal.
    jcLocal *local_ptr = qobject_cast<jcLocal *>(jc_conn.get());
    connect(local_ptr, &jcLocal::dataAvailable, this, &FiJournalCtl::processData);
}

FiJournalCtl::FiJournalCtl(QObject *parent,  const FI_JOURNALCTL::SSH_SETTINGS *ssh_info) : QObject(parent)
{
    // Create our remote object.
    jc_conn = std::unique_ptr<jcRemote>(new jcRemote(this));
    is_remote = true;

    // Connect the local object to our signal. Because this is coming from a thread, we want to
    // process the message from the message queue, not directly.
    jcRemote *remote_ptr = qobject_cast<jcRemote *>(jc_conn.get());
    connect(remote_ptr, &jcRemote::dataAvailable, this, &FiJournalCtl::processData, Qt::QueuedConnection);

    // Pass the security information on to the remote object.
    if (!remote_ptr->setSecurity(ssh_info))
    {
        // See if we can find out why things didn't work.
        error_info = remote_ptr->getErrorString();
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "setSecurity() failed:" << error_info;

        // Shutdown the object.
        jc_conn.reset();
    }
}

FiJournalCtl::~FiJournalCtl()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Are we running?
    if (isRunning())
        close();

    // Close the connection.
    if (jc_conn.get() != nullptr)
        jc_conn.reset();
}

void FiJournalCtl::processData(const QString& log_data)
{
    // Pass the data on.
    emit dataAvailable(log_data);
}

bool FiJournalCtl::isRemote()
{
    return is_remote;
}

bool FiJournalCtl::isRunning()
{
    bool is_running = false;

    // Sanity check.
    if (jc_conn.get() != nullptr)
    {
        // Are we remote or local?
        if (is_remote)
        {
            jcRemote *remote_ptr = qobject_cast<jcRemote *>(jc_conn.get());
            is_running = remote_ptr->isRunning();
        }
        else
        {
            jcLocal *local_ptr = qobject_cast<jcLocal *>(jc_conn.get());
            is_running = local_ptr->isRunning();
        }
    }

    // Return our result.
    return is_running;
}

void FiJournalCtl::setIdentifiers(const QStringList& ids)
{
    // Reset our ID string.
    id_list.clear();

    // Loop through the list and build up the command line.
    foreach (const QString& id, ids)
    {
        id_list.append(QStringLiteral("--identifier=%1").arg(id));
    }
}

void FiJournalCtl::close()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (jc_conn.get() == nullptr)
        return;

    // Are we remote or local?
    if (is_remote)
    {
        // Have the remote object close.
        jcRemote *remote_ptr = qobject_cast<jcRemote *>(jc_conn.get());
        remote_ptr->close();
    }
    else
    {
        jcLocal *local_ptr = qobject_cast<jcLocal *>(jc_conn.get());

        // Are we running as root?
        if (isRoot())
        {
            QStringList args;

            // Because we're running as root, we need to kill the command.
            args.append(QStringLiteral("-n"));
            args.append(QStringLiteral("/usr/bin/killall"));
            args.append(QStringLiteral("journalctl"));
            runAndWait(QStringLiteral("/usr/bin/sudo"), args);
        }

        // Now have the local object close.
        local_ptr->close();
    }
}

/*!
 * @brief Return a string describing the current error condition.
 */
const QString FiJournalCtl::errorStr()
{
    QString err_string;

    // Provide a generic error message if we don't have a connection.
    if (jc_conn.get() == nullptr)
    {
        return error_info;
    }

    if (is_remote)
    {
        // Have the remote object close.
        jcRemote *remote_ptr = qobject_cast<jcRemote *>(jc_conn.get());
        err_string = remote_ptr->getErrorString();
    }
    else
    {
        jcLocal *local_ptr = qobject_cast<jcLocal *>(jc_conn.get());
        err_string = local_ptr->getErrorString();
    }

    // Return the error string.
    return err_string;
}

void FiJournalCtl::run()
{
    QString cmd;
    QStringList args;

    // Sanity check.
    if (jc_conn.get() == nullptr)
        return;

    // Initialize our command string.
    if (need_sudo)
    {
        // For sudo, we have the command and arguments.
        cmd = QString("/usr/bin/sudo");
        args.append("-n");
        args.append("/usr/bin/journalctl");
    }
    else
    {
        // Not sudo, we just have the command.
        cmd = QString("/usr/bin/journalctl");
    }

    // We never want to use the pager.
    args.append("--no-pager");

    // Are we following?
    if (follow)
        args.append("-f");

    // Do we have identifiers.
    if (!id_list.isEmpty())
        args.append(id_list);

    // Do we have extra args?
    if (!extra_args.isEmpty())
        args.append(extra_args);

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cmd:" << cmd << "args:" << args;

    // Are we remote or local?
    if (is_remote)
    {
        jcRemote *remote_ptr = qobject_cast<jcRemote *>(jc_conn.get());
        remote_ptr->run(cmd, args);
    }
    else
    {
        jcLocal *local_ptr = qobject_cast<jcLocal *>(jc_conn.get());
        local_ptr->run(cmd, args);
    }
}

QString FiJournalCtl::runAndWait(const QString& cmd, const QStringList& args)
{
    QString resultStr;

    // Sanity check.
    if (jc_conn.get() == nullptr)
        return QString();

    // Are we remote or local?
    if (is_remote)
    {
        jcRemote *remote_ptr = qobject_cast<jcRemote *>(jc_conn.get());
        resultStr = remote_ptr->runAndWait(cmd, args);
    }
    else
    {
        jcLocal *local_ptr = qobject_cast<jcLocal *>(jc_conn.get());
        resultStr = local_ptr->runAndWait(cmd, args);
    }

    // Return our result.
    return resultStr;
}

