/*!
 * @file fi_loadcap.cpp
 * @brief This is the source file for the Fastec Imaging CAP file loading class.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include <QString>
#include <QTimer>
#include <QStyle>

#include "camera_mdi.h"
#include "camera_stream.h"
#include "fi_comm.h"
#include "fi_loadcap.h"
#include "logger.h"
#include "mainwindow.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

namespace FI_LOADCAP
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The states for loading a CAP file.
enum
{
    CAP_LOAD_IDLE,              //!< Waiting to go.
    CAP_LOAD_STARTUP,           //!< Make sure the camera is in the proper state.
    CAP_LOAD_EMPTY_PART,        //!< Make sure the partition is empty.
    CAP_LOAD_SET_FILE,          //!< Tell the camera server the file to load.
    CAP_LOAD_BEGIN,             //!< Tell the camera server to start loading.
    CAP_LOAD_IN_PROGRESS,       //!< Display the loading progress to the user.
    CAP_LOAD_COMPLETE_PART,     //!< Finish the loading process.
    CAP_LOAD_DONE,              //!< Terminal state: SUCCESS.
    CAP_LOAD_FAIL               //!< Terminal state: FAILURE.
};

//! How frequently we check on the status of the load operation.
const int LOAD_INTERVAL     = 500;

//! Keep a constant of when a sanity check on CameraMDI parent pointer fails.
const QString CAP_LOAD_FAILED = QStringLiteral("An error occurred while loading the CAP file.\nCheck for sensor incompatibility between CAP file and connected camera.");

}   // End of namespace FI_LOADCAP.

using namespace FI_LOADCAP;
using namespace FI_PROGRESS;

/*!
 * @brief The fiLoadCaptureDlg class constructor.
 */
fiLoadCaptureDlg::fiLoadCaptureDlg(QWidget *parent, Qt::WindowFlags flags) :
    fiProgressDialog(parent, flags),
    cur_state(CAP_LOAD_IDLE),
    self_cleanup(true)
{

}

/*!
 * @brief The FiLoadCap class destructor.
 */
fiLoadCaptureDlg::~fiLoadCaptureDlg()
{
}

/*!
* @brief Factory to create an task oriented progress dialog.
*
* @param srcXferTypeTitleName   A string to use as part of the Xfer dialog title.
*
* @returns A std::unique_ptr<fiProgressDialog> containing the task oriented progress dialog.
*/
std::unique_ptr<fiLoadCaptureDlg> fiLoadCaptureDlg::CreateLoadCaptureDlg(const QString& titlePrefixString,
                                                                         QWidget *parent)

{
    FI_comm     *comm_ptr;
    FI_PROGRESS::XFER_TYPES srcXferTypeToSet = FI_PROGRESS::XFER_TYPES::XFER_TYPE_LOAD;
    std::unique_ptr<fiLoadCaptureDlg> loadCaptureDlg;

    // Our parent SHOULD be the camera object.
    CameraMDI *camera_ptr = qobject_cast<CameraMDI *>(parent);
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected NULL camera_ptr!";
        goto createLoadCapDialogFailed;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        goto createLoadCapDialogFailed;
    }

    // If make it here sanity checks have passed create the cap file loading dialog.
    loadCaptureDlg = std::unique_ptr<fiLoadCaptureDlg>(new fiLoadCaptureDlg(parent));

    // Install the configuration setup for the xfer dialog.
    if (loadCaptureDlg)
    {
        loadCaptureDlg->setXferTypeTitle(srcXferTypeToSet, titlePrefixString);

        // Initialize the status icon to a message box information icon and hidden.
        loadCaptureDlg->showStatusIcon(false);

        // Ticket #522: Only want the cancel button for this dialog.
        loadCaptureDlg->setControlButtons(QMessageBox::Cancel);

        // Connect up our signals/slots.
        connect(loadCaptureDlg.get(), &fiLoadCaptureDlg::progressCancelled, loadCaptureDlg.get(), &fiLoadCaptureDlg::userCanceledLoad);
        connect(comm_ptr, &FI_comm::read_reg_complete, loadCaptureDlg.get(), &fiLoadCaptureDlg::read_complete, Qt::QueuedConnection);
        connect(comm_ptr, &FI_comm::write_reg_complete, loadCaptureDlg.get(), &fiLoadCaptureDlg::write_complete, Qt::QueuedConnection);
        connect(camera_ptr, &CameraMDI::closing, loadCaptureDlg.get(), &fiLoadCaptureDlg::userCanceledLoad);

        // We are a "load" type.
        loadCaptureDlg->showXferSrc(true);
        loadCaptureDlg->showXferDest(false);
        loadCaptureDlg->showXferType(false);
        loadCaptureDlg->showStatusValue(false);
        loadCaptureDlg->setBarMinMax(0, 100);
        loadCaptureDlg->showBar(true);
    }

createLoadCapDialogFailed:
    return loadCaptureDlg;
}


/*!
 * @brief Starting point for the dialog.
 */
int fiLoadCaptureDlg::exec()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Initialize our state.
    cur_state = CAP_LOAD_STARTUP;

    // Now that we can go, start the state machine.
    QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);

    // Pass it on.
    return fiProgressDialog::exec();
}

/*!
 * @brief The user has cancelled the load operation.
 */
void fiLoadCaptureDlg::cancelDialog()
{
    // Pass it on.
    fiProgressDialog::cancelDialog();

    // Let's clean ourselves up.
    if (self_cleanup)
        deleteLater();
}

void fiLoadCaptureDlg::stateTransition()
{
    uint32_t    reg_value;
    CameraMDI   *camera_ptr;
    FI_comm     *comm_ptr;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cur_state:" << cur_state;

    // Our parent SHOULD be the camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // State machine.
    switch (cur_state)
    {
    case CAP_LOAD_IDLE:
        break;
    case CAP_LOAD_STARTUP:
        // Make sure we go into LIVE mode.
        reg_value = camera_ptr->read_camera_cache(GIGE_REG_CAMERA_STATE);
        if (reg_value == GIGE_CAMERA_STATE_LIVE_VIDEO)
        {
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_STARTUP: In live, transition to CAP_LOAD_EMPTY_PART";

            // Already in live mode, so transition to the next state.
            cur_state = CAP_LOAD_EMPTY_PART;
            QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);
        }
        else
        {
            // Tell the camera to go live. When we get a state transition,
            // we'll circle back around.
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_STARTUP: Need to go live.";
            camera_ptr->do_camera_live();
        }
        break;
    case CAP_LOAD_EMPTY_PART:
        // Make sure the partition is empty.
        reg_value = camera_ptr->read_camera_cache(GIGE_REG_PART_STATE);
        if (reg_value == GIGE_PART_STATE_EMPTY)
        {
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_EMPTY_PART: Part state is empty, transition to CAP_LOAD_SET_FILE";

            // Move to the next state.
            cur_state = CAP_LOAD_SET_FILE;
            QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);
        }
        else
        {
            // Make sure the partition is marked as empty.
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_EMPTY_PART: Need to set the part state to empty.";
            camera_ptr->write_camera_register(GIGE_REG_PART_STATE, GIGE_PART_STATE_EMPTY);
        }
        break;
    case CAP_LOAD_SET_FILE:
        // Set the file name to load.
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_SET_FILE:" << media_path;
        camera_ptr->write_camera_buffer(GIGE_REG_LOAD_FILE_NAME, media_path.toLocal8Bit());
        break;
    case CAP_LOAD_BEGIN:
        // Is the camera streaming?
        if (camera_ptr->get_streamer()->is_paused())
        {
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_BEGIN: Streamer is paused, transition to CAP_LOAD_IN_PROGRESS";

            // Tell the camera to start loading.
            cur_state = CAP_LOAD_IN_PROGRESS;
            camera_ptr->write_camera_register(GIGE_REG_LOAD, GIGE_LOAD_START);
        }
        else
        {
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_BEGIN: Need to pause streamer";

            // We want to know when the video stream is paused.
            connect(camera_ptr, &CameraMDI::video_is_paused, this, &fiLoadCaptureDlg::streamIsPaused);

            // Pause the streamer.
            camera_ptr->pause_video_request();
        }
        break;
    case CAP_LOAD_IN_PROGRESS:
        // Ask the camera for an update.
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_IN_PROGRESS: Ask for an update";
        comm_ptr->start_read_reg(camera_ptr->get_ip_addr(), GIGE_REG_LOAD);
        break;
    case CAP_LOAD_COMPLETE_PART:
        // Do we care?
        reg_value = camera_ptr->read_camera_cache(GIGE_REG_PART_STATE);
        if (reg_value == GIGE_PART_STATE_COMPLETE)
        {
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_COMPLETE_PART: Part state is complete.";

            // Check to see if we are in review. The camera SHOULD do this for us, but
            // just make sure...
            reg_value = camera_ptr->read_camera_cache(GIGE_REG_CAMERA_STATE);
            if (reg_value != GIGE_CAMERA_STATE_REVIEW)
            {
                // Delay just a little bit before commanding the switch to review.
                QTimer::singleShot(250, camera_ptr, SLOT(do_camera_review()));
            }

            // Let everyone know we're done.
            emit loading_finish();

            // Transition to the terminal state.
            cur_state = CAP_LOAD_DONE;
            QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);
        }
        else
        {
            // Read the current camera state.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "CAP_LOAD_COMPLETE_PART: Partition isn't complete? part_state:" << Qt::showbase << Qt::hex << reg_value;
            comm_ptr->start_read_reg(camera_ptr->get_ip_addr(), GIGE_REG_PART_STATE);
        }
        break;
    case CAP_LOAD_DONE:
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "CAP_LOAD_DONE";
        accept();
        break;

    case CAP_LOAD_FAIL:
    default:
        // We'll just hang out here. We want the user to be able to see the error
        // and actively close the dialog.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "CAP_LOAD_FAIL";
        break;
    }
}

/*!
 * @brief Slot invoked when the CameraMDI object indicates that the stream is paused.
 */
void fiLoadCaptureDlg::streamIsPaused()
{
    CameraMDI   *camera_ptr;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cur_state:" << cur_state;

    // Our parent SHOULD be the camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return;
    }

    // Release the "paused" signal.
    disconnect(camera_ptr, &CameraMDI::video_is_paused, this, &fiLoadCaptureDlg::streamIsPaused);

    // Run the state machine.
    QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);
}

/*!
 * @brief The communications object is indicating that a previously requested register read operation
 *        has completed.
 *
 * @param ip_addr   The IP address of the camera the read operation was for.
 * @param reg_addr  The register that was being read.
 */
void fiLoadCaptureDlg::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr;

    // Our parent SHOULD be the camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return;
    }

    // Is this event for us?
    if (ip_addr != camera_ptr->get_ip_addr())
        return;

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_CAMERA_STATE:
        // Are we starting a load operation?
        if (cur_state == CAP_LOAD_STARTUP)
            QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);
        break;
    case GIGE_REG_LOAD:
        if (cur_state == CAP_LOAD_IN_PROGRESS)
            updateLoadProgress();
        break;
    case GIGE_REG_PART_STATE:
        if ((cur_state == CAP_LOAD_EMPTY_PART) || (cur_state == CAP_LOAD_COMPLETE_PART))
            QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);
        break;
    default:
        break;
    }
}

/*!
 * @brief Update the user interface with the status of the load operation.
 * @param load_status       Status as read from the camera.
 */
void fiLoadCaptureDlg::updateLoadProgress()
{
    bool    is_running;
    bool    is_error;
    uint32_t load_status;
    uint32_t status_value;
    CameraMDI   *camera_ptr;

    // Our parent SHOULD be the camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return;
    }

    // Get the current status.
    load_status = camera_ptr->read_camera_cache(GIGE_REG_LOAD);
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "load_status;" << Qt::showbase << Qt::hex << load_status;

    // Parse the register value.
    is_running = (load_status & GIGE_LOAD_BUSY) == (uint32_t)GIGE_LOAD_BUSY;
    is_error = (load_status & GIGE_LOAD_ERROR) == (uint32_t)GIGE_LOAD_ERROR;
    status_value = load_status & GIGE_LOAD_STATUS_MASK;

    // Did we have an error?
    if (!is_error)
    {
        // Update the progress value.
        setBarValue((int)status_value);
    }

    // Are we still running?
    if (!is_running)
    {
        // Not running...is there an error?
        if (is_error)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error loading:" << Qt::showbase << Qt::hex << status_value;
            cur_state = CAP_LOAD_FAIL;
            setStatusValue(CAP_LOAD_FAILED);
            setStatusIcon(QStyle::SP_MessageBoxWarning);
            setCancelText(QStringLiteral("OK"));
            showStatusIcon(true);
            showCancel(true);   // At this point make sure there is an exit pathway.
            setBarMinMax(0, 0);
            emit load_failed();
            // Stop timer returns and terminate at this point.  The user will have to
            // manually exit the dialog.
            return;
        }
        else
        {
            // Loading is complete.
            cur_state = CAP_LOAD_COMPLETE_PART;
        }
    }

    // Go around again.
    QTimer::singleShot(LOAD_INTERVAL, this, &fiLoadCaptureDlg::stateTransition);
}

/*!
 * @brief The user canceled a load operation.
 */
void fiLoadCaptureDlg::userCanceledLoad()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cur_state:" << cur_state;

    // Is the load running?
    if (cur_state == CAP_LOAD_IN_PROGRESS)
    {
        CameraMDI   *camera_ptr;

        // Our parent SHOULD be the camera object.
        camera_ptr = qobject_cast<CameraMDI *>(parent());
        if (camera_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
            return;
        }

        // Tell the camera to stop loading.
        camera_ptr->write_camera_register(GIGE_REG_LOAD, GIGE_LOAD_STOP);
    }
}

/*!
 * @brief The communications object is indicating that a previously requested register write operation
 *        has completed.
 *
 * @param ip_addr   The IP address of the camera the write operation was for.
 * @param reg_addr  The register that was being written.
 */
void fiLoadCaptureDlg::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Our parent SHOULD be the camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return;
    }

    // Is this event for us?
    if (ip_addr != camera_ptr->get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_CHANGE_CAMERA_STATE:
        // Are we starting a load operation?
        if (cur_state == CAP_LOAD_STARTUP)
        {
            // Read the current camera state.
            comm_ptr->start_read_reg(ip_addr, GIGE_REG_CAMERA_STATE);
        }
        break;
    case GIGE_REG_LOAD:
        // Read the status of the write operation.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Was this a success?
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Start the timer to check on progress.
            QTimer::singleShot(LOAD_INTERVAL, this, &fiLoadCaptureDlg::stateTransition);
        }
        break;
    case GIGE_REG_LOAD_FILE_NAME:
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "GIGE_REG_LOAD_FILE_NAME, cur_state:" << cur_state;

        // Make sure we're in the expected state.
        if (cur_state == CAP_LOAD_SET_FILE)
        {
            // Read the status of the write operation.
            write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
            if (write_status == GEV_STATUS_SUCCESS)
            {
                // Move to the next state.
                cur_state = CAP_LOAD_BEGIN;
                QTimer::singleShot(0, this, &fiLoadCaptureDlg::stateTransition);
            }
        }
        break;
    case GIGE_REG_PART_STATE:
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "GIGE_REG_PART_STATE, cur_state:" << cur_state;
        if (cur_state == CAP_LOAD_EMPTY_PART)
        {
            // Read the current camera state.
            comm_ptr->start_read_reg(ip_addr, GIGE_REG_PART_STATE);
        }
        break;
    default:
        break;
    }
}
