/*!
 * @file fi_mdiarea.cpp
 * @brief This provides the implementation for the fiMdiArea class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_mdiarea.h"

#include <QDragEnterEvent>
#include <QDropEvent>
#include <QHostAddress>
#include <QMdiSubWindow>
#include <QMimeData>

#include <qendian.h>

#include "camera_dock.h"
#include "logger.h"
#include "mainwindow.h"

/*!
 * @brief Our fiMdiArea constructor.
 */
fiMdiArea::fiMdiArea( QWidget *parent ) :
    QMdiArea( parent )
{
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Catch the drag enter event so we can allow a camera object to be dropped
 *        into our QMdiArea widget.
 *
 * @param event     The dragEnter event pointer.
 */
void fiMdiArea::dragEnterEvent( QDragEnterEvent *event )
{
    // Check for a Qt format.
    if ( event->mimeData()->hasFormat( "application/x-qabstractitemmodeldatalist" ))
    {
        // We might like this.
        event->acceptProposedAction();
    }
    else
    {
        // Pass it on.
        QWidget::dragEnterEvent( event );
    }
}

/*!
 * @brief Catch the drop event so we can allow a camera object to be dropped
 *        into our list widget.
 *
 * @param event     The drop event pointer.
 */
void fiMdiArea::dropEvent( QDropEvent *event )
{
    // Check for a Qt format.
    if ( event->mimeData()->hasFormat( "application/x-qabstractitemmodeldatalist" ))
    {
        QByteArray  encoded = event->mimeData()->data( "application/x-qabstractitemmodeldatalist" );
        QDataStream stream( &encoded, QIODevice::ReadOnly );

        // Parse the data. We're looking for an IP address entry.
        while ( !stream.atEnd() )
        {
            int row, col;
            QMap<int, QVariant> roleDataMap;

            // Pull the data into it's constituent parts.
            stream >> row >> col >> roleDataMap;

            // Is this the IP address?
            if ( col == CAMERA_DOCK::CD_IP_ADDR )
            {
                QHostAddress    cam_ip_addr;
                QMdiSubWindow   *mdi_ptr;
                uint32_t        ip_addr;

                // Use the IP address to get the associated camera.
                cam_ip_addr.setAddress( roleDataMap.value(0).toString() );

                // Get the IP address in a straight IPv4 flavor.
                ip_addr = qFromBigEndian( cam_ip_addr.toIPv4Address() );

                // Signal that we have an IP address of interest.
                emit camera_dropped( ip_addr );

                // See if we have an open window.
                mdi_ptr = MainWindow::GetInstance()->find_mdi_child( ip_addr );
                if ( mdi_ptr == NULL )
                {
                    // Ask the dock to open the camera.
                    MainWindow::GetInstance()->get_camera_dock()->open_selected_camera();
                }

                // We're done with this drop event.
                event->acceptProposedAction();
                break;
            }
        }
    }
}
