/*!
 * @file fi_mem_block.h
 *
 * @brief An object to manage a single array of memory in our memory pool.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QMutexLocker>
#include <QString>

#include "logger.h"
#include "fi_mem_block.h"

namespace FI_MEMPOOL
{

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

typedef enum MEMORY_MAP_STATE
{
    NO_MEMORY = 0,
    MEMORY_AVAILABLE,
    MEMORY_IN_USE,
    MEMORY_RECYCLED
} MEMORY_MAP_STATE;

//! Determines whether or not the id settings will write to system log.
const bool SHOW_ID_LOGS = false;

const QString   memBlockIdStr = QStringLiteral("[MEMBLOCK]");

//! Because the memory map objects which are stored in a std::vector could number
//! in the hundreds for a typical jpeg video image stream the number of map entries
//! created for each resize will be larger so that it doesn't create unnecessary
//! allocations.  Especially when image frame sizes may be very small.
static const uint32_t MEMORY_MAP_LIST_ALLOCSIZE = 500;

//! Define a pad-alignment for the entire memory block in memory.
static const uint32_t MEMORY_BLOCK_ALIGN = 64;

//! Define a pad-alignment for every resource chunk reserved in this memory
//! block.  We are using 32-bytes as a bare minimum alignment amount.
static const uint32_t MEMORY_RESOURCE_ALIGN = 32;

//--------------------------------------------------------------------
// Internal Class Declarations
//--------------------------------------------------------------------

class FiMemMapRecord
{
public:
    FiMemMapRecord();
    virtual ~FiMemMapRecord();

    int chunkId;                        //!< Where this entry exists in the memory map list.
    MEMORY_MAP_STATE chunkState;        //!< What is the state of this entry.
    bool pendingWrite;                  //!< Is there a pending write operation on this chunk.
    uint32_t chunkSize;                 //!< How many bytes are reserved for this chunk.
    uint32_t chunkOffset;               //!< Zero-based offset in the memory block for this map entry.
    uint32_t bytesWritten;              //!< Number of bytes written in this chunk.
};

} // End of the FI_MEMPOOL namespace declaration.

// Make things easier for usage.
using namespace std;
using namespace FI_MEMPOOL;

//--------------------------------------------------------------------
// Implementation of FiMemMapRecord
//--------------------------------------------------------------------
uint32_t FiMemoryBlock::nextId = 0; //!< The first id will be zero to mimic an index#.

FiMemMapRecord::FiMemMapRecord() :
    chunkId(NO_MEMORY_MAP),
    chunkState(MEMORY_MAP_STATE::NO_MEMORY),
    pendingWrite(false),
    chunkSize(0),
    chunkOffset(0),
    bytesWritten(0)
{
}
FiMemMapRecord::~FiMemMapRecord()
{
    if (pendingWrite)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "FiMemMapRecord had a pending write outstanding!";
    }
}

//--------------------------------------------------------------------
// Implementation of FiMemoryBlock
//--------------------------------------------------------------------

/*!
 * @brief Constructor for the FiMemoryBlock object.
 */
FiMemoryBlock::FiMemoryBlock() :
    memoryBlockId(NO_MEMORY_BLOCK),
    allowReservations(false),
    warnIfDestructed(false),
    mapRecordsReserved(0),
    mapRecordsInUse(0),
    mapRecordsRecycled(0),
    nextFreeOffset(0),
    memoryBlockCapacity(0)
{
    myId = nextId;
    nextId++;
    if (nextId == 100000)
        nextId = 0;
}

//--------------------------------------------------------------------
// Implementation of FiMemoryBlock
//--------------------------------------------------------------------

/*!
 * @brief Destructor for FiMemoryBlock object.
 */
FiMemoryBlock::~FiMemoryBlock()
{
    QString myLabelStr = QString("%1[~FiMemoryBlock]").arg(memBlockIdStr);
    if (warnIfDestructed)
    {
        // This should never happen with the use of the custom deleter
        // in the memory pool.
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Destructing an FiMemoryBlock object before its time!";
    }

    // Log this information.  Verbosity is handled in the function dependant on the compilation.
    FiMemoryPool::LogVerbose((QString("%1 [memoryBlockId=%2][#Reserved=%3][#InUse=%4][#Recycled=%5][nextOffset=%6][allowReservations=%7]").
                              arg(myLabelStr).arg(memoryBlockId).
                              arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset).
                              arg(allowReservations ? "YES" : "NO")).toUtf8().constData());
}

/*!
 * @brief Factory to create an image data buffer.
 *
 * @param myPoolId - What is the pool id for this block.
 * @param dataSize - How many bytes to create for image data buffer.
 *
 * @returns A newly minted image data object or a nullptr if there is an error.
 */
std::unique_ptr<FI_MEMPOOL::FiMemoryBlock> FiMemoryBlock::CreateMemoryBlock(int myPoolId, uint32_t dataSize)
{
    std::unique_ptr<FI_MEMPOOL::FiMemoryBlock> memoryBlockPtr;
    uint32_t dataSizeAligned = ((dataSize + (MEMORY_BLOCK_ALIGN - 1)) / MEMORY_BLOCK_ALIGN) * MEMORY_BLOCK_ALIGN;  //!< Align on boundary.
    try
    {
        memoryBlockPtr = std::unique_ptr<FI_MEMPOOL::FiMemoryBlock>(new FiMemoryBlock());

        // Report any error in the creation of the master map record.
        if (memoryBlockPtr == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to create FiMemoryBlock object!";
            goto exitCreateMemoryBlock;
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Exception:" << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "(...)Exception!";
    }

    // The heart of the memory block is the actual memory that makes up the block and
    // the memory map that manages reservations of chunk resources over that block.
    try
    {
        // Allocate the internal memory block being managed here.
        memoryBlockPtr->memoryBlock = std::unique_ptr<FI_MEMPOOL::MEMORY_BLOCK>(new FI_MEMPOOL::MEMORY_BLOCK(dataSizeAligned));

        // Report any error in the creation of the memory block.
        if (memoryBlockPtr->memoryBlock == nullptr)
        {
            memoryBlockPtr.reset();
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to create MEMORY_BLOCK!";
            goto exitCreateMemoryBlock;
        }

        // Set the capacity to the requested size (not the aligned size).
        memoryBlockPtr->memoryBlockCapacity = dataSize;

        // Create the memory map list making sure to allocate a large chunk to facilitate
        // faster requests for reserving memory chunks in the memory block.
        memoryBlockPtr->memMapRecords =
            std::unique_ptr<std::vector<FiMemMapRecord>>(new std::vector<FiMemMapRecord>(MEMORY_MAP_LIST_ALLOCSIZE));

        // Report any error in the creation of the memory map records.
        if (memoryBlockPtr->memMapRecords == nullptr)
        {
            // Release any allocations.
            memoryBlockPtr->memoryBlock.reset();
            memoryBlockPtr.reset();
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to create memory map record list!";
            goto exitCreateMemoryBlock;
        }

        memoryBlockPtr->allowReservations = true;
        // AT this point the memory is ready for reservation.  Reservation only moves
        // forward.  Once a chunk is reserved and recycled it is not until ALL reservations
        // are recycled that this memory block will itself be recycled.  There is not
        // reuse on a recycled memory chunk so that is an important design decision.

        // Go ahead and set up the warn if destructed warning message.
        memoryBlockPtr->warnIfDestructed = true;
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Exception:" << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "(...)Exception!";
    }

    // Go ahead and set the pool id for my object.
    memoryBlockPtr->memoryBlockId = myPoolId;

exitCreateMemoryBlock:

    return memoryBlockPtr;
}

/*!
 * @brief Reserves a memory chunk for a resource request.
 *
 * @param resourceSizeRequested: The number of bytes requested for resource.
 *
 * @returns The memory usage id or NO_MEMORY_MAP if unable to reserve the space.
 */
int FiMemoryBlock::GetMemResource(uint32_t resourceSizeRequested)
{
    int reservedId = NO_MEMORY_MAP;
    QString myLabelStr = QString("%1[GetMemResource]").arg(memBlockIdStr);

    // Lock the mutex to protect the master map record data from collisions.
    QMutexLocker memMapMasterRecordMutexLocker(&memMapMasterRecordMutex);

    // Sanity check nextFreeOffset make sure it is within the bounds of
    // the memory block creation size.
    if (nextFreeOffset > memoryBlockCapacity)
    {
        // Log this information.  Verbosity is handled in the function dependant on the compilation.
        FiMemoryPool::LogVerbose((QString("%1[INTERNAL ERROR][memoryBlockId=%2][nextOffset=%3 > memoryBlockCapacity=%4][#Reserved=%5][#InUse=%6][#Recycled=%7]").
                                  arg(myLabelStr).arg(memoryBlockId).arg(nextFreeOffset).arg(memoryBlockCapacity).
                                  arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled)).toUtf8().constData(), VLOG_SHOW_ERROR);

        // This should not happen but we can't have an invalid spaceLeft condition.
        nextFreeOffset = memoryBlockCapacity;
    }

    uint32_t spaceLeft                  = memoryBlockCapacity - nextFreeOffset;
    uint32_t mapSize                    = (uint32_t) memMapRecords->size();
    uint32_t mapSizeNeeded              = mapRecordsReserved + 1;
    uint32_t memMapReSizeAmount         = (uint32_t) memMapRecords->size() + MEMORY_MAP_LIST_ALLOCSIZE;
    uint32_t nextReservedId             = mapRecordsReserved + 1;
    uint32_t resourceSizeAligned       = ((resourceSizeRequested + (MEMORY_RESOURCE_ALIGN - 1)) / MEMORY_RESOURCE_ALIGN) * MEMORY_RESOURCE_ALIGN;

    // Note:  If the resourceSizeRequested was zero it's ok because the size requested will be
    // set to the alignment size - but should this be an error condition?

    // Does this block allow for new reservations?
    if (!allowReservations)
    {
        FiMemoryPool::LogVerbose((QString("%1 NO RESERVATIONS ALLOWED. [memoryBlockId=%2]map[%3 of %4][#Reserved=%5][#InUse=%6][#Recycled=%7][nextOffset=%8]").
                                  arg(myLabelStr).arg(memoryBlockId).arg(mapSizeNeeded).arg(mapSize).
                                  arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset)).toUtf8().constData(), VLOG_SHOW_ERROR);

        goto exitGetMemResource;
    }

    // Does this map have the space for this reservation?
    if (resourceSizeAligned <= spaceLeft)
    {
        // Is there an empty record in memMapRecords for a new reservation?
        if (mapSizeNeeded > mapSize)
        {
            // Need to allocate more reservation map entries.
            // This requires a resize of the memMapRecords list.
            try
            {
                memMapRecords->resize(memMapReSizeAmount);
                mapSize = (uint32_t) memMapRecords->size();  // Reset our mapsize.
            }
            catch (std::exception& e)
            {
                QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
            }
            catch (...)
            {
                QLOG_DEBUG() << myLabelStr << "(...)Exception!";
            }
        }

        // Check the space again.
        if (mapSizeNeeded <= mapSize)
        {
            try
            {
                // Set the data members to the in-use state.
                memMapRecords->at(mapRecordsReserved).chunkId = (int) mapRecordsReserved;
                memMapRecords->at(mapRecordsReserved).chunkState = MEMORY_MAP_STATE::MEMORY_IN_USE;
                memMapRecords->at(mapRecordsReserved).pendingWrite = false;
                memMapRecords->at(mapRecordsReserved).chunkSize = resourceSizeAligned;
                memMapRecords->at(mapRecordsReserved).chunkOffset = nextFreeOffset;
                memMapRecords->at(mapRecordsReserved).bytesWritten = 0;

                // Now that the throw probability is low set the return code.
                reservedId = (int) mapRecordsReserved;

                // Modify the master map information for the new reservation.
                mapRecordsReserved++;
                mapRecordsInUse++;
                nextFreeOffset += resourceSizeAligned;
            }
            catch (std::exception& e)
            {
                QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
            }
            catch (...)
            {
                QLOG_DEBUG() << myLabelStr << "(...)Exception!";
            }
        }
    }

exitGetMemResource:

    FiMemoryPool::LogVerbose((QString("%1 myPoolId[%2] ReservedId[%3] nextId[%4] spaceLeft[%5] chunkSize[%6] map[%7 of %8][#Reserved=%9][#InUse=%10][#Recycled=%11][NextOffset=%12]").
                              arg(myLabelStr).arg(memoryBlockId).arg(reservedId).arg(nextReservedId).
                              arg(spaceLeft).arg(resourceSizeAligned).arg(mapSizeNeeded).arg(mapSize).
                              arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset)).toUtf8().constData(), VLOG_MAP_ALLOCS);
    return reservedId;
}

/*!
 * @brief Log memory pool information for diagnostic purposes.
 *
 * @param moreInfoStr - Provides additional information to log for diagnostic purposes.  Can be a nullptr.
 */
void FiMemoryBlock::LogMyId(const char *moreInfoStr)
{
    if (SHOW_ID_LOGS)
    {
        const char *useNothing = "";
        if (moreInfoStr == nullptr)
        {
            moreInfoStr = useNothing;
        }

        QString myLabelStr = QString("%1[LogMyId]").arg(memBlockIdStr);

        // Destruct the memory map information.
        FiMemoryPool::LogVerbose((QString("%1 [memoryBlockId=%2][#Reserved=%3][#InUse=%4][#Recycled=%5][nextOffset=%6][allowReservations=%7][moreInfo=%8]").
                                  arg(myLabelStr).arg(memoryBlockId).
                                  arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset).
                                  arg(allowReservations ? "YES" : "NO", moreInfoStr)).toUtf8().constData());
    }
}

/*!
 * @brief Gives this object an opportunity to prepare for destructing.
 *
 * At the moment this is for diagnostic purposes.  Because the data
 * belonging to this object is managed by a unique_ptr it
 * is susceptable to a resource-leak in the system if it
 * is not properly transported.  This warning mechanism is strictly
 * that, if the object is destructed and the warning flag is true
 * (which it is after a factory creation) then the warning will
 * appear in the log file.
 */
void FiMemoryBlock::PrepareToDestruct()
{
    // The memory block is being destructed by a known party so will
    // turn off the warning.
    warnIfDestructed = false;
}

/*!
 * @brief Append bytes to the image data.
 *
 * @param memoryMapId - Memory map ID for this piece of the memory block.
 * @param appendSize - Number of bytes to add to the memory buffer.
 * @param bytesToAppend - the bytes to append into memory buffer.
 *
 * @returns The number of bytes appended to the memory buffer.  If it
 * returns 0 there is an error condition.
 */
uint32_t FiMemoryBlock::AppendBytes(int memoryMapId, uint32_t appendSize, const uint8_t *bytesToAppend)
{
    QString myLabelStr = QString("%1[AppendBytes]").arg(memBlockIdStr);
    uint32_t bytesAppended              = 0;
    uint32_t bytesWrittenAfterAppend    = 0;
    uint32_t offsetInChunk              = 0;
    uint32_t memMapIndex                = (uint32_t) memoryMapId;

    // Note:  The memory entry for memoryMapId is unique to itself and the architecture
    // is such that there should NOT be any collision with this data from the multiple
    // threads in the system.

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitAppendBytes;
    }

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // This reserved memory must be in an in-use state.
            if (memMapRecords->at(memMapIndex).chunkState != MEMORY_MAP_STATE::MEMORY_IN_USE)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "chunkState != MEMORY_IN_USE!";
                goto exitAppendBytes;
            }

            // Is there a pending write?  If so this is in the middle of a different append operation.
            if (memMapRecords->at(memMapIndex).pendingWrite)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "pendingWrite!";
                goto exitAppendBytes;
            }

            // Is there space for this append?
            bytesWrittenAfterAppend = memMapRecords->at(memMapIndex).bytesWritten + appendSize;

            // Was there enough room?
            if (bytesWrittenAfterAppend > memMapRecords->at(memMapIndex).chunkSize)
            {
                // There is not enough room here.
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "No Room!";
                goto exitAppendBytes;
            }

            // Looks like it is good to go!  Set the offset into our chunk plus our byteswritten already.
            offsetInChunk = memMapRecords->at(memMapIndex).chunkOffset + memMapRecords->at(memMapIndex).bytesWritten;

            // Append the bytes by the appendSize.
            {
                // Lock the mutex to protect the master map record data from collisions.
                QMutexLocker memMapMasterRecordMutexLocker(&memMapMasterRecordMutex);
                uint8_t *current_position = memoryBlock->data();

                // memoryBlock is a unique_ptr to a std::vector of uint8_t s.
                memcpy(&current_position[offsetInChunk], bytesToAppend, appendSize);
                memMapRecords->at(memMapIndex).bytesWritten += appendSize;
                bytesAppended = appendSize;
            }
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
        goto exitAppendBytes;
    }

exitAppendBytes:

#if 0
    FiMemoryPool::LogVerbose((QString("%1 [memoryBlockId=%2][#Reserved=%3][#InUse=%4][#Recycled=%5][nextOffset=%6][bytesAppended=%7]").
                              arg(myLabelStr).arg(memoryBlockId).
                              arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset).
                              arg(bytesAppended)).toUtf8().constData());

#endif

    return bytesAppended;
}

/*!
 * @brief Append bytes to the image data externally.
 * This will allow access to the append point into our data for external
 * manipulation.  This is necessary in order to facilitate socket reads
 * or other such function calls so that an intermediate buffer is not necessary.
 * This is in the interest of time to copy bytes from network transfers.
 *
 * @param memoryMapId - Memory map ID for this piece of the memory block.
 * @param sizeRequested - Number of bytes to add to the data buffer.
 * @param bytesAvailable - Set before returning to bytes available in buffer.
 *
 * @returns The append point into the image data exposing access to the internal image
 * data - caller MUST be wary of exceeding the image data buffer in memory.  Returns a
 * nullptr if there is not enough space in image data buffer.
 */
uint8_t *FiMemoryBlock::AppendBytesExternally(int memoryMapId, uint32_t sizeRequested, uint32_t& bytesAvailable)
{
    // Initialize to zero.
    bytesAvailable = 0;

    QString myLabelStr = QString("%1[AppendBytesExternally]").arg(memBlockIdStr);

    uint8_t *externalAccessPtr          = nullptr;
    uint32_t bytesWrittenAfterAppend    = 0;
    uint32_t offsetInChunk              = 0;
    uint32_t memMapIndex                = (uint32_t) memoryMapId;

    bytesAvailable = 0; // Initialize this to a zero in case any errors occur.

    // Note:  The memory entry for memoryMapId is unique to itself and the architecture
    // is such that there should NOT be any collision with this data from the multiple
    // threads in the system.

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitAppendBytesExternally;
    }

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // This reserved memory must be in an in-use state.
            if (memMapRecords->at(memMapIndex).chunkState != MEMORY_MAP_STATE::MEMORY_IN_USE)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "chunkState != MEMORY_IN_USE!";
                goto exitAppendBytesExternally;
            }

            // Is there a pending write?  If so this is in the middle of a different append operation.
            if (memMapRecords->at(memMapIndex).pendingWrite)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "pendingWrite!";
                goto exitAppendBytesExternally;
            }

            // Is there space for this append?
            bytesWrittenAfterAppend = memMapRecords->at(memMapIndex).bytesWritten + sizeRequested;

            // Is there enough room?
            if (bytesWrittenAfterAppend > memMapRecords->at(memMapIndex).chunkSize)
            {
                // There is not enough room here.
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "No Room!";
                goto exitAppendBytesExternally;
            }

            // Looks like it is good to go!  Set the offset into our chunk plus our byteswritten already.
            offsetInChunk = memMapRecords->at(memMapIndex).chunkOffset + memMapRecords->at(memMapIndex).bytesWritten;

            // Allow external access to our pointer.
            {
                // Lock the mutex to protect the master map record data from collisions.
                QMutexLocker memMapMasterRecordMutexLocker(&memMapMasterRecordMutex);
                externalAccessPtr = memoryBlock->data() + offsetInChunk;
                memMapRecords->at(memMapIndex).pendingWrite = true;
            }

            bytesAvailable = sizeRequested;
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
        goto exitAppendBytesExternally;
    }

exitAppendBytesExternally:

#if 0
    FiMemoryPool::LogVerbose((QString("%1 [memoryBlockId=%2][#Reserved=%3][#InUse=%4][#Recycled=%5][nextOffset=%6][externalAccessPtr=%7]").
                              arg(myLabelStr).arg(memoryBlockId).
                              arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset).
                              arg((externalAccessPtr ? "VALID" : "NULL"))).toUtf8().constData());
#endif

    return externalAccessPtr;
}

/*!
 * @brief Moves internal data used forward to facilitate an external append
 * operation.  This is a companion function to the AppendBytesExternally function
 * and must be called in order to register the successful external write operation.
 *
 * @param memoryMapId - Memory map ID for this piece of the memory block.
 * @param appendByteSize - How many bytes were actually written in external append operation.
 *
 * @returns The result of the internal modification.  It can fail if internal state or
 * data modification does not pass sanity checks.
 */
bool FiMemoryBlock::CompleteExternalAppend(int memoryMapId, uint32_t appendByteSize)
{
    QString myLabelStr = QString("%1[CompleteExternalAppend]").arg(memBlockIdStr);

    bool externalAppendRegistered   = false;
    uint32_t memMapIndex            = (uint32_t) memoryMapId;
    uint32_t bytesAvailable         = 0;

    // Note:  The memory entry for memoryMapId is unique to itself and the architecture
    // is such that there should NOT be any collision with this data from the multiple
    // threads in the system.

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitCompleteExternalAppend;
    }

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // This reserved memory must be in an in-use state.
            if (memMapRecords->at(memMapIndex).chunkState != MEMORY_MAP_STATE::MEMORY_IN_USE)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "chunkState != MEMORY_IN_USE!";
                goto exitCompleteExternalAppend;
            }

            // This chunk MUST be in a pendingWrite state otherwise this is an error.
            if (!memMapRecords->at(memMapIndex).pendingWrite)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "was expecting pendingWrite to be true but it's false!";
                goto exitCompleteExternalAppend;
            }

            // Allow external access to our pointer.
            {
                // REVISIT_JLC - I don't think this mutex is needed.
                // Lock the mutex to protect the master map record data from collisions.
                QMutexLocker memMapMasterRecordMutexLocker(&memMapMasterRecordMutex);

                // Reset off the pending write flag.
                memMapRecords->at(memMapIndex).pendingWrite = false;

                // Calculate how many bytes are actually available.
                bytesAvailable = memMapRecords->at(memMapIndex).chunkSize - memMapRecords->at(memMapIndex).bytesWritten;

                // If there was a possible overwrite warn the system.
                if (appendByteSize > bytesAvailable)
                {
                    QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "Possible overwrite! bytesAvailable:" << bytesAvailable << "appendSize: " << appendByteSize;
                    appendByteSize = bytesAvailable;        // Don't want to overwrite our bytesWritten.
                }

                memMapRecords->at(memMapIndex).bytesWritten += appendByteSize;
                externalAppendRegistered = true;
            }
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
    }

exitCompleteExternalAppend:
#if 0
    FiMemoryPool::LogVerbose((QString("%1 [memoryBlockId=%2][#Reserved=%3][#InUse=%4][#Recycled=%5][nextOffset=%6][externalAppendRegistered=%7]").
                              arg(myLabelStr).arg(memoryBlockId).
                              arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset).
                              arg((externalAppendRegistered ? "T" : "F"))).toUtf8().constData());
#endif
    return externalAppendRegistered;
}

/*!
 * @brief Provides read-only access to the memory managed by a memory map id in this
 * block.
 *
 * @param memoryMapId - Memory map ID for this piece of the memory block.
 * @param dataLength - How many bytes are available in this chunk.
 *
 * @returns The read-only pointer or a nullptr if error.
 */
const uint8_t *FiMemoryBlock::GetReservedPtr(int memoryMapId, uint32_t& dataLength)
{
    QString myLabelStr = QString("%1[GetReservedPtr]").arg(memBlockIdStr);

    dataLength = 0;
    const uint8_t *memDataPtr = nullptr;
    uint32_t memMapIndex = (uint32_t) memoryMapId;

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitGetReservedPtr;
    }

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // This reserved memory must be in an in-use state.
            if (memMapRecords->at(memMapIndex).chunkState != MEMORY_MAP_STATE::MEMORY_IN_USE)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "chunkState != MEMORY_IN_USE!";
                goto exitGetReservedPtr;
            }

            // Is there a pending write?  If so this is in the middle of an append operation.
            if (memMapRecords->at(memMapIndex).pendingWrite)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "pendingWrite!";
                goto exitGetReservedPtr;
            }

            // Get the size of the memory buffer.
            dataLength = memMapRecords->at(memMapIndex).bytesWritten;

            // Get the read only memory pointer.
            memDataPtr = (const uint8_t *) memoryBlock->data() + memMapRecords->at(memMapIndex).chunkOffset;
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
        goto exitGetReservedPtr;
    }

exitGetReservedPtr:
    return memDataPtr;
}

/*!
 * @brief Provide access to the assigned memory pool id given to this block of memory.
 *
 * @returns The memory pool id given when memory block was created.
 */
int FiMemoryBlock::GetPoolId()
{
    return memoryBlockId;
}

/*!
 * @brief Provides the size of this resource "chunk" in the memory block.
 * This is the full capacity of the chunk not the length of data in-use in
 * the chunk.
 *
 * @param memoryMapId - Memory map ID for reserved chunk in the memory block.
 *
 * @returns The size in bytes of the capacity for this resource, or zero if error.
 */
uint32_t FiMemoryBlock::GetCapacity(int memoryMapId)
{
    QString myLabelStr = QString("%1[GetCapacity]").arg(memBlockIdStr);

    uint32_t dataCapacity   = 0;
    uint32_t memMapIndex    = (uint32_t) memoryMapId;

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitGetLength;
    }

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // Returning the size of this resource.
            dataCapacity = memMapRecords->at(memMapIndex).chunkSize;
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
        goto exitGetLength;
    }

exitGetLength:
    return dataCapacity;
}

/*!
 * @brief Provides the length of the data in use in reserved chunk belonging to a memoryMapId.
 *
 * @param memoryMapId - Memory map ID for reserved chunk in the memory block.
 *
 * @returns The size in bytes of the data in use in reserved chunk or zero if there is an error.
 */
uint32_t FiMemoryBlock::GetLength(int memoryMapId)
{
    QString myLabelStr = QString("%1[GetLength]").arg(memBlockIdStr);

    uint32_t dataLength     = 0;
    uint32_t memMapIndex    = (uint32_t) memoryMapId;

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitGetLength;
    }

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // This reserved memory must be in an in-use state.
            if (memMapRecords->at(memMapIndex).chunkState != MEMORY_MAP_STATE::MEMORY_IN_USE)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "chunkState != MEMORY_IN_USE!";
                goto exitGetLength;
            }

            // Is there a pending write?  If so this is in the middle of an append operation.
            if (memMapRecords->at(memMapIndex).pendingWrite)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "pendingWrite!";
                goto exitGetLength;
            }

            // Returning the number of bytes actually written to the reserved
            // chunk of memory not the chunkSize but how data is in it for use.
            dataLength = memMapRecords->at(memMapIndex).bytesWritten;
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
        goto exitGetLength;
    }

exitGetLength:
    return dataLength;
}

/*!
 * @brief Replaces bytes in the resource with new image bytes.
 *
 * @param memBytesToSet - A read only pointer to the new image bytes.
 * @param memBytesToSetLen - The number of image bytes to set.
 *
 * @returns The number of bytes that are now used in the image data.  If zero
 * the replacment failed.
 */
uint32_t FiMemoryBlock::ReplaceBytes(int memoryMapId, uint32_t memBytesToSetLen, const uint8_t *memBytesToSet)
{
    uint32_t bytesReplaced  = 0;
    QString myLabelStr = QString("%1[ReplaceBytes]").arg(memBlockIdStr);

    uint32_t memMapIndex    = (uint32_t) memoryMapId;

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitReplaceBytes;
    }

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // This reserved memory must be in an in-use state.
            if (memMapRecords->at(memMapIndex).chunkState != MEMORY_MAP_STATE::MEMORY_IN_USE)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "chunkState != MEMORY_IN_USE!";
                goto exitReplaceBytes;
            }

            // Is there a pending write?  If so this is in the middle of an append operation.
            if (memMapRecords->at(memMapIndex).pendingWrite)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "pendingWrite!";
                goto exitReplaceBytes;
            }

            // Simply reset the bytesWritten and append from the beginning this data.
            {
                QMutexLocker memMapMasterRecordMutexLocker(&memMapMasterRecordMutex);
                memMapRecords->at(memMapIndex).bytesWritten = 0;
            }
            bytesReplaced = AppendBytes(memoryMapId, memBytesToSetLen, memBytesToSet);
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
        goto exitReplaceBytes;
    }

exitReplaceBytes:
    return bytesReplaced;
}

/*!
 * @brief Releases a memory chunk as it is no longer needed by a resource.
 *
 * @param memoryMapId - Memory map ID for this piece of the memory block.
 * @param recycleMemoryBlock - Is this memory block recyclable (meaning all outstanding
 * resource chunks have now been recycled.)
 *
 * @returns The result of the release request.  True if successful or false if not.  The
 * setting inUseCount should be considered valid unless a True is returned.
 */
bool FiMemoryBlock::ReleaseMemoryResource(int memoryMapId, bool& recycleMemoryBlock)
{
    QString myLabelStr = QString("%1[ReleaseMemoryResource]").arg(memBlockIdStr);
    bool releasedResource = false;
    uint32_t memMapIndex    = (uint32_t) memoryMapId;
    uint32_t capacity       = 0;
    uint32_t length         = 0;

    // This will get set true if the memory block is no longer making
    // reservations and the in use count is 0 after this resource is processed.
    recycleMemoryBlock = false;

    // Lock the mutex to protect the master map record data from collisions.
    QMutexLocker memMapMasterRecordMutexLocker(&memMapMasterRecordMutex);

    // Note:  The memory entry for memoryMapId is unique to itself and the architecture
    // is such that there should NOT be any collision with this data from the multiple
    // threads in the system.

    // If the memory map id is not valid can't continue.
    if (memoryMapId == NO_MEMORY_MAP)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "NO_MEMORY_MAP!";
        goto exitReleaseReservedChunk;
    }

    // Must lock the access to our internal data because it is imperative that our
    // state be protected against external thread requests.

    // Is the memory map id a reserved one?
    if (memMapIndex < mapRecordsReserved)
    {
        try
        {
            // This reserved memory must be in an in-use state.
            if (memMapRecords->at(memMapIndex).chunkState != MEMORY_MAP_STATE::MEMORY_IN_USE)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "chunkState != MEMORY_IN_USE!";
                goto exitReleaseReservedChunk;
            }

            // Is there a pending write?  If so this is in the middle of a different append operation.
            if (memMapRecords->at(memMapIndex).pendingWrite)
            {
                QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId:" << memoryMapId << "was NOT pendingWrite to be true!";
                //goto exitReleaseReservedChunk;

                // REVISIT_JLC - don't let this condition be an error?  Although
                // if there is a pendingWrite it means the completion call was
                // never made on an exposed external pointer.  It should Not
                // ever happen but the resource needs to be marked recycled?
                // Or should we leave it to be an orphan?
            }

            // Allow external access to our pointer.
            {
                // Reset off the pending write flag.
                memMapRecords->at(memMapIndex).chunkState = MEMORY_MAP_STATE::MEMORY_RECYCLED;

                capacity    = memMapRecords->at(memMapIndex).chunkSize;
                length      = memMapRecords->at(memMapIndex).bytesWritten;

                // This is why we need a mutex because we are now modifying the
                // state of this master map.
                mapRecordsInUse--;
                mapRecordsRecycled++;

                if ((mapRecordsInUse == 0) && (!allowReservations))
                {
                    // This block is fully recycled.  Set the flag passed in.
                    recycleMemoryBlock = true;
                }
                releasedResource = true;
            }
        }
        catch (std::exception& e)
        {
            QLOG_DEBUG() << myLabelStr << "Exception:" << e.what();
        }
        catch (...)
        {
            QLOG_DEBUG() << myLabelStr << "(...)Exception!";
        }
    }
    else
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryMapId out of reserve range:" << mapRecordsReserved;
    }

exitReleaseReservedChunk:

    FiMemoryPool::LogVerbose((QString("%1 [memoryBlockId=%2][#Reserved=%3][#InUse=%4][#Recycled=%5][nextOffset=%6][Capacity=%7][Len=%8] releasedChunk[%9] recycleMemoryBlock[%10]").
                              arg(myLabelStr).
                              arg(memoryBlockId).arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset).
                              arg(capacity).arg(length).arg((releasedResource ? "T" : "F"), (recycleMemoryBlock ? "T" : "F"))).toUtf8().constData(), VLOG_MAP_RELEASE);

    return releasedResource;
}

/*!
 * @brief Turns off the ability to ask this memory block for a chunk reservation.
 *
 * Reservations are turned off when object is constructed, when the memory pool manager
 * or when the memory pool manager cannot get the reservation it requested.  It is a way
 * to signal that this object is consumed and it is used for recycle state.
 *
 * @returns The number of reservations that are marked "in-use".
 */
uint32_t FiMemoryBlock::TurnOffReservations()
{
    QString myLabelStr = QString("%1[TurnOffReservations]").arg(memBlockIdStr);

    //! Mutex to guard access this object's data.
    QMutex memMapMasterRecordMutex;

    allowReservations = false;

    // Log the information.
    FiMemoryPool::LogVerbose((QString("%1 [memoryBlockId=%2][#Reserved=%3][#InUse=%4][#Recycled=%5][nextOffset=%6]").
                              arg(myLabelStr).
                              arg(memoryBlockId).arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset)).toUtf8().constData());
    return mapRecordsInUse;
}
/*!
 * @brief Turns on the ability to ask this memory block for a chunk reservation.
 *
 * Reservations are turned on when a new memory block is created or when a memory block is
 * being recycled.
 *
 * @returns The result of turning on reservations for this memory block.
 */
bool FiMemoryBlock::TurnOnReservations()
{
    bool reservationsTurnedOn = false;
    QString myLabelStr = QString("%1[TurnOnReservations]").arg(memBlockIdStr);

    //! Mutex to guard access this object's data.
    QMutex memMapMasterRecordMutex;

    if (mapRecordsInUse > 0)
    {
        // Put up a warning this should NOT happen!
        FiMemoryPool::LogVerbose((QString("%1 %2 [memoryBlockId=%3][#Reserved=%4][#InUse=%5][#Recycled=%6][nextOffset=%7]").
                                  arg(myLabelStr, memPoolWarnMsg).
                                  arg(memoryBlockId).arg(mapRecordsReserved).arg(mapRecordsInUse).arg(mapRecordsRecycled).arg(nextFreeOffset)).toUtf8().constData());
        //goto exitTurnOnReservations;
    }

    // REVISIT_JLC - there should never be any records in use.  It is a programmatic error if so.
    // Because of this perhaps I should force the recycle?  Right now I ignore and continue.

    allowReservations = true;
    mapRecordsReserved = 0;
    mapRecordsInUse = 0;
    mapRecordsRecycled = 0;
    nextFreeOffset = 0;
    reservationsTurnedOn = true;

//exitTurnOnReservations:
    return reservationsTurnedOn;
}


