/*!
 * @file fi_mem_block.h
 *
 * @brief An object to manage a single array of memory in our memory pool.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FIMEMORYBLOCK_H
#define FIMEMORYBLOCK_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <cstdint>
#include <memory>
#include <vector>
#include <QMutex>

#include "fi_mem_pool.h"

namespace FI_MEMPOOL
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! Define the type of data stored in a memory block managed by FiMemData
typedef std::vector<uint8_t>    MEMORY_BLOCK;

class FiMemMapRecord;   // Forward declaration of embedded record class.

class FiMemoryBlock
{
public:
    // A Note:  - in this destructor there could be a setting used to determing
    // if the object is being destructed too-early - if we start moving these
    // objects through the system and it gets deleted by accident this would be
    // the only place to truly catch that condition.  So the pool could potentially
    // setup a state for when it is ok to destruct.
    virtual ~FiMemoryBlock();

    static std::unique_ptr<FI_MEMPOOL::FiMemoryBlock> CreateMemoryBlock(int myPoolId, uint32_t dataSize);

    int GetMemResource(uint32_t resourceSizeRequested);
    bool ReleaseMemoryResource(int memoryMapId, bool& recycleMemoryBlock);

    uint32_t AppendBytes(int memoryMapId, uint32_t appendSize, const uint8_t *bytesToAppend);
    uint8_t *AppendBytesExternally(int memoryMapId, uint32_t sizeRequested, uint32_t& bytesAvailable);
    bool CompleteExternalAppend(int memoryMapId, uint32_t appendByteSize);

    int GetPoolId();
    uint32_t GetCapacity(int memoryMapId);
    uint32_t GetLength(int memoryMapId);
    const uint8_t *GetReservedPtr(int memoryMapId, uint32_t& dataLength);

    uint32_t ReplaceBytes(int memoryMapId, uint32_t memBytesToSetLen, const uint8_t *memBytesToSet);

    uint32_t TurnOffReservations();
    bool TurnOnReservations();

    void LogMyId(const char *moreInfoStr = nullptr);

    void PrepareToDestruct();

private:
    // This is not a singleton object but the only way to create one of these is to use
    // the factory interface.
    FiMemoryBlock();

    //! This object is non-copyable and non-copy assignable.
    FiMemoryBlock(const FiMemoryBlock& i) = delete;
    FiMemoryBlock& operator=(const FiMemoryBlock& i) = delete;

    //! The memory pool id for this memory block object.
    int memoryBlockId;

    //! Flag that determines if reservations are allowed.
    bool allowReservations;

    //! Flag when set will warn if destructed.
    bool warnIfDestructed;

    //! Mutex to guard access this object's data.
    QMutex memMapMasterRecordMutex;

    uint32_t mapRecordsReserved;    //!< How many map records are reserved for use.
    uint32_t mapRecordsInUse;       //!< How many map records are in use.
    uint32_t mapRecordsRecycled;    //!< How many map records came back for recyling.
    uint32_t nextFreeOffset;        //!< Where in the memory block is the free data?

    //! Size of the memory block from the original create call storing the
    //! requested size as the memory block's capacity.  The actual size of
    //! the memory block will be slightly larger if a byte-alignment occurred
    //! during the memory block creation but we will not include those extra
    //! alignment bytes (if any) in here.
    uint32_t memoryBlockCapacity;

    //! The collection of records mapping usage over a memory block.
    std::unique_ptr<std::vector<FiMemMapRecord>> memMapRecords;  //!< How the memory block is used.

    //! Allocated memory block managed by this master memory map record.
    std::unique_ptr<FI_MEMPOOL::MEMORY_BLOCK> memoryBlock;

    // These were created to give some kind of a unique ID to the constructed object.
    // It is useful to log this information out to the system log when debugging the
    // implementation.
    static uint32_t nextId;
    uint32_t myId;
};

} /* End of namespace FI_MEMPOOL */

#endif // FIMEMORYBLOCK_H
