/*!
 * @file fi_mem_pool.cpp
 *
 * @brief A memory pooling object to manage a uint8_t array buffers in our system
 * memory pool.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */


//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QMutexLocker>
#include <QtGlobal>

#include <vector>

#include "logger.h"
#include "fi_mem_block.h"
#include "fi_mem_pool.h"

namespace FI_MEMPOOL
{

//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------
//const bool VERBOSE_LOGGING = true;
const bool VERBOSE_LOGGING = false;

//! The maximum number of FiMemoryBlock allowed for emergency allocations.
const int   MAX_EMERGENCY_ALLOCATIONS       = 75;

//! The minimum number of FiMemoryBlock allocations for the pool.
const int   MIN_BLOCK_ALLOCATIONS           = 40;

//! The maximum number of FiMemoryBlock allocations for the pool.
const int   MAX_BLOCK_ALLOCATIONS           = 80;

} // End of the FI_MEMPOOL namespace declaration.

// Make things easier for usage.
using namespace std;
using namespace FI_MEMPOOL;

//--------------------------------------------------------------------
// Static Instances privately managed in this module.
//--------------------------------------------------------------------

std::unique_ptr<FI_MEMPOOL::FiMemoryPool> FiMemoryPool::fiMemoryPoolPtr;

//--------------------------------------------------------------------
// Implementation of the FiMemoryResource
//--------------------------------------------------------------------

FiMemoryResource::FiMemoryResource(int poolBlockIdToSet,
                                   int blockMapIdToSet)
{
    poolBlockId         = poolBlockIdToSet;
    blockMapId          = blockMapIdToSet;
}

FiMemoryResource::~FiMemoryResource()
{
}

/*!
 * @brief Append bytes to memory resource
 *
 * @param appendSize - Number of bytes to add to the memory resource buffer.
 * @param bytesToInsert - the bytes to append.
 *
 * @returns The number of bytes appended to the memory resource bufer.  If it
 * returns 0 there is an error condition.
 */
uint32_t FiMemoryResource::AppendBytes(uint32_t appendSize, const uint8_t *bytesToAppend)
{
    uint32_t bytesAppended = 0;
    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    // Is the append data pointer valid?
    if (bytesToAppend ==  nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "bytesToAppend was nullptr!";
        goto exitAppendBytes;
    }

    if (appendSize == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "appendSize was zero!";
        goto exitAppendBytes;
    }

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitAppendBytes;
    }

    // Make the call to the block object to handle the append request.
    bytesAppended = memoryBlockPtr->AppendBytes(blockMapId, appendSize, bytesToAppend);

exitAppendBytes:
    return bytesAppended;
}

/*!
 * @brief Append bytes to the image data externally.
 * This will allow access to the append point into our data for external
 * manipulation.  This is necessary in order to facilitate socket reads
 * or other such function calls so that an intermediate buffer is not necessary.
 * This is in the interest of time to copy bytes from network transfers.
 *
 * @param sizeRequested - Number of bytes to add to the data buffer.
 * @param bytesAvailable - Set before returning to bytes available in buffer.
 *
 * @returns The append point into the image data exposing access to the internal image
 * data - caller MUST be wary of exceeding the image data buffer in memory.  Returns a
 * nullptr if there is not enough space in image data buffer.
 */
uint8_t *FiMemoryResource::AppendBytesExternally(uint32_t sizeRequested, uint32_t& bytesAvailable)
{
    uint8_t *externalAccessPtr      = nullptr;
    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    // Set the bytes available to zero here in case early error condition.
    bytesAvailable = 0;

    // Sanity checks.
    if (sizeRequested == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "sizeRequested was zero!";
        goto exitAppendBytesExternally;
    }

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitAppendBytesExternally;
    }

    // Make the call to the block object to handle the append request.
    externalAccessPtr = memoryBlockPtr->AppendBytesExternally(blockMapId, sizeRequested, bytesAvailable);

exitAppendBytesExternally:
    return externalAccessPtr;
}

/*!
 * @brief Marks an appendBytesExternally as being completed.  Otherwise the memory chunk
 * will remain in a condition of a pending write and that will cause many issues.
 *
 * @param appendByteSize - How many bytes were actually written in external append operation.
 *
 * @returns The result of the internal modification.  It can fail if internal state or
 * data modification does not pass sanity checks.
 */
bool FiMemoryResource::CompleteExternalAppend(uint32_t appendByteSize)
{
    bool externalAppendRegistered = false;

    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitCompleteExternalAppend;
    }

    // Make the call to the block object to handle the completion of the external append operation.
    externalAppendRegistered = memoryBlockPtr->CompleteExternalAppend(blockMapId, appendByteSize);

exitCompleteExternalAppend:

    return externalAppendRegistered;
}

/*!
 * @brief Pulls a read-only pointer to the memory resource and sets the length of that
 * read-only data in dataLength.
 *
 * @param dataLength - Will hold the number of bytes in the read-only pointer returned or
 * be set to zero if an error occurs.
 *
 * @returns The read-only pointer for the resource data or a nullptr if an error occurs.
 */
const uint8_t *FiMemoryResource::GetReservedPtr(uint32_t& dataLength)
{
    dataLength = 0;
    const uint8_t *memDataPtr = nullptr;

    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitGetReservedPtr;
    }

    // The call to retrieve the read-only pointer to the memory.
    memDataPtr = memoryBlockPtr->GetReservedPtr(blockMapId, dataLength);

exitGetReservedPtr:
    return memDataPtr;
}

/*!
 * @brief Provides the size of this resource "chunk" in the memory block.
 * This is the full capacity of the chunk not the length of data in-use in
 * the chunk.
 *
 * @returns The size in bytes of the capacity for this resource, or zero if error.
 */
uint32_t FiMemoryResource::GetCapacity()
{
    uint32_t dataCapacity = 0;

    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitGetLength;
    }

    // How many bytes exist in this resource for use.
    dataCapacity = memoryBlockPtr->GetCapacity(blockMapId);

exitGetLength:
    return dataCapacity;
}

/*!
 * @brief Provides the length of the data in use in this resource.
 *
 * @returns The size in bytes of the data in use in the resource or zero if there is an error.
 */
uint32_t FiMemoryResource::GetLength()
{
    uint32_t dataLength = 0;

    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitGetLength;
    }

    // How many bytes exist in this resource for use.
    dataLength = memoryBlockPtr->GetLength(blockMapId);

exitGetLength:
    return dataLength;
}

/*!
 * @brief Replaces bytes in the resource with new image bytes.
 *
 * @param memBytesToSet - A read only pointer to the new image bytes.
 * @param memBytesToSetLen - The number of image bytes to set.
 *
 * @returns The number of bytes that are now used in the image data.  If zero
 * the replacment failed.
 */
uint32_t FiMemoryResource::ReplaceBytes(uint32_t memBytesToSetLen, const uint8_t *memBytesToSet)
{
    uint32_t bytesReplaced = 0;

    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitReplaceBytes;
    }

    // The call to retrieve the read-only pointer to the memory.
    bytesReplaced = memoryBlockPtr->ReplaceBytes(blockMapId, memBytesToSetLen, memBytesToSet);

exitReplaceBytes:
    return bytesReplaced;
}

/*!
 * @brief Gives this object an opportunity to prepare for destructing.
 *
 * At the moment this is for diagnostic purposes.  Each memory data will
 * get a unique id that recycles at 10,000.  This is being logged right now
 * to track resource movement.
 */
void FiMemoryResource::LogMyId(const QString *moreInfoStr)
{
    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;
    QString myMoreInfoStr;

    // Get the memory block pointer from the pool.
    memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(poolBlockId);

    // If the pool could not find the memory block that is a serious error.
    if (memoryBlockPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Unable to find memory block pointer for: " << poolBlockId;
        goto exitLogMyId;
    }

    // Log the id for this memory block.
    myMoreInfoStr = QString("[poolBlockId=%1][memoryMapId=%2][Info=%3]").arg(poolBlockId).arg(blockMapId).arg(*moreInfoStr);
    memoryBlockPtr->LogMyId(myMoreInfoStr.toUtf8().constData());

exitLogMyId:
    return;
}

/*!
 * @brief Marks a resource as recycled and returns a memory block id to the free pool
 * if all of its resources have been recycled.
 *
 * This is only called by the FiMemoryResource custom deleter.
 *
 * It is also the only place where the resource is recycled.  It should be guaranteed
 * that the resource is always recycled due to the custom deleter mechanism.
 *
 * There should never be any resource leaks in the system with this approach.
 *
 * @param memoryResourcePtr - The FiMemoryResource pointer that holds the information
 * about the resource to recycle.
 *
 * Nothing is returned.
 */
void FiMemoryResource::RecycleResource(FiMemoryResource *memoryResourcePtr)
{
    // Make sure there is a pointer to recycle.
    if (memoryResourcePtr)
    {
        FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr   = nullptr;

        // Get the memory block pointer from the memory pool.
        memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(memoryResourcePtr->poolBlockId);

        // Release the chunk of memory that was reserved for this resource.
        if (memoryBlockPtr)
        {
            bool fullBlockReadyToRecycle = false;

            // Tell the memory block that we are done with our mapped resource data.
            bool releasedResource = memoryBlockPtr->ReleaseMemoryResource(memoryResourcePtr->blockMapId, fullBlockReadyToRecycle);

            // Important:  Even though the inUseCount may be zero it doesn't mean that the
            // FiMemoryBlock is fully consumed.  If the memory pool has stopped taking reservations
            // on this block (this only happens when the activeImagePoolId has moved on to a different
            // one - then it is at that point when the final resource is recycled at this point
            // that we must tell the memory pool to fully recycle the full block.  That is happening
            // right here when that flag goes high.
            if (releasedResource && fullBlockReadyToRecycle)
            {
                FiMemoryPool::RecycleMemoryBlock(memoryResourcePtr->poolBlockId);
            }
        }
    }
}

//--------------------------------------------------------------------
// Implementation of the Image Pool Class
//--------------------------------------------------------------------

/*!
 * @brief Memory pool standard constructor.
 */
FiMemoryPool::FiMemoryPool() :
    emergencyAllocations(0),
    lowWaterMark(0),
    lowWaterMarkPct(0),
    maxMemBufferSize(0),
    activeMemoryBlockId(0),
    memoryBlocksAllocated(0)
{
}

/*!
 * @brief Memory pool standard destructor.
 */
FiMemoryPool::~FiMemoryPool()
{
    QString myLabelStr = QString("%1[FiMemoryPool Destructor]").arg(memPoolLabelStr);

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Destructing the memory pool.";

    // Log any block ids that are on the memoryBlocksFree list.
    {
        // Grab the full size of the memory block list because it shrinks as they
        // are reset and destructed.
        uint32_t numberOfBlocks = (uint32_t) memoryBlockList.size();

        QMutexLocker memoryBlocksLocker(&memoryBlockMutex);

        while (!memoryBlocksFree.empty())
        {
            int memoryBlockId = memoryBlocksFree.at(0);
            memoryBlocksFree.pop_front();
            FiMemoryPool::LogVerbose((QString("%1 [memoryBlocksFree: memoryBlockId:%2]").arg(myLabelStr).arg(memoryBlockId)).toUtf8().constData());
        }

        while (!memoryBlocksToDelete.empty())
        {
            int memoryBlockId = memoryBlocksToDelete.at(0);
            memoryBlocksToDelete.pop_front();
            FiMemoryPool::LogVerbose((QString("%1 [memoryBlocksToDelete: memoryBlockId:%2]").arg(myLabelStr).arg(memoryBlockId)).toUtf8().constData());
        }

        // Traverse the blocklist pulling the pointers and turning of the warning system.
        for (uint32_t blockIndex = 0; blockIndex < numberOfBlocks; blockIndex++)
        {
            FiMemBlockPtr memoryBlockPtr = memoryBlockList[(unsigned long long)blockIndex].get();

            if (memoryBlockPtr)
            {
                memoryBlockPtr->PrepareToDestruct();
                FiMemoryPool::LogVerbose((QString("%1 [PrepareToDestruct: : FiMemoryBlockId=%2]").arg(myLabelStr).arg(blockIndex)).toUtf8().constData());

                // If this is in the non-allocated location in the list pop out a warning.
                if (blockIndex >= memoryBlocksAllocated)
                {
                    QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "[memoryBlockId:" << blockIndex << " was NOT nullptr and outside of allocation range!]";
                }
            }
            else
            {
                // If there was supposed to be a memory allocated block log a warning.
                if (blockIndex < memoryBlocksAllocated)
                {
                    QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "[memoryBlockId:" << blockIndex << " was NULL!]";
                }
                // Otherwise it's an empty slot in a shrunken list and a nullptr was expected.
            }
        }
    }
}

/*!
 * @brief How many image data buffers are available.
 *
 * @returns The count or zero if nothing is available for use.
 */
int FiMemoryPool::AvailableCount()
{
    int theCount = 0;

    if (fiMemoryPoolPtr)
    {
        theCount = (int)fiMemoryPoolPtr->memoryBlocksFree.size();
    }
    return theCount;
}

/*!
 * @brief Calculate a resume point for image pool in low-water conditions.
 *
 * @returns True if there is an image pool and this calculation is valid, false if not.
 */
bool FiMemoryPool::CalculateResumePoint(uint32_t& resumePoint)
{
    bool hasAnImagePool = false;

    resumePoint = 0;
    // Only if the image pool exists and has data can we determine a resume point.
    if (fiMemoryPoolPtr)
    {
        if (fiMemoryPoolPtr->memoryBlocksAllocated > 0)
        {
            // Make sure that the lowWaterMark is representative of the true size of
            // the memory pool using the percentage provided at pool creation time.
            fiMemoryPoolPtr->lowWaterMark = (uint32_t)(fiMemoryPoolPtr->lowWaterMarkPct * (float)fiMemoryPoolPtr->memoryBlocksAllocated);

            // Now set up the resume/available images for reporting purposes.

            // This is the calculation that exists currently in the FSPSocket code.
            // This will provide a handy calculator for stream services using this
            // central location.
            resumePoint = fiMemoryPoolPtr->lowWaterMark + (((uint32_t)fiMemoryPoolPtr->memoryBlocksAllocated - fiMemoryPoolPtr->lowWaterMark) >> 1);
            hasAnImagePool = true;
        }
    }

    return hasAnImagePool;
}

/*!
 * @brief Factory to create the image pool object.
 * This factory will create the single instance of the image pool.
 *
 * @param numberOfMemBuffers - contains the size of the image pool to create.
 * @param maxMemBufferSize - contains the maximum size of the FiMemData buffers.
 * @param lowWaterMarkToSet - contains the point at which the image pool goes
 * into a low-water condition (running low on free image buffers in the pool).
 *
 * @returns Result of the image pool creation.
 */
bool FiMemoryPool::CreateMemoryPool(
    uint32_t numberOfBuffers,
    uint32_t maxSize,
    uint32_t lowWaterMarkToSet)
{
    bool memoryPoolCreatedOk = false;

    // This should onlybe called one time for the application.
    if (fiMemoryPoolPtr == nullptr)
    {
        fiMemoryPoolPtr = std::unique_ptr<FI_MEMPOOL::FiMemoryPool>(new FI_MEMPOOL::FiMemoryPool());
        //fiMemoryPoolPtr = std::make_unique<FI_MEMPOOL::FiMemoryPool>();

        // Was the memory pool object created?
        if (fiMemoryPoolPtr)
        {
            // Allocate the memory pool entries.
            if (!fiMemoryPoolPtr->AllocatePool(numberOfBuffers, maxSize, lowWaterMarkToSet))
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Memory Pool factory creation failure.";
                goto exitCreateMemoryPool;
            }
            // The memory pool has been created and allocated.
            memoryPoolCreatedOk = true;
        }
    }

exitCreateMemoryPool:
    return memoryPoolCreatedOk;
}

/*!
 * @brief Deletes the singleton and releases all memory and resources.
 * Provides opportunity to control destructor of the memory pool.
 */
void FiMemoryPool::DeleteMemoryPool()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Processing request to delete the memory pool.";

    if (fiMemoryPoolPtr != nullptr)
    {
        fiMemoryPoolPtr.reset();
    }
}

/*!
 * @brief Returns a pointer to an FiMemoryBlock record.
 *
 * This is also utilized by our friend FiMemoryResourcePtr which is generated from
 * GetMemoryResource.  It is statically private for this reason.  It serves as the
 * external interface to our singleton FiMemoryPool object and its purpose is to
 * locate and return for use internally by our friend FiMemoryResource.
 *
 * @param memoryBlockId - The memory block id of the memory block pointer requested.
 *
 * @returns The memory block pointer OR a nullptr if not there.
 */
FiMemBlockPtr FiMemoryPool::GetMemoryBlockPtr(int memoryBlockId)
{
    QString myLabelStr = QString("%1[GetMemoryBlockPtr]").arg(memPoolLabelStr);
    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr = nullptr;

    if (fiMemoryPoolPtr == nullptr)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Missing FiMemoryPool singleton!";
        return memoryBlockPtr;
    }

    // Note:  The memoryBlockList doesn't move around in terms of element position.
    // once a memory block is in this list it is non-volatile and we can be assured
    // that we do not need a mutex to lock the list when pulling the block pointer.

    if (memoryBlockId < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "Missing memoryBlockId!";
        goto exitGetMemoryBlockPtr;
    }

    if (fiMemoryPoolPtr)
    {
        // Allow retrieval of entries outside of the "allocated marker".
        int poolSize = (int)fiMemoryPoolPtr->memoryBlockList.size();

        // If outside the allocated block range but within the physical size it is probably
        // a request for a "to-be-deleted" block.  One that was in use when a memory pool
        // resize operation occurred.  Go ahead and log a message if verbosity setting is on.
        if ((uint32_t)memoryBlockId >= fiMemoryPoolPtr->memoryBlocksAllocated)
        {
            FiMemoryPool::LogVerbose((QString("%1 memoryBlockId[%2] exceeds allocated poolsize[%3] probably a to-be-deleted block.").arg(myLabelStr).arg(memoryBlockId).arg(poolSize)).toUtf8().constData());
        }

        // Terminate if outside of the physical poolsize.
        if (memoryBlockId >= poolSize)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << memPoolWarnMsg << "memoryBlockId: " << memoryBlockId << " exceeds physical poolsize: " << poolSize;
            goto exitGetMemoryBlockPtr;
        }

        // If make it here there should be a valid pool entry.
        //
        // Note:  Because the memoryBlockList does not change (it is a list of
        // unique_ptrs to FiMemoryBlock objects) it is safe to return a pointer
        // to this object.  Even if the memoryBlockList is grown it won't matter
        // because the pointer being returned here is the unique_ptr given to the
        // FiMemoryBlock object not the element in the list.
        //
        // Also:
        // We don't need to use the .at(index) here because we've already done the
        // bounds checking above.  It is safe to use the element access[] to get pointer.
        memoryBlockPtr = fiMemoryPoolPtr->memoryBlockList[(unsigned long long)memoryBlockId].get();
    }

exitGetMemoryBlockPtr:
    return memoryBlockPtr;
}

/*!
 * @brief Returns a resource object for a requested data chunk out of the memory
 * block serving up resource reservations.
 *
 * @param requestedSize - How many bytes are needed for the resource reservation.
 * @param okToEmergencyAllocateIfEmpty - Flag that when turned on will allow additional
 * memory block to be created to facilitate this reservation request.
 *
 * @returns The memory resource object or a nullptr if an error occurs.  The ownership
 * of the resource moves to the caller because FiMemoryResourcePtr is a std::unique_ptr.
 */
FiMemoryResourcePtr FiMemoryPool::GetMemoryResource(uint32_t requestedSize, bool okToEmergencyAllocateIfEmpty)
{
    QString myLabelStr = QString("%1[GetMemoryResource]").arg(memPoolLabelStr);

    // Get the memory block id and the chunk id for our resource from the pool.
    int memoryBlockId                        = NO_MEMORY_BLOCK;
    int memoryMapId                         = NO_MEMORY_MAP;

    FiMemoryResourcePtr memoryResourcePtr;

    if (fiMemoryPoolPtr == nullptr)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Missing FiMemoryPool singleton!";
        return memoryResourcePtr;
    }

    // If unable to reserve the memory requested need to try with a new active memory block.
    if (!fiMemoryPoolPtr->ReserveMemoryMap(requestedSize, memoryBlockId, memoryMapId))
    {
        //FiMemoryPool::LogVerbose((QString("%1 Active memory block unable to reserve requestedSize[%2]").arg(myLabelStr).arg(requestedSize)).toUtf8().constData());

        // If the free list is exhausted then keep trying to grow the list until we have
        // an emergency free buffer to use.  This has to be a while loop now because we can
        // have memory blocks on the to-be-deleted list that are in use and growing doesn't
        // necessarily reallocate.  This will grow until there is one to use.
        while (fiMemoryPoolPtr->memoryBlocksFree.empty())
        {
            std::unique_ptr<FI_MEMPOOL::FiMemoryBlock> memoryBlockPtr;
            uint32_t memoryBlockIdAdded = (uint32_t)fiMemoryPoolPtr->memoryBlocksAllocated;

            // If cannot allocate an emergency buffer exit.
            if (!okToEmergencyAllocateIfEmpty)
            {
                QLOG_DEBUG() << myLabelStr << "Out of memoryBlocksFree and !okToEmergencyAllocateIfEmpty!";
                goto exitGetMemoryResource;
            }

            // Have we allocated the maximum emergency allocation buffers?
            if (fiMemoryPoolPtr->emergencyAllocations > MAX_EMERGENCY_ALLOCATIONS)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << myLabelStr << "The emergency allocation count: " << fiMemoryPoolPtr->emergencyAllocations << " has exceeded MAX_EMERGENCY_ALLOCATIONS: " << MAX_EMERGENCY_ALLOCATIONS;
                goto exitGetMemoryResource;
            }

            // Grow our pool by one more.
            try
            {
                fiMemoryPoolPtr->GrowMemoryPool(fiMemoryPoolPtr->memoryBlocksAllocated + 1);
                FiMemoryPool::LogVerbose((QString("%1 memoryBlockId: %2 memoryBlockIdAdded: %3").arg(myLabelStr).arg(memoryBlockId).arg(memoryBlockIdAdded)).toUtf8().constData());
                fiMemoryPoolPtr->emergencyAllocations++;    // Increment the emergency allocation count.
            }
            catch (std::exception& e)
            {
                QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
            }
            catch (...)
            {
                QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
            }
        }

        // Make sure there is something to activate.
        if (!fiMemoryPoolPtr->memoryBlocksFree.empty())
        {
            try
            {
                int releasedActiveId = NO_MEMORY_BLOCK;

                // We are done with the current active memory block and need to pull
                // a new one from the memoryBlocksFree list.
                {
                    QMutexLocker memoryBlocksLocker(&fiMemoryPoolPtr->memoryBlockMutex);

                    // Save the active memory block id being released.
                    releasedActiveId = fiMemoryPoolPtr->activeMemoryBlockId;

                    // Install the new activeMemoryBlockId.
                    fiMemoryPoolPtr->activeMemoryBlockId = fiMemoryPoolPtr->memoryBlocksFree.front();
                    fiMemoryPoolPtr->memoryBlocksFree.pop_front();
                }

                // Tell the released active memory block that it is closed to reservations.
                // If there are no additional resource chunks out in use for this memory block
                // then it must be recycled in the system here or it will be missed.
                if (releasedActiveId != NO_MEMORY_BLOCK)
                {
                    FiMemBlockPtr releasedMemoryBlockPtr =  FiMemoryPool::GetMemoryBlockPtr(releasedActiveId);
                    if (releasedMemoryBlockPtr)
                    {
                        uint32_t reservationsInUse = releasedMemoryBlockPtr->TurnOffReservations();
                        if (reservationsInUse == 0)
                        {
                            // This memory block is no longer "in-use".  Therefore there should not
                            // be any additional reservation recycle events coming through the
                            // system and its block id must be recycled back to the free list.
                            FiMemoryPool::RecycleMemoryBlock(releasedActiveId);
                        }
                    }
                }
            }
            catch (std::exception& e)
            {
                QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
            }
            catch (...)
            {
                QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
            }

            // Make an attempt to service this reservation request.
            if (!fiMemoryPoolPtr->ReserveMemoryMap(requestedSize, memoryBlockId, memoryMapId))
            {
                // There will never be enough space to service this reservation because
                // the memory blocks are not large enough to hold the data.
                QLOG_DEBUG() << myLabelStr << "Requested size:" << requestedSize << " is larger than a pool memory block!";
                goto exitGetMemoryResource;
            }
        }
        else
        {
            QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryBlocksFree is empty!";
            goto exitGetMemoryResource;
        }
    }

    try
    {
        // If arrive here and there are valid block and map ids the memory was reserved then
        // the resource can be made.
        if ((memoryBlockId != NO_MEMORY_BLOCK) && (memoryMapId != NO_MEMORY_MAP))
        {
            memoryResourcePtr = (FiMemoryResourcePtr) new FiMemoryResource(memoryBlockId, memoryMapId);
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

exitGetMemoryResource:
    if (memoryResourcePtr == nullptr)
    {
        if (okToEmergencyAllocateIfEmpty)
        {
            QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Unable to service GetMemoryResource request!";
        }
        else
        {
            QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Unable to service GetMemoryResource request, try turning on okToEmergencyAllocateIfEmpty next time.";
        }
        FiMemoryPool::LogVerbose((QString("%1 activeId[%6] memoryBlockId[%2] memoryMapId[%3] Size[%4], memoryResourcePtr[%5]").arg(myLabelStr).arg(memoryBlockId).arg(memoryMapId).arg(requestedSize).arg(memoryResourcePtr ? "VALID" : "NULL").arg(fiMemoryPoolPtr->activeMemoryBlockId)).toUtf8().constData());
    }

    return memoryResourcePtr;
}

/*!
 * @brief Is there a memory pool available for use?
 *
 * @returns True if the memory pool exists, false if not.
 */
bool FiMemoryPool::HasAMemoryPool()
{
    bool hasMemoryPool = false;
    if (fiMemoryPoolPtr)
    {
        hasMemoryPool = true;
    }
    return hasMemoryPool;
}

/*!
 * @brief Returns the low water mark condition.
 *
 * @returns True if the low water mark condition is in effect, false if not.
 */
bool FiMemoryPool::IsPoolLow()
{
    bool lowWaterMarkFlag = false;
    if (fiMemoryPoolPtr)
    {
        lowWaterMarkFlag = ((uint32_t)fiMemoryPoolPtr->memoryBlocksFree.size() <= fiMemoryPoolPtr->lowWaterMark) ? true : false;
    }
    return lowWaterMarkFlag;
}

/*!
 * @brief Handles a central point to funnel all logging that is considered
 * "verbose" for debugging purposes.  It can be turned on and off based on
 * the setting of VERBOSE_LOGGING.
 */
void FiMemoryPool::LogVerbose(const char *msg, uint32_t vlogFlag)
{
    bool logIt = false;

    // There are two conditions that will supercede a VERBOSE_LOGGING setting
    // and that will be VLOG_SHOW_ERROR and VLOG_SHOW_ALWAYS.
    // - if that is a setting we don't look at the VERBOSE_LOGGING conditional.
    if (vlogFlag & VLOG_SHOW_ERROR)
    {
        logIt = true;
    }
    else if (vlogFlag & VLOG_SHOW_ALWAYS)
    {
        logIt = true;
    }
    else if (VERBOSE_LOGGING)
    {
        if (vlogFlag & VLOG_TO_PRINT)
            logIt = true;
    }

    // Log the message if configuration says to do so.
    if (logIt)
    {
        QLOG_INFO() << msg;
    }
}

/*!
 * @brief Makes a reservation for a memory resource.
 *
 * @param requestedSize - the size requested for the memory resource.
 * @param memoryBlockId - a reference to set for the memory block id reserved.  It will
 * be set to NO_MEMORY_BLOCK if the reservation cannot be made.
 * @param memoryMapId - a reference to set for the memory map id reserved in the memory block.
 * It will be set to NO_MEMORY_MAP if the reservation cannot be made.
 *
 * @returns True if the reservation is made for the memory otherwise false.  Note the return state
 * of the two reference parameters.
 */
bool FiMemoryPool::ReserveMemoryMap(uint32_t requestedSize, int& memoryBlockId, int& memoryMapId)
{
    QString myLabelStr = QString("%1[ReserveMemoryMap]").arg(memPoolLabelStr);
    bool reservedAChunk = false;

    FI_MEMPOOL::FiMemoryBlock *memoryBlockPtr     = nullptr;
    int activeBlockId   = NO_MEMORY_BLOCK;

    // Set the request reference data to an empty state.
    memoryBlockId        = NO_MEMORY_BLOCK;
    memoryMapId         = NO_MEMORY_MAP;

    try
    {
        activeBlockId = activeMemoryBlockId;
        // Does the active memory block have space for this reservation?
        if (activeBlockId > NO_MEMORY_BLOCK)
        {
            memoryBlockPtr = FiMemoryPool::GetMemoryBlockPtr(activeBlockId);
            if (memoryBlockPtr)
            {
                memoryMapId = memoryBlockPtr->GetMemResource(requestedSize);
                if (memoryMapId != NO_MEMORY_MAP)
                {
                    memoryBlockId = activeBlockId;
                    reservedAChunk = true;
                }
            }
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr  << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr  << exceptionWarnMsg ;
    }

#if 0
    FiMemoryPool::LogVerbose((QString("%1 reservedAChunk[%2] activeBlockId[%3] memoryBlockId[%4] memoryMapId[%5] requestedSize[%6]").
                              arg(myLabelStr).arg(reservedAChunk ? "T" : "F").arg(activeBlockId).arg(memoryBlockId).arg(memoryMapId).arg(requestedSize)).toUtf8().constData());
#endif

    if (!reservedAChunk)
    {
        FiMemoryPool::LogVerbose((QString("%1 reservedAChunk[%2] activeBlockId[%3] memoryBlockId[%4] memoryMapId[%5] requestedSize[%6]").
                                  arg(myLabelStr, (reservedAChunk ? "T" : "F")).arg(activeBlockId).arg(memoryBlockId).arg(memoryMapId).arg(requestedSize)).toUtf8().constData());
    }

    return reservedAChunk;
}

/*!
 * @brief Handles the process of putting consumed blocks back on the
 * free list.
 *
 * Because it is possible to have a memory block recycled before the
 * entire memory block was consumed it is important not to recycle the
 * block too early.
 *
 * Will use this point in the processing to push all blocks that are
 * consumed and fully recycled back to the free list.
 *
 * @param memoryBlockId - the memory block to process.
 *
 */
void FiMemoryPool::RecycleMemoryBlock(int memoryBlockId)
{
    QString myLabelStr = QString("%1[RECYCLE MEMORY BLOCK]").arg(memPoolLabelStr);

    // Make sure the singleton exists.
    if (fiMemoryPoolPtr == nullptr)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Missing FiMemoryPool singleton!";
        goto exitRecycleMemoryBlock;
    }

    // Make sure the memory block id seems valid.
    if (memoryBlockId == NO_MEMORY_BLOCK)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "memoryBlockId received was set to NO_MEMORY_BLOCK!";
        goto exitRecycleMemoryBlock;
    }

    try
    {
        QMutexLocker memoryBlockLocker(&fiMemoryPoolPtr->memoryBlockMutex);
        FiMemBlockPtr blockToRecyclePtr = nullptr;
        bool markedForDeletion          = false;

        // If this memory block id is in the "to-be-deleted" list we don't want
        // to recycle the memoryBlockId back to the free list and we may want to
        // reset the memory pool block pointer.
        markedForDeletion = fiMemoryPoolPtr->TakeOffToBeDeleted((uint32_t)memoryBlockId);

        // Is our memory block id within the physical size of the memory pool list?
        if ((uint32_t)memoryBlockId < (uint32_t)fiMemoryPoolPtr->memoryBlockList.size())
        {
            // Pull the local pointer.  Remember we are recycling or deleting this entire block.
            blockToRecyclePtr = fiMemoryPoolPtr->GetMemoryBlockPtr(memoryBlockId);

            // Check for complete recycling of consumed block.
            if (blockToRecyclePtr)
            {
                // Only recycle this memory block if it wasn't marked for deletion.
                if (!markedForDeletion)
                {
                    // Turn back on reservation system for this block.
                    blockToRecyclePtr->TurnOnReservations();
                }
                else
                {
                    // Delete the memory by resetting the block list entry.
                    fiMemoryPoolPtr->ResetMemoryBlock((uint32_t)memoryBlockId);
                }

                FiMemoryPool::LogVerbose((QString("%1 memoryBlockId[%2]").arg(myLabelStr).arg(memoryBlockId)).toUtf8().constData());
            }
            else
            {
                FiMemoryPool::LogVerbose((QString("%1 %2 memoryBlockId[%3] Query produced nullptr.  Marked for deletion[%4].").arg(myLabelStr, memPoolWarnMsg).arg(memoryBlockId).arg(markedForDeletion ? "Yes" : "No")).toUtf8().constData(), VLOG_SHOW_ERROR);
            }

            // If not marked for deletion put it back on the free list.
            if (!markedForDeletion)
            {
                // Put the id back on our free list it's ready to be recycled.
                fiMemoryPoolPtr->memoryBlocksFree.push_back(memoryBlockId);
            }
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

exitRecycleMemoryBlock:
    return;
}

/*!
 * @brief Shrinks the memory pool.
 *
 * @param memoryPoolNewSize - how many buffers has the user set for system.
 *
 * @returns True if the memory pool was resized, false if error or other condition.
 */
bool FiMemoryPool::ShrinkMemoryPool(uint32_t memoryPoolNewSize)
{
    bool resizedPool = false;
    uint32_t numberOfBlocksReset = 0;

    QString myLabelStr = QString("%1[ShrinkMemoryPool]").arg(memPoolLabelStr);

    // If not shrinking warn and get out of here.
    if (memoryPoolNewSize >= memoryBlocksAllocated)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Requesting to shrink a memory pool: " << memoryPoolNewSize << " greater than or equal to existing size: " << memoryBlocksAllocated << "!";
        goto exitShrinkMemoryPool;
    }

    // Shrink the pool.
    try
    {
        int         numberToDelete      = (int)memoryBlocksAllocated - (int)memoryPoolNewSize;
        int         firstBlockIdToDel   = (int)memoryBlocksAllocated - numberToDelete;
        int         curBlockId;
        uint32_t    resumePoint;

        // Lock down the lists.
        QMutexLocker memoryBlockLocker(&fiMemoryPoolPtr->memoryBlockMutex);

        // Want to examine and remove the end numberToDelete memoryBlockList entries.
        // We process each block based on its reservation status and number of in-use reservations.
        // We must also remove any blocks that are outside of the new size from the free list.
        // If the active reservation block id is outside of the new size it must be reset to NO_MEMORY_BLOCK.
        for (curBlockId = firstBlockIdToDel; curBlockId < (int)memoryBlocksAllocated; curBlockId++)
        {
            FiMemBlockPtr memoryBlockPtr =  FiMemoryPool::GetMemoryBlockPtr(curBlockId);
            if (memoryBlockPtr)
            {
                uint32_t reservationsInUse = memoryBlockPtr->TurnOffReservations();
                if (reservationsInUse == 0)
                {
                    // This memory block is no longer "in-use".  Therefore there should not
                    // be any additional reservation recycle events coming through the
                    // system and its actual pointer can be reset now.  We have no way of
                    // knowing whether or not the user will grow the pool again and we have
                    // to assume they want memory resources back which is why they shrunk the
                    // pool in the first place so we will facilitate removing this memory now.
                    ResetMemoryBlock((uint32_t)curBlockId);
                }
                else
                {
                    // Theoretically this current block id could still be sitting on the
                    // to-be-deleted list so remove it if so.
                    TakeOffToBeDeleted((uint32_t)curBlockId);

                    // Push the memoryBlockId on the to-be-deleted list so that when the
                    // last resource to be recyled comes through the block is managed properly.
                    memoryBlocksToDelete.push_back(curBlockId);
                }

                // Even if the memory block was placed on the to-be-deleted list
                // Count the number of "reset" blocks here.
                numberOfBlocksReset++;
            }

            // If this is the active block id then set it to NO_MEMORY_BLOCK
            if (curBlockId == activeMemoryBlockId)
            {
                activeMemoryBlockId = NO_MEMORY_BLOCK;
            }

            // Make sure that this is block is not on the free list.
            TakeOffMemoryBlockFreeList((uint32_t)curBlockId);
        }

        // Shrink the memoryBlocksAllocated at this point.  It was
        // being used as a loop control value above and only now is
        // it safe to set it.
        if (numberOfBlocksReset > 0)
        {
            memoryBlocksAllocated -= numberOfBlocksReset;
            resizedPool = true;
        }

        // Regenerate low water mark and pull the resume point.
        CalculateResumePoint(resumePoint);

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "memoryBlocksAllocated:" << memoryBlocksAllocated << "backoff:"
                     << lowWaterMark << "resume:" << resumePoint << " resizedPool:" << (resizedPool ? "True" : "False");

    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

exitShrinkMemoryPool:

    return resizedPool;
}

/*!
 * @brief Takes an entry off the free list.
 *
 * !! IMPORTANT !! Caller sets and holds a lock on queues.  This is a private
 * utility function that is reused internally so it was made into a function.
 *
 * @param memoryBlockIdToFree - the id of the memory block to look for in free list.
 *
 * @returns True if the memoryBlockId was found in the list, false if not.
 */
bool FiMemoryPool::TakeOffMemoryBlockFreeList(uint32_t memoryBlockIdToFree)
{
    bool foundItInList      = false;
    QString myLabelStr  = QString("%1[TakeOffToFreeList]").arg(memPoolLabelStr);

    try
    {
        uint32_t theIndex;

        uint32_t theSize = (uint32_t) memoryBlocksFree.size();
        for (theIndex = 0; theIndex < theSize && !foundItInList; theIndex++)
        {
            if ((int)memoryBlockIdToFree == memoryBlocksFree.at(theIndex))
            {
                // The block id exists on the to be deleted queue.
                // Set the memory block id to a NO_MEMORY_BLOCK state.
                memoryBlocksFree.at(theIndex) = NO_MEMORY_BLOCK;

                // Found this in the list.
                foundItInList = true;
            }
        }

        // Go through and remove any NO_MEMORY_BLOCK entries.
        if (foundItInList)
        {
            // Grab the size so we know how many to process.
            theSize = (uint32_t) memoryBlocksFree.size();

            // Run through the number of entries that are in the list.
            for (theIndex = 0; theIndex < theSize; theIndex++)
            {
                int memoryBlockId = memoryBlocksFree.front();
                memoryBlocksFree.pop_front();
                if (memoryBlockId != NO_MEMORY_BLOCK)
                {
                    // Put the memory block back on the free list.
                    memoryBlocksFree.push_back(memoryBlockId);
                }
                // Else it is simply removed from the list.
            }
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

    return foundItInList;
}

/*!
 * @brief Takes an entry off the to be deleted queue.
 *
 * !! IMPORTANT !! Caller sets and holds a lock on queues.  This is a private
 * utility function that is reused internally so it was made into a function.
 *
 * @param memoryBlockId - the id of the memory block to look for in to-be deleted list.
 *
 * @returns True if the memoryBlockId was found in the list, false if not.
 */
bool FiMemoryPool::TakeOffToBeDeleted(uint32_t memoryBlockIdToRemove)
{
    bool foundItInList      = false;
    QString myLabelStr  = QString("%1[TakeOffToBeDeleted]").arg(memPoolLabelStr);

    try
    {
        uint32_t theSize = (uint32_t) memoryBlocksToDelete.size();
        uint32_t theIndex;

        for (theIndex = 0; theIndex < theSize && !foundItInList; theIndex++)
        {
            if ((int)memoryBlockIdToRemove == memoryBlocksToDelete.at(theIndex))
            {
                // The block id exists on the to be deleted queue.
                // Set the memory block id to a NO_MEMORY_BLOCK state.
                memoryBlocksToDelete.at(theIndex) = NO_MEMORY_BLOCK;

                FiMemoryPool::LogVerbose((QString("[%1] Taking off [memoryBlockId=%2] from the to be deleted queue.").arg(myLabelStr).arg(memoryBlockIdToRemove)).toUtf8().constData());

                // Found this in the list.
                foundItInList = true;
            }
        }

        // Go through and remove any NO_MEMORY_BLOCK entries.
        if (foundItInList)
        {
            theSize = (uint32_t) memoryBlocksToDelete.size();
            for (theIndex = 0; theIndex < theSize; theIndex++)
            {
                int memoryBlockIdFromList = memoryBlocksToDelete.front();
                memoryBlocksToDelete.pop_front();
                if (memoryBlockIdFromList != NO_MEMORY_BLOCK)
                {
                    // Data contains a memory block so put it back at the end of the list.
                    memoryBlocksToDelete.push_back(memoryBlockIdFromList);
                }
                // Else it is simply removed from the list.
            }
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

    return foundItInList;
}

/*!
 * @brief Resets an FiMemoryBlock unique pointer in the memoryBlockList corresponding
 * to a memory block id.
 *
 * !! IMPORTANT !! Caller sets and holds a lock on queues.  This is a static private
 * utility function that is reused internally so it was made into a function.
 *
 * @param memoryBlockId - the id of the memory block to reset.
 *
 * @returns True if the memoryBlockId was found in the list, false if not.
 */
bool FiMemoryPool::ResetMemoryBlock(uint32_t memoryBlockId)
{
    bool foundItInList      = false;
    QString myLabelStr  = QString("%1[ResetMemoryBlock]").arg(memPoolLabelStr);

    FiMemoryPool::LogVerbose((QString("%1 Resetting (freeing) the FiMemoryBlock for memoryBlockId[%2]").arg(myLabelStr).arg(memoryBlockId)).toUtf8().constData());

    try
    {
        // Make sure this isn't in the to-be-deleted list just in case.
        TakeOffToBeDeleted(memoryBlockId);

        // Is our memory block id within the physical size of the memory pool list?
        if (memoryBlockId < (uint32_t)fiMemoryPoolPtr->memoryBlockList.size())
        {
            // Pull the local pointer.  Remember we are recycling or deleting this entire block.
            FiMemBlockPtr blockToRecyclePtr = fiMemoryPoolPtr->GetMemoryBlockPtr((int)memoryBlockId);

            // Found this in the list.
            foundItInList = true;

            // Check for complete recycling of consumed block.
            if (blockToRecyclePtr)
            {
                uint32_t numberOfReservations = blockToRecyclePtr->TurnOffReservations();

                // If there were reservations it is an error to try to reset the memory block.
                if (numberOfReservations > 0)
                {
                    // As long as all of the programmatic logic is correct this should never be in the log.
                    FiMemoryPool::LogVerbose((QString("%1 %2 memoryBlockId[%3] Has %4 Reservations!").arg(myLabelStr, memPoolWarnMsg).arg(memoryBlockId).arg(numberOfReservations)).toUtf8().constData(), VLOG_SHOW_ERROR);
                }

                // Reset the block.
                blockToRecyclePtr->PrepareToDestruct();
                memoryBlockList.at((unsigned long long)memoryBlockId).reset();
            }
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

    return foundItInList;
}

/*!
 * @brief Grows the memory pool.
 *
 * @param memoryPoolNewSize - how many buffers has the user set for system.
 *
 * @returns True if the memory pool was resized, false if error or other condition.
 */
bool FiMemoryPool::GrowMemoryPool(uint32_t memoryPoolNewSize)
{
    bool resizedPool            = false;
    QString myLabelStr      = QString("%1[GrowMemoryPool]").arg(memPoolLabelStr);

    // If not growing warn and get out of here.
    if (memoryPoolNewSize <= memoryBlocksAllocated)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Requesting to grow a memory pool: " << memoryPoolNewSize << " less than or equal to size: " << memoryBlocksAllocated << "!";
        goto exitGrowMemoryPool;
    }

    // Grow the pool and reinstate any deleted buffers that still exist.
    try
    {
        uint32_t curBlockListSize   = (uint32_t) memoryBlockList.size();
        uint32_t curBlockId;        // This will be the first memory block id in new list.
        uint32_t resumePoint;

        // Lock down access.
        QMutexLocker memoryBlockLocker(&memoryBlockMutex);

        // Move through the block list queue at the end from current allocation to new size.
        for (curBlockId = memoryBlocksAllocated; curBlockId < memoryPoolNewSize; curBlockId++)
        {
            bool blockExists = false;   // Reset for loop pass.

            // Is there a place for this pointer in the existing memory block list already?
            if (curBlockListSize > curBlockId)
            {
                (void)TakeOffToBeDeleted(curBlockId);    // Remove from to-be-deleted list if there.

                // Pull the local pointer.
                FiMemBlockPtr blockToRecyclePtr = fiMemoryPoolPtr->memoryBlockList[(unsigned long long)curBlockId].get();

                // If there is a block of memory here reclaim it for use.
                if (blockToRecyclePtr)
                {
                    uint32_t numberOfReservations = blockToRecyclePtr->TurnOffReservations();
                    // Determine if there are outstanding reservations.
                    if (numberOfReservations == 0)
                    {
                        // Go ahead and recycle the memory block to the free list.
                        blockToRecyclePtr->TurnOnReservations();
                        memoryBlocksFree.push_back((int)curBlockId);
                        memoryBlocksAllocated++;
                    }
                    // The else part of this just lets the reservations stay
                    // "in-use" because they will eventually come back to be
                    // recycled when their resources are deleted.

                    // Without regard to reservations - we don't reallocate the block of memory.
                    blockExists = true;
                }
            }

            // If block doesn't exist already then allocate a new one.
            if (!blockExists)
            {
                // Note:  memoryBlocksAllocated and memoryBlocksFree lists are
                // modified in this function and this function relies on the
                // mutex lock being held by this funtion it does not grab a lock.
                blockExists = AllocBlock(curBlockId, maxMemBufferSize);

                // Report any issues with the allocation.
                if (!blockExists)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "AllocBlock(" << curBlockId << ", " << maxMemBufferSize << ") FAILED!";
                }
            }
        }

        // Regenerate low water mark and pull the resume point.
        CalculateResumePoint(resumePoint);

        resizedPool = (memoryBlocksAllocated == memoryPoolNewSize);

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "memoryBlocksAllocated:" << memoryBlocksAllocated << "backoff:"
                     << lowWaterMark << "resume:" << resumePoint << " resizedPool:" << (resizedPool ? "True" : "False");
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

exitGrowMemoryPool:

    return resizedPool;
}

/*!
 * @brief Resizes the memory pool.
 *
 * @param memoryPoolNewSize - how many buffers has the user set for system.
 *
 * @returns True if the memory pool was resized, false if error or other condition.
 */
bool FiMemoryPool::ResizeMemoryPool(uint32_t memoryPoolNewSize)
{
    bool resizedPool = false;
    QString myLabelStr = QString("%1[ResizeMemoryPool]").arg(memPoolLabelStr);

    // Make sure the singleton exists.
    if (fiMemoryPoolPtr == nullptr)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Missing FiMemoryPool singleton!";
        goto exitResizeMemoryPool;
    }

    // Make sure that the new pool size is not less than our minimum size.
    if (memoryPoolNewSize < MIN_BLOCK_ALLOCATIONS)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Requesting a memory pool: " << memoryPoolNewSize << " less than minimum: " << MIN_BLOCK_ALLOCATIONS << "!  Setting to minimum.";
        memoryPoolNewSize = MIN_BLOCK_ALLOCATIONS;
    }

    // Make sure that the new pool size is not greater than our maximum size.
    if (memoryPoolNewSize > MAX_BLOCK_ALLOCATIONS)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "Requesting a memory pool: " << memoryPoolNewSize << " greater than maximum: " << MAX_BLOCK_ALLOCATIONS << "!  Setting to maximum.";
        memoryPoolNewSize = MAX_BLOCK_ALLOCATIONS;
    }

    try
    {
        // If growing the pool add the new blocks.
        if (memoryPoolNewSize > fiMemoryPoolPtr->memoryBlocksAllocated)
        {
            resizedPool = fiMemoryPoolPtr->GrowMemoryPool(memoryPoolNewSize);
        }
        // If shrinking pool handle the marking for deletion or resetting necessary.
        else if (memoryPoolNewSize < fiMemoryPoolPtr->memoryBlocksAllocated)
        {
            resizedPool = fiMemoryPoolPtr->ShrinkMemoryPool(memoryPoolNewSize);
        }
        else
        {
            // Pool equal to requested setting so mark the flag as being
            // resized even though nothing changed.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requested resize but current size is equal to the existing size.";
            resizedPool = true;
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

exitResizeMemoryPool:

    return resizedPool;
}

/*!
 * @brief Allocates a single memory block for the memory pool.
 *
 * !!IMPORTANT!!
 * The caller will hold a mutex lock for the internal lists.  This is a private
 * function that is only used by the singleton object we support.  In all of the
 * places that call this function the mutex locker is managed by that caller.
 *
 * @param memoryBlockId - unique identifier used for this memory block.
 * @param memoryBlockSize - the size of this memory block.
 *
 * @returns True if the block was created and added to our list.
 * Modifies memoryBlockList, memoryBlocksFree, and memoryBlocksAllocated.
 */
bool FiMemoryPool::AllocBlock(uint32_t memoryBlockId, uint32_t memoryBlockSize)
{
    bool blockCreated           = false;
    QString myLabelStr      = QString("%1[AllocBlock]").arg(memPoolLabelStr);

    // Allocate the image data object.
    try
    {
        std::unique_ptr<FI_MEMPOOL::FiMemoryBlock> memoryBlockPtr;

        memoryBlockPtr = FiMemoryBlock::CreateMemoryBlock((int)memoryBlockId, memoryBlockSize);
        if (memoryBlockPtr)
        {
            // Does the list have a spot for this memory block?
            if (memoryBlockList.size() > memoryBlockId)
            {
                memoryBlockList[(unsigned long long)memoryBlockId] = std::move(memoryBlockPtr);
            }
            else
            {
                memoryBlockList.push_back(std::move(memoryBlockPtr));
            }

            memoryBlocksFree.push_back((int)memoryBlockId);
            memoryBlocksAllocated++;
            blockCreated = true;
            FiMemoryPool::LogVerbose((QString("[ALLOCATE MEMORY BLOCK][memoryBlockId=%1][#Alloc=%2][ListSize=%3][blocksize=%4 , 0x%5]").arg(memoryBlockId).arg(memoryBlocksAllocated).arg(memoryBlockList.size()).arg(memoryBlockSize).arg(memoryBlockSize, 8, 16, QLatin1Char('0'))).toUtf8().constData());
        }
        else
        {
            FiMemoryPool::LogVerbose((QString("%1 Unable to allocate block of memory for pool. [memoryBlockId=%2][memoryBlockSize=%3]").
                                      arg(myLabelStr).arg(memoryBlockId).arg(memoryBlockSize)).toUtf8().constData(), VLOG_SHOW_ERROR);
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << myLabelStr << exceptWarnWhatMsg << e.what();
    }
    catch (...)
    {
        QLOG_DEBUG() << myLabelStr << exceptionWarnMsg ;
    }

    return blockCreated;
}

/*!
 * @brief Allocates the image pool.
 *
 * @param numberOfBuffersToAllocate - contains the size of the memory pool to create.
 * @param maxImageSizeToSet - contains the maximum size of the FiMemData buffers.
 * @param lowWaterMarkToSet - contains the point at which the image pool goes
 * into a low-water condition (running low on free image buffers in the pool).
 *
 * @returns The state of the memory pool allocation operation.
 */
bool FiMemoryPool::AllocatePool(uint32_t numberOfBuffersToAllocate,
                                uint32_t maxBufferSizeToSet,
                                uint32_t lowWaterMarkToSet)
{
    bool allocatedPool      = false;
    QString myLabelStr  = QString("%1[AllocatePool]").arg(memPoolLabelStr);
    uint32_t memoryBlockId;
    uint32_t resumePoint;

    // Make sure there is a singleton memory pool object.
    if (fiMemoryPoolPtr == nullptr)
    {
        QLOG_DEBUG() << myLabelStr << memPoolWarnMsg << "fiMemoryPoolPtr was a nullptr!";
        goto exitAllocPool;
    }

    // Make sure the user requested buffers.
    if (numberOfBuffersToAllocate == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The caller didn't specify any buffers to allocate setting it to MIN_BLOCK_ALLOCATIONS.";
        numberOfBuffersToAllocate = MIN_BLOCK_ALLOCATIONS;
    }

    // Make sure that there is at least a minimum number of FiMemoryBlocks.
    if (numberOfBuffersToAllocate < MIN_BLOCK_ALLOCATIONS)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The caller wanted less FiMemoryBlocks than MIN_BLOCK_ALLOCATIONS.";
        numberOfBuffersToAllocate = MIN_BLOCK_ALLOCATIONS;
    }

    // Make sure the pool request paramaters are valid.
    if (lowWaterMarkToSet >= numberOfBuffersToAllocate)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The number of image pool entries to allocate [" << numberOfBuffersToAllocate << "] should be greater than the backoff:" << lowWaterMarkToSet;
        goto exitAllocPool;
    }

    // Persist the low water mark percentage so that the low water mark can be
    // monitored when poolsize is grown or shrunk.
    lowWaterMarkPct = (float)lowWaterMarkToSet / (float)numberOfBuffersToAllocate;

    // Allocate the pool in this scope
    {
        // Lock the entire thing down while building up the memory pool.
        QMutexLocker memoryBlockLocker(&memoryBlockMutex);

        // Set the maximum size of the image data.
        maxMemBufferSize = maxBufferSizeToSet;

        // Make sure the memory block id free and consumed lists are empty.
        memoryBlocksFree.clear();

        // Allocate our buffer(s) and build up the queues.
        for (memoryBlockId = 0; memoryBlockId < numberOfBuffersToAllocate; memoryBlockId++)
        {
            // Keep flags each time through this loop to understand what is created.
            bool addedMemoryBlockThisPass = AllocBlock(memoryBlockId, maxMemBufferSize);

            // Cut short the allocation effort if we cannot get an image data record.
            if (!addedMemoryBlockThisPass)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "AllocBlock error.  Reducing numberOfMemBuffers to:" << fiMemoryPoolPtr->memoryBlocksFree.size() << " Requested: " << numberOfBuffersToAllocate;

                // Terminate the loop.
                break;
            }
        }

        // Set up the activeMemoryBlockId to be the top spot on the list.
        if (fiMemoryPoolPtr->memoryBlocksFree.size() > 0)
        {
            fiMemoryPoolPtr->activeMemoryBlockId = fiMemoryPoolPtr->memoryBlocksFree.front();
            fiMemoryPoolPtr->memoryBlocksFree.pop_front();
        }
        else
        {
            fiMemoryPoolPtr->activeMemoryBlockId = NO_MEMORY_BLOCK;
        }

        allocatedPool = (fiMemoryPoolPtr->memoryBlocksAllocated > 0) ? true : false;

        // Regenerate low water mark and pull the resume point.
        fiMemoryPoolPtr->CalculateResumePoint(resumePoint);

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "numberOfMemBuffers:" << fiMemoryPoolPtr->memoryBlocksAllocated << "backoff:"
                     << lowWaterMark << "resume:" << resumePoint;
    }
exitAllocPool:

    return allocatedPool;
}

