/*!
 * @file fi_mem_pool.h
 *
 * @brief A memory pooling object to manage memory buffers for the
 * system and generate resources to manage data in those buffers.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_MEMORY_POOL_H
#define FI_MEMORY_POOL_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <memory>
#include <vector>
#include <deque>
#include <QMutex>
#include <QString>

namespace FI_MEMPOOL
{
//--------------------------------------------------------------------
// Constants, Typedefs and Macros
//--------------------------------------------------------------------

//! Verbose Logging finer-tune control.
const uint32_t VLOG_MAP_ALLOCS          = 1;
const uint32_t VLOG_MAP_RELEASE         = 2;
const uint32_t VLOG_BLOCK_RELEASE       = 4;
const uint32_t VLOG_SHOW_ALWAYS         = 8;
const uint32_t VLOG_SHOW_ERROR          = 16;

//! Modify this setting to be a combination of the type of logging to report when verbose
//! logging.  The error conditions will always report no matter what but there is some
//! level of control over the other settings.
const uint32_t VLOG_TO_PRINT            = VLOG_BLOCK_RELEASE;

//! General warning strings for log messages dealing with FI_MEMPOOL.
const QString memPoolWarnMsg            = QStringLiteral("!!WARNING!!");
const QString memPoolLabelStr           = QStringLiteral("[MEMPOOL]");
const QString exceptWarnWhatMsg         = QStringLiteral("Exception what(e):");
const QString exceptionWarnMsg          = QStringLiteral("Exception(...)");

//! When a memory block id is invalid it will be set to NO_MEMORY_BLOCK.
const int NO_MEMORY_BLOCK               = -1;
const int NO_MEMORY_MAP                 = -1;

//! Define a maximum size for our memory block.
//! Currently this is using the old setting from gige_socket.cpp (GiGESocket::MAX_IMAGE_size).
const uint32_t MAX_MEMORY_BLOCKSIZE     = 0x1410000;     // We are using 20Mb as the maximum data to be sent. Add something in for the overhead.

//--------------------------------------------------------------------
// Class Declarations
//--------------------------------------------------------------------

class FiMemoryResource
{
public:
    FiMemoryResource(int poolBlockIdToSet,
                     int blockMapIdToSet);
    virtual ~FiMemoryResource();

    // Interface available for the external world.
    uint32_t AppendBytes(uint32_t appendSize, const uint8_t *bytesToAppend);
    uint8_t *AppendBytesExternally(uint32_t sizeRequested, uint32_t& bytesAvailable);
    bool CompleteExternalAppend(uint32_t appendByteSize);

    uint32_t GetCapacity();
    uint32_t GetLength();
    const uint8_t *GetReservedPtr(uint32_t& dataLength);

    void LogMyId(const QString *moreInfoStr = nullptr);

    uint32_t ReplaceBytes(uint32_t memBytesToSetLen, const uint8_t *memBytesToSet);

private:
    FiMemoryResource()
    {
        poolBlockId     = NO_MEMORY_BLOCK;
        blockMapId      = NO_MEMORY_MAP;
    }
    int poolBlockId;            //!< The id of the block of memory in the memory pool.
    int blockMapId;             //!< The id of memory map in the block of memory.

    //! The FiMemoryResource custom deleter calls this to handle resource recycling.
    static void RecycleResource(FiMemoryResource *memoryResourcePtr);

    //! The FiMemoryPool object is a friend and and manages FiMemoryResource.
    friend class FiMemoryPool;  //!< The memory pool object is a friend.

    //! TheFiMemoryResourceDeleter is where the custom deleter for FiMemoryResource
    //! handles the call to our private RecycleResource method and must be declared
    //! as a friend to us and us to them.
    friend struct FiMemoryResourceDeleter;
};

//! This structure is what is used to facilitate the custom deleter for FiMemoryResource
//! objects when they are created and this structure provides the operation for the
//! custom deleter to ensure resources are always recycled properly as they move through
//! the system.
struct FiMemoryResourceDeleter
{
    void operator()(FiMemoryResource *memoryResourcePtr) const
    {
        if (memoryResourcePtr)
        {
            // Call the recycler to give the memory resource a chance
            // to do anything necessary before the deletion occurs on
            // this resource object.
            FiMemoryResource::RecycleResource(memoryResourcePtr);

            // Perform the delete call.
            delete memoryResourcePtr;
        }
    }
};

//! The definition of a memory resource pointer for use in the system to make
//! it easier to set up declarations using this data type.  The important part
//! of this is the insertion of the custom deleter "FiMemoryResourceDeleter"
//! adding this to the template is what glues the custom deleter to the type.
using FiMemoryResourcePtr = std::unique_ptr<FI_MEMPOOL::FiMemoryResource, FiMemoryResourceDeleter>;

// Provide a forward declaration of the FiMemoryBlock because of circular reference
// issues when they two headers are including each other.
class FiMemoryBlock;
using FiMemBlockPtr = FiMemoryBlock*;

//! The FiMemoryPool class manages the list of FiMemData objects available in the
//! system's image pool.
class FiMemoryPool
{
public:
    virtual ~FiMemoryPool();

    static int AvailableCount();

    static bool CreateMemoryPool(
        uint32_t numberOfBuffers,
        uint32_t maxBufferSize,
        uint32_t lowWaterMarkToSet);

    static void DeleteMemoryPool();

    static FiMemoryResourcePtr GetMemoryResource(uint32_t requestedSize, bool okToEmergencyAllocateIfEmpty);
    static void RecycleMemoryBlock(int memoryBlockId);

    static bool HasAMemoryPool();

    static bool IsPoolLow();
    static void LogVerbose(const char *msg, uint32_t vlogFlag = VLOG_TO_PRINT);

    static bool ResizeMemoryPool(uint32_t numberOfBuffers);

private:
    // This is a singleton object.
    FiMemoryPool();

    // This object is non-copyable and non-copy assignable.
    FiMemoryPool(const FiMemoryPool& i) = delete;
    FiMemoryPool& operator=(const FiMemoryPool& i) = delete;

    bool AllocBlock(uint32_t memoryBlockId, uint32_t memoryBlockSize);
    bool AllocatePool(uint32_t numberOfBuffers,
                      uint32_t maxBufferSize,
                      uint32_t lowWaterMarkToSet);

    bool GrowMemoryPool(uint32_t memoryPoolNewSize);
    bool ShrinkMemoryPool(uint32_t memoryPoolNewSize);
    bool TakeOffToBeDeleted(uint32_t memoryBlockId);
    bool TakeOffMemoryBlockFreeList(uint32_t memoryBlockId);
    bool ResetMemoryBlock(uint32_t memoryBlockId);

    bool CalculateResumePoint(uint32_t& resumePoint);

    static FiMemBlockPtr GetMemoryBlockPtr(int memoryBlockId);

    bool ReserveMemoryMap(uint32_t requestedSize, int& memoryBlockId, int& memoryMapId);

    uint32_t    emergencyAllocations;   //!< How many emergency allocations have occurred.

    uint32_t    lowWaterMark;           //!< A low-water mark for free image buffers available.
    float       lowWaterMarkPct;        //!< At pool creation time what is the low water percentage.
    uint32_t    maxMemBufferSize;       //!< Maximum size of an FiMemData buffer.

    //! Identifies which block of memory is serving up memory resources.
    int activeMemoryBlockId;

    //! A list of memory blocks that are available for use.  It is a FIFO
    //! list of zero-based index values into the memDataBlockList queue.
    std::deque<int> memoryBlocksFree;

    //! A list of memory blocks that are earmarked for destruction when
    //! they are recycled.  This list gets populated with any memory blocks
    //! that have reserved resources in use when shrinking the memory pool.
    //! It is only then that a memory block's id will be placed on this list
    //! otherwise the block was already reset and removed from memoryBlocksFree
    //! when the size of the memory pool was reduced.  When these memory blocks
    //! end up back to the pool for recycling (when all their resources are no
    //! longer in use) OR if the memory pool is grown again that these memory block
    //! ids will be removed and their respective allocated blocks deleted or recycled.
    std::deque<int> memoryBlocksToDelete;

    //! The non-volitale memory block list containing the data comprising
    //! the memory pool's memory blocks.  The order never changes and the
    //! availability of this block list is managed by its zero-based id
    //! in the list maintained in the FIFO list of free memory blocks.  If it is
    //! grown or shrunk that happens strictly to the end range of blocks.
    std::vector<std::unique_ptr<FI_MEMPOOL::FiMemoryBlock>> memoryBlockList;

    //! The size of the memoryBlockList in terms of allocated FiMemoryBlock
    //! pointers.  We will keep the size of memoryBlockList as it grows but
    //! when the memory pool is shrunk by system user the actual FiMemoryBlock
    //! pointers at the end will be "reset" thereby setting them to nullptrs.
    //! It is important that the memoryBlockList.size() is never used to determine
    //! the size of the pool except in cases where the physical size of the list
    //! is needed for growing or shrinking or recycling blocks.  This data setting
    //! will always be the accurate one for the number of blocks in the pool that
    //! have memory blocks for resource allocations.
    uint32_t memoryBlocksAllocated;

    //!< Protects threadsafe access to memoryBlockList, memoryBlocksFree and activeMemoryBlockId.
    QMutex  memoryBlockMutex;

    //! The FiMemoryPool object is the single instance created to manage an
    //! image pool for the system.  It is stored in this static instance.
    static std::unique_ptr<FI_MEMPOOL::FiMemoryPool> fiMemoryPoolPtr;

    //! The FiMemoryResource object is a friend and is managed by the FiMemoryPool.
    friend class FiMemoryResource;
};

} /* End of namespace FI_MEMPOOL */

#endif // FI_MEMORY_POOL_H
