/*!
 * @file fi_metamap.cpp
 * @brief This is the header file for the fiMetaMap class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QByteArray>
#include <QJsonArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QString>
#include <QStringLiteral>

#include "client_math.h"
#include "eventfilter.h"
#include "fi_saver.h"
#include "fi_metamap.h"
#include "logger.h"
#include "mainwindow.h"

namespace FI_METAMAP
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------


}   // End of namespace FI_METAMAP.

using namespace FI_METAMAP;

//--------------------------------------------------------------------
// Private static data members.
//--------------------------------------------------------------------

//! The "name" for the metamap.txt file that houses CAP stack capture information.
const QString FiMetaMap::METAMAP_FILE = QStringLiteral("metamap.txt");

//! The metamap.txt key "psize" represents the size of partition used for FFLR capture.
const QString FiMetaMap::MM_PARTSIZE      = QStringLiteral("psize");

//! The metamap.txt key "recmode" represents FFLR Capture recording type.
const QString FiMetaMap::MM_RECMODE         = QStringLiteral("recmode");
const QString FiMetaMap::MM_RECMODE_IMEM    = QStringLiteral("imem");
const QString FiMetaMap::MM_RECMODE_BASIC   = QStringLiteral("basic");
const QString FiMetaMap::MM_RECMODE_ROC     = QStringLiteral("roc");
const QString FiMetaMap::MM_RECMODE_BROC    = QStringLiteral("broc");

//! The metamap.txt key "start_cap" represents number for first CAP file in stack.
const QString FiMetaMap::MM_START_CAP     = QStringLiteral("start_cap");

//! The metamap.txt key "end_cap" represents number for last CAP file in stack.
const QString FiMetaMap::MM_END_CAP       = QStringLiteral("end_cap");

//! The metamap.txt key "blen" if set represents the BROC FFLR recording burst length.
const QString FiMetaMap::MM_BLEN          = QStringLiteral("blen");

//! The metamap.txt key "trigger_cap" represents number for CAP containing trigger frame.
//! This is only set for basic FFLR capture mode.
const QString FiMetaMap::MM_TRIGGER_CAP   = QStringLiteral("trigger_cap");

//! The metamap.txt key "tframes" represents total number of frames for FFLR capture
const QString FiMetaMap::MM_TFRAMES       = QStringLiteral("tframes");

//! The metamap.txt key "map" represents the mapping of each cap file in a cap stack.
const QString FiMetaMap::MM_MAP             = QStringLiteral("map");
const QString FiMetaMap::MM_MAP_EFRAME      = QStringLiteral("eframe");
const QString FiMetaMap::MM_MAP_SFRAME      = QStringLiteral("sframe");
const QString FiMetaMap::MM_MAP_TFRAME      = QStringLiteral("tframe");
const QString FiMetaMap::MM_MAP_NAME        = QStringLiteral("name");
const QString FiMetaMap::MM_MAP_FCOUNT      = QStringLiteral("fcount");

//! The maximum size of the PFMD LRU cache.
const int CAP_CACHE_MAX_SIZE = 10;

//--------------------------------------------------------------------
// Internal Class Implementation for CapCache
//--------------------------------------------------------------------

/*!
 * @brief The CapCache class is used to have an Least Recently Used (LRU) cache for the
 *          fiRawFile CAP files that are opened and being processed for playback.
 *          This was blantantly taken from the LRU algorithm set for pfmd of CameraMDI.
 */
class CapLRUCache
{
private:
    int max_cache_size;         //!< The maximum cache size, configured at cache creation time.

    QList<uint32_t> cap_lru;   //!< The ordered list of cap file ids. The most used cap file is at the front, least at the back.
    QHash<uint32_t, QSharedPointer<fiRawFile>>   cap_hash;  //!< Associating the cap_id to the fiRawFile information.

public:
    CapLRUCache(int capacity) { max_cache_size = capacity; }
    ~CapLRUCache()            { clear();  }

    /*!
     * @brief Removes all elements from the cache (which are destroyed).
     */
    void clear()
    {
        // Remove all the entries from the hash table.
        QList<uint32_t>::iterator i;
        for (i = cap_lru.begin(); i != cap_lru.end(); ++i)
        {
            QSharedPointer<fiRawFile> oldCapRawFile = cap_hash.take(*i);
        }

        // Reset our list.
        cap_lru.clear();
    }

    /*!
     * @brief Add the FiRawFile pointer to the cache if it is new (and there is room). Move the
     *        entry to the front if it already exists. If there is no room, delete the
     *        "oldest" value, and add the new value.
     *
     * @param frame_num     The frame number associated with the PFMD.
     * @param pfmd_ptr      The pointer to the QBtyeArray containing the PFMD.
     */
    void insert(uint32_t capId, QSharedPointer<fiRawFile>capRawFile)
    {
        int data_index;

        // Does it already exist?
        data_index = cap_lru.indexOf(capId);
        if (data_index < 0)
        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_num:" << frame_num << "Not found. List size:" << pfmd_lru.size() << "max_cache_size:" << max_cache_size;

            // It does not exist. Do we have room?
            if (((int)cap_lru.size()) == max_cache_size)
            {
                int last_index = cap_lru.takeLast();
                QSharedPointer<fiRawFile> oldCapRawFile = cap_hash.take(last_index);
                // We don't have to delete it as it's a shared pointer and it will go away on its own.
            }

            // Add our data.
            cap_hash.insert(capId, capRawFile);
            cap_lru.prepend(capId);
        }
        else if (data_index != 0)
        {
            // Move from where it is to the front of the list.
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Move frame_num:" << frame_num << "from data_index:" << data_index;
            cap_lru.move(data_index, 0);
        }
//        else
//        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_num:" << frame_num << "is at the front of the list";
//        }

//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "LRU:" << pfmd_lru;
    }

    /*!
     * @brief Return the PFMD associated with the supplied frame number, frame_num, if it exists
     *        in the hash.
     *
     * If the frame number is found, then not only is the associated PFMD data retruned, but the
     * frame is moved to the front of the LRU queue.
     *
     * @param[in] capId     The frame number we are interested in.
     *
     * @return The QByteArray pointer to the PFMD, or nullptr if not found.
     */
    QSharedPointer<fiRawFile> retrieve(uint32_t capId)
    {
        int data_index;
        QSharedPointer<fiRawFile> capRawFile;

        // Does it already exist?
        data_index = cap_lru.indexOf(capId);
        if (data_index >= 0)
        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Found frame_num:" << frame_num << "at:" << data_index;
            capRawFile = cap_hash.value(capId);

            // Update our LRU cache.
            insert(capId, capRawFile);
        }
//        else
//        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find frame_num:" << frame_num;
//        }

        // Return our capRawFile (could be null).
        return capRawFile;
    }
};

//--------------------------------------------------------------------
// Internal Class Implementation for CapMapRecord
//--------------------------------------------------------------------

/*!
 * @brief Factory method to create a Cap file record object.
 */
QSharedPointer<CapMapRecord> CapMapRecord::CreateCapMapRecord(uint32_t capIdToSet,
                                                              uint32_t sFrameToSet,
                                                              uint32_t eFrameToSet,
                                                              uint32_t tFrameToSet,
                                                              uint32_t frameCountToSet,
                                                              bool tFrameValidToSet,
                                                              const QString& locationDir)
{
    bool createdCapRec = false;

    // Used to query the pfmd file for starting and ending frame information.
    QFile capPMFile;
    uint32_t eFramePFMDOffset;  //!< Ending frame offset into pfmd ".pm" file.

    // Ask our math class what size a pfmd record is for non-series3 camera.
    uint32_t pfmdRecSize = ClientMath::get_pfmd_size(false);

    // Make a buffer that can hold one pfmd record.
    QVector<uint32_t> pfmdRecord(pfmdRecSize / sizeof(uint32_t));

    QSharedPointer<CapMapRecord> capRecord = QSharedPointer<CapMapRecord>(new CapMapRecord());

    // Was the cap record created ok?
    if (capRecord.isNull())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to create a new CapMapRecord for" << capRecord->capPath;
        goto createCAPMapRecordExit;
    }

    // Populate the cap record.
    capRecord->capId            = capIdToSet;
    capRecord->sFrameId         = sFrameToSet;
    capRecord->eFrameId         = eFrameToSet;
    capRecord->tFrameId         = tFrameToSet;
    capRecord->frameCount       = frameCountToSet;
    capRecord->tFrameValid      = tFrameValidToSet;

    // Validate that the CAP file does exist.
    if (!CapMapRecord::CreateCapPath(locationDir, capIdToSet, capRecord->capPath))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to verify existence of" << capRecord->capPath;
        goto createCAPMapRecordExit;
    }

    // Generate the path to the .PM file (the pfmd (Per-Frame_metadata) file) which should exist
    // when working with FasFireLR.
    if (!CapMapRecord::CreateCapPMPath(locationDir, capIdToSet, capRecord->capPFMDPath))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to verify existence of" << capRecord->capPFMDPath;
        goto createCAPMapRecordExit;
    }

    // In order to help facilitate time searches when dealing with a CAP stack we will store the pfmd
    // sbs information for the starting and ending frames.
    capPMFile.setFileName(capRecord->capPFMDPath);

    // What is our ending frame's zero-based location in the file.
    eFramePFMDOffset = (capRecord->eFrameId - capRecord->sFrameId) * pfmdRecSize;

    // Open our file the starting frame is always at position zero.
    if (!capPMFile.open(QIODevice::ReadOnly))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to open " << capRecord->capPFMDPath << "to read pfmd for starting frame" << capRecord->sFrameId << "!";
        goto createCAPMapRecordExit;
    }

    // Seek to our first record (the starting one).
    if (!capPMFile.seek(0))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to seek " << capRecord->capPFMDPath << "to read pfmd for starting frame" << capRecord->sFrameId << "!";
        goto createCAPMapRecordExit;
    }

    // Read the starting frame's metadata.
    if (capPMFile.read((char *)pfmdRecord.data(), pfmdRecSize) < 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read ending frame" << capRecord->eFrameId << "pfmd (per-frame_metadata) record in" << capRecord->capPFMDPath;
        goto createCAPMapRecordExit;
    }

    // Save the time information for our starting frame.
    capRecord->sFrameSBS = pfmdRecord[0];
    capRecord->sFrameSBSUSec = (pfmdRecord[1] & 0xFFFFF);

    // Now get the ending time information for our last frame.
    if (!capPMFile.seek(eFramePFMDOffset))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to seek " << capRecord->capPFMDPath << "to read pfmd for end frame" << capRecord->eFrameId << "!";
        goto createCAPMapRecordExit;
    }

    // Read the ending frame's metadata.
    if (capPMFile.read((char *)pfmdRecord.data(), pfmdRecSize) < 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read ending frame" << capRecord->eFrameId << "pfmd (per-frame_metadata) record in" << capRecord->capPFMDPath;
        goto createCAPMapRecordExit;
    }

    // Save the time information for our ending frame.
    capRecord->eFrameSBS = pfmdRecord[0];
    capRecord->eFrameSBSUSec = (pfmdRecord[1] & 0xFFFFF);

    // Made it here it was created without an error condition.
    createdCapRec = true;

createCAPMapRecordExit:

    // If didn't fully create the CAP record reset it as an error occurred.
    if (!createdCapRec && !capRecord.isNull())
    {
        capRecord.reset();
    }

    // Close the pfmd file if it was opened.
    if (capPMFile.isOpen())
    {
        capPMFile.close();
    }

    return capRecord;
}

/*!
 * @brief Factory method to create the full path for a cap file.
 *
 * @param[in] locationDir       Where is the full path to the location for the CAP file.
 * @param[in] capId             Zero-based cap numeric Id used to build the cap filename.
 * @param[out] capFullPathName  The fully formed CAP path to use for opening a CAP file.
 *
 * @returns True if file exists, false if the cap file does not exist.
 */
bool CapMapRecord::CreateCapPath(const QString& locationDir, uint32_t capId, QString& capFullPathName)
{
    bool foundIt = false;
    QString capName = QString::fromUtf8("%1.%2").arg(capId, FI_SAVER::IMAGE_SEQ_NUM_LEN, 10, QLatin1Char('0')).arg(FiSaver::IMAGE_EXT_CAPTURE);

    // Install the location directory into the pathToCapCreated.
    capFullPathName = locationDir;

    // Make sure our location directory ends with the directory separator.
    if (!capFullPathName.endsWith(QLatin1Char('/')))
        capFullPathName.append(QLatin1Char('/'));

    // Build up our CAP path.
    capFullPathName.append(capName);

    // Use the QFileInfo to query information about the cap file we have.
    QFileInfo       capFileInfo;

    // If allocated determine if the file exists.
    capFileInfo.setFile(capFullPathName);

    // Does it exist?
    foundIt = capFileInfo.exists();

    // Make the CAP name.
    return foundIt;
}

/*!
 * @brief Factory method to create the full path for a cap's .PM (per-frame-metadata) file.
 *
 * @param[in] locationDir           Where is the full path to the location for the CAP file.
 * @param[in] capId                 Zero-based cap numeric Id used to build the cap filename.
 * @param[out] capPMFullPathName    The fully formed CAP path to use for opening a CAP file's .pm file.
 *
 * @returns True if file exists, false if the cap file does not exist.
 */
bool CapMapRecord::CreateCapPMPath(const QString& locationDir, uint32_t capId, QString& capPMFullPathName)
{
    bool foundIt = false;
    QString capName = QString::fromUtf8("%1.%2").arg(capId, FI_SAVER::IMAGE_SEQ_NUM_LEN, 10, QLatin1Char('0')).arg(FiSaver::IMAGE_EXT_META_RAW);

    // Install the location directory into the pathToCapCreated.
    capPMFullPathName = locationDir;

    // Make sure our location directory ends with the directory separator.
    if (!capPMFullPathName.endsWith(QLatin1Char('/')))
        capPMFullPathName.append(QLatin1Char('/'));

    // Build up our CAP path.
    capPMFullPathName.append(capName);

    // Use the QFileInfo to query information about the cap file we have.
    QFileInfo capPMFileInfo;

    // If allocated determine if the file exists.
    capPMFileInfo.setFile(capPMFullPathName);

    // Does it exist?
    foundIt = capPMFileInfo.exists();

    // Successfully made the full path to the CAP file's .pm (per-frame-metadata) filename.
    return foundIt;
}

//--------------------------------------------------------------------
// Class Implementation for FiMetaMap
//--------------------------------------------------------------------

/*!
 * @brief Constructor for our metadata map object.
 */
FiMetaMap::FiMetaMap()
{
    recType = FFLR_REC_TYPE::FFLR_IMEM;

    partSize = 0;
    startCapId = 0;
    endCapId = 0;
    brocLen = 0;
    trigCap = 0;
    totalFrames = 0;

    hasBrocLen = false;
    hasTrigCap = false;
}

/*!
 * @brief Destructor for our metadata map object.
 */
FiMetaMap::~FiMetaMap()
{
    // All internal data is self-managing upon destruction at this time I believe...
}

/*!
 * @brief Factory method to create an FiMetaMap object.
 *
 * @param metaMapDir    The full path to the directory where the metamap.txt file should reside.
 *
 * @returns an FiMetaMap allocated unique_ptr for the metamap.txt file if it exists, an empty
 * pointer if it does not exist.
 */
QSharedPointer<FI_METAMAP::FiMetaMap> FiMetaMap::CreateMetaMap(const QString& mmPath)
{
    QSharedPointer<FI_METAMAP::FiMetaMap> metaMapPtr;
    QByteArray      mmContents;         // The metamap file's contents.
    QJsonDocument   mmDoc;              // The metamap file as JSON document.
    QJsonObject     mmObj;              // The metamap document as an object.
    QJsonParseError jpError;            // Any JSON parse error.
    QFileInfo       mmFileInfo;         // The metamap file information object.
    QFile           mmFile;             // The metamap file object.
    QString         mmFileName;         // The full path to the metamap.txt file.
    QString         mmRecType;          // How was this cap stack recorded?

    bool            createdMetaMap      = false; // Did we make it through a full factory process?

    int             brocLenValidator    = -1;
    int             trigCapIdValidator  = -1;

    // Allocate an FiMetaMap object and use it to populate its contents.
    metaMapPtr = QSharedPointer<FI_METAMAP::FiMetaMap>(new FI_METAMAP::FiMetaMap());

    // If we were unable to allocate the metamap object output a warning in the log and exit.
    if (metaMapPtr.isNull())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to create the FiMetaMap for" << mmPath << "!";
        goto CreateMetaMapExit;
    }

    // Build up the metamap filename.
    metaMapPtr->mmFilePath = mmPath;

    // Make sure our location directory ends with the directory separator.
    if (!metaMapPtr->mmFilePath.endsWith(QLatin1Char('/')))
        metaMapPtr->mmFilePath.append(QLatin1Char('/'));

    // Create the full filename for the metamap.txt file.
    mmFileName = metaMapPtr->mmFilePath + FiMetaMap::METAMAP_FILE;

    // If allocated determine if the file exists.
    mmFileInfo.setFile(mmFileName);

    // Does it exist?
    if (!mmFileInfo.exists())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("%1 does not exist!").arg(mmFileName);
        goto CreateMetaMapExit;
    }

    // It exists so set up the metamapFile object to be able to read and JSON parse it.
    mmFile.setFileName(mmFileName);

    // Open the file it should be a JSON data file it needs to be interpreted and set here.
    mmFile.open(QIODevice::ReadOnly | QIODevice::Text);

    // Read the file into our buffer.
    mmContents = mmFile.readAll();

    // Turn the buffer read into a QJsonDocument.
    mmDoc = QJsonDocument::fromJson(mmContents, &jpError);
    if (mmDoc.isNull())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("%1 has JSON Parse Error:%2!").arg(mmFileName, jpError.errorString());
        goto CreateMetaMapExit;
    }

    // Interpret the contents of the JSON document setting up the state of the
    // manufactured FiMetaMap object.
    mmObj = mmDoc.object();

    metaMapPtr->partSize    = mmObj[MM_PARTSIZE].toInt();
    metaMapPtr->startCapId  = mmObj[MM_START_CAP].toInt();
    metaMapPtr->endCapId    = mmObj[MM_END_CAP].toInt();
    brocLenValidator        = mmObj[MM_BLEN].toInt(-1);
    trigCapIdValidator      = mmObj[MM_TRIGGER_CAP].toInt(-1);
    metaMapPtr->totalFrames = mmObj[MM_TFRAMES].toInt();
    metaMapPtr->brocLen     = mmObj[MM_BLEN].toInt();
    metaMapPtr->trigCap     = mmObj[MM_TRIGGER_CAP].toInt();

    mmRecType               = mmObj[MM_RECMODE].toString("imem");

    // Set the record type for this CAP file.
    if (mmRecType == MM_RECMODE_BASIC)
    {
        metaMapPtr->recType = FI_METAMAP::FFLR_REC_TYPE::FFLR_BASIC;

        // This is expecting a trigger frame, validate it.
        if (trigCapIdValidator == -1)
        {
            // Ticket #519: If there is a missing trigger frame, make the last frame in the last cap the trigger.
            metaMapPtr->trigCap = metaMapPtr->endCapId;
        }
        else
        {
            metaMapPtr->trigCap = trigCapIdValidator;
        }
        metaMapPtr->hasTrigCap = true;
    }
    else if (mmRecType == MM_RECMODE_BROC)
    {
        metaMapPtr->recType = FI_METAMAP::FFLR_REC_TYPE::FFLR_BROC;
        // This is expecting a broc burst length, validate it.
        if (brocLenValidator == -1)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << QString("%1 is missing %2 for a BROC FasFireLR recording!").arg(mmFileName, MM_BLEN);
            goto CreateMetaMapExit;
        }
        metaMapPtr->brocLen = brocLenValidator;
        metaMapPtr->hasBrocLen = true;
    }
    else if (mmRecType == MM_RECMODE_ROC)
    {
        metaMapPtr->recType = FI_METAMAP::FFLR_REC_TYPE::FFLR_ROC;
    }
    // Default always to image memory.
    else
    {
        metaMapPtr->recType = FI_METAMAP::FFLR_REC_TYPE::FFLR_IMEM;
    }

    // Now build up the map of CAP data.
    if (mmObj.contains(MM_MAP) && mmObj[MM_MAP].isArray())
    {
        QJsonArray mapArray = mmObj[MM_MAP].toArray();

        // Loop through the map, adding map records.
        for (uint32_t mapItem = 0; mapItem < mapArray.size(); ++mapItem)
        {
            int fieldValidator;
            QJsonObject mapObject = mapArray[mapItem].toObject();

            // Pull out the map data.
            uint32_t sFrame = mapObject[MM_MAP_SFRAME].toInt(0);
            uint32_t eFrame = mapObject[MM_MAP_EFRAME].toInt(0);
            uint32_t tFrame = 0;
            bool tFrameValid = false;
            fieldValidator = mapObject[MM_MAP_TFRAME].toInt(-1);
            if (fieldValidator >= 0)
            {
                tFrame = fieldValidator;
                tFrameValid = true;
            }

            uint32_t capId  = mapObject[MM_MAP_NAME].toInt(0);
            uint32_t fCount = mapObject[MM_MAP_FCOUNT].toInt(0);

            // Ticket #519 if there is no trigger frame make sure we make the last one the trigger frame.
            if ((trigCapIdValidator == -1) && (capId == metaMapPtr->trigCap) && !tFrameValid)
            {
                tFrame = eFrame;
                tFrameValid = true;
                QLOG_WARN() << __PRETTY_FUNCTION__ << QString("%1 was missing %2 for a Basic FasFireLR recording, using %3: %4 frame: %5").arg(mmFileName).arg(MM_TRIGGER_CAP).arg(MM_END_CAP).arg(metaMapPtr->endCapId).arg(tFrame);
            }

            QSharedPointer<CapMapRecord> mapRecord = CapMapRecord::CreateCapMapRecord(capId, sFrame, eFrame, tFrame, fCount, tFrameValid, mmPath);

            // Was the mapRecord created without any error?
            if (mapRecord.isNull())
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create mapRecord for capId:" << capId;
                goto CreateMetaMapExit;
            }
            metaMapPtr->capRecords.insert(capId, mapRecord);
        }
    }

    // Create the empty LRU Cache for Cap Files.
    metaMapPtr->capLRUFiles = QSharedPointer<CapLRUCache>(new CapLRUCache(CAP_CACHE_MAX_SIZE));

    // Did the LRU cache get created?
    if (metaMapPtr->capLRUFiles.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create metaMapPtr->capLRUFiles!";
        goto CreateMetaMapExit;
    }

    // If make it here the full metamap record was created.
    createdMetaMap = true;

CreateMetaMapExit:
    // Close any metamap file object if it was opened.
    if (mmFile.isOpen())
    {
        mmFile.close();
    }

    // If something went wrong reset any partially created metamap record.
    if (!createdMetaMap)
    {
        if (!metaMapPtr.isNull())
        {
            metaMapPtr.reset();
        }
    }

    return metaMapPtr;
}

/*!
 * @brief What is the full path to this object's metamap.txt file.
 *
 * @param[out] mmFileNameFound  The full path to the metamap.txt for this object or empty if a problem.
 *
 * @returns True if all is successful and the file exists, false if an error occurred or the file doesn't exist.
 */
bool FiMetaMap::GetMMFileName(QString& mmFileNameFound)
{
    bool fileExists = false;
    QFileInfo mmFileInfo;

    // Empty the output value.
    mmFileNameFound.clear();

    // Do we have a metamap in this object?
    if (mmFilePath.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to find metamap.txt filepath!";
        goto GetMMFileNameExit;
    }

    // Install the full path to the metamap.txt file used to create this object.
    mmFileNameFound = mmFilePath + FiMetaMap::METAMAP_FILE;

    // Make sure it exists.
    mmFileInfo.setFile(mmFileNameFound);

    if (!mmFileInfo.exists())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to verify existance of" << mmFileNameFound  << "!";
        goto GetMMFileNameExit;
    }

GetMMFileNameExit:
    return fileExists;
}

/*!
 * @brief Creates the full path to the "pm_metadata" (metadata.txt) for the starting cap file
 * belonging to this CAP stack.
 *
 * @param[in] capId                     Which cap in the stack are we asking about.
 * @param[out] metadataTXTNameFound     The full path to the metadata.txt for capId or empty if not found.
 *
 * @returns True if all is successful and the file exists, false if an error occurred or the file doesn't exist.
 */
bool FiMetaMap::GetMetadataTXTName(uint32_t capId, QString& metadataTXTNameFound)
{
    bool fileExists = false;

    // Empty the output value.
    metadataTXTNameFound.clear();

    // Do we have a metamap in this object?
    if (mmFilePath.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to find metamap.txt filepath!";
        goto GetMetadataTXTNameExit;
    }

    // Make sure there is a cache entry for our starting cap.
    if (!capRecords.contains(capId))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to find pm_metadata for capId:" << capId << "!";
        goto GetMetadataTXTNameExit;
    }
    else
    {
        QFileInfo metadataTXTInfo;
        QSharedPointer<CapMapRecord> capRecord = capRecords.value(capId);

        metadataTXTNameFound = capRecord->capPath;

        // Swap the extensions.
        metadataTXTNameFound.replace(metadataTXTNameFound.lastIndexOf(FiSaver::IMAGE_EXT_CAPTURE), FiSaver::IMAGE_EXT_CAPTURE.length(), FiSaver::IMAGE_EXT_META_TEXT);

        // Make sure it exists.
        metadataTXTInfo.setFile(metadataTXTNameFound);

        if (metadataTXTInfo.exists())
        {
            fileExists = true;
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to verify existance of" << metadataTXTNameFound << "for capId:" << capId << "!";
            goto GetMetadataTXTNameExit;
        }
    }

GetMetadataTXTNameExit:
    return fileExists;
}

/*!
 * @brief Creates the full path to the Per-Frame MetaData (PFMD) file associated with the
 *        specified CAP file ID (capId).
 *
 * @param[in] capId             Which cap in the stack are we asking about.
 * @param[out] pfmdNameFound    The full path to the metadata.pm for capId or empty if not found.
 *
 * @returns True if all is successful and the file exists, false if an error occurred or the file doesn't exist.
 */
bool FiMetaMap::GetPFMDName(uint32_t capId, QString& pfmdNameFound)
{
    bool fileExists = false;

    // Empty the output value.
    pfmdNameFound.clear();

    // Make sure there is a cache entry for our starting cap.
    if (!capRecords.contains(capId))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to find pm_metadata for capId:" << capId << "!";
        goto GetPFMDNameExit;
    }
    else
    {
        QFileInfo pfmdFileInfo;
        QSharedPointer<CapMapRecord> capRecord = capRecords.value(capId);

        pfmdNameFound = capRecord->capPFMDPath;

        // Make sure it exists.
        pfmdFileInfo.setFile(pfmdNameFound);

        if (pfmdFileInfo.exists())
        {
            fileExists = true;
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to verify existance of" << pfmdNameFound << "for capId:" << capId << "!";
            goto GetPFMDNameExit;
        }
    }

GetPFMDNameExit:
    return fileExists;
}

/*!
 * @brief Determines the per-frame-metadata (pfmd) offset into the absolute frame's CAP pm file.
 *
 * @param[in] absFrameNum           The absolute frame number for the CAP stack to locate.
 * @param[out] capIdFound           The capId belonging to the frame number.
 * @param[out] pfmdFileName         The full path to the metadata.pm for capId or empty if not found.
 * @param[out] pfmdOffset           The offset in the metadata.pm file for the absFrameNum from the Cap Stack.
 *
 * @returns True if all is successful and the file exists, false if an error occurred or the file doesn't exist.
 */
bool FiMetaMap::GetPFMDOffset(uint32_t absFrameNum, uint32_t& capIdFound, QString& pfmdFileName, uint32_t& pfmdOffset)
{
    bool foundIt = false;
    pfmdFileName.clear();
    pfmdOffset = 0;

    if (GetCapInfoForFrame(absFrameNum, capIdFound, pfmdOffset))
    {
        // What is the filename for the pfmd file for this cap id?
        if (GetPFMDName(capIdFound, pfmdFileName))
        {
            // Get the cap record so we can calculate the true offset in the file.
            // Make sure it is contained in the record list.
            if (capRecords.contains(capIdFound))
            {
                QSharedPointer<CapMapRecord> capRecord = capRecords.value(capIdFound);

                // Make sure the absolute frame number is within our range.
                if ((absFrameNum >= capRecord->sFrameId) && (absFrameNum <= capRecord->eFrameId))
                {
                    // The true offset in metadata.pm file will be the difference between the absFrameNum and sFrameId.
                    pfmdOffset = absFrameNum - capRecord->sFrameId;
                    foundIt = true;
                }
                else
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "absFrameNumber" << absFrameNum << "for capId:" << capIdFound << "Not in sFrame:" << capRecord->sFrameId << "to eFrame:" << capRecord->eFrameId << "Range!";
                }
            }
        }
    }

    return foundIt;
}

/*!
 * @brief Allocates a large buffer to hold all of the events for the CAP stack.
 *
 * @returns Allocated using the new operator a buffer of uint8_t event data where caller assumes responsibility for
 * cleaning up and deallocating the memory. Note: I would like to see this turned into a QVector<uint8_t> but that
 * requires a modification to the camera's review object.
 */
uint8_t *FiMetaMap::GetPFMDEvents(bool do_swap)
{
    bool readAllEvents = false;     // Keeps track of whether or not this function read all of the event data.

    QFile capPMFile;                // The opened CAP PFMD file.

    uint8_t *pfmdEvents;            // The allocated pfmd event data.
    uint32_t pfmdEventIndex = 0;    // A master index used to fill up the event data read from the CAP files.

    // Allocate this pointer if we return it from this function the receiver MUST ensure it is properly
    // de-allocated and the memory is returned to the system.
    pfmdEvents = new uint8_t[ totalFrames ];
    if (!pfmdEvents)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocated pfmdEvents for frameCount:" << totalFrames;
        return nullptr;
    }

    // Ask our math class what size a pfmd record is for non-series3 camera.
    uint32_t pfmdRecSize = ClientMath::get_pfmd_size(false);

    // Make a buffer that can hold one pfmd record.
    QVector<uint32_t> pfmdRecord(pfmdRecSize / sizeof(uint32_t));

    // Buffer is allocated.  Read through our CAP files and store their event information.
    for (uint32_t capRecId = startCapId; capRecId <= endCapId; ++capRecId)
    {
        if (capRecords.contains(capRecId))
        {
            QSharedPointer<CapMapRecord> capRecord = capRecords.value(capRecId);

            // In order to help facilitate time searches when dealing with a CAP stack we will store the pfmd
            // sbs information for the starting and ending frames.
            capPMFile.setFileName(capRecord->capPFMDPath);

            // Open our file the starting frame is always at position zero.
            if (!capPMFile.open(QIODevice::ReadOnly))
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to open " << capRecord->capPFMDPath << "to read pfmd!";
                goto GetPFMDEventsExit;
            }

            // Seek to our first record (the starting one).
            if (!capPMFile.seek(0))
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to seek to record 0 of " << capRecord->capPFMDPath << "!";
                goto GetPFMDEventsExit;
            }

            // Go through our pfmd records and set up the time information for the beginning and the end and pull out
            // the "events" information for each record during this pass too.
            for (uint32_t frameIndex = 0; frameIndex < capRecord->frameCount; frameIndex++)
            {
                // Read the starting frame's metadata.
                if (capPMFile.read((char *)pfmdRecord.data(), pfmdRecSize) < 0)
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read frame" << (capRecord->sFrameId + frameIndex) << "pfmd (per-frame_metadata) record in" << capRecord->capPFMDPath;
                    goto GetPFMDEventsExit;
                }

                // Add the event information.
                uint32_t flag_hold;
                const uint8_t *data_ptr = (const uint8_t *)pfmdRecord.constData();

                // Copy the event flag information located in the 3rd word (offset=2).
                memcpy(&flag_hold, data_ptr + (2 * sizeof(uint32_t)), sizeof(flag_hold));
                if (do_swap)
                    flag_hold = qToBigEndian(flag_hold);

                // Make sure we haven't exceeded our buffer allocation.
                if (pfmdEventIndex >= totalFrames)
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "At frameIndex:" << frameIndex << "for capRecId:" << capRecId << "exceeded expected totalFrames" << totalFrames << "while processing pfmd (per-frame_metadata) record in" << capRecord->capPFMDPath;
                    goto GetPFMDEventsExit;
                }

                // Save the marker information into our buffer.
                pfmdEvents[ pfmdEventIndex++ ] = (uint8_t)((flag_hold >> FM_EVENT_FILTERS::EVENT_FLAG_OFFSET) & 0xFF);
            }

            // Close this CAP PFMD file.
            capPMFile.close();
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "The capRecId:" << capRecId << "was not found in capRecords!";
            goto GetPFMDEventsExit;
        }
    }

    // If you make it here it all of the events were processed.
    readAllEvents = true;

GetPFMDEventsExit:
    // Is there any left-open file?
    if (capPMFile.isOpen())
    {
        capPMFile.close();
    }

    // Did all of the events get read?
    if (!readAllEvents)
    {
        // Make sure to delete the buffer created as it contains incomplete inforamtion.
        if (pfmdEvents != nullptr)
        {
            delete []pfmdEvents;
            pfmdEvents = nullptr;
        }
    }

    // Return the pfmdEvents.
    return pfmdEvents;
}

/*!
 * @brief Retrieves the full pfmd file data from a CAP file in the stack.
 *
 * @param[in] capId:            Which CAP file are we interested in reading?
 * @param[in] appendFileData:   Does caller want the pfmd data appended to pfmdFileData or to only hold this pfmd data.
 * @param[out] pfmdFileData:    The entire contents of the pfmd ".pm" file in a QVector.
 *
 * @returns True if all is successful, false if an error occurred.
 */
bool FiMetaMap::GetPFMD(uint32_t capId, bool appendFileData, QByteArray& pfmdFileData)
{
    QFile capPMFile;
    bool readPFMDFile = false;

    // Ask our math class what size a pfmd record is for non-series3 camera.
    uint32_t pfmdRecSize = ClientMath::get_pfmd_size(false);

    // If we are not appending data clear the buffer presented by user.
    if (!appendFileData)
    {
        pfmdFileData.clear();
    }

    // Does the capId exist in our capRecords?
    if (capRecords.contains(capId))
    {
        QSharedPointer<CapMapRecord> capRecord = capRecords.value(capId);

        // How big should our resultant buffer be sized?
        uint32_t pfmdFileSize = capRecord->frameCount * pfmdRecSize;

        char *pfmdFileDataPtr = nullptr;

        // Are we appending?
        if (!appendFileData)
        {
            // No the pfmdFileData will only be set to the pfmd for this CAP file.
            pfmdFileData.resize(pfmdFileSize);
            pfmdFileDataPtr = pfmdFileData.data();
        }
        else
        {
            // Appending.
            uint32_t pfmdFileDataCurSize = pfmdFileData.size();
            pfmdFileData.resize((pfmdFileDataCurSize + pfmdFileSize));
            pfmdFileDataPtr = pfmdFileData.data();
            pfmdFileDataPtr += pfmdFileDataCurSize;
        }

        // In order to help facilitate time searches when dealing with a CAP stack we will store the pfmd
        // sbs information for the starting and ending frames.
        capPMFile.setFileName(capRecord->capPFMDPath);

        // Open our file the starting frame is always at position zero.
        if (!capPMFile.open(QIODevice::ReadOnly))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to open " << capRecord->capPFMDPath << "to read pfmd!";
            goto GetPFMDExit;
        }

        // Seek to our first record (the starting one).
        if (!capPMFile.seek(0))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to seek to record 0 of " << capRecord->capPFMDPath << "!";
            goto GetPFMDExit;
        }

        // Read this pfmd file.
        if (capPMFile.read(pfmdFileDataPtr, pfmdFileSize) < 0)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read:" << capRecord->capPFMDPath << "!";
            goto GetPFMDExit;
        }

        // Close this CAP PFMD file.
        capPMFile.close();
    }
    else
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The capRecId:" << capId << "was not found in capRecords!";
        goto GetPFMDExit;
    }

    // If you make it here it all of pfmd was read into the buffer.
    readPFMDFile = true;

GetPFMDExit:

    // Did we leave the file open?
    if (capPMFile.isOpen())
        capPMFile.close();

    return readPFMDFile;
}

/*!
 * @brief Determines the per-frame-metadata (pfmd) offset into the absolute frame's CAP pm file.
 *
 * @param [in] timeInSecs           The time represented in seconds matching pfmd sbs time value.
 * @param [in] timeInUSecs          The time represented in useconds matching pfmd usbs time value.
 * @param [out] capIdFound          The capId belonging to the frame number.
 * @param [out] absFrameNumFound    The absolute frame number in entire video for the time found.
 *
 * @returns True if all is successful, false if an error occurred.
 */
bool FiMetaMap::GetAbsFrameUsingTime(uint32_t timeInSecs, uint32_t timeInUSecs, uint32_t& capIdFound, uint32_t& absFrameNumFound)
{
    bool foundAbsFrameNum   = false;
    bool foundCapId         = false;

    QFile capPMFile;
    QSharedPointer<CapMapRecord> capRecord;

    // Set both outputs to zero for lack of a better value.
    capIdFound = 0;
    absFrameNumFound = 0;

    // Is this time in our cap record array?
    // We may have to optimize the search here if there is a very large number of CAP files in the stack.
    // But with a 4GB partition perhaps this may not be an issue.
    for (uint32_t index = startCapId; !foundCapId && (index <= endCapId); ++index)
    {
        capRecord = capRecords.value(index);

        // Assuming absolute time is always moving forward only really need to check the endframe's SBS value.
        if (timeInSecs <= capRecord->eFrameSBS)
        {
            // The time in seconds is encased inside of this cap file but is there a special case
            // where the USec could span two cap files?

            // If the timeInSecs is exactly equal to the end frame's SBS double check the useconds.
            if (timeInSecs == capRecord->eFrameSBS)
            {
                // Where does timeInUsecs fall compared to eFrameSBSUSec of this CAP file?
                // It may be that the next CAP is the better fit but maybe not.
                if (timeInUSecs > capRecord->eFrameSBSUSec)
                {
                    // Is there another cap file?
                    if ((index + 1) <= endCapId)
                    {
                        QSharedPointer<CapMapRecord>nextCapRecord = capRecords.value((index + 1));

                        // Does the next cap record start with the timeInSecs?
                        if (timeInSecs == nextCapRecord->sFrameSBS)
                        {
                            // Yes it does, so the next CAP file is the better fit.
                            // Generate a continue and processing naturally falls out in next for loop pass.
                            continue;
                        }
                    }
                }
            }

            // This is the cap file that contains the time.
            capIdFound = capRecord->capId;
            foundCapId = true;
        }
    }

    // Did we find a CAP record that has this time?
    if (foundCapId)
    {
        capPMFile.setFileName(capRecord->capPFMDPath);

        // Ask our math class what size a pfmd record is for non-series3 camera.
        uint32_t pfmdRecSize = ClientMath::get_pfmd_size(false);

        // Make a buffer that can hold one pfmd record.
        QVector<uint32_t> pfmdRecord(pfmdRecSize / sizeof(uint32_t));

        // Open our file the starting frame is always at position zero.
        if (!capPMFile.open(QIODevice::ReadOnly))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to open " << capRecord->capPFMDPath << "to read pfmd for starting frame" << capRecord->sFrameId << "!";
            goto GetAbsFrameUsingTimeExit;
        }

        // Seek to our first record.
        if (!capPMFile.seek(0))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to seek " << capRecord->capPFMDPath << "to read pfmd for starting frame" << capRecord->sFrameId << "!";
            goto GetAbsFrameUsingTimeExit;
        }

        // Loop through each frame's pfmd record looking for matching time to the passed in value.
        for (uint32_t frameIndex = 0; !foundAbsFrameNum && (frameIndex < capRecord->frameCount); ++frameIndex)
        {
            // Read the frame's metadata.
            if (capPMFile.read((char *)pfmdRecord.data(), pfmdRecSize) < 0)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read frame" << absFrameNumFound << "pfmd (per-frame_metadata) record in" << capRecord->capPFMDPath;
                goto GetAbsFrameUsingTimeExit;
            }

            uint32_t sbsTime = pfmdRecord[0];
            uint32_t sbsUTime = (pfmdRecord[1] & 0xFFFFF);

            // The following condition statements leverage logic found in PlaybackMetaInfo::find_frame.

            // Do we have a time match?
            if ((sbsTime == timeInSecs) && (sbsUTime == timeInUSecs))
            {
                foundAbsFrameNum = true;
                absFrameNumFound = frameIndex + capRecord->sFrameId;
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Matching absFrameNumFound:" << absFrameNumFound
                             << "In CAP:" << capRecord->capPFMDPath << "(MATCH REQUEST: timeInSecs:" << timeInSecs << "timeInUSecs:" << timeInUSecs << ")"
                             << "sFrameSBS:" << capRecord->sFrameSBS << "sFrameUSBS:" << capRecord->sFrameSBSUSec
                             << "eFrameSBS:" << capRecord->eFrameSBS << "eFrameUSBS:" << capRecord->eFrameSBSUSec;
            }
            // Did we go past the time we're looking for?
            else if ((sbsTime > timeInSecs) || ((sbsTime == timeInSecs) && (sbsUTime > timeInUSecs)))
            {
                foundAbsFrameNum = true;
                absFrameNumFound = frameIndex + capRecord->sFrameId;
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Matching absFrameNumFound:" << absFrameNumFound
                             << "In CAP:" << capRecord->capPFMDPath << "(MATCH REQUEST: timeInSecs:" << timeInSecs << "timeInUSecs:" << timeInUSecs << ")"
                             << "sFrameSBS:" << capRecord->sFrameSBS << "sFrameUSBS:" << capRecord->sFrameSBSUSec
                             << "eFrameSBS:" << capRecord->eFrameSBS << "eFrameUSBS:" << capRecord->eFrameSBSUSec;
            }
        }
    }

GetAbsFrameUsingTimeExit:
    // Close any opened pfmd file.
    if (capPMFile.isOpen())
        capPMFile.close();

    return foundAbsFrameNum;
}

/*!
 * @brief Locates the frame number in the CAP stack and sets the CapId for the location.
 *
 * @param[in] absFrameNum       The absolute frame number of the CAP stack to locate.
 * @param[out] capIdFound       The capId found in the capRecords for the absFrameNum.
 * @param[out] absCapFrameNum   The absolute frame number (absFrameNum) translated to the CAP file's frame index.
 *
 * @returns True if all is successful and the frame is in the CapRecords, false if it doesn't exist.
 */
bool FiMetaMap::GetCapInfoForFrame(uint32_t absFrameNum, uint32_t& capIdFound, uint32_t& absCapFrameNum)
{
    bool foundCapId = false;

    // Set it to zero for lack of a better value.
    capIdFound = 0;
    absCapFrameNum = 0;

    // Where is it in the record list?  --
    // We may have to optimize the search here if there is a very large number of CAP files in the stack.
    // But with a 4GB partition perhaps this may not be an issue.  TBD.
    for (uint32_t index = startCapId; !capIdFound && (index <= endCapId); ++index)
    {
        QSharedPointer<CapMapRecord> capRecord = capRecords.value(index);
        if ((absFrameNum >= capRecord->sFrameId) && (absFrameNum <= capRecord->eFrameId))
        {
            // Set the capId.
            capIdFound = capRecord->capId;

            // Set the zero-based frame number for this CAP file's frame entry matching absFrameNum.
            absCapFrameNum = absFrameNum - capRecord->sFrameId;
            foundCapId = true;
        }
    }

    return foundCapId;
}

/*!
 * @brief Creates the full path to the "pm_metadata" (metadata.txt) for the starting cap file
 * belonging to this CAP stack.
 *
 * @param[in] capId             Which cap in the stack are we asking about.
 * @param[out] capNameFound     The full path to the metadata.pm for capId or empty if not found.
 *
 * @returns True if all is successful and the file exists, false if an error occurred or the file doesn't exist.
 */
bool FiMetaMap::GetCAPName(uint32_t capId, QString& capNameFound)
{
    bool fileExists = false;

    // Empty the output value.
    capNameFound.clear();

    // Make sure there is a cache entry for our starting cap.
    if (!capRecords.contains(capId))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to find capRecord for capId:" << capId << "!";
        goto GetCAPNameExit;
    }
    else
    {
        QFileInfo capFileInfo;
        QSharedPointer<CapMapRecord> capRecord = capRecords.value(capId);

        capNameFound = capRecord->capPath;

        // Make sure it exists.
        capFileInfo.setFile(capNameFound);

        if (capFileInfo.exists())
        {
            fileExists = true;
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to verify existance of" << capNameFound << "for capId:" << capId << "!";
            goto GetCAPNameExit;
        }
    }

GetCAPNameExit:
    return fileExists;
}

/*!
 * @brief Retrieves the first frame information for a recording.
 *
 * @param[out] capIdFound           The startCapId.
 * @param[out] absFrameNum          The absolute frame number for the very first frame in the video.
 *
 * @returns True if all is successful or false if can't find the starting frame.
 */
bool FiMetaMap::GetFirstFrameInfo(uint32_t& capIdFound, uint32_t& absFrameNum)
{
    bool foundIt = false;

    // Initialize them to zero even though they aren't known values just to set it to something reasonable
    capIdFound = 0;
    absFrameNum = FASMOTION_SETTINGS::UNSET_UINT32;

    // Make sure it is contained in the record list.
    if (capRecords.contains(startCapId))
    {
        QSharedPointer<CapMapRecord> capRecord = capRecords.value(startCapId);
        absFrameNum = capRecord->sFrameId;
        capIdFound = startCapId;
        foundIt = true;
    }

    return foundIt;
}

/*!
 * @brief Retrieves the last frame information for a recording.
 *
 * @param[out] capIdFound       The endCapId.
 * @param[out] absFrameNum      The absolute frame number for the very first frame in the video.
 *
 * @returns True if all is successful or false if can't find the starting frame.
 */
bool FiMetaMap::GetLastFrameInfo(uint32_t& capIdFound, uint32_t& absFrameNum)
{
    bool foundIt = false;

    // Initialize them to zero even though they aren't known values just to set it to something reasonable
    capIdFound = 0;
    absFrameNum = FASMOTION_SETTINGS::UNSET_UINT32;

    // Make sure it is contained in the record list.
    if (capRecords.contains(endCapId))
    {
        QSharedPointer<CapMapRecord> capRecord = capRecords.value(endCapId);
        absFrameNum = capRecord->eFrameId;
        capIdFound = endCapId;
        foundIt = true;
    }

    return foundIt;
}

/*!
 * @brief Retrieves the trigger cap id for a basic recording.
 *
 * @param[out] triggerCapId     The capId found that contains the trigger frame.
 * @param[out] absTrigFrameNum  The absolute frame number for the trigger CAP file of the trigger frame.
 *
 * @returns True if all is successful or false if can't find the trigger frame.
 */
bool FiMetaMap::GetTriggerInfo(uint32_t& triggerCapId, uint32_t& absTrigFrameNum)
{
    bool foundIt = false;

    // Initialize them to zero even though they aren't known values just to set it to something reasonable
    triggerCapId = 0;
    absTrigFrameNum = FASMOTION_SETTINGS::UNSET_UINT32;

    if (hasTrigCap)
    {
        // Make sure it is contained in the record list.
        if (capRecords.contains(trigCap))
        {
            QSharedPointer<CapMapRecord> capRecord = capRecords.value(trigCap);
            if (capRecord->tFrameValid)
            {
                absTrigFrameNum = capRecord->tFrameId;
                triggerCapId = trigCap;
                foundIt = true;
            }
        }
    }
    return foundIt;
}

/*!
 * @brief Retrieves the total number of frames for this cap stack.
 *
 * @param[out] totalFramesFound     The total number of frames for this recording.
 *
 * @returns True if all is successful or false if can't find the trigger frame.
 */
bool FiMetaMap::GetTotalFrameCount(uint32_t& totalFramesFound)
{
    totalFramesFound = totalFrames;
    return true;
}

/*!
 * @brief Retrieves the CAP file's opened object for use based on a frame number provided.
 *
 * @param[in] absFrameNum           The absolute frame number for the CAP stack to locate.
 * @param[out] capIdFound           The capId of the CAP file found.
 * @param[out] capFrameIndex        The zero-based index of frame number in the cap matching the absFrameNum.
 *
 * @returns The opened CAP file if the frame was found, or a null empty one if not.
 */
QSharedPointer<fiRawFile> FiMetaMap::GetCAPFileUsingFrame(uint32_t absFrameNum, uint32_t& capIdFound, uint32_t& capFrameIndex)
{
    QSharedPointer<fiRawFile> capFile;
    QSharedPointer<CapMapRecord> capRecord;

    // Set the capIdFound to an unitialized state.
    capIdFound = FASMOTION_SETTINGS::UNSET_UINT32;

    // Set the capFrameIndex to match the absFrameNum.
    capFrameIndex = absFrameNum;

    // Returns the CAP Id belonging to the frame, o
    if (!GetCapInfoForFrame(absFrameNum, capIdFound, capFrameIndex))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to locate frame" << absFrameNum << "in CAP stack.";
        goto GetCAPFileUsingFrameExit;
    }

    // Get access to the capId's capRecord.
    if (capRecords.contains(capIdFound))
    {
        capRecord = capRecords.value(capIdFound);
    }
    else
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to locate capId" << capIdFound << "in CAP stack records.";
        goto GetCAPFileUsingFrameExit;
    }

    // Is the cap loaded in our LRU cache?
    capFile = capLRUFiles->retrieve(capIdFound);

    // If not, open the CAP file and process it for use and add it to the cache.
    if (capFile.isNull())
    {
        // Allocate our raw cap file object.
        capFile = QSharedPointer<fiRawFile>(new fiRawFile());
        if (!capFile)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to fiRawFile for" << capRecord->capPath;
            goto GetCAPFileUsingFrameExit;
        }

        // Open the CAP file, put it in the LRU cache and return it.
        if (!capFile->open(capRecord->capPath))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open the cap file:" << capRecord->capPath;
            goto GetCAPFileUsingFrameExit;
        }

        // Add it to the LRU cache.
        capLRUFiles->insert(capIdFound, capFile);
    }

GetCAPFileUsingFrameExit:
    return capFile;
}

/*!
 * @brief Retrieves the CAP file's opened object for use based on a frame number provided.
 *
 * @param[out] absFrameNum      The absolute frame number for the first playable frame based on recording type.
 * @param[out] capIdFound       The capId of the CAP file found housing the first playable frame.
 *
 * @returns True if found, false if unknown.
 */
bool FiMetaMap::GetFirstPlayableFrame(uint32_t& absFrameNum, uint32_t& capIdFound)
{
    bool foundIt = false;

    absFrameNum = FASMOTION_SETTINGS::UNSET_UINT32;    //!< The absolute frame number for the first display.
    capIdFound = 0;

    // What type of recording are we managing?
    // The first frame to display will depend on the recording type.
    if (IsBasicRecording())
    {
        // The playback will start with the absolute trigger frame.
        if (!GetTriggerInfo(capIdFound, absFrameNum))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "This Basic FasFireLR recording is missing a trigger frame!";
            goto GetFirstPlayableFrameExit;
        }
    }
    else if (IsBROCRecording() || IsROCRecording())
    {
        // For both BROC and ROC we start at the first frame in the video.
        if (!GetFirstFrameInfo(capIdFound, absFrameNum))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to get the first frame of the recording!";
            goto GetFirstPlayableFrameExit;
        }
    }
    else
    {
        QLOG_INFO() << __PRETTY_FUNCTION__ << "We don't support image memory CAP stacks at this time.";
        goto GetFirstPlayableFrameExit;
    }

    // If you make it to this execution point all is good.
    foundIt = true;

GetFirstPlayableFrameExit:
    return foundIt;
}





