/*!
 * @file fi_metamap.h
 * @brief This is the header file for the fiMetaMap class.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#ifndef FI_METAMAP_H
#define FI_METAMAP_H

#include <QFileInfo>
#include <QSharedPointer>
#include <QString>

#include "playback_meta_info.h"
#include "fi_rawfile.h"

class CapLRUCache;  //!< Forward declaration.

namespace FI_METAMAP
{

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define the different record types for FFLR.
typedef enum
{
    FFLR_IMEM = 0,
    FFLR_BASIC,
    FFLR_BROC,
    FFLR_ROC
} FFLR_REC_TYPE;

//--------------------------------------------------------------------
// Class Definition CapMapRecord
//--------------------------------------------------------------------

//! Define the map record for FiMetaMap map entries where each record defines a
//! CAP file that can be collected together as a single CAP file.
class CapMapRecord
{
public:
    CapMapRecord() {}
    virtual ~CapMapRecord() = default;

    // This object is non-copyable and non-copy assignable.
    CapMapRecord(const CapMapRecord& i) = delete;
    CapMapRecord& operator=(const CapMapRecord& i) = delete;

    //! Factory creation methods.
    static QSharedPointer<CapMapRecord> CreateCapMapRecord(uint32_t capId,
                                                           uint32_t sFrame,
                                                           uint32_t eFrame,
                                                           uint32_t tFrame,
                                                           uint32_t frameCount,
                                                           bool tFrameValid,
                                                           const QString& locationDir);

    static bool CreateCapPath(const QString& locationDir, uint32_t capId, QString& pathToCapCreated);
    static bool CreateCapPMPath(const QString& locationDir, uint32_t capId, QString& pathToCapPMCreated);

    uint32_t    capId;              //!< The zero-based capId for this record.
    uint32_t    sFrameId;           //!< Absolute frame number for start of this capFile.
    uint32_t    sFrameSBS;          //!< The sbs time of start frame capture from pfmd.
    uint32_t    sFrameSBSUSec;      //!< The sbs usec time of start frame capture from pfmd.
    uint32_t    eFrameId;           //!< Absolute frame number for end of this capFile.
    uint32_t    eFrameSBS;          //!< The sbs time of end frame capture from pfmd.
    uint32_t    eFrameSBSUSec;      //!< The sbs usec time of end frame capture from pfmd.
    uint32_t    tFrameId;           //!< Absolute frame number for trigger frame if one exists.
    uint32_t    frameCount;         //!< The frame count contained in capFile.
    bool        tFrameValid;        //!< Is the tFrameId valid?
    QString     capPath;            //!< The full path name to capFile.
    QString     capPFMDPath;        //!< The full path name to the .pm File associated with this CAP.
    QSharedPointer<fiRawFile> capFile;     //!< The cap file if available.
};

//--------------------------------------------------------------------
// Class Definition FiMetaMap
//--------------------------------------------------------------------

class FiMetaMap
{
private:
    FiMetaMap();

public:
    ~FiMetaMap();

    static const QString METAMAP_FILE;

    static QSharedPointer<FI_METAMAP::FiMetaMap> CreateMetaMap(const QString& mmPath);

    bool GetAbsFrameUsingTime(uint32_t timeInSecs, uint32_t timeInUSecs, uint32_t& capIdFound, uint32_t& absFrameNumFound);

    bool GetCapInfoForFrame(uint32_t absFrameNum, uint32_t& capIdFound, uint32_t& absCapFrameNum);
    bool GetCAPName(uint32_t capId, QString& capNameFound);
    bool GetMetadataTXTName(uint32_t capId, QString& metadataTXTNameFound);
    bool GetMMFileName(QString& mmFileNameFound);
    bool GetPFMDName(uint32_t capId, QString& pfmdNameFound);

    bool GetPFMDOffset(uint32_t absFrameNum, uint32_t& capIdFound, QString& pfmdFileName, uint32_t& pfmdOffset);
    uint8_t *GetPFMDEvents(bool do_swap);
    bool GetPFMD(uint32_t capId, bool appendFileData, QByteArray& pfmdFileData);

    bool GetFirstFrameInfo(uint32_t& capIdFound, uint32_t& absFrameNum);
    bool GetLastFrameInfo(uint32_t& capIdFound, uint32_t& absFrameNum);
    bool GetTriggerInfo(uint32_t& triggerCapId, uint32_t& absTrigFrameNum);

    bool GetTotalFrameCount(uint32_t& totalFramesFound);

    bool IsBasicRecording()     { return (recType == FI_METAMAP::FFLR_BASIC);}
    bool IsBROCRecording()      { return (recType == FI_METAMAP::FFLR_BROC);}
    bool IsROCRecording()       { return (recType == FI_METAMAP::FFLR_ROC);}

    QSharedPointer<fiRawFile> GetCAPFileUsingFrame(uint32_t absFrameNum, uint32_t& capIdFound, uint32_t& capFrameIndex);

    bool GetFirstPlayableFrame(uint32_t& absFrameNum, uint32_t& capIdFound);


private:
    static const QString MM_PARTSIZE;
    static const QString MM_RECMODE;
    static const QString MM_RECMODE_IMEM;
    static const QString MM_RECMODE_BASIC;
    static const QString MM_RECMODE_ROC;
    static const QString MM_RECMODE_BROC;
    static const QString MM_START_CAP;
    static const QString MM_END_CAP;
    static const QString MM_BLEN;
    static const QString MM_TRIGGER_CAP;
    static const QString MM_TFRAMES;
    static const QString MM_MAP;
    static const QString MM_MAP_EFRAME;
    static const QString MM_MAP_SFRAME;
    static const QString MM_MAP_TFRAME;
    static const QString MM_MAP_NAME;
    static const QString MM_MAP_FCOUNT;

    FI_METAMAP::FFLR_REC_TYPE recType;  //!< If FFLR cap stack how was it recorded?
    uint32_t partSize;                  //!< Partition size used for FFLR cap stack capture.
    uint32_t startCapId;                //!< Number for starting cap in FFLR cap stack.
    uint32_t endCapId;                  //!< Number for ending cap in FFLR cap stack.
    uint32_t brocLen;                   //!< If recType is FFLR_BROC this defines burst length.
    uint32_t trigCap;                   //!< If recType is FFLR_BASIC this defines cap with trigger frame.
    uint32_t totalFrames;               //!< Total frame count for the entire FFLR cap stack.

    bool hasBrocLen;                    //!< Will provide some assistence in determining if trigCap is valid.
    bool hasTrigCap;                    //!< Will provide some assistence in determining if brocLen is valid.

    QString mmFilePath;                 //!< Full path to metamap.txt file *with* the ending slash.

    //!< CAP file records belonging to the CAP stack this object is managing.
    QMap<uint32_t, QSharedPointer<CapMapRecord>> capRecords;

    //!< CAP fiRawFile LRU cache.
    QSharedPointer<CapLRUCache> capLRUFiles;

};

} // End of FI_METAMAP namespace.

#endif // FI_METAMAP_H
