/*!
 * @file fi_msgbox.cpp
 * @brief This file contains the implementation of the fiMsgBox
 *        methods. These methods are modeless equivalents to the
 *        Qt MessageBox methods.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_msgbox.h"

#include <QMessageBox>

/*!
 * @brief fiMsgBox constructor. Nothing to do.
 */
fiMsgBox::fiMsgBox()
{
}

/*!
 * @brief A utility method to provide a modeless "critical" message box.
 *
 * @param parent            Pointer to the parent widget.
 * @param text_str          The text string to show.
 * @param informative_str   [optional] Informative text to show.
 */
void fiMsgBox::critical( QWidget *parent, QString text_str, QString informative_str )
{
    // Notify the user.
    QMessageBox *err_box = new QMessageBox( parent );

    // Build up our error message.
    err_box->setText( text_str );

    // Build up our string with more information.
    if ( !informative_str.isEmpty() )
        err_box->setInformativeText( informative_str );

    // Make sure the user knows this is bad.
    err_box->setIcon( QMessageBox::Critical );

    // Catch the closing so we can free up the memory later.
    err_box->connect( err_box, SIGNAL(accepted()), err_box, SLOT(deleteLater()));

    // Show the message as a modeless dialog.
    err_box->show();
}

/*!
 * @brief A utility method to provide a modeless "warning" message box.
 *
 * @param parent            Pointer to the parent widget.
 * @param text_str          The text string to show.
 * @param informative_str   [optional] Informative text to show.
 */
void fiMsgBox::warning( QWidget *parent, QString text_str, QString informative_str )
{
    // Notify the user.
    QMessageBox *warn_box = new QMessageBox( parent );

    // Build up our error message.
    warn_box->setText( text_str );

    // Build up our string with more information.
    if ( !informative_str.isEmpty() )
        warn_box->setInformativeText( informative_str );

    // Make sure the user knows this is bad.
    warn_box->setIcon( QMessageBox::Warning );

    // Catch the closing so we can free up the memory later.
    warn_box->connect( warn_box, SIGNAL(accepted()), warn_box, SLOT(deleteLater()));

    // Show the message as a modeless dialog.
    warn_box->show();
}
