#include "fi_overlay.h"

#include <QApplication>
#include <QBrush>
#include <QColorDialog>
#include <QFileDialog>
#include <QFontDialog>
#include <QGraphicsScene>
#include <QMenu>
#include <QPen>
#include <math.h>

#include "camera_info.h"
#include "camera_mdi.h"
#include "fi_graphicsview.h"
#include "flowlayout.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_meta_info.h"
#include "save_dialog.h"

Q_DECLARE_METATYPE(QList<int>)

/**
  *  This box can be re-sized and it can be moved. For moving, we capture
  *  the mouse move events and move the box location accordingly.
  *
  *  To resize the box, we place small indicator boxes on the four corners that give the user
  *  a visual cue to grab with the mouse. The user then drags the corner to stretch
  *  or shrink the box.  The corner grabbers are implemented with the CornerGrabber class.
  *  The CornerGrabber class captures the mouse when the mouse is over the corner's area,
  *  but the fiOverlay object (which owns the corners) captures and processes the mouse
  *  events on behalf of the CornerGrabbers (because the owner wants to be
  *  resized, not the CornerGrabbers themselves). This is accomplished by installed a scene event filter
  *  on the CornerGrabber objects:
  *       corners[0]->installSceneEventFilter(this);
  */

namespace FI_OVERLAY
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define a space around our overlay to allow for the hover events to fire.
//! NOTE: It's important to keep this an even number!
const int   CORNER_BUFFER               = 20;

//! Lock our overlays onto a grid.
//! NOTE: It's important to keep this an even number!
const int   GRID_SPACE                  = 10;

}   // End of namespace FI_OVERLAY.

using namespace FI_OVERLAY;

fiOverlay::fiOverlay(QGraphicsItem *parent, const fiOverlay *copy_ptr) :
    QGraphicsObject(parent),
    QGraphicsLayoutItem(),
    location(0, 0),
    drag_start(0, 0),
    width(200),
    height(3 * CORNER_BUFFER),
    drawing_width(width - CORNER_BUFFER),
    drawing_height(height - CORNER_BUFFER),
    overlay_color(QColor(243, 255, 216, 255)),
    text_color(QApplication::palette().windowText().color()),
    item_font(QApplication::font())
{
    // Initialize our members.
    user_ftime_text.clear();
    user_text.clear();
    flow_widget = new QGraphicsWidget(this);
    outline_pen = QPen(Qt::NoPen);

    // Initialize our update timer.
    delayed_update.setInterval(100);
    delayed_update.setSingleShot(true);
    delayed_update.callOnTimeout(this, &fiOverlay::do_delayed_update);

    // Set this to zero.
    rel_frame_num_default = 0;

    FlowLayout *flow_lay = new FlowLayout;
    QFontMetrics font_metrics(item_font);
    flow_lay->setContentsMargins(9, 9, 9 + CORNER_BUFFER, 9 + CORNER_BUFFER);
    flow_lay->setSpacing(Qt::Horizontal, 2 * font_metrics.averageCharWidth());

    // Clear the overlay item list.
    overlay_item.clear();

    // Are we making a copy?
    if (copy_ptr != nullptr)
    {
        place_holder = nullptr;
    }
    else
    {
        // Add our place holder to the overlay.
        place_holder = new fiLayoutText(QApplication::translate("overlay", "<add items>"), flow_widget);
        place_holder->setFont(item_font);
        place_holder->setBrush(QBrush(text_color));
        flow_lay->addItem(place_holder);

        // Set the layout.
        flow_widget->setLayout(flow_lay);
    }

    // Make sure we get hover events, so we can turn the corners on/off.
    setAcceptHoverEvents(true);

    // Initialize our corners.
    corners[0] = nullptr;
    corners[1] = nullptr;
    corners[2] = nullptr;
    corners[3] = nullptr;

    // Initialize our actions.
    action_background = new QAction(nullptr);
    action_background->setObjectName(QStringLiteral("action_background"));
    action_background->setCheckable(false);
    action_background->setText(QApplication::translate("overlay", "Overlay Background Color..."));
    action_background->setStatusTip(QApplication::translate("overlay", "Select a background color for the overlay."));
    connect(action_background, SIGNAL(triggered()), this, SLOT(do_set_background_color()));

    action_image = new QAction(nullptr);
    action_image->setObjectName(QStringLiteral("action_image"));
    action_image->setCheckable(false);
    action_image->setText(QApplication::translate("overlay", "Overlay Image..."));
    action_image->setStatusTip(QApplication::translate("overlay", "Select a background image for the overlay."));
    connect(action_image, SIGNAL(triggered()), this, SLOT(do_set_image()));

    action_delete_image = new QAction(nullptr);
    action_delete_image->setObjectName(QStringLiteral("action_delete_image"));
    action_delete_image->setCheckable(false);
    action_delete_image->setText(QApplication::translate("overlay", "Delete Overlay Image..."));
    action_delete_image->setStatusTip(QApplication::translate("overlay", "Remove the background image from the overlay."));
    action_delete_image->setEnabled(false);
    connect(action_delete_image, SIGNAL(triggered()), this, SLOT(do_remove_image()));

    action_delete = new QAction(nullptr);
    action_delete->setObjectName(QStringLiteral("action_delete"));
    action_delete->setCheckable(false);
    action_delete->setText(QApplication::translate("overlay", "Delete Overlay"));
    action_delete->setStatusTip(QApplication::translate("overlay", "Remove the overlay object from the image."));
    connect(action_delete, SIGNAL(triggered()), this, SIGNAL(remove_overlay()));

    action_item_font = new QAction(nullptr);
    action_item_font->setObjectName(QStringLiteral("action_item_font"));
    action_item_font->setCheckable(false);
    action_item_font->setText(QApplication::translate("overlay", "Change Item Font..."));
    action_item_font->setStatusTip(QApplication::translate("overlay", "Select a different font for the items."));
    connect(action_item_font, SIGNAL(triggered()), this, SLOT(do_change_font()));

    action_item_foreground = new QAction(nullptr);
    action_item_foreground->setObjectName(QStringLiteral("action_item_foreground"));
    action_item_foreground->setCheckable(false);
    action_item_foreground->setText(QApplication::translate("overlay", "Item Text Color..."));
    action_item_foreground->setStatusTip(QApplication::translate("overlay", "Select a text color for the items."));
    connect(action_item_foreground, SIGNAL(triggered()), this, SLOT(do_change_item_color()));

    action_item_outline = new QAction(nullptr);
    action_item_outline->setObjectName(QStringLiteral("action_item_outline"));
    action_item_outline->setCheckable(true);
    action_item_outline->setText(QApplication::translate("overlay", "Outline Text"));
    action_item_outline->setStatusTip(QApplication::translate("overlay", "Draw an outine around each text character."));
    connect(action_item_outline, SIGNAL(triggered()), this, SLOT(do_text_outline()));

    action_pin_bottom = new QAction(nullptr);
    action_pin_bottom->setObjectName(QStringLiteral("action_pin_bottom"));
    action_pin_bottom->setCheckable(false);
    action_pin_bottom->setText(QApplication::translate("overlay", "Pin To Bottom Border"));
    action_pin_bottom->setStatusTip(QApplication::translate("overlay", "Pin the overlay to the bottom border of the image."));
    connect(action_pin_bottom, SIGNAL(triggered()), this, SIGNAL(pin_to_bottom()));

    action_pin_top = new QAction(nullptr);
    action_pin_top->setObjectName(QStringLiteral("action_pin_top"));
    action_pin_top->setCheckable(false);
    action_pin_top->setText(QApplication::translate("overlay", "Pin To Top Border"));
    action_pin_top->setStatusTip(QApplication::translate("overlay", "Pin the overlay to the top border of the image."));
    connect(action_pin_top, SIGNAL(triggered()), this, SIGNAL(pin_to_top()));

    action_select_item = new QAction(nullptr);
    action_select_item->setObjectName(QStringLiteral("action_select_item"));
    action_select_item->setCheckable(false);
    action_select_item->setText(QApplication::translate("overlay", "Select Item..."));
    action_select_item->setStatusTip(QApplication::translate("overlay", "Select the items to be in the overlay."));
    connect(action_select_item, SIGNAL(triggered()), this, SLOT(do_select_item()), Qt::QueuedConnection);

    // Do we have an item to copy from?
    if (copy_ptr != nullptr)
    {
        pinned = copy_ptr->pinned;
        rel_frame_num_default = copy_ptr->rel_frame_num_default;
        location = copy_ptr->location;
        width = copy_ptr->width;
        height = copy_ptr->height;
        pinned_height = copy_ptr->pinned_height;
        pinned_width = copy_ptr->pinned_width;
        drawing_width = copy_ptr->drawing_width;
        drawing_height = copy_ptr->drawing_height;

        overlay_color = copy_ptr->overlay_color;
        text_color = copy_ptr->text_color;
        item_font = copy_ptr->item_font;
        outline_pen = copy_ptr->outline_pen;
        back_image = copy_ptr->back_image;

        overlay_item = copy_ptr->overlay_item;
        user_ftime_text = copy_ptr->user_ftime_text;
        user_text = copy_ptr->user_text;

        setPos(copy_ptr->pos());
    }
}

fiOverlay::~fiOverlay()
{
    // Remove the existing items.
    if (flow_widget != nullptr)
    {
        int     item_id;
        QList<QGraphicsItem *> kid_items;

        kid_items = flow_widget->childItems();
        for (item_id = 0; item_id < kid_items.count(); ++item_id)
        {
            kid_items.at(item_id)->setParentItem(nullptr);
            if (scene() != nullptr)
            {
                if (kid_items.at(item_id)->scene() == scene())
                    scene()->removeItem(kid_items.at(item_id));
                else
                    delete kid_items.at(item_id);
            }
        }
    }

    if (place_holder != nullptr)
    {
        place_holder->setParentItem(nullptr);
        if (scene() != nullptr)
        {
            if (place_holder->scene() == scene())
                scene()->removeItem(place_holder);
        }
        place_holder = nullptr;
    }

    if (action_delete)
    {
        delete action_delete;
        action_delete = nullptr;
    }
    if (action_delete_image)
    {
        delete action_delete_image;
        action_delete_image = nullptr;
    }
    if (action_background)
    {
        delete action_background;
        action_background = nullptr;
    }
    if (action_image)
    {
        delete action_image;
        action_image = nullptr;
    }
    if (action_item_outline)
    {
        delete action_item_outline;
        action_item_outline = nullptr;
    }
    if (action_item_font)
    {
        delete action_item_font;
        action_item_font = nullptr;
    }
    if (action_item_foreground)
    {
        delete action_item_foreground;
        action_item_foreground = nullptr;
    }
    if (action_pin_bottom)
    {
        delete action_pin_bottom;
        action_pin_bottom = nullptr;
    }
    if (action_pin_top)
    {
        delete action_pin_top;
        action_pin_top = nullptr;
    }
    if (action_select_item)
    {
        delete action_select_item;
        action_select_item = nullptr;
    }
}

void fiOverlay::add_item(FlowLayout *layout, int item_id, uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr, fiLayoutText *item_ptr)
{
    QString     item_str;
    QStringList item_list;
    uint32_t frameCountRetrieved = 0;

    // Sanity check.
    if (item_id >= SelectItemDlg::OPTION_COUNT)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected ID:" << item_id;
        return;
    }

    // If we're updating, skip the items which don't change
    if (item_ptr != nullptr)
    {
        switch (item_id)
        {
        case SelectItemDlg::CUSTOM_TEXT:
        case SelectItemDlg::MAC:
        case SelectItemDlg::MANUFACTURER:
        case SelectItemDlg::MODEL:
        case SelectItemDlg::VERSION:
            return;
        case -1:
            // This is the "placeholder" item.
            return;
        default:
            break;
        }
    }

    // Process based on the ID.
    switch (item_id)
    {
    case SelectItemDlg::ABS_FRAME:
        item_str = SelectItemDlg::get_label(item_id) + QString(": %1").arg(frame_num);
        item_list << item_str;
        break;
    case SelectItemDlg::ABS_TIME:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_time();
        item_list << item_str;
        break;
    case SelectItemDlg::BIT_SEL:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + metadata->getBitMode();
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_bit_mode(camera_ptr->read_camera_cache(GIGE_REG_BIT_MODE));
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::CUSTOM_TEXT:
        item_list = user_text.split(QLatin1Char(' '), Qt::SkipEmptyParts);
        break;
    case SelectItemDlg::DURATION:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_time((int)metadata->getFrameCount() - 1, 0, (uint32_t)info_ptr->get_fps(), metadata, camera_ptr);
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_time((int)camera_ptr->read_camera_cache(GIGE_REG_PART_NUM_FRAMES) - 1, 0, (uint32_t)info_ptr->get_fps(), metadata, camera_ptr);
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::EVENT:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_markers();
        item_list << item_str;
        break;
    case SelectItemDlg::EXPOSURE:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_exposure();
        item_list << item_str;
        break;
    case SelectItemDlg::FRAME_COUNT:
        if (metadata != nullptr)
        {
            frameCountRetrieved = metadata->getFrameCount();
            item_str = SelectItemDlg::get_label(item_id) + QString(": %1").arg(frameCountRetrieved);
            //QLOG_DEBUG() << QString("fiOverlay::add_item()[FRAMECOUNT]metadata->get_frame_count()[0x%1,%2].").arg(frameCountRetrieved, 8, 16).arg(frameCountRetrieved);
        }
        else if (camera_ptr != nullptr)
        {
            frameCountRetrieved = camera_ptr->read_camera_cache(GIGE_REG_PART_NUM_FRAMES);
            item_str = SelectItemDlg::get_label(item_id) + QString(": %1").arg(frameCountRetrieved);
            //QLOG_DEBUG() << QString("fiOverlay::add_item()[FRAMECOUNT]camera_ptr->read_camera_cache(GIGE_REG_PART_NUM_FRAMES)[0x%1,%2].").arg(frameCountRetrieved, 8, 16).arg(frameCountRetrieved);
        }
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::FRAMERATE:
        item_str = SelectItemDlg::get_label(item_id) + QString(": %1").arg(info_ptr->get_fps());
        item_list << item_str;
        break;
    case SelectItemDlg::GAIN:
        if (!info_ptr->get_gain().isEmpty())
        {
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_gain();
            item_list << item_str;
        }
        break;
    case SelectItemDlg::MAC:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + metadata->getMacAddress();
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + camera_ptr->get_mac();
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::MANUFACTURER:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + metadata->getManufacturer();
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + QString::fromLatin1(camera_ptr->get_camera_info()->man_name, FI_COMM::MANUFACTURERS_NAME_LEN);
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::MODEL:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + metadata->getModel();
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + QString::fromLatin1(camera_ptr->get_camera_info()->model_name, FI_COMM::MODEL_NAME_LEN);
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::NAME:
        if (info_ptr->get_camera_name().isEmpty())
        {
            if (metadata != nullptr)
                item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + metadata->getCameraName();
            else
                item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": <unknown>");
        }
        else
        {
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_camera_name();
        }

        item_list << item_str;
        break;
    case SelectItemDlg::REL_FRAME:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QString(": %1").arg((int)(frame_num - metadata->getTriggerFrame()));
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QString(": %1").arg((int)(frame_num - camera_ptr->read_camera_cache(GIGE_REG_PART_TRIG_FRAME)));
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::REL_TIME:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_time((int)frame_num, (int)metadata->getTriggerFrame(), (uint32_t)info_ptr->get_fps(), metadata, camera_ptr);
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_time((int)frame_num, (int)camera_ptr->read_camera_cache(GIGE_REG_PART_TRIG_FRAME), (uint32_t)info_ptr->get_fps(), metadata, camera_ptr);
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::REL_FTIME:
    {
        // Get the frame the user entered.  It must be within the range of absolute frames.
        int ftime_user_frame = user_ftime_text.toInt();
        int trigger_frame = 0;
        if (metadata != nullptr)
        {
            frameCountRetrieved = metadata->getFrameCount();
            trigger_frame = (int)metadata->getTriggerFrame();
        }
        else if (camera_ptr != nullptr)
        {
            frameCountRetrieved = camera_ptr->read_camera_cache(GIGE_REG_PART_NUM_FRAMES);
            trigger_frame = (int)camera_ptr->read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
        }

        // What is the absolute frame number of the user's relative frame?
        uint32_t abs_ftime_user_frame = (uint32_t)ftime_user_frame + trigger_frame;

        // Make sure within range.
        if (abs_ftime_user_frame > frameCountRetrieved)
            abs_ftime_user_frame = frameCountRetrieved - 1;

        // Note that we are not using the "label" as most of the other cases do. This is because we are
        // changing the format slightly for the overlay, vs. what the user sees as the overlay "label" in
        // the dialog.
        item_str = QStringLiteral("Time (Frel=") + user_ftime_text + QStringLiteral("): ") + get_time((int)frame_num, abs_ftime_user_frame, (uint32_t)info_ptr->get_fps(), metadata, camera_ptr);

        // Add the item to the overlay.
        if (!item_str.isEmpty())
            item_list << item_str;
    }
    break;
    case SelectItemDlg::ROI:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_roi();
        item_list << item_str;
        break;
    case SelectItemDlg::SENSOR_OPT:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + metadata->getSensorOpt();
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_sensor_opt(camera_ptr);
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::SMPTE_TIME:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_smpte_time((int)frame_num, (int)metadata->getTriggerFrame(), (uint32_t)info_ptr->get_fps());
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_smpte_time((int)frame_num, (int)camera_ptr->read_camera_cache(GIGE_REG_PART_TRIG_FRAME), (uint32_t)info_ptr->get_fps());
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::SYNC_SOURCE:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_sync();
        item_list << item_str;
        break;
    case SelectItemDlg::TIME_OFFSET:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + get_time0(info_ptr, metadata, camera_ptr);
        item_list << item_str;
        break;
    case SelectItemDlg::TIME_SOURCE:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_time_source();
        item_list << item_str;
        break;
    case SelectItemDlg::TRIGGER_POS:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + info_ptr->get_trig_pos();
        item_list << item_str;
        break;
    case SelectItemDlg::VERSION:
        if (metadata != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + metadata->getVersion();
        else if (camera_ptr != nullptr)
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + camera_ptr->get_version();
        else
            item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": <unknown>");
        if (!item_str.isEmpty())
            item_list << item_str;
        break;
    case SelectItemDlg::VT_FLAGS:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + QString::number(info_ptr->get_vt_flags());
        if ((info_ptr->get_vt_flags() & (CAMERA_INFO::VTRIG_FLAGS_REFRESH | CAMERA_INFO::VTRIG_FLAGS_TRIGGER)) != 0)
        {
            item_str += QChar(' ');
            if ((info_ptr->get_vt_flags() & CAMERA_INFO::VTRIG_FLAGS_TRIGGER) != 0)
                item_str += QStringLiteral("[TRIG]");
            if ((info_ptr->get_vt_flags() & CAMERA_INFO::VTRIG_FLAGS_REFRESH) != 0)
                item_str += QStringLiteral("[RFLOAD]");
        }
        item_list << item_str;
        break;
    case SelectItemDlg::VT_CHANGE:
        item_str = SelectItemDlg::get_label(item_id) + QStringLiteral(": ") + QString::number(info_ptr->get_vt_area_percent(), 'f', 2) + QStringLiteral("%") + QString(" (%1)").arg(info_ptr->get_vt_area_change());
        item_list << item_str;
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected value:" << item_id;
        break;
    }

    // Are we updating?
    if (item_ptr != nullptr)
    {
        qreal   new_width;
        qreal   new_height;

        // Update the item's text.
        item_ptr->setText(item_str);

        // Update the item's preferred and minimum sizes (they are the same).
        new_width = item_ptr->boundingRect().width();
        new_height = item_ptr->boundingRect().height();
        item_ptr->setPreferredSize(new_width, new_height);
        item_ptr->setMinimumSize(new_width, new_height);
    }
    else
    {
        // We're adding a brand new item.
        for (int i = 0; i < item_list.count(); ++i)
        {
            fiLayoutText  *item_label;

            // Create our object.
            item_label = new fiLayoutText(item_list.at(i), flow_widget);
            item_label->setFont(item_font);
            item_label->setBrush(QBrush(text_color));
            item_label->setPen(outline_pen);
            item_label->id = item_id;

            // Add the item to the layout object.
            layout->addItem(item_label);
        }
    }
}

/*!
 * @brief Adjust the size of our object.
 *
 * To allow the user to grab the corners to re-size, we need to get a hover
 * indication. But if the mouse pointer points to the left, then when the mouse
 * tip is to the left but just outsize the box, we will not get the hover.
 * So the solution is to tell the graphics scene the box is larger than
 * what the painter actually paints in. This way when the user gets the mouse
 * within a few pixels of what appears to be the edge of the box, we get
 * the hover indication.
 *
 *  So the CORNER_BUFFER is a few pixel wide buffer zone around the outside
 *  edge of the box.
 */
void fiOverlay::adjustSize(int x, int y)
{
    QSizeF  min_layout;

    // Adjust our width and height values.
    width += x;
    height += y;

    // Adjust our drawing region.
    if (pinned)
    {
        drawing_width = pinned_width - CORNER_BUFFER;
        drawing_height = pinned_height - CORNER_BUFFER;
    }
    else
    {
        drawing_width = width - CORNER_BUFFER;
        drawing_height = height - CORNER_BUFFER;
    }

    // Ask the layout manager to give us an estimated width.
    min_layout = (static_cast<FlowLayout *>(flow_widget->layout()))->minSize(QSizeF(-1, -1));

    // Use the larger of the two widths to determine the height.
    if (drawing_width < min_layout.width())
    {
        drawing_width = min_layout.width();
        width = drawing_width + CORNER_BUFFER;
    }

    // Calculate the height.
    min_layout = (static_cast<FlowLayout *>(flow_widget->layout()))->minSize(QSizeF(drawing_width, 0));

    // Sanity check the new width/height.
    if (drawing_width < min_layout.width())
    {
        drawing_width = min_layout.width();
        width = drawing_width + CORNER_BUFFER;
    }
    if (drawing_height < min_layout.height())
    {
        drawing_height = min_layout.height();
        height = drawing_height + CORNER_BUFFER;
    }

    // Resize our layout widget.
    flow_widget->layout()->setGeometry(QRectF(0, 0, drawing_width, drawing_height));
    flow_widget->resize(drawing_width, drawing_height);
}

/*!
 * @brief Required method for a QGraphicsItem, returns the outer bounds
 *        of the object.
 */
QRectF fiOverlay::boundingRect() const
{
    QRectF  rc;

    if (pinned)
        rc = QRectF(0, 0, pinned_width, pinned_height);
    else
        rc =  QRectF(0, 0, width, height);

    return rc;
}

const QString fiOverlay::get_bit_mode(uint32_t bit_mode)
{
    int     first = 0;
    int     last = 0;
    QString bit_mode_str;

    switch (bit_mode)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
        bit_mode_str = QApplication::translate("overlay", "Low 8");
        break;
    case GIGE_BITMODE_MIDDLE_8_BITS:
        bit_mode_str = QApplication::translate("overlay", "Middle 8");
        break;
    case GIGE_BITMODE_UPPER_8_BITS:
        bit_mode_str = QApplication::translate("overlay", "Upper 8");
        break;
    case GIGE_BITMODE_10_BITS:
        bit_mode_str = QApplication::translate("overlay", "Full 10");
        break;
    case GIGE_BITMODE_7_0:  first = 1;  last = 8;   break;
    case GIGE_BITMODE_8_1:  first = 2;  last = 9;   break;
    case GIGE_BITMODE_9_2:  first = 3;  last = 10;  break;
    case GIGE_BITMODE_10_3: first = 4;  last = 11;  break;
    case GIGE_BITMODE_11_4: first = 5;  last = 12;  break;
    case GIGE_BITMODE_9_0:  first = 1;  last = 10;  break;
    case GIGE_BITMODE_10_1: first = 2;  last = 11;  break;
    case GIGE_BITMODE_11_2: first = 3;  last = 12;  break;
    case GIGE_BITMODE_11_0: first = 1;  last = 12;  break;
    default:
        bit_mode_str = QApplication::translate("overlay", "Full 10");
        break;
    }

    if (bit_mode_str.isEmpty())
    {
        bit_mode_str = QString("%1:%2").arg(last).arg(first);
    }

    return bit_mode_str;
}

const QString fiOverlay::get_sensor_opt(CameraMDI *camera_ptr)
{
    QString opt_str;

    switch (camera_ptr->read_camera_cache(GIGE_REG_SENSOR_OPTIONS) & (GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK))
    {
    case 0x101:
        opt_str = QApplication::translate("overlay", "Normal");
        break;
    case 0x102:
        opt_str = QApplication::translate("overlay", "2X Binning");
        break;
    case 0x104:
        opt_str = QApplication::translate("overlay", "4X Binning");
        break;
    case 0x201:
        opt_str = QApplication::translate("overlay", "2X Subsampling");
        break;
    case 0x401:
        opt_str = QApplication::translate("overlay", "4X Subsampling");
        break;
    case 0x202:
        opt_str = QApplication::translate("overlay", "2X Bin & 2X Sub");
        break;
    default:
        opt_str = QApplication::translate("overlay", "Unknown");
        break;
    }

    return opt_str;
}

/*!
 * @brief Return a string with the time formatted as SMPTE time: HH:MM:SS:<frame>
 *
 * @param frame_no          The frame number we're interested in.
 * @param trigger_frame     The frame number of the trigger frame.
 * @param frame_rate        The frame rate, so we can calculate time.
 * @return String representing SMPTE formatted time.
 */
const QString fiOverlay::get_smpte_time(int frame_no, int trigger_frame, uint32_t frame_rate)
{
    bool        neg_sign;
    int         fr_hour;
    int         fr_min;
    int         fr_sec;
    int         frames;
    double      frame_time;
    QString     field_str;
    QLatin1Char zero_fill('0');

    frame_time = (double)(frame_no - trigger_frame);
    frame_time = frame_time / (double)frame_rate;

    if (frame_time < 0.0)
    {
        neg_sign = true;
        frame_time = abs(frame_time);
    }
    else
    {
        neg_sign = false;
    }

    fr_hour = (int)(frame_time / (60.0 * 60.0));
    frame_time = frame_time - fr_hour;
    fr_min = (int)(frame_time / 60.0);
    frame_time = frame_time - fr_min;
    fr_sec = (int)frame_time;
//    frame_time = frame_time - fr_sec;
    frames = abs(frame_no - trigger_frame) % (int)frame_rate;

    field_str = QString("%1:%2:%3:%4").arg(fr_hour, 2, 10, zero_fill).arg(fr_min, 2, 10, zero_fill).arg(fr_sec, 2, 10, zero_fill).arg(frames);
    if (neg_sign)
        field_str.prepend(QLatin1Char('-'));

    return field_str;
}

/*!
 * @brief Return a string with the time formatted as a floating value. Depending on how you
 *        pass in the arguments, you can get a relative time or total capture duration.
 *
 * @param frame_no          The frame number we're interested in.
 * @param trigger_frame     The frame number of the trigger frame.
 * @param frame_rate        The frame rate, so we can calculate time.
 * @return String representing time, in seconds.
 */
const QString fiOverlay::get_time(int frame_no, int trigger_frame, uint32_t frame_rate, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr)
{
    double delta_time(0.0);
    double frame_time(0.0);
    double trigger_time(0.0);

    if (camera_ptr != nullptr)
    {
        // Calculate the two frames, and then the difference using the camera data.
        frame_time = camera_ptr->calc_frame_time(frame_no);
        trigger_time = camera_ptr->calc_frame_time(trigger_frame);

        // Sanity check the values. We may not have the information required (yet) to perform the
        // time calculation.
        if (qFuzzyIsNull(frame_time) || qFuzzyIsNull(trigger_time))
        {
            // If the timer isn't already active, start it.
            if (!delayed_update.isActive())
            {
                // Start our timer.
                delayed_update.start();
            }
        }

        // Calculate the difference in times.
        delta_time = frame_time - trigger_time;
    }
    else if (metadata != nullptr)
    {
        uint32_t sbs, micro_sec;

        // Calculate the two frames, and then the difference using the metadata pointer.
        sbs = metadata->getPfmdTime(frame_no, micro_sec);
        frame_time = ((double)sbs) + (((double)micro_sec) / 1000000.0);
        sbs = metadata->getPfmdTime(trigger_frame, micro_sec);
        trigger_time = ((double)sbs) + (((double)micro_sec) / 1000000.0);
        delta_time = frame_time - trigger_time;
    }
    else
    {
        // We have neither a camera pointer or a metadata pointer, so just do the calculations assuming
        // no discontinuities.
        frame_time = (double)(frame_no - trigger_frame);
        delta_time = frame_time / (double)frame_rate;
    }

    return QString("%1").arg(delta_time, 8, 'f', 6);
}

const QString fiOverlay::get_time0(CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr)
{
    QString     field_str;
    double      trig_frame_start = 0.0;
    double      trig_event = 0.0;
    int         trigger_frame = -1;

    if (metadata != nullptr)
    {
        uint32_t    trigger_sbs, trigger_micro_sec;

        // Get the trigger frame.
        trigger_frame = (int)metadata->getTriggerFrame();

        // Get the trigger event time.
        trigger_sbs = metadata->getTimestamp(trigger_micro_sec);
        trig_event = (double)trigger_sbs + (((double)trigger_micro_sec) / 1000000.0);

        // Do we have the per-frame metadata for the trigger frame?
        if ((trigger_frame >= (int)metadata->getStartFrame()) && (trigger_frame <= (int)metadata->getEndFrame()))
        {
            // Read the information from the metadata.
            trigger_sbs = metadata->getPfmdTime((uint32_t)trigger_frame - metadata->getStartFrame(), trigger_micro_sec);
            trig_frame_start = (double)trigger_sbs + (((double)trigger_micro_sec) / 1000000.0);
        }
        else
        {
            uint32_t    frame_sbs, frame_micro_sec;
            double      frame_start;
            double      rel_frame_interval;

            // Get the relative time between this frame and the trigger frame.
            rel_frame_interval = (double)(trigger_frame - (int)metadata->getStartFrame());
            rel_frame_interval = rel_frame_interval / (double)info_ptr->get_fps();

            // Read the frame time for the start frame (i.e. frame 0).
            frame_sbs = metadata->getPfmdTime(0, frame_micro_sec);
            frame_start = (double)frame_sbs + (((double)frame_micro_sec) / 1000000.0);

            // Calculate the trigger frame start time.
            trig_frame_start = frame_start + rel_frame_interval;
        }
    }
    else if (camera_ptr != nullptr)
    {
        // Read the information we need from our camera object.
        trigger_frame = (int)camera_ptr->read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
        trig_frame_start = camera_ptr->get_trig_frame_start();
        trig_event = camera_ptr->get_trigger_event();
    }

    // Do we have the data we need?
    if (trigger_frame != -1)
    {
        double offset_val = trig_event - trig_frame_start;
        field_str = QString("%1").arg(offset_val, 8, 'f', 6);
    }

    return field_str;
}

void fiOverlay::load_settings(QSettings& settings, int index)
{
    int i;
    int alpha;
    QList<QVariant> v_list;

    settings.beginGroup(QString("Overlay.%1").arg(index));
    pinned = settings.value(QStringLiteral("pinned"), false).toBool();
    pin_to_top_border = settings.value(QStringLiteral("pin_to_top_border"), true).toBool();

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "index:" << index << "pinned:" << pinned << "pin_to_top_border:" << pin_to_top_border;

    location = settings.value(QStringLiteral("location")).toPointF();
    height = settings.value(QStringLiteral("height")).toReal();
    width = settings.value(QStringLiteral("width")).toReal();
    drawing_height = settings.value(QStringLiteral("drawing_height")).toReal();
    drawing_width = settings.value(QStringLiteral("drawing_width")).toReal();
    overlay_color = QColor(settings.value(QStringLiteral("overlay_color")).toUInt());
    alpha = settings.value(QStringLiteral("overlay_color_alpha"), 255).toInt();
    overlay_color.setAlpha(alpha);
    text_color = QColor(settings.value(QStringLiteral("text_color")).toUInt());
    alpha = settings.value(QStringLiteral("text_color_alpha"), 255).toInt();
    text_color.setAlpha(alpha);
    item_font.fromString(settings.value(QStringLiteral("item_font")).toString());

    if (settings.value(QStringLiteral("outline_pen")).toBool())
    {
        outline_pen = QPen(QColor(Qt::black));
        outline_pen.setWidth(0);
    }
    else
    {
        outline_pen = QPen(Qt::NoPen);
    }
    v_list = settings.value(QStringLiteral("overlay_item")).toList();
    for (i = 0; i < v_list.count(); ++i)
        overlay_item.append(v_list.at(i).toInt());

    // Did the user save off a relative frame value already?
    user_ftime_text = settings.value(QStringLiteral("user_ftime_text")).toString();
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "user_ftime_text:" << user_ftime_text << "isEmpty()=" << user_ftime_text.isEmpty();

    // If not provide the default one.
    if (user_ftime_text.isEmpty())
    {
        user_ftime_text = QString("%1").arg(rel_frame_num_default);
    }

    // Pull out any custom text that may have existed before.
    user_text = settings.value(QStringLiteral("user_text")).toString();

    back_image = settings.value(QStringLiteral("overlay_back_image")).value<QPixmap>();
    settings.endGroup();

    // Adjust ourselves as best we can.
    setPos(location);

    // Remove the existing items. We only expect a single item (the place holder).
    remove_items();

    // Do we have something to delete now?
    action_delete_image->setEnabled(!back_image.isNull());
}

/*!
 * @brief Required method when creating a QGraphicsItem class.
 *
 * The drop shadow effect will be created by drawing a filled, rounded corner rectangle with a gradient fill.
 * Then on top of this will be drawn  filled, rounded corner rectangle, filled with a solid color, and offset
 * such that the gradient filled box is only visible below for a few pixels on two edges.
 *
 * The total box size is width by height. So the top box will start at (0,0) and go to
 * ( width-shadowThickness, height-shadowThickness ), while the under box will be offset,
 * and start at ( shadowThickness+0, shadowThickness+0 ) and go to (width, height).
 *
 * @param painter   Pointer to our paint object.
 * @param option    [UNUSED] Style options.
 * @param widget    [UNUSED] Pointer to the parent widget.
 */
void fiOverlay::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    QRectF  draw_rect;

    Q_UNUSED(option)
    Q_UNUSED(widget)

    // Are we pinned?
    if (isPinned())
    {
        // The drawing rectangle is the bounding rectangle.
        draw_rect = boundingRect();
    }
    else
    {
        // Our drawing rectangle is the adjusted drawing width/height.
        draw_rect = QRectF(CORNER_BUFFER, CORNER_BUFFER, drawing_width - CORNER_BUFFER, drawing_height - CORNER_BUFFER);
    }

    // Setup our brush.
    QBrush rect_brush(overlay_color, Qt::SolidPattern);
    painter->setPen(Qt::NoPen);
    painter->setBrush(rect_brush);

    // Draw a "normal" rectangle.
    painter->drawRect(draw_rect);

    // Do we have a pixmap?
    if (!back_image.isNull())
    {
        // Draw our pixmap too.
        painter->drawPixmap(draw_rect, back_image, back_image.rect());
    }
}

void fiOverlay::remove_items()
{
    int i;
    QList<QGraphicsItem *> kid_items;

    // Remove the existing items. We only expect a single item (the place holder).
    kid_items = flow_widget->childItems();
    for (i = 0; i < kid_items.count(); ++i)
    {
        kid_items.at(i)->setParentItem(nullptr);
        if (scene() != nullptr)
            scene()->removeItem(kid_items.at(i));
        else
            delete kid_items.at(i);

        // Was this the place holder?
        if (kid_items.at(i) == place_holder)
            place_holder = nullptr;
    }

    // Remove the layout item.
    flow_widget->setLayout(nullptr);

    // Let folks know we need data.
    emit need_update();
}

void fiOverlay::save_settings(QSettings& settings, int index)
{
    int             i;
    QList<QVariant> v_list;

    settings.beginGroup(QString("Overlay.%1").arg(index));
    settings.setValue(QStringLiteral("pinned"), pinned);
    settings.setValue(QStringLiteral("pin_to_top_border"), pin_to_top_border);

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "index:" << index << "pinned:" << pinned << "pin_to_top_border:" << pin_to_top_border;

    settings.setValue(QStringLiteral("location"), location);
    settings.setValue(QStringLiteral("height"), height);
    settings.setValue(QStringLiteral("width"), width);
    settings.setValue(QStringLiteral("drawing_height"), drawing_height);
    settings.setValue(QStringLiteral("drawing_width"), drawing_width);
    settings.setValue(QStringLiteral("overlay_color"), overlay_color.rgb());
    settings.setValue(QStringLiteral("overlay_color_alpha"), overlay_color.alpha());
    settings.setValue(QStringLiteral("text_color"), text_color.rgb());
    settings.setValue(QStringLiteral("text_color_alpha"), text_color.alpha());
    settings.setValue(QStringLiteral("item_font"), item_font.toString());
    settings.setValue(QStringLiteral("outline_pen"), (outline_pen.style() != Qt::NoPen));

    for (i = 0; i < overlay_item.count(); ++i)
        v_list.append(overlay_item.at(i));

    settings.setValue(QStringLiteral("overlay_item"), v_list);
    settings.setValue(QStringLiteral("user_ftime_text"), user_ftime_text);
    settings.setValue(QStringLiteral("user_text"), user_text);
    settings.setValue(QStringLiteral("overlay_back_image"), back_image);

    settings.endGroup();
}

/**
  * A default setting when the overlay was created for the relative frame
  * number is provided by the overlay creator.  This will serve as an initializer
  * for the relative frame time item should it be selected.  If there is a setting
  * in the user's space already it will override this value.
  */
void fiOverlay::set_default_rel_frame_info(int rel_frame_num, int rel_frame_min_to_set, int rel_frame_max_to_set)
{
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rel_frame_num:" << rel_frame_num << "rel_frame_min_to_set:" << rel_frame_min_to_set << "rel_frame_max_to_set:" << rel_frame_max_to_set;
    rel_frame_num_default = rel_frame_num;
    rel_frame_min = rel_frame_min_to_set;
    rel_frame_max = rel_frame_max_to_set;
}

/**
  * This scene event filter has been registered with all four corner grabber items.
  * When called, a pointer to the sending item is provided along with a generic
  * event.  A dynamic_cast is used to determine if the event type is one of the events
  * we are interrested in.
  */
bool fiOverlay::sceneEventFilter(QGraphicsItem *watched, QEvent *event)
{
    CornerGrabber *corner = dynamic_cast<CornerGrabber *>(watched);
    if (corner == nullptr)
        return false; // not expected to get here

    QGraphicsSceneMouseEvent *mevent = dynamic_cast<QGraphicsSceneMouseEvent *>(event);
    if (mevent == nullptr)
    {
        // This is not one of the mouse events we are interested in.
        return false;
    }

    switch (event->type())
    {
    case QEvent::GraphicsSceneMousePress:
        // If the mouse went down, record the x,y coords of the press, record it inside the corner object.
        corner->setMouseState(CornerGrabber::kMouseDown);
        corner->mouseDownX = mevent->pos().x();
        corner->mouseDownY = mevent->pos().y();
        break;
    case QEvent::GraphicsSceneMouseRelease:
        corner->setMouseState(CornerGrabber::kMouseReleased);
        break;
    case QEvent::GraphicsSceneMouseMove:
        corner->setMouseState(CornerGrabber::kMouseMoving);
        break;
    default:
        // We don't care about the rest of the events.
        return false;
    }

    if (corner->getMouseState() == CornerGrabber::kMouseMoving)
    {
        qreal x = mevent->pos().x();
        qreal y = mevent->pos().y();

        // Depending on which corner has been grabbed, we want to move the position
        // of the item as it grows/shrinks accordingly. so we need to either add
        // or subtract the offsets based on which corner this is.
        int XaxisSign = 0;
        int YaxisSign = 0;

        switch (corner->getCorner())
        {
        case 0:
            XaxisSign = 1;
            YaxisSign = 1;
            break;
        case 1:
            XaxisSign = -1;
            YaxisSign = 1;
            break;
        case 2:
            XaxisSign = -1;
            YaxisSign = -1;
            break;
        case 3:
            XaxisSign = 1;
            YaxisSign = -1;
            break;
        default:
            // Unexpected.
            break;
        }

        // If the mouse is being dragged, calculate a new size and also re-position
        // the box to give the appearance of dragging the corner out/in to resize the box
        int xMoved = (int)(corner->mouseDownX - x);
        int yMoved = (int)(corner->mouseDownY - y);

        int newWidth = (int)width + (XaxisSign * xMoved);
        if (newWidth < (2 * CORNER_BUFFER))
            newWidth  = (2 * CORNER_BUFFER);

        int newHeight = (int)height + (YaxisSign * yMoved);
        if (newHeight < (2 * CORNER_BUFFER))
            newHeight = (2 * CORNER_BUFFER);

        int deltaWidth = newWidth - (int)width;
        int deltaHeight = newHeight - (int)height;
        adjustSize(deltaWidth, deltaHeight);

        deltaWidth *= (-1);
        deltaHeight *= (-1);

        if (corner->getCorner() == 0)
        {
            int newXpos = (int)this->pos().x() + deltaWidth;
            int newYpos = (int)this->pos().y() + deltaHeight;
            setPos(newXpos, newYpos);
        }
        else if (corner->getCorner() == 1)
        {
            int newYpos = (int)this->pos().y() + deltaHeight;
            setPos(this->pos().x(), newYpos);
        }
        else if (corner->getCorner() == 3)
        {
            int newXpos = (int)this->pos().x() + deltaWidth;
            setPos(newXpos, this->pos().y());
        }

        setCornerPositions();
        update();
    }

    return true;// true => do not send event to watched - we are finished with this event
}

void fiOverlay::setCornerPositions()
{
    qreal c_wd = corners[0]->boundingRect().width();
    qreal c_ht = corners[0]->boundingRect().height();

    corners[0]->setPos(CORNER_BUFFER - c_wd, CORNER_BUFFER - c_ht);
    corners[1]->setPos(drawing_width, CORNER_BUFFER - c_ht);
    corners[2]->setPos(drawing_width, drawing_height);
    corners[3]->setPos(CORNER_BUFFER - c_wd, drawing_height);
}

void fiOverlay::setGeometry(const QRectF& rect)
{
    setPos(rect.topLeft());
}

void fiOverlay::setPinned(bool is_pinned, bool top_border_pin, qreal requested_width)
{
    pinned = is_pinned;
    pin_to_top_border = top_border_pin;

    if (pinned)
    {
        int new_height;
        int new_width = (int)requested_width;

        // Make sure the requested width is even.
        if ((new_width & 1) == 1)
            new_width++;

        // Adjust the widgets location.
        flow_widget->setPos(0, 0);

        // Adjust the width as necessary.
        if (new_width > 0)
            pinned_width = new_width;
        else
            pinned_width = width;

        // Figure out the height.
        new_height = (int)(static_cast<FlowLayout *>(flow_widget->layout()))->minSize(QSizeF(pinned_width, 0)).height() - CORNER_BUFFER;

        // Make sure we meet our "minimum" standards!
        if (new_height < CORNER_BUFFER)
            new_height = CORNER_BUFFER;

        // Make sure the new height is even.
        if ((new_height & 1) == 1)
            new_height++;
        pinned_height = new_height;

        // Resize our layout widget.
        flow_widget->resize(pinned_width, pinned_height);

        // Adjust our minimum height information.
        setMinimumHeight(pinned_height);
    }
    else
    {
        // Not pinned, so reset our widgets location.
        flow_widget->setPos(CORNER_BUFFER, CORNER_BUFFER);

        // Reset our layout for the "normal" region.
        adjustSize(0, 0);
    }
}

void fiOverlay::setPos(const QPointF& pos)
{
    location = pos;
    QGraphicsItem::setPos(pos);
}

void fiOverlay::setPos(qreal x, qreal y)
{
    location.setX(x);
    location.setY(y);
    QGraphicsItem::setPos(x, y);
}

QSizeF fiOverlay::sizeHint(Qt::SizeHint which, const QSizeF& constraint) const
{
    Q_UNUSED(which)
    Q_UNUSED(constraint)
    return boundingRect().size();
}

void fiOverlay::update_items(uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr)
{
    int     item_id;
    QList<QGraphicsItem *> kid_items;

    // Remove the existing items.
    kid_items = flow_widget->childItems();

    // Do we already have items
    if (kid_items.count() == 0)
    {
        // Allocate our new layout.
        FlowLayout *flow_lay = new FlowLayout;
        QFontMetrics font_metrics(item_font);
        flow_lay->setContentsMargins(9, 9, 9 + CORNER_BUFFER, 9 + CORNER_BUFFER);
        flow_lay->setSpacing(Qt::Horizontal, 2 * font_metrics.averageCharWidth());

        // See if the user selected anything.
        for (item_id = 0; item_id < overlay_item.count(); ++item_id)
        {
            // Add our item.
            add_item(flow_lay, overlay_item.at(item_id), frame_num, info_ptr, metadata, camera_ptr);
        }

        // Are there any elements in our layout?
        if (flow_lay->count() == 0)
        {
            // We need to add a place holder into the layout. Do we already
            // have the place holder object?
            if (place_holder == nullptr)
            {
                if (back_image.isNull())
                    place_holder = new fiLayoutText(QApplication::translate("overlay", "<add items>"), flow_widget);
                else
                    place_holder = new fiLayoutText(QApplication::translate("overlay", "  "), flow_widget);
                place_holder->setFont(item_font);
                place_holder->setBrush(QBrush(text_color));
            }

            // Put the place holder into the layout.
            flow_lay->addItem(place_holder);
        }

        // Set the new layout.
        flow_widget->setLayout(flow_lay);
    }
    else
    {
        // Loop through all the items and update them.
        for (item_id = 0; item_id < kid_items.count(); ++item_id)
        {
            fiLayoutText    *text_item = qgraphicsitem_cast<fiLayoutText *>(kid_items.at(item_id));
            if (text_item != nullptr)
            {
                update_item(text_item, text_item->id, frame_num, info_ptr, metadata, camera_ptr);
            }
        }

        // Invalidate our layout.
        flow_widget->layout()->invalidate();
    }

    // Adjust our size.
    setPinned(pinned, pin_to_top_border, pinned_width);
}

/**
  * There is a possible change in the relative frame clip information.
  */
void fiOverlay::update_rel_frame_range(int rel_frame_min_to_set, int rel_frame_max_to_set)
{
    QString user_ftime_changed;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rel_frame_min_to_set:" << rel_frame_min_to_set  << "rel_frame_max_to_set:" << rel_frame_max_to_set;
    rel_frame_min = rel_frame_min_to_set;
    rel_frame_max = rel_frame_max_to_set;

    // Do we have a default relative frame?
    if (rel_frame_num_default < rel_frame_min)
        rel_frame_num_default = rel_frame_min;
    else if (rel_frame_num_default > rel_frame_max)
        rel_frame_num_default = rel_frame_max;

    // Do we have an item that is selected in the overlay?
    if (!user_ftime_text.isEmpty())
    {
        int user_ftime = user_ftime_text.toInt();
        if (user_ftime < rel_frame_min)
        {
            user_ftime = rel_frame_min;
            user_ftime_changed = QString("%1").arg(user_ftime);
        }
        else if (user_ftime > rel_frame_max)
        {
            user_ftime = rel_frame_max;
            user_ftime_changed = QString("%1").arg(user_ftime);
        }
        // Do we need to update any UI data?
        if (!user_ftime_changed.isEmpty())
        {
            user_ftime_text = user_ftime_changed;
        }
    }
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

void fiOverlay::contextMenuEvent(QGraphicsSceneContextMenuEvent *event)
{
    QMenu   context_menu;

    // Build up our menu.
    context_menu.addAction(action_select_item);
    context_menu.addSeparator();
    context_menu.addAction(action_pin_top);
    context_menu.addAction(action_pin_bottom);
    context_menu.addSeparator();
    context_menu.addAction(action_background);
    context_menu.addAction(action_image);
    context_menu.addAction(action_delete_image);
    context_menu.addSeparator();
    context_menu.addAction(action_item_font);
    context_menu.addAction(action_item_foreground);
    context_menu.addAction(action_item_outline);
    context_menu.addSeparator();
    context_menu.addAction(action_delete);

    // Pop it up.
    context_menu.exec(event->screenPos());
}

// create the corner grabbers
void fiOverlay::hoverEnterEvent(QGraphicsSceneHoverEvent *)
{
    if (!isPinned())
    {
        corners[0] = new CornerGrabber(this, 0);
        corners[1] = new CornerGrabber(this, 1);
        corners[2] = new CornerGrabber(this, 2);
        corners[3] = new CornerGrabber(this, 3);

        corners[0]->installSceneEventFilter(this);
        corners[1]->installSceneEventFilter(this);
        corners[2]->installSceneEventFilter(this);
        corners[3]->installSceneEventFilter(this);

        setCornerPositions();
    }

    // When the cursor enters our "space" make sure we have an arrow pointer.
    setCursor(Qt::ArrowCursor);
}

// remove the corner grabbers
void fiOverlay::hoverLeaveEvent(QGraphicsSceneHoverEvent *)
{
    if (!isPinned() && (corners[0] != nullptr))
    {
        corners[0]->setParentItem(nullptr);
        corners[1]->setParentItem(nullptr);
        corners[2]->setParentItem(nullptr);
        corners[3]->setParentItem(nullptr);

        delete corners[ 0 ];
        corners[0] = nullptr;
        delete corners[ 1 ];
        corners[1] = nullptr;
        delete corners[ 2 ];
        corners[2] = nullptr;
        delete corners[ 3 ];
        corners[3] = nullptr;
    }
}

// for supporting moving the box across the scene
void fiOverlay::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    // We only deal with left mouse button events. The move event
    // reports this info in the "buttonS" call, not "button".
    if (event->buttons() != Qt::LeftButton)
        return;

    // Get the current mouse position.
    drag_pos = event->pos();

    // Adjust our location.
    location += (drag_pos - drag_start);
    setPos(location);
}

// for supporting moving the box across the scene
void fiOverlay::mousePressEvent(QGraphicsSceneMouseEvent *event)
{
    // We only deal with left mouse button events.
    if (event->button() != Qt::LeftButton)
        return;

    // Accept the event.
    event->setAccepted(true);

    // Remember where it all started.
    drag_pos = drag_start = event->pos();
}

// for supporting moving the box across the scene
void fiOverlay::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    int new_x;
    int new_y;

    // We only deal with left mouse button events.
    if (event->button() != Qt::LeftButton)
        return;

    // Accept the event.
    event->setAccepted(true);

    // Did we move at all?
    if (event->pos() == drag_pos)
        return;

    // Get our position as an integer.
    new_x = static_cast<int>(location.x());
    new_y = static_cast<int>(location.y());

    // Make sure the coordinates are even.
    if ((new_x & 1) == 1)
        new_x++;
    if ((new_y & 1) == 1)
        new_y++;

    // Set the location, aligned to our grid.
    location.setX((new_x / GRID_SPACE) * GRID_SPACE);
    location.setY((new_y / GRID_SPACE) * GRID_SPACE);
    setPos(location);

    // We are no longer "pinned".
    setPinned(false, false);

    // Let folks know that we've moved.
    emit overlay_moved();
}

//--------------------------------------------------------------------
// Slots
//--------------------------------------------------------------------

void fiOverlay::do_change_font()
{
    int     i;
    bool    ok;

    // Have the user select a new font for the items.
    QFont font = QFontDialog::getFont(&ok, item_font, nullptr, QApplication::translate("overlay", "Select Font"));
    if (!ok)
        return;

    // Save the new font for later use.
    item_font = font;

    QFontMetrics font_metrics(item_font);
    FlowLayout *flow_lay = static_cast<FlowLayout *>(flow_widget->layout());
    if (flow_lay == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Not using a FlowLayout?!";
        return;
    }
    flow_lay->setSpacing(Qt::Horizontal, 2 * font_metrics.averageCharWidth());

    // Get our list of child items so we can change their font.
    QList<QGraphicsItem *> kid_items = flow_widget->childItems();

    // Loop through the existing items and change the font.
    for (i = 0; i < kid_items.count(); ++i)
    {
        fiLayoutText    *text_item = qgraphicsitem_cast<fiLayoutText *>(kid_items.at(i));
        if (text_item != nullptr)
        {
            qreal   new_width;
            qreal   new_height;

            // Switch to the new font.
            text_item->setFont(item_font);

            // Update the item's preferred and minimum sizes (they are the same).
            new_width = text_item->boundingRect().width();
            new_height = text_item->boundingRect().height();
            text_item->setPreferredSize(new_width, new_height);
            text_item->setMinimumSize(new_width, new_height);
        }
    }

    // Re-layout.
    flow_lay->invalidate();
    adjustSize(0, 0);

    // Are we pinned?
    if (pinned)
    {
        // Have our parent re-pin us to force a re-layout of the parents widget.
        if (qFuzzyIsNull(pos().y()))
            emit pin_to_top();
        else
            emit pin_to_bottom();
    }
}

void fiOverlay::do_change_item_color()
{
    int     i;

    // Why are we here?
    QAction *action_ptr = qobject_cast<QAction *>(sender());
    if (action_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null sender?";
        return;
    }
    else if (action_ptr == action_item_foreground)
    {
        const QColor color = QColorDialog::getColor(text_color, MainWindow::GetInstance().get(), QApplication::translate("overlay", "Select Text Color"), QColorDialog::ShowAlphaChannel);
        if (!color.isValid())
            return;

        text_color = color;
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected action!";
        return;
    }

    // Get our list of child items so we can change their font.
    QList<QGraphicsItem *> kid_items = flow_widget->childItems();

    // Loop through the existing items and change the font.
    for (i = 0; i < kid_items.count(); ++i)
    {
        fiLayoutText    *text_item = qgraphicsitem_cast<fiLayoutText *>(kid_items.at(i));
        if (text_item != nullptr)
            text_item->setBrush(QBrush(text_color));
    }

    update();
}

void fiOverlay::do_delayed_update()
{
    // Check to see if anything has attached to our signal.
    if (isSignalConnected(QMetaMethod::fromSignal(&fiOverlay::need_update)))
    {
        // Let folks know we need an update.
        emit need_update();
    }
    else
    {
        // Try again later.
        delayed_update.start();
    }
}

void fiOverlay::do_remove_image()
{
    // Remove the background image.
    back_image = QPixmap();
    action_delete_image->setEnabled(false);

    // Do we have our placeholder?
    if (place_holder != nullptr)
    {
        // Remove the place holder.
        remove_items();
    }

    // Refresh the overlay.
    update();
}

void fiOverlay::do_select_item()
{
    int     item_id;

    // Ask the user what to display.
    SelectItemDlg   dlg;

    // Initialize the dialog.
    for (item_id = 0; item_id < overlay_item.count(); ++item_id)
        dlg.select_item(overlay_item.at(item_id));

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "user_ftime_text.isEmpty()" << user_ftime_text.isEmpty();

    // Is there custom frame relative time?
    if (!user_ftime_text.isEmpty())
    {
        // There is a user installed user frame time.
        dlg.set_custom_ftime_frame(user_ftime_text, rel_frame_min, rel_frame_max);
        dlg.select_item(SelectItemDlg::REL_FTIME);
    }
    else
    {
        // We want to provide a relative frame based on the one set for default purposes.
        QString rel_frame_str = QString("%1").arg(rel_frame_num_default);
        dlg.set_custom_ftime_frame(rel_frame_str, rel_frame_min, rel_frame_max);
    }

    // Is there custom text?
    if (!user_text.isEmpty())
    {
        dlg.set_custom_text(user_text);
        dlg.select_item(SelectItemDlg::CUSTOM_TEXT);
    }

    // Pop up the dialog.
    if (dlg.exec() == QDialog::Rejected)
        return;

    // Get the custom text, if any.
    user_ftime_text = dlg.custom_ftime_frame();

    // Get the custom text, if any.
    user_text = dlg.custom_text();

    // See if the user selected anything.
    overlay_item.clear();
    overlay_item = dlg.get_visible_list();

    // Remove the existing items.
    remove_items();
}

void fiOverlay::do_set_background_color()
{
    const QColor color = QColorDialog::getColor(overlay_color, MainWindow::GetInstance().get(), QApplication::translate("overlay", "Select Background Color"), QColorDialog::ShowAlphaChannel);
    if (color.isValid())
    {
        overlay_color = color;
        update();
    }
}

void fiOverlay::do_set_image()
{
    QString filename;
    QString last_path;
    QSettings   settings;
    QFileInfo   file_info;

    // Initialize the default path based on the saved settings.
    settings.beginGroup("PathInfo");
    last_path = settings.value("overlay_image_path").toString();
    settings.endGroup();

    // If the path is empty, try the default download path.
    if (last_path.isEmpty())
        last_path = SaveDialog::loadPath(0);

    // If that's still empty, how about the system documents path?
    if (last_path.isEmpty())
        last_path = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0);

    // Ask the user for a configuration file.
    filename = QFileDialog::getOpenFileName(nullptr, tr("Select An Image to Load"),
                                            last_path, tr("Image File (*.bmp *.jpg *.png *.tif)"));
    if (filename.isEmpty())
        return;

    // Get the path for the configuration file.
    file_info.setFile(filename);

    // Make sure the path ends with a '/' seperator.
    last_path = file_info.canonicalPath();
    if (!last_path.endsWith(QLatin1Char('/')))
        last_path.append(QLatin1Char('/'));

    // Save the path for the configuration files.
    settings.beginGroup("PathInfo");
    settings.setValue("overlay_image_path", last_path);
    settings.endGroup();

    // Load up the image.
    back_image.load(filename);

    // Do we have something to delete now?
    action_delete_image->setEnabled(!back_image.isNull());

    // Do we just have the place holder?
    if (place_holder != nullptr)
        remove_items();

    // Update the overlay.
    update();
}

void fiOverlay::do_text_outline()
{
    int     i;

    // Create a pen if the user wants outlines, otherwise create a null pen.
    if (action_item_outline->isChecked())
    {
        outline_pen = QPen(QColor(Qt::black));
        outline_pen.setWidth(0);
    }
    else
    {
        outline_pen = QPen(Qt::NoPen);
    }

    // Get our list of child items so we can change their font.
    QList<QGraphicsItem *> kid_items = flow_widget->childItems();

    // Loop through the existing items and change the font.
    for (i = 0; i < kid_items.count(); ++i)
    {
        fiLayoutText    *text_item = qgraphicsitem_cast<fiLayoutText *>(kid_items.at(i));
        if (text_item != nullptr)
            text_item->setPen(outline_pen);
    }

    update();
}

