/*!
 * @file fi_overlay.h
 * @brief This is the header file for the fiOverlay class.
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_OVERLAY_H
#define FI_OVERLAY_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QAction>
#include <QBitArray>
#include <QGraphicsObject>
#include <QGraphicsLayoutItem>
#include <QGraphicsRectItem>
#include <QGraphicsTextItem>
#include <QGraphicsSceneHoverEvent>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsWidget>
#include <QColor>
#include <QFont>
#include <QPainter>
#include <QPen>
#include <QPixmap>
#include <QPointF>
#include <QSettings>
#include <QTimer>
#include <stdint.h>

#include "cornergrabber.h"
#include "select_item.h"

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class CameraInfo;
class CameraMDI;
class fiLayoutText;
class FlowLayout;
class PlaybackMetaInfo;
class QLabel;

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class fiOverlay : public QGraphicsObject, public QGraphicsLayoutItem
{
    Q_OBJECT
    Q_INTERFACES(QGraphicsLayoutItem)

public:
    fiOverlay(QGraphicsItem *parent = nullptr, const fiOverlay *copy_ptr = nullptr);
    ~fiOverlay();

    // Inherited from QGraphicsLayoutItem
    virtual void    setGeometry(const QRectF& geom);
    virtual QSizeF  sizeHint(Qt::SizeHint which, const QSizeF& constraint = QSizeF()) const;

    // Inherited from QGraphicsItem
    virtual QRectF  boundingRect() const;
    virtual void    paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget);

    QPointF get_location() const          { return location; }
    bool    isPinned()                    { return pinned; }
    bool    isPinnedAtTop()               { return pin_to_top_border; }

    void    load_settings(QSettings& settings, int index);
    void    save_settings(QSettings& settings, int index);
    void    set_default_rel_frame_info(int rel_frame_num, int rel_frame_min, int rel_frame_max);
    void    setPinned(bool is_pinned, bool top_border_pin, qreal requested_width = 0);
    void    setPos(const QPointF& pos);
    void    setPos(qreal x, qreal y);
    void    update_items(uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr);
    void    update_rel_frame_range(int rel_frame_min, int rel_frame_max);

public slots:
    void    do_select_item();

signals:
    void    need_update();
    void    pin_to_top();
    void    pin_to_bottom();
    void    remove_overlay();
    void    overlay_moved();

protected:
    virtual void contextMenuEvent(QGraphicsSceneContextMenuEvent *event);

    virtual void hoverEnterEvent(QGraphicsSceneHoverEvent *event);
    virtual void hoverLeaveEvent(QGraphicsSceneHoverEvent *event);

    virtual void mouseMoveEvent(QGraphicsSceneMouseEvent *event);
    virtual void mousePressEvent(QGraphicsSceneMouseEvent *event);
    virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent *event);

    virtual bool sceneEventFilter(QGraphicsItem *watched, QEvent *event);

private slots:
    void    do_change_font();
    void    do_change_item_color();
    void    do_delayed_update();
    void    do_remove_image();
    void    do_set_background_color();
    void    do_set_image();
    void    do_text_outline();

private:
    void    add_item(FlowLayout *layout, int item_id, uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr, fiLayoutText *item_ptr = nullptr);
    inline void update_item(fiLayoutText *item_ptr, int item_id, uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr);
    void    adjustSize(int x, int y);
    void    remove_items();

    const QString   get_bit_mode(uint32_t bit_mode);
    const QString   get_sensor_opt(CameraMDI *camera_ptr);
    const QString   get_smpte_time(int frame_no, int trigger_frame, uint32_t frame_rate);
    const QString   get_time(int frame_no, int trigger_frame, uint32_t frame_rate, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr);
    const QString   get_time0(CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr);

    void    setCornerPositions();

private:
    bool    pinned;                     //!< True if we're pinned to the top or bottom edge.
    bool    pin_to_top_border;          //!< True if user pinned to top border, assume bottom if false and pinned is true.
    int     rel_frame_num_default;      //!< If one is not listed in settings this will be set when overlay was created.
    int     rel_frame_min;              //!< What is the rel_frame minimum value allowed.
    int     rel_frame_max;              //!< What is the rel_frame maximum value allowed.
    QPointF location;                   //!< Origin of the overlay "box".
    QPointF drag_start;                 //!< Where the mouse "down" event started a drag operation.
    QPointF drag_pos;                   //!< The mouse "move" position.
    qreal   width;                      //!< Width of the overlay "box", including the drag corners.
    qreal   height;                     //!< Height of the overlay "box", including the drag corners.
    qreal   pinned_width;
    qreal   pinned_height;
    qreal   drawing_width;              //!< Width of the overlay "box" only (no drag corners).
    qreal   drawing_height;             //!< Height of the overlay "box" only (no drag corners).
    CornerGrabber   *corners[ 4 ];      //!< Starting at x=0, y=0 and moving clockwise around the box.
    QGraphicsWidget *flow_widget;
    QTimer  delayed_update;             //!< Sometimes we need to request an update at a later time.

    QColor  overlay_color;
    QColor  text_color;
    QFont   item_font;
    QPen    outline_pen;
    QPixmap back_image;

    QAction *action_background;
    QAction *action_image;
    QAction *action_delete;
    QAction *action_delete_image;
    QAction *action_item_outline;
    QAction *action_item_font;
    QAction *action_item_foreground;
    QAction *action_pin_bottom;
    QAction *action_pin_top;
    QAction *action_select_item;

    fiLayoutText    *place_holder;
    QList<int>  overlay_item;
    QString     user_ftime_text;            //!< The text custom frame for "frame" relative time.
    QString     user_text;                  //!< The custom text string.
};

void fiOverlay::update_item(fiLayoutText *item_ptr, int item_id, uint32_t frame_num, CameraInfo *info_ptr, PlaybackMetaInfo *metadata, CameraMDI *camera_ptr)
{
    add_item(nullptr, item_id, frame_num, info_ptr, metadata, camera_ptr, item_ptr);
}

#endif // FI_OVERLAY_H
