/*!
 * @file fi_pixmap.cpp
 * @brief This file contains the implementation of the fiPixmap class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QBuffer>
#include <QByteArray>
#include <QColor>
#include <QElapsedTimer>
#include <QFile>
#include <QFileInfo>
#include <QImage>
#include <qmath.h>
#include <QTemporaryFile>
#include <QTime>

#include "fi_pixmap.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"
#include "dcrawutil.h"

namespace FI_PIXMAP
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The maximum number of pixels we'll use to calculate our white balance gains.
const int   MAX_AUTOWB_METRICS          = 10;

//! The threshold for "bright" pixels. We don't want anything above this value.
//! This is about 80% of "white".
const int   MAX_AUTOWB_GREY             = 0xD0;

//! A structure used to track down pixels of interest for auto white balance.
typedef struct AUTOWB_METRICS
{
    int     grey;                   //!< The greyscale value of the pixel.
    int     x;                      //!< The column (x-coordinate) of this pixel.
    int     y;                      //!< The row (y-coordinate) of this pixel.
} AUTOWB_METRICS;

//! Cool White Flourescent illuminant.
static double AL5C_cwf_ccm[][4] =
{
    {-0.044966, 1.397403, -0.418906, 0.021502},
    {-0.041379, -0.267418, 1.228331, 0.039087},
    {-0.051119, 0.117922, -0.767874, 1.649952 }
};
static double X9HC_cwf_ccm[][4] =
{
    {-0.1645, 1.8674, -0.8518, 0.1489},
    {-0.1829, -0.2886, 1.1619, 0.3096},
    {-0.2041, 0.0790, -0.7577, 1.8828}
};

//! D65 illuminant.
static double AL5C_day_ccm[][4] =
{
    {-0.056408, 1.215392, -0.124575, -0.090818},
    {-0.049300, -0.132548, 1.307234, -0.174686},
    {-0.048630, 0.077666, -0.559067, 1.481401}
};
static double X9HC_day_ccm[][4] =
{
    {-0.045251, 1.535080, -0.331541, -0.158288},
    {-0.042715, -0.225976, 1.497115, -0.228423},
    {-0.052206, 0.030874, -0.683932, 1.705264}
};

//! Tungsten illuminant.
static double AL5C_tun_ccm[][4] =
{
    {-0.027663, 1.063022, 0.035831, -0.098853},
    {-0.021364, -0.396392, 1.294551, 0.101842},
    {-0.036942, 0.021363, -0.986430, 1.965067}
};
static double X9HC_tun_ccm[][4] =
{
    {-0.026387, 1.530243, -0.203461, -0.300395},
    {-0.022123, -0.482243, 1.557487, -0.053121},
    {-0.025158, -0.106007, -1.540637, 2.671802}
};

//! "Normal" 1:1:1 matrix.
static double norm_ccm[][4] =
{
    {0, 1.0, 0, 0},
    {0, 0, 1.0, 0},
    {0, 0, 0, 1.0}
};

}   // End of namespace FI_PIXMAP.

using namespace FI_PIXMAP;

/*!
 * @brief Class constructor.
 */
fiPixmap::fiPixmap() : QPixmap()
{
    ccm_updated = false;
    jpeg_quality = -1;
    gamma = 1.0;
    raw_rgb_image = nullptr;
    whitebal = GIGE_WHITE_BAL_DAYLIGHT;
    sensor_type = FI_SENSOR::SENSOR_TYPE_GENERIC_COLOR;
}

fiPixmap::fiPixmap(const QPixmap& src_pixmap) :
    QPixmap(src_pixmap)
{
    ccm_updated = false;
    jpeg_quality = -1;
    gamma = 1.0;
    raw_rgb_image = nullptr;
    whitebal = GIGE_WHITE_BAL_DAYLIGHT;
}

fiPixmap::fiPixmap(int width, int height) :
    QPixmap(width, height)
{
    ccm_updated = false;
    jpeg_quality = -1;
    gamma = 1.0;
    raw_rgb_image = nullptr;
    whitebal = GIGE_WHITE_BAL_DAYLIGHT;
}

/*!
 * @brief Class constructor.
 */
fiPixmap::fiPixmap(const QString& fileName, const char *format, Qt::ImageConversionFlags flags) :
    QPixmap(fileName, format, flags)
{
    ccm_updated = false;
    jpeg_quality = -1;
    gamma = 1.0;
    raw_rgb_image = nullptr;
    whitebal = GIGE_WHITE_BAL_DAYLIGHT;
}

/*!
 * @brief Class destructor.
 */
fiPixmap::~fiPixmap()
{
    if (raw_rgb_image != nullptr)
    {
        delete raw_rgb_image;
        raw_rgb_image = nullptr;
    }
}

/*!
 * @brief Loop through our image and pass each pixel through the related lookup table.
 *
 * @param [in]  lookup_r      Red channel lookup table.
 * @param [in]  lookup_g      Green channel lookup table.
 * @param [in]  lookup_b      Blue channel lookup table.
 */
void fiPixmap::apply_lookup(const uint8_t *lookup_r, const uint8_t *lookup_g, const uint8_t *lookup_b)
{
    int x;
    int y;
    int image_width;
    int image_height;
    QImage  my_image;
    QRgb *my_bits;

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Get the pixmap as an image.
    my_image = toImage();

    // Do we have a valid image?
    if (my_image.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "The toImage() function returned a null QImage!";
        return;
    }

    image_width = width();
    image_height = height();
    my_bits = reinterpret_cast<QRgb *>(my_image.bits());

    // Loop through all the rows.
// *INDENT-OFF*
#pragma omp parallel for collapse(2)
// *INDENT-ON*
    for (y = 0; y < image_height; ++y)
    {
        // Loop through all the columns.
        for (x = 0; x < image_width; ++x)
        {
            QColor  old_color;
            int r, g, b;
            int index;

            // Index into our pixel array.
            index = (y * image_width) + x;

            // Get a color object for this pixel.
            old_color = QColor(my_bits[index]);

            // Pass it through the lookup tables.
            r = old_color.red();
            if (lookup_r != nullptr)
                r = lookup_r[ r ];
            g = old_color.green();
            if (lookup_g != nullptr)
                g = lookup_g[ g ];
            b = old_color.blue();
            if (lookup_b != nullptr)
                b = lookup_b[ b ];

            // Modify our pixel.
            my_bits[index] = qRgb(r, g, b);
        }
    }

    // Update our image.
    convertFromImage(my_image);

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif
}

void fiPixmap::apply_matrix(const double ccm[][4])
{
    int rows, cols;
    QImage work_image;

    // Sanity check.
    if (raw_rgb_image == nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No stage3 image to work with!";
        return;
    }

    // Initialize our work image.
    work_image = QImage(raw_rgb_image->width(), raw_rgb_image->height(), raw_rgb_image->format());

    // Speed things up by using OpenMP to parrallel-ize the nested for loops.
// *INDENT-OFF*
#pragma omp parallel for collapse(2)
// *INDENT-ON*
    for (rows = 0; rows < work_image.height(); ++rows)
    {
        for (cols = 0; cols < work_image.width(); ++cols)
        {
            QRgb pixel = raw_rgb_image->pixel(cols, rows);
            int r_final, g_final, b_final;
            double r, g, b;
            double r_val, g_val, b_val;

            // Get our R, G, B values.
            r_val = (qRed(pixel) / 255.0) * r_wb_gain;
            g_val = (qGreen(pixel) / 255.0) * g_wb_gain;
            b_val = (qBlue(pixel) / 255.0) * b_wb_gain;

            // Extract the results.
            r = (ccm[0][0] * (1.0 / 255.0)) + (ccm[0][1] * r_val) + (ccm[0][2] * g_val) + (ccm[0][3] * b_val);
            g = (ccm[1][0] * (1.0 / 255.0)) + (ccm[1][1] * r_val) + (ccm[1][2] * g_val) + (ccm[1][3] * b_val);
            b = (ccm[2][0] * (1.0 / 255.0)) + (ccm[2][1] * r_val) + (ccm[2][2] * g_val) + (ccm[2][3] * b_val);

            // Are we supposed to apply gamma? We know the "knee" for sRGB, others we'll just
            // use a straight calculation.
            if (qFuzzyCompare(gamma, 2.2))
            {
                // Gamma for sRGB is defined to be ~2.2. But it's actually a multistep function.
                r = ((r > 0.0031308) ? (1.055 * pow(r, 1.0 / 2.4) - 0.055) : (12.92 * r)) * 255.0;
                g = ((g > 0.0031308) ? (1.055 * pow(g, 1.0 / 2.4) - 0.055) : (12.92 * g)) * 255.0;
                b = ((b > 0.0031308) ? (1.055 * pow(b, 1.0 / 2.4) - 0.055) : (12.92 * b)) * 255.0;
            }
            else
            {
                // Get our values into the proper range.
                r = pow(r, 1.0 / gamma) * 255.0;
                g = pow(g, 1.0 / gamma) * 255.0;
                b = pow(b, 1.0 / gamma) * 255.0;
            }

            // Convert to 0..255
            r_final = qFloor(r);
            if (r_final > 0xFF)
                r_final = 0xFF;
            else if (r_final < 0)
                r_final = 0;
            g_final = qFloor(g);
            if (g_final > 0xFF)
                g_final = 0xFF;
            else if (g_final < 0)
                g_final = 0;
            b_final = qFloor(b);
            if (b_final > 0xFF)
                b_final = 0xFF;
            else if (b_final < 0)
                b_final = 0;

            // Set the pixel into the image.
            work_image.setPixel(cols, rows, qRgb(r_final, g_final, b_final));
        }
    }

    // Set our pixmap to the corrected image.
    if (!convertFromImage(work_image))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "convertFromImage() failed!";
        return;
    }
}

/*!
 * @brief Perform an automatic white balance on our image. The new gains are written
 *        to the appropriate parameter.
 *
 * The algorithm is:
 * 1. Find MAX_AUTOWB_MATRICS pixels closest to the MAX_AUTOWB_GREY value.
 * 2. Convert those pixels from RGB to HSV.
 * 3. Average the Hue, Saturation and Value values.
 * 4. Convert the average HSV back to the RGB domain.
 * 5. Determine the maximum of the red, blue and green channels.
 * 6. Calculate the gains of the other (not maximum) channels to "equalize" them
 *    to the "maximum" channel - producing an even "grey" value.
 *
 * @param [out] r_gain  The white balance gain for the red channel.
 * @param [out] g_gain  The white balance gain for the green channel.
 * @param [out] b_gain  The white balance gain for the blue channel.
 */
void fiPixmap::auto_wb(double *r_gain, double *g_gain, double *b_gain)
{
    int     i;
    int     x;
    int     y;
    int     m_index;
    QImage  my_image;
    QColor  my_color;
    AUTOWB_METRICS  metrics[ MAX_AUTOWB_METRICS ];

    QLOG_TRACE() << __PRETTY_FUNCTION__;

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Get the pixmap as an image.
    my_image = toImage();

    // Clear our array.
    memset(metrics, 0, sizeof(AUTOWB_METRICS) * MAX_AUTOWB_METRICS);
    m_index = 0;

    //================================================
    // Find the brightest pixels on the image. We don't want "white", but maybe 80% white.
    //================================================

    // Loop through all the rows.
    for (y = 0; y < height(); ++y)
    {
        // Loop through all the columns.
        for (x = 0; x < width(); ++x)
        {
            int grey_value;

            // Get a color object for this pixel.
            grey_value = qGray(my_image.pixel(x, y));

            // Make sure this isn't TOO white.
            if (grey_value > MAX_AUTOWB_GREY)
                continue;

            // Can we just insert it into the array?
            if (metrics[ m_index ].grey < grey_value)
            {
                metrics[ m_index ].grey = grey_value;
                metrics[ m_index ].y = y;
                metrics[ m_index ].x = x;

                // Update our index.
                ++m_index;
                if (m_index == MAX_AUTOWB_METRICS)
                    m_index = 0;
            }
            else
            {
                // Find a suitable location.
                for (i = 0; i < MAX_AUTOWB_METRICS; ++i)
                {
                    int new_index;

                    // Get our index into the array.
                    new_index = (m_index + i) % MAX_AUTOWB_METRICS;

                    // Is this a good spot?
                    if (metrics[ new_index ].grey < grey_value)
                    {
                        metrics[ new_index ].grey = grey_value;
                        metrics[ new_index ].y = y;
                        metrics[ new_index ].x = x;

                        // Update our index.
                        m_index = new_index + 1;
                        if (m_index >= MAX_AUTOWB_METRICS)
                            m_index = 0;

                        // We're done here.
                        break;
                    }
                }
            }
        }
    }

    //====================================================================
    // Now that we have an array of "bright" pixels, average their HSV values
    //====================================================================

    // Loop through the brightest pixels we found and average their hues.
    int h_accum = 0;
    int s_accum = 0;
    int v_accum = 0;
    for (i = 0; i < MAX_AUTOWB_METRICS; ++i)
    {
        if (metrics[i].grey == 0)
            break;

        my_color = QColor(my_image.pixel(metrics[i].x, metrics[i].y));

        // Accumulate the information.
        h_accum += my_color.hue();
        s_accum += my_color.saturation();
        v_accum += my_color.value();

    }

    // It's not likely, but to avoid any possibility of a divide-by-zero...
    if (i == 0)
        i = 1;

    int h_value = h_accum / i;
    int s_value = s_accum / i;
    int v_value = v_accum / i;

    //====================================================================
    // Create a pixel from the HSV values. Convert that to RGB and find
    // the channel with the highest value. The ratios of the lower channels
    // to that higher channel will be our gains.
    //====================================================================

    // Create the HSV value.
    my_color.setHsv(h_value, s_value, v_value);

    // Convert to RGB and find the maximum value.
    int max_value;
    max_value = my_color.red();
    if (my_color.green() > max_value)
        max_value = my_color.green();
    if (my_color.blue() > max_value)
        max_value = my_color.blue();

    // Now build up the gains.
    *r_gain = ((double)max_value) / my_color.red();
    *g_gain = ((double)max_value) / my_color.green();
    *b_gain = ((double)max_value) / my_color.blue();

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif
}

/*!
 * @brief Determines if the file_type is a "raw" file type.
 *
 * @param file_type     The file type string, e.g. "dng", "jpg", "bmp", etc.
 * @return True if the file type is a "raw" file, false otherwise.
 */
bool fiPixmap::is_raw(const QString& file_type)
{
    // A DNG file is raw.
    if (file_type.compare(FiSaver::IMAGE_EXT_DNG, Qt::CaseInsensitive) == 0)
        return true;

    // Must not be raw.
    return false;
}

/*!
 * @brief Loads a pixmap from the file with the given fileName.
 *
 * @param fileName  The file name is a file on disk, or a resource.
 * @param format    [OPTIONAL] Specify the format of the file.
 * @param flags     [OPTIONAL] Used to control the conversion, if required.
 *
 * @return Returns true if the pixmap was successfully loaded, false otherwise.
 */
bool fiPixmap::load(const QString& fileName, const char *format, Qt::ImageConversionFlags flags)
{
    bool        load_ok;
    QFileInfo   file_info(fileName);

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Check to see if we have a raw file.
    if (is_raw(file_info.suffix()) || is_raw(QString::fromUtf8(format)))
    {
        load_ok = load_raw(fileName, format, flags);
        if (!load_ok)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "load_raw(" << fileName << ") Failed!";
        }
    }
    else
    {
        load_ok = QPixmap::load(fileName, format, flags);
        if (!load_ok)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "QPixmap::load(" << fileName << ") Failed!";
        }
    }

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = elapsed_time.addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Duration:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Return our result.
    return load_ok;
}

/*!
 * @brief Loads a raw pixmap from the file with the given fileName.
 *
 * @param fileName  The file name is a file on disk, or a resource.
 * @param format    [UNUSED] Specify the format of the file.
 * @param flags     [UNUSED] Used to control the conversion, if required.
 *
 * @return Returns true if the pixmap was successfully loaded, false otherwise.
 */
bool fiPixmap::load_raw(const QString& fileName, const char *format, Qt::ImageConversionFlags flags)
{
    QByteArray dng_buf;
    QFile dng_file(fileName);

    Q_UNUSED(format)
    Q_UNUSED(flags)

    // Try to open the file.
    if (!dng_file.open(QIODevice::ReadOnly))
    {
        QString msg_str = QString::fromUtf8("Error opening DNG: '%1' code:%2 (%3)").arg(dng_file.fileName()).arg(dng_file.error()).arg(dng_file.errorString());
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        return false;
    }

    // Read the file into our buffer.
    dng_buf = dng_file.readAll();

    // Done with the file.
    dng_file.close();

    // Parse the stream.
    return load_from_buf(dng_buf, FAS_VIDEO::VIDEO_CODEC::FI_DNG_STACK_CODEC);
}

/*!
 * @brief Loads a pixmap from the buffer with the specified format.
 *
 * @param buf       The buffer we want to use to load the data.
 * @param format    The format of the data in the buffer.
 *
 * @return True on success, false otherwise.
 */
bool fiPixmap::load_from_buf(QByteArray& buf, FAS_VIDEO::VIDEO_CODEC fvCodec)
{
    bool rc = false;
    DcrawUtil dc;

    // Is this a something other than a DNG?
    if (fvCodec != FAS_VIDEO::VIDEO_CODEC::FI_DNG_STACK_CODEC)
    {
        char    format_str[ 8 ];
        memset(format_str, 0, 8);

        // Figure out the format string for the codec used in the buffer.
        switch (fvCodec)
        {
        case FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC:
            memcpy(format_str, "TIF", 4);
            break;
        case FAS_VIDEO::VIDEO_CODEC::FI_BMP_STACK_CODEC:
            memcpy(format_str, "BMP", 4);
            break;
        default:
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unsupported codec type:" << fvCodec;
            return false;
        }

        // Pass it on to the parent class.
        return QPixmap::loadFromData(buf, format_str);
    }

    // Do we already have a stage3 image?
    if (raw_rgb_image != nullptr)
    {
        delete raw_rgb_image;
        raw_rgb_image = nullptr;
    }

    // Sanity check the sensor type is set appropriately.
    switch (sensor_type)
    {
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR:
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Invalid sensor type:" << Qt::showbase << Qt::hex << sensor_type;
        break;
    }


#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Get the raw RGB samples.
    raw_rgb_image = dc.debayer((uint8_t *)buf.data(), buf.size());

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "libraw:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Apply our correction matrix.
    if (ccm_updated)
    {
        // Use the CCM provided.
        apply_matrix(ccm);
    }
    else
    {
        switch (whitebal)
        {
        case GIGE_WHITE_BAL_DAYLIGHT:
            if (sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR)
                apply_matrix(AL5C_day_ccm);
            else if (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR)
                apply_matrix(X9HC_day_ccm);
            break;
        case GIGE_WHITE_BAL_FLOURESCENT:
            if (sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR)
                apply_matrix(AL5C_cwf_ccm);
            else if (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR)
                apply_matrix(X9HC_cwf_ccm);
            break;
        case GIGE_WHITE_BAL_TUNGSTEN:
            if (sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR)
                apply_matrix(AL5C_tun_ccm);
            else if (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR)
                apply_matrix(X9HC_tun_ccm);
            break;
        case GIGE_WHITE_BAL_NORMAL:
            apply_matrix(norm_ccm);
            break;
        default:
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unsupported white balance value:" << whitebal;
            break;
        }
    }

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs(elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "CCM applied:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    elapsed_timer.restart();
    // Statistics
    //-------------------------------------------------------------------
#endif
    rc = true;

    // Return our result.
    return rc;
}

/*!
 * @brief Build an output buffer of the image formatted as specified by format_type.
 *
 * @param dest_buf      Where to store the data.
 * @param imageType     What type of image is being saved.
 */
void fiPixmap::save_image(QByteArray *dest_buf, FI_SAVER::IMAGE_TYPE imageType)
{
    bool    save_result = false;
    int     quality(-1);
    QString format_str;
    QBuffer image_buf(dest_buf);

    // How do we want to save this image?
    switch (imageType)
    {
    case FI_SAVER::IMAGE_TYPE::JPEG:
        format_str = FiSaver::IMAGE_EXT_JPG;

        //================================================================================================
        // Test to catch strange values of JPEG quality. Can be removed when deemed appropriate.
        if (jpeg_quality < 10)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "JPEG Quality < 10%:" << jpeg_quality;
        }
        //================================================================================================

        quality = jpeg_quality;
        break;
    case FI_SAVER::IMAGE_TYPE::BMP:
        format_str = FiSaver::IMAGE_EXT_BMP;
        break;
    case FI_SAVER::IMAGE_TYPE::TIF:
    case FI_SAVER::IMAGE_TYPE::TIF_RAW:
        format_str = FiSaver::IMAGE_EXT_TIF;
        break;
    case FI_SAVER::IMAGE_TYPE::DNG:
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected format_type:" << imageType;
        return;
    }

    // Open up our buffer.
    image_buf.open(QIODevice::WriteOnly);

    // Write the bytes in the appropriate format.
    save_result = save(&image_buf, format_str.toUtf8().constData(), quality);
    if (!save_result)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "QPixmap::save failed to transcode using QIODevice interface!";
    }
}

void fiPixmap::set_ccm(double ccm[][4])
{
    int i, j;

    // Copy over the CCM data.
    for (i = 0; i < 3; i++)
    {
        for (j = 0; j < 4; j++)
        {
            this->ccm[i][j] = ccm[i][j];
        }
    }

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << ccm[0][0] << ccm[0][1] << ccm[0][2] << ccm[0][3];
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << ccm[1][0] << ccm[1][1] << ccm[1][2] << ccm[1][3];
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << ccm[2][0] << ccm[2][1] << ccm[2][2] << ccm[2][3];
    ccm_updated = true;
}


void fiPixmap::set_jpeg_quality(int quality)
{
    //================================================================================================
    // Test to catch strange values of JPEG quality. Can be removed when deemed appropriate.
    if (quality < 10)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "JPEG Quality < 10%:" << quality;
        if (quality == -1)
            quality = 80;
    }
    //================================================================================================

    jpeg_quality = quality;
}
