/*!
 * @file fi_plain_textedit.cpp
 *
 * @brief Provides ability to add line numbers and toggle bookmarks on
 * a QPlainTextEdit widget.
 *
 * Initial code provided by QT Code Editor Example:
 * https://doc.qt.io/qt-5/qtwidgets-widgets-codeeditor-example.html
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QMenu>
#include <QKeyEvent>
#include <QPainter>
#include <QPalette>
#include <QPen>
#include <QTextBlock>

#include "logger.h"
#include "fi_plain_textedit.h"
#include "camera_logs.h"

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

const int FI_LINE_NUMBER_PADDING    = 3;    //!< Padding used for line number widget.

const QColor fiPlainTextEdit::activeHighlightColor       = QColor(Qt::yellow).lighter(160);
const QColor fiPlainTextEdit::bookmarkHighlightColor     = QColor(Qt::blue).lighter(160);
const QColor fiPlainTextEdit::noHighlightColor           = QColor(Qt::lightGray);

using namespace FI_TEXT_EDIT;

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

/*!
 * @brief fiPlainTextEdit constructor.
 */
fiPlainTextEdit::fiPlainTextEdit(QWidget *parent) : QPlainTextEdit(parent)
{
    lineNumberArea = new LineNumberArea(this);
    parentContextIsSubMenu = true;

    connect(this, &fiPlainTextEdit::blockCountChanged, this, &fiPlainTextEdit::updateLineNumberAreaWidth);
    connect(this, &fiPlainTextEdit::updateRequest, this, &fiPlainTextEdit::updateLineNumberArea);
    connect(this, &fiPlainTextEdit::cursorPositionChanged, this, &fiPlainTextEdit::highlightCurrentLine);

    updateLineNumberAreaWidth(0);
    highlightCurrentLine();

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "parent object name:" << parent->objectName();
}

/*!
 * @brief Handle a right click from the user - this will pop-up a context menu.
 * @param event_ptr     Pointer to the right-click event.
 */
void fiPlainTextEdit::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QMenu *menu = createStandardContextMenu(event_ptr->globalPos());
    if (menu)
    {
        menu->addSeparator();
        menu->addAction(tr("Toggle Bookmark"), QKeySequence(Qt::CTRL | Qt::Key_F2), this, &fiPlainTextEdit::doBookmarkToggle);
        menu->addAction(tr("Next Bookmark"), QKeySequence(Qt::Key_F2), this, &fiPlainTextEdit::doBookmarkForward);
        menu->addAction(tr("Previous Bookmark"), QKeySequence(Qt::SHIFT | Qt::Key_F2), this, &fiPlainTextEdit::doBookmarkBackwards);
        menu->addSeparator();

        if (parentContextMenu)
        {
            QMenu *pMenu = parentContextMenu.get();
            menu->addMenu(pMenu);   // Ownership remains with parent according to QT documentation!
        }

        menu->exec(event_ptr->globalPos());
        delete menu;
    }
}

/*!
 * @brief Key release event handler.
 */
void fiPlainTextEdit::keyReleaseEvent(QKeyEvent *event)
{
    // Is this an F2 key press?
    if (event->key() == Qt::Key_F2)
    {
        // How was it modified if at all?
        if (event->modifiers() == Qt::ControlModifier)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "F2+Ctl:" << Qt::showbase << Qt::hex << event->key();
            toggleBookmark();
        }
        else if (event->modifiers() == Qt::ShiftModifier)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "F2+Shift:" << Qt::showbase << Qt::hex << event->key();
            nextBookmark(false);
        }
        else if (event->modifiers() == Qt::NoModifier)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "F2:" << Qt::showbase << Qt::hex << event->key();
            nextBookmark(true);
        }
    }
    // Let it go through.
    QPlainTextEdit::keyReleaseEvent(event);
}



/*!
 * @brief Tells the log to move toggle the bookmark at active cursor.
 */
void fiPlainTextEdit::doBookmarkToggle()
{
    toggleBookmark();
}

/*!
 * @brief Tells the log to move forward to next bookmark wrapping around if necessary.
 */
void fiPlainTextEdit::doBookmarkForward()
{
    nextBookmark(true);
}

/*!
 * @brief Tells the log to move backwards to next bookmark wrapping around if necessary.
 */
void fiPlainTextEdit::doBookmarkBackwards()
{
    nextBookmark(false);
}

/*!
 * @brief Calculates the area width for the line number widget.
 */
int fiPlainTextEdit::lineNumberAreaWidth()
{
    int digits = 1;
    int max = qMax(1, blockCount());
    while (max >= 10)
    {
        max /= 10;
        ++digits;
    }

    int space = FI_LINE_NUMBER_PADDING + fontMetrics().horizontalAdvance(QLatin1Char('9')) * digits;

    return space;
}

/*!
 * @brief Adds bookmark if not set or removes bookmark if set.
 * @param bookmarkPtr - A pointer to the bookmark in question.
 */
void fiPlainTextEdit::lineNumberToggleBookmark(FI_TEXT_EDIT::FI_TEXT_BOOKMARK *bookmarkPtr)
{
    if (bookmarkPtr)
    {
        // Insert the bookmark information if it isn't already in the map.
        if (!bookmarkMap.contains(bookmarkPtr->bmBlockId))
        {
            // Add our bookmark to our map.
            bookmarkMap.insert(bookmarkPtr->bmBlockId, bookmarkPtr->bmCursor);
        }
        else
        {
            // Remove bookmark from our map.
            bookmarkMap.remove(bookmarkPtr->bmBlockId);
        }
    }
}

/*!
 * @brief Installs a shared context menu provided by parent owning this class.
 *
 * @param parentContextMenuToSet - the parent context menu (ownership remains with parent).
 * @param asSubMenu - True if added to standard context menu for QPlainTextEdit otherwise replaces standard context menu.
 *
 */
void fiPlainTextEdit::setParentContextMenu(std::shared_ptr<QMenu> parentContextMenuToSet, bool asSubMenu)
{
    parentContextMenu = parentContextMenuToSet;
    parentContextIsSubMenu = asSubMenu;
}


/*!
 * @brief Viewport margins need to change accomodate line number widget.
 *
 * This is connected slot handler for the QPlainTextEdit::blockCountChanged.
 */
void fiPlainTextEdit::updateLineNumberAreaWidth(int /* newBlockCount */)
{
    setViewportMargins(lineNumberAreaWidth(), 0, 0, 0);
}


/*!
 * @brief Text edit needs update this connects the request to the line
 * number widget.
 *
 * This is connected slot handler for the QPlainTextEdit::updateRequest.
 */
void fiPlainTextEdit::updateLineNumberArea(const QRect& rect, int dy)
{
    if (dy)
        lineNumberArea->scroll(0, dy);
    else
        lineNumberArea->update(0, rect.y(), lineNumberArea->width(), rect.height());

    if (rect.contains(viewport()->rect()))
        updateLineNumberAreaWidth(0);
}


/*!
 * @brief Handles resize request to take into account our line number widget.
 */
void fiPlainTextEdit::resizeEvent(QResizeEvent *e)
{
    QPlainTextEdit::resizeEvent(e);

    QRect cr = contentsRect();
    lineNumberArea->setGeometry(QRect(cr.left(), cr.top(), lineNumberAreaWidth(), cr.height()));
}

/*!
 * @brief Navigates active cursor to the next bookmark based on direction.
 */
void fiPlainTextEdit::nextBookmark(bool movingForwards)
{
    // Check for any bookmarks.
    if (bookmarkMap.isEmpty())
    {
        // We don't have any bookmarks.
        // Nothing to do here.
        return;
    }

    QTextCursor activeCursor = textCursor();
    int activeBlockId   = activeCursor.blockNumber();

    QTextCursor bookmarkCursor;
    int bookmarkBlockId = -1;

    int lowestBM        = -1;
    int highestBM       = -1;
    int closestAfterBM  = -1;
    int closestBeforeBM = -1;

    // Iterate through the map looking for bookmarks we need for navigation.
    QMapIterator<int, QTextCursor> i(bookmarkMap);
    while (i.hasNext())
    {
        i.next();

        // Collect the closest before and after the current block id.
        if (i.key() > activeBlockId)
        {
            // Take care of any initializations.
            if (highestBM == -1)
            {
                highestBM = closestAfterBM = i.key();
            }
            else
            {
                // Are we closer to active block than the last time through.
                if (i.key() < closestAfterBM)
                    closestAfterBM = i.key();
                // Is there a higher bookmark than last time.
                if (i.key() > highestBM)
                    highestBM = i.key();
            }
        }
        else if (i.key() < activeBlockId)
        {
            // Take care of any initializations.
            if (lowestBM == -1)
            {
                lowestBM = closestBeforeBM = i.key();
            }
            else
            {
                // Are we closer to active block than the last time through.
                if (i.key() > closestBeforeBM)
                    closestBeforeBM = i.key();
                // Is there a lower bookmark than last time.
                if (i.key() < lowestBM)
                    lowestBM = i.key();
            }
        }
    }

    // After iteration we should have 4 pieces of inforamtion.
    // The closest Book Marks to the active block id before and after and
    // the lower and upper Book Marks.

    // What direction are we going?
    if (movingForwards)
    {
        // Do we have a bookmark after the active block id?
        if (closestAfterBM != -1)
        {
            // Yes - this is the one.
            bookmarkBlockId = closestAfterBM;
        }
        else if (lowestBM != -1)
        {
            // We wrap backwards to go around again.
            bookmarkBlockId = lowestBM;
        }
        else
        {
            // We don't have any bookmarks - maybe only one is set and that is the active one.
            return;
        }
    }
    else
    {
        // Do we have a bookmark before the active block id?
        if (closestBeforeBM != -1)
        {
            // Yes - this is the one.
            bookmarkBlockId = closestBeforeBM;
        }
        else if (highestBM != -1)
        {
            // We wrap forwards to go around again.
            bookmarkBlockId = highestBM;
        }
        else
        {
            // We don't have any bookmarks - maybe only one is set and that is the active one.
            return;
        }
    }

    // Grab the cursor that we are setting up.
    bookmarkCursor = bookmarkMap.value(bookmarkBlockId);
    setTextCursor(bookmarkCursor);
    performLineHighlight(bookmarkHighlightColor, bookmarkCursor);
}

/*!
 * @brief The user has requested a toggle bookmark on the current line.
 */
void fiPlainTextEdit::toggleBookmark()
{
    FI_TEXT_EDIT::FI_TEXT_BOOKMARK ourBookmark;
    // Grab the cursor first then grab the cursor's block id.
    // Order is important.
    ourBookmark.bmCursor    = textCursor();
    ourBookmark.bmBlockId   = ourBookmark.bmCursor.blockNumber();

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ourBookmark.bmBlockId:" << ourBookmark.bmBlockId;

    // Insert the bookmark information if it isn't already in the map.
    if (!bookmarkMap.contains(ourBookmark.bmBlockId))
    {
        // Add our bookmark to our list.
        bookmarkMap.insert(ourBookmark.bmBlockId, ourBookmark.bmCursor);
        performLineHighlight(bookmarkHighlightColor, ourBookmark.bmCursor);
    }
    else
    {
        // Remove bookmark from our map.
        bookmarkMap.remove(ourBookmark.bmBlockId);
        performLineHighlight(activeHighlightColor, ourBookmark.bmCursor);
    }
}

/*!
 * @brief Common logic to highlight a line in the plain text edit box.
 *
 * @param highlightColor - what color to use.
 * @param lineToHighlight - where to color.
 */
void fiPlainTextEdit::performLineHighlight(const QColor& highlightColor, const QTextCursor& lineToHighlight)
{
    QList<QTextEdit::ExtraSelection> extraSelections;
    QTextEdit::ExtraSelection selection;
    selection.format.setBackground(highlightColor);
    selection.format.setProperty(QTextFormat::FullWidthSelection, true);
    selection.cursor = lineToHighlight;
    selection.cursor.clearSelection();
    extraSelections.append(selection);

    setExtraSelections(extraSelections);
}

/*!
 * @brief Common logic to highlight a line in the plain text edit box.
 * Slot handler for changes to the active line in the log.
 */
void fiPlainTextEdit::highlightCurrentLine()
{
    // Is this a bookmark line?
    QTextCursor currentTextCursor = textCursor();

    // Insert the bookmark information if it isn't already in the map.
    if (bookmarkMap.contains(currentTextCursor.blockNumber()))
    {
        performLineHighlight(bookmarkHighlightColor, currentTextCursor);
    }
    else
    {
        performLineHighlight(activeHighlightColor, currentTextCursor);
    }
}

/*!
 * @brief Clears any bookmarks that were set.
 */
void fiPlainTextEdit::clearBookmarks()
{
    bookmarkMap.clear();
}

/*!
 * @brief Line number area has changed.  Need to update the line numbers.
 */
void fiPlainTextEdit::lineNumberAreaPaintEvent(QPaintEvent *event)
{
    QPainter painter(lineNumberArea);
    painter.fillRect(event->rect(), Qt::lightGray);

    QTextBlock block = firstVisibleBlock();
    int blockNumber = block.blockNumber();
    int top = qRound(blockBoundingGeometry(block).translated(contentOffset()).top());
    int bottom = top + qRound(blockBoundingRect(block).height());

    while (block.isValid() && top <= event->rect().bottom())
    {
        if (block.isVisible() && bottom >= event->rect().top())
        {
            QString number = QString::number(blockNumber + 1);

            QRect textRect(0, top, lineNumberArea->width(), fontMetrics().height());

            // Is this a bookmark?
            if (bookmarkMap.contains(blockNumber))
            {
                QColor bmColor = QColor(Qt::blue).lighter(160);
                painter.fillRect(textRect, bmColor);
            }
            else
            {
                painter.fillRect(textRect, Qt::lightGray);
            }

            painter.setPen(Qt::black);
            painter.drawText(0, top, lineNumberArea->width(), fontMetrics().height(),
                             Qt::AlignRight, number);
        }

        block = block.next();
        top = bottom;
        bottom = top + qRound(blockBoundingRect(block).height());
        ++blockNumber;
    }
}
