/*!
 * @file fi_plain_textedit.h
 *
 * @brief Provides ability to add line numbers and toggle bookmarks on
 * a QPlainTextEdit widget.
 *
 * Initial code provided by QT Code Editor Example:
 * https://doc.qt.io/qt-5/qtwidgets-widgets-codeeditor-example.html
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FIPLAINTEXTEDIT_H
#define FIPLAINTEXTEDIT_H

//--------------------------------------------------------------------
// Include Files
//--------------------------------------------------------------------
#include <QPlainTextEdit>
#include <QMap>
#include <QMenu>
#include "logger.h"

namespace FI_TEXT_EDIT
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define a bookmark structure used to keep track of places in a plain
//! text edit that the user wants to be able to revisit quickly.
typedef struct
{
    QTextCursor bmCursor;
    int bmBlockId;
} FI_TEXT_BOOKMARK;

}   // End of namespace FI_TEXT_EDIT.

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class LineNumberArea;
class CameraLogs;

//--------------------------------------------------------------------
// Class fiPlainTextEdit
//--------------------------------------------------------------------

class fiPlainTextEdit : public QPlainTextEdit
{
    Q_OBJECT

public:
    fiPlainTextEdit(QWidget *parent = nullptr);

    void            clearBookmarks();
    void            lineNumberAreaPaintEvent(QPaintEvent *event);
    int             lineNumberAreaWidth();
    void            lineNumberToggleBookmark(FI_TEXT_EDIT::FI_TEXT_BOOKMARK *bookmarkPtr);
    void            setParentContextMenu(std::shared_ptr<QMenu> parentContextMenuToSet, bool asSubMenu = true);

public slots:
    void            doBookmarkToggle();
    void            doBookmarkForward();
    void            doBookmarkBackwards();

protected:
    void            resizeEvent(QResizeEvent *event) override;
    virtual void    keyReleaseEvent(QKeyEvent *e) override;
    virtual void    contextMenuEvent(QContextMenuEvent *event_ptr) override;

    virtual void    nextBookmark(bool movingForwards = true);
    virtual void    toggleBookmark();

private slots:
    void            highlightCurrentLine();
    void            updateLineNumberArea(const QRect& rect, int dy);
    void            updateLineNumberAreaWidth(int newBlockCount);

private:
    static const QColor activeHighlightColor;
    static const QColor bookmarkHighlightColor;
    static const QColor noHighlightColor;

    void performLineHighlight(const QColor& highlightColor, const QTextCursor& lineToHighlight);

    std::shared_ptr<QMenu> parentContextMenu;    //!< A context menu to add to ours if parent wants additional support.
    bool parentContextIsSubMenu;

    QWidget *lineNumberArea;
    QMap<int, QTextCursor> bookmarkMap;
};

class LineNumberArea : public QWidget
{
public:
    LineNumberArea(fiPlainTextEdit *editor) : QWidget(editor), codeEditor(editor)
    {}

    QSize sizeHint() const override
    {
        return QSize(codeEditor->lineNumberAreaWidth(), 0);
    }

protected:
    void contextMenuEvent(QContextMenuEvent *event_ptr) override
    {
        QMenu menu(this);
        QPoint whereClicked = event_ptr->pos();

        FI_TEXT_EDIT::FI_TEXT_BOOKMARK ourBM;
        FI_TEXT_EDIT::FI_TEXT_BOOKMARK *ourBM_Ptr = &ourBM;

        ourBM.bmCursor = codeEditor->cursorForPosition(whereClicked);
        ourBM.bmBlockId = ourBM.bmCursor.blockNumber();

        menu.addAction(tr("Toggle Bookmark"), QKeySequence(Qt::CTRL | Qt::Key_F2), this, [this, ourBM_Ptr]() { toggleLineBookmark(ourBM_Ptr); });
        menu.addAction(tr("Next Bookmark"), QKeySequence(Qt::Key_F2), codeEditor, &fiPlainTextEdit::doBookmarkForward);
        menu.addAction(tr("Previous Bookmark"), QKeySequence(Qt::SHIFT | Qt::Key_F2), codeEditor, &fiPlainTextEdit::doBookmarkBackwards);
        menu.exec(event_ptr->globalPos());
    }

    void toggleLineBookmark(FI_TEXT_EDIT::FI_TEXT_BOOKMARK *ourBM_Ptr)
    {
        codeEditor->lineNumberToggleBookmark(ourBM_Ptr);
    }

    void mouseDoubleClickEvent(QMouseEvent *event) override
    {
        QPoint whereClicked = event->pos();

        FI_TEXT_EDIT::FI_TEXT_BOOKMARK ourBM;

        ourBM.bmCursor = codeEditor->cursorForPosition(whereClicked);
        ourBM.bmBlockId = ourBM.bmCursor.blockNumber();

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "pos:" << whereClicked << "blockId:" << ourBM.bmBlockId;

        codeEditor->lineNumberToggleBookmark(&ourBM);

        // Force the area to update - We should do the whole widget because the x,y coordinate may not
        // catch the top of the line number.
        repaint(0, 0, width(), height());
    }


    void paintEvent(QPaintEvent *event) override
    {
        codeEditor->lineNumberAreaPaintEvent(event);
    }

private:
    fiPlainTextEdit *codeEditor;
};

#endif // FIPLAINTEXTEDIT_H
