/*!
 * @file fi_progbar.cpp
 * @brief This file contains the implementation of the fiProgBar
 *        class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_progbar.h"
#include "ui_fi_progbar.h"

#include <QTimer>

#include "logger.h"

namespace FI_PROGBAR
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Our color style for when we're approaching the trigger.
static const char *GREEN_STYLE        = "QProgressBar { text-align: center; } ::chunk { background-color: #00CC00; }";

//! Define a color style for when we're waiting for a trigger.
static const char *YELLOW_MELLOW_STYLE = "QProgressBar { text-align: center; } ::chunk { background-color: #FFCC00; }";
static const char *YELLOW_BRIGHT_STYLE = "QProgressBar { text-align: center; } ::chunk { background-color: #FFFF00; }";

//! Our post-trigger style.
static const char *RED_STYLE           = "QProgressBar { text-align: center; } ::chunk { background-color: #CC0000; }";

}   // End of namespace FI_PROGBAR.

using namespace FI_PROGBAR;

/*!
 * @brief The fiProgBar class constructor.
 */
fiProgBar::fiProgBar(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::fiProgBar)
{
    ui->setupUi(this);

    // Connect our signals/slots.
    connect(&progress_poll_timer, &QTimer::timeout, this, &fiProgBar::time_to_update);

    // Reset our elements.
    reset();
}

/*!
 * @brief The fiProgBar class destructor.
 */
fiProgBar::~fiProgBar()
{
    delete ui;
}

/*!
 * @brief Return true if the polling timer is active, false otherwise.
 */
bool fiProgBar::polling_is_active()
{
    return progress_poll_timer.isActive();
}

/*!
 * @brief Reset the class.
 */
void fiProgBar::reset()
{
    // Stop the update polling.
    progress_poll_timer.stop();

    // Reset our control.
    trigger_frame = 0;
    total_frame_count = 100;
    toggle_style = false;

    ui->prg_position->setMinimum(0);
    ui->prg_position->setMaximum(total_frame_count);
    ui->prg_position->setValue(0);
    ui->prg_position->setStyleSheet(QString::fromUtf8(GREEN_STYLE));

    update_label();
}

/*!
 * @brief Set the current frame number.
 *
 * Based on the current frame number, we'll change how we are displaying the
 * progress bar.
 *
 * @param frame_num     The current frame number.
 */
void fiProgBar::set_current_frame(int frame_num)
{
    // Save the value.
    ui->prg_position->setValue(frame_num);

    // Is the current frame the same as the trigger frame?
    if (frame_num == trigger_frame)
    {
        if (toggle_style)
            ui->prg_position->setStyleSheet(QString::fromUtf8(YELLOW_BRIGHT_STYLE));
        else
            ui->prg_position->setStyleSheet(QString::fromUtf8(YELLOW_MELLOW_STYLE));

        toggle_style = !toggle_style;
    }
    else if (frame_num > trigger_frame)
    {
        ui->prg_position->setStyleSheet(QString::fromUtf8(RED_STYLE));
    }
    else
    {
        ui->prg_position->setStyleSheet(QString::fromUtf8(GREEN_STYLE));
    }

    // Update our label.
    update_label();
}

/*!
 * @brief Specify the number of frames for the maximum value of our progress bar.
 */
void fiProgBar::set_num_frames(int num_frames)
{
    // Save the total number of frames we're expecting.
    total_frame_count = num_frames;

    // Update our control.
    ui->prg_position->setMaximum(total_frame_count);

    // Update our label.
    update_label();
}

/*!
 * @brief Specify the frame number for the "trigger" frame.
 */
void fiProgBar::set_trigger(int trig_frame)
{
    // Save the trigger frame.
    trigger_frame = trig_frame;

    // Update our label.
    update_label();
}

/*!
 * @brief Start the polling timer. When the timer expires it fires a
 *        [signal](@ref fiProgBar::time_to_update) so the parent object
 *        can query for progress.
 *
 * @param polling_period    How frequently to fire the time_to_update signal.
 */
void fiProgBar::start_polling(int polling_period)
{
    // Start/restart our timer running.
    progress_poll_timer.start(polling_period);
}

/*!
 * @brief Stop the polling timer.
 */
void fiProgBar::stop_polling()
{
    // Stop our timer.
    progress_poll_timer.stop();
}

/*!
 * @brief Update the progress label.
 */
void fiProgBar::update_label()
{
    QString range_str;

    // Build up our label.
    range_str = QString::number(ui->prg_position->value()) + QLatin1Char('/') + QString::number(total_frame_count);
    ui->lbl_range->setText(range_str);

    // Check to see if we want to hide the label.
    ui->lbl_range->setVisible(total_frame_count != 0);
}

