/*!
 * @file fi_progress_dialog.cpp
 * @brief This is the source file for the Fastec Imaging Progress Dialog class.
 *
 * This class is a progress bar along with text to show the user how things are going.
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include <QCloseEvent>
#include <QString>
#include <QStyle>

#include "fi_progress_dialog.h"
#include "ui_fi_progress_dialog.h"

#include "logger.h"
#include "save_dialog.h"

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

namespace FI_PROGRESS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
const QString   XFER_COPY_LABEL(fiProgressDialog::tr("Copying"));
const QString   XFER_LOAD_LABEL(fiProgressDialog::tr("Loading"));
const QString   XFER_SAVE_LABEL(fiProgressDialog::tr("Saving"));
const QString   XFER_STATUS_LABEL(fiProgressDialog::tr("Status"));

//! The maximum number of characters to display in a path. Anything
//! longer will be elided.
const int       MAX_PATH_LEN            = 48;

//! If you want to turn on verbose logging for this object set this to true
//static const bool VERBOSE_DEBUG = true;
static const bool VERBOSE_DEBUG = false;

}   // End of namespace FI_PROGRESS.

using namespace FI_PROGRESS;

/*!
 * @brief The fiProgressDialog class constructor.
 */
fiProgressDialog::fiProgressDialog(QWidget *parent, Qt::WindowFlags flags) :
    QDialog(parent, flags | Qt::CustomizeWindowHint | Qt::WindowTitleHint),
    ui(new Ui::fiProgressDialog)
{
    // Initialize the UI.
    ui->setupUi(this);

    got_cancel_sig          = false;
    got_yes_sig             = false;
    got_no_sig              = false;
    iAmAXferDlg             = false;
    pollingForStatus        = false;

    // Typically, we want to be application modal.
    setWindowModality(Qt::ApplicationModal);

    // Connect up our signal and slots.
    connect(ui->pb_yes, &QPushButton::clicked, this, &fiProgressDialog::controlYesPushed);
    connect(ui->pb_no, &QPushButton::clicked, this, &fiProgressDialog::controlNoPushed);
    connect(ui->pb_cancel, &QPushButton::clicked, this, &fiProgressDialog::cancelDialog);
    connect(this, &QDialog::rejected, this, &fiProgressDialog::cancelDialog);

    //QIcon icon = style()->standardIcon(QStyle::SP_MessageBoxWarning);
    //or whatever icon you choose
    //QPixmap pixmap = icon.pixmap(QSize(60, 60));
    ui->statusLineIcon->setPixmap(style()->standardPixmap(QStyle::SP_MessageBoxWarning));
    ui->statusLineIcon->setVisible(false);
}

/*!
 * @brief The fiProgressDialog class destructor.
 */
fiProgressDialog::~fiProgressDialog()
{
    //QLOG_DEBUG() << __PRETTY_FUNCTION__;
    if (VERBOSE_DEBUG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "I was a: " << (iAmAXferDlg ? "Transfer" : "Simple") << "called by:" << whoCreatedMe << "Visible:" << (isVisible() ? "Yes" : "No");
    }

    // Make sure all connections are terminated etc.
    disconnect();
    hide();

    delete ui;
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Leaving destructor.";
}

/*!
* @brief Factory to create a simple progress dialog.
*
* @param titleString        What is the title for this simple progress dialog.
* @param statusInfoString   What is the status text to display for this simple progress dialog.
* @param useCancelButton    Does the user want to have a cancel button?
*
* @returns A std::unique_ptr<fiProgressDialog> containing a pointer to the progress dialog.
*/
std::unique_ptr<fiProgressDialog> fiProgressDialog::CreateSimpleProgressDialog(const QString& dialogTitleStr,
                                                                               const QString& statusInfoString,
                                                                               QMessageBox::StandardButtons buttonsYesNoCancelFlags,
                                                                               const QString& whoCalledMe,
                                                                               QWidget *parent, Qt::WindowFlags flags)
{
    std::unique_ptr<fiProgressDialog> theSimpleDlg =
        std::unique_ptr<fiProgressDialog>(new fiProgressDialog(parent, flags));

    if (VERBOSE_DEBUG)
    {
        QLOG_DEBUG() << "[SIMPLE Progress Factory: " << dialogTitleStr << " statusInfoStr:"
                     << statusInfoString << "called by: " << whoCalledMe << "]";
    }

    // Install the configuration setup for the xfer dialog.
    if (theSimpleDlg)
    {
        theSimpleDlg->setWindowTitle(dialogTitleStr);

        theSimpleDlg->setWindowModality(Qt::ApplicationModal);
        theSimpleDlg->whoCreatedMe = whoCalledMe;
        theSimpleDlg->iAmAXferDlg = false;

        theSimpleDlg->setControlButtons(buttonsYesNoCancelFlags);
        theSimpleDlg->showXferSrc(false);
        theSimpleDlg->showXferDest(false);
        theSimpleDlg->showXferType(false);
        theSimpleDlg->showStatusIcon(false);
        theSimpleDlg->showStatusValue(true);
        theSimpleDlg->setStatusValue(statusInfoString);
        theSimpleDlg->setCancelText(tr("Cancel"));
        theSimpleDlg->setStatusIcon(QStyle::SP_MessageBoxInformation);
    }

    return theSimpleDlg;
}

/*!
* @brief Factory to create a transfer styled progress dialog.
*
* @param srcXferTypeTitleName   A string to use as part of the Xfer dialog title.
*
* @returns A std::unique_ptr<fiProgressDialog> containing a pointer to the progress dialog.
*/
std::unique_ptr<fiProgressDialog> fiProgressDialog::CreateXferProgressDialog(FI_PROGRESS::XFER_TYPES srcXferTypeToSet,
                                                                             const QString& titlePrefixString,
                                                                             const QString& whoCalledMe,
                                                                             QWidget *parent, Qt::WindowFlags flags)
{
    std::unique_ptr<fiProgressDialog> theXferDlg =
        std::unique_ptr<fiProgressDialog>(new fiProgressDialog(parent, flags));

    if (VERBOSE_DEBUG)
    {
        QLOG_DEBUG() << "[CREATING XFER PROGRESS DIALOG: " << titlePrefixString <<
                     " called by:" << whoCalledMe << "]";
    }

    // Install the configuration setup for the xfer dialog.
    if (theXferDlg)
    {
        theXferDlg->setXferTypeTitle(srcXferTypeToSet, titlePrefixString);
        theXferDlg->whoCreatedMe = whoCalledMe;
        theXferDlg->iAmAXferDlg = true;
        theXferDlg->setControlButtons(QMessageBox::Cancel); // We only have a cancel button for this one.

        // Initialize the status icon to a message box information icon and hidden.
        theXferDlg->showStatusIcon(false);
        theXferDlg->setStatusIcon(QStyle::SP_MessageBoxInformation);
    }

    return theXferDlg;
}

/*!
 * @brief This progress dialog is not longer needed.  It is time to
 * cleanup the resources in use.  It is very important to call this
 * when relinquishing the smart "unique_ptr" because it will mark the
 * raw pointer to be deleteLater status.
 *
 * @param progressDialogToRecycle   The unique pointer that is to be recycled.
 *                                  Ownership comes to this method and it is properly handled.
 *
 * @returns True if all cleanup occurred or false if invalid unique pointer was passed.
 */
bool fiProgressDialog::RecycleProgressDialog(std::unique_ptr<fiProgressDialog>& progressDialogToRecycle)
{
    bool rc = false;

    if (progressDialogToRecycle)
    {
        // Just in case it is visible on the screen.
        progressDialogToRecycle->hideDialog();

        // Get our pointer directly so we can mark it to be deleted later.
        fiProgressDialog *progressDialogRawPtr = progressDialogToRecycle.release();
        if (progressDialogRawPtr)
        {
            progressDialogRawPtr->deleteLater();
        }
        rc = true;  // We have cleaned up and marked for deletion.
    }

    return rc;
}

/*!
 * @brief Return the value of the progress bar.
 */
int fiProgressDialog::getBarValue()
{
    return ui->progress->value();
}

/*!
 * @brief Return the maximum value allowed by the progress bar.
 */
int fiProgressDialog::getBarMax()
{
    return ui->progress->maximum();
}

/*!
 * @brief Return the minimum value allowed by the progress bar.
 */
int fiProgressDialog::getBarMin()
{
    return ui->progress->minimum();
}

/*!
 * @brief Increment the progress bar. If the wrap flag is set,
 * the value will wrap from maximum to minimum. Otherwise
 * the value will stop at maximum.
 */
void fiProgressDialog::incrementBar()
{
    int     value = 0;

    value = ui->progress->value();
    value++;

    if (value > getBarMax())
        value = getBarMax();

    setBarValue(value);
}

/*!
 * @brief Set the progress bar to the "new_value".
 * @param new_value     New value for the progress bar.
 */
void fiProgressDialog::setBarValue(int new_value)
{
    if (new_value < getBarMin())
        new_value = getBarMin();
    else if (new_value > getBarMax())
        new_value = getBarMax();

    ui->progress->setValue(new_value);
}

/*!
 * @brief Set the contents of the destination label to the new_label parameter.
 * @param new_label     The string we'll set the label to.
 */
void fiProgressDialog::setXferDestLabel(const QString& new_label)
{
    ui->labelXferDest->setText(new_label);
}

/*!
 * @brief Update the destination value control with the new_value parameter.
 * @param new_value     The string we'll set the value to.
 */
void fiProgressDialog::setXferDestValue(const QString& new_value)
{
    QString dest_str;

    // Make sure the path fits within our length restriction.
    if (new_value.length() > MAX_PATH_LEN)
    {
        dest_str = QString::fromUtf8("...");
        dest_str.append(QStringView{new_value}.right(MAX_PATH_LEN).toString());
    }
    else
    {
        // All good.
        dest_str = new_value;
    }

    // Display the path information.
    ui->infoXferDest->setText(dest_str);
}

/*!
 * @brief Update the status icon for the progress dialog.
 * @param spMessageBoxIconType  The QStyle::SP_MessageBox* value to use for the icon.
 * The available icons that make sense are
 *     QStyle::SP_MessageBoxInformation
 *     QStyle::SP_MessageBoxWarning
 *     QStyle::SP_MessageBoxCritical
 *     QStyle::SP_MessageBoxQuestion
 */
void fiProgressDialog::setStatusIcon(int spMessageBoxIconType)
{
    // Set the icon - we only support the 4 settings as listed above and will
    // default to the information icon if not one of the other ones.
    switch (spMessageBoxIconType)
    {
    case QStyle::SP_MessageBoxWarning:
        ui->statusLineIcon->setPixmap(style()->standardPixmap(QStyle::SP_MessageBoxWarning));
        break;
    case QStyle::SP_MessageBoxCritical:
        ui->statusLineIcon->setPixmap(style()->standardPixmap(QStyle::SP_MessageBoxCritical));
        break;
    case QStyle::SP_MessageBoxQuestion:
        ui->statusLineIcon->setPixmap(style()->standardPixmap(QStyle::SP_MessageBoxQuestion));
        break;
    case QStyle::SP_MessageBoxInformation:
    default:
        ui->statusLineIcon->setPixmap(style()->standardPixmap(QStyle::SP_MessageBoxInformation));
        break;
    }

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Update the status value control with the new_value parameter.
 * @param new_value     The string we'll set the value to.
 */
void fiProgressDialog::setStatusValue(const QString& new_value)
{
    ui->statusLineData->setText(new_value);

    // Update our dialog's size.
    adjustSize();
}

void fiProgressDialog::setBarMinMax(int min_value, int max_value)
{
    ui->progress->setMinimum(min_value);
    ui->progress->setMaximum(max_value);
}

void fiProgressDialog::setBarMax(int max_value)
{
    ui->progress->setMaximum(max_value);
}

void fiProgressDialog::setBarMin(int min_value)
{
    ui->progress->setMinimum(min_value);
}

/*!
 * @brief Allow user to change the push buttons available on the dialog.
 */
void fiProgressDialog::setControlButtons(QMessageBox::StandardButtons buttonsYesNoCancelFlags)
{
    bool useCancel  = (buttonsYesNoCancelFlags & QMessageBox::Cancel);
    bool useYes     = (buttonsYesNoCancelFlags & QMessageBox::Yes);
    bool useNo      = (buttonsYesNoCancelFlags & QMessageBox::No);

    // Make visible only the ones that are requested.
    ui->pb_yes->setVisible(useYes);
    ui->pb_no->setVisible(useNo);
    ui->pb_cancel->setVisible(useCancel);

    // If nothing is visible hide the frame.
    if (!useCancel && !useYes && !useNo)
    {
        showControlButtons(false);
        return;
    }

    // Otherwise make visible only the ones that are requested.
    ui->pb_yes->setVisible(useYes);
    ui->pb_no->setVisible(useNo);
    ui->pb_cancel->setVisible(useCancel);

    // Show the control button frame.
    showControlButtons(true);
}

/*!
 * @brief Change the text string of the cancel button to "new_text".
 * @param cancelTextStr  The string to use for the cancel button text.
 */
void fiProgressDialog::setCancelText(const QString& cancelTextStr)
{
    // Do we have a string to use for the cancel button?
    if (!cancelTextStr.isEmpty())
    {
        ui->pb_cancel->setText(cancelTextStr);
    }

    // Resize the button if neceessary.
    ui->pb_cancel->adjustSize();
}

/*!
 * @brief Set the contents of the source label to the new_label parameter.
 * @param new_label     The string we'll set the label to.
 */
void fiProgressDialog::setXferSrcLabel(const QString& new_label)
{
    ui->labelXferSrc->setText(new_label);
}

/*!
 * @brief Update the source value control with the new_value parameter.
 * @param new_value     The string we'll set the value to.
 */
void fiProgressDialog::setXferSrcValue(const QString& new_value)
{
    QString src_str;

    // Make sure the path fits within our length restriction.
    if (new_value.length() > MAX_PATH_LEN)
    {
        src_str = QString::fromUtf8("...");
        src_str.append(QStringView{new_value}.right(MAX_PATH_LEN).toString());
    }
    else
    {
        // All good.
        src_str = new_value;
    }

    // Set the value.
    ui->infoXferSrc->setText(src_str);
}

/*!
 * @brief Specify the type of transfer that we're doing.
 *
 * @param xferTypeToSet     One of the FI_PROGRESS::XFER_TYPE_xxx values.
 * @param prefixLabel       [OPTIONAL] Name to pre-fix on to the title.
 */
void fiProgressDialog::setXferTypeTitle(FI_PROGRESS::XFER_TYPES xferTypeToSet, const QString& prefixLabel)
{
    QString     titleString;

    // Did the user specify a name?
    if (!prefixLabel.isEmpty())
    {
        // Build up the beginning of the title.
        titleString = prefixLabel + QString::fromUtf8(": ");
    }

    // Append the transfer type to the title.
    switch (xferTypeToSet)
    {
    case XFER_TYPE_COPY:
        titleString.append(XFER_COPY_LABEL);
        break;
    case XFER_TYPE_LOAD:
        titleString.append(XFER_LOAD_LABEL);
        break;
    case XFER_TYPE_SAVE:
        titleString.append(XFER_SAVE_LABEL);
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected transfer type:" << xferTypeToSet;
        break;
    }

    // Set our title.
    setWindowTitle(titleString);
}

void fiProgressDialog::setXferTypeValue(const QString& new_value)
{
    ui->infoXferType->setText(new_value);
}

/*!
 * @brief For progress dialogs that poll for status the user wants to
 * pause the polling timer in a wait cycle they will invoke the function
 * setXferToPollAgain() when they are ready to resume.
 */
void fiProgressDialog::setXferToPollWait()
{
    pollTimer.stop();
}

void fiProgressDialog::setXferToPollAgain()
{
    pollTimer.start();
}

/*!
 * @brief Show the dialog and start polling for status updates.
 * @param intervalMS        How many milliseconds to set polling timer.
 * @param updateReceiver    Who is getting the poll request (class)
 * @param updateMember      Who is getting the poll request (class member).
 */
void fiProgressDialog::showDialogWithPolling(int intervalMS)

{
    // Don't want this running - should never be in this state.
    if (pollTimer.isActive())
        pollTimer.stop();

    // Set the interval time.
    pollTimer.setInterval(intervalMS);

    // Connect an internal poll receiver.
    connect(&pollTimer, &QTimer::timeout, this, &fiProgressDialog::pollTimerHit);

    // Make the dialog visible to the user
    if (!isVisible())
        show();

    // Mark the dialog as polling for status.
    pollingForStatus = true;

    // Start the timer.
    pollTimer.start();
}

/*!
 * @brief Show the dialog and wait for caller to update the contents until
 * user cancels or caller hides or deletes.
 */
void fiProgressDialog::showDialogWithoutPolling()
{
    if (pollTimer.isActive())
    {
        pollTimer.stop();
    }
    pollingForStatus = false;

    // Show it if it is hidden.
    if (!isVisible())
    {
        show();
    }
}


/*!
 * @brief The dialog is getting "shown". Create any UI that needs
 * to have the dialog created and showing.
 *
 * @param event     Poiner to the actual event object.
 */
void fiProgressDialog::showEvent(QShowEvent *event)
{
    // Pass it on.
    QDialog::showEvent(event);
}

void fiProgressDialog::hideDialog()
{
    // Make sure there is no timer running.
    if (pollTimer.isActive())
        pollTimer.stop();

    pollingForStatus = false;

    // Hide it if it is visible.
    if (isVisible())
    {
        hide();
    }
}

/*!
 * @brief Show or hide the frame that contains control buttons.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showControlButtons(bool do_show)
{
    // Show or hide the control.
    ui->frameControlBtns->setVisible(do_show);

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Show or hide the cancel button.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showCancel(bool use_cancel_pb)
{
    // Show or hide the control.
    ui->frameControlBtns->setVisible(use_cancel_pb);

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Show or hide the source information line.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showXferSrc(bool do_show)
{
    // Show or hide the control.
    ui->frameXferSrc->setVisible(do_show);

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Show or hide the destination information line.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showXferDest(bool do_show)
{
    // Show or hide the control.
    ui->frameXferDest->setVisible(do_show);

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Show or hide the save type line.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showXferType(bool do_show)
{
    // Show or hide the control.
    ui->frameXferType->setVisible(do_show);

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Show or hide the extra information line.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showStatusIcon(bool do_show)
{
    // Show or hide the control.
    ui->statusLineIcon->setVisible(do_show);

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Show or hide the status value line.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showStatusValue(bool do_show)
{
    // Show or hide the control.
    ui->statusLineData->setVisible(do_show);

    // Update our dialog's size.
    adjustSize();
}

/*!
 * @brief Show or hide the progress bar.
 * @param do_show   True to show, false to hide.
 */
void fiProgressDialog::showBar(bool do_show)
{
    // Show or hide the control.
    ui->progress->setVisible(do_show);

    // Update our dialog's size.
    adjustSize();
}

//--------------------------------------------------------------------
// Event Methods
//--------------------------------------------------------------------

/*!
 * @brief If the user hits the close button, we want to ignore it.
 *    They have to hit the cancel button.
 *
 * @param event Our Close event.
 */
void fiProgressDialog::closeEvent(QCloseEvent *event)
{
    event->ignore();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has pressed the cancel button.
 */
void fiProgressDialog::cancelDialog()
{
    // Remember we've gotten canceled.
    got_cancel_sig = true;

    // Hide our dialog.
    hideDialog();

    // Let folks know we're canceled.
    emit progressCancelled();
}

/*!
 * @brief The user has pressed the yes button.
 */
void fiProgressDialog::controlYesPushed()
{
    // Remember we've gotten canceled.
    got_yes_sig = true;

    // Hide our dialog.
    hideDialog();

    // Let folks know we're canceled.
    emit progressYesPushed();
}

/*!
 * @brief The user has pressed the no button.
 */
void fiProgressDialog::controlNoPushed()
{
    // Remember we've gotten canceled.
    got_no_sig = true;

    // Hide our dialog.
    hideDialog();

    // Let folks know we're canceled.
    emit progressNoPushed();
}

/*!
 * @brief The polling timer has fired.
 */
void fiProgressDialog::pollTimerHit()
{
    // Send the signal that it's time for progress update.
    emit progressUpdatePollCycle();
}

