/*!
 * @file fi_progress_dialog.cpp
 * @brief This is the source file for the Fastec Imaging Progress Dialog class.
 *
 * This class is a progress bar along with text to show the user how things are going.
 *
 * Copyright 2021 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_PROGRESS_DIALOG_H
#define FI_PROGRESS_DIALOG_H

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include <QDialog>
#include <QMessageBox>
#include <QTimer>
#include <stdint.h>

namespace Ui
{
class fiProgressDialog;
}

namespace FI_PROGRESS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

typedef enum XFER_TYPES
{
    XFER_TYPE_NONE,                         //!< Not performing specific data transfer.
    XFER_TYPE_COPY,                         //!< Performing a copy operation.
    XFER_TYPE_LOAD,                         //!< Performing a load operation.
    XFER_TYPE_SAVE,                         //!< Performing a save operation.
} XFER_TYPES;

}   // End of namespace FI_PROGRESS.

//-----------------------------------------------------------------------------
// fiProgressDialog Class Declaration
//-----------------------------------------------------------------------------

class fiProgressDialog : public QDialog
{
    Q_OBJECT

protected:
    explicit fiProgressDialog(QWidget *parent = nullptr, Qt::WindowFlags flags = Qt::WindowFlags());

public:
    ~fiProgressDialog();

    static std::unique_ptr<fiProgressDialog> CreateSimpleProgressDialog(const QString& dialogTitleStr,
                                                                        const QString& statusInfoString,
                                                                        QMessageBox::StandardButtons buttonsYesNoCancelFlags,
                                                                        const QString& whoCalledMe,
                                                                        QWidget *parent = nullptr,
                                                                        Qt::WindowFlags flags = Qt::WindowFlags());

    static std::unique_ptr<fiProgressDialog> CreateXferProgressDialog(FI_PROGRESS::XFER_TYPES srcXferTypeToSet,
                                                                      const QString& titlePrefixString,
                                                                      const QString& whoCalledMe,
                                                                      QWidget *parent = nullptr,
                                                                      Qt::WindowFlags flags = Qt::WindowFlags());

    static bool RecycleProgressDialog(std::unique_ptr<fiProgressDialog>& progressDialogToRecycle);

    int     getBarValue();
    int     getBarMax();
    int     getBarMin();

    void    incrementBar();

    // The set methods are listed in the order they appear in the dialog top to bottom.
    void    setXferSrcLabel(const QString& new_label);
    void    setXferSrcValue(const QString& new_value);
    void    setXferDestLabel(const QString& new_label);
    void    setXferDestValue(const QString& new_value);
    void    setXferTypeTitle(FI_PROGRESS::XFER_TYPES xferTypeToSet, const QString& prefixLabel = QString());
    void    setXferTypeValue(const QString& new_value);
    void    setStatusIcon(int spMessageBoxIconType);
    void    setStatusValue(const QString& new_value);
    void    setBarValue(int new_value);
    void    setBarMinMax(int min_value, int max_value);
    void    setBarMax(int max_value);
    void    setBarMin(int min_value);

    void    setControlButtons(QMessageBox::StandardButtons buttonsYesNoCancelFlags);
    void    setCancelText(const QString& cancelTextStr);

    // Round trip again if polling external source is driving updates.
    // Make sure that the external polling object connects to the emit
    // signal pollTimerHit and provides a polling slot and its polling
    // slot it should call both of these first the wait then the poll
    // again so that the time loops through again.
    void    setXferToPollWait();
    void    setXferToPollAgain();

    // Show the dialog the way you want to provide progress.  If you
    // call show with polling it will set a timer and you must connect
    // the signal pollTimerHit to external polling handler.
    void    showDialogWithPolling(int intervalMS);
    void    showDialogWithoutPolling();
    void    hideDialog();

    // Listed in the order available in the dialog top to bottom.
    void    showXferSrc(bool do_show);
    void    showXferDest(bool do_show);
    void    showXferType(bool do_show);
    void    showStatusIcon(bool do_show);
    void    showStatusValue(bool do_show);
    void    showBar(bool do_show);

    void    showControlButtons(bool do_show);

    void    showCancel(bool do_cancel);
    bool    wasCanceled()                   { return got_cancel_sig; }
    bool    yesSignaled()                   { return got_yes_sig; }
    bool    noSignaled()                    { return got_no_sig; }

public slots:
    virtual void    controlYesPushed();
    virtual void    controlNoPushed();
    virtual void    cancelDialog();
    virtual void    pollTimerHit();

protected:
    void    closeEvent(QCloseEvent *event);

signals:
    void    progressYesPushed();
    void    progressNoPushed();
    void    progressCancelled();
    void    progressUpdatePollCycle();

protected:
    virtual void showEvent(QShowEvent *event);

private:
    Ui::fiProgressDialog *ui;

    bool    got_cancel_sig;             //!< True if we got a cancel signal.
    bool    got_yes_sig;                //!< True if we got a yes push button signal.
    bool    got_no_sig;                 //!< True if we got a no push signal.
    bool    iAmAXferDlg;                //!< I was created as a transfer dialog.
    bool    pollingForStatus;           //!< Is this dialog polling for status?
    QTimer  pollTimer;                  //!< If dialog is polling for status this will be allocated.

    QString whoCreatedMe;               //!< Pretty function string identifying my creator.  Using for debug purposes.
};

#endif // FI_PROGRESS_DIALOG_H
