/*!
 * @file fi_rawdrive.cpp
 * @brief This file contains the implementation of the fiRawDrive class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QApplication>
#include <QByteArray>
#include <QCryptographicHash>
#include <QDir>
#include <QElapsedTimer>
#include <QFile>
#include <QtEndian>
#include <QStorageInfo>
#include <QStringList>
#include <QVector>
#include <stdint.h>
#include <qmath.h>

#ifdef Q_OS_WIN
#include <qt_windows.h>
#include <cinttypes>
#endif // Q_OS_WIN

#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
#include <stdio.h>
#include <stdlib.h>
#include <errno.h>
#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <inttypes.h>

#include <sys/ioctl.h>
#include <sys/stat.h>
#include <sys/types.h>

#ifdef Q_OS_LINUX
#include <linux/fs.h>
#endif
#endif // Q_OS_LINUX or Q_OS_MAC

#include <cstddef>
#include <memory>

#include <fas_video.h>
#include <fas_filesystem.h>

#include "fi_comm.h"
#include "fi_pixmap.h"
#include "fi_rawdrive.h"
#include "fi_sensor.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"

#include "client_math.h"

// Make this a "1" to turn on a debug log message for function entry points or set to "0" to remove from code.
#define RAWDRIVE_LOG_FUNC_ENTRY 0
const QString log_entry_str = QStringLiteral("Entry");

//#define RIGHT_SHIFT_PIXELS 1
namespace FI_RAW_DRIVE
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Number of 32-bit words in a frame metadata block.
const uint32_t  FMETA_SIZE_WORDS            = (GIGE_FRAME_METADATA_LEN / 4);

//! Our ceiling for subtracting 8-bit pixel FPN. If the pixel value is higher than this, we
//! call it "saturated" and we won't subtract the FPN.
const uint8_t   FPN_CEIL_8_BIT              = 0xFC;

#ifdef RIGHT_SHIFT_PIXELS
//! Our ceiling for subtracting 10-bit pixel FPN. If the pixel value is higher than this, we
//! call it "saturated" and we won't subtract the FPN.
const uint16_t  FPN_CEIL_10_BIT             = 0x3F7;

//! Our ceiling for subtracting 12-bit pixel FPN. If the pixel value is higher than this, we
//! call it "saturated" and we won't subtract the FPN.
const uint16_t  FPN_CEIL_12_BIT             = 0xFDC;
#endif /* RIGHT_SHIFT_PIXELS */

//! Our ceiling for subtracting 16-bit pixel FPN. If the pixel value is higher than this, we
//! call it "saturated" and we won't subtract the FPN.
const uint16_t  FPN_CEIL_16_BIT             = 0xFDC0;

//! A starting length for buffer size.
const uint32_t  INIT_LIST_LEN               = 65536;

//! The maximum number of bytes we'll support for header information.
const uint32_t  MAX_IMAGE_HEADER_LEN        = 0x800;

//! On the camera, there are this many sectors per page.
const uint32_t SECTORS_PER_PAGE             = 8;

#ifdef Q_OS_WIN
//! Physical device identifier.
const QString   PHYSICAL_STR("PhysicalDrive");
#endif

#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
//! An invalid handle.
const int       INVALID_HANDLE_VALUE        = -1;

#define DKIOCGETPHYSICALBLOCKSIZE           _IOR('d', 77, uint32_t)
#endif

//--------------------------------------------------------------------
// Private class to hold OS specific logic.
//--------------------------------------------------------------------
class fiRawDrivePrivate : public QObject
{
public:
    explicit fiRawDrivePrivate(QObject *parent);
    ~fiRawDrivePrivate();

#ifdef Q_OS_WIN
    QStringList all_physical_drives();
    void    close_device();
    bool    is_device_open()          { return (device_handle != INVALID_HANDLE_VALUE); }
    bool    open_device(const QString& device_uri);
    bool    query_device_info();
    bool    read_data(uint64_t offset, uint8_t *data_ptr, uint64_t bytes_to_read);
    bool    read_sectors(uint64_t start_sector, uint8_t *data_ptr, uint32_t num_sectors);

    uint32_t    get_sector_size()     { return (uint32_t)geometry.BytesPerSector; }

private:
    HANDLE          device_handle;
    DISK_GEOMETRY   geometry;

#endif // Q_OS_WIN
#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
    QStringList all_physical_drives();
    void    close_device();
    bool    is_device_open()          { return (device_handle != FI_RAW_DRIVE::INVALID_HANDLE_VALUE); }
    bool    open_device(const QString& device_uri);
    bool    query_device_info();
    bool    read_data(uint64_t offset, uint8_t *data_ptr, uint64_t bytes_to_read);
    bool    read_sectors(uint64_t start_sector, uint8_t *data_ptr, uint32_t num_sectors);
    uint32_t    get_sector_size()     { return bytes_per_sector; }

private:
    int         device_handle;
    uint32_t    bytes_per_sector;

#endif // Q_OS_LINUX or Q_OS_MAC

};

}   // End of namespace FI_RAW_DRIVE.

using namespace FI_RAW_DRIVE;

/*!
 * @brief fiRawDrive constructor.
 */
fiRawDrive::fiRawDrive(QObject *parent) :
    QObject(parent)
{
    priv_ptr = std::make_unique<fiRawDrivePrivate>(new fiRawDrivePrivate(this));

    // Initialize our member variables.
    drive_str.clear();
}

fiRawDrive::~fiRawDrive()
{
}

/*!
 * @brief Use the correction table to remove the fixed pixel noise (FPN) from
 *        the data in the image_buf parameter.
 *
 * @param image_buf     The image buffer.
 */
void fiRawDrive::applyFPN(QByteArray& image_buf)
{
    int         ct_shift;
    uint32_t    image_align_bytes;
    uint32_t    ct_align_bytes;
    uint32_t    row;
    uint32_t    column;
    uint8_t     *image_ptr;
    const uint8_t   *ct_ptr;
    const uint8_t   *prnu_ptr;
#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif
#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Sanity check.
    if (ct_buffer.isEmpty())
    {
        // Try to recover the correction table.
        if (!recoverCT())
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the correction table!";
            return;
        }
        if (ct_buffer.isEmpty())
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No correction table?!";
            return;
        }
    }
    if (metadata_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "metdata_ptr is nullptr!";
        return;
    }

    // Move often used values into local variables.
    const uint8_t bit_sel = metadata_ptr->bit_sel;
    const uint32_t height = metadata_ptr->roi_height;
    const uint32_t width = metadata_ptr->roi_width;

    // Figure out our iamge alignment bytes, if any.
    image_align_bytes = 0;
    switch (bit_sel)
    {
    case GIGE_BITMODE_LOWER_8_BITS:
    case GIGE_BITMODE_7_0:
        // For the low 8-bit pixels, there's no shifting of correction table values.
        ct_shift = 0;

        // Figure out how much padding we need, if any.
        if ((width & 3) != 0)
            image_align_bytes = 4 - (width & 3);
        break;
    case GIGE_BITMODE_MIDDLE_8_BITS:
    case GIGE_BITMODE_8_1:
        // For the middle 8-bit captures, we need to shift down the FPN by one bit.
        ct_shift = 1;

        // Figure out how much padding we need, if any.
        if ((width & 3) != 0)
            image_align_bytes = 4 - (width & 3);
        break;
    case GIGE_BITMODE_UPPER_8_BITS:
    case GIGE_BITMODE_9_2:
        // For the upper 8-bit captures, we need to shift the FPN by 2 bits.
        ct_shift = 2;

        // Figure out how much padding we need, if any.
        if ((width & 3) != 0)
            image_align_bytes = 4 - (width & 3);
        break;
    case GIGE_BITMODE_10_BITS:
    case GIGE_BITMODE_9_0:
#ifdef RIGHT_SHIFT_PIXELS
        // For the bottom 10-bit pixels, there is no shift required.
        ct_shift = 0;
#else
        // For 10-bit pixels, we need to shift the correction table values up so
        // that they are aligned with the pixel data.
        ct_shift = 6;
#endif /* RIGHT_SHIFT_PIXELS */

        // In 10-bit mode, since we expand to 16-bit pixels, an alignment can only
        // be two bytes.
        if (((width * 2) & 3) != 0)
            image_align_bytes = 2;
        break;
    case GIGE_BITMODE_10_3:
        // For the 8-bit captures on bits 10:3, we need to shift the FPN by 3 bits.
        ct_shift = 3;

        // Figure out how much padding we need, if any.
        if ((width & 3) != 0)
            image_align_bytes = 4 - (width & 3);
        break;
    case GIGE_BITMODE_11_4:
        // For the 8-bit captures on bits 11:4, we need to shift the FPN by 4 bits.
        ct_shift = 4;

        // Figure out how much padding we need, if any.
        if ((width & 3) != 0)
            image_align_bytes = 4 - (width & 3);
        break;
    case GIGE_BITMODE_10_1:
#ifdef RIGHT_SHIFT_PIXELS
        // For 10-bit pixels on bits 10_1, we need to shift the correction table values down so
        // that they are aligned with the pixel data.
        ct_shift = 1;
#else
        // For 10-bit pixels on bits 10_1, we need to shift the correction table values up so
        // that they are aligned with the pixel data.
        ct_shift = 5;
#endif /* RIGHT_SHIFT_PIXELS */

        // In 10-bit mode, since we expand to 16-bit pixels, an alignment can only
        // be two bytes.
        if (((width * 2) & 3) != 0)
            image_align_bytes = 2;
        break;
    case GIGE_BITMODE_11_2:
#ifdef RIGHT_SHIFT_PIXELS
        // For 10-bit pixels on bits 11_2, we need to shift the correction table values down so
        // that they are aligned with the pixel data.
        ct_shift = 2;
#else
        // For 10-bit pixels on bits 11_2, we need to shift the correction table values up so
        // that they are aligned with the pixel data.
        ct_shift = 4;
#endif /* RIGHT_SHIFT_PIXELS */

        // In 10-bit mode, since we expand to 16-bit pixels, an alignment can only
        // be two bytes.
        if (((width * 2) & 3) != 0)
            image_align_bytes = 2;
        break;
    case GIGE_BITMODE_11_0:
#ifdef RIGHT_SHIFT_PIXELS
        // For 12-bit pixels, do not need to shift the pixels.
        ct_shift = 0;
#else
        // For 12-bit pixels, we need to shift the correction table values up so
        // that they are aligned with the pixel data. The MS bit of the correction
        // byte is in bit7. Bit7 of the 12-bit pixel is at bit11 of the 16-bit raw
        // pixel. To get from bit7 up to bit11, that's a (left) shift of 4.
        ct_shift = 4;
#endif /* RIGHT_SHIFT_PIXELS */

        // In 12-bit mode, since we expand to 16-bit pixels, an alignment can only
        // be two bytes.
        if (((width * 2) & 3) != 0)
            image_align_bytes = 2;
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unknown bit selection:" << bit_sel;
        return;
    }

    // Figure out our correction table alignment bytes. The correction
    // table rows will be 32-bit aligned.
    ct_align_bytes = 0;
    if ((width & 3) != 0)
        ct_align_bytes = 4 - (width & 3);

    // Initialize our correction table pointer.
    ct_ptr = (const uint8_t *)ct_buffer.data();

    // Initialize our PRNU gain table pointer.
    if (prnu_buffer.isEmpty())
    {
        prnu_ptr = nullptr;
    }
    else
    {
        prnu_ptr = (const uint8_t *)prnu_buffer.data();
    }

    // Initialize our image pointer.
    image_ptr = (uint8_t *)image_buf.data();

    // *INDENT-OFF*
#pragma omp parallel for collapse(2)
    // *INDENT-ON*
    // Loop through each row of the data.
    for (row = 0; row < height; ++row)
    {
        // Loop through each column of the data.
        for (column = 0; column < width; ++column)
        {
            float prnu_gain;

            // Where does the row start in the expanded buffer.
            const int pixel_offset = (int)(row * (width + ct_align_bytes)) + (int)(column);

            // Do we have a PRNU gain?
            if (prnu_ptr == nullptr)
            {
                // No gain.
                prnu_gain = 1.0;
            }
            else
            {
                // Our gains are 1.7 fixed point with an implied +1.0.
                prnu_gain = ((float)(prnu_ptr[pixel_offset]) / 128.0) + 1.0;
            }

            // Process based on bit depth.
            switch (bit_sel)
            {
            case GIGE_BITMODE_LOWER_8_BITS:
            case GIGE_BITMODE_7_0:
            case GIGE_BITMODE_MIDDLE_8_BITS:
            case GIGE_BITMODE_8_1:
            case GIGE_BITMODE_UPPER_8_BITS:
            case GIGE_BITMODE_9_2:
            case GIGE_BITMODE_10_3:
            case GIGE_BITMODE_11_4:
            {
                uint8_t fpn_value;
                uint8_t pixel_value;
                uint16_t output_value;

                // Calculate the image offset for where we'll find our pixel value.
                const int image_offset = (int)(row * (width + image_align_bytes)) + (int)(column);
                pixel_value = image_ptr[image_offset];
                fpn_value = ct_ptr[pixel_offset] >> ct_shift;

                // Do we want to correct this pixel?
                if (pixel_value < FPN_CEIL_8_BIT)
                {
                    // Is the image data less than the FPN correction value?
                    if (pixel_value < fpn_value)
                    {
                        pixel_value = 0;
                    }
                    else
                    {
                        output_value = (uint16_t)(pixel_value - fpn_value) * prnu_gain;
                        if (output_value > 0xFF)
                            pixel_value = 0xFF;
                        else
                            pixel_value = (uint8_t)output_value;
                    }

                    // Assign the new pixel value.
                    image_ptr[image_offset] = pixel_value;
                }
            }
            break;
            case GIGE_BITMODE_10_BITS:
            case GIGE_BITMODE_9_0:
            case GIGE_BITMODE_10_1:
            case GIGE_BITMODE_11_2:
            {
                uint16_t pixel_value;
                uint16_t output_value;

                // Because we are dealing with pixel values which are larger than 1 byte, we need a separate
                // offset into the image memory.
                const int image_offset = (int)(row * ((width * 2) + image_align_bytes)) + (int)(column * 2);

                // Get our pixel (2-bytes). Because it may not be aligned,
                // we'll do a memcpy.
                memcpy(&pixel_value, &image_ptr[image_offset], sizeof(pixel_value));

                // Do we want to correct this pixel?
#ifdef RIGHT_SHIFT_PIXELS
                if (pixel_value < FPN_CEIL_10_BIT)
#else
                if (pixel_value < FPN_CEIL_16_BIT)
#endif /* RIGHT_SHIFT_PIXELS */
                {
                    uint16_t fpn_value;

                    // Get the BYTE from the correction table.
                    fpn_value = ((uint16_t)(ct_ptr[pixel_offset]));

                    // Shift our correction value so it'll be aligned with the
                    // pixel value.
#ifdef RIGHT_SHIFT_PIXELS
                    fpn_value = (uint16_t)(fpn_value >> ct_shift);
#else
                    fpn_value = (uint16_t)(fpn_value << ct_shift);
#endif /* RIGHT_SHIFT_PIXELS */

                    // Is the image data less than the FPN correction value?
                    if (pixel_value < fpn_value)
                        pixel_value = 0;
                    else
                        pixel_value = pixel_value - fpn_value;

                    // Right justify the value and apply the gain.
                    output_value = ((pixel_value >> 6) * prnu_gain);

                    // See if we need to saturate. Also, left justify.
                    if (output_value > 0x03FF)
                        pixel_value = 0xFFC0;
                    else
                        pixel_value = (output_value << 6);

                    // Assign the new pixel value.
                    memcpy(&image_ptr[image_offset], &pixel_value, sizeof(pixel_value));
                }
            }
            break;
            default:
            {
                uint16_t pixel_value;
                uint16_t output_value;

                // Because we are dealing with pixel values which are larger than 1 byte, we need a separate
                // offset into the image memory.
                const int image_offset = (int)(row * ((width * 2) + image_align_bytes)) + (int)(column * 2);

                // Get our pixel (2-bytes). Because it may not be aligned,
                // we'll do a memcpy.
                memcpy(&pixel_value, &image_ptr[image_offset], sizeof(pixel_value));

                // Do we want to correct this pixel?
#ifdef RIGHT_SHIFT_PIXELS
                if (pixel_value < FPN_CEIL_12_BIT)
#else
                if (pixel_value < FPN_CEIL_16_BIT)
#endif /* RIGHT_SHIFT_PIXELS */
                {
                    uint16_t fpn_value;

                    // Get the BYTE from the correction table.
                    fpn_value = ((uint16_t)(ct_ptr[pixel_offset]));

                    // Shift our correction value so it'll be aligned with the
                    // pixel value.
#ifdef RIGHT_SHIFT_PIXELS
                    fpn_value = (uint16_t)(fpn_value >> ct_shift);
#else
                    fpn_value = (uint16_t)(fpn_value << ct_shift);
#endif /* RIGHT_SHIFT_PIXELS */

                    // Is the image data less than the FPN correction value?
                    if (pixel_value < fpn_value)
                        pixel_value = 0;
                    else
                        pixel_value = pixel_value - fpn_value;

                    // Right justify the value and apply the gain.
                    output_value = ((pixel_value >> 4) * prnu_gain);

                    // See if we need to saturate. Also, left justify.
                    if (output_value > 0x0FFF)
                        pixel_value = 0xFFF0;
                    else
                        pixel_value = (output_value << 4);

                    // Assign the new pixel value.
                    memcpy(&image_ptr[image_offset], &pixel_value, sizeof(pixel_value));
                }
            }
            break;
            }
        }
    }

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif
}

/*!
 * @brief Convert a byte count to a number of sectors.
 * @param byte_count    The number of bytes we're interested in.
 * @return The number of sectors containing the byte_count bytes.
 */
uint32_t fiRawDrive::bytesToSectors(uint32_t byte_count)
{
    // Sanity check.
    if ((priv_ptr == nullptr) || (priv_ptr->get_sector_size() == 0))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Bad private information!";
        return 0;
    }

    // Return the number of sectors.
    return ((byte_count + (priv_ptr->get_sector_size() - 1)) / priv_ptr->get_sector_size());
}

/*!
 * @brief Close a previously opened drive.
 */
void fiRawDrive::close()
{
#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Close the drive.
    priv_ptr->close_device();

    // Clean up after ourselves.
    metadata_ptr.reset();
    drive_str.clear();
}

/*!
 * @brief Return an absolute frame number.
 *
 *    If "relative", meaning absolute=false, the frame number passed in is relative to the
 *    oldest frame in the partition, which can change since it's a circular buffer.
 *    If it is "absolute", meaning absolute=true, it is relative to the physical start of
 *    the partition; in that case, we just sanity check it and return what was passed.
 *
 * @param frame_num The frame number we're interested in.
 * @param absolute  TRUE if we passed the "absolute" frame_num,
 *          or FALSE if we passed a "relative" frame_num.
 *
 * @return The absolute frame number if everything is good, otherwise (uint32_t)(-1).
 */
uint32_t fiRawDrive::convertToAbsFrameNum(uint32_t frame_num, bool absolute)
{
    uint32_t abs_frame;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_info;

    // Make sure we have our partition information.
    if (part_info.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No partition information!?";
        return -1;
    }

    // Get a pointer to our capture header.
    cap_info = &part_info[0].caps[0].hdr.lr_cap;

    // Did we receive an "absolute" frame number?
    if (absolute)
    {
        // Just sanity check the frame number.
        if (frame_num < cap_info->frame_count)
        {
            abs_frame = frame_num;
        }
        else
        {
            // We should never be asked to do this.
            QLOG_WARN() << __PRETTY_FUNCTION__ << QString("absolute=true but frame_num=%1 allocated_frames=%2").arg(frame_num).arg(cap_info->frame_count);

            // It's not likely, but let's be safe.
            if (cap_info->frame_count == 0)
                abs_frame = frame_num;
            else
                abs_frame = frame_num % cap_info->frame_count;
        }
    }
    else
    {
        // Check to see if the frame being requested goes beyond (i.e. "wraps")
        // the last frame.
        if ((frame_num + cap_info->first_frame_num) < cap_info->frame_count)
        {
            // Nope - we're good. Just add the frame number to our first frame
            // position.
            abs_frame = frame_num + cap_info->first_frame_num;
        }
        else
        {
            // Yes it does. We need to subtract the total number of frames so
            // that the frame number will wrap around to the beginning again.
            abs_frame = (frame_num + cap_info->first_frame_num) - cap_info->frame_count;
        }
    }

    return abs_frame;
}


/*!
 * @brief The idea for this method is that it will return a qualified list
 *        of URIs for Fastec formatted drives connected to the PC. For
 *        Windows systems, this will look like "\\.\PhysicalDrive0",
 *        "\\.\PhysicalDrive1", etc.
 *
 * @return A list of Fastec formatted drives connected to the PC. Empty if
 *         no such drives are found.
 */
QStringList fiRawDrive::drives()
{
    QStringList     all_drives;
    QStringList     fastec_drives;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Get a list of all the physical drives on the system.
    all_drives = priv_ptr->all_physical_drives();

    // Which of those are removable?
    for (int i = 0; i < all_drives.count(); ++i)
    {
        // Is it a Fastec drive?
        if (is_fastec_drive(all_drives.at(i)))
        {
            // Found a drive.
            fastec_drives.append(all_drives.at(i));
        }
    }

    // Return our list of Fastec drives.
    return fastec_drives;
}

/*!
 * @brief Expand from the packed 10-bit format our cameras capture to a format we can
 *        actually use.
 *
 * @param [in]  buffer          Our packed data buffer.
 * @param [out] expanded_buf    The expanded data buffer.
 * @param [in]  needs_rotate    True if we need to rotate the bytes prior to expanding them.
 */
void fiRawDrive::expand_10Bits(QByteArray& buffer, QByteArray& expanded_buf, bool needs_rotate)
{
    bool need_alignment_shim;
    uint32_t row_byte_count;
    const uint8_t *buffer_ptr;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str << "need_rotate:" << needs_rotate;
#endif

    //-------------------------------------------------------------------
    // Statistics
#ifdef DO_STATS
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;

    elapsed_timer.start();
#endif
    // Statistics
    //-------------------------------------------------------------------

    // How many bytes of 16-bit pixels will we have per row.
    row_byte_count = metadata_ptr->roi_width * 2;

    // Make sure the row is 32-bit aligned.
    if ((row_byte_count & 3) != 0)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Need to 32-bit align the expanded row.";

        // We need an alignment "shim" to get the row to 32-bit alignment.
        need_alignment_shim = true;
        row_byte_count = ((row_byte_count >> 2) + 1) << 2;
    }
    else
    {
        // We're 32-bit aligned on the row. No shim needed.
        need_alignment_shim = false;
    }

    // Reserve the amount of space we'll need.
    std::unique_ptr<uint8_t[]> expanded_ptr(new uint8_t[(int)row_byte_count * metadata_ptr->roi_height]);

    // Setup our pointer.
    buffer_ptr = (const uint8_t *)buffer.constData();

    // Loop through the data, swapping the bytes.
// *INDENT-OFF*
#pragma omp parallel for collapse(2)
// *INDENT-ON*
    for (uint32_t row = 0; row < metadata_ptr->roi_height; ++row)
    {
        // Work through all the pixels in this row.
        for (uint32_t column = 0; column < metadata_ptr->roi_width; column += 3)
        {
            int pixel_offset;
            bool done;
            size_t buffer_index;
            uint16_t pixel_short;
            uint32_t packed_word;
            uint32_t current_pixel;

            // Not done yet!
            done = false;

            // Initialize the current pixel position in the column.
            current_pixel = column;

            // Where does the row start in the expanded buffer.
            pixel_offset = (int)(row * row_byte_count) + (int)(column * 2);

            // Where are we getting the raw data from? Remember we are reading
            // 3 pixels (columns) for each 32-bit raw word.
            buffer_index = row * ((metadata_ptr->roi_width + 2) / 3) * 4;
            buffer_index += ((current_pixel / 3) * 4);

            // Grab a word of data.
            memcpy(&packed_word, &buffer_ptr[buffer_index], sizeof(packed_word));

            // Do we need to re-order the word?
            if (needs_rotate)
            {
                // Get the byte ordering correct.
                packed_word = (uint32_t)qToBigEndian((uint32_t)packed_word);
            }

#ifdef RIGHT_SHIFT_PIXELS
            // Extract the pixel from the packed word. We want the
            // pixel to be right-justified in the word.
            pixel_short = (uint16_t)((packed_word & 0x3FF00000) >> 20);
#else
            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & 0x3FF00000) >> 14);
#endif /* RIGHT_SHIFT_PIXELS */

            expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
            expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

            // Bump our column count and see if we're done.
            current_pixel++;
            if (current_pixel == metadata_ptr->roi_width)
            {
                done = true;
            }
            else
            {
#if 0
                // An example for masking from various PIXEL positions
                9 8 7 6   5 4 3 2   1 0 9 8   7 6 5 4   3 2 1 0
                0 0 0 0   0 0 0 0   0 0 x x   x x x x   x x x x
                x x x x   x x x x   x x 0 0   0 0 0 0   0 0 0 0
                0 0 0 0   x x x x   x x x x   x x 0 0   0 0 0 0
#endif
#ifdef RIGHT_SHIFT_PIXELS
                // Extract the pixel from the packed word. We want the
                // pixel to be right-justified in the word.
                pixel_short = (uint16_t)((packed_word & 0xFFC00) >> 10);
#else
                // Extract the pixel from the packed word. We want the
                // pixel to be left-justified in the word.
                pixel_short = (uint16_t)((packed_word & 0xFFC00) >> 4);
#endif /* RIGHT_SHIFT_PIXELS */

                expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
                expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

                // Bump our column count and see if we're done.
                current_pixel++;
                if (current_pixel == metadata_ptr->roi_width)
                {
                    done = true;
                }
                else
                {
                    // Extract the pixel from the packed word. We want the
                    // pixel to be right-justified in the word.
#ifdef RIGHT_SHIFT_PIXELS
                    pixel_short = (uint16_t)(packed_word & 0x3FF);
#else
                    // Extract the pixel from the packed word. We want the
                    // pixel to be left-justified in the word.
                    pixel_short = (uint16_t)((packed_word & 0x3FF) << 6);
#endif /* RIGHT_SHIFT_PIXELS */

                    expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
                    expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

                    // Bump our column count and see if we're done.
                    current_pixel++;
                    if (current_pixel == metadata_ptr->roi_width)
                    {
                        done = true;
                    }
                }
            }

            // Do we need an alignment word?
            if (done && need_alignment_shim)
            {
                // Add an extra false "pixel" to pad the row to be 32-bit aligned.
                expanded_ptr[row_byte_count - 1] = 0;
                expanded_ptr[row_byte_count - 2] = 0;
            }
        }
    }

    // Move the data over.
    expanded_buf = QByteArray((const char *)expanded_ptr.get(), (int)row_byte_count * metadata_ptr->roi_height);

    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
#ifdef DO_STATS
    elapsed = elapsed_timer.elapsed();
    elapsed_time = elapsed_time.addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Duration:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
#endif
    // Statistics
    //-------------------------------------------------------------------
}

/*!
 * @brief Expand from the packed 12-bit format our cameras capture to a format we can
 *        actually use.
 *
 * @param [in]  buffer          Our packed data buffer.
 * @param [out] expanded_buf    The expanded data buffer.
 */
void fiRawDrive::expand_12Bits(QByteArray& buffer, QByteArray& expanded_buf, bool needs_swap)
{
    bool need_alignment_shim;
    uint32_t row_byte_count;
    uint8_t *buffer_ptr;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    //-------------------------------------------------------------------
    // Statistics
#ifdef DO_STATS
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;

    elapsed_timer.start();
#endif
    // Statistics
    //-------------------------------------------------------------------

    // How many bytes of 16-bit pixels will we have per row.
    row_byte_count = metadata_ptr->roi_width * 2;

    // Make sure the row is 32-bit aligned.
    if ((row_byte_count & 3) != 0)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Need to 32-bit align the expanded row.";

        // We need an alignment "shim" to get the row to 32-bit alignment.
        need_alignment_shim = true;
        row_byte_count = ((row_byte_count >> 2) + 1) << 2;
    }
    else
    {
        // We're 32-bit aligned on the row. No shim needed.
        need_alignment_shim = false;
    }

    // Reserve the amount of space we'll need.
    std::unique_ptr<uint8_t[]> expanded_ptr(new uint8_t[(int)row_byte_count * metadata_ptr->roi_height]);

    // Setup our pointer.
    buffer_ptr = (uint8_t *)buffer.data();

    // Loop through the data, swapping the bytes.
// *INDENT-OFF*
#pragma omp parallel for collapse(2)
// *INDENT-ON*
    for (uint32_t row = 0; row < metadata_ptr->roi_height; ++row)
    {
        // Work through all the pixels in this row.
        // Plus 5 because there are 5 * 12bits = 60 packed bits in 64 bits of a uint64_t?
        for (uint32_t column = 0; column < metadata_ptr->roi_width; column += 5)
        {
            int pixel_offset;
            bool done;
            size_t buffer_index;
            uint16_t pixel_short;
            uint64_t packed_word;
            uint32_t current_pixel;

            // Not done yet!
            done = false;

            // Initialize the current pixel position in the column.
            current_pixel = column;

            // Where does the row start in the expanded buffer.
            pixel_offset = (int)(row * row_byte_count) + (int)(column * 2);

            // Where are we getting the raw data from? Remember we are reading
            // five 12-bit (columns) for each 64-bit raw word.
            buffer_index = row * ((metadata_ptr->roi_width + 4) / 5) * 8;
            buffer_index += ((current_pixel / 5) * 8);

            // Grab a word of data. - packed_word is a uint64_t not a uint32_t?
            memcpy(&packed_word, &buffer_ptr[buffer_index], sizeof(packed_word));

            if (needs_swap)
            {
                uint64_t swapper;
                // Swap the 32-bit words around.
                swapper = packed_word << 32;
                packed_word = packed_word >> 32;
                packed_word |= swapper;
            }
            else
            {
                // Believe this is supposed to translate the entire 64 bits?
                // changing from this:
                //packed_word = (uint32_t)qToBigEndian((uint32_t)packed_word);
                // to this:
                uint32_t leftWord = (uint32_t) qFromBigEndian((uint32_t)(packed_word >> 32));
                uint32_t rightWord = (uint32_t) qFromBigEndian((uint32_t)(packed_word));
                packed_word = (uint64_t)leftWord | (uint64_t)((uint64_t)rightWord << 32);
            }

#ifdef RIGHT_SHIFT_PIXELS
            // Extract the pixel from the packed word.
            pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 48)) >> 48);
#else
            // Extract the pixel from the packed word. We want the
            // pixel to be left-justified in the word.
            pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 48)) >> 44);
#endif /* RIGHT_SHIFT_PIXELS */

            expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
            expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

            // Bump our column count and see if we're done.
            current_pixel++;
            if (current_pixel == metadata_ptr->roi_width)
            {
                done = true;
            }
            else
            {
#ifdef RIGHT_SHIFT_PIXELS
                // Extract the pixel from the packed word.
                pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 36)) >> 36);
#else
                // Extract the pixel from the packed word. We want the
                // pixel to be left-justified in the word.
                pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 36)) >> 32);
#endif /* RIGHT_SHIFT_PIXELS */

                expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
                expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

                // Bump our column count and see if we're done.
                current_pixel++;
                if (current_pixel == metadata_ptr->roi_width)
                {
                    done = true;
                }
                else
                {
#ifdef RIGHT_SHIFT_PIXELS
                    // Extract the pixel from the packed word.
                    pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 24)) >> 24);
#else
                    // Extract the pixel from the packed word. We want the
                    // pixel to be left-justified in the word.
                    pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 24)) >> 20);
#endif /* RIGHT_SHIFT_PIXELS */

                    expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
                    expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

                    // Bump our column count and see if we're done.
                    current_pixel++;
                    if (current_pixel == metadata_ptr->roi_width)
                    {
                        done = true;
                    }
                    else
                    {
#ifdef RIGHT_SHIFT_PIXELS
                        // Extract the pixel from the packed word.
                        pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 12)) >> 12);
#else
                        // Extract the pixel from the packed word. We want the
                        // pixel to be left-justified in the word.
                        pixel_short = (uint16_t)((packed_word & (((uint64_t)0xFFF) << 12)) >> 8);
#endif /* RIGHT_SHIFT_PIXELS */

                        expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
                        expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

                        // Bump our column count and see if we're done.
                        current_pixel++;
                        if (current_pixel == metadata_ptr->roi_width)
                        {
                            done = true;
                        }
                        else
                        {
#ifdef RIGHT_SHIFT_PIXELS
                            // Extract the pixel from the packed word.
                            pixel_short = (uint16_t)(packed_word & ((uint64_t)0xFFF));
#else
                            // Extract the pixel from the packed word. We want the
                            // pixel to be left-justified in the word.
                            pixel_short = (uint16_t)((packed_word & ((uint64_t)0xFFF)) << 4);
#endif /* RIGHT_SHIFT_PIXELS */

                            expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)pixel_short;
                            expanded_ptr[static_cast<uint64_t>(pixel_offset++)] = (uint8_t)(pixel_short >> 8);

                            // Bump our column count and see if we're done.
                            current_pixel++;
                            if (current_pixel == metadata_ptr->roi_width)
                            {
                                done = true;
                            }
                        }
                    }
                }
            }

            // Do we need an alignment word?
            if (done && need_alignment_shim)
            {
                // Add an extra false "pixel" to pad the row to be 32-bit aligned.
                expanded_ptr[row_byte_count - 2] = 0;
                expanded_ptr[row_byte_count - 1] = 0;
            }
        }
    }

    // Move the data over.
    expanded_buf = QByteArray((const char *)expanded_ptr.get(), (int)row_byte_count * metadata_ptr->roi_height);

    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
#ifdef DO_STATS
    elapsed = elapsed_timer.elapsed();
    elapsed_time = elapsed_time.addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Duration:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
#endif
    // Statistics
    //-------------------------------------------------------------------
}

/*!
 * @brief Expand the pixels from their captured size to be 16-bit. Also,
 *        convert from Big Endian to Little Endian data if appropriate.
 *
 * @param [in, out] buffer              The buffer with the data.
 * @param [in]      byte_count          The size of the image data.
 * @param [in]      needs_10bit_rotate  True if we need to rotate the packed 10-bit bytes prior to expanding them.
 * @param [in]      needs_12bit_swap    True if we need to swap the 12-bit bytes prior to expanding them.
 */
void fiRawDrive::expandThePixels(QByteArray& buffer, uint32_t byte_count, bool needs_10bit_rotate, bool needs_12bit_swap)
{
    QByteArray  expanded_buf;

    Q_UNUSED(byte_count);

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str << "size of image frame:" << byte_count << "needs_10bit_rotate:" << needs_10bit_rotate << "needs_12bit_swap:" << needs_12bit_swap;
#endif

    // Sanity check.
    if (metadata_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "metdata_ptr is nullptr!";
        return;
    }
    if (IS_8_BIT_MODE((char)metadata_ptr->bit_sel))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No expansion necessary!";
        return;
    }

    // Are we 10-bit or 12-bit?
    if (IS_10_BIT_MODE(metadata_ptr->bit_sel))
    {
        expand_10Bits(buffer, expanded_buf, needs_10bit_rotate);
    }
    else
    {
        expand_12Bits(buffer, expanded_buf, needs_12bit_swap);
    }

    // Copy the expanded buffer over.
    buffer.swap(expanded_buf);
}

/*!
 * @brief Return the number of metadata records that fit per sector.
 */
uint32_t fiRawDrive::pfmdPerSector()
{
    return (priv_ptr->get_sector_size() / GIGE_FRAME_METADATA_LEN);
}

/*!
 * @brief Return the 64-bit time extracted from a frame metadata record.
 *
 * @param metadata_buf  Buffer containing FMETA_SIZE_WORDS words of frame metadata.
 *
 * @return (1M * SBS) + SBUS
 */
uint64_t fiRawDrive::fmetaTime(uint32_t *metadata_buf)
{
    return ((uint64_t)metadata_buf[ FAS_FILESYSTEM::FMETA_TIME_SECONDS ] * (1000 * 1000)) +
           (uint64_t)metadata_buf[ FAS_FILESYSTEM::FMETA_TIME_USECS ];
}

/*!
 * @brief Return the correction table.
 */
const QByteArray& fiRawDrive::getCT() const
{
    return ct_buffer;
}

uint64_t fiRawDrive::getPFMDOffset(FAS_FILESYSTEM::LR_CAPTURE_HDR *cap, uint32_t frame_num)
{
    uint64_t meta_offset;

    // Calculate the byte offset to the metadata. We can be confident that
    // the PFMD is located 64 bytes from the end, because this is not a
    // series3 camera. We know that because the fmeta_lba field is 0.
    meta_offset = ((uint64_t)frame_num) * cap->frame_size;
    meta_offset += (cap->image_lba * priv_ptr->get_sector_size());
    meta_offset += (cap->frame_size - 64);

    // Return the offset.
    return meta_offset;
}

/*!
 * @brief Compute the position of the per-frame metadata for a specified frame
 *
 * @param frame_num The frame number whose metadata position is to be reported
 * @param absolute  true if frame_num is absolute, false if relative to oldest frame.
 * @param p_lba     On return, contains the lba of the sector containing that metadata
 * @param p_ofs     On return, contains the byte offset within that sector
 */
void fiRawDrive::getPFMDPosition(FAS_FILESYSTEM::LR_CAPTURE_HDR *cap, uint32_t frame_num, bool absolute, uint64_t *p_lba, uint32_t *p_ofs)
{
    uint32_t    abs_frame;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Figure out where the metadata for this frame is on the drive.
    abs_frame = convertToAbsFrameNum(frame_num, absolute);
    if (cap->fmeta_lba == 0)
    {
        uint64_t meta_offset;

        // Get the offset for the metadata associated with this frame.
        meta_offset = getPFMDOffset(cap, abs_frame);

        // Break the down into sector and offset in the sector. Let make sure we're not dividing by zero.
        if (priv_ptr->get_sector_size() == 0)
        {
            *p_lba = meta_offset;
            *p_ofs = 0;
        }
        else
        {
            *p_lba = meta_offset / priv_ptr->get_sector_size();
            *p_ofs = meta_offset % priv_ptr->get_sector_size();
        }
    }
    else
    {
        // Make sure we're not dividing by zero.
        if (pfmdPerSector() == 0)
        {
            *p_lba = abs_frame + cap->fmeta_lba;
            *p_ofs = 0;
        }
        else
        {
            *p_lba = (abs_frame / pfmdPerSector()) + cap->fmeta_lba;
            *p_ofs = (abs_frame % pfmdPerSector()) * GIGE_FRAME_METADATA_LEN;
        }
    }
}

/*!
 * @brief Return the number of frames in the capture.
 */
uint32_t fiRawDrive::getFrameCount()
{
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_info;

    // Make sure we have our partition information.
    if (part_info.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No partition information!?";
        return 0;
    }

    // Get a pointer to our capture header.
    cap_info = &part_info[0].caps[0].hdr.lr_cap;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str << "num_frames" << cap_info->num_frames;
#endif

    // Return the number of frames in the capture.
    return cap_info->num_frames;
}

bool fiRawDrive::isSeries3()
{
    bool series3_found = false;
    QString camera_model;

    // Sanity check. Probably the safe thing to do is return true.
    if (metadata_ptr == nullptr)
        return true;

    // Get the serial number in a format easier for us to work with.
    camera_model = QString(metadata_ptr->model_name);

    // Are we a "TSx" or "ILx" camera?
    // TODO: Replace with a RegEx.
    if (camera_model.contains(QStringLiteral("TS")))
        series3_found = true;
    else if (camera_model.contains(QStringLiteral("IL")))
        series3_found = true;

    // Return our result.
    return series3_found;
}

/*!
 * @brief Create a type appropriate image header for the raw data.
 *
 * If the sensor is mono and the capture was 8-bit, then we're doing a BMP
 * header; for other mono captures we use a TIF header.
 *
 * If the sensor is color, we're doing a DNG header.
 *
 * @param [in]  forced8Bits     True if mono sensor image was forced into 8-bits.
 * @param [in]  forceCodec      What codec is requested it is a way to override not just a forced 8-bit scenario but raw codec type.
 * @param [out] hdr_buffer      The buffer to hold the header information.
 * @param [out] codec_type      The type of codec to be used.
 *
 * @return  FAS_VIDEO::VIDEO_CODEC for the image header, FI_INVALID_CODEC
 * can be returned if it unable to create the header.
 */
FAS_VIDEO::VIDEO_CODEC  fiRawDrive::getImageHeader(bool forced8Bits, FAS_VIDEO::VIDEO_CODEC& forceCodec, QByteArray& hdr_buffer)
{
    uint32_t    pix_fourcc;
    std::unique_ptr<FAS_VIDEO::HEADER_DATA> image_hdr;
    FAS_VIDEO::VIDEO_CODEC fv_codec = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;

    bool modifiedBitSel = false;
    uint8_t saveBitSelIfForced8Bit = 0;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Make sure we have our metadata pointer.
    if (metadata_ptr == nullptr)
    {
        pm_metadata dummy_structure;        // Need a place to hold the initial data.

        // Initialize our capture metadata structure (the CLASS structure, we
        // don't much care about the passed in parameter.)
        if (!readPmMetadata(&dummy_structure))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the pm_metadata!";
            goto exit_get_image_header;
        }
    }

    // If we are forcing 8 bit image data here then the metadata we will use
    // must reflect this setup.
    if (forced8Bits)
    {
        saveBitSelIfForced8Bit = metadata_ptr->bit_sel;
        metadata_ptr->bit_sel = GIGE_BITMODE_UPPER_8_BITS;
        modifiedBitSel = true;
    }

    // Are we a color sensor?
    if ((metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX160_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR))
    {
        // Were we provided a codec to use for our raw data?
        if (forceCodec == FAS_VIDEO::FI_TIFF_STACK_CODEC)
        {
            fv_codec = forceCodec;
        }
        else
        {
            // Default raw codec is a DNG stack.  No need to check as we would
            // default to it anyway as nothing else is valid.
            fv_codec = FAS_VIDEO::FI_DNG_STACK_CODEC;
        }

        // What is our capture bit depth?
        if (IS_8_BIT_MODE((char)metadata_ptr->bit_sel))
        {
            // 8-bit pixels.
            pix_fourcc = FAS_VIDEO::FOURCC_BYR1;
        }
        else
        {
            // We make these bit depths 16 bit pixels.
            pix_fourcc = FAS_VIDEO::FOURCC_BYR2;
        }
    }
    // A Monochrome sensor was used to capture the data.
    else
    {
        // What is our capture depth?
        // Note: The caller may have forced an 8-bit condition even
        // though metadata indicates capture is greater than 8-bits.
        if (IS_8_BIT_MODE((char)metadata_ptr->bit_sel))
        {
            // Was a force codec provided?
            if (forceCodec == FAS_VIDEO::FI_TIFF_STACK_CODEC)
            {
                fv_codec = FAS_VIDEO::FI_TIFF_STACK_CODEC;
            }
            else
            {
                // For 8-bit mono captures, we're defaulting to gray BMP.
                fv_codec = FAS_VIDEO::FI_BMP_STACK_CODEC;
            }
            pix_fourcc = FAS_VIDEO::FOURCC_Y8;
        }
        else
        {
            // For larger than 8-bit mono captures, we use a TIF header.
            fv_codec = FAS_VIDEO::FI_TIFF_STACK_CODEC;
            pix_fourcc = FAS_VIDEO::FOURCC_BYR2;
        }
    }

    // Get the header.
    image_hdr = FAS_VIDEO::FiVideo::CreateImageHeader(fv_codec, pix_fourcc, metadata_ptr.get());
    if (image_hdr.get() == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get the header information!";
        goto exit_get_image_header;
    }

    // Resize our buffer.
    hdr_buffer = QByteArray(reinterpret_cast<const char *>(image_hdr->data()), (int)image_hdr->size());

exit_get_image_header:

    // Restore the bit_sel if it was modified here.
    if (modifiedBitSel)
    {
        metadata_ptr->bit_sel = saveBitSelIfForced8Bit;
    }

    // All done.
    return fv_codec;
}

/*!
 * @brief Read the sector containing metadata for a specific frame.
 *
 * @param cap       Capture header already recovered from the drive.
 * @param sector_buf    Sector buffer for reading metadata, must be big enough for a sector.
 * @param abs_frame Absolute frame number of the frame whose metadata is to be read.
 *
 * @return Pointer to frame metadata for abs_frame within sector_buf or nullptr if read failed.
 */
uint32_t *fiRawDrive::getDrivePFMD(FAS_FILESYSTEM::LR_CAPTURE_HDR *cap, uint32_t *sector_buf, uint32_t abs_frame)
{
    uint32_t *pfmd_ptr = nullptr;

    // Are we a new camera?
    if (cap->fmeta_lba == 0)
    {
        uint64_t meta_offset;

        // Get the offset to the metadata.
        meta_offset = getPFMDOffset(cap, abs_frame);

        // Read the data.
        if (priv_ptr->read_data(meta_offset, (uint8_t *)sector_buf, 64))
            pfmd_ptr = sector_buf;
    }
    else
    {
        uint64_t meta_lba;
        uint32_t meta_offset;

        // Where is the PFMD located on the drive.
        getPFMDPosition(cap, abs_frame, true, &meta_lba, &meta_offset);

        // Read the sector with the data.
        if (priv_ptr->read_sectors(meta_lba, (uint8_t *)sector_buf, 1))
        {
            // Point to the location in the buffer that contains the data.
            pfmd_ptr = (sector_buf + meta_offset);
        }
    }

//#define DUMP_PFMD 1
#ifdef DUMP_PFMD
    if (pfmd_ptr != nullptr)
    {
        QString msg_str;
        msg_str = tr("%1 %2 %3 %4").arg(pfmd_ptr[0], 8, 16, QChar('0')).arg(pfmd_ptr[1], 8, 16, QChar('0')).arg(pfmd_ptr[2], 8, 16, QChar('0')).arg(pfmd_ptr[3], 8, 16, QChar('0'));
        QLOG_TRACE() << __PRETTY_FUNCTION__ << msg_str;
        msg_str = tr("%1 %2 %3 %4").arg(pfmd_ptr[4], 8, 16, QChar('0')).arg(pfmd_ptr[5], 8, 16, QChar('0')).arg(pfmd_ptr[6], 8, 16, QChar('0')).arg(pfmd_ptr[7], 8, 16, QChar('0'));
        QLOG_TRACE() << __PRETTY_FUNCTION__ << msg_str;
    }
#endif

    return pfmd_ptr;
}

/*!
 * @brief Return the raw image pixels.
 */
const QByteArray& fiRawDrive::getRawImage() const
{

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    return raw_image_buf;
}


/*!
 * @brief Determines if the drive is, in fact, a Fastec formatted drive.
 *
 * @param drive     The physical drive identifier, aka "\\.\PhysicalDrive0".
 * @return True if the drive is Fastec formatted.
 */
bool fiRawDrive::is_fastec_drive(const QString& drive_uri)
{
    bool        rc;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Try to open the drive. It won't open if it's not a Fastec formatted drive.
    rc = open(drive_uri);

    // Close the drive.
    close();

    // Return our status
    return rc;
}

/*!
 * @brief Attempt to open the device specified by driver_uri. Sanity check to make
 *        sure the drive is a Fastec formatted drive.
 *
 * @param drive_uri     The URI of the drive we're interested in.
 * @return True on success, false otherwise.
 */
bool fiRawDrive::open(const QString& drive_uri)
{
    bool rc;
    uint64_t drv_space;
    std::string vol_label;
    pm_metadata dummy_structure;        // Need a place to hold the initial data.

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Assume this is not a Fastec drive.
    rc = false;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "drive_uri:" << drive_uri;

    // Make sure there's no capture information yet.
    metadata_ptr.reset();
    drive_str.clear();
    part_info.clear();

    // See if we can open the drive.
    if (!priv_ptr->open_device(drive_uri))
    {
        // Unable to open the device.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to open the device";
        goto open_bail;
    }

    // Query the drive for device specific information.
    if (!priv_ptr->query_device_info())
    {
        // Certainly not an error. There are physical devices, such as SDCard readers, which don't
        // have their media installed. They will fail here.
        goto open_bail;
    }

    // Flush anthing that may be hanging around.
    FAS_FILESYSTEM::FiFilesystem::getInstance()->clearCache(std::string());

    // Request an analysis.
    if (!FAS_FILESYSTEM::FiFilesystem::getInstance()->analyze(drive_uri.toStdString()))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "analyze() failed.";
        goto open_bail;
    }

    // We should now have information. Assuming we have a Fastec filesystem.
    if (!FAS_FILESYSTEM::FiFilesystem::getInstance()->getDriveInfo(drive_uri.toStdString(), part_info, drv_space, vol_label))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "getPartitionInfo() failed.";
        goto open_bail;
    }

    // For now, we're only interested in the first partition and the first capture.
    // Do some sanity checks.
    if (part_info.size() == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No partitions found.";
        goto open_bail;
    }
    if (part_info.at(0).part.type != FAS_FILESYSTEM::LR_PART_SIG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected type for the first partition:" << Qt::showbase << Qt::hex << part_info.at(0).part.type;
        goto open_bail;
    }
    if (part_info.at(0).caps.size() == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No captures found.";
        goto open_bail;
    }
    if (part_info.at(0).caps.at(0).type != FAS_FILESYSTEM::LR_CAPTURE_SIG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected type for the first capture:" << Qt::showbase << Qt::hex << part_info.at(0).caps.at(0).type;
        goto open_bail;
    }

    // Initialize our capture metadata structure (the CLASS structure, we
    // don't much care about the passed in parameter.)
    if (!readPmMetadata(&dummy_structure))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to read pm_metadata!";
        goto open_bail;
    }

    // Save the drive path.
    drive_str = drive_uri;

    // Looks good.
    rc = true;

open_bail:
    // Return our status.
    return rc;
}

// Create a define to enable/disable the omp parallel for collapse optimization.
// Uncomment it to enable omp parallel optimization and comment it to disable the omp optimization.
#define USE_OMP_PARALLEL 1
/*!
 * @brief Walks a row of 16 bits per pixel rows and trancates using the
 * most significant byte into 8 bits per pixel rows.
 *
 * @param buffer        The image buffer.
 */
void fiRawDrive::shrink16BitsTo8(QByteArray& buffer, const uint8_t *lookup_bcg)
{
    uint32_t rowByteCount;
    uint32_t frameCount;
    uint32_t row, column;
    bool haveBCGLookup = (lookup_bcg != nullptr) ? true : false;

    QByteArray the8BitArray;    //!< Will hold the shrunken frame data.

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // How many bytes of 16-bit pixels will we have per row.
    rowByteCount = metadata_ptr->roi_width * 2;

    // What is the expected frame count when 8 bpp.
    frameCount = metadata_ptr->roi_width * metadata_ptr->roi_height;

    // Reserve the amount of space we'll need.
    the8BitArray.resize((int)frameCount);

#ifdef USE_OMP_PARALLEL
// *INDENT-OFF*
#pragma omp parallel for collapse(2)
// *INDENT-ON*
#endif
    // Loop through the data, pulling only the 8 bits we need.
    for (row = 0; row < metadata_ptr->roi_height; ++row)
    {
        // Loop through the 16-bit columns.
        for (column = 0; column < metadata_ptr->roi_width; ++column)
        {
            // We are pulling the most-significant byte of a 16 bit word
            // assuming Big-endian memory layout.  Position the source byte
            // index value.
            uint32_t srcIndex = ((row * rowByteCount) + (column * 2)) + 1;

            // Read the byte at this location.
            uchar theByte = (uchar)buffer[srcIndex];
            if (haveBCGLookup)
            {
                theByte = lookup_bcg[theByte];
            }

            // Store the byte we pulled in our 8 bit frame.
            uint32_t destIndex = (row * metadata_ptr->roi_width) + column;
            the8BitArray[destIndex] = (char)theByte;
        }
    }

    // Swap the buffers installing the 8-bit frame into the buffer.
    buffer.swap(the8BitArray);
}

/*!
 * @brief Read the specified frame from the Fastec formatted drive and load it into
 *        the dest_buf object.
 *
 * @param [in]  frame_num       The absolute frame number to be read.
 * @param [in]  do_correction   True if we're supposed to do FPN correction, false otherwise.
 * @param [in]  lookup_bcg      A Brightness/Contrast/Gamma modification for frame.
 * @param [in]  force8BitMono   True if monochrome sensor and we must force 16 bits into 8.
 * @param [in/out] fvCodecPtr   Pointer to where we're to put the codec type for the data.
 * @param [out] dest_buf        Pointer to the QByteArray object for the data.
 *
 * @return True on success, false otherwise.
 */
bool fiRawDrive::readFrame(uint32_t frame_num, bool do_correction, const uint8_t *lookup_bcg,
                           bool force8BitMono, FAS_VIDEO::VIDEO_CODEC *fvCodecPtr, QByteArray *dest_buf)
{
    QByteArray  image_buf;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Pull off any user setting in case we are forcing 8-bits or we are requesting raw frame data.  We need
    // a bit more information to make decision when applying headers.
    FAS_VIDEO::VIDEO_CODEC forceCodec = *fvCodecPtr;

    // Set the default codec to an invalid state.
    *fvCodecPtr = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;

    // Read the raw image.
    if (!readRawImage(frame_num, image_buf))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read raw image for frame:" << frame_num;
        return false;
    }

    // Do we want to do FPN correction on the buffer?
    if (do_correction)
    {
        // Adjust the pixels in the buffer.
        applyFPN(image_buf);
    }

    // Determine if we need to force 16-bit frame to an 8-bit frame.
    if (force8BitMono)
    {
        // If the capture is only 8 bpp deep don't try to shrink up the data.
        force8BitMono = !IS_8_BIT_MODE((char)metadata_ptr->bit_sel);
        if (!force8BitMono)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Caller requested force8BitMono but capture is 8-bit capture! Ignoring request.";
        }
    }

    // Does the image data need to be shrunken down to 8bpp world?
    if (force8BitMono)
    {
        // Shrink up the actual image buffer and apply any bcg transitions
        // then it will be ready for the image header operation.
        shrink16BitsTo8(image_buf, lookup_bcg);
    }

    // #define ALM_NEW_DNG     1
#ifdef ALM_NEW_DNG
    // Are we a color sensor?
    if ((metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX160_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR) ||
        (metadata_ptr->sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR))
    {
        // All color sensor will use a DNG codec.
        *fvCodecPtr = FAS_VIDEO::VIDEO_CODEC::FI_DNG_STACK_CODEC;

        // Build up our DNG "file" in memory.
        if (!build_dng(frame_num, image_buf, dest_buf))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to build DNG.";
            return false;
        }
    }
    else
#endif
    {
        // Get an image header appropriate for the data which also returns codec type.
        *fvCodecPtr = getImageHeader(force8BitMono, forceCodec, *dest_buf);
        if (*fvCodecPtr == FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get an image header!";
            return false;
        }

        // Append the image buffer to our header.
        dest_buf->append(image_buf);
    }

    // All done.
    return true;
}

/*!
 * @brief Read the specified frame from the Fastec formatted drive and load it into
 *        the dest_pixmap object.
 *
 * @param [in]  frame_num       The absolute frame number to be read.
 * @param [in]  do_correction   True if we're supposed to do FPN correction, false otherwise.
 * @param [out] dest_pixmap     Where to put the pixmap data.
 *
 * @return True on success, false otherwise.
 */
bool fiRawDrive::readFrame(uint32_t frame_num, bool do_correction, fiPixmap *dest_pixmap)
{
    QByteArray  image_hdr;
    FAS_VIDEO::VIDEO_CODEC fvCodec = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str << "do_correction:" << do_correction;
#endif

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Read the raw image.
    if (!readRawImage(frame_num, raw_image_buf))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read raw image for frame:" << frame_num;
        return false;
    }

    // Do we want to do FPN correction on the buffer?
    if (do_correction)
    {
        // Adjust the pixels in the buffer.
        applyFPN(raw_image_buf);
    }

    // Get an image header appropriate for the data.
    fvCodec = getImageHeader(false, fvCodec, image_hdr);
    if (fvCodec == FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get an image header!";
        return false;
    }

    // Append the data to the header and parse it.
    QByteArray full_image = image_hdr.append(raw_image_buf);

//#define DO_SAVE_FILE 1
#ifdef DO_SAVE_FILE
    {
        QString file_name = "./test_image_" + QString::number(frame_num) + ".dng";
        QFile new_doc(file_name);
        if (new_doc.open(QIODevice::WriteOnly))
            new_doc.write(full_image);
        new_doc.close();
    }
#endif

    if (!dest_pixmap->load_from_buf(full_image, fvCodec))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the image!";
        return false;
    }

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif

    // We're all good.
    return true;
}

/*!
 * @brief Read the capture metadata from the previously opened Fastec formatted dirve.
 * @param metadata_ptr  Pointer for the storage location of the metadata.
 * @return True on success, false otherwise.
 */
bool fiRawDrive::readPmMetadata(struct pm_metadata_type *metadata_ptr)
{
    bool        rc = false;
    uint32_t    buf_size;
    uint32_t    num_sectors;
    uint32_t    pm_bytes;
    uint64_t    caphd_lba;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_info;
    std::unique_ptr<uint8_t[]> buffer_ptr;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Sanity check.
    if (metadata_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null metadata_ptr passed in!";
        return false;
    }

    // Do we already have the data loaded?
    if (this->metadata_ptr != nullptr)
    {
        // Copy the data over.
        memcpy(metadata_ptr, this->metadata_ptr.get(), sizeof(pm_metadata));

        // All done.
        return true;
    }

    // Make sure we have our partition information.
    if (part_info.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No partition information!?";
        return false;
    }

    // Get a pointer to our capture header.
    caphd_lba = part_info[0].caps[0].lba;
    cap_info = &part_info[0].caps[0].hdr.lr_cap;

    // Is the drive open?
    if (priv_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No private pointer!?";
        return false;
    }
    if (!priv_ptr->is_device_open())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device not opened!";
        return false;
    }

    // Allocate enough buffer to read the PM.
    buf_size = cap_info->pm_offset + cap_info->pm_bytes;
    num_sectors = bytesToSectors(buf_size);
    buf_size = num_sectors * priv_ptr->get_sector_size();
    // TODO Optimize by taking advantage of the offset being on a sector boundary
    // so we only need to read the PM and not reread the capture header.

    // Allocate our buffer.
    buffer_ptr = std::make_unique<uint8_t[]>(buf_size);
    if (buffer_ptr.get() == nullptr)
        return false;

    // Read the data.
    if (!priv_ptr->read_sectors(caphd_lba, buffer_ptr.get(), num_sectors))
        goto pm_bail;

    pm_bytes = cap_info->pm_bytes;
    if (pm_bytes != sizeof(pm_metadata))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected PM size on SSD, expected:" << sizeof(pm_metadata) << "got:" << cap_info->pm_bytes;

        // Did we get anything?
        if (cap_info->pm_bytes == 0)
            goto pm_bail;

        if (pm_bytes < sizeof(pm_metadata))
        {
            // If what is on the disk is smaller than expected, zero the extra space.
            memset(metadata_ptr, 0, sizeof(pm_metadata));
        }
        else
        {
            // If if it is larger than expected, only copy what we know about.
            pm_bytes = sizeof(pm_metadata);
        }
    }
    memcpy(metadata_ptr, buffer_ptr.get() + cap_info->pm_offset, pm_bytes);

    // Now fill in the fields that were updated post-capture in the capture header
    // but not re-written in the PM.
    metadata_ptr->frame_count = cap_info->num_frames;   // Number captured, not allocated.
    metadata_ptr->first_frame = cap_info->first_frame_num;
    metadata_ptr->trig_time_s = cap_info->trigger_time_s;
    metadata_ptr->trig_time_ns = cap_info->trigger_time_ns;

    // Everywhere except pm_metadata we maintain the trigger frame as relative to
    // the first frame number. But in pm_metadata it is an absolute frame number
    // so we need to convert from relative to absolute.
    if ((cap_info->trigger_frame_num + cap_info->first_frame_num) < cap_info->frame_count)
        metadata_ptr->trig_frame = cap_info->trigger_frame_num + cap_info->first_frame_num;
    else
        metadata_ptr->trig_frame = (cap_info->trigger_frame_num + cap_info->first_frame_num) - cap_info->frame_count;

    // Allocate our copy. This will also delete any previous copy.
    this->metadata_ptr.reset(new pm_metadata);
    if (this->metadata_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a pm_metadata object!";
        goto pm_bail;
    }

    // Copy the data over.
    memcpy(this->metadata_ptr.get(), metadata_ptr, sizeof(pm_metadata));

    // All is well...
    rc = true;

pm_bail:
    // Return our status.
    return rc;
}

/*!
 * @brief Load the raw image from the Fastec drive into the raw_image buffer.
 *
 * @param [in]  frame_num       The absolute frame number of the image we want.
 * @param [out] raw_image       The destination buffer for the raw image.
 * @return True on success, false otherwise.
 */
bool fiRawDrive::readRawImage(uint32_t frame_num, QByteArray& raw_image)
{
    uint32_t    abs_frame;
    uint64_t    frame_lba;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_info;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;

    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------

    // Sanity check.
    if (priv_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No priv_ptr!?";
        return false;
    }

    // Make sure we have our partition information.
    if (part_info.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No partition information!?";
        return false;
    }

    // Get a pointer to our capture header.
    cap_info = &part_info[0].caps[0].hdr.lr_cap;

    // Do we have the capture metadata loaded?
    if (metadata_ptr == nullptr)
    {
        pm_metadata dummy_structure;        // Need a place to hold the initial data.

        // Initialize our capture metadata structure (the CLASS structure, we
        // don't much care about the passed in parameter.)
        if (!readPmMetadata(&dummy_structure))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the pm_metadata!";
            return false;
        }
    }

    // Get the absolute frame number.
    abs_frame = convertToAbsFrameNum(frame_num, false);

    // Note:  Background, for series3 cameras the pfmd (per frame meta data) is stored in block
    // together on lr disk but for h series cameras the pfmd is written at the end of every frame.
    // The test below is looking at the fmeta_lba (file's metadata logical block address) and if
    // it is set to zero it is always going to be a non-series3 camera as series3 cameras will have
    // this information populated with the special pfmd file's logical block address.
    if (cap_info->fmeta_lba == 0)
    {
        uint64_t frame_offset;
        uint32_t frame_size;

        // We want the frame size without the metadata at the end.
        // Note:  For series3 cameras the metadata length is only 32 bytes but
        // for h-series cameras it twice that length.  Because this pathway
        // through the read is only for h-series cameras we need to make it
        // 2 X pfmd length as you see below.
        frame_size = cap_info->frame_size - (2 * GIGE_FRAME_METADATA_LEN);

        // Setup our buffer to be only as large as the frame of data minus any pfmd.
        raw_image.resize(frame_size);

        // Get the offset to the image data.
        frame_offset = ((uint64_t)abs_frame) * cap_info->frame_size;
        frame_offset += (cap_info->image_lba * priv_ptr->get_sector_size());

        // Read the data.
        if (!priv_ptr->read_data(frame_offset, (uint8_t *)raw_image.data(), frame_size))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read frame:" << frame_num << "frame_offset:" << frame_offset;
            return false;
        }
    }
    else
    {
        // Get our LBA.
        frame_lba = (abs_frame * cap_info->frame_sectors) + cap_info->image_lba;

        // Make sure our buffer is big enough.
        raw_image.resize((int)(cap_info->frame_sectors * priv_ptr->get_sector_size()));

#if 0
        //============================================================================
        // TEST TEST TEST
        QElapsedTimer   read_timer;
        read_timer.start();
#endif

        // Read the data.
        if (!priv_ptr->read_sectors(frame_lba, (uint8_t *)raw_image.data(), cap_info->frame_sectors))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read frame:" << frame_num << "abs_frame:" << abs_frame << "frame_lba:" << frame_lba;
            return false;
        }

#if 0
        qint64 elapsed = read_timer.elapsed();
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Read took" << elapsed << "milliseconds, bytes/sec:" << (((qint64)(cap_info->frame_sectors * 512000)) / elapsed);
        //============================================================================
#endif

        // Sanity check.
        if ((cap_info->frame_size & 3) != 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Frame size:" << cap_info->frame_size << "is not word aligned!";
        }
    }

    // If the image is greater than 8-bit, then we need to expand the pixels to be 16-bit.
    if (IS_8_BIT_MODE((char)metadata_ptr->bit_sel))
    {
        // Swap the bytes as appropriate for each camera system.
        if (isSeries3())
            swapTheBytes((uint8_t *)raw_image.data(), (uint32_t)ClientMath::qword_to_bytes((uint64_t)metadata_ptr->frame_size));
        else
            swapTheBytes((uint8_t *)raw_image.data(), raw_image.size());
    }
    else
    {
        // This will expand and swap, if necessary.
        if (isSeries3())
            expandThePixels(raw_image, (uint32_t)ClientMath::qword_to_bytes((uint64_t)metadata_ptr->frame_size), false, true);
        else
            expandThePixels(raw_image, raw_image.size(), false, true);
    }

    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = elapsed_time.addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Duration:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------

    // All done.
    return true;
}


/*!
 * @brief Read the per-frame metadata from the previously opened Fastec formatted dirve.
 * @param buffer    QByteArray to store the data.
 * @return The number of bytes in the buffer, 0 on error.
 */
uint32_t fiRawDrive::readPerFrameMetadata(QByteArray& raw_buffer)
{
    uint32_t    return_count = 0;
    uint32_t    raw_bufsize;
    uint64_t    start_lba;
    uint32_t    start_offset;
    uint64_t    end_lba;
    uint32_t    end_offset;
    FAS_FILESYSTEM::LR_CAPTURE_HDR *cap_info;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Make sure we have our partition information.
    if (part_info.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No partition information!?";
        return 0;
    }

    // Get a pointer to our capture header.
    cap_info = &part_info[0].caps[0].hdr.lr_cap;

    // Is the drive open?
    if (priv_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No private pointer!?";
        return 0;
    }
    if (!priv_ptr->is_device_open())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device not opened!";
        return 0;
    }

    // We can tell if we are a series3 or a seriesNC by looking at the fmeta_lba field.
    // The newer cameras do not use this; the PFMD for a frame is immediately located
    // after the image data for the newer cameras.
    if (cap_info->fmeta_lba == 0)
    {
        uint32_t loop_frame;
        QByteArray sector_buf;

        // Reset our return buffer.
        raw_buffer.clear();

        // Make sure our sector buffer is big enough.
        sector_buf.resize(priv_ptr->get_sector_size());

        // Loop through the frames and read up the metadata.
        for (loop_frame = 0; loop_frame < cap_info->num_frames; ++loop_frame)
        {
            uint32_t *meta_ptr;

            // Yield some time every so often.
            if ((loop_frame % YIELD_THRESHOLD) == 0)
                QApplication::processEvents();

            // Read the PFMD for this frame.
            meta_ptr = getDrivePFMD(cap_info, (uint32_t *)sector_buf.data(), loop_frame);
            if (meta_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "get_metadata() failed!";
                break;
            }

            // Append the PFMD to our buffer. The seriesNC cameras have 64 bytes of PFMD.
            raw_buffer.append((const char *)meta_ptr, 64);
        }
        if (loop_frame != cap_info->num_frames)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to load the PFMD.";
            return_count = 0;
            raw_buffer.clear();
        }
        else
        {
            // Looking good.
            return_count = raw_buffer.length();
        }
    }
    else
    {
        // Series3 camera filesystem.
        // Figure the number of bytes/sectors to encompass the full per-frame metadata information.
        raw_bufsize = (bytesToSectors(cap_info->num_frames * GIGE_FRAME_METADATA_LEN) + 2) * priv_ptr->get_sector_size();

        // Make sure our buffer is big enough.
        raw_buffer.resize((int)raw_bufsize);

        // Locate the metadata for these frames on the drive.
        getPFMDPosition(cap_info, 0, false, &start_lba, &start_offset);
        getPFMDPosition(cap_info, cap_info->num_frames, false, &end_lba, &end_offset);

        // If the end is past the start, we can just read it all at once.
        if ((end_lba > start_lba) ||
            ((end_lba == start_lba) && (end_offset > start_offset)))
        {
            if (!priv_ptr->read_sectors(start_lba, (uint8_t *)raw_buffer.data(), (uint32_t)((end_lba - start_lba) + 1)))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read from start:" << start_lba << "for count:" << ((end_lba - start_lba) + 1);
                goto raw_bail;
            }

            // Do we need to slide things around?
            if (start_offset != 0)
            {
                // Slide the data up in the buffer.
                memmove(raw_buffer.data(), raw_buffer.data() + start_offset, (cap_info->num_frames * GIGE_FRAME_METADATA_LEN));
            }
        }

        // We have a wrap situation so we need to do two IOs.
        // Our first and last frames are probably both in the same sector.
        else
        {
            uint64_t    abs_lba;
            uint32_t    abs_offset;
            uint8_t     *append_ptr;

            // Locate the last metadata record, the one written immediately before wrapping.
            getPFMDPosition(cap_info, cap_info->frame_count - 1, true, &abs_lba, &abs_offset);

            // The offset for the last frame must be the last one in the sector!
            // If not, we'll get crazy data.
            if (abs_offset != (priv_ptr->get_sector_size() - GIGE_FRAME_METADATA_LEN))
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString("Got crazy numbers, lba=%1 ofs=%2 for last frame #%3 metadata!").arg(abs_lba).arg(abs_offset).arg(cap_info->frame_count - 1);
            }

            // Read the data.
            if (!priv_ptr->read_sectors(start_lba, (uint8_t *)raw_buffer.data(), (uint32_t)((abs_lba - start_lba) + 1)))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read from start:" << start_lba << "for count:" << ((abs_lba - start_lba) + 1);
                goto raw_bail;
            }

            // Do we need to slide things around?
            if (start_offset != 0)
            {
                // Slide the data up in the buffer.
                memmove(raw_buffer.data(), raw_buffer.data() + start_offset, (cap_info->num_frames * GIGE_FRAME_METADATA_LEN));
            }

            // Get the pointer for the next "chunk".
            append_ptr = (uint8_t *)raw_buffer.data() + (((abs_lba - start_lba) * priv_ptr->get_sector_size()) + abs_offset);

            // Locate the first metadata record.
            getPFMDPosition(cap_info, 0, true, &abs_lba, &abs_offset);

            // The offset for the absolute first frame must be the first one in the sector!
            // If not, we'll get crazy data.
            if (abs_offset != 0)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString("Got crazy numbers, lba=%1 ofs=%2 for first frame #0 metadata!").arg(abs_lba).arg(abs_offset);
            }

            // Read the next "chunk" of data.
            if (!priv_ptr->read_sectors(abs_lba, append_ptr, (uint32_t)((end_lba - abs_lba) + 1)))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read from start:" << abs_lba << "for count:" << ((end_lba - abs_lba) + 1);
                goto raw_bail;
            }
        }

        // Looking good.
        return_count = cap_info->num_frames * GIGE_FRAME_METADATA_LEN;

        // Swap the bytes in the raw metadata buffer.
        swapTheBytes((uint8_t *)raw_buffer.data(), return_count);
    }

raw_bail:
    // Return our status.
    return return_count;
}

/*!
 * @brief Load correction table for a capture.
 * @return True on success, false otherwise.
 */
bool fiRawDrive::recoverCT()
{
    uint64_t    caphd_lba;
    std::string dev_path;
    std::vector<uint8_t> xfer_vec;

#if RAWDRIVE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << log_entry_str;
#endif

    // Make sure we have our partition information.
    if (part_info.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No partition information!?";
        return false;
    }

    // Get a pointer to our capture header.
    caphd_lba = part_info[0].caps[0].lba;
    //cap_info = &part_info[0].caps[0].hdr.lr_cap;
    dev_path = drive_str.toStdString();

    // Read the correction table from the filesystem.
    if (!FAS_FILESYSTEM::FiFilesystem::getInstance()->getCaptureCT(dev_path, caphd_lba, xfer_vec))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FiFileSystem's getCaptureCT failed!";
        return false;
    }

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FiFileSystem's getCaptureCT reports size:" << xfer_vec.size();

    // Make sure we got something.
    if (xfer_vec.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to recover correction table!";
        return false;
    }

    // Transfer the correction table into the ct_buffer.
    ct_buffer.clear();
    ct_buffer.append((const char *)xfer_vec.data(), xfer_vec.size());

    // Now do the same for the PRNU gain table.

    // Read the gain table from the filesystem.  The xfer_vec will be cleared inside of the call before setting it.
    if (!FAS_FILESYSTEM::FiFilesystem::getInstance()->getCaptureGT(dev_path, caphd_lba, xfer_vec))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FiFileSystem's getCaptureGT failed!";
        return false;
    }

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FiFileSystem's getCaptureGT reports size:" << xfer_vec.size();

    // Make sure we got something.
    if (xfer_vec.empty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to recover PRNU gain table!";
        return false;
    }

    // Copy the PRNU correction data.
    prnu_buffer.clear();
    prnu_buffer.append((const char *)xfer_vec.data(), xfer_vec.size());

    // All done.
    return true;
}

/*!
 * @brief Convert from Big Endian to Little Endian data.
 * @param buffer        The buffer with the data.
 * @param byte_count    The number of 8-bit entries to be swapped.
 */
void fiRawDrive::swapTheBytes(uint8_t *buffer, uint32_t byte_count)
{
    uint32_t index;
    uint32_t long_count;

    // Initialize our count.
    long_count = (byte_count + (sizeof(uint32_t) - 1)) / sizeof(uint32_t);

    // Loop through the data, swapping the bytes.
// *INDENT-OFF*
#pragma omp parallel for
// *INDENT-ON*
    for (index = 0; index < long_count; ++index)
    {
        uint32_t temp_word;

        // Swap the bytes.
        memcpy(&temp_word, &buffer[index * sizeof(temp_word)], sizeof(temp_word));
        temp_word = qFromBigEndian(temp_word);
        memcpy(&buffer[index * sizeof(temp_word)], &temp_word, sizeof(temp_word));
    }
}

/*!
 * @brief Swap words of data.
 * @param buffer        The buffer with the data.
 */
void fiRawDrive::swapTheWords(uint32_t *buffer, uint32_t word_count)
{
    uint32_t    index;

    // Loop through the data, swapping the words.
    for (index = 0; index < word_count; index += 4)
    {
        uint32_t    temp_word;
        temp_word = buffer[index];
        buffer[index] = buffer[index + 3];
        buffer[index + 3] = temp_word;
        temp_word = buffer[index + 1];
        buffer[index + 1] = buffer[index + 2];
        buffer[index + 2] = temp_word;
    }
}

//=============================================================================
// fiRawDrivePrivate implementation(s)
//=============================================================================

#ifdef Q_OS_WIN

fiRawDrivePrivate::fiRawDrivePrivate(QObject *parent) :
    QObject(parent),
    device_handle(INVALID_HANDLE_VALUE)
{
}

fiRawDrivePrivate::~fiRawDrivePrivate()
{
    if (device_handle != INVALID_HANDLE_VALUE)
    {
        CloseHandle(device_handle);
        device_handle = INVALID_HANDLE_VALUE;
    }
}

/*!
 * @brief Detect all physical drives in the system, mounted or not.
 * @return A list of URIs for the drives found.
 */
QStringList fiRawDrivePrivate::all_physical_drives()
{
    TCHAR   *phys_list;
    DWORD   list_len;
    DWORD   result;
    QString dev_name;
    QStringList drive_list;

    // Start out with a pretty big buffer.
    list_len = INIT_LIST_LEN;
    result = 0;
    drive_list.clear();

    // Loop until we get something valid.
    while (1)
    {
        // Allocate our lists.
        phys_list = new TCHAR[ list_len ];

        // Get our list.
        result = QueryDosDevice(nullptr, phys_list, list_len);
        if (result == 0)
        {
            DWORD   last_err;

            // What was the error?
            last_err = GetLastError();
            if (last_err == ERROR_INSUFFICIENT_BUFFER)
            {
                // Release our buffer.
                delete []phys_list;

                // Increased our buffer size.
                list_len *= 4;
            }
            else
            {
                LPTSTR  error_message;
                DWORD   error_code = GetLastError();

                FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS,
                              nullptr, error_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                              (LPTSTR)&error_message, 0, nullptr);

                // Log the error.
                QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("QueryDosDevice() failed. Error %1: %2")
                             .arg(error_code).arg(QString::fromWCharArray(error_message));
                LocalFree(error_message);

                // We're done.
                break;
            }
        }
        else
        {
            // We're done.
            break;
        }
    }

    // Did we get anything?
    if (result != 0)
    {
        // Walk our list, building up our return information.
        dev_name.clear();

        // Use a list_ptr to walk the string list. Don't forget to increment the pointer
        // past the null terminator!
        for (TCHAR *list_ptr = phys_list; *list_ptr != 0; list_ptr += (dev_name.length() + 1))
        {
            // Get our device name.
            dev_name = QString::fromWCharArray(list_ptr);

            // Is this a "physical" device?
            if (dev_name.contains(PHYSICAL_STR))
            {
                QString device_uri;

                // Building our device name into an URI.
                device_uri = QString::fromUtf8("\\\\.\\%1").arg(dev_name);

                // Add it to our list.
                drive_list.append(device_uri);
            }
        }
    }

    // Release our buffer.
    delete []phys_list;

    // Return our list.
    return drive_list;
}

void fiRawDrivePrivate::close_device()
{
    // Sanity check.
    if (device_handle != INVALID_HANDLE_VALUE)
    {
        CloseHandle(device_handle);
        device_handle = INVALID_HANDLE_VALUE;
    }
}

bool fiRawDrivePrivate::open_device(const QString& device_uri)
{
    // Sanity check.
    if (device_handle != INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is valid!?";
    }

    // Open the device.
    device_handle = CreateFile((const wchar_t *)device_uri.utf16(), GENERIC_READ, FILE_SHARE_READ /* | FILE_SHARE_WRITE */, nullptr, OPEN_EXISTING, 0, nullptr);
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        LPTSTR  error_message;
        DWORD   error_code = GetLastError();

        if (error_code != EIO)
        {
            FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS,
                          nullptr, error_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                          (LPTSTR)&error_message, 0, nullptr);

            QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to open %1. Error %2: %3")
                         .arg(device_uri).arg(error_code).arg(QString::fromWCharArray(error_message));
            LocalFree(error_message);
        }

        // Didn't work.
        return false;
    }

    // Good to go.
    return true;
}

/*!
 * @brief Query the opened device for it's drive geometry.
 */
bool fiRawDrivePrivate::query_device_info()
{
    DWORD   junk;
    BOOL    bResult;

    // Sanity check.
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is invalid!";
        return false;
    }

    // Clear our geometry structure.
    memset(&geometry, 0, sizeof(geometry));

    // Query the device for the information.
    bResult = DeviceIoControl(device_handle, IOCTL_DISK_GET_DRIVE_GEOMETRY, nullptr, 0, &geometry, sizeof(geometry), &junk, nullptr);
    if (!bResult)
    {
        LPTSTR  error_message;
        DWORD   error_code = GetLastError();

        // Skip these errors.
        if (error_code != ERROR_NOT_READY)
        {
            FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS,
                          nullptr, error_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                          (LPTSTR)&error_message, 0, nullptr);

            // Log the message - may not be an "error" though. Device Not Ready, for example, will happen when
            // a SDCard reader doesn't have an SDCard.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to query device. Error %1: %2")
                         .arg(error_code).arg(QString::fromWCharArray(error_message));
            LocalFree(error_message);
        }

        // Nothing here.
        return false;
    }

    // Looks like we're good.
    return true;
}

/*!
 * @brief Read data from the drive.
 * @param offset            Where on the drive to read the data from.
 * @param data_ptr          Where to put the data.
 * @param bytes_to_read     How many bytes to read.
 *
 * @return True if the read is successful, false otherwise.
 */
bool fiRawDrivePrivate::read_data(uint64_t offset, uint8_t *data_ptr, uint64_t bytes_to_read)
{
    DWORD   bytes_read;
    DWORD   sfp_result;
    BOOL    bResult;
    LARGE_INTEGER   li_offset;

    // Sanity check.
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is invalid!";
        return false;
    }
    if (data_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "data_ptr is nullptr!";
        return false;
    }

    // Convert our starting offset to a Windows happy value.
    li_offset.QuadPart = (int64_t)offset;

    // Adjust the location on the drive.
    sfp_result = SetFilePointer(device_handle, (int32_t)li_offset.LowPart, &li_offset.HighPart, FILE_BEGIN);
    if (sfp_result == INVALID_SET_FILE_POINTER)
    {
        LPTSTR  error_message;
        DWORD   error_code = GetLastError();

        // Skip these errors.
        if (error_code != NO_ERROR)
        {
            FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS,
                          nullptr, error_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                          (LPTSTR)&error_message, 0, nullptr);

            // Log the message - may not be an "error" though. Device Not Ready, for example, will happen when
            // a SDCard reader doesn't have an SDCard.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to query device. Error %1: %2")
                         .arg(error_code).arg(QString::fromWCharArray(error_message));
            LocalFree(error_message);
        }
    }

    // Read the block(s).
    bResult = ReadFile(device_handle, data_ptr, bytes_to_read, &bytes_read, nullptr);

    // Return the result.
    return (bResult == TRUE);
}

/*!
 * @brief Read one or more sectors of data from the drive.
 *
 * @param[out]  data_ptr        Where to put the data.
 * @param[in]   start_sector    The starting sector number.
 * @param[in]   num_sectors     The number of sectors to read.
 *
 * @return True if the read is successful, false otherwise.
 */
bool fiRawDrivePrivate::read_sectors(uint64_t start_sector, uint8_t *data_ptr, uint32_t num_sectors)
{
    DWORD   bytes_to_read;
    DWORD   bytes_read;
    DWORD   sfp_result;
    BOOL    bResult;
    LARGE_INTEGER   li_offset;

    // Sanity check.
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is invalid!";
        return false;
    }
    if (data_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "data_ptr is nullptr!";
        return false;
    }

    // How many bytes do we want to read?
    bytes_to_read = geometry.BytesPerSector * (DWORD)num_sectors;

    // Convert our starting offset to a Windows happy value.
    li_offset.QuadPart = (int64_t)(start_sector * geometry.BytesPerSector);

    // Adjust the location on the drive.
    sfp_result = SetFilePointer(device_handle, (int32_t)li_offset.LowPart, &li_offset.HighPart, FILE_BEGIN);
    if (sfp_result == INVALID_SET_FILE_POINTER)
    {
        LPTSTR  error_message;
        DWORD   error_code = GetLastError();

        // Skip these errors.
        if (error_code != NO_ERROR)
        {
            FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS,
                          nullptr, error_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                          (LPTSTR)&error_message, 0, nullptr);

            // Log the message - may not be an "error" though. Device Not Ready, for example, will happen when
            // a SDCard reader doesn't have an SDCard.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to query device. Error %1: %2")
                         .arg(error_code).arg(QString::fromWCharArray(error_message));
            LocalFree(error_message);
        }
    }

    // Read the block(s).
    bResult = ReadFile(device_handle, data_ptr, bytes_to_read, &bytes_read, nullptr);

    // Return the result.
    return (bResult == TRUE);
}

#endif // Q_OS_WIN
#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )

fiRawDrivePrivate::fiRawDrivePrivate(QObject *parent) :
    QObject(parent),
    device_handle(INVALID_HANDLE_VALUE),
    bytes_per_sector(0)
{
}

fiRawDrivePrivate::~fiRawDrivePrivate()
{
    if (device_handle != INVALID_HANDLE_VALUE)
    {
        close(device_handle);
        device_handle = INVALID_HANDLE_VALUE;
    }
}

/*!
 * @brief Detect all physical drives in the system, mounted or not.
 * @return A list of URIs for the drives found.
 */
QStringList fiRawDrivePrivate::all_physical_drives()
{
    int     i;
    QStringList temp_list;

#ifdef Q_OS_MAC
    QDir    dev_dir("/dev/");
    QStringList drive_list;

    // Need a string list for the files we want to find.
    temp_list << QString::fromUtf8("disk*");

    // Return a list of drives/disks we might be interested in.
    drive_list = dev_dir.entryList(temp_list, QDir::AllEntries | QDir::Hidden | QDir::System);

    // Walk through the list of found entries (if any) and prepend the path.
    temp_list.clear();
    for (i = 0; i < drive_list.count(); ++i)
    {
        temp_list.append(QString::fromUtf8("/dev/") + drive_list.at(i));
    }
#else
    QDir    dev_dir("/sys/block");
    QStringList drive_list;

    // Need a string list for the files we want to find.
    temp_list << QString::fromUtf8("*");

    // Return a list of drives/disks we might be interested in.
    drive_list = dev_dir.entryList(temp_list, QDir::AllEntries | QDir::Hidden | QDir::System | QDir::NoDotAndDotDot);
    QLOG_DEBUG() << "drive_list:" << drive_list;

    // Walk through the list of found entries (if any) and prepend the path.
    temp_list.clear();
    for (i = 0; i < drive_list.count(); ++i)
    {
        temp_list.append(QString::fromUtf8("/dev/") + drive_list.at(i));
    }
    QLOG_DEBUG() << "temp_list:" << temp_list;
#endif

    // Return our list.
    return temp_list;
}

void fiRawDrivePrivate::close_device()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (device_handle != INVALID_HANDLE_VALUE)
    {
        close(device_handle);
        device_handle = INVALID_HANDLE_VALUE;
    }
}

bool fiRawDrivePrivate::open_device(const QString& device_uri)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "device:" << device_uri;

    // Sanity check.
    if (device_handle != INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is valid!?";
    }

    // Open the device.
    device_handle = open(device_uri.toLocal8Bit().constData(), O_RDONLY, 0);
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to open %1. Error %2: %3")
                     .arg(device_uri).arg(errno).arg(QString::fromUtf8(strerror(errno)));

        // Didn't work.
        return false;
    }

    // Good to go.
    return true;
}

/*!
 * @brief Query the opened device for it's drive geometry.
 */
bool fiRawDrivePrivate::query_device_info()
{
    int result;

    // Sanity check.
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is invalid!";
        return false;
    }

    // Clear our geometry structure.
    bytes_per_sector = 0;

    // Query the device for the information.
#ifdef Q_OS_MAC
    result = ioctl(device_handle, DKIOCGETPHYSICALBLOCKSIZE, &bytes_per_sector);
#else
    result = ioctl(device_handle, BLKPBSZGET, &bytes_per_sector);
#endif
    if (result == -1)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to query block size. Error %2: %3")
                     .arg(errno).arg(QString::fromUtf8(strerror(errno)));

        // Nothing here.
        return false;
    }

    // Looks like we're good.
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "bytes_per_sector:" << bytes_per_sector;
    return true;
}

/*!
 * @brief Read data from the drive.
 * @param offset            Where on the drive to read the data from.
 * @param data_ptr          Where to put the data.
 * @param bytes_to_read     How many bytes to read.
 *
 * @return True if the read is successful, false otherwise.
 */
bool fiRawDrivePrivate::read_data(uint64_t offset, uint8_t *data_ptr, uint64_t bytes_to_read)
{
    ssize_t read_count;

    // Sanity check.
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is invalid!";
        return false;
    }
    if (data_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "data_ptr is nullptr!";
        return false;
    }

    // Adjust the location on the drive.
    lseek(device_handle, offset, SEEK_SET);

    // Read the block(s).
    read_count = read(device_handle, data_ptr, bytes_to_read);
    if (read_count == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Error reading disk. Err:%1(%2)").arg(errno).arg(QString::fromUtf8(strerror(errno)));
    }

    // Return the result.
    return (read_count != 0);
}

/*!
 * @brief Read one or more sectors of data from the drive.
 *
 * @param[out]  data_ptr        Where to put the data.
 * @param[in]   start_sector    The starting sector number.
 * @param[in]   num_sectors     The number of sectors to read.
 *
 * @return True if the read is successful, false otherwise.
 */
bool fiRawDrivePrivate::read_sectors(uint64_t start_sector, uint8_t *data_ptr, uint32_t num_sectors)
{
    size_t  bytes_to_read;
    uint64_t    offset;
    ssize_t read_count;

    // Sanity check.
    if (device_handle == INVALID_HANDLE_VALUE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Device handle is invalid!";
        return false;
    }
    if (data_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "data_ptr is nullptr!";
        return false;
    }

    // How many bytes do we want to read?
    bytes_to_read = bytes_per_sector * num_sectors;

    // Convert our starting offset to a Windows happy value.
    offset = (start_sector * bytes_per_sector);

    // Adjust the location on the drive.
    lseek(device_handle, offset, SEEK_SET);

    // Read the block(s).
    read_count = read(device_handle, data_ptr, bytes_to_read);
    if (read_count == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Error reading disk. Err:%1(%2)").arg(errno).arg(QString::fromUtf8(strerror(errno)));
    }

    // Return the result.
    return (read_count != 0);
}


#endif // Q_OS_LINUX or Q_OS_MAC

