/*!
 * @file fi_rawdrive.h
 * @brief This is the header file for the fiRawDrive class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_RAWDRIVE_H
#define FI_RAWDRIVE_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QObject>
#include <QStringList>
#include <stdint.h>
#include <memory>

#include <fas_video.h>
#include <fas_filesystem.h>

namespace FI_RAW_DRIVE
{
class fiRawDrivePrivate;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QByteArray;

class fiPixmap;
struct FI_SIG;
struct CAPTURE_HD;
struct pm_metadata_type;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class fiRawDrive : public QObject
{
    Q_OBJECT

public:
    explicit fiRawDrive(QObject *parent = nullptr);
    ~fiRawDrive();

    QStringList drives();
    bool        is_fastec_drive(const QString& drive_uri);

    virtual void        close();
    virtual const QByteArray& getCT() const;
    virtual uint32_t    getFrameCount();
    virtual const QByteArray& getRawImage() const;
    virtual bool        isSeries3();
    virtual bool        open(const QString& drive_uri);
    virtual bool        readFrame(uint32_t frame_num, bool do_correction, const uint8_t *lookup_bcg,
                                  bool force8BitMono, FAS_VIDEO::VIDEO_CODEC *fvCodecPtr, QByteArray *dest_buf);
    virtual bool        readFrame(uint32_t frame_num, bool do_correction, fiPixmap *dest_pixmap);
    virtual uint32_t    readPerFrameMetadata(QByteArray& raw_buffer);
    virtual bool        readPmMetadata(struct pm_metadata_type *metadata_ptr);
    virtual bool        readRawImage(uint32_t frame_num, QByteArray& raw_image);
    virtual bool        recoverCT();

protected:
    void        applyFPN(QByteArray& image_buf);
    uint32_t    convertToAbsFrameNum(uint32_t frame_num, bool absolute);
    void        expandThePixels(QByteArray& buffer, uint32_t byte_count, bool needs_10bit_rotate = false, bool needs_12bit_swap = false);
    uint64_t    fmetaTime(uint32_t *metadata_buf);
    FAS_VIDEO::VIDEO_CODEC getImageHeader(bool forced8Bits, FAS_VIDEO::VIDEO_CODEC& forced8BitCodec, QByteArray& hdr_buffer);
    void        swapTheBytes(uint8_t *buffer, uint32_t byte_count);
    void        swapTheWords(uint32_t *buffer, uint32_t word_count);

    //! Number of frames before we yield the processor.
    const int YIELD_THRESHOLD = 1024;

protected:
    QByteArray  ct_buffer;                      //!< The correction table used for FPN correction.
    QByteArray  prnu_buffer;                    //!< The PRNU gain table. May be empty.
    QString     drive_str;                      //!< The drive we're working with.
    std::unique_ptr<struct pm_metadata_type> metadata_ptr;      //!< The capture metadata.

private:
    uint32_t    bytesToSectors(uint32_t byte_count);
    void        expand_10Bits(QByteArray& buffer, QByteArray& expanded_buf, bool needs_rotate = false);
    void        expand_12Bits(QByteArray& buffer, QByteArray& expanded_buf, bool needs_swap = false);
    uint32_t    pfmdPerSector();
    uint64_t    getPFMDOffset(FAS_FILESYSTEM::LR_CAPTURE_HDR *cap, uint32_t frame_num);
    void        getPFMDPosition(FAS_FILESYSTEM::LR_CAPTURE_HDR *cap, uint32_t frame_num, bool absolute, uint64_t *p_lba, uint32_t *p_ofs);
    uint32_t    *getDrivePFMD(FAS_FILESYSTEM::LR_CAPTURE_HDR *cap, uint32_t *sector_buf, uint32_t abs_frame);
    void        shrink16BitsTo8(QByteArray& buffer, const uint8_t *lookup_bcg);

private:
    std::unique_ptr<FI_RAW_DRIVE::fiRawDrivePrivate> priv_ptr;      //!< Machine dependent object.
    FAS_FILESYSTEM::PARTITION_INFO_VEC part_info;   //!< Fastec filesystem information.
    QByteArray  raw_image_buf;                      //!< The raw image data.
};

#endif // FI_RAWDRIVE_H
