/*!
 * @file fi_rawfile.cpp
 * @brief This file contains the implementation of the fiRawFile class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QApplication>
#include <QByteArray>
#include <QElapsedTimer>
#include <QFile>
#include <QFileInfo>
#include <QString>
#include <QTime>

#include "fi_rawfile.h"

#include "client_math.h"
#include "gige_msg.h"
#include "logger.h"

// Make this a "1" to turn on a debug log message for function entry points or set to "0" to remove from code.
#define RAWFILE_LOG_FUNC_ENTRY 0

namespace FI_RAW_FILE
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The extension for a spanned capture file.
const QString   CAP_SPAN_EXT = QStringLiteral("_1.cap");

//! The maximum length for a single capture file. Any capture longer than
//! this will need to span two files. This is 4G - 1 sector.
const uint64_t  MAX_CAP_LENGTH = 0xFFFFFE00;
}

using namespace FI_RAW_FILE;

/*!
 * @brief fiRawFile constructor.
 */
fiRawFile::fiRawFile(QObject *parent) :
    fiRawDrive(parent)
{
    root_file = nullptr;
    span_file = nullptr;
}

fiRawFile::~fiRawFile()
{
    // Make sure our files are closed.
    close();
}

/*!
 * @brief What is the name of our file on disk?
 *
 * @param raw_file_name[Out]: Will hold the result of the filename or empty string if no file is open.
 *
 * @returns True if successful, false if no file is open and there is no name to return.
 */
bool fiRawFile::getFileName(QString& raw_file_name)
{
    bool found_it = false;

    // Empty the output parameter.
    raw_file_name.clear();

    // Check the root_file to see if it is valid, and open.
    if (root_file)
    {
        if (root_file->isOpen())
        {
            raw_file_name = root_file->fileName();
            found_it = true;
        }
    }
    return found_it;
}

/*!
 * @brief Close a previously opened file.
 */
void fiRawFile::close()
{
#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif
    if (root_file != nullptr)
    {
        root_file->close();
        delete root_file;
        root_file = nullptr;
    }
    if (span_file != nullptr)
    {
        span_file->close();
        delete span_file;
        span_file = nullptr;
    }
}

/*!
 * @brief Return an absolute frame number.
 *
 *    If "relative", meaning absolute=false, the frame number passed in is relative to the
 *    oldest frame in the partition, which can change since it's a circular buffer.
 *    If it is "absolute", meaning absolute=true, it is relative to the physical start of
 *    the partition; in that case, we just sanity check it and return what was passed.
 *
 * @param frame_num The frame number we're interested in.
 * @param absolute  TRUE if we passed the "absolute" frame_num,
 *                  or FALSE if we passed a "relative" frame_num.
 *
 * @return The absolute frame number if everything is good, otherwise (uint32_t)(-1).
 */
uint32_t fiRawFile::convertToAbsFrameNum(uint32_t frame_num, bool absolute)
{
    uint32_t abs_frame;
#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    // Did we receive an "absolute" frame number?
    if (absolute)
    {
        // Just sanity check the frame number.
        if (frame_num < getFrameCount())
            abs_frame = frame_num;
        else
        {
            // We should never be asked to do this.
            QLOG_WARN() << __PRETTY_FUNCTION__ << "absolute=T but frame_num =" << frame_num << "total_frames =" << getFrameCount();
            abs_frame = frame_num % getFrameCount();
        }
    }
    else
    {
        uint32_t skip_count = get_skip_count();

        // If we have a skip count, then we are looking at a partial CAP save.
        if (skip_count != 0)
        {
            // A partial CAP save has frames arranged linearly from the starting
            // frame to the ending frame. There is no wrapping.
            abs_frame = frame_num;
        }
        else
        {
            // Check to see if the frame being requested goes beyond (i.e. "wraps")
            // the last frame.
            if ((frame_num + getFirstFrame()) < getFrameCount())
            {
                // Nope - we're good. Just add the frame number to our trigger frame
                // position.
                abs_frame = frame_num + getFirstFrame();
            }
            else
            {
                // Yes it does. We need to subtract the total number of frames so
                // that the frame number will wrap around to the beginning again.
                abs_frame = (frame_num + getFirstFrame()) - getFrameCount();
            }
        }
    }

    return abs_frame;
}

/*!
 * @brief Return the number of the first frame in the capture.
 */
uint32_t fiRawFile::getFirstFrame()
{
#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif
    // Sanity check.
    if (metadata_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No metadata_ptr!";
        return 0;
    }

    // Return the number of frames in the capture.
    return metadata_ptr->first_frame;
}

/*!
 * @brief Return the number of frames in the capture.
 */
uint32_t fiRawFile::getFrameCount()
{
#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif
    // Sanity check.
    if (metadata_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No metadata_ptr!";
        return 0;
    }

    // Return the number of frames in the capture.
    return metadata_ptr->frame_count;
}

/*!
 * @brief Return the number of frames to skip when figuring out where a frame is located.
 */
uint32_t fiRawFile::get_skip_count()
{
#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    // There is no skip count prior to 3.1.8, so no skip count
    // for series3 cameras.
    if (isSeries3())
        return 0;

    // Sanity check.
    if (metadata_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No metadata_ptr!";
        return 0;
    }

    // Return the number of frames in the capture.
    return metadata_ptr->skip_count;
}


/*!
 * @brief Attempt to open the file specified by cap_uri. Sanity check to make
 *        sure the file is a Fastec CAP file.
 *
 * @param cap_uri     The URI of the file we're interested in.
 * @return True on success, false otherwise.
 */
bool fiRawFile::open(const QString& cap_uri)
{
    bool        rc;
    pm_metadata dummy_pm;

#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    // Assume this is not a Fastec CAP file.
    rc = false;

    // Make sure any previously opened files are closed.
    close();

    // Allocate our root file object.
    root_file = new QFile(cap_uri);
    if (root_file == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate QFile for:" << cap_uri;
        return false;
    }

    // Open the root file.
    if (!root_file->open(QIODevice::ReadOnly | QIODevice::Unbuffered))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open:" << cap_uri;
        goto rfo_bail;
    }

    // Initialize our metadata structures.
    if (!readPmMetadata(&dummy_pm))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read the PM metadata.";
        goto rfo_bail;
    }

    // If this is a series3 file, check to see if we have a "span" file.
    if (isSeries3())
    {
        uint64_t    total_cap_len;

        // Should there be a "span" file?
        total_cap_len = (uint64_t)metadata_ptr->frame_count * ClientMath::qword_to_bytes((uint64_t)metadata_ptr->frame_size);
        if (total_cap_len > MAX_CAP_LENGTH)
        {
            QString     span_uri;

            // Check to see if we have a "span" file.
            span_uri = cap_uri.left(cap_uri.lastIndexOf(QLatin1Char('.')));
            span_uri.append(CAP_SPAN_EXT);

            // Allocate the span file object.
            span_file = new QFile(span_uri);
            if (root_file == nullptr)
            {
                QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate QFile for:" << span_uri;
                goto rfo_bail;
            }

            // Try to open the span file.
            if (!span_file->open(QIODevice::ReadOnly | QIODevice::Unbuffered))
            {
                // No span file.
                QLOG_WARN() << __PRETTY_FUNCTION__ << "No span file. Maybe post-series3?";
                delete span_file;
                span_file = nullptr;
            }
        }
    }

    // All is well.
    rc = true;

rfo_bail:
    // If things didn't go well, make sure we close the file(s).
    if (!rc)
        close();

    // Return our status.
    return rc;
}

/*!
 * @brief Extract the raw per-frame metadata from the CAP file for a particular frame.
 *
 * The assumption is that the frame number is the ABSOLUTE frame number, already adjusted
 * for any wrapping that occurs. Therefore, the frame number used in this method must be
 * correct for the file (i.e. frame number 0 is the first frame in the file, not necessarily
 * the recording!)
 *
 * @param frame_num     The frame we're interested in.
 * @param event_data    Buffer to put the event data.
 * @return True on success, false otherwise.
 */
bool fiRawFile::readFrameMetadata(uint32_t frame_num, QByteArray& event_data)
{
    uint64_t    frame_len;
    uint64_t    split_len(0);
    uint64_t    frame_offset;
    bool        use_span(false);

#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    // Do we have the capture metadata loaded?
    if (metadata_ptr == nullptr)
    {
        pm_metadata dummy_structure;        // Need a place to hold the initial data.

        // Initialize our capture metadata structure (the CLASS structure, we
        // don't much care about the passed in parameter.)
        if (!readPmMetadata(&dummy_structure))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the pm_metadata!";
            return false;
        }
    }

    // Are we a series3 camera?
    if (isSeries3())
    {
        // What is the size of the frame?
        frame_len = ClientMath::qword_to_bytes((uint64_t)metadata_ptr->frame_size);

        // Figure out where we need to go to read the image. Go just past the frame
        // and then back up to where the metadata is.
        frame_offset = (frame_len * (uint64_t)(frame_num + 1)) + ClientMath::qword_to_bytes((uint64_t)metadata_ptr->sov_off);
        frame_offset -= (uint64_t)GIGE_FRAME_METADATA_LEN;

        // The amount of data we want to read is the size of the metadata array.
        frame_len = (uint64_t)GIGE_FRAME_METADATA_LEN;

        // Do we need to go into the span file?
        if (frame_offset > MAX_CAP_LENGTH)
        {
            // We need to read from the span file.
            use_span = true;

            // Adjust the offset appropriately.
            frame_offset -= MAX_CAP_LENGTH;
        }

        // Check for a split buffer, one that is split between files.
        else if ((frame_offset + frame_len) > MAX_CAP_LENGTH)
        {
            // Calculate the number of bytes in the span file.
            split_len = (frame_offset + frame_len) - MAX_CAP_LENGTH;

            // Adjust the number of bytes to read from the root file.
            frame_len -= split_len;
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "split buffer 1st:" << frame_len << "2nd:" << split_len;
        }
    }
    else
    {
        // What is the size of the frame?
        frame_len = (uint64_t)metadata_ptr->frame_size;

        // Figure out where we need to go to read the image. Go just past the frame
        // and then back up to where the metadata is.
        frame_offset = (frame_len * (uint64_t)(frame_num + 1)) + (uint64_t)metadata_ptr->sov_off;
        frame_offset -= (uint64_t)(2 * GIGE_FRAME_METADATA_LEN);

        // The amount of data we want to read is the size of the metadata array.
        frame_len = (uint64_t)(2 * GIGE_FRAME_METADATA_LEN);
    }

    // Adjust the buffer size appropriately.
    event_data.resize((int)frame_len);

    // Are we reading from the span file?
    if (use_span)
    {
        // Sanity check.
        if ((span_file == nullptr) || !span_file->isOpen())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need the span file but we don't have one!";
            return false;
        }

        // Seek to the right place in the span file.
        span_file->seek((int64_t)frame_offset);

        // Read the image data.
        span_file->read(event_data.data(), (int64_t)frame_len);
    }
    else
    {
        // Sanity check.
        if ((root_file == nullptr) || !root_file->isOpen())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need the root file but we don't have one!";
            return false;
        }

        // Seek to the right place in the span file.
        root_file->seek((int64_t)frame_offset);

        // Read the image data.
        root_file->read(event_data.data(), (int64_t)frame_len);

        // Do we have a split file?
        if (split_len > 0)
        {
            // Sanity check.
            if ((span_file == nullptr) || !span_file->isOpen())
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need the span file but we don't have one!";
                return false;
            }

            // Seek to the right place in the span file.
            span_file->seek(0);

            // Read the image data and append it to our image buffer.
            event_data.append(span_file->read((int64_t)split_len));
        }
    }

    // All done.
    return true;
}

/*!
 * @brief Read the specified frame from the Fastec formatted drive and load it into
 *        the dest_buf object.
 *
 * @param [in]  frame_num       The absolute frame number to be read.
 * @param [in]  do_correction   True if we're supposed to do FPN correction, false otherwise.
 * @param [in]  lookup_bcg      A Brightness/Contrast/Gamma modification for frame.
 * @param [in]  force8BitMono   True if monochrome sensor and we must force 16 bits into 8.
 * @param [out] fvCodecPtr      Pointer to where we're to put the codec type for the data.
 * @param [out] dest_buf        Pointer to the QByteArray object for the data.
 *
 * @return True on success, false otherwise.
 */
bool fiRawFile::readFrame(uint32_t frame_num, bool do_correction, const uint8_t *lookup_bcg,
                          bool force8BitMono, FAS_VIDEO::VIDEO_CODEC *fvCodecPtr, QByteArray *dest_buf)

{
#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    return fiRawDrive::readFrame(convertToAbsFrameNum(frame_num, false), do_correction, lookup_bcg, force8BitMono, fvCodecPtr, dest_buf);
}

/*!
 * @brief Read the specified frame from the Fastec formatted drive and load it into
 *        the dest_pixmap object.
 *
 * @param [in]  frame_num       The absolute frame number to be read.
 * @param [in]  do_correction   True if we're supposed to do FPN correction, false otherwise.
 * @param [out] dest_pixmap     Where to put the pixmap data.
 *
 * @return True on success, false otherwise.
 */
bool fiRawFile::readFrame(uint32_t frame_num, bool do_correction, fiPixmap *dest_pixmap)
{
#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    return fiRawDrive::readFrame(convertToAbsFrameNum(frame_num, false), do_correction, dest_pixmap);
}

/*!
 * @brief Read the capture metadata from the previously opened Fastec formatted drive.
 * @param metadata_ptr  Pointer for the storage location of the metadata.
 * @return True on success, false otherwise.
 */
bool fiRawFile::readPmMetadata(struct pm_metadata_type *metadata_ptr)
{
    bool        rc = false;
    qint64      read_len;

#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    // Sanity check.
    if (metadata_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null metadata_ptr!";
        return false;
    }

    // Do we already have the data loaded?
    if (this->metadata_ptr != nullptr)
    {
        // Copy the data over.
        memcpy(metadata_ptr, this->metadata_ptr.get(), sizeof(pm_metadata));

        // All done.
        return true;
    }

    // Make sure we have a root file and it's open.
    if (root_file == nullptr)
        return false;
    if (!root_file->isOpen())
        return false;

    // The PM metadata structure is at the front of the file.
    root_file->seek(0);

    // Read the data into our structure.
    read_len = root_file->read((char *)metadata_ptr, sizeof(pm_metadata));
    if (read_len <= 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read the metadata, read_len=" << read_len;
        goto pm_bail;
    }

    // Allocate our copy. This will also delete any previous copy.
    this->metadata_ptr.reset(new pm_metadata);
    if (this->metadata_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a pm_metadata object!";
        goto pm_bail;
    }

    // Copy the data over.
    memcpy(this->metadata_ptr.get(), metadata_ptr, sizeof(pm_metadata));

    // All is well...
    rc = true;

pm_bail:
    // Return our status.
    return rc;
}

/*!
 * @brief Load the raw image from the Fastec CAP file into the raw_image buffer.
 *
 * The assumption is that the frame number is the ABSOLUTE frame number, already adjusted
 * for any wrapping that occurs. Therefore, the frame number used in this method must be
 * correct for the file (i.e. frame number 0 is the first frame in the file, not necessarily
 * the recording!)
 *
 * @param [in]  frame_num       The absolute frame number of the image we want.
 * @param [out] raw_image       The destination buffer for the raw image.
 * @return True on success, false otherwise.
 */
bool fiRawFile::readRawImage(uint32_t frame_num, QByteArray& raw_image)
{
    uint64_t    frame_len;
    uint64_t    split_len(0);
    uint64_t    frame_offset;
    bool        use_span(false);

#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Do we have the capture metadata loaded?
    if (metadata_ptr == nullptr)
    {
        pm_metadata dummy_structure;        // Need a place to hold the initial data.

        // Initialize our capture metadata structure (the CLASS structure, we
        // don't much care about the passed in parameter.)
        if (!readPmMetadata(&dummy_structure))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the pm_metadata!";
            return false;
        }
    }

    // Are we dealing with a series3 file?
    if (isSeries3())
    {
        // What is the size of the frame?
        frame_len = ClientMath::qword_to_bytes((uint64_t)metadata_ptr->frame_size);

        // Figure out where we need to go to read the image.
        frame_offset = (frame_len * (uint64_t)frame_num) + ClientMath::qword_to_bytes((uint64_t)metadata_ptr->sov_off);

        // Do we need to go into the span file?
        if (frame_offset > MAX_CAP_LENGTH)
        {
            // We need to read from the span file.
            use_span = true;

            // Adjust the offset appropriately.
            frame_offset -= MAX_CAP_LENGTH;
        }

        // Check for a split frame, one that is split between files.
        else if ((frame_offset + frame_len) > MAX_CAP_LENGTH)
        {
            // Calculate the number of bytes in the span file.
            split_len = (frame_offset + frame_len) - MAX_CAP_LENGTH;

            // Adjust the number of bytes to read from the root file.
            frame_len -= split_len;
        }
    }
    else
    {
        // What is the size of the frame?
        frame_len = (uint64_t)metadata_ptr->frame_size;

        // Figure out where we need to go to read the image.
        frame_offset = (frame_len * (uint64_t)(frame_num - get_skip_count())) + (uint64_t)metadata_ptr->sov_off;

        // Now we need to adjust the frame length so it doesn't include the per-frame metadata.
        frame_len -= (uint64_t)(2 * GIGE_FRAME_METADATA_LEN);
    }

    // Adjust the buffer size appropriately.
    raw_image.resize((int)frame_len);

#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_len:" << frame_len << "frame_offset:" << frame_offset;
#endif

    // Are we reading from the span file?
    if (use_span)
    {
        // Sanity check.
        if ((span_file == nullptr) || !span_file->isOpen())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need the span file but we don't have one!";
            return false;
        }

        // Seek to the right place in the span file.
        span_file->seek((int64_t)frame_offset);

        // Read the image data.
        span_file->read(raw_image.data(), (int64_t)frame_len);
    }
    else
    {
        // Sanity check.
        if ((root_file == nullptr) || !root_file->isOpen())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need the root file but we don't have one!";
            return false;
        }

        // QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame:" << frame_num << "offset: " << Qt::showbase << Qt::hex << frame_offset;

        // Seek to the right place in the main capture file.
        root_file->seek((int64_t)frame_offset);

        // Read the image data.
        root_file->read(raw_image.data(), (int64_t)frame_len);

        // Do we have a split file?
        if (split_len > 0)
        {
            // Sanity check.
            if ((span_file == nullptr) || !span_file->isOpen())
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need the span file but we don't have one!";
                return false;
            }

            // Seek to the right place in the span file.
            span_file->seek(0);

            // Read the image data and append it to our image buffer.
            raw_image.append(span_file->read((int64_t)split_len));
        }
    }

// #define DO_SAVE_FILE 1
#ifdef DO_SAVE_FILE
    {
        QString file_name = "./test_image_" + QString::number(frame_num) + ".raw";
        QFile new_doc(file_name);
        if (new_doc.open(QIODevice::WriteOnly))
            new_doc.write(raw_image);
        new_doc.close();
    }
#endif

    // If the image is greater than 8-bit, then we need to expand the pixels to be 16-bit.
    if (IS_8_BIT_MODE((char)metadata_ptr->bit_sel))
    {
        // Swap the bytes for series3 cameras. Don't need this step for the newer cameras.
        if (isSeries3())
            swapTheBytes((uint8_t *)raw_image.data(), (uint32_t)ClientMath::qword_to_bytes((uint64_t)metadata_ptr->frame_size));
    }
    else
    {
        // This will expand and swap, if necessary.
        if (isSeries3())
        {
            // Note:  For CAP files on series3 cameras there is always no rotation for 10-bit
            // but there is always swapping for 12-bit.
            expandThePixels(raw_image, (uint32_t)ClientMath::qword_to_bytes((uint64_t)metadata_ptr->frame_size), false, true);
        }
        else
        {
            // Note:  For CAP files on h-series cameras there is a rotation of word data for
            // captures greater than 8-bits.  The "true" below indicates that this rotation
            // needs to occur.  And by "rotation" it simply means a reording of byte order.
            // This is handled by calling qToBigEndian((uint32_t)packed_word) prior to expansion.
            // JC - 10/20/2021.
            expandThePixels(raw_image, (uint32_t)frame_len, true, false);
        }
    }

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif

    return true;
}

/*!
 * @brief Read the per-frame metadata from the previously opened Fastec CAP file.
 * @param buffer    QByteArray to store the data.
 * @return The number of bytes in the buffer, 0 on error.
 */
uint32_t fiRawFile::readPerFrameMetadata(QByteArray& raw_buffer)
{
    uint32_t    frame_num;
    uint32_t    return_count = 0;
    uint32_t    pfmd_len = 0;

#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    // Reset the target buffer.
    raw_buffer.clear();

    // How many bytes are we talking about?
    if (isSeries3())
        pfmd_len = GIGE_FRAME_METADATA_LEN;
    else
        pfmd_len = (GIGE_FRAME_METADATA_LEN * 2);

    // Loop through all the frames and gather the event data.
    for (frame_num = 0; frame_num < metadata_ptr->frame_count; ++frame_num)
    {
        QByteArray  event_data;

        // Yield some time every so often.
        if ((frame_num % YIELD_THRESHOLD) == 0)
            QApplication::processEvents();

        // Get the event data for the frame.
        if (!readFrameMetadata(convertToAbsFrameNum(frame_num, false), event_data))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Error reading frame metadata, frame_num:" << frame_num;
            return 0;
        }

        // Append it to our buffer.
        raw_buffer.append(event_data);

        // Update our byte count.
        return_count += pfmd_len;
    }

    // Make sure we have something.
    if (return_count == 0)
    {
        // Let the user know we didn't find anything.
        QLOG_INFO() << __PRETTY_FUNCTION__ << "No images found!";
    }
    else
    {
        // Swap our bytes.
        swapTheBytes((uint8_t *)raw_buffer.data(), return_count);
    }

    // Return the number of bytes stored.
    return return_count;
}

/*!
 * @brief Load correction table for a capture.
 * @return True on success, false otherwise.
 */
bool fiRawFile::recoverCT()
{
    uint64_t offset;
    uint64_t length;

#if RAWFILE_LOG_FUNC_ENTRY
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Entry";
#endif

    // Sanity check.
    if ((root_file == nullptr) || !root_file->isOpen())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need the span file but we don't have one!";
        return false;
    }

    // Do we have the capture metadata loaded?
    if (metadata_ptr == nullptr)
    {
        pm_metadata dummy_structure;        // Need a place to hold the initial data.

        // Initialize our capture metadata structure (the CLASS structure, we
        // don't much care about the passed in parameter.)
        if (!readPmMetadata(&dummy_structure))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the pm_metadata!";
            return false;
        }
    }

    offset = metadata_ptr->corr_off;
    length = metadata_ptr->corr_len;
    if (isSeries3())
    {
        offset = ClientMath::qword_to_bytes(offset);
        length = ClientMath::qword_to_bytes(length);
    }

    // Seek to the correction table location.
    root_file->seek((int64_t)offset);

    // Read the correction table.
    ct_buffer = root_file->read((int64_t)length);

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ct_offset:" << offset << "ct_len:" << length;

    // Is there a PRNU gain table?
    if (!isSeries3())
    {
        uint64_t aligned_len;

        // Make sure our length is on a hexword boundary.
        aligned_len = ((length + 63) >> 6) << 6;

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "metadata_ptr->sov_off:" << metadata_ptr->sov_off << "aligned_len:" << aligned_len;


        if ((offset + aligned_len) < metadata_ptr->sov_off)
        {
            // Seek to the correction table location.
            root_file->seek((int64_t)(offset + aligned_len));

            // Read the correction table.
            prnu_buffer = root_file->read((int64_t)length);

            // Swap our bytes.
            swapTheBytes((uint8_t *)prnu_buffer.data(), (uint32_t)length);
        }
    }

    // Swap our bytes.
    swapTheBytes((uint8_t *)ct_buffer.data(), (uint32_t)length);

    // All done.
    return true;
}
