/*!
 * @file fi_rectitem.cpp
 * @brief This file contains the implementation of the fiRectItem class.
 *
 *  This box can be re-sized and it can be moved. For moving, we capture
 *  the mouse move events and move the box location accordingly.
 *
 *  To resize the box, we place small indicator boxes on the four corners that give the user
 *  a visual cue to grab with the mouse. The user then drags the corner to stretch
 *  or shrink the box.  The corner grabbers are implemented with the CornerGrabber class.
 *  The CornerGrabber class captures the mouse when the mouse is over the corner's area,
 *  but the fiRectItem object (which owns the corners) captures and processes the mouse
 *  events on behalf of the CornerGrabbers (because the owner wants to be
 *  resized, not the CornerGrabbers themselves). This is accomplished by installed a scene event filter
 *  on the CornerGrabber objects:
 *       corners[0]->installSceneEventFilter(this);
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QApplication>

#include "cornergrabber.h"
#include "logger.h"

#include "fi_rectitem.h"
#include "mainwindow.h"

namespace FI_RECTITEM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define a space around our overlay to allow for the hover events to fire.
const int   CORNER_BUFFER               = 20;

}   // End of namespace FI_OVERLAY.

const QEvent::Type fiRectItem::CHANGE_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_RECTITME_CHANGED_EVENT);

using namespace FI_RECTITEM;

/*!
 * @brief Basic constructor.
 */
fiRectItem::fiRectItem(QGraphicsItem *parent) :
    QGraphicsRectItem(parent),
    cb_object(nullptr),
    moveable(true)
{
    initialize();
}

/*!
 * @brief Constructor with a rectangle.
 */
fiRectItem::fiRectItem(const QRectF& rect, QGraphicsItem *parent) :
    QGraphicsRectItem(parent),
    cb_object(nullptr)
{
    setRect(rect);
    initialize();
}

/*!
 * @brief Constructor with a rectangle as separate parameters.
 */
fiRectItem::fiRectItem(qreal x, qreal y, qreal w, qreal h, QGraphicsItem *parent) :
    QGraphicsRectItem(parent),
    cb_object(nullptr)
{
    setRect(QRectF(x, y, w, h));
    initialize();
}

/*!
 * @brief Basic destructor. Nothing to tear down.
 */
fiRectItem::~fiRectItem()
{
}

/*!
 * @brief Common initialization routine.
 */
void fiRectItem::initialize()
{
    // Item is moveable by default.
    moveable = true;

    // Make sure we get hover events, so we can turn the corners on/off.
    setAcceptHoverEvents(true);

    // Initialize our corners.
    corners[0] = nullptr;
    corners[1] = nullptr;
    corners[2] = nullptr;
    corners[3] = nullptr;
    clip_rect = QRectF();

    // Setup our default pens and brushes. Enabled pen and brush are both
    // cyan, with the brush being transparent. No disabled brush, and the
    // pen is gray to be unobtrussive.
    enabled_pen = QPen(Qt::cyan);
    enabled_pen.setWidth(3);
    enabled_brush = QBrush(QColor(0x00, 0xff, 0xff, 0x40), Qt::SolidPattern);
    disabled_pen = QPen(Qt::gray);
    disabled_pen.setWidth(1);
    disabled_brush = QBrush(Qt::NoBrush);
    max_area = 0;
}

/*!
 * @brief Required method when creating a QGraphicsItem class.
 *
 * We need to override the base class implementation to account for the
 * CORNER_BUFFER margins.
 *
 * @param painter   Pointer to our paint object.
 * @param option    [UNUSED] Style options.
 * @param widget    [UNUSED] Pointer to the parent widget.
 */
void fiRectItem::paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget)
{
    Q_UNUSED(option);
    Q_UNUSED(widget);

    // Draw the rectangle differently if we are "enabled".
    if (isEnabled())
    {
        // Setup our pen & brush.
        painter->setPen(enabled_pen);
        painter->setBrush(enabled_brush);
    }
    else
    {
        // Setup our pen & brush.
        painter->setPen(disabled_pen);
        painter->setBrush(disabled_brush);
    }

    // Draw a "normal" rectangle.
    painter->drawRect(adjrect());
}

QRectF fiRectItem::adjrect()
{
    QRectF real_rect;
    real_rect = rect().marginsRemoved(QMarginsF(CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER));
    return real_rect;
}

/*!
 * @brief Setup the locations for the corner "handles".
 */
void fiRectItem::set_corner_positions()
{
    QRectF draw_rect;
    draw_rect = rect().marginsRemoved(QMarginsF(CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER));

    qreal c_wd = corners[0]->boundingRect().width();
    qreal c_ht = corners[0]->boundingRect().height();

    corners[0]->setPos(draw_rect.left() - c_wd, draw_rect.top() - c_ht);
    corners[1]->setPos(draw_rect.right(), draw_rect.top() - c_ht);
    corners[2]->setPos(draw_rect.right(), draw_rect.bottom());
    corners[3]->setPos(draw_rect.left() - c_wd,  draw_rect.bottom());
}

void fiRectItem::set_edge(QPointF edge)
{
    clip_rect.setLeft((rect().width() - edge.x()) / 2);
    clip_rect.setTop((rect().height() - edge.y()) / 2);
    clip_rect.setRight((edge.x() - rect().width()) / 2);
    clip_rect.setBottom((edge.y() - rect().height()) / 2);
}

/*!
 * @brief Set the size of the rectangle, but add our margin around it first.
 * @param rect  The "drawing" rectangle we are supposed to use (without margins).
 */
void fiRectItem::setRect(const QRectF& rect)
{
    // Adjust for our margin.
    QRectF margin_rect = rect.marginsAdded(QMarginsF(CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER));

    // Pass the adjust rectangle onwards.
    QGraphicsRectItem::setRect(margin_rect);
}

void fiRectItem::setMaxArea(int area)
{
    max_area = area;
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Create the corner "grabbers".
 */
void fiRectItem::hoverEnterEvent(QGraphicsSceneHoverEvent *)
{
    // Sanity check.
    if (corners[0] != nullptr)
    {
        // This is a programming error, so just yell and carry on.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Why are there grabbers already?!";
    }

    // If this is not moveable we don't do any of this.
    if (!moveable)
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "IGNORING not moveable!";
        return;
    }

    corners[0] = new CornerGrabber(this, 0);
    corners[1] = new CornerGrabber(this, 1);
    corners[2] = new CornerGrabber(this, 2);
    corners[3] = new CornerGrabber(this, 3);

    corners[0]->setWidth(CORNER_BUFFER >> 1);
    corners[1]->setWidth(CORNER_BUFFER >> 1);
    corners[2]->setWidth(CORNER_BUFFER >> 1);
    corners[3]->setWidth(CORNER_BUFFER >> 1);

    corners[0]->setHeight(CORNER_BUFFER >> 1);
    corners[1]->setHeight(CORNER_BUFFER >> 1);
    corners[2]->setHeight(CORNER_BUFFER >> 1);
    corners[3]->setHeight(CORNER_BUFFER >> 1);

    corners[0]->installSceneEventFilter(this);
    corners[1]->installSceneEventFilter(this);
    corners[2]->installSceneEventFilter(this);
    corners[3]->installSceneEventFilter(this);

    // Locate the "handles" on the corners of our rectangle.
    set_corner_positions();

    // When the cursor enters our "space" make sure we have an arrow pointer.
    setCursor(Qt::ArrowCursor);
}

/*!
 * @brief Remove the corner "grabbers".
 */
void fiRectItem::hoverLeaveEvent(QGraphicsSceneHoverEvent *)
{
    // Sanity check.
    if (corners[0] == nullptr)
        return;

    corners[0]->setParentItem(nullptr);
    corners[1]->setParentItem(nullptr);
    corners[2]->setParentItem(nullptr);
    corners[3]->setParentItem(nullptr);

    delete corners[0];
    corners[0] = nullptr;
    delete corners[1];
    corners[1] = nullptr;
    delete corners[2];
    corners[2] = nullptr;
    delete corners[3];
    corners[3] = nullptr;
}

/*!
 * @brief Hook into the moving event so we can let the user move
 *        the rectangle across the scene.
 * @note  The corner "handles" have their own logic, that's why you
 *        won't see any of that here.
 *
 * @param event Our event information.
 */
void fiRectItem::mouseMoveEvent(QGraphicsSceneMouseEvent *event)
{
    // Pass it on up.
    QGraphicsItem::mouseMoveEvent(event);

    // We only react to left mouse button events. The move event
    // reports this info in the "buttonS" call, not "button".
    if (event->buttons() != Qt::LeftButton)
        return;

    // If we do not have any "modifiers", then we are simply moving
    // the reticle.
    if (event->modifiers() == Qt::NoModifier)
    {
        // Maintain our boundaries.
        if (!clip_rect.isEmpty())
        {
            QLOG_DEBUG() << "clip rect is not empty";

            if (x() < clip_rect.left())
                setPos(clip_rect.left(), y());
            else if (x() > clip_rect.right())
                setPos(clip_rect.right(), y());

            if (y() < clip_rect.top())
                setPos(x(), clip_rect.top());
            else if (y() > clip_rect.bottom())
                setPos(x(), clip_rect.bottom());
        }
        else
        {
            QRectF draw_rect;
            QRectF bounds = scene()->sceneRect();
            draw_rect = sceneBoundingRect().marginsRemoved(QMarginsF(CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER));

            if (draw_rect.left() < bounds.left())
                setX(bounds.left());
            else if (draw_rect.right() > bounds.right())
                setX(bounds.right() - draw_rect.width());

            if (draw_rect.top() < bounds.top())
                setY(bounds.top());
            else if (draw_rect.bottom() > bounds.bottom())
                setY(bounds.bottom() - draw_rect.height());
        }
    }
}

void fiRectItem::mouseReleaseEvent(QGraphicsSceneMouseEvent *event)
{
    // Pass it on up.
    QGraphicsItem::mouseReleaseEvent(event);

    // Let folks know things have changed.
    if (cb_object != nullptr)
        QApplication::postEvent(cb_object, new QEvent(CHANGE_EVENT));
}

/*!
 * @brief This scene event filter has been registered with all four corner grabber items.
 *
 * When called, a pointer to the sending item is provided along with a generic
 * event.  A dynamic_cast is used to determine if the event type is one of the events
 * we are interrested in.
 */
bool fiRectItem::sceneEventFilter(QGraphicsItem *watched, QEvent *event)
{
    CornerGrabber *corner = dynamic_cast<CornerGrabber *>(watched);
    if (corner == nullptr)
        return false; // not expected to get here

    QGraphicsSceneMouseEvent *mevent = dynamic_cast<QGraphicsSceneMouseEvent *>(event);
    if (mevent == nullptr)
    {
        // This is not one of the mouse events we are interested in.
        return false;
    }

    // Process based on the event type.
    switch (event->type())
    {
    case QEvent::GraphicsSceneMousePress:
        // If the mouse went down, record the x,y coords of the press, record it inside the corner object.
        corner->setMouseState(CornerGrabber::kMouseDown);
        corner->mouseDownX = mevent->pos().x();
        corner->mouseDownY = mevent->pos().y();
        break;
    case QEvent::GraphicsSceneMouseRelease:
        corner->setMouseState(CornerGrabber::kMouseReleased);

        // Let folks know things have changed.
        if (cb_object != nullptr)
            QApplication::postEvent(cb_object, new QEvent(CHANGE_EVENT));
        break;
    case QEvent::GraphicsSceneMouseMove:
        corner->setMouseState(CornerGrabber::kMouseMoving);
        break;
    default:
        // We don't care about the rest of the events.
        return false;
    }

    // We only REALLY care about the mouse moving around.
    if (corner->getMouseState() == CornerGrabber::kMouseMoving)
    {
        QRectF draw_rect;
        QRectF bounds = scene()->sceneRect();

        // Make sure the rectangle stays in bounds.
        if (!bounds.contains(mevent->scenePos()))
            return true;

        // Our drawing rectangle is the adjusted drawing width/height.
        draw_rect = rect().marginsRemoved(QMarginsF(CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER, CORNER_BUFFER));

        // Get our x and y values.
        qreal x = mevent->pos().x();
        qreal y = mevent->pos().y();

        // Depending on which corner has been grabbed, we want to move the position
        // of the item as it grows/shrinks accordingly. so we need to either add
        // or subtract the offsets based on which corner this is.
        int XaxisSign = 0;
        int YaxisSign = 0;

        switch (corner->getCorner())
        {
        case 0:
            XaxisSign = 1;
            YaxisSign = 1;
            break;
        case 1:
            XaxisSign = -1;
            YaxisSign = 1;
            break;
        case 2:
            XaxisSign = -1;
            YaxisSign = -1;
            break;
        case 3:
            XaxisSign = 1;
            YaxisSign = -1;
            break;
        default:
            // Unexpected.
            break;
        }

        // If the mouse is being dragged, calculate a new size and also re-position
        // the box to give the appearance of dragging the corner out/in to resize the box
        int xMoved = (int)(corner->mouseDownX - x);
        int yMoved = (int)(corner->mouseDownY - y);

        int newWidth = (int)(draw_rect.width() + (XaxisSign * xMoved));
        if (newWidth < (2 * CORNER_BUFFER))
            newWidth  = (2 * CORNER_BUFFER);

        // Did we exceed the size allowed?
        if ((max_area > 0) && ((newWidth * draw_rect.height()) > max_area))
            newWidth = max_area / draw_rect.height();

        int newHeight = (int)(draw_rect.height() + (YaxisSign * yMoved));
        if (newHeight < (2 * CORNER_BUFFER))
            newHeight = (2 * CORNER_BUFFER);

        // Did we exceed the size allowed?
        if ((max_area > 0) && ((newWidth * newHeight) > max_area))
            newHeight = max_area / newWidth;

        int deltaWidth = newWidth - (int)draw_rect.width();
        int deltaHeight = newHeight - (int)draw_rect.height();
        draw_rect.adjust(0, 0, deltaWidth, deltaHeight);
        setRect(draw_rect);

        deltaWidth *= (-1);
        deltaHeight *= (-1);

        if (corner->getCorner() == 0)
        {
            int newXpos = (int)this->pos().x() + deltaWidth;
            int newYpos = (int)this->pos().y() + deltaHeight;
            setPos(newXpos, newYpos);
        }
        else if (corner->getCorner() == 1)
        {
            int newYpos = (int)this->pos().y() + deltaHeight;
            setPos(this->pos().x(), newYpos);
        }
        else if (corner->getCorner() == 3)
        {
            int newXpos = (int)this->pos().x() + deltaWidth;
            setPos(newXpos, this->pos().y());
        }

        set_corner_positions();
        update();
    }

    return true;// true => do not send event to watched - we are finished with this event
}
