/*!
 * @file fi_rectitem.h
 * @brief This is the header file for the fiRectItem class.
 *
 * fiRectItem subclasses the QGraphicsRectItem so that we can force our
 * region (e.g. a reticle) to stay within the bounds of the scene/image.
 *
 * Copyright 2017 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_RECTITEM_H
#define FI_RECTITEM_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QGraphicsObject>
#include <QGraphicsLayoutItem>
#include <QGraphicsRectItem>
#include <QGraphicsTextItem>
#include <QGraphicsScene>
#include <QGraphicsSceneHoverEvent>
#include <QGraphicsSceneMouseEvent>
#include <QGraphicsWidget>
#include <QPointF>
#include <stdint.h>

//--------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------

class CornerGrabber;

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------

class fiRectItem : public QGraphicsRectItem
{
public:
    //! Define a custom event to indicate when we have been changed.
    static const QEvent::Type CHANGE_EVENT;

public:
    explicit fiRectItem(QGraphicsItem *parent = nullptr);
    explicit fiRectItem(const QRectF& rect, QGraphicsItem *parent = nullptr);
    explicit fiRectItem(qreal x, qreal y, qreal w, qreal h, QGraphicsItem *parent = nullptr);
    ~fiRectItem();

    QRectF adjrect();
    void paint(QPainter *painter, const QStyleOptionGraphicsItem *option, QWidget *widget = nullptr);
    void set_edge(QPointF edge);

    void registerCallback(QObject *cb_ptr)      { cb_object = cb_ptr; }
    void setDisabledBrush(const QBrush& brush)  { disabled_brush = brush; }
    void setDisabledPen(const QPen& pen)        { disabled_pen = pen; }
    void setEnabledBrush(const QBrush& brush)   { enabled_brush = brush; }
    void setEnabledPen(const QPen& pen)         { enabled_pen = pen; }
    void setRect(const QRectF& rect);
    void setMaxArea(int area);
    void setMoveable(bool is_moveable)          { moveable = is_moveable; }

protected:
    virtual void hoverEnterEvent(QGraphicsSceneHoverEvent *event);
    virtual void hoverLeaveEvent(QGraphicsSceneHoverEvent *event);

    virtual void mouseMoveEvent(QGraphicsSceneMouseEvent *event);
    virtual void mouseReleaseEvent(QGraphicsSceneMouseEvent *event);

    virtual bool sceneEventFilter(QGraphicsItem *watched, QEvent *event);

private:
    void initialize();
    void set_corner_positions();

private:
    QRectF  clip_rect;
    QObject *cb_object;

    CornerGrabber *corners[4];          //!< Starting at x=0, y=0 and moving clockwise around the box.
    QPointF drag_start;                 //!< Where the mouse "down" event started a drag operation.

    QPen    disabled_pen;               //!< Pen to use to draw rectangle boarder when disabled.
    QBrush  disabled_brush;             //!< Brush to use to draw rectangle interior when disabled.
    QPen    enabled_pen;                //!< Pen to use to draw rectangle boarder when enabled.
    QBrush  enabled_brush;              //!< Brush to use to draw rectangle interior when enabled.
    int     max_area;                   //!< Maximum area allowed for the rectangle.

    bool    moveable;                   //!< Is this item moveable?
};


#endif // FI_RECTITEM_H
