/*!
 * @file fi_saver.cpp
 * @brief This is a utility class to facilitate some of the more
 *        mundane aspects of saving images.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QMessageBox>
#include <QApplication>
#include <QByteArray>
#include <QDateTime>
#include <QDir>
#include <QFile>
#include <QList>
#include <QStandardPaths>
#include <QString>
#include <QThread>

#include <errno.h>
#include <string.h>
#include <time.h>

#include <string>
#include <memory>

#include "fi_saver.h"
#include "fi_pixmap.h"

#include "fas_video.h"
#include "fi_image_pkg.h"
#include "fi_image.h"
#include "fi_internal.h"
#include "fi_bmp.h"
#include "fi_bmp_planar.h"
#include "fi_jpeg.h"

#include "client_math.h"
#include "logger.h"
#include "gige_msg.h"
#include "raw_to_xml.h"

#include "log_if.h"

#include "mainwindow.h"
#include "camera_mdi.h"
#include "control_panel.h"
#include "panel_lens.h"

namespace FI_SAVER
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
// Set the define to 1 if you want logging turned on, 0 if you want it turned off.
#define LOG_VERBOSE_SAVE        0
#define LOG_RETRY_DEBUG         0

//! A default playback rate to use.
const int DEFAULT_PLAYBACK_RATE                 = 25;

//! The maximum number of file/folder names we'll try in an attempt to find an "open" one.
const int   MAX_TRY_COUNT                       =   1000000;

//! The bottom threshold for warning about JPEG Quality values.
const uint32_t MIN_JPEG_QUALITY_THRESHOLD        = 10;   // Right now warn at something under 10%.

//! Verbose logging support.
#if LOG_VERBOSE_SAVE
static const bool VERBOSE_SAVE_LOG = true;
#else
static const bool VERBOSE_SAVE_LOG = false;
#endif

//! Retry logic logging support.
#if LOG_RETRY_DEBUG
static const bool RETRY_DEBUG_LOG = true;
#else
static const bool RETRY_DEBUG_LOG = false;
#endif

}   // End of namespace FI_SAVER.

//! The default "tag" for our file names.
const QString FiSaver::DEFAULT_TAG = QStringLiteral("fi");

//! The default metadata file name.
const QString FiSaver::DEFAULT_METADATA_FILE_NAME = QStringLiteral("metadata.txt");

//! The name of our statistics file.
const QString FiSaver::STATS_FILE_NAME = QStringLiteral("stats.csv");

//! File extension for JPG files.
const QString FiSaver::IMAGE_EXT_JPG = QStringLiteral("jpg");

//! File extension for TIF files.
const QString FiSaver::IMAGE_EXT_TIF = QStringLiteral("tif");

//! File extension for BMP files.
const QString FiSaver::IMAGE_EXT_BMP = QStringLiteral("bmp");

//! File extension for BMP Planar native files.
const QString FiSaver::IMAGE_EXT_BPP = QStringLiteral("bpp");

//! File extension for DNG files.
const QString FiSaver::IMAGE_EXT_DNG = QStringLiteral("dng");

//! File extension for CAP files.
const QString FiSaver::IMAGE_EXT_CAPTURE = QStringLiteral("cap");

//! File extension for CAP stack files.
const QString FiSaver::IMAGE_EXT_CAPSTACK = QStringLiteral("CAPSTK");

//! File extension for AVI files.
const QString FiSaver::IMAGE_EXT_AVI = QStringLiteral("avi");

//! File extension for MP4 files.
const QString FiSaver::IMAGE_EXT_MP4 = QStringLiteral("mp4");

//! File extension for RDI "root" file.
const QString FiSaver::IMAGE_EXT_RDI_PACKED = QStringLiteral("rdi");

//! File extension for TXT metadata files.
const QString FiSaver::IMAGE_EXT_META_TEXT = QStringLiteral("txt");

//! File extension for CFG metadata files.
const QString FiSaver::IMAGE_EXT_META_CFG = QStringLiteral("cfg");

//! File extension for XML metadata files.
const QString FiSaver::IMAGE_EXT_META_XML = QStringLiteral("xml");

//! File extension for RAW per-frame metadata files.
const QString FiSaver::IMAGE_EXT_META_RAW = QStringLiteral("pm");

//! Prefix for sub-folders in stacks.
const QString FiSaver::SUB_FOLDER_PREFIX = QStringLiteral("img");

// String associated with the STREAM_TYPE settings we support.
const QString FiSaver::FORMAT_JPG_STR       = QStringLiteral("JPG Stack");
const QString FiSaver::FORMAT_BMP_STR       = QStringLiteral("BMP Stack");
const QString FiSaver::FORMAT_TIF_STR       = QStringLiteral("TIF Stack");
const QString FiSaver::FORMAT_TIF_RAW_STR   = QStringLiteral("TIF (Raw) Stack");
const QString FiSaver::FORMAT_DNG_STR       = QStringLiteral("DNG Stack");
const QString FiSaver::FORMAT_AVI_STR       = QStringLiteral("AVI");
const QString FiSaver::FORMAT_MP4_STR       = QStringLiteral("MP4");
const QString FiSaver::FORMAT_CAP_STR       = QStringLiteral("CAP (Partition)");
const QString FiSaver::FORMAT_CAPSTK_STR    = QStringLiteral("CAP Stack");
const QString FiSaver::FORMAT_META_STR      = QStringLiteral("METADATA");
const QString FiSaver::FORMAT_RDI_PACKED_STR = QStringLiteral("RDI Packed");
const QString FiSaver::FORMAT_UNKNOWN_STR   = QStringLiteral("Unknown");

using namespace FI_SAVER;
using namespace FAS_VIDEO;

/*!
 * @brief The class constructor.
 */
FiSaver::FiSaver(QObject *parent) :
    QObject(parent),
    absRetryFrame(0),
    streamImageHeight(0),
    streamPlaybackRate(FI_SAVER::DEFAULT_PLAYBACK_RATE),
    streamImageWidth(0),
    blobFile(nullptr),
    blobSize(FSP_COMM::BLOB_CHUNK_SIZE),
    clipStart(0),
    clipEnd(0),
    convertRawToXml(false),
    destMedia(0),
    frameCount(0),
    jpegQuality_4QTPixmaps(-1),
    streamIsClosed(false),
    streamType(FI_SAVER::STREAM_TYPE::NO_STREAM),
    imageTypeForEncoder(FI_SAVER::IMAGE_TYPE::NO_IMAGE),
    imageInputType(FI_SAVER::IMAGE_TYPE::NO_IMAGE),
    h264Speed(FAS_VIDEO::DEFAULT_H264_SPEED),
    h264Quality(FAS_VIDEO::DEFAULT_H264_QUALITY_PCT),
    h264EnableAllKeyframes(false),
    retryInProgress(false),
    startIndex(0),
    usingSubfolders(false),
    saveMetadataCfgFile(FI_SAVER::DEFAULT_SAVE_METADATA_CFG),
    savePfmd(false),
    rawToXml(nullptr)
{
    streamWriter.reset();
    streamImage.reset();
    imageByteArray.clear();
    imageFilename.clear();
    streamHandle = 0;

    // Allocate a buffer for our comment, if any.
    commentBuf = new QByteArray;
    if (commentBuf == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate comment_buf!";
        return;
    }

    // For debug runs, make sure we have a video log file ready to catch any AVI errors.
    QString vidlog;
#ifdef QT_DEBUG
    vidlog = qApp->applicationDirPath() + QStringLiteral("/fm_video.log");
#else
    vidlog = QDir::tempPath() + QStringLiteral("/fm_video.log");
#endif

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "videoLogPath:" << vidlog;

        // Mark the construction start time.
        QString time_str = QDateTime::currentDateTime().toString();
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "===================================" << time_str.toLocal8Bit().constData();
    }
}

/*!
 * @brief The class destructor.
 */
FiSaver::~FiSaver()
{
    // Relase allocated memory.
    if (commentBuf != nullptr)
    {
        delete commentBuf;
        commentBuf = nullptr;
    }

    imageByteArray.clear();
    if (VERBOSE_SAVE_LOG)
    {
        // Mark the deconstruction time.
        QString time_str = QDateTime::currentDateTime().toString();
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "===================================" << time_str.toLocal8Bit().constData();
    }
}

/*!
 * @brief Is the save stream type a stack of images.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns A boolean true if the stream type is a stack, false if not.
 */
bool FiSaver::isStackOfImages(FI_SAVER::STREAM_TYPE streamType)
{
    bool isAStack = false;

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        isAStack = true;
        break;
    default:
        isAStack = false;
        break;
    }

    return isAStack;
}

/*!
 * @brief Is the save stream type a stack of CAP files.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns A boolean true if the stream type is a stack, false if not.
 */
bool FiSaver::isStackOfCaps(FI_SAVER::STREAM_TYPE streamType)
{
    bool isAStack = false;

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
        isAStack = true;
        break;
    default:
        isAStack = false;
        break;
    }

    return isAStack;
}

/*!
 * @brief Is the save stream type a stream video.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns A boolean true if the stream type is a video, false if not.
 */
bool FiSaver::isStreamVideo(FI_SAVER::STREAM_TYPE streamType)
{
    bool isAVideo = false;

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::AVI:
    case FI_SAVER::STREAM_TYPE::MP4:
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        isAVideo = true;
        break;
    default:
        isAVideo = false;
        break;
    }

    return isAVideo;
}

/*!
 * @brief Is the save stream type an ffmpeg stream video.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns A boolean true if the stream type is an ffmpeg video, false if not.
 */
bool FiSaver::isStreamFFMpegVideo(FI_SAVER::STREAM_TYPE streamType)
{
    bool isAFFmpegVideo = false;

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::MP4:
        isAFFmpegVideo = true;
        break;
    default:
        isAFFmpegVideo = false;
        break;
    }

    return isAFFmpegVideo;
}

/*!
 * @brief Is the save stream type a camara capture file.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns A boolean true if the stream type is a capture file, false if not.
 */
bool FiSaver::isCaptureStream(FI_SAVER::STREAM_TYPE streamType)
{
    bool isACaptureFile = false;

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        isACaptureFile = true;
        break;
    default:
        isACaptureFile = false;
        break;
    }

    return isACaptureFile;
}

/*!
 * @brief Is the setup valid for a snapshot image.
 * Wanted to pull this logic to static function in case in future it changes.
 *
 * @param imageType     The type of image being requested.
 * @param frameCount    The number of frames being requested.
 *
 * @returns A boolean true if the snapshot conditions are set.
 */
bool FiSaver::isSnapshotImage(FI_SAVER::IMAGE_TYPE imageType, uint32_t frameCount)
{
    bool snapShotCondition = false;

    // Right now this common logic is used to determine if a snapshot condition is set.
    if ((imageType == FI_SAVER::IMAGE_TYPE::JPEG) && (frameCount == 1))
        snapShotCondition = true;

    return snapShotCondition;
}

/*!
 * @brief Translate internal stream type to camera GIGE type.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns The GIGE stream format value that matches our stream type or zero
 * if unable to translate OR the camera does not support the streamType.
 */
uint32_t FiSaver::streamTypeToCameraStreamType(FI_SAVER::STREAM_TYPE streamType, bool is_series3)
{
    uint32_t gigeStreamType = 0;      // A zero means no translation was determined.

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        gigeStreamType = GIGE_STREAM_FORMAT_BMP;
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        gigeStreamType = GIGE_STREAM_FORMAT_JPEG;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        gigeStreamType = GIGE_STREAM_FORMAT_TIF;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
        gigeStreamType = (GIGE_STREAM_FORMAT_TIF | GIGE_STREAM_FORMAT_RAW);
        break;
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        gigeStreamType = GIGE_STREAM_FORMAT_DNG;
        break;
    case FI_SAVER::STREAM_TYPE::AVI:
        gigeStreamType = GIGE_STREAM_FORMAT_AVI;
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
        // NOTE_JLC:  I am leaving this comment in place because the caller of this
        // function needs to understand that if MP4 format is returned it better make
        // sure that the camera system can handle an MP4 stream save.
        gigeStreamType = GIGE_STREAM_FORMAT_MP4;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        gigeStreamType = GIGE_STREAM_FORMAT_CAPTURE;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
        gigeStreamType = GIGE_STREAM_FORMAT_CAPSTK;
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        // For series3 cameras we will use raw TIFFs.
        if (is_series3)
            gigeStreamType = (GIGE_STREAM_FORMAT_TIF | GIGE_STREAM_FORMAT_RAW);
        else
            gigeStreamType = GIGE_STREAM_FORMAT_PACK;
        break;
    default:
        gigeStreamType = 0;   // Unable to translate to a camera save setting.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "UNKNOWN FI_SAVER::STREAM_TYPE[" << streamType << "].";
        break;
    }

    return gigeStreamType;
}

/*!
 * @brief Translate internal stream type to camera GIGE type.
 *
 * @param streamType    The type of stream being saved by fiSaver.
 * @param imageType     The type of images being used for the streamType.
 *
 * @returns The GIGE stream format value that matches our stream type or zero
 * if unable to translate.
 */
uint32_t FiSaver::imageTypeToCameraImageType(FI_SAVER::STREAM_TYPE streamType, FI_SAVER::IMAGE_TYPE imageType, bool is_series3)
{
    uint32_t gigeImageType = 0;      // A zero means no translation was determined.

    // The reason the streamType is passed in here is that there are associations with stream
    // types and images that are compatible with the stream type.  This makes sure that they
    // match and if not raises an error condition.

    switch (imageType)
    {
    //#MP4IMAGEINPUT - fiSaver::ImageTypeToCameraImageType - PLANAR support is here.
    case FI_SAVER::IMAGE_TYPE::BMP_PLANAR:
        // This is only valid for MP4 streams.
        if (streamType == FI_SAVER::STREAM_TYPE::MP4)
            gigeImageType = GIGE_STREAM_FORMAT_PLANAR;
        break;
    case FI_SAVER::IMAGE_TYPE::BMP:
        gigeImageType = GIGE_STREAM_FORMAT_BMP;
        break;
    case FI_SAVER::IMAGE_TYPE::JPEG:
        gigeImageType = GIGE_STREAM_FORMAT_JPEG;
        break;
    // RECHECK_JLC - these weren't here so when saving to the camera itself the process
    // error'd out and it should be allowable.
    case FI_SAVER::IMAGE_TYPE::TIF:
        // This is only valid for TIF stacks.
        if (streamType == FI_SAVER::STREAM_TYPE::TIF_STACK)
            gigeImageType = GIGE_STREAM_FORMAT_TIF;
        break;
    case FI_SAVER::IMAGE_TYPE::TIF_RAW:
        // This is only valid for TIF stacks.
        if ((streamType == FI_SAVER::STREAM_TYPE::TIF_RAW_STACK) || ((streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED) && is_series3))
            gigeImageType = (GIGE_STREAM_FORMAT_TIF | GIGE_STREAM_FORMAT_RAW);
        break;
    case FI_SAVER::IMAGE_TYPE::PACKED:
        // Only for RDI_PACKED streams.
        if (streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
        {
            // For series3 cameras we will use raw TIFFs.
            if (is_series3)
                gigeImageType = (GIGE_STREAM_FORMAT_TIF | GIGE_STREAM_FORMAT_RAW);
            else
                gigeImageType = GIGE_STREAM_FORMAT_PACK;
        }
        break;
    case FI_SAVER::IMAGE_TYPE::DNG:
        // This is only valid for DNG stacks.
        if (streamType == FI_SAVER::STREAM_TYPE::DNG_STACK)
            gigeImageType = GIGE_STREAM_FORMAT_DNG;
        break;
    default:
        gigeImageType = 0;   // Unable to translate to a camera save setting.
        break;
    }

    // This is a programming error!
    if (gigeImageType == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "NO GIGE Image Type Translatable" << "IMAGE TYPE: " << imageType << " STREAM TYPE: " << streamType << "!";
    }

    return gigeImageType;
}

/*!
 * @brief Translate internal stream type to user readable string.
 *
 * @param streamType The type of FI Saver stream.
 *
 * @returns The user readable string created here for UI purposes.
 */
QString FiSaver::streamTypeToQString(FI_SAVER::STREAM_TYPE streamType, bool forLog)
{
    QString streamTypeStr;

    if (forLog)
    {
        streamTypeStr = "Video Stream[";
    }

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        streamTypeStr += FiSaver::FORMAT_BMP_STR;
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        streamTypeStr += FiSaver::FORMAT_JPG_STR;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        streamTypeStr += FiSaver::FORMAT_TIF_STR;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
        streamTypeStr += FiSaver::FORMAT_TIF_RAW_STR;
        break;
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        streamTypeStr += FiSaver::FORMAT_DNG_STR;
        break;
    case FI_SAVER::STREAM_TYPE::AVI:
        streamTypeStr += FiSaver::FORMAT_AVI_STR;
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
        streamTypeStr += FiSaver::FORMAT_MP4_STR;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        streamTypeStr += FiSaver::FORMAT_CAP_STR;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
        streamTypeStr += FiSaver::FORMAT_CAPSTK_STR;
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        streamTypeStr += FiSaver::FORMAT_RDI_PACKED_STR;
        break;
    case FI_SAVER::STREAM_TYPE::NO_STREAM:
        streamTypeStr += FiSaver::FORMAT_UNKNOWN_STR;
        break;
    default:
        streamTypeStr += FiSaver::FORMAT_UNKNOWN_STR;
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "UNKNOWN FI_SAVER::STREAM_TYPE[" << streamType << "].";
        break;
    }

    if (forLog)
    {
        streamTypeStr += "]";
    }

    return streamTypeStr;
}

/*!
 * @brief Translate internal image type to user readable string.
 *
 * @param imageType The type of FI Saver stream.
 *
 * @returns The user readable string created here for UI purposes.
 */
QString FiSaver::imageTypeToQString(FI_SAVER::IMAGE_TYPE imageType)
{
    QString imageTypeStr = "Video Stream Image[";

    switch (imageType)
    {
    case FI_SAVER::IMAGE_TYPE::JPEG:
        imageTypeStr += "JPEG]";
        break;
    case FI_SAVER::IMAGE_TYPE::BMP:
        imageTypeStr += "BMP]";
        break;
    case FI_SAVER::IMAGE_TYPE::BMP_PLANAR:
        imageTypeStr += "BMP_PLANAR]";
        break;
    case FI_SAVER::IMAGE_TYPE::TIF:
        imageTypeStr += "TIF]";
        break;
    case FI_SAVER::IMAGE_TYPE::TIF_RAW:
        imageTypeStr += "TIF_RAW]";
        break;
    case FI_SAVER::IMAGE_TYPE::DNG:
        imageTypeStr += "DNG]";
        break;
    default:
        imageTypeStr += "NO_IMAGE]";
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "UNKNOWN FI_SAVER::IMAGE_TYPE[" << imageType << "].";
        }
        break;
    }

    return imageTypeStr;
}

/*!
 * @brief Provide a translate from a GIGE_STREAM_FORMAT setting into a user readable string.
 */
QString FiSaver::getGIGEStreamStr(uint32_t gigeStreamFormat)
{
    QString theUserFriendlyStreamFormat;

    // Switch the GIGE stream format register setting.
    switch (gigeStreamFormat)
    {
    case GIGE_STREAM_FORMAT_JPEG:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_JPEG";
        break;
    case GIGE_STREAM_FORMAT_BMP:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_BMP";
        break;
    case GIGE_STREAM_FORMAT_TIF:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_TIF";
        break;
    case (GIGE_STREAM_FORMAT_TIF | GIGE_STREAM_FORMAT_RAW):
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_TIF + RAW";
        break;
    case GIGE_STREAM_FORMAT_DNG:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_DNG";
        break;
    case GIGE_STREAM_FORMAT_AVI:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_AVI";
        break;
    case GIGE_STREAM_FORMAT_CAPTURE:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_CAPTURE";
        break;
    case GIGE_STREAM_FORMAT_META_ONLY:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_META_ONLY";
        break;
    case GIGE_STREAM_FORMAT_MP4:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_MP4";
        break;
    case GIGE_STREAM_FORMAT_PLANAR:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_PLANAR";
        break;
    case GIGE_STREAM_FORMAT_PACK:
        theUserFriendlyStreamFormat = "GIGE_STREAM_FORMAT_PACK";
        break;

    default:
        theUserFriendlyStreamFormat = "UNKNOWN GIGE_STREAM_FORMAT VALUE";
        break;
    }
    return theUserFriendlyStreamFormat;
}

/*!
 * @brief A static verification module to output a warning if the data is invalid.
 * @returns State of verification, true if passed, false if did not pass.
 */
bool FiSaver::verifyJpegQuality(const char *pretty_str, const char *variable_name, uint32_t value_to_check, bool showMBWarning)
{
    bool jpeg_verified_ok = true;
    QString the_message;

    if (pretty_str == nullptr)
    {
        pretty_str = __PRETTY_FUNCTION__;
    }

    if (value_to_check < MIN_JPEG_QUALITY_THRESHOLD)
    {
        jpeg_verified_ok = false;
        the_message = QString("[JPEG QUALITY WARNING] Setting [%1] is below %2 percent!").arg(value_to_check, MIN_JPEG_QUALITY_THRESHOLD);

        QLOG_WARN() << pretty_str  << " Name[" << variable_name << "]" << the_message;

        if (showMBWarning)
        {
            // Pop up a warning dialog because caller asked for the display.
            QMessageBox::warning(MainWindow::GetInstance().get(), "JPEG Quality Warning", tr("JPEG Quality Factor ."));
        }
    }
    return jpeg_verified_ok;
}

/*!
 * @brief A helper function used to build the target path based on user settings for
 *        incorporating the camera name, a user-specified tag value and a unique
 *        index number.
 */
void FiSaver::buildDestPath()
{
    bool    do_snapshot;

    // Are we doing a snapshot copy here?
    do_snapshot = FiSaver::isSnapshotImage(imageInputType, frameCount);

    // Build up our target path.
    if (cameraName.isEmpty())
    {
        if (tagString.isEmpty())
        {
            // No camera name, no tag.
            destPath = basePath + DEFAULT_TAG;
        }
        else
        {
            // No camera name, but use the tag.
            destPath = basePath + tagString;
        }
    }
    else
    {
        if (tagString.isEmpty())
        {
            // Use the camera name, but no tag.
            destPath = basePath + cameraName;
        }
        else
        {
            // Use the camera name AND the tag name.
            destPath = basePath + cameraName + QStringLiteral("_") + tagString;
        }
    }

    // Do we want an index?
    if (!do_snapshot && (startIndex != (uint32_t) -1))
        destPath = destPath + QStringLiteral("_%1").arg(startIndex, PATH_SEQ_NUM_LEN, 10, QLatin1Char('0'));

    // Make sure we're properly formatted.
    destPath = QDir::fromNativeSeparators(destPath);

    // Make sure it ends with a delimiter.
    if (!destPath.endsWith(QLatin1Char('/')))
        destPath.append(QLatin1Char('/'));
}

/*!
 * @brief A helper function to build up a file name for an image file.
 *
 * @param[out] file_name            The returned fully qualified file name path.
 * @param[in]  file_seq_num         Sequence number of the image we're going to be saving.
 * @param[in]  do_retry             Is this a retry frame?
 */
void FiSaver::buildFileName(QString& file_name, uint32_t file_seq_num, bool do_retry)
{
    uint32_t    frame_num;
    QString     folder_path;

    // If it's a metadata sequence number, then the passed in value will be off by one.
    switch (file_seq_num + 1)
    {
    case METADATA_TXT_SEQ_NUM:
        // Special handling for video stream writers and capture streams.
        if (FiSaver::isStreamVideo(streamType) || FiSaver::isCaptureStream(streamType))
        {
            if (VERBOSE_SAVE_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "METADATA TEXT SEQ dest_path:" << destPath << "fileExtension:" << fileExtension;
            }
            file_name = destPath;
            file_name.replace(file_name.lastIndexOf(fileExtension), fileExtension.length(), IMAGE_EXT_META_TEXT);
        }
        else
        {
            file_name = destPath + DEFAULT_METADATA_FILE_NAME;
        }
        break;
    case METADATA_CFG_SEQ_NUM:
        // Special handling for AVIs & CAPs.
        if (FiSaver::isStreamVideo(streamType) || FiSaver::isCaptureStream(streamType))
        {
            file_name = destPath;
            file_name.replace(file_name.lastIndexOf(fileExtension), fileExtension.length(), IMAGE_EXT_META_CFG);
        }
        else
        {
            file_name = destPath + DEFAULT_METADATA_FILE_NAME;
            file_name.replace(file_name.lastIndexOf(IMAGE_EXT_META_TEXT), IMAGE_EXT_META_TEXT.length(), IMAGE_EXT_META_CFG);
        }
        break;
    case METADATA_XML_SEQ_NUM:
        // Special handling for AVIs & CAPs.
        if (FiSaver::isStreamVideo(streamType) || FiSaver::isCaptureStream(streamType))
        {
            file_name = destPath;
            file_name.replace(file_name.lastIndexOf(fileExtension), fileExtension.length(), IMAGE_EXT_META_XML);
        }
        else
        {
            file_name = destPath + DEFAULT_METADATA_FILE_NAME;
            file_name.replace(file_name.lastIndexOf(IMAGE_EXT_META_TEXT), IMAGE_EXT_META_TEXT.length(), IMAGE_EXT_META_XML);
        }
        break;
    case METADATA_RAW_SEQ_NUM:
        // Special handling for AVIs & CAPs.
        if (FiSaver::isStreamVideo(streamType) || FiSaver::isCaptureStream(streamType))
        {
            file_name = destPath;
            file_name.replace(file_name.lastIndexOf(fileExtension), fileExtension.length(), IMAGE_EXT_META_RAW);
        }
        else
        {
            file_name = destPath + DEFAULT_METADATA_FILE_NAME;
            file_name.replace(file_name.lastIndexOf(IMAGE_EXT_META_TEXT), IMAGE_EXT_META_TEXT.length(), IMAGE_EXT_META_RAW);
        }
        break;
    default:
        // Calculate the frame number based on the sequence number and the starting frame number.
        frame_num = clipStart + file_seq_num;

        // Are we building sub-folders?
        if (usingSubfolders)
        {
            int     folder_index;

            // Which sub-folder are we going to?
            folder_index = file_seq_num / SUB_FOLDER_MAX_IMAGES;
            generateSubfolderName(folder_path, folder_index);
        }
        else
        {
            // Are we saving a snapshot?  If in a retry case it's not a snapshot.
            if (!do_retry && FiSaver::isSnapshotImage(imageInputType, frameCount))
            {
                int     try_count;

                // Remove the terminating '/' and replace it with our extension.
                folder_path = destPath.left(destPath.size() - 1);
                folder_path.append(QLatin1Char('_'));

                // Loop until we find an available file name.
                for (try_count = 0; try_count < MAX_TRY_COUNT; ++try_count, ++frame_num)
                {
                    QFileInfo   file_info;

                    // Build up our file name.
                    file_name = QString::fromUtf8("%1%2.%3").arg(folder_path).arg(frame_num, IMAGE_SEQ_NUM_LEN, 10, QLatin1Char('0')).arg(fileExtension);

                    // Does the file already exist?
                    file_info.setFile(file_name);
                    if (!file_info.exists())
                        break;
                }
                if (try_count == MAX_TRY_COUNT)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find an available name!";
                }
            }

            // Are we saving one file and it's not the retry frame?
            else if ((frameCount == 1) && !retryInProgress)
            {
                // Set the base folder path and the base file name.
                folder_path = destPath + DEFAULT_TAG + QString::fromUtf8("_");
            }
            else
            {
                // Set the base folder path.
                folder_path = destPath;
            }
        }

        // Build up our file name.
        file_name = QString::fromUtf8("%1%2.%3").arg(folder_path).arg(frame_num, IMAGE_SEQ_NUM_LEN, 10, QLatin1Char('0')).arg(fileExtension);
        break;
    }
}

/*!
 * @brief After the user has specified all the settings for a save operation,
 *        this method initializes the application and camera to perform the
 *        save operation.
 *
 * @param save_info     The FI_SAVER::SAVE_PARAMS structure.
 * @return True on success, false otherwise.
 */
bool FiSaver::configure(const SAVE_PARAMS& save_info)
{
    bool            rc = false;

    // Setup the comment. If the user wants one or not, we have to do this to make
    // sure that the comment is set or cleared, as required.
    commentBuf->clear();
    commentBuf->append(save_info.comment.toUtf8());
    commentBuf->append('\0');

    // Reset our filename.
    imageFilename.clear();

    // Set up data that can be used for status of last save to stream operation.
    lastFilenameWritten       = "No files have been written yet.";
    lastSaveHadError         = false;

    // Keep track of the last image sequence number received and the last one successfully saved.
    // Note:  A setting of zero is the sequence number of a live stream and is invalid for a save
    // so if that is what is set here nothing was processed.
    lastSeqNbrSaverReceived = 0;
    lastSeqNbrSaverSaved = 0;
    streamSaveBufferCount = 0;

    // Remember the destination.
    destMedia = save_info.dest_media;
    destLabel = save_info.dest_label;

    // Are we supposed to create an XML?
    convertRawToXml = save_info.create_xml;

    // Do we want to save the PFMD?
    savePfmd = save_info.save_pfmd;

    // Catch the metadata CFG save logic.
    // Note that these are only used when the call is made to fiSaver::save_partation_metadata
    // so if this process does not call it then it won't matter - but I am not sure how many
    // places this is used to transfer the data and I don't want to miss this state.
    setMetadataCfgSave(save_info.save_metadata_cfg);

    // Return our result.
    return rc;
}

/*!
 * @brief Convert a file extension string to an image type.
 * @param fileExtensionStr  The file extension string to translate.
 * @return The IMAGE_TYPE associated with a file extension.
 */
FI_SAVER::IMAGE_TYPE FiSaver::fileExtToImageType(QString fileExtensionStr)
{
    FI_SAVER::IMAGE_TYPE imageType;

    // Compare the extension until we find one we know.
    if (fileExtensionStr.compare(IMAGE_EXT_BMP, Qt::CaseInsensitive) == 0)
        imageType = FI_SAVER::IMAGE_TYPE::BMP;
    else if (fileExtensionStr.compare(IMAGE_EXT_DNG, Qt::CaseInsensitive) == 0)
        imageType = FI_SAVER::IMAGE_TYPE::DNG;
    else if ((fileExtensionStr.compare(IMAGE_EXT_JPG, Qt::CaseInsensitive) == 0) ||
             (fileExtensionStr.compare(QString::fromUtf8("jpeg"), Qt::CaseInsensitive) == 0))
        imageType = FI_SAVER::IMAGE_TYPE::JPEG;
    else if ((fileExtensionStr.compare(IMAGE_EXT_TIF, Qt::CaseInsensitive) == 0) ||
             (fileExtensionStr.compare(QString::fromUtf8("tiff"), Qt::CaseInsensitive) == 0))
        imageType = FI_SAVER::IMAGE_TYPE::TIF;  // Cannot determine raw from other tif just by extension.
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to convert file extension:" << fileExtensionStr;
        imageType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;     // Cannot determine based on extension.
    }

    // Return our type.
    return imageType;
}

/*!
 * @brief Convert a file extension string to an image type.
 * @param fileExtensionStr  The file extension string to translate.
 * @return The IMAGE_TYPE associated with a file extension.
 */
FI_SAVER::STREAM_TYPE FiSaver::fileExtToStreamType(QString fileExtensionStr)
{
    FI_SAVER::STREAM_TYPE streamType;

    // Compare the extension until we find one we know.
    if (fileExtensionStr.compare(IMAGE_EXT_AVI, Qt::CaseInsensitive) == 0)
        streamType = FI_SAVER::STREAM_TYPE::AVI;
    else if (fileExtensionStr.compare(IMAGE_EXT_MP4, Qt::CaseInsensitive) == 0)
        streamType = FI_SAVER::STREAM_TYPE::MP4;
    else if (fileExtensionStr.compare(IMAGE_EXT_BMP, Qt::CaseInsensitive) == 0)
        streamType = FI_SAVER::STREAM_TYPE::BMP_STACK;
    //TODO_JLC FFLR How does a cap stack impact this function?
    else if (fileExtensionStr.compare(IMAGE_EXT_CAPTURE, Qt::CaseInsensitive) == 0)
        streamType = FI_SAVER::STREAM_TYPE::FI_CAP;
    else if (fileExtensionStr.compare(IMAGE_EXT_DNG, Qt::CaseInsensitive) == 0)
        streamType = FI_SAVER::STREAM_TYPE::DNG_STACK;
    else if ((fileExtensionStr.compare(IMAGE_EXT_JPG, Qt::CaseInsensitive) == 0) ||
             (fileExtensionStr.compare(QString::fromUtf8("jpeg"), Qt::CaseInsensitive) == 0))
        streamType = FI_SAVER::STREAM_TYPE::JPEG_STACK;
    else if ((fileExtensionStr.compare(IMAGE_EXT_TIF, Qt::CaseInsensitive) == 0) ||
             (fileExtensionStr.compare(QString::fromUtf8("tiff"), Qt::CaseInsensitive) == 0))
        streamType = FI_SAVER::STREAM_TYPE::TIF_STACK;
    else if (fileExtensionStr.compare(IMAGE_EXT_RDI_PACKED, Qt::CaseInsensitive) == 0)
        streamType = FI_SAVER::STREAM_TYPE::RDI_PACKED;
    else
    {
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to convert file extension:" << fileExtensionStr;
        streamType = FI_SAVER::STREAM_TYPE::NO_STREAM;
    }

    // Return our type.
    return streamType;
}

FI_SAVER::STREAM_TYPE FiSaver::streamFormatToStreamType(uint32_t stream_format)
{
    STREAM_TYPE ret_type;

    switch (stream_format & GIGE_STREAM_FORMAT_MASK)
    {
    case GIGE_STREAM_FORMAT_JPEG:
        ret_type = JPEG_STACK;
        break;
    case GIGE_STREAM_FORMAT_BMP:
        ret_type = BMP_STACK;
        break;
    case GIGE_STREAM_FORMAT_TIF:
        if ((stream_format & GIGE_STREAM_FORMAT_RAW) == GIGE_STREAM_FORMAT_RAW)
            ret_type = TIF_RAW_STACK;
        else
            ret_type = TIF_STACK;
        break;
    case GIGE_STREAM_FORMAT_DNG:
        ret_type = DNG_STACK;
        break;
    case GIGE_STREAM_FORMAT_AVI:
        ret_type = AVI;
        break;
    case GIGE_STREAM_FORMAT_CAPTURE:
        ret_type = FI_CAP;
        break;
    case GIGE_STREAM_FORMAT_CAPSTK:
        ret_type = FI_CAPSTK;
        break;
    case GIGE_STREAM_FORMAT_MP4:
        ret_type = MP4;
        break;
    case GIGE_STREAM_FORMAT_PACK:
        ret_type = RDI_PACKED;
        break;
    default:
        ret_type = NO_STREAM;
        break;
    }

    // Return the stream type.
    return ret_type;
}

/*!
 * @brief Create a unique destination path to save images to. The class variable [dest_path](@ref SaveVideo::dest_path) contains
 *        the fully qualified path.
 *
 * @return 0 on success, <0 on error.
 */
int FiSaver::createDestPath()
{
    int     try_count;
    QDir    dir_info;

    // Loop until we find an available folder name.
    for (try_count = 0; try_count < MAX_TRY_COUNT; ++try_count, ++startIndex)
    {
        // Clear out our destination path.
        destPath.clear();

        // Build the path.
        buildDestPath();

        // Just use the default path if we're only dealing with one file.
        if (frameCount == 1)
            break;

        // Is this folder available?
        dir_info.setPath(destPath);
        if (!dir_info.exists())
        {
            // Got an available name.
            break;
        }
    }
    if (try_count == MAX_TRY_COUNT)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find an available name!";
        return -1;
    }

    // Are we doing a snapshot copy here?
    if (FiSaver::isSnapshotImage(imageInputType, frameCount))
    {
        usingSubfolders = false;
        return 0;
    }

    // We have our base path. See if we can create the folder.
    if (!dir_info.mkpath(destPath))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create folder:" + destPath;
        return -1;
    }

    // Check to see if we need sub-folders.
    createSubFolders();

    // All is well.
    return 0;
}

/*!
 * @brief Create a unique destination file.
 *
 * Used for Video stream saves, this guarantees that we are not overwriting
 * a pre-existing video.
 *
 * @return 0 on success, errno otherwise.
 */
int FiSaver::createFilePath()
{
    int     try_count;
    QDir    dir_info;
    QString file_name;
    QFileInfo   file_info;

    // Check to see if the destination folder exists.
    dir_info.setPath(basePath);
    if (!dir_info.exists())
    {
        // See if we can create the folder.
        if (!dir_info.mkpath(basePath))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create folder:" << basePath;
            return EACCES;
        }
    }

    // Loop until we find an available file name.
    for (try_count = 0; try_count < MAX_TRY_COUNT; ++try_count, ++startIndex)
    {
        QString ancillary_file_to_dot;

        // Build up our filename.
        buildDestPath();

        // Make sure this folder doesn't exist.
        dir_info.setPath(destPath);
        if (dir_info.exists())
            continue;

        // Remove the terminating '/' and replace it with our extension.
        file_name = destPath.left(destPath.size() - 1);
        file_name.append(QLatin1Char('.'));

        // Save a full path copy to the dot before of file extenstion.
        ancillary_file_to_dot = file_name;

        // Pop the name of the actual file onto the check case.
        file_name.append(fileExtension);

        // Does the file already exist?
        file_info.setFile(file_name);
        if (!file_info.exists())
        {
            QString ancillary_file_name;

            // Check for TXT metadata files.
            ancillary_file_name = ancillary_file_to_dot + IMAGE_EXT_META_TEXT;
            file_info.setFile(ancillary_file_name);
            if (file_info.exists())
                continue;   // Can't use this one.

            // Check for CFG metadata files.
            ancillary_file_name = ancillary_file_to_dot + IMAGE_EXT_META_CFG;
            file_info.setFile(ancillary_file_name);
            if (file_info.exists())
                continue;   // Can't use this one.

            // Check for XML metadata files.
            ancillary_file_name = ancillary_file_to_dot + IMAGE_EXT_META_XML;
            file_info.setFile(ancillary_file_name);
            if (file_info.exists())
                continue;   // Can't use this one.

            // Check for RAW metadata files.
            ancillary_file_name = ancillary_file_to_dot + IMAGE_EXT_META_RAW;
            file_info.setFile(ancillary_file_name);
            if (file_info.exists())
                continue;   // Can't use this one.

            // With the addition of allowing for CFG or TXT files to become
            // configurable (to have or not to have that is the question) then
            // if we are creating an MP4 here we need to also check for AVI
            // and treat it like an "ancillary" file and vice-versa.

            // Check for AVI files with this name.
            ancillary_file_name = ancillary_file_to_dot + IMAGE_EXT_AVI;
            file_info.setFile(ancillary_file_name);
            if (file_info.exists())
                continue;   // Can't use this one.

            // Check for MP4 files with this name.
            ancillary_file_name = ancillary_file_to_dot + IMAGE_EXT_MP4;
            file_info.setFile(ancillary_file_name);
            if (file_info.exists())
                continue;   // Can't use this one.

            // Check for RDI "root" file with this name.
            ancillary_file_name = ancillary_file_to_dot + IMAGE_EXT_RDI_PACKED;
            file_info.setFile(ancillary_file_name);
            if (file_info.exists())
                continue;   // Can't use this one.

            // If make it here we are good to go!!
            break;
        }
    }
    if (try_count == MAX_TRY_COUNT)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find an available name!";
        return EACCES;
    }

    // Make our name available.
    destPath = file_name;

    // All is well.
    return 0;
}

/*!
 * @brief Helper function used to create subfolders. This is necessary in the event that we're
 *        saving [a lot](@ref SAVE_VIDEO::SUB_FOLDER_MAX_IMAGES) of images.
 */
void FiSaver::createSubFolders()
{
    int     index;
    int     folder_count;

    // Assume we'll not be using sub-folders;
    usingSubfolders = false;

    // If we're not saving stacks, then we don't need to worry about this.
    if (FiSaver::isStreamVideo(streamType))
        return;

    // Are we beyond our threshold?
    if (frameCount < (uint32_t)SUB_FOLDER_THRESHOLD)
        return;

    // We are using sub-folders.
    usingSubfolders = true;

    // How many folders do we need? Round up.
    folder_count = (frameCount + (SUB_FOLDER_MAX_IMAGES - 1)) / SUB_FOLDER_MAX_IMAGES;

    // Loop through and create all the sub-folders.
    for (index = 0; index < folder_count; ++index)
    {
        QDir    dir_info;
        QString sub_folder;

        // Build up the folder name.
        generateSubfolderName(sub_folder, index);

        // Create the folder.
        dir_info.setPath(sub_folder);
        if (!dir_info.mkpath(sub_folder))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to create folder:" << sub_folder;
        }
    }
}

/*!
 * @brief When all the images have been downloaded, this method flushes
 * out the video stream writer and closes the video file.
 *
 * @note This is only used when saving an video file.
 * @note This is run in the image_saver() thread context.
 */
void FiSaver::finishSave()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "streamType: " << FiSaver::streamTypeToQString(streamType);

    // Only have extra stuff to do if we're saving Video or Capture BLOBs.
    if (!FiSaver::isStreamVideo(streamType) && !FiSaver::isCaptureStream(streamType))
    {
        return;
    }

    // Sanity checks.
    if (blobFile != nullptr)
    {
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Closing the BLOB file.";
        }
        blobFile->close();
        blobFile->deleteLater();
        blobFile = nullptr;
    }

    // Blobs are done.
    if (FiSaver::isCaptureStream(streamType))
        return;

    // Close the video stream writer.
    closeStreamWriter();
}

/*!
 * @brief A utility function to generate the subfolder name based on the "index" parameter.
 *
 * @param sub_folder    The newly created subfolder name.
 * @param index         The index to use when building the name.
 */
void FiSaver::generateSubfolderName(QString& sub_folder, int index)
{
    sub_folder = QString::fromUtf8("%1%2%3").arg(destPath, SUB_FOLDER_PREFIX).arg(index, 4, 10, QLatin1Char('0'));
    sub_folder = QDir::fromNativeSeparators(sub_folder);

    if (!sub_folder.endsWith(QLatin1Char('/')))
        sub_folder.append(QLatin1Char('/'));
}

/*!
 * @brief Check to see if a sequence number is a live image.
 * @param seqNumberToCheck - the sequence number to check.
 * @return True if the sequence number is a live image, false otherwise.
 */
bool FiSaver::isLiveSeqNumber(uint32_t seqNumberToCheck)
{
    if (seqNumberToCheck == LIVE_IMAGE_SEQ_NUM)
        return true;
    return false;
}

/*!
 * @brief Check to see if the seq_num parameter is one of the metadata file sequence numbers.
 * @param seq_num   Sequence number we're interested in.
 * @return True if the sequence number is one of the metadata file numbers, false otherwise.
 */
bool FiSaver::metadataSeqNum(uint32_t seq_num)
{
    bool    rc;

    // See if the sequence number is a metadata file sequence number.
    switch (seq_num)
    {
    case METADATA_TXT_SEQ_NUM:
    case METADATA_CFG_SEQ_NUM:
    case METADATA_XML_SEQ_NUM:
    case METADATA_RAW_SEQ_NUM:
        rc = true;
        break;
    default:
        rc = false;
        break;
    }

    // Return our findings.
    return rc;
}

/*!
 * @brief Utility function to convert the pm_metadata structure to a text format.
 *
 * @param metadata_ptr  Pointer to a byte array containing the pm_metadata.
 * @param streamType    Type of stream being created.
 * @param imageType     Image type used for streamType.
 * @param clip_start    Starting frame number (inclusive).
 * @param clip_end      Ending frame number (inclusive).
 * @param gainsDataPtr  Read only gains settings.
 *
 * @return The text format of the pm_metadata information.
 */
QString FiSaver::metadataToText(QByteArray *metadata_ptr,
                                FI_SAVER::STREAM_TYPE streamType, FI_SAVER::IMAGE_TYPE imageType,
                                uint32_t clip_start, uint32_t clip_end, const FAS_VIDEO::VIDEO_GAINS *gainsDataPtr)
{
    QString imageFormatString;          // The resultant metadata in QT QString form.
    int     meta_len;                   // How many bytes are in the produced metadata string.
    std::string metadataString;         // The user text friendly metadata string.
    std::string h264SpeedText;          // The speed of the h.264 encoder if MP4 video save.

    // Sanity check.
    if ((metadata_ptr == nullptr) || (metadata_ptr->length() != sizeof(struct pm_metadata_type)))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Bad metadata!";
        return QString();
    }

    // Build up the metadata string.
    meta_len = FAS_VIDEO::FiVideo::CreateMetadataString(metadataString,
                                                        (void *)metadata_ptr->constData(),
                                                        (uint32_t)metadata_ptr->length(),
                                                        clip_start, clip_end,
                                                        gainsDataPtr);
    if (meta_len <= 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create metadata string! Err:" << meta_len;
        return QString();
    }

    // Add format information to the metadata string.
    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::AVI:
        if (imageType == FI_SAVER::IMAGE_TYPE::BMP)
            imageFormatString = QStringLiteral("\traw=0\n\tcodec=DIB\n");
        else
            imageFormatString = QStringLiteral("\traw=0\n\tcodec=MJPEG\n");
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
        FiVideo::H264SpeedSettingToText(h264SpeedText, h264Speed);
        imageFormatString = QString("\traw=0\n\tcodec=H264:speed=%1:quality=%2:keyframes=%3\n").arg(QString::fromStdString(h264SpeedText)).arg(h264Quality).arg(h264EnableAllKeyframes ? "All" : "60");
        break;
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        imageFormatString = QString("\traw=0\n\tcodec=BMP\n");
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
        imageFormatString = QString("\traw=1\n\tcodec=CAPSTK\n");
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        imageFormatString = QString("\traw=1\n\tcodec=CAP\n");
        break;
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        imageFormatString = QString("\traw=1\n\tcodec=DNG\n");
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        imageFormatString = QString("\traw=0\n\tcodec=JPEG\n");
        break;
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        imageFormatString = QString("\traw=0\n\tcodec=TIFF\n");
        break;
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
        imageFormatString = QString("\traw=1\n\tcodec=TIFF\n");
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected stream format:" << streamType;
        return QString();
    }

    // Add in the update information.
    imageFormatString = metadataString.c_str() + imageFormatString;

    return imageFormatString;
}

int FiSaver::preparePackedWriter(const pm_metadata_type *pm_ptr)
{
    FAS_INTERNAL::FIP_STREAM_CONFIGURATION configInfo;

    // Sanity checks.
    if (streamType != FI_SAVER::STREAM_TYPE::RDI_PACKED)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Called with non-packed streamType:" << streamType;
        return EFAULT;
    }
    if (pm_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Missing pm_metadata pointer!";
        return EFAULT;
    }

    // Reset our filename.
    imageFilename.clear();

    // Get a file name we can use. If it works, the result will be in destPath.
    int rc;
    if ((rc = createFilePath()) != 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "createFilePath failed, rc:" << rc;
        return rc;
    }

    // Initialize the FAS_INTERNAL::FIP_STREAM_CONFIGURATION structure.
    configInfo.manufacturer.assign(pm_ptr->man_name);
    configInfo.cameraModel.assign(pm_ptr->model_name);
    configInfo.cameraVersion.assign(pm_ptr->app_ver);
    configInfo.macAddress.assign((const char *)pm_ptr->mac_addr);
    configInfo.userComment.assign((const char *)pm_ptr->comment);

    configInfo.roiX = pm_ptr->roi_x;
    configInfo.roiY = pm_ptr->roi_y;
    configInfo.width = pm_ptr->roi_width;
    configInfo.height = pm_ptr->roi_height;
    configInfo.frameRate = pm_ptr->fps;
    configInfo.exposure = pm_ptr->shutter_speed;
    configInfo.frameCount = getFrameCount();        // This is the number of frames to save, not the number captured.

    if (IS_8_BIT_MODE(pm_ptr->bit_sel))
        configInfo.bitsPerPixel = 8;
    else if (IS_10_BIT_MODE(pm_ptr->bit_sel))
        configInfo.bitsPerPixel = 10;
    else
        configInfo.bitsPerPixel = 12;

    configInfo.sensorType = pm_ptr->sensor_type;
    switch ((pm_ptr->sensor_opts >> GIGE_SENSOPT_BIN_SHIFT) & GIGE_SENSOPT_BIN_MASK)
    {
    case 0:
    case 1:
    default:
        configInfo.sensorBinningOpt = FAS_INTERNAL::NO_BINNING;
        break;
    case 2:
        configInfo.sensorBinningOpt = FAS_INTERNAL::BINNING_2X;
        break;
    case 4:
        configInfo.sensorBinningOpt = FAS_INTERNAL::BINNING_4X;
        break;
    }
    switch ((pm_ptr->sensor_opts >> GIGE_SENSOPT_SUBS_SHIFT) & GIGE_SENSOPT_SUBS_MASK)
    {
    case 0:
    case 1:
    default:
        configInfo.sensorSubSampleOpt = FAS_INTERNAL::NO_SUBSAMPLING;
        break;
    case 2:
        configInfo.sensorSubSampleOpt = FAS_INTERNAL::SUBSAMPLE_2X;
        break;
    case 4:
        configInfo.sensorSubSampleOpt = FAS_INTERNAL::SUBSAMPLE_4X;
        break;
    }

    // Get the lens panel object.
    PanelLens *lens_panel;
    lens_panel = static_cast<PanelLens *>(MainWindow::GetInstance()->get_control_panel()->getPanel(CONTROL_PANEL::PANEL_LENS));
    if (lens_panel == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to get the lens panel object!";

        // We need to gather this information. Provide defaults.
        (void)configInfo.lensManufacturer;
        (void)configInfo.lensModel;
        configInfo.focalLength = 35;
        (void)configInfo.aperture;
    }
    else
    {
        configInfo.lensManufacturer.erase();    // The manufacture may be part of the name, but we don't separate that out.
        lens_panel->getLensName(configInfo.lensModel);
        configInfo.focalLength = lens_panel->getFocalLength();
        configInfo.aperture = (uint32_t)(lens_panel->getAperture() * 10);
        // QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lens:" << configInfo.lensModel.c_str() << "focalLength:" << configInfo.focalLength << "aperture:" << configInfo.aperture;
    }

    // The time in the pm_metadata is the time of the trigger frame, not the time of the
    // first frame of the sequence.
    {
        uint32_t startTimeSec;
        uint32_t startTimeMicrosec;

        double trigTime;
        double clipStartTime;

        struct tm tm_info;

        // Convert the trigger time to a double for easier manipulation.
        trigTime = ((double)pm_ptr->trig_time_s) + (((double)pm_ptr->trig_time_ns) / 1000000000);
        if (pm_ptr->trig_frame == pm_ptr->first_frame)
            clipStartTime = trigTime;
        else if (pm_ptr->trig_frame > clipStart)
            clipStartTime = trigTime - ((pm_ptr->trig_frame - clipStart) * (1.0 / ((double)pm_ptr->fps)));
        else
            clipStartTime = trigTime + ((clipStart - pm_ptr->trig_frame) * (1.0 / ((double)pm_ptr->fps)));

        // Convert the clipStartTime in seconds and microseconds.
        startTimeSec = (uint32_t)clipStartTime;
        startTimeMicrosec = (uint32_t)((clipStartTime - ((double)startTimeSec)) * 1000000.0);

        // Convert our SBS to a time structure.
        ClientMath::sbs_to_tm(startTimeSec, &tm_info);

        // Convert the time struct in to time since epoch.
        configInfo.timeStartOfStreamSeconds = mktime(&tm_info);
        configInfo.timeStartOfStreamMicrosec = startTimeMicrosec;
    }

    // Create an instance of our streamer.
    if (FiInternal::factory() != nullptr)
    {
        FiInternal::GET_HANDLE_FN funcPtr = FiInternal::factory()->getHandleFuncPtr();
        if (funcPtr != nullptr)
        {
            streamHandle = (*funcPtr)(destPath.toStdString(), configInfo);
        }
    }
    if (streamHandle == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fiCreatePackStreamer() failed.";
        return EFAULT;
    }

    // Make sure our closing state is reset.
    streamIsClosed = false;

    // Create the video stream writer.
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Creating Packed Writer for:" << destPath;

    // Return our result.
    return 0;
}

/*!
 * @brief This method prepares the video stream writer.
 *
 * NOTE:  Expects this object to be initialized with the
 * context necessary to prepare the stream writer.
 *
 * @return  0 on success, otherwise an error code.
 */
int FiSaver::prepareStreamWriter()
{
    int rc = EFAULT;                // Return code indicates error preparing stream writer.
    QFile Video_QFile;

    // If not saving to a stream video file cannot continue.
    if (!FiSaver::isStreamVideo(streamType))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Called with non-video streamType:" << streamType << " and imageType:" << imageInputType;
        goto prepare_bail;
    }

    // Reset our filename.
    imageFilename.clear();

    // Sanity check. Called from the GUI context, so protect access.
    if (streamWriter != nullptr)
    {
        // Trying to save another video but the first stream writer did not complete
        // its video stream.
        QLOG_WARN() << __PRETTY_FUNCTION__ << "streamWriter is not nullptr!";
        finishSave();
    }

    // Get a file name we can use. If it works, the result will be in dest_path.
    if ((rc = createFilePath()) != 0)
        goto prepare_bail;

    // Try opening the file.
    Video_QFile.setFileName(destPath);
    if (!Video_QFile.open(QIODevice::WriteOnly))
    {
        // Log the error.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QStringLiteral("Unable to create video container file %1 err: %2").arg(destPath, Video_QFile.errorString());
        rc = EACCES;
        goto prepare_bail;
    }

    // Set the file's permissions, if we can.
    Video_QFile.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // If created the video streamer make sure the playback rate is a valid value.
    if (streamPlaybackRate == 0)
    {
        // Log the error.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QStringLiteral("A playback rate of zero was unexpected while preparing stream writer to save: %1.").arg(destPath);
        rc = EINVAL;
        goto prepare_bail;
    }

    // Create the video stream writer.
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Creating Video Stream Writer for:" << destPath;

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::AVI:
        streamWriter = FAS_VIDEO::FiVideo::CreateAVIVideoStreamWriter(destPath.toStdString(), streamPlaybackRate, streamImageWidth, streamImageHeight, frameCount);
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
    {
        uint32_t h264KeyFrameInterval;

        // If the user wants to enable all keyframes that means we tell the
        // MP4 encoder to force one for every single frame.
        if (h264EnableAllKeyframes)
            h264KeyFrameInterval = 1;
        else
            h264KeyFrameInterval = FAS_VIDEO::DEFAULT_H264_KEYFRAME_INTERVAL;

        streamWriter = FAS_VIDEO::FiVideo::CreateMP4VideoStreamWriter(destPath.toStdString(), streamPlaybackRate,
                                                                      streamImageWidth, streamImageHeight, frameCount,
                                                                      h264Quality, h264Speed, h264KeyFrameInterval);
        break;
    }
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No support for video streamType: " << streamType;
        rc = EINVAL;
        break;
    }

    // If unable to create the video stream writer or its reusable image data object log an error.
    if (streamWriter == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Created empty streamWriter for: " << destPath;
        goto prepare_bail;
    }

    // Make sure our closing state is reset.
    streamIsClosed = false;

    // The video stream writer is prepared and ready to receive frames.
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Successfully created Video Stream Writer for:" << streamWriter->GetVideoName().c_str();
    rc = 0;

prepare_bail:
    // Return our result.
    return rc;
}

/*!
 * @brief Reset the settings in the saver object.
 */
void FiSaver::reset()
{
    absRetryFrame = 0;
    streamImageWidth = 0;
    streamImageHeight = 0;
    streamPlaybackRate = FI_SAVER::DEFAULT_PLAYBACK_RATE;
    convertRawToXml = false;
    savePfmd = false;
    imageFilename.clear();
    streamType = FI_SAVER::STREAM_TYPE::NO_STREAM;
    imageTypeForEncoder = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
    imageInputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
    retryInProgress = false;
    usingSubfolders = false;
}

/*!
 * @brief Called by the worker thread to save the image to either a file or the stream.
 *
 * Note:  Make sure the input format fiSaver::imageInputType of this image data matches
 * the format for encoder fiSaver::imageTypeForEncoder.
 *
 * This should have been setup by the stream feeder (the caller) when it set up this
 * object using the interface fiSaver::SetStreamType.
 *
 * @param imageDataPtr      The read-only buffer of image data to write to the stream.
 * @param imageDataLen      The number of bytes in the image data to write.
 * @param imageDataSeqNbr   The sequence number for this image in the stream.
 * @param do_retry      True if a retry is in progress.
 *
 * @return True on success, false otherwise.
 */
bool FiSaver::saveToStream(const uint8_t *imageDataPtr, uint32_t imageDataLen,
                           uint32_t imageDataSeqNbr, bool do_retry)
{
    FAS_VIDEO::STREAM_STATUS streamStatus = FAS_VIDEO::STREAM_STATUS::STREAM_READY;

    // Likewise if data resides in a xformedImageBuffer it has a different image
    // type than that contained in the image_ptr passed in the interface.  It has
    // been transformed.  But only for some encoders will this transformation need
    // to take place.  The initialization setting for this will be equal to the
    // image format expected in image_ptr.  It will be changed to imageTypeForEncoder
    // if a transformation has taken place.
    FI_SAVER::IMAGE_TYPE imageTypeForWriter = imageInputType;

    // Keep the status of this function in a variable.
    bool saveImageOK = false;   // Assume error condition.

    // If trouble arises when attaching the provided image data to the stream's
    // image package save the error condition.
    QString imageErrorStr;

    // Start the save image statistical gathering for duration.
    if (saveImageStats == nullptr)
        saveImageStats = std::unique_ptr<FiDurations>(new FiDurations());
    else
        saveImageStats->MarkStart();    // Start the Interval timer.

    // Sanity check.
    if (imageDataPtr == nullptr)
        goto exitSaveToStream;

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "image_ptr:" << Qt::showbase << Qt::hex << ((uint64_t)imageDataPtr) << "len:" << Qt::dec << imageDataLen;
    }

    // If saving a video stream through a stream writer the process is different than
    // when saving a stack of images or a capture blob.
    // Note:  This used to have "&& !metadata_seq_num(imageDataSeqNbr)" as part of the
    // the "if" condition however the imageDataSeqNbr is never negative (which it needs
    // to be for the condition to be true as metadata files are identified by -1, -2, ...
    // I am putting this note in place to identify an historical condition that is now
    // removed.
    if (FiSaver::isStreamVideo(streamType))
    {
        if (streamWriter)
        {
            // Note:  There had been a test here to see if the stream writer was in
            // a ready state.  This really causes issues here because for some encoders
            // like MP4 it is possible when in a retry condition the lost frames were
            // handled with duplication as it is not possible to "edit" the MP4 encoded
            // video in place like with AVI video.  Therefore we should only test this
            // condition for non retry cases - however in thinking about this we should
            // allow the process to continue no matter what and let the stream writer
            // set any type of error condition so that it will be processed properly.
            // It is decided we will log the condition only.
            if (!streamWriter->IsStreamReady())
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "!streamWriter->IsStreamReady().  do_retry:"
                            << do_retry << "IsStreamWriterComplete():"
                            << (streamWriter->IsStreamWriterComplete() ? "Video Completed" : "Not Complete");
            }

            // Does this stream write require a transformation of the image data?
            // If so the encoder is expecting a different image format than the
            // one provided by provided to this stream writer and will be transformed
            // using the QT image tools.
            if (imageInputType != imageTypeForEncoder)
            {
                // Mark an interval for collecting durations for image transforms.
                if (imageXlateStats == nullptr)
                    imageXlateStats = std::unique_ptr<FiDurations>(new FiDurations());

                imageByteArray.clear();   // Clear out any old allocations.
                fiPixmap myPixMap;

                if (VERBOSE_SAVE_LOG)
                {
                    QLOG_DEBUG() << "IMAGE TRANSFORM:  imageInputType[" << imageInputType << "]"
                                 << " != imageTypeForEncoder[" << imageTypeForEncoder << "]"
                                 << " imageDataSeqNbr[" << imageDataSeqNbr << "] "
                                 << " imageDataLen[" << imageDataLen << "] "
                                 << " imageDataPtr[" << imageDataPtr << "].";
                }

                if (imageXlateStats)
                    imageXlateStats->MarkStart(); // Mark start of interval.

                // Load the QT pixmap with data from passed in image_ptr.
                if (!myPixMap.loadFromData(imageDataPtr, imageDataLen))
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to fiPixmap::loadFromData the image for video stream writer.";
                    imageErrorStr = tr("[Unable To Load Frame Pixmap]");
                    streamImage.reset();     // Set null to let stream mark itself in error.
                    goto imageInvalidMarkStreamError;
                }
                // Transform the image data passed in interface into the type needed
                // for the video stream writer.
                if (imageTypeForEncoder == FI_SAVER::IMAGE_TYPE::JPEG)
                    myPixMap.set_jpeg_quality(jpegQuality_4QTPixmaps);
                myPixMap.save_image(&imageByteArray, imageTypeForEncoder);

                // Will be working with different image data for the rest
                // of the save because the encoder wanted a different format.
                imageDataPtr = (uint8_t *)imageByteArray.data();
                imageDataLen = (uint32_t) imageByteArray.length();
                imageTypeForWriter = imageTypeForEncoder;

                if (imageXlateStats)
                    imageXlateStats->MarkInterval(); // Mark end of interval.
            }

            // We use an FiImage shared pointer to supply frame data to the stream writer.
            // In this case the FSP_COMM object is managing an external allocated image data
            // pointer to the image data frame.  The FiImage object can attach this externally
            // managed image data frame pointer and it does not take ownership.  It is borrowing
            // the pointer and length for the time it takes to edit or add the image data frame.
            //
            // NOTE_JLC: Eventually we want to be passing the image data in a shared pointer from the
            // camera server stream as an FiImage instead of attaching the externally managed image
            // pointer.  For now it is being attached and provides the image frame data for the AVI
            // streamer adds and edits.  I am leaving this NOTE_JLC in place to keep in mind this is
            // an architectural change for the future system.
            // The specified GIGE image stream format determines type of images we are expecting.
            // #MP4IMAGEINPUT - fiSaver::saver_image - building fiImage object for imageTypeForWriter.
            switch (imageTypeForWriter)
            {
            case FI_SAVER::IMAGE_TYPE::BMP:
                if (!streamImage)
                {
                    if (VERBOSE_SAVE_LOG)
                    {
                        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Allocating FI_BMP FiImage object imageDataLen[" << imageDataLen << "].";
                    }
                    streamImage = std::shared_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMP());
                }
                break;
            case FI_SAVER::IMAGE_TYPE::BMP_PLANAR:
                if (!streamImage)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Allocating FI_BMP_PLANER FiImage object imageDataLen[" << imageDataLen << "].";
                    streamImage = std::shared_ptr<FAS_VIDEO::FiImage>(new FI_BMP::FiBMPPlanar());
                }
                break;
            case FI_SAVER::IMAGE_TYPE::JPEG:
            default:
                if (!streamImage)
                {
                    if (VERBOSE_SAVE_LOG)
                    {
                        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Allocating FI_JPEG FiImage object imageDataLen[" << imageDataLen << "].";
                    }
                    streamImage = std::shared_ptr<FAS_VIDEO::FiImage>(new FI_JPEG::FiJPEG());
                }
                break;
            }

            // Sanity check:  Verify that the image object exists.
            if (streamImage)
            {
                // Note:  This image package wrapper was created to provide a common
                // object we could use to pass around different types of image data
                // whether it be externally managed by the system or internally
                // managed by the fas_video library itself in a std::vector form.
                //
                // Create an image package wrapping the image pointer and length.
                // In this case an external source owns the memory for the image and we are using that factory creator.
                std::unique_ptr<FAS_VIDEO::FiImagePkg> imagePkg = FiImagePkg::Create(imageDataPtr, imageDataLen);

                // Attach the created image package into the image object.  It now takes
                // ownership of that unique pointer for the life of streamImage or until
                // another AttachImagePkg is performed on that streamImage.  If the streamImage
                // destructs at the end of this method all is cleaned up because of the use
                // of the unique pointers in this system.
                streamImage->AttachImagePkg(std::move(imagePkg));
                if (!streamImage->IsImageAttached())
                {
                    imageErrorStr = tr("[Expected Frame Format: %1 but Frame Received is not in that Format.").arg(FiSaver::imageTypeToQString(imageTypeForWriter));
                    streamImage.reset();     // Set null to let stream mark itself in error.
                    goto imageInvalidMarkStreamError;
                }
            }
            else
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "streamImage nullptr frame for stream writer. seq_num:" << imageDataSeqNbr << lastErrMsg;
                goto exitSaveToStream;
            }
        }

// This label was created to allow for the image processing error to percolate through the
// streamer with a nullptr for streamImage so that the error handling will be delivered
// to the actual progress dialog as a user interface message so they know why the stream
// failed.
imageInvalidMarkStreamError:

        // Are we retrying?
        if (do_retry)
        {
            int rel_frame;

            // Get the relative frame number, so the library can figure out where this frame goes in the container.
            rel_frame = (int)(absRetryFrame - clipStart);

            if (RETRY_DEBUG_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rel_frame:" << rel_frame
                             << "abs_retry_frame:" << absRetryFrame
                             << "clip_start:" << clipStart
                             << "image_len:" << imageDataLen << "(ok for BMP to be 56 bytes too long).";
            }
            // Edit the image data frame into the AVI stream writer at the retry frame location.
            if (streamWriter)
            {
                streamStatus = streamWriter->EditVideoFrame((uint32_t) rel_frame, streamImage);
            }
        }
        else
        {
            if (streamWriter)
            {
                streamStatus = streamWriter->AddVideoFrame(streamImage);
            }
        }

        // Check our stream status.
        if ((streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY) &&
            (streamStatus != FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED))
        {
            // Build up an error message.
            std::string logErrorMessage;
            streamWriter->LogStreamStatus(logErrorMessage, true);
            lastErrMsg = tr("%2[Frame: %1]").arg(imageDataSeqNbr).arg(QString::fromUtf8(logErrorMessage.c_str()));
            if (imageErrorStr.size() > 0)
            {
                // Add the image error that occurred to the last error message.
                lastErrMsg += tr("%1").arg(imageErrorStr);
            }
            QLOG_ERROR() << __PRETTY_FUNCTION__ << lastErrMsg;
            goto exitSaveToStream;
        }
    }
    // Note:  The &&!metadata_seq_num method is critical here in order to properly fall
    // to the else statement below to properly process the metadata files associated with
    // a copy to pc operation.
    else if (FiSaver::isCaptureStream(streamType) && !metadataSeqNum(imageDataSeqNbr))
    {
        qint64  write_offset;
        qint64  write_status;

        // Are we retrying?
        if (do_retry)
        {
            // Calculate our offet for the retry.
            write_offset = (qint64)absRetryFrame * (qint64)blobSize;

            if (RETRY_DEBUG_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "abs_retry_frame:" << absRetryFrame
                             << "seq_num:" << imageDataSeqNbr
                             << "image_len:" << imageDataLen
                             << "write_offset:" << write_offset;
            }
        }
        else
        {
            // Note:  If this blob is not a true capture blob but a video
            // in disguise (as in case of video copies) - the blob_size may
            // calculation below should always end up with a write_offset
            // of zero because typically videos are smaller than one blob_size.
            // however should the video copy be larger than a blob_size this
            // logic should fall out properly.

            // Setup our position in the file appropriately.
            write_offset = (qint64)(imageDataSeqNbr - 1) * (qint64)blobSize;
            if (RETRY_DEBUG_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_num:" << imageDataSeqNbr
                             << "clip_start:" << clipStart
                             << "image_len:" << imageDataLen
                             << "write_offset:" << write_offset
                             << "blob_size:" << blobSize
                             << "blob_file:" <<  Qt::showbase << Qt::hex << ((uint64_t)blobFile);

            }
        }

        // Are we saving an CAP file?
        if (blobFile != nullptr)
        {
            // Must be saving a CAP. On series3 cameras, we SHOULD never get here.
            // Seek the file to the appropriate location.
            blobFile->seek(write_offset);

            // Write out the data.
            write_status = blobFile->write((const char *)imageDataPtr, imageDataLen);

            // Check on how we did.
            if (write_status == -1)
            {
                // We had a problem opening the file.
                lastErrMsg = tr("Error: %2").arg(blobFile->errorString());
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "CAP error:" << lastErrMsg;
                goto exitSaveToStream;
            }
        }
    }
    else
    {
        // Note:  Metadata files that are part of a copy to pc operation will end up
        // falling through to this else block.  This is how the metadata file is created
        // and named properly (see build_file_name method.

        // This section of the saveToStream is saving a stack of images so this
        // is just a file in that stack.  It should simply be saved to disk with
        // the proper filename in its stack stream.
        bool    file_status;
        QFile   image_file;
        qint64  write_status;

        // Get the full path to our file. Convert the sequence number from 1-based to 0-based.
        buildFileName(imageFilename, imageDataSeqNbr - 1, do_retry);
        image_file.setFileName(imageFilename);

        //QLOG_INFO() << __PRETTY_FUNCTION__ << "Image Save Stack file:" << image_filename;

        file_status = image_file.open(QIODevice::WriteOnly);
        if (!file_status)
        {
            // We had a problem opening the file.
            lastErrMsg = image_file.errorString();
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Image save error on open:" << lastErrMsg << "file:" << imageFilename;
            goto exitSaveToStream;
        }

        // Set the permissions on our image file.
        image_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

        // Write out the data.
        write_status = image_file.write((const char *)imageDataPtr, imageDataLen);

        // Check on how we did.
        if (write_status == -1)
        {
            // We had a problem opening the file. Grab the error information here, but
            // wait until after we close the file to report it.
            lastErrMsg = tr("Error: %1").arg(image_file.errorString());
        }

        // All done.
        image_file.close();

        // Check on how we did.
        if (write_status == -1)
        {
            // Let folks know there was an error.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Image write error:" << lastErrMsg;
            goto exitSaveToStream;
        }
#if 0
// #SAVE_ERROR - FORCE ONE FOR STACKS HERE
        // Put whatever data sequence number you are testing with.
        //if (imageDataSeqNbr == 288)
        {
            last_err_msg = tr("I am FORCING an Error for Frame:").arg(imageDataSeqNbr);
            goto exitSaveToStream;
        }
#endif
    }

    // Persist this data as it is at this point in time we know we wrote this file to its destination without error.
    lastFilenameWritten = imageFilename;

    // Persist the last image sequence number successfully saved.
    lastSeqNbrSaverSaved = imageDataSeqNbr;

    // Accumulate a total number of image buffers (they can be frames or blobs) that
    // have come through here and been processed.
    streamSaveBufferCount++;

    // If make it to this point in the execution the image was saved without error.
    saveImageOK = true;

exitSaveToStream:

    // Mark the ending interval time.
    if (saveImageStats)
        saveImageStats->MarkInterval();

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "saveImageOK:" << saveImageOK << "seqNbr:" << imageDataSeqNbr;
    }

    // Was there an error during this request?  Persist the state.
    lastSaveHadError = !saveImageOK;

    // What was the last image data sequence number received.
    lastSeqNbrSaverReceived = imageDataSeqNbr;

    // All done.
    return saveImageOK;
}

/*!
 * @brief Called by the worker thread to save the image to either a file or the stream.
 *
 * Note:  Make sure the input format fiSaver::imageInputType of this image data matches
 * the format for encoder fiSaver::imageTypeForEncoder.
 *
 * This should have been setup by the stream feeder (the caller) when it set up this
 * object using the interface fiSaver::SetStreamType.
 *
 * @param imageDataPtr      The read-only buffer of image data to write to the stream.
 * @param imageDataLen      The number of bytes in the image data to write.
 * @param imageDataSeqNbr   The sequence number for this image in the stream.
 * @param retryInProgress   True if a retry is in progress.
 *
 * @return True on success, false otherwise.
 */
bool FiSaver::saveToPackedStream(const uint8_t *imageDataPtr, uint32_t imageDataLen,
                                 uint32_t imageDataSeqNbr, bool retryInProgress)
{
    bool saveOK(false);
    FAS_INTERNAL::FIP_STATUS streamStatus = FAS_INTERNAL::FIP_STATUS::STREAM_TYPE_ERROR;

    // Sanity check.
    if (imageDataPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "imageDataPtr is nullptr!";
        goto save_bail;
    }
    if (streamHandle == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "streamHandle is 0!";
        goto save_bail;
    }
    if (FiInternal::factory() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Packed stream not supported!";
        goto save_bail;
    }

    // Start the save image statistical gathering for duration.
    if (saveImageStats == nullptr)
        saveImageStats = std::unique_ptr<FiDurations>(new FiDurations());
    else
        saveImageStats->MarkStart();    // Start the Interval timer.

    // Is this a retry frame?
    if (retryInProgress)
    {
        int rel_frame;

        // Get the relative frame number, so the library can figure out where this frame goes in the container.
        rel_frame = (int)(absRetryFrame - clipStart);

        if (RETRY_DEBUG_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rel_frame:" << rel_frame
                         << "abs_retry_frame:" << absRetryFrame
                         << "clip_start:" << clipStart
                         << "image_len:" << imageDataLen;
        }

        FiInternal::REPLACE_FRAME_FN funcPtr = FiInternal::factory()->getReplaceFrameFuncPtr();
        if (funcPtr != nullptr)
        {
            // Replace the placeholder frame with the data frame.
            streamStatus = (*funcPtr)(streamHandle, (uint32_t)rel_frame, imageDataPtr);
        }
    }
    else if (imageInputType == FI_SAVER::IMAGE_TYPE::TIF_RAW)
    {
        FiInternal::ADD_TIFF_IMAGE_DATA_FN funcPtr = FiInternal::factory()->getAddTiffFuncPtr();
        if (funcPtr != nullptr)
        {
            // Add the image to the streamer.
            streamStatus = (*funcPtr)(streamHandle, imageDataPtr);
        }
    }
    else if (imageInputType == FI_SAVER::IMAGE_TYPE::PACKED)
    {
        FiInternal::ADD_PACKED_DATA_FN funcPtr = FiInternal::factory()->getAddPackedFuncPtr();
        if (funcPtr != nullptr)
        {
            // Add the image to the streamer.
            streamStatus = (*funcPtr)(streamHandle, imageDataPtr);
        }
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "imageInputType:" << imageInputType << "We only support raw TIFFs";
        goto save_bail;
    }


    // Mark the ending interval time.
    if (saveImageStats)
        saveImageStats->MarkInterval();

    // Check our stream status.
    if ((streamStatus != FAS_INTERNAL::FIP_STATUS::STREAM_READY) &&
        (streamStatus != FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error updating pack streamer, streamStatus:" << streamStatus;
        goto save_bail;
    }

    // Persist the last image sequence number successfully saved.
    lastSeqNbrSaverSaved = imageDataSeqNbr;

    // Accumulate a total number of image buffers (they can be frames or blobs) that
    // have come through here and been processed.
    streamSaveBufferCount++;

    // The save went OK.
    saveOK = true;

save_bail:
    // Was there an error during this request?  Persist the state.
    lastSaveHadError = !saveOK;

    // Return our status.
    return saveOK;
}

/*!
 * @brief Save the TXT and CFG metadata files.
 *
 * @param metadata_ptr      The metadata object to be saved.
 * @return True on success, false otherwise.
 */
bool FiSaver::savePartitionMetadata(QByteArray *metadata_ptr, uint32_t red_gain, uint32_t green_gain, uint32_t blue_gain)
{
    QString meta_text;
    QString meta_filename;
    QFile   meta_file;

    // Gain information for metadata.
    bool            do_gain_info;
    FAS_VIDEO::VIDEO_GAINS  videoGains;
    const FAS_VIDEO::VIDEO_GAINS *videoGainsReadOnlyPtr = &videoGains;

    // The xcite configuration record using our camera metadata in bytes.
    std::unique_ptr<FAS_VIDEO::X_CFG_METADATA_VECTOR> xcfgMetadataInBytes;

    // Sanity check.
    if ((metadata_ptr == nullptr) || (metadata_ptr->length() != sizeof(pm_metadata_type)))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Bad metadata!";
        return false;
    }

    //QLOG_INFO() << __PRETTY_FUNCTION__ << "[METADATA CFG][" << save_metadata_cfg_file << "]";

    // Initialize our gain structure. Not using the parameters if they're all defaults.
    if ((red_gain == (uint32_t) -1) && (green_gain == (uint32_t) -1) && (blue_gain == (uint32_t) -1))
    {
        // Don't use the gain info structure.
        do_gain_info = false;
    }
    else
    {
        // Use the gain information structure.
        do_gain_info = true;
        videoGains.red = red_gain;
        videoGains.green = green_gain;
        videoGains.blue = blue_gain;
    }

    // Build up the basic pathway to metafile name first with the txt file even though
    // we may not be saving it we need to get ready to save it.
    meta_filename = destPath;
    if (streamType == FI_SAVER::STREAM_TYPE::AVI)
    {
        // Replace the ".avi" with ".txt".
        meta_filename.replace(meta_filename.lastIndexOf(IMAGE_EXT_AVI), IMAGE_EXT_AVI.length(), IMAGE_EXT_META_TEXT);
    }
    else if (streamType == FI_SAVER::STREAM_TYPE::MP4)
    {
        // Replace the ".mp4" with ".txt".
        meta_filename.replace(meta_filename.lastIndexOf(IMAGE_EXT_MP4), IMAGE_EXT_MP4.length(), IMAGE_EXT_META_TEXT);
    }
    else if (streamType == FI_SAVER::STREAM_TYPE::FI_CAP)
    {
        // Replace the ".cap" with ".txt".
        meta_filename.replace(meta_filename.lastIndexOf(IMAGE_EXT_CAPTURE), IMAGE_EXT_CAPTURE.length(), IMAGE_EXT_META_TEXT);
    }
    else
    {
        // Just using a standard name.
        meta_filename = destPath + DEFAULT_METADATA_FILE_NAME;
    }

    // Convert the metadata to a text format.
    meta_text = FiSaver::metadataToText(metadata_ptr, streamType, imageInputType,
                                        clipStart, (clipStart + frameCount) - 1,
                                        (do_gain_info ? videoGainsReadOnlyPtr : nullptr));
    if (meta_text.isNull())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to convert metadata";
        return false;
    }

    // Make sure the file doesn't already exist.
    QFile::remove(meta_filename);
    meta_file.setFileName(meta_filename);

    // Open the file for writing.
    if (!meta_file.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << meta_filename << "Err:" << meta_file.errorString();
        return false;
    }
    meta_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
    meta_file.write(meta_text.toUtf8().constData());
    meta_file.close();

    if (saveMetadataCfgFile)
    {
        // Create the xcite formatted configuration metadata.
        int meta_len = FAS_VIDEO::FiVideo::CreateX_CFG_MetadataString(xcfgMetadataInBytes,
                                                                      (void *)metadata_ptr->constData(), (uint32_t)metadata_ptr->length(),
                                                                      clipStart, clipStart + frameCount);
        if (meta_len <= 0)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create the xcite metadata record!" << meta_len;
            return false;
        }

        // Build up our filename. Replace the ".txt" with ".cfg".
        meta_filename.replace(meta_filename.lastIndexOf(IMAGE_EXT_META_TEXT), IMAGE_EXT_META_TEXT.length(), IMAGE_EXT_META_CFG);

        // Make sure the file doesn't already exist before we write the data out.
        QFile::remove(meta_filename);
        meta_file.setFileName(meta_filename);
        if (!meta_file.open(QIODevice::WriteOnly))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << meta_filename << "Err:" << meta_file.errorString();
            return false;
        }
        meta_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
        meta_file.write((const char *)xcfgMetadataInBytes->data(), meta_len);
        meta_file.close();
    }

    // All is well.
    return true;
}

/*!
 * @brief Save the pm_metadata file and start the conversion
 *        to XML if desired.
 *
 * @param raw_metadata_ptr      The metadata object to be saved.
 * @param pm_ptr                Pointer to the pm_metadata information. nullptr for no conversion.
 *
 * @return True on success, false otherwise.
 */
bool FiSaver::saveRawMetadata(const QByteArray *raw_metadata_ptr, const struct pm_metadata_type *pm_ptr, uint32_t red_gain, uint32_t green_gain, uint32_t blue_gain)
{
    QString raw_meta_filename;
    QString xml_meta_filename;
    QFile   meta_file;

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__;
    }

    // Build up the pm_metadata file name.
    raw_meta_filename = destPath;
    if (FiSaver::isStreamVideo(streamType))
    {
        // Make sure the file extension information matches this streamType.
        setFileExtension(streamType);

        // Replace this stream's file extension with the actual stream extension.
        QString streamExtForFile = "." + getFileExtension();

        raw_meta_filename.replace(raw_meta_filename.lastIndexOf(streamExtForFile), 4, QString::fromUtf8(".pm"));
    }
    else
    {
        // Just using a standard name.
        raw_meta_filename = destPath + QString::fromUtf8("metadata.pm");
    }

    // Now that we have the "raw" file name, let's get the XML file name;
    xml_meta_filename = raw_meta_filename;
    xml_meta_filename.replace(xml_meta_filename.lastIndexOf(QString::fromUtf8(".pm")), 3, QString::fromUtf8(".xml"));

    // Make sure the file doesn't already exist.
    QFile::remove(raw_meta_filename);

    // Only create the PM file if streamType is supported.
    if (savePfmd)
    {
        meta_file.setFileName(raw_meta_filename);

        // Open the file for writing.
        if (!meta_file.open(QIODevice::WriteOnly))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << raw_meta_filename << "Err:" << meta_file.errorString();
            return false;
        }
        meta_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
        meta_file.write(*raw_metadata_ptr);
        meta_file.close();
    }

    // Do we want to convert to XML?
    if ((pm_ptr != nullptr) && convertRawToXml)
    {
        QThread     *thread_ptr;

        // Make sure we're not trying to overlap conversion threads.
        if (rawToXml != nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Overlap conversions!";
        }

        // Allocate our thread objects.
        thread_ptr = new QThread;
        rawToXml = new RawToXML();

        // Put the convert object into the thread.
        rawToXml->moveToThread(thread_ptr);

        // Setup our signal/slot connections.
        connect(thread_ptr, SIGNAL(started()), rawToXml, SLOT(start_convert()), Qt::UniqueConnection);
        connect(rawToXml, SIGNAL(finished()), thread_ptr, SLOT(quit()), Qt::UniqueConnection);
        connect(rawToXml, SIGNAL(finished()), this, SLOT(cleanUpConversion()), Qt::UniqueConnection);
        connect(thread_ptr, SIGNAL(finished()), thread_ptr, SLOT(deleteLater()), Qt::UniqueConnection);

        // Pass in the parameters.
        rawToXml->set_params(raw_meta_filename, xml_meta_filename, pm_ptr, clipStart, (clipStart + frameCount) - 1);
        if ((red_gain != (uint32_t) -1) || (green_gain != (uint32_t) -1) || (blue_gain != (uint32_t) -1))
            rawToXml->set_gains(red_gain, green_gain, blue_gain);

        // Start the conversion process.
        thread_ptr->start();
    }

    // All is well.
    return true;
}

/*!
 * @brief Common logic to close down any open FiStream objects.
 *
 * @returns A true boolean value if everything closed down ok.
 */
bool FiSaver::closeStreamWriter()
{
    bool closedOk = true;

    // This stream writer is now closed down.  Set the status as such
    // so that all know it is no longer in use.
    streamIsClosed = true;

    // Report the duration of time taken for image transforms for stream writer.
    if (imageXlateStats)
    {
        if (VERBOSE_SAVE_LOG)
        {
            FiDurations::LogDurationSeparator();
            std::string durationInfoStr = "[fiSaver Image Transforms]";
            imageXlateStats->LogDurationInfo(durationInfoStr.c_str());
            FiDurations::LogDurationSeparator();
        }
        imageXlateStats.reset();
    }

    // Report the duration of time taken for saveToStream() function.
    if (saveImageStats)
    {
        if (VERBOSE_SAVE_LOG)
        {
            FiDurations::LogDurationSeparator();
            std::string durationInfoStr = "[fiSaver::saveToStream() Total Time]";
            saveImageStats->LogDurationInfo(durationInfoStr.c_str());
            FiDurations::LogDurationSeparator();
        }
        saveImageStats.reset();
    }

    if (streamWriter)
    {
        FAS_VIDEO::STREAM_STATUS closeStatus = streamWriter->CloseStream();
        if (closeStatus != FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED)
        {
            std::string streamStatusString;
            streamWriter->LogStreamStatus(streamStatusString, false);

            QLOG_WARN() << __PRETTY_FUNCTION__ << "Stream writer did not produce a complete video:" << streamStatusString.c_str();
            closedOk = false;
        }
        streamWriter.reset();    // Reset calls destructor and sets the unique pointer empty.
    }
    else if (streamHandle)
    {
        if (FiInternal::factory() != nullptr)
        {
            FiInternal::CLOSE_PACK_STREAMER_FN funcPtr = FiInternal::factory()->getClosePackFuncPtr();
            if (funcPtr != nullptr)
            {
                FAS_INTERNAL::FIP_STATUS closeStatus = (*funcPtr)(streamHandle);
                if (closeStatus != FAS_INTERNAL::FIP_STATUS::STREAM_CLOSED)
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "Packed writer did not produce a complete video";
                    closedOk = false;
                }
            }
        }

        // Reset the handle.
        streamHandle = 0;
    }

    // Free the reusable streamImage.
    if (streamImage)
    {
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Closing streamImage.  Usage count=" << streamImage.use_count();
        }
        streamImage.reset();
    }

    if (!closedOk)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << tr("Recommended review of fas_video* log to look for any video creation issues.");
    }

    // Return the status of this close_streams method.
    return closedOk;
}

QString FiSaver::getComment()
{
    QString ret_str;

    // Make sure we have a comment.
    if (commentBuf)
    {
        ret_str = QString(*commentBuf);
        if (ret_str.at(ret_str.size() - 1) == '\0')
            ret_str.chop(1);
    }

    return ret_str;
}

/*!
 * @brief A utility function to associate a file extension with a save format.
 * @param format    The format we'll be using for the save operation.
 */
void FiSaver::setFileExtension(FI_SAVER::STREAM_TYPE streamType)
{
    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::AVI:
        fileExtension = IMAGE_EXT_AVI;
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
        fileExtension = IMAGE_EXT_MP4;
        break;
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        fileExtension = IMAGE_EXT_BMP;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        fileExtension = IMAGE_EXT_CAPTURE;
        break;
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        fileExtension = IMAGE_EXT_DNG;
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        fileExtension = IMAGE_EXT_JPG;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
        fileExtension = IMAGE_EXT_TIF;
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        fileExtension = IMAGE_EXT_RDI_PACKED;
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected format:" << Qt::showbase << Qt::hex << streamType;
        break;
    }
}

/*!
 * @brief Sets the streaming format to use for saving.
 * It will also set the image format expected for the stream.
 *
 * @param saveStreamType        The streaming format we'll be using for the save operation.
 * @param saveImageInputType    The image type expected for input.
 * @param saveImageEncoderType  The image type expected for output.
 *
 * @return A boolean true if the format is valid and was set, false if not.
 */
bool FiSaver::setStreamType(FI_SAVER::STREAM_TYPE saveStreamType,
                            FI_SAVER::IMAGE_TYPE saveImageInputType,
                            FI_SAVER::IMAGE_TYPE saveImageEncoderType)
{
    bool validStreamType = true;
    streamType = saveStreamType;    // Assume this is a valid stream type.

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "saveStreamType:" << FiSaver::streamTypeToQString(saveStreamType) << " saveImageInputType:" << FiSaver::imageTypeToQString(saveImageInputType) << "saveImageEncoderType:" << FiSaver::imageTypeToQString(saveImageEncoderType);
    }

    switch (saveStreamType)
    {
    case FI_SAVER::STREAM_TYPE::AVI:
        imageInputType = saveImageInputType;
        imageTypeForEncoder = saveImageEncoderType;
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
        // #MP4IMAGEINPUT - fiSaver::SetStreamType - building imageInputType and imageTypeForEncoder.
        imageInputType = saveImageInputType;
        //imageTypeForEncoder = FI_SAVER::IMAGE_TYPE::BMP;    // Force BMP as that is all MP4 encoder knows (right now).
        imageTypeForEncoder = saveImageEncoderType;
        break;
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        imageTypeForEncoder = imageInputType = FI_SAVER::IMAGE_TYPE::BMP;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        imageTypeForEncoder = imageInputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
        break;
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        imageTypeForEncoder = imageInputType = FI_SAVER::IMAGE_TYPE::DNG;
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        imageTypeForEncoder = imageInputType = FI_SAVER::IMAGE_TYPE::JPEG;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        imageTypeForEncoder = imageInputType = FI_SAVER::IMAGE_TYPE::TIF;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
        imageTypeForEncoder = imageInputType = FI_SAVER::IMAGE_TYPE::TIF_RAW;
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        imageInputType = saveImageInputType;
        imageTypeForEncoder = saveImageEncoderType;
        break;
    default:
        validStreamType = false;
        streamType = FI_SAVER::STREAM_TYPE::NO_STREAM;
        imageTypeForEncoder = imageInputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected format:" << Qt::showbase << Qt::hex << streamType;
        break;
    }
    return validStreamType;
}

/*!
 * @brief Sets the H.264 option for frame quality in stream.
 *
 * @param h264QualityToSet  The H.264 CRF (Constant Rate Factor) option for encoder.
 */
void FiSaver::setH264Quality(uint32_t h264QualityToSet)
{
    // Make sure the h.264 quality setting is not greater than 100%.
    if (h264QualityToSet > 100)
        h264QualityToSet = 100;

    // BUGFIX #260:  This had mistakenly been DEFAULT_H264_QUALITY_PCT
    // instead of the MINIMUM value making the lowest stored setting
    // for the saver as 70.
    // Make sure the h.264 quality setting is not less than the minimum we allow.
    if (h264QualityToSet < FAS_VIDEO::MINIMUM_H264_QUALITY_PCT)
        h264QualityToSet = FAS_VIDEO::MINIMUM_H264_QUALITY_PCT;

    // Set the quality value.
    h264Quality = h264QualityToSet;
}

/*!
 * @brief Sets the H.264 setting to force keyframes every frame.
 *
 * @param h264EnableAllKeyframes  True if enable this setting, false if not.
 */
void FiSaver::setH264EnableAllKeyframes(bool h264EnableAllKeyframesToSet)
{
    h264EnableAllKeyframes = h264EnableAllKeyframesToSet;
}

/*!
 * @brief Sets the H.264 option for compression speed in stream.
 *
 * @param h264QualityToSet  The H.264 preset option for encoder.
 */
void FiSaver::setH264Speed(FAS_VIDEO::H264_SPEEDS h264SpeedToSet)
{
    switch (h264SpeedToSet)
    {
    case FAS_VIDEO::H264_SPEEDS::H264_SLOWER:
        h264Speed = h264SpeedToSet;
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_MEDIUM:
        h264Speed = h264SpeedToSet;
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_FAST:
        h264Speed = h264SpeedToSet;
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_VERY_FAST:
        h264Speed = h264SpeedToSet;
        break;
    default:
        // Put it to the default because user passed invalid setting.
        h264Speed = FAS_VIDEO::DEFAULT_H264_SPEED;
        break;
    }
}

/*!
 * @brief Set the flag to allow partition metadata CFG file saves.
 * @param save_metadata_cfg Flag is true if the user wants these saved, false if not.
 */
void FiSaver::setMetadataCfgSave(bool save_metadata_cfg)
{
    //QLOG_INFO() << __PRETTY_FUNCTION__ << "[METADATA CFG][" << save_metadata_cfg << "]";
    saveMetadataCfgFile = save_metadata_cfg;
}

/*!
 * @brief Set the base path to the "save_path" parameter. Make sure it is terminated appropriately.
 * @param save_path     The path we want to use as our "root" or "base" folder.
 */
void FiSaver::setPath(const QString& save_path)
{
    // Reset the base path.
    basePath.clear();

    // Sanity check.
    if (save_path.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Empty save path!";
    }

    // Make sure the path is valid for our platform.
    basePath = QDir::fromNativeSeparators(save_path);

    // Make sure it ends with a delimiter.
    if (!basePath.endsWith(QLatin1Char('/')))
        basePath.append(QLatin1Char('/'));
}

/*!
 * @brief Return information that can be used for status updates on video stream
 * save operation by filling in the data for each reference variables passed to the function.
 *
 * @param strMsg [output] provides a place to store additional inforamation for caller about stream state.
 * @param frameCount [output] holds the understood framecount for this stream.
 * @param framesAdded [output] how many frames have been processed by the stream.
 * @param framesEncoded[output] how many frames have been encoded (if video) if not it should equal framesAdded.
 * @param streamIsOk [output] what is the state of this stream ok or error condition.
 * @param streamIsComplete[output] is the state of this stream completed?
 *
 * @returns true if the stream is a video, false if not which means a stack of frames.
 */
bool FiSaver::getVideoStreamSaveStatus(QString& strMsg, uint32_t& frameCount, uint32_t& framesAdded,
                                       uint32_t& framesEncoded, bool& streamIsOk, bool& streamIsComplete)
{
    framesAdded = 0;
    framesEncoded = 0;
    bool isVideoStream      = false;

    streamIsOk = false;         // Assume error condition at beginning.
    streamIsComplete = false;   // Assume the stream is not complete.

    // If our stream writer is closed down cannot query it.
    if (streamIsClosed)
    {
        return isVideoStream;
    }

    // If we are creating a stream video special processing.
    if (FiSaver::isStreamVideo(streamType))
    {
        // Make sure that a stream writer exists.
        if (streamWriter != nullptr)
        {
            isVideoStream = true;   // This is a video stream writer.

            // Go ahead and pull out the information from the stream writer.
            frameCount = streamWriter->GetFrameCount();
            framesAdded = streamWriter->GetLastFrameAdded();
            framesEncoded = streamWriter->GetFramesEncodedCount();

            // Pull the current stream status.
            FAS_VIDEO::STREAM_STATUS streamStatus = streamWriter->IsStreamWriterComplete();

            // Note:  In the MP4 case FAS_VIDEO::STREAM_STATUS::PLACEHOLDER_WAIT
            // is NEVER set it is only set when creating an AVI video if the video
            // is waiting on placeholder frames.  For MP4 video it will either be
            // in a "ready" or "completed" or error state.  The idea that a stream
            // is completed needs to be returned to the caller querying the state
            // of the video streamer.  It should be added to this interface.

            // If ready or in a completed state still get the progress for caller.
            if ((streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_READY) ||
                (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED) ||
                (streamStatus == FAS_VIDEO::STREAM_STATUS::PLACEHOLDER_WAIT))
            {
                // If this is an AVI file the framesEncoded should = framesAdded.
                strMsg = QString::fromUtf8("Writing encoded frame [%1] of [%2] to video.").arg(framesEncoded).arg(frameCount);
                //infoStr = strMsg.toStdString(); // This is how to convert QString to std::string.
                streamIsOk = true;

                // Are we completed?
                if (streamStatus == FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED)
                {
                    streamIsComplete = true;
                }
            }
            else
            {
                // There is something wrong with the stream writer.
                std::string logErrorMessage;
                streamWriter->LogStreamStatus(logErrorMessage, false);
                lastErrMsg = tr("DATA: framesAdded[%1] framesEncoded[%2] frameCount[%3] %4").arg(framesAdded).arg(framesEncoded).arg(frameCount).arg(QString::fromUtf8(logErrorMessage.c_str()));
                //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Stream writer error when polling. " <<  last_err_msg;
                strMsg = lastErrMsg;
                streamIsOk = false;
            }
        }
        else if ((streamHandle != 0) && (FiInternal::factory() != nullptr))
        {
            FAS_INTERNAL::FIP_STATUS packStatus = FAS_INTERNAL::NO_IO_STREAM;
            isVideoStream = true;   // This is a video stream writer.

            FiInternal::GET_PACK_STATISTICS_FN statsFuncPtr = FiInternal::factory()->getGetStatisticsFuncPtr();
            if (statsFuncPtr != nullptr)
            {
                // Go ahead and pull out the information from the stream writer.
                (*statsFuncPtr)(streamHandle, frameCount, framesAdded, framesEncoded);
            }

            FiInternal::IS_PACK_STREAMER_FINISHED_FN finishedFuncPtr = FiInternal::factory()->geIsFinishedFuncPtr();
            if (finishedFuncPtr != nullptr)
            {
                // Pull the current stream status.
                packStatus = (*finishedFuncPtr)(streamHandle);
            }

            // If ready or in a completed state still get the progress for caller.
            if ((packStatus == FAS_INTERNAL::FIP_STATUS::STREAM_READY) ||
                (packStatus == FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED) ||
                (packStatus == FAS_INTERNAL::FIP_STATUS::PLACEHOLDER_WAIT))
            {
                strMsg = QString::fromUtf8("Writing encoded frame [%1] of [%2] to video.").arg(framesEncoded).arg(frameCount);
                streamIsOk = true;

                // Are we completed?
                if (packStatus == FAS_INTERNAL::FIP_STATUS::STREAM_COMPLETED)
                {
                    streamIsComplete = true;
                }
            }
            else
            {
                // There is something wrong with the pack writer.
                lastErrMsg = tr("An error occurred when saving the packed stream: framesAdded[%1] framesEncoded[%2] frameCount[%3]").arg(framesAdded).arg(framesEncoded).arg(frameCount);
                strMsg = lastErrMsg;
                streamIsOk = false;
            }
        }
    }
    else
    {
        framesAdded = framesEncoded = streamSaveBufferCount;
        streamIsOk = !lastSaveHadError;  // For stacks if last_save_had_error is true our stream is NOT ok.
    }
    return isVideoStream;
}


//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief This slot is invoked when the RawToXML class finishes it's conversion of the raw per-frame
 *        metadata to an XML formatted file.
 */
void FiSaver::cleanUpConversion()
{
    // Conversion must be done, so clean up our pointer memory.
    if (rawToXml)
    {
        rawToXml->deleteLater();
        rawToXml = nullptr;
    }
}

