/*!
 * @file fi_saver.h
 * @brief This is the header file for the fiSaver class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_SAVER_H
#define FI_SAVER_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QObject>
#include <QString>

#include <stdint.h>

#include <memory>

#include <fas_video.h>
#include <fi_image.h>
#include <fi_stream.h>
#include <fi_durations.h>

#include "fsp_socket.h"
#include "fi_internal.h"

namespace FI_SAVER
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
//! Sequence number for a live image.
const uint32_t LIVE_IMAGE_SEQ_NUM               = 0;

//! Sequence number for the TXT metadata file.
const uint32_t  METADATA_TXT_SEQ_NUM            = ((uint32_t) -1);

//! Sequence number for the CFG metadata file.
const uint32_t  METADATA_CFG_SEQ_NUM            = ((uint32_t) -2);

//! Default flag setting to indicate saving Xcite CFG partition metadata.
const bool      DEFAULT_SAVE_METADATA_CFG       = false;

//! Sequence number for the XML metadata file.
const uint32_t  METADATA_XML_SEQ_NUM            = ((uint32_t) -3);

//! Sequence number for the RAW metadata file.
const uint32_t  METADATA_RAW_SEQ_NUM            = ((uint32_t) -4);

//! The length of the image file name's sequence number.
const int   IMAGE_SEQ_NUM_LEN                   = 7;

//! The length of the path name's sequence number.
const int   PATH_SEQ_NUM_LEN                    = 6;

//! The number of images can we put in a sub-folder.
const int   SUB_FOLDER_MAX_IMAGES               = 30000;

//! The point when we want to switch over to using sub-folders.
const int   SUB_FOLDER_THRESHOLD                = 60000;

//! Stream Format Types for FiSaver class.
typedef enum STREAM_TYPE
{
    NO_STREAM = 0,      // No stream format type.
    JPEG_STACK,         // Stream JPEG stack of images.
    BMP_STACK,          // Stream BMP stack of images.
    TIF_STACK,          // Stream TIF stack of images.
    TIF_RAW_STACK,      // Stream TIF RAW stack of images.
    DNG_STACK,          // Stream DNG stack of images.
    AVI,                // Stream AVI video.
    MP4,                // Stream MP4 video.
    FI_CAP,             // Stream Fastec Imaging Capture file.
    FI_CAPSTK,          // Stream Fastec Imaging Capture stack of CAP files.
    RDI_PACKED          // Stream RDI packed image format.
} STREAM_TYPE;

//! Image Format Types for FiSaver class.
typedef enum IMAGE_TYPE
{
    NO_IMAGE = 0,       // No image set.
    JPEG,               // Image JPEG file.
    BMP,                // Image BMP file.
    BMP_PLANAR,         // Image BMP Planar file.
    TIF,                // Image TIF file.
    TIF_RAW,            // Image TIF raw uncompressed file.
    DNG,                // Image DNG file.
    PACKED,             // Image PACK raw file.
} IMAGE_TYPE;

//! Define a structure to hold the many parameters required for a save operation.
typedef struct
{
    int         clip_end;                       //!< Relative frame number for the end of the save.
    int         clip_start;                     //!< Relative frame number for the start of the save.
    uint32_t    clip_end_abs;                   //!< Absolute frame number for the end of the save.
    uint32_t    clip_start_abs;                 //!< Absolute frame number for the start of the save.
    int         jpeg_quality;                   //!< If saving a JPEG, the quality factor to be used.
    uint32_t    h264_quality;                   //!< If saving MP4, the quality factory is H.264 crf option.
    FAS_VIDEO::H264_SPEEDS h264_speed;          //!< If saving MP4, the compression speed is H.264 preset options.
    bool        h264_enable_all_keyframes;      //!< If saving MP4, this will set all frames as keyframes if true.

    bool        create_xml;                     //!< True if we're to create an XML file.
    bool        trigger_relative;               //!< True if clip range is relative to the trigger position, false if using the user's bookmark.
    bool        use_camera_name;                //!< True if we're to use the camera's name in the file name(s).
    bool        use_camera_range;               //!< True to use the clip range set on the camera (multicamera mode).
    bool        merge_overlays;                 //!< True to merge any overlays, false otherwise.

    uint32_t    dest_media;                     //!< One of the GIGE_MEDIA_xxx values, or SAVE_DLG::MEDIA_LOCAL_PC.
    QString     dest_label;                     //!< For seriesNC cameras, the json "dest_label" value associated with the dest_media.
    uint32_t    playback_rate;                  //!< The desired playback rate of the video.
    STREAM_TYPE streamType;                     //!< The type of stream this saver is supporting.
    IMAGE_TYPE  imagesInputType;                //!< The type of image received for stream writers.
    IMAGE_TYPE  imagesOutputType;               //!< The type of images pushed to stream writer's encoder.
    bool        useRawFormat;                   //!< If saveType supports raw formats user wants to use them.
    bool        save_metadata_cfg;              //!< True if saving the partition metadata to CFG file.
    bool        save_pfmd;                      //!< True if saving the PFMD to a file.
    uint32_t    ref_frame;                      //!< Trigger or user bookmark absolute frame number.

    QString     comment;                        //!< Comment buffer used for the save, if any.
    QString     dest_path;                      //!< The path to the folder to hold the saved capture.
    QString     tag_label;                      //!< The tag to use, if any, in the file name(s).
} SAVE_PARAMS;

}   // End of namespace FI_SAVER.

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class   QByteArray;
class   QFile;

class   RawToXML;
struct pm_metadata_type;

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class FiSaver : public QObject
{
    Q_OBJECT

public:
    static const QString DEFAULT_TAG;
    static const QString DEFAULT_METADATA_FILE_NAME;
    static const QString STATS_FILE_NAME;
    static const QString IMAGE_EXT_JPG;
    static const QString IMAGE_EXT_TIF;
    static const QString IMAGE_EXT_BMP;
    static const QString IMAGE_EXT_BPP;
    static const QString IMAGE_EXT_DNG;
    static const QString IMAGE_EXT_CAPTURE;
    static const QString IMAGE_EXT_CAPSTACK;
    static const QString IMAGE_EXT_AVI;
    static const QString IMAGE_EXT_MP4;
    static const QString IMAGE_EXT_RDI_PACKED;
    static const QString IMAGE_EXT_META_TEXT;
    static const QString IMAGE_EXT_META_CFG;
    static const QString IMAGE_EXT_META_XML;
    static const QString IMAGE_EXT_META_RAW;
    static const QString SUB_FOLDER_PREFIX;

    // String associated with the STREAM_TYPE settings we support.
    static const QString   FORMAT_JPG_STR;
    static const QString   FORMAT_BMP_STR;
    static const QString   FORMAT_TIF_STR;
    static const QString   FORMAT_TIF_RAW_STR;
    static const QString   FORMAT_DNG_STR;
    static const QString   FORMAT_AVI_STR;
    static const QString   FORMAT_MP4_STR;
    static const QString   FORMAT_CAP_STR;
    static const QString   FORMAT_CAPSTK_STR;
    static const QString   FORMAT_META_STR;
    static const QString   FORMAT_RDI_PACKED_STR;
    static const QString   FORMAT_UNKNOWN_STR;

public:
    explicit FiSaver(QObject *parent = nullptr);
    ~FiSaver();

    void        buildDestPath();
    void        buildFileName(QString& file_name, uint32_t file_seq_num, bool do_retry);
    bool        configure(const FI_SAVER::SAVE_PARAMS& save_info);
    int         createDestPath();
    int         createFilePath();
    void        createSubFolders();
    void        finishSave();
    void        generateSubfolderName(QString& sub_folder, int index);
    QString     metadataToText(QByteArray *metadata_ptr, FI_SAVER::STREAM_TYPE streamType, FI_SAVER::IMAGE_TYPE imageType, uint32_t clip_start, uint32_t clip_end, const FAS_VIDEO::VIDEO_GAINS *gainsPtr = nullptr);

    virtual int preparePackedWriter(const struct pm_metadata_type *pm_ptr = nullptr);
    virtual int prepareStreamWriter();

    void        reset();
    bool        saveToStream(const uint8_t *imageDataPtrToWrite, uint32_t imageDataLenToWrite, uint32_t imageDataSeqNbr, bool do_retry);
    bool        saveToPackedStream(const uint8_t *imageDataPtrToWrite, uint32_t imageDataLenToWrite, uint32_t imageDataSeqNbr, bool retryInProgress);

    bool        savePartitionMetadata(QByteArray *metadata_ptr, uint32_t red_gain = (uint32_t) -1, uint32_t green_gain = (uint32_t) -1, uint32_t blue_gain = (uint32_t) -1);
    bool        saveRawMetadata(const QByteArray *metadata_ptr, const struct pm_metadata_type *pm_ptr = nullptr, uint32_t red_gain = (uint32_t) -1, uint32_t green_gain = (uint32_t) -1, uint32_t blue_gain = (uint32_t) -1);
    bool        closeStreamWriter();

    QString     getBasePath()                                       { return basePath; }
    QString     getCameraName()                                     { return cameraName; }
    uint32_t    getClipStart()                                      { return clipStart; }
    uint32_t    getClipEnd()                                        { return clipEnd; }
    QString     getComment();
    uint32_t    getDestMedia()                                      { return destMedia; }
    QString     getDestLabel()                                      { return destLabel; }
    QString     getErrMsg()                                         { return lastErrMsg; }
    uint32_t    getFrameCount()                                     { return frameCount; }
    const QString& getFileExtension()                               { return fileExtension; }
    const QString& getImageFilename()                               { return imageFilename; }
    bool        getMetadataCfgSave()                                { return saveMetadataCfgFile; }
    FI_SAVER::STREAM_TYPE getStreamType()                           { return streamType; }
    FI_SAVER::IMAGE_TYPE getImageInputType()                        { return imageInputType; }
    FI_SAVER::IMAGE_TYPE getImageEncoderType()                      { return imageTypeForEncoder; }
    QString     getLocalPath()                                      { return destPath; }
    uint32_t    getRetryFrame()                                     { return absRetryFrame; }
    QString     getTag()                                            { return tagString; }
    bool        getSavePfmd()                                       { return savePfmd; }
    uint32_t    getH264Quality()                                    { return h264Quality; }
    FAS_VIDEO::H264_SPEEDS  getH264Speed()                          { return h264Speed; }
    bool        getH264EnableAllKeyframes()                         { return h264EnableAllKeyframes; }

    virtual bool getVideoStreamSaveStatus(QString& infoStr, uint32_t& frameCount, uint32_t& framesAdded,
                                          uint32_t& framesEncoded, bool& streamIsOk, bool& streamIsComplete);

    void        setImageHeight(uint32_t height)                     { streamImageHeight = height; }
    void        setImageWidth(uint32_t width)                       { streamImageWidth = width; }
    void        setPlaybackRate(uint32_t rate)                      { streamPlaybackRate = rate; }
    void        setBlobSize(uint32_t new_size)                      { blobSize = new_size; }
    void        setCameraName(const QString& camera_name)           { cameraName = camera_name; }
    void        setClipStart(uint32_t start_frame)                  { clipStart = start_frame; }
    void        setClipEnd(uint32_t end_frame)                      { clipEnd = end_frame; }
    void        setCount(uint32_t count)                            { frameCount = count; }
    void        setFileExtension(FI_SAVER::STREAM_TYPE streamType);
    void        setMetadataCfgSave(bool save_metadata_cfg);
    void        setPath(const QString& save_path);
    void        setRetryFrame(uint32_t retry_frame)                 { absRetryFrame = retry_frame; retryInProgress = true; }
    void        setStartIndex(uint32_t start_num)                   { startIndex = start_num; }
    void        setTag(const QString& tag_str)                      { tagString = tag_str; }
    bool        setStreamType(FI_SAVER::STREAM_TYPE saveStreamType, FI_SAVER::IMAGE_TYPE saveImageInputType, FI_SAVER::IMAGE_TYPE saveImageEncoderType);
    void        setH264Quality(uint32_t h264QualityToSet);
    void        setH264Speed(FAS_VIDEO::H264_SPEEDS h264SpeedToSet);
    void        setH264EnableAllKeyframes(bool h264EnableAllKeyframesToSet);

    static FI_SAVER::STREAM_TYPE fileExtToStreamType(QString fileExtensionStr);
    static FI_SAVER::IMAGE_TYPE fileExtToImageType(QString fileExtensionStr);
    static bool isLiveSeqNumber(uint32_t seqNumberToCheck);
    static FI_SAVER::STREAM_TYPE streamFormatToStreamType(uint32_t stream_format);
    static bool metadataSeqNum(uint32_t seq_num);

    static bool isCaptureStream(FI_SAVER::STREAM_TYPE streamType);
    static bool isSnapshotImage(FI_SAVER::IMAGE_TYPE imageType, uint32_t frameCount);
    static bool isStackOfCaps(FI_SAVER::STREAM_TYPE streamType);
    static bool isStackOfImages(FI_SAVER::STREAM_TYPE streamType);
    static bool isStreamVideo(FI_SAVER::STREAM_TYPE streamType);
    static bool isStreamFFMpegVideo(FI_SAVER::STREAM_TYPE streamType);

    static uint32_t streamTypeToCameraStreamType(FI_SAVER::STREAM_TYPE streamType, bool is_series3);
    static uint32_t imageTypeToCameraImageType(FI_SAVER::STREAM_TYPE streamType, FI_SAVER::IMAGE_TYPE imageType, bool is_series3);
    static QString streamTypeToQString(FI_SAVER::STREAM_TYPE streamType, bool forLog = true);
    static QString imageTypeToQString(FI_SAVER::IMAGE_TYPE imageType);
    static QString getGIGEStreamStr(uint32_t gigeStreamFormat);

    static bool verifyJpegQuality(const char *pretty_str, const char *variable_name, uint32_t value_to_check, bool showMBWarning = false);

protected:
    std::unique_ptr<FAS_VIDEO::FiVideoStream> streamWriter;     //!< Our Fastec Imaging Stream Writer.
    std::unique_ptr<FAS_VIDEO::FiDurations> imageXlateStats;
    std::unique_ptr<FAS_VIDEO::FiDurations> saveImageStats;     //!< Keeps track of total time spent in saveToStream().
    std::shared_ptr<FAS_VIDEO::FiImage> streamImage;            //!< A reusable FiImage for Stream Writer.
    QByteArray imageByteArray;                                  //!< Some stream writers modify image before sending to encoder.
    FAS_INTERNAL::FIP_HANDLE streamHandle;                      //!< Handle to the packed streamer.

    uint32_t    absRetryFrame;                                  //!< Absolute frame number we're retrying.
    uint32_t    streamImageHeight;                              //!< The height of the image for video streams.
    uint32_t    streamPlaybackRate;                             //!< Playback rate "suggestion" for video streams.
    uint32_t    streamImageWidth;                               //!< The width of the image for video streams.
    QString     basePath;                                       //!< Base folder/file path that we're going to save to.
    QFile       *blobFile;                                      //!< Our blob file object (tyically a CAP file).
    uint32_t    blobSize;                                       //!< Largest "blob"/"chunk" of data we can deal with.
    QString     cameraName;                                     //!< The camera name, if we're using it.
    uint32_t    clipStart;                                      //!< The frame number of the starting frame of the clip.
    uint32_t    clipEnd;                                        //!< The frame number of the last frame of the clip.
    QByteArray  *commentBuf;                                    //!< The comment provided by the user, if any.
    bool        convertRawToXml;                                //!< True if the user wants to convert the raw PFMD to an XML file.
    uint32_t    destMedia;                                      //!< Destination media for the save.
    QString     destLabel;                                      //!< Label for the destination media associated with dest_media.
    QString     destPath;                                       //!< The full path to the file/folder we targeting.
    uint32_t    frameCount;                                     //!< The number of frames we are expecting to receive.
    QString     fileExtension;                                  //!< The file extension associated with the streamType.
    QString     imageFilename;                                  //!< The file name used to save the latest image (typically based on the sequence number).
    int         jpegQuality_4QTPixmaps;                         //!< JPEG quality factory to use if we are translating stream images via QT to JPEGs.
    bool        streamIsClosed;                                 //!< When closing down we want to hold off on any additional calls to stream object.
    FI_SAVER::STREAM_TYPE streamType;                           //!< The stream type set for saving video.
    FI_SAVER::IMAGE_TYPE imageTypeForEncoder;                   //!< The image type for frames sent to video stream writer's encoder.
    FI_SAVER::IMAGE_TYPE imageInputType;                        //!< The image type for frames read from input stream (camera or other).
    FAS_VIDEO::H264_SPEEDS h264Speed;                           //!< The H.264 speed to use if MP4 video is being saved.
    uint32_t    h264Quality;                                    //!< The H.264 frame quality to use if MP4 video is being saved.
    bool h264EnableAllKeyframes;                                //!< The H.264 enable all keyframe setting.
    QString     lastErrMsg;                                     //!< The last error message, if any.
    bool        lastSaveHadError;                               //!< What was the condition of the last stream save?
    uint32_t    lastSeqNbrSaverReceived;                        //!< What was the last image sequence number sent to stream save?
    uint32_t    lastSeqNbrSaverSaved;                           //!< What was the last image sequence number successfully saved?
    QString     lastFilenameWritten;                            //!< The last filename written if saving stacks or metadata files.
    uint32_t    streamSaveBufferCount;                          //!< What is the total count of frames the streamer has processed?

    bool        retryInProgress;                                //!< True if we're in the retry stage of a save.
    uint32_t    startIndex;                                     //!< The index we'll start looking at to name the folder/file.
    QString     tagString;                                      //!< The tag, if we're using it.
    bool        usingSubfolders;                                //!< True if we're using sub-folders, false otherwise.

    bool        saveMetadataCfgFile;                            //!< True if we are also saving the metadata cfg file to disk.
    bool        savePfmd;                                       //!< True if we are saving the PFMD to a file.

private slots:
    void        cleanUpConversion();

private:
    RawToXML    *rawToXml;                                      //!< Our raw metadata to XML conversion object.

};

#endif // FI_SAVER_H
