/*!
 * @file fi_sensor.cpp
 * @brief This is a base class to facilitate abstraction and basic
 *        sensor related functionality.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QTimer>

#include "fi_sensor.h"
#include "fi_sensor_priv.h"
#include "fi_sensor_lince5m.h"
#include "fi_sensor_lupa1300.h"
#include "fi_sensor_lux13hs.h"
#include "fi_sensor_lux19hs.h"

#include <errno.h>

#include "camera_mdi.h"
#include "fi_comm.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"

namespace FI_SENSOR
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

const int   QUERY_DELAY             = 10;

}   // End of namespace FI_SENSOR.

using namespace FI_SENSOR;

//--------------------------------------------------------------------
// fiSensorPrivate class implementation.
//--------------------------------------------------------------------
fiSensorPrivate::fiSensorPrivate(fiSensor *parent, uint32_t sensor_id) :
    my_parent(parent),
    sensor_max_frame_rate(0),
    sensor_max_height(0),
    sensor_max_width(0),
    sensor_min_height(0),
    sensor_min_width(0),
    sensor_type(sensor_id),
    icm_version(QVersionNumber(0, 0, 0)),
    fpga_version(QVersionNumber(0, 0, 0))
{
}

fiSensorPrivate::~fiSensorPrivate()
{
}

/*!
 * @brief Update registers of interest to the object.
 * @param reg_addr      The register address.
 * @param reg_value     The value of the register.
 */
void fiSensorPrivate::reg_update(uint32_t reg_addr, uint32_t reg_value)
{
    // Process all registers we're interested in.
    switch (reg_addr)
    {
    case GIGE_REG_MAX_FRAME_RATE:
        sensor_max_frame_rate = reg_value;
        break;
    case GIGE_REG_MIN_HEIGHT:
        sensor_min_height = (uint16_t)reg_value;
        break;
    case GIGE_REG_MIN_WIDTH:
        sensor_min_width = (uint16_t)reg_value;
        break;
    case GIGE_REG_SENSOR_HEIGHT:
        sensor_max_height = (uint16_t)reg_value;
        break;
    case GIGE_REG_SENSOR_WIDTH:
        sensor_max_width = (uint16_t)reg_value;
        break;
    case GIGE_REG_FPGA_VER:
        fpga_version = QVersionNumber((reg_value >> 24) & 0xFF, (reg_value >> 8) & 0xFFFF, reg_value & 0xFF);
        break;
    case GIGE_REG_ICM_MODULE_VER:
        icm_version = QVersionNumber((reg_value >> 24) & 0xFF, (reg_value >> 8) & 0xFFFF, reg_value & 0xFF);
        break;
    default:
        break;
    }
}


//--------------------------------------------------------------------
// fiSensor class implementation.
//--------------------------------------------------------------------

/*!
 * @brief The class constructor.
 *
 * @param parent        It is required that this point to a CameraMDI object.
 */
fiSensor::fiSensor(QObject *parent) :
    QObject(parent)
{
    FI_comm     *comm_ptr;

    // Get our parent object cast in a usable form.
    mdi_parent = qobject_cast<CameraMDI *>(parent);
    if (mdi_parent == nullptr)
    {
        // Unacceptable!
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Parent is not CameraMDI!";
        return;
    }

    // Initialize our members.
    priv_ptr = nullptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Hook up our slots.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &fiSensor::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::unsol_msg_received, this, &fiSensor::handle_unsol_msg);

    // Ask our parent to read the sensor type ID from the camera.
    mdi_parent->start_camera_read_reg(GIGE_REG_SENSOR_TYPE);
}

/*!
 * @brief fiSensor class destructor.
 */
fiSensor::~fiSensor()
{
    if (priv_ptr != nullptr)
    {
        delete priv_ptr;
        priv_ptr = nullptr;
    }
}

/*!
 * @brief Return the type of sensor we're using on the camera.
 */
uint32_t fiSensor::get_type(void)
{
    // Sanity check.
    if (priv_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null priv_ptr!";
        return SENSOR_TYPE_UNKNOWN;
    }

    // Return the type of sensor we're using.
    return priv_ptr->get_type();
}

/*!
 * @brief Return true if the sensor is color, false for mono.
 */
bool fiSensor::is_color(void)
{
    // Sanity check.
    if (priv_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null priv_ptr!";
        return false;
    }

    // Return if we're color or mono.
    return priv_ptr->is_color();
}

/*!
 * @brief Calculate the maximum frame rate allowed for the specified ROI.
 *
 * @param p_roi     The ROI of interest.
 * @return The maximum allowed frames-per-second.
 */
uint32_t fiSensor::max_frame_rate(QRect *p_roi, uint32_t bit_depth)
{
    // Sanity check.
    if (p_roi == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null ROI?!";
        return 0;
    }

    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        return 0;
    }

    // Get the information.
    return priv_ptr->max_frame_rate(p_roi, bit_depth);
}

int fiSensor::max_pixel_depth(void)
{
    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        return 0;
    }

    // Get the information.
    return priv_ptr->max_pixel_depth();
}

QRect fiSensor::max_roi(uint32_t bit_mode)
{
    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        return QRect();
    }

    // Get the information.
    return priv_ptr->max_roi(bit_mode);
}

/*!
 * @brief Returns true if the sensor has options such as binning and/or subsampling.
 */
bool fiSensor::opts_avail(void)
{
    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        return false;
    }

    // Get the information.
    return priv_ptr->opts_avail();
}

/*!
 * @brief Allocate the appropriate sensor object based on the type received.
 * @param sensor_type   One of FI_SENSOR::SENSOR_TYPE_xxx
 */
void fiSensor::process_sensor_type(uint32_t sensor_type)
{
    // We only ever get one sensor attached to a camera, so we shouldn't
    // have a sensor class allocated yet.
    if (priv_ptr != nullptr)
    {
        // Sanity check.
        if (priv_ptr->get_type() != sensor_type)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Existing sensor type:" << priv_ptr->get_type() << "new type:" << sensor_type;
        }

        return;
    }

    // Allocate our sensor object based on the type.
    switch (sensor_type)
    {
    case SENSOR_TYPE_LUX13HS_COLOR:
    case SENSOR_TYPE_LUX13HS_MONO:
        priv_ptr = new fiSensorLux13HS(this, sensor_type);
        break;
    case SENSOR_TYPE_LUX19HS_COLOR:
    case SENSOR_TYPE_LUX19HS_MONO:
        priv_ptr = new fiSensorLux19HS(this, sensor_type);
        break;
    case SENSOR_TYPE_LUX160_COLOR:
    case SENSOR_TYPE_LUX160_MONO:
        // priv_ptr = new fiSensorLux160(this, sensor_type);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Luxima160 sensor is not yet supported!";
        break;
    case SENSOR_TYPE_LUX2100_MONO:
    case SENSOR_TYPE_LUX2100_COLOR:
        // priv_ptr = new fiSensorLux2100(this, sensor_type);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Lux2100 sensor is not yet supported!";
        break;
    case SENSOR_TYPE_LINCE5M_COLOR:
    case SENSOR_TYPE_LINCE5M_MONO:
        priv_ptr = new fiSensorLince5M(this, sensor_type);
        break;
    case SENSOR_TYPE_LUPA1300_COLOR:
    case SENSOR_TYPE_LUPA1300_MONO:
        priv_ptr = new fiSensorLupa1300(this, sensor_type);
        break;
    default:
        break;
    }

    // Query the intialization parameters after a short delay.
    QTimer::singleShot(QUERY_DELAY, this, &fiSensor::init_query);
}

/*!
 * @brief Set the options currently in use.
 * @param opt_value     The current GIGE_REG_SENSOR_OPTIONS value.
 */
void fiSensor::set_options(uint32_t opt_value)
{
    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No priv_ptr?!";
        return;
    }

    // Pass it on.
    priv_ptr->set_options(opt_value);
}

/*!
 * @brief Returns true if the sensor supports advanced calibration features.
 */
bool fiSensor::supports_adv_cal(void)
{
    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        return false;
    }

    // Get the information.
    return priv_ptr->supports_adv_cal();
}

/*!
 * @brief Returns true if the sensor supports the FPGA pixel FPN correction.
 */
bool fiSensor::supports_pixel_fpn(void)
{
    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        return false;
    }

    // Get the information.
    return priv_ptr->supports_pixel_fpn();
}

/*!
 * @brief Return true if this sensor has a built-in FPN correction, false otherwise.
 */
bool fiSensor::supports_sensor_fpn()
{
    // Make sure we've got our sensor - may not in the early stages.
    if (priv_ptr == nullptr)
    {
        return false;
    }

    // Get the information.
    return priv_ptr->supports_sensor_fpn();
}

bool fiSensor::supports_video_trigger()
{
    // Sanity check.
    if (priv_ptr == nullptr)
        return false;

    // Only newer cameras support video trigger.
    if (mdi_parent == nullptr)
        return false;
    if (mdi_parent->is_series3())
        return false;

    // Get the information.
    return priv_ptr->supports_video_trigger();
}

int fiSensor::height_mod()
{
    int mod = DEFAULT_HEIGHT_MOD;

    // Sanity check.
    if (priv_ptr != nullptr)
        mod = priv_ptr->height_mod();

    // Return the modulus, or step, for the height.
    return mod;
}

int fiSensor::width_mod()
{
    int mod = DEFAULT_WIDTH_MOD;

    // Sanity check.
    if (priv_ptr != nullptr)
        mod = priv_ptr->width_mod();

    // Return the modulus, or step, for the width.
    return mod;
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void fiSensor::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    // Sanity check.
    if (mdi_parent == nullptr)
        return;

    // Is this event for us?
    if (ip_addr != mdi_parent->get_ip_addr())
        return;

    // Sanity check. Make sure we know what we are.
    if (priv_ptr != nullptr)
    {
        // Update the sensor object with the value.
        priv_ptr->reg_update(reg_addr, reg_value);
    }
}

/*!
 * @brief Delayed query for initialiation settings.
 */
void fiSensor::init_query(void)
{
    // Sanity check!
    if (mdi_parent == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "mdi_parent is nullptr!";
        return;
    }

    // Are the comms ready?
    if (!mdi_parent->comms_are_ready())
    {
        // Come back in a little bit.
        QTimer::singleShot(QUERY_DELAY, this, &fiSensor::init_query);
        return;
    }

    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if( cam_register_group.isNull() )
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Prepare the camera register group read.
    cam_register_group->addReg(GIGE_REG_MIN_HEIGHT);
    cam_register_group->addReg(GIGE_REG_MIN_WIDTH);
    cam_register_group->addReg(GIGE_REG_SENSOR_WIDTH);
    cam_register_group->addReg(GIGE_REG_SENSOR_HEIGHT);
    cam_register_group->addReg(GIGE_REG_MAX_FRAME_RATE);
    if (mdi_parent->is_series3())
    {
        cam_register_group->lastReg(GIGE_REG_FPGA_VER);
    }
    else
    {
        cam_register_group->addReg(GIGE_REG_FPGA_VER);
        cam_register_group->lastReg(GIGE_REG_ICM_MODULE_VER);
    }
    mdi_parent->read_camera_registers(cam_register_group);
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void fiSensor::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Sanity check.
    if (mdi_parent == nullptr)
        return;

    // Is this event for us?
    if (ip_addr != mdi_parent->get_ip_addr())
        return;

    // Perform any special action we might want on a particular register.
    switch (reg_addr)
    {
    case GIGE_REG_SENSOR_TYPE:
        process_sensor_type(mdi_parent->read_camera_cache(GIGE_REG_SENSOR_TYPE));
        break;
    default:
        // Sanity check. Make sure we know what we are.
        if (priv_ptr != nullptr)
        {
            // Update the sensor object with the value.
            priv_ptr->reg_update(reg_addr, mdi_parent->read_camera_cache(reg_addr));
        }
        break;
    }
}

