/*!
 * @file fi_sensor_lince5m.cpp
 * @brief This class implements the sensor specific functions for the
 *        Lince5M sensor.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_sensor_lince5m.h"

#include "gige_msg.h"
#include "logger.h"
#include "lince5m_utils.h"

using namespace FI_SENSOR;

/*!
 * @brief Class constructor
 * @param parent        Pointer to our fiSensor base class.
 * @param sensor_id     Our sensor type ID.
 */
fiSensorLince5M::fiSensorLince5M(fiSensor *parent, uint32_t sensor_id) :
    fiSensorPrivate(parent, sensor_id),
    current_options(0)
{
}

/*!
 * @brief Class destructor.
 */
fiSensorLince5M::~fiSensorLince5M()
{
}

/*!
 * @brief Return true if the sensor is a color sensor, false for  mono.
 */
bool fiSensorLince5M::is_color()
{
    return (sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR);
}

/*!
 * @brief Compute the maximum framerate for a given ROI.
 *
 * @param p_roi     Pointer to the ROI of interest.
 * @return The maximum frames-per-second allowed.
 */
uint32_t fiSensorLince5M::max_frame_rate(QRect *p_roi, uint32_t bit_depth)
{
    uint32_t    frame_rate;

    // Sanity check.
    if (p_roi == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null ROI!?";
        return 0;
    }

    // Calculate the maximum frame rate based on the sensor dimensions.
    frame_rate = lince5m_max_fps_setting((uint16_t)p_roi->x(), (uint16_t)p_roi->y(), (uint16_t)p_roi->width(), (uint16_t)p_roi->height(), (uint8_t)bit_depth, current_options);

    // Make sure our maximum rate fits within the limitations of what is allowed.
    if (frame_rate > sensor_max_frame_rate)
        frame_rate = sensor_max_frame_rate;

    // Return the rate.
    return frame_rate;
}

/*!
 * @brief Return the maximum size (number of bits) of a pixel on this sensor.
 */
int fiSensorLince5M::max_pixel_depth()
{
    return 12;
}

/*!
 * @brief Return the maximum ROI supported by the sensor for a given bit mode.
 *
 * @param bit_mode      The pixel depth selected by the user.
 * @return The maximum ROI supported by the sensor for the selected bit mode.
 */
QRect fiSensorLince5M::max_roi(uint32_t bit_mode)
{
    uint16_t    max_height;
    uint16_t    max_width;
    QRect   max_roi(0, 0, 0, 0);

    // Calculate the maximum ROI supported.
    lince5m_max_roi_setting((uint8_t)(bit_mode & 0xFF), &max_width, &max_height);

    // Apply possible limits.
    if (max_height > sensor_max_height)
        max_height = sensor_max_height;
    if (max_width > sensor_max_width)
        max_width = sensor_max_width;

    // Adjust our return value.
    max_roi.setHeight((int)max_height);
    max_roi.setWidth((int)max_width);

    // Return our value.
    return max_roi;
}

/*!
 * @brief Return true if binning and/or subsampling are supported.
 */
bool fiSensorLince5M::opts_avail()
{
    return true;
}

/*!
 * @brief Update registers of interest to the object.
 * @param reg_addr      The register address.
 * @param reg_value     The value of the register.
 */
void fiSensorLince5M::reg_update(uint32_t reg_addr, uint32_t reg_value)
{
    // Pass it on to the parent class.
    fiSensorPrivate::reg_update(reg_addr, reg_value);
}

/*!
 * @brief Set the options currently in use.
 * @param opt_value     The current GIGE_REG_SENSOR_OPTIONS value.
 */
void fiSensorLince5M::set_options(uint32_t opt_value)
{
    // Save this for when we need to do things like max frame rate calculations.
    current_options = opt_value;
}

/*!
 * @brief Return true if we do pixel FPN for this sensor, false otherwise.
 */
bool fiSensorLince5M::supports_pixel_fpn()
{
    return true;
}

/*!
 * @brief Return true if this sensor has a built-in FPN correction, false otherwise.
 */
bool fiSensorLince5M::supports_sensor_fpn()
{
    return true;
}

/*!
 * @brief Return true if this sensor (ICM) supports video trigger.
 */
bool fiSensorLince5M::supports_video_trigger()
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << "FPGA:" << fpga_version << "ICM:" << icm_version;

    // The video trigger was added in FPGA version 1.0.5, ICM version 1.0.3.
    return ((QVersionNumber::compare(fpga_version, QVersionNumber(1, 0, 5)) >= 0) &&
            (QVersionNumber::compare(icm_version, QVersionNumber(1, 0, 3)) >= 0));
}

/*!
 * @brief Return true if we have advanced calibration functions
 *        for this sensor, false otherwise.
 */
bool fiSensorLince5M::supports_adv_cal()
{
    return true;
}
