/*!
 * @file fi_sensor_lupa1300.cpp
 * @brief This class implements the sensor specific functions for the
 *        Lupa-1300-2 sensor.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_sensor_lupa1300.h"
#include "logger.h"

#include <qmath.h>

namespace FI_SENSOR_LUPA1300
{
    //--------------------------------------------------------------------
    // Constant, Macro and Type Definitions
    //--------------------------------------------------------------------

    //! The number of pixels per kernel.
    const int       SENSOR_KERN_WIDTH           = 24;

    //! Sensor clock 63 Mhz.
    const double    CLOCK                       = 6.3E+7;

    //! Frame overhead time, in CLOCKs.
    const double    FOT                         = 465.0;

    //! Row overhead time.
    const double    ROT                         = 12.0;

    //! The per-frame overhead required for frame read-out, in microseconds.
    const uint32_t  DEFAULT_FRAME_OVERHEAD      = 6;

}   // End of namespace FI_SENSOR_LUPA1300.

using namespace FI_SENSOR_LUPA1300;
using namespace FI_SENSOR;

/*!
 * @brief Class constructor
 * @param parent        Pointer to our fiSensor base class.
 * @param sensor_id     Our sensor type ID.
 */
fiSensorLupa1300::fiSensorLupa1300( fiSensor *parent, uint32_t sensor_id ) :
    fiSensorPrivate( parent, sensor_id )
{
}

/*!
 * @brief Class destructor.
 */
fiSensorLupa1300::~fiSensorLupa1300()
{
}

/*!
 * @brief Return true if the sensor is a color sensor, false for  mono.
 */
bool fiSensorLupa1300::is_color( void )
{
    return ( sensor_type == SENSOR_TYPE_LUPA1300_COLOR );
}

/*!
 * @brief Compute the maximum framerate for a given ROI.
 *
 * @param p_roi     Pointer to the ROI of interest.
 * @return The maximum frames-per-second allowed.
 */
uint32_t fiSensorLupa1300::max_frame_rate( QRect *p_roi, uint32_t bit_depth )
{
    int k_start = 0;
    int k_end = 0;
    double f_height;
    double f_kwidth = 0.0;
    double max_framerate = 0.0;

    Q_UNUSED( bit_depth );

    // Sanity check.
    if ( p_roi == nullptr )
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null ROI!?";
        return 0;
    }

    // Get the height in a floating point format.
    f_height = (double)p_roi->height();

    // Figure out the number of kernels.
    pixels_to_kernels( p_roi->x(), p_roi->width(), &k_start, &k_end );
    f_kwidth = (double)(( k_end - k_start ) + 1 );

    // Calculate the maximum framerate.
    max_framerate = CLOCK / (((( f_kwidth * 2 ) + ROT ) * f_height ) + FOT );

    // Make sure it's within what the camera says it can be.
    if ( max_framerate > (double)sensor_max_frame_rate )
        max_framerate = (double)sensor_max_frame_rate;

    // Return our framerate.
    return (uint32_t)qFloor( max_framerate );
}

/*!
 * @brief Return the maximum size (number of bits) of a pixel on this sensor.
 */
int fiSensorLupa1300::max_pixel_depth( void )
{
    return 10;
}

/*!
 * @brief Return the maximum ROI supported by the sensor for a given bit mode.
 *
 * For the LUPA-1300, the maximum ROI is unaffected by the bit mode.
 *
 * @param bit_mode      [UNUSED] The pixel depth selected by the user.
 * @return The maximum ROI supported by the sensor for the selected bit mode.
 */
QRect fiSensorLupa1300::max_roi( uint32_t bit_mode )
{
    Q_UNUSED( bit_mode );
    return QRect( 0, 0, (int)sensor_max_width, (int)sensor_max_height );
}

/*!
 * @brief Return true if binning and/or subsampling are supported.
 */
bool fiSensorLupa1300::opts_avail( void )
{
    return false;
}

/*!
 * @brief Utility function to convert from a pixel based address and width to a kernel
 *        based address and width. This is necessary for accurate frame rate calculations.
 *
 * @param pixels_start  Our pixel start address.
 * @param pixels_width  The size of our pixel (8, or 10 bit).
 *
 * @param k_start       The return kernel start address.
 * @param k_end         The ending kernel address.
 * @param k_width       The number of the kernels covered.
 */
void fiSensorLupa1300::pixels_to_kernels( int pixels_start, int pixels_width, int *k_start, int *k_end )
{
    *k_start = ( pixels_start / SENSOR_KERN_WIDTH );
    *k_end = ( pixels_start + pixels_width - 1 ) / SENSOR_KERN_WIDTH;
}

/*!
 * @brief Update registers of interest to the object.
 * @param reg_addr      The register address.
 * @param reg_value     The value of the register.
 */
void fiSensorLupa1300::reg_update( uint32_t reg_addr, uint32_t reg_value )
{
    // Pass it on to the parent class.
    fiSensorPrivate::reg_update( reg_addr, reg_value );
}

/*!
 * @brief Set the options currently in use.
 * @param opt_value     [unused] The current GIGE_REG_SENSOR_OPTIONS value.
 */
void fiSensorLupa1300::set_options( uint32_t opt_value )
{
    Q_UNUSED( opt_value );
}

/*!
 * @brief Return true if we do pixel FPN for this sensor, false otherwise.
 */
bool fiSensorLupa1300::supports_pixel_fpn( void )
{
    return true;
}

/*!
 * @brief Return true if this sensor has a built-in FPN correction, false otherwise.
 */
bool fiSensorLupa1300::supports_sensor_fpn()
{
    return true;
}

/*!
 * @brief Return true if we have advanced calibration functions
 *        for this sensor, false otherwise.
 */
bool fiSensorLupa1300::supports_adv_cal( void )
{
    return true;
}
