/*!
 * @file fi_sensor_lux19hs.cpp
 * @brief This class implements the sensor specific functions for the
 *        Lux19HS sensor.
 *
 * Copyright 2018 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QByteArray>
#include <QObject>
#include <qmath.h>

#include "fi_sensor_lux19hs.h"

#include "gige_msg.h"
#include "logger.h"

namespace LUX19HS_SENSOR
{
const int SENSOR_PCLK = 175000000;      //!< Sensor's clock rate, in MHz.
const int SENSOR_MCLK = (SENSOR_PCLK / 2);

const int QROW_TIME = 128;              //!< Number of clocks for one row to be read.
const int LAST_ROW_READOUT = (int)(QROW_TIME * 1.5);       //!< Number of clocks to finish reading out all the rows. (QROW_TIME * 1.5)
const int POST_READ_GAP = 82;           //!< Minimum number of clocks between the readout and the TX_n pulse.
const int READOUT_START = 8;            //!< When the FPGA decides it has a start frame.
const int TX_N_WIDTH = 100;             //!< Width of the TX_n pulse, in clocks.
const int TX2_N_WIDTH = 20;             //!< Width of the TX2_n pulse, in clocks.
const int TX2_TX_GAP = 30;              //!< Gap between the end of the TX2_n pulse and the start of the TX_n pulse.
const int PD_OVERLAP = 4;
const int PRE_READ_GAP = 4;

//! Countdown clock value for when to end the TX_n pulse.
const int TX_N_DONE = READOUT_START + PD_OVERLAP + PRE_READ_GAP;

//! Countdown clock value for when to start the TX_n pulse.
const int TX_N_START = TX_N_DONE + TX_N_WIDTH;

//! Countdown clock value, in front of TX_n pulse, for when to end the TX2_n pulse
const int TX2_N_DONE = TX_N_START + TX2_TX_GAP;

//! Countdown clock value for when to start the TX2_n pulse.
const int TX2_N_START = TX2_N_DONE + TX2_N_WIDTH;

//! Number of overhead clocks per frame.
const int FRAME_OVERHEAD = LAST_ROW_READOUT + POST_READ_GAP + TX2_N_START;
}

using namespace LUX19HS_SENSOR;
using namespace FI_SENSOR;

/*!
 * @brief Class constructor
 * @param parent        Pointer to our fiSensor base class.
 * @param sensor_id     Our sensor type ID.
 */
fiSensorLux19HS::fiSensorLux19HS(fiSensor *parent, uint32_t sensor_id) :
    fiSensorPrivate(parent, sensor_id),
    current_options(0)
{
}

/*!
 * @brief Class destructor.
 */
fiSensorLux19HS::~fiSensorLux19HS()
{
}

/*!
 * @brief Return true if the sensor is a color sensor, false for  mono.
 */
bool fiSensorLux19HS::is_color()
{
    return (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR);
}

/*!
 * @brief Compute the maximum framerate for a given ROI.
 *
 * @param p_roi     Pointer to the ROI of interest.
 * @return The maximum frames-per-second allowed.
 */
uint32_t fiSensorLux19HS::max_frame_rate(QRect *p_roi, uint32_t /* bit_depth */)
{
    uint32_t    frame_rate;

    // Sanity check.
    if (p_roi == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null ROI!?";
        return 0;
    }

    // Calculate the maximum frame rate based on the sensor dimensions.
    frame_rate = SENSOR_MCLK / (uint32_t)((QROW_TIME * (p_roi->height() / 4)) + FRAME_OVERHEAD);

    // Make sure our maximum rate fits within the limitations of what is allowed.
    if (frame_rate > sensor_max_frame_rate)
        frame_rate = sensor_max_frame_rate;

    // Return the rate.
    return frame_rate;
}

/*!
 * @brief Return the maximum size (number of bits) of a pixel on this sensor.
 */
int fiSensorLux19HS::max_pixel_depth()
{
    return 10;
}

/*!
 * @brief Return the maximum ROI supported by the sensor for a given bit mode.
 *
 * @param bit_mode      The pixel depth selected by the user.
 * @return The maximum ROI supported by the sensor for the selected bit mode.
 */
QRect fiSensorLux19HS::max_roi(uint32_t /* bit_mode */)
{
    QRect   max_roi(0, 0, sensor_max_width, sensor_max_height);

    // Return our value.
    return max_roi;
}

/*!
 * @brief Return true if binning and/or subsampling are supported.
 */
bool fiSensorLux19HS::opts_avail()
{
    return false;
}

/*!
 * @brief Update registers of interest to the object.
 * @param reg_addr      The register address.
 * @param reg_value     The value of the register.
 */
void fiSensorLux19HS::reg_update(uint32_t reg_addr, uint32_t reg_value)
{
    // Pass it on to the parent class.
    fiSensorPrivate::reg_update(reg_addr, reg_value);
}

/*!
 * @brief Set the options currently in use.
 * @param opt_value     The current GIGE_REG_SENSOR_OPTIONS value.
 */
void fiSensorLux19HS::set_options(uint32_t opt_value)
{
    // Save this for when we need to do things like max frame rate calculations.
    current_options = opt_value;
}

/*!
 * @brief Return true if we do pixel FPN for this sensor, false otherwise.
 */
bool fiSensorLux19HS::supports_pixel_fpn()
{
    return true;
}

/*!
 * @brief Return true if this sensor has a built-in FPN correction, false otherwise.
 */
bool fiSensorLux19HS::supports_sensor_fpn()
{
    return false;
}

/*!
 * @brief Return true if we have advanced calibration functions
 *        for this sensor, false otherwise.
 */
bool fiSensorLux19HS::supports_adv_cal()
{
    return false;
}

/*!
 * @brief Return true if this sensor (ICM) supports video trigger.
 */
bool fiSensorLux19HS::supports_video_trigger()
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << "FPGA:" << fpga_version << "ICM:" << icm_version;

    // The video trigger was added in FPGA version 1.0.5, ICM version 1.0.2.
    return ((QVersionNumber::compare(fpga_version, QVersionNumber(1, 0, 5)) >= 0) &&
            (QVersionNumber::compare(icm_version, QVersionNumber(1, 0, 2)) >= 0));
}

