/*!
 * @file fi_sensor_lux2100.cpp
 * @brief This class implements the sensor specific functions for the
 *        Lux2100 sensor.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QByteArray>
#include <QObject>
#include <qmath.h>

#include "fi_sensor_lux2100.h"

#include "gige_msg.h"
#include "logger.h"

namespace LUX2100_SENSOR
{
const int SENSOR_PCLK = 75000000;               //!< Sensor's clock rate, in MHz.

//======================================================================
// From the datasheet, pg 10, 7.1 "Exposure Timing"
const double Tclk = (1.0 / SENSOR_PCLK);        //!< 13.33 ns
const double Thblank = (2 * Tclk);              //!< "Horizontal blank" period, value from the Luxima framerate calculator spreadsheet.
const int    Twavetable = 66;                   //!< Value from the Luxima framerate calculator spreadsheet.

const int    PIXELS_PER_PPW = 32;               //!< Number of pixels per "pixel processing window" (from Lince5M terminology), aka "Number of Channels".
const int    LINE_READ = 1;                     //!< From the Luxima framerate calculator spreadsheet.
const double TXN = 500e-9;                      //!< The amount of time for the TXN pulse (nsec). Value from the Luxima framerate calculator spreadsheet.
const double TXLN = 300e-9;                     //!< Value from the Luxima framerate calculator spreadsheet.
const double NON_OVERLAP = 100e-9;              //!< Value from the Luxima framerate calculator spreadsheet.
const double SOF = 200e-9;                      //!< Value from the Luxima framerate calculator spreadsheet.

const int    MIN_ROW = (Twavetable + 3);        //!< From the Luxima framerate calculator spreadsheet.
const double VBLANK = (TXN + TXLN + NON_OVERLAP + SOF);

//======================================================================
}

using namespace LUX2100_SENSOR;
using namespace FI_SENSOR;

/*!
 * @brief Class constructor
 * @param parent        Pointer to our fiSensor base class.
 * @param sensor_id     Our sensor type ID.
 */
fiSensorLux2100::fiSensorLux2100(fiSensor *parent, uint32_t sensor_id) :
    fiSensorPrivate(parent, sensor_id),
    current_options(0)
{
}

/*!
 * @brief Class destructor.
 */
fiSensorLux2100::~fiSensorLux2100()
{
}

/*!
 * @brief Return true if the sensor is a color sensor, false for  mono.
 */
bool fiSensorLux2100::is_color()
{
    return (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR);
}

/*!
 * @brief Compute the maximum framerate for a given ROI.
 *
 * @param p_roi     Pointer to the ROI of interest.
 * @return The maximum frames-per-second allowed.
 */
uint32_t fiSensorLux2100::max_frame_rate(QRect *p_roi, uint32_t /* bit_depth */)
{
    uint32_t    frame_rate;
    int num_cols;
    int readout_startx;
    int readout_endx;
    double Tdata;
    double Tread;
    double Tframe;

    // Sanity check.
    if (p_roi == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null ROI!?";
        return 0;
    }

    // Where ACTUALLY is our starting X?
    readout_startx = (p_roi->x() / PIXELS_PER_PPW) * PIXELS_PER_PPW;

    // Where ACTUALLY does our ROI end?
    readout_endx = ((p_roi->x() + p_roi->width() + PIXELS_PER_PPW + 1) / PIXELS_PER_PPW) * PIXELS_PER_PPW;

    // How many columns are we talking about?
    num_cols = readout_endx - readout_startx;

    // Calculate our Tdata.
    Tdata = ceil(((double)num_cols) * ((double)LINE_READ) / ((double)PIXELS_PER_PPW)) * Tclk;

    // Calculate our Tread now. Make sure we're greater than or equal to the MIN_ROW value.
    Tread = Tdata + Thblank;
    if ((MIN_ROW * Tclk) > Tread)
        Tread = MIN_ROW * Tclk;

    // Calculate the time for a frame.
    Tframe = (Tread * ceil((double)p_roi->height() / (double)LINE_READ)) + VBLANK;

    // Calculate the maximum frame rate based on the sensor dimensions. Intentionally
    // round down, so the frame rate is within bounds.
    frame_rate = (uint32_t)floor(1.0 / Tframe);
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_rate:" << frame_rate << "Tframe:" << Tframe << "num_cols:" << num_cols << "Tdata:" << Tdata << "Tread:" << Tread;

    // Make sure our maximum rate fits within the limitations of what is allowed.
    if (frame_rate > sensor_max_frame_rate)
        frame_rate = sensor_max_frame_rate;

    // Return the rate.
    return frame_rate;
}

/*!
 * @brief Return the maximum size (number of bits) of a pixel on this sensor.
 */
int fiSensorLux2100::max_pixel_depth()
{
    return 12;
}

/*!
 * @brief Return the maximum ROI supported by the sensor for a given bit mode.
 *
 * @param bit_mode      The pixel depth selected by the user.
 * @return The maximum ROI supported by the sensor for the selected bit mode.
 */
QRect fiSensorLux2100::max_roi(uint32_t /* bit_mode */)
{
    QRect   max_roi(0, 0, sensor_max_width, sensor_max_height);

    // Return our value.
    return max_roi;
}

/*!
 * @brief Return true if binning and/or subsampling are supported.
 */
bool fiSensorLux2100::opts_avail()
{
    return true;
}

/*!
 * @brief Update registers of interest to the object.
 * @param reg_addr      The register address.
 * @param reg_value     The value of the register.
 */
void fiSensorLux2100::reg_update(uint32_t reg_addr, uint32_t reg_value)
{
    // Pass it on to the parent class.
    fiSensorPrivate::reg_update(reg_addr, reg_value);
}

/*!
 * @brief Set the options currently in use.
 * @param opt_value     The current GIGE_REG_SENSOR_OPTIONS value.
 */
void fiSensorLux2100::set_options(uint32_t opt_value)
{
    // Save this for when we need to do things like max frame rate calculations.
    current_options = opt_value;
}

/*!
 * @brief Return true if we do pixel FPN for this sensor, false otherwise.
 */
bool fiSensorLux2100::supports_pixel_fpn()
{
    return true;
}

/*!
 * @brief Return true if this sensor has a built-in FPN correction, false otherwise.
 */
bool fiSensorLux2100::supports_sensor_fpn()
{
    return false;
}

/*!
 * @brief Return true if we have advanced calibration functions
 *        for this sensor, false otherwise.
 */
bool fiSensorLux2100::supports_adv_cal()
{
    return true;
}
