/*!
 * @file fi_slider.cpp
 * @brief This file contains the implementation of the fiSlider class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QtGui>
#include <QApplication>
#include <QKeyEvent>
#include <QMouseEvent>
#include <QPaintEvent>
#include <QPalette>
#include <QStyleOptionSlider>
#include <QStylePainter>
#include <QToolTip>

#include "fi_slider.h"
#include "logger.h"

using namespace FI_SLIDER;

/*!
 * @brief Class constructor.
 */
fiSlider::fiSlider(QWidget *parent) :
    QSlider(parent)
{
    initClass();
}

/*!
 * @brief Class constructor providing the orientation of the slider.
 */
fiSlider::fiSlider(Qt::Orientation orientation, QWidget *parent) :
    QSlider(orientation, parent)
{
    initClass();
}

/*!
 * @brief Class destructor.
 */
fiSlider::~fiSlider()
{
}

/*!
 * @brief Initialize our class object.
 */
void fiSlider::initClass(void)
{
    // Initialize our members.
    event_color = palette().dark().color();
    event_markers.clear();
    update_period = 0;
    throttle_id = 0;

    // Connect our signals/slots.
    connect(this, &QSlider::sliderPressed, this, &fiSlider::handleSliderPressed);
    connect(this, &QSlider::sliderReleased, this, &fiSlider::handleSliderReleased);
    connect(this, &QSlider::sliderReleased, this, &fiSlider::released);
}

/*!
 * @brief Return the list of event markers.
 */
const QList<int>& fiSlider::markers() const
{
    return event_markers;
}

/*!
 * @brief Set the color we want to associate with the event markers.
 * @param[in] newColor  The color to use when drawing the event markers.
 */
void fiSlider::setEventColor(const QColor& newColor)
{
    event_color = newColor;
}

/*!
 * @brief Set or clear the event marker locations.
 * @param[in] newMarkers    Set (or clear, if @newMarkers is empty) the locations for the events.
 */
void fiSlider::setEventMarkers(const QList<int>& newMarkers)
{
    // Clear out our current copy of markers.
    event_markers.clear();

    // Copy the markers.
    for (int i = 0; i < newMarkers.size(); ++i)
        event_markers.append(newMarkers.at(i));

    // Update the display.
    update();
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Handle the paint event.
 * @param event     Unused.
 */
void fiSlider::paintEvent(QPaintEvent *event)
{
    // Are we using markers?
    if (!event_markers.isEmpty())
    {
        QPen    marker_pen(event_color);
        QStylePainter   painter(this);
        QStyleOptionSlider opt;

        // Initialize our style object.
        initStyleOption(&opt);
        opt.subControls = QStyle::SC_SliderGroove;

        // Have the system draw the basic slider object.
        painter.drawComplexControl(QStyle::CC_Slider, opt);

        // Do we have event markers?
        if (!event_markers.isEmpty())
        {
            int i;

            // Set up our pen for drawing event markers.
            painter.setPen(marker_pen);

            // Loop through all the event markers.
            for (i = 0; i < event_markers.size(); ++i)
            {
                int     event_x;

                // Make sure our marker is in range.
                if (event_markers.at(i) < minimum())
                    event_x = minimum();
                else if (event_markers.at(i) > maximum())
                    event_x = maximum();
                else
                    event_x = event_markers.at(i);

                // Get the marker's location in the control.
                int marker_x = style()->sliderPositionFromValue(minimum(), maximum(), event_x, width());

                // Make sure we're in the range for drawing.
                if (marker_x >= 0)
                {
                    // See if we're drawing a run of events.
                    if ((event_markers.size() > (i + 1)) && (event_markers.at(i + 1) == INT_MAX))
                    {
                        // Sanity check.
                        if (event_markers.size() > (i + 2))
                        {
                            int marker_end_x;
                            int end_marker = event_markers.at(i + 2);

                            // Get the ending marker.
                            if (end_marker < minimum())
                                marker_end_x = style()->sliderPositionFromValue(minimum(), maximum(), minimum(), width());
                            else if (end_marker > maximum())
                                marker_end_x = style()->sliderPositionFromValue(minimum(), maximum(), maximum(), width());
                            else
                                marker_end_x = style()->sliderPositionFromValue(minimum(), maximum(), end_marker, width());

                            // Don't bother drawing this if, because of the min/max values of our slider,
                            // the start and end locations are the same (typically either 0 or max.)
                            if (marker_x < marker_end_x)
                            {
                                // Draw our region.
                                painter.setBrush(Qt::BDiagPattern);
                                painter.drawRect(marker_x, 1, marker_end_x - (marker_x + 1), height() - 2);
                            }

                            // Bump our index.
                            i += 2;
                        }
                    }

                    // Don't bother drawing this if, because of the min/max values of our slider,
                    // the marker is at 0.
                    else if (marker_x != 0)
                    {
                        // Draw the marker.
                        painter.drawLine(marker_x, 0, marker_x, height());
                    }
                }
            }
        }

        // Now we want to draw the handle.
        opt.subControls = QStyle::SC_SliderHandle;
        painter.drawComplexControl(QStyle::CC_Slider, opt);
    }
    else
    {
        // Not using markers, so pass on the Paint event.
        QSlider::paintEvent(event);
    }
}

/*!
 * @brief Respond to a timer expiring.
 * @param event     Pointer to the timer event.
 */
void fiSlider::timerEvent(QTimerEvent *event)
{
    // Is this our playback update timer?
    if (event->timerId() != throttle_id)
        return;

    // It's time to update our current position.
    emit valueSlowChanged(value());
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has pressed the slider. We want to start a timer to
 *        control the number of updates.
 */
void fiSlider::handleSliderPressed()
{
    // Do we already have an active timer?
    if (throttle_id != 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Already have a throttle_id?!";
        killTimer(throttle_id);
        throttle_id = 0;
    }

    // Are we interested in throttling the response?
    if (update_period == 0)
        return;

    // We're interested in throttling back the response update of the slider.
    throttle_id = startTimer(update_period);
}

/*!
 * @brief The user has released the slider, so we can stop the timer.
 */
void fiSlider::handleSliderReleased()
{
    // Sanity check.
    if (throttle_id != 0)
    {
        // Kill the timer which will stop the updates.
        killTimer(throttle_id);
        throttle_id = 0;
    }
}
