/*!
 * @file fi_spanslider.cpp
 * @brief This file contains the implementation of the fiSpanSlider class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <algorithm>

#include <QtGui>
#include <QApplication>
#include <QKeyEvent>
#include <QMouseEvent>
#include <QPalette>
#include <QStyleOptionSlider>
#include <QStylePainter>
#include <QToolTip>

#include "fi_spanslider.h"
#include "logger.h"

namespace FI_SPAN_SLIDER
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
const int   HOVER_DELTA         = 2;        //!< The amount of slop (+/-) around the mouse for a hover event.
const int   MAX_ZOOM_FACTOR     = 16;       //!< The maximum amount (2^MAX_ZOOM_FACTOR) of zoom we support.
const int   MIN_FRAME_COUNT     = 3;        //!< The minimum number of frames we need to have in our zoom window.
}

using namespace FI_SPAN_SLIDER;

/*!
 * @class fiSpanSlider
 * @brief fiSpanSlider is a slider with two handles. fiSpanSlider is
 *        handy for letting user to choose an span between min/max.
 *
 * The span color is calculated based on QPalette::Highlight.
 *
 * The keys are bound according to the following table:
 * |   Orientation  |      Key      | Handle |
 * |----------------|---------------|--------|
 * | Qt::Horizontal | Qt::Key_Left  | lower  |
 * | Qt::Horizontal | Qt::Key_Right | lower  |
 * | Qt::Horizontal | Qt::Key_Up    | upper  |
 * | Qt::Horizontal | Qt::Key_Down  | upper  |
 * | Qt::Vertical   | Qt::Key_Up    | lower  |
 * | Qt::Vertical   | Qt::Key_Down  | lower  |
 * | Qt::Vertical   | Qt::Key_Left  | upper  |
 * | Qt::Vertical   | Qt::Key_Right | upper  |
 *
 * Keys are bound by the time the slider is created. A key is bound
 * to same handle for the lifetime of the slider. So even if the handle
 * representation might change from lower to upper, the same key binding
 * remains.
 *
 * @note fiSpanSlider inherits QSlider for implementation specific
 *       reasons. Adjusting any single handle specific properties like:
 *         - QAbstractSlider::sliderPosition
 *         - QAbstractSlider::value
 * @note has no effect. However, all slider specific properties like:
 *         - QAbstractSlider::invertedAppearance
 *         - QAbstractSlider::invertedControls
 *         - QAbstractSlider::minimum
 *         - QAbstractSlider::maximum
 *         - QAbstractSlider::orientation
 *         - QAbstractSlider::pageStep
 *         - QAbstractSlider::singleStep
 *         - QSlider::tickInterval
 *         - QSlider::tickPosition
 * @note are taken into consideration.
 */

/*!
 * @brief Class constructor.
 */
fiSpanSlider::fiSpanSlider(QWidget *parent) :
    QSlider(parent)
{
    init_class();
}

/*!
 * @brief Class constructor providing the orientation of the slider.
 */
fiSpanSlider::fiSpanSlider(Qt::Orientation orientation, QWidget *parent) :
    QSlider(orientation, parent)
{
    init_class();
}

/*!
 * @brief Class destructor.
 */
fiSpanSlider::~fiSpanSlider()
{
}

/*!
 * @brief Draw the SliderHandle of the slider (lower or upper specified by the @a handle parameter).
 *
 * @param[in] painter   Pointer to the painter object we are to use for the painting.
 * @param[in] handle    Which handle we're drawing, lower or upper.
 */
void fiSpanSlider::drawHandle(QStylePainter *painter, FI_SPAN_SLIDER::SPAN_HANDLE handle) const
{
    QStyleOptionSlider opt;

    // Initialize the style option object for our handle.
    initStyleOption(&opt, handle);

    // Specify that we're only looking at the slider handle.
    opt.subControls = QStyle::SC_SliderHandle;

    // Which handle are we talking about? If "handle" is specifying the lower handle, then
    // this will set "pressed" to be the appropriate QStyle::SubControl value. If the handle
    // we are drawing isn't the handle being selected, then this will have the SC_None value.
    QStyle::SubControl pressed = (handle == LowerHandle) ? lowerPressed : upperPressed;
    if (pressed == QStyle::SC_SliderHandle)
    {
        // Our handle is being "pressed", so show it in a "sunken" style.
        opt.activeSubControls = pressed;
        opt.state |= QStyle::State_Sunken;
    }

    // Draw the handle.
    painter->drawComplexControl(QStyle::CC_Slider, opt);
}

/*!
 * @brief Draw the span between the handles.
 *
 * @param[in] painter   Pointer to the painter object we are to use for the painting.
 * @param[in] rect      The area between the handles.
 */
void fiSpanSlider::drawSpan(QStylePainter *painter, const QRect& rect) const
{
    QStyleOptionSlider opt;

    // Initialize our style option object.
    initStyleOption(&opt, FI_SPAN_SLIDER::UpperHandle);

    // Figure out the area of the standard slider's "groove".
    QRect groove = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderGroove, this);
    if (opt.orientation == Qt::Horizontal)
        groove.adjust(0, 0, -1, 0);
    else
        groove.adjust(0, 0, 0, -1);

    // Initialize our painter object based on our orientation.
    if (opt.orientation == Qt::Horizontal)
        setupPainter(painter, opt.orientation, groove.center().x(), groove.top(), groove.center().x(), groove.bottom());
    else
        setupPainter(painter, opt.orientation, groove.left(), groove.center().y(), groove.right(), groove.center().y());

    // Draw our span where it intersects with the slider's groove.
    painter->drawRect(rect.intersected(groove));
}

/*!
 * @brief Retrieve the starting frame number for the highlighted region. If there
 *        is no highlighted region, then we return a 0.
 */
int fiSpanSlider::get_zoom_start_frame()
{
    int     frame_num;

    // Do we have a high lighter?
    if (hilighter.isEmpty())
        frame_num = 0;
    else
        frame_num = zoom_start;

    // Return our frame number.
    return frame_num;
}

/*!
 * @brief Retrieve the ending frame number for the highlighted region. If there
 *        is no highlighted region, then we return a 0.
 */
int fiSpanSlider::get_zoom_stop_frame()
{
    int     frame_num;

    // Do we have a high lighter?
    if (hilighter.isEmpty())
        frame_num = 0;
    else
        frame_num = zoom_stop;

    // Return our frame number.
    return frame_num;
}

/*!
 * @brief Check to see if the handle specified by the @a handle parameter is being selected.
 *
 * @param[in]       pos         The location we're checking.
 * @param[in,out]   control     The SubControl we're looking at (going in) and the SC_SliderHandle or SC_None (going out).
 * @param[in] value     The current position of the @a handle.
 * @param[in] handle    The handle (lower or upper) that we're interested in.
 */
void fiSpanSlider::handleMousePress(const QPoint& pos, QStyle::SubControl& control, int value, FI_SPAN_SLIDER::SPAN_HANDLE handle)
{
    QStyleOptionSlider opt;
    const QStyle::SubControl oldControl = control;

    // Intialize our style option object for the handle.
    initStyleOption(&opt, handle);

    // See if the position passed in hits on the handle.
    control = style()->hitTestComplexControl(QStyle::CC_Slider, &opt, pos, this);

    // Get the area defined by the handle.
    const QRect sr = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderHandle, this);

    // Did we get a hit on the control?
    if (control == QStyle::SC_SliderHandle)
    {
        // Remember the positional information for the handle.
        position = value;
        offset = pick(pos - sr.topLeft());

        // Remember that this handle (lower or upper) is being pressed.
        lastPressed = handle;

        // Let our parent know that the slider is being pressed.
        setSliderDown(true);

        // Let anyone else who might want to know that the slider is being pressed.
        emit sliderPressed(handle);
    }

    // If the control changed (from being on a slider to off, or vice-versa), update
    // the region.
    if (control != oldControl)
        update(sr);
}

/*!
 * @brief Return the allowed movement mode we're currently set to.
 * @sa FI_SPAN_SLIDER::HANDLE_MOVEMENT_MODE
 */
FI_SPAN_SLIDER::HANDLE_MOVEMENT_MODE fiSpanSlider::handleMovementMode() const
{
    return movement;
}

/*!
 * @brief Initialize our class object.
 */
void fiSpanSlider::init_class()
{
    // Initialize our values.
    event_color = palette().dark().color();
    event_markers.clear();
    trigger_color = Qt::red;
    lower = 0;
    upper = 0;
    lowerPos = 0;
    upperPos = 0;
    offset = 0;
    position = 0;
    lastPressed = NoHandle;
    mainControl = LowerHandle;
    lowerPressed = QStyle::SC_None;
    upperPressed = QStyle::SC_None;
    movement = FreeMovement;
    firstMovement = false;
    blockTracking = false;

    // In general, the trigger frame is always frame 0.
    trigger_pos = 0;

    // Not showing a highlighted area.
    hilighter = QRect(0, 0, 0, 0);
    zoom_factor = 0;
    hilight_in_motion = false;
    last_pos = QPoint(0, 0);

    // By default, we don't want any tick marks.
    setTickPosition(QSlider::NoTicks);

    // Make sure we track the mouse.
    setMouseTracking(true);

    // Setup our signal/slots.
    connect(this, &fiSpanSlider::rangeChanged, this, &fiSpanSlider::updateRange);
    connect(this, &fiSpanSlider::sliderReleased, this, &fiSpanSlider::movePressedHandle);
}

/*!
 * @brief Utility function to initialize the StyleOption object.
 *
 * @param[in,out] option    Pointer to the StyleOption object we want to initialize.
 * @param[in] handle        Which handle (lower or upper) we're initializing for.
 */
void fiSpanSlider::initStyleOption(QStyleOptionSlider *option, FI_SPAN_SLIDER::SPAN_HANDLE handle) const
{
    // Have the slider initialize our option structure.
    QSlider::initStyleOption(option);

    // Update the slider position and value based on the handle information.
    option->sliderPosition = (handle == LowerHandle) ? lowerPos : upperPos;
    option->sliderValue = (handle == LowerHandle) ? lower : upper;
}

/*!
 * @brief Retrieve the lower handle position.
 * @sa fiSpanSlider::lowerValue()
 */
int fiSpanSlider::lowerPosition() const
{
    return lowerPos;
}

/*!
 * @brief Retrieve the lower handle value.
 * @sa fiSpanSlider::lowerPosition()
 */
int fiSpanSlider::lowerValue() const
{
    return qMin(lower, upper);
}

/*!
 * @brief Return the list of event markers.
 */
const QList<int>& fiSpanSlider::markers() const
{
    return event_markers;
}

/*!
 * @brief Painter our slider.
 * @param[in] painter   Pointer to our painter object.
 */
void fiSpanSlider::paint_slider(QStylePainter *painter)
{
    QStyleOptionSlider opt;

    // Initialize our style object.
    initStyleOption(&opt, FI_SPAN_SLIDER::UpperHandle);
    opt.sliderValue = opt.minimum;
    opt.sliderPosition = opt.minimum;
    opt.subControls = QStyle::SC_SliderGroove;

    // Have the system draw the basic slider object.
    painter->drawComplexControl(QStyle::CC_Slider, opt);

    // Initialize the regional information for the lower handle.
    opt.sliderPosition = lowerPos;
    const QRect lr = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderHandle, this);
    const int lrv  = pick(lr.center());

    // Initialize the regional information for the upper handle.
    opt.sliderPosition = upperPos;
    const QRect ur = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderHandle, this);
    const int urv  = pick(ur.center());

    // Calculate the region between the two handles (aka, the "span").
    const int minv = qMin(lrv, urv);
    const int maxv = qMax(lrv, urv);
    const QPoint c = QRect(lr.center(), ur.center()).center();
    QRect spanRect;
    if (orientation() == Qt::Horizontal)
        spanRect = QRect(QPoint(minv, c.y() - 2), QPoint(maxv, c.y() + 1));
    else
        spanRect = QRect(QPoint(c.x() - 2, minv), QPoint(c.x() + 1, maxv));

    // Draw the span between the two handles.
    drawSpan(painter, spanRect);

    // Make sure the trigger gets seen. Paint it so it's "on top" of the groove and tick marks.
    if ((trigger_pos >= minimum()) && (trigger_pos <= maximum()))
    {
        int         trig_draw_x;
        QPen        trigger_pen(triggerColor());

        // Figure out where we need to draw our marker.
        trig_draw_x = style()->sliderPositionFromValue(minimum(), maximum(), trigger_pos, width());

        // Make sure we're in the range for drawing.
        if (trig_draw_x >= 0)
        {
            painter->setPen(trigger_pen);
            painter->drawLine(trig_draw_x, 0, trig_draw_x, height());
        }
    }

    // Finally, draw the handles.
    switch (lastPressed)
    {
    case LowerHandle:
        drawHandle(painter, UpperHandle);
        drawHandle(painter, LowerHandle);
        break;
    case UpperHandle:
    default:
        drawHandle(painter, LowerHandle);
        drawHandle(painter, UpperHandle);
        break;
    }
}

/*!
 * @brief Utility function to convert a given pixel position to a slider value.
 *
 * @param[in] pos   The pixel position we're interested in.
 * @return The slider value assoicated with the pixel position.
 */
int fiSpanSlider::pixelPosToRangeValue(int pos) const
{
    int sliderMin;
    int sliderMax;
    int sliderLength;
    QStyleOptionSlider opt;

    // Initialize our style object.
    initStyleOption(&opt, FI_SPAN_SLIDER::UpperHandle);

    // Get the region of the slider and the region of the handle.
    const QRect gr = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderGroove, this);
    const QRect sr = style()->subControlRect(QStyle::CC_Slider, &opt, QStyle::SC_SliderHandle, this);

    // If we're horizontal then our target area is different (short and wide) than
    // when we're vertical (tall and skinny).
    if (orientation() == Qt::Horizontal)
    {
        sliderLength = sr.width();
        sliderMin = gr.x();
        sliderMax = gr.right() - sliderLength + 1;
    }
    else
    {
        sliderLength = sr.height();
        sliderMin = gr.y();
        sliderMax = gr.bottom() - sliderLength + 1;
    }

    // Now get the slider value based on the position (relative to the slider, not the parent window).
    return QStyle::sliderValueFromPosition(minimum(), maximum(), pos - sliderMin,
                                           sliderMax - sliderMin, opt.upsideDown);
}

/*!
 * @brief Set the color we want to associate with the event markers.
 * @param[in] newColor  The color to use when drawing the event markers.
 */
void fiSpanSlider::setEventColor(const QColor& newColor)
{
    event_color = newColor;
}

/*!
 * @brief Set or clear the event marker locations.
 * @param[in] newMarkers    Set (or clear, if @newMarkers is empty) the locations for the events.
 */
void fiSpanSlider::setEventMarkers(const QList<int>& newMarkers)
{
    // Clear out our current copy of markers.
    event_markers.clear();

    // Copy the markers.
    for (int i = 0; i < newMarkers.size(); ++i)
        event_markers.append(newMarkers.at(i));

    // Update the display.
    update();
}

/*!
 * @brief Sets the span from @a new_lower to @a new_upper (slider position values, not pixels).
 *
 * @param[in] new_lower     The new lower bound for the span.
 * @param[in] new_upper     The new upper bound for the span.
 */
void fiSpanSlider::setSpan(int new_lower, int new_upper)
{
    // Sanity check the bounds.
    const int low = qBound(minimum(), qMin(new_lower, new_upper), maximum());
    const int upp = qBound(minimum(), qMax(new_lower, new_upper), maximum());

    // Make sure there's a change.
    if ((low != lower) || (upp != upper))
    {
        // Something changed, was it the lower bound?
        if (low != lower)
        {
            // Remember the new lower setting and emit a signal letting folks
            // know that it changed.
            lower = low;
            lowerPos = low;
            emit lowerValueChanged(low);
        }

        // Did the upper bound change?
        if (upp != upper)
        {
            // Remember the new upper setting and emit a signal letting folks
            // know that it changed.
            upper = upp;
            upperPos = upp;
            emit upperValueChanged(upp);
        }

        // Finally, regardless of if it was the upper or lower bound, let folks know
        // that the span changed.
        emit spanChanged(lower,  upper);

        // Update the display.
        update();
    }
}

/*!
 * @brief Set the color we want to associate with the trigger mark on the slider.
 * @param[in] newColor  The color to use when drawing the trigger mark.
 */
void fiSpanSlider::setTriggerColor(const QColor& newColor)
{
    trigger_color = newColor;
}

/*!
 * @brief   Modify the trigger position. Normally the trigger is frame 0,
 *          but there might be cases where we need to change that. For
 *          example, if there ISN'T a trigger frame (user canceled the capture).
 */
void fiSpanSlider::setTriggerPos(int newPos)
{
    // Save the new position.
    trigger_pos = newPos;

    // Update the display.
    update();
}

/*!
 * @brief Initialize our painter to draw the span between the two handles.
 *
 * We construct a linear gradient with interpolation area between (x1, y1)
 * and (x2, y2) (these are pixel coordinates). The gradient transitions
 * from a slightly darker value of QPalette::Highlight to a slightly lighter
 * value.
 *
 * We use a darker value of the highlight for our QPen object to draw an
 * outline around the span.
 *
 * @param[in] painter       Pointer to the painter object we are to use for the painting.
 * @param[in] orientation   Is the slider vertical or horizontal.
 * @param[in] x1    This is the x1 parameter for the QLineraGradient object.
 * @param[in] y1    This is the y1 parameter for the QLineraGradient object.
 * @param[in] x2    This is the x2 parameter for the QLineraGradient object.
 * @param[in] y2    This is the y2 parameter for the QLineraGradient object.
 */
void fiSpanSlider::setupPainter(QPainter *painter, Qt::Orientation orientation, qreal x1, qreal y1, qreal x2, qreal y2) const
{
    QColor highlight = palette().color(QPalette::Highlight);
    QLinearGradient gradient(x1, y1, x2, y2);

    // Setup our gradient's colors.
    gradient.setColorAt(0, highlight.darker(120));
    gradient.setColorAt(1, highlight.lighter(120));

    // Make our brush be the gradient.
    painter->setBrush(gradient);

    // Use a slightly darker pen when the slider is vertical than
    // if it's in a horizontal orientation.
    if (orientation == Qt::Horizontal)
        painter->setPen(QPen(highlight.darker(130), 0));
    else
        painter->setPen(QPen(highlight.darker(150), 0));
}

/*!
 * @brief   Allow a caller to set the center (x only) position for our
 *          highlighter region. Coordinates are in slider "units", probably
 *          frames in our case.
 *
 *          We'll do the best we can, limited by the extents of the slider.
 *
 * @param   center_x    The location the caller wants to be the center of the
 *                      highlighter region.
 */
void fiSpanSlider::set_zoom_center(int center_x)
{
    int new_x;
    int     width_count;

    // Figure out how many frames we'll be covering.
    width_count = ((maximum() - minimum()) >> zoom_factor);

    // Adjust our start/stop values.
    zoom_start = center_x - (width_count / 2);
    if (zoom_start < minimum())
        zoom_start = minimum();
    if ((zoom_start + width_count) > maximum())
        zoom_start = maximum() - width_count;
    zoom_stop = zoom_start + width_count;

    // Get the new X position in windows coordinates.
    new_x = style()->sliderPositionFromValue(minimum(), maximum(), center_x, size().width());

    // Adjust the X value as necessary to stay in the slider confines.
    if ((new_x - (hilighter.width() / 2)) < 0)
        new_x = hilighter.width() / 2;
    if ((new_x + (hilighter.width() / 2)) > size().width())
        new_x = size().width() - ((hilighter.width() / 2) + 1);

    // Adjust our highlighter region.
    hilighter.moveCenter(QPoint(new_x, hilighter.center().y()));
    update();
}

/*!
 * @brief Swap the lower and upper handles.
 */
void fiSpanSlider::swapControls()
{
    std::swap(lower, upper);
    std::swap(lowerPressed, upperPressed);

    lastPressed = (lastPressed == LowerHandle) ? UpperHandle : LowerHandle;
    mainControl = (mainControl == LowerHandle) ? UpperHandle : LowerHandle;
}

/*!
 * @brief Update the slider based on the @a action being processed.
 *
 * @param[in] action    The action we're processing.
 * @param[in] main_control_sel  True if the "main" or "primary" control is selected.
 */
void fiSpanSlider::triggerAction(QAbstractSlider::SliderAction action, bool main_control_sel)
{
    int value = 0;
    bool take_action = true;
    bool working_with_upper = false;

    const int min = minimum();
    const int max = maximum();
    const FI_SPAN_SLIDER::SPAN_HANDLE altControl = (mainControl == LowerHandle) ? UpperHandle : LowerHandle;

    // Lock out other routines from tracking, so we can avoid recursions.
    blockTracking = true;

    // What action are we working on?
    switch (action)
    {
    case QAbstractSlider::SliderSingleStepAdd:
        if ((main_control_sel && (mainControl == UpperHandle)) || (!main_control_sel && (altControl == UpperHandle)))
        {
            value = qBound(min, upper + singleStep(), max);
            working_with_upper = true;
        }
        else
        {
            value = qBound(min, lower + singleStep(), max);
        }
        break;
    case QAbstractSlider::SliderSingleStepSub:
        if ((main_control_sel && (mainControl == UpperHandle)) || (!main_control_sel && (altControl == UpperHandle)))
        {
            value = qBound(min, upper - singleStep(), max);
            working_with_upper = true;
        }
        else
        {
            value = qBound(min, lower - singleStep(), max);
        }
        break;
    case QAbstractSlider::SliderToMinimum:
        value = min;
        if ((main_control_sel && (mainControl == UpperHandle)) || (!main_control_sel && (altControl == UpperHandle)))
            working_with_upper = true;
        break;
    case QAbstractSlider::SliderToMaximum:
        value = max;
        if ((main_control_sel && (mainControl == UpperHandle)) || (!main_control_sel && (altControl == UpperHandle)))
            working_with_upper = true;
        break;
    case QAbstractSlider::SliderMove:
        if ((main_control_sel && (mainControl == UpperHandle)) || (!main_control_sel && (altControl == UpperHandle)))
            working_with_upper = true;

        //
        // Fall through intentional...
        //
        [[fallthrough]];

    case QAbstractSlider::SliderNoAction:
        take_action = false;
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unknown action";
        break;
    }

    // Are we going to take action?
    if (take_action)
    {
        // Are we working with the upper handle?
        if (working_with_upper)
        {
            // Figure out the value based on how we're allowed to move.
            switch (movement)
            {
            case NoCrossing:
            case NoOverlapping:
                break;
            case FreeMovement:
                if (value < lower)
                {
                    swapControls();
                    setLowerPosition(value);
                }
                break;
            default:
                setUpperPosition(value);
                break;
            }
        }
        else
        {
            // Figure out the value based on how we're allowed to move.
            switch (movement)
            {
            case NoCrossing:
            case NoOverlapping:
                break;
            case FreeMovement:
                if (value > upper)
                {
                    swapControls();
                    setUpperPosition(value);
                }
                break;
            default:
                setLowerPosition(value);
                break;
            }
        }
    }

    // Don't need to block others from updating.
    blockTracking = false;

    // Save our new positions.
    setLowerValue(lowerPos);
    setUpperValue(upperPos) ;
}

/*!
 * @brief Set the span range.
 *
 * @param min   Unused.
 * @param max   Unused.
 */
void fiSpanSlider::updateRange(int min, int max)
{
    Q_UNUSED(min);
    Q_UNUSED(max);

    // setSpan() takes care of keeping span in range
    setSpan(lower, upper);
}

/*!
 * @brief Update the zoom highlighter.
 * @param[in] shrink_it     True if we're making it smaller, false if it's growing.
 */
void fiSpanSlider::update_zoom(bool shrink_it)
{
    int center_x;
    int shift_amount;
    int     width_count;

    // Adjust the hilight region...
    if (hilighter.isEmpty())
        return;

    // Are we shrinking or growing?
    if (shrink_it)
    {
        // Make sure we're not at our limit.
        if ((zoom_factor + 1) == MAX_ZOOM_FACTOR)
            return;

        // Don't zoom in any closer than a minimum number of frames.
        if (((maximum() - minimum()) >> (zoom_factor + 1)) < MIN_FRAME_COUNT)
            return;

        // Increment our zoom factor to take more out of our width,
        // thus causing our region to shrink.
        ++zoom_factor;
    }
    else
    {
        // Make sure we're not at our limit.
        if (zoom_factor == 1)
            return;

        // Reduce our zoom factor to take less out of our width,
        // thus causing our region to grow.
        --zoom_factor;
    }

    // Center on the current position.
    center_x = style()->sliderPositionFromValue(minimum(), maximum(), value(), size().width());

    // Figure out how big we should be based on our zoom factor.
    shift_amount = size().width() >> zoom_factor;

    // Split that in half, 'cuz we're adjust each side towards/away to/from the center.
    shift_amount = (shift_amount / 2);

    // Check for a VERY small (aka '0') shift amount
    if (shift_amount > 0)
    {
        // Adjust our region.
        hilighter.setRight(center_x + shift_amount);
        hilighter.setLeft(center_x - shift_amount);
    }

    // Figure out how many frames we'll be covering.
    width_count = ((maximum() - minimum()) >> zoom_factor);

    // Now slide things around so it fits.
    if (hilighter.left() < 0)
    {
        // We want to to be left-justified.
        hilighter.moveRight(hilighter.width());

        // Since we're left-justified, we're starting at the minimum.
        zoom_start = minimum();
        zoom_stop = zoom_start + width_count;
    }

    else if (hilighter.right() > size().width())
    {
        // We want to be right-justified.
        hilighter.moveLeft(size().width() - (hilighter.width() + 1));

        // Since we're right-justified, we're ending at the maximum.
        zoom_stop = maximum();
        zoom_start = zoom_stop - width_count;
    }
    else
    {
        // We're somewhere in the middle.
        zoom_start = value() - (width_count / 2);
        if ((zoom_start + width_count) > maximum())
            zoom_start = maximum() - width_count;
        zoom_stop = zoom_start + width_count;
    }

    // Let folks know we're changing.
    emit zoom_changing();
}

/*!
 * @brief Retrieve the upper handle position.
 * @sa fiSpanSlider::upperValue()
 */
int fiSpanSlider::upperPosition() const
{
    return upperPos;
}

/*!
 * @brief Retrieve the upper handle position.
 * @sa fiSpanSlider::upperPosition()
 */
int fiSpanSlider::upperValue() const
{
    return qMax(lower,  upper);
}

/*!
 * @brief Return true if the highlighter region is visible.
 */
bool fiSpanSlider::zoom_is_showing()
{
    // If our high lighter region is not empty, then it's showing.
    return (hilighter.isEmpty() == false);
}

/*!
 * @brief Return the highlighter region.
 */
const QRect& fiSpanSlider::zoom_region() const
{
    return hilighter;
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Handle the tool tip event.
 * @param e     Our event pointer.
 */
bool fiSpanSlider::event(QEvent *e)
{
    // Make sure we're only looking at the tool tip events.
    if (e->type() == QEvent::ToolTip)
    {
        // Cast it so it's easier to use.
        QHelpEvent *helpEvent = static_cast<QHelpEvent *>(e);

        // Do we have a tool tip to display?
        if (toolTip().isEmpty())
        {
            // No tooltip, so hide the bubble and ignore the event.
            QToolTip::hideText();
            e->ignore();
        }
        else
        {
            // Show the tool tip. We generated the text in the mouseMoveEvent handler.
            QToolTip::showText(helpEvent->globalPos(), toolTip());
        }

        // All done.
        return true;
    }

    // Pass the event on to the parent.
    return QSlider::event(e);
}

/*!
 * @brief Handle the key press event.
 * @param event     Our event pointer.
 */
void fiSpanSlider::keyPressEvent(QKeyEvent *event)
{
    QSlider::keyPressEvent(event);

    bool main_control_sel = true;
    SliderAction action = SliderNoAction;
    switch (event->key())
    {
    case Qt::Key_Left:
        main_control_sel = (orientation() == Qt::Horizontal);
        action = !invertedAppearance() ? SliderSingleStepSub : SliderSingleStepAdd;
        break;
    case Qt::Key_Right:
        main_control_sel = (orientation() == Qt::Horizontal);
        action = !invertedAppearance() ? SliderSingleStepAdd : SliderSingleStepSub;
        break;
    case Qt::Key_Up:
        main_control_sel = (orientation() == Qt::Vertical);
        action = invertedControls() ? SliderSingleStepSub : SliderSingleStepAdd;
        break;
    case Qt::Key_Down:
        main_control_sel = (orientation() == Qt::Vertical);
        action = invertedControls() ? SliderSingleStepAdd : SliderSingleStepSub;
        break;
    case Qt::Key_Home:
        main_control_sel = (mainControl ==  LowerHandle);
        action = SliderToMinimum;
        break;
    case Qt::Key_End:
        main_control_sel = (mainControl ==  UpperHandle);
        action = SliderToMaximum;
        break;
    default:
        event->ignore();
        break;
    }

    if (action)
        triggerAction(action, main_control_sel);
}

/*!
 * @brief Handle the mouse double click event.
 * @param e     Our event pointer.
 */
void fiSpanSlider::mouseDoubleClickEvent(QMouseEvent *e)
{
    // Pass it on to the parent.
    QSlider::mouseDoubleClickEvent(e);

    // Do we want to show or hide the zoomer.
    if (hilighter.isEmpty())
    {
        // Initialize our zoom factor.
        zoom_factor = 0;

        // Setup our hilighter.
        hilighter = QRect(0, 0, size().width(), size().height());

        // Update our zoom.
        update_zoom(true);
        emit zoom_showing();
    }
    else
    {
        // Turn off the hilighter.
        hilighter = QRect(0, 0, 0, 0);
        emit zoom_hiding();
    }

    // Update our widget.
    update();
}

/*!
 * @brief Handle the mouse move event.
 * @param e     Our event pointer.
 */
void fiSpanSlider::mouseMoveEvent(QMouseEvent *e)
{
    QString tooltip_msg;
    QStyleOptionSlider opt;

    // Initialize our options object.
    initStyleOption(&opt, FI_SPAN_SLIDER::UpperHandle);

    // Is our highlight region in motion?
    if (hilight_in_motion)
    {
        int delta;

        // How much have we moved?
        delta = e->pos().x() - last_pos.x();

        if ((delta < 0) && ((hilighter.left() + delta) < 0))
            delta = -hilighter.left();
        if ((delta > 0) && ((hilighter.right() + delta) > size().width()))
            delta = size().width() - hilighter.right();

        // Can we move our region?
        if (((delta < 0) && ((hilighter.left() + delta) >= 0)) ||
            ((delta > 0) && ((hilighter.right() + delta) <= size().width())))
        {
            int center_frame;
            int width_count;

            // Figure out how many frames we'll be covering.
            width_count = ((maximum() - minimum()) >> zoom_factor);

            // Adjust our region.
            hilighter.adjust(delta, 0, delta, 0);

            // Update the control.
            update();

            // Calculate our new start/stop values.
            if (hilighter.left() == 0)
                zoom_start = minimum();
            else if (hilighter.right() == size().width())
                zoom_start = maximum() - width_count;
            else
            {
                center_frame = style()->sliderValueFromPosition(minimum(), maximum(), hilighter.center().x(), size().width());
                zoom_start = center_frame - (width_count / 2);
                if ((zoom_start + width_count) > maximum())
                    zoom_start = maximum() - width_count;
            }
            zoom_stop = zoom_start + width_count;

            // Remember the new position.
            last_pos = e->pos();

            // Let folks know we're moving.
            emit zoom_moving();
        }
    }

    // Empty our tool tip message.
    tooltip_msg.clear();

    // Do we have a trigger position.
    if ((trigger_pos >= minimum()) && (trigger_pos <= maximum()))
    {
        int trig_draw_x;

        // Figure out where we need to draw our marker.
        trig_draw_x = style()->sliderPositionFromValue(minimum(), maximum(), trigger_pos, size().width());
        if ((e->position().x() >= (trig_draw_x - HOVER_DELTA)) &&
            (e->position().x() <= (trig_draw_x + HOVER_DELTA)))
        {
            tooltip_msg = QString::fromUtf8("Trigger Frame");
        }
    }

    // Do we have event markers?
    if (!event_markers.isEmpty())
    {
        // Loop through all the event markers and see if any are close to the mouse.
        for (int i = 0; i < event_markers.size(); ++i)
        {
            // Get the marker's location in the control.
            int marker_x = style()->sliderPositionFromValue(minimum(), maximum(), event_markers.at(i), size().width());

            // See if we're drawing a run of events.
            if ((event_markers.size() > (i + 1)) && (event_markers.at(i + 1) == INT_MAX))
            {
                // Sanity check.
                if (event_markers.size() > (i + 2))
                {
                    // Get the ending marker.
                    int marker_end_x = style()->sliderPositionFromValue(minimum(), maximum(), event_markers.at(i + 2), size().width());

                    // Do we have a hit?
                    if ((e->position().x() >= (marker_x - HOVER_DELTA)) &&
                        (e->position().x() <= (marker_end_x + HOVER_DELTA)))
                    {
                        // Back into the frame number.
                        int frame_no = style()->sliderValueFromPosition(minimum(), maximum(), e->position().x(), size().width());
                        tooltip_msg = QString::fromUtf8("Frame %1").arg(frame_no);
                        break;
                    }

                    // Skip this run.
                    i += 2;
                }
            }

            // Is the mouse above this marker?
            else if ((e->position().x() >= (marker_x - HOVER_DELTA)) &&
                     (e->position().x() <= (marker_x + HOVER_DELTA)))
            {
                tooltip_msg = QString::fromUtf8("Frame %1").arg(event_markers.at(i));
                break;
            }
        }
    }

    // Set the tooltip message.
    setToolTip(tooltip_msg);

    // Make sure one of our handles was pressed.
    if ((lowerPressed != QStyle::SC_SliderHandle) && (upperPressed != QStyle::SC_SliderHandle))
    {
        e->ignore();
        return;
    }

    const int m = style()->pixelMetric(QStyle::PM_MaximumDragDistance, &opt, this);
    int newPosition =  pixelPosToRangeValue(pick(e->pos()) - offset);

    if (m >= 0)
    {
        const QRect r = rect().adjusted(-m, -m, m, m);
        if (!r.contains(e->pos()))
        {
            newPosition = position;
        }
    }

    // pick the preferred handle on the first movement
    if (firstMovement)
    {
        if (lower == upper)
        {
            if (newPosition < lowerValue())
            {
                swapControls();
                firstMovement = false;
            }
        }
        else
        {
            firstMovement = false;
        }
    }

    if (lowerPressed == QStyle::SC_SliderHandle)
    {
        if (movement == NoCrossing)
        {
            newPosition = qMin(newPosition, upperValue());
        }
        else if (movement == NoOverlapping)
        {
            newPosition = qMin(newPosition, upperValue() - 1);
        }

        if ((movement == FreeMovement) && (newPosition > upper))
        {
            swapControls();
            setUpperPosition(newPosition);
        }
        else
        {
            setLowerPosition(newPosition);
        }
    }
    else if (upperPressed == QStyle::SC_SliderHandle)
    {
        if (movement == NoCrossing)
        {
            newPosition = qMax(newPosition, lowerValue());
        }
        else if (movement == NoOverlapping)
        {
            newPosition = qMax(newPosition, lowerValue() + 1);
        }

        if ((movement == FreeMovement) && (newPosition <  lower))
        {
            swapControls();
            setLowerPosition(newPosition);
        }
        else
        {
            setUpperPosition(newPosition);
        }
    }

    e->accept();
}

/*!
 * @brief Got a mouse button down event. Check to see if the user is selecting
 *        the highlighted zoom region or one of the handles.
 *
 * @param[in] e     Pointer to the event object.
 */
void fiSpanSlider::mousePressEvent(QMouseEvent *e)
{
    // Sanity check.
    if ((minimum() == maximum()) || (e->buttons() ^ e->button()))
    {
        e->ignore();
        return;
    }

    // Only interested in left-mouse button clicks.
    if (e->button() == Qt::LeftButton)
    {
        // Only interested if we have a non-empty high light region.
        if (!hilighter.isEmpty())
        {
            // Did we hit the high lighter region?
            if (hilighter.contains(e->pos()))
            {
                // Remember we're in motion and our current position.
                hilight_in_motion = true;
                last_pos = e->pos();
            }
        }
    }

    // Check to see if we're processing the upper handle.
    handleMousePress(e->pos(), upperPressed, upper, UpperHandle);

    // If the upper handle isn't being selected, check the lower handle.
    if (upperPressed != QStyle::SC_SliderHandle)
        handleMousePress(e->pos(), lowerPressed, lower, LowerHandle);

    // Remember that we have a mouse down event.
    firstMovement = true;

    // Accept the event.
    e->accept();
}

/*!
 * @brief Handle the mouse release event.
 * @param e     Our event pointer.
 */
void fiSpanSlider::mouseReleaseEvent(QMouseEvent *e)
{
    // Are we done moving the highlight region?
    if (hilight_in_motion)
    {
        // All done.
        hilight_in_motion = false;
    }

    // Pass it on to the parent.
    QSlider::mouseReleaseEvent(e);

    setSliderDown(false);

    lowerPressed = QStyle::SC_None;
    upperPressed = QStyle::SC_None;

    update();
}

/*!
 * @brief Handle the paint event.
 * @param e     Unused.
 */
void fiSpanSlider::paintEvent(QPaintEvent *e)
{
    QPen    marker_pen(eventColor());
    QStylePainter   painter(this);

    // Q_UNUSED( e );
    e->accept();

    // Do we have event markers?
    if (!event_markers.isEmpty())
    {
        // Set up our pen for drawing event markers.
        painter.setPen(marker_pen);

        // Loop through all the event markers.
        for (int i = 0; i < event_markers.size(); ++i)
        {
            // Get the marker's location in the control.
            int marker_x = style()->sliderPositionFromValue(minimum(), maximum(), event_markers.at(i), width());

            // Make sure we're in the range for drawing.
            if (marker_x >= 0)
            {
                // See if we're drawing a run of events.
                if ((event_markers.size() > (i + 1)) && (event_markers.at(i + 1) == INT_MAX))
                {
                    // Sanity check.
                    if (event_markers.size() > (i + 2))
                    {
                        // Get the ending marker.
                        int marker_end_x = style()->sliderPositionFromValue(minimum(), maximum(), event_markers.at(i + 2), width());

                        // Draw our region.
                        painter.setBrush(Qt::BDiagPattern);
                        painter.drawRect(marker_x, 1, marker_end_x - (marker_x + 1), height() - 2);

                        // Bump our index.
                        i += 2;
                    }
                }
                else
                {
                    // Draw the marker.
                    painter.drawLine(marker_x, 0, marker_x, height());
                }
            }
        }
    }

    // Draw the slider.
    paint_slider(&painter);

    // Do we have a hilighted region?
    if (!hilighter.isEmpty())
    {
        painter.fillRect(hilighter, QBrush(QColor(128, 128, 255, 128)));
    }
}

/*!
 * @brief Handle the mouse wheel event.
 * @param e     Our event pointer.
 */
void fiSpanSlider::wheelEvent(QWheelEvent *e)
{
    // Update our zoom.
    update_zoom(e->angleDelta().y() > 0.0);
    update();
}


//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief When we get a QSlider::sliderReleased signal, update the handle
 *        that was being pressed, if any.
 */
void fiSpanSlider::movePressedHandle()
{
    switch (lastPressed)
    {
    case LowerHandle:
        if (lowerPos != lower)
        {
            triggerAction(QAbstractSlider::SliderMove, (mainControl == LowerHandle));
        }
        break;
    case UpperHandle:
        if (upperPos != upper)
        {
            triggerAction(QAbstractSlider::SliderMove, (mainControl == UpperHandle));
        }
        break;
    default:
        break;
    }
}

/*!
 * @brief Set the lower position of the span.
 * @param new_lower     The new lower position.
 */
void fiSpanSlider::setLowerPosition(int new_lower)
{
    if (lowerPos != new_lower)
    {
        lowerPos = new_lower;
        if (!hasTracking())
            update();
        if (isSliderDown())
            emit lowerPositionChanged(new_lower);
        if (hasTracking() && !blockTracking)
        {
            triggerAction(SliderMove, (mainControl == LowerHandle));
        }
    }
}

void fiSpanSlider::setLowerValue(int new_lower)
{
    setSpan(new_lower, upper);
}

/*!
 * @brief Set the upper position of the span.
 * @param new_upper     The new upper position.
 */
void fiSpanSlider::setUpperPosition(int new_upper)
{
    if (upperPos != new_upper)
    {
        upperPos = new_upper;
        if (!hasTracking())
            update();
        if (isSliderDown())
            emit upperPositionChanged(new_upper);
        if (hasTracking() && !blockTracking)
        {
            triggerAction(SliderMove, (mainControl == UpperHandle));
        }
    }
}

void fiSpanSlider::setUpperValue(int new_upper)
{
    setSpan(lower, new_upper);
}

