/*!
 * @file fi_spanslider.h
 * @brief A double-headed slider, heavily derived from the QxtSpanSlider which
 *        no longer appears to be support and won't build under Qt5.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_SPANSLIDER_H
#define FI_SPANSLIDER_H

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include <QColor>
#include <QList>
#include <QSlider>
#include <QStyle>

//--------------------------------------------------------------------
// Namespace: FI_SPAN_SLIDER
//--------------------------------------------------------------------
namespace FI_SPAN_SLIDER
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! This enum describes the available handle movement modes.
enum HANDLE_MOVEMENT_MODE
{
    FreeMovement,                           //!< The handles can be moved freely.
    NoCrossing,                             //!< The handles cannot cross, but they can still overlap each other. The lower and upper values can be the same.
    NoOverlapping                           //!< The handles cannot overlap each other. The lower and upper values cannot be the same.
};

//! This enum describes the available span handles.
enum SPAN_HANDLE
{
    NoHandle,                               //!< Internal only.
    LowerHandle,                            //!< The lower boundary handle.
    UpperHandle                             //!< The upper boundary handle.
};

enum TRIGGER_POS
{
    NoTrigger = -1,
    DefaultTriggerPos = 0
};

}   // End namespace (FI_SPAN_SLIDER)

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QStylePainter;

//--------------------------------------------------------------------
// Class Declaration: fiSpanSlider
//--------------------------------------------------------------------
class fiSpanSlider : public QSlider
{
    Q_OBJECT

public:
    explicit fiSpanSlider(QWidget *parent = nullptr);
    explicit fiSpanSlider(Qt::Orientation orientation, QWidget *parent = nullptr);
    ~fiSpanSlider();

    void    setEventColor(const QColor& newColor);
    QColor  eventColor() const { return event_color; }

    void    setTriggerColor(const QColor& newColor);
    QColor  triggerColor() const { return trigger_color; }

    void    setTriggerPos(int newPos);
    void    setEventMarkers(const QList<int>& newMarkers);
    const QList<int>& markers() const;

    void    setHandleMovementMode(FI_SPAN_SLIDER::HANDLE_MOVEMENT_MODE mode)      { movement = mode; }
    FI_SPAN_SLIDER::HANDLE_MOVEMENT_MODE handleMovementMode() const;

    int     lowerValue() const;
    int     upperValue() const;

    int     lowerPosition() const;
    int     upperPosition() const;

    int     get_zoom_factor()             { return zoom_factor; }
    const QRect&    zoom_region() const;
    int     get_zoom_start_frame();
    int     get_zoom_stop_frame();
    void    set_zoom_center(int center_x);
    bool    zoom_is_showing();
    void    update_zoom(bool shrink_it);

public slots:
    void    setLowerValue(int new_lower);
    void    setUpperValue(int new_upper);
    void    setSpan(int new_lower, int new_upper);

    void    setLowerPosition(int new_lower);
    void    setUpperPosition(int new_upper);

    void    updateRange(int min, int max);
    void    movePressedHandle();

signals:
    void    spanChanged(int lower, int upper);
    void    lowerValueChanged(int lower);
    void    upperValueChanged(int upper);

    void    lowerPositionChanged(int lower);
    void    upperPositionChanged(int upper);

    void    sliderPressed(FI_SPAN_SLIDER::SPAN_HANDLE handle);

    void    zoom_showing();
    void    zoom_hiding();
    void    zoom_changing();
    void    zoom_moving();

protected:
    virtual void    mouseDoubleClickEvent(QMouseEvent *e);
    virtual void    keyPressEvent(QKeyEvent *event);
    virtual void    mouseMoveEvent(QMouseEvent *e);
    virtual void    mousePressEvent(QMouseEvent *e);
    virtual void    mouseReleaseEvent(QMouseEvent *e);
    virtual void    paintEvent(QPaintEvent *e);
    virtual void    wheelEvent(QWheelEvent *e);
    virtual bool    event(QEvent *e);

private:
    void    drawHandle(QStylePainter *painter, FI_SPAN_SLIDER::SPAN_HANDLE handle) const;
    void    drawSpan(QStylePainter *painter, const QRect& rect) const;
    void    handleMousePress(const QPoint& pos, QStyle::SubControl& control, int value, FI_SPAN_SLIDER::SPAN_HANDLE handle);
    void    init_class();
    void    initStyleOption(QStyleOptionSlider *option) const override { initStyleOption(option, FI_SPAN_SLIDER::UpperHandle); }
    void    initStyleOption(QStyleOptionSlider *option, FI_SPAN_SLIDER::SPAN_HANDLE handle = FI_SPAN_SLIDER::UpperHandle) const;
    void    paint_slider(QStylePainter *painter);
    int     pixelPosToRangeValue(int pos) const;
    void    setupPainter(QPainter *painter, Qt::Orientation orientation, qreal x1, qreal y1, qreal x2, qreal y2) const;
    void    swapControls();
    void    triggerAction(QAbstractSlider::SliderAction action, bool main_control_sel);

    int pick(const QPoint& pt) const
    {
        return orientation() == Qt::Horizontal ? pt.x() : pt.y();
    }

private:
    QColor      event_color;                    //!< The color we paint our event markers with.
    QList<int>  event_markers;                  //!< The event markers.

    QColor      trigger_color;                  //!< The color we paint our trigger marker with.
    int         trigger_pos;                    //!< Location of the trigger event.

    QRect       hilighter;                      //!< A highlighted area we're showing as "zoomed".
    int         zoom_start;                     //!< The starting frame of the highlighted region.
    int         zoom_stop;                      //!< The ending frame of the highlighted region.
    int         zoom_factor;                    //!< Zoom, from 1..MAX_ZOOM.
    bool        hilight_in_motion;              //!< True if the user is moving the region, false otherwise.
    QPoint      last_pos;                       //!< Last position of the mouse.

    int         lower;
    int         upper;
    int         lowerPos;
    int         upperPos;
    int         offset;
    int         position;
    bool        firstMovement;
    bool        blockTracking;
    QStyle::SubControl lowerPressed;
    QStyle::SubControl upperPressed;
    FI_SPAN_SLIDER::SPAN_HANDLE lastPressed;
    FI_SPAN_SLIDER::SPAN_HANDLE mainControl;
    FI_SPAN_SLIDER::HANDLE_MOVEMENT_MODE    movement;
};

#endif // FI_SPANSLIDER_H
