/*!
 * @file fi_splinewidget.h
 * @brief This file contains the implementation of the Spline display widget.
 *
 * This widget is built on work described here:
 *      Spline library: http://people.sc.fsu.edu/~jburkardt/c_src/spline/spline.html
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_splinewidget.h"
#include "libspline/aaCurve.h"

#include <QContextMenuEvent>
#include <QMenu>
#include <QResizeEvent>
#include <QVector>

#include <qwt_plot.h>
#include <qwt_plot_curve.h>
#include <qwt_plot_canvas.h>
#include <qwt_plot_item.h>
#include <qwt_plot_layout.h>
#include <qwt_plot_marker.h>
#include <qwt_plot_picker.h>
#include <qwt_picker_machine.h>
#include <qwt_symbol.h>

#include "logger.h"

namespace FI_SPLINEWIDGET
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
const qreal SEL_FUDGE       = 16.0;         //!< Fudge factor for checking to see if a knot is selected.

const int   CURVE_INDEX     = 0;            //!< Index into the plot list for the curve we're drawing.
const int   REFRENCE_INDEX  = 1;            //!< Index into the plot list for the "normal" 1:1 line.
const int   MARKER_INDEX    = 2;            //!< The start index into the plot list for the "knot" markers on the curve.

}   // End of namespace FI_SPLINEWIDGET.

using namespace FI_SPLINEWIDGET;

/*!
 * @class fiSplineWidget
 * @brief The fiSplineWidget widget allows the user to create, display and modify
 *        a spline object.
 *
 * The fiSplineWidget allows the user to interact with a spline. The spline will display
 * on the specified parent widget.
 */

/*!
 * @brief Class constructor.
 */
fiSplineWidget::fiSplineWidget(QWidget *parent)
    : QWidget(parent)
{
    // Do base initialization
    init_controls();
}

fiSplineWidget::fiSplineWidget(double deltaT, QWidget *parent)
    : QWidget(parent)
{
    // Do base initialization
    init_controls();

    // Modify our delta.
    delta_t = deltaT;
}

fiSplineWidget::~fiSplineWidget(void)
{
}

void fiSplineWidget::add_marker(int index)
{
    aaAaa::aaPoint  marker_loc;
    QwtPlot         *p_plot;
    QwtPlotMarker   *p_marker;

    // Sanity check.
    if (index < 0)
        return;

    // Get our plot.
    p_plot = plot_curves.at(0);

    // Get the location.
    marker_loc = spline_data->knots.at((uint64_t)index);

    // Create our marker.
    p_marker = new QwtPlotMarker();
    p_marker->setLineStyle(QwtPlotMarker::NoLine);
    p_marker->setSymbol(new QwtSymbol(QwtSymbol::Diamond, QColor(Qt::green), QColor(Qt::green), QSize(8, 8)));
    p_marker->setValue((double)marker_loc.t, (double)marker_loc.y);
    p_marker->attach(p_plot);
}

void fiSplineWidget::del_marker(int index)
{
    QwtPlot         *p_plot;

    // Sanity check.
    if (index < 0)
        return;

    // Get our plot.
    p_plot = plot_curves.at(0);

    // Get a pointer to the specified marker object.
    QwtPlotMarker *pMarker = static_cast<QwtPlotMarker *>(p_plot->itemList().at(index));

    // Remove the marker from the plot and delete the object.
    pMarker->detach();
    delete pMarker;
}

int fiSplineWidget::find_marker(int index)
{
    QwtPlot *p_plot = plot_curves.at(0);
    QPointF knot_point((double)spline_data->knots[(uint64_t)index].t, (double)spline_data->knots[(uint64_t)index].y);

    // Find the marker associated with this knot.
    for (int i = 0; i < p_plot->itemList().count(); ++i)
    {
        if (p_plot->itemList().at(i)->rtti() == QwtPlotItem::Rtti_PlotMarker)
        {
            QwtPlotMarker *pMarker = static_cast<QwtPlotMarker *>(p_plot->itemList().at(i));
            if (pMarker != nullptr)
            {
                // Is this the marker?
                QPointF markPt = pMarker->value();

                if ((std::abs(markPt.x() - knot_point.x()) < 0.001) &&
                    (std::abs(markPt.y() - knot_point.y()) < 0.001))
                {
                    return i;
                }
            }
        }
    }

    // No marker found.
    return -1;
}

int fiSplineWidget::get_data(int step, QVector<double>& lut)
{
    int     index;
    int     stop_index;
    double  t_val;

    // Make sure we have data.
    if (spline_data == nullptr)
        return 0;
    if (spline_data->size() <= 1)
        return 0;

    // Figure out how many points we need for this curve.
    int  point_len = (int)(spline_data->knots.at(spline_data->knots.size() - 1).t - spline_data->knots.at(0).t + 1);

    // Sanity check.
    stop_index = point_len / step;
    if (lut.size() != stop_index)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut.size:" << lut.size() << "!= stop_index:" << stop_index;

        // Make sure we don't try to blow out the LUT.
        if (lut.size() < stop_index)
            stop_index = lut.size();
    }

    // Get our curve factory.
    aaAaa::aaCurvePtr pspline;
    pspline = aaAaa::aaCurveFactory::createCurve(*spline_data);
    if (!pspline.get())
        return 0;

    // Build up the lookup table array.
    t_val = 0.0;
    for (index = 0; index < stop_index; index++)
    {
        double v;

        pspline->getValue(t_val, v);
        if (spline_data->bLimited)
        {
            if (v > (double)spline_data->limit_top)
                v = (double)spline_data->limit_top;
            else if (v < (double)spline_data->limit_bottom)
                v = (double)spline_data->limit_bottom;
        }

        // Save the value.
        lut[ index ] = v;

        // Increment our t_val?
        t_val += step;
    }

    // All is well.
    return point_len;
}

void fiSplineWidget::init_controls()
{
    // Intialize our member variables.
    ctrl_point_sel = -1;
    spline_data = nullptr;
    delta_t = 1.0;
    zoom_ok = false;
    max_knots = -1;

    QwtPlot         *p_plot;
    QwtPlotCurve    *p_curve;
    QSizePolicy     size_policy(QSizePolicy::Expanding, QSizePolicy::Expanding);

    // Initialize our size policy.
    size_policy.setHorizontalStretch(0);
    size_policy.setVerticalStretch(0);

    // Create our plot widget and attach our curve to it.
    p_plot = new QwtPlot(this);
    p_plot->enableAxis(QwtPlot::yLeft, false);
    p_plot->enableAxis(QwtPlot::xBottom, false);
    p_plot->setLineWidth(1);
    p_plot->setFocusPolicy(Qt::NoFocus);
    p_plot->setSizePolicy(size_policy);
    p_plot->plotLayout()->setCanvasMargin(0, -1);
    p_plot->setFrameStyle(QFrame::Plain | QFrame::NoFrame);
    p_plot->setAutoReplot(true);
    p_plot->canvas()->setCursor(Qt::ArrowCursor);
    p_plot->hide();

    // Create our picker.
    plot_picker = new QwtPlotPicker(QwtPlot::xBottom, QwtPlot::yLeft,
                                    QwtPlotPicker::NoRubberBand, QwtPicker::AlwaysOff,
                                    p_plot->canvas());
    plot_picker->setStateMachine(new QwtPickerDragPointMachine());

    // Create our new plot curve control. It's VERY important
    // that this curve is attached to the plot before our "normal"
    // curve, otherwise our index constants won't be correct.
    p_curve = new QwtPlotCurve();
    p_curve->setStyle(QwtPlotCurve::Lines);
    p_curve->setCurveAttribute(QwtPlotCurve::Fitted, true);
    p_curve->setRenderHint(QwtPlotItem::RenderAntialiased);
    p_curve->setPen(QPen(Qt::magenta));
    p_curve->attach(p_plot);

    // Create our plot curve for the "normal" 1:1 line. It's VERY
    // important that this is attached to the plot AFTER our "working"
    // curve, or else our index constants won't be correct.
    p_curve = new QwtPlotCurve();
    p_curve->setStyle(QwtPlotCurve::Lines);
    p_curve->setCurveAttribute(QwtPlotCurve::Fitted, true);
    p_curve->setRenderHint(QwtPlotItem::RenderAntialiased);
    p_curve->setPen(QPen(Qt::gray));
    p_curve->attach(p_plot);

    // Put the plot widget into our list and add it to the layout control.
    plot_curves.append(p_plot);

    // Make sure we get mouse events.
    setFocusPolicy(Qt::ClickFocus);

    // Connect our signals/slots.
    connect(plot_picker, SIGNAL(appended(QPointF)), this, SLOT(marker_selected(QPointF)));
    connect(plot_picker, SIGNAL(moved(QPointF)), this, SLOT(marker_moved(QPointF)));
}

void fiSplineWidget::update_spline(void)
{
    int             index;
    int             point_len;
    double          max_y_value, max_x_value;
    double          t_val;
    QwtPlot         *p_plot;
    QwtPlotCurve    *p_curve;

    // Get our plot.
    p_plot = plot_curves.at(0);

    // Figure out how many points we need for this curve.
    max_x_value = (double)spline_data->knots.at(spline_data->knots.size() - 1).t;
    point_len = (int)(max_x_value - (double)spline_data->knots.at(0).t + 1);

    // Because our step size is not necessarily 1.0, let's update the point length by that information.
    point_len *= 1.0 / delta_t;

    // Get our arrays allocated.
    QVector<double> x_axis(point_len);
    QVector<double> y_axis(point_len);

    aaAaa::aaCurvePtr pspline;
    pspline = aaAaa::aaCurveFactory::createCurve(*spline_data);
    if (!pspline.get())
        return;

    // Build the curve.
    max_y_value = 0.0;
    t_val = 0.0;
    for (index = 0; index < point_len; index++)
    {
        double v;

        pspline->getValue(t_val, v);
        if (spline_data->bLimited)
        {
            if (v > (double)spline_data->limit_top)
                v = (double)spline_data->limit_top;
            else if (v < (double)spline_data->limit_bottom)
                v = (double)spline_data->limit_bottom;
        }

        // Check to see if this is the maximum y.
        if (v > max_y_value)
            max_y_value = v;

        // Save our value.
        x_axis[ index ] = t_val;
        y_axis[ index ] = v;

        // Update our t_val by the delta.
        t_val += delta_t;
    }

    // Make sure we set the max Y appropriately.
    if (spline_data->bLimited)
    {
        if ((double)spline_data->limit_top > max_y_value)
            max_y_value = (double)spline_data->limit_top;
        else if ((double)spline_data->limit_bottom > max_y_value)
            max_y_value = (double)spline_data->limit_bottom;
    }

    // Get our curve object.
    p_curve = static_cast<QwtPlotCurve *>(p_plot->itemList().at(CURVE_INDEX));
    if (p_curve == nullptr)
        return;

    // Set the data. The setSamples() method uses a "deep copy" to move the samples.
    p_curve->setSamples(x_axis.data(), y_axis.data(), point_len);

    // Now build our linear curve. The X axis will remain the same,
    // we just need a new Y axis.
    for (index = 0; index < point_len; index++)
    {
        y_axis[ index ] = (max_y_value / point_len) * index;
    }

    // Get our curve object.
    p_curve = static_cast<QwtPlotCurve *>(p_plot->itemList().at(REFRENCE_INDEX));
    if (p_curve == nullptr)
        return;

    // Set the data. The setSamples() method uses a "deep copy" to move the samples.
    p_curve->setSamples(x_axis.data(), y_axis.data(), point_len);

    // Update the scale.
    p_plot->setAxisScale(QwtPlot::xBottom, 0.0, max_x_value, delta_t);
    p_plot->setAxisScale(QwtPlot::yLeft, 0.0, max_y_value, 1.0);
    p_plot->updateAxes();
    p_plot->update();
}

/*!
 * @brief Setup the initial curve.
 * @param sdata     Pointer to the spline data for the initial curve.
 */
void fiSplineWidget::set_spline(std::shared_ptr<aaAaa::aaSpline> sdata)
{
    int             index;
    QwtPlot         *p_plot;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "SplineWidget:" << objectName();

    // Save the pointer to the spline.
    spline_data = sdata;

    // Initialize our selection index.
    ctrl_point_sel = -1;
    marker_index = -1;

    // Get our plot.
    p_plot = plot_curves.at(0);

    // Delete all the markers we might have in the current plot. The
    // curves are items 0 & 1, and the markers start at item 2.
    while (p_plot->itemList().count() > MARKER_INDEX)
        del_marker(MARKER_INDEX);

    // Make sure we have data.
    if (!spline_data)
        return;
    if (spline_data->size() <= 1)
        return;

    // Add the markers ("knots") to the plot.
    for (index = 0; index < (int)spline_data->knots.size(); ++index)
    {
        add_marker(index);
    }

    // Update our plot.
    update_spline();

    // Turn on the plot.
    p_plot->show();

    // Let folks know we have values, but none selected.
    emit selectValuesChanged(0, 0);
}

/*!
 * @brief Check to see if one of the knots on the curve is selected.
 * @param point     The point around which to check.
 * @return Index of the selected knot, -1 if none selected. Also,
 *         the "selectValuesChanged" signal is emitted.
 */
int fiSplineWidget::checkSelected(const QPointF& point)
{
    // Make sure we have data.
    if (!spline_data)
        return -1;

    // Loop through all the knots on the curve.
    for (size_t i = 0; i < spline_data->knots.size(); ++i)
    {
        // Are we within the fudge factor of this knot?
        if ((std::abs(point.x() - (double)spline_data->knots[i].t) < SEL_FUDGE) &&
            (std::abs(point.y() - (double)spline_data->knots[i].y) < SEL_FUDGE))
        {
            // Find the marker associated with the knot.
            marker_index = find_marker((int)i);

            // Let folks know that this knot is selected.
            emit selectValuesChanged(spline_data->knots[i].t, spline_data->knots[i].y);
            return (int)i;
        }
    }

    // No knots selected, so no marker selected either.
    marker_index = -1;
    emit selectValuesChanged(0, 0);
    return -1;
}


/*!
 * @brief The user has right mouse clicked in our window. Put up a
 *        context specific menu.
 *
 * @param event_ptr     Pointer to the menu event.
 */
void fiSplineWidget::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QAction *linear;
    QAction *cubic;
    QAction *select;
    QMenu   context_menu(this);

    // Only do this if the ctrl key is down.
    if (event_ptr->modifiers() != Qt::CTRL)
        return;
    if (spline_data == nullptr)
        return;

    // Let the user change the spline "type".
    cubic = context_menu.addAction(tr("Cubic"));
    linear = context_menu.addAction(tr("Linear"));

    // Make sure they are "checkable".
    cubic->setCheckable(true);
    linear->setCheckable(true);

    // Which one is currently selected.
    if (spline_data->type == aaAaa::aaSpline::SPLINE_CUBIC)
        cubic->setChecked(true);
    else
        linear->setChecked(true);

    // Pop it up.
    select = context_menu.exec(event_ptr->globalPos());
    if (select == cubic)
    {
        emit switchType(aaAaa::aaSpline::SPLINE_CUBIC);
    }
    else if (select == linear)
    {
        emit switchType(aaAaa::aaSpline::SPLINE_LINEAR);
    }
}

/*!
 * @brief Handle the event when the user release the mouse button.
 * @param event     Pointer to the event specific information.
 */
void fiSplineWidget::mouseReleaseEvent(QMouseEvent *event)
{
    // Make sure we have data.
    if (spline_data == nullptr)
        return;

    // Ignore if the control key is down.
    if (event->modifiers() == Qt::CTRL)
        return;

    // Get a pointer to our plot.
    QwtPlot *p_plot = plot_curves.at(0);

    // Was this a left mouse button release?
    if (event->button() == Qt::LeftButton)
    {
        // Make sure we had a marker.
        if (marker_index >= 0)
        {
            // Get a pointer to the selected marker.
            QwtPlotMarker *pMarker = static_cast<QwtPlotMarker *>(p_plot->itemList().at(marker_index));

            // Change the color to show that it's selected.
            pMarker->setSymbol(new QwtSymbol(QwtSymbol::Diamond, QColor(Qt::green), QColor(Qt::green), QSize(8, 8)));
        }

        // Reset our control point.
        ctrl_point_sel = -1;
        marker_index = -1;
    }
    else if (event->button() == Qt::RightButton)
    {
        aaAaa::aaSpline::KnotsList& ctrlpoints = spline_data->knots;
        QPointF invT;

        invT.setX(p_plot->invTransform(QwtPlot::xBottom, event->pos().x()));
        invT.setY(p_plot->invTransform(QwtPlot::yLeft, event->pos().y()));

        // Are we on a point?
        ctrl_point_sel = checkSelected(invT);

        // Is there a control point selected?
        if (ctrl_point_sel >= 0)
        {
            // If it's not the first or last point, remove it.
            if ((ctrl_point_sel != 0) && (ctrl_point_sel != (int)(ctrlpoints.size() - 1)))
            {
                ctrlpoints.erase(ctrlpoints.begin() + ctrl_point_sel);

                // Delete the marker at this point.
                del_marker(marker_index);
                marker_index = -1;

                // Check to see if there's a knot around where the user released the mouse button.
                ctrl_point_sel = checkSelected(invT);
            }

            // Otherwise ignore right-clicks on the beginning and ending of our spline curve.
        }
        else
        {
            // Check our knot count, and make sure we're OK to add another.
            if ((max_knots > 0) && ((int)ctrlpoints.size() < max_knots))
            {
                // Does the release point indicate it's somewhere on the curve?
                if ((invT.x() > (double)ctrlpoints[0].t) && (invT.x() < (double)ctrlpoints[ ctrlpoints.size() - 1 ].t))
                {
                    int index;

                    // Find out where on the curve the point belongs.
                    for (index = 1; index < (int)ctrlpoints.size(); ++index)
                    {
                        if (invT.x() < (double)ctrlpoints[(uint64_t)index].t)
                            break;
                    }

                    // Insert the point at the current location.
                    ctrlpoints.insert(ctrlpoints.begin() + index, aaAaa::Vector2((float)invT.x(), (float)invT.y()));

                    // Add a marker at the control point.
                    add_marker(index);
                }
            }

            // Reset our control point.
            ctrl_point_sel = -1;
            marker_index = -1;
        }
    }

    // Update the spline.
    update_spline();
}

void fiSplineWidget::resizeEvent(QResizeEvent *event)
{
    Q_UNUSED(event);

    // Update the plot size.
    plot_curves.at(0)->resize(event->size());
}

void fiSplineWidget::marker_moved(const QPointF& pos)
{
    // Make sure we have data.
    if (spline_data == nullptr)
        return;

    // Make sure we're enabled.
    if (!isEnabled())
        return;

    // Get a pointer to our plot.
    QwtPlot *p_plot = plot_curves.at(0);

    // Make sure we have a point to work with.
    if ((ctrl_point_sel < 0) || (marker_index < 0))
        return;

    // Update our control point.
    aaAaa::aaSpline::KnotsList& ctrlpoints = spline_data->knots;

    // Limit the range of the control point, so that it all makes sense.
    // Is this an end knot?
    if ((ctrl_point_sel == 0) || (ctrl_point_sel == (int)(ctrlpoints.size() - 1)))
    {
        // Only allow vertical movement of the end knot.
        ctrlpoints[(uint64_t)ctrl_point_sel].y = pos.y();
    }
    else
    {
        // Update the coordinates of the selected knot.
        ctrlpoints[(uint64_t)ctrl_point_sel].t = pos.x();
        ctrlpoints[(uint64_t)ctrl_point_sel].y = pos.y();

        // Make sure it doesn't overlap the knot on either side.
        if (ctrlpoints[(uint64_t)ctrl_point_sel].t >= (ctrlpoints[(uint64_t)ctrl_point_sel + 1].t - (float)(SEL_FUDGE / 2)))
            ctrlpoints[(uint64_t)ctrl_point_sel].t = (ctrlpoints[(uint64_t)ctrl_point_sel + 1].t - (float)(SEL_FUDGE / 2));
        if (ctrlpoints[(uint64_t)ctrl_point_sel].t <= (ctrlpoints[(uint64_t)ctrl_point_sel - 1].t + (float)(SEL_FUDGE / 2)))
            ctrlpoints[(uint64_t)ctrl_point_sel].t = ctrlpoints[(uint64_t)ctrl_point_sel - 1].t + (float)(SEL_FUDGE / 2);
    }

    // Do any limiting that's required.
    if (spline_data->bLimited)
    {
        if (ctrlpoints[(uint64_t)ctrl_point_sel].t < spline_data->limit_left)
        {
            ctrlpoints[(uint64_t)ctrl_point_sel].t = spline_data->limit_left;
        }
        else if (ctrlpoints[(uint64_t)ctrl_point_sel].t > spline_data->limit_right)
        {
            ctrlpoints[(uint64_t)ctrl_point_sel].t = spline_data->limit_right;
        }

        if (ctrlpoints[(uint64_t)ctrl_point_sel].y < spline_data->limit_bottom)
        {
            ctrlpoints[(uint64_t)ctrl_point_sel].y = spline_data->limit_bottom;
        }
        else if (ctrlpoints[(uint64_t)ctrl_point_sel].y > spline_data->limit_top)
        {
            ctrlpoints[(uint64_t)ctrl_point_sel].y = spline_data->limit_top;
        }
    }

    // Get a pointer to the selected marker.
    QwtPlotMarker *pMarker = static_cast<QwtPlotMarker *>(p_plot->itemList().at(marker_index));

    // Change the marker location.
    pMarker->setValue(QPointF((double)ctrlpoints[(uint64_t)ctrl_point_sel].t, (double)ctrlpoints[(uint64_t)ctrl_point_sel].y));

    // Update spline.
    update_spline();

    // Let folks know the data has changed.
    emit selectValuesChanged(ctrlpoints[(uint64_t)ctrl_point_sel].t, ctrlpoints[(uint64_t)ctrl_point_sel].y);
}

void fiSplineWidget::marker_selected(const QPointF& pos)
{
    // Make sure we have data.
    if (spline_data == nullptr)
        return;

    // Make sure we're enabled.
    if (!isEnabled())
        return;

    // Get a pointer to our plot.
    QwtPlot *p_plot = plot_curves.at(0);

    // Check to see if we have a point.
    ctrl_point_sel = checkSelected(pos);

    // Did we get a marker?
    if (marker_index >= 0)
    {
        // Get a pointer to the selected marker.
        QwtPlotMarker *pMarker = static_cast<QwtPlotMarker *>(p_plot->itemList().at(marker_index));

        // Change the color to show that it's selected.
        pMarker->setSymbol(new QwtSymbol(QwtSymbol::Diamond, QColor(Qt::black), QColor(Qt::black), QSize(8, 8)));
    }
}
