/*!
 * @file fi_streamsocket.cpp
 * @brief This file contains the implementation of the FiStreamSocket
 *        class.
 *
 * Copyright 2018 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QHostAddress>

#include "fi_streamsocket.h"
#include "logger.h"

/*!
 * @brief FiStreamSocket constructor.
 */
FiStreamSocket::FiStreamSocket(QObject *parent) :
    QObject(parent),
    tcp_socket(nullptr),
    stream_port(0),
    last_errcode(0)
{
}

/*!
 * @brief FiStreamSocket destructor.
 */
FiStreamSocket::~FiStreamSocket()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    if (tcp_socket != nullptr)
    {
        // Make sure the socket is closed down.
        if (tcp_socket->isValid())
            tcp_socket->abort();

        // Clean up our socket.
        tcp_socket->deleteLater();
        tcp_socket = nullptr;
    }
}

/*!
 * @brief Connect to a camera server, as specified by the host and port.
 *
 * @param host  The camera's address information.
 * @param port  The streaming port on the camera.
 */
void FiStreamSocket::connect_cam(const QHostAddress& host, quint16 port)
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "host:" << host.toString() << "port:" << port;

    // Make sure we're not already connected.
    if (tcp_socket != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Already have a TCP Socket object!";

        // Close the socket.
        tcp_socket->close();
        tcp_socket->deleteLater();
        tcp_socket = nullptr;
    }

    // Allocate our socket object.
    tcp_socket = new QTcpSocket(this);
    if (tcp_socket == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a QTcpSocket object!";
        return;
    }

    // Connect up our slots/signals.
    connect(tcp_socket, &QTcpSocket::connected, this, &FiStreamSocket::connected);
    connect(tcp_socket, &QTcpSocket::disconnected, this, &FiStreamSocket::disconnected);
    connect(tcp_socket, &QTcpSocket::errorOccurred, this, &FiStreamSocket::errorOccurred);
    connect(tcp_socket, &QTcpSocket::readyRead, this, &FiStreamSocket::read_pending, Qt::DirectConnection);

    // Remember the port we're connecting to.
    stream_port = port;

    // Connect to our host/camera. This is not a blocking call; we'll get
    // a signal when we've connected.
    tcp_socket->connectToHost(host, port);
}

/*!
 * @brief This method returns the number of bytes queued up and ready to be read.
 * @return The number of bytes available on the socket.
 */
qint64 FiStreamSocket::data_available()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (tcp_socket == nullptr)
        return false;
    return tcp_socket->bytesAvailable();
}

/*!
 * @brief Read data from the socket.
 *
 * @param buffer_ptr    Destination for the data.
 * @param read_len      The maximum number of bytes to read. Actual number may be less.
 * @return How many bytes were actually read.
 */
qint64 FiStreamSocket::read(uint8_t *buffer_ptr, qint64 read_len)
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "buffer:" << Qt::showbase << Qt::hex << buffer_ptr << "len:" << dec << read_len;

    // Sanity check.
    if (tcp_socket == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No tcp_socket!";
        return 0;
    }
    if (buffer_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No buffer!";
        return 0;
    }

    // Read up the data.
    return tcp_socket->read((char *)buffer_ptr, read_len);
}

QByteArray FiStreamSocket::read(qint64 max_size)
{
    // Sanity check.
    if (tcp_socket == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No tcp_socket!";
        return QByteArray();
    }

    // Read up the data.
    return tcp_socket->read(max_size);
}

QByteArray FiStreamSocket::readAll()
{
    // Sanity check.
    if (tcp_socket == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No tcp_socket!";
        return QByteArray();
    }

    // Read up the data.
    return tcp_socket->readAll();
}

/*!
 * @brief Time to close down the socket.
 */
void FiStreamSocket::quit()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if ((tcp_socket == nullptr) || !tcp_socket->isValid())
        return;

    // Close the socket.
    tcp_socket->close();
}

//--------------------------------------------------------------------
// Slots
//--------------------------------------------------------------------

/*!
 * @brief The TCP connection has been successfully established.
 */
void FiStreamSocket::connected()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Tell everyone we're ready.
    emit ready();
}

/*!
 * @brief The TCP socket has been disconnected.
 */
void FiStreamSocket::disconnected()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Clean up our socket.
    tcp_socket->deleteLater();
    tcp_socket = nullptr;

    // Tell everyone we're done.
    emit finished();
}

/*!
 * @brief Catches the TCP socket error() signal.
 * @param socketError This describes the type of error that occurred.
 */
void FiStreamSocket::errorOccurred(QAbstractSocket::SocketError socketError)
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << tcp_socket->errorString();

    switch (socketError)
    {
    case QAbstractSocket::RemoteHostClosedError:
    case QAbstractSocket::HostNotFoundError:
    case QAbstractSocket::ConnectionRefusedError:
    default:
        last_errstr = tcp_socket->errorString();
        last_errcode = (int)socketError;
        break;
    }

    // Let folks know there was an error.
    emit error();
}

/*!
 * @brief The TCP socket has indicated that new data is available for reading.
 */
void FiStreamSocket::read_pending()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Pass it on.
    emit data_ready();
}

