/*!
 * @file fi_stringspinbox.cpp
 * @brief This file contains the implementation of the fiStringSpinBox class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "fi_stringspinbox.h"
#include "logger.h"

/*!
 * @class fiStringSpinBox
 * @brief The fiStringSpinBox widget is a spin box with string items.
 *
 * fiStringSpinBox is spin box that takes strings. fiStringSpinBox allows
 * the user to choose a value by clicking the up and down buttons or by
 * pressing Up or Down on the keyboard to increase or decrease the value
 * currently displayed. The user can also type the value in manually.
 */

/*!
 * @brief Class constructor.
 */
fiStringSpinBox::fiStringSpinBox( QWidget *parent ) :
    QSpinBox( parent )
{
    // Initially we're empty.
    setRange( 0, 0 );

    // Connect our signal/slots.
    connect( this, SIGNAL(valueChanged(int)), this, SIGNAL(value_changed()));
}

/*!
 * @brief Class destructor.
 */
fiStringSpinBox::~fiStringSpinBox()
{
}

/*!
 * @brief Try to find a string from the list which matches the @a input.
 * @param[in,out] input     The partial string we're searching for.
 */
void fiStringSpinBox::fixup( QString& input ) const
{
    // Attempt to change the input string to be valid according to the string list.
    input = input.simplified();

    // Try finding a string from the list which starts with input.
    if ( !input.isEmpty() )
    {
        startsWith( input, input );
    }
}

/*!
 * @brief Set the list of strings to be used in the spin box.
 * @param[in] strings       The list of strings.
 */
void fiStringSpinBox::setStrings( const QStringList& strings )
{
    // Copy the list over.
    string_list = strings;

    // Set our range based on the list.
    if ( strings.isEmpty() )
    {
        setRange( 0, 0 );
    }
    else
    {
        setRange( 0, strings.size() - 1 );
        setValue( 0 );
    }
}

/*!
 * @brief Search through our list of strings and see if we can find one which starts
 *        with the string @a start.
 *
 * @param[in] start     The starting string we're searching for.
 * @param[out] string   The matching string, if we find one.
 *
 * @return The index in the list for the match, or -1 if none found.
 */
int fiStringSpinBox::startsWith( const QString& start, QString& string ) const
{
    int index;

    // Loop through all the strings.
    for ( index = 0; index < string_list.size(); ++index )
    {
        // Do we have a match?
        if ( string_list.at( index ).startsWith( start, Qt::CaseInsensitive ))
        {
            // Found it.
            string = string_list.at( index );
            return index;
        }
    }
    // not found
    return -1;
}

/*!
 * @brief Return the list of strings.
 */
const QStringList& fiStringSpinBox::strings() const
{
    return string_list;
}

/*!
 * @brief Retrieve the text located at the index @a value.
 *
 * @param value     The index for which we want the string.
 * @return The string, or an empty string if there wasn't a match.
 */
QString fiStringSpinBox::textFromValue( int value ) const
{
    // Sanity check.
    if ( string_list.isEmpty() || (( value >= 0 ) && ( value < string_list.size() )))
        return ( string_list.isEmpty() ? QLatin1String( "" ) : string_list.at( value ));

    // Failed the sanity check.
    QLOG_WARN() << __PRETTY_FUNCTION__ << "Sanity check failed!";
    return QLatin1String( "" );
}

/*!
 * @brief Validate the @a input string.
 *
 * Invalid: input is invalid according to the string list
 * Intermediate: it is likely that a little more editing will make the input acceptable
 *              (e.g. the user types "A" and stringlist contains "ABC")
 * Acceptable: the input is valid.
 *
 * @param[in] input The string we want validated.
 * @param pos       Unused.
 * @return  The state as described above.
 */
QValidator::State fiStringSpinBox::validate( QString& input, int& pos ) const
{
    Q_UNUSED(pos);
    QString temp;
    QValidator::State state = QValidator::Invalid;

    if ( string_list.contains( input ))
    {
        // We have an exact match
        state = QValidator::Acceptable;
    }
    else if ( input.isEmpty() || ( startsWith( input, temp ) != -1 ))
    {
        // Still empty or some string in the list starts with input.
        state = QValidator::Intermediate;
    }

    // Return our state.
    return state;
}

/*!
 * @brief Find the string @a text and return it's index in the list.
 *
 * @param[in] text      The string we're interested in.
 * @return The index in the list, or -1 if not found.
 */
int fiStringSpinBox::valueFromText( const QString& text ) const
{
    return string_list.indexOf( text );
}

