/*!
 * @file fi_switchbutton.cpp
 * @brief This provides the implementation for the fiSwitchButton class.
 *
 * Copyright 2019 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 *
 * Derived extensively from: https://stackoverflow.com/questions/14780517/toggle-switch-in-qt/38102598.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QApplication>
#include <QColor>
#include <QKeyEvent>
#include <QLabel>
#include <QPainter>
#include <QPalette>
#include <QPen>
#include <QPropertyAnimation>

#include <memory>

#include "fi_switchbutton.h"
#include "logger.h"

namespace FI_SWITCHBUTTON
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Space (top and bottom) around the "switch" object.
const int SWITCH_BORDER     = 2;

// Radius for the switch.
const int SWITCH_RADIUS     = 12;
}

using namespace FI_SWITCHBUTTON;

/*!
 * @brief fiSwitchButton constructor.
 *
 * The fiSwitchButton class is comprised of the base class, layered below
 * a switch "background" object (SwitchBackground), labels ("on" and "off),
 * and the switch "toggle" itself (SwitchCircle).
 *
 * @param parent    The parent of this widget.
 * @param style     Any additional style information.
 */
fiSwitchButton::fiSwitchButton(QWidget *parent, Style style)
    : QWidget(parent, Qt::FramelessWindowHint)
    , _value(false)
    , _duration(100)
    , _enabled(true)
{
    // setWindowFlags(Qt::FramelessWindowHint);
    setAttribute(Qt::WA_TranslucentBackground);
    setFocusPolicy(Qt::StrongFocus);

    // Get the system palette.
    QPalette my_pal = QApplication::palette();

    // Figure out which group we should use.
    QPalette::ColorGroup pal_group;
    if (hasFocus())
        pal_group = QPalette::Active;
    else if (_enabled)
        pal_group = QPalette::Inactive;
    else
        pal_group = QPalette::Disabled;

    // The color for the button of the switch.
    _offcolor = QColor(my_pal.color(pal_group, QPalette::Button));

    // The background color shown in the "on" position.
    _oncolor = QColor(my_pal.color(pal_group, QPalette::Highlight));

    // The text color for the "on" text.
    _pencolor = QColor(my_pal.color(pal_group, QPalette::HighlightedText));

    // How "curvey" will we be?
    _borderradius = SWITCH_RADIUS;

    // Create the children widgets.
    _labeloff = new QLabel(this);
    _background = new SwitchBackground(this, _oncolor);
    _labelon = new QLabel(this);
    _circle = new SwitchCircle(this, _offcolor);

    // Adjust the labels for color, etc.
    _labelon->setStyleSheet(QString("color: rgb(%1, %2, %3); font-weight: bold;").arg(_pencolor.red()).arg(_pencolor.green()).arg(_pencolor.blue()));

    // Create the animation object for moving the switch button.
    __btn_move = new QPropertyAnimation(this);
    __btn_move->setTargetObject(_circle);
    __btn_move->setPropertyName("pos");

    // Create the animation for drawing the "on" background.
    __back_move = new QPropertyAnimation(this);
    __back_move->setTargetObject(_background);
    __back_move->setPropertyName("size");

    // Initialize our labels according to our "style".
    setStyle(style);

    // Default is OFF, so setup appropriately.
    _labelon->hide();
    _labeloff->show();

    // Put the "switch" into the starting position.
    _background->move(SWITCH_BORDER, SWITCH_BORDER);
    _circle->move(SWITCH_BORDER, SWITCH_BORDER);
}

fiSwitchButton::~fiSwitchButton()
{
    delete _circle;
    delete _background;
    delete _labeloff;
    delete _labelon;
    delete __btn_move;
    delete __back_move;
}

void fiSwitchButton::keyReleaseEvent(QKeyEvent *event)
{
    switch (event->key())
    {
    case Qt::Key_Space:
        // Toggle the value.
        setValue(!_value);
        break;
    case Qt::Key_Left:
        // Turn it OFF.
        setValue(false);
        break;
    case Qt::Key_Right:
        // Turn it ON.
        setValue(true);
        break;
    default:
        // Pass it on.
        QWidget::keyReleaseEvent(event);
        break;
    }
}


void fiSwitchButton::paintEvent(QPaintEvent *)
{
    QPen pen(Qt::NoPen);
    QLinearGradient _lg;
    std::unique_ptr<QPainter> painter;
    QPalette my_pal = QApplication::palette();

    // Allocate our painter.
    painter = std::unique_ptr<QPainter>(new QPainter());
    if (!painter)
        return;

    // Figure out which group we should use.
    QPalette::ColorGroup pal_group;
    if (hasFocus())
        pal_group = QPalette::Active;
    else if (_enabled)
        pal_group = QPalette::Inactive;
    else
        pal_group = QPalette::Disabled;

    // Create our button gradient. This creates a "curve" look to the base of the button
    // going from top to bottom. This is what the user will see in the "off" position.
    _lg = QLinearGradient(width() * .5, height(), width() * .5, 0);
    _lg.setColorAt(0, my_pal.color(pal_group, QPalette::Mid));
    _lg.setColorAt(0.25, my_pal.color(pal_group, QPalette::Button));
    _lg.setColorAt(0.75, my_pal.color(pal_group, QPalette::Button));
    _lg.setColorAt(1, my_pal.color(pal_group, QPalette::Mid));

    // Start our painting.
    painter->begin(this);
    painter->setRenderHint(QPainter::Antialiasing, true);
    painter->setPen(pen);

    // Draw the outline of the widget.
    painter->setBrush(my_pal.color(pal_group, (hasFocus() ? QPalette::Highlight : QPalette::Shadow)));
    painter->drawRoundedRect(0, 0, width(), height(), _borderradius, _borderradius);

    // Draw the background of the button.
    // Inset the background slightly from the outline.
    painter->setBrush(_lg);
    painter->drawRoundedRect(1, 1, width() - 2, height() - 2, _borderradius - 2, _borderradius - 2);

    // Done painting the widget.
    painter->end();
}

void fiSwitchButton::mousePressEvent(QMouseEvent *)
{
    // Skip if we're not enabled.
    if (!_enabled)
        return;

    // Make sure to grab the focus
    setFocus();

    // Assigning new current value
    setValue(!_value);
}

void fiSwitchButton::resizeEvent(QResizeEvent *event)
{
    int offset;
    int lbl_height;
    QRect bound_on;
    QRect bound_off;

    // Pass the information on up the chain.
    QWidget::resizeEvent(event);

    // Resize the background of our "switch". Account for the border (top and bottom).
    _background->resize(height() - (2 * SWITCH_BORDER), height() - (2 * SWITCH_BORDER));
    _circle->setFixedSize(height() - (2 * SWITCH_BORDER), height() - (2 * SWITCH_BORDER));

    // Figure out where to place the labels.
    bound_on = _labelon->fontMetrics().boundingRect(_labelon->text());
    bound_off = _labeloff->fontMetrics().boundingRect(_labeloff->text());
    if (bound_on.height() > bound_off.height())
        lbl_height = bound_on.height();
    else
        lbl_height = bound_off.height();

    // Center the "off" label to the right of the switch.
    offset = (int)(((width() - _circle->width()) / 2) - (bound_off.width() * 0.5)) + _circle->width();
    _labeloff->move(offset, (height() / 2) - (lbl_height / 2));

    // Place the "on" label to the left of the switch.
    offset = (int)(((width() - _circle->width()) / 2) - (bound_on.width() * 0.5));
    _labelon->move(offset, (height() / 2) - (lbl_height / 2));

    // Force a repaint event.
    repaint();
}

/*!
 * @brief Enable/disable the switch.
 * @param flag      True to enable the switch, false otherwise.
 */
void fiSwitchButton::setEnabled(bool flag)
{
    // Save the new setting.
    _enabled = flag;

    // Pass it on to the kids.
    _circle->setEnabled(flag);
    _background->setEnabled(flag);

    // If we're not enabled, make sure we only show the appropriate label.
    if (!flag)
    {
        // Show the "on" label, only if the button is "checked".
        if (value())
        {
            _labelon->show();
            _labeloff->hide();
        }
        else
        {
            _labelon->hide();
            _labeloff->show();
        }
    }

    // Pass it on up the chain.
    _labelon->setEnabled(flag);
    _labeloff->setEnabled(flag);
    QWidget::setEnabled(flag);
}

/*!
 * @brief Specify how quickly the button moves. This is the duration of the animation.
 * @param time  The number of milliseconds to take for the animation(s).
 */
void fiSwitchButton::setDuration(int time)
{
    _duration = time;
}

/*!
 * @brief Initialize our button style based on the parameter.
 *        Not to be confused with the Qt "setStyle" method.
 *
 * @param style One of fiSwitchButton::Style values.
 */
void fiSwitchButton::setStyle(fiSwitchButton::Style style)
{
    switch (style)
    {
    case Style::ONOFF:
        _labeloff->setText(tr("Off"));
        _labelon->setText(tr("On"));
        break;
    case Style::YESNO:
        _labeloff->setText(tr("No"));
        _labelon->setText(tr("Yes"));
        break;
    case Style::BOOL:
        _labeloff->setText(tr("False"));
        _labelon->setText(tr("True"));
        break;
    default:
        _labeloff->setText("");
        _labelon->setText("");
        style = EMPTY;
        break;
    }

    // Remember what the style is set to.
    _style = style;
}

/*!
 * @brief Return the style (fiSwitchButton::Style) of the switch.
 * @return One of fiSwitchButton::Style.
 */
fiSwitchButton::Style fiSwitchButton::style() const
{
    return _style;
}

/*!
 * @brief Set the value of the widget.
 * @param flag  True or false, to set the widget.
 */
void fiSwitchButton::setValue(bool flag)
{
    // If the new value is the same as the old value, just return.
    if (flag == value())
        return;

    // Update our internal copy of the value, and reflect that int
    // the widget.
    _value = flag;

    // Update the UI.
    _update();

    // Let folks know things changed.
    emit valueChanged(_value);
}

/*!
 * @brief Return the value of the switch.
 * @return True or false.
 */
bool fiSwitchButton::value() const
{
    return _value;
}

void fiSwitchButton::_update()
{
    QSize initial_size;
    QSize final_size;

    int hback;
    int xi;
    int xf;
    int y;

    // Skip if we're not enabled.
    if (!_enabled)
        return;

    // Cancel any animations which might be in-progress.
    __btn_move->stop();
    __back_move->stop();

    // Update the animation durations.
    __btn_move->setDuration(_duration);
    __back_move->setDuration(_duration);

    // Initialize our values.
    hback = _circle->width();
    initial_size = QSize(hback, hback);
    y  = SWITCH_BORDER;

    // Are we "on" or "off"?
    if (_value)
    {
        // Turning ON.
        final_size = QSize(width() - (2 * SWITCH_BORDER), hback);
        xi = SWITCH_BORDER;
        xf = width() - (hback + SWITCH_BORDER);

        // Hide the OFF, show the ON.
        _labelon->show();
        _labeloff->hide();
    }
    else
    {
        // We are turning off.
        final_size = QSize(hback, hback);
        xi = width() - (hback + SWITCH_BORDER);
        xf = SWITCH_BORDER;

        // Hide the ON, show the OFF.
        _labelon->hide();
        _labeloff->show();
    }

    // Set the start and end point for the animations.
    __btn_move->setStartValue(QPoint(xi, y));
    __btn_move->setEndValue(QPoint(xf, y));

    __back_move->setStartValue(initial_size);
    __back_move->setEndValue(final_size);

    // Start the animation.
    __btn_move->start();
    __back_move->start();
}

fiSwitchButton::SwitchBackground::SwitchBackground(QWidget *parent, QColor color, bool rect) :
    QWidget(parent),
    _rect(rect),
    _borderradius(SWITCH_RADIUS),
    _color(color)
{
    // If we're supposed to be a rectangle, then no radius.
    if (_rect)
        _borderradius = 0;

    // Default to being enabled.
    _enabled = true;
}

fiSwitchButton::SwitchBackground::~SwitchBackground()
{
}

void fiSwitchButton::SwitchBackground::paintEvent(QPaintEvent *)
{
    std::unique_ptr<QPainter> painter;
    QPen pen(Qt::NoPen);
    QLinearGradient _lg;

    // Allocate our painter.
    painter = std::unique_ptr<QPainter>(new QPainter());
    if (!painter)
        return;

    // Create our gradient. This is the "ON" background.
    _lg = QLinearGradient(0, height(), width(), 0);
    _lg.setColorAt(0, _color);
    _lg.setColorAt(0.45, _color.lighter());
    _lg.setColorAt(1, _color);

    // Start painting.
    painter->begin(this);
    painter->setRenderHint(QPainter::Antialiasing, true);
    painter->setPen(pen);

    // Paint the background.
    painter->setBrush(_lg);
    painter->drawRoundedRect(0, 0, width(), height(), _borderradius - 2, _borderradius - 2);

    // All done.
    painter->end();
}

void fiSwitchButton::SwitchBackground::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
}

void fiSwitchButton::SwitchBackground::setEnabled(bool flag)
{
    _enabled = flag;
}

fiSwitchButton::SwitchCircle::SwitchCircle(QWidget *parent, QColor color, bool rect) :
    QWidget(parent),
    _rect(rect),
    _borderradius(SWITCH_RADIUS),
    _color(color)
{
    _enabled = true;
}

fiSwitchButton::SwitchCircle::~SwitchCircle()
{
}

void fiSwitchButton::SwitchCircle::paintEvent(QPaintEvent *)
{
    std::unique_ptr<QPainter> painter;
    QPen pen(Qt::NoPen);
    QRadialGradient _rg;

    // Allocate our painter.
    painter = std::unique_ptr<QPainter>(new QPainter());
    if (!painter)
        return;

    // Get the system palette.
    QPalette my_pal = QApplication::palette();

    // Figure out which group we should use.
    QPalette::ColorGroup pal_group;
    if (_enabled)
        pal_group = QPalette::Active;
    else
        pal_group = QPalette::Disabled;

    // Make a nice radial for our round switch button.
    _rg = QRadialGradient((width() / 2), (height() / 2), _borderradius);
    _rg.setColorAt(0, _color);
    _rg.setColorAt(0.6, _color.lighter());
    _rg.setColorAt(1, _color.darker());

    // Start the painting.
    painter->begin(this);
    painter->setRenderHint(QPainter::Antialiasing, true);
    painter->setPen(pen);

    // Draw the shadow.
    painter->setBrush(my_pal.color(pal_group, QPalette::Shadow));
    painter->drawEllipse(0, 0, width(), height());

    // Draw the button, slightly inset from the shadow.
    painter->setBrush(_rg);
    painter->drawEllipse(1, 1, width() - 2, height() - 2);

    // All done.
    painter->end();
}

void fiSwitchButton::SwitchCircle::resizeEvent(QResizeEvent *event)
{
    QWidget::resizeEvent(event);
}

void fiSwitchButton::SwitchCircle::setEnabled(bool flag)
{
    _enabled = flag;
}
