/*!
 * @file fi_tableview.cpp
 * @brief This provides the implementation for the fiTableView class.
 *
 * We needed to subclass the QTableView class so that we can catch
 * and implement the drag/drop functionality.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "fi_tableview.h"

#include <QApplication>
#include <QDrag>
#include <QHostAddress>
#include <QMdiSubWindow>
#include <QMimeData>
#include <QMouseEvent>
#include <QStandardItemModel>

#include <qendian.h>

#include "camera_dock.h"
#include "logger.h"

/*!
 * @brief The fiTableView class constructor.
 */
fiTableView::fiTableView(QWidget *parent) :
    QTableView(parent)
{
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief We override the mouseMoveEvent so we can customize the data
 *        associated with a drag event. If the user isn't dragging
 *        we pass the event on to the default QTableView object handler.
 *
 * @param event The Qt provided event structure/information.
 */
void fiTableView::mouseMoveEvent(QMouseEvent *event)
{
    int sel_row;
    QByteArray          mdata;
    QDataStream         stream(&mdata, QIODevice::WriteOnly);
    QMap<int, QVariant> roleDataMap;
    QStandardItem       *item_ptr;
    QStandardItemModel  *model_data;

    // Make sure we have a drag event here.
    if ((event->buttons() != Qt::LeftButton) || ((drag_start_pos - event->pos()).manhattanLength() < QApplication::startDragDistance()))
    {
        // Pass it up.
        QTableView::mouseMoveEvent(event);
        return;
    }

    // Get the item pointer associated with the currently selected row.
    sel_row = currentIndex().row();
    model_data = qobject_cast<QStandardItemModel *>(model());
    item_ptr = model_data->item(sel_row, CAMERA_DOCK::CD_IP_ADDR);

    // Make sure we got something.
    if (item_ptr == nullptr)
    {
        // Pass it up.
        QTableView::mouseMoveEvent(event);
        return;
    }

    // Build up the "drag" data for this row.
    roleDataMap[0] = QVariant(item_ptr->text());
    stream << sel_row << CAMERA_DOCK::CD_IP_ADDR << roleDataMap;

    QMimeData *mimeData = new QMimeData;
    mimeData->setText(item_ptr->text());
    mimeData->setData("application/x-qabstractitemmodeldatalist", mdata);

    // Start the drag operation.
    QDrag *drag = new QDrag(this);
    drag->setMimeData(mimeData);
    drag->exec(Qt::CopyAction);
}

/*!
 * @brief We override the mousePressEvent so we can capture a drag event.
 *        If the user isn't initiating a drag (pressing with the left button),
 *        we pass the event on to the default QTableView object handler.
 *
 * @param event The Qt provided event structure/information.
 */
void fiTableView::mousePressEvent(QMouseEvent *event)
{
    // Save the position that the user started the (possible) drag operation.
    if (event->buttons() == Qt::LeftButton)
        drag_start_pos = event->pos();

    // Pass it on.
    QTableView::mousePressEvent(event);
}

