/*!
 * @file fi_updater.cpp
 * @brief The implementation of the Fastec Updater object.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QDateTime>
#include <QDir>
#include <QFile>
#include <QMessageBox>
#include <QNetworkAccessManager>
#include <QNetworkReply>
#include <QSettings>
#include <QStandardPaths>
#include <QStringLiteral>
#include <QTime>
#include <QTimer>
#include <QtNetwork>
#include <QUrl>
#include <QVersionNumber>

#include "fi_updater.h"
#include "fi_progress_dialog.h"

#include "mainwindow.h"
#include "qstringliteral.h"

namespace FI_UPDATER
{
//! The default URL for the Fastec update server. This may be overwritten
//! by a settings value.
const char UPDATE_SERVER_URL[] = "http://upd.fastecimaging.com";

//! The QSettings key for the update path when downloading an update package.
const char UPDATE_PATH[] = "update_path";

//! Location of the Windows "latest" file.
const char WIN_LATEST[] = "/win/FasMotion/latest";

//! Location of the Mac "latest" file.
const char MAC_LATEST[] = "/mac/FasMotion/latest";

//! Location of the Linux "latest" file for FasMotion.
const char LINUX_FM_LATEST[] = "/linux/FasMotion/latest";

//! Location of the Series3 camera "latest" file.
const char CAM_S3_LATEST[] = "/win/cameraS3/latest";

//! Location of the SeriesNC camera "latest" file.
const char CAM_NC_LATEST[] = "/linux/cameraNC/latest";

//! Location of the Windows FasMotion installer package.
const char WIN_FM_INSTALLER[] = "/win/FasMotion/rev%1/FasMotion_%1.exe";

//! Location of the Mac FasMotion installer package.
const char MAC_FM_INSTALLER[] = "/mac/FasMotion/rev%1/FasMotion_%1.dmg";

//! Location of the Linux FasMotion installer package.
const char LINUX_FM_INSTALLER[] = "/linux/FasMotion/rev%1/FasMotion_%1.tar.xz";

//! Location of the series3 camera update package.
const char WIN_CAM_S3_INSTALLER[] = "/win/cameraS3/rev%1/fastec_%1.exe";

//! Location of the seriesNC camera update package.
const char CAM_NC_INSTALLER[] = "/linux/cameraNC/rev%1/fastec_%1.tar.xz";

//! Define the "states" (i.e., what are we looking at?) for the updater.
typedef enum
{
    US_FM_LATEST,               //!< Update is looking at the FasMotion "latest" file.
    US_CAM_S3_LATEST,           //!< Update is looking at the series3 camera software "latest" file.
    US_CAM_NC_LATEST,           //!< Update is looking at the seriesNC camera software "latest" file.
    US_FM_PACKAGE,              //!< Update is downloading the FasMotion update package.
    US_CAM_S3_PACKAGE,          //!< Update is downloading the series3 camera software update package.
    US_CAM_NC_PACKAGE           //!< Update is downloading the seriesNC camera software update package.
} UPDATE_STATE;

typedef enum
{
    NO_ERROR,
    INVALID_URL,
    MEMORY_FAULT,
    OS_NOT_SUPPORTED,

} UPDATE_INTERNAL_ERRORS;

static const char *INTERNAL_ERROR_STR[] =
{
    "No error",
    "URL is invalid!",
    "Memory allocation error!",
    "Operating system is not supported!",
};

} // End of FI_UPDATER

//--------------------------------------------------------------------
// Class Constants
//--------------------------------------------------------------------
const QString FiUpdater::UPDATE_GROUP_TAG("Updater");
const QString FiUpdater::UPDATE_AUTO_CHECK_KEY("auto_check");
const QString FiUpdater::UPDATE_FREQUENCY_KEY("update_frequency");
const QString FiUpdater::LAST_CHECK_TIME_KEY("last_check");
const QString FiUpdater::FM_UPDATE_VERSION("fm_update_ver");
const QString FiUpdater::CAM_S3_UPDATE_VERSION("cam_s3_update_ver");
const QString FiUpdater::CAM_NC_UPDATE_VERSION("cam_NC_update_ver");
const QString FiUpdater::UPDATE_SERVER_KEY("server_url");

using namespace FI_UPDATER;

FiUpdater::FiUpdater(QObject *parent) :
    QObject(parent),
    nam(std::unique_ptr<QNetworkAccessManager>(new QNetworkAccessManager())),
    netReply(nullptr),
    internal_error(0),
    timer_id(0),
    updates_avail(AVAIL_NONE),
    progressDialog(fiProgressDialog::CreateSimpleProgressDialog(tr("Updater Progress"), tr("Download Progress"), QMessageBox::Cancel, __PRETTY_FUNCTION__)),
    download_file(nullptr)
{
    // Display the progress dialog.
    if (progressDialog)
    {
        progressDialog->setBarMinMax(0, 100);
        progressDialog->showBar(true);
        progressDialog->hideDialog();

        connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &FiUpdater::cancelDownload);
    }

    // Fire a timer to check if an update is available.
    QTimer::singleShot(10, this, &FiUpdater::delayedStartup);
}

/*!
 * @brief Updater class destructor.
 */
FiUpdater::~FiUpdater()
{
    if (netReply != nullptr)
    {
        netReply->deleteLater();
        netReply = nullptr;
    }
    if (progressDialog)
    {
        progressDialog.reset();
    }
    if (download_file)
    {
        download_file->close();
        delete download_file;
        download_file = nullptr;
    }
}

void FiUpdater::delayedStartup()
{
    QSettings settings;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Load information from the settings database.
    settings.beginGroup(UPDATE_GROUP_TAG);

    update_path = settings.value(UPDATE_PATH).toString();
    if (update_path.isEmpty())
        update_path = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0);

    // Are we doing automatic updates?
    bool autoUpdates = settings.value(UPDATE_AUTO_CHECK_KEY, true).toBool();

    // All done with the settings.
    settings.endGroup();

    // No updates available. Are we doing automatic updates?
    if (autoUpdates)
    {
        // Schedule for an update check.
        scheduleCheck();
    }
}

/*!
 * @brief Return the netReply object error code.
 */
int FiUpdater::GetError() const
{
    // Sanity check.
    if (netReply == nullptr)
    {
        // Return our internal error, if any.
        return internal_error;
    }

    // Return the error.
    return netReply->error();
}

/*!
 * @brief Returns the string associated with the netReply object error value.
 */
QString FiUpdater::GetErrorString() const
{
    // Sanity check.
    if (netReply == nullptr)
    {
        // Must be an internal error. Return the associated error string.
        return QString(INTERNAL_ERROR_STR[internal_error]);
    }

    // Return the error string.
    return netReply->errorString();
}

/*!
 * @brief Get the fully qualified name to the specified update
 *        package.
 *
 * @param which_update  Which package are we interested in [\sa FI_UPDATER::UPDATE_AVAIL].
 * @return The fully qualified package name.
 */
QString FiUpdater::GetPackage(int which_update) const
{
    QString package;

    switch (which_update)
    {
    case AVAIL_FM_UPDATE:
        // Build up the path to the FasMotion installer.
        package = update_path + "/FasMotion_" + fm_ver_avail.toString();

#ifdef Q_OS_WIN
        // In Windows, the installer has an ".exe" on the end of it.
        package += ".exe";
#endif
#ifdef Q_OS_MAC
        // In MacOS, the installer has an ".dmg" on the end of it.
        package += ".dmg";
#endif
#ifdef Q_OS_LINUX
        // In Linux, the installer is a tarball and ends with ".tgz".
        package += ".tgz";
#endif
        break;
    case AVAIL_CAM_S3_UPDATE:
        // Build up the path to the series3 camera installer.
        package = update_path + "/fastec_" + camS3_ver_avail.toString();

        // In Windows, the installer has an ".exe" on the end of it.
        package += ".exe";
        break;
    case AVAIL_CAM_NC_UPDATE:
        // Build up the path to the seriesNC camera update package.
        package = update_path + "/fastec_" + camNC_ver_avail.toString() + ".tar.xz";
        break;

    default:
        break;
    }

    // Return the fully qualified package name.
    return package;
}

/*!
 * @brief Return which, if any, updates are available.
 */
int FiUpdater::GetUpdates() const
{
    return updates_avail;
}

/*!
 * @brief Return the base path to where the update packages are located.
 */
QString FiUpdater::GetUpdatePath() const
{
    return update_path;
}

/*!
 * @brief Return the version number for the latest SeriesNC camera software.
 */
QVersionNumber FiUpdater::GetVersionCamNC() const
{
    return camNC_ver_avail;
}

/*!
 * @brief Return the version number for the latest Series3 camera software.
 */
QVersionNumber FiUpdater::GetVersionCamS3() const
{
    return camS3_ver_avail;
}

/*!
 * @brief Return the version number for the latest FasMotion software.
 */
QVersionNumber FiUpdater::GetVersionFM() const
{
    return fm_ver_avail;
}

/*!
 * @brief Determine when the next version check should happen.
 * @param update_freq   How frequently should the checks happen [\sa FI_UPDATER::UPDATE_FREQ].
 * @param last_check    When was the last checked performed.
 * @return The date and time for the next check to happen.
 */
QDateTime FiUpdater::GetNextCheck(int update_freq, const QDateTime& last_check)
{
    QDateTime next_check;

    // Figure out when the next check needs to be.
    switch (update_freq)
    {
    case CHECK_DAILY:
        // Add a day.
        next_check = last_check.addDays(1);
        break;
    default:
    case CHECK_WEEKLY:
        // Add a week (7 days).
        next_check = last_check.addDays(7);
        break;
    case CHECK_MONTHLY:
        // Add a month.
        next_check = last_check.addMonths(1);
        break;
    }

    // Return the calculated date and time for the next check.
    return next_check;
}

/*!
 * @brief Enable or disable automatic update checking.
 * @param do_auto   If true, enable automatic checking.
 *                  If false, disable automatic checking.
 */
void FiUpdater::SetAutomatic(bool do_auto)
{
    QSettings settings;

    // Save the state of the automatic check.
    settings.beginGroup(FiUpdater::UPDATE_GROUP_TAG);
    settings.setValue(FiUpdater::UPDATE_AUTO_CHECK_KEY, do_auto);
    settings.endGroup();

    // Are we allowing automatic updates?
    if (do_auto)
    {
        // Schedule an update check.
        scheduleCheck();
    }
    else
    {
        // Kill the timer, if it exists.
        if (timer_id)
        {
            killTimer(timer_id);
            timer_id = 0;
        }
    }
}

/*!
 * @brief FiUpdater::SetFrequency
 * @param check_freq    How often the automatic check should happen [\sa FI_UPDATER::UPDATE_FREQ].
 */
void FiUpdater::SetFrequency(int check_freq)
{
    QSettings settings;

    // Save the value for the check frequency.
    settings.beginGroup(FiUpdater::UPDATE_GROUP_TAG);
    settings.setValue(FiUpdater::UPDATE_FREQUENCY_KEY, check_freq);
    settings.endGroup();

    // If we have a timer, then we want to update the schedule.
    if (timer_id)
        scheduleCheck();
}

/*!
 * @brief Specify where the update packages should be download to.
 * @param path  The fully qualified path name.
 */
void FiUpdater::SetUpdatePath(const QString& path)
{
    QSettings settings;

    // Save the new path.
    update_path = path;

    // Update our settings with the new value.
    settings.beginGroup(UPDATE_GROUP_TAG);
    settings.setValue(UPDATE_PATH, update_path);
    settings.endGroup();
}

/*!
 * @brief Check for update availability.
 *
 * By firing a timer with a 0 duration, we're effectively requesting
 * that an event be put on the queue for immediate processing.
 */
void FiUpdater::UpdateCheck() const
{
    // Caller wants an immediate update.
    QTimer::singleShot(0, this, &FiUpdater::timeForUpdate);
}

/*!
 * @brief Download an update package.
 * @param which_update  Which package are we interested in [\sa FI_UPDATER::UPDATE_AVAIL].
 * @return TRUE if the download process starts, FALSE otherwise.
 */
bool FiUpdater::UpdateDownload(int which_update)
{
    QSettings   settings;
    QString     base_path;
    QString     os_dep_path;
    QString     download_path;
    QUrl        download_url;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "which_update:" << which_update;

    // Load information from the settings database.
    settings.beginGroup(UPDATE_GROUP_TAG);

    // See if there is an updated base URL in the settings. If not,
    // use the default value.
    base_path = settings.value(UPDATE_SERVER_KEY).toString();
    if (base_path.isEmpty())
        base_path = QString(UPDATE_SERVER_URL);

    // All done with the settings.
    settings.endGroup();

    // What are we asking for.
    switch (which_update)
    {
    case AVAIL_FM_UPDATE:
        // Remember what state we're in.
        state = US_FM_PACKAGE;

#ifdef Q_OS_WIN
        // We want to retrieve the FasMotion installer from the server, so build up the path.
        // We're looking for a Windows FasMotion...
        os_dep_path = QString(WIN_FM_INSTALLER).arg(fm_ver_avail.toString());
#endif /* Q_OS_WIN */
#ifdef Q_OS_MAC
        // We want to retrieve the FasMotion installer from the server, so build up the path.
        // We're looking for a Windows FasMotion...
        os_dep_path = QString(MAC_FM_INSTALLER).arg(fm_ver_avail.toString());
#endif /* Q_OS_MAC */
#ifdef Q_OS_LINUX
        // We want to retrieve the FasMotion installer from the server, so build up the path.
        // We're looking for a Windows FasMotion...
        os_dep_path = QString(LINUX_FM_INSTALLER).arg(fm_ver_avail.toString());
#endif /* Q_OS_LINUX */
        break;
    case AVAIL_CAM_S3_UPDATE:
        // Remember what state we're in.
        state = US_CAM_S3_PACKAGE;

        // We want to retrieve the series3 camera installer from the server, so build up the path.
        // We're looking for a Windows installation package...
        os_dep_path = QString(WIN_CAM_S3_INSTALLER).arg(camS3_ver_avail.toString());
        break;
    case AVAIL_CAM_NC_UPDATE:
        // Remember what state we're in.
        state = US_CAM_NC_PACKAGE;

        // We want to retrieve the seriesNC camera installer from the server, so build up the path.
        // There's not really an OS dependent component here, 'cuz this is a tarball we're grabbing.
        os_dep_path = QString(CAM_NC_INSTALLER).arg(camNC_ver_avail.toString());
        break;

    default:
        break;
    }

    // Put together the entire source path.
    download_path = base_path + os_dep_path;

    // Get a URL from the path we just built.
    download_url = download_path;
    if (!download_url.isValid())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "URL is not valid:" << download_url;
        internal_error = INVALID_URL;
        return false;
    }
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "download_url:" << download_url.toDisplayString();

    // Open the destination file.
    download_file = new QFile(GetPackage(which_update));
    if (!download_file)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error creating file object!";
        internal_error = MEMORY_FAULT;
        return false;
    }

    // Make sure we can open the file for writing.
    if (!download_file->open(QIODevice::WriteOnly))
    {
        QMessageBox::information(nullptr, tr("Error"),
                                 tr("Unable to save the file %1: %2.")
                                 .arg(QDir::toNativeSeparators(GetPackage(which_update)), download_file->errorString()));
        internal_error = NO_ERROR;
        return false;
    }

    // Setup our progress dialog.
    if (progressDialog)
    {
        progressDialog->setStatusValue(tr("Downloading %1").arg(download_url.toDisplayString()));
        progressDialog->setBarValue(0);
        progressDialog->setBarMinMax(0, 100);
        progressDialog->showCancel(true);
        progressDialog->showDialogWithoutPolling();
    }

    // Start the download process.
    return startRequest(download_url);
}

/*!
 * @brief The user has cancelled the download operation.
 */
void FiUpdater::cancelDownload()
{
    // Stop the network operation.
    if (netReply && netReply->isRunning())
        netReply->abort();
}

/*!
 * @brief The QNetworkReply object has terminated with an error.
 * @param err_value The reason why we failed.
 */
void FiUpdater::httpError(int err_value)
{
    QLOG_ERROR() << __PRETTY_FUNCTION__ << "err_value:" << err_value << "Update error:" << netReply->errorString();

    // Hide the progress dialog if it is visible.
    if (progressDialog)
        progressDialog->hideDialog();

    // Show the error in the statusbar.
    MainWindow::GetInstance()->set_statusbar_message(tr("Update error: %1").arg(netReply->errorString()), 5000);

    // Let folks know there was a problem.
    emit error(err_value);
}

/*!
 * @brief The QNetworkReply object has finished the download operation.
 */
void FiUpdater::httpFinished()
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "state:" << state;

    // Hide the progress dialog if it is visible.
    if (progressDialog)
        progressDialog->hideDialog();

    // All done with the file, so close it up.
    if (download_file)
    {
        download_file->close();
        delete download_file;
        download_file = nullptr;
    }

    // Sanity check.
    if (netReply == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Null netReply!?";
        return;
    }

    // Check for an error.
    if (netReply->error())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Download failed:" << netReply->errorString();

        // Let folks know the result of the check for updates.
        emit finished(updates_avail);

        // All done.
        return;
    }

    // What are we doing?
    switch (state)
    {
    case US_FM_LATEST:
    case US_CAM_S3_LATEST:
    case US_CAM_NC_LATEST:
        // *INDENT-OFF*
        {
            QString latest_str;

            // Read up the bytes and convert to our string.
            latest_str = QString(netReply->readAll());

            // Finish processing the "latest" information.
            processLatest(latest_str);
        }
        // *INDENT-ON*
        break;
    case US_FM_PACKAGE:
        emit download_complete(AVAIL_FM_UPDATE);
        break;
    case US_CAM_S3_PACKAGE:
        emit download_complete(AVAIL_CAM_S3_UPDATE);
        break;
    case US_CAM_NC_PACKAGE:
        emit download_complete(AVAIL_CAM_NC_UPDATE);
        break;
    }
}

/*!
 * @brief This slot gets called every time the QNetworkReplay object
 *        has new data.
 *
 * If we are downloading a package (i.e. not the "latest" file), we will
 * want to save the received data to the destination file. That way we
 * use less memory than if we waited until the QNetworkReply::finished() signal.
 */
void FiUpdater::httpReadyRead()
{
    // Sanity check.
    if (netReply == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Null netReply!?";
        return;
    }

    // The latest files are very small, so we can wait until the finished() signal.
    if ((state == US_FM_LATEST) || (state == US_CAM_S3_LATEST) || (state == US_CAM_NC_LATEST))
        return;

    if (download_file)
    {
        // Write the data out to the file.
        download_file->write(netReply->readAll());

        // ALM_TODO: Add md5 checksum?
    }
}

/*!
 * @brief Update our progress dialog, if it is visible. Then pass on the progress signal.
 * @param bytesReceived The number of bytes received.
 * @param bytesTotal    The total number of bytes expected.
 */
void FiUpdater::httpProgress(qint64 bytesReceived, qint64 bytesTotal)
{
    // Update our progress dialog,.
    if (progressDialog)
    {
        progressDialog->setBarMax((int)bytesTotal);
        progressDialog->setBarValue((int)bytesReceived);
    }

    // Pass it on to anyone who might care.
    emit progress(bytesReceived, bytesTotal);
}

/*!
 * @brief It is time to check for update availability!
 */
void FiUpdater::timeForUpdate()
{
    QSettings   settings;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Initialize our internal error indicator.
    internal_error = NO_ERROR;

    // Load information from the settings database.
    settings.beginGroup(UPDATE_GROUP_TAG);

    // All done with the settings.
    settings.endGroup();

    // Currently assuming there are no updates available.
    updates_avail = AVAIL_NONE;

    // Check to see if we have an updated FasMotion.
    fmUpdateCheck();
}

/*!
 * @brief Check to see if we have a camera software update available.
 * @param up_state  Determine which camera we're interested in by the
 *                  update state [\sa FI_UPDATER::UPDATE_STATE].
 */
bool FiUpdater::camUpdateCheck(int up_state)
{
    QSettings   settings;
    QString     base_path;
    QString     latest_path;
    QUrl        latest_url;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "state:" << up_state;

    // Load information from the settings database.
    settings.beginGroup(UPDATE_GROUP_TAG);

    // See if there is an updated base URL in the settings. If not,
    // use the default value.
    base_path = settings.value(UPDATE_SERVER_KEY).toString();
    if (base_path.isEmpty())
        base_path = QString(UPDATE_SERVER_URL);

    // All done with the settings.
    settings.endGroup();

    // We want to retrieve the "latest" file from the server, so build up the path.
    if (up_state == US_CAM_NC_LATEST)
        latest_path = base_path + QString(CAM_NC_LATEST);
    else
        latest_path = base_path + QString(CAM_S3_LATEST);

    // Get a URL from the path we just built.
    latest_url = latest_path;
    if (!latest_url.isValid())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "URL is not valid:" << latest_path;
        internal_error = INVALID_URL;
        return false;
    }

    // Remember what state we're in.
    state = up_state;

    // Start the download process.
    return startRequest(latest_url);
}

/*!
 * @brief Check to see if there's a FasMotion update available.
 * @returns True if the check has started, false if there was an error.
 */
bool FiUpdater::fmUpdateCheck()
{
    QSettings   settings;
    QString     base_path;
    QString     latest_path;
    QUrl        latest_url;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

#ifdef Q_OS_LINUX
    // We ARE running on Linux! If we also have local control, then we
    // do not want to check for a FasMotion update at this time.
    if (MainWindow::local_control())
    {
        // Go check to see if we want to check for series3 camera software updates.
        return camUpdateCheck(US_CAM_S3_LATEST);
    }
#endif /* Q_OS_LINUX */

    // Load information from the settings database.
    settings.beginGroup(UPDATE_GROUP_TAG);

    // See if there is an updated base URL in the settings. If not,
    // use the default value.
    base_path = settings.value(UPDATE_SERVER_KEY).toString();
    if (base_path.isEmpty())
    {
        base_path = QString(UPDATE_SERVER_URL);
    }

    // All done with the settings.
    settings.endGroup();

#ifdef Q_OS_WIN
    // We want to retrieve the "latest" file from the server, so build up the path.
    // We're looking for a Windows FasMotion...
    latest_path = base_path + QString(WIN_LATEST);
#elif defined(Q_OS_MAC)
    // We want to retrieve the "latest" file from the server, so build up the path.
    // We're looking for a Mac FasMotion...
    latest_path = base_path + QString(MAC_LATEST);
#else /* Q_OS_LINUX */
    // We want to retrieve the "latest" file from the server, so build up the path.
    // We're looking for a Linux FasMotion...
    latest_path = base_path + QString(LINUX_FM_LATEST);
#endif /* Q_OS_LINUX */

    // Get a URL from the path we just built.
    latest_url = latest_path;
    if (!latest_url.isValid())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "URL is not valid:" << latest_path;
        internal_error = INVALID_URL;
        return false;
    }

    // Remember what state we're in.
    state = US_FM_LATEST;

    // Start the download process.
    return startRequest(latest_url);
}

/*!
 * @brief Finish checking to see if there is an update.
 * @param latest    The latest information downloaded from the server.
 */
void FiUpdater::processLatest(const QString& latest)
{
    QSettings   settings;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "state:" << state << "latest:" << latest;

    // Start access to the settings database.
    settings.beginGroup(UPDATE_GROUP_TAG);

    // Save the current time to the settings database.
    settings.setValue(LAST_CHECK_TIME_KEY, QDateTime::currentDateTime());

    // Convert the received version information.
    QVersionNumber latest_ver = QVersionNumber::fromString(latest);

    // Are we checking the FasMotion version information?
    switch (state)
    {
    case US_FM_LATEST:
        // Did we actually get a version?
        if (!latest_ver.isNull())
        {
            settings.setValue(FM_UPDATE_VERSION, latest_ver.toString());
            updates_avail |= AVAIL_FM_UPDATE;
            fm_ver_avail = latest_ver;
        }

        // Now go check to see if we want to check for series3 camera software updates.
        camUpdateCheck(US_CAM_S3_LATEST);
        break;
    case US_CAM_S3_LATEST:
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Series3 Latest:" << latest_ver.toString();

        // Did we actually get a version?
        if (!latest_ver.isNull())
        {
            settings.setValue(CAM_S3_UPDATE_VERSION, latest_ver.toString());
            updates_avail |= AVAIL_CAM_S3_UPDATE;
            camS3_ver_avail = latest_ver;
        }

        // Now go check to see if we want to check for seriesNC camera software updates.
        camUpdateCheck(US_CAM_NC_LATEST);
        break;
    case US_CAM_NC_LATEST:
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "SeriesNC Latest:" << latest_ver.toString();

        // Did we actually get a version?
        if (!latest_ver.isNull())
        {
            settings.setValue(CAM_NC_UPDATE_VERSION, latest_ver.toString());
            updates_avail |= AVAIL_CAM_NC_UPDATE;
            camNC_ver_avail = latest_ver;
        }

        // We are done. Let folks know what we've found.
        emit finished(updates_avail);
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected state!";
        break;
    }

    // Done with the settings.
    settings.endGroup();
}

/*!
 * @brief Schedule the time for an update check.
 */
void FiUpdater::scheduleCheck()
{
    int update_freq;
    qint64 time_val;
    QDateTime last_check;
    QDateTime next_check;
    QSettings settings;

    // Start access to the settings database.
    settings.beginGroup(UPDATE_GROUP_TAG);

    update_freq = settings.value(UPDATE_FREQUENCY_KEY, -1).toInt();
    last_check = settings.value(LAST_CHECK_TIME_KEY, QDateTime::currentDateTime()).toDateTime();

    // Have we ever initalized these values?
    if (update_freq == -1)
    {
        // Use default values.
        update_freq = CHECK_WEEKLY;
        settings.setValue(UPDATE_FREQUENCY_KEY, CHECK_WEEKLY);
        last_check = QDateTime::currentDateTime();
        settings.setValue(LAST_CHECK_TIME_KEY, last_check);
    }

    // Due to limitations in the QTimer, we can't set a timer out
    // for a month. Fake it to a week.
    if (update_freq == CHECK_MONTHLY)
        update_freq = CHECK_WEEKLY;

    // Get the next check time based on the update frequency.
    next_check = FiUpdater::GetNextCheck(update_freq, last_check);

    // Make sure there isn't already a timer running.
    if (timer_id)
    {
        killTimer(timer_id);
        timer_id = 0;
    }

    // How many milliseconds are there until we need to check.
    time_val = QDateTime::currentDateTime().msecsTo(next_check);

    // Should we do the check immediately?
    if (time_val <= 0)
    {
        // We missed the check, so do one now.
        QTimer::singleShot(0, this, &FiUpdater::timeForUpdate);
    }
    else
    {
        // Make sure we keep things in bounds.
        if (time_val > INT_MAX)
            time_val = INT_MAX - 1;

        // Start a timer.
        timer_id = startTimer(std::chrono::milliseconds(time_val));
    }

    // All done with the settings.
    settings.endGroup();
}

/*!
 * @brief Start the download request for the object specified in the requestedUrl parameter.
 *
 * @param requestedUrl  The URL for the requested object, e.g. the "latest" file, or update package.
 * @return True if the start was successfull, false otherwise.
 */
bool FiUpdater::startRequest(const QUrl& requestedUrl)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "requestedUrl:" << requestedUrl.toString();

    // Sanity check.
    if (nam.get() == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "NULL Network Access Manager!?";
        internal_error = MEMORY_FAULT;
        return false;
    }

    // Make sure we get rid of any reply objects that could be hanging around.
    // Note that this isn't necessarily an error - we don't automatically clean this
    // up incase the caller wants to retrieve any information after the fact.
    if (netReply != nullptr)
    {
        netReply->deleteLater();
        netReply = nullptr;
    }

    // Start the request and get our reply object.
    netReply = nam->get(QNetworkRequest(requestedUrl));
    if (netReply == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to get a netReply object!";
        internal_error = MEMORY_FAULT;
        return false;
    }

    // Connect up the reply signals to our slots.
    connect(netReply, &QNetworkReply::finished, this, &FiUpdater::httpFinished);
    connect(netReply, &QNetworkReply::readyRead, this, &FiUpdater::httpReadyRead);
    connect(netReply, &QNetworkReply::downloadProgress, this, &FiUpdater::httpProgress);

    // This was taken straight out of the Qt documentation: https://doc.qt.io/qt-5/qnetworkreply.html#error-1.
    // Because "error" is overloaded, we use this convenient helper to obtain the signal function pointer.
    // And really all we want is to re-emit the signal.
    connect(netReply, QOverload<QNetworkReply::NetworkError>::of(&QNetworkReply::errorOccurred), [ = ](QNetworkReply::NetworkError code) { httpError(code); });

    // All done.
    return true;
}

/*!
 * @brief Poll the communications object to see if there are any new cameras
 *        which have responded to our discovery message.
 *
 * @param event_ptr     Pointer to the object which generated this event.
 */
void FiUpdater::timerEvent(QTimerEvent *event_ptr)
{
    // Sanity check.
    if (event_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null event pointer?!";
        return;
    }

    // Proceess based on the event ID.
    if (event_ptr->timerId() == timer_id)
    {
        // We run through the schedule check method again, because we may
        // need to extend the timer.
        scheduleCheck();
    }
}

