/*!
 * @file fi_updater.h
 * @brief This is the header file for the FiUpdater class.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FI_UPDATER_H_
#define FI_UPDATER_H_

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QObject>
#include <QDateTime>
#include <QString>
#include <QVersionNumber>

#include <memory>

namespace FI_UPDATER
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! Define the bitmasks for which, if any, updates are available.
typedef enum
{
    AVAIL_NONE = 0,                 //!< There are no updates available.
    AVAIL_FM_UPDATE = 1,            //!< There is a FasMotion update package available for downloading.
    AVAIL_CAM_S3_UPDATE = 2,        //!< There is a series3 camera software update available for downloading.
    AVAIL_CAM_NC_UPDATE = 4,        //!< There is a seriesNC camera software update available for downloading.
} UPDATE_AVAIL;

//! Define constants for how frequently to check for updates.
typedef enum
{
    CHECK_DAILY,                    //!< Once per day.
    CHECK_WEEKLY,                   //!< Once per week.
    CHECK_MONTHLY                   //!< Once per month.
} UPDATE_FREQ;

} // End of FI_UPDATER

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QFile;
class QNetworkAccessManager;
class QNetworkReply;

class fiProgressDialog;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------

class FiUpdater : public QObject
{
    Q_OBJECT

public:
    explicit FiUpdater(QObject *parent = nullptr);
    ~FiUpdater();

    static const QString UPDATE_GROUP_TAG;           //!< The QSettings group tag.
    static const QString UPDATE_AUTO_CHECK_KEY;      //!< The QSettings key for automatic update checks.
    static const QString UPDATE_FREQUENCY_KEY;       //!< The QSettings key for how frequently we are to check for updates.
    static const QString LAST_CHECK_TIME_KEY;        //!< The QSettings key for the time the last check was performed.
    static const QString FM_UPDATE_VERSION;          //!< The QSettings key for the FasMotion update version. If no update is available, this will be empty.
    static const QString CAM_S3_UPDATE_VERSION;      //!< The QSettings key for the series3 camera update version. If no update is available, this will be empty.
    static const QString CAM_NC_UPDATE_VERSION;      //!< The QSettings key for the seriesNC camera update version. If no update is available, this will be empty.
    static const QString UPDATE_SERVER_KEY;          //!< The QSettings key for the update server URL. If the entry doesn't exist, then the UPDATE_SERVER_URL value is used.

    static QDateTime GetNextCheck(int update_freq, const QDateTime& last_check);

    int GetError() const;
    QString GetErrorString() const;
    QString GetPackage(int which_update) const;
    int GetUpdates() const;
    QString GetUpdatePath() const;
    QVersionNumber GetVersionCamNC() const;
    QVersionNumber GetVersionCamS3() const;
    QVersionNumber GetVersionFM() const;

    void SetAutomatic(bool do_auto);
    void SetFrequency(int check_freq);
    void SetUpdatePath(const QString& path);

    void UpdateCheck() const;
    bool UpdateDownload(int which_update);

signals:
    void download_complete(int which_update);
    void finished(int updates_avail);
    void error(int err_value);
    void progress(qint64 bytesReceived, qint64 bytesTotal);

private slots:
    void cancelDownload();
    void delayedStartup();
    void httpError(int err_value);
    void httpFinished();
    void httpReadyRead();
    void httpProgress(qint64 bytesReceived, qint64 bytesTotal);
    void timeForUpdate();
    void timerEvent(QTimerEvent *event_ptr);

private:
    std::unique_ptr<QNetworkAccessManager> nam;
    QNetworkReply *netReply;

    int internal_error;
    int state;
    int timer_id;
    int updates_avail;
    QString update_path;
    std::unique_ptr<fiProgressDialog> progressDialog;   //!< A simple progress dialog for operational status.
    QFile *download_file;

    QVersionNumber fm_ver_avail;
    QVersionNumber camNC_ver_avail;
    QVersionNumber camS3_ver_avail;

    bool camUpdateCheck(int cam_state);
    bool fmUpdateCheck();
    void processLatest(const QString& latest);
    void scheduleCheck();
    bool startRequest(const QUrl& requestedUrl);
};

#endif /* !FI_UPDATER_H_ */
