/*!
 * @file fi_userdoc_dialog.cpp
 * @brief This is the implementation of the fiUserDocDialog QDialog class.
 *
 * This dialog allows the user to select available Fastec Imaging user documentation
 * and attempts to download them to a local path on the system from the update server.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QFileDialog>
#include <QMessageBox>
#include <QSettings>
#include <QString>

#include "fi_userdoc_dialog.h"
#include "ui_fi_userdoc_dialog.h"

#include "fi_progress_dialog.h"
#include "logger.h"

namespace FI_USERDOC_DLG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

}

using namespace FI_USERDOC_DLG;

//--------------------------------------------------------------------
// Private static data members.
//--------------------------------------------------------------------

//! These constants are for any QSettings used for camera documentation download operation.
const QString fiUserDocDialog::DOC_DOWNLOAD_GROUP            = QStringLiteral("DocDownloader");
const QString fiUserDocDialog::DOC_DOWNLOAD_MANUAL_DIR_KEY   = QStringLiteral("user_manual_path");
const QString fiUserDocDialog::DOC_DOWNLOAD_PATH_KEY         = QStringLiteral("dest_path");
const QString fiUserDocDialog::DOC_DOWNLOAD_SERVER_KEY       = QStringLiteral("server_url");

//! These constants are the default document names for downloading camera documentation and server address.
const QString fiUserDocDialog::DOC_DOWNLOAD_SERVER      = QStringLiteral("http://upd.fastecimaging.com");
const QString fiUserDocDialog::DOC_DOWNLOAD_MANUAL_DIR  = QStringLiteral("/docs/User_Manuals/");
const QString fiUserDocDialog::MANUAL_FASMOTION         = QStringLiteral("FasMotion.pdf");
const QString fiUserDocDialog::MANUAL_HS                = QStringLiteral("HS_Manual.pdf");
const QString fiUserDocDialog::MANUAL_HS_QSG            = QStringLiteral("HS_QSG.pdf");
const QString fiUserDocDialog::MANUAL_IL                = QStringLiteral("IL_Manual.pdf");
const QString fiUserDocDialog::MANUAL_IL_QSG            = QStringLiteral("IL_QSG.pdf");
const QString fiUserDocDialog::MANUAL_TS                = QStringLiteral("TS_Manual.pdf");
const QString fiUserDocDialog::MANUAL_TS_QSG            = QStringLiteral("TS_QSG.pdf");

//--------------------------------------------------------------------
// Class Implementation for fiUserDocDialog
//--------------------------------------------------------------------

/*!
 * @brief The FI User Documentation Dialog constructor.
 */
fiUserDocDialog::fiUserDocDialog(QWidget *parent) :
    QDialog(parent),
    ui(new Ui::fiUserDocDialog)
{
    netReply = nullptr;
    downloadFile = nullptr;
    progressDialog = fiProgressDialog::CreateSimpleProgressDialog(tr("Download Camera Documentation Progress"), tr("Download Progress"), QMessageBox::Cancel, __PRETTY_FUNCTION__);

    ui->setupUi(this);

    // Pull out the QSettings to see if we have a path stored there.
    {
        QSettings settings;
        QString manual_dir;

        settings.beginGroup(fiUserDocDialog::DOC_DOWNLOAD_GROUP);
        manual_dir          = settings.value(fiUserDocDialog::DOC_DOWNLOAD_MANUAL_DIR_KEY).toString();
        downloadPath        = settings.value(fiUserDocDialog::DOC_DOWNLOAD_PATH_KEY).toString();

        // If you need to reset your server_url QSetting value you can uncomment this line and run the program
        // to reset it back to the default or something else for error handling tests.
        // settings.setValue(DOC_DOWNLOAD_SERVER_KEY, DOC_DOWNLOAD_SERVER);
        downloadServerUrl   = settings.value(fiUserDocDialog::DOC_DOWNLOAD_SERVER_KEY).toString();
        settings.endGroup();

        // Do we have a path set up already in the user's settings?
        if (!downloadPath.isEmpty())
        {
            QDir dirChecker(downloadPath);

            if (dirChecker.exists())
            {
                // Make sure the path exists.
                ui->txt_path->setText(downloadPath);
            }
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The path in QSettings:" << downloadPath << "doesn't exist and will be ignored.";
            }
        }

        // Do we have a downloadServerUrl in our QSettings? If not use our hard-wired default.
        if (downloadServerUrl.isEmpty())
        {
            downloadServerUrl = fiUserDocDialog::DOC_DOWNLOAD_SERVER;
        }

        // Do we have a user manual directory in our QSettings? If not use our hard-wired default.
        if (manual_dir.isEmpty())
        {
            manual_dir = fiUserDocDialog::DOC_DOWNLOAD_MANUAL_DIR;
        }

        // Append the user manual directory to our server url.
        downloadServerUrl += manual_dir;

        // Check the URL if it's not valid use our default one.
        {
            QString urlToTest = downloadServerUrl + MANUAL_FASMOTION;
            QUrl CameraManualURL(urlToTest);

            if (!CameraManualURL.isValid())
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The server_url in QSettings:" << downloadServerUrl << " isn't being recognized as valid will be using:" << fiUserDocDialog::DOC_DOWNLOAD_SERVER;

                // Let's go with our hard-wired option.
                downloadServerUrl = fiUserDocDialog::DOC_DOWNLOAD_SERVER;
            }
        }
    }

    connect(ui->pb_cancel, &QPushButton::clicked, this, &fiUserDocDialog::reject);
    connect(ui->chk_fasmotion, &QCheckBox::clicked, this, &fiUserDocDialog::doDownloadEnable);
    connect(ui->chk_hs_camera, &QCheckBox::clicked, this, &fiUserDocDialog::doDownloadEnable);
    connect(ui->chk_hs_quickstart, &QCheckBox::clicked, this, &fiUserDocDialog::doDownloadEnable);
    connect(ui->chk_il_camera, &QCheckBox::clicked, this, &fiUserDocDialog::doDownloadEnable);
    connect(ui->chk_il_quickstart, &QCheckBox::clicked, this, &fiUserDocDialog::doDownloadEnable);
    connect(ui->chk_ts_camera, &QCheckBox::clicked, this, &fiUserDocDialog::doDownloadEnable);
    connect(ui->chk_ts_quickstart, &QCheckBox::clicked, this, &fiUserDocDialog::doDownloadEnable);
    connect(ui->pb_browse, &QPushButton::clicked, this, &fiUserDocDialog::doBrowse);

    connect(ui->pb_download, &QPushButton::clicked, this, &fiUserDocDialog::doDownload);

    // Connect up download document condition handlers.
    connect(this, &fiUserDocDialog::docDownloadFailed, this, &fiUserDocDialog::doDocDownloadFailed);
    connect(this, &fiUserDocDialog::docDownloadFinished, this, &fiUserDocDialog::doDocDownloadFinished);
    connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &fiUserDocDialog::doDocDownloadCanceled);

    // Nothing to download initially.
    ui->pb_download->setEnabled(false);
    ui->txt_path->setEnabled(false);
    ui->pb_browse->setEnabled(false);
}

/*!
 * @brief The FI User Documentation Dialog destructor.
 */
fiUserDocDialog::~fiUserDocDialog()
{
    delete ui;

    if (progressDialog)
    {
        progressDialog.reset();
    }
    if (downloadFile)
    {
        downloadFile->close();
        delete downloadFile;
        downloadFile = nullptr;
    }
}

/*!
 * @brief Checks to see if any of the documents are selected and enables download button if there is at least one.
 * @param checked   True if the item invoking this slot handler was checked.
 */
void fiUserDocDialog::doDownloadEnable(bool checked)
{
    bool one_is_checked = checked;

    // If the item calling this was checked we don't need to check the others.
    if (!checked)
    {
        if (ui->chk_fasmotion->isChecked())
            one_is_checked = true;
        else if (ui->chk_hs_camera->isChecked())
            one_is_checked = true;
        else if (ui->chk_hs_quickstart->isChecked())
            one_is_checked = true;
        else if (ui->chk_il_camera->isChecked())
            one_is_checked = true;
        else if (ui->chk_il_quickstart->isChecked())
            one_is_checked = true;
        else if (ui->chk_ts_camera->isChecked())
            one_is_checked = true;
        else if (ui->chk_ts_quickstart->isChecked())
            one_is_checked = true;
    }

    // Enable or disable based on whether or not at least one is checked.
    ui->pb_download->setEnabled(one_is_checked && !ui->txt_path->text().isEmpty());
    ui->txt_path->setEnabled(one_is_checked);
    ui->pb_browse->setEnabled(one_is_checked);
}

/*!
 * @brief The user clicked on the path browse button and this is the handler to
 * facilitate a selection of the download path.
 */
void fiUserDocDialog::doBrowse()
{
    // Retrieve the current setting for the download path.
    downloadPath = ui->txt_path->text();
    downloadPath = QFileDialog::getExistingDirectory(this, tr("Select A Directory To Save Documents To"), downloadPath);

    // Did we get anything?
    if (downloadPath.isEmpty())
        return;

    // Make sure the path will look "right" to the user.
    downloadPath = QDir::toNativeSeparators(downloadPath);
    ui->txt_path->setText(downloadPath);
    ui->txt_path->setToolTip(tr("Full Path: ") + downloadPath);

    // Just to get a run through of the dialog state.
    doDownloadEnable(false);
}

/*!
 * @brief Responds to the download button being pressed. Determines the list of checked
 * documents to download and begins the process of downloading them.
 */
void fiUserDocDialog::doDownload()
{
    // We have started downloading.
    if (!downloadList.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The download_list is not empty, we are already downloading!";
        return;
    }

    // Make sure there is a path.
    QDir dir_check(ui->txt_path->text());
    if (!dir_check.exists())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The download path:" << ui->txt_path->text() << "cannot be verified.";
        return;
    }

    // Persist this path for future invocations of this dialog.
    {
        QSettings settings;
        settings.beginGroup(fiUserDocDialog::DOC_DOWNLOAD_GROUP);
        settings.setValue(DOC_DOWNLOAD_PATH_KEY, downloadPath);
        settings.endGroup();
    }

    // Build up the list of documents to download.
    if (ui->chk_fasmotion->isChecked())
        downloadList.append(MANUAL_FASMOTION);
    if (ui->chk_hs_camera->isChecked())
        downloadList.append(MANUAL_HS);
    if (ui->chk_hs_quickstart->isChecked())
        downloadList.append(MANUAL_HS_QSG);
    if (ui->chk_il_camera->isChecked())
        downloadList.append(MANUAL_IL);
    if (ui->chk_il_quickstart->isChecked())
        downloadList.append(MANUAL_IL_QSG);
    if (ui->chk_ts_camera->isChecked())
        downloadList.append(MANUAL_TS);
    if (ui->chk_ts_quickstart->isChecked())
        downloadList.append(MANUAL_TS_QSG);

    // Disable the dialog.
    setEnabledState(false);

    // Reset the download status and error codes.
    downloadStatus.clear();
    downloadErrCode = 0;

    // Kickoff the first download in the list.
    if (downloadList.size() > 0)
    {
        downloadADocument(downloadList.front());
    }
}

/*!
 * @brief Begins a request to download a document from the network at url downloadServerUrl.
 * @param documentName    The name of the document which represents the name of the file on the update server.
 */
void fiUserDocDialog::downloadADocument(const QString documentName)
{
    QString urlAddr = downloadServerUrl + documentName;

    // Build up the QT URL object.
    QUrl CameraManualURL(urlAddr);

    // Make sure our download path is set to the one in the dialog.
    downloadPath = ui->txt_path->text();

    // Make sure our document filepath ends with the directory separator.
    if (!downloadPath.endsWith(QLatin1Char('/')))
        downloadPath.append(QLatin1Char('/'));

    // Build up our destination filename.
    QString destFilePath = downloadPath;
    destFilePath.append(documentName);

    // Open the destination file.
    downloadFile = new QFile(destFilePath);
    if (!downloadFile)
    {
        downloadStatus = tr("Resource failure allocating %1.").arg(destFilePath);
        emit docDownloadFailed();
        return;
    }

    // Make sure we can open the file for writing.
    if (!downloadFile->open(QIODevice::WriteOnly))
    {
        downloadStatus = tr("Unable to open %1.").arg(destFilePath);
        QMessageBox::information(nullptr, tr("Error"),
                                 tr("Unable to save the file %1: %2.")
                                 .arg(QDir::toNativeSeparators(destFilePath), downloadFile->errorString()));
        emit docDownloadFailed();
        return;
    }

    // Setup our progress dialog.
    if (progressDialog)
    {
        progressDialog->setStatusValue(tr("Downloading %1").arg(documentName));
        progressDialog->setBarValue(0);
        progressDialog->setBarMinMax(0, 100);
        progressDialog->showCancel(true);
        progressDialog->showDialogWithoutPolling();
    }

    // Start the download process.
    if (!startHTTPDownload(CameraManualURL))
    {
        downloadStatus = tr("startHTTPDownload failure for %1.").arg(destFilePath);
        emit docDownloadFailed();
        return;
    }
}

/*!
 * @brief Slot handler for a cancelled document download completion.
 */
void fiUserDocDialog::doDocDownloadCanceled()
{
    downloadErrCode = -1;   // Denoting -1 as a cancel code.
    downloadStatus = tr("Download cancelled by the user.");
    doDocDownloadFailed();
}

/*!
 * @brief Slot handler for a failed document download completion.
 */
void fiUserDocDialog::doDocDownloadFailed()
{
    docDownloadCompleted(false);
}

/*!
 * @brief Slot handler for a successful document download completion.
 */
void fiUserDocDialog::doDocDownloadFinished()
{
    docDownloadCompleted(true);
}

/*!
 * @brief Called when the document download has completed whether in error, finished, or canceled.
 * @param completedOk   The status of the document download, True if fully downloaded or false if
 * the download was cancelled or in error.
 */
void fiUserDocDialog::docDownloadCompleted(bool completedOk)
{
    QString documentName;

    // Disconnect as all pathways lead here, finish, error, and a cancel.
    disconnect(netReply, &QNetworkReply::finished, this, &fiUserDocDialog::httpFinished);
    disconnect(netReply, &QNetworkReply::readyRead, this, &fiUserDocDialog::httpReadyRead);
    disconnect(netReply, &QNetworkReply::downloadProgress, this, &fiUserDocDialog::httpProgress);
    disconnect(netReply, &QNetworkReply::errorOccurred, this, &fiUserDocDialog::httpError);

    // Install the front entry to identify it in this module.
    documentName = downloadList.front();

    // We remove the top item from the list as it's been processed.
    downloadList.pop_front();

    // Hide the progress dialog if it is visible.
    if (progressDialog)
        progressDialog->hideDialog();

    // Cleanup
    if (downloadFile)
    {
        downloadFile->close();
        delete downloadFile;
        downloadFile = nullptr;
    }

    // A downloadErrCode of -1 means that the user cancelled the download.
    // Only report error in true case of network error condition.
    if (!completedOk && (downloadErrCode > 0))
    {
        QString msg_str = tr("A network error has occurred.\nUnable to download: %1\n\nHttp Error Code:%2\nCheck the log for additional information.").arg(documentName).arg(downloadErrCode);
        QLOG_WARN() << __PRETTY_FUNCTION__ << documentName << "Http error_no:" << downloadErrCode << "download status:" << downloadStatus << "completedOk:" << completedOk;
        QMessageBox::warning(this, "Unable to Download Document", msg_str);
    }

    // Is there another document to attempt to download?
    if (!downloadList.isEmpty())
    {
        // Launch another one.
        downloadADocument(downloadList.front());
    }
    else
    {
        // We are all done with the download list.
        setEnabledState(true);
    }
}

/*!
 * @brief The QNetworkReply object has terminated with an error.
 * @param errorCode The reason why the download failed.
 */
void fiUserDocDialog::httpError(QNetworkReply::NetworkError errorCode)
{
    // Pull some common errors to create some custom error messages but not all.
    switch (errorCode)
    {
    case QNetworkReply::ConnectionRefusedError:
        downloadStatus = tr("Connection refused to %1").arg(downloadServerUrl);
        break;
    case QNetworkReply::RemoteHostClosedError:
        downloadStatus = tr("%1 closed the connection.").arg(downloadServerUrl);
        break;
    case QNetworkReply::HostNotFoundError:
        downloadStatus = tr("Host %1 not found.").arg(downloadServerUrl);
        break;
    case QNetworkReply::TimeoutError:
        downloadStatus = tr("Connection %1 timed out.").arg(downloadServerUrl);
        break;
    case QNetworkReply::OperationCanceledError:
        downloadStatus = tr("%1 operation canceled or closed.").arg(downloadServerUrl);
        break;
    case QNetworkReply::SslHandshakeFailedError:
        downloadStatus = tr("SSL/TLS handshake failed to %1.").arg(downloadServerUrl);
        break;
    case QNetworkReply::TemporaryNetworkFailureError:
        downloadStatus = tr("Temporary network failure to %1.").arg(downloadServerUrl);
        break;
    case QNetworkReply::NetworkSessionFailedError:
        downloadStatus = tr("Connection to %1 was broken or failed to start.").arg(downloadServerUrl);
        break;
    case QNetworkReply::BackgroundRequestNotAllowedError:
        downloadStatus = tr("%1 background request not allowed due to platform policy.").arg(downloadServerUrl);
        break;
    case QNetworkReply::TooManyRedirectsError:
        downloadStatus = tr("Too many redirect errors to %1.").arg(downloadServerUrl);
        break;
    case QNetworkReply::InsecureRedirectError:
        downloadStatus = tr("%1 detected redirect from https to http.").arg(downloadServerUrl);
        break;
    default:
        downloadStatus = tr("Unable to download from %1.").arg(downloadServerUrl);
        break;
    }

    // Give a warning in the log.
    QLOG_WARN() << __PRETTY_FUNCTION__ << downloadStatus;

    // Persist the error code and our translated reason for the failure.
    downloadErrCode = errorCode;
    emit docDownloadFailed();
}

/*!
 * @brief The QNetworkReply object has finished the download operation.
 */
void fiUserDocDialog::httpFinished()
{
    // Sanity check.
    if (netReply == nullptr)
    {
        downloadStatus = tr("httpFinished has null netReply.");
        emit docDownloadFailed();
        return;
    }

    // Check for an error.
    if (netReply->error())
    {
        downloadStatus = tr("Download failed: %1.").arg(netReply->errorString());
        emit docDownloadFailed();
        return;
    }

    downloadStatus = tr("Download successful.");
    emit docDownloadFinished();
}

/*!
 * @brief Writes the download package read from over the network to the local file.
 */
void fiUserDocDialog::httpReadyRead()
{
    // Sanity check.
    if (netReply == nullptr)
    {
        downloadStatus = tr("httpReadyRead uncovered a null netReply.");
        emit docDownloadFailed();
        return;
    }

    if (downloadFile)
    {
        // Write the data out to the file.
        downloadFile->write(netReply->readAll());
    }
}

/*!
 * @brief Update our progress dialog, if it is visible. Then pass on the progress signal.
 * @param bytesReceived The number of bytes received.
 * @param bytesTotal    The total number of bytes expected.
 */
void fiUserDocDialog::httpProgress(qint64 bytesReceived, qint64 bytesTotal)
{
    // Update our progress dialog,.
    if (progressDialog)
    {
        progressDialog->setBarMax((int)bytesTotal);
        progressDialog->setBarValue((int)bytesReceived);
    }

    // Pass it on to anyone who might care.
    emit progress(bytesReceived, bytesTotal);
}

/*!
 * @brief Provides sweeping enable/disable support for the controls on the dialog.
 */
void fiUserDocDialog::setEnabledState(bool enabled)
{
    ui->chk_fasmotion->setEnabled(enabled);
    ui->chk_hs_camera->setEnabled(enabled);
    ui->chk_hs_quickstart->setEnabled(enabled);
    ui->chk_il_camera->setEnabled(enabled);
    ui->chk_il_quickstart->setEnabled(enabled);
    ui->chk_ts_camera->setEnabled(enabled);
    ui->chk_ts_quickstart->setEnabled(enabled);
    ui->pb_browse->setEnabled(enabled);
    ui->pb_download->setEnabled(enabled);
}

/*!
 * @brief Start the download request for the object specified in the requestedUrl parameter.
 *
 * @param requestedUrl  The URL for the requested object, e.g. the "latest" file, or update package.
 * @return True if the started the download without error or false if an error occurred.
 */
bool fiUserDocDialog::startHTTPDownload(const QUrl& requestedUrl)
{
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "requestedUrl:" << requestedUrl.toString();

    // Make sure we get rid of any reply objects that could be hanging around.
    // Note that this isn't necessarily an error - we don't automatically clean this
    // up incase the caller wants to retrieve any information after the fact.
    if (netReply != nullptr)
    {
        netReply->deleteLater();
        netReply = nullptr;
    }

    // Let's get a new connection for our download.
    nam = std::unique_ptr<QNetworkAccessManager>(new QNetworkAccessManager());
    if (!nam)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate QNetworkAccessManager!";
        return false;
    }

    // Start the request and get our reply object.
    netReply = nam->get(QNetworkRequest(requestedUrl));
    if (netReply == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to get a netReply object!";
        return false;
    }

    // Connect up the reply signals to our slots.
    connect(netReply, &QNetworkReply::finished, this, &fiUserDocDialog::httpFinished);
    connect(netReply, &QNetworkReply::readyRead, this, &fiUserDocDialog::httpReadyRead);
    connect(netReply, &QNetworkReply::downloadProgress, this, &fiUserDocDialog::httpProgress);
    connect(netReply, &QNetworkReply::errorOccurred, this, &fiUserDocDialog::httpError);

    // All done.
    return true;
}
