/*!
 * @file force_ip.cpp

 * @brief Implementation of the ForceIPDialog.
 *
 * It takes the new settings and applies them to the camera
 * selected by the user. Allowing changes to IP address, etc.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "force_ip.h"
#include "ui_force_ip.h"

#include <QDialogButtonBox>
#include <QMessageBox>
#include <QPushButton>

#include "logger.h"
#include "gige_msg.h"

// Use a namespac to avoid polluting the global space.
namespace FORCE_IP_DLG
{
    //--------------------------------------------------------------------
    // Contant, Macro and Type Definitions
    //--------------------------------------------------------------------

    const QLatin1Char   zero_fill('0');
}

/*!
 * @brief Constructor for the ForceIPDialog.
 */
ForceIPDialog::ForceIPDialog( QWidget *parent ) :
    QDialog( parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint ),
    ui( new Ui::ForceIPDialog )
{
    ui->setupUi(this);

    // Make the cancel button be the default choice.
    ui->buttonBox->button( QDialogButtonBox::Cancel )->setDefault( true );

    // Connect our signals/slots.
    connect( ui->rb_dhcp, SIGNAL( toggled(bool) ), this, SLOT( dhcp_toggled(bool)) );
}

/*!
 * @brief Destructor for the ForceIPDialog.
 */
ForceIPDialog::~ForceIPDialog()
{
    delete ui;
}

void ForceIPDialog::allow_mac_edit( bool do_allow )
{
    // Enable/disable the MAC control.
    ui->mac_addr->setEnabled( do_allow );
}

void ForceIPDialog::set_dhcp_allowed( bool is_allowed )
{
    ui->rb_dhcp->setEnabled( is_allowed );

    // If DHCP is not allowed (we must use static), then make sure
    // the static radio button is selected.
    if ( !is_allowed )
    {
        ui->rb_static->setChecked( true );
        dhcp_toggled( false );
    }
}

/*!
 * @brief Set the camera information pointer. Update the
 *        dialog's fields to reflect the information in
 *        the object.
 */
void ForceIPDialog::set_info( const FI_COMM::FI_CAMERA_INFO *cam_info )
{
    // Sanity check.
    if ( cam_info == nullptr )
    {
        ui->rb_dhcp->setEnabled( false );
        ui->rb_static->setChecked( true );
        dhcp_toggled( false );
        return;
    }

    // Initialize our display.
    set_ip( cam_info->ip_addr );
    set_netmask( cam_info->subnet_mask );
    set_gateway( cam_info->gateway );
    set_mac( cam_info->mac_addr );

    // Figure out if this is a STATIC or DHCP address.
    if (( cam_info->ip_current_cfg & GIGE_IFCONFIG_DHCP_BE ) == GIGE_IFCONFIG_DHCP_BE )
    {
        ui->rb_dhcp->setChecked( true );
        dhcp_toggled( true );
    }
    else
    {
        ui->rb_static->setChecked( true );
        dhcp_toggled( false );
    }
}

/*!
 * @brief Return TRUE if the user selected DHCP, FALSE if the
 *        user selected STATIC.
 */
bool ForceIPDialog::dhcp_is_enabled( void )
{
    return ui->rb_dhcp->isChecked();
}

/*!
 * @brief Set the IP address display of the dialog.
 *
 * @param ip_addr   The IP to initialize the dialog with.
 */
void ForceIPDialog::set_ip( uint32_t ip_addr )
{
    QHostAddress  addr;

    addr.setAddress( qFromBigEndian( ip_addr ));
    ui->ip_addr->set_addr( addr );
}

/*!
 * @brief Set the subnet mask display of the dialog.
 *
 * @param net_mask  The subnet to set the dialog to.
 */
void ForceIPDialog::set_netmask( uint32_t net_mask )
{
    QHostAddress  mask;

    mask.setAddress( qFromBigEndian( net_mask ));
    ui->net_mask->set_addr( mask );
}

/*!
 * @brief Set the gateway display of the dialog.
 *
 * @param gateway_addr  The gateway to set the dialog to.
 */
void ForceIPDialog::set_gateway( uint32_t gateway_addr )
{
    QHostAddress  gw_addr;

    gw_addr.setAddress( qFromBigEndian( gateway_addr ));
    ui->gateway->set_addr( gw_addr );
}

const uint8_t *ForceIPDialog::get_mac( void )
{
    memcpy( user_mac, ui->mac_addr->get_mac(), MacCtrl::QTUTL_MAC_SIZE );
    return user_mac;
}

/*!
 * @brief Set the MAC address display of the dialog.
 *
 * @param mac_addr  The MAC address to set the dialog to.
 */
void ForceIPDialog::set_mac( const uint8_t *mac_addr )
{
    ui->mac_addr->set_mac( mac_addr );
}

void ForceIPDialog::error_msg( const QString &err_str )
{
    // Throw up an error message box.
    QMessageBox::critical( this, tr( "Invalid Entry"), err_str, QMessageBox::Ok );
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user pressed the OK button.
 *
 * Sanity check all the edit fields, make sure the entries are valid.
 */
void ForceIPDialog::accept( void )
{
    // We only care about the edit controls if the user didn't select DHCP.
    if ( ui->rb_dhcp->isChecked() )
        goto pass_it_on;

    // Retrieve the IP address.
    user_ip = ui->ip_addr->get_addr();

    // Retrieve the network mask.
    user_net_mask = ui->net_mask->get_addr();

    // Retrieve the gateway address.
    user_gw = ui->gateway->get_addr();

pass_it_on:
    // Pass it on.
    QDialog::accept();
}

/*!
 * @brief User pressed the DHCP radio button.
 *
 * If checked, gray out the network setting controls, otherwise enable them.
 *
 * @param checked   True if the DHCP radio button is checked.
 */
void ForceIPDialog::dhcp_toggled( bool checked )
{
    ui->lblIPAddress->setEnabled( !checked );
    ui->ip_addr->setEnabled( !checked );
    ui->lblNetMask->setEnabled( !checked );
    ui->net_mask->setEnabled( !checked );
    ui->lblGateway->setEnabled( !checked );
    ui->gateway->setEnabled( !checked );
}
