/*!
 * @file format_dlg.cpp
 * @brief This object allows the user to control how to format an
 *        external drive on the camera.
 *
 * Copyright 2018 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "format_dlg.h"
#include "ui_format_dlg.h"

#include <QLayoutItem>
#include <QRegularExpression>
#include <QRegularExpressionValidator>

#include "fi_switchbutton.h"
#include "gige_msg.h"
#include "logger.h"

const QString FormatDlg::EXT_FMT_SERIES_3_TXT     = tr("Ext2: Linux only (not compatible with Windows PC or Mac)");
const QString FormatDlg::EXT_FMT_SERIES_HS_TXT    = tr("Ext4: Linux only, very fast for CAP files, slow for stacks");

FormatDlg::FormatDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    media_type(GIGE_MEDIA_SSD),
    internal_ssd(true),
    lrmode(false),
    ui(new Ui::FormatDlg),
    series3(true)
{
    // Setup our UI elements.
    ui->setupUi(this);

    // Setup our switch button.
    sw_erase = new fiSwitchButton(this, fiSwitchButton::Style::YESNO);
    sw_erase->setFixedSize(ui->pb_switch->size());
    ui->erase_layout->replaceWidget(ui->pb_switch, sw_erase);
    ui->pb_switch->setVisible(false);

    // Update the tab ordering.
    QWidget::setTabOrder(ui->edit_volume, sw_erase);
    QWidget::setTabOrder(sw_erase, ui->rb_fat32);
    QWidget::setTabOrder(ui->rb_fat32, ui->rb_ext4);
    QWidget::setTabOrder(ui->rb_ext4, ui->rb_ntfs);
    QWidget::setTabOrder(ui->rb_ntfs, ui->rb_exfat);
    QWidget::setTabOrder(ui->rb_exfat, ui->rb_filr);
    QWidget::setTabOrder(ui->rb_filr, ui->buttonBox);

    // Connect signals/slots.
    connect(ui->buttonBox, &QDialogButtonBox::accepted, this, &QDialog::accept);
    connect(ui->buttonBox, &QDialogButtonBox::rejected, this, &QDialog::reject);
    connect(ui->rb_fat32, &QRadioButton::released, this, &FormatDlg::format_FAT);
    connect(ui->rb_ntfs, &QRadioButton::released, this, &FormatDlg::format_NTFS);
    connect(ui->rb_ext4, &QRadioButton::released, this, &FormatDlg::format_EXT);
    connect(ui->rb_exfat, &QRadioButton::released, this, &FormatDlg::format_EXFAT);
    connect(ui->rb_filr, &QRadioButton::released, this, &FormatDlg::format_FILR);
    connect(sw_erase, &fiSwitchButton::valueChanged, this, &FormatDlg::do_sanitize);

    // Setup a regex to validate the volume label, so the user can't type in
    // an invalid value.
    QRegularExpression vol_regex("[0-9a-zA-Z][0-9a-zA-Z ]+");
    QRegularExpressionValidator *validator = new QRegularExpressionValidator(vol_regex, ui->edit_volume);
    ui->edit_volume->setValidator(validator);
}

FormatDlg::~FormatDlg()
{
    // Get rid of the UI object.
    delete ui;
}

void FormatDlg::accept()
{
    // Retrieve the volume label string from the edit control.
    vol_label = ui->edit_volume->text();

    // Pass it on.
    QDialog::accept();
}

int FormatDlg::exec()
{
    // Make sure we have a label.
    if (vol_label.isEmpty())
        vol_label = QString("FastecMedia");
    ui->edit_volume->setText(vol_label);

    // Perform series3 explicit formatting.
    if (series3)
    {
        ui->rb_fat32->setEnabled(((media_type != GIGE_MEDIA_SSD) || !lrmode) ? true : false);
        ui->rb_fat32->setText(tr("FAT32: For use with all systems and small devices"));
        ui->rb_exfat->setEnabled(false);
        ui->rb_ntfs->setEnabled(false);
        ui->rb_ext4->setEnabled(media_type != GIGE_MEDIA_SSD);
        ui->rb_ext4->setText(FormatDlg::EXT_FMT_SERIES_3_TXT);
        ui->rb_filr->setEnabled((media_type == GIGE_MEDIA_SSD) && lrmode);

        // User is not allowed to change the volume name.
        ui->edit_volume->setEnabled(false);
    }
    else
    {
        // Ext4 is allowed for every device, EXCEPT the internal SSD in LR mode.
        ui->rb_fat32->setEnabled(media_type != GIGE_MEDIA_SSD);
        ui->rb_fat32->setText(tr("FAT32: All systems, for small devices (slow)"));
        ui->rb_exfat->setEnabled(!internal_ssd);
        ui->rb_ntfs->setEnabled(!internal_ssd);
        ui->rb_ext4->setEnabled((media_type != GIGE_MEDIA_SSD) || !internal_ssd || !lrmode);
        ui->rb_ext4->setText(FormatDlg::EXT_FMT_SERIES_HS_TXT);
        ui->rb_filr->setEnabled((media_type == GIGE_MEDIA_SSD) && internal_ssd && lrmode);

        // Allow the user to edit the volume name.
        ui->edit_volume->setEnabled(true);
    }

    // Enable/disable the secure erase control
    ui->lbl_erase->setEnabled((media_type == GIGE_MEDIA_SSD) && sanitize);
    ui->erase_layout->setEnabled((media_type == GIGE_MEDIA_SSD) && sanitize);
    sw_erase->setEnabled((media_type == GIGE_MEDIA_SSD) && sanitize);
    sw_erase->setValue(false);

    // Now that we're done using the "sanitize" field as an input, switch
    // it around as an output. The default (as we see above) is false.
    sanitize = false;

    // Select the suggested format type.
    switch (format_type)
    {
    case GIGE_FORMAT_FAT32:
        ui->rb_fat32->setChecked(true);
        ui->edit_volume->setMaxLength(FAT_MAX_VOLUME_LEN);
        break;
    case GIGE_FORMAT_EXT2:
        ui->rb_ext4->setChecked(true);
        ui->edit_volume->setMaxLength(DEFAULT_MAX_VOLUME_LEN);
        break;
    case GIGE_FORMAT_FILR:
        ui->rb_filr->setChecked(true);
        ui->edit_volume->setMaxLength(DEFAULT_MAX_VOLUME_LEN);
        break;
    case GIGE_FORMAT_EXFAT:
        ui->rb_exfat->setChecked(true);
        ui->edit_volume->setMaxLength(FAT_MAX_VOLUME_LEN);
        break;
    case GIGE_FORMAT_NTFS:
        ui->rb_ntfs->setChecked(true);
        ui->edit_volume->setMaxLength(DEFAULT_MAX_VOLUME_LEN);
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected format type:" << format_type;
        ui->rb_exfat->setChecked(true);
        ui->edit_volume->setMaxLength(FAT_MAX_VOLUME_LEN);
        break;
    }

    // Get this bus rolling...
    return QDialog::exec();
}

void FormatDlg::do_sanitize(bool enable)
{
    sanitize = enable;
}

void FormatDlg::format_FAT()
{
    format_type = GIGE_FORMAT_FAT32;
    ui->edit_volume->setMaxLength(FAT_MAX_VOLUME_LEN);
}

void FormatDlg::format_EXFAT()
{
    format_type = GIGE_FORMAT_EXFAT;
    ui->edit_volume->setMaxLength(FAT_MAX_VOLUME_LEN);
}

void FormatDlg::format_EXT()
{
    format_type = GIGE_FORMAT_EXT2;
    ui->edit_volume->setMaxLength(DEFAULT_MAX_VOLUME_LEN);
}

void FormatDlg::format_FILR()
{
    format_type = GIGE_FORMAT_FILR;
    ui->edit_volume->setMaxLength(DEFAULT_MAX_VOLUME_LEN);
}

void FormatDlg::format_NTFS()
{
    format_type = GIGE_FORMAT_NTFS;
    ui->edit_volume->setMaxLength(DEFAULT_MAX_VOLUME_LEN);
}

void FormatDlg::isSeries3(bool series3)
{
    this->series3 = series3;
}
