/*!
 * @file fsp_socket.cpp
 * @brief This file contains the implementation of the FSPSocket
 *        class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QMutexLocker>

#include "fsp_socket.h"

#include "logger.h"
#include "gige_socket.h"
#include "fcp_socket.h"
#include "fi_streamsocket.h"

#include "fi_mem_pool.h"

namespace FSP_SOCKET
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! If you want to turn on verbose logging for this sequence numbers in receive logic turn this true.
//static const bool VERBOSE_SEQ_NBR_LOG = true;
static const bool VERBOSE_SEQ_NBR_LOG = false;

//! Because the image buffer data structure is very small (a sequence number
//! and a pointer to the actual memory pool resource) we can now allocate
//! a large number of these.  We have moved the control of this queue internal
//! and constant based.
const uint32_t IMAGE_BUFFER_INITIAL_QUEUESIZE   = 256;

//! We now maintain a backoff level for the communication stream based on a
//! percentage of the overall queuesize.
const float IMAGE_BUFFER_BACKOFF_LEVEL_PCT      = 0.25;

//! We now maintain a reallocation chunksize for the image buffer queue internally
//! and because the size of these entries is so small we will reallocate in large
//! blocks.
const uint32_t IMAGE_BUFFER_QUEUE_RESIZE_CHUNK  = 100;

//! Under emergency buffer conditions, raise the backoff threshold so that
//! we issue a PAUSE request faster.
const float EMERGENCY_BACKOFF_THRESHOLD = 0.75;

//! Under emergency buffer allocation conditions, let's try increasing the
//! interpacket delay.
const int   NORMAL_INTERPACKET_DELAY    = 100;

//! Once we exceed our MAX_EMERGENCY_ALLOCATION of buffers, we'll start dropping
//! packets. As a last ditch effort, try increasing the inter-packet delay.
const int   EMERGENCY_INTERPACKET_DELAY = 200;

//! The maximum number of packets we'll allocate if we're having a buffer emergency.
const int   MAX_EMERGENCY_ALLOCATION    = 1000;

//! A maximum number of packets we'll process in one pass.
const int   MAX_PACKET_COUNT            = 128;

//! A payload resource padding amount.  The amount of padding will align
//! on this boundary.  The resource padding is used for image data buffers
//! so that there is a little extra room for an overlay should one be set.
//! If the padding amount is not large enough to accomodate an overlay later
//! then the resource will be reallocated to the new capacity.  This padding
//! will just provide a faster throughput for those resources that don't need
//! that reallocation and where the padding amount increases capacity enough
//! to facilitate an overlay merge.
const uint32_t PAYLOAD_PAD_ALIGN        = 0x400;   // Align on 1K pad as default.

//! Define the states we need to transition through to receive an image.
enum
{
    FSP_TEST_PACKET_STATE,              //!< Processing test packets.
    FSP_LEADER_STATE,                   //!< Expecting to receive a leader packet.
    FSP_PAYLOAD_STATE,                  //!< Expecting to receive a payload packet.
    FSP_TRAILER_STATE,                  //!< Expecting to receive a trailer packet.
    FSP_IDLE,                           //!< State machine is idling.
    FSP_RECOVER
};

//! Backoff stages for increasing the inter-packet delay.
enum
{
    FSP_BACKOFF_STAGE_0,                //!< No backoff - whatever the user wants.
    FSP_BACKOFF_STAGE_1,                //!< Force a "normal" interpacket delay.
    FSP_BACKOFF_STAGE_2,                //!< Emergency interpacket delay required.
};

const int NET_SUCCESS           = 0;    //!< Local status code for processing functions: success.
const int NET_FAILURE           = -1;   //!< Local status code for processing functions: generic error.
const int NET_LOSTSYNC          = -2;   //!< Local status code for processing functions: packet out of sync.

}   // End of namespace FSP_SOCKET.

using namespace FSP_COMM;
using namespace FSP_SOCKET;
using namespace FI_MEMPOOL;

// #define DEBUG_FORMAT_ROUTINE
#ifdef DEBUG_FORMAT_ROUTINE
/*!
 * @brief Utility function to create a Qt::hex dump from a QByteArray.
 *
 * @param ba    The QByteArray we're interested in.
 * @return The character Qt::hex dump.
 */
QString byte_array_to_string(QByteArray& ba)
{
    int     i;
    QString resQt::hex;

    for (i = 0; i < ba.size(); ++i)
    {
        uint8_t value = (uint8_t)ba.at(i);
        resQt::hex.append(QString::fromUtf8("%1 ").arg(value, 2, 16, QLatin1Char('0')));
    }

    return resQt::hex;
}
#endif

//--------------------------------------------------------------------
// FSPSocket: class implementation
//--------------------------------------------------------------------

/*!
 * @brief Constructor for the FSPSocket class.
 * @param parent    The FCPSocket class that is our parent class.
 */
FSPSocket::FSPSocket(FCPSocket *parent) :
    QObject(parent),
    fcp_socket(parent),
    s3_socket(nullptr),
    fi_socket(nullptr),
    backoff_level(0),
    block_id(0),
    current_image_len(0),
    current_state(FSP_IDLE),
    discard_packets(true),
    emergency_allocation(0),
    emergency_delay(0),
    expected_image_len(0),
    initialized(false),
    low_water_mark(255),
    msg_port(FCP_COMM::DEFAULT_FCP_PORT + FCP_COMM::FSP_PORT_OFFSET),
    network_buffer(nullptr),
    num_buffers(0),
    overlay_padding(0),
    packet_id(0),
    packet_resync(false),
    resume_level(0),
    seq_number(0),
    stream_is_paused(false),
    user_delay(0),
    last_errcode(FI_COMM::FI_INIT_FAILURE)
{
    // Initialize our members.
    memset(file_leader, 0, sizeof(fsp_leader_file));
    memset(file_trailer, 0, sizeof(fsp_trailer));

    // Allocate our network buffer.
    network_buffer = new char[ JUMBO_PACKET_SIZE * 2 ];

    // NOTE:  There is no need to reset the image buffer pools because
    // they are already empty at construction time.
}

/*!
 * @brief The FSPSocket class destructor.
 */
FSPSocket::~FSPSocket()
{
    // Shut down the socket.
    close_socket();

    // Make sure the image buffer receive list is reset.
    ResetImageBufferRecvList();

    {
        QMutexLocker imageBufferLocker(&imageBufferFreeMutex);

        if ((int)imageBufferFreeList.size() != num_buffers)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "CHECK FREELIST BALANCE[num_buffers:" << num_buffers <<
                         "imageBufferFreeList.size(): " << imageBufferFreeList.size() <<
                         " activeImageBuffer=" << (activeImageBuffer ? "1]" : "0]");
        }

        // Free the resources used by this object.
        imageBufferFreeList.clear();

        // Is there an activeImageBuffer?  If so free its resource(s).
        if (activeImageBuffer)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Recycling the activeImageBuffer.";
            FI_MEMPOOL::FiMemoryResourcePtr imageResourcePtr = activeImageBuffer->ReleaseImageData();
            if (imageResourcePtr)
            {
                imageResourcePtr.reset();
            }
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESOURCE EMPTY:  imageResourcePtr was nullptr!";
            }
            activeImageBuffer = nullptr;
        }
    }

    // Release our network buffer.
    if (network_buffer != nullptr)
    {
        delete []network_buffer;
        network_buffer = nullptr;
    }
}

void FSPSocket::check_pause()
{
    // Make sure we aren't already streamed. Then check to see if there are image buffers and available memory in the pool.
    if (!stream_is_paused && (((int)imageBufferFreeList.size() <= backoff_level) || FiMemoryPool::IsPoolLow()))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "PAUSE. imageBufferFreeList:" << imageBufferFreeList.size() << "backoff:" << backoff_level
                     << "FiMemoryPool:" << FiMemoryPool::AvailableCount() << "IsPoolLow:" << FiMemoryPool::IsPoolLow();

        // Make sure we have a parent object.
        if (fcp_socket && !fcp_socket->is_offline())
        {
            // Send the pause command to the camera server. We do NOT want a response
            // from this command. We don't want any other thread reacting to this.
            fcp_socket->write_reg(fcp_socket->get_cam_address(), GIGE_REG_STREAM_CONTROL, GIGE_STREAM_PAUSE, false, false);
        }

        // Pause the stream.
        stream_is_paused = true;
    }
}

/*!
 * @brief Have our communications object shutdown.
 */
void FSPSocket::close_socket(void)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Do we have a socket object?
    if (s3_socket != nullptr)
    {
        // Tell the socket object to quit. This will cause it
        // to delete itself, so just clear our pointer.
        s3_socket->quit();
        s3_socket = nullptr;
    }
    if (fi_socket != nullptr)
    {
        // Tell the socket object to quit. This will cause it
        // to delete itself, so just clear our pointer.
        fi_socket->quit();
        fi_socket = nullptr;
    }
}

/*!
 * @brief If we run out of our buffer pool, do an emergency allocation.
 * @return True if successful, false otherwise.
 */
bool FSPSocket::alloc_emergency_buffer(void)
{
    bool allocatedImageBuffer = false;

    // Make sure we're below our threshold.
    if (emergency_allocation == MAX_EMERGENCY_ALLOCATION)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Maximum emergency buffer allocation reached!";
        return false;
    }

    // Allocate a emergency image buffer object for the pool.
    try
    {
        uint32_t count;

        // Allocate a block of image buffers.
        for (count = 0; count < IMAGE_BUFFER_QUEUE_RESIZE_CHUNK; ++count)
        {
            std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr = FiImageBuffer::CreateImageBuffer();
            if (imageBufferPtr)
            {
                // Lock access to the free list while removing image buffers.
                QMutexLocker imageBufferLocker(&imageBufferFreeMutex);
                imageBufferFreeList.push_back(std::move(imageBufferPtr));
                allocatedImageBuffer = true;
            }
        }
    }
    catch (std::exception& e)
    {
        QLOG_DEBUG() << "Allocation exception for IMAGE_BUFFER:" << e.what();
    }
    if (!allocatedImageBuffer)
    {
        return false;
    }

    // Note:  The system level image data pool may or may not have a free buffer
    // but that is not part of this image buffer queue and isolated out of this
    // socket's control.

    // Remember the new number of buffers we have allocate.
    num_buffers += IMAGE_BUFFER_QUEUE_RESIZE_CHUNK;

    // Update our backoff level - keeping it in line with the percentage to hold it at.
    backoff_level = (int)((float)num_buffers * IMAGE_BUFFER_BACKOFF_LEVEL_PCT);

    // Adjust our resume level.
    resume_level = backoff_level + ((num_buffers - backoff_level) / 2);

    // Increment our number of emergency buffer allocations.
    emergency_allocation += IMAGE_BUFFER_QUEUE_RESIZE_CHUNK;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "emergency_allocation:" << emergency_allocation << "num_buf:" << num_buffers << "backoff:" << backoff_level << "resume:" << resume_level;

    // Seems to have worked.
    return true;
}

/*!
 * @brief Return the last error string and code.
 * @param err_msg   A location for the error string.
 * @return The error code.
 */
int FSPSocket::get_last_err(QString& err_msg)
{
    err_msg = last_errstr;
    return last_errcode;
}

QHostAddress FSPSocket::get_local_address(void) const
{
    FCPSocket *socket_ptr = qobject_cast<FCPSocket *>(parent());

    // Sanity check
    if (socket_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No parent!";
        return QHostAddress();
    }

    // Return our local address.
    return socket_ptr->get_local_address();
}

/*!
 * @brief Return the final packet delay used in the transfer.
 */
int FSPSocket::get_packet_delay(void)
{
    int packet_delay;

    switch (emergency_delay)
    {
    case FSP_BACKOFF_STAGE_0:
        packet_delay = (int)user_delay;
        break;
    case FSP_BACKOFF_STAGE_1:
        packet_delay = (int)user_delay + NORMAL_INTERPACKET_DELAY;
        break;
    case FSP_BACKOFF_STAGE_2:
        packet_delay = (int)user_delay + EMERGENCY_INTERPACKET_DELAY;
        break;
    default:
        packet_delay = 0;
        break;
    }

    // Return the value.
    return packet_delay;
}

/*!
 * @brief Return the UDP port we're using for communications.
 */
int FSPSocket::get_port(void)
{
    // If we have a socket object, update our port value with
    // the latest version.
    if (s3_socket != nullptr)
        msg_port = (int)s3_socket->get_port();
    else if (fi_socket != nullptr)
        msg_port = fi_socket->get_port();

    return msg_port;
}

void FSPSocket::handle_emergency_delay(bool force_backoff)
{
    // Sanity check.
    if (fcp_socket == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fcp_socket is null.";
        return;
    }

    // We need to increase our delay.
    switch (emergency_delay)
    {
    case FSP_BACKOFF_STAGE_0:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Increasing delay (stage 1) to camera:" << (user_delay + NORMAL_INTERPACKET_DELAY);

        // Tell the camera to slow down.
        fcp_socket->write_reg(fcp_socket->get_cam_address(), GIGE_REG_SCPD0, (user_delay + NORMAL_INTERPACKET_DELAY), false, false);
        emergency_delay = FSP_BACKOFF_STAGE_1;
        break;
    case FSP_BACKOFF_STAGE_1:
        // Give the backoff a chance to work.
        if (force_backoff || (emergency_allocation > (MAX_EMERGENCY_ALLOCATION / 2)))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Increasing delay (stage 2) to camera:" << (user_delay + EMERGENCY_INTERPACKET_DELAY);

            // Tell the camera to slow down.
            fcp_socket->write_reg(fcp_socket->get_cam_address(), GIGE_REG_SCPD0, (user_delay + EMERGENCY_INTERPACKET_DELAY), false, false);
            emergency_delay = FSP_BACKOFF_STAGE_2;
        }
        break;
    case FSP_BACKOFF_STAGE_2:
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unknown emergency delay value:" << emergency_delay;
        break;
    }
}

/*!
 * @brief We received an image. The active image buffer and active image data should be
 * ready to combine into a full image received and put on the recieve list.  Emit an
 * image received signal if all goes well.
 *
 * @param image_len     The number of bytes in the image data.
 */
void FSPSocket::handle_image_received(int image_len)
{
    QString infoMsg;

    // The activeImageData and activeImageBuffer should be set up for this
    // process.  If they aren't there is something wrong.
    if (activeImageData && activeImageBuffer)
    {
        // Set up the image buffer sequence number.
        activeImageBuffer->SetSequenceNumber(seq_number);

        // Validate that the image_len is equal to the actual image data count.
        if (image_len != (int)activeImageData->GetLength())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The image_len " << image_len << " was not equal to active image resource length: " << activeImageData->GetLength();
        }

        infoMsg = QString("[%1]").arg(__PRETTY_FUNCTION__);

        // Time to attach the image data to the image buffer.
        activeImageData->LogMyId(&infoMsg);
        activeImageBuffer->AttachImageData(std::move(activeImageData));

        // Insert the image buffer on the end of the image buffer receive list.
        {
            QMutexLocker imageBufferLocker(&imageBufferRecvMutex);

            if (VERBOSE_SEQ_NBR_LOG)
            {
                if (activeImageBuffer->GetSequenceNumber() > 0)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_number:" << seq_number << "activeSeqNbr=" << activeImageBuffer->GetSequenceNumber() << "activeImageBuffer:" << Qt::showbase << Qt::hex << activeImageBuffer->GetMemDataPtr();
                }
            }

            imageBufferRecvList.push_back(std::move(activeImageBuffer));
        }
    }
    else
    {
        // Something is wrong!
        if (!activeImageBuffer)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "The active image buffer was a nullptr!";
        }
        if (!activeImageData)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "The active image data was a nullptr!";
        }
        return;
    }

    // Do we need to pause the stream?
    check_pause();

    // Update our free image buffer low water mark indicator.
    if ((int)imageBufferFreeList.size() < low_water_mark)
        low_water_mark = (int)imageBufferFreeList.size();

    // Let folks know there's one available.
    emit image_received();
}

/*!
 * @brief Pop an image off the front of the image buffer received list.
 *
 * !!IMPORTANT!!
 * This pointer is transferring image buffer and its image data ownership to
 * the caller.  These resources must be returned to this object by calling the
 * function image_release when the caller has completed processing the data
 * and the resources can be recycled.  If this does not happen a resource leak
 * will have occurred and the resources will not be reused in their respective
 * pools.
 *
 * @return The pointer to the next available image, or nullptr if there isn't a "next" image.
 */
std::unique_ptr<FI_MEMPOOL::FiImageBuffer> FSPSocket::image_take_first(void)
{
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr;

    // Sanity check.
    if (imageBufferRecvList.size() > 0)
    {
        QMutexLocker imageBufferLocker(&imageBufferRecvMutex);

        // Grab the pointer off the list.
        imageBufferPtr = std::move(imageBufferRecvList.front());
        imageBufferRecvList.pop_front();

        // Log a diagnostic out there for this resource.
        if (imageBufferPtr)
        {
            QString infoMsg = QStringLiteral("[FSPSocket::image_take_first()]");
            imageBufferPtr->LogMyDataId(&infoMsg);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESOURCE EMPTY:  imageBufferPtr was nullptr!";
        }
    }

    // Return the pointer and releases ownership.
    return imageBufferPtr;
}

/*!
 * @brief Recycles an image buffer resource and its image data resource back
 * to their respective pool system.
 *
 * @param imageBufferPtr    The image buffer resource to recycle.
*/
void FSPSocket::image_release(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr)
{
    FI_MEMPOOL::FiMemoryResourcePtr imageResourcePtr;

    // Sanity check.
    if (imageBufferPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESOURCE EMPTY:  imageBufferPtr was nullptr!";
        return;
    }

    // We are now assuming ownership.  Go ahead and release the image resource to us.
    imageResourcePtr = imageBufferPtr->ReleaseImageData();

    // Recycle the image data pointer.
    imageResourcePtr.reset();   // Just reset to recycle the resource.

    // Put the image buffer back to the free pool.
    {
        QMutexLocker imageBufferLocker(&imageBufferFreeMutex);
        imageBufferFreeList.push_back(std::move(imageBufferPtr));
    }

    // Check to see if we've paused the camera.
    if (stream_is_paused)
    {
        // Can we resume the stream?
        if ((int)imageBufferFreeList.size() > resume_level)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESUME. imageBufferFreeList:" << imageBufferFreeList.size() << "resume_level:" << resume_level;

            // Make sure we have a parent object.
            if (fcp_socket != nullptr)
                fcp_socket->write_reg(fcp_socket->get_cam_address(), GIGE_REG_STREAM_CONTROL, GIGE_STREAM_RESUME, false, false);

            // The stream is no longer paused.
            stream_is_paused = false;
        }
    }
}

/*!
 * @brief Sets a padding amount to use to support overlay usage.
 *
 * @param overlay_padding_to_set    The overlay padding to set.
 */
void FSPSocket::set_overlay_padding(uint32_t overlay_padding_to_set)
{
    // Make sure the padding ammount is in multiples of PAYLOAD_PAD_ALIGN.
    overlay_padding = ((overlay_padding_to_set + (PAYLOAD_PAD_ALIGN - 1)) / PAYLOAD_PAD_ALIGN) * PAYLOAD_PAD_ALIGN;
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "overlay_padding has been set to:" << overlay_padding;
}

/*!
 * @brief Initialze our buffer pools, allocating the memory and setting up the lists.
 *
 * @param num_buffers_to_set       The number of image buffers to pre-allocate.
 * @param backoff_level_to_set     When the number of available buffers drops to this level, we
 *                                 need to pause the camera's stream.
 *
 * @return Result of the allocation and initialization for the network buffer pool.
 */
void FSPSocket::init_buffers()
{
    // Make sure the image buffer receive list is reset.
    ResetImageBufferRecvList();

    // If there are more buffers in the free list than needed go ahead and
    // release them from the queue.
    if (imageBufferFreeList.size() > IMAGE_BUFFER_INITIAL_QUEUESIZE)
    {
        // Lock access to the free list while removing image buffers.
        QMutexLocker imageBufferLocker(&imageBufferFreeMutex);

        // traverse until equalize the image buffer queue.
        while (imageBufferFreeList.size() > IMAGE_BUFFER_INITIAL_QUEUESIZE)
        {
            imageBufferFreeList.pop_front();
        }
    }
    // If there are less buffers in the free list then add image buffers to equalize.
    else if (imageBufferFreeList.size() < IMAGE_BUFFER_INITIAL_QUEUESIZE)
    {
        // Lock access to the free list while adding image buffers.
        QMutexLocker imageBufferLocker(&imageBufferFreeMutex);

        // Free the image buffers.
        while (imageBufferFreeList.size() < IMAGE_BUFFER_INITIAL_QUEUESIZE)
        {
            std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr = FiImageBuffer::CreateImageBuffer();
            if (imageBufferPtr)
            {
                imageBufferFreeList.push_back(std::move(imageBufferPtr));
            }
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate image buffer.  Needed: "
                             << IMAGE_BUFFER_INITIAL_QUEUESIZE << " buffers - adjusting to: " << imageBufferFreeList.size() << ".";
                break;
            }
        }
    }

    // Remember the number of buffers we allocate.
    num_buffers = (int)imageBufferFreeList.size();

    // Remember when we should pause the camera's stream.
    backoff_level = (int)((float)num_buffers * IMAGE_BUFFER_BACKOFF_LEVEL_PCT);

    // Our resume level is halfway between the backoff level and the number of buffers.
    resume_level = backoff_level + ((num_buffers - backoff_level) >> 1);

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "imageBufferFreeList.size:" << num_buffers << "backoff:" << backoff_level << "resume:" << resume_level;
}

/*!
 * @brief Start the communications object working. This will take a different
 *        path depending on if we're attaching to the older series3 cameras
 *        or a newer camera that uses TCP to stream video.
 */
void FSPSocket::init_socket(bool series3)
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "series3:" << series3;

    // Series3 cameras are UDP, newer cameras are TCP...
    if (series3)
    {
        // Create our GiGE socket object which does all the UDP packet communications.
        s3_socket = new GiGESocket(nullptr, (uint32_t)msg_port);
        if (s3_socket == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Can't allocate a new GIGESocket!";
            return;
        }

        /*
         * Connect up the slots/signals so everyone can communicate.
         */

        // Make sure we grab any errors.
        connect(s3_socket, SIGNAL(error()), this, SLOT(fsp_error_event()));

        // Have the socket object notify us when it's up and running.
        connect(s3_socket, SIGNAL(started()), this, SLOT(fsp_socket_ready()));

        // Have the socket tell us when a data packet (datagram) is available.
        connect(s3_socket, SIGNAL(readyRead()), this, SLOT(fsp_read_pending_msg()), Qt::DirectConnection);

        // For shutdown, once the GiGE socket object is done have it invoke the delayed
        // cleanup method on exit.
        connect(s3_socket, SIGNAL(finished()), s3_socket, SLOT(deleteLater()));

        // Set the local address for our socket.
        s3_socket->setLocalAddress(fcp_socket->get_local_address());

        // Get the socket started. It'll either emit an error signal or, if everything
        // goes OK, it'll emit a "started" signal and we'll know we're initialized.
        s3_socket->start();
    }
    else
    {
        QHostAddress    cam_addr;

        // Create a streaming socket object.
        fi_socket = new FiStreamSocket(this);
        if (fi_socket == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Can't allocate a new FiStreamSocket!";
            return;
        }

        /*
         * Connect up the slots/signals so everyone can communicate.
         */

        // Make sure we grab any errors.
        connect(fi_socket, SIGNAL(error()), this, SLOT(fsp_error_event()));

        // Have the socket object notify us when it's up and running.
        connect(fi_socket, SIGNAL(ready()), this, SLOT(fsp_socket_ready()));

        // Have the socket tell us when a data packet (datagram) is available.
        connect(fi_socket, SIGNAL(data_ready()), this, SLOT(fsp_read_stream()), Qt::DirectConnection);

        // For shutdown, once the socket object is done have it invoke the delayed
        // cleanup method on exit.
        connect(fi_socket, SIGNAL(finished()), fi_socket, SLOT(deleteLater()));
        connect(fi_socket, SIGNAL(finished()), this, SLOT(fsp_socket_done()));

        // Connect to the camera the the "well known" FSP port. Yes, I know it says
        // "UDP"...but a port is a port.
        cam_addr.setAddress(qFromBigEndian(fcp_socket->get_cam_address()));
        fi_socket->connect_cam(cam_addr, GIGE_FSP_UDP_PORT);
    }
}

/*!
 * @brief Return true if the streamer state is IDLE, false otherwise.
 */
bool FSPSocket::is_idle()
{
    return (discard_packets || (current_state == FSP_IDLE));
}

/*!
 * @brief Reset our lists, moving any dirty buffers back to the available list.
 */
void FSPSocket::ResetImageBufferRecvList(void)
{
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESOURCES: imageBufferRecvList.size(): "
                 << imageBufferRecvList.size() << "imageBufferFreeList.size(): "
                 << imageBufferFreeList.size();

    // Traverse the list of images received and restore them to the free list.
    while (imageBufferRecvList.size() > 0)
    {
        // Take the image buffer off the receive list.
        {
            QMutexLocker imageBufferLocker(&imageBufferRecvMutex);
            imageBufferPtr = std::move(imageBufferRecvList.front());
            imageBufferRecvList.pop_front();
        }

        // Restore the image data to the image data pool.
        if (imageBufferPtr)
        {
            QString myInfoStr = QString("[%1-RECYCLING]").arg(__PRETTY_FUNCTION__);
            imageBufferPtr->LogMyDataId(&myInfoStr);
            FI_MEMPOOL::FiMemoryResourcePtr imageResourcePtr = imageBufferPtr->ReleaseImageData();
            if (imageResourcePtr)
            {
                imageResourcePtr.reset();   // Just reset the pointer and the resource is recycled.
            }
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[RESOURCE EMPTY] memDataPtr was nullptr!";
            }
            {
                QMutexLocker imageBufferLocker(&imageBufferFreeMutex);
                imageBufferFreeList.push_back(std::move(imageBufferPtr));
            }
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[RESOURCE EMPTY] imageBufferPtr was nullptr!";
        }
    }
}

/*!
 * @brief This method processes a message received from a series3 camera.
 *
 * It performs simple sanity checks to make sure this is an "expected"
 * message. Then it passes the message on to packet specific parsing methods.
 *
 * @param msg_ptr   The FSP message we've received.
 */
void FSPSocket::process_received_msg(uint8_t *buf_array, uint32_t buf_len)
{
    int     payload_size;           // Number of bytes in the current payload packet.

    // Process the received packet based on the current state.
    switch (current_state)
    {
    case FSP_LEADER_STATE:
        // Process a leader packet..we presume.
        expected_image_len = recv_leader(buf_array, buf_len);
        if (expected_image_len > 0)
        {
            // Move on to receiving the payload packets.
            current_image_len = 0;
            current_state = FSP_PAYLOAD_STATE;

            // Not re-syncing any more.
            packet_resync = false;
        }
        break;
    case FSP_PAYLOAD_STATE:
        // Process the payload packet.
        payload_size = recv_payload(buf_array, buf_len);
        if (payload_size < 0)
        {
            // Got an error. Reset our state.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error processing payload packets...resetting. payload_size:" << payload_size;
            current_state = FSP_LEADER_STATE;
            packet_resync = true;
            recycleActiveImageData();
            break;
        }

        // Increment the number of bytes we've accumulated.
        current_image_len += payload_size;

        // Are we done?
        if (current_image_len == expected_image_len)
        {
            // Move on to receiving the trailer packet and wrapping this image up.
            current_state = FSP_TRAILER_STATE;
        }
        else if (current_image_len > expected_image_len)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Too much data...resetting.";
            current_state = FSP_LEADER_STATE;
            packet_resync = true;
            recycleActiveImageData();
        }
        break;
    case FSP_TRAILER_STATE:
        // Our next state is to look for a leader packet.
        current_state = FSP_LEADER_STATE;

        // We should be getting a trailer packet now.
        if (recv_trailer(buf_array, buf_len) != NET_SUCCESS)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error processing trailer packet...resetting.";
            packet_resync = true;
            recycleActiveImageData();
            break;
        }

        // Handle the received image.
        handle_image_received(expected_image_len);
        break;
    case FSP_TEST_PACKET_STATE:
        // Carry on.
        emit test_packet_received();
        break;
    default:
    case FSP_IDLE:
        break;
    }
}

/*!
 * @brief Handle the processing required to receive and sanity check a FSP leader packet.
 * @return The expected payload size when everything works out, NET_FAILURE otherwise.
 */
int FSPSocket::recv_leader(const uint8_t *buf_array, uint32_t buf_len)
{
    bool        sentBackoffDelay = false;   //!< Did we have to tell the socket to delay?
//    uint8_t     *by_pkt_id;
    uint32_t    payload_type;
    uint32_t    payload_size;
    uint32_t    payload_size_padding = overlay_padding;
    const fsp_header *header_ptr;
    const fsp_leader *leader_ptr;
    const fsp_leader_file  *file_leader;

    union
    {
        uint32_t l_value;
        uint8_t  b_value[4];
    } by_pkt_id;

    Q_UNUSED(buf_len)

    // Overlay our header structure onto the packet.
    header_ptr = (const fsp_header *)buf_array;

    // Make sure we have a LEADER packet. If not, we'll skip until we get one.
    if (header_ptr->pkt_fmt != FSP_PKT_FMT_LEADER)
    {
        // Only complain on entry.
        if (!packet_resync)
        {
            // Not in re-sync so complain.
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Malformed packet: wrong pkt_fmt =" << header_ptr->pkt_fmt;

            // We SHOULD be in resync. Make it so.
            packet_resync = true;
        }

        // Return that we've got nothing.
        return 0;
    }

    // Get our block ID.
    block_id = qFromBigEndian(header_ptr->block_id);

    // Get the packet ID, so we know where to start. We need it in the
    // correct byte order.
    by_pkt_id.l_value = 0;
    by_pkt_id.b_value[0] = header_ptr->pkt_id[2];
    by_pkt_id.b_value[1] = header_ptr->pkt_id[1];
    by_pkt_id.b_value[2] = header_ptr->pkt_id[0];
    packet_id = by_pkt_id.l_value;

    // Now that we know we have a leader packet, overlay our leader structure.
    leader_ptr = (const fsp_leader *)header_ptr;

    // Make sure (if we're a newer camera) that the magic ID matches.
    if (fi_socket != nullptr)
    {
        uint16_t magic_id = qFromBigEndian(leader_ptr->magic_id);
        if (magic_id != FSP_FASTEC_ID)
        {
            // Only complain on entry.
            if (!packet_resync)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Wrong MAGIC ID =" << Qt::showbase << Qt::hex << magic_id;

                // We SHOULD be in resync. Make it so.
                packet_resync = true;
            }

            // Let the caller know something is amiss.
            return 0;
        }
    }

    // Make sure this is the right payload type.
    payload_type = qFromBigEndian(leader_ptr->payload_type);
    if (payload_type != FSP_PAYLOAD_FILE)
    {
        // Only complain on entry.
        if (!packet_resync)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Wrong PAYLOAD type =" << payload_type;

            // We SHOULD be in resync. Make it so.
            packet_resync = true;
        }

        // Let the caller know something is amiss.
        return 0;
    }

    // Now we know we have the right payload type, overlay our file structure.
    file_leader = (const fsp_leader_file *)leader_ptr;

    // Get the expected length for the payload. We only need to worry about
    // the low word - we'll never exceed that. Just make sure the value read
    // is sane. We do have our limits...
    payload_size = qFromBigEndian(file_leader->ps_low);
    if (payload_size > GiGESocket::MAX_IMAGE_SIZE)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Wrong PAYLOAD size =" << Qt::showbase << Qt::hex << payload_size;

        // We SHOULD be in resync. Make it so.
        packet_resync = true;

        // Let the caller know something is amiss.
        return 0;
    }

    // Get the sequence number, aka the file name.
    bool    ok;
    QString file_name = QString::fromUtf8(file_leader->file_name);
    seq_number = file_name.toUInt(&ok, 16);

    // Make sure we're still good.
    check_pause();

    // Sanity Check:  There should NOT be an activeImageData it should be a nullptr
    // at this point.  Make sure that is the case so we don't resource leak the
    // image data pool resource.
    if (activeImageData)
    {
        // NOTE:  This is ok - there are conditions where orphaned activeImageData blocks
        // can exist.  Just recycle it at this point and take the new one.
        //QLOG_WARN() << __PRETTY_FUNCTION__ << "The activeImageData was not a nullptr.  Attempting to recycle."
        //            << Qt::showbase << Qt::hex << "block_id:" << block_id << "packet_id:" << packet_id << "seq_number:" << seq_number;
        recycleActiveImageData();
    }

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "payload_size_padding:" << payload_size_padding << " payload:" << payload_size;

    // Pull out an image data resource from the system pool but don't allow
    // emergency allocations for the first try.  This is because we will need
    // to slow down the socket traffic if emergency allocation is needed.
    activeImageData = FiMemoryPool::GetMemoryResource((payload_size + payload_size_padding), false);

    // If unable to get it that way try again with an emergency allocation.
    if (activeImageData == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "FiMemoryPool free buffers are empty but we didn't allow emergency allocation on first try so this is ok.  We are now telling the fcp_socket to slow down." << Qt::showbase << Qt::hex << "block_id:" << block_id << "packet_id:" << packet_id << "seq_number:" << seq_number;

        // Make sure we have a parent object and we haven't already tried this.
        if (fcp_socket != nullptr)
        {
            // Back off our data transfer rate.
            handle_emergency_delay();
            sentBackoffDelay = true;
        }

        // Now that the socket has been told to back off to support an emergency
        // allocation go ahead and try but this time with emergency allocation on.
        activeImageData = FiMemoryPool::GetMemoryResource((payload_size + payload_size_padding), true);
        if (activeImageData == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to emergency allocate image data object!";
            packet_resync = true;
            return 0;
        }
    }

    // Now make sure that the active image buffer is in play.
    if (!activeImageBuffer)
    {
        if (imageBufferFreeList.size() == 0)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "image buffer pool is empty!  Telling fcp_socket to slow down." << Qt::showbase << Qt::hex << "block_id:" << block_id << "packet_id:" << packet_id << "seq_number:" << seq_number;

            if (!sentBackoffDelay)
            {
                handle_emergency_delay();
                sentBackoffDelay = true;
            }
            if (!alloc_emergency_buffer())
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to emergency allocate image buffer object!";
            }
        }

        // Test again and pull the buffer.
        if (imageBufferFreeList.size() > 0)
        {
            QMutexLocker imageBufferLocker(&imageBufferFreeMutex);
            activeImageBuffer = std::move(imageBufferFreeList.front());
            imageBufferFreeList.pop_front();
        }
    }
    else
    {
        // NOTE_JLC:  If this should happen and the memDataPtr had been the
        // activeImageData object above that unique_ptr would have been recycled
        // and the unique_ptr would now be a nullptr so that there should not be
        // any side-effects or problems here.

        // This really shouldn't happen - but if it does - does it have an
        // attached image data?  If so it needs to be recyled.
        FI_MEMPOOL::FiMemoryResourcePtr imageResourcePtr = activeImageBuffer->ReleaseImageData();
        if (imageResourcePtr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "There was an activeImageBuffer and it had an attached imageResourcePtr that is being recycled!";
            imageResourcePtr.reset();   // Just reset the pointer to recycle the resource.
        }
    }

    // Make sure there is an active image buffer to work with here.
    if (activeImageBuffer == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to retrieve free image buffer object setting for a packet resync!";
        packet_resync = true;
        return 0;
    }

    // Was a backup delay sent in this function?
    if (sentBackoffDelay)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "handle_emergency_delay() was sent.";
    }

    //NOTE: - marking this place to note that the image buffer pool managed locally here may
    // be out of resources - but that should be handled in a different location.  This logic
    // is strictly to handle the image data array pool.

    // All is well. Return the expected payload size.
    return (int)payload_size;
}

/*!
 * @brief Handle the processing required to receive and sanity check a FSP payload packet.
 *
 * @param data_ptr      Pointer to the location to put the data.
 *
 * @return The received byte count, or NET_FAILURE on error.
 */
int FSPSocket::recv_payload(uint8_t *buf_array, uint32_t buf_len)
{
    int         payload_bytes;
    int         bytesTransferred = 0;
    uint32_t    current_pkt_id;
    uint8_t     *by_pkt_id;
    uint16_t    current_blk_id;
    const fsp_header    *header_ptr;
    const fsp_payload  *payload_ptr;

    // Sanity check.
    if (activeImageData == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "activeImageData is nullptr!";
        return NET_LOSTSYNC;
    }

    // Initialize our variables.
    by_pkt_id = (uint8_t *)&current_pkt_id;

    // Overlay our header structure onto the packet.
    header_ptr = (fsp_header *)buf_array;

    // Make sure we have a PAYLOAD packet. If we got something else, we are out of sync
    // so we'll want to just skip and start again at the next LEADER.
    if (header_ptr->pkt_fmt != FSP_PKT_FMT_PAYLOAD)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Out of sync. Got fmt=" << header_ptr->pkt_fmt << "but expecting PAYLOAD";
        return NET_LOSTSYNC;
    }

    // Get our block ID and verify that we are in the same block.
    current_blk_id = qFromBigEndian(header_ptr->block_id);
    if (current_blk_id != block_id)
    {
        QString err_str = QString::fromUtf8("Block ID mismatch (%1 != %2)").arg(current_blk_id).arg(block_id);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        return NET_LOSTSYNC;
    }

    // Get the packet ID.
    current_pkt_id = 0;
    by_pkt_id[0] = header_ptr->pkt_id[2];
    by_pkt_id[1] = header_ptr->pkt_id[1];
    by_pkt_id[2] = header_ptr->pkt_id[0];

    // Make sure this is the right packet.
    if ((packet_id + 1) != current_pkt_id)
    {
        QString err_str = QString::fromUtf8("Packet ID mismatch (0x%1 != 0x%2 + 1) for block: 0x%3").arg(current_pkt_id, 0, 16).arg(packet_id, 0, 16).arg(current_blk_id, 0, 16);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "IP:" << cam_addr.toString() << "port:" << get_port();

#ifdef DEBUG_FORMAT_ROUTINE
        QByteArray my_array = QByteArray((const char *)buf_array, 32);
        QLOG_DEBUG() << byte_array_to_string(my_array);
#endif

        return NET_LOSTSYNC;
    }

    // Update our packet ID for later.
    packet_id = current_pkt_id;

    // Now that we know we have a payload packet, overlay our structure.
    payload_ptr = (const fsp_payload *)header_ptr;

    // Adjust our count.
    payload_bytes = (int)(buf_len - sizeof(fsp_header));
    if (payload_bytes <= 0)
    {
        QString err_str = QString::fromUtf8("Payload byte count is bad (%1). Buf_len: %2").arg(payload_bytes).arg(buf_len);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        return NET_LOSTSYNC;
    }

    // Copy the data over.
    if (activeImageData)
    {
        bytesTransferred = (int)activeImageData->AppendBytes((uint32_t)payload_bytes, (const uint8_t *)&payload_ptr->data);

        // What if bytesTransferred is zero here?  TODO_JLC - what is correct return code falling out from this condition?
        if (bytesTransferred == 0)
        {
            QString err_str = QString::fromUtf8("Unable to append payload bytes(%1) into active image data.").arg(payload_bytes);
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        }
    }
    else
    {
        // TODO_JLC - this is an error condition that wasn't tested before with active_data_ptr.
    }

    // Return the number of data bytes transferred.
    return bytesTransferred;
}

/*!
 * @brief Handle the processing required to receive and sanity check a FSP trailer packet.
 * @return NET_SUCCESS when everything works out, NET_FAILURE otherwise.
 */
int FSPSocket::recv_trailer(uint8_t *buf_array, uint32_t /* buf_len */)
{
    uint32_t    current_pkt_id;
    uint8_t     *by_pkt_id;
    uint16_t    current_blk_id;
    uint32_t    payload_type;
    const fsp_header   *header_ptr;
    const fsp_trailer  *trailer_ptr;

    // Initialize our variables.
    by_pkt_id = (uint8_t *)&current_pkt_id;

    // Overlay our header structure onto the packet.
    header_ptr = (fsp_header *)buf_array;

    // Make sure we have a TRAILER packet.
    if (header_ptr->pkt_fmt != FSP_PKT_FMT_TRAILER)
        return NET_FAILURE;

    // Get our block ID and verify that we're in the same block.
    current_blk_id = qFromBigEndian(header_ptr->block_id);
    if (current_blk_id != block_id)
    {
        QString err_str = QString::fromUtf8("Block ID mismatch (%1 != %2)").arg(current_blk_id).arg(block_id);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        return NET_FAILURE;
    }

    // Get the packet ID.
    current_pkt_id = 0;
    by_pkt_id[0] = header_ptr->pkt_id[2];
    by_pkt_id[1] = header_ptr->pkt_id[1];
    by_pkt_id[2] = header_ptr->pkt_id[0];

    // Make sure this is the right packet.
    if ((packet_id + 1) != current_pkt_id)
    {
        QString err_str = QString::fromUtf8("Packet ID mismatch (%1 != %2 + 1)").arg(current_pkt_id).arg(packet_id);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        return NET_FAILURE;
    }

    // Update our packet ID for later.
    packet_id = current_pkt_id;

    // Now that we know we have a trailer packet, overlay our structure.
    trailer_ptr = (const fsp_trailer *)header_ptr;
    payload_type = qFromBigEndian(trailer_ptr->payload_type);

    // Sanity check the payload type.
    if (payload_type != FSP_PAYLOAD_FILE)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Wrong PAYLOAD type =" << payload_type;
        return NET_FAILURE;
    }

    // All is well.
    return NET_SUCCESS;
}

/*!
 * @brief Recycle active image data back to image pool.
 *
 * Used in response to various conditions which require the resource to return
 * to the system image pool.
 */
void FSPSocket::recycleActiveImageData()
{
    if (activeImageData)
    {
        activeImageData.reset();    // Just reset the pointer to recycle.
    }
}

/*!
 * @brief Flush out any data residing in the stream socket.
 * @return Any residual data received but not processed by the streaming socket.
 */
QByteArray FSPSocket::flush_stream()
{
    QMutexLocker map_locker(&fsp_mutex);    // Lock access to the stream.

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    if (fi_socket == nullptr)
        return QByteArray();
    return fi_socket->readAll();
}

/*!
 * @brief Reset our statistics information.
 */
void FSPSocket::reset_stats()
{
    // low_water_mark keeps track of the low-point in image buffer queue
    // so when we reset the statistics we set it to the full size.
    low_water_mark = num_buffers;

    // If for some reason the num_buffers is not equivalent to the number of
    // free buffers there could be a lost buffer.  But it's ok because of the
    // way that the release works for the smart pointer it is always recovered.
    if (num_buffers != (int)imageBufferFreeList.size())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "CHECK FREELIST BALANCE[num_buffers:" << num_buffers <<
                     "imageBufferFreeList.size(): " << imageBufferFreeList.size() <<
                     " activeImageBuffer=" << (activeImageBuffer ? "1]" : "0]");
    }
}

/*!
 * @brief Force the stream to resume.
 */
void FSPSocket::resume_stream()
{
    // Check to see if we've paused the camera.
    if (stream_is_paused)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "RESUME level:" << resume_level << "imageBufferFreeList:" << imageBufferFreeList.size();

        // Can we resume the stream?
        if ((int)imageBufferFreeList.size() > resume_level)
        {
            // Make sure we have a parent object.
            if (fcp_socket != nullptr)
                fcp_socket->write_reg(fcp_socket->get_cam_address(), GIGE_REG_STREAM_CONTROL, GIGE_STREAM_RESUME, false, false);

            // The stream is no longer paused.
            stream_is_paused = false;
        }
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Stream is not paused.";
    }
}

/*!
 * @brief Start receiving FSP streaming video.
 */
void FSPSocket::start_stream(bool do_test_packet, uint32_t packet_delay)
{
    QMutexLocker map_locker(&fsp_mutex);    // Lock access to the stream.

    // QLOG_TRACE() << __PRETTY_FUNCTION__ << "test_packet:" << do_test_packet << "current_state:" << current_state;

    // Make sure we're initialized.
    if (!initialized)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Not initialized!";
        return;
    }

    // Remember the packet delay the user wants.
    user_delay = packet_delay;

    // Reset our interpacket delay stage.
    emergency_delay = FSP_BACKOFF_STAGE_0;
    current_image_len = 0;
    memset(file_leader, 0, sizeof(fsp_leader_file));
    memset(file_trailer, 0, sizeof(fsp_trailer));

    // Put ourselves into the FSP processing state machine.
    if (do_test_packet)
        current_state = FSP_TEST_PACKET_STATE;
    else if (s3_socket != nullptr)
        current_state = FSP_LEADER_STATE;
    else if (current_state != FSP_LEADER_STATE)
        current_state = FSP_RECOVER;

    // Don't discard packets anymore.
    discard_packets = false;
}

/*!
 * @brief Don't expect any streaming video from the camera.
 */
void FSPSocket::stop_stream()
{
    QMutexLocker map_locker(&fsp_mutex);    // Lock access to the stream.

    // No streaming video.
    if (s3_socket == nullptr)
    {
        // For new cameras, we want to discard the packets.
        discard_packets = true;
    }
    else
    {
        // For a series3 camera, just set the state to IDLE.
        current_state = FSP_IDLE;
    }

    // Drop packets which might be trickling in.
    packet_resync = true;
}

void FSPSocket::quit()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Close down the streamer.
    close_socket();

    // Reset the receive list.
    ResetImageBufferRecvList();
}

//--------------------------------------------------------------------
// Slots
//--------------------------------------------------------------------

/*!
 * @brief Invoked when an error has happened. Grab the error information
 *        into our local variables, last_errcode and last_errstr.
 */
void FSPSocket::fsp_error_event()
{
    // If we have a socket object, grab the error information.
    if (s3_socket != nullptr)
        last_errcode = s3_socket->get_last_err(last_errstr);
    else if (fi_socket != nullptr)
        last_errcode = fi_socket->get_last_err(last_errstr);

    QString err_str = QString::fromUtf8("Error [%1] = %2").arg(last_errcode).arg(last_errstr);
    QLOG_INFO() << __PRETTY_FUNCTION__ << err_str;

    // Close the socket.
    close_socket();

    // Let folks know.
    emit error();
}

/*!
 * @brief Read and do preliminary processing on any pending datagrams
 *        from the gige_socket object. Note that datagrams (UDP)
 *        implies a series3 camera stream.
 */
void FSPSocket::fsp_read_pending_msg()
{
    bool        allow_slowdown;
    int         packet_count(0);

    // Sanity check.
    if (s3_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No s3_socket!";
        return;
    }
    if (!initialized)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Not initialized!";
        return;
    }

    // Only allow slowdowns if we're not already in a packet resync state.
    allow_slowdown = !packet_resync;

    // Read in any and all pending datagrams.
    while (s3_socket->hasPendingDatagrams())
    {
        // Read our waiting message.
        qint64 read_len = s3_socket->readDatagram(network_buffer, JUMBO_PACKET_SIZE, &cam_addr);
        if (read_len <= 0)
        {
            QString err_msg;

            last_errstr = s3_socket->errorString();
            last_errcode = FI_COMM::FI_SOCKET_ERR;

            err_msg = QString::fromUtf8("%1: readDatagram returned err(%2), errno=%3(%4)").arg(__PRETTY_FUNCTION__, last_errstr).arg(errno).arg(strerror(errno));
            QLOG_ERROR() << err_msg;
            return;
        }

        // Process the message.
        process_received_msg((uint8_t *)network_buffer, (uint32_t)read_len);

        // If we got a packet resync, let's slow things down.
        if (packet_resync && allow_slowdown)
        {
            handle_emergency_delay(true);
            allow_slowdown = false;
        }

        // Increment the number of packets we've processed. If we've hit our maximum,
        // then take a break and come back later.
        ++packet_count;
        if (packet_count >= MAX_PACKET_COUNT)
            break;
    }
}

/*!
 * @brief Read the leader packet from the TCP stream. This packet gives
 *        us ordering information as well as the length of the following
 *        payload.
 *
 * @return The number of bytes we've consumed.
 */
qint64 FSPSocket::fsp_read_leader()
{
    qint64 read_amount(0);

    // Sanity check.
    if (current_image_len >= (int)sizeof(fsp_leader_file))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "current_image_len is too big:" << current_image_len;
        current_image_len = 0;
    }

    // Read the amount of data we need/want. Note that we may not get what
    // we asked for.
    read_amount = fi_socket->read(&file_leader[current_image_len], (int64_t)((int)sizeof(fsp_leader_file) - current_image_len));
    if (read_amount < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Error reading leader!";
        return -1;
    }
    else if (read_amount == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No data?!";
        return 0;
    }

    // Increment the number of bytes we've accumulated.
    current_image_len += read_amount;

    // Are we done?
    if (current_image_len == sizeof(fsp_leader_file))
    {
        // Process the leader package.
        expected_image_len = recv_leader(file_leader, sizeof(fsp_leader_file));
        if (expected_image_len > 0)
        {
            // Move on to receiving the payload packets.
            current_state = FSP_PAYLOAD_STATE;
            current_image_len = 0;
        }
        else
        {
            // We've got an error.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "recv_leader() failed!! Entering recovery...";
            read_amount = 0;
            current_state = FSP_RECOVER;
        }
    }

    // Return the amount of data we've read.
    return read_amount;
}

/*!
 * @brief Read the payload from the TCP stream. The assumption is
 *        that we've received a leader packet telling us how many
 *        bytes are in this payload.
 *
 * @return The number of bytes we've consumed.
 */
qint64 FSPSocket::fsp_read_payload()
{
    int payloadSize                 = (expected_image_len - current_image_len);
    uint32_t bytesAvailable         = 0;    // How many bytes are available in the image data object?
    uint8_t *imageDataInternalPtr   = nullptr;

    // Sanity check.
    if (activeImageData == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "activeImageData is a nullptr!";
        return -1;
    }

    // Set up to use the internal pointer of the activeImageData object.
    imageDataInternalPtr = activeImageData->AppendBytesExternally((uint32_t)payloadSize, bytesAvailable);

    // If there isn't space for the possible payload size pull what we can and see
    // how it differs after the socket read.
    if (imageDataInternalPtr == nullptr)
    {
        uint32_t askForBytesAvailable = bytesAvailable;
        imageDataInternalPtr = activeImageData->AppendBytesExternally(askForBytesAvailable, bytesAvailable);

        // If can't get one now there is no going forward.
        if (imageDataInternalPtr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to append to activeImageData.  Asked for " << askForBytesAvailable << "bytes and has " << bytesAvailable << "available.";
            return -1;
        }
        payloadSize = (int)askForBytesAvailable;
    }

    // NOTE - the read_amount is a long_long type - but I am assuming
    // that the payloadSize since it is an integer will keep that value from
    // exceeding the size an int data type can hold.

    // Read the amount of data we need/want. Note that we may not get what
    // we asked for.
    qint64 read_amount = fi_socket->read(imageDataInternalPtr, payloadSize);
    if (read_amount < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "read() returned an error!";
        goto rp_bail;
    }
    else if (read_amount == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No data to read!";
        goto rp_bail;
    }

    // Update our data pointer.
    activeImageData->CompleteExternalAppend((uint32_t)read_amount);

    // Increment the number of bytes we've accumulated.
    current_image_len += read_amount;

    // Are we done?
    if (current_image_len == expected_image_len)
    {
        // Move on to receiving the next packet.
        current_state = FSP_TRAILER_STATE;
        current_image_len = 0;
    }

rp_bail:
    // Return the amount of data we've read.
    return read_amount;
}

/*!
 * @brief Read and do preliminary processing on any pending data from the fi_socket object.
 */
void FSPSocket::fsp_read_stream()
{
    bool done(false);
    QMutexLocker map_locker(&fsp_mutex);    // Lock access to the stream.

    // Sanity check.
    if (fi_socket == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No fi_socket!";
        return;
    }
    if (!initialized)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Not initialized!";
        return;
    }

    // Loop while we have data available.
    while (!done)
    {
        // How many bytes are available.
        qint64 data_len = fi_socket->data_available();
        if (data_len == 0)
        {
            // Nothing more to do here. Pop out of the loop.
            done = true;
        }
        else
        {
            qint64 read_len;

            // Process the received data based on the current state.
            switch (current_state)
            {
            case FSP_LEADER_STATE:
                read_len = fsp_read_leader();
                if (read_len < 0)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fsp_read_leader() reported an error!";
                    done = true;
                    current_image_len = 0;
                    current_state = FSP_IDLE;
                }
                break;
            case FSP_PAYLOAD_STATE:
                read_len = fsp_read_payload();
                if (read_len < 0)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fsp_read_payload() reported an error!";
                    done = true;
                    current_image_len = 0;
                    current_state = FSP_IDLE;
                }
                break;
            case FSP_TRAILER_STATE:
                read_len = fsp_read_trailer();
                if (read_len < 0)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fsp_read_trailer() reported an error!";
                    done = true;
                    current_image_len = 0;
                    current_state = FSP_IDLE;
                }
                break;
            case FSP_TEST_PACKET_STATE:
            {
                QByteArray junk;

                // Don't care about the actual data...just that we got a packet.
                // Discard the data as expeditiously as possible.
                junk = fi_socket->readAll();
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "FSP_TEST_PACKET_STATE, len:" << junk.size();

                // Indicate the test packet was received.
                emit test_packet_received();
            }
            break;
            case FSP_RECOVER:
                fsp_recover_scan();
                break;
            default:
            case FSP_IDLE:
                discard_packets = true;
                current_state = FSP_RECOVER;
                break;
            }
        }
    }
}

/*!
 * @brief Read the trailer packet from the TCP stream. This packet completes
 *        a data transfer from the camera.
 *
 * @return The number of bytes we've consumed.
 */
qint64 FSPSocket::fsp_read_trailer()
{
    qint64 read_amount(0);

    // Sanity check.
    if (current_image_len >= (int)sizeof(fsp_trailer))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "current_image_len is too big:" << current_image_len;
        current_image_len = 0;
    }

    // Read the amount of data we need/want. Note that we may not get what
    // we asked for.
    read_amount = fi_socket->read(&file_trailer[current_image_len], (int)sizeof(fsp_trailer) - current_image_len);
    if (read_amount < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Error reading trailer!";
        return -1;
    }
    else if (read_amount == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No data?!";
        return 0;
    }

    // Increment the number of bytes we've accumulated.
    current_image_len += read_amount;

    // Are we done?
    if (current_image_len == sizeof(fsp_trailer))
    {
        // Our next state is to look for a leader packet.
        current_state = FSP_LEADER_STATE;
        current_image_len = 0;

        // Process the trailer packet.
        if (recv_trailer(file_trailer, sizeof(fsp_trailer)) == NET_SUCCESS)
        {
            // Do we want to ignore this packet?
            if (!discard_packets)
            {
                // Handle the received image.
                handle_image_received(expected_image_len);
            }

            // Our next state is to look for a leader packet.
            current_state = FSP_LEADER_STATE;
            current_image_len = 0;
        }
        else
        {
            // We've got an error.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "recv_trailer() failed!! Entering recovery...";
            read_amount = 0;
            current_state = FSP_RECOVER;
            current_image_len = 0;
        }
    }

    // Return the amount of data we've read.
    return read_amount;
}

void FSPSocket::fsp_recover_scan()
{
    QByteArray rd_buffer;

    // Check to see if we have data in the holding buffer.
    if (current_image_len > 0)
    {
        // Copy the data over to our buffer.
        rd_buffer.append((const char *)file_leader, current_image_len);
        current_image_len = 0;
    }

    // Append up to another leader packets worth of data.
    rd_buffer.append(fi_socket->read(sizeof(fsp_leader_file)));

    // Loop while we have data to work with.
    while (rd_buffer.size() >= (int)sizeof(fsp_leader_file))
    {
        // Check to see if we've got a leader package now.
        expected_image_len = recv_leader((const uint8_t *)rd_buffer.constData(), sizeof(fsp_leader_file));
        if (expected_image_len > 0)
        {
            // Make sure there is an active image data object in play.
            // Following similar logic to the old active_data_ptr logic.
            if (activeImageData == nullptr)
            {
                // TODO_JLC - revisit this spot.  What if an emergency allocation is
                // needed for the image data resource.  Is this a case where we should
                // just let that happen without slowing down network traffic?  This will
                // require a revisit.

                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "activeImageData was not set!";
                activeImageData = FiMemoryPool::GetMemoryResource((uint32_t)expected_image_len, true);
                if (activeImageData == nullptr)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to get free image data object for processing.";
                    break;
                }
            }

            // Move on to receiving the payload packets.
            current_state = FSP_PAYLOAD_STATE;

            // Remove the leader packet from our buffer.
            rd_buffer.remove(0, sizeof(fsp_leader_file));

            // How much data is left?
            current_image_len = rd_buffer.size();

            // Move over any extra data we've got laying around.
            if ((current_image_len > 0) && (activeImageData != nullptr))
            {
                // Copy over the buffered data.
                uint32_t bytesAppended = activeImageData->AppendBytes((uint32_t)current_image_len, (const uint8_t *)rd_buffer.constData());

                // There could not have been enough room in the image data object
                // for the image.  This is an error condition.
                if (bytesAppended != (uint32_t)current_image_len)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to append full image length:  " << current_image_len << " appended " << bytesAppended << " bytes.";
                    // TODO_JLC - this check wasn't here before - what should the handler
                    // do in this condition?
                }
            }

            // Not in resync anymore.
            packet_resync = false;

            // And carry on.
            return;
        }

        // Shift our data and try again.
        rd_buffer.remove(0, 1);
    }

    // If we're here, we haven't found the packet yet. Buffer any
    // data for the next time we come around.
    current_image_len = rd_buffer.size();

    // Move over any extra data we've got laying around.
    if (current_image_len > 0)
        memcpy(file_leader, rd_buffer.constData(), (size_t)current_image_len);
}

void FSPSocket::fsp_socket_done()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Our streamer is done.
    fi_socket = nullptr;
    emit finished();
}


/*!
 * @brief Invoked when the socket has finished initializing.
 */
void FSPSocket::fsp_socket_ready()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // We are initialized and ready for message processing.
    initialized = true;
    last_errcode = FI_COMM::FI_SUCCESS;

    QHostAddress cam_addr;
    cam_addr.setAddress(qFromBigEndian(fcp_socket->get_cam_address()));

    // Let folks know we're ready.
    emit fsp_is_ready(fcp_socket->get_cam_address());
}
