/*!
 * @file fsp_socket.h
 * @brief This header file declares the FSPSocket class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef FSP_SOCKET_H
#define FSP_SOCKET_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <memory>
#include <deque>

#include <QObject>
#include <QByteArray>
#include <QList>
#include <QMutex>
#include <QtNetwork>
#include <QNetworkInterface>

#include "gige_msg.h"
#include "fi_mem_pool.h"
#include "fi_image_buffer.h"

namespace FSP_COMM
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The largest packet we'll ever expect to receive.
const int       JUMBO_PACKET_SIZE       = 9000;

//! The maximum "chunk" of video we can "manipulate" in one go.
const uint32_t  BLOB_CHUNK_SIZE         = (1024 * 1024 * 5);

}   // End of FSP_COMM namespace.

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class GiGESocket;
class FCPSocket;
class FiStreamSocket;

//--------------------------------------------------------------------
// Class definition for FSPSocket
//--------------------------------------------------------------------

class FSPSocket : public QObject
{
    Q_OBJECT

public:
    FSPSocket(FCPSocket *parent = nullptr);
    ~FSPSocket();

    void    init_socket(bool series3);
    bool    is_initialized()          { return initialized; }
    bool    is_idle();
    bool    is_paused()               { return stream_is_paused; }

    QByteArray flush_stream();
    void    reset_stats();
    void    resume_stream();
    void    start_stream(bool do_test_packet, uint32_t packet_delay);
    void    stop_stream();
    void    quit();

    int     get_backoff()             { return backoff_level; }
    int     get_last_err(QString& err_msg);
    QHostAddress get_local_address() const;
    int     get_low_watermark()       { return low_water_mark; }
    int     get_num_buffers()         { return num_buffers; }
    int     get_packet_delay();

    int     get_port();
    void    set_port(int new_port)        { msg_port = new_port; }

    void    set_overlay_padding(uint32_t overlay_padding_to_set);

    int     image_count()             { return (int)imageBufferRecvList.size(); }
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> image_take_first();
    void    image_release(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr);

    void    init_buffers();

signals:
    void    image_received();
    void    test_packet_received();
    void    fsp_is_ready(uint32_t ip_addr);
    void    error();
    void    finished();

private slots:
    void    fsp_error_event();
    void    fsp_read_pending_msg();
    void    fsp_read_stream();
    void    fsp_socket_ready();
    void    fsp_socket_done();

private:
    void    check_pause();
    void    close_socket();
    bool    alloc_emergency_buffer();

    void    handle_emergency_delay(bool force_backoff = false);
    void    handle_image_received(int image_len);
    void    process_received_msg(uint8_t *buf_array, uint32_t buf_len);

    qint64  fsp_read_leader();
    qint64  fsp_read_payload();
    qint64  fsp_read_trailer();
    void    fsp_recover_scan();

    int     recv_leader(const uint8_t *buf_array, uint32_t buf_len);
    int     recv_payload(uint8_t *buf_array, uint32_t buf_len);
    int     recv_trailer(uint8_t *buf_array, uint32_t buf_len);

    void    recycleActiveImageData();

    void    ResetImageBufferRecvList();

private:
    FCPSocket       *fcp_socket;        //!< The associated FCP socket.
    GiGESocket      *s3_socket;         //!< The actual socket class for the GiGE FSP messages.
    FiStreamSocket  *fi_socket;         //!< Streaming socket class for newer cameras.
    QHostAddress    cam_addr;           //!< Address of the camera we're talking with.
    QMutex          fsp_mutex;          //!< Mutex to guard access to the streaming socket.

    int             backoff_level;      //!< The point at which we have to pause the camera's stream.
    uint16_t        block_id;           //!< Our block ID during the image reception process.
    int             current_image_len;  //!< The accumulated bytes from the payload packets.
    int             current_state;      //!< The processing state.
    bool            discard_packets;    //!< True if we're in IDLE and want to toss anything we receive.
    int             emergency_allocation;   //!< Track the number of emergency buffers we've allocated.
    int             emergency_delay;    //!< Which delay stage we're in.
    int             expected_image_len; //!< The image size as reported by the leader packet.
    uint8_t         file_leader[sizeof(fsp_leader_file)];   // Buffer to hold our leader.
    uint8_t         file_trailer[sizeof(fsp_trailer)];      // Buffer to hold our trailer.
    bool            initialized;        //!< Has the socket finished it's initialization sequence (successful or not)?

    //! The active image data for building an image from network payload data.
    FI_MEMPOOL::FiMemoryResourcePtr activeImageData;

    //! The active image buffer for building an image from the network.
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> activeImageBuffer;

    //! A private list of available FiImageBuffer records to manage network buffer streams.
    std::deque<std::unique_ptr<FI_MEMPOOL::FiImageBuffer>> imageBufferFreeList;

    //! Protects the imageBufferFreeList when modifying data.
    QMutex imageBufferFreeMutex;

    //! A private list of received FiImageBuffer records waiting to be processed.
    std::deque<std::unique_ptr<FI_MEMPOOL::FiImageBuffer>> imageBufferRecvList;

    //! Protects the imageBufferRecvList when modifying data.
    QMutex imageBufferRecvMutex;


    int             low_water_mark;     //!< Statistics: the lowest number of buffers in the imageBufferFreeList.
    int             msg_port;           //!< The port we're supposed to use.
    char            *network_buffer;    //!< Intermediate buffer for network packets.
    int             num_buffers;        //!< The number of allocated buffers.
    uint32_t        overlay_padding;    //!< A padding amount to add to image buffers for overlay support.
    uint32_t        packet_id;          //!< Our packet ID during the image reception process.
    bool            packet_resync;      //!< If we lose sync with the streaming packet, this goes true until we are "re-synced".
    int             resume_level;       //!< The number of available buffers we need before we'll resume the stream.
    uint32_t        seq_number;         //!< The sequence number of our image, if applicable.
    bool            stream_is_paused;   //!< True if we have the stream paused, false otherwise.
    uint32_t        user_delay;         //!< The user specified interpacket delay.

    QString     last_errstr;    //!< last message which occurred
    int         last_errcode;   //!< last error code which occurred
};

#endif // s3_socket_H
