/*!
 * @file gige_socket.cpp
 * @brief This file contains the implementation of the GiGESocket
 *        class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "gige_socket.h"

#include <QObject>
#include <QProcess>

#include "fi_comm.h"
#include "logger.h"

// This has to go after all the Qt Headers because the platform specific defines are defined in those header files.
// These are needed for the setsockopt() calls
#ifdef Q_OS_WIN
#include "winsock2.h"
#endif // Q_WS_WIN

#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/statvfs.h>
#include <errno.h>
#endif // Q_WS_X11

//--------------------------------------------------------------------
// Constant, Macros and Type Definitions
//--------------------------------------------------------------------

const uint32_t  GiGESocket::MAX_IMAGE_SIZE      = 0x1410000;     // We are using 20Mb as the maximum data to be sent. Add something in for the overhead.

namespace GIGE_SOCKET
{
#if defined( Q_OS_MAC )
//! The minimum allowed UDP receive space.
const int   MIN_UDP_RECVSPACE   = 0x200000;
#endif

//! Try this many times to find a port to bind to.
const int   NUM_RETRIES         = 10;

//! We need to skip 3 port addresses, because there's three channels:
//! control, streaming and messages.
const int   PORT_SKIP_COUNT     = 3;

}   // namespace GIGE_SOCKET

using namespace GIGE_SOCKET;

/*!
 * @brief GiGESocket constructor.
 */
GiGESocket::GiGESocket(QObject *parent, uint32_t port_id) :
    QUdpSocket(parent),
    msg_port(port_id),
    last_errcode(0)
{
    // It's OK to copy arguments to internal variables, just
    // don't use NEW!
}

GiGESocket::~GiGESocket()
{
}

/*!
 * @brief This method initialize the GiGE socket object.
 *
 * @note    NEVER allocate heap objects (using new) in the constructor
 *          of the QObject class as this allocation is then performed
 *          on the main thread and not on the new QThread instance,
 *          meaning that the newly created object is then owned by the
 *          main thread and not the QThread instance. This will make
 *          your code fail to work.
 */
void GiGESocket::start()
{
    int     retries;

    // Bind the port for the FCP messages. This could be a multiple step process. If we
    // try to open our default port number and we can't bind to it then we assume it's being
    // used by another instance. We will look through a number of ports before we give up and
    // assume there is a networking issue.
    for (retries = 0; retries < NUM_RETRIES; ++retries)
    {
        // Try to bind.
        if (bind((uint16_t)msg_port))
            break;

        // Skip over other ports and try again.
        msg_port += PORT_SKIP_COUNT;
    }
    if (retries == NUM_RETRIES)
    {
        // Save off a code, the error string for logging.
        last_errcode = FI_COMM::FI_INIT_FAILURE;
        last_errstr = errorString();

        // Inidicate there was an error and bail.
        emit error();
        return;
    }

    // Sanity check.
    if (localPort() != msg_port)
    {
        QString msg_str;
        msg_str = QString::fromUtf8("%1: localPort [%2] != msg_port [%3]").arg(__PRETTY_FUNCTION__).arg(localPort()).arg(msg_port);
        QLOG_WARN() << msg_str;
    }

#ifdef Q_OS_WIN
    // The following is platform dependent code (windows) to increase the size of the receive buffers to
    // avoid missing messages
    int socket_bufsize = -1;  // Unlimited apparently, although I can't find documentation to support this.
    if (setsockopt((SOCKET)socketDescriptor(), SOL_SOCKET, SO_RCVBUF, (char *)&socket_bufsize, sizeof(socket_bufsize)) == -1)
    {
        // We'll note the error - but it's really more of a warning.
        last_errstr = QString(tr("Unable to set the stream socket receive buffer size which increased the potential for lost messages"));
        last_errcode = FI_COMM::FI_INIT_FAILURE;
        emit error();
    }
#endif // Q_OS_WIN
#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
#if defined( Q_OS_MAC )
    // We need to make sure the UDP receive space is large enough to handle our images. Fire up a shell
    // so we can run some commands.
    QProcess    mac_sh;
    QStringList mac_args;

    mac_args << "-c" << "sysctl net.inet.udp.recvspace";
    mac_sh.start("/bin/sh", mac_args);
    if (!mac_sh.waitForFinished())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Couldn't start the shell.";
    }
    else
    {
        int     recv_space;
        QString     sh_str;
        QStringList sh_result;

        // Read the result of our command.
        sh_str = mac_sh.readAll();

        // Parse it to get the value.
        sh_result = sh_str.split(':');

        // Convert the second argument to a number.
        recv_space = sh_result.at(1).toInt();

        // If the value is too low, log it.
        if (recv_space < MIN_UDP_RECVSPACE)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << QString("Your 'net.inet.udp.recvspace' setting (%1) is too small! It should be at least %2.").arg(recv_space).arg(MIN_UDP_RECVSPACE);
        }
    }
#else
    // The following is platform dependent code (Linux) to increase the size of the receive buffers to
    // avoid missing parts of a message
    uint32_t socket_bufsize = GiGESocket::MAX_IMAGE_SIZE;  // should be enough for a full frame
    if (setsockopt(socketDescriptor(), SOL_SOCKET, SO_RCVBUF, (char *)&socket_bufsize, sizeof(socket_bufsize)) == -1)
    {
        // We'll note the error - but it's really more of a warning.
        last_errstr = QString(tr("Unable to set the stream socket receive buffer size which increased the potential for lost messages"));
        last_errcode = FI_COMM::FI_INIT_FAILURE;
        emit error();
    }
#endif /* Q_OS_MAC */
#endif // Q_OS_LINUX or Q_OS_MAC

    // Let folks know we're ready.
    emit started();
}

/*!
 * @brief Time to close down the socket.
 */
void GiGESocket::quit()
{
    // Pass it on up.
    close();
}

/*!
 * @brief We are being requested to send a message.
 */
int GiGESocket::send_msg(const QByteArray& msg_out, const QHostAddress& dest_addr)
{
    qint64  bytes_sent;

    // Clear our error information.
    last_errstr.clear();
    last_errcode = FI_COMM::FI_SUCCESS;

    // Send the message.
    bytes_sent = writeDatagram(msg_out, dest_addr, GIGE_SERVER_UDP_PORT);
    if (bytes_sent < 0)
    {
        QString err_msg;

        last_errstr = errorString();
        last_errcode = FI_COMM::FI_SOCKET_ERR;

        err_msg = QString::fromUtf8("err=%1(%2), errno=%3(%4)").arg(QAbstractSocket::error()).arg(last_errstr).arg(errno).arg(strerror(errno));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << dest_addr.toString() << err_msg;

        // Let folks know
        emit error();
    }

    // Return our status.
    return last_errcode;
}

/*!
 * @brief Retrieve error messages from the class.
 *
 * @param err_msg   Where to copy the error string.
 *
 * @return The error code associated with the error string.
 */
int GiGESocket::get_last_err(QString& err_msg)
{
    // Copy over the error string.
    err_msg = last_errstr;

    // Return the error code.
    return last_errcode;
}
