/*!
 * @file histo_dock.cpp
 * @brief This file contains the implementation of the HistoDock
 *        class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "histo_dock.h"
#include "ui_histo_dock.h"

#include <QApplication>
#include <QByteArray>
#include <QClipboard>
#include <QInputDialog>
#include <QContextMenuEvent>
#include <QMdiSubWindow>
#include <QMenu>
#include <QTimer>

#include "camera_mdi.h"
#include "camera_info.h"
#include "camera_stream.h"
#include "camera_video.h"
#include "fi_histo_proc.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_mdi.h"
#include "playback_meta_info.h"
#include "playback_raw_mdi.h"

namespace HISTO_DOCK
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Our update frequency to gather histogram information.
const int   UPDATE_INTERVAL         = 250;

//! Our basic title.
const QString   WINDOW_TITLE_STR(HistoDock::tr("Histogram"));

}   // End of namespace HISTO_DOCK.

using namespace HISTO_DOCK;

/*!
 * @brief The HistoDock constructor.
 */
HistoDock::HistoDock(QWidget *parent, bool useTimer) :
    QDockWidget(parent),
    ui(new Ui::HistoDock)
{
    // Setup our UI object.
    ui->setupUi(this);

    // Initialize our member variables.
    active_camera = nullptr;
    active_mdi = nullptr;
    histo_size = 0;
    histo_local = false;
    histo_proc = nullptr;
    update_paused = false;
    roi_override = false;
    single_in_progress = false;
    use_timer = useTimer;
    update_timer = nullptr;

    // Single-shot a delayed update of the clip registers. These may be cleared
    // out by the embedded GUI on a transition from Live to Review, but we need
    // to wait for that to happen.
    QTimer::singleShot(FASMOTION_SETTINGS::STARTUP_DELAY, this, &HistoDock::finish_setup);
}

/*!
 * @brief The HistoDock destructor.
 */
HistoDock::~HistoDock()
{
    // Clear out our UI object.
    delete ui;

    // Do we have a timer running?
    if (update_timer != nullptr)
    {
        update_timer->stop();
        update_timer->deleteLater();
        update_timer = nullptr;
    }

    // Do we have a histogram processing object?
    if (histo_proc != nullptr)
    {
        delete histo_proc;
        histo_proc = nullptr;
    }
}

void HistoDock::disconnect_camera()
{
    // Make sure we disconnect from any active camera.
    new_active_camera(nullptr);
}

fiHistogram *HistoDock::get_histogram()
{
    return ui->histo_plot;
}

void HistoDock::enable_timer(bool do_timer)
{
    // Remember the enable flag.
    use_timer = do_timer;
    if (do_timer)
    {
        // Create our timer.
        update_timer = new QTimer(this);
        update_timer->setInterval(UPDATE_INTERVAL);

        // Connect it into the slot.
        connect(update_timer, &QTimer::timeout, this, &HistoDock::time_to_update);

        // Start it rolling.
        if (!update_paused && isVisible())
            update_timer->start(UPDATE_INTERVAL);
    }
    else
    {
        // Sanity check.
        if (update_timer != nullptr)
        {
            // Stop the timer.
            update_timer->stop();

            // Disconnect from the slot.
            disconnect(update_timer, &QTimer::timeout, this, &HistoDock::time_to_update);

            // Clean out the timer.
            update_timer->deleteLater();
            update_timer = nullptr;
        }
    }
}

void HistoDock::override_roi(QRect roi)
{
    if (roi.isEmpty())
    {
        roi_override = false;
    }
    else
    {
        roi_override = true;
        roi_rect = roi;
    }
}


/*!
 * @brief Stop the update timer.
 */
void HistoDock::pause()
{
    update_paused = true;
    if (update_timer != nullptr)
        update_timer->stop();
}

/*!
 * @brief Start reading histogram data from the camera.
 *
 * Because the histogram data can be larger than the number of bytes allowed
 * for a FCP ReadMem command, this process spans a number of methods.
 *
 * This routine, histo_read_start(), initiates the read process and pulls down
 * the first GIGE_HISTOGRAM_READ_SIZE bytes.
 *
 * @ref HistoDock::histo_read_update() continues to pull down "chunks" until all
 * the histogram data has been read from the camera.
 */
void HistoDock::histo_read_start()
{
    uint32_t    cam_ip_addr;
    uint32_t    read_len;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Make sure we have an active camera.
    if (active_camera == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active camera!";
        return;
    }

    // Get the IP address for our camera.
    cam_ip_addr = active_camera->get_ip_addr();

    // Read up the size of the histogram.
    if (ui->histo_plot->is_raw())
        histo_size = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTOGRAM_RAW_SIZE);
    else
        histo_size = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SIZE);

    // Reset our buffer.
    histo_data.clear();
    histo_data.reserve((int)histo_size / 4);

    // How much can we read?
    if (histo_size > GIGE_HISTOGRAM_READ_SIZE)
        read_len = GIGE_HISTOGRAM_READ_SIZE;
    else
        read_len = histo_size;

    // Start reading the histogram data.
    if (ui->histo_plot->is_raw())
        comm_ptr->start_read_mem(cam_ip_addr, GIGE_REG_HISTOGRAM_RAW, read_len);
    else
        comm_ptr->start_read_mem(cam_ip_addr, GIGE_REG_HISTO_RGB, read_len);
}

/*!
 * @brief Continue reading the histogram data from the camera.
 */
void HistoDock::histo_read_update()
{
    int         remaining_bytes;
    uint32_t    cam_ip_addr;
    QByteArray  *read_data;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Make sure we have an active camera.
    if (active_camera == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active camera!";
        return;
    }

    // Get the IP address for our camera.
    cam_ip_addr = active_camera->get_ip_addr();

    // Read the data we got from the camera.
    if (ui->histo_plot->is_raw())
        read_data = comm_ptr->get_reg_buffer(cam_ip_addr, GIGE_REG_HISTOGRAM_RAW);
    else
        read_data = comm_ptr->get_reg_buffer(cam_ip_addr, GIGE_REG_HISTO_RGB);
    if (read_data != nullptr)
    {
        // Make sure we have at least one 32-bit value.
        if (read_data->size() > 4)
        {
            // Mark the old size of the data.
            int old_size = histo_data.size();

            // Update the data vector size. Remember that the read_data
            // size is the number of bytes. Convert that to the number
            // of 32-bit words.
            histo_data.resize(histo_data.size() + (read_data->size() / 4));

            // Now get a pointer to the data and move it into old_size position.
            uint32_t *data_ptr = histo_data.data();
            data_ptr += old_size;

            // Copy the data over.
            memcpy(data_ptr, read_data->constData(), (size_t)read_data->size());
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << active_camera->get_ip_addr_pretty() << "Not enough read_data for histogram, assuming failed read";

            // Reset our single flag.
            single_in_progress = false;

            single_shot();  // Attempt another round.

            // Release the buffer.
            delete read_data;

            // Leave.
            return;
        }

        // Release the buffer.
        delete read_data;
    }

    // Are we done? Don't forget to convert the histo_data array size
    // (which is the number of elements) into a byte count.
    remaining_bytes = (int)histo_size - (histo_data.size() * 4);
    if (remaining_bytes > 0)
    {
        uint32_t    read_len;

        // Ask the camera for the next "chunk".
        if (remaining_bytes > GIGE_HISTOGRAM_READ_SIZE)
            read_len = GIGE_HISTOGRAM_READ_SIZE;
        else
            read_len = (uint32_t)remaining_bytes;

        // Start reading the histogram data.
        if (ui->histo_plot->is_raw())
            comm_ptr->start_read_mem(cam_ip_addr, GIGE_REG_HISTOGRAM_RAW, read_len);
        else
            comm_ptr->start_read_mem(cam_ip_addr, GIGE_REG_HISTO_RGB, read_len);
    }
    else
    {
        uint32_t    sum_value;
        uint32_t    sum_sq_l;
        uint32_t    sum_sq_h;
        uint64_t    sum_sq_value;

        // Read up the "blue"/"mono" statistics information we got from the camera.
        sum_value = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_BLU);
        sum_sq_h = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_H);
        sum_sq_l = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_L);
        sum_sq_value = (((uint64_t)sum_sq_h) << 32) | ((uint64_t)sum_sq_l);

        // Sanity check to see if we actually got something.
        if ((sum_value != 0) && (histo_size > 0))
        {
            // Check to make sure our length agrees with what we expect.
            if ((histo_data.size() * 4) > (int)histo_size)
                histo_data.resize(histo_size / 4);

            // If we're mono, then we need to truncate the data.
            if (ui->histo_plot->is_mono())
                histo_data.resize(histo_data.size() / 3);

            // All done reading the histogram data.
            ui->histo_plot->populate_histogram(histo_data);

            // Only have statistics if we aren't raw.
            if (!ui->histo_plot->is_raw())
            {
                // Populate the statistics information we got from the camera.
                ui->histo_plot->set_statistics(FI_HISTOGRAM::BLUE_CURVE, sum_value, sum_sq_value);

                // If we're color, then read up these other channels as well.
                if (!ui->histo_plot->is_mono())
                {
                    sum_value = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_GRN);
                    sum_sq_h = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_H);
                    sum_sq_l = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_L);
                    sum_sq_value = (((uint64_t)sum_sq_h) << 32) | ((uint64_t)sum_sq_l);
                    ui->histo_plot->set_statistics(FI_HISTOGRAM::GREEN_CURVE, sum_value, sum_sq_value);

                    sum_value = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_RED);
                    sum_sq_h = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_H);
                    sum_sq_l = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_L);
                    sum_sq_value = (((uint64_t)sum_sq_h) << 32) | ((uint64_t)sum_sq_l);
                    ui->histo_plot->set_statistics(FI_HISTOGRAM::RED_CURVE, sum_value, sum_sq_value);
                }
            }
        }

        // Go around for another run?
        if (single_in_progress)
            single_in_progress = false;
        else
            update_histo();
    }
}

/*!
 * @brief Start the update timer.
 */
void HistoDock::resume()
{
    update_paused = false;
    if ((update_timer != nullptr) && isVisible())
        update_timer->start(UPDATE_INTERVAL);
}

/*!
 * @brief Configure the histogram plot to use raw data (immediately after capture) or
 *        processed samples (after the image processing pipeline).
 * @param use_raw   True to use raw samples, false for processed samples.
 */
void HistoDock::set_raw_mode(bool use_raw)
{
    bool    was_paused = update_paused;
    bool    activeCameraCanUseRaw = false;

    // Sanity check.
    if (active_camera == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }
    if (active_camera->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No sensor found.";
        return;
    }

    // Sanity check.
    if (active_camera->get_sensor() == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No sensor?";
        return;
    }

    // What sensor are we connecting to?
    switch (active_camera->get_sensor()->get_type())
    {
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO:
        // Raw histogram is available.
        activeCameraCanUseRaw = true;
        break;
    default:
        activeCameraCanUseRaw = false;
        break;
    }

    // If the active camera does not support a raw histogram can't set it.
    if (!activeCameraCanUseRaw && use_raw)
        use_raw = false;    // Force it back to non-raw support.

    // Should we pause the update?
    if (isVisible() && !update_paused)
    {
        // Pause the update.
        pause();
    }

    // Pass it on.
    ui->histo_plot->set_raw_mode(use_raw);

    // If we're not using raw mode, then restore the mono/color mode.
    if (!use_raw)
    {
        if (active_camera)
        {
            // Determine if the camera is mono or color.
            ui->histo_plot->set_mono(!active_camera->get_sensor()->is_color());
        }
        else if (active_mdi)
        {
            PlaybackMetaInfo *pb_meta;

            // Determine if the camera is mono or color. If we don't have a sensor
            // object, just assume mono.
            pb_meta = active_mdi->getMetadata();
            if (pb_meta == nullptr)
                ui->histo_plot->set_mono(true);
            else
                ui->histo_plot->set_mono(!pb_meta->isColor());
        }
    }

    // Should we resume the update?
    if (isVisible() && !was_paused)
    {
        // Resume the updates.
        resume();
    }
}

/*!
 * @brief The caller wants a single histogram update.
 */
void HistoDock::single_shot()
{
    // Are we visible or already doing a single shot?
    if (!isVisible() || single_in_progress)
        return;

    // Remember we're doing a single shot.
    single_in_progress = true;

    // Start the update after a brief delay.
    update_histo();
}

/*!
 * @brief Translate the UI elements for this object.
 */
void HistoDock::translate_ui()
{
    // Have the UI object perform it's translations.
    ui->retranslateUi(this);
}


//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void HistoDock::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    QDockWidget::changeEvent(event);
}

/*!
 * @brief We are closing the window, so make sure we stop our timer.
 *
 * @param event     Pointer to the event structure.
 */
void HistoDock::closeEvent(QCloseEvent *event)
{
    // Stop our timer.
    if (update_timer != nullptr)
        update_timer->stop();

    // Uncheck our menu item.
    MainWindow::GetInstance()->get_view_histogram()->setChecked(false);

    // Accept the event.
    event->accept();

    // Pass it on.
    QDockWidget::closeEvent(event);
}

void HistoDock::do_dock()
{
    setFloating(false);
}

/*!
 * @brief The user has right mouse clicked in our window. Put up a
 *        context specific menu.
 *
 * @param event_ptr     Pointer to the menu event.
 */
void HistoDock::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QAction     *new_action;
    QMenu       context_menu(this);

    // Sanity check.
    if (active_camera)
    {
        uint32_t    camera_state;
        uint32_t    cam_ip_addr;
        FI_comm     *comm_ptr;

        // Make sure the active camera is really active.
        if (active_camera->get_camera_info()->offline)
            return;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
            return;
        }

        // Get the IP address for our camera.
        cam_ip_addr = active_camera->get_ip_addr();

        // Get the camera's state.
        camera_state = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_CAMERA_STATE);

        // Build up our menu.
        if (update_timer != nullptr)
        {
            new_action = context_menu.addAction(tr("Pause"), this, &HistoDock::pause);
            new_action->setEnabled(update_timer->isActive() && (camera_state != GIGE_CAMERA_STATE_REVIEW));
            new_action = context_menu.addAction(tr("Resume"), this, &HistoDock::resume);
            new_action->setEnabled(!update_timer->isActive() && (camera_state != GIGE_CAMERA_STATE_REVIEW));
            context_menu.addSeparator();
        }
    }
    else if (active_mdi != nullptr)
    {
        // Build up our menu.
        if (update_timer != nullptr)
        {
            new_action = context_menu.addAction(tr("Pause"), this, &HistoDock::pause);
            new_action->setEnabled(update_timer->isActive());
            new_action = context_menu.addAction(tr("Resume"), this, &HistoDock::resume);
            new_action->setEnabled(!update_timer->isActive());
            context_menu.addSeparator();
        }
    }
    else
    {
        return;
    }

    if (ui->histo_plot->stats_are_visible())
        context_menu.addAction(tr("Hide statistics"), this, &HistoDock::toggle_statistics);
    else
        context_menu.addAction(tr("Show statistics"), this, &HistoDock::toggle_statistics);
    context_menu.addSeparator();
    context_menu.addAction(tr("Copy Histogram Data"), this, &HistoDock::copy_data);
    context_menu.addAction(tr("Copy Histogram Image"), this, &HistoDock::copy_image);

    context_menu.addSeparator();
    context_menu.addAction(tr("Zoom In"), this, &HistoDock::zoom_in);
    context_menu.addAction(tr("Zoom Out"), this, &HistoDock::zoom_out);
    context_menu.addAction(tr("Zoom Reset"), this, &HistoDock::zoom_reset);
    context_menu.addAction(tr("Set Zoom..."), this, &HistoDock::zoom_user);

#define ALLOW_LOCAL_HISTOGRAM 1
#ifdef ALLOW_LOCAL_HISTOGRAM
    context_menu.addSeparator();
    if (histo_local)
        context_menu.addAction(tr("Use camera's histogram"), this, &HistoDock::toggle_histogram);
    else
        context_menu.addAction(tr("Use local histogram"), this, &HistoDock::toggle_histogram);
#endif

    // Add a menu to re-dock the panel.
    if (isFloating())
    {
        context_menu.addSeparator();
        context_menu.addAction(tr("Dock"), this, &HistoDock::do_dock);
    }

    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void HistoDock::copy_data()
{
    int         i;
    QString     data_dump;
    QClipboard  *clipboard = QApplication::clipboard();

    // Is this a raw histogram?
    if (ui->histo_plot->is_raw())
    {
        QVector<double> vec_data;

        // Get the histogram data.
        vec_data = ui->histo_plot->get_mono_data();

        // Loop through all the data and build our string.
        for (i = 0; i < vec_data.size(); ++i)
            data_dump.append(QString::fromUtf8("%1\n").arg(vec_data[i]));
    }

    // Are we a mono image?
    else if (ui->histo_plot->is_mono())
    {
        QVector<double> vec_data;

        // Get the histogram data.
        vec_data = ui->histo_plot->get_mono_data();

        // Loop through all the data and build our string.
        for (i = 0; i < vec_data.size(); ++i)
            data_dump.append(QString::fromUtf8("%1\n").arg(vec_data[i]));
    }

    // Must be color.
    else
    {
        QVector<double> r_vec_data;
        QVector<double> g_vec_data;
        QVector<double> b_vec_data;

        // Get the histogram data.
        r_vec_data = ui->histo_plot->get_red_data();
        g_vec_data = ui->histo_plot->get_green_data();
        b_vec_data = ui->histo_plot->get_blue_data();

        // Loop through all the data and build a comma seperated list.
        for (i = 0; i < r_vec_data.size(); ++i)
            data_dump.append(QString::fromUtf8("%1, %2, %3\n").arg(r_vec_data[i]).arg(g_vec_data[i]).arg(b_vec_data[i]));
    }

    // Give it to the clipboard.
    clipboard->setText(data_dump);
}

/*!
 * @brief Copy the histogram image to the clipboard.
 */
void HistoDock::copy_image()
{
    QClipboard  *clipboard = QApplication::clipboard();
    clipboard->setPixmap(ui->histo_plot->grab());
}

/*!
 * @brief Handle the "delay" timer event.
 *
 *        This is needed to let the main window finish its startup sequence
 *        so that we'll have the resource we need to do network communications.
 */
void HistoDock::finish_setup()
{
    FI_comm     *comm_ptr;

    // Are we ready to play yet?
    if (MainWindow::GetInstance().isNull())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr != nullptr)
    {
        // Connect our signals and slots.
        connect(comm_ptr, &FI_comm::read_reg_complete, this, &HistoDock::read_complete, Qt::QueuedConnection);
        connect(MainWindow::GetInstance().get(), &MainWindow::active_child_changed, this, &HistoDock::new_active_camera);
    }

    // Enable/disable our timer.
    enable_timer(use_timer);
}

/*!
 * @brief A new camera (or possibly none) has become active.
 * @param window    The QMdiSubWindow pointer to the related camera window.
 */
void HistoDock::new_active_camera(QMdiSubWindow *window)
{
    // QLOG_TRACE() << __PRETTY_FUNCTION__ << "window:" << window << "active_camera:" << active_camera;

    // Get the "new" active camera.
    if (window == nullptr)
    {
        // Stop our update timer.
        if (update_timer != nullptr)
            update_timer->stop();

        // If we have an active camera, make sure we disconnect from it's signals.
        if (active_camera != nullptr)
        {
            disconnect(active_camera, &CameraMDI::video_is_paused, this, &HistoDock::stream_paused);
            disconnect(active_camera, &CameraMDI::video_is_streaming, this, &HistoDock::stream_live);
            disconnect(active_camera, &CameraMDI::closing, this, &HistoDock::parent_closing);
        }

        // Reset our active camera.
        active_camera = nullptr;
        active_mdi = nullptr;

        // Set our title.
        setWindowTitle(WINDOW_TITLE_STR);
    }
    else
    {
        QString     title_str;
        CameraMDI   *new_camera;

        // Get our new camera object.
        new_camera = qobject_cast<CameraMDI *>(window->widget());

        if (new_camera == nullptr)
        {
            PlaybackMDI *playback_cam = qobject_cast<PlaybackMDI *>(window->widget());

            // Is it a PlaybackMDI child?
            if (playback_cam == nullptr)
            {
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "Not a CameraMDI or PlaybackMDI!";
                return;
            }

            // Is it really new?
            else if (playback_cam != active_mdi)
            {
                PlaybackMetaInfo *pb_meta;

                // Remember our active child.
                active_mdi = playback_cam;

                // Determine if the camera is mono or color. If we don't have a sensor
                // object, just assume mono.
                pb_meta = active_mdi->getMetadata();
                if (pb_meta == nullptr)
                    ui->histo_plot->set_mono(true);
                else
                    ui->histo_plot->set_mono(!pb_meta->isColor());

                // Update our title string.
                title_str = WINDOW_TITLE_STR + QString::fromUtf8(": ") + active_mdi->getCameraInfo()->get_camera_name();
                setWindowTitle(title_str);

                // Make sure we only use the local histogram processor...since we aren't connected to a camera.
                use_local_histogram(true);

                // Are we visible?
                if (isVisible())
                    update_histo();
            }
        }

        // Is it really new?
        else if (new_camera != active_camera)
        {
            // Sanity check.
            QList<CameraMDI *>  child_mdi_list = MainWindow::GetInstance()->get_children();
            if (active_camera != nullptr)
            {
                if (!child_mdi_list.contains(active_camera))
                {
                    // The "active" camera window went away - crashed or something.
                    active_camera = nullptr;
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Active camera has gone away. Resetting.";
                }

                // Disconnect from the old camera.
                disconnect(active_camera, &CameraMDI::video_is_paused, this, &HistoDock::stream_paused);
                disconnect(active_camera, &CameraMDI::video_is_streaming, this, &HistoDock::stream_live);
                disconnect(active_camera, &CameraMDI::closing, this, &HistoDock::parent_closing);
            }

            // Remember our active camera.
            active_camera = new_camera;

            // Connect to the new camera.
            connect(active_camera, &CameraMDI::video_is_paused, this, &HistoDock::stream_paused);
            connect(active_camera, &CameraMDI::video_is_streaming, this, &HistoDock::stream_live);
            connect(active_camera, &CameraMDI::closing, this, &HistoDock::parent_closing);

            // Determine if the camera is mono or color. If we don't have a sensor
            // object, just assume mono.
            if (active_camera->get_sensor() == nullptr)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "We don't have active_camera->get_sensor()!  Setting histo plot as mono for a default.";
                ui->histo_plot->set_mono(true);
            }
            else
                ui->histo_plot->set_mono(!active_camera->get_sensor()->is_color());

            // Update our title string.
            title_str = WINDOW_TITLE_STR + QString::fromUtf8(": ") + active_camera->get_camera_name();
            setWindowTitle(title_str);

            // Are we visible?
            if (isVisible())
                update_histo();
        }
    }
}

void HistoDock::parent_closing()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "===========";
    new_active_camera(nullptr);
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void HistoDock::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Ignore this if we don't have an active camera.
    if (active_camera == nullptr)
        return;

    // Is this event for us?
    if (ip_addr != active_camera->get_ip_addr())
        return;

    // Don't care if we're not visible.
    if (!isVisible())
        return;

    // Perform any special action we might want on a particular register.
    switch (reg_addr)
    {
    case GIGE_REG_HISTO_RGB_SIZE:
    case GIGE_REG_HISTOGRAM_RAW_SIZE:
        // Ignore it if we're not visible.
        histo_read_start();
        break;
    case GIGE_REG_HISTO_RGB:
    case GIGE_REG_HISTOGRAM_RAW:
        // Ignore it if we're not visible.
        histo_read_update();
        break;

    default:
        break;
    }
}

/*!
 * @brief Show, or hide, the histogram dock widget.
 * @param visible   True to show, false to hide.
 */
void HistoDock::setVisible(bool visible)
{
    // Are we visible?
    if (visible)
    {
        // Do we have an IP address?
        if (active_camera != nullptr)
        {
            // Determine if the camera is mono or color. If we don't have a sensor
            // object, just assume mono.
            if (active_camera->get_sensor() == nullptr)
                ui->histo_plot->set_mono(true);
            else
                ui->histo_plot->set_mono(!active_camera->get_sensor()->is_color());

            // Start an update.
            update_histo();
        }
        else if (active_mdi != nullptr)
        {
            PlaybackMetaInfo *pb_meta;

            // Determine if the camera is mono or color. If we don't have a sensor
            // object, just assume mono.
            pb_meta = active_mdi->getMetadata();
            if (pb_meta == nullptr)
                ui->histo_plot->set_mono(true);
            else
                ui->histo_plot->set_mono(!pb_meta->isColor());

            // Start an update.
            update_histo();
        }
        else
        {
            // Can't be visible, don't have a camera!
            visible = false;
        }
    }
    else
    {
        // Stop our timer.
        if (update_timer != nullptr)
            update_timer->stop();
    }

    // Pass it on up.
    QDockWidget::setVisible(visible);
}

/*!
 * @brief Notification slot that the stream has switch to live mode.
 */
void HistoDock::stream_live()
{
    resume();
}

/*!
 * @brief Notification slot that the stream has been paused.
 */
void HistoDock::stream_paused()
{
    pause();
}

/*!
 * @brief It's time to update the histogram data from the camera.
 */
void HistoDock::time_to_update()
{
    // Stop the timer.
    if (update_timer != nullptr)
        update_timer->stop();

    // Make sure we're visible.
    if (!isVisible())
        return;

    // Make sure our histogram is in sync with the sensor color type.
    // NOTE:  This was added because there were conditions where the active camera
    // being set doesn't have full access to the sensor pointer when this function
    // is invoked and ends up setting a default mono state.  This gets called more
    // than once and when that occurs we miss the true setting of the sensor type.
    // This will ensure that we always are representing the proper active_camera in
    // a visible histogram.  One example of when this happens is when the black frame
    // is calibrated and the camera comes back online.
    if (active_camera)
    {
        fiSensor *active_camera_sensor_ptr = active_camera->get_sensor();
        if (active_camera_sensor_ptr != nullptr)
        {
            // If our sensor type is not in sync with the histogram setting we need to update it.
            bool change_histogram_type = (!active_camera_sensor_ptr->is_color() != ui->histo_plot->is_mono());

            // If we need to update our histogram do so.
            if (change_histogram_type)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Syncing the active camera sensor type isColor:" << active_camera_sensor_ptr->is_color()
                             << "with histogram is_mono():" << ui->histo_plot->is_mono() << "They were out of sync possibly due to camera going offline then back online.";

                // Need to reconfigure our histogram.
                ui->histo_plot->set_mono(!active_camera_sensor_ptr->is_color());
            }
        }
    }

    // Are we using the local histogram processor?
    if (histo_local)
    {
        // Update the histogram.
        update_local();
    }
    else
    {
        // Get our communications object pointer.
        FI_comm *comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
            return;
        }

        // Make sure we have an active camera.
        if (active_camera == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No active camera!";
            return;
        }

        // Are we offline?
        if (active_camera->get_camera_info()->offline)
        {
            // Restart our timer so we can swing back around later.
            // Maybe we'll be back on-line by then.
            if (update_timer != nullptr)
                update_timer->start(UPDATE_INTERVAL);
            return;
        }

        // Get the IP address for our camera.
        uint32_t cam_ip_addr = active_camera->get_ip_addr();

        // Start reading the histogram data.
        if (ui->histo_plot->is_raw())
        {
            comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_HISTOGRAM_RAW_SIZE);
        }
        else
        {
            FI_COMM::REG_ADDR_LIST  read_list;

            read_list.append(GIGE_REG_HISTO_RGB_SIZE);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_BLU);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_L);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_H);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_GRN);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_L);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_H);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_RED);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_L);
            read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_H);

            // Start the read.
            comm_ptr->start_read_multireg(cam_ip_addr, read_list);
        }
    }
}

/*!
 * @brief The user wants to toggle the processing of the histogram.
 */
void HistoDock::toggle_histogram()
{
    // Toggle whether we are processing locally or on the camera.
    use_local_histogram(!histo_local);
}

/*!
 * @brief The user wants to toggle the visibility of the histogram.
 */
void HistoDock::toggle_statistics()
{
    // Toggle the visibility of the statistics information.
    ui->histo_plot->show_statistics(!ui->histo_plot->stats_are_visible());
    update();
}

void HistoDock::use_local_histogram(bool use_local)
{
    histo_local = use_local;

    // Allocate a histogram processor if we don't already have one.
    if (histo_local)
    {
        if (histo_proc == nullptr)
        {
            histo_proc = new fiHistoProc(this);
        }
    }
    else
    {
        // Make sure we delete the histogram processor, if we have one.
        if (histo_proc != nullptr)
        {
            delete histo_proc;
            histo_proc = nullptr;
        }
    }
}

/*!
 * @brief Request a histogram update from the camera.
 */
void HistoDock::update_histo()
{
    // Update the histogram appropriately.
    if (histo_local)
        update_local();
    else
        update_remote();
}

void HistoDock::update_local()
{
    CameraVideo *active_player;

    // Sanity check.
    if (histo_proc == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "histo_proc is null?!";
        return;
    }

    // Update the ROI
    if (roi_override)
        histo_proc->set_roi(roi_rect);
    else
        histo_proc->set_roi(QRect(0, 0, 0, 0));

    // Handle raw vs. processed.
    if (active_mdi && ui->histo_plot->is_raw())
    {
        const QByteArray& raw_buffer = active_mdi->getRawImage();

        // Did we get a raw buffer?
        if (raw_buffer.size() == 0)
        {
            active_player = active_mdi->getPlayer();
            if (active_player)
            {
                QPixmap my_pixmap;

                active_player->copy_image(my_pixmap);
                histo_proc->calc(my_pixmap);
            }
        }
        else
        {
            int bpp;
            QRect img_rect;
            PlaybackMetaInfo *pb_meta = active_mdi->getMetadata();

            // Figure out how many bits per pixel we have in the capture.
            if (IS_8_BIT_MODE(active_mdi->getBitSel()))
                bpp = 8;
            else if (IS_10_BIT_MODE(active_mdi->getBitSel()))
                bpp = 10;
            else
                bpp = 12;

            // Get the image ROI.
            img_rect.setX(0);
            img_rect.setY(0);
            img_rect.setWidth((int)pb_meta->getWidth());
            img_rect.setHeight((int)pb_meta->getHeight());

            // Calculate the raw histogram.
            histo_proc->calc(raw_buffer, bpp, img_rect);
        }
    }
    else
    {
        // Make sure the histogram processor has the right mode.
        histo_proc->set_mono(ui->histo_plot->is_mono());

        // If we have an active camera, see if we can get the current image.
        if (active_camera)
            active_player = active_camera->get_player();
        else if (active_mdi)
            active_player = active_mdi->getPlayer();
        else
            active_player = nullptr;

        if (active_player)
        {
            QPixmap my_pixmap;

            active_player->copy_image(my_pixmap);
            histo_proc->calc(my_pixmap);
        }
    }

    // Reset our buffer.
    histo_data.clear();
    histo_data.append(*histo_proc->get_red());

    // If we're color, then read up these other channels as well.
    if (!ui->histo_plot->is_mono() && !ui->histo_plot->is_raw())
    {
        histo_data.append(*histo_proc->get_green());
        histo_data.append(*histo_proc->get_blue());
    }

    // All done reading the histogram data.
    ui->histo_plot->populate_histogram(histo_data);

    if (ui->histo_plot->is_mono() || ui->histo_plot->is_raw())
    {
        // For mono, we'll make use of the blue curve, etc on the histogram plot.
        ui->histo_plot->set_statistics(FI_HISTOGRAM::BLUE_CURVE, histo_proc->get_sum_mono(), histo_proc->get_sumsq_mono());
    }
    else
    {
        // For color, move the data over to the histogram plot.
        ui->histo_plot->set_statistics(FI_HISTOGRAM::RED_CURVE, histo_proc->get_sum_red(), histo_proc->get_sumsq_red());
        ui->histo_plot->set_statistics(FI_HISTOGRAM::GREEN_CURVE, histo_proc->get_sum_green(), histo_proc->get_sumsq_green());
        ui->histo_plot->set_statistics(FI_HISTOGRAM::BLUE_CURVE, histo_proc->get_sum_blue(), histo_proc->get_sumsq_blue());
    }

    // Go around for another run?
    if (single_in_progress)
        single_in_progress = false;

    // Start up our timer.
    if (use_timer && (!update_paused || single_in_progress))
    {
        if (update_timer != nullptr)
            update_timer->start(UPDATE_INTERVAL);
    }
}

void HistoDock::update_remote()
{
    uint32_t    cam_ip_addr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Make sure we have an active camera.
    if (active_camera == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active camera!";
        return;
    }

    // If we are not using the timer, we do not want to force an update.
    if (!use_timer)
        return;

    // Get the IP address for our camera.
    cam_ip_addr = active_camera->get_ip_addr();

    // Are we overriding the ROI?
    if (roi_override)
    {
        uint32_t reg_value;

        // Tell the camera our area of interest.
        reg_value = ((uint32_t)roi_rect.x() << 16) | (uint32_t)roi_rect.width();
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_ROI_X, reg_value);
        reg_value = ((uint32_t)roi_rect.y() << 16) | (uint32_t)roi_rect.height();
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_ROI_Y, reg_value);
    }
    else
    {
        // Get our current image settings.
        uint32_t image_width = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_WIDTH);
        uint32_t image_height = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HEIGHT);

        // Tell the camera our area of interest.
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_ROI_X, image_width);
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_ROI_Y, image_height);
    }

    // Start the gathering proceess. Only need to initiate a "processed" histogram. Raw
    // histograms are running all the time.
    if (!ui->histo_plot->is_raw())
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_GENERATE_RGB_HISTO, GIGE_GENERATE_RGB_HISTO_CHAN_P0);

    // Start up our timer.
    if (use_timer && (!update_paused || single_in_progress))
    {
        if (update_timer != nullptr)
            update_timer->start(UPDATE_INTERVAL);
    }
}

void HistoDock::zoom_in()
{
    double new_scale = ui->histo_plot->get_y_scale();

    // Make sure we don't go too crazy.
    if (new_scale > 0.02)
    {
        if (new_scale < 0.2)
            new_scale -= 0.01;
        else
            new_scale -= 0.1;
    }

    ui->histo_plot->set_y_scale(new_scale);
}

void HistoDock::zoom_out()
{
    double new_scale = ui->histo_plot->get_y_scale();
    new_scale += 0.1;
    ui->histo_plot->set_y_scale(new_scale);
}

void HistoDock::zoom_reset()
{
    ui->histo_plot->set_y_scale(1.0);
}

void HistoDock::zoom_user()
{
    bool ok;
    double new_scale = ui->histo_plot->get_y_scale();

    new_scale = QInputDialog::getDouble(this, tr("Enter the Scale Factor"), tr("Scale:"), new_scale, 0.001, 100.0, 3, &ok);
    if (ok)
        ui->histo_plot->set_y_scale(new_scale);
}
