/*!
 * @file jc_local.cpp
 * @brief The implementation of the jcLocal object.
 *
 * Because of issues with QObject, this must be in its own
 * separate header and source file.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QObject>
#include <QProcess>

#include "jc_local.h"

#include "fi_journalctl.h"
#include "logger.h"

namespace FI_JOURNALCTL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

//! The maximum number of bytes we'll read in one shot.
const int MAX_READ_LEN      = 8192000;

//--------------------------------------------------------------------
// Class definition for jcLocal
//--------------------------------------------------------------------

jcLocal::jcLocal(QObject *parent) : QObject(parent),
    my_process(std::unique_ptr<QProcess>(new QProcess(parent)))
{
    if (my_process.get() == nullptr)
    {
        error_info = tr("Unable to create the communications object.");
        return;
    }

    // Catch the "data available" signal.
    connect(my_process.get(), &QProcess::readyRead, this, &jcLocal::processHasData);
}

jcLocal::~jcLocal()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Shutdown our process if it's running.
    if (isRunning())
        close();

    // Release the process object.
    my_process.reset();
}

void jcLocal::run(const QString& cmd, const QStringList& args)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cmd:" << cmd << "args:" << args;

    // Make sure to terminate an already running process first
    if (isRunning())
    {
        my_process->close();
    }

    // Get the process running.
    my_process->start(cmd, args);
}

QString jcLocal::runAndWait(const QString& cmd, const QStringList& args)
{
    QProcess p;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cmd:" << cmd << "args:" << args;

    // Start our command.
    p.start(cmd, args);

    // Block and wait for the operation to finish.
    p.waitForFinished(-1);

    // Return the information from the standard output.
    return QString(p.readAllStandardOutput());
}

bool jcLocal::isRunning()
{
    // Sanity check.
    if (my_process.get() == nullptr)
        return false;

    // Return our state.
    return ((my_process->state() == QProcess::Running) || (my_process->state() == QProcess::Starting));
}

void jcLocal::close()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (my_process.get() == nullptr)
        return;

    // Close the process.
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Close the process";
    my_process->close();
}

const QString jcLocal::getErrorString()
{
    if (my_process.get() == nullptr)
        return error_info;

    return my_process->errorString();
}

void jcLocal::processHasData()
{
    // Collect the process output.
    QByteArray read = my_process->read(MAX_READ_LEN);

    QString readString = QString(read);
    readString = readString.left(readString.size() - 1);

    emit dataAvailable(readString);
}

} // End of FI_JOURNALCTL
