/*!
 * @file jc_remote.cpp
 * @brief The implementation of the Fastec jcRemote object.
 *
 * Because of issues with QObject, this must be in its own
 * separate header and source file.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QDeadlineTimer>
#include <QObject>
#include <QtConcurrent>

#include "jc_remote.h"
#include "logger.h"

namespace FI_JOURNALCTL
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macros
//--------------------------------------------------------------------

const int MAX_BUFFER_LEN        = (64 * 1024);
const uint SSH_STANDARD_PORT    = 22;
const int WAIT_TIMEOUT_MS       = 500;
const int ZOMBIE_WAIT_MS        = 50;

//--------------------------------------------------------------------
// Class definition for jcRemote
//--------------------------------------------------------------------

jcRemote::jcRemote(QObject *parent) : QObject(parent),
    thread_stop(false),
    new_cmd_pending(false),
    ssh(nullptr),
    channel(nullptr)
{
}

jcRemote::~jcRemote()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Make sure we're closed.
    close();

    // Release the SSH object.
    if (ssh != nullptr)
    {
        ssh_disconnect(ssh);
        ssh_free(ssh);
        ssh = nullptr;
    }
}

bool jcRemote::setSecurity(const FI_JOURNALCTL::SSH_SETTINGS *ssh_info)
{
    bool rc(false);
    int ok;
    uint port = SSH_STANDARD_PORT;

    // Allocate our SSH object.
    ssh = ssh_new();
    if (ssh == nullptr)
    {
        error_info = tr("Unable to allocate the SSH object.");
        QLOG_TRACE() << __PRETTY_FUNCTION__ << error_info;
        return false;
    }

    // Initialize the options for the SSH object.
    ssh_options_set(ssh, SSH_OPTIONS_HOST, ssh_info->hostname.toUtf8().constData());
    ssh_options_set(ssh, SSH_OPTIONS_USER, ssh_info->username.toUtf8().constData());
    ssh_options_set(ssh, SSH_OPTIONS_PORT, &port);

    // See if we can connect.
    ok = ssh_connect(ssh);
    if (ok != SSH_OK)
    {
        error_info = tr("Establishing a connection to the remote host failed.");
        QLOG_TRACE() << __PRETTY_FUNCTION__ << error_info;
        goto security_bail;
    }

    // Check to see if the server is OK.
    ok = ssh_session_is_known_server(ssh);
    switch (ok)
    {
    case SSH_KNOWN_HOSTS_OK:
        // Everything okay, nothing to do!
        break;

    // Remote server is not in known_hosts file
    case SSH_KNOWN_HOSTS_NOT_FOUND:
    case SSH_KNOWN_HOSTS_UNKNOWN:
        // Add the host to the known hosts file
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Adding" << ssh_info->hostname << "to known hosts.";
        ssh_session_update_known_hosts(ssh);
        break;

    case SSH_KNOWN_HOSTS_CHANGED:
    case SSH_KNOWN_HOSTS_OTHER:
        // Remember what happened here.
        error_info = tr("Remote Host Identification has changed.");
        QLOG_TRACE() << __PRETTY_FUNCTION__ << error_info;
        goto security_bail;

    // Uncovered error
    default:
        // Remember what happened here.
        error_info = tr("Key verification using your .ssh/known_hosts file failed.");
        QLOG_TRACE() << __PRETTY_FUNCTION__ << error_info;
        goto security_bail;
    }

    // See if we can log in.
    ok = ssh_userauth_password(ssh, ssh_info->username.toUtf8().constData(), ssh_info->crypt_info.toUtf8().constData());

    // Check if the selected authentication method worked
    if (ok != SSH_AUTH_SUCCESS)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Did not receive SSH_AUTH_SUCCESS: " << ssh_get_error(ssh);
        error_info = tr("SSH Authentication on the remote host failed.");
        goto security_bail;
    }

    // { Authenticated connection is established }
    channel = nullptr;
    initSSHChannel();

    // Start our processing thread. Make sure we clear our command.
    cmd.clear();
    ssh_thread = QtConcurrent::run(sshHandler, this);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "sshHandler() forked.";

    // Everything is fine.
    rc = true;

security_bail:
    if (!rc)
    {
        // Release our SSH object.
        ssh_free(ssh);
        ssh = nullptr;
    }

    // Return our status.
    return rc;
}

void jcRemote::close()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Do we have our SSH thread running?
    if (ssh_thread.isRunning())
    {
        // Set our flag to stop the thread.
        thread_stop = true;
        new_cmd_pending = true;

        // Make sure the thread is awake.
        worker_wakeup.wakeAll();

        // Wait for the thread to finish.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Wait for the thread to finish.";
        ssh_thread.waitForFinished();
    }

    // We've closed down our thread, so we don't need to worry about
    // the mutex here.
    ssh_channel_close(channel);
}

void jcRemote::run(const QString& requested_cmd, const QStringList& args)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "requested_cmd:" << requested_cmd;

    // Let the worker know we want to post a new command.
    new_cmd_pending = true;

    // The ordering here is important. We need to initialize the channel
    // first. If the thread is currently busy working on the channel, then
    // this will block until the thread finishes.
    {
        QMutexLocker locker(&ssh_mutex);

        // Initialize our channel.
        initSSHChannel();
    }

    // Now that we've initialized the channel, we can tell the thread
    // to start executing the command.
    {
        QMutexLocker locker(&cmd_mutex);

        // Make sure we don't stop the thread!
        thread_stop = false;

        // Build up our command.
        cmd = requested_cmd;
        foreach (const QString& arg_str, args)
        {
            cmd.append(QLatin1Char(' ') + arg_str);
        }

        // Make certain that the command is terminated with a newline.
        cmd.append(QLatin1Char('\n'));

        // Tell the worker thread to run.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Wake up the worker. cmd:" << cmd;
        worker_wakeup.wakeAll();
    }

    // Done with the new command.
    new_cmd_pending = false;
}

QString jcRemote::runAndWait(const QString& requested_cmd, const QStringList& args)
{
    std::unique_ptr<char[]> buffer;
    QString resultString;
    QString rnw_cmd;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "requested_cmd:" << requested_cmd << "args:" << args;

    // Let the worker know we want to post a new command.
    new_cmd_pending = true;

    // Make sure we have access to the channel.
    QMutexLocker locker(&ssh_mutex);

    // Initialize our channel.
    initSSHChannel();

    // Make certain that the command is terminated with a newline.
    rnw_cmd = requested_cmd;
    foreach (const QString& arg_str, args)
    {
        rnw_cmd.append(QLatin1Char(' ') + arg_str);
    }

    // Make certain that the command is terminated with a newline.
    rnw_cmd.append(QLatin1Char('\n'));

    // So we have to run the command ourself
    ssh_channel_request_exec(channel, rnw_cmd.toUtf8().constData());

    // Block until the process ended
    ssh_channel_get_exit_status(channel);
    int bytes = ssh_channel_poll(channel, 0);

    // Error?
    if (bytes <= 0)
    {
        resultString.clear();
        goto rnw_bail;
    }

    // Copy received data
    buffer = std::unique_ptr <char[]>(new char[bytes]);
    ssh_channel_read(channel, (void *)buffer.get(), (unsigned int)bytes, 0);
    resultString.append(buffer.get());

rnw_bail:
    // Done with the new command.
    new_cmd_pending = false;

    return resultString;
}

bool jcRemote::isRunning()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Make sure we have access to the channel.
    QMutexLocker locker(&ssh_mutex);
    return (ssh_channel_is_open(channel) && !ssh_channel_is_eof(channel));
}

void jcRemote::processHasData()
{
    QString readString = "\n";
    readString = readString.left(readString.size() - 1);

    emit dataAvailable(readString);
}

/*!
 * @brief Initialize the SSH channel.
 *
 * @note    The assumption is that the caller is holding ssh_mutex!
 * @return  True on success, otherwise false.
 */
bool jcRemote::initSSHChannel()
{
    int ok;

    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Is there a channel already open?
    if ((channel != nullptr) && ssh_channel_is_open(channel))
    {
        // Close down the channel.
        ssh_channel_close(channel);
        ssh_channel_free(channel);
    }

    // Allocate a new channel.
    channel = ssh_channel_new(ssh);
    if (channel == nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "ssh_channel_new() failed!";
        return false;
    }

    // Open the new channel.
    ok = ssh_channel_open_session(channel);
    while (ok != SSH_OK)
    {
        // Wait for "Zombie sessions" to be killed
        QThread::msleep(ZOMBIE_WAIT_MS);

        // See if we can open the channel now.
        ok = ssh_channel_open_session(channel);
    }

    // All done.
    return true;
}


void jcRemote::sshHandler(QObject *parent)
{
    std::unique_ptr<char []> buffer;
    QString active_cmd;
    jcRemote *my_class = qobject_cast<jcRemote *>(parent);

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "==========================";

    // Allocate our buffer.
    buffer = std::unique_ptr<char []>(new char[MAX_BUFFER_LEN]);

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Make sure we're not supposed to stop.
    while (!my_class->thread_stop)
    {
        // Check if there is a new command or we can still read the old output

        // Case 1:      There is a new command available, wait for it
        //              to be fixed. (User might be scrolling through settings)
        // Otherwise:   Directly start reading!
        my_class->cmd_mutex.lock();
        if (my_class->cmd.isEmpty())
        {
            QDeadlineTimer  wait_ms(WAIT_TIMEOUT_MS);

            // On return from our wait, the mutex will be locked.
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "sleeping...";
            my_class->worker_wakeup.wait(&my_class->cmd_mutex, wait_ms);
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "...Wake up! thread_stop:" << my_class->thread_stop;

            // Make sure we're not supposed to stop.
            if (my_class->thread_stop)
            {
                // Release the mutex and bail out of the loop.
                my_class->cmd_mutex.unlock();
                break;
            }
        }

        // Sanity check.
        if (my_class->cmd.isEmpty())
        {
            // Release the mutex and go back to the top of the loop.
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "cmd is still empty.";
            my_class->cmd_mutex.unlock();
            continue;
        }

        // Extract current command and then we can release the mutx.
        active_cmd = my_class->cmd;

        // We're done with the command string, so clear it.
        my_class->cmd.clear();
        my_class->cmd_mutex.unlock();

        // Protect access to the SSH channel.
        my_class->ssh_mutex.lock();

        // Execute the command.
        ssh_channel_request_exec(my_class->channel, active_cmd.toUtf8().constData());

        // If the channel isn't ready yet, wait for it!
        if (ssh_channel_is_eof(my_class->channel) || !ssh_channel_is_open(my_class->channel))
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "Channel isn't ready, or open";
            my_class->ssh_mutex.unlock();
            continue;
        }
        my_class->ssh_mutex.unlock();

        // Loop while we are reading and there is no new command pending.
        bool reading = true;
        int blocks_read = 0;
        while (reading && !my_class->new_cmd_pending && !my_class->thread_stop)
        {
            int bytes_read;

            // Read the result of the command.
            my_class->ssh_mutex.lock();
            bytes_read = ssh_channel_read_nonblocking(my_class->channel, buffer.get(), MAX_BUFFER_LEN, 0);
            my_class->ssh_mutex.unlock();

            // Did we get anything?
            if (bytes_read > 0)
            {
                buffer[bytes_read] = '\0';
                QString dataString(buffer.get());

                // Let folks know what we've got.
                emit my_class->dataAvailable(dataString);

                // Is it time to take a breath?
                ++blocks_read;
                if (!my_class->thread_stop && (blocks_read > 4))
                {
                    // Grab the mutex.
                    my_class->cmd_mutex.lock();

                    // On return from our wait, the mutex will be locked.
                    QDeadlineTimer  wait_ms(50);
                    my_class->worker_wakeup.wait(&my_class->cmd_mutex, wait_ms);

                    // Release the mutex.
                    my_class->cmd_mutex.unlock();

                    // Reset our block read counter.
                    blocks_read = 0;
                }
            }
            else if (bytes_read == SSH_ERROR)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Received SSH_ERROR: " << ssh_get_error(my_class->ssh);
                reading = false;
            }
        }
    }
}

} // End of FI_JOURNALCTL
