/*!
 * @file jump_dialog.cpp
 * @brief This is the source file for the "Go To Time" dialog class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------
#include <QDebug>
#include <qmath.h>

#include "jump_dialog.h"
#include "ui_jump_dialog.h"

#include "logger.h"
#include "playback_meta_info.h"

namespace JUMP_DLG
{
//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

enum
{
    CB_PRE_TRIGGER = 0,         //!< The value is pre-trigger.
    CB_POST_TRIGGER             //!< The value is post-trigger.
};

// Developer's note: The primary difference between const and constexpr variables is that the
// initialization of a const variable can be deferred until run time. A constexpr variable
// must be initialized at compile time.
constexpr double    SEC_PER_MINUTE  = 60.0;
constexpr double    SEC_PER_HOUR    = (SEC_PER_MINUTE * 60.0);
constexpr double    SEC_PER_DAY     = (SEC_PER_HOUR * 24.0);
constexpr double    SEC_PER_YEAR    = (SEC_PER_DAY * 365.0);
}

using namespace JUMP_DLG;

/*!
 * @brief Dialog class constructor.
 */
JumpDialog::JumpDialog(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::JumpDialog)
{
    ui->setupUi(this);

    // Initialize our local variables.
    use_relative_time = true;
    frame_rate = 1;
    trigger_frame = 0;
    cur_frame = 0;
    meta_info = nullptr;

    // Assume there's no meta data and so there's no absolute time.
    ui->rb_absolute->setEnabled(false);

    // Setup our connections.
    connect(ui->rb_absolute, SIGNAL(clicked()), this, SLOT(absolute_clicked()));
    connect(ui->rb_relative, SIGNAL(clicked()), this, SLOT(relative_clicked()));
    connect(ui->sld_frames, SIGNAL(valueChanged(int)), this, SLOT(frame_changed(int)));
    connect(ui->sb_seconds, SIGNAL(valueChanged(double)), this, SLOT(seconds_changed(double)));
    connect(ui->sb_minute, SIGNAL(valueChanged(int)), this, SLOT(minutes_changed(int)));
    connect(ui->sb_hour, SIGNAL(valueChanged(int)), this, SLOT(hours_changed(int)));
    connect(ui->sb_yday, SIGNAL(valueChanged(int)), this, SLOT(days_changed(int)));
    connect(ui->sb_year, SIGNAL(valueChanged(int)), this, SLOT(years_changed(int)));
    connect(ui->cb_relative_pos, SIGNAL(currentIndexChanged(int)), this, SLOT(position_changed(int)));
}

/*!
 * @brief Dialog class destructor.
 */
JumpDialog::~JumpDialog()
{
    delete ui;
}

/*!
 * @brief Set the per-frame metadata object that we'll use to manage our timing
 *        calculations, etc.
 *
 * @param meta_ptr  Pointer to the metadata object.
 */
void JumpDialog::set_pfmeta(PlaybackMetaInfo *meta_ptr)
{
    bool abs_enable;

    // Save the pointer to our meta information.
    meta_info = meta_ptr;
    if ((meta_info != nullptr) && (meta_info->getRawBuffer() != nullptr))
        abs_enable = true;
    else
        abs_enable = false;

    // If we got a pointer, enable the absolute button.
    ui->rb_absolute->setEnabled(abs_enable);
}

/*!
 * @brief Set the relative time constants. Also used for absolute time, these
 *        parameters define the bounds within we'll have to operate.
 *
 * @param min_frame     The minimum absolute frame number (inclusive).
 * @param max_frame     The maximum absolute frame number (inclusive).
 * @param cur_frame     The current frame the user is reviewing.
 * @param trig_frame    The trigger frame.
 * @param fps           The frame per second used for the recording.
 */
void JumpDialog::set_rel_time(int32_t min_frame, int32_t max_frame, int32_t cur_frame, int32_t trig_frame, int32_t fps)
{
    // Save the settings.
    frame_rate = fps;
    trigger_frame = trig_frame;
    this->cur_frame = cur_frame;

    // Update the controls.
    ui->sld_frames->setMinimum(min_frame);
    ui->sld_frames->setMaximum(max_frame);
    ui->sld_frames->setValue(cur_frame);
    ui->sb_seconds->setSingleStep(1.0 / (double)frame_rate);
}

/*!
 * @brief Specify if we're to show relative or absolute time.
 *
 * @param do_rel_time   True to show relative time, false for absolute.
 */
void JumpDialog::use_rel_time(bool do_rel_time)
{
    // Update our radio buttons appropriately.
    if (do_rel_time)
        relative_clicked();
    else
        absolute_clicked();
}

/*!
 * @brief Retrieve and parse for the various controls the absolute time specified
 *        by the frame number (the sld_frames value).
 */
void JumpDialog::parse_abs_time()
{
    QStringList time_list;
    QString     time_str;

    // Make sure our position combo box is hidden.
    ui->cb_relative_pos->hide();

    // Sanity check.
    if (meta_info == nullptr)
        return;
    if (meta_info->getRawBuffer() == nullptr)
        return;

    // Get the absolute time for this frame.
    if (!meta_info->getPfmdTimeStr((uint32_t)(ui->sld_frames->value() - ui->sld_frames->minimum()), time_str))
    {
        qDebug() << __PRETTY_FUNCTION__ << "Can't get the per-frame metadata for frame" << ui->sld_frames->value();
        return;
    }

    // Split the string into it's component parts.
    time_list = time_str.split(QString::fromUtf8(":"));     // Get the YY:DDD:HH:MM:SS.microsec of the time.

    // Make sure we got everything.
    if (time_list.size() != 5)
    {
        qDebug() << __FUNCTION__ << "split failed, only produced" << time_list;
        return;
    }

    // Set the spinboxes for the absolute time.
    block_controls(true);
    ui->sb_year->setValue(time_list.at(0).toInt());
    ui->sb_yday->setValue(time_list.at(1).toInt());
    ui->sb_hour->setValue(time_list.at(2).toInt());
    ui->sb_minute->setValue(time_list.at(3).toInt());
    ui->sb_seconds->setValue(time_list.at(4).toDouble());
    block_controls(false);
}

/*!
 * @brief Retrieve and parse for the various controls the relative time specified
 *        by the frame number (the sld_frames value).
 */
void JumpDialog::parse_rel_time()
{
    int32_t int_time;
    double  temp_time;
    uint32_t pfmd_sec;
    uint32_t pfmd_usec;

    // Make sure our position combo box is displayed.
    ui->cb_relative_pos->show();

    // Get the per-frame metadata based time for the selected frame.
    // Calculate our relative time value.
    pfmd_sec = meta_info->getPfmdTime(ui->sld_frames->value(), pfmd_usec);
    temp_time = ((double)pfmd_sec) + ((double)pfmd_usec / 1000000.0);
    temp_time = temp_time - trigger_time();

    // Prevent our controls from propagating signals.
    block_controls(true);

    // Have the combo box display the correct value.
    if ((ui->sld_frames->value() < trigger_frame) ||
        ((ui->sld_frames->value() == trigger_frame) && (trigger_frame == ui->sld_frames->maximum())))
    {
        // A negative relative time means we're pre-trigger.
        ui->cb_relative_pos->setCurrentIndex(JUMP_DLG::CB_PRE_TRIGGER);
        temp_time = -temp_time;
    }
    else
    {
        // A positive (or zero) time means we're post-trigger.
        ui->cb_relative_pos->setCurrentIndex(JUMP_DLG::CB_POST_TRIGGER);
    }

    // Initialize our time controls.
    int_time = (int)(temp_time / JUMP_DLG::SEC_PER_YEAR);
    temp_time -= (int_time * JUMP_DLG::SEC_PER_YEAR);
    ui->sb_year->setValue(int_time);
    int_time = (int)(temp_time / JUMP_DLG::SEC_PER_DAY);
    temp_time -= (int_time * JUMP_DLG::SEC_PER_DAY);
    ui->sb_yday->setValue(int_time + 1);
    int_time = (int)(temp_time / JUMP_DLG::SEC_PER_HOUR);
    temp_time -= (int_time * JUMP_DLG::SEC_PER_HOUR);
    ui->sb_hour->setValue(int_time);
    int_time = (int)(temp_time / JUMP_DLG::SEC_PER_MINUTE);
    temp_time -= (int_time * JUMP_DLG::SEC_PER_MINUTE);
    ui->sb_minute->setValue(int_time);
    ui->sb_seconds->setValue(temp_time);
    block_controls(false);
}

/*!
 * @brief Convert the time (relative or absolute) to a frame number.
 */
int32_t JumpDialog::time_to_frame()
{
    double  total_time;
    int32_t frame_num = 0;
    uint32_t pfmd_sec;
    uint32_t pfmd_usec;


    // Process based on using relative or absolute time.
    if (use_relative_time)
    {
        // Accumulate the time.
        total_time = ui->sb_seconds->value();
        if (ui->sb_minute->isEnabled())
            total_time += (ui->sb_minute->value() * JUMP_DLG::SEC_PER_MINUTE);
        if (ui->sb_hour->isEnabled())
            total_time += (ui->sb_hour->value() * JUMP_DLG::SEC_PER_HOUR);
        if (ui->sb_yday->isEnabled())
            total_time += ((ui->sb_yday->value() - 1) * JUMP_DLG::SEC_PER_DAY);
        if (ui->sb_year->isEnabled())
            total_time += (ui->sb_year->value() * JUMP_DLG::SEC_PER_YEAR);

        // Convert the relative time to an absolute time.
        total_time += trigger_time();
    }
    else
    {
        // Accumulate the time.
        total_time = ui->sb_seconds->value();
        total_time += (ui->sb_minute->value() * JUMP_DLG::SEC_PER_MINUTE);
        total_time += (ui->sb_hour->value() * JUMP_DLG::SEC_PER_HOUR);
        total_time += ((ui->sb_yday->value() - 1) * JUMP_DLG::SEC_PER_DAY);
        total_time += (ui->sb_year->value() * JUMP_DLG::SEC_PER_YEAR);
    }

    // Split our time into seconds & microseconds.
    pfmd_sec = (uint32_t)qFloor(total_time);
    pfmd_usec = (uint32_t)((total_time - (double)pfmd_sec) * 10000000.0);
    pfmd_usec = ((pfmd_usec + 5) / 10);

    // See if we can find a frame at this time.
    frame_num = (int32_t)meta_info->findFrame(pfmd_sec, pfmd_usec);

    // Return our frame number.
    return frame_num;
}

/*!
 * @brief Update the various dialog controls to reflect the value selected by the user.
 */
void JumpDialog::update_display()
{
    // Process based on using relative or absolute time.
    if (use_relative_time)
    {
        // Parse and update our relative time controls.
        parse_rel_time();
    }
    else
    {
        // Parse and update our absolute time controls.
        parse_abs_time();
    }
}

/*!
 * @brief Return the selected absolute frame number.
 */
uint32_t JumpDialog::get_frame()
{
    return (uint32_t)ui->sld_frames->value();
}

int JumpDialog::exec()
{
    // Sanity checks.
    if (meta_info == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No metadata specified!";
        return QDialog::Rejected;
    }

    // Continue with the parent class.
    return QDialog::exec();
}

/*!
 * @brief Utility function to handle when the user changes one of the time
 *        spinbox controls.
 */
void JumpDialog::time_changed()
{
    int32_t new_frame;

    // Convert time to a frame number.
    new_frame = time_to_frame();

    // Sanity check.
    if (new_frame <= ui->sld_frames->minimum())
    {
        // Change the slider to the minimum value.
        if (ui->sld_frames->value() != ui->sld_frames->minimum())
        {
            // Force the controls to stay at the minimum. Set cur_frame
            // to this bogus value (it could be anything) so that the controls
            // will update.
            cur_frame = new_frame - 1;
            ui->sld_frames->setValue(ui->sld_frames->minimum());
        }
        else
        {
            // Simply force the controls to update.
            update_display();
        }
    }
    else if (new_frame >= ui->sld_frames->maximum())
    {
        // Change the slider to the maximum value.
        if (ui->sld_frames->value() != ui->sld_frames->maximum())
        {
            // Force the controls to stay at the maximum. Set cur_frame
            // to this bogus value (it could be anything) so that the controls
            // will update.
            cur_frame = new_frame + 1;
            ui->sld_frames->setValue(ui->sld_frames->maximum());
        }
        else
        {
            // Simply force the controls to update.
            update_display();
        }
    }

    // Set the slider to reflect the new frame.
    else if (new_frame != cur_frame)
    {
        // Remember the new frame.
        cur_frame = new_frame;

        // Update our frame number display box.
        ui->frame_num->setNum(new_frame - trigger_frame);

        // Change the slider position.
        ui->sld_frames->setValue(new_frame);
    }
}

double JumpDialog::trigger_time()
{
    double trigger_time = 0.0;

    // Make sure we have our pointer.
    if (meta_info != nullptr)
    {
        uint32_t pfmd_sec;
        uint32_t pfmd_usec;

        // Calculate our trigger's absolute time.
        pfmd_sec = meta_info->getPfmdTime(trigger_frame, pfmd_usec);
        trigger_time = ((double)pfmd_sec) + ((double)pfmd_usec / 1000000.0);
    }

    // Return the absolute trigger time.
    return trigger_time;
}

//-----------------------------------------------------------------------------
// Slots
//-----------------------------------------------------------------------------

/*!
 * @brief The user has clicked the "relative" radio button. We need
 *        to make sure our display reflects this mode change.
 */
void JumpDialog::relative_clicked()
{
    uint32_t    pfmd_sec;
    uint32_t    pfmd_usec;
    double      rel_time;
    int32_t     frame_num;

    // Save the current frame number before we switch over.
    frame_num = ui->sld_frames->value();

    // Update our user interface.
    use_relative_time = true;
    ui->rb_absolute->setChecked(false);
    ui->rb_relative->setChecked(true);

    // Calculate our relative time value as defined by the position (pre- or post-trigger).
    if ((ui->cb_relative_pos->currentIndex() == JUMP_DLG::CB_PRE_TRIGGER) || (ui->sld_frames->maximum() == trigger_frame))
    {
        // Get the minimum frame time.
        pfmd_sec = meta_info->getPfmdTime(ui->sld_frames->minimum(), pfmd_usec);
    }
    else
    {
        // Get the maximum frame time.
        pfmd_sec = meta_info->getPfmdTime(ui->sld_frames->maximum(), pfmd_usec);
    }

    // Convert the frame number into a time.
    rel_time = ((double)pfmd_sec) + ((double)pfmd_usec / 1000000.0);
    rel_time = rel_time - trigger_time();

    // Make sure we have an absolute value (don't want it negative for the following calculations).
    if (rel_time < 0)
        rel_time = -rel_time;

    // Don't allow the signals caused by our changes below to propagate.
    block_controls(true);

    // Disable the unnecessary fields. For those fields which are enabled,
    // make sure we set the maximum allowed value.
    if (rel_time < JUMP_DLG::SEC_PER_YEAR)
    {
        ui->sb_year->setValue(0);
        ui->frm_year->setEnabled(false);
    }
    else
    {
        // Enable the control.
        ui->frm_year->setEnabled(true);
    }
    if (rel_time < JUMP_DLG::SEC_PER_DAY)
    {
        ui->sb_yday->setValue(1);
        ui->frm_yday->setEnabled(false);
    }
    else
    {
        // Enable the control.
        ui->frm_yday->setEnabled(true);
    }
    if (rel_time < JUMP_DLG::SEC_PER_HOUR)
    {
        ui->sb_hour->setValue(0);
        ui->frm_hour->setEnabled(false);
    }
    else
    {
        // Enable the control.
        ui->frm_hour->setEnabled(true);
    }
    if (rel_time < JUMP_DLG::SEC_PER_MINUTE)
    {
        ui->sb_minute->setValue(0);
        ui->frm_minute->setEnabled(false);
    }
    else
    {
        // Enable the control.
        ui->frm_minute->setEnabled(true);
    }

    // Re-enable signalling.
    block_controls(false);

    // Now restore the frame number.
    cur_frame = -1;
    ui->sld_frames->setValue(frame_num);
    frame_changed(frame_num);
}

/*!
 * @brief The user has clicked the "absolute" radio button. We need
 *        to make sure our display reflects this mode change.
 */
void JumpDialog::absolute_clicked()
{
    QString     time_str;
    QStringList min_list;
    QStringList max_list;
    int32_t     frame_num;

    // Save the current frame number before we switch over.
    frame_num = ui->sld_frames->value();

    // Update our user interface.
    use_relative_time = false;
    ui->rb_absolute->setChecked(true);
    ui->rb_relative->setChecked(false);

    // Get the time for the minimum and maximum frames. This will help
    // limit the UI selections for the user.
    if (!meta_info->getPfmdTimeStr(0, time_str))
    {
        qDebug() << __PRETTY_FUNCTION__ << "Can't get the per-frame metadata for frame" << ui->sld_frames->minimum();
        return;
    }

    // Split the string into it's component parts.
    min_list = time_str.split(QString::fromUtf8(":"));     // Get the YY:DDD:HH:MM:SS.microsec of the time.

    // Make sure we got everything.
    if (min_list.size() != 5)
    {
        qDebug() << __FUNCTION__ << "split failed, only produced" << min_list;
        return;
    }

    // Get the maximum time.
    if (!meta_info->getPfmdTimeStr((uint32_t)(ui->sld_frames->maximum() - ui->sld_frames->minimum()), time_str))
    {
        qDebug() << __PRETTY_FUNCTION__ << "Can't get the per-frame metadata for frame" << ui->sld_frames->maximum();
        return;
    }

    // Split the string into it's component parts.
    max_list = time_str.split(QString::fromUtf8(":"));     // Get the YY:DDD:HH:MM:SS.microsec of the time.

    // Make sure we got everything.
    if (max_list.size() != 5)
    {
        qDebug() << __FUNCTION__ << "split failed, only produced" << max_list;
        return;
    }

    // Show the year and day fields.
    ui->frm_minute->setEnabled(min_list.at(3) != max_list.at(3));
    ui->frm_hour->setEnabled(min_list.at(2) != max_list.at(2));
    ui->frm_yday->setEnabled(min_list.at(1) != max_list.at(1));
    ui->frm_year->setEnabled(min_list.at(0) != max_list.at(0));

    // Now restore the frame number.
    cur_frame = -1;
    ui->sld_frames->setValue(frame_num);
    frame_changed(frame_num);
}

/*!
 * @brief The user has modified the slider, selecting a new frame value.
 *        We need to update the time which represents that frame value.
 */
void JumpDialog::frame_changed(int new_frame)
{
    // Make sure this is a REAL change and not a reflection from
    // an edit control change.
    if (new_frame != cur_frame)
    {
        // Remember the current frame value.
        cur_frame = new_frame;

        // Update our frame number display box.
        ui->frame_num->setNum(new_frame - trigger_frame);

        // Update our display.
        update_display();
    }
}

/*!
 * @brief The user has modified the seconds value. We need to update
 *        the slider to show the frame which is represented by the new time.
 */
void JumpDialog::seconds_changed(double /* new_value */)
{
    // Handle the time change.
    time_changed();
}

/*!
 * @brief The user has modified the minutes value. We need to update
 *        the slider to show the frame which is represented by the new time.
 */
void JumpDialog::minutes_changed(int /* new_value */)
{
    // Handle the time change.
    time_changed();
}

/*!
 * @brief The user has modified the hours value. We need to update
 *        the slider to show the frame which is represented by the new time.
 */
void JumpDialog::hours_changed(int /* new_value */)
{
    // Handle the time change.
    time_changed();
}

/*!
 * @brief The user has modified the days value. We need to update
 *        the slider to show the frame which is represented by the new time.
 */
void JumpDialog::days_changed(int /* new_value */)
{
    // Handle the time change.
    time_changed();
}

/*!
 * @brief The user has modified the years value. We need to update
 *        the slider to show the frame which is represented by the new time.
 */
void JumpDialog::years_changed(int /* new_value */)
{
    // Handle the time change.
    time_changed();
}

/*!
 * @brief Handle the changing of the position combobox.
 *
 * @param new_value     The newly selected value.
 */
void JumpDialog::position_changed(int new_value)
{
    switch (new_value)
    {
    case JUMP_DLG::CB_PRE_TRIGGER:
    case JUMP_DLG::CB_POST_TRIGGER:
        // Update the time display.
        time_changed();
        break;
    default:
        qDebug() << __PRETTY_FUNCTION__ << "Unexpected value =" << new_value;
        break;
    }
}

/*!
 * @brief Convenience method to call blockSignals() on all the spinbox controls.
 */
void JumpDialog::block_controls(bool do_block)
{
    ui->sb_seconds->blockSignals(do_block);
    ui->sb_minute->blockSignals(do_block);
    ui->sb_hour->blockSignals(do_block);
    ui->sb_yday->blockSignals(do_block);
    ui->sb_year->blockSignals(do_block);
    ui->cb_relative_pos->blockSignals(do_block);
}
