//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "logger.h"
#include "user_preferences.h"

#include <QDateTime>
#include <QDir>
#include <QSettings>

namespace FI_LOGGER
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

static const char *DEBUG_FILE_BASE = "fmdebug_";

//! The maximum number of log files we want to exist on the system.
const int   MAX_LOG_FILE_COUNT  = 10;

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

QsLogging::DestinationPtr fileDestination;
QsLogging::DestinationPtr debugDestination;

#ifdef Q_OS_LINUX
QsLogging::DestinationPtr syslogDestination;
#endif /* Q_OS_LINUX */

}   // End of namespace FI_LOGGER.

using namespace FI_LOGGER;

//--------------------------------------------------------------------
// Global Variable Declarations
//--------------------------------------------------------------------

// Get our logging instance.
const QsLogging::Logger& logger = QsLogging::Logger::instance();

/*!
 * @brief Implement a simple log rotation scheme, only allowing a certain
 *        number of log files on the system.
 *
 * @param log_path  The path to the folder holding the log files.
 */
static void log_rotate(QString log_path)
{
    int         i;
    int         max_files;
    QDir        log_dir(log_path);
    QString     log_filter;
    QStringList filter_list;
    QStringList     file_list;

    // Build up our log filter.
    log_filter = DEBUG_FILE_BASE + QString::fromUtf8("*.txt");
    filter_list << log_filter;

    // Set the filter(s) into our directory object.
    file_list = log_dir.entryList(filter_list, QDir::Files, QDir::Time);

    // What is the maximum number of log files we want on the system?
    max_files = log_get_max_files();

    // Remove any old log file. We allow MAX_LOG_FILE_COUNT log files, so
    // start there to remove old ones.
    for (i = max_files - 1; i < file_list.count(); ++i)
        log_dir.remove(file_list.at(i));
}

/*!
 * @brief Initialize the logging object.
 * @param log_dest  The path to the folder holding the log files.
 */
void log_init(const char *log_dest)
{
    QDir                debug_dir;

    // Set the default logging level.
    logger.setLoggingLevel(QsLogging::TraceLevel);

    // Are we logging to a file?
    if (log_dest != nullptr)
    {
        // Build up the logging path.
        QString log_path = QString::fromUtf8(log_dest);
        if (!log_path.endsWith(QLatin1Char('/')))
            log_path.append(QLatin1Char('/'));

        if (!debug_dir.exists(log_path))
            debug_dir.mkdir(log_path);
        else
            log_rotate(log_path);

        // Now that we have the folder, add in the file name info.
        QString dt_format = QString::fromUtf8("yyyy_MMdd_hhmmss");
        log_path = log_path + QString::fromUtf8(DEBUG_FILE_BASE) + QDateTime::currentDateTime().toString(dt_format) + QString::fromUtf8(".txt");

        // Build up our destination information.
        fileDestination = QsLogging::DestinationFactory::MakeFileDestination(log_path);
        logger.addDestination(fileDestination.get());
    }

#ifdef Q_OS_LINUX
    syslogDestination = QsLogging::DestinationFactory::MakeSyslogDestination();
    logger.addDestination(syslogDestination.get());
#endif /* Q_OS_LINUX */
#ifdef DEBUG
    // Always go to our Debug Output Destination.
    debugDestination = QsLogging::DestinationFactory::MakeDebugOutputDestination();
    logger.addDestination(debugDestination.get());
#endif
}

/*!
 * @brief Allow the application to change the verbosity of the log.
 * @param qs_log_level  The log level to use.
 */
void log_set_level(QsLogging::Level qs_log_level)
{
    // Set the logging level based on user request.
    logger.setLoggingLevel(qs_log_level);
}

/*!
 * @brief Allow the application to specify the maximum number of log files supported.
 * @param max_log_files The maximum number of log files to keep around.
 */
void log_set_max_files(int max_log_files)
{
    // Save the current settings.
    QSettings settings;

    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    settings.setValue("max_log_file_count", max_log_files);
    settings.endGroup();
}

/*!
 * @brief Query the max log file setting.
 * @returns The maximum setting our the constant definition.
 */
int log_get_max_files()
{
    int         max_files = MAX_LOG_FILE_COUNT;
    QSettings   settings;

    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    max_files = settings.value("max_log_file_count", MAX_LOG_FILE_COUNT).toInt();
    settings.endGroup();

    return max_files;
}
