// Copyright (c) 2010, Razvan Petru
// All rights reserved.

// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:

// * Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice, this
//   list of conditions and the following disclaimer in the documentation and/or other
//   materials provided with the distribution.
// * The name of the contributors may not be used to endorse or promote products
//   derived from this software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.

#include "QsLog.h"
#include "QsLogDest.h"

#include <QMutex>
#include <QList>
#include <QDateTime>
#include <QSharedPointer>
#include <QtGlobal>

#include <cstdlib>
#include <stdexcept>

namespace QsLogging
{
typedef QList<Destination *> DestinationList;

const char TraceString[] = "TRACE";
const char DebugString[] = "DEBUG";
const char InfoString[]  = "INFO";
const char WarnString[]  = "WARN";
const char ErrorString[] = "ERROR";
const char FatalString[] = "FATAL";

// not using Qt::ISODate because we need the milliseconds too
const QString fmtDateTime = QStringLiteral("yyyy-MM-ddThh:mm:ss.zzz");

// Our one and only.
QSharedPointer<Logger> Logger::staticLog;

static const char *LevelToText(Level theLevel)
{
    // ALM: Removed the default case (causing warnings):
    // Killian has explained already why clang emits the warning: if you extended the enum,
    // you'd fall into the default case which probably isn't what you want. The correct thing
    // to do is to remove the default case and get warnings for unhandled conditions.
    switch (theLevel)
    {
    case TraceLevel:
        return TraceString;
    case DebugLevel:
        return DebugString;
    case InfoLevel:
        return InfoString;
    case WarnLevel:
        return WarnString;
    case ErrorLevel:
        return ErrorString;
    case FatalLevel:
        return FatalString;
    }

    return InfoString;
}

class LoggerImpl
{
public:
    LoggerImpl() :
        level(InfoLevel)
    {
    }

    ~LoggerImpl()
    {
        destList.clear();
    }

    QMutex logMutex;
    DestinationList destList;
    Level level;
};

Logger::Logger() :
    d(new LoggerImpl),
    active(false)
{
}

Logger::~Logger()
{
    // No longer active.
    active = false;

    // Release the impelementation object.
    delete d;
    d = nullptr;
}

void Logger::close()
{
    Logger::staticLog.clear();
}

const Logger& Logger::instance()
{
    // There is only one logging object.
    if (Logger::staticLog.isNull())
        Logger::staticLog = QSharedPointer<Logger>(new Logger);

    return *Logger::staticLog;
}

void Logger::addDestination(Destination *destination) const
{
    if (destination == nullptr)
    {
        qWarning() << __PRETTY_FUNCTION__ << "Null destination!";
        return;
    }
    if (Logger::staticLog.isNull())
    {
        qWarning() << __PRETTY_FUNCTION__ << "Null staticLog?";
        return;
    }

    if (d != nullptr)
    {
        d->destList.push_back(destination);
        Logger::staticLog->active = true;
    }
}

void Logger::setLoggingLevel(Level newLevel) const
{
    if (active && (d != nullptr))
        d->level = newLevel;
}

Level Logger::loggingLevel() const
{
    if (active && (d != nullptr))
    {
        return d->level;
    }
    return TraceLevel;
}

//! creates the complete log message and passes it to the logger
void Logger::Helper::writeToLog()
{
    const char *const levelName = LevelToText(level);
    const QString completeMessage(QString("%1 %2 %3")
                                  .arg(levelName, 5)
                                  .arg(QDateTime::currentDateTime().toString(fmtDateTime), buffer)
                                 );

    const Logger& logger = Logger::instance();

    if (!logger.active)
    {
        qDebug() << completeMessage;
        return;
    }
    if (logger.d != nullptr)
    {
        QMutexLocker lock(&logger.d->logMutex);
        logger.write(level, completeMessage);
    }
}

Logger::Helper::~Helper()
{
    try
    {
        if (Logger::instance().active)
            writeToLog();
    }
    catch (std::exception& e)
    {
        // you shouldn't throw exceptions from a sink
        qWarning() << __PRETTY_FUNCTION__ << "Exception:" << e.what();
    }
}

//! sends the message to all the destinations
void Logger::write(int level, const QString& message) const
{
    if (!active)
    {
        qDebug() << message;
        return;
    }
    if (d == nullptr)
        return;

    for (DestinationList::iterator it = d->destList.begin(),
         endIt = d->destList.end(); it != endIt; ++it)
    {
        if (!(*it))
        {
            qDebug() << __PRETTY_FUNCTION__ << "Null log destination!";
            continue;
        }
        (*it)->write(level, message);
    }
}

} // end namespace
