// Copyright (c) 2010, Razvan Petru
// All rights reserved.

// Redistribution and use in source and binary forms, with or without modification,
// are permitted provided that the following conditions are met:

// * Redistributions of source code must retain the above copyright notice, this
//   list of conditions and the following disclaimer.
// * Redistributions in binary form must reproduce the above copyright notice, this
//   list of conditions and the following disclaimer in the documentation and/or other
//   materials provided with the distribution.
// * The name of the contributors may not be used to endorse or promote products
//   derived from this software without specific prior written permission.

// THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
// ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
// WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
// IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT,
// INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
// BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
// DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF
// LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
// OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED
// OF THE POSSIBILITY OF SUCH DAMAGE.

#include <QFile>
#include <QTextStream>
#include <QString>

#ifdef Q_OS_LINUX
#include <syslog.h>
#endif /* Q_OS_LINUX */

#include "QsLog.h"
#include "QsLogDest.h"
#include "QsDebugOutput.h"

namespace QsLogging
{

//! file message sink
class FileDestination : public Destination
{
public:
    explicit FileDestination(const QString& filePath);
    virtual void write(int level, const QString& message) override;

private:
    QFile mFile;
    QTextStream mOutputStream;
};


FileDestination::FileDestination(const QString& filePath)
{
    mFile.setFileName(filePath);
    mFile.open(QFile::WriteOnly | QFile::Text); //fixme: should throw on failure
    mOutputStream.setDevice(&mFile);
}

void FileDestination::write(int level, const QString& message)
{
    Q_UNUSED(level)

    mOutputStream << message << Qt::endl;
    mOutputStream.flush();
}

//! debugger sink
class DebugOutputDestination : public Destination
{
public:
    virtual void write(int level, const QString& message) override;
};

void DebugOutputDestination::write(int level, const QString& message)
{
    Q_UNUSED(level)
    QsDebugOutput::output(message);
}

#ifdef Q_OS_LINUX
//! syslog sink
class SyslogDestination : public Destination
{
public:
    explicit SyslogDestination(const QString& identifier);
    virtual void write(int level, const QString& message) override;

private:
    virtual int levelToLog(int level);
};

SyslogDestination::SyslogDestination(const QString& identifier)
{
    openlog((identifier.isEmpty() ? NULL : identifier.toUtf8().constData()), LOG_PID | LOG_NDELAY, LOG_USER);
}

int SyslogDestination::levelToLog(int level)
{
    int log_value;

    switch (level)
    {
    default:
    case DebugLevel:
    case TraceLevel:
        log_value = LOG_DEBUG;
        break;
    case InfoLevel:
        log_value = LOG_INFO;
        break;
    case WarnLevel:
        log_value = LOG_WARNING;
        break;
    case ErrorLevel:
        log_value = LOG_ERR;
        break;
    case FatalLevel:
        log_value = LOG_CRIT;
        break;
    }

    return log_value;
}

void SyslogDestination::write(int level, const QString& message)
{
    syslog(levelToLog(level), "%s", message.toUtf8().constData());
}
#endif /* Q_OS_LINUX */

DestinationPtr DestinationFactory::MakeFileDestination(const QString& filePath)
{
    return DestinationPtr(new FileDestination(filePath));
}

DestinationPtr DestinationFactory::MakeDebugOutputDestination()
{
    return DestinationPtr(new DebugOutputDestination);
}

#ifdef Q_OS_LINUX
DestinationPtr DestinationFactory::MakeSyslogDestination(const QString& identifier)
{
    return DestinationPtr(new SyslogDestination(identifier));
}
#endif /* Q_OS_LINUX */

} // end namespace
