/*!
 * @file mac_ctrl.cpp
 * @brief This file contains the implementation of the MacCtrl
 *        widget, a widget to make the user display/input of
 *        a camera's MAC address easier/better.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QRegularExpression>
#include <QRegularExpressionValidator>
#include "mac_ctrl.h"
#include "logger.h"

/*!
 * @brief The MacCtrl class constructor.
 * @param parent    Pointer to the parent widget.
 */
MacCtrl::MacCtrl(QWidget *parent) :
    QFrame(parent)
{
    // Setup our frame style.
    setFrameShape(QFrame::StyledPanel);
    setFrameShadow(QFrame::Sunken);

    // We need a horizontal layout for the four edit controls.
    QHBoxLayout *pLayout = new QHBoxLayout(this);
    setLayout(pLayout);
    pLayout->setContentsMargins(0, 0, 0, 0);
    pLayout->setSpacing(0);

    // Loop through each of the edit controls.
    for (int i = 0; i != QTUTL_MAC_SIZE; ++i)
    {
        // We need a separator between each of the controls.
        if (i != 0)
        {
            QLabel *pColon = new QLabel(":", this);
            pColon->setStyleSheet("background: white");
            pLayout->addWidget(pColon);
            pLayout->setStretch(pLayout->count(), 0);
        }

        // Create our edit control.
        m_pLineEdit[i] = new QLineEdit(this);
        QLineEdit *pEdit = m_pLineEdit[i];

        // Install an event filter so we can catch key strokes and such.
        pEdit->installEventFilter(this);

        // Add our edit control to our layout widget.
        pLayout->addWidget(pEdit);
        pLayout->setStretch(pLayout->count(), 1);

        // No frames, align the value in the center of the edit control.
        pEdit->setFrame(false);
        pEdit->setAlignment(Qt::AlignCenter);

        // Make sure we use a fixed pitch, mono spaced font.
        QFont font = pEdit->font();
        font.setStyleHint(QFont::Monospace);
        font.setFixedPitch(true);
        pEdit->setFont(font);

        // Setup a regex to validate the MAC address, so the user can't type in
        // an invalid value.
        QRegularExpression rx("^([0-9A-Fa-f]{2})$");
        QRegularExpressionValidator *validator = new QRegularExpressionValidator(rx, pEdit);
        pEdit->setValidator(validator);
    }

    // Setup our maximum width.
    setMaximumWidth(30 * QTUTL_MAC_SIZE);

    // Connect our signal/slots.
    connect(this, SIGNAL(signalTextChanged(QLineEdit *)), this, SLOT(slotTextChanged(QLineEdit *)), Qt::QueuedConnection);
}

/*!
 * @brief The MacCtrl class destructor.
 */
MacCtrl::~MacCtrl()
{
}

/*!
 * @brief The contents of one of the edit controls has changed.
 * @param pEdit     Pointer to the offending edit control.
 */
void MacCtrl::slotTextChanged(QLineEdit *pEdit)
{
    // Loop through each of our edit controls.
    for (unsigned int i = 0; i != QTUTL_MAC_SIZE; ++i)
    {
        // Is this the one issuing the alert?
        if (pEdit == m_pLineEdit[i])
        {
            // Figure out how to auto move.
            if ((pEdit->text().size() == MAX_DIGITS) && (pEdit->text().size() == pEdit->cursorPosition()))
            {
                // Auto-move to next item
                if ((i + 1) != QTUTL_MAC_SIZE)
                {
                    m_pLineEdit[i + 1]->setFocus();
                    m_pLineEdit[i + 1]->selectAll();
                }
            }
        }
    }
}

/*!
 * @brief Our event filter, where we can process key strokes so we can properly move
 *        the cursor around.
 *
 * @param obj       The object issuing the event.
 * @param event     The event being issued.
 *
 * @return True to stop further processing of the event, false to allow the event to propogate.
 */
bool MacCtrl::eventFilter(QObject *obj, QEvent *event)
{
    // Pass the event up to our parent class.
    bool bRes = QFrame::eventFilter(obj, event);

    // Are we looking at a key press event?
    if (event->type() == QEvent::KeyPress)
    {
        QKeyEvent *pEvent = dynamic_cast<QKeyEvent *>(event);

        // Make sure our cast worked.
        if (pEvent)
        {
            // Loop through each of our edit controls.
            for (int i = 0; i != QTUTL_MAC_SIZE; ++i)
            {
                QLineEdit *pEdit = m_pLineEdit[i];

                // Is this the control issuing the event?
                if (pEdit == obj)
                {
                    // What key was pressed?
                    switch (pEvent->key())
                    {
                    case Qt::Key_Left:
                        if (pEdit->cursorPosition() == 0)
                        {
                            // user wants to move to previous item
                            MovePrevLineEdit(i);
                        }
                        break;

                    case Qt::Key_Right:
                        if (pEdit->text().isEmpty() || (pEdit->text().size() == pEdit->cursorPosition()))
                        {
                            // user wants to move to next item
                            MoveNextLineEdit(i);
                        }
                        break;

                    case Qt::Key_Backspace:
                        if (pEdit->text().isEmpty() || (pEdit->cursorPosition() == 0))
                        {
                            // user wants to move to previous item
                            MovePrevLineEdit(i);
                        }
                        break;

                    case Qt::Key_Colon:
                    case Qt::Key_hyphen:
                    case Qt::Key_Minus:
                    case Qt::Key_Period:
                        if (pEdit->text().isEmpty())
                        {
                            pEdit->setText("00");
                            emit signalTextChanged(pEdit);
                        }
                        MoveNextLineEdit(i);
                        break;

                    default:
                        emit signalTextChanged(pEdit);
                        break;

                    }
                }
            }
        }
    }

    return bRes;
}

/*!
 * @brief Return a pointer to the MAC address buffer.
 */
const uint8_t *MacCtrl::get_mac(void)
{
    for (int i = 0; i != QTUTL_MAC_SIZE; ++i)
    {
        bool        ok;
        QLineEdit   *pEdit = m_pLineEdit[i];

        mac_addr[i] = (uint8_t)pEdit->text().toUInt(&ok, 16);
        if (!ok)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Invalid MAC address octet #" << i;
            return nullptr;
        }
    }

    return mac_addr;
}

/*!
 * @brief Set the MAC address.
 */
void MacCtrl::set_mac(const uint8_t *user_mac)
{
    memcpy(mac_addr, user_mac, QTUTL_MAC_SIZE);

    for (unsigned int i = 0; i != QTUTL_MAC_SIZE; ++i)
    {
        QLineEdit *pEdit = m_pLineEdit[i];
        pEdit->setText(QString::fromUtf8("%1").arg(mac_addr[i], MAX_DIGITS, 16, QLatin1Char('0')));
    }
}

/*!
 * @brief Move to the "next" edit control after the i-th one.
 * @param i     The "current" edit control.
 */
void MacCtrl::MoveNextLineEdit(int i)
{
    if ((i + 1) != QTUTL_MAC_SIZE)
    {
        m_pLineEdit[i + 1]->setFocus();
        m_pLineEdit[i + 1]->setCursorPosition(0);
        m_pLineEdit[i + 1]->selectAll();
    }
}

/*!
 * @brief Move to the "previous" edit control before the i-th one.
 * @param i     The "current" edit control.
 */
void MacCtrl::MovePrevLineEdit(int i)
{
    if (i != 0)
    {
        m_pLineEdit[i - 1]->setFocus();
        m_pLineEdit[i - 1]->setCursorPosition(m_pLineEdit[i - 1]->text().size());
        //m_pLineEdit[i-1]->selectAll();
    }
}
