/*!
 * @file main.cpp
 * @brief This is the main entry point for the FasMotion application.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "mainwindow.h"
#include <QApplication>
#include <QGuiApplication>
#include <QScopedPointer>
#include <QScreen>
#include <QtSingleApplication>

#include <stdint.h>

#include "fi_mem_pool.h"
#include "logger.h"
#include "user_preferences.h"

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

static const int   DEFAULT_NET_BUFFERS     = 20;       //!< How many frames do we want to buffer in case the save thread gets behind?
static const int   MIN_NET_BUFFERS         = 10;       //!< The fewest net buffers allowed.
static const uint32_t  DEFAULT_LOG_LEVEL   = QsLogging::DebugLevel;     //!< Default to logging warnings and above.

// These are the command line options. Examples of use are:
//      -f -n=10 -v=10
//      -f -n 10 -v 10
//      --factory --net=10 --video=10
static const char *FACTORY_LONG_OPT            = "--factory";   // Display the factory menu
static const char *FACTORY_SHORT_OPT           = "-f";
static const char *NET_BUFFERS_LONG_OPT        = "--net";   // how many network buffers should we use
static const char *NET_BUFFERS_SHORT_OPT       = "-n";
static const char *LOG_LEVEL_LONG_OPT          = "--log";   // what's our log level
static const char *LOG_LEVEL_SHORT_OPT         = "-l";
static const char *STATS_LONG_OPT              = "--stats"; // should a stat's file be generated.
static const char *STATS_SHORT_OPT             = "-s";
static const char *BACKOFF_LONG_OPT            = "--backoff";   // how many empty buffers left before we start trying to pause the fpga
static const char *BACKOFF_SHORT_OPT           = "-b";
static const char *MULTI_LONG_OPT              = "--multi"; // Allow multiple instances to run.
static const char *MULTI_SHORT_OPT             = "-m";
static const char *LOCAL_LONG_OPT              = "--local"; // Use local control.
static const char *LOCAL_SHORT_OPT             = "-L";

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

bool g_local_control;
bool maintain_net_stats;
int  num_net_buffers;
int  num_backoff_buffers;

static bool allow_multi_inst;
static bool show_factory_menu;
static int  log_level;

// We need to define a derived QApplication so we can override default
// widget sizes on the MAC. I got this information from:
// http://www.mr-edd.co.uk/blog/automatic_wa_macsmallsize
class NativeLookingApp : public QtSingleApplication
{
public:
    NativeLookingApp(int& argc, char **argv) :
        QtSingleApplication(argc, argv)
    {
    }

    ~NativeLookingApp()
    {
    }

    bool notify(QObject *receiver, QEvent *event) override
    {
        if (((event->type() == QEvent::Polish) || (event->type() == QEvent::PolishRequest)) &&
            receiver && receiver->isWidgetType())
        {
            set_smaller_text_osx(static_cast<QWidget *>(receiver));
        }

        return QtSingleApplication::notify(receiver, event);
    }

    // It's safe to call this function on any platform.
    // It will only have an effect on the Mac.
    void set_smaller_text_osx(QWidget *w)
    {
        if (w  == nullptr)
            return;

        // By default, none of these size attributes are set.
        // If any has been set explicitly, we'll leave the widget alone.
        if (!w->testAttribute(Qt::WA_MacMiniSize) &&
            !w->testAttribute(Qt::WA_MacSmallSize) &&
            !w->testAttribute(Qt::WA_MacNormalSize))
        {
            // Is the widget is one of a number of types whose default
            // text size is too large?
            if (w->inherits("QLabel") ||
                w->inherits("QLineEdit") ||
                w->inherits("QComboBox") ||
                w->inherits("QCheckBox") ||
                w->inherits("QRadioButton") ||
                w->inherits("QFrame") ||
                w->inherits("QSpinBox") ||
                w->inherits("QAbstractItemView"))
                // Others could be added here...
            {
                // make the text the 'normal' size
                w->setAttribute(Qt::WA_MacSmallSize);
            }
        }
    }
};

/*!
 * @brief Parse the (optional) command line arguments. We also make sure
 *        the parameters are initialized to their default settings.
 */
void parse_args(void)
{
    int     i;
    int     index;
    QString s_tmp;

    // Initialize the command line parameters to their default values.
    QSettings settings;

    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    show_factory_menu = settings.value("factory_enable", false).toBool();
    maintain_net_stats = settings.value("save_stats", false).toBool();
    num_net_buffers = settings.value("network_bufs", DEFAULT_NET_BUFFERS).toInt();
    num_backoff_buffers =  settings.value("backoff_bufs", DEFAULT_NET_BUFFERS / 4).toInt();
    log_level = settings.value("log_level", DEFAULT_LOG_LEVEL).toInt();
    settings.endGroup();

    allow_multi_inst = false;
    g_local_control = false;

    // Grab the command line arguments.
    QStringList cmd_args = QApplication::arguments();

    // Loop through each of the options.
    for (i = 0; i < cmd_args.count(); ++i)
    {
        if ((cmd_args[i].compare(QString::fromUtf8(FACTORY_LONG_OPT), Qt::CaseInsensitive) == 0) ||
            (cmd_args[i].compare(QString::fromUtf8(FACTORY_SHORT_OPT), Qt::CaseInsensitive) == 0))
        {
            show_factory_menu = true;
        }
        else if ((cmd_args[i].compare(QString::fromUtf8(MULTI_LONG_OPT), Qt::CaseInsensitive) == 0) ||
                 (cmd_args[i].compare(QString::fromUtf8(MULTI_SHORT_OPT), Qt::CaseInsensitive) == 0))
        {
            allow_multi_inst = true;
        }
        else if ((cmd_args[i].compare(QString::fromUtf8(STATS_LONG_OPT), Qt::CaseInsensitive) == 0) ||
                 (cmd_args[i].compare(QString::fromUtf8(STATS_SHORT_OPT), Qt::CaseInsensitive) == 0))
        {
            maintain_net_stats = true;
        }
        else if ((cmd_args[i].startsWith(QString::fromUtf8(NET_BUFFERS_LONG_OPT), Qt::CaseInsensitive)) ||
                 (cmd_args[i].startsWith(QString::fromUtf8(NET_BUFFERS_SHORT_OPT), Qt::CaseInsensitive)))
        {
            // The number of net buffers should be either in this parameter seperated with an
            // equal sign or the next parameter.
            index = cmd_args[i].indexOf(QString::fromUtf8("="));
            if (index >= 0)
            {
                s_tmp = cmd_args[i].mid(index + 1);
                num_net_buffers = s_tmp.toInt();
            }
            else
            {
                num_net_buffers = cmd_args[ i + 1 ].toInt();
                if (num_net_buffers > 0)
                    i++;            // Got a good value so skip over the next parameter.
            }

            // Sanity check the number of buffers.
            if (num_net_buffers < MIN_NET_BUFFERS)
                num_net_buffers = MIN_NET_BUFFERS;
        }
        else if ((cmd_args[i].startsWith(QString::fromUtf8(BACKOFF_LONG_OPT), Qt::CaseInsensitive)) ||
                 (cmd_args[i].startsWith(QString::fromUtf8(BACKOFF_SHORT_OPT), Qt::CaseInsensitive)))
        {
            // The number of empty buffers before we start trying to pause the FPGA. This
            // needs to be a pretty big number because when the firehose is spewing the processor
            // gets backed up and it takes awhile until we can get the message out. The number can
            // either be the next parameter or after the equal sign.
            index = cmd_args[i].indexOf(QString::fromUtf8("="));
            if (index >= 0)
            {
                s_tmp = cmd_args[i].mid(index + 1);
                num_backoff_buffers = s_tmp.toInt();
            }
            else
            {
                num_backoff_buffers = cmd_args[ i + 1 ].toInt();
                if (num_backoff_buffers > 0)
                {
                    i++;            // Got a good value so skip over the next parameter.
                }
            }
        }
        else if ((cmd_args[i].startsWith(QString::fromUtf8(LOG_LEVEL_LONG_OPT), Qt::CaseInsensitive)) ||
                 (cmd_args[i].startsWith(QString::fromUtf8(LOG_LEVEL_SHORT_OPT), Qt::CaseSensitive)))
        {
            // the number of video buffers should be either in this parameter seperated with an
            // equal sign or the next parameter
            index = cmd_args[i].indexOf(QString::fromUtf8("="));
            if (index >= 0)
            {
                s_tmp = cmd_args[i].mid(index + 1);
                log_level = s_tmp.toInt();
            }
            else
            {
                log_level = cmd_args[ i + 1 ].toInt();
                if (log_level >= 0)
                    i++;            // Got a good value so skip over the next parameter.
            }
            if (log_level < 0)
                log_level = DEFAULT_LOG_LEVEL;
        }
        else if ((cmd_args[i].startsWith(QString::fromUtf8(LOCAL_LONG_OPT), Qt::CaseInsensitive)) ||
                 (cmd_args[i].startsWith(QString::fromUtf8(LOCAL_SHORT_OPT), Qt::CaseSensitive)))
        {
            g_local_control = true;
        }
    }

    // Sanity check the backoff buffers.
    if (num_backoff_buffers > num_net_buffers)
        num_backoff_buffers = num_net_buffers / 2;
}

/*!
 * @brief This is the entry point for the FasMotion application.
 */
int main(int argc, char *argv[])
{
    QString log_path;
    int ret_status;

    // Create the application class.
    QScopedPointer<NativeLookingApp> my_app(new NativeLookingApp(argc, argv));

    // Initialize our settings values.
    QCoreApplication::setOrganizationName("Fastec");
    QCoreApplication::setOrganizationDomain("fastecimaging.com");
    QCoreApplication::setApplicationName("FasMotion");

    // Parse the command line arguments.
    // NOTE: The location of this call is VERY important. It must come AFTER you've
    //       created the app above. Otherwise, you won't be able to get to the arguments!
    parse_args();

    // Make sure we're the only ones running.
    if (my_app->sendMessage("Wake Up"))
    {
        if (allow_multi_inst)
        {
            qDebug() << "Another instance is already running, allowing multi-instance.";
        }
        else
        {
            qDebug() << "Exiting: Another instance is already running.";
            return 0;
        }
    }

    // Initialize our application logger.
    log_path = QDir::tempPath();
#ifdef QT_DEBUG
    // ------------ TEST ONLY ----------------------
    qDebug() << __PRETTY_FUNCTION__ << "Not using:" << log_path << "- instead:" << my_app->applicationDirPath();
    log_path = my_app->applicationDirPath();
    // ------------ TEST ONLY ----------------------
#endif
    log_init(log_path.toUtf8().constData());
    log_set_level((QsLogging::Level)log_level);

    // Initialize our fas_video log file.
    // NOTE:  Make sure this occurs after FasMotion's own log_init because the
    // call to InitializeLogger below expects the directory to exist it does not
    // make the directory.
    std::string fas_video_logpath = log_path.toUtf8().constData();
    if (!FAS_VIDEO::FiVideo::InitializeLogger(fas_video_logpath, log_get_max_files()))
    {
        // Unable to initialize the fas_video log file.
        // This is not terminal because the log data should still go to
        // other log files.
        QLOG_ERROR() << "FAS_VIDEO LOG: Unable to initialize logger.";
    }

    // Enable/disable local control of the camera.
    MainWindow::enable_local_control(g_local_control);

    // Create our new main window.
    if (MainWindow::GetInstance().isNull())
    {
        QLOG_ERROR() << "Exiting: Unable to allocate MainWindow class.";
        return -1;
    }

    // Set our activation window to our main window.
    my_app->setActivationWindow(MainWindow::GetInstance().get());
    QLOG_INFO() << "FasMotion starting..." << MainWindow::GetInstance()->version();

    // Enable/disable the factory menu.
    MainWindow::GetInstance()->enable_factory(show_factory_menu);

    // Have the main window initialize the communcations object.
    MainWindow::GetInstance()->init_comm(num_net_buffers, num_backoff_buffers, maintain_net_stats);

    // Display the window.
    MainWindow::GetInstance()->show();

    // Run the application.
    ret_status = my_app->exec();

    // All done.
    QLOG_TRACE() << "FasMotion finished with status" << ret_status;

    // Close down logging.
    QsLogging::Logger::instance().close();

    return ret_status;
}
