/*!
 * @file mainwindow.cpp
 * @brief This file contains the implementation of the MainWindow
 *        class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

//#define QT_SHAREDPOINTER_TRACK_POINTERS
#include "mainwindow.h"
#include "ui_mainwindow.h"

#include <QActionGroup>
#include <QAtomicInt>
#include <QCloseEvent>
#include <QDesktopServices>
#include <QtGlobal>
#include <QFile>
#include <QFileDialog>
#include <QFileInfo>
#include <QFileInfoList>
#include <QInputDialog>
#include <QLabel>
#include <QMdiSubWindow>
#include <QMessageBox>
#include <QProcess>
#include <QPushButton>
#include <QSettings>
#include <QStringLiteral>
#include <QStyleFactory>
#include <QTextStream>
#include <QVersionNumber>

#ifdef Q_OS_WIN
#include <windows.h>
#endif
#ifdef Q_OS_LINUX
#include <mntent.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/statvfs.h>
#endif
#ifdef Q_OS_MAC
#include <sys/types.h>
#include <sys/statvfs.h>

#include <CoreServices/CoreServices.h>
#include <CoreFoundation/CoreFoundation.h>
#include <CoreFoundation/CFURLEnumerator.h>

#include "qcore_mac_p.h"
#endif

#include "camera_dock.h"
#include "camera_info.h"
#include "camera_mdi.h"
#include "camera_review.h"
#include "camera_telemetry.h"
#include "control_panel.h"
#include "camera_logs.h"
#include "fi_comm.h"
#include "fi_internal.h"
#include "fi_msgbox.h"
#include "fi_mem_pool.h"
#include "fi_rawdrive.h"
#include "fi_sensor.h"
#include "fi_updater.h"
#include "fi_userdoc_dialog.h"
#include "force_ip.h"
#include "fcp_socket.h"
#include "histo_dock.h"
#include "logger.h"
#include "multicam_dock.h"
#include "peekpoke.h"
#include "playback_cap.h"
#include "playback_mdi.h"
#include "playback_meta_info.h"
#include "playback_raw_mdi.h"
#include "rel_version.h"
#include "save_dialog.h"
#include "select_drive.h"
#include "update_status.h"
#include "user_preferences.h"

namespace FASMOTION_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// At this point in time there is no implementation it is TBD.
//! Testing User Documentation Links.
const bool TEST_USER_DOCS = false;

//! Wait time (milliseconds) to start the second phase of shutdown.
const int   SHUTDOWN_DELAY      = 50;

//! This is a rough estimate of the overhead in a TIFF file.
const int   TIF_HEADER              = 512;

//! The rough estimate for the BMP overhead.
const int   BMP_HEADER              = 100;

//! Rough overhead for DNG & raw TIFF files.
const int   DNG_HEADER              = 500;

//! We will wait for up to 30 times for the camera server to load.
const int   WAIT_FOR_CS_TRIES       = 30;

//! The controller information file.
const QString CONTROLLER_INFO_FILE  = QStringLiteral("/usr/local/bin/fastec/controller.info");

#ifdef Q_OS_LINUX
//! The file where the Camera Server store its PID.
static const char   *CS_PID_FILE = "/var/run/CameraServer.pid";
#endif

#ifdef Q_OS_MAC
QString rootPath;
QByteArray fileSystemName;
static QString toQString(CFStringRef str)
{
    if (!str)
        return QString();

    CFIndex length = CFStringGetLength(str);
    if (length == 0)
        return QString();

    QString string(length, Qt::Uninitialized);
    CFStringGetCharacters(str, CFRangeMake(0, length), reinterpret_cast<UniChar *>(const_cast<QChar *>(string.unicode())));

    return string;
}

static bool getUrlProperties(bool initRootPath, QString& path_str)
{
    const void *rootPathKey[] = { kCFURLVolumeURLKey };
    const void *propertyKey[] = { kCFURLVolumeLocalizedFormatDescriptionKey };
    CFArrayRef keys = CFArrayCreate(kCFAllocatorDefault,
                                    initRootPath ? rootPathKey : propertyKey,
                                    1,
                                    0);
    if (!keys)
        return false;

    QCFString cfPath;
    if (initRootPath)
    {
        cfPath = path_str;
        rootPath.clear();
    }
    else
    {
        cfPath = rootPath;
    }

    CFURLRef url = CFURLCreateWithFileSystemPath(kCFAllocatorDefault,
                                                 cfPath,
                                                 kCFURLPOSIXPathStyle,
                                                 true);
    if (!url)
    {
        CFRelease(keys);
        return false;
    }

    CFErrorRef error;
    CFDictionaryRef map = CFURLCopyResourcePropertiesForKeys(url, keys, &error);
    CFRelease(url);
    CFRelease(keys);

    if (!map)
        return false;

    if (initRootPath)
    {
        CFURLRef rootUrl = (CFURLRef)CFDictionaryGetValue(map, kCFURLVolumeURLKey);
        if (!rootUrl)
            return false;

        rootPath = QCFString(CFURLCopyFileSystemPath(rootUrl, kCFURLPOSIXPathStyle));
        CFRelease(map);
    }
    else
    {
        fileSystemName = toQString((CFStringRef)CFDictionaryGetValue(map, kCFURLVolumeLocalizedFormatDescriptionKey)).toLatin1();
    }

    // All done.
    return true;
}
#endif /* Q_OS_MAC */

//! True if we're on the camera's compute platform.
static bool local_control_enable = false;

}   // End of namespace FASMOTION_SETTINGS.

using namespace FASMOTION_SETTINGS;

//! Keep a constant of when a sanity check on CameraMDI parent pointer fails.
const QString MainWindow::CAMERA_PTR_SANITY_FAILURE = QStringLiteral("[CHECK FAILED] The camera pointer was null!");

//! Keep a constant of when a sanity check on COMMs pointer fails.
const QString MainWindow::COMMS_PTR_SANITY_FAILURE = QStringLiteral("[CHECK FAILED] The communications pointer was null!");

//! To test the main window destruction process uncomment the one you want to test and comment the other one.
static const bool FORCE_DESTRUCTOR_CALL = true;
//static const bool FORCE_DESTRUCTOR_CALL = false;

//! Our singleton object.
QSharedPointer<MainWindow> MainWindow::main_window;

//! Have we created our main_window before?
bool MainWindow::allocated_main_window_once = false;

/*!
 * @brief The constructor for the MainWindow class.
 */
MainWindow::MainWindow(QWidget *parent) :
    QMainWindow(parent),
    ui(new Ui::main_wnd),
    update_btn(nullptr)
{
    int     i;

    // Make sure we got our UI pointer.
    if (ui == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No ui pointer!";
        return;
    }

    // Initialize the UI elements.
    ui->setupUi(this);

    // Setup our UI style(s).
    init_styles();

    // Initialize local variables.
    factory_enabled = false;
    keep_statistics = false;
    in_shutdown = false;
    multicam_enabled = false;
    translator_installed = false;
    selected_language = LANGUAGE_ENGLISH;
    sensor_correction_start_row = 0;

    histo_dock = nullptr;
    multicam_dock = nullptr;
    p_control_panel = nullptr;
    p_camera_dock = nullptr;
    p_camera_logs_dock = nullptr;
    p_metadata_dock = nullptr;
    p_telemetry_dock = nullptr;

    // Create our seperator action element.
    seperator_action = new QAction(this);
    seperator_action->setSeparator(true);

    // Create our language action group.
    language_group = new QActionGroup(this);
    language_group->addAction(ui->action_chinese_simplified);
    language_group->addAction(ui->action_chinese_traditional);
    language_group->addAction(ui->action_english);
    language_group->addAction(ui->action_french);
    language_group->addAction(ui->action_german);
    language_group->addAction(ui->action_japanese);
    language_group->addAction(ui->action_korean);
    language_group->addAction(ui->action_russian);
    language_group->addAction(ui->action_spanish);

    // Create our zoom action group.
    zoom_group = new QActionGroup(this);
    zoom_group->addAction(ui->action_zoom_1_to_1);
    zoom_group->addAction(ui->action_zoom_in);
    zoom_group->addAction(ui->action_zoom_out);
    zoom_group->addAction(ui->action_zoom_to_fit);
    zoom_group->addAction(ui->action_zoom_to_window);

    // Create our "recent videos" actions.
    for (i = 0; i < MAX_RECENT_VIDEOS; ++i)
    {
        // Allocate and initialize our action(s).
        recent_video_acts[i] = new QAction(this);
        if (recent_video_acts[i] == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Error allocating memory for recent_video:" << i;
            return;
        }
        recent_video_acts[i]->setVisible(false);
        connect(recent_video_acts[i], SIGNAL(triggered()), this, SLOT(open_recent_video()));

        // Add to our menu.
        ui->menu_recent_videos->addAction(recent_video_acts[i]);
    }

    // Create our recent video menu seperator.
    recent_video_sep = new QAction(this);
    recent_video_sep->setSeparator(true);
    recent_video_sep->setVisible(false);
    ui->menu_recent_videos->addAction(recent_video_sep);

    // Add the clear action.
    ui->action_clear_list->setVisible(false);
    ui->menu_recent_videos->addAction(ui->action_clear_list);

    // Default is to fit in the window, maintaining aspect ratio.
    ui->action_zoom_to_fit->setChecked(true);

    // Initially, disable the open raw drive menu.
    ui->action_open_raw_drive->setEnabled(false);
    ui->action_open_raw_drive->setVisible(true);

    // Initially, disable the open storage device menu.
    ui->action_open_device->setEnabled(false);
    ui->action_open_device->setVisible(true);

    // Initially, no comms.
    comm_ptr = nullptr;

    // Restore our settings.
    read_settings();

    // Setup our status bar.
    setup_statusbar();

    // Set up our dock-ables.
    setup_docks();

    // Initialize our signal/slot connections.
    init_connects();

    // Initially have the Factory menu be hidden.
    enable_factory(false);

    // Initialize our actions to reflect no connected camera.
    reflect_no_camera();

    // Initialize our updating object.
    init_updater();

    // Make sure we make things look OK on the Mac.
    setUnifiedTitleAndToolBarOnMac(true);

    // For testing purposes will control whether or not the destructor is called.
    if (FORCE_DESTRUCTOR_CALL)
    {
        // Add the ability to call the destructor on close of the application.
        this->setAttribute(Qt::WA_DeleteOnClose);
    }
}

/*!
 * @brief The Main Window is closing this is opportunity to cleanup resources.
 */
void MainWindow::close_main_window()
{
    // All logging when called from the QSharedPointer custom deleter call stack basically does not go
    // anywhere.  See all the notes in this function for all details.  This function is called by the
    // custom deleter MainWindow::do_delete_main_window
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FORCE_DESTRUCTOR_CALL:" << FORCE_DESTRUCTOR_CALL;

    // When forcing a destructor call and this code is executed from the custom
    // deleter given to main_window's QSharedPointer allocation.  It does not operate
    // with the same luxuries given to the MainWindow::~MainWindow destructor.  Therefore
    // there are issues with segmentation faults for things like setting up a QSettings
    // object or missing log messages when using QLOG_DEBUG() message.  These things break
    // when the "FORCE_DESTRUCTOR_CALL" is false.  See the comments below into specifics.
    // It is basically useless as it currently stands but I have left everything in place
    // to document it.
    if (!FORCE_DESTRUCTOR_CALL)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Custom close_main_window called but real destruction and cleanup did NOT take place!";
        // NOTE:  The reason the code below is commented out is that when it attempts to
        // execute for the deleter process of the main_window QSharedPointer the execution
        // segment faults as the object scope at this point cannot handle QSettings creation.
        // It was left here for testing purposes in case anyone wanted to see in a debug
        // session what happens.  In that case build/debug it with "#if 1".
#if 0
        QSettings settings;
        settings.beginGroup("Closing_Test");
        settings.setValue("custom_deletor", "close_main_window");
        settings.endGroup();
#endif
    }
}

/*!
 * @brief The destructor for the MainWindow class.
 */
MainWindow::~MainWindow()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FORCE_DESTRUCTOR_CALL:" << FORCE_DESTRUCTOR_CALL;

    // Save the current settings.
    QSettings settings;
    settings.beginGroup("Closing_Test");
    settings.setValue("destructor", "main_window_destructor");
    settings.endGroup();

    // The dock objects creation order was:
    //    p_control_panel
    //    p_camera_dock
    //    multicam_dock
    //    histo_dock
    //    p_camera_logs_dock
    //    p_metadata_dock

    // Clean up our dock widgets in the reverse order of creation like C++ does with object destruction.
    // NOTE:  Use "delete" construct rather than the commented out <obj>->deleteLater line.  The reason is that
    // we cannot be sure that a deleteLater call could attempt to access this main_window object and we are in the
    // process of destruction therefore we need to explicitely delete the object right here and right now!
    if (p_metadata_dock != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting p_metadata_dock";
        removeDockWidget(p_metadata_dock);
        //p_metadata_dock->deleteLater();
        delete p_metadata_dock;
        p_metadata_dock = nullptr;
    }
    if (p_camera_logs_dock != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting p_camera_logs_dock";
        removeDockWidget(p_camera_logs_dock);
        //p_camera_logs_dock->deleteLater();
        delete p_camera_logs_dock;
        p_camera_logs_dock = nullptr;
    }
    if (histo_dock != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting histo_dock";
        removeDockWidget(histo_dock);
        //histo_dock->deleteLater();
        delete histo_dock;
        histo_dock = nullptr;
    }
    if (multicam_dock != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting multicam_dock";
        removeDockWidget(multicam_dock);
        //multicam_dock->deleteLater();
        delete multicam_dock;
        multicam_dock = nullptr;
    }
    if (p_camera_dock != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting p_camera_dock";
        removeDockWidget(p_camera_dock);
        //p_camera_dock->deleteLater();
        delete p_camera_dock;
        p_camera_dock = nullptr;
    }
    if (p_control_panel != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting p_control_panel";
        removeDockWidget(p_control_panel);
        //p_control_panel->deleteLater();
        delete p_control_panel;
        p_control_panel = nullptr;
    }

    // Done with the updater.
    if (updater != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Releasing FiUpdater updater.";
        FiUpdater *up_ptr = updater.release();
        if (up_ptr != nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Deleting FiUpdater updater's up_ptr.";
            delete up_ptr;
        }
    }

    // Done with the communications object.
    if (comm_ptr != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting comm_ptr";
        delete comm_ptr;
        comm_ptr = nullptr;
    }

    // Done with the UI class.
    if (ui != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting ui";
        delete ui;
        ui = nullptr;
    }

    // Done with the separator actions.
    if (seperator_action != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Removing and deleting seperator_action";
        seperator_action->deleteLater();
        seperator_action = nullptr;
    }

    // Ticket #362 - Rearranging the columns in the "Found Cameras" window does not persist.
    // When destructing the main window some of our application objects will use the destruction process
    // to modify their QSettings.  When doing this sometimes the changes are not flushed out to permanent
    // settings on disk and are essentially "lost".  In order to make sure that any QSettings modifications
    // persist between application runtimes we will use the sync operation which ensures a persistence to
    // flush any changes to permanent storage for any unwritten modifications.
    settings.sync();
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Settings status:" << settings.status();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Done";
}

/*!
 * @brief Custom deleter for the main window singleton instance.
 */
void MainWindow::do_delete_main_window(MainWindow *obj)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Cleanup any resources we may need to cleanup.
    obj->close_main_window();

    // For testing purposes will control whether or not the destructor is called.
    if (!FORCE_DESTRUCTOR_CALL)
    {
        // If destructor was already called and this call goes forward then the QSharedPointer
        // ends up segment faulting during the deleteLater call stack.  Therefore we will ONLY
        // call the obj->deleteLater when we DID NOT FORCE a destructor call when main_window closed.
        // NOTE:  I also tried to just delete obj here and that does call the MainWindow::~MainWindow
        // however it hits the same problems we have in the "close_main_window" code where the
        // QSettings cannot be utilized in the QSharedPointer scope of deletion...
        obj->deleteLater();
    }
}

/*!
 * @brief Method to return the main window object.
 */
QSharedPointer<MainWindow>& MainWindow::GetInstance()
{
    // There can be only one...
    if (main_window.isNull())
    {
        // Make sure we have never allocated this instance before.  It could be possible
        // that something in the destruction process attempts to call this method when
        // things are shutting down.  If for some reason the main_window object is deleting
        // itself this could cause problems by re-creating a new singleton.  This should not
        // happen but this check will guarantee that it doesn't.
        if (!allocated_main_window_once)
        {
            allocated_main_window_once = true;
            //main_window = QSharedPointer<MainWindow>(new MainWindow, &QObject::deleteLater);
            main_window = QSharedPointer<MainWindow>(new MainWindow, do_delete_main_window);
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Someone is trying to get the main_window when it was already allocated and destructed!";
        }
    }
    // Return our singleton.
    return main_window;
}

/*!
 * @brief Translates a camera state into a user readable string for debugging purposes.
 * @return GIGE message in string readable form.
 */
QString MainWindow::camera_state_to_string(uint32_t camera_state)
{
    QString camera_state_str;
    switch (camera_state)
    {
    case GIGE_CAMERA_STATE_ARMED:
        camera_state_str = "GIGE_CAMERA_STATE_ARMED";
        break;
    case GIGE_CAMERA_STATE_DEARMING:
        camera_state_str = "GIGE_CAMERA_STATE_DEARMING";
        break;
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        camera_state_str = "GIGE_CAMERA_STATE_LIVE_VIDEO";
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        camera_state_str = "GIGE_CAMERA_STATE_REVIEW";
        break;
    case GIGE_CAMERA_STATE_IDLE:
        camera_state_str = "GIGE_CAMERA_STATE_IDLE";
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        camera_state_str = "GIGE_CAMERA_STATE_AUTOSAVE";
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
        camera_state_str = "GIGE_CAMERA_STATE_SNAPSHOT";
        break;
    case GIGE_CAMERA_STATE_SHUTDOWN:
        camera_state_str = "GIGE_CAMERA_STATE_SHUTDOWN";
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        camera_state_str = "GIGE_CAMERA_STATE_TRIGGERED";
        break;
    default:
        camera_state_str = QString("Untranslated: %1").arg(camera_state, 4, 16, QLatin1Char('0'));
        break;
    }
    // Return the camera state readable string.
    return camera_state_str;
}

/*!
 * @brief Check to see if the teamviewer daemon is running.
 * @return True if it is running, false otherwise.
 */
bool MainWindow::teamviewer_is_running()
{
    // Check to see if the "pid" file exists.
    return QFile::exists(QStringLiteral("/run/teamviewerd.pid"));
}

/*!
 * @brief Allows for the explicit deletion of the main window object.
 */
void MainWindow::terminate()
{
    MainWindow *main_ptr;
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[[[[[[[[[IT IS TIME to terminate the Main Window!!!]]]]]]]]]";

    main_ptr = main_window.get();
    main_window.clear();
    delete main_ptr;
}

/*!
 * @brief Return the currently active MDI child, nullptr if there are none.
 */
CameraMDI *MainWindow::active_mdi_child()
{
    CameraMDI       *mdi_child = nullptr;
    QMdiSubWindow   *activeSubWindow;

    // Get the active MDI child.
    activeSubWindow = ui->mdi_area->activeSubWindow();
    if (activeSubWindow != nullptr)
        mdi_child = qobject_cast<CameraMDI *>(activeSubWindow->widget());

    // Return the child pointer.
    return mdi_child;
}

/*!
 * @brief Change the network settings for the camera associated with "cam_ip_addr".
 *
 * @param cam_ip_addr   The camera's current IP address.
 * @param do_dhcp       True if we're enabling DHCP, false for a static IP address.
 * @param ip_addr       The new IP address.
 * @param subnet_mask   The new subnet mask.
 * @param gw_addr       The IP address of the new gateway.
 * @param mac_addr      The camera's MAC address.
 *
 * @return True on success, false otherwise.
 */
bool MainWindow::change_camera_network(uint32_t cam_ip_addr, bool do_dhcp, QHostAddress& ip_addr, QHostAddress& subnet_mask, QHostAddress& gw_addr, const uint8_t *mac_addr)
{
    QString msg_str;

    // Make sure the user knows what they're doing.
    msg_str = QString(tr("Changing network parameters may result in the loss of communications with the camera.\n\nAre you sure you want to continue?"));
    if (QMessageBox::question(this, tr("FasMotion"), msg_str, QMessageBox::Yes | QMessageBox::No) == QMessageBox::No)
        return false;

    // If we're enabling DHCP, then that's all we have to do.
    if (do_dhcp)
    {
        // Sanity check.
        if (cam_ip_addr == FCP_COMM::GLOBAL_BROADCAST_IPADDR)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Can't DHCP with a broadcast IP address!";
            return false;
        }

        // Tell the camera to switch to DHCP for it's address.
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_IP_CUR_CFG, qFromBigEndian(GIGE_IFCONFIG_DHCP_BE));
    }
    else
    {
        // We're going to a static IP address. Ask the communications object to send the address information to
        // the specified MAC address (our target camera).
        comm_ptr->start_force_ip(mac_addr, qToBigEndian(ip_addr.toIPv4Address()),
                                 qToBigEndian(subnet_mask.toIPv4Address()), qToBigEndian(gw_addr.toIPv4Address()));
    }

    // All done.
    return true;
}

/*!
 * @brief Change the currently active MDI child to "child_ptr".
 * @param child_ptr     Pointer to the MDI child we want to make active.
 */
void MainWindow::change_mdi_child(QWidget *child_ptr)
{
    // Walk through all the MDI children and de-activate them, except
    // for the one we want activated.
    foreach (QMdiSubWindow *window_ptr, ui->mdi_area->subWindowList())
    {
        // Is this the one?
        if (window_ptr->widget() == child_ptr)
        {
            set_active_subwindow(window_ptr);
        }
        else
        {
            CameraMDI   *cam_child = qobject_cast<CameraMDI *>(window_ptr->widget());
            if (cam_child == nullptr)
            {
                // Must be a playback object.
                PlaybackMDI *play_child = qobject_cast<PlaybackMDI *>(window_ptr->widget());
                if (play_child == nullptr)
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "MDI child is not a camera or playback object!";
                }
                else
                {
                    // Make the playback child window inactive.
                    play_child->setInactive();
                }
            }
            else
            {
                // Make this camera object inactive.
                cam_child->set_inactive();
            }
        }
    }
}

/*!
 * @brief Utility method to determine if the requested save operation is likely to fit. This isn't
 *        an "exact" check, because many times we don't REALLY know how big the result will be. But
 *        it will catch some situations.
 *
 * @param dest_path     The destination path we need to check.
 * @param num_frames    How many frames are we going to be saving.
 * @param streamType    The stream type used to save the images.
 * @param width         The image width.
 * @param height        The image height.
 * @param bit_depth     The number of bits per pixel (GIGE_BITMODE_xxx value).
 * @param sensor_is_mono    True if we have a mono sensor, false if color.
 * @param space_calc    CAP files must pass this information in.
 *
 * @return True if there looks to be sufficient space, false otherwise.
 */
bool MainWindow::check_avail_space(const QString& dest_path, uint32_t num_frames, FI_SAVER::STREAM_TYPE streamType,
                                   uint32_t width, uint32_t height, uint32_t bit_depth, bool sensor_is_mono,
                                   uint64_t *space_calc)
{
    uint32_t    size_per_frame = 0;
    uint64_t    space_needed = 0;
    uint64_t    space_avail;
    QString     target_path;
    QString     msg_str;

    // Make sure our path is in the proper format.
    target_path = QDir::toNativeSeparators(dest_path);

    // How much space is available at the target?
    space_avail = get_avail_space(target_path);
    if (space_avail == 0)
    {
        msg_str = tr("Unable to determine the amount of space available."
                     "<p>Check to make sure your destination still exists and "
                     "there is sufficient space on the drive.");
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "dest_path:" << dest_path << "Err:" << msg_str;
        QMessageBox::critical(this, tr("Save Error"), msg_str);
        return false;
    }

    // How much space do we need?
    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        if (sensor_is_mono)
            size_per_frame = (width * height) + BMP_HEADER;
        else
            size_per_frame = (width * height * 3) + BMP_HEADER;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        // For the "processed" TIFs, even 10-bit captured data is taken down to 8-bit data.
        // Only "raw" TIFs are promoted to 16-bits (for 10-bit captures).
        if (sensor_is_mono)
            size_per_frame = (width * height) + TIF_HEADER;
        else
            size_per_frame = (width * height * 3) + TIF_HEADER;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        // For anything other than 8bpp, the data is converted to 16bpp.
        if (IS_8_BIT_MODE(bit_depth))
            size_per_frame = (width * height) + DNG_HEADER;
        else
            size_per_frame = (width * height * 2) + DNG_HEADER;
        break;
    // TODO_JLC: Support FI_CAPSTK for FFLR
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        // For CAP files, the size required is well known - it's the size of the recording.
        // That SHOULD have been passed in the space_calc argument.
        if (space_calc == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The space_calc parameter is nullptr?! Can't be for CAP file!";
            return false;
        }

        // This tells us how much space is required.
        space_needed = *space_calc;
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        if (FiInternal::factory() != nullptr)
        {
            FiInternal::GET_PACKED_SIZE_FN funcPtr = FiInternal::factory()->getPackedSizeFuncPtr();
            if (funcPtr != nullptr)
            {
                size_per_frame = (*funcPtr)(width, height, (IS_8_BIT_MODE(bit_depth) ? 8 : (IS_10_BIT_MODE(bit_depth) ? 10 : 12)));
            }
        }
        break;
    default:
        // For the other formats we simply cannot pre-determine the required space.
        size_per_frame = 0;
        space_needed = 0;
        break;
    }

    // For other formats, calculate the space needed.
    if (streamType != FI_SAVER::STREAM_TYPE::FI_CAP)
    {
        // Now calculate the rough amount of space required for the total download.
        space_needed = (uint64_t)size_per_frame * (uint64_t)num_frames;

        // Does the user want the calculated space required returned?
        if (space_calc != nullptr)
            *space_calc = space_needed;
    }

    // Do we fit?
    if (space_needed > space_avail)
    {
        msg_str = tr("Not enough free space on device required:%1 free:%2").arg(space_needed).arg(space_avail);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        QMessageBox::critical(this, tr("Save Error"), msg_str);
        return false;
    }

    // All is well.
    return true;
}

/*!
 * @brief Update the UI controls to reflect the value of "zoom_action".
 * @param zoom_action   The new zoom action.
 */
void MainWindow::check_zoom_actions(int zoom_action)
{
    switch (zoom_action)
    {
    case FASMOTION_SETTINGS::ZOOM_UNKNOWN:
        ui->action_zoom_1_to_1->setChecked(false);
        ui->action_zoom_in->setChecked(false);
        ui->action_zoom_out->setChecked(false);
        ui->action_zoom_to_fit->setChecked(false);
        ui->action_zoom_to_window->setChecked(false);
        break;
    case FASMOTION_SETTINGS::ZOOM_1_TO_1:
        ui->action_zoom_1_to_1->setChecked(true);
        break;
    case FASMOTION_SETTINGS::ZOOM_IN:
        ui->action_zoom_in->setChecked(true);
        break;
    case FASMOTION_SETTINGS::ZOOM_OUT:
        ui->action_zoom_out->setChecked(true);
        break;
    case FASMOTION_SETTINGS::ZOOM_TO_WINDOW:
        ui->action_zoom_to_window->setChecked(true);
        break;
    default:
    case FASMOTION_SETTINGS::ZOOM_TO_FIT:
        ui->action_zoom_to_fit->setChecked(true);
        break;
    }
}

/*!
 * @brief Close the MDI child window associated with "child_ptr".
 * @param child_ptr     Pointer to the MDI child window we want to close.
 */
void MainWindow::close_mdi_child(CameraMDI *child_ptr)
{
    QMdiSubWindow   *child_win_ptr;
    static QAtomicInt block(0);

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << ((child_ptr != nullptr) ? child_ptr->get_ip_addr_pretty() : "");

    // Sanity check.
    if (child_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "NULL child_ptr.";
        return;
    }

    // Make sure we're not already here.
    if (!block.testAndSetRelaxed(0, 1))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Already here...bailing...";
        return;
    }

    // Make sure this isn't the broadcast address "child".
    if (child_ptr->get_ip_addr() != -1U)
    {
        FI_comm     *comm_ptr;

        // Tell anyone who cares, that we're closing this child.
        emit child_ptr->closing();

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
            return;
        }

        // Disconnect from the camera.
        comm_ptr->start_write_reg(child_ptr->get_ip_addr(), GIGE_REG_SCP0, 0);
        comm_ptr->start_write_reg(child_ptr->get_ip_addr(), GIGE_REG_CCP, 0);

        // Disconnect the camera.
        comm_ptr->disconnect_cam(child_ptr->get_ip_addr());
    }

    // Make sure the zoom actions are disabled.
    enable_zoom_actions(false);

    // Remove it from the MDI area.
    child_win_ptr = find_mdi_child(child_ptr->get_ip_addr());
    if (child_win_ptr != nullptr)
    {
        // Tell folks that the active window just went away.
        child_activated(nullptr);

        // Clean out the MDI area.
        ui->mdi_area->removeSubWindow(child_ptr);
        ui->mdi_area->setActiveSubWindow(nullptr);

        // Close the child window now.
        child_win_ptr->close();
    }

    // Make sure we have a camera dock.
    if (get_camera_dock() != nullptr)
    {
        // Remove the camera from the list.
        get_camera_dock()->set_camera_offline(child_ptr->get_ip_addr());
    }

    // Do we have a multicamera dock?
    if (multicam_dock != nullptr)
    {
        // Remove the camera from the multicamera control dock, if it's in there.
        multicam_dock->remove_camera(child_ptr);
    }

    // Do we have a camera log visible?
    if (p_camera_logs_dock->isVisible())
    {
        if (p_camera_logs_dock->isCameraAttached(child_ptr))
        {
            p_camera_logs_dock->removeCameraLog(child_ptr);
            p_camera_logs_dock->setVisible(false);
        }
    }

    // Disconnect from the camera.
    child_ptr->set_camera_info(nullptr);
    child_ptr->deleteLater();

    // Update the menus.
    reflect_camera_state();

    // Are we in the middle of a shutdown?
    if (in_shutdown)
    {
        // Come back in a little bit to see if all the windows are closed.
        QTimer::singleShot(SHUTDOWN_DELAY, this, &MainWindow::shutdown_process);
    }

    // Allow someone else to come in now.
    block = 0;
}

/*!
 * @brief Simple housekeeping function for when the telemetry dock
 *        is closed by the user.
 */
void MainWindow::close_telemetry()
{
    QSettings settings;
    // Sanity check.
    if (p_telemetry_dock == nullptr)
        return;

    settings.beginGroup("MainWindow");
    settings.setValue("telemetryArea", p_telemetry_dock->saveGeometry());
    settings.setValue("telemetryGeometry", dockWidgetArea(p_telemetry_dock));
    settings.endGroup();

    // Remove the widget and delete it.
    removeDockWidget(p_telemetry_dock);
    p_telemetry_dock->deleteLater();
    p_telemetry_dock = nullptr;
}

/*!
 * @brief Run the updater.
 */
void MainWindow::do_update()
{
    QSettings settings;

#ifdef Q_OS_MAC
    QStringList cmd_args;

    // Build up a command string to mount the DMG and open a Finder window to it.
    cmd_args << QStringLiteral("-e") << QStringLiteral("do shell script(\"hdiutil attach %1; open /Volumes/FasMotion\")").arg(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE));
    QProcess::startDetached(QStringLiteral("osascript"), cmd_args);
#else
    // Do the update!
    QProcess::startDetached(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE), QStringList());
#endif /* Not Q_OS_MAC */

    // Clear out the version setting.
    settings.beginGroup(FiUpdater::UPDATE_GROUP_TAG);
    settings.remove(FiUpdater::FM_UPDATE_VERSION);
    settings.endGroup();

    // Time to exit while the installer installs.
    close();
}

/*!
 * @brief Create a new MDI child window.
 * @return The QMdiSubWindow pointer of the new child, or nullptr if there's an error.
 */
QMdiSubWindow *MainWindow::create_mdi_child()
{
    QMdiSubWindow   *child_win_ptr;
    CameraMDI       *child;

    // Create our new child widget.
    child = new CameraMDI;
    if (child == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a new CameraMDI child!";
        return nullptr;
    }

    // Add the child to the MDI area.
    child_win_ptr = ui->mdi_area->addSubWindow(child);

    // Make sure we delete the window when it's closed.
    child_win_ptr->setAttribute(Qt::WA_DeleteOnClose);

    // Return our child window pointer.
    return child_win_ptr;
}

/*!
 * @brief Enable/disable (show/hide) the factory menu.
 *
 * @param do_enable True to enable/show the menu, false to disable/hide it.
 */
void MainWindow::enable_factory(bool do_enable)
{
    // Save the enable flag for later.
    factory_enabled = do_enable;

    // Hide or show the menu.
    ui->menu_factory->setEnabled(factory_enabled);
    ui->menu_factory->menuAction()->setVisible(factory_enabled);
}

/*!
 * @brief Adds our camera user name to the linux group called disk.
 *
 * @returns True if this is a linux system and the camera user was added to the linux group called disk, False if not.
 */
bool MainWindow::add_cam_user_to_group_disk()
{
    bool added_to_group_disk = false;

#ifdef Q_OS_LINUX
    QProcess    myProcess;      // QT process object to manage system commands for group management.
    QSettings   settings;       // Will manage access to the [Preferences] section of our configuration set up.
    QString     user_name;      // Pulled from our [Preferences] section and default will be used if none is set up.

    QString     added_verify_str = QStringLiteral("Adding");    // The cmd_result string should contain this if gpasswd was ran.
    QString     group_disk  = QStringLiteral("disk");           // The group we need to belong to in order to open a LR drive.
    QString     gpasswd_cmd = QStringLiteral("/usr/bin/sudo");  // Must be sudu to run this command.
    QStringList gpasswd_args;
    QString     cmd_result, cmd_err;

    // Double check to make sure that this camera user isn't already part of the linux group called disk.
    if (cam_user_in_group_disk())
        return true;    // It's already there.

    // Get the camera's name from the [Preferences] section. Default user name is returned if none exists.
    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    user_name = settings.value(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
    settings.endGroup();

    // Make an attempt to add the camera user name to the group disk.
    gpasswd_args.append(QStringLiteral("gpasswd"));
    gpasswd_args.append(QStringLiteral("-a"));
    gpasswd_args.append(user_name);
    gpasswd_args.append(group_disk);

    myProcess.start(gpasswd_cmd, gpasswd_args, QIODevice::ReadOnly);
    myProcess.waitForFinished();
    cmd_result = myProcess.readAllStandardOutput();
    cmd_err = myProcess.readAllStandardError();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "myProcess gpasswd -a to disk group cmd_result:" << cmd_result << "cmd_err:" << cmd_err;

    // Should see the words "Adding" and "disk" in the cmd_result.  In fact it should say exactly this:
    // "Adding user fascam to group disk\n" - and cmd_err should be empty.  We will check for this condition
    // and set the result based on it.
    if (cmd_result.contains(added_verify_str) && cmd_result.contains(group_disk) && cmd_err.isEmpty())
    {
        added_to_group_disk = true; // We assume the operation was successful because we can't requery until reloading FasMotion.
    }

    // This won't work because until the FasMotion process is reloaded it doesn't know about the disk addition and this will fail
    // even though the group was probably added.  I left the code in place to document the fact here.
#if 0
    // We will base the success of this operation on whether or not we can verify the addiiton.
    added_to_group_disk = cam_user_in_group_disk();
#endif

#endif

    return added_to_group_disk;
}

/*!
 * @brief Determines if the camera user name belong to a linux group called disk.
 *
 * @returns True if this is a linux system and the camera user belongs to the linux group called disk, False if not.
 */
bool MainWindow::cam_user_in_group_disk()
{
    bool in_group_disk = false;

#ifdef Q_OS_LINUX
    QProcess    myProcess;      // QT process object to manage system commands for group management.
    QSettings   settings;       // Will manage access to the [Preferences] section of our configuration set up.
    QString     user_name;      // Pulled from our [Preferences] section and default will be used if none is set up.

    QString group_cmd = QStringLiteral("groups");    // The command is "group user_name" to see all groups belonging to user_name.
    QString group_disk = QStringLiteral("disk");    // The group we need to belong to in order to open a LR drive.
    QStringList group_args;
    QString cmd_result, cmd_err;

    // Get the camera's name from the [Preferences] section. Default user name is returned if none exists.
    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    user_name = settings.value(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
    settings.endGroup();

    // Ask the system if this user belongs to the group disk.
    myProcess.start(group_cmd, group_args, QIODevice::ReadOnly);
    myProcess.waitForFinished();
    cmd_result = myProcess.readAllStandardOutput();
    cmd_err = myProcess.readAllStandardError();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "myProcess test for group disk membership, cmd_result:" << cmd_result << "cmd_err:" << cmd_err;

    // Is "disk" part of the group result?
    if (cmd_result.contains(group_disk))
    {
        QString group_verified_str  = QString(tr("Verified %1 is a member of group %2.")).arg(user_name).arg(group_disk);
        QLOG_DEBUG() << group_verified_str;
        in_group_disk = true;  // We have verified that the cam_user belongs to the linux group "disk".
    }
#endif

    return in_group_disk;
}

/*!
 * @brief Enable/disable (show/hide) the multicamera support UI.
 *
 * @param do_enable True to enable/show, false to disable/hide it.
 */
void MainWindow::enable_multicamera(bool do_enable)
{
    // Save the enable flag for later.
    multicam_enabled = do_enable;

    // Show or hide the menu.
    if (multicam_enabled)
        ui->main_menu->insertMenu(ui->menu_window->menuAction(), ui->menu_multicam);
    else
        ui->main_menu->removeAction(ui->menu_multicam->menuAction());

    // Hide or show the menu to show the dock.
    ui->action_multicam->setVisible(multicam_enabled);

    // If we're disabling support, make sure we remove any "grouped" cameras.
    if (!multicam_enabled && (multicam_dock != nullptr))
        multicam_dock->delete_all_cameras();
}


/*!
 * @brief Enabe/disable the zoom UI controls.
 * @param do_enable     True to enable, false to disable.
 */
void MainWindow::enable_zoom_actions(bool do_enable)
{
    ui->action_zoom_1_to_1->setEnabled(do_enable);
    ui->action_zoom_in->setEnabled(do_enable);
    ui->action_zoom_out->setEnabled(do_enable);
    ui->action_zoom_to_fit->setEnabled(do_enable);
    ui->action_zoom_to_window->setEnabled(do_enable);
    ui->action_rotate_left->setEnabled(do_enable);
    ui->action_rotate_right->setEnabled(do_enable);
}

/*!
 * @brief Find the MDI child window associated with "ip_addr".
 * @param ip_addr   IP Address of the camera whose window we're looking for.
 * @return Return the QMdiSubWindow pointer to the found window, or nullptr if one isn't found.
 */
QMdiSubWindow *MainWindow::find_mdi_child(uint32_t ip_addr)
{
    foreach (QMdiSubWindow *window, ui->mdi_area->subWindowList())
    {
        CameraMDI *mdiChild = qobject_cast<CameraMDI *>(window->widget());
        if ((mdiChild != nullptr) && (mdiChild->get_ip_addr() == ip_addr))
            return window;
    }

    return nullptr;
}

/*!
 * @brief Determine if the file system at the location pointed to by
 *        the path_utf8 parameter supports large files (>4GB).
 *
 * @param target_path   The path to the system we're interested in.
 * @return True if the file system supports large files, false otherwise.
 */
bool MainWindow::fs_supports_large_files(const QString& target_path)
{
    bool    rc(true);                   // Assume success.
    QByteArray  fs_str;

    // See if we can get the file system name.
    if (!get_fs_name(target_path, fs_str))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to determine the filesystem name.";
        return false;
    }

    // Get our filesystem name in a structure we can use.
    fs_str = QString::fromUtf8(fs_str.constData()).toLower().toUtf8();
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << fs_str.constData();

    // Figure out if this is a "Large File" system.
    if ((fs_str == "vfat") ||
        (fs_str == "fat32") ||
        (fs_str == "fat16") ||
        (fs_str == "fat12") ||
        fs_str.contains("ms-dos"))
    {
        rc = false;
    }

    // All done.
    return rc;
}

bool MainWindow::do_autoconnect()
{
    bool autoconnect = false;

#ifdef Q_OS_LINUX
    QSettings settings;

    // See if the user wants us to automatically connect to the camera server.
    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    autoconnect = settings.value("autoconnect", true).toBool();
    settings.endGroup();
#endif

    return autoconnect;
}


/*!
 * @brief If the application is running on the camera, this will allow the user
 *        to have a "remote application" look and feel. If the application is
 *        already "remote", then there is no effect.
 *
 * Since we are running Linux on the camera, if this isn't a Linux based build
 * we can never be running "local".
 *
 * @note The asssumption is that we are called very early on, before event processing in fact.
 *       This assumption allows us to wait for the camera server to start by the simple process
 *       of sleeping.
 *
 * @param do_enable     True to enable the local control, false otherwise.
 */
void MainWindow::enable_local_control(bool do_enable)
{
#ifdef Q_OS_LINUX
    int pid;
    FILE *pid_file;
    char temp_buf[80];
    char *result_ptr;
    struct stat stat_buf;       // Status information structure.
    QSettings settings;
#else
    Q_UNUSED(do_enable)
#endif /* Q_OS_LINUX */

    // Assume we are remote.
    local_control_enable = false;

#ifdef Q_OS_LINUX
    // If we're not being asked to run local, we can leave.
    if (!do_enable)
        return;

    // It can take the camera server awhile (10-ish seconds?) to
    // start on boot. So, loop around a few times.
    for (int retries = 0; retries < WAIT_FOR_CS_TRIES; ++retries)
    {
        // Check to see if we are running "on-camera". We do that by looking
        // to see if the Camera Server is running.
        pid_file = fopen(CS_PID_FILE, "r");
        if (pid_file)
        {
            // Figure out what the pid is of the existing camera server.
            result_ptr = fgets(temp_buf, 80, pid_file);
            fclose(pid_file);

            // Make sure we got something.
            if (result_ptr)
            {
                // Build up the path to the /proc folder to see if this ID exists.
                pid = strtol(temp_buf, nullptr, 0);
                snprintf(temp_buf, 80, "/proc/%d", pid);

                // Now see if Linux knows about this guy.
                if (::stat(temp_buf, &stat_buf) == 0)
                {
                    QLOG_INFO() << __PRETTY_FUNCTION__ << "CameraServer (pid=" << pid << ") is running.";

                    // Do whatever the user wants.
                    QLOG_INFO() << __PRETTY_FUNCTION__ << "We are LOCAL.";
                    local_control_enable = do_enable;
                    return;
                }
            }
        }

        // Because of the assumption stated earlier, we can simply sleep here. Otherwise, we
        // would need to do something much more complicated.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Retry:" << retries << "Still no CS.";
        QThread::sleep(1);
    }
#endif /* Q_OS_LINUX */
}

bool MainWindow::local_control()
{
    return local_control_enable;
}

/*!
 * @brief Retrieve the total amount of space available on the path passed in.
 * @param target_path   The path to have the space returned for.
 *
 * @return The total space available.
 */
uint64_t MainWindow::get_avail_space(const QString& target_path)
{
    uint64_t    space_avail;

#ifdef Q_OS_WIN
    wchar_t     file_path[ 1024 ];
    ULARGE_INTEGER  space;

    // Initialize our variables.
    memset(file_path, 0, sizeof(file_path));
    space.QuadPart = 0;

    // Convert our target path to a WCHAR format. We need it in WCHAR for the
    // Windows GetDiskFreeSpaceEx() call.
    target_path.toWCharArray(file_path);

    // Get the space.
    if (!GetDiskFreeSpaceEx(file_path, &space, nullptr, nullptr))
    {
        LPTSTR  error_message;
        DWORD   error_code = GetLastError();

        // Skip these errors.
        if (error_code != NO_ERROR)
        {
            FormatMessage(FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_IGNORE_INSERTS,
                          nullptr, error_code, MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
                          (LPTSTR)&error_message, 0, nullptr);

            // Log the message.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString::fromUtf8("GetDiskFreeSpaceEx() failed. Error %1: %2")
                         .arg(error_code).arg(QString::fromWCharArray(error_message));
            LocalFree(error_message);
        }
    }

    // Get the amount of space available.
    space_avail = space.QuadPart;

#elif defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
    struct statvfs  os_stat;
    int             status;

    // Get the statistics for the volume associated with the target path.
    status = statvfs((const char *)target_path.toStdString().c_str(), &os_stat);
    if (status)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Call to statvfs() failed!";
        space_avail = 0;
    }
    else
    {
        // Compute the available space.
        space_avail = os_stat.f_bsize * os_stat.f_bavail;
    }
#else
#pragma message("WARNING: Hard drive space will not be checked during save")
    space_avail = 0;
#endif

    return space_avail;
}

/*!
 * @brief Return the list of actions associated with the Camera menu.
 */
QList<QAction *> MainWindow::get_camera_actions()
{
    QList<QAction *>    ret_list;

    if ((ui != nullptr) && (ui->menu_camera != nullptr))
        ret_list = ui->menu_camera->actions();

    return ret_list;
}

/*!
 * @brief Return a pointer to the autosave setup action.
 */
QAction *MainWindow::get_cam_autosave_setup()
{
    return ui->action_autosave_setup;
}

QAction *MainWindow::get_cam_autosave_enable()
{
    return ui->action_autosave_enable;
}

/*!
 * @brief Return a pointer to the connect to camera action.
 */
QAction *MainWindow::get_cam_connect()
{
    return ui->action_connect;
}

/*!
 * @brief Return a pointer to the add to MultiCamera dock action.
 */
QAction *MainWindow::get_cam_multicam()
{
    return ui->action_multicam;
}

/*!
 * @brief Return a pointer to the File > Open Device... action.
 */
QAction *MainWindow::get_cam_open_device()
{
    return ui->action_open_device;
}

/*!
 * @brief Return a pointer to the camera preferences action.
 */
QAction *MainWindow::get_cam_preferences()
{
    return ui->action_cam_preferences;
}

/*!
 * @brief Return a pointer to the camera update action.
 */
QAction *MainWindow::get_cam_update()
{
    return ui->action_update;
}

/*!
 * @brief Return a list of the MDI child windows.
 */
QList<CameraMDI *> MainWindow::get_children()
{
    QList<CameraMDI *> child_list;

    foreach (QMdiSubWindow *window, ui->mdi_area->subWindowList())
    {
        CameraMDI *mdi_child = qobject_cast<CameraMDI *>(window->widget());
        if (mdi_child != nullptr)
            child_list.append(mdi_child);
    }

    return child_list;
}

/*!
 * @brief Return a pointer to the black level configuration action.
 */
QAction *MainWindow::get_config_black_level()
{
    return ui->action_black_level_calibration;
}

/*!
 * @brief Return a pointer to the configuration menu.
 */
QMenu *MainWindow::get_config_menu()
{
    return ui->menu_configuration;
}

/*!
 * @brief Return a pointer to the WIFI configuration action.
 */
QAction *MainWindow::get_config_wifi()
{
    return ui->action_wifi;
}

/*!
 * @brief Return the action pointer for the "close all" menu item.
 */
QAction *MainWindow::get_cp_close_all()
{
    return ui->action_close_all_panels;
}

/*!
 * @brief Return the action pointer for the display settings menu item.
 */
QAction *MainWindow::get_cp_display()
{
    return ui->action_display_settings;
}

/*!
 * @brief Return the action pointer for the lens control menu item.
 */
QAction *MainWindow::get_cp_lens()
{
    return ui->action_lens_control;
}

/*!
 * @brief Return the action pointer for the record controls menu item.
 */
QAction *MainWindow::get_cp_reccontrols()
{
    return ui->action_record_controls;
}

/*!
 * @brief Return the action pointer for the record settings menu item.
 */
QAction *MainWindow::get_cp_recsettings()
{
    return ui->action_record_settings;
}

/*!
 * @brief Return the action pointer for the capture settings menu item.
 */
QAction *MainWindow::get_cp_storage()
{
    return ui->action_capture_settings;
}

/*!
 * @brief Return the action pointer for the view event markers menu item.
 */
QAction *MainWindow::get_event_marker_action()
{
    return ui->action_event_markers;
}

/*!
 * @brief Return the action pointer for the export videos menu item.
 */
QAction *MainWindow::get_export_action()
{
    return ui->action_export;
}

/*!
 * @brief Retrieve the name of the file system pointed to by the
 *        target_path parameter.
 *
 * @param[in] target_path   Path to the file system we're interested in.
 * @param[out] fs_name_buf  Buffer to hold the name of the file system.
 *
 * @return True if we were successful, false otherwise.
 */
bool MainWindow::get_fs_name(const QString& target_path, QByteArray& fs_name_buf)
{
    bool    result;

    // Clear the output buffer.
    fs_name_buf.clear();

#if defined( Q_OS_WIN )
    wchar_t     nameBuf[ MAX_PATH + 1 ];
    DWORD       fileSystemFlags = 0;
    wchar_t     fileSystemNameBuf[ MAX_PATH + 1 ];
    QString path_str = QDir::toNativeSeparators(target_path);

    // We need to extract the root folder/volume information from the
    // specified path.
    if (path_str.startsWith(QString::fromUtf8("\\\\?\\")))
        path_str = path_str.mid(4);
    if (path_str.isEmpty())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Can't get root information from path!";
        return false;
    }
    path_str[0] = path_str[0].toUpper();
    if (!((path_str.length() >= 2) &&
          (path_str.at(0).unicode() >= 'A') && (path_str.at(0).unicode() <= 'Z') &&
          (path_str.at(1) == QLatin1Char(':'))))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "path not well formed!" << path_str;
        return false;
    }

    // If we're of the format "C:\" then make sure that's
    // all we have.
    if (path_str.at(1) == QLatin1Char(':'))
        path_str.truncate(3);

    // Make sure our path ends with a backslash.
    if (!path_str.endsWith(QLatin1Char('\\')))
        path_str.append(QLatin1Char('\\'));

    // Query windows for the information on the volume associated with the path.
    result = ::GetVolumeInformation(reinterpret_cast<const wchar_t *>(path_str.utf16()),
                                    nameBuf, MAX_PATH,
                                    nullptr, nullptr,
                                    &fileSystemFlags,
                                    fileSystemNameBuf, MAX_PATH);
    if (result)
    {
        fs_name_buf = QString::fromWCharArray(fileSystemNameBuf).toUtf8();
    }
#elif defined( Q_OS_LINUX )
    // Assume failure.
    result = false;

    // Open up the mount table.
    FILE    *fp = ::setmntent("/etc/mtab", "r");
    if (fp == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "setmntent() failed!";
    }
    else
    {
        struct mntent   *mnt_info;

        // Loop through all the entries in the file looking for ours.
        while ((mnt_info = ::getmntent(fp)))
        {
            QString path_str = QDir::toNativeSeparators(target_path);

            // Does our path name contain the mount folder string?
            if (path_str.contains(QFile::decodeName(mnt_info->mnt_dir)))
            {
                // Found it.
                result = true;

                // Get the filesystem name.
                fs_name_buf = QByteArray(mnt_info->mnt_type);

                // All done.
                break;
            }
        }
    }
#elif defined( Q_OS_MAC )
    QString path_str = QDir::toNativeSeparators(target_path);

    // Assume failure.
    result = false;

    // First, initialize our root path information.
    if (!getUrlProperties(true, path_str))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Can't get root information from path!";
    }

    // Now get our file system name.
    else if (getUrlProperties(false, path_str))
    {
        fs_name_buf = fileSystemName;
        result = true;
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Can't get file system info from path!";
    }

#else
#pragma message( "WARNING: Unknown OS. Unable to determine Filesystem info." )
    result = false;
#endif

    // All done.
    return result;
}

/*!
 * @brief Return the action pointer for the aiming cursor menu item.
 */
QAction *MainWindow::get_view_aiming_cursor()
{
    return ui->action_aiming_cursor;
}

/*!
 * @brief Return the action pointer for the info cursor menu item.
 */
QAction *MainWindow::get_view_info_cursor()
{
    return ui->action_info_cursor;
}

/*!
 * @brief Return the action pointer for the load capture menu item.
 */
QAction *MainWindow::get_load()
{
    return ui->action_load_capture;
}

/*!
 * @brief Return the action pointer for the release buffer menu item.
 */
QAction *MainWindow::get_release()
{
    return ui->action_release;
}

/*!
 * @brief Return the action pointer for the release all buffers menu item.
 */
QAction *MainWindow::get_release_all()
{
    return ui->action_release_all;
}

/*!
 * @brief Return the action pointer for the save image menu item.
 */
QAction *MainWindow::get_save()
{
    return ui->action_save;
}

/*!
 * @brief Return a pointer to the saveall action.
 */
QAction *MainWindow::get_save_all()
{
    return ui->action_save_all;
}

/*!
 * @brief Return the action pointer for the release buffer menu item.
 */
QAction *MainWindow::get_truncate()
{
    return ui->action_truncate;
}

/*!
 * @brief Return a pointer to the view histogram action.
 */
QAction *MainWindow::get_view_histogram()
{
    return ui->action_histogram;
}

/*!
 * @brief Return a pointer to the view local frametime action.
 */
QAction *MainWindow::get_view_local_frametime()
{
    return ui->action_local_frametime;
}


/*!
 * @brief Return a pointer to the view menu.
 */
QMenu *MainWindow::get_view_menu()
{
    return ui->menu_view;
}

/*!
 * @brief Return a pointer to the view PFMD action.
 */
QAction *MainWindow::get_view_pfmd()
{
    return ui->action_pfmd;
}

/*!
 * @brief Return a pointer to the view review controls action.
 */
QAction *MainWindow::get_view_review_controls()
{
    return ui->action_review_controls;
}

/*!
 * @brief Return a pointer to the view summary action.
 */
QAction *MainWindow::get_view_summary()
{
    return ui->action_summary;
}

/*!
 * @brief Initialize the communications object.
 *
 * @param net_buffers       The number of network buffers to use.
 * @param backoff_buffers   The backoff point, in empty buffers, before pausing the camera.
 * @param maintain_net_stats    True if we're to ask the communications object to maintain statistics.
 */
void MainWindow::init_comm(int net_buffers, int backoff_buffers, bool maintain_net_stats)
{
    // Do we already have a communications object?
    if (comm_ptr != nullptr)
    {
        // Tell the comms object to shutdown.
        comm_ptr->shutdown();

        // Clean it up.
        delete comm_ptr;
        comm_ptr = nullptr;
    }

    // If we don't have a memory pool create one.  This can be called multiple
    // times if communications are reset.  But once a memory pool is created
    // that memory pool will remain active.
    if (!FI_MEMPOOL::FiMemoryPool::HasAMemoryPool())
    {
        init_image_pool(num_net_buffers, num_backoff_buffers);
    }

    // Allocate a new communications object.
    comm_ptr = new FI_comm(net_buffers, backoff_buffers);
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate FI_comm object!";
        return;
    }

    // Have the comm object maintain statistics (or not).
    keep_statistics = maintain_net_stats;
}

/*!
 * @brief Initialize the image pool object.
 *
 * @param net_buffers       The number of network buffers to use.
 * @param backoff_buffers   The backoff point, in empty buffers, before pausing the camera.
 */
void MainWindow::init_image_pool(int net_buffers, int backoff_buffers)
{
    // Allocate our image pool.
    if (! FI_MEMPOOL::FiMemoryPool::CreateMemoryPool((uint32_t)net_buffers,
                                                     FI_MEMPOOL::MAX_MEMORY_BLOCKSIZE,
                                                     (uint32_t)backoff_buffers))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create the image pool!";
        return;
    }
}

/*!
 * @brief This method initializes the signal/slot connections for
 *        the main window.
 */
void MainWindow::init_connects()
{
    connect(ui->menu_file, SIGNAL(aboutToShow()), this, SLOT(file_menu_showing()));
    connect(ui->menu_camera, SIGNAL(aboutToShow()), this, SIGNAL(camera_menu_showing()));
    connect(ui->menu_multicam, SIGNAL(aboutToShow()), this, SLOT(multicam_menu_showing()));
    connect(ui->menu_factory, SIGNAL(aboutToShow()), this, SLOT(factory_menu_showing()));
    connect(ui->menu_help, SIGNAL(aboutToShow()), this, SLOT(help_menu_showing()));

    connect(ui->action_open, SIGNAL(triggered()), this, SLOT(file_open()));
    connect(ui->action_open_raw_drive, SIGNAL(triggered()), this, SLOT(open_raw_drive()));
    connect(ui->action_open_device, SIGNAL(triggered()), this, SIGNAL(camera_open_device()));
    connect(ui->action_aiming_cursor, SIGNAL(triggered()), this, SIGNAL(view_aiming_cursor()));
    connect(ui->action_info_cursor, SIGNAL(triggered()), this, SIGNAL(view_info_cursor()));
    connect(ui->action_load_capture, SIGNAL(triggered()), this, SIGNAL(camera_load()));
    connect(ui->action_save, SIGNAL(triggered()), this, SIGNAL(camera_save()));
    connect(ui->action_save_all, SIGNAL(triggered()), this, SIGNAL(camera_save_all()));
    connect(ui->action_save_state, SIGNAL(triggered()), this, SIGNAL(save_settings()));
    connect(this, SIGNAL(save_settings()), this, SLOT(write_settings()));
    connect(ui->action_clear_list, SIGNAL(triggered()), this, SLOT(clear_recent_videos()));
    connect(ui->action_exit, SIGNAL(triggered()), this, SLOT(close()));

    connect(ui->action_chinese_simplified, SIGNAL(triggered()), this, SLOT(lang_chinese_simplified()));
    connect(ui->action_chinese_traditional, SIGNAL(triggered()), this, SLOT(lang_chinese_traditional()));
    connect(ui->action_english, SIGNAL(triggered()), this, SLOT(lang_english()));
    connect(ui->action_french, SIGNAL(triggered()), this, SLOT(lang_french()));
    connect(ui->action_german, SIGNAL(triggered()), this, SLOT(lang_german()));
    connect(ui->action_japanese, SIGNAL(triggered()), this, SLOT(lang_japanese()));
    connect(ui->action_korean, SIGNAL(triggered()), this, SLOT(lang_korean()));
    connect(ui->action_russian, SIGNAL(triggered()), this, SLOT(lang_russian()));
    connect(ui->action_spanish, SIGNAL(triggered()), this, SLOT(lang_spanish()));

    connect(this, SIGNAL(connect_to_camera(int)), this, SLOT(connect_camera(int)), Qt::QueuedConnection);
    connect(ui->action_about, SIGNAL(triggered()), this, SLOT(about()));
    connect(ui->action_preferences, SIGNAL(triggered()), this, SLOT(preferences()));
    connect(ui->action_autosave_setup, SIGNAL(triggered()), this, SIGNAL(camera_autosave_setup()));
    connect(ui->action_reboot, SIGNAL(triggered()), this, SIGNAL(camera_reboot()));
    connect(ui->action_power_down, SIGNAL(triggered()), this, SIGNAL(camera_power_down()));
    connect(ui->action_update, SIGNAL(triggered()), this, SIGNAL(camera_update()));
    connect(ui->action_samba, SIGNAL(triggered()), this, SIGNAL(camera_samba()));

    connect(ui->action_zoom_1_to_1, SIGNAL(triggered()), this, SIGNAL(zoom_1_to_1()));
    connect(ui->action_zoom_in, SIGNAL(triggered()), this, SIGNAL(zoom_in()));
    connect(ui->action_zoom_out, SIGNAL(triggered()), this, SIGNAL(zoom_out()));
    connect(ui->action_zoom_to_fit, SIGNAL(triggered()), this, SIGNAL(zoom_to_fit()));
    connect(ui->action_zoom_to_window, SIGNAL(triggered()), this, SIGNAL(zoom_to_window()));

    connect(ui->action_rotate_left, SIGNAL(triggered()), this, SIGNAL(rotate_left()));
    connect(ui->action_rotate_right, SIGNAL(triggered()), this, SIGNAL(rotate_right()));

    connect(ui->action_histogram, SIGNAL(triggered()), this, SLOT(view_histogram()));

    connect(ui->action_arm, SIGNAL(triggered()), this, SIGNAL(camera_arm()));
    connect(ui->action_live_mode, SIGNAL(triggered()), this, SIGNAL(camera_live()));
    connect(ui->action_review_mode, SIGNAL(triggered()), this, SIGNAL(camera_review()));
    connect(ui->action_trigger, SIGNAL(triggered()), this, SIGNAL(camera_trigger()));
    connect(ui->action_autosave_enable, &QAction::triggered, this, &MainWindow::camera_autosave_enable);
    connect(ui->action_cam_preferences, SIGNAL(triggered()), this, SIGNAL(camera_preferences()));
    connect(ui->action_snapshot, SIGNAL(triggered()), this, SIGNAL(camera_snapshot()));

    connect(ui->action_black_level_calibration, SIGNAL(triggered()), this, SIGNAL(config_black_level()));
    connect(ui->action_network, SIGNAL(triggered()), this, SLOT(network_config()));
    connect(ui->action_load_config, SIGNAL(triggered()), this, SIGNAL(config_load()));
    connect(ui->action_save_config, SIGNAL(triggered()), this, SIGNAL(config_save()));
    connect(ui->action_factory_reset, SIGNAL(triggered()), this, SIGNAL(config_factory_reset()));
    connect(ui->action_time_date, SIGNAL(triggered()), this, SIGNAL(config_time_date()));
    connect(ui->action_wifi, SIGNAL(triggered()), this, SIGNAL(config_wifi()));

    connect(ui->action_summary, SIGNAL(triggered()), this, SIGNAL(view_summary()));
    connect(ui->action_review_controls, SIGNAL(triggered()), this, SIGNAL(view_review_controls()));
    connect(ui->action_local_frametime, SIGNAL(triggered()), this, SIGNAL(view_local_frametime()));
    connect(ui->action_pfmd, SIGNAL(triggered()), this, SIGNAL(view_pfmd()));
    connect(ui->action_release, SIGNAL(triggered()), this, SIGNAL(release_buffer()));
    connect(ui->action_release_all, SIGNAL(triggered()), this, SIGNAL(release_all_buffers()));
    connect(ui->action_truncate, SIGNAL(triggered()), this, SIGNAL(truncate_recording()));
    connect(ui->action_event_markers, SIGNAL(triggered()), this, SIGNAL(setup_event_markers()));

    update_window_menu();
    connect(ui->menu_window, SIGNAL(aboutToShow()), this, SLOT(update_window_menu()));
    connect(ui->action_full_screen, SIGNAL(triggered()), this, SLOT(full_screen()));
    connect(ui->action_close, SIGNAL(triggered()), ui->mdi_area, SLOT(closeActiveSubWindow()));
    connect(ui->action_close_all_windows, SIGNAL(triggered()), ui->mdi_area, SLOT(closeAllSubWindows()));
    connect(ui->action_tile, SIGNAL(triggered()), ui->mdi_area, SLOT(tileSubWindows()));
    connect(ui->action_cascade, SIGNAL(triggered()), ui->mdi_area, SLOT(cascadeSubWindows()));
    connect(ui->action_next_window, SIGNAL(triggered()), ui->mdi_area, SLOT(activateNextSubWindow()));
    connect(ui->action_prev_window, SIGNAL(triggered()), ui->mdi_area, SLOT(activatePreviousSubWindow()));

    // Connect the factory menu items.
    connect(ui->action_factory_black_level_calibration, SIGNAL(triggered()), this, SLOT(factory_black_level()));
    connect(ui->action_prnu_calibration, SIGNAL(triggered()), this, SLOT(factory_prnu()));
    connect(ui->action_raw_histogram, SIGNAL(triggered()), this, SLOT(factory_raw_histo()));
    connect(ui->action_system_information, SIGNAL(triggered()), this, SLOT(factory_sys_info()));
    connect(ui->action_open_console, SIGNAL(triggered()), this, SLOT(factory_open_console()));
    connect(ui->action_telemetry, SIGNAL(triggered()), this, SLOT(factory_telemetry()));
    connect(ui->action_sensor_corrections, SIGNAL(triggered()), this, SLOT(factory_sensor_correction()));
    connect(ui->action_row_offset, SIGNAL(triggered()), this, SLOT(factory_row_offset()));

    // Connect the help menu items.
    connect(ui->action_system_logs, SIGNAL(triggered()), this, SLOT(help_system_logs()));
    connect(ui->action_download_user_docs, SIGNAL(triggered()), this, SLOT(help_download_user_docs()));
    connect(ui->action_user_docs, SIGNAL(triggered()), this, SLOT(help_user_docs()));
    connect(ui->action_enable_teamviewer, SIGNAL(triggered()), this, SLOT(help_enable_teamviewer()));
    connect(ui->actionFusionDarkStyle, SIGNAL(triggered()), this, SLOT(fusiondark_style()));

    // Connect the commit notification.
    connect(qApp, &QGuiApplication::commitDataRequest, this, &MainWindow::commitData);
}

/*!
 * @brief Do the actions required to load the CAP file(s) specified by the file name parameter.
 *
 * @param file_name     The name of the CAP file to open.
 * @param force_maximized   True if when opened the window should be maximized.
 * @return True on success, false on failure.
 */
bool MainWindow::load_cap(const QString& file_name, bool force_maximized)
{
    PlaybackCap  *child;
    QMdiSubWindow   *child_win_ptr;

    // This can be rather lengthy, so show a busy cursor.
    QApplication::setOverrideCursor(Qt::BusyCursor);
    QApplication::processEvents();

    // Create our new child widget.
    child = new PlaybackCap(this);
    if (child == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a new PlaybackCap child!";

        // Restore our cursor before exiting.
        QApplication::restoreOverrideCursor();
        return false;
    }

    // Pass the video file to the playback object.
    if (!child->setFile(file_name))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The call to setFile on playback window failed!";

        // Restore our cursor.
        QApplication::restoreOverrideCursor();

        // Cleanup the child.
        delete child;

        // We've failed.
        return false;
    }

    // Add the child to the MDI area.
    child_win_ptr = ui->mdi_area->addSubWindow(child);

    // Make sure there is a child window pointer.
    if (child_win_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to addSubWindow for PlaybackCap.";

        // Restore our cursor.
        QApplication::restoreOverrideCursor();

        // Cleanup the child.
        delete child;

        // We've failed.
        return false;
    }

    // Remember the window associated with our widget.
    child->setWindow(child_win_ptr);

    // Make sure we delete the window when it's closed.
    child_win_ptr->setAttribute(Qt::WA_DeleteOnClose);

    // Make sure the window is visible.
    // Does the caller want this window maximized?
    if (force_maximized)
    {
        child_win_ptr->showMaximized();
    }
    else
    {
        child_win_ptr->showNormal();
    }

    // Jump the frame to the trigger frame.
    emit ui->action_tile->trigger();

    // Restore our cursor.
    QApplication::restoreOverrideCursor();

    // Connect up a signal for the Playback CAP review window closing our own
    // signal indicating the playback cap window is closing.
    connect(child, &PlaybackCap::closing, this, &MainWindow::playback_cap_closing);

    // Return our status.
    return true;
}

/*!
 * @brief Do the actions required to load the image(s) specified by the file name parameter.
 *
 * @param file_name         The name of the file to open.
 * @param force_maximized   True if when opened the window should be maximized.
 * @return True on success, false on failure.
 */
bool MainWindow::load_file(const QString& file_name, bool force_maximized)
{
    bool loadOK = false;
    QSettings       settings;
    QStringList     files;

    // Check to make sure the file name is valid.
    if (!PlaybackMDI::isVideoFileValid(file_name))
        return false;

    // Special case files.
    QFileInfo   file_info(file_name);
    if (file_info.suffix().compare(FiSaver::IMAGE_EXT_CAPTURE, Qt::CaseInsensitive) == 0)
    {
        // Attempt to open a CAP file.
        loadOK = load_cap(file_name, force_maximized);
    }
    else
    {
        // Create our new child widget.
        PlaybackMDI *child = new PlaybackMDI;
        if (child == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a new PlaybackMDI child!";
            return false;
        }

        // Pass the video file to the playback object.
        if (!child->setFile(file_name))
        {
            delete child;
            return false;
        }

        // Add the child to the MDI area.
        QMdiSubWindow   *child_win_ptr = ui->mdi_area->addSubWindow(child);

        // Make sure there is a child window pointer.
        if (child_win_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to addSubWindow for PlaybackMDI child!";
            return false;
        }
        else
        {
            child->setWindow(child_win_ptr);

            // Make sure we delete the window when it's closed.
            child_win_ptr->setAttribute(Qt::WA_DeleteOnClose);

            // Do we want the child window maximized:
            // child->setWindowState( child->windowState() | Qt::WindowMaximized );
            if (force_maximized)
            {
                child_win_ptr->showMaximized();
            }
            else
            {
                child_win_ptr->show();
            }

            // Or just make it the active one, still showing the others.
            emit ui->action_tile->trigger();

            // We are good to go so set the loadOk to true.
            loadOK = true;
        }
    }

    // Make sure we loaded OK.
    if (!loadOK)
        return false;

    // Save the file name in our recently used list.
    settings.beginGroup("MainWindow");
    files = settings.value("recentFileList").toStringList();

    // Make sure there aren't any duplicates.
    files.removeAll(file_name);

    // Put this on the list.
    files.prepend(file_name);

    // We only let a limited number stay around.
    while (files.size() > MAX_RECENT_VIDEOS)
        files.removeLast();

    // Save our list.
    settings.setValue("recentFileList", files);
    settings.endGroup();

    // Update our menu.
    update_recent_menu();

    // All done.
    return true;
}

/*!
 * @brief If the camera dock is active, have it stop discovery polling.
 */
void MainWindow::pause_discovery()
{
    if (p_camera_dock != nullptr)
    {
        p_camera_dock->set_discovery_check(false);
    }
}

/*!
 * @brief If the histogram is active, have it stop.
 */
void MainWindow::pause_histogram()
{
    if (histo_dock != nullptr)
        histo_dock->pause();
}

/*!
 * @brief Request all the cameras pause the video streams. If the "caller_ptr"
 *        parameter is not nullptr, then we'll skip that camera.
 *
 * @param caller_ptr    [OPTIONAL] Pointer to camera object we do not want paused.
 */
void MainWindow::pause_video_streaming(CameraMDI *caller_ptr)
{
    foreach (QMdiSubWindow *window, ui->mdi_area->subWindowList())
    {
        CameraMDI *mdi_child = qobject_cast<CameraMDI *>(window->widget());

        if (mdi_child != nullptr)
        {
            // Make sure this isn't the one we want to skip.
            if (mdi_child != caller_ptr)
                mdi_child->pause_video_request();
        }
    }
}

/*!
 * @brief The caller is requesting a single-shot histogram update.
 */
void MainWindow::push_one_histogram()
{
    if (histo_dock != nullptr)
        histo_dock->single_shot();
}


/*!
 * @brief Restore settings using the QSettings object.
 */
void MainWindow::read_settings()
{
    // Load our previous settings.
    QSettings settings;

    settings.beginGroup("MainWindow");
    restoreGeometry(settings.value("winGeometry").toByteArray());
    restoreState(settings.value("winState").toByteArray());
    restoreLanguage(settings.value("language", LANGUAGE_ENGLISH).toInt());
    enable_multicamera(settings.value("multicam_support", false).toBool());
    settings.endGroup();

    // Update our recent videos menu.
    update_recent_menu();

    // Set the checkbox of the full screen menu item based on geometry restore above.
    ui->action_full_screen->setChecked(isFullScreen());

    // Make sure we hide the metadata dock.
    if (p_metadata_dock != nullptr)
        p_metadata_dock->setVisible(false);

    // Make sure we recalculate our window sizes.
    adjustSize();
    update();
}

/*!
 * @brief Update the user interface to reflect the ARM'd camera state.
 */
void MainWindow::reflect_armed()
{
    ui->action_autosave_enable->setEnabled(false);
    ui->action_autosave_setup->setEnabled(false);
    ui->action_black_level_calibration->setEnabled(false);
    ui->action_event_markers->setEnabled(false);
    ui->action_factory_black_level_calibration->setEnabled(false);
    ui->action_factory_reset->setEnabled(false);
    ui->action_histogram->setEnabled(true);
    ui->action_aiming_cursor->setEnabled(true);
    ui->action_info_cursor->setEnabled(true);
    ui->action_load_capture->setEnabled(false);
    ui->action_load_config->setEnabled(false);
    ui->action_local_frametime->setEnabled(false);
    ui->action_network->setEnabled(false);
    ui->action_open->setEnabled(true);
    ui->action_open_console->setEnabled(true);
    ui->action_open_device->setEnabled(false);
    ui->action_pfmd->setEnabled(false);
    ui->action_power_down->setEnabled(false);
    ui->action_raw_histogram->setEnabled(true);
    ui->action_reboot->setEnabled(false);
    ui->action_release->setEnabled(false);
    ui->action_release_all->setEnabled(false);
    ui->action_review_controls->setEnabled(false);
    ui->action_samba->setEnabled(false);
    ui->action_save->setEnabled(false);
    ui->action_save_all->setEnabled(false);
    ui->action_save_config->setEnabled(false);
    ui->action_snapshot->setEnabled(false);
    ui->action_summary->setEnabled(true);
    ui->action_system_information->setEnabled(false);
    ui->action_system_logs->setEnabled(false);
    ui->action_download_user_docs->setEnabled(false);
    ui->action_user_docs->setEnabled(false);
    ui->action_time_date->setEnabled(false);
    ui->action_truncate->setEnabled(false);
    ui->action_cam_preferences->setEnabled(false);
    ui->action_wifi->setEnabled(false);
}

/*!
 * @brief Update the user interface to reflect the AUTOSAVE'd camera state.
 */
void MainWindow::reflect_autosave()
{
    ui->action_autosave_enable->setEnabled(false);
    ui->action_autosave_setup->setEnabled(false);
    ui->action_black_level_calibration->setEnabled(false);
    ui->action_event_markers->setEnabled(false);
    ui->action_factory_black_level_calibration->setEnabled(false);
    ui->action_factory_reset->setEnabled(false);
    ui->action_histogram->setEnabled(false);
    ui->action_aiming_cursor->setEnabled(true);
    ui->action_info_cursor->setEnabled(true);
    ui->action_load_capture->setEnabled(false);
    ui->action_load_config->setEnabled(false);
    ui->action_local_frametime->setEnabled(false);
    ui->action_network->setEnabled(false);
    ui->action_open->setEnabled(true);
    ui->action_open_console->setEnabled(true);
    ui->action_open_device->setEnabled(false);
    ui->action_pfmd->setEnabled(false);
    ui->action_power_down->setEnabled(true);
    ui->action_raw_histogram->setEnabled(false);
    ui->action_reboot->setEnabled(true);
    ui->action_release->setEnabled(false);
    ui->action_release_all->setEnabled(false);
    ui->action_review_controls->setEnabled(false);
    ui->action_samba->setEnabled(false);
    ui->action_save->setEnabled(false);
    ui->action_save_all->setEnabled(false);
    ui->action_save_config->setEnabled(false);
    ui->action_snapshot->setEnabled(false);
    ui->action_summary->setEnabled(true);
    ui->action_system_information->setEnabled(false);
    ui->action_system_logs->setEnabled(false);
    ui->action_download_user_docs->setEnabled(false);
    ui->action_user_docs->setEnabled(false);
    ui->action_time_date->setEnabled(false);
    ui->action_truncate->setEnabled(false);
    ui->action_cam_preferences->setEnabled(false);
    ui->action_wifi->setEnabled(false);
}

/*!
 * @brief Update the user interface to reflect the current camera state.
 */
void MainWindow::reflect_camera_state()
{
    uint32_t    camera_state;
    CameraMDI   *active_child;

    // Get our communications object pointer.
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child != nullptr)
    {
        if (active_child->get_camera_info()->offline)
        {
            // If it's off-line, then we are effectively not connected to a camera.
            reflect_no_camera();
            return;
        }
        else
        {
            // Get the actual camera state.
            camera_state = comm_ptr->get_reg(active_child->get_ip_addr(), GIGE_REG_CAMERA_STATE);
        }

        // The SAMBA menu item is only visible for series3 cameras.
        ui->action_samba->setVisible(active_child->is_series3());
    }
    else
    {
        // No camera so adjust UI accordingly.
        reflect_no_camera();
        return;
    }

    // Update as required.
    switch (camera_state)
    {
    default:
    case GIGE_CAMERA_STATE_IDLE:
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_SNAPSHOT:
    case GIGE_CAMERA_STATE_SHUTDOWN:
        reflect_idle();
        break;
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        reflect_live();
        break;
    case GIGE_CAMERA_STATE_ARMED:
        reflect_armed();
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
        reflect_triggered();
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        reflect_autosave();
        break;
    case GIGE_CAMERA_STATE_REVIEW:
        reflect_review();
        break;
    }
}

void MainWindow::reset_comm()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Re-initialize the communications object.
    init_comm(num_net_buffers, num_backoff_buffers, maintain_net_stats);
}

/*!
 * @brief Update the user interface to reflect the IDLE camera state.
 */
void MainWindow::reflect_idle()
{
    ui->action_arm->setEnabled(false);
    ui->action_arm->setIcon(QIcon(QString::fromUtf8(":/images/armbtn.png")));
    ui->action_live_mode->setEnabled(false);
    ui->action_review_mode->setEnabled(false);
    ui->action_trigger->setEnabled(false);
    ui->action_autosave_enable->setEnabled(false);

    ui->action_autosave_setup->setEnabled(false);
    ui->action_black_level_calibration->setEnabled(false);
    ui->action_event_markers->setEnabled(false);
    ui->action_factory_black_level_calibration->setEnabled(false);
    ui->action_factory_reset->setEnabled(false);
    ui->action_histogram->setEnabled(false);
    ui->action_aiming_cursor->setEnabled(true);
    ui->action_aiming_cursor->setEnabled(true);
    ui->action_info_cursor->setEnabled(true);
    ui->action_load_capture->setEnabled(false);
    ui->action_load_config->setEnabled(false);
    ui->action_local_frametime->setEnabled(false);
    ui->action_network->setEnabled(false);
    ui->action_open->setEnabled(true);
    ui->action_open_console->setEnabled(true);
    ui->action_open_device->setEnabled(true);
    ui->action_pfmd->setEnabled(false);
    ui->action_power_down->setEnabled(true);
    ui->action_raw_histogram->setEnabled(false);
    ui->action_reboot->setEnabled(true);
    ui->action_release->setEnabled(false);
    ui->action_release_all->setEnabled(false);
    ui->action_review_controls->setEnabled(false);
    ui->action_samba->setEnabled(false);
    ui->action_save->setEnabled(false);
    ui->action_save_all->setEnabled(false);
    ui->action_save_config->setEnabled(false);
    ui->action_snapshot->setEnabled(false);
    ui->action_summary->setEnabled(true);
    ui->action_system_information->setEnabled(false);
    ui->action_system_logs->setEnabled(false);
    ui->action_download_user_docs->setEnabled(false);
    ui->action_user_docs->setEnabled(false);
    ui->action_time_date->setEnabled(false);
    ui->action_truncate->setEnabled(false);
    ui->action_cam_preferences->setEnabled(false);
    ui->action_wifi->setEnabled(false);
}

/*!
 * @brief Update the user interface to reflect the LIVE camera state.
 */
void MainWindow::reflect_live()
{
    CameraMDI   *active_child;

    // Get our communications object pointer.
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the currently active child, if any.
    active_child = active_mdi_child();

    // Initialize the actions.
    update_autosave();

    ui->action_black_level_calibration->setEnabled(true);
    ui->action_event_markers->setEnabled(false);
    ui->action_factory_black_level_calibration->setEnabled(true);
    ui->action_factory_reset->setEnabled(true);
    ui->action_histogram->setEnabled(true);
    ui->action_aiming_cursor->setEnabled(true);
    ui->action_info_cursor->setEnabled(true);
    ui->action_load_capture->setEnabled(active_child->cap_is_allowed());
    ui->action_load_config->setEnabled(true);
    ui->action_local_frametime->setEnabled(false);
    ui->action_network->setEnabled(true);
    ui->action_open->setEnabled(true);
    ui->action_open_console->setEnabled(true);
    ui->action_open_device->setEnabled(true);
    ui->action_pfmd->setEnabled(false);
    ui->action_power_down->setEnabled(true);
    ui->action_raw_histogram->setEnabled(true);
    ui->action_reboot->setEnabled(true);
    ui->action_release->setEnabled(false);
    ui->action_release_all->setEnabled(false);
    ui->action_review_controls->setEnabled(false);
    ui->action_samba->setEnabled(true);
    ui->action_save->setEnabled(false);
    ui->action_save_all->setEnabled(false);
    ui->action_save_config->setEnabled(true);
    ui->action_snapshot->setEnabled(true);
    ui->action_summary->setEnabled(true);
    ui->action_system_information->setEnabled(true);
    //setCameraLogMenuItem();
    ui->action_time_date->setEnabled(true);
    ui->action_truncate->setEnabled(false);
    ui->action_cam_preferences->setEnabled(true);

    active_child->update_wifi_status();
}

/*!
 * @brief Update the user interface to reflect the case when we have no cameras opened.
 */
void MainWindow::reflect_no_camera()
{
    PlaybackMDI     *mdi_child = nullptr;
    QMdiSubWindow   *activeSubWindow;

    // Get the active MDI child.
    activeSubWindow = ui->mdi_area->activeSubWindow();
    if (activeSubWindow != nullptr)
        mdi_child = qobject_cast<PlaybackMDI *>(activeSubWindow->widget());

    // If there is no child window, then we need to configure these controls.
    // Otherwise we'll let the child window configure them.
    if (mdi_child == nullptr)
    {
        ui->action_event_markers->setEnabled(false);
        ui->action_save->setEnabled(false);
    }

    ui->action_arm->setEnabled(false);
    ui->action_arm->setIcon(QIcon(QString::fromUtf8(":/images/armbtn.png")));
    ui->action_live_mode->setEnabled(false);
    ui->action_review_mode->setEnabled(false);
    ui->action_trigger->setEnabled(false);
    ui->action_autosave_enable->setEnabled(false);

    ui->action_autosave_setup->setEnabled(false);
    ui->action_black_level_calibration->setEnabled(false);
    ui->action_factory_black_level_calibration->setEnabled(false);
    ui->action_factory_reset->setEnabled(false);
    ui->action_histogram->setEnabled(false);
    ui->action_aiming_cursor->setEnabled(false);
    ui->action_info_cursor->setEnabled(true);
    ui->action_load_capture->setEnabled(false);
    ui->action_load_config->setEnabled(false);
    ui->action_local_frametime->setEnabled(false);
    ui->action_network->setEnabled(false);
    ui->action_open->setEnabled(true);
    ui->action_open_console->setEnabled(false);
    ui->action_open_device->setEnabled(false);
    ui->action_pfmd->setEnabled(false);
    ui->action_power_down->setEnabled(false);
    ui->action_raw_histogram->setEnabled(false);
    ui->action_reboot->setEnabled(false);
    ui->action_release->setEnabled(false);
    ui->action_release_all->setEnabled(false);
    ui->action_review_controls->setEnabled(false);
    ui->action_samba->setEnabled(false);
    ui->action_save_all->setEnabled(false);
    ui->action_save_config->setEnabled(false);
    ui->action_snapshot->setEnabled(false);
    ui->action_summary->setEnabled(false);
    ui->action_system_information->setEnabled(false);
    ui->action_system_logs->setEnabled(false);
    ui->action_download_user_docs->setEnabled(false);
    ui->action_user_docs->setEnabled(false);
    ui->action_time_date->setEnabled(false);
    ui->action_truncate->setEnabled(false);
    ui->action_update->setEnabled(false);
    ui->action_cam_preferences->setEnabled(false);
    ui->action_wifi->setEnabled(false);
}

/*!
 * @brief Update the user interface to reflect the REVIEW camera state.
 */
void MainWindow::reflect_review()
{
    int         dirty_count;
    uint32_t    reg_value;
    uint32_t    ltr_mode;
    bool        ltr_enabled;
    bool        ltr_append_mode;
    CameraMDI   *active_child;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Review mode with no active child?!";
        return;
    }

    // Get the autosave settings.
    reg_value = active_child->read_camera_cache(GIGE_REG_AUTOSAVE);

    // How many dirty partitions are there?
    dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);

    // We're in review mode - so there's at least ONE dirty partition to free.
    ui->action_release->setEnabled(true);

    // Only enable "Release All" for multiple dirty partitions.
    ui->action_release_all->setEnabled(dirty_count > 1);

    // Read the LTR mode register.
    ltr_mode = active_child->read_camera_cache(GIGE_REG_LTR_MODE);
    ltr_enabled = (ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;
    ltr_append_mode = ltr_enabled && ((ltr_mode & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0);

    // Truncate is only allowed for LTR append-mode recordings.
    ui->action_truncate->setEnabled(ltr_append_mode);

    // Set the others as appropriate.
    ui->action_autosave_enable->setEnabled(false);
    ui->action_autosave_setup->setEnabled(false);
    ui->action_black_level_calibration->setEnabled(false);
    ui->action_event_markers->setEnabled(true);
    ui->action_factory_black_level_calibration->setEnabled(false);
    ui->action_factory_reset->setEnabled(false);
    ui->action_histogram->setEnabled(true);
    ui->action_aiming_cursor->setEnabled(true);
    ui->action_info_cursor->setEnabled(true);
    ui->action_load_capture->setEnabled((active_child->is_series3() ? false : active_child->cap_is_allowed()));
    ui->action_load_config->setEnabled(false);
    ui->action_local_frametime->setEnabled(true);
    ui->action_network->setEnabled(false);
    ui->action_open->setEnabled(true);
    ui->action_open_console->setEnabled(true);
    ui->action_open_device->setEnabled(false);
    ui->action_pfmd->setEnabled(true);
    ui->action_power_down->setEnabled(true);
    ui->action_raw_histogram->setEnabled(true);
    ui->action_reboot->setEnabled(true);
    ui->action_review_controls->setEnabled(true);
    ui->action_samba->setEnabled(false);
    ui->action_save->setEnabled(true);
    ui->action_save_all->setEnabled(true);
    ui->action_save_config->setEnabled(true);
    ui->action_snapshot->setEnabled(false);
    ui->action_summary->setEnabled(true);
    ui->action_system_information->setEnabled(true);
    //setCameraLogMenuItem();
    ui->action_time_date->setEnabled(false);
    ui->action_cam_preferences->setEnabled(false);
    ui->action_wifi->setEnabled(false);
}

/*!
 * @brief Update the user interface to reflect the TRIGGER'd camera state.
 */
void MainWindow::reflect_triggered()
{
    ui->action_autosave_enable->setEnabled(false);
    ui->action_autosave_setup->setEnabled(false);
    ui->action_black_level_calibration->setEnabled(false);
    ui->action_event_markers->setEnabled(false);
    ui->action_factory_black_level_calibration->setEnabled(false);
    ui->action_factory_reset->setEnabled(false);
    ui->action_histogram->setEnabled(true);
    ui->action_aiming_cursor->setEnabled(true);
    ui->action_info_cursor->setEnabled(true);
    ui->action_load_capture->setEnabled(false);
    ui->action_load_config->setEnabled(false);
    ui->action_local_frametime->setEnabled(false);
    ui->action_network->setEnabled(false);
    ui->action_open->setEnabled(true);
    ui->action_open_console->setEnabled(true);
    ui->action_open_device->setEnabled(false);
    ui->action_pfmd->setEnabled(false);
    ui->action_power_down->setEnabled(false);
    ui->action_raw_histogram->setEnabled(true);
    ui->action_reboot->setEnabled(false);
    ui->action_release->setEnabled(false);
    ui->action_release_all->setEnabled(false);
    ui->action_review_controls->setEnabled(false);
    ui->action_samba->setEnabled(false);
    ui->action_save->setEnabled(false);
    ui->action_save_all->setEnabled(false);
    ui->action_save_config->setEnabled(false);
    ui->action_snapshot->setEnabled(false);
    ui->action_summary->setEnabled(false);
    ui->action_system_information->setEnabled(false);
    ui->action_system_logs->setEnabled(false);
    ui->action_download_user_docs->setEnabled(false);
    ui->action_user_docs->setEnabled(false);
    ui->action_time_date->setEnabled(false);
    ui->action_truncate->setEnabled(false);
    ui->action_cam_preferences->setEnabled(false);
    ui->action_wifi->setEnabled(false);
}

/*!
 * @brief Restore our UI to the language the user had selected the last time
 *        we were running.
 *
 * @param sel_language      The selected [language](@ref FASMOTION_SETTINGS::SUPPORTED_LANGUAGES).
 */
void MainWindow::restoreLanguage(int sel_language)
{
    // Process the new selection.
    switch (sel_language)
    {
    case LANGUAGE_CHINESE_SIMPLIFIED:
        lang_chinese_simplified();
        break;
    case LANGUAGE_CHINESE_TRADITIONAL:
        lang_chinese_traditional();
        break;
    case LANGUAGE_FRENCH:
        lang_french();
        break;
    case LANGUAGE_GERMAN:
        lang_german();
        break;
    case LANGUAGE_JAPANESE:
        lang_japanese();
        break;
    case LANGUAGE_KOREAN:
        lang_korean();
        break;
    case LANGUAGE_RUSSIAN:
        lang_russian();
        break;
    case LANGUAGE_SPANISH:
        lang_spanish();
        break;
    default:
    case LANGUAGE_ENGLISH:
        lang_english();
        break;
    }
}

/*!
 * @brief If the camera dock is active, have it start discovery polling.
 */
void MainWindow::resume_discovery()
{
    // Resume camera discovery if we have a camera dock.
    if (p_camera_dock != nullptr)
    {
        p_camera_dock->set_discovery_check(true);
    }
}

/*!
 * @brief If the histogram is active, have it start up again.
 */
void MainWindow::resume_histogram()
{
    // Only resume the histogram if it's visible.
    if ((histo_dock != nullptr) && histo_dock->isVisible())
        histo_dock->resume();
}

/*!
 * @brief Request all the cameras resume the video streams.
 */
void MainWindow::resume_video_streaming()
{
    foreach (QMdiSubWindow *window, ui->mdi_area->subWindowList())
    {
        CameraMDI *mdi_child = qobject_cast<CameraMDI *>(window->widget());
        if (mdi_child != nullptr)
            mdi_child->resume_video_request();
    }
}

/*!
 * @brief Set a message on the status bar.
 * @param str               The message to display.
 * @param timeout_ms        How long (in milliseconds) to display the message.
 */
void MainWindow::set_statusbar_message(const QString& str, int timeout_ms)
{
    statusBar()->showMessage(str, timeout_ms);
}

/*!
 * @brief Display the text on the status bar.
 *
 * @param status_str    The string to be displayed on the status bar.
 */
void MainWindow::set_statusbar_text(const QString& status_str)
{
    if (status_str.isEmpty())
        p_statusbar_label->setText(tr(" Ready"));
    else
        p_statusbar_label->setText(status_str);
}

/*!
 *  @brief  We need a way to get the video control dock to resize, specifically
 *          to "shrink". The only way to do this, apparently, is to force the
 *          widget to a maximum size, wait for the change to be processed, and
 *          then relax the constraints to allow the user to resize.
 *
 *  @param set_width    The new width, -1 to ignore.
 *  @param set_height   The new height, -1 to ignore.
 */
void MainWindow::set_dock_size(QDockWidget *widget_ptr, int set_width, int set_height)
{
    QSize   oldMaxSize;

    // Remember what the current max size is, so we can restore it.
    oldMaxSize = widget_ptr->maximumSize();

    // Adjust as appropriate.
    if (set_width >= 0)
    {
        if (widget_ptr->width() < set_width)
            widget_ptr->setMinimumWidth(set_width);
        else
            widget_ptr->setMaximumWidth(set_width);
    }
    if (set_height >= 0)
    {
        if (widget_ptr->height() < set_height)
            widget_ptr->setMinimumHeight(set_height);
        else
            widget_ptr->setMaximumHeight(set_height);
    }

    // Get the events pushed through.
    qApp->processEvents();

    // We don't want a "fixed" size window, so restore the old max size and allow the min size to shrink.
    widget_ptr->setMinimumSize(QSize(0, 0));
    widget_ptr->setMaximumSize(oldMaxSize);
}

/*!
 * @brief This method will create the metadata display dock if necessary. In addition,
 *        it will set the dock's widget to the passed in object.
 * @param meta_info     The new metadata information widget to use with the dock.
 */
void MainWindow::set_metadata_info(PlaybackMetaInfo *meta_info)
{
    // Sanity check.
    if (p_metadata_dock == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No metadata dock!";
        return;
    }

    // Are we showing the dock?
    if (meta_info == nullptr)
        p_metadata_dock->setVisible(false);
    else
        p_metadata_dock->setWidget(meta_info);
}

/*!
 * @brief Initialize all the docking widgets for the MainWindow class.
 */
void MainWindow::setup_docks()
{
    bool    cpOpen;
    QSettings settings;

    // Setup our display area.
    setCentralWidget(ui->mdi_area);

    // Setup the layout so that the side dock(s) (e.g. the control panel)
    // take the full side. This way the video control window fits beneath
    // the video display correctly.
    setCorner(Qt::TopLeftCorner, Qt::LeftDockWidgetArea);
    setCorner(Qt::TopRightCorner, Qt::RightDockWidgetArea);
    setCorner(Qt::BottomLeftCorner, Qt::LeftDockWidgetArea);
    setCorner(Qt::BottomRightCorner, Qt::RightDockWidgetArea);

    // Create our control panel dock widget. This wants to be hidden until
    // we actually have a camera to control.
    p_control_panel = new ControlPanel(this);
    if (p_control_panel == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Couldn't allocate ControlPanel object!";
        return;
    }
    addDockWidget(Qt::LeftDockWidgetArea, p_control_panel);

    // Get our settings.
    cpOpen = settings.value("MainWindow/cpOpen", true).toBool();
    p_control_panel->restoreGeometry(settings.value("MainWindow/cpGeometry").toByteArray());

    // Restore the control panel.
    p_control_panel->setVisible(cpOpen);

    // Create and show our camera dock.
    p_camera_dock = new CameraDock(this);
    addDockWidget(Qt::TopDockWidgetArea, p_camera_dock);
    p_camera_dock->setVisible(true);

    // Create and show our multicamera dock.
    multicam_dock = new MultiCamDock(this);
    // Position the multicam Dock based on its usage in the past if any.
    if (multicam_dock)
    {
        // Restore the docking location. By default, to the bottom of the docking area.
        Qt::DockWidgetArea area = (Qt::DockWidgetArea)settings.value("MainWindow/multiCamArea").toInt();
        if (area == Qt::NoDockWidgetArea)
            area = Qt::RightDockWidgetArea;
        addDockWidget(area, multicam_dock);

        // Restore the logs dock size.
        multicam_dock->restoreGeometry(settings.value("MainWindow/multiCamGeometry").toByteArray());

        // Initially, the dock will be hidden.
        multicam_dock->setVisible(false);
    }

    // Create our histogram dock. Restore it's dimensions and such, but keep it hidden until
    // the user wants it.
    histo_dock = new HistoDock(this);
    if (histo_dock)
    {
        // Restore the docking location. By default, to the bottom of the docking area.
        Qt::DockWidgetArea area = (Qt::DockWidgetArea)settings.value("MainWindow/histoArea").toInt();
        if (area == Qt::NoDockWidgetArea)
            area = Qt::BottomDockWidgetArea;
        addDockWidget(area, histo_dock);

        // Restore the histogram dock size.
        histo_dock->restoreGeometry(settings.value("MainWindow/histoGeometry").toByteArray());
        histo_dock->setVisible(false);
    }

    // Create the camera logs dock.
    p_camera_logs_dock = new CameraLogs(this);
    if (p_camera_logs_dock)
    {
        // Restore the docking location. By default, to the bottom of the docking area.
        Qt::DockWidgetArea area = (Qt::DockWidgetArea)settings.value("MainWindow/logArea").toInt();
        if (area == Qt::NoDockWidgetArea)
            area = Qt::BottomDockWidgetArea;
        addDockWidget(area, p_camera_logs_dock);

        // Restore the logs dock size.
        p_camera_logs_dock->restoreGeometry(settings.value("MainWindow/logGeometry").toByteArray());

        // Initially, the dock will be hidden.
        p_camera_logs_dock->setVisible(false);
    }

    // Create our metadata dock.
    p_metadata_dock = new QDockWidget(tr("Metadata Display"), this);
    if (p_metadata_dock == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate the metadata_dock!";
        return;
    }
    p_metadata_dock->setObjectName("MetadataDock");
    p_metadata_dock->setAllowedAreas(Qt::RightDockWidgetArea | Qt::LeftDockWidgetArea);
    addDockWidget(Qt::RightDockWidgetArea, p_metadata_dock);
    p_metadata_dock->setVisible(false);

    // Make sure we recalculate our window sizes.
    adjustSize();
    update();
}

/*!
 * @brief Initialize the MainWindow status bar.
 */
void MainWindow::setup_statusbar()
{
    // Create our text/status area on the bar.
    p_statusbar_label = new QLabel(tr(" Ready"));
    if (p_statusbar_label == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Couldn't allocate QLabel object!";
        return;
    }

    // Add style to the label.
    p_statusbar_label->setFrameShape(QFrame::Panel);
    p_statusbar_label->setFrameShadow(QFrame::Sunken);

    // We want it right aligned.
    p_statusbar_label->setAlignment(Qt::AlignRight | Qt::AlignVCenter);

    // Now add it to the statusbar.
    ui->main_statusbar->addPermanentWidget(p_statusbar_label);
}

void MainWindow::fusiondark_style()
{
    // Uncheck the previous style.
    if (active_style != nullptr)
        active_style->setChecked(false);

    // Check the new style and make it active.
    active_style = ui->actionFusionDarkStyle;
    active_style->setChecked(true);

    qApp->setStyleSheet(QString(""));
    qApp->setStyle(QStyleFactory::create("Fusion"));
    qApp->setPalette(qApp->style()->standardPalette());

    QPalette palette;
    palette.setColor(QPalette::Window, Qt::darkGray);
    palette.setColor(QPalette::WindowText, Qt::white);
    palette.setColor(QPalette::Base, Qt::darkGray);
    palette.setColor(QPalette::AlternateBase, Qt::gray);
    palette.setColor(QPalette::ToolTipBase, Qt::white);
    palette.setColor(QPalette::ToolTipText, Qt::white);
    palette.setColor(QPalette::Text, Qt::white);

    palette.setColor(QPalette::Button, Qt::darkGray);
    palette.setColor(QPalette::ButtonText, Qt::white);
    palette.setColor(QPalette::BrightText, Qt::red);
    palette.setColor(QPalette::Link, QColor(42, 130, 218));

    palette.setColor(QPalette::Light, QColor(0xb3, 0xb3, 0xb3));
    palette.setColor(QPalette::Midlight, QColor(0x99, 0x99, 0x99));
    palette.setColor(QPalette::Button, Qt::gray);
    palette.setColor(QPalette::Mid, QColor(0x66, 0x66, 0x66));
    palette.setColor(QPalette::Dark, QColor(0x4d, 0x4d, 0x4d));
    palette.setColor(QPalette::Shadow, Qt::black);

    palette.setColor(QPalette::Disabled, QPalette::Text, Qt::lightGray);
    palette.setColor(QPalette::Disabled, QPalette::ButtonText, Qt::lightGray);
    palette.setColor(QPalette::Disabled, QPalette::WindowText, Qt::lightGray);

    palette.setColor(QPalette::Highlight, Qt::darkCyan);
    palette.setColor(QPalette::HighlightedText, Qt::white);
    qApp->setPalette(palette);
    qApp->setStyleSheet("QToolTip { color: #ffffff; background-color: #2a82da; border: 1px solid white; }");
}

void MainWindow::init_styles()
{
    QStringList styles_avail = QStyleFactory::keys();

    // Currently no active style selected.
    active_style = nullptr;

    // Make sure everyone is checkable.
    ui->actionDarkStyle->setCheckable(true);
    ui->actionFusionDarkStyle->setCheckable(true);

    // Add all the styles supported.
    for (int i = 0; i < styles_avail.count(); ++i)
    {
        QAction *new_action;
        QString action_label = styles_avail.at(i);

        new_action = ui->menu_styles->addAction(action_label);
        new_action->setCheckable(true);
        connect(new_action, SIGNAL(triggered()), this, SLOT(set_style()));
        style_items += new_action;

        // Make sure the UI style is set to the default.
        if (action_label.compare("Fusion", Qt::CaseInsensitive) == 0)
        {
            // That's the one.
            new_action->trigger();
        }
    }
}

void MainWindow::init_updater()
{
    // Allocate our updater object.
    updater = std::unique_ptr<FiUpdater>(new FiUpdater(this));
    if (!updater)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate the updater object!";
        return;
    }

    // Hook up the updater signals to our slots.
    connect(updater.get(), &FiUpdater::finished, this, &MainWindow::update_avail);
}

void MainWindow::setCameraLogMenuItem()
{
    bool cameraLogMenuEnabled = false;
    CameraMDI   *active_child = active_mdi_child();

    // Check the active child to see if the menu item should be enabled.
    if (active_child != nullptr)
    {
        cameraLogMenuEnabled = !active_child->is_offline();

        // If the active child is a series 3 camera we want to let this
        // be available for saving to SDCards.
        if (!active_child->is_series3())
        {
            // Is there a docked camera log window already displayed?
            if (p_camera_logs_dock)
            {
                if (p_camera_logs_dock->isVisible())
                {
                    cameraLogMenuEnabled = false;
                }
            }
        }
    }
    else
    {
        // Is there a docked camera log window already displayed?
        if (p_camera_logs_dock)
        {
            if (p_camera_logs_dock->isVisible())
            {
                cameraLogMenuEnabled = false;
            }
        }
    }

    // Set state of the menu enable.
    ui->action_system_logs->setEnabled(cameraLogMenuEnabled);
}

void MainWindow::set_style()
{
    QAction *style_action = qobject_cast<QAction *>(sender());
    QString sel_style;

    // Uncheck the previous style.
    if (active_style != nullptr)
        active_style->setChecked(false);

    // Check the new style and make it active.
    active_style = style_action;
    active_style->setChecked(true);

    // We need the text so we can select the style.
    sel_style = style_action->text();
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Setting style to:" << sel_style;

    qApp->setStyleSheet(QString(""));
    qApp->setStyle(QStyleFactory::create(sel_style));
    qApp->setPalette(qApp->style()->standardPalette());
}

void MainWindow::update_style(QString sel_style)
{
    QFile style_file(sel_style);
    if (!style_file.exists())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Style not found!" << sel_style;
        return;
    }

    style_file.open(QFile::ReadOnly | QFile::Text);
    qApp->setStyleSheet(style_file.readAll());
}

void MainWindow::update_autosave()
{
    CameraMDI   *active_child;
    bool        action_autosave_enabled_state = false;    // Should the action_autosave_enable be menu enabled or disabled?

    // Get our communications object pointer.
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child != nullptr)
    {
        bool ltr_capture_modes = false;

        // Get the LTR mode register.
        uint32_t reg_value = comm_ptr->get_reg(active_child->get_ip_addr(), GIGE_REG_LTR_MODE);

        // Is the LTR mode enabled?
        bool ltr_enabled = (reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;
        bool fflr_enabled = (reg_value & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR;

        if (ltr_enabled || fflr_enabled)
        {
            ltr_capture_modes = true;
        }

        // If LTR mode is enabled, then we are not allowed to do autosave.
        ui->action_autosave_setup->setEnabled(!ltr_capture_modes);

        // Check on our autosave value.
        reg_value = comm_ptr->get_reg(active_child->get_ip_addr(), GIGE_REG_AUTOSAVE);

        // If autosave is configured OR camera has autosave enabled set make sure to keep the action_autosave_enabled
        // selectable so that the user can turn it off...
        if (active_child->is_autosave_configured() || ((reg_value & GIGE_AUTOSAVE_ENABLE) == GIGE_AUTOSAVE_ENABLE))
        {
            // NOTE:  If the autosave is enabled on the camera but our active mdi window doesn't have a setup
            // configured for it we must provide a way to disable the autosave on the camera.  Therefore that
            // condition will fall through to here.  If autosave is not enabled on the camera but our active
            // mdi window does have an autosave configuration then we also want to allow the menu system to
            // set the action enabled.
            action_autosave_enabled_state = true;
        }

        ui->action_autosave_enable->setEnabled(!ltr_capture_modes && action_autosave_enabled_state);
        ui->action_autosave_enable->setChecked(!ltr_capture_modes && ((reg_value & GIGE_AUTOSAVE_ENABLE) == GIGE_AUTOSAVE_ENABLE));
    }
}

/*!
 * @brief Update the recent video menu.
 */
void MainWindow::update_recent_menu()
{
    int     i;
    int     recent_file_count;
    QStringList files;

    // Load our previous settings.
    QSettings settings;
    settings.beginGroup("MainWindow");
    files = settings.value("recentFileList").toStringList();
    settings.endGroup();

    // Figure out how many recent files we have.
    recent_file_count = qMin(files.size(), (int)MAX_RECENT_VIDEOS);

    // Add the files to our menu.
    for (i = 0; i < recent_file_count; ++i)
    {
        QString menu_text = tr("&%1 %2").arg(i + 1).arg(files[i]);

        recent_video_acts[i]->setText(menu_text);
        recent_video_acts[i]->setData(files[i]);
        recent_video_acts[i]->setVisible(true);
    }
    for (i = recent_file_count; i < MAX_RECENT_VIDEOS; ++i)
        recent_video_acts[i]->setVisible(false);

    ui->menu_recent_videos->setEnabled(recent_file_count > 0);
    recent_video_sep->setVisible(recent_file_count > 0);
    ui->action_clear_list->setVisible(recent_file_count > 0);
}

void MainWindow::update_review(CameraMDI *active_child)
{
    uint32_t    part_state;
    uint32_t    camera_state;

    // If there's an active child, check the partition state to determine if we have something to review.
    if (active_child != nullptr)
    {
        // Get the actual camera state.
        camera_state = comm_ptr->get_reg(active_child->get_ip_addr(), GIGE_REG_CAMERA_STATE);

        // Get the current parition state on the camera.
        part_state = comm_ptr->get_reg(active_child->get_ip_addr(), GIGE_REG_PART_STATE);
    }
    else
    {
        // No camera, no partition.
        camera_state = GIGE_CAMERA_STATE_IDLE;
        part_state = GIGE_PART_STATE_EMPTY;
    }

    // Does the camera have something to review?
    ui->action_review_mode->setEnabled((camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO) && (part_state == GIGE_PART_STATE_COMPLETE));
}

QString MainWindow::version()
{
    QString ver_str;
#if defined( _WIN64 ) || defined( __x86_64__ )
    QString build_str("64-bit");
#else
    QString build_str("32-bit");
#endif

    // Do we have version information?
    if ((APP_VERSION_PATCH == 0) && (APP_VERSION_MINOR == 0) && (APP_VERSION_MAJOR == 0))
    {
        // Use our build time instead.
        QFileInfo   our_info(qApp->applicationFilePath());
        ver_str = QString("Built: %1 (%2)").arg(our_info.lastModified().toLocalTime().toString(Qt::TextDate), build_str);
    }
    else
    {
        // Use the version information.
        ver_str = QString("Version: %1.%2.%3 (%4)").arg(APP_VERSION_MAJOR).arg(APP_VERSION_MINOR).arg(APP_VERSION_PATCH).arg(build_str);
    }

    // Return our string.
    return ver_str;
}

QVersionNumber MainWindow::version_num()
{
    QVersionNumber ver_no;

    // Do we have version information?
    if ((APP_VERSION_PATCH == 0) && (APP_VERSION_MINOR == 0) && (APP_VERSION_MAJOR == 0))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Must be a dev build. Faking version number";
        ver_no = QVersionNumber(3, 0, 0);
    }
    else
    {
        // Use the version information.
        ver_no = QVersionNumber(APP_VERSION_MAJOR, APP_VERSION_MINOR, APP_VERSION_PATCH);
    }

    // Return our string.
    return ver_no;
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void MainWindow::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        ui->retranslateUi(this);
    }

    // Pass it on.
    QMainWindow::changeEvent(event);
}

/*!
 * @brief The user wants to close the application.
 *
 * The process of disconnecting from the cameras is started here.
 * If there are camera windows open, we don't want to close the main window
 * just yet. Instead, we'll set the "in_shutdown" flag. As the camera windows
 * close, the invoke MainWindow::close_mdi_child(). This will look at the
 * flag, and if it is set will start a timer which calls MainWindow::shutdown_process().
 *
 * When all the child windows are closed, MainWindow::shutdown_process() will
 * call the object close() method and away we go...
 *
 * @param event     Pointer to the close event object.
 *
 * @sa CameraMDI::closeEvent()
 * @sa MainWindow::commitData()
 * @sa MainWindow::shutdown_process()
 * @sa MainWindow::close_mdi_child()
 */
void MainWindow::closeEvent(QCloseEvent *event)
{
    // Only do this if we're not already in shutdown.
    if (!in_shutdown)
    {
        // Remember we're in shutdown.
        in_shutdown = true;

        // Let folks know that we are going away.
        emit closing();

        // Save our settings.
        write_settings();

        // Do we have any children around?
        if (ui->mdi_area->subWindowList().count() > 0)
        {
            bool    camera_mdi_children = false;

            // Are any of the sub-windows, camera sub-windows?
            foreach (QMdiSubWindow *window, ui->mdi_area->subWindowList())
            {
                CameraMDI *mdiChild = qobject_cast<CameraMDI *>(window->widget());
                if (mdiChild != nullptr)
                {
                    camera_mdi_children = true;
                    break;
                }
            }

            // Close all of the subwindows.
            ui->mdi_area->closeAllSubWindows();

            // If there are camera sub-windows, ignore the event for now. We'll get
            // re-invoked via the shutdown_process() method.
            if (camera_mdi_children)
                event->ignore();
        }
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user selected the "About" menu.
 */
void MainWindow::about()
{
    QString about_str;
    QString serial_no = serialNumber();
    QString model_no = modelNumber();

    // Build up the string to show the user.
    about_str = version() + "\n\n";
    if (!serial_no.isEmpty())
        about_str += "Serial #: " + serial_no + "\n";
    if (!model_no.isEmpty())
        about_str += "Model #: " + model_no + "\n";
    about_str += "Copyright (c) 2014-2022 Fastec Imaging Inc.";

    // Display our info.
    QMessageBox::about(this, tr("About FasMotion"), about_str);
}

/*!
 * @brief The "window" child has been activated. We need to let other
 *        objects know about this new camera.
 *
 * @param window    Pointer to the newly activated child window.
 */
void MainWindow::child_activated(QMdiSubWindow *window)
{
    // Pass on a signal letting folks know that the active child has changed.
    emit active_child_changed(window);
}

/*!
 * @brief Clear out the "recent videos" menu.
 */
void MainWindow::clear_recent_videos()
{
    QSettings   settings;

    // Clear the recent file list.
    settings.beginGroup("MainWindow");
    settings.remove("recentFileList");
    settings.endGroup();

    // Update the menu.
    update_recent_menu();
}

/*!
 * @brief The session manager wants us to close all open files.
 *
 * In our world, this means we are going to be closing down.
 *
 * @param manager   Our session manager.
 */
void MainWindow::commitData(QSessionManager& manager)
{
    Q_UNUSED(manager)

    // Let folks know that we are going away.
    emit closing();

    // Remember we're in shutdown.
    in_shutdown = true;

    // Save our settings.
    write_settings();

    // Do we have any children around?
    if (ui->mdi_area->subWindowList().count() > 0)
    {
        // Close all of the subwindows.
        ui->mdi_area->closeAllSubWindows();
    }
}

/*!
 * @brief Connect to the camera specified by "row_index". This parameter is an index
 *        into the [camera dock](@ref CameraDock) object.
 *
 * @param row_index     Index to the camera information we are interested in.
 */
void MainWindow::connect_camera(int row_index)
{
    CameraMDI       *child;
    QMdiSubWindow   *existing_mdi;
    FI_COMM::FI_CAMERA_INFO *cam_info;

    // Get the camera information for the camera selected by the user.
    cam_info = p_camera_dock->get_info_by_row(row_index);
    if (cam_info == nullptr)
    {
        QString strMsg = QString::fromUtf8("%1: CameraDock::get_info_by_row( %2 ) returned nullptr!").arg(__PRETTY_FUNCTION__).arg(row_index);
        QLOG_ERROR() << strMsg;
        return;
    }

    // Do we have an active camera window with this IP?
    existing_mdi = find_mdi_child(cam_info->ip_addr);
    if (existing_mdi == nullptr)
    {
        QString     status_str;
        CameraMDI   *active_child;

        // Get the currently active child, if any, and tell it to go inactive.
        active_child = active_mdi_child();
        if (active_child != nullptr)
            active_child->set_inactive();

        // Create our new MDI child.
        QMdiSubWindow *child_win_ptr = create_mdi_child();

        // Make sure that the child window was created.
        if (child_win_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create an MDI child!";
            return;
        }

        // Get our MDI widget.
        child = qobject_cast<CameraMDI *>(child_win_ptr->widget());
        if (child != nullptr)
        {
            // Pass the window the camera information associated with it.
            child->set_window(child_win_ptr);
            child->set_camera_info(cam_info);
            status_str = QString(tr("%1 connecting...")).arg(child->user_friendly_name());
            set_statusbar_message(status_str, 2000);

            // Do we want the child window maximized:
            // child->setWindowState( child->windowState() | Qt::WindowMaximized );
            child_win_ptr->show();
            child_activated(child_win_ptr);
        }

        // Or just make it the active one, still showing the others.
        emit ui->action_tile->trigger();
    }
    else
    {
        // Bring up the one we want.
        child = qobject_cast<CameraMDI *>(existing_mdi->widget());
        if (child != nullptr)
            change_mdi_child(child);
    }
}

/*!
 * @brief The user is requesting a factory black reference frame capture.
 */
void MainWindow::factory_black_level()
{
    CameraMDI   *active_child;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active child!";
        return;
    }
    if (active_child->is_offline())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Camera is offline!";
        return;
    }

    // Ask the camera object to start a black reference frame capture.
    active_child->fpn_calibration(true);
}

/*!
 * @brief The user is requesting a factory PRNU calibration.
 */
void MainWindow::factory_prnu()
{
    CameraMDI   *active_child;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active child!";
        return;
    }
    if (active_child->is_offline())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Camera is offline!";
        return;
    }

    // Ask the camera object to start a PRNU calibration.
    active_child->prnu_calibration();
}

/*!
 * @brief The user has selected the factory menu and it's about to be displayed.
 */
void MainWindow::factory_menu_showing()
{
    CameraMDI *active_child;
    fiSensor *sensorPtr = nullptr;  // To check sensor type for raw histogram.
    bool isRawHistogramVisible = false;    // Is displaying raw histogram data visible.
    bool isPrnuEnabled = false;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child != nullptr)
    {
        // What sensor are we connecting to?
        sensorPtr = active_child->get_sensor();
        if (sensorPtr != nullptr)
        {
            switch (sensorPtr->get_type())
            {
            case FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR:
            case FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO:
                // Raw histogram is available ONLY for LUPA1300 sensors.
                isRawHistogramVisible = true;
                break;
            case FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR:
            case FI_SENSOR::SENSOR_TYPE_LUX13HS_MONO:
            case FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR:
            case FI_SENSOR::SENSOR_TYPE_LUX19HS_MONO:
            case FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR:
            case FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO:
                // PRNU is only available for seriesNC cameras.
                if (active_child->is_series3())
                    break;

                // PRNU is only available for these sensors.
                isPrnuEnabled = true;
                break;
            default:
                break;
            }
        }
    }

    // Setup the Histogram menu.
    ui->action_raw_histogram->setVisible(isRawHistogramVisible);
    if (!isRawHistogramVisible)
    {
        // Make sure it is not checked or enabled not that it matters because
        // the menu item is not visible but want to make sure residual state
        // doesn't impact regular histogram usage.
        if (ui->action_raw_histogram->isChecked())
        {
            ui->action_raw_histogram->setChecked(false);
        }
        ui->action_raw_histogram->setEnabled(false);
    }

    // Setup the PRNU calibration menu.
    ui->action_prnu_calibration->setEnabled(isPrnuEnabled);
}

/*!
 * @brief The user is requesting that we open a console window to the active camera.
 */
void MainWindow::factory_open_console()
{
    CameraMDI       *active_child;
    QString         url_string;
    QHostAddress    our_addr;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active child!";
        return;
    }

    // convert the packed binary format into a human readable ip address.
    our_addr.setAddress(qFromBigEndian(active_child->get_ip_addr()));

    // If we are a series3 camera, use "Shell In A Box"
    if (active_child->is_series3())
    {
        // We get there by specifying the port number.
        url_string = QString::fromUtf8("http://%1:4200").arg(our_addr.toString());
        QDesktopServices::openUrl(url_string);
    }
    else
    {
        QSettings settings;
        QString user_name;

        // Get the username. Default is fascam.
        settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
        user_name = settings.value(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
        settings.endGroup();

#if defined( Q_OS_LINUX )
        // For newer cameras, just open up an SSH terminal.
        QStringList ssh_args;

        // We are making an assumption that 'gnome-terminal' and 'ssh' are both available.
        ssh_args << QStringLiteral("--") << QStringLiteral("ssh") << user_name + QString("@") + our_addr.toString();
        QProcess::startDetached(QStringLiteral("gnome-terminal"), ssh_args);
#elif defined( Q_OS_MAC )
        QStringList ssh_args;

        // Build up the command.
        ssh_args << "-e" << QString("tell app \"Terminal\" to do script \"/usr/bin/ssh ") + user_name + QString("@") + our_addr.toString() + QString("\"");
        QProcess::startDetached("/usr/bin/osascript", ssh_args);
#else
        QString putty_path;
        QStringList putty_args;

        // Get the full path to PuTTY.
        settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
        putty_path = settings.value("putty_path").toString();
        settings.endGroup();

        // WinOS: For newer cameras, just open up an SSH terminal.
        // Assumption: putty is on the path.
        putty_args << "-ssh" << user_name + QString("@") + our_addr.toString();
        QProcess::startDetached(putty_path, putty_args);
#endif
    }
}

/*!
 * @brief The user wants to see a raw histogram.
 */
void MainWindow::factory_raw_histo()
{
    // Sanity check.
    if (histo_dock == nullptr)
        return;

    // Tell the histogram whether to use raw or processed samples.
    histo_dock->set_raw_mode(ui->action_raw_histogram->isChecked());

    // Is the histogram visible?
    if (!histo_dock->isVisible() && ui->action_raw_histogram->isChecked())
    {
        // Make the histogram visible.
        ui->action_histogram->setChecked(true);
        histo_dock->setVisible(true);
        histo_dock->raise();
    }
}

void MainWindow::factory_row_offset()
{
    bool ok;
    int new_offset;

    // Is the menu being un-checked?
    if (!ui->action_row_offset->isChecked())
        return;

    // Ask the user for the starting row.
    new_offset = QInputDialog::getInt(this, "Enter Row Offset", "Offset [0: automatic]", row_offset_value, -100, 100, 1, &ok, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint);
    if (ok)
    {
        // Save the new value.
        row_offset_value = new_offset;
    }
}

void MainWindow::factory_sensor_correction()
{
    bool ok;
    int new_start;
    double new_gain;

    PlaybackMDI *active_playback = nullptr;
    QMdiSubWindow   *activeSubWindow;

    // Is the menu being un-checked?
    if (!ui->action_sensor_corrections->isChecked())
        return;

    // Is the menu being un-checked?
    if (!ui->action_sensor_corrections->isChecked())
        return;

    // Ask the user for the starting row.
    new_gain = QInputDialog::getDouble(this, "Enter Odd-Row Gain", "Offset [0.0: automatic]", row_gain_value, 0.0, 10.0, 4, &ok, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint);
    if (ok)
    {
        // Save the new value.
        row_gain_value = new_gain;
    }

    // Get the active MDI child.
    activeSubWindow = ui->mdi_area->activeSubWindow();
    if (activeSubWindow != nullptr)
        active_playback = qobject_cast<PlaybackMDI *>(activeSubWindow->widget());
    if (active_playback == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No playback window found!";
        return;
    }

    // Get the ROI.
    QRect roi = active_playback->getCameraInfo()->get_roi_rect();

    // Ask the user for the starting row.
    new_start = QInputDialog::getInt(this, "Enter Starting Row", "Starting Row", sensor_correction_start_row, 0, roi.height() - 1, 1, &ok, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint);
    if (ok)
    {
        // Save the new value.
        sensor_correction_start_row = new_start;
    }
    else
    {
        // Uncheck the selection.
        ui->action_sensor_corrections->setChecked(false);
    }
}

/*!
 * @brief The user would like to see system information for the active camera.
 */
void MainWindow::factory_sys_info()
{
    CameraMDI   *active_child;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active child!";
        return;
    }
    if (active_child->is_offline())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Camera is offline!";
        return;
    }

    PeekPokeDialog  dlg(active_child);
    dlg.set_ip(active_child->get_ip_addr());
    dlg.exec();
}

void MainWindow::factory_telemetry()
{
    QSettings settings;
    if (p_telemetry_dock)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Telemetry dock already exists.";
        return;
    }

    // Create and show our camera dock.
    p_telemetry_dock = new CameraTelemetry(this);
    if (p_telemetry_dock)
    {
        // Restore the docking location. By default, to the bottom of the docking area.
        Qt::DockWidgetArea area = (Qt::DockWidgetArea)settings.value("MainWindow/telemetryArea").toInt();
        if (area == Qt::NoDockWidgetArea)
            area = Qt::TopDockWidgetArea;
        addDockWidget(area, p_telemetry_dock);

        // Restore the telemetry dock size.
        p_telemetry_dock->restoreGeometry(settings.value("MainWindow/telemetryGeometry").toByteArray());

        // Initially, the dock will be shown.
        p_telemetry_dock->setVisible(true);
    }
}

/*!
 * @brief The user has selected the file menu and it's about to be displayed.
 */
void MainWindow::file_menu_showing()
{
#define RAW_DRIVE_ENABLE 1
#ifdef RAW_DRIVE_ENABLE
    fiRawDrive  raw_drive;

#ifdef Q_OS_LINUX
    // Ticket #347 added this logic to make sure local SeriesNC camera user is part of our the system's disk group.
    // Note:  The reason this was put here is that if the SeriesNC camera user is not a part of the system's disk group
    // then when this functionality below attempts to determine the drive set up the fascam user does not have permission
    // to look at that and there is no way to rectify it even after setting the group membership as we really need to
    // have a reboot of the system for it to take effect.  But at least an attempt will be made.

    // Is there membership in the group "disk"?
    if (!cam_user_in_group_disk())
    {
        // Attempt to create membership.
        if (!add_cam_user_to_group_disk())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to set up group membership to disk.";
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "The system has added group membership to disk.  System needs to be rebooted, warning user now.";

            // There are all sorts of issues trying to figure out how to restart FasMotion here.  There could
            // be a connected camera or not.  If there is a connected camera it may not be our local connected
            // camera.  The CameraMDI contains the "reboot" method.  For this reason I am choosing to only give
            // the user an informational message.  The likelihood of this ever happening is not very high since
            // we will perform a pp_update for our clients that may not have group membership to "disk".

            // Make sure the user understands that this is going to result in a camera reboot.
            QMessageBox::information(this, tr("Fastec Drive Access"),
                                     tr("The system was not configured for Fastec Drive access.\nIt has been modified to allow access to Fastec Drive(s).\n\nPlease reboot the system to access Fastec Drive(s)."),
                                     QMessageBox::Ok);
        }
    }
#endif

    // Only enable the raw drive menu if there's at least one Fastec raw drive connected.
    ui->action_open_raw_drive->setEnabled(!raw_drive.drives().isEmpty());
#else
    ui->action_open_raw_drive->setVisible(false);
#endif /* !RAW_DRIVE_ENABLE */
}

/*!
 * @brief Respond to the user selecting the "File -> Open" menu.
 */
void MainWindow::file_open()
{
    QString     file_name;
    QString     load_path;
    QSettings   settings;
    QFileInfo   load_info;

    // Initialize the default path based on the saved settings.
    settings.beginGroup("PathInfo");
    load_path = settings.value("load_root").toString();
    settings.endGroup();

    // Make sure the folder exists.
    load_info.setFile(load_path);
    if (!load_info.exists())
    {
        // Path no longer exists. Set it to the download path, if it's not already
        // the download path.
        if (load_path != SaveDialog::loadPath(0))
        {
            load_path = SaveDialog::loadPath(0);
            load_info.setFile(load_path);

            // Does this exist?
            if (!load_info.exists())
            {
                // How about the system documents path?
                load_path = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0);
            }
        }
    }

    // Ask the user for a file to open.
    file_name = QFileDialog::getOpenFileName(this, tr("Select A Video File To Open"), load_path, tr("Fastec Files (*.jpg *.jpeg *.bmp *.tif *.tiff *.dng *.avi *.cap);;All Files (*.*)"));
    if (file_name.isEmpty())
        return;

    // Save the folder the user is at so we can use it the next time they
    // want to open an image.
    settings.beginGroup("PathInfo");
    load_info.setFile(file_name);
    settings.setValue("load_root", load_info.absolutePath());
    settings.endGroup();

    // Load the file.
    load_file(file_name);
}

/*!
 * @brief Display the Find Cameras dialog. [DEPRECATED]
 */
void MainWindow::find_cameras()
{
#if 0
    FindCamerasDialog   *my_dlg;
    my_dlg = new FindCamerasDialog(this);

    // Pop up the dialog and let the user search for their camera.
    if (my_dlg->exec() != QDialog::Accepted)
    {
        // Clean up the dialog and exit.
        delete my_dlg;
        return;
    }

    // Clean up our dialog.
    delete my_dlg;
#endif
}

/*!
 * @brief Responsd to the user selecting the "Full Screen" menu item.
 */
void MainWindow::full_screen()
{
    if (ui->action_full_screen->isChecked())
    {
        QMainWindow::showFullScreen();
    }
    else
    {
        QMainWindow::showNormal();
    }
}

/*!
 * @brief Enable the Teamviewer application.
 *
 * Normally we ship the camera with the teamviewer app disabled. The only
 * way this menu item is enabled is when we're running locally on a Linux
 * system.
 */
void MainWindow::help_enable_teamviewer()
{
#ifdef Q_OS_LINUX
    // We want a temporary file to contain the commands to start the
    // Teamviewer app on bootup.
    QFile tv_batch("/tmp/tv_enable.sh");
    if (!tv_batch.open(QIODevice::WriteOnly | QIODevice::Truncate | QIODevice::Text))
    {
        fiMsgBox::warning(this, tr("Unable to write the Teamviewer startup commands."));
        return;
    }

    // Streaming our commands to the file.
    QTextStream tv_stream(&tv_batch);
    tv_stream << "#/bin/bash\n";
    tv_stream << "sudo systemctl enable teamviewerd\n";
    tv_batch.close();

    // Make sure the batch file is executable.
    tv_batch.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ExeOwner | QFile::ReadUser | QFile::WriteUser | QFile::ExeUser | QFile::ReadGroup | QFile::ExeGroup | QFile::ReadOther | QFile::ExeOther);

    // Run the command(s) to start up teamviewer.
    QProcess::startDetached(tv_batch.fileName(), QStringList());

    // Tell the user that they need to reboot now.
    QMessageBox::information(MainWindow::GetInstance().get(), tr("Need To Reboot"), tr("You need to reboot the system for the changes to take effect."));
#endif
}

/*!
 * @brief The user has selected the help menu and it's about to be displayed.
 */
void MainWindow::help_menu_showing()
{
    CameraMDI   *active_child;
    QString     menuText;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child != nullptr)
    {
        if (active_child->is_series3())
        {
            menuText    = QStringLiteral("Save &Logs to SDCard");
        }
        else
        {
            menuText    = QStringLiteral("System Logs...");
        }
        ui->action_system_logs->setText(menuText);
    }
    // Make sure the enable state is set appropriately.
    setCameraLogMenuItem();

    // Currently we will assume the user can get to the user documentation to download it.
    ui->action_download_user_docs->setVisible(true);
    ui->action_download_user_docs->setEnabled(true);

    ui->action_user_docs->setVisible(TEST_USER_DOCS);
    ui->action_user_docs->setEnabled(TEST_USER_DOCS);

    // We are trying to run locally, so show the Teamviewer menu item.
    ui->action_enable_teamviewer->setVisible(g_local_control);

    // Check to see if the teamviewer daemon is running.
    if (!g_local_control || teamviewer_is_running())
    {
        // If it's running or we're not local, we don't need to re-enable it.
        ui->action_enable_teamviewer->setEnabled(false);
    }
    else
    {
        // It's not running, so let the user enable it.
        ui->action_enable_teamviewer->setEnabled(true);
    }
}

/*!
 * @brief User is requesting the camera to show/save its logs to the SD card if series 3
 * or to a local file system if newer camera system.
 */
void MainWindow::help_system_logs()
{
    CameraMDI   *active_child;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active child!";
        return;
    }
    if (active_child->is_offline())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Camera is offline!";
        return;
    }

    // If this is a series3 camera we do things like we always have done.
    if (active_child->is_series3())
    {
        // Ask the camera to save the logs to the SD card.
        active_child->write_camera_register(GIGE_REG_FILE_ACTION, GIGE_FA_SAVE_LOGS | GIGE_IMAGE_DEST_SDCARD);
    }
    else
    {
        // This camera is not a series 3 camera and the CameraLogs
        // can be used to save off log information.
        if (p_camera_logs_dock)
        {
            if (p_camera_logs_dock->initializeCameraLog(active_child))
            {
                p_camera_logs_dock->setVisible(true);
                p_camera_logs_dock->raise();
            }

            // Catch the finished signal so we can free up the memory later.
            //p_camera_logs_dock->connect(p_camera_logs_dock, SIGNAL(finished(int)), p_camera_logs_dock, SLOT(deleteLater()));
        }
    }
}

/*!
 * @brief User is requesting the download user documentation dialog.
 */
void MainWindow::help_download_user_docs()
{
    fiUserDocDialog dlg(this);
    dlg.exec();
}

/*!
 * @brief User is requesting the user documentation webpage.
 */
void MainWindow::help_user_docs()
{
    CameraMDI   *active_child;

    // Get the currently active child, if any.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active child!";
        return;
    }
    if (active_child->is_offline())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Camera is offline!";
        return;
    }
    // This does nothing right now.  What it will do in the future TBD.
    //QDesktopServices::openUrl(QUrl("https://www.fastecimaging.com/firmware-software/"));
#if 0
    // If this is a series3 camera we do things like we always have done.
    if (active_child->is_series3())
    {

    }
    else
    {

    }
#endif
}

/*!
 * @brief User wants to set the language to Chinese Simplified.
 */
void MainWindow::lang_chinese_simplified()
{
    // Remember the language selected.
    selected_language = LANGUAGE_CHINESE_SIMPLIFIED;

    // Check our menu selection.
    ui->action_chinese_simplified->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_zh_CN.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_zh_CN.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    // install the translators.
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief User wants to set the language to Chinese Traditional.
 */
void MainWindow::lang_chinese_traditional()
{
    // Remember the language selected.
    selected_language = LANGUAGE_CHINESE_TRADITIONAL;

    // Check our menu selection.
    ui->action_chinese_traditional->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_zh_TW.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_zh_TW.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    // install the translators.
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief User wants to set the language to English
 */
void MainWindow::lang_english()
{
    // Remember the language selected.
    selected_language = LANGUAGE_ENGLISH;

    // Check our menu selection.
    ui->action_english->setChecked(true);

    // Did we have a translator installed?
    if (translator_installed)
    {
        // Remove the translator to go back to english.
        qApp->removeTranslator(&qt_translator);
        qApp->removeTranslator(&app_translator);
        translator_installed = false;
    }
}

/*!
 * @brief User wants to set the language to French.
 */
void MainWindow::lang_french()
{
    // Remember the language selected.
    selected_language = LANGUAGE_FRENCH;

    // Check our menu selection.
    ui->action_french->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_fr.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_fr.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    // install the translators.
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief User wants to set the language to German.
 */
void MainWindow::lang_german()
{
    // Remember the language selected.
    selected_language = LANGUAGE_GERMAN;

    // Check our menu selection.
    ui->action_german->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_de.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_de.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    // install the translators.
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief User wants to set the language to Japanese
 */
void MainWindow::lang_japanese()
{
    // Remember the language selected.
    selected_language = LANGUAGE_JAPANESE;

    // Check our menu selection.
    ui->action_japanese->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_ja.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_ja.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    // install the translators.
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief User wants to set the language to Korean (Hangul)
 */
void MainWindow::lang_korean()
{
    // Remember the language selected.
    selected_language = LANGUAGE_KOREAN;

    // Check our menu selection.
    ui->action_korean->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_ko.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_ko.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief User wants to set the language to Russian
 */
void MainWindow::lang_russian()
{
    // Remember the language selected.
    selected_language = LANGUAGE_RUSSIAN;

    // Check our menu selection.
    ui->action_russian->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_ru.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_ru.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    // install the translators.
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief User wants to set the language to Spanish.
 */
void MainWindow::lang_spanish()
{
    // Remember the language selected.
    selected_language = LANGUAGE_SPANISH;

    // Check our menu selection.
    ui->action_spanish->setChecked(true);

    // Load the new strings.
    qt_translator.load(QString::fromUtf8("qt_es.qm"), QString::fromUtf8(":/translations"));
    app_translator.load(QString::fromUtf8("fasmotion_es.qm"), QString::fromUtf8(":/translations"));

    // Was a translator already installed? If there was a translator installed,
    // we've already modified the strings so nothing more to do here. Otherwise,
    // install the translators.
    if (!translator_installed)
    {
        // Install our translators.
        qApp->installTranslator(&qt_translator);
        qApp->installTranslator(&app_translator);
        translator_installed = true;
    }
}

/*!
 * @brief On the camera controller, read and return the model number.
 * @return A QString containing the camera controller model number.
 */
QString MainWindow::modelNumber()
{
    QString model_no;
    QSettings settings;

    // Allow for an override.
    settings.beginGroup("MainWindow");
    model_no = settings.value("model_no", QString()).toString();
    settings.endGroup();

#ifdef Q_OS_LINUX
    // Did we get an override?
    if (model_no.isEmpty())
    {
        // Here's the location of the serial number information.
        QFile controller_file(CONTROLLER_INFO_FILE);

        // Open it.
        if (controller_file.open(QIODevice::ReadOnly | QIODevice::Text))
        {
            QTextStream in(&controller_file);

            // Skip the first line (the serial number). We're interested in the
            // second line which is the model number..
            in.readLine();
            model_no = in.readLine();

            controller_file.close();
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No model number found for the controller!";
        }
    }
#endif

    // Return what we've found.
    return model_no;
}


/*!
 * @brief The user has selected the file menu and it's about to be displayed.
 */
void MainWindow::multicam_menu_showing()
{
    QString connect_string;

    // Sanity check.
    if (multicam_dock == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No multicam_dock!";
        return;
    }

    // Clear out the old menu.
    ui->menu_multicam->clear();

    // Update our menu.
    multicam_dock->build_context_menu(ui->menu_multicam);

    // Do we have an active child?
    if (active_mdi_child() != nullptr)
    {
        // Is there a selected entry in the "found camera" dock?
        if (p_camera_dock && (p_camera_dock->get_selected() >= 0))
        {
            FI_COMM::FI_CAMERA_INFO *sel_info;

            // Get the selected camera's information.
            sel_info = p_camera_dock->get_info_by_row(p_camera_dock->get_selected());
            if (sel_info && (active_mdi_child()->get_ip_addr() != sel_info->ip_addr))
            {
                QString cam_name = multicam_dock->camera_display(sel_info);
                connect_string = tr("Add ");
                connect_string.append(cam_name);
            }
        }
    }
    else
    {
        // Is there a selected entry in the "found camera" dock?
        if (p_camera_dock && (p_camera_dock->get_selected() >= 0))
        {
            FI_COMM::FI_CAMERA_INFO *sel_info;

            // Get the selected camera's information.
            sel_info = p_camera_dock->get_info_by_row(p_camera_dock->get_selected());
            if (sel_info != nullptr)
            {
                QString cam_name = multicam_dock->camera_display(sel_info);
                connect_string = tr("Add ");
                connect_string.append(cam_name);
            }
        }
    }

    // Do we want the "connect to" menu?
    if (!connect_string.isEmpty())
    {
        QList<QAction *>    action_list;

        // Get the list of actions.
        action_list = ui->menu_multicam->actions();

        // Sanity check.
        if (action_list.isEmpty())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No action list?!";
            return;
        }

        // Set the text for our action.
        ui->action_multicam->setText(connect_string);

        // Add the "connect to..." action to the top of the menu.
        ui->menu_multicam->insertAction(action_list.at(0), ui->action_multicam);
    }
}

/*!
 * @brief The user has selected the network configuration menu.
 */
void MainWindow::network_config()
{
    CameraMDI   *active_child;
    const FI_COMM::FI_CAMERA_INFO *cam_info;

    // Get the currently active child, if any, and tell it to go inactive.
    active_child = active_mdi_child();
    if (active_child == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active child. This menu shouldn't be enabled!";
        return;
    }

    // Get the camera specific information.
    cam_info = active_child->get_camera_info();
    if (cam_info == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera information?!";
        return;
    }

    // Create our dialog.
    ForceIPDialog dlg(this);

    // When entering the dialog via this path, we do not allow the user
    // to modify the MAC address.
    dlg.allow_mac_edit(false);

    // We can talk directly to this camera, so DHCP is allowed.
    dlg.set_dhcp_allowed(true);

    // Pass the camera info to the dialog.
    dlg.set_info(cam_info);

    // Run the dialog for the user.
    if (dlg.exec() == QDialog::Accepted)
    {
        bool    changes_sent;
        bool    resume_video(false);

        // Ask the camera window to pause it's streaming.
        if (cam_info->stream_running)
        {
            resume_video = true;
            active_child->pause_video_request();
        }

        // Process the changes requested by the user.
        changes_sent = change_camera_network(cam_info->ip_addr, dlg.dhcp_is_enabled(), dlg.get_ip(), dlg.get_net_mask(), dlg.get_gw(), dlg.get_mac());
        if (changes_sent)
        {
            // Close the window.
            close_mdi_child(active_child);
        }
        else
        {
            // Resume streaming if appropriate.
            if (resume_video)
                active_child->resume_video_request();
        }
    }
}

void MainWindow::open_raw_drive()
{
    fiRawDrive  raw_drive;
    QStringList fas_drives;
    QString     selected_drive;
    SelectDrive sel_dlg(this);
    PlaybackRawMDI  *child;
    QMdiSubWindow   *child_win_ptr;

    // Get the list.
    fas_drives = raw_drive.drives();
    QLOG_TRACE() << __PRETTY_FUNCTION__ << fas_drives;

    // Did we find any?
    if (fas_drives.isEmpty())
    {
        QMessageBox::information(this, tr("FasMotion"), tr("No Fastec Raw drives detected."));
        return;
    }

    // Give the dialog the list of drives.
    sel_dlg.set_drive_list(fas_drives);

    // Present the dialog to the user.
    if (sel_dlg.exec() == QDialog::Accepted)
    {
        // Get the user selected drive.
        selected_drive = sel_dlg.get_selected_drive();
    }
    else
    {
        // All done.
        return;
    }

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Drive selected:" << selected_drive;

    // Create our new child widget.
    child = new PlaybackRawMDI;
    if (child == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a new PlaybackRawMDI child!";
        return;
    }

    // Pass the video file to the playback object.
    if (!child->setFile(selected_drive))
    {
        delete child;
        return;
    }

    // Add the child to the MDI area.
    child_win_ptr = ui->mdi_area->addSubWindow(child);

    // Make sure the addSubWindow didn't have an error condition.
    if (child_win_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to addSubWindow for PlaybackRawMDI child!";
        return;
    }

    // Make sure there is a child window pointer.
    child->setWindow(child_win_ptr);

    // Make sure we delete the window when it's closed.
    child_win_ptr->setAttribute(Qt::WA_DeleteOnClose);

    // Do we want the child window maximized:
    // child->setWindowState( child->windowState() | Qt::WindowMaximized );
    child_win_ptr->show();

    // Or just make it the active one, still showing the others.
    emit ui->action_tile->trigger();
}

/*!
 * @brief Open the selected recent video.
 */
void MainWindow::open_recent_video()
{
    QAction *action = qobject_cast<QAction *>(sender());
    if (action != nullptr)
        load_file(action->data().toString());
}

/*!
 * @brief The user has selected the preferences menu item.
 */
void MainWindow::preferences()
{
    bool autoconnect;
    QString putty_path;
    QString default_user;
    QString default_pass;
    QString workgroup;
    QSettings settings;
    UserPreferences dlg;

    // Complain if we don't have a comm_ptr.
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
    autoconnect = settings.value("autoconnect", true).toBool();
    putty_path = settings.value("putty_path").toString();
    default_user = settings.value(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
    default_pass = settings.value(UserPreferences::SERIESNC_CAM_PASS_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
    workgroup = settings.value(UserPreferences::SERIESNC_CAM_WORKGROUP_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_WORKGROUP).toString();
    settings.endGroup();

    // Initialize the parameters.
    dlg.SetAutoConnect(g_local_control, g_local_control && autoconnect);
    dlg.SetFactoryEnable(factory_enabled);
    dlg.SetPuttyPath(putty_path);
    dlg.SetNetworkBuffers(comm_ptr->get_num_net_buffers());
    dlg.SetBackoffBuffers(comm_ptr->get_backoff_level());
    dlg.SetSaveStatsEnable(keep_statistics);
    dlg.SetLoggingLevel(QsLogging::Logger::instance().loggingLevel());
    dlg.SetMulticamEnable(multicam_enabled);
    dlg.SetPacketDelay(comm_ptr->get_default_pkt_delay());
    dlg.SetCredentials(default_user, default_pass, workgroup);

    // Show the dialog.
    if (dlg.exec() == QDialog::Accepted)
    {
        // Save the new user preferences.
        settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
        settings.setValue("autoconnect", dlg.AutoConnectEnabled());
        settings.setValue("factory_enable", dlg.FactoryMenuEnabled());
        settings.setValue("putty_path", dlg.GetPuttyPath());
        settings.setValue("backoff_bufs", dlg.GetBackoffBuffers());
        settings.setValue("network_bufs", dlg.GetNetworkBufferCount());
        settings.setValue("log_level", dlg.GetLoggingLevel());
        settings.setValue("multicam_support", dlg.MulticamEnabled());
        settings.setValue("save_stats", dlg.SaveStatsEnabled());

        dlg.GetCredentials(default_user, default_pass, workgroup);
        settings.setValue(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, default_user);
        settings.setValue(UserPreferences::SERIESNC_CAM_PASS_PREF_KEY, default_pass);
        settings.endGroup();

        // Update our updater.
        updater->SetFrequency(dlg.GetUpdateFreq());
        updater->SetAutomatic(dlg.AutoUpdateEnabled());

        // Update our logging level.
        if (QsLogging::Logger::instance().loggingLevel() != dlg.GetLoggingLevel())
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "log_level changed to:" << dlg.GetLoggingLevel();
            log_set_level(dlg.GetLoggingLevel());
        }

        // Did the factory enable value change?
        if (factory_enabled != dlg.FactoryMenuEnabled())
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "factory_enabled changed to:" << dlg.FactoryMenuEnabled();
            enable_factory(dlg.FactoryMenuEnabled());
        }

        if (putty_path != dlg.GetPuttyPath())
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "putty_path changed to:" << dlg.GetPuttyPath();
        }

        // Did the multicamera support change?
        if (multicam_enabled != dlg.MulticamEnabled())
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "multicam_enabled changed to:" << dlg.MulticamEnabled();
            enable_multicamera(dlg.MulticamEnabled());
        }

        // Did the statistics enable flag change?
        if (keep_statistics != dlg.SaveStatsEnabled())
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "keep_statistics changed to:" << dlg.SaveStatsEnabled();
            keep_statistics = dlg.SaveStatsEnabled();
        }

        // Did our number of network buffers change?
        if ((comm_ptr->get_num_net_buffers() != dlg.GetNetworkBufferCount()) || (comm_ptr->get_backoff_level() != dlg.GetBackoffBuffers()))
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "changing streaming buffers to:" << dlg.GetNetworkBufferCount() << dlg.GetBackoffBuffers();
            comm_ptr->set_backoff_level(dlg.GetBackoffBuffers());
            comm_ptr->set_net_buffers(dlg.GetNetworkBufferCount());

            // Adjust out system memory pool.
            FI_MEMPOOL::FiMemoryPool::ResizeMemoryPool((uint32_t)dlg.GetNetworkBufferCount());
        }
        if (comm_ptr->get_default_pkt_delay() != dlg.GetPacketDelay())
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "changing inter-packet delay to:" << dlg.GetPacketDelay();
            comm_ptr->set_default_pkt_delay(dlg.GetPacketDelay());
        }
    }
}

/*!
 * @brief On the camera controller, read and return the serial number.
 * @return A QString containing the camera controller serial number.
 */
QString MainWindow::serialNumber()
{
    QString serial_no;
    QSettings settings;

    // Allow for an override.
    settings.beginGroup("MainWindow");
    serial_no = settings.value("serial_no", QString()).toString();
    settings.endGroup();

#ifdef Q_OS_LINUX
    // Did we get an override?
    if (serial_no.isEmpty())
    {
        // Here's the location of the serial number information.
        QFile controller_file(CONTROLLER_INFO_FILE);

        // Open it.
        if (controller_file.open(QIODevice::ReadOnly | QIODevice::Text))
        {
            QTextStream in(&controller_file);

            // The first line is the serial number.
            serial_no = in.readLine();
            controller_file.close();
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No serial number found for the controller!";
        }
    }
#endif

    // Return what we've found.
    return serial_no;
}

/*!
 * @brief Activate one of our child windows.
 * @param window    Pointer to the window we want to activate.
 */
void MainWindow::set_active_subwindow(QWidget *window)
{
    if (!window)
        return;

    ui->mdi_area->setActiveSubWindow(qobject_cast<QMdiSubWindow *>(window));
    child_activated(qobject_cast<QMdiSubWindow *>(window));
}

void MainWindow::show()
{
    // Are we full screen?
    if (ui->action_full_screen->isChecked())
        showFullScreen();
    else
        QMainWindow::show();
}


void MainWindow::showFullScreen()
{
    QMainWindow::showFullScreen();
    ui->action_full_screen->setChecked(true);
}

/*!
 * @brief Show the update status window.
 *
 * If you want to specify the local camera (assuming we're running locally)
 * you can specify the CameraMDI or just pass a nullptr. For remote cameras
 * you must specify the CameraMDI.
 *
 * @param camera_ptr    Pointer to the CameraMDI object. May be nullptr for the local camera.
 * @param do_show       True to show the window, false to hide.
 */
void MainWindow::show_update_status(CameraMDI *camera_ptr)
{
    QSettings settings;
    UpdateStatus *status_dock;

    // Create our status dock.
    status_dock = new UpdateStatus(this);
    if (status_dock == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate the UpdateStatus object!";
        return;
    }
    addDockWidget(Qt::BottomDockWidgetArea, status_dock);
    status_dock->restoreGeometry(settings.value("MainWindow/statusGeometry").toByteArray());
    status_dock->setVisible(true);

    // Tell the status dock which camera to talk to.
    if (!status_dock->setCamera(camera_ptr))
    {
        // There was an error starting up the status dock.
        return;
    }

    // Start the status dock running.
    status_dock->start();
}

/*!
 * @brief The second stage of the shutdown process, we need to check to see
 *        if all the child windows have been closed. If they have, then we
 *        can close. Otherwise, try to shut them down again.
 */
void MainWindow::shutdown_process()
{
    // Do we have any MDI windows?
    if (ui->mdi_area->activeSubWindow() == nullptr)
    {
        // Close our window.
        close();
    }
    else
    {
        // Try closing them again.
        ui->mdi_area->closeAllSubWindows();
    }
}

void MainWindow::update_avail(int updates)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "updates:" << updates;

#ifdef Q_OS_LINUX
    // Is there an update available? First, check to see if we're running local. If so, we're going to
    // let the camera server deal with the update logic.
    if (!local_control())
#endif /* Q_OS_LINUX */
    {
        // We're not running on the camera control. See if there is a generic FasMotion update.
        if (((updates & FI_UPDATER::AVAIL_FM_UPDATE) == FI_UPDATER::AVAIL_FM_UPDATE) && (update_btn == nullptr))
        {
            // Sanity check. Compare our version with what we have for an update.
            if (QVersionNumber::compare(updater->GetVersionFM(), version_num()) > 0)
            {
                update_btn = new QPushButton(tr("Update Now!"));
                ui->main_statusbar->addPermanentWidget(update_btn);
                connect(update_btn, &QPushButton::clicked, this, &MainWindow::update_fasmotion);
            }
        }
    }
}

void MainWindow::update_downloaded(int which_update)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "which_update" << which_update;

    switch (which_update)
    {
    case FI_UPDATER::AVAIL_FM_UPDATE:
        // *INDENT-OFF*
        {
        // *INDENT-ON*

            // Not interested in the signal anymore.
            disconnect(updater.get(), &FiUpdater::download_complete, this, &MainWindow::update_downloaded);

            // Ask the user if they want to run the updater.
            if (QMessageBox::question(this, tr("Run FasMotion Updater"),
                                      tr("Do you want to run the FasMotion Updater?"),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
                == QMessageBox::No)
            {
                return;
            }

            // Do the update.
            do_update();

        // *INDENT-OFF*
        }
        // *INDENT-ON*
        break;
    default:
        break;
    }
}

void MainWindow::update_fasmotion()
{
    QString pkg_dest;

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Get the previous download path.
    pkg_dest = updater->GetUpdatePath();
    if (pkg_dest.isEmpty())
    {
        // Ask the user for the destination for the update package.
        pkg_dest = QFileDialog::getExistingDirectory(this, tr("Select a Directory to save the update package"), pkg_dest);
        if (pkg_dest.isEmpty())
            return;
        updater->SetUpdatePath(pkg_dest);
    }

    // Get the download complete signal to continue updating.
    connect(updater.get(), &FiUpdater::download_complete, this, &MainWindow::update_downloaded);

    // Make sure the file doesn't already exist.
    if (QFile::exists(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE)))
    {
        if (QMessageBox::question(this, tr("Overwrite Existing File"),
                                  tr("There already exists a file called %1. Overwrite?").arg(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE)),
                                  QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
            == QMessageBox::No)
        {
            // See if they want to run the updater.
            update_downloaded(FI_UPDATER::AVAIL_FM_UPDATE);
            return;
        }

        // Go ahead and blow away the file.
        QFile::remove(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE));
    }

    // Have the updater download the package.
    updater->UpdateDownload(FI_UPDATER::AVAIL_FM_UPDATE);
}

/*!
 * @brief Rebuild the Window menu to include or remove any child windows.
 */
void MainWindow::update_window_menu()
{
    int     index;
    QMenu   *window_menu;

    // Get a pointer that's a little easier to use.
    window_menu = ui->menu_window;

    // Now rebuild our "window" menu.
    window_menu->clear();
    window_menu->addAction(ui->action_full_screen);
    window_menu->addSeparator();
    window_menu->addAction(ui->action_close);
    window_menu->addAction(ui->action_close_all_windows);
    window_menu->addAction(ui->action_close_all_panels);
    window_menu->addSeparator();
    window_menu->addAction(ui->action_tile);
    window_menu->addAction(ui->action_cascade);
    window_menu->addSeparator();
    window_menu->addAction(ui->action_next_window);
    window_menu->addAction(ui->action_prev_window);
    window_menu->addSeparator();
    window_menu->addAction(ui->action_capture_settings);
    window_menu->addAction(ui->action_record_settings);
    window_menu->addAction(ui->action_record_controls);
    window_menu->addAction(ui->action_display_settings);
    window_menu->addAction(ui->action_lens_control);
    window_menu->addSeparator();

    // Add in the control panel(s).
    if (p_control_panel)
        window_menu->addAction(p_control_panel->toggleViewAction());
    if (p_camera_dock)
        window_menu->addAction(p_camera_dock->toggleViewAction());
    if (multicam_dock && multicam_enabled)
        window_menu->addAction(multicam_dock->toggleViewAction());
    if (p_metadata_dock)
        window_menu->addAction(p_metadata_dock->toggleViewAction());

    // Add our MDI windows, if we have any.
    window_menu->addAction(seperator_action);
    QList<QMdiSubWindow *> windows = ui->mdi_area->subWindowList();
    seperator_action->setVisible(!windows.isEmpty());

    // Loop through all of our child windows and add them to the menu.
    for (index = 0; index < windows.size(); ++index)
    {
        QAction     *action;
        QString     text;

        // See if this is a camera object.
        CameraMDI   *child = qobject_cast<CameraMDI *>(windows.at(index)->widget());
        if (child != nullptr)
        {
            if (index < 9)
                text = QString::fromUtf8("&%1 %2").arg(index + 1).arg(child->user_friendly_name());
            else
                text = QString::fromUtf8("%1 %2").arg(index + 1).arg(child->user_friendly_name());

            // Add (and create) an action for the menu.
            action = window_menu->addAction(text);
        }
        else
        {
            // Must be a playback object.
            PlaybackMDI *playback_child = qobject_cast<PlaybackMDI *>(windows.at(index)->widget());
            if (playback_child == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpect widget type!";
                action = nullptr;
            }
            else
            {
                // Retrieve the action from the playback window.
                action = playback_child->windowMenuAction();

                // Update the text for the action.
                if (index < 9)
                    text = QString::fromUtf8("&%1 %2").arg(index + 1).arg(playback_child->windowTitle());
                else
                    text = QString::fromUtf8("%1 %2").arg(index + 1).arg(playback_child->windowTitle());
                action->setText(text);

                // Add it to our window menu.
                window_menu->addAction(action);
            }
        }

        // Setup the action.
        if (action)
        {
            action->setCheckable(true);
            action->setChecked(windows.at(index)->widget() == ui->mdi_area->activeSubWindow()->widget());
            connect(action, &QAction::triggered, this, [this, index, windows]() { set_active_subwindow(windows.at(index)); });
        }
    }
}

/*!
 * @brief The user has selected the Histogram menu item.
 */
void MainWindow::view_histogram()
{
    // Make sure we have a histodock to show.
    if (histo_dock == nullptr)
        return;

    // Show or hide the dock.
    histo_dock->setVisible(ui->action_histogram->isChecked());

    // Are we showing the dock?
    if (ui->action_histogram->isChecked())
    {
        CameraMDI   *camera_ptr;

        // Get the histogram's camera.
        camera_ptr = histo_dock->get_camera();
        if (camera_ptr != nullptr)
        {
            uint32_t    camera_state;

            // Check to see if we're in review.
            camera_state = comm_ptr->get_reg(camera_ptr->get_ip_addr(), GIGE_REG_CAMERA_STATE);
            if (camera_state == GIGE_CAMERA_STATE_REVIEW)
            {
                // Ask the camera to update the frame.
                camera_ptr->update_frame();
            }
        }
    }
}

/*!
 * @brief Save the user settings to the non-volatile storage.
 */
void MainWindow::write_settings()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
    // Save the current settings.
    QSettings settings;

    settings.beginGroup("MainWindow");
    settings.setValue("winGeometry", saveGeometry());
    settings.setValue("winState", saveState());
    settings.setValue("language", selected_language);
    settings.setValue("multicam_support", multicam_enabled);

    if (p_control_panel != nullptr)
    {
        settings.setValue("cpGeometry", p_control_panel->saveGeometry());
        settings.setValue("cpOpen", p_control_panel->isVisible());
    }
    if (p_camera_logs_dock != nullptr)
    {
        settings.setValue("logGeometry", p_camera_logs_dock->saveGeometry());
        settings.setValue("logArea", dockWidgetArea(p_camera_logs_dock));
    }
    if (multicam_dock != nullptr)
    {
        settings.setValue("multiCamGeometry", multicam_dock->saveGeometry());
        settings.setValue("multiCamArea", dockWidgetArea(multicam_dock));
    }
    if (p_telemetry_dock != nullptr)
    {
        settings.setValue("telemetryGeometry", p_telemetry_dock->saveGeometry());
        settings.setValue("telemetryArea", dockWidgetArea(p_telemetry_dock));
    }
    if (histo_dock != nullptr)
    {
        settings.setValue("histoGeometry", histo_dock->saveGeometry());
        settings.setValue("histoArea", dockWidgetArea(histo_dock));
    }

    settings.endGroup();
}
