/*!
 * @file mainwindow.h
 * @brief This is the header file for the MainWindow class.
 *
 * The class follows the singleton pattern, allowing for one instance
 * of the MainWindow object.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef MAINWINDOW_H
#define MAINWINDOW_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <QMainWindow>
#include <QSharedPointer>
#include <QTranslator>
#include <QtNetwork>

#include <memory>

#include "logger.h"

#include "fi_saver.h"
#include "fi_mem_pool.h"

namespace Ui
{
class main_wnd;
}

namespace FASMOTION_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! A default bandwidth setting, in case we can't read it from the camera.
const uint32_t  DEFAULT_SSD_BANDWIDTH   = 180000000;

//! The FPGA's clock rate.
const int   FPGA_CLK_RATE               = 125000000;

//! A delay period to allow the main window to finish initializing.
const int   STARTUP_DELAY               = 1000;

//! The maximum number of recent videos we'll remember.
const int   MAX_RECENT_VIDEOS           = 10;

//! Create a reusable constant for an uninitialized uint32_t value using -1.
const uint32_t UNSET_UINT32             = (uint32_t) -1;

//! Declare offsets from QEvent::User for application specific events.
enum
{
    FI_CAP_XFER_EVENT,
    FI_CONVERT_EVENT,
    FI_COPY_EVENT,
    FI_DELETE_EVENT,
    FI_MULTICAM_EVENT,
    FI_REFRESH_EVENT,
    FI_RELEASE_EVENT,
    FI_STREAMER_EVENT,
    FI_RECTITME_CHANGED_EVENT
};

//! Declare ZOOM event values.
enum
{
    ZOOM_1_TO_1,
    ZOOM_IN,
    ZOOM_OUT,
    ZOOM_TO_FIT,
    ZOOM_TO_WINDOW,
    ZOOM_UNKNOWN
};

//! Declare enumerations for our supported languages.
enum SUPPORTED_LANGUAGES
{
    LANGUAGE_CHINESE_SIMPLIFIED,
    LANGUAGE_CHINESE_TRADITIONAL,
    LANGUAGE_ENGLISH,
    LANGUAGE_FRENCH,
    LANGUAGE_GERMAN,
    LANGUAGE_JAPANESE,
    LANGUAGE_KOREAN,
    LANGUAGE_RUSSIAN,
    LANGUAGE_SPANISH
};

}   // End of namespace FASMOTION_SETTINGS.


//----------------------------------------------------------------------------
// Forward declarations
//----------------------------------------------------------------------------
class QActionGroup;
class QLabel;
class QMdiSubWindow;
class QPushButton;
class QSessionManager;

class CameraDock;
class CameraLogs;
class CameraMDI;
class CameraTelemetry;
class ControlPanel;
class FI_comm;
class FiUpdater;
class HistoDock;
class MultiCamDock;
class PlaybackMetaInfo;
class UpdateStatus;
class VideoControls;

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class MainWindow : public QMainWindow
{
    Q_OBJECT

public:
    ~MainWindow();
    static QSharedPointer<MainWindow>& GetInstance();
    static QString camera_state_to_string(uint32_t camera_state);
    static void terminate();
    static void do_delete_main_window(MainWindow *p_main_window);


    // Put common sanity failure constant strings here so all objects can reference them.
    static const QString CAMERA_PTR_SANITY_FAILURE;
    static const QString COMMS_PTR_SANITY_FAILURE;

    CameraMDI       *active_mdi_child();
    void    change_mdi_child(QWidget *child_ptr);
    bool    change_camera_network(uint32_t cam_ip_addr, bool do_dhcp, QHostAddress& ip_addr, QHostAddress& subnet_mask, QHostAddress& gw_addr, const uint8_t *mac_addr);
    bool    check_avail_space(const QString& dest_path, uint32_t num_frames,  FI_SAVER::STREAM_TYPE streamType, uint32_t width, uint32_t height, uint32_t bit_depth, bool sensor_is_mono, uint64_t *space_calc = nullptr);
    void    check_zoom_actions(int zoom_action);
    void    close_mdi_child(CameraMDI *child_ptr);
    void    close_telemetry();
    bool    do_stats()                              { return keep_statistics; }
    void    do_update();
    void    enable_factory(bool do_enable);
    void    enable_zoom_actions(bool do_enable);
    QMdiSubWindow   *find_mdi_child(uint32_t ip_addr);
    bool    fs_supports_large_files(const QString& target_path);

    uint64_t    get_avail_space(const QString& target_path);
    QAction     *get_cam_autosave_setup();
    QAction     *get_cam_autosave_enable();
    QAction     *get_cam_connect();
    QAction     *get_cam_multicam();
    QAction     *get_cam_open_device();
    QAction     *get_cam_preferences();
    QAction     *get_cam_update();
    CameraDock  *get_camera_dock()                { return p_camera_dock; }
    QList<CameraMDI *>  get_children();
    FI_comm     *get_comm()       { return comm_ptr; }
    QAction     *get_config_black_level();
    QMenu       *get_config_menu();
    QAction     *get_config_wifi();
    ControlPanel    *get_control_panel()          { return p_control_panel; }
    QAction     *get_cp_close_all();
    QAction     *get_cp_display();
    QAction     *get_cp_lens();
    QAction     *get_cp_reccontrols();
    QAction     *get_cp_recsettings();
    QAction     *get_cp_storage();
    QAction     *get_event_marker_action();
    QAction     *get_export_action();
    bool        get_fs_name(const QString& target_path, QByteArray& fs_name_buf);
    HistoDock       *get_histo_dock()                 { return histo_dock; }
    QAction         *get_load();
    QDockWidget     *get_metadata_dock()              { return p_metadata_dock; }
    MultiCamDock    *get_multicam_dock()              { return multicam_dock; }
    QAction         *get_release();
    QAction         *get_release_all();
    QAction         *get_save();
    QAction         *get_save_all();
    QAction         *get_truncate();
    FiUpdater       *get_updater()                      { return updater.get(); }
    QAction         *get_view_aiming_cursor();
    QAction         *get_view_histogram();
    QAction         *get_view_info_cursor();
    QAction         *get_view_local_frametime();
    QMenu           *get_view_menu();
    QAction         *get_view_pfmd();
    QAction         *get_view_review_controls();
    QAction         *get_view_summary();

    QList<QAction *> get_camera_actions();

    void    init_comm(int net_buffers, int backoff_buffers, bool maintain_net_stats);
    void    init_image_pool(int net_buffers, int backoff_buffers);
    bool    is_factory_enabled()            { return factory_enabled; }
    bool    load_file(const QString& file_name, bool force_maximized = false);
    void    pause_discovery();
    void    pause_histogram();
    void    pause_video_streaming(CameraMDI *caller_ptr = nullptr);
    void    push_one_histogram();
    void    reflect_camera_state();
    void    reset_comm();
    void    resume_discovery();
    void    resume_histogram();
    void    resume_video_streaming();
    double  row_gain()                      { return row_gain_value; }
    int     row_offset()                    { return row_offset_value; }
    void    set_dock_size(QDockWidget *widget_ptr, int set_width, int set_height);
    void    set_metadata_info(PlaybackMetaInfo *meta_info);
    void    set_statusbar_message(const QString& str, int timeout_ms = 0);
    void    set_statusbar_text(const QString& str);
    void    show_update_status(CameraMDI *camera_ptr);
    void    update_review(CameraMDI *active_child);
    QString version();
    QVersionNumber version_num();

    // Override the QMainWindow version(s).
    void    show();
    void    showFullScreen();

    static bool do_autoconnect();
    static void enable_local_control(bool do_enable);
    static bool local_control();

signals:
    void    active_child_changed(QMdiSubWindow *window);
    void    connect_to_camera(int row_index);
    void    closing();

    void    camera_arm();
    void    camera_autosave_enable(bool checked);
    void    camera_autosave_setup();
    void    camera_live();
    void    camera_load();
    void    camera_menu_showing();
    void    camera_open_device();
    void    camera_power_down();
    void    camera_preferences();
    void    camera_reboot();
    void    camera_review();
    void    camera_samba();
    void    camera_save();
    void    camera_save_all();
    void    camera_snapshot();
    void    camera_trigger();
    void    camera_update();
    void    camera_user_docs();
    void    config_black_level();
    void    config_factory_reset();
    void    config_load();
    void    config_save();
    void    config_time_date();
    void    config_wifi();

    void    setup_event_markers();
    void    view_aiming_cursor();
    void    view_info_cursor();
    void    view_local_frametime();
    void    view_pfmd();
    void    view_review_controls();
    void    view_summary();

    void    playback_cap_closing();

    void    release_buffer();
    void    release_all_buffers();
    void    truncate_recording();

    void    save_settings();
    void    load_settings();

    void    rotate_left();
    void    rotate_right();

    void    zoom_1_to_1();
    void    zoom_in();
    void    zoom_out();
    void    zoom_to_fit();
    void    zoom_to_window();

private slots:
    void    about();
    void    child_activated(QMdiSubWindow *window);
    void    clear_recent_videos();
    void    commitData(QSessionManager& manager);
    void    connect_camera(int row_index);
    void    factory_black_level();
    void    factory_menu_showing();
    void    factory_open_console();
    void    factory_prnu();
    void    factory_raw_histo();
    void    factory_row_offset();
    void    factory_sensor_correction();
    void    factory_sys_info();
    void    factory_telemetry();
    void    file_menu_showing();
    void    file_open();
    void    find_cameras();
    void    full_screen();
    void    help_enable_teamviewer();
    void    help_menu_showing();
    void    help_system_logs();
    void    help_download_user_docs();
    void    help_user_docs();
    void    lang_chinese_simplified();
    void    lang_chinese_traditional();
    void    lang_english();
    void    lang_french();
    void    lang_german();
    void    lang_japanese();
    void    lang_korean();
    void    lang_russian();
    void    lang_spanish();
    void    multicam_menu_showing();
    void    network_config();
    void    open_raw_drive();
    void    open_recent_video();
    void    preferences();
    void    set_active_subwindow(QWidget *window);
    void    shutdown_process();
    void    update_avail(int updates);
    void    update_downloaded(int which_update);
    void    update_fasmotion();
    void    update_window_menu();
    void    view_histogram();
    void    write_settings();

    void    fusiondark_style();
    void    set_style();

protected:
    virtual void    changeEvent(QEvent *event);
    virtual void    closeEvent(QCloseEvent *event);

private:
    explicit MainWindow(QWidget *parent = nullptr);

    // This object is non-copyable and non-copy assignable.
    MainWindow(const MainWindow& i) = delete;
    MainWindow& operator=(const MainWindow& i) = delete;

    bool    add_cam_user_to_group_disk();
    bool    cam_user_in_group_disk();
    void    close_main_window();

    void    enable_multicamera(bool do_enable);
    void    init_connects();
    void    init_styles();
    void    init_updater();
    bool    load_cap(const QString& file_name, bool force_maximized = false);
    void    read_settings();

    void    setCameraLogMenuItem();

    void    setup_docks();
    void    setup_statusbar();
    bool    teamviewer_is_running();
    void    update_autosave();
    void    update_recent_menu();

    void    reflect_armed();
    void    reflect_autosave();
    void    reflect_idle();
    void    reflect_live();
    void    reflect_no_camera();
    void    reflect_review();
    void    reflect_triggered();

    void    restoreLanguage(int sel_language);

    void    update_style(QString sel_style);

    QString modelNumber();
    QString serialNumber();
    QMdiSubWindow   *create_mdi_child();

private:
    Ui::main_wnd    *ui;                        //!< Pointer to our UI element(s) class.

    QActionGroup    *language_group;            //!< Our language action group.
    ControlPanel    *p_control_panel;           //!< Pointer to the control panel dock.
    CameraDock      *p_camera_dock;             //!< Pointer to the found cameras dock.
    CameraLogs      *p_camera_logs_dock;        //!< Pointer to the camera logs dock.
    CameraTelemetry *p_telemetry_dock;          //!< Pointer to the telementry information dock.
    QDockWidget     *p_metadata_dock;           //!< Pointer to the metadata information dock.
    QLabel          *p_statusbar_label;         //!< Pointer to the statusbar text area.
    MultiCamDock    *multicam_dock;             //!< Pointer to the multi-camera dock widget.
    QAction         *recent_video_acts[ FASMOTION_SETTINGS::MAX_RECENT_VIDEOS ];        //!< The actions for the recent video menu.
    QAction         *recent_video_sep;          //!< Recent video seperator.
    QAction         *seperator_action;          //!< Our MDI child seperator.
    QActionGroup    *zoom_group;                //!< Our zoom action group.
    HistoDock       *histo_dock;                //!< Our histogram display.
    QVector<QAction *> style_items;             //!< Items dynmically added to our style menu.
    QAction         *active_style;              //!< Index into style_items for the currently active style.

    FI_comm         *comm_ptr;                  //!< Pointer to the communications class object.

    std::unique_ptr<FiUpdater> updater;         //!< Our updater object.
    QPushButton *update_btn;

    QTranslator     app_translator;             //!< The translator for our application strings.
    QTranslator     qt_translator;              //!< The translator for Qt's strings.
    bool        translator_installed;           //!< Flag to indicate if we've installed a translator.

    bool    factory_enabled;                    //!< True if the Factory menu is enabled.
    bool    keep_statistics;                    //!< True if we're keeping statistics, false otherwise.
    bool    in_shutdown;                        //!< True if we're in the process of shutting down.
    bool    multicam_enabled;                   //!< True if multicamera support is enabled.
    double  row_gain_value;                     //!< User supplied row gain for odd rows.
    int     row_offset_value;                   //!< User supplied row offset for image corrections.
    int     selected_language;                  //!< The language selected for our menus, etc.
    int     sensor_correction_start_row;        //!< User supplied start row for sensor corrections.

    //! Our singleton object.
    static QSharedPointer<MainWindow> main_window;
    static bool allocated_main_window_once;
};

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------
extern bool g_local_control;
extern bool maintain_net_stats;
extern int  num_net_buffers;
extern int  num_backoff_buffers;

#endif // MAINWINDOW_H
