/*!
 * @file multicam_dock.cpp
 * @brief The management and display object for multi-camerea control
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "multicam_dock.h"
#include "ui_multicam_dock.h"

#include <QContextMenuEvent>
#include <qendian.h>
#include <QFileDialog>
#include <QHostAddress>
#include <QMdiSubWindow>
#include <QMessageBox>
#include <QMenu>
#include <QMutexLocker>
#include <QPair>
#include <QTimer>

#include "camera_dock.h"
#include "camera_mdi.h"
#include "camera_review.h"
#include "camera_video.h"
#include "fi_msgbox.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"
#include "save_dialog.h"

// Avoid polluting the global namespace.
namespace MULTICAM_DOCK
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macro Definitions
//--------------------------------------------------------------------

//! Time to update our controls.
static const int UPDATE_PERIOD = 500;

//! Setup a type to associated a string with a CAMERA_DOCK::CD_xxx entry. We'll
//! use this to build up our context menu for how the user wants to display the
//! cameras in the list.
typedef QPair<QString, int>     DISPLAY_ENTRY;

}   // End of namespace MUTLICAM_DOCK.

// Make sure we use our namespace.
using namespace MULTICAM_DOCK;

//--------------------------------------------------------------------
// Variable declarations
//--------------------------------------------------------------------
QList<QPair<QString, int>> MultiCamDock::display_list;

/*!
 * @brief Our class constructor.
 */
MultiCamDock::MultiCamDock(QWidget *parent) :
    QDockWidget(parent),
    ui(new Ui::MultiCamDock)
{
    ui->setupUi(this);

    // Load our non-volatile settings.
    QSettings settings;
    settings.beginGroup("MultiCam_Dock");
    display_index = settings.value("display_index", CAMERA_DOCK::CD_USER_NAME).toInt();
    settings.endGroup();

    // Build up our display list.
    display_list.append(DISPLAY_ENTRY(tr("User Defined Name"), CAMERA_DOCK::CD_USER_NAME));
    display_list.append(DISPLAY_ENTRY(tr("IP"), CAMERA_DOCK::CD_IP_ADDR));
    display_list.append(DISPLAY_ENTRY(tr("Mac"), CAMERA_DOCK::CD_MAC_ADDR));

    // Connect up our signal/slots.
    connect(ui->action_add_current, &QAction::triggered, this, &MultiCamDock::add_current);
    connect(ui->action_add_all_open, &QAction::triggered, this, &MultiCamDock::add_all_open);
    connect(ui->action_delete, &QAction::triggered, this, &MultiCamDock::delete_camera);
    connect(ui->action_delete_all, &QAction::triggered, this, &MultiCamDock::delete_all_cameras);
    connect(ui->action_push_configuration, &QAction::triggered, this, &MultiCamDock::push_config);
    connect(ui->list_cam, &fiCamList::camera_dropped, this, &MultiCamDock::add_ip_addr);
    connect(ui->list_cam, &fiCamList::itemDoubleClicked, this, &MultiCamDock::camera_selected);

    // Setup our signals/slots for the multicamera state controls.
    connect(ui->action_mc_arm, &QAction::triggered, this, &MultiCamDock::arm_cameras);
    connect(ui->action_mc_autosave_enable, &QAction::triggered, this, &MultiCamDock::autosave);
    connect(ui->action_mc_live_mode, &QAction::triggered, this, &MultiCamDock::switch_to_live);
    connect(ui->action_mc_review_mode, &QAction::triggered, this, &MultiCamDock::switch_to_review);
    connect(ui->action_mc_trigger, &QAction::triggered, this, &MultiCamDock::trigger_cameras);

    connect(ui->tb_save, &QToolButton::clicked, this, &MultiCamDock::save_video_start);
    connect(ui->tb_sync_clip, &QToolButton::clicked, this, &MultiCamDock::sync_clip);
    connect(ui->action_save, &QAction::triggered, this, &MultiCamDock::save_video_start);

    // Setup our signals/slots for the multicamera review controls.
    connect(ui->tb_jump_first, &QToolButton::clicked, this, &MultiCamDock::jump_forward);
    connect(ui->tb_jump_last, &QToolButton::clicked, this, &MultiCamDock::jump_reverse);
    connect(ui->tb_play_forward, &QToolButton::clicked, this, &MultiCamDock::play_forward);
    connect(ui->tb_play_reverse, &QToolButton::clicked, this, &MultiCamDock::play_reverse);
    connect(ui->tb_step_forward, &QToolButton::clicked, this, &MultiCamDock::stepping_forward);
    connect(ui->tb_step_reverse, &QToolButton::clicked, this, &MultiCamDock::stepping_reverse);
    connect(ui->tb_stop, &QToolButton::clicked, this, &MultiCamDock::playback_stop);
    connect(ui->tb_trig_pos, &QToolButton::clicked, this, &MultiCamDock::jump_trigger);

    // Allocate our timer.
    update_timer = new QTimer(this);
    if (update_timer == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate update timer!";
        return;
    }
    connect(update_timer, &QTimer::timeout, this, &MultiCamDock::update_controls);

    // Disable/hide everything for now. The update will make sure all is as it should be.
    ui->tb_save->setVisible(false);
    ui->tb_sync_clip->setVisible(false);
    ui->frm_review->setVisible(false);
    ui->tb_lock->setEnabled(false);

    // We need to delay a little bit for the main window to finish it's setup.
    QTimer::singleShot(FASMOTION_SETTINGS::STARTUP_DELAY, this, &MultiCamDock::finish_setup);
}

/*!
 * @brief Our class destructor.
 */
MultiCamDock::~MultiCamDock()
{
    QSettings settings;
    settings.beginGroup("MultiCam_Dock");
    settings.setValue("display_index", display_index);
    settings.endGroup();

    delete ui;
}

/*!
 * @brief Add a camera, object pointed to by camera_ptr, to our multi-camera dock.
 *        If the camera is already in the list, then we'll make it the active one.
 *
 * @param camera_ptr    Pointer to the camera object we're interested in.
 */
void MultiCamDock::add_camera(CameraMDI *camera_ptr)
{
    QString cam_name;
    QList<QListWidgetItem *>    found_items;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return;
    }

    // Get the camera's display information. This might not be, strictly speaking,
    // the "name" of the camera, but it's how the camera is shown in the list.
    cam_name = camera_display(camera_ptr->get_camera_info());

    // Does this camera already exist in the list?
    found_items = ui->list_cam->findItems(cam_name, Qt::MatchExactly);
    if (found_items.isEmpty())
    {
        fiListItem *our_item;

        // Build up our list widget item.
        our_item = new fiListItem(cam_name, ui->list_cam);
        if (our_item == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate listItem for camera:" << cam_name;
            return;
        }

        // Associate the camera with the list item.
        our_item->set_camera(camera_ptr);

        // Mark the camera as "grouped".
        camera_ptr->set_grouped(true);

        // Put our newly created camera into the "found" list, so we can highlight it later.
        found_items.append(our_item);
    }

    // Highlight the camera.
    ui->list_cam->setCurrentItem(found_items.at(0));

    // Make sure our lock button is enabled if we have enough cameras to "lock".
    ui->tb_lock->setEnabled(allow_group());

    // Reflect the autosave enable state of the camera.
    ui->action_mc_autosave_enable->setChecked(camera_ptr->is_autosave_enabled());
}

/*!
 * @brief A camera is indicating that it has completed autosave. This isn't an indication
 *        of success or failure, just completion.
 *
 * @param camera_ptr    The camera which finished autosave.
 */
void MultiCamDock::autosave_complete(CameraMDI *camera_ptr)
{
    // Sanity check.
    if (autosave_queue.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpectedly empty autosave queue!";
        return;
    }
    if (camera_ptr != autosave_queue.at(0))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "camera:" << camera_ptr->get_ip_addr_pretty() << "is not the top of the list!";
        return;
    }

    QLOG_TRACE() << __PRETTY_FUNCTION__ << camera_ptr->get_ip_addr_pretty() << "autosave list count:" << autosave_queue.count();

    // Remove the camera from our queue.
    autosave_dequeue(camera_ptr);

    // Any more on the list?
    if (!autosave_queue.isEmpty())
    {
        {
            QStringList str_out;
            for (int i = 0; i < autosave_queue.count(); ++i)
            {
                str_out << autosave_queue.at(i)->get_ip_addr_pretty();
            }

            QLOG_TRACE() << __PRETTY_FUNCTION__ << "Cameras in queue:" << str_out;
        }

        // Let the first entry start saving.
        autosave_queue.at(0)->allow_multicam_autosave();
    }
}

/*!
 * @brief A camera wants to remove itself from the autosave queue..
 *
 * @param camera_ptr    The camera which finished autosave.
 */
void MultiCamDock::autosave_dequeue(CameraMDI *camera_ptr)
{
    // Lock access to the autosave queue.
    QMutexLocker list_lock(&autosave_mutex);

    QLOG_TRACE() << __PRETTY_FUNCTION__ << camera_ptr->get_ip_addr_pretty() << "autosave list count:" << autosave_queue.count();

    // Check to see if there's anything to do.
    if (autosave_queue.isEmpty())
        return;

    // Remove all occurrences of this camera (should be at most one).
    autosave_queue.removeAll(camera_ptr);

    // Any more on the list?
    if (autosave_queue.isEmpty())
    {
        // That was the last one. Resume streaming.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "queue is empty.";
        MainWindow::GetInstance()->resume_video_streaming();
    }
}

void MultiCamDock::autosave(bool checked)
{
    bool        is_enabled = true;
    int         index;
    CameraMDI   *cam_ptr;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "checked:" << checked;

    // Walk through the cameras in the group and enable/disable autosave.
    for (index = 0; index < ui->list_cam->count(); ++index)
    {
        fiListItem *my_item;

        // Get the item list.
        my_item = (fiListItem *)ui->list_cam->item(index);
        if (my_item == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "A nullptr item in the list at index:" << index;
            continue;
        }

        // Get a pointer to the camera.
        cam_ptr = my_item->get_camera();
        if (cam_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
            return;
        }

        // Are we enabling?
        cam_ptr->do_camera_autosave_enable(checked);

        // It only takes one...
        if (!cam_ptr->is_autosave_enabled())
            is_enabled = false;
    }

    // Update the group autosave checkbox.
    ui->action_mc_autosave_enable->setChecked(is_enabled);
}

/*!
 * @brief A camera is telling us that it is ready to start the autosave to
 *        the local workstation. If this is the first one, then we immediately
 *        allow it to start the operation.
 *
 * @param camera_ptr    Pointer to the camera that wants to start autosaving.
 * @return True if the camera can start saving immediately, false if it must wait.
 */
bool MultiCamDock::autosave_ready(CameraMDI *camera_ptr)
{
    bool    list_is_empty;
    int     i;

    // Lock access to the autosave queue.
    QMutexLocker list_lock(&autosave_mutex);

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Null camera_ptr?!";
        return false;
    }

    QLOG_TRACE() << __PRETTY_FUNCTION__ << camera_ptr->get_ip_addr_pretty();

    // Do we need to kick start the save process?
    list_is_empty = autosave_queue.isEmpty();

    // Check to see if this camera is already in the list.
    for (i = 0; i < autosave_queue.count(); ++i)
    {
        // Is this a match? If it is a match, and it's not
        // element 0, then this camera has to wait.
        if (camera_ptr == autosave_queue.at(i))
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "Camera found in queue at position:" << i;
            return (i == 0);
        }
    }

    // Add this camera to the list.
    autosave_queue.append(camera_ptr);

    {
        QStringList str_out;
        for (i = 0; i < autosave_queue.count(); ++i)
        {
            str_out << autosave_queue.at(i)->get_ip_addr_pretty();
        }

        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Cameras in queue:" << str_out;
    }

    // If this is the first in the list, allow it to start saving immediately.
    return list_is_empty;
}

/*!
 * @brief A utility function to build up the context menu.
 */
void MultiCamDock::build_context_menu(QMenu *context_menu)
{
    QString menu_str;
    QMenu   *display_menu;
    FI_comm     *comm_ptr;
    CameraMDI   *active_cam;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the active camera, if there is one.
    active_cam = MainWindow::GetInstance()->active_mdi_child();

    // Is there an active camera?
    if (active_cam != nullptr)
    {
        // Make sure this camera isn't already in the list.
        QString cam_name = camera_display(active_cam->get_camera_info());
        if (ui->list_cam->findItems(cam_name, Qt::MatchExactly).count() == 0)
        {
            menu_str = tr("Add ");
            menu_str.append(cam_name);
            ui->action_add_current->setText(menu_str);
            context_menu->addAction(ui->action_add_current);
        }
    }
    context_menu->addAction(ui->action_add_all_open);
    context_menu->addSeparator();

    if (ui->tb_save->isVisible())
    {
        context_menu->addAction(ui->action_save);
        context_menu->addSeparator();
    }
    context_menu->addAction(ui->action_delete);
    context_menu->addAction(ui->action_delete_all);
    context_menu->addSeparator();

    // Add in our sub menu.
    display_menu = context_menu->addMenu(tr("Display Using"));
    display_actions.clear();
    if (active_cam != nullptr)
    {
        for (int i = 0; i < display_list.count(); ++i)
        {
            QAction *menu_act;

            // Add this entry as a menu action.
            menu_act = display_menu->addAction(display_list.at(i).first);
            menu_act->setIconVisibleInMenu(false);
            menu_act->setCheckable(true);
            connect(menu_act, &QAction::triggered, this, &MultiCamDock::display_update);

            // Is this our displayed type?
            if (display_list.at(i).second == display_index)
                menu_act->setChecked(true);

            // Save the action to our list.
            display_actions.append(menu_act);
        }
    }
    context_menu->addSeparator();
    context_menu->addAction(ui->action_push_configuration);
    context_menu->addSeparator();
    context_menu->addAction(ui->action_mc_arm);
    context_menu->addAction(ui->action_mc_trigger);
    context_menu->addAction(ui->action_mc_live_mode);
    context_menu->addAction(ui->action_mc_review_mode);
    context_menu->addAction(ui->action_mc_autosave_enable);

    // Determine which menu items should be enabled/disabled.
    ui->action_add_current->setEnabled(active_cam != nullptr);
    ui->action_add_all_open->setEnabled(active_cam != nullptr);
    ui->action_mc_arm->setEnabled(false);
    ui->action_mc_trigger->setEnabled(false);
    ui->action_mc_live_mode->setEnabled(false);
    ui->action_mc_review_mode->setEnabled(false);
    ui->action_mc_autosave_enable->setEnabled(false);
    display_menu->setEnabled(active_cam != nullptr);

    // Are there any entries in the list?
    ui->action_delete->setEnabled(ui->list_cam->count() > 0);
    ui->action_delete_all->setEnabled(ui->list_cam->count() > 0);

    // Any entries selected?
    ui->action_push_configuration->setEnabled(ui->list_cam->selectedItems().count() > 0);

    // Update our controls.
    update_controls();
}

/*!
 * @brief The user may choose to display the camera in the list using a number of
 *        possible identifiers: the camera "name", the IP address or the MAC address.
 *        This returns the proper value.
 *
 * @param camera_ptr    Pointer to the camera object we're interested in.
 * @return The identifier for the camera which is currently used in the list.
 */
QString MultiCamDock::camera_display(const FI_COMM::FI_CAMERA_INFO *camera_info)
{
    QString         ret_str;
    QHostAddress    cam_addr;
    QLatin1Char     zero_fill('0');

    // Sanity check.
    if (camera_info == nullptr)
        goto cd_bail;

    // How are we displaying the camera?
    switch (display_index)
    {
    case CAMERA_DOCK::CD_IP_ADDR:
        cam_addr.setAddress(qFromBigEndian(camera_info->ip_addr));
        ret_str = cam_addr.toString();
        break;
    case CAMERA_DOCK::CD_MAC_ADDR:
        ret_str = QString::fromUtf8("%1:%2:%3:%4:%5:%6").
                  arg(camera_info->mac_addr[0], 2, 16, zero_fill).
                  arg(camera_info->mac_addr[1], 2, 16, zero_fill).
                  arg(camera_info->mac_addr[2], 2, 16, zero_fill).
                  arg(camera_info->mac_addr[3], 2, 16, zero_fill).
                  arg(camera_info->mac_addr[4], 2, 16, zero_fill).
                  arg(camera_info->mac_addr[5], 2, 16, zero_fill);
        break;
    default:
    case CAMERA_DOCK::CD_USER_NAME:
        ret_str = QString::fromUtf8(camera_info->user_name);
        break;
    }

cd_bail:
    return ret_str;
}

/*!
 * @brief Return the list of cameras in the group.
 */
const fiCamList *MultiCamDock::get_group_list()
{
    return ui->list_cam;
}

/*!
 * @brief Determine if the camera object "camera_ptr" is a member of the group.
 *
 * @param camera_ptr        Pointer to the camera object we're interested in.
 * @return True if the camera is in the group, false otherwise.
 */
bool MultiCamDock::is_camera_grouped(CameraMDI *camera_ptr)
{
    QString cam_name;
    QList<QListWidgetItem *>    found_items;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return false;
    }

    // Get the camera's name.
    cam_name = camera_display(camera_ptr->get_camera_info());

    // Find the entry.
    found_items = ui->list_cam->findItems(cam_name, Qt::MatchExactly);

    // Return true if we found it.
    return (found_items.count() > 0);
}

/*!
 * @brief Check if the multicamera group is locked.
 *
 * If the group is locked, then all settings will be sent to
 * all cameras in the group.
 *
 * @return True if the control is enabled and checked.
 */
bool MultiCamDock::is_group_locked()
{
    // The group is locked if the control is enabled and checked.
    return (ui->tb_lock->isEnabled() && ui->tb_lock->isChecked());
}

/*!
 * @brief Remove the specified camera from the list of cameras.
 * @param camera_ptr    Pointer to the camera object we want removed from the list.
 */
void MultiCamDock::remove_camera(CameraMDI *camera_ptr)
{
    int     i;
    QString cam_name;
    QList<QListWidgetItem *>    found_items;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Null camera_ptr!";
        return;
    }

    // Get the camera's name.
    cam_name = camera_display(camera_ptr->get_camera_info());

    // Find the entry.
    found_items = ui->list_cam->findItems(cam_name, Qt::MatchExactly);
    for (i = 0; i < found_items.count(); ++i)
    {
        int my_row;
        QListWidgetItem *my_item;

        // Get the item widget at the specific row.
        my_row = ui->list_cam->row(found_items.at(i));
        my_item = ui->list_cam->takeItem(my_row);

        // Delete the widget.
        delete my_item;

        // Tell the camera that it's no longer in the group.
        camera_ptr->set_grouped(false);
    }

    // Make sure our lock button is disabled if there aren't enough to "lock".
    ui->tb_lock->setEnabled(allow_group());
}

void MultiCamDock::save_camera_finish(void)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Under construction";
}

void MultiCamDock::save_camera_start(void)
{
    int         old_clip_end;
    int         old_clip_start;
    uint32_t    old_ref_frame;
    uint32_t    num_frames;
    uint32_t    trig_frame;
    FI_comm     *comm_ptr;
    CameraMDI   *cam_ptr;
    fiListItem  *my_item;
    SaveVideo   *saver_ptr;
    CameraVideo *video_player;

    // Check to see that we have a camera in our list.
    if (save_list.isEmpty())
        return;

    // Grab the first item in the list. We need to keep the item ON the list
    // because we'll need it when the save is complete.
    my_item = (fiListItem *)save_list.at(0);
    if (my_item == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No items?!";
        return;
    }
    cam_ptr = my_item->get_camera();
    if (cam_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get our saver object.
    saver_ptr = cam_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
        return;
    }

    // Before we go mucking with things, save away the "current" values.
    old_clip_end = save_info.clip_end;
    old_clip_start = save_info.clip_start;
    old_ref_frame = save_info.ref_frame;

    // How many frames did we actually capture.
    num_frames = comm_ptr->get_reg(cam_ptr->get_ip_addr(), GIGE_REG_PART_NUM_FRAMES);

    // Read what the trigger actually is.
    trig_frame = comm_ptr->get_reg(cam_ptr->get_ip_addr(), GIGE_REG_PART_TRIG_FRAME);

    // Are we using the clip range on the camera?
    if (save_info.use_camera_range)
    {
        // Get the range set on the camera.
        save_info.ref_frame = trig_frame;
        save_info.clip_end = (int)(cam_ptr->get_clip_end() - trig_frame);
        save_info.clip_start = (int)(cam_ptr->get_clip_start() - trig_frame);
    }
    else
    {
        // Make sure the trigger is a match.
        save_info.ref_frame = trig_frame;

        // Make sure our clip ending doesn't exceed the number of frames we've captured.
        if ((trig_frame + (uint32_t)save_info.clip_end) >= num_frames)
            save_info.clip_end = (int)((num_frames - 1) - trig_frame);

        // Make sure our clip start is valid. If the start point plus the trigger is negative,
        // then that will be a very large unsigned value.
        if ((trig_frame + (uint32_t)save_info.clip_start) >= num_frames)
            save_info.clip_start = -((int)trig_frame);
    }

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "camera:" << camera_display(cam_ptr->get_camera_info()) << "clip_start:" << save_info.clip_start << "clip_end:" << save_info.clip_end;

    // Sanity check our clip range.
    if (save_info.clip_start >= save_info.clip_end)
    {
        QString text_str;
        QString informative_str;

        text_str = tr("Error saving from camera %1").arg(camera_display(cam_ptr->get_camera_info()));
        informative_str = tr("The camera does not support the range of frames specified [start: %1, end: %2 ]").arg(old_clip_start).arg(old_clip_end);

        QLOG_ERROR() << __PRETTY_FUNCTION__ <<  text_str << informative_str << "use_camera_range:" << save_info.use_camera_range
                     << "trig_frame:" << trig_frame << "num_frames:" << num_frames;
        fiMsgBox::critical(cam_ptr, text_str, informative_str);
        return;
    }

    // Have the saver object configure itself.
    if (!saver_ptr->configure(save_info))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Camera:" << cam_ptr->get_ip_addr_pretty() << "Configuration failed.";
        return;
    }

    // Connect the saver's signal so we know when the save completes.
    connect(cam_ptr, &CameraMDI::save_finished, this, &MultiCamDock::file_save_complete, Qt::QueuedConnection);

    // Restore our settings, in case they were modified.
    save_info.clip_end = old_clip_end;
    save_info.clip_start = old_clip_start;
    save_info.ref_frame = old_ref_frame;

    // Do we need to merge the overlay information?
    video_player = cam_ptr->get_player();
    if (save_info.merge_overlays)
        video_player->render_start();

    // Have the camera object fire off the save.
    cam_ptr->start_local_save(true);
}

/*!
 * @brief Translate the UI elements for this object.
 */
void MultiCamDock::translate_ui(void)
{
    // Have the UI object perform it's translations.
    ui->retranslateUi(this);
}

/*!
 * @brief Update the displayed list of cameras. This would happen if the user
 *        changed how to display the cameras identifier.
 */
void MultiCamDock::update_list(void)
{
    // Turn off sorting.
    ui->list_cam->setSortingEnabled(false);

    // Update all elements in the list to reflect the new naming convention.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;
        my_item = (fiListItem *)ui->list_cam->item(i);
        my_item->setText(camera_display(my_item->get_camera()->get_camera_info()));
    }

    // Turn sorting back on.
    ui->list_cam->setSortingEnabled(true);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void MultiCamDock::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    QDockWidget::changeEvent(event);
}

void MultiCamDock::do_dock()
{
    setFloating(false);
}

/*!
 * @brief Handle a right click from the user - this will pop-up a context menu.
 * @param event_ptr     Pointer to the right-click event.
 */
void MultiCamDock::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QMenu   context_menu(this);

    // Build up our menu.
    build_context_menu(&context_menu);

    // Add a menu to re-dock the panel.
    if (isFloating())
    {
        context_menu.addSeparator();
        context_menu.addAction(tr("Dock"), this, &MultiCamDock::do_dock);
    }

    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

/*!
 * @brief The receive method for the MULTICAM_SAVE_EVENT.
 * @param event     Pointer to the event structure.
 */
void MultiCamDock::customEvent(QEvent *event)
{
    // Only respond to our custom event
    if (event->type() == MULTICAM_SAVE_EVENT)
        save_camera_start();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Take all the open cameras (i.e. cameras which have a window in the UI) and
 *        add them to our dock.
 */
void MultiCamDock::add_all_open(void)
{
    int i;
    QList<CameraMDI *>  camera_list;

    // Get all the open cameras.
    camera_list = MainWindow::GetInstance()->get_children();
    for (i = 0; i < camera_list.count(); ++i)
        add_camera(camera_list[ i ]);
}

/*!
 * @brief Add the camera referenced by the ip_addr parameter to our dock.
 * @param ip_addr   The IP address of the camera we're interested in.
 */
void MultiCamDock::add_ip_addr(uint32_t ip_addr)
{
    QMdiSubWindow   *mdi_ptr;
    CameraMDI       *camera_ptr;

    // See if we have an open window.
    mdi_ptr = MainWindow::GetInstance()->find_mdi_child(ip_addr);
    if (mdi_ptr == nullptr)
    {
        // Make sure we catch the signal from the main window when the camera is activated.
        connect(MainWindow::GetInstance().get(), &MainWindow::active_child_changed, this, &MultiCamDock::camera_activated, Qt::UniqueConnection);

        // Ask the dock to open the camera.
        MainWindow::GetInstance()->get_camera_dock()->open_selected_camera();
        return;
    }

    // Get the camera pointer from the window.
    camera_ptr = qobject_cast<CameraMDI *>(mdi_ptr->widget());
    if (camera_ptr == nullptr)
        return;

    // Add this guy to our list.
    add_camera(camera_ptr);
}

/*!
 * @brief Add the camera associated with the currently active window to our dock.
 */
void MultiCamDock::add_current(void)
{
    CameraMDI   *active_cam;

    // Get the currently active camera.
    active_cam = MainWindow::GetInstance()->active_mdi_child();
    if (active_cam == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No active camera?!";
        return;
    }

    // Add the camera to the list.
    add_camera(active_cam);
}

/*!
 * @brief Arm all the cameras in the dock.
 */
void MultiCamDock::arm_cameras(void)
{
    int         index;
    CameraMDI   *cam_ptr;
    bool        user_queried;
    QList<bool> cam_discards;

    // Haven't asked the user yet.
    user_queried = false;

    // We make two passses here - the first pass is to make sure the user
    // REALLY wants to arm the camera(s).
    for (index = 0; index < ui->list_cam->count(); ++index)
    {
        bool        allow_discard;
        fiListItem  *my_item;

        // Get the item list.
        my_item = (fiListItem *)ui->list_cam->item(index);
        if (my_item == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "A nullptr item in the list at index:" << index;
            continue;
        }

        // Get a pointer to the camera.
        cam_ptr = my_item->get_camera();
        if (cam_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
            return;
        }

        // Query the camera to see if we're safe to ARM/de-arm.
        if (!cam_ptr->safe_to_arm(user_queried, allow_discard))
        {
            // We cannot arm/de-arm the camera right now.
            return;
        }

        // Save our discard information.
        cam_discards.append(allow_discard);
    }

    // If we're here then we're OK to arm the camera(s).
    for (index = 0; index < ui->list_cam->count(); ++index)
    {
        fiListItem *my_item;

        // Get the item list.
        my_item = (fiListItem *)ui->list_cam->item(index);
        if (my_item == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "A nullptr item in the list at index:" << index;
            continue;
        }

        // Get a pointer to the camera.
        cam_ptr = my_item->get_camera();
        if (cam_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
            return;
        }

        // Have the camera arm.
        cam_ptr->process_arm(user_queried, cam_discards.at(index));
    }
}

/*!
 * @brief Mark a camera in the list as being on- or off-line. This will happen when we
 *        push a new configuration to the camera, for example.
 *
 * @param ip_addr   IP Address of the camera we're interested in.
 * @param offline   True if the camera is off-line, false if it's on-line.
 */
void MultiCamDock::camera_offline(uint32_t ip_addr, bool offline)
{
    int index;

    // Loop through all the cameras in our list, and see if we have a match.
    for (index = 0; index < ui->list_cam->count(); ++index)
    {
        fiListItem  *my_item;
        CameraMDI   *cam_ptr;

        // Get the item at this index.
        my_item = (fiListItem *)ui->list_cam->item(index);
        if (my_item == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "A nullptr item in the list at index:" << index;
            continue;
        }

        // Get a pointer to the camera.
        cam_ptr = my_item->get_camera();

        // Do we have a match?
        if (cam_ptr && (cam_ptr->get_ip_addr() == ip_addr))
        {
            QFont   item_font;

            // Is this camera going off-line or coming back on-line? Setup
            // our font appropriately.
            item_font.setStrikeOut(offline);
            my_item->setFont(item_font);
        }
    }
}

/*!
 * @brief This gets called when a camera is activated, which would happen if we had to
 *        request the window be opened to the camera as a result of a drag-drop.
 *
 * @param child_win_ptr Pointer to the MDI window for the camera.
 */
void MultiCamDock::camera_activated(QMdiSubWindow *child_win_ptr)
{
    CameraMDI   *camera_ptr;

    // Sanity check.
    if (child_win_ptr == nullptr)
        return;

    // Get the camera object.
    camera_ptr = qobject_cast<CameraMDI *>(child_win_ptr->widget());
    if (camera_ptr == nullptr)
        return;

    // Don't need the signal anymore.
    disconnect(MainWindow::GetInstance().get(), &MainWindow::active_child_changed, this, &MultiCamDock::camera_activated);

    // Add the camera to our list.
    add_camera(camera_ptr);
}

/*!
 * @brief This slot is called when the user double clicks an item in the list.
 * @param sel_item  A pointer to the selected item.
 */
void MultiCamDock::camera_selected(QListWidgetItem *sel_item)
{
    fiListItem  *my_item;
    CameraMDI   *cam_ptr;

    // Get the item at this index.
    my_item = (fiListItem *)sel_item;
    if (my_item == nullptr)
        return;

    // Get a pointer to the camera.
    cam_ptr = my_item->get_camera();
    if (cam_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera pointer?!";
        return;
    }

    // Is the camera on-line?
    if (!cam_ptr->get_camera_info()->offline)
    {
        // Make this the new active window.
        MainWindow::GetInstance()->change_mdi_child(cam_ptr);
    }
}

/*!
 * @brief Remove all cameras from the dock.
 */
void MultiCamDock::delete_all_cameras(void)
{
    // Delete all items in the list.
    while (ui->list_cam->count() > 0)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(0);
        if (my_item != nullptr)
            remove_camera(my_item->get_camera());
    }
}

/*!
 * @brief Remove the selected camera(s) from the list.
 */
void MultiCamDock::delete_camera(void)
{
    QList<QListWidgetItem *> sel_items;

    // Get the currently selected item.
    sel_items = ui->list_cam->selectedItems();
    while (sel_items.count() > 0)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)sel_items.takeFirst();
        if (my_item != nullptr)
            remove_camera(my_item->get_camera());
    }
}

/*!
 * @brief Change how the cameras are displayed.
 */
void MultiCamDock::display_update(void)
{
    QAction *triggered_act;

    // Get the action which caused us to be called.
    triggered_act = qobject_cast<QAction *>(sender());
    if (triggered_act && triggered_act->isChecked())
    {
        int     index;

        // Loop through all of our display actions to find if this is one.
        for (index = 0; index < display_actions.count(); ++index)
        {
            // Find which action this was, so we can get the index.
            if (triggered_act == display_actions.at(index))
            {
                // Save the index.
                display_index = display_list.at(index).second;

                // Update the list.
                update_list();
            }
        }
    }
}

/*!
 * @brief Move to the next camera in the list.
 */
void MultiCamDock::file_save_complete(void)
{
    CameraMDI   *cam_ptr = qobject_cast<CameraMDI *>(sender());
    fiListItem  *my_item;

    if (cam_ptr != nullptr)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty();
    }
    else
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Unknown sender.";
    }

    // Check to see that we have a camera in our list.
    if (save_list.isEmpty())
        return;

    // We're done with the camera on top of the list, so pop it off the stack.
    my_item = (fiListItem *)save_list.takeFirst();
    if (my_item == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No items?!";
        return;
    }
    cam_ptr = my_item->get_camera();
    if (cam_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Disconnect from the saver's signal.
    disconnect(cam_ptr, &CameraMDI::save_finished, this, &MultiCamDock::file_save_complete);

    // Now run through the loop again.
    QApplication::postEvent(this, new QEvent(MULTICAM_SAVE_EVENT));
}

/*!
 * @brief After a brief delay to let the main window finish it's setup,
 *        this slot is invoked so we can finish our setup.
 */
void MultiCamDock::finish_setup(void)
{
    FI_comm     *comm_ptr;

    // Make sure we have a main window.
    if (MainWindow::GetInstance().isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No main window?!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Catch any on/off-line status changes.
    connect(comm_ptr, &FI_comm::camera_offline_status_changing, this, &MultiCamDock::camera_offline, Qt::QueuedConnection);

    // Enable multicamera mode for the camera controls.
    ui->cam_actions->enable_multicam_mode();
    ui->cam_actions->set_arm_action(ui->action_mc_arm);
    ui->cam_actions->set_autosave_action(ui->action_mc_autosave_enable);
    ui->cam_actions->set_live_action(ui->action_mc_live_mode);
    ui->cam_actions->set_review_action(ui->action_mc_review_mode);
    ui->cam_actions->set_trigger_action(ui->action_mc_trigger);

    // Fire off our update timer.
    update_timer->start(UPDATE_PERIOD);
}

void MultiCamDock::jump_forward(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    control_ptr->jump_forward();
                }
            }
        }
    }
}

void MultiCamDock::jump_reverse(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    control_ptr->jump_reverse();
                }
            }
        }
    }
}

void MultiCamDock::jump_trigger(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    control_ptr->jump_trigger();
                }
            }
        }
    }
}

void MultiCamDock::play_forward(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    emit control_ptr->play_forward();
                }
            }
        }
    }
}

void MultiCamDock::play_reverse(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    emit control_ptr->play_reverse();
                }
            }
        }
    }
}

void MultiCamDock::playback_stop(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    control_ptr->playback_stop();
                }
            }
        }
    }
}

void MultiCamDock::stepping_forward(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    control_ptr->stepping_forward();
                }
            }
        }
    }
}

void MultiCamDock::stepping_reverse(void)
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Index" << i << "has no Camera pointer!";
            }

            // Make sure the camera is on-line.
            else if (!cam_ptr->get_camera_info()->offline)
            {
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    control_ptr->stepping_reverse();
                }
            }
        }
    }
}


/*!
 * @brief The user wants to push a configuration out to a selected group of cameras.
 */
void MultiCamDock::push_config(void)
{
    QString     filename;
    QList<QListWidgetItem *> sel_items;

    // Ask the user for a configuration file.
    filename = CameraMDI::get_config_file();
    if (filename.isEmpty())
        return;

    // Get the currently selected item.
    sel_items = ui->list_cam->selectedItems();
    for (int i = 0; i < sel_items.count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)sel_items.at(i);
        if (my_item != nullptr)
        {
            CameraMDI *cam_ptr;

            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();

            // Push the configuration to the camera.
            if (cam_ptr && !cam_ptr->get_camera_info()->offline)
                cam_ptr->push_config_file(filename);
        }
    }
}

/*!
 * @brief This is the entry point into the video saving process for the group of cameras.
 */
void MultiCamDock::save_video_start()
{
    int         i;
    bool        mixed_group;
    bool        first_sensor_is_color;
    CameraMDI   *cam_ptr;
    fiListItem  *my_item;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the currently selected item.
    save_list.clear();
    save_list = ui->list_cam->selectedItems();
    if (save_list.count() == 0)
    {
        // Use all the items in the listbox.
        save_list = ui->list_cam->findItems(QString("*"), Qt::MatchWildcard | Qt::MatchWrap);
    }

    // Create the save dialog based on the first item in the list.
    my_item = (fiListItem *)save_list.at(0);
    if (my_item == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No items?!";
        return;
    }
    cam_ptr = my_item->get_camera();
    if (cam_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // AT THIS POINT cam_ptr is valid.

    // Is this a mono or color camera?
    first_sensor_is_color = cam_ptr->get_sensor()->is_color();

    // Check to see if we should warn the user about a mixed camera group.
    mixed_group = false;
    for (i = 1; i < save_list.count(); ++i)
    {
        bool        cur_sensor_type;
        CameraMDI   *test_cam;
        fiListItem  *test_item;

        // Get the next item in the list.
        test_item = (fiListItem *)save_list.at(i);
        if (test_item == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No items?!";
            return;
        }
        test_cam = test_item->get_camera();
        if (test_cam == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
            return;
        }

        // Get the sensor type.
        cur_sensor_type = test_cam->get_sensor()->is_color();

        // Is it a match to the first one?
        if (first_sensor_is_color != cur_sensor_type)
        {
            // We've got a mixed group.
            mixed_group = true;
        }
    }

    // Is it a mixed group?
    if (mixed_group)
    {
        // Let the user know.
        QMessageBox::warning(this, tr("MultiCamera Save"), tr("You have selected a mixture of color and monochrome cameras. You will not be able to select DNG as a save type."));
    }

    // Create the save dialog.
    SaveDialog      save_dlg(cam_ptr->get_ip_addr(), cam_ptr);

    // Initialize the save parameters.
    save_dlg.init_controls();
    save_dlg.set_autosave_mode(false);
    save_dlg.set_clip_range(cam_ptr->get_clip_start(), cam_ptr->get_clip_end());

    // Are there overlays?
    if (cam_ptr->get_player())
        save_dlg.enable_overlay_merge(cam_ptr->get_player()->has_overlay());

    // Configure the dialog for multicamera support mode.
    save_dlg.set_multicam(true, save_list.count());
    save_dlg.set_multicam_mixed(mixed_group);

    // Pop up the dialog.
    if (save_dlg.exec() == QDialog::Rejected)
        return;

    // Grab the save settings.
    save_dlg.get_save_params(save_info, cam_ptr->is_series3());

    // Are we saving to the workstation or on the camera?
    if (save_info.dest_media == SAVE_DLG::MEDIA_LOCAL_PC)
    {
        // Do we have a path?
        if (save_info.dest_path.isEmpty())
        {
            QString msg_str = tr("The download path must not be empty!");
            QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
            QMessageBox::critical(this, tr("Configuration Error"), msg_str);
            return;
        }
        else
        {
            // Make sure the path exists.
            QDir download_dir(save_info.dest_path);
            if (!download_dir.exists())
            {
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "Path doesn't exist. Creating:" << download_dir;
                if (!download_dir.mkpath(save_info.dest_path))
                {
                    // Can't create the path, so use the Documents folder instead.
                    QString msg_str = tr("The download path (%1) does not exist.").arg(save_info.dest_path);
                    QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
                    QMessageBox::critical(this, tr("Configuration Error"), msg_str);
                    return;
                }
            }
        }
    }

    // Start the ball rolling.
    QApplication::postEvent(this, new QEvent(MULTICAM_SAVE_EVENT));
}

/*!
 * @brief The user wants to synchronize a video clip from master camera to slave(s).
 */
void MultiCamDock::sync_clip()
{
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Save the name of the active camera.
    QString act_name;

    // The inforamtion for the active camera (actcam) that is useful for synchronizing clip set up.
    uint32_t    act_pb_clip_in           = 0;        // Register read of GIGE_REG_PB_CLIP_IN.
    uint32_t    act_pb_clip_out          = 0;        // Register read of GIGE_REG_PB_CLIP_OUT.
    uint32_t    act_clip_size            = 0;        // Number of frames included in the clip.
    uint32_t    act_trigger              = -1;       // Absolute frame number of trigger frame.

    // These location states indicate a left of trigger or a right of trigger clip location if they
    // are both false then the trigger of the active camera is contained in the clip.
    bool        act_pre_trigger_clip     = false;    // Is clip located to the left of trigger frame.
    bool        act_post_trigger_clip    = false;    // Is clip located to the right of trigger frame.

    // How many frames between trigger and touch of clip either beginning or end depending on the
    // location of the clip (depends on either act_pre_trigger_clip or act_post_trigger_clip
    // being set to true otherwise it will be set to zero).
    uint32_t    act_pre_post_gap_size    = 0;

    // Does the active camera have a valid clip to synchronize?
    bool        act_data_valid   = false;

    QList<QListWidgetItem *> sel_items; // Contains the cameras in our list that are marked selected.
    QList<QString> range_error_list; // Contains the cameras in our list that didn't match our sync clip range.

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the active camera, if there is one.
    CameraMDI *active_cam = MainWindow::GetInstance()->active_mdi_child();
    if (active_cam == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No active camera window available to synchronize clip!";
        return;
    }

    // Get the currently selected items.
    sel_items = ui->list_cam->selectedItems();

    // Determine if the active camera is in our selected list and pull out its clip set-up.
    for (int i = 0; !act_data_valid && i < sel_items.count(); ++i)
    {
        fiListItem *my_item = (fiListItem *)sel_items.at(i);
        cam_ptr = nullptr;

        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();

            // Make sure the camera pointer is valid.
            if (cam_ptr && !cam_ptr->get_camera_info()->offline && (active_cam == cam_ptr))
            {
                uint32_t    part_state;

                act_name = cam_ptr->get_camera_name();

                // Make sure this active camera is in a review state.
                // Pull out the information for this camera's partition.
                part_state = cam_ptr->read_camera_cache(GIGE_REG_PART_STATE);

                // Only interested if the partition state is done.
                if (part_state == GIGE_PART_STATE_COMPLETE)
                {
                    // Save the pertinant information for this clip.
                    act_trigger      = cam_ptr->read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

                    // There are some conditions where the review player is more "up-to-date" than
                    // the register cache.  Always use the review control to set active clip.
                    CameraReview *control_ptr = cam_ptr->get_review_controls();
                    if (control_ptr == nullptr)
                    {
                        QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                    }
                    else
                    {
                        act_pb_clip_in   = control_ptr->get_clip_start();
                        act_pb_clip_out  = control_ptr->get_clip_end();
                    }

                    // Sanity check, make sure that the clip in is less than or equal to the clip out.
                    if (act_pb_clip_out < act_pb_clip_in)
                    {
                        act_pb_clip_out = act_pb_clip_in;  // Make them equal so any calculations end up right.
                    }

                    // How many frames in the clip?
                    act_clip_size = act_pb_clip_out - act_pb_clip_in + 1;

                    // Check for the case where trigger is included within the clip.
                    if ((act_trigger >= act_pb_clip_in) && (act_trigger <= act_pb_clip_out))
                    {
                        // Make sure the state of the active camera is accurately setup.
                        act_pre_trigger_clip     = act_post_trigger_clip = false;
                        act_pre_post_gap_size    = 0;
                    }
                    else
                    {
                        // Clip is completely outside of the trigger frame.
                        // To the left or to the right of the trigger?
                        if (act_trigger > act_pb_clip_out)
                        {
                            // 0 1 2 3 4 5 6 7 8
                            // X C C X T X X X X
                            // Clip is to the left of the Trigger frame.
                            act_pre_trigger_clip     = true;
                            act_post_trigger_clip    = false;
                            // Calculate the gapsize between the clip end and the trigger.
                            act_pre_post_gap_size = act_trigger - act_pb_clip_out - 1;
                        }
                        else if (act_trigger < act_pb_clip_in)
                        {
                            // 0 1 2 3 4 5 6 7 8
                            // X X X X T X C C X
                            // Clip is to the right of the trigger frame.
                            act_pre_trigger_clip     = false;
                            act_post_trigger_clip    = true;
                            // Calculate the gapsize between the trigger and the clip start.
                            act_pre_post_gap_size = act_pb_clip_in - act_trigger - 1;
                        }
                    }

                    // Consider the data valid enough to attempt synchronization.
                    act_data_valid   = true;
                }
            }
        }
    }

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "\n\tact_data_valid:" << act_data_valid
                 << "\n\tact_name:" << act_name
                 << "\n\tact_pb_clip_in:" << act_pb_clip_in
                 << "\n\tact_pb_clip_out:" << act_pb_clip_out
                 << "\n\tact_clip_size:" << act_clip_size
                 << "\n\tact_trigger:" << act_trigger
                 << "\n\tact_pre_trigger_clip:" << act_pre_trigger_clip
                 << "\n\tact_post_trigger_clip:" << act_post_trigger_clip
                 << "\n\tact_pre_post_gap_size:" << act_pre_post_gap_size;

    // If we don't have anything to sync we are done here.
    if (!act_data_valid)
    {
        goto exit_sync_clip;
    }

    // Go through all selected cameras and sync up any clip data.
    for (int i = 0; i < sel_items.count(); ++i)
    {
        cam_ptr                 = nullptr;
        bool fits_just_right    = false;   // Does the active camera clip fit "just-right" into this camera's capture space?
        fiListItem *my_item     = (fiListItem *)sel_items.at(i);

        // Make sure there was an item in the list.
        if (my_item == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "fiListItem for multi-camera entry was a nullptr!";
            continue;
        }

        // Get a pointer to the camera.
        cam_ptr = my_item->get_camera();

        // Does this camera need to be synchronized with active camera clip set up?
        if (cam_ptr && !cam_ptr->get_camera_info()->offline && (cam_ptr != active_cam))
        {
            uint32_t    part_state = cam_ptr->read_camera_cache(GIGE_REG_PART_STATE);

            // Only interested if the partition state is complete.
            if (part_state == GIGE_PART_STATE_COMPLETE)
            {
                // Get the camera values.
                uint32_t num_frames     = cam_ptr->read_camera_cache(GIGE_REG_PART_NUM_FRAMES);
                uint32_t last_frame     = num_frames - 1;
                uint32_t trig_frame     = cam_ptr->read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
                uint32_t pb_clip_in;
                uint32_t pb_clip_out;

                // Is the active camera clip outside of the trigger frame?
                if (act_post_trigger_clip)
                {
                    uint32_t perfect_clip_out = trig_frame + act_pre_post_gap_size + act_clip_size;   // A perfect fit clip out setting.

                    // Clip start is to the right of the active camera trigger frame.
                    // Can we accomodate this clip to the right of this camera's trigger frame?
                    // 0 1 2 3 4 5 6 7 8
                    // X X X X T X C C X

                    // Is perfect_clip_out within the capture?
                    if (perfect_clip_out >= num_frames)
                    {
                        // NO it is not.  Set the pb_clip_in from last frame going backwards.
                        int perfect_clip_in = (int)last_frame - (int)act_clip_size;
                        pb_clip_in = (uint32_t)((perfect_clip_in < 0) ? 0 : perfect_clip_in); // Clip to zero if goes negative.
                    }
                    else
                    {
                        // YES there is plenty of room.  Make sure to add one because we are setting the start of the clip.
                        // without that +1 you end up with the last frame of the "gap".
                        pb_clip_in = trig_frame + act_pre_post_gap_size + 1;
                        fits_just_right = true;
                    }
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Post trigger clip camera:" << cam_ptr->get_camera_name() << "pb_clip_in:" << pb_clip_in << "trig_frame:" << trig_frame << "fits:" << fits_just_right;
                }
                else if (act_pre_trigger_clip)
                {
                    // Clip was to the left of the active camera trigger frame.
                    // 0 1 2 3 4 5 6 7 8
                    // X C C X T X X X X

                    // Can we accomodate the clip to the left of our trigger frame?
                    if ((act_pre_post_gap_size + act_clip_size) > trig_frame)
                    {
                        // Not a perfect fit so use the beginning of the camera capture working forwards.
                        pb_clip_in  = 0;
                    }
                    else
                    {
                        // Yes there is plenty of room.  Set accordingly.
                        pb_clip_in = trig_frame - act_pre_post_gap_size - act_clip_size;
                        fits_just_right = true;
                    }
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Pre trigger clip camera:" << cam_ptr->get_camera_name() << "pb_clip_in:" << pb_clip_in;
                }
                else
                {
                    // Clip included the active camera trigger frame.
                    // If the size of the clip start to the trigger frame is >= trigger frame
                    // then this camera's capture isn't a fit just right scenario.
                    // 0 1 2 3 4 5 6 7 8
                    // X X X C C T C C X
                    // B C C X X X X X X  (B=Both T and Clip In are same frame as in start triggers)
                    // C C C C C C C C B  (B=Both T and Clip Out are same frame as in end triggers)
                    // How many frames in the clip precede the active camera's trigger frame?
                    uint32_t pre_Tframes = act_trigger - act_pb_clip_in;

                    // Can this camera fit the count of pre_Tframes?
                    int pre_Tframes_avail = (int)trig_frame - (int)pre_Tframes;

                    // Can sync camera's capture accomodate the number of pre-trigger frames in active camera?
                    if (pre_Tframes_avail < 0)
                    {
                        // There isn't enough room.  So clip as best we can using frame 0.
                        pb_clip_in = 0;
                    }
                    else
                    {
                        // There is enough room to the left go ahead and set the start of the clip the same as active camera.
                        pb_clip_in = trig_frame - pre_Tframes;
                        fits_just_right = true; // possibly a fit-just-right case but will need to check the right-side later.
                    }

                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "pre_Tframes:" << pre_Tframes << "pre_Tframes_avail:" << pre_Tframes_avail
                                 << "sync_cam:" << cam_ptr->get_camera_name() << "pb_clip_in:" << pb_clip_in << "trig_frame:" << trig_frame << "last_frame:" << last_frame;
                }

                // Set the clip out based on the clip size make sure to take one away as
                // it is inclusive at end of clip.
                pb_clip_out = pb_clip_in + act_clip_size - 1;

                // Make sure to clip to end of the capture if exceeds the last frame.
                if (pb_clip_out > last_frame)
                {
                    pb_clip_out = last_frame;
                    fits_just_right = false;    // Not a perfect fit.
                }

                // If this camera didn't have a fit-just-right clip sync, mark it for a warning indication.
                if (!fits_just_right)
                {
                    range_error_list.append(cam_ptr->get_camera_name());
                }

                // Just tell the review object to change the playback control.
                CameraReview *control_ptr = cam_ptr->get_review_controls();
                if (control_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << cam_ptr->get_ip_addr_pretty() << "has no CameraReview pointer!";
                }
                else
                {
                    // Pass on the request.
                    control_ptr->set_pb_sync_clip(pb_clip_in, pb_clip_out);
                }
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "\n\tsync_camera_name:" << cam_ptr->get_camera_name()
                             << "\n\tpb_clip_in:" << pb_clip_in
                             << "\n\tpb_clip_out:" << pb_clip_out
                             << "\n\tfits_just_right:" << fits_just_right
                             << "\n\ttrig_frame:" << trig_frame
                             << "\n\tnum_frames:" << num_frames;
            }
        }
    }

exit_sync_clip:

    // Are any camera names in our list that didn't "fit-just-right"?
    if (!range_error_list.isEmpty())
    {
        QString title_str = tr("Synchronize Clip");
        QString warn_msg = tr("\nUnable to fully duplicate clipping range set in %1.\nPlease verify the clipping range applied to the following camera(s):\n\n").arg((act_name.isEmpty() ? "active camera" : act_name));
        for (int x = 0; x < range_error_list.count(); x++)
        {
            QString camera_name = range_error_list.at(x);
            warn_msg = warn_msg + camera_name + "\n";
        }
        QMessageBox::warning(MainWindow::GetInstance().get(), title_str, warn_msg);
    }
    return;
}

/*!
 * @brief The user wants to switch all the cameras in the dock to LIVE mode.
 */
void MultiCamDock::switch_to_live(void)
{
    CameraMDI   *cam_ptr;

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
                return;
            }

            // Tell the camera to go live.
            cam_ptr->do_camera_live();
        }
    }
}

/*!
 * @brief The user wants to switch all the cameras in the dock to Review mode.
 */
void MultiCamDock::switch_to_review(void)
{
    CameraMDI   *cam_ptr;

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
                return;
            }

            // Tell the camera to go to review.
            cam_ptr->do_camera_review();
        }
    }
}

/*!
 * @brief The user wants to issue a TRIGGER event to all the cameras in the dock.
 */
void MultiCamDock::trigger_cameras(void)
{
    CameraMDI   *cam_ptr;

    // Loop through all the cameras in the list.
    for (int i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();
            if (cam_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera!";
                return;
            }

            // Have the camera trigger.
            cam_ptr->do_camera_trigger();
        }
    }
}

/*!
 * @brief Query the cameras in the group, and update the controls based upon their states.
 */
void MultiCamDock::update_controls(void)
{
    int         i;
    int         sync_cameras = 0;
    bool        use_dearm(false);
    bool        show_review(false);
    bool        show_save(true);
    bool        autosave_checked(true);
    bool        autosave_configured(true);
    CameraMDI   *cam_ptr;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Initialize our controls.
    ui->cam_actions->start_update();

    // Loop through the cameras and see if all the cameras have autosave configured.
    autosave_configured = (ui->list_cam->count() > 0);
    for (i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();

            // Remember if one of the cameras is not configured.
            if (!cam_ptr->is_autosave_configured())
                autosave_configured = false;
        }
    }

    // Enable/disable based on if autosave is configured on ALL the cameras.
    ui->cam_actions->set_autosave_enabled(autosave_configured);

    // Loop through all the cameras in the list.
    for (i = 0; i < ui->list_cam->count(); ++i)
    {
        fiListItem *my_item;

        my_item = (fiListItem *)ui->list_cam->item(i);
        if (my_item != nullptr)
        {
            // Get a pointer to the camera.
            cam_ptr = my_item->get_camera();

            // Make sure the camera is on-line.
            if (cam_ptr && !cam_ptr->get_camera_info()->offline)
            {
                // Read our state registers.
                uint32_t cam_state = comm_ptr->get_reg(cam_ptr->get_ip_addr(), GIGE_REG_CAMERA_STATE);
                uint32_t ltr_mode = comm_ptr->get_reg(cam_ptr->get_ip_addr(), GIGE_REG_LTR_MODE);
                uint32_t part_state = comm_ptr->get_reg(cam_ptr->get_ip_addr(), GIGE_REG_PART_STATE);
                bool formatting = (comm_ptr->get_reg(cam_ptr->get_ip_addr(), GIGE_REG_FORMAT_MEDIA) == 0);

                // Is autosave enabled (i.e. checked)? It just takes one that's not to turn
                // it off (uncheck it) for the rest of the group.
                if (!cam_ptr->is_autosave_enabled())
                    autosave_checked = false;

                // Disable/enable the action buttons based on formatting status.
                ui->cam_actions->setEnabled(!formatting);

                // Update the enable flags.
                ui->cam_actions->reflect_camera_state(cam_state, part_state, ltr_mode);

                // Also need to update some of our own controls.
                switch (cam_state)
                {
                case GIGE_CAMERA_STATE_REVIEW:
                    show_review = true;
                    // Is this entry in the list "selected"?
                    if (my_item->isSelected())
                    {
                        sync_cameras++;     // Keep accumulation of the number of cameras that are in review state.
                    }
                    break;
                case GIGE_CAMERA_STATE_ARMED:
                    use_dearm = true;
                    //
                    // Fall through intentional...
                    //
                    [[fallthrough]];

                case GIGE_CAMERA_STATE_AUTOSAVE:
                case GIGE_CAMERA_STATE_LIVE_VIDEO:
                case GIGE_CAMERA_STATE_DEARMING:
                case GIGE_CAMERA_STATE_IDLE:
                case GIGE_CAMERA_STATE_SNAPSHOT:
                case GIGE_CAMERA_STATE_TRIGGERED:
                default:
                    show_review = false;
                    show_save = false;
                    break;
                }
            }
        }
    }

    // Did we have any cameras?
    if (i == 0)
    {
        // No cameras - no controls.
        show_review = false;
        show_save = false;
        autosave_checked = false;
    }

    // Update our autosave state, i.e. whether it's checked or not.
    ui->cam_actions->set_autosave_checked(autosave_checked);
    ui->action_mc_autosave_enable->setChecked(autosave_checked);

    // Finialize our controls.
    ui->cam_actions->finish_update();
    if (use_dearm)
        ui->action_mc_arm->setIcon(QIcon(QString::fromUtf8(":/images/dearmbtn.png")));
    else
        ui->action_mc_arm->setIcon(QIcon(QString::fromUtf8(":/images/armbtn.png")));

    // Show or hide the review controls.
    ui->frm_review->setVisible(show_review);

    // Show or hide the save button.
    ui->tb_save->setVisible(show_save);

    // Show or hide the sync clip button.
    ui->tb_sync_clip->setVisible(sync_cameras > 1);
    update();
}

bool MultiCamDock::allow_group()
{
    int index;
    bool allowed = false;
    CameraMDI *cam_ptr;
    QString base_model;
    QString base_ver;
    fiListItem *my_item;
    const FI_COMM::FI_CAMERA_INFO *cam_info;

    // You have to have more than one to have a group.
    if (ui->list_cam->count() < 2)
        goto bail;

    // All the cameras in the list must be the same. Compare to the first
    // camera in the list.

    // Get the first item in the list.
    my_item = (fiListItem *)ui->list_cam->item(0);
    if (my_item == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "A nullptr item in the list at index: 0";
        goto bail;
    }

    // Get a pointer to the camera.
    cam_ptr = my_item->get_camera();
    if (cam_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera at index 0!";
        goto bail;
    }

    // Get a pointer to the camera information structure.
    cam_info = cam_ptr->get_camera_info();
    if (cam_info == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera info at index 0?";
        goto bail;
    }

    // Save the model and firmware version.
    base_model = QString(cam_info->model_name);
    base_ver = QString(cam_info->dev_version);

    // Walk through the cameras in the group and enable/disable autosave.
    for (index = 1; index < ui->list_cam->count(); ++index)
    {
        // Get the item list.
        my_item = (fiListItem *)ui->list_cam->item(index);
        if (my_item == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "A nullptr item in the list at index:" << index;
            break;
        }

        // Get a pointer to the camera.
        cam_ptr = my_item->get_camera();
        if (cam_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera at index:" << index;
            break;
        }

        // Get a pointer to the camera information structure.
        cam_info = cam_ptr->get_camera_info();
        if (cam_info == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera info at index:" << index;
            break;
        }

        // Are the models a match?
        if (base_model.compare(cam_info->model_name) != 0)
            break;

        // Are the versions a match?
        if (base_ver.compare(cam_info->dev_version) != 0)
            break;
    }

    // Did we make it all the way through the list? If so, we can allow a group lock.
    allowed = (index == ui->list_cam->count());

bail:
    return allowed;
}
