/*!
 * @file multicam_dock.h
 * @brief This is the header file for the MultiCamDock class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef MULTICAM_DOCK_H
#define MULTICAM_DOCK_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QDockWidget>
#include <QListWidgetItem>
#include <QMutex>
#include <stdint.h>

#include "mainwindow.h"
#include "save_video.h"

namespace Ui
{
class MultiCamDock;
}

namespace FI_COMM
{
struct FI_CAMERA_INFO;
}

namespace MULTICAM_DOCK
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macro Definitions
//--------------------------------------------------------------------

//! Define a custom event for multicamera events.
const QEvent::Type  MULTICAM_SAVE_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_MULTICAM_EVENT);

/*!
 * @brief We need to subclass the list widget item so we can associate
 *        the camera information with the item.
 */
class fiListItem : public QListWidgetItem
{
public:
    explicit fiListItem(const QString& text, QListWidget *parent = nullptr, int type = Type) : QListWidgetItem(text, parent, type) { my_camera = nullptr; }

    CameraMDI   *get_camera()                 { return my_camera; }
    void    set_camera(CameraMDI *camera_ptr)     { my_camera = camera_ptr; }

private:
    CameraMDI   *my_camera;
};


}   // End of namespace MUTLICAM_DOCK.

//--------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------

class fiCamList;
class QListWidgetItem;
class QMdiSubWindow;
class QMenu;
class QTimer;

class CameraMDI;

//--------------------------------------------------------------------
// Class declaration: MultiCamDock
//--------------------------------------------------------------------
class MultiCamDock : public QDockWidget
{
    Q_OBJECT

public:
    explicit MultiCamDock(QWidget *parent = nullptr);
    ~MultiCamDock();

    void    add_camera(CameraMDI *camera_ptr);
    void    autosave_complete(CameraMDI *camera_ptr);
    void    autosave_dequeue(CameraMDI *camera_ptr);
    bool    autosave_ready(CameraMDI *camera_ptr);
    void    build_context_menu(QMenu *context_menu);
    QString camera_display(const FI_COMM::FI_CAMERA_INFO *camera_info);
    const fiCamList *get_group_list();
    bool    is_camera_grouped(CameraMDI *camera_ptr);
    bool    is_group_locked();
    void    remove_camera(CameraMDI *camera_ptr);
    bool    streaming_allowed()           { return autosave_queue.isEmpty(); }

public slots:
    void    add_ip_addr(uint32_t ip_addr);
    void    autosave(bool checked);
    void    delete_all_cameras();

protected:
    virtual void    changeEvent(QEvent *event);
    virtual void    contextMenuEvent(QContextMenuEvent *event_ptr);
    virtual void    customEvent(QEvent *event);

private slots:
    void    add_current();
    void    add_all_open();

    void    camera_activated(QMdiSubWindow *child_win_ptr);
    void    camera_offline(uint32_t ip_addr, bool offline);
    void    camera_selected(QListWidgetItem *sel_item);
    void    finish_setup();
    void    file_save_complete();

    void    delete_camera();
    void    display_update();
    void    do_dock();
    void    push_config();
    void    save_video_start();
    void    sync_clip();

    void    arm_cameras();
    void    trigger_cameras();
    void    switch_to_live();
    void    switch_to_review();

    void    jump_forward();
    void    jump_reverse();
    void    jump_trigger();
    void    play_forward();
    void    play_reverse();
    void    playback_stop();
    void    stepping_forward();
    void    stepping_reverse();

    void    update_controls();

private:
    bool allow_group();
    void update_list();
    void save_camera_start();
    void save_camera_finish();
    void translate_ui();

private:
    Ui::MultiCamDock *ui;                   //!< Our UI object.

    QList<CameraMDI *>  autosave_queue;     //!< Which cameras have indicated that they're ready to autosave.
    QMutex  autosave_mutex;                 //!< Protect the autosave list.
    int     display_index;                  //!< Which element to display: IP address, Mac address, etc.
    QList<QAction *>    display_actions;    //!< The actions associated with our context menu.
    QTimer  *update_timer;                  //!< Timer which polls the camera states.

    // Save controls.
    QList<QListWidgetItem *>    save_list;      //!< The list of cameras still to be saved.
    FI_SAVER::SAVE_PARAMS       save_info;      //!< The parameters to use for saving.

    //! The list of formats we'll allow the user to select from.
    static QList<QPair<QString, int>> display_list;
};

#endif // MULTICAM_DOCK_H
