/*!
 * @file panel_capture.cpp
 * @brief This file contains the implementation of the PanelCapture
 *        class. This class handles all the settings for the capture
 *        mode supported by the camera.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "panel_capture.h"
#include "ui_panel_capture.h"

#include <inttypes.h>
#include <QAbstractItemView>
#include <QCheckBox>
#include <QJsonArray>
#include <QJsonDocument>
#include <QMenu>
#include <QMessageBox>
#include <qmath.h>
#include <memory>

#include "client_math.h"
#include "camera_dock.h"
#include "camera_explore.h"
#include "camera_info.h"
#include "fi_camreg_group.h"
#include "fi_progress_dialog.h"
#include "mainwindow.h"

namespace CAPTURE_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Log messages for testing purposes.  Set to 1 to have them compiled for output.
#define LOG_GROUP_READS             0
#define LOG_LTRMODE_TRACING         0

const int FORMAT_TIMEOUT            = 120;  //!< Number of wake intervals before we show a "close" button.
const int MIN_SESSION_SIZE          = 1;    //!< Minimum session size in 250Mb segments.
const int SESSION_INCR              = 25;   //!< Define the increment for the session length.

const int OVERHEAD_COUNT            = 2;    //!< Number of partitions we need just for overhead.

//! Minimum number of frames for a BROC recording.
const int       MINIMUM_BROC_LEN            = 10;

//! The default FasFireLR session length is 4 GBytes, or 16 1/4 GBytes in "session length" terms.
const uint32_t DEFAULT_FFLR_SESSION_LEN     = 16;

#if LOG_LTRMODE_TRACING
bool fflr_enabled_last = false;
bool ltr_enabled_last = false;
bool ltr_dual_last = false;
#endif

}   // End of namespace CAPTURE_SETTINGS.

// Initialize our constant strings.
const QString PanelCapture::LTR_MODE_LABEL(tr("LR Mode"));
const QString PanelCapture::IM_MODE_LABEL(tr("GBytes"));
const QString PanelCapture::TXT_STYLE_COLOR_WARN  = QStringLiteral("QLabel {color: red;}");

using namespace CAPTURE_SETTINGS;
using namespace FM_PANEL;

/*!
 * @brief Constructor for the PanelCapture class. Initialize variables.
 */
PanelCapture::PanelCapture(QWidget *parent) :
    FMPanelWidget(parent),
    ui(new Ui::PanelCapture)
{
    camera_connecting = false;

    // Setup our UI controls.
    ui->setupUi(this);

    // The read-only line edit controls need to be enabled
    // so that they draw in enabled form - otherwise they
    // appear disabled and are harder to read.
    ui->eb_frames->setEnabled(true);
    ui->lbl_rec_time_hour->setEnabled(true);
    ui->lbl_rec_time_min->setEnabled(true);
    ui->lbl_rec_time_sec->setEnabled(true);
    ui->lbl_rec_time_ms->setEnabled(true);
    ui->lbl_rec_time_hour->setEnabled(true);
    ui->eb_datarate->setEnabled(true);

    // If external I/O arm-in is enabled the capture settings controls are disabled.
    // This will give the user a notice about why they might be disabled in that case.
    ui->lbl_armin_warn->setVisible(false);
    ui->lbl_armin_warn->setStyleSheet(PanelCapture::TXT_STYLE_COLOR_WARN);

    // Initialize our variables.
    wait_for_devices    = true;
    wait_for_ltr_read   = true;
    wait_for_ltr_mode   = false;
    ltr_mode_waiting    = 0;

    update_timer_id = 0;
    progress_timer_id = 0;
    last_write_addr = FI_COMM::NO_REGISTER;
    fflr_session_len = DEFAULT_FFLR_SESSION_LEN;

    write_list.clear();

    // Make sure we got a parent pointer.
    if (parent != nullptr)
    {
        // My grandparent is the mainwindow.
        MainWindow  *p_mainwnd = qobject_cast<MainWindow *>(parent->parent());

        // Connect the menu item's I'm interested in.
        connect(p_mainwnd->get_cp_storage(), &QAction::triggered, this, &PanelCapture::toggle_visibility);
    }

    // Connect up the slots and signals for our panel.
    connect(ui->sld_session, &QSlider::valueChanged, this, &PanelCapture::session_length_changed);
    connect(ui->sld_session, &QSlider::sliderReleased, this, &PanelCapture::confirm_session_length);
    connect(ui->spin_session, QOverload<double>::of(&QDoubleSpinBox::valueChanged), this, &PanelCapture::spin_length_changed);

    connect(ui->im_enable, &QCheckBox::clicked, this, &PanelCapture::capture_mode_clicked);
    connect(ui->ltr_enable, &QCheckBox::clicked, this, &PanelCapture::capture_mode_clicked);
    connect(ui->fflr_enable, &QCheckBox::clicked, this, &PanelCapture::capture_mode_clicked);
    connect(ui->rb_basic_mode, &QRadioButton::clicked, this, &PanelCapture::ltr_basic_clicked);
    connect(ui->rb_broc_mode, &QRadioButton::clicked, this, &PanelCapture::ltr_broc_clicked);
    connect(ui->rb_roc_mode, &QRadioButton::clicked, this, &PanelCapture::ltr_roc_clicked);
    connect(ui->sld_broc_len, &QSlider::sliderPressed, this, &PanelCapture::slider_lock);
    connect(ui->sld_broc_len, &QSlider::sliderReleased, this, &PanelCapture::slider_unlock);
    connect(ui->sld_broc_len, &QSlider::valueChanged, this, &PanelCapture::ltr_sld_broc_len_changed);
    connect(ui->spin_broc_len, qOverload<int>(&QSpinBox::valueChanged), this, &PanelCapture::ltr_spin_broc_len_changed);

    // Connect up the media buttons.
    connect(ui->cb_devices, qOverload<int>(&QComboBox::currentIndexChanged), this, &PanelCapture::device_selected);

    // Assume that there is no storage media connected to the camera.
    adjustSize();

    // After everything is setup, setup the intial size information.
    set_initial_sizing();

    // Start out with things disabled until we get connected to a camera.
    disable_all_panel_frames();
}

/*!
 * @brief Destructor for the PanelCapture class.
 */
PanelCapture::~PanelCapture()
{
    delete ui;

    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }
    if (progress_timer_id != 0)
    {
        killTimer(progress_timer_id);
        progress_timer_id = 0;
    }
}

/*!
 * @brief We are connecting to a (possibly) new camera.
 * @param ip_addr   The IP address of the camera we're connecting to.
 */
void PanelCapture::camera_connect(uint32_t ip_addr)
{
    FI_comm         *comm_ptr;

    // We are in process of connecting a camera.
    camera_connecting = true;

    // We need to wait for devices to load and GIGE_REG_LTR_MODE
    update_in_progress = false;
    wait_for_devices = true;
    wait_for_ltr_read = true;
    wait_for_ltr_mode = false;

    // Make sure we're not already connected to this camera.
    if (ip_addr == camera_ip_addr)
        return;

    // Pass it on.
    FMPanelWidget::camera_connect(ip_addr);

    // Get our communications pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Connect the communications interfaces that we're interested in.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelCapture::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &PanelCapture::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelCapture::handle_unsol_msg, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelCapture::write_complete, Qt::QueuedConnection);

    // Make sure we have an active camera pointer.
    CameraMDI *camera_ptr = p_control_panel->getCamera();
    if (camera_ptr != nullptr)
    {
        connect(camera_ptr, &CameraMDI::media_info_ready, this, &PanelCapture::handle_json_media, Qt::QueuedConnection);
        connect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelCapture::chk_refresh_regs, Qt::QueuedConnection);
        connect(camera_ptr, &CameraMDI::ext_arm_in_enabling, this, &PanelCapture::do_ext_arm_in_enabling, Qt::QueuedConnection);
        connect(camera_ptr, &CameraMDI::ext_arm_in_disabling, this, &PanelCapture::do_ext_arm_in_disabling, Qt::QueuedConnection);

        // Ask the camera for the available media. This will be caught
        // by the camera object, which will start the ball rolling.
        camera_ptr->start_camera_read_reg(GIGE_REG_AVAIL_MEDIA);

        // If we are connected to a series3 camera there is no FasFireLR support.
        ui->fflr_enable->setVisible(!camera_ptr->is_series3());
        ui->grp_save_device->setVisible(!camera_ptr->is_series3());

        // If we are connected to a seriesNC camera, there is no LR support. For a series3, make sure
        // either LTR mode is enabled, or Dual mode is enabled. If we are not enabled, make sure the
        // control is hidden and NOT checked.
        uint32_t ltr_mode = camera_ptr->read_camera_cache(GIGE_REG_LTR_MODE);
        bool enable_ltr = ((ltr_mode & ((uint32_t)GIGE_LTR_MODE_ENABLE | GIGE_LTR_MODE_DUAL)) != 0);
        ui->ltr_enable->setVisible(camera_ptr->is_series3() && enable_ltr);
        if (!enable_ltr)
            ui->ltr_enable->setChecked(false);

        // Initialize our session length.
        fflr_session_len = camera_ptr->get_fflr_session_len();
        if (fflr_session_len == 0)
            fflr_session_len = DEFAULT_FFLR_SESSION_LEN;
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    camera_connecting = false;
}

/*!
 * @brief Disconnect from our camera.
 */
void PanelCapture::camera_disconnect()
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr = nullptr;

    // Have we been here already?
    if (camera_ip_addr == 0)
        return;

    // Clear our IP address.
    camera_ip_addr = 0;

    // Make sure there is a camera pointer as we need to connect to some signals.
    if (p_control_panel)
    {
        camera_ptr = p_control_panel->getCamera();
    }

    // Disconnect any camera-specific signals.
    if (camera_ptr != nullptr)
    {
        disconnect(camera_ptr, &CameraMDI::media_info_ready, this, &PanelCapture::handle_json_media);
        disconnect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelCapture::chk_refresh_regs);
        disconnect(camera_ptr, &CameraMDI::ext_arm_in_enabling, this, &PanelCapture::do_ext_arm_in_enabling);
        disconnect(camera_ptr, &CameraMDI::ext_arm_in_disabling, this, &PanelCapture::do_ext_arm_in_disabling);
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // Pass it up.
    FMPanelWidget::camera_disconnect();

    // If we have a timer going, be sure to kill it.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }
    if (progress_timer_id != 0)
    {
        killTimer(progress_timer_id);
        progress_timer_id = 0;
    }

    // Do we have a progress dialog up?
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Get our communications pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr != nullptr)
    {
        // Connect the communications interfaces that we're interested in.
        disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelCapture::read_complete);
        disconnect(comm_ptr, &FI_comm::msg_timeout, this, &PanelCapture::msg_timeout_event);
        disconnect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelCapture::handle_unsol_msg);
        disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelCapture::write_complete);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
    }

    // Hide the media controls.
    ui->cb_devices->clear();

    // Make sure we still are not waiting for devices or ltr mode.
    update_in_progress = false;
    wait_for_devices = false;
    wait_for_ltr_read = false;
    wait_for_ltr_mode = false;

    // Disable all controls.
    disable_all_panel_frames();
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_AVAIL_MEDIA register value.
 */
void PanelCapture::handle_json_media(const QJsonDocument& jsonMedia)
{
    int item_index;
    double dev_size;
    uint64_t fflr_part_size;
    uint64_t avail_space;
    uint64_t min_required_space;
    QString base_label;
    QString device_label;
    QString device_sel;
    QString device_mountdir;
    QJsonObject mediaObj;
    CameraMDI   *camera_ptr;

    // If we were waiting for a media update we just got it.
    // If we were waiting for the first media to enable the panel we can enable it now.
    if (wait_for_devices)
    {
        wait_for_devices = false;
    }

    // Don't update if the popup is displayed.
    if (ui->cb_devices->view()->isVisible())
        return;

    // Do not update the session information if we are in the middle of a capture!
    if (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO)
    {
        // Cowardly refusing to update.
        return;
    }

    // Calculate the minimum amount of space required for FFLR mode.
    if (fflr_session_len < DEFAULT_FFLR_SESSION_LEN)
        fflr_part_size = ((uint64_t)fflr_session_len * SESSION_INCR * 1024 * 1024 * 1024) / 100;
    else
        fflr_part_size = ((uint64_t)DEFAULT_FFLR_SESSION_LEN * SESSION_INCR * 1024 * 1024 * 1024) / 100;
    min_required_space = fflr_part_size * (OVERHEAD_COUNT + 1);

#ifdef DUMP_JSON_MEDIA
    {
        QFile temp_file("/tmp/json_dev.txt");
        if (!temp_file.open(QIODevice::WriteOnly))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open '/tmp/json_dev.txt, Err:" << temp_file.errorString();
        }
        else
        {
            temp_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
            temp_file.write(jsonMedia.toJson().toStdString().c_str());
            temp_file.close();
        }
    }
#endif

    // Get the object from the document.
    mediaObj = jsonMedia.object();

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Block updates from the combobox.
    ui->cb_devices->blockSignals(true);

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
    {
        // Remember the selected device so we can restore it, if possible.
        device_sel = ui->cb_devices->currentText();
    }

    // Clear the combobox.
    ui->cb_devices->clear();

    // Register to read the JSON string for what media is attached to the camera.
    //
    // Possibly value tags are:
    // "name"       The raw device name, e.g. "/dev/mmcblk0p1" for the SDCard.
    // "type"       One of GIGE_MEDIA_xxx.
    // "label"      The label, if any, for the volume.
    // "devsize"    The number of bytes on the device (total size, not availability).
    // "avail"      The number of available bytes on the device.
    // "mountdir"   The absolute path where the device is mounted, e.g. "/media/sdcard".
    // "tag"        The string shown to the user, e.g. "FastecMedia [8 GB]".

    // Add back in found devices starting with the SDCard.
    if (mediaObj.contains("sdcard") && mediaObj["sdcard"].isArray())
    {
        QJsonArray dev_array = mediaObj["sdcard"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // The device needs a minimum 2 partitions of available space.
            avail_space = dev_object["avail"].toInteger();
            if (avail_space < min_required_space)
            {
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << dev_object["label"].toString() << "does not have enough available space:" << avail_space;
                continue;
            }

            // Remember this is an sdcard device.
            dev_object["type"] = GIGE_MEDIA_SDCARD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            device_mountdir = dev_object["mountdir"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("SDcard %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // We only support saving to devices that have a mounted directory.
            if (!device_mountdir.isEmpty())
            {
                // Add the device to the combobox.
                ui->cb_devices->addItem(device_label, dev_object);
            }
        }
    }

    // Check for NVME SSD devices.
    if (mediaObj.contains("nvme") && mediaObj["nvme"].isArray())
    {
        QJsonArray dev_array = mediaObj["nvme"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // The device needs a minimum 2 partitions of available space.
            avail_space = dev_object["avail"].toInteger();
            if (avail_space < min_required_space)
            {
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << dev_object["label"].toString() << "does not have enough available space:" << avail_space;
                continue;
            }

            // Remember this is an NVME device.
            dev_object["type"] = GIGE_MEDIA_SSD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            device_mountdir = dev_object["mountdir"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                if (dev_object["internal"].toBool())
                    device_label = QString("SSD %1GB [Internal]").arg(dev_size, 0, 'f', 0);
                else
                    device_label = QString("SSD %1GB [External]").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                if (dev_object["internal"].toBool())
                    device_label = base_label + QString(" [%1GB] [Internal]").arg(dev_size, 0, 'f', 0);
                else
                    device_label = base_label + QString(" [%1GB] [External]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // We only support saving to devices that have a mounted directory.
            if (!device_mountdir.isEmpty())
            {
                // Add the device to the combobox.
                ui->cb_devices->addItem(device_label, dev_object);
            }
        }
    }

    // Check for USB drives.
    if (mediaObj.contains("usb") && mediaObj["usb"].isArray())
    {
        QJsonArray dev_array = mediaObj["usb"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // The device needs a minimum 2 partitions of available space.
            avail_space = dev_object["avail"].toInteger();
            if (avail_space < min_required_space)
            {
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << dev_object["label"].toString() << "does not have enough available space:" << avail_space;
                continue;
            }

            // Remember this is an USB device.
            dev_object["type"] = GIGE_MEDIA_USB;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            device_mountdir = dev_object["mountdir"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("USB %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // We only support saving to devices that have a mounted directory.
            if (!device_mountdir.isEmpty())
            {
                // Add the device to the combobox.
                ui->cb_devices->addItem(device_label, dev_object);
            }
        }
    }

    // Allow signals to flow again.
    ui->cb_devices->blockSignals(false);

    // Do we have any devices available?
    if (ui->cb_devices->count() > 0)
    {
        int sel_index = 0;

        // We have devices, so allow FasFireLR mode.
        ui->fflr_enable->setEnabled(true);

        // For FasFireLR mode, if there was not a previously selected device, maybe there was one from a
        // prior run.
        if (is_fflr_enabled() && device_sel.isEmpty())
        {
            // Ask the camera object if it knows of a device.
            device_sel = camera_ptr->get_fflr_device_tag();
        }

        // Attempt to re-select the old value.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);
            if (sel_index < 0)
                sel_index = 0;
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);
        device_selected(sel_index);
    }
    else
    {
        // There are no devices found, so we cannot do FasFireLR mode.
        ui->fflr_enable->setEnabled(false);

        // If FasFireLR mode was selected, move to IM mode.
        if (ui->fflr_enable->isChecked())
        {
            ui->im_enable->setChecked(true);
            capture_mode_clicked();
        }
    }
}

/*!
 * @brief Process register values we received from the camera.
 *
 * @param reg_addr      The register's address.
 * @param reg_value     The register's value.
 */
void PanelCapture::handle_read(uint32_t reg_addr, uint32_t reg_value)
{
    // Don't update our settings if we're in the middle of updating a value.
    if (update_in_progress)
    {
        //QLOG_WARN() << __PRETTY_FUNCTION__ << "update_in_progress still set!!";
        // Just return - we're in the middle of an update right now.
        return;
    }

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_AVAIL_MEDIA:
        update_media(reg_value);
        break;
    case GIGE_REG_BROC_LEN:
        // Make sure the user isn't sitting on the controls.
        if (!ui->spin_broc_len->hasFocus() && !ui->sld_broc_len->hasFocus())
        {
            // Sanity check.
            if (reg_value < MINIMUM_BROC_LEN)
                reg_value = MINIMUM_BROC_LEN;

            // Update the burst length controls.
            if (ui->spin_broc_len->value() != (int)reg_value)
            {
                ui->spin_broc_len->blockSignals(true);
                ui->spin_broc_len->setValue((int)reg_value);
                ui->spin_broc_len->blockSignals(false);
                set_slider_value(ui->sld_broc_len, (int)reg_value);
            }
        }
        break;
    case GIGE_REG_CAMERA_STATE:
        process_camera_state(reg_value);
        break;
    case GIGE_REG_FILE_ACTION_STATUS:
        if (progressDialog)
        {
            // Are we done?
            if (reg_value != (uint32_t)GIGE_FA_STATUS_BUSY)
            {
                // Let the user know we are done.
                progressDialog->setStatusValue("Save operations have completed.");
                progressDialog->setBarMinMax(0, 100);
                progressDialog->setBarValue(100);

                // Make sure we stop the timer. This is how we can tell if
                // the user canceled or closed.
                killTimer(progress_timer_id);
                progress_timer_id = 0;
            }
        }
        break;
    case GIGE_REG_LTR_MODE:
        // Are we waiting for our first read?
        if (wait_for_ltr_read)
        {
            // We need to have had devices load.
            if (!wait_for_devices)
            {
                wait_for_ltr_read = false;

                // Enable the record mode and session length group.
                ui->grp_rec_mode->setEnabled(is_panel_ok_to_enable());
                ui->grp_internal_storage->setEnabled(is_panel_ok_to_enable());
            }
        }
        // Are we waiting for a particular ltr mode change to come through?
        if (wait_for_ltr_mode)
        {
            CameraMDI *camera_ptr;

            // Get our camera pointer.
            camera_ptr = p_control_panel->getCamera();

            // Check to see if we got the register value we were interested in.
            if (reg_value == ltr_mode_waiting)
            {
                wait_for_ltr_mode = false;
                ltr_mode_waiting = 0;

                // Enable the record mode and session length group.
                ui->grp_rec_mode->setEnabled(is_panel_ok_to_enable());
                ui->grp_internal_storage->setEnabled(is_panel_ok_to_enable());

                // Make sure we HAVE a camera pointer.
                if (camera_ptr)
                    camera_ptr->get_info_widget()->enable_actions();
            }
            else
            {
                // Do some sleuthing here.
                FI_comm     *comm_ptr           = MainWindow::GetInstance()->get_comm();

                if (camera_ptr && comm_ptr)
                {
                    // Is the camera reporting itself busy?
                    if (camera_ptr->is_busy())
                    {
                        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_LTR_MODE STALE REGISTER: CameraMDI::is_busy() is true, Did you forget to disable AUTOSAVE?!";
                    }

                    // Flush the cache entry.
                    comm_ptr->expire_entry(camera_ptr->get_ip_addr(), GIGE_REG_LTR_MODE);

                    // Try again.
                    comm_ptr->start_read_reg(camera_ptr->get_ip_addr(), GIGE_REG_LTR_MODE);
                }
                else
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_LTR_MODE STALE REGISTER: I don't have camera or COMM object!";
                }
            }
        }

        // Allow update to happen if we aren't waiting for LTR mode validation.
        if (!wait_for_ltr_mode)
        {
            // Update our LTR mode controls.
            update_ltr_controls(reg_value);
        }
        break;
    case GIGE_REG_PART_SIZE:
        part_size = ClientMath::qword_to_bytes(reg_value);
        break;
    case GIGE_REG_PART_SIZE_RESERVED:
        part_size_reserved = (uint32_t)ClientMath::qword_to_bytes(reg_value);
        break;
    case GIGE_REG_SESSION_LENGTH:
        // Keep the value in a handy to use variable.
        imem_session_len = reg_value;

        // Don't actually update if the user is camped out on the session slider.
        if (!ui->sld_session->hasFocus())
            update_session(__PRETTY_FUNCTION__);
        break;
    case GIGE_REG_ARM_IN_ENABLE:
        // Update this panel based on the arm-in external I/O state.
        update_arm_in_state();
        break;
    default:
        break;
    }
}

/*!
 * @brief Ensure accurate read from camera of FasFireLR setting.
 *
 * @returns True if it is enabled on the camera, false if not.
 */
bool PanelCapture::is_fflr_enabled()
{
    bool        fflr_enabled    = false;
    uint32_t    reg_value;
    CameraMDI   *camera_ptr     = p_control_panel->getCamera();
    FI_comm     *comm_ptr       = MainWindow::GetInstance()->get_comm();

    // Make sure we have an active camera pointer.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        goto is_fflr_enabled_exit;
    }
    // Make sure we have an active communications object pointer.
    if (comm_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        goto is_fflr_enabled_exit;
    }
    reg_value = comm_ptr->get_reg(camera_ptr->get_ip_addr(), GIGE_REG_LTR_MODE);

    // We want to read this from our cache for better accuracy.
    fflr_enabled = (reg_value & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR;

is_fflr_enabled_exit:
    return fflr_enabled;
}

/*!
 * @brief Ensure accurate read from camera of ltr mode's dual setting.
 *
 * @returns True if it is enabled on the camera, false if not.
 */
bool PanelCapture::is_ltr_dual_enabled()
{
    bool        ltr_dual_enabled    = false;
    uint32_t    reg_value;
    CameraMDI   *camera_ptr         = p_control_panel->getCamera();
    FI_comm     *comm_ptr           = MainWindow::GetInstance()->get_comm();

    // Make sure we have an active camera pointer.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        goto is_ltr_dual_enabled_exit;
    }
    // Make sure we have an active communications object pointer.
    if (comm_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        goto is_ltr_dual_enabled_exit;
    }
    reg_value = comm_ptr->get_reg(camera_ptr->get_ip_addr(), GIGE_REG_LTR_MODE);
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_LTR_MODE:" << Qt::hex << Qt::showbase << reg_value;

    // We want to read this from our cache for better accuracy.
    ltr_dual_enabled = (reg_value & (uint32_t)GIGE_LTR_MODE_DUAL) == (uint32_t)GIGE_LTR_MODE_DUAL;

is_ltr_dual_enabled_exit:
    return ltr_dual_enabled;
}

/*!
 * @brief Ensure accurate read from camera of ltr and fflr mode setting
 * and if both are false it assumes we are image memory capture mode.
 *
 * @returns True if it is enabled on the camera, false if not.
 */
bool PanelCapture::is_imem_enabled()
{
    bool        imem_enabled    = false;
    bool        ltr_enabled     = false;
    bool        fflr_enabled    = false;

    uint32_t    reg_value;

    CameraMDI   *camera_ptr         = p_control_panel->getCamera();
    FI_comm     *comm_ptr           = MainWindow::GetInstance()->get_comm();

    // Make sure we have an active camera pointer.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        goto is_ltr_dual_enabled_exit;
    }
    // Make sure we have an active communications object pointer.
    if (comm_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        goto is_ltr_dual_enabled_exit;
    }
    reg_value = comm_ptr->get_reg(camera_ptr->get_ip_addr(), GIGE_REG_LTR_MODE);

    // We want to read this from our cache for better accuracy.
    fflr_enabled    = (reg_value & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR;
    ltr_enabled     = (reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;

    if (!fflr_enabled && !ltr_enabled)
    {
        imem_enabled = true;
    }

is_ltr_dual_enabled_exit:
    return imem_enabled;
}

/*!
 * @brief Can the panel groups be enabled? There is a condition when external I/O arm-in
 * is enabled that we need to ensure the user doesn't modify the capture settings. Currently
 * this is the only condition and we have provided this function to handle it. In the future
 * there may be more conditions that warrant a disabled Capture Settings panel and this would
 * be a location to facilitate additional cases.
 *
 * @returns True if we are ok to enable based on any overriding condition, false if we are not ok
 * to enable at this time.
 */
bool PanelCapture::is_panel_ok_to_enable()
{
    bool ok_to_enable = true;
    CameraMDI *camera_ptr = p_control_panel->getCamera();

    // Do we have a camera connected?
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // Right now the only condition that warrants a total disable state for this panel overriding
    // any normal group enable/disable logic is if the External I/O Arm-In is enabled. In this condition
    // we don't want the user to modify the capture settings.
    uint32_t reg_value = camera_ptr->read_camera_cache(GIGE_REG_ARM_IN_ENABLE);
    if ((reg_value & GIGE_ARM_TYPE_EXTERNAL) == GIGE_ARM_TYPE_EXTERNAL)
    {
        ok_to_enable = false;
    }

    return ok_to_enable;
}

/*!
 * @brief Ensure accurate read from camera of ltr mode settings
 *
 * @param ltr_enabled_now:      Reference variable to set for ltr_enabled.
 * @param fflr_enabled_now:     Reference variable to set for fflr_enabled.
 * @param ltr_dual_enabled_now: Reference variable to set for fflr_dual_enabled.
 * @param ltr_reg:              Reference variable to set for the register itself.

 * @returns True if retrieved without an error, false if something is
 * wrong like no camera or communication object.
 */
bool PanelCapture::get_ltr_modes(bool& ltr_enabled_now, bool& fflr_enabled_now, bool& ltr_dual_enabled_now, uint32_t& ltr_reg)
{
    bool        processed_ok        = false;
    CameraMDI   *camera_ptr         = p_control_panel->getCamera();
    FI_comm     *comm_ptr           = MainWindow::GetInstance()->get_comm();

    // Set all reference variables to false.
    ltr_enabled_now = fflr_enabled_now = ltr_dual_enabled_now = false;

    // Set the register to zero.
    ltr_reg = 0;

    // Make sure we have an active camera pointer.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        goto get_ltr_modes_exit;
    }
    // Make sure we have an active communications object pointer.
    if (comm_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        goto get_ltr_modes_exit;
    }
    ltr_reg = comm_ptr->get_reg(camera_ptr->get_ip_addr(), GIGE_REG_LTR_MODE);

#if LOG_LTRMODE_TRACING
    {
        uint32_t    reg_value;
        reg_value = read_camera_cache(GIGE_REG_LTR_MODE);
        // See if there is ever a difference here.  This will be removed.
        if (ltr_reg != reg_value)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "!!GIGE_REG_LTR_MODE STALE CACHE!!"
                        << "cache:" << Qt::showbase << Qt::hex << reg_value << "cam query:" << Qt::showbase << Qt::hex << ltr_reg;
        }
    }
#endif

    // We want to read this from our cache for better accuracy.
    ltr_enabled_now         = (ltr_reg & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;
    fflr_enabled_now        = (ltr_reg & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR;
    ltr_dual_enabled_now    = (ltr_reg & (uint32_t)GIGE_LTR_MODE_DUAL) == (uint32_t)GIGE_LTR_MODE_DUAL;

    processed_ok = true;

get_ltr_modes_exit:
    return processed_ok;
}

/*!
 * @brief Enable or disable controls on the panel based on the camera's state.
 *
 * @param new_state     The new camera state.
 */
void PanelCapture::process_camera_state(uint32_t new_state)
{
    // Allow the panel to be accessed ONLY if we are in the LIVE state.
    setEnabled(new_state == GIGE_CAMERA_STATE_LIVE_VIDEO);

    // Update our LTR mode controls.
    update_ltr_controls(read_camera_cache(GIGE_REG_LTR_MODE));
}

/*!
 * @brief Utility function to read all registers this panel is interested in.
 */
void PanelCapture::read_registers()
{
    // Do we have a camera connected?
    if (p_control_panel->getCamera() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // If camera is in a review state don't ask the camera for this group of registers.
    if (camera_state == GIGE_CAMERA_STATE_REVIEW)
    {
        return;
    }

    // Skip this if we are in the middle of an update.
    if (update_in_progress)
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Skipping update_in_progress.";
        return;
    }

    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Prepare the camera register group read.
    cam_register_group->addReg(GIGE_REG_AVAIL_MEDIA);
    cam_register_group->addReg(GIGE_REG_BIT_MODE);
    cam_register_group->addReg(GIGE_REG_CAMERA_STATE);
    cam_register_group->addReg(GIGE_REG_FRAME_RATE);
    cam_register_group->addReg(GIGE_REG_FRAME_SIZE);
    cam_register_group->addReg(GIGE_REG_HEIGHT);
    cam_register_group->addReg(GIGE_REG_WIDTH);
    cam_register_group->addReg(GIGE_REG_LTR_MODE);
    cam_register_group->addReg(GIGE_REG_SSD_CAPABILITIES);
    cam_register_group->addReg(GIGE_REG_LTR_SECTOR_COUNT);
    cam_register_group->addReg(GIGE_REG_MEM_SIZE);
    cam_register_group->addReg(GIGE_REG_MEM_SIZE_RESERVED);
    cam_register_group->addReg(GIGE_REG_PART_STATE);
    cam_register_group->addReg(GIGE_REG_PART_SIZE);
    cam_register_group->addReg(GIGE_REG_PART_SIZE_RESERVED);
    cam_register_group->addReg(GIGE_REG_SESSION_LENGTH);
    cam_register_group->lastReg(GIGE_REG_ARM_IN_ENABLE);

    read_camera_registers(cam_register_group);
}

/*!
 * @brief Update our controls based on arm-in external I/O state.
 */
void PanelCapture::update_arm_in_state()
{
    if (!is_panel_ok_to_enable())
    {
        disable_all_panel_frames();
        ui->lbl_armin_warn->setVisible(true);
    }
    else
    {
        // We do not need the warning visible.
        ui->lbl_armin_warn->setVisible(false);

        // Enable the record mode and session length group.
        ui->grp_rec_mode->setEnabled(true);
        ui->grp_internal_storage->setEnabled(true);
    }
}

/*!
 * @brief Update our BROC length control(s).
 */
void PanelCapture::update_broc_len()
{
    bool        fflr_enabled_now;
    bool        ltr_enabled_now;
    bool        ltr_dual_enabled_now;
    uint32_t    ltr_reg;
    uint32_t    width;
    uint32_t    height;
    uint32_t    bit_mode;
    uint32_t    max_frames = 0;
    uint32_t    part_state;
    uint64_t    part_size_avail;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_reg);

    // Make sure we're in LTR mode or FFLR mode.
    if (!ltr_enabled_now && !fflr_enabled_now)
        return;

    // We're in LTR mode, but can the user modify anything?
    if (!ltr_dual_enabled_now)
        return;

    // In FasFireLR mode, we must have a device selected before we can continue.
    if (fflr_enabled_now && (ui->cb_devices->currentIndex() < 0))
        return;

    // For LTR mode we calculate the information differently.
    if (ltr_enabled_now)
    {
        // Figure out what state our partition is in.
        part_state = read_camera_cache(GIGE_REG_PART_STATE);
        part_size_avail = ClientMath::sector_to_bytes(read_camera_cache(GIGE_REG_LTR_SECTOR_COUNT));

        // Read the camera specific settings.
        bit_mode = read_camera_cache(GIGE_REG_BIT_MODE);
        height = read_camera_cache(GIGE_REG_HEIGHT);
        width = read_camera_cache(GIGE_REG_WIDTH);

        // for broc lengths they are within memory partitions.
        // Update our frame count.
        max_frames = ClientMath::calc_frame_count(part_size_avail, width, height, bit_mode, true, camera_ptr->is_series3());

        // Can we use the entire space?
        if (part_state != GIGE_PART_STATE_EMPTY)
        {
            // Adjust the frame count by the number of frames already recorded.
            uint32_t recorded_frames = read_camera_cache(GIGE_REG_PART_NUM_FRAMES);

            // If they are equal, then just ignore it. One time this happens is if
            // the camera is in the ARMed state. We could check the camera state here,
            // but checking this symptom is easier.
            if (max_frames != recorded_frames)
                max_frames -= recorded_frames;
        }
    }
    // For FFLR we don't look at partition space but at our fflr_session_len.
    else if (fflr_enabled_now)
    {
        // FFLR BROC is only limited by the FFLR session length the user has selected.
        max_frames = calc_max_fflr_frames();
    }

    // Sanity check our max_frames.
    if (max_frames < MINIMUM_BROC_LEN)
        return;

    // Update our slider and spinner.
    ui->sld_broc_len->blockSignals(true);
    ui->sld_broc_len->setMinimum(MINIMUM_BROC_LEN);
    ui->sld_broc_len->setMaximum((int)max_frames);
    ui->sld_broc_len->blockSignals(false);
    ui->spin_broc_len->blockSignals(true);
    ui->spin_broc_len->setMinimum(MINIMUM_BROC_LEN);
    ui->spin_broc_len->setMaximum((int)max_frames);
    ui->spin_broc_len->blockSignals(false);

    // Do not change the BROC length value if one of the length control has focus.
    if (!ui->spin_broc_len->hasFocus() && !ui->sld_broc_len->hasFocus())
    {
        uint32_t current_len = camera_ptr->read_camera_cache(GIGE_REG_BROC_LEN);
        if (current_len != (uint32_t)ui->spin_broc_len->value())
        {
            // Make sure we update the length to fit in the space allocated.
            if (current_len > max_frames)
                current_len = max_frames;
            if (current_len < MINIMUM_BROC_LEN)
                current_len = MINIMUM_BROC_LEN;

            // Update the current setting.
            ui->spin_broc_len->setValue(current_len);
            set_slider_value(ui->sld_broc_len, current_len);
            write_camera_register(GIGE_REG_BROC_LEN, (uint32_t)current_len);
        }
    }

    // Update the time.
    update_broc_time();
}

/*!
 * @brief Update the time that is reflect by the settings of our BROC length control(s).
 */
void PanelCapture::update_broc_time()
{
    double  broc_time;
    uint32_t frame_rate = read_camera_cache(GIGE_REG_FRAME_RATE);

    // Update the amount of time reflected in the BROC length.
    broc_time = double(ui->spin_broc_len->value()) / ClientMath::reg_to_frame_rate(frame_rate);
    ui->eb_broc_time->setText(QString::fromUtf8("%1").arg(broc_time, 0, 'f', 3));
}

/*!
 * @brief Utility method to update the display controls which show the number
 *        of frames and the amount of record time as defined by the selected ROI.
 */
void PanelCapture::update_info()
{
    bool        fflr_enabled_now;
    bool        ltr_enabled_now;
    bool        ltr_dual_enabled_now;
    uint32_t    ltr_reg;
    uint32_t    width;
    uint32_t    height;
    uint32_t    bit_mode;
    uint32_t    frames;
    uint32_t    frame_rate;
    double      fvalue;
    QString     s_field;
    uint64_t    part_size_avail;
    CameraMDI   *camera_ptr;

    //QLOG_TRACE() << __PRETTY_FUNCTION__ << "fflr_session_len:" << fflr_session_len;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_reg);

    // In FasFireLR mode, we must have a device selected before we can continue.
    if (fflr_enabled_now && (ui->cb_devices->currentIndex() < 0))
        return;

    // Read the camera specific settings.
    bit_mode = read_camera_cache(GIGE_REG_BIT_MODE);
    height = read_camera_cache(GIGE_REG_HEIGHT);
    width = read_camera_cache(GIGE_REG_WIDTH);

    // Different partition sizing if LTR is enabled or not.
    if (ltr_enabled_now)
    {
        part_size_avail = ClientMath::sector_to_bytes(read_camera_cache(GIGE_REG_LTR_SECTOR_COUNT));

        // Update our display of the image memory size.
        ui->lbl_session_size->setText(PanelCapture::LTR_MODE_LABEL);
        ui->spin_session->setVisible(false);
        ui->sld_session->setValue(ui->sld_session->maximum());

        frames = ClientMath::calc_frame_count(part_size_avail, width, height, bit_mode, true, camera_ptr->is_series3());
    }
    else if (fflr_enabled_now)
    {
        // Update our display of the image memory size.
        ui->lbl_session_size->setText(PanelCapture::IM_MODE_LABEL);
        ui->spin_session->setVisible(true);
        ui->spin_session->setValue((double)(fflr_session_len * SESSION_INCR) / (double)100.0);

        // Use our common routine that calculates FFLR frame counts.
        frames = calc_max_fflr_frames();
    }
    else
    {
        // Update our display of the image memory size.
        ui->lbl_session_size->setText(PanelCapture::IM_MODE_LABEL);
        ui->spin_session->setVisible(true);

        // Calculate the available size.
        part_size_avail = part_size - part_size_reserved;

        // We want to display the amount of session memory which is currently set. This is
        // expressed in gigabytes "GiB".
        fvalue = (double)(imem_session_len * SESSION_INCR) / (double)100.0;
        ui->spin_session->setValue(fvalue);

        frames = ClientMath::calc_frame_count(part_size_avail, width, height, bit_mode, false, camera_ptr->is_series3());
    }

    // Update the frame count based on the calculations made for the active capture mode.
    s_field = QString(tr("%1")).arg(frames);
    ui->eb_frames->setText(s_field);

    // Calculate the amount of recording time in seconds
    frame_rate = read_camera_cache(GIGE_REG_FRAME_RATE);

    // Calculate and display time components.
    {
        QTime       rec_time;

        // How much time are we talking about?
        ClientMath::calc_record_ms(frames, frame_rate, &rec_time);

        // Update our fields.
        ui->lbl_rec_time_hour->setText(QString::fromUtf8("%1").arg(rec_time.hour(), 2, 10, QLatin1Char('0')));
        ui->lbl_rec_time_min->setText(QString::fromUtf8("%1").arg(rec_time.minute(), 2, 10, QLatin1Char('0')));
        ui->lbl_rec_time_sec->setText(QString::fromUtf8("%1").arg(rec_time.second(), 2, 10, QLatin1Char('0')));
        ui->lbl_rec_time_ms->setText(QString::fromUtf8("%1").arg(rec_time.msec(), 3, 10, QLatin1Char('0')));
    }

    // Calculate the approximate data rate.
    fvalue = ((double)ClientMath::calc_frame_size(width, height, bit_mode) * ClientMath::reg_to_frame_rate(frame_rate)) / 1.0e6;
    //s_field = QString(tr("%1")).arg(fvalue, 0, 'f', 3);
    s_field = QString(tr("%1")).arg(fvalue, 0, 'f', 0);
    ui->eb_datarate->setText(s_field);
}

/*!
 * @brief We've had an LTR mode change, so update our controls.
 * @param ltr_mode      The new LTR mode.
 */
void PanelCapture::update_ltr_controls(uint32_t ltr_mode)
{
    CameraMDI   *camera_ptr;

    // Grab the state of the GIGE_REG_LTR_MODE register as passed in to here in "ltr_mode".
    bool        fflr_enabled_now        = (ltr_mode & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR;
    bool        ltr_enabled_now         = (ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;
    bool        ltr_dual_enabled_now    = (ltr_mode & (uint32_t)GIGE_LTR_MODE_DUAL) == (uint32_t)GIGE_LTR_MODE_DUAL;
    bool        uses_fascorder = false;     // Is ROC or BROC FasCorder mode allowed?

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Have we already come in here?
    if (update_in_progress)
        return;

#if LOG_LTRMODE_TRACING
    if ((ltr_enabled_now != ltr_enabled_last) || (fflr_enabled_now != fflr_enabled_last) || (ltr_dual_enabled_now != ltr_dual_last))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "!!GIGE_REG_LTR_MODE CHANGE!! ltr_mode:"
                     << Qt::showbase << Qt::hex << ltr_mode
                     << "ltr_enable:" << ((ltr_enabled_last != ltr_enabled_now) ? "Changed." : "No Change.")
                     << "fflr_enable:" << ((fflr_enabled_last != fflr_enabled_now) ? "Changed." : "No Change.")
                     << "ltr_dual_enabled:" << ((ltr_dual_last != ltr_dual_enabled_now) ? "Changed." : "No Change.");

        // Reset to catch the next change.
        fflr_enabled_last   = fflr_enabled_now;
        ltr_enabled_last    = ltr_enabled_now;
        ltr_dual_last       = ltr_dual_enabled_now;
    }
#endif

    // Update our LTR mode control if we're a series3 camera.
    if (camera_ptr->is_series3())
    {
        ui->ltr_enable->setVisible(ltr_enabled_now || ltr_dual_enabled_now);
        if (!ltr_enabled_now && !ltr_dual_enabled_now)
        {
            ui->ltr_enable->setChecked(false);
        }
    }

    // Check (or clear) the checkbox.
    if (ltr_enabled_now)
    {
        ui->ltr_enable->blockSignals(true);
        if (!ui->ltr_enable->isChecked())
            ui->ltr_enable->setChecked(true);
        ui->ltr_enable->blockSignals(false);

        // Disable the device selection control.
        ui->grp_save_device->setEnabled(false);

        // Disable the session length control.
        ui->frm_session->setEnabled(false);

        // Enable the LTR group controls.
        ui->grp_ltr_mode->setEnabled(is_panel_ok_to_enable());
    }
    else if (fflr_enabled_now)
    {
        ui->fflr_enable->blockSignals(true);
        if (!ui->fflr_enable->isChecked())
            ui->fflr_enable->setChecked(true);
        ui->fflr_enable->blockSignals(false);

        // Enable the device selection control.
        ui->grp_save_device->setEnabled(is_panel_ok_to_enable() && (ui->cb_devices->currentIndex() >= 0));

        // Enable the LTR mode controls.
        ui->grp_ltr_mode->setEnabled(is_panel_ok_to_enable() && (ui->cb_devices->currentIndex() >= 0));

        // Enable the session length control.
        ui->frm_session->setEnabled(is_panel_ok_to_enable() && (ui->cb_devices->currentIndex() >= 0));
    }
    else
    {
        ui->im_enable->blockSignals(true);
        if (!ui->im_enable->isChecked())
            ui->im_enable->setChecked(true);
        ui->im_enable->blockSignals(false);

        // Disable the device selection control.
        ui->grp_save_device->setEnabled(false);

        // Disable the LTR group controls.
        ui->grp_ltr_mode->setEnabled(false);

        // Enable the session length control.
        ui->frm_session->setEnabled(true);
    }

    // Long Record modes specific initialization.
    if (ltr_enabled_now || fflr_enabled_now)
    {
        // Determine whether or not FasCorder mode can be enabled.
        if ((ltr_enabled_now && ltr_dual_enabled_now) || fflr_enabled_now)
        {
            uses_fascorder = true;
        }

        // Do we need to setup our LTR mode controls?
        if (uses_fascorder)
        {
            // Enable the mode controls.
            ui->frm_ltr_mode->setEnabled(true);

            // Now...which mode is it in?
            if ((ltr_mode & GIGE_LTR_MODE_BROC) == GIGE_LTR_MODE_BROC)
            {
                ui->rb_broc_mode->setChecked(true);
                ui->frm_broc_len->setEnabled(true);
            }
            else if ((ltr_mode & GIGE_LTR_MODE_ROC) == GIGE_LTR_MODE_ROC)
            {
                ui->rb_roc_mode->setChecked(true);
                ui->frm_broc_len->setEnabled(false);
            }
            else
            {
                ui->rb_basic_mode->setChecked(true);
                ui->frm_broc_len->setEnabled(false);
            }

            // Update the maximum number of frames the burst length can support.
            update_broc_len();
        }
        else
        {
            // If we're in dual mode, make sure to check the "basic" control
            if (ltr_dual_enabled_now)
                ui->rb_basic_mode->setChecked(true);

            // Disable the LTR controls.
            ui->frm_broc_len->setEnabled(false);
            ui->frm_ltr_mode->setEnabled(false);
        }
    }

    // Don't actually update if the user is camped out on the session slider.
    if (!ui->sld_session->hasFocus())
    {
        update_session(__PRETTY_FUNCTION__);
        update_info();
    }
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_AVAIL_MEDIA register value.
 */
void PanelCapture::update_media(uint32_t reg_value)
{
    QString device_sel;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // We have been updated with device information.
    if (wait_for_devices)
    {
        wait_for_devices = false;
    }

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
        device_sel = ui->cb_devices->currentText();

    // Only do this if we're a series3 camera.
    if (camera_ptr->is_series3())
    {
        int item_index;
        if ((reg_value & GIGE_MEDIA_EXISTS_SDCARD) == GIGE_MEDIA_EXISTS_SDCARD)
        {
            // Only add the item if it doesn't already exist.
            if (ui->cb_devices->findText(CamExplore::SDCARD_LABEL) < 0)
                ui->cb_devices->addItem(CamExplore::SDCARD_LABEL);
        }
        else
        {
            // Check to see if the item went away. Remove it from the list.
            if ((item_index = ui->cb_devices->findText(CamExplore::SDCARD_LABEL)) >= 0)
                ui->cb_devices->removeItem(item_index);
        }
        if ((reg_value & GIGE_MEDIA_EXISTS_SSD) == GIGE_MEDIA_EXISTS_SSD)
        {
            // Only add the item if it doesn't already exist.
            if (ui->cb_devices->findText(CamExplore::SSD_LABEL) < 0)
                ui->cb_devices->addItem(CamExplore::SSD_LABEL);
        }
        else
        {
            // Check to see if the item went away. Remove it from the list.
            if ((item_index = ui->cb_devices->findText(CamExplore::SSD_LABEL)) >= 0)
                ui->cb_devices->removeItem(item_index);
        }
        if ((reg_value & GIGE_MEDIA_EXISTS_USB) == GIGE_MEDIA_EXISTS_USB)
        {
            // Only add the item if it doesn't already exist.
            if (ui->cb_devices->findText(CamExplore::USB_LABEL) < 0)
                ui->cb_devices->addItem(CamExplore::USB_LABEL);
        }
        else
        {
            // Check to see if the item went away. Remove it from the list.
            if ((item_index = ui->cb_devices->findText(CamExplore::USB_LABEL)) >= 0)
                ui->cb_devices->removeItem(item_index);
        }
    }

    // If there is no external media then disable the format and explore buttons.
    // Enabling is left to when the user clicks on the radio button for the particular device.
    if (ui->cb_devices->count() > 0)
    {
        int sel_index = 0;

        // Attempt to re-select the old value.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);
            if (sel_index < 0)
                sel_index = 0;
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);
        device_selected(sel_index);
    }
}

/*!
 * @brief Utility function to calculate the maximum frame count
 * for FasFireLR capture mode.
 * @returns the maximum frame count calculated.
 */
uint64_t PanelCapture::calc_max_fflr_frames()
{
    int sel_index;
    CameraMDI *camera_ptr;

    // Only applies to FasFireLR capture mode.
    if (!is_fflr_enabled())
        return 0;

    // Is there a device selected?
    sel_index = ui->cb_devices->currentIndex();
    if (sel_index < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No device selected.";
        return 0;
    }

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return 0;
    }

    // Get the device information.
    QJsonObject dev_object  = ui->cb_devices->itemData(sel_index).toJsonObject();
    if (dev_object.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No device JSon object.";
        return 0;
    }

    // Make sure our fflr_session_len is current.
    fflr_session_len = ui->sld_session->value();
    if (fflr_session_len == 0)
        fflr_session_len = DEFAULT_FFLR_SESSION_LEN;

    QString device_label    = dev_object["label"].toString();
    QString device_name     = dev_object["name"].toString();
    QString device_mountdir = dev_object["mountdir"].toString();
    QString device_tag      = dev_object["tag"].toString();;
    uint32_t device_type    = dev_object["type"].toInt(0);

    // Safest place to ensure this is set properly in the camera and it must
    // be set before we can calculate frame counts.
    camera_ptr->set_fflr_info(device_label, device_name, device_mountdir, device_tag, device_type, fflr_session_len);

    // Use the camera to tell me how many frames FasFireLR can support.
    return camera_ptr->calc_max_frames();
}

/*!
 * @brief Utility function to "disable" all of the frames on this panel.
 */
void PanelCapture::disable_all_panel_frames()
{
    ui->grp_rec_mode->setEnabled(false);
    ui->grp_internal_storage->setEnabled(false);
    ui->grp_save_device->setEnabled(false);
    ui->grp_ltr_mode->setEnabled(false);
}

/*!
 * @brief Update our session length value based on the camaera's register settings.
 * Takes into account the capture mode to ensure the controls in the panel represent
 * the capture mode conditions.
 */
void PanelCapture::update_session(QString who_called)
{
    bool        fflr_enabled_now;
    bool        ltr_enabled_now;
    bool        ltr_dual_enabled_now;
    uint32_t    ltr_reg;
    CameraMDI *camera_ptr;

    Q_UNUSED(who_called)
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "who:" << who_called;

    // Using uint64_t here to work with proper significance for all calculations below
    // dealing with very large device spaces.  Without this aroud 256GB problems arise.
    uint64_t    max_session_len = MIN_SESSION_SIZE;
    uint64_t    min_session_len = MIN_SESSION_SIZE;
    uint64_t    session_value   = MIN_SESSION_SIZE;

    // If we are in the middle of an update, skip this.
    if (update_in_progress)
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera_ptr!";
        return;
    }

    // Do not update the session information if we are in the middle of a capture!
    if (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO)
    {
        // Cowardly refusing to update.
        return;
    }

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_reg);

    // What is the session length currently set to?
    //
    // In FasFireLR mode, we have to deal with 2 ideas of "sessions". There is the
    // image memory PARTITION, and there is the session being saved to the device.
    // The IMAGE MEMORY "partition" length is specified by the GIGE_REG_SESSION_LENGTH
    // value. In FasFireLR mode, that never changes.
    //
    // In normal image memory recording, the GIGE_REG_SESSION_LENGTH value changes
    // depending on what the user specifies.
    //
    // In LR mode, the session length is irrelevant.
    imem_session_len = read_camera_cache(GIGE_REG_SESSION_LENGTH);

    // If we are in image memory capture mode the session means image memory partitioning.
    // LR mode just rides along.
    if (!fflr_enabled_now)
    {
        uint32_t    ddr2_size;
        uint32_t    ddr2_reserved;

        // Figure out the maximum number of sessions supported by the memory.
        ddr2_size = read_camera_cache(GIGE_REG_MEM_SIZE);
        ddr2_reserved = read_camera_cache(GIGE_REG_MEM_SIZE_RESERVED);
        max_session_len = ClientMath::calc_max_session_count(ddr2_size - ddr2_reserved);

        // For LR mode, the session value is the maximum value.
        if (ltr_enabled_now)
            session_value = max_session_len;
        else
            session_value = imem_session_len;
    }

    // For FFLR capture mode special processing.
    else if ((imem_session_len > 0) && (part_size > 0))
    {
        // Note:  part_size is the number of bytes for a full partition (includes reserved)
        // and fflr_part_size_qw is the quadwords for a full partition which was stored for
        // convienence.

        // For FFLR the selected device gives us the available space.
        uint64_t    avail_space         = 0;        // Device reported available space in bytes.
        bool        has_space           = false;    // Assume not enough space.
        int         sel_index           = ui->cb_devices->currentIndex();

        // There may be nothing selected.
        if (sel_index < 0)
            return;

        // First get the space available for the device it will be in bytes.
        QJsonObject dev_object = ui->cb_devices->itemData(sel_index).toJsonObject();
        if (dev_object.isEmpty())
            return;

        // Get the available space for the selected device.
        avail_space = dev_object["avail"].toInteger();

        // The minimum session length for FFLR mode is one session, specified by imem_session_len.
        min_session_len = imem_session_len;

        // Take away two full partition sizes from our available space to allow for
        // overhead and overprovision support for FasFireLR.
        uint64_t overhead_reduction = part_size * OVERHEAD_COUNT;

        // Is there enough space here? We need at least the overhead and one more partition for the recording.
        has_space = (avail_space < (overhead_reduction + part_size)) ? false : true;

        // If there isn't enough space pop out a warning label.
        if (has_space)
        {
            // Reduce the avail_space by the overhead_reduction.
            avail_space -= overhead_reduction;

            // Convert the bytes to quadwords rounding downward.
            max_session_len = ClientMath::bytes_to_qwords_rounded(avail_space, false);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Not enough space, avail_space:" << avail_space << "sel_index:" << sel_index << "dev_object:" << dev_object;
            return;
        }

        // Calculate the maximum number of 256 MiB sessions in the space available.
        max_session_len = ClientMath::calc_max_session_count64(max_session_len);

        // Make sure we align the maximum session size on the partition session boundary.
        if ((max_session_len % imem_session_len) != 0)
        {
            max_session_len = (max_session_len / imem_session_len) * imem_session_len;
        }

        // Align the selected fflr_session_len on an imem_session_len boundary.
        if ((fflr_session_len % imem_session_len) != 0)
        {
            // Align downward on imem_session_len boundary.
            fflr_session_len = (fflr_session_len / imem_session_len) * imem_session_len;
        }

        // We don't support less than the minimum partition size.
        if (fflr_session_len < imem_session_len)
            fflr_session_len = imem_session_len;

        // We don't support more than the max_session_len.
        if (fflr_session_len > (uint32_t)max_session_len)
            fflr_session_len = (uint32_t)max_session_len;

        // Set the session value to be used below.
        session_value = fflr_session_len;
    }

    // Update the UI controls for the session group.
    //
    // NOTE: Changing the min/max extents may also end up changing the value, even if
    //       we don't really want the value to change. We've preserved the correct
    //       value (session_value) and we will set it at the end.

    // Setup the session slider extents calculated above if they changed.
    if ((uint32_t)ui->sld_session->minimum() != min_session_len)
        ui->sld_session->setMinimum(min_session_len);
    if ((uint32_t)ui->sld_session->maximum() != max_session_len)
        ui->sld_session->setMaximum(max_session_len);

    // Calculate the min/max spinner extents.
    double min_session_spin = (double)((min_session_len * SESSION_INCR) / 100.0);
    double max_session_spin = (double)((max_session_len * SESSION_INCR) / 100.0);

    // Setup the session spinner extents, but only if they changed.
    if (ui->spin_session->minimum() != min_session_spin)
        ui->spin_session->setMinimum(min_session_spin);
    if (ui->spin_session->maximum() != max_session_spin)
        ui->spin_session->setMaximum(max_session_spin);

    // FFLR needs to have the step values set on the partition boundary.
    if (fflr_enabled_now)
    {
        // Has the step size changed? For FFLR mode, that SHOULD only happen at the
        // very beginning.
        if (ui->sld_session->singleStep() != (int)imem_session_len)
        {
            ui->spin_session->setSingleStep(((imem_session_len * SESSION_INCR) / 100.0));
            ui->sld_session->setSingleStep(imem_session_len);
            ui->sld_session->setPageStep(imem_session_len);
        }
    }
    else
    {
        // Has the step size changed? This SHOULD only happen when we switch modes.
        if (ui->sld_session->singleStep() != 0)
        {
            ui->spin_session->setSingleStep((SESSION_INCR / 100.0));
            ui->sld_session->setSingleStep(0);
            ui->sld_session->setPageStep(0);
        }
    }

    // Update the slider value, if it has changed.
    if ((uint32_t)ui->sld_session->value() != session_value)
        ui->sld_session->setValue(session_value);
}

/*!
 * @brief Central management to write a new GIGE_REG_LTR_MODE register
 * out to the camera and set up the wait state to eliminate user interface
 * control signals while we are waiting.
 *
 * @param ltr_mode:     The GIGE_REG_LTR_MODE register setting to write.
 * @param who_called:   For logging purposes who called this function.
 *
 * @returns True if the register was written to the camera.
 */
void PanelCapture::write_ltr_mode(uint32_t ltr_mode, QString who_called)
{
#if 0
    QString log_info = tr("[%1 called by %2]").arg(__PRETTY_FUNCTION__, who_called);
    if (wait_for_ltr_mode)
    {
        QLOG_WARN() << log_info
                    << "Requested write for " << Qt::showbase << Qt::hex << ltr_mode
                    << "But already WAITING for" << Qt::showbase << Qt::hex << ltr_mode_waiting
                    << "will reset for new request.";
    }
    QLOG_DEBUG() << log_info << "Requested GIGE_REG_LTR_MODE write for " << Qt::showbase << Qt::hex << ltr_mode;
#else
    Q_UNUSED(who_called)
#endif

    // Make sure we have an active camera pointer.
    CameraMDI   *camera_ptr;
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Special handling for the series3 vs. seriesNC cameras.
    if (camera_ptr->is_series3())
    {
        // The camera is going to reboot, so don't block updates.
        update_in_progress = false;
        wait_for_ltr_read = true;
        wait_for_devices = false;
        wait_for_ltr_mode = false;
    }
    else
    {
        FI_comm     *comm_ptr;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
            return;
        }

        // Write off the new mode value.
        wait_for_ltr_mode = true;
        ltr_mode_waiting = ltr_mode;
        last_write_addr = GIGE_REG_LTR_MODE;

        // Expire the entry.
        comm_ptr->expire_entry(p_control_panel->getCamera()->get_ip_addr(), GIGE_REG_LTR_MODE);

        // Make sure the user doesn't do something silly.
        disable_all_panel_frames();
    }

    // Write the new value. Go direct to the camera object because we
    // want to make sure the write happens.
    camera_ptr->write_camera_register(GIGE_REG_LTR_MODE, ltr_mode);
}

/*!
 * @brief Virtual function called by the "parent" widget to have us update the
 *        contents of our panel.
 */
void PanelCapture::update_panel()
{
    CameraMDI *camera_ptr = p_control_panel->getCamera();

    // Do we have a camera connected?
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Stop polling the camera.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }

    // Read registers we're interested in.
    read_registers();

    // Restart our timer.
    update_timer_id = startTimer(POLLING_PERIOD);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void PanelCapture::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        ui->retranslateUi(this);
    }

    // Pass it on.
    FMPanelWidget::changeEvent(event);
}

/*!
 * @brief Event routine which is invoked when our timer expires.
 * @param event     Pointer to the event class.
 */
void PanelCapture::timerEvent(QTimerEvent *event)
{
    // Pass it on up.
    FMPanelWidget::timerEvent(event);

    // Do we have a camera connected?
    if (p_control_panel->getCamera() == nullptr)
        return;

    // Is it the one we're interested in?
    if (event->timerId() == update_timer_id)
    {
        // Are we in the middle of an update?
        if (update_in_progress)
            return;

        // Poll the registers we're interested in.
        read_registers();
    }

    // Are we checking progress?
    if (event->timerId() == progress_timer_id)
    {
        CameraMDI *camera_ptr;

        // Make sure we have an active camera pointer.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
            return;
        }

        // Check to see if our operation is done.
        camera_ptr->start_camera_read_reg(GIGE_REG_FILE_ACTION_STATUS);
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has finished modifying the session length. Make sure
 *        everything is OK before writing the setting out to the camera.
 */
void PanelCapture::confirm_session_length()
{
    uint32_t    curr_len;
    uint32_t    part_state;
    bool        fflr_enabled_now;
    bool        ltr_enabled_now;
    bool        ltr_dual_enabled_now;
    uint32_t    ltr_reg;

    // If we are in the middle of an update, skip this.
    if (update_in_progress)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "update_in_progress still set!!";
        return;
    }

    // If the camera is in review state this makes no sense to change a session length.
    // See Ticket #486: Erroneous warning about changing Capture settings when switching cameras in FasMotion.
    if (read_camera_cache(GIGE_REG_CAMERA_STATE) == GIGE_CAMERA_STATE_REVIEW)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Asking to confirm session length when camera is in review state!";
        return;
    }

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_reg);

    // What kind of session are we working with?
    // If image memory special handling for possible recording loss.
    if (!ltr_enabled_now && !fflr_enabled_now)
    {
        // Was there a change?
        curr_len = read_camera_cache(GIGE_REG_SESSION_LENGTH);
        if (curr_len == (uint32_t)ui->sld_session->value())
            return;

        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // Disable the controls.
        ui->frm_session->setEnabled(false);

        // Check to see if there is a "dirty" partition we should be worried about.
        part_state = read_camera_cache(GIGE_REG_PART_STATE);

        // Only need to ask if the active partition is dirty.
        if (!camera_connecting && (part_state == GIGE_PART_STATE_COMPLETE))
        {
            int answer;

            answer = QMessageBox::question(this,
                                           tr("FasMotion"),
                                           tr("Changing the session length will result in the unrecoverable loss of any unsaved video."
                                              "<p>Are you sure you want to change the session length?"),
                                           QMessageBox::Yes | QMessageBox::No);
            if (answer == QMessageBox::No)
            {
                // Release our "lock".
                update_in_progress = false;

                // The user doesn't want to continue with the new settings. Restore
                // the settings to the values on the camera.
                part_size = ClientMath::qword_to_bytes(read_camera_cache(GIGE_REG_PART_SIZE));
                part_size_reserved = (uint32_t)ClientMath::qword_to_bytes(read_camera_cache(GIGE_REG_PART_SIZE_RESERVED));
                imem_session_len = read_camera_cache(GIGE_REG_SESSION_LENGTH);

                // Re-enable the controls.
                ui->frm_session->setEnabled(true);

                // Restore the displayed information.
                update_session(__PRETTY_FUNCTION__);
                update_info();

                // All done.
                return;
            }
        }

        // Update the camera's session length (number of .25 GB "chunks").
        last_write_addr = GIGE_REG_SESSION_LENGTH;
        write_camera_register(GIGE_REG_SESSION_LENGTH, (uint32_t)ui->sld_session->value());
    }
    else
    {
        // Make sure the ROC/BROC values are valid for the new session length.
        update_broc_len();
    }
}

/*!
 * @brief React to the user selecting a device.
 * @param index The index of the entry in the combobox.
 */
void PanelCapture::device_selected(int index)
{
    bool        fflr_enabled_now;
    bool        ltr_enabled_now;
    bool        ltr_dual_enabled_now;
    uint32_t    ltr_reg;

    // Check for nothing selected.
    if (index < 0)
        return;

    // If we are in the middle of an update, skip this.
    if (update_in_progress)
        return;

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_reg);

    // We only care about the device changing for FFLR mode.
    if (!fflr_enabled_now)
        return;

    // This gets called whenever the user physically selects a different device,
    // or when the camera pushes a JSON media update.  This is always called with
    // the device that is selected (its index).  Therefore we can count on this
    // providing us with up-to-date information for connected/selected device.
    // IMPORTANT: This gets called even when the camera is not strictly in a state
    // of Live Video.  The panel itself will disable when not in Live Video.

    // What we want to accomplish here is to give the session controls an opportunity
    // to respond to a device update.  This is important especially for FFLR where it
    // wants to know about device storage available bytes.
    update_session(__PRETTY_FUNCTION__);
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the read registers group read's last address == GIGE_REG_SESSION_LENGTH.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void PanelCapture::chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr  = p_control_panel->getCamera();

    // If we are not connected to any camera we don't care about this signal.
    if (camera_ptr != nullptr)
    {
        // Is this event for us?
        if (ip_addr != camera_ptr->get_ip_addr())
            return;

        // Verify it's our register.
        if (reg_addr == GIGE_REG_SESSION_LENGTH)
        {
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_SESSION_LENGTH]";
#endif
            // Don't actually update if the user is camped out on the session slider.
            if (!ui->sld_session->hasFocus() && !ui->spin_session->hasFocus() && !update_in_progress)
            {
                // Update the panel display(s).
                update_info();
            }
        }
    }
}

void PanelCapture::do_save_logs()
{
    CameraMDI *camera_ptr;
    QJsonObject json_cmd;
    QJsonObject dev_object;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Get the currently selected object.
    dev_object = ui->cb_devices->currentData().toJsonObject();

    // Build up our copy command. The expectation is that the camera server
    // will handle copying all folders, sub-folders, images, metadata, etc.,
    // as appropriate.
    json_cmd["dstpath"] = dev_object["mountdir"];
    json_cmd["action"] = GIGE_FA_SAVE_LOGS;

    // Send it on it's way.
    QJsonDocument json_doc(json_cmd);
    camera_ptr->write_camera_buffer(GIGE_REG_FA_JSON, json_doc.toJson());

    // Through up a dialog to let the user know it's happening.
    progressDialog = fiProgressDialog::CreateSimpleProgressDialog(tr("Saving Log Files"),
                                                                  (tr("This process can take awhile, and youn\ncannot stop it.\n\nIf you close this message box,\nthe operation will continue in the background.")),
                                                                  QMessageBox::Cancel, __PRETTY_FUNCTION__);
    // Display the progress dialog.
    if (progressDialog)
    {
        progressDialog->setBarMinMax(0, 0);
        progressDialog->showBar(true);
        progressDialog->setCancelText(tr("Close"));
        progressDialog->hideDialog();
        connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &PanelCapture::do_user_canceled, Qt::UniqueConnection);
        progressDialog->showDialogWithoutPolling();
        progress_timer_id = startTimer(PROGRESS_PERIOD);
    }
}

/*!
 * @brief Called when the user cancels/closes the progress dialog.
 */
void PanelCapture::do_user_canceled()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // If we are here because the user said Cancel, then the timer
    // will still be going.
    if (progress_timer_id != 0)
    {
        // Stop the timer.
        killTimer(progress_timer_id);
        progress_timer_id = 0;
    }

    // Close down the dialog.
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }
}

/*!
 * @brief The camera is handling user disable arm-in external I/O action.
 */
void PanelCapture::do_ext_arm_in_disabling()
{
    update_arm_in_state();
}

/*!
 * @brief The camera is handling user enable arm-in external I/O action.
 */
void PanelCapture::do_ext_arm_in_enabling()
{
    update_arm_in_state();
}

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void PanelCapture::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Make sure we're not in the middle of an update.
    if (update_in_progress)
        return;

    // Pass it along like it was a read complete event.
    handle_read(reg_addr, reg_value);

    // For some registers, we should update our display.
    switch (reg_addr)
    {
    case GIGE_REG_BIT_MODE:
    case GIGE_REG_FRAME_RATE:
    case GIGE_REG_FRAME_SIZE:
    case GIGE_REG_HEIGHT:
    case GIGE_REG_WIDTH:
    case GIGE_REG_LTR_SECTOR_COUNT:
    case GIGE_REG_PART_SIZE:
    case GIGE_REG_SESSION_LENGTH:
        // Don't actually update if the user is camped out on the session slider.
        if (!ui->sld_session->hasFocus())
        {
            // First update the session information.
            update_session(__PRETTY_FUNCTION__);

            // Then it is ok to update the display information.
            update_info();
        }
        break;
    case GIGE_REG_ARM_IN_ENABLE:
        // When the Arm-In enable state is modified we need to make sure to update the
        // enabled state of this panel.
        if (!is_panel_ok_to_enable())
        {
            disable_all_panel_frames();
            ui->lbl_armin_warn->setVisible(true);
        }
        else
        {
            // We do not need the warning visible.
            ui->lbl_armin_warn->setVisible(false);
        }
        break;
    default:
        break;
    }
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param ip_addr       The IP address of the camera which didn't respond.
 * @param reg_addr      The address of the register which was read.
 * @param read_op       True if this was a read operation, false for a write operation.
 */
void PanelCapture::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Is there a write in progress?
    if (update_in_progress && !read_op)
    {
        // Make sure this is the address we're concerned about.
        if (reg_addr != last_write.first)
            return;

        // Update our retry counter.
        ++retry_counter;

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "retry #" << retry_counter << "Last write:" << last_write;

        // Should we retry the write?
        if (retry_counter < FM_PANEL::MAX_WRITE_RETRIES)
        {
            // Retry the write.
            write_camera_register(last_write.first, last_write.second, false);
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Maximum write attempts exceeded.";

            // Reset our write list and flags.
            write_list.clear();
            update_in_progress = false;
        }
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void PanelCapture::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;

    // Pass it up.
    FMPanelWidget::read_complete(ip_addr, reg_addr);

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Retrieve the camera state.
    reg_value = read_camera_cache(reg_addr);

    // Handle the read event.
    handle_read(reg_addr, reg_value);
}

/*!
 * @brief The user has changed the session length slider. Update the display
 *        controls to provide feedback.
 */
void PanelCapture::session_length_changed(int value)
{
    int step;
    bool fflr_enabled_now;
    bool ltr_enabled_now;
    bool ltr_dual_enabled_now;
    uint32_t ltr_reg;

    // If we are in the middle of an update, skip this.
    if (update_in_progress)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "update_in_progress still set!!";
        return;
    }

    // What is our "step" granularity?
    step = ui->sld_session->singleStep();
    if (step && ((value % step) != 0))
    {
        // Somehow we got off on a wrong step! Fix it.
        value = (value / step) * step;

        // Update the control.
        ui->sld_session->blockSignals(true);
        ui->sld_session->setValue(value);
        ui->sld_session->blockSignals(false);
    }

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_reg);

    // If we are in image memory capture mode the session means image memory partitioning.
    if (!ltr_enabled_now && !fflr_enabled_now)
    {
        uint32_t    ddr2_size;
        uint32_t    ddr2_reserved;

        // Any real change?
        if (imem_session_len == (uint32_t)value)
            return;

        // Update our local copy of the session length.
        imem_session_len = (uint32_t)value;

        // Get some values we'll need.
        ddr2_size = read_camera_cache(GIGE_REG_MEM_SIZE);
        ddr2_reserved = read_camera_cache(GIGE_REG_MEM_SIZE_RESERVED);

        // Update our local idea of how big the partition for the given session length should be.
        part_size = ClientMath::qword_to_bytes(ClientMath::calc_session_space(imem_session_len, (uint32_t)ui->sld_session->maximum(), ddr2_size, ddr2_reserved));
    }
    else if (fflr_enabled_now && (ui->cb_devices->currentIndex() >= 0))
    {
        // Any real change?
        if (fflr_session_len == (uint32_t)value)
            return;

        // Update our local copy for the fflr session length entered by user.
        fflr_session_len = (uint32_t)value;
        if (fflr_session_len == 0)
            fflr_session_len = DEFAULT_FFLR_SESSION_LEN;
    }

    // Give the user feedback as to what these changes mean.
    update_info();
}

void PanelCapture::spin_length_changed(double value)
{
    int session_len;
    bool fflr_enabled_now;
    bool ltr_enabled_now;
    bool ltr_dual_enabled_now;
    uint32_t ltr_reg;

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_reg);

    // Ignore if we're in the middle of an update.
    // If we are in the middle of an update, skip this.
    if (update_in_progress)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "update_in_progress still set!!";
        return;
    }

    // In FasFireLR mode, we must have a device selected before we can continue.
    if (fflr_enabled_now && (ui->cb_devices->currentIndex() < 0))
        return;

    // Calculate the session length.
    session_len = (int)((value * 100.0) / SESSION_INCR);

    // No need to change if the slider and spinner already match.
    if (session_len == ui->sld_session->value())
        return;

    // Set the slider to the session length and confirm this is what the user wants.
    ui->sld_session->setValue(session_len);

    // Are we in image memory capture mode?
    if (!fflr_enabled_now && !ltr_enabled_now)
    {
        // We only need to do this when we're in image memory mode. In LR mode, we shouldn't even be here.
        // In FFLR mode, a change to the session length is a FasMotion only thing. It doesn't actually change
        // the size of the camera server's partition.
        confirm_session_length();
    }
}

/*!
 * @brief The user has selected the basic LTR mode radio button.
 */
void PanelCapture::ltr_basic_clicked()
{
    uint32_t    ltr_mode;
    uint32_t    part_state;

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_in_progress = true;

    // Read the current LTR mode value.
    ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);

    // Warn the user if in LTR mode.
    if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // What is the partition state.
        part_state = read_camera_cache(GIGE_REG_PART_STATE);
        if ((part_state == GIGE_PART_STATE_COMPLETE) || (part_state == GIGE_PART_STATE_LOCKED))
        {
            if (QMessageBox::question(this, tr("FasMotion"),
                                      tr("Camcorder and LR Burst modes append recordings onto the SSD, but LR Basic mode uses the entire drive."
                                         "<p>If you change modes, the next recording will overwrite any existing data on the SSD!"
                                         "<p>Are you sure you want to change the mode?"),
                                      QMessageBox::Yes | QMessageBox::No) == QMessageBox::No)
            {
                // Restore the selected mode.
                if ((ltr_mode & GIGE_LTR_MODE_BROC) == GIGE_LTR_MODE_BROC)
                    ui->rb_broc_mode->setChecked(true);
                else
                    ui->rb_roc_mode->setChecked(true);

                // No need to block the updates anymore.
                update_in_progress = false;
                return;
            }
        }
    }

    // Disable the burst length controls.
    ui->frm_broc_len->setEnabled(false);

    // Clear out the ROC/BROC bits.
    ltr_mode &= ~((uint32_t)GIGE_LTR_MODE_ROC | (uint32_t)GIGE_LTR_MODE_BROC);

    // Call our central handler to write the new mode.
    write_ltr_mode(ltr_mode, __PRETTY_FUNCTION__);
}

/*!
 * @brief The user has selected the BROC radio button.
 */
void PanelCapture::ltr_broc_clicked()
{
    uint32_t    ltr_mode;
    uint32_t    part_state;

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_in_progress = true;

    // Read the current LTR mode value.
    ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);

    // If LTR mode warn user if partition loss can happen.
    if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // Read the partition state.
        part_state = read_camera_cache(GIGE_REG_PART_STATE);

        // If the partition is "dirty" and we're transitioning from basic mode, we need to warn the user.
        if (((part_state == GIGE_PART_STATE_COMPLETE) || (part_state == GIGE_PART_STATE_LOCKED))
            && ((ltr_mode & GIGE_LTR_MODE_ROC) != GIGE_LTR_MODE_ROC))
        {
            if (QMessageBox::question(this, tr("FasMotion"),
                                      tr("Camcorder and LR Burst modes append recordings onto the SSD, but LR Basic mode uses the entire drive."
                                         "<p>If you change modes, the next recording will overwrite any existing data on the SSD!"
                                         "<p>Are you sure you want to change the mode?"),
                                      QMessageBox::Yes | QMessageBox::No) == QMessageBox::No)
            {
                // Go back to basic mode.
                ui->rb_basic_mode->setChecked(true);

                // No need to block the updates anymore.
                update_in_progress = false;
                return;
            }
        }
    }
    // Otherwise FFLR is just writing to disk so it is non-destructive.

    // Enable the burst length controls.
    ui->frm_broc_len->setEnabled(true);

    // Clear out the ROC bit and set the BROC bit.
    ltr_mode = (ltr_mode & ~((uint32_t)GIGE_LTR_MODE_ROC)) | GIGE_LTR_MODE_BROC;

    // Call our central handler to write the new mode.
    write_ltr_mode(ltr_mode, __PRETTY_FUNCTION__);
}

/*!
 * @brief The user has selected one of the capture mode types.
 */
void PanelCapture::capture_mode_clicked()
{
    bool        show_warn_msg = false;  // If capture mode needs user warning response this goes true.
    int         user_resp;
    bool        fflr_enabled_now;
    bool        ltr_enabled_now;
    bool        ltr_dual_enabled_now;
    uint32_t    ltr_mode;
    uint32_t    reg_value;

    QString     msg_str;
    QString     msg_part;
    QString     device_sel;

    CameraMDI   *camera_ptr;

    // Get the settings for our ltr mode.
    get_ltr_modes(ltr_enabled_now, fflr_enabled_now, ltr_dual_enabled_now, ltr_mode);

    // Make sure the user didn't just click on a radio button that was already selected.
    QRadioButton *rb_ptr = qobject_cast<QRadioButton *>(sender());
    if ((rb_ptr == ui->im_enable) && !ltr_enabled_now && !fflr_enabled_now)
        return;
    if ((rb_ptr == ui->fflr_enable) && fflr_enabled_now)
        return;
    if ((rb_ptr == ui->ltr_enable) && ltr_enabled_now)
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get the name of the device selected in the dropdown in case it is needed
    // for one of the messages.
    device_sel = ui->cb_devices->currentText();

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_in_progress = true;

    // Also, don't allow the user to do anything silly.
    camera_ptr->get_info_widget()->disable_actions();

    // Build up a case for dirty partitions should it be needed for warning messages.
    uint32_t    dirty_count = 0; // If camera has unsaved partitions this will be non-zero.

    // TICKET #435:  Add more information to warning if user also has "dirty" unsaved partitions in
    // image memory that they will lose partition data transitioning to long record mode.
    // Only do this if we aren't already in LTR or FFLR capture mode.
    if (!ltr_enabled_now && !fflr_enabled_now)
    {
        uint32_t    part_state;

        // Do we have a partition sitting unsaved in memory?
        part_state = read_camera_cache(GIGE_REG_PART_STATE);
        if (part_state == GIGE_PART_STATE_COMPLETE)
        {
            // Read the autosave register to determine if there is more than one dirty buffer.
            reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);
            dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);

            // We know we have at least one dirty partition, which is why we are here.
            if (dirty_count == 0)
                dirty_count++;

            QLOG_TRACE() << __PRETTY_FUNCTION__ << "GIGE_AUTOSAVE_TOTAL_COUNT dirty_count:" << dirty_count;
        }
    }

    // If the camera is set in long record capture mode but the user has selected
    // a different capture mode a warning should appear about the loss of any long
    // recordings made on the internal SSD long record drive.
    if (ltr_enabled_now && !ui->ltr_enable->isChecked())
    {
        msg_str     = tr("***POSSIBLE DATA LOSS***\nLeaving Long Recording Mode permanently deletes any long recording. ");
        msg_part    = tr("It also requires a separate step to reformat the storage device.\n\n");

        // Make sure the user knows what they're doing.
        msg_str = msg_str + msg_part + tr("Select \"OK\" to exit Long Recording Mode discarding data.\n"
                                          "Select \"Cancel\" to remain in Long Recording Mode protecting data.");
        show_warn_msg = true;
    }

    // Was the long record capture mode selected?
    if (ui->ltr_enable->isChecked())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Long Record Being Set";

        // Set the enable bit for LTR.
        ltr_mode |= GIGE_LTR_MODE_ENABLE;

        // Clear out the GIGE_LTR_MODE_FFLR bit.
        ltr_mode &= ~GIGE_LTR_MODE_FFLR;

        // TO CHECK:  I believe that we only warn going to long record and that dirty buffers may exist
        // in both image memory and fflr mode.  TBD.

        // Build up the warning about losing any image memory recordings.
        msg_str = tr("***POSSIBLE DATA LOSS***\nLong Recording Mode will permanently delete the following\n\n");

        // Depending on the dirty count in the partition of image memory that could be lost report additional
        // information to the user (as per TICKET #435.
        if (dirty_count > 0)
        {
            msg_part = tr("- Image Memory: containing %1 unsaved recording%2\n").arg(dirty_count).arg((dirty_count > 1 ? "s." : "."));
        }

        msg_part = msg_part + tr("- %1: all saved image stacks, videos and capture files.\n\n").arg(device_sel);

        // Make sure the user knows what they're doing.
        msg_str = msg_str + msg_part + tr("Select \"OK\" to proceed into Long Recording Mode.\n"
                                          "Select \"Cancel\" to remain in current recording mode protecting data.");
        show_warn_msg = true;
    }
    else if (ui->fflr_enable->isChecked())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "FFLR Being Set";

        // Set the enable bit for FFLR.
        ltr_mode |= GIGE_LTR_MODE_FFLR;

        // Clear out the GIGE_LTR_MODE_ENABLE bit.
        ltr_mode &= ~GIGE_LTR_MODE_ENABLE;

        // If we had not been in FFLR mode or LTR mode then we were in image memory mode.
        // There could be a loss of data in that case and we need to warn the user.
        if (!fflr_enabled_now && !ltr_enabled_now && (dirty_count > 0))
        {
            // Build up the warning about losing any image memory recordings.
            msg_str = tr("***POSSIBLE DATA LOSS***\nFasFireLR will permanently delete Image Memory ");
            msg_part = tr("containing %1 unsaved recording%2\n\n").arg(dirty_count).arg((dirty_count > 1 ? "s." : "."));

            // Make sure the user knows what they're doing.
            msg_str = msg_str + msg_part + tr("Select \"OK\" to proceed into FasFireLR Mode.\n"
                                              "Select \"Cancel\" to remain in image memory mode protecting data.");
            show_warn_msg = true;
        }
    }
    else if (ui->im_enable->isChecked())
    {
        // Clear FFLR and LTR bits.
        ltr_mode &= ~(GIGE_LTR_MODE_ENABLE | GIGE_LTR_MODE_FFLR | GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC);
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Image Memory Being Set";
    }

    // Show the warning message if one was built.
    if (show_warn_msg)
    {
        user_resp = QMessageBox::warning(this, tr("Capture Mode Change"), msg_str, QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Cancel);
        if (user_resp == QMessageBox::Cancel)
        {
            // No need to block the updates anymore.
            update_in_progress = false;
            camera_ptr->get_info_widget()->enable_actions();

            // We're done.
            return;
        }
    }

    // Make sure the camera is out of autosave mode.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);
    if (((reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE))
    {
        // This is a disable request. Tell the camera to disable autosave.
        camera_ptr->write_camera_register(GIGE_REG_AUTOSAVE, GIGE_AUTOSAVE_DISABLE);
    }

    // Special handling for the series3 cameras.
    if (camera_ptr->is_series3())
    {
        // Make sure the camera dock knows this camera is rebooting.
        MainWindow::GetInstance()->get_camera_dock()->set_camera_autoconnect(camera_ptr->get_ip_addr());
    }

    // The camera needs to go into the IDLE state.
    camera_ptr->write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_IDLE);

    // Call our central handler to write the new LTR mode.
    write_ltr_mode(ltr_mode, __PRETTY_FUNCTION__);

    // We are switching mode, so reset any raw metadata we might have.
    QByteArray *raw_metaptr;
    raw_metaptr = camera_ptr->get_raw_metadata();
    if (raw_metaptr != nullptr)
        raw_metaptr->clear();

    // The series3 cameras will be rebooting, but not the seriesNC cameras. They need
    // to be transitioned to live mode.
    if (!camera_ptr->is_series3())
    {
        // Make sure we don't have any outstanding partitions.
        camera_ptr->free_partition(true);

        // Back to live.
        camera_ptr->write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_LIVE_VIDEO);
    }

    // TICKET #435:  Make sure any gas gauges are cleared out.
    CameraInfo *camera_info_ptr = camera_ptr->get_info_widget();
    if (camera_info_ptr)
    {
        // This should close out any gas gauges on screen for autosaving as we will be going long-record mode.
        camera_info_ptr->display_autosave_settings(0);
    }
}

/*!
 * @brief The user has selected the ROC radio button.
 */
void PanelCapture::ltr_roc_clicked()
{
    uint32_t    ltr_mode;
    uint32_t    part_state;

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_in_progress = true;

    // Read the current LTR mode value.
    ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);

    // If LTR mode warn the user about data loss.
    if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        // Read the partition state.
        part_state = read_camera_cache(GIGE_REG_PART_STATE);

        // If the partition is "dirty" and we're transitioning from basic mode, we need to warn the user.
        if (((part_state == GIGE_PART_STATE_COMPLETE) || (part_state == GIGE_PART_STATE_LOCKED))
            && ((ltr_mode & GIGE_LTR_MODE_BROC) != GIGE_LTR_MODE_BROC))
        {
            if (QMessageBox::question(this, tr("FasMotion"),
                                      tr("Camcorder and LR Burst modes append recordings onto the SSD, but LR Basic mode uses the entire drive."
                                         "<p>If you change modes, the next recording will overwrite any existing data on the SSD!"
                                         "<p>Are you sure you want to change the mode?"),
                                      QMessageBox::Yes | QMessageBox::No) == QMessageBox::No)
            {
                // Go back to basic mode.
                ui->rb_basic_mode->setChecked(true);

                // No need to block the updates anymore.
                update_in_progress = false;
                return;
            }
        }
    }

    // Disable the burst length controls.
    ui->frm_broc_len->setEnabled(false);

    // Clear out the BROC bit and set the ROC bit.
    ltr_mode = (ltr_mode & ~((uint32_t)GIGE_LTR_MODE_BROC)) | GIGE_LTR_MODE_ROC;

    // Call our central handler to write the new mode.
    write_ltr_mode(ltr_mode, __PRETTY_FUNCTION__);
}

/*!
 * @brief The user has modified the BROC length using the slider.
 * @param new_value     The new BROC burst length.
 */
void PanelCapture::ltr_sld_broc_len_changed(int new_value)
{
    // Sanity check.
    if (new_value < MINIMUM_BROC_LEN)
        new_value = MINIMUM_BROC_LEN;

    // Make sure the spinner is in sync.
    ui->spin_broc_len->setValue(new_value);
}

/*!
 * @brief The user has modified the BROC length with the spinner. This also is invoked when the
 *        spinner control's value is modified programatically.
 *
 * @param new_value     The new BROC burst length.
 */
void PanelCapture::ltr_spin_broc_len_changed(int new_value)
{
    // Sanity check.
    if (new_value < MINIMUM_BROC_LEN)
        new_value = MINIMUM_BROC_LEN;

    // Make sure the slider is updated.
    set_slider_value(ui->sld_broc_len, new_value);

    // Tell the camera the new choice.
    update_in_progress = true;
    last_write_addr = GIGE_REG_BROC_LEN;
    write_camera_register(GIGE_REG_BROC_LEN, (uint32_t)new_value);
}

/*!
 * @brief Toggle the visiblity of the panel. Optionally, pass
 *        the event up to the parent.
 *
 * @param notify_parent  True if we're supposed to pass this up to the parent.
 */
void PanelCapture::toggle_visibility(bool notify_parent)
{
    // Pass it on up.
    FMPanelWidget::toggle_visibility(notify_parent);

    // If we're opened up, make sure we refresh our controls
    if (isVisible())
    {
        // Do we have a camera connected?
        if (p_control_panel->getCamera() != nullptr)
        {
            // Update our panel.
            update_panel();
        }
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void PanelCapture::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Only care if we're visible.
    if (!isVisible())
    {
        // Make sure we're reset.
        update_in_progress = false;

        // No need to keep going.
        return;
    }

    // Are we doing an update?
    if (!update_in_progress)
    {
        if (last_write_addr == reg_addr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Whoops!  The update_in_progress flag wasn't set!";
        }
        return;
    }

    // Get our communications object pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Was it a successful write?
    write_status = comm_ptr->get_write_status(camera_ip_addr, reg_addr);
    if ((write_status != GEV_STATUS_SUCCESS) && (last_write_addr == reg_addr))
    {
        QHostAddress    our_addr;
        our_addr.setAddress(qFromBigEndian(camera_ip_addr));

        QString err_msg = QString::fromUtf8("Unable to write register 0x%1 to IP address %2 [status=%3]").arg(reg_addr, 0, 16).arg(our_addr.toString()).arg(write_status, 0, 16);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;

        // Which register had a problem.
        switch (reg_addr)
        {
        case GIGE_REG_LTR_MODE:
            // Refresh the register.
            wait_for_ltr_mode = false;

            // Enable the record mode and session length group.
            ui->grp_rec_mode->setEnabled(is_panel_ok_to_enable());
            ui->grp_internal_storage->setEnabled(is_panel_ok_to_enable());

            // Make sure we update the register setting.
            comm_ptr->start_read_reg(camera_ip_addr, GIGE_REG_LTR_MODE);

            // Notify the user.
            err_msg = FI_comm::gev_status_to_msg(write_status);
            QMessageBox::information(MainWindow::GetInstance().get(), tr("Switching Mode"), tr("Unable to switch, camera reports: ") + err_msg);
            break;
        default:
            break;
        }
    }

    switch (reg_addr)
    {
    // If the register written was GIGE_REG_SESSION_LENGTH, good or bad we need to re-enable the controls.
    case GIGE_REG_SESSION_LENGTH:
        // Re-enable the controls.
        ui->frm_session->setEnabled(true);
        break;
    case GIGE_REG_BROC_LEN:
        // Make sure we update the register setting.
        comm_ptr->start_read_reg(camera_ip_addr, GIGE_REG_BROC_LEN);
        break;
    default:
        break;
    }

    // Are there still write operations to perform?
    if (!write_list.isEmpty())
    {
        REG_VALUE_PAIR  reg_pair;

        // Grab the next operation off the list.
        reg_pair = write_list.takeFirst();

        // Issue the write command.
        write_camera_register(reg_pair.first, reg_pair.second);
    }
    else
    {
        // Do we need to re-enable the session slider?
        if (reg_addr == last_write_addr)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "last_write_addr:" << Qt::showbase << Qt::hex << last_write_addr;

            // Everything looks good. Reset our updating flag.
            update_in_progress = false;

            // We have processed the write requested.
            last_write_addr = FI_COMM::NO_REGISTER;

            // This will refresh all of our registers.
            update_panel();
        }
    }
}
