/*!
 * @file panel_storage.h
 * @brief This is the header file for the Storage configuration panel.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PANEL_STORAGE_H
#define PANEL_STORAGE_H

#include <QWidget>
#include <QString>
#include "panel_widget.h"

namespace Ui
{
class PanelCapture;
}

namespace CAPTURE_SETTINGS
{
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class fiProgressDialog;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class PanelCapture : public FMPanelWidget
{
    Q_OBJECT

public:
    explicit PanelCapture(QWidget *parent = 0);
    ~PanelCapture();

    // Handle camera connect/disconnects.
    void camera_connect(uint32_t ip_addr) override;
    void camera_disconnect() override;

    // Update the panel contents.
    void update_panel() override;

protected:
    void    changeEvent(QEvent *event) override;

protected slots:
    void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value) override;
    void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op) override;
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr) override;
    void    toggle_visibility(bool notify_parent = true) override;
    void    timerEvent(QTimerEvent *event) override;
    void    write_complete(uint32_t ip_addr, uint32_t reg_addr) override;

private slots:
    void    chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr);

    void    do_save_logs();
    void    do_user_canceled();

    void    device_selected(int index);
    void    handle_json_media(const QJsonDocument& jsonMedia);

    void    do_ext_arm_in_disabling();
    void    do_ext_arm_in_enabling();

    void    confirm_session_length();
    void    session_length_changed(int value);
    void    spin_length_changed(double value);

    void    ltr_basic_clicked();
    void    ltr_broc_clicked();
    void    capture_mode_clicked();
    void    ltr_roc_clicked();
    void    ltr_sld_broc_len_changed(int new_value);
    void    ltr_spin_broc_len_changed(int new_value);


private:
    uint64_t calc_max_fflr_frames();
    void    disable_all_panel_frames();
    void    handle_read(uint32_t reg_addr, uint32_t reg_value);
    bool    is_fflr_enabled();
    bool    is_ltr_dual_enabled();
    bool    is_imem_enabled();
    bool    is_panel_ok_to_enable();

    bool    get_ltr_modes(bool& ltr_enabled_now, bool& fflr_enabled_now, bool& ltr_dual_enabled_now, uint32_t& ltr_reg);

    void    process_camera_state(uint32_t new_state);

    void    read_registers();
    void    update_arm_in_state();
    void    update_broc_len();
    void    update_broc_time();

    void    update_info();
    void    update_ltr_controls(uint32_t ltr_mode);
    void    update_media(uint32_t reg_value);
    void    update_session(QString who_called);

    void    write_ltr_mode(uint32_t ltr_mode, QString who_called);

private:
    Ui::PanelCapture *ui;               //!< Our UI object.

    static const QString IM_MODE_LABEL;         //!< Text label used when in image memory mode.
    static const QString LTR_MODE_LABEL;        //!< Text label used when in LR mode.
    static const QString TXT_STYLE_COLOR_WARN;  //!< Color for the warning messages in this panel.

    bool        camera_connecting;      //!< Are we in the process of connecting a camera?

    // FasFireLR Capture Mode:  Session variables to manage session controls.
    uint32_t    fflr_session_len;       //!< The size of a session, in 1/4 GiB units.
    // Image Memory Capture Mode:  Session variables to manage session controls.
    uint32_t    imem_session_len;       //!< The size of a session, in 1/4 GiB units.

    uint32_t    ltr_mode_waiting;       //!< What is the GIGE_REG_LTR_MODE we are waiting to see?
    uint32_t    last_write_addr;        //!< Used to terminate a block register write, this contains the last register address in the block.

    // Keep a copy of so we can show dynamic changes without actually making changes to the camera.
    uint64_t    part_size;              //!< The size of a paritions.
    uint32_t    part_size_reserved;     //!< How much of the partition is reserved.

    std::unique_ptr<fiProgressDialog> progressDialog;   //!< A simple progress dialog for operational status.

    int         progress_timer_id;      //!< Timer to poll progress.
    bool        update_in_progress;     //!< Flag to indicate that an update (write operation) is in progress.
    int         update_timer_id;        //!< Timer ID to trigger an update.

    bool        wait_for_devices;       //!< When connecting a camera we need to wait for device query before enabling panel.
    bool        wait_for_ltr_read;      //!< When connecting a camera we need to have a full read of GIGE_REG_LTR_MODE before enabling.
    bool        wait_for_ltr_mode;      //!< When modifying the GIGE_REG_LTR_MODE this can flag a cycle wait until we see the mode set.
};

#endif // PANEL_STORAGE_H
