/*!
 * @file panel_display.cpp
 * @brief This file contains the implementation of the PanelDisplay
 *        class. This class handles all the settings for the display
 *        elements of the camera (white balance, BCG, etc).
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "panel_display.h"
#include "ui_panel_display.h"

#include <QAction>
#include <QColorDialog>
#include <QObject>
#include <qendian.h>
#include <utility>

#include "camera_mdi.h"
#include "camera_video.h"
#include "curve_settings.h"
#include "fi_comm.h"
#include "fi_lutfile.h"
#include "fi_sensor.h"
#include "fi_splinewidget.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_mdi.h"
#include "playback_meta_info.h"
#include "whitebal_dialog.h"

namespace DISP_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Log messages for testing purposes.  Set to 1 to have them compiled for output.
#define LOG_GROUP_READS 0

//! Define some enumerations to reflect the various HDMI modes we support.
//! It is VERY important that these agree with how the strings are layed out
//! in the control.
enum
{
    HDMI_INDEX_DISABLED = 0,            //!< HDMI is disabled.
    HDMI_INDEX_1080P60,                 //!< HDMI is set to 1080p60.
    HDMI_INDEX_1080P30,                 //!< HDMI is set to 1080p30.
    HDMI_INDEX_720P,                    //!< HDMI is set to 720P.
    HDMI_INDEX_640                      //!< HDMI is set to 640x480.
};

const   uint32_t    DEFAULT_WB_GAIN     = 0x1000;

//! The brightness and contrast register values range from 0..200. We want
//! to present this to the user as -100..100, so we need an offset.
const   int     BC_OFFSET               = 100;

const   int     RGB_GAIN_MULT           = 4;            //!< Conversion from CS value for the RGB gains to what is displayed to the user.
const   double  RGB_GAIN_INCREMENT      = 256.0;        //!< RGB gain units. The gain is defined as an 8.8 fixed point value.

//! Delay period to wait for frames to trickle down from the camera.
const   int     SPLINE_DIALOG_DELAY     = 500;

const   int     MAX_SPLINE_BUFFER_LEN   = 512;

}   // End of namespace DISP_SETTINGS.

using namespace DISP_SETTINGS;
using namespace FM_PANEL;

/*!
 * @brief Common logic to determine if a slider control in the object is engaged.
 */
static bool is_ctl_engaged(fiSlider *slider_ptr, QAbstractSpinBox *spin_ptr, QFrame *frame_ptr)
{
    bool is_engaged = false;

    if ((slider_ptr == nullptr) || (spin_ptr == nullptr) || (frame_ptr == nullptr))
        return is_engaged;

    if (frame_ptr->isEnabled())
    {
        // Is slider focused?
        if (slider_ptr->hasFocus())
        {
            is_engaged = slider_ptr->isSliderDown();
        }
        else if (spin_ptr->hasFocus() && spin_ptr->underMouse())
        {
            is_engaged = true;
        }
    }
    return is_engaged;
}

/*!
 * @brief Constructor for the PanelDisplay class. Initialize variables.
 */
PanelDisplay::PanelDisplay(QWidget *parent) :
    FMPanelWidget(parent),
    ui(new Ui::PanelDisplay),
    camera_ip_addr(0),
    last_lut_id(GIGE_LUT_NONE)
{
    // Setup our UI controls.
    ui->setupUi(this);
    starting_spline_dlg = false;

    // By default, hide the auto whitebalance button.
    ui->pb_autowb->setVisible(false);
    ui->lbl_hdmi_status->setVisible(false);

    // Hide the HDMI layout area.
    ui->frm_hdmi->setVisible(false);

    // Make sure we got a parent pointer.
    if (parent != nullptr)
    {
        // My grandparent is the mainwindow.
        MainWindow  *p_mainwnd = qobject_cast<MainWindow *>(parent->parent());

        // Connect the menu item's I'm interested in.
        connect(p_mainwnd->get_cp_display(), &QAction::triggered, this, &PanelDisplay::toggle_visibility);
    }

    // Initialize our variables.
    custom_whitebal_in_progress = false;
    update_timer_id = 0;
    update_in_progress = false;
    retry_counter = 0;
    playback_ptr = nullptr;

    // Initialize our controls.
    init_controls();

    // After everything is setup, setup the intial size information.
    set_initial_sizing();

    // Start out with things disabled until we get connected to a camera.
    setEnabled(false);
}

/*!
 * @brief Destructor for the PanelDisplay class.
 */
PanelDisplay::~PanelDisplay()
{
    delete ui;
}

/*!
 * @brief We are connecting to a (possibly) new camera.
 * @param ip_addr   The IP address of the camera we're connecting to.
 */
void PanelDisplay::camera_connect(uint32_t ip_addr)
{
    FI_comm *comm_ptr;
    CameraMDI *camera_ptr;

    last_lut_id = GIGE_LUT_NONE;

    // Make sure we're not already connected to this camera.
    if (ip_addr == camera_ip_addr)
        return;

    // Save the IP address.
    camera_ip_addr = ip_addr;

    // Get our communications pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Connect the communications interfaces that we're interested in.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelDisplay::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &PanelDisplay::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelDisplay::handle_unsol_msg, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelDisplay::write_complete, Qt::QueuedConnection);

    // Pass it on.
    FMPanelWidget::camera_connect(ip_addr);

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr != nullptr)
    {
        // Show the HDMI controls if the camera is Series3.
        ui->frm_hdmi->setVisible(camera_ptr->is_series3());
        connect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelDisplay::chk_refresh_regs, Qt::QueuedConnection);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Enable the panel.
    setEnabled(true);
}

/*!
 * @brief Disconnect from our camera.
 */
void PanelDisplay::camera_disconnect(void)
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr = nullptr;

    // Have we been here already?
    if (camera_ip_addr == 0)
        return;

    // Clear our IP address.
    camera_ip_addr = 0;

    // Make sure there is a camera pointer as we need to connect to some signals.
    if (p_control_panel)
    {
        camera_ptr = p_control_panel->getCamera();
    }

    // Disconnect any camera-specific signals.
    if (camera_ptr != nullptr)
    {
        disconnect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelDisplay::chk_refresh_regs);
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // Pass it up.
    FMPanelWidget::camera_disconnect();

    // Hide the HDMI layout area.
    ui->frm_hdmi->setVisible(false);

    // Get our communications pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr != nullptr)
    {
        // Connect the communications interfaces that we're interested in.
        disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelDisplay::read_complete);
        disconnect(comm_ptr, &FI_comm::msg_timeout, this, &PanelDisplay::msg_timeout_event);
        disconnect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelDisplay::handle_unsol_msg);
        disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelDisplay::write_complete);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
    }

    // Disable all controls.
    setEnabled(false);

    // Make sure our timer is stopped.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }
}

/*!
 * @brief Process register values we received from the camera.
 *
 * @param reg_addr      The register's address.
 * @param reg_value     The register's value.
 */
void PanelDisplay::handle_read(uint32_t reg_addr, uint32_t reg_value)
{
    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_BRIGHTNESS:
    case GIGE_REG_CONTRAST:
    case GIGE_REG_GAMMA:
        if (!ignore_updates && !update_in_progress)
            update_bcg();
        break;

    case GIGE_REG_RED:
    case GIGE_REG_GREEN:
    case GIGE_REG_BLUE:
        if (!ignore_updates && !update_in_progress)
            update_rgb();
        break;

    case GIGE_REG_HEIGHT:
        if (reg_value > 1080)
            update_hdmi(read_camera_cache(GIGE_REG_HDMI));
        break;

    case GIGE_REG_WIDTH:
        if (reg_value > 1920)
            update_hdmi(read_camera_cache(GIGE_REG_HDMI));
        break;

    case GIGE_REG_SENSOR_TYPE:
        update_sensor_type(reg_value);
        break;

    case GIGE_REG_WHITE_BAL:
        update_whitebal(reg_value);
        break;

    case GIGE_REG_HDMI:
        update_hdmi(reg_value);
        break;

    case GIGE_REG_CAMERA_STATE:
        // Don't change the white balance except in live or review modes.
        ui->cb_white_balance->setEnabled((reg_value == GIGE_CAMERA_STATE_LIVE_VIDEO) || (reg_value == GIGE_CAMERA_STATE_REVIEW));
        ui->pb_whitebal_cal->setEnabled((reg_value == GIGE_CAMERA_STATE_LIVE_VIDEO) || (reg_value == GIGE_CAMERA_STATE_REVIEW));
        break;

    case GIGE_REG_LUT:
        update_lut(reg_value);
        break;

    default:
        break;
    }
}

/*!
 * @brief Initialize the signal/slot connections and initializations for the controls.
 */
void PanelDisplay::init_controls(void)
{
    // Hide our "calibrate" button. It only gets displayed for Custom white balance.
    ui->pb_whitebal_cal->hide();
    white_bal_blue = DEFAULT_WB_GAIN;
    white_bal_green = DEFAULT_WB_GAIN;
    white_bal_red = DEFAULT_WB_GAIN;

    ui->cb_hdmi->setItemData(HDMI_INDEX_DISABLED, -1);
    ui->cb_hdmi->setItemData(HDMI_INDEX_1080P60, GIGE_HDMI_1080P60);
    ui->cb_hdmi->setItemData(HDMI_INDEX_1080P30, GIGE_HDMI_1080P30);
    ui->cb_hdmi->setItemData(HDMI_INDEX_720P, GIGE_HDMI_720P);
    ui->cb_hdmi->setItemData(HDMI_INDEX_640, GIGE_HDMI_640);

    connect(ui->pb_autowb, &QPushButton::clicked, this, &PanelDisplay::do_autowb);
    connect(ui->cb_white_balance, qOverload<int>(&QComboBox::currentIndexChanged), this, &PanelDisplay::cb_white_bal_changed);
    connect(ui->cb_hdmi, qOverload<int>(&QComboBox::currentIndexChanged), this, &PanelDisplay::cb_hdmi_changed);
    connect(ui->pb_reset, &QPushButton::clicked, this, &PanelDisplay::reset_to_defaults);
    connect(ui->pb_whitebal_cal, &QPushButton::clicked, this, &PanelDisplay::custom_whitebal);

    // We want all of our sliders to throttle their feedback when the user is dragging the "thumb".
    ui->sld_blue->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    ui->sld_brightness->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    ui->sld_contrast->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    ui->sld_gamma->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    ui->sld_green->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    ui->sld_red->setThrottle(FI_SLIDER::SCRUB_UPDATE);

    connect(ui->spin_brightness, qOverload<int>(&QSpinBox::valueChanged), this, &PanelDisplay::spin_brightness_changed);
    connect(ui->sld_brightness, &fiSlider::valueSlowChanged, this, &PanelDisplay::sld_brightness_changed);
    connect(ui->sld_brightness, &fiSlider::sliderPressed, this, &PanelDisplay::do_ignore_updates);
    connect(ui->sld_brightness, &fiSlider::released, this, &PanelDisplay::do_allow_updates);
    connect(ui->spin_contrast, qOverload<int>(&QSpinBox::valueChanged), this, &PanelDisplay::spin_contrast_changed);
    connect(ui->sld_contrast, &fiSlider::valueSlowChanged, this, &PanelDisplay::sld_contrast_changed);
    connect(ui->sld_contrast, &fiSlider::sliderPressed, this, &PanelDisplay::do_ignore_updates);
    connect(ui->sld_contrast, &fiSlider::released, this, &PanelDisplay::do_allow_updates);
    connect(ui->spin_red, qOverload<double>(&QDoubleSpinBox::valueChanged), this, &PanelDisplay::spin_red_changed);
    connect(ui->sld_red, &fiSlider::valueSlowChanged, this, &PanelDisplay::sld_red_changed);
    connect(ui->sld_red, &fiSlider::sliderPressed, this, &PanelDisplay::do_ignore_updates);
    connect(ui->sld_red, &fiSlider::released, this, &PanelDisplay::do_allow_updates);
    connect(ui->spin_green, qOverload<double>(&QDoubleSpinBox::valueChanged), this, &PanelDisplay::spin_green_changed);
    connect(ui->sld_green, &fiSlider::valueSlowChanged, this, &PanelDisplay::sld_green_changed);
    connect(ui->sld_green, &fiSlider::sliderPressed, this, &PanelDisplay::do_ignore_updates);
    connect(ui->sld_green, &fiSlider::released, this, &PanelDisplay::do_allow_updates);
    connect(ui->spin_blue, qOverload<double>(&QDoubleSpinBox::valueChanged), this, &PanelDisplay::spin_blue_changed);
    connect(ui->sld_blue, &fiSlider::valueSlowChanged, this, &PanelDisplay::sld_blue_changed);
    connect(ui->sld_blue, &fiSlider::sliderPressed, this, &PanelDisplay::do_ignore_updates);
    connect(ui->sld_blue, &fiSlider::released, this, &PanelDisplay::do_allow_updates);
    connect(ui->spin_gamma, qOverload<double>(&QDoubleSpinBox::valueChanged), this, &PanelDisplay::spin_gamma_changed);
    connect(ui->sld_gamma, &fiSlider::valueSlowChanged, this, &PanelDisplay::sld_gamma_changed);
    connect(ui->sld_gamma, &fiSlider::sliderPressed, this, &PanelDisplay::do_ignore_updates);
    connect(ui->sld_gamma, &fiSlider::released, this, &PanelDisplay::do_allow_updates);

    connect(ui->pb_use_curves, &QPushButton::clicked, this, &PanelDisplay::switch_to_spline);
    connect(ui->chk_colorize, &QPushButton::clicked, this, &PanelDisplay::do_colorize);
    connect(ui->pb_colorize_color, &QPushButton::clicked, this, &PanelDisplay::do_colorize_color);
    connect(ui->chk_focus_assist, &QPushButton::clicked, this, &PanelDisplay::do_focus_assist);
    connect(ui->pb_focus_assist_color, &QPushButton::clicked, this, &PanelDisplay::do_focus_assist_color);

    // Enable the assistant controls.
    update_colorize(true);
    update_focus_assist(true);
}

void PanelDisplay::playback_connect(QWidget *playback_ptr)
{
    PlaybackMDI *activePlaybackPtr  = this->playback_ptr = qobject_cast<PlaybackMDI *>(playback_ptr);
    if (activePlaybackPtr != nullptr)
    {
        uint32_t bcgValue;
        uint32_t rgbValue;

        // Enable the panel.
        setEnabled(true);

        // But not HDMI.
        ui->cb_hdmi->setEnabled(false);

        // Pass the new value on.
        activePlaybackPtr->colorize(ui->chk_colorize->isChecked());

        // Update the display brightness setting.
        bcgValue = activePlaybackPtr->getDisplayBrightness();
        updateDisplayBrightness(bcgValue);

        // Update the display contrast setting.
        bcgValue = activePlaybackPtr->getDisplayContrast();
        updateDisplayContrast(bcgValue);

        // Update the display contrast setting.
        bcgValue = activePlaybackPtr->getDisplayGamma();
        updateDisplayGamma(bcgValue);

        // Update the red gain setting.
        rgbValue = activePlaybackPtr->getDisplayRed();
        updateDisplayRed(rgbValue);

        // Update the green gain setting.
        rgbValue = activePlaybackPtr->getDisplayGreen();
        updateDisplayGreen(rgbValue);

        // Update the blue gain setting.
        rgbValue = activePlaybackPtr->getDisplayBlue();
        updateDisplayBlue(rgbValue);

        // If we have turned off applying the cap file's custom lut we just make sure the setting
        // reflects no custom lut.
        if (!activePlaybackPtr->getApplyingCapLuts())
        {
            reset_lut_empty();
        }
        // No, so pull out any information that may exist from the metadata.
        else
        {
            // Pull a custom lut from playback pointer if one exists.
            playbackLUT = activePlaybackPtr->GetMetadataLut();

            // Set up information in our panel reflecting the proper display settings.
            if (playbackLUT)
            {
                if (playbackLUT->GetLutId() > 0)
                {
#if 0
// This code is useful if you want to verify the playback LUT table sizes and if they built properly.
                    bool builtRedLUT, builtGreenLUT, builtBlueLUT = false;
                    QVector<double> save_10bit_red_lut     = playbackLUT->GetRedOrGammaCameraLut(builtRedLUT);
                    QVector<double> save_10bit_green_lut   = playbackLUT->GetGreenCameraLut(builtGreenLUT);
                    QVector<double> save_10bit_blue_lut    = playbackLUT->GetBlueCameraLut(builtBlueLUT);

                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "builtRedLut:" << builtRedLUT << "builtGreenLut:" << builtGreenLUT << "builtBlueLut:" << builtBlueLUT;
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_10bit_red_lut:" << save_10bit_red_lut.count() << "save_10bit_green_lut:" << save_10bit_green_lut.count() << "save_10bit_blue_lut:" << save_10bit_blue_lut.count();
#endif
                    update_lut_display(playbackLUT->GetLutId(), playbackLUT->GetLutDesc());
                }
            }
        }
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "QWidget:" << playback_ptr << "does not appear to be a PlaybackMDI object.";
    }
}

void PanelDisplay::playback_disconnect(void)
{
    // Reset our pointer.
    playback_ptr = nullptr;
    playbackLUT = nullptr;

    // Re-enable the HDMI.
    ui->cb_hdmi->setEnabled(true);

    // Do we need to disable the panel?
    if (camera_ip_addr == 0)
        setEnabled(false);
}

/*!
 * @brief Utility function to read all registers this panel is interested in.
 */
void PanelDisplay::read_registers(void)
{
    CameraMDI   *my_camera;

    // Do we have a camera connected?
    if ((my_camera = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }
    if (my_camera->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Are we a color sensor?
    if (my_camera->get_sensor()->is_color())
    {
        // Prepare the camera register group read.
        cam_register_group->addReg(GIGE_REG_RED);
        cam_register_group->addReg(GIGE_REG_GREEN);
        cam_register_group->addReg(GIGE_REG_BLUE);

        // For the white balance group, make sure the WHTE_BAL read happens after the RGB gain
        // reads. That way the gains will be in the cache if we need them.
        cam_register_group->addReg(GIGE_REG_WHITE_BALANCE_RED);
        cam_register_group->addReg(GIGE_REG_WHITE_BALANCE_GRN);
        cam_register_group->addReg(GIGE_REG_WHITE_BALANCE_BLU);
        cam_register_group->addReg(GIGE_REG_WHITE_BAL);
    }

    // Now add the others.
    cam_register_group->addReg(GIGE_REG_BRIGHTNESS);
    cam_register_group->addReg(GIGE_REG_CONTRAST);
    cam_register_group->addReg(GIGE_REG_GAMMA);
    cam_register_group->addReg(GIGE_REG_HEIGHT);
    cam_register_group->addReg(GIGE_REG_WIDTH);
    cam_register_group->lastReg(GIGE_REG_HDMI);

    // Pass on the request but only increment our counter if we sent it to the camera.
    read_camera_registers(cam_register_group);
}

/*!
 * @brief If we're in review, we need to refresh the frame being displayed.
 */
void PanelDisplay::refresh_frame(void)
{
    CameraMDI   *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = MainWindow::GetInstance()->active_mdi_child();
    if (camera_ptr != nullptr)
    {
        // Request the current frame be sent again. The update_frame()
        // method makes sure we're in the corect state, etc.
        camera_ptr->update_frame();
    }
}

void PanelDisplay::send_lut(int lut_id, QByteArray *lut_table)
{
    QByteArray  xfer_buf;

    if (camera_ip_addr != 0)
    {
        int count;
        uint32_t lut_offset;
        uint32_t reg_value;
        CameraMDI   *camera_ptr;

        // Get a pointer to our camera object.
        camera_ptr = MainWindow::GetInstance()->active_mdi_child();
        if (camera_ptr == nullptr)
            return;

        // Tell the camera the we're interested in uploading a new table.
        reg_value = (((uint32_t)lut_id) & GIGE_LUT_MASK);
        camera_ptr->write_camera_register(GIGE_REG_LUT_DATA, reg_value);

        // Send the LUT in chunks.
        count = lut_table->size();
        lut_offset = 0;
        while (count > 0)
        {
            uint8_t *data_ptr;
            uint32_t offset_val;
            size_t xfer_len;

            // Figure out how many bytes we can send of the data this time around.
            // Count is the total number of bytes of data, but this doesn't include
            // the "offset" (4 bytes).
            if (count > (MAX_SPLINE_BUFFER_LEN))
                xfer_len = MAX_SPLINE_BUFFER_LEN;
            else
                xfer_len = (size_t)(count + 4);

            // Build up our transfer buffer.
            xfer_buf.fill(0, (int)xfer_len);

            // Get a pointer to the buffer.
            data_ptr = (uint8_t *)xfer_buf.data();

            // Save the offset.
            offset_val = qToBigEndian(lut_offset);
            memcpy(data_ptr, &offset_val, sizeof(offset_val));
            data_ptr += sizeof(offset_val);
            xfer_len -= sizeof(offset_val);

            // Copy the data over. Don't forget to adjust for the offset so we copy
            // over the right "chunk" of data.
            memcpy(data_ptr, lut_table->constData() + lut_offset, xfer_len);

            // Send it on it's way.
            camera_ptr->write_camera_buffer(GIGE_REG_LUT_DATA, xfer_buf);

            // Adjust our offset.
            lut_offset += xfer_len;

            // Adjust our remaining count.
            count -= xfer_len;
        }

        // Tell the camera to use the new LUT.
        camera_ptr->write_camera_register(GIGE_REG_LUT, (uint32_t)lut_id);
    }
    else
    {
        int i;
        LUT_FILE        *lutData;       //!< Overlays QByteArray (the_array) as LUT_FILE.
        int             lutTableSize = LUT_MAX_TABLE_COUNT / 4; //!< Size of a LUT table vector.

        // Only do this if we have a playback pointer.
        if (playback_ptr == nullptr)
        {
            // There is no place to install the LUT data so just return.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to install LUT data we have no attached camera and no playback_ptr!";
            return;
        }

        // Allocate an array to hold the data. The lookup table takes a byte
        // value in, and converts that to a byte value going out, so it's
        // smaller than if we were on the camera.
        QVector<double> lut8BitRedOrMono(lutTableSize); //!< Red or monochrome LUT table.
        QVector<double> lut8BitGreen(lutTableSize);     //!< Green LUT table.
        QVector<double> lut8BitBlue(lutTableSize);      //!< Blue LUT table.

        // Overlay our structure onto the QByteArray.
        lutData = (LUT_FILE *)lut_table->data();

        // Handle the three channel scenario.
        if (lutData->hdr.lutDataType == LUT_SERIESNC_COLOR)
        {
            //! Overlays QByteArray (the_array) as LUT_RGB_FILE.
            LUT_RGB_FILE *lutRGBData = (LUT_RGB_FILE *)lut_table->data();

            // Subsample each red value and place it in the lut to process for red channel.
            for (i = 0; i < lutTableSize; ++i)
                lut8BitRedOrMono[i] = (double)(lutRGBData->redTable[i * 4]);

            // Subsample each green value and place it in the lut to process for green channel.
            for (i = 0; i < lutTableSize; ++i)
                lut8BitGreen[i] = (double)(lutRGBData->greenTable[i * 4]);

            // Subsample each blue value and place it in the lut to process for blue channel.
            for (i = 0; i < lutTableSize; ++i)
                lut8BitBlue[i] = (double)(lutRGBData->blueTable[i * 4]);

            // Pass the RGB LUTs to the playback engine.
            playback_ptr->set8bitRgbLuts(lut8BitRedOrMono, lut8BitGreen, lut8BitBlue);
        }
        // Otherwise we treat this is a gamma lut.
        else
        {
            // Subsample the LUT information and place it in our vector.
            for (i = 0; i < lutTableSize; ++i)
                lut8BitRedOrMono[i] = (double)(lutData->table[i * 4]);

            // Pass the LUT onwards.
            playback_ptr->set8bitGammaLut(lut8BitRedOrMono);
        }

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

/*!
 * @brief Translate the UI elements for this object.
 */
void PanelDisplay::translate_ui(void)
{
    int i;
    int hdmi_index;
    int wb_index;
    QList<QWidget *> widgetList = findChildren<QWidget *>();

    // Block all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(true);

    // Remember the current settings.
    hdmi_index = ui->cb_hdmi->currentIndex();
    wb_index = ui->cb_white_balance->currentIndex();

    // Translate our user interface.
    ui->retranslateUi(this);

    // Restore the settings.
    ui->cb_hdmi->setCurrentIndex(hdmi_index);
    ui->cb_white_balance->setCurrentIndex(wb_index);

    // Unblock all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(false);
}

/*!
 * @brief Update our display brightness based on the value read from the camera or the
 * playback object.
 *
 * @param camRegValue - Value read from camera register or playback data in range
 * of 0...200 and stored in an unsigned integer.
 */
void PanelDisplay::updateDisplayBrightness(uint32_t camRegValue)
{
    // The displayBrightness received will be in a range of 0...200 unsigned integer.
    // This value needs to be converted to a display UI range of -100 to +100.
    // on what was in camera register or in playback data.  We need to turn it
    // back into a display value of -100 to +100.
    int theValueToSet = (int) camRegValue - BC_OFFSET;

    // Skip the update if the control has the focus.
    if (!is_ctl_engaged(ui->sld_brightness, ui->spin_brightness, ui->frm_bcg))
    {
        ui->spin_brightness->blockSignals(true);
        ui->sld_brightness->blockSignals(true);
        ui->spin_brightness->setValue(theValueToSet);
        ui->sld_brightness->setValue(theValueToSet);
        ui->sld_brightness->blockSignals(false);
        ui->spin_brightness->blockSignals(false);
    }
}

/*!
 * @brief Update our display contrast based on the value read from the camera or the
 * playback object.
 *
 * @param camRegValue - Value read from camera register or playback data in range
 * of 0...200 and stored in an unsigned integer.
 */
void PanelDisplay::updateDisplayContrast(uint32_t camRegValue)
{
    // The displayContrast received will be in a range of 0...200 unsigned integer.
    // This value needs to be converted to a display UI range of -100 to +100.
    // on what was in camera register or in playback data.  We need to turn it
    // back into a display value of -100 to +100.
    int theValueToSet = (int) camRegValue - BC_OFFSET;

    // Skip the update if the control has the focus.
    if (!is_ctl_engaged(ui->sld_contrast, ui->spin_contrast, ui->frm_bcg))
    {
        ui->spin_contrast->blockSignals(true);
        ui->sld_contrast->blockSignals(true);
        ui->spin_contrast->setValue(theValueToSet);
        ui->sld_contrast->setValue(theValueToSet);
        ui->spin_contrast->blockSignals(false);
        ui->sld_contrast->blockSignals(false);
    }
}

/*!
 * @brief Update our display gamma based on the value read from the camera or the
 * playback object.
 *
 * @param camRegValue - Value read from camera register or playback data that is
 * multiplied by 100.
 */
void PanelDisplay::updateDisplayGamma(uint32_t camRegValue)
{
    // Extract the current setting and convert to a floating value for display to the user.
    int gam_int_value = camRegValue & 0xFFFF;
    double gamma_value = double(gam_int_value) / 100.0;

    // Skip the update if the user has the control engaged.
    if (!is_ctl_engaged(ui->sld_gamma, ui->spin_gamma, ui->frm_gamma))
    {
        ui->spin_gamma->blockSignals(true);
        ui->sld_gamma->blockSignals(true);
        ui->sld_gamma->setValue((int)gam_int_value);
        ui->spin_gamma->setValue(gamma_value);
        ui->spin_gamma->blockSignals(false);
        ui->sld_gamma->blockSignals(false);
    }
}

/*!
 * @brief Update our Brightness, Contrast & Gamma UI controls based on the value(s)
 *        read from the camera or the playback object.
 */
void PanelDisplay::update_bcg(void)
{
    uint32_t bcgValue;

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Read the brightness setting.
        bcgValue = (int)read_camera_cache(GIGE_REG_BRIGHTNESS);
    }
    else if (playback_ptr != nullptr)
    {
        bcgValue = playback_ptr->getDisplayBrightness();
    }
    else
    {
        // Just use the default value.
        bcgValue = GIGE_DEFAULT_BRIGHTNESS;
    }

    // Update the display UI elements for brightness.
    updateDisplayBrightness(bcgValue);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Read the contrast setting.
        bcgValue = (int)read_camera_cache(GIGE_REG_CONTRAST);
    }
    else if (playback_ptr != nullptr)
    {
        bcgValue = playback_ptr->getDisplayContrast();
    }
    else
    {
        // Just use the default value.
        bcgValue = GIGE_DEFAULT_CONTRAST;
    }

    // Update the display UI elements for contrast.
    updateDisplayContrast(bcgValue);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Read the gamma setting.
        bcgValue = (int)read_camera_cache(GIGE_REG_GAMMA);
    }
    else if (playback_ptr != nullptr)
    {
        bcgValue = playback_ptr->getDisplayGamma();
    }
    else
    {
        // Just use the default value.
        bcgValue = GIGE_DEFAULT_GAMMA;
    }

    // Update the display UI elements for gamma.
    updateDisplayGamma(bcgValue);

    // Finally, if we are not connected to a camera, update our LUT setting.
    if (camera_ip_addr == 0)
    {
        reset_lut_empty();
    }
}

void PanelDisplay::update_colorize(bool do_colorize)
{
    QSettings   settings;

    // Only do this if the application supports it.
    ui->chk_colorize->setVisible(do_colorize);
    ui->chk_colorize->setEnabled(do_colorize);
    ui->pb_colorize_color->setEnabled(do_colorize);

    if (do_colorize)
    {
        uint32_t pb_color;

        // Restore the colorize checkbox.
        settings.beginGroup("Colorize");
        ui->chk_colorize->setChecked(settings.value("enable", false).toBool());
        pb_color = settings.value("color", qRgb(255, 0, 0)).toUInt();
        settings.endGroup();

        // Set the background of the button to the value the user selected.
        ui->pb_colorize_color->setStyleSheet(QString("background-color:rgb(%1,%2,%3)").arg(qRed(pb_color)).arg(qGreen(pb_color)).arg(qBlue(pb_color)));
    }
    else
    {
        // Probably redundant, but let's be safe.
        ui->chk_colorize->setChecked(false);
    }
}

void PanelDisplay::update_focus_assist(bool do_assist)
{
    QSettings   settings;

    // Only do this if the application supports it.
    ui->chk_focus_assist->setVisible(do_assist);
    ui->chk_focus_assist->setEnabled(do_assist);
    ui->pb_focus_assist_color->setEnabled(do_assist);

    if (do_assist)
    {
        uint32_t pb_color;

        // Restore the focus assist checkbox.
        settings.beginGroup("FocusAssist");
        ui->chk_focus_assist->setChecked(settings.value("enable", false).toBool());
        pb_color = settings.value("color", qRgb(255, 0, 0)).toUInt();
        settings.endGroup();

        // Set the background of the button to the value the user selected.
        ui->pb_focus_assist_color->setStyleSheet(QString("background-color:rgb(%1,%2,%3)").arg(qRed(pb_color)).arg(qGreen(pb_color)).arg(qBlue(pb_color)));
    }
    else
    {
        // Probably redundant, but let's be safe.
        ui->chk_focus_assist->setChecked(false);
    }
}

/*!
 * @brief Update our HDMI control(s) based on the value read from the camera.
 * @param new_setting       The value read from the camera.
 */
void PanelDisplay::update_hdmi(uint32_t new_setting)
{
    int new_index;

    // If we have an ROI with a dimension larger than 1080P, disable HDMI.
    if ((read_camera_cache(GIGE_REG_HEIGHT) > 1080) || (read_camera_cache(GIGE_REG_WIDTH) > 1920))
    {
        // Are the HDMI controls enabled?
        if (ui->cb_hdmi->isEnabled())
        {
            // Turn off the HDMI controls.
            ui->cb_hdmi->setEnabled(false);
            if (ui->cb_hdmi->currentIndex() != HDMI_INDEX_DISABLED)
                ui->cb_hdmi->setCurrentIndex(HDMI_INDEX_DISABLED);

            // Tell the camera to shut it down.
            if ((new_setting & GIGE_HDMI_ENABLE_MASK) == (uint32_t)GIGE_HDMI_ENABLE)
                write_camera_register(GIGE_REG_HDMI, 0);
        }

        // Let the user know what we've done.
        ui->lbl_hdmi_status->setVisible(true);

        // All done here.
        return;
    }

    // Clear the status bar.
    ui->lbl_hdmi_status->setVisible(false);

    // Make sure we're enabled.
    ui->cb_hdmi->setEnabled(true);

    // Is the HDMI enabled?
    if ((new_setting & GIGE_HDMI_ENABLE_MASK) == (uint32_t)GIGE_HDMI_ENABLE)
    {
        // Translate from the FCP setting to our combobox index.
        new_index = ui->cb_hdmi->findData(new_setting & GIGE_HDMI_RESOLUTION_MASK);
    }
    else
    {
        // HDMI is disabled.
        new_index = HDMI_INDEX_DISABLED;
    }

    // Update the combobox if need be.
    if (ui->cb_hdmi->currentIndex() != new_index)
        ui->cb_hdmi->setCurrentIndex(new_index);
}

/*!
 * @brief Update custom LUT based on the value read from the camera.
 * @param reg_value       The GIGE_REG_LUT value read from the camera.
 */
void PanelDisplay::update_lut(uint32_t reg_value)
{
    CameraMDI   *my_camera;
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "starting_spline_dlg:"  << starting_spline_dlg;

    // Do we have a camera connected?
    if ((my_camera = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // If not starting the spline dialog then go ahead and service the lut
    // setting from the camera.
    if (!starting_spline_dlg)
    {
        // Since we're not starting the dialog, let's update our display.
        if ((reg_value & GIGE_LUT_MASK) == GIGE_LUT_NONE)
        {
            reset_lut_empty();
        }
        else if (last_lut_id != (reg_value & GIGE_LUT_MASK))
        {
            // Remember the new LUT number(id).
            last_lut_id = (reg_value & GIGE_LUT_MASK);

            // Ask the camera to send us down the first bit of the LUT data.
            // That's where the description field is.
            // First, we need to tell the camera the we want to read the LUT data.
            my_camera->write_camera_register(GIGE_REG_LUT_DATA, last_lut_id);

            // Next we'll want to read the LUT data. Just the description though. Make
            // sure we account for the offset field before the LUT, and keep it in multiples
            // of 4 bytes.
            my_camera->start_camera_read_buf(GIGE_REG_LUT_DATA, ((LUT_DESCRIPTION_LEN + 7) / 4) * 4);
        }

        // All done.
        return;
    }

    // If fall into this code then the custom curve settings dialog is being
    // deployed and we need to preserve the settings in case the user cancels
    // the operation.

    // Save the register value if we haven't already done so.
    if (lut_reg == GIGE_LUT_NONE)
        lut_reg = reg_value;

    // Preserve the BCG values, in case the user cancels.
    save_contrast = read_camera_cache(GIGE_REG_CONTRAST);
    save_gamma = read_camera_cache(GIGE_REG_GAMMA) & 0xFFFF;
    save_brightness = read_camera_cache(GIGE_REG_BRIGHTNESS);

    // Preserve the RGB values, in case the user cancels.
    save_red = read_camera_cache(GIGE_REG_RED);
    save_green = read_camera_cache(GIGE_REG_GREEN);
    save_blue = read_camera_cache(GIGE_REG_BLUE);

    // Adjust the camera settings so that it's in a "pass-through" mode.
    // Build up the list of register we need to reset.
    write_list.append(REG_VALUE_PAIR(GIGE_REG_CONTRAST, GIGE_DEFAULT_CONTRAST));
    write_list.append(REG_VALUE_PAIR(GIGE_REG_GAMMA, GIGE_DEFAULT_GAMMA));
    write_list.append(REG_VALUE_PAIR(GIGE_REG_RED, GIGE_DEFAULT_RGB_GAIN));
    write_list.append(REG_VALUE_PAIR(GIGE_REG_GREEN, GIGE_DEFAULT_RGB_GAIN));
    write_list.append(REG_VALUE_PAIR(GIGE_REG_BLUE, GIGE_DEFAULT_RGB_GAIN));

    // We are starting an update process.
    update_in_progress = true;

    // Start the write operation with the first operation.
    write_camera_register(GIGE_REG_BRIGHTNESS, GIGE_DEFAULT_BRIGHTNESS);
}

/*!
 * @brief Update custom LUT based on the value read from the camera.
 * @param reg_value       The GIGE_REG_LUT_DATA value read from the camera.
 */
void PanelDisplay::update_lut_display(QByteArray *lut_data)
{
    // Sanity check.
    if (lut_data != nullptr)
    {
        char lut_description[LUT_DESCRIPTION_LEN + 1] = {0};
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut_data.size()" << lut_data->size();
        // sanity check #2
        if (lut_data->size() > (LUT_DESCRIPTION_LEN + 4))
        {
            const char *desc_ptr = lut_data->constData() + 4;   // skip offset.
            for (int i = 0; i < LUT_DESCRIPTION_LEN; i++)
            {
                lut_description[i] = desc_ptr[i];
            }
            lut_description[LUT_DESCRIPTION_LEN] = '\0';
        }

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "last_lut_id:" << last_lut_id << "lut_description:" << lut_description;

        // Update the label.
        ui->lbl_lut->setText(QString("%1: %2").arg(last_lut_id).arg(lut_description));
    }
    else
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "lut_data was null!";
    }
}

/*!
 * @brief Update custom LUT based on information from playback object.
 * @param lutId     The number of the custom LUT in use for playback object.
 * @param lutLabel  The name of the custom LUT or its status if creating it.
 */
void PanelDisplay::update_lut_display(int lutId, const QString& lutLabel)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lutId:" << lutId << "lutLabel:" << lutLabel;

    // Update the label.
    ui->lbl_lut->setText(QString("%1: %2").arg(lutId).arg(lutLabel));
}

/*!
 * @brief Virtual function called by the "parent" widget to have us update the
 *        contents of our panel.
 */
void PanelDisplay::update_panel(void)
{
    bool            reg_ok;
    uint32_t        sensor_type;
    FI_comm         *comm_ptr;

    // Do we have a camera connected?
    if (p_control_panel->getCamera() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "no camera";
        return;
    }

    // Stop polling the camera.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }

    // Get our communications pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get a FI_comm object!";
        return;
    }

    // See if we know the sensor type.
    sensor_type = read_camera_cache(GIGE_REG_SENSOR_TYPE, &reg_ok);
    if (!reg_ok)
    {
        // Read the camera sensor type to get things rolling.
        comm_ptr->start_read_reg(camera_ip_addr, GIGE_REG_SENSOR_TYPE);
    }
    else
    {
        // Update our display based on the sensor type.
        update_sensor_type(sensor_type);

        // Read the registers from the camera now.
        read_registers();
    }

    // Read the LUT register once on startup.
    comm_ptr->start_read_reg(camera_ip_addr, GIGE_REG_LUT);

    // Restart our timer.
    update_timer_id = startTimer(POLLING_PERIOD);
}

/*!
 * @brief Update our color display red UI based on camera register value range.
 *
 * @param camRegValue: The value representative of what is stored on camera for red color gain.
 */
void PanelDisplay::updateDisplayRed(uint32_t camRegValue)
{
    int sld_value = camRegValue / RGB_GAIN_MULT;
    double spin_value = double(camRegValue) / RGB_GAIN_INCREMENT;

    // Skip the update if the control has the focus.
    if (!ui->spin_red->hasFocus())
    {
        ui->sld_red->blockSignals(true);
        ui->spin_red->blockSignals(true);
        ui->sld_red->setValue(sld_value);
        ui->spin_red->setValue(spin_value);
        ui->spin_red->blockSignals(false);
        ui->sld_red->blockSignals(false);
    }
}

/*!
 * @brief Update our color display green UI based on camera register value range.
 *
 * @param camRegValue: The value representative of what is stored on camera for green color gain.
 */
void PanelDisplay::updateDisplayGreen(uint32_t camRegValue)
{
    int sld_value = camRegValue / RGB_GAIN_MULT;
    double spin_value = double(camRegValue) / RGB_GAIN_INCREMENT;

    // Skip the update if the control has the focus.
    if (!ui->spin_green->hasFocus())
    {
        ui->sld_green->blockSignals(true);
        ui->spin_green->blockSignals(true);
        ui->sld_green->setValue(sld_value);
        ui->spin_green->setValue(spin_value);
        ui->spin_green->blockSignals(false);
        ui->sld_green->blockSignals(false);
    }
}

/*!
 * @brief Update our color display blue UI based on camera register value range.
 *
 * @param camRegValue: The value representative of what is stored on camera for blue color gain.
 */
void PanelDisplay::updateDisplayBlue(uint32_t camRegValue)
{
    int sld_value = camRegValue / RGB_GAIN_MULT;
    double spin_value = double(camRegValue) / RGB_GAIN_INCREMENT;

    // Skip the update if the control has the focus.
    if (!ui->spin_blue->hasFocus())
    {
        ui->sld_blue->blockSignals(true);
        ui->spin_blue->blockSignals(true);
        ui->sld_blue->setValue(sld_value);
        ui->spin_blue->setValue(spin_value);
        ui->spin_blue->blockSignals(false);
        ui->sld_blue->blockSignals(false);
    }
}

/*!
 * @brief Update our RGB gain controls based on the values read from the camera.
 */
void PanelDisplay::update_rgb(void)
{
    bool    reg_ok = true;
    int     reg_value;

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Read the register setting.
        reg_value = (int)read_camera_cache(GIGE_REG_RED, &reg_ok);
    }
    else if (playback_ptr != nullptr)
    {
        reg_value = playback_ptr->getDisplayRed();
    }
    else
    {
        // Just set the reg_ok false here so we get all of the defaults set up.
        reg_ok = false;
    }

    // If the register isn't ok we can just set everything to a default setting.
    if (!reg_ok)
    {
        reg_value = GIGE_DEFAULT_RGB_GAIN;
    }

    // Update the ui elements for red gain.
    updateDisplayRed(reg_value);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Read the register setting.
        reg_value = (int)read_camera_cache(GIGE_REG_GREEN, &reg_ok);
    }
    else if (playback_ptr != nullptr)
    {
        reg_value = playback_ptr->getDisplayGreen();
    }
    else
    {
        // Just set the reg_ok false here so we get all of the defaults set up.
        reg_ok = false;
    }

    // If the register isn't ok we can just set everything to a default setting.
    if (!reg_ok)
    {
        reg_value = GIGE_DEFAULT_RGB_GAIN;
    }

    // Update the ui elements for green gain.
    updateDisplayGreen(reg_value);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Read the register setting.
        reg_value = (int)read_camera_cache(GIGE_REG_BLUE, &reg_ok);
    }
    else if (playback_ptr != nullptr)
    {
        reg_value = playback_ptr->getDisplayBlue();
    }
    else
    {
        // Just set the reg_ok false here so we get all of the defaults set up.
        reg_ok = false;
    }

    // If the register isn't ok we can just set everything to a default setting.
    if (!reg_ok)
    {
        reg_value = GIGE_DEFAULT_RGB_GAIN;
    }

    // Update the ui elements for blue gain.
    updateDisplayBlue(reg_value);
}

/*!
 * @brief Update our sensor specific controls based on the type of sensor on the camera.
 *
 * If the camera has a monochrome sensor, then we want to hide the color specific controls,
 * such as the controls for RGB gain and white balance.
 *
 * @param sensor_type       The type of sensor on the camera.
 */
void PanelDisplay::update_sensor_type(uint32_t sensor_type)
{
    // Process based on sensor type.
    switch (sensor_type)
    {
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO:
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO:
    case FI_SENSOR::SENSOR_TYPE_LUX13HS_MONO:
    case FI_SENSOR::SENSOR_TYPE_LUX19HS_MONO:
    case FI_SENSOR::SENSOR_TYPE_LUX160_MONO:
    case FI_SENSOR::SENSOR_TYPE_LUX2100_MONO:
    case FI_SENSOR::SENSOR_TYPE_GENERIC_MONO:
        // Hide all the color specific options.
        ui->frm_rgb->setVisible(false);
        ui->frm_whitebal->setVisible(false);
        ui->pb_autowb->setVisible(false);
        break;
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX160_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR:
    case FI_SENSOR::SENSOR_TYPE_GENERIC_COLOR:
        // Show all the color specific options.
        ui->frm_rgb->setVisible(true);
        if (camera_ip_addr != 0)
        {
            // User needs/wants the white balance choices.
            ui->frm_whitebal->setVisible(true);

            // Hide the auto-white blanace button.
            ui->pb_autowb->setVisible(false);
            ui->frm_gamma->setEnabled(true);
        }
        else if (playback_ptr != nullptr)
        {
            ui->frm_whitebal->setVisible(false);
            ui->pb_autowb->setVisible(true);
            ui->frm_gamma->setEnabled(true);
        }
        break;
    default:
        // Disable all controls.
        setEnabled(false);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unknown sensor type:" << sensor_type;
        break;
    }
}

/*!
 * @brief Update our white balance controls based on the setting read from the camera.
 * @param new_setting       The white balance setting read from the camera.
 */
void PanelDisplay::update_whitebal(uint32_t new_setting)
{
    // Don't bother if nothing has changed.
    if (new_setting == (uint32_t)ui->cb_white_balance->currentIndex())
        return;

    // Skip the update if the control has the focus.
    if (ui->cb_white_balance->hasFocus() || ui->pb_whitebal_cal->hasFocus() || custom_whitebal_in_progress)
        return;

    // The camera has changed the white balance setting, update our
    // combo box so we match.
    ui->cb_white_balance->blockSignals(true);
    ui->cb_white_balance->setCurrentIndex((int)new_setting);
    ui->cb_white_balance->blockSignals(false);

    // Is the camera is using a custom white balance?
    if (new_setting == GIGE_WHITE_BAL_CUSTOM)
    {
        // User selected custom white balance. Show the calibrate button.
        ui->pb_whitebal_cal->show();
    }
    else
    {
        // Hide the calibrate button.
        ui->pb_whitebal_cal->hide();
    }
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void PanelDisplay::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    FMPanelWidget::changeEvent(event);
}

/*!
 * @brief Event routine which is invoked when our timer expires.
 * @param event     Pointer to the event class.
 */
void PanelDisplay::timerEvent(QTimerEvent *event)
{
    // Pass it on up.
    FMPanelWidget::timerEvent(event);

    // Is it the one we're interested in?
    if (event->timerId() != update_timer_id)
        return;

    // Do we have a camera connected?
    if (p_control_panel->getCamera() == nullptr)
        return;

    // Are we in the middle of an update?
    if (update_in_progress)
        return;

    // Are we visible?
    if (isVisible())
    {
        // Update our registers.
        read_registers();
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has modified the HDMI setting.
 * @param index     The newly selected value in the HDMI combobox.
 */
void PanelDisplay::cb_hdmi_changed(int index)
{
    int     hdmi_resolution;
    bool    hdmi_enabled;
    uint32_t    reg_value;
    uint32_t    old_value;

    // What is the HDMI resolution selected by the user?
    hdmi_resolution = ui->cb_hdmi->itemData(index).toInt();
    if (hdmi_resolution < 0)
    {
        hdmi_enabled = false;
        hdmi_resolution = GIGE_HDMI_1080P60;
    }
    else
    {
        hdmi_enabled = true;
    }

    // Get what the camera is currently set to.
    old_value = read_camera_cache(GIGE_REG_HDMI);

    // When we do this, the camera may appear to go off-line. Sometimes it
    // takes a very long time.

    // Update the camera.
    reg_value = (uint32_t)hdmi_resolution;
    if (hdmi_enabled)
        reg_value |= GIGE_HDMI_ENABLE;

    // Any change?
    if (old_value != reg_value)
    {
        write_camera_register(GIGE_REG_HDMI, reg_value);
    }
}

/*!
 * @brief The user has changed the white balance setting.
 * @param index     The newly selected value in the white balance combobox.
 */
void PanelDisplay::cb_white_bal_changed(int index)
{
    uint32_t    old_value;

    // Are we offline?
    if (playback_ptr != nullptr)
    {
        // Tell the playback object we have a new white balance selection.
        playback_ptr->setWhitebal(index);

        // Update the frame.
        playback_ptr->refreshFrame();

        // All done.
        return;
    }

    // Get what the camera is currently set to.
    old_value = read_camera_cache(GIGE_REG_WHITE_BAL);
    if (old_value == GIGE_WHITE_BAL_CUSTOM)
    {
        // Save the current gains before we go changing things around.
        white_bal_blue = read_camera_cache(GIGE_REG_WHITE_BALANCE_BLU);
        white_bal_green = read_camera_cache(GIGE_REG_WHITE_BALANCE_GRN);
        white_bal_red = read_camera_cache(GIGE_REG_WHITE_BALANCE_RED);
    }

    // Tell the camera the new white balance selected.
    write_camera_register(GIGE_REG_WHITE_BAL, (uint32_t)index);

    // The combo box index changed so update member variable
    if (index == GIGE_WHITE_BAL_CUSTOM)
    {
        // User selected custom white balance. Show the calibrate button.
        ui->pb_whitebal_cal->show();

        // Restore the last custom white balance gain settings.
        write_camera_register(GIGE_REG_WHITE_BALANCE_BLU, white_bal_blue);
        write_camera_register(GIGE_REG_WHITE_BALANCE_GRN, white_bal_green);
        write_camera_register(GIGE_REG_WHITE_BALANCE_RED, white_bal_red);
    }
    else
    {
        // Hide the calibrate button.
        ui->pb_whitebal_cal->hide();
    }

    // Make sure we refresh the frame if need be.
    refresh_frame();
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the read registers group read's last address == GIGE_REG_HDMI.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void PanelDisplay::chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr  = p_control_panel->getCamera();

    // If we are not connected to any camera we don't care about this signal.
    if (camera_ptr != nullptr)
    {
        // Is this event for us?
        if (ip_addr != camera_ptr->get_ip_addr())
            return;

        // Verify it's our register.
        if (reg_addr == GIGE_REG_HDMI)
        {
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_HDMI]";
#endif
            // This is a actionless group read there is no action that takes place...
            // There is a handler in the "handle_read" function for GIGE_REG_HDMI.
        }
    }
}

/*!
 * @brief The user has chosen to perform a custom white balance.
 */
void PanelDisplay::custom_whitebal(void)
{
    WhiteBalDlg dlg(MainWindow::GetInstance().get());
    CameraMDI   *camera_ptr;
    QAction     *view_histogram;

    // Get a pointer to our camera object.
    camera_ptr = MainWindow::GetInstance()->active_mdi_child();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get the CameraMDI object!";
        return;
    }

    // Remember that we're in the middle of a custom white balance operation.
    custom_whitebal_in_progress = true;

    // Get the histo_dock pointer, and make sure it's hidden.
    view_histogram = MainWindow::GetInstance()->get_view_histogram();
    if (view_histogram && view_histogram->isChecked())
    {
        // Turn off the histogram.
        view_histogram->setChecked(false);
        emit view_histogram->triggered();
    }

    // Reset the custom white balance gain settings.
    write_camera_register(GIGE_REG_WHITE_BALANCE_BLU, DEFAULT_WB_GAIN);
    write_camera_register(GIGE_REG_WHITE_BALANCE_GRN, DEFAULT_WB_GAIN);
    write_camera_register(GIGE_REG_WHITE_BALANCE_RED, DEFAULT_WB_GAIN);

    // Give the dialog the curent scene.
    dlg.set_scene(camera_ptr->get_player()->get_scene());
    dlg.set_ip_addr(camera_ptr->get_ip_addr());
    dlg.setGeometry(MainWindow::GetInstance()->geometry());

    // Time to hide the main window.
    MainWindow::GetInstance()->setVisible(false);

    // Bring up the dialog.
    if (dlg.exec() == QDialog::Accepted)
    {
        // Generate the gains based on the current ROI used by the histogram logic.
        write_camera_register(GIGE_REG_GENERATE_CUSTOM_WB, dlg.get_cct());
        ui->lbl_cct->setText(QString("[CCT=") + QString::number(dlg.get_cct()) + QString("K]"));
    }
    else
    {
        // Restore the last custom white balance gain settings.
        write_camera_register(GIGE_REG_WHITE_BALANCE_BLU, white_bal_blue);
        write_camera_register(GIGE_REG_WHITE_BALANCE_GRN, white_bal_green);
        write_camera_register(GIGE_REG_WHITE_BALANCE_RED, white_bal_red);
    }

    // All done.
    custom_whitebal_in_progress = false;

    // Bring back the main window.
    MainWindow::GetInstance()->setVisible(true);
}

void PanelDisplay::do_autowb()
{
    // Sanity check.
    if (playback_ptr == nullptr)
        return;

    // Tell the playback object the new white balance to use.
    playback_ptr->performAutoWb();

    // Update the frame.
    playback_ptr->refreshFrame();
}

void PanelDisplay::do_colorize()
{
    QSettings   settings;

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        CameraMDI   *camera_ptr;

        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr != nullptr)
        {
            if (camera_ptr->get_player() != nullptr)
            {
                camera_ptr->get_player()->colorize(ui->chk_colorize->isChecked());
            }
        }
    }
    else if (playback_ptr != nullptr)
    {
        // Pass the new value on.
        playback_ptr->colorize(ui->chk_colorize->isChecked());

        // Force a refresh.
        playback_ptr->refreshFrame();
    }

    // Save the setting
    settings.beginGroup("Colorize");
    settings.setValue("enable", ui->chk_colorize->isChecked());
    settings.endGroup();

}

/*!
 * @brief Let the user change the color for the colorize operation.
 */
void PanelDisplay::do_colorize_color()
{
    QSettings   settings;
    uint32_t pb_color;

    // Get the current color value.
    settings.beginGroup("Colorize");
    pb_color = settings.value("color", qRgb(255, 0, 0)).toUInt();
    settings.endGroup();

    const QColor color = QColorDialog::getColor(pb_color, MainWindow::GetInstance().get(), tr("Select Color"));
    if (color.isValid())
    {
        // Set the background of the button to the value the user selected.
        ui->pb_colorize_color->setStyleSheet(QString("background-color:rgb(%1,%2,%3)").arg(color.red()).arg(color.green()).arg(color.blue()));

        // Save this for next time.
        settings.beginGroup("Colorize");
        settings.setValue("color", color.rgb());
        settings.endGroup();
    }
}

void PanelDisplay::do_focus_assist()
{
    QSettings   settings;

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        CameraMDI   *camera_ptr;

        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr != nullptr)
        {
            if (camera_ptr->get_player() != nullptr)
            {
                camera_ptr->get_player()->enable_focus_assist(ui->chk_focus_assist->isChecked());
            }
        }
    }

    // Save the setting
    settings.beginGroup("FocusAssist");
    settings.setValue("enable", ui->chk_focus_assist->isChecked());
    settings.endGroup();

}

/*!
 * @brief Let the user change the color for the focus assist operation.
 */
void PanelDisplay::do_focus_assist_color()
{
    QSettings   settings;
    uint32_t pb_color;

    // Get the current color value.
    settings.beginGroup("FocusAssist");
    pb_color = settings.value("color", qRgb(255, 0, 0)).toUInt();
    settings.endGroup();

    const QColor color = QColorDialog::getColor(pb_color, MainWindow::GetInstance().get(), tr("Select Color"));
    if (color.isValid())
    {
        // Set the background of the button to the value the user selected.
        ui->pb_focus_assist_color->setStyleSheet(QString("background-color:rgb(%1,%2,%3)").arg(color.red()).arg(color.green()).arg(color.blue()));

        // Save this for next time.
        settings.beginGroup("FocusAssist");
        settings.setValue("color", color.rgb());
        settings.endGroup();
    }
}

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void PanelDisplay::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Only care if we're visible.
    if (!isVisible())
        return;

    // Pass it along like it was a read complete event.
    handle_read(reg_addr, reg_value);
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param ip_addr       The IP address of the camera which didn't respond.
 * @param reg_addr      The address of the register which was read.
 * @param read_op       True if this was a read operation, false for a write operation.
 */
void PanelDisplay::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Only care if we're visible.
    if (!isVisible())
        return;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << ip_addr << reg_addr << read_op;
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void PanelDisplay::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Pass it up.
    FMPanelWidget::read_complete(ip_addr, reg_addr);

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Only care if we're visible.
    if (!isVisible())
        return;

    // Handle the read mems different than read regs.
    if (reg_addr == GIGE_REG_LUT_DATA)
    {
        QByteArray  *lut_data;
        CameraMDI   *my_camera;

        // Do we have a camera connected?
        if ((my_camera = p_control_panel->getCamera()) == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
            return;
        }

        // Get the lut_data.
        lut_data = my_camera->read_camera_buffer(GIGE_REG_LUT_DATA);
        if (lut_data != nullptr)
        {
            update_lut_display(lut_data);
            delete lut_data;
        }
    }
    else
    {
        // Retrieve the camera state.
        uint32_t reg_value = read_camera_cache(reg_addr);

        // Handle the read event.
        handle_read(reg_addr, reg_value);
    }
}

/*!
 * @brief Reset lut control empty.
 */
void PanelDisplay::reset_lut_empty(void)
{
    lut_reg = GIGE_LUT_NONE;
    ui->lbl_lut->setText(tr("No curve selected."));
}

/*!
 * @brief Reset our controls to their default values.
 */
void PanelDisplay::reset_to_defaults(void)
{

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        CameraMDI   *my_camera;
        if ((my_camera = p_control_panel->getCamera()) == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null camera pointer!";
            return;
        }
        if (my_camera->get_sensor() == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
            return;
        }

        // Build the default gamma from the register cache.
        uint32_t default_gamma = ((read_camera_cache(GIGE_REG_GAMMA) >> 16) & 0xFFFF);

        // Build up the list of register we need to reset.
        write_list.append(REG_VALUE_PAIR(GIGE_REG_CONTRAST, GIGE_DEFAULT_CONTRAST));
        write_list.append(REG_VALUE_PAIR(GIGE_REG_GAMMA, default_gamma));

        // Are we a color sensor?
        write_list.append(REG_VALUE_PAIR(GIGE_REG_RED, GIGE_DEFAULT_RGB_GAIN));
        write_list.append(REG_VALUE_PAIR(GIGE_REG_GREEN, GIGE_DEFAULT_RGB_GAIN));
        write_list.append(REG_VALUE_PAIR(GIGE_REG_BLUE, GIGE_DEFAULT_RGB_GAIN));

        // We are starting an update process.
        update_in_progress = true;

        // Start the write operation with the first operation.
        write_camera_register(GIGE_REG_BRIGHTNESS, GIGE_DEFAULT_BRIGHTNESS);
    }
    else if (playback_ptr != nullptr)
    {
        // First set the default display settings in the playback object.
        playback_ptr->setBrightness(GIGE_DEFAULT_BRIGHTNESS);
        playback_ptr->setContrast(GIGE_DEFAULT_CONTRAST);

        if (ui->frm_gamma->isEnabled())
            playback_ptr->setGamma(GIGE_DEFAULT_GAMMA);

        playback_ptr->setRed(GIGE_DEFAULT_RGB_GAIN);
        playback_ptr->setGreen(GIGE_DEFAULT_RGB_GAIN);
        playback_ptr->setBlue(GIGE_DEFAULT_RGB_GAIN);

        // Tell the playback object the new white balance to use.
        playback_ptr->resetAutoWbGains();

        // Reset the display controls to match the default bcg and rgb just set up in the playback object.
        update_bcg();
        update_rgb();

        // Anytime we reset the controls here we don't "apply" the playback lut everytime focus is
        // regained.  We have stepped into a modification "zone" if you will where the display panel
        // has overwritten the default capture lut that may have been in use.  The user can always see
        // it if they create a custom curve but we no longer automatically set it when playback connect
        // operation occurs.
        playback_ptr->setApplyingCapLuts(false);

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

void PanelDisplay::show_spline_dialog(void)
{
    CameraMDI   *my_camera      = nullptr;
    QWidget     *mdi_widget;
    fiSensor    *ourSensor      = nullptr;  // If connected camera this is our sensor.
    bool        isMono          = true;     // Assume Mono as fallback.
    bool        isSeries3       = true;     // Assume Series3 as default.
    int         initialLutId    = 1;        // A default LUT id value.

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "starting_spline_dlg:" << starting_spline_dlg;

    // Are we supposed to be bringing up the dialog?
    if (!starting_spline_dlg)
        return;

    // Grab a copy of the current video frame.
    fiPixmap         snap_pixmap;

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Do we have a camera connected?
        if ((my_camera = p_control_panel->getCamera()) == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
            return;
        }

        // Make sure we have a video player.
        if (my_camera->get_player() == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "video_player object not allocated!";
            return;
        }

        // The parent widget is the camera MDI widget.
        mdi_widget = my_camera;

        // Grab a copy of the current image.
        my_camera->get_player()->copy_image(snap_pixmap);

        // Now that we've gotten our "base" image, go ahead and restore the LUT register.
        my_camera->write_camera_register(GIGE_REG_LUT, lut_reg);

        // Pause our video streaming.
        my_camera->pause_video_request();

        // What type of camera is this?
        ourSensor = my_camera->get_sensor();
        if (ourSensor != nullptr)
        {
            isMono = ourSensor->is_color() ? false : true;
        }

        // Is this a series 3 camera?
        isSeries3 = my_camera->is_series3();

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut_reg:"  << Qt::showbase << Qt::hex << lut_reg
                     << "ourSensor:" << ourSensor << "isMono:" << isMono;

        // If a lut selection is set in camera, pass it on to the dialog.
        if ((lut_reg & GIGE_LUT_MASK) != GIGE_LUT_NONE)
            initialLutId = (lut_reg & GIGE_LUT_MASK);
    }

    // Must just be playing a saved video.
    else if (playback_ptr != nullptr)
    {
        // Pull information from our metadata about this playback stream.
        PlaybackMetaInfo *pbMetaData = playback_ptr->getMetadata();

        // If there is metadata that can yield our state use it.
        if (pbMetaData != nullptr)
        {
            // Is this a monochrome or color sensor.
            isMono = pbMetaData->isColor() ? false : true;
            // Is this a series 3 camera?
            isSeries3 = pbMetaData->isSeries3();
        }

        // The parent widget is the playback MDI widget.
        mdi_widget = playback_ptr;

        // Get a copy of the last frame shown.
        playback_ptr->loadFrame(playback_ptr->getLastFrame(), &snap_pixmap, true);

        // Is there a custom lut associated with this playback object?
        // If so set up our initial LUT id to match it.
        if (playbackLUT)
        {
            initialLutId = playbackLUT->GetLutId();
        }
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera or playback?";
        return;
    }

    // Create our curve settings dialog.
    std::unique_ptr<CurveSettings> curveSettingsDlg =
        CurveSettings::CreateCurveDialog(isMono, isSeries3,
                                         snap_pixmap,
                                         initialLutId,
                                         mdi_widget);

    // Did we create a curve settings dialog?
    if (!curveSettingsDlg)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "CurveSettings::CreateCurveDialog(" << (isMono ? "Monochrome," : "Color,")
                     << (isSeries3 ? "Series3" : "SeriesNC") << "...) Failed!";
        return;
    }

    // We can now safely ignore any messages about the LUT register.
    starting_spline_dlg = false;

    // Restore our cursor.
    QApplication::restoreOverrideCursor();

    // Run the dialog.
    if (curveSettingsDlg->exec() == QDialog::Accepted)
    {
        // Get the lut label from out dialog.
        QString lut_label = QString::fromUtf8("%1: %2").arg(curveSettingsDlg->GetCheckedRadioLutId()).arg(curveSettingsDlg->GetRadioLutSplineRec().desc);

        // Install the new information about he custom curve into our panel.
        QByteArray *cameraLut = curveSettingsDlg->GetCameraLut();

        // Update the label information for the custom curve.
        ui->lbl_lut->setText(lut_label);

        // Send it on it's way.
        send_lut(curveSettingsDlg->GetCheckedRadioLutId(), cameraLut);

        // If we are a playback window go ahead and take the accepted display lut object and install
        // it in the playbackLUT object that we maintain so that the settings move along with the
        // playback window until it is re-opened.
        if (camera_ip_addr == 0)
        {
            // We are a playback scenario.
            try
            {
                // Get access to the accepted data.
                std::shared_ptr<FiDisplayLut> acceptedLutData = curveSettingsDlg->GetAcceptedFiDisplayLut();

                // Was there an accepted lut?
                if (acceptedLutData)
                {
                    QString acceptedLutLabel = QString::fromUtf8("%1: %2").arg(acceptedLutData->GetLutId()).arg(acceptedLutData->GetLutDesc());

                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Accept id:" << acceptedLutData->GetLutId()
                                 << "desc:" << acceptedLutData->GetLutDesc()
                                 << "lut_label:" << lut_label
                                 << "acceptedLutLabel:" << acceptedLutLabel;

                    // Ok at this point if we have a playbackLUT we want to install our
                    // accepted settings into it.
                    if (playbackLUT)
                    {
                        // Morph the contents of the playback LUT to take the
                        // form of the accepted data LUT - this will persist for
                        // the playback LUT until it is re-opened.
                        playbackLUT->CopyDisplayLut(acceptedLutData);
                    }
                }
            }
            catch (std::exception& e)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Exception:" << e.what();
            }
            catch (...)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "(...)Exception!";
            }
        }

        // All done.
        delete cameraLut;
    }
    // User has hit [Cancel] on the Curve Settings Dialog.
    else
    {
        int lutTableSize = LUT_MAX_TABLE_COUNT / 4; //!< Expected size of a LUT table vector.

        // The user cancelled. Are we connected to a camera?
        if (camera_ip_addr != 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cancel was hit.  lut_reg:"  << Qt::showbase << Qt::hex << lut_reg;

            // Do we want to restore the last used LUT, or the BCG settings?
            if ((lut_reg & GIGE_LUT_MASK) != GIGE_LUT_NONE)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut_reg != GIGE_LUT_NONE calling write_camera_register.";

                // I think the proper thing to do here is to update_lut not directly write to the register?
                update_lut(lut_reg);
                // Restore the previous LUT.
                //write_camera_register(GIGE_REG_LUT, lut_reg);
            }
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut_reg == GIGE_LUT_NONE restoring save_*";
                // Restore the BCG and RGB settings.
                write_list.append(REG_VALUE_PAIR(GIGE_REG_CONTRAST, save_contrast));
                write_list.append(REG_VALUE_PAIR(GIGE_REG_GAMMA, save_gamma));

                // Do not write the RGB settings if you have a series3 or Monochrome sensor.
                if (!isSeries3 && !isMono)
                {
                    write_list.append(REG_VALUE_PAIR(GIGE_REG_RED, save_red));
                    write_list.append(REG_VALUE_PAIR(GIGE_REG_GREEN, save_green));
                    write_list.append(REG_VALUE_PAIR(GIGE_REG_BLUE, save_blue));
                }
                // We are starting an update process.
                update_in_progress = true;

                // Start the write operation with the first operation.
                write_camera_register(GIGE_REG_BRIGHTNESS, save_brightness);
            }
        }
        else if (playback_ptr)
        {
            // We are a Playback MDI connection.

            // Restore Brightness.
            ui->spin_brightness->blockSignals(true);
            ui->spin_brightness->setValue((int)save_brightness);
            ui->spin_brightness->blockSignals(false);
            ui->sld_brightness->setValue((int)save_brightness);

            // Restore Contrast.
            ui->spin_contrast->blockSignals(true);
            ui->spin_contrast->setValue((int)save_contrast);
            ui->spin_contrast->blockSignals(false);
            ui->sld_contrast->setValue((int)save_contrast);

            // Restore Gamma.
            ui->spin_gamma->blockSignals(true);
            ui->spin_gamma->setValue(save_gamma / 100.0);
            ui->spin_gamma->blockSignals(false);
            ui->sld_gamma->setValue((int)save_gamma);

            // Restore the curve label information.
            ui->lbl_lut->setText(save_curve_label);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "saved_curve_label:" << save_curve_label;

            // We restore based on camera type and monochrome.
            if (isSeries3 || isMono)
            {
                // Is there a gamma lut?
                if (save_8bit_gamma_lut.count() == lutTableSize)
                {
                    playback_ptr->set8bitGammaLut(save_8bit_gamma_lut);
                }
                else
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "Expected save_gamma lut to be " << lutTableSize
                                << "but it was " << save_8bit_gamma_lut.count() << "size.";
                }
            }
            else
            {
                // This is a color sensor restoration catch RGB.

                // Restore Red Gain
                ui->spin_red->blockSignals(true);
                sld_red_changed((int)save_red);
                ui->spin_red->blockSignals(false);
                ui->sld_red->setValue((int)save_red);

                // Restore Green Gain
                ui->spin_green->blockSignals(true);
                sld_green_changed((int)save_green);
                ui->spin_green->blockSignals(false);
                ui->sld_green->setValue((int)save_green);

                // Restore Blue Gain
                ui->spin_blue->blockSignals(true);
                sld_blue_changed((int)save_blue);
                ui->spin_blue->blockSignals(false);
                ui->sld_blue->setValue((int)save_blue);

                // Restore the RGB LUTs.
                if (save_8bit_red_lut.count() == lutTableSize &&
                    save_8bit_green_lut.count() == lutTableSize &&
                    save_8bit_blue_lut.count() == lutTableSize)
                {
                    // Note:  The vector size for this call is a lut of 256 bytes in length
                    // per channel.
                    playback_ptr->set8bitRgbLuts(save_8bit_red_lut, save_8bit_green_lut, save_8bit_blue_lut);
                }
                else
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "Expected rgb luts to be " << lutTableSize
                                << "but save_red_lut: " << save_8bit_red_lut.count()
                                << "save_green_lut:" << save_8bit_green_lut.count()
                                << "save_blue_lut:" << save_8bit_blue_lut.count();
                }

                // Did we have a custom LUT in play?
                if (playbackLUT)
                {
                    QString lut_label = QString::fromUtf8("%1: %2").arg(playbackLUT->GetLutId()).arg(playbackLUT->GetLutDesc());
                    ui->lbl_lut->setText(lut_label);
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lut_label from playbackLUT:" << lut_label;
                }
            }
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Dialog was cancelled but there is no camera or playback pointer!";
        }
    } // End of the cancel block.

    // Update our display.
    if (camera_ip_addr != 0)
    {
        // Are we connected to a camera?
        if (my_camera == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "my_camera is null?";
        }
        else
        {
            // Resume our stream.
            my_camera->resume_video_request();

            // Update the current frame, if we're in review mode.
            my_camera->update_frame();

            // Restore panel updates.
            if (update_timer_id == 0)
            {
                update_timer_id = startTimer(POLLING_PERIOD);
            }
        }
    }
    else
    {
        // Reload the frame.
        if (playback_ptr)
        {
            playback_ptr->refreshFrame();
        }
    }

}

/*!
 * @brief The user has changed the blue gain values using the slider.
 * @param value     The new blue gain value.
 */
void PanelDisplay::sld_blue_changed(int value)
{
    double  d_value;

    // Convert from the slider value to something the user will see.
    value = value * RGB_GAIN_MULT;
    d_value = (double)value / RGB_GAIN_INCREMENT;
    ui->spin_blue->setValue(d_value);
}

/*!
 * @brief The user has changed the brightness value using the slider.
 * @param value     The new brightness value.
 */
void PanelDisplay::sld_brightness_changed(int value)
{
    ui->spin_brightness->setValue(value);
}

/*!
 * @brief The user has changed the contrast value using the slider.
 * @param value     The new contrast value.
 */
void PanelDisplay::sld_contrast_changed(int value)
{
    ui->spin_contrast->setValue(value);
}

/*!
 * @brief The user has changed the gamma value using the slider.
 * @param value     The new gamma value.
 */
void PanelDisplay::sld_gamma_changed(int value)
{
    double  d_val = ((double)value) / 100.0;

    // Only update if there's a change.
    if (!qFuzzyCompare(ui->spin_gamma->value(), d_val))
    {
        ui->spin_gamma->setValue(d_val);
    }
}

/*!
 * @brief The user has changed the green gain value using the slider.
 * @param value     The new green gain value.
 */
void PanelDisplay::sld_green_changed(int value)
{
    double  d_value;

    // Convert from the slider value to something the user will see.
    value = value * RGB_GAIN_MULT;
    d_value = (double)value / RGB_GAIN_INCREMENT;
    ui->spin_green->setValue(d_value);
}

/*!
 * @brief The user has changed the red gain value using the slider.
 * @param value     The new red gain value.
 */
void PanelDisplay::sld_red_changed(int value)
{
    double  d_value;

    // Convert from the slider value to something the user will see.
    value = value * RGB_GAIN_MULT;
    d_value = (double)value / RGB_GAIN_INCREMENT;
    ui->spin_red->setValue(d_value);
}

/*!
 * @brief The user has changed the blue gain values using the spinner. This
 *        method also is invoked when the slider method updates the spinner
 *        control.
 *
 * @param value     The new blue gain value.
 */
void PanelDisplay::spin_blue_changed(double d_value)
{
    uint32_t        value;
    uint32_t        remain;

    value = (uint32_t)(d_value * RGB_GAIN_INCREMENT);

    // Make sure it's an increment of the RGB multiplier, this can be off somewhat because
    // of using floating point numbers, if not round up
    remain = value % RGB_GAIN_MULT;
    if (remain)
        value += (RGB_GAIN_MULT - remain);

    // Make sure our slider value is consistent with the spinner.
    ui->sld_blue->setValue(value / RGB_GAIN_MULT);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // Send the new value to the camera.
        write_camera_register(GIGE_REG_BLUE, value);
    }
    else if (playback_ptr != nullptr)
    {
        // Pass the new value on.
        playback_ptr->setBlue(value);

        // No longer using a custom lut.
        reset_lut_empty();

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

/*!
 * @brief The user has changed the brightness value using the spinner. This
 *        method also is invoked when the slider method updates the spinner
 *        control.
 *
 * @param value     The new brightness value.
 */
void PanelDisplay::spin_brightness_changed(int value)
{
    ui->sld_brightness->setValue(value);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // Send the new value to the camera.
        write_camera_register(GIGE_REG_BRIGHTNESS, (uint32_t)(value + BC_OFFSET));
    }
    else if (playback_ptr != nullptr)
    {
        // Pass the new value on.
        playback_ptr->setBrightness((uint32_t)(value + BC_OFFSET));

        // Empty our LUT setting.
        reset_lut_empty();

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

/*!
 * @brief The user has changed the contrast value using the spinner. This
 *        method also is invoked when the slider method updates the spinner
 *        control.
 *
 * @param value     The new contrast value.
 */
void PanelDisplay::spin_contrast_changed(int value)
{
    ui->sld_contrast->setValue(value);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // Send the new value to the camera.
        write_camera_register(GIGE_REG_CONTRAST, (uint32_t)(value + BC_OFFSET));
    }
    else if (playback_ptr != nullptr)
    {
        // Pass the new value on.
        playback_ptr->setContrast((uint32_t)(value + BC_OFFSET));

        // Empty our LUT setting.
        reset_lut_empty();

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

/*!
 * @brief The user has changed the gamma value using the spinner. This
 *        method also is invoked when the slider method updates the spinner
 *        control.
 *
 * @param value     The new gamma value.
 */
void PanelDisplay::spin_gamma_changed(double d_val)
{
    int     value = (int)(d_val * 100.0);

    // Update the slider's value.
    ui->sld_gamma->setValue(value);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // Send the new value to the camera.
        write_camera_register(GIGE_REG_GAMMA, (uint32_t)value);
    }
    else if (playback_ptr != nullptr)
    {
        // Pass the new value on.
        playback_ptr->setGamma((uint32_t)value);

        // No longer using a custom lut.
        reset_lut_empty();

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

/*!
 * @brief The user has changed the green gain value using the spinner. This
 *        method also is invoked when the slider method updates the spinner
 *        control.
 *
 * @param value     The new green gain value.
 */
void PanelDisplay::spin_green_changed(double d_value)
{
    uint32_t        value;
    uint32_t        remain;

    value = (uint32_t)(d_value * RGB_GAIN_INCREMENT);

    // Make sure it's an increment of the RGB multiplier, this can be off somewhat because
    // of using floating point numbers, if not round up
    remain = value % RGB_GAIN_MULT;
    if (remain)
        value += (RGB_GAIN_MULT - remain);

    ui->sld_green->setValue(value / RGB_GAIN_MULT);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // Send the new value to the camera.
        write_camera_register(GIGE_REG_GREEN, value);
    }
    else if (playback_ptr != nullptr)
    {
        // Pass the new value on.
        playback_ptr->setGreen(value);

        // No longer using a custom lut.
        reset_lut_empty();

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

/*!
 * @brief The user has changed the red gain value using the spinner. This
 *        method also is invoked when the slider method updates the spinner
 *        control.
 *
 * @param value     The new red gain value.
 */
void PanelDisplay::spin_red_changed(double d_value)
{
    uint32_t        value;
    uint32_t        remain;

    value = (uint32_t)(d_value * RGB_GAIN_INCREMENT);

    // Make sure it's an increment of the RGB multiplier, this can be off somewhat because
    // of using floating point numbers, if not round up
    remain = value % RGB_GAIN_MULT;
    if (remain)
        value += (RGB_GAIN_MULT - remain);

    ui->sld_red->setValue(value / RGB_GAIN_MULT);

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // Send the new value to the camera.
        write_camera_register(GIGE_REG_RED, value);
    }
    else if (playback_ptr != nullptr)
    {
        // Pass the new value on.
        playback_ptr->setRed(value);

        // No longer using a custom lut.
        reset_lut_empty();

        // Force a refresh.
        playback_ptr->refreshFrame();
    }
}

void PanelDisplay::switch_to_spline(void)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
    // Are we in the middle of starting a spline curve settings dialog already?
    if (starting_spline_dlg)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected starting_spline_dlg was true!";
        return;
    }

    // Are we connected to a camera?
    if (camera_ip_addr != 0)
    {
        CameraMDI   *my_camera;

        // Do we have a camera connected?
        if ((my_camera = p_control_panel->getCamera()) == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
            return;
        }

        // This can be rather lengthy, so show a busy cursor.
        QApplication::setOverrideCursor(Qt::BusyCursor);

        // Remember that we're starting the curve settings dialog.
        starting_spline_dlg = true;

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Setting lut_reg to GIGE_LUT_NONE and calling start_camera_read_reg(GIGE_REG_LUT) current lut_reg:"  << Qt::showbase << Qt::hex << lut_reg;
        lut_reg = GIGE_LUT_NONE;

        // Read the LUT register from the camera. We'll continue processing
        // after we get the value back.
        my_camera->start_camera_read_reg(GIGE_REG_LUT);
    }
    else
    {
        // Sanity check.
        if (playback_ptr == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera AND no playback?!";
            return;
        }

        // Since we're not connected to a camera, get straight to it.
        starting_spline_dlg = true;

        // Save the name of the last custom lut if any.
        save_curve_label = ui->lbl_lut->text();

        // Save the gamma lut.
        save_8bit_gamma_lut = playback_ptr->getGammaLut();

        // Save the current BCG settings.
        save_contrast = (uint32_t)ui->sld_contrast->value();
        save_brightness = (uint32_t)ui->sld_brightness->value();
        save_gamma = (uint32_t)ui->sld_gamma->value();

        // Reset the BCG to defaults.
        playback_ptr->setBrightness(GIGE_DEFAULT_BRIGHTNESS);
        playback_ptr->setContrast(GIGE_DEFAULT_CONTRAST);
        playback_ptr->setGamma(GIGE_DEFAULT_GAMMA);

        // Update the display UI elements to match the default bcg settings just set up in the playback object.
        update_bcg();

        // Save off RGB lut and values if the frame is visible.
        if (ui->frm_rgb->isVisible())
        {
            // These are returning 8-bit LUT vectors.
            save_8bit_red_lut = playback_ptr->getRedLut();
            save_8bit_green_lut = playback_ptr->getGreenLut();
            save_8bit_blue_lut = playback_ptr->getBlueLut();

            save_red = (uint32_t)ui->sld_red->value();
            save_green = (uint32_t)ui->sld_green->value();
            save_blue = (uint32_t)ui->sld_blue->value();

            // Reset the playback rgb gains to default settings.
            playback_ptr->setRed(GIGE_DEFAULT_RGB_GAIN);
            playback_ptr->setGreen(GIGE_DEFAULT_RGB_GAIN);
            playback_ptr->setBlue(GIGE_DEFAULT_RGB_GAIN);

            // Reset the RGB ui elements on the display panel match defaults that were just set.
            update_rgb();
        }

        // Force a refresh.
        playback_ptr->refreshFrame();

        // Now show the spline dialog.
        show_spline_dialog();
    }
}

/*!
 * @brief Toggle the visiblity of the panel. Optionally, pass the event up
 *        to the parent.
 *
 * @param notify_parent  True if we're supposed to pass this up to the parent.
 */
void PanelDisplay::toggle_visibility(bool notify_parent)
{
    // Pass it on up.
    FMPanelWidget::toggle_visibility(notify_parent);

    // If we're opened up, make sure we refresh our controls
    if (isVisible())
    {
        // Do we have a camera connected?
        if (p_control_panel->getCamera() != nullptr)
        {
            // Update our panel.
            update_panel();

            // Create our polling timer.
            if (update_timer_id == 0)
            {
                update_timer_id = startTimer(POLLING_PERIOD);
            }
        }
    }
    else
    {
        // Stop polling the camera when we're closed.
        if (update_timer_id != 0)
        {
            killTimer(update_timer_id);
            update_timer_id = 0;
        }
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void PanelDisplay::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Only care if we're visible.
    if (!isVisible())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Only care about our registers.
    switch (reg_addr)
    {
    case GIGE_REG_BLUE:
    case GIGE_REG_BRIGHTNESS:
    case GIGE_REG_CONTRAST:
    case GIGE_REG_GAMMA:
    case GIGE_REG_GREEN:
    case GIGE_REG_GENERATE_CUSTOM_WB:
    case GIGE_REG_HDMI:
    case GIGE_REG_RED:
    case GIGE_REG_WHITE_BAL:
    case GIGE_REG_WHITE_BALANCE_BLU:
    case GIGE_REG_WHITE_BALANCE_GRN:
    case GIGE_REG_WHITE_BALANCE_RED:
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(camera_ip_addr, reg_addr);
        if (write_status != GEV_STATUS_SUCCESS)
        {
            QHostAddress    our_addr;
            our_addr.setAddress(qFromBigEndian(camera_ip_addr));

            QString err_msg = QString::fromUtf8("Unable to write register 0x%1 to IP adddress %2").arg(reg_addr, 0, 16).arg(our_addr.toString());
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }

        // Are there still write operations to perform?
        else if (!write_list.isEmpty())
        {
            REG_VALUE_PAIR  reg_pair;

            // Grab the next operation off the list.
            reg_pair = write_list.takeFirst();

            // Issue the write command.
            write_camera_register(reg_pair.first, reg_pair.second);
        }

        // Are we done with the update?
        else if (update_in_progress)
        {
            // Clear our flag.
            update_in_progress = false;

            // Request a new frame from the camera.
            refresh_frame();

            // Are we starting the spline dialog?
            if (starting_spline_dlg)
            {
                // Let's stop display panel updates until our curve settings dialog completes.
                if (update_timer_id != 0)
                {
                    killTimer(update_timer_id);
                    update_timer_id = 0;
                }

                // And spawn a timer to bring up the CurveSettings dialog.
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Using singleShot timer SPLINE_DIALOG_DELAY to invoke PanelDisplay::show_spline_dialog.";
                QTimer::singleShot(SPLINE_DIALOG_DELAY, this, &PanelDisplay::show_spline_dialog);
            }
            else
            {
                // Update our registers, which will update our controls.
                read_registers();
            }
        }
        break;
    default:
        break;
    }
}
