/*!
 * @file panel_display.h
 * @brief This is the header file for the Display configuration panel.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PANEL_DISPLAY_H
#define PANEL_DISPLAY_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QWidget>
#include "panel_widget.h"
#include "fi_display_lut.h"

namespace Ui
{
class PanelDisplay;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QTimerEvent;
class PlaybackMDI;
class fiSplineWidget;
class CurveSettings;

namespace aaAaa
{
struct aaSpline;
}

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class PanelDisplay : public FMPanelWidget
{
    Q_OBJECT

public:
    explicit PanelDisplay(QWidget *parent = nullptr);
    ~PanelDisplay();

    // Handle camera connect/disconnects.
    virtual void camera_connect(uint32_t ip_addr);
    virtual void camera_disconnect(void);

    // Handle playback object connect/disconnects.
    void playback_connect(QWidget *playback_ptr);
    void playback_disconnect(void);

    // Update the panel contents.
    virtual void update_panel(void);

    void    update_sensor_type(uint32_t sensor_type);
    void    update_whitebal(uint32_t new_setting);

protected:
    virtual void    changeEvent(QEvent *event);

protected slots:
    virtual void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    virtual void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    virtual void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    virtual void    timerEvent(QTimerEvent *event);
    virtual void    toggle_visibility(bool notify_parent = true);
    virtual void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private slots:
    void    cb_hdmi_changed(int index);
    void    cb_white_bal_changed(int index);
    void    chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr);

    void    do_autowb();
    void    do_colorize();
    void    do_colorize_color();
    void    do_focus_assist();
    void    do_focus_assist_color();

    void    sld_blue_changed(int value);
    void    sld_brightness_changed(int value);
    void    sld_contrast_changed(int value);
    void    sld_gamma_changed(int value);
    void    sld_green_changed(int value);
    void    sld_red_changed(int value);

    void    spin_blue_changed(double value);
    void    spin_brightness_changed(int value);
    void    spin_contrast_changed(int value);
    void    spin_gamma_changed(double dval);
    void    spin_green_changed(double value);
    void    spin_red_changed(double value);

    void    switch_to_spline(void);
    void    show_spline_dialog(void);

    void    refresh_frame(void);
    void    reset_lut_empty(void);
    void    reset_to_defaults(void);
    void    custom_whitebal(void);

private:
    void    handle_read(uint32_t reg_addr, uint32_t reg_value);
    void    init_controls(void);

    void    read_registers(void);
    void    send_lut(int lut_id, QByteArray *lut_table);
    void    translate_ui(void);

    void    updateDisplayBrightness(uint32_t camRegValue);
    void    updateDisplayContrast(uint32_t camRegValue);
    void    updateDisplayGamma(uint32_t camRegValue);

    void    update_bcg(void);
    void    update_colorize(bool do_colorize);
    void    update_focus_assist(bool do_assist);
    void    update_lut(uint32_t reg_value);
    void    update_lut_display(QByteArray *lut_data);
    void    update_lut_display(int lutId, const QString& lutLabel);
    void    update_hdmi(uint32_t new_setting);
    void    updateDisplayRed(uint32_t camRegValue);
    void    updateDisplayGreen(uint32_t camRegValue);
    void    updateDisplayBlue(uint32_t camRegValue);
    void    update_rgb(void);

private:
    Ui::PanelDisplay *ui;                       //!< Our UI object.
    PlaybackMDI *playback_ptr;                  //!< Set to the playback object if connected to one, NULL otherwise.
    std::shared_ptr<FiDisplayLut> playbackLUT;  //!< Custom LUT data associated with playback_ptr.

    uint32_t    camera_ip_addr;                 //!< IP address of the camera we're talking to.
    bool        custom_whitebal_in_progress;    //!< Flag to indicate that a custom white-balance operation is in progress.
    int         scrub_update_id;                //!< Timer ID used to "filter", or reduce, the number of updates when the user drags a slider.
    bool        update_in_progress;             //!< Flag to indicate that an update (write operation) is in progress.
    int         update_timer_id;                //!< Timer ID to trigger an update.

    uint32_t    white_bal_blue;                 //!< Blue component of a white balance operation.
    uint32_t    white_bal_green;                //!< Green component of a white balance operation.
    uint32_t    white_bal_red;                  //!< Red component of a white balance operation.

    uint32_t    last_lut_id;                    //!< The last active LUT ID.
    uint32_t    lut_reg;                        //!< Mirror copy of the LUT register so we can restore it when we're done.
    uint32_t    save_brightness;
    uint32_t    save_contrast;
    uint32_t    save_gamma;
    QVector<double> save_8bit_gamma_lut;

    uint32_t    save_red;
    uint32_t    save_green;
    uint32_t    save_blue;

    QVector<double> save_8bit_red_lut;
    QVector<double> save_8bit_green_lut;
    QVector<double> save_8bit_blue_lut;

    QString     save_curve_label;

    bool        starting_spline_dlg;            //!< True if we're in the process of starting the CurveSettings dialog.
};

#endif // PANEL_DISPLAY_H
