/*!
 * @file panel_lens.cpp
 * @brief This file contains the implementation of the PanelLens
 *        class. This class handles controlling the lens.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QAction>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonValue>
#include <QLineEdit>
#include <QTimer>

#include <QString>
#include <QStringLiteral>
#include <string>

#include "panel_lens.h"
#include "ui_panel_lens.h"

#include "camera_mdi.h"
#include "control_panel.h"
#include "fi_comm.h"
#include "fi_switchbutton.h"
#include "logger.h"
#include "mainwindow.h"

namespace PANEL_LENS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Maximum number of times we'll try to get our message across.
const int MAX_RETRIES = 3;

// Milliseconds before we ask for another read.
const int READ_RETRY_TIMEOUT = 100;

// Milliseconds before we ask for a lens status report.
const int STATUS_TIMEOUT = 500;

// Taken from "https://en.wikipedia.org/wiki/F-number".
const float QUARTER_STOP_FNUM_SCALE[]  =
{
    1.0, 1.1, 1.2, 1.3, 1.4, 1.5, 1.7, 1.8, 2, 2.2, 2.4, 2.6, 2.8, 3.1, 3.3, 3.7, 4, 4.4, 4.8, 5.2, 5.6,
    6.2, 6.7, 7.3, 8, 8.7, 9.5, 10, 11, 12, 14, 15, 16, 17, 19, 21, 22, 25, 27, 29, 32
};

} // end of namespace PANEL_LENS.

const QString PanelLens::LENS_CONTROL_ACQUIRING    = QStringLiteral("Aquiring lens information...");
const QString PanelLens::LENS_CONTROL_DISABLED     = QStringLiteral("Lens control disabled");
const QString PanelLens::LENS_NOT_FOUND            = QStringLiteral("No lens detected");
const QString PanelLens::LENS_UNKNOWN              = QStringLiteral("Unknown lens detected");
const QString PanelLens::LENS_WAITING_FOR_POWER    = QStringLiteral("Waiting for power");

using namespace PANEL_LENS;

PanelLens::PanelLens(QWidget *parent) :
    FMPanelWidget(parent),
    ui(new Ui::PanelLens)
{
    // Setup our UI controls.
    ui->setupUi(this);

    aperture_last_index = -1;
    focus_last_index = -1;
    focus_step_size = 1;
    zoom_last_index = -1;
    camera_ip_addr = 0,
    update_allowed = true;
    request_pending = false;
    retry_count = 0;
    status_timer_id = -1;

    read_connected_lens_data = false;

    connected = false;
    powered = false;
    talking = false;

    ez_enabled = false;
    focal_len_wide = 0;
    focal_len_tele = 0;
    zoom_divisions = 0;
    focal_length = 0;

    lens_name.clear();

    // Setup our switch button.
    sw_power = new fiSwitchButton(this, fiSwitchButton::Style::ONOFF);
    sw_power->setFixedSize(ui->pb_power->size());
    ui->headerFrame->layout()->replaceWidget(ui->pb_power, sw_power);
    ui->pb_power->setVisible(false);

    // We want all of our sliders to throttle their feedback when the user is dragging the "thumb".
    ui->aperture_slide->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    ui->focus_slide->setThrottle(FI_SLIDER::SCRUB_UPDATE);
    ui->zoom_slide->setThrottle(FI_SLIDER::SCRUB_UPDATE);

    // Connect our signals and slots.
    connect(ui->aperture_slide, &fiSlider::released, this, &PanelLens::apertureSliderReleased);
    connect(ui->aperture_slide, &fiSlider::valueChanged, this, &PanelLens::apertureSlideSlowChange);
    connect(ui->aperture_slide, &fiSlider::valueSlowChanged, this, &PanelLens::apertureSlideSlowChange);
    connect(ui->aperture_spin, qOverload<double>(&QDoubleSpinBox::valueChanged), this, &PanelLens::apertureSpinValueChanged, Qt::QueuedConnection);

    connect(ui->focus_slide, &fiSlider::released, this, &PanelLens::focusSliderReleased);
    connect(ui->focus_slide, &fiSlider::valueChanged, this, &PanelLens::focusSlideSlowChange);
    connect(ui->focus_slide, &fiSlider::valueSlowChanged, this, &PanelLens::focusSlideSlowChange);
    connect(ui->focus_step_spin, &QSpinBox::editingFinished, this, &PanelLens::focusStepSizeChanged);

    connect(ui->focus_tele_tb, &QToolButton::pressed, this, &PanelLens::focusStepTele);
    connect(ui->focus_wide_tb, &QToolButton::pressed, this, &PanelLens::focusStepWide);
    connect(sw_power, &fiSwitchButton::valueChanged, this, &PanelLens::switchPowerChanged);
    connect(ui->zoom_slide, &fiSlider::released, this, &PanelLens::zoomSliderReleased);
    connect(ui->zoom_slide, &fiSlider::valueChanged, this, &PanelLens::zoomSlideSlowChange);
    connect(ui->zoom_slide, &fiSlider::valueSlowChanged, this, &PanelLens::zoomSlideSlowChange);

    // By default, disable autofocus.
    ui->pb_autofocus->setVisible(false);
    ui->pb_autofocus->setEnabled(false);
    connect(ui->pb_autofocus, &QPushButton::pressed, this, &PanelLens::autoFocus);

    // Make sure we got a parent pointer.
    if (parent != nullptr)
    {
        // My grandparent is the mainwindow.
        MainWindow  *p_mainwnd = qobject_cast<MainWindow *>(parent->parent());

        // Connect the menu item's I'm interested in.
        connect(p_mainwnd->get_cp_lens(), &QAction::triggered, this, &PanelLens::toggle_visibility);
    }

    // After everything is setup, setup the intial size information.
    set_initial_sizing();

    // Start out with things hidden until we get connected to a camera.
    ui->controls_group->setVisible(false);
    sw_power->setVisible(false);
    ui->headerBorder->setVisible(false);
    adjustSize();
    update();

    // Disable the panel.
    setEnabled(false);
}

PanelLens::~PanelLens()
{
    delete ui;
}

/*!
 * @brief We are connecting to a (possibly) new camera.
 * @param ip_addr   The IP address of the camera we're connecting to.
 */
void PanelLens::camera_connect(uint32_t ip_addr)
{
    FI_comm *comm_ptr;
    CameraMDI *camera_ptr;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Make sure we're not already connected to this camera.
    if (ip_addr == camera_ip_addr)
        return;

    // Save the IP address.
    camera_ip_addr = ip_addr;

    // Get our communications pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get a FI_comm object!";
        return;
    }

    // Connect the communications interfaces that we're interested in.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelLens::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &PanelLens::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelLens::handle_unsol_msg, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelLens::write_complete, Qt::QueuedConnection);

    // Pass it on.
    FMPanelWidget::camera_connect(ip_addr);

    // Make certain our autofocus button is hidden away until we know it
    // should be made available.
    ui->pb_autofocus->setVisible(false);
    ui->pb_autofocus->setEnabled(false);

    // Get our camera.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr != nullptr)
    {
        // If we're series3, then we can't do lens control.
        if (camera_ptr->is_series3())
        {
            // Hide the controls.
            ui->controls_group->setVisible(false);
            sw_power->setVisible(false);
            ui->headerBorder->setVisible(false);
            adjustSize();
            update();

            // Make sure the name show's no lens information.
            ui->lens_name->setText(LENS_CONTROL_DISABLED);

            // Disable the panel.
            setEnabled(false);
        }
        else
        {
            // Initialize our variables.
            update_allowed = true;
            connected = false;
            retry_count = 0;
            request_pending = false;

            // Make sure we aren't marked as having read the connected camera's lens data.
            if (read_connected_lens_data)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Panel incorrectly has valid_lens_data while connecting to " << camera_ptr->get_camera_name();
                read_connected_lens_data = false;
            }

            // Make sure the name show's no lens information.
            ui->lens_name->setText(LENS_CONTROL_ACQUIRING);

            // Wait a bit, and then request the lens information.
//            requestInfo();
        }
    }
}

/*!
 * @brief Disconnect from our camera.
 */
void PanelLens::camera_disconnect()
{
    FI_comm     *comm_ptr;

    read_connected_lens_data = false;    // Disconnected to any lens reset read_connected_lens_data state.

    // Have we been here already?
    if (camera_ip_addr == 0)
        return;

    // Make sure we cancel the timer, if one is going.
    if (status_timer_id != -1)
    {
        killTimer(status_timer_id);
        status_timer_id = -1;
    }

    // Clear our IP address.
    camera_ip_addr = 0;

    // Pass it up.
    FMPanelWidget::camera_disconnect();

    // Get our communications pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get a FI_comm object!";
        return;
    }

    // Connect the communications interfaces that we're interested in.
    disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelLens::read_complete);
    disconnect(comm_ptr, &FI_comm::msg_timeout, this, &PanelLens::msg_timeout_event);
    disconnect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelLens::handle_unsol_msg);
    disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelLens::write_complete);

    // Make sure we restore the cursor, in case we overrode it.
//    QApplication::restoreOverrideCursor();

    // Hide the controls.
    ui->controls_group->setVisible(false);
    sw_power->setVisible(false);
    ui->headerBorder->setVisible(false);
    adjustSize();
    update();

    // Make sure the name show's no lens information.
    ui->lens_name->setText(LENS_CONTROL_DISABLED);

    // Disable the panel.
    setEnabled(false);

    // Not connected to a camera.
    camera_ip_addr = 0;
}

void PanelLens::update_panel()
{
    CameraMDI *camera_ptr;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Pass it up.
    FMPanelWidget::update_panel();

    // Make sure we are connected to a camera.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
        return;

    // Is there a request already pending?
    if (request_pending)
        return;

    // If we're series3, then we can't do lens control.
    if (camera_ptr->is_series3())
    {
        // Hide the controls.
        ui->controls_group->setVisible(false);
        sw_power->setVisible(false);
        ui->headerBorder->setVisible(false);
        adjustSize();
        update();

        // Make sure the name show's no lens information.
        ui->lens_name->setText(LENS_CONTROL_DISABLED);

        // Disable the panel.
        setEnabled(false);
    }
    else
    {
        // Only get the status if we've already connected and received the lens information.
        if (isEnabled())
        {
            // Ask the camera for the current lens status.
            requestStatus();
        }
        else
        {
            // Initialize our variables.
            update_allowed = true;
            connected = false;
            retry_count = 0;

            // Make sure the name show's no lens information.
            ui->lens_name->setText(LENS_CONTROL_ACQUIRING);

            // Wait a bit, and then request the lens information.
            requestInfo();
        }
    }
}

double PanelLens::getAperture()
{
    return ui->aperture_spin->value();
}

int PanelLens::getFocalLength()
{
    return focal_length;
}

void PanelLens::getLensName(std::string& lens_name)
{
    lens_name.assign(this->lens_name.toStdString());
}

/*!
 * @brief Initialize the signal/slot connections and initializations for the controls.
 */
void PanelLens::init_controls()
{
}

/*!
 * @brief Translate the UI elements for this object.
 */
void PanelLens::translate_ui()
{
    int i;
    QList<QWidget *> widgetList = findChildren<QWidget *>();

    // Block all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(true);

    // Translate our user interface.
    ui->retranslateUi(this);

    // Unblock all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(false);
}

int PanelLens::findFnumIndex(double f_num)
{
    int i;

    // Find the first f-number greater-than or equal to f_num.
    for (i = 0; i < (int)sizeof(QUARTER_STOP_FNUM_SCALE); ++i)
    {
        if (qFuzzyCompare((double)QUARTER_STOP_FNUM_SCALE[i], f_num) || (QUARTER_STOP_FNUM_SCALE[i] > f_num))
            break;
    }

    // Return what we found.
    return i;
}

void PanelLens::parseLensInfo(const QJsonObject& lensObj)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lensObj:" << lensObj;

    // If we're not connected, we need to hide all the controls.
    connected = lensObj["connected"].toBool();
    if (!connected)
    {
        // Make sure the interface shows that we have no lens, or at least none
        // we can communicate with.
        showNoLens();
        return;
    }

    // We're connected, so make things visible.
    sw_power->setVisible(true);
    ui->headerBorder->setVisible(true);

    // See if we are powered or not.
    powered = false;
    if (!lensObj["powered"].isNull())
    {
        powered = lensObj["powered"].toBool();
    }
    sw_power->blockSignals(true);
    sw_power->setValue(powered);
    sw_power->blockSignals(false);
    if (!powered)
    {
        // We're connected to a lens, but there's no power.
        lens_name.clear();
        ui->lens_name->setText(LENS_WAITING_FOR_POWER);
        return;
    }

    talking = false;
    if (!lensObj["talking"].isNull())
    {
        talking = lensObj["talking"].toBool();
    }
    if (!talking)
    {
        // We're connected and powered, but not talking. This is likely
        // not an MFT lens then.
        lens_name.clear();
        ui->lens_name->setText(LENS_UNKNOWN);
        return;
    }

    // Set the name.
    lens_name.clear();
    if (!lensObj["name"].isNull())
    {
        QString temp_name;

        // Get the name from the object.
        temp_name = lensObj["name"].toString().simplified();

        // Only save the printable characters.
        lens_name.clear();
        foreach (QChar temp, temp_name)
        {
            if (temp.isPrint())
                lens_name += temp;
        }

        // Make sure to strip any leading/trailing spaces.
        lens_name = lens_name.simplified();

        // Display the name.
        ui->lens_name->setText(lens_name);
    }

    // We are talking to the lens, so we can show the rest of the controls.
    ui->controls_group->setVisible(true);

    // Save the rest of the information.
    if (!lensObj["focal_len_tele"].isNull())
    {
        focal_len_tele = lensObj["focal_len_tele"].toInt();
        ui->zoom_max_lbl->setText(QString::number(focal_len_tele) + QStringLiteral(" mm"));
        ui->zoom_spin->setMaximum(focal_len_tele);
    }
    if (!lensObj["focal_len_wide"].isNull())
    {
        focal_len_wide = lensObj["focal_len_wide"].toInt();
        ui->zoom_min_lbl->setText(QString::number(focal_len_wide) + QStringLiteral(" mm"));
        ui->zoom_spin->setMinimum(focal_len_wide);
    }
    if (!lensObj["zoom_divisions"].isNull())
    {
        zoom_divisions = lensObj["zoom_divisions"].toInt();
        if (zoom_divisions > 1)
        {
            // Make sure the zoom controls are visible.
            ui->zoom_lbl->setVisible(true);
            ui->zoom_max_lbl->setVisible(true);
            ui->zoom_min_lbl->setVisible(true);
            ui->zoom_slide->setVisible(true);
            ui->zoom_spin->setVisible(true);

            // Update our maximum values.
            ui->zoom_slide->setMaximum(zoom_divisions - 1);
        }
        else
        {
            // There is no zoom. Hide the controls.
            ui->zoom_lbl->setVisible(false);
            ui->zoom_max_lbl->setVisible(false);
            ui->zoom_min_lbl->setVisible(false);
            ui->zoom_slide->setVisible(false);
            ui->zoom_spin->setVisible(false);
        }
    }

    // Only enable the zoom slider if the lens has an electric zoom.
    if (!lensObj["ez_enable"].isNull())
        ez_enabled = lensObj["ez_enable"].toBool();
    else
        ez_enabled = false;
    if (ez_enabled)
        ui->zoom_lbl->setText(tr("Zoom [EZ]:"));
    else
        ui->zoom_lbl->setText(tr("Zoom [MZ]:"));
    ui->zoom_slide->setEnabled(ez_enabled);

    // Now ask for the current settings.
    requestStatus();
}

/*!
 * @brief Read response to GIGE_REG_LENS_STATUS_JSON to handle translation
 * of lens information to user interface elements in the panel for the lens.
 *
 * @param lensObj [I] the JSON information for the lens connected to the camera.
 */
void PanelLens::parseLensStatus(const QJsonObject& lensObj)
{
    double f_min;
    double f_max;
    double f_number;
    int index;
    int min_index;
    int max_index;
    int focus_min;
    int16_t focus_near;
    int16_t focus_inf;
    int16_t focus_pos;
    int zoom_pos;
    bool manual_focus;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "lensObj:" << lensObj;

    // Reset our controls.
    powered = false;
    talking = false;

    // Make sure we cancel the timer, if one is going.
    if (status_timer_id != -1)
    {
        killTimer(status_timer_id);
        status_timer_id = -1;
    }

    // If we're not connected, we need to hide all the controls.
    connected = lensObj["connected"].toBool();
    if (!connected)
    {
        // Hide the controls.
        ui->controls_group->setVisible(false);
        sw_power->setVisible(false);
        ui->headerBorder->setVisible(false);
        adjustSize();
        update();

        // Make sure the name show's no lens information.
        ui->lens_name->setText(LENS_NOT_FOUND);

        // Disable the panel.
        setEnabled(false);
        return;
    }

    // Check to see if we toggled from not connected to connected. Also, check
    // to see if we've already read the lens information.
    if (!isEnabled() && lens_name.isEmpty())
    {
        CameraMDI *camera_ptr;
        camera_ptr = p_control_panel->getCamera();

        // We are disabled, so that means we disconnected from the lens. Maybe
        // the user reset the lens or maybe the user changed lenses. Either way,
        // we need to get the lens information.
        if (camera_ptr != nullptr)
            camera_ptr->start_camera_read_buf(GIGE_REG_LENS_INFO_JSON, GIGE_LENS_INFO_JSON_LEN);

        // We're done for now.
        return;
    }

    // See if we are powered or not.
    if (!lensObj["powered"].isNull())
    {
        powered = lensObj["powered"].toBool();
    }
    sw_power->blockSignals(true);
    sw_power->setValue(powered);
    sw_power->blockSignals(false);

    // If we're not powered, then we are done.
    if (!powered)
    {
        // We're connected to a lens, but there's no power.
        lens_name.clear();
        ui->lens_name->setText(LENS_WAITING_FOR_POWER);

        // Hide the rest of the controls.
        ui->controls_group->setVisible(false);
        return;
    }

    if (!lensObj["talking"].isNull())
    {
        talking = lensObj["talking"].toBool();
    }
    if (!talking)
    {
        // We're connected and powered, but not talking. This is likely
        // not an MFT lens then.
        lens_name.clear();
        ui->lens_name->setText(LENS_UNKNOWN);

        // Hide the rest of the controls.
        ui->controls_group->setVisible(false);
        return;
    }

    // Do we have the name of the lens?
    if (lens_name.isEmpty())
    {
        // Ask the camera for more information.
        requestInfo();
    }

    // Show the rest of the controls.
    ui->controls_group->setVisible(true);

    // Get the values in a form we can use.
    f_min = lensObj["f_minimum"].toDouble();
    f_max = lensObj["f_maximum"].toDouble();
    f_number = lensObj["f_number"].toDouble();
    focus_pos = lensObj["focus_pos"].toInt();
    focus_min = lensObj["focus_min"].toInt();
    focus_near = lensObj["focus_near"].toInt();
    focus_inf = lensObj["focus_inf"].toInt();
    zoom_pos = lensObj["zoom_pos"].toInt();
    focal_length = lensObj["focal_len"].toInt();
    manual_focus = lensObj["manual_focus"].toBool();

    // Initialize the aperture controls.
    ui->aperture_min_lbl->setText(QLatin1Char('F') + QString::number(f_min, 'g', 2));
    ui->aperture_max_lbl->setText(QLatin1Char('F') + QString::number(f_max, 'g', 3));

    ui->aperture_spin->blockSignals(true);
    ui->aperture_spin->setMinimum(f_min);
    ui->aperture_spin->setMaximum(f_max);
    ui->aperture_spin->blockSignals(false);

    // Set the minimum and maximum values for the aperture setting.
    min_index = findFnumIndex(f_min);
    if (min_index > 0)
        --min_index;
    if ((f_number < f_min) || qFuzzyCompare(f_number, f_min))
        index = min_index;
    else
        index = findFnumIndex(f_number);
    max_index = findFnumIndex(f_max);
    if (index != aperture_last_index)
    {
        ui->aperture_slide->blockSignals(true);
        ui->aperture_slide->setMinimum(min_index);
        ui->aperture_slide->setMaximum(max_index);
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "min:" << min_index << "max:" << max_index << "index:" << index << "f-number:" << f_number;

        ui->aperture_slide->setValue(index);
        aperture_last_index = index;
        ui->aperture_slide->blockSignals(false);

        // Make sure all the aperture controls are updated.
        apertureValueChanged(index);
    }

    // The min/max for the focus slider goes from "near" to "infinity". However, the
    // actual values for infinity are smaller than those for "near". So we've inverted
    // the appearance (in the designer). This keeps the UI oriented with near to the left
    // and infinity to the right.

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "focus_inf:" << focus_inf << "focus_near:" << focus_near << "read_connected_lens_data:" << read_connected_lens_data << "focus_pos:" << focus_pos << "focus_last_index:" << focus_last_index;

    // Make sure that we don't have a case where we are out of range for the lens.
    if ((focus_pos < focus_inf) || (focus_pos > focus_near))
    {
        // Not sure why the lens would give us a focus_pos outside of the range - it was happening when FasMotion
        // was giving it bogus values to set but this will ensure it is within the lens' range.

        // Set the focus_pos equal to the infinity setting so that we are at least in-range of the settings.
        focus_pos = focus_inf;
    }

    // If we have never set up the lens data for the focus make sure the slider ranges reflect the lens data.
    if (!read_connected_lens_data || (focus_pos != focus_last_index))
    {
        ui->focus_slide->blockSignals(true);
        if (focus_inf < focus_near)
        {
            if ((focus_pos >= focus_inf) && (focus_pos <= focus_near))
            {
                ui->focus_slide->setInvertedAppearance(true);
                if (focus_inf != ui->focus_slide->minimum())
                    ui->focus_slide->setMinimum(focus_inf);
                if (focus_near != ui->focus_slide->maximum())
                    ui->focus_slide->setMaximum(focus_near);
            }
            else
            {
                // The focus_pos value reported is invald.
                focus_pos = focus_last_index;
            }
        }
        else
        {
            if ((focus_pos <= focus_inf) && (focus_pos >= focus_near))
            {
                ui->focus_slide->setInvertedAppearance(false);
                if (focus_inf != ui->focus_slide->maximum())
                    ui->focus_slide->setMaximum(focus_inf);
                if (focus_near != ui->focus_slide->minimum())
                    ui->focus_slide->setMinimum(focus_near);
            }
            else
            {
                // The focus_pos value reported is invald.
                focus_pos = focus_last_index;
            }
        }

        // Did the focus position REALLY change?
        if (focus_pos != focus_last_index)
        {
            ui->focus_slide->setValue(focus_pos);
            focus_last_index = focus_pos;

            focusValueChanged(focus_pos);
        }
        ui->focus_slide->blockSignals(false);
    }
    ui->focus_slide->setEnabled(!manual_focus);
    ui->focus_tele_tb->setEnabled(!manual_focus);
    ui->focus_wide_tb->setEnabled(!manual_focus);
    if (manual_focus)
    {
        ui->focus_lbl->setText(tr("Focus [MF]:"));
    }
    else
    {
        ui->focus_lbl->setText(tr("Focus [AF]:"));
    }

    // Show and enable the autofocus button.
    ui->pb_autofocus->setVisible(true);
    ui->pb_autofocus->setEnabled(true);

    // Make sure the label is set properly.
    ui->focus_min_lbl->setText(QString::number(focus_min) + QStringLiteral(" mm"));

    // The min/max for the zoom slide was setup in the parseLensInfo() method.
    // We just need to adjust the value.
    ui->zoom_slide->blockSignals(true);
    ui->zoom_spin->blockSignals(true);
    if (zoom_divisions < zoom_pos)
    {
        zoom_divisions = zoom_pos;
        ui->zoom_slide->setMaximum(zoom_divisions);
        ui->zoom_spin->setMaximum(zoom_divisions);
    }
    ui->zoom_slide->setValue(zoom_pos);
    zoom_last_index = zoom_pos;
    ui->zoom_spin->setValue(focal_length);
    ui->zoom_spin->blockSignals(false);
    ui->zoom_slide->blockSignals(false);

    // We have read connected lens data.
    read_connected_lens_data = true;
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void PanelLens::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    FMPanelWidget::changeEvent(event);
}

/*!
 * @brief Event routine which is invoked when our timer expires.
 * @param event     Pointer to the event class.
 */
void PanelLens::timerEvent(QTimerEvent *event)
{
    // Pass it on up.
    FMPanelWidget::timerEvent(event);

    // Is it the one we're interested in?
    if (event->timerId() != status_timer_id)
        return;

    // Stop the timer.
    killTimer(status_timer_id);
    status_timer_id = -1;

    // Ask for the current settings.
    requestStatus();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void PanelLens::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    CameraMDI   *camera_ptr;

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Do we have a camera connected?
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // The only unsolicited message we care about is when the lens
    // status changes.
    switch (reg_addr)
    {
    case GIGE_REG_CAMERA_STATE:
        // Disable the panel if the state is wrong. Enable it if we have a
        // lens connected.
        if ((reg_value == GIGE_CAMERA_STATE_IDLE) ||
            (reg_value == GIGE_CAMERA_STATE_REVIEW) ||
            (reg_value == GIGE_CAMERA_STATE_SNAPSHOT) ||
            (reg_value == GIGE_CAMERA_STATE_AUTOSAVE))
            setEnabled(false);
        else
            setEnabled(connected);
        break;
    case GIGE_REG_LENS_STATUS_JSON:
        // Series3 cameras don't get to play.
        if (!camera_ptr->is_series3())
        {
            // We need to query the lens for updated status information.
            requestStatus();
        }
        break;
    default:
        break;
    }
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param ip_addr       The IP address of the camera which didn't respond.
 * @param reg_addr      The address of the register which was read.
 * @param read_op       True if this was a read operation, false for a write operation.
 */
void PanelLens::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Only care if we're visible.
    if (!isVisible())
        return;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << ip_addr << reg_addr << read_op;
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void PanelLens::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    QJsonDocument jsonMedia;
    QJsonObject lensObj;
    QJsonParseError jp_err;
    QString json_str;
    QByteArray *json_data = nullptr;
    CameraMDI   *camera_ptr;
    uint32_t reg_value;

    // Pass it up.
    FMPanelWidget::read_complete(ip_addr, reg_addr);

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Do we have a camera connected?
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    switch (reg_addr)
    {
    case GIGE_REG_CAMERA_STATE:
        // Disable the panel if the state is wrong. Enable it if we have a
        // lens connected.
        reg_value = read_camera_cache(reg_addr);
        if ((reg_value == GIGE_CAMERA_STATE_IDLE) ||
            (reg_value == GIGE_CAMERA_STATE_REVIEW) ||
            (reg_value == GIGE_CAMERA_STATE_SNAPSHOT) ||
            (reg_value == GIGE_CAMERA_STATE_AUTOSAVE))
            setEnabled(false);
        else
            setEnabled(connected);
        break;
    case GIGE_REG_LENS_INFO_JSON:
        // Series3 cameras don't get to play.
        if (camera_ptr->is_series3())
            break;

        // Get the JSON information.
        json_data = camera_ptr->read_camera_buffer(GIGE_REG_LENS_INFO_JSON);
        if (json_data == nullptr)
        {
            retry_addr = GIGE_REG_LENS_INFO_JSON;
            QTimer::singleShot(READ_RETRY_TIMEOUT, this, &PanelLens::retry_read);
            break;
        }

        // Make a QString out of the data.
        json_str = QString::fromUtf8(*json_data);
        if (json_str.length() <= 1)
        {
            retry_addr = GIGE_REG_LENS_INFO_JSON;
            QTimer::singleShot(READ_RETRY_TIMEOUT, this, &PanelLens::retry_read);
            break;
        }

        // Strip any trailing NULL characters.
        if (json_str.at(json_str.size() - 1) == '\0')
            json_str.chop(1);

        // Convert to a JSON document.
        jsonMedia = QJsonDocument::fromJson(json_str.toUtf8(), &jp_err);
        if (jsonMedia.isNull())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_LENS_INFO_JSON Try: " << retry_count << "Got a parse err:" << jp_err.errorString() << "str:" << json_str;
            if (retry_count == MAX_RETRIES)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Giving up.";

                // Make sure the interface shows that we have no lens, or at least none
                // we can communicate with.
                showNoLens();
                request_pending = false;
            }
            else
            {
                // Try the request again.
                ++retry_count;
                QTimer::singleShot(100, this, &PanelLens::requestInfo);
            }
        }
        else
        {
            // Clear our request pending flag and retry counter.
            request_pending = false;
            retry_count = 0;

            // Get the object from the document.
            lensObj = jsonMedia.object();
            parseLensInfo(lensObj);
        }
        break;
    case GIGE_REG_LENS_STATUS_JSON:
        // Series3 cameras don't get to play.
        if (camera_ptr->is_series3())
            break;

        // Get the lut_data.
        json_data = camera_ptr->read_camera_buffer(GIGE_REG_LENS_STATUS_JSON);
        if (json_data == nullptr)
        {
            retry_addr = GIGE_REG_LENS_STATUS_JSON;
            QTimer::singleShot(READ_RETRY_TIMEOUT, this, &PanelLens::retry_read);
            break;
        }

        // Make a QString out of the data.
        json_str = QString::fromUtf8(*json_data);
        if (json_str.length() <= 1)
        {
            retry_addr = GIGE_REG_LENS_STATUS_JSON;
            QTimer::singleShot(READ_RETRY_TIMEOUT, this, &PanelLens::retry_read);
            break;
        }

        // Strip any trailing NULL characters.
        if (json_str.at(json_str.size() - 1) == '\0')
            json_str.chop(1);

        // Convert to a JSON document.
        jsonMedia = QJsonDocument::fromJson(json_str.toUtf8(), &jp_err);
        if (jsonMedia.isNull())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_LENS_STATUS_JSON Try: " << retry_count << "Got a parse err:" << jp_err.errorString() << "str:" << json_str;
            if (retry_count == MAX_RETRIES)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Giving up.";
                request_pending = false;
            }
            else
            {
                // Try the request again.
                ++retry_count;
                QTimer::singleShot(100, this, &PanelLens::requestStatus);
            }
        }
        else
        {
            // Clear our request pending flag and retry counter.
            request_pending = false;
            retry_count = 0;

            // Get the object from the document.
            lensObj = jsonMedia.object();
            parseLensStatus(lensObj);
        }
        break;
    default:
        break;
    }

    // Do we need to cleanup?
    if (json_data != nullptr)
    {
        delete json_data;
        json_data = nullptr;
    }
}

void PanelLens::requestInfo()
{
    CameraMDI   *camera_ptr;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Do we have a camera connected?
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Series3 cameras don't get to play.
    if (!camera_ptr->is_series3())
    {
        // Get the lens information.
        camera_ptr->start_camera_read_buf(GIGE_REG_LENS_INFO_JSON, GIGE_LENS_INFO_JSON_LEN);

        // Remember that we've a request pending.
        request_pending = true;
    }
}

void PanelLens::requestStatus()
{
    CameraMDI   *camera_ptr;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Do we have a camera connected?
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Series3 cameras don't get to play.
    if (!camera_ptr->is_series3())
    {
        // Get the lens information.
        camera_ptr->start_camera_read_buf(GIGE_REG_LENS_STATUS_JSON, GIGE_LENS_STATUS_JSON_LEN);

        // Remember that we've a request pending.
        request_pending = true;
    }
}

void PanelLens::retry_read()
{
    CameraMDI   *camera_ptr;

    // Do we have a camera connected?
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Series3 cameras don't get to play.
    if (camera_ptr->is_series3())
        return;

    // What could we be retrying?
    switch (retry_addr)
    {
    case GIGE_REG_LENS_INFO_JSON:
        camera_ptr->start_camera_read_buf(GIGE_REG_LENS_INFO_JSON, GIGE_LENS_INFO_JSON_LEN);
        break;
    case GIGE_REG_LENS_STATUS_JSON:
        camera_ptr->start_camera_read_buf(GIGE_REG_LENS_STATUS_JSON, GIGE_LENS_STATUS_JSON_LEN);
        break;
    default:
        break;
    }

    // Reset our retry register address.
    retry_addr = 0;
}

void PanelLens::switchPowerChanged(bool value)
{
    CameraMDI *camera_ptr;
    QJsonObject json_obj;

    // Make sure we have a camera connected.
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Build our JSON value and send it on to the camera.
    json_obj.insert("power", value);
    QJsonDocument ap_doc(json_obj);
    camera_ptr->write_camera_buffer(GIGE_REG_LENS_SETTINGS_JSON, ap_doc.toJson());
}

/*!
 * @brief Update the interface to show that we have no lens, or at least none
 *        that we can communicate with.
 */
void PanelLens::showNoLens()
{
    // Hide the controls.
    ui->controls_group->setVisible(false);
    sw_power->setVisible(false);
    ui->headerBorder->setVisible(false);
    adjustSize();
    update();

    // Make sure the name show's no lens information.
    ui->lens_name->setText(LENS_NOT_FOUND);

    // Disable the panel.
    setEnabled(false);
}

/*!
 * @brief Toggle the visiblity of the panel. Optionally, pass the event up
 *        to the parent.
 *
 * @param notify_parent  True if we're supposed to pass this up to the parent.
 */
void PanelLens::toggle_visibility(bool notify_parent)
{
    // Pass it on up.
    FMPanelWidget::toggle_visibility(notify_parent);

    // If we're opened up, make sure we refresh our controls
    if (isVisible())
    {
        // Do we have a camera connected?
        if (p_control_panel->getCamera() != nullptr)
        {
            // Update our panel.
            update_panel();
        }
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void PanelLens::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Only care about our registers.
    switch (reg_addr)
    {
    case GIGE_REG_LENS_SETTINGS_JSON:
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(camera_ip_addr, reg_addr);
        if (write_status != GEV_STATUS_SUCCESS)
        {
            QHostAddress    our_addr;
            our_addr.setAddress(qFromBigEndian(camera_ip_addr));

            QString err_msg = QString::fromUtf8("Unable to write register 0x%1 to IP adddress %2").arg(reg_addr, 0, 16).arg(our_addr.toString());
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }

        // Allow updates now. What this means is that, the next status message we
        // get will re-enable the slider.
        update_allowed = true;
        break;
    default:
        break;
    }
}

void PanelLens::apertureValueChanged(int value)
{
    float f_number;

    // Convert the index to an f-number.
    f_number = QUARTER_STOP_FNUM_SCALE[value];

    // Reflect our expected f-number in the spinner.
    ui->aperture_spin->setValue(f_number);

    // De-select the value.
    ui->aperture_spin->findChild<QLineEdit *>()->deselect();
}

void PanelLens::apertureSliderReleased()
{
    int value;
    float f_number;
    CameraMDI *camera_ptr;
    QJsonObject json_obj;

    // Are we allowing updates?
    if (!update_allowed)
        return;

    // Make sure we have a camera connected.
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Get the current value.
    value = ui->aperture_slide->value();

    // Did we move at all?
    if (value == aperture_last_index)
        return;

    // Remember where we are.
    aperture_last_index = value;

    // Convert the index to an f-number.
    f_number = QUARTER_STOP_FNUM_SCALE[value];

    // Build our JSON value and send it on to the camera.
    json_obj.insert("f_number", f_number);
    QJsonDocument ap_doc(json_obj);
    camera_ptr->write_camera_buffer(GIGE_REG_LENS_SETTINGS_JSON, ap_doc.toJson());

    // If the aperture doesn't change, the lens won't send a status update. In that
    // case, we need to force a status request. We'll use a timer for this.
    if (status_timer_id == -1)
        status_timer_id = startTimer(STATUS_TIMEOUT);

    // Disable the slider so the user can change it while we're communicating with the camera.
    update_allowed = false;
}

void PanelLens::apertureSlideSlowChange(int value)
{
    // Make sure there was a change.
    if (aperture_last_index == value)
        return;

    // Update the displayed information.
    apertureValueChanged(value);

    // Update the lens.
    apertureSliderReleased();
}

void PanelLens::apertureSpinValueChanged(double value)
{
    Q_UNUSED(value)

    // De-select the value.
    ui->aperture_spin->findChild<QLineEdit *>()->deselect();
}

void PanelLens::focusValueChanged(int value)
{
    // Only enable the "wide" and "tele" toolbutton if we're not already at the limits.
    if (ui->focus_slide->invertedAppearance())
    {
        ui->focus_wide_tb->setEnabled(value != ui->focus_slide->maximum());
        ui->focus_tele_tb->setEnabled(value != ui->focus_slide->minimum());
    }
    else
    {
        ui->focus_wide_tb->setEnabled(value != ui->focus_slide->minimum());
        ui->focus_tele_tb->setEnabled(value != ui->focus_slide->maximum());
    }

}

void PanelLens::focusSliderReleased()
{
    int value;
    CameraMDI *camera_ptr;
    QJsonObject json_obj;

    // Are we allowing updates?
    if (!update_allowed)
        return;

    // Make sure we have a camera connected.
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Get the current value.
    value = ui->focus_slide->value();

    // Did we move at all?
    if (value == focus_last_index)
        return;

    // Remember where we are.
    focus_last_index = value;

    // This should not happen but we want to make sure that we have read the lens information of the
    // connected camera before we send it data.  Camera server will now look for bad data but lets
    // output a warning in this instance.
    if (!read_connected_lens_data)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Writing to GIGE_REG_LENS_SETTINGS_JSON before read_connected_lens_data for " << camera_ptr->get_camera_name();
    }

    // Build our JSON value and send it on to the camera.
    json_obj.insert("focus_dist", value);
    QJsonDocument ap_doc(json_obj);
    camera_ptr->write_camera_buffer(GIGE_REG_LENS_SETTINGS_JSON, ap_doc.toJson());

    // Disable the slider so the user can change it while we're communicating with the camera.
    update_allowed = false;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "min:" << ui->focus_slide->minimum() << "max:" << ui->focus_slide->maximum() << "value:" << ui->focus_slide->value();
}

void PanelLens::focusSlideSlowChange(int value)
{
    // Make sure there was a change.
    if (focus_last_index == value)
        return;

    // Update the displayed informaiton.
    focusValueChanged(value);

    // Update the lens.
    focusSliderReleased();
}

/*!
 * @brief The user has modified the adjustment step size for each [N] or [F] button press on the Focus[AF] slider.
 */
void PanelLens::focusStepSizeChanged()
{
    // Modify the slider to use the new step size for Focus[AF] slider changes represented by the [N] and [F] buttons.
    focus_step_size = ui->focus_step_spin->value();
    ui->focus_slide->setPageStep(focus_step_size);
}

void PanelLens::focusStepTele()
{
    // Move one step towards infinity.
    ui->focus_slide->setValue(ui->focus_slide->value() - focus_step_size);
    focusSliderReleased();

    // Update the UI.
    focusValueChanged(ui->focus_slide->value());
}

void PanelLens::focusStepWide()
{
    // Move one step closer.
    ui->focus_slide->setValue(ui->focus_slide->value() + focus_step_size);
    focusSliderReleased();

    // Update the UI.
    focusValueChanged(ui->focus_slide->value());
}

void PanelLens::zoomSliderReleased()
{
    int value;
    CameraMDI *camera_ptr;
    QJsonObject json_obj;

    // Are we allowing updates?
    if (!update_allowed)
        return;

    // Make sure we have a camera connected.
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Get the current value.
    value = ui->zoom_slide->value();

    // Did we move at all?
    if (value == zoom_last_index)
        return;

    // Remember where we are.
    zoom_last_index = value;

    // Build our JSON value and send it on to the camera.
    json_obj.insert("zoom_enc", value);
    QJsonDocument ap_doc(json_obj);
    camera_ptr->write_camera_buffer(GIGE_REG_LENS_SETTINGS_JSON, ap_doc.toJson());

    // If the focal length doesn't change, the lens won't send a status update. In that
    // case, we need to force a status request. We'll use a timer for this.
    if (status_timer_id == -1)
        status_timer_id = startTimer(STATUS_TIMEOUT);

    // Disable the slider so the user can change it while we're communicating with the camera.
    update_allowed = false;
}

void PanelLens::zoomSlideSlowChange(int value)
{
    // Make sure there was a change.
    if (zoom_last_index == value)
        return;

    // Update the lens.
    zoomSliderReleased();
}

void PanelLens::autoFocus()
{
    CameraMDI *camera_ptr;
    QJsonObject json_obj;

    // Are we allowing updates?
    if (!update_allowed)
        return;

    // Make sure we have a camera connected.
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Build our JSON value and send it on to the camera.
    json_obj.insert("autofocus", true);
    QJsonDocument ap_doc(json_obj);
    camera_ptr->write_camera_buffer(GIGE_REG_LENS_SETTINGS_JSON, ap_doc.toJson());
}
