/*!
 * @file panel_lens.h
 * @brief This is the header file for the Lens control panel.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PANEL_LENS_H
#define PANEL_LENS_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QWidget>
#include "panel_widget.h"

namespace Ui
{
class PanelLens;
}

//--------------------------------------------------------------------
// Forward declarations
//--------------------------------------------------------------------
class QJsonObject;
class fiSwitchButton;

//--------------------------------------------------------------------
// Class declaration
//--------------------------------------------------------------------
class PanelLens : public FMPanelWidget
{
    Q_OBJECT

public:
    explicit PanelLens(QWidget *parent = nullptr);
    ~PanelLens();

    // Handle camera connect/disconnects.
    virtual void camera_connect(uint32_t ip_addr);
    virtual void camera_disconnect();

    // Update the panel contents.
    virtual void update_panel();

    double getAperture();
    int getFocalLength();
    void getLensName(std::string& lens_name);

protected:
    virtual void    changeEvent(QEvent *event);

protected slots:
    virtual void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    virtual void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    virtual void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    virtual void    timerEvent(QTimerEvent *event);
    virtual void    toggle_visibility(bool notify_parent = true);
    virtual void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private slots:
    void    apertureSliderReleased();
    void    apertureSlideSlowChange(int value);
    void    apertureSpinValueChanged(double value);
    void    focusSliderReleased();
    void    focusSlideSlowChange(int value);
    void    focusStepSizeChanged();
    void    focusStepWide();
    void    focusStepTele();
    void    requestInfo();
    void    requestStatus();
    void    retry_read();
    void    switchPowerChanged(bool value);
    void    zoomSliderReleased();
    void    zoomSlideSlowChange(int value);

    void    autoFocus();

private:
    void    init_controls();
    void    showNoLens();
    void    translate_ui();

    void    apertureValueChanged(int value);
    void    focusValueChanged(int value);
    int     findFnumIndex(double f_num);
    void    parseLensInfo(const QJsonObject& lensObj);
    void    parseLensStatus(const QJsonObject& lensObj);

private:
    Ui::PanelLens *ui;
    fiSwitchButton *sw_power;   // The erase Yes/No button.

    static const QString LENS_CONTROL_ACQUIRING;
    static const QString LENS_CONTROL_DISABLED;
    static const QString LENS_NOT_FOUND;
    static const QString LENS_UNKNOWN;
    static const QString LENS_WAITING_FOR_POWER;

    uint32_t    camera_ip_addr;                 //!< IP address of the camera we're talking to.

    int aperture_last_index;
    int focus_last_index;
    int focus_step_size;                        //!< Size of each [N] or [F] increment on the Focus[AF] slider.
    int zoom_last_index;
    int status_timer_id;                        //!< Timer ID to trigger a status request.

    bool update_allowed;
    bool request_pending;
    int retry_count;

    uint32_t retry_addr;

    // Keep track of when cameras connect so that we know that at least
    // the initial lens data has been read from the connected camera.
    bool read_connected_lens_data;

    // Lens information.
    bool connected;
    bool powered;
    bool talking;
    bool ez_enabled;

    int focal_len_wide;
    int focal_len_tele;
    int zoom_divisions;
    int focal_length;                           //!< Focal length, given the current zoom encoder position.

    QString lens_name;
};

#endif // PANEL_LENS_H
