/*!
 * @file panel_reccontrols.h
 * @brief This is the header file for the PanelRecControls class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PANEL_RECCONTROLS_H
#define PANEL_RECCONTROLS_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QWidget>
#include "panel_widget.h"

namespace Ui
{
class PanelRecControls;
}

namespace REC_CONTROLS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! An enumeration to define the various types of external IO we can have.
enum EXT_IO_TYPES
{
    EXTIO_AO = 0,               //!< ARM-OUT.
    EXTIO_AI,                   //!< ARM-IN.
    EXTIO_SO,                   //!< SYNC-OUT.
    EXTIO_SI,                   //!< SYNC-IN.
    EXTIO_TO,                   //!< TRIGGER-OUT.
    EXTIO_TI,                   //!< TRIGGER-IN.
    EXTIO_NONE                  //!< No external I/O is defined.
};

}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QFocusEvent;
class QLabel;
class QLineEdit;
class QRadioButton;
class QTimerEvent;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class PanelRecControls : public FMPanelWidget
{
    Q_OBJECT

public:
    explicit PanelRecControls(QWidget *parent = nullptr);
    ~PanelRecControls();

    static const QString EXTIO_ACT_ACTIVE_STYLE;
    static const QString EXTIO_ACT_INACTIVE_STYLE;
    static const QString TAB_STYLE_SELECTED;

    static const QString TXT_STYLE_COLOR_DISABLED;
    static const QString TXT_STYLE_COLOR_ACTSIG;
    static const QString TXT_STYLE_COLOR_NOSIG;
    static const QString TXT_STYLE_COLOR_WAITSIG;
    static const QString TXT_STYLE_COLOR_WARN;
    static const QString RECORD_CONTROL_QSETTINGS_GROUP;
    static const QString RECORD_CONTROL_SYNC_OUT_REG;
    static const QString RECORD_CONTROL_SYNC_OUT_DATA_LIST;
    static const QString RECORD_CONTROL_SYNC_OUT_IP;
    static const QString RECORD_CONTROL_SYNC_OUT_DATA;
    static const QString RECORD_CONTROL_SYNC_OUT_TIME;

    // Handle camera connect/disconnects.
    virtual void camera_connect(uint32_t ip_addr);
    virtual void camera_disconnect();

    void enableVtControls(bool do_enable);

    QString getExtioLabel(REC_CONTROLS::EXT_IO_TYPES ext_io);
    REC_CONTROLS::EXT_IO_TYPES getType(int channel);

    bool syncInIsEnabled()          { return (get_sync_in_channel() >= 0); }

    void setLtrMode(uint32_t ltr_mode);


    // Update the panel contents.
    virtual void update_panel();

signals:
    void settingExtArmInEnabled();
    void settingExtArmInDisabled();

protected:
    virtual void    changeEvent(QEvent *event);
    virtual void    focusInEvent(QFocusEvent *event);

protected slots:
    virtual void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    virtual void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    virtual void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    virtual void    toggle_visibility(bool notify_parent = true);
    virtual void    timerEvent(QTimerEvent *event);
    virtual void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private slots:
    void        chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr);
    uint32_t    get_master_framerate();
    uint32_t    get_max_framerate();
    void        irig_clicked(bool checked);
    void        irig_year_clicked(bool checked);
    void        rb_use_frame_clicked(bool checked);
    void        rb_use_percent_clicked(bool checked);

    void        read_retry(uint32_t reg_addr);
    void        request_vt_info();

    void        sld_trigger_pos_changed(int new_value);
    void        spin_trigger_pos_changed(int new_value);
    void        tab_enable_clicked(QWidget *frame_widget);
    void        trigger_pos_action(int action);
    void        trigger_pos_update();

    void        trig_in_channel(bool checked);
    void        trig_in_enable_clicked(bool checked);
    void        trig_in_falling_clicked(bool checked);
    void        trig_in_rising_clicked(bool checked);

    void        arm_in_allow_discard(bool checked);
    void        arm_in_channel(bool checked);
    void        arm_in_edge_type_clicked(bool checked);
    void        arm_in_enable_clicked(bool checked);
    void        arm_in_falling_low_clicked(bool checked);
    void        arm_in_level_type_clicked(bool checked);
    void        arm_in_rising_high_clicked(bool checked);

    void        arm_out_active_high_clicked(bool checked);
    void        arm_out_active_low_clicked(bool checked);
    void        arm_out_channel(bool checked);
    void        arm_out_enable_clicked(bool checked);
    void        arm_out_invert_signal_clicked(bool checked);
    void        arm_out_pass_thru_clicked(bool checked);
    void        arm_out_pulse_width_changed(int value);
    void        arm_out_pulse_width_clicked(bool checked);

    void        sync_in_channel(bool checked);
    void        sync_in_delay_changed(int value);
    void        sync_in_enable_clicked(bool checked);
    void        sync_in_falling_clicked(bool checked);
    void        sync_in_mfr_changed(int value);
    void        sync_in_per_frame_clicked(bool checked);
    void        sync_in_per_sec_clicked(bool checked);
    void        sync_in_rate_div_changed(int value);
    void        sync_in_rising_clicked(bool checked);

    void        sync_out_active_low_clicked(bool checked);
    void        sync_out_active_high_clicked(bool checked);
    void        sync_out_channel(bool checked);
    void        sync_out_duty_cycle_changed(int value);
    void        sync_out_duty_cycle_clicked(bool checked);
    void        sync_out_enable_clicked(bool checked);
    void        sync_out_invert_signals_clicked(bool checked);
    void        sync_out_pass_thru_clicked(bool checked);
    void        sync_out_per_frame_clicked(bool checked);
    void        sync_out_per_sec_clicked(bool checked);
    void        sync_out_shutter_clicked(bool checked);

    void        trig_out_active_low_clicked(bool checked);
    void        trig_out_active_high_clicked(bool checked);
    void        trig_out_channel(bool checked);
    void        trig_out_enable_clicked(bool checked);
    void        trig_out_invert_signal_clicked(bool checked);
    void        trig_out_pass_thru_clicked(bool checked);
    void        trig_out_pulse_width_changed(int value);

    void        tt_enable(bool checked);
    void        tt_incr();
    void        tt_hours_changed(int value);
    void        tt_minutes_changed(int value);
    void        tt_seconds_changed(int value);
    void        tt_restart();
    void        tt_start();
    void        tt_stop();

    void        vt_area_changed(int new_area);
    void        vt_change_changed(int new_change);
    void        vt_enable(bool checked);
    void        vt_mark_only(bool checked);
    void        vt_refresh_changed(int new_refresh);
    void        vt_reset();

    void        write_arm_in();
    void        write_arm_out();
    void        write_sync_in();
    void        write_sync_out();
    void        write_trig_out();

private:
    void        display_time(double time_value, QLabel *lbl_time, QLabel *eb_time);
    void        handle_read(uint32_t reg_addr, uint32_t reg_value);
    void        init_controls();
    void        update_duty_cycle();
    void        update_vt_info();
    void        process_camera_state(uint32_t new_state);
    void        read_registers();
    void        release_channel(int channel);
    int         reserve_channel(int asking_channel = -1);
    void        set_tab_icon(int tab_index, bool is_enabled, int channel = -1);
    void        translate_ui();
    void        update_channel_buttons(QRadioButton *chan1, QRadioButton *chan2, QRadioButton *chan3, int used_channel);

    int         get_arm_in_channel();
    int         get_arm_out_channel();
    int         get_sync_in_channel();
    int         get_sync_out_channel();
    int         get_trig_in_channel();
    int         get_trig_out_channel();

    void        set_custom_styles(bool enabled);

    void        poll_extio_activity();
    void        start_extio_activity_polling();

    void        style_label_warning(QLabel* label_to_style);
    void        style_label_normal(QLabel* label_to_style);

    void        sync_in_connect_camera();
    void        sync_in_disconnect_camera();

    void        sync_out_gige_reg_to_ui(uint32_t sync_out_reg);
    bool        sync_out_load_settings(uint32_t& sync_out_reg_data);
    void        sync_out_save_settings(uint32_t sync_out_reg);
    uint32_t    sync_out_ui_to_gige_reg();

    void        update_arm_in();
    void        update_arm_out();
    void        update_extio_activity();
    void        update_ext_trigger();
    void        update_irig();
    void        update_num_frames();
    void        update_sync_in();
    void        update_sync_out();
    void        update_trig_out();
    void        update_trigger();

private:
    Ui::PanelRecControls    *ui;

    int arm_in_chan;
    int arm_out_chan;
    int sync_in_chan;
    int sync_out_chan;
    int trig_in_chan;
    int trig_out_chan;
    uint32_t    channels_in_use;                //!< Bitmap of the channels being used: bit0 = channel_1, bit1 = channel2, etc.

    uint32_t    camera_ip_addr;                 //!< IP address of the camera we're talking to.
    uint32_t    duty_cycle_fps;                 //!< What was the duty cycle frame rate when last calculated.

    uint32_t    num_frames;                     //!< Used frequently throughout the code, this is the maximum number of frames possible in the session.
    int         extio_read_pending;             //!< Number of ticks we wait before issuing another read.
    QTimer      extio_activity_timer;           //!< A polling timer to query the GIGE_REG_EXTIO_ACTIVITY register at periodic intervals.

    bool        update_in_progress;             //!< A flag to indicate we're writing values out to the camera.
    uint32_t    update_reg_addr;                //!< Register we're updating.
    uint32_t    update_reg_value;               //!< Register value we're waiting for.
    int         update_retry_count;             //!< How many times we've tried to update a register.
    int         update_timer_id;                //!< A timer to used for polling register values.
    bool        vt_allowed;                     //!< A flag to remember if video trigger is allowed or not.
};

#endif // PANEL_RECCONTROLS_H
