/*!
 * @file panel_recsettings.cpp
 * @brief This file contains the implementation of the PanelRecSettings
 *        class. This class handles the record settings for the camera.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "panel_recsettings.h"
#include "ui_panel_recsettings.h"

#include <QAction>
#include <QHostAddress>
#include <QMdiSubWindow>
#include <QMessageBox>
#include <QObject>
#include <qmath.h>
#include <QString>
#include <QStringLiteral>

#include "client_math.h"
#include "camera_dock.h"
#include "camera_info.h"
#include "camera_mdi.h"
#include "camera_video.h"
#include "fi_sensor.h"
#include "mainwindow.h"
#include "panel_reccontrols.h"

namespace REC_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Log messages for testing purposes.  Set to 1 to have them compiled for output.
#define LOG_GROUP_READS  0

//! An enumeration to reflect the various aspect ratios we support.
//! NOTE: This needs to stay in sync with the ".ui" settings.
enum
{
    ASPECT_CUSTOM = 0,                  //!< Custom aspect ratio defined by the width and height.
    ASPECT_1x1,                         //!< 1:1 aspect ratio.
    ASPECT_4x3,                         //!< 4:3 aspect ratio.
    ASPECT_5x4,                         //!< 5:4 aspect ratio (native LUPA-1300).
    ASPECT_16x9                         //!< 16:9 aspect ratio.
};

//! The minimum frame rate to calculate low light shutter speed.
const uint32_t LOW_LIGHT_FRAMERATE_MIN  = 24;

//! The maximum frame rate to calculate low light shutter speed.
const uint32_t LOW_LIGHT_FRAMERATE_MAX  = 125;

//! Amount of time to wait for the sensor object to be ready.
const int       UPDATE_DELAY                = 100;

//! +/- this amount to check for a floating point equality.
const double    FLOAT_EPSLION               = 0.0001;

}   // End of namespace REC_SETTINGS.

using namespace REC_SETTINGS;
using namespace FM_PANEL;

/*!
 * @brief Constructor for the PanelRecSettings class. Initialize variables.
 */
PanelRecSettings::PanelRecSettings(QWidget *parent) :
    FMPanelWidget(parent),
    ui(new Ui::PanelRecSettings),
    camera_ip_addr(0)
{
    // Setup our UI controls.
    ui->setupUi(this);

    // Make sure we got a parent pointer.
    if (parent != nullptr)
    {
        // My grandparent is the mainwindow.
        MainWindow  *p_mainwnd = qobject_cast<MainWindow *>(parent->parent());
        p_mainwnd->get_cp_recsettings();

        // Connect the menu item's I'm interested in.
        connect(p_mainwnd->get_cp_recsettings(), &QAction::triggered, this, &PanelRecSettings::toggle_visibility);
    }

    // Initialize our variables.
    ae_allowed = true;
    connecting_camera = false;
    custom_height = 0;
    custom_width = 0;
    last_reg_addr = FI_COMM::NO_REGISTER;
    ltr_enabled = false;
    retry_counter = 0;
    update_timer_id = 0;
    update_in_progress = false;
    update_pending = false;
    write_list.clear();
    wait_for_video_pause = false;
    fpn_disabled_index = -1;
    fpn_pixel_index = -1;
    fpn_sensor_index = -1;

    camera_hold = false;

    // Initialize our controls.
    init_controls();

    // After everything is setup, setup the intial size information.
    set_initial_sizing();

    // Start out with things disabled until we get connected to a camera.
    setEnabled(false);
}

/*!
 * @brief Destructor for the PanelRecSettings class.
 */
PanelRecSettings::~PanelRecSettings()
{
    delete ui;
}

/*!
 * @brief Utility function to calculate either the height or width of an ROI based
 *        on the width or height, respectively, and the aspect ratio.
 *
 * @param calc_height   True to calculate the height, false for the width.
 * @param base_value    Either the width or height of the ROI...whichever one is known.
 * @param aspect_type   The aspect ratio we're to use.
 *
 * @return The calculated height or width, as selected by the calc_height parameter.
 */
int PanelRecSettings::aspect_update(bool calc_height, int base_value, int aspect_type)
{
    uint32_t    ret_value;

    // Process based on the aspect ratio selected.
    switch (aspect_type)
    {
    case ASPECT_1x1:
        ret_value = base_value;
        break;
    case ASPECT_4x3:
        if (calc_height)
            ret_value = (uint32_t)((((base_value * 30) / 4) + 5) / 10);
        else
            ret_value = (uint32_t)((((base_value * 40) / 3) + 5) / 10);
        break;
    case ASPECT_5x4:
        if (calc_height)
            ret_value = (uint32_t)((((base_value * 40) / 5) + 5) / 10);
        else
            ret_value = (uint32_t)((((base_value * 50) / 4) + 5) / 10);
        break;
    case ASPECT_16x9:
        if (calc_height)
            ret_value = (uint32_t)((((base_value * 90) / 16) + 5) / 10);
        else
            ret_value = (uint32_t)((((base_value * 160) / 9) + 5) / 10);
        break;
    default:
    case ASPECT_CUSTOM:
        if (calc_height)
        {
            // Sanity check.
            if (custom_width == 0)
                ret_value = 0;
            else
                ret_value = (uint32_t)(((((uint32_t)(base_value) * custom_height * 10) / custom_width) + 5) / 10);
        }
        else
        {
            // Sanity check.
            if (custom_height == 0)
                ret_value = 0;
            else
                ret_value = (uint32_t)(((((uint32_t)(base_value) * custom_width * 10) / custom_height) + 5) / 10);
        }
        break;
    }

    // Make sure we have an even value.
    if (ret_value & 1)
        ret_value++;

    // Return our calculated value.
    return (int)(ret_value);
}

/*!
 * @brief Utility method to calculate the maximum ROI which will support the specified frame rate.
 * @param framerate     The frame rate we are interested in.
 */
void PanelRecSettings::auto_max_roi(uint32_t framerate)
{
    uint32_t    offset_x = 0;
    uint32_t    offset_y = 0;
    uint32_t    ssd_bandwidth;
    uint32_t    max_fps;

    QRect       max_roi;
    CameraMDI   *camera_ptr;

    double  new_width = 0.0;
    double  new_height = 0.0;
    double  aspect_ratio;

    bool    found_it = false;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Read the bandwidth allowed for the SSD transfers.
    if (ltr_enabled)
    {
        bool        ok;

        // Get the SSD bandwidth used by the camera.
        ssd_bandwidth = read_camera_cache(GIGE_REG_SSD_BANDWIDTH, &ok);
        if (!ok)
            ssd_bandwidth = FASMOTION_SETTINGS::DEFAULT_SSD_BANDWIDTH;
    }
    else
    {
        // No SSD bandwidth to worry about.
        ssd_bandwidth = 0;
    }

    // Figure out the aspect ratio based on the user's selection.
    switch (ui->cb_aspect->currentIndex())
    {
    case ASPECT_1x1:
        aspect_ratio = 1.0;
        break;
    case ASPECT_4x3:
        aspect_ratio = 4.0 / 3.0;
        break;
    case ASPECT_5x4:
        aspect_ratio = 5.0 / 4.0;
        break;
    case ASPECT_16x9:
        aspect_ratio = 16.0 / 9.0;
        break;
    default:
    case ASPECT_CUSTOM:
        // Custom aspect ratio uses the saved aspect ratio.
        aspect_ratio = (double)custom_width / (double)custom_height;
        break;
    }

    // Get the maximum ROI supported by the selected bitmode.
    max_roi = camera_ptr->get_sensor()->max_roi((uint32_t)ui->cb_bit_depth->currentData().toUInt());

    if (ui->spin_width->value() > ui->spin_height->value())
    {
        // Initialize to our maximum allowed width.
        new_width = (double)ui->spin_width->maximum();

        // Move through width until we find the maximum ROI at the current aspect ratio
        // that will support the requested framerate.
        while (!found_it && (((int)new_width) >= ui->spin_width->minimum()))
        {
            // The regular rounding function rint() and floor() seem to cause drift, ceil
            // doesn't seem to cause the same drift
            new_height = (qCeil(new_width / aspect_ratio) / camera_ptr->get_sensor()->height_mod()) * camera_ptr->get_sensor()->height_mod();

            // Center the ROI.
            offset_x = (((uint32_t)(max_roi.width()) >> 1) - (((uint32_t)new_width) >> 1)) >> 1;
            offset_x = (offset_x / camera_ptr->get_sensor()->width_mod()) * camera_ptr->get_sensor()->width_mod();
            offset_y = (((uint32_t)(max_roi.height()) >> 1) - (((uint32_t)new_height) >> 1)) >> 1;
            offset_y = (offset_y / camera_ptr->get_sensor()->height_mod()) * camera_ptr->get_sensor()->height_mod();

            max_fps = ClientMath::calc_max_framerate(camera_ptr->get_sensor(), offset_x,
                                                     (uint32_t)new_width, (uint32_t)new_height,
                                                     ui->cb_bit_depth->currentData().toInt(),
                                                     ltr_enabled, ssd_bandwidth, camera_ptr->is_series3());
            if (framerate <= max_fps)
            {
                found_it = true;
            }
            else
            {
                // try a smaller width and keep it even
                new_width -= camera_ptr->get_sensor()->width_mod();
            }
        }
    }
    else
    {
        // Initialize to our maximum allowed height.
        new_height = (double)ui->spin_height->maximum();

        // Move through the height until we find the maximum roi at the current aspect ratio
        while (!found_it && ((int)new_height >= ui->spin_height->minimum()))
        {
            // The regular rounding function rint() and floor() seem to cause drift, ceil
            // doesn't seem to cause the same drift
            new_width = (qCeil(new_height * aspect_ratio) / camera_ptr->get_sensor()->width_mod()) * camera_ptr->get_sensor()->width_mod();

            // Center the ROI.
            offset_x = (((uint32_t)(max_roi.width()) >> 1) - (((uint32_t)new_width) >> 1)) >> 1;
            offset_x = (offset_x / camera_ptr->get_sensor()->width_mod()) * camera_ptr->get_sensor()->width_mod();
            offset_y = (((uint32_t)(max_roi.height()) >> 1) - (((uint32_t)new_height) >> 1)) >> 1;
            offset_y = (offset_y / camera_ptr->get_sensor()->height_mod()) * camera_ptr->get_sensor()->height_mod();

            max_fps = ClientMath::calc_max_framerate(camera_ptr->get_sensor(), offset_x,
                                                     (uint32_t)new_width, (uint32_t)new_height,
                                                     ui->cb_bit_depth->currentData().toInt(),
                                                     ltr_enabled, ssd_bandwidth, camera_ptr->is_series3());
            if (framerate <= max_fps)
                found_it = true;
            else
                new_height -= camera_ptr->get_sensor()->height_mod(); // try a smaller height and keep it even
        }
    }

    // If we were sucessful then set the new values into the controls
    if (found_it)
    {
        int new_value;

        new_value = (int)((new_width / camera_ptr->get_sensor()->width_mod()) * camera_ptr->get_sensor()->width_mod());
        if (new_value != ui->spin_width->value())
        {
            ui->spin_width->blockSignals(true);
            ui->spin_width->setValue(new_value);
            ui->spin_width->blockSignals(false);
        }
        set_slider_value(ui->sld_width, new_value / camera_ptr->get_sensor()->width_mod());

        new_value = (int)((new_height / camera_ptr->get_sensor()->height_mod()) * camera_ptr->get_sensor()->height_mod());
        if (new_value != ui->spin_height->value())
        {
            ui->spin_height->blockSignals(true);
            ui->spin_height->setValue(new_value);
            ui->spin_height->blockSignals(false);
        }
        set_slider_value(ui->sld_height, new_value / camera_ptr->get_sensor()->height_mod());

        // The offset value is the CENTERed offset. Multiply by 2 to get the maximum offset.
        // Remember that our sliders are 1/modulo the values of the spinners. This is so
        // we can force the values to be appropriate values.
        if (((offset_x * 2) / camera_ptr->get_sensor()->width_mod()) != (uint32_t)ui->sld_offsetx->maximum())
        {
            int offset = (int)((offset_x * 2) / camera_ptr->get_sensor()->width_mod());
            ui->sld_offsetx->setMaximum(offset);
            ui->spin_offsetx->setMaximum(offset * camera_ptr->get_sensor()->width_mod());
        }
        if (offset_x != (uint32_t)ui->spin_offsetx->value())
            ui->spin_offsetx->setValue((int)offset_x);
        set_slider_value(ui->sld_offsetx, offset_x / camera_ptr->get_sensor()->width_mod());

        if (((offset_y * 2) / camera_ptr->get_sensor()->height_mod()) != (uint32_t)ui->sld_offsety->maximum())
        {
            int offset = (int)((offset_y * 2) / camera_ptr->get_sensor()->height_mod());
            ui->sld_offsety->setMaximum(offset);
            ui->spin_offsety->setMaximum(offset * camera_ptr->get_sensor()->height_mod());
        }
        if (offset_y != (uint32_t)ui->spin_offsety->value())
            ui->spin_offsety->setValue((int)offset_y);
        set_slider_value(ui->sld_offsety, offset_y / camera_ptr->get_sensor()->height_mod());
    }
}

/*!
 * @brief Determine the aspect ratio based on the width and height values.
 *
 * @return The new aspect ration.
 */
int PanelRecSettings::calc_aspect_ratio(int width, int height)
{
    int new_index;

    // Walk through all the aspect ratios and see if we can make a match.
    if (aspect_update(true, width, ASPECT_1x1) == height)
        new_index = ASPECT_1x1;
    else if (aspect_update(true, width, ASPECT_4x3) == height)
        new_index = ASPECT_4x3;
    else if (aspect_update(true, width, ASPECT_5x4) == height)
        new_index = ASPECT_5x4;
    else if (aspect_update(true, width, ASPECT_16x9) == height)
        new_index = ASPECT_16x9;
    else
        new_index = ASPECT_CUSTOM;

    // Return the aspect ratio.
    return new_index;
}

/*!
 * @brief Calculate the min and max of the lowlight spin and slider.
 * It sets them without sending any signals.
 */
void PanelRecSettings::calc_lowlight_range(uint32_t frame_overhead)
{
    uint32_t    lowlight_exposure;
    lowlight_exposure = ClientMath::calc_max_exposure(LOW_LIGHT_FRAMERATE_MAX, frame_overhead);
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ll min using fr of 125:" << lowlight_exposure << "frame_overhead:" << frame_overhead;

    // We are just updating the range so we don't want any effect with any signals.
    ui->sld_ss_ll->blockSignals(true);
    ui->spin_ss_ll->blockSignals(true);

    ui->sld_ss_ll->setMinimum((int)lowlight_exposure);
    ui->spin_ss_ll->setMinimum((int)lowlight_exposure);

    lowlight_exposure = ClientMath::calc_max_exposure(LOW_LIGHT_FRAMERATE_MIN, frame_overhead);
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ll max using fr of 24:" << lowlight_exposure << "frame_overhead:" << frame_overhead;
    ui->sld_ss_ll->setMaximum((int)lowlight_exposure);
    ui->spin_ss_ll->setMaximum((int)lowlight_exposure);

    // We are done updating the range making sure to restore them.
    ui->sld_ss_ll->blockSignals(false);
    ui->spin_ss_ll->blockSignals(false);
}

/*!
 * @brief Utility function to calculate the maximum number of frames represented by
 *        the memory and ROI settings.
 *
 * @return The maximum number of frames supported by the memory and ROI settings.
 */
uint32_t PanelRecSettings::calc_max_total_frames()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return 0;
    }

    uint32_t max_frames = (uint32_t) camera_ptr->calc_max_frames();

    // Return our result.
    //return ClientMath::calc_max_total_frames(camera_ip_addr, (uint32_t)ui->spin_width->value(), (uint32_t)ui->spin_height->value(), ltr_enabled, camera_ptr->is_series3());
    return max_frames;
}

/*!
 * @brief Invoked when we connect the panel to a camera.
 * @param ip_addr   The IP address of the camera we're connecting to. This may be nullptr if we're not connecting
 *                  to any cameras.
 */
void PanelRecSettings::camera_connect(uint32_t ip_addr)
{
    FI_comm         *comm_ptr;
    CameraMDI       *camera_ptr = nullptr;

    // Make sure we're not already connected to this camera.
    if (ip_addr == camera_ip_addr)
        return;

    // Save the IP address.
    camera_ip_addr = ip_addr;

    // Get our communications pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get a FI_comm object!";
        return;
    }

    // Make sure there is a camera pointer as we need to connect to some signals.
    if (p_control_panel)
    {
        camera_ptr = p_control_panel->getCamera();
    }
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    connecting_camera = true;   // We are starting the connect to camera process.

    // Disable the center checkbox it will be set based on the camera's setting in the
    // update control function that happens as slot of the timer below.
    ui->chk_center_roi->setChecked(false);

    // Connect the communications interfaces that we're interested in.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelRecSettings::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &PanelRecSettings::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelRecSettings::handle_unsol_msg, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelRecSettings::write_complete, Qt::QueuedConnection);

    // Connect up to any loss or resume of sync-in stream.
    connect(camera_ptr, &CameraMDI::sync_in_resumed, this, &PanelRecSettings::handle_sync_in_resume, Qt::QueuedConnection);
    connect(camera_ptr, &CameraMDI::sync_in_timeout, this, &PanelRecSettings::handle_sync_in_timeout, Qt::QueuedConnection);

    // Connect up the group read handler.
    connect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelRecSettings::chk_refresh_regs, Qt::QueuedConnection);
    connect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelRecSettings::chk_broc_regs, Qt::QueuedConnection);

    // Pass it on.
    FMPanelWidget::camera_connect(ip_addr);

    // Enable the panel.
    setEnabled(true);

    // No updates pending just yet.
    update_in_progress = false;

    // Note:  I added a *3 time delay to the UPDATE_DELAY for this particular case
    // because when they are exactly equal the special handling that determines centeredness
    // ends up being overwritten by the camera update.  Sometimes it sticks and sometimes it
    // doesn't.  Delaying just a little bit more allows time for the GIGE_REG_SENSOR_TYPE to be
    // read properly.

    // Update our controls.
    QTimer::singleShot((UPDATE_DELAY * 3), this, &PanelRecSettings::update_controls);
}

/*!
 * @brief Disconnect from our camera.
 */
void PanelRecSettings::camera_disconnect()
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr = nullptr;

    // Have we been here already?
    if (camera_ip_addr == 0)
        return;

    // Clear our IP address.
    camera_ip_addr = 0;

    // Make sure there is a camera pointer as we need to connect to some signals.
    if (p_control_panel)
    {
        camera_ptr = p_control_panel->getCamera();
    }

    // Disconnect any camera-specific signals.
    if (camera_ptr != nullptr)
    {
        disconnect(camera_ptr, &CameraMDI::sync_in_resumed, this, &PanelRecSettings::handle_sync_in_resume);
        disconnect(camera_ptr, &CameraMDI::sync_in_timeout, this, &PanelRecSettings::handle_sync_in_timeout);
        disconnect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelRecSettings::chk_refresh_regs);
        disconnect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelRecSettings::chk_broc_regs);
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // Pass it up.
    FMPanelWidget::camera_disconnect();

    // Get our communications pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr != nullptr)
    {
        // Connect the communications interfaces that we're interested in.
        disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelRecSettings::read_complete);
        disconnect(comm_ptr, &FI_comm::msg_timeout, this, &PanelRecSettings::msg_timeout_event);
        disconnect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelRecSettings::handle_unsol_msg);
        disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelRecSettings::write_complete);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
    }

    // Disable all controls.
    setEnabled(false);

    // Make sure our timer is stopped.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }
}

/*!
 * @brief Pushes current frame rate and exposure from our panel settings
 * to the camera if different from what is already on the camera.
 */
void PanelRecSettings::push_framerate_and_exposure()
{
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_FRAME_RATE:" << read_camera_cache(GIGE_REG_FRAME_RATE) << "GIGE_REG_SHUTTER_SPEED" << read_camera_cache(GIGE_REG_SHUTTER_SPEED);
    framerate_update();
    exposure_update();
}

void PanelRecSettings::enable_ae_controls(bool do_enable)
{
    // Remember if we're allowed to be enabled or not.
    ae_allowed = do_enable;

    // Only allow changes when in the correct camera state.
    if ((camera_state == GIGE_CAMERA_STATE_IDLE) ||
        (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO) ||
        (camera_state == GIGE_CAMERA_STATE_ARMED) ||
        (camera_state == GIGE_CAMERA_STATE_TRIGGERED))
    {
        ui->frm_ae->setEnabled(do_enable);
        if (!do_enable)
        {
            CameraMDI   *camera_ptr;

            // Make sure we have an active camera pointer.
            camera_ptr = p_control_panel->getCamera();
            if (camera_ptr == nullptr)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
                return;
            }

            // Turn it off.
            ui->chk_ae_enable->setChecked(false);
            camera_ptr->ae_reticle_show(false);
            ui->chk_track_ss->setChecked(false);
            camera_ptr->track_exposure(false);
        }
    }
}

/*!
 * @brief Utility function to update the label which shows the frame rate information.
 */
void PanelRecSettings::framerate_label_update()
{
    double  ms;
    QString time_str;

    // Update the framerate text
    ms = 1000.0 / (double)ui->spin_framerate->value();
    time_str = QString(tr("Time: %1 mSec")).arg(ms);
    ui->lbl_frame_time->setText(time_str);
}

/*!
 * @brief Returns the center ROI offset values based on the camera sensor/roi setting.
 *
 * @param offset_x   The calculated center X.
 * @param offsetY   The calculated center Y.
 *
 * @returns True if the center was calculated, false if not.
 */
bool PanelRecSettings::get_center_roi_offsets(int& offset_x, int& offset_y)
{
    bool        calculated_center = false;
    QRect       max_roi;
    CameraMDI   *camera_ptr;

    offset_x = offset_y = 0;    // Initialize to some neutral point in the ROI.

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return calculated_center;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return calculated_center;
    }

    // Get the maximum ROI supported by the selected bitmode.
    max_roi = camera_ptr->get_sensor()->max_roi((uint32_t)ui->cb_bit_depth->currentData().toUInt());

    // Figure out where the center would be in the X and Y direction.
    offset_x = ((max_roi.width() >> 1) - (ui->spin_width->value() >> 1));
    offset_y = ((max_roi.height() >> 1) - (ui->spin_height->value() >> 1));

    // We need these offsets to be on a proper boundary.
    offset_x = (offset_x / camera_ptr->get_sensor()->width_mod()) * camera_ptr->get_sensor()->width_mod();
    offset_y = (offset_y / camera_ptr->get_sensor()->height_mod()) * camera_ptr->get_sensor()->height_mod();

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString("max_roi[%1,%2,%3,%4] offset_x:%5 offset_y:%6").arg(max_roi.x()).arg(max_roi.y()).arg(max_roi.width()).arg(max_roi.height()).arg(offset_x).arg(offset_y);

    // We have calculated the center.
    calculated_center = true;

    return calculated_center;
}

/*!
 * @brief The height value has changed, either as a result of reading the camera's register
 *        or the user specifying a new value.
 *
 * @param new_value         The new height value.
 * @param update_aspect     Flag to indicate if we should update the aspect ratio.
 */
void PanelRecSettings::handle_height_change(int new_value, bool update_aspect)
{
    int offset_max;
    int new_mod_value;
    bool need_rezoom;
    QRect max_roi;
    CameraMDI *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Assume we do not need to "re-zoom".
    need_rezoom = false;

    // Divide by the sensor's modulo for the width.
    new_mod_value = new_value / camera_ptr->get_sensor()->height_mod();

    // Get the maximum ROI supported by the selected bitmode.
    max_roi = camera_ptr->get_sensor()->max_roi((uint32_t)ui->cb_bit_depth->currentData().toUInt());

    // If we're not in auto-ROI mode, maintain aspect ratio.
    if (!ui->chk_autoset_roi->isChecked() && update_aspect)
    {
        // Don't mess around if we're using a custom ratio.
        if (ui->cb_aspect->currentIndex() != ASPECT_CUSTOM)
        {
            int new_width;
            int new_height;

            // Get the width based on the height and aspect ratio.
            new_width = aspect_update(false, new_value, ui->cb_aspect->currentIndex());
            if (new_width != ui->spin_width->value())
            {
                int offset_max;

                // Update the width controls.
                set_slider_value(ui->sld_width, new_width / camera_ptr->get_sensor()->width_mod());
                ui->spin_width->blockSignals(true);
                ui->spin_width->setValue(new_width);
                ui->spin_width->blockSignals(false);

                offset_max = (max_roi.width() - new_width) / camera_ptr->get_sensor()->width_mod();
                ui->sld_offsetx->blockSignals(true);
                ui->sld_offsetx->setMinimum(0);
                ui->sld_offsetx->setMaximum(offset_max);
                ui->sld_offsetx->blockSignals(false);
                ui->spin_offsetx->setMaximum(offset_max * camera_ptr->get_sensor()->width_mod());

                // We need a rezoom.
                need_rezoom  = true;
            }

            // Use the width to calculate what our height "should" be based on the aspect
            // ratio. This might be different due to rounding errors.
            new_height = aspect_update(true, new_width, ui->cb_aspect->currentIndex());
            if (new_height != new_value)
            {
                // Update to our new height.
                new_value = new_height;
                new_mod_value = new_value / camera_ptr->get_sensor()->height_mod();
                ui->spin_height->blockSignals(true);
                ui->spin_height->setValue(new_height);
                ui->spin_height->blockSignals(false);

                // We need a rezoom.
                need_rezoom  = true;
            }
        }
    }

    // Since the height changed, that means our max offsety
    // needs to change. Only need to do it here, because we
    // have setup the slider to be a slave to the spinner.
    offset_max = (max_roi.height() - new_value) / camera_ptr->get_sensor()->height_mod();
    ui->sld_offsety->blockSignals(true);
    ui->sld_offsety->setMinimum(0);
    ui->sld_offsety->setMaximum(offset_max);
    ui->sld_offsety->blockSignals(false);
    ui->spin_offsety->setMaximum(offset_max * camera_ptr->get_sensor()->height_mod());

    // To keep values correct, we make the slider 1/modulo of the value of the spin.
    set_slider_value(ui->sld_height, new_mod_value, true);

    // Do we need to update the offset value(s)?
    update_offsets();

    // Update our framerate information.
    update_framerate_limits();

    // Update our information fields.
    update_info();

    // Let's schedule the camera video player object to adjust it's zoom.
    if (camera_ptr && camera_ptr->get_player() && need_rezoom)
        camera_ptr->get_player()->need_rezoom();
}

/*!
 * @brief Process register values we received from the camera.
 *
 * @param reg_addr      The register's address.
 * @param reg_value     The register's value.
 */
void PanelRecSettings::handle_read(uint32_t reg_addr, uint32_t reg_value)
{
    CameraMDI   *camera_ptr;
    QString     who_am_i = QString("%1 [0x%2][0x%3]").arg(__PRETTY_FUNCTION__).arg(reg_addr, 0, 16).arg(reg_value, 0, 16);

    // Don't update our settings if we're in the middle of updating a value.
    if (update_in_progress || update_pending)
        return;

    // If the user is moving a slider around, we can ignore the reads.
    if (slider_changing)
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_MIN_WIDTH:
        // Update our minimum allowed values.
        if (ui->sld_width->minimum() != (int)(reg_value / camera_ptr->get_sensor()->width_mod()))
            ui->sld_width->setMinimum(reg_value / camera_ptr->get_sensor()->width_mod());
        if (ui->spin_width->minimum() != (int)reg_value)
            ui->spin_width->setMinimum((int)reg_value);
        break;
    case GIGE_REG_SENSOR_WIDTH:
        // Update our maximum allowed values.
        if (ui->sld_width->maximum() != (int)(reg_value / camera_ptr->get_sensor()->width_mod()))
            ui->sld_width->setMaximum(reg_value / camera_ptr->get_sensor()->width_mod());
        if (ui->spin_width->maximum() != (int)reg_value)
            ui->spin_width->setMaximum((int)reg_value);
        break;
    case GIGE_REG_WIDTH:
        // Make sure the user isn't sitting on the controls.
        if (!roi_control_has_focus())
        {
            // Update our control value(s).
            if (ui->spin_width->value() != (int)reg_value)
            {
                // Update our control value(s).
                ui->spin_width->blockSignals(true);
                ui->spin_width->setValue((int)reg_value);
                ui->spin_width->blockSignals(false);

                // Make sure we update the aspect ratio too. We don't want the change to cause a
                // recalculation though.
                ui->cb_aspect->blockSignals(true);
                set_aspect_ratio();
                ui->cb_aspect->blockSignals(false);

                // By resetting the center checkbox if already checked the
                // function will modify the offsets to be centered to adjust
                // to the width modification (if any).
                if (ui->chk_center_roi->isChecked())
                {
                    set_center_roi(true, who_am_i);
                }
                update_offsets();
            }

            // Make sure all the controls are consistent.
            handle_width_change((int)reg_value, false);
        }
        break;

    case GIGE_REG_MIN_HEIGHT:
        // Update our minimum allowed values.
        if (ui->sld_height->minimum() != (int)(reg_value / camera_ptr->get_sensor()->height_mod()))
            ui->sld_height->setMinimum(reg_value / camera_ptr->get_sensor()->height_mod());
        if (ui->spin_height->minimum() != (int)reg_value)
            ui->spin_height->setMinimum((int)reg_value);
        break;
    case GIGE_REG_SENSOR_HEIGHT:
        // Update our maximum allowed values.
        if (ui->sld_height->maximum() != (int)(reg_value / camera_ptr->get_sensor()->height_mod()))
            ui->sld_height->setMaximum(reg_value / camera_ptr->get_sensor()->height_mod());
        if (ui->spin_height->maximum() != (int)reg_value)
            ui->spin_height->setMaximum((int)reg_value);
        break;
    case GIGE_REG_HEIGHT:
        // Make sure the user isn't sitting on the controls.
        if (!roi_control_has_focus())
        {
            if (ui->spin_height->value() != (int)reg_value)
            {
                // Update our control value(s).
                ui->spin_height->blockSignals(true);
                ui->spin_height->setValue((int)reg_value);
                ui->spin_height->blockSignals(false);

                // Make sure we update the aspect ratio too.
                ui->cb_aspect->blockSignals(true);
                set_aspect_ratio();
                ui->cb_aspect->blockSignals(false);

                // By resetting the center checkbox if already checked the
                // function will modify the offsets to be centered to adjust
                // to the width modification (if any).
                if (ui->chk_center_roi->isChecked())
                {
                    set_center_roi(true, who_am_i);
                }
                update_offsets();
            }

            // Make sure all the controls are consistent.
            handle_height_change((int)reg_value, false);
        }
        break;

    case GIGE_REG_OFFSET_X:
        // Make sure the user isn't sitting on the controls.
        if (!roi_control_has_focus())
        {
            // Update our control value(s).
            if (ui->spin_offsetx->value() != (int)reg_value)
            {
                ui->spin_offsetx->setValue((int)reg_value);
                spin_offsetx_changed((int)reg_value);

                // By resetting the center checkbox if already checked the
                // function will modify the offsets to be centered to adjust
                // to the width modification (if any).
                if (ui->chk_center_roi->isChecked())
                {
                    set_center_roi(true, who_am_i);
                }
                update_offsets();
            }
        }
        break;
    case GIGE_REG_OFFSET_Y:
        // Make sure the user isn't sitting on the controls.
        if (!roi_control_has_focus())
        {
            // Update our control value(s).
            if (ui->spin_offsety->value() != (int)reg_value)
            {
                ui->spin_offsety->setValue((int)reg_value);
                spin_offsety_changed((int)reg_value);

                // By resetting the center checkbox if already checked the
                // function will modify the offsets to be centered to adjust
                // to the width modification (if any).
                if (ui->chk_center_roi->isChecked())
                {
                    set_center_roi(true, who_am_i);
                }
                update_offsets();
            }
        }
        break;

    case GIGE_REG_MAX_FRAME_RATE:
        // Update our maximum allowed values.
        if (ui->sld_framerate->maximum() > (int)reg_value)
        {
            ui->sld_framerate->blockSignals(true);
            ui->sld_framerate->setMinimum(LOW_LIGHT_FRAMERATE_MIN);
            ui->sld_framerate->setMaximum((int)reg_value);
            ui->sld_framerate->blockSignals(false);
            ui->spin_framerate->setMinimum(LOW_LIGHT_FRAMERATE_MIN);
            ui->spin_framerate->setMaximum((int)reg_value);
        }
        break;
    case GIGE_REG_FRAME_RATE:
        // Make sure the user isn't sitting on the controls.
        if (!ui->sld_framerate->hasFocus() && !ui->spin_framerate->hasFocus())
        {
            int frame_rate = (int)ClientMath::reg_to_frame_rate(reg_value);

            // Update our controls.
            if (ui->sld_framerate->value() != frame_rate)
            {
                set_slider_value(ui->sld_framerate, frame_rate);
                ui->spin_framerate->setValue(frame_rate);
                framerate_label_update();
            }
        }
        break;

    case GIGE_REG_MIN_SHUTTER_SPEED:
        // Update our minimum allowed values.
        if (ui->sld_ss->minimum() != (int)reg_value)
        {
            ui->sld_ss->setMinimum((int)reg_value);
            ui->spin_ss->setMinimum((int)reg_value);
        }
        break;
    case GIGE_REG_MAX_SHUTTER_SPEED:
        // Update our maximum allowed values.
        if (ui->sld_ss->maximum() != (int)reg_value)
        {
            ui->sld_ss->setMaximum((int)reg_value);
            ui->spin_ss->setMaximum((int)reg_value);
        }
        break;
    case GIGE_REG_SHUTTER_SPEED:
        // Make sure the user isn't sitting on the controls.
        if (!ui->sld_ss->hasFocus() && !ui->spin_ss->hasFocus())
        {
            if (ui->sld_ss->value() != (int)reg_value)
            {
                ui->sld_ss->setValue((int)reg_value);
                slide_exposure_changed((int)reg_value);
            }
        }
        break;

    case GIGE_REG_LOW_LIGHT_ENABLE:
        // Make sure the user isn't sitting on the controls.
        if (!ui->sld_ss_ll->hasFocus() && !ui->spin_ss_ll->hasFocus() && !ui->chk_ll->hasFocus())
        {
            uint32_t    camera_state;

            // What is our state.
            camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

            // Only update if we're in the correct state.
            if ((camera_state == GIGE_CAMERA_STATE_IDLE) || (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO))
            {
                // Update the controls based on the value we got.
                ui->chk_ll->setChecked(reg_value);

                // Enable/disable the low light exposure controls.
                ui->sld_ss_ll->setEnabled(reg_value);
                ui->spin_ss_ll->setEnabled(reg_value);
            }
        }
        break;
    case GIGE_REG_LOW_LIGHT_SHUTTER:
        // Make sure the user isn't sitting on the controls.
        if (!ui->sld_ss_ll->hasFocus() && !ui->spin_ss_ll->hasFocus() && !ui->chk_ll->hasFocus())
        {
            // Update our shutter controls.
            if (ui->sld_ss_ll->value() != (int)reg_value)
            {
                ui->sld_ss_ll->setValue((int)reg_value);
                ui->spin_ss_ll->setValue((int)reg_value);
            }
        }
        break;

    case GIGE_REG_BIT_MODE:
        // Make sure the user isn't sitting on the controls.
        if (!ui->cb_bit_depth->view()->isVisible())
        {
            // Update our dipslay, but don't allow that update to send a signal.
            // If we send the signal, we think the user changed the setting and we
            // write it back out.
            if (ui->cb_bit_depth->currentData().toUInt() != reg_value)
            {
                int index;

                index = ui->cb_bit_depth->findData(reg_value);
                if (index < 0)
                {
                    switch (reg_value)
                    {
                    case GIGE_BITMODE_7_0:
                        index = ui->cb_bit_depth->findData(GIGE_BITMODE_LOWER_8_BITS);
                        break;
                    case GIGE_BITMODE_8_1:
                        index = ui->cb_bit_depth->findData(GIGE_BITMODE_MIDDLE_8_BITS);
                        break;
                    case GIGE_BITMODE_9_2:
                        index = ui->cb_bit_depth->findData(GIGE_BITMODE_UPPER_8_BITS);
                        break;
                    case GIGE_BITMODE_9_0:
                        index = ui->cb_bit_depth->findData(GIGE_BITMODE_10_BITS);
                        break;
                    default:
                        break;
                    }
                }

                if (index >= 0)
                {
                    ui->cb_bit_depth->blockSignals(true);
                    ui->cb_bit_depth->setCurrentIndex(index);
                    ui->cb_bit_depth->blockSignals(false);
                }
            }
        }
        break;

    case GIGE_REG_FPN:
        // Make sure the user isn't sitting on the controls.
        if (!ui->cb_fpn->view()->isVisible())
        {
            ui->cb_fpn->blockSignals(true);
            if ((reg_value & GIGE_FPN_SENSOR_ENABLE_MASK) == GIGE_FPN_SENSOR_ENABLE_MASK)
            {
                ui->cb_fpn->setCurrentIndex(fpn_sensor_index);
            }
            else if ((reg_value & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK)
            {
                ui->cb_fpn->setCurrentIndex(fpn_pixel_index);
            }
            else if ((reg_value & GIGE_FPN_FPGA_UNCALIBRATED) == GIGE_FPN_FPGA_UNCALIBRATED)
            {
                ui->cb_fpn->setCurrentIndex(fpn_pixel_index);
            }
            else
            {
                ui->cb_fpn->setCurrentIndex(fpn_disabled_index);
            }
            ui->cb_fpn->blockSignals(false);

            // Update the display.
            update_fpn();
        }
        break;

    case GIGE_REG_SENSOR_OPTIONS:
        // Make sure the user isn't sitting on the controls.
        if (!ui->cb_bin_sample->hasFocus())
        {
            // Update our display, but don't allow that update to send a signal.
            // If we send the signal, we think the user changed the setting and we
            // write it back out.
            if (ui->cb_bin_sample->currentData().toUInt() != (reg_value & (GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK)))
            {
                ui->cb_bin_sample->blockSignals(true);
                ui->cb_bin_sample->setCurrentIndex(ui->cb_bin_sample->findData(reg_value & (GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK)));
                ui->cb_bin_sample->blockSignals(false);
            }
        }

        // Update our sensor object.
        if (p_control_panel && p_control_panel->getCamera() && p_control_panel->getCamera()->get_sensor())
            p_control_panel->getCamera()->get_sensor()->set_options(reg_value);
        break;

    case GIGE_REG_CAMERA_STATE:
        process_camera_state(reg_value);
        break;

    case GIGE_REG_LTR_MODE:
        // Update the ltr_enabled bit.
        ltr_enabled = (reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;

        // Update the camera as some of the limits will be affected by this.
        update_framerate_limits();
        break;

    case GIGE_REG_SENSOR_TYPE:
        // The sensor type register has been read from the camera, so we can be
        // confident that there will be a sensor object setup in a little while.
        QTimer::singleShot(UPDATE_DELAY, this, &PanelRecSettings::update_controls);
        break;

    default:
        break;
    }

    // Are we done with reading our register space?
    if (reg_addr == last_reg_addr)
    {
        // Update the panel display(s).
        update_info();

        // Reset our last register address value.
        last_reg_addr = FI_COMM::NO_REGISTER;
    }
}

/*!
 * @brief The width value has changed, either as a result of reading the camera's register
 *        or the user specifying a new value.
 *
 * @param new_value         The new width value.
 * @param update_aspect     Flag to indicate if we should update the aspect ratio.
 */
void PanelRecSettings::handle_width_change(int new_value, bool update_aspect)
{
    int offset_max;
    int new_mod_value;
    bool need_rezoom;
    QRect max_roi;
    CameraMDI *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // If aspect ratio set to 1:1 then check for ROI width exceeding the
    // maximum height for a sensor.
    //
    // Note:  An assumption is being made that sensor ROIs are always larger
    // in width than in height.  Therefore this same test will not be added
    // to the PanelRecSettings::handle_height_change method.
    // Ticket #304 - Aspect 1:1 widths allowed to exceed maximum height.
    if (ui->cb_aspect->currentIndex() == ASPECT_1x1)
    {
        // What is the maximum height value?
        int max_height = ui->sld_height->maximum() * camera_ptr->get_sensor()->height_mod();

        // We may have to limit the width value.
        if (new_value > max_height)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FORCED 1:1" << "width_value:" << new_value << "reducing to max_height:" << max_height;

            // Limit the "new_value" width to the maximum height setting.
            new_value = max_height;
            set_slider_value(ui->sld_width, new_value / camera_ptr->get_sensor()->width_mod());
            ui->spin_width->blockSignals(true);
            ui->spin_width->setValue(new_value);
            ui->spin_width->blockSignals(false);
        }
    }

    // Assume we do not need to "re-zoom".
    need_rezoom = false;

    // Divide by the sensor's modulo for the width.
    new_mod_value = new_value / camera_ptr->get_sensor()->width_mod();

    // Get the maximum ROI supported by the selected bitmode.
    max_roi = camera_ptr->get_sensor()->max_roi((uint32_t)ui->cb_bit_depth->currentData().toUInt());

    // Since the width changed, that means our max offsetx
    // needs to change. Only need to do it here, because we
    // have setup the slider to be a slave to the spinner.
    offset_max = (max_roi.width() - new_value) / camera_ptr->get_sensor()->width_mod();
    ui->sld_offsetx->setMinimum(0);
    ui->sld_offsetx->setMaximum(offset_max);
    ui->spin_offsetx->setMaximum(offset_max * camera_ptr->get_sensor()->width_mod());

    // To keep values correct, we make the slider 1/modulo of the value of the spin.
    if (ui->sld_width->value() != new_mod_value)
        set_slider_value(ui->sld_width, new_mod_value, true);

    // If we're not in auto-ROI mode, maintain aspect ratio.
    if (!ui->chk_autoset_roi->isChecked() && update_aspect)
    {
        // Don't mess around if we're using a custom ratio.
        if (ui->cb_aspect->currentIndex() != ASPECT_CUSTOM)
        {
            int new_height;

            // Get the height based on the width and aspect ratio.
            new_height = aspect_update(true, new_value, ui->cb_aspect->currentIndex());

            if (new_height != ui->spin_height->value())
            {
                // Update the height controls.
                set_slider_value(ui->sld_height, new_height / camera_ptr->get_sensor()->height_mod());
                ui->spin_height->blockSignals(true);
                ui->spin_height->setValue(new_height);
                ui->spin_height->blockSignals(false);

                offset_max = (max_roi.height() - new_height) / camera_ptr->get_sensor()->height_mod();
                ui->sld_offsety->setMinimum(0);
                ui->sld_offsety->setMaximum(offset_max);
                ui->spin_offsety->setMaximum(offset_max * camera_ptr->get_sensor()->height_mod());

                // We need a rezoom.
                need_rezoom  = true;
            }
        }
    }

    // Do we need to update the offset value(s)?
    update_offsets();

    // Update our framerate information.
    update_framerate_limits();

    // Update our information fields.
    update_info();

    // Let's schedule the camera video player object to adjust it's zoom.
    if (camera_ptr && camera_ptr->get_player() && need_rezoom)
        camera_ptr->get_player()->need_rezoom();
}

/*!
 * @brief Initialize the signal/slot connections and initializations for the controls.
 */
void PanelRecSettings::init_controls()
{
    connect(ui->cb_aspect, qOverload<int>(&QComboBox::currentIndexChanged), this, &PanelRecSettings::aspect_changed);
    connect(ui->cb_bin_sample, qOverload<int>(&QComboBox::currentIndexChanged), this, &PanelRecSettings::bin_sample_changed);
    connect(ui->cb_bit_depth, qOverload<int>(&QComboBox::currentIndexChanged), this, &PanelRecSettings::bitdepth_changed);
    connect(ui->cb_fpn, qOverload<int>(&QComboBox::currentIndexChanged), this, &PanelRecSettings::fpn_changed);

    connect(ui->spin_height, &QSpinBox::editingFinished, this, &PanelRecSettings::height_update);
    connect(ui->sld_height, &QSlider::sliderPressed, this, &PanelRecSettings::slider_lock);
    connect(ui->sld_height, &QSlider::sliderReleased, this, &PanelRecSettings::slider_unlock);
    connect(ui->sld_height, &QSlider::sliderReleased, this, &PanelRecSettings::height_update);
    connect(ui->sld_height, &QSlider::valueChanged, this, &PanelRecSettings::height_changed);

    connect(ui->spin_width, &QSpinBox::editingFinished, this, &PanelRecSettings::width_update);
    connect(ui->sld_width, &QSlider::sliderPressed, this, &PanelRecSettings::slider_lock);
    connect(ui->sld_width, &QSlider::sliderReleased, this, &PanelRecSettings::slider_unlock);
    connect(ui->sld_width, &QSlider::sliderReleased, this, &PanelRecSettings::width_update);
    connect(ui->sld_width, &QSlider::valueChanged, this, &PanelRecSettings::width_changed);

    connect(ui->spin_offsetx, &QSpinBox::editingFinished, this, &PanelRecSettings::offsetx_update);
    connect(ui->sld_offsetx, &QSlider::sliderPressed, this, &PanelRecSettings::slider_lock);
    connect(ui->sld_offsetx, &QSlider::sliderReleased, this, &PanelRecSettings::slider_unlock);
    connect(ui->sld_offsetx, &QSlider::sliderReleased, this, &PanelRecSettings::offsetx_update);
    connect(ui->sld_offsetx, &QSlider::valueChanged, this, &PanelRecSettings::slide_offsetx_changed);

    connect(ui->spin_offsety, &QSpinBox::editingFinished, this, &PanelRecSettings::offsety_update);
    connect(ui->sld_offsety, &QSlider::sliderPressed, this, &PanelRecSettings::slider_lock);
    connect(ui->sld_offsety, &QSlider::sliderReleased, this, &PanelRecSettings::slider_unlock);
    connect(ui->sld_offsety, &QSlider::sliderReleased, this, &PanelRecSettings::offsety_update);
    connect(ui->sld_offsety, &QSlider::valueChanged, this, &PanelRecSettings::slide_offsety_changed);

    connect(ui->spin_framerate, &QSpinBox::editingFinished, this, &PanelRecSettings::framerate_update);
    connect(ui->sld_framerate, &QSlider::sliderPressed, this, &PanelRecSettings::slider_lock);
    connect(ui->sld_framerate, &QSlider::sliderReleased, this, &PanelRecSettings::slider_unlock);
    connect(ui->sld_framerate, &QSlider::sliderReleased, this, &PanelRecSettings::framerate_update);
    connect(ui->sld_framerate, &QSlider::valueChanged, this, &PanelRecSettings::slide_framerate_changed);

    connect(ui->spin_ss, &QSpinBox::editingFinished, this, &PanelRecSettings::exposure_update);
    connect(ui->sld_ss, &QSlider::sliderPressed, this, &PanelRecSettings::slider_lock);
    connect(ui->sld_ss, &QSlider::sliderReleased, this, &PanelRecSettings::slider_unlock);
    connect(ui->sld_ss, &QSlider::sliderReleased, this, &PanelRecSettings::exposure_update);
    connect(ui->sld_ss, &QSlider::valueChanged, this, &PanelRecSettings::slide_exposure_changed);

    connect(ui->spin_ss_ll, &QSpinBox::editingFinished, this, &PanelRecSettings::lowlight_update);
    connect(ui->sld_ss_ll, &QSlider::sliderPressed, this, &PanelRecSettings::slider_lock);
    connect(ui->sld_ss_ll, &QSlider::sliderReleased, this, &PanelRecSettings::slider_unlock);
    connect(ui->sld_ss_ll, &QSlider::sliderReleased, this, &PanelRecSettings::lowlight_update);
    connect(ui->sld_ss_ll, &QSlider::valueChanged, this, &PanelRecSettings::slide_lowlight_changed);

    connect(ui->chk_center_roi, &QCheckBox::clicked, this, &PanelRecSettings::center_clicked);
    connect(ui->chk_autoset_roi, &QCheckBox::clicked, this, &PanelRecSettings::roi_autoset_clicked);
    connect(ui->chk_autoset_rate, &QCheckBox::clicked, this, &PanelRecSettings::rate_autoset_clicked);


    connect(ui->chk_ae_enable, &QCheckBox::clicked, this, &PanelRecSettings::ae_enable);
    connect(ui->chk_track_ss, &QCheckBox::clicked, this, &PanelRecSettings::track_exposure);
    connect(ui->pb_ae_reset, &QPushButton::clicked, this, &PanelRecSettings::ae_reset_region);

    // Set up the ranges for the lowlight spin and slider controls.
    // We don't know the frame_overhead at this point so the default 0 is ok.
    calc_lowlight_range(0);

    connect(ui->chk_ll, &QCheckBox::clicked, this, &PanelRecSettings::lowlight_clicked);
    ui->chk_ll->setChecked(false);
    ui->sld_ss_ll->setEnabled(false);
    ui->spin_ss_ll->setEnabled(false);
}

/*!
 * @brief Check to see if our X and Y offsets are centered.
 * @return True if we are centered, false otherwise.
 */
bool PanelRecSettings::is_offset_centered()
{
    int center_x;
    int center_y;
    int offset_x;
    int offset_y;
    int cam_width;
    int cam_height;
    int cam_max_width;
    int cam_max_height;
    bool    rc;
    QRect       max_roi;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return false;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return false;
    }

    // Get the maximum ROI supported by the selected bitmode.
    max_roi = camera_ptr->get_sensor()->max_roi((uint32_t)ui->cb_bit_depth->currentData().toUInt());

    // Get the width/height value for the camera.
    cam_width = (int)read_camera_cache(GIGE_REG_WIDTH);
    cam_max_width = max_roi.width();
    cam_height = (int)read_camera_cache(GIGE_REG_HEIGHT);
    cam_max_height = max_roi.height();
    offset_x = (int)read_camera_cache(GIGE_REG_OFFSET_X);
    offset_y = (int)read_camera_cache(GIGE_REG_OFFSET_Y);

    // Figure out where the center offset would be in the X and Y direction.
    center_x = ((cam_max_width >> 1) - (cam_width >> 1)) >> 1;
    center_x = (center_x / camera_ptr->get_sensor()->width_mod()) * camera_ptr->get_sensor()->width_mod();
    center_y = ((cam_max_height >> 1) - (cam_height >> 1)) >> 1;
    center_y = (center_y / camera_ptr->get_sensor()->height_mod()) * camera_ptr->get_sensor()->height_mod();

    // Check to see if the offset is for the center position.
    if ((offset_x == center_x) && (offset_y == center_y))
        rc = true;
    else
        rc = false;

    // Return our findings.
    return rc;
}

/*!
 * @brief Enable or disable controls on the panel based on the camera's state.
 *
 * @param new_state     The new camera state.
 */
void PanelRecSettings::process_camera_state(uint32_t new_state)
{
    uint32_t    lr_type;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Enable the controls if we do not have autosave enabled (except in FFLR mode) and we are in IDLE or LIVE view.
    if ((!camera_ptr->is_autosave_enabled() || camera_ptr->is_fflr_enabled(lr_type)) && ((new_state == GIGE_CAMERA_STATE_IDLE) || (new_state == GIGE_CAMERA_STATE_LIVE_VIDEO)))
    {
        // All controls are enabled in these states.
        ui->frm_advanced_settings->setEnabled(true);
        ui->frm_frame_rate->setEnabled(true);
        ui->frm_partition->setEnabled(true);
        ui->frm_ROI->setEnabled(true);

        ui->lbl_ss->setEnabled(true);
        ui->spin_ss->setEnabled(true);
        ui->sld_ss->setEnabled(true);
        ui->spin_degrees->setEnabled(true);

        ui->chk_ll->setEnabled(true);
        ui->sld_ss_ll->setEnabled(ui->chk_ll->isChecked());
        ui->spin_ss_ll->setEnabled(ui->chk_ll->isChecked());

        ui->chk_ae_enable->setEnabled(ae_allowed);
        ui->chk_track_ss->setEnabled(ae_allowed);
    }
    else
    {
        // Most of the controls are disabled in these states.
        ui->frm_advanced_settings->setEnabled(false);
        ui->frm_frame_rate->setEnabled(false);
        ui->frm_partition->setEnabled(false);
        ui->frm_ROI->setEnabled(false);
        ui->chk_ll->setEnabled(false);
        ui->chk_ll->setChecked(false);
        ui->sld_ss_ll->setEnabled(false);
        ui->spin_ss_ll->setEnabled(false);

        // Special handling for the exposure controls.
        if ((new_state == GIGE_CAMERA_STATE_ARMED) || (new_state == GIGE_CAMERA_STATE_ARMED) || (new_state == GIGE_CAMERA_STATE_AUTOSAVE))
        {
            // The exposure controls are enabled in these states too.
            ui->lbl_ss->setEnabled(true);
            ui->spin_ss->setEnabled(true);
            ui->sld_ss->setEnabled(true);
            ui->spin_degrees->setEnabled(true);

            ui->pb_ae_reset->setEnabled(ae_allowed);
            ui->chk_ae_enable->setEnabled(ae_allowed);
            ui->chk_track_ss->setEnabled(ae_allowed);
        }
        else
        {
            // The exposure controls are disabled like everyone else here.
            ui->lbl_ss->setEnabled(false);
            ui->spin_ss->setEnabled(false);
            ui->sld_ss->setEnabled(false);
            ui->spin_degrees->setEnabled(false);

            ui->pb_ae_reset->setEnabled(false);
            ui->chk_ae_enable->setEnabled(false);
            ui->chk_track_ss->setEnabled(false);
        }
    }
}

/*!
 * @brief Utility function to read all registers this panel is interested in.
 */
void PanelRecSettings::read_registers()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Prepare the camera register group read for our calculations.
    cam_register_group->addReg(GIGE_REG_MIN_SHUTTER_SPEED);
    cam_register_group->addReg(GIGE_REG_MAX_SHUTTER_SPEED);

    // Now get the actual settings.
    cam_register_group->addReg(GIGE_REG_WIDTH);
    cam_register_group->addReg(GIGE_REG_HEIGHT);
    cam_register_group->addReg(GIGE_REG_OFFSET_X);
    cam_register_group->addReg(GIGE_REG_OFFSET_Y);
    cam_register_group->addReg(GIGE_REG_FRAME_RATE);
    cam_register_group->addReg(GIGE_REG_SHUTTER_SPEED);
    cam_register_group->addReg(GIGE_REG_LOW_LIGHT_ENABLE);
    cam_register_group->addReg(GIGE_REG_LOW_LIGHT_SHUTTER);

    // We're also interested in the camera state.
    cam_register_group->addReg(GIGE_REG_CAMERA_STATE);

    // We need these for calculations.
    cam_register_group->addReg(GIGE_REG_BIT_MODE);
    if (camera_ptr->get_sensor()->opts_avail())
        cam_register_group->addReg(GIGE_REG_SENSOR_OPTIONS);
    cam_register_group->addReg(GIGE_REG_FRAME_OVERHEAD);
    cam_register_group->addReg(GIGE_REG_HDMI);
    cam_register_group->addReg(GIGE_REG_LTR_SECTOR_COUNT);
    cam_register_group->addReg(GIGE_REG_PART_STATE);
    cam_register_group->addReg(GIGE_REG_PART_NUM_FRAMES);
    cam_register_group->addReg(GIGE_REG_SSD_BANDWIDTH);
    cam_register_group->addReg(GIGE_REG_PART_SIZE);
    cam_register_group->addReg(GIGE_REG_PART_SIZE_RESERVED);
    cam_register_group->addReg(GIGE_REG_BROC_LEN);

    // Make sure we read LTR mode last, 'cuz it expects some of the other
    // values to be in the cache.
    cam_register_group->lastReg(GIGE_REG_LTR_MODE);

    // Pass on the request but only increment our counter if we sent it to the camera.
    read_camera_registers(cam_register_group);
}

/*!
 * @brief Return true if any of the ROI controls has the focus.
 */
bool PanelRecSettings::roi_control_has_focus()
{
    return (ui->spin_width->hasFocus() || ui->spin_height->hasFocus() || ui->spin_offsetx->hasFocus() || ui->spin_offsety->hasFocus() ||
            ui->sld_width->hasFocus() || ui->sld_height->hasFocus() || ui->sld_offsetx->hasFocus() || ui->sld_offsety->hasFocus() ||
            ui->cb_aspect->hasFocus() || ui->chk_center_roi->hasFocus());
}

/*!
 * @brief Determine the aspect ratio based on the width and height values.
 */
void PanelRecSettings::set_aspect_ratio()
{
    int new_index;
    int cam_width;
    int cam_height;

    // Don't go any further if we're disabled.
    if (!ui->frm_ROI->isEnabled())
        return;

    // Get the width/height value for the camera.
    cam_width = (int)read_camera_cache(GIGE_REG_WIDTH);
    cam_height = (int)read_camera_cache(GIGE_REG_HEIGHT);

    // Ignore if we don't have a valid width/height value.
    if ((cam_width == 0) || (cam_height == 0))
        return;

    // Walk through all the aspect ratios and see if we can make a match.
    new_index = calc_aspect_ratio(cam_width, cam_height);
    if (new_index != ui->cb_aspect->currentIndex())
    {
        // Set the new aspect ratio.
        ui->cb_aspect->setCurrentIndex(new_index);
    }

    // Update our maximum ROI value.
    update_min_max_roi(new_index);
}

/*!
 * @brief Make a common place where to set the center checkbox to allow for
 * tracking of who is calling and when in order to isolate problems with the
 * handling of this checkbox.
 *
 * @param check_it      True if the center checkbox is to be checked, false if not.
 * @param who_called_me Who called this function - useful for log messages to track usage.
 */
void PanelRecSettings::set_center_roi(bool check_it, QString who_called_me)
{
    bool modify_it = (ui->chk_center_roi->isChecked() != check_it); // Does it need to be modified?
    bool show_trace = false;    // Set this true if you need to start tracing this information.
    if (show_trace)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << (modify_it ? (check_it ? "MODIFYING->CHECKED" : "MODIFYNG->UNCHECKED") : (ui->chk_center_roi->isChecked() ? "STAYING CHECKED" : "STAYING UNCHECKED")) << "called by:" << who_called_me;
    }

    if (modify_it)
    {
        ui->chk_center_roi->setChecked(check_it);
    }
}

void PanelRecSettings::start_update_write(bool roi_change)
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Did the ROI change?
    if (roi_change)
    {
        uint32_t    reg_value;

        // If the ROI changed, then we'll need to update the HDMI to account for
        // the new settings.
        reg_value = read_camera_cache(GIGE_REG_HDMI);

        // Check to see if HDMI is enabled.
        if ((reg_value & GIGE_HDMI_ENABLE_MASK) == GIGE_HDMI_ENABLE)
        {
            // If we have an ROI with a dimension larger than 1080P, disable HDMI.
            if ((ui->spin_height->value() > 1280) || (ui->spin_width->value() > 1920))
            {
                // Turn off the HDMI.
                write_list.append(REG_VALUE_PAIR(GIGE_REG_HDMI, 0));

                // Let the user know what we've done.
                MainWindow::GetInstance()->set_statusbar_text(QStringLiteral("ROI exceeds 1080P, HDMI disabled!"));
            }
            else
            {
                // If it was enabled, by re-enabling it we will force the HDMI logic to recalculate the scaling
                // required for the new ROI.
                write_list.append(REG_VALUE_PAIR(GIGE_REG_HDMI, reg_value));

                // Clear the status message.
                MainWindow::GetInstance()->set_statusbar_text(QString());
            }
        }
        else
        {
            // Clear the status message.
            MainWindow::GetInstance()->set_statusbar_text(QString());
        }

        // Make sure we pause the histogram.
        MainWindow::GetInstance()->pause_histogram();
    }

    //==============================================================================
    // TODO TODO TODO TODO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "============== Need to worry about SyncOut. =================";

    // 11/29/22 An Update Here. Ticket #455 yielded a change in the handling of Sync-Out within the
    // Record Control Panel. It will catch a change in a frame rate and will rewrite the sync out
    // register so that it can be handled if duty cycle is active. I am leaving this all in place
    // until we can verify that duty cycle produces a truly modified signal wave with an oscilloscope.

    // - If sync_out is enabled, we may need to update that.
    // TODO TODO TODO TODO
    //==============================================================================

    // We need to wait for the video to pause before we can start writing our registers.
    wait_for_video_pause = true;

    // If we're not streaming at all, or the video stream has already been paused, then
    // just start writing the registers. Otherwise we need to pause the video first.
    if (!camera_ptr->get_camera_info()->stream || video_paused)
        stream_paused();
    else
        camera_ptr->pause_video_request();
}

/*!
 * @brief Translate the UI elements for this object.
 */
void PanelRecSettings::translate_ui()
{
    int i;
    int aspect_index;
    int bin_sample_index;
    int bitdepth_index;
    int fpn_index;
    QList<QWidget *> widgetList = findChildren<QWidget *>();

    // Block all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(true);

    // Remember the current settings.
    aspect_index = ui->cb_aspect->currentIndex();
    bin_sample_index = ui->cb_bin_sample->currentIndex();
    bitdepth_index = ui->cb_bit_depth->currentIndex();
    fpn_index = ui->cb_fpn->currentIndex();

    // Translate our user interface.
    ui->retranslateUi(this);

    // Restore the settings.
    ui->cb_aspect->setCurrentIndex(aspect_index);
    ui->cb_bin_sample->setCurrentIndex(bin_sample_index);
    ui->cb_bit_depth->setCurrentIndex(bitdepth_index);
    ui->cb_fpn->setCurrentIndex(fpn_index);

    // Unblock all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(false);
}

/*!
 * @brief Update the bit-depth combobox based on what the sensor supports.
 */
void PanelRecSettings::update_bin_sample()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Do we have a sensor pointer?
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No sensor pointer!";
        return;
    }

    // Do we support binning & sampling?
    if (camera_ptr->get_sensor()->opts_avail())
    {
        // Reset and rebuild the bit-depth combobox based on the sensor information.
        ui->cb_bin_sample->blockSignals(true);
        ui->cb_bin_sample->clear();
        ui->cb_bin_sample->addItem(tr("Normal"), 0x101);
        ui->cb_bin_sample->addItem(tr("2X Binning"), 0x102);
        ui->cb_bin_sample->addItem(tr("4X Binning"), 0x104);
        ui->cb_bin_sample->addItem(tr("2X Subsampling"), 0x201);
        ui->cb_bin_sample->addItem(tr("4X Subsampling"), 0x401);
        ui->cb_bin_sample->addItem(tr("2X Bin & 2X Sub"), 0x202);

        // Get the current bitmode setting.
        uint32_t reg_value = read_camera_cache(GIGE_REG_SENSOR_OPTIONS);

        // Tell the sensor object what options are in use.
        camera_ptr->get_sensor()->set_options(reg_value);

        // Restore the index.
        ui->cb_bin_sample->setCurrentIndex(ui->cb_bin_sample->findData(reg_value & (GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK)));

        // Restore signals.
        ui->cb_bin_sample->blockSignals(false);

        // Make sure the control is visible..
        ui->lbl_bin_sample->setVisible(true);
        ui->cb_bin_sample->setVisible(true);
    }
    else
    {
        // Hide the control.
        ui->lbl_bin_sample->setVisible(false);
        ui->cb_bin_sample->setVisible(false);
    }
}

/*!
 * @brief Update the bit-depth combobox based on what the sensor supports.
 */
void PanelRecSettings::update_bitdepth()
{
    bool        do_enable;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Do we have a sensor pointer?
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No sensor pointer!";
        ui->cb_bit_depth->setEnabled(false);
        return;
    }

    // Reset and rebuild the bit-depth combobox based on the sensor information.
    ui->cb_bit_depth->blockSignals(true);
    ui->cb_bit_depth->clear();

    // Figure out what the pixel width is.
    switch (camera_ptr->get_sensor()->max_pixel_depth())
    {
    case 12:
        // Initialize the combobox.
        ui->cb_bit_depth->addItem(tr("12 Bits [12:1]"), GIGE_BITMODE_11_0);
        ui->cb_bit_depth->addItem(tr("10 Bits [12:3]"), GIGE_BITMODE_11_2);
        ui->cb_bit_depth->addItem(tr("10 Bits"), GIGE_BITMODE_11_2);
        ui->cb_bit_depth->addItem(tr("8 Bits [12:5]"), GIGE_BITMODE_11_4);
        ui->cb_bit_depth->addItem(tr("Upper 8 Bits"), GIGE_BITMODE_11_4);
        ui->cb_bit_depth->addItem(tr("10 Bits [11:2]"), GIGE_BITMODE_10_1);
        ui->cb_bit_depth->addItem(tr("8 Bits [11:4]"), GIGE_BITMODE_10_3);
        ui->cb_bit_depth->addItem(tr("10 Bits [10:1]"), GIGE_BITMODE_9_0);
        ui->cb_bit_depth->addItem(tr("8 Bits [10:3]"), GIGE_BITMODE_9_2);
        ui->cb_bit_depth->addItem(tr("Middle 8 Bits"), GIGE_BITMODE_9_2);
        ui->cb_bit_depth->addItem(tr("8 Bits [9:2]"), GIGE_BITMODE_8_1);
        ui->cb_bit_depth->addItem(tr("8 Bits [8:1]"), GIGE_BITMODE_7_0);
        ui->cb_bit_depth->addItem(tr("Lower 8 Bits"), GIGE_BITMODE_7_0);
        do_enable = true;
        break;

    case 10:
        // Initialize the combobox.
        ui->cb_bit_depth->insertItem(0, tr("10 Bits"), GIGE_BITMODE_10_BITS);
        ui->cb_bit_depth->insertItem(1, tr("Upper 8 Bits"), GIGE_BITMODE_UPPER_8_BITS);
        ui->cb_bit_depth->insertItem(2, tr("Middle 8 Bits"), GIGE_BITMODE_MIDDLE_8_BITS);
        ui->cb_bit_depth->insertItem(3, tr("Lower 8 Bits"), GIGE_BITMODE_LOWER_8_BITS);
        do_enable = true;
        break;

    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected max_pixel_depth:" << camera_ptr->get_sensor()->max_pixel_depth();
        do_enable = false;
        break;
    }

    // Are we enabling the control?
    if (do_enable)
    {
        // Get the current bitmode setting.
        uint32_t reg_value = read_camera_cache(GIGE_REG_BIT_MODE);

        // Restore the index.
        ui->cb_bit_depth->setCurrentIndex(ui->cb_bit_depth->findData(reg_value));
    }

    // Enable/disable the control.
    ui->cb_bit_depth->setEnabled(do_enable);

    // Restore signals.
    ui->cb_bit_depth->blockSignals(false);
}



/*!
 * @brief The assumption is that, by the time we're called all verification about needing
 *        to update the camera has been done. We're only called when we need to update the
 *        camera.
 *
 *        Sometimes we need to notify the user that making this change will be a big deal.
 *        Use the 'b_quiet' flag to indicate if we should tell the user or not.
 *
 *        Note: The "b_quiet" param has a default setting of 'false' if the caller doesn't
 *        provide it in the call.
 *
 * @param b_quiet   True to just make the changes without notifying the user, false to notify the user.
 * @return True if update is in progress, false if update was canceled.
 */
bool PanelRecSettings::update_camera(bool b_quiet)
{
    bool            roi_changed(false);
    uint32_t        part_state;
    uint32_t        reg_value;
    CameraMDI       *camera_ptr;

    if (connecting_camera)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "connecting_camera is true ignoring request to update camera.";
    }

    // Are we waiting for the video to pause, or are we in the middle of an update?
    if (wait_for_video_pause || update_in_progress || connecting_camera)
    {
        // Ignore the request, and set the focus to the panel.
        setFocus();
        return false;
    }

    // Get a pointer to the active camera.
    camera_ptr = MainWindow::GetInstance()->active_mdi_child();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No active camera!";
        return false;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return false;
    }

    // Remember that an update is pending.
    update_pending = true;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "b_quiet:" << b_quiet;

    // What is the partition state?
    part_state = read_camera_cache(GIGE_REG_PART_STATE);

    // Before we go off and make changes, check to see if we should warn the user.
    if ((part_state == GIGE_PART_STATE_COMPLETE) || (part_state == GIGE_PART_STATE_LOCKED))
    {
        if (!b_quiet)
        {
            int     answer;

            // Check with the user to make sure this is the right thing to do.
            answer = QMessageBox::question(this,
                                           tr("FasMotion"),
                                           tr("Changing the ROI or frame rate will result in the unrecoverable loss of any unsaved video."
                                              "<p>Are you sure you want to change the ROI or frame rate?"),
                                           QMessageBox::Yes | QMessageBox::No);

            // How did the user respond?
            if (answer == QMessageBox::No)
            {
                // Reset our flag.
                update_pending = false;

                // Reset our values.
                set_slider_value(ui->sld_height, read_camera_cache(GIGE_REG_HEIGHT) / camera_ptr->get_sensor()->height_mod());
                ui->spin_height->setValue((int)read_camera_cache(GIGE_REG_HEIGHT));
                set_slider_value(ui->sld_width, read_camera_cache(GIGE_REG_WIDTH) / camera_ptr->get_sensor()->width_mod());
                ui->spin_width->setValue((int)read_camera_cache(GIGE_REG_WIDTH));
                set_slider_value(ui->sld_offsetx, read_camera_cache(GIGE_REG_OFFSET_X) / camera_ptr->get_sensor()->width_mod());
                ui->spin_offsetx->setValue((int)read_camera_cache(GIGE_REG_OFFSET_X));
                set_slider_value(ui->sld_offsety, read_camera_cache(GIGE_REG_OFFSET_Y) / camera_ptr->get_sensor()->height_mod());
                ui->spin_offsety->setValue((int)read_camera_cache(GIGE_REG_OFFSET_Y));

                // Do we need to restore the bit depth?
                if (ui->cb_bit_depth->currentIndex() != ui->cb_bit_depth->findData(read_camera_cache(GIGE_REG_BIT_MODE)))
                {
                    // We don't want the change/restore of the bit depth control to cause a recursion,
                    // so block the signals.
                    ui->cb_bit_depth->blockSignals(true);
                    ui->cb_bit_depth->setCurrentIndex(ui->cb_bit_depth->findData(read_camera_cache(GIGE_REG_BIT_MODE)));
                    ui->cb_bit_depth->blockSignals(false);
                }
                if (camera_ptr->get_sensor()->opts_avail())
                {
                    reg_value = read_camera_cache(GIGE_REG_SENSOR_OPTIONS);
                    ui->cb_bin_sample->setCurrentIndex(ui->cb_bin_sample->findData(reg_value & (GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK)));
                    camera_ptr->get_sensor()->set_options(reg_value);
                }

                int frame_rate = (int)ClientMath::reg_to_frame_rate(read_camera_cache(GIGE_REG_FRAME_RATE));
                set_slider_value(ui->sld_framerate, frame_rate);
                ui->spin_framerate->setValue(frame_rate);

                // Were we updating due to the AutoRATE button being checked?
                if (checking_auto_rate)
                {
                    // Clear the checkbox.
                    ui->chk_autoset_rate->setChecked(false);
                    rate_autoset_clicked();
                }

                // Update our label.
                framerate_label_update();

                // Restore any limits that might have changed.
                update_framerate_limits();
                return false;
            }
        }

        // Empty the partition(s).
        camera_ptr->free_partition(true);
    }

    // Are we changing the ROI?
    if (ui->spin_width->value() != (int)read_camera_cache(GIGE_REG_WIDTH))
    {
        roi_changed = true;
        write_list.append(REG_VALUE_PAIR(GIGE_REG_WIDTH, (uint32_t)ui->spin_width->value()));
    }
    if (ui->spin_height->value() != (int)read_camera_cache(GIGE_REG_HEIGHT))
    {
        roi_changed = true;
        write_list.append(REG_VALUE_PAIR(GIGE_REG_HEIGHT, (uint32_t)ui->spin_height->value()));
    }
    if (ui->spin_offsetx->value() != (int)read_camera_cache(GIGE_REG_OFFSET_X))
    {
        roi_changed = true;
        write_list.append(REG_VALUE_PAIR(GIGE_REG_OFFSET_X, (uint32_t)ui->spin_offsetx->value()));
    }
    if (ui->spin_offsety->value() != (int)read_camera_cache(GIGE_REG_OFFSET_Y))
    {
        roi_changed = true;
        write_list.append(REG_VALUE_PAIR(GIGE_REG_OFFSET_Y, (uint32_t)ui->spin_offsety->value()));
    }

    // Did either exposure or framerate change?
    if (ui->spin_ss->value() != (int)read_camera_cache(GIGE_REG_SHUTTER_SPEED))
        write_list.append(REG_VALUE_PAIR(GIGE_REG_SHUTTER_SPEED, (uint32_t)ui->spin_ss->value()));
    int frame_rate = (int)ClientMath::reg_to_frame_rate(read_camera_cache(GIGE_REG_FRAME_RATE));
    if (ui->spin_framerate->value() != frame_rate)
        write_list.append(REG_VALUE_PAIR(GIGE_REG_FRAME_RATE, ClientMath::frame_rate_to_reg((float)ui->spin_framerate->value())));

    // Did the bit depth change?
    if (ui->cb_bit_depth->currentData().toUInt() != read_camera_cache(GIGE_REG_BIT_MODE))
    {
        // Tell the camera the bit-depth we're interested in using.
        write_list.append(REG_VALUE_PAIR(GIGE_REG_BIT_MODE, ui->cb_bit_depth->currentData().toUInt()));
    }

    // Did the sensor options change?
    if (camera_ptr->get_sensor()->opts_avail())
    {
        reg_value = read_camera_cache(GIGE_REG_SENSOR_OPTIONS);
        if (ui->cb_bin_sample->currentData().toUInt() != (reg_value  & (GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK)))
        {
            // Get the new options settings, preserving the other bits.
            reg_value = (reg_value & ~((uint32_t)(GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK))) | ui->cb_bin_sample->currentData().toUInt();

            // Tell the camera the bit-depth we're interested in using.
            write_list.append(REG_VALUE_PAIR(GIGE_REG_SENSOR_OPTIONS, reg_value));

            // Update our sensor object with the new options settings.
            camera_ptr->get_sensor()->set_options(reg_value);
        }
    }

    // Start the write process.
    start_update_write(roi_changed);

    // Update in progress.
    return true;
}

/*!
 * @brief The maximum allowed shutter speed (exposure) changes
 *        based on the selected frame rate.
 */
void PanelRecSettings::update_exposure_limits()
{
    uint32_t    max_shutter;
    double      ss_degree;

    // Given the frame rate we can set the boundaries for the
    // shutter speed.
    max_shutter = ClientMath::calc_max_exposure((uint32_t)ui->spin_framerate->value(), read_camera_cache(GIGE_REG_FRAME_OVERHEAD));

    // Make sure the calculated value makes sense.
    if (max_shutter <= (uint32_t)ui->sld_ss->minimum())
        return;

    // Did it change?
    if (max_shutter == (uint32_t)ui->spin_ss->maximum())
        return;

    // Now set the maximum limits.
    ui->sld_ss->setMaximum((int)max_shutter);
    ui->spin_ss->setMaximum((int)max_shutter);

    // How many degrees is this?
    ss_degree = (double(ui->spin_ss->value()) / double(max_shutter + read_camera_cache(GIGE_REG_FRAME_OVERHEAD))) * 360.0;
    if ((ss_degree < (ui->spin_degrees->value() - FLOAT_EPSLION)) || (ss_degree > (ui->spin_degrees->value() + FLOAT_EPSLION)))
        ui->spin_degrees->setValue(ss_degree);
}

/*!
 * @brief Update the FPN combobox based on what the sensor supports.
 */
void PanelRecSettings::update_fpn()
{
    int         sel_index = 0;
    QString     sel_text;
    CameraMDI   *camera_ptr;

    // Have we already come in here?
    if (update_pending)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__;
        return;
    }

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Do we have a sensor pointer?
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No sensor pointer!";
        return;
    }

    // Get the currently selected index.
    sel_text = ui->cb_fpn->currentText();

    // Reset and rebuild the FPN combobox based on the sensor information.
    ui->cb_fpn->blockSignals(true);
    ui->cb_fpn->clear();
    ui->cb_fpn->addItem(tr("Disabled"));
    fpn_disabled_index = 0;

    if (camera_ptr->get_sensor()->supports_sensor_fpn())
    {
        // Add the label.
        ui->cb_fpn->addItem(tr("Sensor"));

        // Set the sensor index value.
        fpn_sensor_index = sel_index + 1;
        sel_index = fpn_sensor_index;
    }
    else
    {
        // Make the sensor index illegal.
        fpn_sensor_index = -1;
    }


    // Does this sensor support pixel FPN?
    if (camera_ptr->get_sensor()->supports_pixel_fpn())
    {
        QString     item_str = tr("Pixel");
        uint32_t    reg_value;

        // Get the FPN register.
        reg_value = camera_ptr->read_camera_cache(GIGE_REG_FPN);
        if ((reg_value & GIGE_FPN_FPGA_UNCALIBRATED) == GIGE_FPN_FPGA_UNCALIBRATED)
            item_str.append(QLatin1Char('*'));
        ui->cb_fpn->addItem(item_str);

        // Set the pixel index value.
        fpn_pixel_index = sel_index + 1;
    }
    else
    {
        // Make the "pixel" option illegal.
        fpn_pixel_index = -1;
    }

    // Find the previously selected item.
    sel_index = ui->cb_fpn->findText(sel_text, (Qt::MatchStartsWith | Qt::MatchCaseSensitive));
    if (sel_index < 0)
    {
        // If the option doesn't exist, select disabled.
        sel_index = fpn_disabled_index;
    }

    // Restore the index.
    ui->cb_fpn->setCurrentIndex(sel_index);

    // Restore signals.
    ui->cb_fpn->blockSignals(false);
}

/*!
 * @brief The maximum frame rate changes based on the ROI the user
 *        has defined. This method calculates the limit based on
 *        the current ROI.
 */
void PanelRecSettings::update_framerate_limits()
{
    uint32_t    max_framerate;
    uint32_t    ssd_bandwidth;
    CameraMDI   *camera_ptr;

    // Ignore this if we're in auto-roi.
    if (ui->chk_autoset_roi->isChecked())
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Are we in LTR mode?
    if (ltr_enabled)
    {
        bool        ok;

        // Get the SSD bandwidth used by the camera.
        ssd_bandwidth = read_camera_cache(GIGE_REG_SSD_BANDWIDTH, &ok);
        if (!ok)
            ssd_bandwidth = FASMOTION_SETTINGS::DEFAULT_SSD_BANDWIDTH;
    }
    else
    {
        // No SSD bandwidth to worry about.
        ssd_bandwidth = 0;
    }

    // Sanity checks.
    if ((ui->spin_width->value() == 0) || (ui->spin_height->value() == 0))
        return;

    // Figure out the maximum allowed framerate.
    max_framerate = ClientMath::calc_max_framerate(camera_ptr->get_sensor(), (uint32_t)ui->spin_offsetx->value(),
                                                   (uint32_t)ui->spin_width->value(), (uint32_t)ui->spin_height->value(),
                                                   ui->cb_bit_depth->currentData().toInt(),
                                                   ltr_enabled, ssd_bandwidth, camera_ptr->is_series3());

    // Update our controls.
    ui->sld_framerate->blockSignals(true);
    if (ui->sld_framerate->minimum() != (int)LOW_LIGHT_FRAMERATE_MIN)
        ui->sld_framerate->setMinimum(LOW_LIGHT_FRAMERATE_MIN);
    if (ui->sld_framerate->maximum() != (int)max_framerate)
        ui->sld_framerate->setMaximum((int)max_framerate);
    ui->sld_framerate->blockSignals(false);
    if (ui->spin_framerate->minimum() != (int)LOW_LIGHT_FRAMERATE_MIN)
        ui->spin_framerate->setMinimum(LOW_LIGHT_FRAMERATE_MIN);
    if (ui->spin_framerate->maximum() != (int)max_framerate)
        ui->spin_framerate->setMaximum((int)max_framerate);

    // Calculate the low-light range because we need to have the frame overhead
    // and at this point even though it's overkill we must use the frame overhead
    // or risk the possibility of calculating the wrong low light max exposure by
    // using zero.
    calc_lowlight_range(read_camera_cache(GIGE_REG_FRAME_OVERHEAD));

    // Are we in autoset rate mode?
    if (ui->chk_autoset_rate->isChecked() && (ui->sld_framerate->value() != (int)max_framerate))
    {
        // Maximize the framerate.
        set_slider_value(ui->sld_framerate, (int) max_framerate);
        ui->spin_framerate->setValue((int)max_framerate);
        framerate_label_update();

        // Update our exposure controls.
        update_exposure_limits();

        // If we are not in a pending update, let's update the camera.
        if (!update_pending)
        {
            // Do the update.
            update_camera(true);
        }
    }
    else
    {
        // Update our exposure controls.
        update_exposure_limits();
    }
}

/*!
 * @brief Utility method to update the display controls which show the number
 *        of frames and the amount of record time as defined by the selected ROI.
 */
void PanelRecSettings::update_info()
{
    uint32_t    frame_count;
    QTime       rec_time;

    // Get the maximum number of frames supported by our configuration.
    frame_count = calc_max_total_frames();

    // Update the info field.
    ui->sess_frames->setText(QString::number(frame_count));

    // How much time are we talking about?
    ClientMath::calc_record_ms(frame_count, (uint32_t)ui->spin_framerate->value(), &rec_time);

    // Update our fields.
    ui->rec_time_hrs->setText(QString::fromUtf8("%1").arg(rec_time.hour(), 2, 10, QLatin1Char('0')));
    ui->rec_time_min->setText(QString::fromUtf8("%1").arg(rec_time.minute(), 2, 10, QLatin1Char('0')));
    ui->rec_time_sec->setText(QString::fromUtf8("%1").arg(rec_time.second(), 2, 10, QLatin1Char('0')));
    ui->rec_time_ms->setText(QString::fromUtf8("%1").arg(rec_time.msec(), 3, 10, QLatin1Char('0')));

    double fvalue = ((double)ClientMath::calc_frame_size(ui->spin_width->value(), ui->spin_height->value(), ui->cb_bit_depth->currentData().toUInt()) * (double)ui->spin_framerate->value()) / 1.0e6;
    //ui->sess_datarate->setText(QString(tr("%1")).arg(fvalue, 0, 'f', 3));
    ui->sess_datarate->setText(QString(tr("%1")).arg(fvalue, 0, 'f', 0));
}


/*!
 * @brief When the aspect ratio changes, we need to update what we'll allow as a
 *        maximum ROI setting.
 *
 * @param aspect_ratio  The new aspect ratio selected.
 */
void  PanelRecSettings::update_min_max_roi(int aspect_ratio)
{
    int         max_height;
    int         max_width;
    QRect       max_roi;
    int         min_height;
    int         min_width;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Get the maximum ROI supported by the selected bitmode.
    max_roi = camera_ptr->get_sensor()->max_roi((uint32_t)ui->cb_bit_depth->currentData().toUInt());

    // Initialize to use the maximum ROI supported by the sensor.
    max_height = max_roi.height();
    max_width = max_roi.width();

    // If we are using a custom aspect ratio, then that's good enough. Otherwise,
    // we probably need to limit one or the other dimension.
    if (aspect_ratio != ASPECT_CUSTOM)
    {
        int camera_min_height;
        int camera_min_width;

        // Figure out the maximum values based on the new aspect ratio.
        max_height = aspect_update(true, max_roi.width(), aspect_ratio);
        if (max_height > max_roi.height())
            max_height = max_roi.height();
        max_width = aspect_update(false, max_height, aspect_ratio);
        if (max_width > max_roi.width())
        {
            max_width = max_roi.width();
            max_height = aspect_update(true, max_roi.width(), aspect_ratio);
        }

        // Read the camera minimums.
        camera_min_height = (int)camera_ptr->read_camera_cache(GIGE_REG_MIN_HEIGHT);
        camera_min_width = (int)camera_ptr->read_camera_cache(GIGE_REG_MIN_WIDTH);

        // Figure out the minimum values based on the new aspect ratio.
        min_height = aspect_update(true, camera_min_width, aspect_ratio);
        if (min_height < camera_min_height)
            min_height = camera_min_height;
        min_width = aspect_update(false, min_height, aspect_ratio);
        if (min_width < camera_min_width)
        {
            min_width = camera_min_width;
            min_height = aspect_update(true, camera_min_width, aspect_ratio);
        }
    }
    else
    {
        // Read the camera minimums.
        min_height = (int)camera_ptr->read_camera_cache(GIGE_REG_MIN_HEIGHT);
        min_width = (int)camera_ptr->read_camera_cache(GIGE_REG_MIN_WIDTH);
    }

    // Update our controls.
    ui->spin_height->setMaximum(max_height);
    ui->spin_height->setMinimum(min_height);
    ui->sld_height->setMaximum(max_height / camera_ptr->get_sensor()->height_mod());
    ui->sld_height->setMinimum(min_height / camera_ptr->get_sensor()->height_mod());
    ui->spin_width->setMaximum(max_width);
    ui->spin_width->setMinimum(min_width);
    ui->sld_width->setMaximum(max_width / camera_ptr->get_sensor()->width_mod());
    ui->sld_width->setMinimum(min_width / camera_ptr->get_sensor()->width_mod());
}

/*!
 * @brief The user has modified a control which affects the x- and y-offsets, or we've
 *        finished reading a register from the camera. Update our controls to reflect
 *        the new settings.
 */
void PanelRecSettings::update_offsets()
{
    bool    is_checked;
    bool    has_range;
    int     offset_x;
    int     offset_y;

    if (!get_center_roi_offsets(offset_x, offset_y))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to calculate center ROI offsets!";
        return;
    }

    // Is the check box "checked"?
    is_checked = ui->chk_center_roi->isChecked();
    if (is_checked)
    {
        // Update the offset information.
        if (offset_x != ui->spin_offsetx->value())
        {
            ui->spin_offsetx->blockSignals(true);
            ui->spin_offsetx->setValue(offset_x);
            ui->spin_offsetx->blockSignals(false);
        }
        spin_offsetx_changed(offset_x);

        if (offset_y != ui->spin_offsety->value())
        {
            ui->spin_offsety->blockSignals(true);
            ui->spin_offsety->setValue(offset_y);
            ui->spin_offsety->blockSignals(false);
        }
        spin_offsety_changed(offset_y);
    }

    // Enable the offset controls, if appropriate.
    //
    // The sliders only need to be enabled if the maximum and minimum are different.
    has_range = ui->sld_offsetx->maximum() != ui->sld_offsetx->minimum();
    ui->sld_offsetx->setEnabled(!is_checked && has_range);
    ui->spin_offsetx->setEnabled(!is_checked && has_range);
    ui->lbl_offsetx->setEnabled(!is_checked && has_range);

    has_range = ui->sld_offsety->maximum() != ui->sld_offsety->minimum();
    ui->sld_offsety->setEnabled(!is_checked && has_range);
    ui->spin_offsety->setEnabled(!is_checked && has_range);
    ui->lbl_offsety->setEnabled(!is_checked && has_range);
}

/*!
 * @brief Virtual function called by the "parent" widget to have us update the
 *        contents of our panel.
 */
void PanelRecSettings::update_panel()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Is the camera tracking exposure?
    ui->chk_ae_enable->setChecked(camera_ptr->ae_reticle_visible());
    ui->chk_track_ss->setChecked(camera_ptr->tracking_exposure());
    ae_enable();

    // Stop polling the camera.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }

    // Read registers we're interested in.
    read_registers();

    // Restart our timer.
    update_timer_id = startTimer(POLLING_PERIOD);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void PanelRecSettings::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    FMPanelWidget::changeEvent(event);
}

/*!
 * @brief Event routine which is invoked when our timer expires.
 * @param event     Pointer to the event class.
 */
void PanelRecSettings::timerEvent(QTimerEvent *event)
{
    CameraMDI   *camera_ptr;

    // Pass it on up.
    FMPanelWidget::timerEvent(event);

    // Is it the one we're interested in?
    if (event->timerId() != update_timer_id)
        return;

    // Do we have a camera connected?
    if (p_control_panel->getCamera() == nullptr)
        return;

    // Are we in the middle of an update?
    if (update_in_progress)
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr->get_sensor() == nullptr)
        return;

    // Are we visible?
    if (isVisible())
    {
        if (camera_ptr->get_sensor() == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No sensor available.";
            return;
        }

        QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
        if (cam_register_group.isNull())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
            return;
        }

        // These don't come as unsolicited messages, so we need to poll them.
        // Prepare the camera register group read for our calculations.
        cam_register_group->addReg(GIGE_REG_BIT_MODE);
        if (camera_ptr->get_sensor()->opts_avail())
            cam_register_group->addReg(GIGE_REG_SENSOR_OPTIONS);
        cam_register_group->addReg(GIGE_REG_LOW_LIGHT_ENABLE);
        cam_register_group->addReg(GIGE_REG_LOW_LIGHT_SHUTTER);
        cam_register_group->addReg(GIGE_REG_HDMI);
        cam_register_group->addReg(GIGE_REG_PART_NUM_FRAMES);

        // Last register address in this group is the GIGE_REG_BROC_LEN.
        // Do not change it as the handler depends on this being a unique last register in group.
        cam_register_group->lastReg(GIGE_REG_BROC_LEN);

        // Pass on the request but only increment our counter if we sent it to the camera.
        read_camera_registers(cam_register_group);
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The aspect ratio combobox has changed.
 * @param new_value     The new value in the combobox.
 */
void PanelRecSettings::aspect_changed(int new_value)
{
    int cur_index;
    int cur_height;
    int cur_width;

    // Don't go any further if we're disabled.
    if (!ui->frm_ROI->isEnabled())
        return;

    // Don't allow changes while we're working with the control.
    ui->cb_aspect->blockSignals(true);

    // Get the current height & width;
    cur_height = ui->spin_height->value();
    cur_width = ui->spin_width->value();

    // Figure out the "old" aspect ratio.
    cur_index = calc_aspect_ratio(cur_width, cur_height);
    if (cur_index == new_value)
        goto ac_bail;

    // Update our maximum ROI allowed based on the aspect ratio selected.
    update_min_max_roi(new_value);

    // Don't mess around if we're using a custom ratio.
    if (new_value != ASPECT_CUSTOM)
    {
        int new_height;

        // Get the height based on the width and aspect ratio.
        new_height = aspect_update(true, cur_width, new_value);
        if (new_height != cur_height)
        {
            // Make sure we will fit.
            if (new_height <= ui->spin_height->maximum())
            {
                // Update our controls.
                ui->spin_height->setValue(new_height);
                handle_height_change(new_height);
            }
            else
            {
                int new_width;

                // Adjust the width instead.
                new_width = aspect_update(false, cur_height, new_value);

                // Update our controls.
                ui->spin_width->setValue(new_width);
                handle_width_change(new_width);
            }
        }

        // Better update the camera with the new info.
        update_camera();
    }

ac_bail:
    // Re-enable the control.
    ui->cb_aspect->blockSignals(false);
}

/*!
 * @brief The bin and sample combobox has changed.
 * @param new_value     The new value in the combobox.
 */
void PanelRecSettings::bin_sample_changed(int new_value)
{
    uint32_t old_value;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Does the camera support options?
    if (!camera_ptr->get_sensor()->opts_avail())
        return;

    // Did the value really change?
    old_value = read_camera_cache(GIGE_REG_SENSOR_OPTIONS) & (GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK);
    if (old_value == (uint32_t)new_value)
        return;

    // Make sure we're centered.
    set_center_roi(true, __PRETTY_FUNCTION__);

    // Are we going back to "normal"?
    if (new_value == 0)
    {
        uint32_t    sensor_opts;

        // Get the new options settings, preserving the other bits.
        sensor_opts = ((uint32_t)new_value & ~((uint32_t)(GIGE_SENSOPT_BIN_MASK | GIGE_SENSOPT_SUBS_MASK))) | ui->cb_bin_sample->currentData().toUInt();

        // Update our sensor object with the new options settings. We need to do this
        // before updating the offsets so that our sensor will return the correct
        // "maximum" ROI values.
        camera_ptr->get_sensor()->set_options(sensor_opts);

        // Now that we've told our sensor about the new option settings, we can
        // update our offsets based on the new "maximum" ROI.
        update_offsets();

        // Remember that an update is pending.
        update_pending = true;

        // Build up a list of writes to set the X & Y offsets and the new mode.
        write_list.append(REG_VALUE_PAIR(GIGE_REG_OFFSET_X, (uint32_t)ui->spin_offsetx->value()));
        write_list.append(REG_VALUE_PAIR(GIGE_REG_OFFSET_Y, (uint32_t)ui->spin_offsety->value()));
        write_list.append(REG_VALUE_PAIR(GIGE_REG_SENSOR_OPTIONS, sensor_opts));

        // Start the update writing process.
        start_update_write(true);
    }
    else
    {
        // Just update the camera with the new sensor option value.
        update_camera();
    }
}

/*!
 * @brief The bit depth combobox has changed.
 * @param new_value     The new value in the combobox.
 */
void PanelRecSettings::bitdepth_changed(int new_value)
{
    uint32_t    old_value;
    QRect       max_roi;
    CameraMDI   *camera_ptr;

    // Did the value really change?
    old_value = read_camera_cache(GIGE_REG_BIT_MODE);
    if (old_value == ui->cb_bit_depth->itemData(new_value).toUInt())
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Remember that an update is pending.
    update_pending = true;

    // We need to be careful about changing bit-depth. Verify
    // that our ROI will work for the newly selected mode.
    max_roi = camera_ptr->get_sensor()->max_roi((uint32_t)ui->cb_bit_depth->currentData().toUInt());
    if ((max_roi.width() < ui->spin_width->value()) || (max_roi.height() < ui->spin_height->value()))
    {
        // We need to change the ROI to work with this bit depth. Ask the user if this is OK.
        int     answer;

        // Check with the user to make sure this is the right thing to do.
        answer = QMessageBox::question(this,
                                       tr("FasMotion"),
                                       tr("The bit depth you selected will not work with the currently specified ROI."
                                          "<p>Are you sure you want to change the bit depth?"),
                                       QMessageBox::Yes | QMessageBox::No);

        // How did the user respond?
        if (answer == QMessageBox::No)
        {
            update_pending = false;
            return;
        }
    }

    // Get the maximum ROI supported by the selected bitmode.
    update_min_max_roi(ui->cb_aspect->currentIndex());

    // Update our offsets.
    update_offsets();

    // Update our framerate information.
    update_framerate_limits();

    // Update our information fields.
    update_info();

    // Update the camera with the new info.
    update_camera();
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the read registers group read's last address == GIGE_REG_BROC_LEN.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void PanelRecSettings::chk_broc_regs(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr  = p_control_panel->getCamera();

    // If we are not connected to any camera we don't care about this signal.
    if (camera_ptr != nullptr)
    {
        // Is this event for us?
        if (ip_addr != camera_ptr->get_ip_addr())
            return;

        // Verify it's our register.
        if (reg_addr == GIGE_REG_BROC_LEN)
        {
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_BROC_LEN]";
#endif
            // Update the panel display(s).
            // NOTE: Even though these two distinct group read handlers (GIGE_REG_BROC_LEN and GIGE_REG_LTR_MODE)
            // only end up calling update_info - I kept them separate if in the future we do something different
            // for the action.  For now it's just updating the information.
            update_info();
        }
    }
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the read registers group read's last address == GIGE_REG_LTR_MODE.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void PanelRecSettings::chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr  = p_control_panel->getCamera();

    // If we are not connected to any camera we don't care about this signal.
    if (camera_ptr != nullptr)
    {
        // Is this event for us?
        if (ip_addr != camera_ptr->get_ip_addr())
            return;

        // Verify it's our register.
        if (reg_addr == GIGE_REG_LTR_MODE)
        {
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_LTR_MODE]";
#endif
            // Update the panel display(s).
            // NOTE: Even though these two distinct group read handlers (GIGE_REG_BROC_LEN and GIGE_REG_LTR_MODE)
            // only end up calling update_info - I kept them separate if in the future we do something different
            // for the action.  For now it's just updating the information.
            update_info();
        }
    }
}

/*!
 * @brief The user clicked on the "center" checkbox to either force
 *        the ROI to be centered in the sensor, or to allow the user to
 *        adjust the placement of the ROI.
 */
void PanelRecSettings::center_clicked()
{
    // Update the offsets.
    update_offsets();

    // Are we checked and was there a change in at least one of the offsets?
    if (ui->chk_center_roi->isChecked() &&
        ((ui->spin_offsetx->value() != (int)read_camera_cache(GIGE_REG_OFFSET_X)) || (ui->spin_offsety->value() != (int)read_camera_cache(GIGE_REG_OFFSET_Y))))
    {
        // We probably need to update the camera too.
        if (!update_camera())
        {
            // The settings have been restored, just reverse the check box.
            bool set_center_check = !ui->chk_center_roi->isChecked();
            set_center_roi(set_center_check, __PRETTY_FUNCTION__);
            roi_autoset_clicked();
        }
    }
}

/*!
 * @brief The user has finished mucking around with the exposure (shutter). Update the
 *        related controls and the camera with the new setting.
 */
void PanelRecSettings::exposure_update()
{
    uint32_t    old_value;

    // Have we already come in here?
    if (update_pending)
        return;

    // Update the controls now that the user has finished editing.
    spin_exposure_changed(ui->spin_ss->value());

    // Did the value change?
    old_value = read_camera_cache(GIGE_REG_SHUTTER_SPEED);
    if (old_value == (uint32_t)ui->spin_ss->value())
        return;

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_in_progress = true;
    last_reg_addr = GIGE_REG_SHUTTER_SPEED;

    // Update the camera's shutter speed.
    write_camera_register(GIGE_REG_SHUTTER_SPEED, (uint32_t)ui->spin_ss->value());
}

/*!
 * @brief The FPN combobox has changed.
 * @param new_value     The new value in the combobox.
 */
void PanelRecSettings::fpn_changed(int new_value)
{
    uint32_t        reg_value;
    CameraMDI       *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Get the old value from the cache.
    reg_value = read_camera_cache(GIGE_REG_FPN);

    // Figure out how to transition from the old state to the new one.
    if ((reg_value & GIGE_FPN_SENSOR_ENABLE_MASK) == GIGE_FPN_SENSOR_ENABLE_MASK)
    {
        // Are we staying in Sensor FPN mode?
        if (new_value == fpn_sensor_index)
            return;

        // Does the user want to turn on Pixel FPN or turn off FPN all together?
        if (new_value == fpn_pixel_index)
            write_list.append(REG_VALUE_PAIR(GIGE_REG_FPN, GIGE_FPN_FPGA_ENABLE_MASK));
        else
            write_list.append(REG_VALUE_PAIR(GIGE_REG_FPN, 0));
    }
    else if ((reg_value & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK)
    {
        // Are we staying in Pixel FPN mode?
        if (new_value == fpn_pixel_index)
            return;

        // Does the user want to turn on Sensor FPN or turn off FPN all together?
        if (new_value == fpn_sensor_index)
            write_list.append(REG_VALUE_PAIR(GIGE_REG_FPN, GIGE_FPN_SENSOR_ENABLE_MASK));
        else
            write_list.append(REG_VALUE_PAIR(GIGE_REG_FPN, 0));
    }
    else
    {
        // Are we staying in the off state?
        if (new_value == fpn_disabled_index)
            return;

        // Which mode do we want to enable?
        if (new_value == fpn_pixel_index)
        {
            // Enable Pixel FPN.
            write_list.append(REG_VALUE_PAIR(GIGE_REG_FPN, GIGE_FPN_FPGA_ENABLE_MASK));
        }
        else
        {
            // Enable the column FPN.
            write_list.append(REG_VALUE_PAIR(GIGE_REG_FPN, GIGE_FPN_SENSOR_ENABLE_MASK));
        }
    }

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_pending = true;
    last_reg_addr = GIGE_REG_FPN;

    // We need to wait for the video to pause before we can start writing our registers.
    wait_for_video_pause = true;

    // If we're not streaming at all, or the video stream has already been paused, then
    // just start writing the registers. Otherwise we need to pause the video first.
    if (!camera_ptr->get_camera_info()->stream || video_paused)
        stream_paused();
    else
        camera_ptr->pause_video_request();
}

/*!
 * @brief The connected camera has had sync in stream resumed.
 * Need to re-enable the panel if the sync-in enabled stream is resumed.
 */
void PanelRecSettings::handle_sync_in_resume()

{
    setEnabled(true);
}

/*!
 * @brief The connected camera has had sync in stream paused (timed out).
 * Need to disable the panel if the sync-in enabled stream is timed-out (stopped).
 */
void PanelRecSettings::handle_sync_in_timeout()
{
    setEnabled(false);
}


/*!
 * @brief The user has finished mucking around with the frame rate. Update the
 *        related controls and the camera with the new setting.
 */
void PanelRecSettings::framerate_update()
{
    uint32_t    old_value;

    // Have we already come in here?
    if (update_pending)
        return;

    // Update the controls now that the user has finished editing.
    spin_framerate_changed(ui->spin_framerate->value());

    // Did the value change?
    old_value = (uint32_t)ClientMath::reg_to_frame_rate(read_camera_cache(GIGE_REG_FRAME_RATE));
    if (old_value == (uint32_t)ui->sld_framerate->value())
        return;

    // Our shutter speed limits changes with the frame rate. Don't let the
    // changes generate signals.
    update_exposure_limits();

    // Update the camera.
    update_camera();
}

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void PanelRecSettings::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Pass it along like it was a read complete event.
    handle_read(reg_addr, reg_value);
}

/*!
 * @brief The user is modifying the height, dragging the slider around.
 * @param new_value     The new height.
 */
void PanelRecSettings::height_changed(int new_value)
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // The slider is half the spinner, so double the value before
    // putting it up to the spinner control.
    ui->spin_height->blockSignals(true);
    ui->spin_height->setValue(new_value * camera_ptr->get_sensor()->height_mod());
    ui->spin_height->blockSignals(false);

    // Don't go any further if we're disabled.
    if (ui->frm_ROI->isEnabled())
    {
        // Update the controls now that the user has finished editing.
        handle_height_change(ui->spin_height->value());

        // Update our framerate information.
        update_framerate_limits();
    }
}

/*!
 * @brief The user has finished modifying the height, so udpate the camera
 *        with the new value.
 */
void PanelRecSettings::height_update()
{
    // Have we already come in here?
    if (update_pending)
        return;

    // Don't go any further if we're disabled.
    if (ui->frm_ROI->isEnabled())
    {
        int new_value;
        CameraMDI   *camera_ptr;

        // Make sure we have an active camera pointer.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
            return;
        }
        if (camera_ptr->get_sensor() == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
            return;
        }

        // Make sure the value fits our modulo requirements.
        new_value = ((ui->spin_height->value() / camera_ptr->get_sensor()->height_mod()) * camera_ptr->get_sensor()->height_mod());
        if (new_value != ui->spin_height->value())
        {
            ui->spin_height->blockSignals(true);
            ui->spin_height->setValue(new_value);
            ui->spin_height->blockSignals(false);
        }

        // Update the controls now that the user has finished editing.
        handle_height_change(new_value);

        // Did the value change?
        uint32_t old_value = read_camera_cache(GIGE_REG_HEIGHT);
        if (old_value == (uint32_t)new_value)
            return;

        // Update the camera.
        update_camera();
    }
}

/*!
 * @brief The user clicked on the low-light checkbox to enable or disable
 *        the low light option.
 */
void PanelRecSettings::lowlight_clicked()
{
    // Enable/disable the low light exposure controls.
    ui->sld_ss_ll->setEnabled(ui->chk_ll->isChecked());
    ui->spin_ss_ll->setEnabled(ui->chk_ll->isChecked());

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_in_progress = true;
    last_reg_addr = GIGE_REG_LOW_LIGHT_ENABLE;

    // Tell the camera to enable/disable the low light feature.
    write_camera_register(GIGE_REG_LOW_LIGHT_ENABLE, ui->chk_ll->isChecked());
}

/*!
 * @brief The user has finished modifying the low-light setting, so udpate the camera
 *        with the new value.
 */
void PanelRecSettings::lowlight_update()
{
    // Have we already come in here?
    if (update_in_progress)
        return;

    // Anything change?
    if (ui->spin_ss_ll->value() == (int)read_camera_cache(GIGE_REG_LOW_LIGHT_SHUTTER))
        return;

    // Update the controls now that the user has finished editing.
    spin_lowlight_changed(ui->spin_ss_ll->value());

    // Lock out any updates from the camera - don't want the GUI to flash to wrong
    // values before I get a chance to make my change! This gets reset once we get
    // the write ACK.
    update_in_progress = true;
    last_reg_addr = GIGE_REG_LOW_LIGHT_SHUTTER;

    // Update the low light exposure setting.
    write_camera_register(GIGE_REG_LOW_LIGHT_SHUTTER, (uint32_t)ui->spin_ss_ll->value());
}





/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param ip_addr       The IP address of the camera which didn't respond.
 * @param reg_addr      The address of the register which was read.
 * @param read_op       True if this was a read operation, false for a write operation.
 */
void PanelRecSettings::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Is there a write in progress?
    if (update_in_progress && !read_op)
    {
        // Make sure this is the address we're concerned about.
        if (reg_addr != last_write.first)
            return;

        // Update our retry counter.
        ++retry_counter;

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "retry #" << retry_counter << "Last write:" << last_write;

        // Should we retry the write?
        if (retry_counter < FM_PANEL::MAX_WRITE_RETRIES)
        {
            // Retry the write.
            write_camera_register(last_write.first, last_write.second, false);
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Maximum write attempts exceeded.";

            // Reset our write list and flags.
            write_list.clear();
            update_in_progress = false;
        }
    }
}

/*!
 * @brief The user has finished mucking around with the X-offset. Update the
 *        related controls and the camera with the new setting.
 */
void PanelRecSettings::offsetx_update()
{
    uint32_t    old_value;

    // Have we already come in here?
    if (update_pending)
        return;

    // Don't go any further if we're disabled.
    if (!ui->frm_ROI->isEnabled())
        return;

    // Update the controls now that the user has finished editing.
    spin_offsetx_changed(ui->spin_offsetx->value());

    // Did the value change?
    old_value = read_camera_cache(GIGE_REG_OFFSET_X);
    if (old_value == (uint32_t)ui->spin_offsetx->value())
        return;

    // Update our framerate information.
    update_framerate_limits();

    // Update the camera.
    update_camera();
}

/*!
 * @brief The user has finished mucking around with the Y-offset. Update the
 *        related controls and the camera with the new setting.
 */
void PanelRecSettings::offsety_update()
{
    uint32_t    old_value;

    // Have we already come in here?
    if (update_pending)
        return;

    // Don't go any further if we're disabled.
    if (!ui->frm_ROI->isEnabled())
        return;

    // Update the controls now that the user has finished editing.
    spin_offsety_changed(ui->spin_offsety->value());

    // Did the value change?
    old_value = read_camera_cache(GIGE_REG_OFFSET_Y);
    if (old_value == (uint32_t)ui->spin_offsety->value())
        return;

    // Update our framerate information.
    update_framerate_limits();

    // Update the camera.
    update_camera();
}

/*!
 * @brief The user clicked on the rate autoset checkbox to enable
 *        or disable automatically adjusting the frame rate based
 *        on the ROI.
 */
void PanelRecSettings::rate_autoset_clicked()
{
    // Enable/disable the rate controls based on the check state of the Autoset button.
    ui->lbl_frame_rate_grp->setEnabled(!ui->chk_autoset_rate->isChecked());
    ui->lbl_time_grp->setEnabled(!ui->chk_autoset_rate->isChecked());
    ui->sld_framerate->setEnabled(!ui->chk_autoset_rate->isChecked());
    ui->spin_framerate->setEnabled(!ui->chk_autoset_rate->isChecked());

    // Do special handling if we're checked.
    if (ui->chk_autoset_rate->isChecked())
    {
        // Remember that we're processing the AutoRATE button being checked.
        checking_auto_rate = true;

        // Make sure we un-check the ROI autoset button.
        if (ui->chk_autoset_roi->isChecked())
        {
            ui->chk_autoset_roi->setChecked(false);
            roi_autoset_clicked();
        }

        // Set the frame rate to it's maximum value for the current ROI.
        if (ui->sld_framerate->value() != ui->sld_framerate->maximum())
        {
            ui->sld_framerate->setValue(ui->sld_framerate->maximum());
            framerate_update();
        }

        // We're done with the auto rate processing.
        checking_auto_rate = false;
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void PanelRecSettings::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;

    // Pass it up.
    FMPanelWidget::read_complete(ip_addr, reg_addr);

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Retrieve the camera state.
    reg_value = read_camera_cache(reg_addr);

    // Handle the read event.
    handle_read(reg_addr, reg_value);
}

/*!
 * @brief The ROI "autoset" check box has been selected. If
 *        checked, the user wants the computer to calculate
 *        the "best" ROI for a given framerate.
 */
void PanelRecSettings::roi_autoset_clicked()
{
    // Enable/disable the ROI controls based on the check state of the Autoset button.
    ui->lbl_width->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->spin_width->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->sld_width->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->lbl_height->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->spin_height->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->sld_height->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->lbl_aspect->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->cb_aspect->setEnabled(!ui->chk_autoset_roi->isChecked());
    ui->chk_center_roi->setEnabled(!ui->chk_autoset_roi->isChecked());

    // Do special handling if we're checked.
    if (ui->chk_autoset_roi->isChecked())
    {
        uint32_t    max_framerate;
        uint32_t    ssd_bandwidth;
        uint32_t    offset_x;
        CameraMDI   *camera_ptr;

        // Make sure we have an active camera pointer.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
            return;
        }

        // When we're checked, the offset controls are disabled.
        ui->lbl_offsetx->setEnabled(false);
        ui->spin_offsetx->setEnabled(false);
        ui->sld_offsetx->setEnabled(false);
        ui->lbl_offsety->setEnabled(false);
        ui->spin_offsety->setEnabled(false);
        ui->sld_offsety->setEnabled(false);

        // Make sure our centered checkbox is selected.
        set_center_roi(true, __PRETTY_FUNCTION__);

        // If we're not visible, then we're done.
        if (!isVisible())
            return;

        // Make sure we un-check the rate autoset button.
        if (ui->chk_autoset_rate->isChecked())
        {
            ui->chk_autoset_rate->setChecked(false);
            rate_autoset_clicked();
        }

        // Are we in LTR mode?
        if (ltr_enabled)
        {
            bool        ok;

            // Get the SSD bandwidth used by the camera.
            ssd_bandwidth = read_camera_cache(GIGE_REG_SSD_BANDWIDTH, &ok);
            if (!ok)
                ssd_bandwidth = FASMOTION_SETTINGS::DEFAULT_SSD_BANDWIDTH;
        }
        else
        {
            // No SSD bandwidth to worry about.
            ssd_bandwidth = 0;
        }

        // Do we want to remember our custom width/height value?
        if (ui->cb_aspect->currentIndex() == ASPECT_CUSTOM)
        {
            custom_width = (uint32_t)ui->spin_width->value();
            custom_height = (uint32_t)ui->spin_height->value();
        }

        // Get the center offset for the minimum allowed ROI.
        offset_x = (((uint32_t)ui->spin_width->maximum() >> 1) - ((uint32_t)ui->spin_width->minimum() >> 1)) >> 1;

        // Figure out the maximum allowed framerate.
        max_framerate = ClientMath::calc_max_framerate(camera_ptr->get_sensor(), offset_x,
                                                       (uint32_t)ui->spin_width->minimum(), (uint32_t)ui->spin_height->minimum(),
                                                       ui->cb_bit_depth->currentData().toInt(),
                                                       ltr_enabled, ssd_bandwidth, camera_ptr->is_series3());

        // Allow our slider to go to this value.
        if (ui->sld_framerate->maximum() != (int)max_framerate)
        {
            ui->sld_framerate->blockSignals(true);
            ui->sld_framerate->setMaximum((int)max_framerate);
            ui->sld_framerate->blockSignals(false);
        }

        if (ui->spin_framerate->maximum() != (int)max_framerate)
            ui->spin_framerate->setMaximum((int)max_framerate);

        // Go to the maximum ROI for the current framerate.
        auto_max_roi((uint32_t)ui->spin_framerate->value());

        // Update the offsets to be centered.
        update_offsets();

        // Update the camera.
        if (!update_camera())
        {
            // The settings have been restored, just reverse the check box.
            ui->chk_autoset_roi->setChecked(false);
            roi_autoset_clicked();
        }
    }
    else if (isVisible())
    {
        // Update our offset controls.
        update_offsets();

        // Update our limits.
        update_framerate_limits();
    }
}

/*!
 * @brief The user has modified the exposure (shutter) using the slider.
 * @param new_value     The new exposure.
 */
void PanelRecSettings::slide_exposure_changed(int new_value)
{
    double  ss_degree;

    // Update the spinner.
    ui->spin_ss->setValue(new_value);

    // How many degrees is this?
    ss_degree = (double(new_value) / double((uint32_t)ui->spin_ss->maximum() + read_camera_cache(GIGE_REG_FRAME_OVERHEAD))) * 360.0;
    if ((ss_degree < (ui->spin_degrees->value() - FLOAT_EPSLION)) || (ss_degree > (ui->spin_degrees->value() + FLOAT_EPSLION)))
        ui->spin_degrees->setValue(ss_degree);
}

void PanelRecSettings::ae_enable()
{
    CameraMDI   *camera_ptr;
    PanelRecControls *controls_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Enable/Disable the exposure tracking controls based on this selection.
    controls_ptr = static_cast<PanelRecControls *>(p_control_panel->getPanel(CONTROL_PANEL::PANEL_REC_CONTROLS));
    controls_ptr->enableVtControls(!ui->chk_ae_enable->isChecked());

    // Show/hide the reticle
    camera_ptr->ae_reticle_show(ui->chk_ae_enable->isChecked());
}

void PanelRecSettings::ae_reset_region()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Show/hide the reticle
    camera_ptr->ae_reticle_reset();
    ui->chk_ae_enable->setChecked(true);
}

void PanelRecSettings::track_exposure()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // Is the track box checked?
    camera_ptr->track_exposure(ui->chk_track_ss->isChecked());
}

/*!
 * @brief The user has modified the frame rate using the slider.
 * @param new_value     The new frame rate.
 */
void PanelRecSettings::slide_framerate_changed(int new_value)
{
    // Update the spinner.
    if (ui->spin_framerate->value() != new_value)
        ui->spin_framerate->setValue(new_value);

    framerate_label_update();

    // Are we in auto-ROI mode?
    if (ui->chk_autoset_roi->isChecked())
    {
        // Adjust the maximum ROI.
        auto_max_roi((uint32_t)new_value);

        // Update the offsets to keep us centered.
        update_offsets();
    }
}

/*!
 * @brief The user has modified the low-light setting using the slider.
 * @param new_value     The new low-light setting.
 */
void PanelRecSettings::slide_lowlight_changed(int new_value)
{
    // Update the spinner.
    ui->spin_ss_ll->setValue(new_value);
}

/*!
 * @brief The user has modified the X-offset using the slider.
 * @param new_value     The new x-offset.
 */
void PanelRecSettings::slide_offsetx_changed(int new_value)
{
    CameraMDI       *camera_ptr;

    // Get a pointer to the active camera.
    if (p_control_panel == nullptr)
        return;
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
        return;
    if (camera_ptr->get_sensor() == nullptr)
        return;

    // The slider is 1/width_mod() the spinner, so multiply the value by width_mod() before
    // putting it up to the spinner control.
    ui->spin_offsetx->blockSignals(true);
    ui->spin_offsetx->setValue(new_value * camera_ptr->get_sensor()->width_mod());
    ui->spin_offsetx->blockSignals(false);
}

/*!
 * @brief The user has modified the Y-offset using the slider.
 * @param new_value     The new y-offset.
 */
void PanelRecSettings::slide_offsety_changed(int new_value)
{
    CameraMDI       *camera_ptr;

    // Get a pointer to the active camera.
    if (p_control_panel == nullptr)
        return;
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
        return;
    if (camera_ptr->get_sensor() == nullptr)
        return;

    // The slider is 1/height_mod() the spinner, so multiply the value by height_mod() before
    // putting it up to the spinner control.
    ui->spin_offsety->blockSignals(true);
    ui->spin_offsety->setValue(new_value * camera_ptr->get_sensor()->height_mod());
    ui->spin_offsety->blockSignals(false);
}

/*!
 * @brief The user has modified the exposure (shutter) using the spinner.
 * @param new_value     The new exposure.
 */
void PanelRecSettings::spin_exposure_changed(int new_value)
{
    double  ss_degree;

    // Update the slider.
    ui->sld_ss->setValue(new_value);

    // How many degrees is this?
    ss_degree = (double(new_value) / double((uint32_t)ui->sld_ss->maximum() + read_camera_cache(GIGE_REG_FRAME_OVERHEAD))) * 360.0;
    if ((ss_degree < (ui->spin_degrees->value() - FLOAT_EPSLION)) || (ss_degree > (ui->spin_degrees->value() + FLOAT_EPSLION)))
        ui->spin_degrees->setValue(ss_degree);
}

/*!
 * @brief The user has modified the frame rate using the spinner.
 * @param new_value     The new frame rate.
 */
void PanelRecSettings::spin_framerate_changed(int new_value)
{
    // Update the slider.
    if (ui->sld_framerate->value() != new_value)
        set_slider_value(ui->sld_framerate, new_value);

    // Update the label.
    framerate_label_update();

    // Are we in auto-ROI mode?
    if (ui->chk_autoset_roi->isChecked())
    {
        // Adjust the maximum ROI.
        auto_max_roi((uint32_t)new_value);

        // Update the offsets to keep us centered.
        update_offsets();
    }
}

/*!
 * @brief The user has modified the low-light setting using the spinner.
 * @param new_value     The new low-light setting.
 */
void PanelRecSettings::spin_lowlight_changed(int new_value)
{
    // Update the slider.
    ui->sld_ss_ll->setValue(new_value);
}

/*!
 * @brief The user has modified the X-offset using the spinner.
 * @param new_value     The new x-offset.
 */
void PanelRecSettings::spin_offsetx_changed(int new_value)
{
    CameraMDI       *camera_ptr;

    // Get a pointer to the active camera.
    if (p_control_panel == nullptr)
        return;
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
        return;
    if (camera_ptr->get_sensor() == nullptr)
        return;

    // Don't go any further if we're disabled.
    if (!ui->frm_ROI->isEnabled())
        return;

    // To keep values correct, we make the slider 1/modulo of the
    // value of the spin.
    set_slider_value(ui->sld_offsetx, new_value / camera_ptr->get_sensor()->width_mod());
}

/*!
 * @brief The user has modified the Y-offset using the spinner.
 * @param new_value     The new y-offset.
 */
void PanelRecSettings::spin_offsety_changed(int new_value)
{
    CameraMDI       *camera_ptr;

    // Get a pointer to the active camera.
    if (p_control_panel == nullptr)
        return;
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
        return;
    if (camera_ptr->get_sensor() == nullptr)
        return;

    // Don't go any further if we're disabled.
    if (!ui->frm_ROI->isEnabled())
        return;

    // To keep values correct, we make the slider 1/modulo of the
    // value of the spin.
    set_slider_value(ui->sld_offsety, new_value / camera_ptr->get_sensor()->height_mod());
}

/*!
 * @brief Notification slot that the stream has switch to live mode.
 */
void PanelRecSettings::stream_live()
{
    CameraMDI       *camera_ptr;

    // Get a pointer to the active camera.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera?!";
        return;
    }

    // Pass it up.
    FMPanelWidget::stream_live();

    // Make sure our cursor is restored.
    QApplication::restoreOverrideCursor();
    setEnabled(true);
    if (camera_ptr->get_info_widget())
        camera_ptr->get_info_widget()->enable_actions();

    // We can resume the histogram again.
    MainWindow::GetInstance()->resume_histogram();
}

/*!
 * @brief Notification slot that the stream has been paused.
 */
void PanelRecSettings::stream_paused()
{
    // Pass it up.
    FMPanelWidget::stream_paused();

    // Were we waiting for the video to pause?
    if (!wait_for_video_pause)
        return;

    // Now that the video is paused, we don't have to wait for it to pause.
    wait_for_video_pause = false;
    update_pending = false;

    // Do we have any updates pending?
    if (!write_list.isEmpty())
    {
        REG_VALUE_PAIR  reg_pair;
        CameraMDI       *camera_ptr;

        // Get a pointer to the active camera.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera?!";
            return;
        }

        // Lock out any updates from the camera - don't want the GUI to flash to wrong
        // values before I get a chance to make my change! This gets reset once we get
        // the write ACK.
        update_in_progress = true;

        // This could take awhile.
        QApplication::setOverrideCursor(Qt::BusyCursor);
        setEnabled(false);
        if (camera_ptr->get_info_widget())
            camera_ptr->get_info_widget()->disable_actions();

        // The first thing we want to do is grab the lock.
        write_camera_register(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE);

        // Next up to grab the hold register.
        write_camera_register(GIGE_REG_HOLD, 1);
        camera_hold = true;

        // Start the write operation with the first operation.
        reg_pair = write_list.takeFirst();
        write_camera_register(reg_pair.first, reg_pair.second);

        // Remember this address so we can compare in the write complete logic.
        last_reg_addr = reg_pair.first;
    }
}

/*!
 * @brief Toggle the visiblity of the panel. Optionally, pass
 *        the event up to the parent.
 *
 * @param notify_parent  True if we're supposed to pass this up to the parent.
 */
void PanelRecSettings::toggle_visibility(bool notify_parent)
{
    // Pass it on up.
    FMPanelWidget::toggle_visibility(notify_parent);

    // If we're opened up, make sure we refresh our controls
    if (isVisible())
    {
        CameraMDI       *camera_ptr;

        // Get a pointer to the active camera.
        camera_ptr = p_control_panel->getCamera();

        // Do we have a camera connected?
        if (camera_ptr != nullptr)
        {
            // Update our panel.
            read_registers();

            // Create our polling timer if we don't already have one.
            if (update_timer_id == 0)
            {
                update_timer_id = startTimer(POLLING_PERIOD);
            }
        }
    }
    else
    {
        // Stop polling the camera when we're closed.
        killTimer(update_timer_id);
        update_timer_id = 0;

        // Un-check any of the auto settings.
        ui->chk_autoset_rate->setChecked(false);
        rate_autoset_clicked();
        ui->chk_autoset_roi->setChecked(false);
        roi_autoset_clicked();
    }
}

/*!
 * @brief Update the panel's controls.
 */
void PanelRecSettings::update_controls()
{
    CameraMDI   *camera_ptr;
    QSharedPointer<FiCamRegGroup> cam_register_group;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if ((camera_ptr == nullptr) || (camera_ptr->get_sensor() == nullptr))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer or sensor object!";
        goto exitUpdateControls;
    }

    cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        goto exitUpdateControls;
    }

    // Prepare the camera register group read for our calculations.
    cam_register_group->addReg(GIGE_REG_MIN_HEIGHT);
    cam_register_group->addReg(GIGE_REG_MIN_WIDTH);
    cam_register_group->addReg(GIGE_REG_SENSOR_WIDTH);
    cam_register_group->addReg(GIGE_REG_SENSOR_HEIGHT);
    cam_register_group->addReg(GIGE_REG_MAX_FRAME_RATE);
    cam_register_group->lastReg(GIGE_REG_FPN);
    read_camera_registers(cam_register_group);

    // Update our spinners based on the sensor we've attached to.
    ui->spin_height->setSingleStep(camera_ptr->get_sensor()->height_mod());
    ui->spin_width->setSingleStep(camera_ptr->get_sensor()->width_mod());
    ui->spin_offsety->setSingleStep(camera_ptr->get_sensor()->height_mod());
    ui->spin_offsetx->setSingleStep(camera_ptr->get_sensor()->width_mod());

    // Update our UI elements based on the sensor information.
    update_bin_sample();
    update_bitdepth();
    update_fpn();
    set_aspect_ratio();

exitUpdateControls:
    // If the state of connecting to a camera was set prior to this call
    // go ahead and reset it.
    if (connecting_camera)
    {
        int offset_x;
        int offset_y;

        // Reset as we have fully connected the camera.
        connecting_camera = false;

        // Determine if the connected camera is centered.  If so make the centered
        // checkbox checked.  If not make the centered checkbox unchecked.  Remember
        // that the checkbox was manually set to unchecked at the start of the camera
        // connection operation.

        // Note:  An assumption made that this is only place to make this check right
        // at the very end of the connection to the camera operation.

        // What is the calculated center point?
        if (get_center_roi_offsets(offset_x, offset_y))
        {
            bool matches = true;

            // Does the camera's x,y offsets match the calculated one?
            if ((ui->spin_offsetx->value() != offset_x) || (ui->spin_offsety->value() != offset_y))
            {
                matches = false;
            }

            // Keep this readable for documentation's sake and perform a check or uncheck
            // of the center ROI checkbox using the logic below.
            if (!matches)
            {
                // Was the center checked?  If so we want to uncheck it.
                if (ui->chk_center_roi->isChecked())
                    set_center_roi(false, __PRETTY_FUNCTION__);
            }
            else
            {
                // Was the center unchecked?  If so we want to check it.
                if (!ui->chk_center_roi->isChecked())
                    set_center_roi(true, __PRETTY_FUNCTION__);
            }
        }
    }
}

/*!
 * @brief The user is modifying the width, dragging the slider around.
 * @param new_value     The new width.
 */
void PanelRecSettings::width_changed(int new_value)
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
        return;
    }

    // The slider is 1/modulo the spinner, so convert the value before
    // putting it up to the spinner control.
    ui->spin_width->blockSignals(true);
    ui->spin_width->setValue(new_value * camera_ptr->get_sensor()->width_mod());
    ui->spin_width->blockSignals(false);

    // Don't go any further if we're disabled.
    if (ui->frm_ROI->isEnabled())
    {
        // Update the controls now that the user has finished editing.
        handle_width_change(ui->spin_width->value());

        // Update our framerate information.
        update_framerate_limits();
    }
}

/*!
 * @brief The user has finished modifying the width, so udpate the camera
 *        with the new value.
 */
void PanelRecSettings::width_update()
{
    // Have we already come in here?
    if (update_pending)
        return;

    // Don't go any further if we're disabled.
    if (ui->frm_ROI->isEnabled())
    {
        int new_value;
        CameraMDI   *camera_ptr;

        // Make sure we have an active camera pointer.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
            return;
        }
        if (camera_ptr->get_sensor() == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera connected.";
            return;
        }

        // Make sure the value fits our modulo requirements.
        new_value = ((ui->spin_width->value() / camera_ptr->get_sensor()->width_mod()) * camera_ptr->get_sensor()->width_mod());
        if (new_value != ui->spin_width->value())
        {
            ui->spin_width->blockSignals(true);
            ui->spin_width->setValue(new_value);
            ui->spin_width->blockSignals(false);
        }

        // Update the controls now that the user has finished editing.
        handle_width_change(new_value);

        // Did the value change?
        uint32_t old_value = read_camera_cache(GIGE_REG_WIDTH);
        if (old_value == (uint32_t)new_value)
            return;

        // Update the camera.
        update_camera();
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void PanelRecSettings::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Were we waiting for the video to pause, or were we even doing an update?
    if (wait_for_video_pause || !update_in_progress)
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Is this a register we care about?
    // The single register write we care about is for a seriesNC camera, the LR mode registrer.
    if (!camera_ptr->is_series3() && (reg_addr == GIGE_REG_LTR_MODE))
    {
        // Read back the value to update our cache.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_LTR_MODE finished, read back. update_in_progress:" << update_in_progress;
        update_in_progress = false;
        start_camera_read(reg_addr);
        return;
    }

    // The only other thing we care about is that the block of registers
    // that we write for an update has completed.
    else if (last_reg_addr != reg_addr)
        return;

    // Get our communications object pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Was it a successful write?
    write_status = comm_ptr->get_write_status(camera_ip_addr, reg_addr);
    if (write_status != GEV_STATUS_SUCCESS)
    {
        QHostAddress    our_addr;
        our_addr.setAddress(qFromBigEndian(camera_ip_addr));

        QString err_msg = QString::fromUtf8("Unable to write register 0x%1 to IP address %2 status[%3] write_list count=%4 camera_hold:%5").arg(reg_addr, 0, 16).arg(our_addr.toString()).arg(FI_comm::gev_status_to_msg(write_status)).arg(write_list.count()).arg(camera_hold);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;

        // Notes:
        // The write operation failed.  It is at this point that maybe any additional
        // registers sitting in the write_list should be cleared?  But it's a complex
        // issue because we don't know what type of error nor the relationship between
        // the registers in a collective list.  What happens today is that control simply
        // falls through and continues execution.  To mark one known condition to generate
        // GEV_STATUS_ACCESS_DENIED status can be setup by having a sync-in signal go away
        // when changing an ROI set up.  This is a condition where multiple registers may
        // be modified but the camera_server cannot facilitate the request.  Right now
        // this condition should be mitigated by external forces controlling the panel
        // itself - if a sync-in signal goes away the panel is disabled and if it resumes
        // the panel is enabled.  So theoretically we will not fall into this condition.
        // but there may be other ways writes can fail and I am just making a note here in
        // case I end up back here on another day for another ticket...
    }
    else
    {
        // Read back the value to update our cache.
        start_camera_read(reg_addr);
    }

    // Are there still write operations to perform?
    if (!write_list.isEmpty())
    {
        REG_VALUE_PAIR  reg_pair;

        // Grab the next operation off the list.
        reg_pair = write_list.takeFirst();

        // Issue the write command.
        write_camera_register(reg_pair.first, reg_pair.second);

        // Remember this address so we can compare in the write complete logic.
        last_reg_addr = reg_pair.first;
    }
    else
    {
        // Are we holding the camera for a bulk update?
        if (camera_hold)
        {
            // Make sure we release the hold register.
            write_camera_register(GIGE_REG_HOLD, 0);
            camera_hold = false;
        }

        // The last thing we want to do is release the lock.
        write_camera_register(GIGE_REG_FASTEC_CCP, 0);

        // Everything looks good. Reset our updating flag.
        update_in_progress = false;
        last_reg_addr = FI_COMM::NO_REGISTER;

        // Make sure our cursor is restored.
        QApplication::restoreOverrideCursor();

        // Do we want to resume our video stream?
        CameraMDI *camera_ptr = MainWindow::GetInstance()->active_mdi_child();
        if (camera_ptr && camera_ptr->get_camera_info()->stream && video_paused)
        {
            // Resume our video stream.
            camera_ptr->resume_video_request();
        }
        else
        {
            // Go ahead an enable our panel.
            setEnabled(true);
            if (camera_ptr && camera_ptr->get_info_widget())
                camera_ptr->get_info_widget()->enable_actions();
        }

        // I'm going to assume that the update we just did changed the ROI in some way.
        // Let's schedule the camera video player object to adjust it's zoom.
        if (camera_ptr && camera_ptr->get_player())
            camera_ptr->get_player()->need_rezoom();
    }
}

