/*!
 * @file panel_recsettings.h
 * @brief This is the header file for the PanelRecSettings class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PANEL_RECSETTINGS_H
#define PANEL_RECSETTINGS_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QWidget>
#include "panel_widget.h"

namespace Ui
{
class PanelRecSettings;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QSlider;
class QTimerEvent;

//--------------------------------------------------------------------
// Class Declaration
//--------------------------------------------------------------------
class PanelRecSettings : public FMPanelWidget
{
    Q_OBJECT

public:
    explicit PanelRecSettings(QWidget *parent = nullptr);
    ~PanelRecSettings();

    // Handle camera connect/disconnects.
    virtual void camera_connect(uint32_t ip_addr);
    virtual void camera_disconnect();

    // Provide interface to push the current settings on this panel
    // to the camera settings for frame rate and exposure if they
    // are different from what is already set on the camera.
    virtual void push_framerate_and_exposure();

    // Enable the autoexposure controls.
    void enable_ae_controls(bool do_enable);
    bool ae_enabled()   { return ae_allowed; }

    // Update the panel contents.
    virtual void update_panel();

protected slots:
    virtual void    changeEvent(QEvent *event);
    virtual void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    virtual void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    virtual void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    virtual void    stream_live();
    virtual void    stream_paused();
    virtual void    toggle_visibility(bool notify_parent = true);
    virtual void    timerEvent(QTimerEvent *event);
    virtual void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private slots:
    void    aspect_changed(int new_value);
    void    bin_sample_changed(int new_value);
    void    bitdepth_changed(int new_value);
    void    chk_broc_regs(uint32_t ip_addr, uint32_t reg_addr);
    void    chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr);
    void    fpn_changed(int new_value);

    void    handle_sync_in_resume();
    void    handle_sync_in_timeout();

    void    height_changed(int new_value);
    void    height_update();

    void    width_changed(int new_value);
    void    width_update();

    void    offsetx_update();
    void    spin_offsetx_changed(int new_value);
    void    slide_offsetx_changed(int new_value);

    void    offsety_update();
    void    spin_offsety_changed(int new_value);
    void    slide_offsety_changed(int new_value);

    void    framerate_update();
    void    spin_framerate_changed(int new_value);
    void    slide_framerate_changed(int new_value);

    void    exposure_update();
    void    spin_exposure_changed(int new_value);
    void    slide_exposure_changed(int new_value);
    void    ae_enable();
    void    ae_reset_region();
    void    track_exposure();

    void    lowlight_update();
    void    spin_lowlight_changed(int new_value);
    void    slide_lowlight_changed(int new_value);

    void    lowlight_clicked();
    void    rate_autoset_clicked();
    void    roi_autoset_clicked();

    void    center_clicked();


    void    update_controls();

private:
    int         aspect_update(bool calc_height, int base_value, int aspect_type);
    void        auto_max_roi(uint32_t framerate);
    int         calc_aspect_ratio(int width, int height);
    void        calc_lowlight_range(uint32_t frame_overhead = 0);
    uint32_t    calc_max_total_frames();
    void        framerate_label_update();
    bool        get_center_roi_offsets(int& offset_x, int& offset_y);
    void        handle_height_change(int new_value, bool update_aspect = true);
    void        handle_read(uint32_t reg_addr, uint32_t reg_value);
    void        handle_width_change(int new_value, bool update_aspect = true);
    void        init_controls();
    bool        is_offset_centered();
    void        process_camera_state(uint32_t new_state);
    void        read_registers();
    bool        roi_control_has_focus();
    void        set_aspect_ratio();
    void        set_center_roi(bool check_it, QString who_called_me);
    void        start_update_write(bool roi_change);
    void        translate_ui();
    void        update_bin_sample();
    void        update_bitdepth();
    bool        update_camera(bool b_quiet = false);
    void        update_exposure_limits();
    void        update_fpn();
    void        update_framerate_limits();
    void        update_info();
    void        update_min_max_roi(int aspect_ratio);
    void        update_offsets();

private:
    Ui::PanelRecSettings *ui;               //!< Our UI object.

    uint32_t    camera_ip_addr;             //!< IP address of the camera we're talking to.
    bool        connecting_camera;          //!< True if connecting a camera and waiting to update controls.
    bool        checking_auto_rate;         //!< True if we're processing an auto rate check.

    uint32_t    custom_height;              //!< The height when the user has specified a custom aspect.
    uint32_t    custom_width;               //!< The width when the user has specified a custom aspect.

    int         fpn_disabled_index;         //!< Index in the combo box for the "disabled" option.
    int         fpn_sensor_index;           //!< Index in the combo box for the "sensor" option.
    int         fpn_pixel_index;            //!< Index in the combo box for the "pixel" option.

    bool        ae_allowed;                 //!< A flag to remember if exposure tracking is allowed or not.
    uint32_t    last_reg_addr;              //!< Used to terminate a block register read, this contains the last register address in the block.
    bool        ltr_enabled;                //!< Is Long Record enabled?
    bool        update_in_progress;         //!< Flag to indicate that an update (write operation) is in progress.
    bool        update_pending;             //!< When an update is pending, we'll ignore values from the camera.
    int         update_timer_id;            //!< Timer ID to trigger an update.
    bool        wait_for_video_pause;       //!< Flag to indicate that we're waiting for the video from the camera to stop.
    bool        camera_hold;                //!< Flag to indicate if we've held the camera for a bulk register update.
};

#endif // PANEL_RECSETTINGS_H
