/*!
 * @file panel_widget.cpp
 *
 * @brief This file contains the base class for control panel accordian widgets.
 *
 * This is the source file for the FMPanelWidget class. This is the
 * base class for all the accordian tabs on the control panel. The
 * purpose was to encapsulate sizing issue.
 *
 * Copyright 2010, 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#include "panel_widget.h"

#include <QMdiSubWindow>
#include <QToolButton>

#include "camera_mdi.h"
#include "control_panel.h"
#include "fi_camlist.h"
#include "mainwindow.h"
#include "multicam_dock.h"

/*!
 * @brief Constructor for the base class. Initialize variables.
 */
FMPanelWidget::FMPanelWidget(QWidget *parent) :
    QWidget(parent)
{
    // Save off the control panel which at this point is our parent but
    // our parent will at some point change to the splitter.
    p_control_panel = qobject_cast<ControlPanel *>(parent);
    ignore_updates = false;
    retry_counter = 0;
    camera_ip_addr = 0;
    video_paused = false;
    camera_state = GIGE_CAMERA_STATE_IDLE;
    max_height = 0;
    is_panel_open = false;
    slider_changing = false;

    // We are going to use a Queued connection type later, and Qt needs to know about uint32_t types.
    qRegisterMetaType<uint32_t>("uint32_t");
}

/*!
 * @brief Destructor for the base class.
 */
FMPanelWidget::~FMPanelWidget()
{
}

/*!
 * @brief Set the initial sizing for the panel.
 *
 * This function calculates and saves the size of the panel. If we're supposed
 * to start closed (bStartMinimized is true), then change our state.
 *
 * @param bStartMinimized   Should the dialog start minimized?
 */
void FMPanelWidget::set_initial_sizing(bool bStartMinimized)
{
    // We really don't allow the user to adjust our height (compress us).
    // We're either open or we're closed. So get our maximum height,
    // which is the height when we're open.
    max_height = geometry().height();

    // Assume we start open, but if the caller wants us closed (bStartMinimized is true)
    // then we need to adjust ourself.
    is_panel_open = true;
    if (bStartMinimized)
    {
        // The do_change_size() slot toggles our state. If we're open (and we are)
        // this will cause us to close.
        toggle_visibility(false);
    }
}


/*!
 * @brief Close the panel up so that just the title and sizing button are visible.
 */
void FMPanelWidget::close_panel()
{
    // If we are open, then toggle the panel to close it.
    if (is_panel_open)
        toggle_visibility(true);
}

/*!
 * @brief Open the panel so the whole thing is visible.
 */
void FMPanelWidget::open_panel()
{
    // call the slot function because it if virtual and that lets the derived
    // class do something if needed.
    if (!is_panel_open)
        toggle_visibility(true);
}

/*!
 * @brief Enable or disable the users ability to open or close a panel. Right now
 * this is only affecting the sizing button.
 */
void FMPanelWidget::enable_panel(bool do_enable)
{
    // Make sure we have all our pointers setup.
    if (!p_control_panel)
        return;

    // Request that the control panel adjust our button appropriately.
    p_control_panel->panelButtonEnable(this, do_enable);
}

/*!
 * @brief We are connecting to a (possibly) new camera.
 * @param ip_addr   The IP address of the camera we're connecting to.
 */
void FMPanelWidget::camera_connect(uint32_t ip_addr)
{
    QMdiSubWindow   *child_win;

    // Make sure we're not already connected to this camera.
    if (ip_addr == camera_ip_addr)
        return;

    // Save the IP address.
    camera_ip_addr = ip_addr;

    // Get the camera window so we can connect to it.
    child_win = MainWindow::GetInstance()->find_mdi_child(ip_addr);
    if (child_win != nullptr)
    {
        // Actually we need to the CameraMDI widget.
        CameraMDI       *child_ptr = qobject_cast<CameraMDI *>(child_win->widget());
        if (child_ptr != nullptr)
        {
            // Catch when the video is live or paused.
            connect(child_ptr, &CameraMDI::video_is_paused, this, &FMPanelWidget::stream_paused);
            connect(child_ptr, &CameraMDI::video_is_streaming, this, &FMPanelWidget::stream_live);

            // Initialize our streaming state.
            video_paused = !child_ptr->get_camera_info()->stream_running;
        }
    }
}

/*!
 * @brief Disconnect from our camera.
 */
void FMPanelWidget::camera_disconnect()
{
    QMdiSubWindow   *child_win;

    // Have we been here already?
    if (camera_ip_addr == 0)
        return;

    // Clear our IP address.
    camera_ip_addr = 0;

    // Get the camera window so we can disconnect from it.
    child_win = MainWindow::GetInstance()->find_mdi_child(camera_ip_addr);
    if (child_win != nullptr)
    {
        CameraMDI *child_ptr = qobject_cast<CameraMDI *>(child_win->widget());
        if (child_ptr != nullptr)
        {
            disconnect(child_ptr, &CameraMDI::video_is_paused, this, &FMPanelWidget::stream_paused);
            disconnect(child_ptr, &CameraMDI::video_is_streaming, this, &FMPanelWidget::stream_live);
        }
    }
}

/*!
 * @brief Update the panel. The base class has nothing to update.
 */
void FMPanelWidget::update_panel()
{
}

/*!
 * @brief Utility method to get the communications pointer.
 */
FI_comm *FMPanelWidget::get_comm_ptr()
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
    }

    // Return our pointer.
    return comm_ptr;
}

/*!
 * @brief Return the pointer to the panel object index by the panel_index parameter.
 *
 * @param panel_index       Which [panel](@ref CONTROL_PANEL::PANEL_ID) we're interested in.
 * @return Pointer to the requested panel, or nullptr.
 */
const FMPanelWidget *FMPanelWidget::get_panel(int panel_index) const
{
    const QWidget       *widget_ptr;
    const FMPanelWidget *panel_ptr;

    widget_ptr = p_control_panel->getPanel(panel_index);
    if (widget_ptr != nullptr)
    {
        // Get our pointer cast to the right object.
        panel_ptr = qobject_cast< const FMPanelWidget * >(widget_ptr);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Invalid panel_index:" << panel_index;
        panel_ptr = nullptr;
    }

    // Return our pointer.
    return panel_ptr;
}

/*!
 * @brief Utility method to read the camera's register cache. The assumption
 *        being made is that the register value has already been loaded from
 *        the camera.
 *
 * @param reg_addr      The address of the register we're interested in.
 * @param ok            [OPTIONAL] pointer to a bool to indicate if the register is in cache.
 *
 * @return The register value or 0 if an error (*ok == false).
 */
uint32_t FMPanelWidget::read_camera_cache(uint32_t reg_addr, bool *ok)
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";

        if (ok != nullptr)
            *ok = false;
        return 0;
    }

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";

        if (ok != nullptr)
            *ok = false;
        return 0;
    }

    // Return the cached value.
    return comm_ptr->get_reg(camera_ptr->get_ip_addr(), reg_addr, ok);
}

/*!
 * @brief Utility method to do a block read of a number of registers.
 * @param cam_reg_group  The camera register group read list object.
 * @returns True if the group read request was sent to the camera, false if not.
 */
bool FMPanelWidget::read_camera_registers(QSharedPointer<FiCamRegGroup> cam_reg_group)
{
    bool        sent_to_camera = false;
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return sent_to_camera;
    }

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return sent_to_camera;
    }

    // Only do this if the camera isn't busy.
    if (!camera_ptr->is_busy())
    {
        // Start a camera group register read.
        if (cam_reg_group->getCount() > 0)
        {
            camera_ptr->read_camera_registers(cam_reg_group);
            sent_to_camera = true;
        }
    }
    return sent_to_camera;
}

/*!
 * @brief Utility method to incorporate the blocking/unblocking of signals around setting
 *        a slider's value. Often times we do NOT want that action to trigger a signal,
 *        thus the need to block/unblock.
 *
 * @param p_slider      Pointer to the slider control we're manipulating.
 * @param new_value     The new value we want to set the slider's value to.
 * @param block         True if we want to block signals, false otherwise.
 */
void FMPanelWidget::set_slider_value(QSlider *p_slider, int new_value, bool block)
{
    // Sanity check.
    if (p_slider == nullptr)
        return;

    // Are we supposed to block signals from the control?
    if (block)
        p_slider->blockSignals(true);

    // Set the new value if it's different.
    if (new_value != p_slider->value())
        p_slider->setValue(new_value);

    // If we blocked, we need to restore signalling.
    if (block)
        p_slider->blockSignals(false);
}

void FMPanelWidget::slider_lock()
{
    // Lock out any updates.
    slider_changing = true;
}

void FMPanelWidget::slider_unlock()
{
    // Release the lockout.
    slider_changing = false;
}


/*!
 * @brief Utility method to initiate a read operation of a register from our camera.
 *
 * The intent here is to actually go TO the camera, and not just read from the
 * register cache. In order to force a read from the actual device, we expire
 * the cache entry.
 *
 * @param reg_addr      The register address we're interested in.
 */
void FMPanelWidget::start_camera_read(uint32_t reg_addr)
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Only do this if the camera isn't busy.
    if (!camera_ptr->is_busy())
    {
        comm_ptr->expire_entry(camera_ptr->get_ip_addr(), reg_addr);
        comm_ptr->start_read_reg(camera_ptr->get_ip_addr(), reg_addr);
    }
}

/*!
 * @brief Utility method to write a register value to the camera(s).
 *
 * If the camera is in a group, then it is assumed the 'active' camera is the 'master'.
 * If the group is 'locked' for synchronization, then the write will be propogated to
 * all the other cameras in the group.
 *
 * @param reg_addr      The register address we want to modify.
 * @param reg_value     The new value we want the register set to.
 * @param reset_retry_count     True if we're supposed to reset our retry counter, false to let it ride.
 */
void FMPanelWidget::write_camera_register(uint32_t reg_addr, uint32_t reg_value, bool reset_retry_count)
{
    int i;
    const fiCamList *cam_list;
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;
    MultiCamDock *multicam;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Only do this if the camera isn't busy.
    if (!camera_ptr->is_busy())
    {
        // Are we supposed to reset our retry counter?
        if (reset_retry_count)
            retry_counter = 0;

        // Remember the information in case we do need to retry.
        last_write.first = reg_addr;
        last_write.second = reg_value;

        // Issue the write.
        comm_ptr->start_write_reg(camera_ptr->get_ip_addr(), reg_addr, reg_value);
    }

    // Let's see if we're supposed to sync with other cameras.
    // Get the multicamera control.
    multicam = MainWindow::GetInstance()->get_multicam_dock();
    if (multicam == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null multicam dock!";
        return;
    }

    // Make sure the group is locked.
    if (!multicam->is_group_locked())
        return;

    // Get the list of cameras.
    cam_list = multicam->get_group_list();
    if (cam_list == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera list!";
        return;
    }

    // Push the write out to everyone in the group.
    for (i = 0; i < cam_list->count(); ++i)
    {
        CameraMDI *group_cam;
        MULTICAM_DOCK::fiListItem *my_item;
        my_item = (MULTICAM_DOCK::fiListItem *)cam_list->item(i);

        // Sanity check.
        if (my_item == nullptr)
        {
            QLOG_ERROR() <<  __PRETTY_FUNCTION__ << "Null my_item at index:" << i;
            break;
        }

        // Get the group camera
        group_cam = my_item->get_camera();
        if (group_cam == nullptr)
        {
            QLOG_ERROR() <<  __PRETTY_FUNCTION__ << "Null camera pointer at index:" << i;
            break;
        }

        // Don't send to the master camera again.
        if (group_cam != camera_ptr)
        {
            // Issue the write.
            comm_ptr->start_write_reg(group_cam->get_ip_addr(), reg_addr, reg_value);
        }
    }
}


//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void FMPanelWidget::do_allow_updates()
{
    ignore_updates = false;
}

void FMPanelWidget::do_ignore_updates()
{
    ignore_updates = true;
}

/*!
 * @brief Slot invoked when we receive an unsolicited message.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void FMPanelWidget::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    Q_UNUSED(ip_addr)
    Q_UNUSED(reg_addr)
    Q_UNUSED(reg_value)
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param reg_addr      The address of the register which was read.
 */
void FMPanelWidget::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    Q_UNUSED(ip_addr)
    Q_UNUSED(reg_addr)
    Q_UNUSED(read_op)
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 */
void FMPanelWidget::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Save the state information. It may prove useful.
    if (reg_addr == GIGE_REG_CAMERA_STATE)
        camera_state = read_camera_cache(reg_addr);
}

/*!
 * @brief Notification slot that the stream has switch to live mode.
 */
void FMPanelWidget::stream_live()
{
    // We're no longer paused.
    video_paused = false;
}

/*!
 * @brief Notification slot that the stream has been paused.
 */
void FMPanelWidget::stream_paused()
{
    // Remember that the video is now paused.
    video_paused = true;
}

/*!
 * @brief Toggle the visiblity of the panel. Optionally, pass
 *        the event up to the parent.
 *
 * @param notify_parent  True if we're supposed to pass this up to the parent.
 */
void FMPanelWidget::toggle_visibility(bool notify_parent)
{
    // Make sure we have all our pointers setup.
    if (!p_control_panel)
        return;

    // Set our main body of controls visible based on the flag.
    setVisible(!is_panel_open);

    // Remember our state.
    is_panel_open = isVisible();

    // Update the button's icon.
    p_control_panel->panelButtonIcon(this, (is_panel_open ? Qt::DownArrow : Qt::UpArrow));

    // Do we want to pass this up to our parent?
    if (notify_parent)
        p_control_panel->childResized(this);
}

/*!
 * @brief Slot invoked when one of the write operations complete.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 */
void FMPanelWidget::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    Q_UNUSED(ip_addr);
    Q_UNUSED(reg_addr);
}

