/*!
 * @file panel_widget.h
 * @brief Header of the base class for the accordian tabs.
 *
 * This is the header file for the FMPanelWidget class. This is the
 * base class for all the accordian tabs on the control panel. The
 * purpose was to encapsulate sizing issue.
 *
 * @author Ian Crabtree
 * @date 2010-07-01
 *
 * Copyright 2010 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PANEL_WIDGET_H
#define PANEL_WIDGET_H

//-----------------------------------------------------------------------------
// Includes
//-----------------------------------------------------------------------------
#include <stdint.h>
#include <QWidget>
#include <QFrame>
#include <QList>
#include <QPair>
#include <QSlider>

#include "camera_mdi.h"
#include "control_panel.h"
#include "fi_comm.h"
#include "fi_camreg_group.h"
#include "mainwindow.h"

//--------------------------------------------------------------------
// Namespace: FM_PANEL
//--------------------------------------------------------------------
namespace FM_PANEL
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The maximum number of times we'll retry a write attempt.
const int   MAX_WRITE_RETRIES           = 3;

//! Our polling frequency, in milliseconds.
const int   POLLING_PERIOD              = 2000;

//! Polling period for checking progress.
const int   PROGRESS_PERIOD             = 500;

//! A type for a register address and value pair.
typedef QPair< uint32_t, uint32_t >     REG_VALUE_PAIR;

//! A type for a list of register/value pairs.
typedef QList< REG_VALUE_PAIR >         REG_VALUE_LIST;
}

//-----------------------------------------------------------------------------
// Forward declarations
//-----------------------------------------------------------------------------
class QToolButton;

//--------------------------------------------------------------------
// The FMPanelWidget class definition
//--------------------------------------------------------------------
class FMPanelWidget : public QWidget
{
    Q_OBJECT

public:
    explicit FMPanelWidget(QWidget *parent = nullptr);
    ~FMPanelWidget();

    // Setup initial sizing
    void set_initial_sizing(bool bStartMinimized = true);

    // An unsolicited event happened on the camera. It's up to the derived classes to do anything with it.
    virtual void camera_event_happened(uint32_t /* ip_addr */, uint32_t /* event_type */, uint32_t /* event_value */) {}

    // Handle camera connect/disconnects.
    virtual void camera_connect(uint32_t ip_addr);
    virtual void camera_disconnect();

    // Close or open the panel
    void close_panel();
    void open_panel();

    // Enable/disable the resizing button.
    void enable_panel(bool do_enable);

    // Update the panel contents.
    virtual void update_panel();

public slots:
    virtual void    toggle_visibility(bool notify_parent = true);

protected slots:
    virtual void    do_allow_updates();
    virtual void    do_ignore_updates();
    virtual void    handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value);
    virtual void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    virtual void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    virtual void    stream_live();
    virtual void    stream_paused();
    virtual void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

protected:
    FI_comm     *get_comm_ptr();
    const FMPanelWidget *get_panel(int panel_index) const;

    uint32_t    read_camera_cache(uint32_t reg_addr, bool *ok = nullptr);
    bool        read_camera_registers(QSharedPointer<FiCamRegGroup> camGroupRead);
    void        set_slider_value(QSlider *p_slider, int new_value, bool block = true);
    void        slider_lock();
    void        slider_unlock();
    void        start_camera_read(uint32_t reg_addr);
    void        write_camera_register(uint32_t reg_addr, uint32_t reg_value, bool reset_retry_count = true);

protected:
    ControlPanel    *p_control_panel;   //!< Our container object, different than hierarchy parent.

    bool    ignore_updates;             //!< Flag to ignore or allow updates.
    int     retry_counter;              //!< Number of write retry attempts.
    int     max_height;                 //!< Maximum height for the panel when expanded.
    bool    is_panel_open;              //!< Is the panel expanded (i.e. showing all the controls)?
    bool    video_paused;               //!< True if the video stream is paused, false otherwise.
    bool    slider_changing;            //!< True if the user is manipulating a slider control.

    uint32_t    camera_ip_addr;         //!< IP address of the camera we're talking to.
    uint32_t    camera_state;           //!< Last report camera state.
    FM_PANEL::REG_VALUE_PAIR    last_write;     //!< The last write attempt.
    FM_PANEL::REG_VALUE_LIST    write_list;     //!< List of register and value pairs to write.
};

#endif // PANEL_WIDGET_H
