/*!
 * @file peekpoke.cpp
 * @brief Implementation file for the PeekPokeDialog class.
 *
 * This is the source file for the PeekPokeDialog. It is a test dialog which
 * allows registers to be peeked or poked.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "peekpoke.h"
#include "ui_peekpoke.h"

#include <QtEndian>
#include <QtNetwork>
#include <QTimer>

#include "camera_mdi.h"
#include "fi_comm.h"
#include "mainwindow.h"

// Use a namespac to avoid polluting the global space.
namespace PP_DLG
{
//! Update our temperature values at this rate.
const int       UPDATE_DELAY        = 1000;

const int       FPGA_TAB            = 0;
const int       FCP_TAB             = 1;
const int       TEMP_TAB            = 2;

const QLatin1Char   zero_fill('0');
}

/*!
 * @brief Constructor for the PeekPokeDialog class.
 */
PeekPokeDialog::PeekPokeDialog(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::PeekPokeDialog)
{
    FI_comm     *p_comm;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent);

    // Setup our user interface object(s).
    ui->setupUi(this);

    // Initialize our variable(s).
    ip_addr = 0;

    // Get our communications object pointer.
    p_comm = MainWindow::GetInstance()->get_comm();
    if (p_comm == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Setup our signals/slots.
    connect(ui->pbRead, &QPushButton::clicked, this, &PeekPokeDialog::read_reg);
    connect(ui->pbWrite, &QPushButton::clicked, this, &PeekPokeDialog::write_reg);

    // Connect the communications interfaces that we're interested in.
    connect(p_comm, &FI_comm::read_reg_complete, this, &PeekPokeDialog::read_complete, Qt::QueuedConnection);
    connect(p_comm, &FI_comm::write_reg_complete, this, &PeekPokeDialog::write_complete, Qt::QueuedConnection);

    // Did we get a parent?
    if (my_parent && my_parent->is_series3())
    {
        // Series3 cameras have temperature sensors we can query, but not FPGA peek/poke.
        ui->pp_tab->setTabVisible(PP_DLG::TEMP_TAB, true);
        ui->pp_tab->setTabVisible(PP_DLG::FPGA_TAB, false);
        ui->pp_tab->setCurrentIndex(PP_DLG::FCP_TAB);

        // Setup our timer.
        update_timer = new QTimer(this);
        if (update_timer == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate timer!";
            return;
        }

        // We want to poll the temperature information.
        connect(update_timer, SIGNAL(timeout()), this, SLOT(update_camera_info()));

        // Start our update timer.
        update_timer->start(PP_DLG::UPDATE_DELAY);

        // Force an update right away.
        update_camera_info();
    }
    else
    {
        // New cameras do not have the temperature sensors but do support FPGA peek/poke.
        ui->pp_tab->setTabVisible(PP_DLG::TEMP_TAB, false);
        ui->pp_tab->setTabVisible(PP_DLG::FPGA_TAB, true);
        ui->pp_tab->setCurrentIndex(PP_DLG::FPGA_TAB);
    }

    // Update the dialog layout.
    adjustSize();
    update();
}

/*!
 * @brief Destructor for the PeekPokeDialog class.
 */
PeekPokeDialog::~PeekPokeDialog()
{
    delete ui;
}

/*!
 * @brief Set the IP address for the camera we want to peek/poke at.
 * @param ip_addr   The target camera's IP address.
 */
void PeekPokeDialog::set_ip(uint32_t ip_addr)
{
    QHostAddress    ip_ha;

    // Convert the IP address to a string by using the QHostAddress class.
    ip_ha.setAddress(qFromBigEndian(ip_addr));

    // Let the user know what we're using.
    ui->status_msg->setText(QString::fromUtf8("Using IP Addr: ") + ip_ha.toString());

    // Save it for later.
    this->ip_addr = ip_addr;
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void PeekPokeDialog::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;
    FI_comm     *p_comm;

    // Is this message for us?
    if (this->ip_addr != ip_addr)
        return;

    // Get our communications object pointer.
    p_comm = MainWindow::GetInstance()->get_comm();
    if (p_comm == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Was this an FPGA read?
    if (reg_addr == GIGE_REG_PEEKPOKE_JSON)
    {
        QByteArray  *json_message;

        // Read what was returned.
        json_message = p_comm->get_reg_buffer(ip_addr, GIGE_REG_PEEKPOKE_JSON);
        if (json_message != nullptr)
        {
            QString json_str;
            QJsonParseError jp_err;
            QJsonDocument jsonDoc;

            // Make a QString out of the data.
            json_str = QString::fromUtf8(*json_message);
            if (json_str.length())
            {
                // Strip any trailing NULL characters.
                if (json_str.at(json_str.size() - 1) == '\0')
                    json_str.chop(1);

                // Make sure we have a well formed JSON string. Converting to a JSON document.
                jsonDoc = QJsonDocument::fromJson(json_str.toUtf8(), &jp_err);
                if (jsonDoc.isNull())
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << "Got a parse err:" << jp_err.errorString();
                }
                else
                {
                    QJsonObject peekObj = jsonDoc.object();
                    if (peekObj.contains("value"))
                    {
                        reg_value = (uint32_t)peekObj["value"].toInt();
                        ui->fpga_value->setText(QString::fromUtf8("%1").arg(reg_value, 8, 16, PP_DLG::zero_fill));
                        ui->status_msg->setText(QString::fromUtf8("Read completed successfully!"));
                    }
                }
            }

            // Don't need the buffer anymore.
            delete json_message;
        }
    }

    // Is this the register address the user requested?
    else if (this->reg_addr == reg_addr)
    {
        // Get the returned value and display it.
        reg_value = p_comm->get_reg(ip_addr, reg_addr);
        ui->ebValue->setText(QString::fromUtf8("%1").arg(reg_value, 8, 16, PP_DLG::zero_fill));
        ui->status_msg->setText(QString::fromUtf8("Read completed successfully!"));
    }
    else if (reg_addr == GIGE_REG_TEMPERATURE)
    {
        int img_side;
        int comp_side;
        double  f_tmp;
        QHostAddress    ip_ha;

        // Convert the IP address to a string by using the QHostAddress class.
        ip_ha.setAddress(qFromBigEndian(ip_addr));

        // Read the register value.
        reg_value = p_comm->get_reg(ip_addr, reg_addr);

        // Parse the register value.
        img_side = reg_value & GIGE_TEMPERATURE_MASK;
        comp_side = (reg_value >> GIGE_CSIDE_SHIFT) & GIGE_TEMPERATURE_MASK;

        // The temperatures are in 1/10 of a degree.
        f_tmp = (double)comp_side / 10.0;
        ui->eb_comp_side_temp->setText(QString::fromUtf8("%1").arg(f_tmp, 6, 'f', 1));
        f_tmp = (double)img_side / 10.0;
        ui->eb_img_side_temp->setText(QString::fromUtf8("%1").arg(f_tmp, 6, 'f', 1));

        // Log the information as well.
        QLOG_INFO() << ip_ha.toString() << "component_side:" << ((double)comp_side / 10.0) << "imager_side:" << ((double)img_side / 10.0);
    }
}

/*!
 * @brief The user has request a read from a register.
 */
void PeekPokeDialog::read_reg(void)
{
    QString     reg_str;
    bool        ok_status;
    int         status;
    FI_comm     *p_comm;

    // Clear out the status field.
    ui->status_msg->clear();

    // Sanity check.
    if (ip_addr == 0)
    {
        ui->status_msg->setText(QString::fromUtf8("No IP address!"));
        return;
    }

    // Get our communications object pointer.
    p_comm = MainWindow::GetInstance()->get_comm();
    if (p_comm == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Are we doing a FCP register or FPGA register?
    if (ui->pp_tab->currentIndex() == PP_DLG::FCP_TAB)
    {
        // Get our register address.
        reg_str = ui->ebReg->text();
        if (reg_str.isEmpty())
        {
            ui->status_msg->setText(QString::fromUtf8("Register address is empty!"));
            ui->ebReg->setFocus();
            return;
        }
        reg_addr = reg_str.toUInt(&ok_status, 16);

        // Read our register.
        status = p_comm->start_read_reg(ip_addr, reg_addr);
        if (status != FI_COMM::FI_SUCCESS)
        {
            ui->status_msg->setText(QString::fromUtf8("start_read_reg() failed: %1").arg(status));
            return;
        }
    }
    else
    {
        // Get our register address.
        reg_str = ui->fpga_reg->text();
        if (reg_str.isEmpty())
        {
            ui->status_msg->setText(QString::fromUtf8("Register address is empty!"));
            ui->fpga_reg->setFocus();
            return;
        }
        reg_addr = reg_str.toUInt(&ok_status, 16);

        // This is a 2-step process. First we write the command to the camera telling it what we
        // want to read, then we will do a read command to retrieve the results.
        QJsonObject json_cmd;
        json_cmd["poke"] = false;
        json_cmd["reg"] = (qint64)reg_addr;

        // We are waiting to do the actual read.
        wait_for_read = true;
        reg_addr = GIGE_REG_PEEKPOKE_JSON;

        // Send the JSON command.
        QJsonDocument json_doc(json_cmd);
        p_comm->start_write_mem(ip_addr, GIGE_REG_PEEKPOKE_JSON, json_doc.toJson());
    }
}

void PeekPokeDialog::update_camera_info(void)
{
    FI_comm     *p_comm;

    // Get our communications object pointer.
    p_comm = MainWindow::GetInstance()->get_comm();
    if (p_comm == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Ask the camera for the temperature.
    p_comm->start_read_reg(ip_addr, GIGE_REG_TEMPERATURE);
}

void PeekPokeDialog::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint16_t    write_status;
    FI_comm     *p_comm;

    // Is this message for us?
    if ((this->ip_addr != ip_addr) || (this->reg_addr != reg_addr))
        return;

    // Get our communications object pointer.
    p_comm = MainWindow::GetInstance()->get_comm();
    if (p_comm == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Were we successful?
    write_status = (uint16_t)p_comm->get_write_status(ip_addr, reg_addr);
    if (write_status == GEV_STATUS_SUCCESS)
    {
        ui->status_msg->setText(QString::fromUtf8("Write completed successfully!"));

        // Are we waiting to do the read?
        if ((reg_addr == GIGE_REG_PEEKPOKE_JSON) && wait_for_read)
        {
            // Read back the data.
            p_comm->start_read_mem(ip_addr, GIGE_REG_PEEKPOKE_JSON, GIGE_PEEKPOKE_JSON_LEN);
            wait_for_read = false;
        }
    }
    else
    {
        ui->status_msg->setText(QString::fromUtf8("Write error = %1").arg(write_status));
    }
}

/*!
 * @brief The user has requested to write a register.
 */
void PeekPokeDialog::write_reg(void)
{
    QString     temp_str;
    uint32_t    reg_value;
    bool        ok_status;
    int         status;
    FI_comm     *p_comm;

    // Clear out the status field.
    ui->status_msg->clear();

    // Sanity check.
    if (ip_addr == 0)
    {
        ui->status_msg->setText(QString::fromUtf8("No IP address!"));
        return;
    }

    // Get our communications object pointer.
    p_comm = MainWindow::GetInstance()->get_comm();
    if (p_comm == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Are we doing a FCP register or FPGA register?
    if (ui->pp_tab->currentIndex() == PP_DLG::FCP_TAB)
    {
        // Get our register address.
        temp_str = ui->ebReg->text();
        if (temp_str.isEmpty())
        {
            ui->status_msg->setText(QString::fromUtf8("Register address is empty!"));
            ui->ebReg->setFocus();
            return;
        }
        reg_addr = temp_str.toUInt(&ok_status, 16);

        // Get the value to write to the register.
        temp_str = ui->ebValue->text();
        if (temp_str.isEmpty())
        {
            ui->status_msg->setText(QString::fromUtf8("Register value is empty!"));
            ui->ebValue->setFocus();
            return;
        }
        reg_value = temp_str.toUInt(&ok_status, 16);

        // Write our register.
        status = p_comm->start_write_reg(ip_addr, reg_addr, reg_value);
        if (status != FI_COMM::FI_SUCCESS)
        {
            ui->status_msg->setText(QString::fromUtf8("start_write_reg() failed: %1").arg(status));
            return;
        }
    }
    else
    {
        // Get our register address.
        temp_str = ui->fpga_reg->text();
        if (temp_str.isEmpty())
        {
            ui->status_msg->setText(QString::fromUtf8("Register address is empty!"));
            ui->fpga_reg->setFocus();
            return;
        }
        reg_addr = temp_str.toUInt(&ok_status, 16);

        // Get the value to write to the register.
        temp_str = ui->fpga_value->text();
        if (temp_str.isEmpty())
        {
            ui->status_msg->setText(QString::fromUtf8("Register value is empty!"));
            ui->fpga_value->setFocus();
            return;
        }
        reg_value = temp_str.toUInt(&ok_status, 16);

        // Build up the poke command.
        QJsonObject json_cmd;
        json_cmd["poke"] = true;
        json_cmd["reg"] = (qint64)reg_addr;
        json_cmd["value"] = (qint64)reg_value;

        // We are waiting to do the actual read.
        wait_for_read = false;
        reg_addr = GIGE_REG_PEEKPOKE_JSON;

        // Send the JSON command.
        QJsonDocument json_doc(json_cmd);
        p_comm->start_write_mem(ip_addr, GIGE_REG_PEEKPOKE_JSON, json_doc.toJson());
    }
}

