/*!
 * @file playback_cap.cpp
 * @brief This file contains the implementation of the PlaybackCap
 *        class. This is the child window class which allows the user
 *        to playback a captured video from a Fastec CAP file.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QApplication>
#include <QAction>
#include <QContextMenuEvent>
#include <QDateTime>
#include <QDockWidget>
#include <QElapsedTimer>
#include <QFile>
#include <QGraphicsScene>
#include <QImageWriter>
#include <QJsonDocument>
#include <QMenu>
#include <QMessageBox>
#include <QPixmap>
#include <QPixmapCache>
#include <QtConcurrent>
#include <QStyle>
#include <QThread>
#include <QTimer>
#include <QTimerEvent>

#include <fas_video.h>

#include "camera_info.h"
#include "camera_review.h"
#include "camera_video.h"
#include "eventfilter.h"
#include "fi_metamap.h"
#include "fi_pixmap.h"
#include "fi_progress_dialog.h"
#include "fi_rawfile.h"
#include "fi_saver.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_meta_info.h"
#include "playback_meta_info_fflr.h"
#include "playback_cap.h"
#include "save_dialog.h"
#include "save_video.h"

namespace PLAYBACK_CAP_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

const QString MISSING_RAWFILE  = QStringLiteral("The capRawFile has not been created properly!");

//! Rate at which we poll the saver object's progress.
const int   SAVER_POLL_RATE             = 100;


}   // End of namespace PLAYBACK_CAP_SETTINGS.

using namespace PLAYBACK_CAP_SETTINGS;
using namespace FI_METAMAP;


PlaybackCap::PlaybackCap(QWidget *parent) :
    PlaybackMDI(parent)
{
    // Initialize our members.

    // We don't have an active capRawFile yet.
    capIdActive = FASMOTION_SETTINGS::UNSET_UINT32;
}

PlaybackCap::~PlaybackCap()
{
    // Any opened raw cap files will close when their unique pointer
    // is destructed.  There is no need to explicitly close them here.
}

const QByteArray& PlaybackCap::getRawImage() const
{
    static QByteArray null_array;

    // Sanity check.
    if (!capRawFile.isNull())
        return capRawFile->getRawImage();

    // Sanity check.
    return null_array;
}

/*!
 * @brief Extract the various meta data from the drive and save it to files for use by
 *        the application.
 *
 * @return True on success, false otherwise.
 */
bool PlaybackCap::init_metadata()
{
    bool rc(false);
    QString file_name;
    QString tm_label;
    QString rawPFMDFileName;

    // This can take awhile if there's lots of frames.
    // Create the progress dialog if we don't have one.
    if (!progressDialog)
    {
        progressDialog = fiProgressDialog::CreateSimpleProgressDialog(camera_info->get_camera_name(), tr("Accumulating Metadata..."), QMessageBox::NoButton, __PRETTY_FUNCTION__);
    }

    // Display the progress dialog.
    if (progressDialog)
    {
        progressDialog->setBarMinMax(0, 0);
        progressDialog->showBar(true);
        progressDialog->showDialogWithoutPolling();

        // Give the progress bar a chance to start.
        QApplication::processEvents();
    }

    // Allocate our metadata structure.
    if (pbMetaInfoPtr == nullptr)
    {
        pbMetaInfoPtr = new PlaybackMetaInfo(this);
        if (pbMetaInfoPtr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a PlaybackMetaInfo object!";
            return false;
        }
    }

    // Build up the file name for the pm_metadata buffer it is preparing to write this
    // to the temporary path in the name metadata_hhmmss.txt.
    tm_label = QDateTime::currentDateTime().time().toString(QStringLiteral("hhmmss"));
    file_name = QDir::tempPath() + QString::fromUtf8("/metadata_%1.txt").arg(tm_label);

    // This has a side-effect of refreshing the pm_metadata object in pmMetaDataInfo in
    // addition to writing out a temporary metadata.txt file.
    if (!save_pm_metadata(file_name))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to save the pm_metadata to:" << file_name;
        goto im_bail;
    }

    // Install the name of the temporary metadata.txt file into the playback meta information object.
    pbMetaInfoPtr->setMetadataFile(file_name);

    // Do we already have the raw pfmd filename installed?
    // This could be true if in a FasFireLR playback of a CAP stack.
    rawPFMDFileName = pbMetaInfoPtr->getRawfile();
    if (rawPFMDFileName.isEmpty())
    {
        // Build up the file name for the raw per-frame metadata.
        file_name = QDir::tempPath() + QString::fromUtf8("/metadata_%1.pm").arg(tm_label);

        // Save the per-frame metadata.
        if (!save_raw_metadata(file_name))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to save the pm_metadata to:" << file_name;
            goto im_bail;
        }
        // Pass the info to our metadata class.
        pbMetaInfoPtr->setRawMetadataFile(file_name, getPfmdLen());
    }

    // Everything is fine.
    rc = true;

im_bail:
    // Recycle any resources and release the progress dialog.
    fiProgressDialog::RecycleProgressDialog(progressDialog);

    // All done.
    return rc;
}

/*!
 * @brief We need to load a frame from the drive to be displayed by the video_player object.
 *
 * @param [in]  absFrameNum: The absolute frame number of the frame we want displayed.
 * @param [out] capFrameIndex: The zero-based index into capRawFile for absFrameNum.
 *
 * @return True if prepared to load a frame, false if not.
 */
bool PlaybackCap::prepare_to_frameload(uint32_t absFrameNum, uint32_t& capFrameIndex)
{
    bool isPrepared = false;

    // Set the output to the absFrameNum.  It will be overwritten if we have a CAP stack
    // and that frame resides at a different place in that file.
    capFrameIndex = absFrameNum;

    // It is here that we will handle CAP file spans for CAP stacks.
    // If we are a metamap CAP stack ask it for information.
    if (!metamap.isNull())
    {
        uint32_t capId = FASMOTION_SETTINGS::UNSET_UINT32;
        capRawFile = metamap->GetCAPFileUsingFrame(absFrameNum, capId, capFrameIndex);

        // Check for empty capRawFile, uninitalized capId, and whether or not a change in CAP file has happened.
        if (!capRawFile.isNull() && (capId != FASMOTION_SETTINGS::UNSET_UINT32) && (capId != capIdActive))
        {
            QString pfmdFileName;       //!< Filename for the "metadata.pm" equivalent for this CAP file.

            // CAP file transtion has occurred.  Update our active cap id and make sure supporting objects
            // are initialized to support this new CAP file.
            capIdActive = capId;

            // Each CAP transition the following needs to occur.
            // - The pm_metadata header needs to be read and populated into pmMetaDataInfo object.
            // - The raw pfmd file information must be installed into the PlaybackMetaInfoFFLR object.

            // Refresh the pm_metadata from the new active capRawFile object.
            // This function will validate the new capRawFile and supporting objects.
            if (! refresh_pm_metadata())
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to refresh_pm_metadata, unable to process capId:" << capIdActive << "for frame:" << absFrameNum;
                goto PrepareForFrameLoadExit;
            }

            // Install the new CAP file's pfmd (per-frame-metadata) filename into the
            // Playback Metadata Information object.
            if (metamap->GetPFMDName(capIdActive, pfmdFileName))
            {
                if (pbMetaInfoPtr != nullptr)
                {
                    pbMetaInfoPtr->setRawMetadataFile(pfmdFileName, getPfmdLen());
                }
            }

            // Set our title to the drive URI + camera name.
            if (!pmMetaDataInfo.isNull())
            {
                QString capName;
                if (!metamap->GetCAPName(capIdActive, capName))
                {
                    capName = tr("CapId:%1 Unverified").arg(capIdActive);
                }
                setWindowTitle(QDir::toNativeSeparators(capName) + QString::fromUtf8(" [%1]").arg(pmMetaDataInfo->GetMetaInfo()->camera_name));
            }
        }
    }

    // Check for capRawFile validity.
    if (capRawFile.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The capRawFile is nullptr!?";
        goto PrepareForFrameLoadExit;
    }

    // Do we have the pm_metadata from the capRawFile header?
    if (pmMetaDataInfo.isNull() || (pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        // Try to refresh the pm_metadata.
        if (!refresh_pm_metadata())
        {
            QString capFileName;
            capRawFile->getFileName(capFileName);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA << "For CAP:" << capFileName;
            goto PrepareForFrameLoadExit;
        }
    }

    // Set the 0-based index.
    capFrameIndex = absFrameNum % pmMetaDataInfo->GetMetaInfo()->frame_count;

#if 0
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "absFrameNum:" << absFrameNum << "capFrameIndex:" << capFrameIndex << "capIdActive" << capIdActive
                 << "pmMetaDataInfo->frame_count:" << pmMetaDataInfo->GetMetaInfo()->frame_count
                 << "pmMetaDataInfo->first_frame:" << pmMetaDataInfo->GetMetaInfo()->first_frame;
#endif

    // If make it here we know we are prepared.
    isPrepared = true;

PrepareForFrameLoadExit:
    return isPrepared;
}

/*!
 * @brief We need to load a frame from the drive to be displayed by the video_player object.
 *
 * @param frame_num     The absolute frame number of the frame we want displayed.
 * @param pixmap_ptr    Pixmap to load the frame into.
 * @param force_load    True to clear the cache so the frame is reloaded.
 *
 * @return True on success, false otherwise.
 */
bool PlaybackCap::loadFrame(uint32_t frame_num, fiPixmap *pixmap_ptr, bool force_load)
{
    bool        do_fpn;
    uint32_t    cap_frame_index;

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Sanity check.
    if (pixmap_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "pixmap_ptr is nullptr!";
        return false;
    }

    // Ignore if it's loading the same frame again, and we're not forcing it.
    if ((frame_num == last_frame_loaded) && !force_load)
        return false;

    // Make sure we are prepared to load this frame.
    if (!prepare_to_frameload(frame_num, cap_frame_index))
        return false;

    // If this is a force load, then clear the pixmap cache.
    if (force_load)
    {
        // Clear the cache.
        QPixmapCache::clear();
    }

    // Remember this frame.
    last_frame_loaded = frame_num;

    if (!init_ccm_ctldata(pixmap_ptr, do_fpn))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to initialize CCM control data. Frame_num: " << frame_num;
        return false;
    }

    // Because we support multiple CAP files now with a metamap it is possible that
    // the frame_num is incorrect if we have spanned a CAP boundary. For that reason
    // we use the cap_frame_index calculated when we prepared to load the frame.

    // Read the correct frame from the drive.
    if (!capRawFile->readFrame(cap_frame_index, do_fpn, pixmap_ptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the image. Frame_num: " << frame_num;
        return false;
    }

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Pixmap load:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif

    // All done.
    return true;
}

/*!
 * @brief Save the images as specified by the SaveDialog object.
 * @param parent    Pointer to the [parent](@ref PlaybackCap) object.
 */
void PlaybackCap::SaveImages(QObject *parent)
{
    bool        force8BitMono  = false;     // Set if the frame read needs to be forced 8-bit instead of 16-bit.
    bool        usingBMPFrames = false;     // Can be creating BMP for stacks or AVI containers.
    QString     save_frame_err = tr("Saving error has occurred ");
    FAS_VIDEO::VIDEO_CODEC force8BitMonoCodec   = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;
    FAS_VIDEO::VIDEO_CODEC rawFrameCodec        = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;
    QByteArray  metadata_array;
    pm_metadata metadata_copy;
    quint64     byte_count_stat;
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    double      bytes_per_sec;
    double      mbytes_per_sec;
    QString     stats_msg;
    FI_SAVER::SAVE_PARAMS   save_settings;
    QElapsedTimer   save_timer;
    PlaybackCap  *my_class = qobject_cast<PlaybackCap *>(parent);

    bool saveStatusOK = true;       // If fiSaver object has error catch the state.

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Make sure that we have the pm_metadata.
    if (my_class->pmMetaDataInfo.isNull() || (my_class->pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA;
        return;
    }

    // We need to have a capRawFile.
    if (my_class->capRawFile.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MISSING_RAWFILE;
        return;
    }

    // Get a pointer to the save dialog. Makes things more legible later on.
    std::shared_ptr<SaveDialog> saveDialog = my_class->save_dlg;
    if (!saveDialog)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_SAVE_DIALOG;
        return;
    }

    // Get our settings.
    saveDialog->get_save_params(save_settings, my_class->isSeries3());

    // If creating an MP4 video file special processing is needed to prepare for
    // the save dialog configuration.
    if (save_settings.streamType == FI_SAVER::STREAM_TYPE::MP4)
    {
        // Set the input to TIF and the output to BMP for CAP->MP4 video creation.
        save_settings.imagesInputType = FI_SAVER::IMAGE_TYPE::TIF;
        save_settings.imagesOutputType = FI_SAVER::IMAGE_TYPE::BMP;
    }
    else if (save_settings.streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
    {
        // Saving a CAP file as an RDI packed format, we are going to be using TIFs.
        save_settings.imagesInputType = FI_SAVER::IMAGE_TYPE::TIF_RAW;
    }

    // Allocate our saver object.
    my_class->saver = std::unique_ptr<FiSaver>(new FiSaver());

    // Make sure we have a saver to move forward with.
    if (!my_class->saver)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate saver object!";
        return;
    }

    // Initialize the saver object.
    my_class->saver->configure(save_settings);
    my_class->saver->setPath(save_settings.dest_path);
    my_class->saver->setFileExtension(save_settings.streamType);
    my_class->saver->setStreamType(save_settings.streamType, save_settings.imagesInputType, save_settings.imagesOutputType);
    my_class->saver->setPlaybackRate(save_settings.playback_rate);
    my_class->saver->setH264Speed(save_settings.h264_speed);
    my_class->saver->setH264Quality(save_settings.h264_quality);
    my_class->saver->setH264EnableAllKeyframes(save_settings.h264_enable_all_keyframes);
    my_class->saver->setMetadataCfgSave(save_settings.save_metadata_cfg);
    my_class->saver->setStartIndex(0);
    if (save_settings.tag_label.isEmpty())
        my_class->saver->setTag(QString());
    else
        my_class->saver->setTag(save_settings.tag_label);
    if (save_settings.use_camera_name)
        my_class->saver->setCameraName(my_class->pmMetaDataInfo->GetMetaInfo()->camera_name);

    // Set the clip start to zero, 'cuz we'll be using absolute frame numbers as our sequence numbers.
    my_class->saver->setClipStart(0);

    // Tell the saver the number of frames.
    my_class->saver->setCount((saveDialog->get_clip_end() - saveDialog->get_clip_start()) + 1);
    my_class->saver->setClipEnd(my_class->saver->getFrameCount() - 1);

    // Do we need to merge the overlay information?
    if (save_settings.merge_overlays)
    {
        my_class->video_player->render_start();
    }

    // Set video streaming settings.
    if (FiSaver::isStreamVideo(save_settings.streamType))
    {
        // Set the video stream image width & height.
        if (save_settings.merge_overlays)
        {
            QGraphicsScene  *scene_ptr;

            // Get a pointer to our "scene".
            scene_ptr = my_class->video_player->render_get_scene();
            if (scene_ptr == nullptr)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "scene_ptr is nullptr?!";
                goto save_bail;
            }

            my_class->saver->setImageHeight((uint32_t)scene_ptr->sceneRect().height());
            my_class->saver->setImageWidth((uint32_t)scene_ptr->sceneRect().width());
        }
        else
        {
            my_class->saver->setImageHeight(my_class->pmMetaDataInfo->GetMetaInfo()->roi_height);
            my_class->saver->setImageWidth(my_class->pmMetaDataInfo->GetMetaInfo()->roi_width);
        }

        // Are we saving a packed stream?
        if (save_settings.streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
        {
            const pm_metadata *pm_ptr = nullptr;

            // Do we have our partition metadata structure?
            pm_ptr = my_class->pbMetaInfoPtr->getPmMetadata();
            if (pm_ptr == nullptr)
            {
                // See if we have the metadata from a CAP file.
                if (!my_class->pmMetaDataInfo.isNull())
                {
                    // Try retrieving the metadata pointer.
                    pm_ptr = my_class->pmMetaDataInfo->GetMetaInfo();
                }

                // If we got the metadata pointer, we're done. Otherwise, there
                // is something else we can try.
                if (pm_ptr == nullptr)
                {
                    // Recover the pm_metadata structure.
                    my_class->pbMetaInfoPtr->recoverPmMetadata();
                    pm_ptr = my_class->pbMetaInfoPtr->getPmMetadata();
                }
            }

            // The packed writer needs to know what the starting frame number of the clip is.
            my_class->saver->setClipStart(saveDialog->get_clip_start());

            // Prepare the packed stream writer object.
            if (my_class->saver->preparePackedWriter(pm_ptr) < 0)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "PrepareStreamWriter() failed!";
                goto save_bail;
            }
        }
        else
        {
            // Have it prepare for a video stream writer save.
            if (my_class->saver->prepareStreamWriter() < 0)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "PrepareStreamWriter() failed!";
                goto save_bail;
            }
        }
    }
    else if (my_class->saver->createDestPath() < 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "create_dest_path() failed!";
        goto save_bail;
    }

    // Get our metadata array ready for the raw per-frame information.
    metadata_array.clear();

    // Reset the save complete flags for stream video saving.
    my_class->wait_for_save_complete    = false;
    my_class->save_had_error            = false;
    my_class->save_is_complete          = false;
    my_class->saver_count = my_class->saver->getFrameCount();

    // Set a flag to determine if creating BMP frames for the save process.
    // AVI/BMP videos are considered as well.
    if (save_settings.streamType == FI_SAVER::STREAM_TYPE::BMP_STACK ||
        (save_settings.streamType == FI_SAVER::STREAM_TYPE::AVI && save_settings.imagesOutputType == FI_SAVER::IMAGE_TYPE::BMP))
    {
        usingBMPFrames = true;
    }

    // If we are working with raw image frames we will need to tell the frame reader which one so that the proper
    // image header is applied to the resultant frame.
    if (save_settings.streamType == FI_SAVER::STREAM_TYPE::TIF_RAW_STACK)
    {
        rawFrameCodec = FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC;
    }
    else if (save_settings.streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
    {
        if (my_class->isSeries3())
            rawFrameCodec = FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC;
        else
            rawFrameCodec = FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC;
    }
    else if (save_settings.streamType == FI_SAVER::STREAM_TYPE::DNG_STACK)
    {
        rawFrameCodec = FAS_VIDEO::VIDEO_CODEC::FI_DNG_STACK_CODEC;
    }

    // If saving BMP or TIF images and not merging overlays and the capture
    // data is from a monochrome sensor we will force the save operation to
    // produce 8-bpp frame data.
    // NOTE:  FI_SAVER::STREAM_TYPE::TIF_STACK is NOT FI_SAVER::STREAM_TYPE::TIF_RAW_STACK
    if (usingBMPFrames || (save_settings.streamType == FI_SAVER::STREAM_TYPE::TIF_STACK))
    {
        // Only force if not merging overlays.
        if (!save_settings.merge_overlays)
        {
            if ((my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX13HS_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX160_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX2100_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO))
            {
                // Setup to able to notify the frame reader what type we want...
                if (usingBMPFrames)
                {
                    force8BitMonoCodec = FAS_VIDEO::VIDEO_CODEC::FI_BMP_STACK_CODEC;
                }
                else
                {
                    force8BitMonoCodec = FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC;
                }

                // We only want to force an 8-bit monochrome shrinkage if our captured frames > 8 Bit Mode Capture.
                force8BitMono = !IS_8_BIT_MODE((char)my_class->pmMetaDataInfo->GetMetaInfo()->bit_sel);
            }
        }
    }

    // Initialize our values for our statistics.
    byte_count_stat = 0;
    save_timer.start();

    // Loop through all the frames. Make sure we're using absolute (not relative) frame numbers.
    for (my_class->saver_frame = saveDialog->get_clip_start();
         !my_class->thread_stop && (my_class->saver_frame <= saveDialog->get_clip_end()) && saveStatusOK; ++my_class->saver_frame)
    {
        FAS_VIDEO::VIDEO_CODEC fvCodec;
        QByteArray  my_buffer;
        QByteArray  *pfmd_ptr;
        QString     saveStatusOK_error_str;
        fiPixmap    my_pixmap;
        uint32_t    my_seqNbr;
        uint32_t    cap_frame_index;

        bool do_fpn;

        // Make sure we are prepared to load this frame.
        if (!my_class->prepare_to_frameload(my_class->saver_frame, cap_frame_index))
        {
            saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Preparation to read frame failed."));
            saveStatusOK = false;   // Will terminate the saver processing loop.
            goto saveStatusOK_check_error;   // Stop processing and loop around.
        }

        // Handle things differently for packed data.
        if (save_settings.streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
        {
            QImage my_packedData;

            // For packed data, the rawFrameCodec needs to be to be FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC.
            // This way, we can easily read the frame from the CAP file as a raw TIFF and feed that to the
            // RDI streamer directly.
            if (rawFrameCodec != FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "rawFrameCode was not FI_TIFF_STACK_CODEC! Instead:" << rawFrameCodec;
                rawFrameCodec = FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC;
            }
            fvCodec = rawFrameCodec;

            // Load the frame. Make sure it's the "raw" data, so we need to force the load.
            if (!my_class->capRawFile->readFrame(cap_frame_index, false, nullptr, false, &fvCodec, &my_buffer))
            {
                saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Frame read error."));
                saveStatusOK = false;   // Will terminate the saver processing loop.
                goto saveStatusOK_check_error;   // Stop processing and loop around.
            }

            // Load the data into our QImage.
            my_packedData.loadFromData(my_buffer, "TIF");

            // Accumulate the byte count.
            byte_count_stat += (quint64)my_packedData.sizeInBytes();

            // Set the sequence number. For packed data the sequence starts at zero.
            my_seqNbr = my_class->saver_frame - saveDialog->get_clip_start();

            // Save the data.
            // However if it fails we will still process through the images
            // but the update progress will display error condition and the
            // fact that nothing is being updated should cause user to cancel.
            if (!my_class->saver->saveToPackedStream(my_packedData.constBits(), my_packedData.sizeInBytes(), my_seqNbr, false))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "saveToPackedStream() failed! seqNbr:" << my_seqNbr << "frame:" << my_class->saver_frame;
            }

            // Reset our buffer.
            my_buffer.clear();
        }
        else
        {
            // With color processing we need to install information in the saver frame for Fastec CCM Processing.
            // NOTE:  The parameter "do_fpn" is set to the CAP file's metadata value for FPN correction as a result
            // of this call.  If set in the CAP file's metadata field fpn_status by looking at the GIGE_FPN_FPGA_ENABLE_MASK
            // mask then that fixed pattern noise setting is maintained while loading the frame.
            if (!my_class->init_ccm_ctldata(&my_pixmap, do_fpn))
            {
                saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("CCM initialization failed."));
                saveStatusOK = false;   // Will terminate the saver processing loop.
                goto saveStatusOK_check_error;   // Stop processing and loop around.
            }

            // There are two conditions to check for when telling the frame reader which type of header
            // to provide (this is accomplished with the fvCodec setting).
            // First check for a raw frame codec.  This will differentiate between RAW TIFF and DNG.
            if (rawFrameCodec != FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC)
            {
                fvCodec = rawFrameCodec;
            }
            // If not saving in a raw frame format then check for a monochrome 8-bit request.
            // This can be either a BMP or a TIFF (non-raw) frame being saved.
            else if (force8BitMonoCodec != FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC)
            {
                fvCodec = force8BitMonoCodec;
            }

            // Load the frame. Make sure it's the "raw" data, so we need to force the load.
            if (!my_class->capRawFile->readFrame(cap_frame_index, do_fpn, my_class->r_lut, force8BitMono, &fvCodec, &my_buffer))
            {
                saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Frame read error."));
                saveStatusOK = false;   // Will terminate the saver processing loop.
                goto saveStatusOK_check_error;   // Stop processing and loop around.
            }

            // How are we saving this frame.
            switch (save_settings.streamType)
            {
            case FI_SAVER::STREAM_TYPE::BMP_STACK:
            case FI_SAVER::STREAM_TYPE::TIF_STACK:
            case FI_SAVER::STREAM_TYPE::AVI:
                // If the image was modified nothing to do it is ready for processing.
                if (force8BitMono)
                {
                    break;
                }
                // The else condition will fall through to the normal handling.
                [[fallthrough]];
            case FI_SAVER::STREAM_TYPE::MP4:
            case FI_SAVER::STREAM_TYPE::JPEG_STACK:
                // Load these into a pixmap.
                if (!my_pixmap.load_from_buf(my_buffer, fvCodec))
                {
                    saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Frame image loading error."));
                    saveStatusOK = false;   // Will terminate the saver processing loop.
                    goto saveStatusOK_check_error;   // Stop processing and loop around.
                }

                // Pass the image through the lookup table.
                my_pixmap.apply_lookup(my_class->r_lut, my_class->g_lut, my_class->b_lut);

                // Reset our buffer.
                my_buffer.clear();

                // Do we need to merge the overlay information?
                if (save_settings.merge_overlays)
                {
                    my_pixmap = my_class->mergeOverlay(my_class->saver_frame, my_pixmap);
                }

                // Update the quality factor. This will only be used if we're saving JPEGs.
                my_pixmap.set_jpeg_quality(save_settings.jpeg_quality);

                // Get the pixmap as a qImage.
                my_pixmap.save_image(&my_buffer, save_settings.imagesOutputType);

                // Make sure that the transcoded frame has a frame size > 0.
                if (my_buffer.length() <= 0)
                {
                    // Something went wrong during the transcode operation.  Warn the user.
                    saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Frame transcoding error."));
                    saveStatusOK = false;   // Will terminate the saver processing loop.
                    goto saveStatusOK_check_error;   // Stop processing and loop around.
                }
                break;
            case FI_SAVER::STREAM_TYPE::DNG_STACK:
            case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
                // Nothing to do my_buffer is already set up for this operation.
                break;
            default:
                break;
            }

            // Accumulate the byte count.
            byte_count_stat += (quint64)my_buffer.length();

            // Set the sequence number.
            my_seqNbr = my_class->saver_frame + 1;

            // Save the image.
            // However if it fails we will still process through the images
            // but the update progress will display error condition and the
            // fact that nothing is being updated should cause user to cancel.
            if (!my_class->saver->saveToStream((const uint8_t *)my_buffer.data(), (uint32_t)my_buffer.length(), my_seqNbr, false))
            {
                saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Save to stream error."));
                saveStatusOK = false;   // Will terminate the saver processing loop.
                goto saveStatusOK_check_error;   // Stop processing and loop around.
            }

            // Get the frame's raw metadata. Note that regardless of series3 or not, we don't have to byte swap.
            // NOTE: we cannot pass the GIGE_FRAME_METADATA_LEN here because it produces too small of a per frame
            // metadata file.
            pfmd_ptr = my_class->pbMetaInfoPtr->getRawPFMD(my_class->saver_frame, false);
            if (pfmd_ptr != nullptr)
            {
                // Append this frame's metadata.
                metadata_array.append(*pfmd_ptr);

                // Release the pointer.
                delete pfmd_ptr;
            }
saveStatusOK_check_error:
            // If the saveStatusOK is not ok then tell the user about it.
            if (!saveStatusOK)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << saveStatusOK_error_str;
                // We don't have to worry about timers and stream timeouts so just popup a message box
                // that tells the user what happened.
                //
                // NOTE:  In addition the xferDialog will close right after the user acknowledges this
                // as well so no need to put any information in it either.  Because this is a thread we
                // don't want to be modifying the xferDialog that is being updated by a polling timer
                // running in our base class outside of this thread.
                QMessageBox::warning(my_class, "Error Saving File", saveStatusOK_error_str);
            }
        }
    }
    //-------------------------------------------------------------
    // End of the frame processing scope.  This very long for loop!
    //-------------------------------------------------------------

    // If saving a stream video there could be outstanding frame flushes that exist and
    // progress reporting that still is happening (assuming no error has occurred).
    if (my_class->saver->isStreamVideo(save_settings.streamType))
    {
        my_class->wait_for_save_complete = true;

        // If an error had occurred above then set the flag condition.
        if (!saveStatusOK)
        {
            my_class->save_had_error = true;
        }

        while (!my_class->thread_stop && !my_class->save_is_complete)
        {
            // If the saver does not show an error condition and the
            // we are still waiting for save_is_complete pause our
            // thread.
            if (!my_class->save_had_error)
            {
                my_class->saver_thread.suspend();
            }
        }
    }

    // Do we need to clean overlay stuff up?
    if (save_settings.merge_overlays)
        my_class->video_player->render_end();

    // Get the elapsed time and generate statistics.
    elapsed = save_timer.elapsed();
    elapsed_time = elapsed_time.addMSecs((int)elapsed);

    // How many bytes per second.
    bytes_per_sec = byte_count_stat / ((double)elapsed / 1000.0);
    mbytes_per_sec = (bytes_per_sec / (1024.0 * 1024.0));

    // Build up our output string.
    stats_msg = QString::fromUtf8("Duration: %1  TotalBytes: %2  TotalFrames: %3  EffectiveRate: %4 MB/s  Format: %5")
                .arg(elapsed_time.toString(QString("hh:mm:ss.zzz"))).arg(byte_count_stat).arg((saveDialog->get_clip_end() - saveDialog->get_clip_start()) + 1).
                arg(mbytes_per_sec).arg(my_class->saver->streamTypeToQString(my_class->saver->getStreamType(), false));

    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");
    QLOG_INFO() << __PRETTY_FUNCTION__ << my_class->saver->getLocalPath();
    QLOG_INFO() << __PRETTY_FUNCTION__ << stats_msg;
    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");

    // The packed stream does not have a separate metadata text file.
    if (save_settings.streamType != FI_SAVER::STREAM_TYPE::RDI_PACKED)
    {
        // Save the capture metadata. Start off with a copy.
        memcpy(&metadata_copy, my_class->pmMetaDataInfo->GetMetaInfo(), sizeof(pm_metadata));

        // Was there a comment from the user?
        if (!save_settings.comment.isEmpty())
            memcpy(metadata_copy.comment, save_settings.comment.toUtf8().constData(), GIGE_COMMENT_MAX_LEN);

        // Save the raw metadata.
        if (!metadata_array.isEmpty())
            my_class->saver->saveRawMetadata(&metadata_array, &metadata_copy, ((uint32_t)(my_class->r_gain * 256.0)), ((uint32_t)(my_class->g_gain * 256.0)), ((uint32_t)(my_class->b_gain * 256.0)));

        // Set the clip start value back to the actual start frame number. The metadata saving
        // code needs to know this information.
        my_class->saver->setClipStart(saveDialog->get_clip_start());
        my_class->saver->setClipEnd(saveDialog->get_clip_end());

        // Put the capture metadata into a byte array and have the saver object save it.
        metadata_array.clear();
        metadata_array.setRawData((const char *)&metadata_copy, sizeof(metadata_copy));
        my_class->saver->savePartitionMetadata(&metadata_array);
    }

    // Wrap up the save.
    my_class->saver->finishSave();

save_bail:
    // We are done and leaving let our class know.
    emit my_class->saveImagesComplete(saveStatusOK);
}


/*!
 * @brief Refreshes the pm_metadata information for the capRawFile.
 *
 * @return True on success and pmMetaDataInfo is created along with custom lut
 * processing applied, false otherwise.
 */
bool PlaybackCap::refresh_pm_metadata()
{
    bool isRefreshed = false;

    // We need to have a capRawFile.
    if (capRawFile.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MISSING_RAWFILE;
        goto refresh_pm_metadata_exit;
    }

    // Must have the Playback Metadata Information pointer allocated as well.
    if (pbMetaInfoPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "There is no Playback Metadata Information pointer allocated!";
        goto refresh_pm_metadata_exit;
    }

    // Refresh our pm_metadata information from the beginning of new capRawFile.
    pmMetaDataInfo = CaptureMetaInfo::CreateCaptureMetaInfo(capRawFile);

    // Did we refresh the pm_metadata from the capRawFile header?
    if (pmMetaDataInfo.isNull() || (pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        QString capFileName;
        capRawFile->getFileName(capFileName);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA << "For CAP:" << capFileName;
        goto refresh_pm_metadata_exit;
    }

    // Process any custom LUT for the pm_metadata that has been refreshed.
    if (pmMetaDataInfo->GetMetaInfo()->lut_id > 0)
    {
        bool isMono = pbMetaInfoPtr->isColor() ? false : true;
        bool isSeries3 = pbMetaInfoPtr->isSeries3();

        QVector<uint16_t> lutKnots;
        int knotDataSize = ((LUT_MAX_KNOT_COUNT * 2) * 3);

        // Make sure the knots data is large enough to hold the metadata lut_knots.
        lutKnots.resize(knotDataSize);

        // Just copy over the knots to a safe QVector.
        for (int knotId = 0; knotId < knotDataSize; knotId++)
        {
            lutKnots[knotId] = pmMetaDataInfo->GetMetaInfo()->lut_knots[knotId];
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[knot " << knotId << ":" << pmMetadataPtr->lut_knots[knotId] << "]";
        }
        // Create the cap file's custom luts.
        // Note:  This builds the 10-bit LUT vectors so they are 1024 bytes in length.
        capLuts = FiDisplayLut::CreateDisplayLut(isMono, isSeries3, pmMetaDataInfo->GetMetaInfo()->lut_id, lutKnots);

        // If the cap file custom luts were created set up the display luts here.
        if (capLuts)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FiDisplayLut::CreateDisplayLut() created capLuts isUsingColor():" << capLuts->IsUsingColor();

            if (capLuts->IsUsingColor())
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Installing CAP custom RGB luts.";
                bool builtRedOrGammaLuts    = false;
                bool builtGreenLuts         = false;
                bool builtBlueLuts          = false;
                bool installedRGBLuts       = false;

                // Right now the capLuts contains the 10-bit LUT vectors.  There is a new function
                // that can take in the 10-bit vectors for the rgb lut vectors and it subsamples
                // the 10-bit data turning it into the 8-bit LUT data needed by the playback engine.
                // Ticket #259
                installedRGBLuts = set10BitRgbLuts(capLuts->GetRedOrGammaCameraLut(builtRedOrGammaLuts),
                                                   capLuts->GetGreenCameraLut(builtGreenLuts),
                                                   capLuts->GetBlueCameraLut(builtBlueLuts));

                // Warn if there is something wrong with our luts.
                if (!builtRedOrGammaLuts || !builtGreenLuts || !builtBlueLuts || !installedRGBLuts)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The custom LUT data was not set [RBuilt:"
                                 << builtRedOrGammaLuts << "][GBuilt:" << builtGreenLuts << "][BBuilt:"
                                 << builtBlueLuts << "][LUT installed:" << installedRGBLuts << "]";
                }
            }
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Installing CAP custom gamma luts.";
                bool builtGammaLuts     = false;
                bool installed_gamma    = false;

                // Right now the capLuts contains the 10-bit LUT vectors.  There is a new function
                // that can take in the 10-bit vectors for the rgb lut vectors and it subsamples
                // the 10-bit data turning it into the 8-bit LUT data needed by the playback engine.
                // Ticket #259
                installed_gamma = set10BitGammaLut(capLuts->GetRedOrGammaCameraLut(builtGammaLuts));

                // Warn if there is something wrong with our luts.
                if (!builtGammaLuts || !installed_gamma)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The custom LUT data was not read properly [built:" << builtGammaLuts << "][installed:" << installed_gamma << "]";
                }
            }
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to FiDisplayLut::CreateDisplayLut()!";
        }
    }

    isRefreshed = true;

refresh_pm_metadata_exit:
    return isRefreshed;
}

/*!
 * @brief Retrieve the capture metadata (pm_metadata) from the drive and save it to
 *        the specified file.
 *
 * @param filename  Path for the pm_metadata.
 * @return True on success, false otherwise.
 */
bool PlaybackCap::save_pm_metadata(const QString& filename)
{
    QFile       meta_file;
    QString     meta_text;
    QByteArray  meta_buffer;
    SaveVideo   saver;
    FAS_VIDEO::VIDEO_GAINS gainsData;
    const FAS_VIDEO::VIDEO_GAINS *gainsDataPtr = &gainsData;

    // We need to have a capRawFile.
    if (capRawFile.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MISSING_RAWFILE;
        return false;
    }

    // Do we have the pm_metadata from the capRawFile header?
    if (pmMetaDataInfo.isNull() || (pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        // Refresh our pm_metadata information from the beginning of the capRawFile.
        if (!refresh_pm_metadata())
        {
            QString capFileName;
            capRawFile->getFileName(capFileName);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA << "For CAP:" << capFileName;
            return false;
        }
    }

    // Make a buffer from the structure.
    meta_buffer.setRawData((const char *)pmMetaDataInfo->GetMetaInfo(), sizeof(pm_metadata));

    // Build our gain information structure.
    gainsData.red = (uint32_t)(r_gain * 256.0);
    gainsData.green = (uint32_t)(g_gain * 256.0);
    gainsData.blue = (uint32_t)(b_gain * 256.0);

    // Convert the metadata to a text format.
    if (pmMetaDataInfo->GetMetaInfo()->skip_count > 0)
        meta_text = saver.metadataToText(&meta_buffer, FI_SAVER::STREAM_TYPE::FI_CAP, FI_SAVER::IMAGE_TYPE::NO_IMAGE, pmMetaDataInfo->GetMetaInfo()->first_frame, pmMetaDataInfo->GetMetaInfo()->first_frame + capRawFile->getFrameCount() - 1, gainsDataPtr);
    else
        meta_text = saver.metadataToText(&meta_buffer, FI_SAVER::STREAM_TYPE::FI_CAP, FI_SAVER::IMAGE_TYPE::NO_IMAGE, 0, capRawFile->getFrameCount() - 1, gainsDataPtr);

    // Make sure the file doesn't already exist.
    QFile::remove(filename);
    meta_file.setFileName(filename);

    // Open the file for writing.
    if (!meta_file.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << filename << "Err:" << meta_file.errorString();
        return false;
    }
    meta_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
    meta_file.write(meta_text.toUtf8());
    meta_file.close();

    return true;
}

/*!
 * @brief Retrieve the per-frame metadata (raw metadata) from the raw cap file and save it to
 *        the specified file.
 *
 * @param filename  Path for the per-frame metadata.
 * @return True on success, false otherwise.
 */
bool PlaybackCap::save_raw_metadata(const QString& filename)
{
    uint32_t    meta_len;
    QFile       meta_file;
    QByteArray  meta_buffer;

    // Yield some time.
    QApplication::processEvents();

    if (capRawFile.isNull())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << QString("No opened cap file exists!");
        return false;
    }

    // Read the capture metadata.
    meta_len = capRawFile->readPerFrameMetadata(meta_buffer);
    if (meta_len == 0)
    {
        QMessageBox::critical(this, tr("CAP File"), tr("No images were found in the CAP file: %1").arg(root_file));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << root_file << ": Unable to read the per-frame metadata from the drive.";
        return false;
    }

    // Make sure the file doesn't already exist.
    QFile::remove(filename);
    meta_file.setFileName(filename);

    // Open the file for writing.
    if (!meta_file.open(QIODevice::WriteOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << filename << "Err:" << meta_file.errorString();
        return false;
    }
    meta_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
    meta_file.write(meta_buffer.constData(), meta_len);
    meta_file.close();

    return true;
}

/*!
 * @brief In our case, this method is used to specify the file URI for where the
 *        raw captured video is located. If the capture is split across multiple
 *        files, this is the URI for the "root", or first, file.
 *
 * @param filename  The device URI to the Fastec (root) CAP file.
 */
bool PlaybackCap::setFile(const QString& cap_filename)
{
    bool        rc = false;                     // Assume return code is false until reach the end.
    QFileInfo   cap_fileinfo(cap_filename);     //!< Installed with user provided cap filename.

    PlaybackMetaInfoFFLR *playbackMetaInfoFFLR = nullptr;   // The metadata helper class for playback engine.

    // Make sure there is no left-over playback meta information object.
    if (pbMetaInfoPtr != nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "pbMetaInfoPtr is not null!";
        delete pbMetaInfoPtr;
    }
    pbMetaInfoPtr = nullptr;

    // Make sure there is no left-over capRawFile object.
    capRawFile.reset();

    // Save the file name for later.
    root_file = cap_filename;

    // Make sure we are working with a CAP file.
    file_ext = cap_fileinfo.suffix().toLower();
    if (file_ext.compare(FiSaver::IMAGE_EXT_CAPTURE, Qt::CaseInsensitive) != 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "This function is for processing CAP files!";
        goto setFileExit;
    }

    // Build up the smart metamap object which will set up cap playback dependending on
    // the type of CAP file whether it's a stand alone one or a member of a CAP stack.
    metamap = FiMetaMap::CreateMetaMap(cap_fileinfo.path());

    // If working with cap stacks, create the special FasFireLR meta information object and
    // initialize it in preparation for playback operations.
    if (!metamap.isNull())
    {
        // Create a FasFireLR playbackMetaInfo object so that it can provide information to
        // the existing playback codebase providing in a seamless-as-possible manner playback
        // support across the CAP file spans.
        playbackMetaInfoFFLR = new PlaybackMetaInfoFFLR(this);
        if (playbackMetaInfoFFLR == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a playbackMetaInfoFFLR object!";
            goto setFileExit;
        }

        // Install our metamap into the playback meta info object which prepares it in the process.
        if (playbackMetaInfoFFLR->installMetaMap(metamap))
        {
            // Assign the base class pointer to our base class object as it "should"
            // operate by calling the virtual functions that handle FFLR CAP file spans.
            pbMetaInfoPtr = qobject_cast<PlaybackMetaInfo *>(playbackMetaInfoFFLR);

            // Prepare the rest for playing the first playable (fp) frame.
            uint32_t fpFrameNum;
            uint32_t fpFrameCapId;
            uint32_t fpCapFrameIndex;

            if (!metamap->GetFirstPlayableFrame(fpFrameNum, fpFrameCapId))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to locate first playable frame for" << cap_filename;
                goto setFileExit;
            }

            // Set up for the first playable frame according to the metamap.
            if (!prepare_to_frameload(fpFrameNum, fpCapFrameIndex))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to prepare_to_frameload for:" << cap_filename << "frame:" << fpFrameNum;
                goto setFileExit;
            }
        }
        else
        {
            // Unable to operate as a CAP stack.  Let it fall into the normal cap playback
            // to attempt to open this as a single CAP file.
            delete playbackMetaInfoFFLR;
            playbackMetaInfoFFLR = nullptr;
            metamap.reset();
        }
    }

    // Check to see if we have a metamap.  If we don't then either the raw cap file presented is a stand-along
    // raw cap file or the cap stack was not validated and we fallback to opening the single raw cap file.
    if (metamap.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Opening file:" << cap_filename << "for single CAP playback operation.";
        // This is a single CAP file presented for playback.
        // Allocate our raw cap file object.
        capRawFile = QSharedPointer<fiRawFile>(new fiRawFile(this));
        if (!capRawFile)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a fiRawFile for" << cap_filename;
            goto setFileExit;
        }

        // Make sure we can open the CAP file.
        if (!capRawFile->open(cap_filename))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open the cap file:" << cap_filename;
            goto setFileExit;
        }

        // Initialize our metadata structures.
        if (!init_metadata())
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to initialize the metadata structures!";
            goto setFileExit;
        }

        // Set our title to the drive URI + camera name.
        setWindowTitle(QDir::toNativeSeparators(cap_filename) + QString::fromUtf8(" [%1]").arg(pmMetaDataInfo->GetMetaInfo()->camera_name));
    }

    // Save the file URI.
    root_file = cap_filename;

    // Tell the main window about our widget.
    MainWindow::GetInstance()->set_metadata_info(pbMetaInfoPtr);

    // Remember the number of frames in the video.
    frame_count = pbMetaInfoPtr->getFrameCount();

    // Make sure there IS a review controls object.
    if (review_controls == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The review_controls object was not allocated!";
        goto setFileExit;
    }

    // Setup our review controls.
    review_controls->set_pb_trigger(pbMetaInfoPtr->getTriggerFrame());
    review_controls->set_pb_clip_start(pbMetaInfoPtr->getStartFrame());
    review_controls->set_pb_clip_end(pbMetaInfoPtr->getEndFrame());
    review_controls->set_pb_frame_rate(pbMetaInfoPtr->getFramerate());

    // Make sure there IS a camera information object.
    if (camera_info == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The camera_info object was not allocated!";
        goto setFileExit;
    }

    // Update the summary camera information.
    camera_info->set_summary(pbMetaInfoPtr->getWidth(), pbMetaInfoPtr->getHeight(), pbMetaInfoPtr->getFramerate(),
                             pbMetaInfoPtr->getExposure(), pbMetaInfoPtr->getTriggerFrame() | GIGE_TRIGGER_REQ_FRAME_MASK, pbMetaInfoPtr->getSensorGain());

    // Save away some of the information for later.
    review_controls->set_pb_frame_count(frame_count);

    // Enable the playback buttons.
    review_controls->update_buttons(true);

    // Allocate an action for the window's menu to use.
    if (playback_action != nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "playback_action is not null!";
        delete playback_action;
    }
    playback_action = new QAction(this);

    // Make sure the playback_action was allocated.
    if (playback_action == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate playback_action!";
        goto setFileExit;
    }

    playback_action->setCheckable(true);
    connect(playback_action, &QAction::triggered, this, [this]() { show(); });
    connect(playback_action, &QAction::triggered, this, [this]() { setFocus(); });
    playback_action->setText(windowTitle());

    // Jump to the trigger frame.
    review_controls->set_current_frame_num(pbMetaInfoPtr->getTriggerFrame());

    // Turn off the autosave settings.
    camera_info->display_autosave_settings(0);

    // Show/hide the display info.
    MainWindow::GetInstance()->get_view_summary()->setChecked(camera_info->summary_is_enabled());
    MainWindow::GetInstance()->get_view_pfmd()->setChecked(camera_info->get_pfmd_setting());
    MainWindow::GetInstance()->get_view_local_frametime()->setChecked(camera_info->localtime_is_enabled());
    camera_info->show_info();

    // If we're enabling the display, make sure we grab the metadata for the current frame.
    if (camera_info->pfmd_is_enabled())
    {
        // Update the per-frame metadata display.
        updatePfmd();

        // Update our event information.
        updateEventInfo();
    }


    // Make sure the video player was allocated.
    if (video_player == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The video_player object was not allocated!";
        goto setFileExit;
    }

    // Allow overlays
    video_player->enable_overlay(true);

    // Make sure our gain information is set appropriately.
    updateGainInfo();

    if (!pmMetaDataInfo.isNull() && (pmMetaDataInfo->GetMetaInfo() != nullptr))
    {
        if ((pmMetaDataInfo->GetMetaInfo()->fpn_status & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FPN ENABLED";
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FPN DISABLED";
        }
    }

    // If execution makes it here the setFile was processed without error.
    rc = true;

setFileExit:

    // If we did not make it here with a valid return code we have some cleanup to do.
    if (!rc)
    {
        // Cleanup any metamap object created.
        if (!metamap.isNull())
        {
            metamap.reset();
        }

        // Cleanup any created playback metadata information object.
        if (playbackMetaInfoFFLR != nullptr)
        {
            delete playbackMetaInfoFFLR;
            // If we had an FFLR one then the pbMetaInfoPtr is a cast pointer to it.
            pbMetaInfoPtr = nullptr;
        }
        else if (pbMetaInfoPtr)
        {
            delete pbMetaInfoPtr;
            pbMetaInfoPtr = nullptr;
        }

        // Close and deallocate any CAP file opened.
        if (!capRawFile.isNull())
        {
            // This will call the destructor and close any opened file.
            capRawFile.reset();
        }
    }

    // All done.
    return rc;
}

/*!
 * @brief We are being made inactive. If we are actively streaming, we need to slow the rate down by
 *        switching to background mode. In addition, we do NOT want to respond to certain signals.
 */
void PlaybackCap::setInactive()
{
    // Pass this to our parent class.
    PlaybackMDI::setInactive();
}

bool PlaybackCap::set_vtrig_marker(int index, uint32_t& event_word, uint32_t *compare_ptr, uint32_t vtrig_word, uint32_t last_vtrig_word)
{
    bool changed;
    uint32_t vtrig_flags;
    uint32_t last_flags;

    // Assume we will not make any changes.
    changed = false;

    // Get our flags "normalized".
    vtrig_flags = (vtrig_word & CAMERA_INFO::VTRIG_FLAGS_MASK) >> CAMERA_INFO::VTRIG_FLAGS_SHIFT;
    last_flags = (last_vtrig_word & CAMERA_INFO::VTRIG_FLAGS_MASK) >> CAMERA_INFO::VTRIG_FLAGS_SHIFT;

    // If there is a refresh event, we'd like to show that as a glitch.
    if ((vtrig_flags & CAMERA_INFO::VTRIG_FLAGS_REFRESH) == CAMERA_INFO::VTRIG_FLAGS_REFRESH)
    {
        // Mark this.
        event_word |= (1 << (FM_EVENT_FILTERS::EDGE_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));
    }

    // Look to see if we had a video trigger event.
    if ((vtrig_flags & CAMERA_INFO::VTRIG_FLAGS_TRIGGER) == CAMERA_INFO::VTRIG_FLAGS_TRIGGER)
    {
        // We have a trigger event, so set the PFMD event data.
        event_word |= ((uint32_t)1 << (FM_EVENT_FILTERS::EVENT_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));

        // Are we at the beginning of our search?
        if (index == 0)
        {
            // Is there a compare word?
            if (compare_ptr != nullptr)
            {
                // OR in the VTRIG marker into the compare_event.
                *compare_ptr |= ((uint32_t)1 << (FM_EVENT_FILTERS::EVENT_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));
            }
        }
        else
        {
            // We should have a last trigger value, so see if we should set an edge. We have a trigger
            // in the current frame, so if we didn't have one in the previous frame there is an edge.
            if ((last_flags & GIGE_VTRIG_CONTROL_TRIGGER) != GIGE_VTRIG_CONTROL_TRIGGER)
            {
                // There is a "rising" edge here.
                event_word |= (1 << (FM_EVENT_FILTERS::EDGE_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));
            }
        }

        // Remember that we've changed the PFMD event data.
        changed = true;
    }
    else if (index > 0)
    {
        // Check to see if we just lost the trigger. We do NOT have it in this frame, but did we
        // have it in the previous frame?
        if ((last_flags & GIGE_VTRIG_CONTROL_TRIGGER) == GIGE_VTRIG_CONTROL_TRIGGER)
        {
            // There is a "falling" edge here.
            event_word |= (1 << (FM_EVENT_FILTERS::EDGE_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));

            // Remember that we've changed the PFMD event data.
            changed = true;
        }
    }

    // Return our changed flag.
    return changed;
}


/*!
 * @brief Extract the event marker information from the raw metadata file.
 * Note:
 * This routine will process the recording in chunk intervals allowing other threads
 * to get a slice of processing time. This method must be tolerant of the re-entrant
 * nature associated with this implmentation. It is not recursively re-entrant but through
 * the use of a lambda and a timer it is re-entrant after a time interval. Because it is
 * not recursively re-entrant the local stack state is not maintained and cannot be counted
 * on to have state from the last interval processed.
 *
 * @param starting_frame_num    The frame number we're to start with.
 * @param frame_count           The number of frames we're supposed to process.
 *
 * @return True if the entire file has been processed (aka done), false if there's more to do.
 */
bool PlaybackCap::getRawMarkers(uint32_t starting_frame_num, uint32_t frame_count, bool do_swap)
{
    uint32_t last_marker_word(0);
    uint32_t frames_processed(0);
    uint32_t last_vtrig_pfmd(0);
    const QByteArray    *raw_data = nullptr;
    const uint8_t       *data_ptr = nullptr;

    // Force the swap.
    do_swap = true;

    // Sanity check.
    if (!pbMetaInfoPtr)
        return true;
    if (!pbMetaInfoPtr->captureMetadataIsAvail())
        return true;

    // Process for CAP stack is different than for a stand-alone CAP file.
    if (!metamap.isNull())
    {
        bool readPFMD = true;

        uint32_t startCapId;
        uint32_t endCapId;
        uint32_t firstPlayableFrame;
        uint32_t firstCapId;
        uint32_t lastPlayableFrame;
        uint32_t lastFrameNum = starting_frame_num + frame_count;
        uint32_t totalFramesInEntireVideo;
        uint32_t capAbsFrameNum;

        // Have we read all of the frame data in the entire video?
        metamap->GetTotalFrameCount(totalFramesInEntireVideo);

        // Get the starting frame number of the video.
        metamap->GetFirstPlayableFrame(firstPlayableFrame, firstCapId);

        // Get the last frame number of the video.
        lastPlayableFrame = totalFramesInEntireVideo - 1;

        // If we have processed all of the frames in the video we are done here!
        if (starting_frame_num >= totalFramesInEntireVideo)
        {
            // We are done here.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "starting_frame_num:" << starting_frame_num << "is bigger than total frames in video:" << totalFramesInEntireVideo;
            return true;
        }

        // Make sure we don't try to read beyond our last frame number.
        if (lastFrameNum > lastPlayableFrame)
        {
            lastFrameNum = lastPlayableFrame;
        }

        // Find the starting cap ID for the starting_frame_num provided.
        if (!metamap->GetCapInfoForFrame(starting_frame_num, startCapId, capAbsFrameNum))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to locate capId for starting_frame_num:" << starting_frame_num;
            return true;
        }

        // Locate the capId for the starting_frame_num + frame_count.
        if (!metamap->GetCapInfoForFrame(lastFrameNum, endCapId, capAbsFrameNum))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to locate capId for last frame:" << lastFrameNum;
            return true;
        }

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "capActiveId:" << capIdActive
                     << "startCapId:" << startCapId << "starting_frame_num:" << starting_frame_num << "frame_count:" << frame_count << "do_swap:" << do_swap;

        // Is this the first time through the event search?
        if (starting_frame_num == 0)
        {
            // We are starting over for our pfmd accumulation into a single QByteArray.
            capEventsPFMD.clear();
            capEventsPFMDLastCapId = FASMOTION_SETTINGS::UNSET_UINT32;
        }
        else
        {
            bool seemsOkToMe = true;

            // Make some simple sanity checks.
            if (capEventsPFMDLastCapId == FASMOTION_SETTINGS::UNSET_UINT32)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Searching starting_frame_num:" << starting_frame_num << "but capEventsPFMDLastCapId has UNSET_UINT32 value.";
                seemsOkToMe = false;
            }
            else
            {
                // Based on the starting_frame_num-1 does it appear that the number of bytes in our
                // capEventsPFMD match what we might expect for that frame?
                uint32_t framesSoFar = starting_frame_num - 1;
                uint32_t pfmdSizeExpected = framesSoFar * getPfmdLen();

                if (capEventsPFMD.length() < pfmdSizeExpected)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Searching starting_frame_num:" << starting_frame_num << "but capEventsPFMD.length():"
                                 << capEventsPFMD.length() << "< pfmdSizeExpected:" << pfmdSizeExpected;
                    seemsOkToMe = false;
                }
            }

            // If it doesn't seem ok don't continue.
            if (!seemsOkToMe)
                return true;
        }

        // The idea here is to just build up the QByteArray for the range of CAP ids.
        for (uint32_t capId = startCapId; readPFMD && (capId <= endCapId); ++capId)
        {
            // Continue on appending the pfmd to prepare for this search pass.
            if ((capEventsPFMDLastCapId == FASMOTION_SETTINGS::UNSET_UINT32) || (capId > capEventsPFMDLastCapId))
            {
                // We are adding this capId to our collected pfmd array.
                capEventsPFMDLastCapId = capId;

                // We will append to the existing byte array so hopefully at the end we have one
                // big QByteArray that we can take the address of for the old processing logic.
                readPFMD = metamap->GetPFMD(capId, true, capEventsPFMD);
            }
        }
        // Ok.  If we read everything ok we should be good to set up our data.
        if (readPFMD)
        {
            raw_data = &capEventsPFMD;
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable read pfmd file for startCapId:" << startCapId << "through endCapId:" << endCapId;
            return true;
        }
    }
    else
    {
        // The else condition means that we are dealing with a single CAP file for review as no metamap exists.
        if (frame_count > pbMetaInfoPtr->getFrameCount())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_count:" << frame_count << "is bigger than total number of frames:" << pbMetaInfoPtr->getFrameCount();
            return true;
        }

        // Get the raw data.
        raw_data = pbMetaInfoPtr->getRawBuffer();
        if (raw_data == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No raw data.";
            return true;
        }
    }

    // Get our pointer into the raw data.
    data_ptr = (const uint8_t *)raw_data->constData();

    // Seek to the right spot in the buffer.
    if (starting_frame_num != 0)
    {
        // Move to the last frame we processed.
        data_ptr += ((uint32_t)getPfmdLen() * (starting_frame_num - 1));

        // Initialize our "last" markers value.
        memcpy(&last_marker_word, data_ptr + (2 * sizeof(uint32_t)), sizeof(last_marker_word));
        if (do_swap)
            last_marker_word = qFromBigEndian(last_marker_word);

        if (!isSeries3())
        {
            uint32_t pfmd_event;

            // Get the event marker word from the PFMD in the right order.
            pfmd_event = last_marker_word;

            // Read the video trigger flags. The flags are in the 8th word (offset=7).
            memcpy(&last_vtrig_pfmd, data_ptr + (7 * sizeof(uint32_t)), sizeof(last_vtrig_pfmd));

            // Update the event marker word with the video trigger information. If we
            // actually change the PFMD, then update the raw event too.
            if (set_vtrig_marker(starting_frame_num, pfmd_event, nullptr, last_vtrig_pfmd, 0))
            {
                if (do_swap)
                    last_marker_word = qFromBigEndian(pfmd_event);
                else
                    last_marker_word = pfmd_event;
            }
        }

        // Bump our pointer up.
        data_ptr += getPfmdLen();
    }

    // Loop through the file building up our markers.
    for (frames_processed = starting_frame_num; frames_processed < (starting_frame_num + frame_count); ++frames_processed, data_ptr += getPfmdLen())
    {
        uint32_t raw_event;
        uint32_t marker_word;

        // Extract the word containing the markers. This is the 3rd word (offset=2).
        memcpy(&raw_event, data_ptr + (2 * sizeof(uint32_t)), sizeof(raw_event));
        if (do_swap)
            marker_word = qFromBigEndian(raw_event);
        else
            marker_word = raw_event;

        // If this is the first frame, simply initialize the "last" markers
        // to the value we just read.
        if (frames_processed == 0)
        {
            if (do_swap)
                last_marker_word = qFromBigEndian(raw_event);
            else
                last_marker_word = raw_event;
        }

        // Are we series3?
        if (!isSeries3())
        {
            uint32_t vt_pfmd;

            // Newer cameras may have video trigger markers.
            // Read the video trigger flags. The flags are in the 8th word (offset=7).
            memcpy(&vt_pfmd, data_ptr + (7 * sizeof(uint32_t)), sizeof(vt_pfmd));

            // Update the event marker word with the video trigger information. If we
            // actually change the PFMD, then update the raw event too.
            if (set_vtrig_marker(frames_processed, marker_word, &last_marker_word, vt_pfmd, last_vtrig_pfmd))
            {
                if (do_swap)
                    raw_event = qFromBigEndian(marker_word);
                else
                    raw_event = marker_word;
            }

            // Remember the video trigger PFMD for next time.
            last_vtrig_pfmd = vt_pfmd;
        }

        // Does this frame meet the event critera we're searching for?
        if (eventFilter::match_criteria(last_marker_word, marker_word, event_selections, event_logic_is_and, isSeries3()))
        {
            bool    handled(false);
            int     cur_count = marker_list.size();

            // Get the trigger relative index for this frame.
            int rel_index = ((int)pbMetaInfoPtr->getStartFrame() - (int)pbMetaInfoPtr->getTriggerFrame()) + (int)frames_processed;

#if 0
            uint32_t *pfmd_ptr = (uint32_t *) data_ptr;
            QLOG_DEBUG() << "###########EVENT MARKER PFMD RECORD#############"
                         << "\nRAW: [00]:" << Qt::showbase << Qt::hex << pfmd_ptr[0]
                         << "\nRAW: [01]:" << Qt::showbase << Qt::hex << pfmd_ptr[1]
                         << "\nRAW: [02]:" << Qt::showbase << Qt::hex << pfmd_ptr[2]
                         << "\nRAW: [03]:" << Qt::showbase << Qt::hex << pfmd_ptr[3]
                         << "\nRAW: [04]:" << Qt::showbase << Qt::hex << pfmd_ptr[4]
                         << "\nRAW: [05]:" << Qt::showbase << Qt::hex << pfmd_ptr[5]
                         << "\nRAW: [06]:" << Qt::showbase << Qt::hex << pfmd_ptr[6]
                         << "\nRAW: [07]:" << Qt::showbase << Qt::hex << pfmd_ptr[7]
                         << "\nRAW: [08]:" << Qt::showbase << Qt::hex << pfmd_ptr[8]
                         << "\nRAW: [09]:" << Qt::showbase << Qt::hex << pfmd_ptr[9]
                         << "\nRAW: [10]:" << Qt::showbase << Qt::hex << pfmd_ptr[10]
                         << "\nRAW: [11]:" << Qt::showbase << Qt::hex << pfmd_ptr[11]
                         << "\nRAW: [12]:" << Qt::showbase << Qt::hex << pfmd_ptr[12]
                         << "\nRAW: [13]:" << Qt::showbase << Qt::hex << pfmd_ptr[13]
                         << "\nRAW: [14]:" << Qt::showbase << Qt::hex << pfmd_ptr[14]
                         << "\nRAW: [15]:" << Qt::showbase << Qt::hex << pfmd_ptr[15]
                         << "\n###############################################";
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "\nGot A Match marker_word:" << Qt::showbase << Qt::hex << marker_word << "frame:" << frames_processed << "rel_index:" << rel_index << "do_swap:" << do_swap;
#endif

            // We want to add this frame to our list of markers, but squeeze
            // each run into a sequence of { first_frame, RUN_FLAG, last_frame }.

            // If it's the first or second, just copy - no squeezing yet.
            if (cur_count < 2)
            {
                marker_list.append(rel_index);
                handled = true;
            }

            // At this point, we know we have at least 3 entries in our list
            // (cur_count >= 2), so we can safely do the following math.
            if (!handled &&
                (marker_list.at(cur_count - 1) == (rel_index - 1)) &&
                (marker_list.at(cur_count - 2) == (rel_index - 2)))
            {
                // Now we know we have a run of three consecutive frame numbers,
                // so it's time to switch to the run encoding.
                marker_list.replace(cur_count - 1, INT_MAX);       // Use the maximum positive value - we're not likely to have 2GBytes worth of frames.
                marker_list.append(rel_index);
                handled = true;
            }

            // See if we're extending the run.
            if (!handled &&
                (marker_list.at(cur_count - 1) == (rel_index - 1)) &&
                (marker_list.at(cur_count - 2) == INT_MAX))
            {
                // Replace the last_frame value.
                marker_list.replace(cur_count - 1, rel_index);
                handled = true;
            }

            // No conditions apply, just copy it over.
            if (!handled)
                marker_list.append(rel_index);
        }

        // Remember this marker value.
        if (do_swap)
            last_marker_word = qFromBigEndian(raw_event);
        else
            last_marker_word = raw_event;
    }

    bool allFramesProcessed = (frames_processed == pbMetaInfoPtr->getFrameCount());

    // If we have read all of the frames clear out our multi-interval data.
    if (allFramesProcessed)
    {
        capEventsPFMD.clear();
        capEventsPFMDLastCapId = FASMOTION_SETTINGS::UNSET_UINT32;
    }

    // Return true if all of the frames were processed or false if not.
    return (allFramesProcessed);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief The user has right mouse clicked in our window. Put up a
 *        context specific menu.
 *
 * @param event_ptr     Pointer to the menu event.
 */
void PlaybackCap::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QMenu   context_menu(this);

    context_menu.addAction(MainWindow::GetInstance()->get_view_summary());
    context_menu.addAction(MainWindow::GetInstance()->get_view_review_controls());
    context_menu.addAction(MainWindow::GetInstance()->get_view_pfmd());
    context_menu.addAction(MainWindow::GetInstance()->get_view_local_frametime());
    context_menu.addAction(MainWindow::GetInstance()->get_event_marker_action());
    context_menu.addAction(MainWindow::GetInstance()->get_view_info_cursor());
    context_menu.addAction(video_player->get_action_overlay());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_metadata_dock()->toggleViewAction());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_save());

    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

/*!
 * @brief Respond to getting focus. We need to tell the main window to make other "children" inactive
 *        so we can boost the frame rate to our camera.
 * @param event     Pointer to the event structure.
 */
void PlaybackCap::focusInEvent(QFocusEvent *event)
{
    // Pass it on up.
    PlaybackMDI::focusInEvent(event);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has initiated a save operation.
 *
 * @sa MainWindow::camera_save()
 */
void PlaybackCap::doImageSave()
{
    bool ready_to_save = false;
    uint32_t    num_frames;

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Sanity checks.
    if (pbMetaInfoPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No metadata pointer!";
        return;
    }
    if (review_controls == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No review_controls pointer!";
        return;
    }

    // Not allowed to save if playback is active.
    if (playback_timer && playback_timer->isActive())
    {
        QMessageBox::warning(this, "Error Saving File", tr("Unable to save during playback"));
        return;
    }

    // Get our dialog allocated.
    save_dlg = std::shared_ptr<SaveDialog>(new SaveDialog(0, this));
    if (!save_dlg)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_SAVE_DIALOG;
        return;
    }

    // Initialize our dialog. The ordering of the following lines of code is very
    // important. We need to setup the trigger and frame count BEFORE calling the
    // set_direct_transfer() method, so that the limits will be initialized and allow
    // the ranges to be setup for the set_clip_range() call.
    save_dlg->set_sensor_type(pbMetaInfoPtr->getSensorType());
    save_dlg->set_num_frames(pbMetaInfoPtr->getFrameCount());
    save_dlg->set_trig_frame(pbMetaInfoPtr->getTriggerFrame());
    save_dlg->set_direct_transfer(true);
    save_dlg->set_clip_range(review_controls->get_clip_start(), review_controls->get_clip_end());

    // Finally we can have the dialog initialize it's controls.
    save_dlg->init_controls();

    // Are there overlays?
    if (video_player)
        save_dlg->enable_overlay_merge(video_player->has_overlay());

    // Pop up the dialog.
    if (save_dlg->exec() == QDialog::Rejected)
    {
        save_dlg.reset();
        return;
    }

    // Make sure there's enough space.
    num_frames = (save_dlg->get_clip_end() - save_dlg->get_clip_start()) + 1;
    if (!MainWindow::GetInstance()->check_avail_space(save_dlg->get_dest_path(), num_frames, save_dlg->GetStreamType(), pbMetaInfoPtr->getWidth(),
                                                      pbMetaInfoPtr->getHeight(), pmMetaDataInfo->GetMetaInfo()->bit_sel, pbMetaInfoPtr->getSensorType()))
        return;

    // Create and deploy our transfer progress dialog.
    if (CreateXferDialog())
    {
        if (xferDialog)
        {
            xferDialog->setXferSrcLabel(tr("Loading image:"));
            xferDialog->setXferDestValue(save_dlg->get_dest_path());
            xferDialog->setBarMin(save_dlg->get_clip_start());
            xferDialog->setBarMax(save_dlg->get_clip_end());
            xferDialog->setBarValue(save_dlg->get_clip_start());
            xferDialog->showDialogWithPolling(SAVER_POLL_RATE);
            ready_to_save = true;
        }
        else
        {
            QString xfer_create_err_str = tr("Unable to create progress dialog for save process.");
            QLOG_ERROR() << __PRETTY_FUNCTION__ << xfer_create_err_str;
            QMessageBox::warning(this, tr("Save Error"), xfer_create_err_str);
        }
    }
    if (ready_to_save)
    {
        // We want to hook up a signal catch for when SaveImages finishes.
        connect(this, &PlaybackCap::saveImagesComplete, this, &PlaybackCap::saveImagesCompleted, Qt::UniqueConnection);

        // Save the data.
        thread_stop = false;
        saver_thread = QtConcurrent::run(SaveImages, this);
    }
}

/*!
 * @brief The saver has completed its image save and the thread completed.
 */
void PlaybackCap::saveImagesCompleted(bool savedOK)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "savedOK:" << savedOK;

    // There should have only been one connection set up when it was started so we are done with it now.
    disconnect(this, &PlaybackCap::saveImagesComplete, this, &PlaybackCap::saveImagesCompleted);

    // Close down any progress dialogs that were up.
    DoneWithProgressDialogs();

    // For testing purposes only.  This will popup a message box about status of the save.
    //QMessageBox::information(this, tr("Save Images"), tr("Status: %1").arg(savedOK));
}

