/*!
 * @file playback_mdi.cpp
 * @brief This file contains the implementation of the PlaybackMDI
 *        class. This is the child window class which allows the user
 *        to playback a captured video or "stack".
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "playback_mdi.h"

#include <QContextMenuEvent>
#include <QDirIterator>
#include <QDockWidget>
#include <QFile>
#include <QFileInfo>
#include <QGraphicsScene>
#include <QImageWriter>
#include <QMdiSubWindow>
#include <QMenu>
#include <QMessageBox>
#include <QPainter>
#include <QPixmap>
#include <QPixmapCache>
#include <QtConcurrent>
#include <QStyle>
#include <QThread>
#include <QTimer>
#include <QTimerEvent>
#include <QVBoxLayout>
#include <math.h>

#include "camera_info.h"
#include "camera_review.h"
#include "camera_video.h"
#include "client_math.h"
#include "control_panel.h"
#include "eventfilter.h"
#include "export_dlg.h"
#include "fi_msgbox.h"
#include "fi_pixmap.h"
#include "fi_progress_dialog.h"
#include "fi_saver.h"
#include "fi_sensor.h"
#include "gige_socket.h"
#include "jump_dialog.h"
#include "logger.h"
#include "mainwindow.h"
#include "panel_display.h"
#include "playback_meta_info.h"
#include "save_dialog.h"
#include "save_video.h"
#include "user_preferences.h"

#include "log_if.h"

namespace PLAYBACK_MDI_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Default record rate if we don't have a metadata file.
const int   CAM_DEF_REC_RATE            = 500;

//! The default auto-white balance gain.
const double    DEFAULT_AWB_GAIN        = 1.0;

//! This is the rate that we are really playing frames at.
const int   DISPLAY_RATE                = 30;

//! The number of entries in our 8-bit lookup tables.
const int   LUT_LENGTH                  = 256;

//! Arbitrary point at which we release control of the UI.
const int   MAX_SEARCH_COUNT            = 5000;

//! Wait 33 ms for next frame if playback rate is above this.
const int   PLAY_INTERVAL               = 33;

//! Rate at which we poll the saver object's progress.
const int   SAVER_POLL_RATE             = 100;

//! Time to pause between searching bursts.
const int   SEARCH_DELAY                = 10;


// Set this to true if you want a verbose debug log.
const bool PLAYBACK_VERBOSE_LOG         = false;

}   // End of namespace PLAYBACK_MDI_SETTINGS.

using namespace PLAYBACK_MDI_SETTINGS;

#define DEBUG_FORMAT_ROUTINE
#ifdef DEBUG_FORMAT_ROUTINE
/*!
 * @brief Utility function to create a hex dump from a uint8_t pointer.
 *
 * @param ba    The uint8_t pointer we're interested in.
 * @param len   The number of bytes to dump.
 * @return The character hex dump.
 */
QString byte_array_to_string(uint8_t *ba, int len)
{
    int     i;
    QString resHex;

    for (i = 0; i < len; ++i)
    {
        uint8_t value = ba[i];
        resHex.append(QString::fromUtf8("%1 ").arg(value, 2, 16, QLatin1Char('0')));
    }

    return resHex;
}
#endif

//! Keep a constant of when a sanity check on the save dialog fails.
const QString PlaybackMDI::MISSING_SAVE_DIALOG = QStringLiteral("There is no save dialog holding save information!");

//! Keep a constant for when the pm_metadata management object is not valid.
const QString PlaybackMDI::MISSING_PM_METADATA = QStringLiteral("The pm_metadata managed by pmMetaDataInfo is invalid!");

//--------------------------------------------------------------------
// Internal Class Implementation for CaptureMetaInfo
//--------------------------------------------------------------------

CaptureMetaInfo::CaptureMetaInfo()
{
    pm_metadata_ptr = nullptr;
}
CaptureMetaInfo::~CaptureMetaInfo()
{
    if (pm_metadata_ptr != nullptr)
    {
        delete pm_metadata_ptr;
        pm_metadata_ptr = nullptr;
    }
}

/*!
 * @brief Factory method to create the pm_metadata for an fiRawFile.
 *
 * @param[in] capRawFile     The fiRawFile for a Fastec Imaging raw CAP file.
 *
 * @returns The created CapMetaInfo shared pointer for use.
 */
QSharedPointer<CaptureMetaInfo> CaptureMetaInfo::CreateCaptureMetaInfo(QSharedPointer<fiRawFile> capFile)
{
    QSharedPointer<CaptureMetaInfo> capMetaInfo;

    // Process the raw cap file to read and ready its pm_metadata.
    if (!capFile.isNull())
    {
        QString capFileName;

        // Retrieve the capFile's filename which also verifies that it is open and valid.
        if (!capFile->getFileName(capFileName))
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to validate capFile provided. Is it closed?";
            goto CreateCaptureMetaInfoExit;
        }

        // Create the object for use managing the pm_metadata allocation and content.
        capMetaInfo = QSharedPointer<CaptureMetaInfo>(new CaptureMetaInfo);

        // Allocated ok?
        if (!capMetaInfo.isNull())
        {
            // Allocate our pm_metadata_ptr.
            capMetaInfo->pm_metadata_ptr = new pm_metadata;

            // Allocated ok?
            if (capMetaInfo->pm_metadata_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate pm_metadata for" << capFileName << "!";
                goto CreateCaptureMetaInfoExit;
            }

            // Install the name of the raw cap file.
            capMetaInfo->rawFileName = capFileName;

            // Read the pm_metadata from the raw CAP file.
            if (!capFile->readPmMetadata(capMetaInfo->pm_metadata_ptr))
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read pm_metadata for" << capFileName << "!";
                capMetaInfo.reset();    // Can't use this one.
                goto CreateCaptureMetaInfoExit;
            }

            // And finally run it through the "cleanup" to upgrade old versions etc.
            capMetaInfo->cleanupMetadataVersion();
        }
    }

CreateCaptureMetaInfoExit:
    return capMetaInfo;
}

/*!
 * @brief Factory method to create the pm_metadata for an fiRawDrife.
 *
 * @param[in] capRawDrive   The fiRawDrive for a Fastec Imaging Long Record Drive.
 * @returns The created CapMetaInfo shared pointer for use.
 */
QSharedPointer<CaptureMetaInfo> CaptureMetaInfo::CreateCaptureMetaInfo(QSharedPointer<fiRawDrive> capRawDrive)
{
    QSharedPointer<CaptureMetaInfo> capMetaInfo;

    // Process the raw cap file to read and ready its pm_metadata.
    if (!capRawDrive.isNull())
    {
        QString capDriveName = "fiRawDrive"; // For lack of a drive name yet... TBD.

        // Create the object for use managing the pm_metadata allocation and content.
        capMetaInfo = QSharedPointer<CaptureMetaInfo>(new CaptureMetaInfo);

        // Allocated ok?
        if (!capMetaInfo.isNull())
        {
            // Allocate our pm_metadata_ptr.
            capMetaInfo->pm_metadata_ptr = new pm_metadata;

            // Allocated ok?
            if (capMetaInfo->pm_metadata_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to allocate pm_metadata for" << capDriveName << "!";
                goto CreateCaptureMetaInfoExit;
            }

            // Install the name of the raw drive.
            capMetaInfo->rawFileName = capDriveName;

            // Read the pm_metadata from the raw CAP file.
            if (!capRawDrive->readPmMetadata(capMetaInfo->pm_metadata_ptr))
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to read pm_metadata for" << capDriveName << "!";
                capMetaInfo.reset();    // Can't use this one.
                goto CreateCaptureMetaInfoExit;
            }

            // And finally run it through the "cleanup" to upgrade old versions etc.
            capMetaInfo->cleanupMetadataVersion();
        }
    }

CreateCaptureMetaInfoExit:
    return capMetaInfo;
}

/*!
 * @brief Utility to see if this capFile is my capFile.
 *
 * @param[in] capFile   The fiRawFile for a Fastec Imaging raw cap file.
 * @returns True if they are the same, false if not.
 */
bool CaptureMetaInfo::containsMyData(QSharedPointer<fiRawFile> capFile)
{
    bool isMine = false;

    // Process the raw cap file to read and ready its pm_metadata.
    if (!capFile.isNull())
    {
        QString capFileName;

        // Retrieve the capFile's filename which also verifies that it is open and valid.
        if (capFile->getFileName(capFileName))
        {
            // Is it mine?
            if (capFileName.compare(rawFileName) == 0)
            {
                isMine = true;
            }
        }
    }

    // Return the state of the comparison if it happened.
    return isMine;
}

/*!
 * @brief Clean up partition metadata based on the app_ver.
 *
 * If we loaded metadata in a CAP file from a previous version of the firmware,
 * some of the fields might not be initialized as expected. This will clean up
 * a pm_metadata structure so any new fields have sensible values.
 *
 */
void CaptureMetaInfo::cleanupMetadataVersion()
{
    unsigned major;
    unsigned minor;
    unsigned patch;
    uint32_t version;

    // Do we have the capture metadata?
    if (pm_metadata_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Our pm_metadata_ptr object was null!";
        return;
    }

    if (sscanf(pm_metadata_ptr->app_ver, "%u.%u.%u", &major, &minor, &patch) != 3)
    {
        // Log the error.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString("Unexpected version: %1").arg(pm_metadata_ptr->app_ver, 31);

        // Force "unknown" to be older than the current version.
        version = 1;
    }
    else if (((major == 0) && (minor == 0) && (patch == 0)) ||
             ((major == 255) && (minor == 255) && (patch == 255)))
    {
        // Development version.
        version = CURRENT_FIRMWARE_VERSION;
    }
    else
    {
        version = FIRMWARE_VERSION(major, minor, patch);
    }

    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(1, 7, 5)))
    {
        memset(((uint8_t *)pm_metadata_ptr) + GIGE_PM_META_SIZE_1_7_5, 0, sizeof(pm_metadata) - GIGE_PM_META_SIZE_1_7_5);

        // Old PM doesn't have the trigger setting so we will construct one.
        // We stored the trigger frame as an absolute frame number in the PM.
        pm_metadata_ptr->trig_set = 0x7f;
        if (pm_metadata_ptr->trig_frame < pm_metadata_ptr->first_frame)
            pm_metadata_ptr->trig_set_frames = pm_metadata_ptr->first_frame + pm_metadata_ptr->frame_count -
                                               pm_metadata_ptr->trig_frame;
        else
            pm_metadata_ptr->trig_set_frames = pm_metadata_ptr->trig_frame - pm_metadata_ptr->first_frame;
    }
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(1, 9, 99)))
    {
        memset(((uint8_t *)pm_metadata_ptr) + GIGE_PM_META_SIZE_1_9_99, 0, sizeof(pm_metadata) - GIGE_PM_META_SIZE_1_9_99);

        // Old PM doesn't have the sensor options field so fill in default value.
        pm_metadata_ptr->sensor_opts = (1 << GIGE_SENSOPT_BIN_SHIFT) | (1 << GIGE_SENSOPT_SUBS_SHIFT);
    }
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(2, 0, 4)))
    {
        memset(((uint8_t *)pm_metadata_ptr) + GIGE_PM_META_SIZE_2_0_4, 0, sizeof(pm_metadata) - GIGE_PM_META_SIZE_2_0_4);

        pm_metadata_ptr->black_offset = 0;
        pm_metadata_ptr->sensor_dac = 0;
        pm_metadata_ptr->sensor_vrefcol = 0;
        pm_metadata_ptr->sensor_gain = 0;

        // We were trying to stuff 10 pounds of sensor gain into an 8-pound bag.
        // Try to convert the old values to something that makes sense.
        if ((pm_metadata_ptr->old_sensor_gain == 100) ||
            (pm_metadata_ptr->old_sensor_gain == 150) ||
            (pm_metadata_ptr->old_sensor_gain == 200) ||
            (pm_metadata_ptr->old_sensor_gain == 250))
            pm_metadata_ptr->sensor_gain = pm_metadata_ptr->old_sensor_gain;
        else if (pm_metadata_ptr->old_sensor_gain == (300 - 256))
            pm_metadata_ptr->sensor_gain = 300;
        else if (pm_metadata_ptr->old_sensor_gain == (400 - 256))
            pm_metadata_ptr->sensor_gain = 400;
    }
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(2, 2, 1)))
    {
        // Clear out everything added since 2.2.1.
        memset(((uint8_t *)pm_metadata_ptr) + GIGE_PM_META_SIZE_2_2_1, 0, sizeof(pm_metadata) - GIGE_PM_META_SIZE_2_2_1);

        // Old RGB gains were completely invalid so use default.
        pm_metadata_ptr->red = 0x100;
        pm_metadata_ptr->green = 0x100;
        pm_metadata_ptr->blue = 0x100;
        pm_metadata_ptr->gamma = pm_metadata_ptr->old_gamma;

        // If it is less than 100, chances are good that it was a rollover when
        // the value >255 was written to the 8-bit field.
        if (pm_metadata_ptr->gamma < 100)
            pm_metadata_ptr->gamma = 100;

        // We don't really have anyway to figure this out (not like the camera)
        // so just pick something "reasonable".
        pm_metadata_ptr->white_bal = GIGE_WHITE_BAL_DAYLIGHT;

        // No know DC offset values.
        pm_metadata_ptr->dc_offset_red = 0;
        pm_metadata_ptr->dc_offset_grn = 0;
        pm_metadata_ptr->dc_offset_blu = 0;
    }
    if ((version != CURRENT_FIRMWARE_VERSION) && (version <= FIRMWARE_VERSION(3, 1, 8)))
    {
        // No skip count.
        pm_metadata_ptr->skip_count = 0;
    }
    // Otherwise, nothing to do.
}

//--------------------------------------------------------------------
// Class Implementation for PlaybackMDI
//--------------------------------------------------------------------

/*!
 * @brief The class constructor.
 */
PlaybackMDI::PlaybackMDI(QWidget *parent) :
    QWidget(parent)
{
    // In the constructor we want to instantiate all the subforms which are
    // contained within our space.
    setObjectName(QString::fromUtf8("PlaybackMDI"));

    // Our default is to open the display panel when focus event occurs.
    open_display_panel_on_focus = true;

    // Allocate our layout. We want our display split into a top pane for the
    // camera information, middle for the video and bottom for the playback controls.
    layout_ptr = new QVBoxLayout(this);
    if (layout_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate vertical layout!";
        return;
    }

    // Initialize our variables.
    do_colorize = false;
    last_frame_loaded = (uint32_t) -1;
    pbMetaInfoPtr = nullptr;
    play_frames_inc = 1;
    playback_action = nullptr;
    playback_forward = true;
    playback_timer = nullptr;
    record_rate = 0;
    saver_frame = 0;
    trigger_frame = 0;
    using_subfolders = false;
    whitebal = GIGE_WHITE_BAL_DAYLIGHT;
    swap_needed = false;

    r_wb_gain = DEFAULT_AWB_GAIN;
    g_wb_gain = DEFAULT_AWB_GAIN;
    b_wb_gain = DEFAULT_AWB_GAIN;

    r_gain_old = r_gain = ((double)GIGE_DEFAULT_RGB_GAIN) / 256.0;
    g_gain = ((double)GIGE_DEFAULT_RGB_GAIN) / 256.0;
    b_gain = ((double)GIGE_DEFAULT_RGB_GAIN) / 256.0;
    brightness = ((double)GIGE_DEFAULT_BRIGHTNESS);
    contrast = ((double)GIGE_DEFAULT_CONTRAST);
    gamma = ((double)GIGE_DEFAULT_GAMMA) / 100.0;
    do_whitebal = false;

    // Allocate memory for the lookup tables.
    r_lut = new uint8_t[ LUT_LENGTH ];
    g_lut = new uint8_t[ LUT_LENGTH ];
    b_lut = new uint8_t[ LUT_LENGTH ];

    // Initialize the lookup tables based on the default values.
    buildLuts(__PRETTY_FUNCTION__);

    // Initially we will apply any cap file custom LUT that may exist.
    applyingCapLuts = true;

    has_pfmd_pm_file        = false;
    has_metadata_txt_file   = false;
    is_video_file           = false;

    // Allocate our various panes of information.
    camera_info = new CameraInfo(this);
    video_player = new CameraVideo(this);
    review_controls = new CameraReview(this);

    // Figure out which cursor to use.
    doViewInfoCursor();

    // Tell the controls that we're in playback mode.
    camera_info->enable_player_mode();
    review_controls->enable_player_mode();

    // We are going to use a Queued connection type later, and Qt needs to know about uint32_t types.
    qRegisterMetaType<uint32_t>("uint32_t");

    // Connect our signals/slots to the review controls.
    connect(review_controls, &CameraReview::play_reverse, this, &PlaybackMDI::videoPlayReverse, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::review_stop, this, &PlaybackMDI::videoReviewStop, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::play_forward, this, &PlaybackMDI::videoPlayForward, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::jump_time, this, &PlaybackMDI::videoJumpTime, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::current_frame_changed, this, &PlaybackMDI::videoCurrentFrameChanged, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::play_update, this, &PlaybackMDI::updatePfmd);
    connect(review_controls, &CameraReview::playback_rate_changed, this, &PlaybackMDI::videoRateChanged);

    // Add our panes into our splitter.
    layout_ptr->addWidget(camera_info);
    layout_ptr->addWidget(video_player);
    layout_ptr->addWidget(review_controls);

    // Give all extra space to the player window. The other windows are fixed
    layout_ptr->setStretchFactor(camera_info, QSizePolicy::Fixed);
    layout_ptr->setStretchFactor(video_player, QSizePolicy::Expanding);
    layout_ptr->setStretchFactor(review_controls, QSizePolicy::Fixed);

    // Load the non-volatile settings.
    readSettings();

    // Make this our layout.
    setLayout(layout_ptr);

    // Make sure we get focus notifications.
    setFocusPolicy(Qt::StrongFocus);
}

/*!
 * @brief The class destructor.
 */
PlaybackMDI::~PlaybackMDI()
{
    // Clean up our widgets.
    if (layout_ptr != nullptr)
    {
        layout_ptr->removeWidget(review_controls);
        layout_ptr->removeWidget(video_player);
        layout_ptr->removeWidget(camera_info);
        review_controls->deleteLater();
        review_controls = nullptr;
        video_player->deleteLater();
        video_player = nullptr;
        camera_info->deleteLater();
        camera_info = nullptr;
        layout_ptr->deleteLater();
        layout_ptr = nullptr;
    }

    if (r_lut != nullptr)
    {
        delete []r_lut;
        r_lut = nullptr;
    }
    if (g_lut != nullptr)
    {
        delete []g_lut;
        g_lut = nullptr;
    }
    if (b_lut != nullptr)
    {
        delete []b_lut;
        b_lut = nullptr;
    }

    if (pbMetaInfoPtr != nullptr)
    {
        // Clean up our widget.
        pbMetaInfoPtr->deleteLater();
        pbMetaInfoPtr = nullptr;
    }
    if (playback_timer != nullptr)
    {
        playback_timer->stop();
        playback_timer->deleteLater();
        playback_timer = nullptr;
    }
    if (playback_action != nullptr)
    {
        delete playback_action;
        playback_action = nullptr;
    }

    // Close down the stream reader.
    if (AVIStreamReader)
    {
        FAS_VIDEO::STREAM_STATUS closeStatus = AVIStreamReader->CloseStream();
        if (closeStatus != FAS_VIDEO::STREAM_STATUS::STREAM_CLOSED)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Closing AVIStreamReader was not clean.  Stream status=" << closeStatus;
        }
        AVIStreamReader.reset();    // Reset calls destructor and sets the unique pointer empty.
    }
}

/*!
 * @brief Walk through all of the metadata in the [buffer](@ref raw_metadata) and
 *        copy only the event marker information over to a newly allocated buffer.
 *
 * @note    It is the responsibility of the [video review object](@ref CameraReview)
 *          to release the allocated buffer.
 */
void PlaybackMDI::buildData()
{
    uint8_t     *data_ptr;

    // Retrieve the raw event marker information.
    data_ptr = pbMetaInfoPtr->getRawEventFlags(getPfmdLen());
    if (data_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_raw_event_flags() failed!";
        return;
    }

    // Pass the data on to the video control panel. It is the responsibility
    // of the review controls object to release the data_ptr memory.
    review_controls->set_graph_data(data_ptr, (int)frame_count);
}

/*!
 * @brief Build the red, green and blue lookup tables (LUTs) to apply the
 *        changes to the image as specified by the user.
 */
void PlaybackMDI::buildLuts(const QString whoCalledMe)
{
    double  c_norm;
    double  b_norm;
    double  g_norm = 1.0;
    double  pix_div;
    double  pix_pow;
    double  pix_mul;
    uint32_t    r_value;
    uint32_t    g_value;
    uint32_t    b_value;
    int         i;

    if (PLAYBACK_VERBOSE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "whoCalledMe:" << whoCalledMe;
    }

    // Sanity check.
    if ((r_lut == nullptr) || (g_lut == nullptr) || (b_lut == nullptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "One of the LUTs is nullptr!";
        return;
    }

    // Normalize the brightness, contrast and gamma values. This algorithm
    // is based on trial and error - thus the "magic" numbers. Although, we
    // subtract the 100 from brightness and contast because their
    // ranges are from -100 to +100. Gamma's range is 0.10 to 2.20
    // with a step of 0.01 - thus the division with 100.0.
    c_norm = ((contrast - 100.0) * .025) + 1.0;
    b_norm = brightness - 100.0;
    if (!qFuzzyIsNull(gamma))
    {
        // We actually want the inverse of the gamma.
        g_norm = 1.0 / gamma;
    }
    else
    {
        // Make sure we have a gamma LUT instead.
        if (gamma_lut.size() == 0)
        {
            // Default to a gamma of 1.0.
            g_norm = 1.0;
        }
    }

    // Now generate the LUT values for every pixel 0-255.
    for (i = 0; i < LUT_LENGTH; ++i)
    {
        // This is essential mx+b. We adjust the "center" point to
        // be 128 instead of 0. So "c_norm" is our slope (i.e. 'm'),
        // adjust our 'x' to be at the middle of our range (128), and
        // "b_norm" is our 'b'.
        double pix_val = ((c_norm * ((double)i - 128.0)) + b_norm) + 128.0;

        // Make sure we keep our data in-range.
        if (pix_val < 0.0)
            pix_val = 0.0;
        if (pix_val > 255.0)
            pix_val = 255.0;

        // Now we want to apply gamma.
        if (gamma_lut.size() == 0)
        {
            // Use the straight value.
            pix_div = (double)(pix_val / 255.0);
            pix_pow = pow(pix_div, g_norm);
            pix_mul = pix_pow * 255.0;
        }
        else
        {
            // Use the lookup table value.
            pix_mul = gamma_lut[(int)pix_val];
        }

        // Compute the individual RGB values, adjusting for any gain changes as well.
        if (r_gain == 0.0)
            r_value = (uint32_t)(r_gain_lut[(int)pix_mul] + 0.5);
        else
            r_value = (uint32_t)((pix_mul * r_gain) + 0.5);
        if (g_gain == 0.0)
            g_value = (uint32_t)(g_gain_lut[(int)pix_mul] + 0.5);
        else
            g_value = (uint32_t)((pix_mul * g_gain) + 0.5);
        if (b_gain == 0.0)
            b_value = (uint32_t)(b_gain_lut[(int)pix_mul] + 0.5);
        else
            b_value = (uint32_t)((pix_mul * b_gain) + 0.5);

        // Make sure we're clipped at 255 - the maximum
        // output value.
        if (r_value > 0xFF)
            r_value = 0xFF;
        if (g_value > 0xFF)
            g_value = 0xFF;
        if (b_value > 0xFF)
            b_value = 0xFF;

        // Check for colorization.
        if (do_colorize &&
            ((r_value > CAMERA_VIDEO::COLORIZE_THRESHOLD) || (g_value > CAMERA_VIDEO::COLORIZE_THRESHOLD) || (b_value > CAMERA_VIDEO::COLORIZE_THRESHOLD)))
        {
            // Force the colorization value.
            r_value = 0xFF;
            g_value = 0;
            b_value = 0;
        }

        // Update the LUTs.
        r_lut[ i ] = (uint8_t)r_value;
        g_lut[ i ] = (uint8_t)g_value;
        b_lut[ i ] = (uint8_t)b_value;
    }
}

/*!
 * @brief Initializes CCM control data in our fiPixmap frame and determines whether or
 * not to perform pixel corrections (fpn) on the data.
 *
 * @param pixmap_ptr    [I/O] The fiPixmap object to initialize.
 * @param do_fpn        [O] What is the fpn_status in the metadata?  True if correcting, false if not.
 *
 * @return True on success, false otherwise.
 */
bool PlaybackMDI::init_ccm_ctldata(fiPixmap *pixmap_ptr, bool& do_fpn)
{
    double ccm[3][4];

    // Do we have the capture metadata?
    if (pmMetaDataInfo.isNull() || (pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA;
        return false;
    }

    // Do we need to do FPN?
    do_fpn = ((pmMetaDataInfo->GetMetaInfo()->fpn_status & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK);

    // If we don't have a pixmap_ptr we can't initialize it.
    if (pixmap_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "The frame's pixmap_ptr was null!";
        return false;
    }

    // To be complete we are going to set the gamma in the frame data.
    pixmap_ptr->set_gamma(gamma);

    // Note:  This setup was created when Fastec imaging CCM processing was added to fiPixmap::load_from_buf.
    // This is only for Fastec imaging CCM processing.  For reference, the default settings in fiPixmap's
    // constructor are currently being set as shown below.  We are only setting the gamma in all cases outside
    // of this condition (see the setting of gamma above).
    //
    // From fiPixmap::fiPixmap() construction:
    //    fiPixmap::ccm_updated = false;
    //    fiPixmap::jpeg_quality = -1;
    //    fiPixmap::gamma = 1.0;
    //    fiPixmap::whitebal = GIGE_WHITE_BAL_DAYLIGHT;
    //    fiPixmap::sensor_type = FI_SENSOR::SENSOR_TYPE_GENERIC_COLOR;
    //

    // Make sure our gain information is set appropriately.
    updateGainInfo();
    if (!pbMetaInfoPtr->getCCM(ccm))
    {
        // Initialize the CCM to a normal matrix.
        memset(ccm, 0, sizeof(double) * 3 * 4);
        ccm[0][1] = 1.0;
        ccm[1][2] = 1.0;
        ccm[2][3] = 1.0;
        whitebal = GIGE_WHITE_BAL_NORMAL;
        r_wb_gain = g_wb_gain = b_wb_gain = 1.0;
    }

    // Tell the pixmap object the white balance setting.
    pixmap_ptr->set_sensor_type(pbMetaInfoPtr->getSensorType());
    pixmap_ptr->set_whitebal(whitebal, r_wb_gain, g_wb_gain, b_wb_gain);
    pixmap_ptr->set_ccm(ccm);

    return true;    // We are set to go.
}

/*!
 * @brief Figure out the min/max frame number and the number of frames in a directory.
 *
 * @param[in] video_filename    The filename to split up.
 * @param[out] min_framnum      The smallest frame number found.
 * @param[out] max_framnum      The largets frame number found.
 *
 * @return The number of frames total found in the directory structure.
 */
int PlaybackMDI::calculateMinMaxFrameNumber(const QString& video_filename, int& min_framenum, int& max_framenum)
{
    int frame_num;
    QFileInfo   file_info;
    QString file_name;
    QString     path;
    uint32_t    count = 0;

    // Check to see if we have a proper path.
    if (!splitFileName(video_filename, path, frame_num))
        return 0;

    // Initialize our variables.
    min_framenum = -1;
    max_framenum = -1;

    // Walk the whole directory tree to figure out the maximum and minimum frame number. The split routine
    // figured out if we were using subfolders and if so positioned us at the parent folder.
    QDirIterator    dir_iter(path, QDirIterator::Subdirectories);
    while (dir_iter.hasNext())
    {
        dir_iter.next();
        file_info = dir_iter.fileInfo();
        if (file_info.isFile())
        {
            file_name = dir_iter.fileName();
            if (splitFileName(file_name, path, frame_num))
            {
                if ((min_framenum == -1) || (frame_num < min_framenum))
                    min_framenum = frame_num;
                if ((max_framenum == -1) || (frame_num > max_framenum))
                    max_framenum = frame_num;
                count++;
            }
        }
    }

    // Return the number of frames we found.
    return (int)count;
}



/*!
 * @brief Try to locate the next frame, starting at frame_num, and load it into the pixmap_dest.
 *
 * @param frame_num     Frame number we'll start looking. If it doesn't exist we'll keep moving.
 * @param pixmap_dest   Pointer to a Pixmap for the image.
 * @return True on success, false otherwise.
 */
bool PlaybackMDI::findImageFile(uint32_t frame_num, fiPixmap *pixmap_dest)
{
    uint32_t    i;
    QString     image_filename;

    // Sanity check.
    if (pixmap_dest == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Invalid parameters.";
        return false;
    }
    if (review_controls == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "What? No review_controls?!";
        return false;
    }
    if (pbMetaInfoPtr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "What? No metadata?!";
        return false;
    }

    // We need to handle the case where files might be missing.
    for (i = 0; i < frame_count; i++)
    {
        QFileInfo   file_info;

        // Build our path differently if we're using subfolders.
        if (using_subfolders)
        {
            uint32_t    subdir_index;

            subdir_index = (frame_num - review_controls->get_clip_start()) / FI_SAVER::SUB_FOLDER_MAX_IMAGES;
            image_filename = QString::fromUtf8("%1%2%3/%4.%5").arg(stack_path, FiSaver::SUB_FOLDER_PREFIX)
                             .arg(subdir_index, 4, 10, QLatin1Char('0')).arg(frame_num, 7, 10, QLatin1Char('0'))
                             .arg(file_ext);
        }
        else
        {
            image_filename = QString::fromUtf8("%1%2.%3").arg(stack_path).arg(frame_num, 7, 10, QLatin1Char('0')).arg(file_ext);
        }

        // Check to see if the image file exists.
        file_info.setFile(image_filename);
        if (!file_info.exists())
        {
            // Do a more extensive search if we're using sub-folders.
            if (using_subfolders)
            {
                QString         tmp_file;

                // The file didn't exist where we thought it should. Let's see if we can
                // find it anywhere below our root folder. Build up the expected image filename.
                tmp_file = QString::fromUtf8("%1.%2").arg(frame_num, 7, 10, QLatin1Char('0')).arg(file_ext);
                QDirIterator    dir_walker(stack_path, QStringList(tmp_file), QDir::Files | QDir::NoSymLinks, QDirIterator::Subdirectories);

                // Walk through the folders.
                if (dir_walker.hasNext())
                {
                    // We found it!
                    image_filename = dir_walker.next();
                }
                else
                {
                    // Clear our file name string.
                    image_filename.clear();
                }
            }
            else
            {
                // Not using sub-folders, so the file simply doesn't exist.
                image_filename.clear();
            }
        }

        // Do we have a file name?
        if (!image_filename.isEmpty())
        {
            // Tell the pixmap object the white balance setting.
            pixmap_dest->set_gamma(gamma);
            pixmap_dest->set_sensor_type(pbMetaInfoPtr->getSensorType());
            pixmap_dest->set_whitebal(whitebal);

            r_wb_gain = pbMetaInfoPtr->getBalanceRed() / 4096.0;
            g_wb_gain = pbMetaInfoPtr->getBalanceGreen() / 4096.0;
            b_wb_gain = pbMetaInfoPtr->getBalanceBlue() / 4096.0;
//                QLOG_TRACE() << __PRETTY_FUNCTION__ << "r_wb_gain:" << r_wb_gain << "g_wb_gain:" << g_wb_gain << "b_wb_gain:" << b_wb_gain;

            // Update our LUTs.
            buildLuts(__PRETTY_FUNCTION__);

            // Load the image.
            if (!pixmap_dest->load(image_filename))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the image: " << image_filename;
                image_filename.clear();
            }
            else
            {
                // We're done.
                break;
            }
        }

        // Move to the next frame number.
        frame_num++;
        if (frame_num > pbMetaInfoPtr->getEndFrame())
            frame_num = pbMetaInfoPtr->getStartFrame();
    }

    // Return true if we have a filename.
    return !image_filename.isEmpty();
}



/*!
 * @brief Common logic to generate a transfer progress dialog used by
 * this class hierarchy.
 *
 * @param starting_frame_num    The frame number we're to start with.
 * @param frame_count           The number of frames we're supposed to process.
 *
 * @return True the dialog was created successfully, false if an error occurred.
 */
bool PlaybackMDI::CreateXferDialog(QWidget *parent)
{
    bool rc = false;

    // Create our transfer dialog for use when saving images.
    xferDialog = fiProgressDialog::CreateXferProgressDialog(FI_PROGRESS::XFER_TYPES::XFER_TYPE_SAVE,
                                                            camera_info->get_camera_name(),
                                                            __PRETTY_FUNCTION__,
                                                            parent);

    // Do we have the transfer progress dialog?
    if (xferDialog && save_dlg)
    {
        // Set up the source information.
        xferDialog->setXferSrcLabel(SAVE_VIDEO::SAVE_SRC_RECEIVE_FRAME);
        xferDialog->showXferSrc(true);

        // Set up the destination information.
        xferDialog->setXferDestValue(save_dlg->get_dest_path());
        xferDialog->showXferDest(true);

        // Set up the save type information.
        FI_SAVER::STREAM_TYPE saveStreamType = save_dlg->GetStreamType();
        QString saveStreamTypeStr = FiSaver::streamTypeToQString(saveStreamType);
        xferDialog->setXferTypeValue(saveStreamTypeStr);
        xferDialog->showXferType(true);

        // Set up the status icon.
        xferDialog->showStatusIcon(false);
        xferDialog->setStatusIcon(QStyle::SP_MessageBoxInformation);

        // Set up the status.
        saveStreamTypeStr.prepend(tr("Saving "));
        if (FiSaver::isStreamVideo(saveStreamType))
        {
            saveStreamTypeStr.append(tr(" Video"));
        }
        xferDialog->setStatusValue(saveStreamTypeStr);
        xferDialog->showStatusValue(true);

        xferDialog->setBarMinMax((int)save_dlg->get_clip_start(), (int)save_dlg->get_clip_end());
        xferDialog->setBarValue((int)save_dlg->get_clip_start());
        xferDialog->showBar(true);

        // Allow the user to cancel this process and catch the cancelled signal.
        xferDialog->showCancel(true);
        connect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &PlaybackMDI::UserCancelledSave, Qt::UniqueConnection);

        // And finally we will be showing this dialog with a polling effort where
        // we catch the a poll request for updates from the dialog within our
        // control.  The dialog will emit the progress update poll cycle signal
        // when it is time to ask for progress.
        connect(xferDialog.get(), &fiProgressDialog::progressUpdatePollCycle, this, &PlaybackMDI::UpdateSaveProgress, Qt::UniqueConnection);

        rc = true;
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a transfer progress dialog!";
    }

    return rc;
}

/*!
 * @brief Common logic to recycle the xfer progress dialog created for use while saving
 * things in this object hierarchy.
 */
void PlaybackMDI::DoneWithProgressDialogs()
{
    // Clean up any progress dialog(s) that were allocated or displayed.
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }
    if (xferDialog)
    {
        // Recycle any resources and release the transfer progress dialog.
        fiProgressDialog::RecycleProgressDialog(xferDialog);
    }
}

const QByteArray& PlaybackMDI::getRawImage() const
{
    static QByteArray null_array;

    // Sanity check.
    return null_array;
}


/*!
 * @brief Extract the event marker information from the raw metadata file.
 *
 * @param starting_frame_num    The frame number we're to start with.
 * @param frame_count           The number of frames we're supposed to process.
 *
 * @return True if the entire file has been processed (aka done), false if there's more to do.
 */
bool PlaybackMDI::getRawMarkers(uint32_t starting_frame_num, uint32_t frame_count, bool do_swap)
{
    uint32_t    index;
    uint32_t    last_marker_word(0);
    const uint8_t *data_ptr;
    const QByteArray *raw_data;

    // Sanity check.
    if (!pbMetaInfoPtr)
        return true;
    if (!pbMetaInfoPtr->captureMetadataIsAvail())
        return true;
    if (frame_count > pbMetaInfoPtr->getFrameCount())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_count:" << frame_count << "is bigger than total number of frames:" << pbMetaInfoPtr->getFrameCount();
        return true;
    }

    // Get the raw data.
    raw_data = pbMetaInfoPtr->getRawBuffer();
    if (raw_data == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No raw data.";
        return true;
    }

    // Get our pointer into the raw data.
    data_ptr = (const uint8_t *)raw_data->constData();

    // Seek to the right spot in the file.
    if (starting_frame_num != 0)
    {
        // Move to the last frame we processed.
        data_ptr += ((uint32_t)getPfmdLen() * (starting_frame_num - 1));

        // Initialize our "last" markers value.
        memcpy(&last_marker_word, data_ptr + (2 * sizeof(uint32_t)), sizeof(last_marker_word));
        if (do_swap)
            last_marker_word = qFromBigEndian(last_marker_word);

        // Bump our pointer up.
        data_ptr += getPfmdLen();
    }

    // Loop through the file building up our markers.
    for (index = starting_frame_num; index < (starting_frame_num + frame_count); ++index, data_ptr += getPfmdLen())
    {
        uint32_t    marker_word;

        // Extract the word containing the markers. This is the 3rd word (offset=2).
        memcpy(&marker_word, data_ptr + (2 * sizeof(uint32_t)), sizeof(marker_word));
        if (do_swap)
            marker_word = qFromBigEndian(marker_word);

        // If this is the first frame, simply initialize the "last" markers
        // to the value we just read.
        if (index == 0)
            last_marker_word = marker_word;

        // Does this frame meet the event critera we're searching for?
        if (eventFilter::match_criteria(last_marker_word, marker_word, event_selections, event_logic_is_and, isSeries3()))
        {
            bool    handled(false);
            int     cur_count = marker_list.size();
#if 0
            uint32_t *pfmd_ptr = (uint32_t *) data_ptr;
            QLOG_DEBUG() << "###########EVENT MARKER PFMD RECORD#############"
                         << "\nRAW: [00]:" << Qt::showbase << Qt::hex << pfmd_ptr[0]
                         << "\nRAW: [01]:" << Qt::showbase << Qt::hex << pfmd_ptr[1]
                         << "\nRAW: [02]:" << Qt::showbase << Qt::hex << pfmd_ptr[2]
                         << "\nRAW: [03]:" << Qt::showbase << Qt::hex << pfmd_ptr[3]
                         << "\nRAW: [04]:" << Qt::showbase << Qt::hex << pfmd_ptr[4]
                         << "\nRAW: [05]:" << Qt::showbase << Qt::hex << pfmd_ptr[5]
                         << "\nRAW: [06]:" << Qt::showbase << Qt::hex << pfmd_ptr[6]
                         << "\nRAW: [07]:" << Qt::showbase << Qt::hex << pfmd_ptr[7]
                         << "\nRAW: [08]:" << Qt::showbase << Qt::hex << pfmd_ptr[8]
                         << "\nRAW: [09]:" << Qt::showbase << Qt::hex << pfmd_ptr[9]
                         << "\nRAW: [10]:" << Qt::showbase << Qt::hex << pfmd_ptr[10]
                         << "\nRAW: [11]:" << Qt::showbase << Qt::hex << pfmd_ptr[11]
                         << "\nRAW: [12]:" << Qt::showbase << Qt::hex << pfmd_ptr[12]
                         << "\nRAW: [13]:" << Qt::showbase << Qt::hex << pfmd_ptr[13]
                         << "\nRAW: [14]:" << Qt::showbase << Qt::hex << pfmd_ptr[14]
                         << "\nRAW: [15]:" << Qt::showbase << Qt::hex << pfmd_ptr[15]
                         << "\n###############################################";
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "\nGot A Match marker_word:" << Qt::showbase << Qt::hex << marker_word << "frame:" << index << "do_swap:" << do_swap;
#endif

            // Get the trigger relative index for this frame.
            int rel_index = ((int)pbMetaInfoPtr->getStartFrame() - (int)trigger_frame) + (int)index;

            // We want to add this frame to our list of markers, but squeeze
            // each run into a sequence of { first_frame, RUN_FLAG, last_frame }.

            // If it's the first or second, just copy - no squeezing yet.
            if (cur_count < 2)
            {
                marker_list.append(rel_index);
                handled = true;
            }

            // At this point, we know we have at least 3 entries in our list
            // (cur_count >= 2), so we can safely do the following math.
            if (!handled &&
                (marker_list.at(cur_count - 1) == (rel_index - 1)) &&
                (marker_list.at(cur_count - 2) == (rel_index - 2)))
            {
                // Now we know we have a run of three consecutive frame numbers,
                // so it's time to switch to the run encoding.
                marker_list.replace(cur_count - 1, INT_MAX);       // Use the maximum positive value - we're not likely to have 2GBytes worth of frames.
                marker_list.append(rel_index);
                handled = true;
            }

            // See if we're extending the run.
            if (!handled &&
                (marker_list.at(cur_count - 1) == (rel_index - 1)) &&
                (marker_list.at(cur_count - 2) == INT_MAX))
            {
                // Replace the last_frame value.
                marker_list.replace(cur_count - 1, rel_index);
                handled = true;
            }

            // No conditions apply, just copy it over.
            if (!handled)
                marker_list.append(rel_index);
        }

        // Remember this marker value.
        last_marker_word = marker_word;
    }

    // Return true if all of the frames were processed or false if not.
    return ((index == pbMetaInfoPtr->getFrameCount()) ? true : false);
}

char PlaybackMDI::getBitSel()
{
    if (pmMetaDataInfo.isNull() || (pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        return GIGE_BITMODE_UPPER_8_BITS;
    }

    // Return the bit selection from the metadata.
    return (char)pmMetaDataInfo->GetMetaInfo()->bit_sel;
}

/*!
 * @brief Provides external query for brightness setting provided in the last
 * setBrightness call (or the default when the object was constructed).
 *
 * @return The brightness value in the form of 0...200 representative of the
 * value a user might expect from reading register GIGE_REG_BRIGHTNESS.
 */
uint32_t PlaybackMDI::getDisplayBrightness()
{
    uint32_t displayBrightness = (uint32_t)brightness;
    return displayBrightness;
}

/*!
 * @brief Provides external query for contrast setting provided in the last
 * setBrightness call (or the default when the object was constructed).
 *
 * @return The contrast value in the form of 0...200 representative of the
 * value a user might expect from reading register GIGE_REG_CONTRAST.
 */
uint32_t PlaybackMDI::getDisplayContrast()
{
    uint32_t displayContrast = (uint32_t)contrast;
    return displayContrast;
}

/*!
 * @brief Provides external query for gamma setting provided in the last
 * setGamma call (or the default when the object was constructed).
 *
 * @return The gamma value in the form of xx.xx representative of the
 * value a user might expect from reading register GIGE_REG_GAMMA.
 */
uint32_t PlaybackMDI::getDisplayGamma()
{
    uint32_t displayGamma = (uint32_t)(gamma * 100.0);
    return displayGamma;
}

/*!
 * @brief Provides external query for display red gain setting provided in the last
 * setRed call (or the default when the object was constructed).
 *
 * @return The display red value representative of the
 * value a user might expect from reading register GIGE_REG_RED.
 */
uint32_t PlaybackMDI::getDisplayRed()
{
    uint32_t displayRed = GIGE_DEFAULT_RGB_GAIN;

    // We want to mimic a camera register setting.  When our r_gain is set to 0.0
    // it means we are using a custom lut and in that case for this call we will
    // return a default RGB Gain setting.  Otherwise calculate the display pseudo
    // register setting.
    if (r_gain != 0.0)
    {
        displayRed = (uint32_t)(r_gain * 256.0);
    }

    return displayRed;
}

/*!
 * @brief Provides external query for display green gain setting provided in the last
 * setGreen call (or the default when the object was constructed).
 *
 * @return The display green gain value representative of the
 * value a user might expect from reading register GIGE_REG_GREEN.
 */
uint32_t PlaybackMDI::getDisplayGreen()
{
    uint32_t displayGreen = GIGE_DEFAULT_RGB_GAIN;

    // We want to mimic a camera register setting.  When our g_gain is set to 0.0
    // it means we are using a custom lut and in that case for this call we will
    // return a default RGB Gain setting.  Otherwise calculate the display pseudo
    // register setting.
    if (g_gain != 0.0)
    {
        displayGreen = (uint32_t)(g_gain * 256.0);
    }

    return displayGreen;
}

/*!
 * @brief Provides external query for display blue gain setting provided in the last
 * setBlue call (or the default when the object was constructed).
 *
 * @return The display blue gain value representative of the
 * value a user might expect from reading register GIGE_REG_BLUE.
 */
uint32_t PlaybackMDI::getDisplayBlue()
{
    uint32_t displayBlue = GIGE_DEFAULT_RGB_GAIN;

    // We want to mimic a camera register setting.  When our b_gain is set to 0.0
    // it means we are using a custom lut and in that case for this call we will
    // return a default RGB Gain setting.  Otherwise calculate the display pseudo
    // register setting.
    if (b_gain != 0.0)
    {
        displayBlue = (uint32_t)(b_gain * 256.0);
    }

    return displayBlue;
}

int PlaybackMDI::getPfmdLen()
{
    // Sanity check. Probably the safe thing to do is return default length.
    if (pbMetaInfoPtr == nullptr)
        return GIGE_FRAME_METADATA_LEN;

    // Return our result.
    return pbMetaInfoPtr->pfmdLength();
}

/*!
 * @brief Determines if a custom lut is being applied to the playback
 * video.
 *
 * @return True if a custom lut is being applied.
 */
bool PlaybackMDI::isPlaybackLutInUse()
{
    // Really only care about the capLut - is it in use.
    if (capLuts)
    {
        // If data is from a series 3 camera or a monochrome sensor then only care about gamma.
        if (isSeries3() || isMono())
        {
            if (gamma == 0.0)
                return true;    // A custom gamma lut is in use.
        }
        else
        {
            // If any gain is not 0.0 we are using gain luts not custom luts.
            if (r_gain != 0.0 || g_gain != 0.0 || b_gain != 0.0)
            {
                return false;
            }
            // Falling into here means that all gains were 0.0 and a custom lut is in use.
            return true;
        }
    }
    // Not using custom luts.
    return false;
}

/*!
 * @brief Determines if the metadata associated with the playback is
 * supporting a monochrome sensor.  If no metadata is found we fallback
 * to a monochrome sensor.
 *
 * @return True if supporting a monochrome sensor.
 */
bool PlaybackMDI::isMono()
{
    // Sanity check. Probably the safe thing to do is return true.
    if (pbMetaInfoPtr == nullptr)
        return true;

    // Return our result.
    return pbMetaInfoPtr->isColor() ? false : true;
}

/*!
 * @brief Determines if the metadata associated with the playback is
 * supporting a series3 camera.  If no metadata is found we fallback
 * to a series3 camera as a default.
 *
 * @return True if supporting a series3 camera.
 */
bool PlaybackMDI::isSeries3()
{
    // Sanity check. Probably the safe thing to do is return true.
    if (pbMetaInfoPtr == nullptr)
        return true;

    // Return our result.
    return pbMetaInfoPtr->isSeries3();
}



/*!
 * @brief Shares a pointer to our metadata lut information if any exists.
 * This is basically a read-only shared pointer because there is no way to
 * modify the contents of an FiDisplayLut once created in its factory.
 *
 * @return The shared pointer to our metadata lut information or nullptr if
 * none exists.
 */
std::shared_ptr<FiDisplayLut> PlaybackMDI::GetMetadataLut()
{
    // Right now there is no additional processing we just return what
    // we have set in our object.  If no cap file is installed this will
    // be null and empty.
    return capLuts;
}

/*!
 * @brief We need to load a frame from the file to be displayed by the video_player object.
 *
 * @param frame_num     The absolute frame number of the frame we want displayed.
 * @param pixmap_ptr    Pixmap to load the frame into.
 * @param force_load    True to clear the cache so the frame is reloaded.
 *
 * @return True on success, false otherwise.
 */
bool PlaybackMDI::loadFrame(uint32_t frame_num, fiPixmap *pixmap_ptr, bool force_load)
{
    bool usingVideoContext = false;

    // Sanity check.
    if (pixmap_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "pixmap_ptr is nullptr!";
        return false;
    }

    // Ignore if it's loading the same frame again.
    if ((frame_num == last_frame_loaded) && !force_load)
        return false;

    // Remember this frame.
    last_frame_loaded = frame_num;

    // Make sure our gain information is set appropriately.
    updateGainInfo();

    // Tell the pixmap object the white balance setting.
    pixmap_ptr->set_gamma(gamma);
    pixmap_ptr->set_sensor_type(pbMetaInfoPtr->getSensorType());
    pixmap_ptr->set_whitebal(whitebal);

    // Do we have an AVI video stream reader?
    // We can only read AVI video files not any other types we may have saved like MP4s.
    if (AVIStreamReader)
    {
        usingVideoContext = true;

        // Adjust the frame number.
        frame_num = frame_num - pbMetaInfoPtr->getStartFrame();

        // Read the correct frame from the video stream.
        std::unique_ptr<FAS_VIDEO::FiImage> loadedFramePtr;

        loadedFramePtr = AVIStreamReader->LoadVideoFrame(frame_num);

        if (!loadedFramePtr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read video frame: " << frame_num;
            return false;
        }

        uint32_t imageDataSize = loadedFramePtr->GetAttachedImageSize();
        const uint8_t *memDataPtr = loadedFramePtr->GetAttachedImageData();

        // Try to load the pixmap data from the buffer.
        if (!pixmap_ptr->loadFromData(memDataPtr, imageDataSize))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the image for frame: " << frame_num;
            return false;
        }
    }

    // If not using a video context (AVI file) then process request differently.
    if (!usingVideoContext)
    {
        // Are we showing a single file?
        if (frame_count == 1)
        {
            // Load the image.
            if (!pixmap_ptr->load(root_file))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the image: " << root_file;
                return false;
            }
        }
        else
        {
            // Try to find the pixmap in our folder(s).
            if (!findImageFile(frame_num, pixmap_ptr))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to find the next image from frame: " << frame_num;
                return false;
            }
        }
    }

    // All done.
    return true;
}

/*!
 * @brief Try to locate the next frame, starting at frame_num, and load it into the destImage object.
 *
 * @param frame_num     Frame number we'll start looking. If it doesn't exist we'll keep moving.
 * @param destImage     Reference to the QImage we want loaded.
 * @return True on success, false otherwise.
 */
bool PlaybackMDI::loadPackedFrame(uint32_t frame_num, QImage& destImage)
{
    uint32_t    i;
    QString     image_filename;

    // Sanity check.
    if (review_controls == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "What? No review_controls?!";
        return false;
    }
    if (pbMetaInfoPtr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "What? No metadata?!";
        return false;
    }

    // We need to handle the case where files might be missing.
    for (i = 0; i < frame_count; i++)
    {
        QFileInfo   file_info;

        // Build our path differently if we're using subfolders.
        if (using_subfolders)
        {
            uint32_t    subdir_index;

            subdir_index = (frame_num - review_controls->get_clip_start()) / FI_SAVER::SUB_FOLDER_MAX_IMAGES;
            image_filename = QString::fromUtf8("%1%2%3/%4.%5").arg(stack_path, FiSaver::SUB_FOLDER_PREFIX)
                             .arg(subdir_index, 4, 10, QLatin1Char('0')).arg(frame_num, 7, 10, QLatin1Char('0'))
                             .arg(file_ext);
        }
        else
        {
            image_filename = QString::fromUtf8("%1%2.%3").arg(stack_path).arg(frame_num, 7, 10, QLatin1Char('0')).arg(file_ext);
        }

        // Check to see if the image file exists.
        file_info.setFile(image_filename);
        if (!file_info.exists())
        {
            // Do a more extensive search if we're using sub-folders.
            if (using_subfolders)
            {
                QString         tmp_file;

                // The file didn't exist where we thought it should. Let's see if we can
                // find it anywhere below our root folder. Build up the expected image filename.
                tmp_file = QString::fromUtf8("%1.%2").arg(frame_num, 7, 10, QLatin1Char('0')).arg(file_ext);
                QDirIterator    dir_walker(stack_path, QStringList(tmp_file), QDir::Files | QDir::NoSymLinks, QDirIterator::Subdirectories);

                // Walk through the folders.
                if (dir_walker.hasNext())
                {
                    // We found it!
                    image_filename = dir_walker.next();
                }
                else
                {
                    // Clear our file name string.
                    image_filename.clear();
                }
            }
            else
            {
                // Not using sub-folders, so the file simply doesn't exist.
                image_filename.clear();
            }
        }

        // Do we have a file name?
        if (!image_filename.isEmpty())
        {
            // Try loading the file.
            destImage.load(image_filename);
            if (destImage.isNull())
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the image: " << image_filename;
                image_filename.clear();
            }
            else
            {
                // We're done.
                break;
            }
        }

        // Move to the next frame number.
        frame_num++;
        if (frame_num > pbMetaInfoPtr->getEndFrame())
            frame_num = pbMetaInfoPtr->getStartFrame();
    }

    // Return true if we have a filename.
    return !image_filename.isEmpty();
}

/*!
 * @brief Read the per-frame metadata for the frame specified by the "frame_num"
 *        parameter.
 *
 * @param frame_num     The frame number of the frame we want the per-frame metadata for.
 */
void PlaybackMDI::loadPfmd(uint32_t frame_num)
{
    QByteArray  *pfmd_array;
    uint32_t    abs_frame;

    // Get the raw PFMD array.
    pfmd_array = pbMetaInfoPtr->getRawPFMD(frame_num, isSeries3());
    if (pfmd_array != nullptr)
    {
        // Pass it on.
        camera_info->set_pfmd(pfmd_array);
        delete pfmd_array;
    }

    // Convert the relative frame number into an absolute frame number.
    abs_frame = frame_num + pbMetaInfoPtr->getStartFrame();

    // Display the text version in the dock.
    pbMetaInfoPtr->displayPerFrameMetadata(abs_frame);

    // Have the video player update any overlay information.
    video_player->update_overlay(abs_frame, camera_info, pbMetaInfoPtr);
}

fiPixmap PlaybackMDI::mergeOverlay(uint32_t frame_num, fiPixmap merge_pixmap)
{
    QGraphicsScene  *scene_ptr;
    QByteArray      *pfmd_array;
    uint32_t        frame_adj;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_number:" << frame_num;

    // Do we need to adjust the frame number?
    if (frame_num < pbMetaInfoPtr->getStartFrame())
        frame_adj = frame_num;
    else
        frame_adj = frame_num - pbMetaInfoPtr->getStartFrame();

    // Get the raw PFMD array.
    pfmd_array = pbMetaInfoPtr->getRawPFMD(frame_adj, isSeries3());
    if (pfmd_array != nullptr)
    {
        // Pass it on.
        camera_info->set_pfmd(pfmd_array);
        delete pfmd_array;
    }

    // Add the pixmap to the scene.
    video_player->render_add_image(merge_pixmap);
    video_player->render_update(frame_num, camera_info, pbMetaInfoPtr);

    scene_ptr = video_player->render_get_scene();
    if (scene_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "scene_ptr is nullptr?!";
        return merge_pixmap;
    }

    // Grow our pixmap.
    fiPixmap    scene_pixmap((int)scene_ptr->sceneRect().width(), (int)scene_ptr->sceneRect().height());
    QPainter    scene_painter(&scene_pixmap);
    scene_painter.setRenderHint(QPainter::Antialiasing);
    scene_ptr->render(&scene_painter);

    // Copy the new pixmap over.
    return scene_pixmap;
}

void PlaybackMDI::updateGainInfo()
{
    // Sanity check.
    if (pbMetaInfoPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null metadata?!";
        return;
    }

    // See if the metadata indicates a "normal" CCM.
    if (pbMetaInfoPtr->isWbNormal())
    {
        // We have a "normal" CCM.
        whitebal = GIGE_WHITE_BAL_NORMAL;
    }

    // Grab the balance gains used.
    r_wb_gain = pbMetaInfoPtr->getBalanceRed() / 4096.0;
    g_wb_gain = pbMetaInfoPtr->getBalanceGreen() / 4096.0;
    b_wb_gain = pbMetaInfoPtr->getBalanceBlue() / 4096.0;

    // Update our LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief Play the video.
 */
void PlaybackMDI::playVideo()
{
    int playback_rate;

    // Do we have a playback timer? If we do, make sure it's stopped.
    if (playback_timer != nullptr)
        playback_timer->stop();

    if (review_controls->get_playback_rate() < DISPLAY_RATE)
    {
        playback_rate = 1000 / review_controls->get_playback_rate();
        play_frames_inc = 1;
    }
    else
    {
        playback_rate = PLAY_INTERVAL;
        play_frames_inc = review_controls->get_playback_rate() / DISPLAY_RATE;
    }

    // Do we need to allocate our timer?
    if (playback_timer == nullptr)
    {
        playback_timer = new QTimer(this);
        if (playback_timer == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a playback_timer!";
            return;
        }

        // Setup our connections.
        connect(playback_timer, &QTimer::timeout, this, &PlaybackMDI::timeToPlay);
    }

    // Start up our timer.
    playback_timer->start(playback_rate);

    // Get the buttons in the right state.
    review_controls->update_buttons(false);

    // Don't allow saves if the user is playing the video.
    MainWindow::GetInstance()->get_save()->setEnabled(false);
    MainWindow::GetInstance()->get_export_action()->setEnabled(false);
}

/*!
 * @brief Read the QSettings information to restore the non-volatile parameters from
 *        the "last" run.
 */
void PlaybackMDI::readSettings()
{
    QVariantList    settings_list;

    // Load our previous settings.
    QSettings settings;

    settings.beginGroup("EventMarkers");
    use_event_markers = settings.value("useEventMarkers", false).toBool();
    event_logic_is_and = settings.value("logicIsAnd", true).toBool();
    settings_list = settings.value("eventSelections").toList();
    settings.endGroup();

    // Restore the colorize checkbox.
    settings.beginGroup("Colorize");
    colorize(settings.value("enable", false).toBool());
    settings.endGroup();

    // Reset our list.
    event_selections.clear();

    // Now add in each of the entries we read from the settings.
    foreach (QVariant v, settings_list)
    {
        event_selections.append(v.toInt());
    }

    // Make sure we have a selection list.
    while (event_selections.count() < FM_EVENT_FILTERS::INDEX_COUNT)
    {
        event_selections.append(FM_EVENT_FILTERS::SELECT_IGNORE);
    }
}

/*!
 * @brief Forcefully re-load the current frame. You might want to do this if you've
 *        changed the look-up table, one of the gains, the gamma or some such setting.
 */
void PlaybackMDI::refreshFrame()
{
    fiPixmap    my_pixmap;

    try
    {
        // Clear the cache.
        QPixmapCache::clear();
    }
    catch (...)
    {
        // Don't really care about any errors from the clear.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "QPixmapCache::clear() emitted an error.";
    }

    // Re-Load the current frame.
    if (!loadFrame(last_frame_loaded, &my_pixmap, true))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the frame:" << last_frame_loaded;
        return;
    }

    // Make sure we got something.
    if (my_pixmap.isNull())
        return;

    // Are we supposed to white balance?
    if (do_whitebal)
    {
        // Auto white white balance.
        my_pixmap.auto_wb(&r_wb_gain, &g_wb_gain, &b_wb_gain);

        QLOG_TRACE() << __PRETTY_FUNCTION__ << "r_wb_gain:" << r_wb_gain << "g_wb_gain:" << g_wb_gain << "b_wb_gain:" << b_wb_gain;

        // Update our LUTs.
        buildLuts(__PRETTY_FUNCTION__);

        // Reset our flag.
        do_whitebal = false;
    }

    // Pass the image through the lookup table.
    my_pixmap.apply_lookup(r_lut, g_lut, b_lut);

    // Pass it on to the player.
    video_player->display_image(my_pixmap);

    // Are we showing per-frame metadata?
    if (pbMetaInfoPtr->perFrameIsAvail())
    {
        // Request the metadata.
        loadPfmd(last_frame_loaded - pbMetaInfoPtr->getStartFrame());
    }
}

/*!
 * @brief Reset the auto white balance gains back to default.
 */
void PlaybackMDI::resetAutoWbGains()
{
    // Put the white balance gains back to their defaults.
    r_wb_gain = DEFAULT_AWB_GAIN;
    g_wb_gain = DEFAULT_AWB_GAIN;
    b_wb_gain = DEFAULT_AWB_GAIN;

    // Rebuild our look-up tables.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief Save the images as specified by the SaveDialog object.
 * @param parent    Pointer to the [parent](@ref PlaybackRawMDI) object.
 */
void PlaybackMDI::SaveImages(QObject *parent)
{
    quint64     byte_count_stat;
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    double      bytes_per_sec;
    double      mbytes_per_sec;
    QString     stats_msg;
    QString     meta_filename;
    const QByteArray        *raw_buffer;
    FI_SAVER::SAVE_PARAMS   save_settings;
    QElapsedTimer   save_timer;
    PlaybackMDI  *my_class = qobject_cast<PlaybackMDI *>(parent);

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Get a pointer to the save dialog. Makes things more legible later on.
    std::shared_ptr<SaveDialog> saveDialog = my_class->save_dlg;
    if (!saveDialog)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_SAVE_DIALOG;
        return;
    }

    // Get our settings.
    // NOTE_JLC:  The get_save_params takes care of making sure that the setup for the
    // save is correct meaning there is no need to sanity check the imagesOutputType
    // or the imagesInputType because this function will make sure that they are valid
    // for the save operation.
    saveDialog->get_save_params(save_settings, my_class->isSeries3());

    // Allocate our saver object.
    my_class->saver = std::unique_ptr<FiSaver>(new FiSaver());

    // Make sure we have a saver to move forward with.
    if (!my_class->saver)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate saver object!";
        return;
    }

    // Initialize the saver object.
    my_class->saver->configure(save_settings);
    my_class->saver->setPath(save_settings.dest_path);
    my_class->saver->setFileExtension(save_settings.streamType);

    // NOTE:  It is important to set this streamtype where the output matches the
    // input in this case.  It is because this function takes care of translating the
    // image frame data internally and by the time it makes its way to the saver object
    // for saving the image data is already setup for the save.
    my_class->saver->setStreamType(save_settings.streamType, save_settings.imagesOutputType, save_settings.imagesOutputType);
    my_class->saver->setPlaybackRate(save_settings.playback_rate);

    my_class->saver->setH264Speed(save_settings.h264_speed);
    my_class->saver->setH264Quality(save_settings.h264_quality);
    my_class->saver->setH264EnableAllKeyframes(save_settings.h264_enable_all_keyframes);
    my_class->saver->setMetadataCfgSave(save_settings.save_metadata_cfg);
    my_class->saver->setStartIndex(0);
    if (save_settings.tag_label.isEmpty())
        my_class->saver->setTag(QString());
    else
        my_class->saver->setTag(save_settings.tag_label);
    if (save_settings.use_camera_name)
        my_class->saver->setCameraName(my_class->pbMetaInfoPtr->getCameraName());

    // Set the clip start to zero, 'cuz we'll be using absolute frame numbers as our sequence numbers.
    my_class->saver->setClipStart(0);

    // Tell the saver the number of frames.
    my_class->saver->setCount((saveDialog->get_clip_end() - saveDialog->get_clip_start()) + 1);
    my_class->saver->setClipEnd(my_class->saver->getFrameCount() - 1);

    // Do we need to merge the overlay information?
    if (save_settings.merge_overlays)
        my_class->video_player->render_start();

    // Set video stream specific parameters.
    if (FiSaver::isStreamVideo(save_settings.streamType))
    {
        const pm_metadata *pm_ptr = nullptr;

        // Determine the width and height of images for video streamers.
        if (save_settings.merge_overlays)
        {
            QGraphicsScene  *scene_ptr;

            // Get a pointer to our "scene".
            scene_ptr = my_class->video_player->render_get_scene();
            if (scene_ptr == nullptr)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "scene_ptr is nullptr?!";
                goto save_bail;
            }

            my_class->saver->setImageHeight((uint32_t)scene_ptr->sceneRect().height());
            my_class->saver->setImageWidth((uint32_t)scene_ptr->sceneRect().width());
        }
        else
        {
            my_class->saver->setImageHeight(my_class->pbMetaInfoPtr->getHeight());
            my_class->saver->setImageWidth(my_class->pbMetaInfoPtr->getWidth());
        }

        // Are we saving a packed stream?
        if (save_settings.streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
        {
            // Do we have our partition metadata structure?
            pm_ptr = my_class->pbMetaInfoPtr->getPmMetadata();
            if (pm_ptr == nullptr)
            {
                // See if we have the metadata from a CAP file.
                if (!my_class->pmMetaDataInfo.isNull())
                {
                    // Try retrieving the metadata pointer.
                    pm_ptr = my_class->pmMetaDataInfo->GetMetaInfo();
                }

                // If we got the metadata pointer, we're done. Otherwise, there
                // is something else we can try.
                if (pm_ptr == nullptr)
                {
                    // Recover the pm_metadata structure.
                    my_class->pbMetaInfoPtr->recoverPmMetadata();
                    pm_ptr = my_class->pbMetaInfoPtr->getPmMetadata();
                }
            }

            // The packed writer needs to know what the starting frame number of the clip is.
            my_class->saver->setClipStart(saveDialog->get_clip_start());

            // Prepare the packed stream writer object.
            if (my_class->saver->preparePackedWriter(pm_ptr) < 0)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "PrepareStreamWriter() failed!";
                goto save_bail;
            }
        }
        else
        {
            // Prepare the video stream writer object.
            if (my_class->saver->prepareStreamWriter() < 0)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "PrepareStreamWriter() failed!";
                goto save_bail;
            }
        }

        // Update our dialog.
        if (my_class->xferDialog)
            my_class->xferDialog->setXferDestValue(my_class->saver->getLocalPath());
    }
    else if (my_class->saver->createDestPath() < 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "create_dest_path() failed!";
        goto save_bail;
    }

    // Reset the save complete flags for stream video saving.
    my_class->wait_for_save_complete    = false;
    my_class->save_had_error            = false;
    my_class->save_is_complete          = false;
    my_class->saver_count = my_class->saver->getFrameCount();

    // Initialize our values for our statistics.
    byte_count_stat = 0;
    save_timer.start();

    try
    {
        // Clear the cache.
        QPixmapCache::clear();
    }
    catch (...)
    {
        // Don't really care about any errors from the clear.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "QPixmapCache::clear() emitted an error.";
    }

    // Tell our metadata object to ignore the "export" information.
    my_class->pbMetaInfoPtr->ignoreExport(true);

    // Loop through all the frames. Make sure we're using absolute (not relative) frame numbers.
    for (my_class->saver_frame = saveDialog->get_clip_start(); !my_class->thread_stop && (my_class->saver_frame <= saveDialog->get_clip_end()); ++my_class->saver_frame)
    {
        QImage      my_packedData;
        QByteArray  my_buffer;
        fiPixmap    my_pixmap;
        uint32_t    my_seqNbr;
        const uint8_t *dataPtr;
        uint32_t    dataLen;

        // Handle things differently for packed data.
        if (save_settings.streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
        {
            // Load the frame. Make sure it's the "raw" data, so we need to force the load.
            if (!my_class->loadPackedFrame(my_class->saver_frame, my_packedData))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the frame.";
                goto save_bail;
            }

            // Setup our pointer and length.
            dataPtr = my_packedData.constBits();
            dataLen = my_packedData.sizeInBytes();

            // Accumulate the byte count.
            byte_count_stat += (quint64)dataLen;

            // Set the sequence number. For packed data the sequence starts at zero.
            my_seqNbr = my_class->saver_frame - saveDialog->get_clip_start();

            // Save the data.
            // However if it fails we will still process through the images
            // but the update progress will display error condition and the
            // fact that nothing is being updated should cause user to cancel.
            if (!my_class->saver->saveToPackedStream(dataPtr, dataLen, my_seqNbr, false))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "saveToPackedStream() failed! seqNbr:" << my_seqNbr << "frame:" << my_class->saver_frame;
            }
        }
        else
        {
            // Load the frame. Make sure it's the "raw" data, so we need to force the load.
            if (!my_class->loadFrame(my_class->saver_frame, &my_pixmap, true))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the frame.";
                goto save_bail;
            }

            // Do we need to merge the overlay information?
            if (save_settings.merge_overlays)
            {
                my_pixmap = my_class->mergeOverlay(my_class->saver_frame, my_pixmap);
            }

            // How are we saving this frame.
            switch (save_settings.streamType)
            {
            case FI_SAVER::STREAM_TYPE::MP4:
            case FI_SAVER::STREAM_TYPE::AVI:
            case FI_SAVER::STREAM_TYPE::BMP_STACK:
            case FI_SAVER::STREAM_TYPE::JPEG_STACK:
            case FI_SAVER::STREAM_TYPE::TIF_STACK:
                // Reset our buffer.
                my_buffer.clear();

                // Update the quality factor. This will only be used if we're saving JPEGs.
                my_pixmap.set_jpeg_quality(save_settings.jpeg_quality);

                // Get the pixmap as a qImage.  Note that if the image format has a header
                // this data will include that header information.  When the frame is loaded
                // above from the stack it does not include the header.
                my_pixmap.save_image(&my_buffer, save_settings.imagesOutputType);
                break;
            case FI_SAVER::STREAM_TYPE::DNG_STACK:
            case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
                // Nothing to do my_buffer is already set up for this operation.
                break;
            default:
                break;
            }

            // Setup our pointer and length.
            dataPtr = (const uint8_t *)my_buffer.data();
            dataLen = (uint32_t)my_buffer.length();

            // Accumulate the byte count.
            byte_count_stat += (quint64)dataLen;

            // Set the sequence number.
            my_seqNbr = my_class->saver_frame + 1;

            // Save the image.
            // However if it fails we will still process through the images
            // but the update progress will display error condition and the
            // fact that nothing is being updated should cause user to cancel.
            if (!my_class->saver->saveToStream(dataPtr, dataLen, my_seqNbr, false))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "SaveToStream() failed!";
            }

            // Are we saving a stack of images?
            if (FiSaver::isStackOfImages(save_settings.streamType))
            {
                // Update our image file name for next image in the frame stack.
                my_class->saver_image_filename = my_class->saver->getImageFilename();
            }
        }
    } // End of the for loop processing the images to save.

    // If saving a stream video there are frame flushes that may
    // occur and this must be handled.
    if (my_class->saver->isStreamVideo(save_settings.streamType))
    {
        my_class->wait_for_save_complete = true;
        my_class->save_had_error = true;
        while (!my_class->thread_stop && !my_class->save_is_complete)
        {
            // If the saver does not show an error condition and the
            // we are still waiting for save_is_complete pause our
            // thread.
            if (!my_class->save_had_error)
            {
                my_class->saver_thread.suspend();
            }
        }
    }

    // Do we need to clean overlay stuff up?
    if (save_settings.merge_overlays)
        my_class->video_player->render_end();

    // Are we cancelling?
    if (my_class->thread_stop)
        goto save_bail;

    // Get the elapsed time and generate statistics.
    elapsed = save_timer.elapsed();
    elapsed_time = elapsed_time.addMSecs((int)elapsed);

    // How many bytes per second.
    bytes_per_sec = byte_count_stat / ((double)elapsed / 1000.0);
    mbytes_per_sec = (bytes_per_sec / (1024.0 * 1024.0));

    // Build up our output string.
    stats_msg = QString::fromUtf8("Duration: %1  TotalBytes: %2  TotalFrames: %3  EffectiveRate: %4 MB/s  Format: %5")
                .arg(elapsed_time.toString(QString("hh:mm:ss.zzz"))).arg(byte_count_stat).arg((saveDialog->get_clip_end() - saveDialog->get_clip_start()) + 1).arg(mbytes_per_sec).arg(my_class->saver->getStreamType(), 0, 10);

    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");
    QLOG_INFO() << __PRETTY_FUNCTION__ << my_class->saver->getLocalPath();
    QLOG_INFO() << __PRETTY_FUNCTION__ << stats_msg;
    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");

    // The packed stream does not have a separate metadata text file.
    if (save_settings.streamType != FI_SAVER::STREAM_TYPE::RDI_PACKED)
    {
        // Build up the TXT metafile name.
        meta_filename = my_class->saver->getLocalPath();
        if (save_settings.streamType == FI_SAVER::STREAM_TYPE::AVI)
        {
            // Replace the ".avi" with ".txt".
            meta_filename.replace(meta_filename.lastIndexOf(FiSaver::IMAGE_EXT_AVI), FiSaver::IMAGE_EXT_AVI.length(), FiSaver::IMAGE_EXT_META_TEXT);
        }
        else if (save_settings.streamType == FI_SAVER::STREAM_TYPE::MP4)
        {
            // Replace the ".mp4" with ".txt".
            meta_filename.replace(meta_filename.lastIndexOf(FiSaver::IMAGE_EXT_MP4), FiSaver::IMAGE_EXT_MP4.length(), FiSaver::IMAGE_EXT_META_TEXT);
        }
        else
        {
            // Just using a standard name.
            meta_filename = meta_filename + FiSaver::DEFAULT_METADATA_FILE_NAME;
        }

        // Save the capture metadata.
        my_class->pbMetaInfoPtr->saveMetadata(meta_filename);

        // Save the raw metadata.
        raw_buffer = my_class->pbMetaInfoPtr->getRawBuffer();
        if (raw_buffer != nullptr)
        {
            const char  *raw_ptr;
            uint32_t    frame_count;
            uint32_t    frame_num;
            QByteArray  metadata_array;

            // Get a pointer to our data.
            raw_ptr = raw_buffer->constData();

            // How many frames are we saving.
            frame_count = saveDialog->get_clip_end() - saveDialog->get_clip_start();

            // Build up the array.
            for (frame_num = 0; frame_num <= frame_count; ++frame_num)
            {
                uint32_t    index;

                // Calculate the index for this frame in our raw data array.
                index = (saveDialog->get_clip_start() - my_class->pbMetaInfoPtr->getStartFrame()) + frame_num;

                // Copy the PFMD for this frame over to our new array.
                metadata_array.append(&raw_ptr[index * (uint32_t)my_class->getPfmdLen()], my_class->getPfmdLen());
            }

            // Save the data.
            my_class->saver->saveRawMetadata(&metadata_array);
        }
    }

    // Wrap up the save.
    my_class->saver->finishSave();

save_bail:
    // Tell our metadata object to pay attention to the "exports" section now.
    my_class->pbMetaInfoPtr->ignoreExport(false);

    // We are done and leaving let our class know.
    // TODO:  Need to put a valid valud in the return but for now it's ok.
    emit my_class->saveImagesComplete(true);
}

/*!
 * @brief The saver has completed its image save and the thread completed.
 */
void PlaybackMDI::saveImagesCompleted(bool savedOK)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "savedOK:" << savedOK;

    // There should have only been one connection set up when it was started so we are done with it now.
    disconnect(this, &PlaybackMDI::saveImagesComplete, this, &PlaybackMDI::saveImagesCompleted);

    // Close down any progress dialogs that were up.
    DoneWithProgressDialogs();

    // Clear out any "export" metadata we may have created.
    if (pbMetaInfoPtr != nullptr)
        pbMetaInfoPtr->resetExports();

    // For testing purposes only.  This will popup a message box about status of the save.
    //QMessageBox::information(this, tr("Save Images"), tr("Status: %1").arg(savedOK));
}

/*!
 * @brief Save the user specified settings to the non-volatile parameter storage.
 */
void PlaybackMDI::saveSettings()
{
    int     index;
    QVariantList    settings_list;

    // Initialize our variant list.
    for (index = 0; index < event_selections.count(); ++index)
        settings_list.append(event_selections.at(index));

    QSettings settings;
    settings.beginGroup("EventMarkers");
    settings.setValue("useEventMarkers", use_event_markers);
    settings.setValue("logicIsAnd", event_logic_is_and);
    settings.setValue("eventSelections", settings_list);
    settings.endGroup();
}

/*!
 * @brief Adjust the blue gain value based on the new_value parameter.
 *        This function is a simulation of the Camera Server GIGE_REG_BLUE
 *        functionality.
 *
 * @param new_value     The new gain, formated as an 8.8 value.
 */
void PlaybackMDI::setBlue(uint32_t new_value)
{
    // Save our new gain value.
    b_gain = (double)new_value / 256.0;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "b_gain" << b_gain;

    // Clear the LUT.
    b_gain_lut.clear();

    // Update the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief Installs a Red, Blue and Green LUT array and builds the luts one time in one call.
 *
 * @param lutRed    The array of red LUTs.
 * @param lutGreen  The array of green LUTS.
 * @param lutBlue   The array of blue LUTS.
 *
 * @returns True if the 8-bit rgb luts were processed or false if there was an error.
 */
bool PlaybackMDI::set8bitRgbLuts(const QVector<double>& lut8BitRed, const QVector<double>& lut8BitGreen, const QVector<double>& lut8BitBlue)

{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "r_gain_old" << r_gain_old << "r_gain" << r_gain;
    int             lut8BitTableSize = LUT_MAX_TABLE_COUNT / 4; //!< Size of an 8-bit LUT table vector.

    // Make sure the vectors provided are sized correctly.
    if (lut8BitRed.count() != lut8BitTableSize ||
        lut8BitGreen.count() != lut8BitTableSize ||
        lut8BitBlue.count() != lut8BitTableSize)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "One or more of the RGB LUT vectors was too short!";
        return false;
    }

    // Save the last "good" r_gain to use if colorization is selected.
    if (r_gain != 0.0)
        r_gain_old = r_gain;

    // Marking the gain settings as 0.0 indicates a custom lut table exists for that channel.
    r_gain = 0.0;
    g_gain = 0.0;
    b_gain = 0.0;

    // Install all the gains at once.
    r_gain_custom_lut = r_gain_lut = lut8BitRed;    // For red channel also create a backup of the custom LUT.
    g_gain_lut = lut8BitGreen;
    b_gain_lut = lut8BitBlue;

    // Now build the LUTs one time for all three settings.
    buildLuts(__PRETTY_FUNCTION__);
    return true;
}

/*!
 * @brief Installs a Red, Blue and Green LUT array and builds the luts one time in one call.
 * Important:  This expects 10-bit vector LUT arrays as input.
 *
 * @param lutRed    The array of 10-bit red LUTs.
 * @param lutGreen  The array of 10-bit green LUTS.
 * @param lutBlue   The array of 10-bit blue LUTS.
 */
bool PlaybackMDI::set10BitRgbLuts(const QVector<double>& lut10BitRed, const QVector<double>& lut10BitGreen, const QVector<double>& lut10BitBlue)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "r_gain_old" << r_gain_old << "r_gain" << r_gain;

    // Because we only work in 8-bit LUT world these 10-bit LUTs have to be "subsampled" and converted
    // into playback's 8-bit LUTs.
    int             lut8BitTableSize = LUT_MAX_TABLE_COUNT / 4; //!< Size of an 8-bit LUT table vector.

    // Allocate the 8-bit LUT vectors to hold the data. The lookup table takes a byte
    // value in, and converts that to a byte value going out, so it's smaller than if we
    // were on the camera.
    QVector<double> lut8BitRedOrMono(lut8BitTableSize); //!< Red or monochrome LUT table.
    QVector<double> lut8BitGreen(lut8BitTableSize);     //!< Green LUT table.
    QVector<double> lut8BitBlue(lut8BitTableSize);      //!< Blue LUT table.

    // Make sure the input passed in is a vector LUT containing 10-bit data.
    if ((lut10BitRed.count() != LUT_MAX_TABLE_COUNT) || (lut10BitGreen.count() != LUT_MAX_TABLE_COUNT) || (lut10BitBlue.count() != LUT_MAX_TABLE_COUNT))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "One or more of the 10-bit LUT vectors was an incorrect size!";
        return false;
    }
    // Subsample each red value and place it in the lut to process for red channel.
    for (int i = 0; i < lut8BitTableSize; ++i)
        lut8BitRedOrMono[i] = (double)(lut10BitRed[i * 4]);

    // Subsample each green value and place it in the lut to process for green channel.
    for (int i = 0; i < lut8BitTableSize; ++i)
        lut8BitGreen[i] = (double)(lut10BitGreen[i * 4]);

    // Subsample each blue value and place it in the lut to process for blue channel.
    for (int i = 0; i < lut8BitTableSize; ++i)
        lut8BitBlue[i] = (double)(lut10BitBlue[i * 4]);

    // Pass the RGB LUTs to the playback engine.
    set8bitRgbLuts(lut8BitRedOrMono, lut8BitGreen, lut8BitBlue);

    return true;
}


/*!
 * @brief Provides ability to turn on or off the ability to apply any internal
 * CAP custom LUT and it is mainly used for external reasons.
 *
 * NOTE: There are conditions that can occur externally where we would no longer
 * wish to apply a custom capture lut during the lifetime scope of this playback
 * window.  This function provides support to turn this on or off.  Initially
 * the setting of the PlaybackMDI::applyingCapLuts flag is true because we would
 * want to apply any custom lut that exists in the metadata but as a user interacts
 * with a display panel over this window it may become important to no longer apply
 * this custom lut.  One case where we would turn this off would be when a user
 * resets display settings - this will reset any custom lut condition.  When we
 * focus in and out of playback windows there exists a condition where we need to
 * set up the custom lut applied if it is applied - but resetting display settings
 * severed that condition as long as the window is still displayed.  The user can
 * choose to create a new lut, use the old lut, or reset the display settings.
 *
 * @param applyingCapLut - how to set, true or false.
 */
void PlaybackMDI::setApplyingCapLuts(bool applyingCapLutState)
{
    applyingCapLuts = applyingCapLutState;
}

/*!
 * @brief Adjust the brightness setting based on the new_value parameter.
 *        This function is a simulation of the Camera Server GIGE_REG_BRIGHTNESS
 *        functionality.
 *
 * @param new_value     The new brightness value.
 */
void PlaybackMDI::setBrightness(uint32_t new_value)
{
    // Save the new value.
    brightness = (double)new_value;

    // By updating this value, we're are no longer using a LUT.
    if (!gamma_lut.isEmpty())
    {
        gamma = GIGE_DEFAULT_GAMMA / 100.0;
        gamma_lut.clear();
    }

    // Update the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief Adjust the contrast setting based on the new_value parameter.
 *        This function is a simulation of the Camera Server GIGE_REG_CONTRAST
 *        functionality.
 *
 * @param new_value     The new contrast value.
 */
void PlaybackMDI::setContrast(uint32_t new_value)
{
    // Save the new value.
    contrast = (double)new_value;

    // By updating this value, we're are no longer using a LUT.
    if (!gamma_lut.isEmpty())
    {
        gamma = GIGE_DEFAULT_GAMMA / 100.0;
        gamma_lut.clear();
    }

    // Update the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief Adjust the gamma setting based on the new_value parameter.
 *        This function is a simulation of the Camera Server GIGE_REG_GAMMA
 *        functionality.
 *
 * @param new_value     The new gamma value.
 */
void PlaybackMDI::setGamma(uint32_t new_value)
{
    // Save the new value. The incoming format is xx.xx, and we
    // want to change that to a double.
    gamma = (double)new_value / 100.0;

    // Clear the LUT.
    gamma_lut.clear();

    // Update the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

bool PlaybackMDI::set8bitGammaLut(const QVector<double>& lut8BitGamma)
{
    int lut8BitTableSize = LUT_MAX_TABLE_COUNT / 4; //!< Size of an 8-bit LUT table vector.

    // Make sure the vectors provided are sized correctly.
    if (lut8BitGamma.count() != lut8BitTableSize)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The Gamma LUT vector was too short!";
        return false;
    }

    // Set the gamma value to a known bad value.
    gamma = 0.0;

    // Copy the 8-bit LUT to our gamma_lut.
    gamma_lut = lut8BitGamma;

    // Build up the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
    return true;
}

bool PlaybackMDI::set10BitGammaLut(const QVector<double>& lut10BitGamma)
{
    // Because we only work in 8-bit gamma LUT world this 10-bit LUTs has to be "subsampled" and converted
    // into playback's 8-bit gamma LUTs.
    int lut8BitTableSize = LUT_MAX_TABLE_COUNT / 4; //!< Size of an 8-bit LUT table vector.

    // Allocate the 8-bit LUT gamma vector to hold the data. The lookup table takes a byte
    // value in, and converts that to a byte value going out, so it's smaller than if we
    // were on the camera.
    QVector<double> lut8BitGamma(lut8BitTableSize);     //!< 8-bit Gamma LUT table.

    // Make sure the input passed in is a vector LUT containing 10-bit data.
    if ((lut10BitGamma.count() != LUT_MAX_TABLE_COUNT))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The gamma 10-bit LUT vector was an incorrect size!";
        return false;
    }

    // Subsample the gamma 10-bit LUT information and place it in our 8-bit gamma vector.
    for (int i = 0; i < lut8BitTableSize; ++i)
        lut8BitGamma[i] = (double)(lut10BitGamma[i * 4]);

    // Set the gamma value to a known bad value.
    gamma = 0.0;

    // Copy the 8-bit subsampled LUT over to our gamma_lut.
    gamma_lut = lut8BitGamma;

    // Build up the LUTs.
    buildLuts(__PRETTY_FUNCTION__);

    return true;
}

/*!
 * @brief Specify the file name we're supposed to use for playback.
 * @param filename  The fully qualified path to the video we want to load.
 */
bool PlaybackMDI::setFile(const QString& filename)
{
    int         cur_framenum;
    QString     file_basename;
    QString     meta_filename;
    QFileInfo   metadata_file_info;
    QFileInfo   file_info(filename);

    // Save the file name for later.
    root_file = filename;

    // Set up some local state.
    is_video_file = true;
    has_pfmd_pm_file = false;
    has_metadata_txt_file = false;

    // Allocate our metadata structure.
    if (pbMetaInfoPtr == nullptr)
    {
        pbMetaInfoPtr = new PlaybackMetaInfo(this);
        if (pbMetaInfoPtr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a PlaybackMetaInfo object!";
            return false;
        }
    }

    // Tell the main window about our widget.
    MainWindow::GetInstance()->set_metadata_info(pbMetaInfoPtr);

    // Determine if we are playing an AVI or a stack of images. This will affect everything we do.
    file_ext = file_info.suffix().toLower();
    if (file_ext.compare(FiSaver::IMAGE_EXT_AVI, Qt::CaseInsensitive) == 0)
    {
        // Set our title to the file we are playing.
        setWindowTitle(QDir::toNativeSeparators(filename));

        // Create the video stream reader.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Creating Video Stream Reader for video:  " << filename;

        // We don't know the type of images yet for the stream reader but we do know it
        // will be one of the motion jpeg codec types because it's an AVI file.  When we
        // add support for the MP4 video streams we will need to differentiate.
        if (!AVIStreamReader)
            AVIStreamReader = FAS_VIDEO::FiVideo::CreateAVIVideoStreamReader(filename.toStdString());

        // If unable to create the video stream writer log an error.
        if (!AVIStreamReader)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create AVI Video Stream Reader for video: " << filename;
            return false;
        }

        // Setup our metadata files.
        file_basename = file_info.path() + QString::fromUtf8("/") + file_info.baseName() + QString::fromUtf8(".");

        // Process the partition metadata TXT file.
        meta_filename = file_basename + FiSaver::IMAGE_EXT_META_TEXT;
        metadata_file_info.setFile(meta_filename);
        if (metadata_file_info.exists())
            has_metadata_txt_file = true;
        pbMetaInfoPtr->setMetadataFile(meta_filename);

        // Process the XML metadata file if it exists.
        meta_filename = file_basename + FiSaver::IMAGE_EXT_META_XML;
        pbMetaInfoPtr->setXmlMetadataFile(meta_filename);

        // Process the PM - per-frame metadata file.
        meta_filename = file_basename + FiSaver::IMAGE_EXT_META_RAW;
        metadata_file_info.setFile(meta_filename);
        if (metadata_file_info.exists())
            has_pfmd_pm_file = true;
        pbMetaInfoPtr->setRawMetadataFile(meta_filename, getPfmdLen());

        // Set the state of existence for these ancillary metadata files.
        camera_info->set_metadata_file_info(has_pfmd_pm_file, has_metadata_txt_file);

        // Remember the number of frames in the video.
        frame_count = AVIStreamReader->GetFrameCount();

        // Is there a valid metadata file?
        if (pbMetaInfoPtr->captureMetadataIsAvail())
        {
            trigger_frame = (int)pbMetaInfoPtr->getTriggerFrame();
            review_controls->set_pb_trigger(pbMetaInfoPtr->getTriggerFrame());
            review_controls->set_pb_clip_start(pbMetaInfoPtr->getStartFrame());
            review_controls->set_pb_clip_end(pbMetaInfoPtr->getEndFrame());
            review_controls->set_pb_frame_rate(pbMetaInfoPtr->getFramerate());

            // Update the summary camera information.
            camera_info->set_summary(pbMetaInfoPtr->getWidth(), pbMetaInfoPtr->getHeight(), pbMetaInfoPtr->getFramerate(), pbMetaInfoPtr->getExposure(),
                                     pbMetaInfoPtr->getTriggerFrame() | GIGE_TRIGGER_REQ_FRAME_MASK, pbMetaInfoPtr->getSensorGain());
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No capture metadata available.";

            trigger_frame = 0;
            review_controls->set_pb_trigger(0);
            review_controls->set_pb_clip_start(0);
            review_controls->set_pb_clip_end(frame_count - 1);
            review_controls->set_pb_frame_rate(AVIStreamReader->GetFrameRate());

            // Do not enable the PFMD display.
            camera_info->enable_pfmd(false);

            // If there is no TXT file do not allow for summary.  It has side effects
            // that ripple through the system.
            camera_info->enable_summary(has_metadata_txt_file);
        }

        // Save the trigger as our "current" frame.
        cur_framenum = trigger_frame;
    }
    else
    {
        // We are working on an image stack, we need to know if sub folders are being used.
        if (filename.indexOf(QString("/") + FiSaver::SUB_FOLDER_PREFIX, Qt::CaseInsensitive) == -1)
            using_subfolders = false;
        else
            using_subfolders = true;

        // Figure out what image number we are working on. All image stack files we generate end
        // with "_nnn". If we are using subfolders this will also return the parent as the path
        // rather than the real directory.
        if (splitFileName(filename, stack_path, cur_framenum))
        {
            // Since the user "might" have removed files, walk through the folder(s) and find the
            // actual start/end frame numbers.
            frame_count = (uint32_t)calculateMinMaxFrameNumber(filename, (int&)min_framenum, (int&)max_framenum);
        }
        else
        {
            cur_framenum = 0; // since it's invalid, it indicates that we are only playing a single image.
            frame_count = 1;
            min_framenum = 0;
            max_framenum = 0;
        }

        // Set our title to the folder we are playing.
        setWindowTitle(QDir::toNativeSeparators(stack_path));

        // Load up the metadata.

        // Process the metadata.txt file.
        meta_filename = stack_path + FiSaver::DEFAULT_METADATA_FILE_NAME;
        pbMetaInfoPtr->setMetadataFile(meta_filename);
        metadata_file_info.setFile(meta_filename);
        if (metadata_file_info.exists())
            has_metadata_txt_file = true;

        // Process the metadata.xml file.
        meta_filename.replace(meta_filename.lastIndexOf(FiSaver::IMAGE_EXT_META_TEXT), FiSaver::IMAGE_EXT_META_TEXT.length(), FiSaver::IMAGE_EXT_META_XML);
        pbMetaInfoPtr->setXmlMetadataFile(meta_filename);

        // Process the metadata.pm file.
        meta_filename.replace(meta_filename.lastIndexOf(FiSaver::IMAGE_EXT_META_XML), FiSaver::IMAGE_EXT_META_XML.length(), FiSaver::IMAGE_EXT_META_RAW);
        pbMetaInfoPtr->setRawMetadataFile(meta_filename, getPfmdLen());
        metadata_file_info.setFile(meta_filename);
        if (metadata_file_info.exists())
            has_pfmd_pm_file = true;

        // Set the state of existence for these ancillary metadata files.
        camera_info->set_metadata_file_info(has_pfmd_pm_file, has_metadata_txt_file);

        // Setup our variables.
        trigger_frame = (int)pbMetaInfoPtr->getTriggerFrame();

        // Is there a valid metadata file?
        if (pbMetaInfoPtr->captureMetadataIsAvail())
        {
            review_controls->set_pb_trigger(pbMetaInfoPtr->getTriggerFrame());
            review_controls->set_pb_frame_rate(pbMetaInfoPtr->getFramerate());

            // Update the summary camera information.
            camera_info->set_summary(pbMetaInfoPtr->getWidth(), pbMetaInfoPtr->getHeight(), pbMetaInfoPtr->getFramerate(), pbMetaInfoPtr->getExposure(),
                                     pbMetaInfoPtr->getTriggerFrame() | GIGE_TRIGGER_REQ_FRAME_MASK, pbMetaInfoPtr->getSensorGain());
        }
        else
        {
            review_controls->set_pb_trigger((uint32_t)cur_framenum);
            review_controls->set_pb_frame_rate(CAM_DEF_REC_RATE);

            // Do not enable the PFMD display.
            camera_info->enable_pfmd(false);

            // If there is no metadata.txt file do not allow for summary.  It has side effects
            // that ripple through the system.
            camera_info->enable_summary(has_metadata_txt_file);
        }

        review_controls->set_pb_clip_start(min_framenum);
        review_controls->set_pb_clip_end(max_framenum);
    }

    // Save away some of the information for later.
    review_controls->set_pb_frame_count(frame_count);

    // Enable the playback buttons.
    review_controls->update_buttons(true);

    // Allocate an action for the window's menu to use.
    if (playback_action != nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "playback_action is not null!";
        delete playback_action;
    }
    playback_action = new QAction(this);
    playback_action->setCheckable(true);
    connect(playback_action, &QAction::triggered, this, [this]() { show(); });
    connect(playback_action, &QAction::triggered, this, [this]() { setFocus(); });
    playback_action->setText(filename);

    // Jump to the trigger frame.
    review_controls->set_current_frame_num((uint32_t)cur_framenum);

    // Update the review windows graphing information.
    review_controls->update_graph();

    // Turn off the autosave settings.
    camera_info->display_autosave_settings(0);

    // Show/hide the display info.
    MainWindow::GetInstance()->get_view_summary()->setChecked(camera_info->summary_is_enabled());
    MainWindow::GetInstance()->get_view_pfmd()->setChecked(camera_info->get_pfmd_setting());
    MainWindow::GetInstance()->get_view_local_frametime()->setChecked(camera_info->localtime_is_enabled());
    camera_info->show_info();

    // If we're enabling the display, make sure we grab the metadata for the current frame.
    if (camera_info->pfmd_is_enabled())
    {
        // Update the per-frame metadata display.
        updatePfmd();

        // Update our event information.
        updateEventInfo();
    }

    // Only enable the save if there's metadata available.
    if (pbMetaInfoPtr && pbMetaInfoPtr->captureMetadataIsAvail())
    {
        if (is_video_file)
        {
            MainWindow::GetInstance()->get_save()->setEnabled(has_metadata_txt_file);
            MainWindow::GetInstance()->get_export_action()->setEnabled(has_metadata_txt_file);
            camera_info->enable_summary(has_metadata_txt_file);
            MainWindow::GetInstance()->get_view_summary()->setEnabled(has_metadata_txt_file);
        }
        else
        {
            MainWindow::GetInstance()->get_save()->setEnabled(true);
            MainWindow::GetInstance()->get_export_action()->setEnabled(true);
            camera_info->enable_summary(true);
            MainWindow::GetInstance()->get_view_summary()->setEnabled(true);
        }
        MainWindow::GetInstance()->get_view_summary()->setChecked(camera_info->summary_is_enabled());
    }
    else
    {
        MainWindow::GetInstance()->get_save()->setEnabled(false);
        MainWindow::GetInstance()->get_export_action()->setEnabled(false);
        camera_info->enable_summary(false);
        MainWindow::GetInstance()->get_view_summary()->setEnabled(false);
        MainWindow::GetInstance()->get_view_summary()->setChecked(false);
    }
    camera_info->show_info();

    // Allow overlays
    video_player->enable_overlay(true);

    // Make sure our gain information is set appropriately.
    updateGainInfo();

    // All done.
    return true;
}

/*!
 * @brief Adjust the green gain value based on the new_value parameter.
 *        This function is a simulation of the Camera Server GIGE_REG_GREEN
 *        functionality.
 *
 * @param new_value     The new gain, formated as an 8.8 value.
 */
void PlaybackMDI::setGreen(uint32_t new_value)
{
    // Save our new gain value.
    g_gain = (double)new_value / 256.0;

    // Clear the LUT.
    g_gain_lut.clear();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "g_gain" << g_gain;

    // Update the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief We are being made inactive. If we are actively streaming, we need to slow the rate down by
 *        switching to background mode. In addition, we do NOT want to respond to certain signals.
 */
void PlaybackMDI::setInactive()
{
    ControlPanel    *control_panel;
    PanelDisplay    *display_panel;

    // Disable the zoom actions.
    MainWindow::GetInstance()->enable_zoom_actions(false);

    // Disconnect the "active" actions we want.
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_save, this, &PlaybackMDI::doImageSave);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_review_controls, this, &PlaybackMDI::showControls);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_summary, this, &PlaybackMDI::showSummary);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_aiming_cursor, this, &PlaybackMDI::doViewAimingCursor);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_info_cursor, this, &PlaybackMDI::doViewInfoCursor);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_pfmd, this, &PlaybackMDI::showPfmd);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::setup_event_markers, this, &PlaybackMDI::doPlaybackEvents);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_local_frametime, this, &PlaybackMDI::showLocalFrametime);

    disconnect(video_player->get_action_overlay(), &QAction::triggered, video_player, &CameraVideo::do_add_overlay);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_1_to_1, video_player, &CameraVideo::do_zoom_1_to_1);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_in, video_player, &CameraVideo::do_zoom_in);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_out, video_player, &CameraVideo::do_zoom_out);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_fit, video_player, &CameraVideo::do_zoom_to_fit);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_window, video_player, &CameraVideo::do_zoom_to_window);

    disconnect(MainWindow::GetInstance().get(), &MainWindow::rotate_left, video_player, &CameraVideo::do_rotate_counter);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::rotate_right, video_player, &CameraVideo::do_rotate_clockwise);

    disconnect(MainWindow::GetInstance()->get_export_action(), &QAction::triggered, this, &PlaybackMDI::exportVideo);
    MainWindow::GetInstance()->get_export_action()->setEnabled(false);

    // Get the control panel object.
    control_panel = MainWindow::GetInstance()->get_control_panel();
    if (control_panel == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No control panel?";
        return;
    }

    // Get the display panel object.
    display_panel = qobject_cast< PanelDisplay * >(control_panel->getPanel(CONTROL_PANEL::PANEL_DISPLAY));
    if (display_panel == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No display panel?";
        return;
    }

    // Tell the display panel we're going away.
    display_panel->playback_disconnect();
}

void PlaybackMDI::colorize(bool enable)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "r_gain_old" << r_gain_old << "r_gain" << r_gain;

    // If no change in state of colorization nothing to do here.
    if (do_colorize == enable)
        return;

    // We are changing colorization state.  Mark the new setting.
    do_colorize = enable;

    // If r_gain is marked for customization reset it to last known valid setting.
    if (r_gain == 0.0)
    {
        r_gain = r_gain_old;
    }

    // Clear out the current red LUT no matter what.
    r_gain_lut.clear();

    // If we are disabling colorization and had a custom LUT
    // setting then restore its state prior to building the luts.
    if (!do_colorize)
    {
        // Is there a custom red LUT table out there?
        if (!r_gain_custom_lut.isEmpty())
        {
            // Restore the state of red LUT customization.
            r_gain = 0.0;
            r_gain_lut = r_gain_custom_lut;
        }
    }

    // Update the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief Adjust the red gain value based on the new_value parameter.
 *        This function is a simulation of the Camera Server GIGE_REG_RED
 *        functionality.
 *
 * @param new_value     The new gain, formated as an 8.8 value.
 */
void PlaybackMDI::setRed(uint32_t new_value)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "r_gain_old" << r_gain_old << "r_gain" << r_gain;

    // Save our new gain value.
    r_gain_old = r_gain = (double)new_value / 256.0;

    // Clear the LUT.
    r_gain_lut.clear();
    r_gain_custom_lut.clear();  // A red gain setting removes any custom curve setting.

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "r_gain" << r_gain;

    // Update the LUTs.
    buildLuts(__PRETTY_FUNCTION__);
}

/*!
 * @brief Interface method to be told what our MDI Subwindow is. We need this to be able to hook into
 *        the sub-window's "aboutToActivate" signal.
 *
 * @param win_ptr   Our QMdiSubWindow "parent".
 */
void PlaybackMDI::setWindow(QMdiSubWindow *win_ptr)
{
    // If the win_ptr passed is null there is nothing to disconnect.
    if (win_ptr)
    {
        connect(win_ptr, &QMdiSubWindow::aboutToActivate, this, &PlaybackMDI::windowActivating);
    }
}

/*!
 * @brief Split the filename into its component parts
 *
 * @param[in] video_filename    The filename to split up.
 * @param[out] path_basename    The filename without the extension or frame number but including path.
 * @param[out] frame_num        The frame number or -1 if an error.
 *
 * @return True if the filename was a valid stack filename otherwise false.
 */
bool PlaybackMDI::splitFileName(QString video_filename, QString& path_basename, int& frame_num)
{
    QString     file_path;
    QFileInfo   file_info(video_filename);
    QString basename;
    bool    valid = true;

    // We are working on an image stack, so we need to figure out what image number we are
    // working on. All image stack files we generate end with "nnn"
    file_path = file_info.path();
    basename = file_info.baseName();

    // If the basename is all numeric then it's a stack filename. Otherwise the user must have clicked
    // on an individual file.
    frame_num = (int)basename.toUInt(&valid);
    if (!valid)
    {
        frame_num = 0;
        return false;
    }

    // Are we using subfolders?
    if (using_subfolders)
    {
        // If we are using sub folders, then what we really want for the path is the parent directory rather
        // than the sub folder we are currently in.
        file_path = QDir::fromNativeSeparators(file_path);
        int last_delim = file_path.lastIndexOf(QLatin1Char('/'));
        if (last_delim != -1)
            file_path = file_path.left(last_delim);
    }

    // Since file info stripped off the path, add it back in.
    if (!file_path.endsWith(QLatin1Char('/')))
        file_path.append(QLatin1Char('/'));

    // Return the path.
    path_basename = file_path;
    return true;
}

/*!
 * @brief Build up the event information.
 */
void PlaybackMDI::updateEventInfo()
{
    // Do we already have the metadata?
    if (!pbMetaInfoPtr->perFrameIsAvail())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No PFMD is available!";
        return;
    }

    // Did we turn on the IO graph display?
    if (review_controls->graph_is_enabled())
    {
        // We need to build up the data, 'cuz we just turning on the graph.
        buildData();
    }

    // Only start searching if we're interested in event markers.
    if (use_event_markers)
    {
        search_start_index = 0;
        searchForEvents(swap_needed);
    }
}

/*!
 * @brief Update our progress display.
 */
void PlaybackMDI::UpdateSaveProgress()
{
    uint32_t frameCount = 0;
    uint32_t framesAdded = 0;
    uint32_t framesEncoded = 0;
    QString strMsg;
    bool isStreamVideo = false;
    bool streamIsOk = false;
    bool streamIsComplete = false;

    // Make sure we hae a transfer progress dialog, if not it makes no sense to be here.
    if (!xferDialog)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Polling timer set for progress updates but we have no progress dialog!";
        return;
    }

    // First order of business is to stop a polling event so we don't get overloaded.
    // Make sure that there is a transfer dialog!
    xferDialog->setXferToPollWait();

    // Has the thread stopped?
    if (!saver_thread.isRunning())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Thread isn't running.";

        // The thread is no longer running. Shut down the progress dialog, if necessary.
        UserCancelledSave();
        goto exitUpdateProgress;
    }
    else if (!thread_stop)
    {
        // If the save is complete we are done here and could be in the process
        // of shutting down the stream when a timer kicks off for an update.  This
        // can cause issues if we try to get the video stream status and it's closing
        // up shop.
        if (save_is_complete)
            goto exitUpdateProgress;

        // Do we have a saver object?
        if (saver != nullptr)
        {
            isStreamVideo = saver->getVideoStreamSaveStatus(strMsg, frameCount, framesAdded, framesEncoded, streamIsOk, streamIsComplete);

            if (isStreamVideo)
            {
                // Update the progress information for the save.
                if (xferDialog && xferDialog->isVisible())
                {
                    // This should have the proper information to add to the output stream.
                    xferDialog->setXferSrcValue(QString::number(framesAdded));
                    xferDialog->setBarMinMax(0, (int)frameCount);
                    xferDialog->setBarValue((int)framesEncoded);
                    xferDialog->showStatusValue(true);
                    if (!streamIsOk)
                    {
                        xferDialog->setStatusIcon(QStyle::SP_MessageBoxWarning);
                    }
                    xferDialog->showStatusIcon(!streamIsOk);
                    xferDialog->setStatusValue(strMsg);
                }

                // If this video is complete stop the thread.
                if (framesEncoded >= saver_count)
                {
                    // Only mark this save completed if the stream is completed.
                    if (streamIsComplete)
                    {
                        save_is_complete = true;
                    }
                    if (saver_thread.isSuspended())
                        saver_thread.resume();
                }

                // If the stream isn't ok we need to flag the error condition.
                if (!streamIsOk)
                {
                    save_had_error = true;  // We had an error occur in the stream.
                }
                // The else condition here checks for completion and if found it
                // assumes we have more frames being encoded and go another round.
                // Note:  Make sure this is after check of framesEncoded above.
                else if (!save_is_complete)
                {
                    if (xferDialog)
                        xferDialog->setXferToPollAgain();
                }
            }
            // The else means we are processing stacks.
            else
            {
                // If the data being saved is not in a stream video then the frame number is
                // the actual progress value so update accordingly.
                if (saver_frame != (uint32_t)xferDialog->getBarValue())
                {
                    // Update the frame number processed in the progress dialog.
                    xferDialog->setXferSrcValue(QString::number(saver_frame));
                    xferDialog->setBarValue((int)saver_frame);

                    if (!saver_image_filename.isEmpty())
                        xferDialog->setXferDestValue(saver_image_filename);

                }
                // Set the poller to poll again.
                xferDialog->setXferToPollAgain();
            }
        }
    }

exitUpdateProgress:
    return;
}

/*!
 * @brief Provide a static method with which a caller can validate a video file.
 *        If an error occurs this method displays a message box.
 *
 * @param video_filename    The name of the video file.
 * @return Returns true if the file is valid, or false otherwise.
 */
bool PlaybackMDI::isVideoFileValid(const QString& video_filename)
{
    bool            rc;
    QString         msg_str;
    QString         msg_info_str;
    QString         file_ext;
    QFileInfo       file_info(video_filename);

    std::unique_ptr<FAS_VIDEO::FiVideoStream> AVILocalStreamReader;  //! Our AVI stream reader.

    // Make sure it exists.
    if (!file_info.exists())
    {
        // Notify the user.
        msg_str = tr("Cannot find file '%1'.").arg(video_filename);
        QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
        fiMsgBox::critical(MainWindow::GetInstance().get(), msg_str);
        return false;
    }

    // Make sure this video is complete.
    QString video_file_incomplete = video_filename + (FAS_VIDEO::FiVideo::GetIncompleteVideoFileString()).c_str();
    QFileInfo file_incomplete(video_file_incomplete);
    // Make sure the incomplete indicator file isn't there.
    if (file_incomplete.exists())
    {
        // Notify the user.
        msg_str = tr("The video '%1' is incomplete and cannot be opened.").arg(video_filename);
        QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
        fiMsgBox::critical(MainWindow::GetInstance().get(), msg_str);
        return false;
    }

    // Get the extension.
    file_ext = file_info.suffix();

    // What are the valid file extensions that we support?
    QStringList valid_extensions(QStringList()
                                 << FiSaver::IMAGE_EXT_AVI
                                 << FiSaver::IMAGE_EXT_BMP
                                 << FiSaver::IMAGE_EXT_CAPTURE
                                 << FiSaver::IMAGE_EXT_DNG
                                 << FiSaver::IMAGE_EXT_JPG
                                 << FiSaver::IMAGE_EXT_TIF);

    // Is this one of the supported extensions?
    if (!valid_extensions.contains(file_ext, Qt::CaseInsensitive))
    {
        // Notify the user.
        msg_str = tr("File '%1' is not supported.").arg(video_filename);
        QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
        fiMsgBox::critical(MainWindow::GetInstance().get(), msg_str);
        return false;
    }

    // Assume life is good.
    rc = true;

    // Determine if we are playing an AVI or a stack of images. This will affect everything we do.
    if (file_ext.compare(FiSaver::IMAGE_EXT_AVI, Qt::CaseInsensitive) == 0)
    {
        // Create the video stream reader.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Creating local AVI Stream Reader:  " << video_filename;
        AVILocalStreamReader = FAS_VIDEO::FiVideo::CreateAVIVideoStreamReader(video_filename.toStdString());

        // If unable to create the video stream writer log an error.
        if (AVILocalStreamReader == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create local AVI Stream Reader for: " << video_filename;

            // Notify the user.
            msg_str = tr("Unable to validate the video file %1").arg(video_filename);
            QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;

            // Show the message as a modeless dialog.
            fiMsgBox::critical(MainWindow::GetInstance().get(), msg_str, msg_info_str);

            // No good.
            rc = false;
        }
    }

#if 0
    else
    {
        // An individual video frame. Make sure it can be loaded into a pixelmap.
        fiPixmap    my_pixelmap;
        if (!my_pixelmap.load(video_filename))
        {
            // Notify the user.
            msg_str = tr("Unable to load image file %1").arg(video_filename);
            fiMsgBox::critical(MainWindow::GetInstance().get(), msg_str);
            rc = false;
        }
    }
#endif

    // Return our status.
    return rc;
}


//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Respond to the window Close event.
 *
 * @param event Pointer to the event structure.
 */
void PlaybackMDI::closeEvent(QCloseEvent *event)
{
    // We're closing, so go inactive.
    setInactive();

    // Let folks know we're closing.
    emit closing();

    // Accept the event.
    event->accept();
}

/*!
 * @brief The user has right mouse clicked in our window. Put up a
 *        context specific menu.
 *
 * @param event_ptr     Pointer to the menu event.
 */
void PlaybackMDI::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QMenu   context_menu(this);

    // Make sure the review controls are enabled.
    MainWindow::GetInstance()->get_view_review_controls()->setEnabled(true);

    if (pbMetaInfoPtr != nullptr)
    {
        // Enable the PFMD view if it's supported.
        MainWindow::GetInstance()->get_view_pfmd()->setEnabled(pbMetaInfoPtr->perFrameIsAvail());

        // If this is a file - is there a partition metadata txt file?
        if (is_video_file)
        {
            MainWindow::GetInstance()->get_view_summary()->setEnabled(has_metadata_txt_file);
        }
        else
        {
            MainWindow::GetInstance()->get_view_summary()->setEnabled(true);
        }

        // Enable the local time view if it's supported.
        MainWindow::GetInstance()->get_view_local_frametime()->setEnabled(pbMetaInfoPtr->perFrameIsAvail());

        // Enable the event marker menu, if we have raw metadata.
        MainWindow::GetInstance()->get_event_marker_action()->setEnabled(pbMetaInfoPtr->perFrameIsAvail());

        // Don't allow saves if there is an "export" section in the metadata.
        MainWindow::GetInstance()->get_save()->setEnabled(!pbMetaInfoPtr->isExported());
        MainWindow::GetInstance()->get_export_action()->setEnabled(!pbMetaInfoPtr->isExported());
    }
    else
    {
        MainWindow::GetInstance()->get_view_summary()->setEnabled(false);
    }

    context_menu.addAction(MainWindow::GetInstance()->get_view_summary());
    context_menu.addAction(MainWindow::GetInstance()->get_view_review_controls());
    context_menu.addAction(MainWindow::GetInstance()->get_view_pfmd());
    context_menu.addAction(MainWindow::GetInstance()->get_view_local_frametime());
    context_menu.addAction(MainWindow::GetInstance()->get_event_marker_action());
    context_menu.addAction(MainWindow::GetInstance()->get_view_aiming_cursor());
    context_menu.addAction(MainWindow::GetInstance()->get_view_info_cursor());
    context_menu.addAction(video_player->get_action_overlay());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_metadata_dock()->toggleViewAction());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_save());
    context_menu.addAction(MainWindow::GetInstance()->get_export_action());


    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

/*!
 * @brief Respond to getting focus.
 * @param event     Pointer to the event structure.
 */
void PlaybackMDI::focusInEvent(QFocusEvent *event)
{
    ControlPanel    *control_panel;
    PanelDisplay    *display_panel;
    bool            allow_export = false;

    // Pass it on up.
    QWidget::focusInEvent(event);

    // Tell the main window about getting the focus.
    MainWindow::GetInstance()->change_mdi_child(this);

    // Reflect our current state.
    MainWindow::GetInstance()->reflect_camera_state();

    // Connect our signals/slots.
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_save, this, &PlaybackMDI::doImageSave, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_review_controls, this, &PlaybackMDI::showControls, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_summary, this, &PlaybackMDI::showSummary, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_aiming_cursor, this, &PlaybackMDI::doViewAimingCursor, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_info_cursor, this, &PlaybackMDI::doViewInfoCursor, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_pfmd, this, &PlaybackMDI::showPfmd, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_local_frametime, this, &PlaybackMDI::showLocalFrametime, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::setup_event_markers, this, &PlaybackMDI::doPlaybackEvents, Qt::UniqueConnection);

    connect(video_player->get_action_overlay(), &QAction::triggered, video_player, &CameraVideo::do_add_overlay, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_1_to_1, video_player, &CameraVideo::do_zoom_1_to_1, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_in, video_player, &CameraVideo::do_zoom_in, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_out, video_player, &CameraVideo::do_zoom_out, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_fit, video_player, &CameraVideo::do_zoom_to_fit, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_window, video_player, &CameraVideo::do_zoom_to_window, Qt::UniqueConnection);

    connect(MainWindow::GetInstance().get(), &MainWindow::rotate_left, video_player, &CameraVideo::do_rotate_counter, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::rotate_right, video_player, &CameraVideo::do_rotate_clockwise, Qt::UniqueConnection);

    // Make sure the zoom actions are enabled.
    MainWindow::GetInstance()->enable_zoom_actions(true);

    // Make sure our view controls are active and reflect our current state.
    MainWindow::GetInstance()->get_view_review_controls()->setEnabled(true);
    MainWindow::GetInstance()->get_view_review_controls()->setChecked(review_controls->isVisible());

    // Only enable the save if there's metadata available.
    if (pbMetaInfoPtr && pbMetaInfoPtr->captureMetadataIsAvail())
    {
        // Don't allow saves or exports if there is an "export" section in the metadata.
        MainWindow::GetInstance()->get_save()->setEnabled(!pbMetaInfoPtr->isExported());
        allow_export = !pbMetaInfoPtr->isExported();

        MainWindow::GetInstance()->get_view_summary()->setEnabled(true);
        MainWindow::GetInstance()->get_view_summary()->setChecked(camera_info->summary_is_enabled());

        // Enable the PFMD view if it's supported.
        //QLOG_INFO() << __PRETTY_FUNCTION__ << "[FOCUS-IN NEW SET enable_pfd][" << pbMetaInfoPtr->per_frame_is_avail() << "]";

        // What is the current setting of the menu's per-frame metadata checkbox?
        bool pfmd_checked = MainWindow::GetInstance()->get_view_pfmd()->isChecked();

        if (pbMetaInfoPtr->perFrameIsAvail())
        {
            // In this situation we have pfmd available but was menu checked?
            camera_info->enable_pfmd(pfmd_checked);
        }
        else
        {
            // In this situation there is no pfmd available so force false.
            camera_info->enable_pfmd(false);
        }
        MainWindow::GetInstance()->get_view_pfmd()->setEnabled(pbMetaInfoPtr->perFrameIsAvail());
        MainWindow::GetInstance()->get_view_pfmd()->setChecked(camera_info->get_pfmd_setting());
        MainWindow::GetInstance()->get_view_local_frametime()->setEnabled(pbMetaInfoPtr->perFrameIsAvail());

        // Enable the event marker menu, if we have raw metadata.
        MainWindow::GetInstance()->get_event_marker_action()->setEnabled(!pbMetaInfoPtr->getRawfile().isEmpty());
    }
    else
    {
        MainWindow::GetInstance()->get_save()->setEnabled(false);
        allow_export = false;
        camera_info->enable_summary(false);
        MainWindow::GetInstance()->get_view_summary()->setEnabled(false);
        MainWindow::GetInstance()->get_view_summary()->setChecked(false);
        camera_info->enable_pfmd(false);
        MainWindow::GetInstance()->get_view_pfmd()->setEnabled(false);
        MainWindow::GetInstance()->get_view_pfmd()->setChecked(false);
        MainWindow::GetInstance()->get_view_local_frametime()->setEnabled(false);
        MainWindow::GetInstance()->get_event_marker_action()->setEnabled(false);
    }

    // See if we want to allow for exports. Only allow for certain types.
    if ((file_ext.compare(FiSaver::IMAGE_EXT_AVI, Qt::CaseInsensitive) == 0) ||
        (file_ext.compare(FiSaver::IMAGE_EXT_JPG, Qt::CaseInsensitive) == 0) ||
        (file_ext.compare(FiSaver::IMAGE_EXT_TIF, Qt::CaseInsensitive) == 0) ||
        (file_ext.compare(FiSaver::IMAGE_EXT_BMP, Qt::CaseInsensitive) == 0))
    {
        // Make sure we have the information necessary to export.
        if (is_video_file)
        {
            allow_export = has_metadata_txt_file;
        }
        else
        {
            allow_export = true;
        }
    }

    // If it is ok to export a video file with the loaded video frames set it up to do so.
    if (allow_export)
    {
        // Enable and connect the export action.
        MainWindow::GetInstance()->get_export_action()->setEnabled(true);
        connect(MainWindow::GetInstance()->get_export_action(), &QAction::triggered, this, &PlaybackMDI::exportVideo, Qt::UniqueConnection);
    }

    // Update the selected zoom.
    MainWindow::GetInstance()->check_zoom_actions(video_player->get_zoom_setting());

    // Make sure the metadata dock is showing OUR information.
    MainWindow::GetInstance()->set_metadata_info(pbMetaInfoPtr);

    // Get the control panel object.
    control_panel = MainWindow::GetInstance()->get_control_panel();
    if (control_panel == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No control panel?";
        return;
    }

    // Get the display panel object.
    display_panel = qobject_cast<PanelDisplay *>(control_panel->getPanel(CONTROL_PANEL::PANEL_DISPLAY));
    if (display_panel == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No display panel?";
        return;
    }

    // Tell the panel to connect to us.
    display_panel->playback_connect(this);

    // Tell the panel our sensor type.
    display_panel->update_sensor_type(pbMetaInfoPtr->getSensorType());

    // Make sure we open the panel if our flag indicates to do so.
    if (open_display_panel_on_focus)
    {
        control_panel->openPanel(CONTROL_PANEL::PANEL_DISPLAY);
        open_display_panel_on_focus = false;    // Only force the open one time during the life of this playback window.
    }

    // Update shared view menu settings.
    camera_info->show_info();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user would like to setup the event marker display.
 *
 * @sa MainWindow::setup_event_markers()
 */
void PlaybackMDI::doPlaybackEvents()
{
    bool            io_graph_was_displayed;
    eventFilterDlg  dlg(this);

    // Remember if the graph is being displayed or not.
    io_graph_was_displayed = review_controls->graph_is_enabled();

    // Initialize the dialog with our current settings.
    dlg.set_event_markers_enable(use_event_markers);
    dlg.set_display_graph_enable(io_graph_was_displayed);
    dlg.set_logic_to_and(event_logic_is_and);
    dlg.set_selects(event_selections);

    // Pop up the dialog.
    if (dlg.exec() == QDialog::Accepted)
    {
        bool    criteria_changed(false);
        int     i;

        // Read the user settings.
        use_event_markers = dlg.event_markers_are_enabled();
        event_logic_is_and = dlg.use_logic_and();

        // First, see if our criteria changed.
        for (i = 0; i < FM_EVENT_FILTERS::INDEX_COUNT; ++i)
        {
            if (event_selections.at(i) != dlg.event_selects().at(i))
            {
                criteria_changed = true;
                break;
            }
        }

        // Did the criteria change?
        if (criteria_changed)
        {
            // Clear out the old selections and copy in the new.
            event_selections.clear();
            for (i = 0; i < FM_EVENT_FILTERS::INDEX_COUNT; ++i)
                event_selections.append(dlg.event_selects().at(i));
        }

        // Do we want to reset the marker list?
        if (criteria_changed || !use_event_markers)
        {
            // Reset our maker list.
            marker_list.clear();

            // Are we interested in event markers?
            if (!use_event_markers)
                review_controls->set_markers(marker_list);
        }

        // Only need to download the raw per-frame metadata if we care about it.
        if (use_event_markers || dlg.display_graph_enabled())
        {
            // Did we turn on the IO graph display?
            if (!io_graph_was_displayed && dlg.display_graph_enabled())
            {
                // Enable the graph.
                review_controls->enable_graph(true);
            }

            // Update our event information, downloading the per-frame metadata if necessary.
            updateEventInfo();
        }

        // Are we supposed to turn it off?
        if (io_graph_was_displayed && !dlg.display_graph_enabled())
            review_controls->enable_graph(false);

        // Save all the settings so we can come up next time with the same values.
        saveSettings();
    }
}

/*!
 * @brief The user has initiated a save operation.
 *
 * @sa MainWindow::camera_save()
 */
void PlaybackMDI::doImageSave()
{
    uint32_t    num_frames;

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Sanity checks.
    if (pbMetaInfoPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No metadata pointer!";
        return;
    }
    if (review_controls == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No review_controls pointer!";
        return;
    }

    // If you get here and there is no metadata - you cannot continue.
    // This will hold us over until I can get this fixed.
    if (is_video_file)
    {
        if (!has_metadata_txt_file)
        {
            QMessageBox::warning(this, "Error Saving File", tr("Unable to save without partition metadata."));
            return;
        }
    }

    // Not allowed to save if playback is active.
    if (playback_timer && playback_timer->isActive())
    {
        QMessageBox::warning(this, "Error Saving File", tr("Unable to save during playback"));
        return;
    }

    // Get our dialog allocated.
    save_dlg = std::shared_ptr<SaveDialog>(new SaveDialog(0, this));
    if (!save_dlg)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_SAVE_DIALOG;
        return;
    }

    // Initialize our dialog. The ordering of the following lines of code is very
    // important. We need to setup the trigger and frame count BEFORE calling the
    // set_direct_transfer() method, so that the limits will be initialized and allow
    // the ranges to be setup for the set_clip_range() call.
    save_dlg->set_sensor_type(pbMetaInfoPtr->getSensorType());
    save_dlg->set_num_frames(pbMetaInfoPtr->getFrameCount());
    save_dlg->set_trig_frame(pbMetaInfoPtr->getTriggerFrame());
    save_dlg->set_direct_transfer(true);
    save_dlg->disallow_raw();
    save_dlg->disallow_xml();
    save_dlg->set_clip_range(review_controls->get_clip_start(), review_controls->get_clip_end());

    // Finally we can have the dialog initialize it's controls.
    save_dlg->init_controls();

    // Are there overlays?
    if (video_player)
        save_dlg->enable_overlay_merge(video_player->has_overlay());

    // Pop up the dialog.
    if (save_dlg->exec() == QDialog::Rejected)
    {
        save_dlg.reset();
        return;
    }

    // Make sure there's enough space.
    num_frames = (save_dlg->get_clip_end() - save_dlg->get_clip_start()) + 1;
    if (!MainWindow::GetInstance()->check_avail_space(save_dlg->get_dest_path(), num_frames, save_dlg->GetStreamType(),
                                                      pbMetaInfoPtr->getWidth(), pbMetaInfoPtr->getHeight(),
                                                      pbMetaInfoPtr->getConvertedBitMode(), pbMetaInfoPtr->getSensorType()))
    {
        return;
    }

    // Update the metadata with the export values.
    pbMetaInfoPtr->setStartFrame(save_dlg->get_clip_start());
    pbMetaInfoPtr->setEndFrame(save_dlg->get_clip_end());
    pbMetaInfoPtr->setFrameCount(num_frames);
    pbMetaInfoPtr->setCodec(save_dlg->GetStreamType());
    pbMetaInfoPtr->setComment(save_dlg->get_comment());

    // Create and deploy our transfer progress dialog.
    if (CreateXferDialog())
    {
        xferDialog->showDialogWithPolling(SAVER_POLL_RATE);
    }

    // Start our polling timer.
    saver_image_filename.clear();

    // We want to hook up a signal catch for when SaveImages finishes.
    connect(this, &PlaybackMDI::saveImagesComplete, this, &PlaybackMDI::saveImagesCompleted, Qt::UniqueConnection);

    // Save the data.
    thread_stop = false;
    saver_thread = QtConcurrent::run(SaveImages, this);
}

/*!
 * @brief The user is toggling the aiming cursor on or off.
 *
 * @sa MainWindow::view_aiming_cursor()
 */
void PlaybackMDI::doViewAimingCursor()
{
    if (video_player != nullptr)
    {
        video_player->enable_aiming_cursor(MainWindow::GetInstance()->get_view_aiming_cursor()->isChecked());
        video_player->rezoom();
    }
}

/*!
 * @brief The user is toggling the info cursor on or off.
 *
 * @sa MainWindow::view_info_cursor()
 */
void PlaybackMDI::doViewInfoCursor()
{
    if (video_player != nullptr)
        video_player->enable_info_cursor(MainWindow::GetInstance()->get_view_info_cursor()->isChecked());
}

/*!
 * @brief The user wants to cancel the search for events.
 *
 * @sa PlaybackMDI::search_for_events()
 */
void PlaybackMDI::eventSearchCanceled()
{
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }
}

void PlaybackMDI::exportVideo()
{

    QMessageBox::information(
        this,
        tr("Export Video Dialog"),
        tr("The Export Video Dialog will go here when implemented."));

#if 0
    int         rel_end;
    int         rel_start;
    ExportDlg   dlg(this);

    // Initialize our dialog.
    dlg.set_source_path(root_file);
    dlg.set_num_frames(get_metadata()->get_frame_count());
    dlg.set_trig_frame((uint32_t)trigger_frame);
    dlg.set_clip_range(get_metadata()->get_start_frame(), get_metadata()->get_end_frame());

    rel_start = (int)review_controls->get_clip_start() - trigger_frame;
    rel_end = (int)review_controls->get_clip_end() - trigger_frame;
    dlg.set_rel_range(rel_start, rel_end);

    if (!AVIStreamReader)
        dlg.set_fps(get_metadata()->get_framerate());
    else
        dlg.set_fps(AVIStreamReader->GetFrameRate());

    // Go.
    dlg.exec();
#endif
}

/*!
 * @brief Search through the [raw metadata buffer](@ref PlaybackMetaInfo::raw_metadata) looking
 *        for events which match the [user specified criteria](@ref eventFilterDlg).
 *
 * The results are stored in a [list](@ref PlaybackMDI::marker_list) which is passed to
 * the [video review object(@ref CameraReview) for display.
 *
 * Since this processing can take quite a long time (there can be a large number of frames),
 * this is a "distributed" function. Once we've processed a [certain number of frames](@ref PLAYBACK_MDI_SETTINGS::MAX_SEARCH_COUNT),
 * a [timer](@ref PLAYBACK_MDI_SETTINGS::SEARCH_DELAY) is set and the processing loop is exited.
 *
 * After the timer expires, processing is resumed.
 */
void PlaybackMDI::searchForEvents(bool do_swap)
{
    bool        search_done;
    uint32_t    search_count;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "search_start_index:" << search_start_index << "do_swap:" << do_swap;

    // Seek to the starting position.
    if (search_start_index == 0)
    {
        // Reset our marker list.
        marker_list.clear();

        // Do we need to give the user progress on the event search through the metadata?
        if (pbMetaInfoPtr->getFrameCount() > (2 * MAX_SEARCH_COUNT))
        {
            // Create the progress dialog if we don't have one.
            if (!progressDialog)
            {
                progressDialog = fiProgressDialog::CreateSimpleProgressDialog(camera_info->get_camera_name(), tr("Event Marker Processing"), QMessageBox::Cancel, __PRETTY_FUNCTION__, this);
            }

            // Display the progress dialog.
            if (progressDialog)
            {
                connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &PlaybackMDI::eventSearchCanceled, Qt::UniqueConnection);
                progressDialog->setBarMinMax(0, (int)pbMetaInfoPtr->getFrameCount());
                progressDialog->showBar(true);
                progressDialog->setCancelText(tr("Cancel Search"));
                progressDialog->showDialogWithoutPolling();
            }
        }
    }

    // How many frames can we search before we need to give the UI a break.
    search_count = pbMetaInfoPtr->getFrameCount() - (uint32_t)search_start_index;
    if (search_count > MAX_SEARCH_COUNT)
        search_count = MAX_SEARCH_COUNT;

    // Add to our marker list.
    search_done = getRawMarkers((uint32_t)search_start_index, search_count, do_swap);

    // Have we exceeded our maximum number of found markers?
    if (marker_list.size() >= FM_EVENT_FILTERS::MAX_MARKER_LIST_SIZE)
    {
        // Clean up any displayed or allocated progress dialog.
        if (progressDialog)
        {
            // Recycle any resources and release the progress dialog.
            fiProgressDialog::RecycleProgressDialog(progressDialog);
        }

        // We've reached the limit.
        QMessageBox::information(this, tr("Event Marker Processing"),
                                 tr("The maximum number of event markers has been exceeded.\n"
                                    "Please refine your criteria, or review your hardware setup."));

        // Go with what we now have.
        review_controls->set_markers(marker_list);
    }

    // Are we done?
    else if (search_done)
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "search_done and completed. search_start_index:" << search_start_index << "search_count:" << search_count << "do_swap:" << do_swap;

        // Clean up any progress dialog that was allocated or displayed.
        if (progressDialog)
        {
            // Recycle any resources and release the progress dialog.
            fiProgressDialog::RecycleProgressDialog(progressDialog);
        }

        // Set the new event markers.
        review_controls->set_markers(marker_list);
        review_controls->enable_panel(true);
    }
    else
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "SEARCH_DELAY TIME search_start_index:" << search_start_index << "search_count:" << search_count << "do_swap:" << do_swap;

        // Come back after a short timeout.
        QTimer::singleShot(SEARCH_DELAY, this, [ = ]() { searchForEvents(do_swap);});

        // Move up our search start index.
        search_start_index += search_count;

        // Update our progress bar, if we have one.
        if (progressDialog)
        {
            // Update the progress dialog.
            progressDialog->setBarValue(search_start_index);
        }
    }
}

/*!
 * @brief The user wants to show or hide the review control panel.
 */
void PlaybackMDI::showControls()
{
    // Show/hide the review controls panel.
    if (review_controls != nullptr)
    {
        review_controls->setVisible(!review_controls->isVisible());
    }
}

/*!
 * @brief The user would like to view time using a local format.
 *
 * @sa MainWindow::view_local_frametime()
 */
void PlaybackMDI::showLocalFrametime()
{
    // Toggle the enable of the local frametime display.
    bool    enable_local = !camera_info->localtime_is_enabled();

    // Enable/disable the local frametime display.
    camera_info->enable_local_frametime(enable_local);
    MainWindow::GetInstance()->get_view_local_frametime()->setChecked(enable_local);

    // Show/hide the display info.
    camera_info->show_info();

    // Update the per-frame metadata display.
    updatePfmd();
}

/*!
 * @brief The user would like to view the per-frame metadata.
 *
 * @sa MainWindow::view_pfmd()
 */
void PlaybackMDI::showPfmd()
{
    // Toggle the enable of the per-frame metadata display.
    bool    enable_pfmd = !camera_info->pfmd_is_enabled();

    //JLCTEMP
    QLOG_INFO() << __PRETTY_FUNCTION__ << "[SHOW_PFMD NEW SETTING MARKING enable_pfmd][" << enable_pfmd << "]";

    // Enable/disable the per-frame metadata display.
    camera_info->enable_pfmd(enable_pfmd);
    MainWindow::GetInstance()->get_view_pfmd()->setChecked(enable_pfmd);

    // Show/hide the display info.
    camera_info->show_info();

    // If we're enabling the display, make sure we grab the metadata for the current frame.
    if (enable_pfmd)
    {
        // Update the per-frame metadata display.
        updatePfmd();
    }
}

/*!
 * @brief The user wants to show or hide the summary information panel.
 */
void PlaybackMDI::showSummary()
{
    bool    enable = !camera_info->summary_is_enabled();

    // Enable/disable the per-frame metadata display.
    camera_info->enable_summary(enable);
    MainWindow::GetInstance()->get_view_summary()->setChecked(enable);

    // Show/hide the display info.
    camera_info->show_info();
}

/*!
 * @brief Our [play back timer](@ref PlaybackMDI::playback_timer) has expired.
 */
void PlaybackMDI::timeToPlay()
{
    uint32_t    current_frame;

    // Get the current frame.
    current_frame = review_controls->get_current_frame_num();
    if (playback_forward)
    {
        // going forward, if reach the end then wrap
        current_frame += (uint32_t)play_frames_inc;
        if (current_frame > review_controls->get_clip_end())
            current_frame = review_controls->get_clip_start();
    }
    else
    {
        // going backwards, if reach the begining then wrap
        current_frame -= (uint32_t)play_frames_inc;
        if ((current_frame < review_controls->get_clip_start()) || (current_frame > review_controls->get_clip_end()))
            current_frame = review_controls->get_clip_end();
    }

    // Go to the new frame.
    review_controls->set_current_frame_num(current_frame);
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that it needs
 *        the per-frame metadata to be updated.
 *
 * @sa CameraReview::play_update()
 */
void PlaybackMDI::updatePfmd()
{
    // Show whatever information should be visible.
    camera_info->show_info();

    // Get the metadata for the current frame.
    loadPfmd(review_controls->get_current_frame_num() - pbMetaInfoPtr->getStartFrame());
}

void PlaybackMDI::UserCancelledSave()
{
    // Tell the thread to stop.
    thread_stop = true;
    if (saver_thread.isRunning())
    {
        // Wait for the saving thread to stop.
        saver_thread.waitForFinished();
    }

    // Close down any progress dialogs that were up.
    DoneWithProgressDialogs();

    // Clear out any "export" metadata we may have created.
    if (pbMetaInfoPtr != nullptr)
        pbMetaInfoPtr->resetExports();
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the current
 *        frame has changed.
 *
 * @sa CameraReview::current_frame_changed()
 */
void PlaybackMDI::videoCurrentFrameChanged()
{
    uint32_t    cur_frame;
    fiPixmap    my_pixmap;

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    QElapsedTimer   elapsed_timer;
    elapsed_timer.start();
    // Statistics
    //-------------------------------------------------------------------
#endif

    // Don't single step if we don't want streaming.
    // Get the "new" current frame.
    cur_frame = review_controls->get_current_frame_num();

    // Don't need to get the frame if it's the same as the last one.
    if (cur_frame == last_frame_loaded)
        return;

    // Load up the requested frame.
    if (!loadFrame(cur_frame, &my_pixmap))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the frame:" << cur_frame;
        return;
    }

    // Make sure we got something.
    if (my_pixmap.isNull())
        return;

    // Are we supposed to white balance?
    if (do_whitebal)
    {
        // Auto white white balance.
        my_pixmap.auto_wb(&r_wb_gain, &g_wb_gain, &b_wb_gain);

        QLOG_TRACE() << __PRETTY_FUNCTION__ << "r_wb_gain:" << r_wb_gain << "g_wb_gain:" << g_wb_gain << "b_wb_gain:" << b_wb_gain;

        // Update our LUTs.
        buildLuts(__PRETTY_FUNCTION__);

        // Reset our flag.
        do_whitebal = false;
    }

    // Pass the image through the lookup table.
    my_pixmap.apply_lookup(r_lut, g_lut, b_lut);

    // Pass it on to the player.
    video_player->display_image(my_pixmap);

    // Are we showing per-frame metadata?
    if (pbMetaInfoPtr->perFrameIsAvail())
    {
        // Request the metadata.
        loadPfmd(cur_frame - pbMetaInfoPtr->getStartFrame());
    }

#ifdef DO_STATS
    //-------------------------------------------------------------------
    // Statistics
    // Get the elapsed time and generate statistics.
    elapsed = elapsed_timer.elapsed();
    elapsed_time = QTime(0, 0, 0, 0).addMSecs((int)elapsed);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << elapsed_time.toString(QString("hh:mm:ss.zzz"));
    // Statistics
    //-------------------------------------------------------------------
#endif
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to jump
 *        to a particular time/frame in the recorded video.
 *
 * @sa CameraReview::jump_time()
 */
void PlaybackMDI::videoJumpTime()
{
    JumpDialog  jump_dlg(this);

    // Do we have any raw metadata?
    if (pbMetaInfoPtr->getRawfile().isEmpty())
        return;

    // Get the current frame, trigger frame and frame rate so we can
    // calculate the relative time of the current frame.
    jump_dlg.set_pfmeta(pbMetaInfoPtr);
    jump_dlg.set_rel_time((int)review_controls->get_clip_start(), (int)review_controls->get_clip_end(), (int)review_controls->get_current_frame_num(),
                          (int)pbMetaInfoPtr->getTriggerFrame(), (int)pbMetaInfoPtr->getFramerate());
    jump_dlg.use_rel_time();

    // Pop up the dialog.
    if (jump_dlg.exec() == QDialog::Accepted)
    {
        // Go to the selected frame.
        review_controls->set_current_frame_num(jump_dlg.get_frame());
    }
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to playback
 *        the recorded video in the normal (i.e. "forward") direction.
 *
 * @sa CameraReview::play_forward()
 */
void PlaybackMDI::videoPlayForward()
{
    // Play the video going forward.
    playback_forward = true;
    playVideo();
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to playback
 *        the recorded video in the backward (i.e. "reverse") direction.
 *
 * @sa CameraReview::play_reverse()
 */
void PlaybackMDI::videoPlayReverse()
{
    // Play the video going reverse.
    playback_forward = false;
    playVideo();
}

/*!
 * @brief The user has changed the playback rate.
 * @param new_rate  [UNUSED] The new rate the user wants to use for playback.
 */
void PlaybackMDI::videoRateChanged(int new_rate)
{
    Q_UNUSED(new_rate)

    // Is playback currently active?
    if (playback_timer && playback_timer->isActive())
    {
        // Update the playback operation.
        playVideo();
    }
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to stop playback.
 *
 * @sa CameraReview::playback_stop()
 */
void PlaybackMDI::videoReviewStop()
{
    if (playback_timer != nullptr)
    {
        playback_timer->stop();
    }

    // Enable the playback buttons.
    review_controls->update_buttons(true);

    // Allow saves again.
    if (is_video_file)
    {
        MainWindow::GetInstance()->get_save()->setEnabled(has_metadata_txt_file);
        MainWindow::GetInstance()->get_export_action()->setEnabled(has_metadata_txt_file);
    }
    else
    {
        MainWindow::GetInstance()->get_save()->setEnabled(true);
        MainWindow::GetInstance()->get_export_action()->setEnabled(true);
    }
}

/*!
 * @brief Our window is activating. Make sure we get focus.
 */
void PlaybackMDI::windowActivating()
{
    // Check to see if we have focus.
    if (!hasFocus())
    {
        // We don't have focus, so grab focus.
        setFocus();
    }
}

