/*!
 * @file playback_mdi.h
 * @brief This is the header file for the PlaybackMDI class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PLAYBACK_MDI_H
#define PLAYBACK_MDI_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <stdint.h>
#include <QFuture>
#include <QWidget>

#include <fas_video.h>
#include <fi_image.h>
#include <fi_rawdrive.h>
#include <fi_rawfile.h>
#include <fi_saver.h>
#include <fi_stream.h>
#include <fi_display_lut.h>

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class QFile;
class QMdiSubWindow;
class QTimer;
class QTimerEvent;
class QVBoxLayout;

class CameraInfo;
class CameraReview;
class CameraStream;
class CameraVideo;
class FiDisplayLut;
class fiPixmap;
class fiProgressDialog;
class FiSaver;
class PlaybackMetaInfo;
class SaveDialog;


//--------------------------------------------------------------------
// Class Definition wrapper for struct pm_metadata support.
//--------------------------------------------------------------------

//! Management for the structure pm_metadata_type which represents the
//! header record of a Fastec Drive or a Fastec Raw CAP file.
//!
//! The pm_metadata struct is defined and maintained in gigemsg.h
class CaptureMetaInfo
{
private:
    CaptureMetaInfo();

    // This object is non-copyable and non-copy assignable.
    CaptureMetaInfo(const CaptureMetaInfo& i) = delete;
    CaptureMetaInfo& operator=(const CaptureMetaInfo& i) = delete;

public:
    virtual ~CaptureMetaInfo();

    //! Factory creation methods.
    static QSharedPointer<CaptureMetaInfo> CreateCaptureMetaInfo(QSharedPointer<fiRawFile> capFile);
    static QSharedPointer<CaptureMetaInfo> CreateCaptureMetaInfo(QSharedPointer<fiRawDrive> rawDrive);

    bool containsMyData(QSharedPointer<fiRawFile> capFile);

    const struct pm_metadata_type *GetMetaInfo() { return pm_metadata_ptr; }

private:
    void cleanupMetadataVersion();

    QString rawFileName;            //!< Name of the fiRawFile or fiRawDrive owning this data.

    pm_metadata *pm_metadata_ptr;   //!< Our internal pm_metadata being managed.
};

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class PlaybackMDI : public QWidget
{
    Q_OBJECT

public:
    explicit PlaybackMDI(QWidget *parent = nullptr);
    virtual ~PlaybackMDI();

    static const QString MISSING_SAVE_DIALOG;
    static const QString MISSING_PM_METADATA;

    virtual bool CreateXferDialog(QWidget *parent = nullptr);
    virtual void DoneWithProgressDialogs();

    virtual QString getFile()                   { return root_file; }
    virtual const QByteArray& getRawImage() const;
    virtual bool loadFrame(uint32_t frame_num, fiPixmap *pixmap_ptr, bool force_load = false);
    virtual bool loadPackedFrame(uint32_t frame_num, QImage& destImage);
    virtual void refreshFrame();
    virtual bool setFile(const QString& filename);
    virtual void setInactive();

    static bool isVideoFileValid(const QString& video_filename);

    void colorize(bool enable);
    char getBitSel();
    CameraInfo *getCameraInfo()                 { return camera_info; }
    bool getApplyingCapLuts()                   { return applyingCapLuts; }
    uint32_t getDisplayBrightness();
    uint32_t getDisplayContrast();
    uint32_t getDisplayGamma();
    uint32_t getDisplayRed();
    uint32_t getDisplayGreen();
    uint32_t getDisplayBlue();
    double getGamma()                           { return gamma; }
    QVector<double>& getGammaLut()              { return gamma_lut; }
    QVector<double>& getRedLut()                { return r_gain_custom_lut; }
    QVector<double>& getGreenLut()              { return g_gain_lut; }
    QVector<double>& getBlueLut()               { return b_gain_lut; }
    uint32_t getLastFrame()                     { return last_frame_loaded; }

    PlaybackMetaInfo *getMetadata()             { return pbMetaInfoPtr; }
    std::shared_ptr<FiDisplayLut> GetMetadataLut();

    int getPfmdLen();
    CameraVideo *getPlayer()                    { return video_player; }
    CameraReview *getReviewControls()           { return review_controls; }
    int getWhitebal()                           { return whitebal; }

    bool isPlaybackLutInUse();
    bool isMono();
    bool isSeries3();

    void setWindow(QMdiSubWindow *win_ptr);
    QAction *windowMenuAction() const           { return playback_action; }

    void setApplyingCapLuts(bool applyingCapLut);
    void setBrightness(uint32_t new_value);
    void setContrast(uint32_t new_value);
    void setGamma(uint32_t new_value);
    void setRed(uint32_t new_value);
    void setGreen(uint32_t new_value);
    void setBlue(uint32_t new_value);
    void setWhitebal(int new_value)             { whitebal = new_value; }

    bool set8bitGammaLut(const QVector<double>& lut8BitGamma);
    bool set10BitGammaLut(const QVector<double>& lut10BitGamma);
    bool set8bitRgbLuts(const QVector<double>& lut8BitRed, const QVector<double>& lut8BitGreen, const QVector<double>& lut8BitBlue);
    bool set10BitRgbLuts(const QVector<double>& lut10BitRed, const QVector<double>& lut10BitGreen, const QVector<double>& lut10BitBlue);

    void performAutoWb()                        { do_whitebal = true; }
    void resetAutoWbGains();

signals:
    void closing();
    void saveImagesComplete(bool savedOK);

protected:
    void buildLuts(const QString whoCalledMe);

    virtual void closeEvent(QCloseEvent *event) override;
    virtual void contextMenuEvent(QContextMenuEvent *event_ptr) override;
    virtual void focusInEvent(QFocusEvent *event) override;

    virtual bool getRawMarkers(uint32_t starting_frame_num, uint32_t frame_count, bool do_swap);
    virtual bool init_ccm_ctldata(fiPixmap *pixmap_ptr, bool& do_fpn);
    virtual fiPixmap mergeOverlay(uint32_t frame_num, fiPixmap merge_pixmap);
    virtual void updateEventInfo();
    virtual void updateGainInfo();

protected slots:
    void updatePfmd();
    void doViewAimingCursor();
    void doViewInfoCursor();

    virtual void doImageSave();
    virtual void saveImagesCompleted(bool savedOK);
    virtual void UpdateSaveProgress();
    virtual void UserCancelledSave();

private slots:
    void doPlaybackEvents();
    void eventSearchCanceled();
    void exportVideo();
    void showLocalFrametime();
    void showPfmd();
    void searchForEvents(bool do_swap);
    void showControls();
    void showSummary();
    void timeToPlay();
    void videoCurrentFrameChanged();
    void videoJumpTime();
    void videoPlayForward();
    void videoPlayReverse();
    void videoRateChanged(int new_rate);
    void videoReviewStop();
    void windowActivating();

private:
    void buildData();
    int calculateMinMaxFrameNumber(const QString& video_filename, int& min_framenum, int& max_framenum);
    bool findImageFile(uint32_t frame_num, fiPixmap *pixmap_dest);
    void loadPfmd(uint32_t frame_num);
    void playVideo();
    void readSettings();
    void saveSettings();
    bool splitFileName(QString video_filename, QString& path_basename, int& frame_num);

    static void SaveImages(QObject *parent);

protected:
    std::unique_ptr<FAS_VIDEO::FiVideoStream> AVIStreamReader;  //! Our AVI stream reader.
    std::unique_ptr<FiSaver> saver;             //!< The main saver object for Playback hierarchy.

    CameraInfo          *camera_info;           //!< Top pane of the window.
    uint32_t            frame_count;            //!< Number of frames we're playing with.
    uint32_t            last_frame_loaded;      //!< Frame number of the last frame we loaded.
    PlaybackMetaInfo    *pbMetaInfoPtr;         //!< Playback Metadata Information object.
    bool            open_display_panel_on_focus;   //!< True if we open display_panel_on_focus, false if not.
    QTimer          *playback_timer;            //!< Timer used to playback the video.
    QAction         *playback_action;           //!< Our window's menu action.
    CameraReview    *review_controls;           //!< Bottom pane.
    QString         root_file;                  //!< Our "root" file of the stack, or the video file, we're playing.
    CameraVideo     *video_player;              //!< Middle pane of the window.

    std::shared_ptr<SaveDialog> save_dlg;       //!< The save dialog allowing user to drive save operation.

    std::unique_ptr<fiProgressDialog> progressDialog;   //!< A simple progress dialog for operational status.
    std::unique_ptr<fiProgressDialog> xferDialog;       //!< A transfer progress dialog for save operations.

    QSharedPointer<CaptureMetaInfo> pmMetaDataInfo;     //!< The pm_metadata capture header inforamation.

    std::shared_ptr<FiDisplayLut> capLuts;      //!< Capture metadata custom luts.
    bool applyingCapLuts;                       //!< Allow the ability to turn off capLuts usage.
    volatile bool   thread_stop;                //!< Flag to tell our processing thread to stop.
    volatile bool   wait_for_save_complete;     //!< Flag to tell our processing thread to wait for save complete.
    volatile bool   save_had_error;             //!< Flag indicates that save found error condition.
    volatile bool   save_is_complete;           //!< Flag to tell our processing thread that save is complete.
    uint32_t        saver_count;                //!< How many frames are being saved here.
    uint32_t        saver_frame;                //!< The frame our saving thread is currently working on.
    QFuture<void>   saver_thread;               //!< Our saving thread "handle".
    QString         saver_image_filename;       //!< The filename that is/was saved.
    bool            swap_needed;                //!< True if we need to swap data.

    uint8_t         *r_lut;                     //!< The red channel LUT.
    uint8_t         *g_lut;                     //!< The green channel LUT.
    uint8_t         *b_lut;                     //!< The blue channel LUT.

    bool            do_whitebal;                //!< Flag to indicate that, the next time around, perform a white balance.
    int             whitebal;                   //!< One of GIGE_WHITE_BAL_XXX.
    double          r_wb_gain;
    double          g_wb_gain;
    double          b_wb_gain;

    bool            do_colorize;                //!< True if we're colorizing saturated pixels, false otherwise.
    double          r_gain_old;                 //!< The last known "good" r_gain setting.
    double          r_gain;                     //!< The gain for the red channel.
    QVector<double> r_gain_lut;
    QVector<double> r_gain_custom_lut;          //!< A custom red gain lut for custom curve of shared red channel with colorizer.
    double          g_gain;                     //!< The gain for the green channel.
    QVector<double> g_gain_lut;
    double          b_gain;                     //!< The gain for the blue channel.
    QVector<double> b_gain_lut;

    QString         file_ext;                   //!< The file extension we're working with.
    bool            has_pfmd_pm_file;
    bool            has_metadata_txt_file;

    bool        event_logic_is_and;             //!< True if we're using AND for the event logic, false for OR.
    QList<int>  marker_list;                    //!< List of frames with events which pass the selection criteria.
    QList<int>  event_selections;               //!< Which events we're interested in on which lines.


private:
    QVBoxLayout *layout_ptr;                    //!< Our window layout.
    uint32_t    min_framenum;                   //!< Minimum frame number from the file(s) we're playing.
    uint32_t    max_framenum;                   //!< Maximum frame number from the file(s) we're playing.
    bool        playback_forward;               //!< True if playback is in the "normal" direction, false if backwards/reverse.
    int         play_frames_inc;                //!< How many frames should be skipped based upon the playback rate.
    int         record_rate;                    //!< The rate (frames/second) for the recording.
    QString     stack_path;                     //!< Base path to the stack of images.
    int         trigger_frame;                  //!< The trigger frame for the recording.
    bool        using_subfolders;               //!< True if our stack has sub-folders.

    bool        use_event_markers;              //!< True if we're using event markers.
    int         search_start_index;             //!< Which frame to start searching on.

    bool        is_video_file;

    double      brightness;                     //!< The brightness for the image.
    double      contrast;                       //!< The contrast for the image.
    double      gamma;                          //!< The gamma for the image.
    QVector<double> gamma_lut;
};

#endif // PLAYBACK_MDI_H
