/*!
 * @file playback_meta_info.h
 * @brief This is the source file for displaying metafile information about stored video.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include "playback_meta_info.h"
#include "ui_playback_meta_info.h"

#include <errno.h>
#include <QFile>
#include <QHostAddress>
#include <QMessageBox>
#include <QSettings>
#include <QTextEdit>
#include <QVector>
#include <QtEndian>

#include "camera_mdi.h"
#include "eventfilter.h"
#include "fi_sensor.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"
#include "save_dialog.h"

namespace PLAYBACK_META_INFO
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
//#define DUMP_PFMD 1

//! The per-frame metadata SYNC MODE values.
enum
{
    TS3_SYNC_MODE_MASTER = 0,           //!< The camera was the sync master.
    TS3_SYNC_MODE_1PPF,                 //!< The camera was receiving a pulse per-frame.
    TS3_SYNC_MODE_1PPS                  //!< The camera was receiving a pulse per-second.
};

//! The per-frame metadata TIME STATE values.
const int       TS3_TIME_STATE_IRIG_MIN     = 4;


}   // End of namespace PLAYBACK_META_INFO.

using namespace PLAYBACK_META_INFO;

// Define the tags used in the metadata TXT file.
const QString   PlaybackMetaInfo::META_TAG_IMAGE            = QStringLiteral("Image");
const QString   PlaybackMetaInfo::META_TAG_CAMERA           = QStringLiteral("Camera");
const QString   PlaybackMetaInfo::META_TAG_RECORD           = QStringLiteral("Record");
const QString   PlaybackMetaInfo::META_TAG_NORMALIZATION    = QStringLiteral("Normalization");
const QString   PlaybackMetaInfo::META_TAG_EXPORT           = QStringLiteral("Export");
// Define the keys used in the metadata TXT file.
const QString   PlaybackMetaInfo::KEY_BIT_MODE            = QStringLiteral("bit_mode");
const QString   PlaybackMetaInfo::KEY_BLUE_BALANCE        = QStringLiteral("blue_balance");
const QString   PlaybackMetaInfo::KEY_BLUE_GAIN           = QStringLiteral("blue_gain");
const QString   PlaybackMetaInfo::KEY_BLUE_MATRIX         = QStringLiteral("blue_matrix");
const QString   PlaybackMetaInfo::KEY_BLUE_OFFSET         = QStringLiteral("blue_offset");
const QString   PlaybackMetaInfo::KEY_BRIGHTNESS          = QStringLiteral("brightness");
const QString   PlaybackMetaInfo::KEY_CAMERA_NAME         = QStringLiteral("camera_name");
const QString   PlaybackMetaInfo::KEY_CODEC               = QStringLiteral("codec");
const QString   PlaybackMetaInfo::KEY_COMMENT             = QStringLiteral("comment");
const QString   PlaybackMetaInfo::KEY_CONTRAST            = QStringLiteral("contrast");
const QString   PlaybackMetaInfo::KEY_END_FRAME           = QStringLiteral("end_frame");
const QString   PlaybackMetaInfo::KEY_FPS                 = QStringLiteral("fps");
const QString   PlaybackMetaInfo::KEY_FPGA_REV            = QStringLiteral("fpga_rev");
const QString   PlaybackMetaInfo::KEY_FRAME_COUNT         = QStringLiteral("frame_count");
const QString   PlaybackMetaInfo::KEY_GAMMA               = QStringLiteral("gamma");
const QString   PlaybackMetaInfo::KEY_GREEN_BALANCE       = QStringLiteral("green_balance");
const QString   PlaybackMetaInfo::KEY_GREEN_GAIN          = QStringLiteral("green_gain");
const QString   PlaybackMetaInfo::KEY_GREEN_MATRIX        = QStringLiteral("green_matrix");
const QString   PlaybackMetaInfo::KEY_GREEN_OFFSET        = QStringLiteral("green_offset");
const QString   PlaybackMetaInfo::KEY_HEIGHT              = QStringLiteral("height");
const QString   PlaybackMetaInfo::KEY_MAC_ADDRESS         = QStringLiteral("mac_address");
const QString   PlaybackMetaInfo::KEY_MAKE                = QStringLiteral("make");
const QString   PlaybackMetaInfo::KEY_MODEL               = QStringLiteral("model");
const QString   PlaybackMetaInfo::KEY_MULTI_SLOPE         = QStringLiteral("multi_slope");
const QString   PlaybackMetaInfo::KEY_RAW                 = QStringLiteral("raw");
const QString   PlaybackMetaInfo::KEY_RED_BALANCE         = QStringLiteral("red_balance");
const QString   PlaybackMetaInfo::KEY_RED_GAIN            = QStringLiteral("red_gain");
const QString   PlaybackMetaInfo::KEY_RED_MATRIX          = QStringLiteral("red_matrix");
const QString   PlaybackMetaInfo::KEY_RED_OFFSET          = QStringLiteral("red_offset");
const QString   PlaybackMetaInfo::KEY_ROI_X               = QStringLiteral("roi_x");
const QString   PlaybackMetaInfo::KEY_ROI_Y               = QStringLiteral("roi_y");
const QString   PlaybackMetaInfo::KEY_SENSOR_GAIN         = QStringLiteral("sensor_gain");
const QString   PlaybackMetaInfo::KEY_SENSOR_OPTS         = QStringLiteral("sensor_options");
const QString   PlaybackMetaInfo::KEY_SENSOR_TYPE         = QStringLiteral("sensor_type");
const QString   PlaybackMetaInfo::KEY_SHUTTER_SPEED       = QStringLiteral("shutter_speed");
const QString   PlaybackMetaInfo::KEY_SOFTWARE_VERSION    = QStringLiteral("software_version");
const QString   PlaybackMetaInfo::KEY_START_FRAME         = QStringLiteral("start_frame");
const QString   PlaybackMetaInfo::KEY_SYNC_IN             = QStringLiteral("sync_in");
const QString   PlaybackMetaInfo::KEY_SYNC_OUT            = QStringLiteral("sync_out");
const QString   PlaybackMetaInfo::KEY_TIME_STAMP          = QStringLiteral("time_stamp");
const QString   PlaybackMetaInfo::KEY_TRIGGER_FRAME       = QStringLiteral("trigger_frame");
const QString   PlaybackMetaInfo::KEY_TRIGGER_SETTING     = QStringLiteral("trigger_setting");
const QString   PlaybackMetaInfo::KEY_WIDTH               = QStringLiteral("width");


/*!
 * @brief Constructor for the window it sets up the user interface which consists of an text edit inside a scroll area
 *
 * @param parent - window who is our parent
 *
 * @return none
 */
PlaybackMetaInfo::PlaybackMetaInfo(QWidget *parent) :
    QWidget(parent),
    ui(new Ui::PlaybackMetaInfo)
{
    ui->setupUi(this);

    // Initialize our variables.
    meta_database.clear();
    per_frame_avail = false;
    raw_buffer.clear();
    raw_file.clear();
    use_export = true;
    pfmd_len_specified = 0;
}

/*!
 * @brief Destructor for the camera info dialog. Does nothing at this point
 */
PlaybackMetaInfo::~PlaybackMetaInfo()
{
    delete ui;
}

/*!
 * @brief   Build the raw per-frame metadata file from the XML data.
 *
 * @param[in] raw_file      The full filename for the new per-frame metadata file.
 * @param[in] pfmd_len      Number of bytes in the per-frame metadata.
 *
 * @return  True on success, false otherwise.
 */
bool PlaybackMetaInfo::buildRawData(const QString& raw_file, int pfmd_len)
{
    bool            series3;
    QDomElement     dom_root;           // xml token "root"
    QDomElement     dom_frames;         // xml token "frames"
    QDomElement     dom_frame;          // xml token "frame"
    QString         msg_str;
    QFile           raw_metafile(raw_file);

    // Only build the file if one doesn't exist.
    if (raw_metafile.exists())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << raw_file << "already exists!";
        return false;
    }

    // Sanity check the pfmd_len.
    if (pfmd_len == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "pfmd_len is 0!?";
        return false;
    }

    // Make sure we have XML data.
    if (!per_frame_avail)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No XML file to build from!";
        return false;
    }

    // Create our file.
    if (!raw_metafile.open(QIODevice::WriteOnly))
    {
        msg_str = QStringLiteral("Unable to open %2, err=%3").arg(raw_file, raw_metafile.errorString());
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        return false;
    }
    raw_metafile.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadOther);

    // Clear our buffer.
    QVector<uint32_t> pm_buffer((uint32_t)pfmd_len / sizeof(uint32_t));

    // Remember if this is a series3 camera or not.
    series3 = isSeries3();

    // Run through the XML and build up our raw metadata file.
    // Get ourselves to the frame part of the file, then start
    // saving out the metadata.
    dom_root = doc.firstChildElement(QStringLiteral("FastecMetadata"));
    dom_frames = dom_root.firstChildElement(QStringLiteral("frames"));
    dom_frame = dom_frames.firstChildElement(QStringLiteral("frame"));
    while (!dom_frame.isNull())
    {
        uint32_t    markers;
        QString     tmp_str;
        uint32_t    tmp_val;
        uint32_t    xml_frame_num;      // the frame number from the xml file as a number

        // Get the time.
        tmp_str = dom_frame.firstChildElement(QStringLiteral("time")).text();
        tmp_val = 0;
        if (series3)
        {
            pm_buffer[ 0 ] = qToBigEndian(stringToSBS(tmp_str, tmp_val));
            pm_buffer[ 1 ] = qToBigEndian(tmp_val);
        }
        else
        {
            pm_buffer[ 0 ] = stringToSBS(tmp_str, tmp_val);
            pm_buffer[ 1 ] = tmp_val;
        }

        // Get the frame number.
        tmp_str = dom_frame.firstChildElement(QStringLiteral("number")).text();
        xml_frame_num = tmp_str.toUInt();
        if (series3)
            pm_buffer[ 3 ] = qToBigEndian(xml_frame_num);
        else
            pm_buffer[ 3 ] = xml_frame_num;

        // Get the markers for this frame.
        tmp_str = dom_frame.firstChildElement(QStringLiteral("markers")).text();
        markers = tmp_str.toUInt(nullptr, 16);
        if (series3)
            pm_buffer[ 2 ] = qToBigEndian(markers);
        else
            pm_buffer[ 2 ] = markers;

        // Write the data.
        raw_metafile.write((const char *)pm_buffer.constData(), pm_buffer.size() * sizeof(uint32_t));

        // Move to the next frame.
        dom_frame = dom_frame.nextSiblingElement(QStringLiteral("frame"));
    }

    // Close our file.
    raw_metafile.close();

    // Return success.
    return true;
}

/*!
 * @brief Display per frame metadata for a specific frame. Note that this will only
 *        display information in the ui->eb_metaframe if there was an XML file because that
 *        is the only way the private data member PlaybackMetaInfo::per_frame_avail is true.
 *
 * @param[in] frame_num     The number of the frame for which we want to display the PFMD information.
 */
void PlaybackMetaInfo::displayPerFrameMetadata(uint32_t frame_num)
{
    QDomElement     dom_root;           // xml token "root"
    QDomElement     dom_frames;         // xml token "frames"
    QDomElement     dom_frame;          // xml token "frame"
    QString     msg_str;
    QString     time;
    QString     tmp_str;
    QString     sync_str;
    QString     time_state_str;

    if (per_frame_avail)
    {
        // we know that per frame data is available so get ourselves to the
        // frame part of the file. Then look through all the frames until we
        // find the one we want.
        dom_root = doc.firstChildElement(QStringLiteral("FastecMetadata"));
        dom_frames = dom_root.firstChildElement(QStringLiteral("frames"));
        dom_frame = dom_frames.firstChildElement(QStringLiteral("frame"));
        while (!dom_frame.isNull())
        {
            uint32_t    xml_frame_num;      // the frame number from the xml file as a number

            tmp_str = dom_frame.firstChildElement(QStringLiteral("number")).text();
            xml_frame_num = tmp_str.toUInt();
            if (xml_frame_num == frame_num)
            {
                int         irig;
                int         markers;
                int         sync_mode;
                int         time_state;

                // found the frame we are looking for so display it on the screen and leave.
                time = dom_frame.firstChildElement(QStringLiteral("time")).text();
                tmp_str = dom_frame.firstChildElement(QStringLiteral("irig")).text();
                irig = tmp_str.toInt();
                tmp_str = dom_frame.firstChildElement(QStringLiteral("markers")).text();
                markers = tmp_str.toInt(nullptr, 16);
                tmp_str = dom_frame.firstChildElement(QStringLiteral("sync_mode")).text();
                sync_mode = tmp_str.toInt();
                tmp_str = dom_frame.firstChildElement(QStringLiteral("time_state")).text();
                time_state = tmp_str.toInt();

                switch (sync_mode)
                {
                case TS3_SYNC_MODE_MASTER:  sync_str = tr("Internal"); break;
                case TS3_SYNC_MODE_1PPF:    sync_str = tr("1PPF"); break;
                case TS3_SYNC_MODE_1PPS:    sync_str = tr("1PPS"); break;
                default:            sync_str = tr("Unknown"); break;
                }
                if (time_state == 0)
                {
                    time_state_str = tr("Internal");
                }
                else if (time_state < TS3_TIME_STATE_IRIG_MIN)
                {
                    time_state_str = tr("1PPS");
                }
                else
                {
                    time_state_str = tr("IRIG");
                }
                msg_str = tr("Time:%1 Markers:%2%3%4%5%6%7 IRIG:%8 Sync:%9 Time:%10")
                          .arg(time).arg((markers >> 5) & 1)
                          .arg((markers >> 4) & 1)
                          .arg((markers >> 3) & 1)
                          .arg((markers >> 2) & 1)
                          .arg((markers >> 1) & 1)
                          .arg(markers & 1)
                          .arg(irig).arg(sync_str, time_state_str);
                ui->eb_meta_frame->setText(msg_str);
                ui->eb_meta_frame->setCursorPosition(0);
                return;
            }
            dom_frame = dom_frame.nextSiblingElement(QStringLiteral("frame"));
        }
    }

    // we either don't have the per frame metadata file or we went through all the frames and didn't
    // find the one we are lookinig for so just put the placeholder text in the box.
    ui->eb_meta_frame->setText(tr("No per frame metadata"));
}

/*!
 * @brief Find the frame at the specified time.
 *
 * @param[in] time_sec      The number of seconds for the frame.
 * @param[in] time_usec     The number of microseconds for the frame.
 * @param[opt-in] pfmd_len  [OPTIONAL] Number of bytes in the per-frame metadata, 0 if not known.
 * @return The frame number found, or 0 if there was some error.
 */
uint32_t PlaybackMetaInfo::findFrame(uint32_t time_sec, uint32_t time_usec, int pfmd_len)
{
    uint32_t    frame_num = 0;
    uint32_t    sbs;
    uint32_t    sbs_usec;

    // How many bytes is our per-frame metadata?
    if (pfmd_len == 0)
        pfmd_len = pfmdLength();

    // Are we going to the XML or the raw metadata?
    if (raw_file.isEmpty() == false)
    {
        uint32_t    frame_count;
        QFile       raw_file(getRawfile());
        QVector<uint32_t> raw_buffer((uint32_t)pfmd_len / sizeof(uint32_t));

        // Open our file.
        raw_file.open(QIODevice::ReadOnly);
        raw_file.seek(0);

        // Get our frame count.
        frame_count = (uint32_t)(raw_file.size() / pfmd_len);

        // Loop through each of the frame's metadata.
        for (frame_num = 0; frame_num < frame_count; ++frame_num)
        {
            // Read the frame's metadata.
            if (raw_file.read((char *)raw_buffer.data(), pfmd_len) < 0)
            {
                qDebug() << __FUNCTION__ << "error reading the file!";
                break;
            }

            // Get the seconds & microseconds in a format we can compare to.
            if (isSeries3())
            {
                sbs = qFromBigEndian(raw_buffer[0]);
                sbs_usec = qFromBigEndian(raw_buffer[1]) & 0xFFFFF;
            }
            else
            {
                sbs = raw_buffer[0];
                sbs_usec = (raw_buffer[1] & 0xFFFFF);
            }

            // Do we have a match?
            if ((sbs == time_sec) && (sbs_usec == time_usec))
            {
                break;
            }

            // Did we go past the time we're looking for?
            else if ((sbs > time_sec) || ((sbs == time_sec) && (sbs_usec > time_usec)))
            {
                break;
            }
        }

        // Close the file.
        raw_file.close();
    }
    else if (per_frame_avail)
    {
        QDomElement     dom_root;           // xml token "root"
        QDomElement     dom_frames;         // xml token "frames"
        QDomElement     dom_frame;          // xml token "frame"

        // We know that per frame data is available so get ourselves to the
        // frame part of the file. Then look through all the frames until we
        // find the one we want.
        dom_root = doc.firstChildElement(QStringLiteral("FastecMetadata"));
        dom_frames = dom_root.firstChildElement(QStringLiteral("frames"));
        dom_frame = dom_frames.firstChildElement(QStringLiteral("frame"));
        while (!dom_frame.isNull())
        {
            QString time_str;

            // Found the frame we are looking for so grab the time.
            time_str = dom_frame.firstChildElement(QStringLiteral("time")).text();

            // Convert the string into it's binary equivalent.
            sbs_usec = 0;
            sbs = stringToSBS(time_str, sbs_usec);

            // Do we have a match?
            if ((sbs == time_sec) &&
                (sbs_usec == time_usec))
            {
                // Get the frame number from the XML.
                frame_num = dom_frame.firstChildElement(QStringLiteral("number")).text().toUInt();
                break;
            }

            // Did we go past the time we're looking for?
            else if ((sbs > time_sec) || ((sbs == time_sec) && (sbs_usec > time_usec)))
            {
                // Get the frame number from the XML.
                frame_num = dom_frame.firstChildElement(QStringLiteral("number")).text().toUInt();
                break;
            }

            // Move to the next one.
            dom_frame = dom_frame.nextSiblingElement(QStringLiteral("frame"));
        }
    }

    // Return the frame we found.
    return frame_num;
}

/*!
 * @brief Find a frame given a time as a string value.
 * @param time_str      The time formated as "YY:DDD:HH:MM:SS.microseconds".
 * @return The result of the [search](@ref PlaybackMetaInfo::find_frame).
 */
uint32_t PlaybackMetaInfo::findFrame(const QString& time_str)
{
    uint32_t    time_sec;
    uint32_t    time_usec(0);

    // Convert the string into it's binary equivalent.
    time_sec = stringToSBS(time_str, time_usec);

    // Now do the search.
    return findFrame(time_sec, time_usec);
}

const QVariant PlaybackMetaInfo::getExport(const QString& key_name)
{
    static QVariant rc_value;

    // Reset our return value.
    rc_value.clear();

    // Are we ignoring exports?
    if (!use_export)
        return rc_value;

    // Is there an export section?
    if (meta_database.contains(META_TAG_EXPORT))
    {
        META_DB     tag_db;

        // Retrieve the appropriate entries.
        tag_db = meta_database.value(META_TAG_EXPORT).toMap();

        // Is our key in the export section?
        if (tag_db.contains(key_name))
            rc_value = tag_db.value(key_name);
    }

    return rc_value;
}

uint32_t PlaybackMetaInfo::getPfmdTime(uint32_t frame_num, uint32_t& micro_sec, int pfmd_len)
{
    uint32_t sbs = 0;

    Q_UNUSED(pfmd_len)

    // Initialize the microseconds.
    micro_sec = 0;

    // Are we going to the XML or the raw metadata?
    if (!raw_file.isEmpty())
    {
        QByteArray *pfmd_array = nullptr;
        uint32_t raw_buffer[4];

        pfmd_array = getRawPFMD(frame_num, isSeries3());
        if (pfmd_array != nullptr)
        {
            // Copy the data over to our array, the release the memory.
            memcpy(raw_buffer, pfmd_array->constData(), sizeof(raw_buffer));
            delete pfmd_array;

            // Grab the fields we need.
            sbs = raw_buffer[0];
            micro_sec = (raw_buffer[1] & 0xFFFFF);

        }
    }
    else if (per_frame_avail)
    {
        QDomElement     dom_root;           // xml token "root"
        QDomElement     dom_frames;         // xml token "frames"
        QDomElement     dom_frame;          // xml token "frame"

        // We know that per frame data is available so get ourselves to the
        // frame part of the file. Then look through all the frames until we
        // find the one we want.
        dom_root = doc.firstChildElement(QStringLiteral("FastecMetadata"));
        dom_frames = dom_root.firstChildElement(QStringLiteral("frames"));
        dom_frame = dom_frames.firstChildElement(QStringLiteral("frame"));
        while (!dom_frame.isNull())
        {
            QString     tmp_str;
            uint32_t        xml_frame_num;      // the frame number from the xml file as a number

            // Get the frame number from the XML.
            tmp_str = dom_frame.firstChildElement(QStringLiteral("number")).text();
            xml_frame_num = tmp_str.toUInt();

            // Is this the one we're interested in?
            if (xml_frame_num == frame_num)
            {
                QString time_str;

                // Found the frame we are looking for so grab the time.
                time_str = dom_frame.firstChildElement(QStringLiteral("time")).text();

                // Convert to something we can use.
                sbs = stringToSBS(time_str, micro_sec);
                break;
            }

            // Move to the next one.
            dom_frame = dom_frame.nextSiblingElement(QStringLiteral("frame"));
        }
    }

    // Return the time.
    return sbs;
}

/*!
 * @brief Get the timestamp associated with the "frame_num" parameter from the per-frame
 *        metadata.
 *
 * @param[in] frame_num     The frame number we're interested in.
 * @param[out] time_str     The timestamp associated with the frame number.
 * @param[opt-in] pfmd_len  [OPTIONAL] Number of bytes in the per-frame metadata, 0 if not known.
 * @return True if we were successful, false on error.
 */
bool PlaybackMetaInfo::getPfmdTimeStr(uint32_t frame_num, QString& time_str, int pfmd_len)
{
    bool rc;

    Q_UNUSED(pfmd_len)

    // Initialize our return value.
    time_str.clear();
    rc = false;

    // Are we going to the XML or the raw metadata?
    if (!raw_file.isEmpty())
    {
        QByteArray *pfmd_array = nullptr;
        uint32_t raw_buffer[4];

        pfmd_array = getRawPFMD(frame_num, isSeries3());
        if (pfmd_array != nullptr)
        {
            uint32_t    sbs;
            uint32_t    sbs_msec;
            struct tm   sbs_tm;

            // Copy the data over to our array, the release the memory.
            memcpy(raw_buffer, pfmd_array->constData(), sizeof(raw_buffer));
            delete pfmd_array;

            // Clear out the structure.
            memset(&sbs_tm, 0, sizeof(struct tm));

#ifdef DUMP_PFMD
            {
                QString msg_str;
                msg_str = tr("%1 %2 %3 %4").arg(raw_buffer[0], 8, 16, QChar('0')).arg(raw_buffer[1], 8, 16, QChar('0')).arg(raw_buffer[2], 8, 16, QChar('0')).arg(raw_buffer[3], 8, 16, QChar('0'));
                QLOG_TRACE() << __PRETTY_FUNCTION__ << msg_str;
                msg_str = tr("%1 %2 %3 %4").arg(raw_buffer[4], 8, 16, QChar('0')).arg(raw_buffer[5], 8, 16, QChar('0')).arg(raw_buffer[6], 8, 16, QChar('0')).arg(raw_buffer[7], 8, 16, QChar('0'));
                QLOG_TRACE() << __PRETTY_FUNCTION__ << msg_str;
            }
#endif

            sbs = raw_buffer[0];
            sbs_msec = raw_buffer[1] & 0xFFFFF;

            // Get years.
            sbs_tm.tm_year = sbs / (365 * 24 * 60 * 60);
            sbs -= ((uint32_t)sbs_tm.tm_year * (365 * 24 * 60 * 60));
            sbs_tm.tm_yday = sbs / (24 * 60 * 60);
            sbs -= ((uint32_t)sbs_tm.tm_yday * (24 * 60 * 60));
            sbs_tm.tm_hour = sbs / (60 * 60);
            sbs -= ((uint32_t)sbs_tm.tm_hour * (60 * 60));
            sbs_tm.tm_min = sbs / 60;
            sbs_tm.tm_sec = (int)sbs - (sbs_tm.tm_min * 60);

            // Convert the time to a string.
            time_str = tr("%1:%2:%3:%4:%5.%6").arg(sbs_tm.tm_year, 2, 10, QLatin1Char('0'))
                       .arg(sbs_tm.tm_yday + 1, 3, 10, QLatin1Char('0'))
                       .arg(sbs_tm.tm_hour, 2, 10, QLatin1Char('0'))
                       .arg(sbs_tm.tm_min, 2, 10, QLatin1Char('0'))
                       .arg(sbs_tm.tm_sec, 2, 10, QLatin1Char('0'))
                       .arg(sbs_msec, 6, 10, QLatin1Char('0'));
            rc = true;
        }
    }
    else if (per_frame_avail)
    {
        QDomElement     dom_root;           // xml token "root"
        QDomElement     dom_frames;         // xml token "frames"
        QDomElement     dom_frame;          // xml token "frame"

        // We know that per frame data is available so get ourselves to the
        // frame part of the file. Then look through all the frames until we
        // find the one we want.
        dom_root = doc.firstChildElement(QStringLiteral("FastecMetadata"));
        dom_frames = dom_root.firstChildElement(QStringLiteral("frames"));
        dom_frame = dom_frames.firstChildElement(QStringLiteral("frame"));
        while (!dom_frame.isNull())
        {
            QString     tmp_str;
            uint32_t        xml_frame_num;      // the frame number from the xml file as a number

            // Get the frame number from the XML.
            tmp_str = dom_frame.firstChildElement(QStringLiteral("number")).text();
            xml_frame_num = tmp_str.toUInt();

            // Is this the one we're interested in?
            if (xml_frame_num == frame_num)
            {
                // Found the frame we are looking for so grab the time.
                time_str = dom_frame.firstChildElement(QStringLiteral("time")).text();
                rc = true;
                break;
            }

            // Move to the next one.
            dom_frame = dom_frame.nextSiblingElement(QStringLiteral("frame"));
        }
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Return the fully qualified path to the raw per-frame metadata file.
 */
const QString& PlaybackMetaInfo::getRawfile() const
{
    return raw_file;
}

/*!
 * @brief Read the raw metadata file and build up an array of the event flags.
 *
 * @param[opt-in] pfmd_len  [OPTIONAL] Number of bytes in the per-frame metadata, 0 if not known.
 * @return The array of event flags, or nullptr on error. It is the responsibility of the
 *         caller to release the array memory.
 */
uint8_t *PlaybackMetaInfo::getRawEventFlags(int pfmd_len, bool do_swap)
{
    uint8_t     *event_flag_ptr;
    uint32_t    frame_no;
    const uint8_t *data_ptr;
    const QByteArray *raw_data;

    // Make sure we have a raw file.
    if (getRawfile().isEmpty())
        return nullptr;

    // Sanity check.
    if (getFrameCount() == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No frames!";
        return nullptr;
    }

    // Get the raw data.
    raw_data = getRawBuffer();
    if (raw_data == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No raw data.";
        return nullptr;
    }

    // Allocate our data buffer.
    event_flag_ptr = new uint8_t[ getFrameCount() ];
    if (event_flag_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate memory!";
        return nullptr;
    }

    // How many bytes is our per-frame metadata?
    if (pfmd_len == 0)
        pfmd_len = pfmdLength();

    // Get our pointer into the raw data.
    data_ptr = (const uint8_t *)raw_data->constData();

    // Loop through the file and pull in the meta data information.
    for (frame_no = 0; frame_no < getFrameCount(); ++frame_no, data_ptr += pfmd_len)
    {
        uint32_t flag_hold;

        // Copy the event flag information located in the 3rd word (offset=2).
        memcpy(&flag_hold, data_ptr + (2 * sizeof(uint32_t)), sizeof(flag_hold));
        if (do_swap)
            flag_hold = qToBigEndian(flag_hold);

        // Save the marker information into our buffer.
        event_flag_ptr[ frame_no ] = (uint8_t)((flag_hold >> FM_EVENT_FILTERS::EVENT_FLAG_OFFSET) & 0xFF);
    }

    // Return our array pointer.
    return event_flag_ptr;
}

/*!
 * @brief Load the raw per-frame metadata into a buffer so our processing of it can be faster.
 *
 * @return Pointer to the buffer if successful, nullptr if an error happened.
 */
const QByteArray *PlaybackMetaInfo::getRawBuffer()
{
    // Make sure we have a raw file.
    if (getRawfile().isEmpty())
        return nullptr;

    // Do we already have the data loaded?
    if (raw_buffer.isEmpty())
    {
        // Open our file.
        QFile       raw_qfile(getRawfile());
        raw_qfile.open(QIODevice::ReadOnly);

        // Read the file into the buffer.
        raw_buffer = raw_qfile.readAll();
        if (raw_buffer.isEmpty())
        {
            // Try to log something about what just happened.
            QLOG_WARN() << __PRETTY_FUNCTION__ << "No data available? Err:" << raw_qfile.errorString();
        }

        // Close the file.
        raw_qfile.close();
    }

    // Return our buffer.
    return &raw_buffer;
}

/*!
 * @brief Read the raw byte array for the per-frame metadata associated with the specified frame.
 * @param frame_num     The frame we are interested in.
 * @param do_swap       Are the first 4 bytes of metadata supposed to be swapped?
 * @return Pointer the the byte array, or nullptr if an error.
 */
QByteArray *PlaybackMetaInfo::getRawPFMD(uint32_t frame_num, bool do_swap)
{
    QByteArray  *pfmd_array = nullptr;
    qint64      frame_pos;
    int pfmd_len = 0;           //!< The size of the per frame metadata data length is calculated below.

    // Make sure we have a raw file.
    if (getRawfile().isEmpty())
        return nullptr;

    // How many bytes is our per-frame metadata?
    pfmd_len = pfmdLength();

    // Open our file.
    QVector<uint32_t> raw_buffer((uint32_t)pfmd_len / sizeof(uint32_t));
    QFile       raw_qfile(getRawfile());
    raw_qfile.open(QIODevice::ReadOnly);

    // Calculate our frame position within the raw file.
    frame_pos = (qint64)(frame_num) * pfmd_len;

    // Seek to the right spot in the file.
    if (!raw_qfile.seek(frame_pos))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString("Seek failed, bad frame_num [%1]?").arg(frame_num);
    }

    // Read the frame's metadata.
    else if (raw_qfile.read((char *)raw_buffer.data(), pfmd_len) < 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error reading the file:" << raw_qfile.errorString();
    }
    else
    {
        // Are we supposed to swap?
        if (do_swap)
        {
            // Swap the words in our raw data buffer.
            raw_buffer[0] = qToBigEndian(raw_buffer[0]);
            raw_buffer[1] = qToBigEndian(raw_buffer[1]);
            raw_buffer[2] = qToBigEndian(raw_buffer[2]);
            raw_buffer[3] = qToBigEndian(raw_buffer[3]);
        }

        // Allocate our QByteArray.
        pfmd_array = new QByteArray((const char *)raw_buffer.constData(), pfmd_len);
    }

    // Close the file.
    raw_qfile.close();

    // Return our PFMD pointer.
    return pfmd_array;
}

const QString PlaybackMetaInfo::getBitMode()
{
    QString bm_str;
    QString ret_str;
    QStringList tokens;

    // Get the string.
    bm_str = getValue(META_TAG_IMAGE, KEY_BIT_MODE).toString();
    if (bm_str.isEmpty())
        return QString();

    // Try to split the string at the colon.
    tokens = bm_str.split(QChar::fromLatin1(':'));

    // Did we get our two numbers?
    if (tokens.count() != 2)
    {
        // Keep the information as it is.
        ret_str = bm_str;
    }
    else
    {
        // Need to convert numeric modes, such as 11:4, to 1-based.
        int first_num = tokens.at(0).toInt();
        int second_num = tokens.at(1).toInt();

        // Build up our return string.
        ret_str = QString("%1:%2").arg(first_num + 1).arg(second_num + 1);
    }

    // Return our string.
    return ret_str;
}

const QString PlaybackMetaInfo::getCameraName()
{
    return getValue(META_TAG_CAMERA, KEY_CAMERA_NAME).toString();
}

const QString PlaybackMetaInfo::getComment()
{
    return getValue(META_TAG_IMAGE, KEY_COMMENT).toString();
}

uint32_t PlaybackMetaInfo::getEndFrame()
{
    QVariant    rc_value;

    // Is there an end frame in the export section?
    rc_value = getExport(KEY_END_FRAME);

    // Did we get anything?
    if (rc_value.isNull())
        rc_value = getValue(META_TAG_IMAGE, KEY_END_FRAME);

    // Return our value.
    return rc_value.toUInt();
}

uint32_t PlaybackMetaInfo::getExposure()
{
    return getValue(META_TAG_RECORD, KEY_SHUTTER_SPEED).toUInt();
}

uint32_t PlaybackMetaInfo::getFrameCount()
{
    QVariant    rc_value;

    // Is there a frame count in the export section?
    rc_value = getExport(KEY_FRAME_COUNT);

    // Did we get anything?
    if (rc_value.isNull())
        rc_value = getValue(META_TAG_IMAGE, KEY_FRAME_COUNT);

    // Return our value.
    return rc_value.toUInt();
}

uint32_t PlaybackMetaInfo::getFramerate()
{
    return getValue(META_TAG_RECORD, KEY_FPS).toUInt();
}

uint32_t PlaybackMetaInfo::getHeight()
{
    QVariant    rc_value;

    // Is there a height in the export section?
    rc_value = getExport(KEY_HEIGHT);

    // Did we get anything?
    if (rc_value.isNull())
        rc_value = getValue(META_TAG_IMAGE, KEY_HEIGHT);

    // Return our value.
    return rc_value.toUInt();
}

const QString PlaybackMetaInfo::getMacAddress()
{
    return getValue(META_TAG_CAMERA, KEY_MAC_ADDRESS).toString();
}

const QString PlaybackMetaInfo::getManufacturer()
{
    return getValue(META_TAG_CAMERA, KEY_MAKE).toString();
}

const QString PlaybackMetaInfo::getModel()
{
    return getValue(META_TAG_CAMERA, KEY_MODEL).toString();
}

int PlaybackMetaInfo::pfmdLength()
{
    int ret_len;

    // Was the pfmd_len already specified?
    if (pfmd_len_specified)
        return pfmd_len_specified;

    // Try to determine if it is a series3 or a new model.
    if (isSeries3())
        ret_len = GIGE_FRAME_METADATA_LEN;
    else
        ret_len = (GIGE_FRAME_METADATA_LEN * 2);

    return ret_len;
}

/*!
 * @brief Convert the user-friendly bit-mode string to the appropriate
 *        constant.
 * @return The appropriate GIGE_BITMODE_xxx constant.
 */
uint8_t PlaybackMetaInfo::getConvertedBitMode()
{
    uint8_t bit_sel = GIGE_BITMODE_UPPER_8_BITS;    // Set a default value.

    // Convert the bit-mode string to a bit select value.
    QStringList bm_values = getBitMode().split(':');
    if (bm_values.count() != 2)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "I expected a bit mode of <m:n>, but I got:" << getBitMode();
    }
    else
    {
        // Get the bounds as integers. Remember that the values have been bumped by 1.
        int upper = bm_values.at(0).toInt() - 1;
        int lower = bm_values.at(1).toInt() - 1;

        switch (upper)
        {
        case 7:
            bit_sel = GIGE_BITMODE_7_0;
            break;
        case 8:
            bit_sel = GIGE_BITMODE_8_1;
            break;
        case 9:
            switch (lower)
            {
            case 0:
                bit_sel = GIGE_BITMODE_9_0;
                break;
            case 2:
                bit_sel = GIGE_BITMODE_9_2;
                break;
            default:
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected lower bit-mode:" << getBitMode();
                break;
            }
            break;
        case 10:
            switch (lower)
            {
            case 1:
                bit_sel = GIGE_BITMODE_10_1;
                break;
            case 3:
                bit_sel = GIGE_BITMODE_10_3;
                break;
            default:
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected lower bit-mode:" << getBitMode();
                break;
            }
            break;
        case 11:
            switch (lower)
            {
            case 0:
                bit_sel = GIGE_BITMODE_11_0;
                break;
            case 2:
                bit_sel = GIGE_BITMODE_11_2;
                break;
            case 4:
                bit_sel = GIGE_BITMODE_11_4;
                break;
            default:
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected lower bit-mode:" << getBitMode();
                break;
            }
            break;
        default:
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected upper bit-mode:" << getBitMode();
            break;
        }
    }

    // Return our value.
    return bit_sel;
}

/*!
 * @brief Convert the user-friendly sensor options string to the appropriate constant(s).
 * @return The appropriate GIGE_SENSOPT_xxx value(s).
 */
uint32_t PlaybackMetaInfo::convertSensorOpts()
{
    uint32_t bin_value = 1;
    uint32_t subs_value = 1;

    QStringList opt_values = getSensorOpt().split(':');
    if (opt_values.count() == 2)
    {
        if (opt_values.at(0).compare(QStringLiteral("bin2x"), Qt::CaseInsensitive) == 0)
            bin_value = 2;
        else if (opt_values.at(0).compare(QStringLiteral("bin4x"), Qt::CaseInsensitive) == 0)
            bin_value = 4;

        if (opt_values.at(1).compare(QStringLiteral("subs2x"), Qt::CaseInsensitive) == 0)
            subs_value = 2;
        else if (opt_values.at(1).compare(QStringLiteral("subs4x"), Qt::CaseInsensitive) == 0)
            subs_value = 4;
    }

    return ((subs_value << GIGE_SENSOPT_SUBS_SHIFT) | (bin_value << GIGE_SENSOPT_BIN_SHIFT));
}

/*!
 * @brief Recover the partition's capture metadata.
 */
void PlaybackMetaInfo::recoverPmMetadata()
{
    int i;

    // Sanity check.
    if (pm_ptr.get() != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "We already have the pm_metadata.";
        return;
    }

    // Allocate our structure.
    pm_ptr = std::unique_ptr<pm_metadata>(new pm_metadata);
    if (pm_ptr.get() == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate pm_metadata";
        return;
    }

    // Clear the structure.
    std::memset(pm_ptr.get(), 0, sizeof(pm_metadata));

    // Fill in the structure as best we can.
    std::strncpy(pm_ptr->man_name, getManufacturer().toLatin1().constData(), GIGE_DSC_ACK_VENDOR_B_LEN);
    std::strncpy(pm_ptr->model_name, getModel().toLatin1().constData(), GIGE_DSC_ACK_MODEL_B_LEN);
    std::strncpy(pm_ptr->app_ver, getVersion().toLatin1().constData(), GIGE_DSC_ACK_VERSION_B_LEN);
    std::strncpy(pm_ptr->camera_name, getCameraName().toLatin1().constData(), GIGE_DSC_ACK_USER_NAME_B_LEN);

    QStringList mac_values = getMacAddress().split(':');
    if (mac_values.count() != GIGE_DSC_ACK_MAC_B_LEN)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Didn't get 6 values from mac:" << getMacAddress();
    }
    else
    {
        // Convert each entry to an integer.
        for (i = 0; i < GIGE_DSC_ACK_MAC_B_LEN; ++i)
            pm_ptr->mac_addr[i] = mac_values.at(i).toInt(nullptr, 16);
    }

    // Get our bit selection. This is represented as a user-friendly
    // string, so we need to convert that back into one of our constants.
    if (isRaw())
    {
        // This is a raw image. The bit depth is how it was captured.
        pm_ptr->bit_sel = getConvertedBitMode();
    }
    else
    {
        // This is a processed image. It had better be monochrome!
        if (isColor())
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Processed color images are not currently supported!";
            return;
        }
        else
        {
            // A processed monochrome image will be 8-bits.
            pm_ptr->bit_sel = GIGE_BITMODE_7_0;
        }
    }

    pm_ptr->roi_x = getX();
    pm_ptr->roi_y = getY();
    pm_ptr->roi_width = getWidth();
    pm_ptr->roi_height = getHeight();

    pm_ptr->fps = getFramerate();
    pm_ptr->shutter_speed = getExposure();
    pm_ptr->sensor_type = getSensorType();

    pm_ptr->frame_count = getFrameCount();
    // TODO: pm_ptr->frame_size = ???
    pm_ptr->first_frame = getStartFrame();
    pm_ptr->trig_frame = getTriggerFrame();

    uint32_t microsec;
    pm_ptr->trig_time_s = getTimestamp(microsec);
    pm_ptr->trig_time_ns = microsec * 1000;

    std::strncpy((char *)&pm_ptr->comment[0], getComment().toLatin1().constData(), sizeof(pm_ptr->comment));

    // The sensor options (binning, subsampling) are represented as a user-friendly
    // string. We need to convert that back into the appropriate constants.
    pm_ptr->sensor_opts = convertSensorOpts();
}

uint32_t PlaybackMetaInfo::getSensorGain()
{
    uint32_t sensor_gain = getValue(META_TAG_NORMALIZATION, KEY_SENSOR_GAIN).toUInt();

    // See if we need to convert.
    if (sensor_gain == (300 - 256))
        sensor_gain = 300;
    else if (sensor_gain == (400 - 256))
        sensor_gain = 400;

    return sensor_gain;
}

const QString PlaybackMetaInfo::getSensorOpt()
{
    return getValue(META_TAG_IMAGE, KEY_SENSOR_OPTS).toString();
}

uint32_t PlaybackMetaInfo::getSensorType()
{
    uint32_t type_val;

    memcpy(&type_val, getValue(META_TAG_CAMERA, KEY_SENSOR_TYPE).toString().toUtf8().constData(), sizeof(type_val));
    return type_val;
}

uint32_t PlaybackMetaInfo::getStartFrame()
{
    QVariant    rc_value;

    // Is there a start frame in the export section?
    rc_value = getExport(KEY_START_FRAME);

    // Did we get anything?
    if (rc_value.isNull())
        rc_value = getValue(META_TAG_IMAGE, KEY_START_FRAME).toUInt();

    // Return our findings.
    return rc_value.toUInt();
}

const QString PlaybackMetaInfo::getTimestamp()
{
    return getValue(META_TAG_IMAGE, KEY_TIME_STAMP).toString();
}

uint32_t PlaybackMetaInfo::getTimestamp(uint32_t& micro_sec)
{
    return stringToSBS(getValue(META_TAG_IMAGE, KEY_TIME_STAMP).toString(), micro_sec);
}

uint32_t PlaybackMetaInfo::getTriggerFrame()
{
    return getValue(META_TAG_IMAGE, KEY_TRIGGER_FRAME).toUInt();
}

const QVariant PlaybackMetaInfo::getValue(const QString& tag_name, const QString& key_name)
{
    static QVariant rc_value;

    // Reset our return value.
    rc_value.clear();

    // Make sure we have metadata.
    if (captureMetadataIsAvail())
    {
        if (meta_database.contains(tag_name))
        {
            META_DB     tag_db;

            // Retrieve the appropriate entries.
            tag_db = meta_database.value(tag_name).toMap();
            if (tag_db.contains(key_name))
                rc_value = tag_db.value(key_name);
        }
    }

    // Return the value.
    return rc_value;
}

const QString PlaybackMetaInfo::getVersion()
{
    return getValue(META_TAG_CAMERA, KEY_SOFTWARE_VERSION).toString();
}

uint32_t PlaybackMetaInfo::getX()
{
    QVariant    rc_value;

    // Is there a height in the export section?
    rc_value = getExport(KEY_ROI_X);

    // Did we get anything?
    if (rc_value.isNull())
        rc_value = getValue(META_TAG_IMAGE, KEY_ROI_X);

    // Return our value.
    return rc_value.toUInt();
}

uint32_t PlaybackMetaInfo::getY()
{
    QVariant    rc_value;

    // Is there a height in the export section?
    rc_value = getExport(KEY_ROI_Y);

    // Did we get anything?
    if (rc_value.isNull())
        rc_value = getValue(META_TAG_IMAGE, KEY_ROI_Y);

    // Return our value.
    return rc_value.toUInt();
}

uint32_t PlaybackMetaInfo::getWidth()
{
    QVariant    rc_value;

    // Is there a height in the export section?
    rc_value = getExport(KEY_WIDTH);

    // Did we get anything?
    if (rc_value.isNull())
        rc_value = getValue(META_TAG_IMAGE, KEY_WIDTH);

    // Return our value.
    return rc_value.toUInt();
}

bool PlaybackMetaInfo::isColor()
{
    bool rc;
    uint32_t sensor_type = getSensorType();

    switch (sensor_type)
    {
    case FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX160_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR:
    case FI_SENSOR::SENSOR_TYPE_GENERIC_COLOR:
        rc = true;
        break;
    default:
        rc = false;
        break;
    }

    return rc;
}

bool PlaybackMetaInfo::isExported()
{
    // If we don't have capture metadata, then we don't report the PFMD.
    if (!captureMetadataIsAvail())
        return false;

    // Is there an export section?
    return meta_database.contains(META_TAG_EXPORT);
}

bool PlaybackMetaInfo::isRaw()
{
    QString raw_str;

    // Get the value. This is a raw save if the value is '1'.
    raw_str = getValue(META_TAG_NORMALIZATION, KEY_RAW).toString();
    return (raw_str.toInt() == 1);
}

bool PlaybackMetaInfo::isSeries3()
{
    bool series3_found = false;
    QString camera_model;

    // Do we have capture metadata?
    if (!captureMetadataIsAvail())
    {
        // There's no capture metadata to determine model from. Let's
        // see if we have a registered parent. If we're connected, then
        // we can query the parent.
        CameraMDI *my_parent = dynamic_cast<CameraMDI *>(parent());
        if (my_parent)
        {
            return my_parent->is_series3();
        }
    }

    // Get the serial number in a format easier for us to work with.
    camera_model = getModel();
    if (camera_model.isEmpty())
    {
        // Assume a series3.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera model found!?";
        return true;
    }

    // Are we a "TSx" or "ILx" camera?
    // TODO: Replace with a RegEx.
    if (camera_model.contains(QStringLiteral("TS")))
        series3_found = true;
    else if (camera_model.contains(QStringLiteral("IL")))
        series3_found = true;

    // Return our result.
    return series3_found;
}

/*!
 * @brief Determine if the CCM is "normal".
 * @return True if the CCM is "normal".
 */
bool PlaybackMetaInfo::isWbNormal()
{
    bool rc = false;

    // First, are we color?
    if (isColor())
    {
        bool r_norm = false;
        bool g_norm = false;
        bool b_norm = false;

        int32_t     r_val, g_val, b_val;
        QString     matrix_val;
        QStringList matrix_array;
        QVariant    matrix_var;

        // Retrieve the value from the tag area.
        matrix_var = getValue(META_TAG_NORMALIZATION, KEY_RED_MATRIX);
        matrix_array = matrix_var.toStringList();
        if (matrix_array.count() == 3)
        {
            // The first value looks like "[xxx", so strip out the '[' character.
            matrix_val = matrix_array.at(0);
            matrix_val.remove(0, 1);
            r_val = matrix_val.toInt();

            // The second value is just a number.
            g_val = matrix_array.at(1).toInt();

            // The thirs value looks like "xxx]", so strip out the ']' character.
            matrix_val = matrix_array.at(2);
            matrix_val = matrix_val.left(matrix_val.length() - 1);
            b_val = matrix_val.toInt();

            // See if this was a "normal" entry.
            r_norm = ((r_val == 4096) && (g_val == 0) && (b_val == 0));
        }

        matrix_var = getValue(META_TAG_NORMALIZATION, KEY_GREEN_MATRIX);
        matrix_array = matrix_var.toStringList();
        if (matrix_array.count() == 3)
        {
            // The first value looks like "[xxx", so strip out the '[' character.
            matrix_val = matrix_array.at(0);
            matrix_val.remove(0, 1);
            r_val = matrix_val.toInt();

            // The second value is just a number.
            g_val = matrix_array.at(1).toInt();

            // The thirs value looks like "xxx]", so strip out the ']' character.
            matrix_val = matrix_array.at(2);
            matrix_val = matrix_val.left(matrix_val.length() - 1);
            b_val = matrix_val.toInt();

            // See if this was a "normal" entry.
            g_norm = ((r_val == 0) && (g_val == 4096) && (b_val == 0));
        }

        matrix_var = getValue(META_TAG_NORMALIZATION, KEY_BLUE_MATRIX);
        matrix_array = matrix_var.toStringList();
        if (matrix_array.count() == 3)
        {
            // The first value looks like "[xxx", so strip out the '[' character.
            matrix_val = matrix_array.at(0);
            matrix_val.remove(0, 1);
            r_val = matrix_val.toInt();

            // The second value is just a number.
            g_val = matrix_array.at(1).toInt();

            // The thirs value looks like "xxx]", so strip out the ']' character.
            matrix_val = matrix_array.at(2);
            matrix_val = matrix_val.left(matrix_val.length() - 1);
            b_val = matrix_val.toInt();

            // See if this was a "normal" entry.
            b_norm = ((r_val == 0) && (g_val == 0) && (b_val == 4096));
        }

        // We have a "normal" matrix if R, G, and B entries are "normal".
        rc = r_norm && g_norm && b_norm;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Return true if the per-frame metadata file is available, false otherwise.
 */
bool PlaybackMetaInfo::perFrameIsAvail()
{
    // If we don't have capture metadata, then we don't report the PFMD.
    if (!captureMetadataIsAvail())
        return false;

    // If we already know it's available, then just indicate it is so.
    if (per_frame_avail)
        return true;

    // Per-frame metadata could also be available if we have a raw file.
    return (raw_file.isEmpty() == false);
}

uint32_t PlaybackMetaInfo::getBalanceRed()
{
    QVariant    rc_value;

    // Retrieve the value from the tag area.
    rc_value = getValue(META_TAG_NORMALIZATION, KEY_RED_BALANCE);

    // Return our value.
    return rc_value.toUInt();
}

uint32_t PlaybackMetaInfo::getBalanceGreen()
{
    QVariant    rc_value;

    // Retrieve the value from the tag area.
    rc_value = getValue(META_TAG_NORMALIZATION, KEY_GREEN_BALANCE);

    // Return our value.
    return rc_value.toUInt();
}

uint32_t PlaybackMetaInfo::getBalanceBlue()
{
    QVariant rc_value;

    // Retrieve the value from the tag area.
    rc_value = getValue(META_TAG_NORMALIZATION, KEY_BLUE_BALANCE);

    // Return our value.
    return rc_value.toUInt();
}

bool PlaybackMetaInfo::getCCM(double ccm[][4])
{
    bool rc = false;
    int32_t dc_offset;
    QVariant dc_var;
    int32_t val1, val2, val3;
    QString matrix_val;
    QStringList matrix_array;
    QVariant matrix_var;

    // First, are we color?
    if (!isColor())
        goto gc_bail;

    // Retrieve the value from the tag area.
    matrix_var = getValue(META_TAG_NORMALIZATION, KEY_RED_MATRIX);
    matrix_array = matrix_var.toStringList();
    if (matrix_array.count() != 3)
        goto gc_bail;

    // The first value looks like "[xxx", so strip out the '[' character.
    matrix_val = matrix_array.at(0);
    matrix_val.remove(0, 1);
    val1 = matrix_val.toInt();

    // The second value is just a number.
    val2 = matrix_array.at(1).toInt();

    // The thirs value looks like "xxx]", so strip out the ']' character.
    matrix_val = matrix_array.at(2);
    matrix_val = matrix_val.left(matrix_val.length() - 1);
    val3 = matrix_val.toInt();

    // Get the offset.
    dc_var = getValue(META_TAG_NORMALIZATION, KEY_RED_OFFSET);
    if (dc_var.isNull())
        dc_offset = 0;
    else
        dc_offset = dc_var.toInt();

    // Convert to doubles and store.
    ccm[0][0] = ((double)dc_offset / 4096.0);
    ccm[0][1] = ((double)val1) / 4096.0;
    ccm[0][2] = ((double)val2) / 4096.0;
    ccm[0][3] = ((double)val3) / 4096.0;

    matrix_var = getValue(META_TAG_NORMALIZATION, KEY_GREEN_MATRIX);
    matrix_array = matrix_var.toStringList();
    if (matrix_array.count() != 3)
        goto gc_bail;

    // The first value looks like "[xxx", so strip out the '[' character.
    matrix_val = matrix_array.at(0);
    matrix_val.remove(0, 1);
    val1 = matrix_val.toInt();

    // The second value is just a number.
    val2 = matrix_array.at(1).toInt();

    // The thirs value looks like "xxx]", so strip out the ']' character.
    matrix_val = matrix_array.at(2);
    matrix_val = matrix_val.left(matrix_val.length() - 1);
    val3 = matrix_val.toInt();

    // Get the offset.
    dc_var = getValue(META_TAG_NORMALIZATION, KEY_GREEN_OFFSET);
    if (dc_var.isNull())
        dc_offset = 0;
    else
        dc_offset = dc_var.toInt();

    // Convert to doubles and store.
    ccm[1][0] = ((double)dc_offset / 4096.0);
    ccm[1][1] = ((double)val1) / 4096.0;
    ccm[1][2] = ((double)val2) / 4096.0;
    ccm[1][3] = ((double)val3) / 4096.0;

    matrix_var = getValue(META_TAG_NORMALIZATION, KEY_BLUE_MATRIX);
    matrix_array = matrix_var.toStringList();
    if (matrix_array.count() != 3)
        goto gc_bail;

    // The first value looks like "[xxx", so strip out the '[' character.
    matrix_val = matrix_array.at(0);
    matrix_val.remove(0, 1);
    val1 = matrix_val.toInt();

    // The second value is just a number.
    val2 = matrix_array.at(1).toInt();

    // The thirs value looks like "xxx]", so strip out the ']' character.
    matrix_val = matrix_array.at(2);
    matrix_val = matrix_val.left(matrix_val.length() - 1);
    val3 = matrix_val.toInt();

    // Get the offset.
    dc_var = getValue(META_TAG_NORMALIZATION, KEY_BLUE_OFFSET);
    if (dc_var.isNull())
        dc_offset = 0;
    else
        dc_offset = dc_var.toInt();

    // Convert to doubles and store.
    ccm[2][0] = ((double)dc_offset / 4096.0);
    ccm[2][1] = ((double)val1) / 4096.0;
    ccm[2][2] = ((double)val2) / 4096.0;
    ccm[2][3] = ((double)val3) / 4096.0;

    // All is well.
    rc = true;

gc_bail:
    // Return our result.
    return rc;
}

void PlaybackMetaInfo::resetExports()
{
    // Don't do this if there's no capture metadata.
    if (!captureMetadataIsAvail())
        return;

    // Is there an "exports" map?
    if (meta_database.contains(META_TAG_EXPORT))
    {
        // Remove the entry from the database.
        meta_database.remove(META_TAG_EXPORT);
    }
}

bool PlaybackMetaInfo::saveMetadata(const QString& meta_filename)
{
    QStringList tag_list;

    // Make sure we have metadata.
    if (!captureMetadataIsAvail())
        return false;

    // Make sure the file doesn't already exist.
    QFile::remove(meta_filename);

    // Start saving our metadata file.
    QSettings   settings(meta_filename, QSettings::IniFormat);

    // Build our tag list.
    tag_list << META_TAG_IMAGE << META_TAG_CAMERA << META_TAG_RECORD << META_TAG_NORMALIZATION << META_TAG_EXPORT;

    // Work through the list...
    foreach (const QString& tag_name, tag_list)
    {
        // Does our database contain this tag?
        if (meta_database.contains(tag_name))
        {
            META_DB     tag_db;

            // Start this grouping.
            settings.beginGroup(tag_name);

            // Retrieve the appropriate entries.
            tag_db = meta_database.value(tag_name).toMap();

            // Loop through all the values.
            META_DB::const_iterator i = tag_db.constBegin();
            while (i != tag_db.constEnd())
            {
                settings.setValue(i.key(), i.value());
                ++i;
            }

            // Finish this tag.
            settings.endGroup();
        }
    }

    // All went well.
    return true;
}

void PlaybackMetaInfo::setEndFrame(uint32_t new_end)
{
    setValue(META_TAG_EXPORT, KEY_END_FRAME, QVariant::fromValue(new_end));
}

void PlaybackMetaInfo::setFrameCount(uint32_t new_count)
{
    setValue(META_TAG_EXPORT, KEY_FRAME_COUNT, QVariant::fromValue(new_count));
}

void PlaybackMetaInfo::setStartFrame(uint32_t new_start)
{
    setValue(META_TAG_EXPORT, KEY_START_FRAME, QVariant::fromValue(new_start));
}

void PlaybackMetaInfo::setCodec(uint32_t save_type)
{
    switch (save_type)
    {
    case GIGE_STREAM_FORMAT_AVI:
        setValue(META_TAG_EXPORT, KEY_CODEC, QVariant::fromValue(FiSaver::FORMAT_AVI_STR));
        break;
    case GIGE_STREAM_FORMAT_BMP:
        setValue(META_TAG_EXPORT, KEY_CODEC, QVariant::fromValue(FiSaver::FORMAT_BMP_STR));
        break;
    case GIGE_STREAM_FORMAT_JPEG:
        setValue(META_TAG_EXPORT, KEY_CODEC, QVariant::fromValue(FiSaver::FORMAT_JPG_STR));
        break;
    case GIGE_STREAM_FORMAT_TIF:
        setValue(META_TAG_EXPORT, KEY_CODEC, QVariant::fromValue(FiSaver::FORMAT_TIF_STR));
        break;
    default:
        break;
    }
}

void PlaybackMetaInfo::setComment(const QString& new_comment)
{
    // Skip it if the new comment is empty.
    if (!new_comment.isEmpty())
        setValue(META_TAG_EXPORT, KEY_COMMENT, QVariant::fromValue(new_comment));
}

/*!
 * @brief Set the metadata file to read. This causes the file to be read and parsed.
 * @param meta_filename     Fully qualified filename for a metadata file.
 */
void PlaybackMetaInfo::setMetadataFile(const QString& meta_filename)
{
    QStringList tag_list;
    QSettings   settings(meta_filename, QSettings::IniFormat);

    // Reset our database & UI.
    meta_database.clear();
    ui->eb_meta_file->clear();

    // Build our tag list.
    tag_list << META_TAG_IMAGE << META_TAG_CAMERA << META_TAG_RECORD << META_TAG_NORMALIZATION << META_TAG_EXPORT;

    // Work through the list...
    foreach (const QString& tag_name, tag_list)
    {
        META_DB     tag_db;
        QStringList child_keys;

        // Work through the INI file and pull in the data.
        settings.beginGroup(tag_name);
        child_keys = settings.childKeys();
        foreach (const QString& child_key, child_keys)
        {
            // Insert the key and value into our tag database.
            tag_db.insert(child_key, settings.value(child_key));
        }
        settings.endGroup();

        // Now insert the tag database into our meta database, if we have anything.
        if (!tag_db.isEmpty())
            meta_database.insert(tag_name, tag_db);
    }

    // Do we have anything?
    if (captureMetadataIsAvail())
    {
        QFile       meta_file(meta_filename);

        // Open and read the file into our buffer.
        if (meta_file.open(QIODevice::ReadOnly | QIODevice::Text))
        {
            // Read the file into our edit control so we can display it.
            ui->eb_meta_file->append(meta_file.readAll());
        }

        // Close the file.
        meta_file.close();
    }
}

/*!
 * @brief Set the raw per-frame metadata file. If the file doesn't exist, but the XML
 *        file does, then we'll build the raw metadata file from the XML file.
 *
 * @param[in] meta_filename     Fully qualified filename for a raw metadata file.
 * @param[in] pfmd_len          Number of bytes in the per-frame metadata, 0 if not known.
 */
void PlaybackMetaInfo::setRawMetadataFile(const QString& meta_filename, int pfmd_len)
{
    QFile   metafile(meta_filename);

    // If the PFMD length is specified, remember it.
    if (pfmd_len)
        pfmd_len_specified = pfmd_len;

    // Reset our buffer.
    raw_buffer.clear();

    // Does the file exist?
    if (metafile.exists())
    {
        // Save the file name for later use.
        raw_file = meta_filename;
    }
    else
    {
        // Does the XML file exist?
        if (per_frame_avail)
        {
            // Build the raw information from the XML data.
            if (buildRawData(meta_filename, pfmd_len))
            {
                // We've built the raw file successfully. Save the file for later use.
                raw_file = meta_filename;
            }
        }
    }
}

void PlaybackMetaInfo::setValue(const QString& tag_name, const QString& key_name, const QVariant& key_value)
{
    META_DB     tag_db;

    // Don't do this if there's no capture metadata.
    if (!captureMetadataIsAvail())
        return;

    // If the tag already exists, grab it.
    if (meta_database.contains(tag_name))
        tag_db = meta_database.value(tag_name).toMap();

    // Update the entry.
    tag_db.insert(key_name, key_value);

    // Update our tag.
    meta_database.insert(tag_name, tag_db);
}

/*!
 * @brief Set the xml metadata file to read for per frame data. If the file doesn't exist then
 * per frame metadata will not be displayed.. This causes the file to be read and parsed.
 *
 * @param meta_filename     Fully qualified filename for a metadata file.
 */
void PlaybackMetaInfo::setXmlMetadataFile(const QString& meta_filename)
{
    QFile   metafile(meta_filename);
    QString msg_str;

    // open the file and then put it into the DOM
    if (metafile.open(QIODevice::ReadOnly))
    {
        if (!doc.setContent(&metafile))
        {
            msg_str = QStringLiteral("Error parsing per frame metadata xml file %3 code:%4 (%5)")
                      .arg(metafile.fileName())
                      .arg(errno).arg(QString::fromUtf8(strerror(errno)));
            QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        }
        else
        {
            // okay we sucessfully parse the xml per frame metadata file
            per_frame_avail = true;
        }

        // Close the file.
        metafile.close();
    }
}

/*!
 * @brief   Convert the XML time string ("YY:DDD:HH:MM:SS.microseconds") into a
 *          straight-binary-seconds value and a binary microseconds value.
 *
 * @param   time_str    The XML time string.
 * @param   micro_sec   The output value for the binary microseconds.
 *
 * @return  The straight-binary-seconds value.
 */
uint32_t PlaybackMetaInfo::stringToSBS(const QString& time_str, uint32_t& micro_sec)
{
    uint32_t    sbs;
    QStringList time_list = time_str.split(QStringLiteral(":"));     // Get the YY:DDD:HH:MM:SS.microsec of the time.
    QStringList micro_str;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "time_str:" << time_str;

    // Make sure we got everything.
    if (time_list.size() != 5)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "split failed, only produced" << time_list;
        return 0;
    }

    // The last time_list element contains "<SEC>.<MICROSEC>", so split that out.
    micro_str = time_list.at(4).split(QStringLiteral("."));

    // Convert the rest to straight-binary-seconds.
    sbs = time_list.at(0).toUInt() * (365 * 24 * 60 * 60);      // Year.
    sbs += (time_list.at(1).toUInt() - 1) * (24 * 60 * 60);     // Day - don't forget to subtract 1 so we are 0 based.
    sbs += time_list.at(2).toUInt() * (60 * 60);                // Hour.
    sbs += time_list.at(3).toUInt() * 60;                       // Minute.
    sbs += micro_str.at(0).toUInt();                            // Second.

    // Convert the microseconds to a value.
    micro_sec = micro_str.at(1).toUInt();

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << QString("sbs: %1 micro_sec: %2").arg(sbs, 8, 16, QChar('0')).arg(micro_sec, 8, 16, QChar('0'));

    // Return our seconds.
    return sbs;
}

