/*!
 * @file playback_meta_info.h
 * @brief This is the header file for displaying meta file information about the stored video.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef PLAYBACK_META_INFO_H
#define PLAYBACK_META_INFO_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>
#include <QDomDocument>
#include <QMap>
#include <QString>
#include <QVariant>
#include <QWidget>

#include "gige_msg.h"

namespace Ui
{
class PlaybackMetaInfo;
}

//--------------------------------------------------------------------
// Namespace: FI_COMM
//--------------------------------------------------------------------
namespace PLAYBACK_META_INFO
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

typedef QMap< QString, QVariant >  META_DB;        //!< A map of setting [label, value] pairs we've read/written.

}   // End namespace (PLAYBACK_META_INFO)


class PlaybackMetaInfo : public QWidget
{
    Q_OBJECT

public:
    explicit PlaybackMetaInfo(QWidget *parent = nullptr);
    virtual ~PlaybackMetaInfo();

    void setRawMetadataFile(const QString& meta_filename, int pfmd_len = GIGE_FRAME_METADATA_LEN);
    void setXmlMetadataFile(const QString& meta_filename);
    void displayPerFrameMetadata(uint32_t frame_num);

    //------------------------------------------------------------------------
    // Create a virtual interface to allow for customization of the following:
    //------------------------------------------------------------------------
    virtual uint32_t getStartFrame();
    virtual uint32_t getEndFrame();
    virtual uint32_t getTriggerFrame();
    virtual uint32_t getFrameCount();

    virtual uint8_t *getRawEventFlags(int pfmd_len = 0, bool do_swap = true);
    virtual QByteArray *getRawPFMD(uint32_t frame_num, bool do_swap = true);

    virtual uint32_t getPfmdTime(uint32_t frame_num, uint32_t& micro_sec, int pfmd_len = 0);
    virtual bool getPfmdTimeStr(uint32_t frame_num, QString& time_str, int pfmd_len = 0);

    virtual uint32_t findFrame(uint32_t time_sec, uint32_t time_usec, int pfmd_len = 0);
    virtual void setMetadataFile(const QString& meta_filename);

    //------------------------------------------------------------------------
    // Back to normal.
    //------------------------------------------------------------------------
    bool captureMetadataIsAvail()           { return !meta_database.isEmpty(); }
    bool perFrameIsAvail();

    bool isColor();
    bool isExported();
    bool isRaw();
    bool isSeries3();
    bool isWbNormal();

    uint32_t        getBalanceRed();
    uint32_t        getBalanceGreen();
    uint32_t        getBalanceBlue();
    bool            getCCM(double ccm[][4]);

    const QString   getBitMode();
    uint8_t         getConvertedBitMode();
    const QString   getCameraName();
    const QString   getComment();
    uint32_t        getExposure();
    uint32_t        getFramerate();
    uint32_t        getHeight();
    const QString   getMacAddress();
    const QString   getManufacturer();
    const QString   getModel();
    const pm_metadata *getPmMetadata() { return pm_ptr.get(); }
    uint32_t        getSensorGain();
    const QString   getSensorOpt();
    uint32_t        getSensorType();
    const QString   getTimestamp();
    uint32_t        getTimestamp(uint32_t& micro_sec);
    const QString   getVersion();
    uint32_t        getX();
    uint32_t        getY();
    uint32_t        getWidth();

    const QByteArray *getRawBuffer();
    const QString& getRawfile() const;

    uint32_t findFrame(const QString& time_str);
    int pfmdLength();

    void    ignoreExport(bool do_ignore)         { use_export = !do_ignore; }
    void    recoverPmMetadata();
    void    resetExports();
    bool    saveMetadata(const QString& meta_filename);

    void    setEndFrame(uint32_t new_end);
    void    setFrameCount(uint32_t new_count);
    void    setStartFrame(uint32_t new_start);
    void    setCodec(uint32_t save_type);
    void    setComment(const QString& new_comment);

private:
    const QVariant  getExport(const QString& key_name);
    const QVariant  getValue(const QString& tag_name, const QString& key_name);
    void            setValue(const QString& tag_name, const QString& key_name, const QVariant& key_value);

    uint32_t convertSensorOpts();

    bool buildRawData(const QString& raw_file, int pfmd_len);
    uint32_t stringToSBS(const QString& time_str, uint32_t& micro_sec);

protected:
    PLAYBACK_META_INFO::META_DB meta_database;  //!< Our metadata information database.
    Ui::PlaybackMetaInfo    *ui;

    // Define the tags used in the metadata TXT file.
    static const QString   META_TAG_IMAGE;
    static const QString   META_TAG_CAMERA;
    static const QString   META_TAG_RECORD;
    static const QString   META_TAG_NORMALIZATION;
    static const QString   META_TAG_EXPORT;

    // Define the keys used in the metadata TXT file.
    static const QString   KEY_BIT_MODE;
    static const QString   KEY_BLUE_BALANCE;
    static const QString   KEY_BLUE_GAIN;
    static const QString   KEY_BLUE_MATRIX;
    static const QString   KEY_BLUE_OFFSET;
    static const QString   KEY_BRIGHTNESS;
    static const QString   KEY_CAMERA_NAME;
    static const QString   KEY_CODEC;
    static const QString   KEY_COMMENT;
    static const QString   KEY_CONTRAST;
    static const QString   KEY_END_FRAME;
    static const QString   KEY_FPS;
    static const QString   KEY_FPGA_REV;
    static const QString   KEY_FRAME_COUNT;
    static const QString   KEY_GAMMA;
    static const QString   KEY_GREEN_BALANCE;
    static const QString   KEY_GREEN_GAIN;
    static const QString   KEY_GREEN_MATRIX;
    static const QString   KEY_GREEN_OFFSET;
    static const QString   KEY_HEIGHT;
    static const QString   KEY_MAC_ADDRESS;
    static const QString   KEY_MAKE;
    static const QString   KEY_MODEL;
    static const QString   KEY_MULTI_SLOPE;
    static const QString   KEY_RAW;
    static const QString   KEY_RED_BALANCE;
    static const QString   KEY_RED_GAIN;
    static const QString   KEY_RED_MATRIX;
    static const QString   KEY_RED_OFFSET;
    static const QString   KEY_ROI_X;
    static const QString   KEY_ROI_Y;
    static const QString   KEY_SENSOR_GAIN;
    static const QString   KEY_SENSOR_OPTS;
    static const QString   KEY_SENSOR_TYPE;
    static const QString   KEY_SHUTTER_SPEED;
    static const QString   KEY_SOFTWARE_VERSION;
    static const QString   KEY_START_FRAME;
    static const QString   KEY_SYNC_IN;
    static const QString   KEY_SYNC_OUT;
    static const QString   KEY_TIME_STAMP;
    static const QString   KEY_TRIGGER_FRAME;
    static const QString   KEY_TRIGGER_SETTING;
    static const QString   KEY_WIDTH;

private:
    QDomDocument    doc;                //!< The XML file parsed inside the DOM.
    QString         raw_file;           //!< The raw per-frame metadata.
    QByteArray      raw_buffer;         //!< The raw per-frame metadata loaded into a buffer.
    bool            per_frame_avail;    //!< Flag indicating if the per frame metadata is available.
    bool            use_export;         //!< True to use the exports section, if it exists.
    int             pfmd_len_specified; //!< The specified length used when raw metadata is specified.
    std::unique_ptr<pm_metadata> pm_ptr;    //!< The partition metadata structure.
};

#endif // PLAYBACK_META_INFO_HC
