/*!
 * @file playback_meta_info_fflr.cpp
 *
 * @brief This is the header file for displaying meta file information about a CAP
 *        stack of files created with FasFireLR.
 *
 * Copyright 2022 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QFile>
#include <QSettings>
#include <QSharedPointer>
#include <QtEndian>

#include "ui_playback_meta_info.h"
#include "playback_meta_info_fflr.h"

#include "fi_metamap.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_meta_info.h"

//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------
#define LOG_PFMD_INFO       0

using namespace FI_METAMAP;

/*!
 * @brief Constructor for the PlaybackMetaInfoFFLR object.
 *
 * @param parent    Window who is our parent.
 */
PlaybackMetaInfoFFLR::PlaybackMetaInfoFFLR(QWidget *parent) :
    PlaybackMetaInfo(parent)
{
}

/*!
 * @brief Destructor for the PlaybackMetaInfoFFLR object.
 */
PlaybackMetaInfoFFLR::~PlaybackMetaInfoFFLR()
{
}

/*!
 * @brief Assigns the FasFireLR top level metadata (metamap) information to our object.
 *
 * @param fflrMMPtrToSet    Pointer to the top level metadata map object to use.
 * @return True if everything went well, false otherwise.
 */
bool PlaybackMetaInfoFFLR::installMetaMap(QSharedPointer<FI_METAMAP::FiMetaMap> fflrMMPtrToSet)
{
    bool isInstalled = false;
    uint32_t activeAbsFrame;    // The absolute frame number for the first display.
    uint32_t activeCapId;       // The active capId for first playable frame.
    QString metadataTXTName;
    QString pfmdFileName;

    // Were we passed a valid metamap object?
    if (fflrMMPtrToSet.isNull())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The metamap.txt object was null!";
        goto installMetaMapExit;
    }

    // The FasFireLR meta map object contains information about the FasFireLR CAP stack.
    // It is used to provide information for playback across CAP file spans in the CAP stack.
    fflrMMPtr = fflrMMPtrToSet;

    // Get our first playable frame in this cap stack.
    if (!fflrMMPtr->GetFirstPlayableFrame(activeAbsFrame, activeCapId))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to get the first playable frame information!";
        goto installMetaMapExit;
    }

    // Using the metamap object get the equivalent path to the starting cap's metadata.txt file
    // and install it as our internal metadata database.
    if (fflrMMPtr->GetMetadataTXTName(activeCapId, metadataTXTName))
    {
        // Install this metadata database (this is the TAG database) into our object.
        setMetadataFile(metadataTXTName);
    }

    // There should already be a .pm file for us that houses the pfmd for this cap file.
    if (fflrMMPtr->GetPFMDName(activeCapId, pfmdFileName))
    {
        // Install the pfmd file information into our object.
        setRawMetadataFile(pfmdFileName, pfmdLength());
    }

    // The fflrMMPtr is installed and ready for use in this object.
    isInstalled = true;

installMetaMapExit:
    return isInstalled;
}

/*!
 * @brief Returns the absolute starting frame number for the FFLR stack.
 */
uint32_t PlaybackMetaInfoFFLR::getStartFrame()
{
    uint32_t sFrame = FASMOTION_SETTINGS::UNSET_UINT32;
    uint32_t capId;

    // Pull out the starting frame from the metamap.
    if (!fflrMMPtr.isNull())
        fflrMMPtr->GetFirstFrameInfo(capId, sFrame);

    // Were we successful?
    if (sFrame == FASMOTION_SETTINGS::UNSET_UINT32)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to determine starting frame.";
        sFrame = 0;
    }

    // Return our starting frame number.
    return sFrame;
}

/*!
 * @brief Returns the ending frame for the FFLR stack.
 */
uint32_t PlaybackMetaInfoFFLR::getEndFrame()
{
    uint32_t eFrame = FASMOTION_SETTINGS::UNSET_UINT32;
    uint32_t capId;

    // Pull out the starting frame from the metamap.
    if (!fflrMMPtr.isNull())
        fflrMMPtr->GetLastFrameInfo(capId, eFrame);

    // Were we successful?
    if (eFrame == FASMOTION_SETTINGS::UNSET_UINT32)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to determine ending frame.";
        eFrame = 0;
    }

    // Return our ending frame number.
    return eFrame;
}

/*!
 * @brief Returns the number of frames in the FFLR stack.
 */
uint32_t PlaybackMetaInfoFFLR::getFrameCount()
{
    uint32_t totalFrameCount = FASMOTION_SETTINGS::UNSET_UINT32;

    // Pull out the number of frames in the stack.
    if (!fflrMMPtr.isNull())
        fflrMMPtr->GetTotalFrameCount(totalFrameCount);

    // Were we successful?
    if (totalFrameCount == FASMOTION_SETTINGS::UNSET_UINT32)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to determine frame count.";
        totalFrameCount = 0;
    }

    // Return our total frame count.
    return totalFrameCount;
}

/*!
 * @brief Returns the trigger frame for the cap stack.
 */
uint32_t PlaybackMetaInfoFFLR::getTriggerFrame()
{
    uint32_t tFrame = -1U;
    uint32_t capId;

    // Pull out the starting frame from the metamap.
    if (!fflrMMPtr.isNull())
        fflrMMPtr->GetTriggerInfo(capId, tFrame);

    // Were we successful?
    if (tFrame == FASMOTION_SETTINGS::UNSET_UINT32)
    {
        // It may not have an official trigger frame so use the start frame instead.
        tFrame = getStartFrame();
    }

    // Return our trigger frame.
    return tFrame;
}

/*!
 * @brief Read the raw byte array for the per-frame metadata associated with the specified frame.
 *
 * @param absFrameNum   The CAP stack's absolute frame number to locate.
 * @param do_swap       Are the first 4 bytes of metadata supposed to be swapped?
 *
 * @return Pointer the the byte array, or nullptr if an error.
 */
QByteArray *PlaybackMetaInfoFFLR::getRawPFMD(uint32_t absFrameNum, bool do_swap)
{
    int pfmd_len = 0;           // The size of the per frame metadata data length is calculated below.
    uint32_t capId;
    uint32_t pfmdOffset;        // The file offset in the metadata.pm file for the absFrameNum.
    QString pfmdFileName;
    QByteArray *pfmd_array = nullptr;

    // Sanity check.
    if (absFrameNum >= getFrameCount())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "absFrameNum:" << absFrameNum << "is bigger than count:" << getFrameCount();
    }

    // How many bytes is our per-frame metadata?
    pfmd_len = pfmdLength();

    // Ask the metamap object for the pfmd offset belonging to this frame.
    if (!fflrMMPtr.isNull() && fflrMMPtr->GetPFMDOffset(absFrameNum, capId, pfmdFileName, pfmdOffset))
    {
        // Open our file.
        QVector<uint32_t> raw_buffer((uint32_t)pfmd_len / sizeof(uint32_t));
        QFile raw_qfile(pfmdFileName);

        // We are almost certain that the filename is valid, but check the return result
        // just to be safe.
        if (!raw_qfile.open(QIODevice::ReadOnly))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to open:" << pfmdFileName << ", err:" << raw_qfile.errorString();
        }
        else
        {
            // Calculate our frame position within the raw .pm file.
            // The pfmdOffset is a zero-based index for the frame inside of the .pm file.
            qint64 frame_pos = (qint64)(pfmdOffset) * pfmd_len;

#if LOG_PFMD_INFO
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "absFrameNum:" << absFrameNum << "capId:" << capId
                         << "pfmdFileName:" << pfmdFileName << "pfmdOffset:" << pfmdOffset
                         << "frame_pos:" << frame_pos
                         << "pfmd_len:" << pfmd_len;
#endif
            // Seek to the right spot in the file.
            raw_qfile.seek(frame_pos);

            // Read the frame's metadata.
            if (raw_qfile.read((char *)raw_buffer.data(), pfmd_len) < 0)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error reading the file:" << raw_qfile.errorString();
            }
            else
            {
                // Are we supposed to swap?
                if (do_swap)
                {
                    // Swap the words in our raw data buffer.
                    raw_buffer[0] = qToBigEndian(raw_buffer[0]);
                    raw_buffer[1] = qToBigEndian(raw_buffer[1]);
                    raw_buffer[2] = qToBigEndian(raw_buffer[2]);
                    raw_buffer[3] = qToBigEndian(raw_buffer[3]);
                }

                // Allocate our QByteArray.
                pfmd_array = new QByteArray((const char *)raw_buffer.constData(), pfmd_len);
            }

            // Close the file.
            raw_qfile.close();
        }
    }

    // Return our PFMD pointer.
    return pfmd_array;
}

/*!
 * @brief Find the specified time for a frame in our data.
 *
 * @param[in] absFrameNum   The absolute frame number in the CAP stack to locate.
 * @param[out] time_usec    The place to store the number of microseconds for the frame.
 * @param[opt-in] pfmd_len  Number of bytes in the per-frame metadata, 0 if not known.
 *
 * @return The seconds part of the time value for the specified frame.
 */
uint32_t PlaybackMetaInfoFFLR::getPfmdTime(uint32_t absFrameNum, uint32_t& time_usec, int pfmd_len)
{
    QString pfmdFileName;
    uint32_t capId;
    uint32_t frameIndexInCap;
    uint32_t sbs;

#if LOG_PFMD_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "<<ENTRY>> absFrameNum:" << absFrameNum;
#endif

    // Initialize our return values.
    sbs = time_usec = 0;

    // Make sure we have the metamap data.
    if (fflrMMPtr.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected null FasFireLR metamap object!";
        return sbs;
    }

    // Locate the frame in our stack.
    if (!fflrMMPtr->GetCapInfoForFrame(absFrameNum, capId, frameIndexInCap))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to locate absFrameNum:" << absFrameNum << "in CAP stack!";
        return sbs;
    }

    // Get the filename associated with the metadata for the particular capId.
    if (fflrMMPtr->GetPFMDName(capId, pfmdFileName))
    {
        // Install the PFMD file information into our object.
        setRawMetadataFile(pfmdFileName, pfmdLength());

        // We only need to call the base class' implemetation.
        // IMPORTANT: Pass the CAP stack's absolute frame number here.
        // NOTE: This ends up calling PlaybackMetaInfoFFLR::getPFMD because it will be called
        // from this base-class function and we have virtualized getPFMD.
        sbs = PlaybackMetaInfo::getPfmdTime(absFrameNum, time_usec, pfmd_len);
    }

#if LOG_PFMD_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "<<EXIT>>capId:" << capId << "absFrameNum:" << absFrameNum << "frameIndexInCap:"
                 << frameIndexInCap << "pfmdFileName:" << pfmdFileName << "sbs:" << sbs << "time_usec:" << time_usec;
#endif

    // Return the time.
    return sbs;
}

/*!
 * @brief Get the timestamp associated with the "frame_num" parameter from the per-frame
 *        metadata.
 *
 * @param[in] absFrameNum   The absolute frame number in the CAP stack to locate.
 * @param[out] time_str     The timestamp associated with the frame number.
 * @param[opt-in] pfmd_len  [OPTIONAL] Number of bytes in the per-frame metadata, 0 if not known.
 *
 * @return True if we were successful, false on error.
 */
bool PlaybackMetaInfoFFLR::getPfmdTimeStr(uint32_t absFrameNum, QString& time_str, int pfmd_len)
{
    bool rc = false;
    uint32_t capId;
    uint32_t frameNumInCapFile;
    QString pfmdFileName;

    // Initialize our return value.
    time_str.clear();

    // We just want to turn this frame_num which will be the absolute frame number in the entire video
    // into the cap frame number so that the processing occurs properly for this request.
    if (fflrMMPtr.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected null FasFireLR metamap object!";
        return rc;
    }

    // Locate the frame in our stack.
    if (!fflrMMPtr->GetCapInfoForFrame(absFrameNum, capId, frameNumInCapFile))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to locate absFrameNum:" << absFrameNum << "in CAP stack!";
        return rc;
    }

    // Get the filename associated with the metadata for the particular capId.
    if (fflrMMPtr->GetPFMDName(capId, pfmdFileName))
    {
        // Install the pfmd file information into our object.
        setRawMetadataFile(pfmdFileName, pfmdLength());

        // We only need to call the base class' implemetation.
        // IMPORTANT: Pass the CAP stack's absolute frame number here.
        // NOTE: This ends up calling PlaybackMetaInfoFFLR::getPFMD because it will be called
        // from this base-class function and we have virtualized getPFMD.
        rc = PlaybackMetaInfo::getPfmdTimeStr(absFrameNum, time_str, pfmd_len);
    }

    // Return our status.
    return rc;
}

/*!
 * @brief Find the frame at the specified time.
 *
 * @param[in] time_sec      The number of seconds for the frame.
 * @param[in] time_usec     The number of microseconds for the frame.
 * @param[opt-in] pfmd_len  [OPTIONAL] Number of bytes in the per-frame metadata, 0 if not known.
 *
 * @return The frame number found, or 0 if there was some error.
 */
uint32_t PlaybackMetaInfoFFLR::findFrame(uint32_t time_sec, uint32_t time_usec, int pfmd_len)
{
    uint32_t absFrameNum = 0;
    uint32_t capId;

    Q_UNUSED(pfmd_len)

    // Ask the metamap to find the cap file id that holds this record frame based on time.
    if (!fflrMMPtr.isNull() && fflrMMPtr->GetAbsFrameUsingTime(time_sec, time_usec, capId, absFrameNum))
    {
        // Found the frame...
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Found Frame:" << absFrameNum << "in capId:" << capId << "Matching time_sec:" << time_sec << "time_usec:" << time_usec;
    }

    return absFrameNum;
}

/*!
 * @brief Read the raw metadata file and build up an array of the event flags.
 *
 * @param[opt-in] pfmd_len  Number of bytes in the per-frame metadata, 0 if not known.
 *
 * @return The array of event flags, or nullptr on error. It is the responsibility of the
 *         caller to release the array memory.
 */
uint8_t *PlaybackMetaInfoFFLR::getRawEventFlags(int pfmd_len, bool do_swap)
{
    uint8_t *event_flag_ptr = nullptr;

    Q_UNUSED(pfmd_len)

    // Make sure we have a raw file.
    if (getRawfile().isEmpty())
        return nullptr;

    // Sanity check.
    if (getFrameCount() == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No frames!";
        goto get_raw_event_flags_exit;
    }

    // If there isn't a metamap object something is wrong.
    if (fflrMMPtr.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No Metamap object exists!";
        goto get_raw_event_flags_exit;
    }

    // Ask the metamap to pull all of the event information for all of its cap files...
    event_flag_ptr = fflrMMPtr->GetPFMDEvents(do_swap);

    // If it was not allocated output a message.
    if (event_flag_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to read PFMD events!";
        goto get_raw_event_flags_exit;
    }

get_raw_event_flags_exit:
    // Return our array pointer.
    return event_flag_ptr;
}

/*!
 * @brief Set the metadata file to read. This causes the file to be read and parsed.
 * @param meta_filename Fully qualified filename for a metadata file.
 */
void PlaybackMetaInfoFFLR::setMetadataFile(const QString& meta_filename)
{
    QStringList tag_list;
    QSettings   settings(meta_filename, QSettings::IniFormat);

    // Reset our database & UI.
    meta_database.clear();
    ui->eb_meta_file->clear();

    // Build our tag list.
    tag_list << META_TAG_IMAGE << META_TAG_CAMERA << META_TAG_RECORD << META_TAG_NORMALIZATION << META_TAG_EXPORT;

    // Work through the list...
    foreach (const QString& tag_name, tag_list)
    {
        PLAYBACK_META_INFO::META_DB tag_db;
        QStringList child_keys;

        // Work through the INI file and pull in the data.
        settings.beginGroup(tag_name);
        child_keys = settings.childKeys();
        foreach (const QString& child_key, child_keys)
        {
            bool overrodeKey = false;

            // Override the key value if we are a cap stack.
            if (!fflrMMPtr.isNull())
            {
                uint32_t capId;
                uint32_t overrideValue = 0;

                // Override the start frame key.
                if (child_key.compare(PlaybackMetaInfo::KEY_START_FRAME) == 0)
                {
                    if (fflrMMPtr->GetFirstFrameInfo(capId, overrideValue))
                    {
                        overrodeKey = true;
                        tag_db.insert(child_key, overrideValue);
                    }
                }
                else if (child_key.compare(PlaybackMetaInfo::KEY_END_FRAME) == 0)
                {
                    if (fflrMMPtr->GetLastFrameInfo(capId, overrideValue))
                    {
                        overrodeKey = true;
                        tag_db.insert(child_key, overrideValue);
                    }
                }
                else if (child_key.compare(PlaybackMetaInfo::KEY_FRAME_COUNT) == 0)
                {
                    if (fflrMMPtr->GetTotalFrameCount(overrideValue))
                    {
                        overrodeKey = true;
                        tag_db.insert(child_key, overrideValue);
                    }
                }
                else if (child_key.compare(PlaybackMetaInfo::KEY_TRIGGER_FRAME) == 0)
                {
                    if (fflrMMPtr->GetTriggerInfo(capId, overrideValue))
                    {
                        overrodeKey = true;
                        tag_db.insert(child_key, overrideValue);
                    }
                }
                // Go ahead and use the first KEY_TIME_STAMP as the database setting.
            }
            if (!overrodeKey)
            {
                // Insert the key and value into our tag database.
                tag_db.insert(child_key, settings.value(child_key));
            }
        }
        settings.endGroup();

        // Now insert the tag database into our meta database, if we have anything.
        if (!tag_db.isEmpty())
            meta_database.insert(tag_name, tag_db);
    }

    // Do we have anything?
    if (captureMetadataIsAvail())
    {
        QFile       meta_file(meta_filename);

        // Open and read the file into our buffer.
        if (meta_file.open(QIODevice::ReadOnly | QIODevice::Text))
        {
            // Read the file into our edit control so we can display it.
            ui->eb_meta_file->append(meta_file.readAll());
        }

        // Close the file.
        meta_file.close();
    }
}
