/*!
 * @file playback_raw_mdi.cpp
 * @brief This file contains the implementation of the PlaybackRawMDI
 *        class. This is the child window class which allows the user
 *        to playback a captured video from a Fastec formatted drive.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "playback_raw_mdi.h"

#include <QApplication>
#include <QAction>
#include <QContextMenuEvent>
#include <QDockWidget>
#include <QElapsedTimer>
#include <QGraphicsScene>
#include <QImageWriter>
#include <QMenu>
#include <QMessageBox>
#include <QPixmap>
#include <QPixmapCache>
#include <QtConcurrent>
#include <QThread>
#include <QTimer>
#include <QTimerEvent>

#include "camera_info.h"
#include "camera_review.h"
#include "camera_video.h"
#include "fi_pixmap.h"
#include "fi_rawdrive.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"
#include "playback_meta_info.h"
#include "save_dialog.h"
#include "save_video.h"

namespace PLAYBACK_RAW_MDI_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Rate at which we poll the saver object's progress.
const int   SAVER_POLL_RATE             = 500;

const QString MISSING_RAWDRIVE  = QStringLiteral("The rawDrive has not been created properly!");

}   // End of namespace PLAYBACK_RAW_MDI_SETTINGS.

using namespace PLAYBACK_RAW_MDI_SETTINGS;

PlaybackRawMDI::PlaybackRawMDI(QWidget *parent) :
    PlaybackMDI(parent)
{
}

PlaybackRawMDI::~PlaybackRawMDI()
{
    // Do we need to close the raw device?
    if (!rawDrive.isNull())
    {
        rawDrive->close();
        rawDrive.reset();
    }
}

const QByteArray& PlaybackRawMDI::getRawImage() const
{
    static QByteArray null_array;

    // Sanity check.
    if (rawDrive.isNull())
        return null_array;

    return rawDrive->getRawImage();
}

/*!
 * @brief Extract the various meta data from the drive and save it to files for use by
 *        the application.
 *
 * @return True on success, false otherwise.
 */
bool PlaybackRawMDI::initMetadata()
{
    bool rc(false);
    QString file_name;
    QString tm_label;

    // This can take awhile if there's lots of frames.
    // Create the progress dialog if we don't have one.
    if (!progressDialog)
    {
        progressDialog = fiProgressDialog::CreateSimpleProgressDialog(camera_info->get_camera_name(), tr("Accumulating Metadata..."), QMessageBox::NoButton, __PRETTY_FUNCTION__);
    }

    // Display the progress dialog.
    if (progressDialog)
    {
        progressDialog->setBarMinMax(0, 0);
        progressDialog->showBar(true);
        progressDialog->showDialogWithoutPolling();

        // Give the progress bar a chance to start.
        QApplication::processEvents();
    }

    // Allocate our Playback Metadata Information pointer.
    if (pbMetaInfoPtr == nullptr)
    {
        pbMetaInfoPtr = new PlaybackMetaInfo(this);
        if (pbMetaInfoPtr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a PlaybackMetaInfo object!";
            goto im_bail;
        }
    }

    // Build up the file name for the capture metadata.
    tm_label = QDateTime::currentDateTime().time().toString(QStringLiteral("hhmmss"));
    file_name = QDir::tempPath() + QString::fromUtf8("/metadata_%1.txt").arg(tm_label);

    // Save the capture metadata and allocates the pmMetaDataInfo object.
    if (!savePmMetadata(file_name))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to save the pm_metadata to:" << file_name;
        goto im_bail;
    }

    // Pass the info to our metadata class.
    pbMetaInfoPtr->setMetadataFile(file_name);

    // Build up the file name for the raw per-frame metadata.
    file_name = QDir::tempPath() + QString::fromUtf8("/metadata_%1.pm").arg(tm_label);

    // Save the per-frame metadata.
    if (!saveRawPerFrameMetadata(file_name))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to save the pm_metadata to:" << file_name;
        goto im_bail;
    }

    // Pass the info to our metadata class.
    pbMetaInfoPtr->setRawMetadataFile(file_name, getPfmdLen());

    // Everything is fine.
    rc = true;

im_bail:
    // Recycle any resources and release the progress dialog.
    fiProgressDialog::RecycleProgressDialog(progressDialog);

    // All done.
    return rc;
}

/*!
 * @brief We need to load a frame from the drive to be displayed by the video_player object.
 *
 * @param frame_num     The absolute frame number of the frame we want displayed.
 * @param pixmap_ptr    Pixmap to load the frame into.
 * @param force_load    True to clear the cache so the frame is reloaded.
 *
 * @return True on success, false otherwise.
 */
bool PlaybackRawMDI::loadFrame(uint32_t frame_num, fiPixmap *pixmap_ptr, bool force_load)
{
    bool        do_fpn;

    // Sanity check.
    if (pixmap_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "pixmap_ptr is nullptr!";
        return false;
    }
    if (rawDrive.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MISSING_RAWDRIVE;
        return false;
    }
    if (pmMetaDataInfo.isNull())
    {
        // Create the capture pm_metadata object.
        pmMetaDataInfo->CreateCaptureMetaInfo(rawDrive);
    }

    // Is our pmMetaDataInfo valid to use?
    if (pmMetaDataInfo.isNull() || (pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA;
        return false;
    }

    // Ignore if it's loading the same frame again, and we're not forcing it.
    if ((frame_num == last_frame_loaded) && !force_load)
        return false;

    // Do we need to do FPN?
    do_fpn = ((pmMetaDataInfo->GetMetaInfo()->fpn_status & GIGE_FPN_FPGA_ENABLE_MASK) == GIGE_FPN_FPGA_ENABLE_MASK);

    // With color processing we need to install information in the saver frame for Fastec CCM Processing.
    // NOTE:  The parameter "do_fpn" is set to the CAP file's metadata value for FPN correction as a result
    // of this call.  If set in the CAP file's metadata field fpn_status by looking at the GIGE_FPN_FPGA_ENABLE_MASK
    // mask then that fixed pattern noise setting is maintained while loading the frame.
    if (!init_ccm_ctldata(pixmap_ptr, do_fpn))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "init_ccm_ctldata failed!";
    }

    // If this is a force load, then clear the pixmap cache.
    if (force_load)
    {
        // Clear the cache.
        QPixmapCache::clear();
    }

    // Remember this frame.
    last_frame_loaded = frame_num;

    // Read the correct frame from the drive.
    if (!rawDrive->readFrame(frame_num, do_fpn, pixmap_ptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to load the image. Frame_num: " << frame_num;
        return false;
    }

    // All done.
    return true;
}

/*!
 * @brief Save the images as specified by the SaveDialog object.
 * @param parent    Pointer to the [parent](@ref PlaybackRawMDI) object.
 */
void PlaybackRawMDI::SaveImages(QObject *parent)
{
    bool        force8BitMono  = false;     // Set if the frame read needs to be forced 8-bit instead of 16-bit.
    bool        usingBMPFrames = false;     // Can be creating BMP for stacks or AVI containers.
    QString     save_frame_err = tr("Saving error has occurred ");
    FAS_VIDEO::VIDEO_CODEC force8BitMonoCodec   = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;
    FAS_VIDEO::VIDEO_CODEC rawFrameCodec        = FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC;
    QByteArray  metadata_array;
    pm_metadata metadata_copy;
    quint64     byte_count_stat;
    qint64      elapsed;
    QTime       elapsed_time(0, 0, 0, 0);
    double      bytes_per_sec;
    double      mbytes_per_sec;
    QString     stats_msg;
    FI_SAVER::SAVE_PARAMS   save_settings;
    QElapsedTimer   save_timer;
    PlaybackRawMDI  *my_class = qobject_cast<PlaybackRawMDI *>(parent);

    bool saveStatusOK = true;       // If fiSaver object has error catch the state.

    // Sanity checks.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }
    if (my_class->pmMetaDataInfo.isNull() || (my_class->pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA;
        return;
    }

    // Get a pointer to the save dialog. Makes things more legible later on.
    std::shared_ptr<SaveDialog> saveDialog = my_class->save_dlg;
    if (!saveDialog)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_SAVE_DIALOG;
        return;
    }

    // Get our settings.
    saveDialog->get_save_params(save_settings, my_class->isSeries3());

    // If creating an MP4 video file special processing is needed to prepare for
    // the save dialog configuration.
    if (save_settings.streamType == FI_SAVER::STREAM_TYPE::MP4)
    {
        // Set the input to TIF and the output to BMP for CAP->MP4 video creation.
        save_settings.imagesInputType = FI_SAVER::IMAGE_TYPE::TIF;
        save_settings.imagesOutputType = FI_SAVER::IMAGE_TYPE::BMP;
    }

    // Allocate our saver object.
    my_class->saver = std::unique_ptr<FiSaver>(new FiSaver());

    // Make sure we have a saver to move forward with.
    if (!my_class->saver)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate saver object!";
        return;
    }

    // Initialize the saver object.
    my_class->saver->configure(save_settings);
    my_class->saver->setPath(save_settings.dest_path);
    my_class->saver->setFileExtension(save_settings.streamType);
    my_class->saver->setStreamType(save_settings.streamType, save_settings.imagesInputType, save_settings.imagesOutputType);
    my_class->saver->setPlaybackRate(save_settings.playback_rate);
    my_class->saver->setH264Speed(save_settings.h264_speed);
    my_class->saver->setH264Quality(save_settings.h264_quality);
    my_class->saver->setH264EnableAllKeyframes(save_settings.h264_enable_all_keyframes);
    my_class->saver->setMetadataCfgSave(save_settings.save_metadata_cfg);
    my_class->saver->setStartIndex(0);
    if (save_settings.tag_label.isEmpty())
        my_class->saver->setTag(QString());
    else
        my_class->saver->setTag(save_settings.tag_label);
    if (save_settings.use_camera_name)
        my_class->saver->setCameraName(my_class->pmMetaDataInfo->GetMetaInfo()->camera_name);

    // Set the clip start to zero, 'cuz we'll be using absolute frame numbers as our sequence numbers.
    my_class->saver->setClipStart(0);

    // Tell the saver the number of frames.
    my_class->saver->setCount((saveDialog->get_clip_end() - saveDialog->get_clip_start()) + 1);
    my_class->saver->setClipEnd(my_class->saver->getFrameCount() - 1);

    // Do we need to merge the overlay information?
    if (save_settings.merge_overlays)
    {
        my_class->video_player->render_start();
    }

    // Set video streaming parameters if applicable.
    if (FiSaver::isStreamVideo(save_settings.streamType))
    {
        // Set the video stream image width & height.
        if (save_settings.merge_overlays)
        {
            QGraphicsScene  *scene_ptr;

            // Get a pointer to our "scene".
            scene_ptr = my_class->video_player->render_get_scene();
            if (scene_ptr == nullptr)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "scene_ptr is nullptr?!";
                goto save_bail;
            }

            my_class->saver->setImageHeight((uint32_t)scene_ptr->sceneRect().height());
            my_class->saver->setImageWidth((uint32_t)scene_ptr->sceneRect().width());
        }
        else
        {
            my_class->saver->setImageHeight(my_class->pmMetaDataInfo->GetMetaInfo()->roi_height);
            my_class->saver->setImageWidth(my_class->pmMetaDataInfo->GetMetaInfo()->roi_width);
        }

        // Have it prepare for a video stream writer save.
        if (my_class->saver->prepareStreamWriter() < 0)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "PrepareStreamWriter() failed!";
            goto save_bail;
        }
    }
    else if (my_class->saver->createDestPath() < 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "create_dest_path() failed!";
        goto save_bail;
    }

    // Get our metadata array ready for the raw per-frame information.
    metadata_array.clear();

    // Reset the save complete flags for stream video saving.
    my_class->wait_for_save_complete    = false;
    my_class->save_had_error            = false;
    my_class->save_is_complete          = false;
    my_class->saver_count = my_class->saver->getFrameCount();

    // Set a flag to determine if creating BMP frames for the save process.
    // AVI/BMP videos are considered as well.
    if (save_settings.streamType == FI_SAVER::STREAM_TYPE::BMP_STACK ||
        (save_settings.streamType == FI_SAVER::STREAM_TYPE::AVI && save_settings.imagesOutputType == FI_SAVER::IMAGE_TYPE::BMP))
    {
        usingBMPFrames = true;
    }

    // If we are working with raw image frames we will need to tell the frame reader which one so that the proper
    // image header is applied to the resultant frame.
    if (save_settings.streamType == FI_SAVER::STREAM_TYPE::TIF_RAW_STACK)
    {
        rawFrameCodec = FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC;
    }
    else if (save_settings.streamType == FI_SAVER::STREAM_TYPE::DNG_STACK)
    {
        rawFrameCodec = FAS_VIDEO::VIDEO_CODEC::FI_DNG_STACK_CODEC;
    }

    // If saving BMP or TIF images and not merging overlays if the capture
    // data is from a monochrome sensor we will force the save operation to
    // produce 8-bpp frame data.
    // NOTE:  FI_SAVER::STREAM_TYPE::TIF_STACK is NOT FI_SAVER::STREAM_TYPE::TIF_RAW_STACK
    if (usingBMPFrames || (save_settings.streamType == FI_SAVER::STREAM_TYPE::TIF_STACK))
    {
        // Only force if not merging overlays.
        if (!save_settings.merge_overlays)
        {
            if ((my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX13HS_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX160_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUX2100_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO) ||
                (my_class->pmMetaDataInfo->GetMetaInfo()->sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO))
            {
                // Setup to able to notify the frame reader what type we want...
                if (usingBMPFrames)
                {
                    force8BitMonoCodec = FAS_VIDEO::VIDEO_CODEC::FI_BMP_STACK_CODEC;
                }
                else
                {
                    force8BitMonoCodec = FAS_VIDEO::VIDEO_CODEC::FI_TIFF_STACK_CODEC;
                }

                // We only want to force an 8-bit monochrome shrinkage if our captured frames > 8 Bit Mode Capture.
                force8BitMono = !IS_8_BIT_MODE((char)my_class->pmMetaDataInfo->GetMetaInfo()->bit_sel);
            }
        }
    }

    // Initialize our values for our statistics.
    byte_count_stat = 0;
    save_timer.start();

    // Loop through all the frames. Make sure we're using absolute (not relative) frame numbers.
    for (my_class->saver_frame = saveDialog->get_clip_start();
         !my_class->thread_stop && (my_class->saver_frame <= saveDialog->get_clip_end()) && saveStatusOK; ++my_class->saver_frame)
    {
        FAS_VIDEO::VIDEO_CODEC fvCodec;
        QByteArray  my_buffer;
        QByteArray  *pfmd_ptr;
        QString     saveStatusOK_error_str;
        fiPixmap    my_pixmap;
        uint32_t    my_seqNbr;
        bool do_fpn;

        // With color processing we need to install information in the saver frame for Fastec CCM Processing.
        // NOTE:  The parameter "do_fpn" is set to the CAP file's metadata value for FPN correction as a result
        // of this call.  If set in the CAP file's metadata field fpn_status by looking at the GIGE_FPN_FPGA_ENABLE_MASK
        // mask then that fixed pattern noise setting is maintained while loading the frame.
        if (!my_class->init_ccm_ctldata(&my_pixmap, do_fpn))
        {
            saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("CCM initialization failed."));
            saveStatusOK = false;   // Will terminate the saver processing loop.
            goto saveStatusOK_check_error;   // Stop processing and loop around.
        }

        // There are two conditions to check for when telling the frame reader which type of header
        // to provide (this is accomplished with the fvCodec setting).
        // First check for a raw frame codec.  This will differentiate between RAW TIFF and DNG.
        if (rawFrameCodec != FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC)
        {
            fvCodec = rawFrameCodec;
        }
        // If not saving in a raw frame format then check for a monochrome 8-bit request.
        // This can be either a BMP or a TIFF (non-raw) frame being saved.
        else if (force8BitMonoCodec != FAS_VIDEO::VIDEO_CODEC::FI_INVALID_CODEC)
        {
            fvCodec = force8BitMonoCodec;
        }

        // Load the frame. Make sure it's the "raw" data, so we need to force the load.
        if (!my_class->rawDrive->readFrame(my_class->saver_frame, do_fpn, my_class->r_lut, force8BitMono, &fvCodec, &my_buffer))
        {
            saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Frame read error."));
            saveStatusOK = false;   // Will terminate the saver processing loop.
            goto saveStatusOK_check_error;   // Stop processing and loop around.
        }

        // How are we saving this frame.
        switch (save_settings.streamType)
        {
        case FI_SAVER::STREAM_TYPE::BMP_STACK:
        case FI_SAVER::STREAM_TYPE::TIF_STACK:
        case FI_SAVER::STREAM_TYPE::AVI:
            // If the image was modified nothing to do it is ready for processing.
            if (force8BitMono)
            {
                break;
            }
            // The else condition will fall through to the normal handling.
            [[fallthrough]];
        case FI_SAVER::STREAM_TYPE::MP4:
        case FI_SAVER::STREAM_TYPE::JPEG_STACK:
            // Load these into a pixmap.
            if (!my_pixmap.load_from_buf(my_buffer, fvCodec))
            {
                saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Frame image loading error."));
                saveStatusOK = false;   // Will terminate the saver processing loop.
                goto saveStatusOK_check_error;   // Stop processing and loop around.
            }

            // Pass the image through the lookup table.
            my_pixmap.apply_lookup(my_class->r_lut, my_class->g_lut, my_class->b_lut);

            // Reset our buffer.
            my_buffer.clear();

            // Do we need to merge the overlay information?
            if (save_settings.merge_overlays)
            {
                my_pixmap = my_class->mergeOverlay(my_class->saver_frame, my_pixmap);
            }

            // Update the quality factor. This will only be used if we're saving JPEGs.
            my_pixmap.set_jpeg_quality(save_settings.jpeg_quality);

            // Get the pixmap as a qImage.
            my_pixmap.save_image(&my_buffer, save_settings.imagesOutputType);

            // Make sure that the transcoded frame has a frame size > 0.
            if (my_buffer.length() <= 0)
            {
                // Something went wrong during the transcode operation.  Warn the user.
                saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Frame transcoding error."));
                saveStatusOK = false;   // Will terminate the saver processing loop.
                goto saveStatusOK_check_error;   // Stop processing and loop around.
            }
            break;
        case FI_SAVER::STREAM_TYPE::DNG_STACK:
        case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
            // Nothing to do my_buffer is already set up for this operation.
            break;
        default:
            break;
        }

        // Accumulate the byte count.
        byte_count_stat += (quint64)my_buffer.length();

        // Set the sequence number.
        my_seqNbr = my_class->saver_frame + 1;

        // Save the image.
        // However if it fails we will still process through the images
        // but the update progress will display error condition and the
        // fact that nothing is being updated should cause user to cancel.
        if (!my_class->saver->saveToStream((const uint8_t *)my_buffer.data(), (uint32_t)my_buffer.length(), my_seqNbr, false))
        {
            saveStatusOK_error_str = save_frame_err + tr("[frame=%1 %2]").arg(my_class->saver_frame).arg(tr("Save to stream error."));
            saveStatusOK = false;   // Will terminate the saver processing loop.
            goto saveStatusOK_check_error;   // Stop processing and loop around.
        }

        // Get the frame's raw metadata.
        // NOTE:  The last parameter in the call is a boolean which indicates whether or not to "do_swap"
        // meaning to swap the words using qToBigEndian.
        pfmd_ptr = my_class->pbMetaInfoPtr->getRawPFMD(my_class->saver_frame, my_class->isSeries3());
        if (pfmd_ptr != nullptr)
        {
            // Append this frame's metadata.
            metadata_array.append(*pfmd_ptr);

            // Release the pointer.
            delete pfmd_ptr;
        }

saveStatusOK_check_error:
        // If the saveStatusOK is not ok then tell the user about it.
        if (!saveStatusOK)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << saveStatusOK_error_str;
            // We don't have to worry about timers and stream timeouts so just popup a message box
            // that tells the user what happened.
            //
            // NOTE:  In addition the xferDialog will close right after the user acknowledges this
            // as well so no need to put any information in it either.  Because this is a thread we
            // don't want to be modifying the xferDialog that is being updated by a polling timer
            // running in our base class outside of this thread.
            QMessageBox::warning(my_class, "Error Saving File", saveStatusOK_error_str);
        }
    }
    //-------------------------------------------------------------
    // End of the frame processing scope.  This very long for loop!
    //-------------------------------------------------------------

    // If saving a stream video there could be outstanding frame flushes that exist and
    // progress reporting that still is happening (assuming no error has occurred).
    if (my_class->saver->isStreamVideo(save_settings.streamType))
    {
        my_class->wait_for_save_complete = true;

        // If an error had occurred above then set the flag condition.
        if (!saveStatusOK)
        {
            my_class->save_had_error = true;
        }

        while (!my_class->thread_stop && !my_class->save_is_complete)
        {
            // If the saver does not show an error condition and the
            // we are still waiting for save_is_complete pause our
            // thread.
            if (!my_class->save_had_error)
            {
                my_class->saver_thread.suspend();
            }
        }
    }

    // Do we need to clean overlay stuff up?
    if (save_settings.merge_overlays)
        my_class->video_player->render_end();

    // Get the elapsed time and generate statistics.
    elapsed = save_timer.elapsed();
    elapsed_time = elapsed_time.addMSecs((int)elapsed);

    // How many bytes per second.
    bytes_per_sec = byte_count_stat / ((double)elapsed / 1000.0);
    mbytes_per_sec = (bytes_per_sec / (1024.0 * 1024.0));

    // Build up our output string.
    stats_msg = QString::fromUtf8("Duration: %1  TotalBytes: %2  TotalFrames: %3  EffectiveRate: %4 MB/s  Format: %5")
                .arg(elapsed_time.toString(QString("hh:mm:ss.zzz"))).arg(byte_count_stat).arg((saveDialog->get_clip_end() - saveDialog->get_clip_start()) + 1).
                arg(mbytes_per_sec).arg(my_class->saver->streamTypeToQString(my_class->saver->getStreamType(), false));

    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");
    QLOG_INFO() << __PRETTY_FUNCTION__ << my_class->saver->getLocalPath();
    QLOG_INFO() << __PRETTY_FUNCTION__ << stats_msg;
    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");

    // Save the capture metadata. Start off with a copy.
    memcpy(&metadata_copy, my_class->pmMetaDataInfo->GetMetaInfo(), sizeof(pm_metadata));

    // Was there a comment from the user?
    if (!save_settings.comment.isEmpty())
        memcpy(metadata_copy.comment, save_settings.comment.toUtf8().constData(), GIGE_COMMENT_MAX_LEN);

    // Save the raw metadata.
    if (!metadata_array.isEmpty())
        my_class->saver->saveRawMetadata(&metadata_array, &metadata_copy, ((uint32_t)(my_class->r_gain * 256.0)), ((uint32_t)(my_class->g_gain * 256.0)), ((uint32_t)(my_class->b_gain * 256.0)));

    // Set the clip start value back to the actual start frame number. The metadata saving
    // code needs to know this information.
    my_class->saver->setClipStart(saveDialog->get_clip_start());
    my_class->saver->setClipEnd(saveDialog->get_clip_end());

    // Put the capture metadata into a byte array and have the saver object save it.
    metadata_array.clear();
    metadata_array.setRawData((const char *)&metadata_copy, sizeof(metadata_copy));
    my_class->saver->savePartitionMetadata(&metadata_array);

    // Wrap up the save.
    my_class->saver->finishSave();

save_bail:
    // We are done and leaving let our class know.
    emit my_class->saveImagesComplete(saveStatusOK);
}

/*!
 * @brief Retrieve the capture metadata (pm_metadata) from the drive and save it to
 *        the specified file.
 *
 * @param filename  Path for the pm_metadata.
 * @return True on success, false otherwise.
 */
bool PlaybackRawMDI::savePmMetadata(const QString& filename)
{
    QFile       meta_file;
    QString     meta_text;
    QByteArray  meta_buffer;
    SaveVideo   saver;
    FAS_VIDEO::VIDEO_GAINS  gainsData;
    const FAS_VIDEO::VIDEO_GAINS *gainsDataPtr = &gainsData;

    // Do we have a rawDrive?
    if (rawDrive.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MISSING_RAWDRIVE;
        return false;
    }

    // Allocate our buffer.
    pmMetaDataInfo = CaptureMetaInfo::CreateCaptureMetaInfo(rawDrive);
    if (pmMetaDataInfo.isNull() || (pmMetaDataInfo->GetMetaInfo() == nullptr))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_PM_METADATA << "for fiRawDrive:" << root_file << "!";
        return false;
    }

    // Make a buffer from the structure.
    meta_buffer.setRawData((const char *)pmMetaDataInfo->GetMetaInfo(), sizeof(pm_metadata));

    // Build our gain information structure.
    gainsData.red = (uint32_t)(r_gain * 256.0);
    gainsData.green = (uint32_t)(g_gain * 256.0);
    gainsData.blue = (uint32_t)(b_gain * 256.0);

    // Convert the metadata to a text format.
    meta_text = saver.metadataToText(&meta_buffer, FI_SAVER::STREAM_TYPE::FI_CAP, FI_SAVER::IMAGE_TYPE::NO_IMAGE, 0, rawDrive->getFrameCount() - 1, gainsDataPtr);

    // Make sure the file doesn't already exist.
    QFile::remove(filename);
    meta_file.setFileName(filename);

    // Open the file for writing.
    if (!meta_file.open(QIODevice::WriteOnly | QIODevice::Text))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << filename << "Err:" << meta_file.errorString();
        return false;
    }
    meta_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
    meta_file.write(meta_text.toUtf8());
    meta_file.close();

    return true;
}

/*!
 * @brief Retrieve the per-frame metadata (raw metadata) from the drive and save it to
 *        the specified file.
 *
 * @param filename  Path for the per-frame metadata.
 * @return True on success, false otherwise.
 */
bool PlaybackRawMDI::saveRawPerFrameMetadata(const QString& filename)
{
    uint32_t    meta_len;
    QFile       meta_file;
    QByteArray  meta_buffer;

    // Do we have a rawDrive?
    if (rawDrive.isNull())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MISSING_RAWDRIVE;
        return false;
    }

    // Read the capture metadata.
    meta_len = rawDrive->readPerFrameMetadata(meta_buffer);
    if (meta_len == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << root_file << ": Unable to read the per-frame metadata from the drive.";
        return false;
    }

    // Make sure the file doesn't already exist.
    QFile::remove(filename);
    meta_file.setFileName(filename);

    // Open the file for writing.
    if (!meta_file.open(QIODevice::WriteOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << filename << "Err:" << meta_file.errorString();
        return false;
    }
    meta_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
    meta_file.write(meta_buffer.constData(), meta_len);
    meta_file.close();

    return true;
}

/*!
 * @brief In our case, this method is used to specify the Drive URI for where the
 *        raw captured video is located.
 *
 * @param filename  The device URI to the Fastec formatted drive.
 */
bool PlaybackRawMDI::setFile(const QString& filename)
{
    // Allocate our raw drive.
    rawDrive = QSharedPointer<fiRawDrive>(new fiRawDrive(this));
    // Do we have a rawDrive?
    if (rawDrive.isNull())
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate a fiRawDrive!";
        return false;
    }

    // Make sure we can open the drive.
    if (!rawDrive->open(filename))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open the drive:" << filename;
        rawDrive.reset();
        return false;
    }

    // Save the drive URI.
    root_file = filename;

    // Initialize our metadata structures.
    if (!initMetadata())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to initialize the metadata structures!";
        return false;
    }

    // Tell the main window about our widget.
    MainWindow::GetInstance()->set_metadata_info(pbMetaInfoPtr);

    // Set our title to the drive URI + camera name.
    setWindowTitle(QDir::toNativeSeparators(filename) + QString::fromUtf8(" [%1]").arg(pmMetaDataInfo->GetMetaInfo()->camera_name));

    // Remember the number of frames in the video.
    frame_count = pbMetaInfoPtr->getFrameCount();

    // Setup our review controls.
    review_controls->set_pb_trigger(pbMetaInfoPtr->getTriggerFrame());
    review_controls->set_pb_clip_start(pbMetaInfoPtr->getStartFrame());
    review_controls->set_pb_clip_end(pbMetaInfoPtr->getEndFrame());
    review_controls->set_pb_frame_rate(pbMetaInfoPtr->getFramerate());

    // Update the summary camera information.
    camera_info->set_summary(pbMetaInfoPtr->getWidth(), pbMetaInfoPtr->getHeight(), pbMetaInfoPtr->getFramerate(),
                             pbMetaInfoPtr->getExposure(), pbMetaInfoPtr->getTriggerFrame() | GIGE_TRIGGER_REQ_FRAME_MASK,
                             pbMetaInfoPtr->getSensorGain());

    // Save away some of the information for later.
    review_controls->set_pb_frame_count(frame_count);

    // Enable the playback buttons.
    review_controls->update_buttons(true);

    // Allocate an action for the window's menu to use.
    if (playback_action != nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "playback_action is not null!";
        delete playback_action;
    }
    playback_action = new QAction(this);
    playback_action->setCheckable(true);
    connect(playback_action, &QAction::triggered, this, [this]() { show(); });
    connect(playback_action, &QAction::triggered, this, [this]() { setFocus(); });
    playback_action->setText(windowTitle());

    // Jump to the trigger frame.
    review_controls->set_current_frame_num(pbMetaInfoPtr->getTriggerFrame());

    // Turn off the autosave settings.
    camera_info->display_autosave_settings(0);

    // Show/hide the display info.
    MainWindow::GetInstance()->get_view_summary()->setChecked(camera_info->summary_is_enabled());
    MainWindow::GetInstance()->get_view_pfmd()->setChecked(camera_info->get_pfmd_setting());
    MainWindow::GetInstance()->get_view_local_frametime()->setChecked(camera_info->localtime_is_enabled());
    camera_info->show_info();

    // If we're enabling the display, make sure we grab the metadata for the current frame.
    if (camera_info->pfmd_is_enabled())
    {
        bool old_swap = swap_needed;

        // Update the per-frame metadata display.
        updatePfmd();

        // Update our event information.
        swap_needed = true;
        updateEventInfo();
        swap_needed = old_swap;
    }

    // Allow overlays
    video_player->enable_overlay(true);

    // All done.
    return true;
}

/*!
 * @brief We are being made inactive. If we are actively streaming, we need to slow the rate down by
 *        switching to background mode. In addition, we do NOT want to respond to certain signals.
 */
void PlaybackRawMDI::setInactive()
{
    // Pass this to our parent class.
    PlaybackMDI::setInactive();
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief The user has right mouse clicked in our window. Put up a
 *        context specific menu.
 *
 * @param event_ptr     Pointer to the menu event.
 */
void PlaybackRawMDI::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QMenu   context_menu(this);

    context_menu.addAction(MainWindow::GetInstance()->get_view_summary());
    context_menu.addAction(MainWindow::GetInstance()->get_view_review_controls());
    context_menu.addAction(MainWindow::GetInstance()->get_view_pfmd());
    context_menu.addAction(MainWindow::GetInstance()->get_view_local_frametime());
    context_menu.addAction(MainWindow::GetInstance()->get_event_marker_action());
    context_menu.addAction(MainWindow::GetInstance()->get_view_aiming_cursor());
    context_menu.addAction(MainWindow::GetInstance()->get_view_info_cursor());
    context_menu.addAction(video_player->get_action_overlay());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_metadata_dock()->toggleViewAction());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_save());

    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

/*!
 * @brief Respond to getting focus. We need to tell the main window to make other "children" inactive
 *        so we can boost the frame rate to our camera.
 * @param event     Pointer to the event structure.
 */
void PlaybackRawMDI::focusInEvent(QFocusEvent *event)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
    // Pass it on up.
    PlaybackMDI::focusInEvent(event);

    // Connect the "active" actions we want.
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_save, this, &PlaybackRawMDI::doImageSave, Qt::UniqueConnection);

    // Enable the save functionality.
    MainWindow::GetInstance()->get_save()->setEnabled(true);
    MainWindow::GetInstance()->get_export_action()->setEnabled(true);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user has initiated a save operation.
 *
 * @sa MainWindow::camera_save()
 */
void PlaybackRawMDI::doImageSave()
{
    bool        sensor_is_mono;
    uint32_t    num_frames;

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Sanity checks.
    if (pbMetaInfoPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No Playback Metadata Information pointer!";
        return;
    }
    if (review_controls == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No review_controls pointer!";
        return;
    }

    // Get our dialog allocated.
    save_dlg = std::shared_ptr<SaveDialog>(new SaveDialog(0, this));
    if (!save_dlg)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << PlaybackMDI::MISSING_SAVE_DIALOG;
        return;
    }

    // Initialize our dialog. The ordering of the following lines of code is very
    // important. We need to setup the trigger and frame count BEFORE calling the
    // set_direct_transfer() method, so that the limits will be initialized and allow
    // the ranges to be setup for the set_clip_range() call.
    save_dlg->set_sensor_type(pbMetaInfoPtr->getSensorType());
    save_dlg->set_num_frames(pbMetaInfoPtr->getFrameCount());
    save_dlg->set_trig_frame(pbMetaInfoPtr->getTriggerFrame());
    save_dlg->set_direct_transfer(true);
    save_dlg->set_clip_range(review_controls->get_clip_start(), review_controls->get_clip_end());

    // Finally we can have the dialog initialize it's controls.
    save_dlg->init_controls();

    // Are there overlays?
    if (video_player)
        save_dlg->enable_overlay_merge(video_player->has_overlay());

    // Pop up the dialog.
    if (save_dlg->exec() == QDialog::Rejected)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Cancelled Save Dialog.";
        save_dlg.reset();
        return;
    }

    // Make sure there's enough space.
    num_frames = (save_dlg->get_clip_end() - save_dlg->get_clip_start()) + 1;
    sensor_is_mono = ((pbMetaInfoPtr->getSensorType() == FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO) ||
                      (pbMetaInfoPtr->getSensorType() == FI_SENSOR::SENSOR_TYPE_LUX13HS_MONO) ||
                      (pbMetaInfoPtr->getSensorType() == FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO));
    if (!MainWindow::GetInstance()->check_avail_space(save_dlg->get_dest_path(), num_frames, save_dlg->GetStreamType(), pbMetaInfoPtr->getWidth(), pbMetaInfoPtr->getHeight(), pmMetaDataInfo->GetMetaInfo()->bit_sel, sensor_is_mono))
        return;

    // Create and deploy our transfer progress dialog.
    if (CreateXferDialog())
    {
        xferDialog->setXferSrcLabel(tr("Reading image:"));
        xferDialog->showDialogWithPolling(SAVER_POLL_RATE);
    }

    // We want to hook up a signal catch for when SaveImages finishes.
    connect(this, &PlaybackRawMDI::saveImagesComplete, this, &PlaybackRawMDI::saveImagesCompleted, Qt::UniqueConnection);

    // Save the data.
    thread_stop = false;
    saver_thread = QtConcurrent::run(SaveImages, this);
}

/*!
 * @brief The saver has completed its image save and the thread completed.
 */
void PlaybackRawMDI::saveImagesCompleted(bool savedOK)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "savedOK:" << savedOK;

    // There should have only been one connection set up when it was started so we are done with it now.
    disconnect(this, &PlaybackRawMDI::saveImagesComplete, this, &PlaybackRawMDI::saveImagesCompleted);

    // Close down any progress dialogs that were up.
    DoneWithProgressDialogs();

    // For testing purposes only.  This will popup a message box about status of the save.
    //QMessageBox::information(this, tr("Save Images"), tr("Status: %1").arg(savedOK));
}
