/*!
 * @file raw_to_xml.cpp
 * @brief Implementation of the RawToXML class. This class converts the
 *        raw per-frame metadata to an XML formatted file.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//-----------------------------------------------------------------------------
// Include Files
//-----------------------------------------------------------------------------

#include <QDebug>
#include <QFile>
#include <QVector>

#include "raw_to_xml.h"
#include "logger.h"

namespace RAW_TO_XML
{

//-----------------------------------------------------------------------------
// Constant Definitions
//-----------------------------------------------------------------------------

static const QLatin1Char   ZERO_FILL('0');

static const char   *FASTEC_XMLMETA_ROOT    = "FastecMetadata";

static const char   *CAP_XMLMETA_TOP    = "capture";
static const char   *CAP_XMLMETA_VER_STR    = "version";
static const char   *CAP_XMLMETA_VERION = "1.0";

static const char   *IMAGE_XMLMETA_TOP  = "image";
static const char   *IMAGE_XMLMETA_BITMODE  = "bit_mode";
static const char   *IMAGE_XMLMETA_FRAME_COUNT  = "frame_count";
static const char   *IMAGE_XMLMETA_START_FRAME  = "start_frame";
static const char   *IMAGE_XMLMETA_END_FRAME    = "end_frame";
static const char   *IMAGE_XMLMETA_COMMENT  = "comment";
static const char   *IMAGE_XMLMETA_SENSOR_OPTS      = "sensor_options";

static const char   *ROI_XMLMETA_TOP    = "roi";
static const char   *ROI_XMLMETA_X      = "x";
static const char   *ROI_XMLMETA_Y      = "y";
static const char   *ROI_XMLMETA_WIDTH  = "width";
static const char   *ROI_XMLMETA_HEIGHT = "height";

static const char   *TRIGGER_XMLMETA_TOP    = "trigger";
static const char   *TRIGGER_XMLMETA_FRAME  = "frame";
static const char   *TRIGGER_XMLMETA_TIME   = "time";

static const char   *CAMERA_XMLMETA_TOP = "camera";
static const char   *CAMERA_XMLMETA_MAKE    = "make";
static const char   *CAMERA_XMLMETA_MODEL   = "model";
static const char   *CAMERA_XMLMETA_FPGA    = "fpga_rev";
static const char   *CAMERA_XMLMETA_SWREV   = "software_version";
static const char   *CAMERA_XMLMETA_MAC = "mac_address";
static const char   *CAMERA_XMLMETA_NAME    = "camera_name";
static const char   *CAMERA_XMLMETA_SENSOR  = "sensor_type";

static const char   *RECORD_XMLMETA_TOP = "record";
static const char   *RECORD_XMLMETA_FPS = "fps";
static const char   *RECORD_XMLMETA_SHUTTER = "shutter_speed";
static const char   *RECORD_XMLMETA_MULTISLOPE  = "multi_slope";
static const char   *RECORD_XMLMETA_SYNC_IN = "sync_in";
static const char   *RECORD_XMLMETA_SYNC_OUT    = "sync_out";

static const char   *NORMAL_XMLMETA_TOP = "normalization";
static const char   *NORMAL_XMLMETA_RED_BAL = "red_balance";
static const char   *NORMAL_XMLMETA_GRN_BAL = "green_balance";
static const char   *NORMAL_XMLMETA_BLU_BAL = "blue_balance";
static const char   *NORMAL_XMLMETA_BRIGHTNESS  = "brightness";
static const char   *NORMAL_XMLMETA_CONTRAST    = "contrast";
static const char   *NORMAL_XMLMETA_GAMMA           = "gamma";
static const char   *NORMAL_XMLMETA_SENSOR_GAIN = "sensor_gain";
static const char   *NORMAL_XMLMETA_RED_GAIN    = "red_gain";
static const char   *NORMAL_XMLMETA_GRN_GAIN    = "green_gain";
static const char   *NORMAL_XMLMETA_BLU_GAIN    = "blue_gain";
static const char   *NORMAL_XMLMETA_RED_MATRIX  = "red_matrix";
static const char   *NORMAL_XMLMETA_GRN_MATRIX  = "green_matrix";
static const char   *NORMAL_XMLMETA_BLU_MATRIX  = "blue_matrix";

static const char   *FRAMES_XMLMETA_TOP         = "frames";
static const char   *FRAME_XMLMETA_TOP          = "frame";
static const char   *FRAME_XMLMETA_NUMBER       = "number";
static const char   *FRAME_XMLMETA_HW_NUMBER    = "hw_number";
static const char   *FRAME_XMLMETA_IRIG         = "irig";
static const char   *FRAME_XMLMETA_SYNC_MODE    = "sync_mode";
static const char   *FRAME_XMLMETA_TIME_STATE   = "time_state";
static const char   *FRAME_XMLMETA_MARKERS      = "markers";
static const char   *FRAME_XMLMETA_TIME         = "time";
static const char   *FRAME_XMLMETA_VTRIG_AREA   = "vtrig_area";
static const char   *FRAME_XMLMETA_VTRIG_FLAGS  = "vtrig_flags";

// Defines the bit mode strings for use in the metadata text.
static const char   *TS3_BIT_MODE_LOWER_8       = "low 8";
static const char   *TS3_BIT_MODE_MIDDLE_8      = "middle 8";
static const char   *TS3_BIT_MODE_UPPER_8       = "upper 8";
static const char   *TS3_BIT_MODE_FULL_10       = "full 10";
static const char   *TS3_BIT_MODE_UNKNOWN       = "unknown";

//! The per-frame metadata IRIG decode lock bit.
static const uint32_t   IRIG_DECODE_LOCK         = (1U << 31);

//! The per-frame metadata SYNC MODE values.
static const uint32_t   TS3_SYNC_MODE_SHIFT     = 24;
static const uint32_t   TS3_SYNC_MODE_MASK      = (3U << TS3_SYNC_MODE_SHIFT);

//! The per-frame metadata TIME STATE values.
static const uint32_t   TS3_TIME_STATE_SHIFT    = 20;
static const uint32_t   TS3_TIME_STATE_MASK     = (7U << TS3_TIME_STATE_SHIFT);

//! The per-frame metadata VIDEO TRIGGER values.
static const uint32_t VTRIG_FLAGS_SHIFT                    = 20;
static const uint32_t VTRIG_FLAGS_MASK                     = (3 << VTRIG_FLAGS_SHIFT);
static const uint32_t VTRIG_ROI_PIXEL_CHANGE_COUNT_MASK    = 0x0001FFFF;

// Extracting the marker information from the per-frame metadata.
#define STATE_MARKERS( x )      (( x ) & 0xff )
#define EDGE_MARKERS( x )       (( x ) & 0xff00 )

}   // End of namespace RAW_TO_XML.

using namespace RAW_TO_XML;


/*!
 * @brief Constructor for the RawToXML class.
 */
RawToXML::RawToXML(QObject *parent) :
    QObject(parent)
{
    series3_found = false;
    stop_processing = false;
    start_frame = 0;
    end_frame = 0;
    red_gain = 0;
    green_gain = 0;
    blue_gain = 0;
    raw_file_path.clear();
    xml_file_path.clear();
    memset(&session_metadata, 0, sizeof(pm_metadata));

    // This may change, but default to Series3 value.
    pfmd_word_len = GIGE_FRAME_METADATA_LEN / sizeof(uint32_t);
}

/*!
 * @brief Standard destructor for the RawToXML class.
 */
RawToXML::~RawToXML()
{
}

/*!
 * @brief Override the (incorrect) gains found in the pm_metadata structure.
 *
 * @param red       The red gain.
 * @param green     The green gain.
 * @param blue      The blue gain.
 */
void RawToXML::set_gains(uint32_t red, uint32_t green, uint32_t blue)
{
    this->red_gain = red;
    this->green_gain = green;
    this->blue_gain = blue;
}

/*!
 * @brief Specify the parameters associated with the conversion process.
 */
void RawToXML::set_params(const QString& pm_file_path, const QString& xml_file_path, const pm_metadata *session_meta_ptr, uint32_t start_frame, uint32_t end_frame)
{
    QString camera_model;

    // Save our parameters.
    this->raw_file_path = pm_file_path;
    this->xml_file_path = xml_file_path;
    this->start_frame = start_frame;
    this->end_frame = end_frame;
    memcpy(&this->session_metadata, session_meta_ptr, sizeof(pm_metadata));

    // Initialize the gains.
    set_gains(session_meta_ptr->red, session_meta_ptr->green, session_meta_ptr->blue);

    // Get the serial number in a format easier for us to work with.
    camera_model = QString::fromUtf8(session_meta_ptr->model_name);

    // Are we a "TSx" or "ILx" camera?
    // TODO: Replace with a RegEx.
    if (camera_model.contains(QStringLiteral("TS")))
        series3_found = true;
    else if (camera_model.contains(QStringLiteral("IL")))
        series3_found = true;

    // Calculate the PFMD length.
    if (series3_found)
        pfmd_word_len = GIGE_FRAME_METADATA_LEN / sizeof(uint32_t);
    else
        pfmd_word_len = (2 * GIGE_FRAME_METADATA_LEN) / sizeof(uint32_t);
}

/*!
 * @brief If the user cancels, this method will stop the convert process.
 */
void RawToXML::cancel_convert()
{
    stop_processing = true;
}

/*!
 * @brief This is the main entry point for the conversion thread.
 */
void RawToXML::start_convert()
{
    QString             err_str;
    QFile       xml_file;

    // Open our XML file.
    xml_file.setFileName(xml_file_path);
    if (!xml_file.open(QFile::WriteOnly | QFile::Text))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to open the XML file %1, err=%2").arg(xml_file_path, xml_file.errorString());
        return;
    }
    xml_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // Setup our XML writer to work with our file.
    writer.setDevice(&xml_file);
    writer.setAutoFormatting(true);
    writer.writeStartDocument();
    writer.writeStartElement(QString::fromUtf8(FASTEC_XMLMETA_ROOT));

    // Save the capture settings metadata.
    save_capture_metadata();

    // Now save all the per-frame metadata information.
    save_frame_metadata();

    // All done.
    writer.writeEndDocument();
    xml_file.close();

    // Make sure everything is OK.
    if (xml_file.error() != QFile::NoError)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Error writing XML file %1, err=%2").arg(xml_file_path, xml_file.errorString());
        emit error(err_str);
    }

    // Signal that we're done.
    emit finished();
}

/*!
 * @brief Save the partition metadata.
 */
void RawToXML::save_capture_metadata(void)
{
    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Start with our top level element.
    writer.writeStartElement(QString::fromUtf8(CAP_XMLMETA_TOP));

    // Add a version attribute attribute.
    writer.writeAttribute(QString::fromUtf8(CAP_XMLMETA_VER_STR), QString::fromUtf8(CAP_XMLMETA_VERION));

    // Add the image data.
    save_image_metadata();

    // Add the camera specific metadata.
    save_camera_metadata();

    // Add the recording information.
    save_record_metadata();

    // Add the normalization information.
    save_normalization_metadata();

    // End the capture element.
    writer.writeEndElement();
}

/*!
 * @brief Convert straight-binary-seconds to a struct tm.
 */
void RawToXML::sbs_to_tm(uint32_t sbs, struct tm *tm_ptr)
{
    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Clear out the structure.
    memset(tm_ptr, 0, sizeof(struct tm));

    // Get years.
    tm_ptr->tm_year = sbs / (365 * 24 * 60 * 60);
    sbs -= ((uint32_t)tm_ptr->tm_year * (365 * 24 * 60 * 60));
    tm_ptr->tm_yday = sbs / (24 * 60 * 60);
    sbs -= ((uint32_t)tm_ptr->tm_yday * (24 * 60 * 60));
    tm_ptr->tm_hour = sbs / (60 * 60);
    sbs -= ((uint32_t)tm_ptr->tm_hour * (60 * 60));
    tm_ptr->tm_min = sbs / 60;
    tm_ptr->tm_sec = (int)sbs - (tm_ptr->tm_min * 60);
}

/*!
 * @brief Save the normalization metadata.
 */
void RawToXML::save_normalization_metadata(void)
{
    QString out_str;

    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Start with our top level element.
    writer.writeStartElement(QString::fromUtf8(NORMAL_XMLMETA_TOP));

    // Write out our elements
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_RED_BAL), QString::number(session_metadata.wb_red));
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_GRN_BAL), QString::number(session_metadata.wb_grn));
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_BLU_BAL), QString::number(session_metadata.wb_blu));
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_BRIGHTNESS), QString::number(session_metadata.brightness));
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_CONTRAST), QString::number(session_metadata.contrast));
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_GAMMA), QString::number(session_metadata.gamma));
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_SENSOR_GAIN), QString::number(session_metadata.sensor_gain));
    out_str = QString::fromUtf8("%1.%2").arg(red_gain >> 8).arg(((((red_gain & 0xFF) * 10000) / 256) + 5) / 10);
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_RED_GAIN), out_str);
    out_str = QString::fromUtf8("%1.%2").arg(green_gain >> 8).arg(((((green_gain & 0xFF) * 10000) / 256) + 5) / 10);
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_GRN_GAIN), out_str);
    out_str = QString::fromUtf8("%1.%2").arg(blue_gain >> 8).arg(((((blue_gain & 0xFF) * 10000) / 256) + 5) / 10);
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_BLU_GAIN), out_str);

    out_str = QString::fromUtf8("[%1,%2,%3]").arg((int16_t)session_metadata.matrix_coeff1_red).arg((int16_t)session_metadata.matrix_coeff2_red).arg((int16_t)session_metadata.matrix_coeff3_red);
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_RED_MATRIX), out_str);
    out_str = QString::fromUtf8("[%1,%2,%3]").arg((int16_t)session_metadata.matrix_coeff1_grn).arg((int16_t)session_metadata.matrix_coeff2_grn).arg((int16_t)session_metadata.matrix_coeff3_grn);
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_GRN_MATRIX), out_str);
    out_str = QString::fromUtf8("[%1,%2,%3]").arg((int16_t)session_metadata.matrix_coeff1_blu).arg((int16_t)session_metadata.matrix_coeff2_blu).arg((int16_t)session_metadata.matrix_coeff3_blu);
    writer.writeTextElement(QString::fromUtf8(NORMAL_XMLMETA_BLU_MATRIX), out_str);

    // End the camera element.
    writer.writeEndElement();
}

/*!
 * @brief Save the recording specific metadata.
 */
void RawToXML::save_record_metadata(void)
{
    QString out_str;

    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Start with our top level element.
    writer.writeStartElement(QString::fromUtf8(RECORD_XMLMETA_TOP));

    // Write out our elements
    writer.writeTextElement(QString::fromUtf8(RECORD_XMLMETA_FPS), QString::number(session_metadata.fps));
    writer.writeTextElement(QString::fromUtf8(RECORD_XMLMETA_SHUTTER), QString::number(session_metadata.shutter_speed));

    out_str = QString::fromUtf8("%1:%2").arg(session_metadata.multslope[0]).arg(session_metadata.multslope[1]);
    writer.writeTextElement(QString::fromUtf8(RECORD_XMLMETA_MULTISLOPE), out_str);
    out_str = QString::fromUtf8("0x%1").arg(session_metadata.sync_in_set, 0, 16);
    writer.writeTextElement(QString::fromUtf8(RECORD_XMLMETA_SYNC_IN), out_str);
    out_str = QString::fromUtf8("0x%1").arg(session_metadata.sync_out_set, 0, 16);
    writer.writeTextElement(QString::fromUtf8(RECORD_XMLMETA_SYNC_OUT), out_str);

    // End the camera element.
    writer.writeEndElement();
}

/*!
 * @brief Save the camera specific metadata.
 */
void RawToXML::save_camera_metadata(void)
{
    QString out_str;

    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Start with our top level element.
    writer.writeStartElement(QString::fromUtf8(CAMERA_XMLMETA_TOP));

    // Write out our elements
    writer.writeTextElement(QString::fromUtf8(CAMERA_XMLMETA_MAKE), QString::fromUtf8(session_metadata.man_name));
    writer.writeTextElement(QString::fromUtf8(CAMERA_XMLMETA_MODEL), QString::fromUtf8(session_metadata.model_name));
    out_str = QString::fromUtf8("0x%1").arg(session_metadata.fpga_rev, 0, 16);
    writer.writeTextElement(QString::fromUtf8(CAMERA_XMLMETA_FPGA), out_str);
    writer.writeTextElement(QString::fromUtf8(CAMERA_XMLMETA_SWREV), QString::fromUtf8(session_metadata.app_ver));

    out_str = QString::fromUtf8("%1:%2:%3:%4:%5:%6").arg(session_metadata.mac_addr[0], 2, 16, ZERO_FILL)
              .arg(session_metadata.mac_addr[1], 2, 16, ZERO_FILL)
              .arg(session_metadata.mac_addr[2], 2, 16, ZERO_FILL)
              .arg(session_metadata.mac_addr[3], 2, 16, ZERO_FILL)
              .arg(session_metadata.mac_addr[4], 2, 16, ZERO_FILL)
              .arg(session_metadata.mac_addr[5], 2, 16, ZERO_FILL);
    writer.writeTextElement(QString::fromUtf8(CAMERA_XMLMETA_MAC), out_str);
    writer.writeTextElement(QString::fromUtf8(CAMERA_XMLMETA_NAME), QString::fromUtf8(session_metadata.camera_name));

    out_str = QString::fromUtf8((const char *)(&session_metadata.sensor_type));
    writer.writeTextElement(QString::fromUtf8(CAMERA_XMLMETA_SENSOR), out_str);

    // End the camera element.
    writer.writeEndElement();
}

/*!
 * @brief Save the Image specific metadata.
 */
void RawToXML::save_image_metadata(void)
{
    int first = -1;
    int last = -1;
    QString     out_str;
    uint32_t    frame_count;
    uint32_t    trig_frame_num;
    const char  *bit_mode = nullptr;
    struct tm   sbs_tm;

    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Adjust frame count and trigger frame for selected clip
    frame_count = end_frame - start_frame + 1;

    // Now calculate the "relative" trigger offset.
    if (session_metadata.trig_frame < session_metadata.first_frame)
        trig_frame_num = session_metadata.frame_count - (session_metadata.first_frame - session_metadata.trig_frame);
    else
        trig_frame_num = session_metadata.trig_frame - session_metadata.first_frame;

    // Start with our top level element.
    writer.writeStartElement(QString::fromUtf8(IMAGE_XMLMETA_TOP));

    writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_FRAME_COUNT), QString::number(frame_count));
    writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_START_FRAME), QString::number(start_frame));
    writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_END_FRAME), QString::number(end_frame));

    // Make string for bit_sel
    switch (session_metadata.bit_sel)
    {
    case GIGE_BITMODE_LOWER_8_BITS: bit_mode = TS3_BIT_MODE_LOWER_8; break;
    case GIGE_BITMODE_MIDDLE_8_BITS:    bit_mode = TS3_BIT_MODE_MIDDLE_8; break;
    case GIGE_BITMODE_UPPER_8_BITS: bit_mode = TS3_BIT_MODE_UPPER_8; break;
    case GIGE_BITMODE_10_BITS:      bit_mode = TS3_BIT_MODE_FULL_10; break;
    case GIGE_BITMODE_7_0:  first = 0; last = 7; break;
    case GIGE_BITMODE_8_1:  first = 1; last = 8; break;
    case GIGE_BITMODE_9_2:  first = 2; last = 9; break;
    case GIGE_BITMODE_10_3: first = 3; last = 10; break;
    case GIGE_BITMODE_11_4: first = 4; last = 11; break;
    case GIGE_BITMODE_9_0:  first = 0; last = 9; break;
    case GIGE_BITMODE_10_1: first = 1; last = 10; break;
    case GIGE_BITMODE_11_2: first = 2; last = 11; break;
    case GIGE_BITMODE_11_0: first = 0; last = 11; break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected bit_sel:" << Qt::showbase << Qt::hex << session_metadata.bit_sel;
        break;
    }

    if (bit_mode != nullptr)
    {
        writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_BITMODE), QString::fromUtf8(bit_mode));
    }
    else if ((first >= 0) & (last >= 0))
    {
        writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_BITMODE), QString("%1:%2").arg(last).arg(first));
    }
    else
    {
        writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_BITMODE), QString::fromUtf8(TS3_BIT_MODE_UNKNOWN));
    }

    // Make string for sensor options.
    writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_SENSOR_OPTS), QString("bin%1x:subs%2x")
                            .arg((session_metadata.sensor_opts & GIGE_SENSOPT_BIN_MASK) >> GIGE_SENSOPT_BIN_SHIFT)
                            .arg((session_metadata.sensor_opts & GIGE_SENSOPT_SUBS_MASK) >> GIGE_SENSOPT_SUBS_SHIFT));

    // Write our ROI elements.
    writer.writeStartElement(QString::fromUtf8(ROI_XMLMETA_TOP));
    writer.writeTextElement(QString::fromUtf8(ROI_XMLMETA_X), QString::number(session_metadata.roi_x));
    writer.writeTextElement(QString::fromUtf8(ROI_XMLMETA_Y), QString::number(session_metadata.roi_y));
    writer.writeTextElement(QString::fromUtf8(ROI_XMLMETA_WIDTH), QString::number(session_metadata.roi_width));
    writer.writeTextElement(QString::fromUtf8(ROI_XMLMETA_HEIGHT), QString::number(session_metadata.roi_height));
    writer.writeEndElement();

    // Add the trigger information.  Start with the top level element.
    writer.writeStartElement(QString::fromUtf8(TRIGGER_XMLMETA_TOP));
    writer.writeTextElement(QString::fromUtf8(TRIGGER_XMLMETA_FRAME), QString::number(trig_frame_num));

    // Calculate time/date of trigger
    memset(&sbs_tm, 0, sizeof(sbs_tm));
    sbs_to_tm(session_metadata.trig_time_s, &sbs_tm);
    out_str = QString("%1:%2:%3:%4:%5.%6").arg(sbs_tm.tm_year, 2, 10, ZERO_FILL)
              .arg(sbs_tm.tm_yday + 1, 3, 10, ZERO_FILL)
              .arg(sbs_tm.tm_hour, 2, 10, ZERO_FILL)
              .arg(sbs_tm.tm_min, 2, 10, ZERO_FILL)
              .arg(sbs_tm.tm_sec, 2, 10, ZERO_FILL)
              .arg(((session_metadata.trig_time_ns + 500) / 1000), 6, 10, ZERO_FILL);
    writer.writeTextElement(QString::fromUtf8(TRIGGER_XMLMETA_TIME), out_str);

    // End the trigger element.
    writer.writeEndElement();

    // Save the comment field.
    writer.writeTextElement(QString::fromUtf8(IMAGE_XMLMETA_COMMENT), QString::fromUtf8((const char *)session_metadata.comment));

    // End the image element.
    writer.writeEndElement();
}

/*!
 * @brief Write the per-frame metadata for a single frame.
 *
 * @param frame_words   The raw per-frame metadata.
 * @param frame_no  The frame number.
 */
void RawToXML::save_one_frame_metadata(const uint32_t *frame_words, uint32_t frame_no)
{
    QString     out_str;
    struct tm   sbs_tm;
    uint32_t area_change;
    int vtrig_flags;

    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Convert the SBS to something we can use.
    sbs_to_tm(frame_words[0], &sbs_tm);

    // Start with our top level element.
    writer.writeStartElement(QString::fromUtf8(FRAME_XMLMETA_TOP));

    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_NUMBER), QString::number(frame_no));
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_HW_NUMBER), QString::number(frame_words[3]));
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_IRIG), QString::number((frame_words[1] & IRIG_DECODE_LOCK) == IRIG_DECODE_LOCK));
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_SYNC_MODE), QString::number((frame_words[1] & TS3_SYNC_MODE_MASK) >> TS3_SYNC_MODE_SHIFT));
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_TIME_STATE), QString::number(((frame_words[1] & TS3_TIME_STATE_MASK) >> TS3_TIME_STATE_SHIFT)));

    out_str = QString::fromUtf8("0x%1").arg((STATE_MARKERS(frame_words[2]) | EDGE_MARKERS(frame_words[2])), 4, 16, ZERO_FILL);
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_MARKERS), out_str);

    // Linux tm day-of-year is zero based, but we report days starting with 1=Jan 1.
    out_str = QString::fromUtf8("%1:%2:%3:%4:%5.%6").arg(sbs_tm.tm_year, 2, 10, ZERO_FILL)
              .arg(sbs_tm.tm_yday + 1, 3, 10, ZERO_FILL)
              .arg(sbs_tm.tm_hour, 2, 10, ZERO_FILL)
              .arg(sbs_tm.tm_min, 2, 10, ZERO_FILL)
              .arg(sbs_tm.tm_sec, 2, 10, ZERO_FILL)
              .arg((frame_words[ 1 ] & 0xFFFFF), 6, 10, ZERO_FILL);
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_TIME), out_str);

    // Save the number of pixels counter so we can use it a few times.
    area_change = frame_words[7] & VTRIG_ROI_PIXEL_CHANGE_COUNT_MASK;
    vtrig_flags = (frame_words[7] & VTRIG_FLAGS_MASK) >> VTRIG_FLAGS_SHIFT;
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_VTRIG_AREA), QString::number(area_change));
    writer.writeTextElement(QString::fromUtf8(FRAME_XMLMETA_VTRIG_FLAGS), QString::number(vtrig_flags));

    writer.writeEndElement();
}

/*!
 * @brief Loop through the raw per-frame metadata file located
 *        at raw_file_path, and format the data.
 */
void RawToXML::save_frame_metadata(void)
{
    QFile       raw_file;
    uint32_t    frame_num;
    QVector<uint32_t> frame_words((int)pfmd_word_len);

    // Check to see if the user is cancelling.
    if (stop_processing)
        return;

    // Open the per-frame metadata file.
    raw_file.setFileName(raw_file_path);
    if (!raw_file.open(QFile::ReadOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Unable to open the per-frame metadata file %1, err=%2").arg(raw_file_path, raw_file.errorString());
        return;
    }

    // Start with our top level element.
    writer.writeStartElement(QString::fromUtf8(FRAMES_XMLMETA_TOP));

    // Loop through all our frames.
    for (frame_num = start_frame; !stop_processing && (frame_num <= end_frame); ++frame_num)
    {
        if (raw_file.read((char *)frame_words.data(), pfmd_word_len * sizeof(uint32_t)) < 0)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << QString::fromUtf8("Error reading raw metadata, err=%1").arg(raw_file.errorString());
            break;
        }

        // If this is a series3 camera, byte swap the data.
        if (series3_found)
        {
            for (int i = 0; i < (int)pfmd_word_len; ++i)
            {
                frame_words[i] = qToBigEndian(frame_words[i]);
            }
        }

        // Save this frames worth of metadata.
        save_one_frame_metadata(frame_words.constData(), frame_num);
    }

    // Close our frame element.
    writer.writeEndElement();

    // Close the raw file.
    raw_file.close();
}
