/*!
 * @file regcache.cpp
 * @brief This provides the implementation of the RegCache class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QtAlgorithms>

#include "regcache.h"
#include "logger.h"

namespace REG_CACHE
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define a type for the actual entry in the register cache.
typedef struct CACHE_ENTRY
{
    QVariant    entry;              //!< The last value read from the camera.

    uint32_t    flags;              //!< Any special requirements for this entry.
    uint32_t    expire_sec;         //!< Expiration time in seconds (if applicable).
    uint32_t    expire_msec;        //!< Expiration time in milliseconds (if applicable).
    QTime       last_read;          //!< "0" indicates never read, otherwise the time we last updated the value.
} CACHE_ENTRY;

const uint32_t  DEFAULT_ENTRY_FLAGS     = 0;        //!< Default flags, when none are specified.
const uint32_t  DEFAULT_LIFETIME_SEC    = 0;        //!< Default lifetime in seconds.
const uint32_t  DEFAULT_LIFETIME_MSEC   = 100;      //!< Default lifetime in milliseconds.

//--------------------------------------------------------------------
// Variable Declarations
//--------------------------------------------------------------------

}   // End of namespace REG_CACHE

using namespace REG_CACHE;

/*!
 * @brief RegCache class constructor.
 */
RegCache::RegCache(QObject *parent) :
    QObject(parent)
{
}

/*!
 * @brief RegCache class constructor with an initialization list.
 */
RegCache::RegCache(REG_CACHE::CACHE_LIST& cache_list, QObject *parent) :
    QObject(parent)
{
    // Initialize our cache from the list.
    set_entries(cache_list);
}

RegCache::~RegCache()
{
    // Make sure we delete all the entries in the cache.
    clear();
}

/*!
 * @brief This clears out the cache, deleting all the entries and resetting
 *        the map to an empty state.
 */
void RegCache::clear()
{
    // Finally, flush the entire cache contents.
    cache.clear();
}

/*!
 * @brief Allow the client to expire a cache entry.
 * @param reg_addr  The cache entry we're interested in.
 */
void RegCache::expire_entry(uint32_t reg_addr)
{
    QSharedPointer<CACHE_ENTRY> entry_ptr;

    // Check to see if we already have an entry for this address.
    if (cache.contains(reg_addr))
    {
        // Get a pointer to our entry.
        entry_ptr = cache.value(reg_addr);

        // Mark this entry as never read.
        entry_ptr->last_read = QTime(0, 0);
    }
}

/*!
 * @brief Initializes the cache based on the provided list of initialization values.
 * @param cache_list    The list of initializers for the cache.
 */
void RegCache::set_entries(REG_CACHE::CACHE_LIST& cache_list)
{
    int i;
    QSharedPointer<CACHE_ENTRY> entry_ptr;

    // Make sure our cache is empty.
    clear();

    // Loop through each entry in the list and build up our cache from it.
    for (i = 0; i < cache_list.count(); ++i)
    {
        // Allocate a new entry.
        entry_ptr = QSharedPointer<CACHE_ENTRY>(new CACHE_ENTRY);
        if (entry_ptr.get() == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate memory!";
            return;
        }

        // Get the current entry in the init list.
        CACHE_SETTING init_value = cache_list.at(i);

        // Assign the values to our cache entry.
        entry_ptr->flags = init_value.flags;
        entry_ptr->expire_sec = init_value.expire_sec;
        entry_ptr->expire_msec = init_value.expire_msec;
        entry_ptr->last_read = QTime(0, 0);

        // Set the entry into our map.
        cache.insert(init_value.addr, entry_ptr);
    }
}

/*!
 * @brief Set a single entry into the cache based on the values passed in.
 * @param cache_setting     Initializer for a cache entry.
 */
void RegCache::set_entry(REG_CACHE::CACHE_SETTING& cache_setting)
{
    QSharedPointer<CACHE_ENTRY> entry_ptr;

    // Check to see if we already have an entry for this address.
    if (cache.contains(cache_setting.addr))
    {
        // Get a pointer to our entry.
        entry_ptr = cache.value(cache_setting.addr);
    }
    else
    {
        // Allocate a new entry.
        entry_ptr = QSharedPointer<CACHE_ENTRY>(new CACHE_ENTRY);
        if (entry_ptr.get() == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate memory!";
            return;
        }

        // Insert our new entry into the cache.
        cache.insert(cache_setting.addr, entry_ptr);
    }

    // Initialize our entry.
    entry_ptr->flags = cache_setting.flags;
    entry_ptr->expire_sec = cache_setting.expire_sec;
    entry_ptr->expire_msec = cache_setting.expire_msec;
    entry_ptr->last_read = QTime(0, 0);
}

/*!
 * @brief Update an integer value in the cache. If an entry is not found, a default
 *        entry is created.
 *
 * @param reg_addr      The address of the register we're interested in.
 * @param reg_value     The value of the register.
 */
void RegCache::set_reg(uint32_t reg_addr, uint32_t reg_value)
{
    QSharedPointer<CACHE_ENTRY> entry_ptr;

    // Check to see if we already have an entry for this address.
    if (cache.contains(reg_addr))
    {
        // Get a pointer to our entry.
        entry_ptr = cache.value(reg_addr);
    }
    else
    {
        // Allocate a new entry.
        entry_ptr = QSharedPointer<CACHE_ENTRY>(new CACHE_ENTRY);
        if (entry_ptr.get() == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate memory!";
            return;
        }

        // Initialize our entry.
        entry_ptr->flags = DEFAULT_ENTRY_FLAGS;
        entry_ptr->expire_sec = DEFAULT_LIFETIME_SEC;
        entry_ptr->expire_msec = DEFAULT_LIFETIME_MSEC;

        // Insert our new entry into the cache.
        cache.insert(reg_addr, entry_ptr);
    }

    // Update our entry values.
    entry_ptr->entry.setValue(reg_value);
    entry_ptr->last_read = QTime::currentTime();
}

/*!
 * @brief Update an array value in the cache. If an entry is not found, a default
 *        entry is created.
 *
 * @param reg_addr      The address of the register we're interested in.
 * @param mem_value     The array for the register.
 */
void RegCache::set_reg(uint32_t reg_addr, const QByteArray *mem_value)
{
    QSharedPointer<CACHE_ENTRY> entry_ptr;

    // Check to see if we already have an entry for this address.
    if (cache.contains(reg_addr))
    {
        // Get a pointer to our entry.
        entry_ptr = cache.value(reg_addr);
    }
    else
    {
        // Allocate a new entry.
        entry_ptr = QSharedPointer<CACHE_ENTRY>(new CACHE_ENTRY);
        if (entry_ptr.get() == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate memory!";
            return;
        }

        // Initialize our entry.
        entry_ptr->flags = DEFAULT_ENTRY_FLAGS;
        entry_ptr->expire_sec = DEFAULT_LIFETIME_SEC;
        entry_ptr->expire_msec = DEFAULT_LIFETIME_MSEC;

        // Insert our new entry into the cache.
        cache.insert(reg_addr, entry_ptr);
    }

    // Update our entry values.
    entry_ptr->entry.setValue(*mem_value);
    entry_ptr->last_read = QTime::currentTime();
}

/*!
 * @brief Retrieve the value associated with the register at reg_addr.
 *
 * @param reg_addr  The register we're interested in.
 * @param ok        [OPTIONAL] Pointer to bool: true if everything is OK, false if there was an error.
 * @param expired   [OPTIONAL] Pointer to bool: true if the value has expired, false if it's OK.
 *
 * @return The value of the register.
 */
uint32_t RegCache::get_reg(uint32_t reg_addr, bool *ok, bool *expired)
{
    uint32_t    reg_value;
    QSharedPointer<CACHE_ENTRY> entry_ptr;

    // Check to see if we already have an entry for this address.
    if (cache.contains(reg_addr))
    {
        // Get a pointer to our entry.
        entry_ptr = cache.value(reg_addr);

        // Has this entry ever been accessed?
        if (entry_ptr->last_read == QTime(0, 0))
        {
            reg_value = 0;
            if (ok != nullptr)
                *ok = false;
        }
        else
        {
            // Get our value.
            reg_value = entry_ptr->entry.toUInt(ok);

            // Do we want to return our expiration state?
            if (expired != nullptr)
            {
                // Is this entry expired?
                *expired = entry_has_expired(entry_ptr.get());
            }
        }
    }
    else
    {
        // No entry.
        reg_value = 0;
        if (ok != nullptr)
            *ok = false;
    }

    // Return our value.
    return reg_value;
}

/*!
 * @brief Retrieve the buffer associated with the register at reg_addr.
 *
 * @param reg_addr  The register we're interested in.
 * @param ok        [OPTIONAL] Pointer to bool: true if everything is OK, false if there was an error.
 * @param expired   [OPTIONAL] Pointer to bool: true if the value has expired, false if it's OK.
 *
 * @return Pointer to the memory buffer associated with the register. It's the responsibility of the
 *         caller to free the memory.
 */
QByteArray *RegCache::get_mem(uint32_t reg_addr, bool *ok, bool *expired)
{
    QByteArray  *mem_value;
    QSharedPointer<CACHE_ENTRY> entry_ptr;

    // Check to see if we already have an entry for this address.
    if (cache.contains(reg_addr))
    {
        // Get a pointer to our entry.
        entry_ptr = cache.value(reg_addr);

        // Has this entry ever been read?
        if (entry_ptr->last_read == QTime(0, 0))
        {
            mem_value = nullptr;
            if (ok != nullptr)
                *ok = false;
        }
        else
        {
            // Get our value.
            mem_value = new QByteArray(entry_ptr->entry.toByteArray());
            if (ok != nullptr)
                *ok = mem_value != nullptr;

            // Do we want to return our expiration state?
            if (expired != nullptr)
            {
                // Is this entry expired?
                *expired = entry_has_expired(entry_ptr.get());
            }
        }
    }
    else
    {
        // No entry.
        mem_value = nullptr;
        if (ok != nullptr)
            *ok = false;
    }

    // Return our value.
    return mem_value;
}

/*!
 * @brief Utility method to determine if an entry has expired.
 *
 * @param entry_ptr The entry we're interested in.
 * @return True if the entry is expired, false otherwise.
 */
bool RegCache::entry_has_expired(const REG_CACHE::CACHE_ENTRY *entry_ptr)
{
    bool    expired;

    // Sanity check.
    if (entry_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null entry_ptr!?";
        return false;
    }

    // Are we worried about expiration?
    if ((entry_ptr->flags & ENTRY_NEVER_EXPIRES) == ENTRY_NEVER_EXPIRES)
    {
        // It never expires.
        expired = false;
    }

    else if ((entry_ptr->flags & ENTRY_NO_CACHE) == ENTRY_NO_CACHE)
    {
        // It is never to be cached.
        expired = true;
    }

    else
    {
        QTime   expiration_time = entry_ptr->last_read;

        // Calculate our expieration time.
        expiration_time = expiration_time.addMSecs((int)entry_ptr->expire_msec);
        expiration_time = expiration_time.addSecs((int)entry_ptr->expire_sec);

        // Did we expire?
        if (expiration_time < QTime::currentTime())
        {
            // This entry has aged out.
            expired = true;
        }

        else
        {
            // Must be OK.
            expired = false;
        }
    }

    // Return our findings.
    return expired;
}
