/*!
 * @file regcache.h
 * @brief This file contains the definitions for the register caching
 *        class, RegCache.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef REGCACHE_H
#define REGCACHE_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <stdint.h>

#include <QByteArray>
#include <QMap>
#include <QObject>
#include <QSharedPointer>
#include <QTime>
#include <QVariant>

//--------------------------------------------------------------------
// Namespace: FI_COMM
//--------------------------------------------------------------------
namespace REG_CACHE
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

const uint32_t  ENTRY_NEVER_EXPIRES     = 1 << 0;       //!< If set, this value is read once and never expires.
const uint32_t  ENTRY_NO_CACHE          = 1 << 1;       //!< If set, this entry is never to be cached.

//! Define a type for each entry in the register cache.
typedef struct CACHE_SETTING
{
    uint32_t    addr;               //!< The address for the register.
    uint32_t    flags;              //!< Flags to define special needs for this register.
    uint32_t    expire_sec;         //!< Expiration time in seconds (if applicable).
    uint32_t    expire_msec;        //!< Expiration time in milliseconds (if applicable).
} CACHE_SETTING;

//! Define a type for a list of entries. This would be used for
//! initializing the cache, for example.
typedef QList<CACHE_SETTING>    CACHE_LIST;

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

struct CACHE_ENTRY;

}   // End of namespace REG_CACHE


//--------------------------------------------------------------------
// Class RegCache
//--------------------------------------------------------------------
class RegCache : public QObject
{
    Q_OBJECT

public:
    explicit RegCache(QObject *parent = nullptr);
    RegCache(REG_CACHE::CACHE_LIST& cache_list, QObject *parent = nullptr);
    virtual ~RegCache();

    void    clear();
    void    expire_entry(uint32_t reg_addr);

    void    set_entries(REG_CACHE::CACHE_LIST& cache_list);
    void    set_entry(REG_CACHE::CACHE_SETTING& cache_setting);

    void    set_reg(uint32_t reg_addr, uint32_t reg_value);
    void    set_reg(uint32_t reg_addr, const QByteArray *mem_value);

    uint32_t    get_reg(uint32_t reg_addr, bool *ok = nullptr, bool *expired = nullptr);
    QByteArray *get_mem(uint32_t reg_addr, bool *ok = nullptr, bool *expired = nullptr);

private:
    QMap<uint32_t, QSharedPointer<REG_CACHE::CACHE_ENTRY>> cache;      //!< A map of [ reg_addr, cache_entry ].

    bool    entry_has_expired(const REG_CACHE::CACHE_ENTRY *entry_ptr);
};

#endif // REGCACHE_H
