/*!
 * @file save_dialog.cpp
 * @brief This Save dialog, requesting the user to provide all the information
 *        necessary to save the captured data.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "save_dialog.h"
#include "ui_save_dialog.h"

#include <QRegularExpression>
#include <QRegularExpressionValidator>
#include <QFileDialog>
#include <QHostInfo>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonParseError>
#include <QMessageBox>
#include <QSettings>
#include <QStandardPaths>

#include "camera_mdi.h"
#include "camera_explore.h"
#include "client_math.h"
#include "fi_comm.h"
#include "fi_internal.h"
#include "fi_sensor.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"
#include "panel_capture.h"
#include "playback_mdi.h"

namespace SAVE_DLG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

#if 0
#define MP4_DEBUG 1;
#endif

const int   DEFAULT_JPEG_QUAL           = 80;       //!< Default JPEG quality setting.
const int   MP4_JPEG_QUALITY            = 100;      //!< Saving MP4s requires streams of 100% JPEG quality.
const int   MAX_AVI_PLAY_RATE           = 999;      //!< The maximum AVI playback rate we'll support.

const int   MAX_PATH_ENTRIES            = 16;       //!< The maximum number of path entries we'll support.


//! The path entries for downloading images.
struct DOWNLOAD_PATH
{
    uint32_t ip_addr;
    QString path;
    QTime last_accessed;
};

/*!
 * @brief The PathRepository class is used to abstract the storage and retrieval
 *        of the "path" information.
 *
 * We wanted to be able to save and restore the paths used to save images based on
 * the IP address of the camera. This class abstracts that for the SaveDialog object.
 */
class PathRepository
{
private:
    PathRepository() {}

public:

    //! Don't allow this object to be cloned.
    PathRepository(PathRepository& other) = delete;

    //! Not assignable.
    void operator=(const PathRepository&) = delete;

    /*!
     * @brief Find the path associated with the cam_ip_addr.
     * @param cam_ip_addr   The IP address we're interested in.
     * @return The download path associated with cam_ip_addr, or an empty string.
     */
    static const QString loadPath(uint32_t cam_ip_addr)
    {
        int i;
        int found_index = -1;
        QString ret_path;
        QList<DOWNLOAD_PATH> path_list;

//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cam_ip_addr:" << Qt::showbase << Qt::hex << cam_ip_addr;

        // Save the current settings.
        QSettings settings;

        settings.beginGroup("PathInfo");

        // Load the existing array.
        int list_size = settings.beginReadArray("download_list");
        for (i = 0; i < list_size; ++i)
        {
            DOWNLOAD_PATH entry;

            // Let the system know which index we're working with.
            settings.setArrayIndex(i);

            // Read this entry.
            entry.ip_addr = settings.value("ip_addr").toUInt();
            entry.path = settings.value("path").toString();
            entry.last_accessed = settings.value("access_time").toTime();

            // Put it on the list.
            path_list.append(entry);

//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "i:" << i << "entry.ip_addr:" << Qt::showbase << Qt::hex << entry.ip_addr;

            // Is this a match to our IP address?
            if (entry.ip_addr == cam_ip_addr)
            {
                // Remember this entry
                found_index = path_list.size() - 1;
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "found_index:" << found_index;
            }
        }
        settings.endArray();

        // Did we find a result?
        if (found_index >= 0)
        {
            // Remember this path to return.
            ret_path = path_list.at(found_index).path;

            // Update this entrys access time.
            path_list[found_index].last_accessed = QTime::currentTime();

            // Save off the list.
            // NOTE: This might be overkill, re-writing the entire list. I am just not
            //       certain how growing the list would affect the writing. I know this
            //       will work, and I don't have time for optimization at this point.
            settings.beginWriteArray("download_list");
            for (i = 0; i < path_list.size(); ++i)
            {
                settings.setArrayIndex(i);
                settings.setValue("ip_addr", path_list.at(i).ip_addr);
                settings.setValue("path", path_list.at(i).path);
                settings.setValue("access_time", path_list.at(i).last_accessed);
            }
            settings.endArray();
        }
        settings.endGroup();

        // If no path was found, return our default.
        if (ret_path.isEmpty())
        {
            ret_path = QStandardPaths::standardLocations(QStandardPaths::MoviesLocation).at(0);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Picking the standard path:" << ret_path;
        }

        // Return our result.
        return ret_path;
    }

    /*!
     * @brief Save the download path to our settings database.
     * @param cam_ip_addr         The IP address to associate the download_base_path with.
     * @param download_base_path  The new path we want to use when downloading videos.
     */
    static void savePath(uint32_t cam_ip_addr, const QString& download_base_path)
    {
        int i;
        int found_index = -1;
        QString the_path;
        QList<DOWNLOAD_PATH> path_list;

        // Make sure the path is properly formatted.
        the_path = QDir::fromNativeSeparators(download_base_path);
        if (!the_path.endsWith(QLatin1Char('/')))
            the_path.append(QLatin1Char('/'));

        // Save the current settings.
        QSettings settings;
        settings.beginGroup("PathInfo");

        // Load the existing array.
        int list_size = settings.beginReadArray("download_list");
        for (i = 0; i < list_size; ++i)
        {
            DOWNLOAD_PATH entry;

            // Let the system know which index we're working with.
            settings.setArrayIndex(i);

            // Read this entry.
            entry.ip_addr = settings.value("ip_addr").toUInt();
            entry.path = settings.value("path").toString();
            entry.last_accessed = settings.value("access_time").toTime();

            // Put it on the list.
            path_list.append(entry);

            // Is this a match to our IP address?
            if (entry.ip_addr == cam_ip_addr)
                found_index = i;
        }
        settings.endArray();

        // Did we find our IP address in the list?
        if (found_index >= 0)
        {
            // Modify the entiry.
            path_list[found_index].path = the_path;
            path_list[found_index].last_accessed = QTime::currentTime();
        }

        // It wasn't already in the list, so can we add it?
        else if (path_list.size() < MAX_PATH_ENTRIES)
        {
            DOWNLOAD_PATH entry;

            // Create a new entry to append to the list.
            entry.ip_addr = cam_ip_addr;
            entry.path = the_path;
            entry.last_accessed = QTime::currentTime();
            path_list.append(entry);
        }

        // Find out which entry in the list we can replace.
        else
        {
            int oldest_entry;
            QTime oldest_time;

            // Read up the index of the oldest entry. Default to 0.
            oldest_time = QTime::currentTime();
            oldest_entry = 0;

            // Loop through all the entries to find the oldest.
            for (i = 0; i < path_list.size(); ++i)
            {
                if (path_list.at(i).last_accessed < oldest_time)
                {
                    oldest_entry = i;
                    oldest_time = path_list.at(i).last_accessed;
                }
            }

            // Replace the oldest entry with the new value.
            path_list[oldest_entry].ip_addr = cam_ip_addr;
            path_list[oldest_entry].path = the_path;
            path_list[oldest_entry].last_accessed = QTime::currentTime();
        }

        // Save off the list.
        // NOTE: This might be overkill, re-writing the entire list. I am just not
        //       certain how growing the list would affect the writing. I know this
        //       will work, and I don't have time for optimization at this point.
        settings.beginWriteArray("download_list");
        for (i = 0; i < path_list.size(); ++i)
        {
            settings.setArrayIndex(i);
            settings.setValue("ip_addr", path_list.at(i).ip_addr);
            settings.setValue("path", path_list.at(i).path);
            settings.setValue("access_time", path_list.at(i).last_accessed);
        }
        settings.endArray();

        // All done with the group.
        settings.endGroup();
    }
};

}   // End of namespace SAVE_DLG.

using namespace SAVE_DLG;

//--------------------------------------------------------------------
// Constant QString Definitions Global to Application
//--------------------------------------------------------------------
const QString SaveDialog::SAVE_SETTINGS_SECTION     = QStringLiteral("SaveSettings");
const QString SaveDialog::SETTINGS_AVI_SAVE_JPEG    = QStringLiteral("save_avi_jpeg");
const QString SaveDialog::SETTINGS_MP4_H264_QUALITY = QStringLiteral("mp4_quality");
const QString SaveDialog::SETTINGS_MP4_H264_SPEED   = QStringLiteral("mp4_speed");
const QString SaveDialog::SETTINGS_MP4_H264_ALLKEY  = QStringLiteral("mp4_allkey");

/*!
 * @brief The SaveDialog constructor.
 * @param ip_addr   IP address of the camera we're interested in.
 * @param parent    Our parent widget.
 */
SaveDialog::SaveDialog(uint32_t ip_addr, QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::SaveDialog)
{
    // Setup our UI.
    ui->setupUi(this);

    // Want the path which is read-only to be enabled so it's not greyed out.
    ui->txt_path->setEnabled(true);

    // Initialize the default values.
    set_jpeg_quality(DEFAULT_JPEG_QUAL);
    reg_jpeg_quality = DEFAULT_JPEG_QUAL;

    // Make sure the quality for MP4 video has minimum set.
    ui->sld_h264_quality->setMinimum(FAS_VIDEO::MINIMUM_H264_QUALITY_PCT);

    // Setup the MP4 saves based on any save settings we have.  It will set
    // them up to the default if no values have ever been set by the user.
    {
        FAS_VIDEO::H264_SPEEDS h264Speed;
        uint32_t h264Quality;
        bool h264AllKeys;
        SaveDialog::getH264Settings(h264Speed, h264Quality, h264AllKeys);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "h264Quality:" << h264Quality;
        set_h264_quality((int)h264Quality);
        set_h264_speed(h264Speed);
        set_h264_allkeyframe(h264AllKeys);    // As a default this is always unchecked!
    }

    // Hide all the optional stuff.
    ui->grp_h264_options->setVisible(false);
    ui->grp_save_options->setVisible(false);

    // Hide the multicamera controls until we know we're multicamera enabled.
    ui->chk_use_camera_range->setVisible(false);

    // Initialize our variables.
    allow_overlay = false;
    allow_raw = true;
    allow_xml = true;
    allow_planar = false;
    allow_query_cam = false;    // This will be set to query_cam in init_controls().
    autosave_mode = false;
    cap_transfer_mode = false;
    cam_ip_addr = ip_addr;
    cam_connected = false;
    clip_start = 0;
    clip_end = 0;
    fflr_enabled = false;
    initialized = false;
    ltr_enabled = false;
    multicam_mode = false;
    multicam_mixed = false;
    metadata_cfg_selected = FI_SAVER::DEFAULT_SAVE_METADATA_CFG;
    firstJpegRegRead = 0;
    view_only = false;

    // Initialize our signals/slots.
    connect(ui->chk_tag_name, &QCheckBox::stateChanged, this, &SaveDialog::do_chk_tag_label);
    connect(ui->chk_comment, &QCheckBox::stateChanged, this, &SaveDialog::do_chk_comment);
    connect(ui->eb_comment, &QLineEdit::textChanged, this, &SaveDialog::do_comment_changed);
    connect(ui->chk_overlay, &QCheckBox::clicked, this, &SaveDialog::do_chk_overlay);
    connect(ui->rb_pc, &QRadioButton::toggled, this, &SaveDialog::do_pc_toggle);
    connect(ui->cb_devices, QOverload<int>::of(&QComboBox::currentIndexChanged), this, &SaveDialog::device_selected);
    connect(ui->cb_devices, QOverload<int>::of(&QComboBox::highlighted), this, &SaveDialog::device_selected);
    connect(ui->rb_devices, &QRadioButton::clicked, this, &SaveDialog::do_device_select);
    connect(ui->cb_save_type, QOverload<int>::of(&QComboBox::currentIndexChanged), this, &SaveDialog::do_save_type_changed);
    connect(ui->cb_tif_raw, &QCheckBox::clicked, this, &SaveDialog::do_raw_select);
    connect(ui->cb_metadata_cfg, &QCheckBox::clicked, this, &SaveDialog::do_metadata_cfg_select);
    connect(ui->chk_pfmd, &QCheckBox::clicked, this, &SaveDialog::do_pfmd_select);
    connect(ui->chk_xml, &QCheckBox::clicked, this, &SaveDialog::do_xml_select);
    connect(ui->pb_browse, &QPushButton::clicked, this, &SaveDialog::do_browse);
    connect(ui->chk_use_camera_range, &QCheckBox::clicked, this, &SaveDialog::do_cam_range);
    connect(ui->rb_jpg, &QRadioButton::clicked, this, &SaveDialog::do_jpg_select);
    connect(ui->rb_bmp, &QRadioButton::clicked, this, &SaveDialog::do_bmp_select);
    connect(ui->spin_jpeg, QOverload<int>::of(&QSpinBox::valueChanged), this, &SaveDialog::do_jpeg_quality_changed);
    connect(ui->sld_jpeg, &QSlider::valueChanged, this, &SaveDialog::do_jpeg_quality_changed);
    connect(ui->spin_h264_quality, QOverload<int>::of(&QSpinBox::valueChanged), ui->sld_h264_quality, QOverload<int>::of(&QSlider::setValue));
    connect(ui->sld_h264_quality, QOverload<int>::of(&QSlider::valueChanged), ui->spin_h264_quality, QOverload<int>::of(&QSpinBox::setValue));
    connect(ui->sld_h264_speed, QOverload<int>::of(&QSlider::valueChanged), this, &SaveDialog::do_h264_speed_changed);
    connect(ui->cb_keyframe_all, &QCheckBox::clicked, this, &SaveDialog::do_h264_allkeyframes_select);

    // Update the layout.
    adjustSize();
    update();
}

/*!
 * @brief SaveDialog destructor.
 */
SaveDialog::~SaveDialog()
{
    // Cleanup our UI object.
    delete ui;

    // Disconnect our signals/slots.
    disconnect_camera();
}

/*!
 * @brief Find the path associated with the cam_ip_addr.
 * @param cam_ip_addr   The IP address we're interested in.
 * @return The download path associated with cam_ip_addr, or an empty string.
 */
const QString SaveDialog::loadPath(uint32_t cam_ip_addr)
{
    // Pass it on.
    return PathRepository::loadPath(cam_ip_addr);
}

/*!
 * @brief Save the download path to our settings database.
 * @param cam_ip_addr         The IP address to associate the download_base_path with.
 * @param download_base_path  The new path we want to use when downloading videos.
 */
void SaveDialog::savePath(uint32_t cam_ip_addr, const QString& download_base_path)
{
    // Pass it on.
    PathRepository::savePath(cam_ip_addr, download_base_path);
}


/*!
 * @brief Return true if the user has selected to create an XML file.
 */
bool SaveDialog::create_xml()
{
    return (allow_xml && ui->chk_xml->isChecked());
}

/*!
 * @brief Invoked by the user of the dialog to enable the overlay control(s).
 * @param do_enable True to enable, false to disable.
 */
void SaveDialog::enable_overlay_merge(bool do_enable)
{
    // Remember the flag.
    allow_overlay = do_enable;

    // Setup the controls.
    ui->chk_overlay->setEnabled(do_enable);
    ui->chk_overlay->setChecked(do_enable);

    // If the caller wants to enable overlay merge, then they are specifying a
    // save to the PC.
    if (do_enable && !ui->rb_pc->isChecked())
    {
        // Set the destination as the PC.
        ui->rb_pc->setChecked(true);
        ui->rb_devices->setChecked(false);
        fill_save_type(MEDIA_LOCAL_PC);
    }
    else
    {
        // Process as though the user clicked the overlay enable button.
        do_chk_overlay();
    }
}

/*!
 * @brief Populate the save type combobox based on the media_sel value.
 * @param media_sel     Which media is selected.
 */
void SaveDialog::fill_save_type(uint32_t media_sel)
{
    int cur_selection;
    bool series3;
    bool cap_allowed;
    bool mp4_allowed_from_series3 = false;
    bool camera_is_connected = false;    // Ticket #450: can't stream planar if not connected to a camera.
    CameraMDI   *cam_ptr = qobject_cast<CameraMDI *>(parent());

#ifdef MP4_DEBUG
    mp4_allowed_from_series3 = true;
#endif

    // We might not be connected to a camera...
    if (cam_ptr == nullptr)
    {
        PlaybackMDI *pb_ptr = qobject_cast<PlaybackMDI *>(parent());

        series3 = pb_ptr->isSeries3();
        cap_allowed = false;
        mp4_allowed_from_series3 = true;
    }
    else
    {
        series3 = cam_ptr->is_series3();
        cap_allowed = cam_ptr->cap_is_allowed();

        // Make one additional check on the ip address to make sure we are connected to a camera.
        if (cam_ip_addr != 0)
        {
            camera_is_connected = true;
        }
    }

    // Remember the current selection, so we can restore it later.
    if (ui->cb_save_type->count() > 0)
    {
        int index;

        index = ui->cb_save_type->currentIndex();
        cur_selection = ui->cb_save_type->itemData(index).toInt();
    }
    else
        cur_selection = -1;

    // Turn off the update signal from the combobox.
    ui->cb_save_type->blockSignals(true);

    // Reset the combobox.
    ui->cb_save_type->clear();

    // Series3 cameras have more restrictions...
    if (series3)
    {
        // Customize the allowed local save to PC formats.
        if (media_sel == MEDIA_LOCAL_PC)
        {
            // Add the streaming types as appropriate.
            ui->cb_save_type->addItem(FiSaver::FORMAT_AVI_STR, FI_SAVER::STREAM_TYPE::AVI);
            if (mp4_allowed_from_series3)
            {
                ui->cb_save_type->addItem(FiSaver::FORMAT_MP4_STR, FI_SAVER::STREAM_TYPE::MP4);
            }
            if ((cur_selection < 0) || (cur_selection >= FI_SAVER::STREAM_TYPE::FI_CAP))
            {
                cur_selection = FI_SAVER::STREAM_TYPE::AVI;
            }

            // Add RDI packed, if it is allowed.
            if (FiInternal::factory() != nullptr)
                ui->cb_save_type->addItem(FiSaver::FORMAT_RDI_PACKED_STR, FI_SAVER::STREAM_TYPE::RDI_PACKED);
        }
        // Ticket #256: We don't allow AVI saves to SSD at all.  Remove any
        // possible setup for AVI saves in this case.
        else if (media_sel == GIGE_MEDIA_SSD)
        {
            // Make sure the current selection is valid for this setup.
            if ((cur_selection < 0) || (cur_selection == FI_SAVER::STREAM_TYPE::AVI) || (cur_selection >= FI_SAVER::STREAM_TYPE::MP4))
            {
                if (cap_allowed)
                {
                    cur_selection = FI_SAVER::STREAM_TYPE::FI_CAP;
                }
                else
                {
                    // This is safest bet.
                    cur_selection = FI_SAVER::STREAM_TYPE::JPEG_STACK;
                }
            }
        }
        else
        {
            // Other media types (usb and sdcard) will allow for AVI saving but not MP4.
            ui->cb_save_type->addItem(FiSaver::FORMAT_AVI_STR, FI_SAVER::STREAM_TYPE::AVI);

            // Make sure the current selection is valid for this setup.
            if ((cur_selection < 0) || (cur_selection >= FI_SAVER::STREAM_TYPE::FI_CAP))
            {
                cur_selection = FI_SAVER::STREAM_TYPE::AVI;
            }
        }
    }
    else
    {
        // Allow the user to save anything anywhere.
        ui->cb_save_type->addItem(FiSaver::FORMAT_AVI_STR, FI_SAVER::STREAM_TYPE::AVI);
        ui->cb_save_type->addItem(FiSaver::FORMAT_MP4_STR, FI_SAVER::STREAM_TYPE::MP4);

        // Add RDI packed, if it is allowed.
        if (FiInternal::factory() != nullptr)
            ui->cb_save_type->addItem(FiSaver::FORMAT_RDI_PACKED_STR, FI_SAVER::STREAM_TYPE::RDI_PACKED);

        // If we don't have a previous selection, then use AVI as the default.
        if (cur_selection < 0)
        {
            cur_selection = FI_SAVER::STREAM_TYPE::AVI;
        }
    }

    // We can only save DNG stacks if we're a color sensor.
    if (camera_is_connected)
    {
        // Read the sensor type value from our cache.
        if (cam_ptr && cam_ptr->get_sensor())
            sensor_type = cam_ptr->get_sensor()->get_type();
        else
            sensor_type = 0;
    }

    // We can't use the "FI_SENSOR::is_color()" method here because we might not be connected to a sensor!
    if ((sensor_type == FI_SENSOR::SENSOR_TYPE_LUPA1300_COLOR) ||
        (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR) ||
        (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR) ||
        (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX160_COLOR) ||
        (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX2100_COLOR) ||
        (sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR))
    {
        // We can only support DNGs and PLANAR data if we're not in multicamera
        // mode (and we are a color camera) OR our group of cameras is not mixed
        // (they are all color cameras).
        if (!multicam_mode || !multicam_mixed)
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "allow_raw:" << allow_raw << "chk_overlay:" << (ui->chk_overlay->isChecked() ? "CHECKED" : "UNCHECKED") << (ui->chk_overlay->isEnabled() ? "ENABLED" : "DISABLED");

            // If merging overlays cannot allow DNG or NV12 Planar streams.
            if (!ui->chk_overlay->isEnabled())
            {
                // Only add DNG option if allowing raw images.
                if (allow_raw)
                {
                    ui->cb_save_type->addItem(FiSaver::FORMAT_DNG_STR, FI_SAVER::STREAM_TYPE::DNG_STACK);
                }
                // Only allow nv12 planar image streams if not series 3 camera.
                if (!series3 && camera_is_connected)
                {
                    allow_planar = true;
                }
            }
        }
    }
    // Check for a MONO sensor and see if we can stream PLANAR data.
    else if ((sensor_type == FI_SENSOR::SENSOR_TYPE_LUX13HS_MONO) ||
             (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX160_MONO) ||
             (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX19HS_MONO) ||
             (sensor_type == FI_SENSOR::SENSOR_TYPE_LUX2100_MONO) ||
             (sensor_type == FI_SENSOR::SENSOR_TYPE_LUPA1300_MONO) ||
             (sensor_type == FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO))
    {
        // We can only support PLANAR data if we're not in multicamera
        // mode (and we are a mono camera) OR our group of cameras is not mixed
        // (they are all mono cameras).
        if (!multicam_mode || !multicam_mixed)
        {
            // If merging overlays cannot allow DNG or NV12 Planar streams.
            if (!ui->chk_overlay->isChecked())
            {
                // Only allow Y8 planar image streams if not series 3 camera.
                if (!series3 && camera_is_connected)
                {
                    allow_planar = true;
                }
            }
        }
    }

    // Add in the generic save types.
    ui->cb_save_type->addItem(FiSaver::FORMAT_BMP_STR, FI_SAVER::STREAM_TYPE::BMP_STACK);
    ui->cb_save_type->addItem(FiSaver::FORMAT_JPG_STR, FI_SAVER::STREAM_TYPE::JPEG_STACK);
    ui->cb_save_type->addItem(FiSaver::FORMAT_TIF_STR, FI_SAVER::STREAM_TYPE::TIF_STACK);

    // Are CAP files an option?
    // Series3 cameras have more restrictions...
    if (series3)
    {
        if ((media_sel == GIGE_MEDIA_SSD) && cap_allowed)
        {
            ui->cb_save_type->addItem(FiSaver::FORMAT_CAP_STR, FI_SAVER::STREAM_TYPE::FI_CAP);
        }
    }
    else
    {
        // We cannot save CAP files if we're in LR mode, or we are not connected to a camera.
        if (!ltr_enabled && camera_is_connected)
        {
            uint32_t camera_state;

            // What is the current state of the camera?
            if (cam_ptr)
                camera_state = cam_ptr->read_camera_cache(GIGE_REG_CAMERA_STATE);
            else
                camera_state = GIGE_CAMERA_STATE_IDLE;

            // Is fflr enabled while we are in live?  If so we can only be saving to a CAP Stack.
            if (fflr_enabled && (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO))
            {
                //ui->cb_save_type->addItem(fiSaver::FORMAT_CAP_STR, FI_SAVER::STREAM_TYPE::FI_CAP);
                ui->cb_save_type->addItem(FiSaver::FORMAT_CAP_STR, FI_SAVER::STREAM_TYPE::FI_CAPSTK);
            }
            else
            {
                ui->cb_save_type->addItem(FiSaver::FORMAT_CAP_STR, FI_SAVER::STREAM_TYPE::FI_CAP);
            }
        }
    }

    // Turn on the update signal from the combobox.
    ui->cb_save_type->blockSignals(false);

    // Do we have a previous selection?
    select_save_type(cur_selection);

    // Have the combobox resize.
    ui->cb_save_type->setSizeAdjustPolicy(QComboBox::AdjustToContents);
}

/*!
 * @brief Static public method will return the "SaveSettings" values for
 * MP4 video settings or the default if no settings exist.
 *
 * @param h264Speed     - Will hold SETTINGS_MP4_H264_SPEED value or DEFAULT_H264_SPEED
 * @param h264Quality   - Will hold SETTINGS_MP4_H264_QUALITY value or DEFAULT_H264_QUALITY_PCT
 * @param h264Speed     - Will hold SETTINGS_MP4_H264_ALLKEY value or false
 */
void SaveDialog::getH264Settings(FAS_VIDEO::H264_SPEEDS& h264Speed, uint32_t& h264Quality, bool& h264AllKeys)
{
    // Pull the SaveSettings values out and set up our saver based
    // on those settings.
    QSettings settings;
    settings.beginGroup(SaveDialog::SAVE_SETTINGS_SECTION);

    uint32_t h264SpeedAsInt = settings.value(SaveDialog::SETTINGS_MP4_H264_SPEED, FAS_VIDEO::DEFAULT_H264_SPEED).toUInt();
    h264Quality             = settings.value(SaveDialog::SETTINGS_MP4_H264_QUALITY, FAS_VIDEO::DEFAULT_H264_QUALITY_PCT).toUInt();
    h264AllKeys             = settings.value(SaveDialog::SETTINGS_MP4_H264_ALLKEY, false).toBool();

    // Validate the quality.
    if (h264Quality < FAS_VIDEO::MINIMUM_H264_QUALITY_PCT)
        h264Quality = FAS_VIDEO::MINIMUM_H264_QUALITY_PCT;
    else if (h264Quality > 100)
        h264Quality = 100;

    // Validate the speed.
    // Since it's uint32_t 0 is valid but nothing at or over end of speeds is valid.
    if (h264SpeedAsInt >= FAS_VIDEO::H264_SPEEDS::H264_END_OF_SPEEDS)
        h264SpeedAsInt = FAS_VIDEO::DEFAULT_H264_SPEED;

    h264Speed = (FAS_VIDEO::H264_SPEEDS) h264SpeedAsInt;

    settings.endGroup();
}

/*!
 * @brief Return the absolute frame number for the end of the clip.
 */
uint32_t SaveDialog::get_clip_end()
{
    return (trig_frame + (uint32_t)ui->spin_clip_end->value());
}

/*!
 * @brief Return the absolute frame number for the start of the clip.
 */
uint32_t SaveDialog::get_clip_start()
{
    return (trig_frame + (uint32_t)ui->spin_clip_start->value());
}

/*!
 * @brief Return the user specified comment string, or QString() for none.
 */
const QString SaveDialog::get_comment() const
{
    if (ui->chk_comment->isChecked())
        return ui->eb_comment->text();

    return QString();
}

/*!
 * @brief Return the "label" for the destination media (it's "name").
 *
 * If the destination is to the PC, or we're talking to a series3 camera,
 * then there is no destination "label".
 */
QString SaveDialog::get_dest_label()
{
    QString    media_label;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = qobject_cast<CameraMDI *>(parent());

    // First check the PC radio button. If it's checked, then that's where
    // we are going. Otherwise check the combobox.
    if (ui->rb_pc->isChecked() || camera_ptr->is_series3())
    {
        media_label.clear();
    }
    else
    {
        QJsonObject dev_object = ui->cb_devices->currentData().toJsonObject();

        // Get the device type field.
        media_label = dev_object["label"].toString();
    }

    // Return the selection.
    return media_label;
}

/*!
 * @brief Return the enumeration for the selected media.
 */
uint32_t SaveDialog::get_dest_media()
{
    uint32_t    media_sel;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = qobject_cast<CameraMDI *>(parent());

    // First check the PC radio button. If it's checked, then that's where
    // we are going. Otherwise check the combobox.
    if (ui->rb_pc->isChecked())
    {
        media_sel = MEDIA_LOCAL_PC;
    }
    else if ((camera_ptr == nullptr) || camera_ptr->is_series3())
    {
        QString sel_device;

        // Get the selected text.
        sel_device = ui->cb_devices->currentText();

        // Compare it with known values to determine the device selected.
        if (sel_device.compare(CamExplore::SDCARD_LABEL) == 0)
            media_sel = GIGE_MEDIA_SDCARD;
        else if (sel_device.compare(CamExplore::SSD_LABEL) == 0)
            media_sel = GIGE_MEDIA_SSD;
        else
            media_sel = GIGE_MEDIA_USB;
    }
    else
    {
        QJsonObject dev_object = ui->cb_devices->currentData().toJsonObject();

        // Get the device type field.
        media_sel = (uint32_t)dev_object["type"].toInt();
    }

    // Return the selection.
    return media_sel;
}

/*!
 * @brief Return the user specified destination path, if any.
 */
const QString SaveDialog::get_dest_path() const
{
    CameraMDI   *camera_ptr;
    QJsonObject dev_object;

    // Make sure we have an active camera pointer.
    camera_ptr = qobject_cast<CameraMDI *>(parent());

    // If there is no camera pointer or the PC radio button is checked then
    // we are using a "path".
    if ((camera_ptr == nullptr) || ui->rb_pc->isChecked())
    {
        return ui->txt_path->text();
    }
    else if (camera_ptr->is_series3())
    {
        // For series three cameras we want to return the value in the text edit
        // location of the combo-box.
        return ui->cb_devices->currentText();
    }

    // Get the selected object.
    dev_object = ui->cb_devices->currentData().toJsonObject();

    // Return the mount path (this is where we want to write to).
    return dev_object["mountdir"].toString();
}

/*!
 * @brief Return the JPEG quality value the user selected.
 */
int SaveDialog::get_jpeg_quality()
{
    int quality = ui->sld_jpeg->value();
    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "dlg->sld_jpeg", (uint32_t)quality);
    return quality;
}

/*!
 * @brief Return the playback rate the user has selected.
 */
uint32_t SaveDialog::get_playback_rate()
{
    return (uint32_t)ui->spin_play_fps->value();
}

/*!
 * @brief Return the trigger-relative end of the clip.
 */
int SaveDialog::get_rel_clip_end()
{
    return ui->spin_clip_end->value();
}

/*!
 * @brief Return the trigger-relative start of the clip.
 */
int SaveDialog::get_rel_clip_start()
{
    return ui->spin_clip_start->value();
}

/*!
 * @brief Fill in the elements of the save_params structure.
 * @param[out] save_params   Structure to be filled.
 */
void SaveDialog::get_save_params(FI_SAVER::SAVE_PARAMS& save_params, bool is_series3)
{
    // If the camera connected is a seriesNC with local control then it makes sense
    // to have a planar format drive the video file save operation.
    bool usePlanarInput = false;

    // Fill in the SAVE_PARAMS structure record.
    save_params.clip_end = get_rel_clip_end();
    save_params.clip_start = get_rel_clip_start();
    save_params.clip_end_abs = get_clip_end();
    save_params.clip_start_abs = get_clip_start();
    save_params.jpeg_quality = get_jpeg_quality();
    save_params.h264_quality = get_h264_quality();
    save_params.h264_speed = GetH264Speed();
    save_params.h264_enable_all_keyframes = GetH264EnableAllKeyFrames();
    save_params.create_xml = create_xml();
    save_params.dest_media = get_dest_media();
    save_params.dest_label = get_dest_label();
    save_params.playback_rate = get_playback_rate();
    save_params.dest_path = get_dest_path();
    save_params.ref_frame = trig_frame;
    save_params.comment = get_comment();

    save_params.streamType = GetStreamType();
    save_params.useRawFormat = false;
    save_params.save_metadata_cfg = ui->cb_metadata_cfg->isChecked();
    save_params.save_pfmd = ui->chk_pfmd->isChecked();

    switch (save_params.streamType)
    {
    case FI_SAVER::STREAM_TYPE::MP4:
        // When creating an MP4 video stream the input files coming
        // from a camera are expected to be a JPEG 100% quality
        // and because the MP4 video stream writer's encoder expects
        // a Windows Bitmap uncompressed frame of data this JPEG
        // image will be translated into a BMP image using fiPixMap.
        // This is of note because it is one place where the input
        // and output image formats are different.

        // If the local control is true this FasMotion implementation is
        // running on a new camera system.  In this case the streaming
        // feed should use the uncompressed planar stream for best quality
        // of the resultant MP4 video file.
        //if (MainWindow::local_control())
        //{
        //QLOG_INFO() << __PRETTY_FUNCTION__ << "MainWindow::local_control() is TRUE";
        // If sensor and conditions support planar input go ahead and
        // allow it to stream.

        // We used to use local control (if (MainWindow::local_control()))
        // to determine if we want to stream planar data from the camera server.
        // Now it's more sophisticated.  If we get here and the allow_planar
        // flag is true and we are not merging overlays then we will ask
        // for BPP (PLANAR) data from connected camera.
        if (allow_planar && !merge_overlays())
        {
            QLOG_INFO() << __PRETTY_FUNCTION__ << "Requesting Planar Data Input Stream.";
            usePlanarInput = true;
        }

        // When streaming input video streams if they happen to be JPEG the quality
        // we want to receive is 100 (as close to lossless as possible).
        save_params.jpeg_quality = MP4_JPEG_QUALITY;

        // #MP4IMAGEINPUT SIG - SaveDialog::get_save_params - sets usePlanarInput.
        //usePlanarInput = true;    // Force Planar testing on.
        //usePlanarInput = false;     // Force Planar testing off.

        // If using the planar input the input and output type will be different.
        if (usePlanarInput)
        {
            save_params.imagesInputType = FI_SAVER::IMAGE_TYPE::BMP_PLANAR;
            save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::BMP_PLANAR;
        }
        else // All remote camera systems or series 3 and earlier cameras will stream JPEG images.
        {
            save_params.imagesInputType = FI_SAVER::IMAGE_TYPE::JPEG;
            save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::BMP;
        }
        break;
    case FI_SAVER::STREAM_TYPE::AVI:
        if (ui->rb_bmp->isChecked())
        {
            save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::BMP;
        }
        else    // DEFAULT will go to JPEG.
        {
            save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::JPEG;
            save_params.jpeg_quality = (int)reg_jpeg_quality;
        }
        QLOG_INFO() << __PRETTY_FUNCTION__ << "AVI SELECTED WITH " << FiSaver::imageTypeToQString(save_params.imagesInputType);
        break;
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::BMP;
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::JPEG;
        save_params.jpeg_quality = (int)reg_jpeg_quality;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::TIF;
        break;
    case FI_SAVER::STREAM_TYPE::TIF_RAW_STACK:
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        // The setting allow_raw should be true in this case and this should not happen but if it
        // does then force the setting to be regular TIF.
        if (!allow_raw)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "RAW STACK SELECTED BUT NOT SUPPORTED!!!";

            save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::TIF;
            save_params.streamType = FI_SAVER::STREAM_TYPE::TIF_STACK;
        }
        else
        {
            // If was a TIF raw stack selection then set to TIF_RAW.
            if (save_params.streamType == FI_SAVER::STREAM_TYPE::TIF_RAW_STACK)
                save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::TIF_RAW;
            else    // Must have been DNG stack set to DNG.
                save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::DNG;
            save_params.useRawFormat = true;
        }
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        if (is_series3)
        {
            save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::TIF_RAW;
            save_params.useRawFormat = true;
        }
        else
        {
            save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::PACKED;
            save_params.useRawFormat = true;
        }
        break;
    default:
        // Cap files fall into here.
        save_params.imagesInputType = save_params.imagesOutputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
        break;
    }

    save_params.tag_label = get_tag_label();
    save_params.trigger_relative = true;
    save_params.use_camera_name = use_camera_name();
    save_params.use_camera_range = use_camera_range();
    save_params.merge_overlays = merge_overlays();

    // Run quick check on the jpeg_quality to make sure it leaves here valid.
    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "save_params.jpeg_quality", (uint32_t)save_params.jpeg_quality);
}

/*!
 * @brief Return the user selected stream type.
 */
FI_SAVER::STREAM_TYPE SaveDialog::GetStreamType()
{
    int index;
    FI_SAVER::STREAM_TYPE streamType;

    // Get the format information from the combobox entry.
    index = ui->cb_save_type->currentIndex();
    streamType = static_cast<FI_SAVER::STREAM_TYPE>(ui->cb_save_type->itemData(index).toInt());
    if (streamType == FI_SAVER::STREAM_TYPE::TIF_STACK)
    {
        if (allow_raw && ui->cb_tif_raw->isChecked() && !ui->chk_overlay->isChecked())
            streamType = FI_SAVER::STREAM_TYPE::TIF_RAW_STACK;
    }

    // Return the stream type the user selected.
    return streamType;
}

/*!
 * @brief Return the user selected H.264 speed setting.
 */
FAS_VIDEO::H264_SPEEDS SaveDialog::GetH264Speed()
{
    int                     index;
    FAS_VIDEO::H264_SPEEDS  speedSetting;

    // Translate the index into the H.264 speed setting.
    index = ui->sld_h264_speed->value();

    // If the index exceeds the H.264 settings set it to the default.
    if (index >= FAS_VIDEO::H264_SPEEDS::H264_END_OF_SPEEDS)
        speedSetting = FAS_VIDEO::DEFAULT_H264_SPEED;
    else
        speedSetting = (FAS_VIDEO::H264_SPEEDS) index;

    return speedSetting;
}

/*!
 * @brief Return the user selected H.264 enable all keyframe setting.
 */
bool SaveDialog::GetH264EnableAllKeyFrames()
{
    bool enableAllKeyFrames = false;

    if (ui->cb_keyframe_all->isChecked())
        enableAllKeyFrames = true;
    return enableAllKeyFrames;
}

/*!
 * @brief Return the user selected H.264 frame quality setting.
 */
uint32_t SaveDialog::get_h264_quality()
{
    uint32_t pctQuality;

    // Translate the index into the H.264 speed setting.
    pctQuality = (uint32_t) ui->sld_h264_quality->value();

    return pctQuality;
}

/*!
 * @brief Return the user specified tag.
 */
const QString SaveDialog::get_tag_label() const
{
    if (ui->chk_tag_name->isChecked())
        return ui->eb_tag_label->text();
    return QString();
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_AVAIL_MEDIA register value.
 */
void SaveDialog::handle_json_media(const QJsonDocument& jsonMedia)
{
    int item_index;
    double dev_size;
    QString base_label;
    QString device_label;
    QString device_sel;
    QJsonObject mediaObj;

    // Get the object from the document.
    mediaObj = jsonMedia.object();

    // Block updates from the combobox.
    ui->cb_devices->blockSignals(true);

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
        device_sel = ui->cb_devices->currentText();

    // Clear the combobox.
    ui->cb_devices->clear();

    // Add back in found devices starting with the SDCard.
    if (mediaObj.contains("sdcard") && mediaObj["sdcard"].isArray())
    {
        QJsonArray dev_array = mediaObj["sdcard"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an sdcard device.
            dev_object["type"] = GIGE_MEDIA_SDCARD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("SDcard %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // Check for NVME SSD devices.
    if (mediaObj.contains("nvme") && mediaObj["nvme"].isArray())
    {
        QJsonArray dev_array = mediaObj["nvme"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an NVME device.
            dev_object["type"] = GIGE_MEDIA_SSD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                if (dev_object["internal"].toBool())
                    device_label = QString("SSD %1GB [Internal]").arg(dev_size, 0, 'f', 0);
                else
                    device_label = QString("SSD %1GB [External]").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                if (dev_object["internal"].toBool())
                    device_label = base_label + QString(" [%1GB] [Internal]").arg(dev_size, 0, 'f', 0);
                else
                    device_label = base_label + QString(" [%1GB] [External]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // Check for USB drives.
    if (mediaObj.contains("usb") && mediaObj["usb"].isArray())
    {
        QJsonArray dev_array = mediaObj["usb"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an USB device.
            dev_object["type"] = GIGE_MEDIA_USB;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("USB %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // If there is no external media then disable the devices controls.
    if (ui->cb_devices->count() == 0)
    {
        // Make sure we're disabled, and force the PC selection.
        ui->cb_devices->setEnabled(false);
        ui->rb_devices->setEnabled(false);
        ui->rb_pc->setChecked(true);
    }
    else
    {
        int sel_index = 0;

        // Make sure we're enabled.
        ui->cb_devices->setEnabled(true);
        ui->rb_devices->setEnabled(true);

        // Attempt to re-select the old value.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);
            if (sel_index < 0)
                sel_index = 0;
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);
    }

    // Allow signals to flow again.
    ui->cb_devices->blockSignals(false);

    // Now we can restore our settings.
    restore_settings(false);

    // Enable the dialog.
    enable_controls(true);

    // We're done with the initialization.
    disconnect_camera();
}

/*!
 * @brief Initialize our dialog boxes controls.
 *
 * @param query_cam Should the camera information be queried from
 *                  the camera in order to setup the dialog controls?
 */
void SaveDialog::init_controls(bool query_cam)
{
    bool wait_for_response = false;
    bool read_OK;
    bool is_series3 = false;
    bool is_seriesNC = false;
    bool restoreSaveSettings = false;

    // Persist the query_cam state in this object.
    allow_query_cam = query_cam;

    // By default, hide the autosave button.
    ui->pb_autosave->setVisible(false);

    // Are we connected to a camera?
    if (cam_ip_addr != 0)
    {
        CameraMDI   *camera_ptr = qobject_cast<CameraMDI *>(parent());

        if (camera_ptr == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Has cam_ip_addr but camera_ptr is nullptr!";
        }

        // Are we connected to a series3 camera?
        if ((camera_ptr == nullptr) || camera_ptr->is_series3())
        {
            // Remember that we're a series3 camera.
            is_series3 = true;

            // What media is available on the camera?
            uint32_t reg_value = read_camera_cache(GIGE_REG_AVAIL_MEDIA);
            update_media(reg_value);
        }
        else
        {
            // Mark the fact that a seriesNC Camera is connected.
            is_seriesNC = true;
            wait_for_response = query_cam;

            // If we're allowed, ask the camera for the media information.
            if (query_cam)
            {
                // Ask the camera for the available media. This will be caught
                // by the camera object, which will start the ball rolling.
                camera_ptr->start_camera_read_reg(GIGE_REG_AVAIL_MEDIA);
            }
        }

        // What is the JPEG quality factor used on the camera?
        uint32_t reg_value = read_camera_cache(GIGE_REG_JPEG_QUALITY, &read_OK);
        if (read_OK)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "read_camera_cache() GIGE_REG_JPEG_QUALITY:" << reg_value;
            set_jpeg_quality((int)reg_value);
            reg_jpeg_quality = reg_value;   // Set our register value from the camera.
        }
        else if (camera_ptr != nullptr)
        {
            // If we're allowed to query the camera, then that's where we need to get the JPEG
            // quality factor. Otherwise, we'll just use the default value.
            if (query_cam)
            {
                // Ask the camera for the qfactor.
                camera_ptr->start_camera_read_reg(GIGE_REG_JPEG_QUALITY);
            }
        }
    }

    // Default to the local PC for our destination.
    ui->rb_pc->setEnabled(true);
    ui->rb_pc->setChecked(true);
    ui->pb_browse->setEnabled(true);
    ui->rb_devices->setChecked(false);

    // Setup a regex to validate the tag label, so the user can't type in
    // an invalid value.
    QRegularExpression rx("[-_.A-Za-z0-9]+");
    QRegularExpressionValidator *validator = new QRegularExpressionValidator(rx, ui->eb_tag_label);
    ui->eb_tag_label->setValidator(validator);

    // Initialize the ranges of our spin controls.
    init_ranges();

    // Initialize our save type combo.
    fill_save_type(MEDIA_LOCAL_PC);

    // Restore our settings. For series3 cameras, we ALWAYS restore our settings from
    // previous saves, if possible.
    //
    // For seriesNC cameras we do as we're asked (query_cam). Normally, we will have a
    // pending GIGE_REG_MEDIA_JSON read request (see above). Somethimes, this simply isn't
    // possible. If we can't do the read from the camera, then we'll try to restore from
    // previous saves.
    if (is_series3)
    {
        // An older series 3 camera is connected.  Query to restore any saved
        // save settings without looking at the query_cam state.
        restoreSaveSettings = true;
    }
    else if (is_seriesNC)
    {
        // We are working with a seriesNC camera, but we may not be allowed to ask the
        // camera for the parameters (query_cam). This would be the case, as an example,
        // if we were transcoding from a CAP file.
        restoreSaveSettings = !query_cam;
    }
    else
    {
        // Restore based on the query state passed into function.
        restoreSaveSettings = query_cam;
    }

    // Reset the dialog settings based on any stored save settings.
    if (restoreSaveSettings)
        restore_settings(is_series3);

    // For series3 cameras, we are now initialized. Enable the dialog.
    // Otherwise, enable the dialog if we're not waiting on a response.
    if (is_series3 || !wait_for_response)
    {
        // Go ahead and enable the controls
        enable_controls(true);
    }

    // Make sure that we have a text entry for the media and if not take the first one in
    // the list.
    if (ui->cb_devices->count() > 0)
    {
        int cur_device_index = ui->cb_devices->currentIndex();

        // Make sure we have an index selected, if not use zero.
        if (cur_device_index == -1)
            cur_device_index = 0;

        // Make sure to set the current index so the text will display.
        ui->cb_devices->blockSignals(true);
        ui->cb_devices->setCurrentIndex(cur_device_index);
        ui->cb_devices->blockSignals(false);
    }

    // Remember that we're initialized, as long as we're not waiting for a response..
    initialized = !wait_for_response;
}

/*!
 * @brief Initialize the ranges allowed for the stop and end clip controls
 *        as well as the playback frame rate control.
 */
void SaveDialog::init_ranges()
{
    int         rel_min;
    int         rel_max;
    QSettings   settings;

    // Initialize our spinner for the play FPS rate.
    ui->spin_play_fps->setMinimum(1);
    ui->spin_play_fps->setMaximum(MAX_AVI_PLAY_RATE);

    settings.beginGroup(SAVE_SETTINGS_SECTION);
    ui->spin_play_fps->setValue(settings.value("playback_rate", DEF_AVI_PLAY_RATE).toInt());
    settings.endGroup();

    // Check to see if we're in LTR mode.
    if (cam_ip_addr != 0)
    {
        uint32_t ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);
        ltr_enabled = (ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;
        fflr_enabled = (ltr_mode & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR;
    }
    else
    {
        // We're not connected to a camera, so we can't be in LR mode or FasFireLR (FFLR).
        ltr_enabled = false;
        fflr_enabled = false;
    }

    // Are we in autosave mode? If we are, then we need to calculate what the
    // trigger "should" be, otherwise we can read what the trigger actually is.
    if (autosave_mode)
    {
        uint32_t        trig_pos;
        uint32_t        width;
        uint32_t        height;
        CameraMDI       *cam_ptr = qobject_cast<CameraMDI *>(parent());

        // How many frames are we talking about here? Read all the values we'll need
        // for our calculations.
        width = read_camera_cache(GIGE_REG_WIDTH);
        height = read_camera_cache(GIGE_REG_HEIGHT);

        if (cam_ptr)
        {
            num_frames = cam_ptr->calc_max_frames();
        }
        else
        {
            PlaybackMDI *pb_ptr = qobject_cast<PlaybackMDI *>(parent());
            num_frames = ClientMath::calc_max_total_frames(cam_ip_addr, width, height, ltr_enabled, pb_ptr->isSeries3());
        }

        // Get the trigger position.
        trig_pos = read_camera_cache(GIGE_REG_TRIGGER_REQ_POS);

        // Is the trigger position frame number or percent?
        if ((trig_pos & GIGE_TRIGGER_REQ_FRAME_MASK) == GIGE_TRIGGER_REQ_FRAME_MASK)
        {
            // Get the actual trigger frame number.
            trig_frame = trig_pos & GIGE_TRIGGER_REQ_POS_MASK;
        }
        else
        {
            // It's percent based, so calculate which frame it should be.
            trig_frame = ((num_frames - 1) * trig_pos) / 100;
        }

        // Since trigger changed, all the other displayed fields need to change as well
        // since they are relative to the trigger.
        rel_min = 0 - (int)trig_frame;
        rel_max = (int)(num_frames - trig_frame) - 1;
    }
    else if (cam_ip_addr != 0)
    {
        // How many frames did we actually capture.
        num_frames = read_camera_cache(GIGE_REG_PART_NUM_FRAMES);

        // Read what the trigger actually is.
        trig_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

        // Since trigger changed, all the other displayed fields need to change as well
        // since they are relative to the trigger.
        rel_min = 0 - (int)trig_frame;
        rel_max = (int)(num_frames - trig_frame) - 1;
    }
    else
    {
        // Must be transcoding a captured video. If we do not have clip start/end values, use what we know.
        if (clip_start == clip_end)
        {
            // Hopefully we have number of frames and the trigger frame.
            rel_min = 0 - (int)trig_frame;
            rel_max = (int)(num_frames - trig_frame) - 1;
        }
        else
        {
            // Recalculate what our min/max values are.
            rel_min = (int)(clip_start - trig_frame);
            rel_max = (int)(clip_end - trig_frame);
        }
    }

    ui->spin_clip_start->blockSignals(true);
    ui->spin_clip_end->blockSignals(true);
    ui->spin_clip_start->setMinimum(rel_min);
    ui->spin_clip_end->setMinimum(rel_min + 1);
    ui->spin_clip_start->setMaximum(rel_max - 1);
    ui->spin_clip_end->setMaximum(rel_max);
    ui->spin_clip_start->blockSignals(false);
    ui->spin_clip_end->blockSignals(false);
}

bool SaveDialog::merge_overlays()
{
    return ui->chk_overlay->isChecked();
}

/*!
 * @brief Utility method to read the camera's register cache. The assumption
 *        being made is that the register value has already been loaded from
 *        the camera.
 *
 * @param reg_addr      The address of the register we're interested in.
 * @param ok            [OPTIONAL] pointer to a bool to indicate if the register is in cache.
 *
 * @return The register value or 0 if an error (*ok == false).
 */
uint32_t SaveDialog::read_camera_cache(uint32_t reg_addr, bool *ok)
{
    FI_comm     *comm_ptr;

    // Make sure we have a camera's IP address.
    if (cam_ip_addr == 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cam_ip_addr is 0!";
        if (ok != nullptr)
            *ok = false;
        return 0;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";

        if (ok != nullptr)
            *ok = false;
        return 0;
    }

    // Return the cached value.
    return comm_ptr->get_reg(cam_ip_addr, reg_addr, ok);
}

/*!
 * @brief Restore any saved settings so that we can present the user with
 *        the settings last used.
 */
void SaveDialog::restore_settings(bool series3)
{
    bool        use_saver;
    QSettings   settings;
    QString     download_path;
    QString     tag_str;
    QString     comment_str;
    QString     media_label;
    int         save_type = FI_SAVER::STREAM_TYPE::NO_STREAM;
    int         save_type_found = FI_SAVER::STREAM_TYPE::NO_STREAM;
    bool        use_jpeg_codec;
    CameraMDI   *camera_ptr = nullptr;

    // Whether from saver or settings we need these variables in this scope.
    FAS_VIDEO::H264_SPEEDS h264Speed;
    uint32_t h264Quality;
    bool h264AllKeys;

    // Are we connected to a camera AND is it ok to query that camera if so.
    if (cam_ip_addr != 0)
    {
        // Make sure we have the camera object and it's associated saver object.
        // Addition of "allow_query_cam" addresses Ticket#258 and Ticket#66
        // in the case where we aren't allowed to query the camera as in the
        // case of transfer CAP file data we want to load the path from our
        // saved settings.
        camera_ptr = qobject_cast<CameraMDI *>(parent());
        if (camera_ptr && camera_ptr->get_saver() && allow_query_cam)
        {
            // Initialize our path to what we had previously. Note that this may be empty!
            download_path = camera_ptr->get_saver()->getBasePath();
        }
    }

    // Verify that we've initialized the download path.
    if (download_path.isEmpty())
    {
        uint32_t ip_addr = 0;

        // Make sure we have a camera attached.
        if (camera_ptr != nullptr)
            ip_addr = camera_ptr->get_ip_addr();

        // Initialize the default path based on the saved settings.
        download_path = SaveDialog::loadPath(ip_addr);
    }

    settings.beginGroup(SAVE_SETTINGS_SECTION);

    // Check to see if we should use the settings values or what is in the saver object.
    // Addition of "allow_query_cam" addresses Ticket#258 and Ticket#66
    if (camera_ptr && camera_ptr->get_saver() && allow_query_cam)
    {
        // Restore the format the user last selected for saving the recording.
        save_type = camera_ptr->get_saver()->getStreamType();
        use_saver = (save_type != FI_SAVER::STREAM_TYPE::NO_STREAM);
    }
    else
    {
        // Use the settings since we don't have a saver object.
        use_saver = false;
    }

    // Restore the selected media.
    if (series3)
    {
        media_label.clear();
    }
    else
    {
        if (use_saver)
            media_label = camera_ptr->get_saver()->getDestLabel();
        if (media_label.isEmpty())
            media_label = settings.value("dest_label").toString();
    }

    // Determine the last media saved to, and configure what we can save accordingly.
    if (use_saver)
        set_dest_media(camera_ptr->get_saver()->getDestMedia(), media_label);
    else
        set_dest_media(settings.value("dest_media").toUInt(), media_label);
    fill_save_type(get_dest_media());

    // Restore the checkboxes.
    if (use_saver)
        ui->chk_cam_name->setChecked(!camera_ptr->get_saver()->getCameraName().isEmpty());
    else
        ui->chk_cam_name->setChecked(settings.value("use_camera_name", false).toBool());

    // Are we allowing XML translations?
    if (allow_xml)
    {
        // Check to see if we used XML last time.
        if (use_saver)
            ui->chk_xml->setChecked(camera_ptr->get_saver()->saving_xml());
        else
            ui->chk_xml->setChecked(settings.value("create_xml", false).toBool());
    }
    else
    {
        // XML not allowed.
        ui->chk_xml->setChecked(false);
    }

    if (use_saver)
        ui->chk_pfmd->setChecked(camera_ptr->get_saver()->getSavePfmd());
    else
        ui->chk_pfmd->setChecked(settings.value("save_pfmd", false).toBool());

    // Ticket# 143 after discussion with TimB we will always override and
    // automatically select PFMD - because - there are issues he comes across
    // out in the field where people uncheck it then forget and it stays unchecked
    // and when a PM file is needed it's not there.  He would like us to
    // override and always create one (of course it won't for CAP saves) but
    // at this point we will ignore a user settings value.
    if (ui->chk_pfmd->isEnabled())
    {
        ui->chk_pfmd->setChecked(true);
    }

    // Make sure if XML is checked that the PFMD is also checked.
    if (ui->chk_xml->isChecked())
    {
        do_xml_select();    // This will handle the logic.
    }

    if (use_saver)
    {
        // Retrieve the strings which may have been saved.
        comment_str = camera_ptr->get_saver()->getComment();
        tag_str = camera_ptr->get_saver()->getTag();

        // Did we save raw files the last time?
        if ((save_type == FI_SAVER::STREAM_TYPE::TIF_RAW_STACK) ||
            (save_type == FI_SAVER::STREAM_TYPE::DNG_STACK) ||
            (save_type == FI_SAVER::STREAM_TYPE::RDI_PACKED))
        {
            raw_selected = true;
        }
        else
        {
            raw_selected = false;
        }

        // Are we using the JPEG or BMP codec with AVIs?
        if (save_type == FI_SAVER::STREAM_TYPE::AVI)
        {
            use_jpeg_codec = (camera_ptr->get_saver()->getImageInputType() == FI_SAVER::IMAGE_TYPE::JPEG);
        }
        else
        {
            use_jpeg_codec = (save_type == FI_SAVER::STREAM_TYPE::JPEG_STACK);
        }

        // What are the MP4 settings if set.
        // Ticket#260 Restore MP4 settings from saver.
        if (save_type == FI_SAVER::STREAM_TYPE::MP4)
        {
            h264Speed       = camera_ptr->get_saver()->getH264Speed();
            h264Quality     = camera_ptr->get_saver()->getH264Quality();
            h264AllKeys     = camera_ptr->get_saver()->getH264EnableAllKeyframes();

            // Validate the quality.
            if (h264Quality < FAS_VIDEO::MINIMUM_H264_QUALITY_PCT)
                h264Quality = FAS_VIDEO::MINIMUM_H264_QUALITY_PCT;
            else if (h264Quality > 100)
                h264Quality = 100;

            // Validate the speed.
            // Since it's uint32_t 0 is valid but nothing at or over end of speeds is valid.
            if (h264Speed >= FAS_VIDEO::H264_SPEEDS::H264_END_OF_SPEEDS)
                h264Speed = FAS_VIDEO::DEFAULT_H264_SPEED;

            set_h264_quality((int)h264Quality);
            set_h264_speed((int)h264Speed);
            set_h264_allkeyframe(h264AllKeys);
#if 0
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "get_saver() quality:" << h264Quality
                         << "speed:" << h264Speed << "allkeys:" << h264AllKeys;
#endif
        }

        // Set up the flag for saving partition metadata in CFG file to default.
        metadata_cfg_selected = camera_ptr->get_saver()->getMetadataCfgSave();
    }
    else
    {
        // Retrieve the strings which may have been saved.
        comment_str = settings.value("comment").toString();
        tag_str = settings.value("tag_string").toString();

        // Restore the format the user last selected for saving the recording.
        save_type = settings.value("save_type").toInt();

        // Did we save raw files the last time?
        raw_selected = settings.value("save_raw", false).toBool();

        // Did we save metadata cfg files the last time?
        metadata_cfg_selected = settings.value("save_metadata_cfg", false).toBool();

        // Are we saving AVIs?
        if (save_type == FI_SAVER::STREAM_TYPE::AVI)
        {
            // This will, by default, be true. But if we saved AVI/BMPs, then this will be false.
            use_jpeg_codec = settings.value(SETTINGS_AVI_SAVE_JPEG, true).toBool();
        }
        else
        {
            // Not using the JPEG information.
            use_jpeg_codec = (save_type == FI_SAVER::STREAM_TYPE::JPEG_STACK);
        }

        // What are the MP4 settings if set.
        // Ticket#260 Restore MP4 settings from settings.
        SaveDialog::getH264Settings(h264Speed, h264Quality, h264AllKeys);

        set_h264_quality((int)h264Quality);
        set_h264_speed((int)h264Speed);
        set_h264_allkeyframe(h264AllKeys);
#if 0
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "settings quality:" << h264Quality
                     << "speed:" << h264Speed << "allkeys:" << h264AllKeys;
#endif

    }

    settings.endGroup();

    // Did we use the tag?
    ui->chk_tag_name->setChecked(!tag_str.isEmpty());

    // Did we use the comment?
    ui->eb_comment->setText(comment_str);
    set_use_comment(validate_comment());

    // Adjust the JPEG related controls.
    set_avi_jpeg_codec(use_jpeg_codec);

    // Restore the last selected save type.
    save_type_found = select_save_type(save_type);

    // If we cannot restore to the one saved use the one that was found.
    if (save_type_found != save_type)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_type_requested:" << save_type << "not = save_type_found:" << save_type_found;
        save_type = save_type_found;

        // TICKET #433: There are instances when the last save operation was a CAP but when opened a CAP file
        // outside of a camera connection we don't allow saving to CAP files from within that playback scenario.
        // When that was happening the default save type was reset to an AVI/JPEG save type but because the INI
        // settings indicated a CAP file the "use_jpeg_codec" setting was causing the quality slider to hide.
        // We re-evaluate and reset use_jpeg_codec when the save type had to change.

        // Reset the use_jpeg_codec to match the proper save type.
        if (save_type == FI_SAVER::STREAM_TYPE::AVI)
        {
            // We will rely on the INI settings to make the call on whether AVI is a JPEG or a BMP.
            use_jpeg_codec = settings.value(SETTINGS_AVI_SAVE_JPEG, true).toBool();
        }
        else
        {
            // Set the jpeg codec based on the image stack being a JPEG image stack.
            use_jpeg_codec = (save_type == FI_SAVER::STREAM_TYPE::JPEG_STACK);
        }
    }

    // Update our display.
    do_save_type_changed(ui->cb_save_type->currentIndex());

    // The last time we saved TIF, was it RAW?
    if ((save_type == GIGE_STREAM_FORMAT_TIF) || (save_type == GIGE_STREAM_FORMAT_DNG))
        ui->cb_tif_raw->setChecked(raw_selected);
    else
        ui->cb_tif_raw->setChecked(false);

    // Set the save cfg file checkboxes.
    ui->cb_metadata_cfg->setChecked(metadata_cfg_selected);

    // Display the JPEG controls if we are saving JPEGs.
    ui->frm_jpg_quality->setEnabled(use_jpeg_codec);
    ui->frm_jpg_quality->setVisible(use_jpeg_codec);

    // Do we have a path?
    if (!download_path.isEmpty())
    {
        // Make sure the path exists.
        QDir download_dir(download_path);
        if (!download_dir.exists())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Path doesn't exist. Creating:" << download_dir;
            if (!download_dir.mkpath(download_path))
            {
                // Can't create the path, so use the Documents folder instead.
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to create the path:" << download_path;
                download_path = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0);
            }
        }

        // Make sure the path will look "right" to the user.
        download_path = QDir::toNativeSeparators(download_path);
        ui->txt_path->setText(download_path);
        ui->txt_path->setToolTip(tr("Full Path: ") + download_path);
    }

    // Is the tag empty?
    if (tag_str.isEmpty())
    {
        QDate   today = QDate::currentDate();
        tag_str = today.toString(Qt::ISODate);
    }

    // Set the tag value.
    ui->eb_tag_label->setText(tag_str);
}

/*!
 * @brief Select the entry in the combobox which has the same format type as the parameter.
 * @param save_type     Format type we're interested in.
 *
 * @returns save_type_found - It is possible we couldn't find the save_type requested.  This provides
 * a mechanism to return the different save type to the caller.
 */
int SaveDialog::select_save_type(int save_type)
{
    int     index;
    int     save_type_found = save_type;

    // Setup defaults.
    if (save_type == FI_SAVER::STREAM_TYPE::TIF_RAW_STACK)
        save_type = FI_SAVER::STREAM_TYPE::TIF_STACK;
    if (save_type == FI_SAVER::STREAM_TYPE::NO_STREAM)
        save_type = FI_SAVER::STREAM_TYPE::JPEG_STACK;

    // Find which entry has the save type information.
    for (index = 0; index < ui->cb_save_type->count(); ++index)
    {
        int cb_save_type;

        // Get the format information from the combobox entry.
        cb_save_type = ui->cb_save_type->itemData(index).toInt();

        // Do we have a match?
        if (cb_save_type == save_type)
        {
            // Found it.
            break;
        }
    }

    // Sanity check.
    if (index == ui->cb_save_type->count())
    {
        QString str_msg = QString::fromUtf8("The format type [%1] was not found!").arg(save_type);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << str_msg;

        // Recursively switch to a known save type.
        save_type_found = select_save_type(ui->cb_save_type->itemData(0).toInt());
        return save_type_found;
    }

    // Select the found entry.
    ui->cb_save_type->blockSignals(true);
    ui->cb_save_type->setCurrentIndex(index);
    ui->cb_save_type->blockSignals(false);

    // Return the save type that was found.
    return save_type_found;
}

/*!
 * @brief Setup the autosave mode, initializing the range the start/end clip values
 *        are allowed to have.
 *
 * @param do_enable     True to enable autosave mode, false to disable.
 */
void SaveDialog::set_autosave_mode(bool do_enable)
{
    // Remember the flag.
    autosave_mode = do_enable;
    if (do_enable)
    {
        // Update the dialog's title.
        setWindowTitle(tr("Autosave Settings"));
    }

    // Re-initialize our ranges.
    init_ranges();
}

void SaveDialog::set_avi_jpeg_codec(bool use_jpeg_codec)
{
    // Specify the AVI/JPEG or AVI/BMP codec.
    ui->rb_jpg->setChecked(use_jpeg_codec);
    ui->rb_bmp->setChecked(!use_jpeg_codec);

    if (use_jpeg_codec)
        do_jpg_select();
    else
        do_bmp_select();
}

/*!
 * @brief Enable (or disable) CAP file transfer mode.
 * @param do_enable     True to enable the mode, false otherwise.
 */
void SaveDialog::set_cap_transfer(bool do_enable)
{
    // Remember the setting.
    cap_transfer_mode = do_enable;
    if (cap_transfer_mode)
    {
        // The user isn't allowed to select a destination. It must
        // be to the local workstation.
        ui->rb_pc->setChecked(true);
        do_pc_toggle(true);
        ui->grp_devices->hide();

        // User can't select a range either.
        ui->frm_range->hide();
    }
    else
    {
        // User is allowed to select a destination and range.
        ui->grp_devices->show();
        ui->frm_range->show();
    }

    // Update our dialog.
    adjustSize();
    update();
}

/*!
 * @brief Initialize the start/end clip values using absolute frame numbers.
 *
 * @param start_clip    The start clip frame to use.
 * @param end_clip      The ending clip frame to use.
 *
 * @sa SaveDialog::set_rel_range
 */
void SaveDialog::set_clip_range(uint32_t start_clip, uint32_t end_clip)
{
    int rel_start;
    int rel_end;

    // Save the settings.
    clip_start = start_clip;
    clip_end = end_clip;

    // If the user specified a -1 as our end clip, that means to use the number of frames.
    if (end_clip == (uint32_t) -1)
        end_clip = num_frames;

    // Sanity check.
    if (end_clip < start_clip)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Something isn't right, end_clip < start_clip!";
    }

    // Check to see if the trigger is outside of our range.
    if (cam_ip_addr == 0)
    {
        int rel_max;
        int rel_min;

        // Recalculate what our min/max values are.
        rel_min = (int)(start_clip - trig_frame);
        rel_start = rel_min;
        rel_max = (int)(end_clip - trig_frame);
        rel_end = rel_max;

        // Update our controls.
        ui->spin_clip_start->blockSignals(true);
        ui->spin_clip_end->blockSignals(true);
        ui->spin_clip_start->setMinimum(rel_min);
        ui->spin_clip_end->setMinimum(rel_min + 1);
        ui->spin_clip_start->setMaximum(rel_max - 1);
        ui->spin_clip_end->setMaximum(rel_max);
        ui->spin_clip_start->blockSignals(false);
        ui->spin_clip_end->blockSignals(false);
    }

    else
    {
        // Convert our values to relative (to trigger) values for the user.
        rel_start = (int)(start_clip - trig_frame);
        rel_end = (int)(end_clip - trig_frame);
    }

    // Update the starting/ending positions.
    ui->spin_clip_start->setValue(rel_start);
    ui->spin_clip_end->setValue(rel_end);
}

/*!
 * @brief Let the user select a media type.
 * @param media_sel Which of the media we want to go to.
 */
void SaveDialog::set_dest_media(uint32_t media_sel, QString& media_label)
{
    int sel_index;
    int media_avail;
    QJsonObject dev_object;

    // Ignore this for CAP file transfer mode.
    if (cap_transfer_mode)
        return;

    // Are we connected to a camera?
    if (cam_ip_addr != 0)
    {
        // Which media are available?
        media_avail = (int)read_camera_cache(GIGE_REG_AVAIL_MEDIA);
    }
    else
    {
        media_avail = 0;
    }

    // Check to see if we're saving to the local PC.
    if (media_label.isEmpty())
    {
        // Handle the specified media request.
        switch (media_sel)
        {
        case GIGE_MEDIA_SDCARD:
            if ((media_avail & GIGE_MEDIA_SDCARD) == GIGE_MEDIA_SDCARD)
            {
                sel_index = ui->cb_devices->findText(CamExplore::SDCARD_LABEL);
                ui->cb_devices->setCurrentIndex(sel_index);
                ui->rb_devices->setChecked(true);
            }
            break;
        case GIGE_MEDIA_SSD:
            if ((media_avail & GIGE_MEDIA_SSD) == GIGE_MEDIA_SSD)
            {
                sel_index = ui->cb_devices->findText(CamExplore::SSD_LABEL);
                ui->cb_devices->setCurrentIndex(sel_index);
                ui->rb_devices->setChecked(true);
            }
            break;
        case GIGE_MEDIA_USB:
            if ((media_avail & GIGE_MEDIA_USB) == GIGE_MEDIA_USB)
            {
                sel_index = ui->cb_devices->findText(CamExplore::USB_LABEL);
                ui->cb_devices->setCurrentIndex(sel_index);
                ui->rb_devices->setChecked(true);
            }
            break;
        default:
        case MEDIA_LOCAL_PC:
            ui->rb_pc->setChecked(true);
            break;
        }
    }
    else
    {
        // Loop through all the devices we know about.
        for (sel_index = 0; sel_index < ui->cb_devices->count(); ++sel_index)
        {
            // Get the object associated with the selected index.
            dev_object = ui->cb_devices->itemData(sel_index).toJsonObject();

            // Is this of a type we're interested in?
            if ((dev_object["type"].toInt() == (int)media_sel) && (dev_object["label"].toString().compare(media_label) == 0))
            {
                break;
            }
        }

        // Did we find the device requested?
        if (sel_index < ui->cb_devices->count())
        {
            ui->cb_devices->setCurrentIndex(sel_index);
            ui->rb_devices->setChecked(true);
        }
    }
}

/*!
 * @brief Modify the dialog so that it will support the direct from disk
 *        transfer mode.
 * @param do_enable     True to enable the mode, false otherwise.
 */
void SaveDialog::set_direct_transfer(bool do_enable)
{
    // Are we doing a direct from drive transfer?
    if (do_enable)
    {
        // The user isn't allowed to select a destination. It must
        // be to the local workstation.
        ui->rb_pc->setChecked(true);
        do_pc_toggle(true);
        ui->grp_devices->hide();
    }
    else
    {
        // User is allowed to select a destination and range.
        ui->grp_devices->show();
    }

    // Re-initialize the ranges we can use.
    init_ranges();

    // Update our dialog.
    adjustSize();
}

/*!
 * @brief Set the UI controls to the JPEG value passed in.
 * @param quality_value     The current JPEG quality value.
 */
void SaveDialog::set_jpeg_quality(int quality_value)
{
    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "quality_value", (uint32_t)quality_value);

    if (ui->sld_jpeg->value() != quality_value)
        ui->sld_jpeg->setValue(quality_value);
    if (ui->spin_jpeg->value() != quality_value)
        ui->spin_jpeg->setValue(quality_value);
}

/*!
 * @brief Set the UI controls to H.264 setting provided.
 * @param quality_value     The current H.264 quality setting.
 */
void SaveDialog::set_h264_quality(int quality_value)
{
    if (ui->sld_h264_quality->value() != quality_value)
        ui->sld_h264_quality->setValue(quality_value);
    if (ui->spin_h264_quality->value() != quality_value)
        ui->spin_h264_quality->setValue(quality_value);
}

/*!
 * @brief Set the UI controls to H.264 setting provided.
 * @param speed_value     The current H.264 speed setting.
 */
void SaveDialog::set_h264_speed(int speed_value)
{
    if (ui->sld_h264_speed->value() != speed_value)
        ui->sld_h264_speed->setValue(speed_value);

    // Sync up the label to match the speed value set.
    do_h264_speed_changed(speed_value);
}

/*!
 * @brief Set the UI controls to H.264 setting provided.
 * @param enable_allkeyframe    True if turning on, false if not.
 */
void SaveDialog::set_h264_allkeyframe(bool enable_allkeyframe)
{
    if (enable_allkeyframe)
        ui->cb_keyframe_all->setChecked(true);
    else
        ui->cb_keyframe_all->setChecked(false);
}

void SaveDialog::set_multicam(bool do_multicam, int camera_count)
{
    // Remember the new mode.
    multicam_mode = do_multicam;

    // Are we in multicamera mode?
    if (multicam_mode)
    {
        // The user isn't allowed to select a destination. It must
        // be to the local workstation.
        ui->rb_pc->setChecked(true);
        do_pc_toggle(true);

        // Do we have a count?
        if (camera_count > 0)
        {
            QString title_str;

            // Modify our title to include the number of cameras.
            title_str = this->windowTitle() + tr(" (%1 cameras selected)").arg(camera_count);
            setWindowTitle(title_str);
        }
    }

    // Display or hide our controls.
    ui->chk_use_camera_range->setVisible(multicam_mode);
    ui->grp_devices->setVisible(!multicam_mode);

    // Update the dialog.
    adjustSize();
    update();
}

/*!
 * @brief Set our mixed group flag according to the passed in parameter.
 * @param mixed_group       True if we have a mixed group of cameras, false if they are all the same.
 */
void SaveDialog::set_multicam_mixed(bool mixed_group)
{
    // Save our flag.
    multicam_mixed = mixed_group;

    // Initialize our save type combo.
    fill_save_type(MEDIA_LOCAL_PC);
}

/*!
 * @brief Update the number of frames.
 * @param frame_count   The new frame count.
 */
void SaveDialog::set_num_frames(uint32_t frame_count)
{
    // Save the number of frames.
    num_frames = frame_count;
}

/*!
 * @brief Initialize the start/end clip values using trigger-relative
 *        values rather than absolute frame numbers.
 *
 * @param start_clip    The start clip frame to use.
 * @param end_clip      The ending clip frame to use.
 *
 * @sa SaveDialog::set_clip_range
 */
void SaveDialog::set_rel_range(int start_clip, int end_clip)
{
    // Sanity check.
    if (start_clip < ui->spin_clip_start->minimum())
        start_clip = ui->spin_clip_start->minimum();
    if (end_clip > ui->spin_clip_end->maximum())
        end_clip = ui->spin_clip_end->maximum();

    // Are they the same?
    if (start_clip == end_clip)
    {
        // Use the full range.
        start_clip = ui->spin_clip_start->minimum();
        end_clip = ui->spin_clip_end->maximum();
    }

    // Sanity check.
    if (end_clip < start_clip)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Something isn't right, end_clip < start_clip!";
    }

    // Set the controls.
    ui->spin_clip_start->setValue(start_clip);
    ui->spin_clip_end->setValue(end_clip);
}

/*!
 * @brief Enable/disable the comment field.
 * @param use_it    True to enable, false to disable.
 */
void SaveDialog::set_use_comment(bool use_it)
{
    ui->chk_comment->setChecked(use_it);
    do_chk_comment(use_it ? Qt::Checked : Qt::Unchecked);
}

/*!
 * @brief Enable/disable the tag field.
 * @param use_it    True to enable, false to disable.
 */
void SaveDialog::set_use_tag_label(bool use_it)
{
    ui->chk_tag_name->setChecked(use_it);
    do_chk_tag_label(use_it ? Qt::Checked : Qt::Unchecked);
}

/*!
 * @brief If the parameter is set true, make the dialog settings read-only. Also
 *        show the autosave disable button.
 *
 * @param view_only     True to make the view read-only.
 */
void SaveDialog::set_view_only(bool view_only)
{
    // Are we view only?
    if (view_only)
    {
        // Enable/disable the controls based on if we're only allowed to view.
        ui->frm_control->setEnabled(false);

        // If we're view only, hide the OK and Cancel buttons, and replace them
        // with a Close button. Also
        ui->buttonBox->setStandardButtons(QDialogButtonBox::Close);

        // Show our autosave disable button.
        ui->pb_autosave->setVisible(true);

        // Setup our signal/slots.
        connect(ui->buttonBox->button(QDialogButtonBox::Close), &QPushButton::clicked, this, &SaveDialog::reject);
        connect(ui->pb_autosave, &QPushButton::clicked, this, &SaveDialog::accept);

        // Remember we're view only.
        this->view_only = view_only;
    }
}

/*!
 * @brief Update the Image Format and JPEG Quality controls for our Save Option
 * based on the stream type being saved.  These include the JPEG and BMP radios
 * and the JPEG quality slider and spin.
 *
 * @param streamType  What type of stream is being saved.
 */
void SaveDialog::update_imgfmt_ctls(FI_SAVER::STREAM_TYPE streamType, bool isSeries3, uint32_t dest_media)
{
    // These are the default settings.  We rely on them for the default case.
    bool hide_grp_save_option   = false;
    bool frm_avi_option_enabled = false;
    bool use_jpeg_quality       = false;
    bool rb_bmp_avail           = false;
    bool rb_jpg_avail           = false;
    bool cfg_available          = true;
    bool pfmd_available         = !isSeries3;
    bool xml_available          = allow_xml;

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::AVI:
        frm_avi_option_enabled  = true;
        use_jpeg_quality        = ui->rb_jpg->isChecked();
        rb_bmp_avail            = true;
        rb_jpg_avail            = true;
        cfg_available           = true;
        hide_grp_save_option    = false;
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        frm_avi_option_enabled  = false;
        use_jpeg_quality        = true;
        rb_bmp_avail            = false;
        rb_jpg_avail            = true;
        cfg_available           = true;
        ui->rb_jpg->setChecked(true);
        hide_grp_save_option    = false;
        break;
    case FI_SAVER::STREAM_TYPE::BMP_STACK:
        frm_avi_option_enabled  = false;
        use_jpeg_quality        = false;
        rb_bmp_avail            = true;
        rb_jpg_avail            = false;
        cfg_available           = true;
        ui->rb_bmp->setChecked(true);
        hide_grp_save_option    = true;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAP:
        frm_avi_option_enabled  = false;
        use_jpeg_quality        = false;
        rb_bmp_avail            = false;
        rb_jpg_avail            = false;
        cfg_available           = false;
        pfmd_available          = false;
        hide_grp_save_option    = true;
        xml_available           = false;
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
        frm_avi_option_enabled  = false;
        use_jpeg_quality        = false;
        rb_bmp_avail            = false;
        rb_jpg_avail            = false;
        cfg_available           = false;
        pfmd_available          = false;
        hide_grp_save_option    = true;
        xml_available           = false;
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
        frm_avi_option_enabled  = false;
        use_jpeg_quality        = false;
        rb_bmp_avail            = false;
        rb_jpg_avail            = false;
        cfg_available           = true;
        hide_grp_save_option    = true;
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        frm_avi_option_enabled  = false;
        use_jpeg_quality        = false;
        rb_bmp_avail            = false;
        rb_jpg_avail            = false;
        cfg_available           = false;
        pfmd_available          = false;
        hide_grp_save_option    = true;
        xml_available           = false;
        break;
    default:
        frm_avi_option_enabled  = false;
        use_jpeg_quality        = false;
        rb_bmp_avail            = false;
        rb_jpg_avail            = false;
        cfg_available           = true;
        hide_grp_save_option    = false;
        break;
    }

    // Setup the UI based on the flags.
    ui->frm_avi_options->setEnabled(frm_avi_option_enabled);
    ui->frm_avi_options->setVisible(frm_avi_option_enabled);
    ui->rb_bmp->setEnabled(rb_bmp_avail);
    ui->rb_jpg->setEnabled(rb_jpg_avail);
    ui->frm_jpg_quality->setEnabled(use_jpeg_quality);
    ui->frm_jpg_quality->setVisible(use_jpeg_quality);

    // Special handling for the metadata CFG file.
    if (cfg_available)
    {
        // I am assuming that non-series-3 cameras will be able to take advantage of this
        // setting when we mark it in the save json to the new camera servers.
        bool howToCheck     = ((dest_media == MEDIA_LOCAL_PC) || !isSeries3) ? FI_SAVER::DEFAULT_SAVE_METADATA_CFG : true;
        bool howToEnable    = ((dest_media == MEDIA_LOCAL_PC) || !isSeries3) ? true : false;

        ui->cb_metadata_cfg->setChecked(howToCheck);
        ui->cb_metadata_cfg->setEnabled(howToEnable);
        //ui->cb_metadata_cfg->setHidden(false);
    }
    else
    {
        ui->cb_metadata_cfg->setChecked(false);
        ui->cb_metadata_cfg->setEnabled(false);
        //ui->cb_metadata_cfg->setHidden(true);
    }

    // Enable/disable our PFMD checkbox.
    ui->chk_pfmd->setEnabled(pfmd_available);

    // If the ancillary file creation checkboxes are all disabled
    // also disable the label for them.
    ui->lbl_create_ancillaries->setEnabled((xml_available || ui->cb_metadata_cfg->isEnabled() || pfmd_available) ? true : false);

    // CAP file saves don't need save options.
    ui->grp_save_options->setHidden(hide_grp_save_option);
}

/*!
 * @brief Update the Raw(unprocessed) checkbox for our Save Option
 * based on the stream type being saved.
 *
 * @param streamType  What type of stream is being saved.
 */
void SaveDialog::update_rawcb_ctls(FI_SAVER::STREAM_TYPE streamType)
{
    bool set_enabled    = false;
    bool set_hidden     = false;
    bool merge_overlay  = ui->chk_overlay->isChecked();

    switch (streamType)
    {
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        if (allow_raw && !merge_overlay)
        {
            // What did it use to be?
            set_enabled = true;
            ui->cb_tif_raw->setChecked(raw_selected);
        }
        else
        {
            // Raw not allowed.
            set_enabled = false;
            ui->cb_tif_raw->setChecked(false);
        }
        break;
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        // Can only have raw images downloaded.
        ui->cb_tif_raw->setChecked(true);

        // It can not be enabled, because it can not be changed.
        // But we want to show that this (DNG) is a raw format.
        set_enabled = false;
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        // Can only have raw images downloaded.
        ui->cb_tif_raw->setChecked(true);

        // It can not be enabled, because it can not be changed.
        // But we want to show that this (RDI_PACKED) is a raw format.
        set_enabled = false;
        break;
    default:
        // Leave as set in local variable initializers.
        ui->cb_tif_raw->setChecked(false);
        set_enabled = false;
        set_hidden  = true;
        break;
    }

    // Setup the UI based on the flags.
    ui->frm_tif_options->setEnabled(set_enabled);
    ui->cb_tif_raw->setEnabled(set_enabled);
    ui->frm_tif_options->setHidden(set_hidden);
}

/*!
 * @brief Update the H.264 Option group for our Save Option
 * based on the stream type being saved.
 *
 * @param streamType  What type of stream is being saved.
 */
void SaveDialog::update_h264_ctls(FI_SAVER::STREAM_TYPE streamType)
{
    // We only want the H264 options if we're saving an MP4.
    bool set_enabled;
    set_enabled = (streamType == FI_SAVER::STREAM_TYPE::MP4);

    // H.264 options are only for MP4s.
    ui->grp_h264_options->setEnabled(set_enabled);
    ui->grp_h264_options->setVisible(set_enabled);
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_AVAIL_MEDIA register value.
 */
void SaveDialog::update_media(uint32_t reg_value)
{
    int item_index;
    QString device_sel;

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
        device_sel = ui->cb_devices->currentText();

    // We want to check GIGE_MEDIA_xxx and not GIGE_MEDIA_EXISTS_xxx.
    // GIGE_MEDIA_EXISTS_xxx means the device exists, but doesn't necessarily
    // mean it's formatted for us to write on. GIGE_MEDIA_xxx means
    // we can write to the device.
    if ((reg_value & GIGE_MEDIA_SDCARD) == GIGE_MEDIA_SDCARD)
    {
        // Only add the item if it doesn't already exist.
        if (ui->cb_devices->findText(CamExplore::SDCARD_LABEL) < 0)
            ui->cb_devices->addItem(CamExplore::SDCARD_LABEL);
    }
    else
    {
        // Check to see if the item went away. Remove it from the list.
        if ((item_index = ui->cb_devices->findText(CamExplore::SDCARD_LABEL)) >= 0)
            ui->cb_devices->removeItem(item_index);
    }
    if ((reg_value & GIGE_MEDIA_SSD) == GIGE_MEDIA_SSD)
    {
        // Only add the item if it doesn't already exist.
        if (ui->cb_devices->findText(CamExplore::SSD_LABEL) < 0)
            ui->cb_devices->addItem(CamExplore::SSD_LABEL);
    }
    else
    {
        // Check to see if the item went away. Remove it from the list.
        if ((item_index = ui->cb_devices->findText(CamExplore::SSD_LABEL)) >= 0)
            ui->cb_devices->removeItem(item_index);
    }
    if ((reg_value & GIGE_MEDIA_USB) == GIGE_MEDIA_USB)
    {
        // Only add the item if it doesn't already exist.
        if (ui->cb_devices->findText(CamExplore::USB_LABEL) < 0)
            ui->cb_devices->addItem(CamExplore::USB_LABEL);
    }
    else
    {
        // Check to see if the item went away. Remove it from the list.
        if ((item_index = ui->cb_devices->findText(CamExplore::USB_LABEL)) >= 0)
            ui->cb_devices->removeItem(item_index);
    }

    // If there is no external media then hide the devices controls.
    if (ui->cb_devices->count() == 0)
    {
        // Make sure we're disabled, and force the PC selection.
        ui->cb_devices->setEnabled(false);
        ui->rb_devices->setEnabled(false);
        ui->rb_pc->setChecked(true);
    }
    else
    {
        int sel_index = 0;

        // Make sure we're enabled.
        ui->cb_devices->setEnabled(true);
        ui->rb_devices->setEnabled(true);

        // Attempt to re-select the old value.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);
            if (sel_index < 0)
                sel_index = 0;
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);
    }
}

/*!
 * @brief Checks a comment string to make sure that there is at least one non-blank character.
 */
bool SaveDialog::validate_comment()
{
    QString comment_str = ui->eb_comment->text();

    // Is the comment valid? First remove all whitespace from the start and end of the string.
    comment_str = comment_str.trimmed();

    // Get rid of any trailing null characters.
    if (!comment_str.isEmpty() && (comment_str.at(comment_str.size() - 1) == '\0'))
        comment_str.chop(1);

    // Set the updated string to our control.
    ui->eb_comment->setText(comment_str);

    // Return true if the string is not empty.
    return !comment_str.isEmpty();
}

/*!
 * @brief Return true if the user wants to use the JPEG codec, false for the BMP codec.
 */
bool SaveDialog::use_avi_jpeg_codec()
{
    return ui->rb_jpg->isChecked();
}

/*!
 * @brief Return true if the user wants to incorporate the camera's name into the saved file name.
 */
bool SaveDialog::use_camera_name()
{
    return ui->chk_cam_name->isChecked();
}

/*!
 * @brief Return true if the user wants to use the range setup on each individual camera (in multicamera mode).
 */
bool SaveDialog::use_camera_range()
{
    return ui->chk_use_camera_range->isChecked();
}

/*!
 * @brief Return true if the user wants to provide a comment.
 */
bool SaveDialog::use_comment()
{
    return ui->chk_comment->isChecked();
}

/*!
 * @brief Return true if the user wants to utilize a tag in the saved file name.
 */
bool SaveDialog::use_tag_label()
{
    return ui->chk_tag_name->isChecked();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user pressed the OK button. Save the settings entered by the user
 *        so we can re-use them the next time we present the save dialog.
 */
void SaveDialog::accept()
{
    bool        has_a_comment = false;  // Just because the box may be checked if the comment isn't valid we don't want it.
    int         index;
    QSettings   settings;
    FI_SAVER::STREAM_TYPE save_type;

    // Save the values for re-use later.
    if (get_dest_media() == MEDIA_LOCAL_PC)
        SaveDialog::savePath(cam_ip_addr, ui->txt_path->text());

    settings.beginGroup(SAVE_SETTINGS_SECTION);
    settings.setValue("dest_media", get_dest_media());
    settings.setValue("dest_label", get_dest_label());
    if (use_tag_label())
        settings.setValue("tag_string", ui->eb_tag_label->text());
    else
        settings.remove("tag_string");

    // We don't want to use the comment if there is nothing in the comment.  It is possible for the
    // user to have blanks in the comment and we don't want that to be allowed.
    if (use_comment())
    {
        // Is our comment valid?
        has_a_comment = validate_comment();
    }

    // Only set the comment if we have one.
    if (has_a_comment)
    {
        settings.setValue("comment", ui->eb_comment->text());
    }
    else
    {
        settings.remove("comment");
    }

    settings.setValue("use_camera_name", use_camera_name());
    settings.setValue("create_xml", create_xml());
    settings.setValue("playback_rate", ui->spin_play_fps->value());
    settings.setValue("tag_enable", ui->chk_tag_name->isChecked());

    index = ui->cb_save_type->currentIndex();
    save_type = static_cast<FI_SAVER::STREAM_TYPE>(ui->cb_save_type->itemData(index).toInt());
    settings.setValue("save_type", save_type);
    settings.setValue("save_raw", ui->cb_tif_raw->isChecked());
    settings.setValue("save_metadata_cfg", ui->cb_metadata_cfg->isChecked());
    settings.setValue("save_pfmd", ui->chk_pfmd->isChecked());

    // Only change this if the save type is AVI.
    if (save_type == FI_SAVER::AVI)
        settings.setValue(SETTINGS_AVI_SAVE_JPEG, ui->rb_jpg->isChecked());

    // Only change this if the save type is MP4
    // Ticket#260
    if (save_type == FI_SAVER::MP4)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "h264 Quality:" << get_h264_quality();
        settings.setValue(SETTINGS_MP4_H264_QUALITY, (int)get_h264_quality());
        settings.setValue(SETTINGS_MP4_H264_SPEED, (int)GetH264Speed());
        settings.setValue(SETTINGS_MP4_H264_ALLKEY, GetH264EnableAllKeyFrames());
    }

    settings.endGroup();

    // Pass it on up.
    QDialog::accept();
}

void SaveDialog::connect_camera()
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Don't need to connect, if we've already connected.
    if (cam_connected)
        return;

    // Get our communications object pointer. Note that we might not have a
    // camera connected.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
        return;
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
        return;

    // Connect the communications interfaces that we're interested in.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &SaveDialog::read_complete, Qt::QueuedConnection);
    connect(camera_ptr, &CameraMDI::media_info_ready, this, &SaveDialog::handle_json_media, Qt::QueuedConnection);

    // Remember we're connected.
    cam_connected = true;
}

void SaveDialog::disconnect_camera()
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Are we actually connected?
    if (!cam_connected)
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
        return;
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
        return;

    // Disconnect from the camera.
    disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &SaveDialog::read_complete);
    disconnect(camera_ptr, &CameraMDI::media_info_ready, this, &SaveDialog::handle_json_media);

    // Not connected any more.
    cam_connected = false;
}

void SaveDialog::enable_controls(bool do_enable)
{
    // This doesn't apply if we're view only.
    if (view_only)
        return;

    // Enable/Disable all controls except for the cancel button.
    ui->frm_control->setEnabled(do_enable);

    // We may not have an OK button up.
    ui->buttonBox->button(QDialogButtonBox::Save)->setEnabled(do_enable);
}

/*!
 * @brief Override the QDialog's exec method so we can do some initialization first.
 * @return Result from QDialog::exec().
 */
int SaveDialog::exec()
{
    CameraMDI   *cam_ptr = qobject_cast<CameraMDI *>(parent());

    // Initialize our controls, if necessary.
    if (!initialized)
    {
        // By default, we want to be disabled until the dialog has been initialized.
        enable_controls(false);

        // Now initialize.
        // Note:  The default parameter query_cam is "true".
        init_controls();
    }

    // Do we have a camera address?
    if (cam_ptr != nullptr)
        connect_camera();

    // Show the dialog.
    return QDialog::exec();
}

void SaveDialog::device_selected(int index)
{
    uint32_t save_device;
    QString sel_device;

    // Make sure our radio button is lit.
    ui->rb_devices->setChecked(true);

    // Overlays are not allowed here.
    ui->chk_overlay->setEnabled(false);
    ui->chk_overlay->setChecked(false);

    // Get the selected text.
    sel_device = ui->cb_devices->itemText(index);

    // Compare it with known values to determine the device selected.
    if (sel_device.compare(CamExplore::SDCARD_LABEL) == 0)
        save_device = GIGE_MEDIA_SDCARD;
    else if (sel_device.compare(CamExplore::SSD_LABEL) == 0)
        save_device = GIGE_MEDIA_SSD;
    else
        save_device = GIGE_MEDIA_USB;

    // Update what's allowed.
    fill_save_type(save_device);

    // Make sure we update the UI based on the type now available.
    do_save_type_changed(ui->cb_save_type->currentIndex());

    // If we're saving AVIs and the destination is NOT the local PC, only JPEGs are allowed.
    // NOTE_JLC - Revisit this for MP4 video.
    if (GetStreamType() == FI_SAVER::STREAM_TYPE::AVI)
    {
        // Make sure we're specifying the JPEG codec.
        set_avi_jpeg_codec(true);
    }
}

/*!
 * @brief The user clicked on the BMP radio button.
 */
void SaveDialog::do_bmp_select()
{
    // Only enable the JPEG quality controls if the JPG option is selected.
    ui->frm_jpg_quality->setEnabled(false);
    ui->frm_jpg_quality->setVisible(false);
}

/*!
 * @brief The user clicked on the path browse button. Let the user select
 *        a different download folder to use.
 */
void SaveDialog::do_browse()
{
    QString     download_path;

    // Retrieve the current setting for the download path.
    download_path = ui->txt_path->text();
    download_path = QFileDialog::getExistingDirectory(this, tr("Select A Directory To Save Files To"), download_path);

    // Did we get anything?
    if (download_path.isEmpty())
        return;

#if 0
    // Check to see if the user has specified a "local media" device for the path.
    if (MainWindow::local_control())
    {
        CameraMDI   *camera_ptr;

        // Get a pointer to our camera object. See if the camera name matches our name.
        camera_ptr = qobject_cast<CameraMDI *>(parent());
        if (camera_ptr && (QHostInfo::localHostName().compare(camera_ptr->get_camera_name(), Qt::CaseInsensitive) == 0))
        {
            // Loop through all the local media devices.
            for (int i = 0; i < ui->cb_devices->count(); ++i)
            {
                QJsonObject dev_object;

                // Get the JSon information associated with this device.
                dev_object = ui->cb_devices->itemData(i).toJsonObject();

                // Is there a match?
                if (download_path.contains(dev_object["mountdir"].toString()))
                {
                    QMessageBox::information(MainWindow::GetInstance().get(), tr("Performance Suggestion"), tr("Your Path matches an existing Device.\nUsing the Device selection provides higher performance."));
                }
            }
        }
    }
#endif

    // Make sure the path will look "right" to the user.
    download_path = QDir::toNativeSeparators(download_path);
    ui->txt_path->setText(download_path);
    ui->txt_path->setToolTip(tr("Full Path: ") + download_path);
}

void SaveDialog::do_cam_range()
{
    // Make sure we're in multicamerea mode.
    if (!multicam_mode)
        return;

    // Enable/disable the clip selection based on if this is checked or not.
    ui->frm_clip_range->setEnabled(!ui->chk_use_camera_range->isChecked());
}

/*!
 * @brief The user clicked on the comment checkbox, either selecting or de-selecting it.
 * @param state     The state of the checkbox.
 */
void SaveDialog::do_chk_comment(int state)
{
    ui->eb_comment->setEnabled(state == Qt::Checked);
}

void SaveDialog::do_chk_overlay()
{
    int index = ui->cb_save_type->currentIndex();
    FI_SAVER::STREAM_TYPE save_type = static_cast<FI_SAVER::STREAM_TYPE>(ui->cb_save_type->itemData(index).toInt());

    // Things affected by the overlay setting.
    update_rawcb_ctls(save_type);
}

/*!
 * @brief The user clicked on the tag checkbox, either selecting or de-selecting it.
 * @param state     The state of the checkbox.
 */
void SaveDialog::do_chk_tag_label(int state)
{
    ui->eb_tag_label->setEnabled(state == Qt::Checked);
}

/*!
 * @brief The user may have changed the comment modify the tooltip if so.
 */
void SaveDialog::do_comment_changed(const QString& comment_str)
{
    ui->eb_comment->setToolTip((tr("Full Comment(inside quotes): \"%1\"").arg(comment_str)));
}

void SaveDialog::do_device_select()
{
    int sel_index;

    // Make sure the device radiobutton is checked.
    ui->rb_devices->setChecked(true);

    // Update the selection choices.
    sel_index = ui->cb_devices->currentIndex();
    device_selected(sel_index);
}

/*!
 * @brief The user clicked on the JPG radio button.
 */
void SaveDialog::do_jpg_select()
{
    // Only enable the JPEG quality controls if the JPG option is selected.
    ui->frm_jpg_quality->setEnabled(true);
    ui->frm_jpg_quality->setVisible(true);
}

/*!
 * @brief The user has modified the jpeg quality setting.
 */
void SaveDialog::do_jpeg_quality_changed(int jpeg_quality_new_value)
{
    FI_SAVER::STREAM_TYPE save_type;
    int index = ui->cb_save_type->currentIndex();
    save_type = static_cast<FI_SAVER::STREAM_TYPE>(ui->cb_save_type->itemData(index).toInt());

    // Persist AVI and JPEG Stack jpeg quality settings.
    if ((save_type == FI_SAVER::STREAM_TYPE::AVI) || (save_type == FI_SAVER::STREAM_TYPE::JPEG_STACK))
    {
        reg_jpeg_quality = (uint32_t)jpeg_quality_new_value;
    }

    // If we have never read a camera jpeg quality setting mark this one.
    // We don't want to overwrite the user's setting.
    if (firstJpegRegRead == 0)
    {
        firstJpegRegRead = reg_jpeg_quality;
    }

    if (ui->sld_jpeg->value() != jpeg_quality_new_value)
        ui->sld_jpeg->setValue(jpeg_quality_new_value);
    if (ui->spin_jpeg->value() != jpeg_quality_new_value)
        ui->spin_jpeg->setValue(jpeg_quality_new_value);
}

/*!
 * @brief The user clicked on the save metadata cfg checkbox.
 */
void SaveDialog::do_metadata_cfg_select()
{
    metadata_cfg_selected = ui->cb_metadata_cfg->isChecked();
}

/*!
 * @brief The user clicked on the save PFMD checkbox.
 */
void SaveDialog::do_pfmd_select()
{
    // If user unchecks the PFMD checkbox, make sure XML creation is unchecked.
    if (!ui->chk_pfmd->isChecked())
    {
        if (ui->chk_xml->isChecked())
        {
            ui->chk_xml->setChecked(false);
        }
    }
}

/*!
 * @brief The user clicked on the RAW checkbox.
 */
void SaveDialog::do_raw_select()
{
    raw_selected = ui->cb_tif_raw->isChecked();
}

/*!
 * @brief The user clicked on the All Keyframes checkbox.
 */
void SaveDialog::do_h264_allkeyframes_select()
{
    // This makes no sense!  ?
    //raw_selected = ui->cb_keyframe_all->isChecked();
}

/*!
 * @brief The user has modified the save type combobox, changing the format they want
 *        to use when saving the images.
 * @param index     The newly selected index into the combobox.
 */
void SaveDialog::do_save_type_changed(int index)
{
    bool        series3;
    bool        useFPS = false; // Is save type one that expects a frames per second setting?
    uint32_t    dest_media;
    CameraMDI   *cam_ptr = qobject_cast<CameraMDI *>(parent());
    FI_SAVER::STREAM_TYPE save_type;
    QSettings   settings;

    // We might not be connected to a camera...
    if (cam_ptr == nullptr)
    {
        PlaybackMDI *pb_ptr = qobject_cast<PlaybackMDI *>(parent());
        series3 = pb_ptr->isSeries3();
    }
    else
    {
        series3 = cam_ptr->is_series3();
    }

    // Get the format information selected.
    save_type = static_cast<FI_SAVER::STREAM_TYPE>(ui->cb_save_type->itemData(index).toInt());

    // What's the destination media the user has selected.
    dest_media = get_dest_media();

    // Setup the general save type option groups.
    update_imgfmt_ctls(save_type, series3, dest_media);
    update_rawcb_ctls(save_type);
    update_h264_ctls(save_type);

    // Is the selected type one with options?
    switch (save_type)
    {
    case FI_SAVER::STREAM_TYPE::TIF_STACK:
        break;
    case FI_SAVER::STREAM_TYPE::AVI:
        // On series3 cameras, if we're saving AVIs and the destination is NOT the local PC, only JPEGs are allowed.
        if (series3 && (dest_media != MEDIA_LOCAL_PC))
        {
            // Make sure we're specifying the JPEG codec.
            set_avi_jpeg_codec(true);
            ui->rb_bmp->setEnabled(false);  // Override any BMP setting from above.
        }

        // AVIs need to retrieve a Frames Per Second setting.
        useFPS = true;

        // AVIs and JPEG STACKS need to save the avi_jpeg_quality not the mp4 one.
        set_jpeg_quality((int)reg_jpeg_quality);

        // For an AVI, see what we saved last (JPEG vs. BMP)
        // This will, by default, be true. But if we saved AVI/BMPs, then this will be false.
        settings.beginGroup(SAVE_SETTINGS_SECTION);
        set_avi_jpeg_codec(settings.value(SETTINGS_AVI_SAVE_JPEG, true).toBool());
        settings.endGroup();
        break;
    case FI_SAVER::STREAM_TYPE::JPEG_STACK:
        // AVIs and JPEG STACKS need to save the avi_jpeg_quality not the mp4 one.
        set_jpeg_quality((int)reg_jpeg_quality);
        break;
    case FI_SAVER::STREAM_TYPE::MP4:
        //NOTE_JLC - what type of restrictions do we impose on series3 cameras.
        //AND are there any restrictions for non-series3 cameras.
        //AND what type of media do we allow - only MEDIA_LOCAL_PC???
        useFPS = true;  // MP4s require Frames Per Second setting.

        ui->sld_h264_speed->setVisible(true);
        ui->lbl_h264_speed->setVisible(true);
        ui->lbl_h264_speed_setting->setVisible(true);
        ui->cb_keyframe_all->setVisible(true);

        // We want to transfer data for MP4 saves in JPG form and 100% quality.
        // Note that this will overwrite default quality set for AVI JPEG's but
        // its value is restored when the AVI is reselected.
        set_jpeg_quality(MP4_JPEG_QUALITY);
        break;
    case FI_SAVER::STREAM_TYPE::FI_CAP:
    case FI_SAVER::STREAM_TYPE::FI_CAPSTK:
    case FI_SAVER::STREAM_TYPE::DNG_STACK:
        // These do not allow overlays.
        ui->chk_overlay->setChecked(false);
        break;
    case FI_SAVER::STREAM_TYPE::RDI_PACKED:
        // RDI packed does not allow overlays.
        ui->chk_overlay->setChecked(false);

        // RDI packed does not use the frames per second setting.
        useFPS = false;
        break;
    default:
        ui->frm_save_options->setEnabled(true);
        break;
    }

    // Enable the Frames Per Second if format requires setting.
    ui->lbl_play_fps->setVisible(useFPS);
    ui->spin_play_fps->setVisible(useFPS);

    // Series3 vs. newer cameras
    if (series3)
    {
        // Series3 cameras don't allow CAP files to have these features.
        ui->chk_cam_name->setEnabled(save_type != FI_SAVER::STREAM_TYPE::FI_CAP);
        ui->chk_tag_name->setEnabled(save_type != FI_SAVER::STREAM_TYPE::FI_CAP);

        // Series3 have to save the entire clip if they are saving a CAP.
        ui->spin_clip_start->setEnabled(save_type != FI_SAVER::STREAM_TYPE::FI_CAP);
        ui->spin_clip_end->setEnabled(save_type != FI_SAVER::STREAM_TYPE::FI_CAP);
    }
    else
    {
        // Newer cameras allow CAP files to have these features. Except if we are doing a
        // CAP->CAP transfer.
        if (save_type == FI_SAVER::STREAM_TYPE::FI_CAP)
        {
            ui->chk_cam_name->setEnabled(!cap_transfer_mode);
            ui->chk_tag_name->setEnabled(!cap_transfer_mode);
            ui->chk_comment->setEnabled(!cap_transfer_mode);
            if (cap_transfer_mode)
            {
                // Make sure these are cleared as well.
                ui->chk_cam_name->setChecked(false);
                ui->chk_tag_name->setChecked(false);
                ui->chk_comment->setChecked(false);
            }
        }
        else
        {
            ui->chk_cam_name->setEnabled(true);
            ui->chk_tag_name->setEnabled(true);
            ui->chk_comment->setEnabled(true);
        }

        // Newer cameras allow the user to specify the number of frames in a CAP.
        ui->spin_clip_start->setEnabled(true);
        ui->spin_clip_end->setEnabled(true);
    }

    // Overlays are never allowed with CAP files and DNG files and only allowed
    // if saving to the PC and even though you can't save CAP files to PC from
    // series3 there is no need to separate out this logic.
    if (ui->rb_pc->isChecked())
        ui->chk_overlay->setEnabled((save_type != FI_SAVER::STREAM_TYPE::FI_CAP) && (save_type != FI_SAVER::STREAM_TYPE::FI_CAPSTK)
                                    && (save_type != FI_SAVER::STREAM_TYPE::DNG_STACK) && allow_overlay);

    // All cameras prevent CAP files from these controls.
    ui->chk_xml->setEnabled(allow_xml && (save_type < FI_SAVER::STREAM_TYPE::FI_CAP));
    ui->chk_pfmd->setEnabled(!series3 && (save_type < FI_SAVER::STREAM_TYPE::FI_CAP));

    // Explicitly clear the XML check if the format is CAP or XML is not allowed.
    if (!allow_xml || (save_type >= FI_SAVER::STREAM_TYPE::FI_CAP))
        ui->chk_xml->setChecked(false);

    // Want to set it checked - will uncheck below for other conditions.
    ui->chk_pfmd->setChecked(true);

    // Explicitly clear the PFMD check if the format is CAP or greater.
    // Otherwise, if this is a series3 save then default to it being
    // checked if we're doing a local save.
    if (save_type >= FI_SAVER::STREAM_TYPE::FI_CAP)
        ui->chk_pfmd->setChecked(false);
    else if (series3)
        ui->chk_pfmd->setChecked(dest_media == MEDIA_LOCAL_PC);

    // Make sure we are sized correctly.
    adjustSize();

    // Update the dialog.
    update();
}

/*!
 * @brief The user has modified the h.264 speed slider, changing the speed
 * they want to use when compressing images into MP4 video.
 *
 * The index values should correspond to the enumeration in
 * FAS_VIDEO::H264_SPEEDS
 *
 * @param index     The newly selected index on the slider.
 *
 */
void SaveDialog::do_h264_speed_changed(int index)
{
    // Change the label to match the slider setting.
    FAS_VIDEO::H264_SPEEDS indexInSpeed = (FAS_VIDEO::H264_SPEEDS) index;

    switch (indexInSpeed)
    {
    case FAS_VIDEO::H264_SPEEDS::H264_SLOWER:
        ui->lbl_h264_speed_setting->setText("Slow");
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_MEDIUM:
        ui->lbl_h264_speed_setting->setText("Medium");
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_FAST:
        ui->lbl_h264_speed_setting->setText("Fast");
        break;
    case FAS_VIDEO::H264_SPEEDS::H264_VERY_FAST:
    default:
        ui->lbl_h264_speed_setting->setText("Very Fast");
        break;
    }
}

/*!
 * @brief The PC radiobutton has been toggled, either checked or cleared.
 * @param checked       True if the radiobutton is checked.
 */
void SaveDialog::do_pc_toggle(bool checked)
{
    if (checked)
    {
        // Enable browsing.
        ui->pb_browse->setEnabled(true);
        ui->rb_devices->setChecked(false);
        fill_save_type(MEDIA_LOCAL_PC);
        do_save_type_changed(ui->cb_save_type->currentIndex());
    }
    else
    {
        // Overlays are not allowed here.
        ui->chk_overlay->setEnabled(false);
        ui->chk_overlay->setChecked(false);

        // No browsing.
        ui->pb_browse->setEnabled(false);
    }
}

/*!
 * @brief The user clicked on the save XML checkbox.
 */
void SaveDialog::do_xml_select()
{
    // If user checks xml checkbox make sure pfmd is checked.
    if (ui->chk_xml->isChecked())
    {
        // Make sure it's allowed.
        if (!allow_xml)
        {
            ui->chk_xml->setChecked(false);
        }
        else if (!ui->chk_pfmd->isChecked())
        {
            ui->chk_pfmd->setChecked(true);
        }
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void SaveDialog::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;
    CameraMDI   *camera_ptr;

    // Is this event for us?
    if (ip_addr != cam_ip_addr)
        return;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
        return;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_JPEG_QUALITY:
        // Read the value from our cache.
        reg_value = read_camera_cache(GIGE_REG_JPEG_QUALITY);

        // Verify that the register setting is within a threshold.
        FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "reg_value", reg_value, false);

        // Don't want to overwrite a jpeg quality that the user has set already.
        if (firstJpegRegRead == 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Initializing with this GIGE_REG_JPEG_QUALITY read:" << reg_value;
            firstJpegRegRead = reg_value;
            // Save the JPEG quality.
            set_jpeg_quality((int)reg_value);
            reg_jpeg_quality = reg_value;   // Set our register value from the camera.
        }
        break;
    default:
        break;
    }
}
