/*!
 * @file save_dialog.h
 * @brief This file contains the definitions for the SaveDialog class.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef SAVE_DIALOG_H
#define SAVE_DIALOG_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include <QDialog>
#include <stdint.h>

#include "save_video.h"

namespace Ui
{
class SaveDialog;
}

namespace SAVE_DLG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! A value to indicate that the user selected the local workstation
//! for the destination location.
const uint32_t  MEDIA_LOCAL_PC          = (uint32_t) -1;

//!< The default framerate for playing back AVI's.
const int       DEF_AVI_PLAY_RATE       = 30;



}   // End of namespace SAVE_DLG.

//--------------------------------------------------------------------
// Class SaveDialog
//--------------------------------------------------------------------
class SaveDialog : public QDialog
{
    Q_OBJECT

public:
    SaveDialog(uint32_t ip_addr = 0, QWidget *parent = nullptr);
    ~SaveDialog();

    static const QString SAVE_SETTINGS_SECTION;
    static const QString SETTINGS_AVI_SAVE_JPEG;
    static const QString SETTINGS_MP4_H264_QUALITY;
    static const QString SETTINGS_MP4_H264_SPEED;
    static const QString SETTINGS_MP4_H264_ALLKEY;

    static void getH264Settings(FAS_VIDEO::H264_SPEEDS& h264Speed, uint32_t& h264Quality, bool& h264AllKeys);

    uint32_t    get_clip_end();
    uint32_t    get_clip_start();
    QString     get_dest_label();
    uint32_t    get_dest_media();
    int         get_jpeg_quality();
    uint32_t    get_h264_quality();
    uint32_t    get_playback_rate();
    int         get_rel_clip_end();
    int         get_rel_clip_start();

    FI_SAVER::STREAM_TYPE GetStreamType();
    FAS_VIDEO::H264_SPEEDS GetH264Speed();
    bool GetH264EnableAllKeyFrames();

    const QString   get_comment() const;
    const QString   get_dest_path() const;
    const QString   get_tag_label() const;

    void    get_save_params(FI_SAVER::SAVE_PARAMS& save_params, bool is_series3);

    bool    create_xml();
    void    disallow_raw()            { allow_raw = false; }
    void    disallow_xml()            { allow_xml = false; }
    void    enable_overlay_merge(bool do_enable);
    void    init_controls(bool query_cam = true);
    bool    merge_overlays();
    bool    use_avi_jpeg_codec();
    bool    use_camera_name();
    bool    use_camera_range();
    bool    use_comment();
    bool    use_tag_label();

    void    set_autosave_mode(bool do_enable);
    void    set_avi_jpeg_codec(bool use_jpeg_codec);
    void    set_cap_transfer(bool do_enable);
    void    set_clip_range(uint32_t start_clip, uint32_t end_clip);
    void    set_dest_media(uint32_t media_sel, QString& media_label);
    void    set_direct_transfer(bool do_enable);
    void    set_jpeg_quality(int quality_value);
    void    set_h264_quality(int quality_value);
    void    set_h264_speed(int speed_value);
    void    set_h264_allkeyframe(bool enable_allkeyframe);
    void    set_multicam(bool do_multicam, int camera_count);
    void    set_multicam_mixed(bool mixed_group);
    void    set_num_frames(uint32_t frame_count);
    void    set_rel_range(int start_clip, int end_clip);
    void    set_sensor_type(uint32_t type)                { sensor_type = type; }
    void    set_trig_frame(uint32_t frame_num)            { trig_frame = frame_num; }
    void    set_use_comment(bool use_it);
    void    set_use_tag_label(bool use_it);
    void    set_view_only(bool view_only);

    static void savePath(uint32_t cam_ip_addr, const QString& download_base_path);
    static const QString loadPath(uint32_t cam_ip_addr);

public slots:
    virtual int exec();

private slots:
    void    device_selected(int index);
    void    do_bmp_select();
    void    do_browse();
    void    do_cam_range();
    void    do_chk_comment(int state);
    void    do_chk_overlay();
    void    do_chk_tag_label(int state);
    void    do_comment_changed(const QString& comment_str);
    void    do_device_select();
    void    do_h264_speed_changed(int index);
    void    do_h264_allkeyframes_select();
    void    do_jpg_select();
    void    do_jpeg_quality_changed(int jpeg_quality_new_value);
    void    do_metadata_cfg_select();
    void    do_pc_toggle(bool checked);
    void    do_pfmd_select();
    void    do_raw_select();
    void    do_save_type_changed(int index);
    void    do_xml_select();
    void    handle_json_media(const QJsonDocument& jsonMedia);
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr);

    virtual void accept();

private:
    void        connect_camera();
    void        disconnect_camera();
    void        enable_controls(bool do_enable);
    void        fill_save_type(uint32_t media_sel);
    void        init_ranges();
    uint32_t    read_camera_cache(uint32_t reg_addr, bool *ok = nullptr);
    void        restore_settings(bool series3);
    int         select_save_type(int save_type);
    void        update_imgfmt_ctls(FI_SAVER::STREAM_TYPE streamType, bool isSeries3, uint32_t dest_media);
    void        update_rawcb_ctls(FI_SAVER::STREAM_TYPE streamType);
    void        update_h264_ctls(FI_SAVER::STREAM_TYPE streamType);
    void        update_media(uint32_t reg_value);
    bool        validate_comment();

private:
    Ui::SaveDialog *ui;

    bool        allow_overlay;                          //!< True if the caller wants to allow overlays.
    bool        allow_raw;                              //!< True if RAW selection is allowed, false otherwise.
    bool        allow_xml;                              //!< False to disable XML selection. True MAY allow XML, depending on other factors.
    bool        allow_planar;                           //!< True if sensor supports BMP Planar streams for NV12 or Y8.
    bool        allow_query_cam;                        //!< True if allowing camera queries.
    bool        autosave_mode;                          //!< True if we're in autosave setup mode, false for actually saving a file.
    uint32_t    cam_ip_addr;                            //!< The camera's IP address.
    bool        cam_connected;                          //!< True if we've connected signals/slots with a camera.
    bool        cap_transfer_mode;                      //!< True if we're in CAP file transfer mode.
    uint32_t    clip_start;                             //!< User specified start of clip.
    uint32_t    clip_end;                               //!< User specified end of clip.
    bool        fflr_enabled;                           //!< True if FasFireLR (FFLR) is enabled.
    bool        initialized;                            //!< True if we're initialized, false otherwise.
    bool        ltr_enabled;                            //!< True if LR mode is enabled.
    bool        multicam_mode;                          //!< True if we're in multicamera mode.
    bool        multicam_mixed;                         //!< True if we have a mixed collection of cameras.
    uint32_t    num_frames;                             //!< Number of frames in the session.
    bool        raw_selected;                           //!< True if TIF RAW was selected last time a save was performed.
    bool        metadata_cfg_selected;                  //!< True if saving partition metadata to .CFG (Xcite) file with save.
    uint32_t    reg_jpeg_quality;                       //!< Persist the register setting for jpeg quality to restore after MP4 saves.
    uint32_t    firstJpegRegRead;                       //!< What value is in the very first jpeg register read?
    uint32_t    sensor_type;                            //!< Type of sensor we're dealing with.
    uint32_t    trig_frame;                             //!< Which frame is the trigger frame.
    bool        view_only;                              //!< True if we're view only.
};

#endif // SAVE_DIALOG_H
