/*!
 * @file save_snapshot.cpp
 * @brief The code to display and save a snapshot image.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "save_snapshot.h"
#include "ui_save_snapshot.h"

#include <stdint.h>
#include <QFile>
#include <QFileDialog>
#include <QGraphicsPixmapItem>
#include <QGraphicsScene>
#include <QPixmap>
#include <QRegularExpression>
#include <QRegularExpressionValidator>
#include <QResizeEvent>
#include <QTimer>

#include "camera_mdi.h"
#include "fi_comm.h"
#include "fi_msgbox.h"
#include "logger.h"
#include "mainwindow.h"
#include "save_dialog.h"

namespace SAVE_SNAPSHOT
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Define the period between save status queries.
const int   QUERY_PERIOD            = 100;

//! Define the period to wait before sending a resize command to dialog.
const int QUERY_PERIOD_FOR_RESIZE   = 300;

}   // End of namespace SAVE_SNAPSHOT.

using namespace SAVE_SNAPSHOT;

/*!
 * @brief The class constructor.
 */
SaveSnapshot::SaveSnapshot(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::SaveSnapshot)
{
    QList<int>  size_hints;

    // Initialize our UI.
    ui->setupUi(this);

    // Adjust the proportions of the splitter window.
    size_hints.append(400);
    size_hints.append(100);
    ui->splitter->setSizes(size_hints);

    // Create our default scene.
    scene_ptr = new QGraphicsScene(this);

    // Add our new scene to the graphics view object.
    ui->image_view->setScene(scene_ptr);

    // Initialize our controls.
    init_controls();
    err_received = false;

    // Initialize our signals/slots.
    connect(ui->chk_tag_name, &QCheckBox::stateChanged, this, &SaveSnapshot::do_chk_tag_label);
    connect(ui->rb_pc, &QRadioButton::toggled, this, &SaveSnapshot::do_toggle_pc);
    connect(ui->buttonBox, &QDialogButtonBox::accepted, this, &SaveSnapshot::do_save);
    connect(ui->buttonBox, &QDialogButtonBox::rejected, this, &SaveSnapshot::reject);

    update();
}

/*!
 * @brief The class destructor.
 */
SaveSnapshot::~SaveSnapshot()
{
    if (scene_ptr != nullptr)
    {
        delete scene_ptr;
        scene_ptr = nullptr;
    }

    delete ui;
}

/*!
 * @brief Ask the user for a file name to save the snapshot.
 */
void SaveSnapshot::get_filename(QString& file_name)
{
    QString     download_path;
    QString image_filename;
    QSettings   settings;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Initialize the default path based on the saved settings.
    download_path = SaveDialog::loadPath(my_parent->get_ip_addr());

    // Make sure the path will look "right" to the user.
    image_filename = QDir::toNativeSeparators(download_path);

    // Ask the user for the file name.
    file_name = QFileDialog::getSaveFileName(this, tr("Select A Directory and Filename to save the snapshot image."), image_filename, tr("Images (*.jpg)"));
}

/*!
 * @brief Initialize our dialog boxes controls.
 */
void SaveSnapshot::init_controls(void)
{
    uint32_t    reg_value;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // What media is available on the camera?
    reg_value = my_parent->read_camera_cache(GIGE_REG_AVAIL_MEDIA);
    ui->rb_sdcard->setEnabled((reg_value & GIGE_MEDIA_SDCARD) == GIGE_MEDIA_SDCARD);
    ui->rb_ssd->setEnabled((reg_value & GIGE_MEDIA_SSD) == GIGE_MEDIA_SSD);
    ui->rb_usb->setEnabled((reg_value & GIGE_MEDIA_USB) == GIGE_MEDIA_USB);

    // Default to the local PC for our destination.
    ui->rb_pc->setEnabled(true);
    ui->rb_pc->setChecked(true);
    do_toggle_pc(true);

    // Setup a regex to validate the tag label, so the user can't type in
    // an invalid value.
    QRegularExpression rx("[-_.A-Za-z0-9]+");
    QRegularExpressionValidator *validator = new QRegularExpressionValidator(rx, ui->eb_tag_label);
    ui->eb_tag_label->setValidator(validator);

    // Restore our settings from previous instances.
    restore_settings();

    // Save is still on-going. Wait a little and then query again.
    QTimer::singleShot(QUERY_PERIOD_FOR_RESIZE, this, &SaveSnapshot::force_a_resize);
}

void SaveSnapshot::process_save_status(void)
{
    uint32_t    reg_value;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Make sure we got the right parent.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Get the save status value.
    reg_value = my_parent->read_camera_cache(GIGE_REG_SAVE_STATE);

    // If we haven't gotten an error, but there is one now, let's log it.
    if (!err_received && ((reg_value & GIGE_SAVE_STATE_ERR) == GIGE_SAVE_STATE_ERR))
    {
        uint32_t    err_value = reg_value & ~GIGE_SAVE_STATE_FLAGS;
        QString     info_str;

        // We had a problem saving the snapshot. Log the error. A message box will be put up
        // elsewhere (specifically SaveVideo::save_state_change).
        info_str = tr("The error code is: %1 (%2)").arg(err_value).arg(strerror((int)err_value));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to save the snapshot." << info_str;

        // Remember that we got an error.
        err_received = true;
    }
    else if ((reg_value & GIGE_SAVE_STATE_IDLE) == (uint32_t)GIGE_SAVE_STATE_IDLE)
    {
        FI_comm     *comm_ptr;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
            return;
        }

        // Don't want anymore read complete status.
        disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &SaveSnapshot::read_complete);

        // All done.
        if (!err_received)
            QDialog::accept();
    }
    else
    {
        // Save is still on-going. Wait a little and then query again.
        QTimer::singleShot(QUERY_PERIOD, this, &SaveSnapshot::query_save_status);
    }
}

/*!
 * @brief Restore any saved settings so that we can present the user with
 *        the settings last used.
 */
void SaveSnapshot::restore_settings(void)
{
    QSettings   settings;
    QString     tag_str;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Make sure we got the right parent.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    settings.beginGroup("SaveSettings");
    {
        // Retrieve the strings which may have been saved.
        tag_str = settings.value("tag_string").toString();

        // Restore the checkboxes.
        ui->chk_cam_name->setChecked(settings.value("use_camera_name").toBool());
    }
    settings.endGroup();

    // Is the tag empty?
    if (tag_str.isEmpty())
    {
        QDate   today = QDate::currentDate();
        tag_str = today.toString(Qt::ISODate);
    }

    // Set the tag value.
    ui->eb_tag_label->setText(tag_str);
}

/*!
 * @brief The user has selected to save to one of the camera's storage devices.
 */
void SaveSnapshot::save_to_camera(void)
{
    uint32_t    reg_value;
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Make sure we got the right parent.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Does the user want a tag in the file name?
    if (ui->chk_tag_name->isChecked())
    {
        QByteArray  tag_name;

        // Get the tag string.
        tag_name = ui->eb_tag_label->text().toUtf8();
        tag_name.append('\0');

        // Write it to the camera.
        my_parent->write_camera_buffer(GIGE_REG_IMAGE_TAG, tag_name);
    }

    // Make sure we tell the camera we're saving a snapshot.
    reg_value = GIGE_IMAGE_SNAPSHOT;

    // Make sure the camera knows we want it in JPEG format.
    reg_value |= GIGE_STREAM_FORMAT_JPEG;

    // Are we using a tag?
    if (ui->chk_tag_name->isChecked())
        reg_value |= GIGE_IMAGE_USE_TAG;

    // Are we using the camera name?
    if (ui->chk_cam_name->isChecked())
        reg_value |= GIGE_IMAGE_USE_CAMERA_NAME;

    // Specify the destination.
    if (ui->rb_sdcard->isChecked())
        reg_value |= GIGE_IMAGE_DEST_SDCARD;
    else if (ui->rb_ssd->isChecked())
        reg_value |= GIGE_IMAGE_DEST_SSD;
    else if (ui->rb_usb->isChecked())
        reg_value |= GIGE_IMAGE_DEST_USB;
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No valid media selections!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Setup to monitor the save proceess.
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &SaveSnapshot::write_complete, Qt::QueuedConnection);

    // Get the camera started.
    my_parent->write_camera_register(GIGE_REG_IMAGE_DESTINATION, reg_value);

    // Start polling the save status.
    QTimer::singleShot(QUERY_PERIOD, this, &SaveSnapshot::query_save_status);
}

/*!
 * @brief The user has selected to save to the PC.
 */
void SaveSnapshot::save_to_pc(void)
{
    bool        file_status;
    QString     file_name;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    QString err_str;

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Get the full path to our file. Convert the sequence number from 1-based to 0-based.
    get_filename(file_name);
    if (file_name.isEmpty())
        return;

    file_status = snap_pixmap.save(file_name, "JPG");
    if (!file_status)
    {
        // We had a problem opening the file.
        err_str = tr("Unable to save the snapshot to '%1'.").arg(file_name);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        fiMsgBox::critical(my_parent, err_str);
    }
    else
    {
        // All done.
        QDialog::accept();
    }
}

/*!
 * @brief Get a copy of the image we're going to use.
 * @param snap_image    The pixmap of our snapshot.
 */
void SaveSnapshot::set_image(const QPixmap& snap_image)
{
    snap_pixmap = snap_image;
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Update our image during a resize of the dialog.
 * @param event     [UNUSED]
 */
void SaveSnapshot::resizeEvent(QResizeEvent *event)
{
    QGraphicsPixmapItem *pixmap_item_ptr;

    Q_UNUSED(event);
    //QLOG_INFO() << __PRETTY_FUNCTION__ << "Resizing...";

    // Sanity check.
    if (scene_ptr == nullptr)
        return;

    // Create the new pixmap item.
    pixmap_item_ptr = scene_ptr->addPixmap(snap_pixmap);

    // We'll start out by resetting the transform values.
    // This has the side effect of getting us to 1:1.
    ui->image_view->resetTransform();

    // Scale the image into our viewport.
    ui->image_view->fitInView(pixmap_item_ptr, Qt::KeepAspectRatio);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user clicked on the tag checkbox, either selecting or de-selecting it.
 * @param state     The state of the checkbox.
 */
void SaveSnapshot::do_chk_tag_label(int state)
{
    ui->eb_tag_label->setEnabled(state == Qt::Checked);
}

/*!
 * @brief The user is ready to save the image.
 */
void SaveSnapshot::do_save(void)
{
    // Are we saving to the PC or the camera's media?
    if (ui->rb_pc->isChecked())
        save_to_pc();
    else
        save_to_camera();
}

/*!
 * @brief The PC radiobutton has been toggled, either checked or cleared.
 * @param checked       True if the radiobutton is checked.
 */
void SaveSnapshot::do_toggle_pc(bool checked)
{
    ui->camera_options->setEnabled(!checked);
}

/*!
 * @brief It's time to ask the camera how the save is going.
 */
void SaveSnapshot::query_save_status(void)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Request the save status value from the camera.
    my_parent->start_camera_read_reg(GIGE_REG_SAVE_STATE);
}


/*!
 * @brief Send a resize request to the dialog to force readjustment of image.
 * This was added because the initialization of the dialog does not properly
 * set the size of the QPixmap in display for the snapshot.  It appears as a
 * small thumbnail.  A one-shot timer is set during the init_controls() function
 * to call this function one time after a set period.
 */
void SaveSnapshot::force_a_resize(void)
{
    //QLOG_INFO() << __PRETTY_FUNCTION__ << "FORCE A RESIZE...";
    resizeEvent(nullptr);
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void SaveSnapshot::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Did you give me a CameraMDI for a parent?!";
        return;
    }

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_SAVE_STATE:
        process_save_status();
        break;
    default:
        break;
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void SaveSnapshot::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    uint32_t    write_status;
    FI_comm     *comm_ptr;

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_IMAGE_DESTINATION:
        // Get the write status.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status != GEV_STATUS_SUCCESS)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to write GIGE_REG_IMAGE_DESTINATION, status:" << Qt::showbase << Qt::hex << write_status;
        }

        // Done with the writes.
        disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &SaveSnapshot::write_complete);

        // Make sure we start seeing the reads.
        connect(comm_ptr, &FI_comm::read_reg_complete, this, &SaveSnapshot::read_complete, Qt::QueuedConnection);
        break;
    default:
        break;
    }
}
