/*!
 * @file save_video.h
 * @brief This is the header file for the SaveVideo class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef SAVE_VIDEO_H
#define SAVE_VIDEO_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QObject>
#include <QString>

#include <stdint.h>

#include <memory>

#include "fi_image_buffer.h"
#include "fi_saver.h"
#include "fi_progress_dialog.h"

namespace SAVE_VIDEO
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Camera device save states for processing a camera device save request.
typedef enum
{
    SAVE_NONE,                  //!< There is no state yet ready for a device save request.
    SAVE_SETUP_ERROR,           //!< Device save set up has encountered an error.
    SAVE_SETUP_DENIED,          //!< Device save is already outstanding cannot request again until receive response.
    SAVE_REQUESTED,             //!< Wait for the write complete indication as to success of device save request.
    SAVE_DENIED,                //!< The camera cannot start the device save operation at this time.
    SAVE_PROGRESS,              //!< Camera has acknowledged the device save request and should be processing it.
    SAVE_DEARM_AND_KEEP,        //!< Camera is in process of dearming but wants to keep saving.
    SAVE_CANCELLED,             //!< The user has cancelled the device save request.
} DEVICE_SAVE_STATE;

//! FasFireLR states for processing underflow condition.
typedef enum
{
    NONE,                       //!< There is no underflow condition.
    WAIT_FOR_IDLE,              //!< Wait for an GEV_SAVE_STATE_IDLE bit before starting recovery operation.
    READY_TO_COMPLETE,          //!< GEV_SAVE_STATE_IDLE has been received ready to set up recovery operation.
    COMPLETE_SETUP_ERROR,       //!< There was a complete set up error.
    COMPLETE_REQUEST,           //!< Asking camera to start underflow recovery.
    COMPLETE_REQUEST_FAILED,    //!< Camera could not start underflow recovery.
    COMPLETE_PROGRESS,          //!< Watching for save state progress updates as camera started recovery.
    COMPLETE_PROGRESS_ERROR,    //!< Camera reports an error while processing recovery.
    COMPLETE_PROGRESS_CANCEL,   //!< User has cancelled while processing progress updates.
    DISCARDED,                  //!< User discarded the partial recording we are done.
} FFLR_UNDERFLOW_STATE;

//! Define a structure to hold the many parameters required for a copy operation.
typedef struct
{
    bool        use_camera_name;                //!< True if we're to use the camera's name in the file name(s).
    bool        use_existing_name;              //!< True to use the existing name, useful for metadata file downloads.
    bool        lastCopyRequest;                //!< Is this the last copy request expected or are there more in same request?
    uint32_t    first_frame_num;                //!< Absolute frame number of the first frame of the sequence.
    uint32_t    last_frame_num;                 //!< Absolute frame number of the last frame of the sequence.
    uint32_t    metaFileCount;                  //!< How many metafiles are expected?
    FI_SAVER::STREAM_TYPE streamType;           //!< The type of stream this saver is supporting.
    FI_SAVER::STREAM_TYPE videoType;            //!< The TRUE type of stream.
    FI_SAVER::IMAGE_TYPE imageInputType;        //!< The type of image frame this stream expects to receive.
    FI_SAVER::IMAGE_TYPE imageOutputType;       //!< The type of image frame this stream expects to produce.
    QString     dest_path;                      //!< The path to the folder to hold the saved capture.
    QString     src_path;                       //!< Fully qualified path of the source to be copied.
    QString     tag_label;                      //!< The tag to use, if any, in the file name(s).
    QString     metaFileExt;                    //!< If use_existing_name this is the extension to use.
} COPY_PARAMS;

const QString SAVE_SRC_RECEIVE_FRAME = QStringLiteral("Received Frame:");
const QString SAVE_SRC_RECEIVE_BLOB  = QStringLiteral("Received Buffer:");

}   // End of namespace SAVE_VIDEO.

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------

class   QByteArray;
class   QTimer;

//--------------------------------------------------------------------
// Class Definition
//--------------------------------------------------------------------
class SaveVideo : public FiSaver
{
    Q_OBJECT

public:
    explicit SaveVideo(QObject *parent = nullptr);
    ~SaveVideo();

    bool    autosave_is_enabled()                         { return autosave_enabled; }
    bool    autosave_is_configured()                      { return autosave_configured; }
    void    autosave_cancelled();
    void    autosave_display();
    bool    autosave_setup();
    bool    autosave_start();
    bool    autosave_stop();
    bool    autosave_wants_pm_metadata();
    bool    cancelling()                                  { return cancel_in_progress; }
    void    cleanup_save();
    void    close(const QString& whoCalledMe);
    bool    configure(FI_SAVER::SAVE_PARAMS save_info);
    void    copy_finished();
    bool    copy_to_pc(SAVE_VIDEO::COPY_PARAMS copy_info);
    bool    copy_to_pc_invoked()                          { return calledByCopyToPC; }
    void    dearm_and_keep();
    bool    do_merge_overlay()                            { return merge_overlays; }
    void    going_live();
    bool    is_fflr_underflow_recovering();
    bool    lost_frame(int seq_num);
    bool    receive_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr);

    void    reset_stats();
    bool    save_recording();
    bool    saving_xml()                                  { return convertRawToXml; }
    void    start(const QString& whoCalledMe);
    int     setup(bool amICopyToPC = false);

    void    setTrigNo4Frame(uint32_t trigNo4FrameToSet)      { trigFrameNo4Save = trigNo4FrameToSet; }

    uint32_t    get_copy_start();
    uint32_t    get_last_save_state()                     { return last_save_state; }
    uint32_t    get_metafile_count()                        { return copyMetaFileCount; }
    uint32_t    get_metafile_processed_count()              { return copyMetaFilesProcessed; }

    bool getSaveStatus(QString& infoStr, uint32_t& frameCount, uint32_t& framesAdded, uint32_t& framesEncoded,
                       bool& saverIsOk, bool& saverIsComplete);
    virtual bool getVideoStreamSaveStatus(QString& infoStr, uint32_t& frameCount, uint32_t& framesAdded,
                                          uint32_t& framesEncoded, bool& streamIsOk, bool& streamIsComplete);
    virtual int prepareStreamWriter() override;

    bool    local_autosave()                              { return autosave_enabled && local_autosave_enabled; }
    bool    save_in_progress()                            { return image_thread.isRunning(); }
    void    saving_finished(const QString& whoCalledMe);
    void    setRetryFrame(uint32_t retry_frame_abs, int retry_frame_rel, uint32_t retry_frames_resent, uint32_t total_retry_frames);

    void    update_image_dest(uint32_t img_dest);

signals:
    // These all should be QT::QueuedConnection!
    void    free_list_ready();                  //!< Make sure this is a Qt::QueuedConnection!
    void    save_complete();                    //!< Make sure this is a Qt::QueuedConnection!
    void    save_error();                       //!< Make sure this is a Qt::QueuedConnection! An error has occurred. Refer to last_err_msg for information.
    void    save_fflr_underflow_error();        //!< A FasFireLR underflow error arose.
    void    save_receive_count_complete();      //!< Signals non-empty frames received equate to frame count expected.
    void    saverBufferReleaseSignal(uint32_t bufferSeqNbr);    //!< Indicate that a saver buffer is about to be released.
    void    work_list_empty();                  //!< Make sure this is a Qt::QueuedConnection!
    void    worker_thread_exiting();            //!< Signals worker thread is exiting. Make sure this is a Qt::QueuedConnection!

    void    userCancelledSaveSignal(void);      //!< Notifies outside world of user save cancellation.

    // These all should be QT::QueuedConnection!
    void    save_json_write_complete(uint32_t status);          //!< Notification of write complete for GIGE_REG_SAVE_JSON.
    void    save_json_response(uint32_t status);                //!< Final response of GIGE_REG_SAVE_JSON request.

public slots:
    void    cancel_save(const QString& whoCalledMe);
    void    raw_metadata_ready();
    void    save_frame_progress(uint32_t last_frame_saved);
    void    save_state_change(uint32_t new_state);
    void    update_progress();
    void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private slots:
    void    do_fflr_underflow_cancelled();
    void    do_fflr_underflow_progress();
    void    do_fflr_underflow_save_json_response(uint32_t write_status);
    void    do_save_json_response(uint32_t status);
    void    do_save_json_write_complete(uint32_t status);
    void    free_buffer();
    void    save_state_poll_timeout();
    void    process_fflr_underflow();
    void    timeForVideoProgressPollUpdate();   //!< Catch time to update the progress of polling save.
    void    timeForStackProgressFrameUpdate(uint32_t bufferSeqNbr);  //!< Catch a time to update progress of a stack update.
    void    userHasCancelled();                 //!< Catch the transfer dialog cancel save signal.
    void    video_is_paused();
    void    do_worker_thread_exit();            //!< Catches exit of the worker thread for GUI thread level processing.
private:
    static void image_saver(QObject *parent);

    bool        calc_blob_count_for_cap(uint32_t num_frames_to_save, uint32_t& num_blobs_calculated, uint64_t& space_needed);
    void        finishSave();
    void        get_autosave_status_msg(QString& status_msg, uint32_t& total_count, uint32_t& dirty_count);
    uint32_t    get_buffer_released_count();
    void        inc_buffer_released_count();
    void        persist_save_settings(FI_SAVER::SAVE_PARAMS save_info);
    bool        PrepXferDialogFromSaveParams(FI_SAVER::SAVE_PARAMS theSaveParmsToUse, bool savingToPC, const QString& whoCalledMe);
    bool        PrepXferDialogFromFSaver(const QString& whoCalledMe);
    bool        queryFreePartition();
    void        release_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr);
    bool        ReleaseImageFreeList(bool stopWorkerThread, bool emitBufferRelease);
    bool        ReleaseImageRecvList();
    void        ResetSaveVideoState();
    void        RestoreJPEGQFactor();
    bool        saveToStream(const uint8_t *imageDataPtrToWrite, uint32_t imageDataLenToWrite, uint32_t imageDataSeqNbr);
    bool        save_to_camera(FI_SAVER::SAVE_PARAMS save_info);
    bool        save_to_camera_json(FI_SAVER::SAVE_PARAMS save_info);
    bool        save_to_pc(FI_SAVER::SAVE_PARAMS save_info);
    bool        set_jpeg_quality_old_camset(int value_to_set, const QString& who_called_me);
    bool        StopWorkerThreadIfRunning();
    void        UpdateXferDialogForCancelButtonHit();
    void        update_autosave_setup();
    void        update_stats();

    SAVE_VIDEO::DEVICE_SAVE_STATE write_save_json_request(const QJsonDocument& json_save_request);

private:

    QMutex      FiSaverStreamMutex;                             //!< Mutex to guard access to the fiSaver stream object.
    QTimer      *save_state_poll_timer;                         //!< Timer used when saving to the camera allowing us to poll the GIGE_REG_SAVE_STATE status.

    std::unique_ptr<fiProgressDialog> xferDialog;               //!< A xfer progress dialog managed internally.

    bool        setupForSave;                                   //!< True if object is setup properly for a save, False if not.
    bool        savingHasFinished;                              //!< True if this save has marked the end of the save operation.
    uint32_t    saveBufferCount;                                //!< What is the expected save buffer count for a setup save operation?
    bool        calledByCopyToPC;                               //!< True if object was setup as a copy to PC.
    bool        emittedSaveComplete;                            //!< True if emitted "save_complete()" for active save process, False if not.  Reset when setupForSave is reset.
    bool        emittedSaveError;                               //!< True if emitted "save_err()" for active save process, False if not.  Reset when setupForSave is reset.
    bool        emittedSaveReceiveCountComplete;                //!< True if emitted "save_receive_count_complete()" for active save process, False if not.  Reset when setupForSave is reset.
    bool        closeInProgress;                                //!< True if we are already closing the active save process, False if not.
    bool        closeCompleted;                                 //!< True if we already closed the active save process, False if not.

    bool        cancel_in_progress;                             //!< True if we're waiting for a cancel to complete.
    FI_SAVER::STREAM_TYPE copyVideoType;                        //!< The TRUE type of stream for copying blobs.

    int         jpeg_quality_old_camset;                        //!< JPEG quality factory that was set prior to our modification of the camera register.

    bool        localSaveToPath;                                //!< True if we're saving locally to path.
    bool        merge_overlays;                                 //!< True if we're going to merge overlay information.
    bool        reuse_path;                                     //!< True to reuse the path information, useful for metafile downloads.
    bool        waiting_for_pause;                              //!< True if we're waiting for the camera to pause, false otherwise.

    uint32_t    copyBlockCount;                                 //!< When copying data keep count of the number of blocks written.
    uint32_t    copyMetaFileCount;                              //!< How many metafiles expected in this copy operation?
    uint32_t    copyMetaFilesProcessed;                         //!< How many metafiles processed in this copy operation?
    uint32_t    copy_start_frame;                               //!< First frame of the copy sequence.
    uint32_t    copy_end_frame;                                 //!< Last frame of the copy sequence.
    uint32_t    last_save_state;                                //!< The last GIGE_REG_SAVE_STATE value, so we can detect changes.

    bool        saving_finished_in_process;                     //!< Don't allow double entries in the saving finished process.

    // Autosave variables.
    bool        autosave_enabled;                               //!< True if autosave is enabled (either local or remote).
    bool        autosave_configured;                            //!< Has autosave been configured?
    bool        local_autosave_enabled;                         //!< True if autosave mode is enabled to the local workstation, false otherwise.
    int         as_clip_end;                                    //!< Relative frame number for the end of the save.
    int         as_clip_start;                                  //!< Relative frame number for the start of the save.
    uint32_t    as_config_width;                                //!< The width at time of autosave setup.
    uint32_t    as_config_height;                               //!< The height at time of autosave setup.
    uint32_t    as_config_bit_depth;                            //!< The bit-depth at time of autosave setup.
    uint32_t    as_config_part_frames;                          //!< The partition frame count at time of autosave setup.

    FI_SAVER::SAVE_PARAMS as_save_info;                         //!< Initialized during autosave_setup.  Reused in autosave_start.

    // Statistics variables.
    uint64_t        download_byte_count;                        //!< How many bytes did we download?
    uint32_t        download_frame_count;                       //!< How many frames did we download and save to the stream?
    uint32_t        download_receive_frame_count;               //!< How many non-empty frames have we received?
    QElapsedTimer   download_duration;                          //!< Time the download took to complete.
    QDateTime       download_start;                             //!< Time the download started.

    //! After the save thread processes the image data it arrives back to the
    //! free list so that it can be released to restore resources.
    std::deque<std::unique_ptr<FI_MEMPOOL::FiImageBuffer>>imageBufferFreeList;
    QMutex imageBufferFreeMutex;                                //!< Mutex to guard access to imageBufferFreeList.

    //! As image data arrives over the network it will be put into this
    //! image receive list so that it can be processed.  The save thread
    //! will monitor this list waiting for something to do.
    std::deque<std::unique_ptr<FI_MEMPOOL::FiImageBuffer>> imageBufferRecvList;
    QMutex imageBufferRecvMutex;                                //!< Mutex to guard access to imageBufferRecvList.

    std::unique_ptr<FI_SAVER::SAVE_PARAMS> saveToCameraJSONData;    //!< If was asked to save to camera JSON this is copy of data provided.

    bool            error_is_logged;                            //!< True if we've already logged the save error.
    volatile bool   save_is_done;                               //!< Flag to tell our processing thread to not expect any more images.
    volatile bool   wait_for_save_complete;                     //!< Flag to tell our processing thread we are waiting for stream to complete.
    uint32_t        last_encoded_count;                         //!< When in wait_for_save_complete stores last encoded count to check for stagnation.
    uint32_t        same_encoded_count;                         //!< Keeps track of the number of status updates where last encoded count stagnant.
    bool            thread_told_video_complete;                 //!< Keeps track of a special condition for video completion to eliminate duplicate stats.
    volatile bool   thread_stop;                                //!< Flag to tell our processing thread to stop.
    QFuture<void>   image_thread;                               //!< Our image processing thread "handle".

    uint32_t        trigFrameNo4Save;                           //!< When save is started save the trigger frame setting.  It is used for absolute frame number calculations.
    uint32_t        releasedBufferCount;                         //!< How many buffers have we released from the process?
    QWaitCondition  worker_wakeup;                              //!< Synchronization object to inform the worker thread that there's something to do.

    // TODO: This is not a fully flushed out idea - it may be unnecessary and I may want to remove it.
    SAVE_VIDEO::DEVICE_SAVE_STATE active_device_save;           //!< Keeps track of a device save request if one is active.

    std::unique_ptr<fiProgressDialog> fflrRecoverDlg;           //!< A simple progress dialog for FasFireLR Recovery status.
    std::unique_ptr<fiProgressDialog> fflrDearmDlg;             //!< A simple progress dialog for FasFireLR Dearming status.
    uint32_t        fflr_underflow_wait_delays;                 //!< When waiting for review window to get at least 1 frame, this keeps track of the number of wait delays so we don't wait indefinitely.
    SAVE_VIDEO::FFLR_UNDERFLOW_STATE fflr_underflow_state;      //!< What is the state of underflow recovery, if any?
    uint32_t        fflr_underflow_dirty_count;                 //!< How many dirty buffers are we expecting for the recovery process?

    bool            write_save_json_requested;                  //!< Prevents overlapping requests for GIGE_REG_SAVE_JSON writes.
    QDateTime       write_save_json_reqtime;                    //!< A write save json request has been made, this is the time it was made.
    QDateTime       write_save_json_rsptime;                    //!< A write save json response has been received, this is the time it was received.
    uint32_t        write_save_json_reqstatus;                  //!< A write save json request status code, GEV_STATUS_UNKNOWN if no response received.
};

// The save video object is derived from the fi saver object but there
// are constants defined in the base class that need to be accessable to
// users of the save video class header.
using namespace FI_SAVER;

#endif // SAVE_VIDEO_H
