#include "select_device.h"
#include "ui_select_device.h"

#include <QByteArray>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonParseError>
#include <QPushButton>

#include "camera_mdi.h"
#include "fi_comm.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"

// Information strings for device operations.
const QString SelectDevice::device_query_label      = QT_TR_NOOP("Asking system for available devices...");
const QString SelectDevice::device_empty_label      = QT_TR_NOOP("No device is available for use.\nMake sure all connected devices are formatted.");
const QString SelectDevice::device_found_label      = QT_TR_NOOP("There are available devices for use.");

SelectDevice::SelectDevice(QWidget *parent) :
    QDialog(parent, Qt::CustomizeWindowHint | Qt::WindowTitleHint),
    ui(new Ui::SelectDevice),
    local_path_allowed(false)
{
    CameraMDI   *camera_ptr;

    // Initialize the UI.
    ui->setupUi(this);

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent);
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Disable the controls for now.
    enableControls(false);

    // Connect the communications interfaces that we're interested in.
    connect(camera_ptr, &CameraMDI::media_info_ready, this, &SelectDevice::handleJsonMedia, Qt::QueuedConnection);

    // Ask the camera for the available media. This will be caught
    // by the camera object, which will start the ball rolling.
    camera_ptr->start_camera_read_reg(GIGE_REG_AVAIL_MEDIA);

    ui->lbl_waiting->setText(SelectDevice::device_query_label);
}

SelectDevice::~SelectDevice()
{
    CameraMDI   *camera_ptr;

    // Clean up the UI elements.
    delete ui;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr != nullptr)
    {
        // Connect the communications interfaces that we're interested in.
        disconnect(camera_ptr, &CameraMDI::media_info_ready, this, &SelectDevice::handleJsonMedia);
    }
}

uint32_t SelectDevice::deviceSelected(QJsonObject& dev_object, QJsonArray& media_array)
{
    int i;
    uint32_t selected_type;
    QSettings settings;

    // Get the selected device.
    dev_object = ui->cb_devices->currentData().toJsonObject();
    if (dev_object.isEmpty())
    {
        // The user selected the local path.
        selected_type = -1U;
    }
    else
    {
        // Save the name of the currently selected device.
        settings.beginGroup("PathInfo");
        settings.setValue("device_selection", dev_object["tag"].toString());
        settings.endGroup();

        // Remember the type.
        selected_type = (uint32_t)dev_object["type"].toInt();
    }

    // Build an array of all entries in the media combobox.
    for (i = 0; i < ui->cb_devices->count(); ++i)
    {
        QJsonObject media_object = ui->cb_devices->itemData(i).toJsonObject();

        // Skip the path.
        if (media_object.isEmpty())
            continue;
        // Skip the select object.
        if (dev_object["name"].toString() == media_object["name"].toString())
            continue;

        // Add the device to our array.
        media_array.append(media_object);
    }

    // Return the device type.
    return selected_type;
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_AVAIL_MEDIA register value.
 */
void SelectDevice::handleJsonMedia(const QJsonDocument& jsonMedia)
{
    int item_index;
    double dev_size;
    QString base_label;
    QString device_label;
    QString device_sel;
    QJsonObject mediaObj;

    // Get the object from the document.
    mediaObj = jsonMedia.object();

    // Block updates from the combobox.
    ui->cb_devices->blockSignals(true);

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
        device_sel = ui->cb_devices->currentText();
    if (device_sel.isEmpty())
    {
        QSettings   settings;

        // See if we have a device selection saved away.
        settings.beginGroup("PathInfo");
        device_sel = settings.value("device_selection").toString();
        settings.endGroup();
    }

    // Clear the combobox.
    ui->cb_devices->clear();

    // Add back in found devices starting with the SDCard.
    if (mediaObj.contains("sdcard") && mediaObj["sdcard"].isArray())
    {
        QJsonArray dev_array = mediaObj["sdcard"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an sdcard device.
            dev_object["type"] = GIGE_MEDIA_SDCARD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("SDcard %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // Check for NVME SSD devices.
    if (mediaObj.contains("nvme") && mediaObj["nvme"].isArray())
    {
        QJsonArray dev_array = mediaObj["nvme"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an NVME device.
            dev_object["type"] = GIGE_MEDIA_SSD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                device_label = QString("SSD %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // Check for USB drives.
    if (mediaObj.contains("usb") && mediaObj["usb"].isArray())
    {
        QJsonArray dev_array = mediaObj["usb"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // ONLY add devices which have a mount path.
            if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
                continue;

            // Remember this is an USB device.
            dev_object["type"] = GIGE_MEDIA_USB;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("USB %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // Add the device to the combobox.
            ui->cb_devices->addItem(device_label, dev_object);
        }
    }

    // Is the local path allowed?
    if (local_path_allowed)
        ui->cb_devices->addItem(tr("Path..."));

    // If there is no external media then disable the devices controls.
    if (ui->cb_devices->count() == 0)
    {
        // Notify user we found no devices.
        ui->lbl_waiting->setText(SelectDevice::device_empty_label);

        // Make sure we're disabled, and force the PC selection.
        ui->cb_devices->setEnabled(false);
    }
    else
    {
        int sel_index = 0;

        // Tell the user we have devices to select.  This really is too
        // fast to disappear but it's being set anyway in case someone
        // wants to use it for any reason.
        ui->lbl_waiting->setText(SelectDevice::device_found_label);

        // Make sure we're enabled.
        ui->cb_devices->setEnabled(true);

        // Attempt to re-select the old value.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);
            if (sel_index < 0)
                sel_index = 0;
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);
    }

    // Allow signals to flow again.
    ui->cb_devices->blockSignals(false);

    // If there's only one device, just pretend the user selected that one.
    if (ui->cb_devices->count() == 1)
    {
        accept();
    }
    else if (ui->cb_devices->count() > 1)
    {
        // Hide the waiting label.
        ui->lbl_waiting->setVisible(false);

        // Update and resize the dialog.
        adjustSize();
        update();

        // Enable the controls.
        enableControls(true);
    }
}

void SelectDevice::enableControls(bool do_enable)
{
    // Enable/Disable all controls except for the cancel button.
    ui->frm_devices->setEnabled(do_enable);

    // We may not have an OK button up.
    ui->buttonBox->button(QDialogButtonBox::Ok)->setEnabled(do_enable);
}

void SelectDevice::changeEvent(QEvent *e)
{
    QDialog::changeEvent(e);
    switch (e->type())
    {
    case QEvent::LanguageChange:
        ui->retranslateUi(this);
        break;
    default:
        break;
    }
}
