/*!
 * @file select_drive.cpp
 * @brief Displays the found Fastec formatted cameras connected to the
 *        PC and lets the user select one of them to open.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "select_drive.h"
#include "ui_select_drive.h"

#include "fi_rawdrive.h"
#include "gige_msg.h"
#include "logger.h"

// Avoid polluting the global namespace.
namespace SELECT_DRIVE_DLG
{
//--------------------------------------------------------------------
// Constant, Typedefs and Macro Definitions
//--------------------------------------------------------------------

//! An enumeration of the various columns of data we'll support in our table.
enum CD_COLUMN_ID
{
    CD_DRIVE_URI = 0,
    CD_SERIAL_NUM,
    CD_CAMERA_NAME,
    CD_COL_COUNT
};
}   // End of namespace SELECT_DRIVE_DLG.

using namespace SELECT_DRIVE_DLG;

/*!
 * @brief SelectDrive class constructor.
 */
SelectDrive::SelectDrive(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::SelectDrive)
{
    QTableWidgetItem    *tbl_item;

    // Initialize our UI.
    ui->setupUi(this);

    // Setup the table headers.
    tbl_item = new QTableWidgetItem(tr("Drive Path"));
    ui->drive_table->setHorizontalHeaderItem(CD_DRIVE_URI, tbl_item);
    tbl_item = new QTableWidgetItem(tr("Serial Number"));
    ui->drive_table->setHorizontalHeaderItem(CD_SERIAL_NUM, tbl_item);
    tbl_item = new QTableWidgetItem(tr("Camera Name"));
    ui->drive_table->setHorizontalHeaderItem(CD_CAMERA_NAME, tbl_item);
}

/*!
 * @brief SelectDrive destructor.
 */
SelectDrive::~SelectDrive()
{
    delete ui;
}

QString SelectDrive::get_selected_drive(void)
{
    int     row_index;
    QString selected_uri;

    // Which is the row selected?
    row_index = ui->drive_table->currentIndex().row();

    // Just retrive the drive column from that row.
    selected_uri = ui->drive_table->item(row_index, CD_DRIVE_URI)->text();

    // Return the string.
    return selected_uri;
}

void SelectDrive::set_drive_list(const QStringList& drive_list)
{
    // Do we need to grow the table?
    if (drive_list.count() > ui->drive_table->rowCount())
        ui->drive_table->setRowCount(drive_list.count());

    // Loop through each of the devices and add to our table.
    for (int row_index = 0; row_index < drive_list.count(); ++row_index)
    {
        QString     drive_uri;
        QString     temp_str;
        fiRawDrive  raw_drive;
        pm_metadata meta_data;
        QTableWidgetItem    *tbl_item;

        // Get the drive URI for this row.
        drive_uri = drive_list.at(row_index);

        // Create a new item.
        tbl_item = new QTableWidgetItem(drive_uri);
        ui->drive_table->setItem(row_index, CD_DRIVE_URI, tbl_item);

        // Open the device.
        if (!raw_drive.open(drive_uri))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open:" << drive_uri;
            continue;
        }

        // Try to read the capture metadata.
        if (!raw_drive.readPmMetadata(&meta_data))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to read pm_metadata from:" << drive_uri;
            raw_drive.close();
            continue;
        }

        // Close the drive.
        raw_drive.close();

        // Extract the serial number.
        temp_str = QString::fromUtf8(meta_data.serial_num);
        tbl_item = new QTableWidgetItem(temp_str);
        ui->drive_table->setItem(row_index, CD_SERIAL_NUM, tbl_item);

        // Extract the camera name.
        temp_str = QString::fromUtf8(meta_data.camera_name);
        tbl_item = new QTableWidgetItem(temp_str);
        ui->drive_table->setItem(row_index, CD_CAMERA_NAME, tbl_item);
    }

    // We've updated all the cameras. Resize the columns in our table view.
    ui->drive_table->resizeRowsToContents();
    ui->drive_table->resizeColumnsToContents();
    ui->drive_table->horizontalHeader()->setStretchLastSection(true);

    // Select the first entry.
    ui->drive_table->setCurrentCell(0, 0);
}
