#include "select_folder.h"
#include "ui_select_folder.h"

#include <QByteArray>
#include <QDomDocument>
#include <QDomElement>
#include <QJsonDocument>
#include <QJsonObject>
#include <QJsonParseError>
#include <QPushButton>

#include "camera_mdi.h"
#include "client_math.h"
#include "fi_comm.h"
#include "fi_msgbox.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"

namespace SELECT_FOLDER_DLG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Our polling period when waiting for a file action to complete.
const int FA_POLLING_PERIOD             = 250;

//! The indecies into our tree widget.
enum
{
    COLUMN_NAME = 0,            //!< The name of the file/folder.
    COLUMN_SIZE,                //!< The size of the file/folder.
    COLUMN_TIME                 //!< The time of the file/folder.
};

}   // End of namespace SELECT_FOLDER_DLG.

using namespace SELECT_FOLDER_DLG;

SelectFolder::SelectFolder(QWidget *parent) :
    QDialog(parent, Qt::CustomizeWindowHint | Qt::WindowTitleHint),
    ui(new Ui::SelectFolder),
    file_timer_id(0),
    listing_len(0)
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Initialize the UI.
    ui->setupUi(this);

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent);
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Connect the communications interfaces that we're interested in.
    connect(camera_ptr, &CameraMDI::closing, this, &SelectFolder::reject);

    connect(comm_ptr, &FI_comm::read_reg_complete, this, &SelectFolder::readComplete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &SelectFolder::writeComplete, Qt::QueuedConnection);

    connect(ui->cam_tree, &QTreeWidget::expanded, this, &SelectFolder::doTreeExpansion);
    connect(ui->cam_tree, &QTreeWidget::currentItemChanged, this, &SelectFolder::doTreeItemChanged);
    connect(ui->cam_tree, &QTreeWidget::itemSelectionChanged, this, &SelectFolder::doTreeSelectionChanged);
    connect(ui->cam_tree, &QTreeWidget::itemDoubleClicked, this, &SelectFolder::doTreeItemDoubleClicked);
}

/*!
 * @brief Standard destructor method.
 */
SelectFolder::~SelectFolder()
{
    // Clean up the UI elements.
    delete ui;

    // Shutdown the timers.
    if (file_timer_id != 0)
    {
        killTimer(file_timer_id);
        file_timer_id = 0;
    }
}

/*!
 * @brief Return the selected entry.
 */
QString SelectFolder::folderSelected()
{
    int row;
    QString selected_path;
    QList<QTreeWidgetItem *> selected_items;
    const CAM_EXPLORE::FILE_INFO *file_info;

    // Get the selected items. There should only be the one.
    selected_items = ui->cam_tree->selectedItems();
    if (selected_items.count() != 1)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "There should only be one!" << selected_items;
        goto bail;
    }

    // Get the information for the selected item.
    file_info = getFileInfo(selected_items.at(0), row);
    if (file_info == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to get the information for the selected item!";
        goto bail;
    }

    // Build up the path.
    selected_path = media_path + QString("/") + file_info->name;

bail:
    return selected_path;
}

/*!
 * @brief Reset the object and allow it to be reused.
 */
void SelectFolder::reset()
{
    // Clear out everything.
    cam_folders.clear();
    media_array = QJsonArray();
    media_info = QJsonObject();
    media_path.clear();

    ui->cam_tree->clear();
    ui->txt_meta->clear();
}

/*!
 * @brief Build up the path to the media base on the media type (SDCard, SSD, USB) and
 *        the image type (stills, stacks, videos or CAPs). The resulting string (media_path)
 *        is where, on the camera, we expect the information to reside.
 *
 * @param root_row      The root row index in our tree structure. This specifies the image type.
 * @return True if media_path was built successfully, false otherwise.
 */
bool SelectFolder::buildMediaPath()
{
    if (media_info.isEmpty())
    {
        // Build up the path on the camera for the media we're interested in.
        media_path = QStringLiteral("/media/");
        media_path.append(CamExplore::DEST_SSD_STR);
    }
    else
    {
        // Pull the media path from the JSON information.
        media_path = media_info["mountdir"].toString();
        if (media_path.isEmpty())
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "device:" << media_info["name"].toString() << "doesn't have a mount path!";
            return false;
        }
    }

    // Finish up the path construction based on which row we're looking at.
    media_path.append(QStringLiteral("/DCIM/100fastc"));

    // All good.
    return true;
}

/*!
 * @brief Format and display the metadata for the currently selected entry based on the
 *        file_info structure passed in.
 *
 * @param file_info     The file information we are going to display.
 */
void SelectFolder::displayMetadata(const CAM_EXPLORE::FILE_INFO *file_info)
{
    QString meta_str = CamExplore::generateMetadataString(file_info);

    // Set the information in our display.
    ui->txt_meta->setPlainText(meta_str);
}

/*!
 * @brief When the user expands the tree, make sure the columns are fully adjusted.
 * @param model_index   Which item was expanded.
 */
void SelectFolder::doTreeExpansion(QModelIndex model_index)
{
    Q_UNUSED(model_index)

    // Resize the column.
    ui->cam_tree->resizeColumnToContents(COLUMN_NAME);
}

/*!
 * @brief This slot is invoked with the QTreeWidget emits a currentItemChanged signal. We
 *        use this to update our metadata display.
 *
 * The signal is emitted when the "current item" changes. The "current item" is specified
 * by the "item_ptr" parameter, and it replaces the "prev_ptr" item (which we don't use).
 *
 * @param item_ptr  The new "current" item in the widget.
 */
void SelectFolder::doTreeItemChanged(QTreeWidgetItem *item_ptr, QTreeWidgetItem * /* prev_ptr */)
{
    int row;
    const CAM_EXPLORE::FILE_INFO *file_info;       // the meta data associated with the item.

    // Now we want to format and display the meta data.
    file_info = getFileInfo(item_ptr, row);
    displayMetadata(file_info);
}

/*!
 * @brief Handle double click on items in the tree.
 * @param item      Unused.
 * @param column    Unused.
 */
void SelectFolder::doTreeItemDoubleClicked(QTreeWidgetItem *item, int column)
{
    Q_UNUSED(item)
    Q_UNUSED(column)

    // Act like the user clicked the OK button.
    QTimer::singleShot(0, this, &SelectFolder::accept);
}

/*!
 * @brief Called when the user selects a new item in the tree widget. This works in concert with the
 *        doTreeItemChanged() method which handles when the user changes the current item.
 */
void SelectFolder::doTreeSelectionChanged()
{
}

/*!
 * @brief Start execution of the dialog.
 * @return A QDialog::DialogCode result.
 */
int SelectFolder::exec()
{
    // Sanity check.
    if (media_info.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "You must call setJsonInfo() prior to exec()!";
        return QDialog::Rejected;
    }

    // Do we need to reload?
    if (cam_folders.count() == 0)
    {
        // Fire a timer to start retrieving the listing.
        QTimer::singleShot(0, this, &SelectFolder::listingStart);
    }

    // Hand off to the base class.
    return QDialog::exec();
}


/*!
 * @brief Passed a tree widget item. This routine will find the file info info for that tree item.
 *
 * @param item      The item to be checked.
 * @param row_out   [OUTPUT] The row the item was found in.
 *
 * @return The file_info structure or NULL if not found.
 */
const CAM_EXPLORE::FILE_INFO *SelectFolder::getFileInfo(const QTreeWidgetItem *item, int& row_out)
{
    int     row;

    // Loop through each row (aka image type) of our listing database.
    for (row = 0; row < cam_folders.count(); ++row)
    {
        const CAM_EXPLORE::FILE_INFO  *list_info = &cam_folders.at(row);

        // Make sure the list_information is valid.
        if (list_info == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Empty list_info returned for row:" << row;
        }
        else
        {
            // Do we have a match?
            if (list_info->tree_widget == item)
            {
                // Return the row (image type) associated with this entry as
                // well as the information associated with the entry.
                row_out = row;
                return list_info;
            }
        }
    }

    // Didn't find it.
    return nullptr;
}

/*!
 * @brief Utility function to process the write completion status of the FileAction register
 *        write performed by one of the operations.
 */
void SelectFolder::handleFAStatus(uint32_t fa_status)
{
    FI_comm *comm_ptr;
    CameraMDI *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Is the camera server still busy?
    if ((fa_status & GIGE_FA_STATUS_BUSY) == (uint32_t)GIGE_FA_STATUS_BUSY)
    {
        // Start our timer to poll for the completion of the file action.
        if (file_timer_id == 0)
            file_timer_id = startTimer(FA_POLLING_PERIOD);
    }
    else if ((fa_status & GIGE_FA_STATUS_ERROR) == (uint32_t)GIGE_FA_STATUS_ERROR)
    {
        QString err_info_str;

        // Turn off the timer. The timer SHOULD have already been killed.
        if (file_timer_id != 0)
        {
            killTimer(file_timer_id);
            file_timer_id = 0;
        }

        // Build up our string with more information.
        err_info_str = tr("The camera has reported an error:\n%1 (%2)").arg(fa_status & GIGE_FA_STATUS_MASK).arg(strerror(fa_status & GIGE_FA_STATUS_MASK));
        QLOG_WARN() << __PRETTY_FUNCTION__ << err_info_str;

        // Show the user the error information.
        fiMsgBox::critical(camera_ptr, tr("An error has occurred."), err_info_str);
    }
    else
    {
        uint32_t    read_len;

        // Turn off the timer.
        if (file_timer_id != 0)
        {
            killTimer(file_timer_id);
            file_timer_id = 0;
        }

        // Extract the number of bytes in the listing.
        listing_len = fa_status & GIGE_FA_STATUS_MASK;

        // How much can we grab?
        if (listing_len > GIGE_FILEPATH_LEN)
            read_len = GIGE_FILEPATH_LEN;
        else
            read_len = listing_len;

        // Start pulling down the listing.
        comm_ptr->start_read_mem(camera_ptr->get_ip_addr(), GIGE_REG_FILEPATH, read_len);
    }
}

/*!
 * @brief This method performs the actual ReadMem calls to download the directory listing produced
 *        on the camera by a previous FileAction register write.
 *
 * The assumption is that everything is setup, the destination file has been created and left in
 * an open, usable state so that this method simply has to pull down the data and write it out.
 */
void SelectFolder::handleListingDownload()
{
    uint32_t    buffer_len;
    FI_comm     *comm_ptr;
    QByteArray  *read_chunk;
    CameraMDI *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the chunk of data we just read from the camera.
    read_chunk = comm_ptr->get_reg_buffer(camera_ptr->get_ip_addr(), GIGE_REG_FILEPATH);
    if (read_chunk == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No buffer available!";
        return;
    }
    if (read_chunk->size() <= 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Bad buffer length: " << read_chunk->size();
        delete read_chunk;
        return;
    }

    // Get the number of bytes we've read...but we only care up to listing_len bytes.
    buffer_len = (uint32_t)read_chunk->size();
    if (buffer_len > listing_len)
        buffer_len = listing_len;

    // Write the chunk out to our file.
    listing_file.write(read_chunk->constData(), read_chunk->size());

    // Done with the chunk.
    delete read_chunk;

    // How much do we have left?
    listing_len -= buffer_len;
    if (listing_len != 0)
    {
        uint32_t    read_len;

        // How much can we grab?
        if (listing_len > GIGE_FILEPATH_LEN)
            read_len = GIGE_FILEPATH_LEN;
        else
            read_len = listing_len;

        // Request the next chunk.
        comm_ptr->start_read_mem(camera_ptr->get_ip_addr(), GIGE_REG_FILEPATH, read_len);
    }
    else
    {
        // We're done with this listing. Close the file.
        listing_file.close();

        // Parse the information we've downloaded.
        listingFinish();
    }
}

/*!
 * @brief We're all done with the downloading of the listing file. Parse
 *        through all the data to provide a useful display for the user.
 */
void SelectFolder::listingFinish()
{
    int i;

    // Parse the listing file.
    parseListing();

    // Now that we have the folder list, display them in the tree.
    for (i = 0; i < cam_folders.count(); ++i)
    {
        CAM_EXPLORE::FiTreeWidgetItem *p_child;

        // Get our file info structure.
        CAM_EXPLORE::FILE_INFO *cur_file_info = &cam_folders[i];

        // Make sure the file information pointer is valid.
        if (cur_file_info == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to retrieve the current file inforamtion for folder index:" << i;
        }
        else
        {
            // Build up our tree structure.
            p_child = new CAM_EXPLORE::FiTreeWidgetItem(ui->cam_tree->invisibleRootItem());
            p_child->setText(COLUMN_NAME, cur_file_info->name);
            p_child->setText(COLUMN_SIZE, cur_file_info->size);
            p_child->setTextAlignment(COLUMN_SIZE, Qt::AlignRight | Qt::AlignVCenter);

            // We are going to override the time stamp if the file is incomplete.
            // We want the user to note right away the file is not valid.
            if (cur_file_info->incomplete.compare(CamExplore::FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
            {
                p_child->setText(COLUMN_TIME, CamExplore::FILE_INFO_INCOMPLETE_STR);
            }
            else
            {
                p_child->setText(COLUMN_TIME, ClientMath::fi_time_to_local(cur_file_info->time_stamp));
            }

            // Hook the widget into our listing database for later.
            cur_file_info->tree_widget = p_child;
        }
    }

    // Sort the entries in this row by the name column.
    ui->cam_tree->invisibleRootItem()->sortChildren(COLUMN_NAME, Qt::AscendingOrder);

    // Resize the columns.
    ui->cam_tree->resizeColumnToContents(COLUMN_NAME);
    ui->cam_tree->resizeColumnToContents(COLUMN_SIZE);

    // Select the first entry.
    cam_folders.at(0).tree_widget->setSelected(true);
}

/*!
 * @brief Start the process of downloading the file listing.
 */
void SelectFolder::listingStart()
{
    QString label_str;
    QString xml_filename;
    FI_comm *comm_ptr;
    CameraMDI *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Build up the path on the camera for the media we're interested in.
    if (!buildMediaPath())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "build_media_path() failed!";
        return;
    }

    // Now build up our local file and our label based on that row.
    xml_filename = QDir::tempPath() + CamExplore::STACKS_DIR_LISTING;
    label_str = tr("Retrieving listing for the STACKs folder");

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Remove any old file that might still be laying around.
    QFile::remove(xml_filename);

    // Get ready for the new file.
    listing_file.setFileName(xml_filename);
    if (!listing_file.open(QIODevice::WriteOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open" << xml_filename << ". Err =" << listing_file.errorString();
        return;
    }
    listing_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // Request the directory listing from the camera. First thing to do is to
    // tell the camera what path we're interested in.
    comm_ptr->start_write_mem(camera_ptr->get_ip_addr(), GIGE_REG_FILEPATH, media_path.toUtf8());
}

/*!
 * @brief Parse the downloaded listing file (as specified by the "listing_file" member), and build
 *        up our listing database with all the associated metadata.
 */
void SelectFolder::parseListing()
{
    QDomDocument    doc(QStringLiteral("dir_list"));
    QDomElement     dom_root;           // xml token "root"
    QDomElement     dom_folder;         // xml token "folder"
    QString     msg_str;
    QString     target_name;
    QString     target_cap;
    QString     target_avail;
    QString     target_inodes;
    CAM_EXPLORE::FILE_INFO cur_file_info;

    if (!listing_file.open(QIODevice::ReadOnly))
    {
        msg_str = QString::fromUtf8("Error opening returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(listing_file.error()).arg(listing_file.errorString());
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        return;
    }

    if (!doc.setContent(&listing_file))
    {
        msg_str = QString::fromUtf8("Error parsing returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(errno).arg(QString::fromUtf8(strerror(errno)));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;

        listing_file.close();
        return;
    }

    // Don't need this anymore it's been read in.
    listing_file.close();

    // we need to get what target we are working on and how much the capacity is.
    dom_root = doc.firstChildElement(QStringLiteral("root"));
    target_name = dom_root.attribute(QStringLiteral("target"));
    target_cap = dom_root.attribute(QStringLiteral("capacity"));
    target_avail = dom_root.attribute(QStringLiteral("available"));
    target_inodes = dom_root.attribute(QStringLiteral("inodes"));

    // We are going to be navigating the document differently if we asked for stacks vs everything
    // else for the stack directory there are folder tags which go into the stack group, but
    // there are also file tags which go into the stills group, all other types there are
    // only files which go in the appropriate tag
    // first stacks
    dom_folder = dom_root.firstChildElement(QStringLiteral("folder"));
    if (dom_folder.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No folder element?";
    }
    while (!dom_folder.isNull())
    {
        cur_file_info.clear();
        cur_file_info.name = dom_folder.attribute(QStringLiteral("name"));
        if (!cur_file_info.name.isEmpty())
        {
            cur_file_info.size = dom_folder.firstChildElement(QStringLiteral("size")).text();
            cur_file_info.width = dom_folder.firstChildElement(QStringLiteral("width")).text();
            cur_file_info.height = dom_folder.firstChildElement(QStringLiteral("height")).text();
            cur_file_info.bit_mode = dom_folder.firstChildElement(QStringLiteral("bit_mode")).text();
            cur_file_info.frame_count = dom_folder.firstChildElement(QStringLiteral("frame_count")).text();
            cur_file_info.time_stamp = dom_folder.firstChildElement(QStringLiteral("time_stamp")).text();
            cur_file_info.fps = dom_folder.firstChildElement(QStringLiteral("fps")).text();
            cur_file_info.file_ext = dom_folder.firstChildElement(QStringLiteral("codec")).text();
            cur_file_info.meta_count = dom_folder.firstChildElement(QStringLiteral("metafile_count")).text();
            cur_file_info.raw = dom_folder.firstChildElement(QStringLiteral("raw")).text();
            cur_file_info.start_frame = dom_folder.firstChildElement(QStringLiteral("start_frame")).text();
            cur_file_info.end_frame = dom_folder.firstChildElement(QStringLiteral("end_frame")).text();
            cur_file_info.comment = dom_folder.firstChildElement(QStringLiteral("comment")).text();

            cam_folders.append(cur_file_info);
        }
        dom_folder = dom_folder.nextSiblingElement(QStringLiteral("folder"));
    }
}

/*!
 * @brief The communications object is indicating that a previously requested register read operation
 *        has completed.
 *
 * @param ip_addr   The IP address of the camera the read operation was for.
 * @param reg_addr  The register that was being read.
 */
void SelectFolder::readComplete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t reg_value;
    FI_comm *comm_ptr;
    CameraMDI *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Is this event for us?
    if (ip_addr != camera_ptr->get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Read the register value.
    reg_value = comm_ptr->get_reg(ip_addr, reg_addr);

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_FILE_ACTION_STATUS:
        handleFAStatus(reg_value);
        break;
    case GIGE_REG_FILEPATH:
        handleListingDownload();
        break;
    default:
        break;
    }
}

/*!
 * @brief Specify the destination media object with a JSON object.
 *
 * This is an optional function, used in post-series3 cameras.
 *
 * @param json_info     The JSON information we can use for the destination media.
 * @param json_array    The JSON information for all media on the camera.
 */
void SelectFolder::setJsonInfo(const QJsonObject& json_info, const QJsonArray& json_array)
{
    // Copy our new media information object.
    media_array = json_array;
    media_info = json_info;

    // Make sure we got something.
    if (media_info.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No media_info!?";
        return;
    }
}

/*!
 * @brief A timer has fired. Used for monitoring the progress of long operations.
 * @param event The actual timer event
 */
void SelectFolder::timerEvent(QTimerEvent *event)
{
    CameraMDI *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Is this our file action timer?
    if (event->timerId() == file_timer_id)
    {
        // Poll the file action status.
        camera_ptr->start_camera_read_reg(GIGE_REG_FILE_ACTION_STATUS);

        // Stop the timer until we get a response.
        killTimer(file_timer_id);
        file_timer_id = 0;
    }
}

/*!
 * @brief The communications object is indicating that a previously requested register write operation
 *        has completed.
 *
 * @param ip_addr   The IP address of the camera the write operation was for.
 * @param reg_addr  The register that was being written.
 */
void SelectFolder::writeComplete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t write_status;
    FI_comm *comm_ptr;
    CameraMDI *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Is this event for us?
    if (ip_addr != camera_ptr->get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_FILE_ACTION:
    case GIGE_REG_FA_JSON:
        // Read the status of the write operation.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Was this a success?
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Start our timer to poll for the completion of the file action.
            if (file_timer_id == 0)
                file_timer_id = startTimer(FA_POLLING_PERIOD);
        }
        break;
    case GIGE_REG_FILEPATH:
        // Read the status of the write operation.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Process the write completion.
        writeFilepathDone(write_status);
        break;
    default:
        break;
    }
}

/*!
 * @brief Handle the write completion processing of the FilePath register.
 * @param write_status  The write completion status.
 */
void SelectFolder::writeFilepathDone(uint32_t write_status)
{
    FI_comm *comm_ptr;
    CameraMDI *camera_ptr;

    // Get a pointer to our camera object.
    camera_ptr = qobject_cast<CameraMDI *>(parent());
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Check for an error.
    if (write_status != GEV_STATUS_SUCCESS)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << camera_ptr->get_ip_addr_pretty() << "Failed to set file path to:" << media_path;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // We've set the path on the camera we're interested in, now request a listing.
    comm_ptr->start_write_reg(camera_ptr->get_ip_addr(), GIGE_REG_FILE_ACTION, GIGE_FA_LISTING);
}
