/*!
 * @file select_item.cpp
 * @brief Implementation of the SelectItemDlg class.
 *
 * Copyright 2015 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "select_item.h"
#include "ui_select_item.h"

#include <QVariant>

#include "logger.h"

// Overlay item labels.
const QString SelectItemDlg::ABS_FRAME_LABEL = QApplication::translate("overlay", "Frame");
const QString SelectItemDlg::ABS_TIME_LABEL = QApplication::translate("overlay", "Time");
const QString SelectItemDlg::BIT_SEL_LABEL = QApplication::translate("overlay", "Bit Selection");
const QString SelectItemDlg::CUSTOM_TEXT_LABEL = QApplication::translate("overlay", "Custom Text");
const QString SelectItemDlg::DURATION_LABEL = QApplication::translate("overlay", "Duration");
const QString SelectItemDlg::EVENT_LABEL = QApplication::translate("overlay", "Events");
const QString SelectItemDlg::EXPOSURE_LABEL = QApplication::translate("overlay", "Exposure");
const QString SelectItemDlg::FRAME_COUNT_LABEL = QApplication::translate("overlay", "Frame Count");
const QString SelectItemDlg::FRAMERATE_LABEL = QApplication::translate("overlay", "Frame Rate");
const QString SelectItemDlg::GAIN_LABEL = QApplication::translate("overlay", "Analog Gain");
const QString SelectItemDlg::MAC_LABEL = QApplication::translate("overlay", "Camera MAC");
const QString SelectItemDlg::MANUFACTURER_LABEL = QApplication::translate("overlay", "Manufacturer");
const QString SelectItemDlg::MODEL_LABEL = QApplication::translate("overlay", "Model");
const QString SelectItemDlg::NAME_LABEL = QApplication::translate("overlay", "Name");
const QString SelectItemDlg::REL_FRAME_LABEL = QApplication::translate("overlay", "Frame (rel)");
const QString SelectItemDlg::REL_TIME_LABEL = QApplication::translate("overlay", "Time (Trel)");
const QString SelectItemDlg::REL_FTIME_LABEL = QApplication::translate("overlay", "Time (Frel)");
const QString SelectItemDlg::REL_FTIME_PROMPT_LABEL = QApplication::translate("overlay", "Enter Relative Frame for Time Zero");
const QString SelectItemDlg::ROI_LABEL = QApplication::translate("overlay", "ROI");
const QString SelectItemDlg::SENSOR_OPT_LABEL = QApplication::translate("overlay", "Bin/subsample");
const QString SelectItemDlg::SMPTE_TIME_LABEL = QApplication::translate("overlay", "Time (SMPTE)");
const QString SelectItemDlg::SYNC_SOURCE_LABEL = QApplication::translate("overlay", "Sync Source");
const QString SelectItemDlg::TIME_OFFSET_LABEL = QApplication::translate("overlay", "Time0 Offset");
const QString SelectItemDlg::TIME_SOURCE_LABEL = QApplication::translate("overlay", "Time Source");
const QString SelectItemDlg::TRIGGER_POS_LABEL = QApplication::translate("overlay", "Trigger");
const QString SelectItemDlg::VERSION_LABEL = QApplication::translate("overlay", "Version");
const QString SelectItemDlg::VT_FLAGS_LABEL = QApplication::translate("overlay", "VT Flags");
const QString SelectItemDlg::VT_CHANGE_LABEL = QApplication::translate("overlay", "VT Change");
/*!
 * @brief SelectItemDlg class constructor.
 */
SelectItemDlg::SelectItemDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::SelectItemDlg)
{
    int     i;

    ui->setupUi(this);

    // Start out with the move buttons disabled.
    ui->move_down_btn->setEnabled(false);
    ui->move_up_btn->setEnabled(false);

    // Initialize our available list to all entries.
    for (i = 0; i < OPTION_COUNT; ++i)
        avail_items.append(get_label(i));

    // Currently there are no "used" labels.
    used_items.clear();

    // Setup the custom frame relative min and max to be zero.  They will be set by the user
    // of this dialog by calling set_custom_ftime_frame().
    custom_ftime_min_rel_frame = 0;
    custom_ftime_max_rel_frame = 0;

    // Initialize our connections.
    connect(ui->add_btn, &QPushButton::clicked, this, &SelectItemDlg::do_add);
    connect(ui->remove_btn, &QPushButton::clicked, this, &SelectItemDlg::do_remove);
    connect(ui->reset_btn, &QPushButton::clicked, this, &SelectItemDlg::do_reset);
    connect(ui->move_up_btn, &QPushButton::clicked, this, &SelectItemDlg::do_move_up);
    connect(ui->move_down_btn, &QPushButton::clicked, this, &SelectItemDlg::do_move_down);
    connect(ui->visible_items_list, &QListWidget::itemSelectionChanged, this, &SelectItemDlg::do_visible_sel_changed);
    connect(ui->custom_ftime_edit, &QLineEdit::textChanged, this, &SelectItemDlg::do_ftime_range_check);
}

/*!
 * @brief SelectItemDlg class destructor.
 */
SelectItemDlg::~SelectItemDlg()
{
    delete ui;
}

/*!
 * @brief Enable/disable controls based on what's in the lists.
 */
void SelectItemDlg::adjust_ui()
{
    ui->add_btn->setEnabled(ui->avail_items_list->count() > 0);
    ui->remove_btn->setEnabled(ui->visible_items_list->count() > 0);

    // Do we have a custom frame relative time selected?
    bool ftime_selected = !ui->visible_items_list->findItems(REL_FTIME_LABEL, Qt::MatchExactly).isEmpty();
    ui->custom_ftime_edit->setEnabled(ftime_selected);
    ui->custom_ftime_edit->setVisible(ftime_selected);
    ui->custom_ftime_edit->setValidator(new QIntValidator());
    ui->custom_ftime_label->setVisible(ftime_selected);

    // Do we have a custom text selected?
    bool custom_text_selected = !ui->visible_items_list->findItems(CUSTOM_TEXT_LABEL, Qt::MatchExactly).isEmpty();
    ui->custom_edit->setEnabled(custom_text_selected);
    ui->custom_edit->setVisible(custom_text_selected);
    ui->custom_edit_label->setVisible(custom_text_selected);

    ui->buttonBox->button(QDialogButtonBox::Ok)->setEnabled(ui->visible_items_list->count() > 0);
}

const QString SelectItemDlg::custom_ftime_frame()
{
    if (used_items.contains(REL_FTIME_LABEL))
    {
        custom_ftime_data = ui->custom_ftime_edit->text();
        return custom_ftime_data;
    }

    // Not selected.
    return QString();
}

const QString SelectItemDlg::custom_text()
{
    if (used_items.contains(CUSTOM_TEXT_LABEL))
    {
        custom_text_data = ui->custom_edit->toPlainText();
        return custom_text_data;
    }

    // Not selected.
    return QString();
}

const QString SelectItemDlg::get_label(int item_id)
{
    switch (item_id)
    {
    case ABS_FRAME:
        return SelectItemDlg::ABS_FRAME_LABEL;
    case ABS_TIME:
        return SelectItemDlg::ABS_TIME_LABEL;
    case BIT_SEL:
        return SelectItemDlg::BIT_SEL_LABEL;
    case CUSTOM_TEXT:
        return SelectItemDlg::CUSTOM_TEXT_LABEL;
    case DURATION:
        return SelectItemDlg::DURATION_LABEL;
    case EVENT:
        return SelectItemDlg::EVENT_LABEL;
    case EXPOSURE:
        return SelectItemDlg::EXPOSURE_LABEL;
    case FRAME_COUNT:
        return SelectItemDlg::FRAME_COUNT_LABEL;
    case FRAMERATE:
        return SelectItemDlg::FRAMERATE_LABEL;
    case GAIN:
        return SelectItemDlg::GAIN_LABEL;
    case MAC:
        return SelectItemDlg::MAC_LABEL;
    case MANUFACTURER:
        return SelectItemDlg::MANUFACTURER_LABEL;
    case MODEL:
        return SelectItemDlg::MODEL_LABEL;
    case NAME:
        return SelectItemDlg::NAME_LABEL;
    case REL_FRAME:
        return SelectItemDlg::REL_FRAME_LABEL;
    case REL_TIME:
        return SelectItemDlg::REL_TIME_LABEL;
    case REL_FTIME:
        return SelectItemDlg::REL_FTIME_LABEL;
    case ROI:
        return SelectItemDlg::ROI_LABEL;
    case SENSOR_OPT:
        return SelectItemDlg::SENSOR_OPT_LABEL;
    case SMPTE_TIME:
        return SelectItemDlg::SMPTE_TIME_LABEL;
    case SYNC_SOURCE:
        return SelectItemDlg::SYNC_SOURCE_LABEL;
    case TIME_OFFSET:
        return SelectItemDlg::TIME_OFFSET_LABEL;
    case TIME_SOURCE:
        return SelectItemDlg::TIME_SOURCE_LABEL;
    case TRIGGER_POS:
        return SelectItemDlg::TRIGGER_POS_LABEL;
    case VERSION:
        return SelectItemDlg::VERSION_LABEL;
    case VT_FLAGS:
        return SelectItemDlg::VT_FLAGS_LABEL;
    case VT_CHANGE:
        return SelectItemDlg::VT_CHANGE_LABEL;
    default:
        break;
    }

    return QString();
}

QList<int> SelectItemDlg::get_visible_list()
{
    int         used_id;
    QList<int>  visible_list;

    // Build our list. Walk through each item in the "used" list.
    for (used_id = 0; used_id < used_items.count(); ++used_id)
    {
        int item_id;

        // Run through all the items and try to find a match.
        for (item_id = 0; item_id < OPTION_COUNT; ++item_id)
        {
            // Is this a matching item?
            if (used_items.at(used_id) == get_label(item_id))
            {
                // Save the ID.
                visible_list.append(item_id);
                break;
            }
        }
    }

    // Return our list.
    return visible_list;
}


bool SelectItemDlg::item_is_selected(int item_id)
{
    // Sanity check.
    if (get_label(item_id) == QString())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected item ID:" << item_id;
        return false;
    }

    // The item is selected if it's in our "used" list.
    return used_items.contains(get_label(item_id));
}

void SelectItemDlg::select_item(int item_id)
{
    int     i;

    // Where is this item in the available list?
    for (i = 0; i < avail_items.count(); ++i)
    {
        // Is this the one we're looking for?
        if (avail_items.at(i) == get_label(item_id))
        {
            // Remove this from the available list and append it to the visible list.
            used_items.append(avail_items.takeAt(i));
            break;
        }
    }

    // If this was the custom text item, we'll want to enable the edit box.
    if (item_id == CUSTOM_TEXT)
    {
        set_up_custom_text();
        ui->custom_edit->setEnabled(true);
    }
    else if (item_id == REL_FTIME)
    {
        set_up_custom_ftime();
        ui->custom_ftime_edit->setEnabled(true);
    }
}

void SelectItemDlg::set_custom_ftime_frame(const QString& user_text, int ftime_frame_min, int ftime_frame_max)
{
    custom_ftime_data = user_text;
    custom_ftime_min_rel_frame = ftime_frame_min;
    custom_ftime_max_rel_frame = ftime_frame_max;

    QString ftime_range_label = QString(" [Range %1 to %2]:").arg(custom_ftime_min_rel_frame).arg(custom_ftime_max_rel_frame);
    ui->custom_ftime_label->setText(REL_FTIME_PROMPT_LABEL+ftime_range_label);

    set_up_custom_ftime();
}

void SelectItemDlg::set_custom_text(const QString& user_text)
{
    custom_text_data = user_text;
    set_up_custom_text();
}

void SelectItemDlg::set_up_custom_text()
{
    ui->custom_edit->setPlainText(custom_text_data);
    ui->custom_edit_label->setText(CUSTOM_TEXT_LABEL);
}

void SelectItemDlg::set_up_custom_ftime()
{
    ui->custom_ftime_edit->setText(custom_ftime_data);    
}

//--------------------------------------------------------------------
// Slots
//--------------------------------------------------------------------

/*!
 * @brief User clicked the OK button. Build up our string lists so that the
 *        calling function can know which columns the user is interested in.
 */
void SelectItemDlg::accept()
{
    int     item_row;
    QListWidgetItem *item_ptr;

    // Reset our lists.
    avail_items.clear();
    used_items.clear();

    // Build up the available and used lists.
    for (item_row = 0; item_row < ui->avail_items_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        item_ptr = ui->avail_items_list->item(item_row);
        avail_items.append(item_ptr->text());
    }
    for (item_row = 0; item_row < ui->visible_items_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        item_ptr = ui->visible_items_list->item(item_row);
        used_items.append(item_ptr->text());
    }

    // Pass it on.
    QDialog::accept();
}

/*!
 * @brief Move the selected item(s) from the available
 *        list box to the visible list box.
 */
void SelectItemDlg::do_add()
{
    int i;
    QList<QListWidgetItem *>    sel_items;

    // Get all of the items selected from the available list.
    sel_items = ui->avail_items_list->selectedItems();

    // Turn off sorting for our available list while we're manipulating it.
    ui->avail_items_list->setSortingEnabled(false);

    // Move the items to the show list.
    for (i = 0; i < sel_items.count(); ++i)
    {
        int item_row;
        QListWidgetItem *item_ptr;

        // Get a pointer for ease of use.
        item_ptr = sel_items.at(i);

        // What row is this item on?
        item_row = ui->avail_items_list->row(item_ptr);

        // Remove the item from the show list and add it to the available list.
        ui->avail_items_list->takeItem(item_row);
        ui->visible_items_list->addItem(item_ptr);
    }

    // Re-enable sorting in the available list.
    ui->avail_items_list->setSortingEnabled(true);

    // Adjust our buttons
    adjust_ui();
}

/*!
 * @brief Is our ftime for relative frame within the range provided?
 */
void SelectItemDlg::do_ftime_range_check()
{
    bool in_range = true;

    // Is the current setting within our range?
    int ftime_frame = ui->custom_ftime_edit->text().toInt();
    if( ftime_frame < custom_ftime_min_rel_frame )
    {
        ftime_frame = custom_ftime_min_rel_frame;
        in_range = false;
    }
    else if( ftime_frame > custom_ftime_max_rel_frame )
    {
        ftime_frame = custom_ftime_max_rel_frame;
        in_range = false;
    }
    // Reset the user's entry.
    if( !in_range )
    {
        ui->custom_ftime_edit->setText(QString("%1").arg(ftime_frame));
    }
}

/*!
 * @brief Move the selected item(s) from the shown list
 *        box to the available list box.
 */
void SelectItemDlg::do_remove()
{
    int i;
    QList<QListWidgetItem *>    sel_items;

    // Get all of the items selected from the show list.
    sel_items = ui->visible_items_list->selectedItems();

    // Turn off sorting for our available list while we're manipulating it.
    ui->avail_items_list->setSortingEnabled(false);

    // Move the items to the available list.
    for (i = 0; i < sel_items.count(); ++i)
    {
        int item_row;
        QListWidgetItem *item_ptr;

        // Get a pointer for ease of use.
        item_ptr = sel_items.at(i);

        // What row is this item on?
        item_row = ui->visible_items_list->row(item_ptr);

        // Remove the item from the show list and add it to the available list.
        ui->visible_items_list->takeItem(item_row);
        ui->avail_items_list->addItem(item_ptr);
    }

    // Re-enable sorting in the available list.
    ui->avail_items_list->setSortingEnabled(true);
    ui->avail_items_list->sortItems();

    // Adjust our controls.
    adjust_ui();
}

/*!
 * @brief Reset the columns to "factory default".
 */
void SelectItemDlg::do_reset()
{
    // Turn off sorting for our available list while we're manipulating it.
    ui->avail_items_list->setSortingEnabled(false);

    // Clear out the lists.
    ui->avail_items_list->clear();
    ui->visible_items_list->clear();
    avail_items.clear();
    used_items.clear();

    // Build up the items, and repopulate the list.
    for (int i = 0; i < OPTION_COUNT; ++i)
        avail_items.append(get_label(i));

    // Re-enable sorting in the available list.
    ui->avail_items_list->setSortingEnabled(true);
    ui->avail_items_list->addItems(avail_items);

    // Adjust our controls.
    adjust_ui();
}

/*!
 * @brief Move the selected item(s) up in the shown listbox.
 */
void SelectItemDlg::do_move_up()
{
    int     item_row;

    // Walk through the selected items to determine our boolean values.
    for (item_row = 0; item_row < ui->visible_items_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        QListWidgetItem *item_ptr = ui->visible_items_list->item(item_row);

        // We only care about selected items.
        if (!item_ptr->isSelected())
            continue;

        // Move the item up one row.
        ui->visible_items_list->takeItem(item_row);
        ui->visible_items_list->insertItem(item_row - 1, item_ptr);
        item_ptr->setSelected(true);
    }
}

/*!
 * @brief Move the selected item(s) down in the shown listbox.
 */
void SelectItemDlg::do_move_down()
{
    int     item_row;

    // Walk through the selected items to determine our boolean values.
    for (item_row = ui->visible_items_list->count() - 1; item_row >= 0; --item_row)
    {
        // Get a pointer for ease of use.
        QListWidgetItem *item_ptr = ui->visible_items_list->item(item_row);

        // We only care about selected items.
        if (!item_ptr->isSelected())
            continue;

        // Move the item down one row.
        ui->visible_items_list->takeItem(item_row);
        ui->visible_items_list->insertItem(item_row + 1, item_ptr);
        item_ptr->setSelected(true);
    }
}

/*!
 * @brief The item(s) selected by the user in the shown listbox has
 *        changed. We need to decide, based on which item(s) is/are
 *        selected, which controls to enable/disable.
 */
void SelectItemDlg::do_visible_sel_changed()
{
    int     item_row;
    int     last_sel_row;
    int     at_least_one_selected;
    bool    do_enable;
    bool    non_contiguous;
    bool    top_item_selected;
    bool    bottom_item_selected;

    // Initialize our flags.
    at_least_one_selected = false;
    non_contiguous = false;
    top_item_selected = false;
    bottom_item_selected = false;

    // Walk through the selected items to determine our boolean values.
    for (item_row = 0, last_sel_row = -1; item_row < ui->visible_items_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        QListWidgetItem *item_ptr = ui->visible_items_list->item(item_row);

        // We only care about selected items.
        if (!item_ptr->isSelected())
            continue;

        // Ok - got a selected item.
        at_least_one_selected = true;

        // Is the top item selected?
        if (item_row == 0)
            top_item_selected = true;

        // Is the bottom item selected?
        if (item_row == (ui->visible_items_list->count() - 1))
            bottom_item_selected = true;

        // Check for contiguous state if not first time through.
        if (last_sel_row != -1)
        {
            // Are we non-contiguous?
            if (item_row != (last_sel_row + 1))
                non_contiguous = true;
        }

        // Is this a custom text item?
        QString item_text = item_ptr->text();
        if (item_text.compare(CUSTOM_TEXT_LABEL) == 0)
        {
            set_up_custom_text();
            ui->custom_edit->setEnabled(true);
        }
        else if (item_text.compare(REL_FTIME_LABEL) == 0)
        {
            set_up_custom_ftime();
            ui->custom_ftime_edit->setEnabled(true);
        }

        // Remember this row.
        last_sel_row = item_row;
    }

    // Enable/disable the move buttons based on what we've discovered.
    do_enable = at_least_one_selected && !bottom_item_selected && !non_contiguous;
    ui->move_down_btn->setEnabled(do_enable);
    do_enable = at_least_one_selected && !top_item_selected && !non_contiguous;
    ui->move_up_btn->setEnabled(do_enable);
}

/*!
 * @brief We override the exec() method so we can initialize the
 *        controls of the dialog.
 */
int SelectItemDlg::exec()
{
    // Setup our lists.
    if (avail_items.count() > 0)
        ui->avail_items_list->addItems(avail_items);
    else
        ui->avail_items_list->clear();
    if (used_items.count() > 0)
        ui->visible_items_list->addItems(used_items);
    else
        ui->visible_items_list->clear();

    // Adjust our controls.
    adjust_ui();

    // Call on up the chain.
    return QDialog::exec();
}



