/*!
 * @file show_columns.cpp
 * @brief Implementation of the ShowColumnsDialog class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "show_columns.h"
#include "ui_show_columns.h"

#include "logger.h"

/*!
 * @brief ShowColumnsDialog class constructor.
 */
ShowColumnsDialog::ShowColumnsDialog(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::ShowColumnsDialog)
{
    ui->setupUi(this);

    // Start out with the move buttons disabled.
    ui->move_down_btn->setEnabled(false);
    ui->move_up_btn->setEnabled(false);

    // Initialize our connections.
    connect(ui->add_btn, SIGNAL(clicked()), this, SLOT(do_add()));
    connect(ui->remove_btn, SIGNAL(clicked()), this, SLOT(do_remove()));
    connect(ui->reset_btn, SIGNAL(clicked()), this, SLOT(do_reset()));
    connect(ui->move_up_btn, SIGNAL(clicked()), this, SLOT(do_move_up()));
    connect(ui->move_down_btn, SIGNAL(clicked()), this, SLOT(do_move_down()));
    connect(ui->show_col_list, SIGNAL(itemSelectionChanged()), this, SLOT(do_show_sel_changed()));
}

/*!
 * @brief ShowColumnsDialog class destructor.
 */
ShowColumnsDialog::~ShowColumnsDialog()
{
    delete ui;
}

/*!
 * @brief Enable/disable controls based on what's in the lists.
 */
void ShowColumnsDialog::adjust_ui(void)
{
    ui->add_btn->setEnabled(ui->avail_col_list->count() > 0);
    ui->remove_btn->setEnabled(ui->show_col_list->count() > 0);
}

//-----------------------------------------------------------------------------
// Slots
//-----------------------------------------------------------------------------

/*!
 * @brief User clicked the OK button. Build up our string lists so that the
 *        calling function can know which columns the user is interested in.
 */
void ShowColumnsDialog::accept()
{
    int     item_row;
    QListWidgetItem *item_ptr;

    // Reset our lists.
    avail_cols.clear();
    used_cols.clear();

    // Build up the available and used lists.
    for (item_row = 0; item_row < ui->avail_col_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        item_ptr = ui->avail_col_list->item(item_row);
        avail_cols.append(item_ptr->text());
    }
    for (item_row = 0; item_row < ui->show_col_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        item_ptr = ui->show_col_list->item(item_row);
        used_cols.append(item_ptr->text());
    }

    // Pass it on.
    QDialog::accept();
}

/*!
 * @brief We override the exec() method so we can initialize the
 *        controls of the dialog.
 */
int ShowColumnsDialog::exec()
{
    // Setup our lists.
    if (avail_cols.count() > 0)
        ui->avail_col_list->addItems(avail_cols);
    else
        ui->avail_col_list->clear();
    if (used_cols.count() > 0)
        ui->show_col_list->addItems(used_cols);
    else
        ui->show_col_list->clear();

    // Adjust our controls.
    adjust_ui();

    // Call on up the chain.
    return QDialog::exec();
}

/*!
 * @brief Move the selected item(s) from the available
 *        list box to the shown list box.
 */
void ShowColumnsDialog::do_add(void)
{
    int i;
    QList<QListWidgetItem *>    sel_items;

    // Get all of the items selected from the available list.
    sel_items = ui->avail_col_list->selectedItems();

    // Turn off sorting for our available list while we're manipulating it.
    ui->avail_col_list->setSortingEnabled(false);

    // Move the items to the show list.
    for (i = 0; i < sel_items.count(); ++i)
    {
        int item_row;
        QListWidgetItem *item_ptr;

        // Get a pointer for ease of use.
        item_ptr = sel_items.at(i);

        // What row is this item on?
        item_row = ui->avail_col_list->row(item_ptr);

        // Remove the item from the show list and add it to the available list.
        ui->avail_col_list->takeItem(item_row);
        ui->show_col_list->addItem(item_ptr);
    }

    // Re-enable sorting in the available list.
    ui->avail_col_list->setSortingEnabled(true);

    // Adjust our buttons
    adjust_ui();
}

/*!
 * @brief Move the selected item(s) from the shown list
 *        box to the available list box.
 */
void ShowColumnsDialog::do_remove(void)
{
    int i;
    QList<QListWidgetItem *>    sel_items;

    // Get all of the items selected from the show list.
    sel_items = ui->show_col_list->selectedItems();

    // Turn off sorting for our available list while we're manipulating it.
    ui->avail_col_list->setSortingEnabled(false);

    // Move the items to the available list.
    for (i = 0; i < sel_items.count(); ++i)
    {
        int item_row;
        QListWidgetItem *item_ptr;

        // Get a pointer for ease of use.
        item_ptr = sel_items.at(i);

        // What row is this item on?
        item_row = ui->show_col_list->row(item_ptr);

        // Remove the item from the show list and add it to the available list.
        ui->show_col_list->takeItem(item_row);
        ui->avail_col_list->addItem(item_ptr);
    }

    // Re-enable sorting in the available list.
    ui->avail_col_list->setSortingEnabled(true);
    ui->avail_col_list->sortItems();

    // Adjust our controls.
    adjust_ui();
}

/*!
 * @brief Reset the columns to "factory default".
 */
void ShowColumnsDialog::do_reset(void)
{
    // Turn off sorting for our available list while we're manipulating it.
    ui->avail_col_list->setSortingEnabled(false);

    // Clear out the lists.
    ui->avail_col_list->clear();
    ui->show_col_list->clear();

    // Move the factory list to the show list.
    ui->show_col_list->addItems(factory_cols);

    // Re-enable sorting in the available list.
    ui->avail_col_list->setSortingEnabled(true);

    // Adjust our controls.
    adjust_ui();
}

/*!
 * @brief Move the selected item(s) up in the shown listbox.
 */
void ShowColumnsDialog::do_move_up(void)
{
    int     item_row;

    // Walk through the selected items to determine our boolean values.
    for (item_row = 0; item_row < ui->show_col_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        QListWidgetItem *item_ptr = ui->show_col_list->item(item_row);

        // We only care about selected items.
        if (!item_ptr->isSelected())
            continue;

        // Move the item up one row.
        ui->show_col_list->takeItem(item_row);
        ui->show_col_list->insertItem(item_row - 1, item_ptr);
        item_ptr->setSelected(true);
    }
}

/*!
 * @brief Move the selected item(s) down in the shown listbox.
 */
void ShowColumnsDialog::do_move_down(void)
{
    int     item_row;

    // Walk through the selected items to determine our boolean values.
    for (item_row = ui->show_col_list->count() - 1; item_row >= 0; --item_row)
    {
        // Get a pointer for ease of use.
        QListWidgetItem *item_ptr = ui->show_col_list->item(item_row);

        // We only care about selected items.
        if (!item_ptr->isSelected())
            continue;

        // Move the item down one row.
        ui->show_col_list->takeItem(item_row);
        ui->show_col_list->insertItem(item_row + 1, item_ptr);
        item_ptr->setSelected(true);
    }
}

/*!
 * @brief The item(s) selected by the user in the shown listbox has
 *        changed. We need to decide, based on which item(s) is/are
 *        selected, which controls to enable/disable.
 */
void ShowColumnsDialog::do_show_sel_changed(void)
{
    int     item_row;
    int     last_sel_row;
    int     at_least_one_selected;
    bool    do_enable;
    bool    non_contiguous;
    bool    top_item_selected;
    bool    bottom_item_selected;

    // Initialize our flags.
    at_least_one_selected = false;
    non_contiguous = false;
    top_item_selected = false;
    bottom_item_selected = false;

    // Walk through the selected items to determine our boolean values.
    for (item_row = 0, last_sel_row = -1; item_row < ui->show_col_list->count(); ++item_row)
    {
        // Get a pointer for ease of use.
        QListWidgetItem *item_ptr = ui->show_col_list->item(item_row);

        // We only care about selected items.
        if (!item_ptr->isSelected())
            continue;

        // Ok - got a selected item.
        at_least_one_selected = true;

        // Is the top item selected?
        if (item_row == 0)
            top_item_selected = true;

        // Is the bottom item selected?
        if (item_row == (ui->show_col_list->count() - 1))
            bottom_item_selected = true;

        // If not the first time through set up the contiguous state.
        if (last_sel_row != -1)
        {
            // Are we non-contiguous?
            if (item_row != (last_sel_row + 1))
                non_contiguous = true;
        }

        // Remember this row.
        last_sel_row = item_row;
    }

    // Enable/disable the move buttons based on what we've discovered.
    do_enable = at_least_one_selected && !bottom_item_selected && !non_contiguous;
    ui->move_down_btn->setEnabled(do_enable);
    do_enable = at_least_one_selected && !top_item_selected && !non_contiguous;
    ui->move_up_btn->setEnabled(do_enable);
}
