/*!
 * @file time_date.cpp
 * @brief This is the implementation of the TimeDateDLG dialog.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "time_date.h"
#include "ui_time_date.h"

#include <QByteArray>
#include <QDateTime>
#include <QList>
#include <QSettings>
#include <QTime>
#include <QTimeZone>

#include "logger.h"

/*!
 * @brief The TimeDateDLG dialog constructor.
 */
TimeDateDLG::TimeDateDLG(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::TimeDateDLG)
{
    QSettings settings;
    QStringList server_list;

    // Initialize the UI objects.
    ui->setupUi(this);

    // Fill in the NTP server combo box with our saved values, if any.
    settings.beginGroup("NTPServers");
    server_list = settings.value("serverList").toStringList();
    settings.endGroup();

    // Did we get anything?
    if (server_list.count() > 0)
    {
        // Reset the combo box and add in the new list.
        ui->ntpServerComboBox->clear();
        ui->ntpServerComboBox->addItems(server_list);
    }

    // Fill in the timezone combo box.
    ui->timeZoneComboBox->clear();
    QList<QByteArray> ids =  QTimeZone::availableTimeZoneIds();
    foreach (QByteArray id, ids)
    {
        // Do not add any "UTC" timezones.
        if (!id.startsWith("UTC"))
            ui->timeZoneComboBox->addItem(id);
    }

    // Get the current timezone for the PC.
    QByteArray myZone = QTimeZone::systemTimeZoneId();
    ui->timeZoneComboBox->setCurrentIndex(ui->timeZoneComboBox->findText(myZone));

    // Connect combo box to slot to update fields.
    connect(ui->timeZoneComboBox, SIGNAL(currentIndexChanged(int)), this, SLOT(update_fields()));
    connect(ui->ntpServerEnable, SIGNAL(clicked()), this, SLOT(ntp_enable()));

    // Update fields for initial value.
    update_fields();
}

/*!
 * @brief The TimeDateDLG dialog destructor.
 */
TimeDateDLG::~TimeDateDLG()
{
    delete ui;
}

/*!
 * @brief Retrieve the NTP server URL. The return value will be an
 *        empty string if NTP is not enabled.
 */
const QString TimeDateDLG::get_ntp_server(void)
{
    QString ret_str;

    // Only return a string if the NTP enable is checked.
    if (ui->ntpServerEnable->isChecked())
        ret_str = ui->ntpServerComboBox->currentText();

    // Return our value.
    return ret_str;
}

/*!
 * @brief Set the NTP server URL. If the parameter is an empty
 *        string, then disable the controls.
 * @param server_url    URL string for the NTP server. May be empty.
 */
void TimeDateDLG::set_ntp_server(const QString& server_url)
{
    // Did we get a URL?
    if (server_url.isEmpty())
    {
        // Make sure the NTP interface is disabled.
        ui->ntpServerGroup->setEnabled(false);
        ui->ntpServerEnable->setChecked(false);

        // Make sure the user can set the time/date manually.
        ui->time_and_date->setEnabled(true);
    }
    else
    {
        int foundIdx;

        // Enable the NTP interface.
        ui->ntpServerGroup->setEnabled(true);
        ui->ntpServerEnable->setChecked(true);

        // Don't let the user set the time/date manually.
        ui->time_and_date->setEnabled(false);

        // Is the URL already in the list?
        foundIdx = ui->ntpServerComboBox->findText(server_url, Qt::MatchExactly);
        if (foundIdx < 0)
        {
            // Add the new URL.
            ui->ntpServerComboBox->addItem(server_url);
            foundIdx = ui->ntpServerComboBox->findText(server_url, Qt::MatchExactly);
        }

        // Make the URL the currently selected one.
        ui->ntpServerComboBox->setCurrentIndex(foundIdx);
    }
}

/*!
 * @brief Retrieve the time/date setting.
 */
uint32_t TimeDateDLG::get_time(void)
{
    uint32_t nSecsSinceJan1st1970 = 0;
    QDateTime   cam_time;
    QByteArray id = ui->timeZoneComboBox->currentText().toLatin1();
    QTimeZone zone = QTimeZone(id);

    // Read the new settings.
    cam_time = QDateTime(ui->dateEdit->date(), ui->timeEdit->time(), zone).toLocalTime();

    // Make sure the date is valid.
    if (cam_time.isValid())
    {
        uint64_t nSecs64bit = cam_time.toSecsSinceEpoch();
        nSecsSinceJan1st1970 = (uint32_t) nSecs64bit;
    }
    // Return the value in seconds.
    return nSecsSinceJan1st1970;
}

/*!
 * @brief Set the current time/date.
 * @param seconds   Time from the camera.
 */
void TimeDateDLG::set_time(uint32_t seconds)
{
    uint64_t secsSinceEpoch = seconds;
    QDateTime   cam_time;
    QByteArray id = ui->timeZoneComboBox->currentText().toLatin1();
    QTimeZone zone = QTimeZone(id);

    // Initialize the camera time.
    cam_time.fromSecsSinceEpoch(secsSinceEpoch);

    // Do we have a valid timezone?
    if (zone.isValid())
    {
        // Use the timezone to display the correct time.
        QDateTime zoneTime = QDateTime(cam_time.date(), cam_time.time(), zone).toLocalTime();
        ui->dateEdit->setDate(zoneTime.date());
        ui->timeEdit->setTime(zoneTime.time());
    }
    else
    {
        // Just display the information.
        ui->dateEdit->setDate(cam_time.date());
        ui->timeEdit->setTime(cam_time.time());
    }
}

/*!
 * @brief Retrieve the timezone identifier.
 */
const QString TimeDateDLG::get_timezone(void)
{
    return ui->timeZoneComboBox->currentText();
}

/*!
 * @brief Return the timezone as a UTC offset string.
 */
const QString TimeDateDLG::get_utc_offset()
{
    QByteArray id = ui->timeZoneComboBox->currentText().toLatin1();
    QTimeZone zone = QTimeZone(id);

    return zone.displayName(QDateTime(QDate::currentDate(), QTime::currentTime(), zone), QTimeZone::OffsetName);
}

/*!
 * @brief Set the current timezone identifier.
 * @param tz_name   The timezone identifier.
 */
void TimeDateDLG::set_timezone(const QString& tz_name)
{
    int foundIdx;

    // Did we get a timezone?
    if (tz_name.isEmpty())
        return;

    // Make sure we have the timezone in our list.
    foundIdx = ui->timeZoneComboBox->findText(tz_name);
    if (foundIdx < 0)
        return;

    // Set the specified timezone.
    ui->timeZoneComboBox->setCurrentIndex(foundIdx);

    // Update the rest of the fields.
    update_fields();
}

/*!
 * @brief Respond to the user pressing the OK buttons. Make sure we save
 *        the NTP server list, in case the user added a new server URL.
 */
void TimeDateDLG::accept()
{
    int     i;
    QSettings settings;
    QStringList server_list;

    // Pull all the strings from the combobox.
    for (i = 0; i < ui->ntpServerComboBox->count(); ++i)
        server_list.append(ui->ntpServerComboBox->itemText(i));

    // Save all the NTP sites away for future use.
    settings.beginGroup("NTPServers");
    settings.setValue("serverList", server_list);
    settings.endGroup();

    // Pass it on.
    QDialog::accept();
}

/*!
 * @brief Respond to the user clicking on the NTP Enable checkbox.
 */
void TimeDateDLG::ntp_enable()
{
    // Is NTP enabled?
    if (ui->ntpServerEnable->isChecked())
    {
        // Enable the NTP interface.
        ui->ntpServerGroup->setEnabled(true);
        ui->ntpServerEnable->setChecked(true);

        // Don't let the user set the time/date manually.
        ui->time_and_date->setEnabled(false);
    }
    else
    {
        // Make sure the NTP interface is disabled.
        ui->ntpServerGroup->setEnabled(false);
        ui->ntpServerEnable->setChecked(false);

        // Make sure the user can set the time/date manually.
        ui->time_and_date->setEnabled(true);
    }
}

/*!
 * @brief Update the timezone related fields.
 */
void TimeDateDLG::update_fields(void)
{
    QByteArray id = ui->timeZoneComboBox->currentText().toLatin1();
    QTimeZone zone = QTimeZone(id);

    // Fill in fields for current time zone.
    if (zone.isValid())
    {
        ui->descriptionLabel->setText(tr("<b>Description:</b> ") + id);
        ui->countryLabel->setText(tr("<b>Country:</b> ") + QLocale::territoryToString(zone.territory()));
        ui->hasDaylightTimeCheckBox->setChecked(zone.hasDaylightTime());
        ui->isDaylightTimeCheckBox->setChecked(zone.isDaylightTime(QDateTime::currentDateTime()));
        ui->hasTransitionsCheckBox->setChecked(zone.hasTransitions());
        QDateTime zoneTime = QDateTime::currentDateTimeUtc().toTimeZone(zone);
        ui->dateEdit->setDate(zoneTime.date());
        ui->timeEdit->setTime(zoneTime.time());
        ui->utcOffsetLabel->setText(tr("<b>UTC Offset:</b> ") + zone.displayName(QDateTime(QDate::currentDate(), QTime::currentTime(), zone), QTimeZone::OffsetName));

        QTimeZone::OffsetData offset = zone.nextTransition(QDateTime::currentDateTime());
        if (offset.atUtc != QDateTime())
        {
            ui->nextTransitionLabel->setEnabled(true);
            ui->nextTransitionLabel->setText(tr("<b>Next transition:</b> %1").arg(offset.atUtc.toString()));
        }
        else
        {
            ui->nextTransitionLabel->setEnabled(false);
            ui->nextTransitionLabel->setText(tr("<b>Next transition:</b> none"));
        }
    }
}
