/*!
 * @file update_status.cpp
 * @brief This file contains the implementation of the UpdateStatus class.
 *
 * Copyright 2020 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "update_status.h"
#include "ui_update_status.h"

#include "camera_mdi.h"
#include "fi_journalctl.h"
#include "logger.h"
#include "user_preferences.h"

namespace UPDATE_STATUS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! The identifier we want to filter on in the journal.
const QString   UPDATE_ID = QStringLiteral("update.sh");

}   // End of namespace UPDATE_STATUS.

using namespace UPDATE_STATUS;

/*!
 * @brief Standard constructor for the UpdateStatus object.
 * @param parent    Used only for Qt stuff. Most likely should be the MainWindow.
 */
UpdateStatus::UpdateStatus(QWidget *parent) :
    QDockWidget(parent),
    ui(new Ui::UpdateStatus)
{
    // Initialize the UI.
    ui->setupUi(this);
}

/*!
 * @brief Destructor for the UpdateStatus object.
 *
 * We need to make sure we save the geometry of the window so we can restore it later.
 *
 * Also, clean up the journal object. Now the std::unique_ptr will probably take care
 * of this for us, but let's do it explicitly here.
 *
 * Finally, clean up the UI object.
 */
UpdateStatus::~UpdateStatus()
{
    QSettings settings;

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Save our last settings.
    settings.setValue("MainWindow/statusGeometry", saveGeometry());

    // Shutdown our journal control object.
    if (journal.get() != nullptr)
    {
        journal->close();
        journal.reset();
    }

    // Clean up the UI object.
    delete ui;
}

/*!
 * @brief Associate this object with a camera.
 *
 * With the camera_ptr, we can determine if this is a local camera or something
 * we need to SSH into. We use the FiJournalCtl object to manage that connection.
 *
 * @param camera_ptr    Pointer to the CameraMDI object we're interested in.
 * @return True if everything goes well, false if there was a problem.
 */
bool UpdateStatus::setCamera(CameraMDI *camera_ptr)
{
    QString title_str;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << (camera_ptr ? camera_ptr->get_ip_addr_pretty() : "Local Camera");

    // Do we have a journal control object?
    if (journal.get() != nullptr)
    {
        // Close and reset the control object.
        journal->close();
        journal.reset();
    }

    // Is the camera remote?
    if (camera_ptr && !camera_ptr->is_local())
    {
        QSettings settings;
        FI_JOURNALCTL::SSH_SETTINGS ssh_info;

        // Build up a title for our window.
        title_str = QStringLiteral("Update Status [%1]").arg(camera_ptr->get_ip_addr_pretty());

        // Initialize our information.
        ssh_info.hostname = camera_ptr->get_ip_addr_pretty();
        settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
        ssh_info.username = settings.value(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
        ssh_info.crypt_info = settings.value(UserPreferences::SERIESNC_CAM_PASS_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
        settings.endGroup();

        // Create our remote journal control object;
        journal = std::unique_ptr<FiJournalCtl>(new FiJournalCtl(this, &ssh_info));
        if (journal.get() == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate an FiJournalCtl object!";
            return false;
        }
    }
    else
    {
        // Build up a title for our window.
        title_str = QStringLiteral("Update Status [local]");

        // Create our journal control object;
        journal = std::unique_ptr<FiJournalCtl>(new FiJournalCtl(this));
        if (journal.get() == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to allocate an FiJournalCtl object!";
            return false;
        }
    }

    // Set our new title.
    setWindowTitle(title_str);

    // Is the journal control object OK?
    if (!journal->isCreated())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "JournalCtl object was not created successfully.";

        // Show an error message.
        ui->statusLog->appendPlainText(journal->errorStr());

        // Close the journal control object.
        journal->close();
        journal.reset();
        return false;
    }

    // Connect the signals up.
    connect(journal.get(), &FiJournalCtl::dataAvailable, this, &UpdateStatus::dataAvailable, Qt::QueuedConnection);

    // Good to go.
    return true;
}

/*!
 * @brief Begin the collection of journalctl events from the camera.
 *
 * We need to make sure to filter on the update process (only want those
 * entries), and follow the journal in realtime to get the update progress.
 */
void UpdateStatus::start()
{
    QStringList my_id;
    QStringList my_args;

    // Sanity check.
    if (journal.get() == nullptr)
        return;

    // For the update status we want to follow the journal.
    journal->setFollow(true);

    // We also need to run as root.
    journal->setUseSudo(true);

    // We're only interested in the update status, so filter on that.
    my_id.append(UPDATE_ID);
    journal->setIdentifiers(my_id);

    // We're only interested in logs from this latest boot period.
    my_args.append("-b");
    journal->setAdditionalArgs(my_args);

    // Tell the journal control object to go!
    journal->run();
}

/*!
 * @brief Our window has been closed.
 *
 * We want to make sure we delete ourselves later on.
 *
 * @param event [UNUSED] Pass this on to the parent object.
 */
void UpdateStatus::closeEvent(QCloseEvent *event)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Delete ourselves later.
    deleteLater();

    // Pass it on.
    QDockWidget::closeEvent(event);
}

/*!
 * @brief The FiJournalCtl object is notifying us that it received data.
 * @param log_data  The received log data.
 */
void UpdateStatus::dataAvailable(QString log_data)
{
    // Add the new entries to our log display.
    ui->statusLog->appendPlainText(log_data);
}
