/*!
 * @file user_preferences.cpp
 * @brief This is the implementation of the UserPreferences dialog.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "user_preferences.h"
#include "ui_user_preferences.h"

#include <QComboBox>
#include <QDateTime>
#include <QFile>
#include <QFileDialog>
#include <QDir>
#include <QMessageBox>
#include <QVersionNumber>

#include "cam_update_dlg.h"
#include "fi_comm.h"
#include "fi_updater.h"
#include "mainwindow.h"

using namespace USER_PREFERENCES;

//! Keep a constant of our default camera user name.
const QString UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME = QStringLiteral("fascam");

//! Keep a constant of our default camera workgroup.
const QString UserPreferences::SERIESNC_CAM_USER_DEFAULT_WORKGROUP = QStringLiteral("FASCAM");

//! Keep a constant of our camera preferences key for the camera user name for the system.
const QString UserPreferences::SERIESNC_CAM_USER_PREF_KEY = QStringLiteral("cam_user");

//! Keep a constant of our camera preferences key for the camera password for the system.
const QString UserPreferences::SERIESNC_CAM_PASS_PREF_KEY = QStringLiteral("cam_pass");

//! Keep a constant of our camera preferences key for the camera workgroup for the system.
const QString UserPreferences::SERIESNC_CAM_WORKGROUP_KEY = QStringLiteral("cam_workgroup");

//! Keep a constant of our camera preferences section.
const QString UserPreferences::SERIESNC_PREFERENCES_SECTION = QStringLiteral("Preferences");

/*!
 * @brief The UserPreferences dialog constructor.
 */
UserPreferences::UserPreferences(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::UserPreferences)
{
    ui->setupUi(this);
    factoryMenuEnable = nullptr;

    // Don't need putty on Linux.
#if defined( Q_OS_LINUX ) || defined( Q_OS_MAC )
    ui->frm_putty->setVisible(false);
#else
    ui->frm_putty->setVisible(true);
#endif

    // Initialize our controls.
    ui->packet_delay_spin->setMinimum(FI_COMM::PKT_DELAY_MIN);
    ui->packet_delay_spin->setMaximum(FI_COMM::PKT_DELAY_MAX);

    ui->setting_list->setCurrentRow(0);
    ui->optionStack->setCurrentIndex(0);

    // Make the knob standout visually so that it's easier to see on some monitors.
    ui->log_level_knob->setKnobStyle(QwtKnob::Styled);
    ui->log_level_knob->setMarkerStyle(QwtKnob::Triangle);

    // Setup our signals/slots.
    connect(ui->putty_browse, &QPushButton::clicked, this, &UserPreferences::browsePutty);
    connect(ui->net_buffers_spin, QOverload<int>::of(&QSpinBox::valueChanged), [ = ](int num_buffers) { updateBackoff(num_buffers); });
    connect(ui->setting_list, &QListWidget::currentItemChanged, this, &UserPreferences::changePage);
    connect(ui->show_checkbox, &QPushButton::clicked, this, &UserPreferences::showPassword);
    connect(MainWindow::GetInstance().get(), &MainWindow::closing, this, &UserPreferences::reject);

    // Sync up the restrictions for the backoff threshold and network buffers.
    updateBackoff(ui->net_buffers_spin->value());

    // Initialize our pages.
    initUpdatePage();

    // Initialize our network interfaces.
    initNetworkInterface();

    // Make sure we are sized correctly.
    adjustSize();
    update();
}

/*!
 * @brief The UserPreferences dialog destructor.
 */
UserPreferences::~UserPreferences()
{
    delete ui;

    if (factoryMenuEnable != nullptr)
    {
        delete factoryMenuEnable;
        factoryMenuEnable = nullptr;
    }
}

/*!
 * @brief Return the number of backoff buffers the user has selected.
 */
int UserPreferences::GetBackoffBuffers()
{
    return ui->backoff_buffers_spin->value();
}

void UserPreferences::GetCredentials(QString& username, QString& password, QString& workgroup)
{
    username = ui->username_edit->text();
    password = ui->pass_edit->text();
    workgroup = ui->workgroup_edit->text();
}

/*!
 * @brief Return true if the factory menu is to be enabled, false otherwise.
 */
bool UserPreferences::FactoryMenuEnabled()
{
    if (factoryMenuEnable != nullptr)
        return factoryMenuEnable->isChecked();
    return false;
}

QString UserPreferences::GetPuttyPath()
{
    return ui->putty_path->text();
}

void UserPreferences::initUpdatePage()
{
    int check_freq;
    QDateTime last_check;
    QSettings settings;
    QVersionNumber fm_latest;
    FiUpdater *updater;

    // Get the updater object.
    updater = MainWindow::GetInstance()->get_updater();

    // Load information from the settings database.
    settings.beginGroup(FiUpdater::UPDATE_GROUP_TAG);

    // Initialize the auto update group.
    ui->automatic_group->setChecked(settings.value(FiUpdater::UPDATE_AUTO_CHECK_KEY, true).toBool());

    // How frequently are we supposed to do updates?
    check_freq = settings.value(FiUpdater::UPDATE_FREQUENCY_KEY).toInt();

    // When did we last check for an update?
    last_check = settings.value(FiUpdater::LAST_CHECK_TIME_KEY).toDateTime();

    // Are we going to an alternate server? This is really for testing purposes, but
    // it helps to show where we are pulling from.
    serverUrl = settings.value(FiUpdater::UPDATE_SERVER_KEY).toString();
    settings.endGroup();

    // Display which updates are available.
    displayUpdates();

    // Update our times.
    ui->cb_check_interval->setCurrentIndex(check_freq);
    ui->last_check->setText(last_check.toString());
    ui->check_date->setText(FiUpdater::GetNextCheck(check_freq, last_check).toString());
    if (serverUrl.isEmpty())
        ui->lbl_checking->setText(tr("Waiting..."));
    else
        ui->lbl_checking->setText(tr("Waiting...[%1]").arg(serverUrl));

    // Enable the check button.
    ui->pb_check->setEnabled(true);

    // Can't create if we don't have a package.
    ui->pb_download->setEnabled(false);

    // Set the default path.
    ui->download_path->setText(QDir::toNativeSeparators(updater->GetUpdatePath()));

    // Connect the update signals.
    connect(ui->cb_check_interval, QOverload<int>::of(&QComboBox::currentIndexChanged), [ = ](int index) { checkTimeChanged(index); });
    connect(ui->pb_check, &QPushButton::clicked, this, &UserPreferences::doCheck);
    connect(ui->pb_download, &QPushButton::clicked, this, &UserPreferences::doDownloadPackage);

    connect(ui->chk_series3, &QCheckBox::clicked, this, &UserPreferences::doSelectPackage);
    connect(ui->chk_seriesNC, &QCheckBox::clicked, this, &UserPreferences::doSelectPackage);
    connect(ui->chk_fasmotion, &QCheckBox::clicked, this, &UserPreferences::doSelectPackage);

    connect(ui->download_browse, &QPushButton::clicked, this, &UserPreferences::doBrowseDownloadPath);
}


/*!
 * @brief Initialize the network selection combo box.
 */
void UserPreferences::initNetworkInterface()
{
    // Not implemented right now.  On hold.
    ui->network_interface_group->setVisible(false);
    return;
#if 0
    // NOTE:  This is code used in the Camera Dock area to fill up its cb_ni with the
    // available network interfaces in the system.  This should really be incorporated
    // into a network interface class and if we are going to make an attempt to improve
    // the network interface UI to be more wireshark-like we would use our new class.  I
    // am going to go ahead and check this into svn this way to leave a bread-crumb as
    // to my initial experimentation with adding network interface selection to the
    // FasMotion User Preferences.
    QList<QNetworkAddressEntry> all_address;

    // Do not propogate any signals from the combobox while we're doing this!
    ui->cb_ni->blockSignals(true);

    // Find out all the interfaces which are on the system.
    QList<QNetworkInterface> net_interfaces = QNetworkInterface::allInterfaces();

    // Since there may be multiple network interfaces on the client machine we need
    // to loop through all of them and let the user pick which one they want to
    // send the discovery message on.
    for (int i = 0; i < net_interfaces.size(); ++i)
    {
        const QNetworkInterface             net_interface = net_interfaces.at(i);
        QNetworkInterface::InterfaceFlags   flags;

        // Get the network interface's flags.
        flags = net_interface.flags();

        // Make sure the interface is valid.
        if (((flags & QNetworkInterface::IsUp) == 0) ||                     // Network isn't up.
            ((flags & QNetworkInterface::IsRunning) == 0))                  // Network isn't running.
        {
            // Not interested.
            continue;
        }

        // The network is a loopback interface and we're not interested in local only operation.
        if (((flags & QNetworkInterface::IsLoopBack) == QNetworkInterface::IsLoopBack) && !MainWindow::local_control())
            continue;

        // Get the address information for this interface.
        all_address = net_interface.addressEntries();

        // Get the relevant IP info. We will need this to setup which interface to use.
        // Note that we are interested in the IPv4 address and we want to skip the IPv6
        // addresses.
        for (int j = 0; j < all_address.size(); ++j)
        {
            QString     cb_str;                 // The string we show in the combo box.
            QString     inter_name_str;         // The name of the interface.
            QStringList interface_info;         // A string list containing all the information we'll need on this interface.
            QHostAddress    ip, netmask;        // IP address and netmask information.

            // Not interested if it doesn't have a broadcast address.
            if (all_address.at(j).broadcast().isNull())
                continue;

            // We found a good network. Get the name of the interface.
            inter_name_str = net_interface.humanReadableName();

            // Get the IP address information, as well as the netmask and broadcast address.
            ip = all_address.at(j).ip();
            netmask = all_address.at(j).netmask();

            // Build up our string list for the combo box.
            interface_info << inter_name_str << QString::fromUtf8("%1").arg(netmask.toIPv4Address(), 0, 16) <<
                           QString::fromUtf8("%1").arg(ip.toIPv4Address(), 0, 16);

            // Build up the combo box string.
            cb_str = inter_name_str + QString::fromUtf8(" (") + ip.toString() + QString::fromUtf8(")");
            ui->cb_ni->addItem(cb_str, interface_info);
            break;
        }
    }

    // It's ok to turn signals back on now.
    ui->cb_ni->blockSignals(false);
#endif

}

/*!
 * @brief Return the logging level the user would like to use.
 */
QsLogging::Level UserPreferences::GetLoggingLevel()
{
    QsLogging::Level actual_level;

    // Convert from the user value to the actual value.
    switch ((int)ui->log_level_knob->value())
    {
    case 0:
        actual_level = QsLogging::FatalLevel;
        break;
    case 1:
        actual_level = QsLogging::ErrorLevel;
        break;
    default:
    case 2:
        actual_level = QsLogging::WarnLevel;
        break;
    case 3:
        actual_level = QsLogging::InfoLevel;
        break;
    case 4:
        actual_level = QsLogging::DebugLevel;
        break;
    case 5:
        actual_level = QsLogging::TraceLevel;
        break;
    }

    // Return the level.
    return actual_level;
}

/*!
 * @brief Return true if the multicamera support is to be enabled, false otherwise.
 */
bool UserPreferences::MulticamEnabled()
{
    return ui->multicam_enable->isChecked();
}

/*!
 * @brief Return the number of network buffers the user would like to use.
 */
int UserPreferences::GetNetworkBufferCount()
{
    return ui->net_buffers_spin->value();
}

/*!
 * @brief Return the inter-packet delay the user wants to use.
 */
int UserPreferences::GetPacketDelay()
{
    return ui->packet_delay_spin->value();
}

/*!
 * @brief Return true if the user would like to maintain statistics on saves
 *        to the workstation, false otherwise.
 */
bool UserPreferences::SaveStatsEnabled()
{
    return ui->statistics_enable->isChecked();
}

void UserPreferences::SetAutoConnect(bool show, bool enable)
{
    ui->autoconnect->setVisible(show);
    ui->autoconnect->setChecked(enable);
}

int UserPreferences::GetUpdateFreq()
{
    return ui->cb_check_interval->currentIndex();
}

void UserPreferences::displayUpdates()
{
    int updates_avail;
    const FiUpdater *updater;

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    // See if there are any updates available.
    updates_avail = updater->GetUpdates();

    // Enable the packager group if there is a camera update available.
    ui->grp_download->setEnabled(updates_avail != 0);
    ui->grp_download->setVisible(true);

    // Is there a FasMotion update available?
    if (!MainWindow::local_control() && ((updates_avail & FI_UPDATER::AVAIL_FM_UPDATE) == FI_UPDATER::AVAIL_FM_UPDATE))
    {
        // Make the checkbox visible and set the text appropriately.
        ui->chk_fasmotion->setText(tr("FasMotion: Rev%1").arg(updater->GetVersionFM().toString()));
        ui->chk_fasmotion->setEnabled(true);
    }
    else
    {
        // Hide the checkbox visible and set the text appropriately.
        ui->chk_fasmotion->setEnabled(false);
        ui->chk_fasmotion->setText(tr("FasMotion: None available"));
    }

    // Is there a series3 camera update available?
    if ((updates_avail & FI_UPDATER::AVAIL_CAM_S3_UPDATE) == FI_UPDATER::AVAIL_CAM_S3_UPDATE)
    {
        // Make the checkbox visible and set the text appropriately.
        ui->chk_series3->setText(tr("IL/TS: Rev%1").arg(updater->GetVersionCamS3().toString()));
        ui->chk_series3->setEnabled(true);
    }
    else
    {
        // Hide the checkbox visible and set the text appropriately.
        ui->chk_series3->setEnabled(false);
        ui->chk_series3->setText(tr("IL/TS: None available"));
    }

    // Is there a seriesNC camera update available?
    if ((updates_avail & FI_UPDATER::AVAIL_CAM_NC_UPDATE) == FI_UPDATER::AVAIL_CAM_NC_UPDATE)
    {
        // Make the checkbox visible and set the text appropriately.
        ui->chk_seriesNC->setText(tr("HS: Rev%1").arg(updater->GetVersionCamNC().toString()));
        ui->chk_seriesNC->setVisible(true);

        ui->chk_seriesNC->setEnabled(false);
        ui->chk_seriesNC->setChecked(false);
    }
    else
    {
        // Hide the checkbox visible and set the text appropriately.
        ui->chk_seriesNC->setEnabled(false);
        ui->chk_seriesNC->setText(tr("HS: None available"));
    }
}

/*!
 * @brief Specify the number of backoff buffers we're currently using.
 * @param backoff_bufs  The number of backoff buffers.
 */
void UserPreferences::SetBackoffBuffers(int backoff_bufs)
{
    ui->backoff_buffers_spin->setValue(backoff_bufs);
}

/*!
 * @brief Set the factory menu enable checkbox.
 * @param do_enable     True if the factory menu is enabled, false otherwise.
 */
void UserPreferences::SetFactoryEnable(bool do_enable)
{
    if (do_enable)
    {
        if (factoryMenuEnable == nullptr)
        {
            factoryMenuEnable = new QCheckBox(this);
            factoryMenuEnable->setObjectName(QStringLiteral("factory_menu_enable"));
            factoryMenuEnable->setText(tr("Enable Factory Menu"));

            // Insert the checkbox above the vertical spacer.
            ui->enablesLayout->insertWidget(ui->enablesLayout->count() - 1, factoryMenuEnable);
        }

        factoryMenuEnable->setChecked(true);
        factoryMenuEnable->setVisible(true);
    }
    else if (factoryMenuEnable != nullptr)
    {
        factoryMenuEnable->setChecked(false);
        factoryMenuEnable->setVisible(false);
    }
}

void UserPreferences::SetPuttyPath(const QString& the_path)
{
    ui->putty_path->setText(QDir::toNativeSeparators(the_path));
}

/*!
 * @brief Set the current [logging level](@ref QsLogging::Level).
 * @param log_level     The current [logging level](@ref QsLogging::Level).
 */
void UserPreferences::SetLoggingLevel(QsLogging::Level log_level)
{
    int user_level;

    // Convert from the actual level to what the user sees.
    switch (log_level)
    {
    case QsLogging::FatalLevel:
        user_level = 0;
        break;
    case QsLogging::ErrorLevel:
        user_level = 1;
        break;
    default:
    case QsLogging::WarnLevel:
        user_level = 2;
        break;
    case QsLogging::InfoLevel:
        user_level = 3;
        break;
    case QsLogging::DebugLevel:
        user_level = 4;
        break;
    case QsLogging::TraceLevel:
        user_level = 5;
        break;
    }

    // Update our contrl.
    ui->log_level_knob->setValue(user_level);
}

/*!
 * @brief Set the multicamera support enable checkbox.
 * @param do_enable     True if the multicamera support is enabled, false otherwise.
 */
void UserPreferences::SetMulticamEnable(bool do_enable)
{
    ui->multicam_enable->setChecked(do_enable);
}

/*!
 * @brief Specify the number of network buffers currently in use.
 * @param net_bufs      The number of network buffers currently used.
 */
void UserPreferences::SetNetworkBuffers(int net_bufs)
{
    // Update the control
    ui->net_buffers_spin->setValue(net_bufs);
}

/*!
 * @brief Specify the inter-packet delay, in microseconds.
 * @param pkt_delay     The current default inter-packet delay value.
 */
void UserPreferences::SetPacketDelay(int pkt_delay)
{
    ui->packet_delay_spin->setValue(pkt_delay);
}

/*!
 * @brief Specify if we're maintaining statistics on saves to the workstation.
 * @param do_enable     True if we're maintaining statistics, false otherwise.
 */
void UserPreferences::SetSaveStatsEnable(bool do_enable)
{
    ui->statistics_enable->setChecked(do_enable);
}

void UserPreferences::SetCredentials(const QString& username, const QString& password, const QString& workgroup)
{
    if (username.isEmpty())
        ui->username_edit->setText(SERIESNC_CAM_USER_DEFAULT_NAME);
    else
        ui->username_edit->setText(username);

    if (password.isEmpty())
        ui->pass_edit->setText(SERIESNC_CAM_USER_DEFAULT_NAME);
    else
        ui->pass_edit->setText(password);

    if (workgroup.isEmpty())
        ui->workgroup_edit->setText(SERIESNC_CAM_USER_DEFAULT_WORKGROUP);
    else
        ui->workgroup_edit->setText(workgroup);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

void UserPreferences::accept()
{
    // Pass it on.
    QDialog::accept();
}

bool UserPreferences::AutoConnectEnabled()
{
    return ui->autoconnect->isChecked();
}

bool UserPreferences::AutoUpdateEnabled()
{
    return ui->automatic_group->isChecked();
}

void UserPreferences::browsePutty()
{
    QString fileName;

    fileName = QFileDialog::getOpenFileName(this, tr("PuTTY Path"), ui->putty_path->text(), tr("Files (*)"));
    if (!fileName.isEmpty() && QFile::exists(fileName))
    {
        ui->putty_path->setText(QDir::toNativeSeparators(fileName));
    }
}

void UserPreferences::changePage(QListWidgetItem *current, QListWidgetItem *previous)
{
    QString my_label;

    // Make sure there is a current page selected.
    if (!current)
        current = previous;

    // Just a sanity check.
    if (!current)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No current, no previous?";
        return;
    }

    // Tell the stack to switch pages.
    ui->optionStack->setCurrentIndex(ui->setting_list->row(current));

    // Change the title.
    my_label = QString("<h2>") + current->text() + QString("</h2>");
    ui->page_label->setText(my_label);
}

void UserPreferences::checkTimeChanged(int index)
{
    QDateTime last_check = QDateTime::fromString(ui->last_check->text());
    ui->check_date->setText(FiUpdater::GetNextCheck(index, last_check).toString());
}

/*!
 * @brief The user has requested that we perform an update check.
 */
void UserPreferences::doCheck()
{
    const FiUpdater *updater;

    // Disable the button.
    ui->pb_check->setEnabled(false);

    // Show the check in progress controls.
    if (serverUrl.isEmpty())
        ui->lbl_checking->setText(tr("Check in progress..."));
    else
        ui->lbl_checking->setText(tr("Check in progress...[%1]").arg(serverUrl));
    ui->grp_download->setEnabled(false);

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();
    connect(updater, &FiUpdater::error, this, &UserPreferences::updateError);
    connect(updater, &FiUpdater::finished, this, &UserPreferences::updateFinished);

    // Request an update check.
    updater->UpdateCheck();
}

void UserPreferences::doBrowseDownloadPath()
{
    QString dirName;
    FiUpdater *updater;

    // Get the updater object.
    updater = MainWindow::GetInstance()->get_updater();

    // Get the previous download path.
    dirName = updater->GetUpdatePath();

    // Ask the user for the destination for the update package.
    dirName = QFileDialog::getExistingDirectory(this, tr("Select a Directory to save the update package"), dirName);
    if (dirName.isEmpty())
        return;

    // Set the path.
    ui->download_path->setText(QDir::toNativeSeparators(dirName));

    // Tell the updater object about the path.
    updater->SetUpdatePath(dirName);
}

/*!
 * @brief The user has requested to download a camera update package.
 *
 * We'll need to try to download the update package for the Fastec PPA. Assuming
 * that goes well, well need to create a camera-suitable update.
 *
 * This means different things for different camera series. For the series3
 * cameras (TS/IL) it means we run the updater which will create an SDCard
 * for the user to perform an update with.
 */
void UserPreferences::doDownloadPackage()
{
    int first;
    FiUpdater *updater;

    // Clear our download list.
    download_list.clear();

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    // Make sure we have a path.
    if (ui->download_path->text().isEmpty())
    {
        ui->download_path->setFocus();
        QMessageBox::information(nullptr, "Required Parameter Missing", "Where do you want to save the package to?");
        return;
    }

    // Are we downloading the FasMotion update package?
    if (ui->chk_fasmotion->isChecked())
    {
        // Make sure the file doesn't already exist.
        if (QFile::exists(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE)))
        {
            if (QMessageBox::question(this, tr("Overwrite Existing File"),
                                      tr("There already exists a file called %1. Overwrite?").arg(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE)),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
                == QMessageBox::No)
            {
                // Exit.
                goto dp_bail;
            }

            // Go ahead and blow away the file.
            QFile::remove(updater->GetPackage(FI_UPDATER::AVAIL_FM_UPDATE));
        }

        // Push this on our queue.
        download_list.append(FI_UPDATER::AVAIL_FM_UPDATE);
    }

    // Are we downloading the series3 update package?
    if (ui->chk_series3->isChecked())
    {
        // Make sure the file doesn't already exist.
        if (QFile::exists(updater->GetPackage(FI_UPDATER::AVAIL_CAM_S3_UPDATE)))
        {
            if (QMessageBox::question(this, tr("Overwrite Existing File"),
                                      tr("There already exists a file called %1. Overwrite?").arg(updater->GetPackage(FI_UPDATER::AVAIL_CAM_S3_UPDATE)),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
                == QMessageBox::No)
            {
                // Exit.
                goto dp_bail;
            }

            // Go ahead and blow away the file.
            QFile::remove(updater->GetPackage(FI_UPDATER::AVAIL_CAM_S3_UPDATE));
        }

        // Push this on our queue.
        download_list.append(FI_UPDATER::AVAIL_CAM_S3_UPDATE);

    }

    // Are we downloading the series3 update package?
    if (ui->chk_seriesNC->isChecked())
    {
        // Make sure the file doesn't already exist.
        if (QFile::exists(updater->GetPackage(FI_UPDATER::AVAIL_CAM_NC_UPDATE)))
        {
            if (QMessageBox::question(this, tr("Overwrite Existing File"),
                                      tr("There already exists a file called %1. Overwrite?").arg(updater->GetPackage(FI_UPDATER::AVAIL_CAM_NC_UPDATE)),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
                == QMessageBox::No)
            {
                // Exit.
                goto dp_bail;
            }

            // Go ahead and blow away the file.
            QFile::remove(updater->GetPackage(FI_UPDATER::AVAIL_CAM_NC_UPDATE));
        }

        // Push this on our queue.
        download_list.append(FI_UPDATER::AVAIL_CAM_NC_UPDATE);
    }

    // Now pop the first download off the list.
    first = download_list.takeFirst();

    // Connect to download complete, so we know when to move on.
    connect(updater, &FiUpdater::download_complete, this, &UserPreferences::downloadNext);

    // Have the updater download the package.
    updater->UpdateDownload(first);
    return;

dp_bail:
    if (!download_list.isEmpty())
        download_list.clear();
}

/*!
 * @brief The user has checked/unchecked one of the package selections.
 *
 * Only enable the download button if at least one of the packages is selected.
 */
void UserPreferences::doSelectPackage()
{
    ui->pb_download->setEnabled(ui->chk_series3->isChecked() || ui->chk_seriesNC->isChecked() || ui->chk_fasmotion->isChecked());
    if (ui->pb_download->isEnabled())
        ui->pb_download->setFocus();
}

void UserPreferences::downloadNext(int which_update)
{
    int next_pkg;
    FiUpdater *updater;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "which_update:" << which_update;

    switch (which_update)
    {
    case FI_UPDATER::AVAIL_FM_UPDATE:
        // Clear the checkbox for the installer we just downloaded.
        ui->chk_fasmotion->setChecked(false);

        // Ask the user if they'd like to run the downloaded updater.
        if (QMessageBox::question(this, tr("FasMotion Update"), tr("Would you like to start the update now?"),
                                  QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
            == QMessageBox::Yes)
        {
            // Clear the download_list, if it isn't already.
            if (!download_list.isEmpty())
                download_list.clear();

            // Ask the main window to start the update. This will cause the application to close.
            MainWindow::GetInstance()->do_update();
        }
        break;
    case FI_UPDATER::AVAIL_CAM_S3_UPDATE:
        // Clear the checkbox for the installer we just downloaded.
        ui->chk_series3->setChecked(false);

#ifdef Q_OS_WIN
        // Ask the user if they'd like to run the downloaded updater.
        if (QMessageBox::question(this, tr("TL/IL Update"), tr("Would you like to build the SDCard now?"),
                                  QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
            == QMessageBox::Yes)
        {
            // Start building the SDCard.
            CamUpdateDlg::BuildSDCard();
        }
#endif
        break;
    case FI_UPDATER::AVAIL_CAM_NC_UPDATE:
        // Clear the checkbox for the installer we just downloaded.
        ui->chk_seriesNC->setChecked(false);
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected which_update:" << which_update;
        break;
    }

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    // Are we done?
    if (download_list.isEmpty())
    {
        // If the list is empty, disconnect from the download complete signal.
        disconnect(updater, &FiUpdater::download_complete, this, &UserPreferences::downloadNext);
        return;
    }

    // Get the next on the list.
    next_pkg = download_list.takeFirst();

    // Have the updater download the package.
    updater->UpdateDownload(next_pkg);
}

void UserPreferences::showPassword()
{
    if (ui->show_checkbox->isChecked())
        ui->pass_edit->setEchoMode(QLineEdit::Normal);
    else
        ui->pass_edit->setEchoMode(QLineEdit::Password);
}

/*!
 * @brief The number of backoff buffers we support is directly related
 *        to the number of network buffers are configured.
 *
 * @param num_net_buffers       The number of network buffers.
 */
void UserPreferences::updateBackoff(int num_net_buffers)
{
    int     backoff_value;

    // Limit the backoff buffers to 50% of the number of net buffers
    // these are now the number of buffers in our memory pool and we
    // need to contain to a subset of that pool.
    backoff_value = (num_net_buffers / 2) + 1;

    // Before we update the maximum backoff value, let's make sure that
    // the current setting is not too big.
    if (ui->backoff_buffers_spin->value() > backoff_value)
        ui->backoff_buffers_spin->setValue(backoff_value);

    // Update the maximum.
    ui->backoff_buffers_spin->setMaximum(backoff_value);
}

void UserPreferences::updateFinished(int updates_avail)
{
    const FiUpdater *updater;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "updates_avail:" << updates_avail;

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    // Disconnect from the updater.
    disconnect(updater, &FiUpdater::error, this, &UserPreferences::updateError);
    disconnect(updater, &FiUpdater::finished, this, &UserPreferences::updateFinished);

    // Enable the button.
    ui->pb_check->setEnabled(true);
    ui->pb_check->setFocus();

    // If there's an error, then we're done.
    if (updater->GetError() == 0)
    {
        // Are there any updates available?
        if (updates_avail == 0)
        {
            // Make sure the available labels are hidden.
            ui->grp_download->setVisible(false);

            // Indicate that there are no updates available.
            if (serverUrl.isEmpty())
                ui->lbl_checking->setText(tr("No updates available"));
            else
                ui->lbl_checking->setText(tr("No updates available[%1]").arg(serverUrl));
        }
        else
        {
            // Indicate that we are done.
            if (serverUrl.isEmpty())
                ui->lbl_checking->setText(tr("Check complete..."));
            else
                ui->lbl_checking->setText(tr("Check complete...[%1]").arg(serverUrl));

            // Display which updates are available.
            displayUpdates();
        }
    }
}

void UserPreferences::updateError(int err_value)
{
    QString err_msg;
    const FiUpdater *updater;

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "err_value:" << err_value << "[" << updater->GetErrorString() << "]";
    err_msg = tr("Update check failed [err=%1]").arg(err_value);
    ui->lbl_checking->setText(err_msg);
}
