/*!
 * @file video_controls.h
 * @brief This is the header file for the VideoControls class.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef VIDEO_CONTROLS_H
#define VIDEO_CONTROLS_H

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QDockWidget>
#include <stdint.h>

namespace Ui {
    class VideoControls;
}

namespace VIDEO_CONTROLS {

    //-----------------------------------------------------------------------------
    // Constant Definitions
    //-----------------------------------------------------------------------------

    const int   MAX_GRAPH_COUNTROLS     = 3;        //!< The maximum number of graphing controls displayed.
    const int   DEF_PLAYBACK_RATE       = 20;       //!< The default playback rate.

}   // End of namespace VIDEO_CONTROLS

//-----------------------------------------------------------------------------
// Class Forward Declarations
//-----------------------------------------------------------------------------

class QMdiSubWindow;
class QPushButton;
class QSignalMapper;
class QSlider;
class QSpacerItem;
class QTimerEvent;
class QToolButton;
class QwtPlot;

class fiSlider;
class fiStringSpinBox;

class VideoControls : public QDockWidget
{
    Q_OBJECT
    
public:
    explicit VideoControls( QWidget *parent = 0 );
    ~VideoControls();

    void        enable_graph( bool do_enable );
    uint32_t    get_clip_end( void );
    uint32_t    get_clip_start( void );
    uint32_t    get_current_frame_num( void );
    int         get_playback_rate( void );
    bool        graph_is_enabled( void )                { return graph_enable; }
    void        reset_clip( void );
    void        set_camera_ip( uint32_t ip_addr );
    void        set_current_frame_num( uint32_t cur_frame );
    void        set_graph_data( uint8_t *data_ptr, int data_len );
    void        set_markers( const QList<int> &markers );

    const QList<int>& markers( void ) const;

signals:
    void    clip_end_changed( void );
    void    clip_start_changed( void );
    void    current_frame_changed( void );

    void    jump_time( void );
    void    play_forward( void );
    void    play_reverse( void );
    void    play_update( void );
    void    review_stop( void );

public slots:
    void    playback_stop( void );
    void    read_panel_regs( void );

    virtual void    setVisible( bool visible );

protected:
    virtual void    timerEvent( QTimerEvent *event );

private slots:
    void    handle_unsol_msg( uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value );
    void    msg_timeout_event( uint32_t ip_addr, uint32_t reg_addr, bool read_op );
    void    read_complete( uint32_t ip_addr, uint32_t reg_addr );
    void    write_complete( uint32_t ip_addr, uint32_t reg_addr );

    void    clip_changed( void );
    void    clip_end_changing( int new_value );
    void    clip_end_clicked( void );
    void    clip_slider_released( void );
    void    clip_start_changing( int new_value );
    void    clip_start_clicked( void );

    void    cur_frame_action( int action );
    void    cur_frame_changing( int new_value );
    void    cur_frame_pressed( void );
    void    cur_frame_released( void );

    void    add_chart( void );
    void    remove_chart( QWidget *p_caller );
    void    spin_signal_change( QWidget *p_caller );

    void    delayed_update( void );

    void    jump_forward( void );
    void    jump_reverse( void );
    void    jump_trigger( void );
    void    stepping_forward( void );
    void    stepping_reverse( void );

    void    zoom_action( int action );
    void    zoom_change( void );
    void    zoom_hide( void );
    void    zoom_in( void );
    void    zoom_move( void );
    void    zoom_out( void );
    void    zoom_show( void );

private:
    void        create_chart_controls( void );
    void        create_zoom_controls( void );
    void        init_controls( void );
    uint32_t    read_camera_cache( uint32_t reg_addr, bool *ok = NULL );
    void        read_playback_regs( void );
    void        start_blackout( void );
    void        update_buttons( bool do_enable );
    void        update_frame_info( void );
    void        update_playback_frame( void );
    void        update_playback_info( void );
    void        update_plot( fiStringSpinBox *p_active_spin = NULL );
    void        update_roi_rate( void );
    void        update_time_display( int frame_no );
    void        write_register( uint32_t reg_addr, uint32_t reg_value );

private:
    Ui::VideoControls *ui;                          //!< Our UI object.

    bool        allow_frame_updates;                //!< Flag to allow/disallow frame updates.
    int         blackout_id;                        //!< Timer ID to blackout updates when we're dynamically moving around a video.
    int         play_update_id;                     //!< Timer ID for pacing the playback of the video.
    int         scrub_update_id;                    //!< Timer ID to pace the updates when the user is "scrubbing" the slider.

    bool        ignore_updates;                     //!< Flag to ignore updates until we get connected.
    uint32_t    cam_ip_addr;                        //!< IP address of the connected camera.
    int         last_frame_num;                     //!< Used to remember the last (relative) frame the user selected.
    uint32_t    last_reg_addr;                      //!< Used to terminate a block register read, this contains the last register address in the block.

    // Signal graph support.
    QList<fiStringSpinBox *>      spin_signal;      //!< List of spinner controls to allow the user to specify a signal (Arm-In, Sync-Out, etc.)
    QSignalMapper           *spin_mapper;           //!< Signal mapper to manage the signals generated by the "spin_signal" controls.
    QList<QwtPlot *>        plot_curve;             //!< List of plots showing the signals values.
    QList<QPushButton *>    pb_remove_chart;        //!< List of push button controls to remove an associated IO graph.
    QSignalMapper           *remove_mapper;         //!< Signal mapper to manage the signals generated by the "pb_remove_chart" controls.
    QPushButton     *pb_add_chart;                  //!< Push button to allow the user to add an IO graph.
    QSpacerItem     *left_spacer;                   //!< Spacer control required when more than one IO graph is displayed.
    QSpacerItem     *right_spacer;                  //!< Spacer control required when more than one IO graph is displayed.
    QStringList     signal_names;                   //!< List of the various signals (Arm-In, Sync-Out, etc.)
    bool            graph_enable;                   //!< Flag to indicate if IO graphs are enabled (true) or disabled (false).
    uint8_t         *raw_data_ptr;                  //!< Pointer to the raw per-frame metadata buffer.
    uint32_t        raw_data_len;                   //!< The number of bytes in the raw per-frame metadata buffer.
    int             axis_len;                       //!< The number of entries (aka frames) along the x-axis.
    double          *x_axis;                        //!< The array of x-axis values.
    double          *y_ao_axis;                     //!< The array of Arm-Out values.
    double          *y_ai_axis;                     //!< The array of Arm-In values.
    double          *y_so_axis;                     //!< The array of Sync-Out values.
    double          *y_si_axis;                     //!< The array of Sync-In values.
    double          *y_to_axis;                     //!< The array of Trigger-Out values.
    double          *y_ti_axis;                     //!< The array of Trigger-In values.

    // Zoom controls we need.
    QPushButton     *pb_zoom_in;                    //!< Push button control to allow the user to zoom in on the frame bar.
    QPushButton     *pb_zoom_out;                   //!< Push button control to allow the user to zoom out of the frame bar.
    fiSlider        *zoom_scrollbar;                //!< When zoomed in, the "magnified" frame bar.
};

#endif // VIDEO_CONTROLS_H
