/*!
 * @file whitebal_dialog.cpp
 * @brief This file contains the implementation of the WhiteBalDlg
 *        class. This class allows the user to select a region of
 *        the image to use for calculating the custom white balance
 *        gains.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "whitebal_dialog.h"
#include "ui_whitebal_dialog.h"

#include <QGraphicsRectItem>
#include <QGraphicsSceneMouseEvent>
#include <QTimerEvent>

#include "fi_comm.h"
#include "fi_rectitem.h"
#include "logger.h"
#include "mainwindow.h"
#include "color_tool.h"

namespace WHITEBAL_DLG
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Our update frequency to gather histogram information.
const int   UPDATE_INTERVAL         = 250;

}   // End of namespace WHITEBAL_DLG.

using namespace WHITEBAL_DLG;

/*!
 * @brief The WhiteBalDlg constructor.
 */
WhiteBalDlg::WhiteBalDlg(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::WhiteBalDlg)
{
    FI_comm         *comm_ptr;

    // Setup the UI controls.
    ui->setupUi(this);

    // Initialize our controls.
    cct = 0;
    cam_ip_addr = 0;
    histo_size = 0;
    reticle = nullptr;

    // Allocate our timer.
    update_timer = new QTimer(this);
    update_timer->setInterval(UPDATE_INTERVAL);

    ui->image_view->setRenderHints(QPainter::Antialiasing);
    ui->image_view->setBackgroundRole(QPalette::Dark);
    ui->roi_histo->setVisible(false);
    ui->cb_show_histogam->setChecked(false);

    QList<int> initial_sizes;
    initial_sizes.append(ui->image_view->minimumWidth());
    initial_sizes.append(ui->roi_histo->minimumWidth());
    ui->splitter->setSizes(initial_sizes);

    // Get our communications pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to get a FI_comm object!";
        return;
    }

    // Connect our signals and slots.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &WhiteBalDlg::read_complete, Qt::QueuedConnection);
    connect(ui->cb_show_histogam, &QCheckBox::clicked, this, &WhiteBalDlg::show_histo);
    connect(update_timer, &QTimer::timeout, this, &WhiteBalDlg::time_to_update);
    connect(ui->splitter, &QSplitter::splitterMoved, this, &WhiteBalDlg::update_image);
}

/*!
 * @brief The WhiteBalDlg destructor.
 */
WhiteBalDlg::~WhiteBalDlg()
{
    FI_comm         *comm_ptr;
    QGraphicsScene *scene_ptr;

    // Do we have a timer running?
    if (update_timer != nullptr)
    {
        update_timer->stop();
        update_timer->deleteLater();
        update_timer = nullptr;
    }

    // Get our communications pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr != nullptr)
    {
        // Disconnect from the communiations object.
        disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &WhiteBalDlg::read_complete);
    }

    // Get a pointer to our scene.
    scene_ptr = ui->image_view->scene();

    // If there's a scene and a reticle, remove the reticle.
    if ((scene_ptr != nullptr) && (reticle != nullptr))
    {
        scene_ptr->removeItem(reticle);
        delete reticle;
        reticle = nullptr;
    }

    // Clean up the UI object.
    delete ui;
}

/*!
 * @brief Return the reticle used for the white balance histogram.
 */
QRect WhiteBalDlg::get_roi()
{
    QRect our_roi = QRect(0, 0, 0, 0);

    // Sanity check.
    if (reticle != nullptr)
    {
        QRectF user_rect;

        // Get the ROI in scene coordinates.
        user_rect = reticle->mapRectToScene(reticle->rect());

        // Convert to integer values.
        our_roi = user_rect.toAlignedRect();
    }

    // Return the ROI.
    return our_roi;
}

/*!
 * @brief Start to read the white balance histogram.
 */
void WhiteBalDlg::histo_read_start()
{
    uint32_t    read_len;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Read up the size of the histogram.
    histo_size = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SIZE);

    // Reset our buffer.
    histo_data.clear();
    histo_data.reserve((int)histo_size / 4);

    // How much can we read?
    if (histo_size > GIGE_HISTOGRAM_READ_SIZE)
        read_len = GIGE_HISTOGRAM_READ_SIZE;
    else
        read_len = histo_size;

    // Start reading the histogram data.
    comm_ptr->start_read_mem(cam_ip_addr, GIGE_REG_HISTO_RGB, read_len);
}

/*!
 * @brief Continue to read the white balance histogram.
 */
void WhiteBalDlg::histo_read_update()
{
    int         remaining_bytes;
    QByteArray  *read_data;
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Read the data we got from the camera.
    read_data = comm_ptr->get_reg_buffer(cam_ip_addr, GIGE_REG_HISTO_RGB);
    if ((read_data != nullptr) && (read_data->size() > 0))
    {
        // Save the old size of our existing data.
        int old_size = histo_data.size();

        // Update the data vector size. Remember that the read_data
        // size is the number of bytes. Convert that to the number
        // of 32-bit words.
        histo_data.resize(histo_data.size() + (read_data->size() / 4));

        // Get a pointer to where the new data is to be placed.
        uint32_t *data_ptr = histo_data.data();
        data_ptr += old_size;

        // Copy the data over.
        memcpy(data_ptr, read_data->constData(), (size_t)read_data->size());
    }

    // Release the buffer.
    delete read_data;

    // Are we done?
    remaining_bytes = (int)histo_size - (histo_data.size() * 4);
    if (remaining_bytes > 0)
    {
        uint32_t    read_len;

        // Ask the camera for the next "chunk".
        if (remaining_bytes > GIGE_HISTOGRAM_READ_SIZE)
            read_len = GIGE_HISTOGRAM_READ_SIZE;
        else
            read_len = (uint32_t)remaining_bytes;

        // Start reading the histogram data.
        comm_ptr->start_read_mem(cam_ip_addr, GIGE_REG_HISTO_RGB, read_len);
    }
    else
    {
        uint32_t    sum_value;
        uint32_t    sum_sq_l;
        uint32_t    sum_sq_h;
        uint64_t    sum_sq_value;

        // Check to make sure our length agrees with what we expect.
        if ((histo_data.size() * 4) > (int)histo_size)
            histo_data.resize(histo_size / 4);

        // All done reading the histogram data.
        ui->roi_histo->populate_histogram(histo_data);

        // Read up and populate the statistics information we got from the camera.
        sum_value = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_BLU);
        sum_sq_h = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_H);
        sum_sq_l = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_L);
        sum_sq_value = (((uint64_t)sum_sq_h) << 32) | ((uint64_t)sum_sq_l);
        ui->roi_histo->set_statistics(FI_HISTOGRAM::BLUE_CURVE, sum_value, sum_sq_value);

        sum_value = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_GRN);
        sum_sq_h = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_H);
        sum_sq_l = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_L);
        sum_sq_value = (((uint64_t)sum_sq_h) << 32) | ((uint64_t)sum_sq_l);
        ui->roi_histo->set_statistics(FI_HISTOGRAM::GREEN_CURVE, sum_value, sum_sq_value);

        sum_value = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_RED);
        sum_sq_h = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_H);
        sum_sq_l = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_L);
        sum_sq_value = (((uint64_t)sum_sq_h) << 32) | ((uint64_t)sum_sq_l);
        ui->roi_histo->set_statistics(FI_HISTOGRAM::RED_CURVE, sum_value, sum_sq_value);

        // =================================================================================
        // ALM: I'm putting the CCT calculations into it's own block because I'm not certain
        //      this is the right place. It works here, but it might be better somewhere else.
        {
            uint32_t peaks = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_PEAK);
            FI_COLORTOOL::RGBColor rgb_color((peaks & GIGE_HISTO_RGB_PEAK_RED_MASK) >> GIGE_HISTO_RGB_PEAK_RED_SHIFT,
                                             (peaks & GIGE_HISTO_RGB_PEAK_GRN_MASK) >> GIGE_HISTO_RGB_PEAK_GRN_SHIFT,
                                             (peaks & GIGE_HISTO_RGB_PEAK_BLU_MASK) >> GIGE_HISTO_RGB_PEAK_BLU_SHIFT);
            FI_COLORTOOL::XYZColor xyz_color;

            // Convert the RGB color to XYZ so we can calculate the CCT. Use D50, since that seems
            // to be what the DNG spec calls for (see "Camera to XYZ (D50) Transform").
            xyz_color = ColorTool::rgbToXYZ(rgb_color, true);
            // QLOG_DEBUG() << __PRETTY_FUNCTION__ << "peaks:" << Qt::showbase << Qt::hex << peaks << "rgb:" << rgb_color.toString() << "xyz:" << xyz_color.toString();

            // Calculate the CCT.
            cct = ColorTool::xyzToCCT(xyz_color);
        }
        // =================================================================================

        // Go around for another run.
        update_histo();
    }
}

/*!
 * @brief Build up the scene with the white balance reticle.
 * @param scene_ptr     Pointer to our graphics scene object.
 */
void WhiteBalDlg::set_scene(QGraphicsScene *scene_ptr)
{
    // Make sure we have a scene.
    if (scene_ptr != nullptr)
    {
        QRectF  scene_rect;
        QRectF  reticle_rect;
        fiRectItem  *my_rect;

        // Get the bounding rectangle for the scene.
        scene_rect = scene_ptr->sceneRect();
        scene_ptr->setSceneRect(scene_rect);

        // We want a reticle the user can drag around to specify the region
        // to use for calculating the custom white balance gains. Center the
        // reticle and make it 10% the size of the scene.
        reticle_rect = scene_rect;
        reticle_rect.setHeight(scene_rect.height() * 0.05);
        reticle_rect.setWidth(scene_rect.width() * 0.05);
        reticle_rect.moveLeft((scene_rect.width() - reticle_rect.width()) / 2);
        reticle_rect.moveTop((scene_rect.height() - reticle_rect.height()) / 2);

        reticle = new fiRectItem(reticle_rect);
        reticle->setPen(QPen(Qt::white, 2));

        // We need a fiRectItem for setting the bounds.
        my_rect = static_cast<fiRectItem *>(reticle);
        my_rect->set_edge(scene_rect.bottomRight());

        // Make sure the user can grab the rectangle and move it around.
        reticle->setFlag(QGraphicsItem::ItemIsMovable, true);

        // Add it to the scene.
        scene_ptr->addItem(reticle);
        reticle->setZValue(1);

        // Scale the scene so it fits in our view, maintaining the aspect ratio.
        ui->image_view->resetTransform();
        ui->image_view->fitInView(scene_rect, Qt::KeepAspectRatio);
    }

    // Set the scene into our view.
    ui->image_view->setScene(scene_ptr);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief The window has resized. We need to update the image dimensions.
 * @param event     Pointer to the event structure.
 */
void WhiteBalDlg::resizeEvent(QResizeEvent *event)
{
    Q_UNUSED(event);
    update_image(0, 0);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Start up the white balance dialog. We use this opportunity to
 *        start a single shot time to begin updating our histogram.
 *
 * @return The result of QDialog::exec();
 */
int WhiteBalDlg::exec()
{
    // In a little bit, start getting histograms from the camera.
    QTimer::singleShot(UPDATE_INTERVAL, this, &WhiteBalDlg::finish_init);

    // Pass it on.
    return QDialog::exec();
}

/*!
 * @brief Time delayed finish to the initialization process.
 */
void WhiteBalDlg::finish_init()
{
    update_image(0, 0);
    QTimer::singleShot(UPDATE_INTERVAL, this, &WhiteBalDlg::update_histo);
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void WhiteBalDlg::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Is this event for us?
    if (ip_addr != cam_ip_addr)
        return;

    // Perform any special action we might want on a particular register.
    switch (reg_addr)
    {
    case GIGE_REG_HISTO_RGB_SIZE:
        histo_read_start();
        break;
    case GIGE_REG_HISTO_RGB:
        histo_read_update();
        break;
    default:
        break;
    }
}

/*!
 * @brief Show the white balance histogram.
 */
void WhiteBalDlg::show_histo()
{
    QRectF scene_rect;

    // Change if the histogram is visible or not.
    ui->roi_histo->setVisible(ui->cb_show_histogam->isChecked());

    // Let the histogram change take effect.
    qApp->processEvents();

    // Get the bounding rectangle for the scene.
    scene_rect = ui->image_view->scene()->sceneRect();

    // Scale the scene so it fits in our view, maintaining the aspect ratio.
    ui->image_view->resetTransform();
    ui->image_view->fitInView(scene_rect, Qt::KeepAspectRatio);
}

/*!
 * @brief It is time to update the white balance historam.
 */
void WhiteBalDlg::time_to_update()
{
    FI_comm     *comm_ptr;
    FI_COMM::REG_ADDR_LIST  read_list;

    // Stop the timer.
    update_timer->stop();

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Start reading the histogram data.
    read_list.append(GIGE_REG_HISTO_RGB_SIZE);
    read_list.append(GIGE_REG_HISTO_RGB_PEAK);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_BLU);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_L);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_BLU_H);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_GRN);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_L);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_GRN_H);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_RED);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_L);
    read_list.append(GIGE_REG_HISTO_RGB_SUM_X_SQRD_RED_H);

    // Start the read.
    comm_ptr->start_read_multireg(cam_ip_addr, read_list);
}

/*!
 * @brief Tell the camera to start generating a histogram within
 *        the ROI of the white balance reticle.
 */
void WhiteBalDlg::update_histo()
{
    QRect       roi_rect;
    uint32_t    reg_value;
    FI_comm     *comm_ptr;

    // Sanity check.
    if (cam_ip_addr == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No IP address specified!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Get the ROI location.
    roi_rect = get_roi();

    // Tell the camera our area of interest.
    reg_value = ((uint32_t)roi_rect.x() << 16) | (uint32_t)roi_rect.width();
    comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_ROI_X, reg_value);
    reg_value = ((uint32_t)roi_rect.y() << 16) | (uint32_t)roi_rect.height();
    comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_HISTO_RGB_ROI_Y, reg_value);

    // Start the gathering proceess.
    comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_GENERATE_RGB_HISTO, GIGE_GENERATE_RGB_HISTO_CHAN_P0);

    // Start up our timer.
    update_timer->start();
}

/*!
 * @brief Update the size of the image.
 *
 * @param pos       [UNUSED]
 * @param index     [UNUSED]
 */
void WhiteBalDlg::update_image(int pos, int index)
{
    QRectF scene_rect;

    Q_UNUSED(pos);
    Q_UNUSED(index);

    // Get the bounding rectangle for the scene.
    scene_rect = ui->image_view->scene()->sceneRect();

    // Scale the scene so it fits in our view, maintaining the aspect ratio.
    ui->image_view->resetTransform();
    ui->image_view->fitInView(scene_rect, Qt::KeepAspectRatio);
}
