/*!
 * @file wifi_config.cpp

 * @brief Implementation of the Wifi configuration dialog.
 *
 * It takes the new settings and applies them to the camera
 * selected by the user. Allowing changes to IP address, etc.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "wifi_config.h"
#include "ui_wifi_config.h"

#include <QApplication>

#include "camera_mdi.h"
#include "fi_comm.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"
#include "wifi_scan.h"

// Use a namespace to avoid polluting the global space.
namespace WIFI_CONFIG
{
//--------------------------------------------------------------------
// Contant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Give the camera a chunk of time to perform the scan operation.
const int       SCAN_WAKE_INTERVAL  = 500;

//! Maximum number of times we'll retry a write operation.
const int       MAX_WRITE_RETRIES   = 3;
}

/*!
 * @brief Constructor for the WifiConfig.
 */
WifiConfig::WifiConfig(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::WifiConfig),
    write_retries(0)
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent);

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Parent is not a CameraMDI object!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Initialize the UI elements.
    ui->setupUi(this);

    // Limit the length of the passphrase string.
    ui->eb_password->setMaxLength(GIGE_WIFI_PASSPHRASE_MAX);

    // Connect our signals/slots.
    connect(ui->cb_security, qOverload<int>(&QComboBox::currentIndexChanged), this, &WifiConfig::security_changed);
    connect(ui->pb_cancel, &QPushButton::clicked, this, &WifiConfig::reject);
    connect(ui->pb_done, &QPushButton::clicked, this, &WifiConfig::on_connect);
    connect(ui->pb_scan, &QPushButton::clicked, this, &WifiConfig::on_scan);
    connect(ui->rb_dhcp, &QRadioButton::toggled, this, &WifiConfig::dhcp_toggled);

    connect(comm_ptr, &FI_comm::read_reg_complete, this, &WifiConfig::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &WifiConfig::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &WifiConfig::write_complete, Qt::QueuedConnection);

    // If the parent is closing, we need to bail.
    connect(my_parent, &CameraMDI::closing, this, &WifiConfig::reject);

    // Initialize the dialog.
    init_dialog();

    // Startup our poller.
    connect(&status_timer, &QTimer::timeout, this, &WifiConfig::status_update);
    status_timer.start(WIFI_CONFIG::SCAN_WAKE_INTERVAL);
}

/*!
 * @brief Destructor for the WifiConfig.
 */
WifiConfig::~WifiConfig()
{
    // Make sure the cursor is restored.
    QApplication::restoreOverrideCursor();

    // Shut down our poller.
    status_timer.stop();

    // Get rid of the UI object.
    delete ui;
}

/*!
 * @brief Return true if the user wants to connect using AdHoc mode.
 */
bool WifiConfig::adhoc()
{
    return ui->chk_adhoc->isChecked();
}

/*!
 * @brief Write the registers necessary to configure the camera to connect
 *        to the user specified SSID.
 */
void WifiConfig::connect_to_ap()
{
    uint32_t    reg_value;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Parent is nullptr!";
        ui->control_frame->setEnabled(false);
        ui->pb_done->setEnabled(false);
        ui->pb_scan->setEnabled(false);
        return;
    }

    // Reset our list of register we want to write.
    write_list.clear();

    // Setup the write to the configuration register. This will actually be written after
    // the password goes out, after the SSID goes out.
    //
    //---------------------------------------------------------------------------
    // Note that this register does NOT need to be byte swapped first!
    //---------------------------------------------------------------------------
    reg_value = get_security();
    if (adhoc())
        reg_value |= GIGE_WIFI_AP_CFG_ADHOC;
    write_list.append(FM_PANEL::REG_VALUE_PAIR(GIGE_REG_WIFI_AP_CFG, reg_value));

    // Do we want to use DHCP?
    if (use_dhcp())
    {
        // Enable DHCP.
        reg_value = GIGE_IFCONFIG_DHCP_BE;
        write_list.append(FM_PANEL::REG_VALUE_PAIR(GIGE_REG_WIFI_IP_CUR_CFG, qToBigEndian(reg_value)));
    }
    else
    {
        // Send all the static IP settings to the camera.
        write_list.append(FM_PANEL::REG_VALUE_PAIR(GIGE_REG_WIFI_IP_ADDR, qToBigEndian(user_ip.toIPv4Address())));
        write_list.append(FM_PANEL::REG_VALUE_PAIR(GIGE_REG_WIFI_DEF_GATEWAY, qToBigEndian(user_gw.toIPv4Address())));
        write_list.append(FM_PANEL::REG_VALUE_PAIR(GIGE_REG_WIFI_SUBNET_MASK, qToBigEndian(user_net_mask.toIPv4Address())));

        // Tell the camera that we're using a Static IP.
        reg_value = GIGE_IFCONFIG_PERSISTENT_BE;
        write_list.append(FM_PANEL::REG_VALUE_PAIR(GIGE_REG_WIFI_IP_CUR_CFG, qToBigEndian(reg_value)));
    }

    // Change to a busy cursor.
    QApplication::setOverrideCursor(Qt::BusyCursor);

    // Don't let the user get into trouble.
    ui->control_frame->setEnabled(false);
    ui->pb_done->setEnabled(false);
    ui->pb_scan->setEnabled(false);

    // The camera expects to get the null terminator at the end of the string.
    QByteArray ssid_array = user_ssid.toUtf8();
    ssid_array.append('\0');

    // Tell the camera the new SSID.
    my_parent->write_camera_buffer(GIGE_REG_WIFI_SSID, ssid_array);
}

/*!
 * @brief Convert the security index selected by the user to a GIGE_WIFI_AP_CFG_SEC_XXX value.
 *
 * @note    It is assumed that the combobox indecies are equivalent to the GIGE_WIFI_AP_CFG_SEC_xxx values!
 *
 * @return The user selected security level.
 */
uint32_t WifiConfig::get_security()
{
    return (uint32_t)ui->cb_security->currentIndex();
}

bool WifiConfig::ignore_updates()
{
    bool        ignore(false);
    bool        field_state;
    uint32_t    ap_cfg;
    uint32_t    wifi_cfg;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return false;
    }

    // Get the registers we'll be interested in.
    ap_cfg = my_parent->read_camera_cache(GIGE_REG_WIFI_AP_CFG);
    wifi_cfg = qToBigEndian(my_parent->read_camera_cache(GIGE_REG_WIFI_IP_CUR_CFG));

    // Check to see if ADhoc has changed.
    if ((ui->chk_adhoc->isChecked() && ((ap_cfg & GIGE_WIFI_AP_CFG_ADHOC) == GIGE_WIFI_AP_CFG_ADHOC)) ||
        (! ui->chk_adhoc->isChecked() && ((ap_cfg & GIGE_WIFI_AP_CFG_ADHOC) != GIGE_WIFI_AP_CFG_ADHOC)))
    {
        field_state = false;
    }
    else
    {
        field_state = true;
    }
    ignore = ignore || field_state;

    // Check to see if DHCP/Static has changed.
    if ((ui->rb_dhcp->isChecked() && ((wifi_cfg & GIGE_IFCONFIG_DHCP_BE) == GIGE_IFCONFIG_DHCP_BE)) ||
        (ui->rb_static->isChecked() && ((wifi_cfg & GIGE_IFCONFIG_PERSISTENT_BE) == GIGE_IFCONFIG_PERSISTENT_BE)))
    {
        field_state = false;
    }
    else
    {
        field_state = true;
    }
    ignore = ignore || field_state;

    // Has the security setting changed?
    ignore = ignore || (ui->cb_security->currentIndex() != (int)(ap_cfg & GIGE_WIFI_AP_CFG_SEC_MASK));

    // Has the SSID changed?
    ignore = ignore || (user_ssid != ui->eb_ssid->text());

    // Ignore updates if the user has modified the IP address, gateway or netmask.
    ignore = ignore || (user_ip != ui->ip_addr->get_addr());
    ignore = ignore || (user_gw != ui->gateway->get_addr());
    ignore = ignore || (user_net_mask != ui->net_mask->get_addr());

    // Return our flag.
    return ignore;
}

/*!
 * @brief Initialize the dialog's fields to reflect the information in
 *        the camera.
 */
void WifiConfig::init_dialog()
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Parent is nullptr!";
        ui->control_frame->setEnabled(false);
        ui->pb_done->setEnabled(false);
        ui->pb_scan->setEnabled(false);
        return;
    }

    // Do we already have an SSID?
    if (user_ssid.isEmpty())
    {
        uint32_t    wifi_cfg;
        uint32_t    ap_cfg;
        QByteArray  *ssid_array;

        // Update our SSID. Make sure we release the buffer when we're done.
        ssid_array = my_parent->read_camera_buffer(GIGE_REG_WIFI_SSID);
        set_ssid(ssid_array);
        delete ssid_array;

        // Initialize our display.
        set_ip(my_parent->read_camera_cache(GIGE_REG_WIFI_IP_ADDR));
        set_netmask(my_parent->read_camera_cache(GIGE_REG_WIFI_SUBNET_MASK));
        set_gateway(my_parent->read_camera_cache(GIGE_REG_WIFI_DEF_GATEWAY));

        // Get our configurations.
        wifi_cfg = qToBigEndian(my_parent->read_camera_cache(GIGE_REG_WIFI_IP_CUR_CFG));
        ap_cfg = my_parent->read_camera_cache(GIGE_REG_WIFI_AP_CFG);

        // Figure out which security protocol we're using.
        ui->cb_security->setCurrentIndex(ap_cfg & GIGE_WIFI_AP_CFG_SEC_MASK);

        // Are we adhoc?
        ui->chk_adhoc->setChecked((ap_cfg & GIGE_WIFI_AP_CFG_ADHOC) == GIGE_WIFI_AP_CFG_ADHOC);

        // Figure out if this is a STATIC or DHCP address.
        if ((wifi_cfg & GIGE_IFCONFIG_DHCP_BE) == GIGE_IFCONFIG_DHCP_BE)
        {
            ui->rb_dhcp->setChecked(true);
            update_ip_controls(true);
        }
        else
        {
            ui->rb_static->setChecked(true);
            update_ip_controls(false);
        }
    }
}

/*!
 * @brief Set the IP address display of the dialog.
 *
 * @param ip_addr   The IP to initialize the dialog with.
 */
void WifiConfig::set_ip(uint32_t ip_addr)
{
    user_ip.setAddress(ip_addr);
    ui->ip_addr->set_addr(user_ip);
}

/*!
 * @brief Set the gateway display of the dialog.
 *
 * @param gateway_addr  The gateway to set the dialog to.
 */
void WifiConfig::set_gateway(uint32_t gateway_addr)
{
    user_gw.setAddress(gateway_addr);
    ui->gateway->set_addr(user_gw);
}

/*!
 * @brief Set the subnet mask display of the dialog.
 *
 * @param net_mask  The subnet to set the dialog to.
 */
void WifiConfig::set_netmask(uint32_t net_mask)
{
    user_net_mask.setAddress(net_mask);
    ui->net_mask->set_addr(user_net_mask);
}

/*!
 * @brief Set the SSID edit control.
 * @param ssid_array    Pointer to the SSID name array.
 */
void WifiConfig::set_ssid(QByteArray *ssid_array)
{
    // Reset our string.
    user_ssid.clear();

    // Check for a null pointer.
    if (ssid_array != nullptr)
        user_ssid.append(*ssid_array);

    // Pass the text on to the control.
    ui->eb_ssid->setText(user_ssid);
}

void WifiConfig::update_ip_controls(bool dhcp_enabled)
{
    ui->lbl_address->setEnabled(!dhcp_enabled);
    ui->ip_addr->setEnabled(!dhcp_enabled);
    ui->lbl_netmask->setEnabled(!dhcp_enabled);
    ui->net_mask->setEnabled(!dhcp_enabled);
    ui->lbl_gateway->setEnabled(!dhcp_enabled);
    ui->gateway->setEnabled(!dhcp_enabled);
}

/*!
 * @brief Return true if the user wants to use DHCP.
 */
bool WifiConfig::use_dhcp()
{
    return ui->rb_dhcp->isChecked();
}


//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief User pressed the DHCP radio button.
 *
 * If checked, gray out the network setting controls, otherwise enable them.
 *
 * @param checked   True if the DHCP radio button is checked.
 */
void WifiConfig::dhcp_toggled(bool checked)
{
    // Update the controls.
    update_ip_controls(checked);
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param reg_addr      The address of the register which was read.
 */
void WifiConfig::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_WIFI_AP_CFG:
    case GIGE_REG_WIFI_IP_CUR_CFG:
    case GIGE_REG_WIFI_IP_ADDR:
    case GIGE_REG_WIFI_DEF_GATEWAY:
    case GIGE_REG_WIFI_SUBNET_MASK:
    case GIGE_REG_WIFI_SSID:
    case GIGE_REG_WIFI_PASSPHRASE:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Timed out on reg_addr:" << Qt::showbase << Qt::hex << qToBigEndian(reg_addr) << "read:" << read_op;

        // Increment our retry counter.
        ++write_retries;

        // Make sure our register is still in the list.
        if (!write_list.isEmpty() && (write_retries < WIFI_CONFIG::MAX_WRITE_RETRIES))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Retrying:" << Qt::showbase << Qt::hex << write_list.at(0);

            // Continue the process.
            my_parent->write_camera_register(write_list.at(0).first, write_list.at(0).second);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Retries exceeded! Unable to write:" << Qt::showbase << Qt::hex << write_list.at(0);

            // Make sure the cursor is restored.
            QApplication::restoreOverrideCursor();
            ui->control_frame->setEnabled(true);
            ui->pb_done->setEnabled(true);
            ui->pb_scan->setEnabled(true);
        }
        break;
    default:
        break;
    }
}

/*!
 * @brief The user pressed the "connect" button.
 *
 * Sanity check all the edit fields, make sure the entries are valid.
 * Send the new values to the camera and try to connect to the AP.
 */
void WifiConfig::on_connect()
{
    // Save the SSID.
    user_ssid = ui->eb_ssid->text();

    // Save the passphrase.
    user_pw = ui->eb_password->text();

    // We only care about the edit controls if the user didn't select DHCP.
    if (ui->rb_dhcp->isChecked())
        goto pass_it_on;

    // Retrieve the IP address.
    user_ip = ui->ip_addr->get_addr();

    // Retrieve the network mask.
    user_net_mask = ui->net_mask->get_addr();

    // Retrieve the gateway address.
    user_gw = ui->gateway->get_addr();

pass_it_on:

    // Connect to the SSID specified.
    connect_to_ap();
}

/*!
 * @brief User pressed the scan button, which means we want to bring up a dialog to
 *        list the access points and to allow the user to pick which one to associate with.
 */
void WifiConfig::on_scan()
{
    const wifi_scan_results *selected_ap;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Parent is nullptr!";
        ui->control_frame->setEnabled(false);
        ui->pb_done->setEnabled(false);
        ui->pb_scan->setEnabled(false);
        return;
    }

    // Stop our polling timer.
    status_timer.stop();

    // Setup our dialog.
    WifiScan        dlg(my_parent);
    int             dlg_result;
    dlg.set_ssid(ui->eb_ssid->text());
    dlg_result = dlg.exec();

    // Restart our timer.
    status_timer.start();

    // How did the scan go?
    if (dlg_result != QDialog::Accepted)
        return;

    // Get the selected access point.
    selected_ap = dlg.get_ap_info();
    if (selected_ap == nullptr)
        return;

    // Update our configuration dialog with the selected access point's information.
    QByteArray  sel_ssid(selected_ap->ssid);
    set_ssid(&sel_ssid);
    ui->cb_security->setCurrentIndex(selected_ap->security);
    ui->chk_adhoc->setChecked(selected_ap->adhoc);
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void WifiConfig::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Pick the registers we're interested in.
    switch (reg_addr)
    {
    case GIGE_REG_WIFI_STATUS:
        if (!ignore_updates())
            init_dialog();
        break;
    default:
        break;
    }
}

/*!
 * @brief The security selection has changed.
 * @param new_index     Index to the new selection.
 */
void WifiConfig::security_changed(int new_index)
{
    // Disable the password control if it's not needed.
    if ((new_index == GIGE_WIFI_AP_CFG_SEC_OPEN) || (new_index == GIGE_WIFI_AP_CFG_SEC_UNKNOWN))
        ui->eb_password->setEnabled(false);
    else
        ui->eb_password->setEnabled(true);
}

/*!
 * @brief Called when it's time to poll the WIFI status register.
 */
void WifiConfig::status_update()
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Is there an AP specified?
    if (user_ssid.isEmpty())
    {
        QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
        if( cam_register_group.isNull() )
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
            return;
        }

        // Prepare the camera register group read.
        cam_register_group->addReg(GIGE_REG_WIFI_IP_ADDR);
        cam_register_group->addReg(GIGE_REG_WIFI_SUBNET_MASK);
        cam_register_group->addReg(GIGE_REG_WIFI_DEF_GATEWAY);
        cam_register_group->addReg(GIGE_REG_WIFI_IP_CUR_CFG);
        cam_register_group->addReg(GIGE_REG_WIFI_AP_CFG);
        cam_register_group->lastReg(GIGE_REG_WIFI_STATUS);

        // Start reading those registers.
        my_parent->read_camera_registers(cam_register_group);
    }
    else
    {
        // Only poll the signal strength value.
        my_parent->start_camera_read_reg(GIGE_REG_WIFI_STATUS);
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void WifiConfig::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Parent is nullptr!";
        ui->control_frame->setEnabled(false);
        ui->pb_done->setEnabled(false);
        ui->pb_scan->setEnabled(false);
        return;
    }

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ip_addr:" << Qt::showbase << Qt::hex << ip_addr << "my_ip_addr:" << my_parent->get_ip_addr() << "reg_addr:" << reg_addr;

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Not for us.";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Reset our retry counter.
    write_retries = 0;

    // Process the register.
    if (reg_addr == GIGE_REG_WIFI_SSID)
    {
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // The camera expects to get the null terminator at the end of the string.
            QByteArray pass_array = user_pw.toUtf8();
            pass_array.append('\0');

            // Time to send the password.
            my_parent->write_camera_buffer(GIGE_REG_WIFI_PASSPHRASE, pass_array);
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to write GIGE_REG_WIFI_SSID!";

            // Make sure the cursor is restored.
            QApplication::restoreOverrideCursor();
            ui->control_frame->setEnabled(true);
            ui->pb_done->setEnabled(true);
            ui->pb_scan->setEnabled(true);
        }
    }
    else if (reg_addr == GIGE_REG_WIFI_PASSPHRASE)
    {
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Anything more to write?
            if (!write_list.isEmpty())
            {
                // Start the process.
                my_parent->write_camera_register(write_list.at(0).first, write_list.at(0).second);
            }
            else
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpectedly empty write_list!";

                // Make sure the cursor is restored.
                QApplication::restoreOverrideCursor();
                ui->control_frame->setEnabled(true);
                ui->pb_done->setEnabled(true);
                ui->pb_scan->setEnabled(true);
            }
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to write GIGE_REG_WIFI_PASSPHRASE!";

            // Make sure the cursor is restored.
            QApplication::restoreOverrideCursor();
            ui->control_frame->setEnabled(true);
            ui->pb_done->setEnabled(true);
            ui->pb_scan->setEnabled(true);
        }
    }
    else if (!write_list.isEmpty() && (reg_addr == write_list.at(0).first))
    {
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Pop this register off the stack...it's done.
            (void)write_list.takeFirst();

            // Anything more to write?
            if (!write_list.isEmpty())
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Success! Now send:" << Qt::showbase << Qt::hex << write_list.at(0);

                // Continue the process.
                my_parent->write_camera_register(write_list.at(0).first, write_list.at(0).second);
            }
            else
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__  << "All done!";

                // We're done! Make sure the cursor is restored.
                QApplication::restoreOverrideCursor();
                ui->control_frame->setEnabled(true);
                ui->pb_done->setEnabled(true);
                ui->pb_scan->setEnabled(true);

                // Close our dialog.
                accept();
            }
        }
    }
}
