/*!
 * @file wifi_config.h
 * @brief Header for the WIFI configuration dialog.
 *
 * It takes the new settings and applies them to the camera. Allowing changes to
 * IP address, etc.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

#ifndef WIFI_CONFIG_H
#define WIFI_CONFIG_H

#include <QDialog>
#include <stdint.h>

#include <QtEndian>
#include <QtNetwork>

#include "panel_widget.h"

namespace Ui
{
class WifiConfig;
}

//--------------------------------------------------------------------
// Forward Declarations
//--------------------------------------------------------------------
class QLineEdit;

class WifiConfig : public QDialog
{
    Q_OBJECT

public:
    explicit WifiConfig(QWidget *parent = nullptr);
    ~WifiConfig();

    bool            adhoc();
    bool            dhcp_is_enabled();

    QHostAddress&    get_ip()             { return user_ip; }
    QHostAddress&    get_gw()             { return user_gw; }
    QHostAddress&    get_net_mask()       { return user_net_mask; }
    QString&         get_passphrase()     { return user_pw; }
    QString&         get_ssid()           { return user_ssid; }
    uint32_t        get_security();

    bool            use_dhcp();

private slots:
    void    dhcp_toggled(bool checked);
    void    msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op);
    void    on_connect();
    void    on_scan();
    void    read_complete(uint32_t ip_addr, uint32_t reg_addr);
    void    security_changed(int new_index);
    void    status_update();
    void    write_complete(uint32_t ip_addr, uint32_t reg_addr);

private:
    void    connect_to_ap();
    bool    ignore_updates();
    void    init_dialog();
    void    set_ip(uint32_t ip_addr);
    void    set_gateway(uint32_t gateway_addr);
    void    set_netmask(uint32_t net_mask);
    void    set_ssid(QByteArray *ssid_array);
    void    update_ip_controls(bool dhcp_enabled);

private:
    Ui::WifiConfig      *ui;
    QTimer              status_timer;           //!< Timer to poll the WIFI status register.
    QHostAddress        user_ip;                //!< The user supplied IP address if "static IP" is selected.
    QHostAddress        user_gw;                //!< The user supplied gateway address if "static IP" is selected.
    QHostAddress        user_net_mask;          //!< The user supplied network mask if "static IP" is selected.
    QString             user_pw;                //!< The password to connect to the camera, if required by the security type selected.
    QString             user_ssid;              //!< The SSID selected by the user.

    FM_PANEL::REG_VALUE_LIST    write_list;     //!< List of register and value pairs to write.
    int write_retries;                          //!< Retry counter if a write times out.
};

#endif // WIFI_CONFIG_H
