/*!
 * @file wifi_scan.cpp

 * @brief Implementation of the Wifi scanning dialog.
 *
 * This dialog requests the camera to initiate a WIFI scan, and
 * then displays the results for the user to select.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------

#include <QMessageBox>

#include "wifi_scan.h"
#include "ui_wifi_scan.h"

#include "camera_mdi.h"
#include "fi_comm.h"
#include "gige_msg.h"
#include "logger.h"
#include "mainwindow.h"

// Use a namespace to avoid polluting the global space.
namespace WIFI_SCAN
{
//--------------------------------------------------------------------
// Contant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Give the camera a chunk of time to perform the scan operation.
const int       SCAN_WAKE_INTERVAL  = 5000;
}

/*!
 * @brief Constructor for the WifiScan.
 */
WifiScan::WifiScan(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::WifiScan)
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent);

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Initialize the UI elements.
    ui->setupUi(this);

    // Initialize our members.
    cur_ssid.clear();
    scan_list.clear();

    // Connect our signals/slots.
    connect(ui->pb_cancel, &QPushButton::clicked, this, &WifiScan::on_cancel);
    connect(ui->pb_OK, &QPushButton::clicked, this, &WifiScan::on_ok);
    connect(ui->pb_scan, &QPushButton::clicked, this, &WifiScan::on_scan);
    connect(ui->ap_list, &QListWidget::doubleClicked, this, &WifiScan::ap_selected);

    connect(comm_ptr, &FI_comm::read_reg_complete, this, &WifiScan::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &WifiScan::write_complete, Qt::QueuedConnection);

    // If the parent is closing, we need to bail.
    connect(my_parent, &CameraMDI::closing, this, &WifiScan::on_cancel);
}

/*!
 * @brief Destructor for the WifiConfig.
 */
WifiScan::~WifiScan()
{
    delete ui;
}

/*!
 * @brief Return the pointer to the selected access point information.
 */
const wifi_scan_results *WifiScan::get_ap_info(void) const
{
    int     index;

    // Which entry did the user select?
    index = ui->ap_list->currentRow();
    if (index < 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Negative selected row!";
        return nullptr;
    }
    if (index >= scan_list.count())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Selected row too large!";
        return nullptr;
    }

    // Return the selected information.
    return &scan_list.at(index);
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user double clicked on the list box, thus selecting an AP.
 * @param model_index       [UNUSED]
 */
void WifiScan::ap_selected(QModelIndex model_index)
{
    Q_UNUSED(model_index)

    // Pass it on as an OK selection.
    on_ok();
}

int WifiScan::exec()
{
    // Start a scan.
    on_scan();

    // Pass it on.
    return QDialog::exec();
}

/*!
 * @brief The user pressed the cancel button.
 */
void WifiScan::on_cancel(void)
{
    // Make sure the cursor is restored.
    QApplication::restoreOverrideCursor();

    // Pass it on.
    QDialog::reject();
}

/*!
 * @brief The user pressed the OK button.
 *
 * Sanity check all the edit fields, make sure the entries are valid.
 */
void WifiScan::on_ok(void)
{
    // Make sure the cursor is restored.
    QApplication::restoreOverrideCursor();

    // Pass it on.
    QDialog::accept();
}

/*!
 * @brief User pressed the scan button, which means we want to bring up a dialog to
 *        list the access points and to allow the user to pick which one to associate with.
 */
void WifiScan::on_scan(void)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Have the camera initiate a scan for us.
    my_parent->write_camera_register(GIGE_REG_WIFI_START_SCAN, 1);
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void WifiScan::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    int         i;
    QByteArray  *results_buf;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    const wifi_scan_results *scan_info;

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // Only interested in scan results.
    if (reg_addr != GIGE_REG_WIFI_SCAN_RESULTS)
        return;

    // Clear our list.
    ui->ap_list->clear();
    scan_list.clear();

    // Get a pointer to our results buffer.
    results_buf = my_parent->read_camera_buffer(GIGE_REG_WIFI_SCAN_RESULTS);
    if ((results_buf == nullptr) || results_buf->isEmpty())
    {
        // Show that there are no results from the scan.
        ui->ap_list->addItem(tr("[no results]"));

        // If we got an empty buffer, delete it.
        if (results_buf != nullptr)
            delete results_buf;

        // All done.
        return;
    }

    // Get our byte array into a more useable form.
    scan_info = (const wifi_scan_results *)results_buf->constData();
    for (i = 0; i < (int)(GIGE_WIFI_SCAN_RESULTS_READ_SIZE / sizeof(wifi_scan_results)); ++i)
    {
        QString     sec_str;
        QString     adhoc_str;
        QString     ssid_active;
        QString     item_str;

        // Add this entry to our list.
        scan_list.append(*scan_info);

        // Figure out what security is being used.
        switch (scan_info->security)
        {
        case GIGE_WIFI_AP_CFG_SEC_OPEN:
            sec_str = tr("Open");
            break;
        case GIGE_WIFI_AP_CFG_SEC_WEP:
            sec_str = tr("WEP");
            break;
        case GIGE_WIFI_AP_CFG_SEC_WPA:
            sec_str = tr("WPA-PSK");
            break;
        default:
            sec_str = tr("Unknown");
            break;
        }

        // Is this an ADHoc access point?
        if (scan_info->adhoc)
            adhoc_str = tr("(ad-hoc)");
        else
            adhoc_str.clear();

        // Is this the access point we're currently using?
        if (cur_ssid == QString::fromUtf8(scan_info->ssid))
            ssid_active = QString::fromUtf8("*");
        else
            ssid_active.clear();

        // Build up our item string.
        item_str = QString::fromUtf8("%1%2 (%3%) %4 %5").arg(ssid_active).arg(QString::fromUtf8(scan_info->ssid))
                   .arg(scan_info->signal_level).arg(sec_str).arg(adhoc_str);
        ui->ap_list->addItem(item_str);

        // Bump to the next slot.
        ++scan_info;

        // Is there an entry here?
        if (scan_info->ssid[0] == '\0')
        {
            // We're done.
            break;
        }
    }

    // Sanity check.
    if (scan_list.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpectedly no results!";

        // Show that there are no results from the scan.
        ui->ap_list->addItem(tr("[no results]"));
    }

    // All done with the buffer.
    delete results_buf;
}

/*!
 * @brief Our scan timer has expired.
 */
void WifiScan::scan_timeout(void)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null parent!";
        return;
    }

    // Make sure the cursor is restored.
    QApplication::restoreOverrideCursor();
    ui->pb_OK->setEnabled(true);
    ui->pb_scan->setEnabled(true);

    // Start reading the results.
    my_parent->start_camera_read_buf(GIGE_REG_WIFI_SCAN_RESULTS, GIGE_WIFI_SCAN_RESULTS_READ_SIZE);
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void WifiScan::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Parent is nullptr!";
        setEnabled(false);
        return;
    }

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ip_addr:" << Qt::showbase << Qt::hex << ip_addr << "my_ip_addr:" << my_parent->get_ip_addr() << "reg_addr:" << reg_addr;

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Not for us.";
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Process the register.
    if (reg_addr == GIGE_REG_WIFI_START_SCAN)
    {
        // Was it a successful write?
        uint32_t write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Start up our timer.
            QTimer::singleShot(WIFI_SCAN::SCAN_WAKE_INTERVAL, this, &WifiScan::scan_timeout);

            // Let the user know that we're busy.
            QApplication::setOverrideCursor(Qt::BusyCursor);
            ui->pb_OK->setEnabled(false);
            ui->pb_scan->setEnabled(false);
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Error writing GIGE_REG_WIFI_START_SCAN!";

            // Put up an error message for the user.
            QMessageBox::critical(MainWindow::GetInstance().get(), tr("WIFI Scanner"), tr("Unable to perform a WIFI scan at this time."));
        }
    }
}
