/*!
 * @file camera_explore.cpp
 * @brief This provides the "explore" functionality to show the files stored
 *        on the camera's media and allow the user to copy, move, delete and
 *        otherwise access and manipulate those images/videos.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "camera_explore.h"
#include "ui_camera_explore.h"
#include "gige_msg.h"

#include <errno.h>
#include <QApplication>
#include <QDomDocument>
#include <QDomElement>
#include <QFileDialog>
#include <QFuture>
#include <QFutureWatcher>
#include <QHostAddress>
#include <QJsonObject>
#include <QMenu>
#include <QMessageBox>
#include <QProcess>
#include <QStandardItem>
#include <QString>
#include <QStringList>
#include <QStringView>
#include <QtConcurrent>
#include <QTemporaryFile>
#include <QTimer>

#include <fas_video.h>

#include "camera_mdi.h"
#include "camera_stream.h"
#include "camera_video.h"
#include "client_math.h"
#include "fi_comm.h"
#include "fi_file.h"
#include "fi_loadcap.h"
#include "fi_msgbox.h"
#include "fi_progress_dialog.h"
#include "fi_seldestdlg.h"
#include "format_dlg.h"
#include "logger.h"
#include "mainwindow.h"
#include "save_dialog.h"
#include "save_video.h"

namespace CAM_EXPLORE
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

//! Provide a way to turn off the batch download log messages.
//! Set this to 1 if you want to turn on these trace log messages.
#define LOG_BATCH_DOWNLOAD_TRACES  0

//! Number of milliseconds before the camera complains about not hearing from us.
const int COPY_HEARTBEAT_TIMEOUT        = 30000;

//! How often should a copy worker thread update progress on single file copy operations?
//! Right now I have this set at 1% to get updates at that frequency. It can be bumped up here.
const uint32_t COPY_WORKER_PCT_SIGNAL_FREQUENCY = 1;

//! Our polling period when waiting for a file action to complete.
const int FA_POLLING_PERIOD             = 250;

//! Our polling period when we are formatting a device.
const int FORMAT_POLLING_PERIOD         = 250;

//! The polling period for reading registers.
const int REGISTER_POLLING_PERIOD       = 2000;

//! Amount of time to wait for the camera to enter live video mode.
//! This value indicates how long will we allow for the camera state to
//! transfer to live video before we consider too much of a delay and
//! consider the load process a failure.
const int LIVE_DELAY                    = 20000;

//! Number of times to wait for the switch to the live video state.
const int LIVE_STATE_DELAY              = 6;

//! Amount of time to wait for the camera to enter review mode.
//! This value indicates how long will we allow for the camera state to
//! transfer to live video before we consider too much of a delay and
//! consider the load process a failure.
const int REVIEW_DELAY                  = 20000;

//! Number of times to wait for the switch to the review state.
const int REVIEW_STATE_DELAY            = 6;

//! Define the various targets for our copy/move operations.
enum
{
    DEST_SDCARD,                //!< Targeting the SD card.
    DEST_SSD,                   //!< Targeting the SSD.
    DEST_USB,                   //!< Targeting the USB drive.
    DEST_PC_PATH,               //!< Targeting a path on the local workstation.
    DEST_NONE                   //!< No destination specified.
};

//! States for our various state machines.
enum
{
    CAP_XFER_STARTUP,
    CAP_XFER_WAIT_FOR_LIVE,
    CAP_XFER_WAIT_FOR_REVIEW,
    CAP_XFER_LOAD,
    CAP_XFER_SAVE,
    CAP_XFER_COMPLETE,

    CAP_LOAD,

    COPY_STARTUP,
    COPY_SET_FILE,
    COPY_METAFILE,
    COPY_PAUSED,
    COPY_WAIT,
    COPY_COMPLETE,
    COPY_DIRECT,
    COPY_FAULT,

    DELETE_STARTUP,
    DELETE_SET_FILE,
    DELETE_WAIT,
    DELETE_COMPLETE,

    REFRESH_STARTUP,
    REFRESH_STILLS,
    REFRESH_STACKS,
    REFRESH_VIDEO,
    REFRESH_CAPTURE,
    REFRESH_DOWNLOAD_COMPLETE,

    EXPLORE_IDLE
};

//! Define a custom event to indicate when to transition to a new CAP transfer state.
const QEvent::Type CAP_XFER_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_CAP_XFER_EVENT);

//! Define a custom event to indicate when to transition to a new copy state.
const QEvent::Type COPY_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_COPY_EVENT);

//! Define a custom event to indicate when to transition to a new delete state.
const QEvent::Type DELETE_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_DELETE_EVENT);

//! Define a custom event to indicate when to transition to a new refresh state.
const QEvent::Type REFRESH_EVENT = static_cast<QEvent::Type>(QEvent::User + FASMOTION_SETTINGS::FI_REFRESH_EVENT);

//! Subclass QEvent for our custom events.
class ExploreEvent : public QEvent
{
public:
    ExploreEvent(QEvent::Type event_type, int new_state, uint32_t arg1 = 0) :
        QEvent(event_type)
    {
        state = new_state;
        this->arg1 = arg1;
    }

    uint32_t get_arg1() const   { return arg1; }
    int get_state() const       { return state; }

private:
    uint32_t    arg1;                   //!< Optional argument.
    int         state;                  //!< The state we are transitioning to.
};

}   // End of namespace CAM_EXPLORE.

using namespace CAM_EXPLORE;

// Initialize our static const members.
const QString CamExplore::STACKS_DIR_LISTING("/camera_stacks.xml");
const QString CamExplore::VIDEOS_DIR_LISTING("/camera_videos.xml");
const QString CamExplore::CAPTURES_DIR_LISTING("/camera_captures.xml");
const QString CamExplore::CAPTURES_BASE_PATH("/captures/");
const QString CamExplore::CAMERA_EXPLORE_GROUP_NAME("CameraExplore");
const QString CamExplore::LAST_DEVICE("LastDevice");


//! If a file is split it will have this suffix, and we want to ignore such files
//! for the purposes of displaying them to the user.
const QString CamExplore::IGNORE_FILENAME("_1.");

//! A constant for the source label of our progress dialog for copying.
const QString CamExplore::COPY_SRC_RECEIVE_FRAME = (CamExplore::tr("Received:"));

//! The FILE_INFO is incomplete for parsed file.
const QString CamExplore::FILE_INFO_INCOMPLETE = QStringLiteral("1");

//! If FILE_INFO.incomplete flag is set this is the string to use.
const QString CamExplore::FILE_INFO_INCOMPLETE_STR(CamExplore::tr("Incomplete"));

//! If FILE_INFO.incomplete flag is clear this is the string to use.
const QString CamExplore::FILE_INFO_COMPLETE_STR(CamExplore::tr("Complete"));

//! Reusable string for dialog title asking user to select destination path.
const QString CamExplore::FILE_DIRECTORY_SELECT_STR(CamExplore::tr("Select A Directory To Save Files To"));

// These static const strings are used for paths, expecially with the series3 cameras.
// There is no need from them to be translated.
const QString CamExplore::DEST_SDCARD_STR("sdcard");
const QString CamExplore::DEST_SSD_STR("ssd");
const QString CamExplore::DEST_USB_STR("usb");

// Define some common labels to be used whenever you need one. As the user
// will see these, we want them to be translatable.
const QString CamExplore::SDCARD_LABEL(CamExplore::tr("SDCard"));
const QString CamExplore::SSD_LABEL(CamExplore::tr("SSD"));
const QString CamExplore::USB_LABEL(CamExplore::tr("USB Flash Drive"));

/*!
 * @brief Constructor for the CamExplore class. Initialize variables.
 */
CamExplore::CamExplore(QWidget *parent) :
    QDialog(parent, Qt::WindowTitleHint | Qt::WindowSystemMenuHint | Qt::WindowCloseButtonHint),
    ui(new Ui::CamExplore)
{
    QList<int>  size_hints;
    FiTreeWidgetItem *tree_item;

    ui->setupUi(this);
    ui->cam_tree->header()->setSectionResizeMode(QHeaderView::Interactive /* Stretch */);
    ui->frm_options->hide();

    // Adjust the proportions of the splitter window.
    size_hints.append(200);
    size_hints.append(100);
    ui->splitter->setSizes(size_hints);

    // Initialize our variables.
    camera_ptr = nullptr;
    cam_ip_addr = 0;
    cap_copy_in_progress = false;
    cap_file_load = false;
    cap_load_success = false;
    captures_allowed = false;
    copy_destination = DEST_NONE;
    cur_state = EXPLORE_IDLE;
    file_timer_id = 0;
    force_update = false;
    format_cmd_status = GEV_STATUS_SUCCESS;
    format_in_progress = false;
    live_timer_id = 0;
    review_timer_id = 0;
    last_device.clear();
    listing_len = 0;
    media_type = GIGE_MEDIA_SDCARD;
    we_own_lock = false;
    cancel_in_progress = false;
    cleanup_in_progress = false;
    move_in_progress = false;
    wait_for_state_count = 0;

    // Load up our icons.
    capture_icon.addPixmap(QPixmap(QString::fromUtf8(":/images/document.png")));
    folder_icon.addPixmap(QPixmap(QString::fromUtf8(":/images/folder.png")));
    still_icon.addPixmap(QPixmap(QString::fromUtf8(":/images/still.png")));
    stack_icon.addPixmap(QPixmap(QString::fromUtf8(":/images/stack.png")));
    video_icon.addPixmap(QPixmap(QString::fromUtf8(":/images/video.png")));

    // Populate the tree view with the highest level information. These are the
    // "root" nodes of our tree:
    //  stills - Still images
    //  stacks - Directories which are stacks of images
    //  videos - AVI and MP4 files
    //  capture- Only on A3 and later cameras. We add this node in the set_captures_allowed()
    //           method, once we are told the camera supports CAP files.
    //
    // It's important that we maintain this order so that it agrees with our enumeration above.
    tree_item = new FiTreeWidgetItem(ui->cam_tree);
    tree_item->setText(COLUMN_TITLE, tr("Stills"));
    tree_item->setIcon(COLUMN_TITLE, folder_icon);
    root_items.append(tree_item);

    tree_item = new FiTreeWidgetItem(ui->cam_tree);
    tree_item->setText(COLUMN_TITLE, tr("Stacks"));
    tree_item->setIcon(COLUMN_TITLE, folder_icon);
    root_items.append(tree_item);

    tree_item = new FiTreeWidgetItem(ui->cam_tree);
    tree_item->setText(COLUMN_TITLE, tr("Videos"));
    tree_item->setIcon(COLUMN_TITLE, folder_icon);
    root_items.append(tree_item);

    // Populate our camera folder database with the initial values.
    cam_folders.append(new LISTING_INFO);       // Stills
    cam_folders.append(new LISTING_INFO);       // Stacks
    cam_folders.append(new LISTING_INFO);       // Videos

    // Initialize our download path. For now, just use the default path.
    download_base_path = SaveDialog::loadPath(0);

    // Disallow device operations until we know what our devices are.
    ui->cb_devices->setEnabled(false);
    ui->pb_format->setEnabled(false);
    ui->tb_eject->setEnabled(false);
    ui->lbl_space->clear();

    // Hide the format progress indicator and show the button.
    ui->pb_format->setVisible(true);
    ui->format_progress->setVisible(false);

    // Connect up our signals/slots.
    connect(ui->pb_close, &QPushButton::clicked, this, &CamExplore::reject);
    connect(ui->pb_refresh, &QPushButton::clicked, this, &CamExplore::doRefresh);
    connect(ui->pb_load, &QPushButton::clicked, this, &CamExplore::doLoad);
    connect(ui->pb_copy, &QPushButton::clicked, this, &CamExplore::doCopy);
    connect(ui->pb_move, &QPushButton::clicked, this, &CamExplore::doMove);
    connect(ui->pb_delete, &QPushButton::clicked, this, &CamExplore::doDelete);
    connect(ui->cam_tree, &QTreeWidget::expanded, this, &CamExplore::doTreeExpansion);
    connect(ui->cam_tree, &QTreeWidget::currentItemChanged, this, &CamExplore::doTreeItemChanged);
    connect(ui->cam_tree, &QTreeWidget::itemSelectionChanged, this, &CamExplore::doTreeSelectionChanged);
    connect(ui->cam_tree, &QTreeWidget::itemDoubleClicked, this, &CamExplore::doTreeItemDoubleClicked);
    connect(ui->grp_options, &QGroupBox::clicked, this, &CamExplore::doOptions);
    connect(ui->cb_devices, qOverload<int>(&QComboBox::currentIndexChanged), this, &CamExplore::doDeviceSelected);
    connect(ui->pb_format, &QPushButton::clicked, this, &CamExplore::doFormat);
    connect(ui->tb_eject, &QToolButton::released, this, &CamExplore::doEject);
}

/*!
 * @brief Destructor for our class. Make sure we clean up after ourselves.
 */
CamExplore::~CamExplore()
{
    // Shutdown the timers.
    if (file_timer_id != 0)
    {
        killTimer(file_timer_id);
        file_timer_id = 0;
    }
    if (live_timer_id != 0)
    {
        killTimer(live_timer_id);
        live_timer_id = 0;
    }
    if (review_timer_id != 0)
    {
        killTimer(review_timer_id);
        review_timer_id = 0;
    }

    // Call the recycle interface as we are destructing.
    if (progressDialog)
    {
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Disconnect any selection changed or item changed connections.  We are in the process of removing the
    // tree items and don't want any of these signal connections during the cleanup and object destruction.
    disconnect(ui->cam_tree, &QTreeWidget::itemSelectionChanged, this, &CamExplore::doTreeSelectionChanged);
    disconnect(ui->cam_tree, &QTreeWidget::currentItemChanged, this, &CamExplore::doTreeItemChanged);

    // Clean our folder display.
    while (!cam_folders.isEmpty())
    {
        LISTING_INFO *list_info;

        list_info = cam_folders.takeFirst();
        list_info->clear();
        delete list_info;
    }

    while (!root_items.isEmpty())
    {
        QTreeWidgetItem *tree_item;
        QList<QTreeWidgetItem *> children;

        // Get the root item.
        tree_item = root_items.takeFirst();

        // Get rid of all the children of the root item.
        children = tree_item->takeChildren();
        children.clear();

        delete tree_item;
    }

    // Do we have a worker thread allocated?
    if (copy_worker_thread.get() != nullptr)
    {
        // If the worker thread is still running, we want to cancel it.
        if (copy_worker_thread->isRunning())
        {
            // Cancel the worker thread and wait for it.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpectedly found copy_worker_thread is running!";
            copy_worker_thread->cancel();
            copy_worker_thread->waitForFinished();
        }

        // Now we can release the object.
        copy_worker_thread.reset();
    }

    // Clean up the dynamically allocated elements.
    delete ui;
}

/*!
 * @brief Static method to generate the metadata for the currently selected entry based on
 *        the file_info structure passed in.
 *
 * @param file_info     The file information we are going to display.
 *
 * @returns The formatted metadata for displaying.
 */
QString CamExplore::generateMetadataString(const CAM_EXPLORE::FILE_INFO *file_info)
{
    QString meta_str;

    // Sanity check.
    if (file_info == nullptr)
    {
        // Nothing to display.
        meta_str.clear();
    }
    else
    {
        // Everything displays the name.
        meta_str = QString(tr("name: %1\n")).arg(file_info->name);

        // If the file is incomplete process the main components differently.
        if (file_info->incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
        {
            meta_str += QString(tr("status: %1\n")).arg(FILE_INFO_INCOMPLETE_STR);
        }
        else
        {
            meta_str += QString(tr("status: %1\n")).arg(FILE_INFO_COMPLETE_STR);
        }

        // Only display these if they are not empty.
        if (!file_info->time_stamp.isEmpty())
            meta_str += QString(tr("time stamp: %1\n")).arg(file_info->time_stamp);
        if (!file_info->width.isEmpty())
            meta_str += QString(tr("width: %1\n")).arg(file_info->width);
        if (!file_info->height.isEmpty())
            meta_str += QString(tr("height: %1\n")).arg(file_info->height);
        if (!file_info->frame_count.isEmpty())
            meta_str += QString(tr("frame count: %1\n")).arg(file_info->frame_count);
        if (!file_info->fps.isEmpty())
            meta_str += QString(tr("fps: %1\n")).arg(file_info->fps);
        if (!file_info->file_ext.isEmpty())
            meta_str += QString(tr("codec: %1\n")).arg(file_info->file_ext);
        if (!file_info->raw.isEmpty())
            meta_str += QString(tr("raw: %1\n")).arg(file_info->raw);
        if (!file_info->comment.isEmpty())
            meta_str += QString(tr("comment: %1\n")).arg(file_info->comment);
    }
    return meta_str;
}

bool CamExplore::loadCapFile(const QString& file_path)
{
    bool loaded_ok = false;
    uint32_t reg_value;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        goto exitLoadCapFile;
    }

    // If you arrive here and there is an existing load capture dialog running, then
    // this is an overlap condition.  This should not ever happen but if it does the
    // process cannot handle two simultaneous load requests.
    if (loadCaptureDlg)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Can't start a load capture for " << file_path << "! Already in process of loading:" << loadCaptureDlg->getFilePath();
        goto exitLoadCapFile;
    }

    // Create our load capture progress dialog.
    loadCaptureDlg = fiLoadCaptureDlg::CreateLoadCaptureDlg(camera_ptr->get_camera_name(), camera_ptr);
    if (!loadCaptureDlg)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null load capture progress dialog!";
        goto exitLoadCapFile;
    }

    // Sanity check. We had better be in LIVE mode.

    reg_value = camera_ptr->read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Special camera state requirements for series3 cameras.
    // TICKET #441:  I believe not making sure a series3 camera may be related to the problem
    // where during batch downloads the IPM pipeline module is unable to service data and the
    // batch download process ends as well as any live images until that series3 camera is rebooted.
    // this is an attempt to ensure we are in that state prior to initiating a load request.
    if (camera_ptr->is_series3())
    {
        // For series3 cameras we must be in live video and we must have the stream paused.
        if ((reg_value != GIGE_CAMERA_STATE_LIVE_VIDEO))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Load Aborted! Series3 camera state should be" << MainWindow::camera_state_to_string(GIGE_CAMERA_STATE_LIVE_VIDEO) << "but is set to" << MainWindow::camera_state_to_string(reg_value) << "!";
            goto exitLoadCapFile;
        }
    }
    else
    {
        // I don't think our seriesNC camera has this issue so this is the original code in place
        // during this load process.
        if ((reg_value != GIGE_CAMERA_STATE_LIVE_VIDEO) && (reg_value != GIGE_CAMERA_STATE_REVIEW))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Load Aborted! The camera state should be" << MainWindow::camera_state_to_string(GIGE_CAMERA_STATE_LIVE_VIDEO)
                         << "or" << MainWindow::camera_state_to_string(GIGE_CAMERA_STATE_REVIEW) << "but is set to" << MainWindow::camera_state_to_string(reg_value) << "!";
            goto exitLoadCapFile;
        }
    }

    // Set our state. Since the load operation happens in the context of the FiLoadCap dialog,
    // this is here to remind us of where we are.
    cur_state = CAP_LOAD;

    // Set the dialog with the information we have.
    loadCaptureDlg->setFilePath(file_path);
    loadCaptureDlg->setXferSrcLabel(tr("Source:"));
    loadCaptureDlg->setXferSrcValue(file_info->name);
    loadCaptureDlg->setStatusValue(tr("Loading partition..."));

    // Are we connected to a series3 camera?
    if (camera_ptr->is_series3())
    {
        // Can't cancel a series3 load operation.
        loadCaptureDlg->showCancel(false);
    }
    else
    {
        // Hook up our cancel signal.
        loadCaptureDlg->showCancel(true);
    }

    loadCaptureDlg->showStatusValue(true);
    loadCaptureDlg->setBarMinMax(0, 100);
    loadCaptureDlg->setBarMin(0);
    loadCaptureDlg->enableCleanup(false);

    // Get the transfer running.
    if (loadCaptureDlg->exec() == QDialog::Rejected)
    {
        // Note that the load was not successfull.
        cap_load_success = false;

        // Remember we're canceling.
        cancel_in_progress = true;

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        capXferFinish();

        // Carry on.
        userCanceled(__PRETTY_FUNCTION__);

        // Resume video if appropriate.
        if (video_was_running)
            camera_ptr->resume_video_request();
    }
    else
    {
        // Success!
        cap_load_success = true;

        // We are done.
        emit loadComplete();

        // The CAP file was loaded successfully without any type of error.
        loaded_ok = true;
    }

    // We are done with loading this capture whether it was successful or not and
    // we need to cleanup resources and the object allocated for loadCaptureDlg.
    loadCaptureDlg.reset();

exitLoadCapFile:
    return loaded_ok;
}

/*!
 * @brief Build up the path to the media base on the media type (SDCard, SSD, USB) and
 *        the image type (stills, stacks, videos or CAPs). The resulting string (media_path)
 *        is where, on the camera, we expect the information to reside.
 *
 * @param root_row      The root row index in our tree structure. This specifies the image type.
 * @return True if media_path was built successfully, false otherwise.
 */
bool CamExplore::buildMediaPath(int root_row)
{
    if (media_info.isEmpty())
    {
        // Build up the path on the camera for the media we're interested in.
        media_path = QString::fromUtf8("/media/");
        if (media_type == GIGE_MEDIA_SDCARD)
            media_path.append(CamExplore::DEST_SDCARD_STR);
        else if (media_type == GIGE_MEDIA_SSD)
            media_path.append(CamExplore::DEST_SSD_STR);
        else if (media_type == GIGE_MEDIA_USB)
            media_path.append(CamExplore::DEST_USB_STR);
    }
    else
    {
        // Pull the media path from the JSON information.
        media_path = media_info["mountdir"].toString();
        if (media_path.isEmpty())
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "device:" << media_info["name"].toString() << "doesn't have a mount path!";
            return false;
        }
    }

    // Finish up the path construction based on which row we're looking at.
    switch (root_row)
    {
    case ROW_STILLS:
    case ROW_STACKS:
        media_path.append(QString::fromUtf8("/DCIM/100fastc"));
        break;
    case ROW_VIDEO:
        media_path.append(QString::fromUtf8("/hs_video"));
        break;
    case ROW_CAPTURE:
        media_path.append(QString::fromUtf8("/captures"));
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected row:" << root_row;
        return false;
    }

    // All good.
    return true;
}

QString CamExplore::buildSmbPath(const CAM_EXPLORE::FILE_INFO *file_info, int row)
{
    QString file_ext;
    QString full_path;
    QString file_id;

    // Get the file extension.
    file_ext = file_info->file_ext.toLower();
    if (file_ext.compare(FiSaver::IMAGE_EXT_CAPSTACK, Qt::CaseInsensitive) == 0)
    {
        // We need to fix the file name extension, and set the file ID to
        // the starting CAP file in the stack.
        file_ext = FiSaver::IMAGE_EXT_CAPTURE;
        file_id = file_info->start_cap;
    }
    else if ((file_ext.compare(FiSaver::IMAGE_EXT_JPG, Qt::CaseInsensitive) == 0) ||
             (file_ext.compare(QString::fromUtf8("jpeg"), Qt::CaseInsensitive) == 0))
    {
        // Set the file ID to the starting file in the stack.
        file_ext = FiSaver::IMAGE_EXT_JPG;
        file_id = file_info->start_frame;
    }
    else if ((file_ext.compare(FiSaver::IMAGE_EXT_TIF, Qt::CaseInsensitive) == 0) ||
             (file_ext.compare(QString::fromUtf8("tiff"), Qt::CaseInsensitive) == 0))
    {
        // Set the file ID to the starting file in the stack.
        file_ext = FiSaver::IMAGE_EXT_TIF;
        file_id = file_info->start_frame;
    }
    else
    {
        file_id = file_info->start_frame;
    }

    // Build the path for where the image should be loaded on the camera. The path
    // willl be built into the 'media_path' value.
    buildMediaPath(row);

    // For videos and CAP files, we want to ignore the file ID.
    if ((row == ROW_VIDEO) || (row == ROW_CAPTURE))
    {
        // Ignore the file ID.
        file_id = QStringLiteral("-1");
    }

    // Prepare the path information for "loading" into a playback window.
    camera_ptr->prepare_smb_filepath(media_path, file_info->name, file_id, file_ext, full_path);

    // Do we have a path to something to open?
    if (!full_path.isEmpty())
    {
        // Had some trouble with the network interface where the first check to see if
        // the file exists would fail. Since the second try always seems to work, I thought
        // I'd pre-check here prior to calling the mainwindow to load.
        {
            QFile load(full_path);
            if (!load.exists())
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "File: '" + full_path + "' shows not existing!";
                if (!load.open(QIODeviceBase::ReadOnly))
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "File: '" + full_path + "' won't open:" << load.errorString();

                    // Reset the path.
                    full_path.clear();
                }
            }
        }
    }

    // Return the path.
    return full_path;
}

/*!
 * @brief Method to setup the transfer of one or more CAP files to the local workstation.
 * @return True if we are doing a CAP -> image file transform, False to download the CAP file.
 */
bool CamExplore::capFileTransfer()
{
    bool showCopyXferDialog = false;
    int         jpegQualityBeforeAskingUser;
    FI_SAVER::STREAM_TYPE streamType = FI_SAVER::STREAM_TYPE::NO_STREAM;
    const FI_COMM::FI_CAMERA_INFO   *camera_info;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return true;
    }

    // User needs to specify how to transfer data in the CAP file.
    // We want to persist this dialog for use during this process for the
    // purposes of retrieving the information associated with transcoding
    // the CAP data.  Modified to persist as unique_ptr for Ticket #258 fix.
    xferCapDialog = std::unique_ptr<SaveDialog>(new SaveDialog(cam_ip_addr, camera_ptr));
    if (!xferCapDialog)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create xferCapDialog!";
        return true;
    }

    // Initialize the save dialog but we don't want to query the camera
    xferCapDialog->init_controls(false);
    xferCapDialog->set_cap_transfer(true);

    // Are there overlays?
    if (camera_ptr && camera_ptr->get_player())
        xferCapDialog->enable_overlay_merge(camera_ptr->get_player()->has_overlay());

    // Remember the current JPEG quality setting.
    jpegQualityBeforeAskingUser = xferCapDialog->get_jpeg_quality();

    // Let the user make selections.
    if (xferCapDialog->exec() == QDialog::Rejected)
    {
        return true;
    }

    // Do we need to update the JPEG quality factor?
    streamType = xferCapDialog->GetStreamType();

    // Make sure we grab the path, just in case this is a CAP file.
    download_base_path = xferCapDialog->get_dest_path();

    // Do we have a path?
    if (!download_base_path.isEmpty())
    {
        // Make sure the path exists.
        QDir download_dir(download_base_path);
        if (!download_dir.exists())
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "Path doesn't exist. Creating:" << download_dir;
            if (!download_dir.mkpath(download_base_path))
            {
                // Can't create the path, so use the Documents folder instead.
                QString msg_str = tr("The download path (%1) does not exist.").arg(download_base_path);
                QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
                QMessageBox::critical(this, tr("Configuration Error"), msg_str);
                return false;
            }
        }
    }

    // If we are transferring the CAP file, then we are done here.
    if (!camera_ptr->is_series3() && (streamType == FI_SAVER::STREAM_TYPE::FI_CAP))
    {
        // Let the caller know that we want to transfer the CAP file.
        // "False" because we do not want to do a translation.
        return false;
    }

    // Save the settings for the transfer. We only need to worry about the
    // settings which aren't already persistent.
    enable_merge = xferCapDialog->merge_overlays();

    // Get the list of items the user has selected.
    sel_items.clear();
    sel_items = ui->cam_tree->selectedItems();
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected empty selection list!";
        return true;
    }

    // Is this a JPEG-type transformation?
    if ((streamType == FI_SAVER::STREAM_TYPE::JPEG_STACK) || ((streamType == FI_SAVER::STREAM_TYPE::AVI) && (xferCapDialog->use_avi_jpeg_codec())))
    {
        int jpeg_quality = (uint32_t) xferCapDialog->get_jpeg_quality();
        FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "save_dlg.get_jpeg_quality()", jpeg_quality);

        // Write out the new quality value if user entered a different jpeg_quality.
        if (jpeg_quality != jpegQualityBeforeAskingUser)
        {
            camera_ptr->write_camera_register(GIGE_REG_JPEG_QUALITY, jpeg_quality);
        }
    }

    // I don't think we ever show the xfer copy dialog here at all it is handled in "do_copy".
    // I had a place here where I was showing it depending on the streamType but it's not
    // used I don't believe...  It's left because I don't know if I can safely strip out
    // the creation of the createXferCopyDialog call.
    showCopyXferDialog = false;

    // Create a file transfer progress dialog.
    if (createXferCopyDialog(streamType, download_base_path))
    {
        if (showCopyXferDialog)
        {
            // Show the dialog without any polling.
            xferCopyDialog->showDialogWithoutPolling();
        }
    }

    // Lock access to the camera.
    we_own_lock = false;
    cur_state = CAP_XFER_STARTUP;

    // Get the camera information.
    camera_info = camera_ptr->get_camera_info();

    // Remember the current "paused" state.
    video_was_running = camera_info->stream && camera_info->stream_running;

    // Catch the "paused" signal.
    connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::streamPaused);

    // Make sure the video stream is paused.
    camera_ptr->pause_video_request();

    // Are we merging overlays?
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "enable_merge:" << enable_merge << "player:" << camera_ptr->get_player();
    if (enable_merge && camera_ptr && camera_ptr->get_player())
    {
        camera_ptr->get_player()->render_start();
    }

    // All is proceeding according to plan.
    return true;
}

/*!
 * @brief Load the user selected CAP file back into camera memory for processing.
 *        The assumption is that the file_info member variable has been set.
 */
void CamExplore::capLoadStart(int sel_row)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "file:" << file_info->name;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Are we series3?
    if (camera_ptr->is_series3())
    {
        // Build up the full path for the file we're loading.
        media_path = CAPTURES_BASE_PATH;
        media_path += file_info->name;
    }
    else
    {
        QString operation_label = tr("Loading a CAP file");
        // TICKET #427 - Don't overload a dirty partition without warning the user about it.
        if (!camera_ptr->verify_dirty_overwrite(true, operation_label))
        {
            // The user wants to preserve dirty recorded partition.
            return;
        }

        // Build the media path based on the selected entry.
        buildMediaPath(sel_row);
        media_path += QString("/") + file_info->name;
    }

    // Start the process of loading the specified CAP file to the camera.
    loadCapFile(media_path);
}

/*!
 * @brief The final CAP transfer operation state. Clean up and leave.
 */
void CamExplore::capXferFinish()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ ;

    // Not in a CAP file copy process anymore.
    cap_copy_in_progress = false;

    // Don't need our xferCapDialog anymore.
    if (xferCapDialog)
    {
        xferCapDialog.reset();
    }

    // Clean up after ourselves.
    cleanupState();
}

/*!
 * @brief Continue the CAP transfer operation and save the recently loaded
 *        image using the information provided in the save_dlg object.
 */
void CamExplore::capXferImageSave()
{
    uint32_t    num_frames;
    uint32_t    clip_end;
    SaveVideo   *saver_ptr;
    FI_SAVER::SAVE_PARAMS save_info;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our saver object.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
        return;
    }

    // We should already have a xferCapDialog object to use for the save.
    if (!xferCapDialog)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Missing xferCapDialog used for saving!  Creating another one.";
        // Leaving the old logic in place but setting up our unique_ptr for cap file
        // transfer operations.  Hopefully there will always be a valid one created in
        // cap_file_transfer().  Modified for Ticket #258 fix.
        xferCapDialog = std::unique_ptr<SaveDialog>(new SaveDialog(cam_ip_addr, camera_ptr));
        if (!xferCapDialog)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create Replacement xferCapDialog!";
            return;
        }

        // Restore the rest of the settings. No need to query the camera, because we're only
        // using the save dialog as a means of transferring information to the save_info structure.
        xferCapDialog->init_controls(false);
        xferCapDialog->set_cap_transfer(true);
    }

    // Are there overlays?
    xferCapDialog->enable_overlay_merge(enable_merge);

    // Get the number of frames in the newly loaded image.
    num_frames = camera_ptr->read_camera_cache(GIGE_REG_PART_NUM_FRAMES);

    // Set the real clip-end to be zero-based frame number range.
    clip_end = num_frames - 1;

    // Set our number of frames to the real number of frames.
    xferCapDialog->set_num_frames(num_frames);

    // Set Trigger frame to zero.  We don't want a bogus settings for clip range.
    xferCapDialog->set_trig_frame(0);

    // Set our "clip" range to the full range of the capture.
    xferCapDialog->set_clip_range(0, clip_end);

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "num_frames read from camera:" << num_frames;

    // Retrieve the save settings.
    xferCapDialog->get_save_params(save_info, camera_ptr->is_series3());

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_info.clip_start:" << save_info.clip_start
                 << "save_info.clip_end:" << save_info.clip_end;

    // Sanity Check:  Make sure that the clip range coming out of the xferCapDialog is full capture.
    if (save_info.clip_start != 0)
    {
        save_info.clip_start = 0;
    }
    if (save_info.clip_end != (int)clip_end)
    {
        save_info.clip_end = (int)clip_end;
    }

    // Have the saver object configure itself.
    if (!saver_ptr->configure(save_info))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to configure the save object.";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        capXferFinish();
        return;
    }

    // Connect the saver's signal so we know when the save completes.
    connect(saver_ptr, &SaveVideo::save_complete, this, &CamExplore::capXferFileSaved, Qt::QueuedConnection);

    // And Connect the saver's signal so we know if the user cancels the save.
    connect(saver_ptr, &SaveVideo::userCancelledSaveSignal, this, &CamExplore::handleSaverProgressCancelledSignal, Qt::QueuedConnection);

    // Get started on the save.
    retry_count = 0;
    capStartLocalSave();
}

void CamExplore::capStartLocalSave()
{
    // Check to see if the partition metadata has successfully been downloaded.
    // The download happens as a result of the camera entering into review state.
    // Sometimes it seems that it takes a long time to get the final block from
    // the camera.
    if (camera_ptr->is_pm_downloading())
    {
        // Increment our retry counter.
        retry_count++;
        if (retry_count > 10)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "There must be a problem getting the pm_metadata.";

            // Make sure we're not moving.
            move_in_progress = false;

            // Close out the transfer.
            capXferFinish();
            return;
        }

        // We need to wait for all of the metadata to be received.
        // Let's try again in a little bit.
        QTimer::singleShot(250, this, &CamExplore::capStartLocalSave);
    }
    else
    {
        // Make certain that the streamer has settled down, finishing it's post review
        // state processing.
        CameraStream *streamer = camera_ptr->get_streamer();
        if (streamer && streamer->is_busy())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Streamer is busy. Try again in a little bit.";
            QTimer::singleShot(250, this, &CamExplore::capStartLocalSave);
        }
        else
        {
            // Have the camera object fire off the save.
            camera_ptr->start_local_save(true);
        }
    }
}

/*!
 * @brief Verifies that the camera state is in live video mode and paused before
 * attempting to load the next CAP file into image memory.
 */
void CamExplore::capXferStartCheck()
{
    bool            put_into_live_video = false;
    QTreeWidgetItem *selected_item;

    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get the next CAP filename to transfer and put it into file_info.
    // Make sure there's still something to transfer.
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "sel_items list is empty?!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        capXferFinish();
        return;
    }

    // Grab the first item off the list.
    selected_item = sel_items.takeFirst();

    // Get the information structure associated with the item. The "sel_row"
    // parameter will get set by the get_file_info() call.
    file_info = getFileInfo(selected_item, cap_load_sel_row);
    if (file_info == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_file_info() failed!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        capXferFinish();
        return;
    }

    // What is our current state.
    uint32_t camera_state  = camera_ptr->read_camera_cache(GIGE_REG_CAMERA_STATE);

#if LOG_BATCH_DOWNLOAD_TRACES
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Current camera state" << MainWindow::camera_state_to_string(camera_state)
                 << "is_series3():" << camera_ptr->is_series3() << "CamExplore::cur_state:" << cur_state;
#endif

    // TICKET #441: To be on the safe side let's put the camera in Live Video state prior to loading a cap file.
    // this is the way things have to be in order for Storage/Explore to deploy so let's be consistent for all
    // CAP loads.  It's different depending on whether or not it's a series3 camera.
    if (camera_ptr->is_series3())
    {
        // Only tolerate live video state for Series3 cameras.
        if (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO)
        {
            put_into_live_video = true;
        }
    }
    else
    {
        // SeriesNC cameras seem to tolerate both Live View and Review states for CAP loads.
        if ((camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO) && (camera_state != GIGE_CAMERA_STATE_REVIEW))
        {
            put_into_live_video = true;
        }
    }

    // If we need to put the camera into live video mode do so...
    if (put_into_live_video)
    {
#if LOG_BATCH_DOWNLOAD_TRACES
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Current camera state" << MainWindow::camera_state_to_string(camera_state)
                     << "waiting for GIGE_CAMERA_STATE_LIVE_VIDEO transition.";
#endif
        // Get our communications object pointer.
        FI_comm *comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
            return;
        }

        // Put something on the screen to indicate we are preparing to load.
        // We are inserting a delay - put up something on the screen so the user knows.
        progressDialog = fiProgressDialog::CreateSimpleProgressDialog(camera_ptr->get_camera_name(), tr("  Preparing to load %1  ").arg(file_info->name), QMessageBox::Cancel, __PRETTY_FUNCTION__);

        // Display the progress dialog.
        if (progressDialog)
        {
            connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CamExplore::handleProgressCancelledSignal, Qt::UniqueConnection);
            progressDialog->setBarMinMax(0, 0);
            progressDialog->showBar(true);
            progressDialog->showDialogWithoutPolling();

            // Give the progress bar a chance to start.
            QApplication::processEvents();
        }

        // Setup our timer.
        live_timer_id = startTimer(LIVE_DELAY);

        // Set our state.
        cur_state = CAP_XFER_WAIT_FOR_LIVE;

        // Ticket #537:
        // Set the wait_for_state_count to 1 in order to force a write to occur to change the state
        // right away to LIVE Video when the read complete takes place. The register to change the
        // state back to Live Video MUST be written. Rather than writing it here, keep it centralized
        // where it has always resided. By setting the wait count to 1 it will immediately be decremented to
        // zero in the read complete handler and the write register will take place.
        wait_for_state_count = 1;

        // Read the state.
        comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_CAMERA_STATE);
    }
    else
    {
        // Confirmed Live Video State - make sure the stream is paused.
#if LOG_BATCH_DOWNLOAD_TRACES
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Confirmed camera state for camera series type, ARE WE PAUSED?!";
#endif
        if (cur_state != CAP_XFER_LOAD)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "CamExplore::cur_state != CAP_XFER_LOAD but set to" << cur_state << " resetting it!";
            cur_state = CAP_XFER_LOAD;
        }

        // Make a connection for our special slot processing for series3 cameras.
        // Note:  It will probably also be caught in the normal CamExplore::stream_paused but we
        // are in the CAP_XFER_LOAD state already...
        connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::capXferStartReady);

        // Camera is confirmed in live mode this will pause live video stream or confirm that the
        // stream is paused in either case the signal above will propagate to continue our cap transfer.
        camera_ptr->pause_video_request();
    }
}

/*!
 * @brief This is a special condition to intercept the verification that a live video state
 * is paused for series3 cameras.  It can only be set if cap_xfer_start_delay determines that
 * the connection is to the series3 camera AND in live video mode and we need to verify that
 * it is paused.  It has one pathway to this slot.
 */
void CamExplore::capXferStartReady()
{
#if LOG_BATCH_DOWNLOAD_TRACES
    QLOG_TRACE() << __PRETTY_FUNCTION__;
#endif

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Make sure we are disconnected!
    disconnect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::capXferStartReady);

    // Make sure that no progressDialog is left on the screen.
    if (progressDialog)
    {
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Good to go for series3 cameras.
    capXferStart();
}

/*!
 * @brief The entry for the CAP transfer state machine.
 */
void CamExplore::capXferStart()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "file:" << file_info->name;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // This should not happen but just to be sure it is a benign operation.
    // Don't leave any connection left over because cap_xfer_start_ready calls this function.
    disconnect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::capXferStartReady);

    // Track that we're actually doing a CAP file copy, not just a load.
    cap_copy_in_progress = true;

    // We do not want the event marker information to be generated.
    camera_ptr->disable_event_markers();

    // Disable the cache refresh logic.
    camera_ptr->pause_cache_refresh(true);

    // When we get here, loading is done.
    connect(this, &CamExplore::loadComplete, this, &CamExplore::capLoadedReviewCheck, Qt::UniqueConnection);

    // Independent of CamExplore::load_complete signal we also want to know when the camera signals the review panel is ready.
    camera_review_panel_ready = false;
    connect(camera_ptr, &CameraMDI::review_panel_ready_to_go, this, &CamExplore::capLoadedCamReviewPanelReady, Qt::UniqueConnection);

    // Clear any progress dialog as we will want to use it here for status and to give
    // the user a chance to cancel if we never receive the camera review panel ready signal.
    if (progressDialog)
    {
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Start the loading process.
    capLoadStart(cap_load_sel_row);
}

/*!
 * @brief Utility function containing common clean up code, regardless of the operation
 *        being performed (copy, delete, etc.)
 */
void CamExplore::cleanupState()
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Depending on our state, we'll have different cleanup operations to perform.
    switch (cur_state)
    {
    case CAP_XFER_STARTUP:
    case CAP_XFER_WAIT_FOR_LIVE:
    case CAP_XFER_WAIT_FOR_REVIEW:
    case CAP_XFER_LOAD:
    case CAP_XFER_SAVE:
    case CAP_XFER_COMPLETE:
        break;
    case CAP_LOAD:
        // Resume the cache refresh logic.
        camera_ptr->pause_cache_refresh(false);
        break;
    case COPY_STARTUP:
    case COPY_SET_FILE:
    case COPY_METAFILE:
    case COPY_WAIT:
    case COPY_COMPLETE:
    case COPY_DIRECT:
    case COPY_FAULT:
        // Restore the heartbeat rate.
        camera_ptr->change_heartbeat(0);
        break;
    case DELETE_STARTUP:
    case DELETE_COMPLETE:
        break;
    case REFRESH_STARTUP:
    case REFRESH_STILLS:
    case REFRESH_STACKS:
    case REFRESH_VIDEO:
    case REFRESH_CAPTURE:
    case REFRESH_DOWNLOAD_COMPLETE:
        break;
    case EXPLORE_IDLE:
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected state:" << cur_state;
        break;
    }

    // Change our state back to idle.
    cur_state = EXPLORE_IDLE;

    // Remove any general progress dialog from view and recycle resources.
    if (progressDialog)
    {
        disconnect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CamExplore::handleProgressCancelledSignal);

        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Did we create a copy transfer progress object?  If so clean it up.
    if (xferCopyDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(xferCopyDialog);
    }

    // Disconnect from any saver signal if we may have set up.
    SaveVideo   *saver_ptr = camera_ptr->get_saver();
    if (saver_ptr != nullptr)
    {
        disconnect(saver_ptr, &SaveVideo::userCancelledSaveSignal, this, &CamExplore::handleSaverProgressCancelledSignal);
    }

    // If there is an active timer, kill it.
    if (file_timer_id != 0)
    {
        killTimer(file_timer_id);
        file_timer_id = 0;
    }

    // Release the lock on the camera.
    if (we_own_lock)
        camera_ptr->write_camera_register(GIGE_REG_FASTEC_CCP, 0);

    // Restore our cursor.
    QApplication::restoreOverrideCursor();

    // Now that we're done with that, are we actually doing a move?
    if (move_in_progress)
    {
        // Clear our flag.
        move_in_progress = false;

        // Now delete all the files.
        doDelete();
    }
}

void CamExplore::clearFolders()
{
    int i;

    // Clean our folder display.
    root_folder.clear();
    for (i = 0; i < cam_folders.count(); ++i)
        cam_folders.at(i)->clear();
    for (i = 0; i < root_items.count(); ++i)
    {
        QList<QTreeWidgetItem *> children;

        // Get rid of all the children of the root item.
        children = root_items.at(i)->takeChildren();
        children.clear();
    }
}

/*!
 * @brief The starting routine to the copy operation, setting up the camera to begin
 *        copying one or more files to the local workstation or between the camera's media.
 */
void CamExplore::copyFileStart()
{
    bool        copying_stack = false;
    int         sel_row;
    int         max_value;
    QString     copy_str;
    QTreeWidgetItem *selected_item;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Make sure there's still something to copy.
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "sel_items list is empty?!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer operation because we are in error.
        copyStoppedForError();
        return;
    }

    // Grab the first item off the list.
    selected_item = sel_items.takeFirst();

    // Get the information structure associated with the item. The "sel_row"
    // parameter will get set by the get_file_info() call.
    file_info = getFileInfo(selected_item, sel_row);
    if (file_info == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_file_info() failed!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        copyStoppedForError();
        return;
    }

    // Update the progress bar to reflect the current file/folder being copied.
    if (!xferCopyDialog)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No Copy Transfer Progress Dialog?!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        copyStoppedForError();
        return;
    }

    // If we are copying over a video, then the maximum count is the
    // number of chunks, not the number of frames in the video.
    if ((sel_row == ROW_VIDEO) || (sel_row == ROW_CAPTURE))
    {
        uint64_t    video_size;

        // What type of copy is this?
        copying_stack = false;

        // Update our status line.
        copy_str = file_info->name;

        // See if we're moving more than one chunk.
        video_size = file_info->size.toULongLong();
        if (video_size > (uint64_t)camera_ptr->blob_size())
        {
            blob_count = (int)(video_size / camera_ptr->blob_size());
            if ((video_size % camera_ptr->blob_size()) != 0)
                ++blob_count;
        }
        else
        {
            // We can pull the entire file in one blob.
            blob_count = 1;
        }

        // Save the blob count as our maximum value for the progress bar.
        max_value = blob_count;
    }
    else
    {
        // We are copying a stack of images.
        copying_stack = true;

        // Update our status line.
        copy_str = file_info->name;

        // If we are copying a CAP stack, then the maximum value is the number of files being copied.
        if (file_info->file_ext.compare(FiSaver::IMAGE_EXT_CAPSTACK, Qt::CaseInsensitive) == 0)
        {
            uint64_t    stack_size;

            // See if we're moving more than one chunk.
            stack_size = file_info->size.toULongLong();
            if (stack_size > (uint64_t)camera_ptr->blob_size())
            {
                max_value = (int)(stack_size / camera_ptr->blob_size());
                if ((stack_size % camera_ptr->blob_size()) != 0)
                    ++max_value;
            }
            else
            {
                // We can pull the entire file in one blob.
                max_value = 1;
            }
        }
        else
        {
            // The maximum value is the number of frames being copied.
            max_value = file_info->frame_count.toInt();
            blob_count = 0;
        }
    }

    // Initialize our metafile index. We need to "plus-1" because the code will pre-decrement.
    next_metafile = FI_SAVER::METADATA_TXT_SEQ_NUM + 1;
    last_metafile_ext.clear();

    // How many metafiles are associated with the file being copied?
    // Note:  The old logic was relying on the progress bar to hold the number
    // of files to copy and using that setting to determine if the metafile data
    // had been copied.  This is very dangerous.  The new logic keeps the copy
    // progress dialog showing but each file is unique in the number of frames/blobs
    // that are being transferred and provide a better granularity to the user as to
    // the status of data transfer.  In addition we have no control over the save
    // process (which handles the copy to a PC) and it's possible modification of the
    // progress bar we are "sharing" in order to maintain the dialog on the screen.
    if (camera_ptr->is_series3())
    {
        copyMetafileCount = file_info->meta_count.toUInt();
    }
    else
    {
        // Calculate the number of metafiles we'll copy.
        if (file_info->file_ext.compare(FiSaver::IMAGE_EXT_CAPSTACK, Qt::CaseInsensitive) == 0)
        {
            int cap_count;

            // For a CAP stack, we're copying the metafiles PER CAP file, plus the metamap.txt file.
            cap_count = (file_info->end_cap.toUInt() - file_info->start_cap.toUInt()) + 1;
            copyMetafileCount = (cap_count * getMetafileCountFromFlag(file_info->meta_count.toUInt())) + 1;
        }
        else
        {
            // Everything else has just the one set of metafiles.
            copyMetafileCount = getMetafileCountFromFlag(file_info->meta_count.toUInt());
        }
    }
    copyMetafileId = 0;

    // Initialize our receive counters.
    rcv_retrying        = false;
    rcv_count           = 0;
    rcv_metadata_count  = 0;
    rcv_expected        = -1;

    // Update the maximum value for the progress.
    // Note:  This should never be relied upon to stay in place.
    xferCopyDialog->setXferSrcValue(copy_str);
    xferCopyDialog->setBarValue(0);
    xferCopyDialog->setBarMax(max_value + copyMetafileCount);

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Expecting a total of:" << max_value << "frames/blobs and" << copyMetafileCount << "metafiles.";

    // Build the full file path.
    if (!buildMediaPath(sel_row))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "build_media_path() failed!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        copyStoppedForError();
        return;
    }

    // Append the file name.
    media_path.append(QLatin1Char('/'));
    media_path.append(file_info->name);

    // If we are downloading the file(s) to the local workstation, we have to
    // diverge at this point to do specific "copy-to-local" processing.
    if (copy_destination == DEST_PC_PATH)
    {
        SaveVideo   *saver_ptr;

        // Start the save to our local workstation.
        if (!copyLocalStart(sel_row, file_info))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "copy_local_start() failed!";

            // Make sure we're not moving.
            move_in_progress = false;

            // Clean up and bail.
            copyStoppedForError();
            return;
        }

        // Get our saver object.
        saver_ptr = camera_ptr->get_saver();
        if (saver_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
            return;
        }

        // Let's use different handlers based on the type of copy.
        if (copying_stack)
        {
            // We also tie into the saverBufferReleaseSignal because there is special processing that occurs here.
            connect(saver_ptr, &SaveVideo::saverBufferReleaseSignal, this, &CamExplore::copyStackProgress, Qt::UniqueConnection);
        }
        else
        {
            // We also tie into the saverBufferReleaseSignal because there is special processing that occurs here.
            connect(saver_ptr, &SaveVideo::saverBufferReleaseSignal, this, &CamExplore::copyBlobProgress, Qt::UniqueConnection);
        }


        // Update our destination information.
        if (xferCopyDialog)
            xferCopyDialog->setXferDestValue(saver_ptr->getLocalPath());

        // Slow down the heartbeat rate - transferring data can clog up the network.
        camera_ptr->change_heartbeat(COPY_HEARTBEAT_TIMEOUT);

        // Have the camera object fire off the copy.
        camera_ptr->start_copy(media_path);
    }
    else if (camera_ptr->is_series3())
    {
        // Tell the camera the file/folder we want to copy.
        camera_ptr->write_camera_buffer(GIGE_REG_FILEPATH, media_path.toUtf8());
    }
    else
    {
        QJsonObject json_cmd;
        QString dest_path;

        // Build up the destination path. We need to replace the source "mountdir"
        // with the "copy_dest_path".
        dest_path = media_path;
        dest_path.replace(0, media_info["mountdir"].toString().length(), copy_dest_path);

        // Build up our copy command. The expectation is that the camera server
        // will handle copying all folders, sub-folders, images, metadata, etc.,
        // as appropriate.
        json_cmd["srcpath"] = media_path;
        json_cmd["dstpath"] = dest_path;
        json_cmd["action"] = GIGE_FA_COPY;

        // Send it on it's way.
        QJsonDocument json_doc(json_cmd);
        camera_ptr->write_camera_buffer(GIGE_REG_FA_JSON, json_doc.toJson());

        // Update our state.
        cur_state = COPY_WAIT;
    }
}

/*!
 * @brief The final copy operation state. Clean up and leave.
 *
 * We don't need to worry about transitioning the streamer from the "copy" state
 * to the "paused" state - that's handled by CameraMDI::done_saving().
 */
void CamExplore::copyComplete()
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Are we here after a direct copy?
    if (copy_worker_thread.get() != nullptr)
    {
        // Clean up after ourselves.
        cleanupState();

        // Release the worker thread watcher.
        copy_worker_thread.reset();
    }

    // If we aren't in the middle of the clean up process already, AND
    // the video stream is running AND our destination for the copy operation
    // was the local workstation, then we need to pause the stream.
    else if (!cleanup_in_progress && camera_ptr->get_camera_info()->stream_running && (copy_destination == DEST_PC_PATH))
    {
        // Remember we're in the middle of a cleanup.
        cleanup_in_progress = true;

        // We want to know when the video stream is paused.
        connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::streamPaused);

        // Pause the streamer.
        camera_ptr->pause_video_request();
    }
    else
    {
        // Rest our flag.
        cleanup_in_progress = false;

        // If we were saving to the local workstation, clean up the related objects.
        if (copy_destination == DEST_PC_PATH)
        {
            SaveVideo   *saver_ptr;

            // Get our saver object.
            saver_ptr = camera_ptr->get_saver();
            if (saver_ptr == nullptr)
            {
                QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
                return;
            }

            // Disconnect from the saver's signal.
            disconnect(saver_ptr, &SaveVideo::saverBufferReleaseSignal, this, &CamExplore::copyStackProgress);
            disconnect(saver_ptr, &SaveVideo::saverBufferReleaseSignal, this, &CamExplore::copyBlobProgress);

            // Disconnect from the copy timeout signal.
            disconnect(camera_ptr, &CameraMDI::copy_timeout, this, &CamExplore::handleCopyTimeout);

            // Disconnect from the camera's reported error signal.
            disconnect(camera_ptr, &CameraMDI::save_error_reported, this, &CamExplore::handleSaverErrorReported);

            // Tell the camera object we're done.
            camera_ptr->copy_finished(__PRETTY_FUNCTION__);
        }

        // Clean up after ourselves.
        cleanupState();
    }
}

/*!
 * @brief Continue with the copy-to-camera-media (not to PC) operation.
 */
void CamExplore::copyContinue()
{
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cleanup_in_progress:" << cleanup_in_progress;

    // Do we need to grab some metafiles?
    if ((blob_count > 0) && (copyMetafileId < copyMetafileCount))
    {
        // Special processing if not a series 3 camera.
        if (camera_ptr && !camera_ptr->is_series3())
        {
            if (copyMetafileMap.contains(copyMetafileId))
            {
                next_metafile = copyMetafileMap.value(copyMetafileId);
            }
            else
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Missing metafile index for copyMetafileId:" << copyMetafileId;
                next_metafile--;    // decrement just in case can salvage some copy operation on metafile data.
            }
        }
        else
        {
            next_metafile--;        // Just do what we have always done.
        }

        // Reset our blob counter.
        blob_count = 1;

#if 0
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "blob_count:" << blob_count
                     << "progress:" << copyMetafileId << "max:" << copyMetafileCount
                     << "next_metafile:" << next_metafile;
#endif

        // Request the next metadata file.
        QApplication::postEvent(this, new ExploreEvent(COPY_EVENT, COPY_METAFILE, next_metafile));
    }

    // Are we just starting out?
    else if (rcv_count == 0)
    {
        // We're paused, so now grab the lock.
        camera_ptr->write_camera_register(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE);
    }

    // Are we in the middle of a clean-up, or do we have an empty selection list?
    else if (cleanup_in_progress || sel_items.isEmpty())
    {
        // Wrap up the copy operation.
        QApplication::postEvent(this, new ExploreEvent(COPY_EVENT, COPY_COMPLETE));
    }

    // Move to the next file in the list, if any.
    else
    {
        QApplication::postEvent(this, new ExploreEvent(COPY_EVENT, COPY_SET_FILE));
    }
}

/*!
 * @brief Our worker thread for copying recordings from a camera controller to our
 *        local storage.
 *
 * @param promise   The Qt object used to communicate back to the host.
 * @param parent    Pointer to our parent object for access to class values.
 */
void CamExplore::copyWorker(QPromise<QString>& promise, QWidget *parent)
{
    int total_items;
    int current_progress(0);
    QString copy_error_message;
    CamExplore *my_class = qobject_cast<CamExplore *>(parent);
    QList<QTreeWidgetItem *>::iterator selected_item;

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null parent?!";
        return;
    }

    // Figure out how many things we are copying. At a minimum, we are copying the selected items.
    total_items = my_class->sel_items.count();
    promise.setProgressRange(0, total_items);

    // Start the copy process. Loop through each of the selected items.
    for (selected_item = my_class->sel_items.begin(); !my_class->cancel_in_progress && copy_error_message.isEmpty() && (selected_item != my_class->sel_items.end()); ++selected_item)
    {
        int sel_row;
        QFile full_path;
        QString dest_dir_path;
        QString progress_text;
        QFileInfo source_info;
        FI_SAVER::STREAM_TYPE stream_type;
        const CAM_EXPLORE::FILE_INFO *file_info;

        // Get the information structure associated with the item. The "sel_row"
        // parameter will get set by the get_file_info() call.
        file_info = my_class->getFileInfo(*selected_item, sel_row);
        if (file_info == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_file_info() failed!";
            copy_error_message = QStringLiteral("Unable to access the selection information");
            break;
        }

        // Try to build up a path to the selected item.
        full_path.setFileName(my_class->buildSmbPath(file_info, sel_row));
        if (full_path.fileName().isEmpty())
        {
            copy_error_message = QStringLiteral("Unable to create a path to: ") + file_info->name;
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << copy_error_message;
            break;
        }

        // Get a QFileInfo object for this file.
        source_info.setFile(full_path);

        // What is the destination?
        dest_dir_path = my_class->copy_dest_path;

        // If we are copying a stack, we only want the path to the folder.
        if (sel_row == ROW_STACKS)
        {
            QDir source_dir(source_info.absoluteDir());
            QDir dest_dir(dest_dir_path);

            // Build up our iterator.
            QDirIterator source_it(source_dir, QDirIterator::Subdirectories);

            // Remember the length of the base path for later.
            const int source_path_len = source_dir.absolutePath().length();

            // How many entries are we talking about?
            source_dir.setFilter(QDir::AllEntries | QDir::NoDotAndDotDot);
            total_items += source_dir.count();
            promise.setProgressRange(0, total_items);

            // Build the root folder in the destination path.
            if (!dest_dir.mkdir(source_dir.dirName()))
            {
                copy_error_message = QStringLiteral("Unable to make folder: ") + QDir::toNativeSeparators(dest_dir.absolutePath()) + QDir::separator() + source_dir.dirName();
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << copy_error_message;
                break;
            }

            // Move the destination into the new folder.
            dest_dir.cd(source_dir.dirName());
            dest_dir_path = dest_dir.absolutePath();

            // For a stack, we are copying a folder (and maybe sub-folders).
            while (!my_class->cancel_in_progress && source_it.hasNext())
            {
                // Advance the iterator to the next entry.
                source_info.setFile(source_it.next());

                // Ignore hidden files.
                QString file_name = source_info.fileName();
                if (!source_info.isHidden() && (file_name.compare(QStringLiteral(".")) != 0) && (file_name.compare(QStringLiteral("..")) != 0))
                {
                    const QString sub_path = source_info.absoluteFilePath().mid(source_path_len);
                    const QString constructed_path = dest_dir_path + sub_path;

                    // If the entry is a directory, we need to build a directory in
                    // our target path.
                    if (source_info.isDir())
                    {
                        QDir new_sub(source_info.absoluteFilePath());

                        // Create directory in target folder
                        if (!dest_dir.mkpath(constructed_path))
                        {
                            copy_error_message = QStringLiteral("Unable to make folder: ") + QDir::toNativeSeparators(constructed_path);
                            QLOG_DEBUG() << __PRETTY_FUNCTION__ << copy_error_message;
                            break;
                        }

                        // Update our total item counter.
                        new_sub.setFilter(QDir::AllEntries | QDir::NoDotAndDotDot);
                        total_items += new_sub.count();
                        promise.setProgressRange(0, total_items);
                    }
                    else if (source_info.isFile())
                    {
                        // Update our dialog, but only for "image" files. Not metadata.
                        if ((stream_type = FiSaver::fileExtToStreamType(source_info.suffix())) != FI_SAVER::STREAM_TYPE::NO_STREAM)
                        {
                            if (stream_type == FI_SAVER::STREAM_TYPE::FI_CAP)
                                stream_type = FI_SAVER::STREAM_TYPE::FI_CAPSTK;
                            progress_text = source_info.absoluteFilePath() + QChar('\n') + constructed_path + QChar('\n') + FiSaver::streamTypeToQString(stream_type, false);
                            emit my_class->progressUpdate(progress_text);
                        }

                        // Only circumvent the normal progress updates to use the individual file updates
                        // when copying a CAP STACK.
                        bool cap_stack = (stream_type == FI_SAVER::STREAM_TYPE::FI_CAPSTK);
                        uint32_t cap_stack_update_pct = (cap_stack ? COPY_WORKER_PCT_SIGNAL_FREQUENCY : 0);

                        // Create the source file that will handle the copy operation.
                        FiFile source_file(source_info.absoluteFilePath());

                        // If this is a cap stack we want to give file copy progress updates at cap copy level.
                        if (cap_stack)
                        {
                            // Set the progress bar range from 0 to 100% to catch our copy progression.
                            promise.setProgressRange(0, 100);

                            // We want to catch any progress updates for our file copy to give user an update.
                            connect(&source_file, &FiFile::copyPercentComplete, my_class, &CamExplore::copyWorkerFileProgress, Qt::QueuedConnection);
                        }

                        // Perform the copy operation and signal percentage complete updates at the desired frequency.
                        if (!source_file.copyFile(constructed_path, cap_stack_update_pct))
                        {
                            copy_error_message = QStringLiteral("Unable to copy ") + source_info.absoluteFilePath() + QStringLiteral(" to ") + QDir::toNativeSeparators(constructed_path);
                            QLOG_DEBUG() << __PRETTY_FUNCTION__ << copy_error_message;
                        }

                        // Restore the progress bar range and disconnect copy progress.
                        if (cap_stack)
                        {
                            disconnect(&source_file, &FiFile::copyPercentComplete, my_class, &CamExplore::copyWorkerFileProgress);
                            promise.setProgressRange(0, total_items);
                        }

                        // If an error occurred break out of our loop.
                        if (!copy_error_message.isEmpty())
                            break;
                    }

                    // Update our progress.
                    ++current_progress;
                    promise.setProgressValue(current_progress);
                }
            }
        }
        else
        {
            int copied = 0;
            int i;
            QDir source_dir(source_info.absoluteDir());
            QStringList filters;
            QFileInfoList file_list;
            const int source_path_len = source_dir.absolutePath().length();

            // For anything other than a stack, we are copying single files.
            filters << source_info.baseName() + QStringLiteral(".*");
            source_dir.setNameFilters(filters);
            file_list = source_dir.entryInfoList();

            // Loop through the list.
            for (i = 0; !my_class->cancel_in_progress && (i < file_list.count()); ++i)
            {
                bool metadata_file = true;  // Identify the type of file that is being copied.

                QFileInfo source_info(file_list.at(i));
                const QString sub_path = source_info.absoluteFilePath().mid(source_path_len);
                const QString constructed_path = dest_dir_path + sub_path;

                // Update our dialog, but only for "image" files. Not metadata.
                if ((stream_type = FiSaver::fileExtToStreamType(source_info.suffix())) != FI_SAVER::STREAM_TYPE::NO_STREAM)
                {
                    metadata_file = false;
                    progress_text = source_info.absoluteFilePath() + QChar('\n') + constructed_path + QChar('\n') + FiSaver::streamTypeToQString(stream_type, false);
                    emit my_class->progressUpdate(progress_text);
                }

                // Create the source file that will handle the copy operation.
                FiFile source_file(source_info.absoluteFilePath());

                // If this is a regular file we want to give file copy progress updates.
                if (!metadata_file)
                {
                    // Set the progress bar range from 0 to 100% to catch our copy progression.
                    promise.setProgressRange(0, 100);

                    // We want to catch any progress updates for our file copy to give user an update.
                    connect(&source_file, &FiFile::copyPercentComplete, my_class, &CamExplore::copyWorkerFileProgress, Qt::QueuedConnection);
                }

                // Perform the copy operation and signal percentage complete updates at the desired frequency.
                if (!source_file.copyFile(constructed_path, COPY_WORKER_PCT_SIGNAL_FREQUENCY))
                {
                    copy_error_message = QStringLiteral("Unable to copy ") + source_info.absoluteFilePath() + QStringLiteral(" to ") + QDir::toNativeSeparators(constructed_path);
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << copy_error_message;
                    break;
                }
                else
                {
                    ++copied;   // We have successfully copied this file.
                }

                // Restore the progress bar range and disconnect copy progress.
                if (!metadata_file)
                {
                    disconnect(&source_file, &FiFile::copyPercentComplete, my_class, &CamExplore::copyWorkerFileProgress);
                    promise.setProgressRange(0, file_list.count());
                }

                // Update our local copy progress based on the number of files copied.
                promise.setProgressValue(copied);
            }

            // Restore the original range for overall progress update after processing the non-stack files.
            promise.setProgressRange(0, total_items);
        }

        // Update our progress if no error occurred.
        if (!copy_error_message.isEmpty())
        {
            ++current_progress;
            promise.setProgressValue(current_progress);
        }
    }

    // Are we cancelling? No need to inform the parent of our status if we are.
    if (!my_class->cancel_in_progress)
    {
        // Did we have an error?
        if (!copy_error_message.isEmpty())
        {
            // Return the error message, if there was one.
            promise.addResult(copy_error_message);

            // Wrap up the copy operation.
            QApplication::postEvent(my_class, new ExploreEvent(COPY_EVENT, COPY_FAULT));
        }
        else
        {
            // Wrap up the copy operation.
            QApplication::postEvent(my_class, new ExploreEvent(COPY_EVENT, COPY_COMPLETE));
        }
    }
}

/*!
 * @brief This method is used to setup a direct copy from the connect camera to the local
 *        PC using the Samba interface.
 */
void CamExplore::copyDirect()
{
    // We expect a few things to be setup prior to this method being invoked. Let's
    // make sure we have what we need.
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected empty selection list!";
        return;
    }
    if (xferCopyDialog.get() == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No xferCopyDialog!";
        return;
    }

    // Allocate a new watcher.
    if (copy_worker_thread.get() != nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "copy_worker_thread is not null!?";

        // If the worker thread is still running, we want to cancel it.
        if (copy_worker_thread->isRunning())
        {
            // Cancel the worker thread and wait for it.
            copy_worker_thread->cancel();
            copy_worker_thread->waitForFinished();
        }

        // Now we can release the object.
        copy_worker_thread.reset();
    }

    // We can allocate our new worker object.
    copy_worker_thread = std::make_unique<QFutureWatcher<QString>>();
    if (copy_worker_thread.get() == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate the CopyDirect watcher!";
        return;
    }

    // Because the copy operation can take awhile, we are going to startup a worker thread
    // to do the actual copy in the background. The xferCopyDialog will be used to display
    // progress and to allow the user to cancel the operation if so desired.
    //
    // For the progress updates, we can make use of the features of the QFutureWatcher, but
    // because that "throttles" the events, for the text upates we will go with a custom
    // signal/slot setup.
    xferCopyDialog->setBarValue(0);
    connect(copy_worker_thread.get(), &QFutureWatcherBase::progressRangeChanged, this, [ = ](int min, int max) { if (!cancel_in_progress) xferCopyDialog->setBarMinMax(min, max); });
    connect(copy_worker_thread.get(), &QFutureWatcherBase::progressValueChanged, this, [ = ](int progress) { if (!cancel_in_progress) xferCopyDialog->setBarValue(progress); });
    connect(this, &CamExplore::progressUpdate, this, [ = ](QString progress_text)
    {
        if (!cancel_in_progress)
        {
            // We SHOULD have strings separated by a new-line character: the source, the destination, and the type value.
            QStringList progress_info = progress_text.split("\n");
            if (progress_info.count() > 0)
                xferCopyDialog->setXferSrcValue(progress_info.at(0));
            if (progress_info.count() > 1)
                xferCopyDialog->setXferDestValue(progress_info.at(1));
            if (progress_info.count() > 2)
            {
                xferCopyDialog->setXferTypeValue(progress_info.at(2));
                xferCopyDialog->showXferType(true);
            }
            else
            {
                // No type, so just hide it.
                xferCopyDialog->showXferType(false);
            }
        }
    });

    // Fire up the worker thread and mark our state.
    cur_state = COPY_DIRECT;
    QApplication::setOverrideCursor(Qt::BusyCursor);
    copy_worker_thread->setFuture(QtConcurrent::run(copyWorker, this));
}

/*!
 * @brief This method is the second half of the copy from the camera to the local workstation.
 *        The copy_info structure is populated and passed to the "saver" object.
 *
 * @param sel_row       The ROW_xxx value that was selected.
 * @param file_info     Pointer to the selected entry's information.
 * @return True on success, false on failure.
 */
bool CamExplore::copyLocalSetup(int sel_row, const CAM_EXPLORE::FILE_INFO *file_info)
{
    int         index;
    int         length;
    SaveVideo   *saver_ptr;
    SAVE_VIDEO::COPY_PARAMS copy_info;
    QString firstThreeOfCodecStr;      // Empty until we find the valid type.

    // Sanity check.
    if ((camera_ptr == nullptr) || (file_info == nullptr))
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No camera object or file_info pointer!";
        return false;
    }

    // Is our file incomplete? If so we don't want to try to copy it.
    if (file_info->incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << tr("File:") << file_info->name << tr("Is Incomplete and cannot be copied!");
        return false;
    }

    // Get our saver object.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
        return false;
    }

    // Parse the important bits from the source file name.
    copy_info.tag_label = file_info->name;
    index = copy_info.tag_label.lastIndexOf(QLatin1Char('_'));
    length = copy_info.tag_label.length() - index;
    if (length > 0)
    {
        // Extract the tag info.
        copy_info.tag_label.chop(length);

        // Is this just the default tag?
        if (copy_info.tag_label.compare(FiSaver::DEFAULT_TAG, Qt::CaseInsensitive) == 0)
            copy_info.tag_label.clear();
    }
    else
    {
        // No tag information.
        copy_info.tag_label.clear();
    }

    // Initialize the rest of our save parameters structure.
    copy_info.dest_path = download_base_path;
    copy_info.src_path = media_path;
    copy_info.use_camera_name = ui->chk_use_name->isChecked();
    copy_info.use_existing_name = false;

    // How many metafiles are we expecting?
    copy_info.metaFileCount = copyMetafileCount;

    // Configure our save_type based on the selected row.
    switch (sel_row)
    {
    case ROW_STACKS:
        // Copy the start/stop frame numbers from the file info object.
        copy_info.first_frame_num = file_info->start_frame.toUInt();
        copy_info.last_frame_num = file_info->end_frame.toUInt();

        // Are we converting to an AVI?  The checkbox must also be enabled to use it.
        // #SAVE_ERROR - to force one just use flip flop the commented line below with
        // the one that is active in actual software.  Then copy a TIF stack after having
        // selected Option[X] and [AVI] in your copy options for a JPEG stack.
        // if (ui->chk_convert->isChecked())
        if (ui->chk_convert->isEnabled() && ui->chk_convert->isChecked())
        {
            // Our save type will be AVI.
            copy_info.streamType = FI_SAVER::STREAM_TYPE::AVI;

            // What are the source images formatted as?
            if (file_info->file_ext.compare(FiSaver::IMAGE_EXT_BMP, Qt::CaseInsensitive) == 0)
                copy_info.imageOutputType = copy_info.imageInputType = FI_SAVER::IMAGE_TYPE::BMP;
            else
                copy_info.imageOutputType = copy_info.imageInputType = FI_SAVER::IMAGE_TYPE::JPEG;
        }
        else
        {
            // All others get straight transfers - type to type.
            copy_info.streamType = saver_ptr->fileExtToStreamType(file_info->file_ext);
            copy_info.imageOutputType = copy_info.imageInputType = saver_ptr->fileExtToImageType(file_info->file_ext);
        }
        break;
    case ROW_VIDEO:
        // The first expected sequence number we get from the camera depends on if the video
        // fits in a single blob or not.
        if (blob_count == 1)
        {
            uint32_t    file_id;

            // The first sequence number will be the file ID of the video.
            file_id = parseID(file_info->name);

            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ROW_VIDEO file_id for blob_count of 1:" << file_id;

            // The blob count defines how many "chunks" we'll take to copy over the video.
            copy_info.first_frame_num = file_id;
            copy_info.last_frame_num = file_id;
        }
        else
        {
            // The blob count defines how many "chunks" we'll take to copy over the video.
            copy_info.first_frame_num = 0;
            copy_info.last_frame_num = (uint32_t)(blob_count - 1);
        }

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "blob_count:" << blob_count << "file_ext:" << file_info->file_ext;

        // Transfer the AVI and MP4s as a BLOB. We use the "CAPTURE" format here to distinguish
        // between copying an AVI/CAP file from the camera and copying a stack of images from
        // the camera. The file_ext field contains the name of the codec used. For MP4s that'll be "H264".
        copy_info.streamType = FI_SAVER::STREAM_TYPE::FI_CAP;

        // Note:  Handle two types of MP4 video indicators.  For camera server saves this value is
        // "H264" for FasMotion saves this value is "MP4" - need to support both.  And test for AVI
        // as well.  Here is MP4 Save to PC for example:  "MP4:speed=Medium:quality=70:keyframes=60"
        if (file_info->file_ext.length() < 3)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to determine videoType based on this file_ext:" << file_info->file_ext;

            // Make sure our string is empty.
            firstThreeOfCodecStr.clear();
        }
        else
        {
            // Pull out the first three letters of the file extension.
            firstThreeOfCodecStr = QString("%1%2%3").arg(file_info->file_ext[0]).arg(file_info->file_ext[1]).arg(file_info->file_ext[2]);
            firstThreeOfCodecStr = firstThreeOfCodecStr.toUpper();
        }

        if (firstThreeOfCodecStr.compare("H26") == 0)
        {
            copy_info.videoType = FI_SAVER::STREAM_TYPE::MP4;
        }
        else if (firstThreeOfCodecStr.compare("MP4") == 0)
        {
            copy_info.videoType = FI_SAVER::STREAM_TYPE::MP4;
        }
        else if (firstThreeOfCodecStr.compare("AVI") == 0)
        {
            copy_info.videoType = FI_SAVER::STREAM_TYPE::AVI;
        }
        else if (firstThreeOfCodecStr.compare("MJP") == 0)
        {
            copy_info.videoType = FI_SAVER::STREAM_TYPE::AVI;
        }
        else if (firstThreeOfCodecStr.compare("DIB") == 0)
        {
            copy_info.videoType = FI_SAVER::STREAM_TYPE::AVI;
        }
        else
        {
            // We don't understand the codec name of video file we have here...
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "The video codec for file copy:" << file_info->file_ext
                         << "is not understood!  We are setting to NO_STREAM.";
            copy_info.videoType = FI_SAVER::STREAM_TYPE::NO_STREAM;
        }
        break;
    case ROW_STILLS:
        // Extract the sequence number from the file name of the still we're copying.
        // That will be our "first" frame number.
        copy_info.first_frame_num = parseID(file_info->name);
        copy_info.last_frame_num = copy_info.first_frame_num;

        // Stills are JPEGs.
        copy_info.streamType = FI_SAVER::STREAM_TYPE::JPEG_STACK;
        break;
    case ROW_CAPTURE:
        // On the series3 cameras, we cannot download a CAP file.
        if (camera_ptr->is_series3())
            return false;

        // The first expected sequence number we get from the camera depends on if the video
        // fits in a single blob or not.
        if (blob_count == 1)
        {
            uint32_t    file_id;

            // The first sequence number will be the file ID of the video.
            file_id = parseID(file_info->name);

            // The blob count defines how many "chunks" we'll take to copy over the video.
            copy_info.first_frame_num = file_id;
            copy_info.last_frame_num = file_id;
        }
        else
        {
            // The blob count defines how many "chunks" we'll take to copy over the video.
            copy_info.first_frame_num = 0;
            copy_info.last_frame_num = (uint32_t)(blob_count - 1);
        }

        // Transfer the CAP as a BLOB. We use the "CAPTURE" format here to distinguish
        // between copying an AVI/CAP file from the camera and copying a stack of images from
        // the camera.
        copy_info.streamType = FI_SAVER::STREAM_TYPE::FI_CAP;
        copy_info.videoType = FI_SAVER::STREAM_TYPE::FI_CAP;
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Invalid selection row:" << sel_row;
        return false;
    }

    // Setup the saver object.
    if (!saver_ptr->copy_to_pc(copy_info))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to initialize the copy operation!";
        return false;
    }

    // Good to go.
    return true;
}

/*!
 * @brief The entry routine to starting a copy to the local workstation. This routine sanity checks the destination
 *        media to make sure there's sufficient space for the operation.
 *
 * @param sel_row       The ROW_xxx value that was selected.
 * @param file_info     Pointer to the selected entry's information.
 * @return True on success, false on failure.
 */
bool CamExplore::copyLocalStart(int sel_row, const CAM_EXPLORE::FILE_INFO *file_info)
{
    uint64_t    size_avail;
    uint64_t    video_size;

    // Sanity check.
    if (file_info == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No file_info pointer!";
        return false;
    }

    // Is our file incomplete? If so we don't want to try to copy it.
    if (file_info->incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << tr("File:") << file_info->name << tr("Is Incomplete and cannot be copied!");
        return false;
    }

    // Make sure that there's enough space on the target drive.
    video_size = file_info->size.toULongLong();
    size_avail = MainWindow::GetInstance()->get_avail_space(download_base_path);
    if (video_size > size_avail)
    {
        QString msg_str;

        // Let the user know there's a problem.
        msg_str = tr("Not enough free space on device (required:%1 free:%2)").arg(video_size).arg(size_avail);
        fiMsgBox::warning(this, msg_str);
        return false;
    }

    // Check to see if we need to worry about the underlying file system.
    if (ui->chk_convert->isChecked() || (sel_row == ROW_VIDEO))
    {
        // Do we need to calculate size?
        if (video_size == 0)
        {
            // These files can get very large. Make sure the underlying file system will support
            // the size of our video.
            if ((file_info->file_ext.compare(FiSaver::IMAGE_EXT_JPG, Qt::CaseInsensitive) == 0) ||
                (file_info->file_ext.compare(QString::fromUtf8("jpeg"), Qt::CaseInsensitive) == 0))
            {
                // For JPGs, things are more complicated. Make a very rough estimate for
                // image file size.
                video_size = (uint64_t)((file_info->width.toInt() * file_info->height.toInt()) >> 1) * file_info->frame_count.toULongLong();
            }
            else if (file_info->file_ext.compare(FiSaver::IMAGE_EXT_BMP, Qt::CaseInsensitive) == 0)
            {
                // For BMPs, things are pretty straight forward. This is only an estimate,
                // but it should be good enough.
                video_size = (uint64_t)((file_info->width.toInt() * file_info->height.toInt()) * 3) * file_info->frame_count.toULongLong();
            }
        }

        // Check our size.
        if (video_size > FAS_VIDEO::FI_VIDEO_MAX_FAT32_SIZE)
        {
            // Check to see if the file system supports large files.
            if (!MainWindow::GetInstance()->fs_supports_large_files(download_base_path))
            {
                fiMsgBox::warning(this, tr("The destination folder does not support large (greater than 4GB) files."));
                return false;
            }
        }
    }

    // If we're here, then we've passed all the tests. Setup for the transfer.
    return copyLocalSetup(sel_row, file_info);
}

/*!
 * @brief This method is required when copying video files. The camera doesn't
 *        automatically send the metadata files, so we have to request them individually.
 */
void CamExplore::copyMetafile()
{
    SaveVideo   *saver_ptr;
    QString     image_ext;
    SAVE_VIDEO::COPY_PARAMS copy_info;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Processing next_metafile:" << int(next_metafile);

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        copyStoppedForError();
        return;
    }

    // Get our saver object.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        copyStoppedForError();
        return;
    }

    // Get the extension for the image.
    image_ext = saver_ptr->getFileExtension();

    copy_info.metaFileExt = "";
    copy_info.metaFileCount = 1;    // Only one this go around.

    // Modify our file name to specify the new metadata file name.
    switch (next_metafile)
    {
    case FI_SAVER::METADATA_TXT_SEQ_NUM:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FI_SAVER::METADATA_TXT_SEQ_NUM";
        // This is always the first one processed and always exists.
        media_path.replace(media_path.lastIndexOf(image_ext), image_ext.length(), FiSaver::IMAGE_EXT_META_TEXT);
        last_metafile_ext = copy_info.metaFileExt = FiSaver::IMAGE_EXT_META_TEXT;
        break;
    case FI_SAVER::METADATA_CFG_SEQ_NUM:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FI_SAVER::METADATA_CFG_SEQ_NUM";
        media_path.replace(media_path.lastIndexOf(last_metafile_ext), last_metafile_ext.length(), FiSaver::IMAGE_EXT_META_CFG);
        last_metafile_ext = copy_info.metaFileExt = FiSaver::IMAGE_EXT_META_CFG;
        break;
    case FI_SAVER::METADATA_XML_SEQ_NUM:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FI_SAVER::METADATA_XML_SEQ_NUM";
        media_path.replace(media_path.lastIndexOf(last_metafile_ext), last_metafile_ext.length(), FiSaver::IMAGE_EXT_META_XML);
        last_metafile_ext = copy_info.metaFileExt = FiSaver::IMAGE_EXT_META_XML;
        break;
    case FI_SAVER::METADATA_RAW_SEQ_NUM:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FI_SAVER::METADATA_RAW_SEQ_NUM";
        if (camera_ptr && !camera_ptr->is_series3())
        {
            media_path.replace(media_path.lastIndexOf(last_metafile_ext), last_metafile_ext.length(), FiSaver::IMAGE_EXT_META_RAW);
            last_metafile_ext = copy_info.metaFileExt = FiSaver::IMAGE_EXT_META_RAW;
            break;
        }
        // Otherwise do as it has always done...

        //
        // Fall through intentional...
        //
        [[fallthrough]];

    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected next_metafile:" << int(next_metafile);

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        copyStoppedForError();
        return;
    }

    // Fill in the structure.
    copy_info.dest_path = saver_ptr->getLocalPath();
    copy_info.first_frame_num = 0;
    copy_info.last_frame_num = 0;
    copy_info.imageInputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
    copy_info.imageOutputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
    copy_info.src_path = media_path;
    copy_info.tag_label.clear();
    copy_info.use_existing_name = true;
    copy_info.use_camera_name = ui->chk_use_name->isChecked();

    // This is a metafile - mark the "video" type as the
    // proper video type so that when it is time to replace
    // the name of the extension with the metafile extension
    // type it can find the right extension to replace in the
    // dest_path variable of the saver object.
    if (image_ext.compare(FiSaver::IMAGE_EXT_AVI) == 0)
        copy_info.videoType = FI_SAVER::STREAM_TYPE::AVI;
    else if (image_ext.compare(FiSaver::IMAGE_EXT_MP4) == 0)
        copy_info.videoType = FI_SAVER::STREAM_TYPE::MP4;
    else if (image_ext.compare(FiSaver::IMAGE_EXT_CAPTURE) == 0)
        copy_info.videoType = FI_SAVER::STREAM_TYPE::FI_CAP;

    // We are treating metafile data streamType as a capture
    // blob - but this actually gets turned into a stack file
    // in the actual fiSaver::SaveToStream function because
    // the sequence number is a negative number. A bait and
    // switch if you will.
    copy_info.streamType = FI_SAVER::STREAM_TYPE::FI_CAP;

    // Setup the saver object.
    if (!saver_ptr->copy_to_pc(copy_info))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to initialize the copy operation!";

        // Make sure we're not moving.
        move_in_progress = false;

        // Close out the transfer.
        copyComplete();
        return;
    }

    // We will increase our metafile copy id by 1 so that the condition tested in the
    // function CamExplore::copy_continue() will determine whether or not there are
    // more metafiles to manually copy.
    copyMetafileId++;

    // Connect the saver's signal so we know the progress.
    connect(saver_ptr, &SaveVideo::saverBufferReleaseSignal, this, &CamExplore::copyBlobProgress, Qt::UniqueConnection);

    // Reset our receive counter.
    rcv_retrying        = false;
    rcv_count           = 0;
    rcv_metadata_count  = 0;
    rcv_expected        = -1;

    // Have the camera object fire off the copy.
    camera_ptr->start_copy(media_path);
}

/*!
 * @brief When the copy is between the camera's media, this method is used to initiate
 *        the copying process.
 */
void CamExplore::copyStartAction()
{
    uint32_t    reg_value;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // If we're copying from the camera to the local workstation, we need to
    // go to a different routine.
    if (copy_destination == DEST_PC_PATH)
        return;

    // We're going to be doing a copy operation.
    reg_value = GIGE_FA_COPY;

    // Tell the camera the destination media.
    switch (copy_destination)
    {
    case DEST_SDCARD:   reg_value |= GIGE_IMAGE_DEST_SDCARD;    break;
    case DEST_SSD:      reg_value |= GIGE_IMAGE_DEST_SSD;       break;
    case DEST_USB:      reg_value |= GIGE_IMAGE_DEST_USB;       break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unknown copy destination:" << Qt::showbase << Qt::hex << copy_destination;
        return;
    }

    // Tell the camera to start the copy.
    camera_ptr->write_camera_register(GIGE_REG_FILE_ACTION, reg_value);
    cur_state = COPY_WAIT;
}

/*!
 * @brief The copy operation needs to stop because of error condition.
 *
 */
void CamExplore::copyStoppedForError()
{
    // Right now this only calls the copy_complete function but wanted intermediary
    // to follow flow in the log messages.
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
    copyComplete();
}


/*!
 * @brief The logic that creates and initializes a copy transfer progress dialog.
 *
 * @returns true if created properly, false if something is wrong.
 */
bool CamExplore::createXferCopyDialog(FI_SAVER::STREAM_TYPE streamType, QString& destValue)
{
    bool rc = false;

    // Create a file transfer progress dialog.
    xferCopyDialog = fiProgressDialog::CreateXferProgressDialog(FI_PROGRESS::XFER_TYPES::XFER_TYPE_COPY, camera_ptr->get_camera_name(), __PRETTY_FUNCTION__);

    // Do we have the transfer progress dialog?  If so set it up for the copy.
    if (xferCopyDialog)
    {
        // Set up the source information.
        xferCopyDialog->setXferSrcLabel(COPY_SRC_RECEIVE_FRAME);
        xferCopyDialog->showXferSrc(true);

        // Set up the destination information.
        xferCopyDialog->setXferDestValue(destValue);
        xferCopyDialog->showXferDest(true);

        // Set up the save type information.
        xferCopyDialog->setXferTypeValue(FiSaver::streamTypeToQString(streamType, false));
        xferCopyDialog->showXferType((streamType == FI_SAVER::STREAM_TYPE::NO_STREAM) ? false : true);

        // Set up the status icon.
        xferCopyDialog->showStatusIcon(false);
        xferCopyDialog->setStatusIcon(QStyle::SP_MessageBoxInformation);

        // Set up the status.
        xferCopyDialog->setStatusValue(tr("Transferring images..."));
        xferCopyDialog->showStatusValue(false);

        xferCopyDialog->setBarMinMax(0, 100);
        xferCopyDialog->setBarValue(0);
        xferCopyDialog->showBar(true);

        // Allow the user to cancel this process and catch the cancelled signal.
        xferCopyDialog->showCancel(true);
        connect(xferCopyDialog.get(), &fiProgressDialog::progressCancelled, this, &CamExplore::userCanceledCopy, Qt::UniqueConnection);

        // If the parent is closing, we need to bail.
        connect(camera_ptr, &CameraMDI::closing, xferCopyDialog.get(), &fiProgressDialog::cancelDialog);

        rc = true;
    }
    else
    {
        // If cannot create the transfer progress dialog something is terribly wrong.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a fiProgressDialog!";
    }
    return rc;
}

/*!
 * @brief The final delete operation state. Clean up and leave.
 */
void CamExplore::deleteFinish()
{
    // Clean up after ourselves.
    cleanupState();
}

/*!
 * @brief The entry for the delete state machine.
 */
void CamExplore::deleteStart()
{
    int sel_row;
    QTreeWidgetItem *selected_item;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Make sure there's still something to delete.
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "sel_items list is empty?!";
        deleteFinish();
        return;
    }

    // Grab the first item off the list.
    selected_item = sel_items.takeFirst();

    // Get the information structure associated with the item. The "sel_row"
    // parameter will get set by the get_file_info() call.
    file_info = getFileInfo(selected_item, sel_row);
    if (file_info == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_file_info() failed!";
        deleteFinish();
        return;
    }

    // Update the progress bar to reflect the current file/folder being removed.
    if (progressDialog)
        progressDialog->setStatusValue(file_info->name);

    // Get the full file path.
    if (!buildMediaPath(sel_row))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "build_media_path() failed!";
        deleteFinish();
        return;
    }

    // Append the file name.
    media_path.append(QLatin1Char('/'));
    media_path.append(file_info->name);

    // Tell the camera the file we want to delete.
    camera_ptr->write_camera_buffer(GIGE_REG_FILEPATH, media_path.toUtf8());
}


/*!
 * @brief Format and display the metadata for the currently selected entry based on the
 *        file_info structure passed in.
 *
 * @param file_info     The file information we are going to display.
 */
void CamExplore::displayMetadata(const CAM_EXPLORE::FILE_INFO *file_info)
{
    QString meta_str = CamExplore::generateMetadataString(file_info);

    // Set the information in our display.
    ui->txt_meta->setPlainText(meta_str);
}

int CamExplore::findDevice(QJsonObject& dev_object)
{
    int i;
    int found_index = -1;

    // Loop through the entries in the combobox and see if we have a match.
    for (i = 0; i < ui->cb_devices->count(); ++i)
    {
        QJsonObject test_object = ui->cb_devices->itemData(i).toJsonObject();

        // Check to see if we have a match.
        if ((test_object["type"].toInt() == dev_object["type"].toInt()) &&
            (test_object["label"].toString().compare(dev_object["label"].toString()) == 0))
        {
            // Found it.
            found_index = i;
            break;
        }
    }

    return found_index;
}

void CamExplore::getDestination()
{
    int i;
    uint32_t reg_value;
    QList<QString> avail_media;

    // Default initialization.
    copy_destination = DEST_NONE;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (media_info.isEmpty())
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No selected media info!";
        return;
    }

    // Build up the array of destination targets.
    reg_value = camera_ptr->read_camera_cache(GIGE_REG_LTR_MODE);
    for (i = 0; i < ui->cb_devices->count(); ++i)
    {
        // Get an object for the index element.
        QJsonObject dev_object = ui->cb_devices->itemData(i).toJsonObject();

        // Make sure it's not our source.
        if (dev_object["name"].toString() == media_info["name"].toString())
            continue;

        // Make sure it's a "mountable" device.
        if (dev_object["mountdir"].toString().isEmpty())
            continue;

        // No saving to the SSD if we're in LTR mode.
        if (((reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE) &&
            (dev_object["type"].toInt() == GIGE_MEDIA_SSD))
        {
            // Not allowed.
            continue;
        }

        // Append the name to the
        avail_media.append(dev_object["tag"].toString());
    }

    // Create our dialog.
    fiSelDestDlg dlg(this);
    dlg.avail_media(avail_media);

    // Skip it if there's no media.
    if (avail_media.count() > 0)
    {
        if (dlg.exec() == QDialog::Rejected)
            return;
    }

    // Did the user want to pick a path?
    if (dlg.selected() == avail_media.count())
    {
        // If the download path is empty set it to the standard download path location for system.
        if (download_base_path.isEmpty())
        {
            const QString downloadsFolder = QStandardPaths::writableLocation(QStandardPaths::DownloadLocation);
            download_base_path = downloadsFolder;
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "QStandardPaths::DownloadLocation:" << downloadsFolder;
        }

        // Ask the user where they want to save the data.
        copy_dest_path = QFileDialog::getExistingDirectory(this, FILE_DIRECTORY_SELECT_STR, download_base_path);
        if (copy_dest_path.isEmpty())
        {
            // User must have canceled.
            return;
        }

        // Remember that this is a local PC path.
        copy_destination = DEST_PC_PATH;

        // Save the user selected path.
        SaveDialog::savePath(cam_ip_addr, copy_dest_path);
        download_base_path = copy_dest_path;
    }
    else
    {
        // Find the object associated with the selected string.
        for (i = 0; i < ui->cb_devices->count(); ++i)
        {
            // Get an object for the index element.
            QJsonObject dev_object = ui->cb_devices->itemData(i).toJsonObject();

            // Make sure it's not our source.
            if (dev_object["tag"].toString() == avail_media.at(dlg.selected()))
            {
                // Remember the base path for the device.
                copy_dest_path = dev_object["mountdir"].toString();

                // Translate the type to a DEST_xxx value.
                switch (dev_object["type"].toInt())
                {
                case GIGE_MEDIA_SDCARD:
                    copy_destination = DEST_SDCARD;
                    break;
                case GIGE_MEDIA_SSD:
                    copy_destination = DEST_SSD;
                    break;
                case GIGE_MEDIA_USB:
                    copy_destination = DEST_USB;
                    break;
                default:
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected type:" << dev_object["type"].toInt();
                    break;
                }

                // And done.
                break;
            }
        }
    }
}

/*!
 * @brief Utility function to determine a destination for the copy/move operations.
 * @param num_videos    The number of videos were selected. This is used to determine
 *                      if the SSD is available as a destination.
 */
void CamExplore::getDestination(int num_videos)
{
    bool    ssd_avail;
    bool    usb_avail;
    bool    sdcard_avail;
    uint32_t    reg_value;
    QString     user_base_path;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // See what media is avaialble on the camera. The media is NOT available if
    // it is our source for the data.
    reg_value = camera_ptr->read_camera_cache(GIGE_REG_AVAIL_MEDIA);
    sdcard_avail = ((reg_value & GIGE_MEDIA_EXISTS_SDCARD) == (uint32_t)GIGE_MEDIA_EXISTS_SDCARD) && (media_type != GIGE_MEDIA_SDCARD);
    ssd_avail = ((reg_value & GIGE_MEDIA_EXISTS_SSD) == (uint32_t)GIGE_MEDIA_EXISTS_SSD) && (media_type != GIGE_MEDIA_SSD);
    usb_avail = ((reg_value & GIGE_MEDIA_EXISTS_USB) == (uint32_t)GIGE_MEDIA_EXISTS_USB) && (media_type != GIGE_MEDIA_USB);

    // We can't save videos to the "new" A3+ boards SSDs.
    if ((num_videos > 0) && camera_ptr->is_series3() && camera_ptr->cap_is_allowed())
        ssd_avail = false;

    // No saving to the SSD if we're in LTR mode.
    reg_value = camera_ptr->read_camera_cache(GIGE_REG_LTR_MODE);
    if ((reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        ssd_avail = false;

    // Do we have at least one destination (other than the PC)?
    copy_destination = DEST_PC_PATH;
    reg_value = 0;
    if (sdcard_avail)
        ++reg_value;
    if (ssd_avail)
        ++reg_value;
    if (usb_avail)
        ++reg_value;
    if (reg_value > 0)
    {
        QMessageBox *msgbox_ptr;
        QPushButton *sdcard_button = nullptr;
        QPushButton *ssd_button = nullptr;
        QPushButton *usb_button = nullptr;
        QPushButton *path_button = nullptr;

        // Pop up the question to allow the user to select their location for copying the file
        msgbox_ptr = new QMessageBox(QMessageBox::NoIcon, tr("FasMotion"), tr("Choose Destination"), QMessageBox::Cancel, this);
        if (sdcard_avail)
            sdcard_button = msgbox_ptr->addButton(tr("SD Card"), QMessageBox::YesRole);
        if (ssd_avail)
            ssd_button = msgbox_ptr->addButton(tr("SSD"), QMessageBox::AcceptRole);
        if (usb_avail)
            usb_button = msgbox_ptr->addButton(tr("USB"), QMessageBox::DestructiveRole);
        path_button = msgbox_ptr->addButton(tr("Path..."), QMessageBox::ActionRole);
        msgbox_ptr->exec();

        // Figure out what the user selected.
        if (msgbox_ptr->clickedButton() == sdcard_button)
        {
            copy_destination = DEST_SDCARD;
        }
        else if (msgbox_ptr->clickedButton() == ssd_button)
        {
            copy_destination = DEST_SSD;
        }
        else if (msgbox_ptr->clickedButton() == usb_button)
        {
            copy_destination = DEST_USB;
        }
        else if (msgbox_ptr->clickedButton() == path_button)
        {
            copy_destination = DEST_PC_PATH;
        }
        else
        {
            // User must have canceled.
            copy_destination = DEST_NONE;
        }

        // Clean up our message box.
        delete msgbox_ptr;
    }

    // Did the user cancel?
    if (copy_destination == DEST_NONE)
        return;

    // If the destination is on the camera, then we're done.
    if (copy_destination != DEST_PC_PATH)
        return;

    // If the download path is empty set it to the standard download path location for system.
    if (download_base_path.isEmpty())
    {
        const QString downloadsFolder = QStandardPaths::writableLocation(QStandardPaths::DownloadLocation);
        download_base_path = downloadsFolder;
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "QStandardPaths::DownloadLocation:" << downloadsFolder;
    }

    // Ask the user where they want to save the data.
    user_base_path = QFileDialog::getExistingDirectory(this, FILE_DIRECTORY_SELECT_STR, download_base_path);
    if (user_base_path.isEmpty())
    {
        // User must have canceled.
        copy_destination = DEST_NONE;
        return;
    }

    // Save the user selected path.
    SaveDialog::savePath(cam_ip_addr, user_base_path);
    download_base_path = user_base_path;
}

/*!
 * @brief Passed a tree widget item. This routine will find the file info info for that tree item.
 *
 * @param item      The item to be checked.
 * @param row_out   [OUTPUT] The row the item was found in.
 *
 * @return The file_info structure or NULL if not found.
 */
const CAM_EXPLORE::FILE_INFO *CamExplore::getFileInfo(const QTreeWidgetItem *item, int& row_out)
{
    int     i;
    int     row;

    // Initialize our output value.
    row_out = -1;

    // Loop through each row (aka image type) of our listing database.
    for (row = 0; row < cam_folders.count(); ++row)
    {
        const LISTING_INFO  *list_info = cam_folders.at(row);

        // For each row (image type), look at each of the entries we have.
        for (i = 0; i < list_info->count(); ++i)
        {
            // Do we have a match?
            if (list_info->at(i).tree_widget == item)
            {
                // Return the row (image type) associated with this entry as
                // well as the information associated with the entry.
                row_out = row;
                return &list_info->at(i);
            }
        }
    }

    // Didn't find it.
    return nullptr;
}

/*!
 * @brief Translates a metafile_count set up as a flag into the number of
 * expected metafiles in a count.
 *
 * @param metafile_count_as_flag - the data returned from a seriesNC camera
 * server.
 *
 * @return The count of metafiles in the flag setting.
 */
uint32_t CamExplore::getMetafileCountFromFlag(uint32_t metafile_count_as_flag)
{
    uint32_t metafile_count = 0;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "flag:" << metafile_count_as_flag;

    // Reset the map that translates index into metafile type.
    copyMetafileMap.clear();

    // Go through the flag settings incrementing the count if set.
    if (metafile_count_as_flag & GIGE_FA_META_TXT)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "will copy METADATA_TXT_SEQ_NUM";
        copyMetafileMap.insert(metafile_count,  FI_SAVER::METADATA_TXT_SEQ_NUM);
        metafile_count++;
    }
    if (metafile_count_as_flag & GIGE_FA_META_CFG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "will copy METADATA_CFG_SEQ_NUM";
        copyMetafileMap.insert(metafile_count,  FI_SAVER::METADATA_CFG_SEQ_NUM);
        metafile_count++;
    }
    if (metafile_count_as_flag & GIGE_FA_META_XML)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "will copy METADATA_XML_SEQ_NUM";
        copyMetafileMap.insert(metafile_count,  FI_SAVER::METADATA_XML_SEQ_NUM);
        metafile_count++;
    }
    if (metafile_count_as_flag & GIGE_FA_META_PM)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "will copy METADATA_RAW_SEQ_NUM";
        copyMetafileMap.insert(metafile_count,  FI_SAVER::METADATA_RAW_SEQ_NUM);
        metafile_count++;
    }

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "metafile_count:" << metafile_count;

    return metafile_count;
}

/*!
 * @brief Called to determine if the item passed in is one of the root items. If so the
 *        row number is returned, otherwise -1 is returned.
 *
 * @param item_ptr  The item to be checked.
 * @return The row (aka image type) or -1 if not a root item.
 */
int CamExplore::getRootRow(const QTreeWidgetItem *item_ptr)
{
    int row = -1;

    // Loop through each of the root items.
    for (int i = 0; i < root_items.count(); ++i)
    {
        // Is this the one we're looking for?
        if (item_ptr == root_items.at(i))
        {
            row = i;
            break;
        }
    }

    // Return our finding.
    return row;
}

/*!
 * @brief Get how many of each type are selected.
 *
 * @param num_root      Number of root rows selected.
 * @param num_stills    Number of still rows selected.
 * @param num_stacks    Number of stack rows selected.
 * @param num_videos    Number of videos rows selected.
 * @param num_captures  Number of capture rows selected.
 * @param convertable_found When looking at selected items was one of them a "convertable" (JPEG or BMP) image stack.
 * @param incomplete_found - Did we find any selected incomplete files?
 */
void CamExplore::getSelectionCounts(int& num_roots, int& num_stills, int& num_stacks, int& num_videos, int& num_caps, bool& convertable_found, bool& incomplete_found)
{
    int     count;
    int     i;
    const FILE_INFO             *p_file_info;
    QList<QTreeWidgetItem *>    selected_items;

    // Initialize our return values.
    num_roots = num_stills = num_stacks = num_videos = num_caps = 0;
    convertable_found = true;
    incomplete_found = false;

    // Get the selected items. If there aren't any, go no further.
    selected_items = ui->cam_tree->selectedItems();
    if (selected_items.isEmpty())
        return;

    // Walk through all of the selected items to enumerate the selection information.
    count = selected_items.count();
    for (i = 0; i < count; ++i)
    {
        // Get a pointer to this item.
        const QTreeWidgetItem  *selected_item = selected_items.at(i);

        // Is it in a "root" row?
        int root_row = getRootRow(selected_item);
        if (root_row >= 0)
        {
            // root row selected.
            ++num_roots;
        }
        else
        {
            // We know that we got a "real" row, so figure out what kind it is.
            p_file_info = getFileInfo(selected_item, root_row);

            // Make sure there was a valid file information pointer retrieved.
            if (p_file_info == nullptr)
            {
                // Skip this item as there is no file information pointer!
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "No file_info pointer. Unknown row type:" << root_row;

                // The selected format is not something we can convert.
                convertable_found = false;
                continue;
            }

            switch (root_row)
            {
            case ROW_STILLS:
                ++num_stills;

                // The selected format is not something we can convert.
                convertable_found = false;
                break;
            case ROW_STACKS:
                ++num_stacks;

                //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "p_file_info->incomplete:" << p_file_info->incomplete;
                // Check for incomplete stacks.
                if (p_file_info->incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
                {
                    // We have found at least one.
                    incomplete_found = true;
                }

                // Check the format of the selected image(s).
                if ((p_file_info->file_ext.compare(QString::fromUtf8("JPEG"), Qt::CaseInsensitive) != 0) &&
                    (p_file_info->file_ext.compare(QString::fromUtf8("BMP"), Qt::CaseInsensitive) != 0))
                {
                    // The selected format is not something we can convert.
                    convertable_found = false;
                }
                break;
            case ROW_VIDEO:
                ++num_videos;

                // Check for incomplete videos.
                if (p_file_info->incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
                {
                    // We have found at least one.
                    incomplete_found = true;
                }

                // The selected format is not something we can convert.
                convertable_found = false;
                break;
            case ROW_CAPTURE:
                ++num_caps;

                // The selected format is not something we can convert.
                convertable_found = false;
                break;
            default:
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unknown row type:" << root_row;

                // The selected format is not something we can convert.
                convertable_found = false;
                break;
            }
        }
    }
}

/*!
 * @brief Looks the the cur_state and returns user relatable string for reporting a message.
 */
QString CamExplore::getCurStateMsg()
{
    QString the_state_str;

    switch (cur_state)
    {
    case CAP_XFER_STARTUP:
        the_state_str = "CAP_XFER_STARTUP";
        break;
    case CAP_XFER_WAIT_FOR_LIVE:
        the_state_str = "CAP_XFER_WAIT_FOR_LIVE";
        break;
    case CAP_XFER_WAIT_FOR_REVIEW:
        the_state_str = "CAP_XFER_WAIT_FOR_REVIEW";
        break;
    case CAP_XFER_LOAD:
        the_state_str = "CAP_XFER_LOAD";
        break;
    case CAP_XFER_SAVE:
        the_state_str = "CAP_XFER_SAVE";
        break;
    case CAP_XFER_COMPLETE:
        the_state_str = "CAP_XFER_COMPLETE";
        break;
    case CAP_LOAD:
        the_state_str = "CAP_LOAD";
        break;
    case COPY_STARTUP:
        the_state_str = "COPY_STARTUP";
        break;
    case COPY_SET_FILE:
        the_state_str = "COPY_SET_FILE";
        break;
    case COPY_METAFILE:
        the_state_str = "COPY_METAFILE";
        break;
    case COPY_PAUSED:
        the_state_str = "COPY_PAUSED";
        break;
    case COPY_WAIT:
        the_state_str = "COPY_WAIT";
        break;
    case COPY_COMPLETE:
        the_state_str = "COPY_COMPLETE";
        break;
    case COPY_DIRECT:
        the_state_str = "COPY_DIRECT";
        break;
    case COPY_FAULT:
        the_state_str = "COPY_FAULT";
        break;
    case DELETE_STARTUP:
        the_state_str = "DELETE_STARTUP";
        break;
    case DELETE_SET_FILE:
        the_state_str = "DELETE_SET_FILE";
        break;
    case DELETE_WAIT:
        the_state_str = "DELETE_WAIT";
        break;
    case DELETE_COMPLETE:
        the_state_str = "DELETE_COMPLETE";
        break;
    case REFRESH_STARTUP:
        the_state_str = "REFRESH_STARTUP";
        break;
    case REFRESH_STILLS:
        the_state_str = "REFRESH_STILLS";
        break;
    case REFRESH_STACKS:
        the_state_str = "REFRESH_STACKS";
        break;
    case REFRESH_VIDEO:
        the_state_str = "REFRESH_VIDEO";
        break;
    case REFRESH_CAPTURE:
        the_state_str = "REFRESH_CAPTURE";
        break;
    case REFRESH_DOWNLOAD_COMPLETE:
        the_state_str = "REFRESH_DOWNLOAD_COMPLETE";
        break;

    default:
        the_state_str = tr("UNKNOWN cur_state[%1]").arg(cur_state);
        break;
    }
    return the_state_str;
}



/*!
 * @brief Utility function to process the write completion status of the FileAction register
 *        write performed by one of the operations.
 */
void CamExplore::parseFileActionStatus(uint32_t fa_status)
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    if ((fa_status & GIGE_FA_STATUS_BUSY) == (uint32_t)GIGE_FA_STATUS_BUSY)
    {
        int value(0);

        // Still busy. Do we want to update the progress dialog?
        if (progressDialog && (progressDialog->getBarMax() != 0))
        {
            // Parse the current progress from the register value.
            value = fa_status & GIGE_FA_STATUS_MASK;
            progressDialog->setBarValue(value);
        }

        // Still busy. Do we want to update the transfer dialog?
        if (xferCopyDialog && (xferCopyDialog->getBarMax() != 0))
        {
            // Parse the current progress from the register value.
            value = fa_status & GIGE_FA_STATUS_MASK;
            xferCopyDialog->setBarValue(value);
        }

        // Start our timer to poll for the completion of the file action.
        if (file_timer_id == 0)
            file_timer_id = startTimer(FA_POLLING_PERIOD);
    }
    else if ((fa_status & GIGE_FA_STATUS_ERROR) == (uint32_t)GIGE_FA_STATUS_ERROR)
    {
        QString err_str;
        QString err_info_str;

        // Turn off the timer. The timer SHOULD have already been killed.
        if (file_timer_id != 0)
        {
            killTimer(file_timer_id);
            file_timer_id = 0;
        }

        // Get some sort of idea as to what is happening with this error.
        QString state_info_msg = getCurStateMsg();

        // Build up our error message.
        if (file_info != nullptr)
            err_str = tr("An error has occurred with %1 during %2").arg(file_info->name, state_info_msg);
        else
            err_str = tr("An error has occurred during %1.").arg(state_info_msg);

        // Build up our string with more information.
        err_info_str = tr("The camera has reported an error:\n%1 (%2)").arg(fa_status & GIGE_FA_STATUS_MASK).arg(strerror(fa_status & GIGE_FA_STATUS_MASK));
        QLOG_WARN() << __PRETTY_FUNCTION__ << err_str;

        // Show the user the error information.
        fiMsgBox::critical(camera_ptr, err_str, err_info_str);

        // Treat this like the user canceled.
        if (progressDialog)
        {
            progressDialog->cancelDialog();
        }

        // Treat this like the user canceled.
        if (xferCopyDialog)
        {
            xferCopyDialog->cancelDialog();
        }
    }
    else
    {
        uint32_t    read_len;

        // Turn off the timer.
        if (file_timer_id != 0)
        {
            killTimer(file_timer_id);
            file_timer_id = 0;
        }

        // Process based on the current state.
        switch (cur_state)
        {
        case REFRESH_STILLS:
        case REFRESH_STACKS:
        case REFRESH_VIDEO:
        case REFRESH_CAPTURE:
            // Extract the number of bytes in the listing.
            listing_len = fa_status & GIGE_FA_STATUS_MASK;
            QLOG_TRACE() << __PRETTY_FUNCTION__ << "file:" << listing_file.fileName() << "len:" << listing_len << "cur_state:" << cur_state;

            // How much can we grab?
            if (listing_len > GIGE_FILEPATH_LEN)
                read_len = GIGE_FILEPATH_LEN;
            else
                read_len = listing_len;

            // Start pulling down the listing.
            camera_ptr->start_camera_read_buf(GIGE_REG_FILEPATH, read_len);
            break;

        case COPY_WAIT:
            // Move to the next file in the list, if any.
            if (sel_items.isEmpty())
                QApplication::postEvent(this, new ExploreEvent(COPY_EVENT, COPY_COMPLETE));
            else
                QApplication::postEvent(this, new ExploreEvent(COPY_EVENT, COPY_SET_FILE));
            break;
        case DELETE_WAIT:
            // Move to the next file in the list, if any.
            if (sel_items.isEmpty())
                QApplication::postEvent(this, new ExploreEvent(DELETE_EVENT, DELETE_COMPLETE));
            else
                QApplication::postEvent(this, new ExploreEvent(DELETE_EVENT, DELETE_SET_FILE));
            break;
        default:
            // Other states may happen, especially as a result of a user-cancel. We
            // want to ignore these events.
            break;
        }
    }
}

/*!
 * @brief This method performs the actual ReadMem calls to download the directory listing produced
 *        on the camera by a previous FileAction register write.
 *
 * The assumption is that everything is setup, the destination file has been created and left in
 * an open, usable state so that this method simply has to pull down the data and write it out.
 */
void CamExplore::listingDownloadProcess()
{
    uint32_t    buffer_len;
    QByteArray  *read_chunk;

    //QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get the chunk of data we just read from the camera.
    read_chunk = camera_ptr->read_camera_buffer(GIGE_REG_FILEPATH);
    if (read_chunk == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No buffer available!";
        return;
    }
    if (read_chunk->size() <= 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Bad buffer length: " << read_chunk->size();
        delete read_chunk;
        return;
    }

    // Get the number of bytes we've read...but we only care up to listing_len bytes.
    buffer_len = (uint32_t)read_chunk->size();
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "listing_file:" << listing_file.fileName() << "listing_len:" << listing_len << "chunk size:" << buffer_len;
    if (buffer_len > listing_len)
        buffer_len = listing_len;

    // Write the chunk out to our file.
    listing_file.write(read_chunk->constData(), read_chunk->size());

    // Done with the chunk.
    delete read_chunk;

    // How much do we have left?
    listing_len -= buffer_len;
    if (listing_len != 0)
    {
        uint32_t    read_len;

        // How much can we grab?
        if (listing_len > GIGE_FILEPATH_LEN)
            read_len = GIGE_FILEPATH_LEN;
        else
            read_len = listing_len;
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Need more, read_len:" << read_len;

        // Request the next chunk.
        camera_ptr->start_camera_read_buf(GIGE_REG_FILEPATH, read_len);
    }
    else
    {
        int next_stage;

        // We're done with this listing. Close the file.
        listing_file.close();

        // Parse the information we've downloaded.
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "refresh_row:" << refresh_row;
        listingParse();

        // Move on to the next stage of the refresh operation.
        switch (refresh_row)
        {
        case ROW_STILLS:
            next_stage = REFRESH_STACKS;
            break;
        case ROW_STACKS:
            next_stage = REFRESH_VIDEO;
            break;
        case ROW_VIDEO:
            if (captures_allowed)
                next_stage = REFRESH_CAPTURE;
            else
                next_stage = REFRESH_DOWNLOAD_COMPLETE;
            break;
        case ROW_CAPTURE:
            next_stage = REFRESH_DOWNLOAD_COMPLETE;
            break;
        default:
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected refresh row:" << refresh_row;
            next_stage = REFRESH_DOWNLOAD_COMPLETE;
            break;
        }

        // Post our event for the next stage of processing.
        QApplication::postEvent(this, new ExploreEvent(REFRESH_EVENT, next_stage));
    }
}

/*!
 * @brief We're all done with the refresh logic of downloading the listings for each
 *        of the image types (stills, stacks, etc.) Clean-up the progress information
 *        and parse through all the data to provide a useful display for the user.
 */
void CamExplore::listingFinish()
{
    int     row;
    double  free_space;
    double  capacity;
    QString free_space_str;
    QString capacity_str;
    QString meta_str;
    QString inodes_str;

    // Clean up after ourselves.
    cleanupState();

    // Loop through each of the image types in the "root" of our tree display.
    for (row = 0; row < root_items.count(); ++row)
    {
        int i;
        QTreeWidgetItem     *current_item;

        // Get a handy to use pointer.
        current_item = root_items.at(row);

        // Now that we have the folder list, display them in the tree.
        for (i = 0; i < cam_folders.at(row)->count(); ++i)
        {
            FiTreeWidgetItem *p_child;
            LISTING_INFO    *cur_list_info;

            // Get our listing entry.
            cur_list_info = cam_folders[ row ];
            if (cur_list_info == nullptr)
                continue;

            // Get our file info structure.
            FILE_INFO& cur_file_info = (*cur_list_info)[i];

            //QLOG_DEBUG() << "LISTING[" << i << "]: name=" << cur_file_info.name << "row:" << row << "file_ext:" << cur_file_info.file_ext;

            // Build up our tree structure.
            p_child = new FiTreeWidgetItem(current_item);
            p_child->setText(COLUMN_NAME, cur_file_info.name);
            switch (row)
            {
            case ROW_STILLS:
                p_child->setIcon(COLUMN_NAME, still_icon);
                break;
            case ROW_STACKS:
                p_child->setIcon(COLUMN_NAME, stack_icon);
                break;
            case ROW_VIDEO:
                p_child->setIcon(COLUMN_NAME, video_icon);
                break;
            case ROW_CAPTURE:
                p_child->setIcon(COLUMN_NAME, capture_icon);
                break;
            default:
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected row value:" << row;
                break;
            }
            p_child->setText(COLUMN_SIZE, cur_file_info.size);
            p_child->setTextAlignment(COLUMN_SIZE, Qt::AlignRight | Qt::AlignVCenter);

            // We are going to override the time stamp if the file is incomplete.
            // We want the user to note right away the file is not valid.
            if (cur_file_info.incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
            {
                p_child->setText(COLUMN_TIME, FILE_INFO_INCOMPLETE_STR);
            }
            else
            {
                p_child->setText(COLUMN_TIME, ClientMath::fi_time_to_local(cur_file_info.time_stamp));
            }

            // Hook the widget into our listing database for later.
            cur_file_info.tree_widget = p_child;
        }

        // Sort the entries in this row by the name column.
        current_item->sortChildren(COLUMN_NAME, Qt::AscendingOrder);
    }

    // Setup the free space display.
    free_space = (double)root_folder.free;
    if (free_space >= (1024.0 * 1024.0 * 1024.0))
    {
        free_space = free_space / (1024.0 * 1024.0 * 1024.0);
        free_space_str = QString(tr("Free Space: %1GB ")).arg(free_space, 0, 'f', 1);
    }
    else if (free_space >= (1024.0 * 1024.0))
    {
        free_space = free_space / (1024.0 * 1024.0);
        free_space_str = QString(tr("Free Space: %1MB ")).arg(free_space, 0, 'f', 1);
    }
    else
    {
        free_space = free_space / (1024.0 * 1024.0);
        free_space_str = QString(tr("Free Space: %1KB ")).arg(free_space, 0, 'f', 1);
    }
    capacity = (double)root_folder.capacity;
    if (capacity >= (1024.0 * 1024.0 * 1024.0))
    {
        capacity = capacity / (1024.0 * 1024.0 * 1024.0);
        capacity_str = QString(tr("Capacity: %1GB ")).arg(capacity, 0, 'f', 1);
    }
    else if (capacity >= (1024.0 * 1024.0))
    {
        capacity = capacity / (1024.0 * 1024.0);
        capacity_str = QString(tr("Capacity: %1MB ")).arg(capacity, 0, 'f', 1);
    }
    else
    {
        capacity = capacity / (1024.0 * 1024.0);
        capacity_str = QString(tr("Capacity: %1KB ")).arg(capacity, 0, 'f', 1);
    }
    inodes_str = QString(tr("INodes: %1 ")).arg(root_folder.inodes);
    meta_str = capacity_str + free_space_str + inodes_str;
    ui->lbl_space->setText(meta_str);

    // Are we doing a CAP file load?
    if (cap_file_load)
    {
        // NOTE:
        // While testing Ticket #454 I came across a QList assertion for access beyond count.
        // The function CamExplore::setCapturesAllowed(bool do_allow) is the interface responsible
        // for determining if CamExplore::captures_allowed is true or false. It is possible to end
        // up here with cap_file_load=true but captures_allowed=false. If that is the condition an
        // assertion can take place if there are not enough entries for ROW_CAPTURE. In order to
        // avoid uneccesary assertions the code was modified to log a warning but stop any cap file
        // load operation. This only happened for a series3 camera in testing when the call to
        // CameraMDI::cap_is_allowed() returned false (meaning it can't verify the SSD). It could
        // also occur if the series3 camera does have an SSD but has an old board rev.
        if (root_items.count() > ROW_CAPTURE)
        {
            // If doing a CAP file load, set the CAP row as the currently selected item.
            ui->cam_tree->setCurrentItem(root_items[ROW_CAPTURE]);

            // Make sure the tree is expanded.
            ui->cam_tree->expandItem(root_items[ROW_CAPTURE]);
        }
        else
        {
            // We can't load a cap file if we don't have any in the explorer list.
            cap_file_load = false;
            QLOG_WARN() << __PRETTY_FUNCTION__ << "captures_allowed:" << captures_allowed << "root_items count:" << root_items.count() << "is missing ROW_CAPTURE!";
        }
    }
}

/*!
 * @brief Parse the file ID from the file name. If it's a single file, like
 *    a snap shot, then the file name looks something like "ts3_000000.jpg".
 *    The ID in this example would be 0.
 *
 *    For image stacks, the file name IS the frame number which is the ID.
 *    So a file name might be "0000501.dng" and the ID would be 501.
 *
 *    Very large files, such as CAP files, might be split into two files,
 *    so the split file might look like "ts3_000005_1.cap". The ID of this
 *    file is 5, just like the first half of the file - they both share the
 *    same ID.
 *
 *    Metadata files such as "metadata.txt" don't have a file name, so we
 *    associate a special ID for them. -1 (0xFFFFFFFF) for the ".txt" file,
 *    -2 (0xFFFFFFFE) for the ".cfg" file, -3 (0xFFFFFFFD) for the ".xml" file.
 *
 * @param file_path The fully qualified file name.
 *
 * @return The ID, or 0 if we can't figure one out.
 */
uint32_t CamExplore::parseID(const QString& file_path)
{
    bool    bOK;
    int     index;
    int     end_index;
    uint32_t    ret_value;
    QString     sub_str;

    // Initialize our pointer to the period leading the suffix.
    index = file_path.lastIndexOf(QChar('.'));
    if (index == -1)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Expected suffix period not found:" << file_path;
        return 0;
    }

    // Get the suffix.
    sub_str = file_path.mid(index + 1);

    // Check to see if this is a metadata file.
    if (sub_str == FiSaver::IMAGE_EXT_META_XML)
        return (uint32_t) -3;
    else if (sub_str == FiSaver::IMAGE_EXT_META_CFG)
        return (uint32_t) -2;
    else if (sub_str == FiSaver::IMAGE_EXT_META_TEXT)
        return (uint32_t) -1;

    // Check to see if this is a "split" file - a file that was split into 2 parts
    // because it exceeded 4Gbytes.
    if (file_path.at(index - 2) == QChar('_'))
    {
        QLOG_INFO() << __PRETTY_FUNCTION__ << "Split file detected" << file_path;

        // Back up the index to this spot.
        index -= 2;
    }

    // Index is either at the '.' or the '_'. Save this place for later.
    end_index = index;

    // Now backup 7 characters.
    index -= 7;

    // We're either at an underscore character, or a digit. If it's an underscore
    // we need to move forward 1 character to a digit.
    if (file_path.at(index) == QChar('_'))
    {
        index++;
    }

    // We should be pointed at a digit now.
    sub_str = file_path.mid(index, (end_index - index));

    // Convert the value.
    ret_value = sub_str.toUInt(&bOK);
    if (!bOK)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Expected digit not found:" << file_path;
        return 0;
    }

    // Return our value.
    return ret_value;
}

/*!
 * @brief Start the process of downloading the file listing for the data specified
 *        by the listing_row parameter - stills, stacks, videos, CAPs.
 *
 * @param listing_row   Which image type we're interested in.
 */
void CamExplore::listingStart(int listing_row)
{
    QString         label_str;
    QString         xml_filename;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "listing_row:" << listing_row;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Build up the path on the camera for the media we're interested in.
    if (!buildMediaPath(listing_row))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "build_media_path() failed!";
        return;
    }

    // Now build up our local file and our label based on that row.
    switch (listing_row)
    {
    case ROW_STACKS:
        xml_filename = QDir::tempPath() + STACKS_DIR_LISTING;
        label_str = tr("Retrieving listing for the STACKs folder");
        break;
    case ROW_VIDEO:
        xml_filename = QDir::tempPath() + VIDEOS_DIR_LISTING;
        label_str = tr("Retrieving listing for the VIDEOs folder");
        break;
    case ROW_CAPTURE:
        xml_filename = QDir::tempPath() + CAPTURES_DIR_LISTING;
        label_str = tr("Retrieving listing for the CAPTUREs folder");
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected row:" << listing_row;
        return;
    }

    // Remove any old file that might still be laying around.
    QFile::remove(xml_filename);

    // Get ready for the new file.
    listing_file.setFileName(xml_filename);
    if (!listing_file.open(QIODevice::WriteOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open" << xml_filename << ". Err =" << listing_file.errorString();
        return;
    }
    listing_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // Update our progress dialog information, if we have one.
    if (progressDialog)
    {
        progressDialog->setStatusValue(label_str);
        progressDialog->setBarValue(progressDialog->getBarValue() + 1);
    }

    // Request the directory listing from the camera. First thing to do is to
    // tell the camera what path we're interested in.
    camera_ptr->write_camera_buffer(GIGE_REG_FILEPATH, media_path.toUtf8());
}

/*!
 * @brief Parse the downloaded listing file (as specified by the "listing_file" member), and build
 *        up our listing database with all the associated metadata.
 */
void CamExplore::listingParse()
{
    QDomDocument    doc(QString::fromUtf8("dir_list"));
    QDomElement     dom_root;           // xml token "root"
    QDomElement     dom_folder;         // xml token "folder"
    QDomElement     dom_file;           // xml token "file"
    QString         msg_str;
    FILE_INFO       cur_file_info;
    QString         target_name;
    QString         target_cap;
    QString         target_avail;
    QString         target_inodes;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "listing_file:" << listing_file.fileName();

    if (!listing_file.open(QIODevice::ReadOnly))
    {
        msg_str = QString::fromUtf8("Error opening returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(listing_file.error()).arg(listing_file.errorString());
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        return;
    }

    if (!doc.setContent(&listing_file))
    {
        msg_str = QString::fromUtf8("Error parsing returned XML directory listing %1 code:%2 (%3)")
                  .arg(listing_file.fileName()).arg(errno).arg(QString::fromUtf8(strerror(errno)));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;

        listing_file.close();
        return;
    }

    // Don't need this anymore it's been read in.
    listing_file.close();

    // Make sure the refresh_row is a valid row value.
    switch (refresh_row)
    {
    case ROW_STILLS:
    case ROW_STACKS:
    case ROW_VIDEO:
    case ROW_CAPTURE:
        break;
    default:
        refresh_row = ROW_STACKS;  // Just so we don't crash default it to the stack list.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Invalid refresh_row[" << refresh_row << "] provided for directory listing!";
        break;
    }

    // we need to get what target we are working on and how much the capacity is.
    dom_root = doc.firstChildElement(QString::fromUtf8("root"));
    target_name = dom_root.attribute(QString::fromUtf8("target"));
    target_cap = dom_root.attribute(QString::fromUtf8("capacity"));
    target_avail = dom_root.attribute(QString::fromUtf8("available"));
    target_inodes = dom_root.attribute(QString::fromUtf8("inodes"));
    root_folder.target = target_name;
    root_folder.capacity = target_cap.toULongLong();
    root_folder.free = target_avail.toULongLong();
    root_folder.inodes = target_inodes.toULongLong();
    if (root_folder.capacity == 0)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "file:" << listing_file.fileName() << "Capacity is zero?!";
        return;
    }

    // We are going to be navigating the document differently if we asked for stacks vs everything
    // else for the stack directory there are folder tags which go into the stack group, but
    // there are also file tags which go into the stills group, all other types there are
    // only files which go in the appropriate tag
    // first stacks
    dom_folder = dom_root.firstChildElement(QString::fromUtf8("folder"));
    if (dom_folder.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No folder element?";
    }
    while (!dom_folder.isNull())
    {
        bool ignore_folder = true;     // For now subfolders that aren't in a stack are ignored.

        cur_file_info.clear();
        cur_file_info.name = dom_folder.attribute(QString::fromUtf8("name"));
        if (!cur_file_info.name.isEmpty())
        {
            cur_file_info.size = dom_folder.firstChildElement(QString::fromUtf8("size")).text();
            cur_file_info.width = dom_folder.firstChildElement(QString::fromUtf8("width")).text();
            cur_file_info.height = dom_folder.firstChildElement(QString::fromUtf8("height")).text();
            cur_file_info.bit_mode = dom_folder.firstChildElement(QString::fromUtf8("bit_mode")).text();
            cur_file_info.frame_count = dom_folder.firstChildElement(QString::fromUtf8("frame_count")).text();
            cur_file_info.time_stamp = dom_folder.firstChildElement(QString::fromUtf8("time_stamp")).text();
            cur_file_info.fps = dom_folder.firstChildElement(QString::fromUtf8("fps")).text();
            cur_file_info.file_ext = dom_folder.firstChildElement(QString::fromUtf8("codec")).text();
            cur_file_info.meta_count = dom_folder.firstChildElement(QString::fromUtf8("metafile_count")).text();
            cur_file_info.raw = dom_folder.firstChildElement(QString::fromUtf8("raw")).text();
            cur_file_info.start_frame = dom_folder.firstChildElement(QString::fromUtf8("start_frame")).text();
            cur_file_info.end_frame = dom_folder.firstChildElement(QString::fromUtf8("end_frame")).text();
            cur_file_info.comment = dom_folder.firstChildElement(QString::fromUtf8("comment")).text();

            // BUGFIX #365 - It is possible we have user-created subfolders in non-stack scenarios like the
            // capture or video lists.  In this case we won't assume the folder is invalid.  However it is also
            // possible for the user to create subfolders in a ROW_STACK directory.  These will come out as
            // being incomplete as the metadata won't be there.  So this is still a bug and Ticket #369 addresses
            // this topic.
            if (refresh_row == ROW_STACKS)
            {
                ignore_folder = false;      // We will put in subfolders for stacks.
                // BUGFIX #269
                // This is not currently provided here but query for it anyway to get an empty string
                // and support the future should it ever be added for stacks.
                cur_file_info.incomplete = dom_file.firstChildElement(QString::fromUtf8("incomplete")).text();

                // Is this a CAP stack?
                if (cur_file_info.file_ext.compare(FiSaver::IMAGE_EXT_CAPSTACK, Qt::CaseInsensitive) == 0)
                {
                    // CAP stacks specify which CAP file is the starting one for the recording.
                    cur_file_info.start_cap = dom_folder.firstChildElement(QString::fromUtf8("start_cap")).text();
                    cur_file_info.end_cap = dom_folder.firstChildElement(QString::fromUtf8("end_cap")).text();
                }

                // Perform some sanity checks on the folder data.  If we did not receive all of the frames for
                // the stack (because our metadata is empty) then mark the stack as incompleted.
                if (cur_file_info.incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) != 0)
                {
                    // Run through what we expect to be part of a complete stack set.
                    if (cur_file_info.size.isEmpty() ||
                        cur_file_info.width.isEmpty() ||
                        cur_file_info.height.isEmpty() ||
                        cur_file_info.frame_count.isEmpty() ||
                        cur_file_info.time_stamp.isEmpty() ||
                        cur_file_info.file_ext.isEmpty() ||
                        cur_file_info.meta_count.isEmpty() ||
                        cur_file_info.start_frame.isEmpty() ||
                        cur_file_info.end_frame.isEmpty())
                    {
                        cur_file_info.incomplete = FILE_INFO_INCOMPLETE;
                    }
                }
            }
#if 0
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Folder Information ignore_folder:" << ignore_folder << " incomplete:" << cur_file_info.incomplete << "name:" << cur_file_info.name
                         << "time_stamp:" << cur_file_info.time_stamp << "size:" << cur_file_info.size << "codec:" << cur_file_info.file_ext;
#endif
            if (!ignore_folder)
            {
                // BUGFIX #365:  Before this was putting any "refresh_row" subfolder into the cam_folder list
                // for ROW_STACKS but this is incorrect.  We want to use the global variable "refresh_row" to
                // insert the file information record so the folder is inserted in the proper place.  Right now
                // we will be ignoring the subfolders in other ROW_* entries but in the future we will need to
                // support them (see management of ignore_folder.  This will be important for the FasFire mode to
                // our Fastec File System.
                cam_folders[ refresh_row ]->append(cur_file_info);
            }
        }
        dom_folder = dom_folder.nextSiblingElement(QString::fromUtf8("folder"));
    }

    // Now we want to get information on files.
    dom_file = dom_root.firstChildElement(QString::fromUtf8("file"));
    if (dom_file.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No file element?";
    }
    while (!dom_file.isNull())
    {
        cur_file_info.clear();
        cur_file_info.name = dom_file.attribute(QString::fromUtf8("name"));
        if ((!cur_file_info.name.isEmpty()) && (cur_file_info.name.indexOf(IGNORE_FILENAME) == -1))
        {
            cur_file_info.size = dom_file.firstChildElement(QString::fromUtf8("size")).text();
            cur_file_info.width = dom_file.firstChildElement(QString::fromUtf8("width")).text();
            cur_file_info.height = dom_file.firstChildElement(QString::fromUtf8("height")).text();
            cur_file_info.bit_mode = dom_file.firstChildElement(QString::fromUtf8("bit_mode")).text();
            cur_file_info.frame_count = dom_file.firstChildElement(QString::fromUtf8("frame_count")).text();
            cur_file_info.time_stamp = dom_file.firstChildElement(QString::fromUtf8("time_stamp")).text();
            cur_file_info.fps = dom_file.firstChildElement(QString::fromUtf8("fps")).text();
            cur_file_info.file_ext = dom_file.firstChildElement(QString::fromUtf8("codec")).text();
            cur_file_info.meta_count = dom_file.firstChildElement(QString::fromUtf8("metafile_count")).text();
            cur_file_info.raw = dom_file.firstChildElement(QString::fromUtf8("raw")).text();
            cur_file_info.start_frame = dom_file.firstChildElement(QString::fromUtf8("start_frame")).text();
            cur_file_info.end_frame = dom_file.firstChildElement(QString::fromUtf8("end_frame")).text();
            cur_file_info.comment = dom_file.firstChildElement(QString::fromUtf8("comment")).text();
            cur_file_info.incomplete = dom_file.firstChildElement(QString::fromUtf8("incomplete")).text();
#if 0
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "File Information incomplete:" << cur_file_info.incomplete << "name:" << cur_file_info.name
                         << "time_stamp:" << cur_file_info.time_stamp << "size:" << cur_file_info.size;
#endif
            // Perform some sanity checks on the file data.  If we did not receive all of the data we expected
            // mark the file as incomplete even if the incomplete flag was not set.
            if (cur_file_info.incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) != 0)
            {
                // Run through what we expect to be part of a complete file.
                if (cur_file_info.size.isEmpty() ||
                    cur_file_info.width.isEmpty() ||
                    cur_file_info.height.isEmpty() ||
                    cur_file_info.frame_count.isEmpty() ||
                    cur_file_info.time_stamp.isEmpty() ||
                    cur_file_info.file_ext.isEmpty() ||
                    cur_file_info.meta_count.isEmpty() ||
                    cur_file_info.start_frame.isEmpty() ||
                    cur_file_info.end_frame.isEmpty())
                {
                    cur_file_info.incomplete = FILE_INFO_INCOMPLETE;
                }
            }

            // BUGFIX:  The test was using ROW_STACKS not ROW_STILLS!  Roll in with Ticket #365
            if (refresh_row == ROW_STILLS)
            {
                // Stills can't be incomplete so we will allow the emptiness of metadata
                // in this case - reset it clear.
                if (cur_file_info.incomplete.compare(FILE_INFO_INCOMPLETE, Qt::CaseInsensitive) == 0)
                {
                    cur_file_info.incomplete.clear();    // Make it empty.
                }
                cam_folders[ ROW_STILLS ]->append(cur_file_info);
            }
            else
                cam_folders[ refresh_row ]->append(cur_file_info);
        }
        else if (cur_file_info.name.indexOf(IGNORE_FILENAME) != -1)
        {
            // This kind of file is the second section of the file. What we need to do is add the
            // section size to what was reported for the original file section.
            qulonglong      size_cur = 0;
            qulonglong      size_new = 0;
            qulonglong      size_total = 0;
            QString     root_filename;
            int         ignore_filename_idx = 0;
            int         ext_idx = 0;

            cur_file_info.size = dom_file.firstChildElement(QString::fromUtf8("size")).text();
            size_new = cur_file_info.size.toULongLong();

            // create the base filename which doesn't include the "_1", this is needed to find the correct
            // record to add the size to.
            ignore_filename_idx = cur_file_info.name.indexOf(IGNORE_FILENAME);
            ext_idx = cur_file_info.name.lastIndexOf(QString::fromUtf8("."));
            root_filename = QStringView{cur_file_info.name}.left(ignore_filename_idx).toString();
            root_filename += QStringView{cur_file_info.name}.mid(ext_idx).toString();
            for (int i = 0; i < cam_folders.at(refresh_row)->count(); i++)
            {
                if (cam_folders.at(refresh_row)->at(i).name == root_filename)
                {
                    size_cur = cam_folders.at(refresh_row)->at(i).size.toULongLong();
                    size_total = size_cur + size_new;
                    (*cam_folders[ refresh_row ])[ i ].size = QString::number(size_total);
                }
            }
        }
        dom_file = dom_file.nextSiblingElement(QString::fromUtf8("file"));
    }
}

/*!
 * @brief Used to specify if CAP files are allowed on the media we're looking at.
 * @param do_allow  True if captures are allowed, false otherwise.
 */
void CamExplore::setCapturesAllowed(bool do_allow)
{
    FiTreeWidgetItem *tree_item;

    // Remember the setting.
    captures_allowed = do_allow;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "captures_allowed:" << captures_allowed;

    // If captures are allowed, we need an entry in our tree.
    if (captures_allowed)
    {
        // Check to see if we have a capture entry already.
        for (int i = 0; i < root_items.count(); ++i)
        {
            QString root_text = root_items.at(i)->text(COLUMN_TITLE);
            if (root_text.compare(tr("Captures")) == 0)
            {
                // We already have it, so we can leave.
                return;
            }
        }

        // Build up our tree item for display.
        tree_item = new FiTreeWidgetItem(ui->cam_tree);
        tree_item->setText(COLUMN_TITLE, tr("Captures"));
        tree_item->setIcon(COLUMN_TITLE, folder_icon);
        root_items.append(tree_item);

        // Add the root node in our file listing database.
        cam_folders.append(new LISTING_INFO);       // Captures
    }
    else
    {
        // Check to see if we have a capture entry already.
        for (int i = 0; i < root_items.count(); ++i)
        {
            QString root_text = root_items.at(i)->text(COLUMN_TITLE);
            if (root_text.compare(tr("Captures")) == 0)
            {
                // Remove this entry.
                tree_item = root_items.at(i);
                root_items.removeAt(i);
                delete tree_item;
            }
        }
    }
}

/*!
 * @brief Specify the IP address of the camera we're to work with.
 * @param ip_addr   The IP of our camera.
 */
void CamExplore::setCamera(CameraMDI *camera_ptr)
{
    FI_comm *comm_ptr;

    // Remember the camera pointer.
    this->camera_ptr = camera_ptr;
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    cam_ip_addr = camera_ptr->get_ip_addr();

    // Load the current settings associated with the IP address.
    download_base_path = SaveDialog::loadPath(cam_ip_addr);

    // Reset the selected device.
    last_device.clear();

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Now that we have an address, let's connect the signal/slots to that object.
    connect(comm_ptr, &FI_comm::camera_offline_status_changing, this, &CamExplore::cameraOffline, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &CamExplore::msgTimeoutEvent, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &CamExplore::readComplete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &CamExplore::writeComplete, Qt::QueuedConnection);
    connect(camera_ptr, &CameraMDI::closing, this, &CamExplore::handleCameraClosing);
    connect(camera_ptr, &CameraMDI::media_info_ready, this, &CamExplore::handleJsonMedia, Qt::QueuedConnection);

    // Query the camera for some registers.
    QTimer::singleShot(100, this, &CamExplore::registerPoll);
}

/*!
 * @brief Specify the destination media object with a JSON object.
 *
 * This is an optional function, used in post-series3 cameras.
 *
 * @param json_info     The JSON information we can use for the destination media.
 */
void CamExplore::setJsonInfo(const QJsonObject& json_info)
{
    QString title;

    // Copy our new media information object.
    media_info = json_info;

    // Make sure we got something.
    if (media_info.isEmpty())
        return;

    // Parse some of the information we'll need from the JSON data.
    media_type = media_info["type"].toInt();
    title = tr("Camera Explorer - ") + media_info["label"].toString();

    // Update our title.
    setWindowTitle(title);
}

/*!
 * @brief Specify the media we're interested in (SDcard, SSD, USB).
 * @param sel_type  One of the GIGE_MEDIA_xxx types.
 */
void CamExplore::setMediaType(int sel_type)
{
    QString title;

    // Save the type the user has selected.
    media_type = sel_type;

    // Use that information to make a useful title.
    switch (media_type)
    {
    case GIGE_MEDIA_SDCARD:
        title = QString(tr("Camera Explorer - SDCard"));
        break;
    case GIGE_MEDIA_SSD:
        title = QString(tr("Camera Explorer - SSD"));
        break;
    case GIGE_MEDIA_USB:
        title = QString(tr("Camera Explorer - USB"));
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unknown media type:" << media_type;
        return;
    }

    // Update our titile.
    setWindowTitle(title);
}

/*!
 * @brief State machine for the CAP file transfer operation.
 * @param event The event driving a state transition.
 */
void CamExplore::stateCapXfer(QEvent *event)
{
    ExploreEvent    *our_event;

    // Cast the event to the ExploreEvent type.
    our_event = (ExploreEvent *)event;
    cur_state = our_event->get_state();

    // Figure out our current state.
    switch (cur_state)
    {
    case CAP_XFER_LOAD:
        // Make sure the camera is ready to process the next CAP file.
        capXferStartCheck();
        break;
    case CAP_XFER_SAVE:
        capXferImageSave();
        break;
    case CAP_XFER_COMPLETE:
        capXferFinish();
        break;
    default:
        QLOG_ERROR() <<  __PRETTY_FUNCTION__ << "Unknown state:" << cur_state;
        break;
    }
}


/*!
 * @brief State machine for the copy operation.
 * @param event The event driving a state transition.
 */
void CamExplore::stateCopy(QEvent *event)
{
    ExploreEvent    *our_event;

    // Cast the event to the ExploreEvent type.
    our_event = (ExploreEvent *)event;
    cur_state = our_event->get_state();

    // Figure out our current state.
    switch (cur_state)
    {
    case COPY_SET_FILE:
        copyFileStart();
        break;
    case COPY_METAFILE:
        copyMetafile();
        break;
    case COPY_PAUSED:
        copyContinue();
        break;
    case COPY_WAIT:
        // Shouldn't actually enter here, we're waiting in our timing event handler
        // for the operation to finish.
        break;
    case COPY_COMPLETE:
        // Clean up the copy interface.
        copyComplete();
        break;
    case COPY_DIRECT:
        // Nothing going on here, all the work was done in the copyDirect() method.
        break;
    case COPY_FAULT:
    {
        QString copy_error_msg;
        copy_error_msg = copy_worker_thread->result();
        if (!copy_error_msg.isEmpty())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << copy_error_msg;

            // Hide the src/dest values.
            xferCopyDialog->showXferSrc(false);
            xferCopyDialog->showXferDest(false);

            // Set up the status.
            xferCopyDialog->showStatusIcon(true);
            xferCopyDialog->setStatusIcon(QStyle::SP_MessageBoxInformation);
            xferCopyDialog->setStatusValue(copy_error_msg);
            xferCopyDialog->showStatusValue(true);
            xferCopyDialog->showBar(false);
            QApplication::restoreOverrideCursor();  // A copy error has occurred we are no longer busy.
        }
        break;
    }
    default:
        QLOG_ERROR() <<  __PRETTY_FUNCTION__ << "Unknown state:" << cur_state;
        break;
    }
}


/*!
 * @brief State machine for the delete operation.
 * @param event The event driving a state transition.
 */
void CamExplore::stateDelete(QEvent *event)
{
    ExploreEvent    *our_event;

    // Cast the event to the ExploreEvent type.
    our_event = (ExploreEvent *)event;
    cur_state = our_event->get_state();

    // Figure out our current state.
    switch (cur_state)
    {
    case DELETE_SET_FILE:
        // Start the delete operation on the next entry in the sel_items list.
        deleteStart();
        break;
    case DELETE_WAIT:
        // Shouldn't actually enter here, we're waiting in our timing event handler
        // for the operation to finish.
        break;
    case DELETE_COMPLETE:
        // Clean up the delete interface.
        deleteFinish();

        // Force a refresh.
        doRefresh();
        break;
    default:
        QLOG_ERROR() <<  __PRETTY_FUNCTION__ << "Unknown state:" << cur_state;
        break;
    }
}

/*!
 * @brief State machine for the refresh (aka folder listing) operation.
 * @param event The event driving a state transition.
 */
void CamExplore::stateRefresh(QEvent *event)
{
    bool    do_download;
    ExploreEvent    *our_event;

    // Cast the event to the ExploreEvent type.
    our_event = (ExploreEvent *)event;
    cur_state = our_event->get_state();
    QString state_info_msg = getCurStateMsg();

    QLOG_TRACE() << __PRETTY_FUNCTION__ << state_info_msg;

    // Assume we'll want to download some data.
    do_download = true;

    // Figure out our current state.
    switch (cur_state)
    {
    case REFRESH_STILLS:
        QApplication::postEvent(this, new ExploreEvent(REFRESH_EVENT, REFRESH_STACKS));
        do_download = false;
        break;
    case REFRESH_STACKS:
        refresh_row = ROW_STACKS;
        break;
    case REFRESH_VIDEO:
        refresh_row = ROW_VIDEO;
        break;
    case REFRESH_CAPTURE:
        refresh_row = ROW_CAPTURE;
        break;
    case REFRESH_DOWNLOAD_COMPLETE:
        // Remove the progress dialog and generally cleanup after ourselves.
        listingFinish();
        do_download = false;

        // We're now idle.
        cur_state = EXPLORE_IDLE;
        break;
    default:
        QLOG_ERROR() <<  __PRETTY_FUNCTION__ << "Unknown state:" << cur_state;
        do_download = false;
        break;
    }

    // Do we need to do some downloading?
    if (do_download)
        listingStart(refresh_row);
}

/*!
 * @brief Handle the write completion processing of the FilePath register for the various state machines.
 * @param write_status  The write completion status.
 */
void CamExplore::writeFilepathDone(uint32_t write_status)
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Check for an error.
    if (write_status != GEV_STATUS_SUCCESS)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << camera_ptr->get_ip_addr_pretty() << "Failed to set file path to:" << media_path;
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "status:" << Qt::hex << Qt::showbase << write_status;

        // Clean up.
        QApplication::restoreOverrideCursor();
        if (progressDialog)
        {
            disconnect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CamExplore::handleProgressCancelledSignal);

            // Recycle any resources and release the progress dialog.
            fiProgressDialog::RecycleProgressDialog(progressDialog);
        }

        // All done.
        return;
    }

    // Continue processing based on the current state.
    switch (cur_state)
    {
    case REFRESH_STILLS:
    case REFRESH_STACKS:
    case REFRESH_VIDEO:
    case REFRESH_CAPTURE:
        // We've set the path on the camera we're interested in, now request a listing.
        camera_ptr->write_camera_register(GIGE_REG_FILE_ACTION, GIGE_FA_LISTING);
        break;
    case COPY_SET_FILE:
        copyStartAction();
        break;
    case DELETE_SET_FILE:
        // We've set the path on the camera we're interested in, now ask it to delete the file.
        camera_ptr->write_camera_register(GIGE_REG_FILE_ACTION, GIGE_FA_DELETE);
        cur_state = DELETE_WAIT;
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected state:" << cur_state;
        break;
    }
}

/*!
 * @brief Handle the write completion processing of the FastecCCP register for the various state machines.
 * @param write_status  The write completion status.
 */
void CamExplore::writeLockDone(uint32_t write_status)
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Was the write a success?
    if (write_status != GEV_STATUS_SUCCESS)
    {
        QHostAddress    our_addr;
        our_addr.setAddress(qFromBigEndian(cam_ip_addr));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << our_addr.toString() << "Failed to acquire FastecCCP!";

        // Clean up.
        QApplication::restoreOverrideCursor();
        if (progressDialog)
        {
            disconnect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CamExplore::handleProgressCancelledSignal);
            // Recycle any resources and release the progress dialog.
            fiProgressDialog::RecycleProgressDialog(progressDialog);
        }

        // All done.
        return;
    }

    // Toggle our lock status - we either got or released the lock.
    we_own_lock = !we_own_lock;

    // If we no longer have the lock, then we're done.
    if (!we_own_lock)
        return;

    // Continue processing based on the current state.
    switch (cur_state)
    {
    case COPY_STARTUP:
    case COPY_PAUSED:
        QApplication::postEvent(this, new ExploreEvent(COPY_EVENT, COPY_SET_FILE));
        break;
    case DELETE_STARTUP:
        QApplication::postEvent(this, new ExploreEvent(DELETE_EVENT, DELETE_SET_FILE));
        break;
    case REFRESH_STARTUP:
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Got the lock for REFRESH_STARTUP" << "cap_file_load:" << cap_file_load << "captures_allowed:" << captures_allowed;

        // If we're only looking for CAP files, then that's where we want to start.
        if (cap_file_load && captures_allowed)
            QApplication::postEvent(this, new ExploreEvent(REFRESH_EVENT, REFRESH_CAPTURE));
        else
            QApplication::postEvent(this, new ExploreEvent(REFRESH_EVENT, REFRESH_STILLS));
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected state:" << cur_state;
        break;
    }
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief An event happened, presumably to drive one of our state machines.
 * @param event     Pointer to the event structure.
 */
void CamExplore::customEvent(QEvent *event)
{
    // Only respond to our custom event. We also want to leave if we're in shutdown.
    switch ((int)event->type())
    {
    case CAP_XFER_EVENT:
        stateCapXfer(event);
        break;
    case COPY_EVENT:
        stateCopy(event);
        break;
    case DELETE_EVENT:
        stateDelete(event);
        break;
    case REFRESH_EVENT:
        stateRefresh(event);
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected event:" << event;
        break;
    }
}

/*!
 * @brief The dialog is getting "shown". A perfect time to start the first refresh.
 * @param event     Poiner to the actual event object.
 */
void CamExplore::showEvent(QShowEvent *event)
{
    // Ask the camera to update the media list.
    if (camera_ptr != nullptr)
        camera_ptr->start_camera_read_reg(GIGE_REG_AVAIL_MEDIA);

    // Pass it on.
    QDialog::showEvent(event);
}

/*!
 * @brief A timer has fired. Used for monitoring the progress of long operations.
 * @param event The actual timer event
 */
void CamExplore::timerEvent(QTimerEvent *event)
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Is this our one of our file action, live video state change or
    // review state change timers?
    if (event->timerId() == file_timer_id)
    {
        FI_comm     *comm_ptr;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
            return;
        }

        // Poll the file action status.
        comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_FILE_ACTION_STATUS);

        // Stop the timer until we get a response.
        killTimer(file_timer_id);
        file_timer_id = 0;
    }
    else if (event->timerId() == live_timer_id)
    {
        // Cleanup the timer.
        killTimer(live_timer_id);
        live_timer_id = 0;

        // We'll call this a fail.
        cap_load_success = false;
        userCanceledCopy();
    }
    else if (event->timerId() == review_timer_id)
    {
        // Cleanup the timer.
        killTimer(review_timer_id);
        review_timer_id = 0;

        // We'll call this a fail.
        cap_load_success = false;
        userCanceledCopy();
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief Invoked when the communications object has changed the camera's status to off- or on-line.
 *
 * @sa FI_comm::camera_offline_status_changing()
 *
 * @param ip_addr   The IP address of the camera whose state is changing.
 * @param offline   True if the camera is going off-line, false if on-line.
 */
void CamExplore::cameraOffline(uint32_t ip_addr, bool offline)
{
    // Sanity check.
    if (camera_ptr == nullptr)
        return;

    // Is this for us?
    if (ip_addr != cam_ip_addr)
        return;

    // Did we go off-line?
    if (offline)
    {
        // Shut things down.
        userCanceled(__PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Catch the signal indicating that the file save has completed.
 */
void CamExplore::capXferFileSaved()
{
    int new_state;
    SaveVideo   *saver_ptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our saver object.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
        return;
    }

    // Disconnect from the saver's signal.
    disconnect(saver_ptr, &SaveVideo::save_complete, this, &CamExplore::capXferFileSaved);

    // Are we done, or do we need to continue with CAP file transfers?
    if (sel_items.isEmpty())
    {
        new_state = CAP_XFER_COMPLETE;
    }
    else
    {
        // Switch to the new state.
        new_state = CAP_XFER_LOAD;
    }

    // Continue the CAP transfer process.
    QApplication::postEvent(this, new ExploreEvent(CAP_XFER_EVENT, new_state));
}

/*!
 * @brief Catch the "review_panel_ready_to_go" signal after the CAP file was loaded to provide
 * a mechanism to understand when the camera object should be ready to process a save operation.
 */
void CamExplore::capLoadedCamReviewPanelReady()
{
    camera_review_panel_ready = true;

    // We know the camera says the review panel is ready to go so disconnect the signal connection.
    disconnect(camera_ptr, &CameraMDI::review_panel_ready_to_go, this, &CamExplore::capLoadedCamReviewPanelReady);
}

/*!
 * @brief Catch the "load_complete" signal during CAP transfers and verify that the
 * camera has transitioned into review mode and that it is paused and ready to save.
 */
void CamExplore::capLoadedReviewCheck()
{
    uint32_t camera_state;

    // We know the load is complete now disconnect the signal connection.
    disconnect(this, &CamExplore::loadComplete, this, &CamExplore::capLoadedReviewCheck);

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // What is our current state.
    camera_state = camera_ptr->read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Is the camera in the review state?
    if (camera_state != GIGE_CAMERA_STATE_REVIEW)
    {
        // Get our communications object pointer.
        FI_comm *comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
            return;
        }

        // Setup our timer.
        review_timer_id = startTimer(REVIEW_DELAY);

        // Set our state.
        cur_state = CAP_XFER_WAIT_FOR_REVIEW;

        // Initialize our counter.
        wait_for_state_count = REVIEW_STATE_DELAY;

        // Read the state.
        comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_CAMERA_STATE);
    }
    else
    {
        // Sanity check.
        if (cur_state != CAP_LOAD)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "CamExplore::cur_state != CAP_LOAD!";
        }

        // Has the camera told the world that its review panel is ready?
        if (camera_review_panel_ready)
        {
            // We are good to go into the next phase of our batch download.
            capLoadedReviewReady();
        }
        else
        {
            // Let's recheck for the camera's review panel ready signal after a small wait.
            QTimer::singleShot(100, this, &CamExplore::capLoadedReviewCheck);
        }
    }
}

/*!
 * @brief This is a special condition to intercept the verification that a review state
 * is confirmed and paused for transitioning to the save video state.
 */
void CamExplore::capLoadedReviewReady()
{
    uint32_t delay_to_save = 50;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Make sure that the camera has fully read the pfmd.
    if (!camera_review_panel_ready)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "We didn't receive notification from the camera that its review panel was ready!";
    }

    // Make sure that no progressDialog is left on the screen.
    if (progressDialog)
    {
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Make sure our state is set up for the next step in the batch download process.
    cur_state = CAP_XFER_LOAD;

    // If series3 camera system match the delay in CameraReview::review_panel_ready for live->review transitions.
    if (camera_ptr->is_series3())
    {
        delay_to_save = 750;
    }

    // The next step is to start the image save process to complete the transfer.
    QTimer::singleShot(delay_to_save, this, &CamExplore::capXferImageLoaded);
}

/*!
 * @brief Starts the CAP transfer process after the CAP file was loaded into image memory.
 */
void CamExplore::capXferImageLoaded()
{
    FI_comm     *comm_ptr;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "cap_load_success:" << cap_load_success;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Was there an error?
    if (!cap_load_success)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Load failed...cleanup";

        // Clean up and bail.
        userCanceledCopy();
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // We can show the cancel button now.
    if (xferCopyDialog)
        xferCopyDialog->showCancel(true);

    if (cur_state != CAP_XFER_LOAD)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "CamExplore::cur_state != CAP_XFER_LOAD but set to" << cur_state << " resetting it!";
        cur_state = CAP_XFER_LOAD;
    }

    // Query the camera to figure out how big the capture is.  This should start the ball rolling for the save.
    comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_PART_NUM_FRAMES);
}

/*!
 * @brief We reuse the saver objects "buffer_release" signal as a means to monitor
 *        the progress of the copy operation.
 */
void CamExplore::copyBlobProgress(uint32_t bufferSeqNbr)
{
    SaveVideo   *saver_ptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (!xferCopyDialog)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No Copy Transfer Progress Dialog?!";
        return;
    }

    // Get our saver object.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
        return;
    }

    // If there are no blob counts this progress handler doesn't have any work
    // to do and could have been called for a stack copy.
    if (blob_count <= 0)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No blobs to process was this called by mistake?";
        return;
    }

    // Each time through here we are processing a blob of data.
    // This will keep a count of the blobs of data that have been processed here.
    // It will also be used to give an indication as to blob completion.
    // It does not matter about retrying frames or not - as missed frames do not
    // come through this process.
    // But we only do this if we have a valid blob because a blob with bufferSeqNbr
    // of 0 is an empty buffer and placeholder for a missed frame.  We let it pass
    // through when this happens.
    if (bufferSeqNbr != 0)
    {
        // Increment the number of files/blobs received.
        rcv_count++;

        // Do we care about progress?
        if (xferCopyDialog->getBarMax() != 0)
        {
            // Increment the progress.
            xferCopyDialog->incrementBar();
        }
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "COPY BLOB EMPTY FRAME"
                     << "bufferSeqNbr:" << bufferSeqNbr
                     << "rcv_count:" << rcv_count
                     << "missing:" << camera_ptr->get_missed_frames()
                     << "blob_count:" << blob_count
                     << "frame_count:" << saver_ptr->getFrameCount()
                     << "progress:" << xferCopyDialog->getBarValue()
                     << "max:" << xferCopyDialog->getBarMax()
                     << "retry_count:" << camera_ptr->get_retry_count();
    }

    // Update our transfer dialog with the file we are transfering.
    if (!saver_ptr->getImageFilename().isEmpty())
        xferCopyDialog->setXferDestValue(saver_ptr->getImageFilename());

    // If we haven't received all of our blobs do we need to spawn off a retry
    // process?
    if (!camera_ptr->retrying_frames() && (camera_ptr->get_retry_count() > 0))
    {
        // Have we processed all of our frames we can for this blob copy?
        if ((rcv_count + camera_ptr->get_retry_count()) == blob_count)
        {
#if 0
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "COPY BLOB SPAWNING RETRY"
                         << "bufferSeqNbr:" << bufferSeqNbr
                         << "rcv_count:" << rcv_count
                         << "missing:" << camera_ptr->get_missed_frames()
                         << "blob_count:" << blob_count
                         << "frame_count:" << saver_ptr->get_frame_count()
                         << "progress:" << xferCopyDialog->getBarValue()
                         << "max:" << xferCopyDialog->getBarMax()
                         << "retry_count:" << camera_ptr->get_retry_count();
#endif
            // Retry missing blobs.
            camera_ptr->retry_missing_frames();
        }
    }

    // Are we done transferring the file?
    // The rcv_count should equal the exact number of blobs that have been
    // processed through the save system for the copy stream.
    if (rcv_count == blob_count)
    {
        bool stillHasMetaFiles = false;
#if 0
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "COPY BLOB COMPLETE"
                     << "bufferSeqNbr:" << bufferSeqNbr
                     << "rcv_count:" << rcv_count
                     << "missing:" << camera_ptr->get_missed_frames()
                     << "blob_count:" << blob_count
                     << "frame_count:" << saver_ptr->get_frame_count()
                     << "progress:" << xferCopyDialog->getBarValue()
                     << "max:" << xferCopyDialog->getBarMax()
                     << "retry_count:" << camera_ptr->get_retry_count();
#endif

        // The maximum size of the progress bar is our expected blobs to process.
        // There could still be data to process if the bar isn't at maximum.
        if (xferCopyDialog->getBarValue() < xferCopyDialog->getBarMax())
        {
            stillHasMetaFiles = true;
        }

        // Close the video file we have been working on.
        saver_ptr->copy_finished();

        // Do we still have metafiles to pull down?
        if (stillHasMetaFiles)
        {
            // Disconnect from the saver's signal.
            disconnect(saver_ptr, &SaveVideo::saverBufferReleaseSignal, this, &CamExplore::copyBlobProgress);
        }

        // Connect the "paused" signal.
        connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::streamPaused);

        // We're done copying that video and it's associated metafiles.
        // By pausing the video streamer the copy state machine will process
        // the next file in the copy request based on this pause operation.
        // See CamExplore::stream_paused for details.
        camera_ptr->pause_video_request();
    }
}

/*!
 * @brief React to the user selecting a device.
 * @param index The index of the entry in the combobox.
 */
void CamExplore::doDeviceSelected(int index)
{
    QJsonObject dev_object;

    //QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Check for nothing selected.
    if (index < 0)
    {
        // Disable everything and leave.
        ui->pb_format->setEnabled(false);
        return;
    }

    // Are we series3?
    if (camera_ptr->is_series3())
    {
        QString sel_device;
        uint32_t avail_media;

        // Make sure we hide the eject button. Series3 cameras don't support this.
        ui->tb_eject->setVisible(false);

        // Read the available media register.
        avail_media = camera_ptr->read_camera_cache(GIGE_REG_AVAIL_MEDIA);

        // Get the selected text.
        sel_device = ui->cb_devices->itemText(index);

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Series3 avail_media:" << Qt::hex << Qt::showbase << avail_media << "sel_device:" << sel_device;

        // Compare it with known values.
        if (sel_device.compare(CamExplore::SDCARD_LABEL) == 0)
        {
            // Update our format & explore buttons.
            ui->pb_format->setEnabled(avail_media & GIGE_MEDIA_EXISTS_SDCARD);
            media_type = GIGE_MEDIA_SDCARD;
        }
        else if (sel_device.compare(CamExplore::SSD_LABEL) == 0)
        {
            // Update our format & explore buttons.
            ui->pb_format->setEnabled(avail_media & GIGE_MEDIA_EXISTS_SSD);
            media_type = GIGE_MEDIA_SSD;
        }
        else if (sel_device.compare(CamExplore::USB_LABEL) == 0)
        {
            // Update our format & explore buttons.
            ui->pb_format->setEnabled(avail_media & GIGE_MEDIA_EXISTS_USB);
            media_type = GIGE_MEDIA_USB;
        }

        // Determine if capture files are allowed on this camera.
        setCapturesAllowed((media_type == GIGE_MEDIA_SSD) && camera_ptr->cap_is_allowed());
    }
    else
    {
        bool    enable_ok;
        uint32_t dev_type;

        // Get the device type field.
        dev_object = ui->cb_devices->itemData(index).toJsonObject();
        dev_type = (uint32_t)dev_object["type"].toInt();

        // See if we have something mounted.
        if (dev_object["mountdir"].isNull() || dev_object["mountdir"].toString().isEmpty())
            enable_ok = false;
        else
            enable_ok = true;

        // Compare it with known values.
        if (dev_type == GIGE_MEDIA_SDCARD)
        {
            // Update our format & explore buttons.
            ui->pb_format->setEnabled(true);

            // Can't eject an SDCard.
            ui->tb_eject->setVisible(false);
        }
        else if (dev_type == GIGE_MEDIA_SSD)
        {
            // Update our format & explore buttons.
            ui->pb_format->setEnabled(true);

            // We may need to change this to support external vs. internal SSDs.
            ui->tb_eject->setVisible(false);
        }
        else if (dev_type == GIGE_MEDIA_USB)
        {
            // Update our format & explore buttons.
            ui->pb_format->setEnabled(true);

            // Only eject if there is something to eject.
            ui->tb_eject->setVisible(enable_ok);
            ui->tb_eject->setEnabled(enable_ok);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected device type:" << dev_type;
        }

        // Pass the media specific information to the dialog.
        setJsonInfo(dev_object);

        // Allow CAP files if the camera allows them.
        setCapturesAllowed(camera_ptr->cap_is_allowed());
    }

    // Refresh our listing if the device changed.
    if (last_device.compare(ui->cb_devices->itemText(index)) != 0)
    {
        QSettings settings;

        // Are we series3 or do we have JSON information for the device?
        if (camera_ptr->is_series3() || (!dev_object.isEmpty() && !dev_object["mountdir"].toString().isEmpty()))
        {
            // Start refreshing the listing.
            doRefresh();
            ui->pb_refresh->setEnabled(true);
            ui->cam_tree->setEnabled(true);
            ui->grp_options->setEnabled(true);
        }
        else
        {
            // This is a non-mounted drive. Make sure we clear things out.
            ui->lbl_space->clear();
            ui->pb_refresh->setEnabled(false);
            ui->cam_tree->setEnabled(false);
            doOptions(false);
            ui->grp_options->setEnabled(false);
            ui->grp_options->setChecked(false);
            clearFolders();
        }

        // Remember this as the last selected device.
        last_device = ui->cb_devices->currentText();

        // Save this as our last selected device.
        settings.beginGroup(CamExplore::CAMERA_EXPLORE_GROUP_NAME);
        settings.setValue(CamExplore::LAST_DEVICE, last_device);
        settings.endGroup();
    }

    // Make sure we enable our device combobox if there is anything in there.
    ui->grp_devices->setEnabled(ui->cb_devices->count() > 0);
}

/*!
 * @brief The copy worker thread may output individual file progression when copying
 * a file. This will update the progress dialog with the new value.
 */
void CamExplore::copyWorkerFileProgress(int progress, const QString& fileName)
{
    if (!cancel_in_progress)
    {
        xferCopyDialog->setBarValue(progress);
    }
    else
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "USER HAS CANCELED! Checking fileName key:" << fileName << "List:" << FiFile::copyFileList;
        if (FiFile::copyFileList.contains(fileName))
        {
            // Set the static value true as this copy should be cancelling.
            FiFile::copyFileList[fileName] = true;
        }
    }
}

/*!
 * @brief We reuse the saver objects "buffer_release" signal as a means to monitor
 *        the progress of the copy operation.
 */
void CamExplore::copyStackProgress(uint32_t bufferSeqNbr)
{
    bool should_be_complete = false;
    SaveVideo   *saver_ptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (!xferCopyDialog)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No Copy Transfer Progress Dialog?!";
        return;
    }

    // Get our saver object.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No SaveVideo object!";
        return;
    }

    // NOTE:  This is called everytime an image buffer is released.  THESE include
    // missed frame buffers as they are released through the system in the same manner.

    // Any buffer sequence number of zero is ignored.
    if (bufferSeqNbr == 0)
        return;

    // Have we initialized the expected receive count?
    // One valid frame has to come through here to set this up.
    if (rcv_expected == -1)
    {
        // Calculate the total expected received frames (include missing accumulator).
        if (blob_count > 0)
        {
            rcv_expected = blob_count + (int)copyMetafileCount;
        }
        else
        {
            rcv_expected = (int)saver_ptr->getFrameCount() + (int)copyMetafileCount;
        }
    }

    // Increment the receive count.
    rcv_count++;

    // Is this a metadata buffer release?
    if (FiSaver::metadataSeqNum(bufferSeqNbr))
    {
        rcv_metadata_count++;
    }

#if 0
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "COPY PROGRESS UPDATE"
                 << "bufferSeqNbr:" << bufferSeqNbr
                 << "rcv_count:" << rcv_count
                 << "missing:" << camera_ptr->get_missed_frames()
                 << "blob_count:" << blob_count
                 << "frame_count:" << saver_ptr->get_frame_count()
                 << "progress:" << xferCopyDialog->getBarValue()
                 << "max:" << xferCopyDialog->getBarMax()
                 << "retry_count:" << camera_ptr->get_retry_count();
#endif

    // Are all metafile buffers accounted for?
    if (rcv_metadata_count >= (int)copyMetafileCount)
    {
        // At the point of handling all of the metadata we now look to
        // handle any missing frames that may have occurred.
        if (!rcv_retrying && (camera_ptr->get_retry_count() > 0))
        {
            // The camera shouldn't already be retrying frames.  If so then
            // another process has started the operation.  Flag it in the log.
            if (camera_ptr->retrying_frames())
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Another process had started to retry frames for this Copy operation!";
            }

            // Copy operation is now retrying frames.
            rcv_retrying    = true;

            // Reset the recieve count and change expected to the retry count.
            rcv_count       = 0;
            rcv_expected    = camera_ptr->get_retry_count();
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Transitioning to Retrying Frames:  rcv_expected:" << rcv_expected;
            camera_ptr->retry_missing_frames();
        }
    }

    //Make sure our xfer dialog is adjusted if necessary.
    if (xferCopyDialog->getBarMax() != rcv_expected)
    {
        xferCopyDialog->setBarMax(rcv_expected);
    }

    // Mark the current progress in our xfer dialog.
    xferCopyDialog->setBarValue(rcv_count);

    // Are we retrying frames?  Give the user some more infomation.
    if (rcv_retrying)
    {
        QString retry_message = QString::fromUtf8("Copied missed frame [%1]").arg(bufferSeqNbr);
        xferCopyDialog->showStatusIcon(true);
        xferCopyDialog->setStatusIcon(QStyle::SP_MessageBoxInformation);

        // Set up the status.
        xferCopyDialog->setStatusValue(retry_message);
        xferCopyDialog->showStatusValue(true);

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << retry_message << "rcv_count:" << rcv_count << "rcv_expected:" << rcv_expected << "bufferSeqNum:" << bufferSeqNbr;
    }
    else
    {
        xferCopyDialog->showStatusIcon(false);
        xferCopyDialog->showStatusValue(false);
    }

    // Determine if we should be complete.
    if (rcv_count >= rcv_expected)
    {
        should_be_complete = true;

        // Are we retrying?
        if (rcv_retrying)
        {
            if (camera_ptr->get_retry_count() > 0)
            {
                should_be_complete = false;
            }
        }
    }

    // Should copy be complete?
    if (should_be_complete)
    {
        // Tell the saver this copy is finished.
        saver_ptr->copy_finished();

        // Connect the "paused" signal.
        connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::streamPaused);

        // We're done copying that video and it's associated metafiles. Make
        // sure the video is paused before moving on.
        camera_ptr->pause_video_request();
    }   // End of "if (total_rcv >= expected_rcv)"
}

/*!
 * @brief The user has pressed the Copy button. Start the state machine to
 *        copy the selected images/stacks/etc. to a user specified destination.
 */
void CamExplore::doCopy()
{
    int     num_stills;
    int     num_stacks;
    int     num_videos;
    int     num_caps;
    int     num_roots;
    bool    convert_found;
    bool    incomplete_found;
    const FI_COMM::FI_CAMERA_INFO   *camera_info;
    SaveVideo   *saver_ptr = nullptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our saver object so we can connect to any type of save error that occurs.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << tr("Camera does not have a saver object!");
        return;
    }

    // Get the list of items the user has selected.
    sel_items.clear();
    sel_items = ui->cam_tree->selectedItems();
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected empty selection list!";
        return;
    }

    // Find out what sort of items have been selected.
    getSelectionCounts(num_roots, num_stills, num_stacks, num_videos, num_caps, convert_found, incomplete_found);

    // If we received any incomplete files in our list we can't copy the lot.
    if (incomplete_found)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected incomplete files in selection list!";
        return;
    }

    // Reset the saver object to prepare for the copy/save operation.
    saver_ptr->reset();

    // We do things different for series3 cameras.
    if (camera_ptr->is_series3())
    {
        // If the user is copying CAP files then they are only allowed to save to the PC and we pop
        // up a different dialog to ask them how they want to load and save the CAP files.
        if (num_caps > 0)
        {
            // Setup for a CAP file transfer. Normally this results in a transform from the CAP
            // file to a saved file type. If we're just downloading the CAP file itself, this
            // will return false.
            if (capFileTransfer())
                return;
        }

        // Have the user select a destination.
        getDestination(num_videos);

        // Get a string for our destination.
        switch (copy_destination)
        {
        case DEST_SDCARD:
            copy_dest_path = DEST_SDCARD_STR.toUpper();
            break;
        case DEST_SSD:
            copy_dest_path = DEST_SSD_STR.toUpper();
            break;
        case DEST_USB:
            copy_dest_path = DEST_USB_STR.toUpper();
            break;
        case DEST_PC_PATH:
            copy_dest_path = download_base_path;
            break;
        default:
            break;
        }
    }
    else
    {
        // Have the user select a destination.
        getDestination();

        // If the user is copying CAP files then they are only allowed to save to the PC and we pop
        // up a different dialog to ask them how they want to load and save the CAP files.
        //
        // ALM_TODO: CAPs and CAP stacks both have to be treated "special".
        if ((num_caps > 0) && (copy_destination == DEST_PC_PATH))
        {
            // Setup for a CAP file transfer. Normally this results in a transform from the CAP
            // file to a saved file type. If we're just downloading the CAP file itself, this
            // will return false.
            if (capFileTransfer())
                return;
        }
    }

    // Did the user cancel?
    if (copy_destination == DEST_NONE)
        return;

    // Create our copy transfer progress dialog.
    if (createXferCopyDialog(FI_SAVER::STREAM_TYPE::NO_STREAM, copy_dest_path))
    {
        // Are we creating an AVI?
        if (convert_found && ui->chk_convert->isChecked())
        {
            // Add some text so the user knows we're converting.
            xferCopyDialog->setStatusValue(tr("Converting to AVI"));
            xferCopyDialog->showStatusValue(true);
        }
        else
        {
            // Hide the extra info control(s).
            xferCopyDialog->showStatusValue(false);
        }

        // Modify the min/max.
        xferCopyDialog->setBarMinMax(0, 0);

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Showing xferCopyDialog with NO POLLING!";

        // Show the dialog without polling.
        xferCopyDialog->showDialogWithoutPolling();
    }

    // Initialize our flags.
    cancel_in_progress = false;

    // Lock access to the camera and start the copy process.
    we_own_lock = false;
    cur_state = COPY_STARTUP;

    // Remember our current "paused" state.
    camera_info = camera_ptr->get_camera_info();
    video_was_running = camera_info->stream && camera_info->stream_running;

    // Are we copying locally?
    if (copy_destination == DEST_PC_PATH)
    {
        // Check to see if we are talking to a series3 camera, or we are converting to an AVI.
        if (camera_ptr->is_series3() || (convert_found && ui->chk_convert->isChecked()))
        {

            // We want to know when the video stream is paused.
            connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::streamPaused);

            // Connect the camera's error signal.
            connect(camera_ptr, &CameraMDI::copy_timeout, this, &CamExplore::handleCopyTimeout, Qt::UniqueConnection);

            // Connect to the camera's save error reported signal.
            connect(camera_ptr, &CameraMDI::save_error_reported, this, &CamExplore::handleSaverErrorReported, Qt::UniqueConnection);

            // Make sure we pause all camera video. This will start our state machine rolling.
            MainWindow::GetInstance()->pause_video_streaming();

            // If we weren't running, we need to force the "pause" event.
            if (!video_was_running)
                streamPaused();
        }
        else
        {
            // In this mode, we can copy direct from the camera using the Samba interface.
            copyDirect();
        }
    }
    else
    {
        // No need to pause the video. By locking access, we'll start the copy process.
        camera_ptr->write_camera_register(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE);
    }
}

/*!
 * @brief The user has pressed the delete button. Have the camera delete all of the
 *        user selected images/stacks/etc.
 */
void CamExplore::doDelete()
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get the list of items the user has selected.
    sel_items.clear();
    sel_items = ui->cam_tree->selectedItems();
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected empty selection list!";
        return;
    }

    // Put up the progress dialog.
    progressDialog = fiProgressDialog::CreateSimpleProgressDialog(camera_ptr->get_camera_name(), tr("Deleting Files..."), QMessageBox::Cancel, __PRETTY_FUNCTION__);
    if (progressDialog)
    {
        connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CamExplore::handleProgressCancelledSignal, Qt::UniqueConnection);
        progressDialog->setBarMinMax(0, 0);
        progressDialog->showBar(true);
        progressDialog->showDialogWithoutPolling();
    }

    // Lock access to the camera.
    we_own_lock = false;
    cur_state = DELETE_STARTUP;
    camera_ptr->write_camera_register(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE);
}

/*!
 * @brief React to the user pressing the "eject" button.
 */
void CamExplore::doEject()
{
    QJsonObject dev_object;
    QJsonObject json_action;

    // Get the selected device information.
    dev_object = ui->cb_devices->currentData().toJsonObject();

    // Build up the action.
    json_action.insert("action", GIGE_MA_EJECT);
    json_action.insert("device", dev_object["name"].toString());

    // Send it on it's way.
    QJsonDocument action_doc(json_action);
    camera_ptr->write_camera_buffer(GIGE_REG_MA_JSON, action_doc.toJson());

    // Reset the device control
    ui->cb_devices->clear();
    last_device.clear();

    // Ask the camera to update the media list.
    camera_ptr->start_camera_read_reg(GIGE_REG_AVAIL_MEDIA);
}

/*!
 * @brief The user pressed the "format" button. Perform some sanity checks
 *        and, if everything looks OK, warn the user about the dangers of
 *        formatting a device.
 *
 *        If the user wants to proceed, then command the camera to start a
 *        format operation. We'll start up a timer to show activity to the user.
 */
void CamExplore::doFormat()
{
    bool        user_canceled;
    uint32_t    reg_value;
    QString     device_sel;
    QString     volume_label;
    QJsonObject dev_object;
    std::unique_ptr<FormatDlg> format_dlg;
    QMessageBox::StandardButton     user_response;
    const FI_COMM::FI_CAMERA_INFO   *info_ptr;

    // Get the camera information structure.
    info_ptr = camera_ptr->get_camera_info();
    if (info_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "What? No camera info structure?!";
        return;
    }

    // Allocate our dialog.
    format_dlg = std::unique_ptr<FormatDlg>(new FormatDlg(MainWindow::GetInstance().get()));
    if (!format_dlg)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate the dialog!";
        return;
    }

    // Tell the dialog if we're series3 or not.
    format_dlg->isSeries3(camera_ptr->is_series3());

    // Series3 cameras require different information.
    if (camera_ptr->is_series3())
    {
        // Figure out which media we're interested in.
        device_sel = ui->cb_devices->currentText();
        if (device_sel.compare(SDCARD_LABEL) == 0)
            format_dlg->media_type = GIGE_MEDIA_SDCARD;
        else if (device_sel.compare(SSD_LABEL) == 0)
            format_dlg->media_type = GIGE_MEDIA_SSD;
        else
            format_dlg->media_type = GIGE_MEDIA_USB;

        // Default the format type
        format_dlg->format_type = GIGE_FORMAT_FAT32;
    }
    else
    {
        // Retrieve the info from the selected object.
        dev_object = ui->cb_devices->currentData().toJsonObject();
        format_dlg->media_type = (uint32_t)dev_object["type"].toInt();

        // Let the dialog know if this is an internal device.
        format_dlg->internal_ssd = dev_object["internal"].toBool();

        // Default the format type
        format_dlg->format_type = GIGE_FORMAT_EXFAT;
    }

    // Is the SSD selected.
    if (format_dlg->media_type == GIGE_MEDIA_SSD)
    {
        uint32_t ssd_info;
        uint32_t ltr_mode;

        // See if the camera's SSD supports the sanitize command.
        ssd_info = camera_ptr->read_camera_cache(GIGE_REG_SSD_CAPABILITIES);
        ltr_mode = camera_ptr->read_camera_cache(GIGE_REG_LTR_MODE);

        // Make sure we're either a new camera or the series3 camera is A3+.
        if (camera_ptr->is_series3())
        {
            // Label for the internal SSD.
            format_dlg->vol_label = QString("FastecSSD");

            // Does the version of the camera support LTR mode?
            if ((info_ptr->board_rev[0] == FI_COMM::CAM_BOARD_REV_A) && (info_ptr->board_rev[1] >= FI_COMM::CAM_BOARD_REV_A3))
            {
                // Get the LTR mode register value.
                format_dlg->lrmode = (ltr_mode & ((uint32_t)GIGE_LTR_MODE_ENABLE)) == ((uint32_t)GIGE_LTR_MODE_ENABLE);
                if (format_dlg->lrmode)
                    format_dlg->format_type = GIGE_FORMAT_FILR;

                // Let the dialog know we can sanitize the drive.
                format_dlg->sanitize = ((ssd_info & GIGE_SSD_CAP_SANITIZE) == GIGE_SSD_CAP_SANITIZE);
            }
        }
        else
        {
            // On the new cameras, we can have external SSDs. Is this an internal or external drive?
            if (format_dlg->internal_ssd)
            {
                // Label for the internal SSD.
                format_dlg->vol_label = QString("FastecSSD");

                // Internal SSDs support LR mode. Get the LTR mode register value.
                format_dlg->lrmode = (ltr_mode & ((uint32_t)GIGE_LTR_MODE_ENABLE)) == ((uint32_t)GIGE_LTR_MODE_ENABLE);
                if (format_dlg->lrmode)
                    format_dlg->format_type = GIGE_FORMAT_FILR;
                else
                    format_dlg->format_type = GIGE_FORMAT_EXT2;
            }
            else
            {
                // Label for the external SSD.
                format_dlg->vol_label = QString("VideoSSD");

                // External SSDs do not support LR mode. Default to ExFAT.
                format_dlg->lrmode = false;
                format_dlg->format_type = GIGE_FORMAT_EXFAT;
            }

            // Let the dialog know we can sanitize the drive.
            format_dlg->sanitize = ((ssd_info & GIGE_SSD_CAP_SANITIZE) == GIGE_SSD_CAP_SANITIZE);
        }
    }
    else
    {
        // Pass the volume label to the dialog.
        if (format_dlg->media_type == GIGE_MEDIA_USB)
            format_dlg->vol_label = QString("FastecUSB");
        else if (format_dlg->media_type == GIGE_MEDIA_SDCARD)
            format_dlg->vol_label = QString("FastecSD");
        else
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected media_type:" << Qt::showbase << Qt::hex << format_dlg->media_type;
    }

    // Let the user decide how to format the data.
    if (format_dlg->exec() == QDialog::Rejected)
    {
        // Set the cancel flag.
        user_canceled = true;
    }
    else
    {
        // Grab the user's selections.
        volume_label = format_dlg->vol_label;

        // Make sure we clear the cancel flag.
        user_canceled = false;
    }

    // Are we done?
    if (user_canceled)
        return;

    QString drive_type;
    switch (format_dlg->media_type)
    {
    case GIGE_MEDIA_SDCARD:
        drive_type = tr("SD Card");
        break;
    case GIGE_MEDIA_SSD:
        drive_type = tr("Solid State Drive");
        break;
    case GIGE_MEDIA_USB:
        drive_type = tr("USB Flash Drive");
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unknown media type:" << format_dlg->media_type;
        return;
    }

    // Make sure the user knows what they're doing.
    user_response = QMessageBox::warning(MainWindow::GetInstance().get(), tr("Format Media"), tr("Formatting will erase all data on the camera's %1").arg(drive_type), QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Cancel);
    if (user_response == QMessageBox::Cancel)
        return;

    // The user is good to go. Build up the command register.
    reg_value = format_dlg->format_type;
    reg_value |= format_dlg->media_type;
    if (format_dlg->sanitize)
        reg_value |= GIGE_FORMAT_SANITIZE;

    // Reset our status.
    format_cmd_status = GEV_STATUS_SUCCESS;

    // If we are talking to a Series3 camera, then we're good to go. Otherwise, we still
    // have some work left.
    if (camera_ptr->is_series3())
    {
        // Write the register.
        camera_ptr->write_camera_register(GIGE_REG_FORMAT_MEDIA, reg_value);

        // Start our timer to poll the camera for formatting progress.
        QTimer::singleShot(FORMAT_POLLING_PERIOD, this, &CamExplore::formatPoll);
    }
    else
    {
        QJsonObject json_action;

        // Pause the video streaming while we're formatting.
        camera_ptr->pause_video_request();

        // Build up the action.
        json_action.insert("action", GIGE_MA_FORMAT);
        json_action.insert("device", dev_object["name"].toString());
        json_action.insert("format", (int)reg_value);
        json_action.insert("label", volume_label);

        // Send it on it's way. We will start polling once we've received our
        // write status.
        QJsonDocument action_doc(json_action);
        camera_ptr->write_camera_buffer(GIGE_REG_MA_JSON, action_doc.toJson());
    }

    // Disable the group.
    ui->grp_devices->setEnabled(false);
    ui->cam_tree->setEnabled(false);
    ui->pb_refresh->setEnabled(false);
    doOptions(false);
    ui->grp_options->setChecked(false);
    ui->grp_options->setEnabled(false);

    // Hide the format button and show the progress indicator.
    ui->pb_format->setVisible(false);
    ui->format_progress->setVisible(true);
}

/*!
 * @brief Load the user selected CAP file back into camera memory for processing.
 * The user has pressed the load button or they have double clicked on a CAP file
 * in the tree listing.  This function is only allowed for a CAP file.
 *
 */
void CamExplore::doLoad()
{
    int root_row;
    const QTreeWidgetItem       *selected_item;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get the list of items the user has selected.
    sel_items.clear();
    sel_items = ui->cam_tree->selectedItems();
    if (sel_items.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected empty selection list!";
        return;
    }

    // There can be only one!
    if (sel_items.count() != 1)
    {
        fiMsgBox::warning(this, tr("Only one partition can be selected when loading. Load Aborted."));
        return;
    }

    // Get the currently selected item.
    selected_item = ui->cam_tree->currentItem();
    if (!selected_item)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "What? No selected item?!";
        return;
    }

    // Convert the tree item into a file info.
    file_info = getFileInfo(selected_item, root_row);
    if (file_info == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "What? No file info for the selected item:" << selected_item->text(COLUMN_NAME);
        return;
    }

    // Are we "loading" or "opening"?
    if (ui->pb_load->text().compare(tr("Load"), Qt::CaseInsensitive) == 0)
    {
        // We're just loading - no CAP file copy here.
        cap_copy_in_progress = false;
        copy_destination = DEST_NONE;

        // Disable the cache refresh logic.
        camera_ptr->pause_cache_refresh(true);

        // When we get here, loading is done.
        connect(this, &CamExplore::loadComplete, this, &CamExplore::loadingFinish, Qt::QueuedConnection);

        // Start the loading process.
        capLoadStart(root_row);
    }
    else
    {
        QString full_path;

        // Try to build up a path to the selected item.
        full_path = buildSmbPath(file_info, root_row);
        if (!full_path.isEmpty())
        {
            // Ask our main window to open it for us.
            MainWindow::GetInstance()->load_file(full_path, true);

            // Close our window.
            close();
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to build a full path to the selected item:" << file_info->name;
        }
    }
}

/*!
 * @brief The user has pressed the move button. Copy the selected items to the user
 *        specified destination and then remove them from the source.
 */
void CamExplore::doMove()
{
    // A move is nothing more than a copy followed by a delete.
    move_in_progress = true;
    doCopy();
}

void CamExplore::doOptions(bool clicked)
{
    // Show/hide the options group.
    ui->frm_options->setVisible(clicked);

    // If the group is hidden, make sure the options are de-selected.
    if (!clicked)
    {
        ui->chk_convert->setChecked(false);
        ui->chk_use_name->setChecked(false);
    }
}

/*!
 * @brief The user has pressed the refresh button. Start the process of downloading new
 *        listing files from the camera.
 */
void CamExplore::doRefresh()
{
    QJsonObject dev_object;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (cur_state != EXPLORE_IDLE)
        return;

    // Clean our folder display.
    clearFolders();

    // Determine if the selected device is "mountable", and therefore if we
    // can request listings from it. Or the camera is series3.
    dev_object = ui->cb_devices->currentData().toJsonObject();
    if (!camera_ptr->is_series3() && (dev_object.isEmpty() || dev_object["mountdir"].toString().isEmpty()))
        return;

    // Start the refresh processes. It can be rather lengthy, so show a busy cursor.
    QApplication::setOverrideCursor(Qt::BusyCursor);

    // Check for a progress dialog.
    progressDialog = fiProgressDialog::CreateSimpleProgressDialog(camera_ptr->get_camera_name(), tr("Getting a file listing from the camera"), QMessageBox::Cancel, __PRETTY_FUNCTION__);
    if (progressDialog)
    {
        connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CamExplore::handleProgressCancelledSignal, Qt::UniqueConnection);
        progressDialog->showDialogWithoutPolling();

        // There are 3 steps normally (Stills, Stacks, Videos), or 4 if there are CAPs allowed.
        progressDialog->setBarMinMax(0, (captures_allowed ? 4 : 3));
        progressDialog->setBarValue(0);
    }

    // Lock access to the camera.
    we_own_lock = false;
    cur_state = REFRESH_STARTUP;
    camera_ptr->write_camera_register(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE);
}

/*!
 * @brief When the user expands the tree, make sure the columns are fully adjusted.
 * @param model_index   Which item was expanded.
 */
void CamExplore::doTreeExpansion(QModelIndex model_index)
{
    Q_UNUSED(model_index)

    // Resize the column.
    ui->cam_tree->resizeColumnToContents(COLUMN_NAME);
}

/*!
 * @brief This slot is invoked with the QTreeWidget emits a currentItemChanged signal. We
 *        use this to update our metadata display.
 *
 * The signal is emitted when the "current item" changes. The "current item" is specified
 * by the "item_ptr" parameter, and it replaces the "prev_ptr" item (which we don't use).
 *
 * @param item_ptr  The new "current" item in the widget.
 */
void CamExplore::doTreeItemChanged(QTreeWidgetItem *item_ptr, QTreeWidgetItem * /* prev_ptr */)
{
    int     root_row;

    // Assume the user can't do certain things until proven differently.
    ui->pb_refresh->setEnabled(false);

    // If the current item is a root row then enable refresh.
    root_row = getRootRow(item_ptr);
    if (root_row >= 0)
    {
        // If the current row is a top level row then enable refresh
        ui->pb_refresh->setEnabled(true);
        ui->txt_meta->clear();
    }
    else
    {
        const FILE_INFO *file_info;       // the meta data associated with the item.

        // Now we want to format and display the meta data.
        file_info = getFileInfo(item_ptr, root_row);
        displayMetadata(file_info);
    }
}

/*!
 * @brief Handle double click on items in the tree.
 * @param item      Unused.
 * @param column    Unused.
 */
void CamExplore::doTreeItemDoubleClicked(QTreeWidgetItem *item, int column)
{
    Q_UNUSED(item)
    Q_UNUSED(column)

    // Only interested in double clicks if we're loading CAP files.
    if (!cap_file_load)
        return;

    // Make sure the load button is enabled.
    if (ui->pb_load->isEnabled())
    {
        // Act like the user clicked the Load button.
        QTimer::singleShot(0, this, &CamExplore::doLoad);
    }
}

/*!
 * @brief Called when the user selects a new item in the tree widget. This works in concert with the
 *        do_tree_item_changed() method which handles when the user changes the current item.
 */
void CamExplore::doTreeSelectionChanged()
{
    int     num_selected;
    int     num_stills;
    int     num_stacks;
    int     num_videos;
    int     num_caps;
    int     num_roots;
    bool    convert_found;
    bool    incomplete_found;
    uint32_t reg_value;

    // Assume the user can't do certain things until proven differently.
    ui->pb_load->setEnabled(false);
    ui->pb_move->setEnabled(false);
    ui->pb_copy->setEnabled(false);
    ui->pb_delete->setEnabled(false);
    ui->chk_convert->setEnabled(false);
    ui->chk_use_name->setEnabled(false);

    // Make sure we reset the load button's label.
    ui->pb_load->setText(tr("Load"));

    // Determine how many items are selected. If there aren't any, then leave.
    num_selected = ui->cam_tree->selectedItems().count();
    if (num_selected == 0)
        return;

    // Find out what sort of items have been selected.
    getSelectionCounts(num_roots, num_stills, num_stacks, num_videos, num_caps, convert_found, incomplete_found);

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "incomplete_found:" << incomplete_found;

    // If more than one "root" row has been selected, then leave.
    if (num_roots > 1)
        return;

    // Do we have exactly one "root" row?
    if (num_roots == 1)
    {
        // Allow the user to refresh the information.
        ui->pb_refresh->setEnabled(true);

        // There's no meta data to show.
        ui->txt_meta->clear();
        return;
    }

    // Check LTR mode.
    reg_value = camera_ptr->read_camera_cache(GIGE_REG_LTR_MODE);

    // Check to see if the user has selected exactly one CAP file.
    if ((num_selected == 1) && (num_caps == 1) && ((reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) != (uint32_t)GIGE_LTR_MODE_ENABLE))
    {
        // We could load this.
        ui->pb_load->setEnabled(true);
    }

    // If we're only looking to load CAP files, then we don't need to go any further.
    if (cap_file_load)
        return;

    // Check to see if the user has selected exactly one CAP stack on a camera
    // that is not a series3 camera.
    if (camera_ptr && !camera_ptr->is_series3() && (num_selected == 1) && !incomplete_found)
    {
        int root_row;
        const QTreeWidgetItem  *selected_item = ui->cam_tree->selectedItems().at(0);

        // We know that we got a "real" row, so figure out what kind it is.
        getFileInfo(selected_item, root_row);

        // We are only interested in opening stacks, and only if a single
        // item is selected.
        if (root_row == ROW_STACKS)
        {
            // We could open this.
            ui->pb_load->setText(tr("Open"));
            ui->pb_load->setEnabled(true);
        }
    }
    else if (num_stacks > 1)
    {
        // More than one stack, we do not allow load/open.
        ui->pb_load->setEnabled(false);
    }
    else if (incomplete_found)
    {
        // Do not allow load/open with an incomplete.
        ui->pb_load->setEnabled(false);
    }

    // Determine if we are allowed to delete the selection. No deletes from an A3+ SSD.
    if ((camera_ptr && camera_ptr->is_series3() && camera_ptr->cap_is_allowed()) && (media_type == GIGE_MEDIA_SSD))
        ui->pb_delete->setEnabled(false);
    else
        ui->pb_delete->setEnabled(true);

    // Only allow if no CAP files are selected.
    ui->chk_use_name->setEnabled(num_caps == 0);

    // For copy/move operations, the selected items have to be all of the same type.
    // Count the number of types selected.
    num_selected = 0;
    if (num_stills > 0)
        ++num_selected;
    if (num_stacks > 0)
        ++num_selected;
    if (num_videos > 0)
        ++num_selected;
    if (num_caps > 0)
        ++num_selected;

    // Do we have more than one type?
    if (num_selected > 1)
        return;

    // We now know that the selections are all of one type.
    // Only enable the copy button if there were no incomplete status settings.
    ui->pb_copy->setEnabled(!incomplete_found);

    // Moves, like deletes, aren't allowed on A3+ SSDs.
    if ((camera_ptr && camera_ptr->is_series3() && camera_ptr->cap_is_allowed()) && (media_type == GIGE_MEDIA_SSD))
        ui->pb_move->setEnabled(false);
    else
        ui->pb_move->setEnabled(!incomplete_found);

    // if there are any BMP or JPG stacks in selection list then enable the convert check box
    if (convert_found)
    {
        // Only enable the convert checkbox if we're looking at either a JPEG or BMP file.
        ui->chk_convert->setEnabled(!incomplete_found);
    }
    else
    {
        // BUGFIX #274 - let's disable anytime we are not including a convertable stack.
        ui->chk_convert->setEnabled(false);
    }
}

/*!
* @brief Slot handler for camera closing signal.
*/
void CamExplore::handleCameraClosing()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;
    userCanceled(__PRETTY_FUNCTION__);
}

/*!
 * @brief Slot handler for copy timeout.
 */
void CamExplore::handleCopyTimeout()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;
    userCanceledCopy();
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_MEDIA_JSON value.
 */
void CamExplore::handleJsonMedia(const QJsonDocument& jsonMedia)
{
    int found_index;
    int item_index;
    double dev_size;
    QString base_label;
    QString device_label;
    QString device_sel;
    QJsonObject mediaObj;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "jsonMedia:" << jsonMedia;

    // Don't update if the popup is displayed.
    if (ui->cb_devices->view()->isVisible())
        return;

    // Don't update if we are in the middle of formatting.
    if (format_in_progress)
        return;

    // Get the object from the document.
    mediaObj = jsonMedia.object();

    // Block updates from the combobox so there aren't update flashes.
    ui->cb_devices->blockSignals(true);

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
    {
        // Save the selected devices information.
        device_sel = ui->cb_devices->currentText();
    }
    else
    {
        QSettings settings;

        // Try using the previous device.
        settings.beginGroup(CamExplore::CAMERA_EXPLORE_GROUP_NAME);
        device_sel = settings.value(CamExplore::LAST_DEVICE, QString()).toString();
        settings.endGroup();

        // Make sure we get a listing update.
        force_update = true;
    }

    // Clear the combobox.
    ui->cb_devices->clear();

    // Add back in found devices starting with the SDCard.
    if (mediaObj.contains("sdcard") && mediaObj["sdcard"].isArray())
    {
        QJsonArray dev_array = mediaObj["sdcard"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // Remember this is an sdcard device.
            dev_object["type"] = GIGE_MEDIA_SDCARD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("SDcard %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // See if we already have a device with this name.
            found_index = findDevice(dev_object);
            if (found_index >= 0)
            {
                QJsonObject found_object;

                // It is likely that the found object is either the formatted partition
                // or the raw device. We prefer the formatted partition.
                if (!dev_object["mountdir"].toString().isEmpty())
                {
                    // The new object has a mount folder, so the one in the combobox
                    // needs to be replaced.
                    ui->cb_devices->removeItem(found_index);

                    // Add the device to the combobox.
                    ui->cb_devices->addItem(device_label, dev_object);
                }
            }
            else
            {
                // Add the device to the combobox.
                ui->cb_devices->addItem(device_label, dev_object);
            }
        }
    }

    // Check for NVME SSD devices.
    if (mediaObj.contains("nvme") && mediaObj["nvme"].isArray())
    {
        QJsonArray dev_array = mediaObj["nvme"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // Remember this is an NVME device.
            dev_object["type"] = GIGE_MEDIA_SSD;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                if (dev_object["internal"].toBool())
                    device_label = QString("SSD %1GB [Internal]").arg(dev_size, 0, 'f', 0);
                else
                    device_label = QString("SSD %1GB [External]").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                if (dev_object["internal"].toBool())
                    device_label = base_label + QString(" [%1GB] [Internal]").arg(dev_size, 0, 'f', 0);
                else
                    device_label = base_label + QString(" [%1GB] [External]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // See if we already have a device with this name.
            found_index = findDevice(dev_object);
            if (found_index >= 0)
            {
                // It is likely that the found object is either the formatted partition
                // or the raw device. We prefer the formatted partition.
                if (!dev_object["mountdir"].toString().isEmpty())
                {
                    // The new object has a mount folder, so the one in the combobox
                    // needs to be replaced.
                    ui->cb_devices->removeItem(found_index);

                    // Add the device to the combobox.
                    ui->cb_devices->addItem(device_label, dev_object);
                }
            }
            else
            {
                // Add the device to the combobox.
                ui->cb_devices->addItem(device_label, dev_object);
            }
        }
    }

    // Check for USB drives.
    if (mediaObj.contains("usb") && mediaObj["usb"].isArray())
    {
        QJsonArray dev_array = mediaObj["usb"].toArray();

        // Loop through the array, adding elements.
        for (item_index = 0; item_index < dev_array.size(); ++item_index)
        {
            QJsonObject dev_object = dev_array[item_index].toObject();

            // Remember this is an USB device.
            dev_object["type"] = GIGE_MEDIA_USB;

            // Either retrieve the label or build one.
            base_label = dev_object["label"].toString();
            dev_size = dev_object["devsize"].toDouble() / 1.0e9;
            if (base_label.isEmpty())
            {
                // There was no label, so make one using the device size value.
                device_label = QString("USB %1GB").arg(dev_size, 0, 'f', 0);
            }
            else
            {
                // Make sure the name is unique.
                device_label = base_label + QString(" [%1GB]").arg(dev_size, 0, 'f', 0);
            }

            // Add the label to our JSON object so we can use it later. Note that
            // there is already a "label" object, so we'll refer to this as the "tag".
            dev_object["tag"] = device_label;

            // See if we already have a device with this name.
            found_index = findDevice(dev_object);
            if (found_index >= 0)
            {
                QJsonObject found_object;

                // It is likely that the found object is either the formatted partition
                // or the raw device. We prefer the formatted partition.
                if (!dev_object["mountdir"].toString().isEmpty())
                {
                    // The new object has a mount folder, so the one in the combobox
                    // needs to be replaced.
                    ui->cb_devices->removeItem(found_index);

                    // Add the device to the combobox.
                    ui->cb_devices->addItem(device_label, dev_object);
                }
            }
            else
            {
                // Add the device to the combobox.
                ui->cb_devices->addItem(device_label, dev_object);
            }
        }
    }

    // If there is no external media then disable the devices controls.
    if (ui->cb_devices->count() == 0)
    {
        // Make sure we're disabled.
        ui->grp_devices->setEnabled(false);
        ui->cam_tree->setEnabled(false);
        ui->pb_refresh->setEnabled(false);
        ui->grp_options->setChecked(false);
        ui->grp_options->setEnabled(false);
        doOptions(false);
    }
    else
    {
        int sel_index = 0;

        // Make sure we're enabled.
        ui->cb_devices->setEnabled(true);

        // If we had a previously selected value, then restore that.
        // Otherwise, just select the first entry.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);
            if (sel_index < 0)
                sel_index = 0;
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);

        // If we didn't have anything selected, make sure we update the first entry.
        if (device_sel.isEmpty() || force_update)
        {
            // Reset our force update flag.
            force_update = false;
            last_device.clear();

            // Select the first entry.
            doDeviceSelected(sel_index);

            // Enable the device controls.
            ui->cb_devices->setEnabled(true);
            ui->pb_format->setEnabled(cap_file_load != true);
            ui->tb_eject->setEnabled(true);
        }
    }

    // Allow signals to flow again.
    ui->cb_devices->blockSignals(false);
}

/*!
 * @brief Slot handler for copy timeout.
 */
void CamExplore::handleProgressCancelledSignal()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;
    userCanceled(__PRETTY_FUNCTION__);
}

/*!
 * @brief Slot handler for SaveVideo::userCancelledSaveSignal.
 */
void CamExplore::handleSaverProgressCancelledSignal()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;
    userCanceled(__PRETTY_FUNCTION__);
}

/*!
 * @brief Slot handler for CameraMDI::save_error_reported signal.
 */
void CamExplore::handleSaverErrorReported()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;
    // Call this function as it will do the right thing depending on
    // whether saving to PC or not.
    userCanceledCopy();
}

void CamExplore::formatPoll()
{
    // Is everything OK?
    if (format_cmd_status != GEV_STATUS_SUCCESS)
    {
        QString status_str;

        // Is the formatting process already going?
        if (format_in_progress)
        {
            // Shutdown the format display.
            updateFormatProgress(GIGE_FORMAT_DONE | EBUSY);
        }
        else
        {
            // Have the camera resume video.
            camera_ptr->resume_video_request();

            // Hide the format progress indicator and show the button.
            ui->pb_format->setVisible(true);
            ui->format_progress->setVisible(false);

            // Enable the group.
            ui->grp_devices->setEnabled(true);
            ui->cam_tree->setEnabled(true);
            ui->pb_refresh->setEnabled(true);
            ui->grp_options->setEnabled(true);
        }

        // Let the user know.
        status_str = FI_comm::gev_status_to_msg(format_cmd_status);
        QMessageBox::information(MainWindow::GetInstance().get(), tr("Format Media"), tr("Formatting failed with error: ") + status_str);
    }

    // Make sure we still have a camera pointer.
    else if (camera_ptr != nullptr)
    {
        FI_comm     *comm_ptr;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
            return;
        }

        // Ask the camera for the formatting status.
        comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_FORMAT_MEDIA);
    }
}

/*!
 * @brief On completion of loading the CAP file into the camera's image memory.
 */
void CamExplore::loadingFinish()
{
    FI_comm     *comm_ptr;

    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Was there an error?
    if (!cap_load_success)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to load the CAP file!";
        return;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;

        // Reset our state.
        cur_state = EXPLORE_IDLE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;

        // Reset our state.
        cur_state = EXPLORE_IDLE;
        return;
    }

    // Tell the camera to start loading.
    comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_PART_STATE);

    // Cleanup after ourselves.
    cleanupState();

    // Close our window.
    close();
}

/*!
 * @brief The communications object had a message timeout (although it's not necessarily
 *        for us).
 *
 * @param ip_addr   IP address of the camera that timed out.
 * @param reg_addr  The register which caused the timeout.
 * @param read_op   True if it was on a read, false for a write.
 */
void CamExplore::msgTimeoutEvent(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    Q_UNUSED(ip_addr)
    Q_UNUSED(reg_addr)
    Q_UNUSED(read_op)
}

void CamExplore::registerPoll()
{
    // Right now, we only need this for series3 cameras. But it is here if we ever need it.
    if (camera_ptr && camera_ptr->is_series3())
    {
        // Ask the series3 camera for available media.
        camera_ptr->start_camera_read_reg(GIGE_REG_AVAIL_MEDIA);
    }
}

/*!
 * @brief The communications object is indicating that a previously requested register read operation
 *        has completed.
 *
 * @param ip_addr   The IP address of the camera the read operation was for.
 * @param reg_addr  The register that was being read.
 */
void CamExplore::readComplete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;
    FI_comm     *comm_ptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Is this event for us?
    if (ip_addr != cam_ip_addr)
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Read the register value.
    reg_value = comm_ptr->get_reg(ip_addr, reg_addr);

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_CAMERA_STATE:
        if (cur_state == CAP_XFER_WAIT_FOR_LIVE)
        {
#if LOG_BATCH_DOWNLOAD_TRACES
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Current camera state" << MainWindow::camera_state_to_string(reg_value)
                         << "CAP_XFER_WAIT_FOR_LIVE is waiting for camera state GIGE_CAMERA_STATE_LIVE_VIDEO. The wait_for_state_count:" << wait_for_state_count;
#endif
            if (reg_value != GIGE_CAMERA_STATE_LIVE_VIDEO)
            {
                // Decrement our wait counter.
                --wait_for_state_count;
                if (wait_for_state_count == 0)
                {
#if LOG_BATCH_DOWNLOAD_TRACES
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Current camera state" << MainWindow::camera_state_to_string(reg_value)
                                 << "CAP_XFER_WAIT_FOR_LIVE and writing GIGE_REG_CHANGE_CAMERA_STATE NOW for LIVE VIDEO";
#endif
                    // The state transition didn't happen. Ask for it again.
                    camera_ptr->write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_LIVE_VIDEO);

                    // Wait some more.
                    wait_for_state_count = LIVE_STATE_DELAY;
                }
            }
            else
            {
                // We have transitioned to live video state kill our failure delay timer.
                // Kill our timer.
                if (live_timer_id != 0)
                {
                    killTimer(live_timer_id);
                    live_timer_id = 0;
                }

                // Let's see if we can continue...
                cur_state = CAP_XFER_LOAD;

                // Confirmed Live Video State - make sure the stream is paused.
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "Confirmed LIVE VIDEO camera state ARE WE PAUSED?!";

                // Make a connection for our special slot processing for series3 cameras.
                // Note:  It will probably also be caught in the normal CamExplore::stream_paused but we
                // are in the CAP_XFER_LOAD state already...
                connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::capXferStartReady);

                // Camera is confirmed in live mode this will pause live video stream or confirm that the
                // stream is paused in either case the signal above will propagate to continue our cap transfer.
                camera_ptr->pause_video_request();
            }
        }
        else if (cur_state == CAP_XFER_WAIT_FOR_REVIEW)
        {
            if (reg_value != GIGE_CAMERA_STATE_REVIEW)
            {
                // Decrement our wait counter.
                --wait_for_state_count;
                if (wait_for_state_count == 0)
                {
                    // The state transition didn't happen. Ask for it again.
                    camera_ptr->write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_REVIEW);

                    // Wait some more.
                    wait_for_state_count = REVIEW_STATE_DELAY;
                }
            }
            else
            {
                // We have transitioned to review state kill our failure delay timer.
                // Kill our timer.
                if (review_timer_id != 0)
                {
                    killTimer(review_timer_id);
                    review_timer_id = 0;
                }

                // Reset to the proper state coming into cap_loaded_review_check().
                cur_state = CAP_LOAD;

                // Confirmed Review State - recheck for the batch download process..
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "Confirmed camera transitioned to review state, re-running cap_loaded_review_check().";

                // Recheck so we can transfer to the next step in the CAP xfer process.
                capLoadedReviewCheck();
            }
        }
        break;
    case GIGE_REG_FORMAT_MEDIA:
        updateFormatProgress(reg_value);
        break;
    case GIGE_REG_FILE_ACTION_STATUS:
        parseFileActionStatus(reg_value);
        break;
    case GIGE_REG_FILEPATH:
        listingDownloadProcess();
        break;
    case GIGE_REG_PART_NUM_FRAMES:
        // We only care about this during the CAP file transfer operation.
        if (cur_state == CAP_XFER_LOAD)
        {
            // Got the number of frames, so carry on.
            QApplication::postEvent(this, new ExploreEvent(CAP_XFER_EVENT, CAP_XFER_SAVE));
        }
        break;
    case GIGE_REG_AVAIL_MEDIA:
        updateMedia(reg_value);
        break;
    default:
        break;
    }
}

/*!
 * @brief The user wants to close the explore dialog.
 */
void CamExplore::reject()
{
    // Pass it on.
    QDialog::reject();
}

/*!
 * @brief Slot invoked when the CameraMDI object indicates that the stream is paused.
 */
void CamExplore::streamPaused()
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Release the "paused" signal.
    disconnect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::streamPaused);

    // Are we in the middle of a cancel?
    if (cancel_in_progress)
    {
        // Carry on.
        userCanceled(__PRETTY_FUNCTION__);
        return;
    }

    // Do state-specific processing.
    switch (cur_state)
    {
    case CAP_XFER_STARTUP:
        QApplication::postEvent(this, new ExploreEvent(CAP_XFER_EVENT, CAP_XFER_LOAD));
        return;
    case COPY_SET_FILE:
    case COPY_METAFILE:
    case COPY_COMPLETE:
    case COPY_DIRECT:
    case COPY_FAULT:
        QApplication::postEvent(this, new ExploreEvent(COPY_EVENT, COPY_PAUSED));
        return;
    default:
        break;
    }

    // We're paused, so now grab the lock.
    camera_ptr->write_camera_register(GIGE_REG_FASTEC_CCP, GIGE_FASTEC_CCP_CA_BE);
}

/*!
 * @brief Updates the format progress dialog.
 * @param reg_value     The current progress reported by the camera.
 */
void CamExplore::updateFormatProgress(uint32_t reg_value)
{
    QString msg_str;
    //-------------------------------------------------------------------
    // Statistics
#ifdef DO_STATS
    static QElapsedTimer   elapsed_timer;
#endif
    // Statistics
    //-------------------------------------------------------------------

    // Is the format still going?
    if (reg_value == 0)
    {
        // Make sure we tell the user that we're formatting.
        MainWindow::GetInstance()->set_statusbar_message(tr("Format in progress. Please wait..."));

        // Do we know that we're formatting?
        if (!format_in_progress)
        {
            //-------------------------------------------------------------------
            // Statistics
#ifdef DO_STATS
            elapsed_timer.start();
#endif
            // Statistics
            //-------------------------------------------------------------------

            // Remember we're formatting.
            format_in_progress = true;

            // Disable the group.
            ui->grp_devices->setEnabled(false);
            ui->cam_tree->setEnabled(false);
            ui->pb_refresh->setEnabled(false);
            ui->grp_options->setChecked(false);
            ui->grp_options->setEnabled(false);
            doOptions(false);

            // Hide the format button and show the progress indicator.
            ui->pb_format->setVisible(false);
            ui->format_progress->setVisible(true);
        }

        // Poll again in a little bit.
        QTimer::singleShot(FORMAT_POLLING_PERIOD, this, &CamExplore::formatPoll);

        // All done for now.
        return;
    }

    // Are we aware of a format happening right now?
    if (format_in_progress)
    {
        //-------------------------------------------------------------------
        // Statistics
#ifdef DO_STATS
        qint64      elapsed;
        QTime       elapsed_time(0, 0, 0, 0);

        elapsed = elapsed_timer.elapsed();
        elapsed_time = elapsed_time.addMSecs((int)elapsed);
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Format duration:" <<  elapsed_time.toString(QString("hh:mm:ss.zzz"));
#endif
        // Statistics
        //-------------------------------------------------------------------

        // Make sure we have an active camera pointer.
        if (camera_ptr == nullptr)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
            return;
        }

        // Formatting is finished.
        format_in_progress = false;

        // Have the camera resume video.
        camera_ptr->resume_video_request();

        // Were we successful?
        if (reg_value == (uint32_t)GIGE_FORMAT_DONE)
            msg_str = tr("Format complete.");
        else
            msg_str = tr("There was an error during format. Error code=%1").arg(reg_value & 0xFF);
        MainWindow::GetInstance()->set_statusbar_message(msg_str);

        // Hide the format progress indicator and show the button.
        ui->pb_format->setVisible(true);
        ui->format_progress->setVisible(false);

        // Enable the group.
        ui->grp_devices->setEnabled(true);
        ui->cam_tree->setEnabled(true);
        ui->pb_refresh->setEnabled(true);
        ui->grp_options->setEnabled(true);

        // Start a refresh and force an update.
        force_update = true;
        QTimer::singleShot(0, this, &CamExplore::registerPoll);
    }
}

/*!
 * @brief Update our UI controls based on the value read from the camera.
 * @param reg_value     The GIGE_REG_AVAIL_MEDIA register value.
 */
void CamExplore::updateMedia(uint32_t reg_value)
{
    QString device_sel;

    // Is there a selected device?
    if (ui->cb_devices->currentIndex() >= 0)
        device_sel = ui->cb_devices->currentText();

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "reg_value:" << Qt::hex << Qt::showbase << reg_value << "device_sel:" << device_sel;

    // Only do this if we're a series3 camera.
    if (camera_ptr->is_series3())
    {
        int item_index;
        if ((reg_value & GIGE_MEDIA_EXISTS_SDCARD) == GIGE_MEDIA_EXISTS_SDCARD)
        {
            // Only add the item if it doesn't already exist.
            if (ui->cb_devices->findText(CamExplore::SDCARD_LABEL) < 0)
                ui->cb_devices->addItem(CamExplore::SDCARD_LABEL);
        }
        else
        {
            // Check to see if the item went away. Remove it from the list.
            if ((item_index = ui->cb_devices->findText(CamExplore::SDCARD_LABEL)) >= 0)
                ui->cb_devices->removeItem(item_index);
        }
        if ((reg_value & GIGE_MEDIA_EXISTS_SSD) == GIGE_MEDIA_EXISTS_SSD)
        {
            // Only add the item if it doesn't already exist.
            if (ui->cb_devices->findText(CamExplore::SSD_LABEL) < 0)
                ui->cb_devices->addItem(CamExplore::SSD_LABEL);
        }
        else
        {
            // Check to see if the item went away. Remove it from the list.
            if ((item_index = ui->cb_devices->findText(CamExplore::SSD_LABEL)) >= 0)
                ui->cb_devices->removeItem(item_index);
        }
        if ((reg_value & GIGE_MEDIA_EXISTS_USB) == GIGE_MEDIA_EXISTS_USB)
        {
            // Only add the item if it doesn't already exist.
            if (ui->cb_devices->findText(CamExplore::USB_LABEL) < 0)
                ui->cb_devices->addItem(CamExplore::USB_LABEL);
        }
        else
        {
            // Check to see if the item went away. Remove it from the list.
            if ((item_index = ui->cb_devices->findText(CamExplore::USB_LABEL)) >= 0)
                ui->cb_devices->removeItem(item_index);
        }
    }

    // If there is no external media then disable the format and explore buttons.
    // Enabling is left to when the user clicks on the radio button for the particular device.
    // If there is no external media then disable the devices controls.
    if (ui->cb_devices->count() == 0)
    {
        // Make sure we're disabled.
        ui->grp_devices->setEnabled(false);
        ui->cam_tree->setEnabled(false);
        ui->pb_refresh->setEnabled(false);
        ui->grp_options->setChecked(false);
        ui->grp_options->setEnabled(false);
        doOptions(false);
    }
    else
    {
        int sel_index = -1;     // Set it to unknown.

        // Make sure we're enabled.
        ui->cb_devices->setEnabled(true);

        // Attempt to re-select the old device.
        if (!device_sel.isEmpty())
        {
            // Where is the old selection?
            sel_index = ui->cb_devices->findText(device_sel);

            // If it isn't found clear the device_sel.
            if (sel_index < 0)
                device_sel.clear();
        }

        // If we haven't found a previously selected device, set it to an SSD device.
        if (sel_index < 0)
        {
            // Attempt to set it to an SSD device as a default.
            sel_index = ui->cb_devices->findText(CamExplore::SSD_LABEL);

            // If that did not work fall-back to the first device in the list.
            if (sel_index < 0)
            {
                sel_index = 0;
            }
        }

        // Update the display.
        ui->cb_devices->setCurrentIndex(sel_index);

        // If we didn't have anything selected, make sure we update the first entry.
        if (device_sel.isEmpty())
        {
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Calling doDeviceSelected sel_index:" << sel_index;

            // Select the SSD or first entry.
            doDeviceSelected(sel_index);

            // Enable the device controls.
            ui->cb_devices->setEnabled(true);
            ui->pb_format->setEnabled(cap_file_load != true);
            ui->tb_eject->setEnabled(true);
        }
    }

    // Start a timer to poll the register again.
    QTimer::singleShot(REGISTER_POLLING_PERIOD, this, &CamExplore::registerPoll);
}

/*!
 * @brief The user has pressed the cancel button on the progress dialog.
 */
void CamExplore::userCanceled(const QString& whoCalledMe)
{
    SaveVideo   *saver_ptr;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << " called by[" << whoCalledMe << "]";

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Are we already in the middle of a cancel?
    if (!cancel_in_progress)
    {
        // Tell the camera to stop any file operation it has going, if we haven't already.
        camera_ptr->write_camera_register(GIGE_REG_FILE_ACTION, GIGE_FA_CANCEL);
    }

    // Get our saver object so we can close it.
    saver_ptr = camera_ptr->get_saver();
    if (saver_ptr != nullptr)
    {
        // Tell the saver to close it's file(s).
        saver_ptr->close(__PRETTY_FUNCTION__);
    }

    // We may need to resume live video streaming.
    if (copy_destination == DEST_PC_PATH)
        MainWindow::GetInstance()->resume_video_streaming();

    // Cleanup our display.
    cleanupState();

    // Tell ourselves to shutdown.
    QApplication::postEvent(this, new QCloseEvent());
}

/*!
 * @brief The user canceled a copy operation.
 */
void CamExplore::userCanceledCopy()
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Remember we're canceling.
    cancel_in_progress = true;

    // Tell the camera to stop any file operation it has going.
    camera_ptr->write_camera_register(GIGE_REG_FILE_ACTION, GIGE_FA_CANCEL);

    // If we have a worker thread, we are doing a direct copy. This requires
    // different handling.
    if (copy_worker_thread.get() != nullptr)
    {
        // If the worker thread is still running, we want to cancel it.
        if (copy_worker_thread->isRunning())
        {
            // Make sure we tell the user that we're cancelling.
            MainWindow::GetInstance()->set_statusbar_message(tr("Cancel in progress. Please wait..."));

            // Don't allow the user to do anything silly.
            setEnabled(false);

            // We will need to continue processing after the worker thread has cancelled.
            connect(copy_worker_thread.get(), &QFutureWatcherBase::finished, this, [ = ]()
            {
                // Clear the status.
                MainWindow::GetInstance()->set_statusbar_message("");

                // Cleanup our display.
                cleanupState();

                // Tell ourselves to shutdown.
                QApplication::postEvent(this, new QCloseEvent());
            });

            // Cancel the worker thread.
            copy_worker_thread->cancel();
        }
        else
        {
            // Cleanup our display.
            cleanupState();

            // Tell ourselves to shutdown.
            QApplication::postEvent(this, new QCloseEvent());
        }
    }

    // Was the copy to the PC?
    else if (copy_destination == DEST_PC_PATH)
    {
        // Disconnect from the copy timeout signal.
        disconnect(camera_ptr, &CameraMDI::copy_timeout, this, &CamExplore::handleCopyTimeout);

        // Disconnect from the camera's reported error signal.
        disconnect(camera_ptr, &CameraMDI::save_error_reported, this, &CamExplore::handleSaverErrorReported);

        // We want to know when the video stream is paused.
        connect(camera_ptr, &CameraMDI::video_is_paused, this, &CamExplore::streamPaused);

        // Tell the camera to pause it's stream.
        camera_ptr->pause_video_request();
    }
    else
    {
        // Carry on.
        userCanceled(__PRETTY_FUNCTION__);
    }
}

/*!
 * @brief The communications object is indicating that a previously requested register write operation
 *        has completed.
 *
 * @param ip_addr   The IP address of the camera the write operation was for.
 * @param reg_addr  The register that was being written.
 */
void CamExplore::writeComplete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    write_status;
    FI_comm     *comm_ptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Is this event for us?
    if (ip_addr != cam_ip_addr)
        return;

    // Make sure we're in a state that cares about write status.
    switch (cur_state)
    {
    case CAP_XFER_STARTUP:
    case CAP_XFER_WAIT_FOR_LIVE:
    case CAP_XFER_WAIT_FOR_REVIEW:

    case COPY_METAFILE:
    case COPY_PAUSED:
    case COPY_SET_FILE:
    case COPY_STARTUP:
    case COPY_WAIT:
    case COPY_COMPLETE:
    case COPY_DIRECT:
    case COPY_FAULT:

    case DELETE_SET_FILE:
    case DELETE_STARTUP:
    case DELETE_WAIT:

    case REFRESH_STARTUP:
    case REFRESH_STILLS:
    case REFRESH_STACKS:
    case REFRESH_VIDEO:
    case REFRESH_CAPTURE:

    case EXPLORE_IDLE:
        break;
    case CAP_LOAD:
    case CAP_XFER_LOAD:
    case CAP_XFER_SAVE:
    case CAP_XFER_COMPLETE:
        // Not interested in any write completion status while in these states.
        return;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected cur_state:" << cur_state;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_FASTEC_CCP:
        // Read the status of the write operation.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Process the write completion.
        writeLockDone(write_status);
        break;
    case GIGE_REG_FILE_ACTION:
    case GIGE_REG_FA_JSON:
        // Read the status of the write operation.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Was this a success?
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Start our timer to poll for the completion of the file action.
            if (file_timer_id == 0)
                file_timer_id = startTimer(FA_POLLING_PERIOD);
        }
        else
        {
            // There was a problem. Find out what the problem was.
            comm_ptr->start_read_reg(cam_ip_addr, GIGE_REG_FILE_ACTION_STATUS);
        }
        break;
    case GIGE_REG_MA_JSON:
        // Read the status of the write operation.
        format_cmd_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Start our timer to poll the camera for formatting progress.
        QTimer::singleShot(0, this, &CamExplore::formatPoll);
        break;
    case GIGE_REG_FILEPATH:
        // Read the status of the write operation.
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);

        // Process the write completion.
        writeFilepathDone(write_status);
        break;
    default:
        break;
    }
}
