/*!
 * @file camera_mdi.cpp
 * @brief This file contains the implementation of the CameraMDI
 *        class. This is the child window class which connects and
 *        displays video, etc. from a specified camera.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "camera_mdi.h"

#include <QAction>
#include <QApplication>
#include <QAtomicInt>
#include <QByteArray>
#include <QCheckBox>
#include <QCloseEvent>
#include <QDesktopServices>
#include <QDir>
#include <QtEndian>
#include <QFileDialog>
#include <QFileInfo>
#include <QGraphicsScene>
#include <QJsonDocument>
#include <QLabel>
#include <QMessageBox>
#include <QMdiSubWindow>
#include <QMenu>
#include <QMessageBox>
#include <QProcess>
#include <QPushButton>
#include <QtMath>
#include <QtNetwork>
#include <QString>
#include <QStringList>
#include <QTemporaryFile>
#include <QTextStream>
#include <QTimer>
#include <QVBoxLayout>
#include <QVector>
#include <QVersionNumber>
#include <QWindow>

#include <errno.h>
#include <stdio.h>
#include <stdlib.h>

#include <memory>
#include <sstream>

#include "calibrate_dlg.h"
#include "cam_update_dlg.h"
#include "camera_dock.h"
#include "camera_explore.h"
#include "camera_info.h"
#include "camera_preferences.h"
#include "camera_review.h"
#include "camera_stream.h"
#include "camera_video.h"
#include "client_math.h"
#include "eventfilter.h"
#include "fi_camreg_group.h"
#include "fi_fflr_underflow_dialog.h"
#include "fi_histogram.h"
#include "fi_loadcap.h"
#include "fi_metamap.h"
#include "fi_msgbox.h"
#include "fi_overlay.h"
#include "fi_pixmap.h"
#include "fi_progbar.h"
#include "fi_progress_dialog.h"
#include "fi_rectitem.h"
#include "fi_sensor.h"
#include "fi_updater.h"
#include "gige_socket.h"
#include "histo_dock.h"
#include "jump_dialog.h"
#include "logger.h"
#include "mainwindow.h"
#include "multicam_dock.h"
#include "playback_meta_info.h"
#include "save_dialog.h"
#include "save_snapshot.h"
#include "save_video.h"
#include "select_device.h"
#include "select_folder.h"
#include "time_date.h"
#include "user_preferences.h"
#include "wifi_config.h"

#include <fas_video.h>
#include <fi_tiff.h>

namespace CAMERA_MDI_SETTINGS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Log messages for testing purposes.  Set to 1 to have them compiled for output.
#define LOG_AUTOSAVE_TRACING            0
#define LOG_ACTIVE_METADATA_DOWNLOAD    0
#define LOG_BLOCK_DOWNLOAD_INFO         0
#define LOG_PROCESS_STATE_CHANGES       0
#define LOG_GROUP_READS                 0
#define LOG_SAVER_IDLE_MSGS             0
#define LOG_SYNC_IN_SUSPEND_INFO        0
#define LOG_PAUSE_ALREADY_PAUSED        0
#define LOG_PAUSE_NOT_PAUSED            1
#define LOG_WRITE_COMPLETE_ERRORS       1
#define FORCE_MISSING_FRAMES            0

//! How many histograms do we want to accumulate for a baseline?
const int       AE_BASELINE_COUNT       = 32;

//! How many "bins" shift in the histogram is OK.
const int       AE_RANGE_OK             = 5;

//! How quickly do we change the exposure?
const qreal     AE_MU                   = 0.5;

//! How many histograms do we want to accumulate for a baseline?
const int       VT_BASELINE_COUNT       = 8;

//! How frequently to poll the busy status.
const int       BUSY_POLL_TIMEOUT       = 50;

//! Define a timeout to prevent the progress dialog from locking up the application.
const int       CALIBRATION_TIMEOUT     = 20000;

//! Keep a constant of when a sanity check on camera_info fails.
const QString CAMERA_INFO_SANITY_FAILURE = QStringLiteral("[CHECK FAILED] The camera_info was a nullptr!");

//! Keep a constant of when a sanity check on camera_info fails.
const QString STREAMER_SANITY_FAILURE = QStringLiteral("[CHECK FAILED] The streamer was a nullptr!");

//! Delay the raw PFMD download to allow the image pipeline to flush.
const int       RAW_DELAY_TIMEOUT       = 50;

//! The default heartbeat value is too fast to cover all computers and
//! all situations. We'll use a longer period.
const uint32_t  FASMOTION_HEARTBEAT     = 10000;

//! Define an offset to the standard rate. We need to be faster so we
//! don't timeout.
const uint32_t  HEARTBEAT_OFFSET        = (FASMOTION_HEARTBEAT >> 1) + 100;

//! The maximum "chunk" of per-frame metadata we can read in one go.
//const uint32_t  METADATA_CHUNK_SIZE     = (1280 * 1024 * 4);

//! A short amount of time to delay before streaming the metadata from the camera
//! when transitioning to review mode.
const int       METADATA_STREAM_DELAY   = 100;

//! Number of times to try downloading a frame before giving up.
const int       MAX_RETRY_FRAMES        = 7;

//! Arbitrary point at which we release control of the UI.
const int       MAX_SEARCH_COUNT        = 5000;

//! Can't truncate before this frame.
const uint32_t  MIN_TRUNCATE_FRAME      = 3;

//! Time to pause between searching bursts.
const int       SEARCH_DELAY            = 10;

//! Delay a bit before reading some registers.
const int       DELAY_UPDATE            = 1000;

//! Delay to check for a state update.
const int       STATE_CHECK_DELAY       = 500;

//! Delay a bit before finishing the shutdown process.
const int       SHUTDOWN_DELAY          = 100;

//! Shift amount to get the default gamma value.
const int       DEFAULT_GAMMA_SHIFT     = 16;

//! Delay amount when bouncing from Live to Review states.
const int       LIVE_TO_REVIEW_DELAY    = 750;

//! Delay amount to wait when releasing memory.
const int       RELEASE_MEMORY_DELAY    = 250;

//! Delay amount when bouncing from Review to Live states.
const int       REVIEW_TO_LIVE_DELAY    = 250;

//! Number of milliseconds that we allow in-between sync-in warnings about frame timeouts.
const qint64    SYNC_IN_FRAME_TIMEOUT_WARNING_INTERVAL = 4000;

//! Number of milliseconds that we display in main window status line the
//! sync-in warning when a frame timeout has taken place.
const int       SYNC_IN_FRAME_TIMEOUT_STATUS_INTERVAL = 2000;

//! The maximum number of attempts to download the raw metadata.
const int       MAX_RETRIES_RAW_METADATA    = 3;

//! Timeout for the polling loop looking at the review video stop request.
const int       REVIEW_STOP_TIMEOUT = 50;

//! The maximum size of the PFMD LRU cache.
const int       PFMD_CACHE_MAX_SIZE = 20;

//! Define calibration event types.
enum
{
    CAL_EVENT_NONE,                     //!< No calibration event pending.
    CAL_EVENT_CAMERA,                   //!< Camera initiated calibration event.
    CAL_EVENT_FACTORY,                  //!< Performing a Factory calibration.
    CAL_EVENT_USER,                     //!< A user calibration event.
    CAL_EVENT_PRNU                      //!< Performing a PRNU calibration.
};

//! Define autoexposure states
enum
{
    AE_INIT_TRACKING,                   //!< Initialize our autoexposure process.
    AE_BASELINE,                        //!< Initialize our baseline.
    AE_PROCESS                          //!< Process the data.
};

//! Define video trigger tracking states.
enum
{
    VT_INIT_TRACKING,                   //!< Initialize our video triggering process.
    VT_INIT_BASELINE,                   //!< Initializing our baseline.
    VT_PROCESS                          //!< Process the histogram data.
};

//! Define sync-in warning display states.
enum
{
    SYNC_IN_WARN_NONE,                  //!< Never received a sync-in warning.
    SYNC_IN_WARN_DLG_SHOWN,             //!< Sync-in warning received and user has seen dialog.
    SYNC_IN_WARN_STREAM_RESUMED,        //!< Sync-in warning received dialog shown and stream resumed.
};

// A rough conversion from RGB to Y.
#define RGB_TO_Y(r, g, b)           (((r)+(r)+(r)+(b)+(g)+(g)+(g)+(g)) >> 3)

/*!
 * @brief The PfmdCache class is used to have an Least Recently Used (LRU) cache for the
 *        per-frame metadata.
 */
class PfmdCache
{
private:
    int max_cache_size;         //!< The maximum cache size, configured at cache creation time.

    QList<uint32_t> pfmd_lru;   //!< The ordered list of frames. The most used frame is at the front, least at the back.
    QHash<uint32_t, QByteArray *>   pfmd_hash;  //!< Associating the frame to the PFMD information.

public:
    PfmdCache(int capacity) { max_cache_size = capacity; }
    ~PfmdCache()            { clear();  }

    /*!
     * @brief Removes all elements from the cache (which are destroyed).
     */
    void clear()
    {
        // Remove all the entries from the hash table.
        QList<uint32_t>::iterator i;
        for (i = pfmd_lru.begin(); i != pfmd_lru.end(); ++i)
        {
            QByteArray *data_ptr = pfmd_hash.take(*i);
            delete data_ptr;
        }

        // Reset our list.
        pfmd_lru.clear();
    }

    /*!
     * @brief Add the PFMD data to the cache if it is new (and there is room). Move the
     *        entry to the front if it already exists. If there is no room, delete the
     *        "oldest" value, and add the new value.
     *
     * @param frame_num The frame number associated with the PFMD.
     * @param pfmd_ptr  The pointer to the QBtyeArray containing the PFMD.
     */
    void insert(uint32_t frame_num, QByteArray *pfmd_ptr)
    {
        int data_index;

        // Does it already exist?
        data_index = pfmd_lru.indexOf(frame_num);
        if (data_index < 0)
        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_num:" << frame_num << "Not found. List size:" << pfmd_lru.size() << "max_cache_size:" << max_cache_size;

            // It does not exist. Do we have room?
            if (((int)pfmd_lru.size()) == max_cache_size)
            {
                int last_index = pfmd_lru.takeLast();
                QByteArray *data_ptr = pfmd_hash.take(last_index);
                delete data_ptr;
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Need to delete the last entry, frame_num:" << last_index;
            }

            // Add our data.
            pfmd_hash.insert(frame_num, pfmd_ptr);
            pfmd_lru.prepend(frame_num);
        }
        else if (data_index != 0)
        {
            // Move from where it is to the front of the list.
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Move frame_num:" << frame_num << "from data_index:" << data_index;
            pfmd_lru.move(data_index, 0);
        }
//        else
//        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_num:" << frame_num << "is at the front of the list";
//        }

//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "LRU:" << pfmd_lru;
    }

    /*!
     * @brief Return the PFMD associated with the supplied frame number, frame_num, if it exists
     *        in the hash.
     *
     * If the frame number is found, then not only is the associated PFMD data retruned, but the
     * frame is moved to the front of the LRU queue.
     *
     * @param[in] frame_num     The frame number we are interested in.
     * @return The QByteArray pointer to the PFMD, or nullptr if not found.
     */
    const QByteArray *retrieve(uint32_t frame_num)
    {
        int data_index;
        QByteArray *data_ptr = nullptr;

        // Does it already exist?
        data_index = pfmd_lru.indexOf(frame_num);
        if (data_index >= 0)
        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Found frame_num:" << frame_num << "at:" << data_index;
            data_ptr = pfmd_hash.value(frame_num);

            // Update our LRU cache.
            insert(frame_num, data_ptr);
        }
//        else
//        {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find frame_num:" << frame_num;
//        }

        // Return our data pointer, or nullptr.
        return data_ptr;
    }
};
}   // End of namespace CAMERA_MDI_SETTINGS.

using namespace CAMERA_MDI_SETTINGS;

// Give a common string for the FasFireLR feature for user interface consistency.
const QString CameraMDI::fflr_feature_name            = QT_TR_NOOP("FasFire LR");

// Sync-in warning constant strings reused for both dialog and status line of main window.
const QString CameraMDI::sync_in_warning_msg          = QT_TR_NOOP("Please check external I/O connections and configuration.");
const QString CameraMDI::sync_in_warning_title_msg    = QT_TR_NOOP("Sync-In Signal Suspended");

// PRNU check message.
const QString CameraMDI::prnu_title_str = QT_TR_NOOP("Factory PRNU Calibration");
const QString CameraMDI::prnu_setup_msg = QT_TR_NOOP("It is assumed that:<br>"
                                                     "1) You have the camera pointed at a flat-field, almost saturated image.<br>"
                                                     "2) You are using the full bit-depth of your sensor.<br>"
                                                     "3) Your light source is stable (i.e. NOT fluorescent!)<br>"
                                                     "4) You have verified that the mean pixel values are approximately 80% full "
                                                     "scale under sampling conditions.<br>"
                                                     "5) You have performed a black-frame calibration with the current settings.<br><br>"
                                                     "Are you sure you want to perform a PRNU calibration?<br>"
                                                     "This will require the connection to be closed and reopened. Proceed?");

const QString CameraMDI::prnu_red_path      = QT_TR_NOOP("Select the RED stack to use for PRNU calibration");
const QString CameraMDI::prnu_green_path    = QT_TR_NOOP("Select the GREEN stack to use for PRNU calibration");
const QString CameraMDI::prnu_blue_path     = QT_TR_NOOP("Select the BLUE stack to use for PRNU calibration");

//! The environment variable lable for XDG_RUNTIME_DIR.
const char CameraMDI::XDG_RUNTIME_DIR[] = "XDG_RUNTIME_DIR";

/*!
* @brief The class constructor.
*/
CameraMDI::CameraMDI(QWidget *parent) :
    QWidget(parent)
{
    // Initialize our variables.
    active_metadata_requested = false;
    ae_track_state = AE_INIT_TRACKING;
    auto_offset_cal = true;
    block_download_len = 0;
    block_download_id = (int)FI_COMM::NO_REGISTER;
    cache_refresh_holdoff = false;
    camera_ptr = nullptr;
    downloading_pfmd = false;
    fflr_session_len = 0;
    fflr_underflow_discarded = false;
    fflr_user_discards_parts = false;
    force_frame_update = false;
    heartbeat_rate = FASMOTION_HEARTBEAT - HEARTBEAT_OFFSET;
    json_fflr_str.clear();
    json_build_str.clear();
    calibration_event = CAL_EVENT_NONE;
    in_shutdown = false;
    in_startup = true;
    in_truncate = false;
    last_requested_frame = (uint32_t) -1;
    multicam_save_allowed = false;
    ok_to_talk = false;
    old_camera_state = GIGE_CAMERA_STATE_IDLE;
    requested_state = GIGE_CAMERA_STATE_IDLE;
    partition_metadata.clear();
    pm_download_in_progress = false;
    pm_download_only = false;
    pfmd_frame_requests.clear();
    raw_metadata.clear();
    raw_metadata_len = 0;
    raw_path.clear();
    release_all = false;
    retry_count_raw_metadata = 0;
    save_cleanup = false;
    sensor_ptr = nullptr;
    stream_active = false;
    streamer = nullptr;
    sync_in_warning_state = SYNC_IN_WARN_NONE;
    trigger_frame = -1U;
    trigger_frame_start = 0.0;
    vt_area = 10;
    vt_change = 10;
    vt_mark_only = false;
    vt_refresh = 16;
    vt_track_state = VT_INIT_TRACKING;
    waiting_for_frame = false;

    // Reset our PFMD info array.
    pfmd_info = new CAMERA_MDI_SETTINGS::PfmdCache(PFMD_CACHE_MAX_SIZE);

    xferDialog = nullptr;

    lock_timer_id = 0;

    // Start up our 1/sec timer.
    frame_rate_counter = 0;
    frame_rate_timer = startTimer(1000);

    // Initialize our event variables.
    read_settings();
    marker_list.clear();

    // Create our streamer object.
    streamer = new CameraStream(this);
    if (streamer == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate streamer object!";
        return;
    }

    // Create our saver object.
    autosave_pending = false;
    autosave_lockout = false;
    save_video = new SaveVideo(this);
    if (save_video == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate save_video object!";
        return;
    }

    // Make sure we catch when the main window is closing.
    connect(MainWindow::GetInstance().get(), &MainWindow::closing, this, &CameraMDI::commitData, Qt::DirectConnection);

    // Setup the connections to our save video object.
    connect(save_video, &SaveVideo::save_complete, this, &CameraMDI::done_saving, Qt::QueuedConnection);
    connect(save_video, &SaveVideo::save_error, this, &CameraMDI::error_saving, Qt::QueuedConnection);
    connect(save_video, &SaveVideo::save_receive_count_complete, this, &CameraMDI::save_receive_count_complete, Qt::QueuedConnection);
    connect(save_video, &SaveVideo::work_list_empty, this, &CameraMDI::saver_is_idle, Qt::QueuedConnection);

    // Allocate our layout. We want our display split into a top pane for the
    // camera information, middle for the video and bottom for the playback controls.
    layout_ptr = new QVBoxLayout(this);
    if (layout_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate vertical layout!";
        return;
    }

    // Allocate our various panes of information.
    camera_info = new CameraInfo(this);
    video_player = new CameraVideo(this);
    rec_progress = new fiProgBar(this);
    review_controls = new CameraReview(this);

    // We are going to use a Queued connection type later, and Qt needs to know about uint32_t types.
    qRegisterMetaType<uint32_t>("uint32_t");

    // Connect our signals/slots to the camera information pane.
    connect(camera_info, &CameraInfo::camera_name_changed, this, &CameraMDI::new_camera_name);
    connect(camera_info, &CameraInfo::framerate_changed, this, &CameraMDI::framerate_changed);
    connect(camera_info, &CameraInfo::stream_toggle, this, &CameraMDI::user_stream_request);
    connect(camera_info, &CameraInfo::update_clicked, this, &CameraMDI::do_camera_update);

    // Connect our signals/slots to the review controls.
    connect(review_controls, &CameraReview::play_reverse, this, &CameraMDI::video_play_reverse, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::review_stop, this, &CameraMDI::video_review_stop, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::play_forward, this, &CameraMDI::video_play_forward, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::jump_time, this, &CameraMDI::video_jump_time, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::current_frame_changed, this, &CameraMDI::video_current_frame_changed, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::panel_ready, this, &CameraMDI::review_panel_ready, Qt::UniqueConnection);
    connect(review_controls, &CameraReview::play_update, this, &CameraMDI::update_pfmd);
    connect(review_controls, &CameraReview::playback_rate_changed, this, &CameraMDI::video_rate_changed);

    // Connect to the record progress update signal.
    connect(rec_progress, &fiProgBar::time_to_update, this, &CameraMDI::update_rec_progress, Qt::QueuedConnection);

    // Connect up the time trigger timer.
    tt_time = 0;
    tt_count = 0;
    connect(&tt_timer, &QTimer::timeout, this, &CameraMDI::time_trigger_expired, Qt::QueuedConnection);

    // When the camera emits fflr_status_ready signal we catch it and process the playback window
    // for the newly created FasFireLR video.
    connect(this, &CameraMDI::fflr_status_ready, this, &CameraMDI::do_fflr_review, Qt::QueuedConnection);

    // Initially, have the stream button disabled.
    camera_info->enable_stream(false);

    // Add our panes into our splitter.
    layout_ptr->addWidget(camera_info);
    layout_ptr->addWidget(video_player);
    layout_ptr->addWidget(rec_progress);
    layout_ptr->addWidget(review_controls);

    // Give all extra space to the player window. The other windows are fixed
    layout_ptr->setStretchFactor(camera_info, QSizePolicy::Fixed);
    layout_ptr->setStretchFactor(video_player, QSizePolicy::Expanding);
    layout_ptr->setStretchFactor(rec_progress, QSizePolicy::Fixed);
    layout_ptr->setStretchFactor(review_controls, QSizePolicy::Fixed);

    // Hide the progress control until we're recording.
    rec_progress->setVisible(false);
    review_controls->setVisible(false);

    // Make this our layout.
    setLayout(layout_ptr);

    // Make sure we get focus notifications.
    setFocusPolicy(Qt::StrongFocus);
}

/*!
 * @brief The class destructor.
 */
CameraMDI::~CameraMDI()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Enter";

    // Save any settings we want to preserve.
    save_settings();

    // Clean up timers.
    if (lock_timer_id != 0)
    {
        killTimer(lock_timer_id);
        lock_timer_id = 0;
    }
    if (frame_rate_timer != 0)
    {
        killTimer(frame_rate_timer);
        frame_rate_timer = 0;
    }

    // Make sure we cleanup any left-overs.
    reset_event_info();

    // Clean up our widgets.
    layout_ptr->removeWidget(review_controls);
    layout_ptr->removeWidget(rec_progress);
    layout_ptr->removeWidget(video_player);
    layout_ptr->removeWidget(camera_info);
    review_controls->deleteLater();
    review_controls = nullptr;
    rec_progress->deleteLater();
    rec_progress = nullptr;
    sensor_ptr->deleteLater();
    sensor_ptr = nullptr;
    camera_info->deleteLater();
    camera_info = nullptr;
    layout_ptr->deleteLater();
    layout_ptr = nullptr;
    camera_ptr = nullptr;

    if (video_player != nullptr)
    {
        // Make sure our signal/slots are disconnected.
        disconnect_video_player();

        video_player->deleteLater();
        video_player = nullptr;
    }

    // Close down our streamer.
    if (streamer != nullptr)
    {
        streamer->detach();
        streamer->deleteLater();
        streamer = nullptr;
    }

    // Close down our saver.
    if (save_video != nullptr)
    {
        save_video->close(__PRETTY_FUNCTION__);
        save_video->deleteLater();
        save_video = nullptr;
    }

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Exit";
}

/*!
 * @brief Returns true if this is the active window, false otherwise.
 */
bool CameraMDI::active_window()
{
    return (MainWindow::GetInstance()->active_mdi_child() == this);
}

/*!
 * @brief Adds a frame to our retry list if it isn't there already.
 *
 * @param frame_0_based - the sequence number of frame that was lost.
 *
 * @returns True if added or false if already there.
 */
bool CameraMDI::add_to_retry_list(uint32_t frame_0_based)
{
    bool addedIt = false;
    bool on_retry_frame_list = retry_frame_list.contains(frame_0_based);  // Catch this state and use for logging.

    // Only add it if the frame isn't already in the retry frames requested or the retry frame list.
    if (!on_retry_frame_list && !retry_frames_reqd.contains(frame_0_based))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ADDED frame_0_based:" << frame_0_based;
        retry_frame_list.append(frame_0_based);
        addedIt = true;

        // #MISSING_FRAME - Everytime we add a frame/buffer to the retry list the
        // empty buffer must be sent to the save_video object so that the saver
        // can process placeholder frames if part of the save operation.
        // Not so important for non-AVI saves, but mandatory for AVIs.
        std::unique_ptr<FI_MEMPOOL::FiImageBuffer> emptyImageBuffer =
            FI_MEMPOOL::FiImageBuffer::CreateImageBuffer();
        save_video->receive_image(std::move(emptyImageBuffer));

        // Keep the missed_frame_count accumulator in sync with our retry frame list.
        if (missed_frame_count != retry_frame_list.count())
            missed_frame_count = retry_frame_list.count();
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "DUPLICATE RETRY FRAME frame_0_based:" << frame_0_based << (on_retry_frame_list ? "ON RETRY LIST" : "RETRY ALREADY REQUESTED");
    }
    return addedIt;
}

void CameraMDI::ae_baseline()
{
    int i;
    uint32_t peak_lum_value;
    int peak_lum_index;
    const QVector<uint32_t> *lum;

    // Sanity check.
    if (ae_ready)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ae_ready is set, why are we here?";
        ae_track_state = AE_INIT_TRACKING;
        return;
    }

    // Find the peak luminance value.
    lum = video_player->get_histogram()->get_luminance();
    peak_lum_index = 0;
    peak_lum_value = 0;
    for (i = 0; i < lum->size(); i++)
    {
        // Is this the bigger value?
        if (lum->at(i) > peak_lum_value)
        {
            peak_lum_index = i;
            peak_lum_value = lum->at(i);
        }
    }

    // Add the peak bin to our target luminance.
    ae_target_Y += peak_lum_index;

    // Increment the number of histograms we've accumulated.
    ++ae_histo_count;

    // Are we done?
    if (ae_histo_count == AE_BASELINE_COUNT)
    {
        // Average the target luminance.
        ae_target_Y = ae_target_Y / AE_BASELINE_COUNT;

        // We are now ready to begin processing.
        ae_ready = true;

        // Acquire the next histogram.
        ae_track_state = AE_PROCESS;
    }
}

void CameraMDI::ae_process()
{
    int i;
    int err_lum;
    uint32_t peak_lum_value;
    int peak_lum_index;
    const QVector<uint32_t> *lum;

    // Sanity check.
    if (!ae_ready)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ae_ready is not set, why are we here?";
        ae_track_state = AE_INIT_TRACKING;
        return;
    }

    // Find the peak luminance value.
    lum = video_player->get_histogram()->get_luminance();
    peak_lum_index = 0;
    peak_lum_value = 0;
    for (i = 0; i < lum->size(); i++)
    {
        // Is this the bigger value?
        if (lum->at(i) > peak_lum_value)
        {
            peak_lum_index = i;
            peak_lum_value = lum->at(i);
        }
    }

    // What is our error?
    err_lum = ae_target_Y - peak_lum_index;
//    QLOG_TRACE() << "ae_target_Y:" << ae_target_Y << "peak:" << peak_lum_index << "err_lum:" << err_lum;
    if (abs(err_lum) > AE_RANGE_OK)
    {
        uint32_t min_shutter;
        uint32_t max_shutter;

        // Given the frame rate we can set the boundaries for the
        // shutter speed.
        max_shutter = ClientMath::calc_max_exposure(read_camera_cache(GIGE_REG_FRAME_RATE), read_camera_cache(GIGE_REG_FRAME_OVERHEAD));

        // Calculate our adjusted exposure value.
        uint32_t reg_value = read_camera_cache(GIGE_REG_SHUTTER_SPEED) + (uint32_t)(AE_MU * err_lum);

        // Make sure it's not too small.
        min_shutter = read_camera_cache(GIGE_REG_MIN_SHUTTER_SPEED);
        if (reg_value < min_shutter)
            reg_value = min_shutter;

        // Make sure it's not too big.
        if (reg_value > max_shutter)
            reg_value = max_shutter;

        // Tell the camera to make the change.
        write_camera_register(GIGE_REG_SHUTTER_SPEED, reg_value);
//        QLOG_TRACE() << "exp:" << reg_value << "min:" << min_shutter << "max:" << max_shutter;
    }
}

/*!
 * @brief Very similar to the [play_video](@ref CameraMDI::play_video()) method, this is used
 *        to attach to a stream already being generated by the camera. Typically, this is as
 *        a result of the user playing a recorded video via a different UI.
 *
 * @param play_forward  True if the video stream is moving forward, false for reverse.
 */
void CameraMDI::attach_to_playback(bool play_forward)
{
    CAMERA_STREAM::STREAM_PARAMS    stream_info;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // Don't stream if we don't want streaming.
    if (camera_ptr && !camera_ptr->stream)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No streaming allowed.";
        return;
    }

    // Make sure the stream state is appropriate.
    if (streamer && !streamer->is_paused())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "streamer is not paused!";
        return;
    }

    // Sanity check(s)
    if (camera_info == nullptr)
        return;

    // Clear our structure.
    memset(&stream_info, 0, sizeof(stream_info));

    // Get the "new" current frame.
    stream_info.cur_frame = review_controls->get_current_frame_num();
    stream_info.clip_start = review_controls->get_clip_start();
    stream_info.clip_end = review_controls->get_clip_end();

    // Make sure the current frame is within the clip.
    if (stream_info.cur_frame > stream_info.clip_end)
        stream_info.cur_frame = stream_info.clip_end;
    if (stream_info.cur_frame < stream_info.clip_start)
        stream_info.cur_frame = stream_info.clip_start;

    // Figure out our timing parameters.
    stream_info.desired_fps = (uint32_t)review_controls->get_playback_rate();
    if (stream_info.desired_fps <= (uint32_t)camera_info->get_framerate())
    {
        stream_info.skip_interval = 1;
    }
    else
    {
        stream_info.skip_interval = stream_info.desired_fps / (uint32_t)camera_info->get_framerate();
        stream_info.desired_fps = (uint32_t)camera_info->get_framerate();
    }

    // Figure out the number of milliseconds per frame.
    stream_info.frame_period = 1000 / stream_info.desired_fps;

    // When figuring out the frame delay, make sure we oversample.
    stream_info.frame_delay = ((1000 / stream_info.desired_fps) + 1) / 2;

    // Are we playing in reverse?
    if (!play_forward)
        stream_info.skip_interval |= GIGE_STREAM_REVERSE;

    // We always loop when playing.
    stream_info.skip_interval |= GIGE_STREAM_LOOP;

    // We're attaching to the stream, so let the state machine know.
    stream_info.stream_control = GIGE_STREAM_SLAVE_ENABLE;

    // Tell the streamer to start the playback.
    if (streamer)
        streamer->playback(stream_info);
}

/*!
 * @brief Determine if autoplay is enabled.
 * @return True if autoplay is in effect, false otherwise.
 */
bool CameraMDI::autoplay()
{
    uint32_t    reg_value;
    bool        autosave_enabled;
    bool        autoplay_configured;
    bool        ltr_enabled;
    bool        fflr_enabled;

    // Read the user preferences.
    reg_value = read_camera_cache(GIGE_REG_PREFERENCES);
    autoplay_configured = ((reg_value & (uint32_t)GIGE_PREF_AUTOPLAY) == (uint32_t)GIGE_PREF_AUTOPLAY);

    // Read the autosave settings.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);
    autosave_enabled = ((reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE);

    // Read the LTR mode settings.
    reg_value = read_camera_cache(GIGE_REG_LTR_MODE);
    ltr_enabled = ((reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE);
    fflr_enabled = ((reg_value & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR);

    // We only allow autoplay, if it's configured and neither autosave nor ltr mode are enabled.
    return (autoplay_configured && !autosave_enabled && !ltr_enabled && !fflr_enabled);
}

uint32_t CameraMDI::blob_size()
{
    // Series3 cameras have a smaller transfer "chunk"/blob.
    if (is_series3())
        return FSP_COMM::BLOB_CHUNK_SIZE;

    // New cameras can pull an entire image. Make our blob size as big as our
    // biggest image plus "a skosh." Referenced as DL_CHUNK_SIZE on the camera.
    return (2560 * 2048 * 4);
}
/*!
 * @brief Once all the data has been read from the camera, this method is called to "finalize" things.
 */
void CameraMDI::block_download_finish()
{
#if LOG_BLOCK_DOWNLOAD_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "block_download_id:" << Qt::showbase << Qt::hex << block_download_id << "pm_download_only:" << pm_download_only;
#endif

    // Process based on what we're doing.
    switch (block_download_id)
    {
    case GIGE_REG_ACTIVE_METADATA:
        // All done.
        pm_download_in_progress = false;

        // Did we only want a copy of the partition metadata, or did we
        // want to initiate a save operation?
        if (pm_download_only)
        {
            // Reset our flag.
            pm_download_only = false;
        }

        // Make sure we have our save object.
        else if (save_video != nullptr)
        {
#if LOG_ACTIVE_METADATA_DOWNLOAD
            // Dump the pm_metadata.
            const pm_metadata *meta = (const pm_metadata *)partition_metadata.constData();

            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "pm_metadata model_name:" << meta->model_name << "camera_name:" << meta->camera_name;
#endif
            // We don't save pm_metadata when saving RDI packed formatted streams.
            if (save_video->getStreamType() != FI_SAVER::STREAM_TYPE::RDI_PACKED)
            {
                // Save the metadata.
                save_video->savePartitionMetadata(&partition_metadata, read_camera_cache(GIGE_REG_RED), read_camera_cache(GIGE_REG_GREEN), read_camera_cache(GIGE_REG_BLUE));
            }

            // Are we supposed to automatically download the raw metadata?
            if (camera_ptr && camera_ptr->stream && autodownload_metadata)
            {
                // Connect our signal/slots for raw PFMD download.
                connect(this, &CameraMDI::video_is_paused, this, &CameraMDI::raw_metadata_download, Qt::QueuedConnection);

                // Pause the stream while we're downloading metadata.
                QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "autodownload_metadata";
                pause_video_request(__PRETTY_FUNCTION__);
            }
            else
            {
                QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "No autodownload_metadata.";

                // Pretend that we downloaded the raw metadata, and move on.
                save_video->start(__PRETTY_FUNCTION__);
            }
        }

        // Let folks know that we've gotten the metadata.
        emit pm_metadata_received(get_ip_addr());
        break;
    case GIGE_REG_CONFIGURATION:
        config_file.close();
        break;
    default:
        break;
    }

    // Reset our ID.
    block_download_id = (int)FI_COMM::NO_REGISTER;
    block_download_len = 0;
}

/*!
 * @brief A continuation of the pulling of a memory register's data from the camera.
 */
void CameraMDI::block_download_next()
{
    QByteArray  *buf_array;

#if LOG_BLOCK_DOWNLOAD_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "block_download_id:" << Qt::showbase << Qt::hex << block_download_id;
#endif

    // Sanity check.
    if (block_download_id == (int)FI_COMM::NO_REGISTER)
        return;

    // Read the metadata array.
    buf_array = read_camera_buffer((uint32_t)block_download_id);
    if (buf_array == nullptr)
    {
        QString err_str = QString::fromUtf8("get_reg_buffer( %1 ) returned NULL!").arg(block_download_id);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        block_download_finish();
        return;
    }

    // Do we need to get another chunk?
    block_download_len -= buf_array->size();

#if LOG_BLOCK_DOWNLOAD_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "block_download_len:" << block_download_len;
#endif

    // Sanity check the length.
    if (block_download_len < 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "The block_download_len length went negative!";

        // We are done.
        block_download_id = (int)FI_COMM::NO_REGISTER;
        block_download_len = 0;
        pm_download_only = false;
    }

    else if (block_download_len > 0)
    {
        uint32_t    read_len;

        // Figure out how much data we can read.
        if (block_download_len > block_download_max_xfer)
            read_len = (uint32_t)block_download_max_xfer;
        else
            read_len = (uint32_t)block_download_len;

        // Start the transfer.
        start_camera_read_buf((uint32_t)block_download_id, read_len);
    }

    // Process the data we have.
    if (buf_array->size() > 0)
    {
        switch (block_download_id)
        {
        case GIGE_REG_ACTIVE_METADATA:
            // Append the data to our buffer.
            partition_metadata.append(*buf_array);
            break;
        case GIGE_REG_CONFIGURATION:
            config_file.write(buf_array->constData(), buf_array->size());
            break;
        default:
            break;
        }
    }

    // Are we done?
    if (block_download_len == 0)
        block_download_finish();

    // All done with the data array.
    delete buf_array;
}

/*!
 * @brief Start the process of downloading the "block" data from the camera that
 *        is specified by the block_download_id.
 *
 * @param reg_addr      The register address for the memory block we want to download.
 * @param mem_len       The number of bytes we're expecting to read.
 * @param max_xfer      The maximum number of bytes per transfer.
 */
void CameraMDI::block_download_start(uint32_t reg_addr, int mem_len, int max_xfer)
{
    uint32_t    reg_value = 0;

    if (block_download_id != (int)FI_COMM::NO_REGISTER)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Already downloading id:" << block_download_id;
        return;
    }

    // Save the address and length for later.
    block_download_id = (int)reg_addr;
    block_download_len = mem_len;
    block_download_max_xfer = max_xfer;

#if LOG_BLOCK_DOWNLOAD_INFO
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << "ip_addr:" << get_ip_addr_pretty()
                 << "block_download_id:" << block_download_id
                 << "block_download_len:" << block_download_len
                 << "block_download_max_xfer:" << block_download_max_xfer;
#endif

    // Figure out if we need to read in chunks.
    if (block_download_len > max_xfer)
        reg_value = (uint32_t)max_xfer;
    else
        reg_value = (uint32_t)block_download_len;

    // Start the transfer.
    start_camera_read_buf(reg_addr, reg_value);
}

/*!
 * @brief Walk through all of the metadata in the [buffer](@ref raw_metadata) and
 *        copy only the event marker information over to a newly allocated buffer.
 *
 * @note    It is the responsibility of the [video review object](@ref CameraReview)
 *          to release the allocated buffer.
 */
void CameraMDI::build_data()
{
    int frame_no;
    int num_frames;
    uint8_t  *data_ptr;
    int meta_index;

    // How many frames are we dealing with?
    num_frames = raw_metadata_len / get_pfmd_len();

    // Allocate our data buffer.
    data_ptr = new uint8_t[ num_frames ];
    if (data_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate memory!";
        return;
    }

    // The third value (offset=2) is the value with all the flags.
    meta_index = 2 * sizeof(uint32_t);

    // Loop through all the frames so we can get the marker information.
    for (frame_no = 0; frame_no < num_frames; ++frame_no)
    {
        uint32_t temp_word;

        // Copy the data over.
        memcpy(&temp_word, raw_metadata.constData() + meta_index, sizeof(temp_word));

        // Save the marker information into our buffer.
        if (is_series3())
            data_ptr[ frame_no ] = (uint8_t)((temp_word >> FM_EVENT_FILTERS::EVENT_FLAG_OFFSET) & 0xFF);
        else
            data_ptr[ frame_no ] = (uint8_t)((temp_word >> 8) & 0xFF);

        // Move to the next frame's record.
        meta_index += get_pfmd_len();
    }

    // Pass the data on to the video control panel. It is the responsibility
    // of the review controls object to release the data_ptr memory.
    review_controls->set_graph_data(data_ptr, num_frames);
}

/*!
 * @brief Given a pointer to the per-frame metadata of a frame, pfmd_ptr,
 *        compute the straight-binary-seconds for when that frame was captured.
 *
 * @param[in] pfmd_ptr      Pointer to the frame's PFMD.
 * @return The SBS time the frame was captured, or 0.0 if an error occurred.
 */
double CameraMDI::calc_frame_time_from_PFMD(const uint8_t *pfmd_ptr)
{
    uint32_t frame_sbs, frame_micro_sec;
    double frame_time(0.0);

    // Do we have the raw PFMD array?
    if (pfmd_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Null pfmd_ptr?!";
        goto calc_bail;
    }

    // Get the straight-binary-seconds information from the PFMD.
    memcpy(&frame_sbs, pfmd_ptr, sizeof(frame_sbs));

    // Move to the next field for the microseconds.
    memcpy(&frame_micro_sec, pfmd_ptr + sizeof(frame_sbs), sizeof(frame_micro_sec));

    // Calculate the trigger frame start time.
    frame_time = ((double)frame_sbs + (((double)(frame_micro_sec & 0xFFFFF)) / 1000000.0));

calc_bail:
    return frame_time;
}

/*!
 * @brief Compute the maximum number of frames that can be recorded based on the capture
 * mode that is set for this camera.
 *
 * @return The maximum number of frames based on configured capture mode and record settings.
 */
uint64_t CameraMDI::calc_max_frames()
{
    uint64_t max_frames     = 0;
    uint64_t session_size   = 0;    // How big is the space to record into in bytes.

    // Pull the current ROI and bit depth.
    uint32_t bit_mode       = read_camera_cache(GIGE_REG_BIT_MODE);
    uint32_t height         = read_camera_cache(GIGE_REG_HEIGHT);
    uint32_t width          = read_camera_cache(GIGE_REG_WIDTH);

    // Pull the current Partition information.
    uint64_t part_size = ClientMath::qword_to_bytes(read_camera_cache(GIGE_REG_PART_SIZE));
    uint64_t part_size_reserved = (uint32_t)ClientMath::qword_to_bytes(read_camera_cache(GIGE_REG_PART_SIZE_RESERVED));

    // Calculation will depend on the capture mode of operation.
    uint32_t ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);
    bool ltr_enabled = (ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;
    bool fflr_enabled = (ltr_mode & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR;

    // We want to utilize our existing math routine for frame calculation but it will depend
    // on the capture mode as to how this is calculated.
    if (ltr_enabled)
    {
        // Calculate the maximum frame count for long record capture mode.
        session_size = ClientMath::sector_to_bytes(read_camera_cache(GIGE_REG_LTR_SECTOR_COUNT));

        // Calculate the long record frame count using the ssd_bandwidth as record controls uses.
        max_frames = ClientMath::calc_frame_count(session_size, width, height, bit_mode, true, is_series3());
    }
    else if (fflr_enabled)
    {
        // How many 256 MiB chunks of space has user reserved for FFLR?
        uint64_t sessions_reserved = (fflr_session_len * CLIENT_MATH::SESSION_INCR_QWORDS);

        // How many bytes is sessions_reserved?
        session_size = ClientMath::qword_to_bytes(sessions_reserved);

        // How many partitions fit in the space_to_reserve?
        uint64_t part_count = session_size / part_size;

        // Calculate the size of one FFLR partition.
        uint64_t part_size_for_one = part_size - part_size_reserved;

        // Calculate the max frames for one partition.
        uint64_t max_frames_for_one = ClientMath::calc_frame_count(part_size_for_one, width, height, bit_mode, ltr_enabled, is_series3());

        // Multiply the max_frames_for_one * partition count.
        max_frames = max_frames_for_one * part_count;
#if 0
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fflr_session_len:" << Qt::showbase << Qt::hex << fflr_session_len
                     << "sessions_reserved:" << Qt::showbase << Qt::hex << sessions_reserved
                     << "session_size:" << Qt::showbase << Qt::hex << session_size
                     << "part_count:" << Qt::showbase << Qt::hex << part_count
                     << "max_frames:" << Qt::showbase << Qt::hex << max_frames;
#endif
    }
    else
    {
        // It must be image memory capture mode.  Calculate as normally would.
        max_frames = ClientMath::calc_frame_count((part_size - part_size_reserved), width, height, bit_mode, ltr_enabled, is_series3());
    }
    return max_frames;
}

/*!
 * @brief Compute the straight-binary-seconds of the frame specified by the
 *        frame_num parameter from the per-frame metadata information.
 *
 * @param[in] frame_numThe frame we are interested in.
 * @return The SBS time the frame_num frame was captured, or 0.0 if an error occurred.
 */
double CameraMDI::calc_frame_time(uint32_t frame_num)
{
    double frame_time(0.0);

    // Are they asking for the trigger frame, and do we have the trigger frame time?
    if ((frame_num == review_controls->get_trigger_frame_num()) && !qFuzzyIsNull(trigger_frame_start))
    {
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Returning pre-calculated trigger time:" << trigger_frame_start;
        return trigger_frame_start;
    }

    // Do we have the raw PFMD array?
    if (!raw_metadata.isEmpty())
    {
        int raw_offset;

//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Using raw_metadata for frame:" << frame_num;

        // Calculate the (BYTE) offset to the frame's information.
        raw_offset = (int)frame_num * get_pfmd_len();

        // Sanity check.
        if (raw_offset < raw_metadata.size())
        {
            uint32_t raw_buffer[4];

            // Copy the data into a buffer.
            memcpy(raw_buffer, raw_metadata.constData() + raw_offset, sizeof(raw_buffer));

            // Are we series3?
            if (is_series3())
            {
                // Swap the bytes.
                raw_buffer[0] = qToBigEndian(raw_buffer[0]);
                raw_buffer[1] = qToBigEndian(raw_buffer[1]);
                raw_buffer[2] = qToBigEndian(raw_buffer[2]);
                raw_buffer[3] = qToBigEndian(raw_buffer[3]);
            }

            // Calculate the trigger frame start time.
            frame_time = calc_frame_time_from_PFMD((const uint8_t *)raw_buffer);
        }
    }

    // If we couldn't calculate a frame time, is this a frame we might have the PFMD cached for?
    if (qFuzzyIsNull(frame_time))
    {
        const QByteArray *data_ptr = pfmd_info->retrieve(frame_num);
        if (data_ptr != nullptr)
        {
            // Calculate the frame time using the cached PFMD.
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Using the cached PFMD for frame:" << frame_num;
            frame_time = calc_frame_time_from_PFMD((const uint8_t *)data_ptr->constData());
        }
        else
        {
            // Query the camera server to get the associated PFMD for the frame.
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No PFMD is available for frame:" << frame_num;
            request_pfmd(frame_num, false);
        }
    }

    // Return the time we came up with.
    return frame_time;
}

/*!
 * @brief Calculate the trigger frame time.
 *
 * It is assumed that the trigger_frame value is already initialized.
 *
 * @param[in] start_frame   The INCLUSIVE start frame number of the clip.
 * @param[in] end_frame     The INCLUSIVE ending frame number of the clip.
 */
void CameraMDI::calc_trigger_frame_start(uint32_t start_frame, uint32_t end_frame)
{
    // Do we already have the trigger frame start time calculated?
    if (!qFuzzyIsNull(trigger_frame_start))
    {
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "trigger_frame_start is pre-calculated:" << trigger_frame_start;
        return;
    }

    // Make sure we have the raw PFMD array.
    if (raw_metadata.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "raw_metadata is empty!";
        return;
    }

    // Do we have the per-frame metadata for the trigger frame?
    if ((trigger_frame >= start_frame) && (trigger_frame <= end_frame))
    {
        // Calculate the time for the trigger frame.
        trigger_frame_start = calc_frame_time(trigger_frame);
    }
    else
    {
        double rel_frame_interval;

//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString("The trigger_frame [") + QString::number(trigger_frame) + QString("] is outside the range");

        // The trigger frame is outside of the range of frames, so we will need to
        // make the assumption that between the start (or end) frame of the range
        // and the trigger frame, there are no discontinuities.
        if (trigger_frame < start_frame)
        {
            // Get the relative time between the start frame and the trigger frame.
            rel_frame_interval = (double)(((int)trigger_frame) - ((int)start_frame));
            rel_frame_interval = rel_frame_interval / (double)camera_info->get_fps();

            // Get the time for the start frame and then we'll assume no discontinuities.
            trigger_frame_start = calc_frame_time(start_frame) + rel_frame_interval;
        }
        else
        {
            // Get the relative time between the end frame and the trigger frame.
            rel_frame_interval = (double)(((int)trigger_frame) - ((int)end_frame));
            rel_frame_interval = rel_frame_interval / (double)camera_info->get_fps();

            // Get the time for the end frame and then we'll assume no discontinuities.
            trigger_frame_start = calc_frame_time(end_frame) + rel_frame_interval;
        }
    }

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "trigger_frame_start (calc):" << trigger_frame_start;
}

/*!
 * @brief Utility method which checks to see if a capture file (CAP) is
 *        allowed/supported on the connected camera.
 * @return True if CAP files are allowed, false otherwise.
 */
bool CameraMDI::cap_is_allowed()
{
    bool        rc = false;
    uint32_t    reg_value;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // Nobody allows cap files in LR mode.
    reg_value = read_camera_cache(GIGE_REG_LTR_MODE);
    if ((reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
        return false;

    // Are we a series3 camera?
    if (is_series3())
    {
        // Check to see if the camera has an SSD mounted.
        reg_value = read_camera_cache(GIGE_REG_AVAIL_MEDIA);
        if ((reg_value & GIGE_MEDIA_SSD) == GIGE_MEDIA_SSD)
        {
            // Check to hardware revision of the camera.
            if ((camera_ptr->board_rev[0] == FI_COMM::CAM_BOARD_REV_A) && (camera_ptr->board_rev[1] >= FI_COMM::CAM_BOARD_REV_A3))
            {
                // Capture files are allowed!
                rc = true;
            }
        }
    }
    else
    {
        // Newer cameras allow CAP on all media.
        rc = true;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief A method to allow a caller to modify the camera's expectation of when to expect
 *        "pings". We will always go faster than this rate.
 * @param new_rate  The new rate, in milliseconds. Must be greater than or equal to GIGE_HEARBEAT_TIMEOUT.
 */
void CameraMDI::change_heartbeat(uint32_t new_rate)
{
    // Sanity check.
    if (new_rate < FASMOTION_HEARTBEAT)
        new_rate = FASMOTION_HEARTBEAT;

    // Tell the camera the new rate.
    write_camera_register(GIGE_REG_HEARTBEAT, new_rate);
}

/*!
 * @brief Cleanup the save UI elements.
 */
void CameraMDI::cleanup_save()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Are we in a multicamera save?
    if (multicam_save_allowed)
    {
        MultiCamDock    *dock_ptr;

        // Get a pointer to the multicam dock.
        dock_ptr = MainWindow::GetInstance()->get_multicam_dock();
        if (dock_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No multicam dock!?";
            return;
        }

        // Reset our flag.
        multicam_save_allowed = false;

        // Let the multicamera dock know that we're done.
        dock_ptr->autosave_complete(this);
    }

    // Bring down the progress dialog.
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Make sure our saving object is closed.
    if (save_video != nullptr)
        save_video->close(__PRETTY_FUNCTION__);

    // Do we need to clean overlay stuff up?
    if (save_video && save_video->do_merge_overlay())
        video_player->render_end();

    // We are done with the transfer progress dialog.
    if (xferDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(xferDialog);
    }
}

/*!
 * @brief Assuming we've read all the WIFI related registers from the camera,
 *        display the configuration dialog so the user can set things up.
 */
void CameraMDI::configure_wifi()
{
    WifiConfig  config_dlg(this);
    config_dlg.setMinimumSize(329, 239);

    // Pop up the dialog.
    config_dlg.exec();
}

/*!
 * @brief The basic copy procedure has finished. This method is used to
 *        check for any files or "chunks" which might need to be resent.
 */
void CameraMDI::copy_finished(QString whoCalledMe)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "whoCalledMe:" << whoCalledMe << "retry_in_progress:" << retry_in_progress << "retry_count:" << get_retry_count();

    // Sanity check.
    if (save_video == nullptr)
        return;

    // TODO:  We have disconnected signal/slot handlers - we won't properly
    // process retry frames if we go this way...  I have commented out the call
    // to retry_missing_frames...

    // Any files/chunks to be retried?
    if (!retry_frame_list.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "retry list is not empty:" << retry_frame_list;
        //retry_missing_frames();
        return;
    }

    // Let the video saver object know we're done.
    save_video->saving_finished(__PRETTY_FUNCTION__);
}

/*!
 * @brief Handle timeouts while copying a video to the workstation.
 */
void CameraMDI::copy_video_timeout()
{
    QString     infoStr;
    uint32_t    frameCount;
    uint32_t    framesAdded;
    uint32_t    framesSaved;
    bool        streamIsOk;
    bool        streamIsComplete;
    QString     err_str;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "last_received_frame:" << last_received_frame << "frame_count:" << save_video->getFrameCount();

    // Make sure that we haven't lost the very last frame.
    // We will do this for both series3 and seriesNC cameras because it is
    // theoretically possible seriesNC cameras using TCP/IP could have a missed
    // frame in the streamer process.

    // Did we lose the last frame in the copy stream?
    if (last_frame_missing_check())
    {
        // Start the retry missing frame process.
        if (!retry_in_progress)
            retry_missing_frames();
        return;
    }

    // Otherwise it must be determined if the copy status is possibly completed.  It could be
    // possible that the system the way it is configured might miss a copy finished condition
    // leading to a time-out that isn't really valid.

    // What is the state of this save operation.  A copy is just a save of a blob(s) to be clear here.
    save_video->getSaveStatus(infoStr, frameCount,
                              framesAdded, framesSaved, streamIsOk, streamIsComplete);

    // Is the stream OK and are we completed?
    if (streamIsOk && (frameCount == framesSaved))
    {
        // Make sure there are no retry frames and if so we can assume this copy completed.
        if (retry_frame_list.isEmpty())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "calling copy_finished() frameCount:" << frameCount << "framesSaved:" << framesSaved;
            // All we do at this point is tell the save video we should be done.
            copy_finished(__PRETTY_FUNCTION__);

            // Go ahead and just return here.  We should be ok.  If not
            // the user can cancel the progress dialog.
            return;
        }
    }

    // Let the user know that things have gone badly.
    err_str = tr("The copy operations failed due to a timeout waiting for frames.");
    if (retry_in_progress || (last_received_frame > 0))
    {
        err_str.append(tr("<p>The last frame received was frame %1.").arg(last_received_frame - 1));
    }
    else
    {
        err_str.append(tr("<p>No frames have been received."));
    }
    // Add the information from the save status query.
    err_str.append(tr("<p>%1").arg(infoStr));

    QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
    QMessageBox::critical(this, get_camera_name(), err_str);

    // Let folks know we've had troubles.
    emit copy_timeout();
}

/*!
 * @brief Disarm the camera.
 *
 * @return True if we started the dearm process, false otherwise (e.g., the user cancelled).
 */
bool CameraMDI::dearm_camera(bool user_queried, bool allow_discard)
{
    bool        keep_session(true);
    bool        yes_means_dearm(true);
    int         user_response = QMessageBox::Cancel;
    QMessageBox::StandardButtons buttons;
    bool        fflr_is_enabled;
    bool        is_roc_broc;
    QString     str_text;
    uint32_t    new_state;
    uint32_t    fflr_type;

    // Determine if we're in FFLR mode.
    fflr_is_enabled = is_fflr_enabled(fflr_type);
    is_roc_broc = ((fflr_type & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC)) != 0);

    // Has the user already been asked about this?
    if (!user_queried)
    {
        // Build up our string and which buttons we'll show.
        if (allow_discard)
        {
            bool nothing_to_save;
            uint32_t camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
            uint32_t trig_pos = read_camera_cache(GIGE_REG_TRIGGER_REQ_POS) & GIGE_TRIGGER_REQ_POS_MASK;
            uint32_t cur_frame = read_camera_cache(GIGE_REG_CUR_FRAME_POS);

            // If we are in an "append" mode (i.e., ROC/BROC) we need to look to see if we've
            // captured any frames.
            if (is_roc_broc)
                nothing_to_save = (cur_frame == 0);

            // Otherwise, we are in a "basic" mode. If we are using a start trigger, see if we
            // have triggered.
            else if (trig_pos == 0)
                nothing_to_save = (camera_state != GIGE_CAMERA_STATE_TRIGGERED);

            // At this point, we must have SOMETHING to save.
            else
                nothing_to_save = false;

            // Different message box depending on our state.
            if (nothing_to_save)
            {
                str_text = tr("Are you sure you want to exit Video Capture? You may select: \n"
                              "  * 'Yes' to exit the empty video capture, or  \n"
                              "  * 'Cancel' to keep recording.\n");
                buttons = QMessageBox::Yes | QMessageBox::Cancel;

                // In this start-trigger case, yes means to cancel the capture and go live there is nothing to discard.
                yes_means_dearm = false;
            }
            else
            {
                // Here if we are not a start trigger, or we are FFLR ROC or BROC.
                str_text = tr("Are you sure you want to exit Video Capture? You may select: \n"
                              "  * 'Yes' to retain the current session,  \n"
                              "  * 'No' to discard the current session, or  \n"
                              "  * 'Cancel' to keep recording.\n");
                buttons = QMessageBox::Yes | QMessageBox::No | QMessageBox::Cancel;
            }
        }
        else
        {
            str_text = tr("Are you sure you want to exit Video Capture? You may select: \n"
                          "  * 'Yes' to exit Video Capture so you can Review or Save the recording.\n"
                          "  * 'Cancel' to remain in Video Capture, where you can use the Trigger button to run and pause (ROC) or to capture a burst (BROC).\n");
            buttons = QMessageBox::Yes | QMessageBox::Cancel;
        }

        // Ask the user if they want to save or discard the currently captured frames, even
        // though it's a partial capture.
        std::unique_ptr<fiProgressDialog> dearm_progress_dlg = fiProgressDialog::CreateSimpleProgressDialog(get_camera_name(), str_text, buttons, __PRETTY_FUNCTION__);
        if (dearm_progress_dlg)
        {
            dearm_progress_dlg->showStatusIcon(true);
            dearm_progress_dlg->showStatusValue(true);
            dearm_progress_dlg->showBar(false);
            dearm_progress_dlg->setStatusIcon(QStyle::SP_MessageBoxQuestion);
            dearm_progress_dlg->setStatusValue(str_text);
            if (save_video)
            {
                // Ticket #516: We will end up with this message box and the FasFireLR underflow processing if we don't
                // circumvent here and use our own progress dialog class here. It must go away when FasFireLR is recovering.
                // If this error arises we want force a cancel on this progress dialog as it is handled by a different mechanism.
                connect(save_video, &SaveVideo::save_fflr_underflow_error, dearm_progress_dlg.get(), &fiProgressDialog::cancelDialog, Qt::QueuedConnection);
            }

            // Run the dialog in execute form. We don't care about the result here as we can query our
            // dialog itself to see what was hit to cause the dialog to go away.
            dearm_progress_dlg->exec();

            // Translate our simple progress dialog result into the correct user response code for this function.
            if (dearm_progress_dlg->yesSignaled())
                user_response = QMessageBox::Yes;
            else if (dearm_progress_dlg->noSignaled())
                user_response = QMessageBox::No;
            else
                user_response = QMessageBox::Cancel;

            // Disconnect from the error condition.
            if (save_video)
            {
                disconnect(save_video, &SaveVideo::save_fflr_underflow_error, dearm_progress_dlg.get(), &fiProgressDialog::cancelDialog);
            }
            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "dlg_result:" << dlg_result << "user_response:" << Qt::showbase << Qt::hex << user_response;
            fiProgressDialog::RecycleProgressDialog(dearm_progress_dlg);
        }
        // The next line is the QMessageBox that existed instead of our custom one above in case we decide
        // to do something different.
        //user_response = QMessageBox::question(this, get_camera_name(), str_text, QMessageBox::StandardButtons(buttons));
        if (user_response == QMessageBox::Cancel)
            return false;     // Bail out.
    }
    else
    {
        // The user has already been asked about saving, etc. Just do it now. If the user said "discard", then discard.
        if (allow_discard)
            user_response = QMessageBox::No;
        else
            user_response = QMessageBox::Yes;
    }

    // The user wants to cancel, the only question is how?
    if (yes_means_dearm && user_response == QMessageBox::Yes)
    {
        // Move the camera to the "dearming" state.
        new_state = GIGE_CAMERA_STATE_DEARMING;
        if (save_video)
        {
            save_video->dearm_and_keep();
        }
    }
    else
    {
        // User doesn't want to save the data, so continue on the "stop recording" process.
        // Move the camera to Live video.
        new_state = GIGE_CAMERA_STATE_LIVE_VIDEO;

        // User doesn't want to keep the session.
        keep_session = false;
    }

    //QLOG_TRACE() << __PRETTY_FUNCTION__ << "new_state:" << Qt::showbase << Qt::hex << new_state << "keep_session:" << keep_session;

    // Write out the new state.
    write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, new_state);

    // Cleanup connections which were made to handle automatic transitions. Since
    // we are dearming, there is no longer a need for these.
    disconnect(this, &CameraMDI::raw_metadata_received, this, &CameraMDI::raw_metadata_ready);
    disconnect(this, &CameraMDI::raw_metadata_received, save_video, &SaveVideo::raw_metadata_ready);

    // Emit the disarm condition so any save operations can respond appropriately
    // based on the user's selection.  One instance where this is important is when
    // the system is in an autosave state.  In that case our save video object may be
    // listening for these signals.  Added to address concerns in Ticket# 253.
    if (!keep_session)
    {
        // Is there a slot connected to our signal?
        if (isSignalConnected(QMetaMethod::fromSignal(&CameraMDI::dearm_session_discard)))
        {
            emit dearm_session_discard();
        }
        else
        {
            uint32_t autosave_settings = read_camera_cache(GIGE_REG_AUTOSAVE);

            // See if we are autosaving.
            if (autosave_settings & GIGE_AUTOSAVE_SAVING)
            {
                // If we were autosaving, then we definitely want to cancel it on a de-arm.
                cancel_save();
            }
        }

        // If this is a FasFireLR de-arm let it persist that the user has discarded the recorded session partition(s).
        if (fflr_is_enabled)
        {
            fflr_user_discards_parts = true;
        }
    }
    else
    {
        emit dearm_session_keep();
    }

    // Dearm has started.
    return true;
}

/*!
 * @brief Provides the camera object an opportunity to make sure that if in FasFireLR mode
 * it has prepared the autosave process prior to an arm event whether internal or external.
 *
 * @returns True if the camera is prepared for arming, false if not.
 */
bool CameraMDI::fflr_check_before_arming()
{
    bool prepared = false;

    // Make sure the saver object is allocated.
    if (save_video == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "save_video object not allocated?!";
        goto fflr_prepare_to_arm_exit;
    }

    // This only makes sense if camera is FasFireLR enabled.
    uint32_t fflr_type;
    if (!is_fflr_enabled(fflr_type))
    {
        goto fflr_prepare_to_arm_exit;
    }

    // Only want to run the autosave set up if we aren't already enabled and configured.
    if (!save_video->autosave_is_enabled() || !save_video->autosave_is_configured())
    {
        // If we are not in a live or review state we cannot set up the autosave.
        uint32_t camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
        if ((camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO) && (camera_state != GIGE_CAMERA_STATE_REVIEW))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The camera state is not in live or review!";
            goto fflr_prepare_to_arm_exit;
        }

        // Tell the saver to set up the autosave state.
        // In order to process FasFireLR we set up an autosave for the camera and populate the data
        // with FFLR settings.  The user is not queried for the setting information in this process.
        // Note: autosave_setup() calls save_video->autosave_start() internally.
        if (!save_video->autosave_setup())
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "FasFireLR Error: save_video->autosave_setup() returned false";
            goto fflr_prepare_to_arm_exit;
        }

        // Make sure the camera server resets the FFLR status.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Reset FFLR status +++++++++++++++++++++++++++++++++++++";
        write_camera_register(GIGE_REG_FFLR_JSON, 0);

        // Reset any processing underflow condition and discarded state.
        fflr_underflow_discarded = false;

        // Reset this for this FasFireLR session (this is where it's initialized false other than constructor).
        fflr_user_discards_parts = false;
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "autosave_is_enabled:" << save_video->autosave_is_enabled() << "autosave_is_configured:" << save_video->autosave_is_configured();
    }

    prepared = true;    // READY to go.

fflr_prepare_to_arm_exit:
    return prepared;
}

/*!
 * @brief Start the black reference frame calibration process.
 * @param do_factory    True if this is a factory black frame, false for a "normal" user black frame.
 */
void CameraMDI::fpn_calibration(bool do_factory)
{
    QString title_str;
    CalibrateDlg    cal_dlg(this);

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // warn the user that this is going to cause a camera reboot.
    if (do_factory)
    {
        // Make sure the user knows that this will cause the camera connect to be reset.
        int user_resp = QMessageBox::question(this, tr("Factory Black Level Calibration"),
                                              tr("It is assumed that:<br>"
                                                 "1) You have the camera pointed at a flat-field, gray image.<br>"
                                                 "2) You are using the lowest 8-bits.<br>"
                                                 "3) Your light source is stable (i.e. NOT fluorescent!)<br>"
                                                 "4) You have verified that the mean pixel values are between 50-75% full "
                                                 "scale under sampling conditions.<br><br>"
                                                 "Are you sure you want to perform a factory black level calibration?<br>"
                                                 "This will require the connection to be closed and reopened. Proceed?"),
                                              QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
        if (user_resp == QMessageBox::No)
            return;

        // Setup our title string.
        title_str = tr("Factory Black Level Calibration");
    }
    else
    {
        // Setup our title string.
        title_str = tr("Black Level Calibration");
    }

    // Display the calibration dialog.
    cal_dlg.set_title(title_str);
    if (cal_dlg.exec() == QDialog::Rejected)
        return;

    // Are we doing automatic or manual offset calibration?
    auto_offset_cal = cal_dlg.do_auto_offset();

    // Make sure we indicate that we're initiating a calibration event.
    calibration_event = (do_factory ? CAL_EVENT_FACTORY : CAL_EVENT_USER);

    // Connect our signal/slots for calibration.
    connect(this, &CameraMDI::video_is_paused, this, &CameraMDI::calibration_continuation);

    // If the camera should be off-line, but we have a stream running, then
    // shutdown the stream.
    if (camera_ptr->stream_running)
    {
        // Pause the stream while we're calibrating.
        pause_video_request(__PRETTY_FUNCTION__);

        // Check to see if we're the "active" camera (i.e. not in the background).
        if (active_window())
        {
            // Since we're the active camera, let the user know what's going on.
            MainWindow::GetInstance()->set_statusbar_message(tr("Camera is calibrating...please wait."), 3000);
        }
    }
    else
    {
        // Carry on with the calibration process.
        emit video_is_paused();
    }

    // Create and show a simple progress dialog for calibration.
    progressDialog = fiProgressDialog::CreateSimpleProgressDialog(title_str, tr("Camera is calibrating...please wait."), QMessageBox::NoButton, __PRETTY_FUNCTION__, this);

    // Display the progress dialog.
    if (progressDialog)
    {
        progressDialog->setBarMinMax(0, 0);
        progressDialog->showBar(true);
        progressDialog->showDialogWithoutPolling();
    }

    // Setup a single-shot timer to catch the case when the camera dies. We don't
    // want the progress dialog just hanging the application with no way to recover.
    QTimer::singleShot(CALIBRATION_TIMEOUT, this, &CameraMDI::progress_check);
}

/*!
 * @brief Setup for color PRNU calibration.
 * @return QMessageBox::Yes if the user wants to continue, QMessageBox::No if the user cancels.
 */
int CameraMDI::setup_color_prnu_calibration()
{
    int user_resp = QMessageBox::No;
    QJsonArray media_array;
    QJsonObject json_obj;
    QJsonDocument ap_doc;
    SelectFolder sel_stack(this);
    SelectDevice sel_device(this);
    QString red_path;
    QString green_path;
    QString blue_path;

    // See if we can support a local path.
    sel_device.localPathAllowed(MainWindow::GetInstance()->local_control());

    // Have the user select the device to load from.
    if (sel_device.exec() == QDialog::Rejected)
        goto bail;

    // Retrieve the info from the selected object.
    if (sel_device.deviceSelected(json_obj, media_array) == -1U)
    {
        // Get the red stack.
        red_path = QFileDialog::getExistingDirectory(MainWindow::GetInstance().get(), CameraMDI::prnu_red_path, SaveDialog::loadPath(0));
        if (red_path.isEmpty())
            goto bail;

        // Get the green stack.
        green_path = QFileDialog::getExistingDirectory(MainWindow::GetInstance().get(), CameraMDI::prnu_green_path, SaveDialog::loadPath(0));
        if (green_path.isEmpty())
            goto bail;

        // Get the blue stack.
        blue_path = QFileDialog::getExistingDirectory(MainWindow::GetInstance().get(), CameraMDI::prnu_blue_path, SaveDialog::loadPath(0));
        if (blue_path.isEmpty())
            goto bail;
    }
    else
    {
        // Pass the media specific information to the stack selection dialog.
        sel_stack.setJsonInfo(json_obj, media_array);

        // Have the user select the Red stack.
        sel_stack.setWindowTitle(CameraMDI::prnu_red_path);
        if (sel_stack.exec() == QDialog::Rejected)
            goto bail;
        red_path = sel_stack.folderSelected();
        if (red_path.isEmpty())
            goto bail;

        // Have the user select the Red stack.
        sel_stack.setWindowTitle(CameraMDI::prnu_green_path);
        if (sel_stack.exec() == QDialog::Rejected)
            goto bail;
        green_path = sel_stack.folderSelected();
        if (green_path.isEmpty())
            goto bail;

        // Have the user select the Red stack.
        sel_stack.setWindowTitle(CameraMDI::prnu_blue_path);
        if (sel_stack.exec() == QDialog::Rejected)
            goto bail;
        blue_path = sel_stack.folderSelected();
        if (blue_path.isEmpty())
            goto bail;
    }

    // Tell the camera where to find the images.
    json_obj = QJsonObject();
    json_obj.insert("red", red_path);
    json_obj.insert("green", green_path);
    json_obj.insert("blue", blue_path);

    // Convert to a document.
    ap_doc = QJsonDocument(json_obj);

    // Send it to the camera.
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << ap_doc;
    write_camera_buffer(GIGE_REG_PRNU_JSON, ap_doc.toJson());

    // All is well.
    user_resp = QMessageBox::Yes;

bail:
    // Return our status.
    return user_resp;
}

/*!
 * @brief Generates the usable file_name for caller to allow access to a file in the system whether the
 *        camera is connected to a local (only for Linux) or remote (via Samba) camera server.
 *
 * @param[in] media_path        The path generated by the camera server for the file in question.
 * @param[in] name_part         The name part of the path if a unique name is separate from media_path.
 * @param[in] file_id           The identifier to use for the filename.
 * @param[out] smb_file_name    The resulting fully qualified path name to a Samba file.
 *
 * @returns True if successful, false otherwise.
 */
bool CameraMDI::prepare_smb_filepath(const QString& media_path, const QString& name_part, const QString& file_id, const QString& file_ext, QString& smb_file_name)
{
    bool preparedName(false);

    // Set our output to an empty state.
    smb_file_name.clear();

#if defined(Q_OS_WIN)
    // A Windows FasMotion cannot have a local camera_server because our camera_server currently runs on
    // linux systems. We know this is a remote camera_server situation and must use the camera_server's
    // IP address to create a Windows friendly path our system can understand.
    smb_file_name = QString("//") + get_ip_addr_pretty() + media_path;
#elif defined(Q_OS_LINUX)
    // Are we running local?
    if (is_local())
    {
        // Initialize our file path.
        smb_file_name = media_path;
    }
    else
    {
        QTemporaryFile argFile;
        QProcess mountProcess;
        QString fixPath;
        QString sysCommand;
        QStringList args;

        // Build the arguments file. Note that this MIGHT fail, and if it does then
        // we cannot continue.
        if (argFile.open())
        {
            QSettings settings;
            QString default_user;
            QString default_pass;
            QString workgroup;

            // Retrieve the user credentials to use for the camera.
            settings.beginGroup(UserPreferences::SERIESNC_PREFERENCES_SECTION);
            default_user = settings.value(UserPreferences::SERIESNC_CAM_USER_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
            default_pass = settings.value(UserPreferences::SERIESNC_CAM_PASS_PREF_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_NAME).toString();
            workgroup = settings.value(UserPreferences::SERIESNC_CAM_WORKGROUP_KEY, UserPreferences::SERIESNC_CAM_USER_DEFAULT_WORKGROUP).toString();
            settings.endGroup();

            // Build up the temporary file that contains the requried input for the "gio" mount command.
            // These are: user name, workgroup, and password.
            QTextStream outFile(&argFile);
            outFile << default_user + QString("\n") + workgroup + QString("\n") + default_pass + QString("\n");

            // Close the file so we can pass it on to the process for input.
            argFile.close();

            // Tell the system to mount the Samba share to the camera's media folder.
            sysCommand = QStringLiteral("/usr/bin/gio");
            args << "mount" << QString("smb://%1/media").arg(get_ip_addr_pretty());

            // Run the command. Use our temporary file for the input information.
            mountProcess.setStandardInputFile(argFile.fileName());
            mountProcess.start(sysCommand, args);
            mountProcess.waitForFinished();

            // Any errors?
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Output?" << mountProcess.readAllStandardOutput();
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Errors?" << mountProcess.readAllStandardError();

            // Build up the path to where the file is located.
            fixPath = qEnvironmentVariable(XDG_RUNTIME_DIR);

            // Build up the mount path.
            smb_file_name = fixPath + QString("/gvfs/smb-share:server=%1,share=").arg(get_ip_addr_pretty());

            // Trim the leading '/' off the media path and append it to our file path.
            fixPath = media_path.last(media_path.length() - 1);
            smb_file_name += fixPath;
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to create a temporary file!";
        }
    }
#elif defined(Q_OS_MAC)
    QProcess mountProcess;
    QString sysCommand;
    QStringList args;

    // Tell the system to mount the Samba share to the camera's media folder.
    sysCommand = QStringLiteral("/usr/bin/osascript");
    args << "-e" << QString("mount volume \"smb://guest:@") + get_ip_addr_pretty() + QString("\"");

    // Run the command.
    mountProcess.start(sysCommand, args);
    mountProcess.waitForFinished();

    // Any errors?
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Output?" << mountProcess.readAllStandardOutput();
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Errors?" << mountProcess.readAllStandardError();

    // Build up the path to where the file is located.
    smb_file_name = QStringLiteral("/Volumes") + media_path;
#endif

    // Do we have a path to something to open?
    if (!smb_file_name.isEmpty())
    {
        // Build up the path to where the file is located.
        if (!name_part.isEmpty())
        {
            if (file_id.toInt() >= 0)
                smb_file_name += QString("/") + name_part + QString("/%1.%2").arg(QString(file_id).toInt(), 7, 10, QLatin1Char('0')).arg(file_ext);
            else
                smb_file_name += QString("/") + name_part;
        }
        else
        {
            smb_file_name += QString("/%1.%2").arg(QString(file_id).toInt(), 7, 10, QLatin1Char('0')).arg(file_ext);
        }

        // Make sure we're using native separators.
        smb_file_name = QDir::toNativeSeparators(smb_file_name);

        // We are prepared to open this file in this OS.
        preparedName = true;
    }

    return preparedName;
}

/*!
 * @brief Start the PRNU calibration process.
 */
void CameraMDI::prnu_calibration()
{
    int user_resp;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (sensor_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Null sensor pointer!";
        return;
    }
    if (is_series3())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "PRNU is not supported on series3 cameras.";
        return;
    }

    // PRNU is not supported accross all sensors.
    switch (sensor_ptr->get_type())
    {
    case FI_SENSOR::SENSOR_TYPE_LUX13HS_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LUX19HS_COLOR:
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_COLOR:
        // Color PRNU requires some "special" setup.
        user_resp = setup_color_prnu_calibration();
        break;
    case FI_SENSOR::SENSOR_TYPE_LUX13HS_MONO:
    case FI_SENSOR::SENSOR_TYPE_LUX19HS_MONO:
    case FI_SENSOR::SENSOR_TYPE_LINCE5M_MONO:
        // Make sure the user knows that this will cause the camera connect to be reset.
        user_resp = QMessageBox::question(this, prnu_title_str, prnu_setup_msg, (QMessageBox::Yes | QMessageBox::No), QMessageBox::No);
        break;
    default:
        QLOG_WARN() << __PRETTY_FUNCTION__ << "PRNU is not supported on this sensor!";
        user_resp = QMessageBox::No;
        break;
    }

    // Did the user want to exit?
    if (user_resp == QMessageBox::No)
        return;

    // Make sure we indicate that we're initiating a calibration event.
    calibration_event = CAL_EVENT_PRNU;

    // Connect our signal/slots for calibration.
    connect(this, &CameraMDI::video_is_paused, this, &CameraMDI::calibration_continuation, Qt::QueuedConnection);

    // If the camera should be off-line, but we have a stream running, then
    // shutdown the stream.
    if (camera_ptr->stream_running)
    {
        // Pause the stream while we're calibrating.
        pause_video_request(__PRETTY_FUNCTION__);

        // Check to see if we're the "active" camera (i.e. not in the background).
        if (active_window())
        {
            // Since we're the active camera, let the user know what's going on.
            MainWindow::GetInstance()->set_statusbar_message(tr("Camera is calibrating...please wait."), 3000);
        }
    }
    else
    {
        // Carry on with the calibration process.
        emit video_is_paused();
    }

    // Create and show a simple progress dialog for calibration.
    progressDialog = fiProgressDialog::CreateSimpleProgressDialog(prnu_title_str, tr("Camera is calibrating...please wait."), QMessageBox::NoButton, __PRETTY_FUNCTION__, this);

    // Display the progress dialog.
    if (progressDialog)
    {
        progressDialog->setBarMinMax(0, 0);
        progressDialog->showBar(true);
        progressDialog->showDialogWithoutPolling();
    }

    // Setup a single-shot timer to catch the case when the camera dies. We don't
    // want the progress dialog just hanging the application with no way to recover.
    QTimer::singleShot(CALIBRATION_TIMEOUT, this, &CameraMDI::progress_check);
}

/*!
 * @brief Pulls the latest GIGE_REG_AUTOSAVE register and tells the camera information
 * object to update the buffers based on the register.
 */
void CameraMDI::refresh_autosave_buffers()
{
    uint32_t reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);
    if (camera_info)
        camera_info->display_autosave_settings(reg_value);
}

/*!
 * @brief Examines the partition state to determine if there are any dirty partitions
 * and asks the user about overwriting them.  If they choose to overwrite them then
 * the partition(s) are freed.
 *
 * @param free_all          True if we're to release all partitions, false to only release
 *                          the "active" partition.
 *
 * @param operation_label   A label to use for telling the user the operation information.
 *
 * @returns True if user choose to overwrite dirty partitions OR if none were dirty or false
 * if the caller does not want to overwrite any dirty partitions.
 */
bool CameraMDI::verify_dirty_overwrite(bool free_all_on_overwrite, const QString& operation_label)
{
    bool        verified_to_proceed     = false;
    uint32_t    dirty_count             = 0;        // If camera has unsaved partitions this will be non-zero.
    uint32_t    part_state;
    FI_comm     *comm_ptr;

    // TICKET #427 - Don't overload a dirty partition without warning the user about it.
    // Are there unsaved dirty partitions in image memory?

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        goto exit_verified_dirty_overwrite;
    }

    // Do we have a partition sitting unsaved in memory?
    part_state = comm_ptr->get_reg(get_ip_addr(), GIGE_REG_PART_STATE);
    if (part_state == GIGE_PART_STATE_COMPLETE)
    {
        uint32_t    reg_value;

        // Read the autosave register to determine if there is more than one dirty buffer.
        reg_value = comm_ptr->get_reg(get_ip_addr(), GIGE_REG_AUTOSAVE);
        if (reg_value == 0)
        {
            // There is at least this single partition that is dirty.
            dirty_count = 1;
        }
        else
        {
            // There could be more than one partition that is dirty.
            dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);
        }
    }

    // Ask the user if they want to continue with the load because they may lose data.
    if (dirty_count > 0)
    {
        int     user_resp;
        QString msg_part;
        QString msg_str = tr("***POSSIBLE DATA LOSS***\n%1 will overwrite recorded data.\n\nImage Memory ").arg(operation_label);

        // Build up the message based on the number of "dirty" partitions in image memory.
        msg_part = tr("contains %1 unsaved recording%2\n").arg(dirty_count).arg((dirty_count > 1 ? "s." : "."));

        // Make sure the user knows what they're doing.
        msg_str = msg_str + msg_part + tr("\nSelect \"OK\" to discard recorded data and proceed.\n"
                                          "Select \"Cancel\" to exit and protect unsaved data.");

        // Show the warning message.
        user_resp = QMessageBox::warning(this, tr("Unsaved Image Memory Partition(s)"), msg_str, QMessageBox::Ok | QMessageBox::Cancel, QMessageBox::Cancel);
        if (user_resp == QMessageBox::Cancel)
        {
            goto exit_verified_dirty_overwrite;
        }
        else if (user_resp == QMessageBox::Ok)
        {
            // Free the partition(s) based on the caller's setting.
            free_partition(free_all_on_overwrite);
            verified_to_proceed = true;
        }
    }
    else
    {
        // There are no dirty partitions so it is ok to proceed.
        verified_to_proceed = true;
    }

exit_verified_dirty_overwrite:
    return verified_to_proceed;
}

/*!
 * @brief Tell the camera to release one, or if the "free_all" parameter is set,
 *        all of the "dirty" partitions.
 *
 * If the "free_all" parameter is set, this becomes a "distributed" method.
 * We tell the camera to release the "active" partition and then exit the routine.
 *
 * When we get a [write completion](@ref CameraMDI::write_complete) notification,
 * and the "release_all" member variable is set true, this method is invoked again and
 * the next partition is released.
 *
 * @param free_all      True if we're to release all partitions, false to only release
 *                      the "active" partition.
 */
void CameraMDI::free_partition(bool free_all)
{
    int         dirty_count;
    uint32_t    reg_value;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Remember our flag.
    release_all = free_all;

    // Get the autosave settings.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

    // How many dirty partitions are there?
    dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);
    if (dirty_count > 1)
    {
        // Tell the CS to "pop" the partition off the stack.
        write_camera_register(GIGE_REG_AUTOSAVE, GIGE_AUTOSAVE_QUEUE_UPDATE);

        // Stop any streaming.
        pause_video_request(__PRETTY_FUNCTION__);

        // Reset our flags and variables.
        last_requested_frame = (uint32_t) -1;

        // What is the camera's state.
        uint32_t camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "dirty_count:" << dirty_count << "release_all:" << release_all << "free_all:" << free_all << "camera_state:" << camera_state;

        // Special processing if we're in review.
        if (camera_state == GIGE_CAMERA_STATE_REVIEW)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Begin of stutter - go live";

            // Do a stutter step from review to live back to review. This forces all
            // GUIs to reload their settings properly.
            write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_LIVE_VIDEO);

            // We only want to go back if we're not releasing all the buffers.
            if (!release_all)
            {
                // Delay for a bit, and then switch back to review.
                QTimer::singleShot(LIVE_TO_REVIEW_DELAY, this, &CameraMDI::do_camera_review);
            }
        }
    }
    else
    {
        // Is there 1 dirty buffer?
        if (dirty_count == 1)
        {
            // Tell the CS to "pop" the partition off the stack.
            write_camera_register(GIGE_REG_AUTOSAVE, GIGE_AUTOSAVE_QUEUE_UPDATE);
        }

        // Check to see if the last partition is dirty.
        reg_value = read_camera_cache(GIGE_REG_PART_STATE);
        if (reg_value != GIGE_PART_STATE_EMPTY)
            write_camera_register(GIGE_REG_PART_STATE, GIGE_PART_STATE_EMPTY);

        // Are we supposed to be streaming?
        if (camera_ptr->stream && !camera_ptr->stream_running)
            resume_video_request();

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "dirty_count:" << dirty_count << "release_all:" << release_all << "free_all:" << free_all << "camera_state:" << read_camera_cache(GIGE_REG_CAMERA_STATE);

        // We're done.
        release_all = false;
    }

    // Make sure we release our event information for the partition we just released.
    reset_event_info();

    // We need to reset the captured partition information. This will force a re-download of the
    // partition information when the system transitions to the "new" review state.
    partition_metadata.clear();
}

/*!
 * @brief Return a list of the "camera" actions (Arm, Trigger, Live, Review).
 */
QList<QAction *> CameraMDI::get_camera_actions()
{
    if (camera_info == nullptr)
    {
        QList<QAction *>    bad_ret;

        QLOG_ERROR() << __PRETTY_FUNCTION__ << "camera_info is NULL!";
        return bad_ret;
    }

    return camera_info->get_camera_actions();
}

/*!
 * @brief Return the user specified camera name.
 */
QString CameraMDI::get_camera_name()
{
    if (camera_info == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "camera_info is NULL!";
        return QString();
    }

    return camera_info->get_camera_name();
}

/*!
 * @brief Return the camera information pointer for this child window.
 */
const FI_COMM::FI_CAMERA_INFO *CameraMDI::get_camera_info() const
{
    return camera_ptr;
}

uint32_t CameraMDI::get_clip_end()
{
    return review_controls->get_clip_end();
}

uint32_t CameraMDI::get_clip_start()
{
    return review_controls->get_clip_start();
}

/*!
 * @brief Ask the user for a configuration file.
 * @return The fully qualified file name of the selected file, or an empty string.
 */
QString CameraMDI::get_config_file()
{
    int         user_resp;
    QString     download_path;
    QString     filename;
    QSettings   settings;
    QFileInfo   file_info;

    // Make sure the user understands that this is going to result in a camera reboot.
    user_resp = QMessageBox::question(MainWindow::GetInstance().get(), tr("Confirm Load Camera Configuration"),
                                      tr("Are you sure you want to load the camera configuration? This will require a restart of the camera which can take about 2 minutes."),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
    if (user_resp == QMessageBox::No)
        return QString();

    // Initialize the default path based on the saved settings.
    settings.beginGroup("PathInfo");
    download_path = settings.value("config_root").toString();
    settings.endGroup();

    // If we didn't get anything, just ask for a default path.
    if (download_path.isEmpty())
        download_path = SaveDialog::loadPath(0);

    // Ask the user for a configuration file.
    filename = QFileDialog::getOpenFileName(MainWindow::GetInstance().get(), tr("Select A Configuration File to Load to the Camera"),
                                            download_path, tr("Configuration Files (*.cfg)"));
    if (filename.isEmpty())
        return filename;

    // Get the path for the configuration file.
    file_info.setFile(filename);

    // Make sure the path ends with a '/' seperator.
    download_path = file_info.canonicalPath();
    if (!download_path.endsWith(QLatin1Char('/')))
        download_path.append(QLatin1Char('/'));

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "download_path: " << download_path;  // RECHECK_JLC

    // Save the path for the configuration files.
    settings.beginGroup("PathInfo");
    settings.setValue("config_root", download_path);
    settings.endGroup();

    // Return the path to the file.
    return file_info.canonicalFilePath();
}

/*!
 * @brief Return the current frame number (absolute frame number) being reviewed.
 */
uint32_t CameraMDI::get_current_frame_num()
{
    uint32_t abs_frame_num = 0;
    if (review_controls)
    {
        abs_frame_num = review_controls->get_current_frame_num();
    }
    return abs_frame_num;
}

/*!
 * @brief Return the camera's IP address, or -1 if the camera_ptr hasn't
 *        been assigned yet.
 */
uint32_t CameraMDI::get_ip_addr()
{
    uint32_t    rc = -1U;

    // Make sure we have a camera!
    if (camera_ptr != nullptr)
        rc = camera_ptr->ip_addr;

    return rc;
}

/*!
 * @brief Return the camera's IP address, nicely formatted.
 */
QString CameraMDI::get_ip_addr_pretty()
{
    uint32_t ip_addr = -1U;
    QHostAddress our_addr;

    // Make sure we have a camera!
    if (camera_ptr != nullptr)
        ip_addr = camera_ptr->ip_addr;

    // Make a pretty string out of the address.
    our_addr.setAddress(qFromBigEndian(ip_addr));
    return our_addr.toString();
}

/*!
 * @brief Return the camera's MAC address as a string.
 */
QString CameraMDI::get_mac()
{
    QString     mac_addr_str;
    quint8      mac_addr[6];
    QLatin1Char zeroFill('0');

    // Sanity check.
    if (camera_ptr == nullptr)
        return QString::fromUtf8("");

    // Get the MAC address in a user-friendly format.
    memcpy(&mac_addr, camera_ptr->mac_addr, 6);
    mac_addr_str = QString::fromUtf8("%1:%2:%3:%4:%5:%6").
                   arg(mac_addr[0], 2, 16, zeroFill).
                   arg(mac_addr[1], 2, 16, zeroFill).
                   arg(mac_addr[2], 2, 16, zeroFill).
                   arg(mac_addr[3], 2, 16, zeroFill).
                   arg(mac_addr[4], 2, 16, zeroFill).
                   arg(mac_addr[5], 2, 16, zeroFill);
    return mac_addr_str;
}

int CameraMDI::get_pfmd_len()
{
    int ret_len;

    if (is_series3())
        ret_len = GIGE_FRAME_METADATA_LEN;
    else
        ret_len = (GIGE_FRAME_METADATA_LEN * 2);

    return ret_len;
}

int CameraMDI::get_packet_len()
{
    // Sanity check.
    if (streamer == nullptr)
        return 0;

    // Returned the negotiated packet length.
    return streamer->packet_len();
}

double CameraMDI::get_trigger_event()
{
    uint32_t    sbs, sbs_us;

    // Make sure we've got something.
    if (partition_metadata.length() == 0)
        return 0.0;

    // Get an overlay onto the data.
    const pm_metadata *pm_ptr = (const pm_metadata *)partition_metadata.constData();

    // Get the values exactly as it is done on the camera.
    sbs = pm_ptr->trig_time_s;
    sbs_us = (pm_ptr->trig_time_ns + 500) / 1000;

    // Convert to a double.
    return ((double)sbs + (((double)sbs_us) / 1000000.0));
}

const QString CameraMDI::get_version()
{
    QString rc_str;

    if (camera_ptr != nullptr)
    {
        rc_str = QString::fromUtf8(camera_ptr->dev_version);
    }

    return rc_str;
}

/*!
 * @brief Remove the specified group read entry matching last address equal to reg_addr.
 * Note: This only seems to be called by the camera review object when it is set invisible.
 * @param reg_addr  The register address we are interested in.
 */
void CameraMDI::group_read_cancel(uint32_t reg_addr)
{
    // Loop through all the entries in the FIFO.
    for (int i = 0; i < group_reads.count(); ++i)
    {
        // Do we have a matching group read record in our queue?
        QSharedPointer<FiCamRegGroup> cam_group_read = group_reads.at(i);

        // Remove any entries in our queue that may have emptied out.
        if (cam_group_read->getLastRegAddr() == reg_addr)
        {
            // Theoretically there should not be duplicate last address entries but
            // I am only warning right now - TODO: Come back to this logic.
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[CANCELED:" << cam_group_read->getCamRegGroupSigStr()
                         << "][INDEX:" << i << "][COUNT:" << group_reads.count() - 1 << "]";
#endif
            // Got it, so remove it and be done.
            group_reads.removeAt(i);
        }
    }
}

/*!
 * @brief Returns true if we're in a calibration event, false otherwise.
 */
bool CameraMDI::in_calibration()
{
    return (calibration_event != CAL_EVENT_NONE);
}

/*!
 * @brief Returns true if autosave is configured, false otherwise.
 */
bool CameraMDI::is_autosave_configured()
{
    bool rc;

    // If we have a save video object, is it configured for autosave?
    if (save_video != nullptr)
        rc = save_video->autosave_is_configured();
    else
        rc = false;

    return rc;
}

/*!
 * @brief Returns true if autosave is enabled on the camera, false otherwise.
 * NOTE:  Important to note just because autosave is enabled on the camera it doesn't mean that
 * the camera object (this object) is configured properly for the autosave.  For that condition
 * the caller must also query CameraMDI::is_autosave_configured().
 */
bool CameraMDI::is_autosave_enabled()
{
    bool autosave_enabled = false;
    uint32_t reg_value;

    // Pull the cached value.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

    // Check to see if the camera is set for autosave.
    if ((reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
    {
        autosave_enabled = true;
    }

    // Parse out the fields we're interested in.
    return autosave_enabled;
}

/*!
 * @brief Returns true if autosave is enabled on the camera and is in overlapping mode.
 */
bool CameraMDI::is_autosave_overlapping()
{
    bool overlapping = false;
    uint32_t reg_value;

    // Pull the cached value.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

    // Check to see if the camera is set for autosave.
    if ((reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
    {
        if ((reg_value & (uint32_t)GIGE_AUTOSAVE_OVERLAP) == (uint32_t)GIGE_AUTOSAVE_OVERLAP)
            overlapping = true;
    }

    // Return the state of overlapping.
    return overlapping;
}

/*!
 * @brief A convenience function to determine if the camera is set for
 * FasFireLR capture mode.
 *
 * @param ltr_type: If it is enabled is it basic, roc or broc type?
 *
 * @returns True if the FasFireLR (fflr) capture mode is set and the value in the
 * reference paramater ltr_type will reflect the type of FasFireLR capture mode
 * or false if not.
 */
bool CameraMDI::is_fflr_enabled(uint32_t& ltr_type)
{
    uint32_t    ltr_mode;
    bool fflr_enabled = false;

    // Set the ltr_type to zero which will represent "basic" FFLR.
    ltr_type = 0;

    // Get our LTR mode.
    ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);

    // Is it FFLR?
    if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR)
    {
        fflr_enabled = true;

        // Set the BROC type if its bit is set.
        if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_BROC) == (uint32_t)GIGE_LTR_MODE_BROC)
        {
            ltr_type = GIGE_LTR_MODE_BROC;
        }
        // Set the BROC type if its bit is set.
        else if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_ROC) == (uint32_t)GIGE_LTR_MODE_ROC)
        {
            ltr_type = GIGE_LTR_MODE_ROC;
        }
    }
    return fflr_enabled;
}

/*!
 * @brief Are we in the process of recovering from a FasFireLR underflow error?
 */
bool CameraMDI::is_fflr_underflow_recovering()
{
    if (save_video && save_video->is_fflr_underflow_recovering())
    {
        return true;
    }
    return false;
}

/*!
 * @brief A convenience function to determine if the camera is too busy to talk to.
 */
bool CameraMDI::is_busy()
{
    bool    camera_is_busy = false;     // Assume we're not busy.

    // Sanity check.
    if (camera_ptr == nullptr)
        return true;

    // If any of these are true, then we shouldn't be talking to the camera.
    if (camera_ptr->rebooting || shutdown_in_progress())
    {
        camera_is_busy = true;
    }
    else if (camera_ptr->offline || in_calibration())
    {
        camera_is_busy = true;
    }
    else
    {
        bool    enabled;
        bool    save_in_progress;
        uint32_t    avail_count;
        uint32_t    total_count;
        uint32_t    reg_value;

        // Pull the cached value.
        reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

        // Parse out the fields we're interested in.
        enabled = (reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE;
        save_in_progress = (reg_value & (uint32_t)GIGE_AUTOSAVE_SAVING) == (uint32_t)GIGE_AUTOSAVE_SAVING;
        avail_count = GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);
        total_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value);

        // If the camera is in the middle of an autosave, then we probably shouldn't be talking to it.
        if ((!enabled && save_in_progress) || (enabled && (avail_count != total_count)))
            camera_is_busy = true;
    }

    // Check to see if we're in the middle of a copy.
    if (!camera_is_busy && streamer && streamer->is_copying())
        camera_is_busy = true;

    // Return our busy-ness.
    return camera_is_busy;
}

/*!
 * @brief Return true if we are a member of a multi-camera group.
 */
bool CameraMDI::is_grouped()
{
    // Sanity check.
    if (camera_info == nullptr)
        return false;

    // Return our "grouped" status.
    return camera_info->is_grouped();
}

/*!
 * @brief Return true if this camera is the local camera.
 */
bool CameraMDI::is_local()
{
    QString cam_name;

    // Are we running local?
    if (!MainWindow::local_control())
        return false;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // Put the camera name into an object we can use.
    cam_name = QString::fromUtf8(camera_ptr->user_name);

    // Get our name.
    if (QHostInfo::localHostName().compare(cam_name, Qt::CaseInsensitive) == 0)
        return true;

    // No match.
    return false;
}

/*!
 * @brief A convenience function to determine if the camera is set for
 * LTR (Long Time Record) capture mode.
 *
 * @param ltr_type: If it is enabled is it basic, roc or broc type?
 *
 * @returns True if the Long Record (LTR) capture mode is set and the value in the
 * reference paramater ltr_type will reflect the type of FasFireLR capture mode or
 * false if not.
 */
bool CameraMDI::is_ltr_enabled(uint32_t& ltr_type)
{
    uint32_t    ltr_mode;
    bool ltr_enabled = false;

    // Set the ltr_type to zero which will represent "basic" FFLR.
    ltr_type = 0;

    // Get our LTR mode.
    ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);

    // Is it FFLR?
    if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE)
    {
        ltr_enabled = true;

        // Set the BROC type if its bit is set.
        if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_BROC) == (uint32_t)GIGE_LTR_MODE_BROC)
        {
            ltr_type = GIGE_LTR_MODE_BROC;
        }
        // Set the BROC type if its bit is set.
        else if ((ltr_mode & (uint32_t)GIGE_LTR_MODE_ROC) == (uint32_t)GIGE_LTR_MODE_ROC)
        {
            ltr_type = GIGE_LTR_MODE_ROC;
        }
    }
    return ltr_enabled;
}

/*!
 * @brief A utility method to indiciate if the camera is off-line or not.
 */
bool CameraMDI::is_offline()
{
    // Sanity check.
    if (camera_ptr == nullptr)
        return true;

    // Return our offline status.
    return camera_ptr->offline;
}

bool CameraMDI::is_series3()
{
    // Sanity check. Probably the safe thing to do is return true.
    if (camera_ptr == nullptr)
        return true;

    // Return our findings.
    return CameraMDI::is_series3(camera_ptr);
}

/*!
 * @brief A utility method to indiciate if the camera has external I/O sync-in enabled.
 * @returns True if external I/O is sync-in enabled, false if no camera pointer or not enabled.
 */
bool CameraMDI::is_sync_in_enabled()
{
    bool syncInEnabled = false;

    // Sanity check. Probably the safe thing to do is return false.
    if (camera_ptr != nullptr)
    {
        uint32_t    reg_value;
        uint32_t    sync_in_type;

        // Query the camera cache to see if the sync-in external I/O is enabled.
        reg_value = read_camera_cache(GIGE_REG_SYNC_IN_ENABLE);

        // Parse out the settings.
        sync_in_type = reg_value & GIGE_SYNC_TYPE_MASK;

        // If we using an external sync signal, then we're enabled.
        syncInEnabled = (sync_in_type == GIGE_SYNC_TYPE_EXTERNAL);
    }

    // Return our findings.
    return syncInEnabled;
}

/*!
 * @brief A utility method to indiciate if the camera has external I/O sync-in enabled
 * receiving a non-zero pulse value in the register GIGE_REG_SYNC_IN_TIMING.
 *
 * @returns True if the register reports a non-zero value indicating sync-in pulse timing, false otherwise.
 */
bool CameraMDI::is_sync_in_receiving_pulse()
{
    bool syncPulseTimingReceived = false;

    // Sanity check. Probably the safe thing to do is return false.
    if (camera_ptr != nullptr)
    {
        uint32_t    reg_value;

        // Query the camera cache to see if the sync-in external I/O is receiving a timing pulse.
        reg_value = read_camera_cache(GIGE_REG_SYNC_IN_TIMING);

        // Just check for the value, if it is > 0 there is a pulse being received.
        if (reg_value > 0)
            syncPulseTimingReceived = true;
    }

    // Return our findings.
    return syncPulseTimingReceived;
}

/*!
 * @brief A utility method to determine if the last sync-in suspension warning was serviced
 * within a specified threshold.
 *
 * @param msThreshold   The number of milliseconds to allow for a threshold test.
 *
 * @return true if time within threshold, false if not.
 */
bool CameraMDI::is_sync_in_warning_in_threshold(int msThreshold)
{
    bool inside_threshold = false;

    // Only check if we put up a sync-in suspension warning.
    if (time_of_sync_in_suspension.isValid())
    {
        QDateTime time_right_now = QDateTime::currentDateTime();
        qint64 delta_time = time_of_sync_in_suspension.msecsTo(time_right_now);
        // Is the threshold within the delta_time?
        if ((qint64)msThreshold <= delta_time)
        {
            inside_threshold = true;
        }
    }
    return inside_threshold;
}

/*!
 * @brief A utility method to determine if the camera is a series3 camera type.
 *
 * @param info_ptr  The camera information to query.
 *
 * @return true if it is a series3 camera, false if not.
 */
bool CameraMDI::is_series3(const FI_COMM::FI_CAMERA_INFO *info_ptr)
{
    bool series3_found = false;
    QString camera_model;

    // Sanity check. Probably the safe thing to do is return true.
    if (info_ptr == nullptr)
        return true;

    // Get the serial number in a format easier for us to work with.
    camera_model = QString::fromUtf8(info_ptr->model_name);

    // Are we a "TSx" or "ILx" camera?
    // TODO: Replace with a RegEx.
    if (camera_model.contains(QStringLiteral("TS")))
        series3_found = true;
    else if (camera_model.contains(QStringLiteral("IL")))
        series3_found = true;

    // Return our result.
    return series3_found;
}

/*!
 * @brief Determines if the last frame expected is missing from a copy or save
 * stream that thinks it should have received all frames in the stream.
 *
 * @param start_retry_if_missing: True if caller wants the retry process to start,
 * false if not.
 *
 * @returns True if the last frame was missing.
 */
bool CameraMDI::last_frame_missing_check()
{
    bool last_frame_is_missing = false;
    uint32_t last_frame_0_based;

    // Did we receive the very last frame?
    last_frame_0_based = save_video->getClipStart() + (save_video->getFrameCount() - 1);

    // last_received_frame should equal full count of buffers expected but if it is still waiting
    // on the very last one it was missing from the stream.
    if (last_received_frame == last_frame_0_based)
    {
        bool addIt = true;
        last_frame_is_missing = true;

        addIt = add_to_retry_list(last_frame_0_based);

        // Add the entry if it's not there already.
        if (addIt)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "LAST FRAME WAS MISSING:  last_received_frame:" << last_received_frame
                         << "frame_count:" << save_video->getFrameCount()
                         << "last_frame_0_based:" << last_frame_0_based;
        }
    }
    return last_frame_is_missing;
}

void CameraMDI::load_complete()
{
//   QLOG_TRACE() << __PRETTY_FUNCTION__;

    if (loadCaptureDlg.get() != nullptr)
        loadCaptureDlg.reset();
}

void CameraMDI::merge_overlay(uint32_t seq_number, FI_MEMPOOL::FiImageBuffer *imageBufferPtr)
{
    uint32_t    frame_num;
    fiPixmap    my_pixmap;
    QGraphicsScene  *scene_ptr;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_number:" << seq_number;

    // Skip metadata sequence numbers.
    if (FiSaver::metadataSeqNum(seq_number))
        return;

    // Can only process this if there is an image buffer pointer.
    if (imageBufferPtr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The image buffer pointer is null!";
        return;
    }

    // Load the image into our pixmap.
    my_pixmap.loadFromData(imageBufferPtr->GetMemDataPtr(), imageBufferPtr->GetImageDataLen());

    // Calcualte the absolute frame number.
    frame_num = save_video->getClipStart() + (seq_number - 1);

    // Get the raw PFMD array.
    QByteArray          *pfmd_array = nullptr;
    if (!raw_metadata.isEmpty())
    {
        int raw_offset = ((int)seq_number - 1) * get_pfmd_len();

        if (is_series3())
        {
            uint32_t raw_buffer[4];

            memcpy(raw_buffer, raw_metadata.constData() + raw_offset, sizeof(raw_buffer));

            // Copy the raw data over.
            raw_buffer[0] = qToBigEndian(raw_buffer[0]);
            raw_buffer[1] = qToBigEndian(raw_buffer[1]);
            raw_buffer[2] = qToBigEndian(raw_buffer[2]);
            raw_buffer[3] = qToBigEndian(raw_buffer[3]);

            // Allocate our QByteArray.
            pfmd_array = new QByteArray((const char *)raw_buffer, get_pfmd_len());
        }
        else
        {
            // Allocate our QByteArray.
            pfmd_array = new QByteArray((const char *)raw_metadata.constData() + raw_offset, get_pfmd_len());
        }
    }
    else if (!raw_path.isEmpty())
    {
        PlaybackMetaInfo    metadata(this);
        metadata.setRawMetadataFile(raw_path, get_pfmd_len());
        pfmd_array = metadata.getRawPFMD(frame_num, is_series3());
    }

    if (pfmd_array != nullptr)
    {
        // Pass it on.
        camera_info->set_pfmd(pfmd_array);

        // Save it in case we need to use it later.
        pfmd_info->insert(frame_num, pfmd_array);
    }

    // Add the pixmap to the scene.
    video_player->render_add_image(my_pixmap);
    video_player->render_update(frame_num, camera_info, nullptr);

    // Get the pointer to the graphics "scene".
    scene_ptr = video_player->render_get_scene();
    if (scene_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No scene_ptr - did you render_start?";
        return;
    }

    // Grow our pixmap.
    fiPixmap    scene_pixmap((int)scene_ptr->sceneRect().width(), (int)scene_ptr->sceneRect().height());
    QPainter    scene_painter(&scene_pixmap);
    scene_painter.setRenderHint(QPainter::Antialiasing);
    scene_ptr->render(&scene_painter);

    // Get the pixmap as a qImage.
    QByteArray  my_buffer;
    scene_pixmap.set_jpeg_quality((int)read_camera_cache(GIGE_REG_JPEG_QUALITY));
    scene_pixmap.save_image(&my_buffer, save_video->getImageInputType());

    // Build up our image buffer.
    uint32_t bytesReplaced = imageBufferPtr->ReplaceImageBytes((uint32_t) my_buffer.length(), (const uint8_t *)my_buffer.constData());
    if (bytesReplaced != (uint32_t)my_buffer.length())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The merge-overlay image replacement operation failed!";
    }
    return;
}

/*!
 * @brief Parse a camera's configuration file, looking for a register value specified
 *        by the reg_addr parameter.
 *
 * @note:   This method only looks for/retrieves UINT32_T values. If the register type
 *          is a MEMORY type register (i.e. more than a 32-bit value) this routine will
 *          not work properly.
 *
 * @param file_name     The full path to the camera's configuration file.
 * @param reg_addr      The register address we're interested in retrieving.
 * @param reg_value     The value for reg_addr, if found.
 *
 * @return True, if found, or false on error.
 */
bool CameraMDI::parse_config_file(QString& file_name, uint32_t reg_addr, uint32_t& reg_value)
{
    bool        rc;
    uint32_t    reg_type;
    QFile       config_fd(file_name);

    // Open the file for reading.
    if (!config_fd.open(QIODevice::ReadOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open" << file_name << ". Err:" << config_fd.errorString();
        return false;
    }

    // Go until we find the register we're interested in, or we hit the end of the file.
    reg_value = 0;
    rc = false;
    while (!config_fd.atEnd())
    {
        int         read_len;
        uint32_t    parsed_addr;

        // Read a line from the file.
        QByteArray  file_line = config_fd.readLine();

        // Parse the information.
        read_len = sscanf(file_line.constData(), "%u,%u,%u", &parsed_addr, &reg_type, &reg_value);
        if (read_len <= 0)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Malformed config file!";
            break;
        }

        // Did we find the address we're interested in?
        if (parsed_addr == reg_addr)
        {
            rc = true;
            break;
        }
    }

    // Close our file.
    config_fd.close();

    // Return our result.
    return rc;
}

/*!
 * @brief Parse a camera's configuration file, looking for a register value specified
 *        by the reg_addr parameter.
 *
 * @note:   This method only looks for/retrieves UINT32_T values. If the register type
 *          is a MEMORY type register (i.e. more than a 32-bit value) this routine will
 *          not work properly.
 *
 * @param file_name     The full path to the camera's configuration file.
 * @param reg_addr      The register address we're interested in retrieving.
 * @param reg_value     The value for reg_addr, if found.
 *
 * @return True, if found, or false on error.
 */
bool CameraMDI::parse_config_file(QString& file_name, uint32_t reg_addr, QString& reg_value)
{
    bool        rc;
    QFile       config_fd(file_name);

    // Open the file for reading.
    if (!config_fd.open(QIODevice::ReadOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open" << file_name << ". Err:" << config_fd.errorString();
        return false;
    }

    // Go until we find the register we're interested in, or we hit the end of the file.
    rc = false;
    reg_value.clear();
    while (!config_fd.atEnd())
    {
        uint32_t    parsed_addr;
        QByteArray  raw_value(GIGE_READMEM_B_LEN, 0);

        // Read a line from the file.
        QByteArray  file_line = config_fd.readLine();

        // Parse the information.
        std::string token;
        std::stringstream file_stream(file_line.toStdString());

        std::getline(file_stream, token, ',');
        parsed_addr = std::stoi(token);
        std::getline(file_stream, token, ',');
        //auto reg_type = std::stoi(token);         // We don't do anything with the register type value.
        std::getline(file_stream, token, ',');
        raw_value = QByteArray(token.c_str());

        // Did we find the address we're interested in?
        if (parsed_addr == reg_addr)
        {
            rc = true;
            reg_value = QString(raw_value.constData());
            break;
        }
    }

    // Close our file.
    config_fd.close();

    // Return our result.
    return rc;
}

/*!
 * @brief Enable or disable the cache refresh logic.
 * @param do_pause  True to pause, false to resume.
 */
void CameraMDI::pause_cache_refresh(bool do_pause)
{
    cache_refresh_holdoff = do_pause;
}

/*!
 * @brief Handle the request to pause the video stream.
 */
void CameraMDI::pause_video_request(const QString whoCalledMe)
{
    QString myFunctionName      = __PRETTY_FUNCTION__;
    QString myMessageForEmit    = "Signaling video_is_paused().";

    myFunctionName = myFunctionName + "called by[" + whoCalledMe + "]";

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << myFunctionName << "The camera is offline." << myMessageForEmit;
        emit video_is_paused();
        return;
    }

    // If the camera should be off-line, but if we have a stream running, then
    // shutdown the stream.
    if (streamer == nullptr)
    {
        QLOG_DEBUG() << myFunctionName << "The streamer is offline." << myMessageForEmit;
        emit video_is_paused();
        return;
    }

    // Determine if streamer is already paused or not and act accordingly.
    if (streamer->is_paused())
    {
#if LOG_PAUSE_ALREADY_PAUSED
        QLOG_DEBUG() << myFunctionName << "The streamer is paused already." << myMessageForEmit;
#endif
        // The streamer is paused, let anyone listening know it.
        emit video_is_paused();
        return;
    }

    // The streamer is not paused, send the command to pause it.
#if LOG_PAUSE_NOT_PAUSED
    QLOG_DEBUG() << myFunctionName << "The streamer is running calling streamer->pause_streaming().";
#endif
    streamer->pause_streaming();
}

/*!
 * @brief Build up a parameter block for the stream state machine to play a
 *        captured video.
 *
 * @param play_forward  True to play "forward", false to play in "reverse".
 */
void CameraMDI::play_video(bool play_forward)
{
    CAMERA_STREAM::STREAM_PARAMS    stream_info;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Don't stream if we don't want streaming.
    if (!camera_ptr->stream)
        return;

    // Sanity check(s)
    if (camera_info == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera_info!";
        return;
    }

    // Make sure there are review controls in existence.
    if (!review_controls)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "review_controls is empty!";
        return;
    }

    // Ticket #328:
    // Make sure that we disable the review controls first thing to help mitigate
    // re-entry issues.
    review_controls->update_buttons(false);

    // Remember our direction.
    play_video_forward = play_forward;

    // Clear our parameters structure.
    memset(&stream_info, 0, sizeof(stream_info));

    // Get the "new" current frame.
    stream_info.cur_frame = review_controls->get_current_frame_num();
    stream_info.clip_start = review_controls->get_clip_start();
    stream_info.clip_end = review_controls->get_clip_end();

    // Make sure the current frame is within the clip.
    if (stream_info.cur_frame > stream_info.clip_end)
        stream_info.cur_frame = stream_info.clip_end;
    if (stream_info.cur_frame < stream_info.clip_start)
        stream_info.cur_frame = stream_info.clip_start;

    // Figure out our timing parameters. Get the rate the user would like to
    // playback the video at.
    stream_info.desired_fps = (uint32_t)review_controls->get_playback_rate();

    // Check to see if the "playback" rate (how fast we want the FPGA to playback the video)
    // is less than our "refresh" rate (how many frames we can transfer over the network).
    if (stream_info.desired_fps <= (uint32_t)camera_info->get_framerate())
    {
        // The network can support the playback rate, so no need to skip.
        stream_info.skip_interval = 1;
    }
    else
    {
        // The playback rate would generate too many frames/second for the network to support.
        // We need to approximate the rate the user wants by having the FPGA skip frames.
        stream_info.skip_interval = stream_info.desired_fps / (uint32_t)camera_info->get_framerate();
        stream_info.desired_fps = (uint32_t)camera_info->get_framerate();
    }

    // Figure out the number of milliseconds per frame.
    stream_info.frame_period = 1000 / stream_info.desired_fps;

    // When figuring out the frame delay, make sure we oversample.
    stream_info.frame_delay = ((1000 / stream_info.desired_fps) + 1) / 2;

    // Are we playing in reverse?
    if (!play_forward)
        stream_info.skip_interval |= GIGE_STREAM_REVERSE;

    // We always loop when playing.
    stream_info.skip_interval |= GIGE_STREAM_LOOP;

    // Tell the stream state machine to start the playback.
    if (streamer)
        streamer->playback(stream_info);
}

/*!
 * @brief Perform the actual "arm"/"de-arm" operation as required by whatever state the camera
 *        is currently in.
 *
 * @param allow_discard True if the paritition is allowed to be discard, false otherwise.
 */
void CameraMDI::process_arm(bool user_queried, bool allow_discard)
{
    uint32_t    camera_state;

    // Disable all controls while we wait for things to happen. They will be
    // re-enabled when the process_state_change() method is called.
    camera_info->disable_actions();

    // What state is the camera in?
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    switch (camera_state)
    {
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
    case GIGE_CAMERA_STATE_REVIEW:
        // Ask the video control panel to reset it's clip range.
        review_controls->reset_clip();
        reset_event_info();

        // Are we allowed to discard the partition?
        if (allow_discard)
            free_partition(true);

        // Make sure camera is prepared for arming if FasFireLR is enabled.
        fflr_check_before_arming();

        // Delay for a bit, and then arm the camera.
        QTimer::singleShot(RELEASE_MEMORY_DELAY, this, &CameraMDI::finish_arm);
        break;

    case GIGE_CAMERA_STATE_ARMED:
    case GIGE_CAMERA_STATE_TRIGGERED:
        // De-arm the camera.
        if (!dearm_camera(user_queried, allow_discard))
        {
            // We didn't proceed with the dearm, so make sure we re-enable the camera actions.
            camera_info->enable_actions();
        }
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
        // Are we in the process of saving, either locally or remotely?
        if (save_video && (save_video->save_in_progress() || ((save_video->get_last_save_state() & GIGE_SAVE_STATE_FLAGS) == GIGE_SAVE_STATE_BUSY)))
        {
            // User is trying to cancel the autosave.
            QString whoCalled = __PRETTY_FUNCTION__;
            whoCalled += "GIGE_CAMERA_STATE_AUTOSAVE";
            save_video->cancel_save(whoCalled);
        }
        else
        {
            // Go ahead and switch to the armed state.
            write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_ARMED);
        }
        break;
    case GIGE_CAMERA_STATE_SNAPSHOT:
    default:
        // Go ahead and switch to the armed state.
        write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_ARMED);
        break;
    }

    // If we're in a group, we MAY have queued up for a save. Let's
    // make certain that we are removed from the queue.
    if (camera_info->is_grouped())
    {
        MultiCamDock    *dock_ptr;

        QLOG_TRACE() << __PRETTY_FUNCTION__ << "ARMing, so make sure we're not in the autosave queue:" << Qt::showbase << Qt::hex << get_ip_addr_pretty();

        // Get a pointer to the multicam dock.
        dock_ptr = MainWindow::GetInstance()->get_multicam_dock();
        if (dock_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No multicam dock!?";
            return;
        }

        // Reset our flag.
        multicam_save_allowed = false;

        // Let the multicamera dock know that we're done.
        dock_ptr->autosave_dequeue(this);
    }

    // Reset information about the current capture.
    trigger_frame = -1U;
    trigger_frame_start = 0.0;

    // Reset our cache of PFMD.
    pfmd_info->clear();
}

/*!
 * @brief Common initialization of our data that needs to be set prior to a download operation.
 */
void CameraMDI::prepare_to_download()
{
    // Initialize some variables we'll need to manage the download.
    retry_in_progress = false;
    last_received_frame = 0;    // Note: download sequence numbers are 1-based live images are 0-based.
    missed_frame_count = 0;
    retry_frame_list.clear();
    retry_frames_reqd.clear();
    count_metafiles_copied = 0;
}

/*!
 * @brief Process an autosave "state" change.
 * @param reg_value     The new value for the GIGE_REG_AUTOSAVE register.
 */
void CameraMDI::process_autosave(uint32_t reg_value)
{
    bool enabled;
    bool save_in_progress;
    uint32_t avail_count;
    uint32_t total_count;
    MultiCamDock *dock_ptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Update our autosave display.
    camera_info->display_autosave_settings(reg_value);

    // Parse out the fields we're interested in.
    enabled = (reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE;
    save_in_progress = (reg_value & (uint32_t)GIGE_AUTOSAVE_SAVING) == (uint32_t)GIGE_AUTOSAVE_SAVING;
    avail_count = GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);
    total_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value);

    //QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "enabled:" << enabled << "save_in_progress:" << save_in_progress << "avail_count:" << avail_count << "total_count:" << total_count;

    // If we think autosave is enabled, but the register says no, then disable autosave.
    if (save_video->local_autosave() && !enabled)
    {
        // First check to see if we're trying to start autosave.
        if (!autosave_lockout)
        {
            // Disable autosave to the local workstation.
            save_video->autosave_stop();
        }
    }

    // If autosave is not enabled, we're done.
    if (!enabled)
    {
        // Even though we're not enabled, we may still be in the process
        // of saving captured buffers.
        autosave_pending = save_in_progress;

        // Update our display.
        camera_info->display_save_progress(0);

        // All done for this state.
        return;
    }

    // Is the camera busy saving?
    if (!save_video->local_autosave())
    {
        // If we're here, then autosave is enabled on the camera.
        // Check to see if a save is in progress.
        if (avail_count != total_count)
        {
            // If the camera should be off-line, but we have a stream running, then
            // shutdown the stream.
            if (camera_ptr->stream_running)
            {
                // Ticket #342 Test - only pause when saving locally.
                // Pause the stream while we're saving.
                //pause_video_request(__PRETTY_FUNCTION__);
            }
        }

        // See if we should turn the stream back on.
        else if (camera_ptr->stream && !camera_ptr->stream_running)
        {
            // Try to get the video streaming againg.
            resume_video_request();
        }

        // All done for this case.
        return;
    }

    // Are we locked out?
    if (autosave_lockout)
    {
        // Are we in the ARM'ed state?
        if (old_camera_state == GIGE_CAMERA_STATE_ARMED)
        {
            // Reset the lockout.
            QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "Lock out enabled, but OK to reset.";
            autosave_lockout = false;
        }
        else
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "Locked out. old_camera_state:" << old_camera_state;
            return;
        }
    }

    // Are we supposed to be saving something?
    if (avail_count == total_count)
    {
        // Are we in the autosave state?
        if (old_camera_state != GIGE_CAMERA_STATE_AUTOSAVE)
        {
            // All done for this case.
//            QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "state != _AUTOSAVE:" << Qt::showbase << Qt::hex << old_camera_state;
            return;
        }

        // Are we already saving or is it pending?
        if (save_video->save_in_progress() || autosave_pending)
        {
            //QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "Either we're saving or we're going to be saving";
            return;
        }
    }

    // Are we part of a multicamera group?
    if (camera_info->is_grouped())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "We are grouped! multicam_save_allowed:" << multicam_save_allowed;

        // As part of a group, we need to wait our turn when saving to the local PC.
        if (!multicam_save_allowed)
        {
            // Get a pointer to the multicam dock.
            dock_ptr = MainWindow::GetInstance()->get_multicam_dock();
            if (dock_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "No multicam dock!?";
                return;
            }

            // Let the multicam dock know that we're ready to autosave.
            multicam_save_allowed = dock_ptr->autosave_ready(this);

            // Do we need to wait?
            if (!multicam_save_allowed)
            {
                // Go away until the next polling period.
                return;
            }
        }
    }

    // Ready to start our autosave process.
    start_autosave();
}

/*!
 * @brief Determine what action to take based on the status from the
 *        attempt to write the GIGE_REG_CCP register.
 * @param write_status  The status of the write attempt.
 */
void CameraMDI::process_ccp_write(uint32_t write_status)
{
    // Are we shutting down?
    if (in_shutdown)
    {
        // Give things a change to quiet down, then finish the shutdown process.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Lock released";
        QTimer::singleShot(SHUTDOWN_DELAY, this, &CameraMDI::finish_shutdown);
    }
    else if (write_status == GEV_STATUS_SUCCESS)
    {
        FI_comm     *comm_ptr;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
            return;
        }

        // We're connected, so let's show ourselves.
        MainWindow::GetInstance()->change_mdi_child(this);

        // Make sure our heartbeat rate is what we expect.
        comm_ptr->start_write_reg(get_ip_addr(), GIGE_REG_HEARTBEAT, (uint32_t)heartbeat_rate + HEARTBEAT_OFFSET);

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "in_startup:" << in_startup;

        // Is this the first lock on startup?
        if (in_startup)
        {
            FI_COMM::REG_ADDR_LIST  reg_list;

            // No longer in startup.
            in_startup = false;

            // Connect the communications interfaces that we're interested in. Because a signal/slot invocation is effectively a function
            // call (according to the Qt documentation), to prevent recursive operation we want the read/write completion slots to be
            // invoked later (a "Queued" connection).
            connect(comm_ptr, &FI_comm::read_reg_complete, this, &CameraMDI::read_complete, Qt::QueuedConnection);
            connect(comm_ptr, &FI_comm::unsol_socket_ready, this, &CameraMDI::unsol_socket_is_ready, Qt::QueuedConnection);

            // Connect group read completions managed in this module to their handling slots.
            connect(this, &CameraMDI::group_read_complete, this, &CameraMDI::chk_refresh_cache_read, Qt::QueuedConnection);
            connect(this, &CameraMDI::group_read_complete, this, &CameraMDI::chk_camera_preference_read, Qt::QueuedConnection);
            connect(this, &CameraMDI::group_read_complete, this, &CameraMDI::chk_rec_progress_read, Qt::QueuedConnection);

            // We also want to know the camera state for the camera information panel.
            reg_list.append(GIGE_REG_CAMERA_STATE);

            // We would also like to know the camera's firmware version.
            reg_list.append(GIGE_REG_FIRMWARE_VER);
            reg_list.append(GIGE_REG_FPGA_VER);

            // Grab the boot/update status on startup.
            reg_list.append(GIGE_REG_BOOT_STATUS);

            // Let's make sure that the autosave and other settings are loaded too.
            reg_list.append(GIGE_REG_USER_SAVE_START_NUM);
            reg_list.append(GIGE_REG_USER_SAVE_END_NUM);
            reg_list.append(GIGE_REG_SAVE_START_NUM);
            reg_list.append(GIGE_REG_SAVE_END_NUM);
            reg_list.append(GIGE_REG_IMAGE_DESTINATION);
            reg_list.append(GIGE_REG_STREAM_FORMAT);
            reg_list.append(GIGE_REG_WIDTH);
            reg_list.append(GIGE_REG_HEIGHT);
            reg_list.append(GIGE_REG_OFFSET_X);
            reg_list.append(GIGE_REG_OFFSET_Y);
            reg_list.append(GIGE_REG_SENSOR_WIDTH);
            reg_list.append(GIGE_REG_SENSOR_HEIGHT);
            reg_list.append(GIGE_REG_AUTOSAVE);
            reg_list.append(GIGE_REG_LTR_MODE);
            reg_list.append(GIGE_REG_MEM_SIZE);
            reg_list.append(GIGE_REG_MEM_SIZE_RESERVED);
            reg_list.append(GIGE_REG_PART_STATE);
            reg_list.append(GIGE_REG_PART_SIZE);
            reg_list.append(GIGE_REG_PART_SIZE_RESERVED);
            reg_list.append(GIGE_REG_SESSION_LENGTH);

            // Start reading.
            if (camera_ptr)
                comm_ptr->start_read_multireg(camera_ptr->ip_addr, reg_list);

            // Allocate our sensor object.
            sensor_ptr = new fiSensor(this);
        }

        // Shutdown our old timer.
        stop_lock_timer();

        // Restart the timer with the new rate.
        lock_timer_id = startTimer(heartbeat_rate);
    }
    else
    {
        QString         err_msg;

        // Build up our error message.
        err_msg = QString::fromUtf8("Unable to write GIGE_REG_CCP to IP adddress %1. Status=%2").arg(get_ip_addr_pretty()).arg(write_status, 0, 16);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;

        // Put up an error message for the user.
        QMessageBox::critical(MainWindow::GetInstance().get(), get_camera_name(), tr("Unable to lock the camera. Make sure someone else is not already connected."));

        // Tell the main window we want to close.
        QTimer::singleShot(0, MainWindow::GetInstance().get(), [ = ]() { MainWindow::GetInstance()->close_mdi_child(this); });
    }
}

/*!
 * @brief We received an unsolicieted FPN register change message from the camera.
 * @param reg_value The new GIGE_REG_FPN value.
 */
void CameraMDI::process_fpn_change(uint32_t reg_value)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "calibration_event:" << calibration_event << "reg_value:" << Qt::showbase << Qt::hex << reg_value;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // See if the camera is calibrating. If it is, we need to adjust certain operations.
    camera_ptr->offline = ((reg_value & (GIGE_FPN_FPGA_CALIBRATE_MASK | GIGE_FPN_FPGA_FAC_CALIBRATE_MASK | GIGE_FPN_FPGA_LOAD_MASK)) != 0);
    if (camera_ptr->offline)
    {
        // The camera initiated a calibration event.
        calibration_event = CAL_EVENT_CAMERA;

        // If the camera should be off-line, but we have a stream running, then
        // shutdown the stream.
        if (camera_ptr->stream_running)
        {
            // Pause the stream while we're calibrating.
            pause_video_request(__PRETTY_FUNCTION__);

            // Check to see if we're the "active" camera (i.e. not in the background).
            if (active_window())
            {
                // Since we're the active camera, let the user know what's going on.
                MainWindow::GetInstance()->set_statusbar_message(tr("Camera is calibrating...please wait."), 3000);
            }
        }
    }
    else if (calibration_event != CAL_EVENT_NONE)
    {
        // Make sure the camera is in LIVE mode.
        if (read_camera_cache(GIGE_REG_CAMERA_STATE) != GIGE_CAMERA_STATE_LIVE_VIDEO)
            return;

        // We're no longer in an "off-line" state, so bring back the stream if it's paused.
        if (camera_ptr->stream && (streamer && streamer->is_paused()))
        {
            // Indicate that the camera is back on-line.
            camera_ptr->offline = false;

            // Resume the stream as appropriate.
            streamer->go_live();
        }

        // Do we have a progress dialog up?  If so reset it.
        if (progressDialog)
        {
            // Recycle any resources and release the progress dialog.
            fiProgressDialog::RecycleProgressDialog(progressDialog);
        }

        // No longer calibrating.
        calibration_event = CAL_EVENT_NONE;
    }
}

void CameraMDI::process_json_fflr()
{
    QByteArray  *json_message;

    // Read what was returned.
    json_message = read_camera_buffer(GIGE_REG_FFLR_JSON);
    if (json_message != nullptr)
    {
        // Append to our media buffer.
        json_fflr_str.append(QString::fromLocal8Bit(*json_message));

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "json_message->length():" << json_message->length() << "json_fflr_str.length():" << json_fflr_str.length() << "json_fflr_str:" << json_fflr_str << "json_fflr_str.length():" << json_fflr_str.length();

        // Make sure we got it all.
        if (json_message->length() >= (GIGE_FFLR_JSON_LEN - 1))
        {
            // Keep reading.
            start_camera_read_buf(GIGE_REG_FFLR_JSON, GIGE_FFLR_JSON_LEN);
        }

        // A message of a single byte of 0 means the camera server isn't ready yet.
        else if ((json_message->length() == 1) && (json_message->at(0) == 0))
        {
            // Get rid of the nasty null character.
            json_fflr_str.clear();

            // Try again in a little bit.
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Delay a bit and try again.";
            QTimer::singleShot(100, this, [ = ]() { start_camera_read_buf(GIGE_REG_FFLR_JSON, GIGE_FFLR_JSON_LEN); });
        }

        // Looks like we've got the message.
        else
        {
            QJsonParseError jp_err;

            // Strip any trailing NULL characters.
            if (json_fflr_str.at(json_fflr_str.size() - 1) == '\0')
                json_fflr_str.chop(1);

            // Make sure we have a well formed JSON string. Converting to a JSON document.
            jsonFFLRDoc = QJsonDocument::fromJson(json_fflr_str.toUtf8(), &jp_err);
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "jsonFFLRDoc:" << jsonFFLRDoc;

            if (jsonFFLRDoc.isNull())
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "JSON parse err:" << jp_err.errorString() << "json_fflr_str:" << json_fflr_str << "json_fflr_str.length():" << json_fflr_str.length() << "json_message len:" << json_message->length();
            }
            else
            {
                // Let folks know.
                emit fflr_status_ready(jsonFFLRDoc);

                // Make sure we disable autosave.
                write_camera_register(GIGE_REG_AUTOSAVE, GIGE_AUTOSAVE_DISABLE);

                // Clean up the save process.
                save_video->cleanup_save();
            }

            // It all looks good.
            json_fflr_str.clear();
        }

        // Don't need this now.
        delete json_message;
    }
    else
    {
        // Unexpected.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "JSON info from camera was null?";
    }
}

void CameraMDI::process_json_media()
{
    QByteArray  *json_message;

    // Read what was returned.
    json_message = read_camera_buffer(GIGE_REG_MEDIA_JSON);
    if (json_message != nullptr)
    {
        // Append to our media buffer.
        json_build_str.append(QString::fromLocal8Bit(*json_message));
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "json_message len:" << json_message->length() << "json_build_str len:" << json_build_str.length() << "json_build_str:" << json_build_str << "json_build_str_len:" << json_build_str.length();

        // Make sure we got it all.
        if (json_message->length() >= (GIGE_MEDIA_JSON_MAX_LEN - 1))
        {
            // Keep reading.
            start_camera_read_buf(GIGE_REG_MEDIA_JSON, GIGE_MEDIA_JSON_MAX_LEN);
        }
        else
        {
            QJsonParseError jp_err;

            // Strip any trailing NULL characters.
            if (json_build_str.at(json_build_str.size() - 1) == '\0')
                json_build_str.chop(1);

            // Make sure we have a well formed JSON string. Converting to a JSON document.
            jsonMedia = QJsonDocument::fromJson(json_build_str.toUtf8(), &jp_err);

            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "jsonMedia:" << jsonMedia;

            if (jsonMedia.isNull())
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "JSON parse err:" << jp_err.errorString() << "json_build_str:" << json_build_str << "json_build_str len:" << json_build_str.length() << "json_message len:" << json_message->length();
            }
            else
            {
                // Let folks know.
                emit media_info_ready(jsonMedia);
            }

            // It all looks good.
            json_build_str.clear();
        }

        // Don't need this now.
        delete json_message;
    }
    else
    {
        // Unexpected.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "JSON info from camera was null?";
    }
}

/*!
 * @brief Method to handle the processing required when a partition is captured.
 *
 * This method determines if we need to start an auto-save process, update event
 * markers, or anything else which might need to be done.
 */
void CameraMDI::process_partition_capture()
{
    uint32_t    camera_state;

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_video object not allocated?!";
        return;
    }

    // If we're not doing autosave, then bail.
    if (!save_video->local_autosave())
        return;

    // Is the autosave already in progress?
    if (save_video->save_in_progress())
        return;

    // Is the camera waiting for us?
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    if (camera_state == GIGE_CAMERA_STATE_AUTOSAVE)
    {
        int rc;

        // Initialize the saving object.
        if ((rc = save_video->setup()) != 0)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_video::setup() failed! error:" << rc;
            return;
        }

        // If we are autosaving to a CAP file we don't want to create the metadata.txt
        // file or the .pm file because the pm metadata is embedded inside of the CAP file.
        // Skip the download in this case.
        if (save_video->autosave_wants_pm_metadata())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Autosave started and pm_metadata is being fetched.";

            // We're now configured and ready to save. Initialize our buffer.
            partition_metadata.clear();

            // Ask the camera to re-aquire the metadata for downloading.
            request_active_metadata(__PRETTY_FUNCTION__);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Autosave started for a CAP file and pm_metadata is not needed.";

            // Pretend that we downloaded the raw metadata, and move on.
            save_video->start(__PRETTY_FUNCTION__);
        }
    }

    // Clear our flag. The save isn't pending anymore, it's in full swing.
    autosave_pending = false;
}

void CameraMDI::delayed_state_check()
{
    uint32_t    part_state;
    uint32_t    camera_state;

    // Check to see if the last partition is dirty.
    part_state = read_camera_cache(GIGE_REG_PART_STATE);

    // Get the current camera state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "part_state:" << part_state << "camera_state:" << camera_state;

    // If we're in REVIEW state and there's nothing left to review, then bump to LIVE.
    if ((camera_state == GIGE_CAMERA_STATE_REVIEW) && (part_state == GIGE_PART_STATE_EMPTY))
    {
        // Switch us to live.
        write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_LIVE_VIDEO);
    }
}

void CameraMDI::disconnect_video_player()
{
    // Make sure we have a video player.
    if (video_player != nullptr)
    {
        disconnect(video_player->get_action_overlay(), &QAction::triggered, video_player, &CameraVideo::do_add_overlay);
        disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_1_to_1, video_player, &CameraVideo::do_zoom_1_to_1);
        disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_in, video_player, &CameraVideo::do_zoom_in);
        disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_out, video_player, &CameraVideo::do_zoom_out);
        disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_fit, video_player, &CameraVideo::do_zoom_to_fit);
        disconnect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_window, video_player, &CameraVideo::do_zoom_to_window);
    }
}

/*!
 * @brief Read the partition and camera state and update the UI accordingly.
 */
void CameraMDI::process_partition_state()
{
    uint32_t    ltr_mode;
    uint32_t    part_state;
    uint32_t    camera_state;

    // Check to see if the last partition is dirty.
    part_state = read_camera_cache(GIGE_REG_PART_STATE);
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "part_state:" << part_state;

    // Get the current camera state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Get the LTR mode.
    ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);

    // If we're in REVIEW state and there's nothing left to review, then wait a little bit.
    // We'll check again later to see if one of the other UIs switched us or not.
    if ((camera_state == GIGE_CAMERA_STATE_REVIEW) && (part_state == GIGE_PART_STATE_EMPTY))
    {
        QTimer::singleShot(STATE_CHECK_DELAY, this, &CameraMDI::delayed_state_check);
    }

    // If the partition state is showing empty, make sure we reset our event information.
    if (part_state == GIGE_PART_STATE_EMPTY)
    {
        reset_event_info();
    }

    // Update the review control(s).
    camera_info->reflect_camera_state(camera_state, part_state, ltr_mode);
}

/*!
 * @brief We've received a chunk of a file to be saved to disk.
 *
 *  The actual process of saving the image is being handled by the save_video object.
 *  We need to check for dropped "chunks"
 *
 * @param seq_number    The sequence number of the received image. Note that
 *                      sequence numbers from the camera range from 1..<num_frames>.
 */
void CameraMDI::process_received_chunk(uint32_t seq_number)
{
#if FORCE_MISSING_FRAMES
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_number:" << seq_number << "last_received_frame:" << last_received_frame << "retry_in_progress:" << retry_in_progress;
#endif

    // Skip metadata sequence numbers.
    if (FiSaver::metadataSeqNum(seq_number))
        return;

    // We can theoretically drop frames from the new seriesNC cameras also even though
    // they utilize TCP/IP.  For series3 we definitely can drop frames because it is using
    // UDP.  I have removed any condition to only do this if series3 and made it universal.
    if (!retry_in_progress)
    {
        // Our expected sequence number is the last one received plus 1
        // in the 1-based stream of frames Remember seq_number 0 in stream
        // buffer means that frame is a live-view frame.
        uint32_t expected_seq_number = last_received_frame + 1;

        // If the sequence number we received is greater than the expected one we have
        // a gap in our frame stream and need to archive missed frames in our retry list.
        if (seq_number > expected_seq_number)
        {
            uint32_t seq_number_0_based = seq_number - 1;

            // Add missing frame(s) to retry list.
            for (uint32_t frame_0_based = last_received_frame; frame_0_based < seq_number_0_based; ++frame_0_based)
            {
                // Our retry list contains the absolute (0-based) sequence number of the "chunk"/file we want.
                add_to_retry_list(frame_0_based);
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Added Retry Missing Chunk(0-based):" << frame_0_based;
            }
        }
    }

    // Remember this sequence as our last "good" frame.
    last_received_frame = seq_number;
}

/*!
 * @brief We've received an image to be saved to disk.
 *
 *  The actual process of saving the image is being handled by the save_video object.
 *  We need to do some logical processing here, such as:
 *  - checking for dropped frames
 *  - determining if the last frame of the save has been received.
 *  - perform retry operations if required.
 *  - indicating that the save has completed.
 *
 * @param seq_number    The sequence number of the received image. Note that
 *                      sequence numbers from the camera range from 1..<num_frames>.
 */
void CameraMDI::process_received_image(uint32_t seq_number)
{
    int abs_frame;
    int rel_frame;

#if FORCE_MISSING_FRAMES
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_number:" << seq_number << "last_received_frame:" << last_received_frame;
#endif

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_video is NULL?!";
        return;
    }

    // Check for any dropped frames.
    if (seq_number != (last_received_frame + 1))
    {
        int     index;

        // Add the missing sequence numbers to our list.
        for (index = (int)(last_received_frame + 1); index < (int)seq_number; ++index)
        {
            uint32_t missingFrameId = save_video->getClipStart() + ((uint32_t)index - 1);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Missing frame id(0-based):" << missingFrameId
                         << "clip_start:" << save_video->getClipStart() << "index:" << index
                         << "last_received_frame:" << last_received_frame
                         << "seq_number:" << seq_number;

            // Our retry list contains the absolute (0-based) frame number of the frame we want.
            add_to_retry_list(missingFrameId);
        }
    }

    // Remember this sequence as our last "good" frame.
    last_received_frame = seq_number;

    // Calcualte the absolute frame number and then convert it to a relative frame number.
    abs_frame = (int)save_video->getClipStart() + (int)(seq_number - 1);
    rel_frame = abs_frame - (int)trigger_frame;

    // Update our transfer dialog.
    if (xferDialog)
    {
        xferDialog->setXferSrcValue(QString::number(rel_frame));
        // Special logic here.  If the stream type is not an FFmpeg video type
        // and currently that is only MP4 video - then use the frames read to
        // indicate progress.  With ffmpeg video formats we have an encoder that
        // has to be accomodated in terms of the save progress.
        if (!FiSaver::isStreamFFMpegVideo(save_video->getStreamType()))
        {
            xferDialog->incrementBar();
        }
    }
}

/*!
 * @brief We've received a "frame" of raw per-frame metadata from the camera.
 * @param meta_ptr  Our "frame" of raw metadata.
 */
void CameraMDI::process_received_metadata(std::unique_ptr<FI_MEMPOOL::FiImageBuffer>meta_ptr)
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (meta_ptr->GetImageDataLen() > (uint32_t)raw_metadata_len)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "incoming len:" << meta_ptr->GetImageDataLen() << "is too big! Expecting:" << raw_metadata_len;

        // Release the buffer.
        release_buffer(std::move(meta_ptr));
        return;
    }

    // Append the data to our buffer.
    raw_metadata.append((const char *)meta_ptr->GetMemDataPtr(), (int)meta_ptr->GetImageDataLen());

    // Release the buffer.
    release_buffer(std::move(meta_ptr));
    //QLOG_TRACE() << __PRETTY_FUNCTION__ << "raw_metadata.length()" << raw_metadata.length();

    // Update the progress bar, if we're not a series3 camera.
    if (!is_series3())
    {
        // The number of frames of metadata we've received.
        int num_frames = raw_metadata.length() / get_pfmd_len();
        if (xferDialog)
            xferDialog->setBarValue(num_frames);
        else if (progressDialog)
            progressDialog->setBarValue(num_frames);
    }

    // Have we downloaded all the metadata?
    if (raw_metadata.length() == raw_metadata_len)
    {
        // Wait for the stream to pause.
        connect(this, &CameraMDI::video_is_paused, this, &CameraMDI::raw_metadata_wrapup, Qt::UniqueConnection);
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Metadata download complete";

        // Pause the streamer.
        pause_video_request(__PRETTY_FUNCTION__);
    }
}

/*!
 * @brief We've received notification from the camera that it has changed it's state.
 *        If the new state is different from the old state (sometimes we get multiple
 *        state change messages with the same state), we need to react.
 *
 * @param new_state     The new camera state.
 */
void CameraMDI::process_state_change(uint32_t new_state, bool state_change_error)
{
    bool        fflr_status_needed = false; //!< True if we need to retrieve GIGE_REG_FFLR_JSON update.
    uint32_t    part_state;
    uint32_t    ltr_mode;

#if LOG_PROCESS_STATE_CHANGES
    // Is there a state change?
    if (old_camera_state != new_state)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "!!<<< STATE CHANGE >>>!! old_state:" << old_camera_state << "new_state:" << new_state;
    }
#endif

    uint32_t fflr_type;
    bool fflr_is_enabled = is_fflr_enabled(fflr_type);

    // Sanity check.
    if ((MainWindow::GetInstance().isNull()) || (camera_info == nullptr))
        return;

    // Need a little more info for the UI.
    part_state = read_camera_cache(GIGE_REG_PART_STATE);
    ltr_mode = read_camera_cache(GIGE_REG_LTR_MODE);

    // Update the window to reflect the new state.
    camera_info->reflect_camera_state(new_state, part_state, ltr_mode);
    if (active_window())
        MainWindow::GetInstance()->reflect_camera_state();

    // Show the new state.
    camera_info->display_camera_state(new_state);

    // Did the state actually change?
    if (!state_change_error)
    {
        if (old_camera_state == new_state)
        {
            // Ticket #492: Call the routine that manages the polling and visibility of the capture progress bar.
            // Parameter is "capturing" boolean and this will be set true if ARMED or TRIGGERED.
            // Note: The call is just a housekeeping call to make sure state matches capture progress bar.
            if ((new_state == GIGE_CAMERA_STATE_ARMED) || (new_state == GIGE_CAMERA_STATE_TRIGGERED))
            {
                rec_progress_state_check(true);     // It should be polling and visible.
            }
            else if ((new_state == GIGE_CAMERA_STATE_LIVE_VIDEO) || (new_state == GIGE_CAMERA_STATE_REVIEW))
            {
                rec_progress_state_check(false);    // It should not be polling and invisible.
            }
            // The else could be an ancillary transient state - ignore those states.
            return;
        }
    }

    // Reset our flags and variables.
    last_requested_frame = (uint32_t) -1;

    // Is the camera shutting down?
    if (new_state == GIGE_CAMERA_STATE_SHUTDOWN)
    {
        // Ignore it if we're already in shutdwon or rebooting.
        if (!in_shutdown && (camera_ptr && !camera_ptr->rebooting))
        {
            // We're going offline.
            set_offline(true);
            return;
        }
    }

    // Only display the per-frame metadata in REVIEW.
    else if (new_state == GIGE_CAMERA_STATE_REVIEW)
    {
        // Make sure we disable the state actions until we are stable.
        camera_info->disable_actions();

        // Make sure we get that first frame through.
        force_frame_update = true;

        // Allow the review controls to be hidden/shown.
        MainWindow::GetInstance()->get_view_review_controls()->setEnabled(true);

        // Allow for overlays
        video_player->enable_overlay(true);

        // Restore our review controls.
        do_view_review();

        // Have the review object update it's registers.
        QTimer::singleShot(0, review_controls, &CameraReview::read_panel_regs);
    }
    else
    {
        // Allow the review controls to be hidden/shown.
        MainWindow::GetInstance()->get_view_review_controls()->setEnabled(false);

        // Hide the controls since they aren't of any use now.
        review_controls->setVisible(false);

        // Disallow overlays.
        video_player->enable_overlay(false);

        // Update the camera information pane.
        camera_info->show_info();

        // Make sure we enable our actions.
        camera_info->enable_actions();

        // Special handling if FasFireLR and transitioning to Live Video.
        if (new_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
        {
            // Sync ourselves to the Live View state.
            if (rec_progress)
            {
                // Make sure our recording progress control is sync'd to a live view non-capturing state.
                rec_progress_state_check(false);
            }

            // Are we in FasFireLR mode?
            if (fflr_is_enabled)
            {
                // If the old state is one where we were processing a FasFireLR recording
                // this transition to live video is our chance to request the status of
                // the last FasFireLR recording.
                if ((old_camera_state == GIGE_CAMERA_STATE_DEARMING) || (old_camera_state == GIGE_CAMERA_STATE_ARMED) || (old_camera_state == GIGE_CAMERA_STATE_AUTOSAVE) || (old_camera_state == GIGE_CAMERA_STATE_REVIEW))
                {
                    // We need to get the FFLR status update.
                    fflr_status_needed = true;

                    // Let our saver know we are changing state here.
                    if (save_video)
                    {
                        save_video->going_live();
                    }
                }
            }
        }
    }

    // Make sure we properly handle the stream associated with this state.
    stream_setup(new_state, __PRETTY_FUNCTION__);

    // Remember the new "old" camera state.
    old_camera_state = new_state;

    // Make sure there is a camera group read object in case we need to read a group of camera registers.
    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Handle any state specific requirements.
    switch (new_state)
    {
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        if (fflr_status_needed)
        {
            // Query the FasFireLR status register.
            json_fflr_str.clear();
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Asking for the FFLR JSON status ------------------------";
            start_camera_read_buf(GIGE_REG_FFLR_JSON, GIGE_FFLR_JSON_LEN);
        }
        break;
    case GIGE_CAMERA_STATE_REVIEW:
    case GIGE_CAMERA_STATE_AUTOSAVE:
        // For new cameras, query the session size
        if (!is_series3())
        {
            cam_register_group->addReg(GIGE_REG_SESSION_SIZE);
        }

        /*
         * ... Fall through intentional...
         */
        [[fallthrough]];

    case GIGE_CAMERA_STATE_ARMED:
        // Query the save status register for an error.
        cam_register_group->addReg(GIGE_REG_SAVE_STATE);
        cam_register_group->lastReg(GIGE_REG_AUTOSAVE);
        read_camera_registers(cam_register_group);

        // If in FasFireLR mode is it fully prepared for processing?
        if (fflr_is_enabled)
        {
            // Has the FasFireLR been prepared?
            if (!save_video)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "fflr_is_enabled and No save_video object is allocated!";
            }
#if LOG_AUTOSAVE_TRACING
            else
            {
                uint32_t reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);
                bool gige_autosave_enabled = ((reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE);
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "FasFireLR is enabled state:" << new_state << "GIGE_REG_AUTOSAVE enabled:" << gige_autosave_enabled
                             << "save_video->is_autosave_enabled()=" << save_video->autosave_is_enabled()
                             << "save_video->is_autosave_configured()=" << save_video->autosave_is_configured();
            }
#endif
        }
        break;
    default:
        break;
    }

    // Ticket #492: Call the routine that manages the polling and visibility of the capture progress bar.
    // Parameter is "capturing" boolean and this will be set true if ARMED or TRIGGERED.
    // Note: The call is just a housekeeping call to make sure state matches capture progress bar.
    if ((new_state == GIGE_CAMERA_STATE_ARMED) || (new_state == GIGE_CAMERA_STATE_TRIGGERED))
    {
        rec_progress_state_check(true);     // It should be polling and visible.
    }
    else if ((new_state == GIGE_CAMERA_STATE_LIVE_VIDEO) || (new_state == GIGE_CAMERA_STATE_REVIEW))
    {
        rec_progress_state_check(false);    // It should not be polling and invisible.
    }
    // The else could be an ancillary transient state - ignore those states.
}

/*!
 * @brief Now that we've read the time/date information from the camera, let
 *        the user adjust as required.
 *
 * @param timeofday     Camera's time/date information.
 */
void CameraMDI::process_timeofday(uint32_t timeofday)
{
    QByteArray  *tempBuf;
    QString     tzName;
    QString     ntpServer;
    TimeDateDLG dlg;

    // Give the dialog the timezone ID being used by the camera.
    tempBuf = read_camera_buffer(GIGE_REG_TIMEZONE);
    if (tempBuf != nullptr)
    {
        tzName = QString::fromUtf8(*tempBuf);
        delete tempBuf;
        dlg.set_timezone(tzName);
    }

    // Give the dialog the NTP server URL being used by the camera.
    tempBuf = read_camera_buffer(GIGE_REG_NTP_SERVER);
    if (tempBuf != nullptr)
    {
        ntpServer = QString::fromUtf8(*tempBuf);
        delete tempBuf;
        dlg.set_ntp_server(ntpServer);
    }

    // Pass the time/date info to the dialog if there wasn't an NTP server.
    if (ntpServer.isEmpty())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "UTC Camera Time:" << Qt::showbase << Qt::hex << timeofday;
        dlg.set_time(timeofday);
    }

    // Put up our dialog.
    if (dlg.exec() == QDialog::Accepted)
    {
        // See what the user picked for an NTP server.
        ntpServer = dlg.get_ntp_server();
        if (ntpServer.isEmpty())
        {
            // The user is not using an NTP server. Instead, send
            // the new time/date value to the camera.
            write_camera_register(GIGE_REG_TIMEOFDAY, dlg.get_time());
            QLOG_INFO() << __PRETTY_FUNCTION__ << "UTC time:" << Qt::showbase << Qt::hex << dlg.get_time();
        }
        else
        {
            // Send the NTP server URL to the camera.
            write_camera_buffer(GIGE_REG_NTP_SERVER, ntpServer.toUtf8());
            QLOG_INFO() << __PRETTY_FUNCTION__ << "Set NTP server to:" << ntpServer;
        }

        // Pass the timezone ID to the camera. First we need to build up the string
        // in a format the camera expects.
        tzName = dlg.get_timezone() + QStringLiteral(";") + dlg.get_utc_offset();
        write_camera_buffer(GIGE_REG_TIMEZONE, tzName.toUtf8());
        QLOG_INFO() << __PRETTY_FUNCTION__ << "Set timezone to:" << tzName;
    }
}

/*!
 * @brief Send the configuration file specified by the "file_name" parameter to the camera.
 *
 * Assuming the file passes the sanity checks, it is likely to be larger than we can send
 * in a single write_mem operation. We start the multi-write operation here. When we get
 * status back from the communications object that the write has completed, we'll send the
 * next "chunk" out in [push_config_file_next](@ref CameraMDI::push_config_file_next()).
 *
 * @param file_name     The full path to the configuration file we're supposed to upload to the camera.
 */
void CameraMDI::push_config_file(QString& file_name)
{
    int         xfer_len;
    QByteArray  config_array;

    // Sanity check.
    if (MainWindow::GetInstance().isNull())
        return;

    // Make sure this configuration file is appropriate for our camera.
    if (!valid_config_file(file_name))
    {
        // Let the user know this is a bad config file.
        QMessageBox::warning(this, tr("Config File"), tr("The configuration file is not compatible with this camera [%1]. "
                                                         "Cannot update the camera's configuration.").arg(get_camera_name()));
        return;
    }

    // Initialize our variables to push the configuration to the camera.
    config_file.setFileName(file_name);
    if (!config_file.open(QIODevice::ReadOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open" << file_name << ". Err:" << config_file.errorString();
        return;
    }

    // Initialize our length.
    config_length = (int)config_file.size();

    // Figure out how many bytes we can push.
    if (config_length < GIGE_CONFIG_MAX_XFER_COUNT)
        xfer_len = config_length;
    else
        xfer_len = GIGE_CONFIG_MAX_XFER_COUNT - 1;

    // Adjust the amount of data left after the transfer.
    config_length -= xfer_len;

    // Read up our data.
    config_array.clear();
    config_array.append((char)GIGE_CONFIG_START_PACKET);
    config_array.append(config_file.read(xfer_len));

    // Start the transfer.
    write_camera_buffer(GIGE_REG_CONFIGURATION, config_array);
}

/*!
 * @brief For the multi-step upload of a configuration file to the camera, this method continues the process,
 *        sending the next "chunk" of the configuration data to the camera.
 *
 * @sa CameraMDI::push_config_file()
 */
void CameraMDI::push_config_file_next()
{
    int         xfer_len;
    QByteArray  config_array;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Make sure we need to be here.
    if (config_length <= 0)
    {
        // Pause the stream from the camera.
        if (camera_ptr->stream_running)
        {
            // Pause the stream while we're configuring the camera.
            pause_video_request(__PRETTY_FUNCTION__);
        }

        // Move the camera to the idle state and tell it to load the new configuration.
        write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_IDLE);
        write_camera_register(GIGE_REG_CS_REGS, GIGE_CS_REGS_LOAD);

        // Remember that we are having the camera reboot.
        camera_ptr->rebooting = true;

        // All done.
        return;
    }

    // Figure out how many bytes we can push. We can only push a maximum
    // of (GIGE_CONFIG_MAX_XFER_COUNT - 1) bytes, because the first byte
    // is a flag indicating which packet we're sending.
    if (config_length < GIGE_CONFIG_MAX_XFER_COUNT)
        xfer_len = config_length;
    else
        xfer_len = GIGE_CONFIG_MAX_XFER_COUNT - 1;

    // Adjust the amount of data left after the transfer.
    config_length -= xfer_len;

    // Read up our data.
    config_array.clear();
    if (config_length == 0)
        config_array.append((char)GIGE_CONFIG_LAST_PACKET);
    else
        config_array.append((char)GIGE_CONFIG_DATA_PACKET);
    config_array.append(config_file.read(xfer_len));

    // Start the transfer.
    write_camera_buffer(GIGE_REG_CONFIGURATION, config_array);

    // Are we done with the file?
    if (config_length == 0)
    {
        // Close the file.
        config_file.close();
    }
}

/*!
 * @brief Method to sanity check the raw metadata buffer.
 * @return True if the raw metadata appears to be OK, false otherwise.
 */
bool CameraMDI::raw_metadata_ok()
{
    bool    rc(true);           // Assume life is good.

    // There should be SOME raw metadata.
    if (raw_metadata.isEmpty())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Raw metadata buffer is empty!";
        rc = false;
    }
    else if ((raw_metadata.length() != raw_metadata_len) || ((raw_metadata.length() & (get_pfmd_len() - 1)) != 0))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "bytes received:" << raw_metadata.length() << "expected:" << raw_metadata_len;
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief Command the camera to reboot.
 */
void CameraMDI::reboot()
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    // Pause the stream from the camera.
    if (camera_ptr->stream_running)
        pause_video_request(__PRETTY_FUNCTION__);

    // Remember that we are having the camera reboot.
    camera_ptr->rebooting = true;

    // Have the camera reset to factory defaults.
    write_camera_register(GIGE_REG_RESET_CAMERA, 1);
}

/*!
 * @brief Utility method to read the camera's buffer cache. The assumption
 *        being made is that the buffer has already been transferred from
 *        the camera.
 *
 * @param[in] buf_addr      The FCP register address of the buffer we're interested in.
 * @param[out] ok           [OPTIONAL] pointer to a bool to indicate if the register is in cache.
 *
 * @return  Pointer to the QByteArray structure. It is the responsibility of the calling
 *          party to release the pointer.
 */
QByteArray *CameraMDI::read_camera_buffer(uint32_t buf_addr, bool *ok)
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";

        if (ok != nullptr)
            *ok = false;
        return nullptr;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return nullptr;
    }

    // This isn't going to work if the camera is in calibration.
    if (camera_ptr->offline)
    {
        if (ok != nullptr)
            *ok = false;
        return nullptr;
    }

    // Return the cached value.
    return comm_ptr->get_reg_buffer(get_ip_addr(), buf_addr, ok);
}

/*!
 * @brief Utility method to read the camera's register cache. The assumption
 *        being made is that the register value has already been loaded from
 *        the camera.
 *
 * @param[in]  reg_addr     The address of the register we're interested in.
 * @param[out] ok           [OPTIONAL] pointer to a bool to indicate if the register is in cache.
 *
 * @return The register value or 0 if an error (*ok == false).
 */
uint32_t CameraMDI::read_camera_cache(uint32_t reg_addr, bool *ok)
{
    FI_comm     *comm_ptr;

    // If we return early before call to get the register make sure ok is set to false.
    if (ok != nullptr)
    {
        *ok = false;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return 0;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return 0;
    }

    // This isn't going to work if the camera is in calibration.
    if (camera_ptr->offline)
    {
        return 0;
    }

    // Return the cached value.
    return comm_ptr->get_reg(get_ip_addr(), reg_addr, ok);
}

/*!
 * @brief Utility method to do a block read of a number of registers.
 * @param reg_list  The list of register addresses we're interested in.
 */
void CameraMDI::read_camera_registers(QSharedPointer<FiCamRegGroup> cam_reg_grp)
{
    int         comm_status;
    FI_comm     *comm_ptr;

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Make sure there is a group to read.
    if (cam_reg_grp.isNull() || (cam_reg_grp->getCount() == 0))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // If the camera is offline log a message.
    if (camera_ptr->offline)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << tr("The camera is offline and cannot read register information!");
        return;
    }

    // If we are already waiting for a completion of this same group of registers let's skip this request.
    // Check to see if this group being requested matches a group already being processed.
    for (int i = 0; i < group_reads.count(); i++)
    {
        QSharedPointer<FiCamRegGroup> cam_reg_grp_entry = group_reads.at(i);

        // If last registers are not the same, you don't need to compare signatures.
        if (!cam_reg_grp_entry.isNull() && (cam_reg_grp_entry->getLastRegAddr() == cam_reg_grp->getLastRegAddr()))
        {
            // Check for a duplicate camera group read.
            if (cam_reg_grp_entry->camRegGroupCompare(cam_reg_grp))
            {
#if LOG_GROUP_READS
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[DUPLICATE:" << cam_reg_grp_entry->getCamRegGroupSigStr() << "][INDEX:" << i << "][COUNT:" << group_reads.count() << "]";
#endif
                // Short-circuit the processing we don't want to overlap with the same group read request.
                return;
            }

            // The last register and the camera group signature string were NOT equal.
            // This means we have duplicate last addresses but not group contents and because
            // so much of the system relies on the last register signalling a unique group read
            // completion this should output a warning in the log file.
            if (cam_reg_grp_entry->getLastRegAddr() == cam_reg_grp->getLastRegAddr())
            {
//#if LOG_GROUP_READS
                QLOG_WARN() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[LAST_REG_CONFLICT][NEW:" << cam_reg_grp->getCamRegGroupSigStr() << "][INDEX:" << i << "][COUNT:" << group_reads.count() << "][QUEUED:" << cam_reg_grp_entry->getCamRegGroupSigStr() << "]";
//#endif
            }
        }
    }

    // Arriving here we need to process this group read.
    group_reads.append(cam_reg_grp);
#if LOG_GROUP_READS
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[APPENDED][NEW:" << cam_reg_grp->getCamRegGroupSigStr() << "][COUNT:" << group_reads.count() << "]";
#endif

    // Start the group read request.
    if ((comm_status = comm_ptr->start_read_multireg(get_ip_addr(), cam_reg_grp->getCamRegGroupList())) != FI_COMM::FI_SUCCESS)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << tr("The comm_ptr->start_read_multireg returned error:%1").arg(comm_status);
    }
}

/*!
 * @brief We previously requested the PFMD for the "active" frame. This frame number
 *        is stored in last_requested_frame. The camera server has completed the
 *        transfer of the PFMD, and it now available to us to read...
 *
 * ...OR, it was already in our cache (pfmd_info), and we can simply retrive the
 * data from there.
 */
void CameraMDI::read_frame_metadata()
{
    uint32_t requested_frame;
    const QByteArray *pfmd_array;

    // Sanity check.
    if (active_frame_requests.isEmpty())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "active_frame_requests is empty?!";
        return;
    }

    // Do we already have the PFMD for this frame cached?
    requested_frame = active_frame_requests.takeFirst();
    pfmd_array = pfmd_info->retrieve(requested_frame);
    if (pfmd_array == nullptr)
    {
        QByteArray *pfmd_buffer;

        // Read the metadata array.
        pfmd_buffer = read_camera_buffer((uint32_t)GIGE_REG_FRAME_METADATA);
        if (pfmd_buffer == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_reg_buffer( GIGE_REG_FRAME_METADATA ) returned NULL!";
            return;
        }

        // Save it in case we need to use it later.
        pfmd_info->insert(requested_frame, pfmd_buffer);

        // Point to our new buffer.
        pfmd_array = pfmd_buffer;
    }

    // If we have an information window, then pass on the PFMD information.
    if (camera_info)
    {
        camera_info->set_pfmd(pfmd_array);
    }

    // Handle video player processing.
    if (video_player && review_controls)
    {
        // BUGFIX #434:  It is possible to have a "reset" to -1 setting in last_requested_frame value.
        // In SVN[27028] check-in the if statement above used to read as follows:
        // if (video_player && review_controls && (requested_frame == last_requested_frame))
        //
        // BUT we need to *always* update the overlays otherwise they will remain stale and static
        // when playback or scrubbing. I have removed that condition to go back to the way it was before.
        // This is related to change in CameraMDI::video_current_frame_changed for Ticket #353, SVN[26947]
        // where we reset the last_requested_frame back to an initialized state of -1
        //
        // Make sure that the last_requested_frame is valid.
        // If it is -1 (initialized to never been set state) then we don't want to use that bad frame
        // number to update our overlay information or do anything else with it.  Make sure it gets set
        // to the real frame number here so that the update overlay always represents valid frame data.
        if (last_requested_frame == (uint32_t) -1)
        {
            // NOTE:  When testing this it was validated that the review_controls->get_current_frame_num()
            // was equal to the active_frame_requests.takeFirst() every time.  This means that we should
            // be able to count on the requested_frame being in sync with the review_controls object.
            // uint32_t cur_frame_no = review_controls->get_current_frame_num();
            // QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cur_frame_no:" << cur_frame_no << "requested frame:" << requested_frame << "last_requested_frame was NEGATIVE ONE!";

            // Set the last_requested_frame to the one being processed here.
            last_requested_frame = requested_frame;
        }

        // If the current frame is the trigger frame, we have some work to do.
        if (last_requested_frame == review_controls->get_trigger_frame_num())
        {
            // Get the straight-binary-seconds information from the PFMD.
            trigger_frame_start = calc_frame_time_from_PFMD((const uint8_t *)pfmd_array->constData());
        }

        // Update any overlays we might have.
        video_player->update_overlay(last_requested_frame, camera_info, nullptr);
    }
}

/*!
 * @brief Read the QSettings information to restore the non-volatile parameters from
 *        the "last" run.
 */
void CameraMDI::read_settings()
{
    QVariantList    settings_list;

    // Load our previous settings.
    QSettings settings;

    settings.beginGroup("EventMarkers");
    use_event_markers = settings.value("useEventMarkers", false).toBool();
    event_logic_is_and = settings.value("logicIsAnd", true).toBool();
    settings_list = settings.value("eventSelections").toList();
    settings.endGroup();

    // See if we're autodownloading the metadata.
    settings.beginGroup("SaveSettings");
    autodownload_metadata   = settings.value("autodownload_metadata", true).toBool();

    // Get the FFLR data.
    fflr_device_label       = settings.value("dest_fflr_device_label").toString();
    fflr_device_name        = settings.value("dest_fflr_device_name").toString();
    fflr_device_mountdir    = settings.value("dest_fflr_device_mountdir").toString();
    fflr_device_tag         = settings.value("dest_fflr_device_tag").toString();
    fflr_device_type        = settings.value("dest_fflr_device_type").toUInt();
    fflr_session_len        = settings.value("dest_fflr_session_len").toULongLong();
    settings.endGroup();

    // Get our video tracker settings.
    settings.beginGroup("vt_reticle");
    vt_area = settings.value("area", 50).toInt();
    vt_change = settings.value("change", 50).toInt();
    vt_mark_only = settings.value("mark_only", false).toBool();
    vt_refresh = settings.value("refresh", 16).toInt();
    settings.endGroup();

    // Reset our list.
    event_selections.clear();

    // Now add in each of the entries we read from the settings.
    foreach (QVariant v, settings_list)
    {
        event_selections.append(v.toInt());
    }

    // Make sure we have a selection list.
    while (event_selections.count() < FM_EVENT_FILTERS::INDEX_COUNT)
    {
        event_selections.append(FM_EVENT_FILTERS::SELECT_IGNORE);
    }
}

/*!
 * @brief Sets up rec_progress object to match the capturing condition of our camera.
 * @param capturing_video   Are we in a state that is capturing video right now?
 */
void CameraMDI::rec_progress_state_check(bool capturing_video)
{
    if (!rec_progress)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "rec_progress is null!?";
        return;
    }

    // If the camera should be capturing video make sure our progress bar reflects this condition.
    if (capturing_video)
    {
        QString capture_info_str = QString("CAPTURING VIDEO rec_progress");

        // Are we polling? We should be.
        if (!rec_progress->polling_is_active())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << capture_info_str << "was not polling. Starting now.";
            rec_progress->start_polling();      // Start polling to the default polling interval.
        }

        // Are we visible? We should be.
        if (!rec_progress->isVisible())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << capture_info_str << "was not visible. Making visible now.";
            rec_progress->setVisible(true);     // Make us visible.
        }
    }
    else
    {
        QString capture_info_str = QString("NOT CAPTURING VIDEO rec_progress");

        // Are we polling? We shouldn't be.
        if (rec_progress->polling_is_active())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << capture_info_str << "was polling. Stopping now.";
            rec_progress->stop_polling();       // Start polling to the default polling interval.
        }

        // Are we visible? We shouldn't be.
        if (rec_progress->isVisible())
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << capture_info_str << "was visible. Making invisible now.";
            rec_progress->setVisible(false);    // Make us invisible.
        }

        // Reset the progress control.
        rec_progress->reset();
    }
}


/*!
 * @brief Refresh the register cache for a few registers that we won't get unsolicited messages for.
 */
void CameraMDI::refresh_cache()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << get_ip_addr_pretty();

    // If any of these are true, then we shouldn't be talking to the camera. We
    // don't use the is_busy() helper method, because we do want to read these during
    // autosave.
    if ((camera_ptr && camera_ptr->offline) || shutdown_in_progress() || in_calibration())
        return;

    // Don't do anything else if we're in holdoff or the streamer is busy.
    if (!cache_refresh_holdoff && (streamer && !streamer->is_busy()))
    {
        QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();

        // Was the camera register group object created?
        if (cam_register_group.isNull())
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
            return;
        }

        // These are the registers we want to refresh.
        cam_register_group->addReg(GIGE_REG_CCP);

        // Things to read if we're not a series3 camera.
        if (!is_series3())
        {
            cam_register_group->addReg(GIGE_REG_ICM_MODULE_VER);
        }

        cam_register_group->addReg(GIGE_REG_SYNC_IN_TIMING);
        cam_register_group->addReg(GIGE_REG_AVAIL_MEDIA);
        cam_register_group->addReg(GIGE_REG_BATTERY_CHARGE);
        cam_register_group->addReg(GIGE_REG_FORMAT_MEDIA);
        cam_register_group->addReg(GIGE_REG_GAMMA);
        cam_register_group->addReg(GIGE_REG_JPEG_QUALITY);
        cam_register_group->addReg(GIGE_REG_AUTOSAVE);
        cam_register_group->addReg(GIGE_REG_WIFI_STATUS);
        cam_register_group->addReg(GIGE_REG_SENSOR_GAIN);
        cam_register_group->addReg(GIGE_REG_USER_SAVE_START_NUM);
        cam_register_group->addReg(GIGE_REG_USER_SAVE_END_NUM);
        cam_register_group->lastReg(GIGE_REG_SAVE_STATE);

        // Fire off a group read.
        read_camera_registers(cam_register_group);
    }
    else
    {
        // TODO: Ask Tony if we even need to do this if the above if statement is true.
        // All we need to do is to ask for the GIGE_REG_CCP register.
        start_camera_read_reg(GIGE_REG_CCP);
    }
}

/*!
 * @brief A utility method to allow child widgets (CameraVideo, for example) to release image buffers.
 * @param buf_ptr   The image buffer pointer we're interested in releasing.
 */
void CameraMDI::release_buffer(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr)
{
    // Sanity checks.
    if ((streamer == nullptr) || (imageBufferPtr == nullptr))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Bad parameter!";
        return;
    }

    // Release the image buffer.
    streamer->image_release(std::move(imageBufferPtr));
}

/*!
 * @brief Newer cameras can simply restart the camera server, which is much
 *        faster than a full reboot.
 */
void CameraMDI::reset()
{
    // If we're series3, just do a reboot.
    if (is_series3())
    {
        reboot();
        return;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Pause the stream from the camera.
    if (camera_ptr->stream_running)
        pause_video_request(__PRETTY_FUNCTION__);

    // Remember that we are having the camera reboot.
    camera_ptr->rebooting = true;

    // Have the camera reset to factory defaults.
    write_camera_register(GIGE_REG_RESET_CAMERA, GIGE_RESET_RESTART);
}

/*!
 * @brief Kick off the write process for GIGE_REG_ACTIVE_METADATA.
 * Must receive a confirmation of the write before blocks of metadata can be processed.
 * Created to provide common interface for this request.
 */
void CameraMDI::request_active_metadata(const QString& who_called)
{
#if LOG_ACTIVE_METADATA_DOWNLOAD
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "who_called:" << who_called << "active_metadata_requested:" << active_metadata_requested;
#endif

    // Are we already waiting for a confirmation of GIGE_REG_ACTIVE_METADATA write?
    if (active_metadata_requested)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << who_called << "attempting to write GIGE_REG_ACTIVE_METADATA but already have pending write request. Ignoring request!";
        return;
    }

    // Start the process of writing the GIGE_REG_ACTIVE_METADATA register.
    active_metadata_requested = true;

    // The CameraMDI::write_complete function will continue this process when it
    // receives notification from the camera. This way we ensure that the camera
    // is ready to start sending blocks of pm_metadata our way.
    write_camera_register(GIGE_REG_ACTIVE_METADATA, 0);
}

/*!
 * @brief Read the per-frame metadata for the frame specified by the "frame_num"
 *        parameter from the attached camera.
 *
 * @param frame_num     The absolute frame number of the frame we want the per-frame metadata for.
 * @param active        Flag where true (default) means this is an active frame, false is just some frame
 *                      we are interested in.
 */
void CameraMDI::request_pfmd(uint32_t frame_num, bool active)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frame_num:" << frame_num << "active:" << active << "pfmd_frame_requests:" << pfmd_frame_requests << "active_frame_requests:" << active_frame_requests;

    // If we are a series3 camera or active, we have to work with the "active" registers.
    if (is_series3() || active)
    {
        // With the "active" registers we cannot queue more than one request at a time. Check
        // to see if any requests are queued before writing to the register.
        if (active_frame_requests.isEmpty())
        {
            // This is the first, so write the register.
            write_camera_register(GIGE_REG_ACTIVE_FRAME_NUM, frame_num);

            // Remember the frame we are requesting.
            active_frame_requests.append(frame_num);
        }

        // Make sure we don't already have a request pending.
        else if (!active_frame_requests.contains(frame_num))
        {
            // Remember the frame we are requesting.
            active_frame_requests.append(frame_num);
        }
    }

    // For general purpose, and non-series3 cameras...
    // Make sure we don't already have an pending request for this frame.
    else if (!pfmd_frame_requests.contains(frame_num))
    {
        // ... we can use the new register interface. This interface on the camera server
        // supports queuing of multiple requests, seviced in a FIFO fashion.
        write_camera_register(GIGE_REG_FRAME_PFMD, frame_num);

        // Remember the frame we are requesting.
        pfmd_frame_requests.append(frame_num);
    }
}

/*!
 * @brief   Make sure our event information is reset.
 */
void CameraMDI::reset_event_info()
{
    // Are we using event markers?
    if (use_event_markers)
    {
        // Release the old data.
        marker_list.clear();
        review_controls->set_markers(marker_list);
    }

    // Did we have the I/O graph displayed?
    if (!raw_path.isEmpty())
    {
        QDir raw_dir;

        // Make sure our raw_path file is removed.
        if (raw_dir.exists(raw_path))
        {
            raw_dir.remove(raw_path);
            raw_path.clear();
        }
    }

    // Reset our buffers.
    raw_metadata.clear();
    raw_metadata_len = 0;
}

/*!
 * @brief Process a request to resume the video stream from the camera.
 */
void CameraMDI::resume_video_request()
{
    // Make sure we want to stream and the camera isn't already streaming.
    if (camera_ptr && camera_ptr->stream && !camera_ptr->stream_running)
    {
        uint32_t    reg_value;

        // Get the camera state and update our stream accordingly.
        reg_value = read_camera_cache(GIGE_REG_CAMERA_STATE);
        stream_setup(reg_value, __PRETTY_FUNCTION__);
    }
}

/*!
 * @brief If there are frames in the retry_frame_list, then we pop the first entry off the list
 *        and ask the camera to send it. If the list is empty, then we're done saving and it's
 *        time to wrap up.
 */
void CameraMDI::retry_missing_frames()
{
    bool        blob_processing;
    uint32_t    retry_frame_abs;    // Absolute just means the 0-based frame in our capture versus relative to trigger frame.

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "retry_frame_list.count():" << retry_frame_list.count()
                 << "retry_in_progress:" << retry_in_progress << "retry_frames_reqd.count()" << retry_frames_reqd.count();

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "save_video is NULL!";
        return;
    }

    // Determine if we're processing "blobs" or frames. We use the "CAPTURE" format here
    // to distinguish between copying an AVI file from the camera and copying a stack of
    // images from the camera and saving them as an AVI file on the PC. We're not ACTUALLY
    // copying CAP files to the PC.
    if (save_video->getStreamType() == FI_SAVER::STREAM_TYPE::FI_CAP)
        blob_processing = true;
    else
        blob_processing = false;

    // Are there missing frames to download?
    if (retry_frame_list.isEmpty())
    {
        // Reset our retry flag.
        retry_in_progress = false;

        // Let the video saver object know we're done.
        if (blob_processing)
            save_video->saving_finished(__PRETTY_FUNCTION__);

        // We're done.
        return;
    }

    // If this is the very first frame being processed for retry give the user some information
    // in case there is a delay.
    if (!retry_in_progress)
    {
        // Start the retry process. It can be lengthy sometimes.
        QString statusMsg = tr("Downloading %1 missed frame%2").arg(retry_frame_list.count()).arg(((retry_frame_list.count() > 1) ? "s " : " "));
        MainWindow::GetInstance()->set_statusbar_message(statusMsg);
    }

    // Get the first frame we'll want to retry.
    retry_frame_abs = retry_frame_list.takeFirst();

    // Make a relative frame number out of our retry_start value.
    int retry_frame_rel = (int)retry_frame_abs - (int)trigger_frame;

    // How many frames have we already requested? Grab before it gets appended.
    uint32_t retry_frames_resent  = retry_frames_reqd.count();

    // We are going to process this frame. Add it to our requested list in case somebody
    // tries to re-add the retry frame.
    retry_frames_reqd.append(retry_frame_abs);

    // Grab the total number of retry frames being processed.
    uint32_t total_retry_frames = retry_frame_list.count() + retry_frames_reqd.count();

    // Let the saver object know what we're attempting.
    save_video->setRetryFrame(retry_frame_abs, retry_frame_rel, retry_frames_resent, total_retry_frames);
    save_video->setCount(1);

    // Are we processing a "blob"?
    if (blob_processing)
    {
        // Reset our last received frame. We are going to get a sequence number of "retry_start",
        // so our "last_received_frame" need to be the one before that.
        last_received_frame = retry_frame_abs - 1;
    }
    else
    {
        // We're dealing with frames, so the sequence number will be coming back as what we request
        // plus one. This still keeps "last_received_frame" as one before.
        last_received_frame = retry_frame_abs;
    }

    // Remember that we're in retry mode.
    retry_in_progress = true;

    // We're going to be waiting for the frame.
    waiting_for_frame = true;

    // Make sure our dialog shows that we're in Retry mode.
    QString retryStatus = tr("Retry missing frames");
    if (xferDialog)
    {
        xferDialog->setStatusValue(retryStatus);
        xferDialog->showStatusValue(true);
    }

    // Is the missing piece a file or a chunk? When downloading large files,
    // such as an AVI video, the format is "CAPTURE" (not AVI...that's for when
    // we're CREATING an AVI.)
    if (blob_processing)
        retry_frame_abs |= GIGE_STREAM_RETRY_OFFSET;

    // Reset our retry counter.
    retry_frame_count = 0;

    // Write out our registers.
    streamer->retry(retry_frame_abs);
}

/*!
 * @brief Check to see the review window has received at least one frame!
 * @return True if it has, false if not.
 */
bool CameraMDI::review_has_received_a_frame()
{
    bool received_one_frame = false;
    uint32_t camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Must be in review.
    if (camera_state == GIGE_CAMERA_STATE_REVIEW)
    {
        // Has our video player received at least 1 frame?
        if (video_player)
        {
            // Make sure we have at least one frame in the review panel.
            received_one_frame = video_player->has_received_1_image();
        }
    }

    return received_one_frame;
}

/*!
 * @brief Check to see if it's "safe" to arm the camera.
 *
 * @param[in, out]  user_queried    Set to true if the user was asked, false if no user intervention was required.
 * @return True if it is OK to arm, false otherwise.
 */
bool CameraMDI::safe_to_arm(bool& user_queried, bool& allow_discard)
{
    bool        ok_to_arm;
    uint32_t    ltr_record_type;
    uint32_t    camera_state;
    uint32_t    part_state;

    // Assume we can discard.
    allow_discard = true;

    // Assume it's not OK to ARM the camera.
    ok_to_arm = false;

    // Is LTR capture mode enabled?
    if (is_ltr_enabled(ltr_record_type))
    {
        // Are we in ROC or BROC mode?
        if ((ltr_record_type & (GIGE_LTR_MODE_ROC | GIGE_LTR_MODE_BROC)) != 0)
        {
            // Discarding is not an option.
            allow_discard = false;
        }
    }

    // What state is the camera in?
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    switch (camera_state)
    {
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
    case GIGE_CAMERA_STATE_REVIEW:
    case GIGE_CAMERA_STATE_TRIGGERED:
        // Do we have a dirty partition?
        part_state = read_camera_cache(GIGE_REG_PART_STATE);
        if (!user_queried && allow_discard && ((part_state == GIGE_PART_STATE_COMPLETE) || (part_state == GIGE_PART_STATE_ARMED)))
        {
            uint32_t    reg_autosave;
            uint32_t    as_dirty;
            QString     msg_str;

            // Get the autosave register.
            reg_autosave = read_camera_cache(GIGE_REG_AUTOSAVE);
            as_dirty = GIGE_AUTOSAVE_TOTAL_COUNT(reg_autosave) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_autosave);

            if (as_dirty > 1)
            {
                msg_str = tr("There are %1 unsaved partitions. Arming the camera will result in the unrecoverable loss of the captured video.\n\n"
                             "Are you sure you wish to ARM the camera?").arg(as_dirty);
            }
            else
            {
                msg_str = tr("Arming the camera will result in the unrecoverable loss of the captured video.\n\n"
                             "Are you sure you want to Arm the camera?");
            }

            // We are querying the user.
            user_queried = true;

            // Better make sure the user REALLY wants to do this.
            int user_response = QMessageBox::question(this, get_camera_name(), msg_str,
                                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
            if (user_response == QMessageBox::Yes)
            {
                // User says it's OK.
                ok_to_arm = true;
            }
        }
        else
        {
            // Must be OK.
            ok_to_arm = true;
        }
        break;

    case GIGE_CAMERA_STATE_ARMED:
    case GIGE_CAMERA_STATE_AUTOSAVE:
        ok_to_arm = true;
        break;
    case GIGE_CAMERA_STATE_SHUTDOWN:
        ok_to_arm = false;
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "unexpected camera state:" << camera_state;
        break;
    }

    // Return the answer.
    return ok_to_arm;
}

/*!
 * @brief Save the [raw metadata](@ref CameraMDI::raw_metadata) to a file in the user's temporary
 *        folder. The file name is "raw_<camera-ip-addr>.pm", where <camera-ip-addr> is the Qt::hexadecimal
 *        IP address of the camera we're connected to. Note that the address may also be byte-swapped.
 *
 * @return The fully qualified path name to the raw metadata file. This will be an empty QString
 *         if an error occurred.
 */
QString CameraMDI::save_raw_metadata()
{
    QDir        temp_dir(QDir::temp());
    QFile       temp_file;

    // Do we already have the name?
    if (raw_path.isEmpty())
    {
        // Define the place for the raw metadata.
        raw_path = QString::fromUtf8("raw_%1.pm").arg(get_ip_addr(), 0, 16);
        raw_path = temp_dir.absoluteFilePath(raw_path);
    }

    // Open up the file.
    temp_file.setFileName(raw_path);
    if (!temp_file.open(QIODevice::WriteOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open:" << raw_path << "Err:" << temp_file.errorString();
        raw_path.clear();
    }
    else
    {
        temp_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
        temp_file.write(raw_metadata);
        temp_file.close();
    }

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << raw_path;

    // Return the filename.
    return raw_path;
}

/*!
 * @brief Save the user specified settings to the non-volatile parameter storage.
 */
void CameraMDI::save_settings()
{
    int     index;
    QSettings settings;
    QVariantList    settings_list;

    // Initialize our variant list.
    for (index = 0; index < event_selections.count(); ++index)
        settings_list.append(event_selections.at(index));

    settings.beginGroup("EventMarkers");
    settings.setValue("useEventMarkers", use_event_markers);
    settings.setValue("logicIsAnd", event_logic_is_and);
    settings.setValue("eventSelections", settings_list);
    settings.endGroup();

    settings.beginGroup("vt_reticle");
    settings.setValue("change", vt_change);
    settings.setValue("area", vt_area);
    settings.setValue("mark_only", vt_mark_only);
    settings.setValue("refresh", vt_refresh);
    settings.endGroup();
}

/*!
 * @brief Handle timeouts while saving a video to the workstation.
 */
void CameraMDI::save_video_timeout()
{
    uint32_t    last_frame_0_based = 0;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << get_ip_addr_pretty();

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "What?! No save_video?";
        return;
    }

    // What is the last absolute sequence number for our collection of frames being saved?
    last_frame_0_based = save_video->getFrameCount() - 1;

    // Note:  When retrying frames the save_video->getFrameCount() will be 1 because
    // we retry one frame at a time. It is important not to look for the last frame missing
    // if already retrying the missed frames!
    // If not retrying missed frames, is this a timeout on the last frame?
    if (!retry_in_progress && (last_received_frame == last_frame_0_based))
    {
        uint32_t lost_frame_id = save_video->getClipStart() + last_frame_0_based;

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "LAST FRAME TIMEOUT! last_received_frame:" << last_received_frame << "FiSaver::frame count:" << save_video->getFrameCount() << "lost_frame_id:" << lost_frame_id;

        // Our retry list contains the absolute frame number of the frame we want.
        add_to_retry_list(lost_frame_id);

        // Start retrying the missing frames now.
        retry_missing_frames();
    }

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "last_received_frame:" << last_received_frame << "FiSaver::frame count:" << save_video->getFrameCount();

    // Are we in a retry pass?
    if (retry_in_progress && (retry_frame_count < MAX_RETRY_FRAMES))
    {
        uint32_t    retry_start;

        // Determine if we're processing "blobs" or frames. We use the "CAPTURE" format here
        // to distinguish between copying an AVI file from the camera and copying a stack of
        // images from the camera and saving them as an AVI file on the PC. We're not ACTUALLY
        // copying CAP files to the PC.
        if (save_video->getStreamType() == FI_SAVER::STREAM_TYPE::FI_CAP)
        {
            retry_start = last_received_frame + 1;
            retry_start |= GIGE_STREAM_RETRY_OFFSET;
        }
        else
        {
            retry_start = last_received_frame;
        }

        // Increment our retry counter.
        retry_frame_count++;

        // Write out our registers.
        streamer->retry(retry_start);
    }
    else if (save_video->save_in_progress())
    {
        QString     err_str;
        QString     saveStatusStr;

        // There is the possibility that the save process is consumed
        // with encoding frames and all of the input frames have been
        // received.  This is the case if the last_received_frame is equal
        // to the frame count.  As long as there isn't a retry in progress
        // condition then we can assume the saving_finished state is valid
        // to call.  This sets up the final progress push for the encoder.
        // What was happening was that this code would fall through and assume
        // the video was unable to finish when in fact it had finished and
        // the stream writer is waiting on the encoder.
        // REVISIT_JLC:
        // This is an area that needs to be improved upon - in general the
        // entire SaveVideo object needs some tightening up to manage its
        // stream writer.
        if (!retry_in_progress && (last_received_frame >= save_video->getFrameCount()))
        {
            // All we do at this point is tell the save video we should be done.
            save_video->saving_finished(__PRETTY_FUNCTION__);

            // Go ahead and just return here.  We should be ok.  If not
            // the user can cancel the progress dialog.
            return;
        }
        else
        {
            // What is our actual save status?
            uint32_t    frameCount;
            uint32_t    framesAdded;
            uint32_t    framesSaved;
            bool        streamIsOk;
            bool        streamIsComplete;

            // NOTE:  It is possible to get into a true timeout condition where the saver is still in a
            // good state and is just waiting for frames.  We should eventually consider multiple entry
            // to this location with stale data an error and true loss of the input stream.
            // TODO:  Determine the best strategy here if this is "stuck"...  The user can always cancel

            // What is the state of this save operation.
            save_video->getSaveStatus(saveStatusStr, frameCount,
                                      framesAdded, framesSaved, streamIsOk, streamIsComplete);

            // Is the stream OK and are we completed?
            if (streamIsOk && streamIsComplete)
            {
                // Make sure there are no retry frames and if so we can assume this copy completed.
                if (retry_frame_list.isEmpty())
                {
                    // We really should be finished here.
                    save_video->saving_finished(__PRETTY_FUNCTION__);

                    // Go ahead and just return here.  We should be ok.  If not
                    // the user can cancel the progress dialog.
                    return;
                }
            }
        }

        // Let the user know that things have gone badly.
        err_str = tr("The save operations failed due to a timeout waiting for frames.");
        if (retry_in_progress || (last_received_frame > 0))
        {
            uint32_t    abs_frame;
            int         rel_frame;

            // If we're retrying, then the frame is the start frame.
            if (retry_in_progress)
                abs_frame = save_video->getClipStart();
            else
                abs_frame = save_video->getClipStart() + (last_received_frame - 1);
            rel_frame = (int)abs_frame - (int)trigger_frame;

            err_str.append(tr("<p>The last frame received was frame %1 (frame number=%2).").arg(rel_frame).arg(abs_frame));
        }
        else
        {
            err_str.append(tr("<p>No frames have been received."));
        }

        // Append any save status information we may have pulled.
        err_str.append(tr("<p>%1").arg(saveStatusStr));

        // Let folks know we've had a failure.
        emit save_timeout();

        // ==================================================================================
        // At this point, the camera has stopped streaming. Can we retry at this point, or do
        // we just need to cancel the save?
        cancel_save();
        // ==================================================================================

        // Let folks know we're done.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
        fiMsgBox::critical(nullptr, get_camera_name(), err_str);
    }
    else
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "No save in progress?";

        // Start retrying the missing frames.
        retry_missing_frames();
    }
}

bool CameraMDI::set_vtrig_marker(int index, uint32_t& event_word, uint32_t *compare_ptr, uint32_t vtrig_word, uint32_t last_vtrig_word)
{
    bool changed;
    uint32_t vtrig_flags;
    uint32_t last_flags;

    // Assume we will not make any changes.
    changed = false;

    // Get our flags "normalized".
    vtrig_flags = (vtrig_word & CAMERA_INFO::VTRIG_FLAGS_MASK) >> CAMERA_INFO::VTRIG_FLAGS_SHIFT;
    last_flags = (last_vtrig_word & CAMERA_INFO::VTRIG_FLAGS_MASK) >> CAMERA_INFO::VTRIG_FLAGS_SHIFT;

    // If there is a refresh event, we'd like to show that as a glitch.
    if ((vtrig_flags & CAMERA_INFO::VTRIG_FLAGS_REFRESH) == CAMERA_INFO::VTRIG_FLAGS_REFRESH)
    {
        // Mark this.
        event_word |= (1 << (FM_EVENT_FILTERS::EDGE_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));
    }

    // Look to see if we had a video trigger event.
    if ((vtrig_flags & CAMERA_INFO::VTRIG_FLAGS_TRIGGER) == CAMERA_INFO::VTRIG_FLAGS_TRIGGER)
    {
        // We have a trigger event, so set the PFMD event data.
        event_word |= ((uint32_t)1 << (FM_EVENT_FILTERS::EVENT_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));

        // Are we at the beginning of our search?
        if (index == 0)
        {
            // Is there a compare word?
            if (compare_ptr != nullptr)
            {
                // OR in the VTRIG marker into the compare_event.
                *compare_ptr |= ((uint32_t)1 << (FM_EVENT_FILTERS::EVENT_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));
            }
        }
        else
        {
            // We should have a last trigger value, so see if we should set an edge. We have a trigger
            // in the current frame, so if we didn't have one in the previous frame there is an edge.
            if ((last_flags & GIGE_VTRIG_CONTROL_TRIGGER) != GIGE_VTRIG_CONTROL_TRIGGER)
            {
                // There is a "rising" edge here.
                event_word |= (1 << (FM_EVENT_FILTERS::EDGE_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));
            }
        }

        // Remember that we've changed the PFMD event data.
        changed = true;
    }
    else if (index > 0)
    {
        // Check to see if we just lost the trigger. We do NOT have it in this frame, but did we
        // have it in the previous frame?
        if ((last_flags & GIGE_VTRIG_CONTROL_TRIGGER) == GIGE_VTRIG_CONTROL_TRIGGER)
        {
            // There is a "falling" edge here.
            event_word |= (1 << (FM_EVENT_FILTERS::EDGE_FLAG_OFFSET + FM_EVENT_FILTERS::VT_INDEX));

            // Remember that we've changed the PFMD event data.
            changed = true;
        }
    }

    // Return our changed flag.
    return changed;
}

/*!
 * @brief Search through the [raw metadata buffer](@ref CameraMDI::raw_metadata) looking
 *        for events which match the [user specified criteria](@ref eventFilterDlg).
 *
 * The results are stored in a [list](@ref CameraMDI::marker_list) which is passed to
 * the [video review object(@ref CameraReview) for display.
 *
 * Since this processing can take quite a long time (there can be a large number of frames),
 * this is a "distributed" function. Once we've processed a [certain number of frames](@ref CAMERA_MDI_SETTINGS::MAX_SEARCH_COUNT),
 * a [timer](@ref CAMERA_MDI_SETTINGS::SEARCH_DELAY) is set and the processing loop is exit.
 *
 * After the timer expires, processing is resumed.
 */
void CameraMDI::search_for_events()
{
    int index;
    int num_frames;
    int meta_index;
    uint32_t last_markers(0);
    uint32_t last_vtrig_pfmd(0);

    // How many frames are we dealing with?
    num_frames = raw_metadata_len / get_pfmd_len();

    // Seek to the starting position.
    if (search_start_index == 0)
    {
        // Reset our marker list.
        marker_list.clear();

        // Make sure there are enough frames to warrant a progress dialog
        // and if so create one and show it.
        if (num_frames > (2 * MAX_SEARCH_COUNT))
        {
            // Create and show a simple progress dialog for calibration.
            progressDialog = fiProgressDialog::CreateSimpleProgressDialog(tr("Event Marker Processing"), tr("Searching for event markers..."), QMessageBox::Cancel, __PRETTY_FUNCTION__, this);

            // Display the progress dialog.
            if (progressDialog)
            {
                progressDialog->setCancelText(tr("Cancel Search"));
                progressDialog->setBarMinMax(0, 0);
                progressDialog->showBar(true);
                connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CameraMDI::event_search_canceled);
                progressDialog->showDialogWithoutPolling();
            }
        }

        // Starting from the start.
        meta_index = 0;
    }
    else
    {
        // Move to the last frame we processed.
        meta_index = get_pfmd_len() * (int)(search_start_index - 1);

        // Initialize our "last" markers value. The marker is the third element (offset=2).
        memcpy(&last_markers, raw_metadata.constData() + meta_index + (2 * sizeof(uint32_t)), sizeof(last_markers));

        // series3 cameras do not have video trigger events.
        if (!is_series3())
        {
            uint32_t pfmd_event;

            // Get the event marker word from the PFMD in the right order.
            pfmd_event = qToBigEndian(last_markers);

            // Read the video trigger flags. The flags are in the 8th word (offset=7).
            memcpy(&last_vtrig_pfmd, raw_metadata.constData() + meta_index + (7 * sizeof(uint32_t)), sizeof(last_vtrig_pfmd));

            // Update the event marker word with the video trigger information. If we
            // actually change the PFMD, then update the raw event too.
            if (set_vtrig_marker(search_start_index, pfmd_event, nullptr, last_vtrig_pfmd, 0))
            {
                last_markers = qFromBigEndian(pfmd_event);
            }
        }

        // Bump our pointer up.
        meta_index += get_pfmd_len();
    }

    // Loop through the frames of metadata.
    for (index = search_start_index; index < num_frames; ++index, meta_index += get_pfmd_len())
    {
        uint32_t pfmd_event;
        uint32_t compare_event;
        uint32_t raw_event;

        // Do we need to give the UI a chance to get in?
        if (index == (search_start_index + MAX_SEARCH_COUNT))
        {
            // Pick up where we left off.
            search_start_index = index;
            QTimer::singleShot(SEARCH_DELAY, this, &CameraMDI::search_for_events);
            break;
        }

        // Read the marker flags. The flags are in the 3rd word (offset=2).
        memcpy(&raw_event, raw_metadata.constData() + meta_index + (2 * sizeof(uint32_t)), sizeof(raw_event));

        // If this is the first frame, simply initialize the "last" markers
        // to the value we just read.
        if (index == 0)
            last_markers = raw_event;

        // For non-series3 we need to byte swap the event marker word.
        if (is_series3())
        {
            // Everything is ready to go. No byte swapping needed.
            pfmd_event = raw_event;
            compare_event = last_markers;
        }
        else
        {
            uint32_t vt_pfmd;

            // Get the event marker word from the PFMD, as well as the event markers
            // we used the "last time".
            pfmd_event = qToBigEndian(raw_event);
            compare_event = qToBigEndian(last_markers);

            // Read the video trigger flags. The flags are in the 8th word (offset=7).
            memcpy(&vt_pfmd, raw_metadata.constData() + meta_index + (7 * sizeof(uint32_t)), sizeof(vt_pfmd));

            // Update the event marker word with the video trigger information. If we
            // actually change the PFMD, then update the raw event too.
            if (set_vtrig_marker(index, pfmd_event, &compare_event, vt_pfmd, last_vtrig_pfmd))
            {
                raw_event = qFromBigEndian(pfmd_event);
            }

            // Remember the video trigger PFMD for next time.
            last_vtrig_pfmd = vt_pfmd;
        }

        // Does this frame meet the event critera we're searching for?
        if (eventFilter::match_criteria(compare_event, pfmd_event, event_selections, event_logic_is_and, is_series3()))
        {
            bool    handled(false);
            int     cur_count = marker_list.size();

            // Get the trigger relative index for this frame.
            int rel_index = index - (int)trigger_frame;

            // We want to add this frame to our list of markers, but squeeze
            // each run into a sequence of { first_frame, RUN_FLAG, last_frame }.

            // If it's the first or second, just copy - no squeezing yet.
            if (cur_count < 2)
            {
                marker_list.append(rel_index);
                handled = true;
            }

            // At this point, we know we have at least 3 entries in our list
            // (cur_count >= 2), so we can safely do the following math.
            if (!handled &&
                (marker_list.at(cur_count - 1) == (rel_index - 1)) &&
                (marker_list.at(cur_count - 2) == (rel_index - 2)))
            {
                // Now we know we have a run of three consecutive frame numbers,
                // so it's time to switch to the run encoding.
                marker_list.replace(cur_count - 1, INT_MAX);       // Use the maximum positive value - we're not likely to have 2GBytes worth of frames.
                marker_list.append(rel_index);
                handled = true;
            }

            // See if we're extending the run.
            if (!handled &&
                (marker_list.at(cur_count - 1) == (rel_index - 1)) &&
                (marker_list.at(cur_count - 2) == INT_MAX))
            {
                // Replace the last_frame value.
                marker_list.replace(cur_count - 1, rel_index);
                handled = true;
            }

            // No conditions apply, just copy it over.
            if (!handled)
                marker_list.append(rel_index);
        }

        // Remember this marker value.
        last_markers = raw_event;

        // Have we exceeded our maximum number of found markers?
        if (marker_list.size() >= FM_EVENT_FILTERS::MAX_MARKER_LIST_SIZE)
        {
            // Do we have a progress dialog up?  If so reset it.
            if (progressDialog)
            {
                // Recycle any resources and release the progress dialog.
                fiProgressDialog::RecycleProgressDialog(progressDialog);
            }

            // We've reached the limit.
            QMessageBox::information(this, tr("Event Marker Processing"),
                                     tr("The maximum number of event markers has been exceeded.\n"
                                        "Please refine your criteria, or review your hardware setup."));

            // Go with what we now have.
            review_controls->set_markers(marker_list);

            // Let folks know we're done.
            emit event_search_complete();
            break;
        }
    }

    // Are we done?
    if (index == num_frames)
    {
        // Do we have a progress dialog up?  If so reset it.
        if (progressDialog)
        {
            // Recycle any resources and release the progress dialog.
            fiProgressDialog::RecycleProgressDialog(progressDialog);
        }

        // Set the new event markers.
        review_controls->set_markers(marker_list);

        // Let folks know we're done.
        emit event_search_complete();
    }
    else if (progressDialog)
    {
        // Update the progress dialog.
        progressDialog->setBarValue(index);
    }
}

/*!
 * @brief Called once on window creation to initialize our camera information
 *        pointer.
 *
 * @param cam_info  Points to a (persistent) memory structure containing all
 *                  the information about our camera.
 */
void CameraMDI::set_camera_info(FI_COMM::FI_CAMERA_INFO *cam_info)
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Are we closing down?
    if ((camera_ptr != nullptr) && (cam_info == nullptr))
    {
        // Disconnect the camera.
        comm_ptr->disconnect_cam(camera_ptr->ip_addr);
    }

    // Remember the camera information being passed in.
    camera_ptr = cam_info;

    // This is fine. The caller is probably closing us down.
    if (camera_ptr == nullptr)
        return;

    // Not currently streaming.
    camera_ptr->stream_running = false;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << "serial_no:" << camera_ptr->serial_num;

    // Create a network communications object for this camera.
    if (!comm_ptr->connect_cam(cam_info->ip_addr))
    {
        QHostAddress    dest_addr;
        QString err_msg;

        dest_addr.setAddress(qFromBigEndian(cam_info->ip_addr));
        err_msg = QString::fromUtf8("Unable to connect to camera %1!").arg(dest_addr.toString());
        QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        return;
    }

    // Update our window title to utilize the information we just received.
    setWindowTitle(user_friendly_name());

    // Connect the communications interfaces that we're interested in. Because a signal/slot invocation is effectively a function
    // call (according to the Qt documentation), to prevent recursive operation we want the read/write completion slots to be
    // invoked later (a "Queued" connection).
    connect(comm_ptr, &FI_comm::camera_offline_status_changing, this, &CameraMDI::camera_offline, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &CameraMDI::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraMDI::write_complete, Qt::QueuedConnection);

    // There are specific registers we want the save video to be in charge of so we will establish its write complete connection.
    if (save_video)
    {
        connect(comm_ptr, &FI_comm::write_reg_complete, save_video, &SaveVideo::write_complete, Qt::QueuedConnection);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_video object is NULL?!!!:";
    }

    // Remember that we're in startup.
    in_startup = true;

    // Lock the camera to us.
    comm_ptr->start_write_reg(camera_ptr->ip_addr, GIGE_REG_CCP, GIGE_CCP_CA_BE);

    // Display the camera name in the info display.
    camera_info->display_camera_name();

    // Figure out which cursor to use.
    do_view_info_cursor();

    // Is the aiming cursor enabled?
    do_view_aiming_cursor();

    // Have the review controls update as appropriate.
    review_controls->update_graph();
}

/*!
 * @brief This method is used to set or clear the grouped status of the camera.
 * @param is_grouped    True if the camera is part of a multicamera group, false otherwise.
 */
void CameraMDI::set_grouped(bool is_grouped)
{
    camera_info->set_grouped(is_grouped);
}

/*!
 * @brief We are being made inactive. If we are actively streaming, we need to slow the rate down by
 *        switching to background mode. In addition, we do NOT want to respond to certain signals.
 */
void CameraMDI::set_inactive()
{
    // Disable the zoom actions.
    MainWindow::GetInstance()->enable_zoom_actions(false);

    // Disconnect the "active" actions we want.
    disconnect(MainWindow::GetInstance().get(), &MainWindow::config_black_level, this, &CameraMDI::do_black_level);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_arm, this, &CameraMDI::do_camera_arm);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_autosave_enable, this, &CameraMDI::do_camera_autosave_enable);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_autosave_setup, this, &CameraMDI::do_autosave_setup);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_live, this, &CameraMDI::do_camera_live);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_load, this, &CameraMDI::do_camera_load);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_menu_showing, this, &CameraMDI::update_camera_menu);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_power_down, this, &CameraMDI::do_camera_power_down);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_preferences, this, &CameraMDI::do_camera_preferences);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_reboot, this, &CameraMDI::do_camera_reboot);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_review, this, &CameraMDI::do_camera_review);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_samba, this, &CameraMDI::do_camera_samba);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_save, this, &CameraMDI::do_camera_save);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_save_all, this, &CameraMDI::do_camera_save_all);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_snapshot, this, &CameraMDI::do_camera_snapshot);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_trigger, this, &CameraMDI::do_camera_trigger);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_update, this, &CameraMDI::do_camera_update);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::camera_open_device, this, &CameraMDI::do_open_device);

    disconnect(MainWindow::GetInstance().get(), &MainWindow::config_factory_reset, this, &CameraMDI::do_config_factory_reset);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::config_load, this, &CameraMDI::do_config_load);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::config_save, this, &CameraMDI::do_config_save);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::config_time_date, this, &CameraMDI::do_config_time_date);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::config_wifi, this, &CameraMDI::do_config_wifi);

    disconnect(MainWindow::GetInstance().get(), &MainWindow::release_buffer, this, &CameraMDI::do_release_buffer);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::release_all_buffers, this, &CameraMDI::do_release_all_buffers);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::setup_event_markers, this, &CameraMDI::do_camera_events);

    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_aiming_cursor, this, &CameraMDI::do_view_aiming_cursor);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_info_cursor, this, &CameraMDI::do_view_info_cursor);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_pfmd, this, &CameraMDI::do_camera_pfmd);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_local_frametime, this, &CameraMDI::do_view_localtime);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_review_controls, this, &CameraMDI::do_view_review);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::view_summary, this, &CameraMDI::do_view_summary);

    if (video_player != nullptr)
        disconnect_video_player();

    disconnect(MainWindow::GetInstance().get(), &MainWindow::rotate_left, video_player, &CameraVideo::do_rotate_counter);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::rotate_right, video_player, &CameraVideo::do_rotate_clockwise);

    // One of these won't work - but that doesn't really matter. We just want to disconnect from the signal.
    disconnect(MainWindow::GetInstance().get(), &MainWindow::truncate_recording, this, &CameraMDI::do_truncate_clip);
    disconnect(MainWindow::GetInstance().get(), &MainWindow::truncate_recording, this, &CameraMDI::do_truncate_release);

    // Update the streamer so it can adjust the frame rate if need be.
    if (streamer)
        streamer->update(false);
}

/*!
 * @brief Install the FasFireLR (fflr) information for a device.
 *
 * @param fflr_label        The GIGE_REG_MEDIA_JSON value for the "label" key.
 * @param fflr_name         The GIGE_REG_MEDIA_JSON value for the "name" key.
 * @param fflr_mountdir     The GIGE_REG_MEDIA_JSON value for the "mountdir" key.
 * @param fflr_tag          The user displayed value built from the GIGE_REG_MEDIA_JSON values.
 * @param fflr_type         The GIGE_REG_MEDIA_JSON value for the "type" key. It is one of the GIGE_MEDIA_xxx types.
 * @param fflr_session_len  The number of 256 MiB chunks of device storage to reserve.
 *
 * @returns True if there were not errors or problems, false if they were not set.
 */
bool CameraMDI::set_fflr_info(const QString& fflr_label, const QString& fflr_name, const QString& fflr_mountdir, const QString& fflr_tag,
                              uint32_t fflr_type, uint32_t fflr_session_len)
{
    bool        result = true;
    QSettings   settings;

    this->fflr_device_label         = fflr_label;
    this->fflr_device_name          = fflr_name;
    this->fflr_device_mountdir      = fflr_mountdir;
    this->fflr_device_tag           = fflr_tag;
    this->fflr_device_type          = fflr_type;
    this->fflr_session_len          = fflr_session_len;

    // Persist these settings in our QSettings for [SaveSettings].
    settings.beginGroup("SaveSettings");
    settings.setValue("dest_fflr_device_label", fflr_label);
    settings.setValue("dest_fflr_device_name", fflr_name);
    settings.setValue("dest_fflr_device_mountdir", fflr_mountdir);
    settings.setValue("dest_fflr_device_tag", fflr_tag);
    settings.setValue("dest_fflr_device_type", fflr_type);
    settings.setValue("dest_fflr_session_len", (qint64)fflr_session_len);
    settings.endGroup();

    return result;
}

/*!
 * @brief We have told the camera to go offline - typically as a result of a reboot operation.
 */
void CameraMDI::set_offline(bool close_window)
{
    FI_comm     *comm_ptr;
    static QAtomicInt block(0);

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "close_window:" << close_window << "in_shutdown:" << in_shutdown;

    // Make sure we're not already here.
    if (!block.testAndSetRelaxed(0, 1))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Already here...bailing...";
        return;
    }

    // Are we already shutting down?
    if (in_shutdown)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Already in shutdown...bail";
        goto bail;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        goto bail;
    }

    // Tell the video player that we're going off-line.
    if (video_player != nullptr)
        video_player->stop_video();

    // Push the events through. We want to allow for queued up events from the
    // stopping of the video to flow through.
    //
    // WARNING: Don't allow ourselves to be re-entered! That is the reason for
    //          the "block" above.
    QApplication::processEvents();

    // Close down the stream.
    if (streamer != nullptr)
        streamer->detach();

    // Sanity check.
    if (camera_ptr != nullptr)
    {
        // Disconnect from the camera.
        comm_ptr->start_write_reg(camera_ptr->ip_addr, GIGE_REG_SCP0, 0);
        comm_ptr->start_write_reg(camera_ptr->ip_addr, GIGE_REG_CCP, 0);

        // Tell the communications object that we're offline.
        comm_ptr->set_offline(camera_ptr->ip_addr, true);

        // Disconnect the camera.
        comm_ptr->disconnect_cam(camera_ptr->ip_addr);

        // Mark the camera as being off-line.
        camera_ptr->offline = true;
        camera_ptr->offline_skip = CAMERA_DOCK::SCAN_SKIP_COUNT;
    }

    // Are we supposed to close the window?
    if (close_window)
    {
        // Close the window.
        in_shutdown = true;
        finish_shutdown();
    }

bail:
    // Allow entry again.
    block = 0;
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Done.";
}

/*!
 * @brief Update the record progress widget with the current state of the capture.
 * @param camera_state  The current camera state. We're particularly interested in
 *                      GIGE_CAMERA_STATE_ARMED and GIGE_CAMERA_STATE_TRIGGERED.
 */
void CameraMDI::set_rec_progress(uint32_t camera_state)
{
    uint32_t    cap_count;
    uint32_t    cur_frame;
    uint32_t    frame_count;
    uint32_t    trig_pos;
    uint32_t    ltr_record_type;
    bool        ltr_append_mode = false;

    // Process based on state.
    if ((camera_state != GIGE_CAMERA_STATE_ARMED) && (camera_state != GIGE_CAMERA_STATE_TRIGGERED))
        return;

    // Sanity check.
    if (rec_progress == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No rec_progress pointer?!";
        return;
    }

    // Read up the parameters from the camera.
    trig_pos = read_camera_cache(GIGE_REG_TRIGGER_REQ_POS);
    cur_frame = read_camera_cache(GIGE_REG_CUR_FRAME_POS);
    cap_count = read_camera_cache(GIGE_REG_CUR_CAPTURE_COUNT);

    // Check for LTR or FFLR capture mode.
    if (is_ltr_enabled(ltr_record_type) || is_fflr_enabled(ltr_record_type))
    {
        // Only applicable for BROC or ROC.
        ltr_append_mode = ((ltr_record_type & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0);
    }

    // All we have to do is call our new max frame count interface in our own object.
    frame_count = calc_max_frames();

    // Are we in LTR append mode?
    if (ltr_append_mode)
    {
        // Are we waiting for a trigger?
        if (camera_state == GIGE_CAMERA_STATE_ARMED)
        {
            // In "append" mode, we don't really care about the "trigger" position. It's always a (virtual) start trigger.
            trig_pos = cur_frame;
        }
        else
        {
            // For lack of a better value, once we've trigger'd set the position back to 0.
            trig_pos = 0;
        }
    }
    else
    {
        // Is the trigger position frame number or percent?
        bool frame_based = ((trig_pos & GIGE_TRIGGER_REQ_FRAME_MASK) == GIGE_TRIGGER_REQ_FRAME_MASK);
        if (frame_based)
        {
            // Get the actual trigger frame number.
            trig_pos &= GIGE_TRIGGER_REQ_POS_MASK;
        }
        else
        {
            // It's percent based, so calculate which frame it should be.
            trig_pos = (uint32_t)(((uint64_t)frame_count * (uint64_t)trig_pos) / 100UL);
        }

        // Are we waiting for a trigger?
        if (camera_state == GIGE_CAMERA_STATE_ARMED)
        {
            if (cap_count > trig_pos)
            {
                // Lock the current frame to the trigger position.
                cur_frame = trig_pos;
            }
            else
            {
                // Set the current frame to our capture count.
                cur_frame = cap_count;
            }
        }
        else if (camera_state == GIGE_CAMERA_STATE_TRIGGERED)
        {
            // The frame number we got from the camera is the frame PAST the trigger.
            // Adjust our frame number appropriately.
            cur_frame += trig_pos;
        }
    }

    // Sanity check.
    if (frame_count == 0)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "frame_count is zero?!";
        cur_frame = 0;
    }

    // Set the parameters in the control.
    rec_progress->set_num_frames((int)frame_count);
    rec_progress->set_trigger((int)trig_pos);
    rec_progress->set_current_frame((int)cur_frame);

    // Make sure we are sync'd with our capturing video state.
    rec_progress_state_check(true);
}

/*!
 * @brief Interface method to be told what our MDI Subwindow is. We need this to be able to hook into
 *        the sub-window's "aboutToActivate" signal.
 *
 * @param win_ptr   Our QMdiSubWindow "parent".
 */
void CameraMDI::set_window(QMdiSubWindow *win_ptr)
{
    // If the win_ptr passed is null there is nothing to disconnect.
    if (win_ptr)
    {
        connect(win_ptr, &QMdiSubWindow::aboutToActivate, this, &CameraMDI::window_activating);
    }
}

/*!
 * @brief Kick off the autosave process.
 */
void CameraMDI::start_autosave()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "autosave_pending:" << autosave_pending << "save_video->save_in_progress:" << save_video->save_in_progress();

    // Have we already started our save operation?
    if (autosave_pending || save_video->save_in_progress())
        return;

    // We've started our save operation.
    autosave_pending = true;

    // Reset our retry count.
    retry_count_raw_metadata = 0;

    // Make sure we tell everyone to stop streaming.
    MainWindow::GetInstance()->pause_video_streaming(this);

    // Make sure to instruct the camera_server to update its active partition
    // with the current metadata. It doesn't matter what the value for the
    // register is so any value will accomplish the task.
    write_camera_register(GIGE_REG_UPDATE_METADATA, 1);

    // Do we need to merge the overlay information?
    if (save_video && save_video->do_merge_overlay())
    {
        // Get the render information set up.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Calling render_start()";
        video_player->render_start();

        // Are we saving a video stream like AVI or MP4?
        if (FiSaver::isStreamVideo(save_video->getStreamType()))
        {
            QGraphicsScene  *scene_ptr;

            // Get a pointer to our "scene".
            scene_ptr = video_player->render_get_scene();
            if (scene_ptr == nullptr)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "scene_ptr is NULL?!";
                return;
            }

            // Adjust the AVI width/height value.
            save_video->setImageHeight((uint32_t)scene_ptr->sceneRect().height());
            save_video->setImageWidth((uint32_t)scene_ptr->sceneRect().width());
        }
    }
}

/*!
 * @brief Global method to have the streamer state machine start a copy operation.
 */
void CameraMDI::start_copy(QString& camera_path)
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Initialize some variables we'll need to manage the download.
    prepare_to_download();
    last_received_frame = save_video->get_copy_start();

    // Tell the stream state machine to start the copy.
    if (streamer)
        streamer->copy_from_camera(camera_path);

    // The streamer is running.
    camera_ptr->stream_running = true;
}

/*!
 * @brief Begin the process of saving an image capture to the local workstation.
 */
void CameraMDI::start_local_save(bool reset_retry_count)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Camera:" << camera_info->get_camera_name();

    // Don't start a download, if we're in the middle of a download.
    if (pm_download_in_progress)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Camera:" << camera_info->get_camera_name() << "PM download is already in progress";
        return;
    }

    // TICKET #441:  Log some warnings if these conditions are set.
    if (review_controls)
    {
        if (review_controls->is_playing())
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "The review_controls should not be playing right now!";
        }
    }

    // And our stream should be paused.
    if (streamer && !streamer->is_paused())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The streamer is not paused, is_playing(" << streamer->is_playing()
                    << ") is_stepping(" << streamer->is_stepping() << ") is_copying(" << streamer->is_copying();
    }

    // We're now configured and ready to save. Initialize our buffer.
    partition_metadata.clear();

    // Are we supposed to reset the retry counter?
    if (reset_retry_count)
    {
        // Clear our retry counter.
        retry_count_raw_metadata = 0;
    }

    // Have the camera re-acquire the metadata so we can download it.
    request_active_metadata(__PRETTY_FUNCTION__);
}

/*!
 * @brief Utility method to initiate a read operation of a buffer from our camera.
 * @param reg_addr      The register address we're interested in.
 * @param buf_len       The number of bytes we would like to read.
 * @param do_expire     True if we want to force the address to be read from the camera,
 *                      False if reading data from the cache would be OK.
 */
void CameraMDI::start_camera_read_buf(uint32_t buf_addr, uint32_t buf_len, bool do_expire)
{
    FI_comm     *comm_ptr;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "buf_addr:" << Qt::showbase << Qt::hex << buf_addr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // This isn't going to work if the camera is in calibration.
    if (!camera_ptr->offline)
    {
        // Do we want to force a read from the camera?
        if (do_expire)
        {
            // Because Fastec incorrectly implements "most" buffer reads, i.e. we
            // normally read from the same address like it was a FIFO, instead of
            // treating it like a range of memory, most of the time we want to expire
            // the cache so we don't get the first part of the buffer, when we really
            // want the second part (still sitting on the camera).
            comm_ptr->expire_entry(get_ip_addr(), buf_addr);
        }

        // Start the reading.
        comm_ptr->start_read_mem(get_ip_addr(), buf_addr, buf_len);
    }
}

/*!
 * @brief Utility method to initiate a read operation of a register from our camera.
 * @param reg_addr      The register address we're interested in.
 */
void CameraMDI::start_camera_read_reg(uint32_t reg_addr)
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // This isn't going to work if the camera is in calibration.
    if (!camera_ptr->offline)
    {
        comm_ptr->start_read_reg(get_ip_addr(), reg_addr);
    }
}

/*!
 * @brief Start the process of saving a session to the PC.
 */
void CameraMDI::start_session_save()
{
    CAMERA_STREAM::STREAM_PARAMS stream_info;

    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (is_series3())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Can't do this with a series3 camera!";
        return;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Initialize some variables we'll need to manage the download.
    prepare_to_download();

    // Reset the download statistics and start the timer going.
    save_video->reset_stats();

    // Clear our parameter structure.
    memset(&stream_info, 0, sizeof(stream_info));

    // Session saves are only concerned with clip start and end.
    stream_info.clip_end = save_video->getClipEnd();
    stream_info.clip_start = save_video->getClipStart();

    // Tell the stream state machine to start the copy.
    if (streamer)
        streamer->save_session(stream_info);

    // The streamer is running.
    camera_ptr->stream_running = true;
}

/*!
 * @brief Shut down our lock timer. This is the timer we use to periodically refresh
 *        the GIGE_REG_CCP register so we retain ownership of the camera.
 */
void CameraMDI::stop_lock_timer()
{
    if (lock_timer_id != 0)
    {
        killTimer(lock_timer_id);
        lock_timer_id = 0;
    }
}

/*!
 * @brief Public method to initiate a download/save of the raw per-frame metadata.
 */
void CameraMDI::stream_metadata()
{
    int         num_frames;
    uint32_t    camera_state;
    CAMERA_STREAM::STREAM_PARAMS    stream_info;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "retry_count_raw_metadata:" << retry_count_raw_metadata << "pm_download_in_progress:" << pm_download_in_progress;

    // Hold off if there is a metadata download in progress.
    if (pm_download_in_progress)
    {
        QTimer::singleShot(METADATA_STREAM_DELAY, this, &CameraMDI::stream_metadata);
        return;
    }

    // Initialize some variables we'll need to manage the download.
    prepare_to_download();

    // Our state of retry in this pathway may be different.
    retry_in_progress = (retry_count_raw_metadata != 0);

    // Clear our parameters structure.
    memset(&stream_info, 0, sizeof(stream_info));

    // We haven't received any as of yet.
    raw_metadata.clear();
    raw_metadata_len = 0;

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No save_video?";
        return;
    }

    // Retrieve the register value.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "retry:" << retry_in_progress << "camera_state:" << camera_state;

    // Only read the metadata if the camera is in review or autosave.
    if ((camera_state != GIGE_CAMERA_STATE_REVIEW) && (camera_state != GIGE_CAMERA_STATE_AUTOSAVE))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Wrong camera state:" << camera_state << "=========================================";

        // Is there a transfer dialog?
        if (xferDialog)
        {
            // Recycle any resources and release the progress dialog.
            fiProgressDialog::RecycleProgressDialog(xferDialog);
        }

        // Stop this silliness.
        raw_metadata_wrapup();

        // All done.
        return;
    }

    // Are we saving or processing the data?
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "downloading_pfmd:" << downloading_pfmd;
    if (downloading_pfmd)
    {
        // Setup our parameters.
        stream_info.cur_frame = 0;
        stream_info.clip_start = 0;
        stream_info.clip_end = read_camera_cache(GIGE_REG_PART_NUM_FRAMES) - 1;
    }
    else
    {
        // Use the saving parameters.
        stream_info.cur_frame = save_video->getClipStart();
        stream_info.clip_start = save_video->getClipStart();
        stream_info.clip_end = (save_video->getClipStart() + save_video->getFrameCount()) - 1;
    }

    // Saving, so no "desired" fps. The AVI playback FPS is different. That will go into the AVI container.
    stream_info.desired_fps = 0;

    // We're downloading, so no skipping.
    stream_info.skip_interval = 1;

    // Saving, so no frame period. It's a playback thing.
    stream_info.frame_period = 0;

    // When downloading, we don't need any delay.
    stream_info.frame_delay = 0;

    // How many "frames" can we expect to receive?
    num_frames = (int)(stream_info.clip_end - stream_info.clip_start) + 1;

    // How much data does that work out to be?
    raw_metadata_len = num_frames * get_pfmd_len();

    // Sanity check. If we have too many frames, we'll overflow the variable and make it go negative.
    if (raw_metadata_len < 0)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Per-frame metadata is too big! Frame count:" << num_frames;
        return;
    }

    // Is there a transfer dialog?
    if (xferDialog)
    {
        // Are we in a retry operation?
        if (retry_in_progress)
        {
            // Update the transfer dialog interface.
            xferDialog->setStatusValue(tr("Errors occurred. Retrying download [%1 of %2].").arg(retry_count_raw_metadata).arg(MAX_RETRIES_RAW_METADATA));
        }
        else
        {
            // Make sure we're connected to the cancel signal, but once is enough.
            connect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &CameraMDI::cancel_save, Qt::UniqueConnection);

            // Update the transfer dialog interface. We can actually provide updates if we're not
            // a series3 camera.
            xferDialog->setBarValue(0);
            if (is_series3())
                xferDialog->setBarMinMax(0, 0);
            else
                xferDialog->setBarMinMax(0, num_frames);
            xferDialog->setStatusValue(tr("Downloading per-frame metadata"));
        }
    }
    // Do we have a progress dialog up?
    else if (progressDialog)
    {
        // Modify the progress dialog.
        progressDialog->setStatusValue(tr("Errors occurred. Retrying download [%1 of %2].").arg(retry_count_raw_metadata).arg(MAX_RETRIES_RAW_METADATA));
    }
    else
    {
        // Create and show a simple progress dialog for calibration.
        progressDialog = fiProgressDialog::CreateSimpleProgressDialog(get_camera_name(), tr("Downloading per-frame metadata"), QMessageBox::Cancel, __PRETTY_FUNCTION__, this);

        // Display the progress dialog.
        if (progressDialog)
        {
            progressDialog->setCancelText(tr("Cancel Download"));
            progressDialog->setBarMinMax(0, 0);
            progressDialog->showBar(true);

            // The perframe metadata is kept separate in non-series 3 cameras.
            if (!is_series3())
            {
                progressDialog->setBarMinMax(0, num_frames);
            }

            connect(progressDialog.get(), &fiProgressDialog::progressCancelled, this, &CameraMDI::cancel_save, Qt::UniqueConnection);

            progressDialog->showDialogWithoutPolling();
        }
    }

    // Tell the stream state machine to start the playback.
    if (streamer)
        streamer->get_raw_metadata(stream_info, retry_in_progress);

    // The streamer is running.
    if (camera_ptr)
        camera_ptr->stream_running = true;
}

/*!
 * @brief Public method to initiate a download/save of a recorded stream to the local workstation.
 *
 * @param clip_start        The ABSOLUTE (and inclusive) starting frame number of the video to be saved.
 * @param clip_end          The ABSOLUTE (and inclusive) ending frame number of the video to be saved.
 * @param save_format       The format for the saved image(s).
 */
void CameraMDI::stream_save_video(uint32_t clip_start, uint32_t clip_end,
                                  FI_SAVER::STREAM_TYPE streamType, FI_SAVER::IMAGE_TYPE imageType)
{
    CAMERA_STREAM::STREAM_PARAMS    stream_info;

    uint32_t gigeStreamFormat = 0;  // Will hold the camera's GIGE stream format value.

    // Initialize some variables we'll need to manage the download.
    trigger_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

    // Prepare to download a video stream.
    prepare_to_download();

    // Clear our parameter structure.
    memset(&stream_info, 0, sizeof(stream_info));

    // Get the "new" current frame.
    stream_info.cur_frame = clip_start;
    stream_info.clip_start = clip_start;
    stream_info.clip_end = clip_end;

    // Saving, so no "desired" fps. The AVI playback FPS is different. That will go into the AVI container.
    stream_info.desired_fps = 0;

    // We're downloading, so no skipping.
    stream_info.skip_interval = 1;

    // Saving, so no frame period. It's a playback thing.
    stream_info.frame_period = 0;

    // When downloading, we don't need any delay.
    stream_info.frame_delay = 0;

    // The stream_info data structure communicates with the camera what to start
    // streaming.  It is expecting an GIGE_STREAM_FORMAT_* value.  If a streamType
    // is video such as an AVI or MP4 stream then the camera is interested in the
    // GIGE_STREAM_FORMAT_* value for the type of images the video stream writer
    // will be expecting.
    if (FiSaver::isStreamVideo(streamType))
    {
        // #MP4IMAGEINPUT - CameraMDI::stream_save_video calling fiSaver::ImageTypeToCameraImageType
        // The old code was passing in the avi_image_format which in the old model
        // was either GIGE_STREAM_FORMAT_BMP or GIGE_STREAM_FORMAT_JPEG.  In the
        // new model stream, video can be AVI or MP4 and there will be additional
        // image types possible - right now only 2 are supported in camera but it
        // will need to be updated.
        gigeStreamFormat = FiSaver::imageTypeToCameraImageType(streamType, imageType, is_series3());
    }
    else    // A Stack of images is being saved.
    {
        gigeStreamFormat = FiSaver::streamTypeToCameraStreamType(streamType, is_series3());
    }

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Asking CameraServer to STREAM: "
                 << FiSaver::getGIGEStreamStr(gigeStreamFormat)
                 << FiSaver::streamTypeToQString(streamType)
                 << FiSaver::imageTypeToQString(imageType);

    stream_info.save_format = gigeStreamFormat;

    // Tell the stream state machine to start the playback.
    if (streamer)
        streamer->save_video(stream_info);

    // Install trigger frame value into the save video apparatus to report relative frame #s.
    save_video->setTrigNo4Frame(trigger_frame);

    // We are interested in any user cancellations of the save process to know when to cancel a save operation.
    connect(save_video, &SaveVideo::userCancelledSaveSignal, this, &CameraMDI::cancel_save, Qt::QueuedConnection);

    // The streamer is running.
    if (camera_ptr)
        camera_ptr->stream_running = true;
}

void CameraMDI::ae_reticle_show(bool show_reticle)
{
    // Sanity check.
    if (video_player == nullptr)
        return;

    // Enable/disable the reticle.
    video_player->show_ae_reticle(show_reticle);
}

void CameraMDI::ae_reticle_reset()
{
    if (video_player == nullptr)
        return;
    video_player->reset_ae_reticle();
}

bool CameraMDI::ae_reticle_visible()
{
    if (video_player == nullptr)
        return false;
    return video_player->is_ae_reticle_visible();
}

void CameraMDI::track_exposure(bool enable_track)
{
    bool old_enable;

    // Sanity check.
    if (video_player == nullptr)
        return;

    // Allow the reticle to be moved if we are NOT tracking. I apologize
    // for the inverted logic here, but the video_player's POV is that
    // the ae_reticle is a widget. The CameraMDI's POV is that this is
    // auto exposure functionality.
    old_enable = video_player->is_ae_reticle_enabled();
    video_player->enable_ae_reticle(!enable_track);

    // If we're tracking, turn on the timer. Otherwise,
    // turn it off.
    if (enable_track)
    {
        // Make sure we start at the right place in the state machine.
        ae_track_state = AE_INIT_TRACKING;

        // Do an immediate update.
        ae_track();
    }

    // Only disable if it was previously enabled.
    else if (old_enable)
    {
        QRect empty_rect;

        // Disconnect our histogram ready signal.
        disconnect(video_player->get_histogram(), &fiHistoProc::ready, this, &CameraMDI::ae_track);

        // Restore the histogram.
        MainWindow::GetInstance()->get_histo_dock()->override_roi(empty_rect);

        // Disable the histogram engine.
        video_player->enable_histogram(false);
    }
}

bool CameraMDI::tracking_exposure()
{
    if (video_player == nullptr)
        return false;

    // If the reticle is disabled, we are tracking.
    return !video_player->is_ae_reticle_enabled();
}

void CameraMDI::vt_baseline_calc()
{
    int i;
    const QVector<int> *lum;

    // Sanity check.
    if (vt_ready)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "vt_ready is set, why are we here?";
        vt_track_state = VT_INIT_TRACKING;
        return;
    }

    // Get our luminance histogram vector.
    lum = video_player->get_histogram()->get_lum_map();
    for (i = 0; i < lum->size(); i++)
    {
        // Accumulate into our baseline.
        vt_baseline[i] += lum->at(i);
    }

    // Increment the number of histograms we've accumulated.
    ++vt_histo_count;

    // Are we done?
    if (vt_histo_count == VT_BASELINE_COUNT)
    {
        // Make the baseline vector an AVERAGE of the readings.
        for (i = 0; i < vt_baseline.size(); i++)
            vt_baseline[i] = vt_baseline.at(i) / VT_BASELINE_COUNT;

        // We are now ready to begin processing.
        vt_ready = true;

        // Reset our counter.
        vt_histo_count = 0;

        // Acquire the next histogram.
        vt_track_state = VT_PROCESS;

#ifdef DEBUG_DUMP
        {
            FILE *fp;
            fp = fopen("vt_sample.csv", "w");
            if (fp == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open vt_sample.csv for writing";
            }
            else
            {
                for (i = 0; i < vt_baseline.size(); i++)
                    fprintf(fp, "%u,", vt_baseline.at(i));
                fprintf(fp, "\n");
                fclose(fp);
            }
        }
#endif /* DEBUG_DUMP */
    }
}

void CameraMDI::vt_process()
{
    int i;
    int area_threshold;
    int value_threshold;
    int trigger_count;
    bool triggered;
    uint32_t cam_state;
    uint32_t ltr_record_mode;
    bool ltr_roc_mode = false;
    QPen enabled_pen;
    QRect roi;
    QVector<int> abs_diff;
    const QVector<int> *lum;

    // Sanity check.
    if (!vt_ready)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "vt_ready is NOT set! Why are we here?!!";
        vt_track_state = VT_INIT_TRACKING;
        return;
    }

    // Get the histogram values.
    lum = video_player->get_histogram()->get_lum_map();

    // Initialize our values.
    trigger_count = 0;
    value_threshold = (256 * vt_change) / 100;
    roi = video_player->get_vt_rect();
    area_threshold = (roi.height() * roi.width() * vt_area) / 100;

    abs_diff.resize(vt_baseline.size());
    abs_diff.fill(0);

    // See if we have enough for a trigger.
    for (i = 0; i < vt_baseline.size(); i++)
    {
        // Calculate the difference between the histogram and the baseline.
        abs_diff[i] = abs(lum->at(i) - vt_baseline[i]);

        // Do we exceed the trigger threshold?
        if (abs_diff[i] >= value_threshold)
        {
            // Accumulate this pixel that was different.
            trigger_count++;
        }
        else
        {
            // Zero this entry.
            abs_diff[i] = 0;
        }
    }

    // You can do this more efficiently, but I wanted to be very clear what I was after:
    // LTR ROC mode enabled.
    if (is_ltr_enabled(ltr_record_mode) || is_fflr_enabled(ltr_record_mode))
    {
        ltr_roc_mode = ((ltr_record_mode & GIGE_LTR_MODE_ROC) == GIGE_LTR_MODE_ROC);
    }

    // Get the camera state.
    cam_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    if ((cam_state == GIGE_CAMERA_STATE_REVIEW) && !vt_software())
    {
        // If we're in review, and NOT running the software VT algorithm, then we want to look at
        // what the hardware was thinking.
        triggered = ((camera_info->get_vt_flags() & CAMERA_INFO::VTRIG_FLAGS_TRIGGER) == CAMERA_INFO::VTRIG_FLAGS_TRIGGER);
    }
    else
    {
        // Not in review or running software, so only pay attention to the software VT algorithm.
        triggered = (trigger_count >= area_threshold);
    }

    // Did enough pixels change, or did the hardware indicate a trigger?
    if (triggered)
    {
#ifdef DEBUG_DUMP
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "roi:" << roi << "vt_area:" << vt_area << "area_threshold:" << area_threshold;
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "trigger_count:" << trigger_count;
        {
            FILE *fp;
            fp = fopen("vt_sample.csv", "a");
            if (fp == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open vt_sample.csv for append";
            }
            else
            {
                for (i = 0; i < lum->size(); i++)
                    fprintf(fp, "%d,", lum->at(i));
                fprintf(fp, "\n");
                for (i = 0; i < abs_diff.size(); i++)
                    fprintf(fp, "%d,", abs_diff.at(i));
                fprintf(fp, "\n");
                fclose(fp);
            }
        }
#endif /* DEBUG_DUMP */

        // Got a trigger.
        enabled_pen = QPen(Qt::red);
        enabled_pen.setWidth(3);
        video_player->get_vt_reticle()->setEnabledPen(enabled_pen);
        video_player->get_vt_reticle()->setEnabledBrush(QBrush(QColor(0xff, 0x00, 0x00, 0x40), Qt::SolidPattern));

        // Are we triggering? Only if we software triggering and in the ARMED state.
        if (vt_software() && (cam_state == GIGE_CAMERA_STATE_ARMED))
            do_camera_trigger();
    }
    else
    {
        // All is well, so use green.
        enabled_pen = QPen(Qt::green);
        enabled_pen.setWidth(3);
        video_player->get_vt_reticle()->setEnabledPen(enabled_pen);
        video_player->get_vt_reticle()->setEnabledBrush(QBrush(QColor(0x00, 0xff, 0x00, 0x40), Qt::SolidPattern));

        // If we are in ROC and we have triggered, then issue another trigger to
        // finish up the ROC recording.
        if (ltr_roc_mode && (cam_state == GIGE_CAMERA_STATE_TRIGGERED))
            do_camera_trigger();
    }

    // Do we need to re-baseline?
    ++vt_histo_count;
    if (vt_histo_count == vt_refresh)
    {
        // Reset our tracking variables.
        vt_ready = false;
        vt_histo_count = 0;

        // Reset our baseline.
        vt_baseline.fill(0);

        // Make sure our reticle is setup properly.
        video_player->get_vt_reticle()->setEnabledPen(QPen(Qt::white, 3));
//        video_player->get_vt_reticle()->setEnabledBrush(QBrush(Qt::NoBrush));

        // Acquire the next histogram.
        vt_track_state = VT_INIT_BASELINE;
    }
}

void CameraMDI::vt_reticle_show(bool show_reticle)
{
    bool was_visible;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "show_reticle:" << show_reticle;

    // Sanity check.
    if (video_player == nullptr)
        return;

    // Is the reticle currently visible?
    was_visible = video_player->is_vt_reticle_visible();

    // Enable/disable the reticle.
    video_player->show_vt_reticle(show_reticle);
    video_player->enable_vt_reticle(show_reticle);

    // If we're tracking, turn on the timer. Otherwise,
    // turn it off.
    if (show_reticle)
    {
        // If we are to show the reticle, but it was already showing, then
        // we are done.
        if (was_visible)
            return;

        // Reset our settings.
        vt_ready = false;
        vt_histo_count = 0;

        // Make sure we start at the right place in the state machine.
        vt_track_state = VT_INIT_TRACKING;

        // Do an immediate update.
        vt_track();

        // Connect our signals/slots.
        connect(video_player, &CameraVideo::videoRectChanged, this, &CameraMDI::vt_changed);

        // Are we connect to a camera that supports video trigger?
        if (sensor_ptr && sensor_ptr->supports_video_trigger())
        {
            QRect trig_rect;
            QJsonObject json_obj;

            // Get our area of interest.
            trig_rect = video_player->get_vt_rect();
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "trig_rect:" << trig_rect;

            // Make sure there is a valid reticle.
            if (trig_rect.isEmpty())
            {
                // Reset the reticle to defaults.
                vt_reticle_reset();

                // Pull our area of interest after reset.
                trig_rect = video_player->get_vt_rect();

                // Make sure there is a valid reticle.
                if (trig_rect.isEmpty())
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to pull valid vt_rect after vt_reticle_reset().";
                    return;
                }
            }

            // Make absolute certain that the trigger area is entirely within the
            // image area.
            if (video_player->get_image_rect().contains(trig_rect))
            {
                int control_setting;

                // Pass on the new information.
                control_setting = GIGE_VTRIG_CONTROL_ENABLE;
                if (!vt_mark_only)
                    control_setting |= GIGE_VTRIG_CONTROL_TRIGGER;

                // Build our JSON value and send it on to the camera.
                json_obj.insert("control", control_setting);
                json_obj.insert("roi_x", trig_rect.left());
                json_obj.insert("roi_y", trig_rect.top());
                json_obj.insert("roi_width", trig_rect.width());
                json_obj.insert("roi_height", trig_rect.height());
                json_obj.insert("pix_thresh", vt_change);
                json_obj.insert("area_thresh", vt_area);
                json_obj.insert("interval", vt_refresh);

                QJsonDocument ap_doc(json_obj);
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << ap_doc;
                write_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON, ap_doc.toJson());
            }
            else
            {
                video_player->get_vt_reticle()->setEnabledBrush(QBrush(QColor(0xff, 0x00, 0xff, 0x40), Qt::CrossPattern));
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "trig_rect:" << trig_rect << "is not contained in the image:" << video_player->get_image_rect();
            }
        }
    }

    // Only turn it off if we're visible.
    else if (was_visible)
    {
        QRect empty_rect;

        // Disconnect our histogram ready signal.
        disconnect(video_player->get_histogram(), &fiHistoProc::ready, this, &CameraMDI::vt_track);
        disconnect(video_player, &CameraVideo::videoRectChanged, this, &CameraMDI::vt_changed);

        // Restore the histogram.
        MainWindow::GetInstance()->get_histo_dock()->override_roi(empty_rect);

        // Disable the histogram engine.
        video_player->enable_histogram(false);

        // Are we connect to a camera that supports video trigger?
        if (sensor_ptr && sensor_ptr->supports_video_trigger())
        {
            QJsonObject json_obj;

            // Disable the hardware.
            json_obj.insert("control", 0);

            QJsonDocument ap_doc(json_obj);
            write_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON, ap_doc.toJson());
        }
    }
}

void CameraMDI::set_vt_region(const QRect& new_rect)
{
    // Get the reticle.
    if (video_player == nullptr)
        return;

    // Update the reticle's region.
    video_player->reset_vt_reticle(new_rect);
}

void CameraMDI::vt_reticle_reset()
{
    if (video_player == nullptr)
        return;
    video_player->reset_vt_reticle();

    // Toggle the reticle logic to reset things.
    if (vt_reticle_visible())
    {
        vt_reticle_show(false);
        vt_reticle_show(true);
    }
}

bool CameraMDI::vt_reticle_visible()
{
    if (video_player == nullptr)
        return false;
    return video_player->is_vt_reticle_visible();
}

bool CameraMDI::vt_software()
{
    bool do_soft_trigger;

    // Assume we'll issue a trigger.
    do_soft_trigger = true;

    // Does the camera support hardware video trigger?
    if (sensor_ptr && sensor_ptr->supports_video_trigger())
    {
        uint32_t ltr_record_mode;

        // You can do this more efficiently, but I wanted to be very clear what I was after:
        // Determine if LTR or FFLR are enabled.
        if (is_fflr_enabled(ltr_record_mode) || is_ltr_enabled(ltr_record_mode))
        {
            // Even if we have hardware support, it doesn't work in LR ROC/BROC mode.
            // It does work in LR basic mode though, so make sure to differentiate.
            // And basic mode is set if ltr_record_mode returned from the conditional test
            // above is zero.
            if (ltr_record_mode == 0)
                do_soft_trigger = false;
        }
        else
        {
            // Normal high speed record, we can use hardware video trigger.
            do_soft_trigger = false;
        }
    }

    // Return if we need to soft trigger or if the hardware will take care of things.
    return do_soft_trigger;
}

/*!
 * @brief Setup the streaming interface from the camera.
 */
void CameraMDI::stream_setup(uint32_t new_state, const QString& whoCalledMe)
{
    MultiCamDock    *dock_ptr;

    Q_UNUSED(whoCalledMe)
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString("caller[%1] camera_state[%2]").arg(whoCalledMe).arg(MainWindow::camera_state_to_string(new_state));

    // Sanity checks.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (camera_info == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "CAMERA_INFO_SANITY_FAILURE";
        return;
    }
    if (streamer == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "STREAMER_SANITY_FAILURE";
        return;
    }

    // Does the user want to stream?
    if (!camera_ptr->stream)
        return;

    // Get a pointer to the multicam dock.
    dock_ptr = MainWindow::GetInstance()->get_multicam_dock();
    if (dock_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No multicam dock!?";
        return;
    }

    // Check the camera state to determine how to proceed.
    switch (new_state)
    {
    case GIGE_CAMERA_STATE_ARMED:
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        // If we're busy, just stay in a paused state.
        if (is_busy())
        {
            // Make sure we disable the stream button.
            camera_info->enable_stream(false);

            // Ticket #342 Test - only pause when saving locally.
            if (save_video->local_autosave())
            {
                pause_video_request(__PRETTY_FUNCTION__);
            }

            // Try again in a little bit.
            QTimer::singleShot(BUSY_POLL_TIMEOUT, this, &CameraMDI::check_for_busy);
        }

        // Are we grouped?
        else if (!dock_ptr->is_camera_grouped(this) || dock_ptr->streaming_allowed())
        {
            // If the saver is cancelling, then we do not want to go live.
            if ((save_video != nullptr) && !save_video->cancelling())
            {
                // Start the live stream.
                framerate_changed(camera_info->get_framerate());
                camera_info->enable_stream(true);
                streamer->go_live();
            }
        }
        break;
    case GIGE_CAMERA_STATE_REVIEW:
    case GIGE_CAMERA_STATE_IDLE:
        // The stream button has no meaning in this context.
        camera_info->enable_stream(false);
        pause_video_request(__PRETTY_FUNCTION__);
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
    case GIGE_CAMERA_STATE_SNAPSHOT:
    case GIGE_CAMERA_STATE_SHUTDOWN:
        // The stream button has no meaning in this context.
        camera_info->enable_stream(false);
        break;
    case GIGE_CAMERA_STATE_TRIGGERED:
    default:
        break;
    }
}

/*!
 * @brief   If we are in the REVIEW state, make sure we have the per-frame metadata
 *          downloaded and parsed.
 */
void CameraMDI::update_event_info()
{
    uint32_t    camera_state;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__  << "retry_count_raw_metadata:" << retry_count_raw_metadata;

    // Initially assume we're not downloading the per-frame metadata.
    downloading_pfmd = false;

    // Make sure we're supposed to be streaming!
    if ((camera_ptr == nullptr) || !camera_ptr->stream)
    {
        if (camera_info == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << CAMERA_INFO_SANITY_FAILURE;
            return;
        }
        // Make sure our action panel is enabled so the user can switch to other camera states,
        // even if we couldn't download the metadata.
        camera_info->enable_actions();
        return;
    }

    // Are we in autoplay?
    if (autoplay())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Don't do event processing in autoplay.";

        // Make sure our action panel is enabled so the user can switch to other camera states,
        // even if we couldn't download the metadata.
        camera_info->enable_actions();

        // Mark us as being complete.
        emit event_search_complete();
        return;
    }

    // Get the current camera state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Only do this if we're in review.
    if (camera_state != GIGE_CAMERA_STATE_REVIEW)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Bad state?" << Qt::showbase << Qt::hex << camera_state;
        emit event_search_complete();
        return;
    }

    // We need to initialize our trigger frame.
    trigger_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);

    // Do we already have the metadata?
    if (!raw_metadata.isEmpty())
    {
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "download done, retry_count_raw_metadata:" << retry_count_raw_metadata;

        // Sanity check the raw metadata we've downloaded.
        if (!raw_metadata_ok())
        {
            QString text_msg;

            // Clear the buffer.
            raw_metadata.clear();

            // Build the error message.
            text_msg = tr("Raw metadata download failure: Unable to successfully transfer the entire amount of data.");

            // Notify the user.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << text_msg;
            fiMsgBox::critical(this, text_msg);

            // Mark us as being complete.
            emit event_search_complete();
            return;
        }

        // Did we turn on the IO graph display?
        if (review_controls->graph_is_enabled())
        {
            // We need to build up the data, 'cuz we just turning on the graph.
            build_data();
        }

        // Only start searching if we're interested in event markers.
        if (use_event_markers)
        {
            search_start_index = 0;
            search_for_events();
        }
    }

    else if (retry_count_raw_metadata >= MAX_RETRIES_RAW_METADATA)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Metadata download retry exceeded.";

        // Reset our counter.
        retry_count_raw_metadata = 0;

        // Mark us as being complete.
        emit event_search_complete();
    }

    else
    {
        // Remember that we're downloading the metadata not to save, but to use.
        downloading_pfmd = true;

        // Connect our slot to catch the download complete signal.
        connect(this, &CameraMDI::raw_metadata_received, this, &CameraMDI::raw_metadata_ready, Qt::UniqueConnection);

        // Start downloading the metadata. We need to delay a short period to
        // make sure the camera is ready for us.
        QTimer::singleShot(METADATA_STREAM_DELAY, this, &CameraMDI::stream_metadata);
    }
}

/*!
 * @brief This method is used to request that the current frame be resent from
 *        the camera.
 */
void CameraMDI::update_frame()
{
    uint32_t    cur_frame;
    uint32_t    camera_state;

    // Get the current camera state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // We need to be in review, otherwise there's no need to fetch another frame.
    if (camera_state != GIGE_CAMERA_STATE_REVIEW)
        return;

    // If we're already in a "playback" state, then we don't need to do this.
    if (streamer && streamer->is_playing())
        return;
    if (autoplay() && review_controls->is_playing())
        return;

    // Make sure we force the update.
    force_frame_update = true;

    // Request that the current frame be re-sent.
    cur_frame = review_controls->get_current_frame_num();
#define NO_FSP_RESEND
#ifdef NO_FSP_RESEND
    QLOG_INFO() << __PRETTY_FUNCTION__ << "<<<<<<< NEED FSP RESEND >>>>>>>>";

    uint32_t    dither_frame;
    if (cur_frame == review_controls->get_clip_end())
        dither_frame = cur_frame - 1;
    else
        dither_frame = cur_frame + 1;

    // Ask the state machine to stream these frames.
    streamer->request_frame(dither_frame);
#endif
    review_controls->set_current_frame_num(cur_frame);
}

void CameraMDI::update_media()
{
    // If we're not series3, then we want to get the media information as a JSON
    // formatted string.
    if (!is_series3())
    {
        // Reset our string.
        json_build_str.clear();

        // Before we ask the camera to build up our JSON media string, have
        // it reset.
        write_camera_register(GIGE_REG_MEDIA_JSON, 0);
    }
}

/*!
 * @brief During a recording, while the camera is in the @ref GIGE_CAMERA_STATE_ARMED or @ref GIGE_CAMERA_STATE_TRIGGERED
 *        state, we want to show the [recording progress widget](@ref CameraMDI::rec_progress). If the camera isn't in
 *        either of those states, we want to hide the progress widget.
 */
void CameraMDI::update_rec_info()
{
    uint32_t    camera_state;

    // Get the current camera state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Process based on state.
    // If we're armed or triggered, we want to update the recording progress bar.
    if ((camera_state == GIGE_CAMERA_STATE_ARMED) || (camera_state == GIGE_CAMERA_STATE_TRIGGERED))
    {
        // Set the recording progress information.
        set_rec_progress(camera_state);
    }
    // If back to live video or review states reset the progress bar.
    else if ((camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO) || (camera_state == GIGE_CAMERA_STATE_REVIEW))
    {
        // Sync up with the non-capturing video state.
        rec_progress_state_check(false);
    }
}

/*!
 * @brief During an save session, we want to update the progress indicators (one for the number of buffers
 *        and one for the current save progress) found in the [CameraInfo](@ref CameraInfo) widget. This method
 *        updates the save progress.
 *
 * @param progress      Percent complete for the save progress.
 */
void CameraMDI::update_save_progress(uint32_t progress)
{
    // Sanity check.
    if (camera_info != nullptr)
        camera_info->display_save_progress(progress);
}

/*!
 * @brief Update the status bar's info area with pertinent data about the camera settings.
 */
void CameraMDI::update_status()
{
    uint32_t    width;
    uint32_t    height;
    uint32_t    frame_rate;
    uint32_t    gain;
    uint32_t    ss;
    uint32_t    trig_pos;
    bool        enable_update;
    bool        grouped(false);
    MultiCamDock    *dock_ptr;
    QVersionNumber  cam_version;
    const FiUpdater *updater;

    // Get our update object.
    updater = MainWindow::GetInstance()->get_updater();

    // Sanity checks.
    if (camera_info == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CAMERA_INFO_SANITY_FAILURE;
        return;
    }
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get the values we'll need for the status line.
    width = read_camera_cache(GIGE_REG_WIDTH);
    height = read_camera_cache(GIGE_REG_HEIGHT);
    frame_rate = read_camera_cache(GIGE_REG_FRAME_RATE);
    ss = read_camera_cache(GIGE_REG_SHUTTER_SPEED);
    trig_pos = read_camera_cache(GIGE_REG_TRIGGER_REQ_POS);
    gain = read_camera_cache(GIGE_REG_SENSOR_GAIN);

    // Pass the values on.
    camera_info->set_summary(width, height, frame_rate, ss, trig_pos, gain);

    // See if we're in the group.
    dock_ptr = MainWindow::GetInstance()->get_multicam_dock();
    if (dock_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No multicam dock?!";
    }
    else
    {
        grouped = dock_ptr->is_camera_grouped(this);
    }
    camera_info->set_grouped(grouped);

    // Is there an update available for this camera?
    cam_version = QVersionNumber::fromString(camera_ptr->dev_version);

    // Are there updates available? Assume no updates available.
    enable_update = false;
    if (updater->GetUpdates() != FI_UPDATER::AVAIL_NONE)
    {
        // Are we series3?
        if (is_series3())
        {
            // Do we have an update available?
            enable_update = (QVersionNumber::compare(updater->GetVersionCamS3(), cam_version) > 0);
            if (enable_update)
            {
                // We don't allow automatic updates for cameras prior to version 2.0.0.
                if (QVersionNumber::compare(cam_version, QVersionNumber(2, 0, 0)) < 0)
                    enable_update = false;
            }
        }
        else
        {
            // Do we have an update available?
            enable_update = (QVersionNumber::compare(updater->GetVersionCamNC(), cam_version) > 0);
        }
    }

    // If the factory menu is enabled, we always allow an update. This is to facilitate
    // regressing an update, when an update is less than the current version.
    if (MainWindow::GetInstance()->is_factory_enabled())
        enable_update = true;

    // The camera must be in a state of live view in order to enable update.
    if (enable_update)
    {
        // Make sure the camera is in LIVE mode.  If not we won't enable the update.
        if (read_camera_cache(GIGE_REG_CAMERA_STATE) != GIGE_CAMERA_STATE_LIVE_VIDEO)
            enable_update = false;
    }

    // Do we have an update available?
    camera_info->enable_update(enable_update);
    MainWindow::GetInstance()->get_cam_update()->setEnabled(enable_update);
}

/*!
 * @brief Process the GIGE_REG_WIFI_STATUS register value to determine
 *        if we should enable/disable the WIFI configuration menu.
 */
void CameraMDI::update_wifi_status()
{
    // We only care about this if we're the active window.
    if (active_window())
    {
        // Read up our status
        uint32_t reg_value = read_camera_cache(GIGE_REG_WIFI_STATUS);

        // Update our menu state.
        MainWindow::GetInstance()->get_config_wifi()->setEnabled((reg_value & GIGE_WIFI_STATUS_DONGLE) == GIGE_WIFI_STATUS_DONGLE);
    }
}

/*!
 * @brief Generate a user friendly name for the window.
 * @return A QString with the nice name in it.
 */
QString CameraMDI::user_friendly_name()
{
    QString     user_name;

    // Sanity check.
    if (camera_ptr == nullptr)
        return QString::fromUtf8("");

    // Get the IP address in a form we can use.
    QHostAddress ip_addr(qFromBigEndian(camera_ptr->ip_addr));

    // Build up our friendly string.
    user_name = QString::fromUtf8("%1 [%2]").arg(ip_addr.toString(), get_mac());

    // Return our string.
    return user_name;
}

/*!
 * @brief Reports sync-in suspension signal warning to the user.
 */
void CameraMDI::report_sync_in_suspension()
{
    // Make sure the sync in is enabled for the camera.
    if (is_sync_in_enabled())
    {
        bool show_sync_in_warning = true;

        // Make sure the time is valid.
        if (time_of_sync_in_suspension.isValid())
        {
            QDateTime time_right_now = QDateTime::currentDateTime();
            qint64 delta_time = time_of_sync_in_suspension.msecsTo(time_right_now);

            // Have we waited long enough before popping up another warning?
#if LOG_SYNC_IN_SUSPEND_INFO
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "timeElapsed:" << delta_time;
#endif
            // The elapsed timer is valid so see how much time has elapsed since last warning.
            if (delta_time < SYNC_IN_FRAME_TIMEOUT_WARNING_INTERVAL)
            {
                // Wait a little longer in case the user is in process of resetting sync-in.
                show_sync_in_warning = false;
            }
        }
        else
        {
            if (!sync_in_warning_dlg)
            {
                // This is the first time we have responded to a sync-in frame time out.
                // Ticket #320 asks that the user be given more time for sync-in frame
                // time-outs because sometimes it is ok as in the case where one camera
                // driving sync-out has something modified on it.  During that modification
                // period it is just enough time for the slave camera (the camera with sync-in
                // enabled) gets a frame timeout.  We want to be a little more tolerant so
                // let it go around again.
                time_of_sync_in_suspension = QDateTime::currentDateTime();
                show_sync_in_warning = false;
            }
        }

#if LOG_SYNC_IN_SUSPEND_INFO
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "show_sync_in_warning:" << show_sync_in_warning << "sync_in_warning_state:" << sync_in_warning_state << "time_of_sync_in_suspension.is_valid():" << time_of_sync_in_suspension.isValid();
#endif

        // Is it ok to show this warning dialog to the user?
        if (show_sync_in_warning)
        {
            // Mark the time of this warning.
            time_of_sync_in_suspension = QDateTime::currentDateTime();

            // Have we already displayed the sync in warning to the user in the dialog?
            if (sync_in_warning_state == SYNC_IN_WARN_DLG_SHOWN)
            {
                bool displayOnStatusLine = true;
                // Make sure that the dialog isn't visible to the user right now.
                if (sync_in_warning_dlg)
                {
                    displayOnStatusLine = !sync_in_warning_dlg->isVisible();
                }
                // Show the message on the status line to make sure we keep reminding them about the issue.
                if (displayOnStatusLine)
                {
                    QString camera_name = get_camera_name();
                    QString status_message = QStringLiteral(" [")  + CameraMDI::sync_in_warning_title_msg + QStringLiteral(" for ") + camera_name + QStringLiteral("] ") + CameraMDI::sync_in_warning_msg;
                    MainWindow::GetInstance()->set_statusbar_message(status_message, SYNC_IN_FRAME_TIMEOUT_STATUS_INTERVAL);
                }
            }

            // If state indicates we need to show the user the sync in warning dialog then do so.
            if ((sync_in_warning_state == SYNC_IN_WARN_NONE) || (sync_in_warning_state == SYNC_IN_WARN_STREAM_RESUMED))
            {
                // Create the unique sync-in warning dialog object if it isn't already created.
                if (!sync_in_warning_dlg)
                {
                    QString camera_name = get_camera_name();
                    QString dialog_title = CameraMDI::sync_in_warning_title_msg + QStringLiteral(" for ") + camera_name;
                    sync_in_warning_dlg = fiProgressDialog::CreateSimpleProgressDialog(dialog_title, CameraMDI::sync_in_warning_msg, QMessageBox::Cancel, __PRETTY_FUNCTION__, this);
                    sync_in_warning_dlg->setCancelText(tr("OK"));
                    sync_in_warning_dlg->showBar(false);
                    sync_in_warning_dlg->showStatusIcon(true);
                    sync_in_warning_dlg->setStatusIcon(QStyle::SP_MessageBoxWarning);

                    sync_in_warning_dlg->setModal(false);   // Let's make this modeless.

                    // We want to catch the close of the dialog by the user.
                    connect(sync_in_warning_dlg.get(), &fiProgressDialog::progressCancelled, this, &CameraMDI::sync_in_warning_closed);
                }

                // If we have the sync in warning dialog show it to the user.
                if (sync_in_warning_dlg)
                {
                    sync_in_warning_dlg->showDialogWithoutPolling();
                    sync_in_warning_state = SYNC_IN_WARN_DLG_SHOWN;   // Mark that this dialog has been shown to the user.
                }
            }
        }
        emit sync_in_timeout();
    }
}

/*!
 * @brief Perform some sanity checks on the @ref file_name parameter to determine if it references
 *        a valid camera configuration file.
 *
 * We check the following:
 * - The camera's [firmware version](@ref GIGE_REG_FIRMWARE_VER) matches the firmware version we pull
 *   from the configuration file.
 * - The camera's [model](@ref GIGE_REG_MODEL_NAME) matches the camera's model.
 *
 * @param file_name     The full path to the configuration file in question.
 * @return True if the configuration file is valid, false otherwise.
 */
bool CameraMDI::valid_config_file(QString& file_name)
{
    bool        ok;
    bool        valid = false;
    uint32_t    camera_value;
    uint32_t    uint_value;
    QString     str_value;

    // Get the camera's firmware version.
    camera_value = read_camera_cache(GIGE_REG_FIRMWARE_VER, &ok);
    if (!ok)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No GIGE_REG_FIRMWARE_VER in the cache!";
        return false;
    }

    // Get the firmware version from the file.
    ok = parse_config_file(file_name, GIGE_REG_FIRMWARE_VER, uint_value);

    // Valid if the camera matches the field.
    if (ok)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "camera_value:" << camera_value << "uint_value:" << uint_value;
        valid = (camera_value == uint_value);
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Not OK";
    }

    // Do we want to keep looking?
    if (valid)
    {
        // Get the model from the file.
        ok = parse_config_file(file_name, GIGE_REG_MODEL_NAME, str_value);
        if (ok)
        {
            if (camera_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
                return false;
            }

            if (str_value.compare(camera_ptr->model_name) == 0)
                valid = false;
        }
    }

    // Return our result.
    return valid;
}

/*!
 * @brief Utility method to write a buffer to the camera.
 *
 * @param reg_addr  The register address we want to write to.
 * @param buffer    The QByteArray we want to write to the camera.
 */
void CameraMDI::write_camera_buffer(uint32_t buf_addr, const QByteArray& buffer)
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // This isn't going to work if the camera is in calibration.
    if (camera_ptr && !camera_ptr->offline)
    {
        // Issue the write.
        comm_ptr->start_write_mem(get_ip_addr(), buf_addr, buffer);
    }
}

/*!
 * @brief Utility method to write a register value to the camera.
 *
 * @param reg_addr      The register address we want to modify.
 * @param reg_value     The new value we want the register set to.
 * @param reset_retry_count     True if we're supposed to reset our retry counter, false to let it ride.
 */
void CameraMDI::write_camera_register(uint32_t reg_addr, uint32_t reg_value)
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // This isn't going to work if the camera is in calibration.
    if (camera_ptr && !camera_ptr->offline)
    {
        // Are we trying to change the state?
        if (reg_addr == GIGE_REG_CHANGE_CAMERA_STATE)
            requested_state = reg_value;

        // Issue the write.
        comm_ptr->start_write_reg(get_ip_addr(), reg_addr, reg_value);
    }
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Respond to the window Close event.
 *
 * When we receive the close event, we start the process to disconnect from
 * the camera. Because this process is "distributed", that is it takes a number
 * of separate register writes, we need to tell the caller that we're ignoring
 * the event. We're really "taking it under advisment".
 *
 * The process is:
 * 1. Shutdown the video stream. Once the write to the GIGE_REG_SCP0 register
 *    [completes](@ref CameraMDI::write_complete), we move to the next step.
 * 2. After the video stream has been disconnected, we release our lock on the
 *    camera by writing the GIGE_REG_CCP register.
 * 3. When we have [released our lock](@ref CameraMDI::write_complete), we setup
 *    a [single shot timer](@ref CAMERA_MDI_SETTINGS::SHUTDOWN_DELAY) to invoke
 *    the [finish_shutdown](@ref CameraMDI::finish_shutdown) method.
 *
 * We use the [in_shutdown](@ref CameraMDI::in_shutdown) flag to track that we're
 * in the process of disconnecting from the camera.
 *
 * @param event Pointer to the event structure.
 */
void CameraMDI::closeEvent(QCloseEvent *event)
{
    Q_UNUSED(event)

    // Don't recurse through this.
    if (!in_shutdown)
    {
        // Move this camera offline and close the window.
        set_offline(true);
    }
}

/*!
 * @brief The user has right mouse clicked in our window. Put up a
 *        context specific menu.
 *
 * @param event_ptr     Pointer to the menu event.
 */
void CameraMDI::contextMenuEvent(QContextMenuEvent *event_ptr)
{
    QMenu context_menu(this);

    // Update our menu.
    update_camera_menu();

    // Add the camera actions.
    context_menu.addActions(get_camera_actions());
    context_menu.addSeparator();

    context_menu.addAction(MainWindow::GetInstance()->get_view_summary());
    context_menu.addAction(MainWindow::GetInstance()->get_view_review_controls());
    context_menu.addAction(MainWindow::GetInstance()->get_view_pfmd());
    context_menu.addAction(MainWindow::GetInstance()->get_view_local_frametime());
    context_menu.addAction(MainWindow::GetInstance()->get_event_marker_action());
    context_menu.addAction(MainWindow::GetInstance()->get_view_aiming_cursor());
    context_menu.addAction(MainWindow::GetInstance()->get_view_info_cursor());
    context_menu.addAction(video_player->get_action_overlay());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_load());
    context_menu.addAction(MainWindow::GetInstance()->get_save());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_cam_autosave_setup());
    context_menu.addAction(MainWindow::GetInstance()->get_config_black_level());
    context_menu.addMenu(MainWindow::GetInstance()->get_config_menu());
    context_menu.addAction(MainWindow::GetInstance()->get_cam_preferences());
    context_menu.addSeparator();
    context_menu.addAction(MainWindow::GetInstance()->get_release());
    context_menu.addAction(MainWindow::GetInstance()->get_release_all());
    context_menu.addAction(MainWindow::GetInstance()->get_truncate());

    // Pop it up.
    context_menu.exec(event_ptr->globalPos());
}

/*!
 * @brief Respond to getting focus. We need to tell the main window to make other "children" inactive
 *        so we can boost the frame rate to our camera.
 * @param event     Pointer to the event structure.
 */
void CameraMDI::focusInEvent(QFocusEvent *event)
{
    uint32_t    ltr_record_mode;
    bool        ltr_append_mode = false;

    // Pass it on up.
    QWidget::focusInEvent(event);

    // Tell the main window about getting the focus.
    MainWindow::GetInstance()->change_mdi_child(this);

    // Update streaming interface.
    if (streamer != nullptr)
        streamer->update(true);

    // Connect the "active" actions we want.
    connect(MainWindow::GetInstance().get(), &MainWindow::config_black_level, this, &CameraMDI::do_black_level, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_arm, this, &CameraMDI::do_camera_arm, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_autosave_setup, this, &CameraMDI::do_autosave_setup, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_autosave_enable, this, &CameraMDI::do_camera_autosave_enable, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_live, this, &CameraMDI::do_camera_live, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_load, this, &CameraMDI::do_camera_load, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_menu_showing, this, &CameraMDI::update_camera_menu, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_power_down, this, &CameraMDI::do_camera_power_down, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_preferences, this, &CameraMDI::do_camera_preferences, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_reboot, this, &CameraMDI::do_camera_reboot, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_review, this, &CameraMDI::do_camera_review, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_samba, this, &CameraMDI::do_camera_samba, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_save, this, &CameraMDI::do_camera_save, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_save_all, this, &CameraMDI::do_camera_save_all, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_snapshot, this, &CameraMDI::do_camera_snapshot, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_trigger, this, &CameraMDI::do_camera_trigger, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_update, this, &CameraMDI::do_camera_update, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::camera_open_device, this, &CameraMDI::do_open_device, Qt::UniqueConnection);

    connect(MainWindow::GetInstance().get(), &MainWindow::config_factory_reset, this, &CameraMDI::do_config_factory_reset, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::config_load, this, &CameraMDI::do_config_load, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::config_save, this, &CameraMDI::do_config_save, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::config_time_date, this, &CameraMDI::do_config_time_date, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::config_wifi, this, &CameraMDI::do_config_wifi, Qt::UniqueConnection);

    connect(MainWindow::GetInstance().get(), &MainWindow::release_buffer, this, &CameraMDI::do_release_buffer, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::release_all_buffers, this, &CameraMDI::do_release_all_buffers, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::setup_event_markers, this, &CameraMDI::do_camera_events, Qt::UniqueConnection);

    connect(MainWindow::GetInstance().get(), &MainWindow::view_aiming_cursor, this, &CameraMDI::do_view_aiming_cursor, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_info_cursor, this, &CameraMDI::do_view_info_cursor, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_pfmd, this, &CameraMDI::do_camera_pfmd, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_local_frametime, this, &CameraMDI::do_view_localtime, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_review_controls, this, &CameraMDI::do_view_review, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::view_summary, this, &CameraMDI::do_view_summary, Qt::UniqueConnection);

    connect(video_player->get_action_overlay(), &QAction::triggered, video_player, &CameraVideo::do_add_overlay, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_1_to_1, video_player, &CameraVideo::do_zoom_1_to_1, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_in, video_player, &CameraVideo::do_zoom_in, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_out, video_player, &CameraVideo::do_zoom_out, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_fit, video_player, &CameraVideo::do_zoom_to_fit, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::zoom_to_window, video_player, &CameraVideo::do_zoom_to_window, Qt::UniqueConnection);

    connect(MainWindow::GetInstance().get(), &MainWindow::rotate_left, video_player, &CameraVideo::do_rotate_counter, Qt::UniqueConnection);
    connect(MainWindow::GetInstance().get(), &MainWindow::rotate_right, video_player, &CameraVideo::do_rotate_clockwise, Qt::UniqueConnection);

    // Make sure the zoom actions are enabled.
    MainWindow::GetInstance()->enable_zoom_actions(true);

    // Update the selected zoom.
    MainWindow::GetInstance()->check_zoom_actions(video_player->get_zoom_setting());

    // Update our view menus.
    MainWindow::GetInstance()->get_view_pfmd()->setChecked(camera_info->pfmd_is_enabled());
    MainWindow::GetInstance()->get_view_local_frametime()->setChecked(camera_info->localtime_is_enabled());
    MainWindow::GetInstance()->get_view_summary()->setChecked(camera_info->summary_is_enabled());
    MainWindow::GetInstance()->get_view_review_controls()->setChecked(review_controls->controls_enabled());

    // Reflect our current state.
    MainWindow::GetInstance()->reflect_camera_state();

    // Long Record (LTR capture mode) BROC or ROC are only capture modes supporting a capture append scenario.
    if (is_ltr_enabled(ltr_record_mode))
    {
        ltr_append_mode = ((ltr_record_mode & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0);
    }

    // If we're in LTR mode, then the "truncate" signal means to truncate the recording.
    // In a standard recording, this is used to free the buffer.
    if (ltr_append_mode)
    {
        disconnect(MainWindow::GetInstance().get(), &MainWindow::truncate_recording, this, &CameraMDI::do_truncate_release);
        connect(MainWindow::GetInstance().get(), &MainWindow::truncate_recording, this, &CameraMDI::do_truncate_clip, Qt::UniqueConnection);
    }
    else
    {
        disconnect(MainWindow::GetInstance().get(), &MainWindow::truncate_recording, this, &CameraMDI::do_truncate_clip);
        connect(MainWindow::GetInstance().get(), &MainWindow::truncate_recording, this, &CameraMDI::do_truncate_release, Qt::UniqueConnection);
    }

    // Update our information cursor, in case it's changed.
    do_view_info_cursor();

    // Update our aiming cursor.
    do_view_aiming_cursor();
}

/*!
 * @brief The showEvent is invoked either just prior or just after the window is made visible.
 * @param event     Pointer to the event class.
 */
void CameraMDI::showEvent(QShowEvent *event)
{
    // Pass it on up.
    QWidget::showEvent(event);

    // Make sure we have focus.
    setFocus();
}

/*!
 * @brief Event routine which is invoked when one of our timers expires.
 * @param event     Pointer to the event class.
 */
void CameraMDI::timerEvent(QTimerEvent *event)
{
    // Is it our lock timer?
    if (event->timerId() == lock_timer_id)
    {
        // Refresh our cache.
        refresh_cache();
    }

    // Is it our frame rate timer?
    if (event->timerId() == frame_rate_timer)
    {
        // QLOG_INFO() << "frame rate =" << frame_rate_counter;
        frame_rate_counter = 0;
    }
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief This method implements the autoexposure logic for the application.
 */
void CameraMDI::ae_track()
{
    QRect       roi_rect;

    // Make sure we are tracking. We might get called as a result of simply
    // looking at the histogram. Don't want to execute in that case.
    if (!tracking_exposure())
        return;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "ae_track_state:" << ae_track_state;

    // Process our state machine.
    switch (ae_track_state)
    {
    case AE_INIT_TRACKING:
        ae_target_Y = 0;
        ae_histo_count = 0;
        ae_ready = false;

        // Tell the player where we want the histogram.
        roi_rect = video_player->get_ae_rect();
        video_player->get_histogram()->set_roi(roi_rect);

        // Connect our histogram timer signal.
        connect(video_player->get_histogram(), &fiHistoProc::ready, this, &CameraMDI::ae_track);

        // Tell the player that we want histogram events.
        video_player->enable_histogram(true);

        // Override "normal" histogram.
        MainWindow::GetInstance()->get_histo_dock()->override_roi(roi_rect);

        // Set our state to read the histogram and wait.
        ae_track_state = AE_BASELINE;
        break;
    case AE_BASELINE:
        ae_baseline();
        break;
    case AE_PROCESS:
        ae_process();
        break;
    }
}

/*!
 * @brief Invoked when the communications object has changed the camera's status to off- or on-line.
 *
 * @sa FI_comm::camera_offline_status_changing()
 *
 * @param ip_addr   The IP address of the camera whose state is changing.
 * @param offline   True if the camera is going off-line, false if on-line.
 */
void CameraMDI::camera_offline(uint32_t ip_addr, bool offline)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << qToBigEndian(ip_addr) << "offline:" << offline;

    // Is this for us?
    if (ip_addr != get_ip_addr())
        return;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "offline:" << offline
                 << "is_series3:" << is_series3()
                 << "in_calibration():" << in_calibration();

    // Did we go off-line?
    if (offline)
    {
        // Let anyone who cares know we are closing.
        emit closing();


        // If we are series3 and NOT calibrating, then close the window.
        if (is_series3() && !in_calibration())
        {
            // Close the window for now.
            parentWidget()->close();
        }

        // If we are NOT series3 and NOT rebooting and NOT calibrating, then we should probably close the window.
        else if (!is_series3() && (camera_ptr && !camera_ptr->rebooting) && !in_calibration())
        {
            // Close the window for now.
            parentWidget()->close();
        }
        else
        {
            // At a minimum, we need to disable controls.
            if (camera_info)
                camera_info->setEnabled(false);

            // Tell the main window to go away.
            emit MainWindow::GetInstance()->active_child_changed(nullptr);

            // Have the main window reflect our new state.
            MainWindow::GetInstance()->reflect_camera_state();
        }
    }
    else
    {
        // We're coming back online. Did we have the camera reboot? If we did AND
        // we want to stream, let's get it going again.
        if (camera_ptr && camera_ptr->rebooting)
        {
            // We're not rebooting anymore.
            camera_ptr->rebooting = false;

            // Do we want to stream?
            if (camera_ptr->stream && !camera_ptr->stream_running)
            {
                // Startup our streaming machine.
                if (streamer)
                    streamer->attach();
            }

            // Re-enable the elements we disabled when we went off-line.
            if (camera_info)
                camera_info->setEnabled(true);
        }
    }
}

/*!
 * @brief Slot used to process a cancel request during a save operation.
 *
 * @sa fiProgressDialog::cancelled()
 */
void CameraMDI::cancel_save()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << get_ip_addr_pretty();

    // Make sure we have a save object.
    if (save_video != nullptr)
    {
        // Have we already started to cancel?
        if (save_video->cancelling())
            return;

        // Tell the video object that we're done.
        save_video->cancel_save(__PRETTY_FUNCTION__);
    }

    // Cleanup after ourselves.
    cleanup_save();

    // Reset our raw metdata.
    raw_metadata.clear();
    raw_metadata_len = 0;

    // If this is a cancel while downloading PFMD, setup our cancel function to
    // trigger after the video is paused.
    if (downloading_pfmd)
    {
        // Connect our signal/slots for raw PFMD download.
        connect(this, &CameraMDI::video_is_paused, this, &CameraMDI::raw_metadata_cancel, Qt::QueuedConnection);
    }

    // Stop the video stream we are done here.
    pause_video_request(__PRETTY_FUNCTION__);
}

/*!
 * @brief Timer based polling slot for checking to see if the camera has settled down.
 */
void CameraMDI::check_for_busy()
{
    uint32_t    camera_state;

    // Read the camera state and continue processing it.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    stream_setup(camera_state, __PRETTY_FUNCTION__);
}

void CameraMDI::commitData()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Check to see if we are communicating with the camera. If we are, then we'll
    // try to do a graceful shutdown.
    if (camera_ptr && !camera_ptr->offline)
    {
        // Move this camera offline, but don't close it just yet.
        set_offline(false);
    }
}

void CameraMDI::check_for_metadata()
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "downloading_pfmd:" << downloading_pfmd;

    // Don't do this if we're already doing it.
    if (downloading_pfmd)
        return;

    // Sanity check.
    if (!streamer->is_paused())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Why isn't the streamer paused!!";
        return;
    }

    // Reset our retry counter.
    retry_count_raw_metadata = 0;

    // Are the event markers enabled, and do we have the markers loaded?
    if (use_event_markers && !autoplay() && marker_list.isEmpty() && raw_metadata.isEmpty())
    {
        uint32_t    camera_state;

        // Get the camera state.
        camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
        if (camera_state != GIGE_CAMERA_STATE_REVIEW)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera:" << camera_info->get_camera_name() << "Bad camera state:" << camera_state;
            return;
        }

        // Wait for the end of the search.
        connect(this, &CameraMDI::event_search_complete, this, &CameraMDI::review_panel_ready, Qt::UniqueConnection);

        // Download the raw metadata, and then search it for matches to our event criteria.
        update_event_info();
    }
}

/*!
 * @brief Carry on with the process of calibrating the black reference frame on the camera.
 *
 * This slot is invoked once the video stream has been paused.
 */
void CameraMDI::calibration_continuation()
{
    // Done with this connection.
    disconnect(this, &CameraMDI::video_is_paused, this, &CameraMDI::calibration_continuation);

    // Are we starting a calibration event?
    if ((calibration_event == CAL_EVENT_FACTORY) || (calibration_event == CAL_EVENT_USER))
    {
        uint32_t    reg_value;
        FI_comm     *comm_ptr;

        // Are we doing manual or automatic calibration of the offset?
        if (auto_offset_cal)
            reg_value = 0;
        else
            reg_value = GIGE_FPN_CALIBRATE_MANUAL;

        // Are we doing factory or user calibration?
        if (calibration_event == CAL_EVENT_FACTORY)
            reg_value |= GIGE_FPN_FPGA_FAC_CALIBRATE_MASK;
        else
            reg_value |= GIGE_FPN_FPGA_CALIBRATE_MASK;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
            return;
        }

        // Catch when the write is complete.
        connect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraMDI::calibration_finalize, Qt::QueuedConnection);

        // Tell the camera to start.
        write_camera_register(GIGE_REG_FPN, reg_value);
    }
    else if (calibration_event == CAL_EVENT_PRNU)
    {
        uint32_t    reg_value;
        FI_comm     *comm_ptr;

        // Tell the camera that we need a PRNU calibration performed.
        reg_value = GIGE_FPN_PRNU_CALIBRATE;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
            return;
        }

        // Catch when the write is complete.
        connect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraMDI::calibration_finalize, Qt::QueuedConnection);

        // Tell the camera to start.
        write_camera_register(GIGE_REG_FPN, reg_value);
    }
}

/*!
 * @brief Finish up with the process of calibrating the black reference frame on the camera.
 *
 * This slot is invoked once the GIGE_REG_FPN register has been written.
 */
void CameraMDI::calibration_finalize(uint32_t ip_addr, uint32_t reg_addr)
{
    FI_comm     *comm_ptr;

    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // We're only interested in the FPN register.
    if (reg_addr == GIGE_REG_FPN)
    {
        // Done with this connection.
        disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &CameraMDI::calibration_finalize);

        // We're not "exactly" rebooting, but this will cause the window to
        // automatically re-open when the camera starts talking to us again.
        camera_ptr->rebooting = true;

        // Mark the camera offline and close the window.
        set_offline(true);
    }
}

/*!
 * @brief The user wants to enable/disable autosave.
 */
void CameraMDI::do_camera_autosave_enable(bool do_enable)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "do_enable:" << do_enable;

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "save_video object not allocated?!";
        return;
    }

    // Are we enabling autosave?
    if (do_enable)
    {
        // Enable autosave.
        autosave_lockout = true;
        save_video->autosave_start();
    }
    else
    {
        save_video->autosave_stop();
    }
}

/*!
 * @brief The user wants to configure the autosave process.
 *
 * @sa SaveVideo::autosave_setup
 */
void CameraMDI::do_autosave_setup()
{
    uint32_t    reg_value;

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_video object not allocated?!";
        return;
    }

    // Get the register value.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

    // Is auto save enabled?
    if ((reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
    {
        //QLOG_TRACE() << __PRETTY_FUNCTION__ << "Calling save_video->autosave_display() because we are GIGE_AUTOSAVE_ENABLE'd";

        // Show the user the autosave settings. The user may also disable autosave.
        save_video->autosave_display();
    }
    else
    {
        // Have the user setup autosave.
        if (!save_video->autosave_setup())
        {
            //QLOG_TRACE() << __PRETTY_FUNCTION__ << "save_video->autosave_setup() returned false (we are NOT GIGE_AUTOSAVE_ENABLE)";
            // Reflect that we're not configured.
            MainWindow::GetInstance()->get_cam_autosave_enable()->setEnabled(false);
            return;
        }

        // Trace the condition.
        //QLOG_TRACE() << __PRETTY_FUNCTION__ << "save_video->autosave_setup() returned true (we are NOT GIGE_AUTOSAVE_ENABLE)";

        // Reflect that we're configured.
        MainWindow::GetInstance()->get_cam_autosave_enable()->setEnabled(true);
    }
}

/*!
 * @brief This method is invoked when the user wants to perform a black level calibration.
 *
 * @sa MainWindow::config_black_level()
 */
void CameraMDI::do_black_level()
{
    // Perform the FPN calibration.
    fpn_calibration();
}

/*!
 * @brief The user has requested that we ARM the camera and start a capture.
 *
 * @sa MainWindow::camera_arm()
 */
void CameraMDI::do_camera_arm()
{
    bool        allow_discard;
    bool        user_queried;
    uint32_t    camera_state;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Before we go to live, check to see if we're in review mode.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    if (camera_state == GIGE_CAMERA_STATE_REVIEW)
    {
        // Make sure we're stopped!
        review_controls->playback_stop();
    }

    // Haven't asked the user yet.
    user_queried = false;

    // Make sure it's safe to arm.
    if (!safe_to_arm(user_queried, allow_discard))
        return;

    // Since we've determined it's OK for the ARM to happen, carry on.
    process_arm(user_queried, allow_discard);
}

/*!
 * @brief The user would like to setup the event marker display.
 *
 * @sa MainWindow::setup_event_markers()
 */
void CameraMDI::do_camera_events()
{
    bool            io_graph_was_displayed;
    eventFilterDlg  dlg(this);

    // Remember if the graph is being displayed or not.
    io_graph_was_displayed = review_controls->graph_is_enabled();

    // Initialize the dialog with our current settings.
    dlg.set_event_markers_enable(use_event_markers);
    dlg.set_display_graph_enable(io_graph_was_displayed);
    dlg.set_logic_to_and(event_logic_is_and);
    dlg.set_selects(event_selections);

    // Pop up the dialog.
    if (dlg.exec() == QDialog::Accepted)
    {
        bool    criteria_changed(false);
        int     i;

        // Read the user settings.
        use_event_markers = dlg.event_markers_are_enabled();
        event_logic_is_and = dlg.use_logic_and();

        // First, see if our criteria changed.
        for (i = 0; i < FM_EVENT_FILTERS::INDEX_COUNT; ++i)
        {
            if (event_selections.at(i) != dlg.event_selects().at(i))
            {
                criteria_changed = true;
                break;
            }
        }

        // Did the criteria change?
        if (criteria_changed)
        {
            // Clear out the old selections and copy in the new.
            event_selections.clear();
            for (i = 0; i < FM_EVENT_FILTERS::INDEX_COUNT; ++i)
                event_selections.append(dlg.event_selects().at(i));
        }

        // Do we want to reset the marker list?
        if (criteria_changed || !use_event_markers)
        {
            // Reset our maker list.
            marker_list.clear();

            // Are we interested in event markers?
            if (!use_event_markers)
                review_controls->set_markers(marker_list);
        }

        // Only need to download the raw per-frame metadata if we care about it.
        if (use_event_markers || dlg.display_graph_enabled())
        {
            // Did we turn on the IO graph display?
            if (!io_graph_was_displayed && dlg.display_graph_enabled())
            {
                // Enable the graph.
                review_controls->enable_graph(true);
            }

            // Update our event information, downloading the per-frame metadata if necessary.
            update_event_info();
        }

        // Are we supposed to turn it off?
        if (io_graph_was_displayed && !dlg.display_graph_enabled())
            review_controls->enable_graph(false);

        // Save all the settings so we can come up next time with the same values.
        save_settings();
    }
}

/*!
 * @brief The user wants the camera to switch to the LIVE state.
 *
 * @sa MainWindow::camera_live()
 */
void CameraMDI::do_camera_live()
{
    uint32_t    camera_state;

    // Before we go to live, check to see if we're in review mode.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    if (camera_state == GIGE_CAMERA_STATE_REVIEW)
    {
        // Make sure we have our review controls.
        if (review_controls && review_controls->is_playing())
        {
            // Ask the user to stop playback.
            review_controls->playback_stop();
        }
    }

    // See if we need to wait for the stream to pause.
    wait_to_transition();
}

void CameraMDI::wait_to_transition()
{
    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (streamer == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << STREAMER_SANITY_FAILURE;
        return;
    }

    // Check to see if the stream is stopped.
    if (camera_ptr->stream && !streamer->is_paused())
    {
        // Request that streaming pause.
        pause_video_request(__PRETTY_FUNCTION__);

        // Setup a one-shot.
        QTimer::singleShot(REVIEW_TO_LIVE_DELAY, this, &CameraMDI::wait_to_transition);
    }
    else
    {
        // Request the camera to switch to live mode.
        write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_LIVE_VIDEO);
    }
}

/*!
 * @brief The user wants to load a capture file on the camera.
 *
 * @sa MainWindow::camera_load()
 */
void CameraMDI::do_camera_load()
{
    bool use_explore_dlg(true);
    QJsonArray media_array;
    QJsonObject dev_object;

    // Make sure cap files are allowed on this camera.
    if (!cap_is_allowed())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Why are we here? CAP files are not allowed!";
        return;
    }

    // Newer cameras have more options.
    if (!is_series3())
    {
        QString operation_label = tr("Loading a CAP file");
        // TICKET #427 - Don't overload a dirty partition without warning the user about it.
        if (!verify_dirty_overwrite(true, operation_label))
        {
            // User doesn't want to overwrite any dirty partition(s).
            return;
        }

        // Have the user select the device to load from.
        SelectDevice sel_dlg(this);

        // Can we allow a local path as a selection?
        sel_dlg.localPathAllowed(MainWindow::GetInstance()->local_control());
        if (sel_dlg.exec() == QDialog::Rejected)
            return;

        // Retrieve the info from the selected object.
        if (sel_dlg.deviceSelected(dev_object, media_array) == -1U)
        {
            // The user selected the local path. We don't want to use the
            // explore dialog for that.
            use_explore_dlg = false;
        }
    }

    // Are we using the explore dialog?
    if (use_explore_dlg)
    {
        // Initialize the settings for the explore dialog.
        CamExplore  explore_dlg(this);
        explore_dlg.setCamera(this);

        if (is_series3())
        {
            // Series3 cameras can only load CAP files from the SSD, and they only have one.
            explore_dlg.setMediaType(GIGE_MEDIA_SSD);
        }
        else
        {
            // Pass the media specific information to the dialog.
            explore_dlg.setMediaType(dev_object["type"].toInt());
            explore_dlg.setJsonInfo(dev_object);
        }

        // Indicate that we want to load CAP files.
        explore_dlg.setCapFilesLoad(true);

        // Show the dialog.
        explore_dlg.exec();
    }
    else
    {
        QString file_path;
        uint32_t reg_value;

        // Create our load capture progress dialog.
        loadCaptureDlg = fiLoadCaptureDlg::CreateLoadCaptureDlg(get_camera_name(), this);
        if (!loadCaptureDlg)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Null load capture progress dialog!";
            return;
        }

        // Sanity check. We had better be in LIVE mode.
        reg_value = read_camera_cache(GIGE_REG_CAMERA_STATE);
        if ((reg_value != GIGE_CAMERA_STATE_LIVE_VIDEO) && (reg_value != GIGE_CAMERA_STATE_REVIEW))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Camera isn't in the right mode?!";
            return;
        }

        // Initialize our base path to whatever was used last time.
        file_path = QFileDialog::getOpenFileName(MainWindow::GetInstance().get(), tr("Select A CAP File to Load to the Camera"),
                                                 SaveDialog::loadPath(0), tr("CAP Files (*.cap)"));
        if (file_path.isEmpty())
        {
            // User must have canceled.
            return;
        }

        // Put it in a form we can use.
        QFileInfo file_info(file_path);

        // Set the dialog with the information we have.
        loadCaptureDlg->setFilePath(file_info.canonicalFilePath());
        loadCaptureDlg->setXferSrcLabel(tr("Source:"));
        loadCaptureDlg->setXferSrcValue(file_info.fileName());
        loadCaptureDlg->setStatusValue(tr("Loading partition..."));

        // Are we connected to a series3 camera?
        if (is_series3())
        {
            // Can't cancel a series3 load operation.
            loadCaptureDlg->showCancel(false);
        }
        else
        {
            // Hook up our cancel signal.
            loadCaptureDlg->showCancel(true);
        }

        loadCaptureDlg->showStatusValue(true);
        loadCaptureDlg->setBarMinMax(0, 100);
        loadCaptureDlg->setBarValue(0);

        // Connect the signal from the dialog. This way, we can close the dialog when
        // it is finished.
        connect(loadCaptureDlg.get(), &QDialog::accepted, this, &CameraMDI::load_complete, Qt::QueuedConnection);
        connect(loadCaptureDlg.get(), &fiProgressDialog::progressCancelled, this, &CameraMDI::load_complete, Qt::QueuedConnection);
        loadCaptureDlg->enableCleanup(false);

        // Get the transfer running.
        loadCaptureDlg->exec();
    }

    // Regardless of if we actually load or not, clear out the cache.
    pfmd_info->clear();
}

/*!
 * @brief The user would like to view the per-frame metadata.
 *
 * @sa MainWindow::view_pfmd()
 */
void CameraMDI::do_camera_pfmd()
{
    // Toggle the enable of the per-frame metadata display.
    bool    enable_pfmd = !camera_info->pfmd_is_enabled();

    // QLOG_INFO() << __PRETTY_FUNCTION__ << "[MAIN MENU HIT enable_pfmd][" << enable_pfmd << "]";

    // Enable/disable the per-frame metadata display ignoring any saved settings.
    camera_info->enable_pfmd(enable_pfmd);

    // Show/hide the display info.
    camera_info->show_info();

    // If we're enabling the display, make sure we grab the metadata for the current frame.
    if (enable_pfmd)
    {
        // Update the per-frame metadata display.
        update_pfmd();
    }
}

/*!
 * @brief The user would like to power down the camera.
 */
void CameraMDI::do_camera_power_down()
{
    bool was_streaming;
    int user_resp;

    // Sanity check.
    if (camera_ptr == nullptr)
        return;

    // Pause the stream from the camera.
    was_streaming = camera_ptr->stream_running;
    if (was_streaming)
        pause_video_request(__PRETTY_FUNCTION__);

    // Make sure this is something the user really wants to do.
    user_resp = QMessageBox::question(this, tr("Confirm Camera Power Down"),
                                      tr("Are you sure you want to power down the camera? Communication will be lost."),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
    if (user_resp == QMessageBox::No)
    {
        // Make sure we resume streaming, if required.
        if (was_streaming)
            resume_video_request();

        // All done.
        return;
    }

    // We are not rebooting here - this is a shutdown.
    camera_ptr->rebooting = false;

    // Have the camera power down.
    write_camera_register(GIGE_REG_RESET_CAMERA, GIGE_RESET_POWER_DOWN);
}

/*!
 * @brief The user would like to view/modify camera settings.
 */
void CameraMDI::do_camera_preferences()
{
    bool        gammaOK, jpegOK, prefOK;
    bool        extio_5V(false);
    uint32_t    gamma_value;
    uint32_t    jpeg_quality;
    uint32_t    user_prefs;
    QSettings   settings;
    CameraPreferences   dlg(this);

    // Read the register values we're interested in.
    gamma_value = read_camera_cache(GIGE_REG_GAMMA, &gammaOK);
    jpeg_quality = read_camera_cache(GIGE_REG_JPEG_QUALITY, &jpegOK);
    user_prefs = read_camera_cache(GIGE_REG_PREFERENCES, &prefOK);
    if (!gammaOK || !jpegOK || !prefOK)
    {
        // Make sure there is a camera group read object in case we need to read a group of camera registers.
        QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
        if (cam_register_group.isNull())
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
            return;
        }

        // If one is not updated, they must all be not updated.
        cam_register_group->addReg(GIGE_REG_GAMMA);
        cam_register_group->addReg(GIGE_REG_JPEG_QUALITY);
        cam_register_group->lastReg(GIGE_REG_PREFERENCES);

        // Fire off a group read.
        read_camera_registers(cam_register_group);
        return;
    }

    // Tell the dialog the current gamma setting. We only care about the default setting.
    gamma_value = gamma_value >> DEFAULT_GAMMA_SHIFT;
    dlg.set_gamma((int)gamma_value);

    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "jpeg_quality", jpeg_quality);

    // Set the current JPEG quality value.
    dlg.set_jpeg_quality((int)jpeg_quality);

    // Set the autodownload flag.
    dlg.set_metadata_autodownload(autodownload_metadata);

    // Set the autoplay flag.
    dlg.set_autoplayback((user_prefs & GIGE_PREF_AUTOPLAY) == GIGE_PREF_AUTOPLAY);

    // Not for series3 cameras
    if (!is_series3())
    {
        // Are we using 5V or 3.3V for external I/Os?
        extio_5V = (user_prefs & GIGE_PREF_EXTIO_5V) == GIGE_PREF_EXTIO_5V;
        dlg.extIO_5V(extio_5V);
    }

    // Start up the dialog.
    if (dlg.exec() == QDialog::Accepted)
    {
        uint32_t    dlg_jpeg_quality = (uint32_t) dlg.get_jpeg_quality();

        // Remember what the user preferences were.
        uint32_t old_user_prefs = user_prefs;

        // Update based on the settings in the dialog.
        // Did the gamma change?
        if (gamma_value != (uint32_t)dlg.get_gamma())
        {
            // Build up the register for the write.
            uint32_t reg_value = ((uint32_t)dlg.get_gamma()) | GIGE_GAMMA_SET_DEFAULT;
            write_camera_register(GIGE_REG_GAMMA, reg_value);
        }

        // Did the JPEG quality change?
        if (jpeg_quality != dlg_jpeg_quality)
        {
            FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "dlg_jpeg_quality()", dlg_jpeg_quality);

            // Write out the new quality value.
            write_camera_register(GIGE_REG_JPEG_QUALITY, dlg_jpeg_quality);
        }

        // Save the autodownload preference, if it changed.
        if (autodownload_metadata != dlg.get_autodownload_enable())
        {
            // Remember the new setting.
            autodownload_metadata = dlg.get_autodownload_enable();

            // Make the change persistent.
            settings.beginGroup("SaveSettings");
            settings.setValue("autodownload_metadata", autodownload_metadata);
            settings.endGroup();
        }

        // Is there a change in the autoplayback enable?
        if (((user_prefs & GIGE_PREF_AUTOPLAY) == GIGE_PREF_AUTOPLAY) != dlg.get_autoplayback_enable())
        {
            if (dlg.get_autoplayback_enable())
                user_prefs |= GIGE_PREF_AUTOPLAY;
            else
                user_prefs &= ~GIGE_PREF_AUTOPLAY;
        }

        // Check for settings not applicable to the series3 cameras.
        if (!is_series3())
        {
            // Did the external I/O voltage change?
            if (extio_5V != dlg.isExtIO_5V())
            {
                // Are we using 3.3V or 5V?
                if (dlg.isExtIO_5V())
                    user_prefs |= GIGE_PREF_EXTIO_5V;
                else
                    user_prefs &= ~GIGE_PREF_EXTIO_5V;
            }
        }

        // Was there a change in the user preferences?
        if (old_user_prefs != user_prefs)
        {
            // Write out the new settings.
            write_camera_register(GIGE_REG_PREFERENCES, user_prefs);
        }
    }
}

/*!
 * @brief The user would like to reboot the camera.
 */
void CameraMDI::do_camera_reboot()
{
    int     user_resp;

    // Make sure the user understands that this is going to result in a camera reboot.
    user_resp = QMessageBox::question(this, tr("Confirm Camera Reboot"),
                                      tr("Are you sure you want to reboot the camera? Communication will be lost and will need to be reestablished."),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
    if (user_resp == QMessageBox::No)
        return;

    // Issue the reboot command.
    reboot();
}

/*!
 * @brief The user would like the camera to go into the REVIEW state.
 *
 * @sa MainWindow::camera_review()
 */
void CameraMDI::do_camera_review()
{
    // Request the camera to switch to review mode.
    write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_REVIEW);
}

/*!
 * @brief The user would like to bring up the SAMBA interface on the camera.
 */
void CameraMDI::do_camera_samba()
{
    QString         url_string;

    // convert the packed binary format into a human readable ip address.
    url_string = QString::fromUtf8("http://%1:901").arg(get_ip_addr_pretty());
    QDesktopServices::openUrl(url_string);
}

/*!
 * @brief The user wants to take a snapshot.
 */
void CameraMDI::do_camera_snapshot()
{
    QPixmap         snap_pixmap;
    SaveSnapshot    dlg(this);

    // Make sure we have a video player.
    if (video_player == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "video_player object not allocated!";
        return;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Tell the camera to take a snapshot. We'll only use this "version" of the snapshot
    // if the user wants to save on the camera.
    write_camera_register(GIGE_REG_IMAGE_DESTINATION, GIGE_IMAGE_PREP_SNAP);

    // Grab a copy of the current video frame.
    video_player->copy_image(snap_pixmap);

    // Pass it on to our dialog.
    dlg.set_image(snap_pixmap);

    // Pause our video streaming.
    if (camera_ptr->stream_running)
        pause_video_request(__PRETTY_FUNCTION__);

    // Run the dialog.
    dlg.exec();

    // Resume our stream.
    resume_video_request();
}

/*!
 * @brief The user has initiated a save operation.
 *
 * @sa MainWindow::camera_save()
 */
void CameraMDI::do_camera_save()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__;

    // Sanity check.
    if (save_video == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_video object not allocated?!";
        return;
    }
    if (video_player == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "video_player object not allocated?!";
        return;
    }

    // We are interested in any user cancellations of the save process.  This is similar to
    // when we are saving to our local PC as well (see CameraMDI::stream_save_video)
    connect(save_video, &SaveVideo::userCancelledSaveSignal, this, &CameraMDI::cancel_save, Qt::QueuedConnection);

    // Pass it on.
    if (!save_video->save_recording())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_recording() failed.";
        return;
    }

    // Are we saving on the camera?
    if (save_video->getDestMedia() != SAVE_DLG::MEDIA_LOCAL_PC)
    {
        // We're done.
        return;
    }

    // Are we saving a capture?
    if (save_video->getStreamType() == FI_SAVER::STREAM_TYPE::FI_CAP)
    {
        // Start the process to save the session.
        start_session_save();

        // We're done.
        return;
    }

    // If we're saving an XML or using overlays, make sure our autodownload_metadata flag is set.
    if (!autodownload_metadata && (save_video->saving_xml() || video_player->has_overlay()))
    {
        QSettings   settings;

        // Turn on autodownload.
        autodownload_metadata = true;

        // Make the change persistent.
        settings.beginGroup("SaveSettings");
        settings.setValue("autodownload_metadata", autodownload_metadata);
        settings.endGroup();
    }

    // Start the save process to the local workstation.
    start_local_save(true);
}

/*!
 * @brief The user would like to save all remaining captured videos.
 *
 * @sa MainWindow::camera_save_all()
 */
void CameraMDI::do_camera_save_all()
{
    QLOG_INFO() << __PRETTY_FUNCTION__ << "Under construction.";
}

/*!
 * @brief The user would like to issue a "trigger" event to the camera.
 *
 * @sa MainWindow::camera_trigger()
 */
void CameraMDI::do_camera_trigger()
{
    // The trigger button is overloaded. In LIVE mode, it means "take a snapshot",
    // in ARMED mode, it means "trigger the capture".
    if (read_camera_cache(GIGE_REG_CAMERA_STATE) == GIGE_CAMERA_STATE_LIVE_VIDEO)
    {
        // Process a snapshot request.
        do_camera_snapshot();
    }
    else
    {
        // Request the camera to trigger.
        write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_TRIGGERED);
    }
}

/*!
 * @brief The user would like to update the camera.
 */
void CameraMDI::do_camera_update()
{
    // For series3 cameras, see if we have the update package. Only for series3, because that's the only
    // way these cameras can get an update. The new cameras can go straight to the network.
    if (is_series3())
    {
        FiUpdater *updater;

        // Get our update object.
        updater = MainWindow::GetInstance()->get_updater();

        // Do we have an update package?
        if (!QFile::exists(updater->GetPackage(FI_UPDATER::AVAIL_CAM_S3_UPDATE)))
        {
            // Ask the user if they would like to download it.
            if (QMessageBox::question(this, tr("Camera Update Package"),
                                      tr("Would you like to download the camera update package?"),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No)
                == QMessageBox::Yes)
            {
                QString pkg_dest;

                // Get the previous download path.
                pkg_dest = updater->GetUpdatePath();
                if (pkg_dest.isEmpty())
                {
                    // Ask the user for the destination for the update package.
                    pkg_dest = QFileDialog::getExistingDirectory(this, tr("Select a Directory to save the update package"), pkg_dest);
                    if (pkg_dest.isEmpty())
                        return;
                    updater->SetUpdatePath(pkg_dest);
                }

                // Have the updater download the package.
                updater->UpdateDownload(FI_UPDATER::AVAIL_CAM_S3_UPDATE);
                return;
            }
        }
    }

    // Display the dialog allowing the user to download the package.
    CamUpdateDlg update_dlg(this);
    if (update_dlg.exec() == QDialog::Rejected)
        return;

    if (is_series3())
    {
        // Pause the stream from the camera.
        if (camera_ptr && camera_ptr->stream_running)
            pause_video_request(__PRETTY_FUNCTION__);

        // Have the camera reset to factory defaults.
        write_camera_register(GIGE_REG_RESET_CAMERA, GIGE_RESET_FOR_UPDATE);

        // Remember that we are having the camera reboot.
        if (camera_ptr)
            camera_ptr->rebooting = true;
    }
    else
    {
        QJsonObject json_update;
        const FiUpdater *updater;

        // Create and show the update status window.
        MainWindow::GetInstance()->show_update_status(this);

        // Get our update object.
        updater = MainWindow::GetInstance()->get_updater();

        // Build up the command to tell the camera server where to find the update pacakge.
        json_update.insert("net", update_dlg.DoNetworkUpdate());
        if (!update_dlg.DoNetworkUpdate())
        {
            QFileInfo update_pkg(updater->GetPackage(FI_UPDATER::AVAIL_CAM_NC_UPDATE));

            // Not doing a network update, so specify where to find the update package.
            json_update.insert("device", update_dlg.GetPackagePath());

            // Specify the package name.
            json_update.insert("package", update_pkg.fileName());
        }

        // Add the update version.
        json_update.insert("version", update_dlg.GetUpdateVersion());

        // Send it on it's way.
        QJsonDocument action_doc(json_update);
        write_camera_buffer(GIGE_REG_UPDATE_JSON, action_doc.toJson());
    }
}


/*!
 * @brief The user would like to reset the camera to the factory defaults.
 *
 * @sa MainWindow::config_factory_reset()
 */
void CameraMDI::do_config_factory_reset()
{
    int     user_resp;

    // Make sure the user understands that this is going to result in a camera reboot.
    user_resp = QMessageBox::question(this, tr("Confirm Reset Camera Configuration"),
                                      tr("Are you sure you want to reset the camera configuration? This will require a restart of the camera which can take about 2 minutes."),
                                      QMessageBox::Yes | QMessageBox::No, QMessageBox::No);
    if (user_resp == QMessageBox::No)
        return;

    // Pause the stream from the camera.
    if (camera_ptr && camera_ptr->stream_running)
        pause_video_request(__PRETTY_FUNCTION__);

    // Put the camera into the idle state.
    write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_IDLE);

    // Remember that we are having the camera reboot.
    if (camera_ptr)
        camera_ptr->rebooting = true;

    // Have the camera reset to factory defaults.
    write_camera_register(GIGE_REG_RESET_TO_FACTORY, 1);
}

/*!
 * @brief The user would like to select a configuration file to push to the camera.
 *
 * @sa MainWindow::config_load()
 */
void CameraMDI::do_config_load()
{
    QString     filename;

    // Ask the user for a configuration file.
    filename = get_config_file();
    if (filename.isEmpty())
        return;

    // Push the configuration info to the camera.
    push_config_file(filename);
}

/*!
 * @brief The user would like to download the camera's current configuration and save it to a local (on the workstation) file.
 *
 * @sa MainWindow::config_save()
 */
void CameraMDI::do_config_save()
{
    QString     download_path;
    QString     filename;
    QSettings   settings;
    QFileInfo   file_info;

    // Initialize the default path based on the saved settings.
    settings.beginGroup("PathInfo");
    download_path = settings.value("config_root").toString();
    settings.endGroup();

    // If we didn't get anything, try using the download path for this camera.
    if (download_path.isEmpty())
        download_path = SaveDialog::loadPath(get_ip_addr());

    // If that's still empty, how about the system documents path?
    if (download_path.isEmpty())
        download_path = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0);

    // Build up a sample file name.
    filename = download_path + get_camera_name() + QString::fromUtf8(".cfg");

    // Ask the user for a file name.
    filename = QFileDialog::getSaveFileName(this, tr("Select a Directory and Filename to save the Camera Configuration File"),
                                            filename, tr("Configuration File (*.cfg)"));
    if (filename.isEmpty())
        return;

    // The file may or may not exist. We would prefer it not exist.
    QFile::remove(filename);

    // Get ready for the new file.
    config_file.setFileName(filename);
    if (!config_file.open(QIODevice::WriteOnly))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open" << filename << ". Err =" << config_file.errorString();
        return;
    }
    config_file.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);

    // Tell the camera to flush it's register file.
    write_camera_register(GIGE_REG_CS_REGS, GIGE_CS_REGS_FLUSH);

    // Get the path for the configuration file.
    file_info.setFile(config_file);

    // Make sure the path ends with a '/' seperator.
    download_path = file_info.canonicalPath();
    if (!download_path.endsWith(QLatin1Char('/')))
        download_path.append(QLatin1Char('/'));

    // Save the path for the configuration files.
    settings.beginGroup("PathInfo");
    settings.setValue("config_root", download_path);
    settings.endGroup();
}

/*!
 * @brief The user would like to adjust the camera's time/date.
 *
 * @sa MainWindow::config_time_date()
 */
void CameraMDI::do_config_time_date()
{
    // Read what the camera is using for an NTP server.
    start_camera_read_buf(GIGE_REG_NTP_SERVER, GIGE_NTP_SERVER_LEN);

    // Read what the camera is using for a timezone.
    start_camera_read_buf(GIGE_REG_TIMEZONE, GIGE_TIMEZONE_LEN);

    // Make sure GIGE_REG_TIMEOFDAY is the last one, 'cuz we'll use that
    // to open up the dialog for the user.
    start_camera_read_reg(GIGE_REG_TIMEOFDAY);
}

/*!
 * @brief The user would like to configure the camera's WIFI settings.
 *
 * @sa MainWindow::config_wifi()
 */
void CameraMDI::do_config_wifi()
{
    // Make sure there is a camera group read object in case we need to read a group of camera registers.
    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Build up our list of register we need to read.
    cam_register_group->addReg(GIGE_REG_WIFI_IP_ADDR);
    cam_register_group->addReg(GIGE_REG_WIFI_SUBNET_MASK);
    cam_register_group->addReg(GIGE_REG_WIFI_DEF_GATEWAY);
    cam_register_group->addReg(GIGE_REG_WIFI_STATUS);
    cam_register_group->addReg(GIGE_REG_WIFI_IP_CUR_CFG);
    cam_register_group->lastReg(GIGE_REG_WIFI_AP_CFG);

    // Start reading those registers.
    read_camera_registers(cam_register_group);

    // We also want to read the SSID currently in use. The length is
    // such that we can pull the entire SSID down in one buffer read.
    start_camera_read_buf(GIGE_REG_WIFI_SSID, GIGE_WIFI_SSID_MAX);
}

/*!
 * @brief The camera_server has completed a FasFireLR recording and we can now spawn off
 * the playback CAP review window.
 *
 * @param: json_FFLR_doc: The status of the FasFireLR recording.
 *
 */
void CameraMDI::do_fflr_review(const QJsonDocument& json_FFLR_doc)
{
    bool launched_review = false;    //!< Keeps track of whether or not the PlaybackCap review window was launched.

    uint32_t    cap_id              = FASMOTION_SETTINGS::UNSET_UINT32;
    QString     cap_id_str          = "None";
    QString     cap_filename;
    QString     cap_stack_path;
    QString     empty_string;
    QString     fflr_info_enoent    = tr("The %1 capture was unable to complete.").arg(CameraMDI::fflr_feature_name);
    QString     fflr_info_enospc    = tr("The %1 capture filled available space.").arg(CameraMDI::fflr_feature_name);
    QString     fflr_info_enxio     = tr("The %1 capture lost access to the storage device.").arg(CameraMDI::fflr_feature_name);
    QString     fflr_info_epipe     = tr("The %1 capture was interrupted.").arg(CameraMDI::fflr_feature_name);
    QString     fflr_status_str;
    QString     fflr_log_str;
    QJsonObject fflr_json_object;
    uint32_t    fflr_status         = FASMOTION_SETTINGS::UNSET_UINT32;
    QString     error_or_cancel     = tr("%1 was likely cancelled or is in error.").arg(CameraMDI::fflr_feature_name);
    QString     key_first_cap_id    = "first";
    QString     key_fflr_path       = "path";
    QString     key_fflr_status     = "status";
    QString     msg_title_str       = tr("%1 Information").arg(CameraMDI::fflr_feature_name);

    // Make sure to turn off autosave.
    if (save_video)
    {
        save_video->autosave_stop();
    }

    // Read a copy of the current camera state.
    uint32_t cameraState = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Free any partitions partially recorded if camera is in live view.
    // We only look for this live view state.  It may be possible that the
    // camera is in a review state when an errors occur. For example a device
    // can be ejected by the user before FasFireLR is able to complete.
    if ((cameraState == GIGE_CAMERA_STATE_LIVE_VIDEO) && (fflr_user_discards_parts))
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera In Live Video and User has Discarded Leftover Partitions";
        free_partition(true);
    }

    // Make sure we have valid FFLR JSON data.
    if (json_FFLR_doc.isEmpty() || json_FFLR_doc.isNull())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The jsonFFLRDoc was empty or null.";
        return;
    }

    // Retrieve the QJsonObject from the QJsonDocument.
    fflr_json_object = json_FFLR_doc.object();

    // Extract the cap id if the key exists.
    if (fflr_json_object.contains(key_first_cap_id))
    {
        cap_id       = fflr_json_object[key_first_cap_id].toInt(0);
        cap_id_str    = tr("%1").arg(cap_id);
    }

    // Extract the path to the cap stack if key exists.
    if (fflr_json_object.contains(key_fflr_path))
    {
        cap_stack_path = fflr_json_object[key_fflr_path].toString();
    }

    // Check the status of the FasFireLR operation.
    if (fflr_json_object.contains(key_fflr_status))
    {
        fflr_status = fflr_json_object[key_fflr_status].toInt(EINVAL);  // Return EINVAL if something is wrong pulling out the status code.
    }

    fflr_log_str = tr("%4 folder[%1] capid[%2] status[%3]").arg(cap_stack_path.isEmpty() ? "None" : cap_stack_path).arg(cap_id_str).arg(fflr_status).arg(fflr_feature_name);
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_log_str;

    // The value returned in fflr_status seems to correlate with GIGE_REG_SAVE_STATE values. If for some reason
    // the GIGE_SAVE_STATE_FLAGs are still set, clear them before processing the status code.
    if (fflr_status & GIGE_SAVE_STATE_FLAGS)
    {
        // Ticket #513 shows that we can end up reporting status with high bits still set.
        QLOG_WARN() << __PRETTY_FUNCTION__ << "GIGE_SAVE_STATE_FLAGS still set GIGE_REG_SAVE_STATE:" << Qt::showbase << Qt::hex << fflr_status;
        fflr_status = (fflr_status & ~GIGE_SAVE_STATE_FLAGS);
    }

    // If the user has already indicated that they want to discard the session determine if it's an underflow error
    // or do we need to set up an EPIPE request to ask about deleting directory entries...
    if (fflr_user_discards_parts)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << tr("User has discarded capture buffer(s), fflrStatus[%1]").arg(fflr_status);
    }

    // The camera server returns the following known error conditions which could be translated into
    // a more user-friendly description than an error code.
    // - 02 ENOENT means the camera server never saved a file the "why" it never saved is not clear.
    // - 06 ENXIO means no such device or address exists typically it could mean user ejected a USB drive possibly?
    // - 12 ENOMEM means the camera server hit an underflow condition where image memory was filled and could not capture any more data.
    // - 32 EPIPE means at least one CAP was saved but the metamap.txt file was not created (a cancel or error condition).
    // - 28 ENOSPC means that the camera_server ran out of space on the device and couldn't save the captured data.
    // Check for a specific error condition.

    //fflr_status = ENOSPC;    // Force error.

    // First check, did FasFireLR operation save any files?
    if (fflr_status == ENOENT)
    {
        // If we didn't get a CAP id either the user most likely cancelled the FasFireLR operation.
        if ((cap_id == FASMOTION_SETTINGS::UNSET_UINT32) || (cap_stack_path.isEmpty()))
        {
            // Treat this as an expected condition where user most likely cancelled FasFireLR.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_info_enoent << "[ENOENT]" << error_or_cancel << "\n" << fflr_log_str;
            return;
        }
    }
    else if (fflr_status == ENXIO)
    {
        fflr_status_str = fflr_info_enxio;
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_info_enxio << fflr_log_str;
    }
    else if (fflr_status == EPIPE)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_info_epipe << fflr_log_str;

        // Do we have any information in the capStackPath?  If not we don't have anything to clean up.
        if (cap_stack_path.isEmpty())
        {
            // We are done here.
            return;
        }

        fflr_status_str = fflr_info_epipe + "\n\nPartial recording data has been left in folder:\n" + cap_stack_path;

        // The folder information may be on remote camera_server device.  Generate the cap name for the first file
        // and use that path to attempt deletion if the user requests it.
        if (prepare_smb_filepath(cap_stack_path, empty_string, cap_id_str, FiSaver::IMAGE_EXT_CAPTURE, cap_filename))
        {
            QFileInfo capFileInfo(cap_filename);    // Using the system dependent path name set our file information object.
            QDir capDir(capFileInfo.path());        // We just want the pathname without the filename in our QDir.

            // Make sure the capDir exists.
            if (capDir.exists())
            {
                bool discardPartial = fflr_underflow_discarded;

                // Ticket #518
                // If we have already asked the user about discarding the recording, don't need to ask again.
                if (!fflr_underflow_discarded)
                {
                    // Ask the user if they want to delete any residual files in the folder created.
                    fflr_status_str = fflr_status_str + QString("\n\nDo you want to delete and clean up the partial %1 recording?").arg(CameraMDI::fflr_feature_name);

                    // Pop up a warning dialog and ask the user if they want to cleanup any leftover recording data.
                    // If the user responds yes to removing the leftover recording folder attempt to do so.
                    discardPartial = (QMessageBox::question(MainWindow::GetInstance().get(), msg_title_str, fflr_status_str, QMessageBox::Yes | QMessageBox::No, QMessageBox::No) == QMessageBox::Yes);
                }

                // Discarding?
                if (discardPartial)
                {
                    // Delete the partial FasFireLR recording.
                    if (!capDir.removeRecursively())
                    {
                        fflr_status_str = tr("Unable to remove partial recording data:\n%1\n\nThe action could not be completed possibly because\nanother program is accessing the folder.").arg(cap_stack_path);
                        QMessageBox::information(MainWindow::GetInstance().get(), msg_title_str, fflr_status_str);
                    }
                }

                // We are done here.
                return;
            }
        }

        // If make it to here we cannot access the path to the cap stack partially recorded but we still need to let
        // the user know there is an incomplete recording.
    }
    else if (fflr_status == ENOSPC)
    {
        uint32_t fflr_type;
        bool fflr_is_broc = false;

        // Ask for the type of LTR recording just took place.
        if (is_fflr_enabled(fflr_type))
        {
            fflr_is_broc = ((fflr_type & GIGE_LTR_MODE_BROC) != 0);
        }

        // Pop a debug message to our log.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_info_enospc << fflr_log_str;

        // If BROC mode it may be that the last recording in the available space may be truncated.
        // For the moment this is the only LTR mode where we notify the user about the ENOSPC situation.
        if (fflr_is_broc)
        {
            QString broc_warning = "\n\n" + tr("The last BROC recording could be less than a full burst") + "\n" + ("if available space ran short.");
            fflr_status_str = fflr_info_enospc + broc_warning;
        }
    }
    else if (fflr_status == ENOMEM)
    {
        fflr_status_str = tr("A camera server underflow error condition [%1] occurred, unsaved buffers[%2]").arg(fflr_status).arg((fflr_underflow_discarded ? "DISCARDED" : "SAVED"));
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_status_str << fflr_log_str;

        // Did the user discard the partial recording?
        if (fflr_underflow_discarded)
        {
            // Do we have any information in the capStackPath?  If not we don't have anything to clean up.
            if (!cap_stack_path.isEmpty())
            {
                // The folder information may be on remote camera_server device.  Generate the cap name for the first file
                // and use that path to attempt deletion if the user requests it.
                if (prepare_smb_filepath(cap_stack_path, empty_string, cap_id_str, FiSaver::IMAGE_EXT_CAPTURE, cap_filename))
                {
                    QFileInfo capFileInfo(cap_filename);    // Using the system dependent path name set our file information object.
                    QDir capDir(capFileInfo.path());        // We just want the pathname without the filename in our QDir.

                    // Make sure the capDir exists.
                    if (capDir.exists())
                    {
                        // Delete the partial FasFireLR recording.
                        if (!capDir.removeRecursively())
                        {
                            fflr_status_str = "Unable to remove leftover recording data:\n[" + cap_stack_path + "]";
                            QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_status_str << fflr_log_str;
                        }
                    }
                }
            }
        }
        return; // We return here this has been handled and we are done.
    }
    else if (fflr_status != 0)
    {
        fflr_status_str = tr("Unrecognized %1 status [%1] was returned.").arg(CameraMDI::fflr_feature_name).arg(fflr_status, 0, 16);
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_status_str << fflr_log_str;
    }

    // We will still try to launch the playback if fflrStatus was = ENOSPC but will also pop up
    // the warning message as well at the end in the final check of this module.
    if ((fflr_status == 0) || (fflr_status == ENOSPC))
    {
        // Prepare the path provided for loading in a playback review window.
        if (prepare_smb_filepath(cap_stack_path, empty_string, cap_id_str, FiSaver::IMAGE_EXT_CAPTURE, cap_filename))
        {
            // Do we have a path to something to open?
            if (!cap_filename.isEmpty())
            {
                // Had some trouble with the network interface where the first check to see if
                // the file exists would fail. Since the second try always seems to work, I thought
                // I'd pre-check here prior to calling the mainwindow to load.
                {
                    QFile load(cap_filename);
                    if (!load.exists())
                    {
                        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "File shows not existing!";
                        if (!load.open(QIODeviceBase::ReadOnly))
                        {
                            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "File won't open:" << load.errorString();

                            // Reset the path.
                            cap_filename.clear();
                        }
                    }
                }

                // Minimize the my active window and launch the playback window for the FasFireLR video.
                last_window_state = windowState();
                setWindowState(Qt::WindowMinimized);

                // You should only have to call the main window to load the cap file.
                // For this review player we want it to maximize in the MDI window space and we
                // want it focused for review purposes without any other child windows visible.
                if (MainWindow::GetInstance()->load_file(cap_filename, true))
                {
                    launched_review = true;

                    // We want to know when the FasFireLR playback window is closing down so we
                    // can reset our window state.
                    connect(MainWindow::GetInstance().get(), &MainWindow::playback_cap_closing, this, &CameraMDI::fflr_playback_closing, Qt::QueuedConnection);
                }
                else
                {
                    fflr_status_str = "Unable to open recording.";
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_status_str << fflr_log_str << "cap_filename:" << cap_filename;

                    // Restore our window.
                    fflr_playback_closing();
                }
            }
            else
            {
                fflr_status_str = "No samba file access path found.";
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_status_str << fflr_log_str;
            }
        }
        else
        {
            fflr_status_str = "Unable to prepare samba file access path.";
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << fflr_status_str << fflr_log_str;
        }
    }

    // Let the user know that something went wrong here if we have something to report.
    if (!launched_review)
    {
        // If we have something to tell the user let's do so.
        if (!fflr_status_str.isEmpty())
        {
            // Pop up a warning dialog because caller asked for the display.
            QMessageBox::information(MainWindow::GetInstance().get(), msg_title_str, fflr_status_str);
        }

        // Make sure we note this condition in the log.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Did not launch FasFireLR review player fflr_log_str:" << fflr_log_str << "fflr_status_str:" << fflr_status_str;
    }
}

/*!
 * @brief The camera was in FasFire LR underflow condition. The user has chosen to
 * complete any partial saved recording.
 *
 * @param: discard_buffers: True if the user wants dirty buffers added to partial recording, false if not.
 *
 */
void CameraMDI::do_fflr_underflow_discard()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[[[[[[ DISCARDING PARTIAL RECORDING ]]]]]]]]] discard buffers:" << fflr_user_discards_parts;

    // User wants to get rid of all of the image memory capture buffers.
    if (fflr_user_discards_parts)
    {
        release_all_buffers(true);
    }

    // Keep track of the fact that the user has discarded the partial recording as it will
    // be handled when camera state transition moves from Review into Live.
    fflr_underflow_discarded = true;    // User doesn't want any of the data.
}

void CameraMDI::do_open_device()
{
    CamExplore  *explore_dlg;

    // Allocate our dialog.
    explore_dlg = new CamExplore(MainWindow::GetInstance().get());
    if (explore_dlg == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate CamExplore dialog!";
        return;
    }

    // Set up some actions first.
    connect(explore_dlg, &CamExplore::accepted, explore_dlg, &CamExplore::deleteLater);
    connect(explore_dlg, &CamExplore::rejected, explore_dlg, &CamExplore::deleteLater);

    // Initialize the settings for the explore dialog.
    explore_dlg->setCamera(this);

    // Show the dialog.
    explore_dlg->exec();
}

/*!
 * @brief The user has requested that we release the active partition.
 */
void CameraMDI::do_release_buffer()
{
    release_buffer(false);
}

/*!
 * @brief Release the active partition.
 * @param quiet If false, we'll warn the user. If true we'll just release the partition.
 */
void CameraMDI::release_buffer(bool quiet)
{
    int         user_resp;
    QString     warn_str;

    // Do we need to warn the user?
    if (!quiet)
    {
        // Build up our warning message.
        warn_str = tr("<em>Free</em> permanently erases a recording.\n"
                      "You can choose <em>Yes</em> to free the entire recording "
                      "or choose <em>Cancel</em> to keep it. ");

        // Make sure the user understands what they're trying to do.
        user_resp = QMessageBox::warning(this, tr("Free Recording"), warn_str, QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel);
    }
    else
    {
        // It's already been decided.
        user_resp = QMessageBox::Yes;
    }

    // Just release the currently active partition.
    if (user_resp == QMessageBox::Yes)
        free_partition(false);
}

/*!
 * @brief The user has requested that we release all dirty partitions.
 */
void CameraMDI::do_release_all_buffers()
{
    release_all_buffers(false);
}

void CameraMDI::release_all_buffers(bool quiet)
{
    int         user_resp;
    QString     warn_str;
    uint32_t    reg_value;
    uint32_t    dirty_count;

    // Get the autosave settings.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

    // How many dirty partitions are there?
    dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);

    // Do we need to warn the user?
    if (!quiet)
    {
        // Build up our warning message.
        warn_str = tr("<em>Free</em> permanently erases a recording.\n"
                      "There are currently %1 partitions containing recordings. "
                      "You can choose <em>Yes to All</em> to free all partitions, "
                      "or choose <em>Cancel</em> to keep them all. ").arg(dirty_count);

        // Make sure the user understands what they're trying to do.
        user_resp = QMessageBox::warning(this, tr("Free Recording"), warn_str, QMessageBox::YesToAll | QMessageBox::Cancel, QMessageBox::Cancel);
    }
    else
    {
        // It's already been decided.
        user_resp = QMessageBox::YesToAll;
    }

    // Release all partitions.
    if (user_resp == QMessageBox::YesToAll)
        free_partition(true);
}

/*!
 * @brief The user has requested that we truncate the recording.
 */
void CameraMDI::do_truncate_clip()
{
    uint32_t    camera_state;

    // What is the camera's state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Special processing if we're in review.
    if (camera_state == GIGE_CAMERA_STATE_REVIEW)
    {
        // Don't allow the user to truncate the entire recording.
        if (review_controls->get_current_frame_num() < MIN_TRUNCATE_FRAME)
        {
            QMessageBox::warning(this, tr("Truncate Recording"), tr("You cannot truncate from frame #%1!").arg(review_controls->get_current_frame_num()));
            return;
        }

        // Are we already truncating? Don't ask the user again, if by some fluke we come
        // back in here.
        if (!in_truncate)
        {
            // Make sure the user understands what they're trying to do.
            int user_resp = QMessageBox::question(this, tr("Truncate Recording"), tr("Truncate erases from (and including) the current frame (#%1) to the end of the recording.\n\nProceed?").arg(review_controls->get_current_frame_num()));
            if (user_resp == QMessageBox::No)
                return;
        }

        // Remember that we're in the process of truncating the recording.
        in_truncate = true;

        // Switch back to Live mode.
        write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_LIVE_VIDEO);

        // Delay for a bit, and then come back.
        QTimer::singleShot(REVIEW_TO_LIVE_DELAY, this, &CameraMDI::do_truncate_clip);
    }

    // Is the video paused?
    else if (camera_ptr && streamer && camera_ptr->stream && !streamer->is_paused())
    {
        // Connect our signal/slots for pausing the video.
        connect(this, &CameraMDI::video_is_paused, this, &CameraMDI::do_truncate_clip, Qt::QueuedConnection);

        // Stop any streaming.
        pause_video_request(__PRETTY_FUNCTION__);
    }
    else
    {
        // Connect our signal/slots for pausing the video.
        disconnect(this, &CameraMDI::video_is_paused, this, &CameraMDI::do_truncate_clip);

        // Reset our flags and variables.
        last_requested_frame = (uint32_t) -1;

        // Tell the camera to truncate.
        write_camera_register(GIGE_REG_TRUNCATE, review_controls->get_current_frame_num());

        // Make sure we release our event information for the recording we just truncated.
        reset_event_info();
    }
}

/*!
 * @brief The user has requested that we truncate the recording.
 */
void CameraMDI::do_truncate_release()
{
    int         user_resp;
    uint32_t    camera_state;
    QString     warn_str;

    // What is the camera's state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Special processing if we're in review.
    if (camera_state != GIGE_CAMERA_STATE_REVIEW)
    {
        // Ignore the event. Shouldn't even be here.
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Not in review?! State:" << camera_state;
        return;
    }

    // See if we are releasing one buffer or all buffers.
    if (MainWindow::GetInstance()->get_release_all()->isEnabled())
    {
        uint32_t    reg_value;
        uint32_t    dirty_count;

        // Get the autosave settings.
        reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

        // How many dirty partitions are there?
        dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);

        // Build up our warning message.
        warn_str = tr("<em>Free</em> permanently erases a recording.\n"
                      "There are currently %1 partitions containing recordings. "
                      "You can choose <em>Yes</em> to free the current partition, or <em>Yes to All</em> to free all partitions, "
                      "or choose <em>Cancel</em> to keep them all. ").arg(dirty_count);

        // Make sure the user understands what they're trying to do.
        user_resp = QMessageBox::warning(this, tr("Free Recording"), warn_str, QMessageBox::Yes | QMessageBox::YesToAll | QMessageBox::Cancel, QMessageBox::Cancel);
    }
    else
    {
        // Build up our warning message.
        warn_str = tr("<em>Free</em> permanently erases a recording.\n"
                      "You can choose <em>Yes</em> to free the entire recording "
                      "or choose <em>Cancel</em> to keep it. ");

        // Make sure the user understands what they're trying to do.
        user_resp = QMessageBox::warning(this, tr("Free Recording"), warn_str, QMessageBox::Yes | QMessageBox::Cancel, QMessageBox::Cancel);
    }

    // How did the user respond?
    if (user_resp == QMessageBox::Cancel)
        return;

    // Are we freeing one or all?
    if (user_resp == QMessageBox::Yes)
        release_buffer(true);
    else
        release_all_buffers(true);
}

/*!
 * @brief The user is toggling the aiming cursor on or off.
 *
 * @sa MainWindow::view_aiming_cursor()
 */
void CameraMDI::do_view_aiming_cursor()
{
    if (video_player != nullptr)
        video_player->enable_aiming_cursor(MainWindow::GetInstance()->get_view_aiming_cursor()->isChecked());
}

/*!
 * @brief The user is toggling the info cursor on or off.
 *
 * @sa MainWindow::view_info_cursor()
 */
void CameraMDI::do_view_info_cursor()
{
    if (video_player != nullptr)
        video_player->enable_info_cursor(MainWindow::GetInstance()->get_view_info_cursor()->isChecked());
}

/*!
 * @brief The user would like to view time as local time.
 */
void CameraMDI::do_view_localtime()
{
    // Enable/disable the local frametime display.
    camera_info->enable_local_frametime(MainWindow::GetInstance()->get_view_local_frametime()->isChecked());

    // Show/hide the display info.
    camera_info->show_info();

    // Update the per-frame metadata display.
    update_pfmd();
}

/*!
 * @brief The user would like to view the review controls.
 */
void CameraMDI::do_view_review()
{
    review_controls->view_controls(MainWindow::GetInstance()->get_view_review_controls()->isChecked());
}

/*!
 * @brief The user would like to view the summary of the camera settings.
 *
 * @sa MainWindow::view_summary()
 */
void CameraMDI::do_view_summary()
{
    bool    enable = !camera_info->summary_is_enabled();

    // Enable/disable the partition metadata display display.
    camera_info->enable_summary(enable);

    // Show/hide the display info.
    camera_info->show_info();
}

/*!
 * @brief The [save object](@ref SaveVideo) has signaled that the save operations has completed.
 */
void CameraMDI::done_saving()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // If the streamer is not paused, then we want to transition it before continuing
    // on with the cleanup.
    if (streamer && !streamer->is_paused())
    {
        // Remember that we're in the middle of cleaning up after a save.
        save_cleanup = true;

        // Pause the streamer.
        pause_video_request(__PRETTY_FUNCTION__);

        // Request the autosave register so we have a recent copy.
        start_camera_read_reg(GIGE_REG_AUTOSAVE);
    }
    else
    {
        MultiCamDock    *dock_ptr;

        // Clean up the save.
        cleanup_save();

        // Get a pointer to the multicam dock.
        dock_ptr = MainWindow::GetInstance()->get_multicam_dock();
        if (dock_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No multicam dock!?";
            return;
        }

        // If the camera is in a group, then we'll let the multi-camera dock control
        // when to resume video streaming. Otherwise, we should let folks know it's
        // OK to stream.
        if (!dock_ptr->is_camera_grouped(this))
        {
            // Ask all cameras to resume the video stream.
            QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "Resume streaming.";
            MainWindow::GetInstance()->resume_video_streaming();
        }

        // Are we in autosave?
        if (save_video && save_video->local_autosave())
        {
            // Re-arm the camera.
            QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "Autosaving - need to rearm.";

            // From this point until we are ACTUALLY armed, I want to ignore the AUTOSAVE register.
            autosave_lockout = true;

            // Arm the camera.
            do_camera_arm();
        }
    }

    // Let folks know that we're done saving.
    emit save_finished();
}

/*!
 * @brief The [save object](@ref SaveVideo) has signaled that there was an error during the save operation.
 */
void CameraMDI::error_saving()
{
    QString err_str;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The save process has an error condition.";

    // Terminate the save.
    cancel_save();

    // Notify the user by presenting a modeless dialog to them about the error.
    // Ticket #274 Tighten-Up:
    // But let's use the main window as the parent because there are conditions
    // when copying/transcoding via the CamExplore object where if the stream is
    // paused it will terminate the connection to the camera.  This has the side
    // effect of making the message box disappear so the user never had the option
    // to respond or maybe even see it.
    QMessageBox *err_box = new QMessageBox(MainWindow::GetInstance().get());
    err_box->setText(tr("An error has occurred while saving."));

    // Build up our string with more information.
    err_str = tr("An error occurred when writing to file %1. %2").arg(save_video->getLocalPath(), save_video->getErrMsg());
    QLOG_ERROR() << __PRETTY_FUNCTION__ << err_str;
    err_box->setInformativeText(err_str);

    // Make sure the user knows this is bad.
    err_box->setIcon(QMessageBox::Critical);

    // Catch the closing so we can free up the memory later.
    connect(err_box, &QMessageBox::accepted, err_box, &QMessageBox::deleteLater);

    // Show the message as a modeless dialog.
    err_box->show();

    // Emit the save error reported to anyone that wants to know this.
    emit save_error_reported();
}

/*!
 * @brief The user wants to cancel the search for events.
 *
 * @sa CameraMDI::search_for_events()
 */
void CameraMDI::event_search_canceled()
{
    // Do we have a progress dialog up?  If so reset it.
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }
}

/*!
 * @brief The playback FasFireLR review window is closing.
 */
void CameraMDI::fflr_playback_closing()
{
    // Restore our state.
    setWindowState((Qt::WindowStates)last_window_state);

    // Make sure we get focus. This is important to restore our playback rate, if nothing else.
    setFocus();
}

/*!
 * @brief Finish the ARMing process and tell the camera to switch states.
 */
void CameraMDI::finish_arm()
{
    // Request the camera to switch to ARMed mode.
    write_camera_register(GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_ARMED);
}

/*!
 * @brief This is the final stage of disconnecting from the camera and shutting
 *        down the client (not necessarily the application though).
 */
void CameraMDI::finish_shutdown()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Kill our timers.
    if (lock_timer_id != 0)
    {
        killTimer(lock_timer_id);
        lock_timer_id = 0;
    }

    // Disconnect from the control panels.
    MainWindow::GetInstance()->get_control_panel()->cameraDisconnect();

    // Disconnect from the histogram dock.
    MainWindow::GetInstance()->get_histo_dock()->disconnect_camera();

    // We need to be inactive.
    set_inactive();

    // Make sure we still have an IP address.
    if (get_ip_addr() != -1U)
    {
        FI_comm     *comm_ptr;

        // Get our communications object pointer.
        comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr != nullptr)
        {
            // We don't want unsolicited messages anymore.
            comm_ptr->stop_unsolicited(get_ip_addr());
        }
    }

    // Tell our video object to stop processing video.
    if (video_player != nullptr)
    {
        video_player->stop_video();
        video_player->close();
    }

    // Do we have a streamer?
    if (streamer != nullptr)
    {
        // Disconnect any signal/slots from the streamer.
        disconnect(streamer, &CameraStream::detached, this, &CameraMDI::streamer_is_closed);
        disconnect(streamer, &CameraStream::image_received, this, &CameraMDI::streamer_image_received);
        disconnect(streamer, &CameraStream::paused, this, &CameraMDI::streamer_is_paused);
        disconnect(streamer, &CameraStream::ready, this, &CameraMDI::streamer_is_ready);
        disconnect(streamer, &CameraStream::streaming, this, &CameraMDI::streamer_is_streaming);
        disconnect(streamer, &CameraStream::timeout, this, &CameraMDI::streamer_image_timeout);
    }

    // Tell the main window we're closing. Rather than calling this directly, let's go through the event queue.
    QTimer::singleShot(0, MainWindow::GetInstance().get(), [ = ]() { MainWindow::GetInstance()->close_mdi_child(this); });
}

/*!
 * @brief Change the frame rate from the camera. This is NOT the camera's recording frame rate, this is
 *        the number of frames per second coming from the camera.
 * @param new_value     The new rate.
 */
void CameraMDI::framerate_changed(int new_value)
{
    uint32_t    frame_delay;

    // Sanity check the incoming value.
    if (new_value <= 0)
    {
        frame_delay = 0;
    }

    // For the series3 cameras, the rates are a little more approximate than
    // what we can support on the newer cameras.
    else if (is_series3())
    {
        // Need to sub-sample the camera's display rate. If we set the GOE inter-frame
        // delay to half (so we double the sample rate), we seem to get pretty good
        // approximation of the frame rate we want.
        frame_delay = (uint32_t)((1000 / new_value) + 1) / 2;
    }

    // For newer cameras we can be pretty accurate as to our delay between frames.
    else
    {
        frame_delay = (uint32_t)qCeil(1000.0 / new_value);
    }

    // Tell the streamer.
    streamer->set_interframe_delay(frame_delay);
}

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void CameraMDI::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    QByteArray  *new_name;

    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    // Ignore if we're in shutdown.
    if (in_shutdown)
        return;

    // Sanity check.
    if ((camera_ptr == nullptr) || (camera_info == nullptr))
        return;

    // Are we interested in this register?
    switch (reg_addr)
    {
    case GIGE_REG_AUTOSAVE:
        process_autosave(reg_value);
        break;
    case GIGE_REG_CAMERA_STATE:
        process_state_change(reg_value);
        break;
    case GIGE_REG_FPN:
        process_fpn_change(reg_value);
        break;
    case GIGE_REG_IMAGE_DESTINATION:
        // Update the image destination information.
        if (save_video != nullptr)
            save_video->update_image_dest(reg_value);
        break;
    case GIGE_REG_PART_STATE:
        process_partition_state();
        break;
    case GIGE_REG_PROGRESS:
        // QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_PROGRESS:" << reg_value;
        // Update the frame progression for camera device saves.
        if (save_video != nullptr)
        {
            save_video->save_frame_progress(reg_value);
        }
        break;
    case GIGE_REG_SAVE_STATE:
        // Notify our saver about the save state.
        if (save_video != nullptr)
        {
            save_video->save_state_change(reg_value);
        }
        break;
    case GIGE_REG_USER_DEF_NAME:
        // Get the new name.
        new_name = read_camera_buffer((uint32_t)GIGE_REG_USER_DEF_NAME);

        // Clear out our old name to make sure it is zero filled.
        memset(camera_ptr->user_name, 0, sizeof(camera_ptr->user_name));

        // Copy in the new value.
        strncpy(camera_ptr->user_name, new_name->constData(), sizeof(camera_ptr->user_name) - 1);

        // Done with the buffer.
        delete new_name;

        // Make sure it's null terminated.
        camera_ptr->user_name[ sizeof(camera_ptr->user_name) - 1 ] = '\0';
        break;

    case GIGE_REG_POWER:
        // Ask the user if they want to shutdown.
        do_camera_power_down();
        break;

    case GIGE_REG_BOOT_STATUS:
        if ((reg_value & GIGE_BOOT_STATUS_WDTIMEOUT) == GIGE_BOOT_STATUS_WDTIMEOUT)
        {
            MainWindow::GetInstance()->set_statusbar_message(tr("%1 reports that it disconnected from the head unexpectedly.").arg(camera_ptr->user_name));

            // If the head is disconnected, then hold off on any video streaming.
            if (camera_ptr->stream_running)
                pause_video_request(__PRETTY_FUNCTION__);
        }
        break;

    default:
        break;
    }
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 * @param ip_addr   The IP address of camera interested in msg timeout it may not be this camera.
 * @param reg_addr  The register address belonging to the timeout event.
 * @param read_op   True if this is a read operation, false if it is not.
 */
void CameraMDI::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << "ip_addr:" << get_ip_addr_pretty() << "reg_addr:" << reg_addr << "read_op:" << read_op;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_CCP:
        // Don't really care that we didn't get the write through if we're in shutdown. Also don't care about reads
        // 'cuz we'll try them again next time around.
        if (!in_shutdown && !read_op)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Timeout for GIGE_REG_CCP.";

            // Not shutdown. Try the write again.
            write_camera_register(GIGE_REG_CCP, GIGE_CCP_CA_BE);
        }
        break;
    case GIGE_REG_ACTIVE_METADATA:
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "GIGE_REG_ACTIVE_METADATA";
        break;
    default:
        break;
    }

    if (read_op)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Communication timeout while reading group register:" << Qt::showbase << Qt::hex << reg_addr;

        // Just call the cancel on this register.
        // If it exists in the camera's pending group_reads we will remove it.
        group_read_cancel(reg_addr);
    }
}

/*!
 * @brief Slot to handle when the user changes the camera name via the camera info widget.
 */
void CameraMDI::new_camera_name()
{
    QString     name_str;

    // Retrieve the new camera name.
    name_str = get_camera_name();

    // Let the camera know it's new name.
    write_camera_buffer(GIGE_REG_USER_DEF_NAME, name_str.toUtf8());
}

/*!
 * @brief Invoked to make sure that the progress dialog doesn't lock up the application.
 */
void CameraMDI::progress_check()
{
    // Is the progress dialog still showing?
    if (progressDialog)
    {
        // Time to let the user close the dialog.
        progressDialog->setCancelText(tr("Close"));
        progressDialog->showCancel(true);
    }
}

void CameraMDI::raw_metadata_cancel()
{
    // Disconnect our signal/slots for raw PFMD download.
    disconnect(this, &CameraMDI::video_is_paused, this, &CameraMDI::raw_metadata_cancel);

    // Were we downloading the PFMD?
    if (downloading_pfmd)
    {
        // Turn the downloading flag off.
        downloading_pfmd = false;

        // Since we cancelled, we're not using event data.
        use_event_markers = false;

        // Carry on.
        // review_panel_ready();
        emit event_search_complete();
    }
}
/*!
 * @brief Slot to be called once the stream has paused.
 */
void CameraMDI::raw_metadata_download()
{
    // Disconnect our signal/slots for raw PFMD download.
    disconnect(this, &CameraMDI::video_is_paused, this, &CameraMDI::raw_metadata_download);

    // Start downloading the raw per-frame metadata. Delay for a little bit to allow
    // the image pipeline to flush.
    QTimer::singleShot(RAW_DELAY_TIMEOUT, this, &CameraMDI::raw_metadata_start);
}

/*!
 * @brief The slot invoked when the [raw metadata download](@ref CameraMDI::stream_metadata)
 *        is finished.
 */
void CameraMDI::raw_metadata_ready()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Not interested in any more of these events right now.
    disconnect(this, &CameraMDI::raw_metadata_received, this, &CameraMDI::raw_metadata_ready);

    // We're done downloading the per-frame metadata.

    // Do we have a progress dialog up?  If so reset it.
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // Make sure we're supposed to be streaming!
    if ((camera_ptr == nullptr) || !camera_ptr->stream)
    {
        // Make sure our action panel is enabled so the user can switch to other camera states, even if we couldn't download
        // the metadata.
        camera_info->enable_actions();
        return;
    }

    // If we're in the middle of a save, we are done. The saver object
    // will take it from here.
    if (save_video && save_video->save_in_progress())
        return;

    // We're done with the raw metadata, so pause the streamer.
    if (streamer && !streamer->is_paused())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Want to pause, but not going to do it.";
    }
#ifdef DEBUG_PFMD
    if (save_video)
    {
        save_video->save_raw_metadata(get_raw_metadata(), nullptr,
                                      read_camera_cache(GIGE_REG_RED), read_camera_cache(GIGE_REG_GREEN), read_camera_cache(GIGE_REG_BLUE));
    }
#endif

    // Go process it.
    update_event_info();
}

/*!
 * @brief Delayed startup of the raw PFMD download, to allow the image pipeline to flush.
 */
void CameraMDI::raw_metadata_start()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Make sure we're supposed to be streaming!
    if (camera_ptr && camera_ptr->stream)
    {
        // Connect our slot to catch the download complete signal.
        connect(this, &CameraMDI::raw_metadata_received, this, &CameraMDI::raw_metadata_ready, Qt::UniqueConnection);
        connect(this, &CameraMDI::raw_metadata_received, save_video, &SaveVideo::raw_metadata_ready, Qt::UniqueConnection);

        // Start downloading the raw per-frame metadata.
        stream_metadata();
    }
    else
    {
        // Make sure our action panel is enabled so the user can switch to other camera states, even if we couldn't download
        // the metadata.
        camera_info->enable_actions();
    }
}

/*!
 * @brief Handle timeouts while downloading the raw metadata to the workstation.
 */
void CameraMDI::raw_metadata_timeout()
{
    if (camera_ptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "stream_running:" << camera_ptr->stream_running << "retry_count:" << retry_count_raw_metadata << "downloading_pfmd:" << downloading_pfmd;
    }

    // Do we need/want to retry?
    if (retry_count_raw_metadata < MAX_RETRIES_RAW_METADATA)
    {
        // Not interested in any more of these events right now.
        disconnect(this, &CameraMDI::video_is_paused, this, &CameraMDI::raw_metadata_wrapup);

        // Increment our retry count.
        ++retry_count_raw_metadata;

        // Re-start downloading the raw per-frame metadata. Check to see if we are
        // going to "use" the data (downloading_pfmd will be true) or "saving" the
        // data (downloading_pfmd will be false)
        if (downloading_pfmd)
        {
            // Is the camera playing in review?
            if (review_controls->is_playing())
            {
                // Tell the camera to stop!
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Tell camera to stop playing!";
                review_controls->playback_stop();
            }

            // Using the data, so call straight into the stream function.
            QTimer::singleShot(METADATA_STREAM_DELAY, this, &CameraMDI::stream_metadata);
        }
        else
        {
            // Saving the data, so reset that process.
            start_local_save();
        }
    }
    else
    {
        // Too many errors. Cancel the save operation.
        cancel_save();

        // Notify the user of our troubles.
        QMessageBox::information(this, tr("Camera Network Problem"), tr("There were too many network related errors to download the metadata."));
    }
}

/*!
 * @brief The slot invoked when the video stream has paused after the [raw metadata download](@ref CameraMDI::stream_metadata)
 *        has finished.
 */
void CameraMDI::raw_metadata_wrapup()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Not interested in any more of these events right now.
    disconnect(this, &CameraMDI::video_is_paused, this, &CameraMDI::raw_metadata_wrapup);

    // Let folks know we're done.
    emit raw_metadata_received();
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param reg_addr      The address of the register which was read.
 */
void CameraMDI::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;

    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    // Ignore if we're in shutdown.
    if (in_shutdown)
        return;

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_AVAIL_MEDIA:
        update_media();
        break;
    case GIGE_REG_MEDIA_JSON:
        process_json_media();
        break;

    case GIGE_REG_FFLR_JSON:
        process_json_fflr();
        break;

    case GIGE_REG_ACTIVE_METADATA:
    case GIGE_REG_CONFIGURATION:
        // We only care about this if we're in the middle of a download.
        if (block_download_id != (int)FI_COMM::NO_REGISTER)
        {
            // Continue with the download.
            block_download_next();
        }
        break;
    case GIGE_REG_BATTERY_CHARGE:
        // Retrieve the register value.
        reg_value = read_camera_cache(reg_addr);

        // Display the charge information.
        camera_info->display_battery_state(reg_value);
        break;

    case GIGE_REG_FORMAT_MEDIA:
        // Retrieve the register value.
        reg_value = read_camera_cache(reg_addr);
        if (reg_value == 0)
        {
            // At a minimum, we need to disable controls.
            camera_info->setEnabled(false);
        }
        else
        {
            // Re-enable the camera controls.
            camera_info->setEnabled(true);
        }
        break;

    case GIGE_REG_AUTOSAVE:
    case GIGE_REG_CAMERA_STATE:
    case GIGE_REG_SAVE_STATE:
    case GIGE_REG_IMAGE_DESTINATION:
        // Retrieve the register value.
        reg_value = read_camera_cache(reg_addr);

        // Treat these as though we got an unsolicited update.
        handle_unsol_msg(ip_addr, reg_addr, reg_value);
        break;
    case GIGE_REG_CCP:
        // Retrieve the register value.
        reg_value = read_camera_cache(reg_addr);

        // Make sure we still have lock.
        if ((reg_value & GIGE_CCP_CA_BE) == 0)
        {
            FI_comm     *comm_ptr;

            // We've lost lock. Could happen to anyone.
            QLOG_INFO() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "Lost lock, attempting to reacquire. reg_value:" << Qt::showbase << Qt::hex << reg_value;

            // Get our communications object pointer.
            comm_ptr = MainWindow::GetInstance()->get_comm();
            if (comm_ptr == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
                return;
            }

            // Force a write attempt. Don't use the write_camera_register() utility method, because
            // that'll be blocked sometimes.
            comm_ptr->start_write_reg(get_ip_addr(), GIGE_REG_CCP, GIGE_CCP_CA_BE);
        }
        break;
    case GIGE_REG_CONFIG_SIZE:
        // Retrieve the length of the configuration file.
        config_length = (int)read_camera_cache(GIGE_REG_CONFIG_SIZE);
        block_download_start(GIGE_REG_CONFIGURATION, config_length, GIGE_CONFIG_MAX_XFER_COUNT);
        break;
    case GIGE_REG_FRAME_METADATA:
        // Process the PFMD.
        read_frame_metadata();

        // Any more in the queue?
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "active_frame_requests:" << active_frame_requests;
        if (!active_frame_requests.isEmpty())
        {
            // Write the next request.
            write_camera_register(GIGE_REG_ACTIVE_FRAME_NUM, active_frame_requests.first());
        }
        break;
    case GIGE_REG_FRAME_PFMD:
        // Make sure we are expecting a response.
        if (!pfmd_frame_requests.isEmpty())
        {
            uint32_t pfmd_frame;
            QByteArray  *pfmd_array;

            // This response is satisfying the request for the first entry in the list (assuming there is one!).
            pfmd_frame = pfmd_frame_requests.takeFirst();
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_FRAME_PFMD ready for frame:" << pfmd_frame;

            // Read the metadata array.
            pfmd_array = read_camera_buffer((uint32_t)GIGE_REG_FRAME_PFMD);
            if (pfmd_array == nullptr)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "get_reg_buffer( GIGE_REG_FRAME_METADATA ) returned NULL!";
                return;
            }

            // Add the PFMD to the cache.
            pfmd_info->insert(pfmd_frame, pfmd_array);
        }
        break;

    case GIGE_REG_PART_TRIG_FRAME:
        process_partition_capture();
        break;
    case GIGE_REG_TIMEOFDAY:
        // Retrieve the register value.
        reg_value = read_camera_cache(reg_addr);

        // We've read the time/date info from the camera so the user could then
        // adjust as needed.
        process_timeofday(reg_value);
        break;
    case GIGE_REG_BOOT_STATUS:
        // Retrieve the register value.
        reg_value = read_camera_cache(reg_addr);

        // The series3 camera only reports pass/fail for the update.
        if (is_series3())
        {
            if (reg_value != 1)
            {
                // Display the message box to let the user know that the upgrade failed.
                QMessageBox::information(this, tr("Camera Upgrade"), tr("The camera reports that it failed during it's last upgrade attempt. Please create an upgrade disk using a different SDCard and try again."));
            }
        }
        else
        {
            if ((reg_value & GIGE_BOOT_STATUS_UPDATE_FAIL) == GIGE_BOOT_STATUS_UPDATE_FAIL)
            {
                // Display the message box to let the user know that the upgrade failed.
                QMessageBox::information(this, tr("Camera Upgrade"), tr("The camera reports that it failed during it's last upgrade attempt. Please create an upgrade disk using a different SDCard and try again."));
            }
            if ((reg_value & GIGE_BOOT_STATUS_WDTIMEOUT) == GIGE_BOOT_STATUS_WDTIMEOUT)
            {
                // Display the message box to let the user know that the camera's watchdog has timed out.
                QMessageBox::information(this, tr("Camera Boot Problem"), tr("The camera reports that it disconnected from the head unexpectedly. Please verify that the connection is secure before proceeding."));
            }
        }
        break;
    case GIGE_REG_WIFI_SSID:
        // This was the last register we needed to read. Go configure the WIFI now.
        configure_wifi();
        break;
    case GIGE_REG_FIRMWARE_VER:
        break;
    case GIGE_REG_WIDTH:
    case GIGE_REG_HEIGHT:
    case GIGE_REG_FRAME_RATE:
    case GIGE_REG_SHUTTER_SPEED:
    case GIGE_REG_TRIGGER_REQ_POS:
    case GIGE_REG_SENSOR_GAIN:
        // Time to update our status information.
        update_status();
        break;
    case GIGE_REG_SYNC_IN_TIMING:
        // Are we sync-in enabled and not receiving a sync-in pulse signal?
        if (is_sync_in_enabled() && !is_sync_in_receiving_pulse())
        {
            report_sync_in_suspension();
        }
        else
        {
            // Falling in here could be !is_sync_in_enabled() **OR** is_sync_in_enabled() && is_sync_in_receiving_pulse().
            // Cleanup any prior state of sync-in warnings.

            // Had we shown a sync-in warning dialog?
            if (sync_in_warning_state == SYNC_IN_WARN_DLG_SHOWN)
            {
#if LOG_SYNC_IN_SUSPEND_INFO
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Resetting sync_in_warning_state to SYNC_IN_WARN_STREAM_RESUMED.";
#endif
                // The sync in warning state should reflect that the stream has resumed.
                sync_in_warning_state = SYNC_IN_WARN_STREAM_RESUMED;

                // Cleanup any sync-in enabled loss of frame stream processing.
                sync_in_warning_closed();

                // Let anyone know that was listening we are resumed.
                emit sync_in_resumed();
            }
        }
        break;
    default:
        break;
    }

    // Go through the group reads and tell them about the register we just read.
    for (int i = 0; i < group_reads.count(); i++)
    {
        QSharedPointer<FiCamRegGroup> cam_group_read = group_reads.at(i);

        // Remove any entries in our queue that may have emptied out.
        if (cam_group_read->getCount() == 0)
        {
            uint32_t last_reg_addr = cam_group_read->getLastRegAddr();
            if (last_reg_addr > 0)
            {
#if LOG_GROUP_READS
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME)
                             << "[EMPTY-COMPLETE][" << cam_group_read->getCamRegGroupSigStr()
                             << "][COUNT:" << group_reads.count() << "]";
#endif
                // Go ahead and emit the last_reg addr in this group just in case it was missed.
                emit group_read_complete(ip_addr, last_reg_addr);
                group_reads.removeAt(i);
            }
        }
        // Otherwise if there are still entries in this record see if this register is one of them.
        else if (cam_group_read->delReg(reg_addr))
        {
            // Have we read all of this group?
            if (cam_group_read->getCount() == 0)
            {
#if LOG_GROUP_READS
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME)
                             << "[COMPLETE][" << cam_group_read->getCamRegGroupSigStr()
                             << "][COUNT:" << group_reads.count() << "]";
#endif
                // Emit the signal to anyone listening that this group read is complete.
                uint32_t last_reg_addr = cam_group_read->getLastRegAddr();
                emit group_read_complete(ip_addr, last_reg_addr);
                group_reads.removeAt(i);
            }
        }
    }
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the refresh cache group read's last address == GIGE_REG_SAVE_STATE.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void CameraMDI::chk_refresh_cache_read(uint32_t ip_addr, uint32_t reg_addr)
{
    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    // Verify it's our register.
    if (reg_addr == GIGE_REG_SAVE_STATE)
    {
#if LOG_GROUP_READS
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_SAVE_STATE]";
#endif
        // Time to update our status information.
        update_status();
    }
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the camera_preferences group read's last address == GIGE_REG_PREFERENCES
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void CameraMDI::chk_camera_preference_read(uint32_t ip_addr, uint32_t reg_addr)
{
    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    // Verify it's our register.
    if (reg_addr == GIGE_REG_PREFERENCES)
    {
#if LOG_GROUP_READS
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_PREFERENCES]";
#endif
        QTimer::singleShot(0, this, &CameraMDI::do_camera_preferences);
    }
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the update record progress' group read's last address == GIGE_REG_CUR_FRAME_POS.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void CameraMDI::chk_rec_progress_read(uint32_t ip_addr, uint32_t reg_addr)
{
    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    // Verify it's our register.
    if (reg_addr == GIGE_REG_CUR_FRAME_POS)
    {
#if LOG_GROUP_READS
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_CUR_FRAME_POS]";
#endif
        update_rec_info();
    }
}

/*!
 * @brief The review panel is now ready to operate. Continue with any necessary initializations.
 */
void CameraMDI::review_panel_ready()
{
    uint32_t    ltr_record_mode;
    bool        ltr_append_mode = false;

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Disconnect from the event_search_complete signal. It may or may not have been
    // connected, but either way we would like to be sure it's disconnect now.
    disconnect(this, &CameraMDI::event_search_complete, this, &CameraMDI::review_panel_ready);
    disconnect(this, &CameraMDI::pm_metadata_received, this, nullptr);

    //---------------
    // Step 1.
    //---------------
    // Before we can continue into the review "state", we need to make sure that the video
    // streamer has paused any live video.
    if (!streamer->is_paused())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Wait for streamer to pause.";

        // Wait for a little bit until the stream has paused.
        QTimer::singleShot(DELAY_UPDATE, this, &CameraMDI::review_panel_ready);
        return;
    }

    //---------------
    // Step 2.
    //---------------
    // We need to make sure we have the partition metadata. See if the download is already in progress.
    if (!pm_download_in_progress && partition_metadata.isEmpty())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Camera:" << camera_info->get_camera_name() << "OK to download partition metadata.";

        // We need a copy of the partition metadata. Set our flag so we won't try to save anything.
        pm_download_only = true;
        pm_download_in_progress = true;

        // Make sure the buffer is clear.
        partition_metadata.clear();

        // Catch the event for when the metadata is received.
        connect(this, &CameraMDI::pm_metadata_received, [ = ](uint32_t ip_addr) { if (ip_addr == this->get_ip_addr()) { QTimer::singleShot(0, this, &CameraMDI::review_panel_ready); }});

        // Have the camera re-acquire the metadata so we can download it.
        request_active_metadata(__PRETTY_FUNCTION__);
        return;
    }

    //---------------
    // Step 3.
    //---------------
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Ready to check for PFMD metadata.";

    // Now that we have the per-frame metadata, see if we need the full raw metadata information.
    check_for_metadata();

    // Check to see if we are now downloading the PFMD. If we are, leave until it is finished downloading.
    if (downloading_pfmd)
    {
        // Make sure we reconnect to get notified of the end of the search.
        QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty() << "Downloading the PFMD, so bail...";
        connect(this, &CameraMDI::event_search_complete, this, &CameraMDI::review_panel_ready, Qt::UniqueConnection);
        return;
    }

    //---------------
    // Step 4.
    //---------------
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Ready to wrap things up and go into the review state.";

    // Long Record (LTR capture mode) BROC or ROC are only capture modes supporting a capture append scenario.
    if (is_ltr_enabled(ltr_record_mode))
    {
        ltr_append_mode = ((ltr_record_mode & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0);
    }

    // If we're in LTR mode, then the "truncate" signal means to truncate the recording.
    // In a standard recording, this is used to free the buffer.
    if (ltr_append_mode)
    {
        disconnect(review_controls, &CameraReview::truncate, this, &CameraMDI::do_truncate_release);
        connect(review_controls, &CameraReview::truncate, this, &CameraMDI::do_truncate_clip, Qt::UniqueConnection);
    }
    else
    {
        disconnect(review_controls, &CameraReview::truncate, this, &CameraMDI::do_truncate_clip);
        connect(review_controls, &CameraReview::truncate, this, &CameraMDI::do_truncate_release, Qt::UniqueConnection);
    }

    // Enable our review controls.
    review_controls->enable_panel(true);

    // Make sure our camera actions reflect the current state.
    process_partition_state();

    // Are we interested in our per-frame metadata being displayed?
    if (camera_info->pfmd_is_enabled())
    {
        // Display the per-frame metadata controls. We'll actually retrieve
        // the data in the video_current_frame_changed() call.
        camera_info->show_info();
    }

    // Make sure our action panel is enabled so the user can switch to other camera states.
    camera_info->enable_actions();

    // Request a video frame.
    force_frame_update = true;

    // If we are in auto-play, and we're not a series3 camera, have
    // the camera immediately start playback. If we're not in auto-play,
    // then issue a single-step.
    if (autoplay())
    {
        // Don't need this for series3 cameras.
        if (!is_series3())
        {
            // Start autoplay.
            video_play_forward();
        }
    }
    else
    {
        // Give the camera a chance to change states, etc.
        // QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Set timer for video_current_frame_changed()";
        QTimer::singleShot(LIVE_TO_REVIEW_DELAY, this, &CameraMDI::video_current_frame_changed);
    }

    // Let's signal to other folks that our review panel is considered ready to go.
    emit review_panel_ready_to_go();
}

/*!
 * @brief The [save object](@ref SaveVideo) has signaled that the save download operation has received
 * the number of non-empty buffers that were expected.
 */
void CameraMDI::save_receive_count_complete()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // All we really want to do in this case is pause the streamer if it isn't already paused.
    if (streamer && !streamer->is_paused())
    {
        // Pause the streamer.
        pause_video_request(__PRETTY_FUNCTION__);
    }
}

/*!
 * @brief Handle the signal from the save_video object indiciating that it's worker thread
 *        is idle. We can use this opportunity to see if we have images we need to re-transmit
 *        from the camera.
 */
void CameraMDI::saver_is_idle()
{
    // Process based on the streamer image type.
    switch (streamer->image_type())
    {
    case CAMERA_STREAM::STREAM_SAVE_SESSION:
    case CAMERA_STREAM::STREAM_SAVE_IMAGE:
        // Is this the last frame we're interested in? The sequnce number is
        // 1-based, so the last sequence number will be equal to the total number
        // of expected frames.
        if ((last_received_frame == save_video->getFrameCount()) || retry_in_progress)
        {
#if LOG_SAVER_IDLE_MSGS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "last_received_frame:" << last_received_frame << "get_frame_count():"
                         << save_video->get_frame_count() << "retry_in_progress:" << retry_in_progress << "waiting_for_frame:"
                         << waiting_for_frame;
#endif

            // Do we have frames to be retried?
            if (retry_frame_list.isEmpty())
            {
                // TICKET #439:
                // We could be in a situation where the last retry frame was requested but
                // we still have not received it yet.  Should that be the case then the
                // waiting_for_frame will be true.  We cannot be finished yet as that frame
                // is still expected from the camera streamer.  Therefore if waiting_for_frame
                // we will continue to wait.
                if (!waiting_for_frame)
                {
                    // Reset our last received frame, so we don't come back here by mistake.
                    last_received_frame = 0;

                    // Make sure our retry flag is reset.
                    retry_in_progress = false;

                    // Let the video saver object know we're done.
                    save_video->saving_finished(__PRETTY_FUNCTION__);
                }
            }
            else if (!waiting_for_frame)
            {
                // Start retrying our missing frames.
                retry_missing_frames();
            }
        }
        break;
    case CAMERA_STREAM::STREAM_COPY_IMAGE:
        // Are we in "retry" mode?
        if (retry_in_progress && !waiting_for_frame)
        {
            // Continue with the retry process.
            retry_missing_frames();
        }
        break;
    default:
        break;
    }
}

void CameraMDI::streamer_faulted()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;
    streamer_image_timeout();
}

/*!
 * @brief The streamer has received an image.
 */
void CameraMDI::streamer_image_received()
{
    QString infoMsg = QStringLiteral("[CameraMDI::streamer_image_received()]");
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr;

    // Loop while there are images pending and we're not in shutdown.
    while (!in_shutdown && streamer && (streamer->image_count() > 0))
    {
        uint32_t    seq_number;

        // Pop the first image off the list.
        imageBufferPtr = streamer->image_take_first();

        // Sanity check.  Log a diagnostic out there for this resource.
        if (imageBufferPtr)
        {
            imageBufferPtr->LogMyDataId(&infoMsg);
        }
        else
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "!![RESOURCE EMPTY]!! imageBufferPtr was nullptr!";
            continue;
        }

        // Increment our counter.
        frame_rate_counter++;

        // Process based on the streamer state.
        switch (streamer->image_type())
        {
        case CAMERA_STREAM::STREAM_SAVE_IMAGE:
            // Drop any stray live video frames.
            if (FiSaver::isLiveSeqNumber(imageBufferPtr->GetSequenceNumber()))
            {
                // Drop this frame.
                release_buffer(std::move(imageBufferPtr));
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "STREAM_SAVE_IMAGE: Drop live frame.";
                break;
            }

            // Are we retrying?
            if (retrying_frames() && (imageBufferPtr->GetSequenceNumber() == 1))
            {
                uint32_t retryFrameSeqNbr = 0;
                // Make sure the sequence number is what we expect.
                if (save_video == nullptr)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_video is NULL?";
                }
                else
                {
                    retryFrameSeqNbr = (last_received_frame - save_video->getClipStart()) + 1;
                    imageBufferPtr->SetSequenceNumber(retryFrameSeqNbr);
                }
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Retrying, processing missed frame#:" << imageBufferPtr->GetSequenceNumber()
                             << "last_received_frame:" << last_received_frame << "clip_start:" << save_video->getClipStart();
            }

            // We are not waiting for a frame anymore.
            waiting_for_frame = false;

            // Remember some information about the image.
            seq_number = imageBufferPtr->GetSequenceNumber();
#if FORCE_MISSING_FRAMES
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_number:" << seq_number << "retrying_frames()" << retrying_frames();

            if (!retrying_frames())
            {
                // #MISSING_FRAME - Force a missing frame for the save operation.
                bool    forceMissingFrame   = false;
                bool    isMetafile          = false;
                bool    isBlobData          = false;
                uint32_t test_seq_number    = 0;
                uint32_t final_seq_number   = save_video->getClipEnd() - save_video->getClipStart() + 1;

                // Determine some information about this sequence number.
                if (FiSaver::metadataSeqNum(imageBufferPtr->GetSequenceNumber()))
                {
                    isMetafile = true;
                }

                // Is this a blob buffer.
                if (save_video->getStreamType() == FI_SAVER::STREAM_TYPE::FI_CAP)
                {
                    isBlobData = true;
                }

                // If we aren't dealing with a metafile determine the frame/blob number.
                if (!isMetafile)
                {
                    test_seq_number = seq_number - save_video->getClipStart();
                }

                // Catch the very first frame and force an empty frame there.
                if (test_seq_number == 1)
                    forceMissingFrame = true;

                // Catch the send blob in the copy stream and make it missing.
                //if (test_seq_number == 2)
                //    forceMissingFrame = true;

                // Create a gap in the middle.  Change numbers as you see fit.
                if ((test_seq_number >= 10) && (test_seq_number <= 12))
                    forceMissingFrame = true;

                // Create a gap at the end.
                if (test_seq_number == final_seq_number)
                    forceMissingFrame = true;

                // If forcing an empty frame replace our valid one with the empty one.
                if (forceMissingFrame)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "#MISSING_FRAME - Forcing missed bufSeqNbr:" << seq_number
                                 << "isMetafile:" << isMetafile
                                 << "isBlobData:" << isBlobData
                                 << "seq_number:" << seq_number
                                 << "last_received_frame:" << last_received_frame
                                 << "test_seq_number:" << test_seq_number
                                 << "final_seq_number:" << final_seq_number
                                 << "clip_start:" << save_video->getClipStart()
                                 << "clip_end:" << save_video->getClipEnd();

                    // Release this valid frame we are making it a "missing frame".
                    release_buffer(std::move(imageBufferPtr));
                    return;
                }
            }
#endif
            // Process the received frame to check for any dropped frames, etc.
            process_received_image(seq_number);

            // Do we need to merge the overlay information?
            if (save_video && save_video->do_merge_overlay())
            {
                merge_overlay(seq_number, imageBufferPtr.get());
            }

            // Save the image.
            if (save_video != nullptr)
            {
                if (!save_video->receive_image(std::move(imageBufferPtr)))
                {
                    // We don't really have a way to cancel the operation, so log the error.
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to insert image in saver's image pool, seq_number:" << seq_number;
                }
            }
            break;
        case CAMERA_STREAM::STREAM_SAVE_SESSION:
        case CAMERA_STREAM::STREAM_COPY_IMAGE:
            // Drop any stray live video frames.
            if (FiSaver::isLiveSeqNumber(imageBufferPtr->GetSequenceNumber()))
            {
                // Drop this frame.
                release_buffer(std::move(imageBufferPtr));
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "STREAM_COPY_IMAGE: Drop live frame.";
                break;
            }

            // Remember some information about the "chunk".
            seq_number = imageBufferPtr->GetSequenceNumber();
#if FORCE_MISSING_FRAMES
            // Note:  Don't force a missing frame if this is a retry operation.
            // Note:  For copying stream frames there seems to be a problem with seriesNC
            // streams in that they don't turn off the spicket - the streamer continues to
            // send the same frame stream without regard to a single frame.  For now I am
            // not allowing this to be set for seriesNC which in all likelyhood will never
            // drop frames with TCP/IP.  It should be REVISIT_JLC in the future.
            if (!retrying_frames() && is_series3())
            {
                // #MISSING_FRAME - Force a missing frame for the copy operation.
                bool    forceMissingFrame   = false;
                bool    isMetafile          = false;
                bool    isBlobData          = false;
                uint32_t test_seq_number    = 0;
                uint32_t last_seq_number    = save_video->getClipEnd() - save_video->getClipStart() + 1;

                // Determine some information about this sequence number.
                if (FiSaver::metadataSeqNum(imageBufferPtr->GetSequenceNumber()))
                {
                    isMetafile = true;
                }

                // Is this a blob buffer.
                if (save_video->getStreamType() == FI_SAVER::STREAM_TYPE::FI_CAP)
                {
                    isBlobData = true;
                }

                // If we aren't dealing with a metafile determine the frame/blob number.
                if (!isMetafile)
                {
                    test_seq_number = seq_number - save_video->getClipStart();
                }

                // Catch the very first frame and force an empty frame there.
                if (test_seq_number == 1)
                    forceMissingFrame = true;

                // Catch the send blob in the copy stream and make it missing.
                //if (test_seq_number == 2)
                //    forceMissingFrame = true;

                // Create a gap in the middle.  Change numbers as you see fit.
                //if ((test_seq_number >= 10) && (test_seq_number <= 12))
                //    forceMissingFrame = true;

                // Create a gap at the end.
                if (test_seq_number == last_seq_number)
                    forceMissingFrame = true;

                // If forcing an empty frame replace our valid one with the empty one.
                if (forceMissingFrame)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "#MISSING_FRAME - Forcing missed bufSeqNbr:" << seq_number
                                 << "isMetafile:" << isMetafile
                                 << "isBlobData:" << isBlobData
                                 << "seq_number:" << seq_number
                                 << "last_received_frame:" << last_received_frame
                                 << "test_seq_number:" << test_seq_number
                                 << "last_seq_number:" << last_seq_number
                                 << "clip_start:" << save_video->getClipStart()
                                 << "clip_end:" << save_video->getClipEnd();

                    // Release this valid frame we are making it a "missing frame".
                    release_buffer(std::move(imageBufferPtr));
                    return;
                }
            }
#endif
            // We are not waiting from the frame anymore.
            waiting_for_frame = false;

            // Process the received "chunk" to check for any dropped "chunks", etc.
            process_received_chunk(seq_number);

            // Sanity check.
            if (save_video != nullptr)
            {
                // If this is a meta data frame we need to check to see if we
                // missed the very last frame in the copy stream.  And we need
                // to collect the lost frame in the retry collection.
                if (FiSaver::metadataSeqNum(imageBufferPtr->GetSequenceNumber()))
                {
                    // Only do this if we are processing a stack of images.
                    // The reason is that when copying a stack of images the
                    // camera_server includes any metafiles in the stack as a
                    // part of the whole transfer package and blob_count.
                    // If this is the first metafile in the "copy" package to
                    // be received a check must be made to see if the last frame
                    // is missing.  That occurs here.  This is not checked for
                    // when copying video or cap blob packages because that
                    // mechanism is managed differently where the video/cap is
                    // transferred in one blob(s) package and then when that is
                    // complete the metafile(s) are each copied in their own blob(s)
                    // package individually.  Lost frames in these blob packages
                    // for video/cap copies are handled in each blob(s) copy package.
                    if (FiSaver::isStackOfImages(save_video->getStreamType()))
                    {
                        // If this is our first metafile to copy check for missing last frame.
                        if (count_metafiles_copied == 0)
                        {
                            last_frame_missing_check();
                        }
                    }
                    // Increment our accumulator for metadata chunks processed.
                    count_metafiles_copied++;
                }
                // If not a CAP file copy then adjust the sequence number for the saver.
                else if (save_video->getStreamType() != FI_SAVER::FI_CAP)
                {
                    // Only adjust the sequence number if this is NOT a metadata or CAP file.
                    imageBufferPtr->SetSequenceNumber((seq_number - save_video->getClipStart()));
                }

                // Save the chunk. Make sure we are successful.
                if (!save_video->receive_image(std::move(imageBufferPtr)))
                {
                    // We don't really have a way to cancel the operation, so log the error.
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to insert image chunk in saver's image pool, seq_number:" << seq_number;
                }
            }
            break;
        case CAMERA_STREAM::STREAM_RAW_METADATA:
            // Drop any stray live video frames.
            if (FiSaver::isLiveSeqNumber(imageBufferPtr->GetSequenceNumber()))
            {
                // Drop this frame.
                release_buffer(std::move(imageBufferPtr));
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "STREAM_RAW_METADATA: Drop live frame.";
                break;
            }

            // Got a good sequence number, so go deal with it.
            process_received_metadata(std::move(imageBufferPtr));
            break;
        default:
            // In "live" view, we only really care about the latest image. Do we
            // have any more images queued?
            while (!in_shutdown && streamer && (streamer->image_count() > 0))
            {
                // Release the one we have.
                release_buffer(std::move(imageBufferPtr));

                // Grab the next one image off the list.
                imageBufferPtr = streamer->image_take_first();

                // Sanity check.  Log a diagnostic out there for this resource.
                if (imageBufferPtr)
                {
                    QString infoPlusMore = QString("[%1-LV THROW AWAY]").arg(infoMsg);
                    imageBufferPtr->LogMyDataId(&infoPlusMore);
                }
                else
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "!![RESOURCE EMPTY LV THROW AWAY]!! imageBufferPtr was nullptr!";
                    break;
                }
            }

            // At this point the streamer image count should be 0 and the latest
            // image should be in imageBufferPtr and this object "owns" the pointer.

            // Sanity check.
            if (!in_shutdown && (imageBufferPtr != nullptr))
            {
                // Make sure we're still on-line.
                if (is_offline())
                {
                    // Release the one we have.
                    release_buffer(std::move(imageBufferPtr));
                }
                else
                {
                    FI_TIFF::TIFF_HEADER *temp_hdr = ((FI_TIFF::TIFF_HEADER *)imageBufferPtr->GetMemDataPtr());

                    // Check if we're using TIFs for live view.
                    if (temp_hdr->tif_magic == FI_TIFF::TIFF_MAGIC)
                        MainWindow::GetInstance()->set_statusbar_text(QStringLiteral("TIF-live"));
                    else
                        MainWindow::GetInstance()->set_statusbar_text(QString());

                    // Display the image.
                    video_player->receive_image(std::move(imageBufferPtr));
                }
            }
            break;
        }
    }

    // Are we sync-in enabled and receiving out pulses now?
    if (is_sync_in_enabled() && is_sync_in_receiving_pulse())
    {
        // Had we shown a sync-in warning dialog?
        if (sync_in_warning_state == SYNC_IN_WARN_DLG_SHOWN)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Resetting to SYNC_IN_WARN_STREAM_RESUMED.";
            // The sync in warning state should reflect that the stream has resumed.
            sync_in_warning_state = SYNC_IN_WARN_STREAM_RESUMED;
        }

        // Cleanup any sync-in enabled loss of frame stream processing.
        sync_in_warning_closed();

        // Let anyone know that was listening we are resumed.
        emit sync_in_resumed();
    }
}

/*!
 * @brief Timed out waiting for an image frame from the camera.
 */
void CameraMDI::streamer_image_timeout()
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << get_ip_addr_pretty();

    // Handle the timeout recover based on what type of image we're expecting.
    switch (streamer->image_type())
    {
    case CAMERA_STREAM::STREAM_COPY_IMAGE:
        copy_video_timeout();
        break;
    case CAMERA_STREAM::STREAM_LIVE_IMAGE:
        // We can ignore timeouts in the live stream - the streamer can handle
        // them just fine.
        break;
    case CAMERA_STREAM::STREAM_SAVE_IMAGE:
        save_video_timeout();
        break;
    case CAMERA_STREAM::STREAM_RAW_METADATA:
        raw_metadata_timeout();
        break;
    default:
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unexpected image_type:" << streamer->image_type();
        break;
    }
}

/*!
 * @brief Catch the streamer's "detach" signal.
 */
void CameraMDI::streamer_is_closed()
{
    FI_comm     *comm_ptr;

    // No more streaming from the camera.
    ok_to_talk = false;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Sanity check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Finish the shutdown process by releasing the camera.
    comm_ptr->start_write_reg(camera_ptr->ip_addr, GIGE_REG_CCP, 0);
}

/*!
 * @brief Catch the streamer's paused signal.
 */
void CameraMDI::streamer_is_paused()
{
    if (camera_ptr)
    {
        // No longer streaming.
        camera_ptr->stream_running = false;
    }
    else
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // Update our UI element.
    if (camera_info)
    {
        camera_info->enable_stream(true);
    }
    else
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CAMERA_INFO_SANITY_FAILURE;
    }

    // Are we in the middle of a save cleanup?
    if (save_cleanup)
    {
        // Clear our flag.
        save_cleanup = false;

        // Carry on with the cleanup.
        done_saving();
    }

    // Pass on the event.
    emit video_is_paused();
}

/*!
 * @brief Catch the streamer_is_ready signal.
 */
void CameraMDI::streamer_is_ready()
{
    uint32_t    reg_value;

    // We're ready to stream from the camera now.
    ok_to_talk = true;

    // Check the camera state to determine how to proceed.
    reg_value = read_camera_cache(GIGE_REG_CAMERA_STATE);
    switch (reg_value)
    {
    case GIGE_CAMERA_STATE_ARMED:
    case GIGE_CAMERA_STATE_TRIGGERED:
    case GIGE_CAMERA_STATE_DEARMING:
    case GIGE_CAMERA_STATE_LIVE_VIDEO:
        // Make sure the user can enable/disable streaming.
        camera_info->enable_stream(true);
        streamer->go_live();
        break;
    case GIGE_CAMERA_STATE_AUTOSAVE:
    case GIGE_CAMERA_STATE_REVIEW:
    case GIGE_CAMERA_STATE_SNAPSHOT:
        // The stream button has no meaning in this context.
        camera_info->enable_stream(false);
        break;
    case GIGE_CAMERA_STATE_IDLE:
    default:
        break;
    }
}

/*!
 * @brief Catch the streamer's streaming signal.
 */
void CameraMDI::streamer_is_streaming()
{
    // Sanity Check.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }
    if (camera_info == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CAMERA_INFO_SANITY_FAILURE;
        return;
    }

    // The streaming has started.
    camera_ptr->stream_running = true;

    // Update our UI element.
    camera_info->enable_stream(true);

    // Pass on the event.
    emit video_is_streaming();
}

/*!
 * @brief Catch the close of the sync-in frame time-out warning progress dialog and
 * provides an interface to close the dialog if sync-in signal resumes.
 */
void CameraMDI::sync_in_warning_closed()
{
    // Make sure to reset our sync-in warning time to empty state.
    QDateTime empty_date_time;

    // Are we receiving our pulses now?
    if (is_sync_in_enabled())
    {
        // If receiving the pulse reset the time collector.
        if (is_sync_in_receiving_pulse())
        {
            // Reset the warning date/time invalid
            // Setting to empty_date_time returns false for time_of_sync_in_suspension.isValid() calls.
            time_of_sync_in_suspension   = empty_date_time;
        }
        // the else has nothing to do here...
    }

    // Do we have a warning dialog to cleanup?
    if (sync_in_warning_dlg)
    {
#if LOG_SYNC_IN_SUSPEND_INFO
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Recycling sync_in_warning_dlg.";
#endif
        // Disconnect any connection made.
        disconnect(sync_in_warning_dlg.get(), &fiProgressDialog::progressCancelled, this, &CameraMDI::sync_in_warning_closed);

        // Recycle the sync in warning dialog to the system.
        fiProgressDialog::RecycleProgressDialog(sync_in_warning_dlg);
    }
}

/*!
 * @brief The camera wants to know when a user disables arm-in external I/O.
 */
void CameraMDI::do_ext_arm_in_disabling()
{
    // This only makes sense if camera is FasFireLR enabled.
    uint32_t fflr_type;
    if (is_fflr_enabled(fflr_type))
    {
        if (save_video)
        {
            // If autosave is set up and we are in live-view then need to stop autosave set up.
            if (save_video->autosave_is_enabled() && save_video->autosave_is_configured())
            {
                // We only do this if we are in live-view.
                uint32_t camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
                if (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Configured for Autosave, stopping Autosave now.";
                    save_video->autosave_stop();
                }
            }
        }
    }
    emit ext_arm_in_disabling();
}

/*!
 * @brief The camera wants to know when a user is attempting to enable arm-in external I/O.
 */
void CameraMDI::do_ext_arm_in_enabling()
{
    // Make sure camera is prepared for arming if FasFireLR is enabled.
    fflr_check_before_arming();
    emit ext_arm_in_enabling();
}

/*!
 * @brief Manage the time trigger timer.
 * @param time_value Number of seconds between trigger events.
 */
void CameraMDI::time_trigger(int time_value)
{
    // Save the base value.
    tt_time = time_value;

    // Are we being asked to start or stop?
    if (time_value > 0)
    {
        // Stop our timer and restart it with the new time.
        // Don't forget to convert from seconds to milliseconds.
        tt_timer.start(time_value * 1000);
    }
    else
    {
        // Turn off the time trigger timer.
        tt_timer.stop();
    }
}

void CameraMDI::time_trigger_expired()
{
    // Increment our counter.
    ++tt_count;

    // Are we armed?
    if (read_camera_cache(GIGE_REG_CAMERA_STATE) == GIGE_CAMERA_STATE_ARMED)
        do_camera_trigger();

    // Let folks know.
    emit tt_expired();
}

/*!
 * @brief The FMP socket is ready and will start receiving messages. Make sure we
 *        have our connections setup to handle them.
 */
void CameraMDI::unsol_socket_is_ready()
{
    FI_comm     *comm_ptr;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // Don't want anybody else messages to come through.
    disconnect(comm_ptr, &FI_comm::unsol_socket_ready, this, &CameraMDI::unsol_socket_is_ready);
    connect(comm_ptr, &FI_comm::unsol_msg_received, this, &CameraMDI::handle_unsol_msg);

    // Connect our streamer's signals to our slots.
    connect(streamer, &CameraStream::detached, this, &CameraMDI::streamer_is_closed);
    connect(streamer, &CameraStream::fault, this, &CameraMDI::streamer_faulted);
    connect(streamer, &CameraStream::image_received, this, &CameraMDI::streamer_image_received, Qt::QueuedConnection);
    connect(streamer, &CameraStream::paused, this, &CameraMDI::streamer_is_paused);
    connect(streamer, &CameraStream::ready, this, &CameraMDI::streamer_is_ready);
    connect(streamer, &CameraStream::streaming, this, &CameraMDI::streamer_is_streaming);
    connect(streamer, &CameraStream::timeout, this, &CameraMDI::streamer_image_timeout);

    // Attach our streamer object.
    streamer->attach();
}

/*!
 * @brief Update the camera specific menu based on the current state of the application and/or camera.
 */
void CameraMDI::update_camera_menu()
{
    uint32_t    reg_value;

    // Get the autosave register.
    reg_value = read_camera_cache(GIGE_REG_AUTOSAVE);

    // Update the autosave menu.
    if ((reg_value & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE)
    {
        // Switch the menu text so that the user know it's a disable next time.
        MainWindow::GetInstance()->get_cam_autosave_setup()->setText(tr("Autosave Settings..."));
        MainWindow::GetInstance()->get_cam_autosave_enable()->setText(tr("Autosave Disable"));
    }
    else
    {
        // Switch the menu text so that the user know it's an enable next time.
        MainWindow::GetInstance()->get_cam_autosave_setup()->setText(tr("Autosave Setup..."));
        MainWindow::GetInstance()->get_cam_autosave_enable()->setText(tr("Autosave Enable"));
    }
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that it need the per-frame
 *        metadata to be updated.
 *
 * @sa CameraReview::play_update()
 */
void CameraMDI::update_pfmd()
{
    // Show whatever information should be visible.
    camera_info->show_info();

    // Do we have the metadata for this frame cached?
    if (pfmd_info->retrieve(review_controls->get_current_frame_num()) != nullptr)
    {
        // We already have the metadata. The read_frame_metadata() is handy to use
        // but it is expecting the frame information in active_frame_requests.
        active_frame_requests.emplaceFront(review_controls->get_current_frame_num());
        read_frame_metadata();
    }
    else
    {
        // Get the metadata for the current frame.
        request_pfmd(review_controls->get_current_frame_num());
    }
}

/*!
 * @brief The [record progress object](@ref fiProgBar) has indicated that it's time for an update.
 *
 * @sa fiProgBar::time_to_update()
 */
void CameraMDI::update_rec_progress()
{
    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Get the current frame position.
    cam_register_group->addReg(GIGE_REG_ACTIVE_PART);
    cam_register_group->addReg(GIGE_REG_USER_SAVE_START_NUM);
    cam_register_group->addReg(GIGE_REG_USER_SAVE_END_NUM);
    cam_register_group->addReg(GIGE_REG_AUTOSAVE);
    cam_register_group->addReg(GIGE_REG_TRIGGER_REQ_POS);
    cam_register_group->addReg(GIGE_REG_CUR_CAPTURE_COUNT);
    cam_register_group->lastReg(GIGE_REG_CUR_FRAME_POS);

    // Fire off the reads.
    read_camera_registers(cam_register_group);
}

/*!
 * @brief The user has pressed the stream pause/resume button.
 *
 * @sa CameraInfo::stream_toggle()
 */
void CameraMDI::user_stream_request()
{
    // Sanity Checks.
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }
    if (camera_info == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << CAMERA_INFO_SANITY_FAILURE;
    }
    if (streamer == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << STREAMER_SANITY_FAILURE;
    }

    // Update the stream enable button.
    if (camera_info)
    {
        camera_info->enable_stream(true);
    }

    // User wants to toggle the streaming state.
    if (streamer && !streamer->is_paused())
    {
        // We are currently streaming, so the user wants to pause the stream.
        pause_video_request(__PRETTY_FUNCTION__);
    }
    else if (camera_ptr && streamer && camera_ptr->stream && streamer->is_paused())
    {
        uint32_t    reg_value;

        // Get the camera state and update our stream accordingly.
        reg_value = read_camera_cache(GIGE_REG_CAMERA_STATE);
        stream_setup(reg_value, __PRETTY_FUNCTION__);
    }
    else if (camera_ptr && !camera_ptr->stream)
    {
        // We must have started off by not streaming from the camera. The user
        // has clicked on the STREAM button, so we must want to start streaming
        // now.
        camera_ptr->stream = true;
        camera_ptr->stream_running = false;
        resume_video_request();
    }
}

void CameraMDI::set_vt_area(int new_area, bool send_to_camera)
{
    // Is there a change?
    if (vt_area == new_area)
        return;

    // Save the new area setting.
    vt_area = new_area;
    if (!send_to_camera)
        return;

    // Are we connect to a camera that supports video trigger?
    if (sensor_ptr && sensor_ptr->supports_video_trigger())
    {
        QJsonObject json_obj;

        // Pass on the new information.
        json_obj.insert("area_thresh", new_area);

        QJsonDocument ap_doc(json_obj);
        write_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON, ap_doc.toJson());
    }
}

void CameraMDI::set_vt_change(int new_change, bool send_to_camera)
{
    // Is there a change?
    if (vt_change == new_change)
        return;

    // Save the new pixel change value.
    vt_change = new_change;
    if (!send_to_camera)
        return;

    // Are we connect to a camera that supports video trigger?
    if (sensor_ptr && sensor_ptr->supports_video_trigger())
    {
        QJsonObject json_obj;

        // Pass on the new information.
        json_obj.insert("pix_thresh", new_change);

        QJsonDocument ap_doc(json_obj);
        write_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON, ap_doc.toJson());
    }
}

void CameraMDI::set_vt_mark_only(bool only_mark, bool send_to_camera)
{
    // Is there a change?
    if (vt_mark_only == only_mark)
        return;

    // Save the marker only flag.
    vt_mark_only = only_mark;
    if (!send_to_camera)
        return;

    // Are we connect to a camera that supports video trigger?
    if (sensor_ptr && sensor_ptr->supports_video_trigger())
    {
        int control_setting;
        QJsonObject json_obj;

        // Pass on the new information.
        control_setting = GIGE_VTRIG_CONTROL_ENABLE;
        if (!only_mark)
            control_setting |= GIGE_VTRIG_CONTROL_TRIGGER;
        json_obj.insert("control", control_setting);

        QJsonDocument ap_doc(json_obj);
        write_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON, ap_doc.toJson());
    }
}

void CameraMDI::set_vt_refresh(int new_refresh, bool send_to_camera)
{
    // Is there a change?
    if (vt_refresh == new_refresh)
        return;

    // Save the new baseline refresh value.
    vt_refresh = new_refresh;
    if (!send_to_camera)
        return;

    // Are we connect to a camera that supports video trigger?
    if (sensor_ptr && sensor_ptr->supports_video_trigger())
    {
        QJsonObject json_obj;

        // Pass on the new information.
        json_obj.insert("interval", new_refresh);

        QJsonDocument ap_doc(json_obj);
        write_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON, ap_doc.toJson());
    }
}

void CameraMDI::vt_changed()
{
    // Are we connect to a camera that supports video trigger?
    if (sensor_ptr && sensor_ptr->supports_video_trigger())
    {
        QRect trig_rect;
        QJsonObject json_obj;

        // Get our area of interest.
        trig_rect = video_player->get_vt_rect();

        // Make sure there is a valid reticle.
        if (trig_rect.isEmpty())
        {
            vt_reticle_reset();
        }

        // Pull our area of interest after reset.
        trig_rect = video_player->get_vt_rect();

        // Make sure there is a valid reticle.
        if (trig_rect.isEmpty())
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to pull valid vt_rect after vt_reticle_reset().";
            return;
        }

        // Make absolute certain that the trigger area is entirely within the
        // image area.
        if (video_player->get_image_rect().contains(trig_rect))
        {
            // Build our JSON value and send it on to the camera.
            json_obj.insert("roi_x", (int)trig_rect.left());
            json_obj.insert("roi_y", (int)trig_rect.top());
            json_obj.insert("roi_width", (int)trig_rect.width());
            json_obj.insert("roi_height", (int)trig_rect.height());

            QJsonDocument ap_doc(json_obj);
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << json_obj;
            write_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON, ap_doc.toJson());

            // Toggle the reticle logic to reset things.
            vt_reticle_show(false);
            vt_reticle_show(true);
        }
        else
        {
            video_player->get_vt_reticle()->setEnabledBrush(QBrush(QColor(0xff, 0x00, 0xff, 0x40), Qt::CrossPattern));
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "trig_rect:" << trig_rect << "is not contained in the image:" << video_player->get_image_rect();
        }
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "VT not supported?";
    }
}

void CameraMDI::vt_track()
{
    QRect vt_rect;

    // Make sure we are tracking. We might get called as a result of simply
    // looking at the histogram. Don't want to execute in that case.
    if (!vt_reticle_visible())
        return;

    // Process our state machine.
    switch (vt_track_state)
    {
    default:
    case VT_INIT_TRACKING:
        // Reset our tracking variables.
        vt_ready = false;
        vt_histo_count = 0;

        // Get the rectangle the user wants for the video trigger window.
        vt_rect = video_player->get_vt_rect();

        // It is possible to get an empty rectangle.
        if (vt_rect.isEmpty())
        {
            // Not sure what to set for a default rectangle here, so just return
            // and early terminate the VT_INIT_TRACKING.
            QLOG_WARN() << __PRETTY_FUNCTION__ << "The video_player returned an empty video trigger rectangle!";
            return;
        }

        // Tell the player where we want the histogram.
        video_player->get_histogram()->set_roi(vt_rect);

        // Connect our histogram ready signal.
        connect(video_player->get_histogram(), &fiHistoProc::ready, this, &CameraMDI::vt_track);

        // Tell the player that we want histogram events.
        video_player->enable_histogram(true);

        // Override "normal" histogram.
        MainWindow::GetInstance()->get_histo_dock()->override_roi(vt_rect);

        // Make sure our reticle is setup properly.
        video_player->get_vt_reticle()->setEnabledPen(QPen(Qt::white, 3));
        video_player->get_vt_reticle()->setEnabledBrush(QBrush(Qt::NoBrush));

        // We are starting, so size the array and initialize it.
        vt_baseline.resize(vt_rect.width() * vt_rect.height());
        vt_baseline.fill(0);

        // We are initializing our baseline reading.
        vt_track_state = VT_INIT_BASELINE;
        break;
    case VT_INIT_BASELINE:
        // Initialize our baseline data.
        vt_baseline_calc();
        break;
    case VT_PROCESS:
        // Process the histogram to see if we need to trigger.
        vt_process();
        break;
    }
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the current
 *        frame has changed.
 *
 * @sa CameraReview::current_frame_changed()
 */
void CameraMDI::video_current_frame_changed()
{
    uint32_t        camera_state;
    uint32_t        cur_frame;

//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "streamer->is_paused():" << streamer->is_paused();

    // Don't single step if we don't want streaming.
    if (camera_ptr == nullptr)
        return;
    if (!camera_ptr->stream)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No stream.";
        return;
    }

    // Are we in the middle of a save operation?
    if (save_video && save_video->save_in_progress())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Save video in progress.";
        return;
    }

    // Ignore anything other than PAUSE or SINGLE FRAME events.
    if (streamer && !streamer->is_paused() && !streamer->is_stepping())
    {
        // TICKET #353 when playing through a capture this condition ends up returning here when playback
        // is occurring.  We can't just exit because frames are being fetched and displayed and we get
        // into a state where our last_requested_frame could become stale.  If we don't modify it and once
        // we stop playing and the user hits the jump to trigger frame then the stale "last_requested_frame"
        // happens might be the trigger frame and we will NOT catch the need to refresh to the current frame.
        // Which is what was happening in the ticket.  See its notes for more specifics.
        //
        // The best thing to do here is set the last_requested_frame to its empty value.  We could set it
        // to the current frame but maybe it's safer to set it to an initialized not set state.
        if (review_controls->is_playing())
        {
            last_requested_frame = (uint32_t) -1;   // Reset it to an initialized not-set state.
            //last_requested_frame = review_controls->get_current_frame_num();
        }
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "is_paused:" << streamer->is_paused() << "is_stepping:" << streamer->is_stepping();
        return;
    }

    // Sanity check the camera state.
    camera_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
    if (camera_state != GIGE_CAMERA_STATE_REVIEW)
    {
        QLatin1Char     zero_fill('0');
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << QString("Camera state [0x%1] is not review!").arg(camera_state, 2, 16, zero_fill);
        return;
    }

    // Are we in auto-play? This is often used with "DVR" mode too.
    if (autoplay() && review_controls->is_playing())
    {
        // We're playing in auto playback mode. Nothing to see here.
        return;
    }

    // Get the "new" current frame.
    cur_frame = review_controls->get_current_frame_num();

    // Have the stream machine request the frame, if it hasn't already been requested.
    if (force_frame_update || (last_requested_frame != cur_frame))
    {
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requesting cur_frame refresh:" << cur_frame << "force_frame_update:" << force_frame_update << "last_requested_frame:" << last_requested_frame;
        // Ask the histogram to update. This will prepare the histogram engine for when
        // we actually move a frame through it.
        MainWindow::GetInstance()->push_one_histogram();

        // Remember this frame so we don't double up on the same one.
        last_requested_frame = cur_frame;
        force_frame_update = false;

        // Ask the state machine to stream this frame.
        if (streamer)
        {
            streamer->request_frame(cur_frame);
        }

        // Make sure we re-zoom the image.
        if (video_player)
        {
            video_player->need_rezoom();
        }

        // Are we showing per-frame metadata or do we have overlays?
        if (camera_info && video_player && (camera_info->pfmd_is_enabled() || video_player->has_overlay()))
        {
            // Request the metadata. Delay it a little to allow for the frame request to go through.
            QTimer::singleShot(METADATA_STREAM_DELAY, this, [ = ] { this->request_pfmd(this->last_requested_frame); });
        }
    }
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to jump
 *        to a particular time/frame in the recorded video.
 *
 * @sa CameraReview::jump_time()
 */
void CameraMDI::video_jump_time()
{
    uint32_t    record_rate;
    QString     file_name;
    JumpDialog  jump_dlg(this);
    PlaybackMetaInfo meta_info(this);

    // Do we have the raw metadata?
    if (raw_metadata.isEmpty())
    {
        // Remember that we're downloading the metadata not to save, but to use.
        downloading_pfmd = true;

        // Connect our slot to catch the download complete signal.
        connect(this, &CameraMDI::raw_metadata_received, this, &CameraMDI::video_jump_time, Qt::UniqueConnection);

        // Start downloading the metadata.
        stream_metadata();
        return;
    }

    // Disconnect our slot.
    disconnect(this, &CameraMDI::raw_metadata_received, this, &CameraMDI::video_jump_time);

    // Do we have a progress dialog up?  If so reset it.
    if (progressDialog)
    {
        // Recycle any resources and release the progress dialog.
        fiProgressDialog::RecycleProgressDialog(progressDialog);
    }

    // For the jump dialog, we need to save the raw metadata in a file.
    file_name = save_raw_metadata();
    if (file_name.isEmpty())
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to create raw metadata file.";
        return;
    }

    // Initialize our varaibles.
    trigger_frame = read_camera_cache(GIGE_REG_PART_TRIG_FRAME);
    record_rate = read_camera_cache(GIGE_REG_FRAME_RATE);

    // Set the raw metadata file information into the perframe metadata object.
    meta_info.setRawMetadataFile(file_name, get_pfmd_len());

    // How many frames are we dealing with?
    // meta_info.set_frame_count( raw_metadata_len / get_pfmd_len() );

    // Get the current frame, trigger frame and frame rate so we can
    // calculate the relative time of the current frame.
    jump_dlg.set_pfmeta(&meta_info);
    jump_dlg.set_rel_time((int)review_controls->get_clip_start(), (int)review_controls->get_clip_end(), (int)review_controls->get_current_frame_num(), (int)trigger_frame, (int)ClientMath::reg_to_frame_rate(record_rate));
    jump_dlg.use_rel_time();

    // Pop up the dialog.
    if (jump_dlg.exec() == QDialog::Accepted)
    {
        // Go to the selected frame.
        review_controls->set_current_frame_num(jump_dlg.get_frame());
    }
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to playback
 *        the recorded video in the normal (i.e. "forward") direction.
 *
 * @sa CameraReview::play_forward()
 */
void CameraMDI::video_play_forward()
{
    // Play the video going forward.
    play_video(true);
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to playback
 *        the recorded video in the backward (i.e. "reverse") direction.
 *
 * @sa CameraReview::play_reverse()
 */
void CameraMDI::video_play_reverse()
{
    // Play the video going reverse.
    play_video(false);
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to change
 *        the playback rate.
 *
 * @sa CameraReview::playback_rate_changed()
 */
void CameraMDI::video_rate_changed(int new_rate)
{
    Q_UNUSED(new_rate)

    // We only need to update if we're streaming.
    if (streamer->is_playing())
    {
        // Stop the stream.
        pause_video_request(__PRETTY_FUNCTION__);

        // Restart the playback. It'll pick up the new rate.
        play_video(play_video_forward);
    }
}

/*!
 * @brief The [video review object](@ref CameraReview) has indicated that the user wants to stop playback.
 *
 * @sa CameraReview::playback_stop()
 */
void CameraMDI::video_review_stop()
{
    // Check to see if the stream has stopped yet.
    if (camera_ptr && streamer && camera_ptr->stream && !streamer->is_paused())
    {
        // Stop the stream.
        pause_video_request(__PRETTY_FUNCTION__);

        // Setup a one-shot.
        QTimer::singleShot(REVIEW_STOP_TIMEOUT, this, &CameraMDI::video_review_stop_wait);
    }
}

void CameraMDI::video_review_stop_wait()
{
    // Check to see if the stream has paused yet.
    if (camera_ptr && streamer && camera_ptr->stream && !streamer->is_paused())
    {
        // Stop the stream.
        pause_video_request(__PRETTY_FUNCTION__);

        // Setup a one-shot.
        QTimer::singleShot(REVIEW_STOP_TIMEOUT, this, &CameraMDI::video_review_stop_wait);
    }
}

/*!
 * @brief Our window is activating. Make sure we get focus.
 */
void CameraMDI::window_activating()
{
    // Check to see if we have focus.
    if (!hasFocus())
    {
        // We don't have focus, so grab focus.
        setFocus();
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void CameraMDI::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t                write_status = GEV_STATUS_SUCCESS;
    FI_comm                 *comm_ptr;
    QString                 err_msg;
    QString                 reg_name_str;
    QHostAddress            our_addr;

    // Is this event for us?
    if (ip_addr != get_ip_addr())
        return;

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return;
    }

    // This might be useful below.
    our_addr.setAddress(qFromBigEndian(ip_addr));

    // Process the register.
    switch (reg_addr)
    {
    case GIGE_REG_ACTIVE_FRAME_NUM:
        reg_name_str = "GIGE_REG_ACTIVE_FRAME_NUM";
        // Was the write successful?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            uint32_t requested_frame = active_frame_requests.first();

            // Read back the active frame number to cache it.
            if (requested_frame == last_requested_frame)
                start_camera_read_reg(GIGE_REG_ACTIVE_FRAME_NUM);

            // We only need to ready the PFMD if we don't already have the PFMD.
            if (pfmd_info->retrieve(requested_frame) == nullptr)
            {
                // Ask the camera for the metadata.
                start_camera_read_buf(GIGE_REG_FRAME_METADATA, (uint32_t)get_pfmd_len());
            }
            else
            {
                // Act like we've just received the PFMD.
                read_frame_metadata();

                // Any more in the queue?
                if (!active_frame_requests.empty())
                {
                    // Write the next request.
//                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requesting next on the list:" << active_frame_requests;
                    write_camera_register(GIGE_REG_ACTIVE_FRAME_NUM, active_frame_requests.first());
                }
            }
        }
        break;
    case GIGE_REG_FRAME_PFMD:
        reg_name_str = "GIGE_REG_FRAME_PFMD";
        // Was the write successful?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Ask the camera for the metadata.
            start_camera_read_buf(GIGE_REG_FRAME_PFMD, (uint32_t)get_pfmd_len());
        }
        break;
    case GIGE_REG_AUTOSAVE:
        reg_name_str = "GIGE_REG_AUTOSAVE";
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_AUTOSAVE write_status:" << write_status << "release_all:" << release_all;
        // Are we in the middle of a release of partitions?
        if (release_all)
        {
            // Carry on.
            free_partition(true);
        }
        break;
    case GIGE_REG_CCP:
        reg_name_str = "GIGE_REG_CCP";
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        process_ccp_write(write_status);
        break;
    case GIGE_REG_CHANGE_CAMERA_STATE:
        reg_name_str = "GIGE_REG_CHANGE_CAMERA_STATE";
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_ERROR)
        {
            uint32_t    current_state;
            uint32_t    ltr_record_mode;
            bool        ltr_append_mode = false;
            QString     text_msg;
            QString     info_msg;

            // Figure out if LTR mode is enabled or not.
            if (is_ltr_enabled(ltr_record_mode))
            {
                ltr_append_mode = ((ltr_record_mode & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0);
            }
            // We get this error in LTR append mode when the disk is full.
            if (ltr_append_mode)
            {
                // Make the error code something appropriate.
                int err_code = ENOSPC;

                // Build up our error strings for display to the user.
                text_msg = QStringLiteral("The camera reported an error arming the camera.");
                info_msg = QStringLiteral("The error code is %1 (%2)").arg(err_code).arg(strerror(err_code));
            }
            else
            {
                // Generic message for the error.
                text_msg = QStringLiteral("The camera reported an error.");
                switch (requested_state)
                {
                case GIGE_CAMERA_STATE_ARMED:
                    info_msg = QStringLiteral("The camera was unable to arm the camera.");
                    break;
                case GIGE_CAMERA_STATE_LIVE_VIDEO:
                    info_msg = QStringLiteral("The camera was unable to switch the camera to Live View.");
                    break;
                default:
                    info_msg = QStringLiteral("The camera was unable to switch the camera to state %1.").arg(requested_state);
                    break;
                }
            }

            // Report the error to the user.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << text_msg << info_msg;
            fiMsgBox::critical(this, text_msg, info_msg);

            // Handle the error.
            current_state = read_camera_cache(GIGE_REG_CAMERA_STATE);
            process_state_change(current_state, true);
        }
        break;
    case GIGE_REG_CONFIGURATION:
        reg_name_str = "GIGE_REG_CONFIGURATION";
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Go write the next chunk, if there is one.
            push_config_file_next();
        }
        else
        {
            err_msg = QString::fromUtf8("Unable to write GIGE_REG_CONFIGURATION to IP adddress %1").arg(our_addr.toString());
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }
        break;
    case GIGE_REG_CS_REGS:
        reg_name_str = "GIGE_REG_CS_REGS";
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Are we here for reading or writing the configuration. We can tell,
            // because when we're writing the configuration the camera will be rebooting.
            if (camera_ptr && camera_ptr->rebooting)
            {
                // Go offline.
                set_offline(true);
            }
            else
            {
                // Find out how big the configuration file is.
                comm_ptr->start_read_reg(get_ip_addr(), GIGE_REG_CONFIG_SIZE);
            }
        }
        else
        {
            err_msg = QString::fromUtf8("Unable to write GIGE_REG_CS_REGS to IP adddress %2").arg(our_addr.toString());
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }
        break;
    case GIGE_REG_HEARTBEAT:
        reg_name_str = "GIGE_REG_HEARTBEAT";
        // Was it a successful write?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Tell the camera we're interested in unsolicited messages.
            if (camera_ptr)
            {
                comm_ptr->start_unsolicited(camera_ptr->ip_addr);
            }
        }
        else
        {
            err_msg = QString::fromUtf8("Unable to write GIGE_REG_HEARTBEAT to IP adddress %2").arg(our_addr.toString());
            QLOG_ERROR() << __PRETTY_FUNCTION__ << err_msg;
        }
        break;
    case GIGE_REG_JPEG_QUALITY:
        reg_name_str = "GIGE_REG_JPEG_QUALITY";
        // Start a read to update the cache.
        comm_ptr->start_read_reg(get_ip_addr(), GIGE_REG_JPEG_QUALITY);
        break;
    case GIGE_REG_MEDIA_JSON:
        reg_name_str = "GIGE_REG_MEDIA_JSON";
        // Was the write successful?
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        if (write_status == GEV_STATUS_SUCCESS)
        {
            // Now ask the camera to give us a fresh JSON string.
            start_camera_read_buf(GIGE_REG_MEDIA_JSON, GIGE_MEDIA_JSON_MAX_LEN);
        }
        break;
    case GIGE_REG_RESET_CAMERA:
    case GIGE_REG_RESET_TO_FACTORY:
        reg_name_str = "GIGE_REG_RESET*";
        // Go offline.
        if (camera_ptr)
        {
            set_offline(camera_ptr->rebooting);
        }
        break;

    case GIGE_REG_TRUNCATE:
        reg_name_str = "GIGE_REG_TRUNCATE";
        // Make sure this is a result of truncating.
        if (in_truncate)
        {
            // Not truncating any more.
            in_truncate = false;

            // Switch back to review.
            do_camera_review();
        }
        break;

    case GIGE_REG_UPDATE_METADATA:
    {
        reg_name_str = "GIGE_REG_UPDATE_METADATA";
//        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_UPDATE_METADATA write complete!";
        QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
        if (cam_register_group.isNull())
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
            break;
        }

        // The camera should have updated the active partition's metadata.  Now issue
        // an update of our camera cache for this information.
        cam_register_group->addReg(GIGE_REG_USER_SAVE_START_NUM);
        cam_register_group->addReg(GIGE_REG_USER_SAVE_END_NUM);
        cam_register_group->addReg(GIGE_REG_PART_NUM_FRAMES);
        cam_register_group->lastReg(GIGE_REG_PART_TRIG_FRAME);

        // Fire off a read now that this information has been updated.
        read_camera_registers(cam_register_group);
        break;
    }
    case GIGE_REG_USER_DEF_NAME:
        reg_name_str = "GIGE_REG_USER_DEF_NAME";
        // Sanity check.
        if (camera_ptr != nullptr)
        {
            memset(camera_ptr->user_name, 0, FI_COMM::USER_NAME_LEN);
            strncpy(camera_ptr->user_name, get_camera_name().toUtf8().constData(), FI_COMM::USER_NAME_LEN - 1);
        }
        break;

    case GIGE_REG_ACTIVE_METADATA:
        reg_name_str = "GIGE_REG_ACTIVE_METADATA";
        // Make sure we are waiting for this data.
        if (active_metadata_requested)
        {
#if LOG_ACTIVE_METADATA_DOWNLOAD
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_ACTIVE_METADATA" << "active_metadata_requested:" << active_metadata_requested;
#endif
            // We have received our write complete we were waiting for and can reset it.
            // Note: Because CameraMDI::write_complete connections are made with Qt::QueuedConnection
            // we do not have to protect access to the active_metadata_requested data field with a mutex.
            active_metadata_requested = false;

            // Start the block download for the pm_metadata from the camera.
            block_download_start(GIGE_REG_ACTIVE_METADATA, sizeof(pm_metadata_type), GIGE_MAX_METADATA_BLOCK);
        }
        break;

    default:
        break;
    }

#if LOG_WRITE_COMPLETE_ERRORS
    if (write_status != GEV_STATUS_SUCCESS)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "WRITE STATUS" << reg_name_str << "========>" << FI_comm::gev_status_to_msg(write_status);
    }
#endif
}
