/*!
 * @file panel_reccontrols.cpp
 * @brief This file contains the implementation of the PanelRecControls
 *        class. This class handles the record controls for the camera.
 *
 * Copyright 2013 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "panel_reccontrols.h"
#include "ui_panel_reccontrols.h"

#include <QAction>
#include <QFocusEvent>
#include <QFont>
#include <QIcon>
#include <QMessageBox>
#include <QObject>
#include <QTime>
#include <QTimer>

#include "camera_video.h"
#include "client_math.h"
#include "control_panel.h"
#include "mainwindow.h"
#include "panel_recsettings.h"

namespace REC_CONTROLS
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Log messages for testing purposes.  Set to 1 to have them compiled for output.
#define LOG_GROUP_READS             0
#define LOG_MASTER_FRAME_RATE       0
#define LOG_SYNC_OUT_CHANNEL        0

//! An enumeration to define the various types of external triggers we support.
enum EXT_TRIG_TYPES
{
    EXT_TRIG_RISING,                    //!< The external trigger is enabled and active on the rising edge of the incoming signal.
    EXT_TRIG_FALLING,                   //!< The external trigger is enabled and active on the falling edge of the incoming signal.
    EXT_TRIG_DISABLED                   //!< The external trigger is disabled.
};

//! An enumeration for the tab positions of the I/O controls. It is important that
//! the order of these values agree with the order of the tabs in the .UI file.
enum TAB_POSITION
{
    TAB_TRIG_IN,                        //!< The trigger input control tab.
    TAB_TRIG_OUT,                       //!< The trigger output control tab.
    TAB_SYNC_IN,                        //!< The Sync-In control tab.
    TAB_SYNC_OUT,                       //!< The Sync-Out control tab.
    TAB_ARM_IN,                         //!< The ARM-In control tab.
    TAB_ARM_OUT                         //!< The ARM-Out control tab.
};

// The pulse width should go from 1..8 milliseconds
const int       NO_PULSE            = 0;            //!< No pulse specified.
const int       MIN_PW_VALUE        = 1;            //!< The minimum pulse width, in milliseconds.
const int       MAX_PW_VALUE        = 8;            //!< The maximum pulse width, in milliseconds.
const int       DEFAULT_PW_VALUE    = 2;            //!< The default pulse widht, in milliseconds.

const int       MIN_MASTER_FRAMERATE    = 20;       //!< Our minimum supported "master" frame rate.
const int       MAX_MASTER_FRAMERATE    = 250000;   //!< Our maximum supported "master" frame rate.

const int       MIN_CALC_FRAMERATE      = 2;        //!< Our minimum lowest "calculated" frame rate.

const int       MIN_DIVISOR_VALUE   = 1;            //!< Our minimum frequency divisor value.
const int       MAX_DIVISOR_VALUE   = 8;            //!< Our maximum frequency divisor value.

const int       MIN_DUTY_CYCLE      = 1;            //!< Our minimum duty cycle (in percent).
const int       MAX_DUTY_CYCLE      = 99;           //!< Our maximum duty cycle (in percent).
const int       DEFAULT_DUTY_CYCLE  = 50;           //!< Our default duty cycle (in percent).

const int       MAX_CHANNEL         = 3;            //!< We have a maximum of 3 channels.
const uint32_t  MAX_CHANNEL_MASK    = ((1 << MAX_CHANNEL) - 1);     //!< Our channel bitmask.

const int       MAX_RETRY_COUNT     = 50;           //!< The maximum number of times we'll try to update a register value.
const int       RETRY_DELAY         = 100;          //!< The number of milliseconds to wait before retrying.

const int       EXTIO_MAX_READ_COUNT = 10;          //!< Maximum number of ticks we will wait before retrying the read.
const int       EXTIO_ACT_POLL_INTERVAL = 250;      //!< Provides delay allowing human recognition of external I/O activity.

const int       VT_INFO_DELAY       = 1000;         //!< Delay before retrying the video trigger info.
const bool      VERBOSE_LOGGING     = false;        //!< Set to true to allow for verbose logging when diagnosing issues.
const bool      LOG_SYNC_OUT_TRACES = false;        //!< Set this to true if want sync-out specific log tracing.

const int       MAX_SYNC_OUT_DATA_ENTRIES   = 12;   //!< The maximum number of SYNC_OUT_DATA entries we'll support.

//! @note If we are testing the QSettings mechanism for sync-out we want to use the old logic
//! and not set the rest of the register bits.  But the theory is that if it doesn't make
//! any difference to camera_server code to have a cleared disable bit and other register
//! settings then we don't have to use QSettings as the register bits will persist and we
//! will get the proper settings restored in the panel.  This relies on the boolean in the global
//! scope to turn this on and off.  If we are actively using the QSettings for Sync Out then
//! we will maintain the old logic here.  If we want to remove the QSettings pathway for
//! testing the theory the USE_SYNC_OUT_QSETTINGS should be false and we will end up setting
//! the other bits in the register.
const bool      USE_SYNC_OUT_QSETTINGS      = true;

//! The sync-out data saved for settings preservation.
struct SYNC_OUT_DATA
{
    uint32_t ip_addr;
    uint32_t sync_out_reg;
    QTime last_accessed;
};

/*!
 * @brief The SyncOutDataRepository class is used to abstract the storage and retrieval
 *        of the "sync-out" register settings for a camera.
 *
 * We wanted to be able to save and restore the sync-out data based on the IP address
 * of the camera.
 */

class SyncOutDataRepository
{
private:
    SyncOutDataRepository() {}

public:

    //! Don't allow this object to be cloned.
    SyncOutDataRepository(SyncOutDataRepository& other) = delete;

    //! Not assignable.
    void operator=(const SyncOutDataRepository&) = delete;

    /*!
     * @brief Find the sync-out data associated with the cam_ip_addr.
     * @param cam_ip_addr   The IP address we're interested in.
     * @return True if the sync_out_reg_found contains a valid register.
     */
    static bool loadSyncOutData(uint32_t cam_ip_addr, uint32_t& sync_out_reg_found)
    {
        int     i;
        int     found_index = -1;
        bool    sync_out_data_found = false;      // Assume it isn't in the sync-out settings database.
        QList<SYNC_OUT_DATA> sync_out_data_list;

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "cam_ip_addr:" << Qt::showbase << Qt::hex << cam_ip_addr;

        // Save the current settings.
        QSettings settings;

        settings.beginGroup(PanelRecControls::RECORD_CONTROL_QSETTINGS_GROUP);

        // Load the existing array.
        int list_size = settings.beginReadArray(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA_LIST);
        for (i = 0; i < list_size; ++i)
        {
            SYNC_OUT_DATA entry;

            // Let the system know which index we're working with.
            settings.setArrayIndex(i);

            // Read this entry.
            entry.ip_addr = settings.value(PanelRecControls::RECORD_CONTROL_SYNC_OUT_IP).toUInt();
            entry.sync_out_reg = settings.value(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA).toUInt();
            entry.last_accessed = settings.value(PanelRecControls::RECORD_CONTROL_SYNC_OUT_TIME).toTime();

            // Put it on the list.
            sync_out_data_list.append(entry);

            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "i:" << i << "entry.ip_addr:" << Qt::showbase << Qt::hex << entry.ip_addr;

            // Is this a match to our IP address?
            if (entry.ip_addr == cam_ip_addr)
            {
                // Remember this entry
                found_index = sync_out_data_list.size() - 1;
                //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "sync_out_data_list index:" << found_index;
            }
        }
        settings.endArray();

        // Did we find a result?
        if (found_index >= 0)
        {
            // Install the found setting in the reference passed in.
            sync_out_reg_found = sync_out_data_list.at(found_index).sync_out_reg;

            // It was found in the settings database.
            sync_out_data_found = true;

            // Update this entrys access time.
            sync_out_data_list[found_index].last_accessed = QTime::currentTime();

            // Save off the list.
            // NOTE: This might be overkill, re-writing the entire list. I am just not
            //       certain how growing the list would affect the writing. I know this
            //       will work, and I don't have time for optimization at this point.
            settings.beginWriteArray(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA_LIST);
            for (i = 0; i < sync_out_data_list.size(); ++i)
            {
                settings.setArrayIndex(i);
                settings.setValue(PanelRecControls::RECORD_CONTROL_SYNC_OUT_IP, sync_out_data_list.at(i).ip_addr);
                settings.setValue(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA, sync_out_data_list.at(i).sync_out_reg);
                settings.setValue(PanelRecControls::RECORD_CONTROL_SYNC_OUT_TIME, sync_out_data_list.at(i).last_accessed);
            }
            settings.endArray();
        }
        settings.endGroup();

        // Return whether or not the data was found in the settings database.
        return sync_out_data_found;
    }

    /*!
     * @brief Save the sync-out register to our settings database.
     * @param cam_ip_addr           The IP address to associate the download_base_path with.
     * @param sync_out_reg          The new sync-out register settings for active camera.
     */
    static void saveSyncOutData(uint32_t cam_ip_addr, uint32_t sync_out_reg_to_save)
    {
        int i;
        int found_index = -1;
        QList<SYNC_OUT_DATA> sync_out_data_list;

        // Save the current settings.
        QSettings settings;
        settings.beginGroup(PanelRecControls::RECORD_CONTROL_QSETTINGS_GROUP);

        // Load the existing array.
        int list_size = settings.beginReadArray(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA_LIST);
        for (i = 0; i < list_size; ++i)
        {
            SYNC_OUT_DATA entry;

            // Let the system know which index we're working with.
            settings.setArrayIndex(i);

            // Read this entry.
            entry.ip_addr = settings.value(PanelRecControls::RECORD_CONTROL_SYNC_OUT_IP).toUInt();
            entry.sync_out_reg = settings.value(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA).toUInt();
            entry.last_accessed = settings.value(PanelRecControls::RECORD_CONTROL_SYNC_OUT_TIME).toTime();

            // Put it on the list.
            sync_out_data_list.append(entry);

            // Is this a match to our IP address?
            if (entry.ip_addr == cam_ip_addr)
                found_index = i;
        }
        settings.endArray();

        // Did we find our IP address in the list?
        if (found_index >= 0)
        {
            // Modify the entiry.
            sync_out_data_list[found_index].sync_out_reg = sync_out_reg_to_save;
            sync_out_data_list[found_index].last_accessed = QTime::currentTime();
        }

        // It wasn't already in the list, so can we add it?
        else if (sync_out_data_list.size() < MAX_SYNC_OUT_DATA_ENTRIES)
        {
            SYNC_OUT_DATA entry;

            // Create a new entry to append to the list.
            entry.ip_addr = cam_ip_addr;
            entry.sync_out_reg = sync_out_reg_to_save;
            entry.last_accessed = QTime::currentTime();
            sync_out_data_list.append(entry);
        }

        // Find out which entry in the list we can replace.
        else
        {
            int oldest_entry;
            QTime oldest_time;

            // Read up the index of the oldest entry. Default to 0.
            oldest_time = QTime::currentTime();
            oldest_entry = 0;

            // Loop through all the entries to find the oldest.
            for (i = 0; i < sync_out_data_list.size(); ++i)
            {
                if (sync_out_data_list.at(i).last_accessed < oldest_time)
                {
                    oldest_entry = i;
                    oldest_time = sync_out_data_list.at(i).last_accessed;
                }
            }

            // Replace the oldest entry with the new value.
            sync_out_data_list[oldest_entry].ip_addr = cam_ip_addr;
            sync_out_data_list[oldest_entry].sync_out_reg = sync_out_reg_to_save;
            sync_out_data_list[oldest_entry].last_accessed = QTime::currentTime();
        }

        // Save off the list.
        // NOTE: This might be overkill, re-writing the entire list. I am just not
        //       certain how growing the list would affect the writing. I know this
        //       will work, and I don't have time for optimization at this point.
        // NOTE: We are only writing this back out to catch the "last accessed time".
        settings.beginWriteArray(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA_LIST);
        for (i = 0; i < sync_out_data_list.size(); ++i)
        {
#if 0
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Saving SyncOut index:" << i <<
                         "ip_addr:" << Qt::showbase << Qt::hex << sync_out_data_list.at(i).ip_addr <<
                         "reg:" << Qt::showbase << Qt::hex << sync_out_data_list.at(i).sync_out_reg;
#endif
            settings.setArrayIndex(i);
            settings.setValue(PanelRecControls::RECORD_CONTROL_SYNC_OUT_IP, sync_out_data_list.at(i).ip_addr);
            settings.setValue(PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA, sync_out_data_list.at(i).sync_out_reg);
            settings.setValue(PanelRecControls::RECORD_CONTROL_SYNC_OUT_TIME, sync_out_data_list.at(i).last_accessed);
        }
        settings.endArray();

        // All done with the group.
        settings.endGroup();
    }
};

}   // End of namespace REC_CONTROLS.

//! Style used to highlight activity for external I/O on that channel.
const QString PanelRecControls::EXTIO_ACT_ACTIVE_STYLE      = QStringLiteral("QRadioButton {background-color: lime;border-style: outset;border-width: 1px;border-radius: 3px;border-color: dimgray;padding: 3px;color:black}");

//! Style used to highlight inactivity for external I/O on that channel.
const QString PanelRecControls::EXTIO_ACT_INACTIVE_STYLE    = QStringLiteral("QRadioButton {border-style: outset;border-width: 1px;border-radius: 3px;border-color: dimgray;padding: 3px}");

//! Style used for label text with disabled (meaning not set in external I/O register) state.
const QString PanelRecControls::TXT_STYLE_COLOR_DISABLED = QStringLiteral("QLabel {color: black;}");

//! Style used to color label text with active signal status.
const QString PanelRecControls::TXT_STYLE_COLOR_ACTSIG  = QStringLiteral("QLabel {background-color: lime;color:black;}");

//! Style used to color label text with inactive signal status.
const QString PanelRecControls::TXT_STYLE_COLOR_NOSIG = QStringLiteral("QLabel {background-color: red;color:black;}");

//! Style used to color label text with waiting for signal status.
const QString PanelRecControls::TXT_STYLE_COLOR_WAITSIG = QStringLiteral("QLabel {background-color: darkorange;color:black;}");

//! Style used to color warning text for sync-in warning messages.
const QString PanelRecControls::TXT_STYLE_COLOR_WARN = QStringLiteral("QLabel {color: red;font-style: italic;}");

//! The QSettings group entery name for record controls.
const QString PanelRecControls::RECORD_CONTROL_QSETTINGS_GROUP = QStringLiteral("RecordControls");

//! The Sync Out register setting for record controls.
const QString PanelRecControls::RECORD_CONTROL_SYNC_OUT_REG = QStringLiteral("SyncOutRegister");

//! The Sync Out data list.
const QString PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA_LIST = QStringLiteral("SyncOutDataList");

//! The Sync Out data ip address entry.
const QString PanelRecControls::RECORD_CONTROL_SYNC_OUT_IP = QStringLiteral("SyncOutIP");

//! The Sync Out data entry string.
const QString PanelRecControls::RECORD_CONTROL_SYNC_OUT_DATA = QStringLiteral("SyncOutData");

//! The Sync Out data last accessed time.
const QString PanelRecControls::RECORD_CONTROL_SYNC_OUT_TIME = QStringLiteral("SyncOutLastAccess");

using namespace REC_CONTROLS;
using namespace FM_PANEL;

/*!
 * @brief Constructor for the PanelRecControls class. Initialize variables.
 */
PanelRecControls::PanelRecControls(QWidget *parent) :
    FMPanelWidget(parent),
    ui(new Ui::PanelRecControls),
    camera_ip_addr(0)
{
    // Setup our UI controls.
    ui->setupUi(this);

    // Make sure we got a parent pointer.
    if (parent != nullptr)
    {
        // My grandparent is the mainwindow.
        MainWindow  *p_mainwnd = qobject_cast<MainWindow *>(parent->parent());

        // Connect the menu item's I'm interested in.
        connect(p_mainwnd->get_cp_reccontrols(), &QAction::triggered, this, &PanelRecControls::toggle_visibility);
    }

    // Initialize our variables.
    num_frames = 0;
    retry_counter = 0;
    update_timer_id = 0;
    update_in_progress = false;
    update_reg_addr = 0;
    update_retry_count = 0;
    write_list.clear();
    vt_allowed = true;
    extio_read_pending = 0;

    channels_in_use = 0;
    arm_in_chan = -1;
    arm_out_chan = -1;
    sync_in_chan = -1;
    sync_out_chan = -1;
    trig_in_chan = -1;
    trig_out_chan = -1;

    // Initialize our controls.
    init_controls();

    // After everything is setup, setup the intial size information.
    set_initial_sizing();

    // Start out with things disabled until we get connected to a camera.
    setEnabled(false);
}

/*!
 * @brief Destructor for the PanelRecControls class.
 */
PanelRecControls::~PanelRecControls()
{
    // Make sure our timer is stopped.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }

    // Release our UI object.
    delete ui;
}

/*!
 * @brief Invoked when we connect the panel to a camera.
 * @param ip_addr   The IP address of the camera we're connecting to. This may be nullptr if we're not connecting
 *                  to any cameras.
 */
void PanelRecControls::camera_connect(uint32_t ip_addr)
{
    int base_time;
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr;

    // Make sure we're not already connected to this camera.
    if (ip_addr == camera_ip_addr)
        return;

    // Save the IP address.
    camera_ip_addr = ip_addr;

    // Initialize our per-camera settings.
    channels_in_use = 0;
    arm_in_chan = -1;
    arm_out_chan = -1;
    sync_in_chan = -1;
    sync_out_chan = -1;
    trig_in_chan = -1;
    trig_out_chan = -1;

    // Reset the video trigger checkboxes unchecked each camera connect.
    // They will be re-established based on the camera_ptr later.
    // The reason to do this is because if they are left-over checked from
    // a different camera connection some of the update/enable/read_complete
    // handling could mistakenly think the vt was enabled by a leftover check.
    ui->chk_vt_enable->setChecked(false);
    ui->chk_vt_marker_only->setChecked(false);

    // Get our communications pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Connect the communications interfaces that we're interested in.
    connect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelRecControls::read_complete, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::msg_timeout, this, &PanelRecControls::msg_timeout_event, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelRecControls::handle_unsol_msg, Qt::QueuedConnection);
    connect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelRecControls::write_complete, Qt::QueuedConnection);

    // Pass it on.
    FMPanelWidget::camera_connect(ip_addr);

    // Enable the panel.
    setEnabled(true);
    set_custom_styles(true);

    // Mark the sync-in as indication its reading its data.
    sync_in_connect_camera();

    // Set up the sync-out duty cycle fps to zero to indicate it needs to be established.
    duty_cycle_fps = 0;

    // Get our camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    connect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelRecControls::chk_refresh_regs, Qt::QueuedConnection);
    connect(this, &PanelRecControls::settingExtArmInEnabled, camera_ptr, &CameraMDI::do_ext_arm_in_enabling, Qt::QueuedConnection);
    connect(this, &PanelRecControls::settingExtArmInDisabled, camera_ptr, &CameraMDI::do_ext_arm_in_disabling, Qt::QueuedConnection);

//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera:" << camera_ptr->get_camera_name();

    // Disable the channel selection mechanism if we're on a series3 camera.
    ui->frm_ai_chan->setVisible(!camera_ptr->is_series3());
    ui->frm_ao_chan->setVisible(!camera_ptr->is_series3());
    ui->frm_si_chan->setVisible(!camera_ptr->is_series3());
    ui->frm_so_chan->setVisible(!camera_ptr->is_series3());
    ui->frm_ti_chan->setVisible(!camera_ptr->is_series3());
    ui->frm_to_chan->setVisible(!camera_ptr->is_series3());

    // Get the base time period for the trigger.
    base_time = camera_ptr->get_tt_base();

    // Check to see if this camera is running a time trigger.
    if (base_time > 0)
    {
        // Enable the time trigger controls.
        ui->chk_tt_enable->setChecked(true);

        // Pass the information to our countdown control
        ui->tt_next->setTime(base_time);

        // Make sure the time controls are configured appropriately.
        ui->spin_tt_hours->setValue(base_time / 3600);
        base_time = base_time % 3600;
        ui->spin_tt_minutes->setValue(base_time / 60);
        base_time = base_time % 60;
        ui->spin_tt_seconds->setValue(base_time);

        // Connect the time trigger countdown control to the camera.
        ui->tt_next->setCamera(camera_ptr);

        // How many triggers have occurred.
        ui->tt_count->display(camera_ptr->get_tt_count());

        // Resume the countdown.
        ui->tt_next->resume(camera_ptr->get_tt_remaining());

        // Connect our signals.
        connect(ui->tt_restart, &QPushButton::clicked, this, &PanelRecControls::tt_restart);
        connect(ui->tt_start, &QPushButton::clicked, this, &PanelRecControls::tt_start);
        connect(ui->tt_stop, &QPushButton::clicked, this, &PanelRecControls::tt_stop);
        connect(camera_ptr, &CameraMDI::tt_expired, this, &PanelRecControls::tt_incr, Qt::UniqueConnection);

        // Update the UI.
        ui->frm_tt_time->setEnabled(true);
        ui->frm_tt_entry->setEnabled(true);
        ui->tt_start->setEnabled(false);
        ui->frm_tt_entry->setEnabled(false);
        ui->tt_stop->setEnabled(true);
        ui->tt_restart->setEnabled(true);
    }
    else
    {
        // Disconnect from any cameras.
        ui->tt_next->setCamera(nullptr);

        // Tell the countdown display to stop.
        ui->tt_next->stop();

        // Make sure we de-select the Time Trigger enable.
        ui->chk_tt_enable->setChecked(false);

        // Update the UI.
        tt_enable(ui->chk_tt_enable->isChecked());
    }

    // Series 3 cameras can only run the FasMotion software version of the
    // video trigger. The seriesHS cameras run the algorithm in the hardware.
    if (camera_ptr->is_series3())
    {
        // Query our connected camera object to see if the user has enabled
        // the video trigger.
        ui->chk_vt_enable->setChecked(camera_ptr->vt_reticle_visible());

        // Update the user interface.
        vt_enable(ui->chk_vt_enable->isChecked());
    }
    else
    {
        // For HS series cameras, we need to query the camera server
        // to pull the HW video trigger control information.
        request_vt_info();

        // Start the external I/O activity poller.
        start_extio_activity_polling();
    }
}

/*!
 * @brief Disconnect from our camera.
 */
void PanelRecControls::camera_disconnect()
{
    FI_comm     *comm_ptr;
    CameraMDI   *camera_ptr = nullptr;

    // Have we been here already?
    if (camera_ip_addr == 0)
        return;

    // Clear our IP address.
    camera_ip_addr = 0;

    // Make sure there is a camera pointer as we need to connect to some signals.
    if (p_control_panel)
    {
        camera_ptr = p_control_panel->getCamera();
    }

    // Disconnect any camera-specific signals.
    if (camera_ptr != nullptr)
    {
        disconnect(camera_ptr, &CameraMDI::group_read_complete, this, &PanelRecControls::chk_refresh_regs);
        disconnect(camera_ptr, &CameraMDI::tt_expired, this, &PanelRecControls::tt_incr);
        disconnect(this, &PanelRecControls::settingExtArmInEnabled, camera_ptr, &CameraMDI::do_ext_arm_in_enabling);
        disconnect(this, &PanelRecControls::settingExtArmInDisabled, camera_ptr, &CameraMDI::do_ext_arm_in_disabling);
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // Pass it up.
    FMPanelWidget::camera_disconnect();

    // Get our communications pointer.
    comm_ptr = get_comm_ptr();
    if (comm_ptr != nullptr)
    {
        // Disconnect the communications interfaces that we're interested in.
        disconnect(comm_ptr, &FI_comm::read_reg_complete, this, &PanelRecControls::read_complete);
        disconnect(comm_ptr, &FI_comm::msg_timeout, this, &PanelRecControls::msg_timeout_event);
        disconnect(comm_ptr, &FI_comm::unsol_msg_received, this, &PanelRecControls::handle_unsol_msg);
        disconnect(comm_ptr, &FI_comm::write_reg_complete, this, &PanelRecControls::write_complete);
    }
    else
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
    }

    // By default, hide the channel controls.
    ui->frm_ai_chan->setVisible(false);
    ui->frm_ao_chan->setVisible(false);
    ui->frm_si_chan->setVisible(false);
    ui->frm_so_chan->setVisible(false);
    ui->frm_ti_chan->setVisible(false);
    ui->frm_to_chan->setVisible(false);

    // Disconnect the countdown control.
    ui->tt_next->setCamera(nullptr);

    // Tell the countdown display to stop.
    ui->tt_next->stop();

    // Stop polling the camera.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }

    // Stop polling for external i/o activity.
    if (extio_activity_timer.isActive())
        extio_activity_timer.stop();

    // Disable all controls.
    setEnabled(false);
    set_custom_styles(false);
    sync_in_disconnect_camera();
}

/*!
 * @brief Display the time in the correct format for the user to use. This is so
 *        time display is consistent and keeps the field sizes to a minimum.
 *
 * @param time_value    The value of time in seconds.
 * @param lbl_time      The label associated with the time field.
 * @param eb_time       The label to display the time in.
 */
void PanelRecControls::display_time(double time_value, QLabel *lbl_time, QLabel *eb_time)
{
    QString s_field;

    // If the time is greater than 1000 sec then display in seconds without
    // a decimal place
    if (time_value >= 1000.0)
    {
        s_field = QString::fromUtf8("%1").arg(time_value);
        lbl_time->setText(tr("Sec"));
    }
    else if ((time_value >= 1.0) || (time_value == 0.0))
    {
        s_field = QString::fromUtf8("%1").arg(time_value, 0, 'f', 3);
        lbl_time->setText(tr("Sec"));
    }
    else if (time_value >= 0.001)
    {
        // Display as millesconds
        time_value *= 1000.0;
        s_field = QString::fromUtf8("%1").arg(time_value, 0, 'f', 3);
        lbl_time->setText(tr("mSec"));
    }
    else
    {
        // display in microseconds
        time_value *= 1000000.0;
        s_field = QString::fromUtf8("%1").arg(time_value, 0, 'f', 3);
        lbl_time->setText(tr("uSec"));
    }

    eb_time->setText(s_field);
}

void PanelRecControls::enableVtControls(bool do_enable)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "do_enable:" << do_enable;

    // Sanity check.
    if (do_enable)
    {
        uint32_t reg_value;

        // We're only interested in cases when LTR and FFLR capture modes are enabled.
        reg_value = read_camera_cache(GIGE_REG_LTR_MODE);
        if (((reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE) ||
            ((reg_value & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR))
        {
            // Are we in ROC/BROC mode?
            if ((reg_value & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0)
            {
                CameraMDI   *camera_ptr;

                // Make sure we have an active camera pointer.
                camera_ptr = p_control_panel->getCamera();
                if (camera_ptr == nullptr)
                {
                    QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
                    return;
                }

                // Cannot enable the video trigger functionality if we are a seriesNC camera.
                if (!camera_ptr->is_series3())
                {
                    do_enable = false;
                }
            }
        }
    }

    // Remember if we're allowed to be enabled or not.
    vt_allowed = do_enable;

    // Only allow changes when in the correct camera state.
    if ((camera_state == GIGE_CAMERA_STATE_IDLE) ||
        (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO) ||
        (camera_state == GIGE_CAMERA_STATE_ARMED) ||
        (camera_state == GIGE_CAMERA_STATE_TRIGGERED))
    {
        ui->frm_vt->setEnabled(do_enable);
        if (!do_enable)
        {
            CameraMDI   *camera_ptr;

            // Make sure we have an active camera pointer.
            camera_ptr = p_control_panel->getCamera();
            if (camera_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
                return;
            }

            // Turn it off.
            ui->chk_vt_enable->setChecked(false);
            camera_ptr->vt_reticle_show(false);
        }
    }
}

QString PanelRecControls::getExtioLabel(REC_CONTROLS::EXT_IO_TYPES ext_io)
{
    QString label;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ext_io:" << ext_io;

    // Make sure it's empty.
    label.clear();

    // Figure out what it should be.
    switch (ext_io)
    {
    case REC_CONTROLS::EXT_IO_TYPES::EXTIO_AO:
        label.append("A-O");
        break;
    case REC_CONTROLS::EXT_IO_TYPES::EXTIO_AI:
        label.append("A-I");
        break;
    case REC_CONTROLS::EXT_IO_TYPES::EXTIO_SO:
        label.append("S-O");
        break;
    case REC_CONTROLS::EXT_IO_TYPES::EXTIO_SI:
        label.append("S-I");
        break;
    case REC_CONTROLS::EXT_IO_TYPES::EXTIO_TO:
        label.append("T-O");
        break;
    case REC_CONTROLS::EXT_IO_TYPES::EXTIO_TI:
        label.append("T-I");
        break;
    default:
    case REC_CONTROLS::EXT_IO_TYPES::EXTIO_NONE:
        break;
    }

    return label;
}


/*!
 * @brief Return the type of I/O associated with a channel.
 * @param channel The channel we're interested in.
 * @return One of REC_CONTROLS::EXT_IO_TYPES.
 */
REC_CONTROLS::EXT_IO_TYPES PanelRecControls::getType(int channel)
{
    REC_CONTROLS::EXT_IO_TYPES ret_type = EXT_IO_TYPES::EXTIO_NONE;

    // Sanity check.
    if (channel > MAX_CHANNEL)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Invalid channel:" << channel;
        return EXT_IO_TYPES::EXTIO_NONE;
    }

    if (arm_in_chan == channel)
        ret_type = EXT_IO_TYPES::EXTIO_AI;
    else if (arm_out_chan == channel)
        ret_type = EXT_IO_TYPES::EXTIO_AO;
    else if (sync_in_chan == channel)
        ret_type = EXT_IO_TYPES::EXTIO_SI;
    else if (sync_out_chan == channel)
        ret_type = EXT_IO_TYPES::EXTIO_SO;
    else if (trig_in_chan == channel)
        ret_type = EXT_IO_TYPES::EXTIO_TI;
    else if (trig_out_chan == channel)
        ret_type = EXT_IO_TYPES::EXTIO_TO;

    return ret_type;
}

int PanelRecControls::get_arm_in_channel()
{
    int channel = -1;

    // Which radio button was checked?
    if (ui->rb_ai_chan_0->isChecked())
    {
        channel = 0;
    }
    else if (ui->rb_ai_chan_1->isChecked())
    {
        channel = 1;
    }
    else if (ui->rb_ai_chan_2->isChecked())
    {
        channel = 2;
    }

    return channel;
}

int PanelRecControls::get_arm_out_channel()
{
    int channel = -1;

    // Which radio button was checked?
    if (ui->rb_ao_chan_0->isChecked())
    {
        channel = 0;
    }
    else if (ui->rb_ao_chan_1->isChecked())
    {
        channel = 1;
    }
    else if (ui->rb_ao_chan_2->isChecked())
    {
        channel = 2;
    }

    return channel;
}

int PanelRecControls::get_sync_in_channel()
{
    int channel = -1;

    // Which radio button was checked?
    if (ui->rb_si_chan_0->isChecked())
    {
        channel = 0;
    }
    else if (ui->rb_si_chan_1->isChecked())
    {
        channel = 1;
    }
    else if (ui->rb_si_chan_2->isChecked())
    {
        channel = 2;
    }

    return channel;
}

int PanelRecControls::get_sync_out_channel()
{
    int channel = -1;

    // Which radio button was checked?
    if (ui->rb_so_chan_0->isChecked())
    {
        channel = 0;
    }
    else if (ui->rb_so_chan_1->isChecked())
    {
        channel = 1;
    }
    else if (ui->rb_so_chan_2->isChecked())
    {
        channel = 2;
    }

    return channel;
}

int PanelRecControls::get_trig_in_channel()
{
    int channel = -1;

    // Which radio button was checked?
    if (ui->rb_ti_chan_0->isChecked())
    {
        channel = 0;
    }
    else if (ui->rb_ti_chan_1->isChecked())
    {
        channel = 1;
    }
    else if (ui->rb_ti_chan_2->isChecked())
    {
        channel = 2;
    }

    return channel;
}

int PanelRecControls::get_trig_out_channel()
{
    int channel = -1;

    // Which radio button was checked?
    if (ui->rb_to_chan_0->isChecked())
    {
        channel = 0;
    }
    else if (ui->rb_to_chan_1->isChecked())
    {
        channel = 1;
    }
    else if (ui->rb_to_chan_2->isChecked())
    {
        channel = 2;
    }

    return channel;
}

/*!
 * @brief Slot invoked when a group read has completed. This handler is looking for
 * the read registers group read's last address == GIGE_REG_CAMERA_STATE.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the last register which was read in the group.
 */
void PanelRecControls::chk_refresh_regs(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *camera_ptr  = p_control_panel->getCamera();

    // If we are not connected to any camera we don't care about this signal.
    if (camera_ptr != nullptr)
    {
        // Is this event for us?
        if (ip_addr != camera_ptr->get_ip_addr())
            return;

        // Verify it's our register.
        if (reg_addr == GIGE_REG_ARM_OUT_SELECT)
        {
#if LOG_GROUP_READS
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << qPrintable(FiCamRegGroup::CAMERA_GRP_LOG_NAME) << "[COMPLETE][GIGE_REG_ARM_OUT_SELECT]";
#endif
            // This is a actionless group read there is no action that takes place...
        }
    }
}

/*!
 * @brief Read (or calculate) the frame rate of the "master" camera. This is used when we
 *        are determining our frame rate when we are slaving to another camera.
 */
uint32_t PanelRecControls::get_master_framerate()
{
    uint32_t    master_framerate;

    // Get the master frame rate from the camera.
    master_framerate = read_camera_cache(GIGE_REG_SYNC_IN_MFR);

#if LOG_MASTER_FRAME_RATE
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_SYNC_IN_MFR:" << master_framerate;
#endif

    // Sanity check our calculations.
    if (master_framerate < MIN_MASTER_FRAMERATE)
    {
        // We probably are not getting a sync-in signal, so get the actual framerate.
        master_framerate = (uint32_t)ClientMath::reg_to_frame_rate(read_camera_cache(GIGE_REG_FRAME_RATE));
    }
    else if (master_framerate > MAX_MASTER_FRAMERATE)
    {
        // Clip the framerate at the maximum allowed.
        master_framerate = MAX_MASTER_FRAMERATE;
    }

#if LOG_MASTER_FRAME_RATE
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Returning master_framerate:" << master_framerate;
#endif

    // Return our result.
    return master_framerate;
}

/*!
 * @brief A utility function to calculate the maximum allowed frame rate for a given ROI.
 */
uint32_t PanelRecControls::get_max_framerate()
{
    bool        ltr_enabled;
    uint32_t    reg_value;
    uint32_t    ssd_bandwidth;
    uint32_t    width;
    uint32_t    height;
    uint32_t    offset_x;
    uint32_t    bit_depth;
    uint32_t    max_framerate;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return 0;
    }
    if (camera_ptr->get_sensor() == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No sensor?!";
        return 0;
    }

    // Are we in LTR mode?
    reg_value = read_camera_cache(GIGE_REG_LTR_MODE);
    ltr_enabled = (reg_value & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE;
    if (ltr_enabled)
    {
        bool        ok;

        // Get the SSD bandwidth used by the camera.
        ssd_bandwidth = read_camera_cache(GIGE_REG_SSD_BANDWIDTH, &ok);
        if (!ok)
            ssd_bandwidth = FASMOTION_SETTINGS::DEFAULT_SSD_BANDWIDTH;
    }
    else
    {
        // No SSD bandwidth to worry about.
        ssd_bandwidth = 0;
    }

    // Get our ROI.
    width = read_camera_cache(GIGE_REG_WIDTH);
    height = read_camera_cache(GIGE_REG_HEIGHT);
    offset_x = read_camera_cache(GIGE_REG_OFFSET_X);

    // Get the current bit depth.
    bit_depth = read_camera_cache(GIGE_REG_BIT_MODE);

    // Figure out the maximum allowed framerate.
    max_framerate = ClientMath::calc_max_framerate(camera_ptr->get_sensor(), offset_x, width, height,
                                                   (int)bit_depth, ltr_enabled, ssd_bandwidth, camera_ptr->is_series3());

    return max_framerate;
}

/*!
 * @brief When a register read completes, this invokes an appropriate method to
 *        perform any additional processing which might be required.
 *
 * @param reg_addr      The address of the register we just read.
 * @param reg_value     The value of that register.
 */
void PanelRecControls::handle_read(uint32_t reg_addr, uint32_t reg_value)
{
    // Don't update our settings if we're in the middle of updating a value.
    if (update_in_progress)
        return;

    // Process based on the register address.
    switch (reg_addr)
    {
    case GIGE_REG_ARM_IN_ENABLE:
        update_arm_in();
        break;
    case GIGE_REG_ARM_OUT_SELECT:
        update_arm_out();
        break;
    case GIGE_REG_CAMERA_STATE:
        process_camera_state(reg_value);
        break;
    case GIGE_REG_EXTIO_ACTIVITY:
        update_extio_activity();
        break;
    case GIGE_REG_EXT_TRIG_POLARITY:
        update_ext_trigger();
        break;
    case GIGE_REG_IRIG_CONTROL:
        update_irig();
        break;
    case GIGE_REG_LTR_MODE:
        setLtrMode(reg_value);
        update_num_frames();
        break;
    case GIGE_REG_SYNC_IN_ENABLE:
        update_sync_in();
        break;
    case GIGE_REG_SYNC_OUT_SELECT:
        update_sync_out();
        break;
    case GIGE_REG_TRIG_OUT_SELECT:
        update_trig_out();
        break;
    case GIGE_REG_TRIGGER_REQ_POS:
        update_trigger();
        break;
    case GIGE_REG_VTRIG_SETTINGS_JSON:
        update_vt_info();
        break;
    default:
        break;
    }
}


/*!
 * @brief Initialize the signal/slot connections and initializations for the controls.
 */
void PanelRecControls::init_controls()
{
    // Connect the signals to the slot, using a lambda. This allows us to switch between
    // the pannels when the user clicks on a tab.
    connect(ui->arm_in_enable, &QCheckBox::clicked, this, [this](bool /*checked*/) { tab_enable_clicked(this->ui->frame_arm_in); });
    connect(ui->arm_out_enable, &QCheckBox::clicked, this, [this](bool /*checked*/) { tab_enable_clicked(this->ui->frame_arm_out); });
    connect(ui->sync_in_enable, &QCheckBox::clicked, this, [this](bool /*checked*/) { tab_enable_clicked(this->ui->frame_sync_in); });
    connect(ui->sync_out_enable, &QCheckBox::clicked, this, [this](bool /*checked*/) { tab_enable_clicked(this->ui->frame_sync_out); });
    connect(ui->trig_out_enable, &QCheckBox::clicked, this, [this](bool /*checked*/) { tab_enable_clicked(this->ui->frame_trig_out); });

    // On startup we want the tab control to be at index 0 - the first tab.
    ui->tab_io_controls->setCurrentIndex(0);

    // Set up custom styles.
    set_custom_styles(true);

    // By default, disable all the "in/out" controls.
    ui->frame_arm_in->setEnabled(false);
    ui->frame_arm_out->setEnabled(false);
    ui->frame_sync_in->setEnabled(false);
    ui->frame_sync_out->setEnabled(false);
    ui->frame_trig_in->setEnabled(false);
    ui->frame_trig_out->setEnabled(false);

    // By default, hide the IRIG controls.
    ui->frm_irig->setVisible(false);

    // By default, hide the channel controls.
    ui->frm_ai_chan->setVisible(false);
    ui->frm_ao_chan->setVisible(false);
    ui->frm_si_chan->setVisible(false);
    ui->frm_so_chan->setVisible(false);
    ui->frm_ti_chan->setVisible(false);
    ui->frm_to_chan->setVisible(false);

    connect(ui->rb_use_frames, &QRadioButton::clicked, this, &PanelRecControls::rb_use_frame_clicked);
    connect(ui->rb_use_percent, &QRadioButton::clicked, this, &PanelRecControls::rb_use_percent_clicked);
    connect(ui->sld_trig_pos, &QSlider::valueChanged, this, &PanelRecControls::sld_trigger_pos_changed);
    connect(ui->sld_trig_pos, &QSlider::sliderReleased, this, &PanelRecControls::trigger_pos_update, Qt::QueuedConnection);
    connect(ui->sld_trig_pos, &QSlider::actionTriggered, this, &PanelRecControls::trigger_pos_action);
    connect(ui->spin_trig_pos, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::spin_trigger_pos_changed);
    connect(ui->spin_trig_pos, &QSpinBox::editingFinished, this, &PanelRecControls::trigger_pos_update);

    connect(ui->trig_in_enable, &QCheckBox::clicked, this, &PanelRecControls::trig_in_enable_clicked);
    connect(ui->rb_ti_chan_0, &QRadioButton::clicked, this, &PanelRecControls::trig_in_channel);
    connect(ui->rb_ti_chan_1, &QRadioButton::clicked, this, &PanelRecControls::trig_in_channel);
    connect(ui->rb_ti_chan_2, &QRadioButton::clicked, this, &PanelRecControls::trig_in_channel);
    connect(ui->rb_ti_falling, &QRadioButton::clicked, this, &PanelRecControls::trig_in_falling_clicked);
    connect(ui->rb_ti_rising, &QRadioButton::clicked, this, &PanelRecControls::trig_in_rising_clicked);

    connect(ui->chk_vt_enable, &QCheckBox::clicked, this, &PanelRecControls::vt_enable);
    connect(ui->chk_vt_marker_only, &QCheckBox::clicked, this, &PanelRecControls::vt_mark_only);
    connect(ui->pb_vt_reset, &QPushButton::pressed, this, &PanelRecControls::vt_reset);
    connect(ui->sld_vt_area, &QSlider::valueChanged, ui->spin_vt_area, &QSpinBox::setValue);
    connect(ui->spin_vt_area, qOverload<int>(&QSpinBox::valueChanged), ui->sld_vt_area, &QSlider::setValue);
    connect(ui->spin_vt_area, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::vt_area_changed);
    connect(ui->sld_vt_change, &QSlider::valueChanged, ui->spin_vt_change, &QSpinBox::setValue);
    connect(ui->spin_vt_change, qOverload<int>(&QSpinBox::valueChanged), ui->sld_vt_change, &QSlider::setValue);
    connect(ui->spin_vt_change, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::vt_change_changed);
    connect(ui->spin_refresh, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::vt_refresh_changed);

    connect(ui->chk_tt_enable, &QCheckBox::clicked, this, &PanelRecControls::tt_enable);
    connect(ui->spin_tt_hours, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::tt_hours_changed);
    connect(ui->spin_tt_minutes, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::tt_minutes_changed);
    connect(ui->spin_tt_seconds, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::tt_seconds_changed);

    connect(ui->chk_irig, &QCheckBox::clicked, this, &PanelRecControls::irig_clicked);
    connect(ui->chk_irig_year, &QCheckBox::clicked, this, &PanelRecControls::irig_year_clicked);

    connect(ui->arm_in_enable, &QCheckBox::clicked, this, &PanelRecControls::arm_in_enable_clicked);
    connect(ui->rb_ai_chan_0, &QRadioButton::clicked, this, &PanelRecControls::arm_in_channel);
    connect(ui->rb_ai_chan_1, &QRadioButton::clicked, this, &PanelRecControls::arm_in_channel);
    connect(ui->rb_ai_chan_2, &QRadioButton::clicked, this, &PanelRecControls::arm_in_channel);
    connect(ui->rb_ai_edge_type, &QRadioButton::clicked, this, &PanelRecControls::arm_in_edge_type_clicked);
    connect(ui->rb_ai_falling_low, &QRadioButton::clicked, this, &PanelRecControls::arm_in_falling_low_clicked);
    connect(ui->rb_ai_level_type, &QRadioButton::clicked, this, &PanelRecControls::arm_in_level_type_clicked);
    connect(ui->rb_ai_rising_high, &QRadioButton::clicked, this, &PanelRecControls::arm_in_rising_high_clicked);
    connect(ui->chk_allow_discard, &QCheckBox::clicked, this, &PanelRecControls::arm_in_allow_discard);

    ui->spin_ao_pulse_width->setMinimum(MIN_PW_VALUE);
    ui->spin_ao_pulse_width->setMaximum(MAX_PW_VALUE);

    connect(ui->arm_out_enable, &QCheckBox::clicked, this, &PanelRecControls::arm_out_enable_clicked);
    connect(ui->rb_ao_active_high, &QRadioButton::clicked, this, &PanelRecControls::arm_out_active_high_clicked);
    connect(ui->rb_ao_active_low, &QRadioButton::clicked, this, &PanelRecControls::arm_out_active_low_clicked);
    connect(ui->rb_ao_chan_0, &QRadioButton::clicked, this, &PanelRecControls::arm_out_channel);
    connect(ui->rb_ao_chan_1, &QRadioButton::clicked, this, &PanelRecControls::arm_out_channel);
    connect(ui->rb_ao_chan_2, &QRadioButton::clicked, this, &PanelRecControls::arm_out_channel);
    connect(ui->chk_ao_invert_signal, &QCheckBox::clicked, this, &PanelRecControls::arm_out_invert_signal_clicked);
    connect(ui->chk_ao_pass_thru, &QCheckBox::clicked, this, &PanelRecControls::arm_out_pass_thru_clicked);
    connect(ui->chk_ao_pulse_width, &QCheckBox::clicked, this, &PanelRecControls::arm_out_pulse_width_clicked);
    connect(ui->spin_ao_pulse_width, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::arm_out_pulse_width_changed);

    ui->spin_si_mfr->setMinimum(MIN_MASTER_FRAMERATE);
    ui->spin_si_mfr->setMaximum(MAX_MASTER_FRAMERATE);

    // Per ticket #227, let's disable the sync-in per-second control.
    // TICKET_227
    ui->rb_si_per_sec->setEnabled(false);
    ui->rb_si_per_sec->setVisible(false);
    ui->lbl_si_delay->setVisible(false);
    ui->spin_si_delay->setEnabled(false);
    ui->spin_si_delay->setVisible(false);
    ui->rb_si_per_frame->setChecked(true);
    // TICKET_227

    connect(ui->sync_in_enable, &QCheckBox::clicked, this, &PanelRecControls::sync_in_enable_clicked);
    connect(ui->rb_si_chan_0, &QRadioButton::clicked, this, &PanelRecControls::sync_in_channel);
    connect(ui->rb_si_chan_1, &QRadioButton::clicked, this, &PanelRecControls::sync_in_channel);
    connect(ui->rb_si_chan_2, &QRadioButton::clicked, this, &PanelRecControls::sync_in_channel);
    connect(ui->rb_si_per_frame, &QRadioButton::clicked, this, &PanelRecControls::sync_in_per_frame_clicked);
    connect(ui->rb_si_per_sec, &QRadioButton::clicked, this, &PanelRecControls::sync_in_per_sec_clicked);
    connect(ui->rb_si_falling, &QRadioButton::clicked, this, &PanelRecControls::sync_in_falling_clicked);
    connect(ui->rb_si_rising, &QRadioButton::clicked, this, &PanelRecControls::sync_in_rising_clicked);
    connect(ui->spin_si_delay, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::sync_in_delay_changed);
    connect(ui->spin_si_mfr, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::sync_in_mfr_changed);
    connect(ui->spin_si_rate_div, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::sync_in_rate_div_changed);

    ui->spin_so_duty_cycle->setMinimum(MIN_DUTY_CYCLE);
    ui->spin_so_duty_cycle->setMaximum(MAX_DUTY_CYCLE);

    connect(ui->sync_out_enable, &QCheckBox::clicked, this, &PanelRecControls::sync_out_enable_clicked);
    connect(ui->rb_so_active_high, &QRadioButton::clicked, this, &PanelRecControls::sync_out_active_high_clicked);
    connect(ui->rb_so_active_low, &QRadioButton::clicked, this, &PanelRecControls::sync_out_active_low_clicked);
    connect(ui->rb_so_chan_0, &QRadioButton::clicked, this, &PanelRecControls::sync_out_channel);
    connect(ui->rb_so_chan_1, &QRadioButton::clicked, this, &PanelRecControls::sync_out_channel);
    connect(ui->rb_so_chan_2, &QRadioButton::clicked, this, &PanelRecControls::sync_out_channel);
    connect(ui->rb_so_duty_cycle, &QRadioButton::clicked, this, &PanelRecControls::sync_out_duty_cycle_clicked);
    connect(ui->spin_so_duty_cycle, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::sync_out_duty_cycle_changed);
    connect(ui->rb_so_pass_thru, &QRadioButton::clicked, this, &PanelRecControls::sync_out_pass_thru_clicked);
    connect(ui->rb_so_per_frame, &QRadioButton::clicked, this, &PanelRecControls::sync_out_per_frame_clicked);
    connect(ui->rb_so_per_sec, &QRadioButton::clicked, this, &PanelRecControls::sync_out_per_sec_clicked);
    connect(ui->rb_so_shutter, &QRadioButton::clicked, this, &PanelRecControls::sync_out_shutter_clicked);
    connect(ui->chk_so_invert_signals, &QCheckBox::clicked, this, &PanelRecControls::sync_out_invert_signals_clicked);

    ui->spin_trigout_pulse_width->setMinimum(MIN_PW_VALUE);
    ui->spin_trigout_pulse_width->setMaximum(MAX_PW_VALUE);

    connect(ui->trig_out_enable, &QCheckBox::clicked, this, &PanelRecControls::trig_out_enable_clicked);
    connect(ui->rb_to_chan_0, &QRadioButton::clicked, this, &PanelRecControls::trig_out_channel);
    connect(ui->rb_to_chan_1, &QRadioButton::clicked, this, &PanelRecControls::trig_out_channel);
    connect(ui->rb_to_chan_2, &QRadioButton::clicked, this, &PanelRecControls::trig_out_channel);
    connect(ui->chk_trigout_pass_thru, &QCheckBox::clicked, this, &PanelRecControls::trig_out_pass_thru_clicked);
    connect(ui->chk_trigout_invert_signal, &QCheckBox::clicked, this, &PanelRecControls::trig_out_invert_signal_clicked);
    connect(ui->spin_trigout_pulse_width, qOverload<int>(&QSpinBox::valueChanged), this, &PanelRecControls::trig_out_pulse_width_changed);
    connect(ui->rb_trigout_active_low, &QRadioButton::clicked, this, &PanelRecControls::trig_out_active_low_clicked);
    connect(ui->rb_trigout_active_high, &QRadioButton::clicked, this, &PanelRecControls::trig_out_active_high_clicked);
}

/*!
 * @brief Checks to see if sync-out is enabled and if duty cycle is selected. If it is
 * the frame rate is checked to see if it has changed. If it has the register is rewritten.
 * This is an attempt to refresh the internal value used for the duty cycle signal so that
 * it always matches the camera frame rate.
 */
void PanelRecControls::update_duty_cycle()
{
    // Get the current frame rate of the camera.
    uint32_t current_fps = (uint32_t)ClientMath::reg_to_frame_rate(read_camera_cache(GIGE_REG_FRAME_RATE));

    // If our internal value is out of sync we want to force a refresh write.
    // Note: Each camera connect duty_cycle_fps is reset to zero forcing at least one refresh write.
    if (duty_cycle_fps != current_fps)
    {
        uint32_t reg_value = read_camera_cache(GIGE_REG_SYNC_OUT_SELECT);
        if ((reg_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_ENABLE)
        {
            uint32_t    duty_cycle = (reg_value & GIGE_SYNC_OUT_DUTYCYCLE_MASK) >> GIGE_SYNC_OUT_DUTYCYCLE_SHIFT;

            // If this camera is set for sync-out with a duty cycle then refresh the write.
            if ((duty_cycle >= MIN_DUTY_CYCLE) && (duty_cycle <= MAX_DUTY_CYCLE))
            {
                CameraMDI *camera_ptr = p_control_panel->getCamera();
                QString camera_name = "No Camera Found";
                if (camera_ptr)
                {
                    camera_name = camera_ptr->get_camera_name();
                }

                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Refresh write for Sync Out on camera_name:" << camera_name << "duty_cycle_fps:" << duty_cycle_fps << "current_fps:" << current_fps;

                // Force a write back out to the Sync Out register as the frame rate possibly changed
                // and whatever magic happens in the FPGA to recalculate duty cycle signals maybe this
                // will catch it so that the signal rate is modified.
                write_sync_out();

                // Save the frame rate so we catch future changes while camera is connected.
                duty_cycle_fps = current_fps;
            }
        }
    }
}

/*!
 * @brief Parses the Video Trigger information from camera contained in JSON.
 *
 * @param vtInfoJson     The JSON data for the camera's Video Trigger information.
 */
void PanelRecControls::update_vt_info()
{
    CameraMDI *camera_ptr;
    QJsonDocument jsonMedia;
    QJsonObject vtJsonObj;
    QJsonParseError jp_err;
    QString json_str;
    QByteArray *json_data;
    QRect vt_roi;
    int vt_control;

    // Do we have a camera connected?
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Series3 cameras don't get to play.
    if (camera_ptr->is_series3())
        return;

    // Get the JSON information.
    json_data = camera_ptr->read_camera_buffer(GIGE_REG_VTRIG_SETTINGS_JSON);
    if (json_data == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No HW Video Trigger json_data read.";
        return;
    }

    // Make a QString out of the data.
    json_str = QString::fromLocal8Bit(*json_data);
    if (json_str.length() <= 1)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "HW Video Trigger json_data length <= 1.";
        return;
    }
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "json_str:" << json_str;

    // Strip any trailing NULL characters.
    if (json_str.at(json_str.size() - 1) == '\0')
        json_str.chop(1);

    // Convert to a JSON document.
    jsonMedia = QJsonDocument::fromJson(json_str.toUtf8(), &jp_err);
    if (jsonMedia.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_REG_VTRIG_SETTINGS_JSON Got a parse err:" << jp_err.errorString() << "str:" << json_str;
        return;
    }

    // We ALWAYS expect a control value.
    if (jsonMedia["control"].isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No control entry in VT JSON response!";
        return;
    }

    // Is the video trigger enabled?
    vt_control = jsonMedia["control"].toInt();
    if ((vt_control & GIGE_VTRIG_CONTROL_ENABLE) != GIGE_VTRIG_CONTROL_ENABLE)
    {
        // Make sure we are not enabled.
        ui->chk_vt_enable->setChecked(false);

        // Update the UI.
        vt_enable(ui->chk_vt_enable->isChecked());

        // We are done.
        return;
    }

    // Since we are enabled, let's restore the other video trigger settings.
    // We want to pass the values to the camera controller, but don't send them
    // back to the camera.
    if (!jsonMedia["area_thresh"].isNull())
    {
        int area_thresh = jsonMedia["area_thresh"].toInt();
        camera_ptr->set_vt_area(area_thresh, false);
        ui->spin_vt_area->setValue(area_thresh);
    }
    if (!jsonMedia["pix_thresh"].isNull())
    {
        int pix_thresh = jsonMedia["pix_thresh"].toInt();
        camera_ptr->set_vt_change(pix_thresh, false);
        ui->spin_vt_change->setValue(pix_thresh);
    }
    if (!jsonMedia["interval"].isNull())
    {
        int interval = jsonMedia["interval"].toInt();
        camera_ptr->set_vt_refresh(interval, false);
        ui->spin_refresh->setValue(interval);
    }

    if (!jsonMedia["roi_x"].isNull())
        vt_roi.setLeft(jsonMedia["roi_x"].toInt());
    if (!jsonMedia["roi_y"].isNull())
        vt_roi.setTop(jsonMedia["roi_y"].toInt());
    if (!jsonMedia["roi_width"].isNull())
        vt_roi.setWidth(jsonMedia["roi_width"].toInt());
    if (!jsonMedia["roi_height"].isNull())
        vt_roi.setHeight(jsonMedia["roi_height"].toInt());
    if (!vt_roi.isEmpty())
        camera_ptr->set_vt_region(vt_roi);

    // If the camera was set to be triggered set up the panel to match.
    if ((vt_control & GIGE_VTRIG_CONTROL_ENABLE) == GIGE_VTRIG_CONTROL_ENABLE)
    {
        bool vt_mark_checked = (!camera_ptr->vt_software() && ((vt_control & GIGE_VTRIG_CONTROL_TRIGGER) != GIGE_VTRIG_CONTROL_TRIGGER));

        // We are enabled.
        ui->chk_vt_enable->setChecked(true);

        // Update the UI. This call will mess with the VT marker settings, so we need to do
        // this first, and then update the markers appropriately later.
        vt_enable(ui->chk_vt_enable->isChecked());

        // Are we marking only? We do NOT want to write back to the camera, so make sure we set the "send_to_camera"
        // flag correct in the call to set_vt_mark_only()!
        ui->chk_vt_marker_only->setEnabled(!camera_ptr->vt_software());
        ui->chk_vt_marker_only->setChecked(vt_mark_checked);
        camera_ptr->set_vt_mark_only(vt_mark_checked, false);

        // Reset the algorithm.
        camera_ptr->vt_reticle_show(false);
        camera_ptr->vt_reticle_show(true);
    }
}

/*!
 * @brief Enable or disable controls on the panel based on the camera's state.
 *
 * @param new_state     The new camera state.
 */
void PanelRecControls::process_camera_state(uint32_t new_state)
{
    // Disable the panel unless the camera is in IDLE or LIVE view.
    if ((new_state == GIGE_CAMERA_STATE_IDLE) || (new_state == GIGE_CAMERA_STATE_LIVE_VIDEO))
    {
        CameraMDI   *camera_ptr;

        // Make sure we have an active camera pointer.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
            ui->frm_trigger_mgmt->setEnabled(false);
        }
        else
        {
            // The trigger controls are dependent on the LR mode.
            setLtrMode(camera_ptr->read_camera_cache(GIGE_REG_LTR_MODE));
        }

        // All controls are enabled in these states.
        ui->frm_irig->setEnabled(true);
        ui->frm_tt->setEnabled(true);
        ui->frm_vt->setEnabled(vt_allowed);
    }
    else
    {
        // Most of the controls are disabled in these states.
        ui->frm_irig->setEnabled(false);
        ui->frm_trigger_mgmt->setEnabled(false);

        // Special handling for the video trigger controls.
        if ((new_state == GIGE_CAMERA_STATE_ARMED) || (new_state == GIGE_CAMERA_STATE_TRIGGERED))
        {
            // The time & video trigger controls are enabled in these states too.
            ui->frm_tt->setEnabled(true);
            ui->frm_vt->setEnabled(vt_allowed);
        }
        else
        {
            // The time & video trigger controls are disabled like everyone else here.
            ui->frm_tt->setEnabled(false);
            ui->frm_vt->setEnabled(vt_allowed);
        }
    }

    // If camera is in a review state disable our panel and custom styles.
    if (new_state == GIGE_CAMERA_STATE_REVIEW)
    {
        setEnabled(false);
        set_custom_styles(false);
    }
    else if (!isEnabled())
    {
        // Re-enable because we have a camera connection and we are currently disabled.
        setEnabled(true);
        set_custom_styles(true);
    }
}

/*!
 * @brief A response to the GIGE_REG_EXTIO_ACTIVITY read.  It handles updating any channel
 * style associated with the activity state.
 */
void PanelRecControls::update_extio_activity()
{
    uint32_t extio_reg;

    // Only enter when there is a read pending.
    if (extio_read_pending == 0)
        return;
    extio_read_pending = 0;

    // Read our activity register.
    extio_reg = read_camera_cache(GIGE_REG_EXTIO_ACTIVITY);

    // Apply modifications to the channel 0 radio buttons based on activity setting.
    if (extio_reg & GIGE_EXTIO_ACTIVITY_0)
    {
        // Change the background of channel radio button to show activity.
        ui->rb_ti_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_to_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_ai_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_ao_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_si_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_so_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
    }
    else
    {
        // Change the background of channel radio button to show no activity.
        ui->rb_ti_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_to_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ai_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ao_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_si_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_so_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
    }

    // Apply modifications to the channel 1 radio buttons based on activity setting.
    if (extio_reg & GIGE_EXTIO_ACTIVITY_1)
    {
        // Change the background of channel radio button to show activity.
        ui->rb_ti_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_to_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_ai_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_ao_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_si_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_so_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
    }
    else
    {
        // Change the background of channel radio button to show no activity.
        ui->rb_ti_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_to_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ai_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ao_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_si_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_so_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
    }

    // Apply modifications to the channel 2 radio buttons based on activity setting.
    if (extio_reg & GIGE_EXTIO_ACTIVITY_2)
    {
        // Change the background of channel radio button to show activity.
        ui->rb_ti_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_to_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_ai_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_ao_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_si_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
        ui->rb_so_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_ACTIVE_STYLE);
    }
    else
    {
        // Change the background of channel radio button to show no activity.
        ui->rb_ti_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_to_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ai_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ao_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_si_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_so_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
    }

    // Re-start the timer.
    extio_activity_timer.start();
}

/*!
 * @brief A method invoked periodically (@sa update_timer_id) to poll for registers used by this object.
 */
void PanelRecControls::read_registers()
{
    CameraMDI *camera_ptr = p_control_panel->getCamera();

    // Do we have a camera connected?
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // This is called on a timer-basis.
    // The internal camera_state data member is updated when reads inside of this object
    // get updates for camera state but we are short-circuiting the read here if the
    // camera state is review. This ensures we query the camera's cache for its state.
    uint32_t cur_camera_state = camera_ptr->read_camera_cache(GIGE_REG_CAMERA_STATE);
    if (cur_camera_state == GIGE_CAMERA_STATE_REVIEW)
    {
        // This will get our panel set up for review state because we won't be group reading it.
        process_camera_state(cur_camera_state);
        return;
    }

    QSharedPointer<FiCamRegGroup> cam_register_group = FiCamRegGroup::createCamRegGroup();
    if (cam_register_group.isNull())
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << FiCamRegGroup::CAMERA_GRP_READ_FAILURE;
        return;
    }

    // Prepare the camera register group read for our calculations.
    cam_register_group->addReg(GIGE_REG_CAMERA_STATE);
    cam_register_group->addReg(GIGE_REG_WIDTH);
    cam_register_group->addReg(GIGE_REG_HEIGHT);
    cam_register_group->addReg(GIGE_REG_OFFSET_X);
    cam_register_group->addReg(GIGE_REG_OFFSET_Y);
    cam_register_group->addReg(GIGE_REG_FRAME_RATE);
    cam_register_group->addReg(GIGE_REG_SHUTTER_SPEED);
    cam_register_group->addReg(GIGE_REG_BIT_MODE);
    cam_register_group->addReg(GIGE_REG_FRAME_OVERHEAD);
    cam_register_group->addReg(GIGE_REG_LTR_SECTOR_COUNT);
    cam_register_group->addReg(GIGE_REG_PART_STATE);
    cam_register_group->addReg(GIGE_REG_SSD_BANDWIDTH);
    cam_register_group->addReg(GIGE_REG_PART_SIZE);
    cam_register_group->addReg(GIGE_REG_PART_SIZE_RESERVED);
    cam_register_group->addReg(GIGE_REG_BROC_LEN);
    cam_register_group->addReg(GIGE_REG_LTR_MODE);

    // Get the settings from the camera.
    cam_register_group->addReg(GIGE_REG_TRIGGER_REQ_POS);
    cam_register_group->addReg(GIGE_REG_EXT_TRIG_ENABLE);
    cam_register_group->addReg(GIGE_REG_EXT_TRIG_POLARITY);
    cam_register_group->addReg(GIGE_REG_IRIG_STATUS);
    cam_register_group->addReg(GIGE_REG_IRIG_CONTROL);
    cam_register_group->addReg(GIGE_REG_TRIG_OUT_SELECT);
    cam_register_group->addReg(GIGE_REG_SYNC_IN_TIMING);
    cam_register_group->addReg(GIGE_REG_SYNC_IN_MFR);
    cam_register_group->addReg(GIGE_REG_SYNC_DELAY);
    cam_register_group->addReg(GIGE_REG_SYNC_IN_POLARITY);
    cam_register_group->addReg(GIGE_REG_SYNC_IN_ENABLE);
    cam_register_group->addReg(GIGE_REG_SYNC_OUT_SELECT);
    cam_register_group->addReg(GIGE_REG_ARM_IN_POLARITY);
    cam_register_group->addReg(GIGE_REG_ARM_IN_ENABLE);
    cam_register_group->lastReg(GIGE_REG_ARM_OUT_SELECT);

    // Pass on the request but only increment our counter if we sent it to the camera.
    read_camera_registers(cam_register_group);
}

void PanelRecControls::release_channel(int channel)
{
    // Sanity check.
    if (channel < 0)
        return;

    // Mark the channel as available.
    channels_in_use &= ~(1 << channel);
}

int PanelRecControls::reserve_channel(int asking_channel)
{
    int channel;

    // Check to see if we've already allocated all channels.
    if ((channels_in_use & MAX_CHANNEL_MASK) == MAX_CHANNEL_MASK)
        return -1;

    // Is the caller requesting a specific channel?
    if (asking_channel >= 0)
    {
        // See if the requested channel is available.
        if ((channels_in_use & (1 << asking_channel)) == 0)
        {
            // Mark this channel as in-use.
            channels_in_use |= (1 << asking_channel);
            return asking_channel;
        }

        // That channel is not available.
        return -1;
    }

    // Find an available channel.
    for (channel = 0; channel < MAX_CHANNEL; ++channel)
    {
        // See if this channel is available.
        if ((channels_in_use & (1 << channel)) == 0)
        {
            // We found one. Mark this channel as in-use.
            channels_in_use |= (1 << channel);

            // Return this channel value.
            return channel;
        }
    }

    // This "should" never happen...
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "We think there is an available channel, but didn't find one?!";
    return -1;
}

/*!
 * @brief Invoked when the LTR mode changes. Handle UI changes based on the new setting.
 * @param ltr_mode  The new LTR mode settings.
 */
void PanelRecControls::setLtrMode(uint32_t ltr_mode)
{
    // Ignore this unless we're in these states.
    if ((camera_state == GIGE_CAMERA_STATE_IDLE) || (camera_state == GIGE_CAMERA_STATE_LIVE_VIDEO))
    {
        bool    enable_ctrl = true;             // Assume we'll want to enable the controls.
        CameraMDI   *camera_ptr;

        // Make sure we have an active camera pointer.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
            return;
        }

        // We're only interested in cases when LTR and FFLR capture modes are enabled.
        if (((ltr_mode & (uint32_t)GIGE_LTR_MODE_ENABLE) == (uint32_t)GIGE_LTR_MODE_ENABLE) ||
            ((ltr_mode & (uint32_t)GIGE_LTR_MODE_FFLR) == (uint32_t)GIGE_LTR_MODE_FFLR))
        {
            // Are we in ROC/BROC mode?
            if ((ltr_mode & (GIGE_LTR_MODE_BROC | GIGE_LTR_MODE_ROC)) != 0)
            {
                // We're in a mode where it doesn't make sense to modify the trigger location.
                enable_ctrl = false;

                // In ROC/BROC we need to enforce a start trigger.
                if (ui->spin_trig_pos->value() != 0)
                {
                    ui->spin_trig_pos->setValue(0);
                    trigger_pos_update();
                }
            }
        }

        // Update our controls.
        ui->frm_trigger_mgmt->setEnabled(enable_ctrl);
        if (camera_ptr->is_series3())
        {
            ui->chk_vt_enable->setEnabled(true);
        }
        else
        {
            // Enable/disable the video trigger controls as appropriate.
            // If we are disabling, then make sure we turn off the function too.
            enableVtControls(enable_ctrl);
            if (!enable_ctrl)
            {
                ui->chk_vt_enable->setChecked(false);
                vt_enable(false);
            }
        }
    }
}

/*!
 * @brief A utility function to set an appropriate icon on the tab referenced
 *        by "tab_index" based on if the tab is enabled or not.
 *
 * @param tab_index     The tab of interest.
 * @param is_enabled    True if enabled, false if disabled.
 */
void PanelRecControls::set_tab_icon(int tab_index, bool is_enabled, int channel)
{
    int index;
    QString tab_text;

    // Get the text being shown on the tab.
    tab_text = ui->tab_io_controls->tabText(tab_index);

    // Make sure it's the "basic" text.
    index = tab_text.indexOf(QChar(']'));
    if (index >= 0)
    {
        // Get rid of the channel text and the space following.
        tab_text = tab_text.mid(index + 2);
    }

    // Are we enabling or disabling?
    if (is_enabled)
    {
        // Are we selecting a channel?
        if (channel >= 0)
            tab_text = QString("[%1] ").arg(channel) + tab_text;

        // Add the "checked" icon.
        ui->tab_io_controls->setTabIcon(tab_index, QIcon(":/images/checked.png"));
    }
    else
    {
        // Remove any icon.
        ui->tab_io_controls->setTabIcon(tab_index, QIcon());
    }

    // Update our text.
    ui->tab_io_controls->setTabText(tab_index, tab_text);
}

/*!
 * @brief Set up the custom styles for the control based on whether or not they should
 * be enabled.
 * @param enabled   True if the custom styled controls should be enabled, false if not.
 */
void PanelRecControls::set_custom_styles(bool enabled)
{
    if (enabled)
    {
        // The custom selected tab style.
        // Match the application palette setting for highlighted entry.
        QString text_style  = QApplication::palette().highlightedText().color().name();
        QString bg_style    = QApplication::palette().highlight().color().name();
        QString tab_select_style = "QTabBar::tab:selected {background-color:" + bg_style + ";color:" + text_style + ";}";
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << tab_select_style;
        ui->tab_io_controls->setStyleSheet(tab_select_style);
    }
    else
    {
        // The custom selected tab style.
        // Match the application palette setting for highlighted entry.
        QString text_style  = QApplication::palette().mid().color().name();
        QString bg_style    = QApplication::palette().window().color().name();
        QString tab_select_style = "QTabBar::tab:selected {background-color:" + bg_style + ";color:" + text_style + ";}";
        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << tab_select_style;
        ui->tab_io_controls->setStyleSheet(tab_select_style);

        // The Channel Style.
        // Only need to do this for disabling.
        ui->rb_ti_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_to_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ai_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ao_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_si_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_so_chan_0->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);

        ui->rb_ti_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_to_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ai_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ao_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_si_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_so_chan_1->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);

        ui->rb_ti_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_to_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ai_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_ao_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_si_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
        ui->rb_so_chan_2->setStyleSheet(PanelRecControls::EXTIO_ACT_INACTIVE_STYLE);
    }
}


/*!
 * @brief Polling handler for generating another external I/O activity read of GIGE_REG_EXTIO_ACTIVITY register.
 */
void PanelRecControls::poll_extio_activity()
{
    // All we want to do is read the register to determine if any activity has occurred. Only do this
    // if the panel is open. If it is closed, there is nothing for anyone to see.
    if (is_panel_open)
    {
        CameraMDI   *camera_ptr;

        // Make sure we have an active camera pointer.
        camera_ptr = p_control_panel->getCamera();
        if (camera_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "No camera pointer!";
            return;
        }

        // Is this the first time in?
        if (!extio_read_pending)
        {
            // This is a little different from other register reads. We actually want to show activity EVEN
            // WHEN the camera is busy.
            camera_ptr->start_camera_read_reg(GIGE_REG_EXTIO_ACTIVITY);
        }

        // Increment our counter.
        ++extio_read_pending;

        // Are we waiting for the read to complete?
        if (extio_read_pending == EXTIO_MAX_READ_COUNT)
        {
            // We must have missed the read complete. Reset and go around again.
            extio_read_pending = false;
            return;
        }
    }
}

/*!
 * @brief Starts the periodic polling to read GIGE_REG_EXTIO_ACTIVITY register.
 */
void PanelRecControls::start_extio_activity_polling()
{
    // Don't want this running - should never be in this state.
    if (extio_activity_timer.isActive())
        extio_activity_timer.stop();

    // Set the interval time.
    extio_activity_timer.setInterval(EXTIO_ACT_POLL_INTERVAL);

    // Connect an internal poll receiver.
    connect(&extio_activity_timer, &QTimer::timeout, this, &PanelRecControls::poll_extio_activity);

    // Start the timer.
    extio_read_pending = 0;
    extio_activity_timer.start();
}

/*!
 * @brief Style the QLabel with our warning text style.
 * @param label_to_style    The QLabel to style.
 */
void PanelRecControls::style_label_warning(QLabel *label_to_style)
{
    QString warn_style = TXT_STYLE_COLOR_WARN;
    if (label_to_style)
    {
        label_to_style->setStyleSheet(warn_style);
    }
}

/*!
 * @brief Style the QLabel with our normal text style.
 * @param label_to_style    The QLabel to style.
 */
void PanelRecControls::style_label_normal(QLabel *label_to_style)
{
    // Ask the system for the normal text color.
    QString regular_text = QApplication::palette().text().color().name();
    QString label_style = "QLabel {color:" + regular_text + ";}";
    if (label_to_style)
    {
        label_to_style->setStyleSheet(label_style);
    }
}


/*!
 * @brief For the sync-in panel when a camera disconnects reset the tab's controls to disconnected state.
 */
void PanelRecControls::sync_in_disconnect_camera()
{
    // This is just to help facilitate the disconnect U/I so that when moving back and forth
    // between opened cameras false information isn't displayed as sometimes the first read of
    // a camera connection takes a couple of seconds before the tab panel updates.
    QString label_style     = TXT_STYLE_COLOR_WAITSIG;
    QString label_text      = tr("Disconnected...");

    // Set the style for our master frame rate label based on signal status.
    ui->lbl_si_pulse_rate_value->setText(label_text);
    ui->lbl_si_pulse_rate_value->setStyleSheet(label_style);
    ui->tab_sync_in->setEnabled(false);
}

/*!
 * @brief For the sync-in panel when a camera connects reset the tab's controls to connected state.
 */
void PanelRecControls::sync_in_connect_camera()
{
    // This is just to help facilitate the disconnect U/I so that when moving back and forth
    // between opened cameras false information isn't displayed as sometimes the first read of
    // a camera connection takes a couple of seconds before the tab panel updates.
    QString label_style     = TXT_STYLE_COLOR_WAITSIG;
    QString label_text      = tr("Reading...");

    // Set the style for our master frame rate label based on signal status.
    ui->lbl_si_pulse_rate_value->setText(label_text);
    ui->lbl_si_pulse_rate_value->setStyleSheet(label_style);
    ui->tab_sync_in->setEnabled(true);
}

/*!
 * @brief Translates the UI sync out I/O register to the UI controls for sync out I/O in the panel.
 *
 * @param the gige register value for sync out I/O.
 */
void PanelRecControls::sync_out_gige_reg_to_ui(uint32_t reg_value)
{
    bool        use_duty_cycle;
    uint32_t    duty_cycle;

    // Parse the value from the register.
    duty_cycle = (reg_value & GIGE_SYNC_OUT_DUTYCYCLE_MASK) >> GIGE_SYNC_OUT_DUTYCYCLE_SHIFT;
    use_duty_cycle = true;
    if ((duty_cycle < MIN_DUTY_CYCLE) || (duty_cycle > MAX_DUTY_CYCLE))
    {
        use_duty_cycle = false;
        duty_cycle = DEFAULT_DUTY_CYCLE;
    }

    // Update the duty cycle value, as long as the user isn't sitting here.
    if (!ui->spin_so_duty_cycle->hasFocus())
    {
        // Setup our spinner.
        ui->spin_so_duty_cycle->blockSignals(true);
        ui->spin_so_duty_cycle->setValue((int)duty_cycle);
        ui->spin_so_duty_cycle->blockSignals(false);
    }

    // Setup the controls based on the type.
    switch (reg_value & GIGE_SYNC_OUT_MASK)
    {
    case GIGE_SYNC_OUT_PPS:
        ui->rb_so_per_sec->setChecked(true);
        ui->chk_so_invert_signals->setEnabled(false);
        ui->rb_so_duty_cycle->setEnabled(true);
        ui->rb_so_duty_cycle->setChecked(true);
        ui->spin_so_duty_cycle->setEnabled(true);
        ui->rb_so_shutter->setEnabled(false);
        ui->rb_so_active_low->setEnabled(true);
        ui->rb_so_active_high->setEnabled(true);
        break;
    case GIGE_SYNC_OUT_PPF:
        ui->rb_so_per_frame->setChecked(true);
        ui->chk_so_invert_signals->setEnabled(false);
        ui->rb_so_duty_cycle->setEnabled(true);
        ui->rb_so_shutter->setEnabled(true);
        ui->rb_so_active_low->setEnabled(true);
        ui->rb_so_active_high->setEnabled(true);

        // Are we using the duty cycle?
        if (use_duty_cycle)
            ui->rb_so_duty_cycle->setChecked(true);
        else
            ui->rb_so_shutter->setChecked(true);
        break;
    default:
    case GIGE_SYNC_OUT_IS_SYNC_IN:
        ui->rb_so_pass_thru->setChecked(true);
        ui->chk_so_invert_signals->setEnabled(true);
        ui->rb_so_duty_cycle->setEnabled(false);
        ui->rb_so_shutter->setEnabled(false);
        ui->rb_so_shutter->setChecked(true);
        ui->rb_so_active_low->setEnabled(false);
        ui->rb_so_active_high->setEnabled(false);

        // Are we inverting?
        ui->chk_so_invert_signals->setChecked((reg_value & GIGE_SYNC_OUT_INVERT_SYNCIN) == GIGE_SYNC_OUT_INVERT_SYNCIN);
        break;
    }

    // Are we active high? Only set if the user isn't parked here.
    if (!ui->rb_so_active_high->hasFocus() && !ui->rb_so_active_low->hasFocus())
    {
        if ((reg_value & GIGE_SYNC_OUT_ACTIVE_HIGH) == GIGE_SYNC_OUT_ACTIVE_HIGH)
            ui->rb_so_active_high->setChecked(true);
        else
            ui->rb_so_active_low->setChecked(true);
    }
}

/*!
 * @brief Queries the QSettings for Record Control Sync Out to retrieve the register setting.
 *
 * @param sync_out_reg_data     [OUT] The restored SyncOut settings..
 * @returns True if found an entry in sync out data settings, false if not.
 */
bool PanelRecControls::sync_out_load_settings(uint32_t& sync_out_reg_data)
{
    CameraMDI   *camera_ptr = nullptr;
    bool        sync_out_data_found = false;

    // Initialize the output data.
    sync_out_reg_data = GIGE_SYNC_OUT_DISABLE;

    // Conditionalize this for testing purposes.
    if (USE_SYNC_OUT_QSETTINGS)
    {
        // There should be a control panel pointer.
        if (p_control_panel)
        {
            camera_ptr = p_control_panel->getCamera();
            if (camera_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
            }
            else
            {
                // Attempt to load the settings from the persistent store.
                sync_out_data_found = SyncOutDataRepository::loadSyncOutData(camera_ptr->get_ip_addr(), sync_out_reg_data);
            }
        }
    }

    return sync_out_data_found;
}

/*!
 * @brief Persist the ui setup into the QSettings for Record Controls
 * Sync Out register QSettings.
 */
void PanelRecControls::sync_out_save_settings(uint32_t sync_out_reg)
{
    CameraMDI   *camera_ptr = nullptr;

    // Conditionalize this for testing purposes.
    if (USE_SYNC_OUT_QSETTINGS)
    {
        // There should be a control panel pointer.
        if (p_control_panel)
        {
            camera_ptr = p_control_panel->getCamera();
            if (camera_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
            }
            else
            {
                SyncOutDataRepository::saveSyncOutData(camera_ptr->get_ip_addr(), sync_out_reg);
            }
        }
    }
}

/*!
 * @brief Translates the ui sync_out settings to the
 * value of the sync out register.
 *
 * @returns the gige register value based on sync-out ui or zero if there
 * is something wrong with the ui setup.
 */
uint32_t PanelRecControls::sync_out_ui_to_gige_reg()
{
    uint32_t    reg_value = 0;

    // Are we enabled?
    if (ui->sync_out_enable->isChecked())
    {
        // Initial setting.
        reg_value = GIGE_SYNC_OUT_ENABLE;

        // Add in the channel selected.
        if (sync_out_chan >= 0)
            reg_value |= (1 << (GIGE_SYNC_OUT_CHANNEL_SHIFT + sync_out_chan));
    }
    else
    {
        // Initial setting.
        reg_value = GIGE_SYNC_OUT_DISABLE;
    }

    // Are we in pass-thru mode?
    if (ui->rb_so_pass_thru->isChecked())
    {
        // We're in pass-thru mode.
        reg_value |= GIGE_SYNC_OUT_IS_SYNC_IN;

        // Are we inverting the signal?
        if (ui->chk_so_invert_signals->isChecked())
            reg_value |= GIGE_SYNC_OUT_INVERT_SYNCIN;
        else
            reg_value |= GIGE_SYNC_OUT_PASS_SYNCIN;
    }
    else
    {
        // Did the user specify a duty-cycle?
        if (ui->rb_so_duty_cycle->isChecked())
        {
            uint32_t    duty_value;

            // Read up the value and set it into the register.
            duty_value = (uint32_t)ui->spin_so_duty_cycle->value();
            reg_value |= (duty_value << GIGE_SYNC_OUT_DUTYCYCLE_SHIFT);
        }

        // Are we per frame or second?
        if (ui->rb_so_per_frame->isChecked())
        {
            // We're in per-frame mode.
            reg_value |= GIGE_SYNC_OUT_PPF;
        }
        else
        {
            // We're in per-second mode.
            reg_value |= GIGE_SYNC_OUT_PPS;
        }

        // Are we rising or falling?
        reg_value |= (ui->rb_so_active_high->isChecked() ? GIGE_SYNC_OUT_ACTIVE_HIGH : GIGE_SYNC_OUT_ACTIVE_LOW);
    }

    return reg_value;
}

/*!
 * @brief Translate the UI elements for this object.
 */
void PanelRecControls::translate_ui()
{
    int i;
    QList<QWidget *> widgetList = findChildren<QWidget *>();

    // Block all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(true);

    // Translate our user interface.
    ui->retranslateUi(this);

    // Unblock all signals.
    for (i = 0; i < widgetList.count(); ++i)
        widgetList.at(i)->blockSignals(false);
}

/*!
 * @brief Update the ARM-In controls based on the registers read from the camera.
 */
void PanelRecControls::update_arm_in()
{
    uint32_t    reg_value;

    // Check to make sure we're not in the middle of an update.
    if (update_in_progress)
        return;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_ARM_IN_ENABLE);

    // Are we enabled?
    if ((reg_value & GIGE_ARM_TYPE_EXTERNAL) == GIGE_ARM_TYPE_EXTERNAL)
    {
        // Were we disabled before? Only do this if we've changed.
        if (!ui->arm_in_enable->isChecked() && !ui->arm_in_enable->hasFocus())
        {
            ui->arm_in_enable->setChecked(true);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_arm_in->setEnabled(ui->arm_in_enable->isChecked());

        // The following does not apply to series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            uint32_t channel_bit;

            // Release whatever channel we had before.
            release_channel(arm_in_chan);

            // What channel is being used for the arm-in channel?
            channel_bit = (reg_value >> GIGE_ARM_IN_CHANNEL_SHIFT) & MAX_CHANNEL_MASK;
            for (arm_in_chan = 0; arm_in_chan < MAX_CHANNEL; ++arm_in_chan)
            {
                // Is this the one?
                if (channel_bit == (uint32_t)(1 << arm_in_chan))
                {
                    // Mark this channel as "in use".
                    channels_in_use |= channel_bit;
                    break;
                }
            }
            if (arm_in_chan == MAX_CHANNEL)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find arm_in channel, reg_value:" << Qt::showbase << Qt::hex << reg_value << "channel_bit:" << channel_bit;
                arm_in_chan = -1;
            }

            // Enable the channel radio buttons and check the appropriate one.
            ui->frm_ai_chan->setEnabled(true);
            ui->rb_ai_chan_0->setChecked(arm_in_chan == 0);
            ui->rb_ai_chan_1->setChecked(arm_in_chan == 1);
            ui->rb_ai_chan_2->setChecked(arm_in_chan == 2);

            // Update which buttons the user can now select.
            update_channel_buttons(ui->rb_ai_chan_0, ui->rb_ai_chan_1, ui->rb_ai_chan_2, arm_in_chan);
        }

        // Display the tab icon as appropriate.
        set_tab_icon(TAB_ARM_IN, true, arm_in_chan);
    }
    else
    {
        // Doesn't apply to series3.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Not using any channels.
            ui->rb_ai_chan_0->setChecked(false);
            ui->rb_ai_chan_1->setChecked(false);
            ui->rb_ai_chan_2->setChecked(false);
            ui->frm_ai_chan->setEnabled(false);

            // Release the current channel.
            if (arm_in_chan >= 0)
            {
                release_channel(arm_in_chan);
                arm_in_chan = -1;
            }
        }

        // Were we enabled before?
        if (ui->arm_in_enable->isChecked() && !ui->arm_in_enable->hasFocus())
        {
            ui->arm_in_enable->setChecked(false);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_arm_in->setEnabled(ui->arm_in_enable->isChecked());

        // Make sure we reset the icon.
        set_tab_icon(TAB_ARM_IN, false);
    }

    // Get the preferences register. See if we're allowing automatic discard
    // of unsaved images.
    reg_value = read_camera_cache(GIGE_REG_PREFERENCES);

    // Don't update if the user is parked here.
    if (!ui->chk_allow_discard->hasFocus())
    {
        if ((reg_value & GIGE_PREF_ARM_IN_OVERWRITE) == GIGE_PREF_ARM_IN_OVERWRITE)
            ui->chk_allow_discard->setChecked(true);
        else
            ui->chk_allow_discard->setChecked(false);
    }

    // Now pull in the rest of the register settings.
    reg_value = read_camera_cache(GIGE_REG_ARM_IN_POLARITY);

    // Are we edge or level? Don't update if the user is parked here.
    if (!ui->rb_ai_edge_type->hasFocus() && !ui->rb_ai_level_type->hasFocus())
    {
        if ((reg_value & GIGE_ARM_IN_EDGE_TYPE) == GIGE_ARM_IN_EDGE_TYPE)
        {
            ui->rb_ai_edge_type->setChecked(true);
            ui->rb_ai_rising_high->setEnabled(true);
        }
        else
        {
            ui->rb_ai_level_type->setChecked(true);
            ui->rb_ai_rising_high->setEnabled(false);
        }
    }

    // Are we rising or falling? Don't update if the user is parked here.
    if (!ui->rb_ai_falling_low->hasFocus() && !ui->rb_ai_rising_high->hasFocus())
    {
        if ((reg_value & GIGE_ARM_IN_RISING) == GIGE_ARM_IN_RISING)
            ui->rb_ai_rising_high->setChecked(true);
        else
            ui->rb_ai_falling_low->setChecked(true);
    }
}

/*!
 * @brief Update the ARM-Out controls based on the registers read from the camera.
 */
void PanelRecControls::update_arm_out()
{
    uint32_t    reg_value;

    // Check to make sure we're not in the middle of an update.
    if (update_in_progress)
        return;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_ARM_OUT_SELECT);

    // Are we enabled?
    if ((reg_value & GIGE_ARM_OUT_ENABLE) == GIGE_ARM_OUT_ENABLE)
    {
        // Were we disabled before? Only do this if we've changed.
        if (!ui->arm_out_enable->isChecked() && !ui->arm_out_enable->hasFocus())
        {
            ui->arm_out_enable->setChecked(true);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_arm_out->setEnabled(ui->arm_out_enable->isChecked());

        // The following does not apply to series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            uint32_t channel_bit;

            // Release whatever channel we had before.
            release_channel(arm_out_chan);

            // What channel is being used for the external trigger?
            channel_bit = (reg_value >> GIGE_ARM_OUT_CHANNEL_SHIFT) & MAX_CHANNEL_MASK;
            for (arm_out_chan = 0; arm_out_chan < MAX_CHANNEL; ++arm_out_chan)
            {
                // Is this the one?
                if (channel_bit == (uint32_t)(1 << arm_out_chan))
                {
                    // Mark this channel as "in use".
                    channels_in_use |= channel_bit;
                    break;
                }
            }
            if (arm_out_chan == MAX_CHANNEL)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find arm out channel, reg_value:" << Qt::showbase << Qt::hex << reg_value << "channel_bit:" << channel_bit;
                arm_out_chan = -1;
            }

            // Enable the channel radio buttons and check the appropriate one.
            ui->frm_ao_chan->setEnabled(true);
            ui->rb_ao_chan_0->setChecked(arm_out_chan == 0);
            ui->rb_ao_chan_1->setChecked(arm_out_chan == 1);
            ui->rb_ao_chan_2->setChecked(arm_out_chan == 2);

            // Update which buttons the user can now select.
            update_channel_buttons(ui->rb_ao_chan_0, ui->rb_ao_chan_1, ui->rb_ao_chan_2, arm_out_chan);
        }

        // Display the tab icon as appropriate.
        set_tab_icon(TAB_ARM_OUT, true, arm_out_chan);
    }
    else
    {
        // Doesn't apply to series3.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Not using any channels.
            ui->rb_ao_chan_0->setChecked(false);
            ui->rb_ao_chan_1->setChecked(false);
            ui->rb_ao_chan_2->setChecked(false);
            ui->frm_ao_chan->setEnabled(false);

            // Release the current channel.
            if (arm_out_chan >= 0)
            {
                release_channel(arm_out_chan);
                arm_out_chan = -1;
            }
        }

        // Were we enabled before?
        if (ui->arm_out_enable->isChecked() && !ui->arm_out_enable->hasFocus())
        {
            ui->arm_out_enable->setChecked(false);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_arm_out->setEnabled(ui->arm_out_enable->isChecked());

        // Make sure we reset the icon.
        set_tab_icon(TAB_ARM_OUT, false);
    }

    // Update the pass-thru controls, assuming the user isn't parked there.
    if ((reg_value & GIGE_ARM_OUT_IS_ARM_IN) == GIGE_ARM_OUT_IS_ARM_IN)
    {
        ui->chk_ao_pass_thru->setChecked(true);
        ui->chk_ao_invert_signal->setEnabled(true);
        ui->chk_ao_invert_signal->setChecked((reg_value & GIGE_ARM_OUT_POLARITY_MASK) == GIGE_ARM_OUT_INVERT_ARMIN);

        // Since we're in pass-thru, these controls are disabled.
        ui->rb_ao_active_high->setEnabled(false);
        ui->rb_ao_active_low->setEnabled(false);
        ui->chk_ao_pulse_width->setEnabled(false);
    }
    else
    {
        uint32_t    pulse_width;

        // Not doing pass-through, so make sure it's cleared.
        ui->chk_ao_pass_thru->setChecked(false);
        ui->chk_ao_invert_signal->setEnabled(false);
        ui->chk_ao_invert_signal->setChecked(false);

        // Not doing pass-thru, so enable these controls.
        ui->rb_ao_active_high->setEnabled(true);
        ui->rb_ao_active_low->setEnabled(true);
        ui->chk_ao_pulse_width->setEnabled(true);

        // So...active high or low?
        if ((reg_value & GIGE_ARM_OUT_POLARITY_MASK) == GIGE_ARM_OUT_ACTIVE_HIGH)
            ui->rb_ao_active_high->setChecked(true);
        else
            ui->rb_ao_active_low->setChecked(true);

        // Get the pulse width.
        pulse_width = (reg_value & GIGE_ARM_OUT_PERIOD_MASK) >> GIGE_ARM_OUT_PERIOD_SHIFT;
        if (pulse_width)
        {
            ui->chk_ao_pulse_width->setChecked(true);
            ui->spin_ao_pulse_width->setEnabled(true);

            // Don't pass signals.
            ui->spin_ao_pulse_width->blockSignals(true);
            ui->spin_ao_pulse_width->setValue((int)pulse_width);
            ui->spin_ao_pulse_width->blockSignals(false);
        }
        else
        {
            ui->chk_ao_pulse_width->setChecked(false);
            ui->spin_ao_pulse_width->setEnabled(false);

            // Don't pass signals.
            ui->spin_ao_pulse_width->blockSignals(true);
            ui->spin_ao_pulse_width->setValue(DEFAULT_PW_VALUE);
            ui->spin_ao_pulse_width->blockSignals(false);
        }
    }
}

void PanelRecControls::update_channel_buttons(QRadioButton *chan1, QRadioButton *chan2, QRadioButton *chan3, int used_channel)
{
    int i;

    // Disable channels which are in use.
    for (i = 0; i < MAX_CHANNEL; ++i)
    {
        bool enable(false);

        // If this bit is not set, OR we're looking at the channel
        // we are using, enable the button.
        if (((channels_in_use & (uint32_t)(1 << i)) == 0) || (i == used_channel))
            enable = true;

        // Adjust the corresponding button.
        switch (i)
        {
        case 0:
            chan1->setEnabled(enable);
            break;
        case 1:
            chan2->setEnabled(enable);
            break;
        case 2:
            chan3->setEnabled(enable);
            break;
        default:
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "WHAT!! Unexpected index:" << i;
            break;
        }
    }
}

/*!
 * @brief Update the external trigger controls based on the register settings.
 */
void PanelRecControls::update_ext_trigger()
{
    int         ext_select;
    uint32_t    reg_value;
    bool        rising;

    // Are we in the middle of an update?
    if (update_in_progress)
        return;

    // Determine if we're rising or falling.
    reg_value = read_camera_cache(GIGE_REG_EXT_TRIG_POLARITY);
    if ((reg_value & GIGE_EXT_TRIG_POLARITY_RISING) == GIGE_EXT_TRIG_POLARITY_RISING)
        ext_select = EXT_TRIG_RISING;
    else
        ext_select = EXT_TRIG_FALLING;
    rising = (ext_select == EXT_TRIG_RISING);

    // Get the external trigger enable flag.
    reg_value = read_camera_cache(GIGE_REG_EXT_TRIG_ENABLE);
    if ((reg_value & GIGE_EXT_TRIG_ENABLE_MASK) != GIGE_EXT_TRIG_ENABLE_MASK)
    {
        // External trigger is disabled.
        ext_select = EXT_TRIG_DISABLED;

        // Doesn't apply to series3.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Not using any channels.
            ui->rb_ti_chan_0->setChecked(false);
            ui->rb_ti_chan_1->setChecked(false);
            ui->rb_ti_chan_2->setChecked(false);
            ui->frm_ti_chan->setEnabled(false);

            // Did we have a external trigger channel?
            if (trig_in_chan >= 0)
            {
                release_channel(trig_in_chan);
                trig_in_chan = -1;
            }
        }
    }

    // The following doesn't apply to series3 cameras.
    else if (!p_control_panel->getCamera()->is_series3())
    {
        uint32_t channel_bit;

        // Release whatever channel we had before.
        release_channel(trig_in_chan);

        // What channel is being used for the external trigger?
        channel_bit = (reg_value >> GIGE_EXT_TRIG_CHANNEL_SHIFT) & MAX_CHANNEL_MASK;
        for (trig_in_chan = 0; trig_in_chan < MAX_CHANNEL; ++trig_in_chan)
        {
            // Is this the one?
            if (channel_bit == (uint32_t)(1 << trig_in_chan))
            {
                // Mark this channel as "in use".
                channels_in_use |= channel_bit;
                break;
            }
        }
        if (trig_in_chan == MAX_CHANNEL)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find trig in channel, reg_value:" << Qt::showbase << Qt::hex << reg_value << "channel_bit:" << channel_bit;
            trig_in_chan = -1;
        }

        // Enable the channel radio buttons and check the appropriate one.
        ui->frm_ti_chan->setEnabled(true);
        ui->rb_ti_chan_0->setChecked(trig_in_chan == 0);
        ui->rb_ti_chan_1->setChecked(trig_in_chan == 1);
        ui->rb_ti_chan_2->setChecked(trig_in_chan == 2);

        // Update which buttons the user can now select.
        update_channel_buttons(ui->rb_ti_chan_0, ui->rb_ti_chan_1, ui->rb_ti_chan_2, trig_in_chan);
    }

    // Setup our controls
    ui->trig_in_enable->setChecked(ext_select != EXT_TRIG_DISABLED);
    ui->frame_trig_in->setEnabled(ext_select != EXT_TRIG_DISABLED);
    ui->rb_ti_falling->setChecked(!rising);
    ui->rb_ti_rising->setChecked(rising);

    // Display the tab icon if we're enabled..
    set_tab_icon(TAB_TRIG_IN, (ext_select != EXT_TRIG_DISABLED), trig_in_chan);
}

/*!
 * @brief Update our IRIG controls based on registers we've read from the camera.
 */
void PanelRecControls::update_irig()
{
    bool        is_enabled;
    bool        use_year;
    uint32_t    reg_value;

    // Are we in the middle of an update?
    if (update_in_progress)
        return;

    // Is IRIG allowed on this camera?
    reg_value = read_camera_cache(GIGE_REG_IRIG_STATUS);
    if ((reg_value & GIGE_IRIG_ALLOWED) == 0)
    {
        ui->frm_irig->setVisible(false);
        return;
    }

    // Make sure we make the IRIG controls visible.
    ui->frm_irig->setVisible(true);

    // Read the control register and parse out the information.
    reg_value = read_camera_cache(GIGE_REG_IRIG_CONTROL);
    is_enabled = (reg_value & GIGE_IRIG_ENABLE) == (uint32_t)GIGE_IRIG_ENABLE;
    use_year = (reg_value & GIGE_IRIG_USE_YEAR) == (uint32_t)GIGE_IRIG_USE_YEAR;

    // Is IRIG enabled?
    ui->chk_irig->setChecked(is_enabled);
    ui->chk_irig_year->setEnabled(is_enabled);
    ui->chk_irig_year->setChecked(is_enabled && use_year);

    // Based on if IRIG is enabled or not changes how we label the I/O control.
    if (is_enabled)
        ui->rb_si_per_sec->setText(tr("IRIG"));
    else
        ui->rb_si_per_sec->setText(tr("Per Sec"));
}

/*!
 * @brief Utility function to calculate the maximum number of frames based on the width/height and
 *        if we're in LTR mode or not.
 */
void PanelRecControls::update_num_frames()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera pointer!";
        return;
    }

    // Ask the camera how many frames the settings support.
    num_frames = camera_ptr->calc_max_frames();

    // Update our display.
    sld_trigger_pos_changed(ui->sld_trig_pos->value());
}

/*!
 * @brief Virtual function called by the "parent" widget to have us update the
 *        contents of our panel.
 */
void PanelRecControls::update_panel()
{
    CameraMDI   *camera_ptr;

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "***** Updating Rec Control Panel *****";
    }

    // Do we have a camera connected?
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Stop polling the camera.
    if (update_timer_id != 0)
    {
        killTimer(update_timer_id);
        update_timer_id = 0;
    }

    // Initialize our video trigger controls.
    ui->spin_vt_area->setValue(camera_ptr->get_vt_area());
    ui->spin_vt_change->setValue(camera_ptr->get_vt_change());
    ui->spin_refresh->setValue(camera_ptr->get_vt_refresh());
    ui->chk_vt_enable->setChecked(camera_ptr->vt_reticle_visible());
    ui->chk_vt_marker_only->setChecked(camera_ptr->get_vt_mark_only());

    // Read registers we're interested in.
    read_registers();

    // Make sure we get a current version of GIGE_REG_PREFERENCES.
    // This isn't backwards compatible, so don't poll it.
    start_camera_read(GIGE_REG_PREFERENCES);

    // Restart our timer.
    update_timer_id = startTimer(POLLING_PERIOD);
}

/*!
 * @brief Update the Sync-In controls based on registers read from the camera.
 */
void PanelRecControls::update_sync_in()
{
    uint32_t    divisor;
    uint32_t    divisor_min;
    uint32_t    frame_rate;
    uint32_t    master_framerate;
    uint32_t    max_roi_framerate;
    uint32_t    measured_framerate = 0;
    uint32_t    pps;
    uint32_t    sync_in_type;
    uint32_t    sync_in_time = 0;
    uint32_t    skip_pulse;
    uint32_t    reg_value;
    QString     s_field;
    QString     mfr_nosignal    = tr("No Signal");
    QString     label_str       = mfr_nosignal;
    QString     label_style     = TXT_STYLE_COLOR_NOSIG;

    // Check to make sure we're not in the middle of an update.
    if (update_in_progress)
        return;

    // Read the sync-in settings register.
    reg_value = read_camera_cache(GIGE_REG_SYNC_IN_ENABLE);

    // Read in the sync-in timing in case we are enabled as we want to apply styling to information.
    sync_in_time = read_camera_cache(GIGE_REG_SYNC_IN_TIMING);

    // Calculate the measured framerate based on the sync-in pulse rate.
    if (sync_in_time > 0)
    {
        uint32_t sync_in_time_up = (sync_in_time + 1);  // As per old code, it bumps up register setting by one.

        uint32_t modulo = FASMOTION_SETTINGS::FPGA_CLK_RATE % sync_in_time_up;
        measured_framerate = FASMOTION_SETTINGS::FPGA_CLK_RATE / sync_in_time_up;

        // Check for odd framerates.
        if (modulo % 10)
        {
            // Per request, this was making one person very unhappy when the arithmetic was off-by-one
            // on the integer math result. This is attempt to make it consistent with master frame rate.
            measured_framerate++;
        }
    }

    // Mark the label style for the sync-in pulse rate rate to default.
    ui->lbl_si_pulse_rate_value->setText(label_str);
    ui->lbl_si_pulse_rate_value->setStyleSheet(label_style);

    // Parse out the settings.
    sync_in_type = reg_value & GIGE_SYNC_TYPE_MASK;

// Per ticket #227, don't allow this processing.
#ifdef TICKET_227
    pps = ((reg_value & GIGE_SYNC_IN_PULSE_MASK) == GIGE_SYNC_IN_PPS);
#else
    pps = false;
#endif /* TICKET_227 */
    skip_pulse = (reg_value & GIGE_SYNC_IN_SKIP_MASK) >> GIGE_SYNC_IN_SKIP_SHIFT;

    // If we using an external sync signal, then we're enabled. Otherwise, this
    // panel is disabled.
    if (sync_in_type == GIGE_SYNC_TYPE_EXTERNAL)
    {
        // If the sync-in time is > 0 we will style using active signal status.
        if (sync_in_time > 0)
        {
            label_style = TXT_STYLE_COLOR_ACTSIG;
        }
        else
        {
            label_style = TXT_STYLE_COLOR_NOSIG;
        }

        // Set the style for our master frame rate label based on signal status.
        ui->lbl_si_pulse_rate_value->setStyleSheet(label_style);

        // Were we disabled before? Only do this if we've changed.
        if (!ui->sync_in_enable->isChecked())
        {
            ui->sync_in_enable->setChecked(true);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_sync_in->setEnabled(ui->sync_in_enable->isChecked());

        // The following does not apply to series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            uint32_t channel_bit;

            // Release whatever channel we had before.
            release_channel(sync_in_chan);

            // What channel is being used for the sync in?
            channel_bit = (reg_value >> GIGE_SYNC_IN_CHANNEL_SHIFT) & MAX_CHANNEL_MASK;
            for (sync_in_chan = 0; sync_in_chan < MAX_CHANNEL; ++sync_in_chan)
            {
                // Is this the one?
                if (channel_bit == (uint32_t)(1 << sync_in_chan))
                {
                    // Mark this channel as "in use".
                    channels_in_use |= channel_bit;
                    break;
                }
            }
            if (sync_in_chan == MAX_CHANNEL)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find sync in channel, reg_value:" << Qt::showbase << Qt::hex << reg_value << "channel_bit:" << channel_bit;
                sync_in_chan = -1;
            }

            // Enable the channel radio buttons and check the appropriate one.
            ui->frm_si_chan->setEnabled(true);
            ui->rb_si_chan_0->setChecked(sync_in_chan == 0);
            ui->rb_si_chan_1->setChecked(sync_in_chan == 1);
            ui->rb_si_chan_2->setChecked(sync_in_chan == 2);

            // Update which buttons the user can now select.
            update_channel_buttons(ui->rb_si_chan_0, ui->rb_si_chan_1, ui->rb_si_chan_2, sync_in_chan);
        }

        // Display the tab icon as appropriate.
        set_tab_icon(TAB_SYNC_IN, true, sync_in_chan);
    }
    else
    {
        // Doesn't apply to series3.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Not using any channels.
            ui->rb_si_chan_0->setChecked(false);
            ui->rb_si_chan_1->setChecked(false);
            ui->rb_si_chan_2->setChecked(false);
            ui->frm_si_chan->setEnabled(false);

            // Release the current channel.
            if (sync_in_chan >= 0)
            {
                release_channel(sync_in_chan);
                sync_in_chan = -1;
            }
        }

        // Were we enabled before?
        if (ui->sync_in_enable->isChecked())
        {
            ui->sync_in_enable->setChecked(false);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_sync_in->setEnabled(ui->sync_in_enable->isChecked());

        // Make sure we reset the icon.
        set_tab_icon(TAB_SYNC_IN, false);
    }

    // The register is 0-based, but the UI is 1-based, so adjust.
    divisor = skip_pulse + 1;

    // Get the master framerate.
    master_framerate = get_master_framerate();

    // Calculate our minimum divisor we can support.
    max_roi_framerate = get_max_framerate();
    if (max_roi_framerate > 0)
    {
        divisor_min = (uint32_t)ceil((((double)master_framerate) / ((double)max_roi_framerate)));
        if (divisor_min == 0)
            divisor_min = 1;
    }
    else
    {
        divisor_min = 1;
    }

    // Make sure our divisor value is OK.
    if (divisor < divisor_min)
        divisor = divisor_min;

    // If Sync-In is enabled, write out warnings in the log and on the panel if the divisor exceeds our maximum divisor.
    if (ui->sync_in_enable->isChecked() && (divisor > MAX_DIVISOR_VALUE))
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "SYNC-IN DIVISOR OVER LIMIT:" << divisor;

        // The master framerate will exceed max_roi_framerate with out-of-range maximum divisor value.
        // We aren't going to do anything about it other than show the calculated red and pop up our warning.
        QString warn_txt = tr("Expected Pulse Rate too high for Maximum Framerate ");
        ui->lbl_si_mfr_too_high->setText(warn_txt);
        ui->lbl_si_mfr_too_high->setVisible(true);

        // Set the style as warning text in our panel.
        style_label_warning(ui->lbl_si_mfr_too_high);
        style_label_warning(ui->lbl_si_calc_frame_rate);
    }
    else
    {
        // Just hide our warning it doesn't apply.
        ui->lbl_si_mfr_too_high->setVisible(false);

        // Set the style as normal text in our panel.
        style_label_normal(ui->lbl_si_calc_frame_rate);
    }

    // Display what the user is defining for our framrate
    frame_rate = master_framerate / divisor;
    s_field = QString(tr("Calculated Rate: %1 ")).arg(frame_rate);
    ui->lbl_si_calc_frame_rate->setText(s_field);
    s_field = QString(tr("Maximum Framerate: %1 ")).arg(max_roi_framerate);
    ui->lbl_si_mfr_max->setText(s_field);

// Per ticket #227, don't allow this processing.
#ifdef TICKET_227
    // Are we pulse-per-second or per-frame? Don't update if we're sitting here.
    if (!ui->rb_si_per_frame->hasFocus() && !ui->rb_si_per_sec->hasFocus())
    {
        if (pps)
            ui->rb_si_per_sec->setChecked(true);
        else
            ui->rb_si_per_frame->setChecked(true);
    }
#else
    ui->rb_si_per_frame->setChecked(true);
#endif /* TICKET_227 */

    // Only calculate the master frame rate if in ppf (pulse per frame) and enabled.
    if (!pps && (sync_in_type == GIGE_SYNC_TYPE_EXTERNAL))
    {
        // The sync_in_time was read in the beginning of this process.
        if (sync_in_time > 0)
        {
            label_str = tr("%1").arg(measured_framerate);
            label_style = TXT_STYLE_COLOR_ACTSIG;
        }
        else
        {
            // Use a default label for per-frame mode but drop in status of no signal.
            label_str = mfr_nosignal;
            label_style = TXT_STYLE_COLOR_NOSIG;
        }
    }

    // Set the label to the string and style calculated.
    ui->lbl_si_pulse_rate_value->setText(label_str);
    ui->lbl_si_pulse_rate_value->setStyleSheet(label_style);

    // Initialize our spinners. But don't let the signals propogate from here.
    ui->spin_si_mfr->blockSignals(true);
    if (!ui->spin_si_mfr->hasFocus())
        ui->spin_si_mfr->setValue((int)master_framerate);
    // If not in live-video mode it makes no sense to allow user to modify master frame rate.
    if (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO)
    {
        ui->spin_si_mfr->setEnabled(false);
    }
    else
    {
        // Otherwise set it based on the pps setting.
        ui->spin_si_mfr->setEnabled(!pps);
    }
    ui->spin_si_mfr->blockSignals(false);

    ui->spin_si_rate_div->blockSignals(true);
    ui->spin_si_rate_div->setMinimum((int)divisor_min);
    ui->spin_si_rate_div->setMaximum(MAX_DIVISOR_VALUE);
    if (!ui->spin_si_rate_div->hasFocus())
        ui->spin_si_rate_div->setValue((int)divisor);
    // If not in live-video mode it makes no sense to allow user to modify divisor for master frame rate.
    if (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO)
    {
        ui->spin_si_rate_div->setEnabled(false);
    }
    else
    {
        // Otherwise set it based on the pps setting.
        ui->spin_si_rate_div->setEnabled(!pps);
    }
    ui->spin_si_rate_div->blockSignals(false);

// Per ticket #227, do not process per-second controls.
#ifdef TICKET_227
    // Read the delay from the camera.
    ui->spin_si_delay->blockSignals(true);
    if (!ui->spin_si_delay->hasFocus())
    {
        reg_value = read_camera_cache(GIGE_REG_SYNC_DELAY);
        ui->spin_si_delay->setValue((int)reg_value);
    }
    ui->spin_si_delay->setEnabled(pps);
    ui->spin_si_delay->blockSignals(false);
#endif /* TICKET_227 */

    // Determine if we're rising or falling. Don't update if the user
    // is sitting on these controls.
    if (!ui->rb_si_falling->hasFocus() && !ui->rb_si_rising->hasFocus())
    {
        reg_value = read_camera_cache(GIGE_REG_SYNC_IN_POLARITY);
        if ((reg_value & GIGE_SYNC_IN_POL_MASK) == GIGE_SYNC_IN_RISING)
            ui->rb_si_rising->setChecked(true);
        else
            ui->rb_si_falling->setChecked(true);
    }
}

/*!
 * @brief Update the Sync-Out controls based on registers read from the camera.
 */
void PanelRecControls::update_sync_out()
{
    uint32_t    reg_value;

    // Check to make sure we're not in the middle of an update.
    if (update_in_progress)
        return;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_SYNC_OUT_SELECT);

    // There is a case where we may have never saved this camera's enabled state to
    // the sync-out QSettings data repository.
    // It is for that case that I want to make sure the sync-out data is written to the
    // QSettings repository because if it is not there already then as soon as sync-out is
    // disabled during the operation of the panel these settings will be lost if the only
    // thing that changes is the enable checkbox.  Once a sync-out QSettings data repository
    // entry exists for camera this code will not be invoked for the camera on this system.
    if ((reg_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_ENABLE)
    {
        uint32_t sync_out_reg = GIGE_SYNC_OUT_DISABLE;

        // This determines if there is an entry for this camera.
        if (!sync_out_load_settings(sync_out_reg))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "New Camera Added to Sync-Out Data Repository.";
            sync_out_save_settings(reg_value);
        }

        // The following does not apply to series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            uint32_t channel_bit;

            // Release whatever channel we had before.
            release_channel(sync_out_chan);

            // What channel is being used for the sync out?
            channel_bit = (reg_value >> GIGE_SYNC_OUT_CHANNEL_SHIFT) & MAX_CHANNEL_MASK;
            for (sync_out_chan = 0; sync_out_chan < MAX_CHANNEL; ++sync_out_chan)
            {
                // Is this the one?
                if (channel_bit == (uint32_t)(1 << sync_out_chan))
                {
                    // Mark this channel as "in use".
                    channels_in_use |= channel_bit;
                    break;
                }
            }
            if (sync_out_chan == MAX_CHANNEL)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find sync out channel, reg_value:" << Qt::showbase << Qt::hex << reg_value << "channel_bit:" << channel_bit;
                sync_out_chan = -1;
            }

            // Enable the channel radio buttons and check the appropriate one.
            ui->frm_so_chan->setEnabled(true);
            ui->rb_so_chan_0->setChecked(sync_out_chan == 0);
            ui->rb_so_chan_1->setChecked(sync_out_chan == 1);
            ui->rb_so_chan_2->setChecked(sync_out_chan == 2);

            // Update which buttons the user can now select.
            update_channel_buttons(ui->rb_so_chan_0, ui->rb_so_chan_1, ui->rb_so_chan_2, sync_out_chan);
        }

        // Were we enabled before?
        if (!ui->sync_out_enable->isChecked())
        {
            ui->sync_out_enable->setChecked(true);
        }

        // Make sure we reflect the state of the checkbox.
        ui->frame_sync_out->setEnabled(ui->sync_out_enable->isChecked());

        // Display the tab icon as appropriate.
        set_tab_icon(TAB_SYNC_OUT, true, sync_out_chan);
    }
    else
    {
        // Doesn't apply to series3.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Not using any channels.
            ui->rb_so_chan_0->setChecked(false);
            ui->rb_so_chan_1->setChecked(false);
            ui->rb_so_chan_2->setChecked(false);
            ui->frm_so_chan->setEnabled(false);

            // Release the current channel.
            if (sync_out_chan >= 0)
            {
                release_channel(sync_out_chan);
                sync_out_chan = -1;
            }
        }

        // Were we enabled before?
        if (ui->sync_out_enable->isChecked())
        {
            ui->sync_out_enable->setChecked(false);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_sync_out->setEnabled(ui->sync_out_enable->isChecked());

        // Display the tab icon as appropriate.
        set_tab_icon(TAB_SYNC_OUT, false);
    }

    // Update the UI to match the settings in the GIGE register.
    sync_out_gige_reg_to_ui(reg_value);

    // Make sure our duty cycle is refreshed to match camera fps.
    update_duty_cycle();
}

/*!
 * @brief Update the Trigger output controls based on registers read from the camera.
 */
void PanelRecControls::update_trig_out()
{
    uint32_t    reg_value;

    // Check to make sure we're not in the middle of an update.
    if (update_in_progress)
        return;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_TRIG_OUT_SELECT);

    // Determine if we're enabled or not.
    if ((reg_value & GIGE_TRIG_OUT_ENABLE) == GIGE_TRIG_OUT_ENABLE)
    {
        bool    do_passthru;
        bool    active_high;
        uint32_t    pulse_width;

        // Any change from where we were?
        if (!ui->trig_out_enable->isChecked())
        {
            // We are enabled.
            ui->trig_out_enable->setChecked(true);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_trig_out->setEnabled(ui->trig_out_enable->isChecked());

        // Is passthru setup?
        do_passthru = (reg_value & GIGE_TRIG_OUT_IS_TRIG_IN) == GIGE_TRIG_OUT_IS_TRIG_IN;
        ui->chk_trigout_pass_thru->setChecked(do_passthru);

        // Enable the invert control if pass through is enabled.
        ui->chk_trigout_invert_signal->setEnabled(do_passthru);
        ui->chk_trigout_invert_signal->setChecked((reg_value & GIGE_TRIG_OUT_POLARITY_MASK) == GIGE_TRIG_OUT_INVERT_TRIGIN);

        // If we're not passing the signal through, then enable the sourcing controls.
        ui->rb_trigout_active_low->setEnabled(!do_passthru);
        ui->rb_trigout_active_high->setEnabled(!do_passthru);
        ui->lbl_trigout_pulse_width->setEnabled(!do_passthru);
        ui->spin_trigout_pulse_width->setEnabled(!do_passthru);

        // Are we active high or low?
        active_high = (reg_value & GIGE_TRIG_OUT_POLARITY_MASK) == GIGE_TRIG_OUT_ACTIVE_HIGH;

        // Determine if we're using the pulse width or not.
        if ((reg_value & GIGE_TRIG_OUT_IS_TRIG_IN) == GIGE_TRIG_OUT_INTERNAL)
        {
            // Get the width out of the register.
            pulse_width = (reg_value & GIGE_TRIG_OUT_PULSE_MASK) >> GIGE_TRIG_OUT_PULSE_SHIFT;

            // Adjust the pulse width to be in milliseconds. The register holds it
            // in microseconds.
            pulse_width = pulse_width / 1000;
        }
        else
        {
            // No pulse width.
            pulse_width = NO_PULSE;
        }

        // Setup our radio buttons appropriately.
        ui->rb_trigout_active_high->blockSignals(true);
        ui->rb_trigout_active_low->blockSignals(true);
        ui->rb_trigout_active_high->setChecked(active_high);
        ui->rb_trigout_active_low->setChecked(!active_high);
        ui->rb_trigout_active_low->blockSignals(false);
        ui->rb_trigout_active_high->blockSignals(false);

        // Don't let our changes here cause signals to be propogated.
        ui->spin_trigout_pulse_width->blockSignals(true);

        // Are we using the pulse width?
        if (pulse_width)
        {
            // Enable the pulse width controls.
            ui->lbl_trigout_pulse_width->setEnabled(true);
            ui->spin_trigout_pulse_width->setEnabled(true);
            ui->spin_trigout_pulse_width->setValue((int)pulse_width);
        }
        else
        {
            // Disable the controls.
            ui->lbl_trigout_pulse_width->setEnabled(false);
            ui->spin_trigout_pulse_width->setEnabled(false);
            ui->spin_trigout_pulse_width->setValue(DEFAULT_PW_VALUE);
        }

        // Re-enable our spinner's signals.
        ui->spin_trigout_pulse_width->blockSignals(false);

        // The following does not apply to series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            uint32_t channel_bit;

            // Release whatever channel we had before.
            release_channel(trig_out_chan);

            // What channel is being used for the external trigger?
            channel_bit = (reg_value >> GIGE_TRIG_OUT_CHANNEL_SHIFT) & MAX_CHANNEL_MASK;
            for (trig_out_chan = 0; trig_out_chan < MAX_CHANNEL; ++trig_out_chan)
            {
                // Is this the one?
                if (channel_bit == (uint32_t)(1 << trig_out_chan))
                {
                    // Mark this channel as "in use".
                    channels_in_use |= channel_bit;
                    break;
                }
            }
            if (trig_out_chan == MAX_CHANNEL)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Couldn't find trig out channel, reg_value:" << Qt::showbase << Qt::hex << reg_value << "channel_bit:" << channel_bit;
                trig_out_chan = -1;
            }

            // Enable the channel radio buttons and check the appropriate one.
            ui->frm_to_chan->setEnabled(true);
            ui->rb_to_chan_0->setChecked(trig_out_chan == 0);
            ui->rb_to_chan_1->setChecked(trig_out_chan == 1);
            ui->rb_to_chan_2->setChecked(trig_out_chan == 2);

            // Update which buttons the user can now select.
            update_channel_buttons(ui->rb_to_chan_0, ui->rb_to_chan_1, ui->rb_to_chan_2, trig_out_chan);
        }

        // Display the tab icon as appropriate.
        set_tab_icon(TAB_TRIG_OUT, true, trig_out_chan);
    }
    else
    {
        // Doesn't apply to series3.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Not using any channels.
            ui->rb_to_chan_0->setChecked(false);
            ui->rb_to_chan_1->setChecked(false);
            ui->rb_to_chan_2->setChecked(false);
            ui->frm_to_chan->setEnabled(false);

            // Release the current channel.
            if (trig_out_chan >= 0)
            {
                release_channel(trig_out_chan);
                trig_out_chan = -1;
            }
        }

        // We are not enabled. Any change from where we were?
        if (ui->trig_out_enable->isChecked())
        {
            ui->trig_out_enable->setChecked(false);
        }

        // Make sure the frame reflects the state of the checkbox.
        ui->frame_trig_out->setEnabled(ui->trig_out_enable->isChecked());

        // Reset the tab icon.
        set_tab_icon(TAB_TRIG_OUT, false);
    }
}

/*!
 * @brief Update the trigger position based on registers read from the camera.
 */
void PanelRecControls::update_trigger()
{
    bool        use_frames;
    uint32_t    reg_value;

    // Are we in the middle of an update?
    if (update_in_progress)
        return;

    // Ignore this if focus is on the trigger control.
    if (ui->sld_trig_pos->hasFocus() || ui->spin_trig_pos->hasFocus())
        return;

    // Get the current trigger position.
    reg_value = read_camera_cache(GIGE_REG_TRIGGER_REQ_POS);

    // Figure out if the camera is using frames or percentage.
    use_frames = ((reg_value & GIGE_TRIGGER_REQ_FRAME_MASK) == (uint32_t)GIGE_TRIGGER_REQ_FRAME_MASK);
    if (use_frames)
    {
        int tick_interval;

        // Switch the title so the user knows that we're dealing with frames.
        ui->grp_trig_type->setTitle(tr("Trigger In Frames"));

        // Setup to use frames.
        ui->rb_use_frames->blockSignals(true);
        ui->rb_use_frames->setChecked(true);
        ui->rb_use_frames->blockSignals(false);
        ui->rb_use_percent->setChecked(false);

        // Make the tick interval be 1/10th the number of frames.
        tick_interval = (num_frames / 10);
        ui->sld_trig_pos->setTickInterval(tick_interval);
        ui->sld_trig_pos->setMinimum(0);
        ui->sld_trig_pos->setMaximum((int)(num_frames - 1));
        ui->spin_trig_pos->setMinimum(0);
        ui->spin_trig_pos->setMaximum((int)(num_frames - 1));
    }
    else
    {
        // Switch the title so the user knows that we're dealing with percentages.
        ui->grp_trig_type->setTitle(tr("Trigger In %"));

        // Setup to use percentage.
        ui->rb_use_percent->blockSignals(true);
        ui->rb_use_percent->setChecked(true);
        ui->rb_use_percent->blockSignals(false);
        ui->rb_use_frames->setChecked(false);

        ui->sld_trig_pos->setTickInterval(10);
        ui->sld_trig_pos->setMinimum(0);
        ui->sld_trig_pos->setMaximum(100);
        ui->spin_trig_pos->setMinimum(0);
        ui->spin_trig_pos->setMaximum(100);
    }

    // Update our controls.
    ui->spin_trig_pos->setValue(reg_value & GIGE_TRIGGER_REQ_POS_MASK);
}

//--------------------------------------------------------------------
// Event Handlers
//--------------------------------------------------------------------

/*!
 * @brief Process a change event - specifically looking for a language change.
 * @param event     Pointer to the event we're dealing with.
 */
void PanelRecControls::changeEvent(QEvent *event)
{
    // Was this our event?
    if (event->type() == QEvent::LanguageChange)
    {
        // Translate our user interface.
        translate_ui();
    }

    // Pass it on.
    FMPanelWidget::changeEvent(event);
}

void PanelRecControls::focusInEvent(QFocusEvent *event)
{
    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Update our panel.
    update();

    // Pass it on.
    FMPanelWidget::changeEvent(event);
}

/*!
 * @brief Event routine which is invoked when our timer expires.
 * @param event     Pointer to the event class.
 */
void PanelRecControls::timerEvent(QTimerEvent *event)
{
    CameraMDI   *camera_ptr;

    // Pass it on up.
    FMPanelWidget::timerEvent(event);

    // Is it the one we're interested in?
    if (event->timerId() != update_timer_id)
        return;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
        return;

    // Are we in the middle of an update?
    if (update_in_progress)
        return;

    // Update our registers.
    read_registers();
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user clicked on the "allow discard" button of the Arm-In tab.
 */
void PanelRecControls::arm_in_allow_discard(bool /*checked*/)
{
    uint32_t    reg_value;

    // Start the write sequence.
    update_in_progress = true;

    // Get the current settings.
    reg_value = read_camera_cache(GIGE_REG_PREFERENCES);

    // Are discards allowed?
    if (ui->chk_allow_discard->isChecked())
        reg_value |= GIGE_PREF_ARM_IN_OVERWRITE;
    else
        reg_value &= ~GIGE_PREF_ARM_IN_OVERWRITE;

    // Start the writing.
    write_camera_register(GIGE_REG_PREFERENCES, reg_value);
}

void PanelRecControls::arm_in_channel(bool checked)
{
    int channel;
    int reserve;
    uint32_t reg_value;

    Q_UNUSED(checked)

    // Which radio button was checked? Deselect the others.
    // Can't use get_arm_in_channel() because we need to setup
    // the radio buttons first.
    if (sender() == ui->rb_ai_chan_0)
    {
        channel = 0;
        ui->rb_ai_chan_1->setChecked(false);
        ui->rb_ai_chan_2->setChecked(false);
    }
    else if (sender() == ui->rb_ai_chan_1)
    {
        channel = 1;
        ui->rb_ai_chan_0->setChecked(false);
        ui->rb_ai_chan_2->setChecked(false);
    }
    else
    {
        channel = 2;
        ui->rb_ai_chan_0->setChecked(false);
        ui->rb_ai_chan_1->setChecked(false);
    }

    // Request the channel.
    reserve = reserve_channel(channel);
    if (reserve < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requested:" << channel << "was denied!?";
        return;
    }

    // Release the current channel.
    if (arm_in_chan >= 0)
        release_channel(arm_in_chan);
    arm_in_chan = reserve;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_ARM_IN_ENABLE);

    // Build up our register value.
    reg_value &= ~(MAX_CHANNEL_MASK << GIGE_ARM_IN_CHANNEL_SHIFT);
    reg_value |= (1 << (GIGE_ARM_IN_CHANNEL_SHIFT + arm_in_chan));

    // Start the write sequence.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_ARM_IN_ENABLE;
    update_reg_value = reg_value;

    // Write out the new register value.
    write_camera_register(GIGE_REG_ARM_IN_ENABLE, reg_value);
}

/*!
 * @brief The user clicked on the "edge" radio button of the Arm-In tab.
 */
void PanelRecControls::arm_in_edge_type_clicked(bool /*checked*/)
{
    // Update the camera.
    write_arm_in();
}

/*!
 * @brief The user clicked on the "Enable" button of the Arm-In tab.
 * @param checked   True if the button is checked, false otherwise.
 */
void PanelRecControls::arm_in_enable_clicked(bool checked)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "checked:" << checked << "arm_in_enable:" << ui->arm_in_enable->isChecked() << "arm_in_chan:" << arm_in_chan;

    // Are we enabled?
    if (checked)
    {
        // Does not apply for series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Do we have a channel?
            arm_in_chan = get_arm_in_channel();
            if (arm_in_chan < 0)
            {
                // Request a channel.
                arm_in_chan = reserve_channel();
                if (arm_in_chan < 0)
                {
                    // Can't get a channel, so can't enable.
                    ui->arm_in_enable->setChecked(false);

                    // Make sure any other controls are disabled.
                    ui->frame_arm_in->setEnabled(false);

                    // Let the user know what just happened.
                    QMessageBox::warning(this, tr("Enable Failed"), tr("There are no channels available. Please release a channel and try again."), QMessageBox::Ok, QMessageBox::Ok);
                    return;
                }
            }
        }
    }

    // Update the camera and UI.
    write_arm_in();
}

/*!
 * @brief The user clicked on the "falling/low" radio button of the Arm-In tab.
 */
void PanelRecControls::arm_in_falling_low_clicked(bool /*checked*/)
{
    // Update the camera.
    write_arm_in();
}

/*!
 * @brief The user clicked on the "level" radio button of the Arm-In tab.
 */
void PanelRecControls::arm_in_level_type_clicked(bool /*checked*/)
{
    // Update the camera.
    write_arm_in();
}

/*!
 * @brief The user clicked on the "rising/high" radio button of the Arm-In tab.
 */
void PanelRecControls::arm_in_rising_high_clicked(bool /*checked*/)
{
    // Update the camera.
    write_arm_in();
}

/*!
 * @brief The user clicked on the "active high" radio button of the Arm-Out tab.
 */
void PanelRecControls::arm_out_active_high_clicked(bool /*checked*/)
{
    // Update the camera.
    write_arm_out();
}

/*!
 * @brief The user clicked on the "active low" radio button of the Arm-Out tab.
 */
void PanelRecControls::arm_out_active_low_clicked(bool /*checked*/)
{
    // Update the camera.
    write_arm_out();
}

void PanelRecControls::arm_out_channel(bool checked)
{
    int channel;
    int reserve;
    uint32_t reg_value;

    Q_UNUSED(checked);

    // Which radio button was checked? Deselect the others.
    if (sender() == ui->rb_ao_chan_0)
    {
        channel = 0;
        ui->rb_ao_chan_1->setChecked(false);
        ui->rb_ao_chan_2->setChecked(false);
    }
    else if (sender() == ui->rb_ao_chan_1)
    {
        channel = 1;
        ui->rb_ao_chan_0->setChecked(false);
        ui->rb_ao_chan_2->setChecked(false);
    }
    else
    {
        channel = 2;
        ui->rb_ao_chan_0->setChecked(false);
        ui->rb_ao_chan_1->setChecked(false);
    }

    // Request the channel.
    reserve = reserve_channel(channel);
    if (reserve < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requested:" << channel << "was denied!?";
        return;
    }

    // Release the current channel.
    if (arm_out_chan >= 0)
        release_channel(arm_out_chan);
    arm_out_chan = reserve;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_ARM_OUT_SELECT);

    // Build up our register value.
    reg_value &= ~(MAX_CHANNEL_MASK << GIGE_ARM_OUT_CHANNEL_SHIFT);
    reg_value |= (1 << (GIGE_ARM_OUT_CHANNEL_SHIFT + arm_out_chan));

    // Start the write sequence.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_ARM_OUT_SELECT;
    update_reg_value = reg_value;

    // Write out the new register value.
    write_camera_register(GIGE_REG_ARM_OUT_SELECT, reg_value);
}

/*!
 * @brief The user clicked on the "Enable" button of the Arm-Out tab.
 * @param checked   True if the button is checked, false otherwise.
 */
void PanelRecControls::arm_out_enable_clicked(bool checked)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "checked:" << checked << "arm_out_enable:" << ui->arm_out_enable->isChecked() << "arm_out_chan:" << arm_out_chan;

    // Are we enabled?
    if (checked)
    {
        // Does not apply for series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Do we have a channel?
            arm_out_chan = get_arm_out_channel();
            if (arm_out_chan < 0)
            {
                // Request a channel.
                arm_out_chan = reserve_channel();
                if (arm_out_chan < 0)
                {
                    // Can't get a channel, so can't enable.
                    ui->arm_out_enable->setChecked(false);

                    // Make sure any other controls are disabled.
                    ui->frame_arm_out->setEnabled(false);

                    // Let the user know what just happened.
                    QMessageBox::warning(this, tr("Enable Failed"), tr("There are no channels available. Please release a channel and try again."), QMessageBox::Ok, QMessageBox::Ok);
                    return;
                }
            }
        }
    }

    // Update the camera and UI.
    write_arm_out();
}

/*!
 * @brief The user clicked on the "invert" checkbox of the Arm-Out tab.
 */
void PanelRecControls::arm_out_invert_signal_clicked(bool /*checked*/)
{
    // Update the camera.
    write_arm_out();
}

/*!
 * @brief The user clicked on the "pass-thru" checkbox of the Arm-Out tab.
 */
void PanelRecControls::arm_out_pass_thru_clicked(bool /*checked*/)
{
    bool    selected;

    // Initialize the other controls based on if we're selected or not.
    selected = ui->chk_ao_pass_thru->isChecked();

    ui->chk_ao_invert_signal->setEnabled(selected);
    if (!selected)
        ui->chk_ao_invert_signal->setChecked(false);

    ui->rb_ao_active_high->setEnabled(!selected);
    ui->rb_ao_active_low->setEnabled(!selected);
    ui->chk_ao_pulse_width->setEnabled(!selected);
    ui->spin_ao_pulse_width->setEnabled(!selected);

    // Update the camera.
    write_arm_out();
}

/*!
 * @brief The user changed the value of the "pulse width" spinner of the Arm-Out tab.
 */
void PanelRecControls::arm_out_pulse_width_changed(int value)
{
    Q_UNUSED(value);

    // Update the camera.
    write_arm_out();
}

/*!
 * @brief The user clicked on the "pulse width" checkbox of the Arm-Out tab.
 */
void PanelRecControls::arm_out_pulse_width_clicked(bool /*checked*/)
{
    // Enable the spinner based on if our check box is selected or not.
    ui->spin_ao_pulse_width->setEnabled(ui->chk_ao_pulse_width->isChecked());

    // Update the camera.
    write_arm_out();
}

/*!
 * @brief We've received an unsolicited message from the camera. Update any pertinent controls.
 *
 * @param ip_addr       IP Address of the camera sending the message.
 * @param reg_addr      Address of the register affected.
 * @param reg_value     The new value from the register.
 */
void PanelRecControls::handle_unsol_msg(uint32_t ip_addr, uint32_t reg_addr, uint32_t reg_value)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Pass it along like it was a read complete event.
    handle_read(reg_addr, reg_value);
}

/*!
 * @brief The user clicked on the "Enable IRIG" checkbox.
 */
void PanelRecControls::irig_clicked(bool /*checked*/)
{
    uint32_t    reg_value;

    // Pass the information to the camera.
    reg_value = ui->chk_irig->isChecked() ? GIGE_IRIG_ENABLE : GIGE_IRIG_DISABLE;
    write_camera_register(GIGE_REG_IRIG_CONTROL, reg_value);
}

/*!
 * @brief The user clicked on the "Use Year" checkbox.
 */
void PanelRecControls::irig_year_clicked(bool /*checked*/)
{
    uint32_t    reg_value;

    // Pass the information to the camera.
    reg_value = ui->chk_irig_year->isChecked() ? GIGE_IRIG_USE_YEAR : 0;
    reg_value |= GIGE_IRIG_ENABLE;
    write_camera_register(GIGE_REG_IRIG_CONTROL, reg_value);
}

/*!
 * @brief Slot invoked when one of the messaging operations suffers a timeout.
 *
 * @param ip_addr       The IP address of the camera which didn't respond.
 * @param reg_addr      The address of the register which was read.
 * @param read_op       True if this was a read operation, false for a write operation.
 */
void PanelRecControls::msg_timeout_event(uint32_t ip_addr, uint32_t reg_addr, bool read_op)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Is there a write in progress?
    if (update_in_progress && !read_op)
    {
        // Make sure this is the address we're concerned about.
        if (reg_addr != last_write.first)
            return;

        // Update our retry counter.
        ++retry_counter;

        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "retry #" << retry_counter << "Last write:" << last_write;

        // Should we retry the write?
        if (retry_counter < FM_PANEL::MAX_WRITE_RETRIES)
        {
            // Retry the write.
            write_camera_register(last_write.first, last_write.second, false);
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Maximum write attempts exceeded.";

            // Reset our write list and flags.
            write_list.clear();
            update_in_progress = false;
        }
    }
}

/*!
 * @brief Slot invoked when one of the read operations complete.
 *
 * @param ip_addr       The IP address of the camera.
 * @param reg_addr      The address of the register which was read.
 */
void PanelRecControls::read_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    uint32_t    reg_value;

    // Pass it up.
    FMPanelWidget::read_complete(ip_addr, reg_addr);

    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Retrieve the camera register value.
    reg_value = read_camera_cache(reg_addr);

    // If we have an update in progress and we're waiting for the read back,
    // then close the update process.
    if (update_in_progress && (update_reg_addr == reg_addr))
    {
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Got the update_reg_addr, value:" << Qt::showbase << Qt::hex << reg_value << "update_val:" << update_reg_value;

        // Verify that our register modification "was accepted" by the camera.
        if (update_reg_value == reg_value)
        {
            update_in_progress = false;
            update_reg_addr = 0;
            update_reg_value = 0;
            update_retry_count = 0;
        }
        else if (reg_addr == GIGE_REG_SYNC_OUT_SELECT)
        {
            // TICKET #451: If this is sync-out we have a special condition.  We made sure that the
            // register was zero if we were disabling sync-out.  But on Series3 cameras this register
            // may come back with other bits set based on the Series3 GUI.  We must be tolerant of
            // the other bits being different in that special-case.  As elegant as this code is right now
            // I had to insert this special case because I cannot predict Series3 settings.
            // Were we disabling things?
            if ((update_reg_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_DISABLE)
            {
                // We are tolerant of this setting if it is still disabled meaning the camera_server accepted it.
                if ((reg_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_DISABLE)
                {
                    update_in_progress = false;
                    update_reg_addr = 0;
                    update_reg_value = 0;
                    update_retry_count = 0;
                }
            }
        }

        // We need to try to resend the register setting again if we are still in a state of update_in_progress.
        if (update_in_progress)
        {
            // We were expecting to get back a specific value, see if we can try reading the register again.
            ++update_retry_count;
            if (update_retry_count > MAX_RETRY_COUNT)
            {
                // Too many retries. Just give up...
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Retry count exceeded:" << update_retry_count << "update_reg_addr:" << Qt::showbase << Qt::hex << update_reg_addr;
                update_in_progress = false;
                update_reg_addr = 0;
                update_reg_value = 0;
                update_retry_count = 0;
            }
            else
            {
                uint32_t retry_reg = update_reg_addr;

                // Try reading it again...it's not what we expected.
//                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Retry update_reg_addr:" << Qt::showbase << Qt::hex << update_reg_addr;
                QTimer::singleShot(RETRY_DELAY, this, [this, retry_reg]() { this->read_retry(retry_reg); });
            }
        }
    }

    // Handle the read event.
    handle_read(reg_addr, reg_value);
}

/*!
 * @brief Delayed read retry.
 * @param reg_addr  Register address we're supposed to retry.
 */
void PanelRecControls::read_retry(uint32_t reg_addr)
{
//    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Trying again, retry_count:" << update_retry_count << "reg_addr:" << Qt::showbase << Qt::hex << reg_addr;
    start_camera_read(reg_addr);
}

/*!
 * @brief The user clicked on the "Use Frames" radio button for specifying
 *        the trigger position.
 */
void PanelRecControls::rb_use_frame_clicked(bool checked)
{
    uint32_t    reg_value;

    // Only interested when this one is checked.
    if (!checked)
        return;

    // Update our controls.
    update_trigger();

    // Get the current trigger position.
    reg_value = read_camera_cache(GIGE_REG_TRIGGER_REQ_POS);
    if ((reg_value & GIGE_TRIGGER_REQ_FRAME_MASK) != (uint32_t)GIGE_TRIGGER_REQ_FRAME_MASK)
    {
        // Convert from a percentage.
        uint32_t trigger_frame_pos = (uint32_t)(((uint64_t)num_frames * (uint64_t)reg_value) / 100UL);

        // Make sure we set it for "frames" not "percent".
        trigger_frame_pos |= (uint32_t)GIGE_TRIGGER_REQ_FRAME_MASK;

        // Send it off to the camera.
        write_camera_register(GIGE_REG_TRIGGER_REQ_POS, trigger_frame_pos);
    }
}

/*!
 * @brief The user clicked on the "Use Percent" radio button for specifying
 *        the trigger position.
 */
void PanelRecControls::rb_use_percent_clicked(bool checked)
{
    uint32_t    reg_value;

    // Only interested when this one is checked.
    if (!checked)
        return;

    // Update our controls.
    update_trigger();

    // Get the current trigger position.
    reg_value = read_camera_cache(GIGE_REG_TRIGGER_REQ_POS);
    if ((reg_value & GIGE_TRIGGER_REQ_FRAME_MASK) == (uint32_t)GIGE_TRIGGER_REQ_FRAME_MASK)
    {
        // Convert to a percentage
        uint32_t trigger_percent = (uint32_t)(((uint64_t)(reg_value & GIGE_TRIGGER_REQ_POS_MASK) * 100UL) / (uint64_t)num_frames);

        // Get rid of any rounding issues.
        if ((((uint64_t)(reg_value & GIGE_TRIGGER_REQ_POS_MASK) * 100) % (uint64_t)num_frames) != 0)
            ++trigger_percent;

        // Make sure we aren't too big.
        if (trigger_percent > 100)
            trigger_percent = 100;

        // Send it off to the camera.
        write_camera_register(GIGE_REG_TRIGGER_REQ_POS, trigger_percent);
    }
}

/*!
 * @brief Queries camera for Video Trigger information.
 */
void PanelRecControls::request_vt_info()
{
    CameraMDI   *camera_ptr;
    CameraVideo *camera_video;

    // Do we have a camera connected?
    if ((camera_ptr = p_control_panel->getCamera()) == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Has the video trigger reticle information been loaded?
    camera_video = camera_ptr->get_player();
    if (camera_video == nullptr)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No camera video player!";
        return;
    }

    // Check to see if the reticle has been loaded yet.
    if (camera_video->get_vt_rect().isEmpty())
    {
        // Come back later.
        QTimer::singleShot(VT_INFO_DELAY, this, &PanelRecControls::request_vt_info);
        return;
    }

    // Send a read request to retrieve the HW video trigger control information.
    if (!camera_ptr->is_series3())
    {
        // Send a read request to retrieve the HW video trigger control information.
        camera_ptr->start_camera_read_buf(GIGE_REG_VTRIG_SETTINGS_JSON, GIGE_VTRIG_SETTINGS_JSON_LEN);
    }
}

/*!
 * @brief The user is dragging the slider for specifying a new trigger position.
 */
void PanelRecControls::sld_trigger_pos_changed(int new_value)
{
    uint64_t    frames_before;
    uint64_t    frames_after;
    uint64_t    frame_rate;
    double  time_before, time_after;
    QString s_field;

    // Display the title with the new value in it.
    if (ui->rb_use_frames->isChecked())
    {
        frames_before = (uint64_t)new_value;
        frames_after = ((uint64_t)num_frames - 1) - (uint64_t)new_value;
    }
    else
    {
        frames_before = ((uint64_t)(num_frames - 1) * (uint64_t)new_value) / 100;
        frames_after = (uint64_t)(num_frames - 1) - frames_before;
    }

    frame_rate = read_camera_cache(GIGE_REG_FRAME_RATE);
    time_before = double(frames_before) / ClientMath::reg_to_frame_rate(frame_rate);
    time_after = double(frames_after) / ClientMath::reg_to_frame_rate(frame_rate);

    // Now display the values. Note that for time we are using a generic function
    // which will calculate the time in the correct units.
    s_field = QString::number(frames_before);
    ui->eb_frames_before->setText(s_field);
    s_field = QString::number(frames_after);
    ui->eb_frames_after->setText(s_field);
    display_time(time_before, ui->lbl_time_before, ui->eb_time_before);
    display_time(time_after, ui->lbl_time_after, ui->eb_time_after);

    // Update our spinner.
    ui->spin_trig_pos->blockSignals(true);
    ui->spin_trig_pos->setValue(new_value);
    ui->spin_trig_pos->blockSignals(false);
}

/*!
 * @brief The user has changed the trigger position using the spinner.
 * @param new_value     The new trigger position.
 */
void PanelRecControls::spin_trigger_pos_changed(int new_value)
{
    // User change the spin position by either moving it up/down or
    // typing in a new value. Update the slider
    ui->sld_trig_pos->setValue(new_value);
}

void PanelRecControls::sync_in_channel(bool checked)
{
    int channel;
    int reserve;
    uint32_t reg_value;

    Q_UNUSED(checked);

    // Which radio button was checked? Deselect the others.
    if (sender() == ui->rb_si_chan_0)
    {
        channel = 0;
        ui->rb_si_chan_1->setChecked(false);
        ui->rb_si_chan_2->setChecked(false);
    }
    else if (sender() == ui->rb_si_chan_1)
    {
        channel = 1;
        ui->rb_si_chan_0->setChecked(false);
        ui->rb_si_chan_2->setChecked(false);
    }
    else
    {
        channel = 2;
        ui->rb_si_chan_0->setChecked(false);
        ui->rb_si_chan_1->setChecked(false);
    }

    // Request the channel.
    reserve = reserve_channel(channel);
    if (reserve < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requested:" << channel << "was denied!?";
        return;
    }

    // Release the current channel.
    if (sync_in_chan >= 0)
        release_channel(sync_in_chan);
    sync_in_chan = reserve;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_SYNC_IN_ENABLE);

    // Build up our register value.
    reg_value &= ~(MAX_CHANNEL_MASK << GIGE_SYNC_IN_CHANNEL_SHIFT);
    reg_value |= (1 << (GIGE_SYNC_IN_CHANNEL_SHIFT + sync_in_chan));

    // Start the write sequence.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_SYNC_IN_ENABLE;
    update_reg_value = reg_value;

    // Write out the new register value.
    write_camera_register(GIGE_REG_SYNC_IN_ENABLE, reg_value);
}

/*!
 * @brief The user has modified the Sync-In delay value using the spinner.
 * @param value     [UNUSED] The new Sync-In delay.
 */
void PanelRecControls::sync_in_delay_changed(int value)
{
    Q_UNUSED(value);

    // Per ticket #227, disable the following.
#ifdef TICKET_227
    // Update the camera.
    write_sync_in();
#endif /* TICKET_227 */
}

/*!
 * @brief The user clicked on the "Enable" button of the Sync-In tab.
 * @param checked   True if the button is checked, false otherwise.
 */
void PanelRecControls::sync_in_enable_clicked(bool checked)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "checked:" << checked << "sync_in_enable:" << ui->sync_in_enable->isChecked() << "sync_in_chan:" << sync_in_chan;

    // Initialize our register based on if we're enabled or not.
    if (checked)
    {
        // Does not apply for series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Do we have a channel?
            sync_in_chan = get_sync_in_channel();
            if (sync_in_chan < 0)
            {
                // Request a channel.
                sync_in_chan = reserve_channel();
                if (sync_in_chan < 0)
                {
                    // Can't get a channel, so can't enable.
                    ui->sync_in_enable->setChecked(false);

                    // Make sure any other controls are disabled.
                    ui->frame_sync_in->setEnabled(false);

                    // Let the user know what just happened.
                    QMessageBox::warning(this, tr("Enable Failed"), tr("There are no channels available. Please release a channel and try again."), QMessageBox::Ok, QMessageBox::Ok);
                    return;
                }
            }
        }
    }
    else
    {
        // Disabling, then just hide our warning it doesn't apply.
        ui->lbl_si_mfr_too_high->setVisible(false);

        // Make sure the calculated frame rate is restored to normal.
        style_label_normal(ui->lbl_si_calc_frame_rate);
    }

    // Update the camera and UI.
    write_sync_in();
}

/*!
 * @brief The user has clicked on the "falling" radio button.
 */
void PanelRecControls::sync_in_falling_clicked(bool /*checked*/)
{
    // Update the camera.
    write_sync_in();
}

/*!
 * @brief The user has modified the master frame rate.
 * @param value     The new master frame rate value.
 */
void PanelRecControls::sync_in_mfr_changed(int value)
{
    // If the user entered a zero, this is a special case and the user wants to
    // to check with the FPGA to get timings.
    if (value == 0)
    {
        // Update the control with the value from the camera.
        ui->spin_si_mfr->setValue((int)get_master_framerate());
    }
    else
    {
        // Update the camera.
        write_sync_in();
    }
}

/*!
 * @brief The user has clicked on the "per-frame" radio button.
 */
void PanelRecControls::sync_in_per_frame_clicked(bool /*checked*/)
{
    // Update the dependent controls.
    if (ui->rb_si_per_frame->isChecked())
    {
        ui->spin_si_mfr->setEnabled(true);
        ui->spin_si_rate_div->setEnabled(true);

        // Per ticket #227, disable the following.
#ifdef TICKET_227
        ui->spin_si_delay->setEnabled(false);
#endif /* TICKET_227 */
    }

    // Update the camera.
    write_sync_in();
}

/*!
 * @brief The user has clicked on the "per-second" radio button.
 */
void PanelRecControls::sync_in_per_sec_clicked(bool /*checked*/)
{
    // Per ticket #227, don't allow this processing.
#ifdef TICKET_227
    // Update the dependent controls.
    if (ui->rb_si_per_frame->isChecked())
    {
        ui->spin_si_mfr->setEnabled(false);
        ui->spin_si_rate_div->setEnabled(false);
        ui->spin_si_delay->setEnabled(true);
    }

    // Update the camera.
    write_sync_in();
#endif /* TICKET_227 */
}

/*!
 * @brief The user has modified the Sync-In rate divisor.
 * @param value     [UNUSED] The new rate divisor value.
 */
void PanelRecControls::sync_in_rate_div_changed(int value)
{
    Q_UNUSED(value);

    // Update the camera.
    write_sync_in();
}

/*!
 * @brief The user has clicked on the "rising" radio button.
 */
void PanelRecControls::sync_in_rising_clicked(bool /*checked*/)
{
    // Update the camera.
    write_sync_in();
}

/*!
 * @brief The user has clicked on the "active low" radio button.
 */
void PanelRecControls::sync_out_active_low_clicked(bool /*checked*/)
{
    // Update the camera.
    write_sync_out();
}

/*!
 * @brief The user has clicked on the "active high" radio button.
 */
void PanelRecControls::sync_out_active_high_clicked(bool /*checked*/)
{
    // Update the camera.
    write_sync_out();
}

void PanelRecControls::sync_out_channel(bool checked)
{
    int channel;
    int reserve;
    uint32_t reg_value;

    Q_UNUSED(checked);

    // Which radio button was checked? Deselect the others.
    if (sender() == ui->rb_so_chan_0)
    {
        channel = 0;
        ui->rb_so_chan_1->setChecked(false);
        ui->rb_so_chan_2->setChecked(false);
    }
    else if (sender() == ui->rb_so_chan_1)
    {
        channel = 1;
        ui->rb_so_chan_0->setChecked(false);
        ui->rb_so_chan_2->setChecked(false);
    }
    else
    {
        channel = 2;
        ui->rb_so_chan_0->setChecked(false);
        ui->rb_so_chan_1->setChecked(false);
    }

    // Request the channel.
    reserve = reserve_channel(channel);
    if (reserve < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requested:" << channel << "was denied!?";
        return;
    }

    // Release the current channel.
    if (sync_out_chan >= 0)
        release_channel(sync_out_chan);
    sync_out_chan = reserve;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_SYNC_OUT_SELECT);

    // Build up our register value.
    reg_value &= ~(MAX_CHANNEL_MASK << GIGE_SYNC_OUT_CHANNEL_SHIFT);
    reg_value |= (1 << (GIGE_SYNC_OUT_CHANNEL_SHIFT + sync_out_chan));

    // Don't miss the channel change in QSettings if sync-out is enabled.
    // It is probably impossible to get here if it isn't enabled but to
    // be explicitly clear that it must be enabled before we persist it.
    // NOTE:  This is a different pathway to update the register
    // GIGE_REG_SYNC_OUT_SELECT than the normal method
    // PanelRecControls::write_sync_out() which is utilized by most
    // other controls to update the register.
    if ((reg_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_ENABLE)
    {
        sync_out_save_settings(reg_value);
    }

    // Start the write sequence.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_SYNC_OUT_SELECT;
    update_reg_value = reg_value;

#if LOG_SYNC_OUT_CHANNEL
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "reg_value:" << Qt::showbase << Qt::hex << reg_value;
#endif

    // Write out the new register value.
    write_camera_register(GIGE_REG_SYNC_OUT_SELECT, reg_value);
}

/*!
 * @brief The user has modified the Sync-Out duty cycle.
 * @param value     [UNUSED] The new duty cycle.
 */
void PanelRecControls::sync_out_duty_cycle_changed(int value)
{
    Q_UNUSED(value);

    // Update the camera.
    write_sync_out();
}

/*!
 * @brief The user has clicked on the "duty cycle" radio button.
 */
void PanelRecControls::sync_out_duty_cycle_clicked(bool /*checked*/)
{
    // Enable the spinner if we're selected.
    if (ui->rb_so_duty_cycle->isChecked())
        ui->spin_so_duty_cycle->setEnabled(true);

    // Update the camera.
    write_sync_out();
}

/*!
 * @brief Sync out enable checkbox has changed.
 */
void PanelRecControls::sync_out_enable_clicked(bool checked)
{
    if (LOG_SYNC_OUT_TRACES)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "checked" << checked << "sync_out_enable:" << ui->sync_out_enable->isChecked() << "sync_out_chan:" << sync_out_chan;
    }

    // Did we just transition to enabled-state?
    if (checked)
    {
        uint32_t sync_out_reg;

        // Does not apply for series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Do we have a channel?
            sync_out_chan = get_sync_out_channel();
            if (sync_out_chan < 0)
            {
                // Request a channel.
                sync_out_chan = reserve_channel();
                if (sync_out_chan < 0)
                {
                    // Can't get a channel, so can't enable.
                    ui->sync_out_enable->setChecked(false);

                    // Make sure any other controls are disabled.
                    ui->frame_sync_out->setEnabled(false);

                    // Let the user know what just happened.
                    QMessageBox::warning(this, tr("Enable Failed"), tr("There are no channels available. Please release a channel and try again."), QMessageBox::Ok, QMessageBox::Ok);
                    return;
                }
            }
        }

        // The sync-out logic is a little different, because folks wanted to make the settings persistent, rather than simply
        // startup in a default state. To facilitate that, we need to check to see if there is a saved setting. If there is a
        // saved setting for this camera, use the saved data to restore the controls, etc.
        if (sync_out_load_settings(sync_out_reg))
        {
            if (LOG_SYNC_OUT_TRACES)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << Qt::hex << "Sync-out was just enabled and sync_out_reg QSettings:" << sync_out_reg;
            }

            // We actually have stored settings, so update the UI to reflect this.
            sync_out_gige_reg_to_ui(sync_out_reg);
        }
    }

    // Forward on to the write function which updates the camera.
    write_sync_out();
}

/*!
 * @brief The user has clicked on the "invert signals" radio button.
 */
void PanelRecControls::sync_out_invert_signals_clicked(bool /*checked*/)
{
    // Update the camera.
    write_sync_out();
}

/*!
 * @brief The user has clicked on the "pass thru" radio button.
 */
void PanelRecControls::sync_out_pass_thru_clicked(bool /*checked*/)
{
    if (ui->rb_so_pass_thru->isChecked())
    {
        ui->chk_so_invert_signals->setEnabled(true);

        ui->rb_so_active_high->setEnabled(false);
        ui->rb_so_active_low->setEnabled(false);
        ui->rb_so_duty_cycle->setEnabled(false);
        ui->spin_so_duty_cycle->setEnabled(false);
        ui->rb_so_shutter->setEnabled(false);
    }

    // Update the camera.
    write_sync_out();
}

/*!
 * @brief The user has clicked on the "per-frame" radio button.
 */
void PanelRecControls::sync_out_per_frame_clicked(bool /*checked*/)
{
    if (ui->rb_so_per_frame->isChecked())
    {
        ui->chk_so_invert_signals->setEnabled(false);

        ui->rb_so_active_high->setEnabled(true);
        ui->rb_so_active_low->setEnabled(true);
        ui->rb_so_duty_cycle->setEnabled(true);
        if (ui->rb_so_duty_cycle->isChecked())
            ui->spin_so_duty_cycle->setEnabled(true);
        else
            ui->spin_so_duty_cycle->setEnabled(false);
        ui->rb_so_shutter->setEnabled(true);
    }

    // Update the camera.
    write_sync_out();
}

/*!
 * @brief The user has clicked on the "per-second" radio button.
 */
void PanelRecControls::sync_out_per_sec_clicked(bool /*checked*/)
{
    if (ui->rb_so_per_sec->isChecked())
    {
        ui->chk_so_invert_signals->setEnabled(false);

        ui->rb_so_active_high->setEnabled(true);
        ui->rb_so_active_low->setEnabled(true);
        ui->rb_so_duty_cycle->setEnabled(true);
        ui->rb_so_duty_cycle->setChecked(true);
        ui->spin_so_duty_cycle->setEnabled(true);
        ui->rb_so_shutter->setEnabled(false);
    }

    // Update the camera.
    write_sync_out();
}

/*!
 * @brief The user has clicked on the "shutter" radio button.
 */
void PanelRecControls::sync_out_shutter_clicked(bool /*checked*/)
{
    // Disable the duty cycle spinner, if we're selected.
    if (ui->rb_so_shutter->isChecked())
        ui->spin_so_duty_cycle->setEnabled(false);

    // Update the camera.
    write_sync_out();
}

/*!
 * @brief The user clicked on one of the "enable" buttons on a tab.
 *
 * @param frame_widget  A pointer to the frame widget of the affect tab.
 */
void PanelRecControls::tab_enable_clicked(QWidget *frame_widget)
{
    // Toggle the "enabled-ness" of the tab.
    frame_widget->setEnabled(!frame_widget->isEnabled());
}

/*!
 * @brief The user has performed an action on the trigger position slider.
 * @param action        The action the user performed.
 */
void PanelRecControls::trigger_pos_action(int action)
{
    // We're only interested in a few of the actions. We can
    // safely ignore the others.
    switch (action)
    {
    case QAbstractSlider::SliderSingleStepAdd:
    case QAbstractSlider::SliderSingleStepSub:
    case QAbstractSlider::SliderPageStepAdd:
    case QAbstractSlider::SliderPageStepSub:
        // The change event doesn't propagate, so let's make sure things happen.
        ui->spin_trig_pos->setValue(ui->sld_trig_pos->value());

        // Update our display.
        emit ui->sld_trig_pos->sliderReleased();
        break;
    case QAbstractSlider::SliderMove:
        break;
    default:
        break;
    }
}

/*!
 * @brief The user has finished modifying the trigger position, so write
 *        the new value to the camera.
 */
void PanelRecControls::trigger_pos_update()
{
    uint32_t    reg_value;

    // Build up our register setting.
    reg_value = (uint32_t)ui->spin_trig_pos->value();
    if (ui->rb_use_frames->isChecked())
        reg_value |= GIGE_TRIGGER_REQ_FRAME_MASK;

    // Send it on to the camera.
    update_in_progress = true;
    write_camera_register(GIGE_REG_TRIGGER_REQ_POS, reg_value);
}

void PanelRecControls::trig_in_channel(bool checked)
{
    int channel;
    int reserve;
    uint32_t reg_value;

    Q_UNUSED(checked);

    // Ignore if we have an update in progress.
    if (update_in_progress)
        return;

    // Which radio button was checked? Clear the others.
    if (sender() == ui->rb_ti_chan_0)
    {
        channel = 0;
        ui->rb_ti_chan_1->setChecked(false);
        ui->rb_ti_chan_2->setChecked(false);
    }
    else if (sender() == ui->rb_ti_chan_1)
    {
        channel = 1;
        ui->rb_ti_chan_0->setChecked(false);
        ui->rb_ti_chan_2->setChecked(false);
    }
    else
    {
        channel = 2;
        ui->rb_ti_chan_0->setChecked(false);
        ui->rb_ti_chan_1->setChecked(false);
    }

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "channel:" << channel << "trig_in_chan:" << trig_in_chan;

    // Request the channel.
    reserve = reserve_channel(channel);
    if (reserve < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requested:" << channel << "was denied!?";
        return;
    }

    // Release the current channel.
    release_channel(trig_in_chan);
    trig_in_chan = reserve;

    // Build up our register value.
    reg_value = GIGE_EXT_TRIG_ENABLE_MASK | (1 << (GIGE_EXT_TRIG_CHANNEL_SHIFT + trig_in_chan));

    // Remember an update is pending.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_EXT_TRIG_ENABLE;
    update_reg_value = reg_value;

    // Write out the new register value.
    write_camera_register(GIGE_REG_EXT_TRIG_ENABLE, reg_value);
}

void PanelRecControls::trig_in_enable_clicked(bool checked)
{
    bool enabled;
    uint32_t reg_value(0);

    // Setup our current flag values.
    enabled = ((read_camera_cache(GIGE_REG_EXT_TRIG_ENABLE) & GIGE_EXT_TRIG_ENABLE_MASK) == GIGE_EXT_TRIG_ENABLE_MASK);

    // Do we need to change anything?
    if (checked == enabled)
        return;

    // Are we trying to enable?
    if (checked && !p_control_panel->getCamera()->is_series3())
    {
        // Do we have a channel?
        if (trig_in_chan < 0)
        {
            int channel = get_trig_in_channel();

            // Request a channel.
            trig_in_chan = reserve_channel(channel);
            if (trig_in_chan < 0)
            {
                // There are no channels available. Do not allow the user to enable.
                // These two variables are never used as we warn and return in this pathway.
                //checked = false;
                //reg_value =  0;

                // Let the user know what just happened.
                QMessageBox::warning(this, tr("Enable Failed"), tr("There are no channels available. Please release a channel and try again."), QMessageBox::Ok, QMessageBox::Ok);
                return;
            }
        }

        // We are enabling the external trigger and specifying a channel.
        reg_value = GIGE_EXT_TRIG_ENABLE_MASK | (1 << (GIGE_EXT_TRIG_CHANNEL_SHIFT + trig_in_chan));
    }
    else
    {
        // Here for a series3 camera, or we are disabling the external trigger.
        reg_value = checked ? GIGE_EXT_TRIG_ENABLE_MASK : 0;
    }

    // Remember an update is pending.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_EXT_TRIG_ENABLE;
    update_reg_value = reg_value;

    // Write out the new register value.
    write_camera_register(GIGE_REG_EXT_TRIG_ENABLE, reg_value);

    // If we're not enabled, clear the channel selected.
    if (!checked && (trig_in_chan >= 0))
    {
        release_channel(trig_in_chan);
        trig_in_chan = -1;
    }

    // Update our UI.
    ui->frame_trig_in->setEnabled(checked);
    ui->frm_ti_chan->setEnabled(checked);
}

void PanelRecControls::trig_in_falling_clicked(bool /*checked*/)
{
    bool        rising;

    // Setup our current flag values.
    rising = ((read_camera_cache(GIGE_REG_EXT_TRIG_POLARITY) & GIGE_EXT_TRIG_POLARITY_RISING) == GIGE_EXT_TRIG_POLARITY_RISING);

    // Any change?
    if (rising)
    {
        // Remember an update is pending.
        update_in_progress = true;

        // Write out the new register value.
        write_camera_register(GIGE_REG_EXT_TRIG_POLARITY, GIGE_EXT_TRIG_POLARITY_FALLING);
    }
}

void PanelRecControls::trig_in_rising_clicked(bool /*checked*/)
{
    bool        rising;

    // Setup our current flag values.
    rising = ((read_camera_cache(GIGE_REG_EXT_TRIG_POLARITY) & GIGE_EXT_TRIG_POLARITY_RISING) == GIGE_EXT_TRIG_POLARITY_RISING);

    // Any change?
    if (!rising)
    {
        // Remember an update is pending.
        update_in_progress = true;

        // Write out the new register value.
        write_camera_register(GIGE_REG_EXT_TRIG_POLARITY, GIGE_EXT_TRIG_POLARITY_RISING);
    }
}

/*!
 * @brief The user clicked the "active low" radio button.
 */
void PanelRecControls::trig_out_active_low_clicked(bool /*checked*/)
{
    // Only update the camera if we got "checked".
    if (ui->rb_trigout_active_low->isChecked())
        write_trig_out();
}

/*!
 * @brief The user clicked the "active high" radio button.
 */
void PanelRecControls::trig_out_active_high_clicked(bool /*checked*/)
{
    // Only update the camera if we got "checked".
    if (ui->rb_trigout_active_high->isChecked())
        write_trig_out();
}

void PanelRecControls::trig_out_channel(bool checked)
{
    int channel;
    int reserve;
    uint32_t reg_value;

    Q_UNUSED(checked);

    // Which radio button was checked? Deselect the others.
    if (sender() == ui->rb_to_chan_0)
    {
        channel = 0;
        ui->rb_to_chan_1->setChecked(false);
        ui->rb_to_chan_2->setChecked(false);
    }
    else if (sender() == ui->rb_to_chan_1)
    {
        channel = 1;
        ui->rb_to_chan_0->setChecked(false);
        ui->rb_to_chan_2->setChecked(false);
    }
    else
    {
        channel = 2;
        ui->rb_to_chan_0->setChecked(false);
        ui->rb_to_chan_1->setChecked(false);
    }

    // Request the channel.
    reserve = reserve_channel(channel);
    if (reserve < 0)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Requested:" << channel << "was denied!?";
        return;
    }

    // Release the current channel.
    if (trig_out_chan >= 0)
        release_channel(trig_out_chan);
    trig_out_chan = reserve;

    // Get the register settings.
    reg_value = read_camera_cache(GIGE_REG_TRIG_OUT_SELECT);

    // Build up our register value.
    reg_value &= ~(MAX_CHANNEL_MASK << GIGE_TRIG_OUT_CHANNEL_SHIFT);
    reg_value |= (1 << (GIGE_TRIG_OUT_CHANNEL_SHIFT + trig_out_chan));

    // Start the write sequence.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_TRIG_OUT_SELECT;
    update_reg_value = reg_value;

    // Write out the new register value.
    write_camera_register(GIGE_REG_TRIG_OUT_SELECT, reg_value);
}

/*!
 * @brief The user clicked on the "Enable" button of the Arm-In tab.
 * @param checked   True if the button is checked, false otherwise.
 */
void PanelRecControls::trig_out_enable_clicked(bool checked)
{
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "checked:" << checked << "trig_out_enable:" << ui->trig_out_enable->isChecked() << "trig_out_chan:" << trig_out_chan;

    // First find out if trigger out is enabled or disabled, if disabled there's not much more
    // to check.
    if (checked)
    {
        // Does not apply for series3 cameras.
        if (!p_control_panel->getCamera()->is_series3())
        {
            // Do we have a channel?
            trig_out_chan = get_trig_out_channel();
            if (trig_out_chan < 0)
            {
                // Request a channel.
                trig_out_chan = reserve_channel();
                if (trig_out_chan < 0)
                {
                    // Can't get a channel, so can't enable.
                    ui->trig_in_enable->setChecked(false);

                    // Make sure any other controls are disabled.
                    ui->frame_trig_out->setEnabled(false);

                    // Let the user know what just happened.
                    QMessageBox::warning(this, tr("Enable Failed"), tr("There are no channels available. Please release a channel and try again."), QMessageBox::Ok, QMessageBox::Ok);
                    return;
                }
            }
        }
    }

    // Update the camera and UI.
    write_trig_out();
}

/*!
 * @brief The user clicked the "invert" radio button.
 */
void PanelRecControls::trig_out_invert_signal_clicked(bool /*checked*/)
{
    // Update the camera.
    write_trig_out();
}

/*!
 * @brief The user clicked the "pass thru" radio button.
 */
void PanelRecControls::trig_out_pass_thru_clicked(bool /*checked*/)
{
    bool    is_checked;

    // Are we checked?
    is_checked = ui->chk_trigout_pass_thru->isChecked();

    // Setup the controls appropriately.
    ui->chk_trigout_invert_signal->setEnabled(is_checked);
    ui->chk_trigout_invert_signal->setChecked(false);
    ui->rb_trigout_active_low->setEnabled(!is_checked);
    ui->rb_trigout_active_high->setEnabled(!is_checked);
    ui->lbl_trigout_pulse_width->setEnabled(!is_checked);
    ui->spin_trigout_pulse_width->setEnabled(!is_checked);

    // Update the camera.
    write_trig_out();
}

/*!
 * @brief The user has modified the pulse width value.
 * @param value     [UNUSED] The new pulse width.
 */
void PanelRecControls::trig_out_pulse_width_changed(int value)
{
    Q_UNUSED(value);

    // Update the camera.
    write_trig_out();
}

/*!
 * @brief Toggle the visiblity of the panel. Optionally, pass
 *        the event up to the parent.
 *
 * @param notify_parent  True if we're supposed to pass this up to the parent.
 */
void PanelRecControls::toggle_visibility(bool notify_parent)
{
    // Pass it on up.
    FMPanelWidget::toggle_visibility(notify_parent);
}

void PanelRecControls::tt_enable(bool /*checked*/)
{
    int user_time;
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Enable/disable the controls & timer.
    ui->frm_tt_time->setEnabled(ui->chk_tt_enable->isChecked());
    if (ui->chk_tt_enable->isChecked())
    {
        // Calculate time in seconds.
        user_time = (ui->spin_tt_hours->value() * 3600) + (ui->spin_tt_minutes->value() * 60) + ui->spin_tt_seconds->value();

        // Update our UI.
        ui->tt_start->setEnabled(true);
        ui->frm_tt_entry->setEnabled(true);
        ui->tt_stop->setEnabled(false);
        ui->tt_restart->setEnabled(false);
        ui->tt_next->setEnabled(true);
        ui->tt_count->setEnabled(true);

        // Connect our signals.
        connect(ui->tt_restart, &QPushButton::clicked, this, &PanelRecControls::tt_restart);
        connect(ui->tt_start, &QPushButton::clicked, this, &PanelRecControls::tt_start);
        connect(ui->tt_stop, &QPushButton::clicked, this, &PanelRecControls::tt_stop);
        connect(camera_ptr, &CameraMDI::tt_expired, this, &PanelRecControls::tt_incr);
    }
    else
    {
        // Turn off the timer.
        user_time = 0;

        // Pass it on. A value of zero will stop the timer.
        camera_ptr->time_trigger(user_time);

        // Disonnect our signals.
        disconnect(ui->tt_restart, &QPushButton::clicked, this, &PanelRecControls::tt_restart);
        disconnect(ui->tt_start, &QPushButton::clicked, this, &PanelRecControls::tt_start);
        disconnect(ui->tt_stop, &QPushButton::clicked, this, &PanelRecControls::tt_stop);
        disconnect(camera_ptr, &CameraMDI::tt_expired, this, &PanelRecControls::tt_incr);
    }

    // Display the remaining time.
    ui->tt_next->setTime(user_time);
    ui->tt_count->display(0);
}

void PanelRecControls::tt_incr()
{
    // Increment the counter.
    ui->tt_count->display(ui->tt_count->intValue() + 1);
}

void PanelRecControls::tt_hours_changed(int value)
{
    int user_time;
    CameraMDI   *camera_ptr;

    Q_UNUSED(value);

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Calculate time in seconds.
    user_time = (ui->spin_tt_hours->value() * 3600) + (ui->spin_tt_minutes->value() * 60) + ui->spin_tt_seconds->value();

    // Display the remaining time.
    ui->tt_next->setTime(user_time);
}

void PanelRecControls::tt_minutes_changed(int value)
{
    int user_time;
    CameraMDI   *camera_ptr;

    Q_UNUSED(value);

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Calculate time in seconds.
    user_time = (ui->spin_tt_hours->value() * 3600) + (ui->spin_tt_minutes->value() * 60) + ui->spin_tt_seconds->value();

    // Display the remaining time.
    ui->tt_next->setTime(user_time);
}

void PanelRecControls::tt_seconds_changed(int value)
{
    int user_time;
    CameraMDI   *camera_ptr;

    Q_UNUSED(value);

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Calculate time in seconds.
    user_time = (ui->spin_tt_hours->value() * 3600) + (ui->spin_tt_minutes->value() * 60) + ui->spin_tt_seconds->value();

    // Display the remaining time.
    ui->tt_next->setTime(user_time);
}

void PanelRecControls::tt_restart()
{
    // Restarting the timer means we stop the timer, then start it.
    tt_stop();
    tt_start();
}

void PanelRecControls::tt_start()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Tell the camera to start.
    camera_ptr->time_trigger(ui->tt_next->startTime());

    // Tell the countdown display to start.
    ui->tt_next->start();
    ui->tt_count->display(0);

    // Update our UI.
    ui->tt_start->setEnabled(false);
    ui->frm_tt_entry->setEnabled(false);
    ui->tt_stop->setEnabled(true);
    ui->tt_restart->setEnabled(true);
}

void PanelRecControls::tt_stop()
{
    int user_time;
    CameraMDI   *camera_ptr;

    // Tell the countdown display to stop.
    ui->tt_next->stop();

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Tell the camera to stop.
    camera_ptr->time_trigger(0);

    // Calculate time in seconds.
    user_time = (ui->spin_tt_hours->value() * 3600) + (ui->spin_tt_minutes->value() * 60) + ui->spin_tt_seconds->value();

    // Display the remaining time.
    ui->tt_next->setTime(user_time);

    // Update our UI.
    ui->tt_start->setEnabled(true);
    ui->frm_tt_entry->setEnabled(true);
    ui->tt_stop->setEnabled(false);
    ui->tt_restart->setEnabled(false);
}

void PanelRecControls::vt_area_changed(int new_area)
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Pass along the new information.
    camera_ptr->set_vt_area(new_area);
}

void PanelRecControls::vt_change_changed(int new_change)
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Pass along the new information.
    camera_ptr->set_vt_change(new_change);
}

void PanelRecControls::vt_refresh_changed(int new_refresh)
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Pass along the new information.
    camera_ptr->set_vt_refresh(new_refresh);
}

void PanelRecControls::vt_enable(bool /*checked*/)
{
    CameraMDI   *camera_ptr;
    PanelRecSettings *settings_ptr;

//    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
        return;

    // Enable/Disable the exposure tracking controls based on this selection.
    settings_ptr = static_cast<PanelRecSettings *>(p_control_panel->getPanel(CONTROL_PANEL::PANEL_REC_SETTINGS));

    // Are we enabling video trigger?
    if (ui->chk_vt_enable->isChecked())
    {
        // If the autoexposure is enabled, we need to disable it.
        if (settings_ptr->ae_enabled())
            settings_ptr->enable_ae_controls(false);
    }
    else
    {
        // If the autoexposure is disabled, make sure we enable it.
        if (!settings_ptr->ae_enabled())
            settings_ptr->enable_ae_controls(true);
    }

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[***** vt_reticle_visible():" << camera_ptr->vt_reticle_visible()
                     << "chk_vt_enable->isChecked()" << ui->chk_vt_enable->isChecked()
                     << "vt_software():" << camera_ptr->vt_software()
                     << "Camera:" << camera_ptr->get_camera_name() << "*****]";
    }

    // Show/hide the reticle, but only if it's not already in the proper state.
    if (camera_ptr->vt_reticle_visible() != ui->chk_vt_enable->isChecked())
        camera_ptr->vt_reticle_show(ui->chk_vt_enable->isChecked());

    // Enable/disable the controls
    ui->chk_vt_marker_only->setEnabled(!camera_ptr->vt_software() && ui->chk_vt_enable->isChecked());
    ui->chk_vt_marker_only->setChecked(false);

    ui->pb_vt_reset->setEnabled(ui->chk_vt_enable->isChecked());
    ui->lbl_vt_area->setEnabled(ui->chk_vt_enable->isChecked());
    ui->sld_vt_area->setEnabled(ui->chk_vt_enable->isChecked());
    ui->spin_vt_area->setEnabled(ui->chk_vt_enable->isChecked());
    ui->lbl_vt_change->setEnabled(ui->chk_vt_enable->isChecked());
    ui->sld_vt_change->setEnabled(ui->chk_vt_enable->isChecked());
    ui->spin_vt_change->setEnabled(ui->chk_vt_enable->isChecked());
    ui->lbl_refresh->setEnabled(ui->chk_vt_enable->isChecked());
    ui->spin_refresh->setEnabled(ui->chk_vt_enable->isChecked());
}

void PanelRecControls::vt_mark_only(bool /*checked*/)
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    // Pass along the new information.
    camera_ptr->set_vt_mark_only(ui->chk_vt_marker_only->isChecked());
}

void PanelRecControls::vt_reset()
{
    CameraMDI   *camera_ptr;

    // Make sure we have an active camera pointer.
    camera_ptr = p_control_panel->getCamera();
    if (camera_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera!";
        return;
    }

    if (VERBOSE_LOGGING)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[***** vt_reticle_visible():" << camera_ptr->vt_reticle_visible()
                     << "chk_vt_enable->isChecked()" << ui->chk_vt_enable->isChecked()
                     << "vt_software():" << camera_ptr->vt_software()
                     << "Camera:" << camera_ptr->get_camera_name() << "*****]";
    }

    // Reset the location/size of the reticle.
    camera_ptr->vt_reticle_reset();
}

/*!
 * @brief Write the new Arm-In settings to the camera.
 */
void PanelRecControls::write_arm_in()
{
    uint32_t    reg_value;

    // Start the write sequence.
    update_in_progress = true;

    // Are we enabled?
    if (ui->arm_in_enable->isChecked())
    {
        // Initial setting.
        reg_value = GIGE_ARM_TYPE_EXTERNAL;

        // Add in the channel selected.
        if (arm_in_chan >= 0)
            reg_value |= (1 << (GIGE_ARM_IN_CHANNEL_SHIFT + arm_in_chan));

        // Let any listeners know that the external I/O is attempting to set arm-in.
        emit settingExtArmInEnabled();
    }
    else
    {
        // Initial setting.
        reg_value = GIGE_ARM_TYPE_INTERNAL;

        // Let any listeners know that the external I/O is attempting to set arm-in.
        emit settingExtArmInDisabled();
    }
    write_list.append(REG_VALUE_PAIR(GIGE_REG_ARM_IN_ENABLE, reg_value));

    // Are we edge or level?
    if (ui->rb_ai_edge_type->isChecked())
    {
        // User has selected and edge type event.
        reg_value = GIGE_ARM_IN_EDGE_TYPE;

        // Build up the polarity setting now. Are we rising or falling edge?
        reg_value |= ui->rb_ai_rising_high->isChecked() ? GIGE_ARM_IN_RISING : GIGE_ARM_IN_FALLING;

        // Make sure the UI allows the user to select rising or falling edges.
        ui->rb_ai_rising_high->setEnabled(true);
    }
    else
    {
        // With a level event, we only allow falling/low.
        reg_value = GIGE_ARM_IN_LEVEL_TYPE | GIGE_ARM_IN_FALLING;

        // Force the UI to reflect this.
        ui->rb_ai_falling_low->setChecked(true);
        ui->rb_ai_rising_high->setEnabled(false);
    }

    // Start the writing.
    write_camera_register(GIGE_REG_ARM_IN_POLARITY, reg_value);
}

/*!
 * @brief Write the new Arm-Out settings to the camera.
 */
void PanelRecControls::write_arm_out()
{
    uint32_t    reg_value;

    // Start the write sequence.
    update_in_progress = true;

    // Are we enabled?
    if (ui->arm_out_enable->isChecked())
    {
        // Initial setting.
        reg_value = GIGE_ARM_OUT_ENABLE;

        // Add in the channel selected.
        if (arm_out_chan >= 0)
            reg_value |= (1 << (GIGE_ARM_OUT_CHANNEL_SHIFT + arm_out_chan));
    }
    else
    {
        // Initial setting.
        reg_value = GIGE_ARM_OUT_DISABLE;
    }

    // Are we doing pass-thru?
    if (ui->chk_ao_pass_thru->isChecked())
    {
        reg_value |= GIGE_ARM_OUT_IS_ARM_IN;

        // Are we inverting the signal?
        reg_value |= (ui->chk_ao_invert_signal->isChecked() ? GIGE_ARM_OUT_INVERT_ARMIN : GIGE_ARM_OUT_PASS_ARMIN);
    }
    else
    {
        // Are we using the pulse width?
        if (ui->chk_ao_pulse_width->isChecked())
            reg_value |= ((uint32_t)ui->spin_ao_pulse_width->value() << GIGE_ARM_OUT_PERIOD_SHIFT);

        // Active high or low?
        reg_value |= (ui->rb_ao_active_high->isChecked() ? GIGE_ARM_OUT_ACTIVE_HIGH : GIGE_ARM_OUT_ACTIVE_LOW);
    }

    // Write the control settings to the camera.
    write_camera_register(GIGE_REG_ARM_OUT_SELECT, reg_value);
}

/*!
 * @brief Slot method which receives notification when a write completes.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void PanelRecControls::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    // Is this event for us?
    if (ip_addr != camera_ip_addr)
        return;

    // Are there still write operations to perform?
    if (!write_list.isEmpty())
    {
        REG_VALUE_PAIR  reg_pair;

        // Grab the next operation off the list.
        reg_pair = write_list.takeFirst();

        // Issue the write command.
        write_camera_register(reg_pair.first, reg_pair.second);
    }

    // Did we just finish writing the register we're interested in monitoring?
    if (update_reg_addr == reg_addr)
    {
//        QLOG_TRACE() << __PRETTY_FUNCTION__ << "Updating the update_reg_addr:" << Qt::showbase << Qt::hex << reg_addr;

        // Reset our retry count.
        update_retry_count = 0;

        // Read this back so we can refresh the cache.
        start_camera_read(reg_addr);
    }
    else if (update_in_progress && (update_reg_addr == 0))
    {
        // We weren't waiting for a specific register. Just holding off in general.
        update_in_progress = false;
    }
}

/*!
 * @brief Write the new Sync-In settings to the camera.
 */
void PanelRecControls::write_sync_in()
{
    uint32_t    reg_value;
    REG_VALUE_PAIR  reg_pair;

    // Start the write sequence.
    update_in_progress = true;

    // Initialize our register based on if we're enabled or not.
    if (ui->sync_in_enable->isChecked())
    {
        // Initial setting.
        reg_value = GIGE_SYNC_TYPE_EXTERNAL;

        // Add in the channel selected.
        if (sync_in_chan >= 0)
            reg_value |= (1 << (GIGE_SYNC_IN_CHANNEL_SHIFT + sync_in_chan));
    }
    else
    {
        PanelRecSettings *settings_ptr;

        // Initial setting.
        reg_value = GIGE_SYNC_TYPE_INTERNAL;

        // Get the current record settings panel pointer..
        settings_ptr = static_cast<PanelRecSettings *>(p_control_panel->getPanel(CONTROL_PANEL::PANEL_REC_SETTINGS));

        // Have the pointer?
        if (settings_ptr)
        {
            // Disabling sync-in but it is possible that the frame rate and
            // exposure have been modified so attempt to resync up with the
            // current settings in the record settings panel.
            settings_ptr->push_framerate_and_exposure();
        }
    }

    // Are we per-frame or per-second?
    if (ui->rb_si_per_frame->isChecked())
    {
        uint32_t    calculated_rate;
        uint32_t    divisor;
        uint32_t    frame_rate;
        uint32_t    master_rate;
        uint32_t    skip_pulse;

        // We're per-frame.
        reg_value |= GIGE_SYNC_IN_PPF;

        // Since we're doing per frame, we need to calculate the skip pulses from the
        // master frame rate and the divisor.
        master_rate = (uint32_t)ui->spin_si_mfr->value();
        if (master_rate == 0)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Invalid MFR!";
            update_in_progress = false;
            return;
        }

        // Update the camera with the new rate.
        write_list.append(REG_VALUE_PAIR(GIGE_REG_SYNC_IN_MFR, master_rate));

        // Calculate the rate the user wants.
        divisor = (uint32_t)ui->spin_si_rate_div->value();
        if (divisor < 1)
            divisor = 1;
        calculated_rate = master_rate / divisor;

        // Make sure it's not lower than our minimum allowed.
        if (calculated_rate < MIN_CALC_FRAMERATE)
        {
            calculated_rate = MIN_CALC_FRAMERATE;
        }

        // Update the skip pulse value.
        skip_pulse = divisor - 1;
        reg_value |= (skip_pulse << GIGE_SYNC_IN_SKIP_SHIFT);

        // Are we changing the frame rate?
        frame_rate = (uint32_t)ClientMath::reg_to_frame_rate(read_camera_cache(GIGE_REG_FRAME_RATE));
        if (calculated_rate != frame_rate)
        {
            uint32_t    current_exp;
            uint32_t    frame_overhead;
            uint32_t    max_exposure;

            // Get the frame overhead value.
            frame_overhead = read_camera_cache(GIGE_REG_FRAME_OVERHEAD);

            // Get the current exposure setting.
            current_exp = read_camera_cache(GIGE_REG_SHUTTER_SPEED);

            // If the current shutter speed is not valid for the new frame rate then
            // adjust the shutter speed
            max_exposure = ClientMath::calc_max_exposure(calculated_rate, frame_overhead);
            if (max_exposure < current_exp)
            {
                // Update the camera's exposure.
                write_list.append(REG_VALUE_PAIR(GIGE_REG_SHUTTER_SPEED, max_exposure));
            }

            // The settings the user gave us will require us to change the frame rate.
            write_list.append(REG_VALUE_PAIR(GIGE_REG_FRAME_RATE, calculated_rate));
        }
    }

    // Per ticket #227, the following is disabled.
#ifdef TICKET_227
    else
    {
        uint32_t    delay;

        // We're per-second.
        reg_value |= GIGE_SYNC_IN_PPS;

        // Set the sync-in delay value.
        delay = (uint32_t)ui->spin_si_delay->value();
        write_list.append(REG_VALUE_PAIR(GIGE_REG_SYNC_DELAY, delay));
    }
#endif /* TICKET_227 */

    // Push the control settings on the list.
    write_list.append(REG_VALUE_PAIR(GIGE_REG_SYNC_IN_ENABLE, reg_value));

    // Now push the polarity on the list.
    reg_value = (ui->rb_si_rising->isChecked() ? GIGE_SYNC_IN_RISING : GIGE_SYNC_IN_FALLING);
    write_list.append(REG_VALUE_PAIR(GIGE_REG_SYNC_IN_POLARITY, reg_value));

    // Pop the first value of the list and start the writing.
    reg_pair = write_list.takeFirst();
    write_camera_register(reg_pair.first, reg_pair.second);
}

/*!
 * @brief Write the new Sync-Out settings to the camera.
 */
void PanelRecControls::write_sync_out()
{
    uint32_t reg_value;

    // Initialize the register value from the UI settings.
    reg_value = sync_out_ui_to_gige_reg();

    if (LOG_SYNC_OUT_TRACES)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "after sync_out_ui_to_gige_reg() the reg_value:" << Qt::showbase << Qt::hex << reg_value;
    }

    // Only persist enabled settings.
    if ((reg_value & GIGE_SYNC_OUT_ENABLE) == GIGE_SYNC_OUT_ENABLE)
    {
        // Save the settings to our persistent storage.
        sync_out_save_settings(reg_value);
    }
    else if (p_control_panel->getCamera()->is_series3())
    {
        // We are disabling sync-out in this case. We only want to set the register to GIGE_SYNC_OUT_DISABLE.
        // But the call above to sync_out_ui_to_gige_reg() set reg_value based on UI in the Sync-Out panel which sets other bits.
        // TICKET #451:
        // For our purposes let's only set the register to zero to clear out everything.
        // Only doing this for Series3 cameras because they don't like any extra bits set when disabling sync-out.
        // TODO: Possibly, it may make sense to have an "update_reg_bit_mask" to only check for a bit rather than
        // the entire register setting for equality when the write is verified.  Search on "#451" to find the other
        // part of this ticket fix and relavant location where that would apply.
        reg_value = GIGE_SYNC_OUT_DISABLE;
        if (LOG_SYNC_OUT_TRACES)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "disabling, modified to reg_value:" << Qt::showbase << Qt::hex << reg_value;
        }
    }

    // Indicate that we're updating.
    update_in_progress = true;
    update_reg_addr = GIGE_REG_SYNC_OUT_SELECT;
    update_reg_value = reg_value;

    // Write out the value.
    write_camera_register(GIGE_REG_SYNC_OUT_SELECT, reg_value);
}

/*!
 * @brief Write the new Trigger-Out settings to the camera.
 */
void PanelRecControls::write_trig_out()
{
    uint32_t    reg_value;

    // First find out if trigger out is enabled or disabled, if disabled there's not much more
    // to check.
    if (ui->trig_out_enable->isChecked())
    {
        // Trigout functionality is enabled.
        reg_value = GIGE_TRIG_OUT_ENABLE;

        // Is pass-through enabled?
        if (ui->chk_trigout_pass_thru->isChecked())
        {
            // We are passing trig-in to trig-out.
            reg_value |= GIGE_TRIG_OUT_IS_TRIG_IN;

            // Are we inverting?
            if (ui->chk_trigout_invert_signal->isChecked())
                reg_value |= GIGE_TRIG_OUT_INVERT_TRIGIN;
            else
                reg_value |= GIGE_TRIG_OUT_PASS_TRIGIN;
        }
        else
        {
            uint32_t    pulse_width;

            // Get the value from the control.
            pulse_width = (uint32_t)ui->spin_trigout_pulse_width->value();

            // We present the value in milliseconds, but the register
            // is actually specified in microseconds.
            pulse_width *= 1000;

            // Set the pulse width value into our register.
            reg_value |= (pulse_width << GIGE_TRIG_OUT_PULSE_SHIFT);

            // Are we active-high or -low?
            if (ui->rb_trigout_active_high->isChecked())
                reg_value |= GIGE_TRIG_OUT_ACTIVE_HIGH;
            else
                reg_value |= GIGE_TRIG_OUT_ACTIVE_LOW;
        }

        // Add in the channel selected.
        if (trig_out_chan >= 0)
            reg_value |= (1 << (GIGE_TRIG_OUT_CHANNEL_SHIFT + trig_out_chan));
    }
    else
    {
        // The trigger out functionality is diabled.
        reg_value = GIGE_TRIG_OUT_DISABLE;
    }

    // Let folks know we have an update in progress.
    update_in_progress = true;

    // Write the new register settings.
    write_camera_register(GIGE_REG_TRIG_OUT_SELECT, reg_value);
}

