/*!
 * @file save_video.cpp
 * @brief The code used for saving videos to the workstation.
 *
 * Copyright 2014 (c) Fastec Imaging as an unpublished work.
 * All Rights Reserved.
 *
 * The information contained herein is confidential property of
 * Fastec Imaging. The use, copying, transfer or disclosure of such
 * information is prohibited except by express written agreement
 * with Fastec Imaging.
 */

//--------------------------------------------------------------------
// Includes
//--------------------------------------------------------------------
#include "save_video.h"

#include <QByteArray>
#include <QDeadlineTimer>
#include <QDir>
#include <QFile>
#include <QGraphicsScene>
#include <QJsonDocument>
#include <QJsonObject>
#include <QList>
#include <QMessageBox>
#include <QMutex>
#include <QMutexLocker>
#include <QStandardPaths>
#include <QtConcurrent>
#include <QElapsedTimer>
#include <QWaitCondition>
#include <QStyle>

#include <errno.h>

#include <memory>

#include "camera_mdi.h"
#include "camera_video.h"
#include "client_math.h"
#include "fi_fflr_underflow_dialog.h"
#include "fi_comm.h"
#include "fi_msgbox.h"
#include "fi_progress_dialog.h"
#include "fi_sensor.h"
#include "logger.h"
#include "mainwindow.h"
#include "save_dialog.h"
#include "video_controls.h"

#include "fi_durations.h"

namespace SAVE_VIDEO
{
//--------------------------------------------------------------------
// Constant, Macro and Type Definitions
//--------------------------------------------------------------------

// Log messages for testing purposes.  Set to 1 to have them compiled for output.
#define LOG_AUTOSAVE_TRACING            1
#define LOG_SAVE_STATE_TRACING          0
#define LOG_WRITE_COMPLETE              0
#define LOG_GIGE_REG_PROGRESS_UPDATES   0
#define LOG_VERBOSE_INFO                0

//! Common message for warning about video clip range errors.
const QString CLIP_RANGE_INVALID = QStringLiteral("Video clip range is invalid.");

//! Common message for indicating no camera name exists.
const QString EMPTY_CAMERA_NAME = QStringLiteral("Camera Name Not Found");

//! The polling period for save updates.
const int   GIGE_REG_SAVE_STATE_POLLING_RATE    = 500;

//! The number of milliseconds we will stay in the wait().
const int   WAIT_TIMEOUT_MS                     = 500;

//! Number of milliseconds before the camera server needs a message from us.
const int   SAVE_HEARTBEAT                      = 30000;

//! How long to wait before launch FasFireLR underflow.
const int   FFLR_UNDERFLOW_LAUNCH_DELAY         = 1000;

//! How long to wait before we poll again for FasFireLR underflow progress updates.
const int   FFLR_UNDERFLOW_PROGRESS_POLL        = 250;

//! Retest conditions for launching FasFireLR underflow error recovery message.
const int   FFLR_UNDERFLOW_RECHECK_DELAY        = 2000;

//! The number of wait delays we will allow before just giving up.
const int   FFLR_UNDERFLOW_RECHECK_MAX          = 10;

//! Number of retries we will allow for starting a camera save operation when camera is busy.
const int   WRITE_SAVE_JSON_MAX_RETRIES_IF_BUSY = 5;

//! Time to wait between retries if camera is busy.
const int   WRITE_SAVE_JSON_RETRY_WAIT_TIME     = 1000;

//! If you want to turn on verbose logging for this object set this to 1
#if LOG_VERBOSE_INFO
static const bool VERBOSE_SAVE_LOG = true;
#else
static const bool VERBOSE_SAVE_LOG = false;
#endif

}   // End of namespace SAVE_VIDEO.

using namespace SAVE_VIDEO;
using namespace FI_SAVER;

/*!
 * @brief The class constructor.
 */
SaveVideo::SaveVideo(QObject *parent) :
    FiSaver(parent)
{
    // Make sure our member variables are initialized.
    as_clip_end = 0;
    as_clip_start = 0;
    autosave_enabled = false;
    autosave_configured = false;
    cancel_in_progress = false;
    jpeg_quality_old_camset = -1;
    localSaveToPath = false;
    local_autosave_enabled = false;
    last_save_state = 0;
    frameCount = 0;
    basePath.clear();
    cameraName.clear();
    tagString.clear();
    save_state_poll_timer = nullptr;
    setupForSave = false;
    savingHasFinished = false;
    saveBufferCount = 0;
    saving_finished_in_process = false;
    calledByCopyToPC = false;
    emittedSaveComplete = false;
    emittedSaveError = false;
    emittedSaveReceiveCountComplete = false;
    closeInProgress = false;
    closeCompleted = false;
    thread_stop = false;
    convertRawToXml = true;
    waiting_for_pause = false;
    trigFrameNo4Save = 0;
    releasedBufferCount = 0;

    active_device_save = SAVE_VIDEO::DEVICE_SAVE_STATE::SAVE_NONE;

    fflr_underflow_wait_delays = 0;
    fflr_underflow_state = FFLR_UNDERFLOW_STATE::NONE;
    fflr_underflow_dirty_count = 0;

    write_save_json_requested   = false;
    // write_save_json_reqtime is empty by construction.
    // write_save_json_rsptime is empty by construction.
    write_save_json_reqstatus   = FI_COMM::GEV_STATUS_UNKNOWN;

    as_save_info.clip_end = as_save_info.clip_start = 0;
    as_save_info.clip_end_abs = as_save_info.clip_start_abs = 0;

    ResetSaveVideoState();

    // For construction pull any settings out of the application SaveSettings.
    // Assumes the fiSaver base class is already constructed.
    SaveDialog::getH264Settings(h264Speed, h264Quality, h264EnableAllKeyframes);
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "h264Quality:" << h264Quality;

    // Setup our connections.
    connect(this, SIGNAL(free_list_ready()), this, SLOT(free_buffer()), Qt::QueuedConnection);
    connect(this, &SaveVideo::save_fflr_underflow_error, this, &SaveVideo::process_fflr_underflow, Qt::QueuedConnection);
    connect(this, &SaveVideo::worker_thread_exiting, this, &SaveVideo::do_worker_thread_exit, Qt::QueuedConnection);
}

/*!
 * @brief The class destructor.
 */
SaveVideo::~SaveVideo()
{
    // Cleanup any UI element(s).
    cleanup_save();
}

/*!
 * @brief User has cancelled autosave.
 */
void SaveVideo::autosave_cancelled()
{
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

#if LOG_AUTOSAVE_TRACING
    QString ip_addr_pretty = "IP[No Camera Parent]";
    if (my_parent)
    {
        ip_addr_pretty = "IP[" + my_parent->get_ip_addr_pretty() + "]";
    }
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "[<<<<ENTERED>>>>]" << ip_addr_pretty
                 << "autosave_enabled" << autosave_enabled
                 << "autosave_configured" << autosave_configured;
#endif

    // Make sure we're in autosave.
    if (autosave_configured && autosave_enabled)
    {
        // Sanity check.
        if (my_parent != nullptr)
        {
            // Pass on the cancel.
            QTimer::singleShot(0, my_parent, &CameraMDI::cancel_save);
        }
    }
}

/*!
 * @brief Display the autosave settings and allow the user to disable autosave mode.
 */
void SaveVideo::autosave_display()
{
    uint32_t        cam_ip_addr;
    FI_comm         *comm_ptr;
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Get our camera's IP address.
    cam_ip_addr = my_parent->get_ip_addr();

    // Create the save dialog.
    SaveDialog save_dlg(cam_ip_addr, my_parent);

    // Initialize the save parameters.
    save_dlg.init_controls();
    save_dlg.set_autosave_mode(true);

    // Are there overlays?
    if (my_parent->get_player())
        save_dlg.enable_overlay_merge(my_parent->get_player()->has_overlay());

    // Tell the dialog to disable the settings - this is view only.
    save_dlg.set_view_only(true);

    // Are we saving locally?
    if (local_autosave())
    {
        // Sanity check.
        if (as_clip_end < as_clip_start)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << CLIP_RANGE_INVALID << "[as_clip_start:" << as_clip_start << "as_clip_end:" << as_clip_end << "]";
        }

        // Use the saved values.
        save_dlg.set_rel_range(as_clip_start, as_clip_end);
    }
    else
    {
        // Read up the clip start/end values.
        uint32_t user_clip_start = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_USER_SAVE_START_NUM);
        uint32_t user_clip_end = comm_ptr->get_reg(cam_ip_addr, GIGE_REG_USER_SAVE_END_NUM);

        // Sanity check.
        if (user_clip_end < user_clip_start)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << CLIP_RANGE_INVALID << "[user_clip_start:" << user_clip_start << "user_clip_end:" << user_clip_end << "]";
        }

        save_dlg.set_clip_range(user_clip_start, user_clip_end);
    }

    // Pop up the dialog.
    if (save_dlg.exec() == QDialog::Accepted)
    {
        // Stop autosave.
        autosave_stop();
    }
}

/*!
 * @brief Display the [autosave settings](@ref SaveDialog) and allow the user to enable Autosave.
 *
 * @return True on success, false if there was an error.
 */
bool SaveVideo::autosave_setup()
{
    bool            rc;
    uint32_t        fflr_type;
    uint32_t        cam_ip_addr;
    FI_comm         *comm_ptr;
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

#if LOG_AUTOSAVE_TRACING
    QString ip_addr_pretty = "IP[No Camera Parent]";
    if (my_parent)
    {
        ip_addr_pretty = "IP[" + my_parent->get_ip_addr_pretty() + "]";
    }
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "[<<<<ENTERED>>>>]" << ip_addr_pretty
                 << "autosave_enabled" << autosave_enabled
                 << "autosave_configured" << autosave_configured;
#endif

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return false;
    }

    // Get our camera's IP address.
    cam_ip_addr = my_parent->get_ip_addr();

    // See if we need to free captured images.
    if (!queryFreePartition())
        return false;

    // Reset our clip range.
    as_clip_start = as_clip_end = 0;


    // If this isn't a reconfiguring case for modified width/height/bit depth then ask user
    // for the save params unless FasFireLR capture mode.
    {
        // Create the save dialog.
        SaveDialog save_dlg(cam_ip_addr, my_parent);

        // Initialize the save parameters. This will also restore settings.
        save_dlg.init_controls();
        save_dlg.set_autosave_mode(true);

        // Default to the entire buffer.
        save_dlg.set_clip_range(0, -1);

        // Are there overlays?
        if (my_parent && my_parent->get_player())
            save_dlg.enable_overlay_merge(my_parent->get_player()->has_overlay());

        // Don't execute the dialog if FasFireLR mode.
        if (!my_parent->is_fflr_enabled(fflr_type))
        {
            // Get the user's input on what/how they want the video saved.
            if (save_dlg.exec() == QDialog::Rejected)
                return false;
        }
        // Retrieve the settings.
        save_dlg.get_save_params(as_save_info, my_parent->is_series3());

        // If FasFireLR mode make sure the save parameters are set up properly.
        if (my_parent->is_fflr_enabled(fflr_type))
        {
            as_save_info.dest_label = my_parent->get_fflr_device_label();
            as_save_info.dest_path  = my_parent->get_fflr_device_mountdir();
            as_save_info.dest_media = my_parent->get_fflr_device_type();

            as_save_info.streamType = FI_SAVER::STREAM_TYPE::FI_CAPSTK;
            as_save_info.imagesInputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
            as_save_info.imagesOutputType = FI_SAVER::IMAGE_TYPE::NO_IMAGE;
            as_save_info.comment.clear();
        }
    }

    // Configure our parent class
    if (FiSaver::configure(as_save_info))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "fiSaver::configure() failed!";
        return false;
    }

    // Setup the comment. If the user wants one or not, we have to do this to make
    // sure that the comment is set or cleared, as required.
    comm_ptr->start_write_mem(cam_ip_addr, GIGE_REG_COMMENT, *commentBuf);

    // Get the RELATIVE start/stop values for the clip range.
    as_clip_end = as_save_info.clip_end;
    as_clip_start = as_save_info.clip_start;

    // Do we need to update the JPEG quality setting?
    set_jpeg_quality_old_camset((int)comm_ptr->get_reg(cam_ip_addr, GIGE_REG_JPEG_QUALITY), __PRETTY_FUNCTION__);
    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "jpeg_quality_old_camset", (uint32_t)jpeg_quality_old_camset, false);
    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "save_info.jpeg_quality", (uint32_t)as_save_info.jpeg_quality, false);

    // If we require a certain jpeg quality to stream from the camera we need to tell
    // the camera to stream that jpeg quality.  Keep a copy of the old setting so that
    // it can be restored after the save operation is completed.
    if (as_save_info.jpeg_quality != jpeg_quality_old_camset)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Modifying the JPEG quality factor to " << as_save_info.jpeg_quality;
        // Write out the new quality value.
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_JPEG_QUALITY, (uint32_t)as_save_info.jpeg_quality);
    }

    // NOTE:  The only place where "jpeg_quality_4QT_pixmaps" is actually used is when the frame is
    // sent to fiSaver::SaveToStream and the output of the save operation is a JPEG image
    // after running through a QT image translation. That is when this value is used.
    jpegQuality_4QTPixmaps = as_save_info.jpeg_quality;

    // Autosave is enabled.
    autosave_enabled = true;

    // Are we saving to the workstation or on the camera?
    if (as_save_info.dest_media == SAVE_DLG::MEDIA_LOCAL_PC)
    {
        // Do we have a path?
        if (as_save_info.dest_path.isEmpty())
        {
            QString msg_str = tr("The download path must not be empty!");
            QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
            QMessageBox::critical(my_parent, tr("Configuration Error"), msg_str);
            return false;
        }
        else
        {
            // Make sure the path exists.
            QDir download_dir(as_save_info.dest_path);
            if (!download_dir.exists())
            {
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "Path doesn't exist. Creating:" << download_dir;
                if (!download_dir.mkpath(as_save_info.dest_path))
                {
                    // Can't create the path, so use the Documents folder instead.
                    QString msg_str = tr("The download path (%1) does not exist.").arg(as_save_info.dest_path);
                    QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
                    QMessageBox::critical(my_parent, tr("Configuration Error"), msg_str);
                    return false;
                }
            }
        }

        // Remember we're in local autosave mode.
        local_autosave_enabled = true;

        // Reset our local settings.
        ResetSaveVideoState();

        // Even though it's an autosave we also want to know that we're saving locally to our path.
        localSaveToPath = true;

        // Go through and save all the important settings.
        rc = save_to_pc(as_save_info);

        // Tell the camera that we're saving to the PC.
        comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_IMAGE_DESTINATION, GIGE_IMAGE_DEST_OFF_CAMERA);
    }
    else
    {
        // Remember that we're not saving locally.
        local_autosave_enabled = false;

        // And set our general save to pc flag.
        localSaveToPath = false;

        // Have the camera start saving to it's media.
        if (my_parent->is_series3())
            rc = save_to_camera(as_save_info);
        else
            rc = save_to_camera_json(as_save_info);
    }

    // Keep a record of width, height and bit-depth at the time of this autosave configuration.
    as_config_width = my_parent->read_camera_cache(GIGE_REG_WIDTH);
    as_config_height = my_parent->read_camera_cache(GIGE_REG_HEIGHT);
    as_config_bit_depth = my_parent->read_camera_cache(GIGE_REG_BIT_MODE);
    as_config_part_frames = my_parent->calc_max_frames();

    // We are configured.
    autosave_configured = true;

    // Do we have our transfer progress dialog ready to go?
    if (rc)
    {
        rc = PrepXferDialogFromSaveParams(as_save_info, localSaveToPath, __PRETTY_FUNCTION__);
    }

    // Get our local settings ready to go.
    autosave_start();

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Autosave Setup.  Expecting " << frameCount << "receive buffers.";
    }

    // Ready for autosave.
    return rc;
}

bool SaveVideo::autosave_start()
{
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

#if LOG_AUTOSAVE_TRACING
    QString ip_addr_pretty = "IP[No Camera Parent]";
    if (my_parent)
    {
        ip_addr_pretty = "IP[" + my_parent->get_ip_addr_pretty() + "]";
    }
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "[<<<<ENTERED>>>>]" << ip_addr_pretty
                 << "autosave_enabled" << autosave_enabled
                 << "autosave_configured" << autosave_configured;
#endif

    // Have we been configured?
    if (!autosave_configured)
    {
        QString msg_str;
        QString autosave_enabled_warning = tr("Autosave will disable on camera and ");
        bool camera_set_to_autosave = autosave_is_enabled();

        msg_str = tr("Did the camera go offline?  Autosave configuration must be reset.\nAfter you hit [Ok] %1you be presented with Autosave Setup.").arg(camera_set_to_autosave ? "" : autosave_enabled_warning);
        QMessageBox::information(my_parent, my_parent->get_camera_name(), msg_str);
        QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;

        // If the autosave is enabled on the camera, then stop it.
        if (camera_set_to_autosave)
        {
            autosave_stop();
        }
        // And now reconfigure the Autosave operation.  User will have to re-enable it.
        // NOTE:  It is known that this will end up recursing back into this function.
        // It will unwind properly.
        return (autosave_setup());
    }

    // See if we have partitions to free.
    if (!queryFreePartition())
        return false;

    // Since we know we are configured (autosave_setup had been called) - Is this a re-enable of autosave?
    // autosave_enabled if set false indicates that it has been disabled.
    if (!autosave_enabled)
    {
        uint32_t width = my_parent->read_camera_cache(GIGE_REG_WIDTH);
        uint32_t height = my_parent->read_camera_cache(GIGE_REG_HEIGHT);
        uint32_t bit_depth = my_parent->read_camera_cache(GIGE_REG_BIT_MODE);
        uint32_t part_nframes;

        // Calculate the number of frames that fit into the partition.
        part_nframes = my_parent->calc_max_frames();

        // Has our width/height or bit-depth been modified since we configured for
        // autosave?  If it has the save process will produce bad CAP data in some
        // cases because the saver was setup to operate with stale ROI information.
        //
        // NOTE:
        // THE REASON this has to happen is that the SAVE_PARAMS is out of date from the
        // original configuration.  We have NO way to safely retrieve the SAVE_PARAM data
        // unless the SaveDialog is executed and the SAVE_PARAM data pulled after that
        // hapens.  Therefore the best approach with our toolset to date is to re-run the
        // full autosave_setup process again.  This will ensure that the user has verified
        // the setup, is allowed to make a change, and we have a properly initialized
        // SAVE_PARAM with which to run throught the saver process in autosave_setup and
        // eventually when it calls this at the end of autosave_setup.
        //
        // This isn't the cleanest way to do this but it will work.  The intelligence
        // about saving autosave in the saver object needs to stay encapsulated here.
        // We will know right away with this comparison if we have a different setup
        // for the ROI or bit depth.
        //
        // Without refreshing the SAVE_PARAMS and re-running autosave_setup the code is
        // guaranteed to produce bad CAP files...  If the ROI information has stayed equal
        // then it will work to process the next save given that we are dealing with the
        // same set of SAVE_PARAM data originally generated in the original autosave_setup
        // call.  I know this is confusing but this note is important to understand.
        //
        // autosave_start is called everytime an autosave is enabled or re-enabled.  autosave_stop
        // is called everytime autosave is disabled.  But everytime we disable it the user has
        // the opportunity to muck with the ROI or bit depth and then re-enable autosave with
        // stale parameters.  It works it's not elegant but it works for the moment.  If we
        // ever get the time to rework SAVE_PARAMS and the data that a saver object uses we can
        // revisit this code at that time.
        bool width_changed = (width != as_config_width);
        bool height_changed = (height != as_config_height);
        bool bit_depth_changed = (bit_depth != as_config_bit_depth);
        bool part_nframes_changed = (part_nframes != as_config_part_frames);

        if (width_changed || height_changed || bit_depth_changed || part_nframes_changed)
        {
            QString empty_str;
            QString width_str           = (width_changed)       ? tr("\nFrame Width[%1] was changed to [%2]").arg(as_config_width).arg(width) : empty_str;
            QString height_str          = (height_changed)      ? tr("\nFrame Height[%1] was changed to [%2]").arg(as_config_height).arg(height) : empty_str;
            QString bit_depth_str       = (bit_depth_changed)   ? tr("\nFrame Bit Depth was modified") : empty_str;
            QString part_nframes_str    = (part_nframes_changed) ? tr("\nPartition Size was modified") : empty_str;

            QString msg_str = tr("Autosave configuration reset because record information changed:\n%1%2%3%4").arg(width_str).arg(height_str).arg(bit_depth_str).arg(part_nframes_str);
            QMessageBox::information(my_parent, my_parent->get_camera_name(), msg_str);
            //QString warn_msg = QString::fromUtf8("RE-ENABLE AUTOSAVE RECONFIGURE CHECK w[%1] as_w[%2] h[%3] as_h[%4] b[%5] as_b[%6]").arg(width).arg(as_config_width).arg(height).arg(as_config_height).arg(bit_depth).arg(as_config_bit_depth);
            QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
            // NOTE:  It is known that this will end up recursing back into this function.
            // It will unwind properly.
            return (autosave_setup());
        }
    }

    // Autosave is enabled.
    autosave_enabled = true;

    // Reset our local settings.
    ResetSaveVideoState();

    // NOTE:  These two connect signals were added to respond to Ticket# 253.  There are different
    // conditions associated with each connection.  When we are auto saving to a device there will
    // *not* be a visible xferDialog on the screen and in that case the SaveVideo::userCancelledSaveSignal
    // will not fire.  We can however receive any CameraMDI::dearm_session_discard signal
    // should the user cancel the save-to-device autosave condition by dearming the camera.
    //
    // The other condition is when we are auto saving to the PC.  In this condition we will have
    // a visible xferDialog and the user cancels the auto save session by hitting the "cancel" button
    // on that xferDialog and in that case the SaveVideo::userCancelledSaveSignal is fired.

    // If the user cancels the operation we need to stop the autosave process.
    connect(this, &SaveVideo::userCancelledSaveSignal, this, &SaveVideo::autosave_cancelled, Qt::QueuedConnection);

    // If our parent responds to a dearm condition the user may choose to discard the
    // active autosave.  If this happens we want to transform the signal into a cancel
    // causing the current session save to be discarded.
    connect(my_parent, &CameraMDI::dearm_session_discard, this, &SaveVideo::autosave_cancelled, Qt::QueuedConnection);

    // Write the autosave enable register to the camera.
    my_parent->write_camera_register(GIGE_REG_AUTOSAVE, (uint32_t)GIGE_AUTOSAVE_ENABLE);
    return true;
}

/*!
 * @brief Disable autosave.
 * @return True on success, false otherwise.
 */
bool SaveVideo::autosave_stop()
{
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

#if LOG_AUTOSAVE_TRACING
    QString ip_addr_pretty = "IP[No Camera Parent]";
    if (my_parent)
    {
        ip_addr_pretty = "IP[" + my_parent->get_ip_addr_pretty() + "]";
    }
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "[<<<<ENTERED>>>>]" << ip_addr_pretty
                 << "autosave_enabled" << autosave_enabled
                 << "autosave_configured" << autosave_configured;
#endif

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // This is a disable request. Tell the camera that we're no longer in autosave.
    my_parent->write_camera_register(GIGE_REG_AUTOSAVE, GIGE_AUTOSAVE_DISABLE);

    // Turn off the state of being autosave enabled.
    // NOTE:  The autosave_configured state persists until the camera is de-armed but
    // autosaving may start and stop several times while partition data (Device saves)
    // or PC disk space (Path saves) exists to support the "video" data.
    autosave_enabled = false;

    // We don't need to respond to these conditions now that we have stopped auto save.
    disconnect(this, &SaveVideo::userCancelledSaveSignal, this, &SaveVideo::autosave_cancelled);
    disconnect(my_parent, &CameraMDI::dearm_session_discard, this, &SaveVideo::autosave_cancelled);

    if (fflrDearmDlg)
    {
        fflrDearmDlg->hide();
        fiProgressDialog::RecycleProgressDialog(fflrDearmDlg);
    }

    // We're done.
    return true;
}

/*!
 * @brief Does this autosave save object want to download the
 * per frame metadata when it saves to a local PC?
 *
 * @return True if saver is setup for autosave and the streamType is not a CAP file, or False
 * if the saver is not configured for autosave or if configured the streamType is a CAP file.
 */
bool SaveVideo::autosave_wants_pm_metadata()
{
    bool wants_pm_metadata = false;

    if (local_autosave())
    {
        if (streamType != FI_SAVER::STREAM_TYPE::FI_CAP)
        {
            wants_pm_metadata = true;
        }
    }
    return wants_pm_metadata;
}

/*!
 * @brief Determine how many blobs it takes to save this captured partition as a
 * CAP file on the PC.
 *
 * @param num_frames_to_save    - how many frames in the capture are being saved.
 * @param num_blobs_calculated  - how many blobs does it take to save this partition capture.
 * @param space_needed          - how much space needs to be available on the PC drive system.
 *
 * @returns True if the calculation was successful, false if unable to perform the calculation.
 */
bool SaveVideo::calc_blob_count_for_cap(uint32_t num_frames_to_save, uint32_t& num_blobs_calculated, uint64_t& space_needed)
{
    bool        blob_count_calculated   = false;
    uint32_t    total_frames;
    uint32_t    width;
    uint32_t    height;
    uint32_t    bit_depth;

    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Set the references empty.
    num_blobs_calculated = 0;
    space_needed = 0;

    // Sanity checks.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        goto exit_calc_blob_count_for_cap;
    }
    if (my_parent->get_sensor() == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera sensor!";
        goto exit_calc_blob_count_for_cap;
    }

    // We'll need frame information to try to calculate the storage requirement.
    width = my_parent->read_camera_cache(GIGE_REG_WIDTH);
    height = my_parent->read_camera_cache(GIGE_REG_HEIGHT);
    bit_depth = my_parent->read_camera_cache(GIGE_REG_BIT_MODE);

    // How many frames were captured?
    total_frames = my_parent->read_camera_cache(GIGE_REG_PART_NUM_FRAMES);

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "width:" << width << "height:" << height << "bit_depth:" << bit_depth << "total_frames:" << total_frames;
    }

    // Are we doing a partial or full save? That will change the number of "blobs" we will be expecting.
    if (num_frames_to_save == total_frames)
    {
        // Doing a full save, so get the size of the partion.
        uint64_t max_space = ClientMath::qword_to_bytes(my_parent->read_camera_cache(GIGE_REG_PART_SIZE));

        // Determine the amount of space required to save the session.
        // Assumption is that we've read the GIGE_REG_SESSION_SIZE value previously.
        space_needed = ClientMath::qword_to_bytes(CLIENT_MATH::SESSION_INCR_QWORDS) * (uint64_t)my_parent->read_camera_cache(GIGE_REG_SESSION_SIZE);
        if (space_needed > max_space)
            space_needed = max_space;
    }
    else
    {
        uint32_t frame_size;

        // The amount of space needed is the partition reserved size plus the size of the image data.
        space_needed = ClientMath::qword_to_bytes(my_parent->read_camera_cache(GIGE_REG_PART_SIZE_RESERVED));

        // We need the size of the raw frame in image memory, plus the PFMD.
        frame_size = ClientMath::calc_frame_size(width, height, bit_depth) + (uint32_t)my_parent->get_pfmd_len();
        space_needed += (num_frames_to_save * frame_size);
    }

    // How many blobs are we trying to save?
    if (space_needed > (uint64_t)my_parent->blob_size())
    {
        num_blobs_calculated = (uint32_t)(space_needed / my_parent->blob_size());
        if ((space_needed % my_parent->blob_size()) != 0)
            ++num_blobs_calculated;
    }
    else
    {
        // We can pull the entire file in one blob.
        num_blobs_calculated = 1;
    }

    // We were successful in blob count calculation.
    blob_count_calculated = true;

exit_calc_blob_count_for_cap:
    return blob_count_calculated;
}

/*!
 * @brief Cleanup the saving objects, such as the [polling timer](@ref SaveVideo::save_poller) and the [xferDialog](@ref SaveVideo::xferDialog).
 */
void SaveVideo::cleanup_save()
{
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "my_parent:" << (my_parent != nullptr ? "Allocated" : "Empty")
                     << "save_state_poll_timer:" << (save_state_poll_timer != nullptr ? "Allocated" : "Empty")
                     << "xferDialog:" << (xferDialog != nullptr ? "Allocated" : "Empty")
                     << "localSaveToPath:" << (localSaveToPath ? "True" : "False");
    }

    // Sanity check.
    if (my_parent != nullptr)
    {
        // Allow cache refresh again.
        my_parent->pause_cache_refresh(false);

        // Also, restore the heartbeat requirement.
        my_parent->change_heartbeat(0);
    }

    // If we have a timer running, stop it.
    if (save_state_poll_timer != nullptr)
    {
        disconnect(save_state_poll_timer, SIGNAL(timeout()), this, SLOT(save_state_poll_timeout()));
        save_state_poll_timer->stop();
        save_state_poll_timer->deleteLater();
        save_state_poll_timer = nullptr;
    }

    // Just hide any transfer progress dialog.  If autosaving it is reused.
    // If regular save it will be regenerated when another save operation
    // takes place due to the nature of the smart pointer.
    if (xferDialog)
    {
        xferDialog->hideDialog();

        // Were we cancelled?
        if (xferDialog->wasCanceled())
        {
            fiProgressDialog::RecycleProgressDialog(xferDialog);
        }
    }

    // Do we need to clean up the local saver?
    if (localSaveToPath)
    {
        localSaveToPath = false;
    }

    // Restore any JPEG QFactor if necessary.
    RestoreJPEGQFactor();
}

/*!
 * @brief Shutdown the save object.
 */
void SaveVideo::close(const QString& whoCalledMe)
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Caller[" << whoCalledMe << "] Entry - closeInProgress:" << closeInProgress << "closeCompleted:" << closeCompleted << "setUpForSave:" << setupForSave << "savingHasFinished:" << savingHasFinished;

    if (closeInProgress || closeCompleted)
    {
        goto exitSaveVideoClose;
    }

    closeInProgress = true;     // We are attempting to close down.

    // Make sure we cleanup after ourselves.
    cleanup_save();

    // Do we have our image saver thread running?
    StopWorkerThreadIfRunning();

    // Release our image queues.
    ReleaseImageRecvList();
    ReleaseImageFreeList(false, false);

    // Make sure that the stream writer is closed if it isn't already closed.
    if (!streamIsClosed)
    {
        closeStreamWriter();
    }

    closeInProgress = false;
    closeCompleted = true;

exitSaveVideoClose:
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "Caller[" << whoCalledMe << "] Exit - closeInProgress:" << closeInProgress << "closeCompleted:" << closeCompleted << "setUpForSave:" << setupForSave << "savingHasFinished:" << savingHasFinished;
}

/*!
 * @brief After the user has specified all the settings for a save operation,
 *        this method initializes the application and camera to perform the
 *        save operation.
 *
 * @param save_info     The SAVE_VIDEO::SAVE_PARAMS structure.
 * @return True on success, false otherwise.
 */
bool SaveVideo::configure(FI_SAVER::SAVE_PARAMS save_info)
{
    bool            rc          = false;    // Return code set here.
    FI_comm         *comm_ptr;
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    // Reset our own state to prepare for the save.
    ResetSaveVideoState();

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        goto exitConfigure;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        goto exitConfigure;
    }

    // Since the user wants to save a video, let's pause the stream if it's not already paused.
    my_parent->pause_video_request();

    // Configure our parent class
    if (FiSaver::configure(save_info))
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "fiSaver::configure() failed!";
        goto exitConfigure;
    }

    // The fiSaver object will have initialized our comment buffer, so pass it on to the camera.
    comm_ptr->start_write_mem(my_parent->get_ip_addr(), GIGE_REG_COMMENT, *commentBuf);

    // Do we need to update the JPEG quality setting?
    set_jpeg_quality_old_camset((int)comm_ptr->get_reg(my_parent->get_ip_addr(), GIGE_REG_JPEG_QUALITY), __PRETTY_FUNCTION__);

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "jpeg_quality_old_camset:" << jpeg_quality_old_camset;

    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "jpeg_quality_old_camset", (uint32_t)jpeg_quality_old_camset, false);
    FiSaver::verifyJpegQuality(__PRETTY_FUNCTION__, "save_info.jpeg_quality", (uint32_t)save_info.jpeg_quality, false);

    // If we require a certain jpeg quality to stream from the camera we need to tell
    // the camera to stream that jpeg quality.  Keep a copy of the old setting so that
    // it can be restored after the save operation is completed.
    if (save_info.jpeg_quality != jpeg_quality_old_camset)
    {
        // Write out the new quality value.
        comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_JPEG_QUALITY, (uint32_t)save_info.jpeg_quality);
    }
    // NOTE:  The only place where "jpeg_quality_4QT_pixmaps" is actually used is when the frame is
    // sent to fiSaver::SaveToStream and the output of the save operation is a JPEG image
    // after running through a QT image translation. That is when this value is used.
    jpegQuality_4QTPixmaps = save_info.jpeg_quality;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_info.jpeg_quality:" << save_info.jpeg_quality;

    // Are we saving to the workstation or on the camera?
    if (destMedia == SAVE_DLG::MEDIA_LOCAL_PC)
    {
        localSaveToPath = true;  // We are saving to a local path using our fiSaver.

        // Initiate the save to our workstation.
        rc = save_to_pc(save_info);
        if (!rc)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_to_pc() failed!";
            goto exitConfigure;
        }
    }
    else
    {
        localSaveToPath = false; // We are saving to the camera device media using camera server.

        // Have the camera start saving to it's media.
        if (my_parent->is_series3())
        {
            rc = save_to_camera(save_info);
            if (!rc)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_to_camera() failed!";
                goto exitConfigure;
            }
        }
        else
        {
            rc = save_to_camera_json(save_info);
            if (!rc)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "save_to_camera_json() failed!";
                goto exitConfigure;
            }
        }
    }

    // If the save video state is good to move forward create the transfer progress dialog.
    if (rc)
    {
        rc = PrepXferDialogFromSaveParams(save_info, localSaveToPath, __PRETTY_FUNCTION__);

        // Do we have the transfer progress dialog?
        if (!rc)
        {
            // If cannot create the transfer progress dialog something is terribly wrong.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a transfer progress dialog!";
            goto exitConfigure;
        }

        // Start to show progress on the save operation to the user.
        if (xferDialog)
        {
            // Show the dialog with polling set.
            if (FiSaver::isStreamVideo(streamType))
                xferDialog->showDialogWithPolling(GIGE_REG_SAVE_STATE_POLLING_RATE);
            else
                xferDialog->showDialogWithoutPolling();
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no xferDialog ready for progress updates!";
        }
    }
    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Configured.  Expecting " << frameCount << "receive buffers.";
    }

    rc = true;
exitConfigure:

    // Return our result.
    return rc;
}

/*!
 * @brief The copy operation has finished, and we need to clean things up.
 */
void SaveVideo::copy_finished()
{
    //if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "calledByCopyToPC:" << calledByCopyToPC << "copyBlockCount: " << copyBlockCount;
    }

    // Protect access to the fiSaver stream.
    if (streamWriter)
    {
        QMutexLocker fiSaverStreamLocker(&FiSaverStreamMutex);

        // Close the stream writer.
        closeStreamWriter();
    }

    if (blobFile != nullptr)
    {
        blobFile->close();
        blobFile->deleteLater();
        blobFile = nullptr;
    }
}

/*!
 * @brief Configure the save object to receive the image data from the camera.
 * @param copy_info     The structure containing the settings for the copy operation.
 */
bool SaveVideo::copy_to_pc(SAVE_VIDEO::COPY_PARAMS copy_info)
{
    int             src_index;
    bool            rc(true);
    uint32_t        num_frames;
    QString         temp_str;

    //TODO_JLC: FFLR CAPSTK needs copy support.

    QLOG_DEBUG() << __PRETTY_FUNCTION__;

    // If we are batch copying data - the previous copy may have left the worker
    // thread running.  If that is the case close it down so the next copy can
    // take effect.
    if (setupForSave)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "setupForSave is true.  Are we in a batch copy?  Closing to restart a new copy.";
        close(__PRETTY_FUNCTION__);
    }

    // By definition we're saving to the local PC.
    destMedia = SAVE_DLG::MEDIA_LOCAL_PC;
    destLabel.clear();

    // How many frames are we trying to save?
    num_frames = (copy_info.last_frame_num - copy_info.first_frame_num) + 1;

    // Save the start frame so we can use that later.
    copy_start_frame = copy_info.first_frame_num;
    copy_end_frame = copy_info.last_frame_num;

    // We haven't written any copy blocks yet.  Each block may or may not
    // be a full blob size of data so we cannot count on that mechanism to
    // write in the proper place when saving a "blob" image to the stream.
    copyBlockCount = 0;

    // We haven't processed any metafile data.
    copyMetaFileCount = copy_info.metaFileCount;
    copyMetaFilesProcessed = 0;

    // See if we can parse the index from the source file name. It would be nice
    // if we could use that same number for the destination file.
    temp_str = copy_info.src_path;
    src_index = temp_str.lastIndexOf(QLatin1Char('_'));
    if (src_index >= 0)
    {
        QStringView mid = QStringView{temp_str}.mid(src_index + 1, PATH_SEQ_NUM_LEN);
        src_index = mid.toInt();
    }
    else
    {
        src_index = 0;
    }

    // Reset the save video object to prepare for the save.
    ResetSaveVideoState();

    // And reset the stats - we can keep them for this operation.
    reset_stats();

    //if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << "SaveVideo - COPY TO PC Data:";
        QLOG_DEBUG() << "++++++++++++++++++++++++++++++++++++++++";
        QLOG_DEBUG() << "use_camera_name:" << copy_info.use_camera_name;
        QLOG_DEBUG() << "use_existing_name:" << copy_info.use_existing_name;
        QLOG_DEBUG() << "lastCopyRequest:" << copy_info.lastCopyRequest;
        QLOG_DEBUG() << "first_frame_num:" << copy_info.first_frame_num;
        QLOG_DEBUG() << "last_frame_num:" << copy_info.last_frame_num;
        QLOG_DEBUG() << "streamType:" << FiSaver::streamTypeToQString(copy_info.streamType);
        QLOG_DEBUG() << "videoType:" << FiSaver::streamTypeToQString(copy_info.videoType);
        QLOG_DEBUG() << "imageInputType:" << FiSaver::imageTypeToQString(copy_info.imageInputType);
        QLOG_DEBUG() << "imageOutputType:" << FiSaver::imageTypeToQString(copy_info.imageOutputType);
        QLOG_DEBUG() << "dest_path:" << copy_info.dest_path;
        QLOG_DEBUG() << "src_path:" << copy_info.src_path;
        QLOG_DEBUG() << "tag_label:" << copy_info.tag_label;
        QLOG_DEBUG() << "metaFileExt:" << copy_info.metaFileExt;
        QLOG_DEBUG() << "src_index:" << src_index;
        QLOG_DEBUG() << "num_frames:" << num_frames;
        QLOG_DEBUG() << "++++++++++++++++++++++++++++++++++++++++";
    }

    reuse_path = copy_info.use_existing_name;
    if (!reuse_path)
    {
        setPath(copy_info.dest_path);
    }
#if 0
// This seems to be ineffective - in fiSaver::buildFilename it looks for negative sequence number.
    else if (!copy_info.metaFileExt.isEmpty())
    {
        // A reusable path is for metafile copy operations.  The
        // second component set along with use_existing_name is the
        // file extension for the type of metafile being copied.
        //
        // Replace the existing dest_path's file extension with the
        // appropriate metafile one.
        QString metaFilePath = copy_info.dest_path;
        int index = metaFilePath.lastIndexOf(QLatin1Char('.'));
        metaFilePath.truncate(index);
        metaFilePath = "." + copy_info.metaFileExt;
        set_path(metaFilePath);
    }
#endif

    setCount(num_frames);

    // This will tell us how many buffers make up the copy operation.
    saveBufferCount = num_frames;

    // We SAY we're streaming CAP files, but the truth lies in the videoType.
    if (copy_info.streamType == FI_SAVER::STREAM_TYPE::FI_CAP)
    {
        // We're copying a blob. Is it a video or CAP blob?
        // Note:  The ramification here is that this extension is used
        // later in the process when copying the metadata file(s) associated
        // with the original copy video data.  If we don't set the proper
        // extension then when the metadata file is saved it won't be able
        // to replace the file extension with the appropriate metadata file
        // extension.
        setFileExtension(copy_info.videoType);
        copyVideoType = copy_info.videoType;
    }
    else
    {
        // The extension is stored in the save_type field.
        setFileExtension(copy_info.streamType);
        copyVideoType = copy_info.streamType;
    }
    setStreamType(copy_info.streamType, copy_info.imageInputType, copy_info.imageOutputType);
    setClipStart(copy_start_frame);
    setClipEnd(copy_info.last_frame_num);
    setStartIndex((uint32_t)src_index);
    if (copy_info.tag_label.isEmpty())
        setTag(QString());
    else
        setTag(copy_info.tag_label);
    if (copy_info.use_camera_name)
    {
        CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

        // Sanity check.
        if (my_parent == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
            setCameraName(QString());
        }
        else
        {
            setCameraName(QString::fromUtf8(my_parent->get_camera_info()->user_name));
        }
    }
    else
    {
        setCameraName(QString());
    }

    // If streaming video set image type and playback rate.
    if (FiSaver::isStreamVideo(copy_info.streamType))
    {
        // Set the playback rate to the default.
        setPlaybackRate(SAVE_DLG::DEF_AVI_PLAY_RATE);
    }

    // For copy operations progress is managed outside of our realm.  We are
    // therefore not going to be responsible for updating progress internally.
    // If outside caller wants progress updates they should connect the signal
    // SaveVideo::buffer_release() to a slot handler on their object.
    //
    // The reason this approach was chosen is because a copy operation works
    // in blobs depending on how many file(s) selected in the explorer window
    // and each file will require a save video operation.  If we relied on SaveVideo
    // object to internally manage progress the user would see a lot of hides/shows
    // for each file copied. It's a much better user experince if the copy handler
    // in the CamExplore object manages this GUI.  Again CamExplore can simply plug
    // into SaveVideo::buffer_release signal to update its progress information because
    // SaveVideo becomes a stream service in that paradigm.
    if (xferDialog)
    {
        // Recycle any resources and release the transfer progress dialog.
        fiProgressDialog::RecycleProgressDialog(xferDialog);
    }

    // Get our worker thread started and ready to receive data.
    // This is the only place where we tell this object we are "copyToPC" enabled.
    // And this is only called by the CamExplore object at this point in time.
    int err_code;
    if ((err_code = setup(true)) != 0)
    {
        CameraMDI *my_parent = qobject_cast<CameraMDI *>(parent());

        QString msg_str = tr("The setup operation failed. Error: %1 [%2]").arg(strerror(err_code)).arg(err_code);
        QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
        QMessageBox::critical(my_parent, tr("Save Setup Failed"), msg_str);
        rc = false;
    }

    // Return our result.
    return rc;
}

/*!
 * @brief What is the releasedBufferCount.
 */
uint32_t SaveVideo::get_buffer_released_count()
{
    QMutex          bufferReleaseCountMutex;    //!< Mutex to guard access to releasedBufferCount.
    QMutexLocker    bufferReleaseCountLocker(&bufferReleaseCountMutex);
    return releasedBufferCount;
}

/*!
 * @brief Increment releasedBufferCount.
 */
void SaveVideo::inc_buffer_released_count()
{
    QMutex          bufferReleaseCountMutex;    //!< Mutex to guard access to releasedBufferCount.
    QMutexLocker    bufferReleaseCountLocker(&bufferReleaseCountMutex);
    releasedBufferCount++;
}

/*!
 * @brief When all the images have been downloaded, this method flushes
 * out the video stream writer and closes the video file.
 *
 * @note This is only used when saving an AVI or MP4 video file.
 * @note This is run in the image_saver() thread context.
 */
void SaveVideo::finishSave()
{
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "streamType: " << fiSaver::StreamTypeToQString(streamType);

    // Only have extra stuff to do if we're saving video streams or capture BLOBs.
    if (!FiSaver::isStreamVideo(streamType) && !FiSaver::isCaptureStream(streamType))
    {
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "A stack was saved so finish_save is done.";
        }
        return;
    }

    // Keep this scoped in a block.
    {
        // Protect access to the fiSaver stream.
        QMutexLocker fiSaverStreamLocker(&FiSaverStreamMutex);
        // Flush out the video stream writer.
        FiSaver::finishSave();
    }
}

/*!
 * @brief Return the expected sequence number for the first frame or "chunk" of the
 *        copy operation.
 */
uint32_t SaveVideo::get_copy_start()
{
    return copy_start_frame;
}

/*!
 * @brief This is the worker thread for saving the data. Because the process of saving an image to the
 *        disk is very time consuming, we use this worker thread to work in parallel with the main GUI
 *        thread.
 *
 * @param parent    Pointer to the [parent](@ref SaveVideo) object.
 */
void SaveVideo::image_saver(QObject *parent)
{
    std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr;
    SaveVideo   *my_class = qobject_cast<SaveVideo *>(parent);

    // Sanity check.
    if (my_class == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "SaveVideo* parent was Null!";
        return;
    }

    // Make sure we're not supposed to stop.
    while (!my_class->thread_stop)
    {
        // Are there available image buffers and are we not stopping?
        my_class->imageBufferRecvMutex.lock();
        if (my_class->imageBufferRecvList.size() == 0)
        {
            QDeadlineTimer  wait_ms(WAIT_TIMEOUT_MS);

            // Are we done?
            if (my_class->save_is_done)
            {
                // Reset our flag.
                my_class->save_is_done = false;

                if (VERBOSE_SAVE_LOG)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Worker Thread has acknowledged save_is_done flag.";
                }

                // Do any clean-up required.
                my_class->finishSave();

                // Update our statistics.
                my_class->update_stats();

                // Let folks know that we're done. Hopefully the user has
                // made this a queued connection, since we have the mutex locked.
                emit my_class->save_complete();
                my_class->emittedSaveComplete = true;
            }
            else
            {
                // Release the mutex.
                my_class->imageBufferRecvMutex.unlock();

                // Let folks know that the saver is going idle.
                emit my_class->work_list_empty();

                // Re-take the mutex.
                my_class->imageBufferRecvMutex.lock();
            }

            // On return from our wait, the mutex will be locked.
            my_class->worker_wakeup.wait(&my_class->imageBufferRecvMutex, wait_ms);

            // Make sure we're not supposed to stop.
            if (my_class->thread_stop)
            {
                // Release the mutex and bail out of the loop.
                my_class->imageBufferRecvMutex.unlock();
                break;
            }
        }

        // Look through the image buffer receive list for an image buffer to process.
        while (!my_class->imageBufferRecvList.empty())
        {
            if (my_class->thread_stop)
            {
                if (VERBOSE_SAVE_LOG)
                {
                    QLOG_WARN() << __PRETTY_FUNCTION__ << "Still processing imageBufferRecvList but thread_stop is true!";
                }
            }

            // Grab a new image from the pool.
            imageBufferPtr = std::move(my_class->imageBufferRecvList.front());
            my_class->imageBufferRecvList.pop_front();

            // Release access to the list.
            my_class->imageBufferRecvMutex.unlock();

            // Theoretically the imageBufferPtr could be an empty unique_ptr so only send it to the saver if it's valid.
            // Note:  The unique_ptr for the imageBufferPtr should always be valid however it could contain an empty
            // image data pointer, a data length of 0 and a sequence number of zero for the case where a lost frame is
            // in the stream.
            if (imageBufferPtr)
            {
                // Save the image as appropriate.
                if (!my_class->saveToStream(imageBufferPtr->GetMemDataPtr(), imageBufferPtr->GetImageDataLen(), imageBufferPtr->GetSequenceNumber()))
                {
                    if (VERBOSE_SAVE_LOG)
                    {
                        QLOG_ERROR() << __PRETTY_FUNCTION__ << "my_class->SaveToStream reports error.";
                    }
                }
            }
            else
            {
                // This really is an error condition.
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "We received an unexpected empty imageBufferPtr in our imageBufferRecvList!";
            }

            // Lock access to the to-be-freed list.
            my_class->imageBufferFreeMutex.lock();

            // Put the buffer on the image buffer free list only if the image buffer pointer was a valid unique_ptr.
            if (imageBufferPtr != nullptr)
            {
                my_class->imageBufferFreeList.push_back(std::move(imageBufferPtr));
            }

            // Release the to-be-freed list lock.
            my_class->imageBufferFreeMutex.unlock();

            // For really SLOW drives we build up free list entries and never yield time
            // to actually free them!  This was observed when testing Ticket #441 and saving
            // to a USB thumb drive TIFF raw image stacks.  The progress bar was never moving
            // because the free list was never getting time to free itself.
            if (my_class->imageBufferFreeList.size() > 1)
            {
                emit my_class->free_list_ready();
            }

            // Need to set the ready list mutex again, before we loop back around.
            my_class->imageBufferRecvMutex.lock();
        }

        // Release the mutex and go back to the top of the loop.
        my_class->imageBufferRecvMutex.unlock();

        // Are there entries in the to-be-freed list? We don't need to protect
        // this access - it's OK if we send a couple of spurious signals.
        if (!my_class->imageBufferFreeList.empty())
        {
            // Signal we have at least one buffer to free.
            emit my_class->free_list_ready();
        }
    }

    // Tell our class we are exiting.
    emit my_class->worker_thread_exiting();
}

/*!
 * @brief Camera is dearming.
 */
void SaveVideo::dearm_and_keep()
{
    // Is this a FasFireLR save operation?
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    uint32_t    fflr_type;
    bool        fflr_is_enabled = false;
    if (my_parent)
    {
        fflr_is_enabled = my_parent->is_fflr_enabled(fflr_type);
    }

    // Set the active device save state.
    if (active_device_save != DEVICE_SAVE_STATE::SAVE_NONE)
    {
        active_device_save = DEVICE_SAVE_STATE::SAVE_DEARM_AND_KEEP;

        // Create some kind of progress dialog.
        if (fflr_is_enabled && !fflrDearmDlg)
        {
            QString progress_title  = tr("%1 Saving").arg(CameraMDI::fflr_feature_name);
            QString status_msg;
            uint32_t total_count;
            uint32_t dirty_count;
            get_autosave_status_msg(status_msg, total_count, dirty_count);

            fflrDearmDlg = fiProgressDialog::CreateSimpleProgressDialog(progress_title, status_msg, QMessageBox::Cancel, __PRETTY_FUNCTION__);
            connect(fflrDearmDlg.get(), &fiProgressDialog::progressCancelled, my_parent, &CameraMDI::cancel_save, Qt::QueuedConnection);
        }
    }
}

/*!
 * @brief Are we in the process of recovering from a FasFireLR underflow error?
 */
bool SaveVideo::is_fflr_underflow_recovering()
{
    if (fflrRecoverDlg && (fflr_underflow_state != FFLR_UNDERFLOW_STATE::NONE))
    {
        return true;
    }
    return false;
}

/*!
 * @brief This method is called to inform the SaveVideo object that camera is going live.
 */
void SaveVideo::going_live()
{
    // Are we showing a recovery dialog?
    if (fflrRecoverDlg)
    {
        fflrRecoverDlg->hideDialog();
        fiProgressDialog::RecycleProgressDialog(fflrRecoverDlg);
        fflrRecoverDlg = nullptr;
    }

    // Were we in a state of complete progress updates?
    if (fflr_underflow_state == FFLR_UNDERFLOW_STATE::COMPLETE_PROGRESS)
    {
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_PROGRESS_CANCEL;
    }
}

/*!
 * @brief This method is called to inform the SaveVideo object that we've lost a frame.
 *
 * @param seq_num   Not currently used, I thought it might be useful at some future time.
 * @returns True if the lost frame was facilitated, false if something is wrong with writer.
 */
bool SaveVideo::lost_frame(int seq_num)
{
    bool addedLostFrameOk = true;

    QString streamTypeStr = FiSaver::streamTypeToQString(streamType);
    QString imageTypeStr = FiSaver::imageTypeToQString(imageInputType);

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "seq_num: " << seq_num << "streamType:" << streamTypeStr
                     << " imageType: " << imageTypeStr;
    }

    // Losing frames to create placeholder locations is only applicable for stream writers.
    if (FiSaver::isStreamVideo(streamType))
    {
        // Protect access to the fiSaver stream.
        QMutexLocker fiSaverStreamLocker(&FiSaverStreamMutex);

        // Tell the stream writer to put in a place holder frame for the current frame.
        if (streamWriter)
        {
            FAS_VIDEO::STREAM_STATUS writerStatus = streamWriter->AddPlaceholderFrame();

            // There are only two valid stream status states that correlate to a successful
            // operation to add a placeholder frame - a ready state and a completed state.
            if ((writerStatus != FAS_VIDEO::STREAM_STATUS::STREAM_READY) && (writerStatus != FAS_VIDEO::STREAM_STATUS::STREAM_COMPLETED))
            {
                // Pull stream writer status and build up an error message for the user.
                std::string writerStatusStr;
                streamWriter->LogStreamStatus(writerStatusStr, false);   // Pull the status in readable form.
                lastErrMsg = tr("Unable to process lost frame in stream writer. [%1]").arg(writerStatusStr.c_str());

                // We want to log the error condition as well.
                QLOG_ERROR() << __PRETTY_FUNCTION__ << lastErrMsg;

                addedLostFrameOk = false;
            }
        }
    }
    return addedLostFrameOk;
}

/*!
 * @brief When saving to PC or saving using camera (either series3 or seriesNC) this saver object
 * may be called upon during a future SaveDialog initialization.  If we do not persist the save
 * info we can get out of sync errors associated with the initialization of future SaveDialog(s).
 * This function was created to assist in the management of these settings when the system is
 * trying to leverage this SaveVideo object instead of looking at the QSettings "SaveSettings"
 * for default information.
 *
 * NOTE:  Added for reopened Ticket#260
 *
 * @param save_info - the save information provided by the save dialog.
 */
void SaveVideo::persist_save_settings(FI_SAVER::SAVE_PARAMS save_info)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    merge_overlays = save_info.merge_overlays;

    setFileExtension(save_info.streamType);
    setStreamType(save_info.streamType, save_info.imagesInputType, save_info.imagesOutputType);
    setClipStart(save_info.ref_frame + (uint32_t)save_info.clip_start);
    setClipEnd(save_info.ref_frame + (uint32_t)save_info.clip_end);
    setStartIndex(0);

    if (save_info.tag_label.isEmpty())
        setTag(QString());
    else
        setTag(save_info.tag_label);
    if (save_info.use_camera_name && (my_parent != nullptr))
        setCameraName(QString::fromUtf8(my_parent->get_camera_info()->user_name));
    else
        setCameraName(QString());

    // Set the image type if streaming a video.
    if (FiSaver::isStreamVideo(save_info.streamType))
    {
        // Save the playback rate the user wants.
        setPlaybackRate(save_info.playback_rate);

        // If saving the MP4 video set up the saver to match.
        if (save_info.streamType == STREAM_TYPE::MP4)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "save_info.h264_quality:" << save_info.h264_quality;
            setH264Speed(save_info.h264_speed);
            setH264Quality(save_info.h264_quality);
            setH264EnableAllKeyframes(save_info.h264_enable_all_keyframes);
        }
    }

    // Make sure to persist the selected save partition metadata settings.
    //QLOG_INFO() << __PRETTY_FUNCTION__ << "[METADATA CFG][" << save_info.save_metadata_cfg << "]";
    setMetadataCfgSave(save_info.save_metadata_cfg);
}

/*!
 * @brief This method prepares the video stream writer for saving.
 * @return  0 on success, errno on failure.
 */
int SaveVideo::prepareStreamWriter()
{
    int rc;
    CameraMDI *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return EFAULT;
    }

    // We will query for the image width and height no matter the image format.
    streamImageWidth = my_parent->read_camera_cache(GIGE_REG_WIDTH);
    streamImageHeight = my_parent->read_camera_cache(GIGE_REG_HEIGHT);

    {
        // Protect access to the fiSaver stream.
        QMutexLocker fiSaverStreamLocker(&FiSaverStreamMutex);

        // Pass it on to our parent class.
        rc = FiSaver::prepareStreamWriter();
    }

    // All is well.
    return rc;
}

/*!
 * @brief This method prepares the internal transfer progress dialog
 * for use in this class.  It will be utilized for saves not copies as
 * it expects the base class is already configured for the save.
 *
 * @param whoCalledMe - Name of the function that is calling me to create the Transfer Dialog.
 *
 * @return  True if the dialog was properly prepared and is ready for use.
 */
bool SaveVideo::PrepXferDialogFromFSaver(const QString& whoCalledMe)
{
    CameraMDI       *my_parent      = qobject_cast<CameraMDI *>(parent());
    bool            rc              = false;
    bool            savingToLocalPC = (destMedia == SAVE_DLG::MEDIA_LOCAL_PC) ? true : false;
    bool            isSeries3       = true;
    QString         cameraName      = EMPTY_CAMERA_NAME;
    QString         statusStr;

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }
    else
    {
        cameraName = my_parent->get_camera_name();
        isSeries3 = my_parent->is_series3();
    }

    // Create our transfer progress dialog if we do not have one.
    if (!xferDialog)
    {
        xferDialog = fiProgressDialog::CreateXferProgressDialog(FI_PROGRESS::XFER_TYPES::XFER_TYPE_SAVE,
                                                                cameraName,
                                                                whoCalledMe);
    }

    // Do we have the transfer progress dialog?
    if (xferDialog)
    {
        // Set up the source information.
        xferDialog->setXferSrcLabel((streamType == FI_SAVER::STREAM_TYPE::FI_CAP) ? SAVE_SRC_RECEIVE_BLOB : SAVE_SRC_RECEIVE_FRAME);
        xferDialog->showXferSrc(savingToLocalPC);

        // Set up the destination information.
        xferDialog->setXferDestValue(destPath);
        xferDialog->showXferDest(true);

        // Set up the save type information.
        xferDialog->setXferTypeValue(FiSaver::streamTypeToQString(streamType, false));
        xferDialog->showXferType(true);

        // Set up the status icon.
        xferDialog->showStatusIcon(false);
        xferDialog->setStatusIcon(QStyle::SP_MessageBoxInformation);

        // Set up the status line.
        statusStr = tr("Saving %1 %2").arg((FiSaver::isStreamVideo(streamType) ? "Video" : ""), (savingToLocalPC ? "" : "On Camera Device"));

        xferDialog->setStatusValue(statusStr);
        xferDialog->showStatusValue(FiSaver::isStreamVideo(streamType) ? true : false);
        if (savingToLocalPC)
        {
            xferDialog->setBarMinMax((int)clipStart, (int)clipEnd);
            xferDialog->setBarValue((int)clipStart);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Set the min:" << clipStart << "the max:" << clipEnd << "whoCalledMe:" << whoCalledMe;
        }
        else
        {
            // When saving on camera the progress gets updated in SaveVideo::save_state_change.
            // The progress is going to come back in a percentage of complete from 0 to 100.
            xferDialog->setBarMinMax(0, 100);
            xferDialog->setBarValue(0);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Set the max to 100% whoCalledMe:" << whoCalledMe;
        }
        xferDialog->showBar(true);

        // Allow the user to cancel this process and catch the cancelled signal.
        // Note:  Old code had a condition for series3 cameras and saving a
        // capture file.  The save could not be cancelled.  So in that case
        // we do not want the cancel button to be available...
        if (isSeries3 && (streamType == FI_SAVER::STREAM_TYPE::FI_CAP))
        {
            // Series3 CAP file, hides the cancel button.
            xferDialog->showCancel(false);
            disconnect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &SaveVideo::userHasCancelled);
        }
        else
        {
            xferDialog->showCancel(true);
            connect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &SaveVideo::userHasCancelled, Qt::UniqueConnection);
        }

        // If we are saving to our PC a video file then we poll otherwise we just snag the image release.
        if (savingToLocalPC)
        {
            // Are we a video?
            if (FiSaver::isStreamVideo(streamType))
            {
                // And finally we will be showing this dialog with a polling effort where
                // we catch the a poll request for updates from the dialog within our
                // control.  It will end up in our polling slot handler do_xfer_progress_update().
                connect(xferDialog.get(), &fiProgressDialog::progressUpdatePollCycle, this, &SaveVideo::timeForVideoProgressPollUpdate, Qt::UniqueConnection);
            }
            else if (FiSaver::isStackOfImages(streamType) || FiSaver::isCaptureStream(streamType))
            {
                // For stacks of images we assume each buffer release was a frame in the stack.
                connect(this, &SaveVideo::saverBufferReleaseSignal, this, &SaveVideo::timeForStackProgressFrameUpdate, Qt::UniqueConnection);
            }
        }
        // The transfer dialog is ready to go.
        rc = true;
    }
    else
    {
        // If cannot create the transfer progress dialog something is terribly wrong.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a transfer progress dialog!";
    }
    return rc;
}

/*!
 * @brief This method prepares the internal transfer progress dialog
 * for use in this class.  It will be utilized for saves not copies as
 * it expects the FI_SAVER::SAVE_PARAMS for preparation.
 *
 * @param theSaveParmsToUse - how to set up the transfer progress dialog
 * for the save process.
 * @param savingToPC - True if the save operation is local to the PC or false if remote on camera device.
 * @param whoCalledMe - Name of the function that is calling me to create the Transfer Dialog.
 *
 * @return  True if the dialog was properly prepared and is ready for use.
 */
bool SaveVideo::PrepXferDialogFromSaveParams(FI_SAVER::SAVE_PARAMS theSaveParmsToUse, bool savingToPC, const QString& whoCalledMe)
{
    bool            rc = false;
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());
    bool            isSeries3       = true;
    QString         cameraName      = EMPTY_CAMERA_NAME;
    QString         statusStr;

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }
    else
    {
        cameraName = my_parent->get_camera_name();
        isSeries3 = my_parent->is_series3();
    }

    // Assuming all is good set up the save operation's progress dialog.
    // Create our transfer progress dialog.
    xferDialog = fiProgressDialog::CreateXferProgressDialog(FI_PROGRESS::XFER_TYPES::XFER_TYPE_SAVE,
                                                            cameraName,
                                                            whoCalledMe);

    // Do we have the transfer progress dialog?
    if (xferDialog)
    {
        // Set up the source information.
        xferDialog->setXferSrcLabel((theSaveParmsToUse.streamType == FI_SAVER::STREAM_TYPE::FI_CAP) ? SAVE_SRC_RECEIVE_BLOB : SAVE_SRC_RECEIVE_FRAME);
        xferDialog->showXferSrc(savingToPC);

        // Set up the destination information.
        if (!theSaveParmsToUse.dest_path.isEmpty())
        {
            xferDialog->setXferDestValue(theSaveParmsToUse.dest_path);
            xferDialog->showXferDest(true);
        }

        // Set up the save type information.
        xferDialog->setXferTypeValue(FiSaver::streamTypeToQString(theSaveParmsToUse.streamType, false));
        xferDialog->showXferType(true);

        // Set up the status icon.
        xferDialog->showStatusIcon(false);
        xferDialog->setStatusIcon(QStyle::SP_MessageBoxInformation);

        // Setup the progress bar unit type.
        if (savingToPC)
        {
            // Set up the status line.
            statusStr = tr("Saved [0] frames of [%2]").arg(frameCount);

            xferDialog->setBarMinMax(0, (int)frameCount);
            xferDialog->setBarValue(0);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "isSeries3:" << isSeries3 << "savingToPC:" << savingToPC << "whoCalledMe:" << whoCalledMe << "frame_count:" << frameCount;
        }
        else
        {
            // Set up the status line.
            statusStr = tr("Saved [0] frames of [%2]").arg(theSaveParmsToUse.clip_end - theSaveParmsToUse.clip_start + 1);

            // When saving on camera the progress gets updated in SaveVideo::save_state_change.
            // The progress is going to come back in a percentage of complete from 0 to 100.
            xferDialog->setBarMinMax(0, 100);
            xferDialog->setBarValue(0);
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "isSeries3:" << isSeries3 << "savingToPC:" << savingToPC << "whoCalledMe:" << whoCalledMe << "frame_count:" << frameCount << "Bar is 100%";
        }
        xferDialog->showBar(true);

        // Set the status string from set-up above.
        xferDialog->setStatusValue(statusStr);

        // Let's always show this status line we are a video stream (but not if we're saving from a series3 camera).
        xferDialog->showStatusValue((!isSeries3 && FiSaver::isStreamVideo(theSaveParmsToUse.streamType)));

        // Allow the user to cancel this process and catch the cancelled signal.
        // Note:  Old code had a condition for series3 cameras and saving a
        // capture file.  The save could not be cancelled.  So in that case
        // we do not want the cancel button to be available...
        if (isSeries3 && (streamType == FI_SAVER::STREAM_TYPE::FI_CAP))
        {
            // Series3 CAP file, hides the cancel button.
            xferDialog->showCancel(false);
            disconnect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &SaveVideo::userHasCancelled);
        }
        else
        {
            xferDialog->showCancel(true);
            connect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &SaveVideo::userHasCancelled, Qt::UniqueConnection);
        }

        // If we are saving to our PC a video file then we poll otherwise we just snag the image release.
        if (savingToPC)
        {
            // Are we a video?
            if (FiSaver::isStreamVideo(theSaveParmsToUse.streamType))
            {
                // And finally we will be showing this dialog with a polling effort where
                // we catch the a poll request for updates from the dialog within our
                // control.  It will end up in our polling slot handler do_xfer_progress_update().
                connect(xferDialog.get(), &fiProgressDialog::progressUpdatePollCycle, this, &SaveVideo::timeForVideoProgressPollUpdate, Qt::UniqueConnection);
            }
            else if (FiSaver::isStackOfImages(theSaveParmsToUse.streamType) || FiSaver::isCaptureStream(theSaveParmsToUse.streamType))
            {
                // For stacks of images we assume each buffer release was a frame in the stack.
                connect(this, &SaveVideo::saverBufferReleaseSignal, this, &SaveVideo::timeForStackProgressFrameUpdate, Qt::UniqueConnection);
            }
        }
        // The transfer dialog is ready to go.
        rc = true;
    }
    else
    {
        // If cannot create the transfer progress dialog something is terribly wrong.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to allocate a transfer progress dialog!";
    }
    return rc;
}

bool SaveVideo::queryFreePartition()
{
    QString     operation_label = tr("Autosave operations");
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // Do we have a partition sitting unsaved in memory?
    // This call returns false if the partition cannot be overwritten, otherwise
    // it will return true.  We can just use its return value for our caller.
    return (my_parent->verify_dirty_overwrite(true, operation_label));
}

/*!
 * @brief Return information that can be used for status updates on video stream
 * save operation by filling in the data for each reference variables passed to the function.
 *
 * @param strMsg [output] provides a place to store additional inforamation for caller about stream state.
 * @param frameCount [output] holds the understood framecount for this stream.
 * @param framesAdded [output] how many frames have been processed by the stream.
 * @param framesEncoded[output] how many frames have been encoded (if video) if not it should equal framesAdded.
 * @param saverIsOk [output] what is the state of this saver ok or error condition.
 * @param saverIsComplete[output] is the state of this saver complete?
 *
 * @returns true if the stream is a video, false if not which means a stack of frames.
 */
bool SaveVideo::getSaveStatus(QString& strMsg, uint32_t& frameCount, uint32_t& framesAdded,
                              uint32_t& framesEncoded, bool& saverIsOk, bool& saverIsComplete)
{
    frameCount = 0;
    framesAdded = 0;
    framesEncoded = 0;
    saverIsOk = false;
    saverIsComplete = false;

    bool isVideoStream = isStreamVideo(streamType);     // Save our stream video state.

    // Was this objectset up to save?
    if (!setupForSave)
    {
        strMsg = QString::fromUtf8("SaveVideo was not setup for save operation.");
        return false;
    }

    // I want to improve upon this to make it robust enough to determine what is the actual
    // saver status.  For the moment this is a simple test to see if we are video stream or
    // not.  If we are a video stream then call the interface to get video status update if
    // we are not just look for now at a couple of things and put any type of message together.

    // For video streams we will use the existing status query.
    if (isVideoStream)
    {
        return getVideoStreamSaveStatus(strMsg, frameCount, framesAdded, framesEncoded, saverIsOk, saverIsComplete);
    }

    // Falling through to here means that it is not a video stream.

    // What is the status of this object if it isn't saving a video?
    // The value stored in FiSaver::stream_save_buffer_count is only incremented at the heart of the fiSaver object
    // when an image is written out for a stack and we can rely on its accuracy.

    frameCount      = saveBufferCount;
    framesAdded     = streamSaveBufferCount;
    framesEncoded   = streamSaveBufferCount;

    // If there was an error condition on the last save request note the condition.
    saverIsOk = !lastSaveHadError;
    strMsg = QString::fromUtf8("SaveVideo Status: buffers[expected = %1][received=%2][saved=%3][error=%4].").arg(frameCount).arg(framesAdded).arg(framesEncoded).arg(lastErrMsg);

    if (saverIsOk)
    {
        if (framesEncoded == frameCount)
        {
            saverIsComplete = true; // We should be complete with this simple test.
        }
    }

    return isVideoStream;
}

/*!
 * @brief Return information that can be used for status updates on video stream
 * save operation by filling in the data for each reference variables passed to the function.
 *
 * @param strMsg [output] provides a place to store additional inforamation for caller about stream state.
 * @param frameCount [output] holds the understood framecount for this stream.
 * @param framesAdded [output] how many frames have been processed by the stream.
 * @param framesEncoded[output] how many frames have been encoded (if video) if not it should equal framesAdded.
 * @param streamIsOk [output] what is the state of this stream ok or error condition.
 * @param streamIsComplete[output] is the state of this stream completed?
 *
 * @returns true if the stream is a video, false if not which means a stack of frames.
 */
bool SaveVideo::getVideoStreamSaveStatus(QString& strMsg, uint32_t& frameCount, uint32_t& framesAdded,
                                         uint32_t& framesEncoded, bool& streamIsOk, bool& streamIsComplete)
{
    (void)frameCount;
    framesAdded = frameCount;
    framesEncoded = last_encoded_count;
    bool isClosed = streamIsClosed;                     // Catch a stream is closing state.
    bool isVideoStream = isStreamVideo(streamType);     // Save our stream video state.

    // If copying a stream a call here makes no sense.  Pop up a warning in debug mode.
    if (!isVideoStream)
    {
        framesEncoded = get_buffer_released_count();
        streamIsOk = false;         // We are not a video stream this makes no sense here.
        streamIsComplete = false;   // We are not a video stream this makes no sense here.
        strMsg = QString::fromUtf8("NonVideo Save Status:  frame_count [%1] releasedBufferCount [%2].").arg(frameCount).arg(framesEncoded);
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << strMsg;
        }
        return false;       // We can't query the stream writer here.
    }

    if (VERBOSE_SAVE_LOG)
    {
        if (isClosed)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Streamer is closed!";
        }

        if (!retryInProgress && (framesEncoded >= frameCount))
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Stream should be complete.";
        }
    }

    streamIsOk = true;                  // Assume stream is ok.
    streamIsComplete = streamIsClosed;  // Let's assume it's complete if closed.

    // Refresh the video stream status if not in middle of closing shop.
    if (!isClosed)
    {
        isVideoStream = FiSaver::getVideoStreamSaveStatus(strMsg, frameCount, framesAdded, framesEncoded, streamIsOk, streamIsComplete);
    }

    // Note:  wait_for_save_complete was created to make time for encoders to
    // complete their task. It is set to false here:  SaveVideo::ResetSaveVideoState()
    // and is only set to true here: void SaveVideo::saving_finished() when the stream
    // is a video and the operation is not a copy in progress.  The process of identifying
    // when a stream writer has consumed the frames (the read is complete) and has
    // encoded all of the frames read (the encoder is complete) is in need of refactor
    // to tighten up and encapsulate better in the actual FiSaver object.
    // I am dropping this note here because this is the location that determines
    // when the stream writer has done its job and I don't think this is the best
    // place to catch that state as it currently is dependent on polling to determine
    // and would be much better served with a signal/slot interface.

    // Is the stream completly encoded?
    if (framesEncoded >= frameCount)
    {
        // We are still polling so is the stream good and is it completed?
        if (!isClosed && streamIsOk && streamIsComplete)
        {
            // We should be in a "wait_for_save_complete" state - if we are not then the call to
            // SaveVideo::saving_finished() did not occur.  Let's make that call now.
            if (!wait_for_save_complete)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Video is encoded but wait_for_save_complete was not set.  framesEncoded: " << framesEncoded << " frameCount: " << frameCount << "fiSaver::frame_count:" << frameCount << "retry_in_progress:" << retryInProgress;
                saving_finished(__PRETTY_FUNCTION__);
            }
        }
    }

    // What is our state of completion?
    if (wait_for_save_complete)
    {
        bool stopWaiting = false;

        // Is the stream ok?
        if (!streamIsOk)
        {
            // Something is wrong with the stream.
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Stream Writer Status is not OK!";
            stopWaiting = true;
        }

        // Is the framesEncoded counter moving or stagnant?
        if (last_encoded_count == framesEncoded)
        {
            float secondsBetweenPolls = (float)GIGE_REG_SAVE_STATE_POLLING_RATE / 1000.0;
            // Set up a warning threshold of the number of polling periods/second * 10 seconds.
            uint32_t warnThreshold = ((1000 / GIGE_REG_SAVE_STATE_POLLING_RATE) * 10);
            same_encoded_count++;

            // Typically the save progress is polled every SAVE_POLLING_PERIOD milliseconds.  If we are not
            // able to discern progression on the encoded frame count pop out a warning to the log file
            // so it can be noted upon a review.
            if (same_encoded_count >=  warnThreshold)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "Stream Writer has not increased frames encoded count:"
                            << framesEncoded << "in over " << secondsBetweenPolls << "seconds.";
            }

            // Note:  The user will still be able to cancel out of the progress
            // dialog if nothing is happening and a warning will have output if
            // we get stuck in some sort of stalemate.  Which should not ever
            // happen but you never know it could happen with ffmpeg encoder.
        }
        else
        {
            last_encoded_count = framesEncoded;
            same_encoded_count = 0;     // Reset we have made progress.
        }

        // Is the stream completly encoded?
        if (framesEncoded >= frameCount)
        {
            if (VERBOSE_SAVE_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "stopWaiting CASE.  framesEncoded: " << framesEncoded << " frameCount: " << frameCount << "frame_count:" << frameCount << "retry_in_progress:" << retryInProgress;
            }
            stopWaiting = true;
        }

        // Output a message about the state of thread_told_video_complete.
        if (VERBOSE_SAVE_LOG && stopWaiting)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "stopWaiting[true] streamIsClosed[" << streamIsClosed << "] thread_told_video_complete[" << thread_told_video_complete << "]";
        }

        // If a condition arose to stop waiting for a completion because of error or otherwise
        // then terminate the worker thread and complete the mission. But only do it one time.
        if (stopWaiting && !thread_told_video_complete)
        {
            // Transition to the save_is_done and wake up the worker thread.
            save_is_done = true;
            if (VERBOSE_SAVE_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Stream Writer Is Done Encoding Waking Up Worker Thread! The streamIsClosed:" << streamIsClosed;
            }

            // Make sure our worker thread knows it's time to close the stream.
            worker_wakeup.wakeAll();

            // Keep track that we had told our worker thread that our video was complete.
            thread_told_video_complete = true;
        }
    }

    return isVideoStream;
}

/*!
 * @brief Called when the streamer receives an image, we put the image onto the "queue" (ready_list)
 *        to be proceses by the [worker thread](@ref SaveVideo::image_saver).
 *
 * @param imageBufferPtr     The image buffer we take ownership of if we can receive it.
 *
 * @returns True on success, false otherwise.
 */
bool SaveVideo::receive_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr)
{
    bool unableToProcessImage   = false;
    bool receivedImageOk        = false;
    int absFrameNo              = 0;

    // If user cancelled save process don't process the image release it back.
    if (cancel_in_progress)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The user has cancelled the saver operation will release image back to buffer pool.";
        unableToProcessImage = true;
    }

    // Make sure the thread is running.
    if (!image_thread.isRunning())
    {
        unableToProcessImage = true;
        // Report a little more information to the user.
        if (imageBufferPtr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Received frame:" << ((int)getClipStart() + (int)imageBufferPtr->GetSequenceNumber() - 1)
                        << "but worker thread is not running.  Releasing image back to buffer pool.";
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Received null imageBufferPtr and worker thread is not running.  No image to release back to buffer pool.";
        }
    }

    // In the case of a "missing" frame the caller is expected to provide an empty
    // std::unique_ptr<FI_MEMPOOL::FiImageBuffer> to this method.  By "empty" it means
    // a constructed FiImageBuffer smart pointer but no attached memory pool resource and a
    // sequence number of zero.  This will be the indication that the streamer is pushing
    // a missing frame through the stream.  If we ever get a true null pointer or an
    // empty std::unique_ptr then it means something is wrong with the execution of FasMotion.
    // Therefore we will mark the frame in error and output an error message.
    if (!imageBufferPtr)
    {
        unableToProcessImage = true;
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "imageBufferPtr is nullptr - expected std::unique_ptr<FI_MEMPOOL::FiImageBuffer>!";
    }

    // Make sure we're not in the middle of stopping the thread.
    if (thread_stop)
    {
        unableToProcessImage = true;
        // Report a little more information to the user.
        if (imageBufferPtr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Received frame:" << ((int)getClipStart() + (int)imageBufferPtr->GetSequenceNumber() - 1)
                         << "but unable to process the frame as worker thread reports thread_stop condition.  Releasing image back to buffer pool.";
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Received null imageBufferPtr and worker thread reports thread_stop condition.  Releasing image back to buffer pool.";
        }
    }

    // If unable to process image just release it back.
    if (unableToProcessImage)
    {
        if (imageBufferPtr)
        {
            // Let the saver object take care of releasing this buffer.
            release_image(std::move(imageBufferPtr));
        }

        // Force a cancel operation as it makes no sense to continue to
        // receive frames when there is no worker thread!
        if (!cancel_in_progress)
        {
            // User is trying to cancel the autosave.
            QString whoCalled = __PRETTY_FUNCTION__;
            whoCalled += "No Worker Thread?";
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Forcing SaveVideo::cancel_save()";
            cancel_save(whoCalled);
        }
    }
    // Otherwise take ownership by placing it into our ready list.
    else
    {
        int imageSeqNbr = (int) imageBufferPtr->GetSequenceNumber();

        // Are we processing a valid frame?  Remember missing frames have imageSeqNbr of 0 and
        // metafile data has an imageSeqNbr < 0.  Our frame_count sometimes includes metafiles.
        // TO_TEST:  Verify missing frames and operations that include metafile data.
        if (imageSeqNbr != 0)
        {
            download_receive_frame_count++;
#if LOG_VERBOSE_INFO
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "download_receive_frame_count:" << download_receive_frame_count << "frameCount:" << frameCount;
#endif
            // Have we received all of the expected valid frames for the stream?
            if (download_receive_frame_count == frameCount)
            {
#if LOG_VERBOSE_INFO
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "download_receive_frame_count:" << download_receive_frame_count << "frameCount:" << frameCount;
#endif

                // Don't do this if copy to pc invoked as there is another mechanism for this
                // being managed in the CamExplore object for the moment.
                if (!copy_to_pc_invoked())
                {
                    // Let anyone connected know that the number of non-empty buffers received is equal
                    // to the frame_count that was expected.
                    emit save_receive_count_complete();
                    emittedSaveReceiveCountComplete = true;
                }
            }
        }

        // Lock access to the ready list.
        QMutexLocker    ready_list_lock(&imageBufferRecvMutex);

        absFrameNo = (int)getClipStart() + imageSeqNbr - 1;

        // Append the image to the ready list.
        imageBufferRecvList.push_back(std::move(imageBufferPtr));

        // Make sure our worker thread knows there's data.
        worker_wakeup.wakeAll();

        // We received the image.
        receivedImageOk = true;
    }

    // Update status about the frame received value.
    if (xferDialog && receivedImageOk)
    {
        int relFrameNo = absFrameNo - (int)trigFrameNo4Save;

        // When we receive a buffer update the frame value of the image received.
        xferDialog->setXferSrcLabel((streamType == FI_SAVER::STREAM_TYPE::FI_CAP) ? SAVE_SRC_RECEIVE_BLOB : SAVE_SRC_RECEIVE_FRAME);
        xferDialog->setXferSrcValue(QString::number(relFrameNo));
    }

    // All done.
    return receivedImageOk;
}

/*!
 * @brief Release a consumed image buffer back to the free pool.
 * @param image_ptr     The image buffer we're done with.
 */
void SaveVideo::release_image(std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // We want to keep track of a count of released buffers in the system.
    inc_buffer_released_count();

    // Is this a valid image buffer?  It could be an empty one in which was
    // used as a placeholder for lost frames.  If so we get rid of it here.
    if (imageBufferPtr)
    {
        // Did this image buffer have a valid image buffer?
        uint32_t imageDataLen = imageBufferPtr->GetImageDataLen();
        const uint8_t *imageDataPtr = imageBufferPtr->GetMemDataPtr();
        if (imageBufferPtr->GetSequenceNumber() > 0)
        {
            if (imageDataPtr != nullptr)
            {
                // This is an image resource that needs to be recycled.
                my_parent->release_buffer(std::move(imageBufferPtr));
            }
        }
        else
        {
            // This is an empty image buffer that can be destructed here.
            imageBufferPtr->PrepareToDestruct();
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "SeqNbr=0.  Placeholder Frame Scenario for Missed Frame [imageDataLen:" <<
                         imageDataLen << "][imageDataPtr:" << (imageDataPtr == nullptr ? "EMPTY" : "HAS DATA")
                         << "][releasedBufferCount:" << get_buffer_released_count() << "]";
        }
    }
}

/*!
 * @brief Releases all images on the free list back to the
 * stream's image buffer pool.
 *
 * @returns The result is true if all was handled, false if error.
 */
bool SaveVideo::ReleaseImageFreeList(bool stopWorkerThread, bool emitBufferRelease)

{
    bool imagesReleased = false;
    uint32_t nReleased = 0;

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Releasing Images In To Be Freed Queue."
                     << "stopWorkerThread:" << stopWorkerThread
                     << "emitBufferRelease:" << emitBufferRelease;
    }

    if (stopWorkerThread)
        StopWorkerThreadIfRunning();

    // Release all the entries in the image buffer free list.
    {
        QMutexLocker imageBufferLocker(&imageBufferFreeMutex);
        while (!imageBufferFreeList.empty())
        {
            std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr =
                std::move(imageBufferFreeList.front());
            imageBufferFreeList.pop_front();
            if (imageBufferPtr)
            {
                if (emitBufferRelease)
                {
                    uint32_t imageSeqNbr = imageBufferPtr->GetSequenceNumber();
                    emit saverBufferReleaseSignal(imageSeqNbr);
                }
                release_image(std::move(imageBufferPtr));
                nReleased++;
            }
        }
        imagesReleased = true;
    }

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Number of Free Images Released:" << nReleased;
    }

    return imagesReleased;
}

/*!
 * @brief Releases all images on the receive list back to the
 * stream's image buffer pool.
 *
 * @returns The result is true if all was handled, false if error.
 */
bool SaveVideo::ReleaseImageRecvList()
{
    bool imagesReleased = false;
    uint32_t nReleased  = 0;

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Releasing the Images in Receive Queue.";
    }

    StopWorkerThreadIfRunning();

    // Release all the entries in the image buffer received list.
    {
        QMutexLocker imageBufferLocker(&imageBufferRecvMutex);
        while (!imageBufferRecvList.empty())
        {
            std::unique_ptr<FI_MEMPOOL::FiImageBuffer> imageBufferPtr =
                std::move(imageBufferRecvList.front());
            imageBufferRecvList.pop_front();
            if (imageBufferPtr)
            {
                release_image(std::move(imageBufferPtr));
                nReleased++;
            }
        }
        imagesReleased = true;
    }

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Number of Received Images Released:" << nReleased;
    }

    return imagesReleased;
}

/*!
 * @brief Called to collectively reset all of our state variables to prepare for
 * the next save operation.
 */
void SaveVideo::ResetSaveVideoState()
{
    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Resetting the SaveVideo state.";
    }

    save_is_done = false;
    wait_for_save_complete = false;
    last_encoded_count = same_encoded_count = 0;
    thread_told_video_complete = false;
    streamSaveBufferCount = 0;
    reuse_path = false;
    retryInProgress = false;
    releasedBufferCount = 0;
    copyBlockCount = 0;
    trigFrameNo4Save = 0;
    error_is_logged = false;

    // Prepares for the next save operation.
    active_device_save = SAVE_VIDEO::DEVICE_SAVE_STATE::SAVE_NONE;

    // Reset FasFireLR underflow condition.
    fflr_underflow_state = FFLR_UNDERFLOW_STATE::NONE;
    fflr_underflow_wait_delays = 0;
    fflr_underflow_dirty_count = 0;
    write_save_json_requested = false;
}

/*!
 * @brief Called to collectively reset all of our state variables to prepare for
 * the next save operation.
 */
void SaveVideo::RestoreJPEGQFactor()
{
    // Part of Ticket# 258 and 66 bug-fix.  This had only been in the base class fiSaver
    // but we don't use fiSaver outside of local path saves so saving to devices was
    // missing this restoration.  Should be called when saver closes.

    // Check to see if we modified the old jpeg quality setting on the camera.
    if (jpeg_quality_old_camset != -1)
    {
        uint32_t        cam_ip_addr;
        CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

        // Sanity checks.
        if (my_parent == nullptr)
        {
            // This could happen on shutdown.
            QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
            return;
        }

        // Get our communications object pointer.
        FI_comm *comm_ptr = MainWindow::GetInstance()->get_comm();
        if (comm_ptr == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
            return;
        }

        // Get our camera's IP address.
        cam_ip_addr = my_parent->get_ip_addr();

        // Does the camera match the old setting or was it changed?
        int jpeg_quality_cur_camset = (int)comm_ptr->get_reg(cam_ip_addr, GIGE_REG_JPEG_QUALITY);
        {
            // If there is a difference restore the old camera setting.
            if (jpeg_quality_cur_camset != jpeg_quality_old_camset)
            {
                //if (VERBOSE_SAVE_LOG)
                {
                    QLOG_DEBUG() << __PRETTY_FUNCTION__
                                 << "Resetting jpeg_quality_cur_camset:" << jpeg_quality_cur_camset
                                 << "To jpeg_quality_old_camset:" << jpeg_quality_old_camset;
                }

                // Restore the original JPEG quality on the camera.
                comm_ptr->start_write_reg(cam_ip_addr, GIGE_REG_JPEG_QUALITY, (uint32_t)jpeg_quality_old_camset);
            }

            // We are done with this jpeg setting reset it.
            jpeg_quality_old_camset = -1;
        }
    }
}

/*!
 * @brief Called to reset the statistics associated with a save operation.
 */
void SaveVideo::reset_stats()
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Reset our save video object as well.
    ResetSaveVideoState();

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
    }

    // Reset our statistics.
    download_byte_count = 0;
    download_frame_count = 0;
    download_receive_frame_count = 0;

    // Handle the reset to the socket streamer if we have a connected camera and
    // communications object.
    if (my_parent && comm_ptr)
    {
        FSPSocket *fspSocketStreamer = comm_ptr->get_streamer(my_parent->get_ip_addr());
        if (fspSocketStreamer == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "FSPSocket pointer was NULL from FI_comm pointer!";
        }
        else
        {
            fspSocketStreamer->reset_stats();
        }
    }

    // Start our elapsed timer.
    download_duration.start();

    // Remember the start time.
    download_start = QDateTime::currentDateTime();
}

/*!
 * @brief Called by the worker thread to save the image to either a file or the stream.
 *
 * Note:  Make sure the input format fiSaver::imageInputType of this image data matches
 * the format for encoder fiSaver::imageTypeForEncoder.
 *
 * !!!REMEMBER!!! - THIS IS ONLY CALLED BY THE WORKER THREAD PROCESSING INPUT FRAMES!!!
 *
 * This should have been setup by the stream feeder (the caller) when it set up this
 * object using the interface fiSaver::SetStreamType.
 *
 * @param imageDataPtrToWrite       The read-only buffer of image data to write to the stream.
 * @param imageDataLenToWrite       The number of bytes in the image data to write.
 * @param imageDataSeqNbr           The sequence number for this image in the stream.
 *
 * @return True on success, false otherwise.
 */
bool SaveVideo::saveToStream(const uint8_t *imageDataPtrToWrite, uint32_t imageDataLenToWrite, uint32_t imageDataSeqNbr)
{
    bool        lockedFiStreamMutex = false;
    bool        retryingFrames      = false;
    bool        savedOK             = false;
    bool        saveToStreamCalled  = false;    // Want to know if we called this function.
    CameraMDI   *my_parent = nullptr;

    // If in the middle of canceling the save or if we have an error, just leave.
    if (cancel_in_progress || !lastErrMsg.isEmpty())
        goto exitSaveToStream;

    // If our pointer is null, then we have a missing frame.
    if (imageDataPtrToWrite == nullptr)
    {
        // Reserve space for the frame.
        savedOK = lost_frame(0);

        // We want to process any lost frame failures as if we would normally process failed frame writes.
        goto checkSavedOKLabel;
    }

    // Get the camera object we're associated with.
    my_parent = qobject_cast<CameraMDI *>(parent());
    if (my_parent == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }
    else
    {
        retryingFrames = my_parent->retrying_frames();
    }

    // Note:  The mutex lock below is very important to control access to this save to stream function.
    // The only caller of this function (currently) is the worker thread processing received image buffers.
    // Without this mutex lock, access to the saver object is not mitigated in a multi-threaded environment.
    // What I have observed on Linux is that image buffers come to this process out of sequence order and
    // without this in lock place data are written out of order when copying videos and cap files from the
    // Storage/Explore (camera explorer) dialog.  This was most obvious on a Linux version of FasMotion
    // by copying 2 AVI/BMP files saved to an SSD device into a local path.  The 2nd copied AVI file ended
    // up with corrupted frames. It will manifest in any type of multiple file copy scenario and that is
    // a very easy test case.  This lock should not be removed in future modifications to this file.

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "image_seq_nbr:" << imageDataSeqNbr << "calledByCopyToPC?:" << calledByCopyToPC;

    // Protect access to our stream writer always.
    FiSaverStreamMutex.lock();
    lockedFiStreamMutex = true;

    // Special logic here when copying data blobs to set up a sequence number.
    if (isCaptureStream(streamType) && calledByCopyToPC)
    {
        // Increase the blocks copied by one so that it creates a one-based
        // sequence number that will propagate through the stream save process.
        copyBlockCount++;

        // Note:  Series3 cameras do not save real CAP files this way.  They do however probably
        // save other things as "BLOBS" using the storage/explore dialog like video files
        //
        // I do however want to revisit the question of a retry scenario in a
        // copy condition - does it ever happen?  In revisiting this and testing
        // this using the copy version of #MISSING_FRAME the retry occurs.

        // What is the sequence number provided?
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "imageDataSeqNbr Provided:" << imageDataSeqNbr << "copyBlockCount:" << copyBlockCount;
        }
        // NOTE:  I commented this out because the stream service seems to be providing the
        // proper sequence number and I don't know why I was assuming the data in the queue
        // would correspond to the copyBlockCount - 2/18/21
        //
        // This is just flat out wrong.  I am leaving note in place for posterity.  This routine
        // does get presented with images out of sequence and therefore *must* rely on the fsp socket's
        // sequence number provided in the memory pool buffer.  We have no guarantee images pulled out
        // of the receive cache are in sequence number order.  2/19/21
        //
        // If this is not a metadata file then we need to modify the sequence number
        // to be the copy block count (one-based) in order to have the write logic
        // in the stream saver position a blob that is more than a blob-size.
        // What was happenening in the past was that this sequence number provided
        // ended up being whatever the filename indicates it is for video copies and
        // therefore we end up writing in the wrong places.  We are creating a pseudo
        // sequence number for copy video/capture cases.
        // NOTE:  I am still working out whether or not this needs to be in here for Blob AVI/MP4 video copies.
        //if (IsStreamVideo(copyVideoType) && !metadata_seq_num(imageDataSeqNbr))
        //{
        //    imageDataSeqNbr = copyBlockCount;
        //}
        // There is still a bug in the system when only one blob is copied over and the original
        // video file had a sequence number greater than 0.
        // The criteria is a video type, only 1 blob-frame, and not a metadata sequence number.
        if (isStreamVideo(copyVideoType) && ((copy_end_frame - copy_start_frame) == 0) && (!metadataSeqNum(imageDataSeqNbr)))
        {
            // ONLY one blob is being copied.  If the sequence number is > 1 there is
            // an issue!
            if (imageDataSeqNbr > 1)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[!!OVERRIDE imageDataSeqNbr!!]  Only one BLOB copied and sequence number must be 1.  imageDataSeqNbr Provided:" << imageDataSeqNbr
                             << "copy_start_frame:" << copy_start_frame
                             << "copy_end_frame:" << copy_end_frame
                             << "The source filename must have a nonzero sequence number.";
                imageDataSeqNbr = 1;    // Override!!
            }
        }
    }

    // Do the actual save.
    if (streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
    {
        // For series3 cameras, we convert from using raw TIFF to using packed images.
        if (my_parent->is_series3())
        {
            QImage myImage;
            if (!myImage.loadFromData(imageDataPtrToWrite, imageDataLenToWrite, "TIF"))
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Unable to load the data into a QImage." << "imageDataSeqNbr:" << imageDataSeqNbr;
                savedOK = false;
            }
            else
            {
                savedOK = FiSaver::saveToPackedStream(myImage.constBits(), myImage.sizeInBytes(), imageDataSeqNbr, retryingFrames);
            }
        }
        else
        {
            savedOK = FiSaver::saveToPackedStream(imageDataPtrToWrite, imageDataLenToWrite, imageDataSeqNbr, retryingFrames);
        }
    }
    else
    {
        savedOK = FiSaver::saveToStream(imageDataPtrToWrite, imageDataLenToWrite, imageDataSeqNbr, retryingFrames);
    }

    saveToStreamCalled = true;      // We did call this fiSaver::SaveToStream.

// Note:  Can get here for lost frames as well.
checkSavedOKLabel:
    // Check for an error during the save.
    if (!savedOK)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Failed to process frame in stream writer.";

        // Need to exit here without regard for last_err_msg being empty.
        if (lastErrMsg.isEmpty())
        {
            lastErrMsg = "Saver failed to write frame and the last_err_msg is empty!";
        }
    }

    // Let's accumulate any buffers that have size in our download calculations without regard
    // to the save status.
    if (imageDataLenToWrite > 0)
    {
        // Accumulate the bytes written.
        download_byte_count += imageDataLenToWrite;

        // Increment the number of received frames.
        download_frame_count++;
    }
exitSaveToStream:

    if (saveToStreamCalled && !savedOK)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "saveToStreamCalled and !savedOK.  Asking camera to pause the stream and will emit save_err() signal.";

#if 0
        // Ticket #271 Tightening-Up error handling.
        // Make sure to pause any stream that maybe ongoing.  We want to limit the number
        // of threads operating once a save error occurs.
        {
            CameraMDI *my_parent = qobject_cast<CameraMDI *>(parent());

            // Sanity check.
            if (my_parent != nullptr)
            {
                my_parent->pause_video_request();
            }
        }
#endif
        // Close the stream writer now this stream writer is terminated.
        closeStreamWriter();

        // We have emitted the save_error().
        emittedSaveError = true;

        // Let folks know there was an error.
        emit save_error();
    }

    // Release the mutex lock, if we created one.
    if (lockedFiStreamMutex)
    {
        FiSaverStreamMutex.unlock();
    }

    return savedOK;
}

/*!
 * @brief This is the entry into the save process. We put up a [save dialog](@ref SaveDialog)
 *        to get parameters from the user.
 *
 * This is called as result of File/Save menu item action which is handled by the
 * CameraMDI::do_camera_save function call.
 *
 * @return True on success, false on failure.
 */
bool SaveVideo::save_recording()
{
    bool            rc;
    SAVE_PARAMS     save_info;
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // We are not doing autosave.
    autosave_enabled = false;
    autosave_configured = false;

    // Create the save dialog.
    SaveDialog      save_dlg(my_parent->get_ip_addr(), my_parent);

    // Initialize the save parameters.
    save_dlg.init_controls();
    save_dlg.set_autosave_mode(false);
    save_dlg.set_clip_range(my_parent->get_clip_start(), my_parent->get_clip_end());

    // Are there overlays?
    if (my_parent->get_player())
        save_dlg.enable_overlay_merge(my_parent->get_player()->has_overlay());

    // Get the user's input on what/how they want the video saved.
    if (save_dlg.exec() == QDialog::Rejected)
        return false;

    // Retrieve the settings.
    save_dlg.get_save_params(save_info, my_parent->is_series3());

    // Are we saving to the workstation or on the camera?
    if (save_info.dest_media == SAVE_DLG::MEDIA_LOCAL_PC)
    {
        // Do we have a path?
        if (save_info.dest_path.isEmpty())
        {
            QString msg_str = tr("The download path must not be empty!");
            QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
            QMessageBox::critical(my_parent, tr("Configuration Error"), msg_str);
            return false;
        }
        else
        {
            // Make sure the path exists.
            QDir download_dir(save_info.dest_path);
            if (!download_dir.exists())
            {
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "Path doesn't exist. Creating:" << download_dir;
                if (!download_dir.mkpath(save_info.dest_path))
                {
                    // Can't create the path, so use the Documents folder instead.
                    QString msg_str = tr("The download path (%1) does not exist.").arg(save_info.dest_path);
                    QLOG_WARN() << __PRETTY_FUNCTION__ << msg_str;
                    QMessageBox::critical(my_parent, tr("Configuration Error"), msg_str);
                    return false;
                }
            }
        }
    }

    // Reset our video streaming parameters.
    streamImageWidth = 0;
    streamImageHeight = 0;

    // Do we need to merge the overlay information?
    if (my_parent && save_dlg.merge_overlays())
    {
        // Setup for the rendering of the scene.
        my_parent->get_player()->render_start();

        // Are we saving a video stream?
        if (FiSaver::isStreamVideo(save_info.streamType))
        {
            QGraphicsScene  *scene_ptr;

            // Get a pointer to our "scene".
            scene_ptr = my_parent->get_player()->render_get_scene();
            if (scene_ptr == nullptr)
            {
                QLOG_WARN() << __PRETTY_FUNCTION__ << "scene_ptr is nullptr?!";
                return false;
            }

            setImageHeight((uint32_t)scene_ptr->sceneRect().height());
            setImageWidth((uint32_t)scene_ptr->sceneRect().width());
        }
    }

    // Configure the save process.
    rc = configure(save_info);
    if (rc)
    {
        // Make sure we don't refresh the cache while we're saving.
        my_parent->pause_cache_refresh(true);

        // Also, slow down the heartbeat requirement.
        my_parent->change_heartbeat(SAVE_HEARTBEAT);
    }

    // Return our result.
    return rc;
}

/*!
 * @brief If the user is going to save the recorded video to camera specific media, then we need to
 *        inform the camera about what it needs to do.
 *
 * @param save_info     The SAVE_VIDEO::SAVE_PARAMS structure.
 * @return True for success, false otherwise.
 */
bool SaveVideo::save_to_camera(SAVE_PARAMS save_info)
{
    bool        use_tag;
    uint32_t    end_frame;
    uint32_t    start_frame;
    uint32_t    reg_value;
    uint32_t    gigeStreamType;     // Stores the stream format for GIGE_REG_STREAM_FORMAT register.
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "No comms pointer!";
        return false;
    }

    persist_save_settings(save_info);

    // We're saving on the camera. Setup the camera for the save operation.
    // Are we using a tag?
    use_tag = !save_info.tag_label.isEmpty();
    if (use_tag)
    {
        QByteArray tag_buffer;

        // Get the tag.
        tag_buffer = save_info.tag_label.toUtf8();
        tag_buffer.append('\0');

        // Send the tag over to the camera.
        comm_ptr->start_write_mem(my_parent->get_ip_addr(), GIGE_REG_IMAGE_TAG, tag_buffer);
    }

    // Get the absolute start frame number.
    start_frame = save_info.ref_frame + (uint32_t)save_info.clip_start;
    end_frame = save_info.ref_frame + (uint32_t)save_info.clip_end;

    // Setup the frame information for the save.
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_SAVE_START_NUM, start_frame);
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_USER_SAVE_START_NUM, start_frame);
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_SAVE_END_NUM, end_frame);
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_USER_SAVE_END_NUM, end_frame);

    // Translate our streamType into the GIGE stream format.
    gigeStreamType = FiSaver::streamTypeToCameraStreamType(save_info.streamType, my_parent->is_series3());

    // NOTE_JLC - for now want to report error if series3 and MP4 or PACKED is selected...
    if (((gigeStreamType == GIGE_STREAM_FORMAT_MP4) || (save_info.streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)) && my_parent->is_series3())
    {
        // make the gigeStreamType 0 for the moment.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Series3 cameras don't save MP4 video on camera's media.";
        return false;
    }

    // If the translation wasn't successful something is wrong.
    if (gigeStreamType == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to determine GIGE stream type using:" << save_info.streamType;
        return false;
    }

    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_STREAM_FORMAT, gigeStreamType);

    // Now build up the control word to start the save.
    reg_value = gigeStreamType;

    // Setup the destination location.
    destMedia = save_info.dest_media;
    destLabel = save_info.dest_label;
    switch (destMedia)
    {
    case GIGE_MEDIA_SDCARD:
        reg_value |= GIGE_IMAGE_DEST_SDCARD;
        break;
    case GIGE_MEDIA_SSD:
        reg_value |= GIGE_IMAGE_DEST_SSD;
        break;
    case GIGE_MEDIA_USB:
        reg_value |= GIGE_IMAGE_DEST_USB;
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected destination value:" << Qt::showbase << Qt::hex << save_info.dest_media;
        return false;
    }

    // Go ahead and set the playback rate, even if we're not doing AVIs.
    reg_value |= (save_info.playback_rate << GIGE_PLAYBACK_RATE_SHIFT);

    // Finally, setup the other flags.
    if (use_tag)
        reg_value |= GIGE_IMAGE_USE_TAG;
    if (save_info.use_camera_name)
        reg_value |= GIGE_IMAGE_USE_CAMERA_NAME;
    if (save_info.create_xml)
        reg_value |= GIGE_IMAGE_SAVE_XML;

    // Autosave setup means we don't want the save to happen right now.
    if (autosave_enabled)
        reg_value |= GIGE_IMAGE_SAVE_DELAY;

    // Have the camera start the save.
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_IMAGE_DESTINATION, reg_value);

    // Success.
    return true;
}

/*!
 * @brief If the user is going to save the recorded video to camera specific media, then we need to
 *        inform the camera about what it needs to do.
 *
 * @param save_info     The SAVE_VIDEO::SAVE_PARAMS structure.
 * @return True for success, false otherwise.
 */
bool SaveVideo::save_to_camera_json(SAVE_PARAMS save_info)
{
    bool        use_tag;
    uint32_t    end_frame;
    uint32_t    start_frame;
    uint32_t    reg_value;
    uint32_t    gigeStreamType = 0;     // Stores the stream format for GIGE_REG_STREAM_FORMAT register.
    uint32_t    gigeImageType = 0;      // Stores the image format for the avi component of the JSON data.
    FI_comm     *comm_ptr;
    QJsonObject json_save;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return false;
    }

    // Persist the save information in this object as well before sending the
    // JSON to the camera server.
    persist_save_settings(save_info);

    // Create our unique_ptr to hold this information so we can use it for progress.
    saveToCameraJSONData = std::unique_ptr<FI_SAVER::SAVE_PARAMS>(new FI_SAVER::SAVE_PARAMS());

    // Do we have a place to put our data?
    // Pull this into a copy operation when you put this into an object form.
    if (saveToCameraJSONData)
    {
        saveToCameraJSONData->clip_end = save_info.clip_end;
        saveToCameraJSONData->clip_start = save_info.clip_start;
        saveToCameraJSONData->jpeg_quality = save_info.jpeg_quality;
        saveToCameraJSONData->h264_quality = save_info.h264_quality;
        saveToCameraJSONData->h264_speed = save_info.h264_speed;
        saveToCameraJSONData->h264_enable_all_keyframes = save_info.h264_enable_all_keyframes;
        saveToCameraJSONData->create_xml = save_info.create_xml;
        saveToCameraJSONData->trigger_relative = save_info.trigger_relative;
        saveToCameraJSONData->use_camera_name = save_info.use_camera_name;
        saveToCameraJSONData->use_camera_range = save_info.use_camera_range;
        saveToCameraJSONData->merge_overlays = save_info.merge_overlays;
        saveToCameraJSONData->dest_media = save_info.dest_media;
        saveToCameraJSONData->dest_label = save_info.dest_label;
        saveToCameraJSONData->playback_rate = save_info.playback_rate;
        saveToCameraJSONData->streamType = save_info.streamType;
        saveToCameraJSONData->imagesInputType = save_info.imagesInputType;
        saveToCameraJSONData->imagesOutputType = save_info.imagesOutputType;
        saveToCameraJSONData->useRawFormat = save_info.useRawFormat;
        saveToCameraJSONData->save_metadata_cfg = save_info.save_metadata_cfg;
        saveToCameraJSONData->save_pfmd = save_info.save_pfmd;
        saveToCameraJSONData->ref_frame = save_info.ref_frame;
        saveToCameraJSONData->comment = save_info.comment;
        saveToCameraJSONData->dest_path = save_info.dest_path;
        saveToCameraJSONData->tag_label = save_info.tag_label;
    }

    // We're saving on the camera. Setup the camera for the save operation.
    // Are we using a tag?
    use_tag = !save_info.tag_label.isEmpty();
    if (use_tag)
        json_save.insert("tag", save_info.tag_label);

    // Get the absolute start frame number.
    start_frame = save_info.ref_frame + (uint32_t)save_info.clip_start;
    end_frame = save_info.ref_frame + (uint32_t)save_info.clip_end;

    // Send up the user start/end values.
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_USER_SAVE_START_NUM, start_frame);
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_USER_SAVE_END_NUM, end_frame);

    // Translate our streamType into the GIGE stream format.
    gigeStreamType = FiSaver::streamTypeToCameraStreamType(save_info.streamType, my_parent->is_series3());

    // If the stream type translation wasn't successful cannot tell camera server what to save.
    if (gigeStreamType == 0)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to determine GIGE stream type using:" << save_info.streamType;
        return false;
    }

    // For stream video files the image type must be setup.
    if (FiSaver::isStreamVideo(save_info.streamType))
    {
        // If a stream video type - validate the image format...
        // In the old logic this would set it to either JPEG or BMP image types for AVI files.

        // RECHECK_JLC This is just for validation purposes.
        gigeImageType = FiSaver::imageTypeToCameraImageType(save_info.streamType, save_info.imagesOutputType, my_parent->is_series3());

        // If the translation wasn't successful something is wrong with the save_info data.
        if (gigeImageType == 0)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to determine GIGE Image Type using streamType:" <<
                         save_info.streamType << " imageOutputType: " << save_info.imagesOutputType;
        }
    }

    // If there is no setting for the gigeImageType then default to GIGE_STREAM_FORMAT_JPEG.
    // RECHECK_JLC - what happens in camera server if avi_codec has a JPEG setting could this
    // be contributing to the tif extension but jpeg image bug?
    if (gigeImageType == 0)
    {
        gigeImageType = GIGE_STREAM_FORMAT_JPEG;
    }

    // Setup the destination location.
    json_save.insert("dstpath", save_info.dest_path);

    // If FFLR insert the frame count.
    uint32_t frame_count = save_info.clip_end_abs - save_info.clip_start_abs + 1;
    json_save.insert("frame_count", (qint64) frame_count);

    // Setup the frame information for the save.
    json_save.insert("frame_start", (qint64)start_frame);
    json_save.insert("frame_end", (qint64)end_frame);
    json_save.insert("format", (qint64)gigeStreamType);
    json_save.insert("avi_codec", (qint64)gigeImageType);

    // Now build up the control word to start the save.
    reg_value = gigeStreamType;

    // Setup the destination location.
    destMedia = save_info.dest_media;
    destLabel = save_info.dest_label;
    switch (destMedia)
    {
    case GIGE_MEDIA_SDCARD:
        reg_value |= GIGE_IMAGE_DEST_SDCARD;
        break;
    case GIGE_MEDIA_SSD:
        reg_value |= GIGE_IMAGE_DEST_SSD;
        break;
    case GIGE_MEDIA_USB:
        reg_value |= GIGE_IMAGE_DEST_USB;
        break;
    default:
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unexpected destination value:" << Qt::showbase << Qt::hex << save_info.dest_media;
        return false;
    }

    // Go ahead and set the playback rate, even if we're not doing AVIs.
    reg_value |= (save_info.playback_rate << GIGE_PLAYBACK_RATE_SHIFT);

    // Finally, setup the other flags.
    if (use_tag)
        reg_value |= GIGE_IMAGE_USE_TAG;
    if (save_info.use_camera_name)
        reg_value |= GIGE_IMAGE_USE_CAMERA_NAME;
    if (save_info.create_xml)
        reg_value |= GIGE_IMAGE_SAVE_XML;

    // Autosave setup means we don't want the save to happen right now.
    if (autosave_enabled)
        reg_value |= GIGE_IMAGE_SAVE_DELAY;

    // Are there any flags?
    json_save.insert("dst_info", (qint64)reg_value);

    // Set the flags for saving the partition metadata CFG files.
    json_save.insert("save_metadata_cfg", save_info.save_metadata_cfg);
    json_save.insert("save_pfmd", save_info.save_pfmd);

    // If we're saving MP4 videos, we need some additional parameters.
    if (save_info.streamType == FI_SAVER::STREAM_TYPE::MP4)
    {
        json_save.insert("h264_quality", (int)save_info.h264_quality);
        json_save.insert("h264_speed", (int)save_info.h264_speed);
        json_save.insert("h264_enable_all_keyframes", save_info.h264_enable_all_keyframes);
    }

    // This is actually useful information.  I pulled it out of the condition VERBOSE_LOG
    // So that it will always output in a debug build.
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__;
        QLOG_DEBUG() << "++++++++++++++++++++++++++++++++++++++++";
        QLOG_DEBUG() << json_save;
        QLOG_DEBUG() << "++++++++++++++++++++++++++++++++++++++++";
        QLOG_DEBUG() << "json_save tag:"        << save_info.tag_label;
        QLOG_DEBUG() << "json_save dstpath:"    << save_info.dest_path;
        QLOG_DEBUG() << "json_save frame_count:" << (qint64)frame_count;
        QLOG_DEBUG() << "json_save frame_start:" << (qint64)start_frame;
        QLOG_DEBUG() << "json_save frame_end:"  << (qint64)end_frame;
        QLOG_DEBUG() << "json_save format:"     << Qt::showbase << Qt::hex << (qint64)gigeStreamType;
        QLOG_DEBUG() << "json_save avi_codec:"  << Qt::showbase << Qt::hex << (qint64)gigeImageType;
        QLOG_DEBUG() << "json_save dst_info:"   << Qt::showbase << Qt::hex << (qint64)reg_value;
        QLOG_DEBUG() << "json_save save_metadata_cfg:" << save_info.save_metadata_cfg;
        QLOG_DEBUG() << "json_save save_pfmd:"  << save_info.save_pfmd;
        if (save_info.streamType == FI_SAVER::STREAM_TYPE::MP4)
        {
            QLOG_DEBUG() << "json_h264_quality"     << (int)save_info.h264_quality;
            QLOG_DEBUG() << "json_h264_speed"       << save_info.h264_speed;
            QLOG_DEBUG() << "json_h264_enable_all_keyframes" << save_info.h264_enable_all_keyframes;
        }
        QLOG_DEBUG() << "++++++++++++++++++++++++++++++++++++++++";
    }

    // Build up the proper JSON object.
    QJsonDocument action_doc(json_save);

    // We will connect the response to our slot handler.
    connect(this, &SaveVideo::save_json_response, this, &SaveVideo::do_save_json_response, Qt::QueuedConnection);

    // Send the write request and store the result of that request in our requesting device save status.
    DEVICE_SAVE_STATE req_device_save = write_save_json_request(action_doc);

    // If the request was not sent report the request failure to the user.
    if (req_device_save < DEVICE_SAVE_STATE::SAVE_REQUESTED)
    {
        QString progress_title = tr("%1 Save Request").arg(my_parent->get_camera_name());
        QString status_str;

        if (req_device_save == DEVICE_SAVE_STATE::SAVE_SETUP_DENIED)
        {
            status_str = tr("Access is denied, a save request is already pending.");
        }
        else
        {
            status_str = tr("Unable to complete the save request at this time.");
        }
        QMessageBox::warning(MainWindow::GetInstance().get(), progress_title, status_str, QMessageBox::Ok);
    }

    // Report result of the write request to caller it should be in the state of save requested.
    // TODO: Check callers to see what they expect from this result.
    return (req_device_save == SAVE_VIDEO::DEVICE_SAVE_STATE::SAVE_REQUESTED);
}

/*!
 * @brief If the user is going to save the recorded video to the local workstation, then we need to
 *        inform the camera about what it needs to do.
 *
 * @param save_info     The SAVE_VIDEO::SAVE_PARAMS structure.
 * @returns true if the operation was successful, false if an error occurred.
 */
bool SaveVideo::save_to_pc(FI_SAVER::SAVE_PARAMS save_info)
{
    uint32_t    width;
    uint32_t    height;
    uint32_t    bit_depth;
    uint32_t    num_frames;
    uint64_t    space_needed;
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // TODO_JLC - FFLR CAPSTK needs save_to_pc support.

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return false;
    }
    if (my_parent->get_sensor() == nullptr)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No camera sensor!";
        return false;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return false;
    }

    // We are definitely not doing a copy operation reset our internal setting
    // because I wasn't sure the autosave would catch this.
    calledByCopyToPC = false;

    // Does the user want to save XML?
    convertRawToXml = save_info.create_xml;

    // We'll need some information to try to calculate the storage requirement.
    width = my_parent->read_camera_cache(GIGE_REG_WIDTH);
    height = my_parent->read_camera_cache(GIGE_REG_HEIGHT);
    bit_depth = my_parent->read_camera_cache(GIGE_REG_BIT_MODE);

    // How many frames are we trying to save?
    num_frames = (uint32_t)(save_info.clip_end - save_info.clip_start) + 1;

    // Make sure there's enough space for capture file.
    if (save_info.streamType == FI_SAVER::STREAM_TYPE::FI_CAP)
    {
        uint32_t blob_count;

        // Perform a calculation to determine the number of blobs and space needed for CAP file.
        if (!calc_blob_count_for_cap(num_frames, blob_count, space_needed))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to calculate buffer counts and space needed for CAP file!";
            return false;
        }

        // Reuse the "num_frames" variable.
        num_frames = blob_count;
    }

    // All others we can do an approximate check.
    if (!MainWindow::GetInstance()->check_avail_space(save_info.dest_path, num_frames, save_info.streamType, width, height, bit_depth, !my_parent->get_sensor()->is_color(), &space_needed))
        return false;

    // NOTE_JLC - Revisit this with MP4 in mind.
    // If we're saving an AVI/CAP, make sure we aren't trying to save too much.
    if (((save_info.streamType == FI_SAVER::STREAM_TYPE::AVI) || (save_info.streamType == FI_SAVER::STREAM_TYPE::FI_CAP)) &&
        (space_needed > FAS_VIDEO::FI_VIDEO_MAX_FAT32_SIZE))
    {
        // Check to see if the file system supports large files.
        if (!MainWindow::GetInstance()->fs_supports_large_files(save_info.dest_path))
        {
            fiMsgBox::warning(MainWindow::GetInstance().get(), tr("The destination folder does not support large (greater than 4GB) files."));
            return false;
        }
    }

    // Ask all cameras to pause the video stream.
    // Note: The pause in the stream causes the save to start.
    if (!autosave_enabled)
        MainWindow::GetInstance()->pause_video_streaming();

    // Setup for the save.
    ResetSaveVideoState();

    // Set our PC path.
    setPath(save_info.dest_path);

    persist_save_settings(save_info);
    setCount(num_frames);

    // This will tell us how many buffers make up the save operation.
    saveBufferCount = num_frames;

    // Get our logic started with the save process, if we're not in autosave mode.
    if (!autosave_enabled)
    {
        int err_code;

        if ((err_code = setup()) != 0)
        {
            QString msg_str = tr("The setup operation failed. Error: %1 [%2]").arg(strerror(err_code)).arg(err_code);
            QLOG_ERROR() << __PRETTY_FUNCTION__ << msg_str;
            QMessageBox::critical(my_parent, tr("Save Setup Failed"), msg_str);
            return false;
        }
    }

    // Success.
    return true;
}

/*!
 * @brief This method provides a setter for the jpeg_quality_old_camset with additional tests to make
 * sure that the value being set is valid.
 *
 * @param value_to_set -    The value used to set jpeg_quality_old_camset.
 * @param who_called_me -   The string identifying where this setter was invoked.
 *
 * @returns True if the value was set is reasonable and above 70% or false if the not.
 */
bool SaveVideo::set_jpeg_quality_old_camset(int value_to_set, const QString& who_called_me)
{
    bool seems_valid = false;
    if (value_to_set < 75)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "WARNING: The value_to_set is suspciously LOW:" << value_to_set << "Called by:" << who_called_me;
        if (value_to_set < 25)
            value_to_set = 25;
    }
    else if (value_to_set > 100)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "WARNING: The value_to_set is GREATER THAN 100:" << value_to_set << "Called by:" << who_called_me;
        value_to_set = 100;
    }
    else
    {
        seems_valid = true;
    }
    // This function will set the value requested but any error message will indicate an issue.
    // This has been and still is notoriously difficult to diagnose when we get invalid settings so I have added this
    // intermediary while debugging this old code.
    jpeg_quality_old_camset = value_to_set;
    return seems_valid;
}

/*!
 * @brief This method is called to inform the worker object that we have received all the images
 *        and the save operation is done.
 */
void SaveVideo::saving_finished(const QString& whoCalledMe)
{
    //----------------------------------------------------------------------------------------------------------------------
    // NOTES:
    // This function can be called in many ways.

    // Internally in this SaveVideo object it can be called by the two distinct handlers for status updates:
    // SaveVideo::getVideoStreamSaveStatus for video stream saves (MP4 and AVI)
    // SaveVideo::timeForStackProgressFrameUpdate for BLOB and Image stack saves.

    // When saving BLOB and Image stack saves an internal connection is made when the internal Progress Xfer Dialog
    // is constructed by SaveVideo::PrepXferDialogFromFSaver or SaveVideo::PrepXferDialogFromSaveParams.  The connection:
    // connect(this, &SaveVideo::saverBufferReleaseSignal, this, &SaveVideo::timeForStackProgressFrameUpdate, Qt::UniqueConnection);
    // There is NO disconnection.
    // Therefore all SaveVideo::saverBufferReleaseSignal "signals" end up SaveVideo::timeForStackProgressFrameUpdate and it will
    // end up calling saving_finished if it determines the save is complete.

    // When saving Video streams the SaveVideo::getVideoStreamSaveStatus is called via a polling timer.

    // AND FINALLY it can also be called by CameraMDI in the following interfaces:
    // void CameraMDI::copy_finished(QString whoCalledMe)
    // void CameraMDI::retry_missing_frames() - (If copying CAP files)
    // void CameraMDI::save_video_timeout()
    // void CameraMDI::saver_is_idle()
    //----------------------------------------------------------------------------------------------------------------------

    // If the worker thread has already been told that saving is finished we have nothing to do here.
    if (savingHasFinished)
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "whoCalledMe:" << whoCalledMe << "savingHasFinished already and worker thread is finishing.";
        return;
    }
    else
    {
        QLOG_TRACE() << __PRETTY_FUNCTION__ << "whoCalledMe:" << whoCalledMe << "!savingHasFinished attempting to notify worker thread.";
    }

    if (saving_finished_in_process)
    {
        QLOG_WARN() << "<<<< Already processing SaveVideo::saving_finished returning from this call! >>>>";
        return;
    }

    saving_finished_in_process = true;

    // Go ahead and pause the video request at the connected camera level
    // but if for some reason there is no parent go ahead and let the Save
    // object mark itself as finished.
    {
        CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
        // Sanity check.
        if (my_parent == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        }
        else
        {
            if (VERBOSE_SAVE_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << my_parent->get_ip_addr_pretty() << " streamType:  " << FiSaver::streamTypeToQString(streamType);
            }

            // Don't want to continue to look for frames on the parent streamer as frames
            // are complete in terms of reading from camera.  The flushing operations of
            // some video stream writers takes long enough to cause the stream reader of
            // the camera to time-out.
            my_parent->pause_video_request();
        }
    }

    // Lock access to the ready list. The mutex will be unlocked
    // when the read_list_lock object is destroyed.
    //TODO:  Why was this being locked here?!!
    //QMutexLocker    ready_list_lock(&imageBufferRecvMutex);

    // Note:  This is really the heart of determining whether or not the save operation
    // is done.  For video we currently rely on the polling for video status through the
    // calls to getVideoStreamStatus.  Surely this can be made to be more isolated and
    // fullproof - as it stands this is shaky ground if you modify it in any way.

    // If this saver has not been properly setup we cannot start any save stream.
    if (!setupForSave)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The saver had not been properly setup for saving.  Please call SaveVideo::setup() before starting a save stream.";
    }

    // Set either the wait_for_save_complete if a stream video or save_is_done if not.
    if (FiSaver::isStreamVideo(streamType))
    {
        // We have successfully determined that a stream video save is completed and the worker thread will
        // complete the process.
        savingHasFinished = true;
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "The stream video is now waiting for save complete.";
        }
        // Some stream video writers flush encoders and even though all of the
        // frames were received there may still be work to complete before save_is_done.
        wait_for_save_complete = true;
    }
    else
    {
        // We have successfully determined that a non-stream video save is completed and the worker thread
        // will complete the process.
        savingHasFinished = true;
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << " non-stream video save is finished.";
        }
        // Nothing to wait for the save process is complete.
        save_is_done = true;

        // Make sure our worker thread knows that the save_is_done.
        worker_wakeup.wakeAll();
    }

    saving_finished_in_process = false;
}

/*!
 * @brief We're going to try to download the "retry_frame" frame again.
 * @param retry_frame   The absolute frame number of the frame we're going to try to download.
 */
void SaveVideo::setRetryFrame(uint32_t retry_frame_abs, int retry_frame_rel, uint32_t retry_frames_resent, uint32_t total_retry_frames)
{
    QString streamTypeStr = FiSaver::streamTypeToQString(streamType);
    QString imageTypeStr = FiSaver::imageTypeToQString(imageInputType);

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "retry_frame(absolute):" << retry_frame_abs << "retry_frame_rel:" << retry_frame_rel
                 << "retry_frames_resent:" << retry_frames_resent << "total_retry_frames:" << total_retry_frames
                 << "clip_start:" << clipStart
                 << "retry_in_progress:" << retryInProgress
                 << "calledByCopyToPC:" << calledByCopyToPC
                 << "streamType:" << streamTypeStr
                 << "imageType:" << imageTypeStr;

    // We need to adjust the frame number if we're saving images to an AVI file
    if (streamType == FI_SAVER::STREAM_TYPE::AVI)
    {
        // But only if we're copying.
        if (calledByCopyToPC)
        {
            uint32_t origRetryFrame = retry_frame_abs;
            retry_frame_abs = retry_frame_abs - copy_start_frame;
            if (VERBOSE_SAVE_LOG)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "AVI copy original retry_frame:"
                             << origRetryFrame << "copy_start_frame:" << copy_start_frame
                             << "adjusted retry frame:" << retry_frame_abs;

            }
        }
        else if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "AVI save the retry_frame:" << retry_frame_abs;
        }
    }

    // Do we have our xferDialog up and running?
    if (xferDialog)
    {
        QString resent_info_str = tr("received %1 of %2 missed frame%3.").arg(retry_frames_resent).arg(total_retry_frames).arg((total_retry_frames > 1 ? "s" : ""));
        QString statusMsg = tr("Resending frame [%1] %2").arg(retry_frame_rel).arg(resent_info_str);
        xferDialog->setStatusValue(statusMsg);
        xferDialog->setStatusIcon(QStyle::SP_MessageBoxInformation);
        xferDialog->showStatusIcon(true);
        xferDialog->showStatusValue(true);

        // We can modify the percentage too now that we have the total count...
        xferDialog->setBarMinMax(0, total_retry_frames);
        xferDialog->setBarValue(retry_frames_resent);
    }

    // Save the retry frame as our starting clip.
    // NOTE:  What happens in here is the "retry_in_progress" state becomes active
    // if it wasn't active before and an absolute frame number for retry process
    // is set up.
    FiSaver::setRetryFrame(retry_frame_abs);
}

/*!
 * @brief If the worker thread is running it makes sure that it is
 * stopped.
 *
 * @returns The result is true if all was handled, false if error.
 */
bool SaveVideo::StopWorkerThreadIfRunning()
{
    bool stoppedWorkerThread = false;

    // Do we have our image saver thread running?
    if (image_thread.isRunning())
    {
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Worker thread was running we are stopping it now.";
        }

        // Set our flag to stop the thread.
        thread_stop = true;

        // Make sure the thread is awake.
        worker_wakeup.wakeAll();

        // Wait for the thread to finish.
        image_thread.waitForFinished();
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Worker thread is now stopped.";
        }
    }
    stoppedWorkerThread = true;

    // Reset any setup condition for this object.  Once no worker thread is running there is no
    // way we can run again until another setup occurs.
    setupForSave = false;

    return stoppedWorkerThread;
}

void SaveVideo::update_image_dest(uint32_t img_dest)
{
    // If we have a xfer progress dialog showing then update its information.
    // Note:  This is old code left in place from the old fiProgressDlg logic.
    // It gets invoked from CameraMDI::handle_unsol_msg for GIGE_REG_IMAGE_DESTINATION
    // And I am not sure if this is leftover that is specific to series3 or if this
    // is for all camera server saves?  #REVISIT_JLC
    if (xferDialog)
    {
        // If this is a CAP file save, then the user can't cancel.
        if ((img_dest & GIGE_STREAM_FORMAT_MASK) == GIGE_STREAM_FORMAT_CAPTURE)
        {
            // CAP file, so hide the cancel button.
            xferDialog->showCancel(false);
            disconnect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &SaveVideo::userHasCancelled);
        }
        else
        {
            // Not a CAP file, so allow the user to cancel.
            xferDialog->showCancel(true);
            connect(xferDialog.get(), &fiProgressDialog::progressCancelled, this, &SaveVideo::userHasCancelled, Qt::UniqueConnection);
        }
    }
}

/*!
 * @brief This method is used to setup and initiate the worker thread.
 *
 * @param amICopyToPC - TRUE if this was called by SaveVideo::copyToPC() - which is the only
 * mechanism that this saver gets invoked from the camera explorer object.  This is a default
 * parameter initialized to false.
 *
 * @return 0 on success, errno on failure.
 */
int SaveVideo::setup(bool amICopyToPC)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (basePath.isEmpty())
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "No base path specified!";
    }
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return EFAULT;
    }

    // If this object is setupForSave already then something is wrong.
    if (setupForSave)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "A previous setup was not properly terminated.";
        return EFAULT;
    }

    // I want to know this for the purposes of testing with pausing the streamer or not.
    // We need a much better way to interact with the CamExplore object here then by doing
    // this right now.
    calledByCopyToPC = amICopyToPC;

    QLOG_TRACE() << __PRETTY_FUNCTION__ << Qt::showbase << Qt::hex << my_parent->get_ip_addr_pretty()
                 << "streamType:" << FiSaver::streamTypeToQString(streamType) << "Type:" << (calledByCopyToPC ? "CopyToPC" : "SaveToPC");

    // Are we in autosave mode? If we are, then we need to update our counts
    // before we start setting up containers and such.
    if (local_autosave())
    {
        // Update our settings for a new autosave.
        // Note:  Normally a lot of this setup would occur in the SaveVideo::save_to_pc and
        // SaveVideo::copy_to_pc functions.  Since those are not invoked when autosaving this
        // ensures that all setup data is properly initialized for an autosave to the PC.
        update_autosave_setup();
    }

    // Are we reusing the path information?
    if (!reuse_path)
    {
        int rc;

        // Stream video writers and fastec imaging capture BLOBS take special handling.
        if (FiSaver::isStreamVideo(streamType))
        {
            // Are we saving a packed stream?
            if (streamType == FI_SAVER::STREAM_TYPE::RDI_PACKED)
            {
                const pm_metadata *pm_ptr = nullptr;

                // Sanity checks.
                if (my_parent && (my_parent->get_partition_metadata() != nullptr))
                    pm_ptr = (const pm_metadata *)my_parent->get_partition_metadata()->constData();

                // We will query for the image width and height no matter the image format.
                streamImageWidth = my_parent->read_camera_cache(GIGE_REG_WIDTH);
                streamImageHeight = my_parent->read_camera_cache(GIGE_REG_HEIGHT);

                // Prepare the packed stream writer object.
                rc = preparePackedWriter(pm_ptr);
            }
            else
            {
                // Setup for video stream writing.
                rc = prepareStreamWriter();
            }
        }
        else if (FiSaver::isCaptureStream(streamType))
        {
            // Get a file name we can use. If it works, the result will be in dest_path.
            rc = createFilePath();
            if (rc == 0)
            {
                // Create the file.
                blobFile = new QFile(this);
                if (blobFile == nullptr)
                {
                    QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to allocate blob_file!";
                    rc = ENOMEM;
                }
                else
                {
                    bool    b_status;

                    // Set the file name to our object and open the file.
                    blobFile->setFileName(destPath);

                    if (VERBOSE_SAVE_LOG)
                    {
                        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Setting name of Blob to: " << destPath;
                    }

                    b_status = blobFile->open(QIODevice::WriteOnly);   // this will overwrite if already there
                    if (b_status)
                    {
                        blobFile->setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadGroup | QFile::WriteGroup | QFile::ReadOther);
                    }
                    else
                    {
                        // Log the error.
                        QLOG_ERROR() << __PRETTY_FUNCTION__ << QStringLiteral("Unable to create file %1 err:%2").arg(destPath, blobFile->errorString());
                        rc = blobFile->error();
                    }
                }
            }
        }
        else
        {
            // Get the target path and folder(s) setup for the save.
            rc = createDestPath();
        }

        // Return now if there is an error condition.
        if (rc != 0)
            return rc;
    }

    // Initialize our run variables.
    cancel_in_progress = false;
    lastErrMsg.clear();
    thread_stop = false;

    // Setup our blob size, in case we need it.
    setBlobSize(my_parent->blob_size());

    // Since we are going to be downloading image data from the camera
    // let's pause the camera discovery polling until the thread completes.
    MainWindow::GetInstance()->pause_discovery();

    // We're good to go. Start up our image saving thread.
    image_thread = QtConcurrent::run(image_saver, this);
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "image_saver() forked.";

    // Mark this object in a setup state.
    setupForSave = true;
    savingHasFinished = false;
    emittedSaveComplete = false;
    emittedSaveError = false;
    emittedSaveReceiveCountComplete = false;
    closeInProgress = false;
    closeCompleted = false;

    // All is well.
    return 0;
}

/*!
 * @brief Initiate the local path save using our fiSaver file writer.
 */
void SaveVideo::start(const QString& whoCalledMe)
{
    CameraMDI   *my_parent      = qobject_cast<CameraMDI *>(parent());
    uint32_t gigeStreamFormat   = 0;

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "called by:" << whoCalledMe;

    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // If this saver has not been properly setup we cannot start any save stream.
    if (!setupForSave)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "The saver has not been properly setup for saving.  Please call SaveVideo::setup() before starting a save stream.  Ignoring this request.";
        return;
    }

    // Setup our blob size, in case we need it.
    setBlobSize(my_parent->blob_size());

    // Reset our statistics.
    reset_stats();

    // Set the gige camera format for the streamType.
    gigeStreamFormat = FiSaver::streamTypeToCameraStreamType(streamType, my_parent->is_series3());

    // If this fails I don't know how to gracefully fail
    if (gigeStreamFormat == 0)
    {
        QString streamTypeStr = FiSaver::streamTypeToQString(streamType);
        QLOG_FATAL() << __PRETTY_FUNCTION__ << "Unable to determine gigeStreamFormat for " << streamTypeStr;
        return;
    }

    // Ticket #326 - if a capture file is being saved there is a different save operation
    // initiated at the CameraMDI level.
    if (FiSaver::isCaptureStream(streamType))
    {
        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "CAP file Diversion frame_count:" << frameCount << ".";
        }

        my_parent->start_session_save();
    }
    else
    {
        // Get the stream started.
        my_parent->stream_save_video(clipStart, (clipStart + frameCount) - 1, streamType, imageInputType);
    }

    // If a copy is in progress we do not update any progress dialog because it
    // is being managed outside of SaveVideo.  However if a save is in progress
    // then make sure that the transfer dialog is set up for this save operation.
    if (!calledByCopyToPC)
    {
        // If autosaving make sure the value is set to zero.
        if (autosave_configured)
        {
            // Each time an autosave save is started the general save is initialized here.
            // The transfer dialog should be properly set up when autosave was setup
            // however that happens once and at that time the maximum value will
            // be set.  We need to make sure we reset transfer dialog value to minimum
            // each time through this loop.
            // Make sure the transfer dialog is set up properly for this save operation.
            PrepXferDialogFromFSaver(__PRETTY_FUNCTION__);
            if (xferDialog)
            {
                //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "AUTOSAVE xferDialog Setup- saveBufferCount:" << saveBufferCount;
                // Set to the buffer count we expect for this save transfer.
                xferDialog->setBarMinMax(0, (int)saveBufferCount);
                xferDialog->setBarValue(0);
            }
        }
        else
        {
            // Make sure the transfer dialog is set up properly for this save operation.
            PrepXferDialogFromFSaver(__PRETTY_FUNCTION__);
        }

        // Start to show progress on the save operation to the user.
        if (xferDialog)
        {
            // Show the dialog with polling set.
            if (FiSaver::isStreamVideo(streamType))
                xferDialog->showDialogWithPolling(GIGE_REG_SAVE_STATE_POLLING_RATE);
            else
                xferDialog->showDialogWithoutPolling();
        }
        else
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "There is no xferDialog ready for progress updates!";
        }
    }
    else if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "COPY IN PROGRESS and we don't use internal SaveVideo::xferDialog here.";
    }
}

/*!
 * @brief Common modification of xferDialog if responding to a cancel dialog state.
 * This can be called in multiple places and the need for a central function exists.
 */
void SaveVideo::UpdateXferDialogForCancelButtonHit()
{
    // Do we have a xfer progress dialog set up?  If so perform some status modifications.
    if (xferDialog)
    {
        // Change the message to let the user know we're cancelling.
        xferDialog->setStatusValue("Cancel in progress.  Please wait.");
        xferDialog->showStatusValue(true);
        xferDialog->showCancel(true);
        xferDialog->showBar(false);
    }
}

/*!
 * @brief Update the parameters used for saving.
 */
void SaveVideo::update_autosave_setup()
{
    int         end_frame;
    int         start_frame;
    uint32_t    num_frames;
    uint32_t    trig_frame;
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__;
    }

    // Check our camera pointer.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Setup our variables.
    ResetSaveVideoState();

    // How many frames did we actually capture.
    num_frames = comm_ptr->get_reg(my_parent->get_ip_addr(), GIGE_REG_PART_NUM_FRAMES);

    // Read what the trigger actually is.
    trig_frame = comm_ptr->get_reg(my_parent->get_ip_addr(), GIGE_REG_PART_TRIG_FRAME);

    // Calculate the absolute starting frame number.
    start_frame = (int)trig_frame + as_clip_start;
    if (start_frame < 0)
        start_frame = 0;
    setClipStart((uint32_t)start_frame);

    // Get the absolute ending frame number.
    end_frame = (int)trig_frame + as_clip_end;
    if (end_frame < 0)
        end_frame = 0;
    if (end_frame == start_frame)
    {
        // How many frames did the user request?
        uint32_t requested_length = (uint32_t)(as_clip_end - as_clip_start);

        // Make sure that will fit in the number of frames captured.
        if (num_frames < requested_length)
            requested_length = num_frames;

        // Adjust our end frame appropriately.
        end_frame = start_frame + (int)requested_length;
    }
    setClipEnd((uint32_t)end_frame);

    // Adjust the number of frames being saved.
    if ((uint32_t)((end_frame - start_frame) + 1) < num_frames)
        num_frames = (uint32_t)(end_frame - start_frame) + 1;

    // Make sure we have a minimum number of frames captured.
    if (num_frames < 2)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << "Not enough frames captured!";
        return;
    }

    // If this is a CAP file make sure that we work with blob counts not frame counts.
    // Make sure there's enough space for capture file.
    if (streamType == FI_SAVER::STREAM_TYPE::FI_CAP)
    {
        uint32_t blob_count;
        uint64_t space_needed;

        // Perform a calculation to determine the number of blobs and space needed for CAP file.
        if (!calc_blob_count_for_cap(num_frames, blob_count, space_needed))
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to calculate buffer counts and space needed for CAP file!";
            return;
        }

        // Reuse the "num_frames" variable.
        num_frames = blob_count;

        if (VERBOSE_SAVE_LOG)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Calculated CAP file autosave blob_count:" << blob_count;
        }
    }

    // Update our count.
    setCount(num_frames);

    // This will tell us how many buffers make up the save operation.
    saveBufferCount = num_frames;
}

/*!
 * @brief On the completion of a download, we want to update our statistics about the
 *        download.
 */
void SaveVideo::update_stats()
{
    double  bytes_per_sec;
    double  mbytes_per_sec;
    int     elapsed_ms;
    QTime   elapsed(0, 0, 0, 0);
    QString target_root;
    QString stats_msg;
    QString ext_str;
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // If copying CAP files don't output any statistics.
    // This used to just be "copy_in_progress" but some of the stack
    // copy statistics is helpful.
    if (calledByCopyToPC && FiSaver::isCaptureStream(streamType))
    {
        return;
    }

    // How long did it take?
    elapsed_ms = (int)download_duration.elapsed();
    elapsed = elapsed.addMSecs(elapsed_ms);

    // Check our camera pointer.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // How many bytes per second.
    bytes_per_sec = download_byte_count / (elapsed_ms / 1000.0);
    mbytes_per_sec = (bytes_per_sec / (1024.0 * 1024.0));

#ifdef Q_OS_WIN
    // Grab the drive letter used.
    target_root = destPath.left(2);
#elif defined( Q_OS_MAC ) || defined( Q_OS_LINUX )
    // Grab the root folder.
    target_root = destPath.left(destPath.lastIndexOf(QLatin1Char('/')) + 1);
#endif

    if (FiSaver::isStreamVideo(streamType))
    {
        QString imageTypeForEncoderStr;

        // #MP4IMAGEINPUT - SaveVideo::update_stats - file extension text format.
        // Always output the encoder file type for informational purposes.
        // before it was conditionalized on whether or not the
        // imageInputType != imageTypeForEncoder
        if (imageTypeForEncoder == FI_SAVER::IMAGE_TYPE::BMP)
            imageTypeForEncoderStr = QString::fromUtf8("/Enc:") + IMAGE_EXT_BMP;
        else if (imageTypeForEncoder == FI_SAVER::IMAGE_TYPE::BMP_PLANAR)
            imageTypeForEncoderStr = QString::fromUtf8("/Enc:") + IMAGE_EXT_BPP;
        else if (imageTypeForEncoder == FI_SAVER::IMAGE_TYPE::JPEG)
            imageTypeForEncoderStr = QString::fromUtf8("/Enc:") + IMAGE_EXT_JPG;

        // Build up the general statistic about the extension inforamtion.
        if (imageInputType == FI_SAVER::IMAGE_TYPE::BMP)
            ext_str = fileExtension + QString::fromUtf8("/In:") + IMAGE_EXT_BMP + imageTypeForEncoderStr;
        else if (imageInputType == FI_SAVER::IMAGE_TYPE::BMP_PLANAR)
            ext_str = fileExtension + QString::fromUtf8("/In:") + IMAGE_EXT_BPP + imageTypeForEncoderStr;
        else if (imageInputType == FI_SAVER::IMAGE_TYPE::JPEG)
            ext_str = fileExtension + QString::fromUtf8("/In:") + IMAGE_EXT_JPG + imageTypeForEncoderStr;
        else
            ext_str = fileExtension;    // This actually may be .mp4 or .avi so not right either
    }
    else
    {
        ext_str = fileExtension;
    }

    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "download_frame_count:" << download_frame_count << "saved_buffer_count:" << stream_save_buffer_count << "saveBufferCount:" << saveBufferCount;

    // Build up our output string.
    stats_msg = QString::fromUtf8("Duration: %1  TotalBytes: %2  TotalFrames: %3  EffectiveRate: %4 MB/s  MissedFrames: %5  Drv: %6  Ext: %7")
                .arg(elapsed.toString(QString("hh:mm:ss.zzz"))).arg(download_byte_count).arg(download_frame_count).arg(mbytes_per_sec)
                .arg(my_parent->get_missed_frames())
                .arg(target_root, ext_str);
    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");
    QLOG_INFO() << __PRETTY_FUNCTION__ << destPath;
    QLOG_INFO() << __PRETTY_FUNCTION__ << stats_msg;
    QLOG_INFO() << __PRETTY_FUNCTION__ << QString::fromUtf8("===========================================================================");

    // Check to see if the user wants to save the statistics.
    if (MainWindow::GetInstance()->do_stats())
    {
        QFile   stats_fd;
        QString stats_path;

        // Build up the file path for our statistics file. Make sure the folder exists.
        stats_path = QStandardPaths::standardLocations(QStandardPaths::DocumentsLocation).at(0);
        QDir stats_dir(stats_path);
        stats_dir.mkpath(stats_path);

        // Make sure the path is formatted properly.
        stats_path = QDir::fromNativeSeparators(stats_path);
        if (!stats_path.endsWith(QLatin1Char('/')))
            stats_path.append(QLatin1Char('/'));
        stats_path.append(STATS_FILE_NAME);

        //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "stats_path:" << stats_path;

        // Check to see if the file already exists.
        stats_fd.setFileName(stats_path);
        if (!stats_fd.exists())
        {
            // Try to create the file.
            if (!stats_fd.open(QIODevice::WriteOnly | QIODevice::Text))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << stats_path << "Err:" << stats_fd.errorString();
                return;
            }

            // Set the file permissions.
            stats_fd.setPermissions(QFile::ReadOwner | QFile::WriteOwner | QFile::ReadUser | QFile::WriteUser | QFile::ReadOther);

            // Write out the headers into the file.
            stats_fd.write("Timestamp,Duration,Bytes,Speed,MissFrames,GoodFrames,PktSize,PktDlyBegin,PktDlyEnd,Buffer,Backoff,Min,Drv,Ext,CamIP,Destination\n");
        }
        else
        {
            // Open up the file.
            if (!stats_fd.open(QIODevice::WriteOnly | QIODevice::Append | QIODevice::Text))
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Unable to open file:" << stats_path << "Err:" << stats_fd.errorString();
                return;
            }
        }

        // Build up our string.
        stats_msg = QString::fromUtf8("%1,%2,%3,%4,%5,%6,%7,%8,%9,%10,%11,%12,%13,%14,%15,%16\n")
                    .arg(download_start.toString(QString("yyyy/MM/dd hh:mm:ss")), elapsed.toString(QString("hh:mm:ss.zzz")))
                    .arg(download_byte_count).arg(mbytes_per_sec)
                    .arg(QString::number(my_parent->get_missed_frames())).arg(download_frame_count).arg(my_parent->get_packet_len())
                    .arg(my_parent->get_camera_info()->packet_delay)
                    .arg(comm_ptr->get_streamer(my_parent->get_ip_addr())->get_packet_delay())
                    .arg(comm_ptr->get_streamer(my_parent->get_ip_addr())->get_num_buffers())
                    .arg(comm_ptr->get_streamer(my_parent->get_ip_addr())->get_backoff())
                    .arg(comm_ptr->get_streamer(my_parent->get_ip_addr())->get_low_watermark())
                    .arg(target_root).arg(ext_str)
                    .arg(my_parent->get_ip_addr_pretty(), destPath);

        // Write it out.
        stats_fd.write(stats_msg.toUtf8());

        // All done.
        stats_fd.close();
    }
}

/*!
 * @brief Slot method which receives notification when a write completes.
 * This only monitors our registers which we control.
 *
 * @param ip_addr       IP address of the camera that was being written.
 * @param reg_addr      The address of the register being written.
 */
void SaveVideo::write_complete(uint32_t ip_addr, uint32_t reg_addr)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    bool                    ignore_it = false;
    uint32_t                write_status = GEV_STATUS_SUCCESS;
    FI_comm                 *comm_ptr;
    QString                 reg_name_str;
    QHostAddress            our_addr;

    // Make sure our camera object exists.
    if (my_parent == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_FATAL() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Is this event for us?
    if (ip_addr != my_parent->get_ip_addr())
        return;

    // This might be useful below.
    our_addr.setAddress(qFromBigEndian(ip_addr));

    // **Important**
    // If the function comm_ptr->get_write_status returns FI_COMM::GEV_STATUS_UNKNOWN someone else pulled the
    // write complete status already. This should not happen if we are managing all writes to these
    // registers within this module.

    // We only care about these two registers, otherwise we just ignore it.
    switch (reg_addr)
    {
    case GIGE_REG_SAVE_STATE:
        reg_name_str = "GIGE_REG_SAVE_STATE";
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        break;

    case GIGE_REG_SAVE_JSON:
        reg_name_str = "GIGE_REG_SAVE_JSON";
        write_status = comm_ptr->get_write_status(ip_addr, reg_addr);
        emit save_json_write_complete(write_status);
        break;
    default:
        ignore_it = true;
        break;
    }

    // Default case means we ignored it.
    if (!ignore_it)
    {
#if LOG_WRITE_COMPLETE
        // I want to pop this so I am using a warning for the moment.
        // TODO: Remove the warning usage later and replace with QLOG_DEBUG...
        QLOG_WARN() << __PRETTY_FUNCTION__ << "WRITE COMPLETE" << reg_name_str << "========>" << FI_comm::gev_status_to_msg(write_status);
#endif
    }
}

/*!
 * @brief Responds to FasFireLR underflow recovery request based on camera's write_status.
 */
void SaveVideo::do_fflr_underflow_cancelled()
{
    // If we are in the normal state of processing without any errors mark condition and ask camera to stop saving.
    if (fflr_underflow_state == FFLR_UNDERFLOW_STATE::COMPLETE_PROGRESS)
    {
        // Ask the camera to cancel any save operation.
        CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

        // We are just processing the recovery. Mark the state as canceled.
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_PROGRESS_CANCEL;

        // Pass cancel request on to parent.
        if (my_parent != nullptr)
        {
            my_parent->cancel_save();
        }
    }

    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "fflr_underflow_state:" << fflr_underflow_state;

    // User is done with FasFireLR recovery.
    fiProgressDialog::RecycleProgressDialog(fflrRecoverDlg);
    fflrRecoverDlg = nullptr;
}

/*!
 * @brief Generates a common status message based on the autosave register.
 * @param[out] status_msg   The generated message.
 * @param[out] total_count  How many buffers exist for autosave.
 * @param[out] dirty_count  How many buffers are recorded out of total_count.
 * @returns The count of dirty buffers for the autosave.
 */
void SaveVideo::get_autosave_status_msg(QString& status_msg, uint32_t& total_count, uint32_t& dirty_count)
{
    // Ask the camera to cancel any save operation.
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    uint32_t reg_value;
    QString buffer_count;

    dirty_count = 0;
    total_count = 0;

    // Query the Autosave register.
    if (my_parent)
    {
        reg_value = my_parent->read_camera_cache(GIGE_REG_AUTOSAVE);
        dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);
        total_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value);
    }
    //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "dirty_count:" << dirty_count;

    // Grammer needs to be right.
    if (dirty_count <= 1)
        buffer_count = tr("there is 1 capture buffer to process.");
    else
        buffer_count = tr("there are %1 capture buffers to process.").arg(dirty_count);

    // What is our status?
    status_msg = tr("Completing partial recording, %1").arg(buffer_count);
}

/*!
 * @brief Reports on polling basis the number of dirty partitions being recovered.
 */
void SaveVideo::do_fflr_underflow_progress()
{
    // Ask the camera to cancel any save operation.
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // If we can't get to our parent camera that's a problem.
    if (!my_parent)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Do we have a FasFireLR Dialog?
    if (fflrRecoverDlg)
    {
        // What is our status?
        QString status_str;
        uint32_t dirty_count, total_count;
        get_autosave_status_msg(status_str, total_count, dirty_count);

        fflrRecoverDlg->setBarMax(fflr_underflow_dirty_count * 100);
        fflrRecoverDlg->setStatusValue(status_str);

        // Show the dialog if it is not showing.
        if (!fflrRecoverDlg->isVisible())
        {
            fflrRecoverDlg->showDialogWithoutPolling();
        }

        // Query the camera for save status to keep our pace with the recovery save operation.
        my_parent->start_camera_read_reg(GIGE_REG_SAVE_STATE);
        my_parent->start_camera_read_reg(GIGE_REG_AUTOSAVE);

        // As long as the user hasn't cancelled let's keep updating.
        if (!fflrRecoverDlg->wasCanceled())
        {
            QTimer::singleShot(FFLR_UNDERFLOW_PROGRESS_POLL, this, &SaveVideo::do_fflr_underflow_progress);
        }
    }
}

/*!
 * @brief Responds to FasFireLR underflow recovery request based on camera's write_status.
 *
 * @param[in] write_status  Result of the request to complete the partial recording from the camera.
 *
 */
void SaveVideo::do_fflr_underflow_save_json_response(uint32_t write_status)
{
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    // Status expected for GIGE_REG_SAVE_JSON:
    // GEV_STATUS_ERROR, GEV_STATUS_INVALID_PARAMETER, GEV_STATUS_BUSY, GEV_STATUS_ACCESS_DENIED, or GEV_STATUS_SUCCESS.

    // The user has requested a partial recording completion as a result of FasFireLR underflow error.
    // The GIGE_REG_SAVE_JSON register was written and this is the result of the write complete.

    // Make sure to disconnect this connection!
    disconnect(this, &SaveVideo::save_json_response, this, &SaveVideo::do_fflr_underflow_save_json_response);

    // If we can't get to our parent camera that's a problem.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
    }

    // If all went well start the process of polling for underflow recovery progress.
    // In the function fcp.cpp FCP_LOCAL::fcp_parse_save_json it calls imageSaveWorker::parseSaveSettings which
    // reads the JSON and returns true if the proper keys existed. Then the handler in fcp_parse_save_json will
    // check to see if an instance of imgSaveWorker is running if it is, it returns GEV_STATUS_ACCESS_DENIED if
    // it isn't and it's spawned off it returns GEV_STATUS_BUSY - so that's really what should be returned here.
    // I am removing the check for GEV_STATUS_SUCCESS here as it really should be set to GEV_STATUS_BUSY.
    if ((my_parent != nullptr) && (write_status == GEV_STATUS_BUSY))
    {
        // Start the progress of reporting recovery, initially this will be an activity progress bar until we get updates.
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_PROGRESS;

        // Give things a second to settle out... Then start progress updates.
        QTimer::singleShot(FFLR_UNDERFLOW_LAUNCH_DELAY, this, &SaveVideo::do_fflr_underflow_progress);
    }
    // Otherwise we can't start the recovery process, notify the user and wait for them to hit the cancel button.
    else
    {
        // Mark the state as error.
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_REQUEST_FAILED;
        QString status_str = QString("The camera is unable to complete partial recording. [GEV_STATUS: %1]").arg(FI_comm::gev_status_to_msg(write_status));
        QLOG_ERROR() << __PRETTY_FUNCTION__ << status_str;

        // Modify our recovery dialog to show the error condition.
        fflrRecoverDlg->setStatusValue(status_str);
    }
}

/*!
 * @brief Handles final response status when writing GIGE_REG_SAVE_JSON to start a DeviceSave/fflrRecover operation on camera.
 *
 * @param write_status     The GEV status code from the last write of GIGE_REG_SAVE_JSON.
 *
 */
void SaveVideo::do_save_json_response(uint32_t write_status)

{
    // Status expected for GIGE_REG_SAVE_JSON:
    // GEV_STATUS_ERROR, GEV_STATUS_INVALID_PARAMETER, GEV_STATUS_BUSY, GEV_STATUS_ACCESS_DENIED, or GEV_STATUS_SUCCESS.

    // Make sure to disconnect this connection!
    disconnect(this, &SaveVideo::save_json_response, this, &SaveVideo::do_save_json_response);

    // If an error occurred warn the user.
    if (write_status != GEV_STATUS_SUCCESS)
    {
        QString     camera_name = "camera";
        CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
        if (my_parent)
        {
            camera_name = my_parent->get_camera_name();
        }

        // If the status was not successful, handle the process of cancelling the save on our end.
        QString title_str = QString("%1 Save Request").arg(camera_name);
        QString status_failed;

        if (write_status == GEV_STATUS_ACCESS_DENIED)
        {
            status_failed = QString("%1 has denied access for the save operation requested.").arg(camera_name);
        }
        else
        {
            // TODO: Could probably break out the codes after they are fully understood to give better feedback.
            status_failed = QString("%1 save request has failed, status received[%2]").arg(camera_name).arg(FI_comm::gev_status_to_msg(write_status));
        }

        // Log the error.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << status_failed;

        // Pop up the warning message for the user.
        QMessageBox::warning(MainWindow::GetInstance().get(), title_str,
                             status_failed,
                             QMessageBox::Ok);

        // Sanity check.
        if (my_parent == nullptr)
        {
            QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
            return;
        }
        // Attempt to cancel any partial save operation.
        my_parent->cancel_save();
    }
    else
    {
        // Mark our active_device_save as waiting for progress.
        active_device_save = SAVE_VIDEO::DEVICE_SAVE_STATE::SAVE_PROGRESS;

        // Make sure there is a poll timer running for catching the device save progress.
        if (save_state_poll_timer == nullptr)
        {
            // Start up our polling timer. This happens after the command to start, so that there is less
            // chance of us reading back an "idle" status and thinking the save has completed before it's
            // even begun.
            save_state_poll_timer = new QTimer(this);
            connect(save_state_poll_timer, SIGNAL(timeout()), this, SLOT(save_state_poll_timeout()));
            save_state_poll_timer->start(GIGE_REG_SAVE_STATE_POLLING_RATE);
        }
    }
}

/*!
 * @brief Handles the reception of the write complete status for GIGE_REG_SAVE_JSON.
 *
 * @param[in] write_status     The GEV status code from the last write of GIGE_REG_SAVE_JSON.
 */
void SaveVideo::do_save_json_write_complete(uint32_t write_status)
{
    // We caught our signal and we now disconnect.
    disconnect(this, &SaveVideo::save_json_write_complete, this, &SaveVideo::do_save_json_write_complete);

    // Mark the time and status.
    write_save_json_rsptime = QDateTime::currentDateTime();
    write_save_json_reqstatus = write_status;

    // We are done here.
    write_save_json_requested = false;

    // Report the final status.
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "emitting save_json_response: write_status:" << FI_comm::gev_status_to_msg(write_status);
    emit save_json_response(write_status);
}

/*!
 * @brief Slot handler for responding to an ENOMEM error condition when a FasFireLR capture hits an
 * underflow condition. Signal caught: SaveVideo::save_fflr_underflow_error emitted from the the polling
 * logic for GIGE_REG_SAVE_STATE in SaveVideo::save_state_change.
 */
void SaveVideo::process_fflr_underflow()
{
    uint32_t    camera_state;
    QString     device_name;
    uint32_t    fflr_type;
    bool        fflr_is_enabled;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    QString     progress_title  = tr("%1 Saving").arg(CameraMDI::fflr_feature_name);
    uint32_t    reg_value;
    int         result;
    bool        user_discarded_buffers;

    // If we are already processing underflow condition ignore an overlapping request.
    if (fflr_underflow_state != FFLR_UNDERFLOW_STATE::READY_TO_COMPLETE)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "fflr_underflow_state:" << fflr_underflow_state << "not READY_TO_COMPLETE?!";
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_SETUP_ERROR;
        goto exit_process_fflr_underflow;
    }

    // Sanity check.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_SETUP_ERROR;
        goto exit_process_fflr_underflow;
    }

    // What is our FasFireLR condition.
    fflr_is_enabled = my_parent->is_fflr_enabled(fflr_type);

    // If not processing FasFireLR it makes no sense to be here.
    if (!fflr_is_enabled)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Not processing FasFireLR how come we are here?!";
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_SETUP_ERROR;
        goto exit_process_fflr_underflow;
    }

    // Get the camera state.
    camera_state = my_parent->read_camera_cache(GIGE_REG_CAMERA_STATE);

    // If the camera is not in REVIEW just leave.
    if (camera_state != GIGE_CAMERA_STATE_REVIEW)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Camera not in GIGE_CAMERA_STATE_REVIEW, how come we are here?!";
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_SETUP_ERROR;
        goto exit_process_fflr_underflow;
    }

    // Has our parent received at least one image?
    if (!my_parent->review_has_received_a_frame())
    {
        // Let's give our a parent a little more time to get at least 1 frame before we process the error.
        fflr_underflow_wait_delays++;   // Increase our delay count.

        // Have we waited as long as we will wait?
        if (fflr_underflow_wait_delays < FFLR_UNDERFLOW_RECHECK_MAX)
        {
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Waiting delay for video frame count:" << fflr_underflow_wait_delays;
            // Let's give this a little more time in case it's a restart situation.
            QTimer::singleShot(FFLR_UNDERFLOW_RECHECK_DELAY, this, &SaveVideo::process_fflr_underflow);
            return;
        }
        else
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << "Can't verify review has received one frame?!";
            fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_SETUP_ERROR;
            goto exit_process_fflr_underflow;
        }
    }

    // Make the simple progress dialog that will report on the recovery progress.
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Creating the fflrRecoverDlg";
    fflrRecoverDlg = fiProgressDialog::CreateSimpleProgressDialog(progress_title, tr(""), QMessageBox::Cancel, __PRETTY_FUNCTION__);

    // If we can't create our progress dialog something is wrong.
    if (!fflrRecoverDlg)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unable to create recovery progress dialog?!";
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_SETUP_ERROR;
        goto exit_process_fflr_underflow;
    }

    // Make sure we are not showing a dearming dialog.
    if (fflrDearmDlg)
    {
        fflrDearmDlg->hide();
        fiProgressDialog::RecycleProgressDialog(fflrDearmDlg);
    }

    // Make sure we are good to go.
    if (fflr_underflow_state == FFLR_UNDERFLOW_STATE::READY_TO_COMPLETE)
    {
        // We are officially starting an underflow recovery state first step asking the camera to start one.
        fflr_underflow_state = FFLR_UNDERFLOW_STATE::COMPLETE_REQUEST;

        // Reset our fflr_underflow wait delay count.
        fflr_underflow_wait_delays = 0;
        // The camera server has run out of available capture buffers for the FFLR recording.
        // This represents an "underflow" scenario where image memory is filled faster than the
        // storage device's data rates for write operations. Determine the autosave condition.
        reg_value = my_parent->read_camera_cache(GIGE_REG_AUTOSAVE);
        fflr_underflow_dirty_count = GIGE_AUTOSAVE_TOTAL_COUNT(reg_value) - GIGE_AUTOSAVE_AVAIL_COUNT(reg_value);
        device_name = my_parent->get_fflr_device_label();

        // Customized dialog to report the underflow condition.
        FiFFLRUnderflowDlg fflr_underflow_dlg(fflr_underflow_dirty_count, device_name, my_parent);

        result = fflr_underflow_dlg.exec();
        user_discarded_buffers = fflr_underflow_dlg.userDiscardedBuffers();

        // Tell parent what user chose to do with recorded partitions.
        my_parent->set_fflr_user_discarded_parts(user_discarded_buffers);

        // A return of QDialog::Accepted means they want to recover and complete the partial recording.
        if (result == QDialog::Accepted)
        {
            QJsonObject json_save;
            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "[[[[[[ CAMERA SERVER RECOVERY OPERATION SELECTED! ]]]]]]]]] discard buffers:" << user_discarded_buffers;

            // Tell the camera server we want to save any partial recorded data, if the release and discard
            // checkbox is unchecked, which is the default. Otherwise if it is checked then the camera server
            // will only write out the metamap file and it releases any partial recorded data.
            //
            // This is accomplished by writing out the following JSON data to GIGE_REG_SAVE_JSON register.
            //
            // write_buffers:true or false
            // save_metadata:true or false
            //
            json_save.insert("write_buffers", (user_discarded_buffers ? false : true)); // If discarding we don't write buffers.
            json_save.insert("save_metadata", true);

            // Useful to have in the log file.
            {
                QLOG_DEBUG() << "++++++++++++++++++++++++++++++++++++++++";
                QLOG_DEBUG() << json_save;
                QLOG_DEBUG() << "++++++++++++++++++++++++++++++++++++++++";
            }

            // Send it on it's way.
            QJsonDocument action_doc(json_save);

            // Make the simple progress dialog that will report on the recovery progress.
            progress_title  = tr("%1 Saving").arg(CameraMDI::fflr_feature_name);

            fflrRecoverDlg->setWindowTitle(progress_title);
            QString status_str = tr("Asking %1 to complete partial recording on %2.").arg(my_parent->get_camera_name()).arg(my_parent->get_fflr_device_label());
            fflrRecoverDlg->setStatusValue(status_str);

            // We want to capture the user's cancel request if one is made.
            connect(fflrRecoverDlg.get(), &fiProgressDialog::progressCancelled, this, &SaveVideo::do_fflr_underflow_cancelled, Qt::QueuedConnection);

            // We don't have percentages yet so let's just keep the progress bar as a running one.
            fflrRecoverDlg->setBarMinMax(0, 0);
            fflrRecoverDlg->setBarValue(0);
            fflrRecoverDlg->showBar(true);

            // Show the dialog if it is not showing.
            if (!fflrRecoverDlg->isVisible())
            {
                fflrRecoverDlg->showDialogWithoutPolling();
            }

            // Make sure we update the buffer queue on the information page.
            my_parent->refresh_autosave_buffers();

            // Connect up our handler for result of the write request.
            connect(this, &SaveVideo::save_json_response, this, &SaveVideo::do_fflr_underflow_save_json_response, Qt::QueuedConnection);

            // Start the recovery process by writing our JSON save request.
            write_save_json_request(action_doc);
        }
        // A return of QDialog::Cancel means they want to discard image memory partial recording.
        else
        {
            // Make the simple progress dialog that will report on the recovery progress.
            progress_title  = tr("%1 Discarded Partial Recording").arg(CameraMDI::fflr_feature_name);

            fflrRecoverDlg->setWindowTitle(progress_title);
            QString status_str = tr("Camera %1 been asked to delete any partial recordings.").arg(my_parent->get_camera_name());
            fflrRecoverDlg->setStatusValue(status_str);
            fflrRecoverDlg->setCancelText("Ok");

            // User marks this recording as discarded and end of our underflow processing.
            fflr_underflow_state = FFLR_UNDERFLOW_STATE::DISCARDED;
            my_parent->do_fflr_underflow_discard();
        }
    }
exit_process_fflr_underflow:
    // If arrive here with set up error condition warn the user.
    if (fflr_underflow_state == FFLR_UNDERFLOW_STATE::COMPLETE_SETUP_ERROR)
    {
        QMessageBox::warning(MainWindow::GetInstance().get(), progress_title, tr("The storage device could not keep up with image recording bandwidth. Unable to set up a recovery operation at this time."), QMessageBox::Ok);
    }
    return;
}

/*!
 * @brief Starts the process of asking the camera to begin a device save operation.
 *
 * @param[in] json_save_request         The JSON record describing the camera save operation.
 *
 * @returns The state of writing the save json request.
 */
SAVE_VIDEO::DEVICE_SAVE_STATE SaveVideo::write_save_json_request(const QJsonDocument& json_save_request)
{
    QDateTime   empty_date_time;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    DEVICE_SAVE_STATE request_state = DEVICE_SAVE_STATE::SAVE_SETUP_ERROR;

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        request_state = DEVICE_SAVE_STATE::SAVE_SETUP_ERROR;
        goto exit_write_save_json;
    }

    // Are we already waiting for a GIGE_REG_SAVE_JSON write request?
    if (write_save_json_requested)
    {
        QLOG_WARN() << __PRETTY_FUNCTION__ << "Already waiting for write save request started at" << write_save_json_reqtime;
        request_state = DEVICE_SAVE_STATE::SAVE_SETUP_DENIED;
        goto exit_write_save_json;
    }

    // Mark the time of this request.
    write_save_json_requested   = true;
    write_save_json_reqtime     = QDateTime::currentDateTime();
    write_save_json_rsptime     = empty_date_time;
    write_save_json_reqstatus   = FI_COMM::GEV_STATUS_UNKNOWN;

    // We need to catch the write complete signal to keep processing the request.
    connect(this, &SaveVideo::save_json_write_complete, this, &SaveVideo::do_save_json_write_complete, Qt::QueuedConnection);

    // The request is sent and out in the world.
    request_state = DEVICE_SAVE_STATE::SAVE_REQUESTED;

    // Send our request to the camera and we are done here.
    my_parent->write_camera_buffer(GIGE_REG_SAVE_JSON, json_save_request.toJson());

exit_write_save_json:
    return request_state;
}

//--------------------------------------------------------------------
// Slot Methods
//--------------------------------------------------------------------

/*!
 * @brief The user wants to cancel the save operation.
 */
void SaveVideo::cancel_save(const QString& whoCalledMe)
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());

    QLOG_TRACE() << "[SaveVideo::cancel_save requested by " << whoCalledMe << "]";

    // Check our camera pointer.
    if (my_parent == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Tell the camera to stop saving. We'll notify the user when the camera
    // reports that it's been stopped.
#if LOG_WRITE_COMPLETE
    QLOG_DEBUG() << __PRETTY_FUNCTION__
                 << "Writing GIGE_SAVE_STATE_BUSY a ZERO to GIGE_REG_SAVE_STATE cancel the save!";
#endif
    comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_SAVE_STATE, 0);

    // Remember that we're cancelling.
    cancel_in_progress = true;

    // Are we in autosave mode?
    if (autosave_enabled)
    {
        // Disable autosave.
        autosave_stop();

        // Get the current camera state.
        uint32_t    camera_state;
        camera_state = comm_ptr->get_reg(my_parent->get_ip_addr(), GIGE_REG_CAMERA_STATE);

        // Only switch state if we're armed, triggered or in autosave.
        if ((camera_state == GIGE_CAMERA_STATE_ARMED) || (camera_state == GIGE_CAMERA_STATE_TRIGGERED) || (camera_state == GIGE_CAMERA_STATE_AUTOSAVE))
        {
            // Tell the camera to go to live mode.
            comm_ptr->start_write_reg(my_parent->get_ip_addr(), GIGE_REG_CHANGE_CAMERA_STATE, GIGE_CAMERA_STATE_LIVE_VIDEO);
        }
    }

    // Do we have a xfer progress dialog set up?  If so perform some status modifications.
    UpdateXferDialogForCancelButtonHit();
}

/*!
 * @brief There is a buffer on the to-be-freed list.
 */
void SaveVideo::free_buffer()
{
    ReleaseImageFreeList(false, true);
}

/*!
 * @brief It's time to go read the camera's @ref GIGE_REG_SAVE_STATE register.
 */
void SaveVideo::save_state_poll_timeout()
{
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Query the camera for save status.
    my_parent->start_camera_read_reg(GIGE_REG_SAVE_STATE);
}

/*!
 * @brief The raw per-frame metadata has been downloaded.
 */
void SaveVideo::raw_metadata_ready()
{
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    if (VERBOSE_SAVE_LOG)
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__;
    }

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Not interested in any more of these events right now.
    disconnect(my_parent, SIGNAL(raw_metadata_received()), this, SLOT(raw_metadata_ready()));

    // Sanity check the raw metadata we've downloaded.
    if (!my_parent->raw_metadata_ok())
    {
        QString text_msg;

        // Reset the buffer.
        my_parent->get_raw_metadata()->clear();

        // Build our error message(s).
        text_msg = tr("Raw metadata download failure: Unable to successfully transfer the entire amount of data.");

        // Notify the user.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << text_msg;
        QMessageBox::critical(my_parent, tr("FasMotion"), text_msg);
        return;
    }

    // Save it off.
    saveRawMetadata(my_parent->get_raw_metadata(), (convertRawToXml ? (const pm_metadata *)my_parent->get_partition_metadata()->constData() : nullptr),
                    my_parent->read_camera_cache(GIGE_REG_RED), my_parent->read_camera_cache(GIGE_REG_GREEN), my_parent->read_camera_cache(GIGE_REG_BLUE));

    // Not interested in any more of these events right now.
    connect(my_parent, SIGNAL(video_is_paused()), this, SLOT(video_is_paused()));

    // Remember that we're waiting for the video to pause.
    waiting_for_pause = true;

    // Since the user wants to save a video, let's pause the stream if it's not already paused.
    my_parent->pause_video_request();

    // Calculate the trigger frame start time.
    my_parent->calc_trigger_frame_start(clipStart, (clipStart + frameCount) - 1);
}

/*!
 * @brief Called to process a [save state](@ref GIGE_REG_PROGRESS) change from the camera.
 * Note:  Called when saving to a device on the camera.  This returns the last frame number saved.
 * @param frames_saved - A count of frames saved to the file.
 */
void SaveVideo::save_frame_progress(uint32_t frames_saved)
{
    FI_comm     *comm_ptr;
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    bool        as_enable;
    bool        as_overlap;
    uint32_t    autosave_settings;

    QString     saveStatusStr;
    bool        mp4Save = false;
    uint32_t    start_frame = 0;
    uint32_t    end_frame = 0;
    uint32_t    frameCountExpected = 0;
    uint32_t    progressPercent = 0;

#if LOG_GIGE_REG_PROGRESS_UPDATES
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "frames_saved:" << frames_saved
                 << "xferDialog:" << (xferDialog ? "exists" : "empty")
                 << "cancel_in_progress:" << cancel_in_progress;
#endif
    // Has the user started the cancel process?
    if (cancel_in_progress)
        return;

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Get our communications object pointer.
    comm_ptr = MainWindow::GetInstance()->get_comm();
    if (comm_ptr == nullptr)
    {
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::COMMS_PTR_SANITY_FAILURE;
        return;
    }

    // Read the current autosave settings.
    autosave_settings = comm_ptr->get_reg(my_parent->get_ip_addr(), GIGE_REG_AUTOSAVE);

    // Parse out the fields we're interested in.
    as_enable = (autosave_settings & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE;
    as_overlap = (autosave_settings & (uint32_t)GIGE_AUTOSAVE_OVERLAP) == (uint32_t)GIGE_AUTOSAVE_OVERLAP;

    // Calculate the frame count expected and build our status string for this percentage update.
    if (saveToCameraJSONData)
    {
        // Get the absolute start/end frame numbers.
        start_frame = saveToCameraJSONData->ref_frame + (uint32_t)saveToCameraJSONData->clip_start;
        end_frame = saveToCameraJSONData->ref_frame + (uint32_t)saveToCameraJSONData->clip_end;

        // The count expected is 1-based range used to calculate the percentage complete.
        frameCountExpected = end_frame - start_frame + 1;

        // Make sure that the frames_saved count is within our frameCountExpected range.
        if (frames_saved > frameCountExpected)
        {
            frames_saved = frameCountExpected;
        }

        // Calculate the percentage between 0 and 100%.
        progressPercent = (uint32_t)((float)frames_saved / (float)frameCountExpected * (float)100.0);

#if 0
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "ref_frame:" << saveToCameraJSONData->ref_frame
                     << "clip_start:" << saveToCameraJSONData->clip_start << "clip_end:" << saveToCameraJSONData->clip_end
                     << "startFrame:" << start_frame << "endFrame:" << end_frame
                     << "frames_saved:" << frames_saved
                     << "frameCountExpected:" << frameCountExpected << "progressPercent:" << progressPercent;
#endif

        // Saving MP4 video will cause a final flush which takes time and it will be
        // important to notify the user about the flush later once all frames have been
        // processed through the save operation.
        if (saveToCameraJSONData->streamType == FI_SAVER::STREAM_TYPE::MP4)
        {
            // Note that we are MP4 video save.
            mp4Save = true;
        }
    }

    // How the progress information is displayed is strictly related to the camera state.
//    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "as_enable:" << as_enable << "as_overlap;" << as_overlap << "is_grouped():" << my_parent->is_grouped();
    if (as_enable && (as_overlap || my_parent->is_grouped()))
    {
        // When in fasfire mode we do not have a local xferDialog.  We are allowing trigger events
        // to overlap and therefore we can't have an xferDialog present due to dialog modality.
        // Therefore this is "hidden".  How do you "cancel" or stop the process when there is no
        // xferDialog cancel button.  The way is through the arm button.  If in fasfire the arm
        // button will ask the user whether or not to continue.  This is where the handling should
        // occur.
        if (xferDialog)
        {
            // Hide the transfer dialog.
            xferDialog->hideDialog();

            // Were we cancelled?
            if (xferDialog->wasCanceled())
            {
                fiProgressDialog::RecycleProgressDialog(xferDialog);
            }
        }

        // Do we have a frameCountExpected?  If so calculate the percentage value.
        if (frameCountExpected)
        {
            // Update the display for the autosave fasfire mode.
            my_parent->update_save_progress(progressPercent);
        }
    }
    else
    {
        // If we are enabled, we still want the save progress to update.
        if (as_enable && frameCountExpected)
        {
            // Update the display for the autosave fasfire mode.
            my_parent->update_save_progress(progressPercent);
        }

        // Update our information for the save progression.
        if (xferDialog)
        {
            // Only put out the status if we have frameCountExpected > 0
            if (frameCountExpected > 0)
            {
                // Build up an informational string for the status line.
                saveStatusStr = tr("Saved [%1] frames of [%2]").arg(frames_saved).arg(frameCountExpected);

                // Could we be in an MP4 video final encoder flushing state?
                if (progressPercent == 100)
                {
                    if (mp4Save)
                    {
                        // Tell the user to wait for the final encoding pass to complete.
                        saveStatusStr = tr("Please wait for final encoding pass to complete.");
                    }
                }

                // Keep with the 100% theme for device saves on the progress bar
                // and only move forward on the bar if the percentage is moving forward.
                if (xferDialog->getBarValue() < (int)progressPercent)
                    xferDialog->setBarValue((int)progressPercent);
            }
            else
            {
                // Report only the frame number last saved as we don't have any context to calculate progress.
                saveStatusStr = tr("Frames saved [%1]").arg(frames_saved);
            }

            //QLOG_DEBUG() << __PRETTY_FUNCTION__ << "saveStatusStr:" << saveStatusStr << "isVisible;" << xferDialog->isVisible() << "progressPercent:" << progressPercent;

            // And show the status string in the dialog.
            xferDialog->setStatusValue(saveStatusStr);
            xferDialog->showStatusValue(xferDialog->isVisible());
        }
    }
}

/*!
 * @brief Called to process a [save state](@ref GIGE_REG_SAVE_STATE) change from the camera.
 * Note:  Called when saving to a device on the camera.  This returns the percentage complete.
 * @param new_save_state    The new state information in GIGE_REG_SAVE_STATE.
 */
void SaveVideo::save_state_change(uint32_t new_save_state)
{
    bool        as_enable;
    bool        as_overlap;
    uint32_t    autosave_settings;
    uint32_t    camera_state;
    uint32_t    save_flags;
    uint32_t    save_state_value = 0;               // What is the value of the register minus all flags?
    uint32_t    prev_save_state = last_save_state;  // Save it for logging and delta percentage calculations.
    CameraMDI   *my_parent = qobject_cast<CameraMDI *>(parent());
    uint32_t    fflr_type;
    bool        fflr_is_enabled = false;
    QString     save_state_str;

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_ERROR() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Translate our register state into a displayable state for logging.
    // Accumulate bit states and if BUSY set output to hexidecimal setting.
    // NOTE: GIGE_SAVE_STATE_BUSY is a ZERO, meaning no bits are set.
    if ((new_save_state & GIGE_SAVE_STATE_IDLE) == (uint32_t)GIGE_SAVE_STATE_IDLE)
        save_state_str += "[GIGE_SAVE_STATE_IDLE]";
    if ((new_save_state & GIGE_SAVE_STATE_ERR) == (uint32_t)GIGE_SAVE_STATE_ERR)
        save_state_str = "[GIGE_SAVE_STATE_ERR]";
    if (save_state_str.isEmpty())
        save_state_str = tr("[GIGE_SAVE_STATE_BUSY:%1]").arg(new_save_state, 8, 16, QChar('0'));

    // Save the new state.
    last_save_state = new_save_state;

    // Is this a FasFireLR save operation?
    fflr_is_enabled = my_parent->is_fflr_enabled(fflr_type);

    // Extract the flags.
    save_flags = new_save_state & GIGE_SAVE_STATE_FLAGS;

    // Extract the save value if one exists, this is either progress or an error code.
    save_state_value = (new_save_state & ~GIGE_SAVE_STATE_FLAGS);

    // Get the current camera state.
    camera_state = my_parent->read_camera_cache(GIGE_REG_CAMERA_STATE);

    // Read the current autosave settings.
    autosave_settings = my_parent->read_camera_cache(GIGE_REG_AUTOSAVE);

    // Parse out the fields we're interested in.
    as_enable = (autosave_settings & (uint32_t)GIGE_AUTOSAVE_ENABLE) == (uint32_t)GIGE_AUTOSAVE_ENABLE;
    as_overlap = (autosave_settings & (uint32_t)GIGE_AUTOSAVE_OVERLAP) == (uint32_t)GIGE_AUTOSAVE_OVERLAP;

#if LOG_SAVE_STATE_TRACING
    // If you want to distinguish these on Linux just use QLOG_WARN so they output in journalctl in yellow.
    QLOG_TRACE() << __PRETTY_FUNCTION__ << save_state_str
                 << "curr:" << Qt::showbase << Qt::hex << new_save_state
                 << "prev:" << Qt::showbase << Qt::hex << prev_save_state
                 << "val:" << save_state_value
                 << "ase:" << as_enable << "aso:" << as_overlap
                 << "camstate:" << Qt::showbase << Qt::hex << camera_state
                 << (cancel_in_progress ? "cancel in progress" : "");
#endif

    // Has anything changed?
    if (!cancel_in_progress && (new_save_state == prev_save_state))
    {
        return;
    }

    // Ticket#530: Series3 cameras should not return early if the state equal to
    // GIGE_SAVE_STATE_BUSY (which equates to a zero value in register).
    if ((new_save_state == GIGE_SAVE_STATE_BUSY) && (!my_parent->is_series3()))
    {
        return;
    }

    //--------------------------------------------------------------------------------------------
    // FIRST CHECK FOR GIGE_SAVE_STATE_ERROR
    //--------------------------------------------------------------------------------------------

    // Is there an error?
    // GIGE_SAVE_STATE_ERR      1 << 30 - hex in register [0x4000 0000] binary [0100] leftmost byte.
    // GIGE_SAVE_STATE_IDLE     1 << 31 - hex in register [0x8000 0000] binary [1000] leftmost byte.
    // GIGE_SAVE_STATE_FLAGS    F << 28 - hex in register [0xF000 0000] binary [1111] leftmost byte.
    if ((save_flags & GIGE_SAVE_STATE_ERR) == (uint32_t)GIGE_SAVE_STATE_ERR)
    {
        QString msg_str;

        // Clear the error code right away.
        // This is the official way to notify the camera_server to clear the error code condition.
        // It was validated by a camera_server code check.
        // NOTE: If the camera server's imgSaveWorker is still running this has no effect...
        // It's not until we see the idle bit come in that we can be sure imgSaveWorker has stopped.
#if LOG_WRITE_COMPLETE
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_SAVE_STATE_ERR "
                     << "last:" << Qt::showbase << Qt::hex << prev_save_state
                     << "Writing GIGE_REG_SAVE_STATE with GIGE_SAVE_STATE_ERR to clear error bit!";
#endif
        my_parent->write_camera_register(GIGE_REG_SAVE_STATE, GIGE_SAVE_STATE_ERR);

        // Are we FasFireLR?
        if (fflr_is_enabled)
        {
            // Test for an ENOMEM error situation. This should be for a FasFireLR operation underflow.
            // This may come in with just an error bit, so check for idle bit being set before emitting the
            // underflow condition to our handler. When idle bit is set the camera server's imgSaveWorker
            // thread has exited and it should be safe to clear the error bit and ask for underflow recovery.
            if (save_state_value == ENOMEM)
            {
                // Is this our first indication of an error?
                if (fflr_underflow_state == FFLR_UNDERFLOW_STATE::NONE)
                {
                    // This should be the first indication of the underflow error.
                    // We will start the process but indicate we need to see an idle come through.
                    fflr_underflow_state = FFLR_UNDERFLOW_STATE::WAIT_FOR_IDLE;
                }
#if 0
// This really shouldn't be happening because if it does the error was not cleared yet...
// Let's just let it report the Haven't sent recovery request yet message below.
                else if (fflr_underflow_state == FFLR_UNDERFLOW_STATE::WAIT_FOR_IDLE)
                {
                    // If we get idle notification here process even though error bit was still set
                    // go ahead and push state forward to start the recovery.
                    if ((save_flags & GIGE_SAVE_STATE_IDLE) == (uint32_t)GIGE_SAVE_STATE_IDLE)
                    {
                        // Mark the state as ready to complete and start the underflow process.
                        fflr_underflow_state = FFLR_UNDERFLOW_STATE::READY_TO_COMPLETE;
                        emit save_fflr_underflow_error();
                    }
                }
#endif
                else
                {
                    if (fflr_underflow_state <= FFLR_UNDERFLOW_STATE::COMPLETE_REQUEST)
                    {
                        QLOG_WARN() << __PRETTY_FUNCTION__ << "Haven't sent recovery request yet: fflr_underflow_state:" << fflr_underflow_state;
                    }
                    else
                    {
                        QLOG_WARN() << __PRETTY_FUNCTION__ << "Unexpected error (Maybe error bit is not cleared?!): fflr_underflow_state:" << fflr_underflow_state;
                    }
                }
                return;
            }
        }

        // The camera is in error condition we don't recover from. Shutdown any progress display.
        cleanup_save();

        // Was there an "error" or is the camera just telling us that the save was canceled.
        if (!error_is_logged && (save_state_value != EINTR))
        {
            // Remember that we've logged an error.
            error_is_logged = true;

            // Let the user know about the problem.
            msg_str = tr("The camera reported an error saving the video to it's media. The error code is %1 (%2)").arg(save_state_value).arg(QString::fromUtf8(strerror((int)save_state_value)));
            QTimer::singleShot(0, this, [ = ]() { QMessageBox::warning(my_parent, tr("Camera Error"), msg_str); });
        }

        return; // We are done processing the GIGE_REG_SAVE_STATE register for GIGE_SAVE_STATE_ERR.
    }

    //--------------------------------------------------------------------------------------------
    // IF MAKE IT HERE THEN CHECK FOR GIGE_SAVE_STATE_IDLE
    //--------------------------------------------------------------------------------------------

    // GIGE_SAVE_STATE_IDLE should indicate there is no imgSaveWorker on the camera_server and therefore
    // we should be transitioning to non-save state and cleaning up any cancel state.
    if ((save_flags & GIGE_SAVE_STATE_IDLE) == (uint32_t)GIGE_SAVE_STATE_IDLE)
    {
        // Special processing for FasFire LR capture mode.
        if (fflr_is_enabled)
        {
            // In order to clear any error condition on the camera by writing the clear request (GIGE_SAVE_STATE_ERR)
            // to GIGE_REG_SAVE_STATE there must be no imgSaveWorker threads running on the camera server. If there
            // are then it will have no effect. The only time we can be assured of this is when GIGE_REG_SAVE_STATE goes
            // GIGE_SAVE_STATE_IDLE. That is the reason we are testing the prev_save_state now that we are idle and the
            // reason for this special processing.
            if ((prev_save_state & GIGE_SAVE_STATE_ERR) == (uint32_t)GIGE_SAVE_STATE_ERR)
            {
#if LOG_WRITE_COMPLETE
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "GIGE_SAVE_STATE_IDLE "
                             << "last:" << Qt::showbase << Qt::hex << prev_save_state
                             << "Writing GIGE_REG_SAVE_STATE with GIGE_SAVE_STATE_ERR to clear error bit!";
#endif
                // Clear the error bit now it should take effect because we are idle.
                my_parent->write_camera_register(GIGE_REG_SAVE_STATE, GIGE_SAVE_STATE_ERR);
            }

            // If we were waiting for the idle transition we can now start the
            // FasFireLR underflow recovery process.
            if (fflr_underflow_state == FFLR_UNDERFLOW_STATE::WAIT_FOR_IDLE)
            {
                // Mark the state as ready to complete and start the underflow process.
                fflr_underflow_state = FFLR_UNDERFLOW_STATE::READY_TO_COMPLETE;

                emit save_fflr_underflow_error();
                return;
            }
        }

        // FasFireLR mode might be in the process of saving multiple CAP files. It will shutdown the progress
        // display once all CAP files have been saved.
        if (!fflr_is_enabled || (streamType != FI_SAVER::FI_CAPSTK))
        {
            cleanup_save();
        }

        // Update the display.
        my_parent->update_save_progress(0);

        // Update the menus incase something has changed.
        MainWindow::GetInstance()->reflect_camera_state();

        // Are we at the end of a cancel?
        if (cancel_in_progress)
        {
            // Reset our flag.
            cancel_in_progress = false;

            // Do we have a partition sitting unsaved in memory?
            uint32_t part_state = my_parent->read_camera_cache(GIGE_REG_PART_STATE);

            // Can only go into review if we have a partition to review.
            if ((camera_state != GIGE_CAMERA_STATE_REVIEW) && (part_state == GIGE_PART_STATE_COMPLETE))
            {
                // Switch us over to review.
                my_parent->do_camera_review();
            }
            else
            {
                // We can't go into review, so just turn on streaming.
                my_parent->resume_video_request();
            }
        }

        return; // We are done processing the GIGE_REG_SAVE_STATE register for GIGE_SAVE_STATE_IDLE.
    }

    //--------------------------------------------------------------------------------------------
    // IF MAKE IT HERE THEN WE SHOULD HAVE A PROGRESS VALUE FOR THE CAMERA DEVICE SAVE.
    //--------------------------------------------------------------------------------------------

#if LOG_SAVE_STATE_TRACING
    QLOG_DEBUG() << __PRETTY_FUNCTION__ << "PROGRESS UPDATE progress_value:" << save_state_value << "xferDialog.is_visible():" << ((xferDialog != nullptr) ? (xferDialog->isVisible() ? "Visible" : "Not Visible") : "xferDialog was null")
                 << "as_enable:" << as_enable << "as_overlap:" << as_overlap;
#endif

    // If FasFireLR is enabled and the camera is in a review state we are probably
    // in the underflow recovery save operation.
    if (fflr_is_enabled)
    {
        // Keep showing progress on our buffer recovery save operation.
        my_parent->update_save_progress(save_state_value);

        // Update our buffer display in the camera information panel...
        my_parent->refresh_autosave_buffers();

        // Are we in any kind of an underflow condition?
        if (is_fflr_underflow_recovering())
        {
            if (fflrRecoverDlg)
            {
                // Update the overall progress on the recovery save operation for all buffers.
                if (save_state_value > prev_save_state)
                {
                    uint32_t delta_pct = save_state_value - prev_save_state;
                    uint32_t value_update = fflrRecoverDlg->getBarValue() + delta_pct;
                    fflrRecoverDlg->setBarValue(value_update);
                }
            }
        }
        else if (active_device_save == DEVICE_SAVE_STATE::SAVE_DEARM_AND_KEEP)
        {
            // Show the dearming progress dialog.
            if (fflrDearmDlg)
            {
                // Give the user some kind notification that we must be in process of saving.
                QString status_msg;
                uint32_t dirty_count, total_count;
                get_autosave_status_msg(status_msg, total_count, dirty_count);

                fflrDearmDlg->setBarMinMax(0, total_count);
                fflrDearmDlg->setBarValue((total_count - dirty_count));
                fflrDearmDlg->setStatusIcon(QStyle::SP_MessageBoxInformation);
                fflrDearmDlg->setStatusValue(status_msg);

                // Make sure the dialog is visible.
                if (!fflrDearmDlg->isVisible())
                    fflrDearmDlg->showDialogWithoutPolling();
            }
        }
        else
        {
            if (fflrDearmDlg)
            {
                fflrDearmDlg->hide();
                fiProgressDialog::RecycleProgressDialog(fflrDearmDlg);
            }
        }

        // For FasFireLR don't let the stuff below happen because it can cause bogus
        // XferProgressDialogs to pop-up when they don't need to pop-up!
        return;
    }

    // Catch bogus save status events. If the progress is at 100%, but we don't have
    // a progress dialog up, then this is probably something we want to ignore.
    if (!xferDialog && (save_state_value == 100))
    {
        // Ignore the event.
        return;
    }

    // Looks like we are receiving percentage save updates from camera for device save. Make sure
    // the polling timer for GIGE_REG_SAVE_STATE is running!
    if (save_state_poll_timer == nullptr)
    {
        // Start up our polling timer. This happens after the command to start, so that there is less
        // chance of us reading back an "idle" status and thinking the save has completed before it's
        // even begun.
        save_state_poll_timer = new QTimer(this);
        connect(save_state_poll_timer, SIGNAL(timeout()), this, SLOT(save_state_poll_timeout()));
        save_state_poll_timer->start(GIGE_REG_SAVE_STATE_POLLING_RATE);
    }

    // How the progress information is displayed is strictly related to the camera state.
    // If the camera is in the Live state when we are still saving, then that is the end state for the FasFireLR
    // operation. To keep the user from doing something, we will want to display the xferDialog.
    if (as_enable && (as_overlap || my_parent->is_grouped()) && (camera_state != GIGE_CAMERA_STATE_LIVE_VIDEO))
    {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Hide xferDialog and call update_save_progress";

        // When in fasfire mode we do not have a local xferDialog.  We are allowing trigger events
        // to overlap and therefore we can't have an xferDialog present due to dialog modality.
        // Therefore this is "hidden".  How do you "cancel" or stop the process when there is no
        // xferDialog cancel button.  The way is through the arm button.  If in fasfire the arm
        // button will ask the user whether or not to continue.  This is where the handling should
        // occur.
        if (xferDialog)
        {
            // Hide the transfer dialog.
            xferDialog->hideDialog();

            // Were we cancelled?
            if (xferDialog->wasCanceled())
            {
                fiProgressDialog::RecycleProgressDialog(xferDialog);
            }
        }

        // Update the display.
        my_parent->update_save_progress(save_state_value);

        return;     // We are done here.
    }

    // If there is a progress value and we make it to this point, process it.
    if (save_state_value > 0)
    {
//            QLOG_DEBUG() << __PRETTY_FUNCTION__ << "Camera saving progress:" << save_state_value << "xferDialog:" << (xferDialog ? "[exists]" : "[to be created]");

        // If we are not in the middle of a copy then we need to be responsible
        // for showing progress.
        if (calledByCopyToPC)
            return; // We are done here.

        // If we are enabled, we still want the save progress to update.
        if (as_enable && save_state_value)
        {
            // Update the display for the autosave fasfire mode.
            my_parent->update_save_progress(save_state_value);
        }

        // If there is no xfer progress dialog up we need to create one. This is likely
        // because we are connecting to a camera that is in the middle of a save operation.
        if (!xferDialog)
        {
            // Are we autosaving?
            if (as_enable)
            {
                QLOG_DEBUG() << __PRETTY_FUNCTION__ << "PROGRESS A: Setting xferDialog bar value:" << save_state_value;
                // Do we know what it is we're autosaving? A good indicator is
                // that the clip_start is not the same as the clip_end. The structure
                // may not be initialized if we are connecting to a camera that's already
                // in the middle of an autosave session.
                if (as_save_info.clip_start == as_save_info.clip_end)
                {
                    // Initialize some of these settings prior to creating the xferDialog object.
                    as_save_info.clip_start = my_parent->read_camera_cache(GIGE_REG_SAVE_START_NUM);
                    as_save_info.clip_end = my_parent->read_camera_cache(GIGE_REG_SAVE_END_NUM);
                    as_save_info.dest_media = my_parent->read_camera_cache(GIGE_REG_IMAGE_DESTINATION);
                    as_save_info.streamType = FiSaver::streamFormatToStreamType(my_parent->read_camera_cache(GIGE_REG_STREAM_FORMAT));
                }

                // Create our dialog.
                PrepXferDialogFromSaveParams(as_save_info, localSaveToPath, __PRETTY_FUNCTION__);

                // Is there a slot connected to our signal?
                //
                // It is very likely that there is no slot connected, because of the reason we are
                // here. We are here, most likely, because we are connecting to a camera that is already
                // in the midst of a save operation.
                if (!isSignalConnected(QMetaMethod::fromSignal(&SaveVideo::userCancelledSaveSignal)))
                {
                    // WARNING: We are going to make an assumption about the save, which might not be true.
                    //          It is LIKELY (but not guaranteed) that the camera is saving to a device local
                    //          to the camera. In this case, the controlling entity is the CameraMDI objec.
                    //
                    // Hook the CameraMDI object to the cancel signal.
                    connect(this, &SaveVideo::userCancelledSaveSignal, my_parent, &CameraMDI::cancel_save, Qt::QueuedConnection);
                }
            }
            else
            {
                QLOG_TRACE() << __PRETTY_FUNCTION__ << save_state_str << "PROGRESS B: Setting xferDialog bar value:" << save_state_value;
                // Create a task oriented transfer dialog.
                xferDialog = fiProgressDialog::CreateXferProgressDialog(FI_PROGRESS::XFER_TYPES::XFER_TYPE_SAVE,
                                                                        my_parent->get_camera_name(),
                                                                        __PRETTY_FUNCTION__);

                // Did we create the transfer progress dialog?
                if (xferDialog)
                {
                    // This is not on the PC - don't have source/destination showing.
                    xferDialog->showXferSrc(false);
                    xferDialog->showXferDest(false);

                    // Set up the save type information.
                    // We need to find out what the camera is using as a format for the save. We'll
                    // update the progress dialog when we find out.
                    my_parent->start_camera_read_reg(GIGE_REG_IMAGE_DESTINATION);
                    xferDialog->showXferType(false);

                    // Set up the status.
                    xferDialog->setStatusValue(my_parent->get_camera_name() + tr(" - Saving Files"));
                    xferDialog->showStatusValue(true);

                    xferDialog->setBarMinMax(0, (int)frameCount);
                    xferDialog->setBarValue(0);
                    xferDialog->showBar(true);

                    xferDialog->showCancel(true);
                }
            }
        }

        // We definitely should have a progress dialog at this point.
        // Make sure it is visible and increment the progress bar.
        if (xferDialog)
        {
            QLOG_TRACE() << __PRETTY_FUNCTION__ << save_state_str << "PROGRESS C: Setting xferDialog bar value:" << save_state_value;

            // Is the dialog hidden?
            if (!xferDialog->isVisible())
            {
                // Make it visible.
                xferDialog->showDialogWithoutPolling();
            }

            // Update our display.
            xferDialog->setBarValue((int)save_state_value);
        }
    }
}

/*!
 * @brief Simply update the progress bar.
 */
void SaveVideo::update_progress()
{
    if (xferDialog)
    {
        xferDialog->incrementBar();
    }
    else
    {
        QLOG_DEBUG() << __PRETTY_FUNCTION__ << "No xferDialog was it calledByCopyToPC?:" << calledByCopyToPC;
    }
}

/*!
 * @brief Slot handler for timer notification of transfer progress update.
 */
void SaveVideo::timeForVideoProgressPollUpdate()
{
    uint32_t frameCount = 0;
    uint32_t framesAdded = 0;
    uint32_t framesEncoded = 0;
    QString strMsg;
    bool isStreamVideo = false;
    bool streamIsOk = false;
    bool streamIsComplete = false;

    // Make sure that there is a transfer dialog!
    if (!xferDialog)
        return;

    // Pause the polling process on the transfer dialog.
    xferDialog->setXferToPollWait();

    isStreamVideo = getVideoStreamSaveStatus(strMsg, frameCount, framesAdded, framesEncoded, streamIsOk, streamIsComplete);

    // Ticket #271 - if stream writer is in error and we do not pause the
    // stream we could end up causing a crash in the system because the user
    // is seeing this progress dialog with error condition but the camera's
    // streamer is still trying to push images and possibly timing out in the
    // process.  When we know there is an error we need to stop the input stream.
    //
    // If anything is wrong with the stream try to pause the streamer.
    if (isStreamVideo && !streamIsOk)
    {
        CameraMDI *my_parent = qobject_cast<CameraMDI *>(parent());

        // Sanity check.
        if (my_parent == nullptr)
        {
            QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        }
    }

#if 0
    QLOG_DEBUG() << "POLL_LAST_FRAME:  Back save_video->getVideoStreamSaveStatus isStreamVideo: "
                 << isStreamVideo << "framesAdded: " << framesAdded << "framesEncoded: " << framesEncoded
                 << "frameCount:" << frameCount << "streamIsOk: " << streamIsOk << "streamComplete:" << streamComplete;
#endif

    if (isStreamVideo)
    {
        if (xferDialog->isVisible())
        {
            // This should have the proper information to add to the output stream.
            // We do not update the "source frame" here because framesAdded is a positive
            // frame count not a relative frame number.  This field is updated in the save
            // process when the image is received - a calculation occurs based on the frame
            // image's sequence number relative to the trigger frame and presented to the user
            // in that format because that is what they "see" when they are looking at the clip
            // information in the camera playback window.
            //xferDialog->setXferSrcValue(QString::number(framesAdded));
            if (!streamIsOk)
            {
                xferDialog->setStatusIcon(QStyle::SP_MessageBoxWarning);
                xferDialog->showStatusIcon(true);
                xferDialog->setStatusValue(strMsg);
            }
            else
            {
                // Retry pass puts up a different message in this area.
                if (!retryInProgress)
                {
                    xferDialog->showStatusIcon(!streamIsOk);
                    xferDialog->setStatusValue(strMsg);
                    xferDialog->setBarMinMax(0, (int)frameCount);
                    xferDialog->setBarValue((int)framesEncoded);
                    xferDialog->showStatusValue(true);
                }
            }
        }
#if LOG_VERBOSE_INFO
        // If we are at the end of this video save pop out a message.
        if (framesEncoded == frameCount)
        {
            QLOG_TRACE() << "PROGRESS_POLL:  save_video->getVideoStreamSaveStatus isStreamVideo: "
                         << isStreamVideo << "framesAdded: " << framesAdded
                         << "framesEncoded: " << framesEncoded
                         << "streamIsClosed: " << streamIsClosed
                         << "streamIsOk: " << streamIsOk
                         << "streamIsComplete: " << streamIsComplete
                         << "statusStr:" << strMsg;
        }
#endif
        // If stream is ok and not quite completed keep polling.
        if (streamIsOk && !streamIsComplete)
        {
            xferDialog->setXferToPollAgain();
        }
    }
}

/*!
 * @brief Slot handler for timer notification of transfer progress update.
 */
void SaveVideo::timeForStackProgressFrameUpdate(uint32_t bufferSeqNbr)
{
    if (xferDialog)
    {
        xferDialog->incrementBar();
    }
    if (!copy_to_pc_invoked())
    {
        // What is our actual save status?
        QString infoStr;
        uint32_t frameCount;
        uint32_t framesAdded;
        uint32_t framesSaved;
        bool streamIsOk;
        bool streamIsComplete;

        getSaveStatus(infoStr, frameCount, framesAdded, framesSaved, streamIsOk, streamIsComplete);

        // Is the stream completly encoded?
        if (framesSaved >= frameCount)
        {
            // If any error output it to log.
            if (!streamIsOk)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__ << "Saver reports error: " << infoStr << "bufferSeqNbr:" << bufferSeqNbr;
            }
            // There are no encoders etc.  Frame counts should satisfy completeness bug if discrepency report it.
            if (!streamIsComplete)
            {
                QLOG_ERROR() << __PRETTY_FUNCTION__
                             << "Saver framesSaved == frameCount but reporting it is not complete.  infoStr:"
                             << infoStr << "bufferSeqNbr:" << bufferSeqNbr;
            }

            // NOTES:
            // May need to initiate the saving_finished processing - but it's not straight forward and requires
            // some testing to determine what to do.

            // The value in emittedSaveComplete is only set true when the worker thread has emitted SaveVideo::save_complete
            // signal.  It is the last thing that worker thread does when it is shutting down itself based on the initiation
            // by SaveVideo::saving_finished.

            // Therefore if we get here due to the release of an image buffer (which could happen in the process of the worker
            // thread shutting down and it clears out its queues) then extra tests will ensure whether or not we need to call
            // SaveVideo::saving_finished or not.

            // Also remember that we are only arriving here when SaveVideo has set up the connection between the
            // SaveVideo::saverBufferReleaseSignal to this function when the SaveVideo's ProgressXferDialog has
            // been set up for an image stack save operation.  There are also external ways in which SaveVideo::saving_finished
            // is called by CameraMDI.  But if control ends up here it is a fully encapsulated save operation managed by
            // SaveVideo itself.

            if (!emittedSaveComplete)
            {
                // If the saving_finished process has already run and need to ignore this signal received
                // for this image buffer release.  The worker thread may be in the process of completing the
                // save operation and is already shutting down.
                QLOG_TRACE() << __PRETTY_FUNCTION__ << "[" << (savingHasFinished ? "!savingHasFinished" : "savingHasFinished")
                             << "] and !emittedSaveComplete.  Signalled by bufferSeqNbr:" << bufferSeqNbr
                             << "stream_save_buffer_count" << streamSaveBufferCount;

                if (!savingHasFinished)
                {
                    // Start the process of telling the worker thread the save operation is complete.
                    saving_finished(__PRETTY_FUNCTION__);
                }
            }
        }
    }
}

/*!
 * @brief The user has hit the cancel save button on the transfer progress dialog.
 */
void SaveVideo::userHasCancelled()
{
    QLOG_TRACE() << __PRETTY_FUNCTION__ << "User has cancelled the progress dialog.";
    UpdateXferDialogForCancelButtonHit();

    // Right now there is no action other than to emit the signal.
    emit userCancelledSaveSignal();
}

/*!
 * @brief Get notified once we've paused the video stream.
 */
void SaveVideo::video_is_paused()
{
    CameraMDI       *my_parent = qobject_cast<CameraMDI *>(parent());

    // Sanity check.
    if (my_parent == nullptr)
    {
        // This could happen on shutdown.
        QLOG_WARN() << __PRETTY_FUNCTION__ << MainWindow::CAMERA_PTR_SANITY_FAILURE;
        return;
    }

    // Not interested in any more of these events right now.
    disconnect(my_parent, SIGNAL(video_is_paused()), this, SLOT(video_is_paused()));

    // Ignore this if we're not waiting for the video to pause.
    if (!waiting_for_pause)
        return;

    // Done waiting.
    waiting_for_pause = false;

    // Start downloading the images.
    start(__PRETTY_FUNCTION__);
}

/*!
 * @brief Get notified when the worker thread exits.
 */
void SaveVideo::do_worker_thread_exit()
{
    // Resume any discovery polling that may have been paused while we were processing data from the camera.
    MainWindow::GetInstance()->resume_discovery();
    // Clear any status information that was presented.
    MainWindow::GetInstance()->set_statusbar_message("");
}

